
html, body {
    height: 100%;
}

body {
    margin: auto;
    font-family: monospace;
}

#main {
    display: grid;
    grid-template-columns: 1fr 10px 800px 10px 1fr;
    height: 100%;
}

.gutter.gutter-horizontal:hover {
    transition: background-color 0.5s ease;
    background: #f0f0f0;
}

.gutter.gutter-horizontal {
    transition: background-color 0.5s ease;
    cursor: col-resize;
    height: 100%;
}

.snek-interactor {
    display: flex;
    flex-direction: column;
    height: 100%;
    overflow: scroll;
}

.snek-outer-pane {
    flex-grow: 1;
    padding: 5px;
    border-bottom: 3px solid #999;
    overflow: scroll;

    /* This keeps it scrolled to the bottom. */
    display: flex;
    flex-direction: column-reverse;
}

.snek-pin-pane {
    display: flex;
    flex-direction: column;
    padding: 5px;
    box-sizing: border-box;
    overflow: scroll;
    height: 100%;
}

.snek-pin-pane > div {
    padding: 5px;
}

.snek-input-box {
    position: relative;
    display: flex;
    flex-direction: row;
    padding: 5px;
}

.snek-input {
    flex: 1;
}

.snek-input-mode {
    margin-right: 5px;
    margin-top: 1px;
}

.snek-input-mode-python {
    color: #bbb;
    margin-right: 5px;
    margin-top: 1px;
}

.snek-input-mode-debug {
    color: #f84;
    margin-right: 5px;
    margin-top: 1px;
}

.snek-nav {
    width: 100%;
}

.snek-status-bar {
    height: 1.5em;
    min-height: 1.5em;
    text-align: right;
    color: #888;
    cursor: pointer;
}

.snek-status-normal {
    transition: color 2s ease;
    color: #888;
}

.snek-status-error {
    color: red;
}

.snek-status-flash-error {
    background: #fbb;
}

.snek-line {
    display: flex;
    flex-direction: row;
    margin-top: 5px;
    margin-bottom: 5px;
}

.snek-result {
    flex-grow: 1;
}

.snek-gutter {
    min-height: 1em;
}


.snek-line.snek-t-echo .ace_content {
    background: #f8f8f8;
}

.snek-line.snek-t-echo {
    background: #f8f8f8;
    padding-top: 5px;
    padding-bottom: 5px;
}

.snek-gutter.snek-t-echo {
    /* background: white; */
    width: 20px;
    border-right: 5px solid transparent;
    margin-right: 10px;
}

.snek-gutter.snek-t-expression {
    /* background: #dcdcdc; */
    border-right: 5px solid #88f;
    width: 20px;
    margin-right: 10px;
}

.snek-gutter.snek-t-exception {
    /* background: #dcdcdc; */
    border-right: 5px solid red;
    width: 20px;
    margin-right: 10px;
}

.snek-gutter.snek-t-plain {
    border-right: 5px solid transparent;
    width: 20px;
    margin-right: 10px;
}

.snek-line.snek-t-info {
    background: #ffd;
    padding-top: 5px;
    padding-bottom: 5px;
}

.snek-gutter.snek-t-info {
    border-right: 5px solid orange;
    width: 20px;
    margin-right: 10px;
}

.snek-gutter.snek-t-print {
    border-right: 5px solid transparent;
    width: 20px;
    margin-right: 10px;
    min-height: 0;
}

.snek-line.snek-t-print {
    margin: 0px;
    padding: 0px;
}

.snek-print-sequence {
    display: flex;
    flex-direction: row;
    place-items: center;
}

.snek-print-sequence > * {
    padding-left: 5px;
    padding-right: 5px;
}

.snek-pinbutton:hover {
    cursor: pointer;
}

.snek-block-type {
    margin-left: 5px;
    color: #fa0;
}

.snek-clsname {
    color: #aaa;
    text-align: right;
}

.snek-clsname-principal {
    color: #777;
    text-align: right;
}

.snek-collapsible {
    display: flex;
    flex-direction: column;
    justify-items: stretch;
}

.snek-collapsible-title {
    cursor: pointer;
}

.snek-title-row {
    background-color: #ddd;
    display: flex;
    flex-direction: row;
}

.snek-title-row > *:first-child {
    flex-grow: 1;
}

.snek-title-row.snek-exception-lib {
    color: #888;
}

.hrepr-error {
    width: 100%;
    box-sizing: border-box;
}

.hrepr-error-message {
    flex-grow: 1;
    background: white;
    padding-left: 5px;
}

.snek-bedit-hl {
    background: #fa5;
}

.snek-pin-pane .snek-bedit {
    width: 600px;
}

.snek-bedit {
    width: calc(100% - 10px);
    border: 5px solid black;
    border-bottom: 2px solid;
    box-sizing: border-box;
}

.snek-bedit-editor {
    background: white;
    border: 5px solid white;
}

.snek-bedit-saved { border-color: green; }
.snek-bedit-live { border-color: #8a0; }
.snek-bedit-dirty { border-color: #f60; }
.snek-bedit-error { border-color: red; }

.snek-bedit-saved .snek-bedit-status { background-color: green; }
.snek-bedit-live .snek-bedit-status { background-color: #8a0; }
.snek-bedit-dirty .snek-bedit-status { background-color: #f60; }
.snek-bedit-error .snek-bedit-status { background-color: red; }

.snek-bedit-status {
    display: flex;
    flex-direction: row;
    color: white;
    font-size: 10px;
}

.snek-bedit-filename {
    flex: 1;
}

.snek-bedit-state {
    margin-left: 20px;
}

.snek-hidden {
    display: none;
}

.snek-print-str {
    white-space: pre-wrap;
}

.snek-bedit-readonly {
    min-width: 500px;
    width: 100%;
    border: 1px solid #888;
    box-sizing: border-box;
}

.snek-bedit-readonly > .snek-bedit-status {
    color: #888;
}

.hreprk-classmethod > .hrepr-defn-key,
.hreprk-staticmethod > .hrepr-defn-key,
.hreprk-property > .hrepr-defn-key {
	color: #080;
}

.snek-docstring {
    max-height: 400px;
    overflow: scroll;
}

.hrepr-title .hrepr-defn-key {
    color: #ddd !important;
}

.snek-interpreter-in {
    font-weight: bold;
    color: purple;
}


.snek-popup-nav {
    display: flex;
    flex-direction: column-reverse;
    max-height: 60vh;
    overflow: scroll;
    background: white;
    box-shadow: 5px 5px 10px #bbb;
    border: 1px solid #ddd;
    position: absolute;
    bottom: calc(100% + 10px);
    white-space: pre;
    cursor: pointer;
}

.snek-popup-nav-history {
    width: calc(100% - 20px);
}

.snek-popup-nav-entry {
    border-left: 1px solid white;
    padding-left: 5px;
    padding-right: 5px;
    padding-top: 1px;
    padding-bottom: 1px;
}

.snek-popup-nav-entry:hover {
    border-left: 1px solid green;
    /* background: rgb(244, 244, 232); */
}

.snek-popup-nav-cursor {
    background: beige;
}

.snek-shellout {
    white-space: pre;
}

.snek-input-mode {
    cursor: pointer;
}
