/*!-----------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * monaco-typescript version: 4.2.0(82635e7161723da59d38ab2b1d9c1cc5a8c9051a)
 * Released under the MIT license
 * https://github.com/Microsoft/monaco-typescript/blob/master/LICENSE.md
 *-----------------------------------------------------------------------------*/
/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
var __spreadArrays=this&&this.__spreadArrays||function(){for(var e=0,n=0,t=arguments.length;n<t;n++)e+=arguments[n].length;var r=Array(e),a=0;for(n=0;n<t;n++)for(var i=arguments[n],o=0,s=i.length;o<s;o++,a++)r[a]=i[o];return r},__assign=this&&this.__assign||function(){return(__assign=Object.assign||function(e){for(var n,t=1,r=arguments.length;t<r;t++)for(var a in n=arguments[t])Object.prototype.hasOwnProperty.call(n,a)&&(e[a]=n[a]);return e}).apply(this,arguments)},__makeTemplateObject=this&&this.__makeTemplateObject||function(e,n){return Object.defineProperty?Object.defineProperty(e,"raw",{value:n}):e.raw=n,e},__generator=this&&this.__generator||function(e,n){var t,r,a,i,o={label:0,sent:function(){if(1&a[0])throw a[1];return a[1]},trys:[],ops:[]};return i={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function s(i){return function(s){return function(i){if(t)throw new TypeError("Generator is already executing.");for(;o;)try{if(t=1,r&&(a=2&i[0]?r.return:i[0]?r.throw||((a=r.return)&&a.call(r),0):r.next)&&!(a=a.call(r,i[1])).done)return a;switch(r=0,a&&(i=[2&i[0],a.value]),i[0]){case 0:case 1:a=i;break;case 4:return o.label++,{value:i[1],done:!1};case 5:o.label++,r=i[1],i=[0];continue;case 7:i=o.ops.pop(),o.trys.pop();continue;default:if(!(a=o.trys,(a=a.length>0&&a[a.length-1])||6!==i[0]&&2!==i[0])){o=0;continue}if(3===i[0]&&(!a||i[1]>a[0]&&i[1]<a[3])){o.label=i[1];break}if(6===i[0]&&o.label<a[1]){o.label=a[1],a=i;break}if(a&&o.label<a[2]){o.label=a[2],o.ops.push(i);break}a[2]&&o.ops.pop(),o.trys.pop();continue}i=n.call(e,o)}catch(e){i=[6,e],r=0}finally{t=a=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,s])}}},__rest=this&&this.__rest||function(e,n){var t={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&n.indexOf(r)<0&&(t[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var a=0;for(r=Object.getOwnPropertySymbols(e);a<r.length;a++)n.indexOf(r[a])<0&&Object.prototype.propertyIsEnumerable.call(e,r[a])&&(t[r[a]]=e[r[a]])}return t},__extends=this&&this.__extends||function(){var e=function(n,t){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,n){e.__proto__=n}||function(e,n){for(var t in n)Object.prototype.hasOwnProperty.call(n,t)&&(e[t]=n[t])})(n,t)};return function(n,t){function r(){this.constructor=n}e(n,t),n.prototype=null===t?Object.create(t):(r.prototype=t.prototype,new r)}}();!function(e){function n(){var e={};return e.prev=e,{head:e,tail:e,size:0}}function t(e,n){return e===n||e!=e&&n!=n}function r(e){var n=e.prev;if(!n||n===e)throw new Error("Illegal state");return n}function a(e){for(;e;){var n=!e.prev;if(e=e.next,!n)return e}}function i(e,n){for(var a=e.tail;a!==e.head;a=r(a))if(t(a.key,n))return a}function o(e,n,t){var r=i(e,n);if(!r){var a=function(e,n){return{key:e,value:n,next:void 0,prev:void 0}}(n,t);return a.prev=e.tail,e.tail.next=a,e.tail=a,e.size++,a}r.value=t}function s(e,n){for(var a=e.tail;a!==e.head;a=r(a)){if(void 0===a.prev)throw new Error("Illegal state");if(t(a.key,n)){if(a.next)a.next.prev=a.prev;else{if(e.tail!==a)throw new Error("Illegal state");e.tail=a.prev}return a.prev.next=a.next,a.next=a.prev,a.prev=void 0,e.size--,a}}}function c(e){for(var n=e.tail;n!==e.head;){var t=r(n);n.next=e.head,n.prev=void 0,n=t}e.head.next=void 0,e.tail=e.head,e.size=0}function l(e,n){for(var t=e.head;t;)(t=a(t))&&n(t.value,t.key)}function u(e,n){if(e)for(var t=e.next();!t.done;t=e.next())n(t.value)}function d(e,n){return{current:e.head,selector:n}}function p(e){return e.current=a(e.current),e.current?{value:e.selector(e.current.key,e.current.value),done:!1}:{value:void 0,done:!0}}!function(e){e.createMapShim=function(e){var t=function(){function e(e,n){this._data=d(e,n)}return e.prototype.next=function(){return p(this._data)},e}();return function(){function r(t){var r=this;this._mapData=n(),u(e(t),(function(e){var n=e[0],t=e[1];return r.set(n,t)}))}return Object.defineProperty(r.prototype,"size",{get:function(){return this._mapData.size},enumerable:!1,configurable:!0}),r.prototype.get=function(e){var n;return null===(n=i(this._mapData,e))||void 0===n?void 0:n.value},r.prototype.set=function(e,n){return o(this._mapData,e,n),this},r.prototype.has=function(e){return!!i(this._mapData,e)},r.prototype.delete=function(e){return!!s(this._mapData,e)},r.prototype.clear=function(){c(this._mapData)},r.prototype.keys=function(){return new t(this._mapData,(function(e,n){return e}))},r.prototype.values=function(){return new t(this._mapData,(function(e,n){return n}))},r.prototype.entries=function(){return new t(this._mapData,(function(e,n){return[e,n]}))},r.prototype.forEach=function(e){l(this._mapData,e)},r}()},e.createSetShim=function(e){var t=function(){function e(e,n){this._data=d(e,n)}return e.prototype.next=function(){return p(this._data)},e}();return function(){function r(t){var r=this;this._mapData=n(),u(e(t),(function(e){return r.add(e)}))}return Object.defineProperty(r.prototype,"size",{get:function(){return this._mapData.size},enumerable:!1,configurable:!0}),r.prototype.add=function(e){return o(this._mapData,e,e),this},r.prototype.has=function(e){return!!i(this._mapData,e)},r.prototype.delete=function(e){return!!s(this._mapData,e)},r.prototype.clear=function(){c(this._mapData)},r.prototype.keys=function(){return new t(this._mapData,(function(e,n){return e}))},r.prototype.values=function(){return new t(this._mapData,(function(e,n){return n}))},r.prototype.entries=function(){return new t(this._mapData,(function(e,n){return[e,n]}))},r.prototype.forEach=function(e){l(this._mapData,e)},r}()}}(e.ShimCollections||(e.ShimCollections={}))}(ts||(ts={})),function(e){e.versionMajorMinor="4.1",e.version="4.1.3",function(e){e[e.LessThan=-1]="LessThan",e[e.EqualTo=0]="EqualTo",e[e.GreaterThan=1]="GreaterThan"}(e.Comparison||(e.Comparison={})),function(e){e.tryGetNativeMap=function(){return"undefined"!=typeof Map&&"entries"in Map.prototype&&1===new Map([[0,0]]).size?Map:void 0},e.tryGetNativeSet=function(){return"undefined"!=typeof Set&&"entries"in Set.prototype&&1===new Set([0]).size?Set:void 0}}(e.NativeCollections||(e.NativeCollections={}))}(ts||(ts={})),function(e){function n(n,r,a){var i,o=null!==(i=e.NativeCollections[r]())&&void 0!==i?i:null===e.ShimCollections||void 0===e.ShimCollections?void 0:e.ShimCollections[a](t);if(o)return o;throw new Error("TypeScript requires an environment that provides a compatible native "+n+" implementation.")}function t(n){if(n){if(D(n))return f(n);if(n instanceof e.Map)return n.entries();if(n instanceof e.Set)return n.values();throw new Error("Iteration not supported.")}}function r(e,n,t){if(void 0===t&&(t=O),e)for(var r=0,a=e;r<a.length;r++){if(t(a[r],n))return!0}return!1}function a(e,n){if(e){if(!n)return e.length>0;for(var t=0,r=e;t<r.length;t++){if(n(r[t]))return!0}}return!1}function i(e,n){return a(n)?a(e)?__spreadArrays(e,n):n:e}function o(e,n){return n}function s(e){return e.map(o)}function c(e,n){return void 0===n?e:void 0===e?[n]:(e.push(n),e)}function l(e,n){return n<0?e.length+n:n}function u(e,n,t,r){if(void 0===n||0===n.length)return e;if(void 0===e)return n.slice(t,r);t=void 0===t?0:l(n,t),r=void 0===r?n.length:l(n,r);for(var a=t;a<r&&a<n.length;a++)void 0!==n[a]&&e.push(n[a]);return e}function d(e,n,t){return!r(e,n,t)&&(e.push(n),!0)}function p(e,n,t){n.sort((function(n,r){return t(e[n],e[r])||M(n,r)}))}function m(e,n){return 0===e.length?e:e.slice().sort(n)}function f(e){var n=0;return{next:function(){return n===e.length?{value:void 0,done:!0}:(n++,{value:e[n-1],done:!1})}}}function _(e,n,t,r,a){return g(e,t(n),t,r,a)}function g(e,n,t,r,i){if(!a(e))return-1;for(var o=i||0,s=e.length-1;o<=s;){var c=o+(s-o>>1);switch(r(t(e[c],c),n)){case-1:o=c+1;break;case 0:return c;case 1:s=c-1}}return~o}function y(e,n,t,r,a){if(e&&e.length>0){var i=e.length;if(i>0){var o=void 0===r||r<0?0:r,s=void 0===a||o+a>i-1?i-1:o+a,c=void 0;for(arguments.length<=2?(c=e[o],o++):c=t;o<=s;)c=n(c,e[o],o),o++;return c}}return t}e.Map=n("Map","tryGetNativeMap","createMapShim"),e.Set=n("Set","tryGetNativeSet","createSetShim"),e.getIterator=t,e.emptyArray=[],e.emptyMap=new e.Map,e.emptySet=new e.Set,e.createMap=function(){return new e.Map},e.createMapFromTemplate=function(n){var t=new e.Map;for(var r in n)v.call(n,r)&&t.set(r,n[r]);return t},e.length=function(e){return e?e.length:0},e.forEach=function(e,n){if(e)for(var t=0;t<e.length;t++){var r=n(e[t],t);if(r)return r}},e.forEachRight=function(e,n){if(e)for(var t=e.length-1;t>=0;t--){var r=n(e[t],t);if(r)return r}},e.firstDefined=function(e,n){if(void 0!==e)for(var t=0;t<e.length;t++){var r=n(e[t],t);if(void 0!==r)return r}},e.firstDefinedIterator=function(e,n){for(;;){var t=e.next();if(t.done)return;var r=n(t.value);if(void 0!==r)return r}},e.reduceLeftIterator=function(e,n,t){var r=t;if(e)for(var a=e.next(),i=0;!a.done;a=e.next(),i++)r=n(r,a.value,i);return r},e.zipWith=function(n,t,r){var a=[];e.Debug.assertEqual(n.length,t.length);for(var i=0;i<n.length;i++)a.push(r(n[i],t[i],i));return a},e.zipToIterator=function(n,t){e.Debug.assertEqual(n.length,t.length);var r=0;return{next:function(){return r===n.length?{value:void 0,done:!0}:(r++,{value:[n[r-1],t[r-1]],done:!1})}}},e.zipToMap=function(n,t){e.Debug.assert(n.length===t.length);for(var r=new e.Map,a=0;a<n.length;++a)r.set(n[a],t[a]);return r},e.intersperse=function(e,n){if(e.length<=1)return e;for(var t=[],r=0,a=e.length;r<a;r++)r&&t.push(n),t.push(e[r]);return t},e.every=function(e,n){if(e)for(var t=0;t<e.length;t++)if(!n(e[t],t))return!1;return!0},e.find=function(e,n){for(var t=0;t<e.length;t++){var r=e[t];if(n(r,t))return r}},e.findLast=function(e,n){for(var t=e.length-1;t>=0;t--){var r=e[t];if(n(r,t))return r}},e.findIndex=function(e,n,t){for(var r=t||0;r<e.length;r++)if(n(e[r],r))return r;return-1},e.findLastIndex=function(e,n,t){for(var r=void 0===t?e.length-1:t;r>=0;r--)if(n(e[r],r))return r;return-1},e.findMap=function(n,t){for(var r=0;r<n.length;r++){var a=t(n[r],r);if(a)return a}return e.Debug.fail()},e.contains=r,e.arraysEqual=function(e,n,t){return void 0===t&&(t=O),e.length===n.length&&e.every((function(e,r){return t(e,n[r])}))},e.indexOfAnyCharCode=function(e,n,t){for(var a=t||0;a<e.length;a++)if(r(n,e.charCodeAt(a)))return a;return-1},e.countWhere=function(e,n){var t=0;if(e)for(var r=0;r<e.length;r++){n(e[r],r)&&t++}return t},e.filter=function(e,n){if(e){for(var t=e.length,r=0;r<t&&n(e[r]);)r++;if(r<t){var a=e.slice(0,r);for(r++;r<t;){var i=e[r];n(i)&&a.push(i),r++}return a}}return e},e.filterMutate=function(e,n){for(var t=0,r=0;r<e.length;r++)n(e[r],r,e)&&(e[t]=e[r],t++);e.length=t},e.clear=function(e){e.length=0},e.map=function(e,n){var t;if(e){t=[];for(var r=0;r<e.length;r++)t.push(n(e[r],r))}return t},e.mapIterator=function(e,n){return{next:function(){var t=e.next();return t.done?t:{value:n(t.value),done:!1}}}},e.sameMap=function(e,n){if(e)for(var t=0;t<e.length;t++){var r=e[t],a=n(r,t);if(r!==a){var i=e.slice(0,t);for(i.push(a),t++;t<e.length;t++)i.push(n(e[t],t));return i}}return e},e.flatten=function(e){for(var n=[],t=0,r=e;t<r.length;t++){var a=r[t];a&&(D(a)?u(n,a):n.push(a))}return n},e.flatMap=function(n,t){var r;if(n)for(var a=0;a<n.length;a++){var i=t(n[a],a);i&&(r=D(i)?u(r,i):c(r,i))}return r||e.emptyArray},e.flatMapToMutable=function(e,n){var t=[];if(e)for(var r=0;r<e.length;r++){var a=n(e[r],r);a&&(D(a)?u(t,a):t.push(a))}return t},e.flatMapIterator=function(n,t){var r=n.next();if(r.done)return e.emptyIterator;var a=i(r.value);return{next:function(){for(;;){var e=a.next();if(!e.done)return e;var t=n.next();if(t.done)return t;a=i(t.value)}}};function i(n){var r=t(n);return void 0===r?e.emptyIterator:D(r)?f(r):r}},e.sameFlatMap=function(e,n){var t;if(e)for(var r=0;r<e.length;r++){var a=e[r],i=n(a,r);(t||a!==i||D(i))&&(t||(t=e.slice(0,r)),D(i)?u(t,i):t.push(i))}return t||e},e.mapAllOrFail=function(e,n){for(var t=[],r=0;r<e.length;r++){var a=n(e[r],r);if(void 0===a)return;t.push(a)}return t},e.mapDefined=function(e,n){var t=[];if(e)for(var r=0;r<e.length;r++){var a=n(e[r],r);void 0!==a&&t.push(a)}return t},e.mapDefinedIterator=function(e,n){return{next:function(){for(;;){var t=e.next();if(t.done)return t;var r=n(t.value);if(void 0!==r)return{value:r,done:!1}}}}},e.mapDefinedEntries=function(n,t){if(n){var r=new e.Map;return n.forEach((function(e,n){var a=t(n,e);if(void 0!==a){var i=a[0],o=a[1];void 0!==i&&void 0!==o&&r.set(i,o)}})),r}},e.mapDefinedValues=function(n,t){if(n){var r=new e.Set;return n.forEach((function(e){var n=t(e);void 0!==n&&r.add(n)})),r}},e.getOrUpdate=function(e,n,t){if(e.has(n))return e.get(n);var r=t();return e.set(n,r),r},e.tryAddToSet=function(e,n){return!e.has(n)&&(e.add(n),!0)},e.emptyIterator={next:function(){return{value:void 0,done:!0}}},e.singleIterator=function(e){var n=!1;return{next:function(){var t=n;return n=!0,t?{value:void 0,done:!0}:{value:e,done:!1}}}},e.spanMap=function(e,n,t){var r;if(e){r=[];for(var a=e.length,i=void 0,o=void 0,s=0,c=0;s<a;){for(;c<a;){if(o=n(e[c],c),0===c)i=o;else if(o!==i)break;c++}if(s<c){var l=t(e.slice(s,c),i,s,c);l&&r.push(l),s=c}i=o,c++}}return r},e.mapEntries=function(n,t){if(n){var r=new e.Map;return n.forEach((function(e,n){var a=t(n,e),i=a[0],o=a[1];r.set(i,o)})),r}},e.some=a,e.getRangesWhere=function(e,n,t){for(var r,a=0;a<e.length;a++)n(e[a])?r=void 0===r?a:r:void 0!==r&&(t(r,a),r=void 0);void 0!==r&&t(r,e.length)},e.concatenate=i,e.indicesOf=s,e.deduplicate=function(e,n,t){return 0===e.length?[]:1===e.length?e.slice():t?function(e,n,t){var r=s(e);p(e,r,t);for(var a=e[r[0]],i=[r[0]],o=1;o<r.length;o++){var c=r[o],l=e[c];n(a,l)||(i.push(c),a=l)}return i.sort(),i.map((function(n){return e[n]}))}(e,n,t):function(e,n){for(var t=[],r=0,a=e;r<a.length;r++)d(t,a[r],n);return t}(e,n)},e.insertSorted=function(e,n,t){if(0!==e.length){var r=_(e,n,I,t);r<0&&e.splice(~r,0,n)}else e.push(n)},e.sortAndDeduplicate=function(n,t,r){return function(n,t){if(0===n.length)return e.emptyArray;for(var r=n[0],a=[r],i=1;i<n.length;i++){var o=n[i];switch(t(o,r)){case!0:case 0:continue;case-1:return e.Debug.fail("Array is unsorted.")}a.push(r=o)}return a}(m(n,t),r||t||F)},e.arrayIsSorted=function(e,n){if(e.length<2)return!0;for(var t=e[0],r=0,a=e.slice(1);r<a.length;r++){var i=a[r];if(1===n(t,i))return!1;t=i}return!0},e.arrayIsEqualTo=function(e,n,t){if(void 0===t&&(t=O),!e||!n)return e===n;if(e.length!==n.length)return!1;for(var r=0;r<e.length;r++)if(!t(e[r],n[r],r))return!1;return!0},e.compact=function(e){var n;if(e)for(var t=0;t<e.length;t++){var r=e[t];!n&&r||(n||(n=e.slice(0,t)),r&&n.push(r))}return n||e},e.relativeComplement=function(n,t,r){if(!t||!n||0===t.length||0===n.length)return t;var a=[];e:for(var i=0,o=0;o<t.length;o++){o>0&&e.Debug.assertGreaterThanOrEqual(r(t[o],t[o-1]),0);n:for(var s=i;i<n.length;i++)switch(i>s&&e.Debug.assertGreaterThanOrEqual(r(n[i],n[i-1]),0),r(t[o],n[i])){case-1:a.push(t[o]);continue e;case 0:continue e;case 1:continue n}}return a},e.sum=function(e,n){for(var t=0,r=0,a=e;r<a.length;r++){t+=a[r][n]}return t},e.append=c,e.combine=function(e,n){return void 0===e?n:void 0===n?e:D(e)?D(n)?i(e,n):c(e,n):D(n)?c(n,e):[e,n]},e.addRange=u,e.pushIfUnique=d,e.appendIfUnique=function(e,n,t){return e?(d(e,n,t),e):[n]},e.sort=m,e.arrayIterator=f,e.arrayReverseIterator=function(e){var n=e.length;return{next:function(){return 0===n?{value:void 0,done:!0}:(n--,{value:e[n],done:!1})}}},e.stableSort=function(e,n){var t=s(e);return p(e,t,n),t.map((function(n){return e[n]}))},e.rangeEquals=function(e,n,t,r){for(;t<r;){if(e[t]!==n[t])return!1;t++}return!0},e.elementAt=function(e,n){if(e&&(n=l(e,n))<e.length)return e[n]},e.firstOrUndefined=function(e){return 0===e.length?void 0:e[0]},e.first=function(n){return e.Debug.assert(0!==n.length),n[0]},e.lastOrUndefined=function(e){return 0===e.length?void 0:e[e.length-1]},e.last=function(n){return e.Debug.assert(0!==n.length),n[n.length-1]},e.singleOrUndefined=function(e){return e&&1===e.length?e[0]:void 0},e.singleOrMany=function(e){return e&&1===e.length?e[0]:e},e.replaceElement=function(e,n,t){var r=e.slice(0);return r[n]=t,r},e.binarySearch=_,e.binarySearchKey=g,e.reduceLeft=y;var v=Object.prototype.hasOwnProperty;function h(e,n){return v.call(e,n)}function b(e){var n=[];for(var t in e)v.call(e,t)&&n.push(t);return n}e.hasProperty=h,e.getProperty=function(e,n){return v.call(e,n)?e[n]:void 0},e.getOwnKeys=b,e.getAllKeys=function(e){var n=[];do{for(var t=0,r=Object.getOwnPropertyNames(e);t<r.length;t++){d(n,r[t])}}while(e=Object.getPrototypeOf(e));return n},e.getOwnValues=function(e){var n=[];for(var t in e)v.call(e,t)&&n.push(e[t]);return n};var E=Object.entries||function(e){for(var n=b(e),t=Array(n.length),r=0;r<n.length;r++)t[r]=[n[r],e[n[r]]];return t};function T(e,n){for(var t=[],r=e.next();!r.done;r=e.next())t.push(n?n(r.value):r.value);return t}function S(e,n,t){void 0===t&&(t=I);for(var r=x(),a=0,i=e;a<i.length;a++){var o=i[a];r.add(n(o),t(o))}return r}function x(){var n=new e.Map;return n.add=L,n.remove=A,n}function L(e,n){var t=this.get(e);return t?t.push(n):this.set(e,t=[n]),t}function A(e,n){var t=this.get(e);t&&(W(t,n),t.length||this.delete(e))}function D(e){return Array.isArray?Array.isArray(e):e instanceof Array}function C(e){}function I(e){return e}function N(e){return e.toLowerCase()}e.getEntries=function(e){return e?E(e):[]},e.arrayOf=function(e,n){for(var t=new Array(e),r=0;r<e;r++)t[r]=n(r);return t},e.arrayFrom=T,e.assign=function(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];for(var r=0,a=n;r<a.length;r++){var i=a[r];if(void 0!==i)for(var o in i)h(i,o)&&(e[o]=i[o])}return e},e.equalOwnProperties=function(e,n,t){if(void 0===t&&(t=O),e===n)return!0;if(!e||!n)return!1;for(var r in e)if(v.call(e,r)){if(!v.call(n,r))return!1;if(!t(e[r],n[r]))return!1}for(var r in n)if(v.call(n,r)&&!v.call(e,r))return!1;return!0},e.arrayToMap=function(n,t,r){void 0===r&&(r=I);for(var a=new e.Map,i=0,o=n;i<o.length;i++){var s=o[i],c=t(s);void 0!==c&&a.set(c,r(s))}return a},e.arrayToNumericMap=function(e,n,t){void 0===t&&(t=I);for(var r=[],a=0,i=e;a<i.length;a++){var o=i[a];r[n(o)]=t(o)}return r},e.arrayToMultiMap=S,e.group=function(e,n,t){return void 0===t&&(t=I),T(S(e,n).values(),t)},e.clone=function(e){var n={};for(var t in e)v.call(e,t)&&(n[t]=e[t]);return n},e.extend=function(e,n){var t={};for(var r in n)v.call(n,r)&&(t[r]=n[r]);for(var r in e)v.call(e,r)&&(t[r]=e[r]);return t},e.copyProperties=function(e,n){for(var t in n)v.call(n,t)&&(e[t]=n[t])},e.maybeBind=function(e,n){return n?n.bind(e):void 0},e.createMultiMap=x,e.createUnderscoreEscapedMultiMap=function(){return x()},e.isArray=D,e.toArray=function(e){return D(e)?e:[e]},e.isString=function(e){return"string"==typeof e},e.isNumber=function(e){return"number"==typeof e},e.tryCast=function(e,n){return void 0!==e&&n(e)?e:void 0},e.cast=function(n,t){return void 0!==n&&t(n)?n:e.Debug.fail("Invalid cast. The supplied value "+n+" did not pass the test '"+e.Debug.getFunctionName(t)+"'.")},e.noop=C,e.returnFalse=function(){return!1},e.returnTrue=function(){return!0},e.returnUndefined=function(){},e.identity=I,e.toLowerCase=N;var R=/[^\u0130\u0131\u00DFa-z0-9\\/:\-_\. ]+/g;function k(e){return R.test(e)?e.replace(R,N):e}function O(e,n){return e===n}function P(e,n){return e===n?0:void 0===e?-1:void 0===n?1:e<n?-1:1}function M(e,n){return P(e,n)}function w(e,n){return e===n?0:void 0===e?-1:void 0===n?1:(e=e.toUpperCase())<(n=n.toUpperCase())?-1:e>n?1:0}function F(e,n){return P(e,n)}e.toFileNameLowerCase=k,e.notImplemented=function(){throw new Error("Not implemented")},e.memoize=function(e){var n;return function(){return e&&(n=e(),e=void 0),n}},e.memoizeOne=function(n){var t=new e.Map;return function(e){var r=typeof e+":"+e,a=t.get(r);return void 0!==a||t.has(r)||(a=n(e),t.set(r,a)),a}},e.compose=function(e,n,t,r,a){if(a){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return function(e){return y(i,(function(e,n){return n(e)}),e)}}return r?function(a){return r(t(n(e(a))))}:t?function(r){return t(n(e(r)))}:n?function(t){return n(e(t))}:e?function(n){return e(n)}:function(e){return e}},function(e){e[e.None=0]="None",e[e.Normal=1]="Normal",e[e.Aggressive=2]="Aggressive",e[e.VeryAggressive=3]="VeryAggressive"}(e.AssertionLevel||(e.AssertionLevel={})),e.equateValues=O,e.equateStringsCaseInsensitive=function(e,n){return e===n||void 0!==e&&void 0!==n&&e.toUpperCase()===n.toUpperCase()},e.equateStringsCaseSensitive=function(e,n){return O(e,n)},e.compareValues=M,e.compareTextSpans=function(e,n){return M(null==e?void 0:e.start,null==n?void 0:n.start)||M(null==e?void 0:e.length,null==n?void 0:n.length)},e.min=function(e,n,t){return-1===t(e,n)?e:n},e.compareStringsCaseInsensitive=w,e.compareStringsCaseSensitive=F,e.getStringComparer=function(e){return e?w:F};var G,B,U=function(){var e,n,t=function(){if("object"==typeof Intl&&"function"==typeof Intl.Collator)return a;if("function"==typeof String.prototype.localeCompare&&"function"==typeof String.prototype.toLocaleUpperCase&&"a".localeCompare("B")<0)return i;return o}();return function(r){return void 0===r?e||(e=t(r)):"en-US"===r?n||(n=t(r)):t(r)};function r(e,n,t){if(e===n)return 0;if(void 0===e)return-1;if(void 0===n)return 1;var r=t(e,n);return r<0?-1:r>0?1:0}function a(e){var n=new Intl.Collator(e,{usage:"sort",sensitivity:"variant"}).compare;return function(e,t){return r(e,t,n)}}function i(e){return void 0!==e?o():function(e,t){return r(e,t,n)};function n(e,n){return e.localeCompare(n)}}function o(){return function(n,t){return r(n,t,e)};function e(e,t){return n(e.toUpperCase(),t.toUpperCase())||n(e,t)}function n(e,n){return e<n?-1:e>n?1:0}}}();function V(e,n,t){for(var r=new Array(n.length+1),a=new Array(n.length+1),i=t+1,o=0;o<=n.length;o++)r[o]=o;for(o=1;o<=e.length;o++){var s=e.charCodeAt(o-1),c=o>t?o-t:1,l=n.length>t+o?t+o:n.length;a[0]=o;for(var u=o,d=1;d<c;d++)a[d]=i;for(d=c;d<=l;d++){var p=s===n.charCodeAt(d-1)?r[d-1]:Math.min(r[d]+1,a[d-1]+1,r[d-1]+2);a[d]=p,u=Math.min(u,p)}for(d=l+1;d<=n.length;d++)a[d]=i;if(u>t)return;var m=r;r=a,a=m}var f=r[n.length];return f>t?void 0:f}function j(e,n){var t=e.length-n.length;return t>=0&&e.indexOf(n,t)===t}function H(e,n){for(var t=n;t<e.length-1;t++)e[t]=e[t+1];e.pop()}function K(e,n){e[n]=e[e.length-1],e.pop()}function W(e,n){return function(e,n){for(var t=0;t<e.length;t++)if(n(e[t]))return K(e,t),!0;return!1}(e,(function(e){return e===n}))}function z(e,n){return 0===e.lastIndexOf(n,0)}function q(e,n){var t=e.prefix,r=e.suffix;return n.length>=t.length+r.length&&z(n,t)&&j(n,r)}function J(e,n,t,r){for(var a=0,i=e[r];a<i.length;a++){var o=i[a],s=void 0;t?(s=t.slice()).push(o):s=[o],r===e.length-1?n.push(s):J(e,n,s,r+1)}}e.getUILocale=function(){return B},e.setUILocale=function(e){B!==e&&(B=e,G=void 0)},e.compareStringsCaseSensitiveUI=function(e,n){return(G||(G=U(B)))(e,n)},e.compareProperties=function(e,n,t,r){return e===n?0:void 0===e?-1:void 0===n?1:r(e[t],n[t])},e.compareBooleans=function(e,n){return M(e?1:0,n?1:0)},e.getSpellingSuggestion=function(n,t,r){for(var a,i=Math.min(2,Math.floor(.34*n.length)),o=Math.floor(.4*n.length)+1,s=!1,c=n.toLowerCase(),l=0,u=t;l<u.length;l++){var d=u[l],p=r(d);if(void 0!==p&&Math.abs(p.length-c.length)<=i){var m=p.toLowerCase();if(m===c){if(p===n)continue;return d}if(s)continue;if(p.length<3)continue;var f=V(c,m,o-1);if(void 0===f)continue;f<3?(s=!0,a=d):(e.Debug.assert(f<o),o=f,a=d)}}return a},e.endsWith=j,e.removeSuffix=function(e,n){return j(e,n)?e.slice(0,e.length-n.length):e},e.tryRemoveSuffix=function(e,n){return j(e,n)?e.slice(0,e.length-n.length):void 0},e.stringContains=function(e,n){return-1!==e.indexOf(n)},e.removeMinAndVersionNumbers=function(e){var n=/[.-]((min)|(\d+(\.\d+)*))$/;return e.replace(n,"").replace(n,"")},e.orderedRemoveItem=function(e,n){for(var t=0;t<e.length;t++)if(e[t]===n)return H(e,t),!0;return!1},e.orderedRemoveItemAt=H,e.unorderedRemoveItemAt=K,e.unorderedRemoveItem=W,e.createGetCanonicalFileName=function(e){return e?I:k},e.patternText=function(e){return e.prefix+"*"+e.suffix},e.matchedText=function(n,t){return e.Debug.assert(q(n,t)),t.substring(n.prefix.length,t.length-n.suffix.length)},e.findBestPatternMatch=function(e,n,t){for(var r,a=-1,i=0,o=e;i<o.length;i++){var s=o[i],c=n(s);q(c,t)&&c.prefix.length>a&&(a=c.prefix.length,r=s)}return r},e.startsWith=z,e.removePrefix=function(e,n){return z(e,n)?e.substr(n.length):e},e.tryRemovePrefix=function(e,n,t){return void 0===t&&(t=I),z(t(e),t(n))?e.substring(n.length):void 0},e.and=function(e,n){return function(t){return e(t)&&n(t)}},e.or=function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return function(){for(var n=[],t=0;t<arguments.length;t++)n[t]=arguments[t];for(var r=0,a=e;r<a.length;r++){var i=a[r];if(i.apply(void 0,n))return!0}return!1}},e.not=function(e){return function(){for(var n=[],t=0;t<arguments.length;t++)n[t]=arguments[t];return!e.apply(void 0,n)}},e.assertType=function(e){},e.singleElementArray=function(e){return void 0===e?void 0:[e]},e.enumerateInsertsAndDeletes=function(e,n,t,r,a,i){i=i||C;for(var o=0,s=0,c=e.length,l=n.length,u=!1;o<c&&s<l;){var d=e[o],p=n[s],m=t(d,p);-1===m?(r(d),o++,u=!0):1===m?(a(p),s++,u=!0):(i(p,d),o++,s++)}for(;o<c;)r(e[o++]),u=!0;for(;s<l;)a(n[s++]),u=!0;return u},e.fill=function(e,n){for(var t=Array(e),r=0;r<e;r++)t[r]=n(r);return t},e.cartesianProduct=function(e){var n=[];return J(e,n,void 0,0),n},e.padLeft=function(e,n,t){return void 0===t&&(t=" "),n<=e.length?e:t.repeat(n-e.length)+e},e.padRight=function(e,n,t){return void 0===t&&(t=" "),n<=e.length?e:e+t.repeat(n-e.length)},e.takeWhile=function(e,n){for(var t=e.length,r=0;r<t&&n(e[r]);)r++;return e.slice(0,r)}}(ts||(ts={})),function(e){var n;!function(e){e[e.Off=0]="Off",e[e.Error=1]="Error",e[e.Warning=2]="Warning",e[e.Info=3]="Info",e[e.Verbose=4]="Verbose"}(n=e.LogLevel||(e.LogLevel={})),function(t){var r,a,i=0;function o(){return null!=r?r:r=new e.Version(e.version)}function s(e){return t.currentLogLevel<=e}function c(e,n){t.loggingHost&&s(e)&&t.loggingHost.log(e,n)}function l(e){c(n.Info,e)}t.currentLogLevel=n.Warning,t.isDebugging=!1,t.getTypeScriptVersion=o,t.shouldLog=s,t.log=l,(a=l=t.log||(t.log={})).error=function(e){c(n.Error,e)},a.warn=function(e){c(n.Warning,e)},a.log=function(e){c(n.Info,e)},a.trace=function(e){c(n.Verbose,e)};var u={};function d(e){return i>=e}function p(n,r){return!!d(n)||(u[r]={level:n,assertion:t[r]},t[r]=e.noop,!1)}function m(e,n){var t=new Error(e?"Debug Failure. "+e:"Debug Failure.");throw Error.captureStackTrace&&Error.captureStackTrace(t,n||m),t}function f(e,n,t,r){e||(n=n?"False expression: "+n:"False expression.",t&&(n+="\r\nVerbose Debug Information: "+("string"==typeof t?t:t())),m(n,r||f))}function _(e,n,t){null==e&&m(n,t||_)}function g(e,n,t){return _(e,n,t||g),e}function y(e,n,t){for(var r=0,a=e;r<a.length;r++){_(a[r],n,t||y)}}function v(e,n,t){return y(e,n,t||v),e}function h(e){if("function"!=typeof e)return"";if(e.hasOwnProperty("name"))return e.name;var n=Function.prototype.toString.call(e),t=/^function\s+([\w\$]+)\s*\(/.exec(n);return t?t[1]:""}function b(n,t,r){void 0===n&&(n=0);var a=function(n){var t=[];for(var r in n){var a=n[r];"number"==typeof a&&t.push([a,r])}return e.stableSort(t,(function(n,t){return e.compareValues(n[0],t[0])}))}(t);if(0===n)return a.length>0&&0===a[0][0]?a[0][1]:"0";if(r){for(var i="",o=n,s=0,c=a;s<c.length;s++){var l=c[s],u=l[0],d=l[1];if(u>n)break;0!==u&&u&n&&(i=i+(i?"|":"")+d,o&=~u)}if(0===o)return i}else for(var p=0,m=a;p<m.length;p++){var f=m[p];u=f[0],d=f[1];if(u===n)return d}return n.toString()}function E(n){return b(n,e.SyntaxKind,!1)}function T(n){return b(n,e.NodeFlags,!0)}function S(n){return b(n,e.ModifierFlags,!0)}function x(n){return b(n,e.TransformFlags,!0)}function L(n){return b(n,e.EmitFlags,!0)}function A(n){return b(n,e.SymbolFlags,!0)}function D(n){return b(n,e.TypeFlags,!0)}function C(n){return b(n,e.ObjectFlags,!0)}function I(n){return b(n,e.FlowFlags,!0)}t.getAssertionLevel=function(){return i},t.setAssertionLevel=function(n){var r=i;if(i=n,n>r)for(var a=0,o=e.getOwnKeys(u);a<o.length;a++){var s=o[a],c=u[s];void 0!==c&&t[s]!==c.assertion&&n>=c.level&&(t[s]=c,u[s]=void 0)}},t.shouldAssert=d,t.fail=m,t.failBadSyntaxKind=function e(n,t,r){return m((t||"Unexpected node.")+"\r\nNode "+E(n.kind)+" was unexpected.",r||e)},t.assert=f,t.assertEqual=function e(n,t,r,a,i){n!==t&&m("Expected "+n+" === "+t+". "+(r?a?r+" "+a:r:""),i||e)},t.assertLessThan=function e(n,t,r,a){n>=t&&m("Expected "+n+" < "+t+". "+(r||""),a||e)},t.assertLessThanOrEqual=function e(n,t,r){n>t&&m("Expected "+n+" <= "+t,r||e)},t.assertGreaterThanOrEqual=function e(n,t,r){n<t&&m("Expected "+n+" >= "+t,r||e)},t.assertIsDefined=_,t.checkDefined=g,t.assertDefined=g,t.assertEachIsDefined=y,t.checkEachDefined=v,t.assertEachDefined=v,t.assertNever=function n(t,r,a){return void 0===r&&(r="Illegal value:"),m(r+" "+("object"==typeof t&&e.hasProperty(t,"kind")&&e.hasProperty(t,"pos")&&E?"SyntaxKind: "+E(t.kind):JSON.stringify(t)),a||n)},t.assertEachNode=function n(t,r,a,i){p(1,"assertEachNode")&&f(void 0===r||e.every(t,r),a||"Unexpected node.",(function(){return"Node array did not pass test '"+h(r)+"'."}),i||n)},t.assertNode=function e(n,t,r,a){p(1,"assertNode")&&f(void 0!==n&&(void 0===t||t(n)),r||"Unexpected node.",(function(){return"Node "+E(n.kind)+" did not pass test '"+h(t)+"'."}),a||e)},t.assertNotNode=function e(n,t,r,a){p(1,"assertNotNode")&&f(void 0===n||void 0===t||!t(n),r||"Unexpected node.",(function(){return"Node "+E(n.kind)+" should not have passed test '"+h(t)+"'."}),a||e)},t.assertOptionalNode=function e(n,t,r,a){p(1,"assertOptionalNode")&&f(void 0===t||void 0===n||t(n),r||"Unexpected node.",(function(){return"Node "+E(n.kind)+" did not pass test '"+h(t)+"'."}),a||e)},t.assertOptionalToken=function e(n,t,r,a){p(1,"assertOptionalToken")&&f(void 0===t||void 0===n||n.kind===t,r||"Unexpected node.",(function(){return"Node "+E(n.kind)+" was not a '"+E(t)+"' token."}),a||e)},t.assertMissingNode=function e(n,t,r){p(1,"assertMissingNode")&&f(void 0===n,t||"Unexpected node.",(function(){return"Node "+E(n.kind)+" was unexpected'."}),r||e)},t.getFunctionName=h,t.formatSymbol=function(n){return"{ name: "+e.unescapeLeadingUnderscores(n.escapedName)+"; flags: "+A(n.flags)+"; declarations: "+e.map(n.declarations,(function(e){return E(e.kind)}))+" }"},t.formatEnum=b,t.formatSyntaxKind=E,t.formatNodeFlags=T,t.formatModifierFlags=S,t.formatTransformFlags=x,t.formatEmitFlags=L,t.formatSymbolFlags=A,t.formatTypeFlags=D,t.formatObjectFlags=C,t.formatFlowFlags=I;var N,R,k,O=!1;function P(e){return function(){if(F(),!N)throw new Error("Debugging helpers could not be loaded.");return N}().formatControlFlowGraph(e)}function M(n){"__debugFlowFlags"in n||Object.defineProperties(n,{__tsDebuggerDisplay:{value:function(){var e=2&this.flags?"FlowStart":4&this.flags?"FlowBranchLabel":8&this.flags?"FlowLoopLabel":16&this.flags?"FlowAssignment":32&this.flags?"FlowTrueCondition":64&this.flags?"FlowFalseCondition":128&this.flags?"FlowSwitchClause":256&this.flags?"FlowArrayMutation":512&this.flags?"FlowCall":1024&this.flags?"FlowReduceLabel":1&this.flags?"FlowUnreachable":"UnknownFlow",n=-2048&this.flags;return e+(n?" ("+I(n)+")":"")}},__debugFlowFlags:{get:function(){return b(this.flags,e.FlowFlags,!0)}},__debugToString:{value:function(){return P(this)}}})}function w(e){"__tsDebuggerDisplay"in e||Object.defineProperties(e,{__tsDebuggerDisplay:{value:function(e){return"NodeArray "+(e=String(e).replace(/(?:,[\s\w\d_]+:[^,]+)+\]$/,"]"))}}})}function F(){if(!O){var n,t;Object.defineProperties(e.objectAllocator.getSymbolConstructor().prototype,{__tsDebuggerDisplay:{value:function(){var n=33554432&this.flags?"TransientSymbol":"Symbol",t=-33554433&this.flags;return n+" '"+e.symbolName(this)+"'"+(t?" ("+A(t)+")":"")}},__debugFlags:{get:function(){return A(this.flags)}}}),Object.defineProperties(e.objectAllocator.getTypeConstructor().prototype,{__tsDebuggerDisplay:{value:function(){var n=98304&this.flags?"NullableType":384&this.flags?"LiteralType "+JSON.stringify(this.value):2048&this.flags?"LiteralType "+(this.value.negative?"-":"")+this.value.base10Value+"n":8192&this.flags?"UniqueESSymbolType":32&this.flags?"EnumType":67359327&this.flags?"IntrinsicType "+this.intrinsicName:1048576&this.flags?"UnionType":2097152&this.flags?"IntersectionType":4194304&this.flags?"IndexType":8388608&this.flags?"IndexedAccessType":16777216&this.flags?"ConditionalType":33554432&this.flags?"SubstitutionType":262144&this.flags?"TypeParameter":524288&this.flags?3&this.objectFlags?"InterfaceType":4&this.objectFlags?"TypeReference":8&this.objectFlags?"TupleType":16&this.objectFlags?"AnonymousType":32&this.objectFlags?"MappedType":2048&this.objectFlags?"ReverseMappedType":256&this.objectFlags?"EvolvingArrayType":"ObjectType":"Type",t=524288&this.flags?-2368&this.objectFlags:0;return n+(this.symbol?" '"+e.symbolName(this.symbol)+"'":"")+(t?" ("+C(t)+")":"")}},__debugFlags:{get:function(){return D(this.flags)}},__debugObjectFlags:{get:function(){return 524288&this.flags?C(this.objectFlags):""}},__debugTypeToString:{value:function(){var e=(void 0===n&&"function"==typeof WeakMap&&(n=new WeakMap),n),t=null==e?void 0:e.get(this);return void 0===t&&(t=this.checker.typeToString(this),null==e||e.set(this,t)),t}}});for(var r=0,a=[e.objectAllocator.getNodeConstructor(),e.objectAllocator.getIdentifierConstructor(),e.objectAllocator.getTokenConstructor(),e.objectAllocator.getSourceFileConstructor()];r<a.length;r++){var i=a[r];i.prototype.hasOwnProperty("__debugKind")||Object.defineProperties(i.prototype,{__tsDebuggerDisplay:{value:function(){return(e.isGeneratedIdentifier(this)?"GeneratedIdentifier":e.isIdentifier(this)?"Identifier '"+e.idText(this)+"'":e.isPrivateIdentifier(this)?"PrivateIdentifier '"+e.idText(this)+"'":e.isStringLiteral(this)?"StringLiteral "+JSON.stringify(this.text.length<10?this.text:this.text.slice(10)+"..."):e.isNumericLiteral(this)?"NumericLiteral "+this.text:e.isBigIntLiteral(this)?"BigIntLiteral "+this.text+"n":e.isTypeParameterDeclaration(this)?"TypeParameterDeclaration":e.isParameter(this)?"ParameterDeclaration":e.isConstructorDeclaration(this)?"ConstructorDeclaration":e.isGetAccessorDeclaration(this)?"GetAccessorDeclaration":e.isSetAccessorDeclaration(this)?"SetAccessorDeclaration":e.isCallSignatureDeclaration(this)?"CallSignatureDeclaration":e.isConstructSignatureDeclaration(this)?"ConstructSignatureDeclaration":e.isIndexSignatureDeclaration(this)?"IndexSignatureDeclaration":e.isTypePredicateNode(this)?"TypePredicateNode":e.isTypeReferenceNode(this)?"TypeReferenceNode":e.isFunctionTypeNode(this)?"FunctionTypeNode":e.isConstructorTypeNode(this)?"ConstructorTypeNode":e.isTypeQueryNode(this)?"TypeQueryNode":e.isTypeLiteralNode(this)?"TypeLiteralNode":e.isArrayTypeNode(this)?"ArrayTypeNode":e.isTupleTypeNode(this)?"TupleTypeNode":e.isOptionalTypeNode(this)?"OptionalTypeNode":e.isRestTypeNode(this)?"RestTypeNode":e.isUnionTypeNode(this)?"UnionTypeNode":e.isIntersectionTypeNode(this)?"IntersectionTypeNode":e.isConditionalTypeNode(this)?"ConditionalTypeNode":e.isInferTypeNode(this)?"InferTypeNode":e.isParenthesizedTypeNode(this)?"ParenthesizedTypeNode":e.isThisTypeNode(this)?"ThisTypeNode":e.isTypeOperatorNode(this)?"TypeOperatorNode":e.isIndexedAccessTypeNode(this)?"IndexedAccessTypeNode":e.isMappedTypeNode(this)?"MappedTypeNode":e.isLiteralTypeNode(this)?"LiteralTypeNode":e.isNamedTupleMember(this)?"NamedTupleMember":e.isImportTypeNode(this)?"ImportTypeNode":E(this.kind))+(this.flags?" ("+T(this.flags)+")":"")}},__debugKind:{get:function(){return E(this.kind)}},__debugNodeFlags:{get:function(){return T(this.flags)}},__debugModifierFlags:{get:function(){return S(e.getEffectiveModifierFlagsNoCache(this))}},__debugTransformFlags:{get:function(){return x(this.transformFlags)}},__debugIsParseTreeNode:{get:function(){return e.isParseTreeNode(this)}},__debugEmitFlags:{get:function(){return L(e.getEmitFlags(this))}},__debugGetText:{value:function(n){if(e.nodeIsSynthesized(this))return"";var r=(void 0===t&&"function"==typeof WeakMap&&(t=new WeakMap),t),a=null==r?void 0:r.get(this);if(void 0===a){var i=e.getParseTreeNode(this),o=i&&e.getSourceFileOfNode(i);a=o?e.getSourceTextOfNodeFromSourceFile(o,i,n):"",null==r||r.set(this,a)}return a}}})}try{if(e.sys&&e.sys.require){e.getDirectoryPath(e.resolvePath(e.sys.getExecutingFilePath()));var o=void 0;o.error||(o.module.init(e),N=o.module)}}catch(e){}O=!0}}function G(n,t,r,a,i){var o=t?"DeprecationError: ":"DeprecationWarning: ";return o+="'"+n+"' ",o+=a?"has been deprecated since v"+a:"is deprecated",o+=t?" and can no longer be used.":r?" and will no longer be usable after v"+r+".":".",o+=i?" "+e.formatStringFromArgs(i,[n],0):""}function B(n,t){var r,a;void 0===t&&(t={});var i="string"==typeof t.typeScriptVersion?new e.Version(t.typeScriptVersion):null!==(r=t.typeScriptVersion)&&void 0!==r?r:o(),s="string"==typeof t.errorAfter?new e.Version(t.errorAfter):t.errorAfter,c="string"==typeof t.warnAfter?new e.Version(t.warnAfter):t.warnAfter,u="string"==typeof t.since?new e.Version(t.since):null!==(a=t.since)&&void 0!==a?a:c,d=t.error||s&&i.compareTo(s)<=0,p=!c||i.compareTo(c)>=0;return d?function(e,n,t,r){var a=G(e,!0,n,t,r);return function(){throw new TypeError(a)}}(n,s,u,t.message):p?function(e,n,t,r){var a=!1;return function(){a||(l.warn(G(e,!1,n,t,r)),a=!0)}}(n,s,u,t.message):e.noop}t.printControlFlowGraph=function(e){return console.log(P(e))},t.formatControlFlowGraph=P,t.attachFlowNodeDebugInfo=function(e){O&&("function"==typeof Object.setPrototypeOf?(R||M(R=Object.create(Object.prototype)),Object.setPrototypeOf(e,R)):M(e))},t.attachNodeArrayDebugInfo=function(e){O&&("function"==typeof Object.setPrototypeOf?(k||w(k=Object.create(Array.prototype)),Object.setPrototypeOf(e,k)):w(e))},t.enableDebugInfo=F,t.deprecate=function(e,n){return function(e,n){return function(){return e(),n.apply(this,arguments)}}(B(h(e),n),e)}}(e.Debug||(e.Debug={}))}(ts||(ts={})),function(e){var n=/^(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i,t=/^(?:0|[1-9]\d*|[a-z-][a-z0-9-]*)(?:\.(?:0|[1-9]\d*|[a-z-][a-z0-9-]*))*$/i,r=/^[a-z0-9-]+(?:\.[a-z0-9-]+)*$/i,a=/^(0|[1-9]\d*)$/,i=function(){function n(n,a,i,s,c){if(void 0===a&&(a=0),void 0===i&&(i=0),void 0===s&&(s=""),void 0===c&&(c=""),"string"==typeof n){var l=e.Debug.checkDefined(o(n),"Invalid version");n=l.major,a=l.minor,i=l.patch,s=l.prerelease,c=l.build}e.Debug.assert(n>=0,"Invalid argument: major"),e.Debug.assert(a>=0,"Invalid argument: minor"),e.Debug.assert(i>=0,"Invalid argument: patch"),e.Debug.assert(!s||t.test(s),"Invalid argument: prerelease"),e.Debug.assert(!c||r.test(c),"Invalid argument: build"),this.major=n,this.minor=a,this.patch=i,this.prerelease=s?s.split("."):e.emptyArray,this.build=c?c.split("."):e.emptyArray}return n.tryParse=function(e){var t=o(e);if(t)return new n(t.major,t.minor,t.patch,t.prerelease,t.build)},n.prototype.compareTo=function(n){return this===n?0:void 0===n?1:e.compareValues(this.major,n.major)||e.compareValues(this.minor,n.minor)||e.compareValues(this.patch,n.patch)||function(n,t){if(n===t)return 0;if(0===n.length)return 0===t.length?0:1;if(0===t.length)return-1;for(var r=Math.min(n.length,t.length),i=0;i<r;i++){var o=n[i],s=t[i];if(o!==s){var c=a.test(o),l=a.test(s);if(c||l){if(c!==l)return c?-1:1;if(u=e.compareValues(+o,+s))return u}else{var u;if(u=e.compareStringsCaseSensitive(o,s))return u}}}return e.compareValues(n.length,t.length)}(this.prerelease,n.prerelease)},n.prototype.increment=function(t){switch(t){case"major":return new n(this.major+1,0,0);case"minor":return new n(this.major,this.minor+1,0);case"patch":return new n(this.major,this.minor,this.patch+1);default:return e.Debug.assertNever(t)}},n.prototype.toString=function(){var n=this.major+"."+this.minor+"."+this.patch;return e.some(this.prerelease)&&(n+="-"+this.prerelease.join(".")),e.some(this.build)&&(n+="+"+this.build.join(".")),n},n.zero=new n(0,0,0),n}();function o(e){var a=n.exec(e);if(a){var i=a[1],o=a[2],s=void 0===o?"0":o,c=a[3],l=void 0===c?"0":c,u=a[4],d=void 0===u?"":u,p=a[5],m=void 0===p?"":p;if((!d||t.test(d))&&(!m||r.test(m)))return{major:parseInt(i,10),minor:parseInt(s,10),patch:parseInt(l,10),prerelease:d,build:m}}}e.Version=i;var s=function(){function n(n){this._alternatives=n?e.Debug.checkDefined(m(n),"Invalid range spec."):e.emptyArray}return n.tryParse=function(e){var t=m(e);if(t){var r=new n("");return r._alternatives=t,r}},n.prototype.test=function(e){return"string"==typeof e&&(e=new i(e)),function(e,n){if(0===n.length)return!0;for(var t=0,r=n;t<r.length;t++){if(h(e,r[t]))return!0}return!1}(e,this._alternatives)},n.prototype.toString=function(){return n=this._alternatives,e.map(n,E).join(" || ")||"*";var n},n}();e.VersionRange=s;var c=/\s*\|\|\s*/g,l=/\s+/g,u=/^([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i,d=/^\s*([a-z0-9-+.*]+)\s+-\s+([a-z0-9-+.*]+)\s*$/i,p=/^\s*(~|\^|<|<=|>|>=|=)?\s*([a-z0-9-+.*]+)$/i;function m(e){for(var n=[],t=0,r=e.trim().split(c);t<r.length;t++){var a=r[t];if(a){var i=[],o=d.exec(a);if(o){if(!_(o[1],o[2],i))return}else for(var s=0,u=a.split(l);s<u.length;s++){var m=u[s],f=p.exec(m);if(!f||!g(f[1],f[2],i))return}n.push(i)}}return n}function f(e){var n=u.exec(e);if(n){var t=n[1],r=n[2],a=void 0===r?"*":r,o=n[3],s=void 0===o?"*":o,c=n[4],l=n[5];return{version:new i(y(t)?0:parseInt(t,10),y(t)||y(a)?0:parseInt(a,10),y(t)||y(a)||y(s)?0:parseInt(s,10),c,l),major:t,minor:a,patch:s}}}function _(e,n,t){var r=f(e);if(!r)return!1;var a=f(n);return!!a&&(y(r.major)||t.push(v(">=",r.version)),y(a.major)||t.push(y(a.minor)?v("<",a.version.increment("major")):y(a.patch)?v("<",a.version.increment("minor")):v("<=",a.version)),!0)}function g(e,n,t){var r=f(n);if(!r)return!1;var a=r.version,o=r.major,s=r.minor,c=r.patch;if(y(o))"<"!==e&&">"!==e||t.push(v("<",i.zero));else switch(e){case"~":t.push(v(">=",a)),t.push(v("<",a.increment(y(s)?"major":"minor")));break;case"^":t.push(v(">=",a)),t.push(v("<",a.increment(a.major>0||y(s)?"major":a.minor>0||y(c)?"minor":"patch")));break;case"<":case">=":t.push(v(e,a));break;case"<=":case">":t.push(y(s)?v("<="===e?"<":">=",a.increment("major")):y(c)?v("<="===e?"<":">=",a.increment("minor")):v(e,a));break;case"=":case void 0:y(s)||y(c)?(t.push(v(">=",a)),t.push(v("<",a.increment(y(s)?"major":"minor")))):t.push(v("=",a));break;default:return!1}return!0}function y(e){return"*"===e||"x"===e||"X"===e}function v(e,n){return{operator:e,operand:n}}function h(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(!b(e,a.operator,a.operand))return!1}return!0}function b(n,t,r){var a=n.compareTo(r);switch(t){case"<":return a<0;case"<=":return a<=0;case">":return a>0;case">=":return a>=0;case"=":return 0===a;default:return e.Debug.assertNever(t)}}function E(n){return e.map(n,T).join(" ")}function T(e){return""+e.operator+e.operand}}(ts||(ts={})),function(e){function n(e,n){return"object"==typeof e&&"number"==typeof e.timeOrigin&&"function"==typeof e.mark&&"function"==typeof e.measure&&"function"==typeof e.now&&"function"==typeof n}var t=function(){if("object"==typeof performance&&"function"==typeof PerformanceObserver&&n(performance,PerformanceObserver))return{performance:performance,PerformanceObserver:PerformanceObserver}}()||function(){if("object"==typeof module&&"function"==typeof require)try{var t={},r=t.performance,a=t.PerformanceObserver;if(n(r,a)){var i=new e.Version(process.versions.node);return new e.VersionRange("<12.16.3 || 13 <13.13").test(i)?{performance:{get timeOrigin(){return r.timeOrigin},now:function(){return r.now()},mark:function(e){return r.mark(e)},measure:function(e,n,t){void 0===n&&(n="nodeStart"),void 0===t&&(t="__performance.measure-fix__",r.mark(t)),r.measure(e,n,t),"__performance.measure-fix__"===t&&r.clearMarks("__performance.measure-fix__")}},PerformanceObserver:a}:{performance:r,PerformanceObserver:a}}}catch(e){}}(),r=null==t?void 0:t.performance;e.tryGetNativePerformanceHooks=function(){return t},e.timestamp=r?function(){return r.now()}:Date.now?Date.now:function(){return+new Date}}(ts||(ts={})),function(e){!function(n){var t,r,a;function i(n,t,r){var a=0;return{enter:function(){1==++a&&c(t)},exit:function(){0==--a?(c(r),l(n,t,r)):a<0&&e.Debug.fail("enter/exit count does not match.")}}}n.createTimerIf=function(e,t,r,a){return e?i(t,r,a):n.nullTimer},n.createTimer=i,n.nullTimer={enter:e.noop,exit:e.noop};var o=new e.Map,s=new e.Map;function c(e){null==a||a.mark(e)}function l(e,n,t){null==a||a.measure(e,n,t)}function u(e){for(var n=0,t=e.getEntriesByType("mark");n<t.length;n++){var r=t[n];o.set(r.name,(o.get(r.name)||0)+1)}for(var a=0,i=e.getEntriesByType("measure");a<i.length;a++){var c=i[a];s.set(c.name,(s.get(c.name)||0)+c.duration)}}n.mark=c,n.measure=l,n.getCount=function(e){return o.get(e)||0},n.getDuration=function(e){return s.get(e)||0},n.forEachMeasure=function(e){s.forEach((function(n,t){return e(t,n)}))},n.isEnabled=function(){return!!a},n.enable=function(){if(!a){if(t||(t=e.tryGetNativePerformanceHooks()),!t)return!1;r||(r=new t.PerformanceObserver(u)),r.observe({entryTypes:["mark","measure"]}),a=t.performance}return!0},n.disable=function(){null==r||r.disconnect(),a=void 0,o.clear(),s.clear()}}(e.performance||(e.performance={}))}(ts||(ts={})),function(e){var n,t,r={logEvent:e.noop,logErrEvent:e.noop,logPerfEvent:e.noop,logInfoEvent:e.noop,logStartCommand:e.noop,logStopCommand:e.noop,logStartUpdateProgram:e.noop,logStopUpdateProgram:e.noop,logStartUpdateGraph:e.noop,logStopUpdateGraph:e.noop,logStartResolveModule:e.noop,logStopResolveModule:e.noop,logStartParseSourceFile:e.noop,logStopParseSourceFile:e.noop,logStartReadFile:e.noop,logStopReadFile:e.noop,logStartBindFile:e.noop,logStopBindFile:e.noop,logStartScheduledOperation:e.noop,logStopScheduledOperation:e.noop};try{n=process.env.TS_ETW_MODULE_PATH;t=void 0}catch(e){t=void 0}e.perfLogger=t&&t.logEvent?t:r}(ts||(ts={})),function(e){!function(n){var t,r,a,i=0,o=[];n.startTracing=function(n,s,c){if(e.Debug.assert(!r,"Tracing already started"),void 0===t)try{t=void 0}catch(e){t=!1}if(t){void 0===a&&(a=e.combinePaths(s,"legend.json")),t.existsSync(s)||t.mkdirSync(s,{recursive:!0});var l=c?"."+ ++i:"",u=e.combinePaths(s,"trace"+l+".json"),d=e.combinePaths(s,"types"+l+".json");o.push({configFilePath:n,tracePath:u,typesPath:d}),r=t.openSync(u,"w");var p={cat:"__metadata",ph:"M",ts:1e3*e.timestamp(),pid:1,tid:1};t.writeSync(r,"[\n"+[__assign({name:"process_name",args:{name:"tsc"}},p),__assign({name:"thread_name",args:{name:"Main"}},p),__assign(__assign({name:"TracingStartedInBrowser"},p),{cat:"disabled-by-default-devtools.timeline"})].map((function(e){return JSON.stringify(e)})).join(",\n"))}},n.stopTracing=function(n){r?(e.Debug.assert(t),t.writeSync(r,"\n]\n"),t.closeSync(r),r=void 0,n?function(n){var r,a,i,s,c,u,d,p,m,f,_,g,y,v,h,b;e.Debug.assert(t),e.performance.mark("beginDumpTypes");var E=o[o.length-1].typesPath,T=t.openSync(E,"w"),S=new e.Map;t.writeSync(T,"[");for(var x=n.length,L=0;L<x;L++){var A=n[L],D=A.objectFlags,C=null!==(r=A.aliasSymbol)&&void 0!==r?r:A.symbol,I=null===(a=null==C?void 0:C.declarations)||void 0===a?void 0:a[0],N=I&&e.getSourceFileOfNode(I),R=void 0;if(16&D|2944&A.flags)try{R=null===(i=A.checker)||void 0===i?void 0:i.typeToString(A)}catch(e){R=void 0}var k={};if(8388608&A.flags){var O=A;k={indexedAccessObjectType:null===(s=O.objectType)||void 0===s?void 0:s.id,indexedAccessIndexType:null===(c=O.indexType)||void 0===c?void 0:c.id}}var P={};if(4&D){var M=A;P={instantiatedType:null===(u=M.target)||void 0===u?void 0:u.id,typeArguments:null===(d=M.resolvedTypeArguments)||void 0===d?void 0:d.map((function(e){return e.id}))}}var w={};if(16777216&A.flags){var F=A;w={conditionalCheckType:null===(p=F.checkType)||void 0===p?void 0:p.id,conditionalExtendsType:null===(m=F.extendsType)||void 0===m?void 0:m.id,conditionalTrueType:null!==(_=null===(f=F.resolvedTrueType)||void 0===f?void 0:f.id)&&void 0!==_?_:-1,conditionalFalseType:null!==(y=null===(g=F.resolvedFalseType)||void 0===g?void 0:g.id)&&void 0!==y?y:-1}}var G=void 0,B=A.checker.getRecursionIdentity(A);B&&((G=S.get(B))||(G=S.size,S.set(B,G)));var U=__assign(__assign(__assign(__assign({id:A.id,intrinsicName:A.intrinsicName,symbolName:(null==C?void 0:C.escapedName)&&e.unescapeLeadingUnderscores(C.escapedName),recursionId:G,unionTypes:1048576&A.flags?null===(v=A.types)||void 0===v?void 0:v.map((function(e){return e.id})):void 0,intersectionTypes:2097152&A.flags?A.types.map((function(e){return e.id})):void 0,aliasTypeArguments:null===(h=A.aliasTypeArguments)||void 0===h?void 0:h.map((function(e){return e.id})),keyofType:4194304&A.flags?null===(b=A.type)||void 0===b?void 0:b.id:void 0},k),P),w),{firstDeclaration:I&&{path:N.path,start:l(e.getLineAndCharacterOfPosition(N,I.pos)),end:l(e.getLineAndCharacterOfPosition(e.getSourceFileOfNode(I),I.end))},flags:e.Debug.formatTypeFlags(A.flags).split("|"),display:R});t.writeSync(T,JSON.stringify(U)),L<x-1&&t.writeSync(T,",\n")}t.writeSync(T,"]\n"),t.closeSync(T),e.performance.mark("endDumpTypes"),e.performance.measure("Dump types","beginDumpTypes","endDumpTypes")}(n):o[o.length-1].typesPath=void 0):e.Debug.assert(!t,"Tracing is not in progress")},n.isTracing=function(){return!!r},function(e){e.Parse="parse",e.Program="program",e.Bind="bind",e.Check="check",e.Emit="emit"}(n.Phase||(n.Phase={})),n.begin=function(e,n,t){r&&c("B",e,n,t)},n.end=function(e,n,t){r&&c("E",e,n,t)},n.instant=function(e,n,t){r&&c("I",e,n,t,'"s":"g"')};var s=[];function c(n,a,i,o,s,c){void 0===c&&(c=1e3*e.timestamp()),e.Debug.assert(r),e.Debug.assert(t),e.performance.mark("beginTracing"),t.writeSync(r,',\n{"pid":1,"tid":1,"ph":"'+n+'","cat":"'+a+'","ts":'+c+',"name":"'+i+'"'),s&&t.writeSync(r,","+s),o&&t.writeSync(r,',"args":'+JSON.stringify(o)),t.writeSync(r,"}"),e.performance.mark("endTracing"),e.performance.measure("Tracing","beginTracing","endTracing")}function l(e){return{line:e.line+1,character:e.character+1}}n.push=function(n,t,a){r&&s.push({phase:n,name:t,args:a,time:1e3*e.timestamp()})},n.pop=function(){if(r){e.Debug.assert(s.length>0);var n=s.pop(),t=n.phase,a=n.name,i=n.args,o=n.time;c("X",t,a,i,'"dur":'+(1e3*e.timestamp()-o),o)}},n.dumpLegend=function(){a&&(e.Debug.assert(t),t.writeFileSync(a,JSON.stringify(o)))}}(e.tracing||(e.tracing={}))}(ts||(ts={})),function(e){!function(e){e[e.Unknown=0]="Unknown",e[e.EndOfFileToken=1]="EndOfFileToken",e[e.SingleLineCommentTrivia=2]="SingleLineCommentTrivia",e[e.MultiLineCommentTrivia=3]="MultiLineCommentTrivia",e[e.NewLineTrivia=4]="NewLineTrivia",e[e.WhitespaceTrivia=5]="WhitespaceTrivia",e[e.ShebangTrivia=6]="ShebangTrivia",e[e.ConflictMarkerTrivia=7]="ConflictMarkerTrivia",e[e.NumericLiteral=8]="NumericLiteral",e[e.BigIntLiteral=9]="BigIntLiteral",e[e.StringLiteral=10]="StringLiteral",e[e.JsxText=11]="JsxText",e[e.JsxTextAllWhiteSpaces=12]="JsxTextAllWhiteSpaces",e[e.RegularExpressionLiteral=13]="RegularExpressionLiteral",e[e.NoSubstitutionTemplateLiteral=14]="NoSubstitutionTemplateLiteral",e[e.TemplateHead=15]="TemplateHead",e[e.TemplateMiddle=16]="TemplateMiddle",e[e.TemplateTail=17]="TemplateTail",e[e.OpenBraceToken=18]="OpenBraceToken",e[e.CloseBraceToken=19]="CloseBraceToken",e[e.OpenParenToken=20]="OpenParenToken",e[e.CloseParenToken=21]="CloseParenToken",e[e.OpenBracketToken=22]="OpenBracketToken",e[e.CloseBracketToken=23]="CloseBracketToken",e[e.DotToken=24]="DotToken",e[e.DotDotDotToken=25]="DotDotDotToken",e[e.SemicolonToken=26]="SemicolonToken",e[e.CommaToken=27]="CommaToken",e[e.QuestionDotToken=28]="QuestionDotToken",e[e.LessThanToken=29]="LessThanToken",e[e.LessThanSlashToken=30]="LessThanSlashToken",e[e.GreaterThanToken=31]="GreaterThanToken",e[e.LessThanEqualsToken=32]="LessThanEqualsToken",e[e.GreaterThanEqualsToken=33]="GreaterThanEqualsToken",e[e.EqualsEqualsToken=34]="EqualsEqualsToken",e[e.ExclamationEqualsToken=35]="ExclamationEqualsToken",e[e.EqualsEqualsEqualsToken=36]="EqualsEqualsEqualsToken",e[e.ExclamationEqualsEqualsToken=37]="ExclamationEqualsEqualsToken",e[e.EqualsGreaterThanToken=38]="EqualsGreaterThanToken",e[e.PlusToken=39]="PlusToken",e[e.MinusToken=40]="MinusToken",e[e.AsteriskToken=41]="AsteriskToken",e[e.AsteriskAsteriskToken=42]="AsteriskAsteriskToken",e[e.SlashToken=43]="SlashToken",e[e.PercentToken=44]="PercentToken",e[e.PlusPlusToken=45]="PlusPlusToken",e[e.MinusMinusToken=46]="MinusMinusToken",e[e.LessThanLessThanToken=47]="LessThanLessThanToken",e[e.GreaterThanGreaterThanToken=48]="GreaterThanGreaterThanToken",e[e.GreaterThanGreaterThanGreaterThanToken=49]="GreaterThanGreaterThanGreaterThanToken",e[e.AmpersandToken=50]="AmpersandToken",e[e.BarToken=51]="BarToken",e[e.CaretToken=52]="CaretToken",e[e.ExclamationToken=53]="ExclamationToken",e[e.TildeToken=54]="TildeToken",e[e.AmpersandAmpersandToken=55]="AmpersandAmpersandToken",e[e.BarBarToken=56]="BarBarToken",e[e.QuestionToken=57]="QuestionToken",e[e.ColonToken=58]="ColonToken",e[e.AtToken=59]="AtToken",e[e.QuestionQuestionToken=60]="QuestionQuestionToken",e[e.BacktickToken=61]="BacktickToken",e[e.EqualsToken=62]="EqualsToken",e[e.PlusEqualsToken=63]="PlusEqualsToken",e[e.MinusEqualsToken=64]="MinusEqualsToken",e[e.AsteriskEqualsToken=65]="AsteriskEqualsToken",e[e.AsteriskAsteriskEqualsToken=66]="AsteriskAsteriskEqualsToken",e[e.SlashEqualsToken=67]="SlashEqualsToken",e[e.PercentEqualsToken=68]="PercentEqualsToken",e[e.LessThanLessThanEqualsToken=69]="LessThanLessThanEqualsToken",e[e.GreaterThanGreaterThanEqualsToken=70]="GreaterThanGreaterThanEqualsToken",e[e.GreaterThanGreaterThanGreaterThanEqualsToken=71]="GreaterThanGreaterThanGreaterThanEqualsToken",e[e.AmpersandEqualsToken=72]="AmpersandEqualsToken",e[e.BarEqualsToken=73]="BarEqualsToken",e[e.BarBarEqualsToken=74]="BarBarEqualsToken",e[e.AmpersandAmpersandEqualsToken=75]="AmpersandAmpersandEqualsToken",e[e.QuestionQuestionEqualsToken=76]="QuestionQuestionEqualsToken",e[e.CaretEqualsToken=77]="CaretEqualsToken",e[e.Identifier=78]="Identifier",e[e.PrivateIdentifier=79]="PrivateIdentifier",e[e.BreakKeyword=80]="BreakKeyword",e[e.CaseKeyword=81]="CaseKeyword",e[e.CatchKeyword=82]="CatchKeyword",e[e.ClassKeyword=83]="ClassKeyword",e[e.ConstKeyword=84]="ConstKeyword",e[e.ContinueKeyword=85]="ContinueKeyword",e[e.DebuggerKeyword=86]="DebuggerKeyword",e[e.DefaultKeyword=87]="DefaultKeyword",e[e.DeleteKeyword=88]="DeleteKeyword",e[e.DoKeyword=89]="DoKeyword",e[e.ElseKeyword=90]="ElseKeyword",e[e.EnumKeyword=91]="EnumKeyword",e[e.ExportKeyword=92]="ExportKeyword",e[e.ExtendsKeyword=93]="ExtendsKeyword",e[e.FalseKeyword=94]="FalseKeyword",e[e.FinallyKeyword=95]="FinallyKeyword",e[e.ForKeyword=96]="ForKeyword",e[e.FunctionKeyword=97]="FunctionKeyword",e[e.IfKeyword=98]="IfKeyword",e[e.ImportKeyword=99]="ImportKeyword",e[e.InKeyword=100]="InKeyword",e[e.InstanceOfKeyword=101]="InstanceOfKeyword",e[e.NewKeyword=102]="NewKeyword",e[e.NullKeyword=103]="NullKeyword",e[e.ReturnKeyword=104]="ReturnKeyword",e[e.SuperKeyword=105]="SuperKeyword",e[e.SwitchKeyword=106]="SwitchKeyword",e[e.ThisKeyword=107]="ThisKeyword",e[e.ThrowKeyword=108]="ThrowKeyword",e[e.TrueKeyword=109]="TrueKeyword",e[e.TryKeyword=110]="TryKeyword",e[e.TypeOfKeyword=111]="TypeOfKeyword",e[e.VarKeyword=112]="VarKeyword",e[e.VoidKeyword=113]="VoidKeyword",e[e.WhileKeyword=114]="WhileKeyword",e[e.WithKeyword=115]="WithKeyword",e[e.ImplementsKeyword=116]="ImplementsKeyword",e[e.InterfaceKeyword=117]="InterfaceKeyword",e[e.LetKeyword=118]="LetKeyword",e[e.PackageKeyword=119]="PackageKeyword",e[e.PrivateKeyword=120]="PrivateKeyword",e[e.ProtectedKeyword=121]="ProtectedKeyword",e[e.PublicKeyword=122]="PublicKeyword",e[e.StaticKeyword=123]="StaticKeyword",e[e.YieldKeyword=124]="YieldKeyword",e[e.AbstractKeyword=125]="AbstractKeyword",e[e.AsKeyword=126]="AsKeyword",e[e.AssertsKeyword=127]="AssertsKeyword",e[e.AnyKeyword=128]="AnyKeyword",e[e.AsyncKeyword=129]="AsyncKeyword",e[e.AwaitKeyword=130]="AwaitKeyword",e[e.BooleanKeyword=131]="BooleanKeyword",e[e.ConstructorKeyword=132]="ConstructorKeyword",e[e.DeclareKeyword=133]="DeclareKeyword",e[e.GetKeyword=134]="GetKeyword",e[e.InferKeyword=135]="InferKeyword",e[e.IntrinsicKeyword=136]="IntrinsicKeyword",e[e.IsKeyword=137]="IsKeyword",e[e.KeyOfKeyword=138]="KeyOfKeyword",e[e.ModuleKeyword=139]="ModuleKeyword",e[e.NamespaceKeyword=140]="NamespaceKeyword",e[e.NeverKeyword=141]="NeverKeyword",e[e.ReadonlyKeyword=142]="ReadonlyKeyword",e[e.RequireKeyword=143]="RequireKeyword",e[e.NumberKeyword=144]="NumberKeyword",e[e.ObjectKeyword=145]="ObjectKeyword",e[e.SetKeyword=146]="SetKeyword",e[e.StringKeyword=147]="StringKeyword",e[e.SymbolKeyword=148]="SymbolKeyword",e[e.TypeKeyword=149]="TypeKeyword",e[e.UndefinedKeyword=150]="UndefinedKeyword",e[e.UniqueKeyword=151]="UniqueKeyword",e[e.UnknownKeyword=152]="UnknownKeyword",e[e.FromKeyword=153]="FromKeyword",e[e.GlobalKeyword=154]="GlobalKeyword",e[e.BigIntKeyword=155]="BigIntKeyword",e[e.OfKeyword=156]="OfKeyword",e[e.QualifiedName=157]="QualifiedName",e[e.ComputedPropertyName=158]="ComputedPropertyName",e[e.TypeParameter=159]="TypeParameter",e[e.Parameter=160]="Parameter",e[e.Decorator=161]="Decorator",e[e.PropertySignature=162]="PropertySignature",e[e.PropertyDeclaration=163]="PropertyDeclaration",e[e.MethodSignature=164]="MethodSignature",e[e.MethodDeclaration=165]="MethodDeclaration",e[e.Constructor=166]="Constructor",e[e.GetAccessor=167]="GetAccessor",e[e.SetAccessor=168]="SetAccessor",e[e.CallSignature=169]="CallSignature",e[e.ConstructSignature=170]="ConstructSignature",e[e.IndexSignature=171]="IndexSignature",e[e.TypePredicate=172]="TypePredicate",e[e.TypeReference=173]="TypeReference",e[e.FunctionType=174]="FunctionType",e[e.ConstructorType=175]="ConstructorType",e[e.TypeQuery=176]="TypeQuery",e[e.TypeLiteral=177]="TypeLiteral",e[e.ArrayType=178]="ArrayType",e[e.TupleType=179]="TupleType",e[e.OptionalType=180]="OptionalType",e[e.RestType=181]="RestType",e[e.UnionType=182]="UnionType",e[e.IntersectionType=183]="IntersectionType",e[e.ConditionalType=184]="ConditionalType",e[e.InferType=185]="InferType",e[e.ParenthesizedType=186]="ParenthesizedType",e[e.ThisType=187]="ThisType",e[e.TypeOperator=188]="TypeOperator",e[e.IndexedAccessType=189]="IndexedAccessType",e[e.MappedType=190]="MappedType",e[e.LiteralType=191]="LiteralType",e[e.NamedTupleMember=192]="NamedTupleMember",e[e.TemplateLiteralType=193]="TemplateLiteralType",e[e.TemplateLiteralTypeSpan=194]="TemplateLiteralTypeSpan",e[e.ImportType=195]="ImportType",e[e.ObjectBindingPattern=196]="ObjectBindingPattern",e[e.ArrayBindingPattern=197]="ArrayBindingPattern",e[e.BindingElement=198]="BindingElement",e[e.ArrayLiteralExpression=199]="ArrayLiteralExpression",e[e.ObjectLiteralExpression=200]="ObjectLiteralExpression",e[e.PropertyAccessExpression=201]="PropertyAccessExpression",e[e.ElementAccessExpression=202]="ElementAccessExpression",e[e.CallExpression=203]="CallExpression",e[e.NewExpression=204]="NewExpression",e[e.TaggedTemplateExpression=205]="TaggedTemplateExpression",e[e.TypeAssertionExpression=206]="TypeAssertionExpression",e[e.ParenthesizedExpression=207]="ParenthesizedExpression",e[e.FunctionExpression=208]="FunctionExpression",e[e.ArrowFunction=209]="ArrowFunction",e[e.DeleteExpression=210]="DeleteExpression",e[e.TypeOfExpression=211]="TypeOfExpression",e[e.VoidExpression=212]="VoidExpression",e[e.AwaitExpression=213]="AwaitExpression",e[e.PrefixUnaryExpression=214]="PrefixUnaryExpression",e[e.PostfixUnaryExpression=215]="PostfixUnaryExpression",e[e.BinaryExpression=216]="BinaryExpression",e[e.ConditionalExpression=217]="ConditionalExpression",e[e.TemplateExpression=218]="TemplateExpression",e[e.YieldExpression=219]="YieldExpression",e[e.SpreadElement=220]="SpreadElement",e[e.ClassExpression=221]="ClassExpression",e[e.OmittedExpression=222]="OmittedExpression",e[e.ExpressionWithTypeArguments=223]="ExpressionWithTypeArguments",e[e.AsExpression=224]="AsExpression",e[e.NonNullExpression=225]="NonNullExpression",e[e.MetaProperty=226]="MetaProperty",e[e.SyntheticExpression=227]="SyntheticExpression",e[e.TemplateSpan=228]="TemplateSpan",e[e.SemicolonClassElement=229]="SemicolonClassElement",e[e.Block=230]="Block",e[e.EmptyStatement=231]="EmptyStatement",e[e.VariableStatement=232]="VariableStatement",e[e.ExpressionStatement=233]="ExpressionStatement",e[e.IfStatement=234]="IfStatement",e[e.DoStatement=235]="DoStatement",e[e.WhileStatement=236]="WhileStatement",e[e.ForStatement=237]="ForStatement",e[e.ForInStatement=238]="ForInStatement",e[e.ForOfStatement=239]="ForOfStatement",e[e.ContinueStatement=240]="ContinueStatement",e[e.BreakStatement=241]="BreakStatement",e[e.ReturnStatement=242]="ReturnStatement",e[e.WithStatement=243]="WithStatement",e[e.SwitchStatement=244]="SwitchStatement",e[e.LabeledStatement=245]="LabeledStatement",e[e.ThrowStatement=246]="ThrowStatement",e[e.TryStatement=247]="TryStatement",e[e.DebuggerStatement=248]="DebuggerStatement",e[e.VariableDeclaration=249]="VariableDeclaration",e[e.VariableDeclarationList=250]="VariableDeclarationList",e[e.FunctionDeclaration=251]="FunctionDeclaration",e[e.ClassDeclaration=252]="ClassDeclaration",e[e.InterfaceDeclaration=253]="InterfaceDeclaration",e[e.TypeAliasDeclaration=254]="TypeAliasDeclaration",e[e.EnumDeclaration=255]="EnumDeclaration",e[e.ModuleDeclaration=256]="ModuleDeclaration",e[e.ModuleBlock=257]="ModuleBlock",e[e.CaseBlock=258]="CaseBlock",e[e.NamespaceExportDeclaration=259]="NamespaceExportDeclaration",e[e.ImportEqualsDeclaration=260]="ImportEqualsDeclaration",e[e.ImportDeclaration=261]="ImportDeclaration",e[e.ImportClause=262]="ImportClause",e[e.NamespaceImport=263]="NamespaceImport",e[e.NamedImports=264]="NamedImports",e[e.ImportSpecifier=265]="ImportSpecifier",e[e.ExportAssignment=266]="ExportAssignment",e[e.ExportDeclaration=267]="ExportDeclaration",e[e.NamedExports=268]="NamedExports",e[e.NamespaceExport=269]="NamespaceExport",e[e.ExportSpecifier=270]="ExportSpecifier",e[e.MissingDeclaration=271]="MissingDeclaration",e[e.ExternalModuleReference=272]="ExternalModuleReference",e[e.JsxElement=273]="JsxElement",e[e.JsxSelfClosingElement=274]="JsxSelfClosingElement",e[e.JsxOpeningElement=275]="JsxOpeningElement",e[e.JsxClosingElement=276]="JsxClosingElement",e[e.JsxFragment=277]="JsxFragment",e[e.JsxOpeningFragment=278]="JsxOpeningFragment",e[e.JsxClosingFragment=279]="JsxClosingFragment",e[e.JsxAttribute=280]="JsxAttribute",e[e.JsxAttributes=281]="JsxAttributes",e[e.JsxSpreadAttribute=282]="JsxSpreadAttribute",e[e.JsxExpression=283]="JsxExpression",e[e.CaseClause=284]="CaseClause",e[e.DefaultClause=285]="DefaultClause",e[e.HeritageClause=286]="HeritageClause",e[e.CatchClause=287]="CatchClause",e[e.PropertyAssignment=288]="PropertyAssignment",e[e.ShorthandPropertyAssignment=289]="ShorthandPropertyAssignment",e[e.SpreadAssignment=290]="SpreadAssignment",e[e.EnumMember=291]="EnumMember",e[e.UnparsedPrologue=292]="UnparsedPrologue",e[e.UnparsedPrepend=293]="UnparsedPrepend",e[e.UnparsedText=294]="UnparsedText",e[e.UnparsedInternalText=295]="UnparsedInternalText",e[e.UnparsedSyntheticReference=296]="UnparsedSyntheticReference",e[e.SourceFile=297]="SourceFile",e[e.Bundle=298]="Bundle",e[e.UnparsedSource=299]="UnparsedSource",e[e.InputFiles=300]="InputFiles",e[e.JSDocTypeExpression=301]="JSDocTypeExpression",e[e.JSDocNameReference=302]="JSDocNameReference",e[e.JSDocAllType=303]="JSDocAllType",e[e.JSDocUnknownType=304]="JSDocUnknownType",e[e.JSDocNullableType=305]="JSDocNullableType",e[e.JSDocNonNullableType=306]="JSDocNonNullableType",e[e.JSDocOptionalType=307]="JSDocOptionalType",e[e.JSDocFunctionType=308]="JSDocFunctionType",e[e.JSDocVariadicType=309]="JSDocVariadicType",e[e.JSDocNamepathType=310]="JSDocNamepathType",e[e.JSDocComment=311]="JSDocComment",e[e.JSDocTypeLiteral=312]="JSDocTypeLiteral",e[e.JSDocSignature=313]="JSDocSignature",e[e.JSDocTag=314]="JSDocTag",e[e.JSDocAugmentsTag=315]="JSDocAugmentsTag",e[e.JSDocImplementsTag=316]="JSDocImplementsTag",e[e.JSDocAuthorTag=317]="JSDocAuthorTag",e[e.JSDocDeprecatedTag=318]="JSDocDeprecatedTag",e[e.JSDocClassTag=319]="JSDocClassTag",e[e.JSDocPublicTag=320]="JSDocPublicTag",e[e.JSDocPrivateTag=321]="JSDocPrivateTag",e[e.JSDocProtectedTag=322]="JSDocProtectedTag",e[e.JSDocReadonlyTag=323]="JSDocReadonlyTag",e[e.JSDocCallbackTag=324]="JSDocCallbackTag",e[e.JSDocEnumTag=325]="JSDocEnumTag",e[e.JSDocParameterTag=326]="JSDocParameterTag",e[e.JSDocReturnTag=327]="JSDocReturnTag",e[e.JSDocThisTag=328]="JSDocThisTag",e[e.JSDocTypeTag=329]="JSDocTypeTag",e[e.JSDocTemplateTag=330]="JSDocTemplateTag",e[e.JSDocTypedefTag=331]="JSDocTypedefTag",e[e.JSDocSeeTag=332]="JSDocSeeTag",e[e.JSDocPropertyTag=333]="JSDocPropertyTag",e[e.SyntaxList=334]="SyntaxList",e[e.NotEmittedStatement=335]="NotEmittedStatement",e[e.PartiallyEmittedExpression=336]="PartiallyEmittedExpression",e[e.CommaListExpression=337]="CommaListExpression",e[e.MergeDeclarationMarker=338]="MergeDeclarationMarker",e[e.EndOfDeclarationMarker=339]="EndOfDeclarationMarker",e[e.SyntheticReferenceExpression=340]="SyntheticReferenceExpression",e[e.Count=341]="Count",e[e.FirstAssignment=62]="FirstAssignment",e[e.LastAssignment=77]="LastAssignment",e[e.FirstCompoundAssignment=63]="FirstCompoundAssignment",e[e.LastCompoundAssignment=77]="LastCompoundAssignment",e[e.FirstReservedWord=80]="FirstReservedWord",e[e.LastReservedWord=115]="LastReservedWord",e[e.FirstKeyword=80]="FirstKeyword",e[e.LastKeyword=156]="LastKeyword",e[e.FirstFutureReservedWord=116]="FirstFutureReservedWord",e[e.LastFutureReservedWord=124]="LastFutureReservedWord",e[e.FirstTypeNode=172]="FirstTypeNode",e[e.LastTypeNode=195]="LastTypeNode",e[e.FirstPunctuation=18]="FirstPunctuation",e[e.LastPunctuation=77]="LastPunctuation",e[e.FirstToken=0]="FirstToken",e[e.LastToken=156]="LastToken",e[e.FirstTriviaToken=2]="FirstTriviaToken",e[e.LastTriviaToken=7]="LastTriviaToken",e[e.FirstLiteralToken=8]="FirstLiteralToken",e[e.LastLiteralToken=14]="LastLiteralToken",e[e.FirstTemplateToken=14]="FirstTemplateToken",e[e.LastTemplateToken=17]="LastTemplateToken",e[e.FirstBinaryOperator=29]="FirstBinaryOperator",e[e.LastBinaryOperator=77]="LastBinaryOperator",e[e.FirstStatement=232]="FirstStatement",e[e.LastStatement=248]="LastStatement",e[e.FirstNode=157]="FirstNode",e[e.FirstJSDocNode=301]="FirstJSDocNode",e[e.LastJSDocNode=333]="LastJSDocNode",e[e.FirstJSDocTagNode=314]="FirstJSDocTagNode",e[e.LastJSDocTagNode=333]="LastJSDocTagNode",e[e.FirstContextualKeyword=125]="FirstContextualKeyword",e[e.LastContextualKeyword=156]="LastContextualKeyword"}(e.SyntaxKind||(e.SyntaxKind={})),function(e){e[e.None=0]="None",e[e.Let=1]="Let",e[e.Const=2]="Const",e[e.NestedNamespace=4]="NestedNamespace",e[e.Synthesized=8]="Synthesized",e[e.Namespace=16]="Namespace",e[e.OptionalChain=32]="OptionalChain",e[e.ExportContext=64]="ExportContext",e[e.ContainsThis=128]="ContainsThis",e[e.HasImplicitReturn=256]="HasImplicitReturn",e[e.HasExplicitReturn=512]="HasExplicitReturn",e[e.GlobalAugmentation=1024]="GlobalAugmentation",e[e.HasAsyncFunctions=2048]="HasAsyncFunctions",e[e.DisallowInContext=4096]="DisallowInContext",e[e.YieldContext=8192]="YieldContext",e[e.DecoratorContext=16384]="DecoratorContext",e[e.AwaitContext=32768]="AwaitContext",e[e.ThisNodeHasError=65536]="ThisNodeHasError",e[e.JavaScriptFile=131072]="JavaScriptFile",e[e.ThisNodeOrAnySubNodesHasError=262144]="ThisNodeOrAnySubNodesHasError",e[e.HasAggregatedChildData=524288]="HasAggregatedChildData",e[e.PossiblyContainsDynamicImport=1048576]="PossiblyContainsDynamicImport",e[e.PossiblyContainsImportMeta=2097152]="PossiblyContainsImportMeta",e[e.JSDoc=4194304]="JSDoc",e[e.Ambient=8388608]="Ambient",e[e.InWithStatement=16777216]="InWithStatement",e[e.JsonFile=33554432]="JsonFile",e[e.TypeCached=67108864]="TypeCached",e[e.Deprecated=134217728]="Deprecated",e[e.BlockScoped=3]="BlockScoped",e[e.ReachabilityCheckFlags=768]="ReachabilityCheckFlags",e[e.ReachabilityAndEmitFlags=2816]="ReachabilityAndEmitFlags",e[e.ContextFlags=25358336]="ContextFlags",e[e.TypeExcludesFlags=40960]="TypeExcludesFlags",e[e.PermanentlySetIncrementalFlags=3145728]="PermanentlySetIncrementalFlags"}(e.NodeFlags||(e.NodeFlags={})),function(e){e[e.None=0]="None",e[e.Export=1]="Export",e[e.Ambient=2]="Ambient",e[e.Public=4]="Public",e[e.Private=8]="Private",e[e.Protected=16]="Protected",e[e.Static=32]="Static",e[e.Readonly=64]="Readonly",e[e.Abstract=128]="Abstract",e[e.Async=256]="Async",e[e.Default=512]="Default",e[e.Const=2048]="Const",e[e.HasComputedJSDocModifiers=4096]="HasComputedJSDocModifiers",e[e.Deprecated=8192]="Deprecated",e[e.HasComputedFlags=536870912]="HasComputedFlags",e[e.AccessibilityModifier=28]="AccessibilityModifier",e[e.ParameterPropertyModifier=92]="ParameterPropertyModifier",e[e.NonPublicAccessibilityModifier=24]="NonPublicAccessibilityModifier",e[e.TypeScriptModifier=2270]="TypeScriptModifier",e[e.ExportDefault=513]="ExportDefault",e[e.All=11263]="All"}(e.ModifierFlags||(e.ModifierFlags={})),function(e){e[e.None=0]="None",e[e.IntrinsicNamedElement=1]="IntrinsicNamedElement",e[e.IntrinsicIndexedElement=2]="IntrinsicIndexedElement",e[e.IntrinsicElement=3]="IntrinsicElement"}(e.JsxFlags||(e.JsxFlags={})),function(e){e[e.Succeeded=1]="Succeeded",e[e.Failed=2]="Failed",e[e.Reported=4]="Reported",e[e.ReportsUnmeasurable=8]="ReportsUnmeasurable",e[e.ReportsUnreliable=16]="ReportsUnreliable",e[e.ReportsMask=24]="ReportsMask"}(e.RelationComparisonResult||(e.RelationComparisonResult={})),function(e){e[e.None=0]="None",e[e.Auto=1]="Auto",e[e.Loop=2]="Loop",e[e.Unique=3]="Unique",e[e.Node=4]="Node",e[e.KindMask=7]="KindMask",e[e.ReservedInNestedScopes=8]="ReservedInNestedScopes",e[e.Optimistic=16]="Optimistic",e[e.FileLevel=32]="FileLevel",e[e.AllowNameSubstitution=64]="AllowNameSubstitution"}(e.GeneratedIdentifierFlags||(e.GeneratedIdentifierFlags={})),function(e){e[e.None=0]="None",e[e.PrecedingLineBreak=1]="PrecedingLineBreak",e[e.PrecedingJSDocComment=2]="PrecedingJSDocComment",e[e.Unterminated=4]="Unterminated",e[e.ExtendedUnicodeEscape=8]="ExtendedUnicodeEscape",e[e.Scientific=16]="Scientific",e[e.Octal=32]="Octal",e[e.HexSpecifier=64]="HexSpecifier",e[e.BinarySpecifier=128]="BinarySpecifier",e[e.OctalSpecifier=256]="OctalSpecifier",e[e.ContainsSeparator=512]="ContainsSeparator",e[e.UnicodeEscape=1024]="UnicodeEscape",e[e.ContainsInvalidEscape=2048]="ContainsInvalidEscape",e[e.BinaryOrOctalSpecifier=384]="BinaryOrOctalSpecifier",e[e.NumericLiteralFlags=1008]="NumericLiteralFlags",e[e.TemplateLiteralLikeFlags=2048]="TemplateLiteralLikeFlags"}(e.TokenFlags||(e.TokenFlags={})),function(e){e[e.Unreachable=1]="Unreachable",e[e.Start=2]="Start",e[e.BranchLabel=4]="BranchLabel",e[e.LoopLabel=8]="LoopLabel",e[e.Assignment=16]="Assignment",e[e.TrueCondition=32]="TrueCondition",e[e.FalseCondition=64]="FalseCondition",e[e.SwitchClause=128]="SwitchClause",e[e.ArrayMutation=256]="ArrayMutation",e[e.Call=512]="Call",e[e.ReduceLabel=1024]="ReduceLabel",e[e.Referenced=2048]="Referenced",e[e.Shared=4096]="Shared",e[e.Label=12]="Label",e[e.Condition=96]="Condition"}(e.FlowFlags||(e.FlowFlags={})),function(e){e[e.ExpectError=0]="ExpectError",e[e.Ignore=1]="Ignore"}(e.CommentDirectiveType||(e.CommentDirectiveType={}));var n,t=function(){};e.OperationCanceledException=t,function(e){e[e.Import=0]="Import",e[e.ReferenceFile=1]="ReferenceFile",e[e.TypeReferenceDirective=2]="TypeReferenceDirective"}(e.RefFileKind||(e.RefFileKind={})),function(e){e[e.Not=0]="Not",e[e.SafeModules=1]="SafeModules",e[e.Completely=2]="Completely"}(e.StructureIsReused||(e.StructureIsReused={})),function(e){e[e.Success=0]="Success",e[e.DiagnosticsPresent_OutputsSkipped=1]="DiagnosticsPresent_OutputsSkipped",e[e.DiagnosticsPresent_OutputsGenerated=2]="DiagnosticsPresent_OutputsGenerated",e[e.InvalidProject_OutputsSkipped=3]="InvalidProject_OutputsSkipped",e[e.ProjectReferenceCycle_OutputsSkipped=4]="ProjectReferenceCycle_OutputsSkipped",e[e.ProjectReferenceCycle_OutputsSkupped=4]="ProjectReferenceCycle_OutputsSkupped"}(e.ExitStatus||(e.ExitStatus={})),function(e){e[e.None=0]="None",e[e.Literal=1]="Literal",e[e.Subtype=2]="Subtype"}(e.UnionReduction||(e.UnionReduction={})),function(e){e[e.None=0]="None",e[e.Signature=1]="Signature",e[e.NoConstraints=2]="NoConstraints",e[e.Completions=4]="Completions",e[e.SkipBindingPatterns=8]="SkipBindingPatterns"}(e.ContextFlags||(e.ContextFlags={})),function(e){e[e.None=0]="None",e[e.NoTruncation=1]="NoTruncation",e[e.WriteArrayAsGenericType=2]="WriteArrayAsGenericType",e[e.GenerateNamesForShadowedTypeParams=4]="GenerateNamesForShadowedTypeParams",e[e.UseStructuralFallback=8]="UseStructuralFallback",e[e.ForbidIndexedAccessSymbolReferences=16]="ForbidIndexedAccessSymbolReferences",e[e.WriteTypeArgumentsOfSignature=32]="WriteTypeArgumentsOfSignature",e[e.UseFullyQualifiedType=64]="UseFullyQualifiedType",e[e.UseOnlyExternalAliasing=128]="UseOnlyExternalAliasing",e[e.SuppressAnyReturnType=256]="SuppressAnyReturnType",e[e.WriteTypeParametersInQualifiedName=512]="WriteTypeParametersInQualifiedName",e[e.MultilineObjectLiterals=1024]="MultilineObjectLiterals",e[e.WriteClassExpressionAsTypeLiteral=2048]="WriteClassExpressionAsTypeLiteral",e[e.UseTypeOfFunction=4096]="UseTypeOfFunction",e[e.OmitParameterModifiers=8192]="OmitParameterModifiers",e[e.UseAliasDefinedOutsideCurrentScope=16384]="UseAliasDefinedOutsideCurrentScope",e[e.UseSingleQuotesForStringLiteralType=268435456]="UseSingleQuotesForStringLiteralType",e[e.NoTypeReduction=536870912]="NoTypeReduction",e[e.NoUndefinedOptionalParameterType=1073741824]="NoUndefinedOptionalParameterType",e[e.AllowThisInObjectLiteral=32768]="AllowThisInObjectLiteral",e[e.AllowQualifedNameInPlaceOfIdentifier=65536]="AllowQualifedNameInPlaceOfIdentifier",e[e.AllowAnonymousIdentifier=131072]="AllowAnonymousIdentifier",e[e.AllowEmptyUnionOrIntersection=262144]="AllowEmptyUnionOrIntersection",e[e.AllowEmptyTuple=524288]="AllowEmptyTuple",e[e.AllowUniqueESSymbolType=1048576]="AllowUniqueESSymbolType",e[e.AllowEmptyIndexInfoType=2097152]="AllowEmptyIndexInfoType",e[e.AllowNodeModulesRelativePaths=67108864]="AllowNodeModulesRelativePaths",e[e.DoNotIncludeSymbolChain=134217728]="DoNotIncludeSymbolChain",e[e.IgnoreErrors=70221824]="IgnoreErrors",e[e.InObjectTypeLiteral=4194304]="InObjectTypeLiteral",e[e.InTypeAlias=8388608]="InTypeAlias",e[e.InInitialEntityName=16777216]="InInitialEntityName",e[e.InReverseMappedType=33554432]="InReverseMappedType"}(e.NodeBuilderFlags||(e.NodeBuilderFlags={})),function(e){e[e.None=0]="None",e[e.NoTruncation=1]="NoTruncation",e[e.WriteArrayAsGenericType=2]="WriteArrayAsGenericType",e[e.UseStructuralFallback=8]="UseStructuralFallback",e[e.WriteTypeArgumentsOfSignature=32]="WriteTypeArgumentsOfSignature",e[e.UseFullyQualifiedType=64]="UseFullyQualifiedType",e[e.SuppressAnyReturnType=256]="SuppressAnyReturnType",e[e.MultilineObjectLiterals=1024]="MultilineObjectLiterals",e[e.WriteClassExpressionAsTypeLiteral=2048]="WriteClassExpressionAsTypeLiteral",e[e.UseTypeOfFunction=4096]="UseTypeOfFunction",e[e.OmitParameterModifiers=8192]="OmitParameterModifiers",e[e.UseAliasDefinedOutsideCurrentScope=16384]="UseAliasDefinedOutsideCurrentScope",e[e.UseSingleQuotesForStringLiteralType=268435456]="UseSingleQuotesForStringLiteralType",e[e.NoTypeReduction=536870912]="NoTypeReduction",e[e.AllowUniqueESSymbolType=1048576]="AllowUniqueESSymbolType",e[e.AddUndefined=131072]="AddUndefined",e[e.WriteArrowStyleSignature=262144]="WriteArrowStyleSignature",e[e.InArrayType=524288]="InArrayType",e[e.InElementType=2097152]="InElementType",e[e.InFirstTypeArgument=4194304]="InFirstTypeArgument",e[e.InTypeAlias=8388608]="InTypeAlias",e[e.WriteOwnNameForAnyLike=0]="WriteOwnNameForAnyLike",e[e.NodeBuilderFlagsMask=814775659]="NodeBuilderFlagsMask"}(e.TypeFormatFlags||(e.TypeFormatFlags={})),function(e){e[e.None=0]="None",e[e.WriteTypeParametersOrArguments=1]="WriteTypeParametersOrArguments",e[e.UseOnlyExternalAliasing=2]="UseOnlyExternalAliasing",e[e.AllowAnyNodeKind=4]="AllowAnyNodeKind",e[e.UseAliasDefinedOutsideCurrentScope=8]="UseAliasDefinedOutsideCurrentScope",e[e.DoNotIncludeSymbolChain=16]="DoNotIncludeSymbolChain"}(e.SymbolFormatFlags||(e.SymbolFormatFlags={})),function(e){e[e.Accessible=0]="Accessible",e[e.NotAccessible=1]="NotAccessible",e[e.CannotBeNamed=2]="CannotBeNamed"}(e.SymbolAccessibility||(e.SymbolAccessibility={})),function(e){e[e.UnionOrIntersection=0]="UnionOrIntersection",e[e.Spread=1]="Spread"}(e.SyntheticSymbolKind||(e.SyntheticSymbolKind={})),function(e){e[e.This=0]="This",e[e.Identifier=1]="Identifier",e[e.AssertsThis=2]="AssertsThis",e[e.AssertsIdentifier=3]="AssertsIdentifier"}(e.TypePredicateKind||(e.TypePredicateKind={})),function(e){e[e.Unknown=0]="Unknown",e[e.TypeWithConstructSignatureAndValue=1]="TypeWithConstructSignatureAndValue",e[e.VoidNullableOrNeverType=2]="VoidNullableOrNeverType",e[e.NumberLikeType=3]="NumberLikeType",e[e.BigIntLikeType=4]="BigIntLikeType",e[e.StringLikeType=5]="StringLikeType",e[e.BooleanType=6]="BooleanType",e[e.ArrayLikeType=7]="ArrayLikeType",e[e.ESSymbolType=8]="ESSymbolType",e[e.Promise=9]="Promise",e[e.TypeWithCallSignature=10]="TypeWithCallSignature",e[e.ObjectType=11]="ObjectType"}(e.TypeReferenceSerializationKind||(e.TypeReferenceSerializationKind={})),function(e){e[e.None=0]="None",e[e.FunctionScopedVariable=1]="FunctionScopedVariable",e[e.BlockScopedVariable=2]="BlockScopedVariable",e[e.Property=4]="Property",e[e.EnumMember=8]="EnumMember",e[e.Function=16]="Function",e[e.Class=32]="Class",e[e.Interface=64]="Interface",e[e.ConstEnum=128]="ConstEnum",e[e.RegularEnum=256]="RegularEnum",e[e.ValueModule=512]="ValueModule",e[e.NamespaceModule=1024]="NamespaceModule",e[e.TypeLiteral=2048]="TypeLiteral",e[e.ObjectLiteral=4096]="ObjectLiteral",e[e.Method=8192]="Method",e[e.Constructor=16384]="Constructor",e[e.GetAccessor=32768]="GetAccessor",e[e.SetAccessor=65536]="SetAccessor",e[e.Signature=131072]="Signature",e[e.TypeParameter=262144]="TypeParameter",e[e.TypeAlias=524288]="TypeAlias",e[e.ExportValue=1048576]="ExportValue",e[e.Alias=2097152]="Alias",e[e.Prototype=4194304]="Prototype",e[e.ExportStar=8388608]="ExportStar",e[e.Optional=16777216]="Optional",e[e.Transient=33554432]="Transient",e[e.Assignment=67108864]="Assignment",e[e.ModuleExports=134217728]="ModuleExports",e[e.All=67108863]="All",e[e.Enum=384]="Enum",e[e.Variable=3]="Variable",e[e.Value=111551]="Value",e[e.Type=788968]="Type",e[e.Namespace=1920]="Namespace",e[e.Module=1536]="Module",e[e.Accessor=98304]="Accessor",e[e.FunctionScopedVariableExcludes=111550]="FunctionScopedVariableExcludes",e[e.BlockScopedVariableExcludes=111551]="BlockScopedVariableExcludes",e[e.ParameterExcludes=111551]="ParameterExcludes",e[e.PropertyExcludes=0]="PropertyExcludes",e[e.EnumMemberExcludes=900095]="EnumMemberExcludes",e[e.FunctionExcludes=110991]="FunctionExcludes",e[e.ClassExcludes=899503]="ClassExcludes",e[e.InterfaceExcludes=788872]="InterfaceExcludes",e[e.RegularEnumExcludes=899327]="RegularEnumExcludes",e[e.ConstEnumExcludes=899967]="ConstEnumExcludes",e[e.ValueModuleExcludes=110735]="ValueModuleExcludes",e[e.NamespaceModuleExcludes=0]="NamespaceModuleExcludes",e[e.MethodExcludes=103359]="MethodExcludes",e[e.GetAccessorExcludes=46015]="GetAccessorExcludes",e[e.SetAccessorExcludes=78783]="SetAccessorExcludes",e[e.TypeParameterExcludes=526824]="TypeParameterExcludes",e[e.TypeAliasExcludes=788968]="TypeAliasExcludes",e[e.AliasExcludes=2097152]="AliasExcludes",e[e.ModuleMember=2623475]="ModuleMember",e[e.ExportHasLocal=944]="ExportHasLocal",e[e.BlockScoped=418]="BlockScoped",e[e.PropertyOrAccessor=98308]="PropertyOrAccessor",e[e.ClassMember=106500]="ClassMember",e[e.ExportSupportsDefaultModifier=112]="ExportSupportsDefaultModifier",e[e.ExportDoesNotSupportDefaultModifier=-113]="ExportDoesNotSupportDefaultModifier",e[e.Classifiable=2885600]="Classifiable",e[e.LateBindingContainer=6256]="LateBindingContainer"}(e.SymbolFlags||(e.SymbolFlags={})),function(e){e[e.Numeric=0]="Numeric",e[e.Literal=1]="Literal"}(e.EnumKind||(e.EnumKind={})),function(e){e[e.Instantiated=1]="Instantiated",e[e.SyntheticProperty=2]="SyntheticProperty",e[e.SyntheticMethod=4]="SyntheticMethod",e[e.Readonly=8]="Readonly",e[e.ReadPartial=16]="ReadPartial",e[e.WritePartial=32]="WritePartial",e[e.HasNonUniformType=64]="HasNonUniformType",e[e.HasLiteralType=128]="HasLiteralType",e[e.ContainsPublic=256]="ContainsPublic",e[e.ContainsProtected=512]="ContainsProtected",e[e.ContainsPrivate=1024]="ContainsPrivate",e[e.ContainsStatic=2048]="ContainsStatic",e[e.Late=4096]="Late",e[e.ReverseMapped=8192]="ReverseMapped",e[e.OptionalParameter=16384]="OptionalParameter",e[e.RestParameter=32768]="RestParameter",e[e.DeferredType=65536]="DeferredType",e[e.HasNeverType=131072]="HasNeverType",e[e.Mapped=262144]="Mapped",e[e.StripOptional=524288]="StripOptional",e[e.Synthetic=6]="Synthetic",e[e.Discriminant=192]="Discriminant",e[e.Partial=48]="Partial"}(e.CheckFlags||(e.CheckFlags={})),function(e){e.Call="__call",e.Constructor="__constructor",e.New="__new",e.Index="__index",e.ExportStar="__export",e.Global="__global",e.Missing="__missing",e.Type="__type",e.Object="__object",e.JSXAttributes="__jsxAttributes",e.Class="__class",e.Function="__function",e.Computed="__computed",e.Resolving="__resolving__",e.ExportEquals="export=",e.Default="default",e.This="this"}(e.InternalSymbolName||(e.InternalSymbolName={})),function(e){e[e.TypeChecked=1]="TypeChecked",e[e.LexicalThis=2]="LexicalThis",e[e.CaptureThis=4]="CaptureThis",e[e.CaptureNewTarget=8]="CaptureNewTarget",e[e.SuperInstance=256]="SuperInstance",e[e.SuperStatic=512]="SuperStatic",e[e.ContextChecked=1024]="ContextChecked",e[e.AsyncMethodWithSuper=2048]="AsyncMethodWithSuper",e[e.AsyncMethodWithSuperBinding=4096]="AsyncMethodWithSuperBinding",e[e.CaptureArguments=8192]="CaptureArguments",e[e.EnumValuesComputed=16384]="EnumValuesComputed",e[e.LexicalModuleMergesWithClass=32768]="LexicalModuleMergesWithClass",e[e.LoopWithCapturedBlockScopedBinding=65536]="LoopWithCapturedBlockScopedBinding",e[e.ContainsCapturedBlockScopeBinding=131072]="ContainsCapturedBlockScopeBinding",e[e.CapturedBlockScopedBinding=262144]="CapturedBlockScopedBinding",e[e.BlockScopedBindingInLoop=524288]="BlockScopedBindingInLoop",e[e.ClassWithBodyScopedClassBinding=1048576]="ClassWithBodyScopedClassBinding",e[e.BodyScopedClassBinding=2097152]="BodyScopedClassBinding",e[e.NeedsLoopOutParameter=4194304]="NeedsLoopOutParameter",e[e.AssignmentsMarked=8388608]="AssignmentsMarked",e[e.ClassWithConstructorReference=16777216]="ClassWithConstructorReference",e[e.ConstructorReferenceInClass=33554432]="ConstructorReferenceInClass",e[e.ContainsClassWithPrivateIdentifiers=67108864]="ContainsClassWithPrivateIdentifiers"}(e.NodeCheckFlags||(e.NodeCheckFlags={})),function(e){e[e.Any=1]="Any",e[e.Unknown=2]="Unknown",e[e.String=4]="String",e[e.Number=8]="Number",e[e.Boolean=16]="Boolean",e[e.Enum=32]="Enum",e[e.BigInt=64]="BigInt",e[e.StringLiteral=128]="StringLiteral",e[e.NumberLiteral=256]="NumberLiteral",e[e.BooleanLiteral=512]="BooleanLiteral",e[e.EnumLiteral=1024]="EnumLiteral",e[e.BigIntLiteral=2048]="BigIntLiteral",e[e.ESSymbol=4096]="ESSymbol",e[e.UniqueESSymbol=8192]="UniqueESSymbol",e[e.Void=16384]="Void",e[e.Undefined=32768]="Undefined",e[e.Null=65536]="Null",e[e.Never=131072]="Never",e[e.TypeParameter=262144]="TypeParameter",e[e.Object=524288]="Object",e[e.Union=1048576]="Union",e[e.Intersection=2097152]="Intersection",e[e.Index=4194304]="Index",e[e.IndexedAccess=8388608]="IndexedAccess",e[e.Conditional=16777216]="Conditional",e[e.Substitution=33554432]="Substitution",e[e.NonPrimitive=67108864]="NonPrimitive",e[e.TemplateLiteral=134217728]="TemplateLiteral",e[e.StringMapping=268435456]="StringMapping",e[e.AnyOrUnknown=3]="AnyOrUnknown",e[e.Nullable=98304]="Nullable",e[e.Literal=2944]="Literal",e[e.Unit=109440]="Unit",e[e.StringOrNumberLiteral=384]="StringOrNumberLiteral",e[e.StringOrNumberLiteralOrUnique=8576]="StringOrNumberLiteralOrUnique",e[e.DefinitelyFalsy=117632]="DefinitelyFalsy",e[e.PossiblyFalsy=117724]="PossiblyFalsy",e[e.Intrinsic=67359327]="Intrinsic",e[e.Primitive=131068]="Primitive",e[e.StringLike=402653316]="StringLike",e[e.NumberLike=296]="NumberLike",e[e.BigIntLike=2112]="BigIntLike",e[e.BooleanLike=528]="BooleanLike",e[e.EnumLike=1056]="EnumLike",e[e.ESSymbolLike=12288]="ESSymbolLike",e[e.VoidLike=49152]="VoidLike",e[e.DisjointDomains=469892092]="DisjointDomains",e[e.UnionOrIntersection=3145728]="UnionOrIntersection",e[e.StructuredType=3670016]="StructuredType",e[e.TypeVariable=8650752]="TypeVariable",e[e.InstantiableNonPrimitive=58982400]="InstantiableNonPrimitive",e[e.InstantiablePrimitive=406847488]="InstantiablePrimitive",e[e.Instantiable=465829888]="Instantiable",e[e.StructuredOrInstantiable=469499904]="StructuredOrInstantiable",e[e.ObjectFlagsType=3899393]="ObjectFlagsType",e[e.Simplifiable=25165824]="Simplifiable",e[e.Substructure=469237760]="Substructure",e[e.Narrowable=536624127]="Narrowable",e[e.NotPrimitiveUnion=469647395]="NotPrimitiveUnion",e[e.IncludesMask=205258751]="IncludesMask",e[e.IncludesStructuredOrInstantiable=262144]="IncludesStructuredOrInstantiable",e[e.IncludesNonWideningType=4194304]="IncludesNonWideningType",e[e.IncludesWildcard=8388608]="IncludesWildcard",e[e.IncludesEmptyObject=16777216]="IncludesEmptyObject"}(e.TypeFlags||(e.TypeFlags={})),function(e){e[e.Class=1]="Class",e[e.Interface=2]="Interface",e[e.Reference=4]="Reference",e[e.Tuple=8]="Tuple",e[e.Anonymous=16]="Anonymous",e[e.Mapped=32]="Mapped",e[e.Instantiated=64]="Instantiated",e[e.ObjectLiteral=128]="ObjectLiteral",e[e.EvolvingArray=256]="EvolvingArray",e[e.ObjectLiteralPatternWithComputedProperties=512]="ObjectLiteralPatternWithComputedProperties",e[e.ContainsSpread=1024]="ContainsSpread",e[e.ReverseMapped=2048]="ReverseMapped",e[e.JsxAttributes=4096]="JsxAttributes",e[e.MarkerType=8192]="MarkerType",e[e.JSLiteral=16384]="JSLiteral",e[e.FreshLiteral=32768]="FreshLiteral",e[e.ArrayLiteral=65536]="ArrayLiteral",e[e.ObjectRestType=131072]="ObjectRestType",e[e.PrimitiveUnion=262144]="PrimitiveUnion",e[e.ContainsWideningType=524288]="ContainsWideningType",e[e.ContainsObjectOrArrayLiteral=1048576]="ContainsObjectOrArrayLiteral",e[e.NonInferrableType=2097152]="NonInferrableType",e[e.IsGenericObjectTypeComputed=4194304]="IsGenericObjectTypeComputed",e[e.IsGenericObjectType=8388608]="IsGenericObjectType",e[e.IsGenericIndexTypeComputed=16777216]="IsGenericIndexTypeComputed",e[e.IsGenericIndexType=33554432]="IsGenericIndexType",e[e.CouldContainTypeVariablesComputed=67108864]="CouldContainTypeVariablesComputed",e[e.CouldContainTypeVariables=134217728]="CouldContainTypeVariables",e[e.ContainsIntersections=268435456]="ContainsIntersections",e[e.IsNeverIntersectionComputed=268435456]="IsNeverIntersectionComputed",e[e.IsNeverIntersection=536870912]="IsNeverIntersection",e[e.IsClassInstanceClone=1073741824]="IsClassInstanceClone",e[e.ClassOrInterface=3]="ClassOrInterface",e[e.RequiresWidening=1572864]="RequiresWidening",e[e.PropagatingFlags=3670016]="PropagatingFlags",e[e.ObjectTypeKindMask=2367]="ObjectTypeKindMask"}(e.ObjectFlags||(e.ObjectFlags={})),function(e){e[e.Invariant=0]="Invariant",e[e.Covariant=1]="Covariant",e[e.Contravariant=2]="Contravariant",e[e.Bivariant=3]="Bivariant",e[e.Independent=4]="Independent",e[e.VarianceMask=7]="VarianceMask",e[e.Unmeasurable=8]="Unmeasurable",e[e.Unreliable=16]="Unreliable",e[e.AllowsStructuralFallback=24]="AllowsStructuralFallback"}(e.VarianceFlags||(e.VarianceFlags={})),function(e){e[e.Required=1]="Required",e[e.Optional=2]="Optional",e[e.Rest=4]="Rest",e[e.Variadic=8]="Variadic",e[e.Variable=12]="Variable"}(e.ElementFlags||(e.ElementFlags={})),function(e){e[e.Component=0]="Component",e[e.Function=1]="Function",e[e.Mixed=2]="Mixed"}(e.JsxReferenceKind||(e.JsxReferenceKind={})),function(e){e[e.Call=0]="Call",e[e.Construct=1]="Construct"}(e.SignatureKind||(e.SignatureKind={})),function(e){e[e.None=0]="None",e[e.HasRestParameter=1]="HasRestParameter",e[e.HasLiteralTypes=2]="HasLiteralTypes",e[e.IsInnerCallChain=4]="IsInnerCallChain",e[e.IsOuterCallChain=8]="IsOuterCallChain",e[e.IsUntypedSignatureInJSFile=16]="IsUntypedSignatureInJSFile",e[e.PropagatingFlags=19]="PropagatingFlags",e[e.CallChainFlags=12]="CallChainFlags"}(e.SignatureFlags||(e.SignatureFlags={})),function(e){e[e.String=0]="String",e[e.Number=1]="Number"}(e.IndexKind||(e.IndexKind={})),function(e){e[e.Simple=0]="Simple",e[e.Array=1]="Array",e[e.Function=2]="Function",e[e.Composite=3]="Composite",e[e.Merged=4]="Merged"}(e.TypeMapKind||(e.TypeMapKind={})),function(e){e[e.NakedTypeVariable=1]="NakedTypeVariable",e[e.SpeculativeTuple=2]="SpeculativeTuple",e[e.HomomorphicMappedType=4]="HomomorphicMappedType",e[e.PartialHomomorphicMappedType=8]="PartialHomomorphicMappedType",e[e.MappedTypeConstraint=16]="MappedTypeConstraint",e[e.ContravariantConditional=32]="ContravariantConditional",e[e.ReturnType=64]="ReturnType",e[e.LiteralKeyof=128]="LiteralKeyof",e[e.NoConstraints=256]="NoConstraints",e[e.AlwaysStrict=512]="AlwaysStrict",e[e.MaxValue=1024]="MaxValue",e[e.PriorityImpliesCombination=208]="PriorityImpliesCombination",e[e.Circularity=-1]="Circularity"}(e.InferencePriority||(e.InferencePriority={})),function(e){e[e.None=0]="None",e[e.NoDefault=1]="NoDefault",e[e.AnyDefault=2]="AnyDefault",e[e.SkippedGenericFunction=4]="SkippedGenericFunction"}(e.InferenceFlags||(e.InferenceFlags={})),function(e){e[e.False=0]="False",e[e.Unknown=1]="Unknown",e[e.Maybe=3]="Maybe",e[e.True=-1]="True"}(e.Ternary||(e.Ternary={})),function(e){e[e.None=0]="None",e[e.ExportsProperty=1]="ExportsProperty",e[e.ModuleExports=2]="ModuleExports",e[e.PrototypeProperty=3]="PrototypeProperty",e[e.ThisProperty=4]="ThisProperty",e[e.Property=5]="Property",e[e.Prototype=6]="Prototype",e[e.ObjectDefinePropertyValue=7]="ObjectDefinePropertyValue",e[e.ObjectDefinePropertyExports=8]="ObjectDefinePropertyExports",e[e.ObjectDefinePrototypeProperty=9]="ObjectDefinePrototypeProperty"}(e.AssignmentDeclarationKind||(e.AssignmentDeclarationKind={})),function(e){e[e.Warning=0]="Warning",e[e.Error=1]="Error",e[e.Suggestion=2]="Suggestion",e[e.Message=3]="Message"}(n=e.DiagnosticCategory||(e.DiagnosticCategory={})),e.diagnosticCategoryName=function(e,t){void 0===t&&(t=!0);var r=n[e.category];return t?r.toLowerCase():r},function(e){e[e.Classic=1]="Classic",e[e.NodeJs=2]="NodeJs"}(e.ModuleResolutionKind||(e.ModuleResolutionKind={})),function(e){e[e.FixedPollingInterval=0]="FixedPollingInterval",e[e.PriorityPollingInterval=1]="PriorityPollingInterval",e[e.DynamicPriorityPolling=2]="DynamicPriorityPolling",e[e.UseFsEvents=3]="UseFsEvents",e[e.UseFsEventsOnParentDirectory=4]="UseFsEventsOnParentDirectory"}(e.WatchFileKind||(e.WatchFileKind={})),function(e){e[e.UseFsEvents=0]="UseFsEvents",e[e.FixedPollingInterval=1]="FixedPollingInterval",e[e.DynamicPriorityPolling=2]="DynamicPriorityPolling"}(e.WatchDirectoryKind||(e.WatchDirectoryKind={})),function(e){e[e.FixedInterval=0]="FixedInterval",e[e.PriorityInterval=1]="PriorityInterval",e[e.DynamicPriority=2]="DynamicPriority"}(e.PollingWatchKind||(e.PollingWatchKind={})),function(e){e[e.None=0]="None",e[e.CommonJS=1]="CommonJS",e[e.AMD=2]="AMD",e[e.UMD=3]="UMD",e[e.System=4]="System",e[e.ES2015=5]="ES2015",e[e.ES2020=6]="ES2020",e[e.ESNext=99]="ESNext"}(e.ModuleKind||(e.ModuleKind={})),function(e){e[e.None=0]="None",e[e.Preserve=1]="Preserve",e[e.React=2]="React",e[e.ReactNative=3]="ReactNative",e[e.ReactJSX=4]="ReactJSX",e[e.ReactJSXDev=5]="ReactJSXDev"}(e.JsxEmit||(e.JsxEmit={})),function(e){e[e.Remove=0]="Remove",e[e.Preserve=1]="Preserve",e[e.Error=2]="Error"}(e.ImportsNotUsedAsValues||(e.ImportsNotUsedAsValues={})),function(e){e[e.CarriageReturnLineFeed=0]="CarriageReturnLineFeed",e[e.LineFeed=1]="LineFeed"}(e.NewLineKind||(e.NewLineKind={})),function(e){e[e.Unknown=0]="Unknown",e[e.JS=1]="JS",e[e.JSX=2]="JSX",e[e.TS=3]="TS",e[e.TSX=4]="TSX",e[e.External=5]="External",e[e.JSON=6]="JSON",e[e.Deferred=7]="Deferred"}(e.ScriptKind||(e.ScriptKind={})),function(e){e[e.ES3=0]="ES3",e[e.ES5=1]="ES5",e[e.ES2015=2]="ES2015",e[e.ES2016=3]="ES2016",e[e.ES2017=4]="ES2017",e[e.ES2018=5]="ES2018",e[e.ES2019=6]="ES2019",e[e.ES2020=7]="ES2020",e[e.ESNext=99]="ESNext",e[e.JSON=100]="JSON",e[e.Latest=99]="Latest"}(e.ScriptTarget||(e.ScriptTarget={})),function(e){e[e.Standard=0]="Standard",e[e.JSX=1]="JSX"}(e.LanguageVariant||(e.LanguageVariant={})),function(e){e[e.None=0]="None",e[e.Recursive=1]="Recursive"}(e.WatchDirectoryFlags||(e.WatchDirectoryFlags={})),function(e){e[e.nullCharacter=0]="nullCharacter",e[e.maxAsciiCharacter=127]="maxAsciiCharacter",e[e.lineFeed=10]="lineFeed",e[e.carriageReturn=13]="carriageReturn",e[e.lineSeparator=8232]="lineSeparator",e[e.paragraphSeparator=8233]="paragraphSeparator",e[e.nextLine=133]="nextLine",e[e.space=32]="space",e[e.nonBreakingSpace=160]="nonBreakingSpace",e[e.enQuad=8192]="enQuad",e[e.emQuad=8193]="emQuad",e[e.enSpace=8194]="enSpace",e[e.emSpace=8195]="emSpace",e[e.threePerEmSpace=8196]="threePerEmSpace",e[e.fourPerEmSpace=8197]="fourPerEmSpace",e[e.sixPerEmSpace=8198]="sixPerEmSpace",e[e.figureSpace=8199]="figureSpace",e[e.punctuationSpace=8200]="punctuationSpace",e[e.thinSpace=8201]="thinSpace",e[e.hairSpace=8202]="hairSpace",e[e.zeroWidthSpace=8203]="zeroWidthSpace",e[e.narrowNoBreakSpace=8239]="narrowNoBreakSpace",e[e.ideographicSpace=12288]="ideographicSpace",e[e.mathematicalSpace=8287]="mathematicalSpace",e[e.ogham=5760]="ogham",e[e._=95]="_",e[e.$=36]="$",e[e._0=48]="_0",e[e._1=49]="_1",e[e._2=50]="_2",e[e._3=51]="_3",e[e._4=52]="_4",e[e._5=53]="_5",e[e._6=54]="_6",e[e._7=55]="_7",e[e._8=56]="_8",e[e._9=57]="_9",e[e.a=97]="a",e[e.b=98]="b",e[e.c=99]="c",e[e.d=100]="d",e[e.e=101]="e",e[e.f=102]="f",e[e.g=103]="g",e[e.h=104]="h",e[e.i=105]="i",e[e.j=106]="j",e[e.k=107]="k",e[e.l=108]="l",e[e.m=109]="m",e[e.n=110]="n",e[e.o=111]="o",e[e.p=112]="p",e[e.q=113]="q",e[e.r=114]="r",e[e.s=115]="s",e[e.t=116]="t",e[e.u=117]="u",e[e.v=118]="v",e[e.w=119]="w",e[e.x=120]="x",e[e.y=121]="y",e[e.z=122]="z",e[e.A=65]="A",e[e.B=66]="B",e[e.C=67]="C",e[e.D=68]="D",e[e.E=69]="E",e[e.F=70]="F",e[e.G=71]="G",e[e.H=72]="H",e[e.I=73]="I",e[e.J=74]="J",e[e.K=75]="K",e[e.L=76]="L",e[e.M=77]="M",e[e.N=78]="N",e[e.O=79]="O",e[e.P=80]="P",e[e.Q=81]="Q",e[e.R=82]="R",e[e.S=83]="S",e[e.T=84]="T",e[e.U=85]="U",e[e.V=86]="V",e[e.W=87]="W",e[e.X=88]="X",e[e.Y=89]="Y",e[e.Z=90]="Z",e[e.ampersand=38]="ampersand",e[e.asterisk=42]="asterisk",e[e.at=64]="at",e[e.backslash=92]="backslash",e[e.backtick=96]="backtick",e[e.bar=124]="bar",e[e.caret=94]="caret",e[e.closeBrace=125]="closeBrace",e[e.closeBracket=93]="closeBracket",e[e.closeParen=41]="closeParen",e[e.colon=58]="colon",e[e.comma=44]="comma",e[e.dot=46]="dot",e[e.doubleQuote=34]="doubleQuote",e[e.equals=61]="equals",e[e.exclamation=33]="exclamation",e[e.greaterThan=62]="greaterThan",e[e.hash=35]="hash",e[e.lessThan=60]="lessThan",e[e.minus=45]="minus",e[e.openBrace=123]="openBrace",e[e.openBracket=91]="openBracket",e[e.openParen=40]="openParen",e[e.percent=37]="percent",e[e.plus=43]="plus",e[e.question=63]="question",e[e.semicolon=59]="semicolon",e[e.singleQuote=39]="singleQuote",e[e.slash=47]="slash",e[e.tilde=126]="tilde",e[e.backspace=8]="backspace",e[e.formFeed=12]="formFeed",e[e.byteOrderMark=65279]="byteOrderMark",e[e.tab=9]="tab",e[e.verticalTab=11]="verticalTab"}(e.CharacterCodes||(e.CharacterCodes={})),function(e){e.Ts=".ts",e.Tsx=".tsx",e.Dts=".d.ts",e.Js=".js",e.Jsx=".jsx",e.Json=".json",e.TsBuildInfo=".tsbuildinfo"}(e.Extension||(e.Extension={})),function(e){e[e.None=0]="None",e[e.ContainsTypeScript=1]="ContainsTypeScript",e[e.ContainsJsx=2]="ContainsJsx",e[e.ContainsESNext=4]="ContainsESNext",e[e.ContainsES2020=8]="ContainsES2020",e[e.ContainsES2019=16]="ContainsES2019",e[e.ContainsES2018=32]="ContainsES2018",e[e.ContainsES2017=64]="ContainsES2017",e[e.ContainsES2016=128]="ContainsES2016",e[e.ContainsES2015=256]="ContainsES2015",e[e.ContainsGenerator=512]="ContainsGenerator",e[e.ContainsDestructuringAssignment=1024]="ContainsDestructuringAssignment",e[e.ContainsTypeScriptClassSyntax=2048]="ContainsTypeScriptClassSyntax",e[e.ContainsLexicalThis=4096]="ContainsLexicalThis",e[e.ContainsRestOrSpread=8192]="ContainsRestOrSpread",e[e.ContainsObjectRestOrSpread=16384]="ContainsObjectRestOrSpread",e[e.ContainsComputedPropertyName=32768]="ContainsComputedPropertyName",e[e.ContainsBlockScopedBinding=65536]="ContainsBlockScopedBinding",e[e.ContainsBindingPattern=131072]="ContainsBindingPattern",e[e.ContainsYield=262144]="ContainsYield",e[e.ContainsAwait=524288]="ContainsAwait",e[e.ContainsHoistedDeclarationOrCompletion=1048576]="ContainsHoistedDeclarationOrCompletion",e[e.ContainsDynamicImport=2097152]="ContainsDynamicImport",e[e.ContainsClassFields=4194304]="ContainsClassFields",e[e.ContainsPossibleTopLevelAwait=8388608]="ContainsPossibleTopLevelAwait",e[e.HasComputedFlags=536870912]="HasComputedFlags",e[e.AssertTypeScript=1]="AssertTypeScript",e[e.AssertJsx=2]="AssertJsx",e[e.AssertESNext=4]="AssertESNext",e[e.AssertES2020=8]="AssertES2020",e[e.AssertES2019=16]="AssertES2019",e[e.AssertES2018=32]="AssertES2018",e[e.AssertES2017=64]="AssertES2017",e[e.AssertES2016=128]="AssertES2016",e[e.AssertES2015=256]="AssertES2015",e[e.AssertGenerator=512]="AssertGenerator",e[e.AssertDestructuringAssignment=1024]="AssertDestructuringAssignment",e[e.OuterExpressionExcludes=536870912]="OuterExpressionExcludes",e[e.PropertyAccessExcludes=536870912]="PropertyAccessExcludes",e[e.NodeExcludes=536870912]="NodeExcludes",e[e.ArrowFunctionExcludes=547309568]="ArrowFunctionExcludes",e[e.FunctionExcludes=547313664]="FunctionExcludes",e[e.ConstructorExcludes=547311616]="ConstructorExcludes",e[e.MethodOrAccessorExcludes=538923008]="MethodOrAccessorExcludes",e[e.PropertyExcludes=536875008]="PropertyExcludes",e[e.ClassExcludes=536905728]="ClassExcludes",e[e.ModuleExcludes=546379776]="ModuleExcludes",e[e.TypeExcludes=-2]="TypeExcludes",e[e.ObjectLiteralExcludes=536922112]="ObjectLiteralExcludes",e[e.ArrayLiteralOrCallOrNewExcludes=536879104]="ArrayLiteralOrCallOrNewExcludes",e[e.VariableDeclarationListExcludes=537018368]="VariableDeclarationListExcludes",e[e.ParameterExcludes=536870912]="ParameterExcludes",e[e.CatchClauseExcludes=536887296]="CatchClauseExcludes",e[e.BindingPatternExcludes=536879104]="BindingPatternExcludes",e[e.PropertyNamePropagatingFlags=4096]="PropertyNamePropagatingFlags"}(e.TransformFlags||(e.TransformFlags={})),function(e){e[e.None=0]="None",e[e.SingleLine=1]="SingleLine",e[e.AdviseOnEmitNode=2]="AdviseOnEmitNode",e[e.NoSubstitution=4]="NoSubstitution",e[e.CapturesThis=8]="CapturesThis",e[e.NoLeadingSourceMap=16]="NoLeadingSourceMap",e[e.NoTrailingSourceMap=32]="NoTrailingSourceMap",e[e.NoSourceMap=48]="NoSourceMap",e[e.NoNestedSourceMaps=64]="NoNestedSourceMaps",e[e.NoTokenLeadingSourceMaps=128]="NoTokenLeadingSourceMaps",e[e.NoTokenTrailingSourceMaps=256]="NoTokenTrailingSourceMaps",e[e.NoTokenSourceMaps=384]="NoTokenSourceMaps",e[e.NoLeadingComments=512]="NoLeadingComments",e[e.NoTrailingComments=1024]="NoTrailingComments",e[e.NoComments=1536]="NoComments",e[e.NoNestedComments=2048]="NoNestedComments",e[e.HelperName=4096]="HelperName",e[e.ExportName=8192]="ExportName",e[e.LocalName=16384]="LocalName",e[e.InternalName=32768]="InternalName",e[e.Indented=65536]="Indented",e[e.NoIndentation=131072]="NoIndentation",e[e.AsyncFunctionBody=262144]="AsyncFunctionBody",e[e.ReuseTempVariableScope=524288]="ReuseTempVariableScope",e[e.CustomPrologue=1048576]="CustomPrologue",e[e.NoHoisting=2097152]="NoHoisting",e[e.HasEndOfDeclarationMarker=4194304]="HasEndOfDeclarationMarker",e[e.Iterator=8388608]="Iterator",e[e.NoAsciiEscaping=16777216]="NoAsciiEscaping",e[e.TypeScriptClassWrapper=33554432]="TypeScriptClassWrapper",e[e.NeverApplyImportHelper=67108864]="NeverApplyImportHelper",e[e.IgnoreSourceNewlines=134217728]="IgnoreSourceNewlines"}(e.EmitFlags||(e.EmitFlags={})),function(e){e[e.Extends=1]="Extends",e[e.Assign=2]="Assign",e[e.Rest=4]="Rest",e[e.Decorate=8]="Decorate",e[e.Metadata=16]="Metadata",e[e.Param=32]="Param",e[e.Awaiter=64]="Awaiter",e[e.Generator=128]="Generator",e[e.Values=256]="Values",e[e.Read=512]="Read",e[e.Spread=1024]="Spread",e[e.SpreadArrays=2048]="SpreadArrays",e[e.Await=4096]="Await",e[e.AsyncGenerator=8192]="AsyncGenerator",e[e.AsyncDelegator=16384]="AsyncDelegator",e[e.AsyncValues=32768]="AsyncValues",e[e.ExportStar=65536]="ExportStar",e[e.ImportStar=131072]="ImportStar",e[e.ImportDefault=262144]="ImportDefault",e[e.MakeTemplateObject=524288]="MakeTemplateObject",e[e.ClassPrivateFieldGet=1048576]="ClassPrivateFieldGet",e[e.ClassPrivateFieldSet=2097152]="ClassPrivateFieldSet",e[e.CreateBinding=4194304]="CreateBinding",e[e.FirstEmitHelper=1]="FirstEmitHelper",e[e.LastEmitHelper=4194304]="LastEmitHelper",e[e.ForOfIncludes=256]="ForOfIncludes",e[e.ForAwaitOfIncludes=32768]="ForAwaitOfIncludes",e[e.AsyncGeneratorIncludes=12288]="AsyncGeneratorIncludes",e[e.AsyncDelegatorIncludes=53248]="AsyncDelegatorIncludes",e[e.SpreadIncludes=1536]="SpreadIncludes"}(e.ExternalEmitHelpers||(e.ExternalEmitHelpers={})),function(e){e[e.SourceFile=0]="SourceFile",e[e.Expression=1]="Expression",e[e.IdentifierName=2]="IdentifierName",e[e.MappedTypeParameter=3]="MappedTypeParameter",e[e.Unspecified=4]="Unspecified",e[e.EmbeddedStatement=5]="EmbeddedStatement",e[e.JsxAttributeValue=6]="JsxAttributeValue"}(e.EmitHint||(e.EmitHint={})),function(e){e[e.Parentheses=1]="Parentheses",e[e.TypeAssertions=2]="TypeAssertions",e[e.NonNullAssertions=4]="NonNullAssertions",e[e.PartiallyEmittedExpressions=8]="PartiallyEmittedExpressions",e[e.Assertions=6]="Assertions",e[e.All=15]="All"}(e.OuterExpressionKinds||(e.OuterExpressionKinds={})),function(e){e[e.None=0]="None",e[e.InParameters=1]="InParameters",e[e.VariablesHoistedInParameters=2]="VariablesHoistedInParameters"}(e.LexicalEnvironmentFlags||(e.LexicalEnvironmentFlags={})),function(e){e.Prologue="prologue",e.EmitHelpers="emitHelpers",e.NoDefaultLib="no-default-lib",e.Reference="reference",e.Type="type",e.Lib="lib",e.Prepend="prepend",e.Text="text",e.Internal="internal"}(e.BundleFileSectionKind||(e.BundleFileSectionKind={})),function(e){e[e.None=0]="None",e[e.SingleLine=0]="SingleLine",e[e.MultiLine=1]="MultiLine",e[e.PreserveLines=2]="PreserveLines",e[e.LinesMask=3]="LinesMask",e[e.NotDelimited=0]="NotDelimited",e[e.BarDelimited=4]="BarDelimited",e[e.AmpersandDelimited=8]="AmpersandDelimited",e[e.CommaDelimited=16]="CommaDelimited",e[e.AsteriskDelimited=32]="AsteriskDelimited",e[e.DelimitersMask=60]="DelimitersMask",e[e.AllowTrailingComma=64]="AllowTrailingComma",e[e.Indented=128]="Indented",e[e.SpaceBetweenBraces=256]="SpaceBetweenBraces",e[e.SpaceBetweenSiblings=512]="SpaceBetweenSiblings",e[e.Braces=1024]="Braces",e[e.Parenthesis=2048]="Parenthesis",e[e.AngleBrackets=4096]="AngleBrackets",e[e.SquareBrackets=8192]="SquareBrackets",e[e.BracketsMask=15360]="BracketsMask",e[e.OptionalIfUndefined=16384]="OptionalIfUndefined",e[e.OptionalIfEmpty=32768]="OptionalIfEmpty",e[e.Optional=49152]="Optional",e[e.PreferNewLine=65536]="PreferNewLine",e[e.NoTrailingNewLine=131072]="NoTrailingNewLine",e[e.NoInterveningComments=262144]="NoInterveningComments",e[e.NoSpaceIfEmpty=524288]="NoSpaceIfEmpty",e[e.SingleElement=1048576]="SingleElement",e[e.SpaceAfterList=2097152]="SpaceAfterList",e[e.Modifiers=262656]="Modifiers",e[e.HeritageClauses=512]="HeritageClauses",e[e.SingleLineTypeLiteralMembers=768]="SingleLineTypeLiteralMembers",e[e.MultiLineTypeLiteralMembers=32897]="MultiLineTypeLiteralMembers",e[e.SingleLineTupleTypeElements=528]="SingleLineTupleTypeElements",e[e.MultiLineTupleTypeElements=657]="MultiLineTupleTypeElements",e[e.UnionTypeConstituents=516]="UnionTypeConstituents",e[e.IntersectionTypeConstituents=520]="IntersectionTypeConstituents",e[e.ObjectBindingPatternElements=525136]="ObjectBindingPatternElements",e[e.ArrayBindingPatternElements=524880]="ArrayBindingPatternElements",e[e.ObjectLiteralExpressionProperties=526226]="ObjectLiteralExpressionProperties",e[e.ArrayLiteralExpressionElements=8914]="ArrayLiteralExpressionElements",e[e.CommaListElements=528]="CommaListElements",e[e.CallExpressionArguments=2576]="CallExpressionArguments",e[e.NewExpressionArguments=18960]="NewExpressionArguments",e[e.TemplateExpressionSpans=262144]="TemplateExpressionSpans",e[e.SingleLineBlockStatements=768]="SingleLineBlockStatements",e[e.MultiLineBlockStatements=129]="MultiLineBlockStatements",e[e.VariableDeclarationList=528]="VariableDeclarationList",e[e.SingleLineFunctionBodyStatements=768]="SingleLineFunctionBodyStatements",e[e.MultiLineFunctionBodyStatements=1]="MultiLineFunctionBodyStatements",e[e.ClassHeritageClauses=0]="ClassHeritageClauses",e[e.ClassMembers=129]="ClassMembers",e[e.InterfaceMembers=129]="InterfaceMembers",e[e.EnumMembers=145]="EnumMembers",e[e.CaseBlockClauses=129]="CaseBlockClauses",e[e.NamedImportsOrExportsElements=525136]="NamedImportsOrExportsElements",e[e.JsxElementOrFragmentChildren=262144]="JsxElementOrFragmentChildren",e[e.JsxElementAttributes=262656]="JsxElementAttributes",e[e.CaseOrDefaultClauseStatements=163969]="CaseOrDefaultClauseStatements",e[e.HeritageClauseTypes=528]="HeritageClauseTypes",e[e.SourceFileStatements=131073]="SourceFileStatements",e[e.Decorators=2146305]="Decorators",e[e.TypeArguments=53776]="TypeArguments",e[e.TypeParameters=53776]="TypeParameters",e[e.Parameters=2576]="Parameters",e[e.IndexSignatureParameters=8848]="IndexSignatureParameters",e[e.JSDocComment=33]="JSDocComment"}(e.ListFormat||(e.ListFormat={})),function(e){e[e.None=0]="None",e[e.TripleSlashXML=1]="TripleSlashXML",e[e.SingleLine=2]="SingleLine",e[e.MultiLine=4]="MultiLine",e[e.All=7]="All",e[e.Default=7]="Default"}(e.PragmaKindFlags||(e.PragmaKindFlags={})),e.commentPragmas={reference:{args:[{name:"types",optional:!0,captureSpan:!0},{name:"lib",optional:!0,captureSpan:!0},{name:"path",optional:!0,captureSpan:!0},{name:"no-default-lib",optional:!0}],kind:1},"amd-dependency":{args:[{name:"path"},{name:"name",optional:!0}],kind:1},"amd-module":{args:[{name:"name"}],kind:1},"ts-check":{kind:2},"ts-nocheck":{kind:2},jsx:{args:[{name:"factory"}],kind:4},jsxfrag:{args:[{name:"factory"}],kind:4},jsximportsource:{args:[{name:"factory"}],kind:4},jsxruntime:{args:[{name:"factory"}],kind:4}}}(ts||(ts={})),function(e){e.directorySeparator="/";var n=/\\/g;function t(e){return 47===e||92===e}function r(e){return l(e)>0}function a(e){return 0!==l(e)}function i(e){return/^\.\.?($|[\\/])/.test(e)}function o(n,t){return n.length>t.length&&e.endsWith(n,t)}function s(e){return e.length>0&&t(e.charCodeAt(e.length-1))}function c(e){return e>=97&&e<=122||e>=65&&e<=90}function l(n){if(!n)return 0;var t=n.charCodeAt(0);if(47===t||92===t){if(n.charCodeAt(1)!==t)return 1;var r=n.indexOf(47===t?e.directorySeparator:"\\",2);return r<0?n.length:r+1}if(c(t)&&58===n.charCodeAt(1)){var a=n.charCodeAt(2);if(47===a||92===a)return 3;if(2===n.length)return 2}var i=n.indexOf("://");if(-1!==i){var o=i+"://".length,s=n.indexOf(e.directorySeparator,o);if(-1!==s){var l=n.slice(0,i),u=n.slice(o,s);if("file"===l&&(""===u||"localhost"===u)&&c(n.charCodeAt(s+1))){var d=function(e,n){var t=e.charCodeAt(n);if(58===t)return n+1;if(37===t&&51===e.charCodeAt(n+1)){var r=e.charCodeAt(n+2);if(97===r||65===r)return n+3}return-1}(n,s+2);if(-1!==d){if(47===n.charCodeAt(d))return~(d+1);if(d===n.length)return~d}}return~(s+1)}return~n.length}return 0}function u(e){var n=l(e);return n<0?~n:n}function d(n){var t=u(n=y(n));return t===n.length?n:(n=x(n)).slice(0,Math.max(t,n.lastIndexOf(e.directorySeparator)))}function p(n,t,r){if(u(n=y(n))===n.length)return"";var a=(n=x(n)).slice(Math.max(u(n),n.lastIndexOf(e.directorySeparator)+1)),i=void 0!==t&&void 0!==r?f(a,t,r):void 0;return i?a.slice(0,a.length-i.length):a}function m(n,t,r){if(e.startsWith(t,".")||(t="."+t),n.length>=t.length&&46===n.charCodeAt(n.length-t.length)){var a=n.slice(n.length-t.length);if(r(a,t))return a}}function f(n,t,r){if(t)return function(e,n,t){if("string"==typeof n)return m(e,n,t)||"";for(var r=0,a=n;r<a.length;r++){var i=m(e,a[r],t);if(i)return i}return""}(x(n),t,r?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive);var a=p(n),i=a.lastIndexOf(".");return i>=0?a.substring(i):""}function _(n,t){return void 0===t&&(t=""),function(n,t){var r=n.substring(0,t),a=n.substring(t).split(e.directorySeparator);return a.length&&!e.lastOrUndefined(a)&&a.pop(),__spreadArrays([r],a)}(n=h(t,n),u(n))}function g(n){return 0===n.length?"":(n[0]&&L(n[0]))+n.slice(1).join(e.directorySeparator)}function y(t){return t.replace(n,e.directorySeparator)}function v(n){if(!e.some(n))return[];for(var t=[n[0]],r=1;r<n.length;r++){var a=n[r];if(a&&"."!==a){if(".."===a)if(t.length>1){if(".."!==t[t.length-1]){t.pop();continue}}else if(t[0])continue;t.push(a)}}return t}function h(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];e&&(e=y(e));for(var r=0,a=n;r<a.length;r++){var i=a[r];i&&(i=y(i),e=e&&0===u(i)?L(e)+i:i)}return e}function b(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];return S(e.some(t)?h.apply(void 0,__spreadArrays([n],t)):y(n))}function E(e,n){return v(_(e,n))}function T(e,n){return g(E(e,n))}function S(e){var n=g(v(_(e=y(e))));return n&&s(e)?L(n):n}function x(e){return s(e)?e.substr(0,e.length-1):e}function L(n){return s(n)?n:n+e.directorySeparator}function A(e){return a(e)||i(e)?e:"./"+e}e.isAnyDirectorySeparator=t,e.isUrl=function(e){return l(e)<0},e.isRootedDiskPath=r,e.isDiskPathRoot=function(e){var n=l(e);return n>0&&n===e.length},e.pathIsAbsolute=a,e.pathIsRelative=i,e.pathIsBareSpecifier=function(e){return!a(e)&&!i(e)},e.hasExtension=function(n){return e.stringContains(p(n),".")},e.fileExtensionIs=o,e.fileExtensionIsOneOf=function(e,n){for(var t=0,r=n;t<r.length;t++){if(o(e,r[t]))return!0}return!1},e.hasTrailingDirectorySeparator=s,e.getRootLength=u,e.getDirectoryPath=d,e.getBaseFileName=p,e.getAnyExtensionFromPath=f,e.getPathComponents=_,e.getPathFromPathComponents=g,e.normalizeSlashes=y,e.reducePathComponents=v,e.combinePaths=h,e.resolvePath=b,e.getNormalizedPathComponents=E,e.getNormalizedAbsolutePath=T,e.normalizePath=S,e.getNormalizedAbsolutePathWithoutRoot=function(n,t){return function(n){return 0===n.length?"":n.slice(1).join(e.directorySeparator)}(E(n,t))},e.toPath=function(e,n,t){return t(r(e)?S(e):T(e,n))},e.normalizePathAndParts=function(n){var t=v(_(n=y(n))),r=t[0],a=t.slice(1);if(a.length){var i=r+a.join(e.directorySeparator);return{path:s(n)?L(i):i,parts:a}}return{path:r,parts:a}},e.removeTrailingDirectorySeparator=x,e.ensureTrailingDirectorySeparator=L,e.ensurePathIsNonModuleName=A,e.changeAnyExtension=function(n,t,r,a){var i=void 0!==r&&void 0!==a?f(n,r,a):f(n);return i?n.slice(0,n.length-i.length)+(e.startsWith(t,".")?t:"."+t):n};var D=/(^|\/)\.{0,2}($|\/)/;function C(n,t,r){if(n===t)return 0;if(void 0===n)return-1;if(void 0===t)return 1;var a=n.substring(0,u(n)),i=t.substring(0,u(t)),o=e.compareStringsCaseInsensitive(a,i);if(0!==o)return o;var s=n.substring(a.length),c=t.substring(i.length);if(!D.test(s)&&!D.test(c))return r(s,c);for(var l=v(_(n)),d=v(_(t)),p=Math.min(l.length,d.length),m=1;m<p;m++){var f=r(l[m],d[m]);if(0!==f)return f}return e.compareValues(l.length,d.length)}function I(n,t,r,a){var i,o=v(_(n)),s=v(_(t));for(i=0;i<o.length&&i<s.length;i++){var c=a(o[i]),l=a(s[i]);if(!(0===i?e.equateStringsCaseInsensitive:r)(c,l))break}if(0===i)return s;for(var u=s.slice(i),d=[];i<o.length;i++)d.push("..");return __spreadArrays([""],d,u)}function N(n,t,r){e.Debug.assert(u(n)>0==u(t)>0,"Paths must either both be absolute or both be relative");var a="function"==typeof r?r:e.identity;return g(I(n,t,"boolean"==typeof r&&r?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive,a))}function R(n,t,a,i,o){var s=I(b(a,n),b(a,t),e.equateStringsCaseSensitive,i),c=s[0];if(o&&r(c)){var l=c.charAt(0)===e.directorySeparator?"file://":"file:///";s[0]=l+c}return g(s)}e.comparePathsCaseSensitive=function(n,t){return C(n,t,e.compareStringsCaseSensitive)},e.comparePathsCaseInsensitive=function(n,t){return C(n,t,e.compareStringsCaseInsensitive)},e.comparePaths=function(n,t,r,a){return"string"==typeof r?(n=h(r,n),t=h(r,t)):"boolean"==typeof r&&(a=r),C(n,t,e.getStringComparer(a))},e.containsPath=function(n,t,r,a){if("string"==typeof r?(n=h(r,n),t=h(r,t)):"boolean"==typeof r&&(a=r),void 0===n||void 0===t)return!1;if(n===t)return!0;var i=v(_(n)),o=v(_(t));if(o.length<i.length)return!1;for(var s=a?e.equateStringsCaseInsensitive:e.equateStringsCaseSensitive,c=0;c<i.length;c++){if(!(0===c?e.equateStringsCaseInsensitive:s)(i[c],o[c]))return!1}return!0},e.startsWithDirectory=function(n,t,r){var a=r(n),i=r(t);return e.startsWith(a,i+"/")||e.startsWith(a,i+"\\")},e.getPathComponentsRelativeTo=I,e.getRelativePathFromDirectory=N,e.convertToRelativePath=function(e,n,t){return r(e)?R(n,e,n,t,!1):e},e.getRelativePathFromFile=function(e,n,t){return A(N(d(e),n,t))},e.getRelativePathToDirectoryOrUrl=R,e.forEachAncestorDirectory=function(e,n){for(;;){var t=n(e);if(void 0!==t)return t;var r=d(e);if(r===e)return;e=r}},e.isNodeModulesDirectory=function(n){return e.endsWith(n,"/node_modules")}}(ts||(ts={})),function(e){var n,t;function r(e){var n;return(n={})[t.Low]=e.Low,n[t.Medium]=e.Medium,n[t.High]=e.High,n}e.generateDjb2Hash=function(e){for(var n=5381,t=0;t<e.length;t++)n=(n<<5)+n+e.charCodeAt(t);return n.toString()},e.setStackTraceLimit=function(){Error.stackTraceLimit<100&&(Error.stackTraceLimit=100)},function(e){e[e.Created=0]="Created",e[e.Changed=1]="Changed",e[e.Deleted=2]="Deleted"}(n=e.FileWatcherEventKind||(e.FileWatcherEventKind={})),function(e){e[e.High=2e3]="High",e[e.Medium=500]="Medium",e[e.Low=250]="Low"}(t=e.PollingInterval||(e.PollingInterval={})),e.missingFileModifiedTime=new Date(0);var a={Low:32,Medium:64,High:256},i=r(a);function o(n){if(n.getEnvironmentVariable){var o=function(e,n){var t=s(e);if(t)return r("Low"),r("Medium"),r("High"),!0;return!1;function r(e){n[e]=t[e]||n[e]}}("TSC_WATCH_POLLINGINTERVAL",t);i=c("TSC_WATCH_POLLINGCHUNKSIZE",a)||i,e.unchangedPollThresholds=c("TSC_WATCH_UNCHANGEDPOLLTHRESHOLDS",a)||e.unchangedPollThresholds}function s(e){var t;return r("Low"),r("Medium"),r("High"),t;function r(r){var a=function(e,t){return n.getEnvironmentVariable(e+"_"+t.toUpperCase())}(e,r);a&&((t||(t={}))[r]=Number(a))}}function c(e,n){var t=s(e);return(o||t)&&r(t?__assign(__assign({},n),t):n)}}function s(n){var r=[],a=[],o=u(t.Low),s=u(t.Medium),c=u(t.High);return function(n,t,a){var i={fileName:n,callback:t,unchangedPolls:0,mtime:h(n)};return r.push(i),_(i,a),{close:function(){i.isClosed=!0,e.unorderedRemoveItem(r,i)}}};function u(e){var n=[];return n.pollingInterval=e,n.pollIndex=0,n.pollScheduled=!1,n}function d(n){n.pollIndex=m(n,n.pollingInterval,n.pollIndex,i[n.pollingInterval]),n.length?v(n.pollingInterval):(e.Debug.assert(0===n.pollIndex),n.pollScheduled=!1)}function p(e){m(a,t.Low,0,a.length),d(e),!e.pollScheduled&&a.length&&v(t.Low)}function m(n,r,i,o){for(var s=n.length,c=i,u=0;u<o&&s>0;m(),s--){var d=n[i];if(d)if(d.isClosed)n[i]=void 0;else{u++;var p=l(d,h(d.fileName));d.isClosed?n[i]=void 0:p?(d.unchangedPolls=0,n!==a&&(n[i]=void 0,g(d))):d.unchangedPolls!==e.unchangedPollThresholds[r]?d.unchangedPolls++:n===a?(d.unchangedPolls=1,n[i]=void 0,_(d,t.Low)):r!==t.High&&(d.unchangedPolls++,n[i]=void 0,_(d,r===t.Low?t.Medium:t.High)),n[i]&&(c<i&&(n[c]=d,n[i]=void 0),c++)}}return i;function m(){++i===n.length&&(c<i&&(n.length=c),i=0,c=0)}}function f(e){switch(e){case t.Low:return o;case t.Medium:return s;case t.High:return c}}function _(e,n){f(n).push(e),y(n)}function g(e){a.push(e),y(t.Low)}function y(e){f(e).pollScheduled||v(e)}function v(e){f(e).pollScheduled=n.setTimeout(e===t.Low?p:d,e,f(e))}function h(t){return n.getModifiedTime(t)||e.missingFileModifiedTime}}function c(r,a){var i=e.createMultiMap(),o=new e.Map,s=e.createGetCanonicalFileName(a);return function(a,c,l,u){var d=s(a);i.add(d,c);var p=e.getDirectoryPath(d)||".",m=o.get(p)||function(a,c,l){var u=r(a,1,(function(t,r){if(e.isString(r)){var o=e.getNormalizedAbsolutePath(r,a),c=o&&i.get(s(o));if(c)for(var l=0,u=c;l<u.length;l++){(0,u[l])(o,n.Changed)}}}),!1,t.Medium,l);return u.referenceCount=0,o.set(c,u),u}(e.getDirectoryPath(a)||".",p,u);return m.referenceCount++,{close:function(){1===m.referenceCount?(m.close(),o.delete(p)):m.referenceCount--,i.remove(d,c)}}}}function l(e,n){var t=e.mtime.getTime(),r=n.getTime();return t!==r&&(e.mtime=n,e.callback(e.fileName,u(t,r)),!0)}function u(e,t){return 0===e?n.Created:0===t?n.Deleted:n.Changed}function d(n){var t,r=new e.Map,a=e.createMultiMap(),i=new e.Map,o=e.getStringComparer(!n.useCaseSensitiveFileNames),s=e.createGetCanonicalFileName(n.useCaseSensitiveFileNames);return function(e,t,r,a){return r?c(e,a,t):n.watchDirectory(e,t,r,a)};function c(o,c,f){var _=s(o),g=r.get(_);g?g.refCount++:(g={watcher:n.watchDirectory(o,(function(e){m(e)||((null==c?void 0:c.synchronousWatchDirectory)?(l(_,e),p(o,_,c)):function(e,a,o,s){var c=r.get(a);if(c&&n.directoryExists(e))return void function(e,r,a,o){var s=i.get(r);s?s.fileNames.push(a):i.set(r,{dirName:e,options:o,fileNames:[a]});t&&(n.clearTimeout(t),t=void 0);t=n.setTimeout(u,1e3)}(e,a,o,s);l(a,o),d(c)}(o,_,e,c))}),!1,c),refCount:1,childWatches:e.emptyArray},r.set(_,g),p(o,_,c));var y=f&&{dirName:o,callback:f};return y&&a.add(_,y),{dirName:o,close:function(){var n=e.Debug.checkDefined(r.get(_));y&&a.remove(_,y),n.refCount--,n.refCount||(r.delete(_),e.closeFileWatcherOf(n),n.childWatches.forEach(e.closeFileWatcher))}}}function l(n,t,r){var i,o;e.isString(t)?i=t:o=t,a.forEach((function(t,a){var s;if((!o||!0!==o.get(a))&&(a===n||e.startsWith(n,a)&&n[a.length]===e.directorySeparator))if(o)if(r){var c=o.get(a);c?(s=c).push.apply(s,r):o.set(a,r.slice())}else o.set(a,!0);else t.forEach((function(e){return(0,e.callback)(i)}))}))}function u(){t=void 0,e.sysLog("sysLog:: onTimerToUpdateChildWatches:: "+i.size);for(var n=e.timestamp(),r=new e.Map;!t&&i.size;){var o=i.entries().next(),s=o.value,c=s[0],u=s[1],d=u.dirName,m=u.options,f=u.fileNames,_=o.done;e.Debug.assert(!_),i.delete(c);var g=p(d,c,m);l(c,r,g?void 0:f)}e.sysLog("sysLog:: invokingWatchers:: "+(e.timestamp()-n)+"ms:: "+i.size),a.forEach((function(n,t){var a=r.get(t);a&&n.forEach((function(n){var t=n.callback,r=n.dirName;e.isArray(a)?a.forEach(t):t(r)}))}));var y=e.timestamp()-n;e.sysLog("sysLog:: Elapsed "+y+"ms:: onTimerToUpdateChildWatches:: "+i.size+" "+t)}function d(n){if(n){var t=n.childWatches;n.childWatches=e.emptyArray;for(var a=0,i=t;a<i.length;a++){var o=i[a];o.close(),d(r.get(s(o.dirName)))}}}function p(t,a,i){var s,l=r.get(a);if(!l)return!1;var u=e.enumerateInsertsAndDeletes(n.directoryExists(t)?e.mapDefined(n.getAccessibleSortedChildDirectories(t),(function(r){var a=e.getNormalizedAbsolutePath(r,t);return m(a)||0!==o(a,e.normalizePath(n.realpath(a)))?void 0:a})):e.emptyArray,l.childWatches,(function(e,n){return o(e,n.dirName)}),(function(e){d(c(e,i))}),e.closeFileWatcher,d);return l.childWatches=s||e.emptyArray,u;function d(e){(s||(s=[])).push(e)}}function m(t){return e.some(e.ignoredPaths,(function(r){return function(t,r){return!!e.stringContains(t,r)||!n.useCaseSensitiveFileNames&&e.stringContains(s(t),r)}(t,r)}))}}function p(n,t){return function(r,a){"rename"===r&&t(a?e.normalizePath(e.combinePaths(n,a)):n)}}e.unchangedPollThresholds=r(a),e.setCustomPollingValues=o,e.createDynamicPriorityPollingWatchFile=s,e.createSingleFileWatcherPerName=function(n,t){var r=new e.Map,a=e.createMultiMap(),i=e.createGetCanonicalFileName(t);return function(t,o,s,c){var l=i(t),u=r.get(l);return u?u.refCount++:r.set(l,{watcher:n(t,(function(n,t){return e.forEach(a.get(l),(function(e){return e(n,t)}))}),s,c),refCount:1}),a.add(l,o),{close:function(){var n=e.Debug.checkDefined(r.get(l));a.remove(l,o),n.refCount--,n.refCount||(r.delete(l),e.closeFileWatcherOf(n))}}}},e.onWatchedFileStat=l,e.getFileWatcherEventKind=u,e.ignoredPaths=["/node_modules/.","/.git","/.#"],e.sysLog=e.noop,e.setSysLog=function(n){e.sysLog=n},e.createDirectoryWatcherSupportingRecursive=d,function(e){e[e.File=0]="File",e[e.Directory=1]="Directory"}(e.FileSystemEntryKind||(e.FileSystemEntryKind={})),e.createFileWatcherCallback=function(e){return function(t,r){return e(r===n.Changed?"change":"rename","")}},e.createSystemWatchFunctions=function(r){var a,i,o,l=r.pollingWatchFile,u=r.getModifiedTime,m=r.setTimeout,f=r.clearTimeout,_=r.fsWatch,g=r.fileExists,y=r.useCaseSensitiveFileNames,v=r.fsSupportsRecursiveFsWatch,h=r.directoryExists,b=r.getAccessibleSortedChildDirectories,E=r.realpath,T=r.tscWatchFile,S=r.useNonPollingWatchers,x=r.tscWatchDirectory;return{watchFile:function(r,a,o,s){s=function(n,t){if(n&&void 0!==n.watchFile)return n;switch(T){case"PriorityPollingInterval":return{watchFile:e.WatchFileKind.PriorityPollingInterval};case"DynamicPriorityPolling":return{watchFile:e.WatchFileKind.DynamicPriorityPolling};case"UseFsEvents":return A(e.WatchFileKind.UseFsEvents,e.PollingWatchKind.PriorityInterval,n);case"UseFsEventsWithFallbackDynamicPolling":return A(e.WatchFileKind.UseFsEvents,e.PollingWatchKind.DynamicPriority,n);case"UseFsEventsOnParentDirectory":t=!0;default:return t?A(e.WatchFileKind.UseFsEventsOnParentDirectory,e.PollingWatchKind.PriorityInterval,n):{watchFile:e.WatchFileKind.FixedPollingInterval}}}(s,S);var u=e.Debug.checkDefined(s.watchFile);switch(u){case e.WatchFileKind.FixedPollingInterval:return l(r,a,t.Low,void 0);case e.WatchFileKind.PriorityPollingInterval:return l(r,a,o,void 0);case e.WatchFileKind.DynamicPriorityPolling:return L()(r,a,o,void 0);case e.WatchFileKind.UseFsEvents:return _(r,0,function(e,t,r){return function(a){t(e,"rename"===a?r(e)?n.Created:n.Deleted:n.Changed)}}(r,a,g),!1,o,e.getFallbackOptions(s));case e.WatchFileKind.UseFsEventsOnParentDirectory:return i||(i=c(_,y)),i(r,a,o,e.getFallbackOptions(s));default:e.Debug.assertNever(u)}},watchDirectory:function(n,r,a,i){if(v)return _(n,1,p(n,r),a,t.Medium,e.getFallbackOptions(i));o||(o=d({useCaseSensitiveFileNames:y,directoryExists:h,getAccessibleSortedChildDirectories:b,watchDirectory:D,realpath:E,setTimeout:m,clearTimeout:f}));return o(n,r,a,i)}};function L(){return a||(a=s({getModifiedTime:u,setTimeout:m}))}function A(e,n,t){var r=null==t?void 0:t.fallbackPolling;return{watchFile:e,fallbackPolling:void 0===r?n:r}}function D(n,r,a,i){e.Debug.assert(!a),i=function(n){if(n&&void 0!==n.watchDirectory)return n;switch(x){case"RecursiveDirectoryUsingFsWatchFile":return{watchDirectory:e.WatchDirectoryKind.FixedPollingInterval};case"RecursiveDirectoryUsingDynamicPriorityPolling":return{watchDirectory:e.WatchDirectoryKind.DynamicPriorityPolling};default:var t=null==n?void 0:n.fallbackPolling;return{watchDirectory:e.WatchDirectoryKind.UseFsEvents,fallbackPolling:void 0!==t?t:void 0}}}(i);var o=e.Debug.checkDefined(i.watchDirectory);switch(o){case e.WatchDirectoryKind.FixedPollingInterval:return l(n,(function(){return r(n)}),t.Medium,void 0);case e.WatchDirectoryKind.DynamicPriorityPolling:return L()(n,(function(){return r(n)}),t.Medium,void 0);case e.WatchDirectoryKind.UseFsEvents:return _(n,1,p(n,r),a,t.Medium,e.getFallbackOptions(i));default:e.Debug.assertNever(o)}}},e.patchWriteFileEnsuringDirectory=function(n){var t=n.writeFile;n.writeFile=function(r,a,i){return e.writeFileEnsuringDirectories(r,a,!!i,(function(e,r,a){return t.call(n,e,r,a)}),(function(e){return n.createDirectory(e)}),(function(e){return n.directoryExists(e)}))}},e.getNodeMajorVersion=function(){if("undefined"!=typeof process){var e=process.version;if(e){var n=e.indexOf(".");if(-1!==n)return parseInt(e.substring(1,n))}}},e.sys=void 0,e.sys&&e.sys.getEnvironmentVariable&&(o(e.sys),e.Debug.setAssertionLevel(/^development$/i.test(e.sys.getEnvironmentVariable("NODE_ENV"))?1:0)),e.sys&&e.sys.debugMode&&(e.Debug.isDebugging=!0)}(ts||(ts={})),function(e){function n(e,n,t,r,a,i,o){return{code:e,category:n,key:t,message:r,reportsUnnecessary:a,elidedInCompatabilityPyramid:i,reportsDeprecated:o}}e.Diagnostics={Unterminated_string_literal:n(1002,e.DiagnosticCategory.Error,"Unterminated_string_literal_1002","Unterminated string literal."),Identifier_expected:n(1003,e.DiagnosticCategory.Error,"Identifier_expected_1003","Identifier expected."),_0_expected:n(1005,e.DiagnosticCategory.Error,"_0_expected_1005","'{0}' expected."),A_file_cannot_have_a_reference_to_itself:n(1006,e.DiagnosticCategory.Error,"A_file_cannot_have_a_reference_to_itself_1006","A file cannot have a reference to itself."),The_parser_expected_to_find_a_to_match_the_token_here:n(1007,e.DiagnosticCategory.Error,"The_parser_expected_to_find_a_to_match_the_token_here_1007","The parser expected to find a '}' to match the '{' token here."),Trailing_comma_not_allowed:n(1009,e.DiagnosticCategory.Error,"Trailing_comma_not_allowed_1009","Trailing comma not allowed."),Asterisk_Slash_expected:n(1010,e.DiagnosticCategory.Error,"Asterisk_Slash_expected_1010","'*/' expected."),An_element_access_expression_should_take_an_argument:n(1011,e.DiagnosticCategory.Error,"An_element_access_expression_should_take_an_argument_1011","An element access expression should take an argument."),Unexpected_token:n(1012,e.DiagnosticCategory.Error,"Unexpected_token_1012","Unexpected token."),A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma:n(1013,e.DiagnosticCategory.Error,"A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma_1013","A rest parameter or binding pattern may not have a trailing comma."),A_rest_parameter_must_be_last_in_a_parameter_list:n(1014,e.DiagnosticCategory.Error,"A_rest_parameter_must_be_last_in_a_parameter_list_1014","A rest parameter must be last in a parameter list."),Parameter_cannot_have_question_mark_and_initializer:n(1015,e.DiagnosticCategory.Error,"Parameter_cannot_have_question_mark_and_initializer_1015","Parameter cannot have question mark and initializer."),A_required_parameter_cannot_follow_an_optional_parameter:n(1016,e.DiagnosticCategory.Error,"A_required_parameter_cannot_follow_an_optional_parameter_1016","A required parameter cannot follow an optional parameter."),An_index_signature_cannot_have_a_rest_parameter:n(1017,e.DiagnosticCategory.Error,"An_index_signature_cannot_have_a_rest_parameter_1017","An index signature cannot have a rest parameter."),An_index_signature_parameter_cannot_have_an_accessibility_modifier:n(1018,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_an_accessibility_modifier_1018","An index signature parameter cannot have an accessibility modifier."),An_index_signature_parameter_cannot_have_a_question_mark:n(1019,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_a_question_mark_1019","An index signature parameter cannot have a question mark."),An_index_signature_parameter_cannot_have_an_initializer:n(1020,e.DiagnosticCategory.Error,"An_index_signature_parameter_cannot_have_an_initializer_1020","An index signature parameter cannot have an initializer."),An_index_signature_must_have_a_type_annotation:n(1021,e.DiagnosticCategory.Error,"An_index_signature_must_have_a_type_annotation_1021","An index signature must have a type annotation."),An_index_signature_parameter_must_have_a_type_annotation:n(1022,e.DiagnosticCategory.Error,"An_index_signature_parameter_must_have_a_type_annotation_1022","An index signature parameter must have a type annotation."),An_index_signature_parameter_type_must_be_either_string_or_number:n(1023,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_must_be_either_string_or_number_1023","An index signature parameter type must be either 'string' or 'number'."),readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature:n(1024,e.DiagnosticCategory.Error,"readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature_1024","'readonly' modifier can only appear on a property declaration or index signature."),An_index_signature_cannot_have_a_trailing_comma:n(1025,e.DiagnosticCategory.Error,"An_index_signature_cannot_have_a_trailing_comma_1025","An index signature cannot have a trailing comma."),Accessibility_modifier_already_seen:n(1028,e.DiagnosticCategory.Error,"Accessibility_modifier_already_seen_1028","Accessibility modifier already seen."),_0_modifier_must_precede_1_modifier:n(1029,e.DiagnosticCategory.Error,"_0_modifier_must_precede_1_modifier_1029","'{0}' modifier must precede '{1}' modifier."),_0_modifier_already_seen:n(1030,e.DiagnosticCategory.Error,"_0_modifier_already_seen_1030","'{0}' modifier already seen."),_0_modifier_cannot_appear_on_class_elements_of_this_kind:n(1031,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_class_elements_of_this_kind_1031","'{0}' modifier cannot appear on class elements of this kind."),super_must_be_followed_by_an_argument_list_or_member_access:n(1034,e.DiagnosticCategory.Error,"super_must_be_followed_by_an_argument_list_or_member_access_1034","'super' must be followed by an argument list or member access."),Only_ambient_modules_can_use_quoted_names:n(1035,e.DiagnosticCategory.Error,"Only_ambient_modules_can_use_quoted_names_1035","Only ambient modules can use quoted names."),Statements_are_not_allowed_in_ambient_contexts:n(1036,e.DiagnosticCategory.Error,"Statements_are_not_allowed_in_ambient_contexts_1036","Statements are not allowed in ambient contexts."),A_declare_modifier_cannot_be_used_in_an_already_ambient_context:n(1038,e.DiagnosticCategory.Error,"A_declare_modifier_cannot_be_used_in_an_already_ambient_context_1038","A 'declare' modifier cannot be used in an already ambient context."),Initializers_are_not_allowed_in_ambient_contexts:n(1039,e.DiagnosticCategory.Error,"Initializers_are_not_allowed_in_ambient_contexts_1039","Initializers are not allowed in ambient contexts."),_0_modifier_cannot_be_used_in_an_ambient_context:n(1040,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_in_an_ambient_context_1040","'{0}' modifier cannot be used in an ambient context."),_0_modifier_cannot_be_used_with_a_class_declaration:n(1041,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_with_a_class_declaration_1041","'{0}' modifier cannot be used with a class declaration."),_0_modifier_cannot_be_used_here:n(1042,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_here_1042","'{0}' modifier cannot be used here."),_0_modifier_cannot_appear_on_a_data_property:n(1043,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_data_property_1043","'{0}' modifier cannot appear on a data property."),_0_modifier_cannot_appear_on_a_module_or_namespace_element:n(1044,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_module_or_namespace_element_1044","'{0}' modifier cannot appear on a module or namespace element."),A_0_modifier_cannot_be_used_with_an_interface_declaration:n(1045,e.DiagnosticCategory.Error,"A_0_modifier_cannot_be_used_with_an_interface_declaration_1045","A '{0}' modifier cannot be used with an interface declaration."),Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier:n(1046,e.DiagnosticCategory.Error,"Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier_1046","Top-level declarations in .d.ts files must start with either a 'declare' or 'export' modifier."),A_rest_parameter_cannot_be_optional:n(1047,e.DiagnosticCategory.Error,"A_rest_parameter_cannot_be_optional_1047","A rest parameter cannot be optional."),A_rest_parameter_cannot_have_an_initializer:n(1048,e.DiagnosticCategory.Error,"A_rest_parameter_cannot_have_an_initializer_1048","A rest parameter cannot have an initializer."),A_set_accessor_must_have_exactly_one_parameter:n(1049,e.DiagnosticCategory.Error,"A_set_accessor_must_have_exactly_one_parameter_1049","A 'set' accessor must have exactly one parameter."),A_set_accessor_cannot_have_an_optional_parameter:n(1051,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_an_optional_parameter_1051","A 'set' accessor cannot have an optional parameter."),A_set_accessor_parameter_cannot_have_an_initializer:n(1052,e.DiagnosticCategory.Error,"A_set_accessor_parameter_cannot_have_an_initializer_1052","A 'set' accessor parameter cannot have an initializer."),A_set_accessor_cannot_have_rest_parameter:n(1053,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_rest_parameter_1053","A 'set' accessor cannot have rest parameter."),A_get_accessor_cannot_have_parameters:n(1054,e.DiagnosticCategory.Error,"A_get_accessor_cannot_have_parameters_1054","A 'get' accessor cannot have parameters."),Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value:n(1055,e.DiagnosticCategory.Error,"Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Prom_1055","Type '{0}' is not a valid async function return type in ES5/ES3 because it does not refer to a Promise-compatible constructor value."),Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher:n(1056,e.DiagnosticCategory.Error,"Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher_1056","Accessors are only available when targeting ECMAScript 5 and higher."),An_async_function_or_method_must_have_a_valid_awaitable_return_type:n(1057,e.DiagnosticCategory.Error,"An_async_function_or_method_must_have_a_valid_awaitable_return_type_1057","An async function or method must have a valid awaitable return type."),The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1058,e.DiagnosticCategory.Error,"The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_t_1058","The return type of an async function must either be a valid promise or must not contain a callable 'then' member."),A_promise_must_have_a_then_method:n(1059,e.DiagnosticCategory.Error,"A_promise_must_have_a_then_method_1059","A promise must have a 'then' method."),The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback:n(1060,e.DiagnosticCategory.Error,"The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback_1060","The first parameter of the 'then' method of a promise must be a callback."),Enum_member_must_have_initializer:n(1061,e.DiagnosticCategory.Error,"Enum_member_must_have_initializer_1061","Enum member must have initializer."),Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method:n(1062,e.DiagnosticCategory.Error,"Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method_1062","Type is referenced directly or indirectly in the fulfillment callback of its own 'then' method."),An_export_assignment_cannot_be_used_in_a_namespace:n(1063,e.DiagnosticCategory.Error,"An_export_assignment_cannot_be_used_in_a_namespace_1063","An export assignment cannot be used in a namespace."),The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0:n(1064,e.DiagnosticCategory.Error,"The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_wri_1064","The return type of an async function or method must be the global Promise<T> type. Did you mean to write 'Promise<{0}>'?"),In_ambient_enum_declarations_member_initializer_must_be_constant_expression:n(1066,e.DiagnosticCategory.Error,"In_ambient_enum_declarations_member_initializer_must_be_constant_expression_1066","In ambient enum declarations member initializer must be constant expression."),Unexpected_token_A_constructor_method_accessor_or_property_was_expected:n(1068,e.DiagnosticCategory.Error,"Unexpected_token_A_constructor_method_accessor_or_property_was_expected_1068","Unexpected token. A constructor, method, accessor, or property was expected."),Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces:n(1069,e.DiagnosticCategory.Error,"Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces_1069","Unexpected token. A type parameter name was expected without curly braces."),_0_modifier_cannot_appear_on_a_type_member:n(1070,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_type_member_1070","'{0}' modifier cannot appear on a type member."),_0_modifier_cannot_appear_on_an_index_signature:n(1071,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_an_index_signature_1071","'{0}' modifier cannot appear on an index signature."),A_0_modifier_cannot_be_used_with_an_import_declaration:n(1079,e.DiagnosticCategory.Error,"A_0_modifier_cannot_be_used_with_an_import_declaration_1079","A '{0}' modifier cannot be used with an import declaration."),Invalid_reference_directive_syntax:n(1084,e.DiagnosticCategory.Error,"Invalid_reference_directive_syntax_1084","Invalid 'reference' directive syntax."),Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0:n(1085,e.DiagnosticCategory.Error,"Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0_1085","Octal literals are not available when targeting ECMAScript 5 and higher. Use the syntax '{0}'."),_0_modifier_cannot_appear_on_a_constructor_declaration:n(1089,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_constructor_declaration_1089","'{0}' modifier cannot appear on a constructor declaration."),_0_modifier_cannot_appear_on_a_parameter:n(1090,e.DiagnosticCategory.Error,"_0_modifier_cannot_appear_on_a_parameter_1090","'{0}' modifier cannot appear on a parameter."),Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement:n(1091,e.DiagnosticCategory.Error,"Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement_1091","Only a single variable declaration is allowed in a 'for...in' statement."),Type_parameters_cannot_appear_on_a_constructor_declaration:n(1092,e.DiagnosticCategory.Error,"Type_parameters_cannot_appear_on_a_constructor_declaration_1092","Type parameters cannot appear on a constructor declaration."),Type_annotation_cannot_appear_on_a_constructor_declaration:n(1093,e.DiagnosticCategory.Error,"Type_annotation_cannot_appear_on_a_constructor_declaration_1093","Type annotation cannot appear on a constructor declaration."),An_accessor_cannot_have_type_parameters:n(1094,e.DiagnosticCategory.Error,"An_accessor_cannot_have_type_parameters_1094","An accessor cannot have type parameters."),A_set_accessor_cannot_have_a_return_type_annotation:n(1095,e.DiagnosticCategory.Error,"A_set_accessor_cannot_have_a_return_type_annotation_1095","A 'set' accessor cannot have a return type annotation."),An_index_signature_must_have_exactly_one_parameter:n(1096,e.DiagnosticCategory.Error,"An_index_signature_must_have_exactly_one_parameter_1096","An index signature must have exactly one parameter."),_0_list_cannot_be_empty:n(1097,e.DiagnosticCategory.Error,"_0_list_cannot_be_empty_1097","'{0}' list cannot be empty."),Type_parameter_list_cannot_be_empty:n(1098,e.DiagnosticCategory.Error,"Type_parameter_list_cannot_be_empty_1098","Type parameter list cannot be empty."),Type_argument_list_cannot_be_empty:n(1099,e.DiagnosticCategory.Error,"Type_argument_list_cannot_be_empty_1099","Type argument list cannot be empty."),Invalid_use_of_0_in_strict_mode:n(1100,e.DiagnosticCategory.Error,"Invalid_use_of_0_in_strict_mode_1100","Invalid use of '{0}' in strict mode."),with_statements_are_not_allowed_in_strict_mode:n(1101,e.DiagnosticCategory.Error,"with_statements_are_not_allowed_in_strict_mode_1101","'with' statements are not allowed in strict mode."),delete_cannot_be_called_on_an_identifier_in_strict_mode:n(1102,e.DiagnosticCategory.Error,"delete_cannot_be_called_on_an_identifier_in_strict_mode_1102","'delete' cannot be called on an identifier in strict mode."),A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator:n(1103,e.DiagnosticCategory.Error,"A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator_1103","A 'for-await-of' statement is only allowed within an async function or async generator."),A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement:n(1104,e.DiagnosticCategory.Error,"A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement_1104","A 'continue' statement can only be used within an enclosing iteration statement."),A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement:n(1105,e.DiagnosticCategory.Error,"A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement_1105","A 'break' statement can only be used within an enclosing iteration or switch statement."),Jump_target_cannot_cross_function_boundary:n(1107,e.DiagnosticCategory.Error,"Jump_target_cannot_cross_function_boundary_1107","Jump target cannot cross function boundary."),A_return_statement_can_only_be_used_within_a_function_body:n(1108,e.DiagnosticCategory.Error,"A_return_statement_can_only_be_used_within_a_function_body_1108","A 'return' statement can only be used within a function body."),Expression_expected:n(1109,e.DiagnosticCategory.Error,"Expression_expected_1109","Expression expected."),Type_expected:n(1110,e.DiagnosticCategory.Error,"Type_expected_1110","Type expected."),A_default_clause_cannot_appear_more_than_once_in_a_switch_statement:n(1113,e.DiagnosticCategory.Error,"A_default_clause_cannot_appear_more_than_once_in_a_switch_statement_1113","A 'default' clause cannot appear more than once in a 'switch' statement."),Duplicate_label_0:n(1114,e.DiagnosticCategory.Error,"Duplicate_label_0_1114","Duplicate label '{0}'."),A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement:n(1115,e.DiagnosticCategory.Error,"A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement_1115","A 'continue' statement can only jump to a label of an enclosing iteration statement."),A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement:n(1116,e.DiagnosticCategory.Error,"A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement_1116","A 'break' statement can only jump to a label of an enclosing statement."),An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode:n(1117,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode_1117","An object literal cannot have multiple properties with the same name in strict mode."),An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name:n(1118,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name_1118","An object literal cannot have multiple get/set accessors with the same name."),An_object_literal_cannot_have_property_and_accessor_with_the_same_name:n(1119,e.DiagnosticCategory.Error,"An_object_literal_cannot_have_property_and_accessor_with_the_same_name_1119","An object literal cannot have property and accessor with the same name."),An_export_assignment_cannot_have_modifiers:n(1120,e.DiagnosticCategory.Error,"An_export_assignment_cannot_have_modifiers_1120","An export assignment cannot have modifiers."),Octal_literals_are_not_allowed_in_strict_mode:n(1121,e.DiagnosticCategory.Error,"Octal_literals_are_not_allowed_in_strict_mode_1121","Octal literals are not allowed in strict mode."),Variable_declaration_list_cannot_be_empty:n(1123,e.DiagnosticCategory.Error,"Variable_declaration_list_cannot_be_empty_1123","Variable declaration list cannot be empty."),Digit_expected:n(1124,e.DiagnosticCategory.Error,"Digit_expected_1124","Digit expected."),Hexadecimal_digit_expected:n(1125,e.DiagnosticCategory.Error,"Hexadecimal_digit_expected_1125","Hexadecimal digit expected."),Unexpected_end_of_text:n(1126,e.DiagnosticCategory.Error,"Unexpected_end_of_text_1126","Unexpected end of text."),Invalid_character:n(1127,e.DiagnosticCategory.Error,"Invalid_character_1127","Invalid character."),Declaration_or_statement_expected:n(1128,e.DiagnosticCategory.Error,"Declaration_or_statement_expected_1128","Declaration or statement expected."),Statement_expected:n(1129,e.DiagnosticCategory.Error,"Statement_expected_1129","Statement expected."),case_or_default_expected:n(1130,e.DiagnosticCategory.Error,"case_or_default_expected_1130","'case' or 'default' expected."),Property_or_signature_expected:n(1131,e.DiagnosticCategory.Error,"Property_or_signature_expected_1131","Property or signature expected."),Enum_member_expected:n(1132,e.DiagnosticCategory.Error,"Enum_member_expected_1132","Enum member expected."),Variable_declaration_expected:n(1134,e.DiagnosticCategory.Error,"Variable_declaration_expected_1134","Variable declaration expected."),Argument_expression_expected:n(1135,e.DiagnosticCategory.Error,"Argument_expression_expected_1135","Argument expression expected."),Property_assignment_expected:n(1136,e.DiagnosticCategory.Error,"Property_assignment_expected_1136","Property assignment expected."),Expression_or_comma_expected:n(1137,e.DiagnosticCategory.Error,"Expression_or_comma_expected_1137","Expression or comma expected."),Parameter_declaration_expected:n(1138,e.DiagnosticCategory.Error,"Parameter_declaration_expected_1138","Parameter declaration expected."),Type_parameter_declaration_expected:n(1139,e.DiagnosticCategory.Error,"Type_parameter_declaration_expected_1139","Type parameter declaration expected."),Type_argument_expected:n(1140,e.DiagnosticCategory.Error,"Type_argument_expected_1140","Type argument expected."),String_literal_expected:n(1141,e.DiagnosticCategory.Error,"String_literal_expected_1141","String literal expected."),Line_break_not_permitted_here:n(1142,e.DiagnosticCategory.Error,"Line_break_not_permitted_here_1142","Line break not permitted here."),or_expected:n(1144,e.DiagnosticCategory.Error,"or_expected_1144","'{' or ';' expected."),Declaration_expected:n(1146,e.DiagnosticCategory.Error,"Declaration_expected_1146","Declaration expected."),Import_declarations_in_a_namespace_cannot_reference_a_module:n(1147,e.DiagnosticCategory.Error,"Import_declarations_in_a_namespace_cannot_reference_a_module_1147","Import declarations in a namespace cannot reference a module."),Cannot_use_imports_exports_or_module_augmentations_when_module_is_none:n(1148,e.DiagnosticCategory.Error,"Cannot_use_imports_exports_or_module_augmentations_when_module_is_none_1148","Cannot use imports, exports, or module augmentations when '--module' is 'none'."),File_name_0_differs_from_already_included_file_name_1_only_in_casing:n(1149,e.DiagnosticCategory.Error,"File_name_0_differs_from_already_included_file_name_1_only_in_casing_1149","File name '{0}' differs from already included file name '{1}' only in casing."),const_declarations_must_be_initialized:n(1155,e.DiagnosticCategory.Error,"const_declarations_must_be_initialized_1155","'const' declarations must be initialized."),const_declarations_can_only_be_declared_inside_a_block:n(1156,e.DiagnosticCategory.Error,"const_declarations_can_only_be_declared_inside_a_block_1156","'const' declarations can only be declared inside a block."),let_declarations_can_only_be_declared_inside_a_block:n(1157,e.DiagnosticCategory.Error,"let_declarations_can_only_be_declared_inside_a_block_1157","'let' declarations can only be declared inside a block."),Unterminated_template_literal:n(1160,e.DiagnosticCategory.Error,"Unterminated_template_literal_1160","Unterminated template literal."),Unterminated_regular_expression_literal:n(1161,e.DiagnosticCategory.Error,"Unterminated_regular_expression_literal_1161","Unterminated regular expression literal."),An_object_member_cannot_be_declared_optional:n(1162,e.DiagnosticCategory.Error,"An_object_member_cannot_be_declared_optional_1162","An object member cannot be declared optional."),A_yield_expression_is_only_allowed_in_a_generator_body:n(1163,e.DiagnosticCategory.Error,"A_yield_expression_is_only_allowed_in_a_generator_body_1163","A 'yield' expression is only allowed in a generator body."),Computed_property_names_are_not_allowed_in_enums:n(1164,e.DiagnosticCategory.Error,"Computed_property_names_are_not_allowed_in_enums_1164","Computed property names are not allowed in enums."),A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1165,e.DiagnosticCategory.Error,"A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_t_1165","A computed property name in an ambient context must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1166,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_1166","A computed property name in a class property declaration must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1168,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_ty_1168","A computed property name in a method overload must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1169,e.DiagnosticCategory.Error,"A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_1169","A computed property name in an interface must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type:n(1170,e.DiagnosticCategory.Error,"A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type__1170","A computed property name in a type literal must refer to an expression whose type is a literal type or a 'unique symbol' type."),A_comma_expression_is_not_allowed_in_a_computed_property_name:n(1171,e.DiagnosticCategory.Error,"A_comma_expression_is_not_allowed_in_a_computed_property_name_1171","A comma expression is not allowed in a computed property name."),extends_clause_already_seen:n(1172,e.DiagnosticCategory.Error,"extends_clause_already_seen_1172","'extends' clause already seen."),extends_clause_must_precede_implements_clause:n(1173,e.DiagnosticCategory.Error,"extends_clause_must_precede_implements_clause_1173","'extends' clause must precede 'implements' clause."),Classes_can_only_extend_a_single_class:n(1174,e.DiagnosticCategory.Error,"Classes_can_only_extend_a_single_class_1174","Classes can only extend a single class."),implements_clause_already_seen:n(1175,e.DiagnosticCategory.Error,"implements_clause_already_seen_1175","'implements' clause already seen."),Interface_declaration_cannot_have_implements_clause:n(1176,e.DiagnosticCategory.Error,"Interface_declaration_cannot_have_implements_clause_1176","Interface declaration cannot have 'implements' clause."),Binary_digit_expected:n(1177,e.DiagnosticCategory.Error,"Binary_digit_expected_1177","Binary digit expected."),Octal_digit_expected:n(1178,e.DiagnosticCategory.Error,"Octal_digit_expected_1178","Octal digit expected."),Unexpected_token_expected:n(1179,e.DiagnosticCategory.Error,"Unexpected_token_expected_1179","Unexpected token. '{' expected."),Property_destructuring_pattern_expected:n(1180,e.DiagnosticCategory.Error,"Property_destructuring_pattern_expected_1180","Property destructuring pattern expected."),Array_element_destructuring_pattern_expected:n(1181,e.DiagnosticCategory.Error,"Array_element_destructuring_pattern_expected_1181","Array element destructuring pattern expected."),A_destructuring_declaration_must_have_an_initializer:n(1182,e.DiagnosticCategory.Error,"A_destructuring_declaration_must_have_an_initializer_1182","A destructuring declaration must have an initializer."),An_implementation_cannot_be_declared_in_ambient_contexts:n(1183,e.DiagnosticCategory.Error,"An_implementation_cannot_be_declared_in_ambient_contexts_1183","An implementation cannot be declared in ambient contexts."),Modifiers_cannot_appear_here:n(1184,e.DiagnosticCategory.Error,"Modifiers_cannot_appear_here_1184","Modifiers cannot appear here."),Merge_conflict_marker_encountered:n(1185,e.DiagnosticCategory.Error,"Merge_conflict_marker_encountered_1185","Merge conflict marker encountered."),A_rest_element_cannot_have_an_initializer:n(1186,e.DiagnosticCategory.Error,"A_rest_element_cannot_have_an_initializer_1186","A rest element cannot have an initializer."),A_parameter_property_may_not_be_declared_using_a_binding_pattern:n(1187,e.DiagnosticCategory.Error,"A_parameter_property_may_not_be_declared_using_a_binding_pattern_1187","A parameter property may not be declared using a binding pattern."),Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement:n(1188,e.DiagnosticCategory.Error,"Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement_1188","Only a single variable declaration is allowed in a 'for...of' statement."),The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer:n(1189,e.DiagnosticCategory.Error,"The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer_1189","The variable declaration of a 'for...in' statement cannot have an initializer."),The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer:n(1190,e.DiagnosticCategory.Error,"The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer_1190","The variable declaration of a 'for...of' statement cannot have an initializer."),An_import_declaration_cannot_have_modifiers:n(1191,e.DiagnosticCategory.Error,"An_import_declaration_cannot_have_modifiers_1191","An import declaration cannot have modifiers."),Module_0_has_no_default_export:n(1192,e.DiagnosticCategory.Error,"Module_0_has_no_default_export_1192","Module '{0}' has no default export."),An_export_declaration_cannot_have_modifiers:n(1193,e.DiagnosticCategory.Error,"An_export_declaration_cannot_have_modifiers_1193","An export declaration cannot have modifiers."),Export_declarations_are_not_permitted_in_a_namespace:n(1194,e.DiagnosticCategory.Error,"Export_declarations_are_not_permitted_in_a_namespace_1194","Export declarations are not permitted in a namespace."),export_Asterisk_does_not_re_export_a_default:n(1195,e.DiagnosticCategory.Error,"export_Asterisk_does_not_re_export_a_default_1195","'export *' does not re-export a default."),Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified:n(1196,e.DiagnosticCategory.Error,"Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified_1196","Catch clause variable type annotation must be 'any' or 'unknown' if specified."),Catch_clause_variable_cannot_have_an_initializer:n(1197,e.DiagnosticCategory.Error,"Catch_clause_variable_cannot_have_an_initializer_1197","Catch clause variable cannot have an initializer."),An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive:n(1198,e.DiagnosticCategory.Error,"An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive_1198","An extended Unicode escape value must be between 0x0 and 0x10FFFF inclusive."),Unterminated_Unicode_escape_sequence:n(1199,e.DiagnosticCategory.Error,"Unterminated_Unicode_escape_sequence_1199","Unterminated Unicode escape sequence."),Line_terminator_not_permitted_before_arrow:n(1200,e.DiagnosticCategory.Error,"Line_terminator_not_permitted_before_arrow_1200","Line terminator not permitted before arrow."),Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead:n(1202,e.DiagnosticCategory.Error,"Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_1202","Import assignment cannot be used when targeting ECMAScript modules. Consider using 'import * as ns from \"mod\"', 'import {a} from \"mod\"', 'import d from \"mod\"', or another module format instead."),Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead:n(1203,e.DiagnosticCategory.Error,"Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or__1203","Export assignment cannot be used when targeting ECMAScript modules. Consider using 'export default' or another module format instead."),Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type:n(1205,e.DiagnosticCategory.Error,"Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type_1205","Re-exporting a type when the '--isolatedModules' flag is provided requires using 'export type'."),Decorators_are_not_valid_here:n(1206,e.DiagnosticCategory.Error,"Decorators_are_not_valid_here_1206","Decorators are not valid here."),Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name:n(1207,e.DiagnosticCategory.Error,"Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name_1207","Decorators cannot be applied to multiple get/set accessors of the same name."),_0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_import_export_or_an_empty_export_statement_to_make_it_a_module:n(1208,e.DiagnosticCategory.Error,"_0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_imp_1208","'{0}' cannot be compiled under '--isolatedModules' because it is considered a global script file. Add an import, export, or an empty 'export {}' statement to make it a module."),Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode:n(1210,e.DiagnosticCategory.Error,"Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode_1210","Invalid use of '{0}'. Class definitions are automatically in strict mode."),A_class_declaration_without_the_default_modifier_must_have_a_name:n(1211,e.DiagnosticCategory.Error,"A_class_declaration_without_the_default_modifier_must_have_a_name_1211","A class declaration without the 'default' modifier must have a name."),Identifier_expected_0_is_a_reserved_word_in_strict_mode:n(1212,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_1212","Identifier expected. '{0}' is a reserved word in strict mode."),Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode:n(1213,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_stric_1213","Identifier expected. '{0}' is a reserved word in strict mode. Class definitions are automatically in strict mode."),Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode:n(1214,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode_1214","Identifier expected. '{0}' is a reserved word in strict mode. Modules are automatically in strict mode."),Invalid_use_of_0_Modules_are_automatically_in_strict_mode:n(1215,e.DiagnosticCategory.Error,"Invalid_use_of_0_Modules_are_automatically_in_strict_mode_1215","Invalid use of '{0}'. Modules are automatically in strict mode."),Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules:n(1216,e.DiagnosticCategory.Error,"Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules_1216","Identifier expected. '__esModule' is reserved as an exported marker when transforming ECMAScript modules."),Export_assignment_is_not_supported_when_module_flag_is_system:n(1218,e.DiagnosticCategory.Error,"Export_assignment_is_not_supported_when_module_flag_is_system_1218","Export assignment is not supported when '--module' flag is 'system'."),Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning:n(1219,e.DiagnosticCategory.Error,"Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_t_1219","Experimental support for decorators is a feature that is subject to change in a future release. Set the 'experimentalDecorators' option in your 'tsconfig' or 'jsconfig' to remove this warning."),Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher:n(1220,e.DiagnosticCategory.Error,"Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher_1220","Generators are only available when targeting ECMAScript 2015 or higher."),Generators_are_not_allowed_in_an_ambient_context:n(1221,e.DiagnosticCategory.Error,"Generators_are_not_allowed_in_an_ambient_context_1221","Generators are not allowed in an ambient context."),An_overload_signature_cannot_be_declared_as_a_generator:n(1222,e.DiagnosticCategory.Error,"An_overload_signature_cannot_be_declared_as_a_generator_1222","An overload signature cannot be declared as a generator."),_0_tag_already_specified:n(1223,e.DiagnosticCategory.Error,"_0_tag_already_specified_1223","'{0}' tag already specified."),Signature_0_must_be_a_type_predicate:n(1224,e.DiagnosticCategory.Error,"Signature_0_must_be_a_type_predicate_1224","Signature '{0}' must be a type predicate."),Cannot_find_parameter_0:n(1225,e.DiagnosticCategory.Error,"Cannot_find_parameter_0_1225","Cannot find parameter '{0}'."),Type_predicate_0_is_not_assignable_to_1:n(1226,e.DiagnosticCategory.Error,"Type_predicate_0_is_not_assignable_to_1_1226","Type predicate '{0}' is not assignable to '{1}'."),Parameter_0_is_not_in_the_same_position_as_parameter_1:n(1227,e.DiagnosticCategory.Error,"Parameter_0_is_not_in_the_same_position_as_parameter_1_1227","Parameter '{0}' is not in the same position as parameter '{1}'."),A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods:n(1228,e.DiagnosticCategory.Error,"A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods_1228","A type predicate is only allowed in return type position for functions and methods."),A_type_predicate_cannot_reference_a_rest_parameter:n(1229,e.DiagnosticCategory.Error,"A_type_predicate_cannot_reference_a_rest_parameter_1229","A type predicate cannot reference a rest parameter."),A_type_predicate_cannot_reference_element_0_in_a_binding_pattern:n(1230,e.DiagnosticCategory.Error,"A_type_predicate_cannot_reference_element_0_in_a_binding_pattern_1230","A type predicate cannot reference element '{0}' in a binding pattern."),An_export_assignment_can_only_be_used_in_a_module:n(1231,e.DiagnosticCategory.Error,"An_export_assignment_can_only_be_used_in_a_module_1231","An export assignment can only be used in a module."),An_import_declaration_can_only_be_used_in_a_namespace_or_module:n(1232,e.DiagnosticCategory.Error,"An_import_declaration_can_only_be_used_in_a_namespace_or_module_1232","An import declaration can only be used in a namespace or module."),An_export_declaration_can_only_be_used_in_a_module:n(1233,e.DiagnosticCategory.Error,"An_export_declaration_can_only_be_used_in_a_module_1233","An export declaration can only be used in a module."),An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file:n(1234,e.DiagnosticCategory.Error,"An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file_1234","An ambient module declaration is only allowed at the top level in a file."),A_namespace_declaration_is_only_allowed_in_a_namespace_or_module:n(1235,e.DiagnosticCategory.Error,"A_namespace_declaration_is_only_allowed_in_a_namespace_or_module_1235","A namespace declaration is only allowed in a namespace or module."),The_return_type_of_a_property_decorator_function_must_be_either_void_or_any:n(1236,e.DiagnosticCategory.Error,"The_return_type_of_a_property_decorator_function_must_be_either_void_or_any_1236","The return type of a property decorator function must be either 'void' or 'any'."),The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any:n(1237,e.DiagnosticCategory.Error,"The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any_1237","The return type of a parameter decorator function must be either 'void' or 'any'."),Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression:n(1238,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression_1238","Unable to resolve signature of class decorator when called as an expression."),Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression:n(1239,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression_1239","Unable to resolve signature of parameter decorator when called as an expression."),Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression:n(1240,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression_1240","Unable to resolve signature of property decorator when called as an expression."),Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression:n(1241,e.DiagnosticCategory.Error,"Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression_1241","Unable to resolve signature of method decorator when called as an expression."),abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration:n(1242,e.DiagnosticCategory.Error,"abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration_1242","'abstract' modifier can only appear on a class, method, or property declaration."),_0_modifier_cannot_be_used_with_1_modifier:n(1243,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_with_1_modifier_1243","'{0}' modifier cannot be used with '{1}' modifier."),Abstract_methods_can_only_appear_within_an_abstract_class:n(1244,e.DiagnosticCategory.Error,"Abstract_methods_can_only_appear_within_an_abstract_class_1244","Abstract methods can only appear within an abstract class."),Method_0_cannot_have_an_implementation_because_it_is_marked_abstract:n(1245,e.DiagnosticCategory.Error,"Method_0_cannot_have_an_implementation_because_it_is_marked_abstract_1245","Method '{0}' cannot have an implementation because it is marked abstract."),An_interface_property_cannot_have_an_initializer:n(1246,e.DiagnosticCategory.Error,"An_interface_property_cannot_have_an_initializer_1246","An interface property cannot have an initializer."),A_type_literal_property_cannot_have_an_initializer:n(1247,e.DiagnosticCategory.Error,"A_type_literal_property_cannot_have_an_initializer_1247","A type literal property cannot have an initializer."),A_class_member_cannot_have_the_0_keyword:n(1248,e.DiagnosticCategory.Error,"A_class_member_cannot_have_the_0_keyword_1248","A class member cannot have the '{0}' keyword."),A_decorator_can_only_decorate_a_method_implementation_not_an_overload:n(1249,e.DiagnosticCategory.Error,"A_decorator_can_only_decorate_a_method_implementation_not_an_overload_1249","A decorator can only decorate a method implementation, not an overload."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5:n(1250,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_1250","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode:n(1251,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_d_1251","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Class definitions are automatically in strict mode."),Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode:n(1252,e.DiagnosticCategory.Error,"Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_1252","Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Modules are automatically in strict mode."),_0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag:n(1253,e.DiagnosticCategory.Error,"_0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag_1253","'{0}' tag cannot be used independently as a top level JSDoc tag."),A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference:n(1254,e.DiagnosticCategory.Error,"A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_refere_1254","A 'const' initializer in an ambient context must be a string or numeric literal or literal enum reference."),A_definite_assignment_assertion_is_not_permitted_in_this_context:n(1255,e.DiagnosticCategory.Error,"A_definite_assignment_assertion_is_not_permitted_in_this_context_1255","A definite assignment assertion '!' is not permitted in this context."),A_rest_element_must_be_last_in_a_tuple_type:n(1256,e.DiagnosticCategory.Error,"A_rest_element_must_be_last_in_a_tuple_type_1256","A rest element must be last in a tuple type."),A_required_element_cannot_follow_an_optional_element:n(1257,e.DiagnosticCategory.Error,"A_required_element_cannot_follow_an_optional_element_1257","A required element cannot follow an optional element."),Module_0_can_only_be_default_imported_using_the_1_flag:n(1259,e.DiagnosticCategory.Error,"Module_0_can_only_be_default_imported_using_the_1_flag_1259","Module '{0}' can only be default-imported using the '{1}' flag"),Keywords_cannot_contain_escape_characters:n(1260,e.DiagnosticCategory.Error,"Keywords_cannot_contain_escape_characters_1260","Keywords cannot contain escape characters."),Already_included_file_name_0_differs_from_file_name_1_only_in_casing:n(1261,e.DiagnosticCategory.Error,"Already_included_file_name_0_differs_from_file_name_1_only_in_casing_1261","Already included file name '{0}' differs from file name '{1}' only in casing."),Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module:n(1262,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module_1262","Identifier expected. '{0}' is a reserved word at the top-level of a module."),Declarations_with_initializers_cannot_also_have_definite_assignment_assertions:n(1263,e.DiagnosticCategory.Error,"Declarations_with_initializers_cannot_also_have_definite_assignment_assertions_1263","Declarations with initializers cannot also have definite assignment assertions."),Declarations_with_definite_assignment_assertions_must_also_have_type_annotations:n(1264,e.DiagnosticCategory.Error,"Declarations_with_definite_assignment_assertions_must_also_have_type_annotations_1264","Declarations with definite assignment assertions must also have type annotations."),with_statements_are_not_allowed_in_an_async_function_block:n(1300,e.DiagnosticCategory.Error,"with_statements_are_not_allowed_in_an_async_function_block_1300","'with' statements are not allowed in an async function block."),await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules:n(1308,e.DiagnosticCategory.Error,"await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules_1308","'await' expressions are only allowed within async functions and at the top levels of modules."),Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern:n(1312,e.DiagnosticCategory.Error,"Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_1312","Did you mean to use a ':'? An '=' can only follow a property name when the containing object literal is part of a destructuring pattern."),The_body_of_an_if_statement_cannot_be_the_empty_statement:n(1313,e.DiagnosticCategory.Error,"The_body_of_an_if_statement_cannot_be_the_empty_statement_1313","The body of an 'if' statement cannot be the empty statement."),Global_module_exports_may_only_appear_in_module_files:n(1314,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_in_module_files_1314","Global module exports may only appear in module files."),Global_module_exports_may_only_appear_in_declaration_files:n(1315,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_in_declaration_files_1315","Global module exports may only appear in declaration files."),Global_module_exports_may_only_appear_at_top_level:n(1316,e.DiagnosticCategory.Error,"Global_module_exports_may_only_appear_at_top_level_1316","Global module exports may only appear at top level."),A_parameter_property_cannot_be_declared_using_a_rest_parameter:n(1317,e.DiagnosticCategory.Error,"A_parameter_property_cannot_be_declared_using_a_rest_parameter_1317","A parameter property cannot be declared using a rest parameter."),An_abstract_accessor_cannot_have_an_implementation:n(1318,e.DiagnosticCategory.Error,"An_abstract_accessor_cannot_have_an_implementation_1318","An abstract accessor cannot have an implementation."),A_default_export_can_only_be_used_in_an_ECMAScript_style_module:n(1319,e.DiagnosticCategory.Error,"A_default_export_can_only_be_used_in_an_ECMAScript_style_module_1319","A default export can only be used in an ECMAScript-style module."),Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1320,e.DiagnosticCategory.Error,"Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member_1320","Type of 'await' operand must either be a valid promise or must not contain a callable 'then' member."),Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1321,e.DiagnosticCategory.Error,"Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_cal_1321","Type of 'yield' operand in an async generator must either be a valid promise or must not contain a callable 'then' member."),Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:n(1322,e.DiagnosticCategory.Error,"Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_con_1322","Type of iterated elements of a 'yield*' operand must either be a valid promise or must not contain a callable 'then' member."),Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system_or_umd:n(1323,e.DiagnosticCategory.Error,"Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system__1323","Dynamic imports are only supported when the '--module' flag is set to 'es2020', 'esnext', 'commonjs', 'amd', 'system', or 'umd'."),Dynamic_import_must_have_one_specifier_as_an_argument:n(1324,e.DiagnosticCategory.Error,"Dynamic_import_must_have_one_specifier_as_an_argument_1324","Dynamic import must have one specifier as an argument."),Specifier_of_dynamic_import_cannot_be_spread_element:n(1325,e.DiagnosticCategory.Error,"Specifier_of_dynamic_import_cannot_be_spread_element_1325","Specifier of dynamic import cannot be spread element."),Dynamic_import_cannot_have_type_arguments:n(1326,e.DiagnosticCategory.Error,"Dynamic_import_cannot_have_type_arguments_1326","Dynamic import cannot have type arguments."),String_literal_with_double_quotes_expected:n(1327,e.DiagnosticCategory.Error,"String_literal_with_double_quotes_expected_1327","String literal with double quotes expected."),Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal:n(1328,e.DiagnosticCategory.Error,"Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_li_1328","Property value can only be string literal, numeric literal, 'true', 'false', 'null', object literal or array literal."),_0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0:n(1329,e.DiagnosticCategory.Error,"_0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write__1329","'{0}' accepts too few arguments to be used as a decorator here. Did you mean to call it first and write '@{0}()'?"),A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly:n(1330,e.DiagnosticCategory.Error,"A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly_1330","A property of an interface or type literal whose type is a 'unique symbol' type must be 'readonly'."),A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly:n(1331,e.DiagnosticCategory.Error,"A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly_1331","A property of a class whose type is a 'unique symbol' type must be both 'static' and 'readonly'."),A_variable_whose_type_is_a_unique_symbol_type_must_be_const:n(1332,e.DiagnosticCategory.Error,"A_variable_whose_type_is_a_unique_symbol_type_must_be_const_1332","A variable whose type is a 'unique symbol' type must be 'const'."),unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name:n(1333,e.DiagnosticCategory.Error,"unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name_1333","'unique symbol' types may not be used on a variable declaration with a binding name."),unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement:n(1334,e.DiagnosticCategory.Error,"unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement_1334","'unique symbol' types are only allowed on variables in a variable statement."),unique_symbol_types_are_not_allowed_here:n(1335,e.DiagnosticCategory.Error,"unique_symbol_types_are_not_allowed_here_1335","'unique symbol' types are not allowed here."),An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead:n(1336,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead_1336","An index signature parameter type cannot be a type alias. Consider writing '[{0}: {1}]: {2}' instead."),An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead:n(1337,e.DiagnosticCategory.Error,"An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead_1337","An index signature parameter type cannot be a union type. Consider using a mapped object type instead."),infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type:n(1338,e.DiagnosticCategory.Error,"infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type_1338","'infer' declarations are only permitted in the 'extends' clause of a conditional type."),Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here:n(1339,e.DiagnosticCategory.Error,"Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here_1339","Module '{0}' does not refer to a value, but is used as a value here."),Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0:n(1340,e.DiagnosticCategory.Error,"Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0_1340","Module '{0}' does not refer to a type, but is used as a type here. Did you mean 'typeof import('{0}')'?"),Type_arguments_cannot_be_used_here:n(1342,e.DiagnosticCategory.Error,"Type_arguments_cannot_be_used_here_1342","Type arguments cannot be used here."),The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system:n(1343,e.DiagnosticCategory.Error,"The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system_1343","The 'import.meta' meta-property is only allowed when the '--module' option is 'es2020', 'esnext', or 'system'."),A_label_is_not_allowed_here:n(1344,e.DiagnosticCategory.Error,"A_label_is_not_allowed_here_1344","'A label is not allowed here."),An_expression_of_type_void_cannot_be_tested_for_truthiness:n(1345,e.DiagnosticCategory.Error,"An_expression_of_type_void_cannot_be_tested_for_truthiness_1345","An expression of type 'void' cannot be tested for truthiness."),This_parameter_is_not_allowed_with_use_strict_directive:n(1346,e.DiagnosticCategory.Error,"This_parameter_is_not_allowed_with_use_strict_directive_1346","This parameter is not allowed with 'use strict' directive."),use_strict_directive_cannot_be_used_with_non_simple_parameter_list:n(1347,e.DiagnosticCategory.Error,"use_strict_directive_cannot_be_used_with_non_simple_parameter_list_1347","'use strict' directive cannot be used with non-simple parameter list."),Non_simple_parameter_declared_here:n(1348,e.DiagnosticCategory.Error,"Non_simple_parameter_declared_here_1348","Non-simple parameter declared here."),use_strict_directive_used_here:n(1349,e.DiagnosticCategory.Error,"use_strict_directive_used_here_1349","'use strict' directive used here."),Print_the_final_configuration_instead_of_building:n(1350,e.DiagnosticCategory.Message,"Print_the_final_configuration_instead_of_building_1350","Print the final configuration instead of building."),An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal:n(1351,e.DiagnosticCategory.Error,"An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal_1351","An identifier or keyword cannot immediately follow a numeric literal."),A_bigint_literal_cannot_use_exponential_notation:n(1352,e.DiagnosticCategory.Error,"A_bigint_literal_cannot_use_exponential_notation_1352","A bigint literal cannot use exponential notation."),A_bigint_literal_must_be_an_integer:n(1353,e.DiagnosticCategory.Error,"A_bigint_literal_must_be_an_integer_1353","A bigint literal must be an integer."),readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types:n(1354,e.DiagnosticCategory.Error,"readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types_1354","'readonly' type modifier is only permitted on array and tuple literal types."),A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array_or_object_literals:n(1355,e.DiagnosticCategory.Error,"A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array__1355","A 'const' assertions can only be applied to references to enum members, or string, number, boolean, array, or object literals."),Did_you_mean_to_mark_this_function_as_async:n(1356,e.DiagnosticCategory.Error,"Did_you_mean_to_mark_this_function_as_async_1356","Did you mean to mark this function as 'async'?"),An_enum_member_name_must_be_followed_by_a_or:n(1357,e.DiagnosticCategory.Error,"An_enum_member_name_must_be_followed_by_a_or_1357","An enum member name must be followed by a ',', '=', or '}'."),Tagged_template_expressions_are_not_permitted_in_an_optional_chain:n(1358,e.DiagnosticCategory.Error,"Tagged_template_expressions_are_not_permitted_in_an_optional_chain_1358","Tagged template expressions are not permitted in an optional chain."),Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here:n(1359,e.DiagnosticCategory.Error,"Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here_1359","Identifier expected. '{0}' is a reserved word that cannot be used here."),Did_you_mean_to_parenthesize_this_function_type:n(1360,e.DiagnosticCategory.Error,"Did_you_mean_to_parenthesize_this_function_type_1360","Did you mean to parenthesize this function type?"),_0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type:n(1361,e.DiagnosticCategory.Error,"_0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type_1361","'{0}' cannot be used as a value because it was imported using 'import type'."),_0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type:n(1362,e.DiagnosticCategory.Error,"_0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type_1362","'{0}' cannot be used as a value because it was exported using 'export type'."),A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both:n(1363,e.DiagnosticCategory.Error,"A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both_1363","A type-only import can specify a default import or named bindings, but not both."),Convert_to_type_only_export:n(1364,e.DiagnosticCategory.Message,"Convert_to_type_only_export_1364","Convert to type-only export"),Convert_all_re_exported_types_to_type_only_exports:n(1365,e.DiagnosticCategory.Message,"Convert_all_re_exported_types_to_type_only_exports_1365","Convert all re-exported types to type-only exports"),Split_into_two_separate_import_declarations:n(1366,e.DiagnosticCategory.Message,"Split_into_two_separate_import_declarations_1366","Split into two separate import declarations"),Split_all_invalid_type_only_imports:n(1367,e.DiagnosticCategory.Message,"Split_all_invalid_type_only_imports_1367","Split all invalid type-only imports"),Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types:n(1368,e.DiagnosticCategory.Message,"Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types_1368","Specify emit/checking behavior for imports that are only used for types"),Did_you_mean_0:n(1369,e.DiagnosticCategory.Message,"Did_you_mean_0_1369","Did you mean '{0}'?"),Only_ECMAScript_imports_may_use_import_type:n(1370,e.DiagnosticCategory.Error,"Only_ECMAScript_imports_may_use_import_type_1370","Only ECMAScript imports may use 'import type'."),This_import_is_never_used_as_a_value_and_must_use_import_type_because_the_importsNotUsedAsValues_is_set_to_error:n(1371,e.DiagnosticCategory.Error,"This_import_is_never_used_as_a_value_and_must_use_import_type_because_the_importsNotUsedAsValues_is__1371","This import is never used as a value and must use 'import type' because the 'importsNotUsedAsValues' is set to 'error'."),Convert_to_type_only_import:n(1373,e.DiagnosticCategory.Message,"Convert_to_type_only_import_1373","Convert to type-only import"),Convert_all_imports_not_used_as_a_value_to_type_only_imports:n(1374,e.DiagnosticCategory.Message,"Convert_all_imports_not_used_as_a_value_to_type_only_imports_1374","Convert all imports not used as a value to type-only imports"),await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module:n(1375,e.DiagnosticCategory.Error,"await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_fi_1375","'await' expressions are only allowed at the top level of a file when that file is a module, but this file has no imports or exports. Consider adding an empty 'export {}' to make this file a module."),_0_was_imported_here:n(1376,e.DiagnosticCategory.Message,"_0_was_imported_here_1376","'{0}' was imported here."),_0_was_exported_here:n(1377,e.DiagnosticCategory.Message,"_0_was_exported_here_1377","'{0}' was exported here."),Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher:n(1378,e.DiagnosticCategory.Error,"Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_t_1378","Top-level 'await' expressions are only allowed when the 'module' option is set to 'esnext' or 'system', and the 'target' option is set to 'es2017' or higher."),An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type:n(1379,e.DiagnosticCategory.Error,"An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type_1379","An import alias cannot reference a declaration that was exported using 'export type'."),An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type:n(1380,e.DiagnosticCategory.Error,"An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type_1380","An import alias cannot reference a declaration that was imported using 'import type'."),Unexpected_token_Did_you_mean_or_rbrace:n(1381,e.DiagnosticCategory.Error,"Unexpected_token_Did_you_mean_or_rbrace_1381","Unexpected token. Did you mean `{'}'}` or `&rbrace;`?"),Unexpected_token_Did_you_mean_or_gt:n(1382,e.DiagnosticCategory.Error,"Unexpected_token_Did_you_mean_or_gt_1382","Unexpected token. Did you mean `{'>'}` or `&gt;`?"),Only_named_exports_may_use_export_type:n(1383,e.DiagnosticCategory.Error,"Only_named_exports_may_use_export_type_1383","Only named exports may use 'export type'."),A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list:n(1384,e.DiagnosticCategory.Error,"A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list_1384","A 'new' expression with type arguments must always be followed by a parenthesized argument list."),Function_type_notation_must_be_parenthesized_when_used_in_a_union_type:n(1385,e.DiagnosticCategory.Error,"Function_type_notation_must_be_parenthesized_when_used_in_a_union_type_1385","Function type notation must be parenthesized when used in a union type."),Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type:n(1386,e.DiagnosticCategory.Error,"Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type_1386","Constructor type notation must be parenthesized when used in a union type."),Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type:n(1387,e.DiagnosticCategory.Error,"Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type_1387","Function type notation must be parenthesized when used in an intersection type."),Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type:n(1388,e.DiagnosticCategory.Error,"Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type_1388","Constructor type notation must be parenthesized when used in an intersection type."),_0_is_not_allowed_as_a_variable_declaration_name:n(1389,e.DiagnosticCategory.Error,"_0_is_not_allowed_as_a_variable_declaration_name_1389","'{0}' is not allowed as a variable declaration name."),The_types_of_0_are_incompatible_between_these_types:n(2200,e.DiagnosticCategory.Error,"The_types_of_0_are_incompatible_between_these_types_2200","The types of '{0}' are incompatible between these types."),The_types_returned_by_0_are_incompatible_between_these_types:n(2201,e.DiagnosticCategory.Error,"The_types_returned_by_0_are_incompatible_between_these_types_2201","The types returned by '{0}' are incompatible between these types."),Call_signature_return_types_0_and_1_are_incompatible:n(2202,e.DiagnosticCategory.Error,"Call_signature_return_types_0_and_1_are_incompatible_2202","Call signature return types '{0}' and '{1}' are incompatible.",void 0,!0),Construct_signature_return_types_0_and_1_are_incompatible:n(2203,e.DiagnosticCategory.Error,"Construct_signature_return_types_0_and_1_are_incompatible_2203","Construct signature return types '{0}' and '{1}' are incompatible.",void 0,!0),Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1:n(2204,e.DiagnosticCategory.Error,"Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1_2204","Call signatures with no arguments have incompatible return types '{0}' and '{1}'.",void 0,!0),Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1:n(2205,e.DiagnosticCategory.Error,"Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1_2205","Construct signatures with no arguments have incompatible return types '{0}' and '{1}'.",void 0,!0),Duplicate_identifier_0:n(2300,e.DiagnosticCategory.Error,"Duplicate_identifier_0_2300","Duplicate identifier '{0}'."),Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor:n(2301,e.DiagnosticCategory.Error,"Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor_2301","Initializer of instance member variable '{0}' cannot reference identifier '{1}' declared in the constructor."),Static_members_cannot_reference_class_type_parameters:n(2302,e.DiagnosticCategory.Error,"Static_members_cannot_reference_class_type_parameters_2302","Static members cannot reference class type parameters."),Circular_definition_of_import_alias_0:n(2303,e.DiagnosticCategory.Error,"Circular_definition_of_import_alias_0_2303","Circular definition of import alias '{0}'."),Cannot_find_name_0:n(2304,e.DiagnosticCategory.Error,"Cannot_find_name_0_2304","Cannot find name '{0}'."),Module_0_has_no_exported_member_1:n(2305,e.DiagnosticCategory.Error,"Module_0_has_no_exported_member_1_2305","Module '{0}' has no exported member '{1}'."),File_0_is_not_a_module:n(2306,e.DiagnosticCategory.Error,"File_0_is_not_a_module_2306","File '{0}' is not a module."),Cannot_find_module_0_or_its_corresponding_type_declarations:n(2307,e.DiagnosticCategory.Error,"Cannot_find_module_0_or_its_corresponding_type_declarations_2307","Cannot find module '{0}' or its corresponding type declarations."),Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity:n(2308,e.DiagnosticCategory.Error,"Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambig_2308","Module {0} has already exported a member named '{1}'. Consider explicitly re-exporting to resolve the ambiguity."),An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements:n(2309,e.DiagnosticCategory.Error,"An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements_2309","An export assignment cannot be used in a module with other exported elements."),Type_0_recursively_references_itself_as_a_base_type:n(2310,e.DiagnosticCategory.Error,"Type_0_recursively_references_itself_as_a_base_type_2310","Type '{0}' recursively references itself as a base type."),A_class_may_only_extend_another_class:n(2311,e.DiagnosticCategory.Error,"A_class_may_only_extend_another_class_2311","A class may only extend another class."),An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2312,e.DiagnosticCategory.Error,"An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_me_2312","An interface can only extend an object type or intersection of object types with statically known members."),Type_parameter_0_has_a_circular_constraint:n(2313,e.DiagnosticCategory.Error,"Type_parameter_0_has_a_circular_constraint_2313","Type parameter '{0}' has a circular constraint."),Generic_type_0_requires_1_type_argument_s:n(2314,e.DiagnosticCategory.Error,"Generic_type_0_requires_1_type_argument_s_2314","Generic type '{0}' requires {1} type argument(s)."),Type_0_is_not_generic:n(2315,e.DiagnosticCategory.Error,"Type_0_is_not_generic_2315","Type '{0}' is not generic."),Global_type_0_must_be_a_class_or_interface_type:n(2316,e.DiagnosticCategory.Error,"Global_type_0_must_be_a_class_or_interface_type_2316","Global type '{0}' must be a class or interface type."),Global_type_0_must_have_1_type_parameter_s:n(2317,e.DiagnosticCategory.Error,"Global_type_0_must_have_1_type_parameter_s_2317","Global type '{0}' must have {1} type parameter(s)."),Cannot_find_global_type_0:n(2318,e.DiagnosticCategory.Error,"Cannot_find_global_type_0_2318","Cannot find global type '{0}'."),Named_property_0_of_types_1_and_2_are_not_identical:n(2319,e.DiagnosticCategory.Error,"Named_property_0_of_types_1_and_2_are_not_identical_2319","Named property '{0}' of types '{1}' and '{2}' are not identical."),Interface_0_cannot_simultaneously_extend_types_1_and_2:n(2320,e.DiagnosticCategory.Error,"Interface_0_cannot_simultaneously_extend_types_1_and_2_2320","Interface '{0}' cannot simultaneously extend types '{1}' and '{2}'."),Excessive_stack_depth_comparing_types_0_and_1:n(2321,e.DiagnosticCategory.Error,"Excessive_stack_depth_comparing_types_0_and_1_2321","Excessive stack depth comparing types '{0}' and '{1}'."),Type_0_is_not_assignable_to_type_1:n(2322,e.DiagnosticCategory.Error,"Type_0_is_not_assignable_to_type_1_2322","Type '{0}' is not assignable to type '{1}'."),Cannot_redeclare_exported_variable_0:n(2323,e.DiagnosticCategory.Error,"Cannot_redeclare_exported_variable_0_2323","Cannot redeclare exported variable '{0}'."),Property_0_is_missing_in_type_1:n(2324,e.DiagnosticCategory.Error,"Property_0_is_missing_in_type_1_2324","Property '{0}' is missing in type '{1}'."),Property_0_is_private_in_type_1_but_not_in_type_2:n(2325,e.DiagnosticCategory.Error,"Property_0_is_private_in_type_1_but_not_in_type_2_2325","Property '{0}' is private in type '{1}' but not in type '{2}'."),Types_of_property_0_are_incompatible:n(2326,e.DiagnosticCategory.Error,"Types_of_property_0_are_incompatible_2326","Types of property '{0}' are incompatible."),Property_0_is_optional_in_type_1_but_required_in_type_2:n(2327,e.DiagnosticCategory.Error,"Property_0_is_optional_in_type_1_but_required_in_type_2_2327","Property '{0}' is optional in type '{1}' but required in type '{2}'."),Types_of_parameters_0_and_1_are_incompatible:n(2328,e.DiagnosticCategory.Error,"Types_of_parameters_0_and_1_are_incompatible_2328","Types of parameters '{0}' and '{1}' are incompatible."),Index_signature_is_missing_in_type_0:n(2329,e.DiagnosticCategory.Error,"Index_signature_is_missing_in_type_0_2329","Index signature is missing in type '{0}'."),Index_signatures_are_incompatible:n(2330,e.DiagnosticCategory.Error,"Index_signatures_are_incompatible_2330","Index signatures are incompatible."),this_cannot_be_referenced_in_a_module_or_namespace_body:n(2331,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_module_or_namespace_body_2331","'this' cannot be referenced in a module or namespace body."),this_cannot_be_referenced_in_current_location:n(2332,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_current_location_2332","'this' cannot be referenced in current location."),this_cannot_be_referenced_in_constructor_arguments:n(2333,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_constructor_arguments_2333","'this' cannot be referenced in constructor arguments."),this_cannot_be_referenced_in_a_static_property_initializer:n(2334,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_static_property_initializer_2334","'this' cannot be referenced in a static property initializer."),super_can_only_be_referenced_in_a_derived_class:n(2335,e.DiagnosticCategory.Error,"super_can_only_be_referenced_in_a_derived_class_2335","'super' can only be referenced in a derived class."),super_cannot_be_referenced_in_constructor_arguments:n(2336,e.DiagnosticCategory.Error,"super_cannot_be_referenced_in_constructor_arguments_2336","'super' cannot be referenced in constructor arguments."),Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors:n(2337,e.DiagnosticCategory.Error,"Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors_2337","Super calls are not permitted outside constructors or in nested functions inside constructors."),super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class:n(2338,e.DiagnosticCategory.Error,"super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_der_2338","'super' property access is permitted only in a constructor, member function, or member accessor of a derived class."),Property_0_does_not_exist_on_type_1:n(2339,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_2339","Property '{0}' does not exist on type '{1}'."),Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword:n(2340,e.DiagnosticCategory.Error,"Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword_2340","Only public and protected methods of the base class are accessible via the 'super' keyword."),Property_0_is_private_and_only_accessible_within_class_1:n(2341,e.DiagnosticCategory.Error,"Property_0_is_private_and_only_accessible_within_class_1_2341","Property '{0}' is private and only accessible within class '{1}'."),An_index_expression_argument_must_be_of_type_string_number_symbol_or_any:n(2342,e.DiagnosticCategory.Error,"An_index_expression_argument_must_be_of_type_string_number_symbol_or_any_2342","An index expression argument must be of type 'string', 'number', 'symbol', or 'any'."),This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_version_of_0:n(2343,e.DiagnosticCategory.Error,"This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_ve_2343","This syntax requires an imported helper named '{1}' which does not exist in '{0}'. Consider upgrading your version of '{0}'."),Type_0_does_not_satisfy_the_constraint_1:n(2344,e.DiagnosticCategory.Error,"Type_0_does_not_satisfy_the_constraint_1_2344","Type '{0}' does not satisfy the constraint '{1}'."),Argument_of_type_0_is_not_assignable_to_parameter_of_type_1:n(2345,e.DiagnosticCategory.Error,"Argument_of_type_0_is_not_assignable_to_parameter_of_type_1_2345","Argument of type '{0}' is not assignable to parameter of type '{1}'."),Call_target_does_not_contain_any_signatures:n(2346,e.DiagnosticCategory.Error,"Call_target_does_not_contain_any_signatures_2346","Call target does not contain any signatures."),Untyped_function_calls_may_not_accept_type_arguments:n(2347,e.DiagnosticCategory.Error,"Untyped_function_calls_may_not_accept_type_arguments_2347","Untyped function calls may not accept type arguments."),Value_of_type_0_is_not_callable_Did_you_mean_to_include_new:n(2348,e.DiagnosticCategory.Error,"Value_of_type_0_is_not_callable_Did_you_mean_to_include_new_2348","Value of type '{0}' is not callable. Did you mean to include 'new'?"),This_expression_is_not_callable:n(2349,e.DiagnosticCategory.Error,"This_expression_is_not_callable_2349","This expression is not callable."),Only_a_void_function_can_be_called_with_the_new_keyword:n(2350,e.DiagnosticCategory.Error,"Only_a_void_function_can_be_called_with_the_new_keyword_2350","Only a void function can be called with the 'new' keyword."),This_expression_is_not_constructable:n(2351,e.DiagnosticCategory.Error,"This_expression_is_not_constructable_2351","This expression is not constructable."),Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first:n(2352,e.DiagnosticCategory.Error,"Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the__2352","Conversion of type '{0}' to type '{1}' may be a mistake because neither type sufficiently overlaps with the other. If this was intentional, convert the expression to 'unknown' first."),Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1:n(2353,e.DiagnosticCategory.Error,"Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1_2353","Object literal may only specify known properties, and '{0}' does not exist in type '{1}'."),This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found:n(2354,e.DiagnosticCategory.Error,"This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found_2354","This syntax requires an imported helper but module '{0}' cannot be found."),A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value:n(2355,e.DiagnosticCategory.Error,"A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value_2355","A function whose declared type is neither 'void' nor 'any' must return a value."),An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type:n(2356,e.DiagnosticCategory.Error,"An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type_2356","An arithmetic operand must be of type 'any', 'number', 'bigint' or an enum type."),The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access:n(2357,e.DiagnosticCategory.Error,"The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access_2357","The operand of an increment or decrement operator must be a variable or a property access."),The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter:n(2358,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_paramete_2358","The left-hand side of an 'instanceof' expression must be of type 'any', an object type or a type parameter."),The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type:n(2359,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_F_2359","The right-hand side of an 'instanceof' expression must be of type 'any' or of a type assignable to the 'Function' interface type."),The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol:n(2360,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol_2360","The left-hand side of an 'in' expression must be of type 'any', 'string', 'number', or 'symbol'."),The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter:n(2361,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter_2361","The right-hand side of an 'in' expression must be of type 'any', an object type or a type parameter."),The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type:n(2362,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2362","The left-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type:n(2363,e.DiagnosticCategory.Error,"The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2363","The right-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access:n(2364,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access_2364","The left-hand side of an assignment expression must be a variable or a property access."),Operator_0_cannot_be_applied_to_types_1_and_2:n(2365,e.DiagnosticCategory.Error,"Operator_0_cannot_be_applied_to_types_1_and_2_2365","Operator '{0}' cannot be applied to types '{1}' and '{2}'."),Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined:n(2366,e.DiagnosticCategory.Error,"Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined_2366","Function lacks ending return statement and return type does not include 'undefined'."),This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap:n(2367,e.DiagnosticCategory.Error,"This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap_2367","This condition will always return '{0}' since the types '{1}' and '{2}' have no overlap."),Type_parameter_name_cannot_be_0:n(2368,e.DiagnosticCategory.Error,"Type_parameter_name_cannot_be_0_2368","Type parameter name cannot be '{0}'."),A_parameter_property_is_only_allowed_in_a_constructor_implementation:n(2369,e.DiagnosticCategory.Error,"A_parameter_property_is_only_allowed_in_a_constructor_implementation_2369","A parameter property is only allowed in a constructor implementation."),A_rest_parameter_must_be_of_an_array_type:n(2370,e.DiagnosticCategory.Error,"A_rest_parameter_must_be_of_an_array_type_2370","A rest parameter must be of an array type."),A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation:n(2371,e.DiagnosticCategory.Error,"A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation_2371","A parameter initializer is only allowed in a function or constructor implementation."),Parameter_0_cannot_reference_itself:n(2372,e.DiagnosticCategory.Error,"Parameter_0_cannot_reference_itself_2372","Parameter '{0}' cannot reference itself."),Parameter_0_cannot_reference_identifier_1_declared_after_it:n(2373,e.DiagnosticCategory.Error,"Parameter_0_cannot_reference_identifier_1_declared_after_it_2373","Parameter '{0}' cannot reference identifier '{1}' declared after it."),Duplicate_string_index_signature:n(2374,e.DiagnosticCategory.Error,"Duplicate_string_index_signature_2374","Duplicate string index signature."),Duplicate_number_index_signature:n(2375,e.DiagnosticCategory.Error,"Duplicate_number_index_signature_2375","Duplicate number index signature."),A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_parameter_properties_or_private_identifiers:n(2376,e.DiagnosticCategory.Error,"A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_proper_2376","A 'super' call must be the first statement in the constructor when a class contains initialized properties, parameter properties, or private identifiers."),Constructors_for_derived_classes_must_contain_a_super_call:n(2377,e.DiagnosticCategory.Error,"Constructors_for_derived_classes_must_contain_a_super_call_2377","Constructors for derived classes must contain a 'super' call."),A_get_accessor_must_return_a_value:n(2378,e.DiagnosticCategory.Error,"A_get_accessor_must_return_a_value_2378","A 'get' accessor must return a value."),Getter_and_setter_accessors_do_not_agree_in_visibility:n(2379,e.DiagnosticCategory.Error,"Getter_and_setter_accessors_do_not_agree_in_visibility_2379","Getter and setter accessors do not agree in visibility."),get_and_set_accessor_must_have_the_same_type:n(2380,e.DiagnosticCategory.Error,"get_and_set_accessor_must_have_the_same_type_2380","'get' and 'set' accessor must have the same type."),A_signature_with_an_implementation_cannot_use_a_string_literal_type:n(2381,e.DiagnosticCategory.Error,"A_signature_with_an_implementation_cannot_use_a_string_literal_type_2381","A signature with an implementation cannot use a string literal type."),Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature:n(2382,e.DiagnosticCategory.Error,"Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature_2382","Specialized overload signature is not assignable to any non-specialized signature."),Overload_signatures_must_all_be_exported_or_non_exported:n(2383,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_exported_or_non_exported_2383","Overload signatures must all be exported or non-exported."),Overload_signatures_must_all_be_ambient_or_non_ambient:n(2384,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_ambient_or_non_ambient_2384","Overload signatures must all be ambient or non-ambient."),Overload_signatures_must_all_be_public_private_or_protected:n(2385,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_public_private_or_protected_2385","Overload signatures must all be public, private or protected."),Overload_signatures_must_all_be_optional_or_required:n(2386,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_optional_or_required_2386","Overload signatures must all be optional or required."),Function_overload_must_be_static:n(2387,e.DiagnosticCategory.Error,"Function_overload_must_be_static_2387","Function overload must be static."),Function_overload_must_not_be_static:n(2388,e.DiagnosticCategory.Error,"Function_overload_must_not_be_static_2388","Function overload must not be static."),Function_implementation_name_must_be_0:n(2389,e.DiagnosticCategory.Error,"Function_implementation_name_must_be_0_2389","Function implementation name must be '{0}'."),Constructor_implementation_is_missing:n(2390,e.DiagnosticCategory.Error,"Constructor_implementation_is_missing_2390","Constructor implementation is missing."),Function_implementation_is_missing_or_not_immediately_following_the_declaration:n(2391,e.DiagnosticCategory.Error,"Function_implementation_is_missing_or_not_immediately_following_the_declaration_2391","Function implementation is missing or not immediately following the declaration."),Multiple_constructor_implementations_are_not_allowed:n(2392,e.DiagnosticCategory.Error,"Multiple_constructor_implementations_are_not_allowed_2392","Multiple constructor implementations are not allowed."),Duplicate_function_implementation:n(2393,e.DiagnosticCategory.Error,"Duplicate_function_implementation_2393","Duplicate function implementation."),This_overload_signature_is_not_compatible_with_its_implementation_signature:n(2394,e.DiagnosticCategory.Error,"This_overload_signature_is_not_compatible_with_its_implementation_signature_2394","This overload signature is not compatible with its implementation signature."),Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local:n(2395,e.DiagnosticCategory.Error,"Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local_2395","Individual declarations in merged declaration '{0}' must be all exported or all local."),Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters:n(2396,e.DiagnosticCategory.Error,"Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters_2396","Duplicate identifier 'arguments'. Compiler uses 'arguments' to initialize rest parameters."),Declaration_name_conflicts_with_built_in_global_identifier_0:n(2397,e.DiagnosticCategory.Error,"Declaration_name_conflicts_with_built_in_global_identifier_0_2397","Declaration name conflicts with built-in global identifier '{0}'."),constructor_cannot_be_used_as_a_parameter_property_name:n(2398,e.DiagnosticCategory.Error,"constructor_cannot_be_used_as_a_parameter_property_name_2398","'constructor' cannot be used as a parameter property name."),Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference:n(2399,e.DiagnosticCategory.Error,"Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference_2399","Duplicate identifier '_this'. Compiler uses variable declaration '_this' to capture 'this' reference."),Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference:n(2400,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference_2400","Expression resolves to variable declaration '_this' that compiler uses to capture 'this' reference."),Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference:n(2401,e.DiagnosticCategory.Error,"Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference_2401","Duplicate identifier '_super'. Compiler uses '_super' to capture base class reference."),Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference:n(2402,e.DiagnosticCategory.Error,"Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference_2402","Expression resolves to '_super' that compiler uses to capture base class reference."),Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2:n(2403,e.DiagnosticCategory.Error,"Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_t_2403","Subsequent variable declarations must have the same type.  Variable '{0}' must be of type '{1}', but here has type '{2}'."),The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation:n(2404,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation_2404","The left-hand side of a 'for...in' statement cannot use a type annotation."),The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any:n(2405,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any_2405","The left-hand side of a 'for...in' statement must be of type 'string' or 'any'."),The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access:n(2406,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access_2406","The left-hand side of a 'for...in' statement must be a variable or a property access."),The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0:n(2407,e.DiagnosticCategory.Error,"The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_2407","The right-hand side of a 'for...in' statement must be of type 'any', an object type or a type parameter, but here has type '{0}'."),Setters_cannot_return_a_value:n(2408,e.DiagnosticCategory.Error,"Setters_cannot_return_a_value_2408","Setters cannot return a value."),Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class:n(2409,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class_2409","Return type of constructor signature must be assignable to the instance type of the class."),The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any:n(2410,e.DiagnosticCategory.Error,"The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any_2410","The 'with' statement is not supported. All symbols in a 'with' block will have type 'any'."),Property_0_of_type_1_is_not_assignable_to_string_index_type_2:n(2411,e.DiagnosticCategory.Error,"Property_0_of_type_1_is_not_assignable_to_string_index_type_2_2411","Property '{0}' of type '{1}' is not assignable to string index type '{2}'."),Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2:n(2412,e.DiagnosticCategory.Error,"Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2_2412","Property '{0}' of type '{1}' is not assignable to numeric index type '{2}'."),Numeric_index_type_0_is_not_assignable_to_string_index_type_1:n(2413,e.DiagnosticCategory.Error,"Numeric_index_type_0_is_not_assignable_to_string_index_type_1_2413","Numeric index type '{0}' is not assignable to string index type '{1}'."),Class_name_cannot_be_0:n(2414,e.DiagnosticCategory.Error,"Class_name_cannot_be_0_2414","Class name cannot be '{0}'."),Class_0_incorrectly_extends_base_class_1:n(2415,e.DiagnosticCategory.Error,"Class_0_incorrectly_extends_base_class_1_2415","Class '{0}' incorrectly extends base class '{1}'."),Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2:n(2416,e.DiagnosticCategory.Error,"Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2_2416","Property '{0}' in type '{1}' is not assignable to the same property in base type '{2}'."),Class_static_side_0_incorrectly_extends_base_class_static_side_1:n(2417,e.DiagnosticCategory.Error,"Class_static_side_0_incorrectly_extends_base_class_static_side_1_2417","Class static side '{0}' incorrectly extends base class static side '{1}'."),Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1:n(2418,e.DiagnosticCategory.Error,"Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1_2418","Type of computed property's value is '{0}', which is not assignable to type '{1}'."),Types_of_construct_signatures_are_incompatible:n(2419,e.DiagnosticCategory.Error,"Types_of_construct_signatures_are_incompatible_2419","Types of construct signatures are incompatible."),Class_0_incorrectly_implements_interface_1:n(2420,e.DiagnosticCategory.Error,"Class_0_incorrectly_implements_interface_1_2420","Class '{0}' incorrectly implements interface '{1}'."),A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2422,e.DiagnosticCategory.Error,"A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_memb_2422","A class can only implement an object type or intersection of object types with statically known members."),Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor:n(2423,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_access_2423","Class '{0}' defines instance member function '{1}', but extended class '{2}' defines it as instance member accessor."),Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function:n(2425,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_functi_2425","Class '{0}' defines instance member property '{1}', but extended class '{2}' defines it as instance member function."),Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function:n(2426,e.DiagnosticCategory.Error,"Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_functi_2426","Class '{0}' defines instance member accessor '{1}', but extended class '{2}' defines it as instance member function."),Interface_name_cannot_be_0:n(2427,e.DiagnosticCategory.Error,"Interface_name_cannot_be_0_2427","Interface name cannot be '{0}'."),All_declarations_of_0_must_have_identical_type_parameters:n(2428,e.DiagnosticCategory.Error,"All_declarations_of_0_must_have_identical_type_parameters_2428","All declarations of '{0}' must have identical type parameters."),Interface_0_incorrectly_extends_interface_1:n(2430,e.DiagnosticCategory.Error,"Interface_0_incorrectly_extends_interface_1_2430","Interface '{0}' incorrectly extends interface '{1}'."),Enum_name_cannot_be_0:n(2431,e.DiagnosticCategory.Error,"Enum_name_cannot_be_0_2431","Enum name cannot be '{0}'."),In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element:n(2432,e.DiagnosticCategory.Error,"In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enu_2432","In an enum with multiple declarations, only one declaration can omit an initializer for its first enum element."),A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged:n(2433,e.DiagnosticCategory.Error,"A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merg_2433","A namespace declaration cannot be in a different file from a class or function with which it is merged."),A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged:n(2434,e.DiagnosticCategory.Error,"A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged_2434","A namespace declaration cannot be located prior to a class or function with which it is merged."),Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces:n(2435,e.DiagnosticCategory.Error,"Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces_2435","Ambient modules cannot be nested in other modules or namespaces."),Ambient_module_declaration_cannot_specify_relative_module_name:n(2436,e.DiagnosticCategory.Error,"Ambient_module_declaration_cannot_specify_relative_module_name_2436","Ambient module declaration cannot specify relative module name."),Module_0_is_hidden_by_a_local_declaration_with_the_same_name:n(2437,e.DiagnosticCategory.Error,"Module_0_is_hidden_by_a_local_declaration_with_the_same_name_2437","Module '{0}' is hidden by a local declaration with the same name."),Import_name_cannot_be_0:n(2438,e.DiagnosticCategory.Error,"Import_name_cannot_be_0_2438","Import name cannot be '{0}'."),Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name:n(2439,e.DiagnosticCategory.Error,"Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relati_2439","Import or export declaration in an ambient module declaration cannot reference module through relative module name."),Import_declaration_conflicts_with_local_declaration_of_0:n(2440,e.DiagnosticCategory.Error,"Import_declaration_conflicts_with_local_declaration_of_0_2440","Import declaration conflicts with local declaration of '{0}'."),Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module:n(2441,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_2441","Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module."),Types_have_separate_declarations_of_a_private_property_0:n(2442,e.DiagnosticCategory.Error,"Types_have_separate_declarations_of_a_private_property_0_2442","Types have separate declarations of a private property '{0}'."),Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2:n(2443,e.DiagnosticCategory.Error,"Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2_2443","Property '{0}' is protected but type '{1}' is not a class derived from '{2}'."),Property_0_is_protected_in_type_1_but_public_in_type_2:n(2444,e.DiagnosticCategory.Error,"Property_0_is_protected_in_type_1_but_public_in_type_2_2444","Property '{0}' is protected in type '{1}' but public in type '{2}'."),Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses:n(2445,e.DiagnosticCategory.Error,"Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses_2445","Property '{0}' is protected and only accessible within class '{1}' and its subclasses."),Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1:n(2446,e.DiagnosticCategory.Error,"Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1_2446","Property '{0}' is protected and only accessible through an instance of class '{1}'."),The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead:n(2447,e.DiagnosticCategory.Error,"The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead_2447","The '{0}' operator is not allowed for boolean types. Consider using '{1}' instead."),Block_scoped_variable_0_used_before_its_declaration:n(2448,e.DiagnosticCategory.Error,"Block_scoped_variable_0_used_before_its_declaration_2448","Block-scoped variable '{0}' used before its declaration."),Class_0_used_before_its_declaration:n(2449,e.DiagnosticCategory.Error,"Class_0_used_before_its_declaration_2449","Class '{0}' used before its declaration."),Enum_0_used_before_its_declaration:n(2450,e.DiagnosticCategory.Error,"Enum_0_used_before_its_declaration_2450","Enum '{0}' used before its declaration."),Cannot_redeclare_block_scoped_variable_0:n(2451,e.DiagnosticCategory.Error,"Cannot_redeclare_block_scoped_variable_0_2451","Cannot redeclare block-scoped variable '{0}'."),An_enum_member_cannot_have_a_numeric_name:n(2452,e.DiagnosticCategory.Error,"An_enum_member_cannot_have_a_numeric_name_2452","An enum member cannot have a numeric name."),The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_type_arguments_explicitly:n(2453,e.DiagnosticCategory.Error,"The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_typ_2453","The type argument for type parameter '{0}' cannot be inferred from the usage. Consider specifying the type arguments explicitly."),Variable_0_is_used_before_being_assigned:n(2454,e.DiagnosticCategory.Error,"Variable_0_is_used_before_being_assigned_2454","Variable '{0}' is used before being assigned."),Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0:n(2455,e.DiagnosticCategory.Error,"Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0_2455","Type argument candidate '{1}' is not a valid type argument because it is not a supertype of candidate '{0}'."),Type_alias_0_circularly_references_itself:n(2456,e.DiagnosticCategory.Error,"Type_alias_0_circularly_references_itself_2456","Type alias '{0}' circularly references itself."),Type_alias_name_cannot_be_0:n(2457,e.DiagnosticCategory.Error,"Type_alias_name_cannot_be_0_2457","Type alias name cannot be '{0}'."),An_AMD_module_cannot_have_multiple_name_assignments:n(2458,e.DiagnosticCategory.Error,"An_AMD_module_cannot_have_multiple_name_assignments_2458","An AMD module cannot have multiple name assignments."),Module_0_declares_1_locally_but_it_is_not_exported:n(2459,e.DiagnosticCategory.Error,"Module_0_declares_1_locally_but_it_is_not_exported_2459","Module '{0}' declares '{1}' locally, but it is not exported."),Module_0_declares_1_locally_but_it_is_exported_as_2:n(2460,e.DiagnosticCategory.Error,"Module_0_declares_1_locally_but_it_is_exported_as_2_2460","Module '{0}' declares '{1}' locally, but it is exported as '{2}'."),Type_0_is_not_an_array_type:n(2461,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_2461","Type '{0}' is not an array type."),A_rest_element_must_be_last_in_a_destructuring_pattern:n(2462,e.DiagnosticCategory.Error,"A_rest_element_must_be_last_in_a_destructuring_pattern_2462","A rest element must be last in a destructuring pattern."),A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature:n(2463,e.DiagnosticCategory.Error,"A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature_2463","A binding pattern parameter cannot be optional in an implementation signature."),A_computed_property_name_must_be_of_type_string_number_symbol_or_any:n(2464,e.DiagnosticCategory.Error,"A_computed_property_name_must_be_of_type_string_number_symbol_or_any_2464","A computed property name must be of type 'string', 'number', 'symbol', or 'any'."),this_cannot_be_referenced_in_a_computed_property_name:n(2465,e.DiagnosticCategory.Error,"this_cannot_be_referenced_in_a_computed_property_name_2465","'this' cannot be referenced in a computed property name."),super_cannot_be_referenced_in_a_computed_property_name:n(2466,e.DiagnosticCategory.Error,"super_cannot_be_referenced_in_a_computed_property_name_2466","'super' cannot be referenced in a computed property name."),A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type:n(2467,e.DiagnosticCategory.Error,"A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type_2467","A computed property name cannot reference a type parameter from its containing type."),Cannot_find_global_value_0:n(2468,e.DiagnosticCategory.Error,"Cannot_find_global_value_0_2468","Cannot find global value '{0}'."),The_0_operator_cannot_be_applied_to_type_symbol:n(2469,e.DiagnosticCategory.Error,"The_0_operator_cannot_be_applied_to_type_symbol_2469","The '{0}' operator cannot be applied to type 'symbol'."),Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object:n(2470,e.DiagnosticCategory.Error,"Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object_2470","'Symbol' reference does not refer to the global Symbol constructor object."),A_computed_property_name_of_the_form_0_must_be_of_type_symbol:n(2471,e.DiagnosticCategory.Error,"A_computed_property_name_of_the_form_0_must_be_of_type_symbol_2471","A computed property name of the form '{0}' must be of type 'symbol'."),Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher:n(2472,e.DiagnosticCategory.Error,"Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher_2472","Spread operator in 'new' expressions is only available when targeting ECMAScript 5 and higher."),Enum_declarations_must_all_be_const_or_non_const:n(2473,e.DiagnosticCategory.Error,"Enum_declarations_must_all_be_const_or_non_const_2473","Enum declarations must all be const or non-const."),const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values:n(2474,e.DiagnosticCategory.Error,"const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values_2474","const enum member initializers can only contain literal values and other computed enum values."),const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query:n(2475,e.DiagnosticCategory.Error,"const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_im_2475","'const' enums can only be used in property or index access expressions or the right hand side of an import declaration or export assignment or type query."),A_const_enum_member_can_only_be_accessed_using_a_string_literal:n(2476,e.DiagnosticCategory.Error,"A_const_enum_member_can_only_be_accessed_using_a_string_literal_2476","A const enum member can only be accessed using a string literal."),const_enum_member_initializer_was_evaluated_to_a_non_finite_value:n(2477,e.DiagnosticCategory.Error,"const_enum_member_initializer_was_evaluated_to_a_non_finite_value_2477","'const' enum member initializer was evaluated to a non-finite value."),const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN:n(2478,e.DiagnosticCategory.Error,"const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN_2478","'const' enum member initializer was evaluated to disallowed value 'NaN'."),Property_0_does_not_exist_on_const_enum_1:n(2479,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_const_enum_1_2479","Property '{0}' does not exist on 'const' enum '{1}'."),let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations:n(2480,e.DiagnosticCategory.Error,"let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations_2480","'let' is not allowed to be used as a name in 'let' or 'const' declarations."),Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1:n(2481,e.DiagnosticCategory.Error,"Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1_2481","Cannot initialize outer scoped variable '{0}' in the same scope as block scoped declaration '{1}'."),The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation:n(2483,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation_2483","The left-hand side of a 'for...of' statement cannot use a type annotation."),Export_declaration_conflicts_with_exported_declaration_of_0:n(2484,e.DiagnosticCategory.Error,"Export_declaration_conflicts_with_exported_declaration_of_0_2484","Export declaration conflicts with exported declaration of '{0}'."),The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access:n(2487,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access_2487","The left-hand side of a 'for...of' statement must be a variable or a property access."),Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2488,e.DiagnosticCategory.Error,"Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator_2488","Type '{0}' must have a '[Symbol.iterator]()' method that returns an iterator."),An_iterator_must_have_a_next_method:n(2489,e.DiagnosticCategory.Error,"An_iterator_must_have_a_next_method_2489","An iterator must have a 'next()' method."),The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property:n(2490,e.DiagnosticCategory.Error,"The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property_2490","The type returned by the '{0}()' method of an iterator must have a 'value' property."),The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern:n(2491,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern_2491","The left-hand side of a 'for...in' statement cannot be a destructuring pattern."),Cannot_redeclare_identifier_0_in_catch_clause:n(2492,e.DiagnosticCategory.Error,"Cannot_redeclare_identifier_0_in_catch_clause_2492","Cannot redeclare identifier '{0}' in catch clause."),Tuple_type_0_of_length_1_has_no_element_at_index_2:n(2493,e.DiagnosticCategory.Error,"Tuple_type_0_of_length_1_has_no_element_at_index_2_2493","Tuple type '{0}' of length '{1}' has no element at index '{2}'."),Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher:n(2494,e.DiagnosticCategory.Error,"Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher_2494","Using a string in a 'for...of' statement is only supported in ECMAScript 5 and higher."),Type_0_is_not_an_array_type_or_a_string_type:n(2495,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_2495","Type '{0}' is not an array type or a string type."),The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression:n(2496,e.DiagnosticCategory.Error,"The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_stand_2496","The 'arguments' object cannot be referenced in an arrow function in ES3 and ES5. Consider using a standard function expression."),This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export:n(2497,e.DiagnosticCategory.Error,"This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_2497","This module can only be referenced with ECMAScript imports/exports by turning on the '{0}' flag and referencing its default export."),Module_0_uses_export_and_cannot_be_used_with_export_Asterisk:n(2498,e.DiagnosticCategory.Error,"Module_0_uses_export_and_cannot_be_used_with_export_Asterisk_2498","Module '{0}' uses 'export =' and cannot be used with 'export *'."),An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments:n(2499,e.DiagnosticCategory.Error,"An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments_2499","An interface can only extend an identifier/qualified-name with optional type arguments."),A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments:n(2500,e.DiagnosticCategory.Error,"A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments_2500","A class can only implement an identifier/qualified-name with optional type arguments."),A_rest_element_cannot_contain_a_binding_pattern:n(2501,e.DiagnosticCategory.Error,"A_rest_element_cannot_contain_a_binding_pattern_2501","A rest element cannot contain a binding pattern."),_0_is_referenced_directly_or_indirectly_in_its_own_type_annotation:n(2502,e.DiagnosticCategory.Error,"_0_is_referenced_directly_or_indirectly_in_its_own_type_annotation_2502","'{0}' is referenced directly or indirectly in its own type annotation."),Cannot_find_namespace_0:n(2503,e.DiagnosticCategory.Error,"Cannot_find_namespace_0_2503","Cannot find namespace '{0}'."),Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator:n(2504,e.DiagnosticCategory.Error,"Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator_2504","Type '{0}' must have a '[Symbol.asyncIterator]()' method that returns an async iterator."),A_generator_cannot_have_a_void_type_annotation:n(2505,e.DiagnosticCategory.Error,"A_generator_cannot_have_a_void_type_annotation_2505","A generator cannot have a 'void' type annotation."),_0_is_referenced_directly_or_indirectly_in_its_own_base_expression:n(2506,e.DiagnosticCategory.Error,"_0_is_referenced_directly_or_indirectly_in_its_own_base_expression_2506","'{0}' is referenced directly or indirectly in its own base expression."),Type_0_is_not_a_constructor_function_type:n(2507,e.DiagnosticCategory.Error,"Type_0_is_not_a_constructor_function_type_2507","Type '{0}' is not a constructor function type."),No_base_constructor_has_the_specified_number_of_type_arguments:n(2508,e.DiagnosticCategory.Error,"No_base_constructor_has_the_specified_number_of_type_arguments_2508","No base constructor has the specified number of type arguments."),Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members:n(2509,e.DiagnosticCategory.Error,"Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_2509","Base constructor return type '{0}' is not an object type or intersection of object types with statically known members."),Base_constructors_must_all_have_the_same_return_type:n(2510,e.DiagnosticCategory.Error,"Base_constructors_must_all_have_the_same_return_type_2510","Base constructors must all have the same return type."),Cannot_create_an_instance_of_an_abstract_class:n(2511,e.DiagnosticCategory.Error,"Cannot_create_an_instance_of_an_abstract_class_2511","Cannot create an instance of an abstract class."),Overload_signatures_must_all_be_abstract_or_non_abstract:n(2512,e.DiagnosticCategory.Error,"Overload_signatures_must_all_be_abstract_or_non_abstract_2512","Overload signatures must all be abstract or non-abstract."),Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression:n(2513,e.DiagnosticCategory.Error,"Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression_2513","Abstract method '{0}' in class '{1}' cannot be accessed via super expression."),Classes_containing_abstract_methods_must_be_marked_abstract:n(2514,e.DiagnosticCategory.Error,"Classes_containing_abstract_methods_must_be_marked_abstract_2514","Classes containing abstract methods must be marked abstract."),Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2:n(2515,e.DiagnosticCategory.Error,"Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2_2515","Non-abstract class '{0}' does not implement inherited abstract member '{1}' from class '{2}'."),All_declarations_of_an_abstract_method_must_be_consecutive:n(2516,e.DiagnosticCategory.Error,"All_declarations_of_an_abstract_method_must_be_consecutive_2516","All declarations of an abstract method must be consecutive."),Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type:n(2517,e.DiagnosticCategory.Error,"Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type_2517","Cannot assign an abstract constructor type to a non-abstract constructor type."),A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard:n(2518,e.DiagnosticCategory.Error,"A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard_2518","A 'this'-based type guard is not compatible with a parameter-based type guard."),An_async_iterator_must_have_a_next_method:n(2519,e.DiagnosticCategory.Error,"An_async_iterator_must_have_a_next_method_2519","An async iterator must have a 'next()' method."),Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions:n(2520,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions_2520","Duplicate identifier '{0}'. Compiler uses declaration '{1}' to support async functions."),Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions:n(2521,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions_2521","Expression resolves to variable declaration '{0}' that compiler uses to support async functions."),The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method:n(2522,e.DiagnosticCategory.Error,"The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_usi_2522","The 'arguments' object cannot be referenced in an async function or method in ES3 and ES5. Consider using a standard function or method."),yield_expressions_cannot_be_used_in_a_parameter_initializer:n(2523,e.DiagnosticCategory.Error,"yield_expressions_cannot_be_used_in_a_parameter_initializer_2523","'yield' expressions cannot be used in a parameter initializer."),await_expressions_cannot_be_used_in_a_parameter_initializer:n(2524,e.DiagnosticCategory.Error,"await_expressions_cannot_be_used_in_a_parameter_initializer_2524","'await' expressions cannot be used in a parameter initializer."),Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value:n(2525,e.DiagnosticCategory.Error,"Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value_2525","Initializer provides no value for this binding element and the binding element has no default value."),A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface:n(2526,e.DiagnosticCategory.Error,"A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface_2526","A 'this' type is available only in a non-static member of a class or interface."),The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary:n(2527,e.DiagnosticCategory.Error,"The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary_2527","The inferred type of '{0}' references an inaccessible '{1}' type. A type annotation is necessary."),A_module_cannot_have_multiple_default_exports:n(2528,e.DiagnosticCategory.Error,"A_module_cannot_have_multiple_default_exports_2528","A module cannot have multiple default exports."),Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions:n(2529,e.DiagnosticCategory.Error,"Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_func_2529","Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module containing async functions."),Property_0_is_incompatible_with_index_signature:n(2530,e.DiagnosticCategory.Error,"Property_0_is_incompatible_with_index_signature_2530","Property '{0}' is incompatible with index signature."),Object_is_possibly_null:n(2531,e.DiagnosticCategory.Error,"Object_is_possibly_null_2531","Object is possibly 'null'."),Object_is_possibly_undefined:n(2532,e.DiagnosticCategory.Error,"Object_is_possibly_undefined_2532","Object is possibly 'undefined'."),Object_is_possibly_null_or_undefined:n(2533,e.DiagnosticCategory.Error,"Object_is_possibly_null_or_undefined_2533","Object is possibly 'null' or 'undefined'."),A_function_returning_never_cannot_have_a_reachable_end_point:n(2534,e.DiagnosticCategory.Error,"A_function_returning_never_cannot_have_a_reachable_end_point_2534","A function returning 'never' cannot have a reachable end point."),Enum_type_0_has_members_with_initializers_that_are_not_literals:n(2535,e.DiagnosticCategory.Error,"Enum_type_0_has_members_with_initializers_that_are_not_literals_2535","Enum type '{0}' has members with initializers that are not literals."),Type_0_cannot_be_used_to_index_type_1:n(2536,e.DiagnosticCategory.Error,"Type_0_cannot_be_used_to_index_type_1_2536","Type '{0}' cannot be used to index type '{1}'."),Type_0_has_no_matching_index_signature_for_type_1:n(2537,e.DiagnosticCategory.Error,"Type_0_has_no_matching_index_signature_for_type_1_2537","Type '{0}' has no matching index signature for type '{1}'."),Type_0_cannot_be_used_as_an_index_type:n(2538,e.DiagnosticCategory.Error,"Type_0_cannot_be_used_as_an_index_type_2538","Type '{0}' cannot be used as an index type."),Cannot_assign_to_0_because_it_is_not_a_variable:n(2539,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_not_a_variable_2539","Cannot assign to '{0}' because it is not a variable."),Cannot_assign_to_0_because_it_is_a_read_only_property:n(2540,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_a_read_only_property_2540","Cannot assign to '{0}' because it is a read-only property."),The_target_of_an_assignment_must_be_a_variable_or_a_property_access:n(2541,e.DiagnosticCategory.Error,"The_target_of_an_assignment_must_be_a_variable_or_a_property_access_2541","The target of an assignment must be a variable or a property access."),Index_signature_in_type_0_only_permits_reading:n(2542,e.DiagnosticCategory.Error,"Index_signature_in_type_0_only_permits_reading_2542","Index signature in type '{0}' only permits reading."),Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference:n(2543,e.DiagnosticCategory.Error,"Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_me_2543","Duplicate identifier '_newTarget'. Compiler uses variable declaration '_newTarget' to capture 'new.target' meta-property reference."),Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference:n(2544,e.DiagnosticCategory.Error,"Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta__2544","Expression resolves to variable declaration '_newTarget' that compiler uses to capture 'new.target' meta-property reference."),A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any:n(2545,e.DiagnosticCategory.Error,"A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any_2545","A mixin class must have a constructor with a single rest parameter of type 'any[]'."),The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property:n(2547,e.DiagnosticCategory.Error,"The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_pro_2547","The type returned by the '{0}()' method of an async iterator must be a promise for a type with a 'value' property."),Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2548,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator_2548","Type '{0}' is not an array type or does not have a '[Symbol.iterator]()' method that returns an iterator."),Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator:n(2549,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns__2549","Type '{0}' is not an array type or a string type or does not have a '[Symbol.iterator]()' method that returns an iterator."),Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2_or_later:n(2550,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_c_2550","Property '{0}' does not exist on type '{1}'. Do you need to change your target library? Try changing the `lib` compiler option to '{2}' or later."),Property_0_does_not_exist_on_type_1_Did_you_mean_2:n(2551,e.DiagnosticCategory.Error,"Property_0_does_not_exist_on_type_1_Did_you_mean_2_2551","Property '{0}' does not exist on type '{1}'. Did you mean '{2}'?"),Cannot_find_name_0_Did_you_mean_1:n(2552,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_1_2552","Cannot find name '{0}'. Did you mean '{1}'?"),Computed_values_are_not_permitted_in_an_enum_with_string_valued_members:n(2553,e.DiagnosticCategory.Error,"Computed_values_are_not_permitted_in_an_enum_with_string_valued_members_2553","Computed values are not permitted in an enum with string valued members."),Expected_0_arguments_but_got_1:n(2554,e.DiagnosticCategory.Error,"Expected_0_arguments_but_got_1_2554","Expected {0} arguments, but got {1}."),Expected_at_least_0_arguments_but_got_1:n(2555,e.DiagnosticCategory.Error,"Expected_at_least_0_arguments_but_got_1_2555","Expected at least {0} arguments, but got {1}."),Expected_0_arguments_but_got_1_or_more:n(2556,e.DiagnosticCategory.Error,"Expected_0_arguments_but_got_1_or_more_2556","Expected {0} arguments, but got {1} or more."),Expected_at_least_0_arguments_but_got_1_or_more:n(2557,e.DiagnosticCategory.Error,"Expected_at_least_0_arguments_but_got_1_or_more_2557","Expected at least {0} arguments, but got {1} or more."),Expected_0_type_arguments_but_got_1:n(2558,e.DiagnosticCategory.Error,"Expected_0_type_arguments_but_got_1_2558","Expected {0} type arguments, but got {1}."),Type_0_has_no_properties_in_common_with_type_1:n(2559,e.DiagnosticCategory.Error,"Type_0_has_no_properties_in_common_with_type_1_2559","Type '{0}' has no properties in common with type '{1}'."),Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it:n(2560,e.DiagnosticCategory.Error,"Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it_2560","Value of type '{0}' has no properties in common with type '{1}'. Did you mean to call it?"),Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2:n(2561,e.DiagnosticCategory.Error,"Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_writ_2561","Object literal may only specify known properties, but '{0}' does not exist in type '{1}'. Did you mean to write '{2}'?"),Base_class_expressions_cannot_reference_class_type_parameters:n(2562,e.DiagnosticCategory.Error,"Base_class_expressions_cannot_reference_class_type_parameters_2562","Base class expressions cannot reference class type parameters."),The_containing_function_or_module_body_is_too_large_for_control_flow_analysis:n(2563,e.DiagnosticCategory.Error,"The_containing_function_or_module_body_is_too_large_for_control_flow_analysis_2563","The containing function or module body is too large for control flow analysis."),Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor:n(2564,e.DiagnosticCategory.Error,"Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor_2564","Property '{0}' has no initializer and is not definitely assigned in the constructor."),Property_0_is_used_before_being_assigned:n(2565,e.DiagnosticCategory.Error,"Property_0_is_used_before_being_assigned_2565","Property '{0}' is used before being assigned."),A_rest_element_cannot_have_a_property_name:n(2566,e.DiagnosticCategory.Error,"A_rest_element_cannot_have_a_property_name_2566","A rest element cannot have a property name."),Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations:n(2567,e.DiagnosticCategory.Error,"Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations_2567","Enum declarations can only merge with namespace or other enum declarations."),Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators:n(2569,e.DiagnosticCategory.Error,"Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterati_2569","Type '{0}' is not an array type or a string type. Use compiler option '--downlevelIteration' to allow iterating of iterators."),Object_is_of_type_unknown:n(2571,e.DiagnosticCategory.Error,"Object_is_of_type_unknown_2571","Object is of type 'unknown'."),Rest_signatures_are_incompatible:n(2572,e.DiagnosticCategory.Error,"Rest_signatures_are_incompatible_2572","Rest signatures are incompatible."),Property_0_is_incompatible_with_rest_element_type:n(2573,e.DiagnosticCategory.Error,"Property_0_is_incompatible_with_rest_element_type_2573","Property '{0}' is incompatible with rest element type."),A_rest_element_type_must_be_an_array_type:n(2574,e.DiagnosticCategory.Error,"A_rest_element_type_must_be_an_array_type_2574","A rest element type must be an array type."),No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments:n(2575,e.DiagnosticCategory.Error,"No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments_2575","No overload expects {0} arguments, but overloads do exist that expect either {1} or {2} arguments."),Property_0_is_a_static_member_of_type_1:n(2576,e.DiagnosticCategory.Error,"Property_0_is_a_static_member_of_type_1_2576","Property '{0}' is a static member of type '{1}'."),Return_type_annotation_circularly_references_itself:n(2577,e.DiagnosticCategory.Error,"Return_type_annotation_circularly_references_itself_2577","Return type annotation circularly references itself."),Unused_ts_expect_error_directive:n(2578,e.DiagnosticCategory.Error,"Unused_ts_expect_error_directive_2578","Unused '@ts-expect-error' directive."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode:n(2580,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashno_2580","Cannot find name '{0}'. Do you need to install type definitions for node? Try `npm i --save-dev @types/node`."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery:n(2581,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slash_2581","Cannot find name '{0}'. Do you need to install type definitions for jQuery? Try `npm i --save-dev @types/jquery`."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha:n(2582,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_type_2582","Cannot find name '{0}'. Do you need to install type definitions for a test runner? Try `npm i --save-dev @types/jest` or `npm i --save-dev @types/mocha`."),Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_1_or_later:n(2583,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2583","Cannot find name '{0}'. Do you need to change your target library? Try changing the `lib` compiler option to '{1}' or later."),Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom:n(2584,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2584","Cannot find name '{0}'. Do you need to change your target library? Try changing the `lib` compiler option to include 'dom'."),_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later:n(2585,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_2585","'{0}' only refers to a type, but is being used as a value here. Do you need to change your target library? Try changing the `lib` compiler option to es2015 or later."),Enum_type_0_circularly_references_itself:n(2586,e.DiagnosticCategory.Error,"Enum_type_0_circularly_references_itself_2586","Enum type '{0}' circularly references itself."),JSDoc_type_0_circularly_references_itself:n(2587,e.DiagnosticCategory.Error,"JSDoc_type_0_circularly_references_itself_2587","JSDoc type '{0}' circularly references itself."),Cannot_assign_to_0_because_it_is_a_constant:n(2588,e.DiagnosticCategory.Error,"Cannot_assign_to_0_because_it_is_a_constant_2588","Cannot assign to '{0}' because it is a constant."),Type_instantiation_is_excessively_deep_and_possibly_infinite:n(2589,e.DiagnosticCategory.Error,"Type_instantiation_is_excessively_deep_and_possibly_infinite_2589","Type instantiation is excessively deep and possibly infinite."),Expression_produces_a_union_type_that_is_too_complex_to_represent:n(2590,e.DiagnosticCategory.Error,"Expression_produces_a_union_type_that_is_too_complex_to_represent_2590","Expression produces a union type that is too complex to represent."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig:n(2591,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashno_2591","Cannot find name '{0}'. Do you need to install type definitions for node? Try `npm i --save-dev @types/node` and then add `node` to the types field in your tsconfig."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig:n(2592,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slash_2592","Cannot find name '{0}'. Do you need to install type definitions for jQuery? Try `npm i --save-dev @types/jquery` and then add `jquery` to the types field in your tsconfig."),Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig:n(2593,e.DiagnosticCategory.Error,"Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_type_2593","Cannot find name '{0}'. Do you need to install type definitions for a test runner? Try `npm i --save-dev @types/jest` or `npm i --save-dev @types/mocha` and then add `jest` or `mocha` to the types field in your tsconfig."),This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the_0_flag:n(2594,e.DiagnosticCategory.Error,"This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the__2594","This module is declared with using 'export =', and can only be used with a default import when using the '{0}' flag."),_0_can_only_be_imported_by_using_a_default_import:n(2595,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_using_a_default_import_2595","'{0}' can only be imported by using a default import."),_0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import:n(2596,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import_2596","'{0}' can only be imported by turning on the 'esModuleInterop' flag and using a default import."),_0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import:n(2597,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import_2597","'{0}' can only be imported by using a 'require' call or by using a default import."),_0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import:n(2598,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using__2598","'{0}' can only be imported by using a 'require' call or by turning on the 'esModuleInterop' flag and using a default import."),JSX_element_attributes_type_0_may_not_be_a_union_type:n(2600,e.DiagnosticCategory.Error,"JSX_element_attributes_type_0_may_not_be_a_union_type_2600","JSX element attributes type '{0}' may not be a union type."),The_return_type_of_a_JSX_element_constructor_must_return_an_object_type:n(2601,e.DiagnosticCategory.Error,"The_return_type_of_a_JSX_element_constructor_must_return_an_object_type_2601","The return type of a JSX element constructor must return an object type."),JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist:n(2602,e.DiagnosticCategory.Error,"JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist_2602","JSX element implicitly has type 'any' because the global type 'JSX.Element' does not exist."),Property_0_in_type_1_is_not_assignable_to_type_2:n(2603,e.DiagnosticCategory.Error,"Property_0_in_type_1_is_not_assignable_to_type_2_2603","Property '{0}' in type '{1}' is not assignable to type '{2}'."),JSX_element_type_0_does_not_have_any_construct_or_call_signatures:n(2604,e.DiagnosticCategory.Error,"JSX_element_type_0_does_not_have_any_construct_or_call_signatures_2604","JSX element type '{0}' does not have any construct or call signatures."),JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements:n(2605,e.DiagnosticCategory.Error,"JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements_2605","JSX element type '{0}' is not a constructor function for JSX elements."),Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property:n(2606,e.DiagnosticCategory.Error,"Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property_2606","Property '{0}' of JSX spread attribute is not assignable to target property."),JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property:n(2607,e.DiagnosticCategory.Error,"JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property_2607","JSX element class does not support attributes because it does not have a '{0}' property."),The_global_type_JSX_0_may_not_have_more_than_one_property:n(2608,e.DiagnosticCategory.Error,"The_global_type_JSX_0_may_not_have_more_than_one_property_2608","The global type 'JSX.{0}' may not have more than one property."),JSX_spread_child_must_be_an_array_type:n(2609,e.DiagnosticCategory.Error,"JSX_spread_child_must_be_an_array_type_2609","JSX spread child must be an array type."),_0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property:n(2610,e.DiagnosticCategory.Error,"_0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property_2610","'{0}' is defined as an accessor in class '{1}', but is overridden here in '{2}' as an instance property."),_0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor:n(2611,e.DiagnosticCategory.Error,"_0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor_2611","'{0}' is defined as a property in class '{1}', but is overridden here in '{2}' as an accessor."),Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration:n(2612,e.DiagnosticCategory.Error,"Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_2612","Property '{0}' will overwrite the base property in '{1}'. If this is intentional, add an initializer. Otherwise, add a 'declare' modifier or remove the redundant declaration."),Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead:n(2613,e.DiagnosticCategory.Error,"Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead_2613","Module '{0}' has no default export. Did you mean to use 'import { {1} } from {0}' instead?"),Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead:n(2614,e.DiagnosticCategory.Error,"Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead_2614","Module '{0}' has no exported member '{1}'. Did you mean to use 'import {1} from {0}' instead?"),Type_of_property_0_circularly_references_itself_in_mapped_type_1:n(2615,e.DiagnosticCategory.Error,"Type_of_property_0_circularly_references_itself_in_mapped_type_1_2615","Type of property '{0}' circularly references itself in mapped type '{1}'."),_0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import:n(2616,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import_2616","'{0}' can only be imported by using 'import {1} = require({2})' or a default import."),_0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import:n(2617,e.DiagnosticCategory.Error,"_0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_us_2617","'{0}' can only be imported by using 'import {1} = require({2})' or by turning on the 'esModuleInterop' flag and using a default import."),Source_has_0_element_s_but_target_requires_1:n(2618,e.DiagnosticCategory.Error,"Source_has_0_element_s_but_target_requires_1_2618","Source has {0} element(s) but target requires {1}."),Source_has_0_element_s_but_target_allows_only_1:n(2619,e.DiagnosticCategory.Error,"Source_has_0_element_s_but_target_allows_only_1_2619","Source has {0} element(s) but target allows only {1}."),Target_requires_0_element_s_but_source_may_have_fewer:n(2620,e.DiagnosticCategory.Error,"Target_requires_0_element_s_but_source_may_have_fewer_2620","Target requires {0} element(s) but source may have fewer."),Target_allows_only_0_element_s_but_source_may_have_more:n(2621,e.DiagnosticCategory.Error,"Target_allows_only_0_element_s_but_source_may_have_more_2621","Target allows only {0} element(s) but source may have more."),Element_at_index_0_is_variadic_in_one_type_but_not_in_the_other:n(2622,e.DiagnosticCategory.Error,"Element_at_index_0_is_variadic_in_one_type_but_not_in_the_other_2622","Element at index {0} is variadic in one type but not in the other."),Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity:n(2649,e.DiagnosticCategory.Error,"Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity_2649","Cannot augment module '{0}' with value exports because it resolves to a non-module entity."),A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums:n(2651,e.DiagnosticCategory.Error,"A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_memb_2651","A member initializer in a enum declaration cannot reference members declared after it, including members defined in other enums."),Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead:n(2652,e.DiagnosticCategory.Error,"Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_d_2652","Merged declaration '{0}' cannot include a default export declaration. Consider adding a separate 'export default {0}' declaration instead."),Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1:n(2653,e.DiagnosticCategory.Error,"Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1_2653","Non-abstract class expression does not implement inherited abstract member '{0}' from class '{1}'."),Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_package_author_to_update_the_package_definition:n(2654,e.DiagnosticCategory.Error,"Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_pack_2654","Exported external package typings file cannot contain tripleslash references. Please contact the package author to update the package definition."),Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_the_package_definition:n(2656,e.DiagnosticCategory.Error,"Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_2656","Exported external package typings file '{0}' is not a module. Please contact the package author to update the package definition."),JSX_expressions_must_have_one_parent_element:n(2657,e.DiagnosticCategory.Error,"JSX_expressions_must_have_one_parent_element_2657","JSX expressions must have one parent element."),Type_0_provides_no_match_for_the_signature_1:n(2658,e.DiagnosticCategory.Error,"Type_0_provides_no_match_for_the_signature_1_2658","Type '{0}' provides no match for the signature '{1}'."),super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher:n(2659,e.DiagnosticCategory.Error,"super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_highe_2659","'super' is only allowed in members of object literal expressions when option 'target' is 'ES2015' or higher."),super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions:n(2660,e.DiagnosticCategory.Error,"super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions_2660","'super' can only be referenced in members of derived classes or object literal expressions."),Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module:n(2661,e.DiagnosticCategory.Error,"Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module_2661","Cannot export '{0}'. Only local declarations can be exported from a module."),Cannot_find_name_0_Did_you_mean_the_static_member_1_0:n(2662,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_the_static_member_1_0_2662","Cannot find name '{0}'. Did you mean the static member '{1}.{0}'?"),Cannot_find_name_0_Did_you_mean_the_instance_member_this_0:n(2663,e.DiagnosticCategory.Error,"Cannot_find_name_0_Did_you_mean_the_instance_member_this_0_2663","Cannot find name '{0}'. Did you mean the instance member 'this.{0}'?"),Invalid_module_name_in_augmentation_module_0_cannot_be_found:n(2664,e.DiagnosticCategory.Error,"Invalid_module_name_in_augmentation_module_0_cannot_be_found_2664","Invalid module name in augmentation, module '{0}' cannot be found."),Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented:n(2665,e.DiagnosticCategory.Error,"Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augm_2665","Invalid module name in augmentation. Module '{0}' resolves to an untyped module at '{1}', which cannot be augmented."),Exports_and_export_assignments_are_not_permitted_in_module_augmentations:n(2666,e.DiagnosticCategory.Error,"Exports_and_export_assignments_are_not_permitted_in_module_augmentations_2666","Exports and export assignments are not permitted in module augmentations."),Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module:n(2667,e.DiagnosticCategory.Error,"Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_mod_2667","Imports are not permitted in module augmentations. Consider moving them to the enclosing external module."),export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible:n(2668,e.DiagnosticCategory.Error,"export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always__2668","'export' modifier cannot be applied to ambient modules and module augmentations since they are always visible."),Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations:n(2669,e.DiagnosticCategory.Error,"Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_2669","Augmentations for the global scope can only be directly nested in external modules or ambient module declarations."),Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context:n(2670,e.DiagnosticCategory.Error,"Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambien_2670","Augmentations for the global scope should have 'declare' modifier unless they appear in already ambient context."),Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity:n(2671,e.DiagnosticCategory.Error,"Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity_2671","Cannot augment module '{0}' because it resolves to a non-module entity."),Cannot_assign_a_0_constructor_type_to_a_1_constructor_type:n(2672,e.DiagnosticCategory.Error,"Cannot_assign_a_0_constructor_type_to_a_1_constructor_type_2672","Cannot assign a '{0}' constructor type to a '{1}' constructor type."),Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration:n(2673,e.DiagnosticCategory.Error,"Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration_2673","Constructor of class '{0}' is private and only accessible within the class declaration."),Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration:n(2674,e.DiagnosticCategory.Error,"Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration_2674","Constructor of class '{0}' is protected and only accessible within the class declaration."),Cannot_extend_a_class_0_Class_constructor_is_marked_as_private:n(2675,e.DiagnosticCategory.Error,"Cannot_extend_a_class_0_Class_constructor_is_marked_as_private_2675","Cannot extend a class '{0}'. Class constructor is marked as private."),Accessors_must_both_be_abstract_or_non_abstract:n(2676,e.DiagnosticCategory.Error,"Accessors_must_both_be_abstract_or_non_abstract_2676","Accessors must both be abstract or non-abstract."),A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type:n(2677,e.DiagnosticCategory.Error,"A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type_2677","A type predicate's type must be assignable to its parameter's type."),Type_0_is_not_comparable_to_type_1:n(2678,e.DiagnosticCategory.Error,"Type_0_is_not_comparable_to_type_1_2678","Type '{0}' is not comparable to type '{1}'."),A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void:n(2679,e.DiagnosticCategory.Error,"A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void_2679","A function that is called with the 'new' keyword cannot have a 'this' type that is 'void'."),A_0_parameter_must_be_the_first_parameter:n(2680,e.DiagnosticCategory.Error,"A_0_parameter_must_be_the_first_parameter_2680","A '{0}' parameter must be the first parameter."),A_constructor_cannot_have_a_this_parameter:n(2681,e.DiagnosticCategory.Error,"A_constructor_cannot_have_a_this_parameter_2681","A constructor cannot have a 'this' parameter."),get_and_set_accessor_must_have_the_same_this_type:n(2682,e.DiagnosticCategory.Error,"get_and_set_accessor_must_have_the_same_this_type_2682","'get' and 'set' accessor must have the same 'this' type."),this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation:n(2683,e.DiagnosticCategory.Error,"this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_2683","'this' implicitly has type 'any' because it does not have a type annotation."),The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1:n(2684,e.DiagnosticCategory.Error,"The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1_2684","The 'this' context of type '{0}' is not assignable to method's 'this' of type '{1}'."),The_this_types_of_each_signature_are_incompatible:n(2685,e.DiagnosticCategory.Error,"The_this_types_of_each_signature_are_incompatible_2685","The 'this' types of each signature are incompatible."),_0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead:n(2686,e.DiagnosticCategory.Error,"_0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead_2686","'{0}' refers to a UMD global, but the current file is a module. Consider adding an import instead."),All_declarations_of_0_must_have_identical_modifiers:n(2687,e.DiagnosticCategory.Error,"All_declarations_of_0_must_have_identical_modifiers_2687","All declarations of '{0}' must have identical modifiers."),Cannot_find_type_definition_file_for_0:n(2688,e.DiagnosticCategory.Error,"Cannot_find_type_definition_file_for_0_2688","Cannot find type definition file for '{0}'."),Cannot_extend_an_interface_0_Did_you_mean_implements:n(2689,e.DiagnosticCategory.Error,"Cannot_extend_an_interface_0_Did_you_mean_implements_2689","Cannot extend an interface '{0}'. Did you mean 'implements'?"),_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0:n(2690,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0_2690","'{0}' only refers to a type, but is being used as a value here. Did you mean to use '{1} in {0}'?"),An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead:n(2691,e.DiagnosticCategory.Error,"An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead_2691","An import path cannot end with a '{0}' extension. Consider importing '{1}' instead."),_0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible:n(2692,e.DiagnosticCategory.Error,"_0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible_2692","'{0}' is a primitive, but '{1}' is a wrapper object. Prefer using '{0}' when possible."),_0_only_refers_to_a_type_but_is_being_used_as_a_value_here:n(2693,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_2693","'{0}' only refers to a type, but is being used as a value here."),Namespace_0_has_no_exported_member_1:n(2694,e.DiagnosticCategory.Error,"Namespace_0_has_no_exported_member_1_2694","Namespace '{0}' has no exported member '{1}'."),Left_side_of_comma_operator_is_unused_and_has_no_side_effects:n(2695,e.DiagnosticCategory.Error,"Left_side_of_comma_operator_is_unused_and_has_no_side_effects_2695","Left side of comma operator is unused and has no side effects.",!0),The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead:n(2696,e.DiagnosticCategory.Error,"The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead_2696","The 'Object' type is assignable to very few other types. Did you mean to use the 'any' type instead?"),An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:n(2697,e.DiagnosticCategory.Error,"An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_in_2697","An async function or method must return a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your `--lib` option."),Spread_types_may_only_be_created_from_object_types:n(2698,e.DiagnosticCategory.Error,"Spread_types_may_only_be_created_from_object_types_2698","Spread types may only be created from object types."),Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1:n(2699,e.DiagnosticCategory.Error,"Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1_2699","Static property '{0}' conflicts with built-in property 'Function.{0}' of constructor function '{1}'."),Rest_types_may_only_be_created_from_object_types:n(2700,e.DiagnosticCategory.Error,"Rest_types_may_only_be_created_from_object_types_2700","Rest types may only be created from object types."),The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access:n(2701,e.DiagnosticCategory.Error,"The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access_2701","The target of an object rest assignment must be a variable or a property access."),_0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here:n(2702,e.DiagnosticCategory.Error,"_0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here_2702","'{0}' only refers to a type, but is being used as a namespace here."),The_operand_of_a_delete_operator_must_be_a_property_reference:n(2703,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_must_be_a_property_reference_2703","The operand of a 'delete' operator must be a property reference."),The_operand_of_a_delete_operator_cannot_be_a_read_only_property:n(2704,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_cannot_be_a_read_only_property_2704","The operand of a 'delete' operator cannot be a read-only property."),An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:n(2705,e.DiagnosticCategory.Error,"An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_de_2705","An async function or method in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your `--lib` option."),Required_type_parameters_may_not_follow_optional_type_parameters:n(2706,e.DiagnosticCategory.Error,"Required_type_parameters_may_not_follow_optional_type_parameters_2706","Required type parameters may not follow optional type parameters."),Generic_type_0_requires_between_1_and_2_type_arguments:n(2707,e.DiagnosticCategory.Error,"Generic_type_0_requires_between_1_and_2_type_arguments_2707","Generic type '{0}' requires between {1} and {2} type arguments."),Cannot_use_namespace_0_as_a_value:n(2708,e.DiagnosticCategory.Error,"Cannot_use_namespace_0_as_a_value_2708","Cannot use namespace '{0}' as a value."),Cannot_use_namespace_0_as_a_type:n(2709,e.DiagnosticCategory.Error,"Cannot_use_namespace_0_as_a_type_2709","Cannot use namespace '{0}' as a type."),_0_are_specified_twice_The_attribute_named_0_will_be_overwritten:n(2710,e.DiagnosticCategory.Error,"_0_are_specified_twice_The_attribute_named_0_will_be_overwritten_2710","'{0}' are specified twice. The attribute named '{0}' will be overwritten."),A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:n(2711,e.DiagnosticCategory.Error,"A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES20_2711","A dynamic import call returns a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your `--lib` option."),A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:n(2712,e.DiagnosticCategory.Error,"A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declarat_2712","A dynamic import call in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your `--lib` option."),Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1:n(2713,e.DiagnosticCategory.Error,"Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_p_2713","Cannot access '{0}.{1}' because '{0}' is a type, but not a namespace. Did you mean to retrieve the type of the property '{1}' in '{0}' with '{0}[\"{1}\"]'?"),The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context:n(2714,e.DiagnosticCategory.Error,"The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context_2714","The expression of an export assignment must be an identifier or qualified name in an ambient context."),Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor:n(2715,e.DiagnosticCategory.Error,"Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor_2715","Abstract property '{0}' in class '{1}' cannot be accessed in the constructor."),Type_parameter_0_has_a_circular_default:n(2716,e.DiagnosticCategory.Error,"Type_parameter_0_has_a_circular_default_2716","Type parameter '{0}' has a circular default."),Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2:n(2717,e.DiagnosticCategory.Error,"Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_t_2717","Subsequent property declarations must have the same type.  Property '{0}' must be of type '{1}', but here has type '{2}'."),Duplicate_property_0:n(2718,e.DiagnosticCategory.Error,"Duplicate_property_0_2718","Duplicate property '{0}'."),Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated:n(2719,e.DiagnosticCategory.Error,"Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated_2719","Type '{0}' is not assignable to type '{1}'. Two different types with this name exist, but they are unrelated."),Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass:n(2720,e.DiagnosticCategory.Error,"Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclas_2720","Class '{0}' incorrectly implements class '{1}'. Did you mean to extend '{1}' and inherit its members as a subclass?"),Cannot_invoke_an_object_which_is_possibly_null:n(2721,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_null_2721","Cannot invoke an object which is possibly 'null'."),Cannot_invoke_an_object_which_is_possibly_undefined:n(2722,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_undefined_2722","Cannot invoke an object which is possibly 'undefined'."),Cannot_invoke_an_object_which_is_possibly_null_or_undefined:n(2723,e.DiagnosticCategory.Error,"Cannot_invoke_an_object_which_is_possibly_null_or_undefined_2723","Cannot invoke an object which is possibly 'null' or 'undefined'."),_0_has_no_exported_member_named_1_Did_you_mean_2:n(2724,e.DiagnosticCategory.Error,"_0_has_no_exported_member_named_1_Did_you_mean_2_2724","'{0}' has no exported member named '{1}'. Did you mean '{2}'?"),Class_name_cannot_be_Object_when_targeting_ES5_with_module_0:n(2725,e.DiagnosticCategory.Error,"Class_name_cannot_be_Object_when_targeting_ES5_with_module_0_2725","Class name cannot be 'Object' when targeting ES5 with module {0}."),Cannot_find_lib_definition_for_0:n(2726,e.DiagnosticCategory.Error,"Cannot_find_lib_definition_for_0_2726","Cannot find lib definition for '{0}'."),Cannot_find_lib_definition_for_0_Did_you_mean_1:n(2727,e.DiagnosticCategory.Error,"Cannot_find_lib_definition_for_0_Did_you_mean_1_2727","Cannot find lib definition for '{0}'. Did you mean '{1}'?"),_0_is_declared_here:n(2728,e.DiagnosticCategory.Message,"_0_is_declared_here_2728","'{0}' is declared here."),Property_0_is_used_before_its_initialization:n(2729,e.DiagnosticCategory.Error,"Property_0_is_used_before_its_initialization_2729","Property '{0}' is used before its initialization."),An_arrow_function_cannot_have_a_this_parameter:n(2730,e.DiagnosticCategory.Error,"An_arrow_function_cannot_have_a_this_parameter_2730","An arrow function cannot have a 'this' parameter."),Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String:n(2731,e.DiagnosticCategory.Error,"Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_i_2731","Implicit conversion of a 'symbol' to a 'string' will fail at runtime. Consider wrapping this expression in 'String(...)'."),Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension:n(2732,e.DiagnosticCategory.Error,"Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension_2732","Cannot find module '{0}'. Consider using '--resolveJsonModule' to import module with '.json' extension."),Property_0_was_also_declared_here:n(2733,e.DiagnosticCategory.Error,"Property_0_was_also_declared_here_2733","Property '{0}' was also declared here."),Are_you_missing_a_semicolon:n(2734,e.DiagnosticCategory.Error,"Are_you_missing_a_semicolon_2734","Are you missing a semicolon?"),Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1:n(2735,e.DiagnosticCategory.Error,"Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1_2735","Did you mean for '{0}' to be constrained to type 'new (...args: any[]) => {1}'?"),Operator_0_cannot_be_applied_to_type_1:n(2736,e.DiagnosticCategory.Error,"Operator_0_cannot_be_applied_to_type_1_2736","Operator '{0}' cannot be applied to type '{1}'."),BigInt_literals_are_not_available_when_targeting_lower_than_ES2020:n(2737,e.DiagnosticCategory.Error,"BigInt_literals_are_not_available_when_targeting_lower_than_ES2020_2737","BigInt literals are not available when targeting lower than ES2020."),An_outer_value_of_this_is_shadowed_by_this_container:n(2738,e.DiagnosticCategory.Message,"An_outer_value_of_this_is_shadowed_by_this_container_2738","An outer value of 'this' is shadowed by this container."),Type_0_is_missing_the_following_properties_from_type_1_Colon_2:n(2739,e.DiagnosticCategory.Error,"Type_0_is_missing_the_following_properties_from_type_1_Colon_2_2739","Type '{0}' is missing the following properties from type '{1}': {2}"),Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more:n(2740,e.DiagnosticCategory.Error,"Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more_2740","Type '{0}' is missing the following properties from type '{1}': {2}, and {3} more."),Property_0_is_missing_in_type_1_but_required_in_type_2:n(2741,e.DiagnosticCategory.Error,"Property_0_is_missing_in_type_1_but_required_in_type_2_2741","Property '{0}' is missing in type '{1}' but required in type '{2}'."),The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary:n(2742,e.DiagnosticCategory.Error,"The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_a_2742","The inferred type of '{0}' cannot be named without a reference to '{1}'. This is likely not portable. A type annotation is necessary."),No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments:n(2743,e.DiagnosticCategory.Error,"No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments_2743","No overload expects {0} type arguments, but overloads do exist that expect either {1} or {2} type arguments."),Type_parameter_defaults_can_only_reference_previously_declared_type_parameters:n(2744,e.DiagnosticCategory.Error,"Type_parameter_defaults_can_only_reference_previously_declared_type_parameters_2744","Type parameter defaults can only reference previously declared type parameters."),This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided:n(2745,e.DiagnosticCategory.Error,"This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_pr_2745","This JSX tag's '{0}' prop expects type '{1}' which requires multiple children, but only a single child was provided."),This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided:n(2746,e.DiagnosticCategory.Error,"This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided_2746","This JSX tag's '{0}' prop expects a single child of type '{1}', but multiple children were provided."),_0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2:n(2747,e.DiagnosticCategory.Error,"_0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_t_2747","'{0}' components don't accept text as child elements. Text in JSX has the type 'string', but the expected type of '{1}' is '{2}'."),Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided:n(2748,e.DiagnosticCategory.Error,"Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided_2748","Cannot access ambient const enums when the '--isolatedModules' flag is provided."),_0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0:n(2749,e.DiagnosticCategory.Error,"_0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0_2749","'{0}' refers to a value, but is being used as a type here. Did you mean 'typeof {0}'?"),The_implementation_signature_is_declared_here:n(2750,e.DiagnosticCategory.Error,"The_implementation_signature_is_declared_here_2750","The implementation signature is declared here."),Circularity_originates_in_type_at_this_location:n(2751,e.DiagnosticCategory.Error,"Circularity_originates_in_type_at_this_location_2751","Circularity originates in type at this location."),The_first_export_default_is_here:n(2752,e.DiagnosticCategory.Error,"The_first_export_default_is_here_2752","The first export default is here."),Another_export_default_is_here:n(2753,e.DiagnosticCategory.Error,"Another_export_default_is_here_2753","Another export default is here."),super_may_not_use_type_arguments:n(2754,e.DiagnosticCategory.Error,"super_may_not_use_type_arguments_2754","'super' may not use type arguments."),No_constituent_of_type_0_is_callable:n(2755,e.DiagnosticCategory.Error,"No_constituent_of_type_0_is_callable_2755","No constituent of type '{0}' is callable."),Not_all_constituents_of_type_0_are_callable:n(2756,e.DiagnosticCategory.Error,"Not_all_constituents_of_type_0_are_callable_2756","Not all constituents of type '{0}' are callable."),Type_0_has_no_call_signatures:n(2757,e.DiagnosticCategory.Error,"Type_0_has_no_call_signatures_2757","Type '{0}' has no call signatures."),Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_other:n(2758,e.DiagnosticCategory.Error,"Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_2758","Each member of the union type '{0}' has signatures, but none of those signatures are compatible with each other."),No_constituent_of_type_0_is_constructable:n(2759,e.DiagnosticCategory.Error,"No_constituent_of_type_0_is_constructable_2759","No constituent of type '{0}' is constructable."),Not_all_constituents_of_type_0_are_constructable:n(2760,e.DiagnosticCategory.Error,"Not_all_constituents_of_type_0_are_constructable_2760","Not all constituents of type '{0}' are constructable."),Type_0_has_no_construct_signatures:n(2761,e.DiagnosticCategory.Error,"Type_0_has_no_construct_signatures_2761","Type '{0}' has no construct signatures."),Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_with_each_other:n(2762,e.DiagnosticCategory.Error,"Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_2762","Each member of the union type '{0}' has construct signatures, but none of those signatures are compatible with each other."),Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_send_0:n(2763,e.DiagnosticCategory.Error,"Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_s_2763","Cannot iterate value because the 'next' method of its iterator expects type '{1}', but for-of will always send '{0}'."),Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_always_send_0:n(2764,e.DiagnosticCategory.Error,"Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_al_2764","Cannot iterate value because the 'next' method of its iterator expects type '{1}', but array spread will always send '{0}'."),Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring_will_always_send_0:n(2765,e.DiagnosticCategory.Error,"Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring__2765","Cannot iterate value because the 'next' method of its iterator expects type '{1}', but array destructuring will always send '{0}'."),Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_containing_generator_will_always_send_0:n(2766,e.DiagnosticCategory.Error,"Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_co_2766","Cannot delegate iteration to value because the 'next' method of its iterator expects type '{1}', but the containing generator will always send '{0}'."),The_0_property_of_an_iterator_must_be_a_method:n(2767,e.DiagnosticCategory.Error,"The_0_property_of_an_iterator_must_be_a_method_2767","The '{0}' property of an iterator must be a method."),The_0_property_of_an_async_iterator_must_be_a_method:n(2768,e.DiagnosticCategory.Error,"The_0_property_of_an_async_iterator_must_be_a_method_2768","The '{0}' property of an async iterator must be a method."),No_overload_matches_this_call:n(2769,e.DiagnosticCategory.Error,"No_overload_matches_this_call_2769","No overload matches this call."),The_last_overload_gave_the_following_error:n(2770,e.DiagnosticCategory.Error,"The_last_overload_gave_the_following_error_2770","The last overload gave the following error."),The_last_overload_is_declared_here:n(2771,e.DiagnosticCategory.Error,"The_last_overload_is_declared_here_2771","The last overload is declared here."),Overload_0_of_1_2_gave_the_following_error:n(2772,e.DiagnosticCategory.Error,"Overload_0_of_1_2_gave_the_following_error_2772","Overload {0} of {1}, '{2}', gave the following error."),Did_you_forget_to_use_await:n(2773,e.DiagnosticCategory.Error,"Did_you_forget_to_use_await_2773","Did you forget to use 'await'?"),This_condition_will_always_return_true_since_the_function_is_always_defined_Did_you_mean_to_call_it_instead:n(2774,e.DiagnosticCategory.Error,"This_condition_will_always_return_true_since_the_function_is_always_defined_Did_you_mean_to_call_it__2774","This condition will always return true since the function is always defined. Did you mean to call it instead?"),Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation:n(2775,e.DiagnosticCategory.Error,"Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation_2775","Assertions require every name in the call target to be declared with an explicit type annotation."),Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name:n(2776,e.DiagnosticCategory.Error,"Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name_2776","Assertions require the call target to be an identifier or qualified name."),The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access:n(2777,e.DiagnosticCategory.Error,"The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access_2777","The operand of an increment or decrement operator may not be an optional property access."),The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access:n(2778,e.DiagnosticCategory.Error,"The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access_2778","The target of an object rest assignment may not be an optional property access."),The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access:n(2779,e.DiagnosticCategory.Error,"The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access_2779","The left-hand side of an assignment expression may not be an optional property access."),The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access:n(2780,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access_2780","The left-hand side of a 'for...in' statement may not be an optional property access."),The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access:n(2781,e.DiagnosticCategory.Error,"The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access_2781","The left-hand side of a 'for...of' statement may not be an optional property access."),_0_needs_an_explicit_type_annotation:n(2782,e.DiagnosticCategory.Message,"_0_needs_an_explicit_type_annotation_2782","'{0}' needs an explicit type annotation."),_0_is_specified_more_than_once_so_this_usage_will_be_overwritten:n(2783,e.DiagnosticCategory.Error,"_0_is_specified_more_than_once_so_this_usage_will_be_overwritten_2783","'{0}' is specified more than once, so this usage will be overwritten."),get_and_set_accessors_cannot_declare_this_parameters:n(2784,e.DiagnosticCategory.Error,"get_and_set_accessors_cannot_declare_this_parameters_2784","'get' and 'set' accessors cannot declare 'this' parameters."),This_spread_always_overwrites_this_property:n(2785,e.DiagnosticCategory.Error,"This_spread_always_overwrites_this_property_2785","This spread always overwrites this property."),_0_cannot_be_used_as_a_JSX_component:n(2786,e.DiagnosticCategory.Error,"_0_cannot_be_used_as_a_JSX_component_2786","'{0}' cannot be used as a JSX component."),Its_return_type_0_is_not_a_valid_JSX_element:n(2787,e.DiagnosticCategory.Error,"Its_return_type_0_is_not_a_valid_JSX_element_2787","Its return type '{0}' is not a valid JSX element."),Its_instance_type_0_is_not_a_valid_JSX_element:n(2788,e.DiagnosticCategory.Error,"Its_instance_type_0_is_not_a_valid_JSX_element_2788","Its instance type '{0}' is not a valid JSX element."),Its_element_type_0_is_not_a_valid_JSX_element:n(2789,e.DiagnosticCategory.Error,"Its_element_type_0_is_not_a_valid_JSX_element_2789","Its element type '{0}' is not a valid JSX element."),The_operand_of_a_delete_operator_must_be_optional:n(2790,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_must_be_optional_2790","The operand of a 'delete' operator must be optional."),Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_later:n(2791,e.DiagnosticCategory.Error,"Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_lat_2791","Exponentiation cannot be performed on 'bigint' values unless the 'target' option is set to 'es2016' or later."),Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option:n(2792,e.DiagnosticCategory.Error,"Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_th_2792","Cannot find module '{0}'. Did you mean to set the 'moduleResolution' option to 'node', or to add aliases to the 'paths' option?"),The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_are_not_externally_visible:n(2793,e.DiagnosticCategory.Error,"The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_2793","The call would have succeeded against this implementation, but implementation signatures of overloads are not externally visible."),Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise:n(2794,e.DiagnosticCategory.Error,"Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise_2794","Expected {0} arguments, but got {1}. Did you forget to include 'void' in your type argument to 'Promise'?"),The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types:n(2795,e.DiagnosticCategory.Error,"The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types_2795","The 'intrinsic' keyword can only be used to declare compiler provided intrinsic types."),It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tagged_template_expression_which_cannot_be_invoked:n(2796,e.DiagnosticCategory.Error,"It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tag_2796","It is likely that you are missing a comma to separate these two template expressions. They form a tagged template expression which cannot be invoked."),Import_declaration_0_is_using_private_name_1:n(4e3,e.DiagnosticCategory.Error,"Import_declaration_0_is_using_private_name_1_4000","Import declaration '{0}' is using private name '{1}'."),Type_parameter_0_of_exported_class_has_or_is_using_private_name_1:n(4002,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_class_has_or_is_using_private_name_1_4002","Type parameter '{0}' of exported class has or is using private name '{1}'."),Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1:n(4004,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1_4004","Type parameter '{0}' of exported interface has or is using private name '{1}'."),Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1:n(4006,e.DiagnosticCategory.Error,"Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4006","Type parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1:n(4008,e.DiagnosticCategory.Error,"Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4008","Type parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:n(4010,e.DiagnosticCategory.Error,"Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4010","Type parameter '{0}' of public static method from exported class has or is using private name '{1}'."),Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:n(4012,e.DiagnosticCategory.Error,"Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4012","Type parameter '{0}' of public method from exported class has or is using private name '{1}'."),Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1:n(4014,e.DiagnosticCategory.Error,"Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4014","Type parameter '{0}' of method from exported interface has or is using private name '{1}'."),Type_parameter_0_of_exported_function_has_or_is_using_private_name_1:n(4016,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_function_has_or_is_using_private_name_1_4016","Type parameter '{0}' of exported function has or is using private name '{1}'."),Implements_clause_of_exported_class_0_has_or_is_using_private_name_1:n(4019,e.DiagnosticCategory.Error,"Implements_clause_of_exported_class_0_has_or_is_using_private_name_1_4019","Implements clause of exported class '{0}' has or is using private name '{1}'."),extends_clause_of_exported_class_0_has_or_is_using_private_name_1:n(4020,e.DiagnosticCategory.Error,"extends_clause_of_exported_class_0_has_or_is_using_private_name_1_4020","'extends' clause of exported class '{0}' has or is using private name '{1}'."),extends_clause_of_exported_class_has_or_is_using_private_name_0:n(4021,e.DiagnosticCategory.Error,"extends_clause_of_exported_class_has_or_is_using_private_name_0_4021","'extends' clause of exported class has or is using private name '{0}'."),extends_clause_of_exported_interface_0_has_or_is_using_private_name_1:n(4022,e.DiagnosticCategory.Error,"extends_clause_of_exported_interface_0_has_or_is_using_private_name_1_4022","'extends' clause of exported interface '{0}' has or is using private name '{1}'."),Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4023,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4023","Exported variable '{0}' has or is using name '{1}' from external module {2} but cannot be named."),Exported_variable_0_has_or_is_using_name_1_from_private_module_2:n(4024,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_name_1_from_private_module_2_4024","Exported variable '{0}' has or is using name '{1}' from private module '{2}'."),Exported_variable_0_has_or_is_using_private_name_1:n(4025,e.DiagnosticCategory.Error,"Exported_variable_0_has_or_is_using_private_name_1_4025","Exported variable '{0}' has or is using private name '{1}'."),Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4026,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot__4026","Public static property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4027,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4027","Public static property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:n(4028,e.DiagnosticCategory.Error,"Public_static_property_0_of_exported_class_has_or_is_using_private_name_1_4028","Public static property '{0}' of exported class has or is using private name '{1}'."),Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4029,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_name_4029","Public property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4030,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4030","Public property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_property_0_of_exported_class_has_or_is_using_private_name_1:n(4031,e.DiagnosticCategory.Error,"Public_property_0_of_exported_class_has_or_is_using_private_name_1_4031","Public property '{0}' of exported class has or is using private name '{1}'."),Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4032,e.DiagnosticCategory.Error,"Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4032","Property '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),Property_0_of_exported_interface_has_or_is_using_private_name_1:n(4033,e.DiagnosticCategory.Error,"Property_0_of_exported_interface_has_or_is_using_private_name_1_4033","Property '{0}' of exported interface has or is using private name '{1}'."),Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4034,e.DiagnosticCategory.Error,"Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_mod_4034","Parameter type of public static setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1:n(4035,e.DiagnosticCategory.Error,"Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1_4035","Parameter type of public static setter '{0}' from exported class has or is using private name '{1}'."),Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4036,e.DiagnosticCategory.Error,"Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4036","Parameter type of public setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1:n(4037,e.DiagnosticCategory.Error,"Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1_4037","Parameter type of public setter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4038,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_modul_4038","Return type of public static getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4039,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_4039","Return type of public static getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1:n(4040,e.DiagnosticCategory.Error,"Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1_4040","Return type of public static getter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4041,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_4041","Return type of public getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4042,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4042","Return type of public getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1:n(4043,e.DiagnosticCategory.Error,"Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1_4043","Return type of public getter '{0}' from exported class has or is using private name '{1}'."),Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4044,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_mod_4044","Return type of constructor signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0:n(4045,e.DiagnosticCategory.Error,"Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0_4045","Return type of constructor signature from exported interface has or is using private name '{0}'."),Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4046,e.DiagnosticCategory.Error,"Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4046","Return type of call signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0:n(4047,e.DiagnosticCategory.Error,"Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0_4047","Return type of call signature from exported interface has or is using private name '{0}'."),Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4048,e.DiagnosticCategory.Error,"Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4048","Return type of index signature from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0:n(4049,e.DiagnosticCategory.Error,"Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0_4049","Return type of index signature from exported interface has or is using private name '{0}'."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4050,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module__4050","Return type of public static method from exported class has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:n(4051,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4051","Return type of public static method from exported class has or is using name '{0}' from private module '{1}'."),Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0:n(4052,e.DiagnosticCategory.Error,"Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0_4052","Return type of public static method from exported class has or is using private name '{0}'."),Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4053,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_c_4053","Return type of public method from exported class has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:n(4054,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4054","Return type of public method from exported class has or is using name '{0}' from private module '{1}'."),Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0:n(4055,e.DiagnosticCategory.Error,"Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0_4055","Return type of public method from exported class has or is using private name '{0}'."),Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1:n(4056,e.DiagnosticCategory.Error,"Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4056","Return type of method from exported interface has or is using name '{0}' from private module '{1}'."),Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0:n(4057,e.DiagnosticCategory.Error,"Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0_4057","Return type of method from exported interface has or is using private name '{0}'."),Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:n(4058,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named_4058","Return type of exported function has or is using name '{0}' from external module {1} but cannot be named."),Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1:n(4059,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1_4059","Return type of exported function has or is using name '{0}' from private module '{1}'."),Return_type_of_exported_function_has_or_is_using_private_name_0:n(4060,e.DiagnosticCategory.Error,"Return_type_of_exported_function_has_or_is_using_private_name_0_4060","Return type of exported function has or is using private name '{0}'."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4061,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_can_4061","Parameter '{0}' of constructor from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4062,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2_4062","Parameter '{0}' of constructor from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1:n(4063,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1_4063","Parameter '{0}' of constructor from exported class has or is using private name '{1}'."),Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4064,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_mod_4064","Parameter '{0}' of constructor signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1:n(4065,e.DiagnosticCategory.Error,"Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4065","Parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4066,e.DiagnosticCategory.Error,"Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4066","Parameter '{0}' of call signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1:n(4067,e.DiagnosticCategory.Error,"Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4067","Parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4068,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module__4068","Parameter '{0}' of public static method from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4069,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4069","Parameter '{0}' of public static method from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:n(4070,e.DiagnosticCategory.Error,"Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4070","Parameter '{0}' of public static method from exported class has or is using private name '{1}'."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4071,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_c_4071","Parameter '{0}' of public method from exported class has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:n(4072,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4072","Parameter '{0}' of public method from exported class has or is using name '{1}' from private module '{2}'."),Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:n(4073,e.DiagnosticCategory.Error,"Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4073","Parameter '{0}' of public method from exported class has or is using private name '{1}'."),Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4074,e.DiagnosticCategory.Error,"Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4074","Parameter '{0}' of method from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1:n(4075,e.DiagnosticCategory.Error,"Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4075","Parameter '{0}' of method from exported interface has or is using private name '{1}'."),Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4076,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4076","Parameter '{0}' of exported function has or is using name '{1}' from external module {2} but cannot be named."),Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2:n(4077,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2_4077","Parameter '{0}' of exported function has or is using name '{1}' from private module '{2}'."),Parameter_0_of_exported_function_has_or_is_using_private_name_1:n(4078,e.DiagnosticCategory.Error,"Parameter_0_of_exported_function_has_or_is_using_private_name_1_4078","Parameter '{0}' of exported function has or is using private name '{1}'."),Exported_type_alias_0_has_or_is_using_private_name_1:n(4081,e.DiagnosticCategory.Error,"Exported_type_alias_0_has_or_is_using_private_name_1_4081","Exported type alias '{0}' has or is using private name '{1}'."),Default_export_of_the_module_has_or_is_using_private_name_0:n(4082,e.DiagnosticCategory.Error,"Default_export_of_the_module_has_or_is_using_private_name_0_4082","Default export of the module has or is using private name '{0}'."),Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1:n(4083,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1_4083","Type parameter '{0}' of exported type alias has or is using private name '{1}'."),Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict:n(4090,e.DiagnosticCategory.Error,"Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_librar_4090","Conflicting definitions for '{0}' found at '{1}' and '{2}'. Consider installing a specific version of this library to resolve the conflict."),Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4091,e.DiagnosticCategory.Error,"Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4091","Parameter '{0}' of index signature from exported interface has or is using name '{1}' from private module '{2}'."),Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1:n(4092,e.DiagnosticCategory.Error,"Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1_4092","Parameter '{0}' of index signature from exported interface has or is using private name '{1}'."),Property_0_of_exported_class_expression_may_not_be_private_or_protected:n(4094,e.DiagnosticCategory.Error,"Property_0_of_exported_class_expression_may_not_be_private_or_protected_4094","Property '{0}' of exported class expression may not be private or protected."),Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4095,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_4095","Public static method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4096,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4096","Public static method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_static_method_0_of_exported_class_has_or_is_using_private_name_1:n(4097,e.DiagnosticCategory.Error,"Public_static_method_0_of_exported_class_has_or_is_using_private_name_1_4097","Public static method '{0}' of exported class has or is using private name '{1}'."),Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4098,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4098","Public method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:n(4099,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4099","Public method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),Public_method_0_of_exported_class_has_or_is_using_private_name_1:n(4100,e.DiagnosticCategory.Error,"Public_method_0_of_exported_class_has_or_is_using_private_name_1_4100","Public method '{0}' of exported class has or is using private name '{1}'."),Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:n(4101,e.DiagnosticCategory.Error,"Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4101","Method '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),Method_0_of_exported_interface_has_or_is_using_private_name_1:n(4102,e.DiagnosticCategory.Error,"Method_0_of_exported_interface_has_or_is_using_private_name_1_4102","Method '{0}' of exported interface has or is using private name '{1}'."),Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1:n(4103,e.DiagnosticCategory.Error,"Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1_4103","Type parameter '{0}' of exported mapped object type is using private name '{1}'."),The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1:n(4104,e.DiagnosticCategory.Error,"The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1_4104","The type '{0}' is 'readonly' and cannot be assigned to the mutable type '{1}'."),Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter:n(4105,e.DiagnosticCategory.Error,"Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter_4105","Private or protected member '{0}' cannot be accessed on a type parameter."),Parameter_0_of_accessor_has_or_is_using_private_name_1:n(4106,e.DiagnosticCategory.Error,"Parameter_0_of_accessor_has_or_is_using_private_name_1_4106","Parameter '{0}' of accessor has or is using private name '{1}'."),Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2:n(4107,e.DiagnosticCategory.Error,"Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2_4107","Parameter '{0}' of accessor has or is using name '{1}' from private module '{2}'."),Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:n(4108,e.DiagnosticCategory.Error,"Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4108","Parameter '{0}' of accessor has or is using name '{1}' from external module '{2}' but cannot be named."),Type_arguments_for_0_circularly_reference_themselves:n(4109,e.DiagnosticCategory.Error,"Type_arguments_for_0_circularly_reference_themselves_4109","Type arguments for '{0}' circularly reference themselves."),Tuple_type_arguments_circularly_reference_themselves:n(4110,e.DiagnosticCategory.Error,"Tuple_type_arguments_circularly_reference_themselves_4110","Tuple type arguments circularly reference themselves."),The_current_host_does_not_support_the_0_option:n(5001,e.DiagnosticCategory.Error,"The_current_host_does_not_support_the_0_option_5001","The current host does not support the '{0}' option."),Cannot_find_the_common_subdirectory_path_for_the_input_files:n(5009,e.DiagnosticCategory.Error,"Cannot_find_the_common_subdirectory_path_for_the_input_files_5009","Cannot find the common subdirectory path for the input files."),File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:n(5010,e.DiagnosticCategory.Error,"File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0_5010","File specification cannot end in a recursive directory wildcard ('**'): '{0}'."),Cannot_read_file_0_Colon_1:n(5012,e.DiagnosticCategory.Error,"Cannot_read_file_0_Colon_1_5012","Cannot read file '{0}': {1}."),Failed_to_parse_file_0_Colon_1:n(5014,e.DiagnosticCategory.Error,"Failed_to_parse_file_0_Colon_1_5014","Failed to parse file '{0}': {1}."),Unknown_compiler_option_0:n(5023,e.DiagnosticCategory.Error,"Unknown_compiler_option_0_5023","Unknown compiler option '{0}'."),Compiler_option_0_requires_a_value_of_type_1:n(5024,e.DiagnosticCategory.Error,"Compiler_option_0_requires_a_value_of_type_1_5024","Compiler option '{0}' requires a value of type {1}."),Unknown_compiler_option_0_Did_you_mean_1:n(5025,e.DiagnosticCategory.Error,"Unknown_compiler_option_0_Did_you_mean_1_5025","Unknown compiler option '{0}'. Did you mean '{1}'?"),Could_not_write_file_0_Colon_1:n(5033,e.DiagnosticCategory.Error,"Could_not_write_file_0_Colon_1_5033","Could not write file '{0}': {1}."),Option_project_cannot_be_mixed_with_source_files_on_a_command_line:n(5042,e.DiagnosticCategory.Error,"Option_project_cannot_be_mixed_with_source_files_on_a_command_line_5042","Option 'project' cannot be mixed with source files on a command line."),Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher:n(5047,e.DiagnosticCategory.Error,"Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES_5047","Option 'isolatedModules' can only be used when either option '--module' is provided or option 'target' is 'ES2015' or higher."),Option_0_cannot_be_specified_when_option_target_is_ES3:n(5048,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_when_option_target_is_ES3_5048","Option '{0}' cannot be specified when option 'target' is 'ES3'."),Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided:n(5051,e.DiagnosticCategory.Error,"Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided_5051","Option '{0} can only be used when either option '--inlineSourceMap' or option '--sourceMap' is provided."),Option_0_cannot_be_specified_without_specifying_option_1:n(5052,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_without_specifying_option_1_5052","Option '{0}' cannot be specified without specifying option '{1}'."),Option_0_cannot_be_specified_with_option_1:n(5053,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_with_option_1_5053","Option '{0}' cannot be specified with option '{1}'."),A_tsconfig_json_file_is_already_defined_at_Colon_0:n(5054,e.DiagnosticCategory.Error,"A_tsconfig_json_file_is_already_defined_at_Colon_0_5054","A 'tsconfig.json' file is already defined at: '{0}'."),Cannot_write_file_0_because_it_would_overwrite_input_file:n(5055,e.DiagnosticCategory.Error,"Cannot_write_file_0_because_it_would_overwrite_input_file_5055","Cannot write file '{0}' because it would overwrite input file."),Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files:n(5056,e.DiagnosticCategory.Error,"Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files_5056","Cannot write file '{0}' because it would be overwritten by multiple input files."),Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0:n(5057,e.DiagnosticCategory.Error,"Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0_5057","Cannot find a tsconfig.json file at the specified directory: '{0}'."),The_specified_path_does_not_exist_Colon_0:n(5058,e.DiagnosticCategory.Error,"The_specified_path_does_not_exist_Colon_0_5058","The specified path does not exist: '{0}'."),Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier:n(5059,e.DiagnosticCategory.Error,"Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier_5059","Invalid value for '--reactNamespace'. '{0}' is not a valid identifier."),Pattern_0_can_have_at_most_one_Asterisk_character:n(5061,e.DiagnosticCategory.Error,"Pattern_0_can_have_at_most_one_Asterisk_character_5061","Pattern '{0}' can have at most one '*' character."),Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character:n(5062,e.DiagnosticCategory.Error,"Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character_5062","Substitution '{0}' in pattern '{1}' can have at most one '*' character."),Substitutions_for_pattern_0_should_be_an_array:n(5063,e.DiagnosticCategory.Error,"Substitutions_for_pattern_0_should_be_an_array_5063","Substitutions for pattern '{0}' should be an array."),Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2:n(5064,e.DiagnosticCategory.Error,"Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2_5064","Substitution '{0}' for pattern '{1}' has incorrect type, expected 'string', got '{2}'."),File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0:n(5065,e.DiagnosticCategory.Error,"File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildca_5065","File specification cannot contain a parent directory ('..') that appears after a recursive directory wildcard ('**'): '{0}'."),Substitutions_for_pattern_0_shouldn_t_be_an_empty_array:n(5066,e.DiagnosticCategory.Error,"Substitutions_for_pattern_0_shouldn_t_be_an_empty_array_5066","Substitutions for pattern '{0}' shouldn't be an empty array."),Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name:n(5067,e.DiagnosticCategory.Error,"Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name_5067","Invalid value for 'jsxFactory'. '{0}' is not a valid identifier or qualified-name."),Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig:n(5068,e.DiagnosticCategory.Error,"Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript__5068","Adding a tsconfig.json file will help organize projects that contain both TypeScript and JavaScript files. Learn more at https://aka.ms/tsconfig."),Option_0_cannot_be_specified_without_specifying_option_1_or_option_2:n(5069,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_without_specifying_option_1_or_option_2_5069","Option '{0}' cannot be specified without specifying option '{1}' or option '{2}'."),Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy:n(5070,e.DiagnosticCategory.Error,"Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy_5070","Option '--resolveJsonModule' cannot be specified without 'node' module resolution strategy."),Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext:n(5071,e.DiagnosticCategory.Error,"Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_5071","Option '--resolveJsonModule' can only be specified when module code generation is 'commonjs', 'amd', 'es2015' or 'esNext'."),Unknown_build_option_0:n(5072,e.DiagnosticCategory.Error,"Unknown_build_option_0_5072","Unknown build option '{0}'."),Build_option_0_requires_a_value_of_type_1:n(5073,e.DiagnosticCategory.Error,"Build_option_0_requires_a_value_of_type_1_5073","Build option '{0}' requires a value of type {1}."),Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBuildInfoFile_is_specified:n(5074,e.DiagnosticCategory.Error,"Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBui_5074","Option '--incremental' can only be specified using tsconfig, emitting to single file or when option `--tsBuildInfoFile` is specified."),_0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_constraint_2:n(5075,e.DiagnosticCategory.Error,"_0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_5075","'{0}' is assignable to the constraint of type '{1}', but '{1}' could be instantiated with a different subtype of constraint '{2}'."),_0_and_1_operations_cannot_be_mixed_without_parentheses:n(5076,e.DiagnosticCategory.Error,"_0_and_1_operations_cannot_be_mixed_without_parentheses_5076","'{0}' and '{1}' operations cannot be mixed without parentheses."),Unknown_build_option_0_Did_you_mean_1:n(5077,e.DiagnosticCategory.Error,"Unknown_build_option_0_Did_you_mean_1_5077","Unknown build option '{0}'. Did you mean '{1}'?"),Unknown_watch_option_0:n(5078,e.DiagnosticCategory.Error,"Unknown_watch_option_0_5078","Unknown watch option '{0}'."),Unknown_watch_option_0_Did_you_mean_1:n(5079,e.DiagnosticCategory.Error,"Unknown_watch_option_0_Did_you_mean_1_5079","Unknown watch option '{0}'. Did you mean '{1}'?"),Watch_option_0_requires_a_value_of_type_1:n(5080,e.DiagnosticCategory.Error,"Watch_option_0_requires_a_value_of_type_1_5080","Watch option '{0}' requires a value of type {1}."),Cannot_find_a_tsconfig_json_file_at_the_current_directory_Colon_0:n(5081,e.DiagnosticCategory.Error,"Cannot_find_a_tsconfig_json_file_at_the_current_directory_Colon_0_5081","Cannot find a tsconfig.json file at the current directory: {0}."),_0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1:n(5082,e.DiagnosticCategory.Error,"_0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1_5082","'{0}' could be instantiated with an arbitrary type which could be unrelated to '{1}'."),Cannot_read_file_0:n(5083,e.DiagnosticCategory.Error,"Cannot_read_file_0_5083","Cannot read file '{0}'."),Tuple_members_must_all_have_names_or_all_not_have_names:n(5084,e.DiagnosticCategory.Error,"Tuple_members_must_all_have_names_or_all_not_have_names_5084","Tuple members must all have names or all not have names."),A_tuple_member_cannot_be_both_optional_and_rest:n(5085,e.DiagnosticCategory.Error,"A_tuple_member_cannot_be_both_optional_and_rest_5085","A tuple member cannot be both optional and rest."),A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type:n(5086,e.DiagnosticCategory.Error,"A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_c_5086","A labeled tuple element is declared as optional with a question mark after the name and before the colon, rather than after the type."),A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type:n(5087,e.DiagnosticCategory.Error,"A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type_5087","A labeled tuple element is declared as rest with a `...` before the name, rather than before the type."),The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialized_A_type_annotation_is_necessary:n(5088,e.DiagnosticCategory.Error,"The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialize_5088","The inferred type of '{0}' references a type with a cyclic structure which cannot be trivially serialized. A type annotation is necessary."),Option_0_cannot_be_specified_when_option_jsx_is_1:n(5089,e.DiagnosticCategory.Error,"Option_0_cannot_be_specified_when_option_jsx_is_1_5089","Option '{0}' cannot be specified when option 'jsx' is '{1}'."),Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash:n(5090,e.DiagnosticCategory.Error,"Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash_5090","Non-relative paths are not allowed when 'baseUrl' is not set. Did you forget a leading './'?"),Generates_a_sourcemap_for_each_corresponding_d_ts_file:n(6e3,e.DiagnosticCategory.Message,"Generates_a_sourcemap_for_each_corresponding_d_ts_file_6000","Generates a sourcemap for each corresponding '.d.ts' file."),Concatenate_and_emit_output_to_single_file:n(6001,e.DiagnosticCategory.Message,"Concatenate_and_emit_output_to_single_file_6001","Concatenate and emit output to single file."),Generates_corresponding_d_ts_file:n(6002,e.DiagnosticCategory.Message,"Generates_corresponding_d_ts_file_6002","Generates corresponding '.d.ts' file."),Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations:n(6003,e.DiagnosticCategory.Message,"Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations_6003","Specify the location where debugger should locate map files instead of generated locations."),Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations:n(6004,e.DiagnosticCategory.Message,"Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations_6004","Specify the location where debugger should locate TypeScript files instead of source locations."),Watch_input_files:n(6005,e.DiagnosticCategory.Message,"Watch_input_files_6005","Watch input files."),Redirect_output_structure_to_the_directory:n(6006,e.DiagnosticCategory.Message,"Redirect_output_structure_to_the_directory_6006","Redirect output structure to the directory."),Do_not_erase_const_enum_declarations_in_generated_code:n(6007,e.DiagnosticCategory.Message,"Do_not_erase_const_enum_declarations_in_generated_code_6007","Do not erase const enum declarations in generated code."),Do_not_emit_outputs_if_any_errors_were_reported:n(6008,e.DiagnosticCategory.Message,"Do_not_emit_outputs_if_any_errors_were_reported_6008","Do not emit outputs if any errors were reported."),Do_not_emit_comments_to_output:n(6009,e.DiagnosticCategory.Message,"Do_not_emit_comments_to_output_6009","Do not emit comments to output."),Do_not_emit_outputs:n(6010,e.DiagnosticCategory.Message,"Do_not_emit_outputs_6010","Do not emit outputs."),Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typechecking:n(6011,e.DiagnosticCategory.Message,"Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typech_6011","Allow default imports from modules with no default export. This does not affect code emit, just typechecking."),Skip_type_checking_of_declaration_files:n(6012,e.DiagnosticCategory.Message,"Skip_type_checking_of_declaration_files_6012","Skip type checking of declaration files."),Do_not_resolve_the_real_path_of_symlinks:n(6013,e.DiagnosticCategory.Message,"Do_not_resolve_the_real_path_of_symlinks_6013","Do not resolve the real path of symlinks."),Only_emit_d_ts_declaration_files:n(6014,e.DiagnosticCategory.Message,"Only_emit_d_ts_declaration_files_6014","Only emit '.d.ts' declaration files."),Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_ES2019_ES2020_or_ESNEXT:n(6015,e.DiagnosticCategory.Message,"Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_ES2019_ES2020_or_6015","Specify ECMAScript target version: 'ES3' (default), 'ES5', 'ES2015', 'ES2016', 'ES2017', 'ES2018', 'ES2019', 'ES2020', or 'ESNEXT'."),Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_es2020_or_ESNext:n(6016,e.DiagnosticCategory.Message,"Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_es2020_or_ESNext_6016","Specify module code generation: 'none', 'commonjs', 'amd', 'system', 'umd', 'es2015', 'es2020', or 'ESNext'."),Print_this_message:n(6017,e.DiagnosticCategory.Message,"Print_this_message_6017","Print this message."),Print_the_compiler_s_version:n(6019,e.DiagnosticCategory.Message,"Print_the_compiler_s_version_6019","Print the compiler's version."),Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json:n(6020,e.DiagnosticCategory.Message,"Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json_6020","Compile the project given the path to its configuration file, or to a folder with a 'tsconfig.json'."),Syntax_Colon_0:n(6023,e.DiagnosticCategory.Message,"Syntax_Colon_0_6023","Syntax: {0}"),options:n(6024,e.DiagnosticCategory.Message,"options_6024","options"),file:n(6025,e.DiagnosticCategory.Message,"file_6025","file"),Examples_Colon_0:n(6026,e.DiagnosticCategory.Message,"Examples_Colon_0_6026","Examples: {0}"),Options_Colon:n(6027,e.DiagnosticCategory.Message,"Options_Colon_6027","Options:"),Version_0:n(6029,e.DiagnosticCategory.Message,"Version_0_6029","Version {0}"),Insert_command_line_options_and_files_from_a_file:n(6030,e.DiagnosticCategory.Message,"Insert_command_line_options_and_files_from_a_file_6030","Insert command line options and files from a file."),Starting_compilation_in_watch_mode:n(6031,e.DiagnosticCategory.Message,"Starting_compilation_in_watch_mode_6031","Starting compilation in watch mode..."),File_change_detected_Starting_incremental_compilation:n(6032,e.DiagnosticCategory.Message,"File_change_detected_Starting_incremental_compilation_6032","File change detected. Starting incremental compilation..."),KIND:n(6034,e.DiagnosticCategory.Message,"KIND_6034","KIND"),FILE:n(6035,e.DiagnosticCategory.Message,"FILE_6035","FILE"),VERSION:n(6036,e.DiagnosticCategory.Message,"VERSION_6036","VERSION"),LOCATION:n(6037,e.DiagnosticCategory.Message,"LOCATION_6037","LOCATION"),DIRECTORY:n(6038,e.DiagnosticCategory.Message,"DIRECTORY_6038","DIRECTORY"),STRATEGY:n(6039,e.DiagnosticCategory.Message,"STRATEGY_6039","STRATEGY"),FILE_OR_DIRECTORY:n(6040,e.DiagnosticCategory.Message,"FILE_OR_DIRECTORY_6040","FILE OR DIRECTORY"),Generates_corresponding_map_file:n(6043,e.DiagnosticCategory.Message,"Generates_corresponding_map_file_6043","Generates corresponding '.map' file."),Compiler_option_0_expects_an_argument:n(6044,e.DiagnosticCategory.Error,"Compiler_option_0_expects_an_argument_6044","Compiler option '{0}' expects an argument."),Unterminated_quoted_string_in_response_file_0:n(6045,e.DiagnosticCategory.Error,"Unterminated_quoted_string_in_response_file_0_6045","Unterminated quoted string in response file '{0}'."),Argument_for_0_option_must_be_Colon_1:n(6046,e.DiagnosticCategory.Error,"Argument_for_0_option_must_be_Colon_1_6046","Argument for '{0}' option must be: {1}."),Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1:n(6048,e.DiagnosticCategory.Error,"Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1_6048","Locale must be of the form <language> or <language>-<territory>. For example '{0}' or '{1}'."),Unsupported_locale_0:n(6049,e.DiagnosticCategory.Error,"Unsupported_locale_0_6049","Unsupported locale '{0}'."),Unable_to_open_file_0:n(6050,e.DiagnosticCategory.Error,"Unable_to_open_file_0_6050","Unable to open file '{0}'."),Corrupted_locale_file_0:n(6051,e.DiagnosticCategory.Error,"Corrupted_locale_file_0_6051","Corrupted locale file {0}."),Raise_error_on_expressions_and_declarations_with_an_implied_any_type:n(6052,e.DiagnosticCategory.Message,"Raise_error_on_expressions_and_declarations_with_an_implied_any_type_6052","Raise error on expressions and declarations with an implied 'any' type."),File_0_not_found:n(6053,e.DiagnosticCategory.Error,"File_0_not_found_6053","File '{0}' not found."),File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1:n(6054,e.DiagnosticCategory.Error,"File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1_6054","File '{0}' has an unsupported extension. The only supported extensions are {1}."),Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures:n(6055,e.DiagnosticCategory.Message,"Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures_6055","Suppress noImplicitAny errors for indexing objects lacking index signatures."),Do_not_emit_declarations_for_code_that_has_an_internal_annotation:n(6056,e.DiagnosticCategory.Message,"Do_not_emit_declarations_for_code_that_has_an_internal_annotation_6056","Do not emit declarations for code that has an '@internal' annotation."),Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir:n(6058,e.DiagnosticCategory.Message,"Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir_6058","Specify the root directory of input files. Use to control the output directory structure with --outDir."),File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files:n(6059,e.DiagnosticCategory.Error,"File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files_6059","File '{0}' is not under 'rootDir' '{1}'. 'rootDir' is expected to contain all source files."),Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix:n(6060,e.DiagnosticCategory.Message,"Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix_6060","Specify the end of line sequence to be used when emitting files: 'CRLF' (dos) or 'LF' (unix)."),NEWLINE:n(6061,e.DiagnosticCategory.Message,"NEWLINE_6061","NEWLINE"),Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line:n(6064,e.DiagnosticCategory.Error,"Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line_6064","Option '{0}' can only be specified in 'tsconfig.json' file or set to 'null' on command line."),Enables_experimental_support_for_ES7_decorators:n(6065,e.DiagnosticCategory.Message,"Enables_experimental_support_for_ES7_decorators_6065","Enables experimental support for ES7 decorators."),Enables_experimental_support_for_emitting_type_metadata_for_decorators:n(6066,e.DiagnosticCategory.Message,"Enables_experimental_support_for_emitting_type_metadata_for_decorators_6066","Enables experimental support for emitting type metadata for decorators."),Enables_experimental_support_for_ES7_async_functions:n(6068,e.DiagnosticCategory.Message,"Enables_experimental_support_for_ES7_async_functions_6068","Enables experimental support for ES7 async functions."),Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6:n(6069,e.DiagnosticCategory.Message,"Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6_6069","Specify module resolution strategy: 'node' (Node.js) or 'classic' (TypeScript pre-1.6)."),Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file:n(6070,e.DiagnosticCategory.Message,"Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file_6070","Initializes a TypeScript project and creates a tsconfig.json file."),Successfully_created_a_tsconfig_json_file:n(6071,e.DiagnosticCategory.Message,"Successfully_created_a_tsconfig_json_file_6071","Successfully created a tsconfig.json file."),Suppress_excess_property_checks_for_object_literals:n(6072,e.DiagnosticCategory.Message,"Suppress_excess_property_checks_for_object_literals_6072","Suppress excess property checks for object literals."),Stylize_errors_and_messages_using_color_and_context_experimental:n(6073,e.DiagnosticCategory.Message,"Stylize_errors_and_messages_using_color_and_context_experimental_6073","Stylize errors and messages using color and context (experimental)."),Do_not_report_errors_on_unused_labels:n(6074,e.DiagnosticCategory.Message,"Do_not_report_errors_on_unused_labels_6074","Do not report errors on unused labels."),Report_error_when_not_all_code_paths_in_function_return_a_value:n(6075,e.DiagnosticCategory.Message,"Report_error_when_not_all_code_paths_in_function_return_a_value_6075","Report error when not all code paths in function return a value."),Report_errors_for_fallthrough_cases_in_switch_statement:n(6076,e.DiagnosticCategory.Message,"Report_errors_for_fallthrough_cases_in_switch_statement_6076","Report errors for fallthrough cases in switch statement."),Do_not_report_errors_on_unreachable_code:n(6077,e.DiagnosticCategory.Message,"Do_not_report_errors_on_unreachable_code_6077","Do not report errors on unreachable code."),Disallow_inconsistently_cased_references_to_the_same_file:n(6078,e.DiagnosticCategory.Message,"Disallow_inconsistently_cased_references_to_the_same_file_6078","Disallow inconsistently-cased references to the same file."),Specify_library_files_to_be_included_in_the_compilation:n(6079,e.DiagnosticCategory.Message,"Specify_library_files_to_be_included_in_the_compilation_6079","Specify library files to be included in the compilation."),Specify_JSX_code_generation_Colon_preserve_react_native_or_react:n(6080,e.DiagnosticCategory.Message,"Specify_JSX_code_generation_Colon_preserve_react_native_or_react_6080","Specify JSX code generation: 'preserve', 'react-native', or 'react'."),File_0_has_an_unsupported_extension_so_skipping_it:n(6081,e.DiagnosticCategory.Message,"File_0_has_an_unsupported_extension_so_skipping_it_6081","File '{0}' has an unsupported extension, so skipping it."),Only_amd_and_system_modules_are_supported_alongside_0:n(6082,e.DiagnosticCategory.Error,"Only_amd_and_system_modules_are_supported_alongside_0_6082","Only 'amd' and 'system' modules are supported alongside --{0}."),Base_directory_to_resolve_non_absolute_module_names:n(6083,e.DiagnosticCategory.Message,"Base_directory_to_resolve_non_absolute_module_names_6083","Base directory to resolve non-absolute module names."),Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react_JSX_emit:n(6084,e.DiagnosticCategory.Message,"Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react__6084","[Deprecated] Use '--jsxFactory' instead. Specify the object invoked for createElement when targeting 'react' JSX emit"),Enable_tracing_of_the_name_resolution_process:n(6085,e.DiagnosticCategory.Message,"Enable_tracing_of_the_name_resolution_process_6085","Enable tracing of the name resolution process."),Resolving_module_0_from_1:n(6086,e.DiagnosticCategory.Message,"Resolving_module_0_from_1_6086","======== Resolving module '{0}' from '{1}'. ========"),Explicitly_specified_module_resolution_kind_Colon_0:n(6087,e.DiagnosticCategory.Message,"Explicitly_specified_module_resolution_kind_Colon_0_6087","Explicitly specified module resolution kind: '{0}'."),Module_resolution_kind_is_not_specified_using_0:n(6088,e.DiagnosticCategory.Message,"Module_resolution_kind_is_not_specified_using_0_6088","Module resolution kind is not specified, using '{0}'."),Module_name_0_was_successfully_resolved_to_1:n(6089,e.DiagnosticCategory.Message,"Module_name_0_was_successfully_resolved_to_1_6089","======== Module name '{0}' was successfully resolved to '{1}'. ========"),Module_name_0_was_not_resolved:n(6090,e.DiagnosticCategory.Message,"Module_name_0_was_not_resolved_6090","======== Module name '{0}' was not resolved. ========"),paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0:n(6091,e.DiagnosticCategory.Message,"paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0_6091","'paths' option is specified, looking for a pattern to match module name '{0}'."),Module_name_0_matched_pattern_1:n(6092,e.DiagnosticCategory.Message,"Module_name_0_matched_pattern_1_6092","Module name '{0}', matched pattern '{1}'."),Trying_substitution_0_candidate_module_location_Colon_1:n(6093,e.DiagnosticCategory.Message,"Trying_substitution_0_candidate_module_location_Colon_1_6093","Trying substitution '{0}', candidate module location: '{1}'."),Resolving_module_name_0_relative_to_base_url_1_2:n(6094,e.DiagnosticCategory.Message,"Resolving_module_name_0_relative_to_base_url_1_2_6094","Resolving module name '{0}' relative to base url '{1}' - '{2}'."),Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1:n(6095,e.DiagnosticCategory.Message,"Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1_6095","Loading module as file / folder, candidate module location '{0}', target file type '{1}'."),File_0_does_not_exist:n(6096,e.DiagnosticCategory.Message,"File_0_does_not_exist_6096","File '{0}' does not exist."),File_0_exist_use_it_as_a_name_resolution_result:n(6097,e.DiagnosticCategory.Message,"File_0_exist_use_it_as_a_name_resolution_result_6097","File '{0}' exist - use it as a name resolution result."),Loading_module_0_from_node_modules_folder_target_file_type_1:n(6098,e.DiagnosticCategory.Message,"Loading_module_0_from_node_modules_folder_target_file_type_1_6098","Loading module '{0}' from 'node_modules' folder, target file type '{1}'."),Found_package_json_at_0:n(6099,e.DiagnosticCategory.Message,"Found_package_json_at_0_6099","Found 'package.json' at '{0}'."),package_json_does_not_have_a_0_field:n(6100,e.DiagnosticCategory.Message,"package_json_does_not_have_a_0_field_6100","'package.json' does not have a '{0}' field."),package_json_has_0_field_1_that_references_2:n(6101,e.DiagnosticCategory.Message,"package_json_has_0_field_1_that_references_2_6101","'package.json' has '{0}' field '{1}' that references '{2}'."),Allow_javascript_files_to_be_compiled:n(6102,e.DiagnosticCategory.Message,"Allow_javascript_files_to_be_compiled_6102","Allow javascript files to be compiled."),Option_0_should_have_array_of_strings_as_a_value:n(6103,e.DiagnosticCategory.Error,"Option_0_should_have_array_of_strings_as_a_value_6103","Option '{0}' should have array of strings as a value."),Checking_if_0_is_the_longest_matching_prefix_for_1_2:n(6104,e.DiagnosticCategory.Message,"Checking_if_0_is_the_longest_matching_prefix_for_1_2_6104","Checking if '{0}' is the longest matching prefix for '{1}' - '{2}'."),Expected_type_of_0_field_in_package_json_to_be_1_got_2:n(6105,e.DiagnosticCategory.Message,"Expected_type_of_0_field_in_package_json_to_be_1_got_2_6105","Expected type of '{0}' field in 'package.json' to be '{1}', got '{2}'."),baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1:n(6106,e.DiagnosticCategory.Message,"baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1_6106","'baseUrl' option is set to '{0}', using this value to resolve non-relative module name '{1}'."),rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0:n(6107,e.DiagnosticCategory.Message,"rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0_6107","'rootDirs' option is set, using it to resolve relative module name '{0}'."),Longest_matching_prefix_for_0_is_1:n(6108,e.DiagnosticCategory.Message,"Longest_matching_prefix_for_0_is_1_6108","Longest matching prefix for '{0}' is '{1}'."),Loading_0_from_the_root_dir_1_candidate_location_2:n(6109,e.DiagnosticCategory.Message,"Loading_0_from_the_root_dir_1_candidate_location_2_6109","Loading '{0}' from the root dir '{1}', candidate location '{2}'."),Trying_other_entries_in_rootDirs:n(6110,e.DiagnosticCategory.Message,"Trying_other_entries_in_rootDirs_6110","Trying other entries in 'rootDirs'."),Module_resolution_using_rootDirs_has_failed:n(6111,e.DiagnosticCategory.Message,"Module_resolution_using_rootDirs_has_failed_6111","Module resolution using 'rootDirs' has failed."),Do_not_emit_use_strict_directives_in_module_output:n(6112,e.DiagnosticCategory.Message,"Do_not_emit_use_strict_directives_in_module_output_6112","Do not emit 'use strict' directives in module output."),Enable_strict_null_checks:n(6113,e.DiagnosticCategory.Message,"Enable_strict_null_checks_6113","Enable strict null checks."),Unknown_option_excludes_Did_you_mean_exclude:n(6114,e.DiagnosticCategory.Error,"Unknown_option_excludes_Did_you_mean_exclude_6114","Unknown option 'excludes'. Did you mean 'exclude'?"),Raise_error_on_this_expressions_with_an_implied_any_type:n(6115,e.DiagnosticCategory.Message,"Raise_error_on_this_expressions_with_an_implied_any_type_6115","Raise error on 'this' expressions with an implied 'any' type."),Resolving_type_reference_directive_0_containing_file_1_root_directory_2:n(6116,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_1_root_directory_2_6116","======== Resolving type reference directive '{0}', containing file '{1}', root directory '{2}'. ========"),Resolving_using_primary_search_paths:n(6117,e.DiagnosticCategory.Message,"Resolving_using_primary_search_paths_6117","Resolving using primary search paths..."),Resolving_from_node_modules_folder:n(6118,e.DiagnosticCategory.Message,"Resolving_from_node_modules_folder_6118","Resolving from node_modules folder..."),Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2:n(6119,e.DiagnosticCategory.Message,"Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2_6119","======== Type reference directive '{0}' was successfully resolved to '{1}', primary: {2}. ========"),Type_reference_directive_0_was_not_resolved:n(6120,e.DiagnosticCategory.Message,"Type_reference_directive_0_was_not_resolved_6120","======== Type reference directive '{0}' was not resolved. ========"),Resolving_with_primary_search_path_0:n(6121,e.DiagnosticCategory.Message,"Resolving_with_primary_search_path_0_6121","Resolving with primary search path '{0}'."),Root_directory_cannot_be_determined_skipping_primary_search_paths:n(6122,e.DiagnosticCategory.Message,"Root_directory_cannot_be_determined_skipping_primary_search_paths_6122","Root directory cannot be determined, skipping primary search paths."),Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set:n(6123,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set_6123","======== Resolving type reference directive '{0}', containing file '{1}', root directory not set. ========"),Type_declaration_files_to_be_included_in_compilation:n(6124,e.DiagnosticCategory.Message,"Type_declaration_files_to_be_included_in_compilation_6124","Type declaration files to be included in compilation."),Looking_up_in_node_modules_folder_initial_location_0:n(6125,e.DiagnosticCategory.Message,"Looking_up_in_node_modules_folder_initial_location_0_6125","Looking up in 'node_modules' folder, initial location '{0}'."),Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder:n(6126,e.DiagnosticCategory.Message,"Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_mod_6126","Containing file is not specified and root directory cannot be determined, skipping lookup in 'node_modules' folder."),Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1:n(6127,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1_6127","======== Resolving type reference directive '{0}', containing file not set, root directory '{1}'. ========"),Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set:n(6128,e.DiagnosticCategory.Message,"Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set_6128","======== Resolving type reference directive '{0}', containing file not set, root directory not set. ========"),Resolving_real_path_for_0_result_1:n(6130,e.DiagnosticCategory.Message,"Resolving_real_path_for_0_result_1_6130","Resolving real path for '{0}', result '{1}'."),Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system:n(6131,e.DiagnosticCategory.Error,"Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system_6131","Cannot compile modules using option '{0}' unless the '--module' flag is 'amd' or 'system'."),File_name_0_has_a_1_extension_stripping_it:n(6132,e.DiagnosticCategory.Message,"File_name_0_has_a_1_extension_stripping_it_6132","File name '{0}' has a '{1}' extension - stripping it."),_0_is_declared_but_its_value_is_never_read:n(6133,e.DiagnosticCategory.Error,"_0_is_declared_but_its_value_is_never_read_6133","'{0}' is declared but its value is never read.",!0),Report_errors_on_unused_locals:n(6134,e.DiagnosticCategory.Message,"Report_errors_on_unused_locals_6134","Report errors on unused locals."),Report_errors_on_unused_parameters:n(6135,e.DiagnosticCategory.Message,"Report_errors_on_unused_parameters_6135","Report errors on unused parameters."),The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files:n(6136,e.DiagnosticCategory.Message,"The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files_6136","The maximum dependency depth to search under node_modules and load JavaScript files."),Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1:n(6137,e.DiagnosticCategory.Error,"Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1_6137","Cannot import type declaration files. Consider importing '{0}' instead of '{1}'."),Property_0_is_declared_but_its_value_is_never_read:n(6138,e.DiagnosticCategory.Error,"Property_0_is_declared_but_its_value_is_never_read_6138","Property '{0}' is declared but its value is never read.",!0),Import_emit_helpers_from_tslib:n(6139,e.DiagnosticCategory.Message,"Import_emit_helpers_from_tslib_6139","Import emit helpers from 'tslib'."),Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2:n(6140,e.DiagnosticCategory.Error,"Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using__6140","Auto discovery for typings is enabled in project '{0}'. Running extra resolution pass for module '{1}' using cache location '{2}'."),Parse_in_strict_mode_and_emit_use_strict_for_each_source_file:n(6141,e.DiagnosticCategory.Message,"Parse_in_strict_mode_and_emit_use_strict_for_each_source_file_6141",'Parse in strict mode and emit "use strict" for each source file.'),Module_0_was_resolved_to_1_but_jsx_is_not_set:n(6142,e.DiagnosticCategory.Error,"Module_0_was_resolved_to_1_but_jsx_is_not_set_6142","Module '{0}' was resolved to '{1}', but '--jsx' is not set."),Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1:n(6144,e.DiagnosticCategory.Message,"Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1_6144","Module '{0}' was resolved as locally declared ambient module in file '{1}'."),Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified:n(6145,e.DiagnosticCategory.Message,"Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified_6145","Module '{0}' was resolved as ambient module declared in '{1}' since this file was not modified."),Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h:n(6146,e.DiagnosticCategory.Message,"Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h_6146","Specify the JSX factory function to use when targeting 'react' JSX emit, e.g. 'React.createElement' or 'h'."),Resolution_for_module_0_was_found_in_cache_from_location_1:n(6147,e.DiagnosticCategory.Message,"Resolution_for_module_0_was_found_in_cache_from_location_1_6147","Resolution for module '{0}' was found in cache from location '{1}'."),Directory_0_does_not_exist_skipping_all_lookups_in_it:n(6148,e.DiagnosticCategory.Message,"Directory_0_does_not_exist_skipping_all_lookups_in_it_6148","Directory '{0}' does not exist, skipping all lookups in it."),Show_diagnostic_information:n(6149,e.DiagnosticCategory.Message,"Show_diagnostic_information_6149","Show diagnostic information."),Show_verbose_diagnostic_information:n(6150,e.DiagnosticCategory.Message,"Show_verbose_diagnostic_information_6150","Show verbose diagnostic information."),Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file:n(6151,e.DiagnosticCategory.Message,"Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file_6151","Emit a single file with source maps instead of having a separate file."),Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap_to_be_set:n(6152,e.DiagnosticCategory.Message,"Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap__6152","Emit the source alongside the sourcemaps within a single file; requires '--inlineSourceMap' or '--sourceMap' to be set."),Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule:n(6153,e.DiagnosticCategory.Message,"Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule_6153","Transpile each file as a separate module (similar to 'ts.transpileModule')."),Print_names_of_generated_files_part_of_the_compilation:n(6154,e.DiagnosticCategory.Message,"Print_names_of_generated_files_part_of_the_compilation_6154","Print names of generated files part of the compilation."),Print_names_of_files_part_of_the_compilation:n(6155,e.DiagnosticCategory.Message,"Print_names_of_files_part_of_the_compilation_6155","Print names of files part of the compilation."),The_locale_used_when_displaying_messages_to_the_user_e_g_en_us:n(6156,e.DiagnosticCategory.Message,"The_locale_used_when_displaying_messages_to_the_user_e_g_en_us_6156","The locale used when displaying messages to the user (e.g. 'en-us')"),Do_not_generate_custom_helper_functions_like_extends_in_compiled_output:n(6157,e.DiagnosticCategory.Message,"Do_not_generate_custom_helper_functions_like_extends_in_compiled_output_6157","Do not generate custom helper functions like '__extends' in compiled output."),Do_not_include_the_default_library_file_lib_d_ts:n(6158,e.DiagnosticCategory.Message,"Do_not_include_the_default_library_file_lib_d_ts_6158","Do not include the default library file (lib.d.ts)."),Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files:n(6159,e.DiagnosticCategory.Message,"Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files_6159","Do not add triple-slash references or imported modules to the list of compiled files."),Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files:n(6160,e.DiagnosticCategory.Message,"Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files_6160","[Deprecated] Use '--skipLibCheck' instead. Skip type checking of default library declaration files."),List_of_folders_to_include_type_definitions_from:n(6161,e.DiagnosticCategory.Message,"List_of_folders_to_include_type_definitions_from_6161","List of folders to include type definitions from."),Disable_size_limitations_on_JavaScript_projects:n(6162,e.DiagnosticCategory.Message,"Disable_size_limitations_on_JavaScript_projects_6162","Disable size limitations on JavaScript projects."),The_character_set_of_the_input_files:n(6163,e.DiagnosticCategory.Message,"The_character_set_of_the_input_files_6163","The character set of the input files."),Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files:n(6164,e.DiagnosticCategory.Message,"Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files_6164","Emit a UTF-8 Byte Order Mark (BOM) in the beginning of output files."),Do_not_truncate_error_messages:n(6165,e.DiagnosticCategory.Message,"Do_not_truncate_error_messages_6165","Do not truncate error messages."),Output_directory_for_generated_declaration_files:n(6166,e.DiagnosticCategory.Message,"Output_directory_for_generated_declaration_files_6166","Output directory for generated declaration files."),A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl:n(6167,e.DiagnosticCategory.Message,"A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl_6167","A series of entries which re-map imports to lookup locations relative to the 'baseUrl'."),List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime:n(6168,e.DiagnosticCategory.Message,"List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime_6168","List of root folders whose combined content represents the structure of the project at runtime."),Show_all_compiler_options:n(6169,e.DiagnosticCategory.Message,"Show_all_compiler_options_6169","Show all compiler options."),Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file:n(6170,e.DiagnosticCategory.Message,"Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file_6170","[Deprecated] Use '--outFile' instead. Concatenate and emit output to single file"),Command_line_Options:n(6171,e.DiagnosticCategory.Message,"Command_line_Options_6171","Command-line Options"),Basic_Options:n(6172,e.DiagnosticCategory.Message,"Basic_Options_6172","Basic Options"),Strict_Type_Checking_Options:n(6173,e.DiagnosticCategory.Message,"Strict_Type_Checking_Options_6173","Strict Type-Checking Options"),Module_Resolution_Options:n(6174,e.DiagnosticCategory.Message,"Module_Resolution_Options_6174","Module Resolution Options"),Source_Map_Options:n(6175,e.DiagnosticCategory.Message,"Source_Map_Options_6175","Source Map Options"),Additional_Checks:n(6176,e.DiagnosticCategory.Message,"Additional_Checks_6176","Additional Checks"),Experimental_Options:n(6177,e.DiagnosticCategory.Message,"Experimental_Options_6177","Experimental Options"),Advanced_Options:n(6178,e.DiagnosticCategory.Message,"Advanced_Options_6178","Advanced Options"),Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3:n(6179,e.DiagnosticCategory.Message,"Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3_6179","Provide full support for iterables in 'for-of', spread, and destructuring when targeting 'ES5' or 'ES3'."),Enable_all_strict_type_checking_options:n(6180,e.DiagnosticCategory.Message,"Enable_all_strict_type_checking_options_6180","Enable all strict type-checking options."),List_of_language_service_plugins:n(6181,e.DiagnosticCategory.Message,"List_of_language_service_plugins_6181","List of language service plugins."),Scoped_package_detected_looking_in_0:n(6182,e.DiagnosticCategory.Message,"Scoped_package_detected_looking_in_0_6182","Scoped package detected, looking in '{0}'"),Reusing_resolution_of_module_0_to_file_1_from_old_program:n(6183,e.DiagnosticCategory.Message,"Reusing_resolution_of_module_0_to_file_1_from_old_program_6183","Reusing resolution of module '{0}' to file '{1}' from old program."),Reusing_module_resolutions_originating_in_0_since_resolutions_are_unchanged_from_old_program:n(6184,e.DiagnosticCategory.Message,"Reusing_module_resolutions_originating_in_0_since_resolutions_are_unchanged_from_old_program_6184","Reusing module resolutions originating in '{0}' since resolutions are unchanged from old program."),Disable_strict_checking_of_generic_signatures_in_function_types:n(6185,e.DiagnosticCategory.Message,"Disable_strict_checking_of_generic_signatures_in_function_types_6185","Disable strict checking of generic signatures in function types."),Enable_strict_checking_of_function_types:n(6186,e.DiagnosticCategory.Message,"Enable_strict_checking_of_function_types_6186","Enable strict checking of function types."),Enable_strict_checking_of_property_initialization_in_classes:n(6187,e.DiagnosticCategory.Message,"Enable_strict_checking_of_property_initialization_in_classes_6187","Enable strict checking of property initialization in classes."),Numeric_separators_are_not_allowed_here:n(6188,e.DiagnosticCategory.Error,"Numeric_separators_are_not_allowed_here_6188","Numeric separators are not allowed here."),Multiple_consecutive_numeric_separators_are_not_permitted:n(6189,e.DiagnosticCategory.Error,"Multiple_consecutive_numeric_separators_are_not_permitted_6189","Multiple consecutive numeric separators are not permitted."),Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen:n(6191,e.DiagnosticCategory.Message,"Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen_6191","Whether to keep outdated console output in watch mode instead of clearing the screen."),All_imports_in_import_declaration_are_unused:n(6192,e.DiagnosticCategory.Error,"All_imports_in_import_declaration_are_unused_6192","All imports in import declaration are unused.",!0),Found_1_error_Watching_for_file_changes:n(6193,e.DiagnosticCategory.Message,"Found_1_error_Watching_for_file_changes_6193","Found 1 error. Watching for file changes."),Found_0_errors_Watching_for_file_changes:n(6194,e.DiagnosticCategory.Message,"Found_0_errors_Watching_for_file_changes_6194","Found {0} errors. Watching for file changes."),Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols:n(6195,e.DiagnosticCategory.Message,"Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols_6195","Resolve 'keyof' to string valued property names only (no numbers or symbols)."),_0_is_declared_but_never_used:n(6196,e.DiagnosticCategory.Error,"_0_is_declared_but_never_used_6196","'{0}' is declared but never used.",!0),Include_modules_imported_with_json_extension:n(6197,e.DiagnosticCategory.Message,"Include_modules_imported_with_json_extension_6197","Include modules imported with '.json' extension"),All_destructured_elements_are_unused:n(6198,e.DiagnosticCategory.Error,"All_destructured_elements_are_unused_6198","All destructured elements are unused.",!0),All_variables_are_unused:n(6199,e.DiagnosticCategory.Error,"All_variables_are_unused_6199","All variables are unused.",!0),Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0:n(6200,e.DiagnosticCategory.Error,"Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0_6200","Definitions of the following identifiers conflict with those in another file: {0}"),Conflicts_are_in_this_file:n(6201,e.DiagnosticCategory.Message,"Conflicts_are_in_this_file_6201","Conflicts are in this file."),Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0:n(6202,e.DiagnosticCategory.Error,"Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0_6202","Project references may not form a circular graph. Cycle detected: {0}"),_0_was_also_declared_here:n(6203,e.DiagnosticCategory.Message,"_0_was_also_declared_here_6203","'{0}' was also declared here."),and_here:n(6204,e.DiagnosticCategory.Message,"and_here_6204","and here."),All_type_parameters_are_unused:n(6205,e.DiagnosticCategory.Error,"All_type_parameters_are_unused_6205","All type parameters are unused."),package_json_has_a_typesVersions_field_with_version_specific_path_mappings:n(6206,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_field_with_version_specific_path_mappings_6206","'package.json' has a 'typesVersions' field with version-specific path mappings."),package_json_does_not_have_a_typesVersions_entry_that_matches_version_0:n(6207,e.DiagnosticCategory.Message,"package_json_does_not_have_a_typesVersions_entry_that_matches_version_0_6207","'package.json' does not have a 'typesVersions' entry that matches version '{0}'."),package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2:n(6208,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_ma_6208","'package.json' has a 'typesVersions' entry '{0}' that matches compiler version '{1}', looking for a pattern to match module name '{2}'."),package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range:n(6209,e.DiagnosticCategory.Message,"package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range_6209","'package.json' has a 'typesVersions' entry '{0}' that is not a valid semver range."),An_argument_for_0_was_not_provided:n(6210,e.DiagnosticCategory.Message,"An_argument_for_0_was_not_provided_6210","An argument for '{0}' was not provided."),An_argument_matching_this_binding_pattern_was_not_provided:n(6211,e.DiagnosticCategory.Message,"An_argument_matching_this_binding_pattern_was_not_provided_6211","An argument matching this binding pattern was not provided."),Did_you_mean_to_call_this_expression:n(6212,e.DiagnosticCategory.Message,"Did_you_mean_to_call_this_expression_6212","Did you mean to call this expression?"),Did_you_mean_to_use_new_with_this_expression:n(6213,e.DiagnosticCategory.Message,"Did_you_mean_to_use_new_with_this_expression_6213","Did you mean to use 'new' with this expression?"),Enable_strict_bind_call_and_apply_methods_on_functions:n(6214,e.DiagnosticCategory.Message,"Enable_strict_bind_call_and_apply_methods_on_functions_6214","Enable strict 'bind', 'call', and 'apply' methods on functions."),Using_compiler_options_of_project_reference_redirect_0:n(6215,e.DiagnosticCategory.Message,"Using_compiler_options_of_project_reference_redirect_0_6215","Using compiler options of project reference redirect '{0}'."),Found_1_error:n(6216,e.DiagnosticCategory.Message,"Found_1_error_6216","Found 1 error."),Found_0_errors:n(6217,e.DiagnosticCategory.Message,"Found_0_errors_6217","Found {0} errors."),Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2:n(6218,e.DiagnosticCategory.Message,"Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2_6218","======== Module name '{0}' was successfully resolved to '{1}' with Package ID '{2}'. ========"),Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3:n(6219,e.DiagnosticCategory.Message,"Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3_6219","======== Type reference directive '{0}' was successfully resolved to '{1}' with Package ID '{2}', primary: {3}. ========"),package_json_had_a_falsy_0_field:n(6220,e.DiagnosticCategory.Message,"package_json_had_a_falsy_0_field_6220","'package.json' had a falsy '{0}' field."),Disable_use_of_source_files_instead_of_declaration_files_from_referenced_projects:n(6221,e.DiagnosticCategory.Message,"Disable_use_of_source_files_instead_of_declaration_files_from_referenced_projects_6221","Disable use of source files instead of declaration files from referenced projects."),Emit_class_fields_with_Define_instead_of_Set:n(6222,e.DiagnosticCategory.Message,"Emit_class_fields_with_Define_instead_of_Set_6222","Emit class fields with Define instead of Set."),Generates_a_CPU_profile:n(6223,e.DiagnosticCategory.Message,"Generates_a_CPU_profile_6223","Generates a CPU profile."),Disable_solution_searching_for_this_project:n(6224,e.DiagnosticCategory.Message,"Disable_solution_searching_for_this_project_6224","Disable solution searching for this project."),Specify_strategy_for_watching_file_Colon_FixedPollingInterval_default_PriorityPollingInterval_DynamicPriorityPolling_UseFsEvents_UseFsEventsOnParentDirectory:n(6225,e.DiagnosticCategory.Message,"Specify_strategy_for_watching_file_Colon_FixedPollingInterval_default_PriorityPollingInterval_Dynami_6225","Specify strategy for watching file: 'FixedPollingInterval' (default), 'PriorityPollingInterval', 'DynamicPriorityPolling', 'UseFsEvents', 'UseFsEventsOnParentDirectory'."),Specify_strategy_for_watching_directory_on_platforms_that_don_t_support_recursive_watching_natively_Colon_UseFsEvents_default_FixedPollingInterval_DynamicPriorityPolling:n(6226,e.DiagnosticCategory.Message,"Specify_strategy_for_watching_directory_on_platforms_that_don_t_support_recursive_watching_natively__6226","Specify strategy for watching directory on platforms that don't support recursive watching natively: 'UseFsEvents' (default), 'FixedPollingInterval', 'DynamicPriorityPolling'."),Specify_strategy_for_creating_a_polling_watch_when_it_fails_to_create_using_file_system_events_Colon_FixedInterval_default_PriorityInterval_DynamicPriority:n(6227,e.DiagnosticCategory.Message,"Specify_strategy_for_creating_a_polling_watch_when_it_fails_to_create_using_file_system_events_Colon_6227","Specify strategy for creating a polling watch when it fails to create using file system events: 'FixedInterval' (default), 'PriorityInterval', 'DynamicPriority'."),Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_support_recursive_watching_natively:n(6228,e.DiagnosticCategory.Message,"Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_supp_6228","Synchronously call callbacks and update the state of directory watchers on platforms that don't support recursive watching natively."),Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3:n(6229,e.DiagnosticCategory.Error,"Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3_6229","Tag '{0}' expects at least '{1}' arguments, but the JSX factory '{2}' provides at most '{3}'."),Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line:n(6230,e.DiagnosticCategory.Error,"Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line_6230","Option '{0}' can only be specified in 'tsconfig.json' file or set to 'false' or 'null' on command line."),Could_not_resolve_the_path_0_with_the_extensions_Colon_1:n(6231,e.DiagnosticCategory.Error,"Could_not_resolve_the_path_0_with_the_extensions_Colon_1_6231","Could not resolve the path '{0}' with the extensions: {1}."),Declaration_augments_declaration_in_another_file_This_cannot_be_serialized:n(6232,e.DiagnosticCategory.Error,"Declaration_augments_declaration_in_another_file_This_cannot_be_serialized_6232","Declaration augments declaration in another file. This cannot be serialized."),This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_file:n(6233,e.DiagnosticCategory.Error,"This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_fil_6233","This is the declaration being augmented. Consider moving the augmenting declaration into the same file."),This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without:n(6234,e.DiagnosticCategory.Error,"This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without_6234","This expression is not callable because it is a 'get' accessor. Did you mean to use it without '()'?"),Disable_loading_referenced_projects:n(6235,e.DiagnosticCategory.Message,"Disable_loading_referenced_projects_6235","Disable loading referenced projects."),Arguments_for_the_rest_parameter_0_were_not_provided:n(6236,e.DiagnosticCategory.Error,"Arguments_for_the_rest_parameter_0_were_not_provided_6236","Arguments for the rest parameter '{0}' were not provided."),Generates_an_event_trace_and_a_list_of_types:n(6237,e.DiagnosticCategory.Message,"Generates_an_event_trace_and_a_list_of_types_6237","Generates an event trace and a list of types."),Specify_the_module_specifier_to_be_used_to_import_the_jsx_and_jsxs_factory_functions_from_eg_react:n(6238,e.DiagnosticCategory.Error,"Specify_the_module_specifier_to_be_used_to_import_the_jsx_and_jsxs_factory_functions_from_eg_react_6238","Specify the module specifier to be used to import the `jsx` and `jsxs` factory functions from. eg, react"),Projects_to_reference:n(6300,e.DiagnosticCategory.Message,"Projects_to_reference_6300","Projects to reference"),Enable_project_compilation:n(6302,e.DiagnosticCategory.Message,"Enable_project_compilation_6302","Enable project compilation"),Composite_projects_may_not_disable_declaration_emit:n(6304,e.DiagnosticCategory.Error,"Composite_projects_may_not_disable_declaration_emit_6304","Composite projects may not disable declaration emit."),Output_file_0_has_not_been_built_from_source_file_1:n(6305,e.DiagnosticCategory.Error,"Output_file_0_has_not_been_built_from_source_file_1_6305","Output file '{0}' has not been built from source file '{1}'."),Referenced_project_0_must_have_setting_composite_Colon_true:n(6306,e.DiagnosticCategory.Error,"Referenced_project_0_must_have_setting_composite_Colon_true_6306","Referenced project '{0}' must have setting \"composite\": true."),File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_include_pattern:n(6307,e.DiagnosticCategory.Error,"File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_includ_6307","File '{0}' is not listed within the file list of project '{1}'. Projects must list all files or use an 'include' pattern."),Cannot_prepend_project_0_because_it_does_not_have_outFile_set:n(6308,e.DiagnosticCategory.Error,"Cannot_prepend_project_0_because_it_does_not_have_outFile_set_6308","Cannot prepend project '{0}' because it does not have 'outFile' set"),Output_file_0_from_project_1_does_not_exist:n(6309,e.DiagnosticCategory.Error,"Output_file_0_from_project_1_does_not_exist_6309","Output file '{0}' from project '{1}' does not exist"),Referenced_project_0_may_not_disable_emit:n(6310,e.DiagnosticCategory.Error,"Referenced_project_0_may_not_disable_emit_6310","Referenced project '{0}' may not disable emit."),Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2:n(6350,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2_6350","Project '{0}' is out of date because oldest output '{1}' is older than newest input '{2}'"),Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2:n(6351,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2_6351","Project '{0}' is up to date because newest input '{1}' is older than oldest output '{2}'"),Project_0_is_out_of_date_because_output_file_1_does_not_exist:n(6352,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_output_file_1_does_not_exist_6352","Project '{0}' is out of date because output file '{1}' does not exist"),Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date:n(6353,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date_6353","Project '{0}' is out of date because its dependency '{1}' is out of date"),Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies:n(6354,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies_6354","Project '{0}' is up to date with .d.ts files from its dependencies"),Projects_in_this_build_Colon_0:n(6355,e.DiagnosticCategory.Message,"Projects_in_this_build_Colon_0_6355","Projects in this build: {0}"),A_non_dry_build_would_delete_the_following_files_Colon_0:n(6356,e.DiagnosticCategory.Message,"A_non_dry_build_would_delete_the_following_files_Colon_0_6356","A non-dry build would delete the following files: {0}"),A_non_dry_build_would_build_project_0:n(6357,e.DiagnosticCategory.Message,"A_non_dry_build_would_build_project_0_6357","A non-dry build would build project '{0}'"),Building_project_0:n(6358,e.DiagnosticCategory.Message,"Building_project_0_6358","Building project '{0}'..."),Updating_output_timestamps_of_project_0:n(6359,e.DiagnosticCategory.Message,"Updating_output_timestamps_of_project_0_6359","Updating output timestamps of project '{0}'..."),delete_this_Project_0_is_up_to_date_because_it_was_previously_built:n(6360,e.DiagnosticCategory.Message,"delete_this_Project_0_is_up_to_date_because_it_was_previously_built_6360","delete this - Project '{0}' is up to date because it was previously built"),Project_0_is_up_to_date:n(6361,e.DiagnosticCategory.Message,"Project_0_is_up_to_date_6361","Project '{0}' is up to date"),Skipping_build_of_project_0_because_its_dependency_1_has_errors:n(6362,e.DiagnosticCategory.Message,"Skipping_build_of_project_0_because_its_dependency_1_has_errors_6362","Skipping build of project '{0}' because its dependency '{1}' has errors"),Project_0_can_t_be_built_because_its_dependency_1_has_errors:n(6363,e.DiagnosticCategory.Message,"Project_0_can_t_be_built_because_its_dependency_1_has_errors_6363","Project '{0}' can't be built because its dependency '{1}' has errors"),Build_one_or_more_projects_and_their_dependencies_if_out_of_date:n(6364,e.DiagnosticCategory.Message,"Build_one_or_more_projects_and_their_dependencies_if_out_of_date_6364","Build one or more projects and their dependencies, if out of date"),Delete_the_outputs_of_all_projects:n(6365,e.DiagnosticCategory.Message,"Delete_the_outputs_of_all_projects_6365","Delete the outputs of all projects"),Enable_verbose_logging:n(6366,e.DiagnosticCategory.Message,"Enable_verbose_logging_6366","Enable verbose logging"),Show_what_would_be_built_or_deleted_if_specified_with_clean:n(6367,e.DiagnosticCategory.Message,"Show_what_would_be_built_or_deleted_if_specified_with_clean_6367","Show what would be built (or deleted, if specified with '--clean')"),Build_all_projects_including_those_that_appear_to_be_up_to_date:n(6368,e.DiagnosticCategory.Message,"Build_all_projects_including_those_that_appear_to_be_up_to_date_6368","Build all projects, including those that appear to be up to date"),Option_build_must_be_the_first_command_line_argument:n(6369,e.DiagnosticCategory.Error,"Option_build_must_be_the_first_command_line_argument_6369","Option '--build' must be the first command line argument."),Options_0_and_1_cannot_be_combined:n(6370,e.DiagnosticCategory.Error,"Options_0_and_1_cannot_be_combined_6370","Options '{0}' and '{1}' cannot be combined."),Updating_unchanged_output_timestamps_of_project_0:n(6371,e.DiagnosticCategory.Message,"Updating_unchanged_output_timestamps_of_project_0_6371","Updating unchanged output timestamps of project '{0}'..."),Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed:n(6372,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed_6372","Project '{0}' is out of date because output of its dependency '{1}' has changed"),Updating_output_of_project_0:n(6373,e.DiagnosticCategory.Message,"Updating_output_of_project_0_6373","Updating output of project '{0}'..."),A_non_dry_build_would_update_timestamps_for_output_of_project_0:n(6374,e.DiagnosticCategory.Message,"A_non_dry_build_would_update_timestamps_for_output_of_project_0_6374","A non-dry build would update timestamps for output of project '{0}'"),A_non_dry_build_would_update_output_of_project_0:n(6375,e.DiagnosticCategory.Message,"A_non_dry_build_would_update_output_of_project_0_6375","A non-dry build would update output of project '{0}'"),Cannot_update_output_of_project_0_because_there_was_error_reading_file_1:n(6376,e.DiagnosticCategory.Message,"Cannot_update_output_of_project_0_because_there_was_error_reading_file_1_6376","Cannot update output of project '{0}' because there was error reading file '{1}'"),Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1:n(6377,e.DiagnosticCategory.Error,"Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1_6377","Cannot write file '{0}' because it will overwrite '.tsbuildinfo' file generated by referenced project '{1}'"),Enable_incremental_compilation:n(6378,e.DiagnosticCategory.Message,"Enable_incremental_compilation_6378","Enable incremental compilation"),Composite_projects_may_not_disable_incremental_compilation:n(6379,e.DiagnosticCategory.Error,"Composite_projects_may_not_disable_incremental_compilation_6379","Composite projects may not disable incremental compilation."),Specify_file_to_store_incremental_compilation_information:n(6380,e.DiagnosticCategory.Message,"Specify_file_to_store_incremental_compilation_information_6380","Specify file to store incremental compilation information"),Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_current_version_2:n(6381,e.DiagnosticCategory.Message,"Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_curren_6381","Project '{0}' is out of date because output for it was generated with version '{1}' that differs with current version '{2}'"),Skipping_build_of_project_0_because_its_dependency_1_was_not_built:n(6382,e.DiagnosticCategory.Message,"Skipping_build_of_project_0_because_its_dependency_1_was_not_built_6382","Skipping build of project '{0}' because its dependency '{1}' was not built"),Project_0_can_t_be_built_because_its_dependency_1_was_not_built:n(6383,e.DiagnosticCategory.Message,"Project_0_can_t_be_built_because_its_dependency_1_was_not_built_6383","Project '{0}' can't be built because its dependency '{1}' was not built"),Have_recompiles_in_incremental_and_watch_assume_that_changes_within_a_file_will_only_affect_files_directly_depending_on_it:n(6384,e.DiagnosticCategory.Message,"Have_recompiles_in_incremental_and_watch_assume_that_changes_within_a_file_will_only_affect_files_di_6384","Have recompiles in '--incremental' and '--watch' assume that changes within a file will only affect files directly depending on it."),_0_is_deprecated:n(6385,e.DiagnosticCategory.Suggestion,"_0_is_deprecated_6385","'{0}' is deprecated",void 0,void 0,!0),Performance_timings_for_diagnostics_or_extendedDiagnostics_are_not_available_in_this_session_A_native_implementation_of_the_Web_Performance_API_could_not_be_found:n(6386,e.DiagnosticCategory.Message,"Performance_timings_for_diagnostics_or_extendedDiagnostics_are_not_available_in_this_session_A_nativ_6386","Performance timings for '--diagnostics' or '--extendedDiagnostics' are not available in this session. A native implementation of the Web Performance API could not be found."),The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1:n(6500,e.DiagnosticCategory.Message,"The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1_6500","The expected type comes from property '{0}' which is declared here on type '{1}'"),The_expected_type_comes_from_this_index_signature:n(6501,e.DiagnosticCategory.Message,"The_expected_type_comes_from_this_index_signature_6501","The expected type comes from this index signature."),The_expected_type_comes_from_the_return_type_of_this_signature:n(6502,e.DiagnosticCategory.Message,"The_expected_type_comes_from_the_return_type_of_this_signature_6502","The expected type comes from the return type of this signature."),Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing:n(6503,e.DiagnosticCategory.Message,"Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing_6503","Print names of files that are part of the compilation and then stop processing."),File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option:n(6504,e.DiagnosticCategory.Error,"File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option_6504","File '{0}' is a JavaScript file. Did you mean to enable the 'allowJs' option?"),Include_undefined_in_index_signature_results:n(6800,e.DiagnosticCategory.Message,"Include_undefined_in_index_signature_results_6800","Include 'undefined' in index signature results"),Variable_0_implicitly_has_an_1_type:n(7005,e.DiagnosticCategory.Error,"Variable_0_implicitly_has_an_1_type_7005","Variable '{0}' implicitly has an '{1}' type."),Parameter_0_implicitly_has_an_1_type:n(7006,e.DiagnosticCategory.Error,"Parameter_0_implicitly_has_an_1_type_7006","Parameter '{0}' implicitly has an '{1}' type."),Member_0_implicitly_has_an_1_type:n(7008,e.DiagnosticCategory.Error,"Member_0_implicitly_has_an_1_type_7008","Member '{0}' implicitly has an '{1}' type."),new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type:n(7009,e.DiagnosticCategory.Error,"new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type_7009","'new' expression, whose target lacks a construct signature, implicitly has an 'any' type."),_0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type:n(7010,e.DiagnosticCategory.Error,"_0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type_7010","'{0}', which lacks return-type annotation, implicitly has an '{1}' return type."),Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type:n(7011,e.DiagnosticCategory.Error,"Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7011","Function expression, which lacks return-type annotation, implicitly has an '{0}' return type."),Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type:n(7013,e.DiagnosticCategory.Error,"Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7013","Construct signature, which lacks return-type annotation, implicitly has an 'any' return type."),Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type:n(7014,e.DiagnosticCategory.Error,"Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7014","Function type, which lacks return-type annotation, implicitly has an '{0}' return type."),Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number:n(7015,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number_7015","Element implicitly has an 'any' type because index expression is not of type 'number'."),Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type:n(7016,e.DiagnosticCategory.Error,"Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type_7016","Could not find a declaration file for module '{0}'. '{1}' implicitly has an 'any' type."),Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature:n(7017,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_7017","Element implicitly has an 'any' type because type '{0}' has no index signature."),Object_literal_s_property_0_implicitly_has_an_1_type:n(7018,e.DiagnosticCategory.Error,"Object_literal_s_property_0_implicitly_has_an_1_type_7018","Object literal's property '{0}' implicitly has an '{1}' type."),Rest_parameter_0_implicitly_has_an_any_type:n(7019,e.DiagnosticCategory.Error,"Rest_parameter_0_implicitly_has_an_any_type_7019","Rest parameter '{0}' implicitly has an 'any[]' type."),Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type:n(7020,e.DiagnosticCategory.Error,"Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7020","Call signature, which lacks return-type annotation, implicitly has an 'any' return type."),_0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer:n(7022,e.DiagnosticCategory.Error,"_0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or__7022","'{0}' implicitly has type 'any' because it does not have a type annotation and is referenced directly or indirectly in its own initializer."),_0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions:n(7023,e.DiagnosticCategory.Error,"_0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_reference_7023","'{0}' implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions:n(7024,e.DiagnosticCategory.Error,"Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_ref_7024","Function implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type_annotation:n(7025,e.DiagnosticCategory.Error,"Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_retu_7025","Generator implicitly has yield type '{0}' because it does not yield any values. Consider supplying a return type annotation."),JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists:n(7026,e.DiagnosticCategory.Error,"JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists_7026","JSX element implicitly has type 'any' because no interface 'JSX.{0}' exists."),Unreachable_code_detected:n(7027,e.DiagnosticCategory.Error,"Unreachable_code_detected_7027","Unreachable code detected.",!0),Unused_label:n(7028,e.DiagnosticCategory.Error,"Unused_label_7028","Unused label.",!0),Fallthrough_case_in_switch:n(7029,e.DiagnosticCategory.Error,"Fallthrough_case_in_switch_7029","Fallthrough case in switch."),Not_all_code_paths_return_a_value:n(7030,e.DiagnosticCategory.Error,"Not_all_code_paths_return_a_value_7030","Not all code paths return a value."),Binding_element_0_implicitly_has_an_1_type:n(7031,e.DiagnosticCategory.Error,"Binding_element_0_implicitly_has_an_1_type_7031","Binding element '{0}' implicitly has an '{1}' type."),Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation:n(7032,e.DiagnosticCategory.Error,"Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation_7032","Property '{0}' implicitly has type 'any', because its set accessor lacks a parameter type annotation."),Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation:n(7033,e.DiagnosticCategory.Error,"Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation_7033","Property '{0}' implicitly has type 'any', because its get accessor lacks a return type annotation."),Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined:n(7034,e.DiagnosticCategory.Error,"Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined_7034","Variable '{0}' implicitly has type '{1}' in some locations where its type cannot be determined."),Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0:n(7035,e.DiagnosticCategory.Error,"Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare__7035","Try `npm i --save-dev @types/{1}` if it exists or add a new declaration (.d.ts) file containing `declare module '{0}';`"),Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0:n(7036,e.DiagnosticCategory.Error,"Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0_7036","Dynamic import's specifier must be of type 'string', but here has type '{0}'."),Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for_all_imports_Implies_allowSyntheticDefaultImports:n(7037,e.DiagnosticCategory.Message,"Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for__7037","Enables emit interoperability between CommonJS and ES Modules via creation of namespace objects for all imports. Implies 'allowSyntheticDefaultImports'."),Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead:n(7038,e.DiagnosticCategory.Message,"Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cau_7038","Type originates at this import. A namespace-style import cannot be called or constructed, and will cause a failure at runtime. Consider using a default import or import require here instead."),Mapped_object_type_implicitly_has_an_any_template_type:n(7039,e.DiagnosticCategory.Error,"Mapped_object_type_implicitly_has_an_any_template_type_7039","Mapped object type implicitly has an 'any' template type."),If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1:n(7040,e.DiagnosticCategory.Error,"If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_S_7040","If the '{0}' package actually exposes this module, consider sending a pull request to amend 'https://github.com/DefinitelyTyped/DefinitelyTyped/tree/master/types/{1}`"),The_containing_arrow_function_captures_the_global_value_of_this:n(7041,e.DiagnosticCategory.Error,"The_containing_arrow_function_captures_the_global_value_of_this_7041","The containing arrow function captures the global value of 'this'."),Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used:n(7042,e.DiagnosticCategory.Error,"Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used_7042","Module '{0}' was resolved to '{1}', but '--resolveJsonModule' is not used."),Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7043,e.DiagnosticCategory.Suggestion,"Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7043","Variable '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7044,e.DiagnosticCategory.Suggestion,"Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7044","Parameter '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage:n(7045,e.DiagnosticCategory.Suggestion,"Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7045","Member '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage:n(7046,e.DiagnosticCategory.Suggestion,"Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage_7046","Variable '{0}' implicitly has type '{1}' in some locations, but a better type may be inferred from usage."),Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage:n(7047,e.DiagnosticCategory.Suggestion,"Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage_7047","Rest parameter '{0}' implicitly has an 'any[]' type, but a better type may be inferred from usage."),Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage:n(7048,e.DiagnosticCategory.Suggestion,"Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage_7048","Property '{0}' implicitly has type 'any', but a better type for its get accessor may be inferred from usage."),Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage:n(7049,e.DiagnosticCategory.Suggestion,"Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage_7049","Property '{0}' implicitly has type 'any', but a better type for its set accessor may be inferred from usage."),_0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage:n(7050,e.DiagnosticCategory.Suggestion,"_0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage_7050","'{0}' implicitly has an '{1}' return type, but a better type may be inferred from usage."),Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1:n(7051,e.DiagnosticCategory.Error,"Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1_7051","Parameter has a name but no type. Did you mean '{0}: {1}'?"),Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1:n(7052,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1_7052","Element implicitly has an 'any' type because type '{0}' has no index signature. Did you mean to call '{1}'?"),Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1:n(7053,e.DiagnosticCategory.Error,"Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1_7053","Element implicitly has an 'any' type because expression of type '{0}' can't be used to index type '{1}'."),No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1:n(7054,e.DiagnosticCategory.Error,"No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1_7054","No index signature with a parameter of type '{0}' was found on type '{1}'."),_0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type:n(7055,e.DiagnosticCategory.Error,"_0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type_7055","'{0}', which lacks return-type annotation, implicitly has an '{1}' yield type."),The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_type_annotation_is_needed:n(7056,e.DiagnosticCategory.Error,"The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_ty_7056","The inferred type of this node exceeds the maximum length the compiler will serialize. An explicit type annotation is needed."),You_cannot_rename_this_element:n(8e3,e.DiagnosticCategory.Error,"You_cannot_rename_this_element_8000","You cannot rename this element."),You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library:n(8001,e.DiagnosticCategory.Error,"You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library_8001","You cannot rename elements that are defined in the standard TypeScript library."),import_can_only_be_used_in_TypeScript_files:n(8002,e.DiagnosticCategory.Error,"import_can_only_be_used_in_TypeScript_files_8002","'import ... =' can only be used in TypeScript files."),export_can_only_be_used_in_TypeScript_files:n(8003,e.DiagnosticCategory.Error,"export_can_only_be_used_in_TypeScript_files_8003","'export =' can only be used in TypeScript files."),Type_parameter_declarations_can_only_be_used_in_TypeScript_files:n(8004,e.DiagnosticCategory.Error,"Type_parameter_declarations_can_only_be_used_in_TypeScript_files_8004","Type parameter declarations can only be used in TypeScript files."),implements_clauses_can_only_be_used_in_TypeScript_files:n(8005,e.DiagnosticCategory.Error,"implements_clauses_can_only_be_used_in_TypeScript_files_8005","'implements' clauses can only be used in TypeScript files."),_0_declarations_can_only_be_used_in_TypeScript_files:n(8006,e.DiagnosticCategory.Error,"_0_declarations_can_only_be_used_in_TypeScript_files_8006","'{0}' declarations can only be used in TypeScript files."),Type_aliases_can_only_be_used_in_TypeScript_files:n(8008,e.DiagnosticCategory.Error,"Type_aliases_can_only_be_used_in_TypeScript_files_8008","Type aliases can only be used in TypeScript files."),The_0_modifier_can_only_be_used_in_TypeScript_files:n(8009,e.DiagnosticCategory.Error,"The_0_modifier_can_only_be_used_in_TypeScript_files_8009","The '{0}' modifier can only be used in TypeScript files."),Type_annotations_can_only_be_used_in_TypeScript_files:n(8010,e.DiagnosticCategory.Error,"Type_annotations_can_only_be_used_in_TypeScript_files_8010","Type annotations can only be used in TypeScript files."),Type_arguments_can_only_be_used_in_TypeScript_files:n(8011,e.DiagnosticCategory.Error,"Type_arguments_can_only_be_used_in_TypeScript_files_8011","Type arguments can only be used in TypeScript files."),Parameter_modifiers_can_only_be_used_in_TypeScript_files:n(8012,e.DiagnosticCategory.Error,"Parameter_modifiers_can_only_be_used_in_TypeScript_files_8012","Parameter modifiers can only be used in TypeScript files."),Non_null_assertions_can_only_be_used_in_TypeScript_files:n(8013,e.DiagnosticCategory.Error,"Non_null_assertions_can_only_be_used_in_TypeScript_files_8013","Non-null assertions can only be used in TypeScript files."),Type_assertion_expressions_can_only_be_used_in_TypeScript_files:n(8016,e.DiagnosticCategory.Error,"Type_assertion_expressions_can_only_be_used_in_TypeScript_files_8016","Type assertion expressions can only be used in TypeScript files."),Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0:n(8017,e.DiagnosticCategory.Error,"Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0_8017","Octal literal types must use ES2015 syntax. Use the syntax '{0}'."),Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0:n(8018,e.DiagnosticCategory.Error,"Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0_8018","Octal literals are not allowed in enums members initializer. Use the syntax '{0}'."),Report_errors_in_js_files:n(8019,e.DiagnosticCategory.Message,"Report_errors_in_js_files_8019","Report errors in .js files."),JSDoc_types_can_only_be_used_inside_documentation_comments:n(8020,e.DiagnosticCategory.Error,"JSDoc_types_can_only_be_used_inside_documentation_comments_8020","JSDoc types can only be used inside documentation comments."),JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags:n(8021,e.DiagnosticCategory.Error,"JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags_8021","JSDoc '@typedef' tag should either have a type annotation or be followed by '@property' or '@member' tags."),JSDoc_0_is_not_attached_to_a_class:n(8022,e.DiagnosticCategory.Error,"JSDoc_0_is_not_attached_to_a_class_8022","JSDoc '@{0}' is not attached to a class."),JSDoc_0_1_does_not_match_the_extends_2_clause:n(8023,e.DiagnosticCategory.Error,"JSDoc_0_1_does_not_match_the_extends_2_clause_8023","JSDoc '@{0} {1}' does not match the 'extends {2}' clause."),JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name:n(8024,e.DiagnosticCategory.Error,"JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_8024","JSDoc '@param' tag has name '{0}', but there is no parameter with that name."),Class_declarations_cannot_have_more_than_one_augments_or_extends_tag:n(8025,e.DiagnosticCategory.Error,"Class_declarations_cannot_have_more_than_one_augments_or_extends_tag_8025","Class declarations cannot have more than one `@augments` or `@extends` tag."),Expected_0_type_arguments_provide_these_with_an_extends_tag:n(8026,e.DiagnosticCategory.Error,"Expected_0_type_arguments_provide_these_with_an_extends_tag_8026","Expected {0} type arguments; provide these with an '@extends' tag."),Expected_0_1_type_arguments_provide_these_with_an_extends_tag:n(8027,e.DiagnosticCategory.Error,"Expected_0_1_type_arguments_provide_these_with_an_extends_tag_8027","Expected {0}-{1} type arguments; provide these with an '@extends' tag."),JSDoc_may_only_appear_in_the_last_parameter_of_a_signature:n(8028,e.DiagnosticCategory.Error,"JSDoc_may_only_appear_in_the_last_parameter_of_a_signature_8028","JSDoc '...' may only appear in the last parameter of a signature."),JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type:n(8029,e.DiagnosticCategory.Error,"JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_h_8029","JSDoc '@param' tag has name '{0}', but there is no parameter with that name. It would match 'arguments' if it had an array type."),The_type_of_a_function_declaration_must_match_the_function_s_signature:n(8030,e.DiagnosticCategory.Error,"The_type_of_a_function_declaration_must_match_the_function_s_signature_8030","The type of a function declaration must match the function's signature."),You_cannot_rename_a_module_via_a_global_import:n(8031,e.DiagnosticCategory.Error,"You_cannot_rename_a_module_via_a_global_import_8031","You cannot rename a module via a global import."),Qualified_name_0_is_not_allowed_without_a_leading_param_object_1:n(8032,e.DiagnosticCategory.Error,"Qualified_name_0_is_not_allowed_without_a_leading_param_object_1_8032","Qualified name '{0}' is not allowed without a leading '@param {object} {1}'."),A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags:n(8033,e.DiagnosticCategory.Error,"A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags_8033","A JSDoc '@typedef' comment may not contain multiple '@type' tags."),The_tag_was_first_specified_here:n(8034,e.DiagnosticCategory.Error,"The_tag_was_first_specified_here_8034","The tag was first specified here."),Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_class_extends_clause:n(9002,e.DiagnosticCategory.Error,"Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_clas_9002","Only identifiers/qualified-names with optional type arguments are currently supported in a class 'extends' clause."),class_expressions_are_not_currently_supported:n(9003,e.DiagnosticCategory.Error,"class_expressions_are_not_currently_supported_9003","'class' expressions are not currently supported."),Language_service_is_disabled:n(9004,e.DiagnosticCategory.Error,"Language_service_is_disabled_9004","Language service is disabled."),Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_declaration_emit:n(9005,e.DiagnosticCategory.Error,"Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_9005","Declaration emit for this file requires using private name '{0}'. An explicit type annotation may unblock declaration emit."),Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotation_may_unblock_declaration_emit:n(9006,e.DiagnosticCategory.Error,"Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotati_9006","Declaration emit for this file requires using private name '{0}' from module '{1}'. An explicit type annotation may unblock declaration emit."),JSX_attributes_must_only_be_assigned_a_non_empty_expression:n(17e3,e.DiagnosticCategory.Error,"JSX_attributes_must_only_be_assigned_a_non_empty_expression_17000","JSX attributes must only be assigned a non-empty 'expression'."),JSX_elements_cannot_have_multiple_attributes_with_the_same_name:n(17001,e.DiagnosticCategory.Error,"JSX_elements_cannot_have_multiple_attributes_with_the_same_name_17001","JSX elements cannot have multiple attributes with the same name."),Expected_corresponding_JSX_closing_tag_for_0:n(17002,e.DiagnosticCategory.Error,"Expected_corresponding_JSX_closing_tag_for_0_17002","Expected corresponding JSX closing tag for '{0}'."),JSX_attribute_expected:n(17003,e.DiagnosticCategory.Error,"JSX_attribute_expected_17003","JSX attribute expected."),Cannot_use_JSX_unless_the_jsx_flag_is_provided:n(17004,e.DiagnosticCategory.Error,"Cannot_use_JSX_unless_the_jsx_flag_is_provided_17004","Cannot use JSX unless the '--jsx' flag is provided."),A_constructor_cannot_contain_a_super_call_when_its_class_extends_null:n(17005,e.DiagnosticCategory.Error,"A_constructor_cannot_contain_a_super_call_when_its_class_extends_null_17005","A constructor cannot contain a 'super' call when its class extends 'null'."),An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses:n(17006,e.DiagnosticCategory.Error,"An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_ex_17006","An unary expression with the '{0}' operator is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses:n(17007,e.DiagnosticCategory.Error,"A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Con_17007","A type assertion expression is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),JSX_element_0_has_no_corresponding_closing_tag:n(17008,e.DiagnosticCategory.Error,"JSX_element_0_has_no_corresponding_closing_tag_17008","JSX element '{0}' has no corresponding closing tag."),super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class:n(17009,e.DiagnosticCategory.Error,"super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class_17009","'super' must be called before accessing 'this' in the constructor of a derived class."),Unknown_type_acquisition_option_0:n(17010,e.DiagnosticCategory.Error,"Unknown_type_acquisition_option_0_17010","Unknown type acquisition option '{0}'."),super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class:n(17011,e.DiagnosticCategory.Error,"super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class_17011","'super' must be called before accessing a property of 'super' in the constructor of a derived class."),_0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2:n(17012,e.DiagnosticCategory.Error,"_0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2_17012","'{0}' is not a valid meta-property for keyword '{1}'. Did you mean '{2}'?"),Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor:n(17013,e.DiagnosticCategory.Error,"Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constru_17013","Meta-property '{0}' is only allowed in the body of a function declaration, function expression, or constructor."),JSX_fragment_has_no_corresponding_closing_tag:n(17014,e.DiagnosticCategory.Error,"JSX_fragment_has_no_corresponding_closing_tag_17014","JSX fragment has no corresponding closing tag."),Expected_corresponding_closing_tag_for_JSX_fragment:n(17015,e.DiagnosticCategory.Error,"Expected_corresponding_closing_tag_for_JSX_fragment_17015","Expected corresponding closing tag for JSX fragment."),The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_compiler_option:n(17016,e.DiagnosticCategory.Error,"The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_com_17016","The 'jsxFragmentFactory' compiler option must be provided to use JSX fragments with the 'jsxFactory' compiler option."),An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments:n(17017,e.DiagnosticCategory.Error,"An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments_17017","An @jsxFrag pragma is required when using an @jsx pragma with JSX fragments."),Unknown_type_acquisition_option_0_Did_you_mean_1:n(17018,e.DiagnosticCategory.Error,"Unknown_type_acquisition_option_0_Did_you_mean_1_17018","Unknown type acquisition option '{0}'. Did you mean '{1}'?"),Circularity_detected_while_resolving_configuration_Colon_0:n(18e3,e.DiagnosticCategory.Error,"Circularity_detected_while_resolving_configuration_Colon_0_18000","Circularity detected while resolving configuration: {0}"),A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not:n(18001,e.DiagnosticCategory.Error,"A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not_18001","A path in an 'extends' option must be relative or rooted, but '{0}' is not."),The_files_list_in_config_file_0_is_empty:n(18002,e.DiagnosticCategory.Error,"The_files_list_in_config_file_0_is_empty_18002","The 'files' list in config file '{0}' is empty."),No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2:n(18003,e.DiagnosticCategory.Error,"No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2_18003","No inputs were found in config file '{0}'. Specified 'include' paths were '{1}' and 'exclude' paths were '{2}'."),File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module:n(80001,e.DiagnosticCategory.Suggestion,"File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module_80001","File is a CommonJS module; it may be converted to an ES6 module."),This_constructor_function_may_be_converted_to_a_class_declaration:n(80002,e.DiagnosticCategory.Suggestion,"This_constructor_function_may_be_converted_to_a_class_declaration_80002","This constructor function may be converted to a class declaration."),Import_may_be_converted_to_a_default_import:n(80003,e.DiagnosticCategory.Suggestion,"Import_may_be_converted_to_a_default_import_80003","Import may be converted to a default import."),JSDoc_types_may_be_moved_to_TypeScript_types:n(80004,e.DiagnosticCategory.Suggestion,"JSDoc_types_may_be_moved_to_TypeScript_types_80004","JSDoc types may be moved to TypeScript types."),require_call_may_be_converted_to_an_import:n(80005,e.DiagnosticCategory.Suggestion,"require_call_may_be_converted_to_an_import_80005","'require' call may be converted to an import."),This_may_be_converted_to_an_async_function:n(80006,e.DiagnosticCategory.Suggestion,"This_may_be_converted_to_an_async_function_80006","This may be converted to an async function."),await_has_no_effect_on_the_type_of_this_expression:n(80007,e.DiagnosticCategory.Suggestion,"await_has_no_effect_on_the_type_of_this_expression_80007","'await' has no effect on the type of this expression."),Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers:n(80008,e.DiagnosticCategory.Suggestion,"Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accur_80008","Numeric literals with absolute values equal to 2^53 or greater are too large to be represented accurately as integers."),Add_missing_super_call:n(90001,e.DiagnosticCategory.Message,"Add_missing_super_call_90001","Add missing 'super()' call"),Make_super_call_the_first_statement_in_the_constructor:n(90002,e.DiagnosticCategory.Message,"Make_super_call_the_first_statement_in_the_constructor_90002","Make 'super()' call the first statement in the constructor"),Change_extends_to_implements:n(90003,e.DiagnosticCategory.Message,"Change_extends_to_implements_90003","Change 'extends' to 'implements'"),Remove_unused_declaration_for_Colon_0:n(90004,e.DiagnosticCategory.Message,"Remove_unused_declaration_for_Colon_0_90004","Remove unused declaration for: '{0}'"),Remove_import_from_0:n(90005,e.DiagnosticCategory.Message,"Remove_import_from_0_90005","Remove import from '{0}'"),Implement_interface_0:n(90006,e.DiagnosticCategory.Message,"Implement_interface_0_90006","Implement interface '{0}'"),Implement_inherited_abstract_class:n(90007,e.DiagnosticCategory.Message,"Implement_inherited_abstract_class_90007","Implement inherited abstract class"),Add_0_to_unresolved_variable:n(90008,e.DiagnosticCategory.Message,"Add_0_to_unresolved_variable_90008","Add '{0}.' to unresolved variable"),Remove_variable_statement:n(90010,e.DiagnosticCategory.Message,"Remove_variable_statement_90010","Remove variable statement"),Remove_template_tag:n(90011,e.DiagnosticCategory.Message,"Remove_template_tag_90011","Remove template tag"),Remove_type_parameters:n(90012,e.DiagnosticCategory.Message,"Remove_type_parameters_90012","Remove type parameters"),Import_0_from_module_1:n(90013,e.DiagnosticCategory.Message,"Import_0_from_module_1_90013","Import '{0}' from module \"{1}\""),Change_0_to_1:n(90014,e.DiagnosticCategory.Message,"Change_0_to_1_90014","Change '{0}' to '{1}'"),Add_0_to_existing_import_declaration_from_1:n(90015,e.DiagnosticCategory.Message,"Add_0_to_existing_import_declaration_from_1_90015","Add '{0}' to existing import declaration from \"{1}\""),Declare_property_0:n(90016,e.DiagnosticCategory.Message,"Declare_property_0_90016","Declare property '{0}'"),Add_index_signature_for_property_0:n(90017,e.DiagnosticCategory.Message,"Add_index_signature_for_property_0_90017","Add index signature for property '{0}'"),Disable_checking_for_this_file:n(90018,e.DiagnosticCategory.Message,"Disable_checking_for_this_file_90018","Disable checking for this file"),Ignore_this_error_message:n(90019,e.DiagnosticCategory.Message,"Ignore_this_error_message_90019","Ignore this error message"),Initialize_property_0_in_the_constructor:n(90020,e.DiagnosticCategory.Message,"Initialize_property_0_in_the_constructor_90020","Initialize property '{0}' in the constructor"),Initialize_static_property_0:n(90021,e.DiagnosticCategory.Message,"Initialize_static_property_0_90021","Initialize static property '{0}'"),Change_spelling_to_0:n(90022,e.DiagnosticCategory.Message,"Change_spelling_to_0_90022","Change spelling to '{0}'"),Declare_method_0:n(90023,e.DiagnosticCategory.Message,"Declare_method_0_90023","Declare method '{0}'"),Declare_static_method_0:n(90024,e.DiagnosticCategory.Message,"Declare_static_method_0_90024","Declare static method '{0}'"),Prefix_0_with_an_underscore:n(90025,e.DiagnosticCategory.Message,"Prefix_0_with_an_underscore_90025","Prefix '{0}' with an underscore"),Rewrite_as_the_indexed_access_type_0:n(90026,e.DiagnosticCategory.Message,"Rewrite_as_the_indexed_access_type_0_90026","Rewrite as the indexed access type '{0}'"),Declare_static_property_0:n(90027,e.DiagnosticCategory.Message,"Declare_static_property_0_90027","Declare static property '{0}'"),Call_decorator_expression:n(90028,e.DiagnosticCategory.Message,"Call_decorator_expression_90028","Call decorator expression"),Add_async_modifier_to_containing_function:n(90029,e.DiagnosticCategory.Message,"Add_async_modifier_to_containing_function_90029","Add async modifier to containing function"),Replace_infer_0_with_unknown:n(90030,e.DiagnosticCategory.Message,"Replace_infer_0_with_unknown_90030","Replace 'infer {0}' with 'unknown'"),Replace_all_unused_infer_with_unknown:n(90031,e.DiagnosticCategory.Message,"Replace_all_unused_infer_with_unknown_90031","Replace all unused 'infer' with 'unknown'"),Import_default_0_from_module_1:n(90032,e.DiagnosticCategory.Message,"Import_default_0_from_module_1_90032","Import default '{0}' from module \"{1}\""),Add_default_import_0_to_existing_import_declaration_from_1:n(90033,e.DiagnosticCategory.Message,"Add_default_import_0_to_existing_import_declaration_from_1_90033","Add default import '{0}' to existing import declaration from \"{1}\""),Add_parameter_name:n(90034,e.DiagnosticCategory.Message,"Add_parameter_name_90034","Add parameter name"),Declare_private_property_0:n(90035,e.DiagnosticCategory.Message,"Declare_private_property_0_90035","Declare private property '{0}'"),Replace_0_with_Promise_1:n(90036,e.DiagnosticCategory.Message,"Replace_0_with_Promise_1_90036","Replace '{0}' with 'Promise<{1}>'"),Fix_all_incorrect_return_type_of_an_async_functions:n(90037,e.DiagnosticCategory.Message,"Fix_all_incorrect_return_type_of_an_async_functions_90037","Fix all incorrect return type of an async functions"),Declare_private_method_0:n(90038,e.DiagnosticCategory.Message,"Declare_private_method_0_90038","Declare private method '{0}'"),Remove_unused_destructuring_declaration:n(90039,e.DiagnosticCategory.Message,"Remove_unused_destructuring_declaration_90039","Remove unused destructuring declaration"),Remove_unused_declarations_for_Colon_0:n(90041,e.DiagnosticCategory.Message,"Remove_unused_declarations_for_Colon_0_90041","Remove unused declarations for: '{0}'"),Declare_a_private_field_named_0:n(90053,e.DiagnosticCategory.Message,"Declare_a_private_field_named_0_90053","Declare a private field named '{0}'."),Convert_function_to_an_ES2015_class:n(95001,e.DiagnosticCategory.Message,"Convert_function_to_an_ES2015_class_95001","Convert function to an ES2015 class"),Convert_function_0_to_class:n(95002,e.DiagnosticCategory.Message,"Convert_function_0_to_class_95002","Convert function '{0}' to class"),Convert_0_to_1_in_0:n(95003,e.DiagnosticCategory.Message,"Convert_0_to_1_in_0_95003","Convert '{0}' to '{1} in {0}'"),Extract_to_0_in_1:n(95004,e.DiagnosticCategory.Message,"Extract_to_0_in_1_95004","Extract to {0} in {1}"),Extract_function:n(95005,e.DiagnosticCategory.Message,"Extract_function_95005","Extract function"),Extract_constant:n(95006,e.DiagnosticCategory.Message,"Extract_constant_95006","Extract constant"),Extract_to_0_in_enclosing_scope:n(95007,e.DiagnosticCategory.Message,"Extract_to_0_in_enclosing_scope_95007","Extract to {0} in enclosing scope"),Extract_to_0_in_1_scope:n(95008,e.DiagnosticCategory.Message,"Extract_to_0_in_1_scope_95008","Extract to {0} in {1} scope"),Annotate_with_type_from_JSDoc:n(95009,e.DiagnosticCategory.Message,"Annotate_with_type_from_JSDoc_95009","Annotate with type from JSDoc"),Annotate_with_types_from_JSDoc:n(95010,e.DiagnosticCategory.Message,"Annotate_with_types_from_JSDoc_95010","Annotate with types from JSDoc"),Infer_type_of_0_from_usage:n(95011,e.DiagnosticCategory.Message,"Infer_type_of_0_from_usage_95011","Infer type of '{0}' from usage"),Infer_parameter_types_from_usage:n(95012,e.DiagnosticCategory.Message,"Infer_parameter_types_from_usage_95012","Infer parameter types from usage"),Convert_to_default_import:n(95013,e.DiagnosticCategory.Message,"Convert_to_default_import_95013","Convert to default import"),Install_0:n(95014,e.DiagnosticCategory.Message,"Install_0_95014","Install '{0}'"),Replace_import_with_0:n(95015,e.DiagnosticCategory.Message,"Replace_import_with_0_95015","Replace import with '{0}'."),Use_synthetic_default_member:n(95016,e.DiagnosticCategory.Message,"Use_synthetic_default_member_95016","Use synthetic 'default' member."),Convert_to_ES6_module:n(95017,e.DiagnosticCategory.Message,"Convert_to_ES6_module_95017","Convert to ES6 module"),Add_undefined_type_to_property_0:n(95018,e.DiagnosticCategory.Message,"Add_undefined_type_to_property_0_95018","Add 'undefined' type to property '{0}'"),Add_initializer_to_property_0:n(95019,e.DiagnosticCategory.Message,"Add_initializer_to_property_0_95019","Add initializer to property '{0}'"),Add_definite_assignment_assertion_to_property_0:n(95020,e.DiagnosticCategory.Message,"Add_definite_assignment_assertion_to_property_0_95020","Add definite assignment assertion to property '{0}'"),Convert_all_type_literals_to_mapped_type:n(95021,e.DiagnosticCategory.Message,"Convert_all_type_literals_to_mapped_type_95021","Convert all type literals to mapped type"),Add_all_missing_members:n(95022,e.DiagnosticCategory.Message,"Add_all_missing_members_95022","Add all missing members"),Infer_all_types_from_usage:n(95023,e.DiagnosticCategory.Message,"Infer_all_types_from_usage_95023","Infer all types from usage"),Delete_all_unused_declarations:n(95024,e.DiagnosticCategory.Message,"Delete_all_unused_declarations_95024","Delete all unused declarations"),Prefix_all_unused_declarations_with_where_possible:n(95025,e.DiagnosticCategory.Message,"Prefix_all_unused_declarations_with_where_possible_95025","Prefix all unused declarations with '_' where possible"),Fix_all_detected_spelling_errors:n(95026,e.DiagnosticCategory.Message,"Fix_all_detected_spelling_errors_95026","Fix all detected spelling errors"),Add_initializers_to_all_uninitialized_properties:n(95027,e.DiagnosticCategory.Message,"Add_initializers_to_all_uninitialized_properties_95027","Add initializers to all uninitialized properties"),Add_definite_assignment_assertions_to_all_uninitialized_properties:n(95028,e.DiagnosticCategory.Message,"Add_definite_assignment_assertions_to_all_uninitialized_properties_95028","Add definite assignment assertions to all uninitialized properties"),Add_undefined_type_to_all_uninitialized_properties:n(95029,e.DiagnosticCategory.Message,"Add_undefined_type_to_all_uninitialized_properties_95029","Add undefined type to all uninitialized properties"),Change_all_jsdoc_style_types_to_TypeScript:n(95030,e.DiagnosticCategory.Message,"Change_all_jsdoc_style_types_to_TypeScript_95030","Change all jsdoc-style types to TypeScript"),Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types:n(95031,e.DiagnosticCategory.Message,"Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types_95031","Change all jsdoc-style types to TypeScript (and add '| undefined' to nullable types)"),Implement_all_unimplemented_interfaces:n(95032,e.DiagnosticCategory.Message,"Implement_all_unimplemented_interfaces_95032","Implement all unimplemented interfaces"),Install_all_missing_types_packages:n(95033,e.DiagnosticCategory.Message,"Install_all_missing_types_packages_95033","Install all missing types packages"),Rewrite_all_as_indexed_access_types:n(95034,e.DiagnosticCategory.Message,"Rewrite_all_as_indexed_access_types_95034","Rewrite all as indexed access types"),Convert_all_to_default_imports:n(95035,e.DiagnosticCategory.Message,"Convert_all_to_default_imports_95035","Convert all to default imports"),Make_all_super_calls_the_first_statement_in_their_constructor:n(95036,e.DiagnosticCategory.Message,"Make_all_super_calls_the_first_statement_in_their_constructor_95036","Make all 'super()' calls the first statement in their constructor"),Add_qualifier_to_all_unresolved_variables_matching_a_member_name:n(95037,e.DiagnosticCategory.Message,"Add_qualifier_to_all_unresolved_variables_matching_a_member_name_95037","Add qualifier to all unresolved variables matching a member name"),Change_all_extended_interfaces_to_implements:n(95038,e.DiagnosticCategory.Message,"Change_all_extended_interfaces_to_implements_95038","Change all extended interfaces to 'implements'"),Add_all_missing_super_calls:n(95039,e.DiagnosticCategory.Message,"Add_all_missing_super_calls_95039","Add all missing super calls"),Implement_all_inherited_abstract_classes:n(95040,e.DiagnosticCategory.Message,"Implement_all_inherited_abstract_classes_95040","Implement all inherited abstract classes"),Add_all_missing_async_modifiers:n(95041,e.DiagnosticCategory.Message,"Add_all_missing_async_modifiers_95041","Add all missing 'async' modifiers"),Add_ts_ignore_to_all_error_messages:n(95042,e.DiagnosticCategory.Message,"Add_ts_ignore_to_all_error_messages_95042","Add '@ts-ignore' to all error messages"),Annotate_everything_with_types_from_JSDoc:n(95043,e.DiagnosticCategory.Message,"Annotate_everything_with_types_from_JSDoc_95043","Annotate everything with types from JSDoc"),Add_to_all_uncalled_decorators:n(95044,e.DiagnosticCategory.Message,"Add_to_all_uncalled_decorators_95044","Add '()' to all uncalled decorators"),Convert_all_constructor_functions_to_classes:n(95045,e.DiagnosticCategory.Message,"Convert_all_constructor_functions_to_classes_95045","Convert all constructor functions to classes"),Generate_get_and_set_accessors:n(95046,e.DiagnosticCategory.Message,"Generate_get_and_set_accessors_95046","Generate 'get' and 'set' accessors"),Convert_require_to_import:n(95047,e.DiagnosticCategory.Message,"Convert_require_to_import_95047","Convert 'require' to 'import'"),Convert_all_require_to_import:n(95048,e.DiagnosticCategory.Message,"Convert_all_require_to_import_95048","Convert all 'require' to 'import'"),Move_to_a_new_file:n(95049,e.DiagnosticCategory.Message,"Move_to_a_new_file_95049","Move to a new file"),Remove_unreachable_code:n(95050,e.DiagnosticCategory.Message,"Remove_unreachable_code_95050","Remove unreachable code"),Remove_all_unreachable_code:n(95051,e.DiagnosticCategory.Message,"Remove_all_unreachable_code_95051","Remove all unreachable code"),Add_missing_typeof:n(95052,e.DiagnosticCategory.Message,"Add_missing_typeof_95052","Add missing 'typeof'"),Remove_unused_label:n(95053,e.DiagnosticCategory.Message,"Remove_unused_label_95053","Remove unused label"),Remove_all_unused_labels:n(95054,e.DiagnosticCategory.Message,"Remove_all_unused_labels_95054","Remove all unused labels"),Convert_0_to_mapped_object_type:n(95055,e.DiagnosticCategory.Message,"Convert_0_to_mapped_object_type_95055","Convert '{0}' to mapped object type"),Convert_namespace_import_to_named_imports:n(95056,e.DiagnosticCategory.Message,"Convert_namespace_import_to_named_imports_95056","Convert namespace import to named imports"),Convert_named_imports_to_namespace_import:n(95057,e.DiagnosticCategory.Message,"Convert_named_imports_to_namespace_import_95057","Convert named imports to namespace import"),Add_or_remove_braces_in_an_arrow_function:n(95058,e.DiagnosticCategory.Message,"Add_or_remove_braces_in_an_arrow_function_95058","Add or remove braces in an arrow function"),Add_braces_to_arrow_function:n(95059,e.DiagnosticCategory.Message,"Add_braces_to_arrow_function_95059","Add braces to arrow function"),Remove_braces_from_arrow_function:n(95060,e.DiagnosticCategory.Message,"Remove_braces_from_arrow_function_95060","Remove braces from arrow function"),Convert_default_export_to_named_export:n(95061,e.DiagnosticCategory.Message,"Convert_default_export_to_named_export_95061","Convert default export to named export"),Convert_named_export_to_default_export:n(95062,e.DiagnosticCategory.Message,"Convert_named_export_to_default_export_95062","Convert named export to default export"),Add_missing_enum_member_0:n(95063,e.DiagnosticCategory.Message,"Add_missing_enum_member_0_95063","Add missing enum member '{0}'"),Add_all_missing_imports:n(95064,e.DiagnosticCategory.Message,"Add_all_missing_imports_95064","Add all missing imports"),Convert_to_async_function:n(95065,e.DiagnosticCategory.Message,"Convert_to_async_function_95065","Convert to async function"),Convert_all_to_async_functions:n(95066,e.DiagnosticCategory.Message,"Convert_all_to_async_functions_95066","Convert all to async functions"),Add_missing_call_parentheses:n(95067,e.DiagnosticCategory.Message,"Add_missing_call_parentheses_95067","Add missing call parentheses"),Add_all_missing_call_parentheses:n(95068,e.DiagnosticCategory.Message,"Add_all_missing_call_parentheses_95068","Add all missing call parentheses"),Add_unknown_conversion_for_non_overlapping_types:n(95069,e.DiagnosticCategory.Message,"Add_unknown_conversion_for_non_overlapping_types_95069","Add 'unknown' conversion for non-overlapping types"),Add_unknown_to_all_conversions_of_non_overlapping_types:n(95070,e.DiagnosticCategory.Message,"Add_unknown_to_all_conversions_of_non_overlapping_types_95070","Add 'unknown' to all conversions of non-overlapping types"),Add_missing_new_operator_to_call:n(95071,e.DiagnosticCategory.Message,"Add_missing_new_operator_to_call_95071","Add missing 'new' operator to call"),Add_missing_new_operator_to_all_calls:n(95072,e.DiagnosticCategory.Message,"Add_missing_new_operator_to_all_calls_95072","Add missing 'new' operator to all calls"),Add_names_to_all_parameters_without_names:n(95073,e.DiagnosticCategory.Message,"Add_names_to_all_parameters_without_names_95073","Add names to all parameters without names"),Enable_the_experimentalDecorators_option_in_your_configuration_file:n(95074,e.DiagnosticCategory.Message,"Enable_the_experimentalDecorators_option_in_your_configuration_file_95074","Enable the 'experimentalDecorators' option in your configuration file"),Convert_parameters_to_destructured_object:n(95075,e.DiagnosticCategory.Message,"Convert_parameters_to_destructured_object_95075","Convert parameters to destructured object"),Allow_accessing_UMD_globals_from_modules:n(95076,e.DiagnosticCategory.Message,"Allow_accessing_UMD_globals_from_modules_95076","Allow accessing UMD globals from modules."),Extract_type:n(95077,e.DiagnosticCategory.Message,"Extract_type_95077","Extract type"),Extract_to_type_alias:n(95078,e.DiagnosticCategory.Message,"Extract_to_type_alias_95078","Extract to type alias"),Extract_to_typedef:n(95079,e.DiagnosticCategory.Message,"Extract_to_typedef_95079","Extract to typedef"),Infer_this_type_of_0_from_usage:n(95080,e.DiagnosticCategory.Message,"Infer_this_type_of_0_from_usage_95080","Infer 'this' type of '{0}' from usage"),Add_const_to_unresolved_variable:n(95081,e.DiagnosticCategory.Message,"Add_const_to_unresolved_variable_95081","Add 'const' to unresolved variable"),Add_const_to_all_unresolved_variables:n(95082,e.DiagnosticCategory.Message,"Add_const_to_all_unresolved_variables_95082","Add 'const' to all unresolved variables"),Add_await:n(95083,e.DiagnosticCategory.Message,"Add_await_95083","Add 'await'"),Add_await_to_initializer_for_0:n(95084,e.DiagnosticCategory.Message,"Add_await_to_initializer_for_0_95084","Add 'await' to initializer for '{0}'"),Fix_all_expressions_possibly_missing_await:n(95085,e.DiagnosticCategory.Message,"Fix_all_expressions_possibly_missing_await_95085","Fix all expressions possibly missing 'await'"),Remove_unnecessary_await:n(95086,e.DiagnosticCategory.Message,"Remove_unnecessary_await_95086","Remove unnecessary 'await'"),Remove_all_unnecessary_uses_of_await:n(95087,e.DiagnosticCategory.Message,"Remove_all_unnecessary_uses_of_await_95087","Remove all unnecessary uses of 'await'"),Enable_the_jsx_flag_in_your_configuration_file:n(95088,e.DiagnosticCategory.Message,"Enable_the_jsx_flag_in_your_configuration_file_95088","Enable the '--jsx' flag in your configuration file"),Add_await_to_initializers:n(95089,e.DiagnosticCategory.Message,"Add_await_to_initializers_95089","Add 'await' to initializers"),Extract_to_interface:n(95090,e.DiagnosticCategory.Message,"Extract_to_interface_95090","Extract to interface"),Convert_to_a_bigint_numeric_literal:n(95091,e.DiagnosticCategory.Message,"Convert_to_a_bigint_numeric_literal_95091","Convert to a bigint numeric literal"),Convert_all_to_bigint_numeric_literals:n(95092,e.DiagnosticCategory.Message,"Convert_all_to_bigint_numeric_literals_95092","Convert all to bigint numeric literals"),Convert_const_to_let:n(95093,e.DiagnosticCategory.Message,"Convert_const_to_let_95093","Convert 'const' to 'let'"),Prefix_with_declare:n(95094,e.DiagnosticCategory.Message,"Prefix_with_declare_95094","Prefix with 'declare'"),Prefix_all_incorrect_property_declarations_with_declare:n(95095,e.DiagnosticCategory.Message,"Prefix_all_incorrect_property_declarations_with_declare_95095","Prefix all incorrect property declarations with 'declare'"),Convert_to_template_string:n(95096,e.DiagnosticCategory.Message,"Convert_to_template_string_95096","Convert to template string"),Add_export_to_make_this_file_into_a_module:n(95097,e.DiagnosticCategory.Message,"Add_export_to_make_this_file_into_a_module_95097","Add 'export {}' to make this file into a module"),Set_the_target_option_in_your_configuration_file_to_0:n(95098,e.DiagnosticCategory.Message,"Set_the_target_option_in_your_configuration_file_to_0_95098","Set the 'target' option in your configuration file to '{0}'"),Set_the_module_option_in_your_configuration_file_to_0:n(95099,e.DiagnosticCategory.Message,"Set_the_module_option_in_your_configuration_file_to_0_95099","Set the 'module' option in your configuration file to '{0}'"),Convert_invalid_character_to_its_html_entity_code:n(95100,e.DiagnosticCategory.Message,"Convert_invalid_character_to_its_html_entity_code_95100","Convert invalid character to its html entity code"),Convert_all_invalid_characters_to_HTML_entity_code:n(95101,e.DiagnosticCategory.Message,"Convert_all_invalid_characters_to_HTML_entity_code_95101","Convert all invalid characters to HTML entity code"),Add_class_tag:n(95102,e.DiagnosticCategory.Message,"Add_class_tag_95102","Add '@class' tag"),Add_this_tag:n(95103,e.DiagnosticCategory.Message,"Add_this_tag_95103","Add '@this' tag"),Add_this_parameter:n(95104,e.DiagnosticCategory.Message,"Add_this_parameter_95104","Add 'this' parameter."),Convert_function_expression_0_to_arrow_function:n(95105,e.DiagnosticCategory.Message,"Convert_function_expression_0_to_arrow_function_95105","Convert function expression '{0}' to arrow function"),Convert_function_declaration_0_to_arrow_function:n(95106,e.DiagnosticCategory.Message,"Convert_function_declaration_0_to_arrow_function_95106","Convert function declaration '{0}' to arrow function"),Fix_all_implicit_this_errors:n(95107,e.DiagnosticCategory.Message,"Fix_all_implicit_this_errors_95107","Fix all implicit-'this' errors"),Wrap_invalid_character_in_an_expression_container:n(95108,e.DiagnosticCategory.Message,"Wrap_invalid_character_in_an_expression_container_95108","Wrap invalid character in an expression container"),Wrap_all_invalid_characters_in_an_expression_container:n(95109,e.DiagnosticCategory.Message,"Wrap_all_invalid_characters_in_an_expression_container_95109","Wrap all invalid characters in an expression container"),Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file:n(95110,e.DiagnosticCategory.Message,"Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file_95110","Visit https://aka.ms/tsconfig.json to read more about this file"),Add_a_return_statement:n(95111,e.DiagnosticCategory.Message,"Add_a_return_statement_95111","Add a return statement"),Remove_braces_from_arrow_function_body:n(95112,e.DiagnosticCategory.Message,"Remove_braces_from_arrow_function_body_95112","Remove braces from arrow function body"),Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal:n(95113,e.DiagnosticCategory.Message,"Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal_95113","Wrap the following body with parentheses which should be an object literal"),Add_all_missing_return_statement:n(95114,e.DiagnosticCategory.Message,"Add_all_missing_return_statement_95114","Add all missing return statement"),Remove_braces_from_all_arrow_function_bodies_with_relevant_issues:n(95115,e.DiagnosticCategory.Message,"Remove_braces_from_all_arrow_function_bodies_with_relevant_issues_95115","Remove braces from all arrow function bodies with relevant issues"),Wrap_all_object_literal_with_parentheses:n(95116,e.DiagnosticCategory.Message,"Wrap_all_object_literal_with_parentheses_95116","Wrap all object literal with parentheses"),Move_labeled_tuple_element_modifiers_to_labels:n(95117,e.DiagnosticCategory.Message,"Move_labeled_tuple_element_modifiers_to_labels_95117","Move labeled tuple element modifiers to labels"),Convert_overload_list_to_single_signature:n(95118,e.DiagnosticCategory.Message,"Convert_overload_list_to_single_signature_95118","Convert overload list to single signature"),Generate_get_and_set_accessors_for_all_overriding_properties:n(95119,e.DiagnosticCategory.Message,"Generate_get_and_set_accessors_for_all_overriding_properties_95119","Generate 'get' and 'set' accessors for all overriding properties"),Wrap_in_JSX_fragment:n(95120,e.DiagnosticCategory.Message,"Wrap_in_JSX_fragment_95120","Wrap in JSX fragment"),Wrap_all_unparented_JSX_in_JSX_fragment:n(95121,e.DiagnosticCategory.Message,"Wrap_all_unparented_JSX_in_JSX_fragment_95121","Wrap all unparented JSX in JSX fragment"),Convert_arrow_function_or_function_expression:n(95122,e.DiagnosticCategory.Message,"Convert_arrow_function_or_function_expression_95122","Convert arrow function or function expression"),Convert_to_anonymous_function:n(95123,e.DiagnosticCategory.Message,"Convert_to_anonymous_function_95123","Convert to anonymous function"),Convert_to_named_function:n(95124,e.DiagnosticCategory.Message,"Convert_to_named_function_95124","Convert to named function"),Convert_to_arrow_function:n(95125,e.DiagnosticCategory.Message,"Convert_to_arrow_function_95125","Convert to arrow function"),Remove_parentheses:n(95126,e.DiagnosticCategory.Message,"Remove_parentheses_95126","Remove parentheses"),Could_not_find_a_containing_arrow_function:n(95127,e.DiagnosticCategory.Message,"Could_not_find_a_containing_arrow_function_95127","Could not find a containing arrow function"),Containing_function_is_not_an_arrow_function:n(95128,e.DiagnosticCategory.Message,"Containing_function_is_not_an_arrow_function_95128","Containing function is not an arrow function"),Could_not_find_export_statement:n(95129,e.DiagnosticCategory.Message,"Could_not_find_export_statement_95129","Could not find export statement"),This_file_already_has_a_default_export:n(95130,e.DiagnosticCategory.Message,"This_file_already_has_a_default_export_95130","This file already has a default export"),Could_not_find_import_clause:n(95131,e.DiagnosticCategory.Message,"Could_not_find_import_clause_95131","Could not find import clause"),Could_not_find_namespace_import_or_named_imports:n(95132,e.DiagnosticCategory.Message,"Could_not_find_namespace_import_or_named_imports_95132","Could not find namespace import or named imports"),Selection_is_not_a_valid_type_node:n(95133,e.DiagnosticCategory.Message,"Selection_is_not_a_valid_type_node_95133","Selection is not a valid type node"),No_type_could_be_extracted_from_this_type_node:n(95134,e.DiagnosticCategory.Message,"No_type_could_be_extracted_from_this_type_node_95134","No type could be extracted from this type node"),Could_not_find_property_for_which_to_generate_accessor:n(95135,e.DiagnosticCategory.Message,"Could_not_find_property_for_which_to_generate_accessor_95135","Could not find property for which to generate accessor"),Name_is_not_valid:n(95136,e.DiagnosticCategory.Message,"Name_is_not_valid_95136","Name is not valid"),Can_only_convert_property_with_modifier:n(95137,e.DiagnosticCategory.Message,"Can_only_convert_property_with_modifier_95137","Can only convert property with modifier"),Switch_each_misused_0_to_1:n(95138,e.DiagnosticCategory.Message,"Switch_each_misused_0_to_1_95138","Switch each misused '{0}' to '{1}'"),Convert_to_optional_chain_expression:n(95139,e.DiagnosticCategory.Message,"Convert_to_optional_chain_expression_95139","Convert to optional chain expression"),Could_not_find_convertible_access_expression:n(95140,e.DiagnosticCategory.Message,"Could_not_find_convertible_access_expression_95140","Could not find convertible access expression"),Could_not_find_matching_access_expressions:n(95141,e.DiagnosticCategory.Message,"Could_not_find_matching_access_expressions_95141","Could not find matching access expressions"),Can_only_convert_logical_AND_access_chains:n(95142,e.DiagnosticCategory.Message,"Can_only_convert_logical_AND_access_chains_95142","Can only convert logical AND access chains"),Add_void_to_Promise_resolved_without_a_value:n(95143,e.DiagnosticCategory.Message,"Add_void_to_Promise_resolved_without_a_value_95143","Add 'void' to Promise resolved without a value"),Add_void_to_all_Promises_resolved_without_a_value:n(95144,e.DiagnosticCategory.Message,"Add_void_to_all_Promises_resolved_without_a_value_95144","Add 'void' to all Promises resolved without a value"),No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer:n(18004,e.DiagnosticCategory.Error,"No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer_18004","No value exists in scope for the shorthand property '{0}'. Either declare one or provide an initializer."),Classes_may_not_have_a_field_named_constructor:n(18006,e.DiagnosticCategory.Error,"Classes_may_not_have_a_field_named_constructor_18006","Classes may not have a field named 'constructor'."),JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array:n(18007,e.DiagnosticCategory.Error,"JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array_18007","JSX expressions may not use the comma operator. Did you mean to write an array?"),Private_identifiers_cannot_be_used_as_parameters:n(18009,e.DiagnosticCategory.Error,"Private_identifiers_cannot_be_used_as_parameters_18009","Private identifiers cannot be used as parameters."),An_accessibility_modifier_cannot_be_used_with_a_private_identifier:n(18010,e.DiagnosticCategory.Error,"An_accessibility_modifier_cannot_be_used_with_a_private_identifier_18010","An accessibility modifier cannot be used with a private identifier."),The_operand_of_a_delete_operator_cannot_be_a_private_identifier:n(18011,e.DiagnosticCategory.Error,"The_operand_of_a_delete_operator_cannot_be_a_private_identifier_18011","The operand of a 'delete' operator cannot be a private identifier."),constructor_is_a_reserved_word:n(18012,e.DiagnosticCategory.Error,"constructor_is_a_reserved_word_18012","'#constructor' is a reserved word."),Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier:n(18013,e.DiagnosticCategory.Error,"Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier_18013","Property '{0}' is not accessible outside class '{1}' because it has a private identifier."),The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_private_identifier_with_the_same_spelling:n(18014,e.DiagnosticCategory.Error,"The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_priv_18014","The property '{0}' cannot be accessed on type '{1}' within this class because it is shadowed by another private identifier with the same spelling."),Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2:n(18015,e.DiagnosticCategory.Error,"Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2_18015","Property '{0}' in type '{1}' refers to a different member that cannot be accessed from within type '{2}'."),Private_identifiers_are_not_allowed_outside_class_bodies:n(18016,e.DiagnosticCategory.Error,"Private_identifiers_are_not_allowed_outside_class_bodies_18016","Private identifiers are not allowed outside class bodies."),The_shadowing_declaration_of_0_is_defined_here:n(18017,e.DiagnosticCategory.Error,"The_shadowing_declaration_of_0_is_defined_here_18017","The shadowing declaration of '{0}' is defined here"),The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here:n(18018,e.DiagnosticCategory.Error,"The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here_18018","The declaration of '{0}' that you probably intended to use is defined here"),_0_modifier_cannot_be_used_with_a_private_identifier:n(18019,e.DiagnosticCategory.Error,"_0_modifier_cannot_be_used_with_a_private_identifier_18019","'{0}' modifier cannot be used with a private identifier."),A_method_cannot_be_named_with_a_private_identifier:n(18022,e.DiagnosticCategory.Error,"A_method_cannot_be_named_with_a_private_identifier_18022","A method cannot be named with a private identifier."),An_accessor_cannot_be_named_with_a_private_identifier:n(18023,e.DiagnosticCategory.Error,"An_accessor_cannot_be_named_with_a_private_identifier_18023","An accessor cannot be named with a private identifier."),An_enum_member_cannot_be_named_with_a_private_identifier:n(18024,e.DiagnosticCategory.Error,"An_enum_member_cannot_be_named_with_a_private_identifier_18024","An enum member cannot be named with a private identifier."),can_only_be_used_at_the_start_of_a_file:n(18026,e.DiagnosticCategory.Error,"can_only_be_used_at_the_start_of_a_file_18026","'#!' can only be used at the start of a file."),Compiler_reserves_name_0_when_emitting_private_identifier_downlevel:n(18027,e.DiagnosticCategory.Error,"Compiler_reserves_name_0_when_emitting_private_identifier_downlevel_18027","Compiler reserves name '{0}' when emitting private identifier downlevel."),Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher:n(18028,e.DiagnosticCategory.Error,"Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher_18028","Private identifiers are only available when targeting ECMAScript 2015 and higher."),Private_identifiers_are_not_allowed_in_variable_declarations:n(18029,e.DiagnosticCategory.Error,"Private_identifiers_are_not_allowed_in_variable_declarations_18029","Private identifiers are not allowed in variable declarations."),An_optional_chain_cannot_contain_private_identifiers:n(18030,e.DiagnosticCategory.Error,"An_optional_chain_cannot_contain_private_identifiers_18030","An optional chain cannot contain private identifiers."),The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituents:n(18031,e.DiagnosticCategory.Error,"The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituent_18031","The intersection '{0}' was reduced to 'never' because property '{1}' has conflicting types in some constituents."),The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_private_in_some:n(18032,e.DiagnosticCategory.Error,"The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_pr_18032","The intersection '{0}' was reduced to 'never' because property '{1}' exists in multiple constituents and is private in some."),Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhaustiveness_checks_consider_using_an_object_literal_instead:n(18033,e.DiagnosticCategory.Error,"Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhau_18033","Only numeric enums can have computed members, but this expression has type '{0}'. If you do not need exhaustiveness checks, consider using an object literal instead."),Specify_the_JSX_fragment_factory_function_to_use_when_targeting_react_JSX_emit_with_jsxFactory_compiler_option_is_specified_e_g_Fragment:n(18034,e.DiagnosticCategory.Message,"Specify_the_JSX_fragment_factory_function_to_use_when_targeting_react_JSX_emit_with_jsxFactory_compi_18034","Specify the JSX fragment factory function to use when targeting 'react' JSX emit with 'jsxFactory' compiler option is specified, e.g. 'Fragment'."),Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name:n(18035,e.DiagnosticCategory.Error,"Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name_18035","Invalid value for 'jsxFragmentFactory'. '{0}' is not a valid identifier or qualified-name.")}}(ts||(ts={})),function(e){var n;function t(e){return e>=78}e.tokenIsIdentifierOrKeyword=t,e.tokenIsIdentifierOrKeywordOrGreaterThan=function(e){return 31===e||t(e)};var r=((n={abstract:125,any:128,as:126,asserts:127,bigint:155,boolean:131,break:80,case:81,catch:82,class:83,continue:85,const:84}).constructor=132,n.debugger=86,n.declare=133,n.default=87,n.delete=88,n.do=89,n.else=90,n.enum=91,n.export=92,n.extends=93,n.false=94,n.finally=95,n.for=96,n.from=153,n.function=97,n.get=134,n.if=98,n.implements=116,n.import=99,n.in=100,n.infer=135,n.instanceof=101,n.interface=117,n.intrinsic=136,n.is=137,n.keyof=138,n.let=118,n.module=139,n.namespace=140,n.never=141,n.new=102,n.null=103,n.number=144,n.object=145,n.package=119,n.private=120,n.protected=121,n.public=122,n.readonly=142,n.require=143,n.global=154,n.return=104,n.set=146,n.static=123,n.string=147,n.super=105,n.switch=106,n.symbol=148,n.this=107,n.throw=108,n.true=109,n.try=110,n.type=149,n.typeof=111,n.undefined=150,n.unique=151,n.unknown=152,n.var=112,n.void=113,n.while=114,n.with=115,n.yield=124,n.async=129,n.await=130,n.of=156,n),a=new e.Map(e.getEntries(r)),i=new e.Map(e.getEntries(__assign(__assign({},r),{"{":18,"}":19,"(":20,")":21,"[":22,"]":23,".":24,"...":25,";":26,",":27,"<":29,">":31,"<=":32,">=":33,"==":34,"!=":35,"===":36,"!==":37,"=>":38,"+":39,"-":40,"**":42,"*":41,"/":43,"%":44,"++":45,"--":46,"<<":47,"</":30,">>":48,">>>":49,"&":50,"|":51,"^":52,"!":53,"~":54,"&&":55,"||":56,"?":57,"??":60,"?.":28,":":58,"=":62,"+=":63,"-=":64,"*=":65,"**=":66,"/=":67,"%=":68,"<<=":69,">>=":70,">>>=":71,"&=":72,"|=":73,"^=":77,"||=":74,"&&=":75,"??=":76,"@":59,"`":61}))),o=[170,170,181,181,186,186,192,214,216,246,248,543,546,563,592,685,688,696,699,705,720,721,736,740,750,750,890,890,902,902,904,906,908,908,910,929,931,974,976,983,986,1011,1024,1153,1164,1220,1223,1224,1227,1228,1232,1269,1272,1273,1329,1366,1369,1369,1377,1415,1488,1514,1520,1522,1569,1594,1600,1610,1649,1747,1749,1749,1765,1766,1786,1788,1808,1808,1810,1836,1920,1957,2309,2361,2365,2365,2384,2384,2392,2401,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2524,2525,2527,2529,2544,2545,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2649,2652,2654,2654,2674,2676,2693,2699,2701,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2749,2749,2768,2768,2784,2784,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2870,2873,2877,2877,2908,2909,2911,2913,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,2997,2999,3001,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3168,3169,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3294,3294,3296,3297,3333,3340,3342,3344,3346,3368,3370,3385,3424,3425,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3585,3632,3634,3635,3648,3654,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3760,3762,3763,3773,3773,3776,3780,3782,3782,3804,3805,3840,3840,3904,3911,3913,3946,3976,3979,4096,4129,4131,4135,4137,4138,4176,4181,4256,4293,4304,4342,4352,4441,4447,4514,4520,4601,4608,4614,4616,4678,4680,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4742,4744,4744,4746,4749,4752,4782,4784,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4814,4816,4822,4824,4846,4848,4878,4880,4880,4882,4885,4888,4894,4896,4934,4936,4954,5024,5108,5121,5740,5743,5750,5761,5786,5792,5866,6016,6067,6176,6263,6272,6312,7680,7835,7840,7929,7936,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8319,8319,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8497,8499,8505,8544,8579,12293,12295,12321,12329,12337,12341,12344,12346,12353,12436,12445,12446,12449,12538,12540,12542,12549,12588,12593,12686,12704,12727,13312,19893,19968,40869,40960,42124,44032,55203,63744,64045,64256,64262,64275,64279,64285,64285,64287,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65136,65138,65140,65140,65142,65276,65313,65338,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500],s=[170,170,181,181,186,186,192,214,216,246,248,543,546,563,592,685,688,696,699,705,720,721,736,740,750,750,768,846,864,866,890,890,902,902,904,906,908,908,910,929,931,974,976,983,986,1011,1024,1153,1155,1158,1164,1220,1223,1224,1227,1228,1232,1269,1272,1273,1329,1366,1369,1369,1377,1415,1425,1441,1443,1465,1467,1469,1471,1471,1473,1474,1476,1476,1488,1514,1520,1522,1569,1594,1600,1621,1632,1641,1648,1747,1749,1756,1759,1768,1770,1773,1776,1788,1808,1836,1840,1866,1920,1968,2305,2307,2309,2361,2364,2381,2384,2388,2392,2403,2406,2415,2433,2435,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2492,2492,2494,2500,2503,2504,2507,2509,2519,2519,2524,2525,2527,2531,2534,2545,2562,2562,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2620,2620,2622,2626,2631,2632,2635,2637,2649,2652,2654,2654,2662,2676,2689,2691,2693,2699,2701,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2748,2757,2759,2761,2763,2765,2768,2768,2784,2784,2790,2799,2817,2819,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2870,2873,2876,2883,2887,2888,2891,2893,2902,2903,2908,2909,2911,2913,2918,2927,2946,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,2997,2999,3001,3006,3010,3014,3016,3018,3021,3031,3031,3047,3055,3073,3075,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3134,3140,3142,3144,3146,3149,3157,3158,3168,3169,3174,3183,3202,3203,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3262,3268,3270,3272,3274,3277,3285,3286,3294,3294,3296,3297,3302,3311,3330,3331,3333,3340,3342,3344,3346,3368,3370,3385,3390,3395,3398,3400,3402,3405,3415,3415,3424,3425,3430,3439,3458,3459,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3530,3530,3535,3540,3542,3542,3544,3551,3570,3571,3585,3642,3648,3662,3664,3673,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3769,3771,3773,3776,3780,3782,3782,3784,3789,3792,3801,3804,3805,3840,3840,3864,3865,3872,3881,3893,3893,3895,3895,3897,3897,3902,3911,3913,3946,3953,3972,3974,3979,3984,3991,3993,4028,4038,4038,4096,4129,4131,4135,4137,4138,4140,4146,4150,4153,4160,4169,4176,4185,4256,4293,4304,4342,4352,4441,4447,4514,4520,4601,4608,4614,4616,4678,4680,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4742,4744,4744,4746,4749,4752,4782,4784,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4814,4816,4822,4824,4846,4848,4878,4880,4880,4882,4885,4888,4894,4896,4934,4936,4954,4969,4977,5024,5108,5121,5740,5743,5750,5761,5786,5792,5866,6016,6099,6112,6121,6160,6169,6176,6263,6272,6313,7680,7835,7840,7929,7936,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8255,8256,8319,8319,8400,8412,8417,8417,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8497,8499,8505,8544,8579,12293,12295,12321,12335,12337,12341,12344,12346,12353,12436,12441,12442,12445,12446,12449,12542,12549,12588,12593,12686,12704,12727,13312,19893,19968,40869,40960,42124,44032,55203,63744,64045,64256,64262,64275,64279,64285,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65056,65059,65075,65076,65101,65103,65136,65138,65140,65140,65142,65276,65296,65305,65313,65338,65343,65343,65345,65370,65381,65470,65474,65479,65482,65487,65490,65495,65498,65500],c=[170,170,181,181,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,880,884,886,887,890,893,902,902,904,906,908,908,910,929,931,1013,1015,1153,1162,1319,1329,1366,1369,1369,1377,1415,1488,1514,1520,1522,1568,1610,1646,1647,1649,1747,1749,1749,1765,1766,1774,1775,1786,1788,1791,1791,1808,1808,1810,1839,1869,1957,1969,1969,1994,2026,2036,2037,2042,2042,2048,2069,2074,2074,2084,2084,2088,2088,2112,2136,2208,2208,2210,2220,2308,2361,2365,2365,2384,2384,2392,2401,2417,2423,2425,2431,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2493,2493,2510,2510,2524,2525,2527,2529,2544,2545,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2649,2652,2654,2654,2674,2676,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2749,2749,2768,2768,2784,2785,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2877,2877,2908,2909,2911,2913,2929,2929,2947,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3024,3024,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3133,3133,3160,3161,3168,3169,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3261,3261,3294,3294,3296,3297,3313,3314,3333,3340,3342,3344,3346,3386,3389,3389,3406,3406,3424,3425,3450,3455,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3585,3632,3634,3635,3648,3654,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3760,3762,3763,3773,3773,3776,3780,3782,3782,3804,3807,3840,3840,3904,3911,3913,3948,3976,3980,4096,4138,4159,4159,4176,4181,4186,4189,4193,4193,4197,4198,4206,4208,4213,4225,4238,4238,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4992,5007,5024,5108,5121,5740,5743,5759,5761,5786,5792,5866,5870,5872,5888,5900,5902,5905,5920,5937,5952,5969,5984,5996,5998,6e3,6016,6067,6103,6103,6108,6108,6176,6263,6272,6312,6314,6314,6320,6389,6400,6428,6480,6509,6512,6516,6528,6571,6593,6599,6656,6678,6688,6740,6823,6823,6917,6963,6981,6987,7043,7072,7086,7087,7098,7141,7168,7203,7245,7247,7258,7293,7401,7404,7406,7409,7413,7414,7424,7615,7680,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8305,8305,8319,8319,8336,8348,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11502,11506,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11648,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,11823,11823,12293,12295,12321,12329,12337,12341,12344,12348,12353,12438,12445,12447,12449,12538,12540,12543,12549,12589,12593,12686,12704,12730,12784,12799,13312,19893,19968,40908,40960,42124,42192,42237,42240,42508,42512,42527,42538,42539,42560,42606,42623,42647,42656,42735,42775,42783,42786,42888,42891,42894,42896,42899,42912,42922,43e3,43009,43011,43013,43015,43018,43020,43042,43072,43123,43138,43187,43250,43255,43259,43259,43274,43301,43312,43334,43360,43388,43396,43442,43471,43471,43520,43560,43584,43586,43588,43595,43616,43638,43642,43642,43648,43695,43697,43697,43701,43702,43705,43709,43712,43712,43714,43714,43739,43741,43744,43754,43762,43764,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43968,44002,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64285,64287,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65136,65140,65142,65276,65313,65338,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500],l=[170,170,181,181,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,768,884,886,887,890,893,902,902,904,906,908,908,910,929,931,1013,1015,1153,1155,1159,1162,1319,1329,1366,1369,1369,1377,1415,1425,1469,1471,1471,1473,1474,1476,1477,1479,1479,1488,1514,1520,1522,1552,1562,1568,1641,1646,1747,1749,1756,1759,1768,1770,1788,1791,1791,1808,1866,1869,1969,1984,2037,2042,2042,2048,2093,2112,2139,2208,2208,2210,2220,2276,2302,2304,2403,2406,2415,2417,2423,2425,2431,2433,2435,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2492,2500,2503,2504,2507,2510,2519,2519,2524,2525,2527,2531,2534,2545,2561,2563,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2620,2620,2622,2626,2631,2632,2635,2637,2641,2641,2649,2652,2654,2654,2662,2677,2689,2691,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2748,2757,2759,2761,2763,2765,2768,2768,2784,2787,2790,2799,2817,2819,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2876,2884,2887,2888,2891,2893,2902,2903,2908,2909,2911,2915,2918,2927,2929,2929,2946,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3006,3010,3014,3016,3018,3021,3024,3024,3031,3031,3046,3055,3073,3075,3077,3084,3086,3088,3090,3112,3114,3123,3125,3129,3133,3140,3142,3144,3146,3149,3157,3158,3160,3161,3168,3171,3174,3183,3202,3203,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3260,3268,3270,3272,3274,3277,3285,3286,3294,3294,3296,3299,3302,3311,3313,3314,3330,3331,3333,3340,3342,3344,3346,3386,3389,3396,3398,3400,3402,3406,3415,3415,3424,3427,3430,3439,3450,3455,3458,3459,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3530,3530,3535,3540,3542,3542,3544,3551,3570,3571,3585,3642,3648,3662,3664,3673,3713,3714,3716,3716,3719,3720,3722,3722,3725,3725,3732,3735,3737,3743,3745,3747,3749,3749,3751,3751,3754,3755,3757,3769,3771,3773,3776,3780,3782,3782,3784,3789,3792,3801,3804,3807,3840,3840,3864,3865,3872,3881,3893,3893,3895,3895,3897,3897,3902,3911,3913,3948,3953,3972,3974,3991,3993,4028,4038,4038,4096,4169,4176,4253,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4957,4959,4992,5007,5024,5108,5121,5740,5743,5759,5761,5786,5792,5866,5870,5872,5888,5900,5902,5908,5920,5940,5952,5971,5984,5996,5998,6e3,6002,6003,6016,6099,6103,6103,6108,6109,6112,6121,6155,6157,6160,6169,6176,6263,6272,6314,6320,6389,6400,6428,6432,6443,6448,6459,6470,6509,6512,6516,6528,6571,6576,6601,6608,6617,6656,6683,6688,6750,6752,6780,6783,6793,6800,6809,6823,6823,6912,6987,6992,7001,7019,7027,7040,7155,7168,7223,7232,7241,7245,7293,7376,7378,7380,7414,7424,7654,7676,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8204,8205,8255,8256,8276,8276,8305,8305,8319,8319,8336,8348,8400,8412,8417,8417,8421,8432,8450,8450,8455,8455,8458,8467,8469,8469,8473,8477,8484,8484,8486,8486,8488,8488,8490,8493,8495,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11647,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,11744,11775,11823,11823,12293,12295,12321,12335,12337,12341,12344,12348,12353,12438,12441,12442,12445,12447,12449,12538,12540,12543,12549,12589,12593,12686,12704,12730,12784,12799,13312,19893,19968,40908,40960,42124,42192,42237,42240,42508,42512,42539,42560,42607,42612,42621,42623,42647,42655,42737,42775,42783,42786,42888,42891,42894,42896,42899,42912,42922,43e3,43047,43072,43123,43136,43204,43216,43225,43232,43255,43259,43259,43264,43309,43312,43347,43360,43388,43392,43456,43471,43481,43520,43574,43584,43597,43600,43609,43616,43638,43642,43643,43648,43714,43739,43741,43744,43759,43762,43766,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43968,44010,44012,44013,44016,44025,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65024,65039,65056,65062,65075,65076,65101,65103,65136,65140,65142,65276,65296,65305,65313,65338,65343,65343,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500],u=[65,90,97,122,170,170,181,181,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,880,884,886,887,890,893,895,895,902,902,904,906,908,908,910,929,931,1013,1015,1153,1162,1327,1329,1366,1369,1369,1376,1416,1488,1514,1519,1522,1568,1610,1646,1647,1649,1747,1749,1749,1765,1766,1774,1775,1786,1788,1791,1791,1808,1808,1810,1839,1869,1957,1969,1969,1994,2026,2036,2037,2042,2042,2048,2069,2074,2074,2084,2084,2088,2088,2112,2136,2144,2154,2208,2228,2230,2237,2308,2361,2365,2365,2384,2384,2392,2401,2417,2432,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2493,2493,2510,2510,2524,2525,2527,2529,2544,2545,2556,2556,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2649,2652,2654,2654,2674,2676,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2749,2749,2768,2768,2784,2785,2809,2809,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2877,2877,2908,2909,2911,2913,2929,2929,2947,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3024,3024,3077,3084,3086,3088,3090,3112,3114,3129,3133,3133,3160,3162,3168,3169,3200,3200,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3261,3261,3294,3294,3296,3297,3313,3314,3333,3340,3342,3344,3346,3386,3389,3389,3406,3406,3412,3414,3423,3425,3450,3455,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3585,3632,3634,3635,3648,3654,3713,3714,3716,3716,3718,3722,3724,3747,3749,3749,3751,3760,3762,3763,3773,3773,3776,3780,3782,3782,3804,3807,3840,3840,3904,3911,3913,3948,3976,3980,4096,4138,4159,4159,4176,4181,4186,4189,4193,4193,4197,4198,4206,4208,4213,4225,4238,4238,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4992,5007,5024,5109,5112,5117,5121,5740,5743,5759,5761,5786,5792,5866,5870,5880,5888,5900,5902,5905,5920,5937,5952,5969,5984,5996,5998,6e3,6016,6067,6103,6103,6108,6108,6176,6264,6272,6312,6314,6314,6320,6389,6400,6430,6480,6509,6512,6516,6528,6571,6576,6601,6656,6678,6688,6740,6823,6823,6917,6963,6981,6987,7043,7072,7086,7087,7098,7141,7168,7203,7245,7247,7258,7293,7296,7304,7312,7354,7357,7359,7401,7404,7406,7411,7413,7414,7418,7418,7424,7615,7680,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8305,8305,8319,8319,8336,8348,8450,8450,8455,8455,8458,8467,8469,8469,8472,8477,8484,8484,8486,8486,8488,8488,8490,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11502,11506,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11648,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,12293,12295,12321,12329,12337,12341,12344,12348,12353,12438,12443,12447,12449,12538,12540,12543,12549,12591,12593,12686,12704,12730,12784,12799,13312,19893,19968,40943,40960,42124,42192,42237,42240,42508,42512,42527,42538,42539,42560,42606,42623,42653,42656,42735,42775,42783,42786,42888,42891,42943,42946,42950,42999,43009,43011,43013,43015,43018,43020,43042,43072,43123,43138,43187,43250,43255,43259,43259,43261,43262,43274,43301,43312,43334,43360,43388,43396,43442,43471,43471,43488,43492,43494,43503,43514,43518,43520,43560,43584,43586,43588,43595,43616,43638,43642,43642,43646,43695,43697,43697,43701,43702,43705,43709,43712,43712,43714,43714,43739,43741,43744,43754,43762,43764,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43824,43866,43868,43879,43888,44002,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64285,64287,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65136,65140,65142,65276,65313,65338,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500,65536,65547,65549,65574,65576,65594,65596,65597,65599,65613,65616,65629,65664,65786,65856,65908,66176,66204,66208,66256,66304,66335,66349,66378,66384,66421,66432,66461,66464,66499,66504,66511,66513,66517,66560,66717,66736,66771,66776,66811,66816,66855,66864,66915,67072,67382,67392,67413,67424,67431,67584,67589,67592,67592,67594,67637,67639,67640,67644,67644,67647,67669,67680,67702,67712,67742,67808,67826,67828,67829,67840,67861,67872,67897,67968,68023,68030,68031,68096,68096,68112,68115,68117,68119,68121,68149,68192,68220,68224,68252,68288,68295,68297,68324,68352,68405,68416,68437,68448,68466,68480,68497,68608,68680,68736,68786,68800,68850,68864,68899,69376,69404,69415,69415,69424,69445,69600,69622,69635,69687,69763,69807,69840,69864,69891,69926,69956,69956,69968,70002,70006,70006,70019,70066,70081,70084,70106,70106,70108,70108,70144,70161,70163,70187,70272,70278,70280,70280,70282,70285,70287,70301,70303,70312,70320,70366,70405,70412,70415,70416,70419,70440,70442,70448,70450,70451,70453,70457,70461,70461,70480,70480,70493,70497,70656,70708,70727,70730,70751,70751,70784,70831,70852,70853,70855,70855,71040,71086,71128,71131,71168,71215,71236,71236,71296,71338,71352,71352,71424,71450,71680,71723,71840,71903,71935,71935,72096,72103,72106,72144,72161,72161,72163,72163,72192,72192,72203,72242,72250,72250,72272,72272,72284,72329,72349,72349,72384,72440,72704,72712,72714,72750,72768,72768,72818,72847,72960,72966,72968,72969,72971,73008,73030,73030,73056,73061,73063,73064,73066,73097,73112,73112,73440,73458,73728,74649,74752,74862,74880,75075,77824,78894,82944,83526,92160,92728,92736,92766,92880,92909,92928,92975,92992,92995,93027,93047,93053,93071,93760,93823,93952,94026,94032,94032,94099,94111,94176,94177,94179,94179,94208,100343,100352,101106,110592,110878,110928,110930,110948,110951,110960,111355,113664,113770,113776,113788,113792,113800,113808,113817,119808,119892,119894,119964,119966,119967,119970,119970,119973,119974,119977,119980,119982,119993,119995,119995,119997,120003,120005,120069,120071,120074,120077,120084,120086,120092,120094,120121,120123,120126,120128,120132,120134,120134,120138,120144,120146,120485,120488,120512,120514,120538,120540,120570,120572,120596,120598,120628,120630,120654,120656,120686,120688,120712,120714,120744,120746,120770,120772,120779,123136,123180,123191,123197,123214,123214,123584,123627,124928,125124,125184,125251,125259,125259,126464,126467,126469,126495,126497,126498,126500,126500,126503,126503,126505,126514,126516,126519,126521,126521,126523,126523,126530,126530,126535,126535,126537,126537,126539,126539,126541,126543,126545,126546,126548,126548,126551,126551,126553,126553,126555,126555,126557,126557,126559,126559,126561,126562,126564,126564,126567,126570,126572,126578,126580,126583,126585,126588,126590,126590,126592,126601,126603,126619,126625,126627,126629,126633,126635,126651,131072,173782,173824,177972,177984,178205,178208,183969,183984,191456,194560,195101],d=[48,57,65,90,95,95,97,122,170,170,181,181,183,183,186,186,192,214,216,246,248,705,710,721,736,740,748,748,750,750,768,884,886,887,890,893,895,895,902,906,908,908,910,929,931,1013,1015,1153,1155,1159,1162,1327,1329,1366,1369,1369,1376,1416,1425,1469,1471,1471,1473,1474,1476,1477,1479,1479,1488,1514,1519,1522,1552,1562,1568,1641,1646,1747,1749,1756,1759,1768,1770,1788,1791,1791,1808,1866,1869,1969,1984,2037,2042,2042,2045,2045,2048,2093,2112,2139,2144,2154,2208,2228,2230,2237,2259,2273,2275,2403,2406,2415,2417,2435,2437,2444,2447,2448,2451,2472,2474,2480,2482,2482,2486,2489,2492,2500,2503,2504,2507,2510,2519,2519,2524,2525,2527,2531,2534,2545,2556,2556,2558,2558,2561,2563,2565,2570,2575,2576,2579,2600,2602,2608,2610,2611,2613,2614,2616,2617,2620,2620,2622,2626,2631,2632,2635,2637,2641,2641,2649,2652,2654,2654,2662,2677,2689,2691,2693,2701,2703,2705,2707,2728,2730,2736,2738,2739,2741,2745,2748,2757,2759,2761,2763,2765,2768,2768,2784,2787,2790,2799,2809,2815,2817,2819,2821,2828,2831,2832,2835,2856,2858,2864,2866,2867,2869,2873,2876,2884,2887,2888,2891,2893,2902,2903,2908,2909,2911,2915,2918,2927,2929,2929,2946,2947,2949,2954,2958,2960,2962,2965,2969,2970,2972,2972,2974,2975,2979,2980,2984,2986,2990,3001,3006,3010,3014,3016,3018,3021,3024,3024,3031,3031,3046,3055,3072,3084,3086,3088,3090,3112,3114,3129,3133,3140,3142,3144,3146,3149,3157,3158,3160,3162,3168,3171,3174,3183,3200,3203,3205,3212,3214,3216,3218,3240,3242,3251,3253,3257,3260,3268,3270,3272,3274,3277,3285,3286,3294,3294,3296,3299,3302,3311,3313,3314,3328,3331,3333,3340,3342,3344,3346,3396,3398,3400,3402,3406,3412,3415,3423,3427,3430,3439,3450,3455,3458,3459,3461,3478,3482,3505,3507,3515,3517,3517,3520,3526,3530,3530,3535,3540,3542,3542,3544,3551,3558,3567,3570,3571,3585,3642,3648,3662,3664,3673,3713,3714,3716,3716,3718,3722,3724,3747,3749,3749,3751,3773,3776,3780,3782,3782,3784,3789,3792,3801,3804,3807,3840,3840,3864,3865,3872,3881,3893,3893,3895,3895,3897,3897,3902,3911,3913,3948,3953,3972,3974,3991,3993,4028,4038,4038,4096,4169,4176,4253,4256,4293,4295,4295,4301,4301,4304,4346,4348,4680,4682,4685,4688,4694,4696,4696,4698,4701,4704,4744,4746,4749,4752,4784,4786,4789,4792,4798,4800,4800,4802,4805,4808,4822,4824,4880,4882,4885,4888,4954,4957,4959,4969,4977,4992,5007,5024,5109,5112,5117,5121,5740,5743,5759,5761,5786,5792,5866,5870,5880,5888,5900,5902,5908,5920,5940,5952,5971,5984,5996,5998,6e3,6002,6003,6016,6099,6103,6103,6108,6109,6112,6121,6155,6157,6160,6169,6176,6264,6272,6314,6320,6389,6400,6430,6432,6443,6448,6459,6470,6509,6512,6516,6528,6571,6576,6601,6608,6618,6656,6683,6688,6750,6752,6780,6783,6793,6800,6809,6823,6823,6832,6845,6912,6987,6992,7001,7019,7027,7040,7155,7168,7223,7232,7241,7245,7293,7296,7304,7312,7354,7357,7359,7376,7378,7380,7418,7424,7673,7675,7957,7960,7965,7968,8005,8008,8013,8016,8023,8025,8025,8027,8027,8029,8029,8031,8061,8064,8116,8118,8124,8126,8126,8130,8132,8134,8140,8144,8147,8150,8155,8160,8172,8178,8180,8182,8188,8255,8256,8276,8276,8305,8305,8319,8319,8336,8348,8400,8412,8417,8417,8421,8432,8450,8450,8455,8455,8458,8467,8469,8469,8472,8477,8484,8484,8486,8486,8488,8488,8490,8505,8508,8511,8517,8521,8526,8526,8544,8584,11264,11310,11312,11358,11360,11492,11499,11507,11520,11557,11559,11559,11565,11565,11568,11623,11631,11631,11647,11670,11680,11686,11688,11694,11696,11702,11704,11710,11712,11718,11720,11726,11728,11734,11736,11742,11744,11775,12293,12295,12321,12335,12337,12341,12344,12348,12353,12438,12441,12447,12449,12538,12540,12543,12549,12591,12593,12686,12704,12730,12784,12799,13312,19893,19968,40943,40960,42124,42192,42237,42240,42508,42512,42539,42560,42607,42612,42621,42623,42737,42775,42783,42786,42888,42891,42943,42946,42950,42999,43047,43072,43123,43136,43205,43216,43225,43232,43255,43259,43259,43261,43309,43312,43347,43360,43388,43392,43456,43471,43481,43488,43518,43520,43574,43584,43597,43600,43609,43616,43638,43642,43714,43739,43741,43744,43759,43762,43766,43777,43782,43785,43790,43793,43798,43808,43814,43816,43822,43824,43866,43868,43879,43888,44010,44012,44013,44016,44025,44032,55203,55216,55238,55243,55291,63744,64109,64112,64217,64256,64262,64275,64279,64285,64296,64298,64310,64312,64316,64318,64318,64320,64321,64323,64324,64326,64433,64467,64829,64848,64911,64914,64967,65008,65019,65024,65039,65056,65071,65075,65076,65101,65103,65136,65140,65142,65276,65296,65305,65313,65338,65343,65343,65345,65370,65382,65470,65474,65479,65482,65487,65490,65495,65498,65500,65536,65547,65549,65574,65576,65594,65596,65597,65599,65613,65616,65629,65664,65786,65856,65908,66045,66045,66176,66204,66208,66256,66272,66272,66304,66335,66349,66378,66384,66426,66432,66461,66464,66499,66504,66511,66513,66517,66560,66717,66720,66729,66736,66771,66776,66811,66816,66855,66864,66915,67072,67382,67392,67413,67424,67431,67584,67589,67592,67592,67594,67637,67639,67640,67644,67644,67647,67669,67680,67702,67712,67742,67808,67826,67828,67829,67840,67861,67872,67897,67968,68023,68030,68031,68096,68099,68101,68102,68108,68115,68117,68119,68121,68149,68152,68154,68159,68159,68192,68220,68224,68252,68288,68295,68297,68326,68352,68405,68416,68437,68448,68466,68480,68497,68608,68680,68736,68786,68800,68850,68864,68903,68912,68921,69376,69404,69415,69415,69424,69456,69600,69622,69632,69702,69734,69743,69759,69818,69840,69864,69872,69881,69888,69940,69942,69951,69956,69958,69968,70003,70006,70006,70016,70084,70089,70092,70096,70106,70108,70108,70144,70161,70163,70199,70206,70206,70272,70278,70280,70280,70282,70285,70287,70301,70303,70312,70320,70378,70384,70393,70400,70403,70405,70412,70415,70416,70419,70440,70442,70448,70450,70451,70453,70457,70459,70468,70471,70472,70475,70477,70480,70480,70487,70487,70493,70499,70502,70508,70512,70516,70656,70730,70736,70745,70750,70751,70784,70853,70855,70855,70864,70873,71040,71093,71096,71104,71128,71133,71168,71232,71236,71236,71248,71257,71296,71352,71360,71369,71424,71450,71453,71467,71472,71481,71680,71738,71840,71913,71935,71935,72096,72103,72106,72151,72154,72161,72163,72164,72192,72254,72263,72263,72272,72345,72349,72349,72384,72440,72704,72712,72714,72758,72760,72768,72784,72793,72818,72847,72850,72871,72873,72886,72960,72966,72968,72969,72971,73014,73018,73018,73020,73021,73023,73031,73040,73049,73056,73061,73063,73064,73066,73102,73104,73105,73107,73112,73120,73129,73440,73462,73728,74649,74752,74862,74880,75075,77824,78894,82944,83526,92160,92728,92736,92766,92768,92777,92880,92909,92912,92916,92928,92982,92992,92995,93008,93017,93027,93047,93053,93071,93760,93823,93952,94026,94031,94087,94095,94111,94176,94177,94179,94179,94208,100343,100352,101106,110592,110878,110928,110930,110948,110951,110960,111355,113664,113770,113776,113788,113792,113800,113808,113817,113821,113822,119141,119145,119149,119154,119163,119170,119173,119179,119210,119213,119362,119364,119808,119892,119894,119964,119966,119967,119970,119970,119973,119974,119977,119980,119982,119993,119995,119995,119997,120003,120005,120069,120071,120074,120077,120084,120086,120092,120094,120121,120123,120126,120128,120132,120134,120134,120138,120144,120146,120485,120488,120512,120514,120538,120540,120570,120572,120596,120598,120628,120630,120654,120656,120686,120688,120712,120714,120744,120746,120770,120772,120779,120782,120831,121344,121398,121403,121452,121461,121461,121476,121476,121499,121503,121505,121519,122880,122886,122888,122904,122907,122913,122915,122916,122918,122922,123136,123180,123184,123197,123200,123209,123214,123214,123584,123641,124928,125124,125136,125142,125184,125259,125264,125273,126464,126467,126469,126495,126497,126498,126500,126500,126503,126503,126505,126514,126516,126519,126521,126521,126523,126523,126530,126530,126535,126535,126537,126537,126539,126539,126541,126543,126545,126546,126548,126548,126551,126551,126553,126553,126555,126555,126557,126557,126559,126559,126561,126562,126564,126564,126567,126570,126572,126578,126580,126583,126585,126588,126590,126590,126592,126601,126603,126619,126625,126627,126629,126633,126635,126651,131072,173782,173824,177972,177984,178205,178208,183969,183984,191456,194560,195101,917760,917999],p=/^\s*\/\/\/?\s*@(ts-expect-error|ts-ignore)/,m=/^\s*(?:\/|\*)*\s*@(ts-expect-error|ts-ignore)/;function f(e,n){if(e<n[0])return!1;for(var t,r=0,a=n.length;r+1<a;){if(t=r+(a-r)/2,n[t-=t%2]<=e&&e<=n[t+1])return!0;e<n[t]?a=t:r=t+2}return!1}function _(e,n){return f(e,n>=2?u:1===n?c:o)}e.isUnicodeIdentifierStart=_;var g,y=(g=[],i.forEach((function(e,n){g[e]=n})),g);function v(e){for(var n=new Array,t=0,r=0;t<e.length;){var a=e.charCodeAt(t);switch(t++,a){case 13:10===e.charCodeAt(t)&&t++;case 10:n.push(r),r=t;break;default:a>127&&L(a)&&(n.push(r),r=t)}}return n.push(r),n}function h(n,t,r,a,i){(t<0||t>=n.length)&&(i?t=t<0?0:t>=n.length?n.length-1:t:e.Debug.fail("Bad line number. Line: "+t+", lineStarts.length: "+n.length+" , line map is correct? "+(void 0!==a?e.arraysEqual(n,v(a)):"unknown")));var o=n[t]+r;return i?o>n[t+1]?n[t+1]:"string"==typeof a&&o>a.length?a.length:o:(t<n.length-1?e.Debug.assert(o<n[t+1]):void 0!==a&&e.Debug.assert(o<=a.length),o)}function b(e){return e.lineMap||(e.lineMap=v(e.text))}function E(e,n){var t=T(e,n);return{line:t,character:n-e[t]}}function T(n,t,r){var a=e.binarySearch(n,t,e.identity,e.compareValues,r);return a<0&&(a=~a-1,e.Debug.assert(-1!==a,"position cannot precede the beginning of the file")),a}function S(e){return x(e)||L(e)}function x(e){return 32===e||9===e||11===e||12===e||160===e||133===e||5760===e||e>=8192&&e<=8203||8239===e||8287===e||12288===e||65279===e}function L(e){return 10===e||13===e||8232===e||8233===e}function A(e){return e>=48&&e<=57}function D(e){return A(e)||e>=65&&e<=70||e>=97&&e<=102}function C(e){return e>=48&&e<=55}e.tokenToString=function(e){return y[e]},e.stringToToken=function(e){return i.get(e)},e.computeLineStarts=v,e.getPositionOfLineAndCharacter=function(e,n,t,r){return e.getPositionOfLineAndCharacter?e.getPositionOfLineAndCharacter(n,t,r):h(b(e),n,t,e.text,r)},e.computePositionOfLineAndCharacter=h,e.getLineStarts=b,e.computeLineAndCharacterOfPosition=E,e.computeLineOfPosition=T,e.getLinesBetweenPositions=function(e,n,t){if(n===t)return 0;var r=b(e),a=Math.min(n,t),i=a===t,o=i?n:t,s=T(r,a),c=T(r,o,s);return i?s-c:c-s},e.getLineAndCharacterOfPosition=function(e,n){return E(b(e),n)},e.isWhiteSpaceLike=S,e.isWhiteSpaceSingleLine=x,e.isLineBreak=L,e.isOctalDigit=C,e.couldStartTrivia=function(e,n){var t=e.charCodeAt(n);switch(t){case 13:case 10:case 9:case 11:case 12:case 32:case 47:case 60:case 124:case 61:case 62:return!0;case 35:return 0===n;default:return t>127}},e.skipTrivia=function(n,t,r,a){if(void 0===a&&(a=!1),e.positionIsSynthesized(t))return t;for(;;){var i=n.charCodeAt(t);switch(i){case 13:10===n.charCodeAt(t+1)&&t++;case 10:if(t++,r)return t;continue;case 9:case 11:case 12:case 32:t++;continue;case 47:if(a)break;if(47===n.charCodeAt(t+1)){for(t+=2;t<n.length&&!L(n.charCodeAt(t));)t++;continue}if(42===n.charCodeAt(t+1)){for(t+=2;t<n.length;){if(42===n.charCodeAt(t)&&47===n.charCodeAt(t+1)){t+=2;break}t++}continue}break;case 60:case 124:case 61:case 62:if(N(n,t)){t=R(n,t);continue}break;case 35:if(0===t&&O(n,t)){t=P(n,t);continue}break;default:if(i>127&&S(i)){t++;continue}}return t}};var I="<<<<<<<".length;function N(n,t){if(e.Debug.assert(t>=0),0===t||L(n.charCodeAt(t-1))){var r=n.charCodeAt(t);if(t+I<n.length){for(var a=0;a<I;a++)if(n.charCodeAt(t+a)!==r)return!1;return 61===r||32===n.charCodeAt(t+I)}}return!1}function R(n,t,r){r&&r(e.Diagnostics.Merge_conflict_marker_encountered,t,I);var a=n.charCodeAt(t),i=n.length;if(60===a||62===a)for(;t<i&&!L(n.charCodeAt(t));)t++;else for(e.Debug.assert(124===a||61===a);t<i;){var o=n.charCodeAt(t);if((61===o||62===o)&&o!==a&&N(n,t))break;t++}return t}var k=/^#!.*/;function O(n,t){return e.Debug.assert(0===t),k.test(n)}function P(e,n){return n+=k.exec(e)[0].length}function M(e,n,t,r,a,i,o){var s,c,l,u,d=!1,p=r,m=o;if(0===t){p=!0;var f=B(n);f&&(t=f.length)}e:for(;t>=0&&t<n.length;){var _=n.charCodeAt(t);switch(_){case 13:10===n.charCodeAt(t+1)&&t++;case 10:if(t++,r)break e;p=!0,d&&(u=!0);continue;case 9:case 11:case 12:case 32:t++;continue;case 47:var g=n.charCodeAt(t+1),y=!1;if(47===g||42===g){var v=47===g?2:3,h=t;if(t+=2,47===g)for(;t<n.length;){if(L(n.charCodeAt(t))){y=!0;break}t++}else for(;t<n.length;){if(42===n.charCodeAt(t)&&47===n.charCodeAt(t+1)){t+=2;break}t++}if(p){if(d&&(m=a(s,c,l,u,i,m),!e&&m))return m;s=h,c=t,l=v,u=y,d=!0}continue}break e;default:if(_>127&&S(_)){d&&L(_)&&(u=!0),t++;continue}break e}}return d&&(m=a(s,c,l,u,i,m)),m}function w(e,n,t,r,a){return M(!0,e,n,!1,t,r,a)}function F(e,n,t,r,a){return M(!0,e,n,!0,t,r,a)}function G(e,n,t,r,a,i){return i||(i=[]),i.push({kind:t,pos:e,end:n,hasTrailingNewLine:r}),i}function B(e){var n=k.exec(e);if(n)return n[0]}function U(e,n){return e>=65&&e<=90||e>=97&&e<=122||36===e||95===e||e>127&&_(e,n)}function V(e,n,t){return e>=65&&e<=90||e>=97&&e<=122||e>=48&&e<=57||36===e||95===e||1===t&&(45===e||58===e)||e>127&&function(e,n){return f(e,n>=2?d:1===n?l:s)}(e,n)}e.isShebangTrivia=O,e.scanShebangTrivia=P,e.forEachLeadingCommentRange=function(e,n,t,r){return M(!1,e,n,!1,t,r)},e.forEachTrailingCommentRange=function(e,n,t,r){return M(!1,e,n,!0,t,r)},e.reduceEachLeadingCommentRange=w,e.reduceEachTrailingCommentRange=F,e.getLeadingCommentRanges=function(e,n){return w(e,n,G,void 0,void 0)},e.getTrailingCommentRanges=function(e,n){return F(e,n,G,void 0,void 0)},e.getShebang=B,e.isIdentifierStart=U,e.isIdentifierPart=V,e.isIdentifierText=function(e,n,t){var r=j(e,0);if(!U(r,n))return!1;for(var a=H(r);a<e.length;a+=H(r))if(!V(r=j(e,a),n,t))return!1;return!0},e.createScanner=function(n,r,i,o,s,c,l){void 0===i&&(i=0);var u,d,f,_,g,y,v,h,b=o,E=0;ce(b,c,l);var T={getStartPos:function(){return f},getTextPos:function(){return u},getToken:function(){return g},getTokenPos:function(){return _},getTokenText:function(){return b.substring(_,u)},getTokenValue:function(){return y},hasUnicodeEscape:function(){return 0!=(1024&v)},hasExtendedUnicodeEscape:function(){return 0!=(8&v)},hasPrecedingLineBreak:function(){return 0!=(1&v)},hasPrecedingJSDocComment:function(){return 0!=(2&v)},isIdentifier:function(){return 78===g||g>115},isReservedWord:function(){return g>=80&&g<=115},isUnterminated:function(){return 0!=(4&v)},getCommentDirectives:function(){return h},getNumericLiteralFlags:function(){return 1008&v},getTokenFlags:function(){return v},reScanGreaterToken:function(){if(31===g){if(62===b.charCodeAt(u))return 62===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=71):(u+=2,g=49):61===b.charCodeAt(u+1)?(u+=2,g=70):(u++,g=48);if(61===b.charCodeAt(u))return u++,g=33}return g},reScanAsteriskEqualsToken:function(){return e.Debug.assert(65===g,"'reScanAsteriskEqualsToken' should only be called on a '*='"),u=_+1,g=62},reScanSlashToken:function(){if(43===g||67===g){for(var t=_+1,r=!1,a=!1;;){if(t>=d){v|=4,I(e.Diagnostics.Unterminated_regular_expression_literal);break}var i=b.charCodeAt(t);if(L(i)){v|=4,I(e.Diagnostics.Unterminated_regular_expression_literal);break}if(r)r=!1;else{if(47===i&&!a){t++;break}91===i?a=!0:92===i?r=!0:93===i&&(a=!1)}t++}for(;t<d&&V(b.charCodeAt(t),n);)t++;u=t,y=b.substring(_,u),g=13}return g},reScanTemplateToken:function(n){return e.Debug.assert(19===g,"'reScanTemplateToken' should only be called on a '}'"),u=_,g=q(n)},reScanTemplateHeadOrNoSubstitutionTemplate:function(){return u=_,g=q(!0)},scanJsxIdentifier:function(){if(t(g))for(;u<d;){if(45!==b.charCodeAt(u)){var e=u;if(y+=$(),u===e)break}else y+="-",u++}return g},scanJsxAttributeValue:oe,reScanJsxAttributeValue:function(){return u=_=f,oe()},reScanJsxToken:function(){return u=_=f,g=ie()},reScanLessThanToken:function(){if(47===g)return u=_+1,g=29;return g},reScanQuestionToken:function(){return e.Debug.assert(60===g,"'reScanQuestionToken' should only be called on a '??'"),u=_+1,g=57},scanJsxToken:ie,scanJsDocToken:function(){if(f=_=u,v=0,u>=d)return g=1;var e=j(b,u);switch(u+=H(e),e){case 9:case 11:case 12:case 32:for(;u<d&&x(b.charCodeAt(u));)u++;return g=5;case 64:return g=59;case 13:10===b.charCodeAt(u)&&u++;case 10:return v|=1,g=4;case 42:return g=41;case 123:return g=18;case 125:return g=19;case 91:return g=22;case 93:return g=23;case 60:return g=29;case 62:return g=31;case 61:return g=62;case 44:return g=27;case 46:return g=24;case 96:return g=61;case 92:u--;var t=Z();if(t>=0&&U(t,n))return u+=3,v|=8,y=Y()+$(),g=ee();var r=Q();return r>=0&&U(r,n)?(u+=6,v|=1024,y=String.fromCharCode(r)+$(),g=ee()):(u++,g=0)}if(U(e,n)){for(var a=e;u<d&&V(a=j(b,u),n)||45===b.charCodeAt(u);)u+=H(a);return y=b.substring(_,u),92===a&&(y+=$()),g=ee()}return g=0},scan:re,getText:function(){return b},clearCommentDirectives:function(){h=void 0},setText:ce,setScriptTarget:function(e){n=e},setLanguageVariant:function(e){i=e},setOnError:function(e){s=e},setTextPos:le,setInJSDocType:function(e){E+=e?1:-1},tryScan:function(e){return se(e,!1)},lookAhead:function(e){return se(e,!0)},scanRange:function(e,n,t){var r=d,a=u,i=f,o=_,s=g,c=y,l=v,p=h;ce(b,e,n);var m=t();return d=r,u=a,f=i,_=o,g=s,y=c,v=l,h=p,m}};return e.Debug.isDebugging&&Object.defineProperty(T,"__debugShowCurrentPositionInText",{get:function(){var e=T.getText();return e.slice(0,T.getStartPos())+"║"+e.slice(T.getStartPos())}}),T;function I(e,n,t){if(void 0===n&&(n=u),s){var r=u;u=n,s(e,t||0),u=r}}function k(){for(var n=u,t=!1,r=!1,a="";;){var i=b.charCodeAt(u);if(95!==i){if(!A(i))break;t=!0,r=!1,u++}else v|=512,t?(t=!1,r=!0,a+=b.substring(n,u)):I(r?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),n=++u}return 95===b.charCodeAt(u-1)&&I(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),a+b.substring(n,u)}function M(){var n,t,r=u,a=k();46===b.charCodeAt(u)&&(u++,n=k());var i,o=u;if(69===b.charCodeAt(u)||101===b.charCodeAt(u)){u++,v|=16,43!==b.charCodeAt(u)&&45!==b.charCodeAt(u)||u++;var s=u,c=k();c?(t=b.substring(o,s)+c,o=u):I(e.Diagnostics.Digit_expected)}if(512&v?(i=a,n&&(i+="."+n),t&&(i+=t)):i=b.substring(r,o),void 0!==n||16&v)return w(r,void 0===n&&!!(16&v)),{type:8,value:""+ +i};y=i;var l=te();return w(r),{type:l,value:y}}function w(t,r){if(U(j(b,u),n)){var a=u,i=$().length;1===i&&"n"===b[a]?I(r?e.Diagnostics.A_bigint_literal_cannot_use_exponential_notation:e.Diagnostics.A_bigint_literal_must_be_an_integer,t,a-t+1):(I(e.Diagnostics.An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal,a,i),u=a)}}function F(){for(var e=u;C(b.charCodeAt(u));)u++;return+b.substring(e,u)}function G(e,n){var t=K(e,!1,n);return t?parseInt(t,16):-1}function B(e,n){return K(e,!0,n)}function K(n,t,r){for(var a=[],i=!1,o=!1;a.length<n||t;){var s=b.charCodeAt(u);if(r&&95===s)v|=512,i?(i=!1,o=!0):I(o?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),u++;else{if(i=r,s>=65&&s<=70)s+=32;else if(!(s>=48&&s<=57||s>=97&&s<=102))break;a.push(s),u++,o=!1}}return a.length<n&&(a=[]),95===b.charCodeAt(u-1)&&I(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),String.fromCharCode.apply(String,a)}function z(n){void 0===n&&(n=!1);for(var t=b.charCodeAt(u),r="",a=++u;;){if(u>=d){r+=b.substring(a,u),v|=4,I(e.Diagnostics.Unterminated_string_literal);break}var i=b.charCodeAt(u);if(i===t){r+=b.substring(a,u),u++;break}if(92!==i||n){if(L(i)&&!n){r+=b.substring(a,u),v|=4,I(e.Diagnostics.Unterminated_string_literal);break}u++}else r+=b.substring(a,u),r+=J(),a=u}return r}function q(n){for(var t,r=96===b.charCodeAt(u),a=++u,i="";;){if(u>=d){i+=b.substring(a,u),v|=4,I(e.Diagnostics.Unterminated_template_literal),t=r?14:17;break}var o=b.charCodeAt(u);if(96===o){i+=b.substring(a,u),u++,t=r?14:17;break}if(36===o&&u+1<d&&123===b.charCodeAt(u+1)){i+=b.substring(a,u),u+=2,t=r?15:16;break}92!==o?13!==o?u++:(i+=b.substring(a,u),++u<d&&10===b.charCodeAt(u)&&u++,i+="\n",a=u):(i+=b.substring(a,u),i+=J(n),a=u)}return e.Debug.assert(void 0!==t),y=i,t}function J(n){var t=u;if(++u>=d)return I(e.Diagnostics.Unexpected_end_of_text),"";var r=b.charCodeAt(u);switch(u++,r){case 48:return n&&u<d&&A(b.charCodeAt(u))?(u++,v|=2048,b.substring(t,u)):"\0";case 98:return"\b";case 116:return"\t";case 110:return"\n";case 118:return"\v";case 102:return"\f";case 114:return"\r";case 39:return"'";case 34:return'"';case 117:if(n)for(var a=u;a<u+4;a++)if(a<d&&!D(b.charCodeAt(a))&&123!==b.charCodeAt(a))return u=a,v|=2048,b.substring(t,u);if(u<d&&123===b.charCodeAt(u)){if(u++,n&&!D(b.charCodeAt(u)))return v|=2048,b.substring(t,u);if(n){var i=u,o=B(1,!1),s=o?parseInt(o,16):-1;if(!(s<=1114111&&125===b.charCodeAt(u)))return v|=2048,b.substring(t,u);u=i}return v|=8,Y()}return v|=1024,X(4);case 120:if(n){if(!D(b.charCodeAt(u)))return v|=2048,b.substring(t,u);if(!D(b.charCodeAt(u+1)))return u++,v|=2048,b.substring(t,u)}return X(2);case 13:u<d&&10===b.charCodeAt(u)&&u++;case 10:case 8232:case 8233:return"";default:return String.fromCharCode(r)}}function X(n){var t=G(n,!1);return t>=0?String.fromCharCode(t):(I(e.Diagnostics.Hexadecimal_digit_expected),"")}function Y(){var n=B(1,!1),t=n?parseInt(n,16):-1,r=!1;return t<0?(I(e.Diagnostics.Hexadecimal_digit_expected),r=!0):t>1114111&&(I(e.Diagnostics.An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive),r=!0),u>=d?(I(e.Diagnostics.Unexpected_end_of_text),r=!0):125===b.charCodeAt(u)?u++:(I(e.Diagnostics.Unterminated_Unicode_escape_sequence),r=!0),r?"":W(t)}function Q(){if(u+5<d&&117===b.charCodeAt(u+1)){var e=u;u+=2;var n=G(4,!1);return u=e,n}return-1}function Z(){if(n>=2&&117===j(b,u+1)&&123===j(b,u+2)){var e=u;u+=3;var t=B(1,!1),r=t?parseInt(t,16):-1;return u=e,r}return-1}function $(){for(var e="",t=u;u<d;){var r=j(b,u);if(V(r,n))u+=H(r);else{if(92!==r)break;if((r=Z())>=0&&V(r,n)){u+=3,v|=8,e+=Y(),t=u;continue}if(!((r=Q())>=0&&V(r,n)))break;v|=1024,e+=b.substring(t,u),e+=W(r),t=u+=6}}return e+=b.substring(t,u)}function ee(){var e=y.length;if(e>=2&&e<=12){var n=y.charCodeAt(0);if(n>=97&&n<=122){var t=a.get(y);if(void 0!==t)return g=t}}return g=78}function ne(n){for(var t="",r=!1,a=!1;;){var i=b.charCodeAt(u);if(95!==i){if(r=!0,!A(i)||i-48>=n)break;t+=b[u],u++,a=!1}else v|=512,r?(r=!1,a=!0):I(a?e.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted:e.Diagnostics.Numeric_separators_are_not_allowed_here,u,1),u++}return 95===b.charCodeAt(u-1)&&I(e.Diagnostics.Numeric_separators_are_not_allowed_here,u-1,1),t}function te(){if(110===b.charCodeAt(u))return y+="n",384&v&&(y=e.parsePseudoBigInt(y)+"n"),u++,9;var n=128&v?parseInt(y.slice(2),2):256&v?parseInt(y.slice(2),8):+y;return y=""+n,8}function re(){var t;f=u,v=0;for(var a=!1;;){if(_=u,u>=d)return g=1;var o=j(b,u);if(35===o&&0===u&&O(b,u)){if(u=P(b,u),r)continue;return g=6}switch(o){case 10:case 13:if(v|=1,r){u++;continue}return 13===o&&u+1<d&&10===b.charCodeAt(u+1)?u+=2:u++,g=4;case 9:case 11:case 12:case 32:case 160:case 5760:case 8192:case 8193:case 8194:case 8195:case 8196:case 8197:case 8198:case 8199:case 8200:case 8201:case 8202:case 8203:case 8239:case 8287:case 12288:case 65279:if(r){u++;continue}for(;u<d&&x(b.charCodeAt(u));)u++;return g=5;case 33:return 61===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=37):(u+=2,g=35):(u++,g=53);case 34:case 39:return y=z(),g=10;case 96:return g=q(!1);case 37:return 61===b.charCodeAt(u+1)?(u+=2,g=68):(u++,g=44);case 38:return 38===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=75):(u+=2,g=55):61===b.charCodeAt(u+1)?(u+=2,g=72):(u++,g=50);case 40:return u++,g=20;case 41:return u++,g=21;case 42:if(61===b.charCodeAt(u+1))return u+=2,g=65;if(42===b.charCodeAt(u+1))return 61===b.charCodeAt(u+2)?(u+=3,g=66):(u+=2,g=42);if(u++,E&&!a&&1&v){a=!0;continue}return g=41;case 43:return 43===b.charCodeAt(u+1)?(u+=2,g=45):61===b.charCodeAt(u+1)?(u+=2,g=63):(u++,g=39);case 44:return u++,g=27;case 45:return 45===b.charCodeAt(u+1)?(u+=2,g=46):61===b.charCodeAt(u+1)?(u+=2,g=64):(u++,g=40);case 46:return A(b.charCodeAt(u+1))?(y=M().value,g=8):46===b.charCodeAt(u+1)&&46===b.charCodeAt(u+2)?(u+=3,g=25):(u++,g=24);case 47:if(47===b.charCodeAt(u+1)){for(u+=2;u<d&&!L(b.charCodeAt(u));)u++;if(h=ae(h,b.slice(_,u),p,_),r)continue;return g=2}if(42===b.charCodeAt(u+1)){u+=2,42===b.charCodeAt(u)&&47!==b.charCodeAt(u+1)&&(v|=2);for(var s=!1,c=_;u<d;){var l=b.charCodeAt(u);if(42===l&&47===b.charCodeAt(u+1)){u+=2,s=!0;break}u++,L(l)&&(c=u,v|=1)}if(h=ae(h,b.slice(c,u),m,c),s||I(e.Diagnostics.Asterisk_Slash_expected),r)continue;return s||(v|=4),g=3}return 61===b.charCodeAt(u+1)?(u+=2,g=67):(u++,g=43);case 48:if(u+2<d&&(88===b.charCodeAt(u+1)||120===b.charCodeAt(u+1)))return u+=2,(y=B(1,!0))||(I(e.Diagnostics.Hexadecimal_digit_expected),y="0"),y="0x"+y,v|=64,g=te();if(u+2<d&&(66===b.charCodeAt(u+1)||98===b.charCodeAt(u+1)))return u+=2,(y=ne(2))||(I(e.Diagnostics.Binary_digit_expected),y="0"),y="0b"+y,v|=128,g=te();if(u+2<d&&(79===b.charCodeAt(u+1)||111===b.charCodeAt(u+1)))return u+=2,(y=ne(8))||(I(e.Diagnostics.Octal_digit_expected),y="0"),y="0o"+y,v|=256,g=te();if(u+1<d&&C(b.charCodeAt(u+1)))return y=""+F(),v|=32,g=8;case 49:case 50:case 51:case 52:case 53:case 54:case 55:case 56:case 57:return t=M(),g=t.type,y=t.value,g;case 58:return u++,g=58;case 59:return u++,g=26;case 60:if(N(b,u)){if(u=R(b,u,I),r)continue;return g=7}return 60===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=69):(u+=2,g=47):61===b.charCodeAt(u+1)?(u+=2,g=32):1===i&&47===b.charCodeAt(u+1)&&42!==b.charCodeAt(u+2)?(u+=2,g=30):(u++,g=29);case 61:if(N(b,u)){if(u=R(b,u,I),r)continue;return g=7}return 61===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=36):(u+=2,g=34):62===b.charCodeAt(u+1)?(u+=2,g=38):(u++,g=62);case 62:if(N(b,u)){if(u=R(b,u,I),r)continue;return g=7}return u++,g=31;case 63:return 46!==b.charCodeAt(u+1)||A(b.charCodeAt(u+2))?63===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=76):(u+=2,g=60):(u++,g=57):(u+=2,g=28);case 91:return u++,g=22;case 93:return u++,g=23;case 94:return 61===b.charCodeAt(u+1)?(u+=2,g=77):(u++,g=52);case 123:return u++,g=18;case 124:if(N(b,u)){if(u=R(b,u,I),r)continue;return g=7}return 124===b.charCodeAt(u+1)?61===b.charCodeAt(u+2)?(u+=3,g=74):(u+=2,g=56):61===b.charCodeAt(u+1)?(u+=2,g=73):(u++,g=51);case 125:return u++,g=19;case 126:return u++,g=54;case 64:return u++,g=59;case 92:var T=Z();if(T>=0&&U(T,n))return u+=3,v|=8,y=Y()+$(),g=ee();var S=Q();return S>=0&&U(S,n)?(u+=6,v|=1024,y=String.fromCharCode(S)+$(),g=ee()):(I(e.Diagnostics.Invalid_character),u++,g=0);case 35:if(0!==u&&"!"===b[u+1])return I(e.Diagnostics.can_only_be_used_at_the_start_of_a_file),u++,g=0;if(u++,U(o=b.charCodeAt(u),n)){for(u++;u<d&&V(o=b.charCodeAt(u),n);)u++;y=b.substring(_,u),92===o&&(y+=$())}else y="#",I(e.Diagnostics.Invalid_character);return g=79;default:if(U(o,n)){for(u+=H(o);u<d&&V(o=j(b,u),n);)u+=H(o);return y=b.substring(_,u),92===o&&(y+=$()),g=ee()}if(x(o)){u+=H(o);continue}if(L(o)){v|=1,u+=H(o);continue}return I(e.Diagnostics.Invalid_character),u+=H(o),g=0}}}function ae(n,t,r,a){var i=function(e,n){var t=n.exec(e);if(!t)return;switch(t[1]){case"ts-expect-error":return 0;case"ts-ignore":return 1}return}(t,r);return void 0===i?n:e.append(n,{range:{pos:a,end:u},type:i})}function ie(){if(f=_=u,u>=d)return g=1;var n=b.charCodeAt(u);if(60===n)return 47===b.charCodeAt(u+1)?(u+=2,g=30):(u++,g=29);if(123===n)return u++,g=18;for(var t=0,r=-1;u<d&&(x(n)||(r=u),123!==(n=b.charCodeAt(u)));){if(60===n){if(N(b,u))return u=R(b,u,I),g=7;break}62===n&&I(e.Diagnostics.Unexpected_token_Did_you_mean_or_gt,u,1),125===n&&I(e.Diagnostics.Unexpected_token_Did_you_mean_or_rbrace,u,1),r>0&&r++,L(n)&&0===t?t=-1:S(n)||(t=u),u++}var a=-1===r?u:r;return y=b.substring(f,a),-1===t?12:11}function oe(){switch(f=u,b.charCodeAt(u)){case 34:case 39:return y=z(!0),g=10;default:return re()}}function se(e,n){var t=u,r=f,a=_,i=g,o=y,s=v,c=e();return c&&!n||(u=t,f=r,_=a,g=i,y=o,v=s),c}function ce(e,n,t){b=e||"",d=void 0===t?b.length:n+t,le(n||0)}function le(n){e.Debug.assert(n>=0),u=n,f=n,_=n,g=0,y=void 0,v=0}};var j=String.prototype.codePointAt?function(e,n){return e.codePointAt(n)}:function(e,n){var t=e.length;if(!(n<0||n>=t)){var r=e.charCodeAt(n);if(r>=55296&&r<=56319&&t>n+1){var a=e.charCodeAt(n+1);if(a>=56320&&a<=57343)return 1024*(r-55296)+a-56320+65536}return r}};function H(e){return e>=65536?2:1}var K=String.fromCodePoint?function(e){return String.fromCodePoint(e)}:function(n){if(e.Debug.assert(0<=n&&n<=1114111),n<=65535)return String.fromCharCode(n);var t=Math.floor((n-65536)/1024)+55296,r=(n-65536)%1024+56320;return String.fromCharCode(t,r)};function W(e){return K(e)}e.utf16EncodeAsString=W}(ts||(ts={})),function(e){function n(e){return e.start+e.length}function t(e){return 0===e.length}function r(e,n){var t=i(e,n);return t&&0===t.length?void 0:t}function a(e,n,t,r){return t<=e+n&&t+r>=e}function i(e,t){var r=Math.max(e.start,t.start),a=Math.min(n(e),n(t));return r<=a?s(r,a):void 0}function o(e,n){if(e<0)throw new Error("start < 0");if(n<0)throw new Error("length < 0");return{start:e,length:n}}function s(e,n){return o(e,n-e)}function c(e,n){if(n<0)throw new Error("newLength < 0");return{span:e,newLength:n}}function l(n){return!!Y(n)&&e.every(n.elements,u)}function u(n){return!!e.isOmittedExpression(n)||l(n.name)}function d(n){for(var t=n.parent;e.isBindingElement(t.parent);)t=t.parent.parent;return t.parent}function p(n,t){e.isBindingElement(n)&&(n=d(n));var r=t(n);return 249===n.kind&&(n=n.parent),n&&250===n.kind&&(r|=t(n),n=n.parent),n&&232===n.kind&&(r|=t(n)),r}function m(e){return 0==(8&e.flags)}function f(e){var n=e;return n.length>=3&&95===n.charCodeAt(0)&&95===n.charCodeAt(1)&&95===n.charCodeAt(2)?n.substr(1):n}function _(e){return f(e.escapedText)}function g(n){var t=n.parent.parent;if(t){if(ie(t))return y(t);switch(t.kind){case 232:if(t.declarationList&&t.declarationList.declarations[0])return y(t.declarationList.declarations[0]);break;case 233:var r=t.expression;switch(216===r.kind&&62===r.operatorToken.kind&&(r=r.left),r.kind){case 201:return r.name;case 202:var a=r.argumentExpression;if(e.isIdentifier(a))return a}break;case 207:return y(t.expression);case 245:if(ie(t.statement)||ne(t.statement))return y(t.statement)}}}function y(n){var t=E(n);return t&&e.isIdentifier(t)?t:void 0}function v(e){return e.name||g(e)}function h(e){return!!e.name}function b(n){switch(n.kind){case 78:return n;case 333:case 326:var t=n.name;if(157===t.kind)return t.right;break;case 203:case 216:var r=n;switch(e.getAssignmentDeclarationKind(r)){case 1:case 4:case 5:case 3:return e.getElementOrPropertyAccessArgumentExpressionOrName(r.left);case 7:case 8:case 9:return r.arguments[1];default:return}case 331:return v(n);case 325:return g(n);case 266:var a=n.expression;return e.isIdentifier(a)?a:void 0;case 202:var i=n;if(e.isBindableStaticElementAccessExpression(i))return i.argumentExpression}return n.name}function E(n){if(void 0!==n)return b(n)||(e.isFunctionExpression(n)||e.isClassExpression(n)?T(n):void 0)}function T(n){if(n.parent){if(e.isPropertyAssignment(n.parent)||e.isBindingElement(n.parent))return n.parent.name;if(e.isBinaryExpression(n.parent)&&n===n.parent.right){if(e.isIdentifier(n.parent.left))return n.parent.left;if(e.isAccessExpression(n.parent.left))return e.getElementOrPropertyAccessArgumentExpressionOrName(n.parent.left)}else if(e.isVariableDeclaration(n.parent)&&e.isIdentifier(n.parent.name))return n.parent.name}}function S(n,t){if(n.name){if(e.isIdentifier(n.name)){var r=n.name.escapedText;return I(n.parent,t).filter((function(n){return e.isJSDocParameterTag(n)&&e.isIdentifier(n.name)&&n.name.escapedText===r}))}var a=n.parent.parameters.indexOf(n);e.Debug.assert(a>-1,"Parameters should always be in their parents' parameter list");var i=I(n.parent,t).filter(e.isJSDocParameterTag);if(a<i.length)return[i[a]]}return e.emptyArray}function x(e){return S(e,!1)}function L(n,t){var r=n.name.escapedText;return I(n.parent,t).filter((function(n){return e.isJSDocTemplateTag(n)&&n.typeParameters.some((function(e){return e.name.escapedText===r}))}))}function A(n){return R(n,e.isJSDocReturnTag)}function D(n){var t=R(n,e.isJSDocTypeTag);if(t&&t.typeExpression&&t.typeExpression.type)return t}function C(n){var t=R(n,e.isJSDocTypeTag);return!t&&e.isParameter(n)&&(t=e.find(x(n),(function(e){return!!e.typeExpression}))),t&&t.typeExpression&&t.typeExpression.type}function I(n,t){var r=n.jsDocCache;if(void 0===r||t){var a=e.getJSDocCommentsAndTags(n,t);e.Debug.assert(a.length<2||a[0]!==a[1]),r=e.flatMap(a,(function(n){return e.isJSDoc(n)?n.tags:n})),t||(n.jsDocCache=r)}return r}function N(e){return I(e,!1)}function R(n,t,r){return e.find(I(n,r),t)}function k(e,n){return N(e).filter(n)}function O(e){var n=e.kind;return!!(32&e.flags)&&(201===n||202===n||203===n||225===n)}function P(n){return O(n)&&!e.isNonNullExpression(n)&&!!n.questionDotToken}function M(n){return e.skipOuterExpressions(n,8)}function w(e){switch(e.kind){case 294:case 295:return!0;default:return!1}}function F(e){return e>=157}function G(e){return 8<=e&&e<=14}function B(e){return 14<=e&&e<=17}function U(n){return e.isPropertyDeclaration(n)&&e.isPrivateIdentifier(n.name)}function V(e){switch(e){case 125:case 129:case 84:case 133:case 87:case 92:case 122:case 120:case 121:case 142:case 123:return!0}return!1}function j(n){return!!(92&e.modifierToFlag(n))}function H(e){return e&&W(e.kind)}function K(e){switch(e){case 251:case 165:case 166:case 167:case 168:case 208:case 209:return!0;default:return!1}}function W(e){switch(e){case 164:case 169:case 313:case 170:case 171:case 174:case 308:case 175:return!0;default:return K(e)}}function z(e){var n=e.kind;return 166===n||163===n||165===n||167===n||168===n||171===n||229===n}function q(e){return e&&(252===e.kind||221===e.kind)}function J(e){var n=e.kind;return 170===n||169===n||162===n||164===n||171===n}function X(e){var n=e.kind;return 288===n||289===n||290===n||165===n||167===n||168===n}function Y(e){if(e){var n=e.kind;return 197===n||196===n}return!1}function Q(e){switch(e.kind){case 196:case 200:return!0}return!1}function Z(e){switch(e.kind){case 197:case 199:return!0}return!1}function $(e){switch(e){case 201:case 202:case 204:case 203:case 273:case 274:case 277:case 205:case 199:case 207:case 200:case 221:case 208:case 78:case 13:case 8:case 9:case 10:case 14:case 218:case 94:case 103:case 107:case 109:case 105:case 225:case 226:case 99:return!0;default:return!1}}function ee(e){switch(e){case 214:case 215:case 210:case 211:case 212:case 213:case 206:return!0;default:return $(e)}}function ne(e){return function(e){switch(e){case 217:case 219:case 209:case 216:case 220:case 224:case 222:case 337:case 336:return!0;default:return ee(e)}}(M(e).kind)}function te(n){return e.isExportAssignment(n)||e.isExportDeclaration(n)}function re(e){return 251===e||271===e||252===e||253===e||254===e||255===e||256===e||261===e||260===e||267===e||266===e||259===e}function ae(e){return 241===e||240===e||248===e||235===e||233===e||231===e||238===e||239===e||237===e||234===e||245===e||242===e||244===e||246===e||247===e||232===e||236===e||243===e||335===e||339===e||338===e}function ie(n){return 159===n.kind?n.parent&&330!==n.parent.kind||e.isInJSFile(n):209===(t=n.kind)||198===t||252===t||221===t||166===t||255===t||291===t||270===t||251===t||208===t||167===t||262===t||260===t||265===t||253===t||280===t||165===t||164===t||256===t||259===t||263===t||269===t||160===t||288===t||163===t||162===t||168===t||289===t||254===t||159===t||249===t||331===t||324===t||333===t;var t}function oe(e){return e.kind>=314&&e.kind<=333}e.isExternalModuleNameRelative=function(n){return e.pathIsRelative(n)||e.isRootedDiskPath(n)},e.sortAndDeduplicateDiagnostics=function(n){return e.sortAndDeduplicate(n,e.compareDiagnostics)},e.getDefaultLibFileName=function(e){switch(e.target){case 99:return"lib.esnext.full.d.ts";case 7:return"lib.es2020.full.d.ts";case 6:return"lib.es2019.full.d.ts";case 5:return"lib.es2018.full.d.ts";case 4:return"lib.es2017.full.d.ts";case 3:return"lib.es2016.full.d.ts";case 2:return"lib.es6.d.ts";default:return"lib.d.ts"}},e.textSpanEnd=n,e.textSpanIsEmpty=t,e.textSpanContainsPosition=function(e,t){return t>=e.start&&t<n(e)},e.textRangeContainsPositionInclusive=function(e,n){return n>=e.pos&&n<=e.end},e.textSpanContainsTextSpan=function(e,t){return t.start>=e.start&&n(t)<=n(e)},e.textSpanOverlapsWith=function(e,n){return void 0!==r(e,n)},e.textSpanOverlap=r,e.textSpanIntersectsWithTextSpan=function(e,n){return a(e.start,e.length,n.start,n.length)},e.textSpanIntersectsWith=function(e,n,t){return a(e.start,e.length,n,t)},e.decodedTextSpanIntersectsWith=a,e.textSpanIntersectsWithPosition=function(e,t){return t<=n(e)&&t>=e.start},e.textSpanIntersection=i,e.createTextSpan=o,e.createTextSpanFromBounds=s,e.textChangeRangeNewSpan=function(e){return o(e.span.start,e.newLength)},e.textChangeRangeIsUnchanged=function(e){return t(e.span)&&0===e.newLength},e.createTextChangeRange=c,e.unchangedTextChangeRange=c(o(0,0),0),e.collapseTextChangeRangesAcrossMultipleVersions=function(t){if(0===t.length)return e.unchangedTextChangeRange;if(1===t.length)return t[0];for(var r=t[0],a=r.span.start,i=n(r.span),o=a+r.newLength,l=1;l<t.length;l++){var u=t[l],d=a,p=i,m=o,f=u.span.start,_=n(u.span),g=f+u.newLength;a=Math.min(d,f),i=Math.max(p,p+(_-m)),o=Math.max(g,g+(m-_))}return c(s(a,i),o-a)},e.getTypeParameterOwner=function(e){if(e&&159===e.kind)for(var n=e;n;n=n.parent)if(H(n)||q(n)||253===n.kind)return n},e.isParameterPropertyDeclaration=function(n,t){return e.hasSyntacticModifier(n,92)&&166===t.kind},e.isEmptyBindingPattern=l,e.isEmptyBindingElement=u,e.walkUpBindingElementsAndPatterns=d,e.getCombinedModifierFlags=function(n){return p(n,e.getEffectiveModifierFlags)},e.getCombinedNodeFlagsAlwaysIncludeJSDoc=function(n){return p(n,e.getEffectiveModifierFlagsAlwaysIncludeJSDoc)},e.getCombinedNodeFlags=function(e){return p(e,(function(e){return e.flags}))},e.validateLocaleAndSetLanguage=function(n,t,r){var a=/^([a-z]+)([_\-]([a-z]+))?$/.exec(n.toLowerCase());if(a){var i=a[1];o(i,a[3],r)||o(i,void 0,r),e.setUILocale(n)}else r&&r.push(e.createCompilerDiagnostic(e.Diagnostics.Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1,"en","ja-jp"));function o(n,r,a){var i=e.normalizePath(t.getExecutingFilePath()),o=e.getDirectoryPath(i),s=e.combinePaths(o,n);if(r&&(s=s+"-"+r),s=t.resolvePath(e.combinePaths(s,"diagnosticMessages.generated.json")),!t.fileExists(s))return!1;var c="";try{c=t.readFile(s)}catch(n){return a&&a.push(e.createCompilerDiagnostic(e.Diagnostics.Unable_to_open_file_0,s)),!1}try{e.setLocalizedDiagnosticMessages(JSON.parse(c))}catch(n){return a&&a.push(e.createCompilerDiagnostic(e.Diagnostics.Corrupted_locale_file_0,s)),!1}return!0}},e.getOriginalNode=function(e,n){if(e)for(;void 0!==e.original;)e=e.original;return!n||n(e)?e:void 0},e.findAncestor=function(e,n){for(;e;){var t=n(e);if("quit"===t)return;if(t)return e;e=e.parent}},e.isParseTreeNode=m,e.getParseTreeNode=function(e,n){if(void 0===e||m(e))return e;for(e=e.original;e;){if(m(e))return!n||n(e)?e:void 0;e=e.original}},e.escapeLeadingUnderscores=function(e){return e.length>=2&&95===e.charCodeAt(0)&&95===e.charCodeAt(1)?"_"+e:e},e.unescapeLeadingUnderscores=f,e.idText=_,e.symbolName=function(e){return e.valueDeclaration&&U(e.valueDeclaration)?_(e.valueDeclaration.name):f(e.escapedName)},e.nodeHasName=function n(t,r){return!(!h(t)||!e.isIdentifier(t.name)||_(t.name)!==_(r))||!(!e.isVariableStatement(t)||!e.some(t.declarationList.declarations,(function(e){return n(e,r)})))},e.getNameOfJSDocTypedef=v,e.isNamedDeclaration=h,e.getNonAssignedNameOfDeclaration=b,e.getNameOfDeclaration=E,e.getAssignedName=T,e.getJSDocParameterTags=x,e.getJSDocParameterTagsNoCache=function(e){return S(e,!0)},e.getJSDocTypeParameterTags=function(e){return L(e,!1)},e.getJSDocTypeParameterTagsNoCache=function(e){return L(e,!0)},e.hasJSDocParameterTags=function(n){return!!R(n,e.isJSDocParameterTag)},e.getJSDocAugmentsTag=function(n){return R(n,e.isJSDocAugmentsTag)},e.getJSDocImplementsTags=function(n){return k(n,e.isJSDocImplementsTag)},e.getJSDocClassTag=function(n){return R(n,e.isJSDocClassTag)},e.getJSDocPublicTag=function(n){return R(n,e.isJSDocPublicTag)},e.getJSDocPublicTagNoCache=function(n){return R(n,e.isJSDocPublicTag,!0)},e.getJSDocPrivateTag=function(n){return R(n,e.isJSDocPrivateTag)},e.getJSDocPrivateTagNoCache=function(n){return R(n,e.isJSDocPrivateTag,!0)},e.getJSDocProtectedTag=function(n){return R(n,e.isJSDocProtectedTag)},e.getJSDocProtectedTagNoCache=function(n){return R(n,e.isJSDocProtectedTag,!0)},e.getJSDocReadonlyTag=function(n){return R(n,e.isJSDocReadonlyTag)},e.getJSDocReadonlyTagNoCache=function(n){return R(n,e.isJSDocReadonlyTag,!0)},e.getJSDocDeprecatedTag=function(n){return R(n,e.isJSDocDeprecatedTag)},e.getJSDocDeprecatedTagNoCache=function(n){return R(n,e.isJSDocDeprecatedTag,!0)},e.getJSDocEnumTag=function(n){return R(n,e.isJSDocEnumTag)},e.getJSDocThisTag=function(n){return R(n,e.isJSDocThisTag)},e.getJSDocReturnTag=A,e.getJSDocTemplateTag=function(n){return R(n,e.isJSDocTemplateTag)},e.getJSDocTypeTag=D,e.getJSDocType=C,e.getJSDocReturnType=function(n){var t=A(n);if(t&&t.typeExpression)return t.typeExpression.type;var r=D(n);if(r&&r.typeExpression){var a=r.typeExpression.type;if(e.isTypeLiteralNode(a)){var i=e.find(a.members,e.isCallSignatureDeclaration);return i&&i.type}if(e.isFunctionTypeNode(a)||e.isJSDocFunctionType(a))return a.type}},e.getJSDocTags=N,e.getJSDocTagsNoCache=function(e){return I(e,!0)},e.getAllJSDocTags=k,e.getAllJSDocTagsOfKind=function(e,n){return N(e).filter((function(e){return e.kind===n}))},e.getEffectiveTypeParameterDeclarations=function(n){if(e.isJSDocSignature(n))return e.emptyArray;if(e.isJSDocTypeAlias(n))return e.Debug.assert(311===n.parent.kind),e.flatMap(n.parent.tags,(function(n){return e.isJSDocTemplateTag(n)?n.typeParameters:void 0}));if(n.typeParameters)return n.typeParameters;if(e.isInJSFile(n)){var t=e.getJSDocTypeParameterDeclarations(n);if(t.length)return t;var r=C(n);if(r&&e.isFunctionTypeNode(r)&&r.typeParameters)return r.typeParameters}return e.emptyArray},e.getEffectiveConstraintOfTypeParameter=function(n){return n.constraint?n.constraint:e.isJSDocTemplateTag(n.parent)&&n===n.parent.typeParameters[0]?n.parent.constraint:void 0},e.isIdentifierOrPrivateIdentifier=function(e){return 78===e.kind||79===e.kind},e.isGetOrSetAccessorDeclaration=function(e){return 168===e.kind||167===e.kind},e.isPropertyAccessChain=function(n){return e.isPropertyAccessExpression(n)&&!!(32&n.flags)},e.isElementAccessChain=function(n){return e.isElementAccessExpression(n)&&!!(32&n.flags)},e.isCallChain=function(n){return e.isCallExpression(n)&&!!(32&n.flags)},e.isOptionalChain=O,e.isOptionalChainRoot=P,e.isExpressionOfOptionalChainRoot=function(e){return P(e.parent)&&e.parent.expression===e},e.isOutermostOptionalChain=function(e){return!O(e.parent)||P(e.parent)||e!==e.parent.expression},e.isNullishCoalesce=function(e){return 216===e.kind&&60===e.operatorToken.kind},e.isConstTypeReference=function(n){return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)&&"const"===n.typeName.escapedText&&!n.typeArguments},e.skipPartiallyEmittedExpressions=M,e.isNonNullChain=function(n){return e.isNonNullExpression(n)&&!!(32&n.flags)},e.isBreakOrContinueStatement=function(e){return 241===e.kind||240===e.kind},e.isNamedExportBindings=function(e){return 269===e.kind||268===e.kind},e.isUnparsedTextLike=w,e.isUnparsedNode=function(e){return w(e)||292===e.kind||296===e.kind},e.isJSDocPropertyLikeTag=function(e){return 333===e.kind||326===e.kind},e.isNode=function(e){return F(e.kind)},e.isNodeKind=F,e.isToken=function(e){return e.kind>=0&&e.kind<=156},e.isNodeArray=function(e){return e.hasOwnProperty("pos")&&e.hasOwnProperty("end")},e.isLiteralKind=G,e.isLiteralExpression=function(e){return G(e.kind)},e.isTemplateLiteralKind=B,e.isTemplateLiteralToken=function(e){return B(e.kind)},e.isTemplateMiddleOrTemplateTail=function(e){var n=e.kind;return 16===n||17===n},e.isImportOrExportSpecifier=function(n){return e.isImportSpecifier(n)||e.isExportSpecifier(n)},e.isTypeOnlyImportOrExportDeclaration=function(e){switch(e.kind){case 265:case 270:return e.parent.parent.isTypeOnly;case 263:return e.parent.isTypeOnly;case 262:return e.isTypeOnly;default:return!1}},e.isStringTextContainingNode=function(e){return 10===e.kind||B(e.kind)},e.isGeneratedIdentifier=function(n){return e.isIdentifier(n)&&(7&n.autoGenerateFlags)>0},e.isPrivateIdentifierPropertyDeclaration=U,e.isPrivateIdentifierPropertyAccessExpression=function(n){return e.isPropertyAccessExpression(n)&&e.isPrivateIdentifier(n.name)},e.isModifierKind=V,e.isParameterPropertyModifier=j,e.isClassMemberModifier=function(e){return j(e)||123===e},e.isModifier=function(e){return V(e.kind)},e.isEntityName=function(e){var n=e.kind;return 157===n||78===n},e.isPropertyName=function(e){var n=e.kind;return 78===n||79===n||10===n||8===n||158===n},e.isBindingName=function(e){var n=e.kind;return 78===n||196===n||197===n},e.isFunctionLike=H,e.isFunctionLikeDeclaration=function(e){return e&&K(e.kind)},e.isFunctionLikeKind=W,e.isFunctionOrModuleBlock=function(n){return e.isSourceFile(n)||e.isModuleBlock(n)||e.isBlock(n)&&H(n.parent)},e.isClassElement=z,e.isClassLike=q,e.isAccessor=function(e){return e&&(167===e.kind||168===e.kind)},e.isMethodOrAccessor=function(e){switch(e.kind){case 165:case 167:case 168:return!0;default:return!1}},e.isTypeElement=J,e.isClassOrTypeElement=function(e){return J(e)||z(e)},e.isObjectLiteralElementLike=X,e.isTypeNode=function(n){return e.isTypeNodeKind(n.kind)},e.isFunctionOrConstructorTypeNode=function(e){switch(e.kind){case 174:case 175:return!0}return!1},e.isBindingPattern=Y,e.isAssignmentPattern=function(e){var n=e.kind;return 199===n||200===n},e.isArrayBindingElement=function(e){var n=e.kind;return 198===n||222===n},e.isDeclarationBindingElement=function(e){switch(e.kind){case 249:case 160:case 198:return!0}return!1},e.isBindingOrAssignmentPattern=function(e){return Q(e)||Z(e)},e.isObjectBindingOrAssignmentPattern=Q,e.isArrayBindingOrAssignmentPattern=Z,e.isPropertyAccessOrQualifiedNameOrImportTypeNode=function(e){var n=e.kind;return 201===n||157===n||195===n},e.isPropertyAccessOrQualifiedName=function(e){var n=e.kind;return 201===n||157===n},e.isCallLikeExpression=function(e){switch(e.kind){case 275:case 274:case 203:case 204:case 205:case 161:return!0;default:return!1}},e.isCallOrNewExpression=function(e){return 203===e.kind||204===e.kind},e.isTemplateLiteral=function(e){var n=e.kind;return 218===n||14===n},e.isLeftHandSideExpression=function(e){return $(M(e).kind)},e.isUnaryExpression=function(e){return ee(M(e).kind)},e.isUnaryExpressionWithWrite=function(e){switch(e.kind){case 215:return!0;case 214:return 45===e.operator||46===e.operator;default:return!1}},e.isExpression=ne,e.isAssertionExpression=function(e){var n=e.kind;return 206===n||224===n},e.isNotEmittedOrPartiallyEmittedNode=function(n){return e.isNotEmittedStatement(n)||e.isPartiallyEmittedExpression(n)},e.isIterationStatement=function e(n,t){switch(n.kind){case 237:case 238:case 239:case 235:case 236:return!0;case 245:return t&&e(n.statement,t)}return!1},e.isScopeMarker=te,e.hasScopeMarker=function(n){return e.some(n,te)},e.needsScopeMarker=function(n){return!(e.isAnyImportOrReExport(n)||e.isExportAssignment(n)||e.hasSyntacticModifier(n,1)||e.isAmbientModule(n))},e.isExternalModuleIndicator=function(n){return e.isAnyImportOrReExport(n)||e.isExportAssignment(n)||e.hasSyntacticModifier(n,1)},e.isForInOrOfStatement=function(e){return 238===e.kind||239===e.kind},e.isConciseBody=function(n){return e.isBlock(n)||ne(n)},e.isFunctionBody=function(n){return e.isBlock(n)},e.isForInitializer=function(n){return e.isVariableDeclarationList(n)||ne(n)},e.isModuleBody=function(e){var n=e.kind;return 257===n||256===n||78===n},e.isNamespaceBody=function(e){var n=e.kind;return 257===n||256===n},e.isJSDocNamespaceBody=function(e){var n=e.kind;return 78===n||256===n},e.isNamedImportBindings=function(e){var n=e.kind;return 264===n||263===n},e.isModuleOrEnumDeclaration=function(e){return 256===e.kind||255===e.kind},e.isDeclaration=ie,e.isDeclarationStatement=function(e){return re(e.kind)},e.isStatementButNotDeclaration=function(e){return ae(e.kind)},e.isStatement=function(n){var t=n.kind;return ae(t)||re(t)||function(n){if(230!==n.kind)return!1;if(void 0!==n.parent&&(247===n.parent.kind||287===n.parent.kind))return!1;return!e.isFunctionBlock(n)}(n)},e.isStatementOrBlock=function(e){var n=e.kind;return ae(n)||re(n)||230===n},e.isModuleReference=function(e){var n=e.kind;return 272===n||157===n||78===n},e.isJsxTagNameExpression=function(e){var n=e.kind;return 107===n||78===n||201===n},e.isJsxChild=function(e){var n=e.kind;return 273===n||283===n||274===n||11===n||277===n},e.isJsxAttributeLike=function(e){var n=e.kind;return 280===n||282===n},e.isStringLiteralOrJsxExpression=function(e){var n=e.kind;return 10===n||283===n},e.isJsxOpeningLikeElement=function(e){var n=e.kind;return 275===n||274===n},e.isCaseOrDefaultClause=function(e){var n=e.kind;return 284===n||285===n},e.isJSDocNode=function(e){return e.kind>=301&&e.kind<=333},e.isJSDocCommentContainingNode=function(n){return 311===n.kind||310===n.kind||oe(n)||e.isJSDocTypeLiteral(n)||e.isJSDocSignature(n)},e.isJSDocTag=oe,e.isSetAccessor=function(e){return 168===e.kind},e.isGetAccessor=function(e){return 167===e.kind},e.hasJSDocNodes=function(e){var n=e.jsDoc;return!!n&&n.length>0},e.hasType=function(e){return!!e.type},e.hasInitializer=function(e){return!!e.initializer},e.hasOnlyExpressionInitializer=function(e){switch(e.kind){case 249:case 160:case 198:case 162:case 163:case 288:case 291:return!0;default:return!1}},e.isObjectLiteralElement=function(e){return 280===e.kind||282===e.kind||X(e)},e.isTypeReferenceType=function(e){return 173===e.kind||223===e.kind};var se=1073741823;e.guessIndentation=function(n){for(var t=se,r=0,a=n;r<a.length;r++){var i=a[r];if(i.length){for(var o=0;o<i.length&&o<t&&e.isWhiteSpaceLike(i.charCodeAt(o));o++);if(o<t&&(t=o),0===t)return 0}}return t===se?void 0:t},e.isStringLiteralLike=function(e){return 10===e.kind||14===e.kind}}(ts||(ts={})),function(e){e.resolvingEmptyArray=[],e.externalHelpersModuleNameText="tslib",e.defaultMaximumTruncationLength=160,e.noTruncationMaximumTruncationLength=1e6,e.getDeclarationOfKind=function(e,n){var t=e.declarations;if(t)for(var r=0,a=t;r<a.length;r++){var i=a[r];if(i.kind===n)return i}},e.createUnderscoreEscapedMap=function(){return new e.Map},e.hasEntries=function(e){return!!e&&!!e.size},e.createSymbolTable=function(n){var t=new e.Map;if(n)for(var r=0,a=n;r<a.length;r++){var i=a[r];t.set(i.escapedName,i)}return t},e.isTransientSymbol=function(e){return 0!=(33554432&e.flags)};var n,t,r=(n="",{getText:function(){return n},write:t=function(e){return n+=e},rawWrite:t,writeKeyword:t,writeOperator:t,writePunctuation:t,writeSpace:t,writeStringLiteral:t,writeLiteral:t,writeParameter:t,writeProperty:t,writeSymbol:function(e,n){return t(e)},writeTrailingSemicolon:t,writeComment:t,getTextPos:function(){return n.length},getLine:function(){return 0},getColumn:function(){return 0},getIndent:function(){return 0},isAtStartOfLine:function(){return!1},hasTrailingComment:function(){return!1},hasTrailingWhitespace:function(){return!!n.length&&e.isWhiteSpaceLike(n.charCodeAt(n.length-1))},writeLine:function(){return n+=" "},increaseIndent:e.noop,decreaseIndent:e.noop,clear:function(){return n=""},trackSymbol:e.noop,reportInaccessibleThisError:e.noop,reportInaccessibleUniqueSymbolError:e.noop,reportPrivateInBaseOfClassExpression:e.noop});function a(n,t){return e.moduleResolutionOptionDeclarations.some((function(e){return!Jr(mr(n,e),mr(t,e))}))}function i(e){return e.end-e.pos}function o(n){return function(n){if(!(524288&n.flags)){(0!=(65536&n.flags)||e.forEachChild(n,o))&&(n.flags|=262144),n.flags|=524288}}(n),0!=(262144&n.flags)}function s(e){for(;e&&297!==e.kind;)e=e.parent;return e}function c(n,t){e.Debug.assert(n>=0);var r=e.getLineStarts(t),a=n,i=t.text;if(a+1===r.length)return i.length-1;var o=r[a],s=r[a+1]-1;for(e.Debug.assert(e.isLineBreak(i.charCodeAt(s)));o<=s&&e.isLineBreak(i.charCodeAt(s));)s--;return s}function l(e){return void 0===e||e.pos===e.end&&e.pos>=0&&1!==e.kind}function u(e){return!l(e)}function d(e,n,t){if(void 0===n||0===n.length)return e;for(var r=0;r<e.length&&t(e[r]);++r);return e.splice.apply(e,__spreadArrays([r,0],n)),e}function p(e,n,t){if(void 0===n)return e;for(var r=0;r<e.length&&t(e[r]);++r);return e.splice(r,0,n),e}function m(e){return U(e)||!!(1048576&b(e))}function f(e,n){return 42===e.charCodeAt(n+1)&&33===e.charCodeAt(n+2)}function _(n,t,r){return l(n)?n.pos:e.isJSDocNode(n)||11===n.kind?e.skipTrivia((t||s(n)).text,n.pos,!1,!0):r&&e.hasJSDocNodes(n)?_(n.jsDoc[0],t):334===n.kind&&n._children.length>0?_(n._children[0],t,r):e.skipTrivia((t||s(n)).text,n.pos)}function g(e,n,t){return void 0===t&&(t=!1),y(e.text,n,t)}function y(n,t,r){if(void 0===r&&(r=!1),l(t))return"";var a=n.substring(r?t.pos:e.skipTrivia(n,t.pos),t.end);return function(n){return!!e.findAncestor(n,e.isJSDocTypeExpression)}(t)&&(a=a.replace(/(^|\r?\n|\r)\s*\*\s*/g,"$1")),a}function v(e,n){return void 0===n&&(n=!1),g(s(e),e,n)}function h(e){return e.pos}function b(e){var n=e.emitNode;return n&&n.flags||0}function E(e){var n=_n(e);return 249===n.kind&&287===n.parent.kind}function T(n){return e.isModuleDeclaration(n)&&(10===n.name.kind||x(n))}function S(n){return e.isModuleDeclaration(n)||e.isIdentifier(n)}function x(e){return!!(1024&e.flags)}function L(e){return T(e)&&A(e)}function A(n){switch(n.parent.kind){case 297:return e.isExternalModule(n.parent);case 257:return T(n.parent.parent)&&e.isSourceFile(n.parent.parent.parent)&&!e.isExternalModule(n.parent.parent.parent)}return!1}function D(n,t){switch(n.kind){case 297:case 258:case 287:case 256:case 237:case 238:case 239:case 166:case 165:case 167:case 168:case 251:case 208:case 209:return!0;case 230:return!e.isFunctionLike(t)}return!1}function C(n){switch(n.kind){case 169:case 170:case 164:case 171:case 174:case 175:case 308:case 252:case 221:case 253:case 254:case 330:case 251:case 165:case 166:case 167:case 168:case 208:case 209:return!0;default:return e.assertType(n),!1}}function I(e){switch(e.kind){case 261:case 260:return!0;default:return!1}}function N(e){return e&&0!==i(e)?v(e):"(Missing)"}function R(n){switch(n.kind){case 78:case 79:return n.escapedText;case 10:case 8:case 14:return e.escapeLeadingUnderscores(n.text);case 158:return an(n.expression)?e.escapeLeadingUnderscores(n.expression.text):e.Debug.fail("Text of property name cannot be read from non-literal-valued ComputedPropertyNames");default:return e.Debug.assertNever(n)}}function k(n){switch(n.kind){case 107:return"this";case 79:case 78:return 0===i(n)?e.idText(n):v(n);case 157:return k(n.left)+"."+k(n.right);case 201:return e.isIdentifier(n.name)||e.isPrivateIdentifier(n.name)?k(n.expression)+"."+k(n.name):e.Debug.assertNever(n.name);default:return e.Debug.assertNever(n)}}function O(e,n,t,r,a,i,o){var s=M(e,n);return tr(e,s.start,s.length,t,r,a,i,o)}function P(n,t){var r=e.createScanner(n.languageVersion,!0,n.languageVariant,n.text,void 0,t);r.scan();var a=r.getTokenPos();return e.createTextSpanFromBounds(a,r.getTextPos())}function M(n,t){var r=t;switch(t.kind){case 297:var a=e.skipTrivia(n.text,0,!1);return a===n.text.length?e.createTextSpan(0,0):P(n,a);case 249:case 198:case 252:case 221:case 253:case 256:case 255:case 291:case 251:case 208:case 165:case 167:case 168:case 254:case 163:case 162:r=t.name;break;case 209:return function(n,t){var r=e.skipTrivia(n.text,t.pos);if(t.body&&230===t.body.kind){var a=e.getLineAndCharacterOfPosition(n,t.body.pos).line;if(a<e.getLineAndCharacterOfPosition(n,t.body.end).line)return e.createTextSpan(r,c(a,n)-r+1)}return e.createTextSpanFromBounds(r,t.end)}(n,t);case 284:case 285:var i=e.skipTrivia(n.text,t.pos),o=t.statements.length>0?t.statements[0].pos:t.end;return e.createTextSpanFromBounds(i,o)}if(void 0===r)return P(n,t.pos);e.Debug.assert(!e.isJSDoc(r));var s=l(r),u=s||e.isJsxText(t)?r.pos:e.skipTrivia(n.text,r.pos);return s?(e.Debug.assert(u===r.pos,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),e.Debug.assert(u===r.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809")):(e.Debug.assert(u>=r.pos,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809"),e.Debug.assert(u<=r.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809")),e.createTextSpanFromBounds(u,r.end)}function w(e){return 6===e.scriptKind}function F(n){return!!(2&e.getCombinedNodeFlags(n))}function G(e){return 203===e.kind&&99===e.expression.kind}function B(n){return e.isImportTypeNode(n)&&e.isLiteralTypeNode(n.argument)&&e.isStringLiteral(n.argument.literal)}function U(e){return 233===e.kind&&10===e.expression.kind}function V(e){return!!(1048576&b(e))}function j(n){return e.isIdentifier(n.name)&&!n.initializer}e.changesAffectModuleResolution=function(e,n){return e.configFilePath!==n.configFilePath||a(e,n)},e.optionsHaveModuleResolutionChanges=a,e.forEachAncestor=function(n,t){for(;;){var r=t(n);if("quit"===r)return;if(void 0!==r)return r;if(e.isSourceFile(n))return;n=n.parent}},e.forEachEntry=function(e,n){for(var t=e.entries(),r=t.next();!r.done;r=t.next()){var a=r.value,i=a[0],o=n(a[1],i);if(o)return o}},e.forEachKey=function(e,n){for(var t=e.keys(),r=t.next();!r.done;r=t.next()){var a=n(r.value);if(a)return a}},e.copyEntries=function(e,n){e.forEach((function(e,t){n.set(t,e)}))},e.usingSingleLineStringWriter=function(e){var n=r.getText();try{return e(r),r.getText()}finally{r.clear(),r.writeKeyword(n)}},e.getFullWidth=i,e.getResolvedModule=function(e,n){return e&&e.resolvedModules&&e.resolvedModules.get(n)},e.setResolvedModule=function(n,t,r){n.resolvedModules||(n.resolvedModules=new e.Map),n.resolvedModules.set(t,r)},e.setResolvedTypeReferenceDirective=function(n,t,r){n.resolvedTypeReferenceDirectiveNames||(n.resolvedTypeReferenceDirectiveNames=new e.Map),n.resolvedTypeReferenceDirectiveNames.set(t,r)},e.projectReferenceIsEqualTo=function(e,n){return e.path===n.path&&!e.prepend==!n.prepend&&!e.circular==!n.circular},e.moduleResolutionIsEqualTo=function(e,n){return e.isExternalLibraryImport===n.isExternalLibraryImport&&e.extension===n.extension&&e.resolvedFileName===n.resolvedFileName&&e.originalPath===n.originalPath&&(t=e.packageId,r=n.packageId,t===r||!!t&&!!r&&t.name===r.name&&t.subModuleName===r.subModuleName&&t.version===r.version);var t,r},e.packageIdToString=function(e){var n=e.name,t=e.subModuleName;return(t?n+"/"+t:n)+"@"+e.version},e.typeDirectiveIsEqualTo=function(e,n){return e.resolvedFileName===n.resolvedFileName&&e.primary===n.primary},e.hasChangesInResolutions=function(n,t,r,a){e.Debug.assert(n.length===t.length);for(var i=0;i<n.length;i++){var o=t[i],s=r&&r.get(n[i]);if(s?!o||!a(s,o):o)return!0}return!1},e.containsParseError=o,e.getSourceFileOfNode=s,e.isStatementWithLocals=function(e){switch(e.kind){case 230:case 258:case 237:case 238:case 239:return!0}return!1},e.getStartPositionOfLine=function(n,t){return e.Debug.assert(n>=0),e.getLineStarts(t)[n]},e.nodePosToString=function(n){var t=s(n),r=e.getLineAndCharacterOfPosition(t,n.pos);return t.fileName+"("+(r.line+1)+","+(r.character+1)+")"},e.getEndLinePosition=c,e.isFileLevelUniqueName=function(e,n,t){return!(t&&t(n)||e.identifiers.has(n))},e.nodeIsMissing=l,e.nodeIsPresent=u,e.insertStatementsAfterStandardPrologue=function(e,n){return d(e,n,U)},e.insertStatementsAfterCustomPrologue=function(e,n){return d(e,n,m)},e.insertStatementAfterStandardPrologue=function(e,n){return p(e,n,U)},e.insertStatementAfterCustomPrologue=function(e,n){return p(e,n,m)},e.isRecognizedTripleSlashComment=function(n,t,r){if(47===n.charCodeAt(t+1)&&t+2<r&&47===n.charCodeAt(t+2)){var a=n.substring(t,r);return!!(a.match(e.fullTripleSlashReferencePathRegEx)||a.match(e.fullTripleSlashAMDReferencePathRegEx)||a.match(H)||a.match(K))}return!1},e.isPinnedComment=f,e.createCommentDirectivesMap=function(n,t){var r=new e.Map(t.map((function(t){return[""+e.getLineAndCharacterOfPosition(n,t.range.end).line,t]}))),a=new e.Map;return{getUnusedExpectations:function(){return e.arrayFrom(r.entries()).filter((function(e){var n=e[0];return 0===e[1].type&&!a.get(n)})).map((function(e){e[0];return e[1]}))},markUsed:function(e){if(!r.has(""+e))return!1;return a.set(""+e,!0),!0}}},e.getTokenPosOfNode=_,e.getNonDecoratorTokenPosOfNode=function(n,t){return l(n)||!n.decorators?_(n,t):e.skipTrivia((t||s(n)).text,n.decorators.end)},e.getSourceTextOfNodeFromSourceFile=g,e.isExportNamespaceAsDefaultDeclaration=function(n){return!!(e.isExportDeclaration(n)&&n.exportClause&&e.isNamespaceExport(n.exportClause)&&"default"===n.exportClause.name.escapedText)},e.getTextOfNodeFromSourceText=y,e.getTextOfNode=v,e.indexOfNode=function(n,t){return e.binarySearch(n,t,h,e.compareValues)},e.getEmitFlags=b,e.getScriptTargetFeatures=function(){return{es2015:{Array:["find","findIndex","fill","copyWithin","entries","keys","values"],RegExp:["flags","sticky","unicode"],Reflect:["apply","construct","defineProperty","deleteProperty","get"," getOwnPropertyDescriptor","getPrototypeOf","has","isExtensible","ownKeys","preventExtensions","set","setPrototypeOf"],ArrayConstructor:["from","of"],ObjectConstructor:["assign","getOwnPropertySymbols","keys","is","setPrototypeOf"],NumberConstructor:["isFinite","isInteger","isNaN","isSafeInteger","parseFloat","parseInt"],Math:["clz32","imul","sign","log10","log2","log1p","expm1","cosh","sinh","tanh","acosh","asinh","atanh","hypot","trunc","fround","cbrt"],Map:["entries","keys","values"],Set:["entries","keys","values"],Promise:e.emptyArray,PromiseConstructor:["all","race","reject","resolve"],Symbol:["for","keyFor"],WeakMap:["entries","keys","values"],WeakSet:["entries","keys","values"],Iterator:e.emptyArray,AsyncIterator:e.emptyArray,String:["codePointAt","includes","endsWith","normalize","repeat","startsWith","anchor","big","blink","bold","fixed","fontcolor","fontsize","italics","link","small","strike","sub","sup"],StringConstructor:["fromCodePoint","raw"]},es2016:{Array:["includes"]},es2017:{Atomics:e.emptyArray,SharedArrayBuffer:e.emptyArray,String:["padStart","padEnd"],ObjectConstructor:["values","entries","getOwnPropertyDescriptors"],DateTimeFormat:["formatToParts"]},es2018:{Promise:["finally"],RegExpMatchArray:["groups"],RegExpExecArray:["groups"],RegExp:["dotAll"],Intl:["PluralRules"],AsyncIterable:e.emptyArray,AsyncIterableIterator:e.emptyArray,AsyncGenerator:e.emptyArray,AsyncGeneratorFunction:e.emptyArray},es2019:{Array:["flat","flatMap"],ObjectConstructor:["fromEntries"],String:["trimStart","trimEnd","trimLeft","trimRight"],Symbol:["description"]},es2020:{BigInt:e.emptyArray,BigInt64Array:e.emptyArray,BigUint64Array:e.emptyArray,PromiseConstructor:["allSettled"],SymbolConstructor:["matchAll"],String:["matchAll"],DataView:["setBigInt64","setBigUint64","getBigInt64","getBigUint64"],RelativeTimeFormat:["format","formatToParts","resolvedOptions"]},esnext:{PromiseConstructor:["any"],String:["replaceAll"],NumberFormat:["formatToParts"]}}},function(e){e[e.None=0]="None",e[e.NeverAsciiEscape=1]="NeverAsciiEscape",e[e.JsxAttributeEscape=2]="JsxAttributeEscape",e[e.TerminateUnterminatedLiterals=4]="TerminateUnterminatedLiterals"}(e.GetLiteralTextFlags||(e.GetLiteralTextFlags={})),e.getLiteralText=function(n,t,r){if(!gn(n)&&n.parent&&!(4&r&&n.isUnterminated)&&!(e.isNumericLiteral(n)&&512&n.numericLiteralFlags||e.isBigIntLiteral(n)))return g(t,n);switch(n.kind){case 10:var a=2&r?Mn:1&r||16777216&b(n)?Cn:Nn;return n.singleQuote?"'"+a(n.text,39)+"'":'"'+a(n.text,34)+'"';case 14:case 15:case 16:case 17:a=1&r||16777216&b(n)?Cn:Nn;var i=n.rawText||function(e){return e.replace(En,"\\${")}(a(n.text,96));switch(n.kind){case 14:return"`"+i+"`";case 15:return"`"+i+"${";case 16:return"}"+i+"${";case 17:return"}"+i+"`"}break;case 8:case 9:return n.text;case 13:return 4&r&&n.isUnterminated?n.text+(92===n.text.charCodeAt(n.text.length-1)?" /":"/"):n.text}return e.Debug.fail("Literal kind '"+n.kind+"' not accounted for.")},e.getTextOfConstantValue=function(n){return e.isString(n)?'"'+Nn(n)+'"':""+n},e.makeIdentifierFromModuleName=function(n){return e.getBaseFileName(n).replace(/^(\d)/,"_$1").replace(/\W/g,"_")},e.isBlockOrCatchScoped=function(n){return 0!=(3&e.getCombinedNodeFlags(n))||E(n)},e.isCatchClauseVariableDeclarationOrBindingElement=E,e.isAmbientModule=T,e.isModuleWithStringLiteralName=function(n){return e.isModuleDeclaration(n)&&10===n.name.kind},e.isNonGlobalAmbientModule=function(n){return e.isModuleDeclaration(n)&&e.isStringLiteral(n.name)},e.isEffectiveModuleDeclaration=S,e.isShorthandAmbientModuleSymbol=function(e){return(n=e.valueDeclaration)&&256===n.kind&&!n.body;var n},e.isBlockScopedContainerTopLevel=function(n){return 297===n.kind||256===n.kind||e.isFunctionLike(n)},e.isGlobalScopeAugmentation=x,e.isExternalModuleAugmentation=L,e.isModuleAugmentationExternal=A,e.getNonAugmentationDeclaration=function(n){return e.find(n.declarations,(function(n){return!(L(n)||e.isModuleDeclaration(n)&&x(n))}))},e.isEffectiveExternalModule=function(n,t){return e.isExternalModule(n)||t.isolatedModules||lr(t)===e.ModuleKind.CommonJS&&!!n.commonJsModuleIndicator},e.isEffectiveStrictModeSourceFile=function(n,t){switch(n.scriptKind){case 1:case 3:case 2:case 4:break;default:return!1}return!n.isDeclarationFile&&(!!dr(t,"alwaysStrict")||(!!e.startsWithUseStrict(n.statements)||!(!e.isExternalModule(n)&&!t.isolatedModules)&&(lr(t)>=e.ModuleKind.ES2015||!t.noImplicitUseStrict)))},e.isBlockScope=D,e.isDeclarationWithTypeParameters=function(n){switch(n.kind){case 324:case 331:case 313:return!0;default:return e.assertType(n),C(n)}},e.isDeclarationWithTypeParameterChildren=C,e.isAnyImportSyntax=I,e.isLateVisibilityPaintedStatement=function(e){switch(e.kind){case 261:case 260:case 232:case 252:case 251:case 256:case 254:case 253:case 255:return!0;default:return!1}},e.isAnyImportOrReExport=function(n){return I(n)||e.isExportDeclaration(n)},e.getEnclosingBlockScopeContainer=function(n){return e.findAncestor(n.parent,(function(e){return D(e,e.parent)}))},e.declarationNameToString=N,e.getNameFromIndexInfo=function(e){return e.declaration?N(e.declaration.parameters[0].name):void 0},e.isComputedNonLiteralName=function(e){return 158===e.kind&&!an(e.expression)},e.getTextOfPropertyName=R,e.entityNameToString=k,e.createDiagnosticForNode=function(e,n,t,r,a,i){return O(s(e),e,n,t,r,a,i)},e.createDiagnosticForNodeArray=function(n,t,r,a,i,o,s){var c=e.skipTrivia(n.text,t.pos);return tr(n,c,t.end-c,r,a,i,o,s)},e.createDiagnosticForNodeInSourceFile=O,e.createDiagnosticForNodeFromMessageChain=function(e,n,t){var r=s(e),a=M(r,e);return{file:r,start:a.start,length:a.length,code:n.code,category:n.category,messageText:n.next?n:n.messageText,relatedInformation:t}},e.createDiagnosticForFileFromMessageChain=function(e,n,t){return{file:e,start:0,length:0,code:n.code,category:n.category,messageText:n.next?n:n.messageText,relatedInformation:t}},e.createDiagnosticForRange=function(e,n,t){return{file:e,start:n.pos,length:n.end-n.pos,code:t.code,category:t.category,messageText:t.message}},e.getSpanOfTokenAtPosition=P,e.getErrorSpanForNode=M,e.isExternalOrCommonJsModule=function(e){return void 0!==(e.externalModuleIndicator||e.commonJsModuleIndicator)},e.isJsonSourceFile=w,e.isEnumConst=function(n){return!!(2048&e.getCombinedModifierFlags(n))},e.isDeclarationReadonly=function(n){return!(!(64&e.getCombinedModifierFlags(n))||e.isParameterPropertyDeclaration(n,n.parent))},e.isVarConst=F,e.isLet=function(n){return!!(1&e.getCombinedNodeFlags(n))},e.isSuperCall=function(e){return 203===e.kind&&105===e.expression.kind},e.isImportCall=G,e.isImportMeta=function(n){return e.isMetaProperty(n)&&99===n.keywordToken&&"meta"===n.name.escapedText},e.isLiteralImportTypeNode=B,e.isPrologueDirective=U,e.isCustomPrologue=V,e.isHoistedFunction=function(n){return V(n)&&e.isFunctionDeclaration(n)},e.isHoistedVariableStatement=function(n){return V(n)&&e.isVariableStatement(n)&&e.every(n.declarationList.declarations,j)},e.getLeadingCommentRangesOfNode=function(n,t){return 11!==n.kind?e.getLeadingCommentRanges(t.text,n.pos):void 0},e.getJSDocCommentRanges=function(n,t){var r=160===n.kind||159===n.kind||208===n.kind||209===n.kind||207===n.kind?e.concatenate(e.getTrailingCommentRanges(t,n.pos),e.getLeadingCommentRanges(t,n.pos)):e.getLeadingCommentRanges(t,n.pos);return e.filter(r,(function(e){return 42===t.charCodeAt(e.pos+1)&&42===t.charCodeAt(e.pos+2)&&47!==t.charCodeAt(e.pos+3)}))},e.fullTripleSlashReferencePathRegEx=/^(\/\/\/\s*<reference\s+path\s*=\s*)('|")(.+?)\2.*?\/>/;var H=/^(\/\/\/\s*<reference\s+types\s*=\s*)('|")(.+?)\2.*?\/>/;e.fullTripleSlashAMDReferencePathRegEx=/^(\/\/\/\s*<amd-dependency\s+path\s*=\s*)('|")(.+?)\2.*?\/>/;var K=/^(\/\/\/\s*<reference\s+no-default-lib\s*=\s*)('|")(.+?)\2\s*\/>/;function W(n){if(172<=n.kind&&n.kind<=195)return!0;switch(n.kind){case 128:case 152:case 144:case 155:case 147:case 131:case 148:case 145:case 150:case 141:return!0;case 113:return 212!==n.parent.kind;case 223:return!St(n);case 159:return 190===n.parent.kind||185===n.parent.kind;case 78:(157===n.parent.kind&&n.parent.right===n||201===n.parent.kind&&n.parent.name===n)&&(n=n.parent),e.Debug.assert(78===n.kind||157===n.kind||201===n.kind,"'node' was expected to be a qualified name, identifier or property access in 'isPartOfTypeNode'.");case 157:case 201:case 107:var t=n.parent;if(176===t.kind)return!1;if(195===t.kind)return!t.isTypeOf;if(172<=t.kind&&t.kind<=195)return!0;switch(t.kind){case 223:return!St(t);case 159:case 330:return n===t.constraint;case 163:case 162:case 160:case 249:return n===t.type;case 251:case 208:case 209:case 166:case 165:case 164:case 167:case 168:return n===t.type;case 169:case 170:case 171:case 206:return n===t.type;case 203:case 204:return e.contains(t.typeArguments,n);case 205:return!1}}return!1}function z(e){if(e)switch(e.kind){case 198:case 291:case 160:case 288:case 163:case 162:case 289:case 249:return!0}return!1}function q(e){return 250===e.parent.kind&&232===e.parent.parent.kind}function J(e,n,t){return e.properties.filter((function(e){if(288===e.kind){var r=R(e.name);return n===r||!!t&&t===r}return!1}))}function X(n){if(n&&n.statements.length){var t=n.statements[0].expression;return e.tryCast(t,e.isObjectLiteralExpression)}}function Y(n,t){var r=X(n);return r?J(r,t):e.emptyArray}function Q(n,t){for(e.Debug.assert(297!==n.kind);;){if(!(n=n.parent))return e.Debug.fail();switch(n.kind){case 158:if(e.isClassLike(n.parent.parent))return n;n=n.parent;break;case 161:160===n.parent.kind&&e.isClassElement(n.parent.parent)?n=n.parent.parent:e.isClassElement(n.parent)&&(n=n.parent);break;case 209:if(!t)continue;case 251:case 208:case 256:case 163:case 162:case 165:case 164:case 166:case 167:case 168:case 169:case 170:case 171:case 255:case 297:return n}}}function Z(e){var n=e.kind;return(201===n||202===n)&&105===e.expression.kind}function $(n,t,r){if(e.isNamedDeclaration(n)&&e.isPrivateIdentifier(n.name))return!1;switch(n.kind){case 252:return!0;case 163:return 252===t.kind;case 167:case 168:case 165:return void 0!==n.body&&252===t.kind;case 160:return void 0!==t.body&&(166===t.kind||165===t.kind||168===t.kind)&&252===r.kind}return!1}function ee(e,n,t){return void 0!==e.decorators&&$(e,n,t)}function ne(e,n,t){return ee(e,n,t)||te(e,n)}function te(n,t){switch(n.kind){case 252:return e.some(n.members,(function(e){return ne(e,n,t)}));case 165:case 168:return e.some(n.parameters,(function(e){return ee(e,n,t)}));default:return!1}}function re(e){var n=e.parent;return(275===n.kind||274===n.kind||276===n.kind)&&n.tagName===e}function ae(e){switch(e.kind){case 105:case 103:case 109:case 94:case 13:case 199:case 200:case 201:case 202:case 203:case 204:case 205:case 224:case 206:case 225:case 207:case 208:case 221:case 209:case 212:case 210:case 211:case 214:case 215:case 216:case 217:case 220:case 218:case 222:case 273:case 274:case 277:case 219:case 213:case 226:return!0;case 157:for(;157===e.parent.kind;)e=e.parent;return 176===e.parent.kind||re(e);case 78:if(176===e.parent.kind||re(e))return!0;case 8:case 9:case 10:case 14:case 107:return ie(e);default:return!1}}function ie(e){var n=e.parent;switch(n.kind){case 249:case 160:case 163:case 162:case 291:case 288:case 198:return n.initializer===e;case 233:case 234:case 235:case 236:case 242:case 243:case 244:case 284:case 246:return n.expression===e;case 237:var t=n;return t.initializer===e&&250!==t.initializer.kind||t.condition===e||t.incrementor===e;case 238:case 239:var r=n;return r.initializer===e&&250!==r.initializer.kind||r.expression===e;case 206:case 224:case 228:case 158:return e===n.expression;case 161:case 283:case 282:case 290:return!0;case 223:return n.expression===e&&St(n);case 289:return n.objectAssignmentInitializer===e;default:return ae(n)}}function oe(e){for(;157===e.kind||78===e.kind;)e=e.parent;return 176===e.kind}function se(e){return 260===e.kind&&272===e.moduleReference.kind}function ce(e){return le(e)}function le(e){return!!e&&!!(131072&e.flags)}function ue(n,t){if(203!==n.kind)return!1;var r=n,a=r.expression,i=r.arguments;if(78!==a.kind||"require"!==a.escapedText)return!1;if(1!==i.length)return!1;var o=i[0];return!t||e.isStringLiteralLike(o)}function de(n,t){return 198===n.kind&&(n=n.parent.parent),e.isVariableDeclaration(n)&&!!n.initializer&&ue(Kt(n.initializer),t)}function pe(n){return e.isBinaryExpression(n)||Ht(n)||e.isIdentifier(n)||e.isCallExpression(n)}function me(n){return le(n)&&n.initializer&&e.isBinaryExpression(n.initializer)&&(56===n.initializer.operatorToken.kind||60===n.initializer.operatorToken.kind)&&n.name&&xt(n.name)&&_e(n.name,n.initializer.left)?n.initializer.right:n.initializer}function fe(n,t){if(e.isCallExpression(n)){var r=ze(n.expression);return 208===r.kind||209===r.kind?n:void 0}return 208===n.kind||221===n.kind||209===n.kind||e.isObjectLiteralExpression(n)&&(0===n.properties.length||t)?n:void 0}function _e(n,t){if(dn(n)&&dn(t))return pn(n)===pn(t);if(e.isIdentifier(n)&&Te(t)&&(107===t.expression.kind||e.isIdentifier(t.expression)&&("window"===t.expression.escapedText||"self"===t.expression.escapedText||"global"===t.expression.escapedText))){var r=De(t);return e.isPrivateIdentifier(r)&&e.Debug.fail("Unexpected PrivateIdentifier in name expression with literal-like access."),_e(n,r)}return!(!Te(n)||!Te(t))&&(Ie(n)===Ie(t)&&_e(n.expression,t.expression))}function ge(e){for(;Tt(e,!0);)e=e.right;return e}function ye(n){return e.isIdentifier(n)&&"exports"===n.escapedText}function ve(n){return e.isIdentifier(n)&&"module"===n.escapedText}function he(n){return(e.isPropertyAccessExpression(n)||Se(n))&&ve(n.expression)&&"exports"===Ie(n)}function be(n){var t=function(n){if(e.isCallExpression(n)){if(!Ee(n))return 0;var t=n.arguments[0];return ye(t)||he(t)?8:xe(t)&&"prototype"===Ie(t)?9:7}if(62!==n.operatorToken.kind||!Ht(n.left)||(r=ge(n),e.isVoidExpression(r)&&e.isNumericLiteral(r.expression)&&"0"===r.expression.text))return 0;var r;if(Ae(n.left.expression,!0)&&"prototype"===Ie(n.left)&&e.isObjectLiteralExpression(Re(n)))return 6;return Ne(n.left)}(n);return 5===t||le(n)?t:0}function Ee(n){return 3===e.length(n.arguments)&&e.isPropertyAccessExpression(n.expression)&&e.isIdentifier(n.expression.expression)&&"Object"===e.idText(n.expression.expression)&&"defineProperty"===e.idText(n.expression.name)&&an(n.arguments[1])&&Ae(n.arguments[0],!0)}function Te(n){return e.isPropertyAccessExpression(n)||Se(n)}function Se(n){return e.isElementAccessExpression(n)&&(an(n.argumentExpression)||ln(n.argumentExpression))}function xe(n,t){return e.isPropertyAccessExpression(n)&&(!t&&107===n.expression.kind||e.isIdentifier(n.name)&&Ae(n.expression,!0))||Le(n,t)}function Le(e,n){return Se(e)&&(!n&&107===e.expression.kind||xt(e.expression)||xe(e.expression,!0))}function Ae(e,n){return xt(e)||xe(e,n)}function De(n){return e.isPropertyAccessExpression(n)?n.name:n.argumentExpression}function Ce(n){if(e.isPropertyAccessExpression(n))return n.name;var t=ze(n.argumentExpression);return e.isNumericLiteral(t)||e.isStringLiteralLike(t)?t:n}function Ie(n){var t=Ce(n);if(t){if(e.isIdentifier(t))return t.escapedText;if(e.isStringLiteralLike(t)||e.isNumericLiteral(t))return e.escapeLeadingUnderscores(t.text)}if(e.isElementAccessExpression(n)&&ln(n.argumentExpression))return mn(e.idText(n.argumentExpression.name))}function Ne(n){if(107===n.expression.kind)return 4;if(he(n))return 2;if(Ae(n.expression,!0)){if(At(n.expression))return 3;for(var t=n;!e.isIdentifier(t.expression);)t=t.expression;var r=t.expression;if(("exports"===r.escapedText||"module"===r.escapedText&&"exports"===Ie(t))&&xe(n))return 1;if(Ae(n,!0)||e.isElementAccessExpression(n)&&cn(n))return 5}return 0}function Re(n){for(;e.isBinaryExpression(n.right);)n=n.right;return n.right}function ke(n){switch(n.parent.kind){case 261:case 267:return n.parent;case 272:return n.parent.parent;case 203:return G(n.parent)||ue(n.parent,!1)?n.parent:void 0;case 191:return e.Debug.assert(e.isStringLiteral(n)),e.tryCast(n.parent.parent,e.isImportTypeNode);default:return}}function Oe(e){return 331===e.kind||324===e.kind||325===e.kind}function Pe(n){return e.isExpressionStatement(n)&&e.isBinaryExpression(n.expression)&&0!==be(n.expression)&&e.isBinaryExpression(n.expression.right)&&(56===n.expression.right.operatorToken.kind||60===n.expression.right.operatorToken.kind)?n.expression.right.right:void 0}function Me(e){switch(e.kind){case 232:var n=we(e);return n&&n.initializer;case 163:case 288:return e.initializer}}function we(n){return e.isVariableStatement(n)?e.firstOrUndefined(n.declarationList.declarations):void 0}function Fe(n){return e.isModuleDeclaration(n)&&n.body&&256===n.body.kind?n.body:void 0}function Ge(n){var t=n.parent;return 288===t.kind||266===t.kind||163===t.kind||233===t.kind&&201===n.kind||Fe(t)||e.isBinaryExpression(n)&&62===n.operatorToken.kind?t:t.parent&&(we(t.parent)===n||e.isBinaryExpression(t)&&62===t.operatorToken.kind)?t.parent:t.parent&&t.parent.parent&&(we(t.parent.parent)||Me(t.parent.parent)===n||Pe(t.parent.parent))?t.parent.parent:void 0}function Be(n){var t=Ue(n);return t&&e.isFunctionLike(t)?t:void 0}function Ue(n){var t=Ve(n);return Pe(t)||function(n){return e.isExpressionStatement(n)&&e.isBinaryExpression(n.expression)&&62===n.expression.operatorToken.kind?ge(n.expression):void 0}(t)||Me(t)||we(t)||Fe(t)||t}function Ve(n){return e.Debug.checkDefined(e.findAncestor(n.parent,e.isJSDoc)).parent}function je(n){var t=e.isJSDocParameterTag(n)?n.typeExpression&&n.typeExpression.type:n.type;return void 0!==n.dotDotDotToken||!!t&&309===t.kind}function He(e){for(var n=e.parent;;){switch(n.kind){case 216:var t=n.operatorToken.kind;return ht(t)&&n.left===e?62===t||vt(t)?1:2:0;case 214:case 215:var r=n.operator;return 45===r||46===r?2:0;case 238:case 239:return n.initializer===e?1:0;case 207:case 199:case 220:case 225:e=n;break;case 289:if(n.name!==e)return 0;e=n.parent;break;case 288:if(n.name===e)return 0;e=n.parent;break;default:return 0}n=e.parent}}function Ke(e,n){for(;e&&e.kind===n;)e=e.parent;return e}function We(e){return Ke(e,207)}function ze(n){return e.skipOuterExpressions(n,1)}function qe(n){return xt(n)||e.isClassExpression(n)}function Je(e){return qe(Xe(e))}function Xe(n){return e.isExportAssignment(n)?n.expression:n.right}function Ye(n){var t=Qe(n);if(t&&le(n)){var r=e.getJSDocAugmentsTag(n);if(r)return r.class}return t}function Qe(e){var n=en(e.heritageClauses,93);return n&&n.types.length>0?n.types[0]:void 0}function Ze(n){if(le(n))return e.getJSDocImplementsTags(n).map((function(e){return e.class}));var t=en(n.heritageClauses,116);return null==t?void 0:t.types}function $e(e){var n=en(e.heritageClauses,93);return n?n.types:void 0}function en(e,n){if(e)for(var t=0,r=e;t<r.length;t++){var a=r[t];if(a.token===n)return a}}function nn(e){return 80<=e&&e<=156}function tn(e){return 125<=e&&e<=156}function rn(e){return nn(e)&&!tn(e)}function an(n){return e.isStringLiteralLike(n)||e.isNumericLiteral(n)}function on(n){return e.isPrefixUnaryExpression(n)&&(39===n.operator||40===n.operator)&&e.isNumericLiteral(n.operand)}function sn(n){var t=e.getNameOfDeclaration(n);return!!t&&cn(t)}function cn(n){if(158!==n.kind&&202!==n.kind)return!1;var t=e.isElementAccessExpression(n)?ze(n.argumentExpression):n.expression;return!an(t)&&!on(t)&&!ln(t)}function ln(n){return e.isPropertyAccessExpression(n)&&fn(n.expression)}function un(n){switch(n.kind){case 78:case 79:return n.escapedText;case 10:case 8:return e.escapeLeadingUnderscores(n.text);case 158:var t=n.expression;return ln(t)?mn(e.idText(t.name)):an(t)?e.escapeLeadingUnderscores(t.text):on(t)?40===t.operator?e.tokenToString(t.operator)+t.operand.text:t.operand.text:void 0;default:return e.Debug.assertNever(n)}}function dn(e){switch(e.kind){case 78:case 10:case 14:case 8:return!0;default:return!1}}function pn(n){return e.isIdentifierOrPrivateIdentifier(n)?e.idText(n):n.text}function mn(e){return"__@"+e}function fn(e){return 78===e.kind&&"Symbol"===e.escapedText}function _n(e){for(;198===e.kind;)e=e.parent.parent;return e}function gn(e){return Wr(e.pos)||Wr(e.end)}function yn(e,n,t){switch(e){case 204:return t?0:1;case 214:case 211:case 212:case 210:case 213:case 217:case 219:return 1;case 216:switch(n){case 42:case 62:case 63:case 64:case 66:case 65:case 67:case 68:case 69:case 70:case 71:case 72:case 77:case 73:case 74:case 75:case 76:return 1}}return 0}function vn(e){return 216===e.kind?e.operatorToken.kind:214===e.kind||215===e.kind?e.operator:e.kind}function hn(e,n,t){switch(e){case 337:return 0;case 220:return 1;case 219:return 2;case 217:return 4;case 216:switch(n){case 27:return 0;case 62:case 63:case 64:case 66:case 65:case 67:case 68:case 69:case 70:case 71:case 72:case 77:case 73:case 74:case 75:case 76:return 3;default:return bn(n)}case 214:case 211:case 212:case 210:case 213:return 16;case 215:return 17;case 203:return 18;case 204:return t?19:18;case 205:case 201:case 202:return 19;case 107:case 105:case 78:case 103:case 109:case 94:case 8:case 9:case 10:case 199:case 200:case 208:case 209:case 221:case 13:case 14:case 218:case 207:case 222:case 273:case 274:case 277:return 20;default:return-1}}function bn(e){switch(e){case 60:return 4;case 56:return 5;case 55:return 6;case 51:return 7;case 52:return 8;case 50:return 9;case 34:case 35:case 36:case 37:return 10;case 29:case 31:case 32:case 33:case 101:case 100:case 126:return 11;case 47:case 48:case 49:return 12;case 39:case 40:return 13;case 41:case 43:case 44:return 14;case 42:return 15}return-1}e.isPartOfTypeNode=W,e.isChildOfNodeWithKind=function(e,n){for(;e;){if(e.kind===n)return!0;e=e.parent}return!1},e.forEachReturnStatement=function(n,t){return function n(r){switch(r.kind){case 242:return t(r);case 258:case 230:case 234:case 235:case 236:case 237:case 238:case 239:case 243:case 244:case 284:case 285:case 245:case 247:case 287:return e.forEachChild(r,n)}}(n)},e.forEachYieldExpression=function(n,t){return function n(r){switch(r.kind){case 219:t(r);var a=r.expression;return void(a&&n(a));case 255:case 253:case 256:case 254:return;default:if(e.isFunctionLike(r)){if(r.name&&158===r.name.kind)return void n(r.name.expression)}else W(r)||e.forEachChild(r,n)}}(n)},e.getRestParameterElementType=function(n){return n&&178===n.kind?n.elementType:n&&173===n.kind?e.singleOrUndefined(n.typeArguments):void 0},e.getMembersOfDeclaration=function(e){switch(e.kind){case 253:case 252:case 221:case 177:return e.members;case 200:return e.properties}},e.isVariableLike=z,e.isVariableLikeOrAccessor=function(n){return z(n)||e.isAccessor(n)},e.isVariableDeclarationInVariableStatement=q,e.isValidESSymbolDeclaration=function(n){return e.isVariableDeclaration(n)?F(n)&&e.isIdentifier(n.name)&&q(n):e.isPropertyDeclaration(n)?ct(n)&&st(n):e.isPropertySignature(n)&&ct(n)},e.introducesArgumentsExoticObject=function(e){switch(e.kind){case 165:case 164:case 166:case 167:case 168:case 251:case 208:return!0}return!1},e.unwrapInnermostStatementOfLabel=function(e,n){for(;;){if(n&&n(e),245!==e.statement.kind)return e.statement;e=e.statement}},e.isFunctionBlock=function(n){return n&&230===n.kind&&e.isFunctionLike(n.parent)},e.isObjectLiteralMethod=function(e){return e&&165===e.kind&&200===e.parent.kind},e.isObjectLiteralOrClassExpressionMethod=function(e){return 165===e.kind&&(200===e.parent.kind||221===e.parent.kind)},e.isIdentifierTypePredicate=function(e){return e&&1===e.kind},e.isThisTypePredicate=function(e){return e&&0===e.kind},e.getPropertyAssignment=J,e.getTsConfigObjectLiteralExpression=X,e.getTsConfigPropArrayElementValue=function(n,t,r){return e.firstDefined(Y(n,t),(function(n){return e.isArrayLiteralExpression(n.initializer)?e.find(n.initializer.elements,(function(n){return e.isStringLiteral(n)&&n.text===r})):void 0}))},e.getTsConfigPropArray=Y,e.getContainingFunction=function(n){return e.findAncestor(n.parent,e.isFunctionLike)},e.getContainingFunctionDeclaration=function(n){return e.findAncestor(n.parent,e.isFunctionLikeDeclaration)},e.getContainingClass=function(n){return e.findAncestor(n.parent,e.isClassLike)},e.getThisContainer=Q,e.isInTopLevelContext=function(n){e.isIdentifier(n)&&(e.isClassDeclaration(n.parent)||e.isFunctionDeclaration(n.parent))&&n.parent.name===n&&(n=n.parent);var t=Q(n,!0);return e.isSourceFile(t)},e.getNewTargetContainer=function(e){var n=Q(e,!1);if(n)switch(n.kind){case 166:case 251:case 208:return n}},e.getSuperContainer=function(n,t){for(;;){if(!(n=n.parent))return n;switch(n.kind){case 158:n=n.parent;break;case 251:case 208:case 209:if(!t)continue;case 163:case 162:case 165:case 164:case 166:case 167:case 168:return n;case 161:160===n.parent.kind&&e.isClassElement(n.parent.parent)?n=n.parent.parent:e.isClassElement(n.parent)&&(n=n.parent)}}},e.getImmediatelyInvokedFunctionExpression=function(e){if(208===e.kind||209===e.kind){for(var n=e,t=e.parent;207===t.kind;)n=t,t=t.parent;if(203===t.kind&&t.expression===n)return t}},e.isSuperOrSuperProperty=function(e){return 105===e.kind||Z(e)},e.isSuperProperty=Z,e.isThisProperty=function(e){var n=e.kind;return(201===n||202===n)&&107===e.expression.kind},e.isThisInitializedDeclaration=function(n){var t;return!!n&&e.isVariableDeclaration(n)&&107===(null===(t=n.initializer)||void 0===t?void 0:t.kind)},e.getEntityNameFromTypeNode=function(e){switch(e.kind){case 173:return e.typeName;case 223:return xt(e.expression)?e.expression:void 0;case 78:case 157:return e}},e.getInvokedExpression=function(e){switch(e.kind){case 205:return e.tag;case 275:case 274:return e.tagName;default:return e.expression}},e.nodeCanBeDecorated=$,e.nodeIsDecorated=ee,e.nodeOrChildIsDecorated=ne,e.childIsDecorated=te,e.isJSXTagName=re,e.isExpressionNode=ae,e.isInExpressionContext=ie,e.isPartOfTypeQuery=oe,e.isExternalModuleImportEqualsDeclaration=se,e.getExternalModuleImportEqualsDeclarationExpression=function(n){return e.Debug.assert(se(n)),n.moduleReference.expression},e.getExternalModuleRequireArgument=function(e){return de(e,!0)&&Kt(e.initializer).arguments[0]},e.isInternalModuleImportEqualsDeclaration=function(e){return 260===e.kind&&272!==e.moduleReference.kind},e.isSourceFileJS=ce,e.isSourceFileNotJS=function(e){return!le(e)},e.isInJSFile=le,e.isInJsonFile=function(e){return!!e&&!!(33554432&e.flags)},e.isSourceFileNotJson=function(e){return!w(e)},e.isInJSDoc=function(e){return!!e&&!!(4194304&e.flags)},e.isJSDocIndexSignature=function(n){return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)&&"Object"===n.typeName.escapedText&&n.typeArguments&&2===n.typeArguments.length&&(147===n.typeArguments[0].kind||144===n.typeArguments[0].kind)},e.isRequireCall=ue,e.isRequireVariableDeclaration=de,e.isRequireVariableStatement=function(n,t){return void 0===t&&(t=!0),e.isVariableStatement(n)&&n.declarationList.declarations.length>0&&e.every(n.declarationList.declarations,(function(e){return de(e,t)}))},e.isSingleOrDoubleQuote=function(e){return 39===e||34===e},e.isStringDoubleQuoted=function(e,n){return 34===g(n,e).charCodeAt(0)},e.isAssignmentDeclaration=pe,e.getEffectiveInitializer=me,e.getDeclaredExpandoInitializer=function(e){var n=me(e);return n&&fe(n,At(e.name))},e.getAssignedExpandoInitializer=function(n){if(n&&n.parent&&e.isBinaryExpression(n.parent)&&62===n.parent.operatorToken.kind){var t=At(n.parent.left);return fe(n.parent.right,t)||function(n,t,r){var a=e.isBinaryExpression(t)&&(56===t.operatorToken.kind||60===t.operatorToken.kind)&&fe(t.right,r);if(a&&_e(n,t.left))return a}(n.parent.left,n.parent.right,t)}if(n&&e.isCallExpression(n)&&Ee(n)){var r=function(n,t){return e.forEach(n.properties,(function(n){return e.isPropertyAssignment(n)&&e.isIdentifier(n.name)&&"value"===n.name.escapedText&&n.initializer&&fe(n.initializer,t)}))}(n.arguments[2],"prototype"===n.arguments[1].text);if(r)return r}},e.getExpandoInitializer=fe,e.isDefaultedExpandoInitializer=function(n){var t=e.isVariableDeclaration(n.parent)?n.parent.name:e.isBinaryExpression(n.parent)&&62===n.parent.operatorToken.kind?n.parent.left:void 0;return t&&fe(n.right,At(t))&&xt(t)&&_e(t,n.left)},e.getNameOfExpando=function(n){if(e.isBinaryExpression(n.parent)){var t=56!==n.parent.operatorToken.kind&&60!==n.parent.operatorToken.kind||!e.isBinaryExpression(n.parent.parent)?n.parent:n.parent.parent;if(62===t.operatorToken.kind&&e.isIdentifier(t.left))return t.left}else if(e.isVariableDeclaration(n.parent))return n.parent.name},e.isSameEntityName=_e,e.getRightMostAssignedExpression=ge,e.isExportsIdentifier=ye,e.isModuleIdentifier=ve,e.isModuleExportsAccessExpression=he,e.getAssignmentDeclarationKind=be,e.isBindableObjectDefinePropertyCall=Ee,e.isLiteralLikeAccess=Te,e.isLiteralLikeElementAccess=Se,e.isBindableStaticAccessExpression=xe,e.isBindableStaticElementAccessExpression=Le,e.isBindableStaticNameExpression=Ae,e.getNameOrArgument=De,e.getElementOrPropertyAccessArgumentExpressionOrName=Ce,e.getElementOrPropertyAccessName=Ie,e.getAssignmentDeclarationPropertyAccessKind=Ne,e.getInitializerOfBinaryExpression=Re,e.isPrototypePropertyAssignment=function(n){return e.isBinaryExpression(n)&&3===be(n)},e.isSpecialPropertyDeclaration=function(n){return le(n)&&n.parent&&233===n.parent.kind&&(!e.isElementAccessExpression(n)||Se(n))&&!!e.getJSDocTypeTag(n.parent)},e.setValueDeclaration=function(e,n){var t=e.valueDeclaration;(!t||(!(8388608&n.flags)||8388608&t.flags)&&pe(t)&&!pe(n)||t.kind!==n.kind&&S(t))&&(e.valueDeclaration=n)},e.isFunctionSymbol=function(n){if(!n||!n.valueDeclaration)return!1;var t=n.valueDeclaration;return 251===t.kind||e.isVariableDeclaration(t)&&t.initializer&&e.isFunctionLike(t.initializer)},e.importFromModuleSpecifier=function(n){return ke(n)||e.Debug.failBadSyntaxKind(n.parent)},e.tryGetImportFromModuleSpecifier=ke,e.getExternalModuleName=function(n){switch(n.kind){case 261:case 267:return n.moduleSpecifier;case 260:return 272===n.moduleReference.kind?n.moduleReference.expression:void 0;case 195:return B(n)?n.argument.literal:void 0;default:return e.Debug.assertNever(n)}},e.getNamespaceDeclarationNode=function(n){switch(n.kind){case 261:return n.importClause&&e.tryCast(n.importClause.namedBindings,e.isNamespaceImport);case 260:return n;case 267:return n.exportClause&&e.tryCast(n.exportClause,e.isNamespaceExport);default:return e.Debug.assertNever(n)}},e.isDefaultImport=function(e){return 261===e.kind&&!!e.importClause&&!!e.importClause.name},e.forEachImportClauseDeclaration=function(n,t){var r;if(n.name&&(r=t(n)))return r;if(n.namedBindings&&(r=e.isNamespaceImport(n.namedBindings)?t(n.namedBindings):e.forEach(n.namedBindings.elements,t)))return r},e.hasQuestionToken=function(e){if(e)switch(e.kind){case 160:case 165:case 164:case 289:case 288:case 163:case 162:return void 0!==e.questionToken}return!1},e.isJSDocConstructSignature=function(n){var t=e.isJSDocFunctionType(n)?e.firstOrUndefined(n.parameters):void 0,r=e.tryCast(t&&t.name,e.isIdentifier);return!!r&&"new"===r.escapedText},e.isJSDocTypeAlias=Oe,e.isTypeAlias=function(n){return Oe(n)||e.isTypeAliasDeclaration(n)},e.getSingleInitializerOfVariableStatementOrPropertyDeclaration=Me,e.getSingleVariableOfVariableStatement=we,e.getJSDocCommentsAndTags=function(n,t){var r;z(n)&&e.hasInitializer(n)&&e.hasJSDocNodes(n.initializer)&&(r=e.append(r,e.last(n.initializer.jsDoc)));for(var a=n;a&&a.parent;){if(e.hasJSDocNodes(a)&&(r=e.append(r,e.last(a.jsDoc))),160===a.kind){r=e.addRange(r,(t?e.getJSDocParameterTagsNoCache:e.getJSDocParameterTags)(a));break}if(159===a.kind){r=e.addRange(r,(t?e.getJSDocTypeParameterTagsNoCache:e.getJSDocTypeParameterTags)(a));break}a=Ge(a)}return r||e.emptyArray},e.getParameterSymbolFromJSDoc=function(n){if(n.symbol)return n.symbol;if(e.isIdentifier(n.name)){var t=n.name.escapedText,r=Be(n);if(r){var a=e.find(r.parameters,(function(e){return 78===e.name.kind&&e.name.escapedText===t}));return a&&a.symbol}}},e.getHostSignatureFromJSDoc=Be,e.getEffectiveJSDocHost=Ue,e.getJSDocHost=Ve,e.getTypeParameterFromJsDoc=function(n){var t=n.name.escapedText,r=n.parent.parent.parent.typeParameters;return r&&e.find(r,(function(e){return e.name.escapedText===t}))},e.hasRestParameter=function(n){var t=e.lastOrUndefined(n.parameters);return!!t&&je(t)},e.isRestParameter=je,e.hasTypeArguments=function(e){return!!e.typeArguments},function(e){e[e.None=0]="None",e[e.Definite=1]="Definite",e[e.Compound=2]="Compound"}(e.AssignmentKind||(e.AssignmentKind={})),e.getAssignmentTargetKind=He,e.isAssignmentTarget=function(e){return 0!==He(e)},e.isNodeWithPossibleHoistedDeclaration=function(e){switch(e.kind){case 230:case 232:case 243:case 234:case 244:case 258:case 284:case 285:case 245:case 237:case 238:case 239:case 235:case 236:case 247:case 287:return!0}return!1},e.isValueSignatureDeclaration=function(n){return e.isFunctionExpression(n)||e.isArrowFunction(n)||e.isMethodOrAccessor(n)||e.isFunctionDeclaration(n)||e.isConstructorDeclaration(n)},e.walkUpParenthesizedTypes=function(e){return Ke(e,186)},e.walkUpParenthesizedExpressions=We,e.walkUpParenthesizedTypesAndGetParentAndChild=function(e){for(var n;e&&186===e.kind;)n=e,e=e.parent;return[n,e]},e.skipParentheses=ze,e.isDeleteTarget=function(e){return(201===e.kind||202===e.kind)&&((e=We(e.parent))&&210===e.kind)},e.isNodeDescendantOf=function(e,n){for(;e;){if(e===n)return!0;e=e.parent}return!1},e.isDeclarationName=function(n){return!e.isSourceFile(n)&&!e.isBindingPattern(n)&&e.isDeclaration(n.parent)&&n.parent.name===n},e.getDeclarationFromName=function(n){var t=n.parent;switch(n.kind){case 10:case 14:case 8:if(e.isComputedPropertyName(t))return t.parent;case 78:if(e.isDeclaration(t))return t.name===n?t:void 0;if(e.isQualifiedName(t)){var r=t.parent;return e.isJSDocParameterTag(r)&&r.name===t?r:void 0}var a=t.parent;return e.isBinaryExpression(a)&&0!==be(a)&&(a.left.symbol||a.symbol)&&e.getNameOfDeclaration(a)===n?a:void 0;case 79:return e.isDeclaration(t)&&t.name===n?t:void 0;default:return}},e.isLiteralComputedPropertyDeclarationName=function(n){return an(n)&&158===n.parent.kind&&e.isDeclaration(n.parent.parent)},e.isIdentifierName=function(e){var n=e.parent;switch(n.kind){case 163:case 162:case 165:case 164:case 167:case 168:case 291:case 288:case 201:return n.name===e;case 157:return n.right===e;case 198:case 265:return n.propertyName===e;case 270:case 280:return!0}return!1},e.isAliasSymbolDeclaration=function(n){return 260===n.kind||259===n.kind||262===n.kind&&!!n.name||263===n.kind||269===n.kind||265===n.kind||270===n.kind||266===n.kind&&Je(n)||e.isBinaryExpression(n)&&2===be(n)&&Je(n)||e.isPropertyAccessExpression(n)&&e.isBinaryExpression(n.parent)&&n.parent.left===n&&62===n.parent.operatorToken.kind&&qe(n.parent.right)||289===n.kind||288===n.kind&&qe(n.initializer)},e.getAliasDeclarationFromName=function e(n){switch(n.parent.kind){case 262:case 265:case 263:case 270:case 266:case 260:return n.parent;case 157:do{n=n.parent}while(157===n.parent.kind);return e(n)}},e.isAliasableExpression=qe,e.exportAssignmentIsAlias=Je,e.getExportAssignmentExpression=Xe,e.getPropertyAssignmentAliasLikeExpression=function(e){return 289===e.kind?e.name:288===e.kind?e.initializer:e.parent.right},e.getEffectiveBaseTypeNode=Ye,e.getClassExtendsHeritageElement=Qe,e.getEffectiveImplementsTypeNodes=Ze,e.getAllSuperTypeNodes=function(n){return e.isInterfaceDeclaration(n)?$e(n)||e.emptyArray:e.isClassLike(n)&&e.concatenate(e.singleElementArray(Ye(n)),Ze(n))||e.emptyArray},e.getInterfaceBaseTypeNodes=$e,e.getHeritageClause=en,e.getAncestor=function(e,n){for(;e;){if(e.kind===n)return e;e=e.parent}},e.isKeyword=nn,e.isContextualKeyword=tn,e.isNonContextualKeyword=rn,e.isFutureReservedKeyword=function(e){return 116<=e&&e<=124},e.isStringANonContextualKeyword=function(n){var t=e.stringToToken(n);return void 0!==t&&rn(t)},e.isStringAKeyword=function(n){var t=e.stringToToken(n);return void 0!==t&&nn(t)},e.isIdentifierANonContextualKeyword=function(e){var n=e.originalKeywordKind;return!!n&&!tn(n)},e.isTrivia=function(e){return 2<=e&&e<=7},function(e){e[e.Normal=0]="Normal",e[e.Generator=1]="Generator",e[e.Async=2]="Async",e[e.Invalid=4]="Invalid",e[e.AsyncGenerator=3]="AsyncGenerator"}(e.FunctionFlags||(e.FunctionFlags={})),e.getFunctionFlags=function(e){if(!e)return 4;var n=0;switch(e.kind){case 251:case 208:case 165:e.asteriskToken&&(n|=1);case 209:ot(e,256)&&(n|=2)}return e.body||(n|=4),n},e.isAsyncFunction=function(e){switch(e.kind){case 251:case 208:case 209:case 165:return void 0!==e.body&&void 0===e.asteriskToken&&ot(e,256)}return!1},e.isStringOrNumericLiteralLike=an,e.isSignedNumericLiteral=on,e.hasDynamicName=sn,e.isDynamicName=cn,e.isWellKnownSymbolSyntactically=ln,e.getPropertyNameForPropertyNameNode=un,e.isPropertyNameLiteral=dn,e.getTextOfIdentifierOrLiteral=pn,e.getEscapedTextOfIdentifierOrLiteral=function(n){return e.isIdentifierOrPrivateIdentifier(n)?n.escapedText:e.escapeLeadingUnderscores(n.text)},e.getPropertyNameForUniqueESSymbol=function(n){return"__@"+e.getSymbolId(n)+"@"+n.escapedName},e.getPropertyNameForKnownSymbolName=mn,e.getSymbolNameForPrivateIdentifier=function(n,t){return"__#"+e.getSymbolId(n)+"@"+t},e.isKnownSymbol=function(n){return e.startsWith(n.escapedName,"__@")},e.isESSymbolIdentifier=fn,e.isPushOrUnshiftIdentifier=function(e){return"push"===e.escapedText||"unshift"===e.escapedText},e.isParameterDeclaration=function(e){return 160===_n(e).kind},e.getRootDeclaration=_n,e.nodeStartsNewLexicalEnvironment=function(e){var n=e.kind;return 166===n||208===n||251===n||209===n||165===n||167===n||168===n||256===n||297===n},e.nodeIsSynthesized=gn,e.getOriginalSourceFile=function(n){return e.getParseTreeNode(n,e.isSourceFile)||n},function(e){e[e.Left=0]="Left",e[e.Right=1]="Right"}(e.Associativity||(e.Associativity={})),e.getExpressionAssociativity=function(e){var n=vn(e),t=204===e.kind&&void 0!==e.arguments;return yn(e.kind,n,t)},e.getOperatorAssociativity=yn,e.getExpressionPrecedence=function(e){var n=vn(e),t=204===e.kind&&void 0!==e.arguments;return hn(e.kind,n,t)},e.getOperator=vn,function(e){e[e.Comma=0]="Comma",e[e.Spread=1]="Spread",e[e.Yield=2]="Yield",e[e.Assignment=3]="Assignment",e[e.Conditional=4]="Conditional",e[e.Coalesce=4]="Coalesce",e[e.LogicalOR=5]="LogicalOR",e[e.LogicalAND=6]="LogicalAND",e[e.BitwiseOR=7]="BitwiseOR",e[e.BitwiseXOR=8]="BitwiseXOR",e[e.BitwiseAND=9]="BitwiseAND",e[e.Equality=10]="Equality",e[e.Relational=11]="Relational",e[e.Shift=12]="Shift",e[e.Additive=13]="Additive",e[e.Multiplicative=14]="Multiplicative",e[e.Exponentiation=15]="Exponentiation",e[e.Unary=16]="Unary",e[e.Update=17]="Update",e[e.LeftHandSide=18]="LeftHandSide",e[e.Member=19]="Member",e[e.Primary=20]="Primary",e[e.Highest=20]="Highest",e[e.Lowest=0]="Lowest",e[e.Invalid=-1]="Invalid"}(e.OperatorPrecedence||(e.OperatorPrecedence={})),e.getOperatorPrecedence=hn,e.getBinaryOperatorPrecedence=bn,e.getSemanticJsxChildren=function(n){return e.filter(n,(function(e){switch(e.kind){case 283:return!!e.expression;case 11:return!e.containsOnlyTriviaWhiteSpaces;default:return!0}}))},e.createDiagnosticCollection=function(){var n=[],t=[],r=new e.Map,a=!1;return{add:function(i){var o;i.file?(o=r.get(i.file.fileName))||(o=[],r.set(i.file.fileName,o),e.insertSorted(t,i.file.fileName,e.compareStringsCaseSensitive)):(a&&(a=!1,n=n.slice()),o=n);e.insertSorted(o,i,ir)},lookup:function(t){var a;a=t.file?r.get(t.file.fileName):n;if(!a)return;var i=e.binarySearch(a,t,e.identity,or);if(i>=0)return a[i];return},getGlobalDiagnostics:function(){return a=!0,n},getDiagnostics:function(a){if(a)return r.get(a)||[];var i=e.flatMapToMutable(t,(function(e){return r.get(e)}));if(!n.length)return i;return i.unshift.apply(i,n),i},reattachFileDiagnostics:function(n){e.forEach(r.get(n.fileName),(function(e){return e.file=n}))}}};var En=/\$\{/g;e.hasInvalidEscape=function(n){return n&&!!(e.isNoSubstitutionTemplateLiteral(n)?n.templateFlags:n.head.templateFlags||e.some(n.templateSpans,(function(e){return!!e.literal.templateFlags})))};var Tn=/[\\\"\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g,Sn=/[\\\'\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g,xn=/[\\`]/g,Ln=new e.Map(e.getEntries({"\t":"\\t","\v":"\\v","\f":"\\f","\b":"\\b","\r":"\\r","\n":"\\n","\\":"\\\\",'"':'\\"',"'":"\\'","`":"\\`","\u2028":"\\u2028","\u2029":"\\u2029","":"\\u0085"}));function An(e){return"\\u"+("0000"+e.toString(16).toUpperCase()).slice(-4)}function Dn(e,n,t){if(0===e.charCodeAt(0)){var r=t.charCodeAt(n+e.length);return r>=48&&r<=57?"\\x00":"\\0"}return Ln.get(e)||An(e.charCodeAt(0))}function Cn(e,n){var t=96===n?xn:39===n?Sn:Tn;return e.replace(t,Dn)}e.escapeString=Cn;var In=/[^\u0000-\u007F]/g;function Nn(e,n){return e=Cn(e,n),In.test(e)?e.replace(In,(function(e){return An(e.charCodeAt(0))})):e}e.escapeNonAsciiString=Nn;var Rn=/[\"\u0000-\u001f\u2028\u2029\u0085]/g,kn=/[\'\u0000-\u001f\u2028\u2029\u0085]/g,On=new e.Map(e.getEntries({'"':"&quot;","'":"&apos;"}));function Pn(e){return 0===e.charCodeAt(0)?"&#0;":On.get(e)||"&#x"+e.charCodeAt(0).toString(16).toUpperCase()+";"}function Mn(e,n){var t=39===n?kn:Rn;return e.replace(t,Pn)}e.escapeJsxAttributeString=Mn,e.stripQuotes=function(e){var n,t=e.length;return t>=2&&e.charCodeAt(0)===e.charCodeAt(t-1)&&(39===(n=e.charCodeAt(0))||34===n||96===n)?e.substring(1,t-1):e},e.isIntrinsicJsxName=function(n){var t=n.charCodeAt(0);return t>=97&&t<=122||e.stringContains(n,"-")};var wn=["","    "];function Fn(e){for(var n=wn[1],t=wn.length;t<=e;t++)wn.push(wn[t-1]+n);return wn[e]}function Gn(){return wn[1].length}function Bn(e){return!!e.useCaseSensitiveFileNames&&e.useCaseSensitiveFileNames()}function Un(e,n,t){return n.moduleName||Vn(e,n.fileName,t&&t.fileName)}function Vn(n,t,r){var a=function(e){return n.getCanonicalFileName(e)},i=e.toPath(r?e.getDirectoryPath(r):n.getCommonSourceDirectory(),n.getCurrentDirectory(),a),o=e.getNormalizedAbsolutePath(t,n.getCurrentDirectory()),s=Vr(e.getRelativePathToDirectoryOrUrl(i,o,i,a,!1));return r?e.ensurePathIsNonModuleName(s):s}function jn(e,n,t,r,a){var i=n.declarationDir||n.outDir;return Vr(i?zn(e,i,t,r,a):e)+".d.ts"}function Hn(e){return e.outFile||e.out}function Kn(e,n,t){return!(n.getCompilerOptions().noEmitForJsFiles&&ce(e))&&!e.isDeclarationFile&&!n.isSourceFileFromExternalLibrary(e)&&!(w(e)&&n.getResolvedProjectReferenceToRedirect(e.fileName))&&(t||!n.isSourceOfProjectReferenceRedirect(e.fileName))}function Wn(e,n,t){return zn(e,t,n.getCurrentDirectory(),n.getCommonSourceDirectory(),(function(e){return n.getCanonicalFileName(e)}))}function zn(n,t,r,a,i){var o=e.getNormalizedAbsolutePath(n,r);return o=0===i(o).indexOf(i(a))?o.substring(a.length):o,e.combinePaths(t,o)}function qn(n,t,r){n.length>e.getRootLength(n)&&!r(n)&&(qn(e.getDirectoryPath(n),t,r),t(n))}function Jn(n,t){return e.computeLineOfPosition(n,t)}function Xn(e){if(e&&e.parameters.length>0){var n=2===e.parameters.length&&Yn(e.parameters[0]);return e.parameters[n?1:0]}}function Yn(e){return Qn(e.name)}function Qn(e){return!!e&&78===e.kind&&Zn(e)}function Zn(e){return 107===e.originalKeywordKind}function $n(n){if(le(n)||!e.isFunctionDeclaration(n)){var t=n.type;return t||!le(n)?t:e.isJSDocPropertyLikeTag(n)?n.typeExpression&&n.typeExpression.type:e.getJSDocType(n)}}function et(e,n,t,r){nt(e,n,t.pos,r)}function nt(e,n,t,r){r&&r.length&&t!==r[0].pos&&Jn(e,t)!==Jn(e,r[0].pos)&&n.writeLine()}function tt(e,n,t,r,a,i,o,s){if(r&&r.length>0){a&&t.writeSpace(" ");for(var c=!1,l=0,u=r;l<u.length;l++){var d=u[l];c&&(t.writeSpace(" "),c=!1),s(e,n,t,d.pos,d.end,o),d.hasTrailingNewLine?t.writeLine():c=!0}c&&i&&t.writeSpace(" ")}}function rt(e,n,t,r,a,i){var o=Math.min(n,i-1),s=e.substring(a,o).replace(/^\s+|\s+$/g,"");s?(t.writeComment(s),o!==n&&t.writeLine()):t.rawWrite(r)}function at(n,t,r){for(var a=0;t<r&&e.isWhiteSpaceSingleLine(n.charCodeAt(t));t++)9===n.charCodeAt(t)?a+=Gn()-a%Gn():a++;return a}function it(e,n){return!!lt(e,n)}function ot(e,n){return!!ut(e,n)}function st(e){return ot(e,32)}function ct(e){return it(e,64)}function lt(e,n){return pt(e)&n}function ut(e,n){return mt(e)&n}function dt(e,n,t){return e.kind>=0&&e.kind<=156?0:(536870912&e.modifierFlagsCache||(e.modifierFlagsCache=536870912|_t(e)),!n||4096&e.modifierFlagsCache||!t&&!le(e)||!e.parent||(e.modifierFlagsCache|=4096|ft(e)),-536875009&e.modifierFlagsCache)}function pt(e){return dt(e,!0)}function mt(e){return dt(e,!1)}function ft(n){var t=0;return n.parent&&!e.isParameter(n)&&(le(n)&&(e.getJSDocPublicTagNoCache(n)&&(t|=4),e.getJSDocPrivateTagNoCache(n)&&(t|=8),e.getJSDocProtectedTagNoCache(n)&&(t|=16),e.getJSDocReadonlyTagNoCache(n)&&(t|=64)),e.getJSDocDeprecatedTagNoCache(n)&&(t|=8192)),t}function _t(e){var n=gt(e.modifiers);return(4&e.flags||78===e.kind&&e.isInJSDocNamespace)&&(n|=1),n}function gt(e){var n=0;if(e)for(var t=0,r=e;t<r.length;t++){n|=yt(r[t].kind)}return n}function yt(e){switch(e){case 123:return 32;case 122:return 4;case 121:return 16;case 120:return 8;case 125:return 128;case 92:return 1;case 133:return 2;case 84:return 2048;case 87:return 512;case 129:return 256;case 142:return 64}return 0}function vt(e){return 74===e||75===e||76===e}function ht(e){return e>=62&&e<=77}function bt(e){var n=Et(e);return n&&!n.isImplements?n.class:void 0}function Et(n){return e.isExpressionWithTypeArguments(n)&&e.isHeritageClause(n.parent)&&e.isClassLike(n.parent.parent)?{class:n.parent.parent,isImplements:116===n.parent.token}:void 0}function Tt(n,t){return e.isBinaryExpression(n)&&(t?62===n.operatorToken.kind:ht(n.operatorToken.kind))&&e.isLeftHandSideExpression(n.left)}function St(e){return void 0!==bt(e)}function xt(e){return 78===e.kind||Lt(e)}function Lt(n){return e.isPropertyAccessExpression(n)&&e.isIdentifier(n.name)&&xt(n.expression)}function At(e){return xe(e)&&"prototype"===Ie(e)}e.getIndentString=Fn,e.getIndentSize=Gn,e.createTextWriter=function(n){var t,r,a,i,o,s=!1;function c(n){var r=e.computeLineStarts(n);r.length>1?(i=i+r.length-1,o=t.length-n.length+e.last(r),a=o-t.length==0):a=!1}function l(e){e&&e.length&&(a&&(e=Fn(r)+e,a=!1),t+=e,c(e))}function u(e){e&&(s=!1),l(e)}function d(){t="",r=0,a=!0,i=0,o=0,s=!1}return d(),{write:u,rawWrite:function(e){void 0!==e&&(t+=e,c(e),s=!1)},writeLiteral:function(e){e&&e.length&&u(e)},writeLine:function(e){a&&!e||(i++,o=(t+=n).length,a=!0,s=!1)},increaseIndent:function(){r++},decreaseIndent:function(){r--},getIndent:function(){return r},getTextPos:function(){return t.length},getLine:function(){return i},getColumn:function(){return a?r*Gn():t.length-o},getText:function(){return t},isAtStartOfLine:function(){return a},hasTrailingComment:function(){return s},hasTrailingWhitespace:function(){return!!t.length&&e.isWhiteSpaceLike(t.charCodeAt(t.length-1))},clear:d,reportInaccessibleThisError:e.noop,reportPrivateInBaseOfClassExpression:e.noop,reportInaccessibleUniqueSymbolError:e.noop,trackSymbol:e.noop,writeKeyword:u,writeOperator:u,writeParameter:u,writeProperty:u,writePunctuation:u,writeSpace:u,writeStringLiteral:u,writeSymbol:function(e,n){return u(e)},writeTrailingSemicolon:u,writeComment:function(e){e&&(s=!0),l(e)},getTextPosWithWriteLine:function(){return a?t.length:t.length+n.length}}},e.getTrailingSemicolonDeferringWriter=function(e){var n=!1;function t(){n&&(e.writeTrailingSemicolon(";"),n=!1)}return __assign(__assign({},e),{writeTrailingSemicolon:function(){n=!0},writeLiteral:function(n){t(),e.writeLiteral(n)},writeStringLiteral:function(n){t(),e.writeStringLiteral(n)},writeSymbol:function(n,r){t(),e.writeSymbol(n,r)},writePunctuation:function(n){t(),e.writePunctuation(n)},writeKeyword:function(n){t(),e.writeKeyword(n)},writeOperator:function(n){t(),e.writeOperator(n)},writeParameter:function(n){t(),e.writeParameter(n)},writeSpace:function(n){t(),e.writeSpace(n)},writeProperty:function(n){t(),e.writeProperty(n)},writeComment:function(n){t(),e.writeComment(n)},writeLine:function(){t(),e.writeLine()},increaseIndent:function(){t(),e.increaseIndent()},decreaseIndent:function(){t(),e.decreaseIndent()}})},e.hostUsesCaseSensitiveFileNames=Bn,e.hostGetCanonicalFileName=function(n){return e.createGetCanonicalFileName(Bn(n))},e.getResolvedExternalModuleName=Un,e.getExternalModuleNameFromDeclaration=function(e,n,t){var r=n.getExternalModuleFileFromDeclaration(t);if(r&&!r.isDeclarationFile)return Un(e,r)},e.getExternalModuleNameFromPath=Vn,e.getOwnEmitOutputFilePath=function(e,n,t){var r=n.getCompilerOptions();return(r.outDir?Vr(Wn(e,n,r.outDir)):Vr(e))+t},e.getDeclarationEmitOutputFilePath=function(e,n){return jn(e,n.getCompilerOptions(),n.getCurrentDirectory(),n.getCommonSourceDirectory(),(function(e){return n.getCanonicalFileName(e)}))},e.getDeclarationEmitOutputFilePathWorker=jn,e.outFile=Hn,e.getPathsBasePath=function(n,t){var r,a;if(n.paths)return null!==(r=n.baseUrl)&&void 0!==r?r:e.Debug.checkDefined(n.pathsBasePath||(null===(a=t.getCurrentDirectory)||void 0===a?void 0:a.call(t)),"Encountered 'paths' without a 'baseUrl', config file, or host 'getCurrentDirectory'.")},e.getSourceFilesToEmit=function(n,t,r){var a=n.getCompilerOptions();if(Hn(a)){var i=lr(a),o=a.emitDeclarationOnly||i===e.ModuleKind.AMD||i===e.ModuleKind.System;return e.filter(n.getSourceFiles(),(function(t){return(o||!e.isExternalModule(t))&&Kn(t,n,r)}))}var s=void 0===t?n.getSourceFiles():[t];return e.filter(s,(function(e){return Kn(e,n,r)}))},e.sourceFileMayBeEmitted=Kn,e.getSourceFilePathInNewDir=Wn,e.getSourceFilePathInNewDirWorker=zn,e.writeFile=function(n,t,r,a,i,o){n.writeFile(r,a,i,(function(n){t.add(rr(e.Diagnostics.Could_not_write_file_0_Colon_1,r,n))}),o)},e.writeFileEnsuringDirectories=function(n,t,r,a,i,o){try{a(n,t,r)}catch(s){qn(e.getDirectoryPath(e.normalizePath(n)),i,o),a(n,t,r)}},e.getLineOfLocalPosition=function(n,t){var r=e.getLineStarts(n);return e.computeLineOfPosition(r,t)},e.getLineOfLocalPositionFromLineMap=Jn,e.getFirstConstructorWithBody=function(n){return e.find(n.members,(function(n){return e.isConstructorDeclaration(n)&&u(n.body)}))},e.getSetAccessorValueParameter=Xn,e.getSetAccessorTypeAnnotationNode=function(e){var n=Xn(e);return n&&n.type},e.getThisParameter=function(n){if(n.parameters.length&&!e.isJSDocSignature(n)){var t=n.parameters[0];if(Yn(t))return t}},e.parameterIsThisKeyword=Yn,e.isThisIdentifier=Qn,e.identifierIsThisKeyword=Zn,e.getAllAccessorDeclarations=function(n,t){var r,a,i,o;return sn(t)?(r=t,167===t.kind?i=t:168===t.kind?o=t:e.Debug.fail("Accessor has wrong kind")):e.forEach(n,(function(n){e.isAccessor(n)&&ot(n,32)===ot(t,32)&&(un(n.name)===un(t.name)&&(r?a||(a=n):r=n,167!==n.kind||i||(i=n),168!==n.kind||o||(o=n)))})),{firstAccessor:r,secondAccessor:a,getAccessor:i,setAccessor:o}},e.getEffectiveTypeAnnotationNode=$n,e.getTypeAnnotationNode=function(e){return e.type},e.getEffectiveReturnTypeNode=function(n){return e.isJSDocSignature(n)?n.type&&n.type.typeExpression&&n.type.typeExpression.type:n.type||(le(n)?e.getJSDocReturnType(n):void 0)},e.getJSDocTypeParameterDeclarations=function(n){return e.flatMap(e.getJSDocTags(n),(function(n){return function(n){return e.isJSDocTemplateTag(n)&&!(311===n.parent.kind&&n.parent.tags.some(Oe))}(n)?n.typeParameters:void 0}))},e.getEffectiveSetAccessorTypeAnnotationNode=function(e){var n=Xn(e);return n&&$n(n)},e.emitNewLineBeforeLeadingComments=et,e.emitNewLineBeforeLeadingCommentsOfPosition=nt,e.emitNewLineBeforeLeadingCommentOfPosition=function(e,n,t,r){t!==r&&Jn(e,t)!==Jn(e,r)&&n.writeLine()},e.emitComments=tt,e.emitDetachedComments=function(n,t,r,a,i,o,s){var c,l;if(s?0===i.pos&&(c=e.filter(e.getLeadingCommentRanges(n,i.pos),(function(e){return f(n,e.pos)}))):c=e.getLeadingCommentRanges(n,i.pos),c){for(var u=[],d=void 0,p=0,m=c;p<m.length;p++){var _=m[p];if(d){var g=Jn(t,d.end);if(Jn(t,_.pos)>=g+2)break}u.push(_),d=_}if(u.length){g=Jn(t,e.last(u).end);Jn(t,e.skipTrivia(n,i.pos))>=g+2&&(et(t,r,i,c),tt(n,t,r,u,!1,!0,o,a),l={nodePos:i.pos,detachedCommentEndPos:e.last(u).end})}}return l},e.writeCommentRange=function(n,t,r,a,i,o){if(42===n.charCodeAt(a+1))for(var s=e.computeLineAndCharacterOfPosition(t,a),c=t.length,l=void 0,u=a,d=s.line;u<i;d++){var p=d+1===c?n.length+1:t[d+1];if(u!==a){void 0===l&&(l=at(n,t[s.line],a));var m=r.getIndent()*Gn()-l+at(n,u,p);if(m>0){var f=m%Gn(),_=Fn((m-f)/Gn());for(r.rawWrite(_);f;)r.rawWrite(" "),f--}else r.rawWrite("")}rt(n,i,r,o,u,p),u=p}else r.writeComment(n.substring(a,i))},e.hasEffectiveModifiers=function(e){return 0!==pt(e)},e.hasSyntacticModifiers=function(e){return 0!==mt(e)},e.hasEffectiveModifier=it,e.hasSyntacticModifier=ot,e.hasStaticModifier=st,e.hasEffectiveReadonlyModifier=ct,e.getSelectedEffectiveModifierFlags=lt,e.getSelectedSyntacticModifierFlags=ut,e.getEffectiveModifierFlags=pt,e.getEffectiveModifierFlagsAlwaysIncludeJSDoc=function(e){return dt(e,!0,!0)},e.getSyntacticModifierFlags=mt,e.getEffectiveModifierFlagsNoCache=function(e){return _t(e)|ft(e)},e.getSyntacticModifierFlagsNoCache=_t,e.modifiersToFlags=gt,e.modifierToFlag=yt,e.isLogicalOperator=function(e){return 56===e||55===e||53===e},e.isLogicalOrCoalescingAssignmentOperator=vt,e.isLogicalOrCoalescingAssignmentExpression=function(e){return vt(e.operatorToken.kind)},e.isAssignmentOperator=ht,e.tryGetClassExtendingExpressionWithTypeArguments=bt,e.tryGetClassImplementingOrExtendingExpressionWithTypeArguments=Et,e.isAssignmentExpression=Tt,e.isDestructuringAssignment=function(e){if(Tt(e,!0)){var n=e.left.kind;return 200===n||199===n}return!1},e.isExpressionWithTypeArgumentsInClassExtendsClause=St,e.isEntityNameExpression=xt,e.getFirstIdentifier=function(e){switch(e.kind){case 78:return e;case 157:do{e=e.left}while(78!==e.kind);return e;case 201:do{e=e.expression}while(78!==e.kind);return e}},e.isDottedName=function e(n){return 78===n.kind||107===n.kind||105===n.kind||201===n.kind&&e(n.expression)||207===n.kind&&e(n.expression)},e.isPropertyAccessEntityNameExpression=Lt,e.tryGetPropertyAccessOrIdentifierToString=function n(t){if(e.isPropertyAccessExpression(t)){var r=n(t.expression);if(void 0!==r)return r+"."+k(t.name)}else if(e.isIdentifier(t))return e.unescapeLeadingUnderscores(t.escapedText)},e.isPrototypeAccess=At,e.isRightSideOfQualifiedNameOrPropertyAccess=function(e){return 157===e.parent.kind&&e.parent.right===e||201===e.parent.kind&&e.parent.name===e},e.isEmptyObjectLiteral=function(e){return 200===e.kind&&0===e.properties.length},e.isEmptyArrayLiteral=function(e){return 199===e.kind&&0===e.elements.length},e.getLocalSymbolForExportDefault=function(n){if(function(n){return n&&e.length(n.declarations)>0&&ot(n.declarations[0],512)}(n))for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t];if(a.localSymbol)return a.localSymbol}},e.tryExtractTSExtension=function(n){return e.find(e.supportedTSExtensionsForExtractExtension,(function(t){return e.fileExtensionIs(n,t)}))};var Dt="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";function Ct(n){for(var t,r,a,i,o="",s=function(n){for(var t=[],r=n.length,a=0;a<r;a++){var i=n.charCodeAt(a);i<128?t.push(i):i<2048?(t.push(i>>6|192),t.push(63&i|128)):i<65536?(t.push(i>>12|224),t.push(i>>6&63|128),t.push(63&i|128)):i<131072?(t.push(i>>18|240),t.push(i>>12&63|128),t.push(i>>6&63|128),t.push(63&i|128)):e.Debug.assert(!1,"Unexpected code point")}return t}(n),c=0,l=s.length;c<l;)t=s[c]>>2,r=(3&s[c])<<4|s[c+1]>>4,a=(15&s[c+1])<<2|s[c+2]>>6,i=63&s[c+2],c+1>=l?a=i=64:c+2>=l&&(i=64),o+=Dt.charAt(t)+Dt.charAt(r)+Dt.charAt(a)+Dt.charAt(i),c+=3;return o}e.convertToBase64=Ct,e.base64encode=function(e,n){return e&&e.base64encode?e.base64encode(n):Ct(n)},e.base64decode=function(e,n){if(e&&e.base64decode)return e.base64decode(n);for(var t=n.length,r=[],a=0;a<t&&n.charCodeAt(a)!==Dt.charCodeAt(64);){var i=Dt.indexOf(n[a]),o=Dt.indexOf(n[a+1]),s=Dt.indexOf(n[a+2]),c=Dt.indexOf(n[a+3]),l=(63&i)<<2|o>>4&3,u=(15&o)<<4|s>>2&15,d=(3&s)<<6|63&c;0===u&&0!==s?r.push(l):0===d&&0!==c?r.push(l,u):r.push(l,u,d),a+=4}return function(e){for(var n="",t=0,r=e.length;t<r;){var a=e[t];if(a<128)n+=String.fromCharCode(a),t++;else if(192==(192&a)){for(var i=63&a,o=e[++t];128==(192&o);)i=i<<6|63&o,o=e[++t];n+=String.fromCharCode(i)}else n+=String.fromCharCode(a),t++}return n}(r)},e.readJson=function(n,t){try{var r=t.readFile(n);if(!r)return{};var a=e.parseConfigFileTextToJson(n,r);return a.error?{}:a.config}catch(e){return{}}},e.directoryProbablyExists=function(e,n){return!n.directoryExists||n.directoryExists(e)};var It;function Nt(n,t){return void 0===t&&(t=n),e.Debug.assert(t>=n||-1===t),{pos:n,end:t}}function Rt(e,n){return Nt(n,e.end)}function kt(e){return e.decorators&&e.decorators.length>0?Rt(e,e.decorators.end):e}function Ot(e,n,t){return Pt(Mt(e,t,!1),n.end,t)}function Pt(n,t,r){return 0===e.getLinesBetweenPositions(r,n,t)}function Mt(n,t,r){return Wr(n.pos)?-1:e.skipTrivia(t.text,n.pos,!1,r)}function wt(e){return void 0!==e.initializer}function Ft(e){return 33554432&e.flags?e.checkFlags:0}function Gt(n){var t=n.parent;if(!t)return 0;switch(t.kind){case 207:return Gt(t);case 215:case 214:var r=t.operator;return 45===r||46===r?c():0;case 216:var a=t,i=a.left,o=a.operatorToken;return i===n&&ht(o.kind)?62===o.kind?1:c():0;case 201:return t.name!==n?0:Gt(t);case 288:var s=Gt(t.parent);return n===t.name?function(n){switch(n){case 0:return 1;case 1:return 0;case 2:return 2;default:return e.Debug.assertNever(n)}}(s):s;case 289:return n===t.objectAssignmentInitializer?0:Gt(t.parent);case 199:return Gt(t);default:return 0}function c(){return t.parent&&233===function(e){for(;207===e.kind;)e=e.parent;return e}(t.parent).kind?1:2}}function Bt(e,n,t){var r=t.onDeleteValue,a=t.onExistingValue;e.forEach((function(t,i){var o=n.get(i);void 0===o?(e.delete(i),r(t,i)):a&&a(t,o,i)}))}function Ut(e){if(32&e.flags){var n=Vt(e);return!!n&&ot(n,128)}return!1}function Vt(n){return e.find(n.declarations,e.isClassLike)}function jt(e){return 3899393&e.flags?e.objectFlags:0}function Ht(e){return 201===e.kind||202===e.kind}function Kt(e){for(;Ht(e);)e=e.expression;return e}function Wt(e,n){this.flags=e,this.escapedName=n,this.declarations=void 0,this.valueDeclaration=void 0,this.id=void 0,this.mergeId=void 0,this.parent=void 0}function zt(n,t){this.flags=t,(e.Debug.isDebugging||e.tracing.isTracing())&&(this.checker=n)}function qt(n,t){this.flags=t,e.Debug.isDebugging&&(this.checker=n)}function Jt(e,n,t){this.pos=n,this.end=t,this.kind=e,this.id=0,this.flags=0,this.modifierFlagsCache=0,this.transformFlags=0,this.parent=void 0,this.original=void 0}function Xt(e,n,t){this.pos=n,this.end=t,this.kind=e,this.id=0,this.flags=0,this.transformFlags=0,this.parent=void 0}function Yt(e,n,t){this.pos=n,this.end=t,this.kind=e,this.id=0,this.flags=0,this.transformFlags=0,this.parent=void 0,this.original=void 0,this.flowNode=void 0}function Qt(e,n,t){this.fileName=e,this.text=n,this.skipTrivia=t||function(e){return e}}function Zt(n,t,r){return void 0===r&&(r=0),n.replace(/{(\d+)}/g,(function(n,a){return""+e.Debug.checkDefined(t[+a+r])}))}function $t(n){return e.localizedDiagnosticMessages&&e.localizedDiagnosticMessages[n.key]||n.message}function er(e){return void 0===e.file&&void 0!==e.start&&void 0!==e.length&&"string"==typeof e.fileName}function nr(n,t){var r=t.fileName||"",a=t.text.length;e.Debug.assertEqual(n.fileName,r),e.Debug.assertLessThanOrEqual(n.start,a),e.Debug.assertLessThanOrEqual(n.start+n.length,a);var i={file:t,start:n.start,length:n.length,messageText:n.messageText,category:n.category,code:n.code,reportsUnnecessary:n.reportsUnnecessary};if(n.relatedInformation){i.relatedInformation=[];for(var o=0,s=n.relatedInformation;o<s.length;o++){var c=s[o];er(c)&&c.fileName===r?(e.Debug.assertLessThanOrEqual(c.start,a),e.Debug.assertLessThanOrEqual(c.start+c.length,a),i.relatedInformation.push(nr(c,t))):i.relatedInformation.push(c)}}return i}function tr(n,t,r,a){e.Debug.assertGreaterThanOrEqual(t,0),e.Debug.assertGreaterThanOrEqual(r,0),n&&(e.Debug.assertLessThanOrEqual(t,n.text.length),e.Debug.assertLessThanOrEqual(t+r,n.text.length));var i=$t(a);return arguments.length>4&&(i=Zt(i,arguments,4)),{file:n,start:t,length:r,messageText:i,category:a.category,code:a.code,reportsUnnecessary:a.reportsUnnecessary,reportsDeprecated:a.reportsDeprecated}}function rr(e){var n=$t(e);return arguments.length>1&&(n=Zt(n,arguments,1)),{file:void 0,start:void 0,length:void 0,messageText:n,category:e.category,code:e.code,reportsUnnecessary:e.reportsUnnecessary,reportsDeprecated:e.reportsDeprecated}}function ar(e){return e.file?e.file.path:void 0}function ir(n,t){return or(n,t)||function(n,t){if(!n.relatedInformation&&!t.relatedInformation)return 0;if(n.relatedInformation&&t.relatedInformation)return e.compareValues(n.relatedInformation.length,t.relatedInformation.length)||e.forEach(n.relatedInformation,(function(e,n){return ir(e,t.relatedInformation[n])}))||0;return n.relatedInformation?-1:1}(n,t)||0}function or(n,t){return e.compareStringsCaseSensitive(ar(n),ar(t))||e.compareValues(n.start,t.start)||e.compareValues(n.length,t.length)||e.compareValues(n.code,t.code)||sr(n.messageText,t.messageText)||0}function sr(n,t){if("string"==typeof n&&"string"==typeof t)return e.compareStringsCaseSensitive(n,t);if("string"==typeof n)return-1;if("string"==typeof t)return 1;var r=e.compareStringsCaseSensitive(n.messageText,t.messageText);if(r)return r;if(!n.next&&!t.next)return 0;if(!n.next)return-1;if(!t.next)return 1;for(var a=Math.min(n.next.length,t.next.length),i=0;i<a;i++)if(r=sr(n.next[i],t.next[i]))return r;return n.next.length<t.next.length?-1:n.next.length>t.next.length?1:0}function cr(e){return e.target||0}function lr(n){return"number"==typeof n.module?n.module:cr(n)>=2?e.ModuleKind.ES2015:e.ModuleKind.CommonJS}function ur(e){return!(!e.declaration&&!e.composite)}function dr(e,n){return void 0===e[n]?!!e.strict:!!e[n]}function pr(e){return void 0===e.allowJs?!!e.checkJs:e.allowJs}function mr(e,n){return n.strictFlag?dr(e,n.name):e[n.name]}function fr(e){for(var n=!1,t=0;t<e.length;t++)if(42===e.charCodeAt(t)){if(n)return!1;n=!0}return!0}function _r(){var n,t;return{getSymlinkedFiles:function(){return t},getSymlinkedDirectories:function(){return n},setSymlinkedFile:function(n,r){return(t||(t=new e.Map)).set(n,r)},setSymlinkedDirectory:function(t,r){return(n||(n=new e.Map)).set(t,r)}}}function gr(n,t,r,a){for(var i=e.getPathComponents(e.toPath(n,r,a)),o=e.getPathComponents(e.toPath(t,r,a)),s=!1;!yr(i[i.length-2],a)&&!yr(o[o.length-2],a)&&a(i[i.length-1])===a(o[o.length-1]);)i.pop(),o.pop(),s=!0;return s?[e.getPathFromPathComponents(i),e.getPathFromPathComponents(o)]:void 0}function yr(n,t){return"node_modules"===t(n)||e.startsWith(n,"@")}e.getNewLineCharacter=function(n,t){switch(n.newLine){case 0:return"\r\n";case 1:return"\n"}return t?t():e.sys?e.sys.newLine:"\r\n"},e.createRange=Nt,e.moveRangeEnd=function(e,n){return Nt(e.pos,n)},e.moveRangePos=Rt,e.moveRangePastDecorators=kt,e.moveRangePastModifiers=function(e){return e.modifiers&&e.modifiers.length>0?Rt(e,e.modifiers.end):kt(e)},e.isCollapsedRange=function(e){return e.pos===e.end},e.createTokenRange=function(n,t){return Nt(n,n+e.tokenToString(t).length)},e.rangeIsOnSingleLine=function(e,n){return Ot(e,e,n)},e.rangeStartPositionsAreOnSameLine=function(e,n,t){return Pt(Mt(e,t,!1),Mt(n,t,!1),t)},e.rangeEndPositionsAreOnSameLine=function(e,n,t){return Pt(e.end,n.end,t)},e.rangeStartIsOnSameLineAsRangeEnd=Ot,e.rangeEndIsOnSameLineAsRangeStart=function(e,n,t){return Pt(e.end,Mt(n,t,!1),t)},e.getLinesBetweenRangeEndAndRangeStart=function(n,t,r,a){var i=Mt(t,r,a);return e.getLinesBetweenPositions(r,n.end,i)},e.getLinesBetweenRangeEndPositions=function(n,t,r){return e.getLinesBetweenPositions(r,n.end,t.end)},e.isNodeArrayMultiLine=function(e,n){return!Pt(e.pos,e.end,n)},e.positionsAreOnSameLine=Pt,e.getStartPositionOfRange=Mt,e.getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter=function(n,t,r,a){var i=e.skipTrivia(r.text,n,!1,a),o=function(n,t,r){void 0===t&&(t=0);for(;n-- >t;)if(!e.isWhiteSpaceLike(r.text.charCodeAt(n)))return n}(i,t,r);return e.getLinesBetweenPositions(r,null!=o?o:t,i)},e.getLinesBetweenPositionAndNextNonWhitespaceCharacter=function(n,t,r,a){var i=e.skipTrivia(r.text,n,!1,a);return e.getLinesBetweenPositions(r,n,Math.min(t,i))},e.isDeclarationNameOfEnumOrNamespace=function(n){var t=e.getParseTreeNode(n);if(t)switch(t.parent.kind){case 255:case 256:return t===t.parent.name}return!1},e.getInitializedVariables=function(n){return e.filter(n.declarations,wt)},e.isWatchSet=function(e){return e.watch&&e.hasOwnProperty("watch")},e.closeFileWatcher=function(e){e.close()},e.getCheckFlags=Ft,e.getDeclarationModifierFlagsFromSymbol=function(n){if(n.valueDeclaration){var t=e.getCombinedModifierFlags(n.valueDeclaration);return n.parent&&32&n.parent.flags?t:-29&t}if(6&Ft(n)){var r=n.checkFlags;return(1024&r?8:256&r?4:16)|(2048&r?32:0)}return 4194304&n.flags?36:0},e.skipAlias=function(e,n){return 2097152&e.flags?n.getAliasedSymbol(e):e},e.getCombinedLocalAndExportSymbolFlags=function(e){return e.exportSymbol?e.exportSymbol.flags|e.flags:e.flags},e.isWriteOnlyAccess=function(e){return 1===Gt(e)},e.isWriteAccess=function(e){return 0!==Gt(e)},function(e){e[e.Read=0]="Read",e[e.Write=1]="Write",e[e.ReadWrite=2]="ReadWrite"}(It||(It={})),e.compareDataObjects=function e(n,t){if(!n||!t||Object.keys(n).length!==Object.keys(t).length)return!1;for(var r in n)if("object"==typeof n[r]){if(!e(n[r],t[r]))return!1}else if("function"!=typeof n[r]&&n[r]!==t[r])return!1;return!0},e.clearMap=function(e,n){e.forEach(n),e.clear()},e.mutateMapSkippingNewValues=Bt,e.mutateMap=function(e,n,t){Bt(e,n,t);var r=t.createNewValue;n.forEach((function(n,t){e.has(t)||e.set(t,r(t,n))}))},e.isAbstractConstructorType=function(e){return!!(16&jt(e))&&!!e.symbol&&Ut(e.symbol)},e.isAbstractConstructorSymbol=Ut,e.getClassLikeDeclarationOfSymbol=Vt,e.getObjectFlags=jt,e.typeHasCallOrConstructSignatures=function(e,n){return 0!==n.getSignaturesOfType(e,0).length||0!==n.getSignaturesOfType(e,1).length},e.forSomeAncestorDirectory=function(n,t){return!!e.forEachAncestorDirectory(n,(function(e){return!!t(e)||void 0}))},e.isUMDExportSymbol=function(n){return!!n&&!!n.declarations&&!!n.declarations[0]&&e.isNamespaceExportDeclaration(n.declarations[0])},e.showModuleSpecifier=function(n){var t=n.moduleSpecifier;return e.isStringLiteral(t)?t.text:v(t)},e.getLastChild=function(n){var t;return e.forEachChild(n,(function(e){u(e)&&(t=e)}),(function(e){for(var n=e.length-1;n>=0;n--)if(u(e[n])){t=e[n];break}})),t},e.addToSeen=function(e,n,t){return void 0===t&&(t=!0),n=String(n),!e.has(n)&&(e.set(n,t),!0)},e.isObjectTypeDeclaration=function(n){return e.isClassLike(n)||e.isInterfaceDeclaration(n)||e.isTypeLiteralNode(n)},e.isTypeNodeKind=function(e){return e>=172&&e<=195||128===e||152===e||144===e||155===e||145===e||131===e||147===e||148===e||113===e||150===e||141===e||223===e||303===e||304===e||305===e||306===e||307===e||308===e||309===e},e.isAccessExpression=Ht,e.getNameOfAccessExpression=function(n){return 201===n.kind?n.name:(e.Debug.assert(202===n.kind),n.argumentExpression)},e.isBundleFileTextLike=function(e){switch(e.kind){case"text":case"internal":return!0;default:return!1}},e.isNamedImportsOrExports=function(e){return 264===e.kind||268===e.kind},e.getLeftmostAccessExpression=Kt,e.getLeftmostExpression=function(e,n){for(;;){switch(e.kind){case 215:e=e.operand;continue;case 216:e=e.left;continue;case 217:e=e.condition;continue;case 205:e=e.tag;continue;case 203:if(n)return e;case 224:case 202:case 201:case 225:case 336:e=e.expression;continue}return e}},e.objectAllocator={getNodeConstructor:function(){return Jt},getTokenConstructor:function(){return Xt},getIdentifierConstructor:function(){return Yt},getPrivateIdentifierConstructor:function(){return Jt},getSourceFileConstructor:function(){return Jt},getSymbolConstructor:function(){return Wt},getTypeConstructor:function(){return zt},getSignatureConstructor:function(){return qt},getSourceMapSourceConstructor:function(){return Qt}},e.setObjectAllocator=function(n){e.objectAllocator=n},e.formatStringFromArgs=Zt,e.setLocalizedDiagnosticMessages=function(n){e.localizedDiagnosticMessages=n},e.getLocaleSpecificMessage=$t,e.createDetachedDiagnostic=function(n,t,r,a){e.Debug.assertGreaterThanOrEqual(t,0),e.Debug.assertGreaterThanOrEqual(r,0);var i=$t(a);return arguments.length>4&&(i=Zt(i,arguments,4)),{file:void 0,start:t,length:r,messageText:i,category:a.category,code:a.code,reportsUnnecessary:a.reportsUnnecessary,fileName:n}},e.attachFileToDiagnostics=function(e,n){for(var t=[],r=0,a=e;r<a.length;r++){var i=a[r];t.push(nr(i,n))}return t},e.createFileDiagnostic=tr,e.formatMessage=function(e,n){var t=$t(n);return arguments.length>2&&(t=Zt(t,arguments,2)),t},e.createCompilerDiagnostic=rr,e.createCompilerDiagnosticFromMessageChain=function(e){return{file:void 0,start:void 0,length:void 0,code:e.code,category:e.category,messageText:e.next?e:e.messageText}},e.chainDiagnosticMessages=function(e,n){var t=$t(n);return arguments.length>2&&(t=Zt(t,arguments,2)),{messageText:t,category:n.category,code:n.code,next:void 0===e||Array.isArray(e)?e:[e]}},e.concatenateDiagnosticMessageChains=function(e,n){for(var t=e;t.next;)t=t.next[0];t.next=[n]},e.compareDiagnostics=ir,e.compareDiagnosticsSkipRelatedInformation=or,e.getLanguageVariant=function(e){return 4===e||2===e||1===e||6===e?1:0},e.getEmitScriptTarget=cr,e.getEmitModuleKind=lr,e.getEmitModuleResolutionKind=function(n){var t=n.moduleResolution;return void 0===t&&(t=lr(n)===e.ModuleKind.CommonJS?e.ModuleResolutionKind.NodeJs:e.ModuleResolutionKind.Classic),t},e.hasJsonModuleEmitEnabled=function(n){switch(lr(n)){case e.ModuleKind.CommonJS:case e.ModuleKind.AMD:case e.ModuleKind.ES2015:case e.ModuleKind.ES2020:case e.ModuleKind.ESNext:return!0;default:return!1}},e.unreachableCodeIsError=function(e){return!1===e.allowUnreachableCode},e.unusedLabelIsError=function(e){return!1===e.allowUnusedLabels},e.getAreDeclarationMapsEnabled=function(e){return!(!ur(e)||!e.declarationMap)},e.getAllowSyntheticDefaultImports=function(n){var t=lr(n);return void 0!==n.allowSyntheticDefaultImports?n.allowSyntheticDefaultImports:n.esModuleInterop||t===e.ModuleKind.System},e.getEmitDeclarations=ur,e.isIncrementalCompilation=function(e){return!(!e.incremental&&!e.composite)},e.getStrictOptionValue=dr,e.getAllowJSCompilerOption=pr,e.compilerOptionsAffectSemanticDiagnostics=function(n,t){return t!==n&&e.semanticDiagnosticsOptionDeclarations.some((function(e){return!Jr(mr(t,e),mr(n,e))}))},e.compilerOptionsAffectEmit=function(n,t){return t!==n&&e.affectsEmitOptionDeclarations.some((function(e){return!Jr(mr(t,e),mr(n,e))}))},e.getCompilerOptionValue=mr,e.getJSXTransformEnabled=function(e){var n=e.jsx;return 2===n||4===n||5===n},e.getJSXImplicitImportBase=function(n,t){var r=null==t?void 0:t.pragmas.get("jsximportsource"),a=e.isArray(r)?r[0]:r;return 4===n.jsx||5===n.jsx||n.jsxImportSource||a?(null==a?void 0:a.arguments.factory)||n.jsxImportSource||"react":void 0},e.getJSXRuntimeImport=function(e,n){return e?e+"/"+(5===n.jsx?"jsx-dev-runtime":"jsx-runtime"):void 0},e.hasZeroOrOneAsteriskCharacter=fr,e.createSymlinkCache=_r,e.discoverProbableSymlinks=function(n,t,r){for(var a=_r(),i=0,o=e.flatten(e.mapDefined(n,(function(n){return n.resolvedModules&&e.compact(e.arrayFrom(e.mapIterator(n.resolvedModules.values(),(function(e){return e&&e.originalPath&&e.resolvedFileName!==e.originalPath?[e.resolvedFileName,e.originalPath]:void 0}))))})));i<o.length;i++){var s=o[i],c=gr(s[0],s[1],r,t)||e.emptyArray,l=c[0],u=c[1];l&&u&&a.setSymlinkedDirectory(e.toPath(u,r,t),{real:l,realPath:e.toPath(l,r,t)})}return a},e.tryRemoveDirectoryPrefix=function(n,t,r){var a,i=e.tryRemovePrefix(n,t,r);return void 0===i?void 0:(a=i,e.isAnyDirectorySeparator(a.charCodeAt(0))?a.slice(1):void 0)};var vr=/[^\w\s\/]/g;function hr(e){return"\\"+e}e.regExpEscape=function(e){return e.replace(vr,hr)};var br=[42,63];e.commonPackageFolders=["node_modules","bower_components","jspm_packages"];var Er="(?!("+e.commonPackageFolders.join("|")+")(/|$))",Tr={singleAsteriskRegexFragment:"([^./]|(\\.(?!min\\.js$))?)*",doubleAsteriskRegexFragment:"(/"+Er+"[^/.][^/]*)*?",replaceWildcardCharacter:function(e){return Ir(e,Tr.singleAsteriskRegexFragment)}},Sr={singleAsteriskRegexFragment:"[^/]*",doubleAsteriskRegexFragment:"(/"+Er+"[^/.][^/]*)*?",replaceWildcardCharacter:function(e){return Ir(e,Sr.singleAsteriskRegexFragment)}},xr={singleAsteriskRegexFragment:"[^/]*",doubleAsteriskRegexFragment:"(/.+?)?",replaceWildcardCharacter:function(e){return Ir(e,xr.singleAsteriskRegexFragment)}},Lr={files:Tr,directories:Sr,exclude:xr};function Ar(e,n,t){var r=Dr(e,n,t);if(r&&r.length)return"^("+r.map((function(e){return"("+e+")"})).join("|")+")"+("exclude"===t?"($|/)":"$")}function Dr(n,t,r){if(void 0!==n&&0!==n.length)return e.flatMap(n,(function(n){return n&&function(n,t,r,a){var i=a.singleAsteriskRegexFragment,o=a.doubleAsteriskRegexFragment,s=a.replaceWildcardCharacter,c="",l=!1,u=e.getNormalizedPathComponents(n,t),d=e.last(u);if("exclude"!==r&&"**"===d)return;u[0]=e.removeTrailingDirectorySeparator(u[0]),Cr(d)&&u.push("**","*");for(var p=0,m=0,f=u;m<f.length;m++){var _=f[m];if("**"===_)c+=o;else if("directories"===r&&(c+="(",p++),l&&(c+=e.directorySeparator),"exclude"!==r){var g="";42===_.charCodeAt(0)?(g+="([^./]"+i+")?",_=_.substr(1)):63===_.charCodeAt(0)&&(g+="[^./]",_=_.substr(1)),(g+=_.replace(vr,s))!==_&&(c+=Er),c+=g}else c+=_.replace(vr,s);l=!0}for(;p>0;)c+=")?",p--;return c}(n,t,r,Lr[r])}))}function Cr(e){return!/[.*?]/.test(e)}function Ir(e,n){return"*"===e?n:"?"===e?"[^/]":"\\"+e}function Nr(n,t,r,a,i){n=e.normalizePath(n),i=e.normalizePath(i);var o=e.combinePaths(i,n);return{includeFilePatterns:e.map(Dr(r,o,"files"),(function(e){return"^"+e+"$"})),includeFilePattern:Ar(r,o,"files"),includeDirectoryPattern:Ar(r,o,"directories"),excludePattern:Ar(t,o,"exclude"),basePaths:kr(n,r,a)}}function Rr(e,n){return new RegExp(e,n?"":"i")}function kr(n,t,r){var a=[n];if(t){for(var i=[],o=0,s=t;o<s.length;o++){var c=s[o],l=e.isRootedDiskPath(c)?c:e.normalizePath(e.combinePaths(n,c));i.push(Or(l))}i.sort(e.getStringComparer(!r));for(var u=function(t){e.every(a,(function(a){return!e.containsPath(a,t,n,!r)}))&&a.push(t)},d=0,p=i;d<p.length;d++){u(p[d])}}return a}function Or(n){var t=e.indexOfAnyCharCode(n,br);return t<0?e.hasExtension(n)?e.removeTrailingDirectorySeparator(e.getDirectoryPath(n)):n:n.substring(0,n.lastIndexOf(e.directorySeparator,t))}function Pr(e){switch(e.substr(e.lastIndexOf(".")).toLowerCase()){case".js":return 1;case".jsx":return 2;case".ts":return 3;case".tsx":return 4;case".json":return 6;default:return 0}}e.getRegularExpressionForWildcard=Ar,e.getRegularExpressionsForWildcards=Dr,e.isImplicitGlob=Cr,e.getFileMatcherPatterns=Nr,e.getRegexFromPattern=Rr,e.matchFiles=function(n,t,r,a,i,o,s,c,l){n=e.normalizePath(n),o=e.normalizePath(o);for(var u=Nr(n,r,a,i,o),d=u.includeFilePatterns&&u.includeFilePatterns.map((function(e){return Rr(e,i)})),p=u.includeDirectoryPattern&&Rr(u.includeDirectoryPattern,i),m=u.excludePattern&&Rr(u.excludePattern,i),f=d?d.map((function(){return[]})):[[]],_=new e.Map,g=e.createGetCanonicalFileName(i),y=0,v=u.basePaths;y<v.length;y++){var h=v[y];b(h,e.combinePaths(o,h),s)}return e.flatten(f);function b(n,r,a){var i=g(l(r));if(!_.has(i)){_.set(i,!0);for(var o=c(n),s=o.files,u=o.directories,y=function(a){var i=e.combinePaths(n,a),o=e.combinePaths(r,a);if(t&&!e.fileExtensionIsOneOf(i,t))return"continue";if(m&&m.test(o))return"continue";if(d){var s=e.findIndex(d,(function(e){return e.test(o)}));-1!==s&&f[s].push(i)}else f[0].push(i)},v=0,h=e.sort(s,e.compareStringsCaseSensitive);v<h.length;v++){y(S=h[v])}if(void 0===a||0!=--a)for(var E=0,T=e.sort(u,e.compareStringsCaseSensitive);E<T.length;E++){var S=T[E],x=e.combinePaths(n,S),L=e.combinePaths(r,S);p&&!p.test(L)||m&&m.test(L)||b(x,L,a)}}}},e.ensureScriptKind=function(e,n){return n||Pr(e)||3},e.getScriptKindFromFileName=Pr,e.supportedTSExtensions=[".ts",".tsx",".d.ts"],e.supportedTSExtensionsWithJson=[".ts",".tsx",".d.ts",".json"],e.supportedTSExtensionsForExtractExtension=[".d.ts",".ts",".tsx"],e.supportedJSExtensions=[".js",".jsx"],e.supportedJSAndJsonExtensions=[".js",".jsx",".json"];var Mr=__spreadArrays(e.supportedTSExtensions,e.supportedJSExtensions),wr=__spreadArrays(e.supportedTSExtensions,e.supportedJSExtensions,[".json"]);function Fr(n,t){var r=n&&pr(n);if(!t||0===t.length)return r?Mr:e.supportedTSExtensions;var a=__spreadArrays(r?Mr:e.supportedTSExtensions,e.mapDefined(t,(function(e){return 7===e.scriptKind||r&&(1===(n=e.scriptKind)||2===n)?e.extension:void 0;var n})));return e.deduplicate(a,e.equateStringsCaseSensitive,e.compareStringsCaseSensitive)}function Gr(n,t){return n&&n.resolveJsonModule?t===Mr?wr:t===e.supportedTSExtensions?e.supportedTSExtensionsWithJson:__spreadArrays(t,[".json"]):t}function Br(e,n){return e<2?0:e<n.length?2:n.length}e.getSupportedExtensions=Fr,e.getSuppoertedExtensionsWithJsonIfResolveJsonModule=Gr,e.hasJSFileExtension=function(n){return e.some(e.supportedJSExtensions,(function(t){return e.fileExtensionIs(n,t)}))},e.hasTSFileExtension=function(n){return e.some(e.supportedTSExtensions,(function(t){return e.fileExtensionIs(n,t)}))},e.isSupportedSourceFileName=function(n,t,r){if(!n)return!1;for(var a=0,i=Gr(t,Fr(t,r));a<i.length;a++){var o=i[a];if(e.fileExtensionIs(n,o))return!0}return!1},function(e){e[e.TypeScriptFiles=0]="TypeScriptFiles",e[e.DeclarationAndJavaScriptFiles=2]="DeclarationAndJavaScriptFiles",e[e.Highest=0]="Highest",e[e.Lowest=2]="Lowest"}(e.ExtensionPriority||(e.ExtensionPriority={})),e.getExtensionPriority=function(n,t){for(var r=t.length-1;r>=0;r--)if(e.fileExtensionIs(n,t[r]))return Br(r,t);return 0},e.adjustExtensionPriority=Br,e.getNextLowestExtensionPriority=function(e,n){return e<2?2:n.length};var Ur=[".d.ts",".ts",".js",".tsx",".jsx",".json"];function Vr(e){for(var n=0,t=Ur;n<t.length;n++){var r=jr(e,t[n]);if(void 0!==r)return r}return e}function jr(n,t){return e.fileExtensionIs(n,t)?Hr(n,t):void 0}function Hr(e,n){return e.substring(0,e.length-n.length)}function Kr(n){e.Debug.assert(fr(n));var t=n.indexOf("*");return-1===t?void 0:{prefix:n.substr(0,t),suffix:n.substr(t+1)}}function Wr(e){return!(e>=0)}function zr(e){return".ts"===e||".tsx"===e||".d.ts"===e}function qr(n){return e.find(Ur,(function(t){return e.fileExtensionIs(n,t)}))}function Jr(n,t){return n===t||"object"==typeof n&&null!==n&&"object"==typeof t&&null!==t&&e.equalOwnProperties(n,t,Jr)}function Xr(e,n){return e.pos=n,e}function Yr(e,n){return e.end=n,e}function Qr(e,n,t){return Yr(Xr(e,n),t)}function Zr(e,n){return e&&n&&(e.parent=n),e}e.removeFileExtension=Vr,e.tryRemoveExtension=jr,e.removeExtension=Hr,e.changeExtension=function(n,t){return e.changeAnyExtension(n,t,Ur,!1)},e.tryParsePattern=Kr,e.positionIsSynthesized=Wr,e.extensionIsTS=zr,e.resolutionExtensionIsTSOrJson=function(e){return zr(e)||".json"===e},e.extensionFromPath=function(n){var t=qr(n);return void 0!==t?t:e.Debug.fail("File "+n+" has unknown extension.")},e.isAnySupportedFileExtension=function(e){return void 0!==qr(e)},e.tryGetExtensionFromPath=qr,e.isCheckJsEnabledForFile=function(e,n){return e.checkJsDirective?e.checkJsDirective.enabled:n.checkJs},e.emptyFileSystemEntries={files:e.emptyArray,directories:e.emptyArray},e.matchPatternOrExact=function(n,t){for(var r=[],a=0,i=n;a<i.length;a++){var o=i[a];if(fr(o)){var s=Kr(o);if(s)r.push(s);else if(o===t)return o}}return e.findBestPatternMatch(r,(function(e){return e}),t)},e.sliceAfter=function(n,t){var r=n.indexOf(t);return e.Debug.assert(-1!==r),n.slice(r)},e.addRelatedInfo=function(n){for(var t,r=[],a=1;a<arguments.length;a++)r[a-1]=arguments[a];return r.length?(n.relatedInformation||(n.relatedInformation=[]),e.Debug.assert(n.relatedInformation!==e.emptyArray,"Diagnostic had empty array singleton for related info, but is still being constructed!"),(t=n.relatedInformation).push.apply(t,r),n):n},e.minAndMax=function(n,t){e.Debug.assert(0!==n.length);for(var r=t(n[0]),a=r,i=1;i<n.length;i++){var o=t(n[i]);o<r?r=o:o>a&&(a=o)}return{min:r,max:a}},e.rangeOfNode=function(e){return{pos:_(e),end:e.end}},e.rangeOfTypeParameters=function(e){return{pos:e.pos-1,end:e.end+1}},e.skipTypeChecking=function(e,n,t){return n.skipLibCheck&&e.isDeclarationFile||n.skipDefaultLibCheck&&e.hasNoDefaultLib||t.isSourceOfProjectReferenceRedirect(e.fileName)},e.isJsonEqual=Jr,e.parsePseudoBigInt=function(e){var n;switch(e.charCodeAt(1)){case 98:case 66:n=1;break;case 111:case 79:n=3;break;case 120:case 88:n=4;break;default:for(var t=e.length-1,r=0;48===e.charCodeAt(r);)r++;return e.slice(r,t)||"0"}for(var a=e.length-1,i=(a-2)*n,o=new Uint16Array((i>>>4)+(15&i?1:0)),s=a-1,c=0;s>=2;s--,c+=n){var l=c>>>4,u=e.charCodeAt(s),d=(u<=57?u-48:10+u-(u<=70?65:97))<<(15&c);o[l]|=d;var p=d>>>16;p&&(o[l+1]|=p)}for(var m="",f=o.length-1,_=!0;_;){var g=0;_=!1;for(l=f;l>=0;l--){var y=g<<16|o[l],v=y/10|0;o[l]=v,g=y-10*v,v&&!_&&(f=l,_=!0)}m=g+m}return m},e.pseudoBigIntToString=function(e){var n=e.negative,t=e.base10Value;return(n&&"0"!==t?"-":"")+t},e.isValidTypeOnlyAliasUseSite=function(n){return!!(8388608&n.flags)||oe(n)||function(n){if(78!==n.kind)return!1;var t=e.findAncestor(n.parent,(function(e){switch(e.kind){case 286:return!0;case 201:case 223:return!1;default:return"quit"}}));return 116===(null==t?void 0:t.token)||253===(null==t?void 0:t.parent.kind)}(n)||function(e){for(;78===e.kind||201===e.kind;)e=e.parent;if(158!==e.kind)return!1;if(ot(e.parent,128))return!0;var n=e.parent.parent.kind;return 253===n||177===n}(n)||!ae(n)},e.typeOnlyDeclarationIsExport=function(e){return 270===e.kind},e.isIdentifierTypeReference=function(n){return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)},e.arrayIsHomogeneous=function(n,t){if(void 0===t&&(t=e.equateValues),n.length<2)return!0;for(var r=n[0],a=1,i=n.length;a<i;a++){if(!t(r,n[a]))return!1}return!0},e.setTextRangePos=Xr,e.setTextRangeEnd=Yr,e.setTextRangePosEnd=Qr,e.setTextRangePosWidth=function(e,n,t){return Qr(e,n,n+t)},e.setNodeFlags=function(e,n){return e&&(e.flags=n),e},e.setParent=Zr,e.setEachParent=function(e,n){if(e)for(var t=0,r=e;t<r.length;t++){Zr(r[t],n)}return e},e.setParentRecursive=function(n,t){return n?(e.forEachChildRecursively(n,e.isJSDocNode(n)?r:function(n,t){return r(n,t)||function(n){if(e.hasJSDocNodes(n))for(var t=0,a=n.jsDoc;t<a.length;t++){var i=a[t];r(i,n),e.forEachChildRecursively(i,r)}}(n)}),n):n;function r(e,n){if(t&&e.parent===n)return"skip";Zr(e,n)}}}(ts||(ts={})),function(e){e.createBaseNodeFactory=function(){var n,t,r,a,i;return{createBaseSourceFileNode:function(n){return new(i||(i=e.objectAllocator.getSourceFileConstructor()))(n,-1,-1)},createBaseIdentifierNode:function(n){return new(r||(r=e.objectAllocator.getIdentifierConstructor()))(n,-1,-1)},createBasePrivateIdentifierNode:function(n){return new(a||(a=e.objectAllocator.getPrivateIdentifierConstructor()))(n,-1,-1)},createBaseTokenNode:function(n){return new(t||(t=e.objectAllocator.getTokenConstructor()))(n,-1,-1)},createBaseNode:function(t){return new(n||(n=e.objectAllocator.getNodeConstructor()))(t,-1,-1)}}}}(ts||(ts={})),function(e){e.createParenthesizerRules=function(n){return{parenthesizeLeftSideOfBinary:function(e,n){return r(e,n,!0)},parenthesizeRightSideOfBinary:function(e,n,t){return r(e,t,!1,n)},parenthesizeExpressionOfComputedPropertyName:function(t){return e.isCommaSequence(t)?n.createParenthesizedExpression(t):t},parenthesizeConditionOfConditionalExpression:function(t){var r=e.getOperatorPrecedence(217,57),a=e.skipPartiallyEmittedExpressions(t),i=e.getExpressionPrecedence(a);if(1!==e.compareValues(i,r))return n.createParenthesizedExpression(t);return t},parenthesizeBranchOfConditionalExpression:function(t){var r=e.skipPartiallyEmittedExpressions(t);return e.isCommaSequence(r)?n.createParenthesizedExpression(t):t},parenthesizeExpressionOfExportDefault:function(t){var r=e.skipPartiallyEmittedExpressions(t),a=e.isCommaSequence(r);if(!a)switch(e.getLeftmostExpression(r,!1).kind){case 221:case 208:a=!0}return a?n.createParenthesizedExpression(t):t},parenthesizeExpressionOfNew:function(t){var r=e.getLeftmostExpression(t,!0);switch(r.kind){case 203:return n.createParenthesizedExpression(t);case 204:return r.arguments?t:n.createParenthesizedExpression(t)}return a(t)},parenthesizeLeftSideOfAccess:a,parenthesizeOperandOfPostfixUnary:function(t){return e.isLeftHandSideExpression(t)?t:e.setTextRange(n.createParenthesizedExpression(t),t)},parenthesizeOperandOfPrefixUnary:function(t){return e.isUnaryExpression(t)?t:e.setTextRange(n.createParenthesizedExpression(t),t)},parenthesizeExpressionsOfCommaDelimitedList:function(t){var r=e.sameMap(t,i);return e.setTextRange(n.createNodeArray(r,t.hasTrailingComma),t)},parenthesizeExpressionForDisallowedComma:i,parenthesizeExpressionOfExpressionStatement:function(t){var r=e.skipPartiallyEmittedExpressions(t);if(e.isCallExpression(r)){var a=r.expression,i=e.skipPartiallyEmittedExpressions(a).kind;if(208===i||209===i){var o=n.updateCallExpression(r,e.setTextRange(n.createParenthesizedExpression(a),a),r.typeArguments,r.arguments);return n.restoreOuterExpressions(t,o,8)}}var s=e.getLeftmostExpression(r,!1).kind;if(200===s||208===s)return e.setTextRange(n.createParenthesizedExpression(t),t);return t},parenthesizeConciseBodyOfArrowFunction:function(t){if(!e.isBlock(t)&&(e.isCommaSequence(t)||200===e.getLeftmostExpression(t,!1).kind))return e.setTextRange(n.createParenthesizedExpression(t),t);return t},parenthesizeMemberOfConditionalType:o,parenthesizeMemberOfElementType:s,parenthesizeElementTypeOfArrayType:function(e){switch(e.kind){case 176:case 188:case 185:return n.createParenthesizedType(e)}return s(e)},parenthesizeConstituentTypesOfUnionOrIntersectionType:function(t){return n.createNodeArray(e.sameMap(t,s))},parenthesizeTypeArguments:function(t){if(e.some(t))return n.createNodeArray(e.sameMap(t,c))}};function t(n){if(n=e.skipPartiallyEmittedExpressions(n),e.isLiteralKind(n.kind))return n.kind;if(216===n.kind&&39===n.operatorToken.kind){if(void 0!==n.cachedLiteralKind)return n.cachedLiteralKind;var r=t(n.left),a=e.isLiteralKind(r)&&r===t(n.right)?r:0;return n.cachedLiteralKind=a,a}return 0}function r(r,a,i,o){return 207===e.skipPartiallyEmittedExpressions(a).kind?a:function(n,r,a,i){var o=e.getOperatorPrecedence(216,n),s=e.getOperatorAssociativity(216,n),c=e.skipPartiallyEmittedExpressions(r);if(!a&&209===r.kind&&o>3)return!0;var l=e.getExpressionPrecedence(c);switch(e.compareValues(l,o)){case-1:return!(!a&&1===s&&219===r.kind);case 1:return!1;case 0:if(a)return 1===s;if(e.isBinaryExpression(c)&&c.operatorToken.kind===n){if(function(e){return 41===e||51===e||50===e||52===e}(n))return!1;if(39===n){var u=i?t(i):0;if(e.isLiteralKind(u)&&u===t(c))return!1}}return 0===e.getExpressionAssociativity(c)}}(r,a,i,o)?n.createParenthesizedExpression(a):a}function a(t){var r=e.skipPartiallyEmittedExpressions(t);return e.isLeftHandSideExpression(r)&&(204!==r.kind||r.arguments)?t:e.setTextRange(n.createParenthesizedExpression(t),t)}function i(t){var r=e.skipPartiallyEmittedExpressions(t);return e.getExpressionPrecedence(r)>e.getOperatorPrecedence(216,27)?t:e.setTextRange(n.createParenthesizedExpression(t),t)}function o(e){return 184===e.kind?n.createParenthesizedType(e):e}function s(e){switch(e.kind){case 182:case 183:case 174:case 175:return n.createParenthesizedType(e)}return o(e)}function c(t,r){return 0===r&&e.isFunctionOrConstructorTypeNode(t)&&t.typeParameters?n.createParenthesizedType(t):t}},e.nullParenthesizerRules={parenthesizeLeftSideOfBinary:function(e,n){return n},parenthesizeRightSideOfBinary:function(e,n,t){return t},parenthesizeExpressionOfComputedPropertyName:e.identity,parenthesizeConditionOfConditionalExpression:e.identity,parenthesizeBranchOfConditionalExpression:e.identity,parenthesizeExpressionOfExportDefault:e.identity,parenthesizeExpressionOfNew:function(n){return e.cast(n,e.isLeftHandSideExpression)},parenthesizeLeftSideOfAccess:function(n){return e.cast(n,e.isLeftHandSideExpression)},parenthesizeOperandOfPostfixUnary:function(n){return e.cast(n,e.isLeftHandSideExpression)},parenthesizeOperandOfPrefixUnary:function(n){return e.cast(n,e.isUnaryExpression)},parenthesizeExpressionsOfCommaDelimitedList:function(n){return e.cast(n,e.isNodeArray)},parenthesizeExpressionForDisallowedComma:e.identity,parenthesizeExpressionOfExpressionStatement:e.identity,parenthesizeConciseBodyOfArrowFunction:e.identity,parenthesizeMemberOfConditionalType:e.identity,parenthesizeMemberOfElementType:e.identity,parenthesizeElementTypeOfArrayType:e.identity,parenthesizeConstituentTypesOfUnionOrIntersectionType:function(n){return e.cast(n,e.isNodeArray)},parenthesizeTypeArguments:function(n){return n&&e.cast(n,e.isNodeArray)}}}(ts||(ts={})),function(e){e.createNodeConverters=function(n){return{convertToFunctionBlock:function(t,r){if(e.isBlock(t))return t;var a=n.createReturnStatement(t);e.setTextRange(a,t);var i=n.createBlock([a],r);return e.setTextRange(i,t),i},convertToFunctionExpression:function(t){if(!t.body)return e.Debug.fail("Cannot convert a FunctionDeclaration without a body");var r=n.createFunctionExpression(t.modifiers,t.asteriskToken,t.name,t.typeParameters,t.parameters,t.type,t.body);e.setOriginalNode(r,t),e.setTextRange(r,t),e.getStartsOnNewLine(t)&&e.setStartsOnNewLine(r,!0);return r},convertToArrayAssignmentElement:t,convertToObjectAssignmentElement:r,convertToAssignmentPattern:a,convertToObjectAssignmentPattern:i,convertToArrayAssignmentPattern:o,convertToAssignmentElementTarget:s};function t(t){if(e.isBindingElement(t)){if(t.dotDotDotToken)return e.Debug.assertNode(t.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(n.createSpreadElement(t.name),t),t);var r=s(t.name);return t.initializer?e.setOriginalNode(e.setTextRange(n.createAssignment(r,t.initializer),t),t):r}return e.cast(t,e.isExpression)}function r(t){if(e.isBindingElement(t)){if(t.dotDotDotToken)return e.Debug.assertNode(t.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(n.createSpreadAssignment(t.name),t),t);if(t.propertyName){var r=s(t.name);return e.setOriginalNode(e.setTextRange(n.createPropertyAssignment(t.propertyName,t.initializer?n.createAssignment(r,t.initializer):r),t),t)}return e.Debug.assertNode(t.name,e.isIdentifier),e.setOriginalNode(e.setTextRange(n.createShorthandPropertyAssignment(t.name,t.initializer),t),t)}return e.cast(t,e.isObjectLiteralElementLike)}function a(e){switch(e.kind){case 197:case 199:return o(e);case 196:case 200:return i(e)}}function i(t){return e.isObjectBindingPattern(t)?e.setOriginalNode(e.setTextRange(n.createObjectLiteralExpression(e.map(t.elements,r)),t),t):e.cast(t,e.isObjectLiteralExpression)}function o(r){return e.isArrayBindingPattern(r)?e.setOriginalNode(e.setTextRange(n.createArrayLiteralExpression(e.map(r.elements,t)),r),r):e.cast(r,e.isArrayLiteralExpression)}function s(n){return e.isBindingPattern(n)?a(n):e.cast(n,e.isExpression)}},e.nullNodeConverters={convertToFunctionBlock:e.notImplemented,convertToFunctionExpression:e.notImplemented,convertToArrayAssignmentElement:e.notImplemented,convertToObjectAssignmentElement:e.notImplemented,convertToAssignmentPattern:e.notImplemented,convertToObjectAssignmentPattern:e.notImplemented,convertToArrayAssignmentPattern:e.notImplemented,convertToAssignmentElementTarget:e.notImplemented}}(ts||(ts={})),function(e){var n,t=0;function r(r,p){var m=8&r?a:i,f=e.memoize((function(){return 1&r?e.nullParenthesizerRules:e.createParenthesizerRules(D)})),_=e.memoize((function(){return 2&r?e.nullNodeConverters:e.createNodeConverters(D)})),g=e.memoizeOne((function(e){return function(n,t){return Nn(n,e,t)}})),v=e.memoizeOne((function(e){return function(n){return Cn(e,n)}})),h=e.memoizeOne((function(e){return function(n){return In(n,e)}})),b=e.memoizeOne((function(e){return function(){return function(e){return I(e)}(e)}})),E=e.memoizeOne((function(e){return function(n){return Jt(e,n)}})),T=e.memoizeOne((function(e){return function(n,t){return function(e,n,t){return n.type!==t?m(Jt(e,t),n):n}(e,n,t)}})),S=e.memoizeOne((function(e){return function(n,t){return ur(e,n,t)}})),x=e.memoizeOne((function(e){return function(n,t,r){return function(e,n,t,r){void 0===t&&(t=$t(n));return n.tagName!==t||n.comment!==r?m(ur(e,t,r),n):n}(e,n,t,r)}})),L=e.memoizeOne((function(e){return function(n,t,r){return dr(e,n,t,r)}})),A=e.memoizeOne((function(e){return function(n,t,r,a){return function(e,n,t,r,a){void 0===t&&(t=$t(n));return n.tagName!==t||n.typeExpression!==r||n.comment!==a?m(dr(e,t,r,a),n):n}(e,n,t,r,a)}})),D={get parenthesizer(){return f()},get converters(){return _()},createNodeArray:C,createNumericLiteral:j,createBigIntLiteral:H,createStringLiteral:W,createStringLiteralFromNode:function(n){var t=K(e.getTextOfIdentifierOrLiteral(n),void 0);return t.textSourceNode=n,t},createRegularExpressionLiteral:z,createLiteralLikeNode:function(e,n){switch(e){case 8:return j(n,0);case 9:return H(n);case 10:return W(n,void 0);case 11:return hr(n,!1);case 12:return hr(n,!0);case 13:return z(n);case 14:return Mn(e,n,void 0,0)}},createIdentifier:X,updateIdentifier:function(n,t){return n.typeArguments!==t?m(X(e.idText(n),t),n):n},createTempVariable:Y,createLoopVariable:function(){return J("",2)},createUniqueName:function(n,t){void 0===t&&(t=0);return e.Debug.assert(!(7&t),"Argument out of range: flags"),e.Debug.assert(32!=(48&t),"GeneratedIdentifierFlags.FileLevel cannot be set without also setting GeneratedIdentifierFlags.Optimistic"),J(n,3|t)},getGeneratedNameForNode:Q,createPrivateIdentifier:function(n){e.startsWith(n,"#")||e.Debug.fail("First character of private identifier must be #: "+n);var t=p.createBasePrivateIdentifierNode(79);return t.escapedText=e.escapeLeadingUnderscores(n),t.transformFlags|=4194304,t},createToken:$,createSuper:function(){return $(105)},createThis:ee,createNull:function(){return $(103)},createTrue:ne,createFalse:te,createModifier:re,createModifiersFromModifierFlags:ae,createQualifiedName:ie,updateQualifiedName:function(e,n,t){return e.left!==n||e.right!==t?m(ie(n,t),e):e},createComputedPropertyName:oe,updateComputedPropertyName:function(e,n){return e.expression!==n?m(oe(n),e):e},createTypeParameterDeclaration:se,updateTypeParameterDeclaration:function(e,n,t,r){return e.name!==n||e.constraint!==t||e.default!==r?m(se(n,t,r),e):e},createParameterDeclaration:ce,updateParameterDeclaration:le,createDecorator:ue,updateDecorator:function(e,n){return e.expression!==n?m(ue(n),e):e},createPropertySignature:de,updatePropertySignature:pe,createPropertyDeclaration:me,updatePropertyDeclaration:fe,createMethodSignature:_e,updateMethodSignature:ge,createMethodDeclaration:ye,updateMethodDeclaration:ve,createConstructorDeclaration:he,updateConstructorDeclaration:be,createGetAccessorDeclaration:Ee,updateGetAccessorDeclaration:Te,createSetAccessorDeclaration:Se,updateSetAccessorDeclaration:xe,createCallSignature:Le,updateCallSignature:function(e,n,t,r){return e.typeParameters!==n||e.parameters!==t||e.type!==r?P(Le(n,t,r),e):e},createConstructSignature:Ae,updateConstructSignature:function(e,n,t,r){return e.typeParameters!==n||e.parameters!==t||e.type!==r?P(Ae(n,t,r),e):e},createIndexSignature:De,updateIndexSignature:Ce,createTemplateLiteralTypeSpan:Ie,updateTemplateLiteralTypeSpan:function(e,n,t){return e.type!==n||e.literal!==t?m(Ie(n,t),e):e},createKeywordTypeNode:function(e){return $(e)},createTypePredicateNode:Ne,updateTypePredicateNode:function(e,n,t,r){return e.assertsModifier!==n||e.parameterName!==t||e.type!==r?m(Ne(n,t,r),e):e},createTypeReferenceNode:Re,updateTypeReferenceNode:function(e,n,t){return e.typeName!==n||e.typeArguments!==t?m(Re(n,t),e):e},createFunctionTypeNode:ke,updateFunctionTypeNode:function(e,n,t,r){return e.typeParameters!==n||e.parameters!==t||e.type!==r?P(ke(n,t,r),e):e},createConstructorTypeNode:Oe,updateConstructorTypeNode:function(e,n,t,r){return e.typeParameters!==n||e.parameters!==t||e.type!==r?P(Oe(n,t,r),e):e},createTypeQueryNode:Pe,updateTypeQueryNode:function(e,n){return e.exprName!==n?m(Pe(n),e):e},createTypeLiteralNode:Me,updateTypeLiteralNode:function(e,n){return e.members!==n?m(Me(n),e):e},createArrayTypeNode:we,updateArrayTypeNode:function(e,n){return e.elementType!==n?m(we(n),e):e},createTupleTypeNode:Fe,updateTupleTypeNode:function(e,n){return e.elements!==n?m(Fe(n),e):e},createNamedTupleMember:Ge,updateNamedTupleMember:function(e,n,t,r,a){return e.dotDotDotToken!==n||e.name!==t||e.questionToken!==r||e.type!==a?m(Ge(n,t,r,a),e):e},createOptionalTypeNode:Be,updateOptionalTypeNode:function(e,n){return e.type!==n?m(Be(n),e):e},createRestTypeNode:Ue,updateRestTypeNode:function(e,n){return e.type!==n?m(Ue(n),e):e},createUnionTypeNode:function(e){return Ve(182,e)},updateUnionTypeNode:function(e,n){return je(e,n)},createIntersectionTypeNode:function(e){return Ve(183,e)},updateIntersectionTypeNode:function(e,n){return je(e,n)},createConditionalTypeNode:He,updateConditionalTypeNode:function(e,n,t,r,a){return e.checkType!==n||e.extendsType!==t||e.trueType!==r||e.falseType!==a?m(He(n,t,r,a),e):e},createInferTypeNode:Ke,updateInferTypeNode:function(e,n){return e.typeParameter!==n?m(Ke(n),e):e},createImportTypeNode:ze,updateImportTypeNode:function(e,n,t,r,a){void 0===a&&(a=e.isTypeOf);return e.argument!==n||e.qualifier!==t||e.typeArguments!==r||e.isTypeOf!==a?m(ze(n,t,r,a),e):e},createParenthesizedType:qe,updateParenthesizedType:function(e,n){return e.type!==n?m(qe(n),e):e},createThisTypeNode:function(){var e=I(187);return e.transformFlags=1,e},createTypeOperatorNode:Je,updateTypeOperatorNode:function(e,n){return e.type!==n?m(Je(e.operator,n),e):e},createIndexedAccessTypeNode:Xe,updateIndexedAccessTypeNode:function(e,n,t){return e.objectType!==n||e.indexType!==t?m(Xe(n,t),e):e},createMappedTypeNode:Ye,updateMappedTypeNode:function(e,n,t,r,a,i){return e.readonlyToken!==n||e.typeParameter!==t||e.nameType!==r||e.questionToken!==a||e.type!==i?m(Ye(n,t,r,a,i),e):e},createLiteralTypeNode:Qe,updateLiteralTypeNode:function(e,n){return e.literal!==n?m(Qe(n),e):e},createTemplateLiteralType:We,updateTemplateLiteralType:function(e,n,t){return e.head!==n||e.templateSpans!==t?m(We(n,t),e):e},createObjectBindingPattern:Ze,updateObjectBindingPattern:function(e,n){return e.elements!==n?m(Ze(n),e):e},createArrayBindingPattern:$e,updateArrayBindingPattern:function(e,n){return e.elements!==n?m($e(n),e):e},createBindingElement:en,updateBindingElement:function(e,n,t,r,a){return e.propertyName!==t||e.dotDotDotToken!==n||e.name!==r||e.initializer!==a?m(en(n,t,r,a),e):e},createArrayLiteralExpression:tn,updateArrayLiteralExpression:function(e,n){return e.elements!==n?m(tn(n,e.multiLine),e):e},createObjectLiteralExpression:rn,updateObjectLiteralExpression:function(e,n){return e.properties!==n?m(rn(n,e.multiLine),e):e},createPropertyAccessExpression:4&r?function(n,t){return e.setEmitFlags(an(n,t),131072)}:an,updatePropertyAccessExpression:function(n,t,r){if(e.isPropertyAccessChain(n))return sn(n,t,n.questionDotToken,e.cast(r,e.isIdentifier));return n.expression!==t||n.name!==r?m(an(t,r),n):n},createPropertyAccessChain:4&r?function(n,t,r){return e.setEmitFlags(on(n,t,r),131072)}:on,updatePropertyAccessChain:sn,createElementAccessExpression:cn,updateElementAccessExpression:function(n,t,r){if(e.isElementAccessChain(n))return un(n,t,n.questionDotToken,r);return n.expression!==t||n.argumentExpression!==r?m(cn(t,r),n):n},createElementAccessChain:ln,updateElementAccessChain:un,createCallExpression:dn,updateCallExpression:function(n,t,r,a){if(e.isCallChain(n))return mn(n,t,n.questionDotToken,r,a);return n.expression!==t||n.typeArguments!==r||n.arguments!==a?m(dn(t,r,a),n):n},createCallChain:pn,updateCallChain:mn,createNewExpression:fn,updateNewExpression:function(e,n,t,r){return e.expression!==n||e.typeArguments!==t||e.arguments!==r?m(fn(n,t,r),e):e},createTaggedTemplateExpression:_n,updateTaggedTemplateExpression:function(e,n,t,r){return e.tag!==n||e.typeArguments!==t||e.template!==r?m(_n(n,t,r),e):e},createTypeAssertion:gn,updateTypeAssertion:yn,createParenthesizedExpression:vn,updateParenthesizedExpression:hn,createFunctionExpression:bn,updateFunctionExpression:En,createArrowFunction:Tn,updateArrowFunction:Sn,createDeleteExpression:xn,updateDeleteExpression:function(e,n){return e.expression!==n?m(xn(n),e):e},createTypeOfExpression:Ln,updateTypeOfExpression:function(e,n){return e.expression!==n?m(Ln(n),e):e},createVoidExpression:An,updateVoidExpression:function(e,n){return e.expression!==n?m(An(n),e):e},createAwaitExpression:Dn,updateAwaitExpression:function(e,n){return e.expression!==n?m(Dn(n),e):e},createPrefixUnaryExpression:Cn,updatePrefixUnaryExpression:function(e,n){return e.operand!==n?m(Cn(e.operator,n),e):e},createPostfixUnaryExpression:In,updatePostfixUnaryExpression:function(e,n){return e.operand!==n?m(In(n,e.operator),e):e},createBinaryExpression:Nn,updateBinaryExpression:function(e,n,t,r){return e.left!==n||e.operatorToken!==t||e.right!==r?m(Nn(n,t,r),e):e},createConditionalExpression:kn,updateConditionalExpression:function(e,n,t,r,a,i){return e.condition!==n||e.questionToken!==t||e.whenTrue!==r||e.colonToken!==a||e.whenFalse!==i?m(kn(n,t,r,a,i),e):e},createTemplateExpression:On,updateTemplateExpression:function(e,n,t){return e.head!==n||e.templateSpans!==t?m(On(n,t),e):e},createTemplateHead:function(e,n,t){return Pn(15,e,n,t)},createTemplateMiddle:function(e,n,t){return Pn(16,e,n,t)},createTemplateTail:function(e,n,t){return Pn(17,e,n,t)},createNoSubstitutionTemplateLiteral:function(e,n,t){return Pn(14,e,n,t)},createTemplateLiteralLikeNode:Mn,createYieldExpression:wn,updateYieldExpression:function(e,n,t){return e.expression!==t||e.asteriskToken!==n?m(wn(n,t),e):e},createSpreadElement:Fn,updateSpreadElement:function(e,n){return e.expression!==n?m(Fn(n),e):e},createClassExpression:Gn,updateClassExpression:Bn,createOmittedExpression:function(){return nn(222)},createExpressionWithTypeArguments:Un,updateExpressionWithTypeArguments:function(e,n,t){return e.expression!==n||e.typeArguments!==t?m(Un(n,t),e):e},createAsExpression:Vn,updateAsExpression:jn,createNonNullExpression:Hn,updateNonNullExpression:Kn,createNonNullChain:Wn,updateNonNullChain:zn,createMetaProperty:qn,updateMetaProperty:function(e,n){return e.name!==n?m(qn(e.keywordToken,n),e):e},createTemplateSpan:Jn,updateTemplateSpan:function(e,n,t){return e.expression!==n||e.literal!==t?m(Jn(n,t),e):e},createSemicolonClassElement:function(){var e=I(229);return e.transformFlags|=256,e},createBlock:Xn,updateBlock:function(e,n){return e.statements!==n?m(Xn(n,e.multiLine),e):e},createVariableStatement:Yn,updateVariableStatement:Qn,createEmptyStatement:Zn,createExpressionStatement:$n,updateExpressionStatement:function(e,n){return e.expression!==n?m($n(n),e):e},createIfStatement:et,updateIfStatement:function(e,n,t,r){return e.expression!==n||e.thenStatement!==t||e.elseStatement!==r?m(et(n,t,r),e):e},createDoStatement:nt,updateDoStatement:function(e,n,t){return e.statement!==n||e.expression!==t?m(nt(n,t),e):e},createWhileStatement:tt,updateWhileStatement:function(e,n,t){return e.expression!==n||e.statement!==t?m(tt(n,t),e):e},createForStatement:rt,updateForStatement:function(e,n,t,r,a){return e.initializer!==n||e.condition!==t||e.incrementor!==r||e.statement!==a?m(rt(n,t,r,a),e):e},createForInStatement:at,updateForInStatement:function(e,n,t,r){return e.initializer!==n||e.expression!==t||e.statement!==r?m(at(n,t,r),e):e},createForOfStatement:it,updateForOfStatement:function(e,n,t,r,a){return e.awaitModifier!==n||e.initializer!==t||e.expression!==r||e.statement!==a?m(it(n,t,r,a),e):e},createContinueStatement:ot,updateContinueStatement:function(e,n){return e.label!==n?m(ot(n),e):e},createBreakStatement:st,updateBreakStatement:function(e,n){return e.label!==n?m(st(n),e):e},createReturnStatement:ct,updateReturnStatement:function(e,n){return e.expression!==n?m(ct(n),e):e},createWithStatement:lt,updateWithStatement:function(e,n,t){return e.expression!==n||e.statement!==t?m(lt(n,t),e):e},createSwitchStatement:ut,updateSwitchStatement:function(e,n,t){return e.expression!==n||e.caseBlock!==t?m(ut(n,t),e):e},createLabeledStatement:dt,updateLabeledStatement:pt,createThrowStatement:mt,updateThrowStatement:function(e,n){return e.expression!==n?m(mt(n),e):e},createTryStatement:ft,updateTryStatement:function(e,n,t,r){return e.tryBlock!==n||e.catchClause!==t||e.finallyBlock!==r?m(ft(n,t,r),e):e},createDebuggerStatement:function(){return I(248)},createVariableDeclaration:_t,updateVariableDeclaration:function(e,n,t,r,a){return e.name!==n||e.type!==r||e.exclamationToken!==t||e.initializer!==a?m(_t(n,t,r,a),e):e},createVariableDeclarationList:gt,updateVariableDeclarationList:function(e,n){return e.declarations!==n?m(gt(n,e.flags),e):e},createFunctionDeclaration:yt,updateFunctionDeclaration:vt,createClassDeclaration:ht,updateClassDeclaration:bt,createInterfaceDeclaration:Et,updateInterfaceDeclaration:Tt,createTypeAliasDeclaration:St,updateTypeAliasDeclaration:xt,createEnumDeclaration:Lt,updateEnumDeclaration:At,createModuleDeclaration:Dt,updateModuleDeclaration:Ct,createModuleBlock:It,updateModuleBlock:function(e,n){return e.statements!==n?m(It(n),e):e},createCaseBlock:Nt,updateCaseBlock:function(e,n){return e.clauses!==n?m(Nt(n),e):e},createNamespaceExportDeclaration:Rt,updateNamespaceExportDeclaration:function(e,n){return e.name!==n?m(Rt(n),e):e},createImportEqualsDeclaration:kt,updateImportEqualsDeclaration:Ot,createImportDeclaration:Pt,updateImportDeclaration:Mt,createImportClause:wt,updateImportClause:function(e,n,t,r){return e.isTypeOnly!==n||e.name!==t||e.namedBindings!==r?m(wt(n,t,r),e):e},createNamespaceImport:Ft,updateNamespaceImport:function(e,n){return e.name!==n?m(Ft(n),e):e},createNamespaceExport:Gt,updateNamespaceExport:function(e,n){return e.name!==n?m(Gt(n),e):e},createNamedImports:Bt,updateNamedImports:function(e,n){return e.elements!==n?m(Bt(n),e):e},createImportSpecifier:Ut,updateImportSpecifier:function(e,n,t){return e.propertyName!==n||e.name!==t?m(Ut(n,t),e):e},createExportAssignment:Vt,updateExportAssignment:jt,createExportDeclaration:Ht,updateExportDeclaration:Kt,createNamedExports:Wt,updateNamedExports:function(e,n){return e.elements!==n?m(Wt(n),e):e},createExportSpecifier:zt,updateExportSpecifier:function(e,n,t){return e.propertyName!==n||e.name!==t?m(zt(n,t),e):e},createMissingDeclaration:function(){return N(271,void 0,void 0)},createExternalModuleReference:qt,updateExternalModuleReference:function(e,n){return e.expression!==n?m(qt(n),e):e},get createJSDocAllType(){return b(303)},get createJSDocUnknownType(){return b(304)},get createJSDocNonNullableType(){return E(306)},get updateJSDocNonNullableType(){return T(306)},get createJSDocNullableType(){return E(305)},get updateJSDocNullableType(){return T(305)},get createJSDocOptionalType(){return E(307)},get updateJSDocOptionalType(){return T(307)},get createJSDocVariadicType(){return E(309)},get updateJSDocVariadicType(){return T(309)},get createJSDocNamepathType(){return E(310)},get updateJSDocNamepathType(){return T(310)},createJSDocFunctionType:Xt,updateJSDocFunctionType:function(e,n,t){return e.parameters!==n||e.type!==t?m(Xt(n,t),e):e},createJSDocTypeLiteral:Yt,updateJSDocTypeLiteral:function(e,n,t){return e.jsDocPropertyTags!==n||e.isArrayType!==t?m(Yt(n,t),e):e},createJSDocTypeExpression:Qt,updateJSDocTypeExpression:function(e,n){return e.type!==n?m(Qt(n),e):e},createJSDocSignature:Zt,updateJSDocSignature:function(e,n,t,r){return e.typeParameters!==n||e.parameters!==t||e.type!==r?m(Zt(n,t,r),e):e},createJSDocTemplateTag:nr,updateJSDocTemplateTag:function(e,n,t,r,a){void 0===n&&(n=$t(e));return e.tagName!==n||e.constraint!==t||e.typeParameters!==r||e.comment!==a?m(nr(n,t,r,a),e):e},createJSDocTypedefTag:tr,updateJSDocTypedefTag:function(e,n,t,r,a){void 0===n&&(n=$t(e));return e.tagName!==n||e.typeExpression!==t||e.fullName!==r||e.comment!==a?m(tr(n,t,r,a),e):e},createJSDocParameterTag:rr,updateJSDocParameterTag:function(e,n,t,r,a,i,o){void 0===n&&(n=$t(e));return e.tagName!==n||e.name!==t||e.isBracketed!==r||e.typeExpression!==a||e.isNameFirst!==i||e.comment!==o?m(rr(n,t,r,a,i,o),e):e},createJSDocPropertyTag:ar,updateJSDocPropertyTag:function(e,n,t,r,a,i,o){void 0===n&&(n=$t(e));return e.tagName!==n||e.name!==t||e.isBracketed!==r||e.typeExpression!==a||e.isNameFirst!==i||e.comment!==o?m(ar(n,t,r,a,i,o),e):e},createJSDocCallbackTag:ir,updateJSDocCallbackTag:function(e,n,t,r,a){void 0===n&&(n=$t(e));return e.tagName!==n||e.typeExpression!==t||e.fullName!==r||e.comment!==a?m(ir(n,t,r,a),e):e},createJSDocAugmentsTag:or,updateJSDocAugmentsTag:function(e,n,t,r){void 0===n&&(n=$t(e));return e.tagName!==n||e.class!==t||e.comment!==r?m(or(n,t,r),e):e},createJSDocImplementsTag:sr,updateJSDocImplementsTag:function(e,n,t,r){void 0===n&&(n=$t(e));return e.tagName!==n||e.class!==t||e.comment!==r?m(sr(n,t,r),e):e},createJSDocSeeTag:cr,updateJSDocSeeTag:function(e,n,t,r){return e.tagName!==n||e.name!==t||e.comment!==r?m(cr(n,t,r),e):e},createJSDocNameReference:lr,updateJSDocNameReference:function(e,n){return e.name!==n?m(lr(n),e):e},get createJSDocTypeTag(){return L(329)},get updateJSDocTypeTag(){return A(329)},get createJSDocReturnTag(){return L(327)},get updateJSDocReturnTag(){return A(327)},get createJSDocThisTag(){return L(328)},get updateJSDocThisTag(){return A(328)},get createJSDocEnumTag(){return L(325)},get updateJSDocEnumTag(){return A(325)},get createJSDocAuthorTag(){return S(317)},get updateJSDocAuthorTag(){return x(317)},get createJSDocClassTag(){return S(319)},get updateJSDocClassTag(){return x(319)},get createJSDocPublicTag(){return S(320)},get updateJSDocPublicTag(){return x(320)},get createJSDocPrivateTag(){return S(321)},get updateJSDocPrivateTag(){return x(321)},get createJSDocProtectedTag(){return S(322)},get updateJSDocProtectedTag(){return x(322)},get createJSDocReadonlyTag(){return S(323)},get updateJSDocReadonlyTag(){return x(323)},get createJSDocDeprecatedTag(){return S(318)},get updateJSDocDeprecatedTag(){return x(318)},createJSDocUnknownTag:pr,updateJSDocUnknownTag:function(e,n,t){return e.tagName!==n||e.comment!==t?m(pr(n,t),e):e},createJSDocComment:mr,updateJSDocComment:function(e,n,t){return e.comment!==n||e.tags!==t?m(mr(n,t),e):e},createJsxElement:fr,updateJsxElement:function(e,n,t,r){return e.openingElement!==n||e.children!==t||e.closingElement!==r?m(fr(n,t,r),e):e},createJsxSelfClosingElement:_r,updateJsxSelfClosingElement:function(e,n,t,r){return e.tagName!==n||e.typeArguments!==t||e.attributes!==r?m(_r(n,t,r),e):e},createJsxOpeningElement:gr,updateJsxOpeningElement:function(e,n,t,r){return e.tagName!==n||e.typeArguments!==t||e.attributes!==r?m(gr(n,t,r),e):e},createJsxClosingElement:yr,updateJsxClosingElement:function(e,n){return e.tagName!==n?m(yr(n),e):e},createJsxFragment:vr,createJsxText:hr,updateJsxText:function(e,n,t){return e.text!==n||e.containsOnlyTriviaWhiteSpaces!==t?m(hr(n,t),e):e},createJsxOpeningFragment:function(){var e=I(278);return e.transformFlags|=2,e},createJsxJsxClosingFragment:function(){var e=I(279);return e.transformFlags|=2,e},updateJsxFragment:function(e,n,t,r){return e.openingFragment!==n||e.children!==t||e.closingFragment!==r?m(vr(n,t,r),e):e},createJsxAttribute:br,updateJsxAttribute:function(e,n,t){return e.name!==n||e.initializer!==t?m(br(n,t),e):e},createJsxAttributes:Er,updateJsxAttributes:function(e,n){return e.properties!==n?m(Er(n),e):e},createJsxSpreadAttribute:Tr,updateJsxSpreadAttribute:function(e,n){return e.expression!==n?m(Tr(n),e):e},createJsxExpression:Sr,updateJsxExpression:function(e,n){return e.expression!==n?m(Sr(e.dotDotDotToken,n),e):e},createCaseClause:xr,updateCaseClause:function(e,n,t){return e.expression!==n||e.statements!==t?m(xr(n,t),e):e},createDefaultClause:Lr,updateDefaultClause:function(e,n){return e.statements!==n?m(Lr(n),e):e},createHeritageClause:Ar,updateHeritageClause:function(e,n){return e.types!==n?m(Ar(e.token,n),e):e},createCatchClause:Dr,updateCatchClause:function(e,n,t){return e.variableDeclaration!==n||e.block!==t?m(Dr(n,t),e):e},createPropertyAssignment:Cr,updatePropertyAssignment:function(e,n,t){return e.name!==n||e.initializer!==t?function(e,n){n.decorators&&(e.decorators=n.decorators);n.modifiers&&(e.modifiers=n.modifiers);n.questionToken&&(e.questionToken=n.questionToken);n.exclamationToken&&(e.exclamationToken=n.exclamationToken);return m(e,n)}(Cr(n,t),e):e},createShorthandPropertyAssignment:Ir,updateShorthandPropertyAssignment:function(e,n,t){return e.name!==n||e.objectAssignmentInitializer!==t?function(e,n){n.decorators&&(e.decorators=n.decorators);n.modifiers&&(e.modifiers=n.modifiers);n.equalsToken&&(e.equalsToken=n.equalsToken);n.questionToken&&(e.questionToken=n.questionToken);n.exclamationToken&&(e.exclamationToken=n.exclamationToken);return m(e,n)}(Ir(n,t),e):e},createSpreadAssignment:Nr,updateSpreadAssignment:function(e,n){return e.expression!==n?m(Nr(n),e):e},createEnumMember:Rr,updateEnumMember:function(e,n,t){return e.name!==n||e.initializer!==t?m(Rr(n,t),e):e},createSourceFile:function(e,n,t){var r=p.createBaseSourceFileNode(297);return r.statements=C(e),r.endOfFileToken=n,r.flags|=t,r.fileName="",r.text="",r.languageVersion=0,r.languageVariant=0,r.scriptKind=0,r.isDeclarationFile=!1,r.hasNoDefaultLib=!1,r.transformFlags|=u(r.statements)|l(r.endOfFileToken),r},updateSourceFile:function(n,t,r,a,i,o,s){void 0===r&&(r=n.isDeclarationFile);void 0===a&&(a=n.referencedFiles);void 0===i&&(i=n.typeReferenceDirectives);void 0===o&&(o=n.hasNoDefaultLib);void 0===s&&(s=n.libReferenceDirectives);return n.statements!==t||n.isDeclarationFile!==r||n.referencedFiles!==a||n.typeReferenceDirectives!==i||n.hasNoDefaultLib!==o||n.libReferenceDirectives!==s?m(function(n,t,r,a,i,o,s){var c=p.createBaseSourceFileNode(297);for(var d in n)"emitNode"!==d&&!e.hasProperty(c,d)&&e.hasProperty(n,d)&&(c[d]=n[d]);return c.flags|=n.flags,c.statements=C(t),c.endOfFileToken=n.endOfFileToken,c.isDeclarationFile=r,c.referencedFiles=a,c.typeReferenceDirectives=i,c.hasNoDefaultLib=o,c.libReferenceDirectives=s,c.transformFlags=u(c.statements)|l(c.endOfFileToken),c}(n,t,r,a,i,o,s),n):n},createBundle:kr,updateBundle:function(n,t,r){void 0===r&&(r=e.emptyArray);return n.sourceFiles!==t||n.prepends!==r?m(kr(t,r),n):n},createUnparsedSource:function(n,t,r){var a=I(299);return a.prologues=n,a.syntheticReferences=t,a.texts=r,a.fileName="",a.text="",a.referencedFiles=e.emptyArray,a.libReferenceDirectives=e.emptyArray,a.getLineAndCharacterOfPosition=function(n){return e.getLineAndCharacterOfPosition(a,n)},a},createUnparsedPrologue:function(e){return Or(292,e)},createUnparsedPrepend:function(e,n){var t=Or(293,e);return t.texts=n,t},createUnparsedTextLike:function(e,n){return Or(n?295:294,e)},createUnparsedSyntheticReference:function(e){var n=I(296);return n.data=e.data,n.section=e,n},createInputFiles:function(){var e=I(300);return e.javascriptText="",e.declarationText="",e},createSyntheticExpression:function(e,n,t){void 0===n&&(n=!1);var r=I(227);return r.type=e,r.isSpread=n,r.tupleNameSource=t,r},createSyntaxList:function(e){var n=I(334);return n._children=e,n},createNotEmittedStatement:function(n){var t=I(335);return t.original=n,e.setTextRange(t,n),t},createPartiallyEmittedExpression:Pr,updatePartiallyEmittedExpression:Mr,createCommaListExpression:Fr,updateCommaListExpression:function(e,n){return e.elements!==n?m(Fr(n),e):e},createEndOfDeclarationMarker:function(e){var n=I(339);return n.emitNode={},n.original=e,n},createMergeDeclarationMarker:function(e){var n=I(338);return n.emitNode={},n.original=e,n},createSyntheticReferenceExpression:Gr,updateSyntheticReferenceExpression:function(e,n,t){return e.expression!==n||e.thisArg!==t?m(Gr(n,t),e):e},cloneNode:Br,get createComma(){return g(27)},get createAssignment(){return g(62)},get createLogicalOr(){return g(56)},get createLogicalAnd(){return g(55)},get createBitwiseOr(){return g(51)},get createBitwiseXor(){return g(52)},get createBitwiseAnd(){return g(50)},get createStrictEquality(){return g(36)},get createStrictInequality(){return g(37)},get createEquality(){return g(34)},get createInequality(){return g(35)},get createLessThan(){return g(29)},get createLessThanEquals(){return g(32)},get createGreaterThan(){return g(31)},get createGreaterThanEquals(){return g(33)},get createLeftShift(){return g(47)},get createRightShift(){return g(48)},get createUnsignedRightShift(){return g(49)},get createAdd(){return g(39)},get createSubtract(){return g(40)},get createMultiply(){return g(41)},get createDivide(){return g(43)},get createModulo(){return g(44)},get createExponent(){return g(42)},get createPrefixPlus(){return v(39)},get createPrefixMinus(){return v(40)},get createPrefixIncrement(){return v(45)},get createPrefixDecrement(){return v(46)},get createBitwiseNot(){return v(54)},get createLogicalNot(){return v(53)},get createPostfixIncrement(){return h(45)},get createPostfixDecrement(){return h(46)},createImmediatelyInvokedFunctionExpression:function(e,n,t){return dn(bn(void 0,void 0,void 0,void 0,n?[n]:[],void 0,Xn(e,!0)),void 0,t?[t]:[])},createImmediatelyInvokedArrowFunction:function(e,n,t){return dn(Tn(void 0,void 0,n?[n]:[],void 0,void 0,Xn(e,!0)),void 0,t?[t]:[])},createVoidZero:Ur,createExportDefault:function(e){return Vt(void 0,void 0,!1,e)},createExternalModuleExport:function(e){return Ht(void 0,void 0,!1,Wt([zt(void 0,e)]))},createTypeCheck:function(e,n){return"undefined"===n?D.createStrictEquality(e,Ur()):D.createStrictEquality(Ln(e),W(n))},createMethodCall:Vr,createGlobalMethodCall:jr,createFunctionBindCall:function(e,n,t){return Vr(e,"bind",__spreadArrays([n],t))},createFunctionCallCall:function(e,n,t){return Vr(e,"call",__spreadArrays([n],t))},createFunctionApplyCall:function(e,n,t){return Vr(e,"apply",[n,t])},createArraySliceCall:function(e,n){return Vr(e,"slice",void 0===n?[]:[ea(n)])},createArrayConcatCall:function(e,n){return Vr(e,"concat",n)},createObjectDefinePropertyCall:function(e,n,t){return jr("Object","defineProperty",[e,ea(n),t])},createPropertyDescriptor:function(n,t){var r=[];Hr(r,"enumerable",ea(n.enumerable)),Hr(r,"configurable",ea(n.configurable));var a=Hr(r,"writable",ea(n.writable));a=Hr(r,"value",n.value)||a;var i=Hr(r,"get",n.get);return i=Hr(r,"set",n.set)||i,e.Debug.assert(!(a&&i),"A PropertyDescriptor may not be both an accessor descriptor and a data descriptor."),rn(r,!t)},createCallBinding:function(n,t,r,a){void 0===a&&(a=!1);var i,o,s=e.skipOuterExpressions(n,15);e.isSuperProperty(s)?(i=ee(),o=s):e.isSuperKeyword(s)?(i=ee(),o=void 0!==r&&r<2?e.setTextRange(X("_super"),s):s):4096&e.getEmitFlags(s)?(i=Ur(),o=f().parenthesizeLeftSideOfAccess(s)):e.isPropertyAccessExpression(s)?Kr(s.expression,a)?(i=Y(t),o=an(e.setTextRange(D.createAssignment(i,s.expression),s.expression),s.name),e.setTextRange(o,s)):(i=s.expression,o=s):e.isElementAccessExpression(s)?Kr(s.expression,a)?(i=Y(t),o=cn(e.setTextRange(D.createAssignment(i,s.expression),s.expression),s.argumentExpression),e.setTextRange(o,s)):(i=s.expression,o=s):(i=Ur(),o=f().parenthesizeLeftSideOfAccess(n));return{target:o,thisArg:i}},inlineExpressions:function(n){return n.length>10?Fr(n):e.reduceLeft(n,D.createComma)},getInternalName:function(e,n,t){return Wr(e,n,t,49152)},getLocalName:function(e,n,t){return Wr(e,n,t,16384)},getExportName:zr,getDeclarationName:function(e,n,t){return Wr(e,n,t)},getNamespaceMemberName:qr,getExternalModuleOrNamespaceExportName:function(n,t,r,a){if(n&&e.hasSyntacticModifier(t,1))return qr(n,Wr(t),r,a);return zr(t,r,a)},restoreOuterExpressions:function n(t,r,a){void 0===a&&(a=15);if(t&&e.isOuterExpression(t,a)&&(i=t,!(e.isParenthesizedExpression(i)&&e.nodeIsSynthesized(i)&&e.nodeIsSynthesized(e.getSourceMapRange(i))&&e.nodeIsSynthesized(e.getCommentRange(i)))||e.some(e.getSyntheticLeadingComments(i))||e.some(e.getSyntheticTrailingComments(i))))return function(e,n){switch(e.kind){case 207:return hn(e,n);case 206:return yn(e,e.type,n);case 224:return jn(e,n,e.type);case 225:return Kn(e,n);case 336:return Mr(e,n)}}(t,n(t.expression,r));var i;return r},restoreEnclosingLabel:function n(t,r,a){if(!r)return t;var i=pt(r,r.label,e.isLabeledStatement(r.statement)?n(t,r.statement):t);a&&a(r);return i},createUseStrictPrologue:Jr,copyPrologue:function(e,n,t,r){var a=Xr(e,n,t);return Yr(e,n,a,r)},copyStandardPrologue:Xr,copyCustomPrologue:Yr,ensureUseStrict:function(n){if(!e.findUseStrictPrologue(n))return e.setTextRange(C(__spreadArrays([Jr()],n)),n);return n},liftToBlock:function(n){return e.Debug.assert(e.every(n,e.isStatementOrBlock),"Cannot lift nodes to a Block."),e.singleOrUndefined(n)||Xn(n)},mergeLexicalEnvironment:function(n,t){if(!e.some(t))return n;var r=Qr(n,e.isPrologueDirective,0),a=Qr(n,e.isHoistedFunction,r),i=Qr(n,e.isHoistedVariableStatement,a),o=Qr(t,e.isPrologueDirective,0),s=Qr(t,e.isHoistedFunction,o),c=Qr(t,e.isHoistedVariableStatement,s),l=Qr(t,e.isCustomPrologue,c);e.Debug.assert(l===t.length,"Expected declarations to be valid standard or custom prologues");var u=e.isNodeArray(n)?n.slice():n;l>c&&u.splice.apply(u,__spreadArrays([i,0],t.slice(c,l)));c>s&&u.splice.apply(u,__spreadArrays([a,0],t.slice(s,c)));s>o&&u.splice.apply(u,__spreadArrays([r,0],t.slice(o,s)));if(o>0)if(0===r)u.splice.apply(u,__spreadArrays([0,0],t.slice(0,o)));else{for(var d=new e.Map,p=0;p<r;p++){var m=n[p];d.set(m.expression.text,!0)}for(p=o-1;p>=0;p--){var f=t[p];d.has(f.expression.text)||u.unshift(f)}}if(e.isNodeArray(n))return e.setTextRange(C(u,n.hasTrailingComma),n);return n},updateModifiers:function(n,t){var r;"number"==typeof t&&(t=ae(t));return e.isParameter(n)?le(n,n.decorators,t,n.dotDotDotToken,n.name,n.questionToken,n.type,n.initializer):e.isPropertySignature(n)?pe(n,t,n.name,n.questionToken,n.type):e.isPropertyDeclaration(n)?fe(n,n.decorators,t,n.name,null!==(r=n.questionToken)&&void 0!==r?r:n.exclamationToken,n.type,n.initializer):e.isMethodSignature(n)?ge(n,t,n.name,n.questionToken,n.typeParameters,n.parameters,n.type):e.isMethodDeclaration(n)?ve(n,n.decorators,t,n.asteriskToken,n.name,n.questionToken,n.typeParameters,n.parameters,n.type,n.body):e.isConstructorDeclaration(n)?be(n,n.decorators,t,n.parameters,n.body):e.isGetAccessorDeclaration(n)?Te(n,n.decorators,t,n.name,n.parameters,n.type,n.body):e.isSetAccessorDeclaration(n)?xe(n,n.decorators,t,n.name,n.parameters,n.body):e.isIndexSignatureDeclaration(n)?Ce(n,n.decorators,t,n.parameters,n.type):e.isFunctionExpression(n)?En(n,t,n.asteriskToken,n.name,n.typeParameters,n.parameters,n.type,n.body):e.isArrowFunction(n)?Sn(n,t,n.typeParameters,n.parameters,n.type,n.equalsGreaterThanToken,n.body):e.isClassExpression(n)?Bn(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members):e.isVariableStatement(n)?Qn(n,t,n.declarationList):e.isFunctionDeclaration(n)?vt(n,n.decorators,t,n.asteriskToken,n.name,n.typeParameters,n.parameters,n.type,n.body):e.isClassDeclaration(n)?bt(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members):e.isInterfaceDeclaration(n)?Tt(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members):e.isTypeAliasDeclaration(n)?xt(n,n.decorators,t,n.name,n.typeParameters,n.type):e.isEnumDeclaration(n)?At(n,n.decorators,t,n.name,n.members):e.isModuleDeclaration(n)?Ct(n,n.decorators,t,n.name,n.body):e.isImportEqualsDeclaration(n)?Ot(n,n.decorators,t,n.name,n.moduleReference):e.isImportDeclaration(n)?Mt(n,n.decorators,t,n.importClause,n.moduleSpecifier):e.isExportAssignment(n)?jt(n,n.decorators,t,n.expression):e.isExportDeclaration(n)?Kt(n,n.decorators,t,n.isTypeOnly,n.exportClause,n.moduleSpecifier):e.Debug.assertNever(n)}};return D;function C(n,t){if(void 0===n||n===e.emptyArray)n=[];else if(e.isNodeArray(n))return void 0===n.transformFlags&&d(n),e.Debug.attachNodeArrayDebugInfo(n),n;var r=n.length,a=r>=1&&r<=4?n.slice():n;return e.setTextRangePosEnd(a,-1,-1),a.hasTrailingComma=!!t,d(a),e.Debug.attachNodeArrayDebugInfo(a),a}function I(e){return p.createBaseNode(e)}function N(e,n,t){var r=I(e);return r.decorators=Zr(n),r.modifiers=Zr(t),r.transformFlags|=u(r.decorators)|u(r.modifiers),r.symbol=void 0,r.localSymbol=void 0,r.locals=void 0,r.nextContainer=void 0,r}function R(n,t,r,a){var i=N(n,t,r);if(a=$r(a),i.name=a,a)switch(i.kind){case 165:case 167:case 168:case 163:case 288:if(e.isIdentifier(a)){i.transformFlags|=c(a);break}default:i.transformFlags|=l(a)}return i}function k(e,n,t,r,a){var i=R(e,n,t,r);return i.typeParameters=Zr(a),i.transformFlags|=u(i.typeParameters),a&&(i.transformFlags|=1),i}function O(e,n,t,r,a,i,o){var s=k(e,n,t,r,a);return s.parameters=C(i),s.type=o,s.transformFlags|=u(s.parameters)|l(s.type),o&&(s.transformFlags|=1),s}function P(e,n){return n.typeArguments&&(e.typeArguments=n.typeArguments),m(e,n)}function M(e,n,t,r,a,i,o,s){var c=O(e,n,t,r,a,i,o);return c.body=s,c.transformFlags|=-8388609&l(c.body),s||(c.transformFlags|=1),c}function w(e,n){return n.exclamationToken&&(e.exclamationToken=n.exclamationToken),n.typeArguments&&(e.typeArguments=n.typeArguments),P(e,n)}function F(e,n,t,r,a,i){var o=k(e,n,t,r,a);return o.heritageClauses=Zr(i),o.transformFlags|=u(o.heritageClauses),o}function G(e,n,t,r,a,i,o){var s=F(e,n,t,r,a,i);return s.members=C(o),s.transformFlags|=u(s.members),s}function B(e,n,t,r,a){var i=R(e,n,t,r);return i.initializer=a,i.transformFlags|=l(i.initializer),i}function U(e,n,t,r,a,i){var o=B(e,n,t,r,i);return o.type=a,o.transformFlags|=l(a),a&&(o.transformFlags|=1),o}function V(e,n){var t=Z(e);return t.text=n,t}function j(e,n){void 0===n&&(n=0);var t=V(8,"number"==typeof e?e+"":e);return t.numericLiteralFlags=n,384&n&&(t.transformFlags|=256),t}function H(n){var t=V(9,"string"==typeof n?n:e.pseudoBigIntToString(n)+"n");return t.transformFlags|=4,t}function K(e,n){var t=V(10,e);return t.singleQuote=n,t}function W(e,n,t){var r=K(e,n);return r.hasExtendedUnicodeEscape=t,t&&(r.transformFlags|=256),r}function z(e){return V(13,e)}function q(n,t){void 0===t&&n&&(t=e.stringToToken(n)),78===t&&(t=void 0);var r=p.createBaseIdentifierNode(78);return r.originalKeywordKind=t,r.escapedText=e.escapeLeadingUnderscores(n),r}function J(e,n){var r=q(e,void 0);return r.autoGenerateFlags=n,r.autoGenerateId=t,t++,r}function X(e,n,t){var r=q(e,t);return n&&(r.typeArguments=C(n)),130===r.originalKeywordKind&&(r.transformFlags|=8388608),r}function Y(e,n){var t=1;n&&(t|=8);var r=J("",t);return e&&e(r),r}function Q(n,t){void 0===t&&(t=0),e.Debug.assert(!(7&t),"Argument out of range: flags");var r=J(n&&e.isIdentifier(n)?e.idText(n):"",4|t);return r.original=n,r}function Z(e){return p.createBaseTokenNode(e)}function $(n){e.Debug.assert(n>=0&&n<=156,"Invalid token"),e.Debug.assert(n<=14||n>=17,"Invalid token. Use 'createTemplateLiteralLikeNode' to create template literals."),e.Debug.assert(n<=8||n>=14,"Invalid token. Use 'createLiteralLikeNode' to create literals."),e.Debug.assert(78!==n,"Invalid token. Use 'createIdentifier' to create identifiers");var t=Z(n),r=0;switch(n){case 129:r=96;break;case 122:case 120:case 121:case 142:case 125:case 133:case 84:case 128:case 144:case 155:case 141:case 145:case 147:case 131:case 148:case 113:case 152:case 150:r=1;break;case 123:case 105:r=256;break;case 107:r=4096}return r&&(t.transformFlags|=r),t}function ee(){return $(107)}function ne(){return $(109)}function te(){return $(94)}function re(e){return $(e)}function ae(e){var n=[];return 1&e&&n.push(re(92)),2&e&&n.push(re(133)),512&e&&n.push(re(87)),2048&e&&n.push(re(84)),4&e&&n.push(re(122)),8&e&&n.push(re(120)),16&e&&n.push(re(121)),128&e&&n.push(re(125)),32&e&&n.push(re(123)),64&e&&n.push(re(142)),256&e&&n.push(re(129)),n}function ie(e,n){var t=I(157);return t.left=e,t.right=$r(n),t.transformFlags|=l(t.left)|c(t.right),t}function oe(e){var n=I(158);return n.expression=f().parenthesizeExpressionOfComputedPropertyName(e),n.transformFlags|=33024|l(n.expression),n}function se(e,n,t){var r=R(159,void 0,void 0,e);return r.constraint=n,r.default=t,r.transformFlags=1,r}function ce(n,t,r,a,i,o,s){var c=U(160,n,t,a,o,s&&f().parenthesizeExpressionForDisallowedComma(s));return c.dotDotDotToken=r,c.questionToken=i,e.isThisIdentifier(c.name)?c.transformFlags=1:(c.transformFlags|=l(c.dotDotDotToken)|l(c.questionToken),i&&(c.transformFlags|=1),92&e.modifiersToFlags(c.modifiers)&&(c.transformFlags|=2048),(s||r)&&(c.transformFlags|=256)),c}function le(e,n,t,r,a,i,o,s){return e.decorators!==n||e.modifiers!==t||e.dotDotDotToken!==r||e.name!==a||e.questionToken!==i||e.type!==o||e.initializer!==s?m(ce(n,t,r,a,i,o,s),e):e}function ue(e){var n=I(161);return n.expression=f().parenthesizeLeftSideOfAccess(e),n.transformFlags|=2049|l(n.expression),n}function de(e,n,t,r){var a=R(162,void 0,e,n);return a.type=r,a.questionToken=t,a.transformFlags=1,a}function pe(e,n,t,r,a){return e.modifiers!==n||e.name!==t||e.questionToken!==r||e.type!==a?m(de(n,t,r,a),e):e}function me(n,t,r,a,i,o){var s=U(163,n,t,r,i,o);return s.questionToken=a&&e.isQuestionToken(a)?a:void 0,s.exclamationToken=a&&e.isExclamationToken(a)?a:void 0,s.transformFlags|=l(s.questionToken)|l(s.exclamationToken)|4194304,(e.isComputedPropertyName(s.name)||e.hasStaticModifier(s)&&s.initializer)&&(s.transformFlags|=2048),(a||2&e.modifiersToFlags(s.modifiers))&&(s.transformFlags|=1),s}function fe(n,t,r,a,i,o,s){return n.decorators!==t||n.modifiers!==r||n.name!==a||n.questionToken!==(void 0!==i&&e.isQuestionToken(i)?i:void 0)||n.exclamationToken!==(void 0!==i&&e.isExclamationToken(i)?i:void 0)||n.type!==o||n.initializer!==s?m(me(t,r,a,i,o,s),n):n}function _e(e,n,t,r,a,i){var o=O(164,void 0,e,n,r,a,i);return o.questionToken=t,o.transformFlags=1,o}function ge(e,n,t,r,a,i,o){return e.modifiers!==n||e.name!==t||e.questionToken!==r||e.typeParameters!==a||e.parameters!==i||e.type!==o?P(_e(n,t,r,a,i,o),e):e}function ye(n,t,r,a,i,o,s,c,u){var d=M(165,n,t,a,o,s,c,u);return d.asteriskToken=r,d.questionToken=i,d.transformFlags|=l(d.asteriskToken)|l(d.questionToken)|256,i&&(d.transformFlags|=1),256&e.modifiersToFlags(d.modifiers)?d.transformFlags|=r?32:64:r&&(d.transformFlags|=512),d}function ve(e,n,t,r,a,i,o,s,c,l){return e.decorators!==n||e.modifiers!==t||e.asteriskToken!==r||e.name!==a||e.questionToken!==i||e.typeParameters!==o||e.parameters!==s||e.type!==c||e.body!==l?w(ye(n,t,r,a,i,o,s,c,l),e):e}function he(e,n,t,r){var a=M(166,e,n,void 0,void 0,t,void 0,r);return a.transformFlags|=256,a}function be(e,n,t,r,a){return e.decorators!==n||e.modifiers!==t||e.parameters!==r||e.body!==a?w(he(n,t,r,a),e):e}function Ee(e,n,t,r,a,i){return M(167,e,n,t,void 0,r,a,i)}function Te(e,n,t,r,a,i,o){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.parameters!==a||e.type!==i||e.body!==o?w(Ee(n,t,r,a,i,o),e):e}function Se(e,n,t,r,a){return M(168,e,n,t,void 0,r,void 0,a)}function xe(e,n,t,r,a,i){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.parameters!==a||e.body!==i?w(Se(n,t,r,a,i),e):e}function Le(e,n,t){var r=O(169,void 0,void 0,void 0,e,n,t);return r.transformFlags=1,r}function Ae(e,n,t){var r=O(170,void 0,void 0,void 0,e,n,t);return r.transformFlags=1,r}function De(e,n,t,r){var a=O(171,e,n,void 0,void 0,t,r);return a.transformFlags=1,a}function Ce(e,n,t,r,a){return e.parameters!==r||e.type!==a||e.decorators!==n||e.modifiers!==t?P(De(n,t,r,a),e):e}function Ie(e,n){var t=I(194);return t.type=e,t.literal=n,t.transformFlags=1,t}function Ne(e,n,t){var r=I(172);return r.assertsModifier=e,r.parameterName=$r(n),r.type=t,r.transformFlags=1,r}function Re(e,n){var t=I(173);return t.typeName=$r(e),t.typeArguments=n&&f().parenthesizeTypeArguments(C(n)),t.transformFlags=1,t}function ke(e,n,t){var r=O(174,void 0,void 0,void 0,e,n,t);return r.transformFlags=1,r}function Oe(e,n,t){var r=O(175,void 0,void 0,void 0,e,n,t);return r.transformFlags=1,r}function Pe(e){var n=I(176);return n.exprName=e,n.transformFlags=1,n}function Me(e){var n=I(177);return n.members=C(e),n.transformFlags=1,n}function we(e){var n=I(178);return n.elementType=f().parenthesizeElementTypeOfArrayType(e),n.transformFlags=1,n}function Fe(e){var n=I(179);return n.elements=C(e),n.transformFlags=1,n}function Ge(e,n,t,r){var a=I(192);return a.dotDotDotToken=e,a.name=n,a.questionToken=t,a.type=r,a.transformFlags=1,a}function Be(e){var n=I(180);return n.type=f().parenthesizeElementTypeOfArrayType(e),n.transformFlags=1,n}function Ue(e){var n=I(181);return n.type=e,n.transformFlags=1,n}function Ve(e,n){var t=I(e);return t.types=f().parenthesizeConstituentTypesOfUnionOrIntersectionType(n),t.transformFlags=1,t}function je(e,n){return e.types!==n?m(Ve(e.kind,n),e):e}function He(e,n,t,r){var a=I(184);return a.checkType=f().parenthesizeMemberOfConditionalType(e),a.extendsType=f().parenthesizeMemberOfConditionalType(n),a.trueType=t,a.falseType=r,a.transformFlags=1,a}function Ke(e){var n=I(185);return n.typeParameter=e,n.transformFlags=1,n}function We(e,n){var t=I(193);return t.head=e,t.templateSpans=C(n),t.transformFlags=1,t}function ze(e,n,t,r){void 0===r&&(r=!1);var a=I(195);return a.argument=e,a.qualifier=n,a.typeArguments=t&&f().parenthesizeTypeArguments(t),a.isTypeOf=r,a.transformFlags=1,a}function qe(e){var n=I(186);return n.type=e,n.transformFlags=1,n}function Je(e,n){var t=I(188);return t.operator=e,t.type=f().parenthesizeMemberOfElementType(n),t.transformFlags=1,t}function Xe(e,n){var t=I(189);return t.objectType=f().parenthesizeMemberOfElementType(e),t.indexType=n,t.transformFlags=1,t}function Ye(e,n,t,r,a){var i=I(190);return i.readonlyToken=e,i.typeParameter=n,i.nameType=t,i.questionToken=r,i.type=a,i.transformFlags=1,i}function Qe(e){var n=I(191);return n.literal=e,n.transformFlags=1,n}function Ze(e){var n=I(196);return n.elements=C(e),n.transformFlags|=131328|u(n.elements),8192&n.transformFlags&&(n.transformFlags|=16416),n}function $e(e){var n=I(197);return n.elements=C(e),n.transformFlags|=131328|u(n.elements),n}function en(n,t,r,a){var i=B(198,void 0,void 0,r,a);return i.propertyName=$r(t),i.dotDotDotToken=n,i.transformFlags|=256|l(i.dotDotDotToken),i.propertyName&&(i.transformFlags|=e.isIdentifier(i.propertyName)?c(i.propertyName):l(i.propertyName)),n&&(i.transformFlags|=8192),i}function nn(e){return I(e)}function tn(e,n){var t=nn(199);return t.elements=f().parenthesizeExpressionsOfCommaDelimitedList(C(e)),t.multiLine=n,t.transformFlags|=u(t.elements),t}function rn(e,n){var t=nn(200);return t.properties=C(e),t.multiLine=n,t.transformFlags|=u(t.properties),t}function an(n,t){var r=nn(201);return r.expression=f().parenthesizeLeftSideOfAccess(n),r.name=$r(t),r.transformFlags=l(r.expression)|(e.isIdentifier(r.name)?c(r.name):l(r.name)),e.isSuperKeyword(n)&&(r.transformFlags|=96),r}function on(n,t,r){var a=nn(201);return a.flags|=32,a.expression=f().parenthesizeLeftSideOfAccess(n),a.questionDotToken=t,a.name=$r(r),a.transformFlags|=8|l(a.expression)|l(a.questionDotToken)|(e.isIdentifier(a.name)?c(a.name):l(a.name)),a}function sn(n,t,r,a){return e.Debug.assert(!!(32&n.flags),"Cannot update a PropertyAccessExpression using updatePropertyAccessChain. Use updatePropertyAccess instead."),n.expression!==t||n.questionDotToken!==r||n.name!==a?m(on(t,r,a),n):n}function cn(n,t){var r=nn(202);return r.expression=f().parenthesizeLeftSideOfAccess(n),r.argumentExpression=ea(t),r.transformFlags|=l(r.expression)|l(r.argumentExpression),e.isSuperKeyword(n)&&(r.transformFlags|=96),r}function ln(e,n,t){var r=nn(202);return r.flags|=32,r.expression=f().parenthesizeLeftSideOfAccess(e),r.questionDotToken=n,r.argumentExpression=ea(t),r.transformFlags|=l(r.expression)|l(r.questionDotToken)|l(r.argumentExpression)|8,r}function un(n,t,r,a){return e.Debug.assert(!!(32&n.flags),"Cannot update a ElementAccessExpression using updateElementAccessChain. Use updateElementAccess instead."),n.expression!==t||n.questionDotToken!==r||n.argumentExpression!==a?m(ln(t,r,a),n):n}function dn(n,t,r){var a=nn(203);return a.expression=f().parenthesizeLeftSideOfAccess(n),a.typeArguments=Zr(t),a.arguments=f().parenthesizeExpressionsOfCommaDelimitedList(C(r)),a.transformFlags|=l(a.expression)|u(a.typeArguments)|u(a.arguments),a.typeArguments&&(a.transformFlags|=1),e.isImportKeyword(a.expression)?a.transformFlags|=2097152:e.isSuperProperty(a.expression)&&(a.transformFlags|=4096),a}function pn(n,t,r,a){var i=nn(203);return i.flags|=32,i.expression=f().parenthesizeLeftSideOfAccess(n),i.questionDotToken=t,i.typeArguments=Zr(r),i.arguments=f().parenthesizeExpressionsOfCommaDelimitedList(C(a)),i.transformFlags|=l(i.expression)|l(i.questionDotToken)|u(i.typeArguments)|u(i.arguments)|8,i.typeArguments&&(i.transformFlags|=1),e.isSuperProperty(i.expression)&&(i.transformFlags|=4096),i}function mn(n,t,r,a,i){return e.Debug.assert(!!(32&n.flags),"Cannot update a CallExpression using updateCallChain. Use updateCall instead."),n.expression!==t||n.questionDotToken!==r||n.typeArguments!==a||n.arguments!==i?m(pn(t,r,a,i),n):n}function fn(e,n,t){var r=nn(204);return r.expression=f().parenthesizeExpressionOfNew(e),r.typeArguments=Zr(n),r.arguments=t?f().parenthesizeExpressionsOfCommaDelimitedList(t):void 0,r.transformFlags|=l(r.expression)|u(r.typeArguments)|u(r.arguments)|8,r.typeArguments&&(r.transformFlags|=1),r}function _n(n,t,r){var a=nn(205);return a.tag=f().parenthesizeLeftSideOfAccess(n),a.typeArguments=Zr(t),a.template=r,a.transformFlags|=l(a.tag)|u(a.typeArguments)|l(a.template)|256,a.typeArguments&&(a.transformFlags|=1),e.hasInvalidEscape(a.template)&&(a.transformFlags|=32),a}function gn(e,n){var t=nn(206);return t.expression=f().parenthesizeOperandOfPrefixUnary(n),t.type=e,t.transformFlags|=l(t.expression)|l(t.type)|1,t}function yn(e,n,t){return e.type!==n||e.expression!==t?m(gn(n,t),e):e}function vn(e){var n=nn(207);return n.expression=e,n.transformFlags=l(n.expression),n}function hn(e,n){return e.expression!==n?m(vn(n),e):e}function bn(n,t,r,a,i,o,s){var c=M(208,void 0,n,r,a,i,o,s);return c.asteriskToken=t,c.transformFlags|=l(c.asteriskToken),c.typeParameters&&(c.transformFlags|=1),256&e.modifiersToFlags(c.modifiers)?c.asteriskToken?c.transformFlags|=32:c.transformFlags|=64:c.asteriskToken&&(c.transformFlags|=512),c}function En(e,n,t,r,a,i,o,s){return e.name!==r||e.modifiers!==n||e.asteriskToken!==t||e.typeParameters!==a||e.parameters!==i||e.type!==o||e.body!==s?w(bn(n,t,r,a,i,o,s),e):e}function Tn(n,t,r,a,i,o){var s=M(209,void 0,n,void 0,t,r,a,f().parenthesizeConciseBodyOfArrowFunction(o));return s.equalsGreaterThanToken=null!=i?i:$(38),s.transformFlags|=256|l(s.equalsGreaterThanToken),256&e.modifiersToFlags(s.modifiers)&&(s.transformFlags|=64),s}function Sn(e,n,t,r,a,i,o){return e.modifiers!==n||e.typeParameters!==t||e.parameters!==r||e.type!==a||e.equalsGreaterThanToken!==i||e.body!==o?w(Tn(n,t,r,a,i,o),e):e}function xn(e){var n=nn(210);return n.expression=f().parenthesizeOperandOfPrefixUnary(e),n.transformFlags|=l(n.expression),n}function Ln(e){var n=nn(211);return n.expression=f().parenthesizeOperandOfPrefixUnary(e),n.transformFlags|=l(n.expression),n}function An(e){var n=nn(212);return n.expression=f().parenthesizeOperandOfPrefixUnary(e),n.transformFlags|=l(n.expression),n}function Dn(e){var n=nn(213);return n.expression=f().parenthesizeOperandOfPrefixUnary(e),n.transformFlags|=524384|l(n.expression),n}function Cn(e,n){var t=nn(214);return t.operator=e,t.operand=f().parenthesizeOperandOfPrefixUnary(n),t.transformFlags|=l(t.operand),t}function In(e,n){var t=nn(215);return t.operator=n,t.operand=f().parenthesizeOperandOfPostfixUnary(e),t.transformFlags=l(t.operand),t}function Nn(n,t,r){var a,i=nn(216),o="number"==typeof(a=t)?$(a):a,s=o.kind;return i.left=f().parenthesizeLeftSideOfBinary(s,n),i.operatorToken=o,i.right=f().parenthesizeRightSideOfBinary(s,i.left,r),i.transformFlags|=l(i.left)|l(i.operatorToken)|l(i.right),60===s?i.transformFlags|=8:62===s?e.isObjectLiteralExpression(i.left)?i.transformFlags|=1312|Rn(i.left):e.isArrayLiteralExpression(i.left)&&(i.transformFlags|=1280|Rn(i.left)):42===s||66===s?i.transformFlags|=128:e.isLogicalOrCoalescingAssignmentOperator(s)&&(i.transformFlags|=4),i}function Rn(n){if(16384&n.transformFlags)return 16384;if(32&n.transformFlags)for(var t=0,r=e.getElementsOfBindingOrAssignmentPattern(n);t<r.length;t++){var a=r[t],i=e.getTargetOfBindingOrAssignmentElement(a);if(i&&e.isAssignmentPattern(i)){if(16384&i.transformFlags)return 16384;if(32&i.transformFlags){var o=Rn(i);if(o)return o}}}return 0}function kn(e,n,t,r,a){var i=nn(217);return i.condition=f().parenthesizeConditionOfConditionalExpression(e),i.questionToken=null!=n?n:$(57),i.whenTrue=f().parenthesizeBranchOfConditionalExpression(t),i.colonToken=null!=r?r:$(58),i.whenFalse=f().parenthesizeBranchOfConditionalExpression(a),i.transformFlags|=l(i.condition)|l(i.questionToken)|l(i.whenTrue)|l(i.colonToken)|l(i.whenFalse),i}function On(e,n){var t=nn(218);return t.head=e,t.templateSpans=C(n),t.transformFlags|=l(t.head)|u(t.templateSpans)|256,t}function Pn(t,r,a,i){void 0===i&&(i=0),e.Debug.assert(!(-2049&i),"Unsupported template flags.");var o=void 0;if(void 0!==a&&a!==r&&"object"==typeof(o=function(t,r){n||(n=e.createScanner(99,!1,0));switch(t){case 14:n.setText("`"+r+"`");break;case 15:n.setText("`"+r+"${");break;case 16:n.setText("}"+r+"${");break;case 17:n.setText("}"+r+"`")}var a,i=n.scan();23===i&&(i=n.reScanTemplateToken(!1));if(n.isUnterminated())return n.setText(void 0),s;switch(i){case 14:case 15:case 16:case 17:a=n.getTokenValue()}if(void 0===a||1!==n.scan())return n.setText(void 0),s;return n.setText(void 0),a}(t,a)))return e.Debug.fail("Invalid raw text");if(void 0===r){if(void 0===o)return e.Debug.fail("Arguments 'text' and 'rawText' may not both be undefined.");r=o}else void 0!==o&&e.Debug.assert(r===o,"Expected argument 'text' to be the normalized (i.e. 'cooked') version of argument 'rawText'.");return Mn(t,r,a,i)}function Mn(e,n,t,r){var a=Z(e);return a.text=n,a.rawText=t,a.templateFlags=2048&r,a.transformFlags|=256,a.templateFlags&&(a.transformFlags|=32),a}function wn(n,t){e.Debug.assert(!n||!!t,"A `YieldExpression` with an asteriskToken must have an expression.");var r=nn(219);return r.expression=t&&f().parenthesizeExpressionForDisallowedComma(t),r.asteriskToken=n,r.transformFlags|=262432|(l(r.expression)|l(r.asteriskToken)),r}function Fn(e){var n=nn(220);return n.expression=f().parenthesizeExpressionForDisallowedComma(e),n.transformFlags|=8448|l(n.expression),n}function Gn(e,n,t,r,a,i){var o=G(221,e,n,t,r,a,i);return o.transformFlags|=256,o}function Bn(e,n,t,r,a,i,o){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.typeParameters!==a||e.heritageClauses!==i||e.members!==o?m(Gn(n,t,r,a,i,o),e):e}function Un(e,n){var t=I(223);return t.expression=f().parenthesizeLeftSideOfAccess(e),t.typeArguments=n&&f().parenthesizeTypeArguments(n),t.transformFlags|=l(t.expression)|u(t.typeArguments)|256,t}function Vn(e,n){var t=nn(224);return t.expression=e,t.type=n,t.transformFlags|=l(t.expression)|l(t.type)|1,t}function jn(e,n,t){return e.expression!==n||e.type!==t?m(Vn(n,t),e):e}function Hn(e){var n=nn(225);return n.expression=f().parenthesizeLeftSideOfAccess(e),n.transformFlags|=1|l(n.expression),n}function Kn(n,t){return e.isNonNullChain(n)?zn(n,t):n.expression!==t?m(Hn(t),n):n}function Wn(e){var n=nn(225);return n.flags|=32,n.expression=f().parenthesizeLeftSideOfAccess(e),n.transformFlags|=1|l(n.expression),n}function zn(n,t){return e.Debug.assert(!!(32&n.flags),"Cannot update a NonNullExpression using updateNonNullChain. Use updateNonNullExpression instead."),n.expression!==t?m(Wn(t),n):n}function qn(n,t){var r=nn(226);switch(r.keywordToken=n,r.name=t,r.transformFlags|=l(r.name),n){case 102:r.transformFlags|=256;break;case 99:r.transformFlags|=4;break;default:return e.Debug.assertNever(n)}return r}function Jn(e,n){var t=I(228);return t.expression=e,t.literal=n,t.transformFlags|=l(t.expression)|l(t.literal)|256,t}function Xn(e,n){var t=I(230);return t.statements=C(e),t.multiLine=n,t.transformFlags|=u(t.statements),t}function Yn(n,t){var r=N(232,void 0,n);return r.declarationList=e.isArray(t)?gt(t):t,r.transformFlags|=l(r.declarationList),2&e.modifiersToFlags(r.modifiers)&&(r.transformFlags=1),r}function Qn(e,n,t){return e.modifiers!==n||e.declarationList!==t?m(Yn(n,t),e):e}function Zn(){return I(231)}function $n(e){var n=I(233);return n.expression=f().parenthesizeExpressionOfExpressionStatement(e),n.transformFlags|=l(n.expression),n}function et(e,n,t){var r=I(234);return r.expression=e,r.thenStatement=na(n),r.elseStatement=na(t),r.transformFlags|=l(r.expression)|l(r.thenStatement)|l(r.elseStatement),r}function nt(e,n){var t=I(235);return t.statement=na(e),t.expression=n,t.transformFlags|=l(t.statement)|l(t.expression),t}function tt(e,n){var t=I(236);return t.expression=e,t.statement=na(n),t.transformFlags|=l(t.expression)|l(t.statement),t}function rt(e,n,t,r){var a=I(237);return a.initializer=e,a.condition=n,a.incrementor=t,a.statement=na(r),a.transformFlags|=l(a.initializer)|l(a.condition)|l(a.incrementor)|l(a.statement),a}function at(e,n,t){var r=I(238);return r.initializer=e,r.expression=n,r.statement=na(t),r.transformFlags|=l(r.initializer)|l(r.expression)|l(r.statement),r}function it(e,n,t,r){var a=I(239);return a.awaitModifier=e,a.initializer=n,a.expression=f().parenthesizeExpressionForDisallowedComma(t),a.statement=na(r),a.transformFlags|=l(a.awaitModifier)|l(a.initializer)|l(a.expression)|l(a.statement)|256,e&&(a.transformFlags|=32),a}function ot(e){var n=I(240);return n.label=$r(e),n.transformFlags|=1048576|l(n.label),n}function st(e){var n=I(241);return n.label=$r(e),n.transformFlags|=1048576|l(n.label),n}function ct(e){var n=I(242);return n.expression=e,n.transformFlags|=1048608|l(n.expression),n}function lt(e,n){var t=I(243);return t.expression=e,t.statement=na(n),t.transformFlags|=l(t.expression)|l(t.statement),t}function ut(e,n){var t=I(244);return t.expression=f().parenthesizeExpressionForDisallowedComma(e),t.caseBlock=n,t.transformFlags|=l(t.expression)|l(t.caseBlock),t}function dt(e,n){var t=I(245);return t.label=$r(e),t.statement=na(n),t.transformFlags|=l(t.label)|l(t.statement),t}function pt(e,n,t){return e.label!==n||e.statement!==t?m(dt(n,t),e):e}function mt(e){var n=I(246);return n.expression=e,n.transformFlags|=l(n.expression),n}function ft(e,n,t){var r=I(247);return r.tryBlock=e,r.catchClause=n,r.finallyBlock=t,r.transformFlags|=l(r.tryBlock)|l(r.catchClause)|l(r.finallyBlock),r}function _t(e,n,t,r){var a=U(249,void 0,void 0,e,t,r&&f().parenthesizeExpressionForDisallowedComma(r));return a.exclamationToken=n,a.transformFlags|=l(a.exclamationToken),n&&(a.transformFlags|=1),a}function gt(e,n){void 0===n&&(n=0);var t=I(250);return t.flags|=3&n,t.declarations=C(e),t.transformFlags|=1048576|u(t.declarations),3&n&&(t.transformFlags|=65792),t}function yt(n,t,r,a,i,o,s,c){var u=M(251,n,t,a,i,o,s,c);return u.asteriskToken=r,!u.body||2&e.modifiersToFlags(u.modifiers)?u.transformFlags=1:(u.transformFlags|=1048576|l(u.asteriskToken),256&e.modifiersToFlags(u.modifiers)?u.asteriskToken?u.transformFlags|=32:u.transformFlags|=64:u.asteriskToken&&(u.transformFlags|=512)),u}function vt(e,n,t,r,a,i,o,s,c){return e.decorators!==n||e.modifiers!==t||e.asteriskToken!==r||e.name!==a||e.typeParameters!==i||e.parameters!==o||e.type!==s||e.body!==c?w(yt(n,t,r,a,i,o,s,c),e):e}function ht(n,t,r,a,i,o){var s=G(252,n,t,r,a,i,o);return 2&e.modifiersToFlags(s.modifiers)?s.transformFlags=1:(s.transformFlags|=256,2048&s.transformFlags&&(s.transformFlags|=1)),s}function bt(e,n,t,r,a,i,o){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.typeParameters!==a||e.heritageClauses!==i||e.members!==o?m(ht(n,t,r,a,i,o),e):e}function Et(e,n,t,r,a,i){var o=F(253,e,n,t,r,a);return o.members=C(i),o.transformFlags=1,o}function Tt(e,n,t,r,a,i,o){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.typeParameters!==a||e.heritageClauses!==i||e.members!==o?m(Et(n,t,r,a,i,o),e):e}function St(e,n,t,r,a){var i=k(254,e,n,t,r);return i.type=a,i.transformFlags=1,i}function xt(e,n,t,r,a,i){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.typeParameters!==a||e.type!==i?m(St(n,t,r,a,i),e):e}function Lt(e,n,t,r){var a=R(255,e,n,t);return a.members=C(r),a.transformFlags|=1|u(a.members),a.transformFlags&=-8388609,a}function At(e,n,t,r,a){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.members!==a?m(Lt(n,t,r,a),e):e}function Dt(n,t,r,a,i){void 0===i&&(i=0);var o=N(256,n,t);return o.flags|=1044&i,o.name=r,o.body=a,2&e.modifiersToFlags(o.modifiers)?o.transformFlags=1:o.transformFlags|=l(o.name)|l(o.body)|1,o.transformFlags&=-8388609,o}function Ct(e,n,t,r,a){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.body!==a?m(Dt(n,t,r,a,e.flags),e):e}function It(e){var n=I(257);return n.statements=C(e),n.transformFlags|=u(n.statements),n}function Nt(e){var n=I(258);return n.clauses=C(e),n.transformFlags|=u(n.clauses),n}function Rt(e){var n=R(259,void 0,void 0,e);return n.transformFlags=1,n}function kt(n,t,r,a){var i=R(260,n,t,r);return i.moduleReference=a,i.transformFlags|=l(i.moduleReference),e.isExternalModuleReference(i.moduleReference)||(i.transformFlags|=1),i.transformFlags&=-8388609,i}function Ot(e,n,t,r,a){return e.decorators!==n||e.modifiers!==t||e.name!==r||e.moduleReference!==a?m(kt(n,t,r,a),e):e}function Pt(e,n,t,r){var a=N(261,e,n);return a.importClause=t,a.moduleSpecifier=r,a.transformFlags|=l(a.importClause)|l(a.moduleSpecifier),a.transformFlags&=-8388609,a}function Mt(e,n,t,r,a){return e.decorators!==n||e.modifiers!==t||e.importClause!==r||e.moduleSpecifier!==a?m(Pt(n,t,r,a),e):e}function wt(e,n,t){var r=I(262);return r.isTypeOnly=e,r.name=n,r.namedBindings=t,r.transformFlags|=l(r.name)|l(r.namedBindings),e&&(r.transformFlags|=1),r.transformFlags&=-8388609,r}function Ft(e){var n=I(263);return n.name=e,n.transformFlags|=l(n.name),n.transformFlags&=-8388609,n}function Gt(e){var n=I(269);return n.name=e,n.transformFlags|=4|l(n.name),n.transformFlags&=-8388609,n}function Bt(e){var n=I(264);return n.elements=C(e),n.transformFlags|=u(n.elements),n.transformFlags&=-8388609,n}function Ut(e,n){var t=I(265);return t.propertyName=e,t.name=n,t.transformFlags|=l(t.propertyName)|l(t.name),t.transformFlags&=-8388609,t}function Vt(e,n,t,r){var a=N(266,e,n);return a.isExportEquals=t,a.expression=t?f().parenthesizeRightSideOfBinary(62,void 0,r):f().parenthesizeExpressionOfExportDefault(r),a.transformFlags|=l(a.expression),a.transformFlags&=-8388609,a}function jt(e,n,t,r){return e.decorators!==n||e.modifiers!==t||e.expression!==r?m(Vt(n,t,e.isExportEquals,r),e):e}function Ht(e,n,t,r,a){var i=N(267,e,n);return i.isTypeOnly=t,i.exportClause=r,i.moduleSpecifier=a,i.transformFlags|=l(i.exportClause)|l(i.moduleSpecifier),i.transformFlags&=-8388609,i}function Kt(e,n,t,r,a,i){return e.decorators!==n||e.modifiers!==t||e.isTypeOnly!==r||e.exportClause!==a||e.moduleSpecifier!==i?m(Ht(n,t,r,a,i),e):e}function Wt(e){var n=I(268);return n.elements=C(e),n.transformFlags|=u(n.elements),n.transformFlags&=-8388609,n}function zt(e,n){var t=I(270);return t.propertyName=$r(e),t.name=$r(n),t.transformFlags|=l(t.propertyName)|l(t.name),t.transformFlags&=-8388609,t}function qt(e){var n=I(272);return n.expression=e,n.transformFlags|=l(n.expression),n.transformFlags&=-8388609,n}function Jt(e,n){var t=I(e);return t.type=n,t}function Xt(e,n){return O(308,void 0,void 0,void 0,void 0,e,n)}function Yt(e,n){void 0===n&&(n=!1);var t=I(312);return t.jsDocPropertyTags=Zr(e),t.isArrayType=n,t}function Qt(e){var n=I(301);return n.type=e,n}function Zt(e,n,t){var r=I(313);return r.typeParameters=Zr(e),r.parameters=C(n),r.type=t,r}function $t(n){var t=o(n.kind);return n.tagName.escapedText===e.escapeLeadingUnderscores(t)?n.tagName:X(t)}function er(e,n,t){var r=I(e);return r.tagName=n,r.comment=t,r}function nr(e,n,t,r){var a=er(330,null!=e?e:X("template"),r);return a.constraint=n,a.typeParameters=C(t),a}function tr(n,t,r,a){var i=er(331,null!=n?n:X("typedef"),a);return i.typeExpression=t,i.fullName=r,i.name=e.getJSDocTypeAliasName(r),i}function rr(e,n,t,r,a,i){var o=er(326,null!=e?e:X("param"),i);return o.typeExpression=r,o.name=n,o.isNameFirst=!!a,o.isBracketed=t,o}function ar(e,n,t,r,a,i){var o=er(333,null!=e?e:X("prop"),i);return o.typeExpression=r,o.name=n,o.isNameFirst=!!a,o.isBracketed=t,o}function ir(n,t,r,a){var i=er(324,null!=n?n:X("callback"),a);return i.typeExpression=t,i.fullName=r,i.name=e.getJSDocTypeAliasName(r),i}function or(e,n,t){var r=er(315,null!=e?e:X("augments"),t);return r.class=n,r}function sr(e,n,t){var r=er(316,null!=e?e:X("implements"),t);return r.class=n,r}function cr(e,n,t){var r=er(332,null!=e?e:X("see"),t);return r.name=n,r}function lr(e){var n=I(302);return n.name=e,n}function ur(e,n,t){return er(e,null!=n?n:X(o(e)),t)}function dr(e,n,t,r){var a=er(e,null!=n?n:X(o(e)),r);return a.typeExpression=t,a}function pr(e,n){return er(314,e,n)}function mr(e,n){var t=I(311);return t.comment=e,t.tags=Zr(n),t}function fr(e,n,t){var r=I(273);return r.openingElement=e,r.children=C(n),r.closingElement=t,r.transformFlags|=l(r.openingElement)|u(r.children)|l(r.closingElement)|2,r}function _r(e,n,t){var r=I(274);return r.tagName=e,r.typeArguments=Zr(n),r.attributes=t,r.transformFlags|=l(r.tagName)|u(r.typeArguments)|l(r.attributes)|2,r.typeArguments&&(r.transformFlags|=1),r}function gr(e,n,t){var r=I(275);return r.tagName=e,r.typeArguments=Zr(n),r.attributes=t,r.transformFlags|=l(r.tagName)|u(r.typeArguments)|l(r.attributes)|2,n&&(r.transformFlags|=1),r}function yr(e){var n=I(276);return n.tagName=e,n.transformFlags|=2|l(n.tagName),n}function vr(e,n,t){var r=I(277);return r.openingFragment=e,r.children=C(n),r.closingFragment=t,r.transformFlags|=l(r.openingFragment)|u(r.children)|l(r.closingFragment)|2,r}function hr(e,n){var t=I(11);return t.text=e,t.containsOnlyTriviaWhiteSpaces=!!n,t.transformFlags|=2,t}function br(e,n){var t=I(280);return t.name=e,t.initializer=n,t.transformFlags|=l(t.name)|l(t.initializer)|2,t}function Er(e){var n=I(281);return n.properties=C(e),n.transformFlags|=2|u(n.properties),n}function Tr(e){var n=I(282);return n.expression=e,n.transformFlags|=2|l(n.expression),n}function Sr(e,n){var t=I(283);return t.dotDotDotToken=e,t.expression=n,t.transformFlags|=l(t.dotDotDotToken)|l(t.expression)|2,t}function xr(e,n){var t=I(284);return t.expression=f().parenthesizeExpressionForDisallowedComma(e),t.statements=C(n),t.transformFlags|=l(t.expression)|u(t.statements),t}function Lr(e){var n=I(285);return n.statements=C(e),n.transformFlags=u(n.statements),n}function Ar(n,t){var r=I(286);switch(r.token=n,r.types=C(t),r.transformFlags|=u(r.types),n){case 93:r.transformFlags|=256;break;case 116:r.transformFlags|=1;break;default:return e.Debug.assertNever(n)}return r}function Dr(n,t){var r=I(287);return n=e.isString(n)?_t(n,void 0,void 0,void 0):n,r.variableDeclaration=n,r.block=t,r.transformFlags|=l(r.variableDeclaration)|l(r.block),n||(r.transformFlags|=16),r}function Cr(e,n){var t=R(288,void 0,void 0,e);return t.initializer=f().parenthesizeExpressionForDisallowedComma(n),t.transformFlags|=l(t.name)|l(t.initializer),t}function Ir(e,n){var t=R(289,void 0,void 0,e);return t.objectAssignmentInitializer=n&&f().parenthesizeExpressionForDisallowedComma(n),t.transformFlags|=256|l(t.objectAssignmentInitializer),t}function Nr(e){var n=I(290);return n.expression=f().parenthesizeExpressionForDisallowedComma(e),n.transformFlags|=16416|l(n.expression),n}function Rr(e,n){var t=I(291);return t.name=$r(e),t.initializer=n&&f().parenthesizeExpressionForDisallowedComma(n),t.transformFlags|=l(t.name)|l(t.initializer)|1,t}function kr(n,t){void 0===t&&(t=e.emptyArray);var r=I(298);return r.prepends=t,r.sourceFiles=n,r}function Or(e,n){var t=I(e);return t.data=n,t}function Pr(n,t){var r=I(336);return r.expression=n,r.original=t,r.transformFlags|=1|l(r.expression),e.setTextRange(r,t),r}function Mr(e,n){return e.expression!==n?m(Pr(n,e.original),e):e}function wr(n){if(e.nodeIsSynthesized(n)&&!e.isParseTreeNode(n)&&!n.original&&!n.emitNode&&!n.id){if(e.isCommaListExpression(n))return n.elements;if(e.isBinaryExpression(n)&&e.isCommaToken(n.operatorToken))return[n.left,n.right]}return n}function Fr(n){var t=I(337);return t.elements=C(e.sameFlatMap(n,wr)),t.transformFlags|=u(t.elements),t}function Gr(e,n){var t=I(340);return t.expression=e,t.thisArg=n,t.transformFlags|=l(t.expression)|l(t.thisArg),t}function Br(n){if(void 0===n)return n;var t=e.isSourceFile(n)?p.createBaseSourceFileNode(297):e.isIdentifier(n)?p.createBaseIdentifierNode(78):e.isPrivateIdentifier(n)?p.createBasePrivateIdentifierNode(79):e.isNodeKind(n.kind)?p.createBaseNode(n.kind):p.createBaseTokenNode(n.kind);for(var r in t.flags|=-9&n.flags,t.transformFlags=n.transformFlags,y(t,n),n)!t.hasOwnProperty(r)&&n.hasOwnProperty(r)&&(t[r]=n[r]);return t}function Ur(){return An(j("0"))}function Vr(e,n,t){return dn(an(e,n),void 0,t)}function jr(e,n,t){return Vr(X(e),n,t)}function Hr(e,n,t){return!!t&&(e.push(Cr(n,t)),!0)}function Kr(n,t){var r=e.skipParentheses(n);switch(r.kind){case 78:return t;case 107:case 8:case 9:case 10:return!1;case 199:return 0!==r.elements.length;case 200:return r.properties.length>0;default:return!0}}function Wr(n,t,r,a){void 0===a&&(a=0);var i=e.getNameOfDeclaration(n);if(i&&e.isIdentifier(i)&&!e.isGeneratedIdentifier(i)){var o=e.setParent(e.setTextRange(Br(i),i),i.parent);return a|=e.getEmitFlags(i),r||(a|=48),t||(a|=1536),a&&e.setEmitFlags(o,a),o}return Q(n)}function zr(e,n,t){return Wr(e,n,t,8192)}function qr(n,t,r,a){var i=an(n,e.nodeIsSynthesized(t)?t:Br(t));e.setTextRange(i,t);var o=0;return a||(o|=48),r||(o|=1536),o&&e.setEmitFlags(i,o),i}function Jr(){return e.startOnNewLine($n(W("use strict")))}function Xr(n,t,r){e.Debug.assert(0===t.length,"Prologue directives should be at the first statement in the target statements array");for(var a,i=!1,o=0,s=n.length;o<s;){var c=n[o];if(!e.isPrologueDirective(c))break;a=c,e.isStringLiteral(a.expression)&&"use strict"===a.expression.text&&(i=!0),t.push(c),o++}return r&&!i&&t.push(Jr()),o}function Yr(n,t,r,a,i){void 0===i&&(i=e.returnTrue);for(var o=n.length;void 0!==r&&r<o;){var s=n[r];if(!(1048576&e.getEmitFlags(s)&&i(s)))break;e.append(t,a?e.visitNode(s,a,e.isStatement):s),r++}return r}function Qr(e,n,t){for(var r=t;r<e.length&&n(e[r]);)r++;return r}function Zr(e){return e?C(e):void 0}function $r(e){return"string"==typeof e?X(e):e}function ea(e){return"string"==typeof e?W(e):"number"==typeof e?j(e):"boolean"==typeof e?e?ne():te():e}function na(n){return n&&e.isNotEmittedStatement(n)?e.setTextRange(y(Zn(),n),n):n}}function a(n,t){return n!==t&&e.setTextRange(n,t),n}function i(n,t){return n!==t&&(y(n,t),e.setTextRange(n,t)),n}function o(n){switch(n){case 329:return"type";case 327:return"returns";case 328:return"this";case 325:return"enum";case 317:return"author";case 319:return"class";case 320:return"public";case 321:return"private";case 322:return"protected";case 323:return"readonly";case 330:return"template";case 331:return"typedef";case 326:return"param";case 333:return"prop";case 324:return"callback";case 315:return"augments";case 316:return"implements";default:return e.Debug.fail("Unsupported kind: "+e.Debug.formatSyntaxKind(n))}}!function(e){e[e.None=0]="None",e[e.NoParenthesizerRules=1]="NoParenthesizerRules",e[e.NoNodeConverters=2]="NoNodeConverters",e[e.NoIndentationOnFreshPropertyAccess=4]="NoIndentationOnFreshPropertyAccess",e[e.NoOriginalNode=8]="NoOriginalNode"}(e.NodeFactoryFlags||(e.NodeFactoryFlags={})),e.createNodeFactory=r;var s={};function c(e){return-8388609&l(e)}function l(n){if(!n)return 0;var t,r=n.transformFlags&~p(n.kind);return e.isNamedDeclaration(n)&&e.isPropertyName(n.name)?(t=n.name,r|4096&t.transformFlags):r}function u(e){return e?e.transformFlags:0}function d(e){for(var n=0,t=0,r=e;t<r.length;t++){n|=l(r[t])}e.transformFlags=n}function p(e){if(e>=172&&e<=195)return-2;switch(e){case 203:case 204:case 199:return 536879104;case 256:return 546379776;case 160:return 536870912;case 209:return 547309568;case 208:case 251:return 547313664;case 250:return 537018368;case 252:case 221:return 536905728;case 166:return 547311616;case 163:return 536875008;case 165:case 167:case 168:return 538923008;case 128:case 144:case 155:case 141:case 147:case 145:case 131:case 148:case 113:case 159:case 162:case 164:case 169:case 170:case 171:case 253:case 254:return-2;case 200:return 536922112;case 287:return 536887296;case 196:case 197:return 536879104;case 206:case 224:case 336:case 207:case 105:return 536870912;case 201:case 202:default:return 536870912}}e.getTransformFlagsSubtreeExclusions=p;var m=e.createBaseNodeFactory();function f(e){return e.flags|=8,e}var _,g={createBaseSourceFileNode:function(e){return f(m.createBaseSourceFileNode(e))},createBaseIdentifierNode:function(e){return f(m.createBaseIdentifierNode(e))},createBasePrivateIdentifierNode:function(e){return f(m.createBasePrivateIdentifierNode(e))},createBaseTokenNode:function(e){return f(m.createBaseTokenNode(e))},createBaseNode:function(e){return f(m.createBaseNode(e))}};function y(n,t){if(n.original=t,t){var r=t.emitNode;r&&(n.emitNode=function(n,t){var r=n.flags,a=n.leadingComments,i=n.trailingComments,o=n.commentRange,s=n.sourceMapRange,c=n.tokenSourceMapRanges,l=n.constantValue,u=n.helpers,d=n.startsOnNewLine;t||(t={});a&&(t.leadingComments=e.addRange(a.slice(),t.leadingComments));i&&(t.trailingComments=e.addRange(i.slice(),t.trailingComments));r&&(t.flags=r);o&&(t.commentRange=o);s&&(t.sourceMapRange=s);c&&(t.tokenSourceMapRanges=function(e,n){n||(n=[]);for(var t in e)n[t]=e[t];return n}(c,t.tokenSourceMapRanges));void 0!==l&&(t.constantValue=l);if(u)for(var p=0,m=u;p<m.length;p++){var f=m[p];t.helpers=e.appendIfUnique(t.helpers,f)}void 0!==d&&(t.startsOnNewLine=d);return t}(r,n.emitNode))}return n}e.factory=r(4,g),e.createUnparsedSourceFile=function(n,t,r){var a,i,o,s,c,l,u,d,p,m;e.isString(n)?(o="",s=n,c=n.length,l=t,u=r):(e.Debug.assert("js"===t||"dts"===t),o=("js"===t?n.javascriptPath:n.declarationPath)||"",l="js"===t?n.javascriptMapPath:n.declarationMapPath,d=function(){return"js"===t?n.javascriptText:n.declarationText},p=function(){return"js"===t?n.javascriptMapText:n.declarationMapText},c=function(){return d().length},n.buildInfo&&n.buildInfo.bundle&&(e.Debug.assert(void 0===r||"boolean"==typeof r),a=r,i="js"===t?n.buildInfo.bundle.js:n.buildInfo.bundle.dts,m=n.oldFileOfCurrentEmit));var f=m?function(n){for(var t,r,a=0,i=n.sections;a<i.length;a++){var o=i[a];switch(o.kind){case"internal":case"text":t=e.append(t,e.setTextRange(e.factory.createUnparsedTextLike(o.data,"internal"===o.kind),o));break;case"no-default-lib":case"reference":case"type":case"lib":r=e.append(r,e.setTextRange(e.factory.createUnparsedSyntheticReference(o),o));break;case"prologue":case"emitHelpers":case"prepend":break;default:e.Debug.assertNever(o)}}var s=e.factory.createUnparsedSource(e.emptyArray,r,null!=t?t:e.emptyArray);return e.setEachParent(r,s),e.setEachParent(t,s),s.helpers=e.map(n.sources&&n.sources.helpers,(function(n){return e.getAllUnscopedEmitHelpers().get(n)})),s}(e.Debug.assertDefined(i)):function(n,t,r){for(var a,i,o,s,c,l,u,d,p=0,m=n?n.sections:e.emptyArray;p<m.length;p++){var f=m[p];switch(f.kind){case"prologue":a=e.append(a,e.setTextRange(e.factory.createUnparsedPrologue(f.data),f));break;case"emitHelpers":i=e.append(i,e.getAllUnscopedEmitHelpers().get(f.data));break;case"no-default-lib":d=!0;break;case"reference":o=e.append(o,{pos:-1,end:-1,fileName:f.data});break;case"type":s=e.append(s,f.data);break;case"lib":c=e.append(c,{pos:-1,end:-1,fileName:f.data});break;case"prepend":for(var _=void 0,g=0,y=f.texts;g<y.length;g++){var v=y[g];t&&"internal"===v.kind||(_=e.append(_,e.setTextRange(e.factory.createUnparsedTextLike(v.data,"internal"===v.kind),v)))}l=e.addRange(l,_),u=e.append(u,e.factory.createUnparsedPrepend(f.data,null!=_?_:e.emptyArray));break;case"internal":if(t){u||(u=[]);break}case"text":u=e.append(u,e.setTextRange(e.factory.createUnparsedTextLike(f.data,"internal"===f.kind),f));break;default:e.Debug.assertNever(f)}}if(!u){var h=e.factory.createUnparsedTextLike(void 0,!1);e.setTextRangePosWidth(h,0,"function"==typeof r?r():r),u=[h]}var b=e.parseNodeFactory.createUnparsedSource(null!=a?a:e.emptyArray,void 0,u);return e.setEachParent(a,b),e.setEachParent(u,b),e.setEachParent(l,b),b.hasNoDefaultLib=d,b.helpers=i,b.referencedFiles=o||e.emptyArray,b.typeReferenceDirectives=s,b.libReferenceDirectives=c||e.emptyArray,b}(i,a,c);return f.fileName=o,f.sourceMapPath=l,f.oldFileOfCurrentEmit=m,d&&p?(Object.defineProperty(f,"text",{get:d}),Object.defineProperty(f,"sourceMapText",{get:p})):(e.Debug.assert(!m),f.text=null!=s?s:"",f.sourceMapText=u),f},e.createInputFiles=function(n,t,r,a,i,o,s,c,l,u,d){var p=e.parseNodeFactory.createInputFiles();if(e.isString(n))p.javascriptText=n,p.javascriptMapPath=r,p.javascriptMapText=a,p.declarationText=t,p.declarationMapPath=i,p.declarationMapText=o,p.javascriptPath=s,p.declarationPath=c,p.buildInfoPath=l,p.buildInfo=u,p.oldFileOfCurrentEmit=d;else{var m,f=new e.Map,_=function(e){if(void 0!==e){var t=f.get(e);return void 0===t&&(t=n(e),f.set(e,void 0!==t&&t)),!1!==t?t:void 0}},g=function(e){var n=_(e);return void 0!==n?n:"/* Input file "+e+" was missing */\r\n"};p.javascriptPath=t,p.javascriptMapPath=r,p.declarationPath=e.Debug.assertDefined(a),p.declarationMapPath=i,p.buildInfoPath=o,Object.defineProperties(p,{javascriptText:{get:function(){return g(t)}},javascriptMapText:{get:function(){return _(r)}},declarationText:{get:function(){return g(e.Debug.assertDefined(a))}},declarationMapText:{get:function(){return _(i)}},buildInfo:{get:function(){return function(n){if(void 0===m){var t=n();m=void 0!==t&&e.getBuildInfo(t)}return m||void 0}((function(){return _(o)}))}}})}return p},e.createSourceMapSource=function(n,t,r){return new(_||(_=e.objectAllocator.getSourceMapSourceConstructor()))(n,t,r)},e.setOriginalNode=y}(ts||(ts={})),function(e){function n(t){var r;if(!t.emitNode){if(e.isParseTreeNode(t)){if(297===t.kind)return t.emitNode={annotatedNodes:[t]};n(null!==(r=e.getSourceFileOfNode(e.getParseTreeNode(e.getSourceFileOfNode(t))))&&void 0!==r?r:e.Debug.fail("Could not determine parsed source file.")).annotatedNodes.push(t)}t.emitNode={}}return t.emitNode}function t(e){var n;return null===(n=e.emitNode)||void 0===n?void 0:n.leadingComments}function r(e,t){return n(e).leadingComments=t,e}function a(e){var n;return null===(n=e.emitNode)||void 0===n?void 0:n.trailingComments}function i(e,t){return n(e).trailingComments=t,e}e.getOrCreateEmitNode=n,e.disposeEmitNodes=function(n){var t,r,a=null===(r=null===(t=e.getSourceFileOfNode(e.getParseTreeNode(n)))||void 0===t?void 0:t.emitNode)||void 0===r?void 0:r.annotatedNodes;if(a)for(var i=0,o=a;i<o.length;i++){o[i].emitNode=void 0}},e.removeAllComments=function(e){var t=n(e);return t.flags|=1536,t.leadingComments=void 0,t.trailingComments=void 0,e},e.setEmitFlags=function(e,t){return n(e).flags=t,e},e.addEmitFlags=function(e,t){var r=n(e);return r.flags=r.flags|t,e},e.getSourceMapRange=function(e){var n,t;return null!==(t=null===(n=e.emitNode)||void 0===n?void 0:n.sourceMapRange)&&void 0!==t?t:e},e.setSourceMapRange=function(e,t){return n(e).sourceMapRange=t,e},e.getTokenSourceMapRange=function(e,n){var t,r;return null===(r=null===(t=e.emitNode)||void 0===t?void 0:t.tokenSourceMapRanges)||void 0===r?void 0:r[n]},e.setTokenSourceMapRange=function(e,t,r){var a,i=n(e);return(null!==(a=i.tokenSourceMapRanges)&&void 0!==a?a:i.tokenSourceMapRanges=[])[t]=r,e},e.getStartsOnNewLine=function(e){var n;return null===(n=e.emitNode)||void 0===n?void 0:n.startsOnNewLine},e.setStartsOnNewLine=function(e,t){return n(e).startsOnNewLine=t,e},e.getCommentRange=function(e){var n,t;return null!==(t=null===(n=e.emitNode)||void 0===n?void 0:n.commentRange)&&void 0!==t?t:e},e.setCommentRange=function(e,t){return n(e).commentRange=t,e},e.getSyntheticLeadingComments=t,e.setSyntheticLeadingComments=r,e.addSyntheticLeadingComment=function(n,a,i,o){return r(n,e.append(t(n),{kind:a,pos:-1,end:-1,hasTrailingNewLine:o,text:i}))},e.getSyntheticTrailingComments=a,e.setSyntheticTrailingComments=i,e.addSyntheticTrailingComment=function(n,t,r,o){return i(n,e.append(a(n),{kind:t,pos:-1,end:-1,hasTrailingNewLine:o,text:r}))},e.moveSyntheticComments=function(e,o){r(e,t(o)),i(e,a(o));var s=n(o);return s.leadingComments=void 0,s.trailingComments=void 0,e},e.getConstantValue=function(e){var n;return null===(n=e.emitNode)||void 0===n?void 0:n.constantValue},e.setConstantValue=function(e,t){return n(e).constantValue=t,e},e.addEmitHelper=function(t,r){var a=n(t);return a.helpers=e.append(a.helpers,r),t},e.addEmitHelpers=function(t,r){if(e.some(r))for(var a=n(t),i=0,o=r;i<o.length;i++){var s=o[i];a.helpers=e.appendIfUnique(a.helpers,s)}return t},e.removeEmitHelper=function(n,t){var r,a=null===(r=n.emitNode)||void 0===r?void 0:r.helpers;return!!a&&e.orderedRemoveItem(a,t)},e.getEmitHelpers=function(e){var n;return null===(n=e.emitNode)||void 0===n?void 0:n.helpers},e.moveEmitHelpers=function(t,r,a){var i=t.emitNode,o=i&&i.helpers;if(e.some(o)){for(var s=n(r),c=0,l=0;l<o.length;l++){var u=o[l];a(u)?(c++,s.helpers=e.appendIfUnique(s.helpers,u)):c>0&&(o[l-c]=u)}c>0&&(o.length-=c)}},e.ignoreSourceNewlines=function(e){return n(e).flags|=134217728,e}}(ts||(ts={})),function(e){function n(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return function(t){for(var r="",a=0;a<n.length;a++)r+=e[a],r+=t(n[a]);return r+=e[e.length-1]}}var t;e.createEmitHelperFactory=function(n){var t=n.factory;return{getUnscopedHelperName:r,createDecorateHelper:function(a,i,o,s){n.requestEmitHelper(e.decorateHelper);var c=[];c.push(t.createArrayLiteralExpression(a,!0)),c.push(i),o&&(c.push(o),s&&c.push(s));return t.createCallExpression(r("__decorate"),void 0,c)},createMetadataHelper:function(a,i){return n.requestEmitHelper(e.metadataHelper),t.createCallExpression(r("__metadata"),void 0,[t.createStringLiteral(a),i])},createParamHelper:function(a,i,o){return n.requestEmitHelper(e.paramHelper),e.setTextRange(t.createCallExpression(r("__param"),void 0,[t.createNumericLiteral(i+""),a]),o)},createAssignHelper:function(a){if(n.getCompilerOptions().target>=2)return t.createCallExpression(t.createPropertyAccessExpression(t.createIdentifier("Object"),"assign"),void 0,a);return n.requestEmitHelper(e.assignHelper),t.createCallExpression(r("__assign"),void 0,a)},createAwaitHelper:function(a){return n.requestEmitHelper(e.awaitHelper),t.createCallExpression(r("__await"),void 0,[a])},createAsyncGeneratorHelper:function(a,i){return n.requestEmitHelper(e.awaitHelper),n.requestEmitHelper(e.asyncGeneratorHelper),(a.emitNode||(a.emitNode={})).flags|=786432,t.createCallExpression(r("__asyncGenerator"),void 0,[i?t.createThis():t.createVoidZero(),t.createIdentifier("arguments"),a])},createAsyncDelegatorHelper:function(a){return n.requestEmitHelper(e.awaitHelper),n.requestEmitHelper(e.asyncDelegator),t.createCallExpression(r("__asyncDelegator"),void 0,[a])},createAsyncValuesHelper:function(a){return n.requestEmitHelper(e.asyncValues),t.createCallExpression(r("__asyncValues"),void 0,[a])},createRestHelper:function(a,i,o,s){n.requestEmitHelper(e.restHelper);for(var c=[],l=0,u=0;u<i.length-1;u++){var d=e.getPropertyNameOfBindingOrAssignmentElement(i[u]);if(d)if(e.isComputedPropertyName(d)){e.Debug.assertIsDefined(o,"Encountered computed property name but 'computedTempVariables' argument was not provided.");var p=o[l];l++,c.push(t.createConditionalExpression(t.createTypeCheck(p,"symbol"),void 0,p,void 0,t.createAdd(p,t.createStringLiteral(""))))}else c.push(t.createStringLiteralFromNode(d))}return t.createCallExpression(r("__rest"),void 0,[a,e.setTextRange(t.createArrayLiteralExpression(c),s)])},createAwaiterHelper:function(a,i,o,s){n.requestEmitHelper(e.awaiterHelper);var c=t.createFunctionExpression(void 0,t.createToken(41),void 0,void 0,[],void 0,s);return(c.emitNode||(c.emitNode={})).flags|=786432,t.createCallExpression(r("__awaiter"),void 0,[a?t.createThis():t.createVoidZero(),i?t.createIdentifier("arguments"):t.createVoidZero(),o?e.createExpressionFromEntityName(t,o):t.createVoidZero(),c])},createExtendsHelper:function(a){return n.requestEmitHelper(e.extendsHelper),t.createCallExpression(r("__extends"),void 0,[a,t.createUniqueName("_super",48)])},createTemplateObjectHelper:function(a,i){return n.requestEmitHelper(e.templateObjectHelper),t.createCallExpression(r("__makeTemplateObject"),void 0,[a,i])},createSpreadHelper:function(a){return n.requestEmitHelper(e.readHelper),n.requestEmitHelper(e.spreadHelper),t.createCallExpression(r("__spread"),void 0,a)},createSpreadArraysHelper:function(a){return n.requestEmitHelper(e.spreadArraysHelper),t.createCallExpression(r("__spreadArrays"),void 0,a)},createValuesHelper:function(a){return n.requestEmitHelper(e.valuesHelper),t.createCallExpression(r("__values"),void 0,[a])},createReadHelper:function(a,i){return n.requestEmitHelper(e.readHelper),t.createCallExpression(r("__read"),void 0,void 0!==i?[a,t.createNumericLiteral(i+"")]:[a])},createGeneratorHelper:function(a){return n.requestEmitHelper(e.generatorHelper),t.createCallExpression(r("__generator"),void 0,[t.createThis(),a])},createCreateBindingHelper:function(a,i,o){return n.requestEmitHelper(e.createBindingHelper),t.createCallExpression(r("__createBinding"),void 0,__spreadArrays([t.createIdentifier("exports"),a,i],o?[o]:[]))},createImportStarHelper:function(a){return n.requestEmitHelper(e.importStarHelper),t.createCallExpression(r("__importStar"),void 0,[a])},createImportStarCallbackHelper:function(){return n.requestEmitHelper(e.importStarHelper),r("__importStar")},createImportDefaultHelper:function(a){return n.requestEmitHelper(e.importDefaultHelper),t.createCallExpression(r("__importDefault"),void 0,[a])},createExportStarHelper:function(a,i){void 0===i&&(i=t.createIdentifier("exports"));return n.requestEmitHelper(e.exportStarHelper),n.requestEmitHelper(e.createBindingHelper),t.createCallExpression(r("__exportStar"),void 0,[a,i])},createClassPrivateFieldGetHelper:function(a,i){return n.requestEmitHelper(e.classPrivateFieldGetHelper),t.createCallExpression(r("__classPrivateFieldGet"),void 0,[a,i])},createClassPrivateFieldSetHelper:function(a,i,o){return n.requestEmitHelper(e.classPrivateFieldSetHelper),t.createCallExpression(r("__classPrivateFieldSet"),void 0,[a,i,o])}};function r(n){return e.setEmitFlags(t.createIdentifier(n),4098)}},e.compareEmitHelpers=function(n,t){return n===t||n.priority===t.priority?0:void 0===n.priority?1:void 0===t.priority?-1:e.compareValues(n.priority,t.priority)},e.helperString=n,e.decorateHelper={name:"typescript:decorate",importName:"__decorate",scoped:!1,priority:2,text:'\n            var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {\n                var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;\n                if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);\n                else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;\n                return c > 3 && r && Object.defineProperty(target, key, r), r;\n            };'},e.metadataHelper={name:"typescript:metadata",importName:"__metadata",scoped:!1,priority:3,text:'\n            var __metadata = (this && this.__metadata) || function (k, v) {\n                if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);\n            };'},e.paramHelper={name:"typescript:param",importName:"__param",scoped:!1,priority:4,text:"\n            var __param = (this && this.__param) || function (paramIndex, decorator) {\n                return function (target, key) { decorator(target, key, paramIndex); }\n            };"},e.assignHelper={name:"typescript:assign",importName:"__assign",scoped:!1,priority:1,text:"\n            var __assign = (this && this.__assign) || function () {\n                __assign = Object.assign || function(t) {\n                    for (var s, i = 1, n = arguments.length; i < n; i++) {\n                        s = arguments[i];\n                        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))\n                            t[p] = s[p];\n                    }\n                    return t;\n                };\n                return __assign.apply(this, arguments);\n            };"},e.awaitHelper={name:"typescript:await",importName:"__await",scoped:!1,text:"\n            var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }"},e.asyncGeneratorHelper={name:"typescript:asyncGenerator",importName:"__asyncGenerator",scoped:!1,dependencies:[e.awaitHelper],text:'\n            var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {\n                if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");\n                var g = generator.apply(thisArg, _arguments || []), i, q = [];\n                return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;\n                function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }\n                function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }\n                function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }\n                function fulfill(value) { resume("next", value); }\n                function reject(value) { resume("throw", value); }\n                function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }\n            };'},e.asyncDelegator={name:"typescript:asyncDelegator",importName:"__asyncDelegator",scoped:!1,dependencies:[e.awaitHelper],text:'\n            var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {\n                var i, p;\n                return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;\n                function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }\n            };'},e.asyncValues={name:"typescript:asyncValues",importName:"__asyncValues",scoped:!1,text:'\n            var __asyncValues = (this && this.__asyncValues) || function (o) {\n                if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");\n                var m = o[Symbol.asyncIterator], i;\n                return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);\n                function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }\n                function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }\n            };'},e.restHelper={name:"typescript:rest",importName:"__rest",scoped:!1,text:'\n            var __rest = (this && this.__rest) || function (s, e) {\n                var t = {};\n                for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)\n                    t[p] = s[p];\n                if (s != null && typeof Object.getOwnPropertySymbols === "function")\n                    for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {\n                        if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))\n                            t[p[i]] = s[p[i]];\n                    }\n                return t;\n            };'},e.awaiterHelper={name:"typescript:awaiter",importName:"__awaiter",scoped:!1,priority:5,text:'\n            var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {\n                function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }\n                return new (P || (P = Promise))(function (resolve, reject) {\n                    function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }\n                    function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }\n                    function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }\n                    step((generator = generator.apply(thisArg, _arguments || [])).next());\n                });\n            };'},e.extendsHelper={name:"typescript:extends",importName:"__extends",scoped:!1,priority:0,text:"\n            var __extends = (this && this.__extends) || (function () {\n                var extendStatics = function (d, b) {\n                    extendStatics = Object.setPrototypeOf ||\n                        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||\n                        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };\n                    return extendStatics(d, b);\n                };\n\n                return function (d, b) {\n                    extendStatics(d, b);\n                    function __() { this.constructor = d; }\n                    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());\n                };\n            })();"},e.templateObjectHelper={name:"typescript:makeTemplateObject",importName:"__makeTemplateObject",scoped:!1,priority:0,text:'\n            var __makeTemplateObject = (this && this.__makeTemplateObject) || function (cooked, raw) {\n                if (Object.defineProperty) { Object.defineProperty(cooked, "raw", { value: raw }); } else { cooked.raw = raw; }\n                return cooked;\n            };'},e.readHelper={name:"typescript:read",importName:"__read",scoped:!1,text:'\n            var __read = (this && this.__read) || function (o, n) {\n                var m = typeof Symbol === "function" && o[Symbol.iterator];\n                if (!m) return o;\n                var i = m.call(o), r, ar = [], e;\n                try {\n                    while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);\n                }\n                catch (error) { e = { error: error }; }\n                finally {\n                    try {\n                        if (r && !r.done && (m = i["return"])) m.call(i);\n                    }\n                    finally { if (e) throw e.error; }\n                }\n                return ar;\n            };'},e.spreadHelper={name:"typescript:spread",importName:"__spread",scoped:!1,dependencies:[e.readHelper],text:"\n            var __spread = (this && this.__spread) || function () {\n                for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));\n                return ar;\n            };"},e.spreadArraysHelper={name:"typescript:spreadArrays",importName:"__spreadArrays",scoped:!1,text:"\n            var __spreadArrays = (this && this.__spreadArrays) || function () {\n                for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;\n                for (var r = Array(s), k = 0, i = 0; i < il; i++)\n                    for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)\n                        r[k] = a[j];\n                return r;\n            };"},e.valuesHelper={name:"typescript:values",importName:"__values",scoped:!1,text:'\n            var __values = (this && this.__values) || function(o) {\n                var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;\n                if (m) return m.call(o);\n                if (o && typeof o.length === "number") return {\n                    next: function () {\n                        if (o && i >= o.length) o = void 0;\n                        return { value: o && o[i++], done: !o };\n                    }\n                };\n                throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");\n            };'},e.generatorHelper={name:"typescript:generator",importName:"__generator",scoped:!1,priority:6,text:'\n            var __generator = (this && this.__generator) || function (thisArg, body) {\n                var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;\n                return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;\n                function verb(n) { return function (v) { return step([n, v]); }; }\n                function step(op) {\n                    if (f) throw new TypeError("Generator is already executing.");\n                    while (_) try {\n                        if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;\n                        if (y = 0, t) op = [op[0] & 2, t.value];\n                        switch (op[0]) {\n                            case 0: case 1: t = op; break;\n                            case 4: _.label++; return { value: op[1], done: false };\n                            case 5: _.label++; y = op[1]; op = [0]; continue;\n                            case 7: op = _.ops.pop(); _.trys.pop(); continue;\n                            default:\n                                if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }\n                                if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }\n                                if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }\n                                if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }\n                                if (t[2]) _.ops.pop();\n                                _.trys.pop(); continue;\n                        }\n                        op = body.call(thisArg, _);\n                    } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }\n                    if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };\n                }\n            };'},e.createBindingHelper={name:"typescript:commonjscreatebinding",importName:"__createBinding",scoped:!1,priority:1,text:"\n            var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {\n                if (k2 === undefined) k2 = k;\n                Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });\n            }) : (function(o, m, k, k2) {\n                if (k2 === undefined) k2 = k;\n                o[k2] = m[k];\n            }));"},e.setModuleDefaultHelper={name:"typescript:commonjscreatevalue",importName:"__setModuleDefault",scoped:!1,priority:1,text:'\n            var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {\n                Object.defineProperty(o, "default", { enumerable: true, value: v });\n            }) : function(o, v) {\n                o["default"] = v;\n            });'},e.importStarHelper={name:"typescript:commonjsimportstar",importName:"__importStar",scoped:!1,dependencies:[e.createBindingHelper,e.setModuleDefaultHelper],priority:2,text:'\n            var __importStar = (this && this.__importStar) || function (mod) {\n                if (mod && mod.__esModule) return mod;\n                var result = {};\n                if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);\n                __setModuleDefault(result, mod);\n                return result;\n            };'},e.importDefaultHelper={name:"typescript:commonjsimportdefault",importName:"__importDefault",scoped:!1,text:'\n            var __importDefault = (this && this.__importDefault) || function (mod) {\n                return (mod && mod.__esModule) ? mod : { "default": mod };\n            };'},e.exportStarHelper={name:"typescript:export-star",importName:"__exportStar",scoped:!1,dependencies:[e.createBindingHelper],priority:2,text:'\n            var __exportStar = (this && this.__exportStar) || function(m, exports) {\n                for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);\n            };'},e.classPrivateFieldGetHelper={name:"typescript:classPrivateFieldGet",importName:"__classPrivateFieldGet",scoped:!1,text:'\n            var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {\n                if (!privateMap.has(receiver)) {\n                    throw new TypeError("attempted to get private field on non-instance");\n                }\n                return privateMap.get(receiver);\n            };'},e.classPrivateFieldSetHelper={name:"typescript:classPrivateFieldSet",importName:"__classPrivateFieldSet",scoped:!1,text:'\n            var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {\n                if (!privateMap.has(receiver)) {\n                    throw new TypeError("attempted to set private field on non-instance");\n                }\n                privateMap.set(receiver, value);\n                return value;\n            };'},e.getAllUnscopedEmitHelpers=function(){return t||(t=e.arrayToMap([e.decorateHelper,e.metadataHelper,e.paramHelper,e.assignHelper,e.awaitHelper,e.asyncGeneratorHelper,e.asyncDelegator,e.asyncValues,e.restHelper,e.awaiterHelper,e.extendsHelper,e.templateObjectHelper,e.spreadHelper,e.spreadArraysHelper,e.valuesHelper,e.readHelper,e.generatorHelper,e.importStarHelper,e.importDefaultHelper,e.exportStarHelper,e.classPrivateFieldGetHelper,e.classPrivateFieldSetHelper,e.createBindingHelper,e.setModuleDefaultHelper],(function(e){return e.name})))},e.asyncSuperHelper={name:"typescript:async-super",scoped:!0,text:n(__makeTemplateObject(["\n            const "," = name => super[name];"],["\n            const "," = name => super[name];"]),"_superIndex")},e.advancedAsyncSuperHelper={name:"typescript:advanced-async-super",scoped:!0,text:n(__makeTemplateObject(["\n            const "," = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"],["\n            const "," = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"]),"_superIndex")}}(ts||(ts={})),function(e){e.isNumericLiteral=function(e){return 8===e.kind},e.isBigIntLiteral=function(e){return 9===e.kind},e.isStringLiteral=function(e){return 10===e.kind},e.isJsxText=function(e){return 11===e.kind},e.isRegularExpressionLiteral=function(e){return 13===e.kind},e.isNoSubstitutionTemplateLiteral=function(e){return 14===e.kind},e.isTemplateHead=function(e){return 15===e.kind},e.isTemplateMiddle=function(e){return 16===e.kind},e.isTemplateTail=function(e){return 17===e.kind},e.isIdentifier=function(e){return 78===e.kind},e.isQualifiedName=function(e){return 157===e.kind},e.isComputedPropertyName=function(e){return 158===e.kind},e.isPrivateIdentifier=function(e){return 79===e.kind},e.isSuperKeyword=function(e){return 105===e.kind},e.isImportKeyword=function(e){return 99===e.kind},e.isCommaToken=function(e){return 27===e.kind},e.isQuestionToken=function(e){return 57===e.kind},e.isExclamationToken=function(e){return 53===e.kind},e.isTypeParameterDeclaration=function(e){return 159===e.kind},e.isParameter=function(e){return 160===e.kind},e.isDecorator=function(e){return 161===e.kind},e.isPropertySignature=function(e){return 162===e.kind},e.isPropertyDeclaration=function(e){return 163===e.kind},e.isMethodSignature=function(e){return 164===e.kind},e.isMethodDeclaration=function(e){return 165===e.kind},e.isConstructorDeclaration=function(e){return 166===e.kind},e.isGetAccessorDeclaration=function(e){return 167===e.kind},e.isSetAccessorDeclaration=function(e){return 168===e.kind},e.isCallSignatureDeclaration=function(e){return 169===e.kind},e.isConstructSignatureDeclaration=function(e){return 170===e.kind},e.isIndexSignatureDeclaration=function(e){return 171===e.kind},e.isTypePredicateNode=function(e){return 172===e.kind},e.isTypeReferenceNode=function(e){return 173===e.kind},e.isFunctionTypeNode=function(e){return 174===e.kind},e.isConstructorTypeNode=function(e){return 175===e.kind},e.isTypeQueryNode=function(e){return 176===e.kind},e.isTypeLiteralNode=function(e){return 177===e.kind},e.isArrayTypeNode=function(e){return 178===e.kind},e.isTupleTypeNode=function(e){return 179===e.kind},e.isNamedTupleMember=function(e){return 192===e.kind},e.isOptionalTypeNode=function(e){return 180===e.kind},e.isRestTypeNode=function(e){return 181===e.kind},e.isUnionTypeNode=function(e){return 182===e.kind},e.isIntersectionTypeNode=function(e){return 183===e.kind},e.isConditionalTypeNode=function(e){return 184===e.kind},e.isInferTypeNode=function(e){return 185===e.kind},e.isParenthesizedTypeNode=function(e){return 186===e.kind},e.isThisTypeNode=function(e){return 187===e.kind},e.isTypeOperatorNode=function(e){return 188===e.kind},e.isIndexedAccessTypeNode=function(e){return 189===e.kind},e.isMappedTypeNode=function(e){return 190===e.kind},e.isLiteralTypeNode=function(e){return 191===e.kind},e.isImportTypeNode=function(e){return 195===e.kind},e.isTemplateLiteralTypeSpan=function(e){return 194===e.kind},e.isTemplateLiteralTypeNode=function(e){return 193===e.kind},e.isObjectBindingPattern=function(e){return 196===e.kind},e.isArrayBindingPattern=function(e){return 197===e.kind},e.isBindingElement=function(e){return 198===e.kind},e.isArrayLiteralExpression=function(e){return 199===e.kind},e.isObjectLiteralExpression=function(e){return 200===e.kind},e.isPropertyAccessExpression=function(e){return 201===e.kind},e.isElementAccessExpression=function(e){return 202===e.kind},e.isCallExpression=function(e){return 203===e.kind},e.isNewExpression=function(e){return 204===e.kind},e.isTaggedTemplateExpression=function(e){return 205===e.kind},e.isTypeAssertionExpression=function(e){return 206===e.kind},e.isParenthesizedExpression=function(e){return 207===e.kind},e.isFunctionExpression=function(e){return 208===e.kind},e.isArrowFunction=function(e){return 209===e.kind},e.isDeleteExpression=function(e){return 210===e.kind},e.isTypeOfExpression=function(e){return 211===e.kind},e.isVoidExpression=function(e){return 212===e.kind},e.isAwaitExpression=function(e){return 213===e.kind},e.isPrefixUnaryExpression=function(e){return 214===e.kind},e.isPostfixUnaryExpression=function(e){return 215===e.kind},e.isBinaryExpression=function(e){return 216===e.kind},e.isConditionalExpression=function(e){return 217===e.kind},e.isTemplateExpression=function(e){return 218===e.kind},e.isYieldExpression=function(e){return 219===e.kind},e.isSpreadElement=function(e){return 220===e.kind},e.isClassExpression=function(e){return 221===e.kind},e.isOmittedExpression=function(e){return 222===e.kind},e.isExpressionWithTypeArguments=function(e){return 223===e.kind},e.isAsExpression=function(e){return 224===e.kind},e.isNonNullExpression=function(e){return 225===e.kind},e.isMetaProperty=function(e){return 226===e.kind},e.isSyntheticExpression=function(e){return 227===e.kind},e.isPartiallyEmittedExpression=function(e){return 336===e.kind},e.isCommaListExpression=function(e){return 337===e.kind},e.isTemplateSpan=function(e){return 228===e.kind},e.isSemicolonClassElement=function(e){return 229===e.kind},e.isBlock=function(e){return 230===e.kind},e.isVariableStatement=function(e){return 232===e.kind},e.isEmptyStatement=function(e){return 231===e.kind},e.isExpressionStatement=function(e){return 233===e.kind},e.isIfStatement=function(e){return 234===e.kind},e.isDoStatement=function(e){return 235===e.kind},e.isWhileStatement=function(e){return 236===e.kind},e.isForStatement=function(e){return 237===e.kind},e.isForInStatement=function(e){return 238===e.kind},e.isForOfStatement=function(e){return 239===e.kind},e.isContinueStatement=function(e){return 240===e.kind},e.isBreakStatement=function(e){return 241===e.kind},e.isReturnStatement=function(e){return 242===e.kind},e.isWithStatement=function(e){return 243===e.kind},e.isSwitchStatement=function(e){return 244===e.kind},e.isLabeledStatement=function(e){return 245===e.kind},e.isThrowStatement=function(e){return 246===e.kind},e.isTryStatement=function(e){return 247===e.kind},e.isDebuggerStatement=function(e){return 248===e.kind},e.isVariableDeclaration=function(e){return 249===e.kind},e.isVariableDeclarationList=function(e){return 250===e.kind},e.isFunctionDeclaration=function(e){return 251===e.kind},e.isClassDeclaration=function(e){return 252===e.kind},e.isInterfaceDeclaration=function(e){return 253===e.kind},e.isTypeAliasDeclaration=function(e){return 254===e.kind},e.isEnumDeclaration=function(e){return 255===e.kind},e.isModuleDeclaration=function(e){return 256===e.kind},e.isModuleBlock=function(e){return 257===e.kind},e.isCaseBlock=function(e){return 258===e.kind},e.isNamespaceExportDeclaration=function(e){return 259===e.kind},e.isImportEqualsDeclaration=function(e){return 260===e.kind},e.isImportDeclaration=function(e){return 261===e.kind},e.isImportClause=function(e){return 262===e.kind},e.isNamespaceImport=function(e){return 263===e.kind},e.isNamespaceExport=function(e){return 269===e.kind},e.isNamedImports=function(e){return 264===e.kind},e.isImportSpecifier=function(e){return 265===e.kind},e.isExportAssignment=function(e){return 266===e.kind},e.isExportDeclaration=function(e){return 267===e.kind},e.isNamedExports=function(e){return 268===e.kind},e.isExportSpecifier=function(e){return 270===e.kind},e.isMissingDeclaration=function(e){return 271===e.kind},e.isNotEmittedStatement=function(e){return 335===e.kind},e.isSyntheticReference=function(e){return 340===e.kind},e.isMergeDeclarationMarker=function(e){return 338===e.kind},e.isEndOfDeclarationMarker=function(e){return 339===e.kind},e.isExternalModuleReference=function(e){return 272===e.kind},e.isJsxElement=function(e){return 273===e.kind},e.isJsxSelfClosingElement=function(e){return 274===e.kind},e.isJsxOpeningElement=function(e){return 275===e.kind},e.isJsxClosingElement=function(e){return 276===e.kind},e.isJsxFragment=function(e){return 277===e.kind},e.isJsxOpeningFragment=function(e){return 278===e.kind},e.isJsxClosingFragment=function(e){return 279===e.kind},e.isJsxAttribute=function(e){return 280===e.kind},e.isJsxAttributes=function(e){return 281===e.kind},e.isJsxSpreadAttribute=function(e){return 282===e.kind},e.isJsxExpression=function(e){return 283===e.kind},e.isCaseClause=function(e){return 284===e.kind},e.isDefaultClause=function(e){return 285===e.kind},e.isHeritageClause=function(e){return 286===e.kind},e.isCatchClause=function(e){return 287===e.kind},e.isPropertyAssignment=function(e){return 288===e.kind},e.isShorthandPropertyAssignment=function(e){return 289===e.kind},e.isSpreadAssignment=function(e){return 290===e.kind},e.isEnumMember=function(e){return 291===e.kind},e.isUnparsedPrepend=function(e){return 293===e.kind},e.isSourceFile=function(e){return 297===e.kind},e.isBundle=function(e){return 298===e.kind},e.isUnparsedSource=function(e){return 299===e.kind},e.isJSDocTypeExpression=function(e){return 301===e.kind},e.isJSDocNameReference=function(e){return 302===e.kind},e.isJSDocAllType=function(e){return 303===e.kind},e.isJSDocUnknownType=function(e){return 304===e.kind},e.isJSDocNullableType=function(e){return 305===e.kind},e.isJSDocNonNullableType=function(e){return 306===e.kind},e.isJSDocOptionalType=function(e){return 307===e.kind},e.isJSDocFunctionType=function(e){return 308===e.kind},e.isJSDocVariadicType=function(e){return 309===e.kind},e.isJSDocNamepathType=function(e){return 310===e.kind},e.isJSDoc=function(e){return 311===e.kind},e.isJSDocTypeLiteral=function(e){return 312===e.kind},e.isJSDocSignature=function(e){return 313===e.kind},e.isJSDocAugmentsTag=function(e){return 315===e.kind},e.isJSDocAuthorTag=function(e){return 317===e.kind},e.isJSDocClassTag=function(e){return 319===e.kind},e.isJSDocCallbackTag=function(e){return 324===e.kind},e.isJSDocPublicTag=function(e){return 320===e.kind},e.isJSDocPrivateTag=function(e){return 321===e.kind},e.isJSDocProtectedTag=function(e){return 322===e.kind},e.isJSDocReadonlyTag=function(e){return 323===e.kind},e.isJSDocDeprecatedTag=function(e){return 318===e.kind},e.isJSDocEnumTag=function(e){return 325===e.kind},e.isJSDocParameterTag=function(e){return 326===e.kind},e.isJSDocReturnTag=function(e){return 327===e.kind},e.isJSDocThisTag=function(e){return 328===e.kind},e.isJSDocTypeTag=function(e){return 329===e.kind},e.isJSDocTemplateTag=function(e){return 330===e.kind},e.isJSDocTypedefTag=function(e){return 331===e.kind},e.isJSDocUnknownTag=function(e){return 314===e.kind},e.isJSDocPropertyTag=function(e){return 333===e.kind},e.isJSDocImplementsTag=function(e){return 316===e.kind},e.isSyntaxList=function(e){return 334===e.kind}}(ts||(ts={})),function(e){function n(n,t,r,a){if(e.isComputedPropertyName(r))return e.setTextRange(n.createElementAccessExpression(t,r.expression),a);var i=e.setTextRange(e.isIdentifierOrPrivateIdentifier(r)?n.createPropertyAccessExpression(t,r):n.createElementAccessExpression(t,r),r);return e.getOrCreateEmitNode(i).flags|=64,i}function t(n,t){var r=e.parseNodeFactory.createIdentifier(n||"React");return e.setParent(r,e.getParseTreeNode(t)),r}function r(n,a,i){if(e.isQualifiedName(a)){var o=r(n,a.left,i),s=n.createIdentifier(e.idText(a.right));return s.escapedText=a.right.escapedText,n.createPropertyAccessExpression(o,s)}return t(e.idText(a),i)}function a(e,n,a,i){return n?r(e,n,i):e.createPropertyAccessExpression(t(a,i),"createElement")}function i(n,t){return e.isIdentifier(t)?n.createStringLiteralFromNode(t):e.isComputedPropertyName(t)?e.setParent(e.setTextRange(n.cloneNode(t.expression),t.expression),t.expression.parent):e.setParent(e.setTextRange(n.cloneNode(t),t),t.parent)}function o(n){return e.isStringLiteral(n.expression)&&"use strict"===n.expression.text}function s(e,n){switch(void 0===n&&(n=15),e.kind){case 207:return 0!=(1&n);case 206:case 224:return 0!=(2&n);case 225:return 0!=(4&n);case 336:return 0!=(8&n)}return!1}function c(e,n){for(void 0===n&&(n=15);s(e,n);)e=e.expression;return e}function l(n){return e.setStartsOnNewLine(n,!0)}function u(n){var t=e.getOriginalNode(n,e.isSourceFile),r=t&&t.emitNode;return r&&r.externalHelpersModuleName}function d(n,t,r,a,i){if(r.importHelpers&&e.isEffectiveExternalModule(t,r)){var o=u(t);if(o)return o;var s=e.getEmitModuleKind(r),c=(a||r.esModuleInterop&&i)&&s!==e.ModuleKind.System&&s<e.ModuleKind.ES2015;if(!c){var l=e.getEmitHelpers(t);if(l)for(var d=0,p=l;d<p.length;d++){if(!p[d].scoped){c=!0;break}}}if(c){var m=e.getOriginalNode(t,e.isSourceFile),f=e.getOrCreateEmitNode(m);return f.externalHelpersModuleName||(f.externalHelpersModuleName=n.createUniqueName(e.externalHelpersModuleNameText))}}}function p(n,t,r,a){if(t)return t.moduleName?n.createStringLiteral(t.moduleName):!t.isDeclarationFile&&e.outFile(a)?n.createStringLiteral(e.getExternalModuleNameFromPath(r,t.fileName)):void 0}function m(n){if(e.isDeclarationBindingElement(n))return n.name;if(!e.isObjectLiteralElementLike(n))return e.isAssignmentExpression(n,!0)?m(n.left):e.isSpreadElement(n)?m(n.expression):n;switch(n.kind){case 288:return m(n.initializer);case 289:return n.name;case 290:return m(n.expression)}}function f(n){switch(n.kind){case 198:if(n.propertyName){var t=n.propertyName;return e.isPrivateIdentifier(t)?e.Debug.failBadSyntaxKind(t):e.isComputedPropertyName(t)&&_(t.expression)?t.expression:t}break;case 288:if(n.name){t=n.name;return e.isPrivateIdentifier(t)?e.Debug.failBadSyntaxKind(t):e.isComputedPropertyName(t)&&_(t.expression)?t.expression:t}break;case 290:return n.name&&e.isPrivateIdentifier(n.name)?e.Debug.failBadSyntaxKind(n.name):n.name}var r=m(n);if(r&&e.isPropertyName(r))return r}function _(e){var n=e.kind;return 10===n||8===n}e.createEmptyExports=function(e){return e.createExportDeclaration(void 0,void 0,!1,e.createNamedExports([]),void 0)},e.createMemberAccessForPropertyName=n,e.createJsxFactoryExpression=a,e.createExpressionForJsxElement=function(n,t,r,a,i,o){var s=[r];if(a&&s.push(a),i&&i.length>0)if(a||s.push(n.createNull()),i.length>1)for(var c=0,u=i;c<u.length;c++){var d=u[c];l(d),s.push(d)}else s.push(i[0]);return e.setTextRange(n.createCallExpression(t,void 0,s),o)},e.createExpressionForJsxFragment=function(n,i,o,s,c,u,d){var p=[function(e,n,a,i){return n?r(e,n,i):e.createPropertyAccessExpression(t(a,i),"Fragment")}(n,o,s,u),n.createNull()];if(c&&c.length>0)if(c.length>1)for(var m=0,f=c;m<f.length;m++){var _=f[m];l(_),p.push(_)}else p.push(c[0]);return e.setTextRange(n.createCallExpression(a(n,i,s,u),void 0,p),d)},e.createForOfBindingStatement=function(n,t,r){if(e.isVariableDeclarationList(t)){var a=e.first(t.declarations),i=n.updateVariableDeclaration(a,a.name,void 0,void 0,r);return e.setTextRange(n.createVariableStatement(void 0,n.updateVariableDeclarationList(t,[i])),t)}var o=e.setTextRange(n.createAssignment(t,r),t);return e.setTextRange(n.createExpressionStatement(o),t)},e.insertLeadingStatement=function(n,t,r){return e.isBlock(t)?n.updateBlock(t,e.setTextRange(n.createNodeArray(__spreadArrays([r],t.statements)),t.statements)):n.createBlock(n.createNodeArray([t,r]),!0)},e.createExpressionFromEntityName=function n(t,r){if(e.isQualifiedName(r)){var a=n(t,r.left),i=e.setParent(e.setTextRange(t.cloneNode(r.right),r.right),r.right.parent);return e.setTextRange(t.createPropertyAccessExpression(a,i),r)}return e.setParent(e.setTextRange(t.cloneNode(r),r),r.parent)},e.createExpressionForPropertyName=i,e.createExpressionForObjectLiteralElementLike=function(t,r,a,o){switch(a.name&&e.isPrivateIdentifier(a.name)&&e.Debug.failBadSyntaxKind(a.name,"Private identifiers are not allowed in object literals."),a.kind){case 167:case 168:return function(n,t,r,a,o){var s=e.getAllAccessorDeclarations(t,r),c=s.firstAccessor,l=s.getAccessor,u=s.setAccessor;if(r===c)return e.setTextRange(n.createObjectDefinePropertyCall(a,i(n,r.name),n.createPropertyDescriptor({enumerable:n.createFalse(),configurable:!0,get:l&&e.setTextRange(e.setOriginalNode(n.createFunctionExpression(l.modifiers,void 0,void 0,void 0,l.parameters,void 0,l.body),l),l),set:u&&e.setTextRange(e.setOriginalNode(n.createFunctionExpression(u.modifiers,void 0,void 0,void 0,u.parameters,void 0,u.body),u),u)},!o)),c)}(t,r.properties,a,o,!!r.multiLine);case 288:return function(t,r,a){return e.setOriginalNode(e.setTextRange(t.createAssignment(n(t,a,r.name,r.name),r.initializer),r),r)}(t,a,o);case 289:return function(t,r,a){return e.setOriginalNode(e.setTextRange(t.createAssignment(n(t,a,r.name,r.name),t.cloneNode(r.name)),r),r)}(t,a,o);case 165:return function(t,r,a){return e.setOriginalNode(e.setTextRange(t.createAssignment(n(t,a,r.name,r.name),e.setOriginalNode(e.setTextRange(t.createFunctionExpression(r.modifiers,r.asteriskToken,void 0,void 0,r.parameters,void 0,r.body),r),r)),r),r)}(t,a,o)}},e.isInternalName=function(n){return 0!=(32768&e.getEmitFlags(n))},e.isLocalName=function(n){return 0!=(16384&e.getEmitFlags(n))},e.isExportName=function(n){return 0!=(8192&e.getEmitFlags(n))},e.findUseStrictPrologue=function(n){for(var t=0,r=n;t<r.length;t++){var a=r[t];if(!e.isPrologueDirective(a))break;if(o(a))return a}},e.startsWithUseStrict=function(n){var t=e.firstOrUndefined(n);return void 0!==t&&e.isPrologueDirective(t)&&o(t)},e.isCommaSequence=function(e){return 216===e.kind&&27===e.operatorToken.kind||337===e.kind},e.isOuterExpression=s,e.skipOuterExpressions=c,e.skipAssertions=function(e){return c(e,6)},e.startOnNewLine=l,e.getExternalHelpersModuleName=u,e.hasRecordedExternalHelpers=function(n){var t=e.getOriginalNode(n,e.isSourceFile),r=t&&t.emitNode;return!(!r||!r.externalHelpersModuleName&&!r.externalHelpers)},e.createExternalHelpersImportDeclarationIfNeeded=function(n,t,r,a,i,o,s){if(a.importHelpers&&e.isEffectiveExternalModule(r,a)){var c=void 0,l=e.getEmitModuleKind(a);if(l>=e.ModuleKind.ES2015&&l<=e.ModuleKind.ESNext){var u=e.getEmitHelpers(r);if(u){for(var p=[],m=0,f=u;m<f.length;m++){var _=f[m];if(!_.scoped){var g=_.importName;g&&e.pushIfUnique(p,g)}}if(e.some(p)){p.sort(e.compareStringsCaseSensitive),c=n.createNamedImports(e.map(p,(function(a){return e.isFileLevelUniqueName(r,a)?n.createImportSpecifier(void 0,n.createIdentifier(a)):n.createImportSpecifier(n.createIdentifier(a),t.getUnscopedHelperName(a))})));var y=e.getOriginalNode(r,e.isSourceFile);e.getOrCreateEmitNode(y).externalHelpers=!0}}}else{var v=d(n,r,a,i,o||s);v&&(c=n.createNamespaceImport(v))}if(c){var h=n.createImportDeclaration(void 0,void 0,n.createImportClause(!1,void 0,c),n.createStringLiteral(e.externalHelpersModuleNameText));return e.addEmitFlags(h,67108864),h}}},e.getOrCreateExternalHelpersModuleNameIfNeeded=d,e.getLocalNameForExternalImport=function(n,t,r){var a=e.getNamespaceDeclarationNode(t);if(a&&!e.isDefaultImport(t)&&!e.isExportNamespaceAsDefaultDeclaration(t)){var i=a.name;return e.isGeneratedIdentifier(i)?i:n.createIdentifier(e.getSourceTextOfNodeFromSourceFile(r,i)||e.idText(i))}return 261===t.kind&&t.importClause||267===t.kind&&t.moduleSpecifier?n.getGeneratedNameForNode(t):void 0},e.getExternalModuleNameLiteral=function(n,t,r,a,i,o){var s=e.getExternalModuleName(t);if(10===s.kind)return function(e,n,t,r,a){return p(t,r.getExternalModuleFileFromDeclaration(e),n,a)}(t,a,n,i,o)||function(e,n,t){var r=t.renamedDependencies&&t.renamedDependencies.get(n.text);return r&&e.createStringLiteral(r)}(n,s,r)||n.cloneNode(s)},e.tryGetModuleNameFromFile=p,e.getInitializerOfBindingOrAssignmentElement=function n(t){if(e.isDeclarationBindingElement(t))return t.initializer;if(e.isPropertyAssignment(t)){var r=t.initializer;return e.isAssignmentExpression(r,!0)?r.right:void 0}return e.isShorthandPropertyAssignment(t)?t.objectAssignmentInitializer:e.isAssignmentExpression(t,!0)?t.right:e.isSpreadElement(t)?n(t.expression):void 0},e.getTargetOfBindingOrAssignmentElement=m,e.getRestIndicatorOfBindingOrAssignmentElement=function(e){switch(e.kind){case 160:case 198:return e.dotDotDotToken;case 220:case 290:return e}},e.getPropertyNameOfBindingOrAssignmentElement=function(n){var t=f(n);return e.Debug.assert(!!t||e.isSpreadAssignment(n),"Invalid property name for binding element."),t},e.tryGetPropertyNameOfBindingOrAssignmentElement=f,e.getElementsOfBindingOrAssignmentPattern=function(e){switch(e.kind){case 196:case 197:case 199:return e.elements;case 200:return e.properties}},e.getJSDocTypeAliasName=function(n){if(n)for(var t=n;;){if(e.isIdentifier(t)||!t.body)return e.isIdentifier(t)?t:t.name;t=t.body}},e.canHaveModifiers=function(e){var n=e.kind;return 160===n||162===n||163===n||164===n||165===n||166===n||167===n||168===n||171===n||208===n||209===n||221===n||232===n||251===n||252===n||253===n||254===n||255===n||256===n||260===n||261===n||266===n||267===n},e.isExportModifier=function(e){return 92===e.kind},e.isAsyncModifier=function(e){return 129===e.kind},e.isStaticModifier=function(e){return 123===e.kind}}(ts||(ts={})),function(e){e.setTextRange=function(n,t){return t?e.setTextRangePosEnd(n,t.pos,t.end):n}}(ts||(ts={})),function(e){var n,t,r,a,i,o,s,c,l;function u(e,n){return n&&e(n)}function d(e,n,t){if(t){if(n)return n(t);for(var r=0,a=t;r<a.length;r++){var i=e(a[r]);if(i)return i}}}function p(e,n){return 42===e.charCodeAt(n+1)&&42===e.charCodeAt(n+2)&&47!==e.charCodeAt(n+3)}function m(n,t,r){if(n&&!(n.kind<=156))switch(n.kind){case 157:return u(t,n.left)||u(t,n.right);case 159:return u(t,n.name)||u(t,n.constraint)||u(t,n.default)||u(t,n.expression);case 289:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.questionToken)||u(t,n.exclamationToken)||u(t,n.equalsToken)||u(t,n.objectAssignmentInitializer);case 290:return u(t,n.expression);case 160:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.dotDotDotToken)||u(t,n.name)||u(t,n.questionToken)||u(t,n.type)||u(t,n.initializer);case 163:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.questionToken)||u(t,n.exclamationToken)||u(t,n.type)||u(t,n.initializer);case 162:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.questionToken)||u(t,n.type)||u(t,n.initializer);case 288:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.questionToken)||u(t,n.initializer);case 249:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.exclamationToken)||u(t,n.type)||u(t,n.initializer);case 198:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.dotDotDotToken)||u(t,n.propertyName)||u(t,n.name)||u(t,n.initializer);case 174:case 175:case 169:case 170:case 171:return d(t,r,n.decorators)||d(t,r,n.modifiers)||d(t,r,n.typeParameters)||d(t,r,n.parameters)||u(t,n.type);case 165:case 164:case 166:case 167:case 168:case 208:case 251:case 209:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.asteriskToken)||u(t,n.name)||u(t,n.questionToken)||u(t,n.exclamationToken)||d(t,r,n.typeParameters)||d(t,r,n.parameters)||u(t,n.type)||u(t,n.equalsGreaterThanToken)||u(t,n.body);case 173:return u(t,n.typeName)||d(t,r,n.typeArguments);case 172:return u(t,n.assertsModifier)||u(t,n.parameterName)||u(t,n.type);case 176:return u(t,n.exprName);case 177:return d(t,r,n.members);case 178:return u(t,n.elementType);case 179:return d(t,r,n.elements);case 182:case 183:return d(t,r,n.types);case 184:return u(t,n.checkType)||u(t,n.extendsType)||u(t,n.trueType)||u(t,n.falseType);case 185:return u(t,n.typeParameter);case 195:return u(t,n.argument)||u(t,n.qualifier)||d(t,r,n.typeArguments);case 186:case 188:return u(t,n.type);case 189:return u(t,n.objectType)||u(t,n.indexType);case 190:return u(t,n.readonlyToken)||u(t,n.typeParameter)||u(t,n.nameType)||u(t,n.questionToken)||u(t,n.type);case 191:return u(t,n.literal);case 192:return u(t,n.dotDotDotToken)||u(t,n.name)||u(t,n.questionToken)||u(t,n.type);case 196:case 197:case 199:return d(t,r,n.elements);case 200:return d(t,r,n.properties);case 201:return u(t,n.expression)||u(t,n.questionDotToken)||u(t,n.name);case 202:return u(t,n.expression)||u(t,n.questionDotToken)||u(t,n.argumentExpression);case 203:case 204:return u(t,n.expression)||u(t,n.questionDotToken)||d(t,r,n.typeArguments)||d(t,r,n.arguments);case 205:return u(t,n.tag)||u(t,n.questionDotToken)||d(t,r,n.typeArguments)||u(t,n.template);case 206:return u(t,n.type)||u(t,n.expression);case 207:case 210:case 211:case 212:return u(t,n.expression);case 214:return u(t,n.operand);case 219:return u(t,n.asteriskToken)||u(t,n.expression);case 213:return u(t,n.expression);case 215:return u(t,n.operand);case 216:return u(t,n.left)||u(t,n.operatorToken)||u(t,n.right);case 224:return u(t,n.expression)||u(t,n.type);case 225:return u(t,n.expression);case 226:return u(t,n.name);case 217:return u(t,n.condition)||u(t,n.questionToken)||u(t,n.whenTrue)||u(t,n.colonToken)||u(t,n.whenFalse);case 220:return u(t,n.expression);case 230:case 257:return d(t,r,n.statements);case 297:return d(t,r,n.statements)||u(t,n.endOfFileToken);case 232:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.declarationList);case 250:return d(t,r,n.declarations);case 233:return u(t,n.expression);case 234:return u(t,n.expression)||u(t,n.thenStatement)||u(t,n.elseStatement);case 235:return u(t,n.statement)||u(t,n.expression);case 236:return u(t,n.expression)||u(t,n.statement);case 237:return u(t,n.initializer)||u(t,n.condition)||u(t,n.incrementor)||u(t,n.statement);case 238:return u(t,n.initializer)||u(t,n.expression)||u(t,n.statement);case 239:return u(t,n.awaitModifier)||u(t,n.initializer)||u(t,n.expression)||u(t,n.statement);case 240:case 241:return u(t,n.label);case 242:return u(t,n.expression);case 243:return u(t,n.expression)||u(t,n.statement);case 244:return u(t,n.expression)||u(t,n.caseBlock);case 258:return d(t,r,n.clauses);case 284:return u(t,n.expression)||d(t,r,n.statements);case 285:return d(t,r,n.statements);case 245:return u(t,n.label)||u(t,n.statement);case 246:return u(t,n.expression);case 247:return u(t,n.tryBlock)||u(t,n.catchClause)||u(t,n.finallyBlock);case 287:return u(t,n.variableDeclaration)||u(t,n.block);case 161:return u(t,n.expression);case 252:case 221:case 253:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||d(t,r,n.typeParameters)||d(t,r,n.heritageClauses)||d(t,r,n.members);case 254:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||d(t,r,n.typeParameters)||u(t,n.type);case 255:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||d(t,r,n.members);case 291:return u(t,n.name)||u(t,n.initializer);case 256:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.body);case 260:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.name)||u(t,n.moduleReference);case 261:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.importClause)||u(t,n.moduleSpecifier);case 262:return u(t,n.name)||u(t,n.namedBindings);case 259:case 263:case 269:return u(t,n.name);case 264:case 268:return d(t,r,n.elements);case 267:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.exportClause)||u(t,n.moduleSpecifier);case 265:case 270:return u(t,n.propertyName)||u(t,n.name);case 266:return d(t,r,n.decorators)||d(t,r,n.modifiers)||u(t,n.expression);case 218:return u(t,n.head)||d(t,r,n.templateSpans);case 228:return u(t,n.expression)||u(t,n.literal);case 193:return u(t,n.head)||d(t,r,n.templateSpans);case 194:return u(t,n.type)||u(t,n.literal);case 158:return u(t,n.expression);case 286:return d(t,r,n.types);case 223:return u(t,n.expression)||d(t,r,n.typeArguments);case 272:return u(t,n.expression);case 271:return d(t,r,n.decorators);case 337:return d(t,r,n.elements);case 273:return u(t,n.openingElement)||d(t,r,n.children)||u(t,n.closingElement);case 277:return u(t,n.openingFragment)||d(t,r,n.children)||u(t,n.closingFragment);case 274:case 275:return u(t,n.tagName)||d(t,r,n.typeArguments)||u(t,n.attributes);case 281:return d(t,r,n.properties);case 280:return u(t,n.name)||u(t,n.initializer);case 282:return u(t,n.expression);case 283:return u(t,n.dotDotDotToken)||u(t,n.expression);case 276:return u(t,n.tagName);case 180:case 181:case 301:case 306:case 305:case 307:case 309:return u(t,n.type);case 308:return d(t,r,n.parameters)||u(t,n.type);case 311:return d(t,r,n.tags);case 332:return u(t,n.tagName)||u(t,n.name);case 302:return u(t,n.name);case 326:case 333:return u(t,n.tagName)||(n.isNameFirst?u(t,n.name)||u(t,n.typeExpression):u(t,n.typeExpression)||u(t,n.name));case 317:return u(t,n.tagName);case 316:case 315:return u(t,n.tagName)||u(t,n.class);case 330:return u(t,n.tagName)||u(t,n.constraint)||d(t,r,n.typeParameters);case 331:return u(t,n.tagName)||(n.typeExpression&&301===n.typeExpression.kind?u(t,n.typeExpression)||u(t,n.fullName):u(t,n.fullName)||u(t,n.typeExpression));case 324:return u(t,n.tagName)||u(t,n.fullName)||u(t,n.typeExpression);case 327:case 329:case 328:case 325:return u(t,n.tagName)||u(t,n.typeExpression);case 313:return e.forEach(n.typeParameters,t)||e.forEach(n.parameters,t)||u(t,n.type);case 312:return e.forEach(n.jsDocPropertyTags,t);case 314:case 319:case 320:case 321:case 322:case 323:return u(t,n.tagName);case 336:return u(t,n.expression)}}function f(e){return void 0!==e.externalModuleIndicator}function _(n){return e.fileExtensionIs(n,".d.ts")}function g(n,t){for(var r=[],a=0,i=e.getLeadingCommentRanges(t,0)||e.emptyArray;a<i.length;a++){var o=i[a];T(r,o,t.substring(o.pos,o.end))}n.pragmas=new e.Map;for(var s=0,c=r;s<c.length;s++){var l=c[s];if(n.pragmas.has(l.name)){var u=n.pragmas.get(l.name);u instanceof Array?u.push(l.args):n.pragmas.set(l.name,[u,l.args])}else n.pragmas.set(l.name,l.args)}}function y(n,t){n.checkJsDirective=void 0,n.referencedFiles=[],n.typeReferenceDirectives=[],n.libReferenceDirectives=[],n.amdDependencies=[],n.hasNoDefaultLib=!1,n.pragmas.forEach((function(r,a){switch(a){case"reference":var i=n.referencedFiles,o=n.typeReferenceDirectives,s=n.libReferenceDirectives;e.forEach(e.toArray(r),(function(r){var a=r.arguments,c=a.types,l=a.lib,u=a.path;r.arguments["no-default-lib"]?n.hasNoDefaultLib=!0:c?o.push({pos:c.pos,end:c.end,fileName:c.value}):l?s.push({pos:l.pos,end:l.end,fileName:l.value}):u?i.push({pos:u.pos,end:u.end,fileName:u.value}):t(r.range.pos,r.range.end-r.range.pos,e.Diagnostics.Invalid_reference_directive_syntax)}));break;case"amd-dependency":n.amdDependencies=e.map(e.toArray(r),(function(e){return{name:e.arguments.name,path:e.arguments.path}}));break;case"amd-module":if(r instanceof Array)for(var c=0,l=r;c<l.length;c++){var u=l[c];n.moduleName&&t(u.range.pos,u.range.end-u.range.pos,e.Diagnostics.An_AMD_module_cannot_have_multiple_name_assignments),n.moduleName=u.arguments.name}else n.moduleName=r.arguments.name;break;case"ts-nocheck":case"ts-check":e.forEach(e.toArray(r),(function(e){(!n.checkJsDirective||e.range.pos>n.checkJsDirective.pos)&&(n.checkJsDirective={enabled:"ts-check"===a,end:e.range.end,pos:e.range.pos})}));break;case"jsx":case"jsxfrag":case"jsximportsource":case"jsxruntime":return;default:e.Debug.fail("Unhandled pragma kind")}}))}!function(e){e[e.None=0]="None",e[e.Yield=1]="Yield",e[e.Await=2]="Await",e[e.Type=4]="Type",e[e.IgnoreMissingOpenBrace=16]="IgnoreMissingOpenBrace",e[e.JSDoc=32]="JSDoc"}(n||(n={})),function(e){e[e.TryParse=0]="TryParse",e[e.Lookahead=1]="Lookahead",e[e.Reparse=2]="Reparse"}(t||(t={})),e.parseBaseNodeFactory={createBaseSourceFileNode:function(n){return new(s||(s=e.objectAllocator.getSourceFileConstructor()))(n,-1,-1)},createBaseIdentifierNode:function(n){return new(i||(i=e.objectAllocator.getIdentifierConstructor()))(n,-1,-1)},createBasePrivateIdentifierNode:function(n){return new(o||(o=e.objectAllocator.getPrivateIdentifierConstructor()))(n,-1,-1)},createBaseTokenNode:function(n){return new(a||(a=e.objectAllocator.getTokenConstructor()))(n,-1,-1)},createBaseNode:function(n){return new(r||(r=e.objectAllocator.getNodeConstructor()))(n,-1,-1)}},e.parseNodeFactory=e.createNodeFactory(1,e.parseBaseNodeFactory),e.isJSDocLikeText=p,e.forEachChild=m,e.forEachChildRecursively=function(n,t,r){for(var a=[n];a.length;){var i=a.pop(),o=c(i,s(i));if(o)return o}return;function s(e){var n=[];return m(e,t,t),n;function t(e){n.unshift(e)}}function c(n,i){for(var o=0,s=i;o<s.length;o++){var c=s[o];if(e.isArray(c)){if(r)if(d=r(c,n)){if("skip"===d)continue;return d}for(var l=c.length-1;l>=0;l--){var u=c[l];if(d=t(u,n)){if("skip"===d)continue;return d}a.push(u)}}else{var d;if(a.push(c),d=t(c,n)){if("skip"===d)continue;return d}}}}},e.createSourceFile=function(n,t,r,a,i){void 0===a&&(a=!1);var o,s=["parse","createSourceFile",{path:n}];return e.tracing.begin.apply(e.tracing,s),e.performance.mark("beforeParse"),e.perfLogger.logStartParseSourceFile(n),o=100===r?c.parseSourceFile(n,t,r,void 0,a,6):c.parseSourceFile(n,t,r,void 0,a,i),e.perfLogger.logStopParseSourceFile(),e.performance.mark("afterParse"),e.performance.measure("Parse","beforeParse","afterParse"),e.tracing.end.apply(e.tracing,s),o},e.parseIsolatedEntityName=function(e,n){return c.parseIsolatedEntityName(e,n)},e.parseJsonText=function(e,n){return c.parseJsonText(e,n)},e.isExternalModule=f,e.updateSourceFile=function(e,n,t,r){void 0===r&&(r=!1);var a=l.updateSourceFile(e,n,t,r);return a.flags|=3145728&e.flags,a},e.parseIsolatedJSDocComment=function(e,n,t){var r=c.JSDocParser.parseIsolatedJSDocComment(e,n,t);return r&&r.jsDoc&&c.fixupParentReferences(r.jsDoc),r},e.parseJSDocTypeExpressionForTests=function(e,n,t){return c.JSDocParser.parseJSDocTypeExpressionForTests(e,n,t)},function(n){var t,r,a,i,o,s=e.createScanner(99,!0),c=20480;function u(e){return C++,e}var d,v,h,b,E,T,S,L,A,D,C,I,N,R,k,O,P,M={createBaseSourceFileNode:function(e){return u(new o(e,0,0))},createBaseIdentifierNode:function(e){return u(new a(e,0,0))},createBasePrivateIdentifierNode:function(e){return u(new i(e,0,0))},createBaseTokenNode:function(e){return u(new r(e,0,0))},createBaseNode:function(e){return u(new t(e,0,0))}},w=e.createNodeFactory(11,M),F=!0,G=!1;function B(n,t,r,a,i){void 0===r&&(r=2),void 0===i&&(i=!1),U(n,t,r,a,6),v=P,Ee();var o,s,c=ge();if(1===ve())o=je([],c,c),s=Be();else{var l=void 0;switch(ve()){case 22:l=Sr();break;case 109:case 94:case 103:l=Be();break;case 40:l=Ie((function(){return 8===Ee()&&58!==Ee()}))?Zt():Lr();break;case 8:case 10:if(Ie((function(){return 58!==Ee()}))){l=Pn();break}default:l=Lr()}var u=w.createExpressionStatement(l);He(u,c),o=je([u],c),s=Ge(1,e.Diagnostics.Unexpected_token)}var d=Y(n,2,6,!1,o,s,v);i&&X(d),d.nodeCount=C,d.identifierCount=R,d.identifiers=I,d.parseDiagnostics=e.attachFileToDiagnostics(S,d),L&&(d.jsDocDiagnostics=e.attachFileToDiagnostics(L,d));var p=d;return V(),p}function U(n,c,l,u,p){switch(t=e.objectAllocator.getNodeConstructor(),r=e.objectAllocator.getTokenConstructor(),a=e.objectAllocator.getIdentifierConstructor(),i=e.objectAllocator.getPrivateIdentifierConstructor(),o=e.objectAllocator.getSourceFileConstructor(),d=e.normalizePath(n),h=c,b=l,A=u,E=p,T=e.getLanguageVariant(p),S=[],k=0,I=new e.Map,N=new e.Map,R=0,C=0,v=0,F=!0,E){case 1:case 2:P=131072;break;case 6:P=33685504;break;default:P=0}G=!1,s.setText(h),s.setOnError(_e),s.setScriptTarget(b),s.setLanguageVariant(T)}function V(){s.clearCommentDirectives(),s.setText(""),s.setOnError(void 0),h=void 0,b=void 0,A=void 0,E=void 0,T=void 0,v=0,S=void 0,L=void 0,k=0,I=void 0,O=void 0,F=!0}function j(n,t,r){var a=_(d);a&&(P|=8388608),v=P,Ee();var i=yn(0,Hr);e.Debug.assert(1===ve());var o=J(Be()),c=Y(d,n,r,a,i,o,v);return g(c,h),y(c,(function(n,t,r){S.push(e.createDetachedDiagnostic(d,n,t,r))})),c.commentDirectives=s.getCommentDirectives(),c.nodeCount=C,c.identifierCount=R,c.identifiers=I,c.parseDiagnostics=e.attachFileToDiagnostics(S,c),L&&(c.jsDocDiagnostics=e.attachFileToDiagnostics(L,c)),t&&X(c),c}function H(e,n){return n?J(e):e}n.parseSourceFile=function(n,t,r,a,i,o){if(void 0===i&&(i=!1),6===(o=e.ensureScriptKind(n,o))){var s=B(n,t,r,a,i);return e.convertToObjectWorker(s,s.parseDiagnostics,!1,void 0,void 0),s.referencedFiles=e.emptyArray,s.typeReferenceDirectives=e.emptyArray,s.libReferenceDirectives=e.emptyArray,s.amdDependencies=e.emptyArray,s.hasNoDefaultLib=!1,s.pragmas=e.emptyMap,s}U(n,t,r,a,o);var c=j(r,i,o);return V(),c},n.parseIsolatedEntityName=function(e,n){U("",e,n,void 0,1),Ee();var t=An(!0),r=1===ve()&&!S.length;return V(),r?t:void 0},n.parseJsonText=B;var K,W,z,q=!1;function J(n){e.Debug.assert(!n.jsDoc);var t=e.mapDefined(e.getJSDocCommentRanges(n,h),(function(e){return z.parseJSDocComment(n,e.pos,e.end-e.pos)}));return t.length&&(n.jsDoc=t),q&&(q=!1,n.flags|=134217728),n}function X(n){e.setParentRecursive(n,!0)}function Y(n,t,r,a,i,o,c){var u=w.createSourceFile(i,o,c);return e.setTextRangePosWidth(u,0,h.length),function(n){n.externalModuleIndicator=e.forEach(n.statements,wa)||function(e){return 2097152&e.flags?Fa(e):void 0}(n)}(u),!a&&f(u)&&8388608&u.transformFlags&&(u=function(n){var t=A,r=l.createSyntaxCursor(n);A={currentNode:function(e){var n=r.currentNode(e);return F&&n&&m(n)&&(n.intersectsChange=!0),n}};var a=[],i=S;S=[];for(var o=0,c=f(n.statements,0),u=function(){var t=n.statements[o],r=n.statements[c];e.addRange(a,n.statements,o,c),o=_(n.statements,c);var l=e.findIndex(i,(function(e){return e.start>=t.pos})),u=l>=0?e.findIndex(i,(function(e){return e.start>=r.pos}),l):-1;l>=0&&e.addRange(S,i,l,u>=0?u:void 0),Ce((function(){var e=P;for(P|=32768,s.setTextPos(r.pos),Ee();1!==ve();){var t=s.getStartPos(),i=vn(0,Hr);if(a.push(i),t===s.getStartPos()&&Ee(),o>=0){var c=n.statements[o];if(i.end===c.pos)break;i.end>c.pos&&(o=_(n.statements,o+1))}}P=e}),2),c=o>=0?f(n.statements,o):-1};-1!==c;)u();if(o>=0){var d=n.statements[o];e.addRange(a,n.statements,o);var p=e.findIndex(i,(function(e){return e.start>=d.pos}));p>=0&&e.addRange(S,i,p)}return A=t,w.updateSourceFile(n,e.setTextRange(w.createNodeArray(a),n.statements));function m(e){return!(32768&e.flags||!(8388608&e.transformFlags))}function f(e,n){for(var t=n;t<e.length;t++)if(m(e[t]))return t;return-1}function _(e,n){for(var t=n;t<e.length;t++)if(!m(e[t]))return t;return-1}}(u)),u.text=h,u.bindDiagnostics=[],u.bindSuggestionDiagnostics=void 0,u.languageVersion=t,u.fileName=n,u.languageVariant=e.getLanguageVariant(r),u.isDeclarationFile=a,u.scriptKind=r,u}function Q(e,n){e?P|=n:P&=~n}function Z(e){Q(e,4096)}function $(e){Q(e,8192)}function ee(e){Q(e,16384)}function ne(e){Q(e,32768)}function te(e,n){var t=e&P;if(t){Q(!1,t);var r=n();return Q(!0,t),r}return n()}function re(e,n){var t=e&~P;if(t){Q(!0,t);var r=n();return Q(!1,t),r}return n()}function ae(e){return te(4096,e)}function ie(e){return re(32768,e)}function oe(e){return 0!=(P&e)}function se(){return oe(8192)}function ce(){return oe(4096)}function le(){return oe(16384)}function ue(){return oe(32768)}function de(e,n){me(s.getTokenPos(),s.getTextPos(),e,n)}function pe(n,t,r,a){var i=e.lastOrUndefined(S);i&&n===i.start||S.push(e.createDetachedDiagnostic(d,n,t,r,a)),G=!0}function me(e,n,t,r){pe(e,n-e,t,r)}function fe(e,n,t){me(e.pos,e.end,n,t)}function _e(e,n){pe(s.getTextPos(),n,e)}function ge(){return s.getStartPos()}function ye(){return s.hasPrecedingJSDocComment()}function ve(){return D}function he(){return D=s.scan()}function be(e){return Ee(),e()}function Ee(){return e.isKeyword(D)&&(s.hasUnicodeEscape()||s.hasExtendedUnicodeEscape())&&me(s.getTokenPos(),s.getTextPos(),e.Diagnostics.Keywords_cannot_contain_escape_characters),he()}function Te(){return D=s.scanJsDocToken()}function Se(){return D=s.reScanGreaterToken()}function xe(){return D=s.reScanTemplateHeadOrNoSubstitutionTemplate()}function Le(){return D=s.reScanLessThanToken()}function Ae(){return D=s.scanJsxIdentifier()}function De(){return D=s.scanJsxToken()}function Ce(n,t){var r=D,a=S.length,i=G,o=P,c=0!==t?s.lookAhead(n):s.tryScan(n);return e.Debug.assert(o===P),c&&0===t||(D=r,2!==t&&(S.length=a),G=i),c}function Ie(e){return Ce(e,1)}function Ne(e){return Ce(e,0)}function Re(){return 78===ve()||ve()>115}function ke(){return 78===ve()||(124!==ve()||!se())&&((130!==ve()||!ue())&&ve()>115)}function Oe(n,t,r){return void 0===r&&(r=!0),ve()===n?(r&&Ee(),!0):(t?de(t):de(e.Diagnostics._0_expected,e.tokenToString(n)),!1)}function Pe(n){return ve()===n?(Te(),!0):(de(e.Diagnostics._0_expected,e.tokenToString(n)),!1)}function Me(e){return ve()===e&&(Ee(),!0)}function we(e){if(ve()===e)return Be()}function Fe(e){if(ve()===e)return n=ge(),t=ve(),Te(),He(w.createToken(t),n);var n,t}function Ge(n,t,r){return we(n)||Ke(n,!1,t||e.Diagnostics._0_expected,r||e.tokenToString(n))}function Be(){var e=ge(),n=ve();return Ee(),He(w.createToken(n),e)}function Ue(){return 26===ve()||(19===ve()||1===ve()||s.hasPrecedingLineBreak())}function Ve(){return Ue()?(26===ve()&&Ee(),!0):Oe(26)}function je(n,t,r,a){var i=w.createNodeArray(n,a);return e.setTextRangePosEnd(i,t,null!=r?r:s.getStartPos()),i}function He(n,t,r){return e.setTextRangePosEnd(n,t,null!=r?r:s.getStartPos()),P&&(n.flags|=P),G&&(G=!1,n.flags|=65536),n}function Ke(n,t,r,a){t?pe(s.getStartPos(),0,r,a):r&&de(r,a);var i=ge();return He(78===n?w.createIdentifier("",void 0,void 0):e.isTemplateLiteralKind(n)?w.createTemplateLiteralLikeNode(n,"","",void 0):8===n?w.createNumericLiteral("",void 0):10===n?w.createStringLiteral("",void 0):271===n?w.createMissingDeclaration():w.createToken(n),i)}function We(e){var n=I.get(e);return void 0===n&&I.set(e,n=e),n}function ze(n,t,r){if(R++,n){var a=ge(),i=ve(),o=We(s.getTokenValue());return he(),He(w.createIdentifier(o,void 0,i),a)}if(79===ve())return de(r||e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies),ze(!0);var c=1===ve(),l=s.isReservedWord(),u=s.getTokenText(),d=l?e.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here:e.Diagnostics.Identifier_expected;return Ke(78,c,t||d,u)}function qe(e){return ze(Re(),void 0,e)}function Je(e,n){return ze(ke(),e,n)}function Xe(n){return ze(e.tokenIsIdentifierOrKeyword(ve()),n)}function Ye(){return e.tokenIsIdentifierOrKeyword(ve())||10===ve()||8===ve()}function Qe(e){if(10===ve()||8===ve()){var n=Pn();return n.text=We(n.text),n}return e&&22===ve()?function(){var e=ge();Oe(22);var n=ae(Bt);return Oe(23),He(w.createComputedPropertyName(n),e)}():79===ve()?$e():Xe()}function Ze(){return Qe(!0)}function $e(){var e,n,t=ge(),r=w.createPrivateIdentifier((e=s.getTokenText(),void 0===(n=N.get(e))&&N.set(e,n=e),n));return Ee(),He(r,t)}function en(e){return ve()===e&&Ne(tn)}function nn(){return Ee(),!s.hasPrecedingLineBreak()&&on()}function tn(){switch(ve()){case 84:return 91===Ee();case 92:return Ee(),87===ve()?Ie(sn):149===ve()?Ie(an):rn();case 87:return sn();case 123:case 134:case 146:return Ee(),on();default:return nn()}}function rn(){return 41!==ve()&&126!==ve()&&18!==ve()&&on()}function an(){return Ee(),rn()}function on(){return 22===ve()||18===ve()||41===ve()||25===ve()||Ye()}function sn(){return Ee(),83===ve()||97===ve()||117===ve()||125===ve()&&Ie(wr)||129===ve()&&Ie(Fr)}function cn(n,t){if(hn(n))return!0;switch(n){case 0:case 1:case 3:return!(26===ve()&&t)&&Vr();case 2:return 81===ve()||87===ve();case 4:return Ie(at);case 5:return Ie(ua)||26===ve()&&!t;case 6:return 22===ve()||Ye();case 12:switch(ve()){case 22:case 41:case 25:case 24:return!0;default:return Ye()}case 18:return Ye();case 9:return 22===ve()||25===ve()||Ye();case 7:return 18===ve()?Ie(ln):t?ke()&&!mn():Ft()&&!mn();case 8:return Qr();case 10:return 27===ve()||25===ve()||Qr();case 19:return ke();case 15:switch(ve()){case 27:case 24:return!0}case 11:return 25===ve()||Gt();case 16:return zn(!1);case 17:return zn(!0);case 20:case 21:return 27===ve()||St();case 22:return xa();case 23:return e.tokenIsIdentifierOrKeyword(ve());case 13:return e.tokenIsIdentifierOrKeyword(ve())||18===ve();case 14:return!0}return e.Debug.fail("Non-exhaustive case in 'isListElement'.")}function ln(){if(e.Debug.assert(18===ve()),19===Ee()){var n=Ee();return 27===n||18===n||93===n||116===n}return!0}function un(){return Ee(),ke()}function dn(){return Ee(),e.tokenIsIdentifierOrKeyword(ve())}function pn(){return Ee(),e.tokenIsIdentifierOrKeywordOrGreaterThan(ve())}function mn(){return(116===ve()||93===ve())&&Ie(fn)}function fn(){return Ee(),Gt()}function _n(){return Ee(),St()}function gn(e){if(1===ve())return!0;switch(e){case 1:case 2:case 4:case 5:case 6:case 12:case 9:case 23:return 19===ve();case 3:return 19===ve()||81===ve()||87===ve();case 7:return 18===ve()||93===ve()||116===ve();case 8:return function(){if(Ue())return!0;if(Xt(ve()))return!0;if(38===ve())return!0;return!1}();case 19:return 31===ve()||20===ve()||18===ve()||93===ve()||116===ve();case 11:return 21===ve()||26===ve();case 15:case 21:case 10:return 23===ve();case 17:case 16:case 18:return 21===ve()||23===ve();case 20:return 27!==ve();case 22:return 18===ve()||19===ve();case 13:return 31===ve()||43===ve();case 14:return 29===ve()&&Ie(Na);default:return!1}}function yn(e,n){var t=k;k|=1<<e;for(var r=[],a=ge();!gn(e);)if(cn(e,!1)){var i=vn(e,n);r.push(i)}else if(En(e))break;return k=t,je(r,a)}function vn(e,n){var t=hn(e);return t?bn(t):n()}function hn(n){if(A&&function(e){switch(e){case 5:case 2:case 0:case 1:case 3:case 6:case 4:case 8:case 17:case 16:return!0}return!1}(n)&&!G){var t=A.currentNode(s.getStartPos());if(!(e.nodeIsMissing(t)||t.intersectsChange||e.containsParseError(t)))if((25358336&t.flags)===P&&function(e,n){switch(n){case 5:return function(e){if(e)switch(e.kind){case 166:case 171:case 167:case 168:case 163:case 229:return!0;case 165:var n=e;return!(78===n.name.kind&&132===n.name.originalKeywordKind)}return!1}(e);case 2:return function(e){if(e)switch(e.kind){case 284:case 285:return!0}return!1}(e);case 0:case 1:case 3:return function(e){if(e)switch(e.kind){case 251:case 232:case 230:case 234:case 233:case 246:case 242:case 244:case 241:case 240:case 238:case 239:case 237:case 236:case 243:case 231:case 247:case 245:case 235:case 248:case 261:case 260:case 267:case 266:case 256:case 252:case 253:case 255:case 254:return!0}return!1}(e);case 6:return function(e){return 291===e.kind}(e);case 4:return function(e){if(e)switch(e.kind){case 170:case 164:case 171:case 162:case 169:return!0}return!1}(e);case 8:return function(e){if(249!==e.kind)return!1;return void 0===e.initializer}(e);case 17:case 16:return function(e){if(160!==e.kind)return!1;return void 0===e.initializer}(e)}return!1}(t,n))return t.jsDocCache&&(t.jsDocCache=void 0),t}}function bn(e){return s.setTextPos(e.end),Ee(),e}function En(n){return function(n){switch(n){case 0:case 1:return de(e.Diagnostics.Declaration_or_statement_expected);case 2:return de(e.Diagnostics.case_or_default_expected);case 3:return de(e.Diagnostics.Statement_expected);case 18:case 4:return de(e.Diagnostics.Property_or_signature_expected);case 5:return de(e.Diagnostics.Unexpected_token_A_constructor_method_accessor_or_property_was_expected);case 6:return de(e.Diagnostics.Enum_member_expected);case 7:return de(e.Diagnostics.Expression_expected);case 8:return e.isKeyword(ve())?de(e.Diagnostics._0_is_not_allowed_as_a_variable_declaration_name,e.tokenToString(ve())):de(e.Diagnostics.Variable_declaration_expected);case 9:return de(e.Diagnostics.Property_destructuring_pattern_expected);case 10:return de(e.Diagnostics.Array_element_destructuring_pattern_expected);case 11:return de(e.Diagnostics.Argument_expression_expected);case 12:return de(e.Diagnostics.Property_assignment_expected);case 15:return de(e.Diagnostics.Expression_or_comma_expected);case 17:case 16:return de(e.Diagnostics.Parameter_declaration_expected);case 19:return de(e.Diagnostics.Type_parameter_declaration_expected);case 20:return de(e.Diagnostics.Type_argument_expected);case 21:return de(e.Diagnostics.Type_expected);case 22:return de(e.Diagnostics.Unexpected_token_expected);case 23:case 13:case 14:return de(e.Diagnostics.Identifier_expected);default:;}}(n),!!function(){for(var e=0;e<24;e++)if(k&1<<e&&(cn(e,!0)||gn(e)))return!0;return!1}()||(Ee(),!1)}function Tn(e,n,t){var r=k;k|=1<<e;for(var a=[],i=ge(),o=-1;;)if(cn(e,!1)){var c=s.getStartPos();if(a.push(vn(e,n)),o=s.getTokenPos(),Me(27))continue;if(o=-1,gn(e))break;Oe(27,Sn(e)),t&&26===ve()&&!s.hasPrecedingLineBreak()&&Ee(),c===s.getStartPos()&&Ee()}else{if(gn(e))break;if(En(e))break}return k=r,je(a,i,void 0,o>=0)}function Sn(n){return 6===n?e.Diagnostics.An_enum_member_name_must_be_followed_by_a_or:void 0}function xn(){var e=je([],ge());return e.isMissingList=!0,e}function Ln(e,n,t,r){if(Oe(t)){var a=Tn(e,n);return Oe(r),a}return xn()}function An(e,n){for(var t=ge(),r=e?Xe(n):Je(n),a=ge();Me(24);){if(29===ve()){r.jsdocDotPos=a;break}a=ge(),r=He(w.createQualifiedName(r,Cn(e,!1)),t)}return r}function Dn(e,n){return He(w.createQualifiedName(e,n),e.pos)}function Cn(n,t){if(s.hasPrecedingLineBreak()&&e.tokenIsIdentifierOrKeyword(ve())&&Ie(Mr))return Ke(78,!0,e.Diagnostics.Identifier_expected);if(79===ve()){var r=$e();return t?r:Ke(78,!0,e.Diagnostics.Identifier_expected)}return n?Xe():Je()}function In(e){var n=ge();return He(w.createTemplateExpression(Mn(e),function(e){var n,t=ge(),r=[];do{n=On(e),r.push(n)}while(16===n.literal.kind);return je(r,t)}(e)),n)}function Nn(){var e=ge();return He(w.createTemplateLiteralType(Mn(!1),function(){var e,n=ge(),t=[];do{e=Rn(),t.push(e)}while(16===e.literal.kind);return je(t,n)}()),e)}function Rn(){var e=ge();return He(w.createTemplateLiteralTypeSpan(Pt(),kn(!1)),e)}function kn(n){return 19===ve()?(function(e){D=s.reScanTemplateToken(e)}(n),t=wn(ve()),e.Debug.assert(16===t.kind||17===t.kind,"Template fragment has wrong token kind"),t):Ge(17,e.Diagnostics._0_expected,e.tokenToString(19));var t}function On(e){var n=ge();return He(w.createTemplateSpan(ae(Bt),kn(e)),n)}function Pn(){return wn(ve())}function Mn(n){n&&xe();var t=wn(ve());return e.Debug.assert(15===t.kind,"Template head has wrong token kind"),t}function wn(n){var t=ge(),r=e.isTemplateLiteralKind(n)?w.createTemplateLiteralLikeNode(n,s.getTokenValue(),function(e){var n=14===e||17===e,t=s.getTokenText();return t.substring(1,t.length-(s.isUnterminated()?0:n?1:2))}(n),2048&s.getTokenFlags()):8===n?w.createNumericLiteral(s.getTokenValue(),s.getNumericLiteralFlags()):10===n?w.createStringLiteral(s.getTokenValue(),void 0,s.hasExtendedUnicodeEscape()):e.isLiteralKind(n)?w.createLiteralLikeNode(n,s.getTokenValue()):e.Debug.fail();return s.hasExtendedUnicodeEscape()&&(r.hasExtendedUnicodeEscape=!0),s.isUnterminated()&&(r.isUnterminated=!0),Ee(),He(r,t)}function Fn(){return An(!0,e.Diagnostics.Type_expected)}function Gn(){if(!s.hasPrecedingLineBreak()&&29===Le())return Ln(20,Pt,29,31)}function Bn(){var e=ge();return He(w.createTypeReferenceNode(Fn(),Gn()),e)}function Un(n){switch(n.kind){case 173:return e.nodeIsMissing(n.typeName);case 174:case 175:var t=n,r=t.parameters,a=t.type;return!!r.isMissingList||Un(a);case 186:return Un(n.type);default:return!1}}function Vn(){var e=ge();return Ee(),He(w.createThisTypeNode(),e)}function jn(){var e,n=ge();return 107!==ve()&&102!==ve()||(e=Xe(),Oe(58)),He(w.createParameterDeclaration(void 0,void 0,void 0,e,void 0,Hn(),void 0),n)}function Hn(){s.setInJSDocType(!0);var e=ge();if(Me(139)){var n=w.createJSDocNamepathType(void 0);e:for(;;)switch(ve()){case 19:case 1:case 27:case 5:break e;default:Te()}return s.setInJSDocType(!1),He(n,e)}var t=Me(25),r=kt();return s.setInJSDocType(!1),t&&(r=He(w.createJSDocVariadicType(r),e)),62===ve()?(Ee(),He(w.createJSDocOptionalType(r),e)):r}function Kn(){var e,n,t=ge(),r=Je();Me(93)&&(St()||!Gt()?e=Pt():n=$t());var a=Me(62)?Pt():void 0,i=w.createTypeParameterDeclaration(r,e,a);return i.expression=n,He(i,t)}function Wn(){if(29===ve())return Ln(19,Kn,29,31)}function zn(n){return 25===ve()||Qr()||e.isModifierKind(ve())||59===ve()||St(!n)}function qn(){return Xn(!0)}function Jn(){return Xn(!1)}function Xn(n){var t=ge(),r=ye();if(107===ve())return H(He(w.createParameterDeclaration(void 0,void 0,void 0,ze(!0),void 0,wt(),void 0),t),r);var a=n?ie(ma):ma(),i=F;F=!1;var o=_a(),s=H(He(w.createParameterDeclaration(a,o,we(25),function(n){var t=Zr(e.Diagnostics.Private_identifiers_cannot_be_used_as_parameters);return 0===e.getFullWidth(t)&&!e.some(n)&&e.isModifierKind(ve())&&Ee(),t}(o),we(57),wt(),Ut()),t),r);return F=i,s}function Yn(n,t){if(function(n,t){if(38===n)return Oe(n),!0;if(Me(58))return!0;if(t&&38===ve())return de(e.Diagnostics._0_expected,e.tokenToString(58)),Ee(),!0;return!1}(n,t))return kt()}function Qn(e){var n=se(),t=ue();$(!!(1&e)),ne(!!(2&e));var r=32&e?Tn(17,jn):Tn(16,t?qn:Jn);return $(n),ne(t),r}function Zn(e){if(!Oe(20))return xn();var n=Qn(e);return Oe(21),n}function $n(){Me(27)||Ve()}function et(e){var n=ge(),t=ye();170===e&&Oe(102);var r=Wn(),a=Zn(4),i=Yn(58,!0);return $n(),H(He(169===e?w.createCallSignature(r,a,i):w.createConstructSignature(r,a,i),n),t)}function nt(){return 22===ve()&&Ie(tt)}function tt(){if(Ee(),25===ve()||23===ve())return!0;if(e.isModifierKind(ve())){if(Ee(),ke())return!0}else{if(!ke())return!1;Ee()}return 58===ve()||27===ve()||57===ve()&&(Ee(),58===ve()||27===ve()||23===ve())}function rt(e,n,t,r){var a=Ln(16,Jn,22,23),i=wt();return $n(),H(He(w.createIndexSignature(t,r,a,i),e),n)}function at(){if(20===ve()||29===ve())return!0;for(var n=!1;e.isModifierKind(ve());)n=!0,Ee();return 22===ve()||(Ye()&&(n=!0,Ee()),!!n&&(20===ve()||29===ve()||57===ve()||58===ve()||27===ve()||Ue()))}function it(){if(20===ve()||29===ve())return et(169);if(102===ve()&&Ie(ot))return et(170);var e=ge(),n=ye(),t=_a();return nt()?rt(e,n,void 0,t):function(e,n,t){var r,a=Ze(),i=we(57);if(20===ve()||29===ve()){var o=Wn(),s=Zn(4),c=Yn(58,!0);r=w.createMethodSignature(t,a,i,o,s,c)}else c=wt(),r=w.createPropertySignature(t,a,i,c),62===ve()&&(r.initializer=Ut());return $n(),H(He(r,e),n)}(e,n,t)}function ot(){return Ee(),20===ve()||29===ve()}function st(){return 24===Ee()}function ct(){switch(Ee()){case 20:case 29:case 24:return!0}return!1}function lt(){var e;return Oe(18)?(e=yn(4,it),Oe(19)):e=xn(),e}function ut(){return Ee(),39===ve()||40===ve()?142===Ee():(142===ve()&&Ee(),22===ve()&&un()&&100===Ee())}function dt(){var e,n=ge();Oe(18),142!==ve()&&39!==ve()&&40!==ve()||142!==(e=Be()).kind&&Oe(142),Oe(22);var t,r=function(){var e=ge(),n=Xe();Oe(100);var t=Pt();return He(w.createTypeParameterDeclaration(n,t,void 0),e)}(),a=Me(126)?Pt():void 0;Oe(23),57!==ve()&&39!==ve()&&40!==ve()||57!==(t=Be()).kind&&Oe(57);var i=wt();return Ve(),Oe(19),He(w.createMappedTypeNode(e,r,a,t,i),n)}function pt(){var n=ge();if(Me(25))return He(w.createRestTypeNode(Pt()),n);var t=Pt();if(e.isJSDocNullableType(t)&&t.pos===t.type.pos){var r=w.createOptionalTypeNode(t.type);return e.setTextRange(r,t),r.flags=t.flags,r}return t}function mt(){return 58===Ee()||57===ve()&&58===Ee()}function ft(){return 25===ve()?e.tokenIsIdentifierOrKeyword(Ee())&&mt():e.tokenIsIdentifierOrKeyword(ve())&&mt()}function _t(){if(Ie(ft)){var e=ge(),n=ye(),t=we(25),r=Xe(),a=we(57);Oe(58);var i=pt();return H(He(w.createNamedTupleMember(t,r,a,i),e),n)}return pt()}function gt(){var e=ge(),n=ye(),t=Me(102),r=Wn(),a=Zn(4),i=Yn(38,!1);return H(He(t?w.createConstructorTypeNode(r,a,i):w.createFunctionTypeNode(r,a,i),e),n)}function yt(){var e=Be();return 24===ve()?void 0:e}function vt(e){var n=ge();e&&Ee();var t=109===ve()||94===ve()||103===ve()?Be():wn(ve());return e&&(t=He(w.createPrefixUnaryExpression(40,t),n)),He(w.createLiteralTypeNode(t),n)}function ht(){return Ee(),99===ve()}function bt(){v|=1048576;var e=ge(),n=Me(111);Oe(99),Oe(20);var t=Pt();Oe(21);var r=Me(24)?Fn():void 0,a=Gn();return He(w.createImportTypeNode(t,r,a,n),e)}function Et(){return Ee(),8===ve()||9===ve()}function Tt(){switch(ve()){case 128:case 152:case 147:case 144:case 155:case 148:case 131:case 150:case 141:case 145:return Ne(yt)||Bn();case 65:s.reScanAsteriskEqualsToken();case 41:return t=ge(),Ee(),He(w.createJSDocAllType(),t);case 60:s.reScanQuestionToken();case 57:return function(){var e=ge();return Ee(),27===ve()||19===ve()||21===ve()||31===ve()||62===ve()||51===ve()?He(w.createJSDocUnknownType(),e):He(w.createJSDocNullableType(Pt()),e)}();case 97:return function(){var e=ge(),n=ye();if(Ie(Ia)){Ee();var t=Zn(36),r=Yn(58,!1);return H(He(w.createJSDocFunctionType(t,r),e),n)}return He(w.createTypeReferenceNode(Xe(),void 0),e)}();case 53:return function(){var e=ge();return Ee(),He(w.createJSDocNonNullableType(Tt()),e)}();case 14:case 10:case 8:case 9:case 109:case 94:case 103:return vt();case 40:return Ie(Et)?vt(!0):Bn();case 113:return Be();case 107:var e=Vn();return 137!==ve()||s.hasPrecedingLineBreak()?e:(n=e,Ee(),He(w.createTypePredicateNode(void 0,n,Pt()),n.pos));case 111:return Ie(ht)?bt():function(){var e=ge();return Oe(111),He(w.createTypeQueryNode(An(!0)),e)}();case 18:return Ie(ut)?dt():function(){var e=ge();return He(w.createTypeLiteralNode(lt()),e)}();case 22:return function(){var e=ge();return He(w.createTupleTypeNode(Ln(21,_t,22,23)),e)}();case 20:return function(){var e=ge();Oe(20);var n=Pt();return Oe(21),He(w.createParenthesizedType(n),e)}();case 99:return bt();case 127:return Ie(Mr)?function(){var e=ge(),n=Ge(127),t=107===ve()?Vn():Je(),r=Me(137)?Pt():void 0;return He(w.createTypePredicateNode(n,t,r),e)}():Bn();case 15:return Nn();default:return Bn()}var n,t}function St(e){switch(ve()){case 128:case 152:case 147:case 144:case 155:case 131:case 142:case 148:case 151:case 113:case 150:case 103:case 107:case 111:case 141:case 18:case 22:case 29:case 51:case 50:case 102:case 10:case 8:case 9:case 109:case 94:case 145:case 41:case 57:case 53:case 25:case 135:case 99:case 127:case 14:case 15:return!0;case 97:return!e;case 40:return!e&&Ie(Et);case 20:return!e&&Ie(xt);default:return ke()}}function xt(){return Ee(),21===ve()||zn(!1)||St()}function Lt(){var e=ge();return Oe(135),He(w.createInferTypeNode(function(){var e=ge();return He(w.createTypeParameterDeclaration(Je(),void 0,void 0),e)}()),e)}function At(){var e=ve();switch(e){case 138:case 151:case 142:return function(e){var n=ge();return Oe(e),He(w.createTypeOperatorNode(e,At()),n)}(e);case 135:return Lt()}return function(){for(var e=ge(),n=Tt();!s.hasPrecedingLineBreak();)switch(ve()){case 53:Ee(),n=He(w.createJSDocNonNullableType(n),e);break;case 57:if(Ie(_n))return n;Ee(),n=He(w.createJSDocNullableType(n),e);break;case 22:if(Oe(22),St()){var t=Pt();Oe(23),n=He(w.createIndexedAccessTypeNode(n,t),e)}else Oe(23),n=He(w.createArrayTypeNode(n),e);break;default:return n}return n}()}function Dt(n){if(Nt()){var t=gt();return fe(t,e.isFunctionTypeNode(t)?n?e.Diagnostics.Function_type_notation_must_be_parenthesized_when_used_in_a_union_type:e.Diagnostics.Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type:n?e.Diagnostics.Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type:e.Diagnostics.Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type),t}}function Ct(e,n,t){var r=ge(),a=51===e,i=Me(e),o=i&&Dt(a)||n();if(ve()===e||i){for(var s=[o];Me(e);)s.push(Dt(a)||n());o=He(t(je(s,r)),r)}return o}function It(){return Ct(50,At,w.createIntersectionTypeNode)}function Nt(){return 29===ve()||(!(20!==ve()||!Ie(Rt))||102===ve())}function Rt(){if(Ee(),21===ve()||25===ve())return!0;if(function(){if(e.isModifierKind(ve())&&_a(),ke()||107===ve())return Ee(),!0;if(22===ve()||18===ve()){var n=S.length;return Zr(),n===S.length}return!1}()){if(58===ve()||27===ve()||57===ve()||62===ve())return!0;if(21===ve()&&(Ee(),38===ve()))return!0}return!1}function kt(){var e=ge(),n=ke()&&Ne(Ot),t=Pt();return n?He(w.createTypePredicateNode(void 0,n,t),e):t}function Ot(){var e=Je();if(137===ve()&&!s.hasPrecedingLineBreak())return Ee(),e}function Pt(){return te(40960,Mt)}function Mt(e){if(Nt())return gt();var n=ge(),t=Ct(51,It,w.createUnionTypeNode);if(!e&&!s.hasPrecedingLineBreak()&&Me(93)){var r=Mt(!0);Oe(57);var a=Mt();Oe(58);var i=Mt();return He(w.createConditionalTypeNode(t,r,a,i),n)}return t}function wt(){return Me(58)?Pt():void 0}function Ft(){switch(ve()){case 107:case 105:case 103:case 109:case 94:case 8:case 9:case 10:case 14:case 15:case 20:case 22:case 18:case 97:case 83:case 102:case 43:case 67:case 78:return!0;case 99:return Ie(ct);default:return ke()}}function Gt(){if(Ft())return!0;switch(ve()){case 39:case 40:case 54:case 53:case 88:case 111:case 113:case 45:case 46:case 29:case 130:case 124:case 79:return!0;default:return!!function(){if(ce()&&100===ve())return!1;return e.getBinaryOperatorPrecedence(ve())>0}()||ke()}}function Bt(){var e=le();e&&ee(!1);for(var n,t=ge(),r=Vt();n=we(27);)r=Qt(r,n,Vt(),t);return e&&ee(!0),r}function Ut(){return Me(62)?Vt():void 0}function Vt(){if(function(){if(124===ve())return!!se()||Ie(Gr);return!1}())return function(){var e=ge();return Ee(),s.hasPrecedingLineBreak()||41!==ve()&&!Gt()?He(w.createYieldExpression(void 0,void 0),e):He(w.createYieldExpression(we(41),Vt()),e)}();var n=function(){var e=function(){if(20===ve()||29===ve()||129===ve())return Ie(Ht);if(38===ve())return 1;return 0}();if(0===e)return;return 1===e?zt(!0):Ne(Kt)}()||function(){if(129===ve()&&1===Ie(Wt)){var e=ge(),n=ga();return jt(e,Jt(0),n)}return}();if(n)return n;var t=ge(),r=Jt(0);return 78===r.kind&&38===ve()?jt(t,r,void 0):e.isLeftHandSideExpression(r)&&e.isAssignmentOperator(Se())?Qt(r,Be(),Vt(),t):function(n,t){var r,a=we(57);if(!a)return n;return He(w.createConditionalExpression(n,a,te(c,Vt),r=Ge(58),e.nodeIsPresent(r)?Vt():Ke(78,!1,e.Diagnostics._0_expected,e.tokenToString(58))),t)}(r,t)}function jt(n,t,r){e.Debug.assert(38===ve(),"parseSimpleArrowFunctionExpression should only have been called if we had a =>");var a=w.createParameterDeclaration(void 0,void 0,void 0,t,void 0,void 0,void 0);He(a,t.pos);var i=je([a],a.pos,a.end),o=Ge(38),s=qt(!!r);return J(He(w.createArrowFunction(r,void 0,i,void 0,o,s),n))}function Ht(){if(129===ve()){if(Ee(),s.hasPrecedingLineBreak())return 0;if(20!==ve()&&29!==ve())return 0}var n=ve(),t=Ee();if(20===n){if(21===t)switch(Ee()){case 38:case 58:case 18:return 1;default:return 0}if(22===t||18===t)return 2;if(25===t)return 1;if(e.isModifierKind(t)&&129!==t&&Ie(un))return 1;if(!ke()&&107!==t)return 0;switch(Ee()){case 58:return 1;case 57:return Ee(),58===ve()||27===ve()||62===ve()||21===ve()?1:0;case 27:case 62:case 21:return 2}return 0}return e.Debug.assert(29===n),ke()?1===T?Ie((function(){var e=Ee();if(93===e)switch(Ee()){case 62:case 31:return!1;default:return!0}else if(27===e)return!0;return!1}))?1:0:2:0}function Kt(){var n=s.getTokenPos();if(!(null==O?void 0:O.has(n))){var t=zt(!1);return t||(O||(O=new e.Set)).add(n),t}}function Wt(){if(129===ve()){if(Ee(),s.hasPrecedingLineBreak()||38===ve())return 0;var e=Jt(0);if(!s.hasPrecedingLineBreak()&&78===e.kind&&38===ve())return 1}return 0}function zt(n){var t,r=ge(),a=ye(),i=ga(),o=e.some(i,e.isAsyncModifier)?2:0,s=Wn();if(Oe(20)){if(t=Qn(o),!Oe(21)&&!n)return}else{if(!n)return;t=xn()}var c=Yn(58,!1);if(!c||n||!Un(c)){var l=c&&e.isJSDocFunctionType(c);if(n||38===ve()||!l&&18===ve()){var u=ve(),d=Ge(38),p=38===u||18===u?qt(e.some(i,e.isAsyncModifier)):Je();return H(He(w.createArrowFunction(i,s,t,c,d,p),r),a)}}}function qt(e){if(18===ve())return Ir(e?2:0);if(26!==ve()&&97!==ve()&&83!==ve()&&Vr()&&(18===ve()||97===ve()||83===ve()||59===ve()||!Gt()))return Ir(16|(e?2:0));var n=F;F=!1;var t=e?ie(Vt):te(32768,Vt);return F=n,t}function Jt(e){var n=ge();return Yt(e,$t(),n)}function Xt(e){return 100===e||156===e}function Yt(n,t,r){for(;;){Se();var a=e.getBinaryOperatorPrecedence(ve());if(!(42===ve()?a>=n:a>n))break;if(100===ve()&&ce())break;if(126===ve()){if(s.hasPrecedingLineBreak())break;Ee(),i=t,o=Pt(),t=He(w.createAsExpression(i,o),i.pos)}else t=Qt(t,Be(),Jt(a),r)}var i,o;return t}function Qt(e,n,t,r){return He(w.createBinaryExpression(e,n,t),r)}function Zt(){var e=ge();return He(w.createPrefixUnaryExpression(ve(),be(er)),e)}function $t(){if(function(){switch(ve()){case 39:case 40:case 54:case 53:case 88:case 111:case 113:case 130:return!1;case 29:if(1!==T)return!1;default:return!0}}()){var n=ge(),t=nr();return 42===ve()?Yt(e.getBinaryOperatorPrecedence(ve()),t,n):t}var r=ve(),a=er();if(42===ve()){n=e.skipTrivia(h,a.pos);var i=a.end;206===a.kind?me(n,i,e.Diagnostics.A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses):me(n,i,e.Diagnostics.An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses,e.tokenToString(r))}return a}function er(){switch(ve()){case 39:case 40:case 54:case 53:return Zt();case 88:return e=ge(),He(w.createDeleteExpression(be(er)),e);case 111:return function(){var e=ge();return He(w.createTypeOfExpression(be(er)),e)}();case 113:return function(){var e=ge();return He(w.createVoidExpression(be(er)),e)}();case 29:return function(){var e=ge();Oe(29);var n=Pt();Oe(31);var t=er();return He(w.createTypeAssertion(n,t),e)}();case 130:if(130===ve()&&(ue()||Ie(Gr)))return function(){var e=ge();return He(w.createAwaitExpression(be(er)),e)}();default:return nr()}var e}function nr(){if(45===ve()||46===ve()){var n=ge();return He(w.createPrefixUnaryExpression(ve(),be(tr)),n)}if(1===T&&29===ve()&&Ie(pn))return ar(!0);var t=tr();if(e.Debug.assert(e.isLeftHandSideExpression(t)),(45===ve()||46===ve())&&!s.hasPrecedingLineBreak()){var r=ve();return Ee(),He(w.createPostfixUnaryExpression(t,r),t.pos)}return t}function tr(){var n,t=ge();return 99===ve()?Ie(ot)?(v|=1048576,n=Be()):Ie(st)?(Ee(),Ee(),n=He(w.createMetaProperty(99,Xe()),t),v|=2097152):n=rr():n=105===ve()?function(){var n=ge(),t=Be();if(29===ve()){var r=ge();void 0!==Ne(hr)&&me(r,ge(),e.Diagnostics.super_may_not_use_type_arguments)}if(20===ve()||24===ve()||22===ve())return t;return Ge(24,e.Diagnostics.super_must_be_followed_by_an_argument_list_or_member_access),He(w.createPropertyAccessExpression(t,Cn(!0,!0)),n)}():rr(),yr(t,n)}function rr(){return fr(ge(),br(),!0)}function ar(n,t){var r,a=ge(),i=function(e){var n=ge();if(Oe(29),31===ve())return De(),He(w.createJsxOpeningFragment(),n);var t,r=sr(),a=Sa(),i=function(){var e=ge();return He(w.createJsxAttributes(yn(13,lr)),e)}();31===ve()?(De(),t=w.createJsxOpeningElement(r,a,i)):(Oe(43),e?Oe(31):(Oe(31,void 0,!1),De()),t=w.createJsxSelfClosingElement(r,a,i));return He(t,n)}(n);if(275===i.kind){var o=or(i),s=function(e){var n=ge();Oe(30);var t=sr();e?Oe(31):(Oe(31,void 0,!1),De());return He(w.createJsxClosingElement(t),n)}(n);x(i.tagName,s.tagName)||fe(s,e.Diagnostics.Expected_corresponding_JSX_closing_tag_for_0,e.getTextOfNodeFromSourceText(h,i.tagName)),r=He(w.createJsxElement(i,o,s),a)}else 278===i.kind?r=He(w.createJsxFragment(i,or(i),function(n){var t=ge();Oe(30),e.tokenIsIdentifierOrKeyword(ve())&&fe(sr(),e.Diagnostics.Expected_corresponding_closing_tag_for_JSX_fragment);n?Oe(31):(Oe(31,void 0,!1),De());return He(w.createJsxJsxClosingFragment(),t)}(n)),a):(e.Debug.assert(274===i.kind),r=i);if(n&&29===ve()){var c=void 0===t?r.pos:t,l=Ne((function(){return ar(!0,c)}));if(l){var u=Ke(27,!1);return e.setTextRangePosWidth(u,l.pos,0),me(e.skipTrivia(h,c),l.end,e.Diagnostics.JSX_expressions_must_have_one_parent_element),He(w.createBinaryExpression(r,u,l),a)}}return r}function ir(n,t){switch(t){case 1:if(e.isJsxOpeningFragment(n))fe(n,e.Diagnostics.JSX_fragment_has_no_corresponding_closing_tag);else{var r=n.tagName;me(e.skipTrivia(h,r.pos),r.end,e.Diagnostics.JSX_element_0_has_no_corresponding_closing_tag,e.getTextOfNodeFromSourceText(h,n.tagName))}return;case 30:case 7:return;case 11:case 12:return a=ge(),i=w.createJsxText(s.getTokenValue(),12===D),D=s.scanJsxToken(),He(i,a);case 18:return cr(!1);case 29:return ar(!1);default:return e.Debug.assertNever(t)}var a,i}function or(e){var n=[],t=ge(),r=k;for(k|=16384;;){var a=ir(e,D=s.reScanJsxToken());if(!a)break;n.push(a)}return k=r,je(n,t)}function sr(){var e=ge();Ae();for(var n=107===ve()?Be():Xe();Me(24);)n=He(w.createPropertyAccessExpression(n,Cn(!0,!1)),e);return n}function cr(e){var n,t,r=ge();if(Oe(18))return 19!==ve()&&(n=we(25),t=Bt()),e?Oe(19):Oe(19,void 0,!1)&&De(),He(w.createJsxExpression(n,t),r)}function lr(){if(18===ve())return function(){var e=ge();Oe(18),Oe(25);var n=Bt();return Oe(19),He(w.createJsxSpreadAttribute(n),e)}();Ae();var e=ge();return He(w.createJsxAttribute(Xe(),62!==ve()?void 0:10===(D=s.scanJsxAttributeValue())?Pn():cr(!0)),e)}function ur(){return Ee(),e.tokenIsIdentifierOrKeyword(ve())||22===ve()||_r()}function dr(n){if(32&n.flags)return!0;if(e.isNonNullExpression(n)){for(var t=n.expression;e.isNonNullExpression(t)&&!(32&t.flags);)t=t.expression;if(32&t.flags){for(;e.isNonNullExpression(n);)n.flags|=32,n=n.expression;return!0}}return!1}function pr(n,t,r){var a=Cn(!0,!0),i=r||dr(t),o=i?w.createPropertyAccessChain(t,r,a):w.createPropertyAccessExpression(t,a);return i&&e.isPrivateIdentifier(o.name)&&fe(o.name,e.Diagnostics.An_optional_chain_cannot_contain_private_identifiers),He(o,n)}function mr(n,t,r){var a;if(23===ve())a=Ke(78,!0,e.Diagnostics.An_element_access_expression_should_take_an_argument);else{var i=ae(Bt);e.isStringOrNumericLiteralLike(i)&&(i.text=We(i.text)),a=i}return Oe(23),He(r||dr(t)?w.createElementAccessChain(t,r,a):w.createElementAccessExpression(t,a),n)}function fr(n,t,r){for(;;){var a=void 0,i=!1;if(r&&28===ve()&&Ie(ur)?(a=Ge(28),i=e.tokenIsIdentifierOrKeyword(ve())):i=Me(24),i)t=pr(n,t,a);else if(a||53!==ve()||s.hasPrecedingLineBreak())if(!a&&le()||!Me(22)){if(!_r())return t;t=gr(n,t,a,void 0)}else t=mr(n,t,a);else Ee(),t=He(w.createNonNullExpression(t),n)}}function _r(){return 14===ve()||15===ve()}function gr(e,n,t,r){var a=w.createTaggedTemplateExpression(n,r,14===ve()?(xe(),Pn()):In(!0));return(t||32&n.flags)&&(a.flags|=32),a.questionDotToken=t,He(a,e)}function yr(n,t){for(;;){t=fr(n,t,!0);var r=we(28);if(29===ve()||47===ve()){var a=Ne(hr);if(a){if(_r()){t=gr(n,t,r,a);continue}var i=vr();t=He(r||dr(t)?w.createCallChain(t,r,a,i):w.createCallExpression(t,a,i),n);continue}}else if(20===ve()){i=vr();t=He(r||dr(t)?w.createCallChain(t,r,void 0,i):w.createCallExpression(t,void 0,i),n);continue}if(r){var o=Ke(78,!1,e.Diagnostics.Identifier_expected);t=He(w.createPropertyAccessChain(t,r,o),n)}break}return t}function vr(){Oe(20);var e=Tn(11,Tr);return Oe(21),e}function hr(){if(29===Le()){Ee();var e=Tn(20,Pt);if(Oe(31))return e&&function(){switch(ve()){case 20:case 14:case 15:case 24:case 21:case 23:case 58:case 26:case 57:case 34:case 36:case 35:case 37:case 55:case 56:case 60:case 52:case 50:case 51:case 19:case 1:return!0;case 27:case 18:default:return!1}}()?e:void 0}}function br(){switch(ve()){case 8:case 9:case 10:case 14:return Pn();case 107:case 105:case 103:case 109:case 94:return Be();case 20:return function(){var e=ge(),n=ye();Oe(20);var t=ae(Bt);return Oe(21),H(He(w.createParenthesizedExpression(t),e),n)}();case 22:return Sr();case 18:return Lr();case 129:if(!Ie(Fr))break;return Ar();case 83:return ha(ge(),ye(),void 0,void 0,221);case 97:return Ar();case 102:return function(){var n=ge();if(Oe(102),Me(24)){var t=Xe();return He(w.createMetaProperty(102,t),n)}var r,a,i=ge(),o=br();for(;;){o=fr(i,o,!1),r=Ne(hr),_r()&&(e.Debug.assert(!!r,"Expected a type argument list; all plain tagged template starts should be consumed in 'parseMemberExpressionRest'"),o=gr(i,o,void 0,r),r=void 0);break}20===ve()?a=vr():r&&me(n,s.getStartPos(),e.Diagnostics.A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list);return He(w.createNewExpression(o,r,a),n)}();case 43:case 67:if(13===(D=s.reScanSlashToken()))return Pn();break;case 15:return In(!1)}return Je(e.Diagnostics.Expression_expected)}function Er(){return 25===ve()?function(){var e=ge();Oe(25);var n=Vt();return He(w.createSpreadElement(n),e)}():27===ve()?He(w.createOmittedExpression(),ge()):Vt()}function Tr(){return te(c,Er)}function Sr(){var e=ge();Oe(22);var n=s.hasPrecedingLineBreak(),t=Tn(15,Er);return Oe(23),He(w.createArrayLiteralExpression(t,n),e)}function xr(){var e=ge(),n=ye();if(we(25)){var t=Vt();return H(He(w.createSpreadAssignment(t),e),n)}var r=ma(),a=_a();if(en(134))return la(e,n,r,a,167);if(en(146))return la(e,n,r,a,168);var i,o=we(41),s=ke(),c=Ze(),l=we(57),u=we(53);if(o||20===ve()||29===ve())return oa(e,n,r,a,o,c,l,u);if(s&&58!==ve()){var d=we(62),p=d?ae(Vt):void 0;(i=w.createShorthandPropertyAssignment(c,p)).equalsToken=d}else{Oe(58);var m=ae(Vt);i=w.createPropertyAssignment(c,m)}return i.decorators=r,i.modifiers=a,i.questionToken=l,i.exclamationToken=u,H(He(i,e),n)}function Lr(){var n=ge(),t=s.getTokenPos();Oe(18);var r=s.hasPrecedingLineBreak(),a=Tn(12,xr,!0);if(!Oe(19)){var i=e.lastOrUndefined(S);i&&i.code===e.Diagnostics._0_expected.code&&e.addRelatedInfo(i,e.createDetachedDiagnostic(d,t,1,e.Diagnostics.The_parser_expected_to_find_a_to_match_the_token_here))}return He(w.createObjectLiteralExpression(a,r),n)}function Ar(){var n=le();n&&ee(!1);var t=ge(),r=ye(),a=_a();Oe(97);var i=we(41),o=i?1:0,s=e.some(a,e.isAsyncModifier)?2:0,c=o&&s?re(40960,Dr):o?function(e){return re(8192,e)}(Dr):s?ie(Dr):Dr(),l=Wn(),u=Zn(o|s),d=Yn(58,!1),p=Ir(o|s);return n&&ee(!0),H(He(w.createFunctionExpression(a,i,c,l,u,d,p),t),r)}function Dr(){return Re()?qe():void 0}function Cr(n,t){var r=ge(),a=s.getTokenPos();if(Oe(18,t)||n){var i=s.hasPrecedingLineBreak(),o=yn(1,Hr);if(!Oe(19)){var c=e.lastOrUndefined(S);c&&c.code===e.Diagnostics._0_expected.code&&e.addRelatedInfo(c,e.createDetachedDiagnostic(d,a,1,e.Diagnostics.The_parser_expected_to_find_a_to_match_the_token_here))}return He(w.createBlock(o,i),r)}o=xn();return He(w.createBlock(o,void 0),r)}function Ir(e,n){var t=se();$(!!(1&e));var r=ue();ne(!!(2&e));var a=F;F=!1;var i=le();i&&ee(!1);var o=Cr(!!(16&e),n);return i&&ee(!0),F=a,$(t),ne(r),o}function Nr(){var e=ge();Oe(96);var n,t,r=we(130);if(Oe(20),26!==ve()&&(n=112===ve()||118===ve()||84===ve()?na(!0):re(4096,Bt)),r?Oe(156):Me(156)){var a=ae(Vt);Oe(21),t=w.createForOfStatement(r,n,a,Hr())}else if(Me(100)){a=ae(Bt);Oe(21),t=w.createForInStatement(n,a,Hr())}else{Oe(26);var i=26!==ve()&&21!==ve()?ae(Bt):void 0;Oe(26);var o=21!==ve()?ae(Bt):void 0;Oe(21),t=w.createForStatement(n,i,o,Hr())}return He(t,e)}function Rr(e){var n=ge();Oe(241===e?80:85);var t=Ue()?void 0:Je();return Ve(),He(241===e?w.createBreakStatement(t):w.createContinueStatement(t),n)}function kr(){return 81===ve()?function(){var e=ge();Oe(81);var n=ae(Bt);Oe(58);var t=yn(3,Hr);return He(w.createCaseClause(n,t),e)}():function(){var e=ge();Oe(87),Oe(58);var n=yn(3,Hr);return He(w.createDefaultClause(n),e)}()}function Or(){var e=ge();Oe(106),Oe(20);var n=ae(Bt);Oe(21);var t=function(){var e=ge();Oe(18);var n=yn(2,kr);return Oe(19),He(w.createCaseBlock(n),e)}();return He(w.createSwitchStatement(n,t),e)}function Pr(){var e=ge();Oe(110);var n,t=Cr(!1),r=82===ve()?function(){var e,n=ge();Oe(82),Me(20)?(e=ea(),Oe(21)):e=void 0;var t=Cr(!1);return He(w.createCatchClause(e,t),n)}():void 0;return r&&95!==ve()||(Oe(95),n=Cr(!1)),He(w.createTryStatement(t,r,n),e)}function Mr(){return Ee(),e.tokenIsIdentifierOrKeyword(ve())&&!s.hasPrecedingLineBreak()}function wr(){return Ee(),83===ve()&&!s.hasPrecedingLineBreak()}function Fr(){return Ee(),97===ve()&&!s.hasPrecedingLineBreak()}function Gr(){return Ee(),(e.tokenIsIdentifierOrKeyword(ve())||8===ve()||9===ve()||10===ve())&&!s.hasPrecedingLineBreak()}function Br(){for(;;)switch(ve()){case 112:case 118:case 84:case 97:case 83:case 91:return!0;case 117:case 149:return Ee(),!s.hasPrecedingLineBreak()&&ke();case 139:case 140:return qr();case 125:case 129:case 133:case 120:case 121:case 122:case 142:if(Ee(),s.hasPrecedingLineBreak())return!1;continue;case 154:return Ee(),18===ve()||78===ve()||92===ve();case 99:return Ee(),10===ve()||41===ve()||18===ve()||e.tokenIsIdentifierOrKeyword(ve());case 92:var n=Ee();if(149===n&&(n=Ie(Ee)),62===n||41===n||18===n||87===n||126===n)return!0;continue;case 123:Ee();continue;default:return!1}}function Ur(){return Ie(Br)}function Vr(){switch(ve()){case 59:case 26:case 18:case 112:case 118:case 97:case 83:case 91:case 98:case 89:case 114:case 96:case 85:case 80:case 104:case 115:case 106:case 108:case 110:case 86:case 82:case 95:return!0;case 99:return Ur()||Ie(ct);case 84:case 92:return Ur();case 129:case 133:case 117:case 139:case 140:case 149:case 154:return!0;case 122:case 120:case 121:case 123:case 142:return Ur()||!Ie(Mr);default:return Gt()}}function jr(){return Ee(),ke()||18===ve()||22===ve()}function Hr(){switch(ve()){case 26:return n=ge(),Oe(26),He(w.createEmptyStatement(),n);case 18:return Cr(!1);case 112:return ra(ge(),ye(),void 0,void 0);case 118:if(Ie(jr))return ra(ge(),ye(),void 0,void 0);break;case 97:return aa(ge(),ye(),void 0,void 0);case 83:return va(ge(),ye(),void 0,void 0);case 98:return function(){var e=ge();Oe(98),Oe(20);var n=ae(Bt);Oe(21);var t=Hr(),r=Me(90)?Hr():void 0;return He(w.createIfStatement(n,t,r),e)}();case 89:return function(){var e=ge();Oe(89);var n=Hr();Oe(114),Oe(20);var t=ae(Bt);return Oe(21),Me(26),He(w.createDoStatement(n,t),e)}();case 114:return function(){var e=ge();Oe(114),Oe(20);var n=ae(Bt);Oe(21);var t=Hr();return He(w.createWhileStatement(n,t),e)}();case 96:return Nr();case 85:return Rr(240);case 80:return Rr(241);case 104:return function(){var e=ge();Oe(104);var n=Ue()?void 0:ae(Bt);return Ve(),He(w.createReturnStatement(n),e)}();case 115:return function(){var e=ge();Oe(115),Oe(20);var n=ae(Bt);Oe(21);var t=re(16777216,Hr);return He(w.createWithStatement(n,t),e)}();case 106:return Or();case 108:return function(){var e=ge();Oe(108);var n=s.hasPrecedingLineBreak()?void 0:ae(Bt);return void 0===n&&(R++,n=He(w.createIdentifier(""),ge())),Ve(),He(w.createThrowStatement(n),e)}();case 110:case 82:case 95:return Pr();case 86:return function(){var e=ge();return Oe(86),Ve(),He(w.createDebuggerStatement(),e)}();case 59:return Wr();case 129:case 117:case 149:case 139:case 140:case 133:case 84:case 91:case 92:case 99:case 120:case 121:case 122:case 125:case 123:case 142:case 154:if(Ur())return Wr()}var n;return function(){var n,t=ge(),r=ye(),a=20===ve(),i=ae(Bt);return e.isIdentifier(i)&&Me(58)?n=w.createLabeledStatement(i,Hr()):(Ve(),n=w.createExpressionStatement(i),a&&(r=!1)),H(He(n,t),r)}()}function Kr(e){return 133===e.kind}function Wr(){var n=e.some(Ie((function(){return ma(),_a()})),Kr);if(n){var t=re(8388608,(function(){var e=hn(k);if(e)return bn(e)}));if(t)return t}var r=ge(),a=ye(),i=ma(),o=_a();if(n){for(var s=0,c=o;s<c.length;s++){c[s].flags|=8388608}return re(8388608,(function(){return zr(r,a,i,o)}))}return zr(r,a,i,o)}function zr(n,t,r,a){switch(ve()){case 112:case 118:case 84:return ra(n,t,r,a);case 97:return aa(n,t,r,a);case 83:return va(n,t,r,a);case 117:return function(e,n,t,r){Oe(117);var a=Je(),i=Wn(),o=ba(),s=lt();return H(He(w.createInterfaceDeclaration(t,r,a,i,o,s),e),n)}(n,t,r,a);case 149:return function(e,n,t,r){Oe(149);var a=Je(),i=Wn();Oe(62);var o=136===ve()&&Ne(yt)||Pt();return Ve(),H(He(w.createTypeAliasDeclaration(t,r,a,i,o),e),n)}(n,t,r,a);case 91:return function(e,n,t,r){Oe(91);var a,i=Je();Oe(18)?(a=te(40960,(function(){return Tn(6,La)})),Oe(19)):a=xn();return H(He(w.createEnumDeclaration(t,r,i,a),e),n)}(n,t,r,a);case 154:case 139:case 140:return function(e,n,t,r){var a=0;if(154===ve())return Ca(e,n,t,r);if(Me(140))a|=16;else if(Oe(139),10===ve())return Ca(e,n,t,r);return Da(e,n,t,r,a)}(n,t,r,a);case 99:return function(n,t,r,a){Oe(99);var i,o=s.getStartPos();ke()&&(i=Je());var c,l=!1;153===ve()||"type"!==(null==i?void 0:i.escapedText)||!ke()&&41!==ve()&&18!==ve()||(l=!0,i=ke()?Je():void 0);if(i&&27!==ve()&&153!==ve())return function(n,t,r,a,i,o){Oe(62);var s=143===ve()&&Ie(Ia)?function(){var e=ge();Oe(143),Oe(20);var n=Ra();return Oe(21),He(w.createExternalModuleReference(n),e)}():An(!1);Ve();var c=H(He(w.createImportEqualsDeclaration(r,a,i,s),n),t);o&&fe(c,e.Diagnostics.Only_ECMAScript_imports_may_use_import_type);return c}(n,t,r,a,i,l);(i||41===ve()||18===ve())&&(c=function(e,n,t){var r;e&&!Me(27)||(r=41===ve()?function(){var e=ge();Oe(41),Oe(126);var n=Je();return He(w.createNamespaceImport(n),e)}():ka(264));return He(w.createImportClause(t,e,r),n)}(i,o,l),Oe(153));var u=Ra();return Ve(),H(He(w.createImportDeclaration(r,a,c,u),n),t)}(n,t,r,a);case 92:switch(Ee(),ve()){case 87:case 62:return function(e,n,t,r){var a,i=ue();ne(!0),Me(62)?a=!0:Oe(87);var o=Vt();return Ve(),ne(i),H(He(w.createExportAssignment(t,r,a,o),e),n)}(n,t,r,a);case 126:return function(e,n,t,r){Oe(126),Oe(140);var a=Je();Ve();var i=w.createNamespaceExportDeclaration(a);return i.decorators=t,i.modifiers=r,H(He(i,e),n)}(n,t,r,a);default:return function(e,n,t,r){var a,i,o=ue();ne(!0);var c=Me(149),l=ge();Me(41)?(Me(126)&&(a=function(e){return He(w.createNamespaceExport(Xe()),e)}(l)),Oe(153),i=Ra()):(a=ka(268),(153===ve()||10===ve()&&!s.hasPrecedingLineBreak())&&(Oe(153),i=Ra()));return Ve(),ne(o),H(He(w.createExportDeclaration(t,r,c,a,i),e),n)}(n,t,r,a)}default:if(r||a){var i=Ke(271,!0,e.Diagnostics.Declaration_expected);return e.setTextRangePos(i,n),i.decorators=r,i.modifiers=a,i}return}}function qr(){return Ee(),!s.hasPrecedingLineBreak()&&(ke()||10===ve())}function Jr(e,n){if(18===ve()||!Ue())return Ir(e,n);Ve()}function Xr(){var e=ge();if(27===ve())return He(w.createOmittedExpression(),e);var n=we(25),t=Zr(),r=Ut();return He(w.createBindingElement(n,void 0,t,r),e)}function Yr(){var e,n=ge(),t=we(25),r=Re(),a=Ze();r&&58!==ve()?(e=a,a=void 0):(Oe(58),e=Zr());var i=Ut();return He(w.createBindingElement(t,a,e,i),n)}function Qr(){return 18===ve()||22===ve()||79===ve()||Re()}function Zr(e){return 22===ve()?function(){var e=ge();Oe(22);var n=Tn(10,Xr);return Oe(23),He(w.createArrayBindingPattern(n),e)}():18===ve()?function(){var e=ge();Oe(18);var n=Tn(9,Yr);return Oe(19),He(w.createObjectBindingPattern(n),e)}():qe(e)}function $r(){return ea(!0)}function ea(n){var t,r=ge(),a=Zr(e.Diagnostics.Private_identifiers_are_not_allowed_in_variable_declarations);n&&78===a.kind&&53===ve()&&!s.hasPrecedingLineBreak()&&(t=Be());var i=wt(),o=Xt(ve())?void 0:Ut();return He(w.createVariableDeclaration(a,t,i,o),r)}function na(n){var t,r=ge(),a=0;switch(ve()){case 112:break;case 118:a|=1;break;case 84:a|=2;break;default:e.Debug.fail()}if(Ee(),156===ve()&&Ie(ta))t=xn();else{var i=ce();Z(n),t=Tn(8,n?ea:$r),Z(i)}return He(w.createVariableDeclarationList(t,a),r)}function ta(){return un()&&21===Ee()}function ra(e,n,t,r){var a=na(!1);Ve();var i=w.createVariableStatement(r,a);return i.decorators=t,H(He(i,e),n)}function aa(n,t,r,a){var i=ue(),o=e.modifiersToFlags(a);Oe(97);var s=we(41),c=512&o?Dr():qe(),l=s?1:0,u=256&o?2:0,d=Wn();1&o&&ne(!0);var p=Zn(l|u),m=Yn(58,!1),f=Jr(l|u,e.Diagnostics.or_expected);return ne(i),H(He(w.createFunctionDeclaration(r,a,s,c,d,p,m,f),n),t)}function ia(n,t,r,a){return Ne((function(){if(132===ve()?Oe(132):10===ve()&&20===Ie(Ee)?Ne((function(){var e=Pn();return"constructor"===e.text?e:void 0})):void 0){var i=Wn(),o=Zn(0),s=Yn(58,!1),c=Jr(0,e.Diagnostics.or_expected),l=w.createConstructorDeclaration(r,a,o,c);return l.typeParameters=i,l.type=s,H(He(l,n),t)}}))}function oa(n,t,r,a,i,o,s,c,l){var u=i?1:0,d=e.some(a,e.isAsyncModifier)?2:0,p=Wn(),m=Zn(u|d),f=Yn(58,!1),_=Jr(u|d,l),g=w.createMethodDeclaration(r,a,i,o,s,p,m,f,_);return g.exclamationToken=c,H(He(g,n),t)}function sa(e,n,t,r,a,i){var o=i||s.hasPrecedingLineBreak()?void 0:we(53),c=wt(),l=te(45056,Ut);return Ve(),H(He(w.createPropertyDeclaration(t,r,a,i||o,c,l),e),n)}function ca(n,t,r,a){var i=we(41),o=Ze(),s=we(57);return i||20===ve()||29===ve()?oa(n,t,r,a,i,o,s,void 0,e.Diagnostics.or_expected):sa(n,t,r,a,o,s)}function la(e,n,t,r,a){var i=Ze(),o=Wn(),s=Zn(0),c=Yn(58,!1),l=Jr(0),u=167===a?w.createGetAccessorDeclaration(t,r,i,s,c,l):w.createSetAccessorDeclaration(t,r,i,s,l);return u.typeParameters=o,c&&168===u.kind&&(u.type=c),H(He(u,e),n)}function ua(){var n;if(59===ve())return!0;for(;e.isModifierKind(ve());){if(n=ve(),e.isClassMemberModifier(n))return!0;Ee()}if(41===ve())return!0;if(Ye()&&(n=ve(),Ee()),22===ve())return!0;if(void 0!==n){if(!e.isKeyword(n)||146===n||134===n)return!0;switch(ve()){case 20:case 29:case 53:case 58:case 62:case 57:return!0;default:return Ue()}}return!1}function da(){if(ue()&&130===ve()){var n=ge(),t=Je(e.Diagnostics.Expression_expected);return Ee(),yr(n,fr(n,t,!0))}return tr()}function pa(){var e=ge();if(Me(59)){var n=re(16384,da);return He(w.createDecorator(n),e)}}function ma(){for(var n,t,r=ge();t=pa();)n=e.append(n,t);return n&&je(n,r)}function fa(n){var t=ge(),r=ve();if(84===ve()&&n){if(!Ne(nn))return}else if(!e.isModifierKind(ve())||!Ne(tn))return;return He(w.createToken(r),t)}function _a(n){for(var t,r,a=ge();r=fa(n);)t=e.append(t,r);return t&&je(t,a)}function ga(){var e;if(129===ve()){var n=ge();Ee(),e=je([He(w.createToken(129),n)],n)}return e}function ya(){var n=ge();if(26===ve())return Ee(),He(w.createSemicolonClassElement(),n);var t=ye(),r=ma(),a=_a(!0);if(en(134))return la(n,t,r,a,167);if(en(146))return la(n,t,r,a,168);if(132===ve()||10===ve()){var i=ia(n,t,r,a);if(i)return i}if(nt())return rt(n,t,r,a);if(e.tokenIsIdentifierOrKeyword(ve())||10===ve()||8===ve()||41===ve()||22===ve()){if(e.some(a,Kr)){for(var o=0,s=a;o<s.length;o++){s[o].flags|=8388608}return re(8388608,(function(){return ca(n,t,r,a)}))}return ca(n,t,r,a)}if(r||a){var c=Ke(78,!0,e.Diagnostics.Declaration_expected);return sa(n,t,r,a,c,void 0)}return e.Debug.fail("Should not have attempted to parse class member declaration.")}function va(e,n,t,r){return ha(e,n,t,r,252)}function ha(n,t,r,a,i){var o=ue();Oe(83);var s=!Re()||116===ve()&&Ie(dn)?void 0:ze(Re()),c=Wn();e.some(a,e.isExportModifier)&&ne(!0);var l,u=ba();return Oe(18)?(l=yn(5,ya),Oe(19)):l=xn(),ne(o),H(He(252===i?w.createClassDeclaration(r,a,s,c,u,l):w.createClassExpression(r,a,s,c,u,l),n),t)}function ba(){if(xa())return yn(22,Ea)}function Ea(){var n=ge(),t=ve();e.Debug.assert(93===t||116===t),Ee();var r=Tn(7,Ta);return He(w.createHeritageClause(t,r),n)}function Ta(){var e=ge(),n=tr(),t=Sa();return He(w.createExpressionWithTypeArguments(n,t),e)}function Sa(){return 29===ve()?Ln(20,Pt,29,31):void 0}function xa(){return 93===ve()||116===ve()}function La(){var e=ge(),n=ye(),t=Ze(),r=ae(Ut);return H(He(w.createEnumMember(t,r),e),n)}function Aa(){var e,n=ge();return Oe(18)?(e=yn(1,Hr),Oe(19)):e=xn(),He(w.createModuleBlock(e),n)}function Da(e,n,t,r,a){var i=16&a,o=Je(),s=Me(24)?Da(ge(),!1,void 0,void 0,4|i):Aa();return H(He(w.createModuleDeclaration(t,r,o,s,a),e),n)}function Ca(e,n,t,r){var a,i,o=0;return 154===ve()?(a=Je(),o|=1024):(a=Pn()).text=We(a.text),18===ve()?i=Aa():Ve(),H(He(w.createModuleDeclaration(t,r,a,i,o),e),n)}function Ia(){return 20===Ee()}function Na(){return 43===Ee()}function Ra(){if(10===ve()){var e=Pn();return e.text=We(e.text),e}return Bt()}function ka(e){var n=ge();return He(264===e?w.createNamedImports(Ln(23,Pa,18,19)):w.createNamedExports(Ln(23,Oa,18,19)),n)}function Oa(){return Ma(270)}function Pa(){return Ma(265)}function Ma(n){var t,r,a=ge(),i=e.isKeyword(ve())&&!ke(),o=s.getTokenPos(),c=s.getTextPos(),l=Xe();return 126===ve()?(t=l,Oe(126),i=e.isKeyword(ve())&&!ke(),o=s.getTokenPos(),c=s.getTextPos(),r=Xe()):r=l,265===n&&i&&me(o,c,e.Diagnostics.Identifier_expected),He(265===n?w.createImportSpecifier(t,r):w.createExportSpecifier(t,r),a)}function wa(n){return function(n,t){return e.some(n.modifiers,(function(e){return e.kind===t}))}(n,92)||e.isImportEqualsDeclaration(n)&&e.isExternalModuleReference(n.moduleReference)||e.isImportDeclaration(n)||e.isExportAssignment(n)||e.isExportDeclaration(n)?n:void 0}function Fa(n){return function(n){return e.isMetaProperty(n)&&99===n.keywordToken&&"meta"===n.name.escapedText}(n)?n:m(n,Fa)}n.fixupParentReferences=X,function(e){e[e.SourceElements=0]="SourceElements",e[e.BlockStatements=1]="BlockStatements",e[e.SwitchClauses=2]="SwitchClauses",e[e.SwitchClauseStatements=3]="SwitchClauseStatements",e[e.TypeMembers=4]="TypeMembers",e[e.ClassMembers=5]="ClassMembers",e[e.EnumMembers=6]="EnumMembers",e[e.HeritageClauseElement=7]="HeritageClauseElement",e[e.VariableDeclarations=8]="VariableDeclarations",e[e.ObjectBindingElements=9]="ObjectBindingElements",e[e.ArrayBindingElements=10]="ArrayBindingElements",e[e.ArgumentExpressions=11]="ArgumentExpressions",e[e.ObjectLiteralMembers=12]="ObjectLiteralMembers",e[e.JsxAttributes=13]="JsxAttributes",e[e.JsxChildren=14]="JsxChildren",e[e.ArrayLiteralMembers=15]="ArrayLiteralMembers",e[e.Parameters=16]="Parameters",e[e.JSDocParameters=17]="JSDocParameters",e[e.RestProperties=18]="RestProperties",e[e.TypeParameters=19]="TypeParameters",e[e.TypeArguments=20]="TypeArguments",e[e.TupleElementTypes=21]="TupleElementTypes",e[e.HeritageClauses=22]="HeritageClauses",e[e.ImportOrExportSpecifiers=23]="ImportOrExportSpecifiers",e[e.Count=24]="Count"}(K||(K={})),function(e){e[e.False=0]="False",e[e.True=1]="True",e[e.Unknown=2]="Unknown"}(W||(W={})),function(n){function t(e){var n=ge(),t=(e?Me:Oe)(18),r=re(4194304,Hn);e&&!t||Pe(19);var a=w.createJSDocTypeExpression(r);return X(a),He(a,n)}function r(){var e=ge(),n=Me(18),t=An(!1);n&&Pe(19);var r=w.createJSDocNameReference(t);return X(r),He(r,e)}var a,i;function o(n,a){void 0===n&&(n=0);var i=h,o=void 0===a?i.length:n+a;if(a=o-n,e.Debug.assert(n>=0),e.Debug.assert(n<=o),e.Debug.assert(o<=i.length),p(i,n)){var c,l,u,m=[];return s.scanRange(n+3,a-5,(function(){var e,t,r,a=1,d=n-(i.lastIndexOf("\n",n)+1)+4;function p(n){e||(e=d),m.push(n),d+=n.length}for(Te();U(5););U(4)&&(a=0,d=0);e:for(;;){switch(ve()){case 59:0===a||1===a?(_(m),x(b(d)),a=0,e=void 0):p(s.getTokenText());break;case 4:m.push(s.getTokenText()),a=0,d=0;break;case 41:var g=s.getTokenText();1===a||2===a?(a=2,p(g)):(a=1,d+=g.length);break;case 5:var y=s.getTokenText();2===a?m.push(y):void 0!==e&&d+y.length>e&&m.push(y.slice(e-d)),d+=y.length;break;case 1:break e;default:a=2,p(s.getTokenText())}Te()}return f(m),_(m),t=m.length?m.join(""):void 0,r=c&&je(c,l,u),He(w.createJSDocComment(t,r),n,o)}))}function f(e){for(;e.length&&("\n"===e[0]||"\r"===e[0]);)e.shift()}function _(e){for(;e.length&&""===e[e.length-1].trim();)e.pop()}function g(){for(;;){if(Te(),1===ve())return!0;if(5!==ve()&&4!==ve())return!1}}function y(){if(5!==ve()&&4!==ve()||!Ie(g))for(;5===ve()||4===ve();)Te()}function v(){if((5===ve()||4===ve())&&Ie(g))return"";for(var e=s.hasPrecedingLineBreak(),n=!1,t="";e&&41===ve()||5===ve()||4===ve();)t+=s.getTokenText(),4===ve()?(e=!0,n=!0,t=""):41===ve()&&(e=!1),Te();return n?t:""}function b(n){e.Debug.assert(59===ve());var a=s.getTokenPos();Te();var i,o=V(void 0),l=v();switch(o.escapedText){case"author":i=function(e,n,t,r){var a=Ne((function(){return function(){var e=[],n=!1,t=!1,r=s.getToken();e:for(;;){switch(r){case 78:case 5:case 24:case 59:e.push(s.getTokenText());break;case 29:if(n||t)return;n=!0,e.push(s.getTokenText());break;case 31:if(!n||t)return;t=!0,e.push(s.getTokenText()),s.setTextPos(s.getTokenPos()+1);break e;case 4:case 1:break e}r=Te()}if(n&&t)return 0===e.length?void 0:e.join("")}()}));if(!a){var i=ge();return He(w.createJSDocAuthorTag(n,E(e,i,t,r)),e,i)}var o=a;if(Ie((function(){return 4!==Ee()}))){var c=T(t);c&&(o+=c)}return He(w.createJSDocAuthorTag(n,o),e)}(a,o,n,l);break;case"implements":i=function(e,n,t,r){var a=N(),i=ge();return He(w.createJSDocImplementsTag(n,a,E(e,i,t,r)),e,i)}(a,o,n,l);break;case"augments":case"extends":i=function(e,n,t,r){var a=N(),i=ge();return He(w.createJSDocAugmentsTag(n,a,E(e,i,t,r)),e,i)}(a,o,n,l);break;case"class":case"constructor":i=k(a,w.createJSDocClassTag,o,n,l);break;case"public":i=k(a,w.createJSDocPublicTag,o,n,l);break;case"private":i=k(a,w.createJSDocPrivateTag,o,n,l);break;case"protected":i=k(a,w.createJSDocProtectedTag,o,n,l);break;case"readonly":i=k(a,w.createJSDocReadonlyTag,o,n,l);break;case"deprecated":q=!0,i=k(a,w.createJSDocDeprecatedTag,o,n,l);break;case"this":i=function(e,n,r,a){var i=t(!0);y();var o=ge();return He(w.createJSDocThisTag(n,i,E(e,o,r,a)),e,o)}(a,o,n,l);break;case"enum":i=function(e,n,r,a){var i=t(!0);y();var o=ge();return He(w.createJSDocEnumTag(n,i,E(e,o,r,a)),e,o)}(a,o,n,l);break;case"arg":case"argument":case"param":return C(a,o,2,n);case"return":case"returns":i=function(n,t,r,a){e.some(c,e.isJSDocReturnTag)&&me(t.pos,s.getTokenPos(),e.Diagnostics._0_tag_already_specified,t.escapedText);var i=L(),o=ge();return He(w.createJSDocReturnTag(t,i,E(n,o,r,a)),n,o)}(a,o,n,l);break;case"template":i=function(e,n,r,a){var i=18===ve()?t():void 0,o=function(){var e=ge(),n=[];do{y(),n.push(B()),v()}while(U(27));return je(n,e)}(),s=ge();return He(w.createJSDocTemplateTag(n,i,o,E(e,s,r,a)),e,s)}(a,o,n,l);break;case"type":i=I(a,o,n,l);break;case"typedef":i=function(n,t,r,a){var i,o=L();v();var s=O();y();var c,l=T(r);if(!o||D(o.type)){for(var u=void 0,p=void 0,m=void 0,f=!1;u=Ne((function(){return M(r)}));)if(f=!0,329===u.kind){if(p){de(e.Diagnostics.A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags);var _=e.lastOrUndefined(S);_&&e.addRelatedInfo(_,e.createDetachedDiagnostic(d,0,0,e.Diagnostics.The_tag_was_first_specified_here));break}p=u}else m=e.append(m,u);if(f){var g=o&&178===o.type.kind,h=w.createJSDocTypeLiteral(m,g);c=(o=p&&p.typeExpression&&!D(p.typeExpression.type)?p.typeExpression:He(h,n)).end}}c=c||void 0!==l?ge():(null!==(i=null!=s?s:o)&&void 0!==i?i:t).end,l||(l=E(n,c,r,a));return He(w.createJSDocTypedefTag(t,o,s,l),n,c)}(a,o,n,l);break;case"callback":i=function(n,t,r,a){var i=O();y();var o=T(r),s=function(n){var t,r,a=ge();for(;t=Ne((function(){return F(4,n)}));)r=e.append(r,t);return je(r||[],a)}(r),c=Ne((function(){if(U(59)){var e=b(r);if(e&&327===e.kind)return e}})),l=He(w.createJSDocSignature(void 0,s,c),n),u=ge();o||(o=E(n,u,r,a));return He(w.createJSDocCallbackTag(t,l,i,o),n,u)}(a,o,n,l);break;case"see":i=function(e,n,t,a){var i=r(),o=ge(),s=void 0!==t&&void 0!==a?E(e,o,t,a):void 0;return He(w.createJSDocSeeTag(n,i,s),e,o)}(a,o,n,l);break;default:i=function(e,n,t,r){var a=ge();return He(w.createJSDocUnknownTag(n,E(e,a,t,r)),e,a)}(a,o,n,l)}return i}function E(e,n,t,r){return r||(t+=n-e),T(t,r.slice(t))}function T(n,t){var r,a=[],i=0;function o(e){r||(r=n),a.push(e),n+=e.length}void 0!==t&&(""!==t&&o(t),i=1);var c=ve();e:for(;;){switch(c){case 4:i=0,a.push(s.getTokenText()),n=0;break;case 59:if(3===i){a.push(s.getTokenText());break}s.setTextPos(s.getTextPos()-1);case 1:break e;case 5:if(2===i||3===i)o(s.getTokenText());else{var l=s.getTokenText();void 0!==r&&n+l.length>r&&a.push(l.slice(r-n)),n+=l.length}break;case 18:i=2,Ie((function(){return 59===Te()&&e.tokenIsIdentifierOrKeyword(Te())&&"link"===s.getTokenText()}))&&(o(s.getTokenText()),Te(),o(s.getTokenText()),Te()),o(s.getTokenText());break;case 61:i=3===i?2:3,o(s.getTokenText());break;case 41:if(0===i){i=1,n+=1;break}default:3!==i&&(i=2),o(s.getTokenText())}c=Te()}return f(a),_(a),0===a.length?void 0:a.join("")}function x(e){e&&(c?c.push(e):(c=[e],l=e.pos),u=e.end)}function L(){return v(),18===ve()?t():void 0}function A(){var n=U(22);n&&y();var t,r=U(61),a=function(){var e=V();Me(22)&&Oe(23);for(;Me(24);){var n=V();Me(22)&&Oe(23),e=Dn(e,n)}return e}();return r&&(Fe(t=61)||Ke(t,!1,e.Diagnostics._0_expected,e.tokenToString(t))),n&&(y(),we(62)&&Bt(),Oe(23)),{name:a,isBracketed:n}}function D(n){switch(n.kind){case 145:return!0;case 178:return D(n.elementType);default:return e.isTypeReferenceNode(n)&&e.isIdentifier(n.typeName)&&"Object"===n.typeName.escapedText&&!n.typeArguments}}function C(n,t,r,a){var i=L(),o=!i;v();var c=A(),l=c.name,u=c.isBracketed;y(),o&&(i=L());var d=T(a+s.getStartPos()-n),p=4!==r&&function(n,t,r,a){if(n&&D(n.type)){for(var i=ge(),o=void 0,s=void 0;o=Ne((function(){return F(r,a,t)}));)326!==o.kind&&333!==o.kind||(s=e.append(s,o));if(s){var c=He(w.createJSDocTypeLiteral(s,178===n.type.kind),i);return He(w.createJSDocTypeExpression(c),i)}}}(i,l,r,a);return p&&(i=p,o=!0),He(1===r?w.createJSDocPropertyTag(t,l,u,i,o,d):w.createJSDocParameterTag(t,l,u,i,o,d),n)}function I(n,r,a,i){e.some(c,e.isJSDocTypeTag)&&me(r.pos,s.getTokenPos(),e.Diagnostics._0_tag_already_specified,r.escapedText);var o=t(!0),l=ge(),u=void 0!==a&&void 0!==i?E(n,l,a,i):void 0;return He(w.createJSDocTypeTag(r,o,u),n,l)}function N(){var e=Me(18),n=ge(),t=function(){var e=ge(),n=V();for(;Me(24);){var t=V();n=He(w.createPropertyAccessExpression(n,t),e)}return n}(),r=Sa(),a=He(w.createExpressionWithTypeArguments(t,r),n);return e&&Oe(19),a}function k(e,n,t,r,a){var i=ge();return He(n(t,E(e,i,r,a)),e,i)}function O(n){var t=s.getTokenPos();if(e.tokenIsIdentifierOrKeyword(ve())){var r=V();if(Me(24)){var a=O(!0);return He(w.createModuleDeclaration(void 0,void 0,r,a,n?4:void 0),t)}return n&&(r.isInJSDocNamespace=!0),r}}function P(n,t){for(;!e.isIdentifier(n)||!e.isIdentifier(t);){if(e.isIdentifier(n)||e.isIdentifier(t)||n.right.escapedText!==t.right.escapedText)return!1;n=n.left,t=t.left}return n.escapedText===t.escapedText}function M(e){return F(1,e)}function F(n,t,r){for(var a=!0,i=!1;;)switch(Te()){case 59:if(a){var o=G(n,t);return!(o&&(326===o.kind||333===o.kind)&&4!==n&&r&&(e.isIdentifier(o.name)||!P(r,o.name.left)))&&o}i=!1;break;case 4:a=!0,i=!1;break;case 41:i&&(a=!1),i=!0;break;case 78:a=!1;break;case 1:return!1}}function G(n,t){e.Debug.assert(59===ve());var r=s.getStartPos();Te();var a,i=V();switch(y(),i.escapedText){case"type":return 1===n&&I(r,i);case"prop":case"property":a=1;break;case"arg":case"argument":case"param":a=6;break;default:return!1}return!!(n&a)&&C(r,i,n,t)}function B(){var n=ge(),t=V(e.Diagnostics.Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces);return He(w.createTypeParameterDeclaration(t,void 0,void 0),n)}function U(e){return ve()===e&&(Te(),!0)}function V(n){if(!e.tokenIsIdentifierOrKeyword(ve()))return Ke(78,!n,n||e.Diagnostics.Identifier_expected);R++;var t=s.getTokenPos(),r=s.getTextPos(),a=ve(),i=We(s.getTokenValue()),o=He(w.createIdentifier(i,void 0,a),t,r);return Te(),o}}n.parseJSDocTypeExpressionForTests=function(n,r,a){U("file.js",n,99,void 0,1),s.setText(n,r,a),D=s.scan();var i=t(),o=Y("file.js",99,1,!1,[],w.createToken(1),0),c=e.attachFileToDiagnostics(S,o);return L&&(o.jsDocDiagnostics=e.attachFileToDiagnostics(L,o)),V(),i?{jsDocTypeExpression:i,diagnostics:c}:void 0},n.parseJSDocTypeExpression=t,n.parseJSDocNameReference=r,n.parseIsolatedJSDocComment=function(n,t,r){U("",n,99,void 0,1);var a=re(4194304,(function(){return o(t,r)})),i={languageVariant:0,text:n},s=e.attachFileToDiagnostics(S,i);return V(),a?{jsDoc:a,diagnostics:s}:void 0},n.parseJSDocComment=function(n,t,r){var a=D,i=S.length,s=G,c=re(4194304,(function(){return o(t,r)}));return e.setParent(c,n),131072&P&&(L||(L=[]),L.push.apply(L,S)),D=a,S.length=i,G=s,c},function(e){e[e.BeginningOfLine=0]="BeginningOfLine",e[e.SawAsterisk=1]="SawAsterisk",e[e.SavingComments=2]="SavingComments",e[e.SavingBackticks=3]="SavingBackticks"}(a||(a={})),function(e){e[e.Property=1]="Property",e[e.Parameter=2]="Parameter",e[e.CallbackParameter=4]="CallbackParameter"}(i||(i={}))}(z=n.JSDocParser||(n.JSDocParser={}))}(c||(c={})),function(n){function t(n,t,a,o,s,c){return void(t?u(n):l(n));function l(n){var t="";if(c&&r(n)&&(t=o.substring(n.pos,n.end)),n._children&&(n._children=void 0),e.setTextRangePosEnd(n,n.pos+a,n.end+a),c&&r(n)&&e.Debug.assert(t===s.substring(n.pos,n.end)),m(n,l,u),e.hasJSDocNodes(n))for(var d=0,p=n.jsDoc;d<p.length;d++){l(p[d])}i(n,c)}function u(n){n._children=void 0,e.setTextRangePosEnd(n,n.pos+a,n.end+a);for(var t=0,r=n;t<r.length;t++){l(r[t])}}}function r(e){switch(e.kind){case 10:case 8:case 78:return!0}return!1}function a(n,t,r,a,i){e.Debug.assert(n.end>=t,"Adjusting an element that was entirely before the change range"),e.Debug.assert(n.pos<=r,"Adjusting an element that was entirely after the change range"),e.Debug.assert(n.pos<=n.end);var o=Math.min(n.pos,a),s=n.end>=r?n.end+i:Math.min(n.end,a);e.Debug.assert(o<=s),n.parent&&(e.Debug.assertGreaterThanOrEqual(o,n.parent.pos),e.Debug.assertLessThanOrEqual(s,n.parent.end)),e.setTextRangePosEnd(n,o,s)}function i(n,t){if(t){var r=n.pos,a=function(n){e.Debug.assert(n.pos>=r),r=n.end};if(e.hasJSDocNodes(n))for(var i=0,o=n.jsDoc;i<o.length;i++){a(o[i])}m(n,a),e.Debug.assert(r<=n.end)}}function o(n,t){var r,a=n;if(m(n,(function n(i){if(e.nodeIsMissing(i))return;if(!(i.pos<=t))return e.Debug.assert(i.pos>t),!0;if(i.pos>=a.pos&&(a=i),t<i.end)return m(i,n),!0;e.Debug.assert(i.end<=t),r=i})),r){var i=function(n){for(;;){var t=e.getLastChild(n);if(!t)return n;n=t}}(r);i.pos>a.pos&&(a=i)}return a}function s(n,t,r,a){var i=n.text;if(r&&(e.Debug.assert(i.length-r.span.length+r.newLength===t.length),a||e.Debug.shouldAssert(3))){var o=i.substr(0,r.span.start),s=t.substr(0,r.span.start);e.Debug.assert(o===s);var c=i.substring(e.textSpanEnd(r.span),i.length),l=t.substring(e.textSpanEnd(e.textChangeRangeNewSpan(r)),t.length);e.Debug.assert(c===l)}}function l(n){var t=n.statements,r=0;e.Debug.assert(r<t.length);var a=t[r],i=-1;return{currentNode:function(o){return o!==i&&(a&&a.end===o&&r<t.length-1&&(r++,a=t[r]),a&&a.pos===o||function(e){return t=void 0,r=-1,a=void 0,void m(n,i,o);function i(n){return e>=n.pos&&e<n.end&&(m(n,i,o),!0)}function o(n){if(e>=n.pos&&e<n.end)for(var s=0;s<n.length;s++){var c=n[s];if(c){if(c.pos===e)return t=n,r=s,a=c,!0;if(c.pos<e&&e<c.end)return m(c,i,o),!0}}return!1}}(o)),i=o,e.Debug.assert(!a||a.pos===o),a}}}var u;n.updateSourceFile=function(n,r,u,d){if(s(n,r,u,d=d||e.Debug.shouldAssert(2)),e.textChangeRangeIsUnchanged(u))return n;if(0===n.statements.length)return c.parseSourceFile(n.fileName,r,n.languageVersion,void 0,!0,n.scriptKind);var p=n;e.Debug.assert(!p.hasBeenIncrementallyParsed),p.hasBeenIncrementallyParsed=!0,c.fixupParentReferences(p);var f=n.text,_=l(n),g=function(n,t){for(var r=1,a=t.span.start,i=0;a>0&&i<=r;i++){var s=o(n,a);e.Debug.assert(s.pos<=a);var c=s.pos;a=Math.max(0,c-1)}var l=e.createTextSpanFromBounds(a,e.textSpanEnd(t.span)),u=t.newLength+(t.span.start-a);return e.createTextChangeRange(l,u)}(n,u);s(n,r,g,d),e.Debug.assert(g.span.start<=u.span.start),e.Debug.assert(e.textSpanEnd(g.span)===e.textSpanEnd(u.span)),e.Debug.assert(e.textSpanEnd(e.textChangeRangeNewSpan(g))===e.textSpanEnd(e.textChangeRangeNewSpan(u)));var y=e.textChangeRangeNewSpan(g).length-g.span.length;!function(n,r,o,s,c,l,u,d){return void p(n);function p(n){if(e.Debug.assert(n.pos<=n.end),n.pos>o)t(n,!1,c,l,u,d);else{var _=n.end;if(_>=r){if(n.intersectsChange=!0,n._children=void 0,a(n,r,o,s,c),m(n,p,f),e.hasJSDocNodes(n))for(var g=0,y=n.jsDoc;g<y.length;g++){p(y[g])}i(n,d)}else e.Debug.assert(_<r)}}function f(n){if(e.Debug.assert(n.pos<=n.end),n.pos>o)t(n,!0,c,l,u,d);else{var i=n.end;if(i>=r){n.intersectsChange=!0,n._children=void 0,a(n,r,o,s,c);for(var m=0,f=n;m<f.length;m++){p(f[m])}}else e.Debug.assert(i<r)}}}(p,g.span.start,e.textSpanEnd(g.span),e.textSpanEnd(e.textChangeRangeNewSpan(g)),y,f,r,d);var v=c.parseSourceFile(n.fileName,r,n.languageVersion,_,!0,n.scriptKind);return v.commentDirectives=function(n,t,r,a,i,o,s,c){if(!n)return t;for(var l,u=!1,d=0,p=n;d<p.length;d++){var m=p[d],f=m.range,_=m.type;if(f.end<r)l=e.append(l,m);else if(f.pos>a){y();var g={range:{pos:f.pos+i,end:f.end+i},type:_};l=e.append(l,g),c&&e.Debug.assert(o.substring(f.pos,f.end)===s.substring(g.range.pos,g.range.end))}}return y(),l;function y(){u||(u=!0,l?t&&l.push.apply(l,t):l=t)}}(n.commentDirectives,v.commentDirectives,g.span.start,e.textSpanEnd(g.span),y,f,r,d),v},n.createSyntaxCursor=l,function(e){e[e.Value=-1]="Value"}(u||(u={}))}(l||(l={})),e.isDeclarationFileName=_,e.processCommentPragmas=g,e.processPragmasIntoFields=y;var v=new e.Map;function h(e){if(v.has(e))return v.get(e);var n=new RegExp("(\\s"+e+"\\s*=\\s*)('|\")(.+?)\\2","im");return v.set(e,n),n}var b=/^\/\/\/\s*<(\S+)\s.*?\/>/im,E=/^\/\/\/?\s*@(\S+)\s*(.*)\s*$/im;function T(n,t,r){var a=2===t.kind&&b.exec(r);if(a){var i=a[1].toLowerCase(),o=e.commentPragmas[i];if(!(o&&1&o.kind))return;if(o.args){for(var s={},c=0,l=o.args;c<l.length;c++){var u=l[c],d=h(u.name).exec(r);if(!d&&!u.optional)return;if(d)if(u.captureSpan){var p=t.pos+d.index+d[1].length+d[2].length;s[u.name]={value:d[3],pos:p,end:p+d[3].length}}else s[u.name]=d[3]}n.push({name:i,args:{arguments:s,range:t}})}else n.push({name:i,args:{arguments:{},range:t}})}else{var m=2===t.kind&&E.exec(r);if(m)return S(n,t,2,m);if(3===t.kind)for(var f=/\s*@(\S+)\s*(.*)\s*$/gim,_=void 0;_=f.exec(r);)S(n,t,4,_)}}function S(n,t,r,a){if(a){var i=a[1].toLowerCase(),o=e.commentPragmas[i];if(o&&o.kind&r){var s=function(n,t){if(!t)return{};if(!n.args)return{};for(var r=t.split(/\s+/),a={},i=0;i<n.args.length;i++){var o=n.args[i];if(!r[i]&&!o.optional)return"fail";if(o.captureSpan)return e.Debug.fail("Capture spans not yet implemented for non-xml pragmas");a[o.name]=r[i]}return a}(o,a[2]);"fail"!==s&&n.push({name:i,args:{arguments:s,range:t}})}}}function x(e,n){return e.kind===n.kind&&(78===e.kind?e.escapedText===n.escapedText:107===e.kind||e.name.escapedText===n.name.escapedText&&x(e.expression,n.expression))}e.tagNamesAreEquivalent=x}(ts||(ts={})),function(e){e.compileOnSaveCommandLineOption={name:"compileOnSave",type:"boolean"};var n=new e.Map(e.getEntries({preserve:1,"react-native":3,react:2,"react-jsx":4,"react-jsxdev":5}));e.inverseJsxOptionMap=new e.Map(e.arrayFrom(e.mapIterator(n.entries(),(function(e){var n=e[0];return[""+e[1],n]}))));var t,r,a=[["es5","lib.es5.d.ts"],["es6","lib.es2015.d.ts"],["es2015","lib.es2015.d.ts"],["es7","lib.es2016.d.ts"],["es2016","lib.es2016.d.ts"],["es2017","lib.es2017.d.ts"],["es2018","lib.es2018.d.ts"],["es2019","lib.es2019.d.ts"],["es2020","lib.es2020.d.ts"],["esnext","lib.esnext.d.ts"],["dom","lib.dom.d.ts"],["dom.iterable","lib.dom.iterable.d.ts"],["webworker","lib.webworker.d.ts"],["webworker.importscripts","lib.webworker.importscripts.d.ts"],["webworker.iterable","lib.webworker.iterable.d.ts"],["scripthost","lib.scripthost.d.ts"],["es2015.core","lib.es2015.core.d.ts"],["es2015.collection","lib.es2015.collection.d.ts"],["es2015.generator","lib.es2015.generator.d.ts"],["es2015.iterable","lib.es2015.iterable.d.ts"],["es2015.promise","lib.es2015.promise.d.ts"],["es2015.proxy","lib.es2015.proxy.d.ts"],["es2015.reflect","lib.es2015.reflect.d.ts"],["es2015.symbol","lib.es2015.symbol.d.ts"],["es2015.symbol.wellknown","lib.es2015.symbol.wellknown.d.ts"],["es2016.array.include","lib.es2016.array.include.d.ts"],["es2017.object","lib.es2017.object.d.ts"],["es2017.sharedmemory","lib.es2017.sharedmemory.d.ts"],["es2017.string","lib.es2017.string.d.ts"],["es2017.intl","lib.es2017.intl.d.ts"],["es2017.typedarrays","lib.es2017.typedarrays.d.ts"],["es2018.asyncgenerator","lib.es2018.asyncgenerator.d.ts"],["es2018.asynciterable","lib.es2018.asynciterable.d.ts"],["es2018.intl","lib.es2018.intl.d.ts"],["es2018.promise","lib.es2018.promise.d.ts"],["es2018.regexp","lib.es2018.regexp.d.ts"],["es2019.array","lib.es2019.array.d.ts"],["es2019.object","lib.es2019.object.d.ts"],["es2019.string","lib.es2019.string.d.ts"],["es2019.symbol","lib.es2019.symbol.d.ts"],["es2020.bigint","lib.es2020.bigint.d.ts"],["es2020.promise","lib.es2020.promise.d.ts"],["es2020.sharedmemory","lib.es2020.sharedmemory.d.ts"],["es2020.string","lib.es2020.string.d.ts"],["es2020.symbol.wellknown","lib.es2020.symbol.wellknown.d.ts"],["es2020.intl","lib.es2020.intl.d.ts"],["esnext.array","lib.es2019.array.d.ts"],["esnext.symbol","lib.es2019.symbol.d.ts"],["esnext.asynciterable","lib.es2018.asynciterable.d.ts"],["esnext.intl","lib.esnext.intl.d.ts"],["esnext.bigint","lib.es2020.bigint.d.ts"],["esnext.string","lib.esnext.string.d.ts"],["esnext.promise","lib.esnext.promise.d.ts"],["esnext.weakref","lib.esnext.weakref.d.ts"]];function i(n){var t=new e.Map,r=new e.Map;return e.forEach(n,(function(e){t.set(e.name.toLowerCase(),e),e.shortName&&r.set(e.shortName,e.name)})),{optionsNameMap:t,shortOptionNames:r}}function o(){return t||(t=i(e.optionDeclarations))}function s(e){return e&&void 0!==e.enableAutoDiscovery&&void 0===e.enable?{enable:e.enableAutoDiscovery,include:e.include||[],exclude:e.exclude||[]}:e}function c(n){return l(n,e.createCompilerDiagnostic)}function l(n,t){var r=e.arrayFrom(n.type.keys()).map((function(e){return"'"+e+"'"})).join(", ");return t(e.Diagnostics.Argument_for_0_option_must_be_Colon_1,"--"+n.name,r)}function u(e,n,t){return ue(e,de(n||""),t)}function d(n,t,r){if(void 0===t&&(t=""),t=de(t),!e.startsWith(t,"-")){if(""===t)return[];var a=t.split(",");switch(n.element.type){case"number":return e.map(a,parseInt);case"string":return e.map(a,(function(e){return e||""}));default:return e.mapDefined(a,(function(e){return u(n.element,e,r)}))}}}function p(e){return e.name}function m(n,t,r,a){var i=e.getSpellingSuggestion(n,t.optionDeclarations,p);return i?r(t.unknownDidYouMeanDiagnostic,a||n,i.name):r(t.unknownOptionDiagnostic,a||n)}function f(n,t,r){var a,i={},o=[],s=[];return c(t),{options:i,watchOptions:a,fileNames:o,errors:s};function c(t){for(var r=0;r<t.length;){var c=t[r];if(r++,64===c.charCodeAt(0))l(c.slice(1));else if(45===c.charCodeAt(0)){var u=c.slice(45===c.charCodeAt(1)?2:1),d=y(n.getOptionsNameMap,u,!0);if(d)r=_(t,r,n,d,i,s);else{var p=y(N.getOptionsNameMap,u,!0);p?r=_(t,r,N,p,a||(a={}),s):s.push(m(u,n,e.createCompilerDiagnostic,c))}}else o.push(c)}}function l(n){var t=E(n,r||function(n){return e.sys.readFile(n)});if(e.isString(t)){for(var a=[],i=0;;){for(;i<t.length&&t.charCodeAt(i)<=32;)i++;if(i>=t.length)break;var o=i;if(34===t.charCodeAt(o)){for(i++;i<t.length&&34!==t.charCodeAt(i);)i++;i<t.length?(a.push(t.substring(o+1,i)),i++):s.push(e.createCompilerDiagnostic(e.Diagnostics.Unterminated_quoted_string_in_response_file_0,n))}else{for(;t.charCodeAt(i)>32;)i++;a.push(t.substring(o,i))}}c(a)}else s.push(t)}}function _(n,t,r,a,i,o){if(a.isTSConfigOnly)"null"===(s=n[t])?(i[a.name]=void 0,t++):"boolean"===a.type?"false"===s?(i[a.name]=!1,t++):("true"===s&&t++,o.push(e.createCompilerDiagnostic(e.Diagnostics.Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line,a.name))):(o.push(e.createCompilerDiagnostic(e.Diagnostics.Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line,a.name)),s&&!e.startsWith(s,"-")&&t++);else if(n[t]||"boolean"===a.type||o.push(e.createCompilerDiagnostic(r.optionTypeMismatchDiagnostic,a.name,w(a))),"null"!==n[t])switch(a.type){case"number":i[a.name]=parseInt(n[t]),t++;break;case"boolean":var s=n[t];i[a.name]="false"!==s,"false"!==s&&"true"!==s||t++;break;case"string":i[a.name]=n[t]||"",t++;break;case"list":var c=d(a,n[t],o);i[a.name]=c||[],c&&t++;break;default:i[a.name]=u(a,n[t],o),t++}else i[a.name]=void 0,t++;return t}function g(e,n){return y(o,e,n)}function y(e,n,t){void 0===t&&(t=!1),n=n.toLowerCase();var r=e(),a=r.optionsNameMap,i=r.shortOptionNames;if(t){var o=i.get(n);void 0!==o&&(n=o)}return a.get(n)}e.libs=a.map((function(e){return e[0]})),e.libMap=new e.Map(a),e.optionsForWatch=[{name:"watchFile",type:new e.Map(e.getEntries({fixedpollinginterval:e.WatchFileKind.FixedPollingInterval,prioritypollinginterval:e.WatchFileKind.PriorityPollingInterval,dynamicprioritypolling:e.WatchFileKind.DynamicPriorityPolling,usefsevents:e.WatchFileKind.UseFsEvents,usefseventsonparentdirectory:e.WatchFileKind.UseFsEventsOnParentDirectory})),category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_strategy_for_watching_file_Colon_FixedPollingInterval_default_PriorityPollingInterval_DynamicPriorityPolling_UseFsEvents_UseFsEventsOnParentDirectory},{name:"watchDirectory",type:new e.Map(e.getEntries({usefsevents:e.WatchDirectoryKind.UseFsEvents,fixedpollinginterval:e.WatchDirectoryKind.FixedPollingInterval,dynamicprioritypolling:e.WatchDirectoryKind.DynamicPriorityPolling})),category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_strategy_for_watching_directory_on_platforms_that_don_t_support_recursive_watching_natively_Colon_UseFsEvents_default_FixedPollingInterval_DynamicPriorityPolling},{name:"fallbackPolling",type:new e.Map(e.getEntries({fixedinterval:e.PollingWatchKind.FixedInterval,priorityinterval:e.PollingWatchKind.PriorityInterval,dynamicpriority:e.PollingWatchKind.DynamicPriority})),category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_strategy_for_creating_a_polling_watch_when_it_fails_to_create_using_file_system_events_Colon_FixedInterval_default_PriorityInterval_DynamicPriority},{name:"synchronousWatchDirectory",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_support_recursive_watching_natively}],e.commonOptionsWithBuild=[{name:"help",shortName:"h",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Print_this_message},{name:"help",shortName:"?",type:"boolean"},{name:"watch",shortName:"w",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Watch_input_files},{name:"preserveWatchOutput",type:"boolean",showInSimplifiedHelpView:!1,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen},{name:"listFiles",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Print_names_of_files_part_of_the_compilation},{name:"listEmittedFiles",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Print_names_of_generated_files_part_of_the_compilation},{name:"pretty",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Stylize_errors_and_messages_using_color_and_context_experimental},{name:"traceResolution",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Enable_tracing_of_the_name_resolution_process},{name:"diagnostics",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Show_diagnostic_information},{name:"extendedDiagnostics",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Show_verbose_diagnostic_information},{name:"generateCpuProfile",type:"string",isFilePath:!0,paramType:e.Diagnostics.FILE_OR_DIRECTORY,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Generates_a_CPU_profile},{name:"generateTrace",type:"string",isFilePath:!0,isCommandLineOnly:!0,paramType:e.Diagnostics.DIRECTORY,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Generates_an_event_trace_and_a_list_of_types},{name:"incremental",shortName:"i",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Enable_incremental_compilation,transpileOptionValue:void 0},{name:"assumeChangesOnlyAffectDirectDependencies",type:"boolean",affectsSemanticDiagnostics:!0,affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Have_recompiles_in_incremental_and_watch_assume_that_changes_within_a_file_will_only_affect_files_directly_depending_on_it},{name:"locale",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_locale_used_when_displaying_messages_to_the_user_e_g_en_us}],e.optionDeclarations=__spreadArrays(e.commonOptionsWithBuild,[{name:"all",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Show_all_compiler_options},{name:"version",shortName:"v",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Print_the_compiler_s_version},{name:"init",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file},{name:"project",shortName:"p",type:"string",isFilePath:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,paramType:e.Diagnostics.FILE_OR_DIRECTORY,description:e.Diagnostics.Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json},{name:"build",type:"boolean",shortName:"b",showInSimplifiedHelpView:!0,category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Build_one_or_more_projects_and_their_dependencies_if_out_of_date},{name:"showConfig",type:"boolean",category:e.Diagnostics.Command_line_Options,isCommandLineOnly:!0,description:e.Diagnostics.Print_the_final_configuration_instead_of_building},{name:"listFilesOnly",type:"boolean",category:e.Diagnostics.Command_line_Options,affectsSemanticDiagnostics:!0,affectsEmit:!0,isCommandLineOnly:!0,description:e.Diagnostics.Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing},{name:"target",shortName:"t",type:new e.Map(e.getEntries({es3:0,es5:1,es6:2,es2015:2,es2016:3,es2017:4,es2018:5,es2019:6,es2020:7,esnext:99})),affectsSourceFile:!0,affectsModuleResolution:!0,affectsEmit:!0,paramType:e.Diagnostics.VERSION,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_ECMAScript_target_version_Colon_ES3_default_ES5_ES2015_ES2016_ES2017_ES2018_ES2019_ES2020_or_ESNEXT},{name:"module",shortName:"m",type:new e.Map(e.getEntries({none:e.ModuleKind.None,commonjs:e.ModuleKind.CommonJS,amd:e.ModuleKind.AMD,system:e.ModuleKind.System,umd:e.ModuleKind.UMD,es6:e.ModuleKind.ES2015,es2015:e.ModuleKind.ES2015,es2020:e.ModuleKind.ES2020,esnext:e.ModuleKind.ESNext})),affectsModuleResolution:!0,affectsEmit:!0,paramType:e.Diagnostics.KIND,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_module_code_generation_Colon_none_commonjs_amd_system_umd_es2015_es2020_or_ESNext},{name:"lib",type:"list",element:{name:"lib",type:e.libMap},affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_library_files_to_be_included_in_the_compilation,transpileOptionValue:void 0},{name:"allowJs",type:"boolean",affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Allow_javascript_files_to_be_compiled},{name:"checkJs",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Report_errors_in_js_files},{name:"jsx",type:n,affectsSourceFile:!0,affectsEmit:!0,affectsModuleResolution:!0,paramType:e.Diagnostics.KIND,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_JSX_code_generation_Colon_preserve_react_native_or_react},{name:"declaration",shortName:"d",type:"boolean",affectsEmit:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_corresponding_d_ts_file,transpileOptionValue:void 0},{name:"declarationMap",type:"boolean",affectsEmit:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_a_sourcemap_for_each_corresponding_d_ts_file,transpileOptionValue:void 0},{name:"emitDeclarationOnly",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Only_emit_d_ts_declaration_files,transpileOptionValue:void 0},{name:"sourceMap",type:"boolean",affectsEmit:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Generates_corresponding_map_file},{name:"outFile",type:"string",affectsEmit:!0,isFilePath:!0,paramType:e.Diagnostics.FILE,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Concatenate_and_emit_output_to_single_file,transpileOptionValue:void 0},{name:"outDir",type:"string",affectsEmit:!0,isFilePath:!0,paramType:e.Diagnostics.DIRECTORY,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Redirect_output_structure_to_the_directory},{name:"rootDir",type:"string",affectsEmit:!0,isFilePath:!0,paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir},{name:"composite",type:"boolean",affectsEmit:!0,isTSConfigOnly:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Enable_project_compilation,transpileOptionValue:void 0},{name:"tsBuildInfoFile",type:"string",affectsEmit:!0,isFilePath:!0,paramType:e.Diagnostics.FILE,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Specify_file_to_store_incremental_compilation_information,transpileOptionValue:void 0},{name:"removeComments",type:"boolean",affectsEmit:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Do_not_emit_comments_to_output},{name:"noEmit",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Do_not_emit_outputs,transpileOptionValue:void 0},{name:"importHelpers",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Import_emit_helpers_from_tslib},{name:"importsNotUsedAsValues",type:new e.Map(e.getEntries({remove:0,preserve:1,error:2})),affectsEmit:!0,affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types},{name:"downlevelIteration",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3},{name:"isolatedModules",type:"boolean",category:e.Diagnostics.Basic_Options,description:e.Diagnostics.Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule,transpileOptionValue:!0},{name:"strict",type:"boolean",showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_all_strict_type_checking_options},{name:"noImplicitAny",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Raise_error_on_expressions_and_declarations_with_an_implied_any_type},{name:"strictNullChecks",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_null_checks},{name:"strictFunctionTypes",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_checking_of_function_types},{name:"strictBindCallApply",type:"boolean",strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_bind_call_and_apply_methods_on_functions},{name:"strictPropertyInitialization",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Enable_strict_checking_of_property_initialization_in_classes},{name:"noImplicitThis",type:"boolean",affectsSemanticDiagnostics:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Raise_error_on_this_expressions_with_an_implied_any_type},{name:"alwaysStrict",type:"boolean",affectsSourceFile:!0,strictFlag:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Strict_Type_Checking_Options,description:e.Diagnostics.Parse_in_strict_mode_and_emit_use_strict_for_each_source_file},{name:"noUnusedLocals",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_on_unused_locals},{name:"noUnusedParameters",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_on_unused_parameters},{name:"noImplicitReturns",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_error_when_not_all_code_paths_in_function_return_a_value},{name:"noFallthroughCasesInSwitch",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Report_errors_for_fallthrough_cases_in_switch_statement},{name:"noUncheckedIndexedAccess",type:"boolean",affectsSemanticDiagnostics:!0,showInSimplifiedHelpView:!1,category:e.Diagnostics.Additional_Checks,description:e.Diagnostics.Include_undefined_in_index_signature_results},{name:"moduleResolution",type:new e.Map(e.getEntries({node:e.ModuleResolutionKind.NodeJs,classic:e.ModuleResolutionKind.Classic})),affectsModuleResolution:!0,paramType:e.Diagnostics.STRATEGY,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6},{name:"baseUrl",type:"string",affectsModuleResolution:!0,isFilePath:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Base_directory_to_resolve_non_absolute_module_names},{name:"paths",type:"object",affectsModuleResolution:!0,isTSConfigOnly:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl,transpileOptionValue:void 0},{name:"rootDirs",type:"list",isTSConfigOnly:!0,element:{name:"rootDirs",type:"string",isFilePath:!0},affectsModuleResolution:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime,transpileOptionValue:void 0},{name:"typeRoots",type:"list",element:{name:"typeRoots",type:"string",isFilePath:!0},affectsModuleResolution:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.List_of_folders_to_include_type_definitions_from},{name:"types",type:"list",element:{name:"types",type:"string"},affectsModuleResolution:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Type_declaration_files_to_be_included_in_compilation,transpileOptionValue:void 0},{name:"allowSyntheticDefaultImports",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typechecking},{name:"esModuleInterop",type:"boolean",affectsSemanticDiagnostics:!0,affectsEmit:!0,showInSimplifiedHelpView:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for_all_imports_Implies_allowSyntheticDefaultImports},{name:"preserveSymlinks",type:"boolean",category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Do_not_resolve_the_real_path_of_symlinks},{name:"allowUmdGlobalAccess",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Module_Resolution_Options,description:e.Diagnostics.Allow_accessing_UMD_globals_from_modules},{name:"sourceRoot",type:"string",affectsEmit:!0,paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations},{name:"mapRoot",type:"string",affectsEmit:!0,paramType:e.Diagnostics.LOCATION,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations},{name:"inlineSourceMap",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file},{name:"inlineSources",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Source_Map_Options,description:e.Diagnostics.Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap_to_be_set},{name:"experimentalDecorators",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Experimental_Options,description:e.Diagnostics.Enables_experimental_support_for_ES7_decorators},{name:"emitDecoratorMetadata",type:"boolean",affectsSemanticDiagnostics:!0,affectsEmit:!0,category:e.Diagnostics.Experimental_Options,description:e.Diagnostics.Enables_experimental_support_for_emitting_type_metadata_for_decorators},{name:"jsxFactory",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h},{name:"jsxFragmentFactory",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_JSX_fragment_factory_function_to_use_when_targeting_react_JSX_emit_with_jsxFactory_compiler_option_is_specified_e_g_Fragment},{name:"jsxImportSource",type:"string",affectsSemanticDiagnostics:!0,affectsEmit:!0,affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_module_specifier_to_be_used_to_import_the_jsx_and_jsxs_factory_functions_from_eg_react},{name:"resolveJsonModule",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Include_modules_imported_with_json_extension},{name:"out",type:"string",affectsEmit:!0,isFilePath:!1,category:e.Diagnostics.Advanced_Options,paramType:e.Diagnostics.FILE,description:e.Diagnostics.Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file,transpileOptionValue:void 0},{name:"reactNamespace",type:"string",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react_JSX_emit},{name:"skipDefaultLibCheck",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files},{name:"charset",type:"string",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_character_set_of_the_input_files},{name:"emitBOM",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files},{name:"newLine",type:new e.Map(e.getEntries({crlf:0,lf:1})),affectsEmit:!0,paramType:e.Diagnostics.NEWLINE,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix},{name:"noErrorTruncation",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_truncate_error_messages},{name:"noLib",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_include_the_default_library_file_lib_d_ts,transpileOptionValue:!0},{name:"noResolve",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files,transpileOptionValue:!0},{name:"stripInternal",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_declarations_for_code_that_has_an_internal_annotation},{name:"disableSizeLimit",type:"boolean",affectsSourceFile:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_size_limitations_on_JavaScript_projects},{name:"disableSourceOfProjectReferenceRedirect",type:"boolean",isTSConfigOnly:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_use_of_source_files_instead_of_declaration_files_from_referenced_projects},{name:"disableSolutionSearching",type:"boolean",isTSConfigOnly:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_solution_searching_for_this_project},{name:"disableReferencedProjectLoad",type:"boolean",isTSConfigOnly:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_loading_referenced_projects},{name:"noImplicitUseStrict",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_use_strict_directives_in_module_output},{name:"noEmitHelpers",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_generate_custom_helper_functions_like_extends_in_compiled_output},{name:"noEmitOnError",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_emit_outputs_if_any_errors_were_reported,transpileOptionValue:void 0},{name:"preserveConstEnums",type:"boolean",affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_erase_const_enum_declarations_in_generated_code},{name:"declarationDir",type:"string",affectsEmit:!0,isFilePath:!0,paramType:e.Diagnostics.DIRECTORY,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Output_directory_for_generated_declaration_files,transpileOptionValue:void 0},{name:"skipLibCheck",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Skip_type_checking_of_declaration_files},{name:"allowUnusedLabels",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_report_errors_on_unused_labels},{name:"allowUnreachableCode",type:"boolean",affectsBindDiagnostics:!0,affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Do_not_report_errors_on_unreachable_code},{name:"suppressExcessPropertyErrors",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Suppress_excess_property_checks_for_object_literals},{name:"suppressImplicitAnyIndexErrors",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures},{name:"forceConsistentCasingInFileNames",type:"boolean",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disallow_inconsistently_cased_references_to_the_same_file},{name:"maxNodeModuleJsDepth",type:"number",affectsModuleResolution:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files},{name:"noStrictGenericChecks",type:"boolean",affectsSemanticDiagnostics:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Disable_strict_checking_of_generic_signatures_in_function_types},{name:"useDefineForClassFields",type:"boolean",affectsSemanticDiagnostics:!0,affectsEmit:!0,category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Emit_class_fields_with_Define_instead_of_Set},{name:"keyofStringsOnly",type:"boolean",category:e.Diagnostics.Advanced_Options,description:e.Diagnostics.Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols},{name:"plugins",type:"list",isTSConfigOnly:!0,element:{name:"plugin",type:"object"},description:e.Diagnostics.List_of_language_service_plugins}]),e.semanticDiagnosticsOptionDeclarations=e.optionDeclarations.filter((function(e){return!!e.affectsSemanticDiagnostics})),e.affectsEmitOptionDeclarations=e.optionDeclarations.filter((function(e){return!!e.affectsEmit})),e.moduleResolutionOptionDeclarations=e.optionDeclarations.filter((function(e){return!!e.affectsModuleResolution})),e.sourceFileAffectingCompilerOptions=e.optionDeclarations.filter((function(e){return!!e.affectsSourceFile||!!e.affectsModuleResolution||!!e.affectsBindDiagnostics})),e.transpileOptionValueCompilerOptions=e.optionDeclarations.filter((function(n){return e.hasProperty(n,"transpileOptionValue")})),e.buildOpts=__spreadArrays(e.commonOptionsWithBuild,[{name:"verbose",shortName:"v",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Enable_verbose_logging,type:"boolean"},{name:"dry",shortName:"d",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Show_what_would_be_built_or_deleted_if_specified_with_clean,type:"boolean"},{name:"force",shortName:"f",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Build_all_projects_including_those_that_appear_to_be_up_to_date,type:"boolean"},{name:"clean",category:e.Diagnostics.Command_line_Options,description:e.Diagnostics.Delete_the_outputs_of_all_projects,type:"boolean"}]),e.typeAcquisitionDeclarations=[{name:"enableAutoDiscovery",type:"boolean"},{name:"enable",type:"boolean"},{name:"include",type:"list",element:{name:"include",type:"string"}},{name:"exclude",type:"list",element:{name:"exclude",type:"string"}},{name:"disableFilenameBasedTypeAcquisition",type:"boolean"}],e.createOptionNameMap=i,e.getOptionsNameMap=o,e.defaultInitCompilerOptions={module:e.ModuleKind.CommonJS,target:1,strict:!0,esModuleInterop:!0,forceConsistentCasingInFileNames:!0,skipLibCheck:!0},e.convertEnableAutoDiscoveryToEnable=s,e.createCompilerDiagnosticForInvalidCustomType=c,e.parseCustomTypeOption=u,e.parseListTypeOption=d,e.parseCommandLineWorker=f,e.compilerOptionsDidYouMeanDiagnostics={getOptionsNameMap:o,optionDeclarations:e.optionDeclarations,unknownOptionDiagnostic:e.Diagnostics.Unknown_compiler_option_0,unknownDidYouMeanDiagnostic:e.Diagnostics.Unknown_compiler_option_0_Did_you_mean_1,optionTypeMismatchDiagnostic:e.Diagnostics.Compiler_option_0_expects_an_argument},e.parseCommandLine=function(n,t){return f(e.compilerOptionsDidYouMeanDiagnostics,n,t)},e.getOptionFromName=g;var v={getOptionsNameMap:function(){return r||(r=i(e.buildOpts))},optionDeclarations:e.buildOpts,unknownOptionDiagnostic:e.Diagnostics.Unknown_build_option_0,unknownDidYouMeanDiagnostic:e.Diagnostics.Unknown_build_option_0_Did_you_mean_1,optionTypeMismatchDiagnostic:e.Diagnostics.Build_option_0_requires_a_value_of_type_1};function h(n,t){var r=e.parseJsonText(n,t);return{config:P(r,r.parseDiagnostics),error:r.parseDiagnostics.length?r.parseDiagnostics[0]:void 0}}function b(n,t){var r=E(n,t);return e.isString(r)?e.parseJsonText(n,r):{fileName:n,parseDiagnostics:[r]}}function E(n,t){var r;try{r=t(n)}catch(t){return e.createCompilerDiagnostic(e.Diagnostics.Cannot_read_file_0_Colon_1,n,t.message)}return void 0===r?e.createCompilerDiagnostic(e.Diagnostics.Cannot_read_file_0,n):r}function T(n){return e.arrayToMap(n,p)}e.parseBuildCommand=function(n){var t=f(v,n),r=t.options,a=t.watchOptions,i=t.fileNames,o=t.errors,s=r;return 0===i.length&&i.push("."),s.clean&&s.force&&o.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","force")),s.clean&&s.verbose&&o.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","verbose")),s.clean&&s.watch&&o.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"clean","watch")),s.watch&&s.dry&&o.push(e.createCompilerDiagnostic(e.Diagnostics.Options_0_and_1_cannot_be_combined,"watch","dry")),{buildOptions:s,watchOptions:a,projects:i,errors:o}},e.getDiagnosticText=function(n){for(var t=[],r=1;r<arguments.length;r++)t[r-1]=arguments[r];var a=e.createCompilerDiagnostic.apply(void 0,arguments);return a.messageText},e.getParsedCommandLineOfConfigFile=function(n,t,r,a,i,o){var s=E(n,(function(e){return r.readFile(e)}));if(e.isString(s)){var c=e.parseJsonText(n,s),l=r.getCurrentDirectory();return c.path=e.toPath(n,l,e.createGetCanonicalFileName(r.useCaseSensitiveFileNames)),c.resolvedPath=c.path,c.originalFileName=c.fileName,W(c,r,e.getNormalizedAbsolutePath(e.getDirectoryPath(n),l),t,e.getNormalizedAbsolutePath(n,l),void 0,o,a,i)}r.onUnRecoverableConfigFileDiagnostic(s)},e.readConfigFile=function(n,t){var r=E(n,t);return e.isString(r)?h(n,r):{config:{},error:r}},e.parseConfigFileTextToJson=h,e.readJsonConfigFile=b,e.tryReadFile=E;var S,x={optionDeclarations:e.typeAcquisitionDeclarations,unknownOptionDiagnostic:e.Diagnostics.Unknown_type_acquisition_option_0,unknownDidYouMeanDiagnostic:e.Diagnostics.Unknown_type_acquisition_option_0_Did_you_mean_1};function L(){return S||(S=i(e.optionsForWatch))}var A,D,C,I,N={getOptionsNameMap:L,optionDeclarations:e.optionsForWatch,unknownOptionDiagnostic:e.Diagnostics.Unknown_watch_option_0,unknownDidYouMeanDiagnostic:e.Diagnostics.Unknown_watch_option_0_Did_you_mean_1,optionTypeMismatchDiagnostic:e.Diagnostics.Watch_option_0_requires_a_value_of_type_1};function R(){return A||(A=T(e.optionDeclarations))}function k(){return D||(D=T(e.optionsForWatch))}function O(){return C||(C=T(e.typeAcquisitionDeclarations))}function P(e,n){return M(e,n,!0,void 0,void 0)}function M(n,t,r,a,i){return n.statements.length?c(n.statements[0].expression,a):r?{}:void 0;function o(e){return a&&a.elementOptions===e}function s(a,s,l,d){for(var p=r?{}:void 0,f=function(a){if(288!==a.kind)return t.push(e.createDiagnosticForNodeInSourceFile(n,a,e.Diagnostics.Property_assignment_expected)),"continue";a.questionToken&&t.push(e.createDiagnosticForNodeInSourceFile(n,a.questionToken,e.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files,"?")),u(a.name)||t.push(e.createDiagnosticForNodeInSourceFile(n,a.name,e.Diagnostics.String_literal_with_double_quotes_expected));var f=e.isComputedNonLiteralName(a.name)?void 0:e.getTextOfPropertyName(a.name),_=f&&e.unescapeLeadingUnderscores(f),g=_&&s?s.get(_):void 0;_&&l&&!g&&(s?t.push(m(_,l,(function(t,r,i){return e.createDiagnosticForNodeInSourceFile(n,a.name,t,r,i)}))):t.push(e.createDiagnosticForNodeInSourceFile(n,a.name,l.unknownOptionDiagnostic,_)));var y=c(a.initializer,g);if(void 0!==_&&(r&&(p[_]=y),i&&(d||o(s)))){var v=F(g,y);d?v&&i.onSetValidOptionKeyValueInParent(d,g,y):o(s)&&(v?i.onSetValidOptionKeyValueInRoot(_,a.name,y,a.initializer):g||i.onSetUnknownOptionKeyValueInRoot(_,a.name,y,a.initializer))}},_=0,g=a.properties;_<g.length;_++){f(g[_])}return p}function c(a,i){switch(a.kind){case 109:return g(i&&"boolean"!==i.type),!0;case 94:return g(i&&"boolean"!==i.type),!1;case 103:return g(i&&"extends"===i.name),null;case 10:u(a)||t.push(e.createDiagnosticForNodeInSourceFile(n,a,e.Diagnostics.String_literal_with_double_quotes_expected)),g(i&&e.isString(i.type)&&"string"!==i.type);var o=a.text;if(i&&!e.isString(i.type)){var d=i;d.type.has(o.toLowerCase())||t.push(l(d,(function(t,r,i){return e.createDiagnosticForNodeInSourceFile(n,a,t,r,i)})))}return o;case 8:return g(i&&"number"!==i.type),Number(a.text);case 214:if(40!==a.operator||8!==a.operand.kind)break;return g(i&&"number"!==i.type),-Number(a.operand.text);case 200:g(i&&"object"!==i.type);var p=a;if(i){var m=i;return s(p,m.elementOptions,m.extraKeyDiagnostics,m.name)}return s(p,void 0,void 0,void 0);case 199:return g(i&&"list"!==i.type),f=a.elements,_=i&&i.element,r?e.filter(f.map((function(e){return c(e,_)})),(function(e){return void 0!==e})):f.forEach((function(e){return c(e,_)}))}var f,_;return void(i?g(!0):t.push(e.createDiagnosticForNodeInSourceFile(n,a,e.Diagnostics.Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal)));function g(r){r&&t.push(e.createDiagnosticForNodeInSourceFile(n,a,e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,i.name,w(i)))}}function u(t){return e.isStringLiteral(t)&&e.isStringDoubleQuoted(t,n)}}function w(n){return"list"===n.type?"Array":e.isString(n.type)?n.type:"string"}function F(n,t){return!!n&&(!!q(t)||("list"===n.type?e.isArray(t):typeof t===(e.isString(n.type)?n.type:"string")))}function G(n){return __assign({},e.arrayFrom(n.entries()).reduce((function(e,n){var t;return __assign(__assign({},e),((t={})[n[0]]=n[1],t))}),{}))}function B(n){if(e.length(n)){if(1!==e.length(n))return n;if("**/*"!==n[0])return n}}function U(e){return"string"===e.type||"number"===e.type||"boolean"===e.type||"object"===e.type?void 0:"list"===e.type?U(e.element):e.type}function V(n,t){return e.forEachEntry(t,(function(e,t){if(e===n)return t}))}function j(e,n){return H(e,o(),n)}function H(n,t,r){var a=t.optionsNameMap,i=new e.Map,o=r&&e.createGetCanonicalFileName(r.useCaseSensitiveFileNames),s=function(t){if(e.hasProperty(n,t)){if(a.has(t)&&a.get(t).category===e.Diagnostics.Command_line_Options)return"continue";var s=n[t],c=a.get(t.toLowerCase());if(c){var l=U(c);l?"list"===c.type?i.set(t,s.map((function(e){return V(e,l)}))):i.set(t,V(s,l)):r&&c.isFilePath?i.set(t,e.getRelativePathFromFile(r.configFilePath,e.getNormalizedAbsolutePath(s,e.getDirectoryPath(r.configFilePath)),o)):i.set(t,s)}}};for(var c in n)s(c);return i}function K(e,n,t){if(e&&!q(n))if("list"===e.type){var r=n;if(e.element.isFilePath&&r.length)return r.map(t)}else if(e.isFilePath)return t(n);return n}function W(e,n,t,r,a,i,o,s,c){return X(void 0,e,n,t,r,c,a,i,o,s)}function z(e,n){n&&Object.defineProperty(e,"configFile",{enumerable:!1,writable:!1,value:n})}function q(e){return null==e}function J(n,t){return e.getDirectoryPath(e.getNormalizedAbsolutePath(n,t))}function X(n,t,r,a,i,o,s,c,l,u){void 0===i&&(i={}),void 0===c&&(c=[]),void 0===l&&(l=[]),e.Debug.assert(void 0===n&&void 0!==t||void 0!==n&&void 0===t);var d,p=[],m=ee(n,t,r,a,s,c,p,u),f=m.raw,_=e.extend(i,m.options||{}),g=o&&m.watchOptions?e.extend(o,m.watchOptions):m.watchOptions||o;_.configFilePath=s&&e.normalizeSlashes(s),z(_,t);var y=function(){var n=S("references",(function(e){return"object"==typeof e}),"object");if(e.isArray(n))for(var i=0,o=n;i<o.length;i++){var u=o[i];"string"!=typeof u.path?x(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"reference.path","string"):(d||(d=[])).push({path:e.getNormalizedAbsolutePath(u.path,a),originalPath:u.path,prepend:u.prepend,circular:u.circular})}var m=E(T("files"));if(m){var g="no-prop"===n||e.isArray(n)&&0===n.length,y=e.hasProperty(f,"extends");if(0===m.length&&g&&!y)if(t){var v=s||"tsconfig.json",h=e.Diagnostics.The_files_list_in_config_file_0_is_empty,b=e.firstDefined(e.getTsConfigPropArray(t,"files"),(function(e){return e.initializer})),L=b?e.createDiagnosticForNodeInSourceFile(t,b,h,v):e.createCompilerDiagnostic(h,v);p.push(L)}else x(e.Diagnostics.The_files_list_in_config_file_0_is_empty,s||"tsconfig.json")}var A=E(T("include")),D=T("exclude"),C=E(D);if("no-prop"===D&&f.compilerOptions){var I=f.compilerOptions.outDir,N=f.compilerOptions.declarationDir;(I||N)&&(C=[I,N].filter((function(e){return!!e})))}void 0===m&&void 0===A&&(A=["**/*"]);var R=function(n,t,r,a,i,o,s,c,l){var u,d;a=e.normalizePath(a),t&&(u=ye(t,s,!1,l,"include"));r&&(d=ye(r,s,!0,l,"exclude"));var p=function(n,t,r,a){var i=e.getRegularExpressionForWildcard(t,r,"exclude"),o=i&&new RegExp(i,a?"":"i"),s={};if(void 0!==n){for(var c=[],l=0,u=n;l<u.length;l++){var d=u[l],p=e.normalizePath(e.combinePaths(r,d));if(!o||!o.test(p)){var m=ve(p,a);if(m){var f=m.key,_=m.flags,g=s[f];(void 0===g||g<_)&&(s[f]=_,1===_&&c.push(f))}}}for(var f in s)if(e.hasProperty(s,f))for(var y=0,v=c;y<v.length;y++){var h=v[y];f!==h&&e.containsPath(h,f,r,!a)&&delete s[f]}}return s}(u,d,a,o.useCaseSensitiveFileNames);return ge({filesSpecs:n,includeSpecs:t,excludeSpecs:r,validatedFilesSpec:e.filter(n,e.isString),validatedIncludeSpecs:u,validatedExcludeSpecs:d,wildcardDirectories:p},a,i,o,c)}(m,A,C,s?J(s,a):a,_,r,p,l,t);Q(R,Z(f),c)&&p.push(Y(R.spec,s));return R}(),v=y.fileNames,h=y.wildcardDirectories,b=y.spec;return{options:_,watchOptions:g,fileNames:v,projectReferences:d,typeAcquisition:m.typeAcquisition||ae(),raw:f,errors:p,wildcardDirectories:h,compileOnSave:!!f.compileOnSave,configFileSpecs:b};function E(n){return e.isArray(n)?n:void 0}function T(n){return S(n,e.isString,"string")}function S(n,r,a){if(e.hasProperty(f,n)&&!q(f[n])){if(e.isArray(f[n])){var i=f[n];return t||e.every(i,r)||p.push(e.createCompilerDiagnostic(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,n,a)),i}return x(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,n,"Array"),"not-array"}return"no-prop"}function x(n,r,a){t||p.push(e.createCompilerDiagnostic(n,r,a))}}function Y(n,t){var r=n.includeSpecs,a=n.excludeSpecs;return e.createCompilerDiagnostic(e.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2,t||"tsconfig.json",JSON.stringify(r||[]),JSON.stringify(a||[]))}function Q(e,n,t){return 0===e.fileNames.length&&n&&(!t||0===t.length)}function Z(n){return!e.hasProperty(n,"files")&&!e.hasProperty(n,"references")}function $(e){return!!e.options}function ee(n,t,r,a,i,o,s,c){var l;a=e.normalizeSlashes(a);var u=e.getNormalizedAbsolutePath(i||"",a);if(o.indexOf(u)>=0)return s.push(e.createCompilerDiagnostic(e.Diagnostics.Circularity_detected_while_resolving_configuration_Colon_0,__spreadArrays(o,[u]).join(" -> "))),{raw:n||P(t,s)};var d=n?function(n,t,r,a,i){e.hasProperty(n,"excludes")&&i.push(e.createCompilerDiagnostic(e.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude));var o,s=re(n.compilerOptions,r,i,a),c=ie(n.typeAcquisition||n.typingOptions,r,i,a),l=function(e,n,t){return oe(k(),e,n,void 0,N,t)}(n.watchOptions,r,i);if(n.compileOnSave=function(n,t,r){if(!e.hasProperty(n,e.compileOnSaveCommandLineOption.name))return!1;var a=se(e.compileOnSaveCommandLineOption,n.compileOnSave,t,r);return"boolean"==typeof a&&a}(n,r,i),n.extends)if(e.isString(n.extends)){var u=a?J(a,r):r;o=ne(n.extends,t,u,i,e.createCompilerDiagnostic)}else i.push(e.createCompilerDiagnostic(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,"extends","string"));return{raw:n,options:s,watchOptions:l,typeAcquisition:c,extendedConfigPath:o}}(n,r,a,i,s):function(n,t,r,a,i){var o,s,c,l,u=te(a),d={onSetValidOptionKeyValueInParent:function(n,t,i){var l;switch(n){case"compilerOptions":l=u;break;case"watchOptions":l=c||(c={});break;case"typeAcquisition":l=o||(o=ae(a));break;case"typingOptions":l=s||(s=ae(a));break;default:e.Debug.fail("Unknown option")}l[t.name]=ce(t,r,i)},onSetValidOptionKeyValueInRoot:function(o,s,c,u){switch(o){case"extends":var d=a?J(a,r):r;return void(l=ne(c,t,d,i,(function(t,r){return e.createDiagnosticForNodeInSourceFile(n,u,t,r)})))}},onSetUnknownOptionKeyValueInRoot:function(t,r,a,o){"excludes"===t&&i.push(e.createDiagnosticForNodeInSourceFile(n,r,e.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude))}},p=M(n,i,!0,(void 0===I&&(I={name:void 0,type:"object",elementOptions:T([{name:"compilerOptions",type:"object",elementOptions:R(),extraKeyDiagnostics:e.compilerOptionsDidYouMeanDiagnostics},{name:"watchOptions",type:"object",elementOptions:k(),extraKeyDiagnostics:N},{name:"typingOptions",type:"object",elementOptions:O(),extraKeyDiagnostics:x},{name:"typeAcquisition",type:"object",elementOptions:O(),extraKeyDiagnostics:x},{name:"extends",type:"string"},{name:"references",type:"list",element:{name:"references",type:"object"}},{name:"files",type:"list",element:{name:"files",type:"string"}},{name:"include",type:"list",element:{name:"include",type:"string"}},{name:"exclude",type:"list",element:{name:"exclude",type:"string"}},e.compileOnSaveCommandLineOption])}),I),d);o||(o=s?void 0!==s.enableAutoDiscovery?{enable:s.enableAutoDiscovery,include:s.include,exclude:s.exclude}:s:ae(a));return{raw:p,options:u,watchOptions:c,typeAcquisition:o,extendedConfigPath:l}}(t,r,a,i,s);if((null===(l=d.options)||void 0===l?void 0:l.paths)&&(d.options.pathsBasePath=a),d.extendedConfigPath){o=o.concat([u]);var p=function(n,t,r,a,i,o,s){var c,l,u,d,p=r.useCaseSensitiveFileNames?t:e.toFileNameLowerCase(t);if(s&&(l=s.get(p)))u=l.extendedResult,d=l.extendedConfig;else{if(!(u=b(t,(function(e){return r.readFile(e)}))).parseDiagnostics.length){var m=e.getDirectoryPath(t);if($(d=ee(void 0,u,r,m,e.getBaseFileName(t),i,o,s))){var f=e.convertToRelativePath(m,a,e.identity),_=function(n){return e.isRootedDiskPath(n)?n:e.combinePaths(f,n)},g=function(n){y[n]&&(y[n]=e.map(y[n],_))},y=d.raw;g("include"),g("exclude"),g("files")}}s&&s.set(p,{extendedResult:u,extendedConfig:d})}n&&(n.extendedSourceFiles=[u.fileName],u.extendedSourceFiles&&(c=n.extendedSourceFiles).push.apply(c,u.extendedSourceFiles));if(u.parseDiagnostics.length)return void o.push.apply(o,u.parseDiagnostics);return d}(t,d.extendedConfigPath,r,a,o,s,c);if(p&&$(p)){var m=p.raw,f=d.raw,_=function(e){var n=f[e]||m[e];n&&(f[e]=n)};_("include"),_("exclude"),_("files"),void 0===f.compileOnSave&&(f.compileOnSave=m.compileOnSave),d.options=e.assign({},p.options,d.options),d.watchOptions=d.watchOptions&&p.watchOptions?e.assign({},p.watchOptions,d.watchOptions):d.watchOptions||p.watchOptions}}return d}function ne(n,t,r,a,i){if(n=e.normalizeSlashes(n),e.isRootedDiskPath(n)||e.startsWith(n,"./")||e.startsWith(n,"../")){var o=e.getNormalizedAbsolutePath(n,r);return t.fileExists(o)||e.endsWith(o,".json")||(o+=".json",t.fileExists(o))?o:void a.push(i(e.Diagnostics.File_0_not_found,n))}var s=e.nodeModuleNameResolver(n,e.combinePaths(r,"tsconfig.json"),{moduleResolution:e.ModuleResolutionKind.NodeJs},t,void 0,void 0,!0);if(s.resolvedModule)return s.resolvedModule.resolvedFileName;a.push(i(e.Diagnostics.File_0_not_found,n))}function te(n){return n&&"jsconfig.json"===e.getBaseFileName(n)?{allowJs:!0,maxNodeModuleJsDepth:2,allowSyntheticDefaultImports:!0,skipLibCheck:!0,noEmit:!0}:{}}function re(n,t,r,a){var i=te(a);return oe(R(),n,t,i,e.compilerOptionsDidYouMeanDiagnostics,r),a&&(i.configFilePath=e.normalizeSlashes(a)),i}function ae(n){return{enable:!!n&&"jsconfig.json"===e.getBaseFileName(n),include:[],exclude:[]}}function ie(e,n,t,r){var a=ae(r),i=s(e);return oe(O(),i,n,a,x,t),a}function oe(n,t,r,a,i,o){if(t){for(var s in t){var c=n.get(s);c?(a||(a={}))[c.name]=se(c,t[s],r,o):o.push(m(s,i,e.createCompilerDiagnostic))}return a}}function se(n,t,r,a){if(F(n,t)){var i=n.type;return"list"===i&&e.isArray(t)?function(n,t,r,a){return e.filter(e.map(t,(function(e){return se(n.element,e,r,a)})),(function(e){return!!e}))}(n,t,r,a):e.isString(i)?le(n,r,t):ue(n,t,a)}a.push(e.createCompilerDiagnostic(e.Diagnostics.Compiler_option_0_requires_a_value_of_type_1,n.name,w(n)))}function ce(n,t,r){if(!q(r)){if("list"===n.type){var a=n;return a.element.isFilePath||!e.isString(a.element.type)?e.filter(e.map(r,(function(e){return ce(a.element,t,e)})),(function(e){return!!e})):r}return e.isString(n.type)?le(n,t,r):n.type.get(e.isString(r)?r.toLowerCase():r)}}function le(n,t,r){return n.isFilePath&&""===(r=e.getNormalizedAbsolutePath(r,t))&&(r="."),r}function ue(e,n,t){if(!q(n)){var r=n.toLowerCase(),a=e.type.get(r);if(void 0!==a)return a;t.push(c(e))}}function de(e){return"function"==typeof e.trim?e.trim():e.replace(/^[\s]+|[\s]+$/g,"")}e.convertToObject=P,e.convertToObjectWorker=M,e.convertToTSConfig=function(n,t,r){var a=e.createGetCanonicalFileName(r.useCaseSensitiveFileNames),i=e.map(e.filter(n.fileNames,n.configFileSpecs&&n.configFileSpecs.validatedIncludeSpecs?function(n,t,r,a){if(!t)return function(e){return!0};var i=e.getFileMatcherPatterns(n,r,t,a.useCaseSensitiveFileNames,a.getCurrentDirectory()),o=i.excludePattern&&e.getRegexFromPattern(i.excludePattern,a.useCaseSensitiveFileNames),s=i.includeFilePattern&&e.getRegexFromPattern(i.includeFilePattern,a.useCaseSensitiveFileNames);if(s)return o?function(e){return!(s.test(e)&&!o.test(e))}:function(e){return!s.test(e)};if(o)return function(e){return o.test(e)};return function(e){return!0}}(t,n.configFileSpecs.validatedIncludeSpecs,n.configFileSpecs.validatedExcludeSpecs,r):function(e){return!0}),(function(n){return e.getRelativePathFromFile(e.getNormalizedAbsolutePath(t,r.getCurrentDirectory()),e.getNormalizedAbsolutePath(n,r.getCurrentDirectory()),a)})),o=j(n.options,{configFilePath:e.getNormalizedAbsolutePath(t,r.getCurrentDirectory()),useCaseSensitiveFileNames:r.useCaseSensitiveFileNames}),s=n.watchOptions&&H(n.watchOptions,L());return __assign(__assign({compilerOptions:__assign(__assign({},G(o)),{showConfig:void 0,configFile:void 0,configFilePath:void 0,help:void 0,init:void 0,listFiles:void 0,listEmittedFiles:void 0,project:void 0,build:void 0,version:void 0}),watchOptions:s&&G(s),references:e.map(n.projectReferences,(function(e){return __assign(__assign({},e),{path:e.originalPath?e.originalPath:"",originalPath:void 0})})),files:e.length(i)?i:void 0},n.configFileSpecs?{include:B(n.configFileSpecs.validatedIncludeSpecs),exclude:n.configFileSpecs.validatedExcludeSpecs}:{}),{compileOnSave:!!n.compileOnSave||void 0})},e.generateTSConfig=function(n,t,r){var a=j(e.extend(n,e.defaultInitCompilerOptions));return function(){for(var n=e.createMultiMap(),c=0,l=e.optionDeclarations;c<l.length;c++){var u=l[c],d=u.category;s(u)&&n.add(e.getLocaleSpecificMessage(d),u)}var p=0,m=0,f=[];n.forEach((function(n,t){0!==f.length&&f.push({value:""}),f.push({value:"/* "+t+" */"});for(var r=0,o=n;r<o.length;r++){var s=o[r],c=void 0;c=a.has(s.name)?'"'+s.name+'": '+JSON.stringify(a.get(s.name))+((m+=1)===a.size?"":","):'// "'+s.name+'": '+JSON.stringify(i(s))+",",f.push({value:c,description:"/* "+(s.description&&e.getLocaleSpecificMessage(s.description)||s.name)+" */"}),p=Math.max(c.length,p)}}));var _=o(2),g=[];g.push("{"),g.push(_+'"compilerOptions": {'),g.push(""+_+_+"/* "+e.getLocaleSpecificMessage(e.Diagnostics.Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file)+" */"),g.push("");for(var y=0,v=f;y<v.length;y++){var h=v[y],b=h.value,E=h.description,T=void 0===E?"":E;g.push(b&&""+_+_+b+(T&&o(p-b.length+2)+T))}if(t.length){g.push(_+"},"),g.push(_+'"files": [');for(var S=0;S<t.length;S++)g.push(""+_+_+JSON.stringify(t[S])+(S===t.length-1?"":","));g.push(_+"]")}else g.push(_+"}");return g.push("}"),g.join(r)+r}();function i(n){switch(n.type){case"number":return 1;case"boolean":return!0;case"string":return n.isFilePath?"./":"";case"list":return[];case"object":return{};default:var t=n.type.keys().next();return t.done?e.Debug.fail("Expected 'option.type' to have entries."):t.value}}function o(e){return Array(e+1).join(" ")}function s(n){var t=n.category,r=n.name;return void 0!==t&&t!==e.Diagnostics.Command_line_Options&&(t!==e.Diagnostics.Advanced_Options||a.has(r))}},e.convertToOptionsWithAbsolutePaths=function(n,t){var r={},a=o().optionsNameMap;for(var i in n)e.hasProperty(n,i)&&(r[i]=K(a.get(i.toLowerCase()),n[i],t));return r.configFilePath&&(r.configFilePath=t(r.configFilePath)),r},e.parseJsonConfigFileContent=function(e,n,t,r,a,i,o,s,c){return X(e,void 0,n,t,r,c,a,i,o,s)},e.parseJsonSourceFileConfigFileContent=W,e.setConfigFileInOptions=z,e.canJsonReportNoInputFiles=Z,e.updateErrorForNoInputFiles=function(n,t,r,a,i){var o=a.length;return Q(n,i)?a.push(Y(r,t)):e.filterMutate(a,(function(n){return!function(n){return n.code===e.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2.code}(n)})),o!==a.length},e.convertCompilerOptionsFromJson=function(e,n,t){var r=[];return{options:re(e,n,r,t),errors:r}},e.convertTypeAcquisitionFromJson=function(e,n,t){var r=[];return{options:ie(e,n,r,t),errors:r}};var pe=/(^|\/)\*\*\/?$/,me=/(^|\/)\*\*\/(.*\/)?\.\.($|\/)/,fe=/\/[^/]*?[*?][^/]*\//,_e=/^[^*?]*(?=\/[^/]*[*?])/;function ge(n,t,r,a,i){void 0===i&&(i=e.emptyArray),t=e.normalizePath(t);var o,s=e.createGetCanonicalFileName(a.useCaseSensitiveFileNames),c=new e.Map,l=new e.Map,u=new e.Map,d=n.validatedFilesSpec,p=n.validatedIncludeSpecs,m=n.validatedExcludeSpecs,f=n.wildcardDirectories,_=e.getSupportedExtensions(r,i),g=e.getSuppoertedExtensionsWithJsonIfResolveJsonModule(r,_);if(d)for(var y=0,v=d;y<v.length;y++){var h=v[y],b=e.getNormalizedAbsolutePath(h,t);c.set(s(b),b)}if(p&&p.length>0)for(var E=function(n){if(e.fileExtensionIs(n,".json")){if(!o){var r=p.filter((function(n){return e.endsWith(n,".json")})),i=e.map(e.getRegularExpressionsForWildcards(r,t,"files"),(function(e){return"^"+e+"$"}));o=i?i.map((function(n){return e.getRegexFromPattern(n,a.useCaseSensitiveFileNames)})):e.emptyArray}if(-1!==e.findIndex(o,(function(e){return e.test(n)}))){var d=s(n);c.has(d)||u.has(d)||u.set(d,n)}return"continue"}if(function(n,t,r,a,i){for(var o=e.getExtensionPriority(n,a),s=e.adjustExtensionPriority(o,a),c=0;c<s;c++){var l=a[c],u=i(e.changeExtension(n,l));if(t.has(u)||r.has(u))return!0}return!1}(n,c,l,_,s))return"continue";!function(n,t,r,a){for(var i=e.getExtensionPriority(n,r),o=e.getNextLowestExtensionPriority(i,r);o<r.length;o++){var s=r[o],c=a(e.changeExtension(n,s));t.delete(c)}}(n,l,_,s);var m=s(n);c.has(m)||l.has(m)||l.set(m,n)},T=0,S=a.readDirectory(t,g,m,p,void 0);T<S.length;T++){E(b=S[T])}var x=e.arrayFrom(c.values()),L=e.arrayFrom(l.values());return{fileNames:x.concat(L,e.arrayFrom(u.values())),wildcardDirectories:f,spec:n}}function ye(n,t,r,a,i){return n.filter((function(n){if(!e.isString(n))return!1;var o=function(n,t){if(!t&&pe.test(n))return e.Diagnostics.File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0;if(me.test(n))return e.Diagnostics.File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0}(n,r);return void 0!==o&&t.push(function(n,t){var r=e.getTsConfigPropArrayElementValue(a,i,t);return r?e.createDiagnosticForNodeInSourceFile(a,r,n,t):e.createCompilerDiagnostic(n,t)}(o,n)),void 0===o}))}function ve(n,t){var r=_e.exec(n);return r?{key:t?r[0]:e.toFileNameLowerCase(r[0]),flags:fe.test(n)?1:0}:e.isImplicitGlob(n)?{key:t?n:e.toFileNameLowerCase(n),flags:1}:void 0}function he(n,t){switch(t.type){case"object":case"string":return"";case"number":return"number"==typeof n?n:"";case"boolean":return"boolean"==typeof n?n:"";case"list":var r=t.element;return e.isArray(n)?n.map((function(e){return he(e,r)})):"";default:return e.forEachEntry(t.type,(function(e,t){if(e===n)return t}))}}e.getFileNamesFromConfigSpecs=ge,e.isExcludedFile=function(n,t,r,a,i){var o=t.validatedFilesSpec,s=t.validatedIncludeSpecs,c=t.validatedExcludeSpecs;if(!e.length(s)||!e.length(c))return!1;r=e.normalizePath(r);var l=e.createGetCanonicalFileName(a);if(o)for(var u=0,d=o;u<d.length;u++){var p=d[u];if(l(e.getNormalizedAbsolutePath(p,r))===n)return!1}var m=e.getRegularExpressionForWildcard(c,e.combinePaths(e.normalizePath(i),r),"exclude"),f=m&&e.getRegexFromPattern(m,a);return!!f&&(!!f.test(n)||!e.hasExtension(n)&&f.test(e.ensureTrailingDirectorySeparator(n)))},e.convertCompilerOptionsForTelemetry=function(e){var n={};for(var t in e)if(e.hasOwnProperty(t)){var r=g(t);void 0!==r&&(n[t]=he(e[t],r))}return n}}(ts||(ts={})),function(e){function n(n){n.trace(e.formatMessage.apply(void 0,arguments))}function t(e,n){return!!e.traceResolution&&void 0!==n.trace}function r(n,t){var r;if(t&&n){var a=n.packageJsonContent;"string"==typeof a.name&&"string"==typeof a.version&&(r={name:a.name,subModuleName:t.path.slice(n.packageDirectory.length+e.directorySeparator.length),version:a.version})}return t&&{path:t.path,extension:t.ext,packageId:r}}function a(e){return r(void 0,e)}function i(n){if(n)return e.Debug.assert(void 0===n.packageId),{path:n.path,ext:n.extension}}var o,s;function c(n){if(n)return e.Debug.assert(e.extensionIsTS(n.extension)),{fileName:n.path,packageId:n.packageId}}function l(e,n,t,r){var a;return r?((a=r.failedLookupLocations).push.apply(a,t),r):{resolvedModule:e&&{resolvedFileName:e.path,originalPath:!0===e.originalPath?void 0:e.originalPath,extension:e.extension,isExternalLibraryImport:n,packageId:e.packageId},failedLookupLocations:t}}function u(t,r,a,i){if(e.hasProperty(t,r)){var o=t[r];if(typeof o===a&&null!==o)return o;i.traceEnabled&&n(i.host,e.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2,r,a,null===o?"null":typeof o)}else i.traceEnabled&&n(i.host,e.Diagnostics.package_json_does_not_have_a_0_field,r)}function d(t,r,a,i){var o=u(t,r,"string",i);if(void 0!==o){if(o){var s=e.normalizePath(e.combinePaths(a,o));return i.traceEnabled&&n(i.host,e.Diagnostics.package_json_has_0_field_1_that_references_2,r,o,s),s}i.traceEnabled&&n(i.host,e.Diagnostics.package_json_had_a_falsy_0_field,r)}}function p(e,n,t){return d(e,"typings",n,t)||d(e,"types",n,t)}function m(e,n,t){return d(e,"main",n,t)}function f(t,r){var a=function(t,r){var a=u(t,"typesVersions","object",r);if(void 0!==a)return r.traceEnabled&&n(r.host,e.Diagnostics.package_json_has_a_typesVersions_field_with_version_specific_path_mappings),a}(t,r);if(void 0!==a){if(r.traceEnabled)for(var i in a)e.hasProperty(a,i)&&!e.VersionRange.tryParse(i)&&n(r.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range,i);var o=_(a);if(o){var s=o.version,c=o.paths;if("object"==typeof c)return o;r.traceEnabled&&n(r.host,e.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2,"typesVersions['"+s+"']","object",typeof c)}else r.traceEnabled&&n(r.host,e.Diagnostics.package_json_does_not_have_a_typesVersions_entry_that_matches_version_0,e.versionMajorMinor)}}function _(n){for(var t in s||(s=new e.Version(e.version)),n)if(e.hasProperty(n,t)){var r=e.VersionRange.tryParse(t);if(void 0!==r&&r.test(s))return{version:t,paths:n[t]}}}function g(n,t){return n.typeRoots?n.typeRoots:(n.configFilePath?r=e.getDirectoryPath(n.configFilePath):t.getCurrentDirectory&&(r=t.getCurrentDirectory()),void 0!==r?function(n,t){if(!t.directoryExists)return[e.combinePaths(n,y)];var r;return e.forEachAncestorDirectory(e.normalizePath(n),(function(n){var a=e.combinePaths(n,y);t.directoryExists(a)&&(r||(r=[])).push(a)})),r}(r,t):void 0);var r}e.trace=n,e.isTraceEnabled=t,function(e){e[e.TypeScript=0]="TypeScript",e[e.JavaScript=1]="JavaScript",e[e.Json=2]="Json",e[e.TSConfig=3]="TSConfig",e[e.DtsOnly=4]="DtsOnly"}(o||(o={})),e.getPackageJsonTypesVersionsPaths=_,e.getEffectiveTypeRoots=g;var y=e.combinePaths("node_modules","@types");function v(n){var t=new e.Map,r=new e.Map;return{ownMap:t,redirectsMap:r,getOrCreateMapOfCacheRedirects:function(a){if(!a)return t;var i=a.sourceFile.path,o=r.get(i);o||(o=!n||e.optionsHaveModuleResolutionChanges(n,a.commandLine.options)?new e.Map:t,r.set(i,o));return o},clear:function(){t.clear(),r.clear()},setOwnOptions:function(e){n=e},setOwnMap:function(e){t=e}}}function h(n,t,r,a){return{getOrCreateCacheForDirectory:function(t,o){var s=e.toPath(t,r,a);return i(n,o,s,(function(){return new e.Map}))},getOrCreateCacheForModuleName:function(n,r){return e.Debug.assert(!e.isExternalModuleNameRelative(n)),i(t,r,n,o)},directoryToModuleNameMap:n,moduleNameToDirectoryMap:t};function i(e,n,t,r){var a=e.getOrCreateMapOfCacheRedirects(n),i=a.get(t);return i||(i=r(),a.set(t,i)),i}function o(){var n=new e.Map;return{get:function(t){return n.get(e.toPath(t,r,a))},set:function(t,i){var o=e.toPath(t,r,a);if(n.has(o))return;n.set(o,i);var s=i.resolvedModule&&(i.resolvedModule.originalPath||i.resolvedModule.resolvedFileName),c=s&&function(n,t){var i=e.toPath(e.getDirectoryPath(t),r,a),o=0,s=Math.min(n.length,i.length);for(;o<s&&n.charCodeAt(o)===i.charCodeAt(o);)o++;if(o===n.length&&(i.length===o||i[o]===e.directorySeparator))return n;var c=e.getRootLength(n);if(o<c)return;var l=n.lastIndexOf(e.directorySeparator,o-1);if(-1===l)return;return n.substr(0,Math.max(l,c))}(o,s),l=o;for(;l!==c;){var u=e.getDirectoryPath(l);if(u===l||n.has(u))break;n.set(u,i),l=u}}}}}function b(t,r,a,i,o){var s=function(t,r,a,i){var o=i.compilerOptions,s=o.baseUrl,c=o.paths;if(c&&!e.pathIsRelative(r)){return i.traceEnabled&&(s&&n(i.host,e.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1,s,r),n(i.host,e.Diagnostics.paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0,r)),K(t,r,e.getPathsBasePath(i.compilerOptions,i.host),c,a,!1,i)}}(t,r,i,o);return s?s.value:e.isExternalModuleNameRelative(r)?function(t,r,a,i,o){if(!o.compilerOptions.rootDirs)return;o.traceEnabled&&n(o.host,e.Diagnostics.rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0,r);for(var s,c,l=e.normalizePath(e.combinePaths(a,r)),u=0,d=o.compilerOptions.rootDirs;u<d.length;u++){var p=d[u],m=e.normalizePath(p);e.endsWith(m,e.directorySeparator)||(m+=e.directorySeparator);var f=e.startsWith(l,m)&&(void 0===c||c.length<m.length);o.traceEnabled&&n(o.host,e.Diagnostics.Checking_if_0_is_the_longest_matching_prefix_for_1_2,m,l,f),f&&(c=m,s=p)}if(c){o.traceEnabled&&n(o.host,e.Diagnostics.Longest_matching_prefix_for_0_is_1,l,c);var _=l.substr(c.length);o.traceEnabled&&n(o.host,e.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2,_,c,l);var g=i(t,l,!e.directoryProbablyExists(a,o.host),o);if(g)return g;o.traceEnabled&&n(o.host,e.Diagnostics.Trying_other_entries_in_rootDirs);for(var y=0,v=o.compilerOptions.rootDirs;y<v.length;y++){if((p=v[y])!==s){var h=e.combinePaths(e.normalizePath(p),_);o.traceEnabled&&n(o.host,e.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2,_,p,h);var b=e.getDirectoryPath(h),E=i(t,h,!e.directoryProbablyExists(b,o.host),o);if(E)return E}}o.traceEnabled&&n(o.host,e.Diagnostics.Module_resolution_using_rootDirs_has_failed)}return}(t,r,a,i,o):function(t,r,a,i){var o=i.compilerOptions.baseUrl;if(!o)return;i.traceEnabled&&n(i.host,e.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1,o,r);var s=e.normalizePath(e.combinePaths(o,r));i.traceEnabled&&n(i.host,e.Diagnostics.Resolving_module_name_0_relative_to_base_url_1_2,r,o,s);return a(t,s,!e.directoryProbablyExists(e.getDirectoryPath(s),i.host),i)}(t,r,i,o)}e.resolveTypeReferenceDirective=function(r,a,i,s,l){var u=t(i,s);l&&(i=l.commandLine.options);var d=[],p={compilerOptions:i,host:s,traceEnabled:u,failedLookupLocations:d},m=g(i,s);u&&(void 0===a?void 0===m?n(s,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set,r):n(s,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1,r,m):void 0===m?n(s,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set,r,a):n(s,e.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_2,r,a,m),l&&n(s,e.Diagnostics.Using_compiler_options_of_project_reference_redirect_0,l.sourceFile.fileName));var f,_=function(){if(m&&m.length)return u&&n(s,e.Diagnostics.Resolving_with_primary_search_path_0,m.join(", ")),e.firstDefined(m,(function(t){var a=e.combinePaths(t,r),i=e.getDirectoryPath(a),l=e.directoryProbablyExists(i,s);return!l&&u&&n(s,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,i),c(w(o.DtsOnly,a,!l,p))}));u&&n(s,e.Diagnostics.Root_directory_cannot_be_determined_skipping_primary_search_paths)}(),y=!0;if(_||(_=function(){var t=a&&e.getDirectoryPath(a);if(void 0!==t){u&&n(s,e.Diagnostics.Looking_up_in_node_modules_folder_initial_location_0,t);var i=void 0;if(e.isExternalModuleNameRelative(r)){var l=e.normalizePathAndParts(e.combinePaths(t,r)).path;i=I(o.DtsOnly,l,!1,p,!0)}else{var d=U(o.DtsOnly,r,t,p,void 0,void 0);i=d&&d.value}var m=c(i);return!m&&u&&n(s,e.Diagnostics.Type_reference_directive_0_was_not_resolved,r),m}u&&n(s,e.Diagnostics.Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder)}(),y=!1),_){var v=_.fileName,h=_.packageId,b=i.preserveSymlinks?v:C(v,s,u);u&&(h?n(s,e.Diagnostics.Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3,r,b,e.packageIdToString(h),y):n(s,e.Diagnostics.Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2,r,b,y)),f={primary:y,resolvedFileName:b,packageId:h,isExternalLibraryImport:N(v)}}return{resolvedTypeReferenceDirective:f,failedLookupLocations:d}},e.getAutomaticTypeDirectiveNames=function(n,t){if(n.types)return n.types;var r=[];if(t.directoryExists&&t.getDirectories){var a=g(n,t);if(a)for(var i=0,o=a;i<o.length;i++){var s=o[i];if(t.directoryExists(s))for(var c=0,l=t.getDirectories(s);c<l.length;c++){var u=l[c],d=e.normalizePath(u),p=e.combinePaths(s,d,"package.json");if(!(t.fileExists(p)&&null===e.readJson(p,t).typings)){var m=e.getBaseFileName(d);46!==m.charCodeAt(0)&&r.push(m)}}}}return r},e.createModuleResolutionCache=function(e,n,t){return h(v(t),v(t),e,n)},e.createCacheWithRedirects=v,e.createModuleResolutionCacheWithMaps=h,e.resolveModuleNameFromCache=function(n,t,r){var a=e.getDirectoryPath(t),i=r&&r.getOrCreateCacheForDirectory(a);return i&&i.get(n)},e.resolveModuleName=function(r,a,i,o,s,c){var l=t(i,o);c&&(i=c.commandLine.options),l&&(n(o,e.Diagnostics.Resolving_module_0_from_1,r,a),c&&n(o,e.Diagnostics.Using_compiler_options_of_project_reference_redirect_0,c.sourceFile.fileName));var u=e.getDirectoryPath(a),d=s&&s.getOrCreateCacheForDirectory(u,c),p=d&&d.get(r);if(p)l&&n(o,e.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1,r,u);else{var m=i.moduleResolution;switch(void 0===m?(m=e.getEmitModuleKind(i)===e.ModuleKind.CommonJS?e.ModuleResolutionKind.NodeJs:e.ModuleResolutionKind.Classic,l&&n(o,e.Diagnostics.Module_resolution_kind_is_not_specified_using_0,e.ModuleResolutionKind[m])):l&&n(o,e.Diagnostics.Explicitly_specified_module_resolution_kind_Colon_0,e.ModuleResolutionKind[m]),e.perfLogger.logStartResolveModule(r),m){case e.ModuleResolutionKind.NodeJs:p=A(r,a,i,o,s,c);break;case e.ModuleResolutionKind.Classic:p=X(r,a,i,o,s,c);break;default:return e.Debug.fail("Unexpected moduleResolution: "+m)}p&&p.resolvedModule&&e.perfLogger.logInfoEvent('Module "'+r+'" resolved to "'+p.resolvedModule.resolvedFileName+'"'),e.perfLogger.logStopResolveModule(p&&p.resolvedModule?""+p.resolvedModule.resolvedFileName:"null"),d&&(d.set(r,p),e.isExternalModuleNameRelative(r)||s.getOrCreateCacheForModuleName(r,c).set(u,p))}return l&&(p.resolvedModule?p.resolvedModule.packageId?n(o,e.Diagnostics.Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2,r,p.resolvedModule.resolvedFileName,e.packageIdToString(p.resolvedModule.packageId)):n(o,e.Diagnostics.Module_name_0_was_successfully_resolved_to_1,r,p.resolvedModule.resolvedFileName):n(o,e.Diagnostics.Module_name_0_was_not_resolved,r)),p},e.resolveJSModule=function(e,n,t){var r=L(e,n,t),a=r.resolvedModule,i=r.failedLookupLocations;if(!a)throw new Error("Could not resolve JS module '"+e+"' starting at '"+n+"'. Looked in: "+i.join(", "));return a.resolvedFileName},e.tryResolveJSModule=function(e,n,t){var r=L(e,n,t).resolvedModule;return r&&r.resolvedFileName};var E=[o.JavaScript],T=[o.TypeScript,o.JavaScript],S=__spreadArrays(T,[o.Json]),x=[o.TSConfig];function L(n,t,r){return D(n,t,{moduleResolution:e.ModuleResolutionKind.NodeJs,allowJs:!0},r,void 0,E,void 0)}function A(n,t,r,a,i,o,s){return D(n,e.getDirectoryPath(t),r,a,i,s?x:r.resolveJsonModule?S:T,o)}function D(r,a,i,s,c,u,d){var p,m,f=t(i,s),_=[],g={compilerOptions:i,host:s,traceEnabled:f,failedLookupLocations:_},y=e.forEach(u,(function(t){return function(t){var l=b(t,r,a,(function(e,n,t,r){return I(e,n,t,r,!0)}),g);if(l)return Y({resolved:l,isExternalLibraryImport:N(l.path)});if(e.isExternalModuleNameRelative(r)){var u=e.normalizePathAndParts(e.combinePaths(a,r)),p=u.path,m=u.parts,_=I(t,p,!1,g,!0);return _&&Y({resolved:_,isExternalLibraryImport:e.contains(m,"node_modules")})}f&&n(s,e.Diagnostics.Loading_module_0_from_node_modules_folder_target_file_type_1,r,o[t]);var y=U(t,r,a,g,c,d);if(!y)return;var v=y.value;if(!i.preserveSymlinks&&v&&!v.originalPath){var h=C(v.path,s,f),E=h===v.path?void 0:v.path;v=__assign(__assign({},v),{path:h,originalPath:E})}return{value:v&&{resolved:v,isExternalLibraryImport:!0}}}(t)}));return l(null===(p=null==y?void 0:y.value)||void 0===p?void 0:p.resolved,null===(m=null==y?void 0:y.value)||void 0===m?void 0:m.isExternalLibraryImport,_,g.resultFromCache)}function C(t,r,a){if(!r.realpath)return t;var i=e.normalizePath(r.realpath(t));return a&&n(r,e.Diagnostics.Resolving_real_path_for_0_result_1,t,i),e.Debug.assert(r.fileExists(i),t+" linked to nonexistent file "+i),i}function I(t,a,i,s,c){if(s.traceEnabled&&n(s.host,e.Diagnostics.Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1,a,o[t]),!e.hasTrailingDirectorySeparator(a)){if(!i){var l=e.getDirectoryPath(a);e.directoryProbablyExists(l,s.host)||(s.traceEnabled&&n(s.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,l),i=!0)}var u=O(t,a,i,s);if(u){var d=c?function(n){var t=e.normalizePath(n.path),r=t.lastIndexOf(e.nodeModulesPathPart);if(-1===r)return;var a=r+e.nodeModulesPathPart.length,i=R(t,a);64===t.charCodeAt(a)&&(i=R(t,i));return t.slice(0,i)}(u):void 0;return r(d?F(d,!1,s):void 0,u)}}i||(e.directoryProbablyExists(a,s.host)||(s.traceEnabled&&n(s.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,a),i=!0));return w(t,a,i,s,c)}function N(n){return e.stringContains(n,e.nodeModulesPathPart)}function R(n,t){var r=n.indexOf(e.directorySeparator,t+1);return-1===r?t:r}function k(e,n,t,r){return a(O(e,n,t,r))}function O(t,r,a,i){if(t===o.Json||t===o.TSConfig){var s=e.tryRemoveExtension(r,".json");return void 0===s&&t===o.Json?void 0:P(s||r,t,a,i)}var c=P(r,t,a,i);if(c)return c;if(e.hasJSFileExtension(r)){var l=e.removeFileExtension(r);if(i.traceEnabled){var u=r.substring(l.length);n(i.host,e.Diagnostics.File_name_0_has_a_1_extension_stripping_it,r,u)}return P(l,t,a,i)}}function P(n,t,r,a){if(!r){var i=e.getDirectoryPath(n);i&&(r=!e.directoryProbablyExists(i,a.host))}switch(t){case o.DtsOnly:return s(".d.ts");case o.TypeScript:return s(".ts")||s(".tsx")||s(".d.ts");case o.JavaScript:return s(".js")||s(".jsx");case o.TSConfig:case o.Json:return s(".json")}function s(e){var t=M(n+e,r,a);return void 0===t?void 0:{path:t,ext:e}}}function M(t,r,a){if(!r){if(a.host.fileExists(t))return a.traceEnabled&&n(a.host,e.Diagnostics.File_0_exist_use_it_as_a_name_resolution_result,t),t;a.traceEnabled&&n(a.host,e.Diagnostics.File_0_does_not_exist,t)}a.failedLookupLocations.push(t)}function w(e,n,t,a,i){void 0===i&&(i=!0);var o=i?F(n,t,a):void 0;return r(o,G(e,n,t,a,o&&o.packageJsonContent,o&&o.versionPaths))}function F(t,r,a){var i=a.host,o=a.traceEnabled,s=!r&&e.directoryProbablyExists(t,i),c=e.combinePaths(t,"package.json");if(s&&i.fileExists(c)){var l=e.readJson(c,i);return o&&n(i,e.Diagnostics.Found_package_json_at_0,c),{packageDirectory:t,packageJsonContent:l,versionPaths:f(l,a)}}s&&o&&n(i,e.Diagnostics.File_0_does_not_exist,c),a.failedLookupLocations.push(c)}function G(t,r,s,c,l,u){var f;if(l)switch(t){case o.JavaScript:case o.Json:f=m(l,r,c);break;case o.TypeScript:f=p(l,r,c)||m(l,r,c);break;case o.DtsOnly:f=p(l,r,c);break;case o.TSConfig:f=function(e,n,t){return d(e,"tsconfig",n,t)}(l,r,c);break;default:return e.Debug.assertNever(t)}var _=function(t,r,i,s){var c=M(r,i,s);if(c){var l=function(n,t){var r=e.tryGetExtensionFromPath(t);return void 0!==r&&function(e,n){switch(e){case o.JavaScript:return".js"===n||".jsx"===n;case o.TSConfig:case o.Json:return".json"===n;case o.TypeScript:return".ts"===n||".tsx"===n||".d.ts"===n;case o.DtsOnly:return".d.ts"===n}}(n,r)?{path:t,ext:r}:void 0}(t,c);if(l)return a(l);s.traceEnabled&&n(s.host,e.Diagnostics.File_0_has_an_unsupported_extension_so_skipping_it,c)}return I(t===o.DtsOnly?o.TypeScript:t,r,i,s,!1)},g=f?!e.directoryProbablyExists(e.getDirectoryPath(f),c.host):void 0,y=s||!e.directoryProbablyExists(r,c.host),v=e.combinePaths(r,t===o.TSConfig?"tsconfig":"index");if(u&&(!f||e.containsPath(r,f))){var h=e.getRelativePathFromDirectory(r,f||v,!1);c.traceEnabled&&n(c.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2,u.version,e.version,h);var b=K(t,h,r,u.paths,_,g||y,c);if(b)return i(b.value)}var E=f&&i(_(t,f,g,c));return E||O(t,v,y,c)}function B(n){var t=n.indexOf(e.directorySeparator);return"@"===n[0]&&(t=n.indexOf(e.directorySeparator,t+1)),-1===t?{packageName:n,rest:""}:{packageName:n.slice(0,t),rest:n.slice(t+1)}}function U(e,n,t,r,a,i){return V(e,n,t,r,!1,a,i)}function V(n,t,r,a,i,o,s){var c=o&&o.getOrCreateCacheForModuleName(t,s);return e.forEachAncestorDirectory(e.normalizeSlashes(r),(function(r){if("node_modules"!==e.getBaseFileName(r)){var o=J(c,t,r,a);return o||Y(j(n,t,r,a,i))}}))}function j(t,r,a,i,s){var c=e.combinePaths(a,"node_modules"),l=e.directoryProbablyExists(c,i.host);!l&&i.traceEnabled&&n(i.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,c);var u=s?void 0:H(t,r,c,l,i);if(u)return u;if(t===o.TypeScript||t===o.DtsOnly){var d=e.combinePaths(c,"@types"),p=l;return l&&!e.directoryProbablyExists(d,i.host)&&(i.traceEnabled&&n(i.host,e.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it,d),p=!1),H(o.DtsOnly,function(t,r){var a=z(t);r.traceEnabled&&a!==t&&n(r.host,e.Diagnostics.Scoped_package_detected_looking_in_0,a);return a}(r,i),d,p,i)}}function H(t,i,o,s,c){var l=e.normalizePath(e.combinePaths(o,i)),u=F(l,!s,c);if(u){var d=O(t,l,!s,c);if(d)return a(d);var p=G(t,l,!s,c,u.packageJsonContent,u.versionPaths);return r(u,p)}var m=function(e,n,t,a){var i=O(e,n,t,a)||G(e,n,t,a,u&&u.packageJsonContent,u&&u.versionPaths);return r(u,i)},f=B(i),_=f.packageName,g=f.rest;if(""!==g){var y=e.combinePaths(o,_);if((u=F(y,!s,c))&&u.versionPaths){c.traceEnabled&&n(c.host,e.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2,u.versionPaths.version,e.version,g);var v=s&&e.directoryProbablyExists(y,c.host),h=K(t,g,y,u.versionPaths.paths,m,!v,c);if(h)return h.value}}return m(t,l,!s,c)}function K(t,r,i,o,s,c,l){var u=e.matchPatternOrExact(e.getOwnKeys(o),r);if(u){var d=e.isString(u)?void 0:e.matchedText(u,r),p=e.isString(u)?u:e.patternText(u);return l.traceEnabled&&n(l.host,e.Diagnostics.Module_name_0_matched_pattern_1,r,p),{value:e.forEach(o[p],(function(r){var o=d?r.replace("*",d):r,u=e.normalizePath(e.combinePaths(i,o));l.traceEnabled&&n(l.host,e.Diagnostics.Trying_substitution_0_candidate_module_location_Colon_1,r,o);var p=e.tryGetExtensionFromPath(u);if(void 0!==p){var m=M(u,c,l);if(void 0!==m)return a({path:m,ext:p})}return s(t,u,c||!e.directoryProbablyExists(e.getDirectoryPath(u),l.host),l)}))}}}e.nodeModuleNameResolver=A,e.nodeModulesPathPart="/node_modules/",e.pathContainsNodeModules=N,e.parsePackageName=B;var W="__";function z(n){if(e.startsWith(n,"@")){var t=n.replace(e.directorySeparator,W);if(t!==n)return t.slice(1)}return n}function q(n){return e.stringContains(n,W)?"@"+n.replace(W,e.directorySeparator):n}function J(t,r,a,i){var o=t&&t.get(a);if(o)return i.traceEnabled&&n(i.host,e.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1,r,a),i.resultFromCache=o,{value:o.resolvedModule&&{path:o.resolvedModule.resolvedFileName,originalPath:o.resolvedModule.originalPath||!0,extension:o.resolvedModule.extension,packageId:o.resolvedModule.packageId}}}function X(n,r,a,i,s,c){var u=[],d={compilerOptions:a,host:i,traceEnabled:t(a,i),failedLookupLocations:u},p=e.getDirectoryPath(r),m=f(o.TypeScript)||f(o.JavaScript);return l(m&&m.value,!1,u,d.resultFromCache);function f(t){var r=b(t,n,p,k,d);if(r)return{value:r};if(e.isExternalModuleNameRelative(n)){var a=e.normalizePath(e.combinePaths(p,n));return Y(k(t,a,!1,d))}var i=s&&s.getOrCreateCacheForModuleName(n,c),l=e.forEachAncestorDirectory(p,(function(r){var a=J(i,n,r,d);if(a)return a;var o=e.normalizePath(e.combinePaths(r,n));return Y(k(t,o,!1,d))}));return l||(t===o.TypeScript?function(e,n,t){return V(o.DtsOnly,e,n,t,!0,void 0,void 0)}(n,p,d):void 0)}}function Y(e){return void 0!==e?{value:e}:void 0}e.getTypesPackageName=function(e){return"@types/"+z(e)},e.mangleScopedPackageName=z,e.getPackageNameFromTypesPackageName=function(n){var t=e.removePrefix(n,"@types/");return t!==n?q(t):n},e.unmangleScopedPackageName=q,e.classicNameResolver=X,e.loadModuleFromGlobalCache=function(r,a,i,s,c){var u=t(i,s);u&&n(s,e.Diagnostics.Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2,a,r,c);var d=[],p={compilerOptions:i,host:s,traceEnabled:u,failedLookupLocations:d};return l(j(o.DtsOnly,r,c,p,!1),!0,d,p.resultFromCache)}}(ts||(ts={})),function(e){var n;function t(n,t){return n.body&&!n.body.parent&&(e.setParent(n.body,n),e.setParentRecursive(n.body,!1)),n.body?r(n.body,t):1}function r(n,i){void 0===i&&(i=new e.Map);var o=e.getNodeId(n);if(i.has(o))return i.get(o)||0;i.set(o,void 0);var s=function(n,i){switch(n.kind){case 253:case 254:return 0;case 255:if(e.isEnumConst(n))return 2;break;case 261:case 260:if(!e.hasSyntacticModifier(n,1))return 0;break;case 267:var o=n;if(!o.moduleSpecifier&&o.exportClause&&268===o.exportClause.kind){for(var s=0,c=0,l=o.exportClause.elements;c<l.length;c++){var u=a(l[c],i);if(u>s&&(s=u),1===s)return s}return s}break;case 257:var d=0;return e.forEachChild(n,(function(n){var t=r(n,i);switch(t){case 0:return;case 2:return void(d=2);case 1:return d=1,!0;default:e.Debug.assertNever(t)}})),d;case 256:return t(n,i);case 78:if(n.isInJSDocNamespace)return 0}return 1}(n,i);return i.set(o,s),s}function a(n,t){for(var a=n.propertyName||n.name,i=n.parent;i;){if(e.isBlock(i)||e.isModuleBlock(i)||e.isSourceFile(i)){for(var o=void 0,s=0,c=i.statements;s<c.length;s++){var l=c[s];if(e.nodeHasName(l,a)){l.parent||(e.setParent(l,i),e.setParentRecursive(l,!1));var u=r(l,t);if((void 0===o||u>o)&&(o=u),1===o)return o}}if(void 0!==o)return o}i=i.parent}return 1}function i(n){return e.Debug.attachFlowNodeDebugInfo(n),n}!function(e){e[e.NonInstantiated=0]="NonInstantiated",e[e.Instantiated=1]="Instantiated",e[e.ConstEnumOnly=2]="ConstEnumOnly"}(e.ModuleInstanceState||(e.ModuleInstanceState={})),e.getModuleInstanceState=t,function(e){e[e.None=0]="None",e[e.IsContainer=1]="IsContainer",e[e.IsBlockScopedContainer=2]="IsBlockScopedContainer",e[e.IsControlFlowContainer=4]="IsControlFlowContainer",e[e.IsFunctionLike=8]="IsFunctionLike",e[e.IsFunctionExpression=16]="IsFunctionExpression",e[e.HasLocals=32]="HasLocals",e[e.IsInterface=64]="IsInterface",e[e.IsObjectLiteralOrClassExpressionMethod=128]="IsObjectLiteralOrClassExpressionMethod"}(n||(n={}));var o=function(){var n,r,a,o,u,d,p,m,f,_,g,y,v,h,b,E,T,S,x,L,A,D,C,I,N=!1,R=0,k={flags:1},O={flags:1};function P(t,r,a,i,o){return e.createDiagnosticForNodeInSourceFile(e.getSourceFileOfNode(t)||n,t,r,a,i,o)}return function(t,s){n=t,r=s,a=e.getEmitScriptTarget(r),D=function(n,t){return!(!e.getStrictOptionValue(t,"alwaysStrict")||n.isDeclarationFile)||!!n.externalModuleIndicator}(n,s),I=new e.Set,R=0,C=e.objectAllocator.getSymbolConstructor(),e.Debug.attachFlowNodeDebugInfo(k),e.Debug.attachFlowNodeDebugInfo(O),n.locals||(Me(n),n.symbolCount=R,n.classifiableNames=I,function(){if(f){for(var t=u,r=m,a=p,s=o,l=g,d=0,_=f;d<_.length;d++){var y=_[d],v=e.getJSDocHost(y);u=e.findAncestor(v.parent,(function(e){return!!(1&Te(e))}))||n,p=e.getEnclosingBlockScopeContainer(v)||n,g=i({flags:2}),o=y,Me(y.typeExpression);var h=e.getNameOfDeclaration(y);if((e.isJSDocEnumTag(y)||!y.fullName)&&h&&e.isPropertyAccessEntityNameExpression(h.parent)){var b=Ye(h.parent);if(b){Je(n.symbol,h.parent,b,!!e.findAncestor(h,(function(n){return e.isPropertyAccessExpression(n)&&"prototype"===n.name.escapedText})),!1);var E=u;switch(e.getAssignmentDeclarationPropertyAccessKind(h.parent)){case 1:case 2:u=e.isExternalOrCommonJsModule(n)?n:void 0;break;case 4:u=h.parent.expression;break;case 3:u=h.parent.expression.name;break;case 5:u=c(n,h.parent.expression)?n:e.isPropertyAccessExpression(h.parent.expression)?h.parent.expression.name:h.parent.expression;break;case 0:return e.Debug.fail("Shouldn't have detected typedef or enum on non-assignment declaration")}u&&U(y,524288,788968),u=E}}else e.isJSDocEnumTag(y)||!y.fullName||78===y.fullName.kind?(o=y.parent,Ce(y,524288,788968)):Me(y.fullName)}u=t,m=r,p=a,o=s,g=l}}()),n=void 0,r=void 0,a=void 0,o=void 0,u=void 0,d=void 0,p=void 0,m=void 0,f=void 0,_=!1,g=void 0,y=void 0,v=void 0,h=void 0,b=void 0,E=void 0,T=void 0,x=void 0,L=!1,N=!1,A=0};function M(e,n){return R++,new C(e,n)}function w(n,t,r){n.flags|=r,t.symbol=n,n.declarations=e.appendIfUnique(n.declarations,t),1955&r&&!n.exports&&(n.exports=e.createSymbolTable()),6240&r&&!n.members&&(n.members=e.createSymbolTable()),n.constEnumOnlyModule&&304&n.flags&&(n.constEnumOnlyModule=!1),111551&r&&e.setValueDeclaration(n,t)}function F(n){if(266===n.kind)return n.isExportEquals?"export=":"default";var t=e.getNameOfDeclaration(n);if(t){if(e.isAmbientModule(n)){var r=e.getTextOfIdentifierOrLiteral(t);return e.isGlobalScopeAugmentation(n)?"__global":'"'+r+'"'}if(158===t.kind){var a=t.expression;return e.isStringOrNumericLiteralLike(a)?e.escapeLeadingUnderscores(a.text):e.isSignedNumericLiteral(a)?e.tokenToString(a.operator)+a.operand.text:(e.Debug.assert(e.isWellKnownSymbolSyntactically(a)),e.getPropertyNameForKnownSymbolName(e.idText(a.name)))}if(e.isWellKnownSymbolSyntactically(t))return e.getPropertyNameForKnownSymbolName(e.idText(t.name));if(e.isPrivateIdentifier(t)){var i=e.getContainingClass(n);if(!i)return;var o=i.symbol;return e.getSymbolNameForPrivateIdentifier(o,t.escapedText)}return e.isPropertyNameLiteral(t)?e.getEscapedTextOfIdentifierOrLiteral(t):void 0}switch(n.kind){case 166:return"__constructor";case 174:case 169:case 313:return"__call";case 175:case 170:return"__new";case 171:return"__index";case 267:return"__export";case 297:return"export=";case 216:if(2===e.getAssignmentDeclarationKind(n))return"export=";e.Debug.fail("Unknown binary declaration kind");break;case 308:return e.isJSDocConstructSignature(n)?"__new":"__call";case 160:return e.Debug.assert(308===n.parent.kind,"Impossible parameter parent kind",(function(){return"parent is: "+(e.SyntaxKind?e.SyntaxKind[n.parent.kind]:n.parent.kind)+", expected JSDocFunctionType"})),"arg"+n.parent.parameters.indexOf(n)}}function G(n){return e.isNamedDeclaration(n)?e.declarationNameToString(n.name):e.unescapeLeadingUnderscores(e.Debug.checkDefined(F(n)))}function B(t,r,a,i,o,s){e.Debug.assert(!e.hasDynamicName(a));var c,l=e.hasSyntacticModifier(a,512)||e.isExportSpecifier(a)&&"default"===a.name.escapedText,u=l&&r?"default":F(a);if(void 0===u)c=M(0,"__missing");else if(c=t.get(u),2885600&i&&I.add(u),c){if(s&&!c.isReplaceableByMethod)return c;if(c.flags&o)if(c.isReplaceableByMethod)t.set(u,c=M(0,u));else if(!(3&i&&67108864&c.flags)){e.isNamedDeclaration(a)&&e.setParent(a.name,a);var d=2&c.flags?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,p=!0;(384&c.flags||384&i)&&(d=e.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations,p=!1);var m=!1;e.length(c.declarations)&&(l||c.declarations&&c.declarations.length&&266===a.kind&&!a.isExportEquals)&&(d=e.Diagnostics.A_module_cannot_have_multiple_default_exports,p=!1,m=!0);var f=[];e.isTypeAliasDeclaration(a)&&e.nodeIsMissing(a.type)&&e.hasSyntacticModifier(a,1)&&2887656&c.flags&&f.push(P(a,e.Diagnostics.Did_you_mean_0,"export type { "+e.unescapeLeadingUnderscores(a.name.escapedText)+" }"));var _=e.getNameOfDeclaration(a)||a;e.forEach(c.declarations,(function(t,r){var a=e.getNameOfDeclaration(t)||t,i=P(a,d,p?G(t):void 0);n.bindDiagnostics.push(m?e.addRelatedInfo(i,P(_,0===r?e.Diagnostics.Another_export_default_is_here:e.Diagnostics.and_here)):i),m&&f.push(P(a,e.Diagnostics.The_first_export_default_is_here))}));var g=P(_,d,p?G(a):void 0);n.bindDiagnostics.push(e.addRelatedInfo.apply(void 0,__spreadArrays([g],f))),c=M(0,u)}}else t.set(u,c=M(0,u)),s&&(c.isReplaceableByMethod=!0);return w(c,a,i),c.parent?e.Debug.assert(c.parent===r,"Existing symbol parent should match new one"):c.parent=r,c}function U(n,t,r){var a=!!(1&e.getCombinedModifierFlags(n))||function(n){n.parent&&e.isModuleDeclaration(n)&&(n=n.parent);if(!e.isJSDocTypeAlias(n))return!1;if(!e.isJSDocEnumTag(n)&&n.fullName)return!0;var t=e.getNameOfDeclaration(n);return!!t&&(!(!e.isPropertyAccessEntityNameExpression(t.parent)||!Ye(t.parent))||!!(e.isDeclaration(t.parent)&&1&e.getCombinedModifierFlags(t.parent)))}(n);if(2097152&t)return 270===n.kind||260===n.kind&&a?B(u.symbol.exports,u.symbol,n,t,r):B(u.locals,void 0,n,t,r);if(e.isJSDocTypeAlias(n)&&e.Debug.assert(e.isInJSFile(n)),!e.isAmbientModule(n)&&(a||64&u.flags)){if(!u.locals||e.hasSyntacticModifier(n,512)&&!F(n))return B(u.symbol.exports,u.symbol,n,t,r);var i=111551&t?1048576:0,o=B(u.locals,void 0,n,i,r);return o.exportSymbol=B(u.symbol.exports,u.symbol,n,t,r),n.localSymbol=o,o}return B(u.locals,void 0,n,t,r)}function V(e){j(e,(function(e){return 251===e.kind?Me(e):void 0})),j(e,(function(e){return 251!==e.kind?Me(e):void 0}))}function j(n,t){void 0===t&&(t=Me),void 0!==n&&e.forEach(n,t)}function H(n){e.forEachChild(n,Me,j)}function K(n){var a=N;if(N=!1,function(n){if(!(1&g.flags))return!1;if(g===k){if((e.isStatementButNotDeclaration(n)&&231!==n.kind||252===n.kind||256===n.kind&&function(e){var n=t(e);return 1===n||2===n&&!!r.preserveConstEnums}(n))&&(g=O,!r.allowUnreachableCode)){var a=e.unreachableCodeIsError(r)&&!(8388608&n.flags)&&(!e.isVariableStatement(n)||!!(3&e.getCombinedNodeFlags(n.declarationList))||n.declarationList.declarations.some((function(e){return!!e.initializer})));!function(n,t){if(e.isStatement(n)&&s(n)&&e.isBlock(n.parent)){var r=n.parent.statements,a=e.sliceAfter(r,n);e.getRangesWhere(a,s,(function(e,n){return t(a[e],a[n-1])}))}else t(n,n)}(n,(function(n,t){return Pe(a,n,t,e.Diagnostics.Unreachable_code_detected)}))}}return!0}(n))return H(n),we(n),void(N=a);switch(n.kind>=232&&n.kind<=248&&!r.allowUnreachableCode&&(n.flowNode=g),n.kind){case 236:!function(e){var n=pe(e,Q()),t=Y(),r=Y();ee(n,g),g=n,ue(e.expression,t,r),g=ie(t),de(e.statement,r,n),ee(n,g),g=ie(r)}(n);break;case 235:!function(e){var n=Q(),t=pe(e,Y()),r=Y();ee(n,g),g=n,de(e.statement,r,t),ee(t,g),g=ie(t),ue(e.expression,n,r),g=ie(r)}(n);break;case 237:!function(e){var n=pe(e,Q()),t=Y(),r=Y();Me(e.initializer),ee(n,g),g=n,ue(e.condition,t,r),g=ie(t),de(e.statement,r,n),Me(e.incrementor),ee(n,g),g=ie(r)}(n);break;case 238:case 239:!function(e){var n=pe(e,Q()),t=Y();Me(e.expression),ee(n,g),g=n,239===e.kind&&Me(e.awaitModifier);ee(t,g),Me(e.initializer),250!==e.initializer.kind&&ge(e.initializer);de(e.statement,t,n),ee(n,g),g=ie(t)}(n);break;case 234:!function(e){var n=Y(),t=Y(),r=Y();ue(e.expression,n,t),g=ie(n),Me(e.thenStatement),ee(r,g),g=ie(t),Me(e.elseStatement),ee(r,g),g=ie(r)}(n);break;case 242:case 246:!function(e){Me(e.expression),242===e.kind&&(L=!0,h&&ee(h,g));g=k}(n);break;case 241:case 240:!function(e){if(Me(e.label),e.label){var n=function(e){for(var n=x;n;n=n.next)if(n.name===e)return n;return}(e.label.escapedText);n&&(n.referenced=!0,me(e,n.breakTarget,n.continueTarget))}else me(e,y,v)}(n);break;case 247:!function(n){var t=h,r=T,a=Y(),i=Y(),o=Y();n.finallyBlock&&(h=i);ee(o,g),T=o,Me(n.tryBlock),ee(a,g),n.catchClause&&(g=ie(o),ee(o=Y(),g),T=o,Me(n.catchClause),ee(a,g));if(h=t,T=r,n.finallyBlock){var s=Y();s.antecedents=e.concatenate(e.concatenate(a.antecedents,o.antecedents),i.antecedents),g=s,Me(n.finallyBlock),1&g.flags?g=k:(h&&i.antecedents&&ee(h,Z(s,i.antecedents,g)),T&&o.antecedents&&ee(T,Z(s,o.antecedents,g)),g=a.antecedents?Z(s,a.antecedents,g):k)}else g=ie(a)}(n);break;case 244:!function(n){var t=Y();Me(n.expression);var r=y,a=S;y=t,S=g,Me(n.caseBlock),ee(t,g);var i=e.forEach(n.caseBlock.clauses,(function(e){return 285===e.kind}));n.possiblyExhaustive=!i&&!t.antecedents,i||ee(t,te(S,n,0,0));y=r,S=a,g=ie(t)}(n);break;case 258:!function(e){for(var n=e.clauses,t=W(e.parent.expression),a=k,i=0;i<n.length;i++){for(var o=i;!n[i].statements.length&&i+1<n.length;)Me(n[i]),i++;var s=Y();ee(s,t?te(S,e.parent,o,i+1):S),ee(s,a),g=ie(s);var c=n[i];Me(c),a=g,1&g.flags||i===n.length-1||!r.noFallthroughCasesInSwitch||(c.fallthroughFlowNode=g)}}(n);break;case 284:!function(e){var n=g;g=S,Me(e.expression),g=n,j(e.statements)}(n);break;case 233:!function(e){Me(e.expression),fe(e.expression)}(n);break;case 245:!function(n){var t=Y();x={next:x,name:n.label.escapedText,breakTarget:t,continueTarget:void 0,referenced:!1},Me(n.label),Me(n.statement),x.referenced||r.allowUnusedLabels||function(e,n,t){Pe(e,n,n,t)}(e.unusedLabelIsError(r),n.label,e.Diagnostics.Unused_label);x=x.next,ee(t,g),g=ie(t)}(n);break;case 214:!function(e){if(53===e.operator){var n=b;b=E,E=n,H(e),E=b,b=n}else H(e),45!==e.operator&&46!==e.operator||ge(e.operand)}(n);break;case 215:!function(e){H(e),(45===e.operator||46===e.operator)&&ge(e.operand)}(n);break;case 216:if(e.isDestructuringAssignment(n))return N=a,void function(e){N?(N=!1,Me(e.operatorToken),Me(e.right),N=!0,Me(e.left)):(N=!0,Me(e.left),N=!1,Me(e.operatorToken),Me(e.right));ge(e.left)}(n);!function(n){var t={expr:[n],state:[1],inStrictMode:[void 0],parent:[void 0]},r=0;for(;r>=0;)switch(n=t.expr[r],t.state[r]){case 0:e.setParent(n,o);var a=D;Ge(n);var i=o;o=n,l(1,a,i);break;case 1:if(55===(c=n.operatorToken.kind)||56===c||60===c||e.isLogicalOrCoalescingAssignmentOperator(c)){if(ce(n)){var s=Y();ye(n,s,s),g=ie(s)}else ye(n,b,E);u()}else l(2),d(n.left);break;case 2:27===n.operatorToken.kind&&fe(n.left),l(3),d(n.operatorToken);break;case 3:l(4),d(n.right);break;case 4:var c=n.operatorToken.kind;if(e.isAssignmentOperator(c)&&!e.isAssignmentTarget(n))if(ge(n.left),62===c&&202===n.left.kind)X(n.left.expression)&&(g=re(256,g,n));u();break;default:return e.Debug.fail("Invalid state "+t.state[r]+" for bindBinaryExpressionFlow")}function l(e,n,a){t.state[r]=e,void 0!==n&&(t.inStrictMode[r]=n),void 0!==a&&(t.parent[r]=a)}function u(){void 0!==t.inStrictMode[r]&&(D=t.inStrictMode[r],o=t.parent[r]),r--}function d(n){n&&e.isBinaryExpression(n)&&!e.isDestructuringAssignment(n)?(r++,t.expr[r]=n,t.state[r]=0,t.inStrictMode[r]=void 0,t.parent[r]=void 0):Me(n)}}(n);break;case 210:!function(e){H(e),201===e.expression.kind&&ge(e.expression)}(n);break;case 217:!function(e){var n=Y(),t=Y(),r=Y();ue(e.condition,n,t),g=ie(n),Me(e.questionToken),Me(e.whenTrue),ee(r,g),g=ie(t),Me(e.colonToken),Me(e.whenFalse),ee(r,g),g=ie(r)}(n);break;case 249:!function(n){H(n),(n.initializer||e.isForInOrOfStatement(n.parent.parent))&&ve(n)}(n);break;case 201:case 202:!function(n){e.isOptionalChain(n)?Ee(n):H(n)}(n);break;case 203:!function(n){if(e.isOptionalChain(n))Ee(n);else{var t=e.skipParentheses(n.expression);208===t.kind||209===t.kind?(j(n.typeArguments),j(n.arguments),Me(n.expression)):(H(n),105===n.expression.kind&&(g=ae(g,n)))}if(201===n.expression.kind){var r=n.expression;e.isIdentifier(r.name)&&X(r.expression)&&e.isPushOrUnshiftIdentifier(r.name)&&(g=re(256,g,n))}}(n);break;case 225:!function(n){e.isOptionalChain(n)?Ee(n):H(n)}(n);break;case 331:case 324:case 325:!function(n){e.setParent(n.tagName,n),325!==n.kind&&n.fullName&&(e.setParent(n.fullName,n),e.setParentRecursive(n.fullName,!1))}(n);break;case 297:V(n.statements),Me(n.endOfFileToken);break;case 230:case 257:V(n.statements);break;case 198:!function(n){e.isBindingPattern(n.name)?(j(n.decorators),j(n.modifiers),Me(n.dotDotDotToken),Me(n.propertyName),Me(n.initializer),Me(n.name)):H(n)}(n);break;case 200:case 199:case 288:case 220:N=a;default:H(n)}we(n),N=a}function W(n){switch(n.kind){case 78:case 79:case 107:case 201:case 202:return q(n);case 203:return function(e){if(e.arguments)for(var n=0,t=e.arguments;n<t.length;n++){if(q(t[n]))return!0}if(201===e.expression.kind&&q(e.expression.expression))return!0;return!1}(n);case 207:case 225:return W(n.expression);case 216:return function(n){switch(n.operatorToken.kind){case 62:case 74:case 75:case 76:return q(n.left);case 34:case 35:case 36:case 37:return X(n.left)||X(n.right)||J(n.right,n.left)||J(n.left,n.right);case 101:return X(n.left);case 100:return t=n.left,r=n.right,e.isStringLiteralLike(t)&&W(r);case 27:return W(n.right)}var t,r;return!1}(n);case 214:return 53===n.operator&&W(n.operand);case 211:return W(n.expression)}return!1}function z(n){return 78===n.kind||79===n.kind||107===n.kind||105===n.kind||(e.isPropertyAccessExpression(n)||e.isNonNullExpression(n)||e.isParenthesizedExpression(n))&&z(n.expression)||e.isBinaryExpression(n)&&27===n.operatorToken.kind&&z(n.right)||e.isElementAccessExpression(n)&&e.isStringOrNumericLiteralLike(n.argumentExpression)&&z(n.expression)||e.isAssignmentExpression(n)&&z(n.left)}function q(n){return z(n)||e.isOptionalChain(n)&&q(n.expression)}function J(n,t){return e.isTypeOfExpression(n)&&X(n.expression)&&e.isStringLiteralLike(t)}function X(e){switch(e.kind){case 207:return X(e.expression);case 216:switch(e.operatorToken.kind){case 62:return X(e.left);case 27:return X(e.right)}}return q(e)}function Y(){return i({flags:4,antecedents:void 0})}function Q(){return i({flags:8,antecedents:void 0})}function Z(e,n,t){return i({flags:1024,target:e,antecedents:n,antecedent:t})}function $(e){e.flags|=2048&e.flags?4096:2048}function ee(n,t){1&t.flags||e.contains(n.antecedents,t)||((n.antecedents||(n.antecedents=[])).push(t),$(t))}function ne(n,t,r){return 1&t.flags?t:r?!(109===r.kind&&64&n||94===r.kind&&32&n)||e.isExpressionOfOptionalChainRoot(r)||e.isNullishCoalesce(r.parent)?W(r)?($(t),i({flags:n,antecedent:t,node:r})):t:k:32&n?t:k}function te(e,n,t,r){return $(e),i({flags:128,antecedent:e,switchStatement:n,clauseStart:t,clauseEnd:r})}function re(e,n,t){$(n);var r=i({flags:e,antecedent:n,node:t});return T&&ee(T,r),r}function ae(e,n){return $(e),i({flags:512,antecedent:e,node:n})}function ie(e){var n=e.antecedents;return n?1===n.length?n[0]:e:k}function oe(e){for(;;)if(207===e.kind)e=e.expression;else{if(214!==e.kind||53!==e.operator)return 216===e.kind&&(55===e.operatorToken.kind||56===e.operatorToken.kind||60===e.operatorToken.kind);e=e.operand}}function se(n){return n=e.skipParentheses(n),e.isBinaryExpression(n)&&e.isLogicalOrCoalescingAssignmentOperator(n.operatorToken.kind)}function ce(n){for(;e.isParenthesizedExpression(n.parent)||e.isPrefixUnaryExpression(n.parent)&&53===n.parent.operator;)n=n.parent;return!(function(e){var n=e.parent;switch(n.kind){case 234:case 236:case 235:return n.expression===e;case 237:case 217:return n.condition===e}return!1}(n)||se(n.parent)||oe(n.parent)||e.isOptionalChain(n.parent)&&n.parent.expression===n)}function le(e,n,t,r){var a=b,i=E;b=t,E=r,e(n),b=a,E=i}function ue(n,t,r){le(Me,n,t,r),n&&(se(n)||oe(n)||e.isOptionalChain(n)&&e.isOutermostOptionalChain(n))||(ee(t,ne(32,g,n)),ee(r,ne(64,g,n)))}function de(e,n,t){var r=y,a=v;y=n,v=t,Me(e),y=r,v=a}function pe(e,n){for(var t=x;t&&245===e.parent.kind;)t.continueTarget=n,t=t.next,e=e.parent;return n}function me(e,n,t){var r=241===e.kind?n:t;r&&(ee(r,g),g=k)}function fe(n){if(203===n.kind){var t=n;e.isDottedName(t.expression)&&105!==t.expression.kind&&(g=ae(g,t))}}function _e(e){216===e.kind&&62===e.operatorToken.kind?ge(e.left):ge(e)}function ge(e){if(z(e))g=re(16,g,e);else if(199===e.kind)for(var n=0,t=e.elements;n<t.length;n++){var r=t[n];220===r.kind?ge(r.expression):_e(r)}else if(200===e.kind)for(var a=0,i=e.properties;a<i.length;a++){var o=i[a];288===o.kind?_e(o.initializer):289===o.kind?ge(o.name):290===o.kind&&ge(o.expression)}}function ye(n,t,r){var a=Y();55===n.operatorToken.kind||75===n.operatorToken.kind?ue(n.left,a,r):ue(n.left,t,a),g=ie(a),Me(n.operatorToken),e.isLogicalOrCoalescingAssignmentOperator(n.operatorToken.kind)?(le(Me,n.right,t,r),ge(n.left),ee(t,ne(32,g,n)),ee(r,ne(64,g,n))):ue(n.right,t,r)}function ve(n){var t=e.isOmittedExpression(n)?void 0:n.name;if(e.isBindingPattern(t))for(var r=0,a=t.elements;r<a.length;r++){ve(a[r])}else g=re(16,g,n)}function he(e){switch(e.kind){case 201:Me(e.questionDotToken),Me(e.name);break;case 202:Me(e.questionDotToken),Me(e.argumentExpression);break;case 203:Me(e.questionDotToken),j(e.typeArguments),j(e.arguments)}}function be(n,t,r){var a=e.isOptionalChainRoot(n)?Y():void 0;!function(n,t,r){le(Me,n,t,r),e.isOptionalChain(n)&&!e.isOutermostOptionalChain(n)||(ee(t,ne(32,g,n)),ee(r,ne(64,g,n)))}(n.expression,a||t,r),a&&(g=ie(a)),le(he,n,t,r),e.isOutermostOptionalChain(n)&&(ee(t,ne(32,g,n)),ee(r,ne(64,g,n)))}function Ee(e){if(ce(e)){var n=Y();be(e,n,n),g=ie(n)}else be(e,b,E)}function Te(n){switch(n.kind){case 221:case 252:case 255:case 200:case 177:case 312:case 281:return 1;case 253:return 65;case 256:case 254:case 190:return 33;case 297:return 37;case 165:if(e.isObjectLiteralOrClassExpressionMethod(n))return 173;case 166:case 251:case 164:case 167:case 168:case 169:case 313:case 308:case 174:case 170:case 171:case 175:return 45;case 208:case 209:return 61;case 257:return 4;case 163:return n.initializer?4:0;case 287:case 237:case 238:case 239:case 258:return 2;case 230:return e.isFunctionLike(n.parent)?0:2}return 0}function Se(e){m&&(m.nextContainer=e),m=e}function xe(t,r,a){switch(u.kind){case 256:return U(t,r,a);case 297:return function(t,r,a){return e.isExternalModule(n)?U(t,r,a):B(n.locals,void 0,t,r,a)}(t,r,a);case 221:case 252:return function(n,t,r){return e.hasSyntacticModifier(n,32)?B(u.symbol.exports,u.symbol,n,t,r):B(u.symbol.members,u.symbol,n,t,r)}(t,r,a);case 255:return B(u.symbol.exports,u.symbol,t,r,a);case 177:case 312:case 200:case 253:case 281:return B(u.symbol.members,u.symbol,t,r,a);case 174:case 175:case 169:case 170:case 313:case 171:case 165:case 164:case 166:case 167:case 168:case 251:case 208:case 209:case 308:case 331:case 324:case 254:case 190:return B(u.locals,void 0,t,r,a)}}function Le(n){8388608&n.flags&&!function(n){var t=e.isSourceFile(n)?n:e.tryCast(n.body,e.isModuleBlock);return!!t&&t.statements.some((function(n){return e.isExportDeclaration(n)||e.isExportAssignment(n)}))}(n)?n.flags|=64:n.flags&=-65}function Ae(e){var n=t(e),r=0!==n;return xe(e,r?512:1024,r?110735:0),n}function De(e,n,t){var r=M(n,t);return 106508&n&&(r.parent=u.symbol),w(r,e,n),r}function Ce(n,t,r){switch(p.kind){case 256:U(n,t,r);break;case 297:if(e.isExternalOrCommonJsModule(u)){U(n,t,r);break}default:p.locals||(p.locals=e.createSymbolTable(),Se(p)),B(p.locals,void 0,n,t,r)}}function Ie(t){n.parseDiagnostics.length||8388608&t.flags||4194304&t.flags||e.isIdentifierName(t)||(D&&t.originalKeywordKind>=116&&t.originalKeywordKind<=124?n.bindDiagnostics.push(P(t,function(t){if(e.getContainingClass(t))return e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode;if(n.externalModuleIndicator)return e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode;return e.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode}(t),e.declarationNameToString(t))):130===t.originalKeywordKind?e.isExternalModule(n)&&e.isInTopLevelContext(t)?n.bindDiagnostics.push(P(t,e.Diagnostics.Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module,e.declarationNameToString(t))):32768&t.flags&&n.bindDiagnostics.push(P(t,e.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here,e.declarationNameToString(t))):124===t.originalKeywordKind&&8192&t.flags&&n.bindDiagnostics.push(P(t,e.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here,e.declarationNameToString(t))))}function Ne(t,r){if(r&&78===r.kind){var a=r;if(o=a,e.isIdentifier(o)&&("eval"===o.escapedText||"arguments"===o.escapedText)){var i=e.getErrorSpanForNode(n,r);n.bindDiagnostics.push(e.createFileDiagnostic(n,i.start,i.length,function(t){if(e.getContainingClass(t))return e.Diagnostics.Invalid_use_of_0_Class_definitions_are_automatically_in_strict_mode;if(n.externalModuleIndicator)return e.Diagnostics.Invalid_use_of_0_Modules_are_automatically_in_strict_mode;return e.Diagnostics.Invalid_use_of_0_in_strict_mode}(t),e.idText(a)))}}var o}function Re(e){D&&Ne(e,e.name)}function ke(t){if(a<2&&297!==p.kind&&256!==p.kind&&!e.isFunctionLike(p)){var r=e.getErrorSpanForNode(n,t);n.bindDiagnostics.push(e.createFileDiagnostic(n,r.start,r.length,function(t){return e.getContainingClass(t)?e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode:n.externalModuleIndicator?e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode:e.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5}(t)))}}function Oe(t,r,a,i,o){var s=e.getSpanOfTokenAtPosition(n,t.pos);n.bindDiagnostics.push(e.createFileDiagnostic(n,s.start,s.length,r,a,i,o))}function Pe(t,r,a,i){!function(t,r,a){var i=e.createFileDiagnostic(n,r.pos,r.end-r.pos,a);t?n.bindDiagnostics.push(i):n.bindSuggestionDiagnostics=e.append(n.bindSuggestionDiagnostics,__assign(__assign({},i),{category:e.DiagnosticCategory.Suggestion}))}(t,{pos:e.getTokenPosOfNode(r,n),end:a.end},i)}function Me(n){if(n){e.setParent(n,o);var t=D;if(Ge(n),n.kind>156){var r=o;o=n;var a=Te(n);0===a?K(n):function(n,t){var r=u,a=d,o=p;if(1&t?(209!==n.kind&&(d=u),u=p=n,32&t&&(u.locals=e.createSymbolTable()),Se(u)):2&t&&((p=n).locals=void 0),4&t){var s=g,c=y,l=v,m=h,f=T,b=x,E=L,S=16&t&&!e.hasSyntacticModifier(n,256)&&!n.asteriskToken&&!!e.getImmediatelyInvokedFunctionExpression(n);S||(g=i({flags:2}),144&t&&(g.node=n)),h=S||166===n.kind||e.isInJSFile&&(251===n.kind||208===n.kind)?Y():void 0,T=void 0,y=void 0,v=void 0,x=void 0,L=!1,K(n),n.flags&=-2817,!(1&g.flags)&&8&t&&e.nodeIsPresent(n.body)&&(n.flags|=256,L&&(n.flags|=512),n.endFlowNode=g),297===n.kind&&(n.flags|=A),h&&(ee(h,g),g=ie(h),(166===n.kind||e.isInJSFile&&(251===n.kind||208===n.kind))&&(n.returnFlowNode=g)),S||(g=s),y=c,v=l,h=m,T=f,x=b,L=E}else 64&t?(_=!1,K(n),n.flags=_?128|n.flags:-129&n.flags):K(n);u=r,d=a,p=o}(n,a),o=r}else{r=o;1===n.kind&&(o=n),we(n),o=r}D=t}}function we(n){if(e.hasJSDocNodes(n))if(e.isInJSFile(n))for(var t=0,r=n.jsDoc;t<r.length;t++){Me(o=r[t])}else for(var a=0,i=n.jsDoc;a<i.length;a++){var o=i[a];e.setParent(o,n),e.setParentRecursive(o,!1)}}function Fe(t){if(!D)for(var r=0,a=t;r<a.length;r++){var i=a[r];if(!e.isPrologueDirective(i))return;if(o=i,s=void 0,'"use strict"'===(s=e.getSourceTextOfNodeFromSourceFile(n,o.expression))||"'use strict'"===s)return void(D=!0)}var o,s}function Ge(t){switch(t.kind){case 78:if(t.isInJSDocNamespace){for(var a=t.parent;a&&!e.isJSDocTypeAlias(a);)a=a.parent;Ce(a,524288,788968);break}case 107:return g&&(e.isExpression(t)||289===o.kind)&&(t.flowNode=g),Ie(t);case 105:t.flowNode=g;break;case 79:return function(t){"#constructor"===t.escapedText&&(n.parseDiagnostics.length||n.bindDiagnostics.push(P(t,e.Diagnostics.constructor_is_a_reserved_word,e.declarationNameToString(t))))}(t);case 201:case 202:var i=t;g&&z(i)&&(i.flowNode=g),e.isSpecialPropertyDeclaration(i)&&function(n){107===n.expression.kind?He(n):e.isBindableStaticAccessExpression(n)&&297===n.parent.parent.kind&&(e.isPrototypeAccess(n.expression)?ze(n,n.parent):qe(n))}(i),e.isInJSFile(i)&&n.commonJsModuleIndicator&&e.isModuleExportsAccessExpression(i)&&!l(p,"module")&&B(n.locals,void 0,i.expression,134217729,111550);break;case 216:switch(e.getAssignmentDeclarationKind(t)){case 1:Ve(t);break;case 2:!function(t){if(!Ue(t))return;var r=e.getRightMostAssignedExpression(t.right);if(e.isEmptyObjectLiteral(r)||u===n&&c(n,r))return;if(e.isObjectLiteralExpression(r)&&e.every(r.properties,e.isShorthandPropertyAssignment))return void e.forEach(r.properties,je);var a=e.exportAssignmentIsAlias(t)?2097152:1049092,i=B(n.symbol.exports,n.symbol,t,67108864|a,0);e.setValueDeclaration(i,t)}(t);break;case 3:ze(t.left,t);break;case 6:!function(n){e.setParent(n.left,n),e.setParent(n.right,n),Qe(n.left.expression,n.left,!1,!0)}(t);break;case 4:He(t);break;case 5:var s=t.left.expression;if(e.isInJSFile(t)&&e.isIdentifier(s)){var d=l(p,s.escapedText);if(e.isThisInitializedDeclaration(null==d?void 0:d.valueDeclaration)){He(t);break}}!function(t){var r,a=Ze(t.left.expression,u)||Ze(t.left.expression,p);if(!e.isInJSFile(t)&&!e.isFunctionSymbol(a))return;var i=e.getLeftmostAccessExpression(t.left);if(e.isIdentifier(i)&&2097152&(null===(r=l(u,i.escapedText))||void 0===r?void 0:r.flags))return;if(e.setParent(t.left,t),e.setParent(t.right,t),e.isIdentifier(t.left.expression)&&u===n&&c(n,t.left.expression))Ve(t);else if(e.hasDynamicName(t)){De(t,67108868,"__computed"),We(t,Je(a,t.left.expression,Ye(t.left),!1,!1))}else qe(e.cast(t.left,e.isBindableStaticNameExpression))}(t);break;case 0:break;default:e.Debug.fail("Unknown binary expression special property assignment kind")}return function(n){D&&e.isLeftHandSideExpression(n.left)&&e.isAssignmentOperator(n.operatorToken.kind)&&Ne(n,n.left)}(t);case 287:return function(e){D&&e.variableDeclaration&&Ne(e,e.variableDeclaration.name)}(t);case 210:return function(t){if(D&&78===t.expression.kind){var r=e.getErrorSpanForNode(n,t.expression);n.bindDiagnostics.push(e.createFileDiagnostic(n,r.start,r.length,e.Diagnostics.delete_cannot_be_called_on_an_identifier_in_strict_mode))}}(t);case 8:return function(t){D&&32&t.numericLiteralFlags&&n.bindDiagnostics.push(P(t,e.Diagnostics.Octal_literals_are_not_allowed_in_strict_mode))}(t);case 215:return function(e){D&&Ne(e,e.operand)}(t);case 214:return function(e){D&&(45!==e.operator&&46!==e.operator||Ne(e,e.operand))}(t);case 243:return function(n){D&&Oe(n,e.Diagnostics.with_statements_are_not_allowed_in_strict_mode)}(t);case 245:return function(n){D&&r.target>=2&&(e.isDeclarationStatement(n.statement)||e.isVariableStatement(n.statement))&&Oe(n.label,e.Diagnostics.A_label_is_not_allowed_here)}(t);case 187:return void(_=!0);case 172:break;case 159:return function(n){if(e.isJSDocTemplateTag(n.parent)){var t=e.find(n.parent.parent.tags,e.isJSDocTypeAlias)||e.getHostSignatureFromJSDoc(n.parent);t?(t.locals||(t.locals=e.createSymbolTable()),B(t.locals,void 0,n,262144,526824)):xe(n,262144,526824)}else if(185===n.parent.kind){var r=function(n){var t=e.findAncestor(n,(function(n){return n.parent&&e.isConditionalTypeNode(n.parent)&&n.parent.extendsType===n}));return t&&t.parent}(n.parent);r?(r.locals||(r.locals=e.createSymbolTable()),B(r.locals,void 0,n,262144,526824)):De(n,262144,F(n))}else xe(n,262144,526824)}(t);case 160:return nn(t);case 249:return en(t);case 198:return t.flowNode=g,en(t);case 163:case 162:return function(e){return tn(e,4|(e.questionToken?16777216:0),0)}(t);case 288:case 289:return tn(t,4,0);case 291:return tn(t,8,900095);case 169:case 170:case 171:return xe(t,131072,0);case 165:case 164:return tn(t,8192|(t.questionToken?16777216:0),e.isObjectLiteralMethod(t)?0:103359);case 251:return function(t){n.isDeclarationFile||8388608&t.flags||e.isAsyncFunction(t)&&(A|=2048);Re(t),D?(ke(t),Ce(t,16,110991)):xe(t,16,110991)}(t);case 166:return xe(t,16384,0);case 167:return tn(t,32768,46015);case 168:return tn(t,65536,78783);case 174:case 308:case 313:case 175:return function(n){var t=M(131072,F(n));w(t,n,131072);var r=M(2048,"__type");w(r,n,2048),r.members=e.createSymbolTable(),r.members.set(t.escapedName,t)}(t);case 177:case 312:case 190:return function(e){return De(e,2048,"__type")}(t);case 319:return function(n){H(n);var t=e.getHostSignatureFromJSDoc(n);t&&165!==t.kind&&w(t.symbol,t,32)}(t);case 200:return function(t){var r;if(function(e){e[e.Property=1]="Property",e[e.Accessor=2]="Accessor"}(r||(r={})),D&&!e.isAssignmentTarget(t))for(var a=new e.Map,i=0,o=t.properties;i<o.length;i++){var s=o[i];if(290!==s.kind&&78===s.name.kind){var c=s.name,l=288===s.kind||289===s.kind||165===s.kind?1:2,u=a.get(c.escapedText);if(u){if(1===l&&1===u){var d=e.getErrorSpanForNode(n,c);n.bindDiagnostics.push(e.createFileDiagnostic(n,d.start,d.length,e.Diagnostics.An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode))}}else a.set(c.escapedText,l)}}return De(t,4096,"__object")}(t);case 208:case 209:return function(t){n.isDeclarationFile||8388608&t.flags||e.isAsyncFunction(t)&&(A|=2048);g&&(t.flowNode=g);Re(t);var r=t.name?t.name.escapedText:"__function";return De(t,16,r)}(t);case 203:switch(e.getAssignmentDeclarationKind(t)){case 7:return function(e){var n=Ze(e.arguments[0]),t=297===e.parent.parent.kind;n=Je(n,e.arguments[0],t,!1,!1),Xe(e,n,!1)}(t);case 8:return function(e){if(!Ue(e))return;var n=$e(e.arguments[0],void 0,(function(e,n){return n&&w(n,e,67110400),n}));if(n){var t=1048580;B(n.exports,n,e,t,0)}}(t);case 9:return function(e){var n=Ze(e.arguments[0].expression);n&&n.valueDeclaration&&w(n,n.valueDeclaration,32);Xe(e,n,!0)}(t);case 0:break;default:return e.Debug.fail("Unknown call expression assignment declaration kind")}e.isInJSFile(t)&&function(t){!n.commonJsModuleIndicator&&e.isRequireCall(t,!1)&&Ue(t)}(t);break;case 221:case 252:return D=!0,function(t){if(252===t.kind)Ce(t,32,899503);else{De(t,32,t.name?t.name.escapedText:"__class"),t.name&&I.add(t.name.escapedText)}var r=t.symbol,a=M(4194308,"prototype"),i=r.exports.get(a.escapedName);i&&(t.name&&e.setParent(t.name,t),n.bindDiagnostics.push(P(i.declarations[0],e.Diagnostics.Duplicate_identifier_0,e.symbolName(a))));r.exports.set(a.escapedName,a),a.parent=r}(t);case 253:return Ce(t,64,788872);case 254:return Ce(t,524288,788968);case 255:return function(n){return e.isEnumConst(n)?Ce(n,128,899967):Ce(n,256,899327)}(t);case 256:return function(t){if(Le(t),e.isAmbientModule(t))if(e.hasSyntacticModifier(t,1)&&Oe(t,e.Diagnostics.export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible),e.isModuleAugmentationExternal(t))Ae(t);else{var r=void 0;if(10===t.name.kind){var a=t.name.text;e.hasZeroOrOneAsteriskCharacter(a)?r=e.tryParsePattern(a):Oe(t.name,e.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character,a)}var i=xe(t,512,110735);n.patternAmbientModules=e.append(n.patternAmbientModules,r&&{pattern:r,symbol:i})}else{var o=Ae(t);0!==o&&((i=t.symbol).constEnumOnlyModule=!(304&i.flags)&&2===o&&!1!==i.constEnumOnlyModule)}}(t);case 281:return function(e){return De(e,4096,"__jsxAttributes")}(t);case 280:return function(e,n,t){return xe(e,n,t)}(t,4,0);case 260:case 263:case 265:case 270:return xe(t,2097152,2097152);case 259:return function(t){t.modifiers&&t.modifiers.length&&n.bindDiagnostics.push(P(t,e.Diagnostics.Modifiers_cannot_appear_here));var r=e.isSourceFile(t.parent)?e.isExternalModule(t.parent)?t.parent.isDeclarationFile?void 0:e.Diagnostics.Global_module_exports_may_only_appear_in_declaration_files:e.Diagnostics.Global_module_exports_may_only_appear_in_module_files:e.Diagnostics.Global_module_exports_may_only_appear_at_top_level;r?n.bindDiagnostics.push(P(t,r)):(n.symbol.globalExports=n.symbol.globalExports||e.createSymbolTable(),B(n.symbol.globalExports,n.symbol,t,2097152,2097152))}(t);case 262:return function(e){e.name&&xe(e,2097152,2097152)}(t);case 267:return function(n){u.symbol&&u.symbol.exports?n.exportClause?e.isNamespaceExport(n.exportClause)&&(e.setParent(n.exportClause,n),B(u.symbol.exports,u.symbol,n.exportClause,2097152,2097152)):B(u.symbol.exports,u.symbol,n,8388608,0):De(n,8388608,F(n))}(t);case 266:return function(n){if(u.symbol&&u.symbol.exports){var t=e.exportAssignmentIsAlias(n)?2097152:4,r=B(u.symbol.exports,u.symbol,n,t,67108863);n.isExportEquals&&e.setValueDeclaration(r,n)}else De(n,2097152,F(n))}(t);case 297:return Fe(t.statements),function(){if(Le(n),e.isExternalModule(n))Be();else if(e.isJsonSourceFile(n)){Be();var t=n.symbol;B(n.symbol.exports,n.symbol,n,4,67108863),n.symbol=t}}();case 230:if(!e.isFunctionLike(t.parent))return;case 257:return Fe(t.statements);case 326:if(313===t.parent.kind)return nn(t);if(312!==t.parent.kind)break;case 333:var m=t;return xe(m,m.isBracketed||m.typeExpression&&307===m.typeExpression.type.kind?16777220:4,0);case 331:case 324:case 325:return(f||(f=[])).push(t)}}function Be(){De(n,512,'"'+e.removeFileExtension(n.fileName)+'"')}function Ue(e){return!n.externalModuleIndicator&&(n.commonJsModuleIndicator||(n.commonJsModuleIndicator=e,Be()),!0)}function Ve(n){if(Ue(n)){var t=$e(n.left.expression,void 0,(function(e,n){return n&&w(n,e,67110400),n}));if(t){var r=e.isAliasableExpression(n.right)&&(e.isExportsIdentifier(n.left.expression)||e.isModuleExportsAccessExpression(n.left.expression))?2097152:1048580;e.setParent(n.left,n),B(t.exports,t,n.left,r,0)}}}function je(e){B(n.symbol.exports,n.symbol,e,69206016,0)}function He(n){if(e.Debug.assert(e.isInJSFile(n)),!(e.isBinaryExpression(n)&&e.isPropertyAccessExpression(n.left)&&e.isPrivateIdentifier(n.left.name)||e.isPropertyAccessExpression(n)&&e.isPrivateIdentifier(n.name))){var t=e.getThisContainer(n,!1);switch(t.kind){case 251:case 208:var r=t.symbol;if(e.isBinaryExpression(t.parent)&&62===t.parent.operatorToken.kind){var a=t.parent.left;e.isBindableStaticAccessExpression(a)&&e.isPrototypeAccess(a.expression)&&(r=Ze(a.expression.expression,d))}r&&r.valueDeclaration&&(r.members=r.members||e.createSymbolTable(),e.hasDynamicName(n)?Ke(n,r):B(r.members,r,n,67108868,0),w(r,r.valueDeclaration,32));break;case 166:case 163:case 165:case 167:case 168:var i=t.parent,o=e.hasSyntacticModifier(t,32)?i.symbol.exports:i.symbol.members;e.hasDynamicName(n)?Ke(n,i.symbol):B(o,i.symbol,n,67108868,0,!0);break;case 297:if(e.hasDynamicName(n))break;t.commonJsModuleIndicator?B(t.symbol.exports,t.symbol,n,1048580,0):xe(n,1,111550);break;default:e.Debug.failBadSyntaxKind(t)}}}function Ke(e,n){De(e,4,"__computed"),We(e,n)}function We(n,t){t&&(t.assignmentDeclarationMembers||(t.assignmentDeclarationMembers=new e.Map)).set(e.getNodeId(n),n)}function ze(n,t){var r=n.expression,a=r.expression;e.setParent(a,r),e.setParent(r,n),e.setParent(n,t),Qe(a,n,!0,!0)}function qe(n){e.Debug.assert(!e.isIdentifier(n)),e.setParent(n.expression,n),Qe(n.expression,n,!1,!1)}function Je(t,r,a,i,o){if(2097152&(null==t?void 0:t.flags))return t;if(a&&!i){var s=67110400;t=$e(r,t,(function(t,r,a){return r?(w(r,t,s),r):B(a?a.exports:n.jsGlobalAugmentations||(n.jsGlobalAugmentations=e.createSymbolTable()),a,t,s,110735)}))}return o&&t&&t.valueDeclaration&&w(t,t.valueDeclaration,32),t}function Xe(n,t,r){if(t&&function(n){if(1072&n.flags)return!0;var t=n.valueDeclaration;if(t&&e.isCallExpression(t))return!!e.getAssignedExpandoInitializer(t);var r=t?e.isVariableDeclaration(t)?t.initializer:e.isBinaryExpression(t)?t.right:e.isPropertyAccessExpression(t)&&e.isBinaryExpression(t.parent)?t.parent.right:void 0:void 0;if(r=r&&e.getRightMostAssignedExpression(r)){var a=e.isPrototypeAccess(e.isVariableDeclaration(t)?t.name:e.isBinaryExpression(t)?t.left:t);return!!e.getExpandoInitializer(!e.isBinaryExpression(r)||56!==r.operatorToken.kind&&60!==r.operatorToken.kind?r:r.right,a)}return!1}(t)){var a=r?t.members||(t.members=e.createSymbolTable()):t.exports||(t.exports=e.createSymbolTable()),i=0,o=0;e.isFunctionLikeDeclaration(e.getAssignedExpandoInitializer(n))?(i=8192,o=103359):e.isCallExpression(n)&&e.isBindableObjectDefinePropertyCall(n)&&(e.some(n.arguments[2].properties,(function(n){var t=e.getNameOfDeclaration(n);return!!t&&e.isIdentifier(t)&&"set"===e.idText(t)}))&&(i|=65540,o|=78783),e.some(n.arguments[2].properties,(function(n){var t=e.getNameOfDeclaration(n);return!!t&&e.isIdentifier(t)&&"get"===e.idText(t)}))&&(i|=32772,o|=46015)),0===i&&(i=4,o=0),B(a,t,n,67108864|i,-67108865&o)}}function Ye(n){return e.isBinaryExpression(n.parent)?297===function(n){for(;e.isBinaryExpression(n.parent);)n=n.parent;return n.parent}(n.parent).parent.kind:297===n.parent.parent.kind}function Qe(e,n,t,r){var a=Ze(e,u)||Ze(e,p),i=Ye(n);Xe(n,a=Je(a,n.expression,i,t,r),t)}function Ze(n,t){if(void 0===t&&(t=u),e.isIdentifier(n))return l(t,n.escapedText);var r=Ze(n.expression);return r&&r.exports&&r.exports.get(e.getElementOrPropertyAccessName(n))}function $e(t,r,a){if(c(n,t))return n.symbol;if(e.isIdentifier(t))return a(t,Ze(t),r);var i=$e(t.expression,r,a),o=e.getNameOrArgument(t);return e.isPrivateIdentifier(o)&&e.Debug.fail("unexpected PrivateIdentifier"),a(o,i&&i.exports&&i.exports.get(e.getElementOrPropertyAccessName(t)),i)}function en(n){D&&Ne(n,n.name),e.isBindingPattern(n.name)||(e.isInJSFile(n)&&e.isRequireVariableDeclaration(n,!0)&&!e.getJSDocTypeTag(n)?xe(n,2097152,2097152):e.isBlockOrCatchScoped(n)?Ce(n,2,111551):e.isParameterDeclaration(n)?xe(n,1,111551):xe(n,1,111550))}function nn(n){if((326!==n.kind||313===u.kind)&&(!D||8388608&n.flags||Ne(n,n.name),e.isBindingPattern(n.name)?De(n,1,"__"+n.parent.parameters.indexOf(n)):xe(n,1,111551),e.isParameterPropertyDeclaration(n,n.parent))){var t=n.parent.parent;B(t.symbol.members,t.symbol,n,4|(n.questionToken?16777216:0),0)}}function tn(t,r,a){return n.isDeclarationFile||8388608&t.flags||!e.isAsyncFunction(t)||(A|=2048),g&&e.isObjectLiteralOrClassExpressionMethod(t)&&(t.flowNode=g),e.hasDynamicName(t)?De(t,r,"__computed"):xe(t,r,a)}}();function s(n){return!(e.isFunctionDeclaration(n)||function(n){switch(n.kind){case 253:case 254:return!0;case 256:return 1!==t(n);case 255:return e.hasSyntacticModifier(n,2048);default:return!1}}(n)||e.isEnumDeclaration(n)||e.isVariableStatement(n)&&!(3&e.getCombinedNodeFlags(n))&&n.declarationList.declarations.some((function(e){return!e.initializer})))}function c(n,t){for(var r=0,a=[t];a.length&&r<100;){if(r++,t=a.shift(),e.isExportsIdentifier(t)||e.isModuleExportsAccessExpression(t))return!0;if(e.isIdentifier(t)){var i=l(n,t.escapedText);if(i&&i.valueDeclaration&&e.isVariableDeclaration(i.valueDeclaration)&&i.valueDeclaration.initializer){var o=i.valueDeclaration.initializer;a.push(o),e.isAssignmentExpression(o,!0)&&(a.push(o.left),a.push(o.right))}}}return!1}function l(n,t){var r=n.locals&&n.locals.get(t);return r?r.exportSymbol||r:e.isSourceFile(n)&&n.jsGlobalAugmentations&&n.jsGlobalAugmentations.has(t)?n.jsGlobalAugmentations.get(t):n.symbol&&n.symbol.exports&&n.symbol.exports.get(t)}e.bindSourceFile=function(n,t){var r=["bind","bindSourceFile",{path:n.path}];e.tracing.begin.apply(e.tracing,r),e.performance.mark("beforeBind"),e.perfLogger.logStartBindFile(""+n.fileName),o(n,t),e.perfLogger.logStopBindFile(),e.performance.mark("afterBind"),e.performance.measure("Bind","beforeBind","afterBind"),e.tracing.end.apply(e.tracing,r)},e.isExportsOrModuleExportsOrAlias=c}(ts||(ts={})),function(e){e.createGetSymbolWalker=function(n,t,r,a,i,o,s,c,l,u,d){return function(p){void 0===p&&(p=function(){return!0});var m=[],f=[];return{walkType:function(n){try{return _(n),{visitedTypes:e.getOwnValues(m),visitedSymbols:e.getOwnValues(f)}}finally{e.clear(m),e.clear(f)}},walkSymbol:function(n){try{return v(n),{visitedTypes:e.getOwnValues(m),visitedSymbols:e.getOwnValues(f)}}finally{e.clear(m),e.clear(f)}}};function _(n){if(n&&(!m[n.id]&&(m[n.id]=n,!v(n.symbol)))){if(524288&n.flags){var t=n,r=t.objectFlags;4&r&&function(n){_(n.target),e.forEach(d(n),_)}(n),32&r&&function(e){_(e.typeParameter),_(e.constraintType),_(e.templateType),_(e.modifiersType)}(n),3&r&&(y(i=n),e.forEach(i.typeParameters,_),e.forEach(a(i),_),_(i.thisType)),24&r&&y(t)}var i;262144&n.flags&&function(e){_(l(e))}(n),3145728&n.flags&&function(n){e.forEach(n.types,_)}(n),4194304&n.flags&&function(e){_(e.type)}(n),8388608&n.flags&&function(e){_(e.objectType),_(e.indexType),_(e.constraint)}(n)}}function g(a){var i=t(a);i&&_(i.type),e.forEach(a.typeParameters,_);for(var o=0,s=a.parameters;o<s.length;o++){v(s[o])}_(n(a)),_(r(a))}function y(e){_(c(e,0)),_(c(e,1));for(var n=i(e),t=0,r=n.callSignatures;t<r.length;t++){g(r[t])}for(var a=0,o=n.constructSignatures;a<o.length;a++){g(o[a])}for(var s=0,l=n.properties;s<l.length;s++){v(l[s])}}function v(n){if(!n)return!1;var t=e.getSymbolId(n);return!f[t]&&(f[t]=n,!p(n)||(_(o(n)),n.exports&&n.exports.forEach(v),e.forEach(n.declarations,(function(e){if(e.type&&176===e.type.kind){var n=e.type;v(s(u(n.exprName)))}})),!1))}}}}(ts||(ts={})),function(e){var n,t,r,a,i=/^".+"$/,o="(anonymous)",s=1,c=1,l=1,u=1;!function(e){e[e.AllowsSyncIterablesFlag=1]="AllowsSyncIterablesFlag",e[e.AllowsAsyncIterablesFlag=2]="AllowsAsyncIterablesFlag",e[e.AllowsStringInputFlag=4]="AllowsStringInputFlag",e[e.ForOfFlag=8]="ForOfFlag",e[e.YieldStarFlag=16]="YieldStarFlag",e[e.SpreadFlag=32]="SpreadFlag",e[e.DestructuringFlag=64]="DestructuringFlag",e[e.PossiblyOutOfBounds=128]="PossiblyOutOfBounds",e[e.Element=1]="Element",e[e.Spread=33]="Spread",e[e.Destructuring=65]="Destructuring",e[e.ForOf=13]="ForOf",e[e.ForAwaitOf=15]="ForAwaitOf",e[e.YieldStar=17]="YieldStar",e[e.AsyncYieldStar=19]="AsyncYieldStar",e[e.GeneratorReturnType=1]="GeneratorReturnType",e[e.AsyncGeneratorReturnType=2]="AsyncGeneratorReturnType"}(n||(n={})),function(e){e[e.Yield=0]="Yield",e[e.Return=1]="Return",e[e.Next=2]="Next"}(t||(t={})),function(e){e[e.Normal=0]="Normal",e[e.FunctionReturn=1]="FunctionReturn",e[e.GeneratorNext=2]="GeneratorNext",e[e.GeneratorYield=3]="GeneratorYield"}(r||(r={})),function(e){e[e.None=0]="None",e[e.TypeofEQString=1]="TypeofEQString",e[e.TypeofEQNumber=2]="TypeofEQNumber",e[e.TypeofEQBigInt=4]="TypeofEQBigInt",e[e.TypeofEQBoolean=8]="TypeofEQBoolean",e[e.TypeofEQSymbol=16]="TypeofEQSymbol",e[e.TypeofEQObject=32]="TypeofEQObject",e[e.TypeofEQFunction=64]="TypeofEQFunction",e[e.TypeofEQHostObject=128]="TypeofEQHostObject",e[e.TypeofNEString=256]="TypeofNEString",e[e.TypeofNENumber=512]="TypeofNENumber",e[e.TypeofNEBigInt=1024]="TypeofNEBigInt",e[e.TypeofNEBoolean=2048]="TypeofNEBoolean",e[e.TypeofNESymbol=4096]="TypeofNESymbol",e[e.TypeofNEObject=8192]="TypeofNEObject",e[e.TypeofNEFunction=16384]="TypeofNEFunction",e[e.TypeofNEHostObject=32768]="TypeofNEHostObject",e[e.EQUndefined=65536]="EQUndefined",e[e.EQNull=131072]="EQNull",e[e.EQUndefinedOrNull=262144]="EQUndefinedOrNull",e[e.NEUndefined=524288]="NEUndefined",e[e.NENull=1048576]="NENull",e[e.NEUndefinedOrNull=2097152]="NEUndefinedOrNull",e[e.Truthy=4194304]="Truthy",e[e.Falsy=8388608]="Falsy",e[e.All=16777215]="All",e[e.BaseStringStrictFacts=3735041]="BaseStringStrictFacts",e[e.BaseStringFacts=12582401]="BaseStringFacts",e[e.StringStrictFacts=16317953]="StringStrictFacts",e[e.StringFacts=16776705]="StringFacts",e[e.EmptyStringStrictFacts=12123649]="EmptyStringStrictFacts",e[e.EmptyStringFacts=12582401]="EmptyStringFacts",e[e.NonEmptyStringStrictFacts=7929345]="NonEmptyStringStrictFacts",e[e.NonEmptyStringFacts=16776705]="NonEmptyStringFacts",e[e.BaseNumberStrictFacts=3734786]="BaseNumberStrictFacts",e[e.BaseNumberFacts=12582146]="BaseNumberFacts",e[e.NumberStrictFacts=16317698]="NumberStrictFacts",e[e.NumberFacts=16776450]="NumberFacts",e[e.ZeroNumberStrictFacts=12123394]="ZeroNumberStrictFacts",e[e.ZeroNumberFacts=12582146]="ZeroNumberFacts",e[e.NonZeroNumberStrictFacts=7929090]="NonZeroNumberStrictFacts",e[e.NonZeroNumberFacts=16776450]="NonZeroNumberFacts",e[e.BaseBigIntStrictFacts=3734276]="BaseBigIntStrictFacts",e[e.BaseBigIntFacts=12581636]="BaseBigIntFacts",e[e.BigIntStrictFacts=16317188]="BigIntStrictFacts",e[e.BigIntFacts=16775940]="BigIntFacts",e[e.ZeroBigIntStrictFacts=12122884]="ZeroBigIntStrictFacts",e[e.ZeroBigIntFacts=12581636]="ZeroBigIntFacts",e[e.NonZeroBigIntStrictFacts=7928580]="NonZeroBigIntStrictFacts",e[e.NonZeroBigIntFacts=16775940]="NonZeroBigIntFacts",e[e.BaseBooleanStrictFacts=3733256]="BaseBooleanStrictFacts",e[e.BaseBooleanFacts=12580616]="BaseBooleanFacts",e[e.BooleanStrictFacts=16316168]="BooleanStrictFacts",e[e.BooleanFacts=16774920]="BooleanFacts",e[e.FalseStrictFacts=12121864]="FalseStrictFacts",e[e.FalseFacts=12580616]="FalseFacts",e[e.TrueStrictFacts=7927560]="TrueStrictFacts",e[e.TrueFacts=16774920]="TrueFacts",e[e.SymbolStrictFacts=7925520]="SymbolStrictFacts",e[e.SymbolFacts=16772880]="SymbolFacts",e[e.ObjectStrictFacts=7888800]="ObjectStrictFacts",e[e.ObjectFacts=16736160]="ObjectFacts",e[e.FunctionStrictFacts=7880640]="FunctionStrictFacts",e[e.FunctionFacts=16728e3]="FunctionFacts",e[e.UndefinedFacts=9830144]="UndefinedFacts",e[e.NullFacts=9363232]="NullFacts",e[e.EmptyObjectStrictFacts=16318463]="EmptyObjectStrictFacts",e[e.AllTypeofNE=556800]="AllTypeofNE",e[e.EmptyObjectFacts=16777215]="EmptyObjectFacts"}(a||(a={}));var d,p,m,f,_,g,y,v,h,b=new e.Map(e.getEntries({string:1,number:2,bigint:4,boolean:8,symbol:16,undefined:65536,object:32,function:64})),E=new e.Map(e.getEntries({string:256,number:512,bigint:1024,boolean:2048,symbol:4096,undefined:524288,object:8192,function:16384}));!function(e){e[e.Type=0]="Type",e[e.ResolvedBaseConstructorType=1]="ResolvedBaseConstructorType",e[e.DeclaredType=2]="DeclaredType",e[e.ResolvedReturnType=3]="ResolvedReturnType",e[e.ImmediateBaseConstraint=4]="ImmediateBaseConstraint",e[e.EnumTagType=5]="EnumTagType",e[e.ResolvedTypeArguments=6]="ResolvedTypeArguments",e[e.ResolvedBaseTypes=7]="ResolvedBaseTypes"}(d||(d={})),function(e){e[e.Normal=0]="Normal",e[e.Contextual=1]="Contextual",e[e.Inferential=2]="Inferential",e[e.SkipContextSensitive=4]="SkipContextSensitive",e[e.SkipGenericFunctions=8]="SkipGenericFunctions",e[e.IsForSignatureHelp=16]="IsForSignatureHelp"}(p||(p={})),function(e){e[e.None=0]="None",e[e.NoIndexSignatures=1]="NoIndexSignatures",e[e.Writing=2]="Writing",e[e.CacheSymbol=4]="CacheSymbol",e[e.NoTupleBoundsCheck=8]="NoTupleBoundsCheck",e[e.ExpressionPosition=16]="ExpressionPosition"}(m||(m={})),function(e){e[e.BivariantCallback=1]="BivariantCallback",e[e.StrictCallback=2]="StrictCallback",e[e.IgnoreReturnTypes=4]="IgnoreReturnTypes",e[e.StrictArity=8]="StrictArity",e[e.Callback=3]="Callback"}(f||(f={})),function(e){e[e.None=0]="None",e[e.Source=1]="Source",e[e.Target=2]="Target",e[e.PropertyCheck=4]="PropertyCheck",e[e.InPropertyCheck=8]="InPropertyCheck"}(_||(_={})),function(e){e[e.IncludeReadonly=1]="IncludeReadonly",e[e.ExcludeReadonly=2]="ExcludeReadonly",e[e.IncludeOptional=4]="IncludeOptional",e[e.ExcludeOptional=8]="ExcludeOptional"}(g||(g={})),function(e){e[e.None=0]="None",e[e.Source=1]="Source",e[e.Target=2]="Target",e[e.Both=3]="Both"}(y||(y={})),function(e){e.resolvedExports="resolvedExports",e.resolvedMembers="resolvedMembers"}(v||(v={})),function(e){e[e.Local=0]="Local",e[e.Parameter=1]="Parameter"}(h||(h={}));var T,S,x,L,A=e.and(P,(function(n){return!e.isAccessor(n)}));!function(e){e[e.GetAccessor=1]="GetAccessor",e[e.SetAccessor=2]="SetAccessor",e[e.PropertyAssignment=4]="PropertyAssignment",e[e.Method=8]="Method",e[e.GetOrSetAccessor=3]="GetOrSetAccessor",e[e.PropertyAssignmentOrMethod=12]="PropertyAssignmentOrMethod"}(T||(T={})),function(e){e[e.None=0]="None",e[e.ExportValue=1]="ExportValue",e[e.ExportType=2]="ExportType",e[e.ExportNamespace=4]="ExportNamespace"}(S||(S={})),function(e){e[e.None=0]="None",e[e.StrongArityForUntypedJS=1]="StrongArityForUntypedJS",e[e.VoidIsNonOptional=2]="VoidIsNonOptional"}(x||(x={})),function(e){e[e.Uppercase=0]="Uppercase",e[e.Lowercase=1]="Lowercase",e[e.Capitalize=2]="Capitalize",e[e.Uncapitalize=3]="Uncapitalize"}(L||(L={}));var D,C=new e.Map(e.getEntries({Uppercase:0,Lowercase:1,Capitalize:2,Uncapitalize:3}));function I(){}function N(){this.flags=0}function R(e){return e.id||(e.id=c,c++),e.id}function k(e){return e.id||(e.id=s,s++),e.id}function O(n,t){var r=e.getModuleInstanceState(n);return 1===r||t&&2===r}function P(e){return 251!==e.kind&&165!==e.kind||!!e.body}function M(n){switch(n.parent.kind){case 265:case 270:return e.isIdentifier(n);default:return e.isDeclarationName(n)}}function w(e){switch(e.kind){case 262:case 260:case 263:case 265:return!0;case 78:return 265===e.parent.kind;default:return!1}}function F(e){switch(e){case 0:return"yieldType";case 1:return"returnType";case 2:return"nextType"}}function G(e){return!!(1&e.flags)}function B(e){return!!(2&e.flags)}e.getNodeId=R,e.getSymbolId=k,e.isInstantiatedModule=O,e.createTypeChecker=function(n,t){var r,a,s,c,d=e.memoize((function(){var t=new e.Set;return n.getSourceFiles().forEach((function(n){n.resolvedModules&&e.forEachEntry(n.resolvedModules,(function(e){e&&e.packageId&&t.add(e.packageId.name)}))})),t})),p=e.objectAllocator.getSymbolConstructor(),m=e.objectAllocator.getTypeConstructor(),f=e.objectAllocator.getSignatureConstructor(),_=0,g=0,y=0,v=0,h=0,T=0,S=0,x=[],L=e.createSymbolTable(),U=[1],V=n.getCompilerOptions(),j=e.getEmitScriptTarget(V),H=e.getEmitModuleKind(V),K=e.getAllowSyntheticDefaultImports(V),W=e.getStrictOptionValue(V,"strictNullChecks"),z=e.getStrictOptionValue(V,"strictFunctionTypes"),q=e.getStrictOptionValue(V,"strictBindCallApply"),J=e.getStrictOptionValue(V,"strictPropertyInitialization"),X=e.getStrictOptionValue(V,"noImplicitAny"),Y=e.getStrictOptionValue(V,"noImplicitThis"),Q=!!V.keyofStringsOnly,Z=V.suppressExcessPropertyErrors?0:32768,$=function(){var t,r=n.getResolvedTypeReferenceDirectives();r&&(t=new e.Map,r.forEach((function(e,t){if(e&&e.resolvedFileName){var r=n.getSourceFile(e.resolvedFileName);r&&i(r,t)}})));return{getReferencedExportContainer:NT,getReferencedImportDeclaration:RT,getReferencedDeclarationWithCollidingName:OT,isDeclarationWithCollidingName:PT,isValueAliasDeclaration:function(n){var t=e.getParseTreeNode(n);return!t||MT(t)},hasGlobalName:eS,isReferencedAliasDeclaration:function(n,t){var r=e.getParseTreeNode(n);return!r||BT(r,t)},getNodeCheckFlags:function(n){var t=e.getParseTreeNode(n);return t?WT(t):0},isTopLevelValueImportEqualsWithEntityName:wT,isDeclarationVisible:yi,isImplementationOfOverload:UT,isRequiredInitializedParameter:VT,isOptionalUninitializedParameterProperty:jT,isExpandoFunctionDeclaration:HT,getPropertiesOfContainerFunction:KT,createTypeOfDeclaration:QT,createReturnTypeOfSignatureDeclaration:ZT,createTypeOfExpression:$T,createLiteralConstValue:aS,isSymbolAccessible:Ya,isEntityNameVisible:ti,getConstantValue:function(n){var t=e.getParseTreeNode(n,qT);return t?JT(t):void 0},collectLinkedAliases:vi,getReferencedValueDeclaration:tS,getTypeReferenceSerializationKind:YT,isOptionalParameter:bc,moduleExportsSomeValue:IT,isArgumentsLocalBinding:CT,getExternalModuleFileFromDeclaration:sS,getTypeReferenceDirectivesForEntityName:function(e){if(!t)return;var n=790504;(78===e.kind&&_f(e)||201===e.kind&&!function(e){return e.parent&&223===e.parent.kind&&e.parent.parent&&286===e.parent.parent.kind}(e))&&(n=1160127);var r=ca(e,n,!0);return r&&r!==he?a(r,n):void 0},getTypeReferenceDirectivesForSymbol:a,isLiteralConstDeclaration:rS,isLateBound:function(n){var t=e.getParseTreeNode(n,e.isDeclaration),r=t&&xa(t);return!!(r&&4096&e.getCheckFlags(r))},getJsxFactoryEntity:iS,getJsxFragmentFactoryEntity:oS,getAllAccessorDeclarations:function(n){var t=168===(n=e.getParseTreeNode(n,e.isGetOrSetAccessorDeclaration)).kind?167:168,r=e.getDeclarationOfKind(xa(n),t);return{firstAccessor:r&&r.pos<n.pos?r:n,secondAccessor:r&&r.pos<n.pos?n:r,setAccessor:168===n.kind?n:r,getAccessor:167===n.kind?n:r}},getSymbolOfExternalModuleSpecifier:function(e){return da(e,e,void 0)},isBindingCapturedByNode:function(n,t){var r=e.getParseTreeNode(n),a=e.getParseTreeNode(t);return!!r&&!!a&&(e.isVariableDeclaration(a)||e.isBindingElement(a))&&function(n,t){var r=Sr(n);return!!r&&e.contains(r.capturedBlockScopeBindings,xa(t))}(r,a)},getDeclarationStatementsForSourceFile:function(n,t,r,a){var i=e.getParseTreeNode(n);e.Debug.assert(i&&297===i.kind,"Non-sourcefile node passed into getDeclarationsForSourceFile");var o=xa(n);return o?o.exports?ee.symbolTableToDeclarationStatements(o.exports,n,t,r,a):[]:n.locals?ee.symbolTableToDeclarationStatements(n.locals,n,t,r,a):[]},isImportRequiredByAugmentation:function(n){var t=e.getSourceFileOfNode(n);if(!t.symbol)return!1;var r=sS(n);if(!r)return!1;if(r===t)return!1;for(var a=ba(t.symbol),i=0,o=e.arrayFrom(a.values());i<o.length;i++){var s=o[i];if(s.mergeId)for(var c=0,l=Sa(s).declarations;c<l.length;c++){var u=l[c];if(e.getSourceFileOfNode(u)===r)return!0}}return!1}};function a(n,r){if(t&&function(n){if(!n.declarations)return!1;var r=n;for(;;){var a=La(r);if(!a)break;r=a}if(r.valueDeclaration&&297===r.valueDeclaration.kind&&512&r.flags)return!1;for(var i=0,o=n.declarations;i<o.length;i++){var s=o[i],c=e.getSourceFileOfNode(s);if(t.has(c.path))return!0}return!1}(n)){for(var a,i=0,o=n.declarations;i<o.length;i++){var s=o[i];if(s.symbol&&s.symbol.flags&r){var c=e.getSourceFileOfNode(s),l=t.get(c.path);if(!l)return;(a||(a=[])).push(l)}}return a}}function i(r,a){if(!t.has(r.path)){t.set(r.path,a);for(var o=0,s=r.referencedFiles;o<s.length;o++){var c=s[o].fileName,l=e.resolveTripleslashReference(c,r.fileName),u=n.getSourceFile(l);u&&i(u,a)}}}}(),ee=function(){return{typeToTypeNode:function(e,n,r,a){return t(n,r,a,(function(n){return o(e,n)}))},indexInfoToIndexSignatureDeclaration:function(e,n,r,a,i){return t(r,a,i,(function(t){return u(e,n,t,void 0)}))},signatureToSignatureDeclaration:function(e,n,r,a,i){return t(r,a,i,(function(t){return d(e,n,t)}))},symbolToEntityName:function(e,n,r,a,i){return t(r,a,i,(function(t){return L(e,t,n,!1)}))},symbolToExpression:function(e,n,r,a,i){return t(r,a,i,(function(t){return A(e,t,n)}))},symbolToTypeParameterDeclarations:function(e,n,r,a){return t(n,r,a,(function(n){return v(e,n)}))},symbolToParameterDeclaration:function(e,n,r,a){return t(n,r,a,(function(n){return f(e,n)}))},typeParameterToDeclaration:function(e,n,r,a){return t(n,r,a,(function(n){return m(e,n)}))},symbolTableToDeclarationStatements:function(n,r,a,i,s){return t(r,a,i,(function(t){return function(n,t,r){var a=te(e.factory.createPropertyDeclaration,165,!0),i=te((function(n,t,r,a,i){return e.factory.createPropertySignature(t,r,a,i)}),164,!1),s=t.enclosingDeclaration,c=[],l=new e.Set,p=[],f=t;t=__assign(__assign({},f),{usedSymbolNames:new e.Set(f.usedSymbolNames),remappedSymbolNames:new e.Map,tracker:__assign(__assign({},f.tracker),{trackSymbol:function(e,n,r){if(0===Ya(e,n,r,!1).accessibility){var a=y(e,t,r);4&e.flags||F(a[0])}else f.tracker&&f.tracker.trackSymbol&&f.tracker.trackSymbol(e,n,r)}})}),e.forEachEntry(n,(function(n,t){de(n,e.unescapeLeadingUnderscores(t))}));var _=!r,g=n.get("export=");g&&n.size>1&&2097152&g.flags&&(n=e.createSymbolTable()).set("export=",g);return R(n),x(c);function v(e){return!!e&&78===e.kind}function h(n){return e.isVariableStatement(n)?e.filter(e.map(n.declarationList.declarations,e.getNameOfDeclaration),v):e.filter([e.getNameOfDeclaration(n)],v)}function b(n){var t=e.find(n,e.isExportAssignment),r=e.findIndex(n,e.isModuleDeclaration),a=-1!==r?n[r]:void 0;if(a&&t&&t.isExportEquals&&e.isIdentifier(t.expression)&&e.isIdentifier(a.name)&&e.idText(a.name)===e.idText(t.expression)&&a.body&&e.isModuleBlock(a.body)){var i=e.filter(n,(function(n){return!!(1&e.getEffectiveModifierFlags(n))})),o=a.name,s=a.body;if(e.length(i)&&(a=e.factory.updateModuleDeclaration(a,a.decorators,a.modifiers,a.name,s=e.factory.updateModuleBlock(s,e.factory.createNodeArray(__spreadArrays(a.body.statements,[e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports(e.map(e.flatMap(i,(function(e){return h(e)})),(function(n){return e.factory.createExportSpecifier(void 0,n)}))),void 0)])))),n=__spreadArrays(n.slice(0,r),[a],n.slice(r+1))),!e.find(n,(function(n){return n!==a&&e.nodeHasName(n,o)}))){c=[];var l=!e.some(s.statements,(function(n){return e.hasSyntacticModifier(n,1)||e.isExportAssignment(n)||e.isExportDeclaration(n)}));e.forEach(s.statements,(function(e){B(e,l?1:0)})),n=__spreadArrays(e.filter(n,(function(e){return e!==a&&e!==t})),c)}}return n}function T(n){var t=e.filter(n,(function(n){return e.isExportDeclaration(n)&&!n.moduleSpecifier&&!!n.exportClause&&e.isNamedExports(n.exportClause)}));if(e.length(t)>1){var r=e.filter(n,(function(n){return!e.isExportDeclaration(n)||!!n.moduleSpecifier||!n.exportClause}));n=__spreadArrays(r,[e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports(e.flatMap(t,(function(n){return e.cast(n.exportClause,e.isNamedExports).elements}))),void 0)])}var a=e.filter(n,(function(n){return e.isExportDeclaration(n)&&!!n.moduleSpecifier&&!!n.exportClause&&e.isNamedExports(n.exportClause)}));if(e.length(a)>1){var i=e.group(a,(function(n){return e.isStringLiteral(n.moduleSpecifier)?">"+n.moduleSpecifier.text:">"}));if(i.length!==a.length)for(var o=function(t){t.length>1&&(n=__spreadArrays(e.filter(n,(function(e){return-1===t.indexOf(e)})),[e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports(e.flatMap(t,(function(n){return e.cast(n.exportClause,e.isNamedExports).elements}))),t[0].moduleSpecifier)]))},s=0,c=i;s<c.length;s++){o(c[s])}}return n}function S(n){var t=e.findIndex(n,(function(n){return e.isExportDeclaration(n)&&!n.moduleSpecifier&&!!n.exportClause&&e.isNamedExports(n.exportClause)}));if(t>=0){var r=n[t],a=e.mapDefined(r.exportClause.elements,(function(t){if(!t.propertyName){var r=e.indicesOf(n),a=e.filter(r,(function(r){return e.nodeHasName(n[r],t.name)}));if(e.length(a)&&e.every(a,(function(e){return D(n[e])}))){for(var i=0,o=a;i<o.length;i++){var s=o[i];n[s]=C(n[s])}return}}return t}));e.length(a)?n[t]=e.factory.updateExportDeclaration(r,r.decorators,r.modifiers,r.isTypeOnly,e.factory.updateNamedExports(r.exportClause,a),r.moduleSpecifier):e.orderedRemoveItemAt(n,t)}return n}function x(n){return n=S(n=T(n=b(n))),s&&(e.isSourceFile(s)&&e.isExternalOrCommonJsModule(s)||e.isModuleDeclaration(s))&&(!e.some(n,e.isExternalModuleIndicator)||!e.hasScopeMarker(n)&&e.some(n,e.needsScopeMarker))&&n.push(e.createEmptyExports(e.factory)),n}function D(n){return e.isEnumDeclaration(n)||e.isVariableStatement(n)||e.isFunctionDeclaration(n)||e.isClassDeclaration(n)||e.isModuleDeclaration(n)&&!e.isExternalModuleAugmentation(n)&&!e.isGlobalScopeAugmentation(n)||e.isInterfaceDeclaration(n)||fT(n)}function C(n){var t=-3&(1|e.getEffectiveModifierFlags(n));return e.factory.updateModifiers(n,t)}function N(n){var t=-2&e.getEffectiveModifierFlags(n);return e.factory.updateModifiers(n,t)}function R(n,t,r){t||p.push(new e.Map),n.forEach((function(e){P(e,!1,!!r)})),t||(p[p.length-1].forEach((function(e){P(e,!0,!!r)})),p.pop())}function P(n,r,a){var i=Sa(n);if(!l.has(k(i))&&(l.add(k(i)),!r||e.length(n.declarations)&&e.some(n.declarations,(function(n){return!!e.findAncestor(n,(function(e){return e===s}))})))){var o=t;t=function(n){var t=__assign({},n);t.typeParameterNames&&(t.typeParameterNames=new e.Map(t.typeParameterNames));t.typeParameterNamesByText&&(t.typeParameterNamesByText=new e.Set(t.typeParameterNamesByText));t.typeParameterSymbolList&&(t.typeParameterSymbolList=new e.Set(t.typeParameterSymbolList));return t}(t);var c=w(n,r,a);return t=o,c}}function w(n,a,i){var o=e.unescapeLeadingUnderscores(n.escapedName),c="default"===n.escapedName;if(!a||131072&t.flags||!e.isStringANonContextualKeyword(o)||c){var l=c&&!!(-113&n.flags||16&n.flags&&e.length(Gs(lo(n))))&&!(2097152&n.flags),u=!l&&!a&&e.isStringANonContextualKeyword(o)&&!c;(l||u)&&(a=!0);var d=(a?0:1)|(c&&!l?512:0),p=1536&n.flags&&7&n.flags&&"export="!==n.escapedName,m=p&&ne(lo(n),n);if((8208&n.flags||m)&&J(lo(n),n,de(n,o),d),524288&n.flags&&U(n,o,d),7&n.flags&&"export="!==n.escapedName&&!(4194304&n.flags)&&!(32&n.flags)&&!m)if(i){ee(n)&&(u=!1,l=!1)}else{var f=lo(n),_=de(n,o);if(16&n.flags||!ne(f,n)){var g=2&n.flags?E_(n)?2:1:void 0,y=!l&&4&n.flags?le(_,n):_,v=n.declarations&&e.find(n.declarations,(function(n){return e.isVariableDeclaration(n)}));v&&e.isVariableDeclarationList(v.parent)&&1===v.parent.declarations.length&&(v=v.parent.parent);var h=e.find(n.declarations,e.isPropertyAccessExpression);if(h&&e.isBinaryExpression(h.parent)&&e.isIdentifier(h.parent.right)&&f.symbol&&e.isSourceFile(f.symbol.valueDeclaration)){var b=_===h.parent.right.escapedText?void 0:h.parent.right;B(e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([e.factory.createExportSpecifier(b,_)])),0),t.tracker.trackSymbol(f.symbol,t.enclosingDeclaration,111551)}else{B(e.setTextRange(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(y,void 0,M(t,f,n,s,F,r))],g)),v),y!==_?-2&d:d),y===_||a||(B(e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([e.factory.createExportSpecifier(y,_)])),0),u=!1,l=!1)}}else J(f,n,_,d)}if(384&n.flags&&q(n,o,d),32&n.flags&&(4&n.flags&&e.isBinaryExpression(n.valueDeclaration.parent)&&e.isClassExpression(n.valueDeclaration.parent.right)?Z(n,de(n,o),d):Q(n,de(n,o),d)),(1536&n.flags&&(!p||W(n))||m)&&z(n,o,d),64&n.flags&&!(32&n.flags)&&H(n,o,d),2097152&n.flags&&Z(n,de(n,o),d),4&n.flags&&"export="===n.escapedName&&ee(n),8388608&n.flags)for(var T=0,S=n.declarations;T<S.length;T++){var x=S[T],L=ua(x,x.moduleSpecifier);L&&B(e.factory.createExportDeclaration(void 0,void 0,!1,void 0,e.factory.createStringLiteral(E(L,t))),0)}l?B(e.factory.createExportAssignment(void 0,void 0,!1,e.factory.createIdentifier(de(n,o))),0):u&&B(e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([e.factory.createExportSpecifier(de(n,o),o)])),0)}else t.encounteredError=!0}function F(n){if(!e.some(n.declarations,e.isParameterDeclaration)){e.Debug.assertIsDefined(p[p.length-1]),le(e.unescapeLeadingUnderscores(n.escapedName),n);var t=!!(2097152&n.flags)&&!e.some(n.declarations,(function(n){return!!e.findAncestor(n,e.isExportDeclaration)||e.isNamespaceExport(n)||e.isImportEqualsDeclaration(n)&&!e.isExternalModuleReference(n.moduleReference)}));p[t?0:p.length-1].set(k(n),n)}}function G(n){return e.isSourceFile(n)&&(e.isExternalOrCommonJsModule(n)||e.isJsonSourceFile(n))||e.isAmbientModule(n)&&!e.isGlobalScopeAugmentation(n)}function B(n,r){if(e.canHaveModifiers(n)){var a=0;1&r&&t.enclosingDeclaration&&(G(t.enclosingDeclaration)||e.isModuleDeclaration(t.enclosingDeclaration))&&D(n)&&(a|=1),!_||1&a||t.enclosingDeclaration&&8388608&t.enclosingDeclaration.flags||!(e.isEnumDeclaration(n)||e.isVariableStatement(n)||e.isFunctionDeclaration(n)||e.isClassDeclaration(n)||e.isModuleDeclaration(n))||(a|=2),512&r&&(e.isClassDeclaration(n)||e.isInterfaceDeclaration(n)||e.isFunctionDeclaration(n))&&(a|=512),a&&(n=e.factory.updateModifiers(n,a|e.getEffectiveModifierFlags(n)))}c.push(n)}function U(n,r,a){var i=Io(n),s=Tr(n).typeParameters,c=e.map(s,(function(e){return m(e,t)})),l=e.find(n.declarations,e.isJSDocTypeAlias),u=l?l.comment||l.parent.comment:void 0,d=t.flags;t.flags|=8388608,B(e.setSyntheticLeadingComments(e.factory.createTypeAliasDeclaration(void 0,void 0,de(n,r),c,o(i,t)),u?[{kind:3,text:"*\n * "+u.replace(/\n/g,"\n * ")+"\n ",pos:-1,end:-1,hasTrailingNewLine:!0}]:[]),a),t.flags=d}function H(n,r,a){var i=Co(n),o=yo(n),s=e.map(o,(function(e){return m(e,t)})),c=Lo(i),l=e.length(c)?tu(c):void 0,u=e.flatMap(Gs(i),(function(e){return re(e,l)})),d=ae(0,i,l,169),p=ae(1,i,l,170),f=ie(i,l),_=e.length(c)?[e.factory.createHeritageClause(93,e.mapDefined(c,(function(e){return se(e,111551)})))]:void 0;B(e.factory.createInterfaceDeclaration(void 0,void 0,de(n,r),s,_,__spreadArrays(f,p,d,u)),a)}function K(n){return n.exports?e.filter(e.arrayFrom(n.exports.values()),Y):[]}function W(n){return e.every(K(n),(function(e){return!(111551&$r(e).flags)}))}function z(n,r,a){var i=K(n),o=e.arrayToMultiMap(i,(function(e){return e.parent&&e.parent===n?"real":"merged"})),s=o.get("real")||e.emptyArray,c=o.get("merged")||e.emptyArray;e.length(s)&&X(s,u=de(n,r),a,!!(67108880&n.flags));if(e.length(c)){var l=e.getSourceFileOfNode(t.enclosingDeclaration),u=de(n,r),d=e.factory.createModuleBlock([e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports(e.mapDefined(e.filter(c,(function(e){return"export="!==e.escapedName})),(function(r){var a,i,o=e.unescapeLeadingUnderscores(r.escapedName),s=de(r,o),c=r.declarations&&Br(r);if(!l||(c?l===e.getSourceFileOfNode(c):e.some(r.declarations,(function(n){return e.getSourceFileOfNode(n)===l})))){var u=c&&Qr(c,!0);F(u||r);var d=u?de(u,e.unescapeLeadingUnderscores(u.escapedName)):s;return e.factory.createExportSpecifier(o===d?void 0:d,o)}null===(i=null===(a=t.tracker)||void 0===a?void 0:a.reportNonlocalAugmentation)||void 0===i||i.call(a,l,n,r)}))))]);B(e.factory.createModuleDeclaration(void 0,void 0,e.factory.createIdentifier(u),d,16),0)}}function q(n,t,r){B(e.factory.createEnumDeclaration(void 0,e.factory.createModifiersFromModifierFlags(_h(n)?2048:0),de(n,t),e.map(e.filter(Gs(lo(n)),(function(e){return!!(8&e.flags)})),(function(n){var t=n.declarations&&n.declarations[0]&&e.isEnumMember(n.declarations[0])?JT(n.declarations[0]):void 0;return e.factory.createEnumMember(e.unescapeLeadingUnderscores(n.escapedName),void 0===t?void 0:"string"==typeof t?e.factory.createStringLiteral(t):e.factory.createNumericLiteral(t))}))),r)}function J(n,a,i,o){for(var s=0,c=uc(n,0);s<c.length;s++){var l=c[s],u=d(l,251,t,{name:e.factory.createIdentifier(i),privateSymbolVisitor:F,bundledImports:r});B(e.setTextRange(u,l.declaration&&e.isVariableDeclaration(l.declaration.parent)&&l.declaration.parent.parent||l.declaration),o)}1536&a.flags&&a.exports&&a.exports.size||X(e.filter(Gs(n),Y),i,o,!0)}function X(n,r,a,i){if(e.length(n)){var o=e.arrayToMultiMap(n,(function(n){return!e.length(n.declarations)||e.some(n.declarations,(function(n){return e.getSourceFileOfNode(n)===e.getSourceFileOfNode(t.enclosingDeclaration)}))?"local":"remote"})).get("local")||e.emptyArray,l=e.parseNodeFactory.createModuleDeclaration(void 0,void 0,e.factory.createIdentifier(r),e.factory.createModuleBlock([]),16);e.setParent(l,s),l.locals=e.createSymbolTable(n),l.symbol=n[0].parent;var u=c;c=[];var d=_;_=!1;var p=__assign(__assign({},t),{enclosingDeclaration:l}),m=t;t=p,R(e.createSymbolTable(o),i,!0),t=m,_=d;var f=c;c=u;var g=e.map(f,(function(n){return e.isExportAssignment(n)&&!n.isExportEquals&&e.isIdentifier(n.expression)?e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([e.factory.createExportSpecifier(n.expression,e.factory.createIdentifier("default"))])):n})),y=e.every(g,(function(n){return e.hasSyntacticModifier(n,1)}))?e.map(g,N):g;B(l=e.factory.updateModuleDeclaration(l,l.decorators,l.modifiers,l.name,e.factory.createModuleBlock(y)),a)}}function Y(n){return!!(2887656&n.flags)||!(4194304&n.flags||"prototype"===n.escapedName||n.valueDeclaration&&32&e.getEffectiveModifierFlags(n.valueDeclaration)&&e.isClassLike(n.valueDeclaration.parent))}function Q(n,r,i){var o,s=yo(n),c=e.map(s,(function(e){return m(e,t)})),l=Co(n),u=Lo(l),d=e.mapDefined(function(n){for(var t=e.emptyArray,r=0,a=n.symbol.declarations;r<a.length;r++){var i=a[r],o=e.getEffectiveImplementsTypeNodes(i);if(o)for(var s=0,c=o;s<c.length;s++){var l=sd(c[s]);l!==xe&&(t===e.emptyArray?t=[l]:t.push(l))}}return t}(l),ce),p=lo(n),f=!!(null===(o=p.symbol)||void 0===o?void 0:o.valueDeclaration)&&e.isClassLike(p.symbol.valueDeclaration),_=f?So(p):Ee,g=__spreadArrays(e.length(u)?[e.factory.createHeritageClause(93,e.map(u,(function(e){return oe(e,_,r)})))]:[],e.length(d)?[e.factory.createHeritageClause(116,d)]:[]),y=function(n,t,r){if(!e.length(t))return r;var a=new e.Map;e.forEach(r,(function(e){a.set(e.escapedName,e)}));for(var i=0,o=t;i<o.length;i++)for(var s=0,c=Gs(rs(o[i],n.thisType));s<c.length;s++){var l=c[s],u=a.get(l.escapedName);u&&!wp(u,l)&&a.delete(l.escapedName)}return e.arrayFrom(a.values())}(l,u,Gs(l)),v=e.filter(y,(function(n){var t=n.valueDeclaration;return t&&!(e.isNamedDeclaration(t)&&e.isPrivateIdentifier(t.name))})),h=e.some(y,(function(n){var t=n.valueDeclaration;return t&&e.isNamedDeclaration(t)&&e.isPrivateIdentifier(t.name)}))?[e.factory.createPropertyDeclaration(void 0,void 0,e.factory.createPrivateIdentifier("#private"),void 0,void 0,void 0)]:e.emptyArray,b=e.flatMap(v,(function(e){return a(e,!1,u[0])})),E=e.flatMap(e.filter(Gs(p),(function(e){return!(4194304&e.flags||"prototype"===e.escapedName||Y(e))})),(function(e){return a(e,!0,_)})),T=!f&&!!n.valueDeclaration&&e.isInJSFile(n.valueDeclaration)&&!e.some(uc(p,1))?[e.factory.createConstructorDeclaration(void 0,e.factory.createModifiersFromModifierFlags(8),[],void 0)]:ae(1,p,u[0],166),S=ie(l,u[0]);B(e.setTextRange(e.factory.createClassDeclaration(void 0,void 0,r,c,g,__spreadArrays(S,E,T,b,h)),n.declarations&&e.filter(n.declarations,(function(n){return e.isClassDeclaration(n)||e.isClassExpression(n)}))[0]),i)}function Z(n,r,a){var i,o,s,c,l,u=Br(n);if(!u)return e.Debug.fail();var d=Sa(Qr(u,!0));if(d){var p=e.unescapeLeadingUnderscores(d.escapedName);"export="===p&&(V.esModuleInterop||V.allowSyntheticDefaultImports)&&(p="default");var m=de(d,p);switch(F(d),u.kind){case 198:if(249===(null===(o=null===(i=u.parent)||void 0===i?void 0:i.parent)||void 0===o?void 0:o.kind)){var f=E(d.parent||d,t),_=u.propertyName;B(e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,void 0,e.factory.createNamedImports([e.factory.createImportSpecifier(_&&e.isIdentifier(_)?e.factory.createIdentifier(e.idText(_)):void 0,e.factory.createIdentifier(r))])),e.factory.createStringLiteral(f)),0);break}e.Debug.failBadSyntaxKind((null===(s=u.parent)||void 0===s?void 0:s.parent)||u,"Unhandled binding element grandparent kind in declaration serialization");break;case 289:216===(null===(l=null===(c=u.parent)||void 0===c?void 0:c.parent)||void 0===l?void 0:l.kind)&&$(e.unescapeLeadingUnderscores(n.escapedName),m);break;case 249:if(e.isPropertyAccessExpression(u.initializer)){var g=u.initializer,y=e.factory.createUniqueName(r),v=E(d.parent||d,t);B(e.factory.createImportEqualsDeclaration(void 0,void 0,y,e.factory.createExternalModuleReference(e.factory.createStringLiteral(v))),0),B(e.factory.createImportEqualsDeclaration(void 0,void 0,e.factory.createIdentifier(r),e.factory.createQualifiedName(y,g.name)),a);break}case 260:if("export="===d.escapedName&&e.some(d.declarations,e.isJsonSourceFile)){ee(n);break}var h=!(512&d.flags||e.isVariableDeclaration(u));B(e.factory.createImportEqualsDeclaration(void 0,void 0,e.factory.createIdentifier(r),h?L(d,t,67108863,!1):e.factory.createExternalModuleReference(e.factory.createStringLiteral(E(d,t)))),h?a:0);break;case 259:B(e.factory.createNamespaceExportDeclaration(e.idText(u.name)),0);break;case 262:B(e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,e.factory.createIdentifier(r),void 0),e.factory.createStringLiteral(E(d.parent||d,t))),0);break;case 263:B(e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,void 0,e.factory.createNamespaceImport(e.factory.createIdentifier(r))),e.factory.createStringLiteral(E(d,t))),0);break;case 269:B(e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamespaceExport(e.factory.createIdentifier(r)),e.factory.createStringLiteral(E(d,t))),0);break;case 265:B(e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,void 0,e.factory.createNamedImports([e.factory.createImportSpecifier(r!==p?e.factory.createIdentifier(p):void 0,e.factory.createIdentifier(r))])),e.factory.createStringLiteral(E(d.parent||d,t))),0);break;case 270:var b=u.parent.parent.moduleSpecifier;$(e.unescapeLeadingUnderscores(n.escapedName),b?p:m,b&&e.isStringLiteralLike(b)?e.factory.createStringLiteral(b.text):void 0);break;case 266:ee(n);break;case 216:case 201:case 202:"default"===n.escapedName||"export="===n.escapedName?ee(n):$(r,m);break;default:return e.Debug.failBadSyntaxKind(u,"Unhandled alias declaration kind in symbol serializer!")}}}function $(n,t,r){B(e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([e.factory.createExportSpecifier(n!==t?t:void 0,n)]),r),0)}function ee(n){if(4194304&n.flags)return!1;var a=e.unescapeLeadingUnderscores(n.escapedName),i="export="===a,o=i||"default"===a,l=n.declarations&&Br(n),u=l&&Qr(l,!0);if(u&&e.length(u.declarations)&&e.some(u.declarations,(function(n){return e.getSourceFileOfNode(n)===e.getSourceFileOfNode(s)}))){var d=l&&(e.isExportAssignment(l)||e.isBinaryExpression(l)?e.getExportAssignmentExpression(l):e.getPropertyAssignmentAliasLikeExpression(l)),p=d&&e.isEntityNameExpression(d)?function(n){switch(n.kind){case 78:return n;case 157:do{n=n.left}while(78!==n.kind);return n;case 201:do{if(e.isModuleExportsAccessExpression(n.expression)&&!e.isPrivateIdentifier(n.name))return n.name;n=n.expression}while(78!==n.kind);return n}}(d):void 0,m=p&&ca(p,67108863,!0,!0,s);(m||u)&&F(m||u);var f=t.tracker.trackSymbol;if(t.tracker.trackSymbol=e.noop,o)c.push(e.factory.createExportAssignment(void 0,void 0,i,A(u,t,67108863)));else if(p===d&&p)$(a,e.idText(p));else if(d&&e.isClassExpression(d))$(a,de(u,e.symbolName(u)));else{var _=le(a,n);B(e.factory.createImportEqualsDeclaration(void 0,void 0,e.factory.createIdentifier(_),L(u,t,67108863,!1)),0),$(a,_)}return t.tracker.trackSymbol=f,!0}_=le(a,n);var g=Im(lo(Sa(n)));return ne(g,n)?J(g,n,_,o?0:1):B(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(_,void 0,M(t,g,n,s,F,r))],2)),u&&4&u.flags&&"export="===u.escapedName?2:a===_?1:0),o?(c.push(e.factory.createExportAssignment(void 0,void 0,i,e.factory.createIdentifier(_))),!0):a!==_&&($(a,_),!0)}function ne(n,r){var a=e.getSourceFileOfNode(t.enclosingDeclaration);return 48&e.getObjectFlags(n)&&!mc(n,0)&&!mc(n,1)&&!ui(n)&&!(!e.length(e.filter(Gs(n),Y))&&!e.length(uc(n,0)))&&!e.length(uc(n,1))&&!O(r,s)&&!(n.symbol&&e.some(n.symbol.declarations,(function(n){return e.getSourceFileOfNode(n)!==a})))&&!e.some(Gs(n),(function(e){return Xo(e.escapedName)}))&&!e.some(Gs(n),(function(n){return e.some(n.declarations,(function(n){return e.getSourceFileOfNode(n)!==a}))}))&&e.every(Gs(n),(function(n){return e.isIdentifierText(e.symbolName(n),j)}))}function te(n,a,i){return function(o,c,l){var u=e.getDeclarationModifierFlagsFromSymbol(o),p=!!(8&u);if(c&&2887656&o.flags)return[];if(4194304&o.flags||l&&cc(l,o.escapedName)&&oh(cc(l,o.escapedName))===oh(o)&&(16777216&o.flags)==(16777216&cc(l,o.escapedName).flags)&&Vd(lo(o),xi(l,o.escapedName)))return[];var m=-257&u|(c?32:0),f=I(o,t),_=e.find(o.declarations,e.or(e.isPropertyDeclaration,e.isAccessor,e.isVariableDeclaration,e.isPropertySignature,e.isBinaryExpression,e.isPropertyAccessExpression));if(98304&o.flags&&i){var g=[];if(65536&o.flags&&g.push(e.setTextRange(e.factory.createSetAccessorDeclaration(void 0,e.factory.createModifiersFromModifierFlags(m),f,[e.factory.createParameterDeclaration(void 0,void 0,void 0,"arg",void 0,p?void 0:M(t,lo(o),o,s,F,r))],void 0),e.find(o.declarations,e.isSetAccessor)||_)),32768&o.flags){var y=8&u;g.push(e.setTextRange(e.factory.createGetAccessorDeclaration(void 0,e.factory.createModifiersFromModifierFlags(m),f,[],y?void 0:M(t,lo(o),o,s,F,r),void 0),e.find(o.declarations,e.isGetAccessor)||_))}return g}if(98311&o.flags)return e.setTextRange(n(void 0,e.factory.createModifiersFromModifierFlags((oh(o)?64:0)|m),f,16777216&o.flags?e.factory.createToken(57):void 0,p?void 0:M(t,lo(o),o,s,F,r),void 0),e.find(o.declarations,e.or(e.isPropertyDeclaration,e.isVariableDeclaration))||_);if(8208&o.flags){var v=uc(lo(o),0);if(8&m)return e.setTextRange(n(void 0,e.factory.createModifiersFromModifierFlags((oh(o)?64:0)|m),f,16777216&o.flags?e.factory.createToken(57):void 0,void 0,void 0),e.find(o.declarations,e.isFunctionLikeDeclaration)||v[0]&&v[0].declaration||o.declarations[0]);for(var h=[],b=0,E=v;b<E.length;b++){var T=E[b],S=d(T,a,t,{name:f,questionToken:16777216&o.flags?e.factory.createToken(57):void 0,modifiers:m?e.factory.createModifiersFromModifierFlags(m):void 0});h.push(e.setTextRange(S,T.declaration))}return h}return e.Debug.fail("Unhandled class member kind! "+(o.__debugFlags||o.flags))}}function re(e,n){return i(e,!1,n)}function ae(n,r,a,i){var o=uc(r,n);if(1===n){if(!a&&e.every(o,(function(n){return 0===e.length(n.parameters)})))return[];if(a){var s=uc(a,1);if(!e.length(s)&&e.every(o,(function(n){return 0===e.length(n.parameters)})))return[];if(s.length===o.length){for(var c=!1,l=0;l<s.length;l++)if(!Gp(o[l],s[l],!1,!1,!0,jd)){c=!0;break}if(!c)return[]}}for(var u=0,p=0,m=o;p<m.length;p++){var f=m[p];f.declaration&&(u|=e.getSelectedEffectiveModifierFlags(f.declaration,24))}if(u)return[e.setTextRange(e.factory.createConstructorDeclaration(void 0,e.factory.createModifiersFromModifierFlags(u),[],void 0),o[0].declaration)]}for(var _=[],g=0,y=o;g<y.length;g++){var v=y[g],h=d(v,i,t);_.push(e.setTextRange(h,v.declaration))}return _}function ie(e,n){for(var r=[],a=0,i=[0,1];a<i.length;a++){var o=i[a],s=mc(e,o);if(s){if(n){var c=mc(n,o);if(c&&Vd(s.type,c.type))continue}r.push(u(s,o,t,void 0))}}return r}function oe(n,r,a){var i=se(n,111551);if(i)return i;var s=le(a+"_base");return B(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(s,void 0,o(r,t))],2)),0),e.factory.createExpressionWithTypeArguments(e.factory.createIdentifier(s),void 0)}function se(n,r){var a,i;if(n.target&&Ja(n.target.symbol,s,r)?(a=e.map(nl(n),(function(e){return o(e,t)})),i=A(n.target.symbol,t,788968)):n.symbol&&Ja(n.symbol,s,r)&&(i=A(n.symbol,t,788968)),i)return e.factory.createExpressionWithTypeArguments(i,a)}function ce(n){var r=se(n,788968);return r||(n.symbol?e.factory.createExpressionWithTypeArguments(A(n.symbol,t,788968),void 0):void 0)}function le(e,n){var r,a,i=n?k(n):void 0;if(i&&t.remappedSymbolNames.has(i))return t.remappedSymbolNames.get(i);n&&(e=ue(n,e));for(var o=0,s=e;null===(r=t.usedSymbolNames)||void 0===r?void 0:r.has(e);)e=s+"_"+ ++o;return null===(a=t.usedSymbolNames)||void 0===a||a.add(e),i&&t.remappedSymbolNames.set(i,e),e}function ue(n,r){if("default"===r||"__class"===r||"__function"===r){var a=t.flags;t.flags|=16777216;var i=gi(n,t);t.flags=a,r=i.length>0&&e.isSingleOrDoubleQuote(i.charCodeAt(0))?e.stripQuotes(i):i}return"default"===r?r="_default":"export="===r&&(r="_exports"),r=e.isIdentifierText(r,j)&&!e.isStringANonContextualKeyword(r)?r:"_"+r.replace(/[^a-zA-Z0-9]/g,"_")}function de(e,n){var r=k(e);return t.remappedSymbolNames.has(r)?t.remappedSymbolNames.get(r):(n=ue(e,n),t.remappedSymbolNames.set(r,n),n)}}(n,t,s)}))}};function t(t,r,a,i){var o,s;e.Debug.assert(void 0===t||0==(8&t.flags));var c={enclosingDeclaration:t,flags:r||0,tracker:a&&a.trackSymbol?a:{trackSymbol:e.noop,moduleResolverHost:134217728&r?{getCommonSourceDirectory:n.getCommonSourceDirectory?function(){return n.getCommonSourceDirectory()}:function(){return""},getSourceFiles:function(){return n.getSourceFiles()},getCurrentDirectory:function(){return n.getCurrentDirectory()},getSymlinkCache:e.maybeBind(n,n.getSymlinkCache),useCaseSensitiveFileNames:e.maybeBind(n,n.useCaseSensitiveFileNames),redirectTargetsMap:n.redirectTargetsMap,getProjectReferenceRedirect:function(e){return n.getProjectReferenceRedirect(e)},isSourceOfProjectReferenceRedirect:function(e){return n.isSourceOfProjectReferenceRedirect(e)},fileExists:function(e){return n.fileExists(e)}}:void 0},encounteredError:!1,visitedTypes:void 0,symbolDepth:void 0,inferTypeParameters:void 0,approximateLength:0},l=i(c);return c.truncating&&1&c.flags&&(null===(s=null===(o=c.tracker)||void 0===o?void 0:o.reportTruncationError)||void 0===s||s.call(o)),c.encounteredError?void 0:l}function a(n){return n.truncating?n.truncating:n.truncating=n.approximateLength>(1&n.flags?e.noTruncationMaximumTruncationLength:e.defaultMaximumTruncationLength)}function o(n,t){r&&r.throwIfCancellationRequested&&r.throwIfCancellationRequested();var i=8388608&t.flags;if(t.flags&=-8388609,!n)return 262144&t.flags?(t.approximateLength+=3,e.factory.createKeywordTypeNode(128)):void(t.encounteredError=!0);if(536870912&t.flags||(n=rc(n)),1&n.flags)return t.approximateLength+=3,e.factory.createKeywordTypeNode(n===Ae?136:128);if(2&n.flags)return e.factory.createKeywordTypeNode(152);if(4&n.flags)return t.approximateLength+=6,e.factory.createKeywordTypeNode(147);if(8&n.flags)return t.approximateLength+=6,e.factory.createKeywordTypeNode(144);if(64&n.flags)return t.approximateLength+=6,e.factory.createKeywordTypeNode(155);if(16&n.flags)return t.approximateLength+=7,e.factory.createKeywordTypeNode(131);if(1024&n.flags&&!(1048576&n.flags)){var m=La(n.symbol),f=T(m,t,788968);if(Fo(m)===n)return f;var _=e.symbolName(n.symbol);return e.isIdentifierText(_,0)?K(f,e.factory.createTypeReferenceNode(_,void 0)):e.isImportTypeNode(f)?(f.isTypeOf=!0,e.factory.createIndexedAccessTypeNode(f,e.factory.createLiteralTypeNode(e.factory.createStringLiteral(_)))):e.isTypeReferenceNode(f)?e.factory.createIndexedAccessTypeNode(e.factory.createTypeQueryNode(f.typeName),e.factory.createLiteralTypeNode(e.factory.createStringLiteral(_))):e.Debug.fail("Unhandled type node kind returned from `symbolToTypeNode`.")}if(1056&n.flags)return T(n.symbol,t,788968);if(128&n.flags)return t.approximateLength+=n.value.length+2,e.factory.createLiteralTypeNode(e.setEmitFlags(e.factory.createStringLiteral(n.value,!!(268435456&t.flags)),16777216));if(256&n.flags){var g=n.value;return t.approximateLength+=(""+g).length,e.factory.createLiteralTypeNode(g<0?e.factory.createPrefixUnaryExpression(40,e.factory.createNumericLiteral(-g)):e.factory.createNumericLiteral(g))}if(2048&n.flags)return t.approximateLength+=e.pseudoBigIntToString(n.value).length+1,e.factory.createLiteralTypeNode(e.factory.createBigIntLiteral(n.value));if(512&n.flags)return t.approximateLength+=n.intrinsicName.length,e.factory.createLiteralTypeNode("true"===n.intrinsicName?e.factory.createTrue():e.factory.createFalse());if(8192&n.flags){if(!(1048576&t.flags)){if(qa(n.symbol,t.enclosingDeclaration))return t.approximateLength+=6,T(n.symbol,t,111551);t.tracker.reportInaccessibleUniqueSymbolError&&t.tracker.reportInaccessibleUniqueSymbolError()}return t.approximateLength+=13,e.factory.createTypeOperatorNode(151,e.factory.createKeywordTypeNode(148))}if(16384&n.flags)return t.approximateLength+=4,e.factory.createKeywordTypeNode(113);if(32768&n.flags)return t.approximateLength+=9,e.factory.createKeywordTypeNode(150);if(65536&n.flags)return t.approximateLength+=4,e.factory.createLiteralTypeNode(e.factory.createNull());if(131072&n.flags)return t.approximateLength+=5,e.factory.createKeywordTypeNode(141);if(4096&n.flags)return t.approximateLength+=6,e.factory.createKeywordTypeNode(148);if(67108864&n.flags)return t.approximateLength+=6,e.factory.createKeywordTypeNode(145);if(xu(n))return 4194304&t.flags&&(t.encounteredError||32768&t.flags||(t.encounteredError=!0),t.tracker.reportInaccessibleThisError&&t.tracker.reportInaccessibleThisError()),t.approximateLength+=4,e.factory.createThisTypeNode();if(!i&&n.aliasSymbol&&(16384&t.flags||za(n.aliasSymbol,t.enclosingDeclaration))){var y=l(n.aliasTypeArguments,t);return!Ga(n.aliasSymbol.escapedName)||32&n.aliasSymbol.flags?T(n.aliasSymbol,t,788968,y):e.factory.createTypeReferenceNode(e.factory.createIdentifier(""),y)}var v=e.getObjectFlags(n);if(4&v)return e.Debug.assert(!!(524288&n.flags)),n.node?V(n,H):H(n);if(262144&n.flags||3&v){if(262144&n.flags&&e.contains(t.inferTypeParameters,n))return t.approximateLength+=e.symbolName(n.symbol).length+6,e.factory.createInferTypeNode(p(n,t,void 0));if(4&t.flags&&262144&n.flags&&!za(n.symbol,t.enclosingDeclaration)){var h=x(n,t);return t.approximateLength+=e.idText(h).length,e.factory.createTypeReferenceNode(e.factory.createIdentifier(e.idText(h)),void 0)}return n.symbol?T(n.symbol,t,788968):e.factory.createTypeReferenceNode(e.factory.createIdentifier("?"),void 0)}if(3145728&n.flags){var b=1048576&n.flags?function(e){for(var n=[],t=0,r=0;r<e.length;r++){var a=e[r];if(t|=a.flags,!(98304&a.flags)){if(1536&a.flags){var i=512&a.flags?Ue:Oo(a);if(1048576&i.flags){var o=i.types.length;if(r+o<=e.length&&ed(e[r+o-1])===ed(i.types[o-1])){n.push(i),r+=o-1;continue}}}n.push(a)}}65536&t&&n.push(Re);32768&t&&n.push(Ce);return n||e}(n.types):n.types;if(1===e.length(b))return o(b[0],t);var E=l(b,t,!0);return E&&E.length>0?1048576&n.flags?e.factory.createUnionTypeNode(E):e.factory.createIntersectionTypeNode(E):void(t.encounteredError||262144&t.flags||(t.encounteredError=!0))}if(48&v)return e.Debug.assert(!!(524288&n.flags)),U(n);if(4194304&n.flags){var S=n.type;t.approximateLength+=6;var L=o(S,t);return e.factory.createTypeOperatorNode(138,L)}if(134217728&n.flags){var A=n.texts,D=n.types,C=e.factory.createTemplateHead(A[0]),I=e.factory.createNodeArray(e.map(D,(function(n,r){return e.factory.createTemplateLiteralTypeSpan(o(n,t),(r<D.length-1?e.factory.createTemplateMiddle:e.factory.createTemplateTail)(A[r+1]))})));return t.approximateLength+=2,e.factory.createTemplateLiteralType(C,I)}if(268435456&n.flags){var N=o(n.type,t);return T(n.symbol,t,788968,[N])}if(8388608&n.flags){var O=o(n.objectType,t);L=o(n.indexType,t);return t.approximateLength+=2,e.factory.createIndexedAccessTypeNode(O,L)}if(16777216&n.flags){var P=o(n.checkType,t),M=t.inferTypeParameters;t.inferTypeParameters=n.root.inferTypeParameters;var w=o(n.extendsType,t);t.inferTypeParameters=M;var F=B(wu(n)),G=B(Fu(n));return t.approximateLength+=15,e.factory.createConditionalTypeNode(P,w,F,G)}return 33554432&n.flags?o(n.baseType,t):e.Debug.fail("Should be unreachable.");function B(e){var n,r,a;return 1048576&e.flags?(null===(n=t.visitedTypes)||void 0===n?void 0:n.has(Hl(e)))?(131072&t.flags||(t.encounteredError=!0,null===(a=null===(r=t.tracker)||void 0===r?void 0:r.reportCyclicStructureError)||void 0===a||a.call(r)),s(t)):V(e,(function(e){return o(e,t)})):o(e,t)}function U(n){var r,a=n.id,i=n.symbol;if(i){var o=ui(n)?788968:111551;if(fv(i.valueDeclaration))return T(i,t,o);if(32&i.flags&&!io(i)&&!(221===i.valueDeclaration.kind&&2048&t.flags)||896&i.flags||function(){var n,r=!!(8192&i.flags)&&e.some(i.declarations,(function(n){return e.hasSyntacticModifier(n,32)})),o=!!(16&i.flags)&&(i.parent||e.forEach(i.declarations,(function(e){return 297===e.parent.kind||257===e.parent.kind})));if(r||o)return(!!(4096&t.flags)||(null===(n=t.visitedTypes)||void 0===n?void 0:n.has(a)))&&(!(8&t.flags)||qa(i,t.enclosingDeclaration))}())return T(i,t,o);if(null===(r=t.visitedTypes)||void 0===r?void 0:r.has(a)){var c=function(n){if(n.symbol&&2048&n.symbol.flags){var t=e.walkUpParenthesizedTypes(n.symbol.declarations[0].parent);if(254===t.kind)return xa(t)}return}(n);return c?T(c,t,788968):s(t)}return V(n,j)}return j(n)}function V(n,r){var a,i=n.id,o=16&e.getObjectFlags(n)&&n.symbol&&32&n.symbol.flags,c=4&e.getObjectFlags(n)&&n.node?"N"+R(n.node):n.symbol?(o?"+":"")+k(n.symbol):void 0;if(t.visitedTypes||(t.visitedTypes=new e.Set),c&&!t.symbolDepth&&(t.symbolDepth=new e.Map),c){if((a=t.symbolDepth.get(c)||0)>10)return s(t);t.symbolDepth.set(c,a+1)}t.visitedTypes.add(i);var l=r(n);return t.visitedTypes.delete(i),c&&t.symbolDepth.set(c,a),l}function j(n){if(Os(n)||n.containsError)return function(n){e.Debug.assert(!!(524288&n.flags));var r,a=n.declaration.readonlyToken?e.factory.createToken(n.declaration.readonlyToken.kind):void 0,i=n.declaration.questionToken?e.factory.createToken(n.declaration.questionToken.kind):void 0;r=Cs(n)?e.factory.createTypeOperatorNode(138,o(Is(n),t)):o(xs(n),t);var s=p(Ss(n),t,r),c=n.declaration.nameType?o(Ls(n),t):void 0,l=o(As(n),t),u=e.factory.createMappedTypeNode(a,s,c,i,l);return t.approximateLength+=10,e.setEmitFlags(u,1)}(n);var r=Ps(n);if(!r.properties.length&&!r.stringIndexInfo&&!r.numberIndexInfo){if(!r.callSignatures.length&&!r.constructSignatures.length)return t.approximateLength+=2,e.setEmitFlags(e.factory.createTypeLiteralNode(void 0),1);if(1===r.callSignatures.length&&!r.constructSignatures.length)return d(r.callSignatures[0],174,t);if(1===r.constructSignatures.length&&!r.callSignatures.length)return d(r.constructSignatures[0],175,t)}var i=t.flags;t.flags|=4194304;var l=function(n){if(a(t))return[e.factory.createPropertySignature(void 0,"...",void 0,void 0)];for(var r=[],i=0,o=n.callSignatures;i<o.length;i++){var l=o[i];r.push(d(l,169,t))}for(var p=0,m=n.constructSignatures;p<m.length;p++){l=m[p];r.push(d(l,170,t))}if(n.stringIndexInfo){var f=void 0;f=2048&n.objectFlags?u(Wc(Ee,n.stringIndexInfo.isReadonly,n.stringIndexInfo.declaration),0,t,s(t)):u(n.stringIndexInfo,0,t,void 0),r.push(f)}n.numberIndexInfo&&r.push(u(n.numberIndexInfo,1,t,void 0));var _=n.properties;if(!_)return r;for(var g=0,y=0,v=_;y<v.length;y++){var h=v[y];if(g++,2048&t.flags){if(4194304&h.flags)continue;24&e.getDeclarationModifierFlagsFromSymbol(h)&&t.tracker.reportPrivateInBaseOfClassExpression&&t.tracker.reportPrivateInBaseOfClassExpression(e.unescapeLeadingUnderscores(h.escapedName))}if(a(t)&&g+2<_.length-1){r.push(e.factory.createPropertySignature(void 0,"... "+(_.length-g)+" more ...",void 0,void 0)),c(_[_.length-1],t,r);break}c(h,t,r)}return r.length?r:void 0}(r);t.flags=i;var m=e.factory.createTypeLiteralNode(l);return t.approximateLength+=2,e.setEmitFlags(m,1024&t.flags?0:1)}function H(n){var r=nl(n);if(n.target===En||n.target===Tn){if(2&t.flags){var a=o(r[0],t);return e.factory.createTypeReferenceNode(n.target===En?"Array":"ReadonlyArray",[a])}var i=o(r[0],t),s=e.factory.createArrayTypeNode(i);return n.target===En?s:e.factory.createTypeOperatorNode(142,s)}if(!(8&n.target.objectFlags)){if(2048&t.flags&&n.symbol.valueDeclaration&&e.isClassLike(n.symbol.valueDeclaration)&&!qa(n.symbol,t.enclosingDeclaration))return U(n);var c=n.target.outerTypeParameters,u=(S=0,void 0);if(c)for(var d=c.length;S<d;){var p=S,m=Xc(c[S]);do{S++}while(S<d&&Xc(c[S])===m);if(!e.rangeEquals(c,r,p,S)){var f=l(r.slice(p,S),t),_=t.flags;t.flags|=16;var g=T(m,t,788968,f);t.flags=_,u=u?K(u,g):g}}var y=void 0;if(r.length>0){var v=(n.target.typeParameters||e.emptyArray).length;y=l(r.slice(S,v),t)}x=t.flags;t.flags|=16;var h=T(n.symbol,t,788968,y);return t.flags=x,u?K(u,h):h}if(r.length>0){var b=tl(n),E=l(r.slice(0,b),t);if(E){if(n.target.labeledElementDeclarations)for(var S=0;S<E.length;S++){var x=n.target.elementFlags[S];E[S]=e.factory.createNamedTupleMember(12&x?e.factory.createToken(25):void 0,e.factory.createIdentifier(e.unescapeLeadingUnderscores(Iv(n.target.labeledElementDeclarations[S]))),2&x?e.factory.createToken(57):void 0,4&x?e.factory.createArrayTypeNode(E[S]):E[S])}else for(var S=0;S<Math.min(b,E.length);S++){var x=n.target.elementFlags[S];E[S]=12&x?e.factory.createRestTypeNode(4&x?e.factory.createArrayTypeNode(E[S]):E[S]):2&x?e.factory.createOptionalTypeNode(E[S]):E[S]}var L=e.setEmitFlags(e.factory.createTupleTypeNode(E),1);return n.target.readonly?e.factory.createTypeOperatorNode(142,L):L}}if(t.encounteredError||524288&t.flags){L=e.setEmitFlags(e.factory.createTupleTypeNode([]),1);return n.target.readonly?e.factory.createTypeOperatorNode(142,L):L}t.encounteredError=!0}function K(n,t){if(e.isImportTypeNode(n)){var r=n.typeArguments,a=n.qualifier;a&&(a=e.isIdentifier(a)?e.factory.updateIdentifier(a,r):e.factory.updateQualifiedName(a,a.left,e.factory.updateIdentifier(a.right,r))),r=t.typeArguments;for(var i=0,o=W(t);i<o.length;i++){var s=o[i];a=a?e.factory.createQualifiedName(a,s):s}return e.factory.updateImportTypeNode(n,n.argument,a,r,n.isTypeOf)}r=n.typeArguments;var c=n.typeName;c=e.isIdentifier(c)?e.factory.updateIdentifier(c,r):e.factory.updateQualifiedName(c,c.left,e.factory.updateIdentifier(c.right,r)),r=t.typeArguments;for(var l=0,u=W(t);l<u.length;l++){s=u[l];c=e.factory.createQualifiedName(c,s)}return e.factory.updateTypeReferenceNode(n,c,r)}function W(n){for(var t=n.typeName,r=[];!e.isIdentifier(t);)r.unshift(t.right),t=t.left;return r.unshift(t),r}}function s(n){return n.approximateLength+=3,1&n.flags?e.factory.createKeywordTypeNode(128):e.factory.createTypeReferenceNode(e.factory.createIdentifier("..."),void 0)}function c(n,t,r){var a=!!(8192&e.getCheckFlags(n)),i=a&&33554432&t.flags?Ee:lo(n),o=t.enclosingDeclaration;if(t.enclosingDeclaration=void 0,t.tracker.trackSymbol&&4096&e.getCheckFlags(n)){var c=e.first(n.declarations);if(Yo(c))if(e.isBinaryExpression(c)){var l=e.getNameOfDeclaration(c);l&&e.isElementAccessExpression(l)&&e.isPropertyAccessEntityNameExpression(l.argumentExpression)&&_(l.argumentExpression,o,t)}else _(c.name.expression,o,t)}t.enclosingDeclaration=o;var u=I(n,t);t.approximateLength+=e.symbolName(n).length+1;var p=16777216&n.flags?e.factory.createToken(57):void 0;if(8208&n.flags&&!Ms(i).length&&!oh(n))for(var m=0,f=uc(Jf(i,(function(e){return!(32768&e.flags)})),0);m<f.length;m++){var g=d(f[m],164,t,{name:u,questionToken:p});r.push(E(g))}else{var y=t.flags;t.flags|=a?33554432:0;var v=void 0;v=a&&33554432&y?s(t):i?M(t,i,n,o):e.factory.createKeywordTypeNode(128),t.flags=y;var h=oh(n)?[e.factory.createToken(142)]:void 0;h&&(t.approximateLength+=9);var b=e.factory.createPropertySignature(h,u,p,v);r.push(E(b))}function E(t){if(e.some(n.declarations,(function(e){return 333===e.kind}))){var r=e.find(n.declarations,(function(e){return 333===e.kind})).comment;r&&e.setSyntheticLeadingComments(t,[{kind:3,text:"*\n * "+r.replace(/\n/g,"\n * ")+"\n ",pos:-1,end:-1,hasTrailingNewLine:!0}])}else n.valueDeclaration&&e.setCommentRange(t,n.valueDeclaration);return t}}function l(n,t,r){if(e.some(n)){if(a(t)){if(!r)return[e.factory.createTypeReferenceNode("...",void 0)];if(n.length>2)return[o(n[0],t),e.factory.createTypeReferenceNode("... "+(n.length-2)+" more ...",void 0),o(n[n.length-1],t)]}for(var i=!(64&t.flags)?e.createUnderscoreEscapedMultiMap():void 0,s=[],c=0,l=0,u=n;l<u.length;l++){var d=u[l];if(c++,a(t)&&c+2<n.length-1){s.push(e.factory.createTypeReferenceNode("... "+(n.length-c)+" more ...",void 0));var p=o(n[n.length-1],t);p&&s.push(p);break}t.approximateLength+=2;var m=o(d,t);m&&(s.push(m),i&&e.isIdentifierTypeReference(m)&&i.add(m.typeName.escapedText,[d,s.length-1]))}if(i){var f=t.flags;t.flags|=64,i.forEach((function(n){if(!e.arrayIsHomogeneous(n,(function(e,n){return function(e,n){return e===n||!!e.symbol&&e.symbol===n.symbol||!!e.aliasSymbol&&e.aliasSymbol===n.aliasSymbol}(e[0],n[0])})))for(var r=0,a=n;r<a.length;r++){var i=a[r],c=i[0],l=i[1];s[l]=o(c,t)}})),t.flags=f}return s}}function u(n,t,r,a){var i=e.getNameFromIndexInfo(n)||"x",s=e.factory.createKeywordTypeNode(0===t?147:144),c=e.factory.createParameterDeclaration(void 0,void 0,void 0,i,void 0,s,void 0);return a||(a=o(n.type||Ee,r)),n.type||2097152&r.flags||(r.encounteredError=!0),r.approximateLength+=i.length+4,e.factory.createIndexSignature(void 0,n.isReadonly?[e.factory.createToken(142)]:void 0,[c],a)}function d(n,t,r,a){var i,s,c,l,u,d,p=256&r.flags;p&&(r.flags&=-257),32&r.flags&&n.target&&n.mapper&&n.target.typeParameters?d=n.target.typeParameters.map((function(e){return o(kd(e,n.mapper),r)})):u=n.typeParameters&&n.typeParameters.map((function(e){return m(e,r)}));var _,g=ls(n,!0)[0],y=(e.some(g,(function(n){return n!==g[g.length-1]&&!!(32768&e.getCheckFlags(n))}))?n.parameters:g).map((function(e){return f(e,r,166===t,null==a?void 0:a.privateSymbolVisitor,null==a?void 0:a.bundledImports)}));if(n.thisParameter){var v=f(n.thisParameter,r);y.unshift(v)}var h=Rc(n);if(h){var b=2===h.kind||3===h.kind?e.factory.createToken(127):void 0,E=1===h.kind||3===h.kind?e.setEmitFlags(e.factory.createIdentifier(h.parameterName),16777216):e.factory.createThisTypeNode(),T=h.type&&o(h.type,r);_=e.factory.createTypePredicateNode(b,E,T)}else{var S=kc(n);!S||p&&Li(S)?p||(_=e.factory.createKeywordTypeNode(128)):_=function(n,t,r,a,i){if(t!==xe&&n.enclosingDeclaration){var s=r.declaration&&e.getEffectiveReturnTypeNode(r.declaration);if(e.findAncestor(s,(function(e){return e===n.enclosingDeclaration}))&&s&&kd(sd(s),r.mapper)===t&&P(s,t)){var c=w(n,s,a,i);if(c)return c}}return o(t,n)}(r,S,n,null==a?void 0:a.privateSymbolVisitor,null==a?void 0:a.bundledImports)}r.approximateLength+=3;var x=169===t?e.factory.createCallSignature(u,y,_):170===t?e.factory.createConstructSignature(u,y,_):164===t?e.factory.createMethodSignature(null==a?void 0:a.modifiers,null!==(i=null==a?void 0:a.name)&&void 0!==i?i:e.factory.createIdentifier(""),null==a?void 0:a.questionToken,u,y,_):165===t?e.factory.createMethodDeclaration(void 0,null==a?void 0:a.modifiers,void 0,null!==(s=null==a?void 0:a.name)&&void 0!==s?s:e.factory.createIdentifier(""),void 0,u,y,_,void 0):166===t?e.factory.createConstructorDeclaration(void 0,null==a?void 0:a.modifiers,y,void 0):167===t?e.factory.createGetAccessorDeclaration(void 0,null==a?void 0:a.modifiers,null!==(c=null==a?void 0:a.name)&&void 0!==c?c:e.factory.createIdentifier(""),y,_,void 0):168===t?e.factory.createSetAccessorDeclaration(void 0,null==a?void 0:a.modifiers,null!==(l=null==a?void 0:a.name)&&void 0!==l?l:e.factory.createIdentifier(""),y,void 0):171===t?e.factory.createIndexSignature(void 0,null==a?void 0:a.modifiers,y,_):308===t?e.factory.createJSDocFunctionType(y,_):174===t?e.factory.createFunctionTypeNode(u,y,null!=_?_:e.factory.createTypeReferenceNode(e.factory.createIdentifier(""))):175===t?e.factory.createConstructorTypeNode(u,y,null!=_?_:e.factory.createTypeReferenceNode(e.factory.createIdentifier(""))):251===t?e.factory.createFunctionDeclaration(void 0,null==a?void 0:a.modifiers,void 0,(null==a?void 0:a.name)?e.cast(a.name,e.isIdentifier):e.factory.createIdentifier(""),u,y,_,void 0):208===t?e.factory.createFunctionExpression(null==a?void 0:a.modifiers,void 0,(null==a?void 0:a.name)?e.cast(a.name,e.isIdentifier):e.factory.createIdentifier(""),u,y,_,e.factory.createBlock([])):209===t?e.factory.createArrowFunction(null==a?void 0:a.modifiers,u,y,_,void 0,e.factory.createBlock([])):e.Debug.assertNever(t);return d&&(x.typeArguments=e.factory.createNodeArray(d)),x}function p(n,t,r){var a=t.flags;t.flags&=-513;var i=x(n,t),s=Ys(n),c=s&&o(s,t);return t.flags=a,e.factory.createTypeParameterDeclaration(i,r,c)}function m(e,n,t){return void 0===t&&(t=Us(e)),p(e,n,t&&o(t,n))}function f(n,t,r,a,i){var o=e.getDeclarationOfKind(n,160);o||e.isTransientSymbol(n)||(o=e.getDeclarationOfKind(n,326));var s,c=lo(n);o&&VT(o)&&(c=fm(c)),1073741824&t.flags&&o&&!e.isJSDocParameterTag(o)&&(s=o,W&&bc(s)&&!s.initializer)&&(c=If(c,524288));var l=M(t,c,n,t.enclosingDeclaration,a,i),u=!(8192&t.flags)&&r&&o&&o.modifiers?o.modifiers.map(e.factory.cloneNode):void 0,d=o&&e.isRestParameter(o)||32768&e.getCheckFlags(n)?e.factory.createToken(25):void 0,p=o&&o.name?78===o.name.kind?e.setEmitFlags(e.factory.cloneNode(o.name),16777216):157===o.name.kind?e.setEmitFlags(e.factory.cloneNode(o.name.right),16777216):function n(r){t.tracker.trackSymbol&&e.isComputedPropertyName(r)&&Jo(r)&&_(r.expression,t.enclosingDeclaration,t);var a=e.visitEachChild(r,n,e.nullTransformationContext,void 0,n);return e.isBindingElement(a)&&(a=e.factory.updateBindingElement(a,a.dotDotDotToken,a.propertyName,a.name,void 0)),e.nodeIsSynthesized(a)||(a=e.factory.cloneNode(a)),e.setEmitFlags(a,16777217)}(o.name):e.symbolName(n),m=o&&bc(o)||16384&e.getCheckFlags(n)?e.factory.createToken(57):void 0,f=e.factory.createParameterDeclaration(void 0,u,d,p,m,l,void 0);return t.approximateLength+=e.symbolName(n).length+3,f}function _(n,t,r){if(r.tracker.trackSymbol){var a=e.getFirstIdentifier(n),i=Cr(a,a.escapedText,1160127,void 0,void 0,!0);i&&r.tracker.trackSymbol(i,t,111551)}}function g(e,n,t,r){return n.tracker.trackSymbol(e,n.enclosingDeclaration,t),y(e,n,t,r)}function y(n,t,r,a){var i;return 262144&n.flags||!(t.enclosingDeclaration||64&t.flags)||134217728&t.flags?i=[n]:(i=e.Debug.checkDefined(function n(r,i,o){var s,c=Ka(r,t.enclosingDeclaration,i,!!(128&t.flags));if(!c||Wa(c[0],t.enclosingDeclaration,1===c.length?i:Ha(i))){var l=Aa(c?c[0]:r,t.enclosingDeclaration,i);if(e.length(l)){s=l.map((function(n){return e.some(n.declarations,ei)?E(n,t):void 0}));var u=l.map((function(e,n){return n}));u.sort(_);for(var d=0,p=u.map((function(e){return l[e]}));d<p.length;d++){var m=p[d],f=n(m,Ha(i),!1);if(f){if(m.exports&&m.exports.get("export=")&&Ia(m.exports.get("export="),r)){c=f;break}c=f.concat(c||[Ca(m,r)||r]);break}}}}if(c)return c;if(o||!(6144&r.flags)){if(!o&&!a&&e.forEach(r.declarations,ei))return;return[r]}function _(n,t){var r=s[n],a=s[t];if(r&&a){var i=e.pathIsRelative(a);return e.pathIsRelative(r)===i?e.moduleSpecifiers.countPathComponents(r)-e.moduleSpecifiers.countPathComponents(a):i?-1:1}return 0}}(n,r,!0)),e.Debug.assert(i&&i.length>0)),i}function v(n,t){var r;return 524384&UE(n).flags&&(r=e.factory.createNodeArray(e.map(yo(n),(function(e){return m(e,t)})))),r}function h(n,t,r){var a;e.Debug.assert(n&&0<=t&&t<n.length);var i=n[t],o=k(i);if(!(null===(a=r.typeParameterSymbolList)||void 0===a?void 0:a.has(o))){var s;if((r.typeParameterSymbolList||(r.typeParameterSymbolList=new e.Set)).add(o),512&r.flags&&t<n.length-1){var c=i,u=n[t+1];if(1&e.getCheckFlags(u)){var d=function(n){return e.concatenate(go(n),yo(n))}(2097152&c.flags?ea(c):c);s=l(e.map(d,(function(e){return md(e,u.mapper)})),r)}else s=v(i,r)}return s}}function b(n){return e.isIndexedAccessTypeNode(n.objectType)?b(n.objectType):n}function E(n,t){var r,a=e.getDeclarationOfKind(n,297);if(!a){var o=e.firstDefined(n.declarations,(function(e){return Da(e,n)}));o&&(a=e.getDeclarationOfKind(o,297))}if(a&&void 0!==a.moduleName)return a.moduleName;if(!a){if(t.tracker.trackReferencedAmbientModule){var s=e.filter(n.declarations,e.isAmbientModule);if(e.length(s))for(var c=0,l=s;c<l.length;c++){var u=l[c];t.tracker.trackReferencedAmbientModule(u,n)}}if(i.test(n.escapedName))return n.escapedName.substring(1,n.escapedName.length-1)}if(!t.enclosingDeclaration||!t.tracker.moduleResolverHost)return i.test(n.escapedName)?n.escapedName.substring(1,n.escapedName.length-1):e.getSourceFileOfNode(e.getNonAugmentationDeclaration(n)).fileName;var d=e.getSourceFileOfNode(e.getOriginalNode(t.enclosingDeclaration)),p=Tr(n),m=p.specifierCache&&p.specifierCache.get(d.path);if(!m){var f=!!e.outFile(V),_=t.tracker.moduleResolverHost,g=f?__assign(__assign({},V),{baseUrl:_.getCommonSourceDirectory()}):V;m=e.first(e.moduleSpecifiers.getModuleSpecifiers(n,g,d,_,{importModuleSpecifierPreference:f?"non-relative":"relative",importModuleSpecifierEnding:f?"minimal":void 0})),null!==(r=p.specifierCache)&&void 0!==r||(p.specifierCache=new e.Map),p.specifierCache.set(d.path,m)}return m}function T(n,t,r,a){var i=g(n,t,r,!(16384&t.flags)),o=111551===r;if(e.some(i[0].declarations,ei)){var s=i.length>1?y(i,i.length-1,1):void 0,c=a||h(i,0,t),l=E(i[0],t);!(67108864&t.flags)&&e.getEmitModuleResolutionKind(V)===e.ModuleResolutionKind.NodeJs&&l.indexOf("/node_modules/")>=0&&(t.encounteredError=!0,t.tracker.reportLikelyUnsafeImportRequiredError&&t.tracker.reportLikelyUnsafeImportRequiredError(l));var u=e.factory.createLiteralTypeNode(e.factory.createStringLiteral(l));if(t.tracker.trackExternalModuleSymbolOfImportTypeNode&&t.tracker.trackExternalModuleSymbolOfImportTypeNode(i[0]),t.approximateLength+=l.length+10,!s||e.isEntityName(s)){if(s)(f=e.isIdentifier(s)?s:s.right).typeArguments=void 0;return e.factory.createImportTypeNode(u,s,c,o)}var d=b(s),p=d.objectType.typeName;return e.factory.createIndexedAccessTypeNode(e.factory.createImportTypeNode(u,p,c,o),d.indexType)}var m=y(i,i.length-1,0);if(e.isIndexedAccessTypeNode(m))return m;if(o)return e.factory.createTypeQueryNode(m);var f,_=(f=e.isIdentifier(m)?m:m.right).typeArguments;return f.typeArguments=void 0,e.factory.createTypeReferenceNode(m,_);function y(n,r,i){var o,s=r===n.length-1?a:h(n,r,t),c=n[r],l=n[r-1];if(0===r)t.flags|=16777216,o=gi(c,t),t.approximateLength+=(o?o.length:0)+1,t.flags^=16777216;else if(l&&ha(l)){var u=ha(l);e.forEachEntry(u,(function(n,t){if(Ia(n,c)&&!Xo(t)&&"export="!==t)return o=e.unescapeLeadingUnderscores(t),!0}))}if(o||(o=gi(c,t)),t.approximateLength+=o.length+1,!(16&t.flags)&&l&&ns(l)&&ns(l).get(c.escapedName)&&Ia(ns(l).get(c.escapedName),c)){var d=y(n,r-1,i);return e.isIndexedAccessTypeNode(d)?e.factory.createIndexedAccessTypeNode(d,e.factory.createLiteralTypeNode(e.factory.createStringLiteral(o))):e.factory.createIndexedAccessTypeNode(e.factory.createTypeReferenceNode(d,s),e.factory.createLiteralTypeNode(e.factory.createStringLiteral(o)))}var p=e.setEmitFlags(e.factory.createIdentifier(o,s),16777216);if(p.symbol=c,r>i){d=y(n,r-1,i);return e.isEntityName(d)?e.factory.createQualifiedName(d,p):e.Debug.fail("Impossible construct - an export of an indexed access cannot be reachable")}return p}}function S(e,n,t){var r=Cr(n.enclosingDeclaration,e,788968,void 0,e,!1);return!!r&&!(262144&r.flags&&r===t.symbol)}function x(n,t){var r;if(4&t.flags&&t.typeParameterNames){var a=t.typeParameterNames.get(Hl(n));if(a)return a}var i=L(n.symbol,t,788968,!0);if(!(78&i.kind))return e.factory.createIdentifier("(Missing type parameter)");if(4&t.flags){for(var o=i.escapedText,s=0,c=o;(null===(r=t.typeParameterNamesByText)||void 0===r?void 0:r.has(c))||S(c,t,n);)c=o+"_"+ ++s;c!==o&&(i=e.factory.createIdentifier(c,i.typeArguments)),(t.typeParameterNames||(t.typeParameterNames=new e.Map)).set(Hl(n),i),(t.typeParameterNamesByText||(t.typeParameterNamesByText=new e.Set)).add(i.escapedText)}return i}function L(n,t,r,a){var i=g(n,t,r);return!a||1===i.length||t.encounteredError||65536&t.flags||(t.encounteredError=!0),function n(r,a){var i=h(r,a,t),o=r[a];0===a&&(t.flags|=16777216);var s=gi(o,t);0===a&&(t.flags^=16777216);var c=e.setEmitFlags(e.factory.createIdentifier(s,i),16777216);return c.symbol=o,a>0?e.factory.createQualifiedName(n(r,a-1),c):c}(i,i.length-1)}function A(n,t,r){var a=g(n,t,r);return function n(r,a){var i=h(r,a,t),o=r[a];0===a&&(t.flags|=16777216);var s=gi(o,t);0===a&&(t.flags^=16777216);var c=s.charCodeAt(0);if(e.isSingleOrDoubleQuote(c)&&e.some(o.declarations,ei))return e.factory.createStringLiteral(E(o,t));var l=35===c?s.length>1&&e.isIdentifierStart(s.charCodeAt(1),j):e.isIdentifierStart(c,j);if(0===a||l){var u=e.setEmitFlags(e.factory.createIdentifier(s,i),16777216);return u.symbol=o,a>0?e.factory.createPropertyAccessExpression(n(r,a-1),u):u}91===c&&(s=s.substring(1,s.length-1),c=s.charCodeAt(0));var d=void 0;return e.isSingleOrDoubleQuote(c)?d=e.factory.createStringLiteral(s.substring(1,s.length-1).replace(/\\./g,(function(e){return e.substring(1)})),39===c):""+ +s===s&&(d=e.factory.createNumericLiteral(+s)),d||((d=e.setEmitFlags(e.factory.createIdentifier(s,i),16777216)).symbol=o),e.factory.createElementAccessExpression(n(r,a-1),d)}(a,a.length-1)}function D(n){var t=e.getNameOfDeclaration(n);return!!t&&e.isStringLiteral(t)}function C(n){var t=e.getNameOfDeclaration(n);return!!(t&&e.isStringLiteral(t)&&(t.singleQuote||!e.nodeIsSynthesized(t)&&e.startsWith(e.getTextOfNode(t,!1),"'")))}function I(n,t){var r=!!e.length(n.declarations)&&e.every(n.declarations,C),a=function(n,t,r){var a=Tr(n).nameType;if(a){if(384&a.flags){var i=""+a.value;return e.isIdentifierText(i,V.target)||yg(i)?yg(i)&&e.startsWith(i,"-")?e.factory.createComputedPropertyName(e.factory.createNumericLiteral(+i)):N(i):e.factory.createStringLiteral(i,!!r)}if(8192&a.flags)return e.factory.createComputedPropertyName(A(a.symbol,t,111551))}}(n,t,r);return a||(e.isKnownSymbol(n)?e.factory.createComputedPropertyName(e.factory.createPropertyAccessExpression(e.factory.createIdentifier("Symbol"),n.escapedName.substr(3))):N(e.unescapeLeadingUnderscores(n.escapedName),!!e.length(n.declarations)&&e.every(n.declarations,D),r))}function N(n,t,r){return e.isIdentifierText(n,V.target)?e.factory.createIdentifier(n):!t&&yg(n)&&+n>=0?e.factory.createNumericLiteral(+n):e.factory.createStringLiteral(n,!!r)}function O(n,t){return n.declarations&&e.find(n.declarations,(function(n){return!(!e.getEffectiveTypeAnnotationNode(n)||t&&!e.findAncestor(n,(function(e){return e===t})))}))}function P(n,t){return!(4&e.getObjectFlags(t))||!e.isTypeReferenceNode(n)||e.length(n.typeArguments)>=Sc(t.target.typeParameters)}function M(n,t,r,a,i,s){if(t!==xe&&a){var c=O(r,a);if(c&&!e.isFunctionLikeDeclaration(c)){var l=e.getEffectiveTypeAnnotationNode(c);if(sd(l)===t&&P(l,t)){var u=w(n,l,i,s);if(u)return u}}}var d=n.flags;8192&t.flags&&t.symbol===r&&(!n.enclosingDeclaration||e.some(r.declarations,(function(t){return e.getSourceFileOfNode(t)===e.getSourceFileOfNode(n.enclosingDeclaration)})))&&(n.flags|=1048576);var p=o(t,n);return n.flags=d,p}function w(t,a,i,s){r&&r.throwIfCancellationRequested&&r.throwIfCancellationRequested();var c=!1,l=e.getSourceFileOfNode(a),u=e.visitNode(a,(function r(a){var u,d;if(e.isJSDocAllType(a)||310===a.kind)return e.factory.createKeywordTypeNode(128);if(e.isJSDocUnknownType(a))return e.factory.createKeywordTypeNode(152);if(e.isJSDocNullableType(a))return e.factory.createUnionTypeNode([e.visitNode(a.type,r),e.factory.createLiteralTypeNode(e.factory.createNull())]);if(e.isJSDocOptionalType(a))return e.factory.createUnionTypeNode([e.visitNode(a.type,r),e.factory.createKeywordTypeNode(150)]);if(e.isJSDocNonNullableType(a))return e.visitNode(a.type,r);if(e.isJSDocVariadicType(a))return e.factory.createArrayTypeNode(e.visitNode(a.type,r));if(e.isJSDocTypeLiteral(a))return e.factory.createTypeLiteralNode(e.map(a.jsDocPropertyTags,(function(n){var i=e.isIdentifier(n.name)?n.name:n.name.right,s=xi(sd(a),i.escapedText),c=s&&n.typeExpression&&sd(n.typeExpression.type)!==s?o(s,t):void 0;return e.factory.createPropertySignature(void 0,i,n.typeExpression&&e.isJSDocOptionalType(n.typeExpression.type)?e.factory.createToken(57):void 0,c||n.typeExpression&&e.visitNode(n.typeExpression.type,r)||e.factory.createKeywordTypeNode(128))})));if(e.isTypeReferenceNode(a)&&e.isIdentifier(a.typeName)&&""===a.typeName.escapedText)return e.setOriginalNode(e.factory.createKeywordTypeNode(128),a);if((e.isExpressionWithTypeArguments(a)||e.isTypeReferenceNode(a))&&e.isJSDocIndexSignature(a))return e.factory.createTypeLiteralNode([e.factory.createIndexSignature(void 0,void 0,[e.factory.createParameterDeclaration(void 0,void 0,void 0,"x",void 0,e.visitNode(a.typeArguments[0],r))],e.visitNode(a.typeArguments[1],r))]);if(e.isJSDocFunctionType(a)){var p;return e.isJSDocConstructSignature(a)?e.factory.createConstructorTypeNode(e.visitNodes(a.typeParameters,r),e.mapDefined(a.parameters,(function(n,t){return n.name&&e.isIdentifier(n.name)&&"new"===n.name.escapedText?void(p=n.type):e.factory.createParameterDeclaration(void 0,void 0,g(n),y(n,t),n.questionToken,e.visitNode(n.type,r),void 0)})),e.visitNode(p||a.type,r)||e.factory.createKeywordTypeNode(128)):e.factory.createFunctionTypeNode(e.visitNodes(a.typeParameters,r),e.map(a.parameters,(function(n,t){return e.factory.createParameterDeclaration(void 0,void 0,g(n),y(n,t),n.questionToken,e.visitNode(n.type,r),void 0)})),e.visitNode(a.type,r)||e.factory.createKeywordTypeNode(128))}if(e.isTypeReferenceNode(a)&&e.isInJSDoc(a)&&(!P(a,sd(a))||fl(a)||he===ol(il(a),788968,!0)))return e.setOriginalNode(o(sd(a),t),a);if(e.isLiteralImportTypeNode(a))return e.factory.updateImportTypeNode(a,e.factory.updateLiteralTypeNode(a.argument,function(r,a){if(s){if(t.tracker&&t.tracker.moduleResolverHost){var i=sS(r);if(i){var o={getCanonicalFileName:e.createGetCanonicalFileName(!!n.useCaseSensitiveFileNames),getCurrentDirectory:function(){return t.tracker.moduleResolverHost.getCurrentDirectory()},getCommonSourceDirectory:function(){return t.tracker.moduleResolverHost.getCommonSourceDirectory()}},c=e.getResolvedExternalModuleName(o,i);return e.factory.createStringLiteral(c)}}}else if(t.tracker&&t.tracker.trackExternalModuleSymbolOfImportTypeNode){var l=da(a,a,void 0);l&&t.tracker.trackExternalModuleSymbolOfImportTypeNode(l)}return a}(a,a.argument.literal)),a.qualifier,e.visitNodes(a.typeArguments,r,e.isTypeNode),a.isTypeOf);if(e.isEntityName(a)||e.isEntityNameExpression(a)){var m=e.getFirstIdentifier(a);if(e.isInJSFile(a)&&(e.isExportsIdentifier(m)||e.isModuleExportsAccessExpression(m.parent)||e.isQualifiedName(m.parent)&&e.isModuleIdentifier(m.parent.left)&&e.isExportsIdentifier(m.parent.right)))return c=!0,a;var f=ca(m,67108863,!0,!0);if(f&&(0!==Ya(f,t.enclosingDeclaration,67108863,!1).accessibility?c=!0:(null===(d=null===(u=t.tracker)||void 0===u?void 0:u.trackSymbol)||void 0===d||d.call(u,f,t.enclosingDeclaration,67108863),null==i||i(f)),e.isIdentifier(a))){var _=262144&f.flags?x(Fo(f),t):e.factory.cloneNode(a);return _.symbol=f,e.setEmitFlags(e.setOriginalNode(_,a),16777216)}}l&&e.isTupleTypeNode(a)&&e.getLineAndCharacterOfPosition(l,a.pos).line===e.getLineAndCharacterOfPosition(l,a.end).line&&e.setEmitFlags(a,1);return e.visitEachChild(a,r,e.nullTransformationContext);function g(n){return n.dotDotDotToken||(n.type&&e.isJSDocVariadicType(n.type)?e.factory.createToken(25):void 0)}function y(n,t){return n.name&&e.isIdentifier(n.name)&&"this"===n.name.escapedText?"this":g(n)?"args":"arg"+t}}));if(!c)return u===a?e.setTextRange(e.factory.cloneNode(a),a):u}}(),ne=e.createSymbolTable(),te=mr(4,"undefined");te.declarations=[];var re=mr(1536,"globalThis",8);re.exports=ne,re.declarations=[],ne.set(re.escapedName,re);var ae,ie=mr(4,"arguments"),oe=mr(4,"require"),se={getNodeCount:function(){return e.sum(n.getSourceFiles(),"nodeCount")},getIdentifierCount:function(){return e.sum(n.getSourceFiles(),"identifierCount")},getSymbolCount:function(){return e.sum(n.getSourceFiles(),"symbolCount")+g},getTypeCatalog:function(){return x},getTypeCount:function(){return _},getInstantiationCount:function(){return v},getRelationCacheSizes:function(){return{assignable:tr.size,identity:ar.size,subtype:er.size,strictSubtype:nr.size}},isUndefinedSymbol:function(e){return e===te},isArgumentsSymbol:function(e){return e===ie},isUnknownSymbol:function(e){return e===he},getMergedSymbol:Sa,getDiagnostics:pT,getGlobalDiagnostics:function(){return mT(),Yt.getGlobalDiagnostics()},getRecursionIdentity:Mp,getTypeOfSymbolAtLocation:function(n,t){var r=e.getParseTreeNode(t);return r?function(n,t){if(n=n.exportSymbol||n,78===t.kind&&(e.isRightSideOfQualifiedNameOrPropertyAccess(t)&&(t=t.parent),e.isExpressionNode(t)&&!e.isAssignmentTarget(t))){var r=Uh(t);if(Na(Sr(t).resolvedSymbol)===n)return r}return lo(n)}(n,r):xe},getSymbolsOfParameterPropertyDeclaration:function(n,t){var r=e.getParseTreeNode(n,e.isParameter);return void 0===r?e.Debug.fail("Cannot get symbols of a synthetic parameter that cannot be resolved to a parse-tree node."):function(n,t){var r=n.parent,a=n.parent.parent,i=Lr(r.locals,t,111551),o=Lr(ns(a.symbol),t,111551);if(i&&o)return[i,o];return e.Debug.fail("There should exist two symbols, one as property declaration and one as parameter declaration")}(r,e.escapeLeadingUnderscores(t))},getDeclaredTypeOfSymbol:Fo,getPropertiesOfType:Gs,getPropertyOfType:function(n,t){return cc(n,e.escapeLeadingUnderscores(t))},getPrivateIdentifierPropertyOfType:function(n,t,r){var a=e.getParseTreeNode(r);if(a){var i=oy(e.escapeLeadingUnderscores(t),a);return i?sy(n,i):void 0}},getTypeOfPropertyOfType:function(n,t){return xi(n,e.escapeLeadingUnderscores(t))},getIndexInfoOfType:mc,getSignaturesOfType:uc,getIndexTypeOfType:fc,getBaseTypes:Lo,getBaseTypeOfLiteralType:Qp,getWidenedType:Im,getTypeFromTypeNode:function(n){var t=e.getParseTreeNode(n,e.isTypeNode);return t?sd(t):xe},getParameterType:Ov,getPromisedTypeOfPromise:db,getAwaitedType:function(e){return mb(e)},getReturnTypeOfSignature:kc,isNullableType:Yg,getNullableType:mm,getNonNullableType:_m,getNonOptionalType:vm,getTypeArguments:nl,typeToTypeNode:ee.typeToTypeNode,indexInfoToIndexSignatureDeclaration:ee.indexInfoToIndexSignatureDeclaration,signatureToSignatureDeclaration:ee.signatureToSignatureDeclaration,symbolToEntityName:ee.symbolToEntityName,symbolToExpression:ee.symbolToExpression,symbolToTypeParameterDeclarations:ee.symbolToTypeParameterDeclarations,symbolToParameterDeclaration:ee.symbolToParameterDeclaration,typeParameterToDeclaration:ee.typeParameterToDeclaration,getSymbolsInScope:function(n,t){var r=e.getParseTreeNode(n);return r?function(n,t){if(16777216&n.flags)return[];var r=e.createSymbolTable(),a=!1;return i(),r.delete("this"),yc(r);function i(){for(;n;){switch(n.locals&&!xr(n)&&s(n.locals,t),n.kind){case 297:if(!e.isExternalOrCommonJsModule(n))break;case 256:s(xa(n).exports,2623475&t);break;case 255:s(xa(n).exports,8&t);break;case 221:n.name&&o(n.symbol,t);case 252:case 253:a||s(ns(xa(n)),788968&t);break;case 208:n.name&&o(n.symbol,t)}e.introducesArgumentsExoticObject(n)&&o(ie,t),a=e.hasSyntacticModifier(n,32),n=n.parent}s(ne,t)}function o(n,t){if(e.getCombinedLocalAndExportSymbolFlags(n)&t){var a=n.escapedName;r.has(a)||r.set(a,n)}}function s(e,n){n&&e.forEach((function(e){o(e,n)}))}}(r,t):[]},getSymbolAtLocation:function(n){var t=e.getParseTreeNode(n);return t?bT(t,!0):void 0},getShorthandAssignmentValueSymbol:function(n){var t=e.getParseTreeNode(n);return t?function(e){if(e&&289===e.kind)return ca(e.name,2208703);return}(t):void 0},getExportSpecifierLocalTargetSymbol:function(n){var t=e.getParseTreeNode(n,e.isExportSpecifier);return t?function(n){return e.isExportSpecifier(n)?n.parent.parent.moduleSpecifier?qr(n.parent.parent,n):ca(n.propertyName||n.name,2998271):ca(n,2998271)}(t):void 0},getExportSymbolOfSymbol:function(e){return Sa(e.exportSymbol||e)},getTypeAtLocation:function(n){var t=e.getParseTreeNode(n);return t?ET(t):xe},getTypeOfAssignmentPattern:function(n){var t=e.getParseTreeNode(n,e.isAssignmentPattern);return t&&TT(t)||xe},getPropertySymbolOfDestructuringAssignment:function(n){var t=e.getParseTreeNode(n,e.isIdentifier);return t?function(n){var t=TT(e.cast(n.parent.parent,e.isAssignmentPattern));return t&&cc(t,n.escapedText)}(t):void 0},signatureToString:function(n,t,r,a){return ai(n,e.getParseTreeNode(t),r,a)},typeToString:function(n,t,r){return ii(n,e.getParseTreeNode(t),r)},symbolToString:function(n,t,r,a){return ri(n,e.getParseTreeNode(t),r,a)},typePredicateToString:function(n,t,r){return di(n,e.getParseTreeNode(t),r)},writeSignature:function(n,t,r,a,i){return ai(n,e.getParseTreeNode(t),r,a,i)},writeType:function(n,t,r,a){return ii(n,e.getParseTreeNode(t),r,a)},writeSymbol:function(n,t,r,a,i){return ri(n,e.getParseTreeNode(t),r,a,i)},writeTypePredicate:function(n,t,r,a){return di(n,e.getParseTreeNode(t),r,a)},getAugmentedPropertiesOfType:AT,getRootSymbols:function n(t){var r=function(n){if(6&e.getCheckFlags(n))return e.mapDefined(Tr(n).containingType.types,(function(e){return cc(e,n.escapedName)}));if(33554432&n.flags){var t=n,r=t.leftSpread,a=t.rightSpread,i=t.syntheticOrigin;return r?[r,a]:i?[i]:e.singleElementArray(function(e){var n,t=e;for(;t=Tr(t).target;)n=t;return n}(n))}return}(t);return r?e.flatMap(r,n):[t]},getSymbolOfExpando:gv,getContextualType:function(n,t){var r=e.getParseTreeNode(n,e.isExpression);if(r){var a=e.findAncestor(r,e.isCallLikeExpression),i=a&&Sr(a).resolvedSignature;if(4&t&&a){var o=r;do{Sr(o).skipDirectInference=!0,o=o.parent}while(o&&o!==a);Sr(a).resolvedSignature=void 0}var s=ag(r,t);if(4&t&&a){o=r;do{Sr(o).skipDirectInference=void 0,o=o.parent}while(o&&o!==a);Sr(a).resolvedSignature=i}return s}},getContextualTypeForObjectLiteralElement:function(n){var t=e.getParseTreeNode(n,e.isObjectLiteralElementLike);return t?Y_(t):void 0},getContextualTypeForArgumentAtIndex:function(n,t){var r=e.getParseTreeNode(n,e.isCallLikeExpression);return r&&K_(r,t)},getContextualTypeForJsxAttribute:function(n){var t=e.getParseTreeNode(n,e.isJsxAttributeLike);return t&&$_(t)},isContextSensitive:wd,getFullyQualifiedName:sa,getResolvedSignature:function(e,n,t){return ce(e,n,t,0)},getResolvedSignatureForSignatureHelp:function(e,n,t){return ce(e,n,t,16)},getExpandedParameters:ls,hasEffectiveRestParameter:Gv,getConstantValue:function(n){var t=e.getParseTreeNode(n,qT);return t?JT(t):void 0},isValidPropertyAccess:function(n,t){var r=e.getParseTreeNode(n,e.isPropertyAccessOrQualifiedNameOrImportTypeNode);return!!r&&function(e,n){switch(e.kind){case 201:return by(e,105===e.expression.kind,n,Im(Hh(e.expression)));case 157:return by(e,!1,n,Im(Hh(e.left)));case 195:return by(e,!1,n,sd(e))}}(r,e.escapeLeadingUnderscores(t))},isValidPropertyAccessForCompletions:function(n,t,r){var a=e.getParseTreeNode(n,e.isPropertyAccessExpression);return!!a&&function(e,n,t){return by(e,201===e.kind&&105===e.expression.kind,t.escapedName,n)}(a,t,r)},getSignatureFromDeclaration:function(n){var t=e.getParseTreeNode(n,e.isFunctionLike);return t?Lc(t):void 0},isImplementationOfOverload:function(n){var t=e.getParseTreeNode(n,e.isFunctionLike);return t?UT(t):void 0},getImmediateAliasedSymbol:bg,getAliasedSymbol:ea,getEmitResolver:function(e,n){return pT(e,n),$},getExportsOfModule:ya,getExportsAndPropertiesOfModule:function(n){var t=ya(n),r=fa(n);r!==n&&e.addRange(t,Gs(lo(r)));return t},getSymbolWalker:e.createGetSymbolWalker((function(e){return Mc(e)||Ee}),Rc,kc,Lo,Ps,lo,ff,pc,Us,e.getFirstIdentifier,nl),getAmbientModules:function(){fn||(fn=[],ne.forEach((function(e,n){i.test(n)&&fn.push(e)})));return fn},getJsxIntrinsicTagNamesAt:function(n){var t=Ig(D.IntrinsicElements,n);return t?Gs(t):e.emptyArray},isOptionalParameter:function(n){var t=e.getParseTreeNode(n,e.isParameter);return!!t&&bc(t)},tryGetMemberInModuleExports:function(n,t){return va(e.escapeLeadingUnderscores(n),t)},tryGetMemberInModuleExportsAndProperties:function(n,t){return function(n,t){var r=va(n,t);if(r)return r;var a=fa(t);if(a===t)return;var i=lo(a);return 131068&i.flags||1&e.getObjectFlags(i)||qp(i)?void 0:cc(i,n)}(e.escapeLeadingUnderscores(n),t)},tryFindAmbientModuleWithoutAugmentations:function(e){return hc(e,!1)},getApparentType:Zs,getUnionType:Xl,isTypeAssignableTo:zd,createAnonymousType:Va,createSignature:is,createSymbol:mr,createIndexInfo:Wc,getAnyType:function(){return Ee},getStringType:function(){return Oe},getNumberType:function(){return Pe},createPromiseType:zv,createArrayType:Nl,getElementTypeOfArrayType:Hp,getBooleanType:function(){return Ue},getFalseType:function(e){return e?we:Fe},getTrueType:function(e){return e?Ge:Be},getVoidType:function(){return je},getUndefinedType:function(){return Ce},getNullType:function(){return Re},getESSymbolType:function(){return Ve},getNeverType:function(){return He},getOptionalType:function(){return Ne},isSymbolAccessible:Ya,isArrayType:Up,isTupleType:tm,isArrayLikeType:Kp,isTypeInvalidDueToUnionDiscriminant:function(e,n){return n.properties.some((function(n){var t=n.name&&ou(n.name),r=t&&qo(t)?Zo(t):void 0,a=void 0===r?void 0:xi(e,r);return!!a&&Yp(a)&&!zd(ET(n),a)}))},getAllPossiblePropertiesOfTypes:function(n){var t=Xl(n);if(!(1048576&t.flags))return AT(t);for(var r=e.createSymbolTable(),a=0,i=n;a<i.length;a++)for(var o=i[a],s=0,c=AT(o);s<c.length;s++){var l=c[s].escapedName;if(!r.has(l)){var u=ec(t,l);u&&r.set(l,u)}}return e.arrayFrom(r.values())},getSuggestedSymbolForNonexistentProperty:my,getSuggestionForNonexistentProperty:_y,getSuggestedSymbolForNonexistentJSXAttribute:fy,getSuggestedSymbolForNonexistentSymbol:function(n,t,r){return gy(n,e.escapeLeadingUnderscores(t),r)},getSuggestionForNonexistentSymbol:function(n,t,r){return function(n,t,r){var a=gy(n,t,r);return a&&e.symbolName(a)}(n,e.escapeLeadingUnderscores(t),r)},getSuggestedSymbolForNonexistentModule:yy,getSuggestionForNonexistentExport:function(n,t){var r=yy(n,t);return r&&e.symbolName(r)},getBaseConstraintOfType:Ws,getDefaultFromTypeParameter:function(e){return e&&262144&e.flags?Ys(e):void 0},resolveName:function(n,t,r,a){return Cr(t,e.escapeLeadingUnderscores(n),r,void 0,void 0,!1,a)},getJsxNamespace:function(n){return e.unescapeLeadingUnderscores(sr(n))},getAccessibleSymbolChain:Ka,getTypePredicateOfSignature:Rc,resolveExternalModuleName:function(n){var t=e.getParseTreeNode(n,e.isExpression);return t&&ua(t,t,!0)},resolveExternalModuleSymbol:fa,tryGetThisTypeAt:function(n,t){var r=e.getParseTreeNode(n);return r&&k_(r,t)},getTypeArgumentConstraint:function(n){var t=e.getParseTreeNode(n,e.isTypeNode);return t&&function(n){var t=e.tryCast(n.parent,e.isTypeReferenceType);if(!t)return;var r=tb(t),a=Us(r[t.typeArguments.indexOf(n)]);return a&&kd(a,pd(r,eb(t,r)))}(t)},getSuggestionDiagnostics:function(t,a){var i,o=e.getParseTreeNode(t,e.isSourceFile)||e.Debug.fail("Could not determine parsed source file.");if(e.skipTypeChecking(o,V,n))return e.emptyArray;try{return r=a,lT(o),e.Debug.assert(!!(1&Sr(o).flags)),i=e.addRange(i,Qt.getDiagnostics(o.fileName)),Lb(dT(o),(function(n,t,r){e.containsParseError(n)||uT(t,!!(8388608&n.flags))||(i||(i=[])).push(__assign(__assign({},r),{category:e.DiagnosticCategory.Suggestion}))})),i||e.emptyArray}finally{r=void 0}},runWithCancellationToken:function(e,n){try{return r=e,n(se)}finally{r=void 0}},getLocalTypeParametersOfClassOrInterfaceOrTypeAlias:yo,isDeclarationVisible:yi};function ce(n,t,r,a){var i=e.getParseTreeNode(n,e.isCallLikeExpression);ae=r;var o=i?mv(i,t,a):void 0;return ae=void 0,o}var le=new e.Map,ue=new e.Map,de=new e.Map,pe=new e.Map,me=new e.Map,fe=new e.Map,_e=new e.Map,ge=new e.Map,ye=[],ve=new e.Map,he=mr(4,"unknown"),be=mr(0,"__resolving__"),Ee=Pa(1,"any"),Te=Pa(1,"any"),Se=Pa(1,"any"),xe=Pa(1,"error"),Le=Pa(1,"any",524288),Ae=Pa(1,"intrinsic"),De=Pa(2,"unknown"),Ce=Pa(32768,"undefined"),Ie=W?Ce:Pa(32768,"undefined",524288),Ne=Pa(32768,"undefined"),Re=Pa(65536,"null"),ke=W?Re:Pa(65536,"null",524288),Oe=Pa(4,"string"),Pe=Pa(8,"number"),Me=Pa(64,"bigint"),we=Pa(512,"false"),Fe=Pa(512,"false"),Ge=Pa(512,"true"),Be=Pa(512,"true");Ge.regularType=Be,Ge.freshType=Ge,Be.regularType=Be,Be.freshType=Ge,we.regularType=Fe,we.freshType=we,Fe.regularType=Fe,Fe.freshType=we;var Ue=Ma([Fe,Be]);Ma([Fe,Ge]),Ma([we,Be]),Ma([we,Ge]);var Ve=Pa(4096,"symbol"),je=Pa(16384,"void"),He=Pa(131072,"never"),Ke=Pa(131072,"never"),We=Pa(131072,"never",2097152),ze=Pa(131072,"never"),qe=Pa(131072,"never"),Je=Pa(67108864,"object"),Xe=Xl([Oe,Pe,Ve]),Ye=Q?Oe:Xe,Qe=Xl([Pe,Me]),Ze=Xl([Oe,Pe,Ue,Me,Re,Ce]),$e=_d((function(e){return 262144&e.flags?(n=e).constraint===De?n:n.restrictiveInstantiation||(n.restrictiveInstantiation=Fa(n.symbol),n.restrictiveInstantiation.constraint=De,n.restrictiveInstantiation):e;var n})),en=_d((function(e){return 262144&e.flags?Se:e})),nn=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0),tn=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0);tn.objectFlags|=4096;var rn=mr(2048,"__type");rn.members=e.createSymbolTable();var an=Va(rn,L,e.emptyArray,e.emptyArray,void 0,void 0),on=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0);on.instantiations=new e.Map;var sn=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0);sn.objectFlags|=2097152;var cn=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0),ln=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0),un=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0),dn=Fa(),pn=Fa();pn.constraint=dn;var mn,fn,_n,gn,yn,vn,hn,bn,En,Tn,Sn,xn,Ln,An,Dn,Cn,In,Nn,Rn,kn,On,Pn,Mn,wn,Fn,Gn,Bn,Un,Vn,jn,Hn,Kn,Wn,zn,qn,Jn,Xn,Yn,Qn,Zn,$n,et,nt,tt,rt,at,it,ot=Fa(),st=Tc(1,"<<unresolved>>",0,Ee),ct=is(void 0,void 0,void 0,e.emptyArray,Ee,void 0,0,0),lt=is(void 0,void 0,void 0,e.emptyArray,xe,void 0,0,0),ut=is(void 0,void 0,void 0,e.emptyArray,Ee,void 0,0,0),dt=is(void 0,void 0,void 0,e.emptyArray,Ke,void 0,0,0),pt=Wc(Oe,!0),mt=new e.Map,ft={get yieldType(){return e.Debug.fail("Not supported")},get returnType(){return e.Debug.fail("Not supported")},get nextType(){return e.Debug.fail("Not supported")}},_t=sE(Ee,Ee,Ee),gt=sE(Ee,Ee,De),yt=sE(He,Ee,Ce),vt={iterableCacheKey:"iterationTypesOfAsyncIterable",iteratorCacheKey:"iterationTypesOfAsyncIterator",iteratorSymbolName:"asyncIterator",getGlobalIteratorType:function(e){return zn||(zn=El("AsyncIterator",3,e))||on},getGlobalIterableType:function(e){return Wn||(Wn=El("AsyncIterable",1,e))||on},getGlobalIterableIteratorType:function(e){return qn||(qn=El("AsyncIterableIterator",1,e))||on},getGlobalGeneratorType:function(e){return Jn||(Jn=El("AsyncGenerator",3,e))||on},resolveIterationType:mb,mustHaveANextMethodDiagnostic:e.Diagnostics.An_async_iterator_must_have_a_next_method,mustBeAMethodDiagnostic:e.Diagnostics.The_0_property_of_an_async_iterator_must_be_a_method,mustHaveAValueDiagnostic:e.Diagnostics.The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property},ht={iterableCacheKey:"iterationTypesOfIterable",iteratorCacheKey:"iterationTypesOfIterator",iteratorSymbolName:"iterator",getGlobalIteratorType:function(e){return Un||(Un=El("Iterator",3,e))||on},getGlobalIterableType:Al,getGlobalIterableIteratorType:function(e){return Vn||(Vn=El("IterableIterator",1,e))||on},getGlobalGeneratorType:function(e){return jn||(jn=El("Generator",3,e))||on},resolveIterationType:function(e,n){return e},mustHaveANextMethodDiagnostic:e.Diagnostics.An_iterator_must_have_a_next_method,mustBeAMethodDiagnostic:e.Diagnostics.The_0_property_of_an_iterator_must_be_a_method,mustHaveAValueDiagnostic:e.Diagnostics.The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property},bt=new e.Map,Et=!1,Tt=new e.Map,St=0,xt=0,Lt=0,At=!1,Dt=0,Ct=td(""),It=td(0),Nt=td({negative:!1,base10Value:"0"}),Rt=[],kt=[],Ot=[],Pt=0,Mt=[],wt=[],Ft=[],Gt=[],Bt=[],Ut=[],Vt=[],jt=[],Ht=[],Kt=[],Wt=[],zt=[],qt=[],Jt=[],Xt=[],Yt=e.createDiagnosticCollection(),Qt=e.createDiagnosticCollection(),Zt=new e.Map(e.getEntries({string:Oe,number:Pe,bigint:Me,boolean:Ue,symbol:Ve,undefined:Ce})),$t=Xl(e.arrayFrom(b.keys(),td)),er=new e.Map,nr=new e.Map,tr=new e.Map,rr=new e.Map,ar=new e.Map,ir=new e.Map,or=e.createSymbolTable();return or.set(te.escapedName,te),function(){for(var t=0,r=n.getSourceFiles();t<r.length;t++){var a=r[t];e.bindSourceFile(a,V)}var i;mn=new e.Map;for(var o=0,s=n.getSourceFiles();o<s.length;o++){if(!(a=s[o]).redirectInfo){if(!e.isExternalOrCommonJsModule(a)){var c=a.locals.get("globalThis");if(c)for(var l=0,u=c.declarations;l<u.length;l++){var d=u[l];Yt.add(e.createDiagnosticForNode(d,e.Diagnostics.Declaration_name_conflicts_with_built_in_global_identifier_0,"globalThis"))}br(ne,a.locals)}if(a.jsGlobalAugmentations&&br(ne,a.jsGlobalAugmentations),a.patternAmbientModules&&a.patternAmbientModules.length&&(_n=e.concatenate(_n,a.patternAmbientModules)),a.moduleAugmentations.length&&(i||(i=[])).push(a.moduleAugmentations),a.symbol&&a.symbol.globalExports)a.symbol.globalExports.forEach((function(e,n){ne.has(n)||ne.set(n,e)}))}}if(i)for(var p=0,m=i;p<m.length;p++)for(var f=m[p],_=0,g=f;_<g.length;_++){var y=g[_];e.isGlobalScopeAugmentation(y.parent)&&Er(y)}(function(n,t,r){t.forEach((function(t,a){var i=n.get(a);i?e.forEach(i.declarations,function(n,t){return function(r){return Yt.add(e.createDiagnosticForNode(r,t,n))}}(e.unescapeLeadingUnderscores(a),r)):n.set(a,t)}))})(ne,or,e.Diagnostics.Declaration_name_conflicts_with_built_in_global_identifier_0),Tr(te).type=Ie,Tr(ie).type=El("IArguments",0,!0),Tr(he).type=xe,Tr(re).type=wa(16,re),En=El("Array",1,!0),yn=El("Object",0,!0),vn=El("Function",0,!0),hn=q&&El("CallableFunction",0,!0)||vn,bn=q&&El("NewableFunction",0,!0)||vn,Sn=El("String",0,!0),xn=El("Number",0,!0),Ln=El("Boolean",0,!0),An=El("RegExp",0,!0),Cn=Nl(Ee),(In=Nl(Te))===nn&&(In=Va(void 0,L,e.emptyArray,e.emptyArray,void 0,void 0));if(Tn=Dl("ReadonlyArray",1)||En,Nn=Tn?Cl(Tn,[Ee]):Cn,Dn=Dl("ThisType",1),i)for(var v=0,h=i;v<h.length;v++){f=h[v];for(var b=0,E=f;b<E.length;b++){y=E[b];e.isGlobalScopeAugmentation(y.parent)||Er(y)}}mn.forEach((function(n){var t=n.firstFile,r=n.secondFile,a=n.conflictingSymbols;if(a.size<8)a.forEach((function(n,t){for(var r=n.isBlockScoped,a=n.firstFileLocations,i=n.secondFileLocations,o=r?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,s=0,c=a;s<c.length;s++){hr(c[s],o,t,i)}for(var l=0,u=i;l<u.length;l++){hr(u[l],o,t,a)}}));else{var i=e.arrayFrom(a.keys()).join(", ");Yt.add(e.addRelatedInfo(e.createDiagnosticForNode(t,e.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0,i),e.createDiagnosticForNode(r,e.Diagnostics.Conflicts_are_in_this_file))),Yt.add(e.addRelatedInfo(e.createDiagnosticForNode(r,e.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0,i),e.createDiagnosticForNode(t,e.Diagnostics.Conflicts_are_in_this_file)))}})),mn=void 0}(),se;function sr(n){if(n){var t=e.getSourceFileOfNode(n);if(t)if(e.isJsxOpeningFragment(n)){if(t.localJsxFragmentNamespace)return t.localJsxFragmentNamespace;var r=t.pragmas.get("jsxfrag");if(r){var a=e.isArray(r)?r[0]:r;if(t.localJsxFragmentFactory=e.parseIsolatedEntityName(a.arguments.factory,j),e.visitNode(t.localJsxFragmentFactory,s),t.localJsxFragmentFactory)return t.localJsxFragmentNamespace=e.getFirstIdentifier(t.localJsxFragmentFactory).escapedText}var i=oS(n);if(i)return t.localJsxFragmentFactory=i,t.localJsxFragmentNamespace=e.getFirstIdentifier(i).escapedText}else{if(t.localJsxNamespace)return t.localJsxNamespace;var o=t.pragmas.get("jsx");if(o){a=e.isArray(o)?o[0]:o;if(t.localJsxFactory=e.parseIsolatedEntityName(a.arguments.factory,j),e.visitNode(t.localJsxFactory,s),t.localJsxFactory)return t.localJsxNamespace=e.getFirstIdentifier(t.localJsxFactory).escapedText}}}return rt||(rt="React",V.jsxFactory?(at=e.parseIsolatedEntityName(V.jsxFactory,j),e.visitNode(at,s),at&&(rt=e.getFirstIdentifier(at).escapedText)):V.reactNamespace&&(rt=e.escapeLeadingUnderscores(V.reactNamespace))),at||(at=e.factory.createQualifiedName(e.factory.createIdentifier(e.unescapeLeadingUnderscores(rt)),"createElement")),rt;function s(n){return e.setTextRangePosEnd(n,-1,-1),e.visitEachChild(n,s,e.nullTransformationContext)}}function cr(e,n,t,r,a,i,o){var s=lr(n,t,r,a,i,o);return s.skippedOn=e,s}function lr(n,t,r,a,i,o){var s=n?e.createDiagnosticForNode(n,t,r,a,i,o):e.createCompilerDiagnostic(t,r,a,i,o);return Yt.add(s),s}function ur(n,t){n?Yt.add(t):Qt.add(__assign(__assign({},t),{category:e.DiagnosticCategory.Suggestion}))}function dr(n,t,r,a,i,o,s){if(t.pos<0||t.end<0){if(!n)return;var c=e.getSourceFileOfNode(t);ur(n,"message"in r?e.createFileDiagnostic(c,0,0,r,a,i,o,s):e.createDiagnosticForFileFromMessageChain(c,r))}else ur(n,"message"in r?e.createDiagnosticForNode(t,r,a,i,o,s):e.createDiagnosticForNodeFromMessageChain(t,r))}function pr(n,t,r,a,i,o,s){var c=lr(n,r,a,i,o,s);if(t){var l=e.createDiagnosticForNode(n,e.Diagnostics.Did_you_forget_to_use_await);e.addRelatedInfo(c,l)}return c}function mr(e,n,t){g++;var r=new p(33554432|e,n);return r.checkFlags=t||0,r}function fr(e){var n=0;return 2&e&&(n|=111551),1&e&&(n|=111550),4&e&&(n|=0),8&e&&(n|=900095),16&e&&(n|=110991),32&e&&(n|=899503),64&e&&(n|=788872),256&e&&(n|=899327),128&e&&(n|=899967),512&e&&(n|=110735),8192&e&&(n|=103359),32768&e&&(n|=46015),65536&e&&(n|=78783),262144&e&&(n|=526824),524288&e&&(n|=788968),2097152&e&&(n|=2097152),n}function _r(e,n){n.mergeId||(n.mergeId=l,l++),Mt[n.mergeId]=e}function gr(n){var t=mr(n.flags,n.escapedName);return t.declarations=n.declarations?n.declarations.slice():[],t.parent=n.parent,n.valueDeclaration&&(t.valueDeclaration=n.valueDeclaration),n.constEnumOnlyModule&&(t.constEnumOnlyModule=!0),n.members&&(t.members=new e.Map(n.members)),n.exports&&(t.exports=new e.Map(n.exports)),_r(t,n),t}function yr(n,t,r){if(void 0===r&&(r=!1),!(n.flags&fr(t.flags))||67108864&(t.flags|n.flags)){if(t===n)return n;if(!(33554432&n.flags)){var a=$r(n);if(a===he)return t;n=gr(a)}512&t.flags&&512&n.flags&&n.constEnumOnlyModule&&!t.constEnumOnlyModule&&(n.constEnumOnlyModule=!1),n.flags|=t.flags,t.valueDeclaration&&e.setValueDeclaration(n,t.valueDeclaration),e.addRange(n.declarations,t.declarations),t.members&&(n.members||(n.members=e.createSymbolTable()),br(n.members,t.members,r)),t.exports&&(n.exports||(n.exports=e.createSymbolTable()),br(n.exports,t.exports,r)),r||_r(n,t)}else if(1024&n.flags)n!==re&&lr(e.getNameOfDeclaration(t.declarations[0]),e.Diagnostics.Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity,ri(n));else{var i=!!(384&n.flags||384&t.flags),o=!!(2&n.flags||2&t.flags),s=i?e.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations:o?e.Diagnostics.Cannot_redeclare_block_scoped_variable_0:e.Diagnostics.Duplicate_identifier_0,c=t.declarations&&e.getSourceFileOfNode(t.declarations[0]),l=n.declarations&&e.getSourceFileOfNode(n.declarations[0]),u=ri(t);if(c&&l&&mn&&!i&&c!==l){var d=-1===e.comparePaths(c.path,l.path)?c:l,p=d===c?l:c,m=e.getOrUpdate(mn,d.path+"|"+p.path,(function(){return{firstFile:d,secondFile:p,conflictingSymbols:new e.Map}})),f=e.getOrUpdate(m.conflictingSymbols,u,(function(){return{isBlockScoped:o,firstFileLocations:[],secondFileLocations:[]}}));_(f.firstFileLocations,t),_(f.secondFileLocations,n)}else vr(t,s,u,n),vr(n,s,u,t)}return n;function _(n,t){for(var r=0,a=t.declarations;r<a.length;r++){var i=a[r];e.pushIfUnique(n,i)}}}function vr(n,t,r,a){e.forEach(n.declarations,(function(e){hr(e,t,r,a.declarations)}))}function hr(n,t,r,a){for(var i=(e.getExpandoInitializer(n,!1)?e.getNameOfExpando(n):e.getNameOfDeclaration(n))||n,o=function(n,t,r,a,i,o){var s=n?e.createDiagnosticForNode(n,t,r,a,i,o):e.createCompilerDiagnostic(t,r,a,i,o);return Yt.lookup(s)||(Yt.add(s),s)}(i,t,r),s=function(n){var t=(e.getExpandoInitializer(n,!1)?e.getNameOfExpando(n):e.getNameOfDeclaration(n))||n;if(t===i)return"continue";o.relatedInformation=o.relatedInformation||[];var a=e.createDiagnosticForNode(t,e.Diagnostics._0_was_also_declared_here,r),s=e.createDiagnosticForNode(t,e.Diagnostics.and_here);if(e.length(o.relatedInformation)>=5||e.some(o.relatedInformation,(function(n){return 0===e.compareDiagnostics(n,s)||0===e.compareDiagnostics(n,a)})))return"continue";e.addRelatedInfo(o,e.length(o.relatedInformation)?s:a)},c=0,l=a||e.emptyArray;c<l.length;c++){s(l[c])}}function br(e,n,t){void 0===t&&(t=!1),n.forEach((function(n,r){var a=e.get(r);e.set(r,a?yr(a,n,t):n)}))}function Er(n){var t,r,a=n.parent;if(a.symbol.declarations[0]===a)if(e.isGlobalScopeAugmentation(a))br(ne,a.symbol.exports);else{var i=da(n,n,8388608&n.parent.parent.flags?void 0:e.Diagnostics.Invalid_module_name_in_augmentation_module_0_cannot_be_found,!0);if(!i)return;if(1920&(i=fa(i)).flags)if(e.some(_n,(function(e){return i===e.symbol}))){var o=yr(a.symbol,i,!0);gn||(gn=new e.Map),gn.set(n.text,o)}else{if((null===(t=i.exports)||void 0===t?void 0:t.get("__export"))&&(null===(r=a.symbol.exports)||void 0===r?void 0:r.size))for(var s=es(i,"resolvedExports"),c=0,l=e.arrayFrom(a.symbol.exports.entries());c<l.length;c++){var u=l[c],d=u[0],p=u[1];s.has(d)&&!i.exports.has(d)&&yr(s.get(d),p)}yr(i,a.symbol)}else lr(n,e.Diagnostics.Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity,n.text)}else e.Debug.assert(a.symbol.declarations.length>1)}function Tr(e){if(33554432&e.flags)return e;var n=k(e);return wt[n]||(wt[n]=new I)}function Sr(e){var n=R(e);return Ft[n]||(Ft[n]=new N)}function xr(n){return 297===n.kind&&!e.isExternalOrCommonJsModule(n)}function Lr(n,t,r){if(r){var a=Sa(n.get(t));if(a){if(e.Debug.assert(0==(1&e.getCheckFlags(a)),"Should never get an instantiated symbol here."),a.flags&r)return a;if(2097152&a.flags){var i=ea(a);if(i===he||i.flags&r)return a}}}}function Ar(t,r){var a=e.getSourceFileOfNode(t),i=e.getSourceFileOfNode(r),o=e.getEnclosingBlockScopeContainer(t);if(a!==i){if(H&&(a.externalModuleIndicator||i.externalModuleIndicator)||!e.outFile(V)||_f(r)||8388608&t.flags)return!0;if(l(r,t))return!0;var s=n.getSourceFiles();return s.indexOf(a)<=s.indexOf(i)}if(t.pos<=r.pos&&(!e.isPropertyDeclaration(t)||!e.isThisProperty(r.parent)||t.initializer||t.exclamationToken)){if(198===t.kind){var c=e.getAncestor(r,198);return c?e.findAncestor(c,e.isBindingElement)!==e.findAncestor(t,e.isBindingElement)||t.pos<c.pos:Ar(e.getAncestor(t,249),r)}return 249===t.kind?!function(n,t){switch(n.parent.parent.kind){case 232:case 237:case 239:if(Fr(t,n,o))return!0}var r=n.parent.parent;return e.isForInOrOfStatement(r)&&Fr(t,r.expression,o)}(t,r):e.isClassDeclaration(t)?!e.findAncestor(r,(function(n){return e.isComputedPropertyName(n)&&n.parent.parent===t})):e.isPropertyDeclaration(t)?!u(t,r,!1):!e.isParameterPropertyDeclaration(t,t.parent)||!(99===V.target&&V.useDefineForClassFields&&e.getContainingClass(t)===e.getContainingClass(r)&&l(r,t))}return!!(270===r.parent.kind||266===r.parent.kind&&r.parent.isExportEquals)||(!(266!==r.kind||!r.isExportEquals)||(!!(4194304&r.flags||_f(r)||e.findAncestor(r,(function(n){return e.isInterfaceDeclaration(n)||e.isTypeAliasDeclaration(n)})))||!!l(r,t)&&(99!==V.target||!V.useDefineForClassFields||!e.getContainingClass(t)||!e.isPropertyDeclaration(t)&&!e.isParameterPropertyDeclaration(t,t.parent)||!u(t,r,!0))));function l(n,t){return!!e.findAncestor(n,(function(r){if(r===o)return"quit";if(e.isFunctionLike(r))return!0;if(r.parent&&163===r.parent.kind&&r.parent.initializer===r)if(e.hasSyntacticModifier(r.parent,32)){if(165===t.kind)return!0}else if(!(163===t.kind&&!e.hasSyntacticModifier(t,32))||e.getContainingClass(n)!==e.getContainingClass(t))return!0;return!1}))}function u(n,t,r){return!(t.end>n.end)&&void 0===e.findAncestor(t,(function(t){if(t===n)return"quit";switch(t.kind){case 209:return!0;case 163:return!r||!(e.isPropertyDeclaration(n)&&t.parent===n.parent||e.isParameterPropertyDeclaration(n,n.parent)&&t.parent===n.parent.parent)||"quit";case 230:switch(t.parent.kind){case 167:case 165:case 168:return!0;default:return!1}default:return!1}}))}}function Dr(n,t,r){var a=e.getEmitScriptTarget(V),i=t;if(e.isParameter(r)&&i.body&&n.valueDeclaration.pos>=i.body.pos&&n.valueDeclaration.end<=i.body.end&&a>=2){var o=Sr(i);return void 0===o.declarationRequiresScopeChange&&(o.declarationRequiresScopeChange=e.forEach(i.parameters,(function(e){return s(e.name)||!!e.initializer&&s(e.initializer)}))||!1),!o.declarationRequiresScopeChange}return!1;function s(n){switch(n.kind){case 209:case 208:case 251:case 166:return!1;case 165:case 167:case 168:case 288:return s(n.name);case 163:return e.hasStaticModifier(n)?a<99||!V.useDefineForClassFields:s(n.name);default:return e.isNullishCoalesce(n)||e.isOptionalChain(n)?a<7:e.isBindingElement(n)&&n.dotDotDotToken&&e.isObjectBindingPattern(n.parent)?a<4:!e.isTypeNode(n)&&(e.forEachChild(n,s)||!1)}}}function Cr(e,n,t,r,a,i,o,s){return void 0===o&&(o=!1),Ir(e,n,t,r,a,i,o,Lr,s)}function Ir(n,t,r,a,i,o,s,c,l){var u,d,p,m,f,_,g=n,y=!1,v=n,h=!1;e:for(;n;){if(n.locals&&!xr(n)&&(u=c(n.locals,t,r))){var b=!0;if(e.isFunctionLike(n)&&d&&d!==n.body?(r&u.flags&788968&&311!==d.kind&&(b=!!(262144&u.flags)&&(d===n.type||160===d.kind||159===d.kind)),r&u.flags&3&&(Dr(u,n,d)?b=!1:1&u.flags&&(b=160===d.kind||d===n.type&&!!e.findAncestor(u.valueDeclaration,e.isParameter)))):184===n.kind&&(b=d===n.trueType),b)break e;u=void 0}switch(y=y||Nr(n,d),n.kind){case 297:if(!e.isExternalOrCommonJsModule(n))break;h=!0;case 256:var E=xa(n).exports||L;if(297===n.kind||e.isModuleDeclaration(n)&&8388608&n.flags&&!e.isGlobalScopeAugmentation(n)){if(u=E.get("default")){var T=e.getLocalSymbolForExportDefault(u);if(T&&u.flags&r&&T.escapedName===t)break e;u=void 0}var S=E.get(t);if(S&&2097152===S.flags&&(e.getDeclarationOfKind(S,270)||e.getDeclarationOfKind(S,269)))break}if("default"!==t&&(u=c(E,t,2623475&r))){if(!e.isSourceFile(n)||!n.commonJsModuleIndicator||u.declarations.some(e.isJSDocTypeAlias))break e;u=void 0}break;case 255:if(u=c(xa(n).exports,t,8&r))break e;break;case 163:if(!e.hasSyntacticModifier(n,32)){var x=ka(n.parent);x&&x.locals&&c(x.locals,t,111551&r)&&(m=n)}break;case 252:case 221:case 253:if(u=c(xa(n).members||L,t,788968&r)){if(!Or(u,n)){u=void 0;break}if(d&&e.hasSyntacticModifier(d,32))return void lr(v,e.Diagnostics.Static_members_cannot_reference_class_type_parameters);break e}if(221===n.kind&&32&r){var A=n.name;if(A&&t===A.escapedText){u=n.symbol;break e}}break;case 223:if(d===n.expression&&93===n.parent.token){var D=n.parent.parent;if(e.isClassLike(D)&&(u=c(xa(D).members,t,788968&r)))return void(a&&lr(v,e.Diagnostics.Base_class_expressions_cannot_reference_class_type_parameters))}break;case 158:if(_=n.parent.parent,(e.isClassLike(_)||253===_.kind)&&(u=c(xa(_).members,t,788968&r)))return void lr(v,e.Diagnostics.A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type);break;case 209:if(V.target>=2)break;case 165:case 166:case 167:case 168:case 251:if(3&r&&"arguments"===t){u=ie;break e}break;case 208:if(3&r&&"arguments"===t){u=ie;break e}if(16&r){var C=n.name;if(C&&t===C.escapedText){u=n.symbol;break e}}break;case 161:n.parent&&160===n.parent.kind&&(n=n.parent),n.parent&&(e.isClassElement(n.parent)||252===n.parent.kind)&&(n=n.parent);break;case 331:case 324:case 325:n=e.getJSDocHost(n);break;case 160:d&&(d===n.initializer||d===n.name&&e.isBindingPattern(d))&&(f||(f=n));break;case 198:d&&(d===n.initializer||d===n.name&&e.isBindingPattern(d))&&e.isParameterDeclaration(n)&&!f&&(f=n);break;case 185:if(262144&r){var I=n.typeParameter.name;if(I&&t===I.escapedText){u=n.typeParameter.symbol;break e}}}Rr(n)&&(p=n),d=n,n=n.parent}if(!o||!u||p&&u===p.symbol||(u.isReferenced|=r),!u){if(d&&(e.Debug.assert(297===d.kind),d.commonJsModuleIndicator&&"exports"===t&&r&d.symbol.flags))return d.symbol;s||(u=c(ne,t,r))}if(!u&&g&&e.isInJSFile(g)&&g.parent&&e.isRequireCall(g.parent,!1))return oe;if(u){if(a){if(m&&(99!==V.target||!V.useDefineForClassFields)){var N=m.name;return void lr(v,e.Diagnostics.Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor,e.declarationNameToString(N),kr(i))}if(v&&(2&r||(32&r||384&r)&&111551==(111551&r))){var R=Na(u);(2&R.flags||32&R.flags||384&R.flags)&&function(n,t){if(e.Debug.assert(!!(2&n.flags||32&n.flags||384&n.flags)),67108881&n.flags&&32&n.flags)return;var r=e.find(n.declarations,(function(n){return e.isBlockOrCatchScoped(n)||e.isClassLike(n)||255===n.kind}));if(void 0===r)return e.Debug.fail("checkResolvedBlockScopedVariable could not find block-scoped declaration");if(!(8388608&r.flags||Ar(r,t))){var a=void 0,i=e.declarationNameToString(e.getNameOfDeclaration(r));2&n.flags?a=lr(t,e.Diagnostics.Block_scoped_variable_0_used_before_its_declaration,i):32&n.flags?a=lr(t,e.Diagnostics.Class_0_used_before_its_declaration,i):256&n.flags?a=lr(t,e.Diagnostics.Enum_0_used_before_its_declaration,i):(e.Debug.assert(!!(128&n.flags)),V.preserveConstEnums&&(a=lr(t,e.Diagnostics.Enum_0_used_before_its_declaration,i))),a&&e.addRelatedInfo(a,e.createDiagnosticForNode(r,e.Diagnostics._0_is_declared_here,i))}}(R,v)}if(u&&h&&111551==(111551&r)&&!(4194304&g.flags)){var k=Sa(u);e.length(k.declarations)&&e.every(k.declarations,(function(n){return e.isNamespaceExportDeclaration(n)||e.isSourceFile(n)&&!!n.symbol.globalExports}))&&dr(!V.allowUmdGlobalAccess,v,e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead,e.unescapeLeadingUnderscores(t))}if(u&&f&&!y&&111551==(111551&r)){var O=Sa(ts(u)),P=e.getRootDeclaration(f);O===xa(f)?lr(v,e.Diagnostics.Parameter_0_cannot_reference_itself,e.declarationNameToString(f.name)):O.valueDeclaration&&O.valueDeclaration.pos>f.pos&&P.parent.locals&&c(P.parent.locals,O.escapedName,r)===O&&lr(v,e.Diagnostics.Parameter_0_cannot_reference_identifier_1_declared_after_it,e.declarationNameToString(f.name),e.declarationNameToString(v))}u&&v&&111551&r&&2097152&u.flags&&function(n,t,r){if(!e.isValidTypeOnlyAliasUseSite(r)){var a=ra(n);if(a){var i=e.typeOnlyDeclarationIsExport(a),o=i?e.Diagnostics._0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type:e.Diagnostics._0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type,s=i?e.Diagnostics._0_was_exported_here:e.Diagnostics._0_was_imported_here,c=e.unescapeLeadingUnderscores(t);e.addRelatedInfo(lr(r,o,c),e.createDiagnosticForNode(a,s,c))}}}(u,t,v)}return u}if(a&&!(v&&(function(n,t,r){if(!e.isIdentifier(n)||n.escapedText!==t||_T(n)||_f(n))return!1;var a=e.getThisContainer(n,!1),i=a;for(;i;){if(e.isClassLike(i.parent)){var o=xa(i.parent);if(!o)break;if(cc(lo(o),t))return lr(n,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0,kr(r),ri(o)),!0;if(i===a&&!e.hasSyntacticModifier(i,32))if(cc(Fo(o).thisType,t))return lr(n,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0,kr(r)),!0}i=i.parent}return!1}(v,t,i)||Pr(v)||function(n,t,r){var a=1920|(e.isInJSFile(n)?111551:0);if(r===a){var i=$r(Cr(n,t,788968&~a,void 0,void 0,!1)),o=n.parent;if(i){if(e.isQualifiedName(o)){e.Debug.assert(o.left===n,"Should only be resolving left side of qualified name as a namespace");var s=o.right.escapedText;if(cc(Fo(i),s))return lr(o,e.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1,e.unescapeLeadingUnderscores(t),e.unescapeLeadingUnderscores(s)),!0}return lr(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here,e.unescapeLeadingUnderscores(t)),!0}}return!1}(v,t,r)||function(n,t){if(wr(t)&&270===n.parent.kind)return lr(n,e.Diagnostics.Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module,t),!0;return!1}(v,t)||function(n,t,r){if(111551&r){if(wr(t))return lr(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here,e.unescapeLeadingUnderscores(t)),!0;var a=$r(Cr(n,t,788544,void 0,void 0,!1));if(a&&!(1024&a.flags)){var i=e.unescapeLeadingUnderscores(t);return!function(e){switch(e){case"Promise":case"Symbol":case"Map":case"WeakMap":case"Set":case"WeakSet":return!0}return!1}(t)?!function(n,t){var r=e.findAncestor(n.parent,(function(n){return!e.isComputedPropertyName(n)&&!e.isPropertySignature(n)&&(e.isTypeLiteralNode(n)||"quit")}));if(r&&1===r.members.length){var a=Fo(t);return!!(1048576&a.flags)&&mh(a,384,!0)}return!1}(n,a)?lr(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here,i):lr(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0,i,"K"===i?"P":"K"):lr(n,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later,i),!0}}return!1}(v,t,r)||function(n,t,r){if(111127&r){if($r(Cr(n,t,1024,void 0,void 0,!1)))return lr(n,e.Diagnostics.Cannot_use_namespace_0_as_a_value,e.unescapeLeadingUnderscores(t)),!0}else if(788544&r){if($r(Cr(n,t,1536,void 0,void 0,!1)))return lr(n,e.Diagnostics.Cannot_use_namespace_0_as_a_type,e.unescapeLeadingUnderscores(t)),!0}return!1}(v,t,r)||function(n,t,r){if(788584&r){var a=$r(Cr(n,t,111127,void 0,void 0,!1));if(a&&!(1920&a.flags))return lr(n,e.Diagnostics._0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0,e.unescapeLeadingUnderscores(t)),!0}return!1}(v,t,r)))){var M=void 0;if(l&&Pt<10&&(M=gy(g,t,r))){var w=ri(M),F=lr(v,l,kr(i),w);M.valueDeclaration&&e.addRelatedInfo(F,e.createDiagnosticForNode(M.valueDeclaration,e.Diagnostics._0_is_declared_here,w))}if(!M&&i){var G=function(n){for(var t=kr(n),r=e.getScriptTargetFeatures(),a=e.getOwnKeys(r),i=0,o=a;i<o.length;i++){var s=o[i],c=e.getOwnKeys(r[s]);if(void 0!==c&&e.contains(c,t))return s}}(i);G?lr(v,a,kr(i),G):lr(v,a,kr(i))}Pt++}}function Nr(n,t){return 209!==n.kind&&208!==n.kind?e.isTypeQueryNode(n)||(e.isFunctionLikeDeclaration(n)||163===n.kind&&!e.hasSyntacticModifier(n,32))&&(!t||t!==n.name):(!t||t!==n.name)&&(!(!n.asteriskToken&&!e.hasSyntacticModifier(n,256))||!e.getImmediatelyInvokedFunctionExpression(n))}function Rr(e){switch(e.kind){case 251:case 252:case 253:case 255:case 254:case 256:return!0;default:return!1}}function kr(n){return e.isString(n)?e.unescapeLeadingUnderscores(n):e.declarationNameToString(n)}function Or(n,t){for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(159===i.kind)if((e.isJSDocTemplateTag(i.parent)?e.getJSDocHost(i.parent):i.parent)===t)return!(e.isJSDocTemplateTag(i.parent)&&e.find(i.parent.parent.tags,e.isJSDocTypeAlias))}return!1}function Pr(n){var t=Mr(n);return!(!t||!ca(t,64,!0))&&(lr(n,e.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements,e.getTextOfNode(t)),!0)}function Mr(n){switch(n.kind){case 78:case 201:return n.parent?Mr(n.parent):void 0;case 223:if(e.isEntityNameExpression(n.expression))return n.expression;default:return}}function wr(e){return"any"===e||"string"===e||"number"===e||"boolean"===e||"never"===e||"unknown"===e}function Fr(n,t,r){return!!t&&!!e.findAncestor(n,(function(n){return n===r||e.isFunctionLike(n)?"quit":n===t}))}function Gr(e){switch(e.kind){case 260:return e;case 262:return e.parent;case 263:return e.parent.parent;case 265:return e.parent.parent.parent;default:return}}function Br(n){return e.find(n.declarations,Ur)}function Ur(n){return 260===n.kind||259===n.kind||262===n.kind&&!!n.name||263===n.kind||269===n.kind||265===n.kind||270===n.kind||266===n.kind&&e.exportAssignmentIsAlias(n)||e.isBinaryExpression(n)&&2===e.getAssignmentDeclarationKind(n)&&e.exportAssignmentIsAlias(n)||e.isAccessExpression(n)&&e.isBinaryExpression(n.parent)&&n.parent.left===n&&62===n.parent.operatorToken.kind&&Vr(n.parent.right)||289===n.kind||288===n.kind&&Vr(n.initializer)||e.isRequireVariableDeclaration(n,!0)}function Vr(n){return e.isAliasableExpression(n)||e.isFunctionExpression(n)&&fv(n)}function jr(n,t){var r=Jr(n);if(r){var a=e.getLeftmostAccessExpression(r.expression).arguments[0];return e.isIdentifier(r.name)?$r(cc(Ic(a),r.name.escapedText)):void 0}if(e.isVariableDeclaration(n)||272===n.moduleReference.kind){var i=ua(n,e.getExternalModuleRequireArgument(n)||e.getExternalModuleImportEqualsDeclarationExpression(n)),o=fa(i);return na(n,i,o,!1),o}var s=oa(n.moduleReference,t);return function(n,t){if(na(n,void 0,t,!1)){var r=ra(xa(n)),a=e.typeOnlyDeclarationIsExport(r),i=a?e.Diagnostics.An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type:e.Diagnostics.An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type,o=a?e.Diagnostics._0_was_exported_here:e.Diagnostics._0_was_imported_here,s=e.unescapeLeadingUnderscores(r.name.escapedText);e.addRelatedInfo(lr(n.moduleReference,i),e.createDiagnosticForNode(r,o,s))}}(n,s),s}function Hr(e,n,t,r){var a=e.exports.get("export="),i=a?cc(lo(a),n):e.exports.get(n),o=$r(i,r);return na(t,i,o,!1),o}function Kr(n){return e.isExportAssignment(n)&&!n.isExportEquals||e.hasSyntacticModifier(n,512)||e.isExportSpecifier(n)}function Wr(n,t,r){if(!K)return!1;if(!n||n.isDeclarationFile){var a=Hr(t,"default",void 0,!0);return(!a||!e.some(a.declarations,Kr))&&!Hr(t,e.escapeLeadingUnderscores("__esModule"),void 0,r)}return e.isSourceFileJS(n)?!n.externalModuleIndicator&&!Hr(t,e.escapeLeadingUnderscores("__esModule"),void 0,r):ga(t)}function zr(n,t){var r=ua(n,n.parent.moduleSpecifier);if(r){var a=void 0;a=e.isShorthandAmbientModuleSymbol(r)?r:Hr(r,"default",n,t);var i=Wr(e.find(r.declarations,e.isSourceFile),r,t);if(a||i){if(i){var o=fa(r,t)||$r(r,t);return na(n,r,o,!1),o}}else if(ga(r)){var s=H>=e.ModuleKind.ES2015?"allowSyntheticDefaultImports":"esModuleInterop",c=r.exports.get("export=").valueDeclaration,l=lr(n.name,e.Diagnostics.Module_0_can_only_be_default_imported_using_the_1_flag,ri(r),s);e.addRelatedInfo(l,e.createDiagnosticForNode(c,e.Diagnostics.This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the_0_flag,s))}else!function(n,t){var r,a;if(null===(r=n.exports)||void 0===r?void 0:r.has(t.symbol.escapedName))lr(t.name,e.Diagnostics.Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead,ri(n),ri(t.symbol));else{var i=lr(t.name,e.Diagnostics.Module_0_has_no_default_export,ri(n)),o=null===(a=n.exports)||void 0===a?void 0:a.get("__export");if(o){var s=e.find(o.declarations,(function(n){var t,r;return!!(e.isExportDeclaration(n)&&n.moduleSpecifier&&(null===(r=null===(t=ua(n,n.moduleSpecifier))||void 0===t?void 0:t.exports)||void 0===r?void 0:r.has("default")))}));s&&e.addRelatedInfo(i,e.createDiagnosticForNode(s,e.Diagnostics.export_Asterisk_does_not_re_export_a_default))}}}(r,n);return na(n,a,void 0,!1),a}}function qr(n,t,r){var a;void 0===r&&(r=!1);var i=e.getExternalModuleRequireArgument(n)||n.moduleSpecifier,o=ua(n,i),s=!e.isPropertyAccessExpression(t)&&t.propertyName||t.name;if(e.isIdentifier(s)){var c=_a(o,i,r,"default"===s.escapedText&&!(!V.allowSyntheticDefaultImports&&!V.esModuleInterop));if(c&&s.escapedText){if(e.isShorthandAmbientModuleSymbol(o))return o;var l=void 0;l=o&&o.exports&&o.exports.get("export=")?cc(lo(c),s.escapedText):function(e,n){if(3&e.flags){var t=e.valueDeclaration.type;if(t)return $r(cc(sd(t),n))}}(c,s.escapedText),l=$r(l,r);var u=function(e,n,t,r){if(1536&e.flags){var a=ha(e).get(n.escapedText),i=$r(a,r);return na(t,a,i,!1),i}}(c,s,t,r);if(void 0===u&&"default"===s.escapedText)Wr(e.find(o.declarations,e.isSourceFile),o,r)&&(u=fa(o,r)||$r(o,r));var d=u&&l&&u!==l?function(n,t){if(n===he&&t===he)return he;if(790504&n.flags)return n;var r=mr(n.flags|t.flags,n.escapedName);return r.declarations=e.deduplicate(e.concatenate(n.declarations,t.declarations),e.equateValues),r.parent=n.parent||t.parent,n.valueDeclaration&&(r.valueDeclaration=n.valueDeclaration),t.members&&(r.members=new e.Map(t.members)),n.exports&&(r.exports=new e.Map(n.exports)),r}(l,u):u||l;if(!d){var p=sa(o,n),m=e.declarationNameToString(s),f=yy(s,c);if(void 0!==f){var _=ri(f),g=lr(s,e.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2,p,m,_);f.valueDeclaration&&e.addRelatedInfo(g,e.createDiagnosticForNode(f.valueDeclaration,e.Diagnostics._0_is_declared_here,_))}else(null===(a=o.exports)||void 0===a?void 0:a.has("default"))?lr(s,e.Diagnostics.Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead,p,m):function(n,t,r,a,i){var o,s=null===(o=a.valueDeclaration.locals)||void 0===o?void 0:o.get(t.escapedText),c=a.exports;if(s){var l=null==c?void 0:c.get("export=");if(l)Ia(l,s)?function(n,t,r,a){if(H>=e.ModuleKind.ES2015){lr(t,V.esModuleInterop?e.Diagnostics._0_can_only_be_imported_by_using_a_default_import:e.Diagnostics._0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import,r)}else{if(e.isInJSFile(n))lr(t,V.esModuleInterop?e.Diagnostics._0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import:e.Diagnostics._0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import,r);else lr(t,V.esModuleInterop?e.Diagnostics._0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import:e.Diagnostics._0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import,r,r,a)}}(n,t,r,i):lr(t,e.Diagnostics.Module_0_has_no_exported_member_1,i,r);else{var u=c?e.find(yc(c),(function(e){return!!Ia(e,s)})):void 0,d=u?lr(t,e.Diagnostics.Module_0_declares_1_locally_but_it_is_exported_as_2,i,r,ri(u)):lr(t,e.Diagnostics.Module_0_declares_1_locally_but_it_is_not_exported,i,r);e.addRelatedInfo.apply(void 0,__spreadArrays([d],e.map(s.declarations,(function(n,t){return e.createDiagnosticForNode(n,0===t?e.Diagnostics._0_is_declared_here:e.Diagnostics.and_here,r)}))))}}else lr(t,e.Diagnostics.Module_0_has_no_exported_member_1,i,r)}(n,s,m,o,p)}return d}}}function Jr(n){if(e.isVariableDeclaration(n)&&n.initializer&&e.isPropertyAccessExpression(n.initializer))return n.initializer}function Xr(e,n,t){var r=e.parent.parent.moduleSpecifier?qr(e.parent.parent,e,t):ca(e.propertyName||e.name,n,!1,t);return na(e,void 0,r,!1),r}function Yr(n,t){if(e.isClassExpression(n))return Lh(n).symbol;if(e.isEntityName(n)||e.isEntityNameExpression(n)){var r=ca(n,901119,!0,t);return r||(Lh(n),Sr(n).resolvedSymbol)}}function Qr(n,t){switch(void 0===t&&(t=!1),n.kind){case 260:case 249:return jr(n,t);case 262:return zr(n,t);case 263:return function(e,n){var t=e.parent.parent.moduleSpecifier,r=ua(e,t),a=_a(r,t,n,!1);return na(e,r,a,!1),a}(n,t);case 269:return function(e,n){var t=e.parent.moduleSpecifier,r=t&&ua(e,t),a=t&&_a(r,t,n,!1);return na(e,r,a,!1),a}(n,t);case 265:case 198:return function(n,t){var r=e.isBindingElement(n)?e.getRootDeclaration(n):n.parent.parent.parent,a=Jr(r),i=qr(r,a||n,t),o=n.propertyName||n.name;return a&&i&&e.isIdentifier(o)?$r(cc(lo(i),o.escapedText),t):(na(n,void 0,i,!1),i)}(n,t);case 270:return Xr(n,901119,t);case 266:case 216:return function(n,t){var r=Yr(e.isExportAssignment(n)?n.expression:n.right,t);return na(n,void 0,r,!1),r}(n,t);case 259:return function(e,n){var t=fa(e.parent.symbol,n);return na(e,void 0,t,!1),t}(n,t);case 289:return ca(n.name,901119,!0,t);case 288:return function(e,n){return Yr(e.initializer,n)}(n,t);case 202:case 201:return function(n,t){if(e.isBinaryExpression(n.parent)&&n.parent.left===n&&62===n.parent.operatorToken.kind)return Yr(n.parent.right,t)}(n,t);default:return e.Debug.fail()}}function Zr(e,n){return void 0===n&&(n=901119),!!e&&(2097152==(e.flags&(2097152|n))||!!(2097152&e.flags&&67108864&e.flags))}function $r(e,n){return!n&&Zr(e)?ea(e):e}function ea(n){e.Debug.assert(0!=(2097152&n.flags),"Should only get Alias here.");var t=Tr(n);if(t.target)t.target===be&&(t.target=he);else{t.target=be;var r=Br(n);if(!r)return e.Debug.fail();var a=Qr(r);t.target===be?t.target=a||he:lr(r,e.Diagnostics.Circular_definition_of_import_alias_0,ri(n))}return t.target}function na(n,t,r,a){if(!n||e.isPropertyAccessExpression(n))return!1;var i=xa(n);if(e.isTypeOnlyImportOrExportDeclaration(n))return Tr(i).typeOnlyDeclaration=n,!0;var o=Tr(i);return ta(o,t,a)||ta(o,r,a)}function ta(n,t,r){var a,i,o;if(t&&(void 0===n.typeOnlyDeclaration||r&&!1===n.typeOnlyDeclaration)){var s=null!==(i=null===(a=t.exports)||void 0===a?void 0:a.get("export="))&&void 0!==i?i:t,c=s.declarations&&e.find(s.declarations,e.isTypeOnlyImportOrExportDeclaration);n.typeOnlyDeclaration=null!==(o=null!=c?c:Tr(s).typeOnlyDeclaration)&&void 0!==o&&o}return!!n.typeOnlyDeclaration}function ra(e){if(2097152&e.flags)return Tr(e).typeOnlyDeclaration||void 0}function aa(e){var n=xa(e),t=ea(n);t&&((t===he||111551&t.flags&&!GT(t)&&!ra(n))&&ia(n))}function ia(n){var t=Tr(n);if(!t.referenced){t.referenced=!0;var r=Br(n);if(!r)return e.Debug.fail();if(e.isInternalModuleImportEqualsDeclaration(r)){var a=$r(n);(a===he||111551&a.flags)&&Lh(r.moduleReference)}}}function oa(n,t){return 78===n.kind&&e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent),78===n.kind||157===n.parent.kind?ca(n,1920,!1,t):(e.Debug.assert(260===n.parent.kind),ca(n,901119,!1,t))}function sa(e,n){return e.parent?sa(e.parent,n)+"."+ri(e):ri(e,n,void 0,20)}function ca(n,t,r,a,i){if(!e.nodeIsMissing(n)){var o,s=1920|(e.isInJSFile(n)?111551&t:0);if(78===n.kind){var c=t===s||e.nodeIsSynthesized(n)?e.Diagnostics.Cannot_find_namespace_0:mf(e.getFirstIdentifier(n)),l=e.isInJSFile(n)&&!e.nodeIsSynthesized(n)?function(n,t){if(pl(n.parent)){var r=function(n){if(e.findAncestor(n,(function(n){return e.isJSDocNode(n)||4194304&n.flags?e.isJSDocTypeAlias(n):"quit"})))return;var t=e.getJSDocHost(n);if(e.isExpressionStatement(t)&&e.isBinaryExpression(t.expression)&&3===e.getAssignmentDeclarationKind(t.expression)){if(a=xa(t.expression.left))return la(a)}if((e.isObjectLiteralMethod(t)||e.isPropertyAssignment(t))&&e.isBinaryExpression(t.parent.parent)&&6===e.getAssignmentDeclarationKind(t.parent.parent)){if(a=xa(t.parent.parent.left))return la(a)}var r=e.getEffectiveJSDocHost(n);if(r&&e.isFunctionLike(r)){var a;return(a=xa(r))&&a.valueDeclaration}}(n.parent);if(r)return Cr(r,n.escapedText,t,void 0,n,!0)}}(n,t):void 0;if(!(o=Sa(Cr(i||n,n.escapedText,t,r||l?void 0:c,n,!0))))return Sa(l)}else{if(157!==n.kind&&201!==n.kind)throw e.Debug.assertNever(n,"Unknown entity name kind.");var u=157===n.kind?n.left:n.expression,d=157===n.kind?n.right:n.name,p=ca(u,s,r,!1,i);if(!p||e.nodeIsMissing(d))return;if(p===he)return p;if(e.isInJSFile(n)&&p.valueDeclaration&&e.isVariableDeclaration(p.valueDeclaration)&&p.valueDeclaration.initializer&&Tv(p.valueDeclaration.initializer)){var m=p.valueDeclaration.initializer.arguments[0],f=ua(m,m);if(f){var _=fa(f);_&&(p=_)}}if(!(o=Sa(Lr(ha(p),d.escapedText,t)))){if(!r){var g=sa(p),y=e.declarationNameToString(d),v=yy(d,p);v?lr(d,e.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2,g,y,ri(v)):lr(d,e.Diagnostics.Namespace_0_has_no_exported_member_1,g,y)}return}}return e.Debug.assert(0==(1&e.getCheckFlags(o)),"Should never get an instantiated symbol here."),!e.nodeIsSynthesized(n)&&e.isEntityName(n)&&(2097152&o.flags||266===n.parent.kind)&&na(e.getAliasDeclarationFromName(n),o,void 0,!0),o.flags&t||a?o:ea(o)}}function la(n){var t=n.parent.valueDeclaration;if(t)return(e.isAssignmentDeclaration(t)?e.getAssignedExpandoInitializer(t):e.hasOnlyExpressionInitializer(t)?e.getDeclaredExpandoInitializer(t):void 0)||t}function ua(n,t,r){var a=e.getEmitModuleResolutionKind(V)===e.ModuleResolutionKind.Classic?e.Diagnostics.Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option:e.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations;return da(n,t,r?void 0:a)}function da(n,t,r,a){return void 0===a&&(a=!1),e.isStringLiteralLike(t)?pa(n,t.text,r,t,a):void 0}function pa(t,r,a,i,o){(void 0===o&&(o=!1),e.startsWith(r,"@types/"))&&lr(i,e.Diagnostics.Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1,e.removePrefix(r,"@types/"),r);var s=hc(r,!0);if(s)return s;var c=e.getSourceFileOfNode(t),l=e.getResolvedModule(c,r),u=l&&e.getResolutionDiagnostic(V,l),d=l&&!u&&n.getSourceFile(l.resolvedFileName);if(d)return d.symbol?(l.isExternalLibraryImport&&!e.resolutionExtensionIsTSOrJson(l.extension)&&ma(!1,i,l,r),Sa(d.symbol)):void(a&&lr(i,e.Diagnostics.File_0_is_not_a_module,d.fileName));if(_n){var p=e.findBestPatternMatch(_n,(function(e){return e.pattern}),r);if(p){var m=gn&&gn.get(r);return Sa(m?m:p.symbol)}}if(l&&!e.resolutionExtensionIsTSOrJson(l.extension)&&void 0===u||u===e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type)o?lr(i,e.Diagnostics.Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented,r,l.resolvedFileName):ma(X&&!!a,i,l,r);else if(a){if(l){var f=n.getProjectReferenceRedirect(l.resolvedFileName);if(f)return void lr(i,e.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1,f,l.resolvedFileName)}if(u)lr(i,u,r,l.resolvedFileName);else{var _=e.tryExtractTSExtension(r);if(_)lr(i,e.Diagnostics.An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead,_,e.removeExtension(r,_));else!V.resolveJsonModule&&e.fileExtensionIs(r,".json")&&e.getEmitModuleResolutionKind(V)===e.ModuleResolutionKind.NodeJs&&e.hasJsonModuleEmitEnabled(V)?lr(i,e.Diagnostics.Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension,r):lr(i,a,r)}}}function ma(n,t,r,a){var i,o=r.packageId,s=r.resolvedFileName,c=!e.isExternalModuleNameRelative(a)&&o?(i=o.name,d().has(e.getTypesPackageName(i))?e.chainDiagnosticMessages(void 0,e.Diagnostics.If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1,o.name,e.mangleScopedPackageName(o.name)):e.chainDiagnosticMessages(void 0,e.Diagnostics.Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0,a,e.mangleScopedPackageName(o.name))):void 0;dr(n,t,e.chainDiagnosticMessages(c,e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type,a,s))}function fa(n,t){if(null==n?void 0:n.exports){var r=function(n,t){if(!n||n===he||n===t||1===t.exports.size||2097152&n.flags)return n;var r=Tr(n);if(r.cjsExportMerged)return r.cjsExportMerged;var a=33554432&n.flags?n:gr(n);a.flags=512|a.flags,void 0===a.exports&&(a.exports=e.createSymbolTable());return t.exports.forEach((function(e,n){"export="!==n&&a.exports.set(n,a.exports.has(n)?yr(a.exports.get(n),e):e)})),Tr(a).cjsExportMerged=a,r.cjsExportMerged=a}(Sa($r(n.exports.get("export="),t)),Sa(n));return Sa(r)||n}}function _a(n,t,r,a){var i=fa(n,r);if(!r&&i){if(!(a||1539&i.flags||e.getDeclarationOfKind(i,297))){var o=H>=e.ModuleKind.ES2015?"allowSyntheticDefaultImports":"esModuleInterop";return lr(t,e.Diagnostics.This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export,o),i}if(V.esModuleInterop){var s=t.parent;if(e.isImportDeclaration(s)&&e.getNamespaceDeclarationNode(s)||e.isImportCall(s)){var c=lo(i),l=lc(c,0);if(l&&l.length||(l=lc(c,1)),l&&l.length){var u=Ev(c,i,n),d=mr(i.flags,i.escapedName);d.declarations=i.declarations?i.declarations.slice():[],d.parent=i.parent,d.target=i,d.originatingImport=s,i.valueDeclaration&&(d.valueDeclaration=i.valueDeclaration),i.constEnumOnlyModule&&(d.constEnumOnlyModule=!0),i.members&&(d.members=new e.Map(i.members)),i.exports&&(d.exports=new e.Map(i.exports));var p=Ps(u);return d.type=Va(d,p.members,e.emptyArray,e.emptyArray,p.stringIndexInfo,p.numberIndexInfo),d}}}}return i}function ga(e){return void 0!==e.exports.get("export=")}function ya(e){return yc(ba(e))}function va(e,n){var t=ba(n);if(t)return t.get(e)}function ha(e){return 6256&e.flags?es(e,"resolvedExports"):1536&e.flags?ba(e):e.exports||L}function ba(e){var n=Tr(e);return n.resolvedExports||(n.resolvedExports=Ta(e))}function Ea(n,t,r,a){t&&t.forEach((function(t,i){if("default"!==i){var o=n.get(i);if(o){if(r&&a&&o&&$r(o)!==$r(t)){var s=r.get(i);s.exportsWithDuplicate?s.exportsWithDuplicate.push(a):s.exportsWithDuplicate=[a]}}else n.set(i,t),r&&a&&r.set(i,{specifierText:e.getTextOfNode(a.moduleSpecifier)})}}))}function Ta(n){var t=[];return function n(r){if(!(r&&r.exports&&e.pushIfUnique(t,r)))return;var a=new e.Map(r.exports),i=r.exports.get("__export");if(i){for(var o=e.createSymbolTable(),s=new e.Map,c=0,l=i.declarations;c<l.length;c++){var u=l[c],d=ua(u,u.moduleSpecifier),p=n(d);Ea(o,p,s,u)}s.forEach((function(n,t){var r=n.exportsWithDuplicate;if("export="!==t&&r&&r.length&&!a.has(t))for(var i=0,o=r;i<o.length;i++){var c=o[i];Yt.add(e.createDiagnosticForNode(c,e.Diagnostics.Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity,s.get(t).specifierText,e.unescapeLeadingUnderscores(t)))}})),Ea(a,o)}return a}(n=fa(n))||L}function Sa(e){var n;return e&&e.mergeId&&(n=Mt[e.mergeId])?n:e}function xa(e){return Sa(e.symbol&&ts(e.symbol))}function La(e){return Sa(e.parent&&ts(e.parent))}function Aa(t,r,a){var i=La(t);if(i&&!(262144&t.flags)){var o=e.mapDefined(i.declarations,(function(e){return i&&Da(e,i)})),s=r&&function(t,r){var a,i=e.getSourceFileOfNode(r),o=R(i),s=Tr(t);if(s.extendedContainersByFile&&(a=s.extendedContainersByFile.get(o)))return a;if(i&&i.imports){for(var c=0,l=i.imports;c<l.length;c++){var u=l[c];if(!e.nodeIsSynthesized(u)){var d=ua(r,u,!0);d&&Ca(d,t)&&(a=e.append(a,d))}}if(e.length(a))return(s.extendedContainersByFile||(s.extendedContainersByFile=new e.Map)).set(o,a),a}if(s.extendedContainers)return s.extendedContainers;for(var p=0,m=n.getSourceFiles();p<m.length;p++){var f=m[p];if(e.isExternalModule(f)){var _=xa(f);Ca(_,t)&&(a=e.append(a,_))}}return s.extendedContainers=a||e.emptyArray}(t,r),c=function(n,t){var r=!!e.length(n.declarations)&&e.first(n.declarations);if(111551&t&&r&&r.parent&&e.isVariableDeclaration(r.parent)&&(e.isObjectLiteralExpression(r)&&r===r.parent.initializer||e.isTypeLiteralNode(r)&&r===r.parent.type))return xa(r.parent)}(i,a);if(r&&Ka(i,r,1920,!1))return e.append(e.concatenate(e.concatenate([i],o),s),c);var l=e.append(e.append(o,i),c);return e.concatenate(l,s)}var u=e.mapDefined(t.declarations,(function(n){return!e.isAmbientModule(n)&&n.parent&&ei(n.parent)?xa(n.parent):e.isClassExpression(n)&&e.isBinaryExpression(n.parent)&&62===n.parent.operatorToken.kind&&e.isAccessExpression(n.parent.left)&&e.isEntityNameExpression(n.parent.left.expression)?e.isModuleExportsAccessExpression(n.parent.left)||e.isExportsIdentifier(n.parent.left.expression)?xa(e.getSourceFileOfNode(n)):(Lh(n.parent.left.expression),Sr(n.parent.left.expression).resolvedSymbol):void 0}));if(e.length(u))return e.mapDefined(u,(function(e){return Ca(e,t)?e:void 0}))}function Da(e,n){var t=Za(e),r=t&&t.exports&&t.exports.get("export=");return r&&Ia(r,n)?t:void 0}function Ca(n,t){if(n===La(t))return t;var r=n.exports&&n.exports.get("export=");if(r&&Ia(r,t))return n;var a=ha(n),i=a.get(t.escapedName);return i&&Ia(i,t)?i:e.forEachEntry(a,(function(e){if(Ia(e,t))return e}))}function Ia(e,n){if(Sa($r(Sa(e)))===Sa($r(Sa(n))))return e}function Na(e){return Sa(e&&0!=(1048576&e.flags)?e.exportSymbol:e)}function Ra(e){return!!(111551&e.flags||2097152&e.flags&&111551&ea(e).flags&&!ra(e))}function ka(n){for(var t=0,r=n.members;t<r.length;t++){var a=r[t];if(166===a.kind&&e.nodeIsPresent(a.body))return a}}function Oa(e){var n=new m(se,e);return _++,n.id=_,x.push(n),n}function Pa(e,n,t){void 0===t&&(t=0);var r=Oa(e);return r.intrinsicName=n,r.objectFlags=t,r}function Ma(e){var n=Xl(e);return n.flags|=16,n.intrinsicName="boolean",n}function wa(e,n){var t=Oa(524288);return t.objectFlags=e,t.symbol=n,t.members=void 0,t.properties=void 0,t.callSignatures=void 0,t.constructSignatures=void 0,t.stringIndexInfo=void 0,t.numberIndexInfo=void 0,t}function Fa(e){var n=Oa(262144);return e&&(n.symbol=e),n}function Ga(e){return 95===e.charCodeAt(0)&&95===e.charCodeAt(1)&&95!==e.charCodeAt(2)&&64!==e.charCodeAt(2)&&35!==e.charCodeAt(2)}function Ba(n){var t;return n.forEach((function(e,n){!Ga(n)&&Ra(e)&&(t||(t=[])).push(e)})),t||e.emptyArray}function Ua(n,t,r,a,i,o){return n.members=t,n.properties=t===L?e.emptyArray:Ba(t),n.callSignatures=r,n.constructSignatures=a,n.stringIndexInfo=i,n.numberIndexInfo=o,n}function Va(e,n,t,r,a,i){return Ua(wa(16,e),n,t,r,a,i)}function ja(n,t){for(var r,a=function(n){if(n.locals&&!xr(n)&&(r=t(n.locals)))return{value:r};switch(n.kind){case 297:if(!e.isExternalOrCommonJsModule(n))break;case 256:var a=xa(n);if(r=t((null==a?void 0:a.exports)||L))return{value:r};break;case 252:case 221:case 253:var i;if((xa(n).members||L).forEach((function(n,t){788968&n.flags&&(i||(i=e.createSymbolTable())).set(t,n)})),i&&(r=t(i)))return{value:r}}},i=n;i;i=i.parent){var o=a(i);if("object"==typeof o)return o.value}return t(ne)}function Ha(e){return 111551===e?111551:1920}function Ka(n,t,r,a,i){if(void 0===i&&(i=new e.Map),n&&!function(e){if(e.declarations&&e.declarations.length){for(var n=0,t=e.declarations;n<t.length;n++){switch(t[n].kind){case 163:case 165:case 167:case 168:continue;default:return!1}}return!0}return!1}(n)){var o=k(n),s=i.get(o);return s||i.set(o,s=[]),ja(t,c)}function c(r,i){if(e.pushIfUnique(s,r)){var o=function(r,i){if(u(r.get(n.escapedName),void 0,i))return[n];return e.forEachEntry(r,(function(r){if(2097152&r.flags&&"export="!==r.escapedName&&"default"!==r.escapedName&&!(e.isUMDExportSymbol(r)&&t&&e.isExternalModule(e.getSourceFileOfNode(t)))&&(!a||e.some(r.declarations,e.isExternalModuleImportEqualsDeclaration))&&(i||!e.getDeclarationOfKind(r,270))){var o=d(r,ea(r),i);if(o)return o}if(r.escapedName===n.escapedName&&r.exportSymbol&&u(Sa(r.exportSymbol),void 0,i))return[n]}))||(r===ne?d(re,re,i):void 0)}(r,i);return s.pop(),o}}function l(e,n){return!Wa(e,t,n)||!!Ka(e.parent,t,Ha(n),a,i)}function u(t,a,i){return(n===(a||t)||Sa(n)===Sa(a||t))&&!e.some(t.declarations,ei)&&(i||l(Sa(t),r))}function d(e,n,t){if(u(e,n,t))return[e];var a=ha(n),i=a&&c(a,!0);return i&&l(e,Ha(r))?[e].concat(i):void 0}}function Wa(n,t,r){var a=!1;return ja(t,(function(t){var i=Sa(t.get(n.escapedName));return!!i&&(i===n||!!((i=2097152&i.flags&&!e.getDeclarationOfKind(i,270)?ea(i):i).flags&r)&&(a=!0,!0))})),a}function za(e,n){return 0===Qa(e,n,788968,!1,!0).accessibility}function qa(e,n){return 0===Qa(e,n,111551,!1,!0).accessibility}function Ja(e,n,t){return 0===Qa(e,n,t,!1,!1).accessibility}function Xa(n,t,r,a,i,o){if(e.length(n)){for(var s,c=!1,l=0,u=n;l<u.length;l++){var d=u[l],p=Ka(d,t,a,!1);if(p){s=d;var m=ni(p[0],i);if(m)return m}else if(o&&e.some(d.declarations,ei)){if(i){c=!0;continue}return{accessibility:0}}var f=Xa(Aa(d,t,a),t,r,r===d?Ha(a):a,i,o);if(f)return f}return c?{accessibility:0}:s?{accessibility:1,errorSymbolName:ri(r,t,a),errorModuleName:s!==r?ri(s,t,1920):void 0}:void 0}}function Ya(e,n,t,r){return Qa(e,n,t,r,!0)}function Qa(n,t,r,a,i){if(n&&t){var o=Xa([n],t,n,r,a,i);if(o)return o;var s=e.forEach(n.declarations,Za);if(s)if(s!==Za(t))return{accessibility:2,errorSymbolName:ri(n,t,r),errorModuleName:ri(s)};return{accessibility:1,errorSymbolName:ri(n,t,r)}}return{accessibility:0}}function Za(n){var t=e.findAncestor(n,$a);return t&&xa(t)}function $a(n){return e.isAmbientModule(n)||297===n.kind&&e.isExternalOrCommonJsModule(n)}function ei(n){return e.isModuleWithStringLiteralName(n)||297===n.kind&&e.isExternalOrCommonJsModule(n)}function ni(n,t){var r;if(e.every(e.filter(n.declarations,(function(e){return 78!==e.kind})),(function(t){var r,i;if(!yi(t)){var o=Gr(t);return o&&!e.hasSyntacticModifier(o,1)&&yi(o.parent)?a(t,o):e.isVariableDeclaration(t)&&e.isVariableStatement(t.parent.parent)&&!e.hasSyntacticModifier(t.parent.parent,1)&&yi(t.parent.parent.parent)?a(t,t.parent.parent):e.isLateVisibilityPaintedStatement(t)&&!e.hasSyntacticModifier(t,1)&&yi(t.parent)?a(t,t):!!(2097152&n.flags&&e.isBindingElement(t)&&e.isInJSFile(t)&&(null===(r=t.parent)||void 0===r?void 0:r.parent)&&e.isVariableDeclaration(t.parent.parent)&&(null===(i=t.parent.parent.parent)||void 0===i?void 0:i.parent)&&e.isVariableStatement(t.parent.parent.parent.parent)&&!e.hasSyntacticModifier(t.parent.parent.parent.parent,1)&&t.parent.parent.parent.parent.parent&&yi(t.parent.parent.parent.parent.parent))&&a(t,t.parent.parent.parent.parent)}return!0})))return{accessibility:0,aliasesToMakeVisible:r};function a(n,a){return t&&(Sr(n).isVisible=!0,r=e.appendIfUnique(r,a)),!0}}function ti(n,t){var r;r=176===n.parent.kind||e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)||158===n.parent.kind?1160127:157===n.kind||201===n.kind||260===n.parent.kind?1920:788968;var a=e.getFirstIdentifier(n),i=Cr(t,a.escapedText,r,void 0,void 0,!1);return i&&262144&i.flags&&788968&r?{accessibility:0}:i&&ni(i,!0)||{accessibility:1,errorSymbolName:e.getTextOfNode(a),errorNode:a}}function ri(n,t,r,a,i){void 0===a&&(a=4);var o=70221824;2&a&&(o|=128),1&a&&(o|=512),8&a&&(o|=16384),16&a&&(o|=134217728);var s=4&a?ee.symbolToExpression:ee.symbolToEntityName;return i?c(i).getText():e.usingSingleLineStringWriter(c);function c(a){var i=s(n,r,t,o),c=297===(null==t?void 0:t.kind)?e.createPrinter({removeComments:!0,neverAsciiEscape:!0}):e.createPrinter({removeComments:!0}),l=t&&e.getSourceFileOfNode(t);return c.writeNode(4,i,l,a),a}}function ai(n,t,r,a,i){return void 0===r&&(r=0),i?o(i).getText():e.usingSingleLineStringWriter(o);function o(i){var o;o=262144&r?1===a?175:174:1===a?170:169;var s=ee.signatureToSignatureDeclaration(n,o,t,70222336|li(r)),c=e.createPrinter({removeComments:!0,omitTrailingSemicolon:!0}),l=t&&e.getSourceFileOfNode(t);return c.writeNode(4,s,l,e.getTrailingSemicolonDeferringWriter(i)),i}}function ii(n,t,r,a){void 0===r&&(r=1064960),void 0===a&&(a=e.createTextWriter(""));var i=V.noErrorTruncation||1&r,o=ee.typeToTypeNode(n,t,70221824|li(r)|(i?1:0),a);if(void 0===o)return e.Debug.fail("should always get typenode");var s=e.createPrinter({removeComments:!0}),c=t&&e.getSourceFileOfNode(t);s.writeNode(4,o,c,a);var l=a.getText(),u=i?2*e.noTruncationMaximumTruncationLength:2*e.defaultMaximumTruncationLength;return u&&l&&l.length>=u?l.substr(0,u-"...".length)+"...":l}function oi(e,n){var t=ci(e.symbol)?ii(e,e.symbol.valueDeclaration):ii(e),r=ci(n.symbol)?ii(n,n.symbol.valueDeclaration):ii(n);return t===r&&(t=si(e),r=si(n)),[t,r]}function si(e){return ii(e,void 0,64)}function ci(n){return n&&n.valueDeclaration&&e.isExpression(n.valueDeclaration)&&!wd(n.valueDeclaration)}function li(e){return void 0===e&&(e=0),814775659&e}function ui(n){return!!(n.symbol&&32&n.symbol.flags&&(n===Co(n.symbol)||1073741824&e.getObjectFlags(n)))}function di(n,t,r,a){return void 0===r&&(r=16384),a?i(a).getText():e.usingSingleLineStringWriter(i);function i(a){var i=e.factory.createTypePredicateNode(2===n.kind||3===n.kind?e.factory.createToken(127):void 0,1===n.kind||3===n.kind?e.factory.createIdentifier(n.parameterName):e.factory.createThisTypeNode(),n.type&&ee.typeToTypeNode(n.type,t,70222336|li(r))),o=e.createPrinter({removeComments:!0}),s=t&&e.getSourceFileOfNode(t);return o.writeNode(4,i,s,a),a}}function pi(e){return 8===e?"private":16===e?"protected":"public"}function mi(n){return n&&n.parent&&257===n.parent.kind&&e.isExternalModuleAugmentation(n.parent.parent)}function fi(n){return 297===n.kind||e.isAmbientModule(n)}function _i(n,t){var r=Tr(n).nameType;if(r){if(384&r.flags){var a=""+r.value;return e.isIdentifierText(a,V.target)||yg(a)?yg(a)&&e.startsWith(a,"-")?"["+a+"]":a:'"'+e.escapeString(a,34)+'"'}if(8192&r.flags)return"["+gi(r.symbol,t)+"]"}}function gi(n,t){if(t&&"default"===n.escapedName&&!(16384&t.flags)&&(!(16777216&t.flags)||!n.declarations||t.enclosingDeclaration&&e.findAncestor(n.declarations[0],fi)!==e.findAncestor(t.enclosingDeclaration,fi)))return"default";if(n.declarations&&n.declarations.length){var r=e.firstDefined(n.declarations,(function(n){return e.getNameOfDeclaration(n)?n:void 0})),a=r&&e.getNameOfDeclaration(r);if(r&&a){if(e.isCallExpression(r)&&e.isBindableObjectDefinePropertyCall(r))return e.symbolName(n);if(e.isComputedPropertyName(a)&&!(4096&e.getCheckFlags(n))){var i=Tr(n).nameType;if(i&&384&i.flags){var o=_i(n,t);if(void 0!==o)return o}}return e.declarationNameToString(a)}if(r||(r=n.declarations[0]),r.parent&&249===r.parent.kind)return e.declarationNameToString(r.parent.name);switch(r.kind){case 221:case 208:case 209:return!t||t.encounteredError||131072&t.flags||(t.encounteredError=!0),221===r.kind?"(Anonymous class)":"(Anonymous function)"}}var s=_i(n,t);return void 0!==s?s:e.symbolName(n)}function yi(n){if(n){var t=Sr(n);return void 0===t.isVisible&&(t.isVisible=!!function(){switch(n.kind){case 324:case 331:case 325:return!!(n.parent&&n.parent.parent&&n.parent.parent.parent&&e.isSourceFile(n.parent.parent.parent));case 198:return yi(n.parent.parent);case 249:if(e.isBindingPattern(n.name)&&!n.name.elements.length)return!1;case 256:case 252:case 253:case 254:case 251:case 255:case 260:if(e.isExternalModuleAugmentation(n))return!0;var t=Si(n);return 1&e.getCombinedModifierFlags(n)||260!==n.kind&&297!==t.kind&&8388608&t.flags?yi(t):xr(t);case 163:case 162:case 167:case 168:case 165:case 164:if(e.hasEffectiveModifier(n,24))return!1;case 166:case 170:case 169:case 171:case 160:case 257:case 174:case 175:case 177:case 173:case 178:case 179:case 182:case 183:case 186:case 192:return yi(n.parent);case 262:case 263:case 265:return!1;case 159:case 297:case 259:return!0;case 266:default:return!1}}()),t.isVisible}return!1}function vi(n,t){var r,a,i;return n.parent&&266===n.parent.kind?r=Cr(n,n.escapedText,2998271,void 0,n,!1):270===n.parent.kind&&(r=Xr(n.parent,2998271)),r&&((i=new e.Set).add(k(r)),function n(r){e.forEach(r,(function(r){var o=Gr(r)||r;if(t?Sr(r).isVisible=!0:(a=a||[],e.pushIfUnique(a,o)),e.isInternalModuleImportEqualsDeclaration(r)){var s=r.moduleReference,c=Cr(r,e.getFirstIdentifier(s).escapedText,901119,void 0,void 0,!1);c&&i&&e.tryAddToSet(i,k(c))&&n(c.declarations)}}))}(r.declarations)),a}function hi(e,n){var t=bi(e,n);if(t>=0){for(var r=Rt.length,a=t;a<r;a++)kt[a]=!1;return!1}return Rt.push(e),kt.push(!0),Ot.push(n),!0}function bi(e,n){for(var t=Rt.length-1;t>=0;t--){if(Ei(Rt[t],Ot[t]))return-1;if(Rt[t]===e&&Ot[t]===n)return t}return-1}function Ei(n,t){switch(t){case 0:return!!Tr(n).type;case 5:return!!Sr(n).resolvedEnumType;case 2:return!!Tr(n).declaredType;case 1:return!!n.resolvedBaseConstructorType;case 3:return!!n.resolvedReturnType;case 4:return!!n.immediateBaseConstraint;case 6:return!!n.resolvedTypeArguments;case 7:return!!n.baseTypesResolved}return e.Debug.assertNever(t)}function Ti(){return Rt.pop(),Ot.pop(),kt.pop()}function Si(n){return e.findAncestor(e.getRootDeclaration(n),(function(e){switch(e.kind){case 249:case 250:case 265:case 264:case 263:case 262:return!1;default:return!0}})).parent}function xi(e,n){var t=cc(e,n);return t?lo(t):void 0}function Li(e){return e&&0!=(1&e.flags)}function Ai(e){var n=xa(e);return n&&Tr(n).type||Mi(e,!1)}function Di(n,t,r){if(131072&(n=Jf(n,(function(e){return!(98304&e.flags)}))).flags)return nn;if(1048576&n.flags)return Yf(n,(function(e){return Di(e,t,r)}));var a=Xl(e.map(t,ou));if(Tu(n)||Su(a)){if(131072&a.flags)return n;var i=Zn||(Zn=bl("Omit",524288,e.Diagnostics.Cannot_find_global_type_0));return i?al(i,[n,a]):xe}for(var o=e.createSymbolTable(),s=0,c=Gs(n);s<c.length;s++){var l=c[s];zd(su(l,8576),a)||24&e.getDeclarationModifierFlagsFromSymbol(l)||!Xu(l)||o.set(l.escapedName,Yu(l,!1))}var u=mc(n,0),d=mc(n,1),p=Va(r,o,e.emptyArray,e.emptyArray,u,d);return p.objectFlags|=131072,p}function Ci(e,n){var t=Ii(e);return t?y_(t,n):n}function Ii(n){var t=function(e){var n=e.parent.parent;switch(n.kind){case 198:case 288:return Ii(n);case 199:return Ii(e.parent);case 249:return n.initializer;case 216:return n.right}}(n);if(t&&t.flowNode){var r=function(e){var n=e.parent;if(198===e.kind&&196===n.kind)return Ni(e.propertyName||e.name);if(288===e.kind||289===e.kind)return Ni(e.name);return""+n.elements.indexOf(e)}(n);if(r){var a=e.setTextRange(e.parseNodeFactory.createStringLiteral(r),n),i=e.isLeftHandSideExpression(t)?t:e.parseNodeFactory.createParenthesizedExpression(t),o=e.setTextRange(e.parseNodeFactory.createElementAccessExpression(i,a),n);return e.setParent(a,o),e.setParent(o,n),i!==t&&e.setParent(i,o),o.flowNode=t.flowNode,o}}}function Ni(e){var n=ou(e);return 384&n.flags?""+n.value:void 0}function Ri(n){var t,r=n.parent,a=Ai(r.parent);if(!a||Li(a))return a;if(W&&8388608&n.flags&&e.isParameterDeclaration(n)?a=_m(a):!W||!r.parent.initializer||65536&Cf(Gf(r.parent.initializer))||(a=If(a,524288)),196===r.kind)if(n.dotDotDotToken){if(2&(a=rc(a)).flags||!Tg(a))return lr(n,e.Diagnostics.Rest_types_may_only_be_created_from_object_types),xe;for(var i=[],o=0,s=r.elements;o<s.length;o++){var c=s[o];c.dotDotDotToken||i.push(c.propertyName||c.name)}t=Di(a,i,n.symbol)}else{var l=n.propertyName||n.name;t=Ci(n,x_(Iu(a,p=ou(l),void 0,l,void 0,void 0,16),n.name))}else{var u=aE(65|(n.dotDotDotToken?0:128),a,Ce,r),d=r.elements.indexOf(n);if(n.dotDotDotToken)t=qf(a,tm)?Yf(a,(function(e){return Vl(e,d)})):Nl(u);else if(Kp(a)){var p=td(d),m=pg(n)?8:0;t=Ci(n,x_(Ru(a,p,void 0,n.name,16|m)||xe,n.name))}else t=u}return n.initializer?e.getEffectiveTypeAnnotationNode(e.walkUpBindingElementsAndPatterns(n))?!W||32768&um(Dh(n))?t:If(t,524288):Ch(n,Xl([If(t,524288),Dh(n)],2)):t}function ki(n){var t=e.getJSDocType(n);if(t)return sd(t)}function Oi(n){var t=e.skipParentheses(n);return 199===t.kind&&0===t.elements.length}function Pi(e,n){return void 0===n&&(n=!0),W&&n?fm(e):e}function Mi(n,t){if(e.isVariableDeclaration(n)&&238===n.parent.parent.kind){var r=uu(Qg(Hh(n.parent.parent.expression)));return 4456448&r.flags?du(r):Oe}if(e.isVariableDeclaration(n)&&239===n.parent.parent.kind)return rE(n.parent.parent)||Ee;if(e.isBindingPattern(n.parent))return Ri(n);var a,i,o=t&&(e.isParameter(n)&&vc(n)||Ec(n)||!e.isBindingElement(n)&&!e.isVariableDeclaration(n)&&!!n.questionToken),s=Zi(n);if(s)return Pi(s,o);if((X||e.isInJSFile(n))&&e.isVariableDeclaration(n)&&!e.isBindingPattern(n.name)&&!(1&e.getCombinedModifierFlags(n))&&!(8388608&n.flags)){if(!(2&e.getCombinedNodeFlags(n)||n.initializer&&(a=n.initializer,i=e.skipParentheses(a),103!==i.kind&&(78!==i.kind||ff(i)!==te))))return Te;if(n.initializer&&Oi(n.initializer))return In}if(e.isParameter(n)){var c=n.parent;if(168===c.kind&&!Qo(c)){var l=e.getDeclarationOfKind(xa(n.parent),167);if(l){var u=Lc(l),d=LS(c);return d&&n===d?(e.Debug.assert(!d.type),lo(u.thisParameter)):kc(u)}}if(e.isInJSFile(n)){var p=e.getJSDocType(c);if(p&&e.isFunctionTypeNode(p)){var m=Lc(p),f=c.parameters.indexOf(n);return n.dotDotDotToken?Mv(m,f):Ov(m,f)}}if(g="this"===n.symbol.escapedName?F_(c):G_(n))return Pi(g,o)}if(e.hasOnlyExpressionInitializer(n)&&n.initializer){if(e.isInJSFile(n)&&!e.isParameter(n)){var _=ji(n,xa(n),e.getDeclaredExpandoInitializer(n));if(_)return _}return Pi(g=Ch(n,Dh(n)),o)}if(e.isPropertyDeclaration(n)&&!e.hasStaticModifier(n)&&(X||e.isInJSFile(n))){var g,y=ka(n.parent);return(g=y?Bi(n.symbol,y):2&e.getEffectiveModifierFlags(n)?Op(n.symbol):void 0)&&Pi(g,o)}return e.isJsxAttribute(n)?Ge:e.isBindingPattern(n.name)?Ji(n.name,!1,!0):void 0}function wi(n){if(n.valueDeclaration&&e.isBinaryExpression(n.valueDeclaration)){var t=Tr(n);return void 0===t.isConstructorDeclaredProperty&&(t.isConstructorDeclaredProperty=!1,t.isConstructorDeclaredProperty=!!Gi(n)&&e.every(n.declarations,(function(t){return e.isBinaryExpression(t)&&z_(t)&&(202!==t.left.kind||e.isStringOrNumericLiteralLike(t.left.argumentExpression))&&!Hi(void 0,t,n,t)}))),t.isConstructorDeclaredProperty}return!1}function Fi(n){var t=n.valueDeclaration;return t&&e.isPropertyDeclaration(t)&&!e.getEffectiveTypeAnnotationNode(t)&&!t.initializer&&(X||e.isInJSFile(t))}function Gi(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t],i=e.getThisContainer(a,!1);if(i&&(166===i.kind||fv(i)))return i}}function Bi(n,t){var r=e.startsWith(n.escapedName,"__#")?e.factory.createPrivateIdentifier(n.escapedName.split("@")[1]):e.unescapeLeadingUnderscores(n.escapedName),a=e.factory.createPropertyAccessExpression(e.factory.createThis(),r);e.setParent(a.expression,a),e.setParent(a,t),a.flowNode=t.returnFlowNode;var i=Ui(a,n);return!X||i!==Te&&i!==In||lr(n.valueDeclaration,e.Diagnostics.Member_0_implicitly_has_an_1_type,ri(n),ii(i)),qf(i,Yg)?void 0:Wb(i)}function Ui(n,t){var r=t&&(!Fi(t)||2&e.getEffectiveModifierFlags(t.valueDeclaration))&&Op(t)||Ce;return y_(n,Te,r)}function Vi(n,t){var r,a=e.getAssignedExpandoInitializer(n.valueDeclaration);if(a){var i=e.getJSDocTypeTag(a);return i&&i.typeExpression?sd(i.typeExpression):ji(n.valueDeclaration,n,a)||Zp(Lh(a))}var o=!1,s=!1;if(wi(n)&&(r=Bi(n,Gi(n))),!r){for(var c=void 0,l=void 0,u=0,d=n.declarations;u<d.length;u++){var p=d[u],m=e.isBinaryExpression(p)||e.isCallExpression(p)?p:e.isAccessExpression(p)?e.isBinaryExpression(p.parent)?p.parent:p:void 0;if(m){var f=e.isAccessExpression(m)?e.getAssignmentDeclarationPropertyAccessKind(m):e.getAssignmentDeclarationKind(m);(4===f||e.isBinaryExpression(m)&&z_(m,f))&&(Wi(m)?o=!0:s=!0),e.isCallExpression(m)||(c=Hi(c,m,n,p)),c||(l||(l=[])).push(e.isBinaryExpression(m)||e.isCallExpression(m)?Ki(n,t,m,f):He)}}if(!(r=c)){if(!e.length(l))return xe;var _=o?function(n,t){return e.Debug.assert(n.length===t.length),n.filter((function(n,r){var a=t[r],i=e.isBinaryExpression(a)?a:e.isBinaryExpression(a.parent)?a.parent:void 0;return i&&Wi(i)}))}(l,n.declarations):void 0;if(s){var g=Op(n);g&&((_||(_=[])).push(g),o=!0)}r=Xl(e.some(_,(function(e){return!!(-98305&e.flags)}))?_:l,2)}}var y=Im(Pi(r,s&&!o));return Jf(y,(function(e){return!!(-98305&e.flags)}))===He?(km(n.valueDeclaration,Ee),Ee):y}function ji(n,t,r){var a,i;if(e.isInJSFile(n)&&r&&e.isObjectLiteralExpression(r)&&!r.properties.length){for(var o=e.createSymbolTable();e.isBinaryExpression(n)||e.isPropertyAccessExpression(n);){var s=xa(n);(null===(a=null==s?void 0:s.exports)||void 0===a?void 0:a.size)&&br(o,s.exports),n=e.isBinaryExpression(n)?n.parent:n.parent.parent}var c=xa(n);(null===(i=null==c?void 0:c.exports)||void 0===i?void 0:i.size)&&br(o,c.exports);var l=Va(t,o,e.emptyArray,e.emptyArray,void 0,void 0);return l.objectFlags|=16384,l}}function Hi(n,t,r,a){var i=e.getEffectiveTypeAnnotationNode(t.parent);if(i){var o=Im(sd(i));if(!n)return o;n===xe||o===xe||Vd(n,o)||qb(void 0,n,a,o)}if(r.parent){var s=e.getEffectiveTypeAnnotationNode(r.parent.valueDeclaration);if(s)return xi(sd(s),r.escapedName)}return n}function Ki(n,t,r,a){if(e.isCallExpression(r)){if(t)return lo(t);var i=Lh(r.arguments[2]),o=xi(i,"value");if(o)return o;var s=xi(i,"get");if(s){var c=Py(s);if(c)return kc(c)}var l=xi(i,"set");if(l){var u=Py(l);if(u)return Vv(u)}return Ee}if(function(n,t){return e.isPropertyAccessExpression(n)&&107===n.expression.kind&&e.forEachChildRecursively(t,(function(e){return yf(n,e)}))}(r.left,r.right))return Ee;var d=t?lo(t):Zp(Lh(r.right));if(524288&d.flags&&2===a&&"export="===n.escapedName){var p=Ps(d),m=e.createSymbolTable();e.copyEntries(p.members,m);var f=m.size;t&&!t.exports&&(t.exports=e.createSymbolTable()),(t||n).exports.forEach((function(n,t){var r,a=m.get(t);if(a&&a!==n)if(111551&n.flags&&111551&a.flags){if(e.getSourceFileOfNode(n.valueDeclaration)!==e.getSourceFileOfNode(a.valueDeclaration)){var i=e.unescapeLeadingUnderscores(n.escapedName),o=(null===(r=e.tryCast(a.valueDeclaration,e.isNamedDeclaration))||void 0===r?void 0:r.name)||a.valueDeclaration;e.addRelatedInfo(lr(n.valueDeclaration,e.Diagnostics.Duplicate_identifier_0,i),e.createDiagnosticForNode(o,e.Diagnostics._0_was_also_declared_here,i)),e.addRelatedInfo(lr(o,e.Diagnostics.Duplicate_identifier_0,i),e.createDiagnosticForNode(n.valueDeclaration,e.Diagnostics._0_was_also_declared_here,i))}var s=mr(n.flags|a.flags,t);s.type=Xl([lo(n),lo(a)]),s.valueDeclaration=a.valueDeclaration,s.declarations=e.concatenate(a.declarations,n.declarations),m.set(t,s)}else m.set(t,yr(n,a));else m.set(t,n)}));var _=Va(f!==m.size?void 0:p.symbol,m,p.callSignatures,p.constructSignatures,p.stringIndexInfo,p.numberIndexInfo);return _.objectFlags|=16384&e.getObjectFlags(d),_.symbol&&32&_.symbol.flags&&d===Co(_.symbol)&&(_.objectFlags|=1073741824),_}return Wp(d)?(km(r,Cn),Cn):d}function Wi(n){var t=e.getThisContainer(n,!1);return 166===t.kind||251===t.kind||208===t.kind&&!e.isPrototypePropertyAssignment(t.parent)}function zi(n,t,r){return n.initializer?Pi(Ch(n,Dh(n,e.isBindingPattern(n.name)?Ji(n.name,!0,!1):De))):e.isBindingPattern(n.name)?Ji(n.name,t,r):(r&&!Qi(n)&&km(n,Ee),t?Le:Ee)}function qi(n,t,r){var a,i=n.elements,o=e.lastOrUndefined(i),s=o&&198===o.kind&&o.dotDotDotToken?o:void 0;if(0===i.length||1===i.length&&s)return j>=2?(a=Ee,Cl(Al(!0),[a])):Cn;var c=e.map(i,(function(n){return e.isOmittedExpression(n)?Ee:zi(n,t,r)})),l=e.findLastIndex(i,(function(n){return!(n===s||e.isOmittedExpression(n)||pg(n))}),i.length-1)+1,u=Fl(c,e.map(i,(function(e,n){return e===s?4:n>=l?2:1})));return t&&((u=$c(u)).pattern=n,u.objectFlags|=1048576),u}function Ji(n,t,r){return void 0===t&&(t=!1),void 0===r&&(r=!1),196===n.kind?function(n,t,r){var a,i=e.createSymbolTable(),o=1048704;e.forEach(n.elements,(function(e){var n=e.propertyName||e.name;if(e.dotDotDotToken)a=Wc(Ee,!1);else{var s=ou(n);if(qo(s)){var c=Zo(s),l=mr(4|(e.initializer?16777216:0),c);l.type=zi(e,t,r),l.bindingElement=e,i.set(l.escapedName,l)}else o|=512}}));var s=Va(void 0,i,e.emptyArray,e.emptyArray,a,void 0);return s.objectFlags|=o,t&&(s.pattern=n,s.objectFlags|=1048576),s}(n,t,r):qi(n,t,r)}function Xi(e,n){return Yi(Mi(e,!0),e,n)}function Yi(n,t,r){return n?(r&&Om(t,n),8192&n.flags&&(e.isBindingElement(t)||!t.type)&&n.symbol!==xa(t)&&(n=Ve),Im(n)):(n=e.isParameter(t)&&t.dotDotDotToken?Cn:Ee,r&&(Qi(t)||km(t,n)),n)}function Qi(n){var t=e.getRootDeclaration(n);return ob(160===t.kind?t.parent:t)}function Zi(n){var t=e.getEffectiveTypeAnnotationNode(n);if(t)return sd(t)}function $i(n){var t=Tr(n);if(!t.type){var r=function(n){if(4194304&n.flags)return(t=Fo(La(n))).typeParameters?Zc(t,e.map(t.typeParameters,(function(e){return Ee}))):t;var t;if(n===oe)return Ee;if(134217728&n.flags){var r=xa(e.getSourceFileOfNode(n.valueDeclaration)),a=e.createSymbolTable();return a.set("exports",r),Va(n,a,e.emptyArray,e.emptyArray,void 0,void 0)}var i,o=n.valueDeclaration;if(e.isCatchClauseVariableDeclarationOrBindingElement(o)){var s=o;if(!s.type)return Ee;var c=ET(s.type);return Li(c)||c===De?c:xe}if(e.isSourceFile(o)&&e.isJsonSourceFile(o))return o.statements.length?Im(Zp(Hh(o.statements[0].expression))):nn;if(!hi(n,0))return 512&n.flags&&!(67108864&n.flags)?oo(n):co(n);if(266===o.kind)i=Yi(Lh(o.expression),o);else if(e.isBinaryExpression(o)||e.isInJSFile(o)&&(e.isCallExpression(o)||(e.isPropertyAccessExpression(o)||e.isBindableStaticElementAccessExpression(o))&&e.isBinaryExpression(o.parent)))i=Vi(n);else if(e.isPropertyAccessExpression(o)||e.isElementAccessExpression(o)||e.isIdentifier(o)||e.isStringLiteralLike(o)||e.isNumericLiteral(o)||e.isClassDeclaration(o)||e.isFunctionDeclaration(o)||e.isMethodDeclaration(o)&&!e.isObjectLiteralMethod(o)||e.isMethodSignature(o)||e.isSourceFile(o)){if(9136&n.flags)return oo(n);i=e.isBinaryExpression(o.parent)?Vi(n):Zi(o)||Ee}else if(e.isPropertyAssignment(o))i=Zi(o)||kh(o);else if(e.isJsxAttribute(o))i=Zi(o)||Lg(o);else if(e.isShorthandPropertyAssignment(o))i=Zi(o)||Rh(o.name,0);else if(e.isObjectLiteralMethod(o))i=Zi(o)||Oh(o,0);else if(e.isParameter(o)||e.isPropertyDeclaration(o)||e.isPropertySignature(o)||e.isVariableDeclaration(o)||e.isBindingElement(o)||e.isJSDocPropertyLikeTag(o))i=Xi(o,!0);else if(e.isEnumDeclaration(o))i=oo(n);else if(e.isEnumMember(o))i=so(n);else{if(!e.isAccessor(o))return e.Debug.fail("Unhandled declaration kind! "+e.Debug.formatSyntaxKind(o.kind)+" for "+e.Debug.formatSymbol(n));i=ao(n)}if(!Ti())return 512&n.flags&&!(67108864&n.flags)?oo(n):co(n);return i}(n);t.type||(t.type=r)}return t.type}function eo(n){if(n)return 167===n.kind?e.getEffectiveReturnTypeNode(n):e.getEffectiveSetAccessorTypeAnnotationNode(n)}function no(e){var n=eo(e);return n&&sd(n)}function to(e){return Nc(Lc(e))}function ro(n){var t=Tr(n);return t.type||(t.type=function(n){if(!hi(n,0))return xe;var t=ao(n);if(!Ti()){if(t=Ee,X)lr(e.getDeclarationOfKind(n,167),e.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions,ri(n))}return t}(n))}function ao(n){var t=e.getDeclarationOfKind(n,167),r=e.getDeclarationOfKind(n,168);if(t&&e.isInJSFile(t)){var a=ki(t);if(a)return a}var i=no(t);if(i)return i;var o=no(r);return o||(t&&t.body?Xv(t):(r?ob(r)||dr(X,r,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation,ri(n)):(e.Debug.assert(!!t,"there must exist a getter as we are current checking either setter or getter in this function"),ob(t)||dr(X,t,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation,ri(n))),Ee))}function io(n){var t=So(Co(n));return 8650752&t.flags?t:2097152&t.flags?e.find(t.types,(function(e){return!!(8650752&e.flags)})):void 0}function oo(n){var t=Tr(n),r=t;if(!t.type){var a=n.valueDeclaration&&gv(n.valueDeclaration,!1);if(a){var i=_v(n,a);i&&(n=t=i)}r.type=t.type=function(n){var t=n.valueDeclaration;if(1536&n.flags&&e.isShorthandAmbientModuleSymbol(n))return Ee;if(t&&(216===t.kind||e.isAccessExpression(t)&&216===t.parent.kind))return Vi(n);if(512&n.flags&&t&&e.isSourceFile(t)&&t.commonJsModuleIndicator){var r=fa(n);if(r!==n){if(!hi(n,0))return xe;var a=Sa(n.exports.get("export=")),i=Vi(a,a===r?void 0:r);return Ti()?i:co(n)}}var o=wa(16,n);if(32&n.flags){var s=io(n);return s?tu([o,s]):o}return W&&16777216&n.flags?fm(o):o}(n)}return t.type}function so(e){var n=Tr(e);return n.type||(n.type=Mo(e))}function co(n){var t=n.valueDeclaration;return e.getEffectiveTypeAnnotationNode(t)?(lr(n.valueDeclaration,e.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_type_annotation,ri(n)),xe):(X&&(160!==t.kind||t.initializer)&&lr(n.valueDeclaration,e.Diagnostics._0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer,ri(n)),Ee)}function lo(n){var t=e.getCheckFlags(n);return 65536&t?function(n){var t=Tr(n);return t.type||(e.Debug.assertIsDefined(t.deferralParent),e.Debug.assertIsDefined(t.deferralConstituents),t.type=1048576&t.deferralParent.flags?Xl(t.deferralConstituents):tu(t.deferralConstituents)),t.type}(n):1&t?function(e){var n=Tr(e);if(!n.type){if(!hi(e,0))return n.type=xe;var t=kd(lo(n.target),n.mapper);Ti()||(t=co(e)),n.type=t}return n.type}(n):262144&t?function(n){if(!n.type){var t=n.mappedType;if(!hi(n,0))return t.containsError=!0,xe;var r=kd(As(t.target||t),Ed(t.mapper,Ss(t),n.keyType)),a=W&&16777216&n.flags&&!dh(r,49152)?fm(r):524288&n.checkFlags?If(r,524288):r;Ti()||(lr(c,e.Diagnostics.Type_of_property_0_circularly_references_itself_in_mapped_type_1,ri(n),ii(t)),a=xe),n.type=a}return n.type}(n):8192&t?function(e){return Jm(e.propertyType,e.mappedType,e.constraintType)}(n):7&n.flags?$i(n):9136&n.flags?oo(n):8&n.flags?so(n):98304&n.flags?ro(n):2097152&n.flags?function(e){var n=Tr(e);if(!n.type){var t=ea(e);n.type=111551&t.flags?lo(t):xe}return n.type}(n):xe}function uo(n,t){return void 0!==n&&void 0!==t&&0!=(4&e.getObjectFlags(n))&&n.target===t}function po(n){return 4&e.getObjectFlags(n)?n.target:n}function mo(n,t){return function n(r){if(7&e.getObjectFlags(r)){var a=po(r);return a===t||e.some(Lo(a),n)}if(2097152&r.flags)return e.some(r.types,n);return!1}(n)}function fo(n,t){for(var r=0,a=t;r<a.length;r++){var i=a[r];n=e.appendIfUnique(n,wo(xa(i)))}return n}function _o(n,t){for(;;){if((n=n.parent)&&e.isBinaryExpression(n)){var r=e.getAssignmentDeclarationKind(n);if(6===r||3===r){var a=xa(n.left);a&&a.parent&&!e.findAncestor(a.parent.valueDeclaration,(function(e){return n===e}))&&(n=a.parent.valueDeclaration)}}if(!n)return;switch(n.kind){case 232:case 252:case 221:case 253:case 169:case 170:case 164:case 174:case 175:case 308:case 251:case 165:case 208:case 209:case 254:case 330:case 331:case 325:case 324:case 190:case 184:var i=_o(n,t);if(190===n.kind)return e.append(i,wo(xa(n.typeParameter)));if(184===n.kind)return e.concatenate(i,Gu(n));if(232===n.kind&&!e.isInJSFile(n))break;var o=fo(i,e.getEffectiveTypeParameterDeclarations(n)),s=t&&(252===n.kind||221===n.kind||253===n.kind||fv(n))&&Co(xa(n)).thisType;return s?e.append(o,s):o;case 326:var c=e.getParameterSymbolFromJSDoc(n);c&&(n=c.valueDeclaration)}}}function go(n){var t=32&n.flags?n.valueDeclaration:e.getDeclarationOfKind(n,253);return e.Debug.assert(!!t,"Class was missing valueDeclaration -OR- non-class had no interface declarations"),_o(t)}function yo(n){for(var t,r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(253===i.kind||252===i.kind||221===i.kind||fv(i)||e.isTypeAlias(i)){var o=i;t=fo(t,e.getEffectiveTypeParameterDeclarations(o))}}return t}function vo(e){var n=uc(e,1);if(1===n.length){var t=n[0];return!t.typeParameters&&1===t.parameters.length&&G(t)&&Hp(Cv(t.parameters[0]))===Ee}return!1}function ho(e){if(uc(e,1).length>0)return!0;if(8650752&e.flags){var n=Ws(e);return!!n&&vo(n)}return!1}function bo(n){return e.getEffectiveBaseTypeNode(n.symbol.valueDeclaration)}function Eo(n,t,r){var a=e.length(t),i=e.isInJSFile(r);return e.filter(uc(n,1),(function(n){return(i||a>=Sc(n.typeParameters))&&a<=e.length(n.typeParameters)}))}function To(n,t,r){var a=Eo(n,t,r),i=e.map(t,sd);return e.sameMap(a,(function(n){return e.some(n.typeParameters)?wc(n,i,e.isInJSFile(r)):n}))}function So(n){if(!n.resolvedBaseConstructorType){var t=n.symbol.valueDeclaration,r=e.getEffectiveBaseTypeNode(t),a=bo(n);if(!a)return n.resolvedBaseConstructorType=Ce;if(!hi(n,1))return xe;var i=Hh(a.expression);if(r&&a!==r&&(e.Debug.assert(!r.typeArguments),Hh(r.expression)),2621440&i.flags&&Ps(i),!Ti())return lr(n.symbol.valueDeclaration,e.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_base_expression,ri(n.symbol)),n.resolvedBaseConstructorType=xe;if(!(1&i.flags||i===ke||ho(i))){var o=lr(a.expression,e.Diagnostics.Type_0_is_not_a_constructor_function_type,ii(i));if(262144&i.flags){var s=Jc(i),c=De;if(s){var l=uc(s,1);l[0]&&(c=kc(l[0]))}e.addRelatedInfo(o,e.createDiagnosticForNode(i.symbol.declarations[0],e.Diagnostics.Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1,ri(i.symbol),ii(c)))}return n.resolvedBaseConstructorType=xe}n.resolvedBaseConstructorType=i}return n.resolvedBaseConstructorType}function xo(n,t){lr(n,e.Diagnostics.Type_0_recursively_references_itself_as_a_base_type,ii(t,void 0,2))}function Lo(n){if(!n.baseTypesResolved){if(hi(n,7)&&(8&n.objectFlags?n.resolvedBaseTypes=[Ao(n)]:96&n.symbol.flags?(32&n.symbol.flags&&function(n){n.resolvedBaseTypes=e.resolvingEmptyArray;var t=Zs(So(n));if(!(2621441&t.flags))return n.resolvedBaseTypes=e.emptyArray;var r,a=bo(n),i=t.symbol?Fo(t.symbol):void 0;if(t.symbol&&32&t.symbol.flags&&function(e){var n=e.outerTypeParameters;if(n){var t=n.length-1,r=nl(e);return n[t].symbol!==r[t].symbol}return!0}(i))r=rl(a,t.symbol);else if(1&t.flags)r=t;else{var o=To(t,a.typeArguments,a);if(!o.length)return lr(a.expression,e.Diagnostics.No_base_constructor_has_the_specified_number_of_type_arguments),n.resolvedBaseTypes=e.emptyArray;r=kc(o[0])}if(r===xe)return n.resolvedBaseTypes=e.emptyArray;var s=rc(r);if(!Do(s)){var c=sc(void 0,r),l=e.chainDiagnosticMessages(c,e.Diagnostics.Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members,ii(s));return Yt.add(e.createDiagnosticForNodeFromMessageChain(a.expression,l)),n.resolvedBaseTypes=e.emptyArray}if(n===s||mo(s,n))return lr(n.symbol.valueDeclaration,e.Diagnostics.Type_0_recursively_references_itself_as_a_base_type,ii(n,void 0,2)),n.resolvedBaseTypes=e.emptyArray;n.resolvedBaseTypes===e.resolvingEmptyArray&&(n.members=void 0);n.resolvedBaseTypes=[s]}(n),64&n.symbol.flags&&function(n){n.resolvedBaseTypes=n.resolvedBaseTypes||e.emptyArray;for(var t=0,r=n.symbol.declarations;t<r.length;t++){var a=r[t];if(253===a.kind&&e.getInterfaceBaseTypeNodes(a))for(var i=0,o=e.getInterfaceBaseTypeNodes(a);i<o.length;i++){var s=o[i],c=rc(sd(s));c!==xe&&(Do(c)?n===c||mo(c,n)?xo(a,n):n.resolvedBaseTypes===e.emptyArray?n.resolvedBaseTypes=[c]:n.resolvedBaseTypes.push(c):lr(s,e.Diagnostics.An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members))}}}(n)):e.Debug.fail("type must be class or interface"),!Ti()))for(var t=0,r=n.symbol.declarations;t<r.length;t++){var a=r[t];252!==a.kind&&253!==a.kind||xo(a,n)}n.baseTypesResolved=!0}return n.resolvedBaseTypes}function Ao(n){return Nl(Xl(e.sameMap(n.typeParameters,(function(e,t){return 8&n.elementFlags[t]?Iu(e,Pe):e}))||e.emptyArray),n.readonly)}function Do(n){if(262144&n.flags){var t=Ws(n);if(t)return Do(t)}return!!(67633153&n.flags&&!Os(n)||2097152&n.flags&&e.every(n.types,Do))}function Co(n){var t,r,a,i,o,s=Tr(n),c=s;if(!s.declaredType){var l=32&n.flags?1:2,u=_v(n,(t=n.valueDeclaration,a=t&&gv(t,!0),i=null===(r=null==a?void 0:a.exports)||void 0===r?void 0:r.get("prototype"),(o=(null==i?void 0:i.valueDeclaration)&&function(n){if(!n.parent)return!1;for(var t=n.parent;t&&201===t.kind;)t=t.parent;if(t&&e.isBinaryExpression(t)&&e.isPrototypeAccess(t.left)&&62===t.operatorToken.kind){var r=e.getInitializerOfBinaryExpression(t);return e.isObjectLiteralExpression(r)&&r}}(i.valueDeclaration))?xa(o):void 0));u&&(n=s=u);var d=c.declaredType=s.declaredType=wa(l,n),p=go(n),m=yo(n);(p||m||1===l||!function(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t];if(253===a.kind){if(128&a.flags)return!1;var i=e.getInterfaceBaseTypeNodes(a);if(i)for(var o=0,s=i;o<s.length;o++){var c=s[o];if(e.isEntityNameExpression(c.expression)){var l=ca(c.expression,788968,!0);if(!l||!(64&l.flags)||Co(l).thisType)return!1}}}}return!0}(n))&&(d.objectFlags|=4,d.typeParameters=e.concatenate(p,m),d.outerTypeParameters=p,d.localTypeParameters=m,d.instantiations=new e.Map,d.instantiations.set(Yc(d.typeParameters),d),d.target=d,d.resolvedTypeArguments=d.typeParameters,d.thisType=Fa(n),d.thisType.isThisType=!0,d.thisType.constraint=d)}return s.declaredType}function Io(n){var t=Tr(n);if(!t.declaredType){if(!hi(n,2))return xe;var r=e.Debug.checkDefined(e.find(n.declarations,e.isTypeAlias),"Type alias symbol with no valid declaration found"),a=e.isJSDocTypeAlias(r)?r.typeExpression:r.type,i=a?sd(a):xe;if(Ti()){var o=yo(n);o&&(t.typeParameters=o,t.instantiations=new e.Map,t.instantiations.set(Yc(o),i))}else i=xe,lr(e.isNamedDeclaration(r)?r.name:r||r,e.Diagnostics.Type_alias_0_circularly_references_itself,ri(n));t.declaredType=i}return t.declaredType}function No(n){return!!e.isStringLiteralLike(n)||216===n.kind&&(No(n.left)&&No(n.right))}function Ro(n){var t=n.initializer;if(!t)return!(8388608&n.flags);switch(t.kind){case 10:case 8:case 14:return!0;case 214:return 40===t.operator&&8===t.operand.kind;case 78:return e.nodeIsMissing(t)||!!xa(n.parent).exports.get(t.escapedText);case 216:return No(t);default:return!1}}function ko(n){var t=Tr(n);if(void 0!==t.enumKind)return t.enumKind;for(var r=!1,a=0,i=n.declarations;a<i.length;a++){var o=i[a];if(255===o.kind)for(var s=0,c=o.members;s<c.length;s++){var l=c[s];if(l.initializer&&e.isStringLiteralLike(l.initializer))return t.enumKind=1;Ro(l)||(r=!0)}}return t.enumKind=r?0:1}function Oo(e){return 1024&e.flags&&!(1048576&e.flags)?Fo(La(e.symbol)):e}function Po(e){var n=Tr(e);if(n.declaredType)return n.declaredType;if(1===ko(e)){y++;for(var t=[],r=0,a=e.declarations;r<a.length;r++){var i=a[r];if(255===i.kind)for(var o=0,s=i.members;o<s.length;o++){var c=s[o],l=zT(c),u=$u(td(void 0!==l?l:0,y,xa(c)));Tr(xa(c)).declaredType=u,t.push(ed(u))}}if(t.length){var d=Xl(t,1,e,void 0);return 1048576&d.flags&&(d.flags|=1024,d.symbol=e),n.declaredType=d}}var p=Oa(32);return p.symbol=e,n.declaredType=p}function Mo(e){var n=Tr(e);if(!n.declaredType){var t=Po(La(e));n.declaredType||(n.declaredType=t)}return n.declaredType}function wo(e){var n=Tr(e);return n.declaredType||(n.declaredType=Fa(e))}function Fo(e){return Go(e)||xe}function Go(e){return 96&e.flags?Co(e):524288&e.flags?Io(e):262144&e.flags?wo(e):384&e.flags?Po(e):8&e.flags?Mo(e):2097152&e.flags?function(e){var n=Tr(e);return n.declaredType||(n.declaredType=Fo(ea(e)))}(e):void 0}function Bo(e){switch(e.kind){case 128:case 152:case 147:case 144:case 155:case 131:case 148:case 145:case 113:case 150:case 141:case 191:return!0;case 178:return Bo(e.elementType);case 173:return!e.typeArguments||e.typeArguments.every(Bo)}return!1}function Uo(n){var t=e.getEffectiveConstraintOfTypeParameter(n);return!t||Bo(t)}function Vo(n){var t=e.getEffectiveTypeAnnotationNode(n);return t?Bo(t):!e.hasInitializer(n)}function jo(n){if(n.declarations&&1===n.declarations.length){var t=n.declarations[0];if(t)switch(t.kind){case 163:case 162:return Vo(t);case 165:case 164:case 166:case 167:case 168:return r=t,a=e.getEffectiveReturnTypeNode(r),i=e.getEffectiveTypeParameterDeclarations(r),(166===r.kind||!!a&&Bo(a))&&r.parameters.every(Vo)&&i.every(Uo)}}var r,a,i;return!1}function Ho(n,t,r){for(var a=e.createSymbolTable(),i=0,o=n;i<o.length;i++){var s=o[i];a.set(s.escapedName,r&&jo(s)?s:xd(s,t))}return a}function Ko(e,n){for(var t=0,r=n;t<r.length;t++){var a=r[t];e.has(a.escapedName)||Wo(a)||e.set(a.escapedName,a)}}function Wo(n){return!!n.valueDeclaration&&e.isPrivateIdentifierPropertyDeclaration(n.valueDeclaration)&&e.hasSyntacticModifier(n.valueDeclaration,32)}function zo(n){if(!n.declaredProperties){var t=n.symbol,r=ns(t);n.declaredProperties=Ba(r),n.declaredCallSignatures=e.emptyArray,n.declaredConstructSignatures=e.emptyArray,n.declaredCallSignatures=Cc(r.get("__call")),n.declaredConstructSignatures=Cc(r.get("__new")),n.declaredStringIndexInfo=zc(t,0),n.declaredNumberIndexInfo=zc(t,1)}return n}function qo(e){return!!(8576&e.flags)}function Jo(n){if(!e.isComputedPropertyName(n)&&!e.isElementAccessExpression(n))return!1;var t=e.isComputedPropertyName(n)?n.expression:n.argumentExpression;return e.isEntityNameExpression(t)&&qo(e.isComputedPropertyName(n)?vg(n):Lh(t))}function Xo(e){return 95===e.charCodeAt(0)&&95===e.charCodeAt(1)&&64===e.charCodeAt(2)}function Yo(n){var t=e.getNameOfDeclaration(n);return!!t&&Jo(t)}function Qo(n){return e.hasDynamicName(n)&&!Yo(n)}function Zo(n){return 8192&n.flags?n.escapedName:384&n.flags?e.escapeLeadingUnderscores(""+n.value):e.Debug.fail()}function $o(n,t,r,a){e.Debug.assert(!!a.symbol,"The member is expected to have a symbol.");var i=Sr(a);if(!i.resolvedSymbol){i.resolvedSymbol=a.symbol;var o=e.isBinaryExpression(a)?a.left:a.name,s=e.isElementAccessExpression(o)?Lh(o.argumentExpression):vg(o);if(qo(s)){var c=Zo(s),l=a.symbol.flags,u=r.get(c);u||r.set(c,u=mr(0,c,4096));var d=t&&t.get(c);if(u.flags&fr(l)||d){var p=d?e.concatenate(d.declarations,u.declarations):u.declarations,m=!(8192&s.flags)&&e.unescapeLeadingUnderscores(c)||e.declarationNameToString(o);e.forEach(p,(function(n){return lr(e.getNameOfDeclaration(n)||n,e.Diagnostics.Property_0_was_also_declared_here,m)})),lr(o||a,e.Diagnostics.Duplicate_property_0,m),u=mr(0,c,4096)}return u.nameType=s,function(n,t,r){e.Debug.assert(!!(4096&e.getCheckFlags(n)),"Expected a late-bound symbol."),n.flags|=r,Tr(t.symbol).lateSymbol=n,n.declarations?n.declarations.push(t):n.declarations=[t],111551&r&&(n.valueDeclaration&&n.valueDeclaration.kind===t.kind||(n.valueDeclaration=t))}(u,a,l),u.parent?e.Debug.assert(u.parent===n,"Existing symbol parent should match new one"):u.parent=n,i.resolvedSymbol=u}}return i.resolvedSymbol}function es(n,t){var r=Tr(n);if(!r[t]){var a="resolvedExports"===t,i=a?1536&n.flags?Ta(n):n.exports:n.members;r[t]=i||L;for(var o=e.createSymbolTable(),s=0,c=n.declarations||e.emptyArray;s<c.length;s++){var l=c[s],u=e.getMembersOfDeclaration(l);if(u)for(var d=0,p=u;d<p.length;d++){var m=p[d];a===e.hasStaticModifier(m)&&Yo(m)&&$o(n,i,o,m)}}var f=n.assignmentDeclarationMembers;if(f)for(var _=0,g=e.arrayFrom(f.values());_<g.length;_++){m=g[_];var y=e.getAssignmentDeclarationKind(m);a===!(3===y||e.isBinaryExpression(m)&&z_(m,y)||9===y||6===y)&&Yo(m)&&$o(n,i,o,m)}r[t]=function(n,t){if(!(null==n?void 0:n.size))return t;if(!(null==t?void 0:t.size))return n;var r=e.createSymbolTable();return br(r,n),br(r,t),r}(i,o)||L}return r[t]}function ns(e){return 6256&e.flags?es(e,"resolvedMembers"):e.members||L}function ts(n){if(106500&n.flags&&"__computed"===n.escapedName){var t=Tr(n);if(!t.lateSymbol&&e.some(n.declarations,Yo)){var r=Sa(n.parent);e.some(n.declarations,e.hasStaticModifier)?ha(r):ns(r)}return t.lateSymbol||(t.lateSymbol=n)}return n}function rs(n,t,r){if(4&e.getObjectFlags(n)){var a=n.target,i=nl(n);if(e.length(a.typeParameters)===e.length(i)){var o=Zc(a,e.concatenate(i,[t||a.thisType]));return r?Zs(o):o}}else if(2097152&n.flags)return tu(e.map(n.types,(function(e){return rs(e,t,r)})));return r?Zs(n):n}function as(n,t,r,a){var i,o,s,c,l,u;e.rangeEquals(r,a,0,r.length)?(o=t.symbol?ns(t.symbol):e.createSymbolTable(t.declaredProperties),s=t.declaredCallSignatures,c=t.declaredConstructSignatures,l=t.declaredStringIndexInfo,u=t.declaredNumberIndexInfo):(i=pd(r,a),o=Ho(t.declaredProperties,i,1===r.length),s=dd(t.declaredCallSignatures,i),c=dd(t.declaredConstructSignatures,i),l=Md(t.declaredStringIndexInfo,i),u=Md(t.declaredNumberIndexInfo,i));var d=Lo(t);if(d.length){t.symbol&&o===ns(t.symbol)&&(o=e.createSymbolTable(t.declaredProperties)),Ua(n,o,s,c,l,u);for(var p=e.lastOrUndefined(a),m=0,f=d;m<f.length;m++){var _=f[m],g=p?rs(kd(_,i),p):_;Ko(o,Gs(g)),s=e.concatenate(s,uc(g,0)),c=e.concatenate(c,uc(g,1)),l||(l=g===Ee?Wc(Ee,!1):mc(g,0)),u=u||mc(g,1)}}Ua(n,o,s,c,l,u)}function is(e,n,t,r,a,i,o,s){var c=new f(se,s);return c.declaration=e,c.typeParameters=n,c.parameters=r,c.thisParameter=t,c.resolvedReturnType=a,c.resolvedTypePredicate=i,c.minArgumentCount=o,c.resolvedMinArgumentCount=void 0,c.target=void 0,c.mapper=void 0,c.unionSignatures=void 0,c}function os(e){var n=is(e.declaration,e.typeParameters,e.thisParameter,e.parameters,void 0,void 0,e.minArgumentCount,19&e.flags);return n.target=e.target,n.mapper=e.mapper,n.unionSignatures=e.unionSignatures,n}function ss(e,n){var t=os(e);return t.unionSignatures=n,t.target=void 0,t.mapper=void 0,t}function cs(n,t){if((12&n.flags)===t)return n;n.optionalCallSignatureCache||(n.optionalCallSignatureCache={});var r=4===t?"inner":"outer";return n.optionalCallSignatureCache[r]||(n.optionalCallSignatureCache[r]=function(n,t){e.Debug.assert(4===t||8===t,"An optional call signature can either be for an inner call chain or an outer call chain, but not both.");var r=os(n);return r.flags|=t,r}(n,t))}function ls(n,t){if(G(n)){var r=n.parameters.length-1,a=lo(n.parameters[r]);if(tm(a))return[i(a,r)];if(!t&&1048576&a.flags&&e.every(a.types,tm))return e.map(a.types,(function(e){return i(e,r)}))}return[n.parameters];function i(t,r){var a=nl(t),i=t.target.labeledElementDeclarations,o=e.map(a,(function(e,a){var o=!!i&&Iv(i[a])||Nv(n,r+a,t),s=t.target.elementFlags[a],c=mr(1,o,12&s?32768:2&s?16384:0);return c.type=4&s?Nl(e):e,c}));return e.concatenate(n.parameters.slice(0,r),o)}}function us(e,n,t,r,a){for(var i=0,o=e;i<o.length;i++){var s=o[i];if(Gp(s,n,t,r,a,t?Kd:jd))return s}}function ds(n,t,r){if(t.typeParameters){if(r>0)return;for(var a=1;a<n.length;a++)if(!us(n[a],t,!1,!1,!1))return;return[t]}var i;for(a=0;a<n.length;a++){var o=a===r?t:us(n[a],t,!0,!1,!0);if(!o)return;i=e.appendIfUnique(i,o)}return i}function ps(n){for(var t,r,a=0;a<n.length;a++){if(0===n[a].length)return e.emptyArray;n[a].length>1&&(r=void 0===r?a:-1);for(var i=0,o=n[a];i<o.length;i++){var s=o[i];if(!t||!us(t,s,!1,!1,!0)){var c=ds(n,s,a);if(c){var l=s;if(c.length>1){var u=s.thisParameter,d=e.forEach(c,(function(e){return e.thisParameter}));if(d)u=Tm(d,tu(e.mapDefined(c,(function(e){return e.thisParameter&&lo(e.thisParameter)}))));(l=ss(s,c)).thisParameter=u}(t||(t=[])).push(l)}}}}if(!e.length(t)&&-1!==r){for(var p=n[void 0!==r?r:0],m=p.slice(),f=function(n){if(n!==p){var t=n[0];if(e.Debug.assert(!!t,"getUnionSignatures bails early on empty signature lists and should not have empty lists on second pass"),!(m=t.typeParameters&&e.some(m,(function(e){return!!e.typeParameters}))?void 0:e.map(m,(function(n){return function(n,t){var r=n.declaration,a=function(e,n){for(var t=wv(e),r=wv(n),a=t>=r?e:n,i=a===e?n:e,o=a===e?t:r,s=Gv(e)||Gv(n),c=s&&!Gv(a),l=new Array(o+(c?1:0)),u=0;u<o;u++){var d=tu([Pv(a,u),Pv(i,u)||De]),p=s&&!c&&u===o-1,m=u>=Fv(a)&&u>=Fv(i),f=u>=t?void 0:Nv(e,u),_=u>=r?void 0:Nv(n,u),g=mr(1|(m&&!p?16777216:0),(f===_?f:f?_?void 0:f:_)||"arg"+u);g.type=p?Nl(d):d,l[u]=g}if(c){var y=mr(1,"args");y.type=Nl(Ov(i,o)),l[o]=y}return l}(n,t),i=function(e,n){if(!e||!n)return e||n;var t=tu([lo(e),lo(n)]);return Tm(e,t)}(n.thisParameter,t.thisParameter),o=Math.max(n.minArgumentCount,t.minArgumentCount),s=is(r,n.typeParameters||t.typeParameters,i,a,void 0,void 0,o,19&(n.flags|t.flags));return s.unionSignatures=e.concatenate(n.unionSignatures||[n],[t]),s}(n,t)}))))return"break"}},_=0,g=n;_<g.length;_++){if("break"===f(g[_]))break}t=m}return t||e.emptyArray}function ms(e,n){for(var t=[],r=!1,a=0,i=e;a<i.length;a++){var o=mc(Zs(i[a]),n);if(!o)return;t.push(o.type),r=r||o.isReadonly}return Wc(Xl(t,2),r)}function fs(e,n){return e?n?tu([e,n]):e:n}function _s(e,n){return e?n?Wc(tu([e.type,n.type]),e.isReadonly&&n.isReadonly):e:n}function gs(e,n){return e&&n&&Wc(Xl([e.type,n.type]),e.isReadonly||n.isReadonly)}function ys(n){var t=e.countWhere(n,(function(e){return uc(e,1).length>0})),r=e.map(n,vo);if(t>0&&t===e.countWhere(r,(function(e){return e}))){var a=r.indexOf(!0);r[a]=!1}return r}function vs(n){for(var t,r,a,i,o=n.types,s=ys(o),c=e.countWhere(s,(function(e){return e})),l=function(l){var u=n.types[l];if(!s[l]){var d=uc(u,1);d.length&&c>0&&(d=e.map(d,(function(e){var n=os(e);return n.resolvedReturnType=function(e,n,t,r){for(var a=[],i=0;i<n.length;i++)i===r?a.push(e):t[i]&&a.push(kc(uc(n[i],1)[0]));return tu(a)}(kc(e),o,s,l),n}))),r=hs(r,d)}t=hs(t,uc(u,0)),a=_s(a,mc(u,0)),i=_s(i,mc(u,1))},u=0;u<o.length;u++)l(u);Ua(n,L,t||e.emptyArray,r||e.emptyArray,a,i)}function hs(n,t){for(var r=function(t){n&&!e.every(n,(function(e){return!Gp(e,t,!1,!1,!1,jd)}))||(n=e.append(n,t))},a=0,i=t;a<i.length;a++){r(i[a])}return n}function bs(n){var t=Sa(n.symbol);if(n.target)Ua(n,L,e.emptyArray,e.emptyArray,void 0,void 0),Ua(n,i=Ho(Ms(n.target),n.mapper,!1),r=dd(uc(n.target,0),n.mapper),a=dd(uc(n.target,1),n.mapper),o=Md(mc(n.target,0),n.mapper),l=Md(mc(n.target,1),n.mapper));else if(2048&t.flags){Ua(n,L,e.emptyArray,e.emptyArray,void 0,void 0);var r=Cc((i=ns(t)).get("__call")),a=Cc(i.get("__new"));Ua(n,i,r,a,o=zc(t,0),l=zc(t,1))}else{var i=L,o=void 0;if(t.exports&&(i=ha(t),t===re)){var s=new e.Map;i.forEach((function(e){418&e.flags||s.set(e.escapedName,e)})),i=s}if(Ua(n,i,e.emptyArray,e.emptyArray,void 0,void 0),32&t.flags){var c=So(Co(t));11272192&c.flags?Ko(i=e.createSymbolTable(Ba(i)),Gs(c)):c===Ee&&(o=Wc(Ee,!1))}var l=384&t.flags&&(32&Fo(t).flags||e.some(n.properties,(function(e){return!!(296&lo(e).flags)})))?pt:void 0;if(Ua(n,i,e.emptyArray,e.emptyArray,o,l),8208&t.flags&&(n.callSignatures=Cc(t)),32&t.flags){var u=Co(t);a=t.members?Cc(t.members.get("__constructor")):e.emptyArray;16&t.flags&&(a=e.addRange(a.slice(),e.mapDefined(n.callSignatures,(function(e){return fv(e.declaration)?is(e.declaration,e.typeParameters,e.thisParameter,e.parameters,u,void 0,e.minArgumentCount,19&e.flags):void 0})))),a.length||(a=function(n){var t=uc(So(n),1);if(0===t.length)return[is(void 0,n.localTypeParameters,void 0,e.emptyArray,n,void 0,0,0)];for(var r=bo(n),a=e.isInJSFile(r),i=gl(r),o=e.length(i),s=[],c=0,l=t;c<l.length;c++){var u=l[c],d=Sc(u.typeParameters),p=e.length(u.typeParameters);if(a||o>=d&&o<=p){var m=p?Gc(u,xc(i,u.typeParameters,d,a)):os(u);m.typeParameters=n.localTypeParameters,m.resolvedReturnType=n,s.push(m)}}return s}(u)),n.constructSignatures=a}}}function Es(n){if(4194304&n.flags){var t=Zs(n.type);return rm(t)?jl(t):uu(t)}if(16777216&n.flags){if(n.root.isDistributive){var r=n.checkType,a=Es(r);if(a!==r)return Rd(n,bd(n.root.checkType,a,n.mapper))}return n}return 1048576&n.flags?Xl(e.sameMap(n.types,Es)):2097152&n.flags?tu(e.sameMap(n.types,Es)):n}function Ts(n){var t,r,a=e.createSymbolTable();Ua(n,L,e.emptyArray,e.emptyArray,void 0,void 0);var i=Ss(n),o=xs(n),s=Ls(n.target||n),c=As(n.target||n),l=Zs(Is(n)),u=Ns(n),d=Q?128:8576;if(Cs(n)){for(var p=0,m=Gs(l);p<m.length;p++){f(su(m[p],d))}(1&l.flags||mc(l,0))&&f(Oe),!Q&&mc(l,1)&&f(Pe)}else zf(Es(o),f);function f(e){zf(s?kd(s,Ed(n.mapper,i,e)):e,(function(o){return function(e,o){if(qo(o)){var s=Zo(o),d=a.get(s);if(d)d.nameType=Xl([d.nameType,o]),d.keyType=Xl([d.keyType,e]);else{var p=qo(e)?cc(l,Zo(e)):void 0,m=!!(4&u||!(8&u)&&p&&16777216&p.flags),f=mr(4|(m?16777216:0),s,262144|(!!(1&u||!(2&u)&&p&&oh(p))?8:0)|(W&&!m&&p&&16777216&p.flags?524288:0));f.mappedType=n,f.nameType=o,f.keyType=e,p&&(f.syntheticOrigin=p,f.declarations=p.declarations),a.set(s,f)}}else if(45&o.flags){var _=kd(c,Ed(n.mapper,i,e));5&o.flags?t=Wc(t?Xl([t.type,_]):_,!!(1&u)):r=Wc(r?Xl([r.type,_]):_,!!(1&u))}}(e,o)}))}Ua(n,a,e.emptyArray,e.emptyArray,t,r)}function Ss(e){return e.typeParameter||(e.typeParameter=wo(xa(e.declaration.typeParameter)))}function xs(e){return e.constraintType||(e.constraintType=Us(Ss(e))||xe)}function Ls(e){return e.declaration.nameType?e.nameType||(e.nameType=kd(sd(e.declaration.nameType),e.mapper)):void 0}function As(e){return e.templateType||(e.templateType=e.declaration.type?kd(Pi(sd(e.declaration.type),!!(4&Ns(e))),e.mapper):xe)}function Ds(n){return e.getEffectiveConstraintOfTypeParameter(n.declaration.typeParameter)}function Cs(e){var n=Ds(e);return 188===n.kind&&138===n.operator}function Is(e){if(!e.modifiersType)if(Cs(e))e.modifiersType=kd(sd(Ds(e).type),e.mapper);else{var n=xs(Ou(e.declaration)),t=n&&262144&n.flags?Us(n):n;e.modifiersType=t&&4194304&t.flags?kd(t.type,e.mapper):De}return e.modifiersType}function Ns(e){var n=e.declaration;return(n.readonlyToken?40===n.readonlyToken.kind?2:1:0)|(n.questionToken?40===n.questionToken.kind?8:4:0)}function Rs(e){var n=Ns(e);return 8&n?-1:4&n?1:0}function ks(e){var n=Rs(e),t=Is(e);return n||(Os(t)?Rs(t):0)}function Os(n){return!!(32&e.getObjectFlags(n))&&Su(xs(n))}function Ps(n){return n.members||(524288&n.flags?4&n.objectFlags?function(n){var t=zo(n.target),r=e.concatenate(t.typeParameters,[t.thisType]),a=nl(n);as(n,t,r,a.length===r.length?a:e.concatenate(a,[n]))}(n):3&n.objectFlags?function(n){as(n,zo(n),e.emptyArray,e.emptyArray)}(n):2048&n.objectFlags?function(n){for(var t=mc(n.source,0),r=Ns(n.mappedType),a=!(1&r),i=4&r?0:16777216,o=t&&Wc(Jm(t.type,n.mappedType,n.constraintType),a&&t.isReadonly),s=e.createSymbolTable(),c=0,l=Gs(n.source);c<l.length;c++){var u=l[c],d=8192|(a&&oh(u)?8:0),p=mr(4|u.flags&i,u.escapedName,d);p.declarations=u.declarations,p.nameType=Tr(u).nameType,p.propertyType=lo(u),p.mappedType=n.mappedType,p.constraintType=n.constraintType,s.set(u.escapedName,p)}Ua(n,s,e.emptyArray,e.emptyArray,o,void 0)}(n):16&n.objectFlags?bs(n):32&n.objectFlags&&Ts(n):1048576&n.flags?function(n){var t=ps(e.map(n.types,(function(e){return e===vn?[lt]:uc(e,0)}))),r=ps(e.map(n.types,(function(e){return uc(e,1)}))),a=ms(n.types,0),i=ms(n.types,1);Ua(n,L,t,r,a,i)}(n):2097152&n.flags&&vs(n)),n}function Ms(n){return 524288&n.flags?Ps(n).properties:e.emptyArray}function ws(e,n){if(524288&e.flags){var t=Ps(e).members.get(n);if(t&&Ra(t))return t}}function Fs(n){if(!n.resolvedProperties){for(var t=e.createSymbolTable(),r=0,a=n.types;r<a.length;r++){for(var i=a[r],o=0,s=Gs(i);o<s.length;o++){var c=s[o];if(!t.has(c.escapedName)){var l=tc(n,c.escapedName);l&&t.set(c.escapedName,l)}}if(1048576&n.flags&&!mc(i,0)&&!mc(i,1))break}n.resolvedProperties=Ba(t)}return n.resolvedProperties}function Gs(e){return 3145728&(e=$s(e)).flags?Fs(e):Ms(e)}function Bs(e){return 262144&e.flags?Us(e):8388608&e.flags?function(e){return qs(e)?function(e){var n=Vs(e.indexType);if(n&&n!==e.indexType){var t=Ru(e.objectType,n,e.noUncheckedIndexedAccessCandidate);if(t)return t}var r=Vs(e.objectType);if(r&&r!==e.objectType)return Ru(r,e.indexType,e.noUncheckedIndexedAccessCandidate);return}(e):void 0}(e):16777216&e.flags?function(e){return qs(e)?Ks(e):void 0}(e):Ws(e)}function Us(e){return qs(e)?Jc(e):void 0}function Vs(e){var n=Lu(e,!1);return n!==e?n:Bs(e)}function js(e){if(!e.resolvedDefaultConstraint){var n=function(e){return e.resolvedInferredTrueType||(e.resolvedInferredTrueType=e.combinedMapper?kd(sd(e.root.node.trueType),e.combinedMapper):wu(e))}(e),t=Fu(e);e.resolvedDefaultConstraint=Li(n)?t:Li(t)?n:Xl([n,t])}return e.resolvedDefaultConstraint}function Hs(e){if(e.root.isDistributive&&e.restrictiveInstantiation!==e){var n=Lu(e.checkType,!1),t=n===e.checkType?Bs(n):n;if(t&&t!==e.checkType){var r=Rd(e,bd(e.root.checkType,t,e.mapper));if(!(131072&r.flags))return r}}}function Ks(e){return Hs(e)||js(e)}function Ws(e){if(464781312&e.flags){var n=Js(e);return n!==cn&&n!==ln?n:void 0}return 4194304&e.flags?Ye:void 0}function zs(e){return Ws(e)||e}function qs(e){return Js(e)!==ln}function Js(n){if(n.resolvedBaseConstraint)return n.resolvedBaseConstraint;var t=!1,r=[];return n.resolvedBaseConstraint=rs(a(n),n);function a(a){if(!a.immediateBaseConstraint){if(!hi(a,4))return ln;if(S>=50)return e.tracing.instant("check","getImmediateBaseConstraint_DepthLimit",{typeId:a.id,originalTypeId:n.id,depth:S}),lr(c,e.Diagnostics.Type_instantiation_is_excessively_deep_and_possibly_infinite),t=!0,a.immediateBaseConstraint=cn;var o=void 0;if(Pp(a,r,r.length)||(r.push(a),S++,o=function(n){if(262144&n.flags){var t=Jc(n);return n.isThisType||!t?t:i(t)}if(3145728&n.flags){for(var r=[],a=0,o=c=n.types;a<o.length;a++){var s=i(o[a]);s&&r.push(s)}return 1048576&n.flags&&r.length===c.length?Xl(r):2097152&n.flags&&r.length?tu(r):void 0}if(4194304&n.flags)return Ye;if(134217728&n.flags){var c=n.types,l=e.mapDefined(c,i);return l.length===c.length?pu(n.texts,l):Oe}if(268435456&n.flags){return(t=i(n.type))?fu(n.symbol,t):Oe}if(8388608&n.flags){var u=i(n.objectType),d=i(n.indexType),p=u&&d&&Ru(u,d,n.noUncheckedIndexedAccessCandidate);return p&&i(p)}if(16777216&n.flags){t=Ks(n);S++;var m=t&&i(t);return S--,m}if(33554432&n.flags)return i(n.substitute);return n}(Lu(a,!1)),S--,r.pop()),!Ti()){if(262144&a.flags){var s=qc(a);if(s){var l=lr(s,e.Diagnostics.Type_parameter_0_has_a_circular_constraint,ii(a));!c||e.isNodeDescendantOf(s,c)||e.isNodeDescendantOf(c,s)||e.addRelatedInfo(l,e.createDiagnosticForNode(c,e.Diagnostics.Circularity_originates_in_type_at_this_location))}}o=ln}t&&(o=ln),a.immediateBaseConstraint=o||cn}return a.immediateBaseConstraint}function i(e){var n=a(e);return n!==cn&&n!==ln?n:void 0}}function Xs(n){if(n.default)n.default===un&&(n.default=ln);else if(n.target){var t=Xs(n.target);n.default=t?kd(t,n.mapper):cn}else{n.default=un;var r=n.symbol&&e.forEach(n.symbol.declarations,(function(n){return e.isTypeParameterDeclaration(n)&&n.default})),a=r?sd(r):cn;n.default===un&&(n.default=a)}return n.default}function Ys(e){var n=Xs(e);return n!==cn&&n!==ln?n:void 0}function Qs(e){return e.resolvedApparentType||(e.resolvedApparentType=function(e){var n=Ad(e);if(n&&!e.declaration.nameType){var t=Us(n);if(t&&(Up(t)||tm(t)))return kd(e,bd(n,t,e.mapper))}return e}(e))}function Zs(n){var t,r=465829888&n.flags?Ws(n)||De:n;return 32&e.getObjectFlags(r)?Qs(r):2097152&r.flags?function(e){return e.resolvedApparentType||(e.resolvedApparentType=rs(e,e,!0))}(r):402653316&r.flags?Sn:296&r.flags?xn:2112&r.flags?(t=j>=7,$n||($n=El("BigInt",0,t))||nn):528&r.flags?Ln:12288&r.flags?Sl(j>=2):67108864&r.flags?nn:4194304&r.flags?Ye:2&r.flags&&!W?nn:r}function $s(e){return rc(Zs(rc(e)))}function ec(n,t){for(var r,a,i,o=1048576&n.flags,s=o?0:16777216,c=4,l=0,u=0,d=n.types;u<d.length;u++){if(!((x=Zs(d[u]))===xe||131072&x.flags)){var p=(S=cc(x,t))?e.getDeclarationModifierFlagsFromSymbol(S):0;if(S){if(o?s|=16777216&S.flags:s&=S.flags,r){if(S!==r){a||(a=new e.Map).set(k(r),r);var m=k(S);a.has(m)||a.set(m,S)}}else r=S;l|=(oh(S)?8:0)|(24&p?0:256)|(16&p?512:0)|(8&p?1024:0)|(32&p?2048:0),zg(S)||(c=2)}else if(o){var f=!Xo(t)&&(yg(t)&&mc(x,1)||mc(x,0));f?(l|=32|(f.isReadonly?8:0),i=e.append(i,tm(x)?im(x)||Ce:f.type)):of(x)?(l|=32,i=e.append(i,Ce)):l|=16}}}if(r&&!(o&&(a||48&l)&&1536&l)){if(!(a||16&l||i))return r;for(var _,g,y,v,h=[],b=!1,E=0,T=a?e.arrayFrom(a.values()):[r];E<T.length;E++){var S=T[E];v?S.valueDeclaration&&S.valueDeclaration!==v&&(b=!0):v=S.valueDeclaration,_=e.addRange(_,S.declarations);var x=lo(S);g?x!==g&&(l|=64):(g=x,y=Tr(S).nameType),Yp(x)&&(l|=128),131072&x.flags&&(l|=131072),h.push(x)}e.addRange(h,i);var L=mr(4|s,t,c|l);return L.containingType=n,!b&&v&&(L.valueDeclaration=v,v.symbol.parent&&(L.parent=v.symbol.parent)),L.declarations=_,L.nameType=y,h.length>2?(L.checkFlags|=65536,L.deferralParent=n,L.deferralConstituents=h):L.type=o?Xl(h):tu(h),L}}function nc(n,t){var r=n.propertyCache||(n.propertyCache=e.createSymbolTable()),a=r.get(t);return a||(a=ec(n,t))&&r.set(t,a),a}function tc(n,t){var r=nc(n,t);return!r||16&e.getCheckFlags(r)?void 0:r}function rc(n){return 1048576&n.flags&&268435456&n.objectFlags?n.resolvedReducedType||(n.resolvedReducedType=function(n){var t=e.sameMap(n.types,rc);if(t===n.types)return n;var r=Xl(t);1048576&r.flags&&(r.resolvedReducedType=r);return r}(n)):2097152&n.flags?(268435456&n.objectFlags||(n.objectFlags|=268435456|(e.some(Fs(n),ac)?536870912:0)),536870912&n.objectFlags?He:n):n}function ac(e){return ic(e)||oc(e)}function ic(n){return!(16777216&n.flags||192!=(131264&e.getCheckFlags(n))||!(131072&lo(n).flags))}function oc(n){return!n.valueDeclaration&&!!(1024&e.getCheckFlags(n))}function sc(n,t){if(536870912&e.getObjectFlags(t)){var r=e.find(Fs(t),ic);if(r)return e.chainDiagnosticMessages(n,e.Diagnostics.The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituents,ii(t,void 0,536870912),ri(r));var a=e.find(Fs(t),oc);if(a)return e.chainDiagnosticMessages(n,e.Diagnostics.The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_private_in_some,ii(t,void 0,536870912),ri(a))}return n}function cc(e,n){if(524288&(e=$s(e)).flags){var t=Ps(e),r=t.members.get(n);if(r&&Ra(r))return r;var a=t===sn?vn:t.callSignatures.length?hn:t.constructSignatures.length?bn:void 0;if(a){var i=ws(a,n);if(i)return i}return ws(yn,n)}if(3145728&e.flags)return tc(e,n)}function lc(n,t){if(3670016&n.flags){var r=Ps(n);return 0===t?r.callSignatures:r.constructSignatures}return e.emptyArray}function uc(e,n){return lc($s(e),n)}function dc(e,n){if(3670016&e.flags){var t=Ps(e);return 0===n?t.stringIndexInfo:t.numberIndexInfo}}function pc(e,n){var t=dc(e,n);return t&&t.type}function mc(e,n){return dc($s(e),n)}function fc(e,n){return pc($s(e),n)}function _c(n,t){if(Em(n)){for(var r=[],a=0,i=Gs(n);a<i.length;a++){var o=i[a];(0===t||yg(o.escapedName))&&r.push(lo(o))}if(0===t&&e.append(r,fc(n,1)),r.length)return Xl(r)}}function gc(n){for(var t,r=0,a=e.getEffectiveTypeParameterDeclarations(n);r<a.length;r++){var i=a[r];t=e.appendIfUnique(t,wo(i.symbol))}return t}function yc(e){var n=[];return e.forEach((function(e,t){Ga(t)||n.push(e)})),n}function vc(n){return e.isInJSFile(n)&&(n.type&&307===n.type.kind||e.getJSDocParameterTags(n).some((function(e){var n=e.isBracketed,t=e.typeExpression;return n||!!t&&307===t.type.kind})))}function hc(n,t){if(!e.isExternalModuleNameRelative(n)){var r=Lr(ne,'"'+n+'"',512);return r&&t?Sa(r):r}}function bc(n){if(e.hasQuestionToken(n)||Ec(n)||vc(n))return!0;if(n.initializer){var t=Lc(n.parent),r=n.parent.parameters.indexOf(n);return e.Debug.assert(r>=0),r>=Fv(t,3)}var a=e.getImmediatelyInvokedFunctionExpression(n.parent);return!!a&&(!n.type&&!n.dotDotDotToken&&n.parent.parameters.indexOf(n)>=a.arguments.length)}function Ec(n){if(!e.isJSDocPropertyLikeTag(n))return!1;var t=n.isBracketed,r=n.typeExpression;return t||!!r&&307===r.type.kind}function Tc(e,n,t,r){return{kind:e,parameterName:n,parameterIndex:t,type:r}}function Sc(n){var t,r=0;if(n)for(var a=0;a<n.length;a++)(t=n[a]).symbol&&e.forEach(t.symbol.declarations,(function(n){return e.isTypeParameterDeclaration(n)&&n.default}))||(r=a+1);return r}function xc(n,t,r,a){var i=e.length(t);if(!i)return[];var o=e.length(n);if(a||o>=r&&o<=i){for(var s=n?n.slice():[],c=o;c<i;c++)s[c]=xe;var l=df(a);for(c=o;c<i;c++){var u=Ys(t[c]);a&&u&&(Vd(u,De)||Vd(u,nn))&&(u=Ee),s[c]=u?kd(u,pd(t,s)):l}return s.length=t.length,s}return n&&n.slice()}function Lc(n){var t,r=Sr(n);if(!r.resolvedSignature){var a=[],i=0,o=0,s=void 0,c=!1,l=e.getImmediatelyInvokedFunctionExpression(n),u=e.isJSDocConstructSignature(n);!l&&e.isInJSFile(n)&&e.isValueSignatureDeclaration(n)&&!e.hasJSDocParameterTags(n)&&!e.getJSDocType(n)&&(i|=16);for(var d=u?1:0;d<n.parameters.length;d++){var p=n.parameters[d],m=p.symbol,f=e.isJSDocParameterTag(p)?p.typeExpression&&p.typeExpression.type:p.type;if(m&&4&m.flags&&!e.isBindingPattern(p.name))m=Cr(p,m.escapedName,111551,void 0,void 0,!1);0===d&&"this"===m.escapedName?(c=!0,s=p.symbol):a.push(m),f&&191===f.kind&&(i|=2),Ec(p)||p.initializer||p.questionToken||p.dotDotDotToken||l&&a.length>l.arguments.length&&!f||vc(p)||(o=a.length)}if(!(167!==n.kind&&168!==n.kind||Qo(n)||c&&s)){var _=167===n.kind?168:167,g=e.getDeclarationOfKind(xa(n),_);g&&(s=(t=LS(g))&&t.symbol)}var y=166===n.kind?Co(Sa(n.parent.symbol)):void 0,v=y?y.localTypeParameters:gc(n);(e.hasRestParameter(n)||e.isInJSFile(n)&&function(n,t){if(e.isJSDocSignature(n)||!Dc(n))return!1;var r=e.lastOrUndefined(n.parameters),a=r?e.getJSDocParameterTags(r):e.getJSDocTags(n).filter(e.isJSDocParameterTag),i=e.firstDefined(a,(function(n){return n.typeExpression&&e.isJSDocVariadicType(n.typeExpression.type)?n.typeExpression.type:void 0})),o=mr(3,"args",32768);o.type=i?Nl(sd(i.type)):Cn,i&&t.pop();return t.push(o),!0}(n,a))&&(i|=1),r.resolvedSignature=is(n,v,s,a,void 0,void 0,o,i)}return r.resolvedSignature}function Ac(n){if(e.isInJSFile(n)&&e.isFunctionLikeDeclaration(n)){var t=e.getJSDocTypeTag(n),r=t&&t.typeExpression&&Py(sd(t.typeExpression));return r&&Bc(r)}}function Dc(n){var t=Sr(n);return void 0===t.containsArgumentsReference&&(8192&t.flags?t.containsArgumentsReference=!0:t.containsArgumentsReference=function n(t){if(!t)return!1;switch(t.kind){case 78:return"arguments"===t.escapedText&&e.isExpressionNode(t);case 163:case 165:case 167:case 168:return 158===t.name.kind&&n(t.name);default:return!e.nodeStartsNewLexicalEnvironment(t)&&!e.isPartOfTypeNode(t)&&!!e.forEachChild(t,n)}}(n.body)),t.containsArgumentsReference}function Cc(n){if(!n)return e.emptyArray;for(var t=[],r=0;r<n.declarations.length;r++){var a=n.declarations[r];if(e.isFunctionLike(a)){if(r>0&&a.body){var i=n.declarations[r-1];if(a.parent===i.parent&&a.kind===i.kind&&a.pos===i.end)continue}t.push(Lc(a))}}return t}function Ic(e){var n=ua(e,e);if(n){var t=fa(n);if(t)return lo(t)}return Ee}function Nc(e){if(e.thisParameter)return lo(e.thisParameter)}function Rc(n){if(!n.resolvedTypePredicate){if(n.target){var t=Rc(n.target);n.resolvedTypePredicate=t?(o=t,s=n.mapper,Tc(o.kind,o.parameterName,o.parameterIndex,kd(o.type,s))):st}else if(n.unionSignatures)n.resolvedTypePredicate=function(e){for(var n,t=[],r=0,a=e;r<a.length;r++){var i=Rc(a[r]);if(i&&2!==i.kind&&3!==i.kind){if(n){if(!Yl(n,i))return}else n=i;t.push(i.type)}}if(!n)return;var o=Xl(t);return Tc(n.kind,n.parameterName,n.parameterIndex,o)}(n.unionSignatures)||st;else{var r=n.declaration&&e.getEffectiveReturnTypeNode(n.declaration),a=void 0;if(!r&&e.isInJSFile(n.declaration)){var i=Ac(n.declaration);i&&n!==i&&(a=Rc(i))}n.resolvedTypePredicate=r&&e.isTypePredicateNode(r)?function(n,t){var r=n.parameterName,a=n.type&&sd(n.type);return 187===r.kind?Tc(n.assertsModifier?2:0,void 0,void 0,a):Tc(n.assertsModifier?3:1,r.escapedText,e.findIndex(t.parameters,(function(e){return e.escapedName===r.escapedText})),a)}(r,n):a||st}e.Debug.assert(!!n.resolvedTypePredicate)}var o,s;return n.resolvedTypePredicate===st?void 0:n.resolvedTypePredicate}function kc(n){if(!n.resolvedReturnType){if(!hi(n,3))return xe;var t=n.target?kd(kc(n.target),n.mapper):n.unionSignatures?Xl(e.map(n.unionSignatures,kc),2):Oc(n.declaration)||(e.nodeIsMissing(n.declaration.body)?Ee:Xv(n.declaration));if(4&n.flags?t=gm(t):8&n.flags&&(t=fm(t)),!Ti()){if(n.declaration){var r=e.getEffectiveReturnTypeNode(n.declaration);if(r)lr(r,e.Diagnostics.Return_type_annotation_circularly_references_itself);else if(X){var a=n.declaration,i=e.getNameOfDeclaration(a);i?lr(i,e.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions,e.declarationNameToString(i)):lr(a,e.Diagnostics.Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions)}}t=Ee}n.resolvedReturnType=t}return n.resolvedReturnType}function Oc(n){if(166===n.kind)return Co(Sa(n.parent.symbol));if(e.isJSDocConstructSignature(n))return sd(n.parameters[0].type);var t,r=e.getEffectiveReturnTypeNode(n);if(r)return sd(r);if(167===n.kind&&!Qo(n)){var a=e.isInJSFile(n)&&ki(n);if(a)return a;var i=no(e.getDeclarationOfKind(xa(n),168));if(i)return i}return(t=Ac(n))&&kc(t)}function Pc(e){return!e.resolvedReturnType&&bi(e,3)>=0}function Mc(e){if(G(e)){var n=lo(e.parameters[e.parameters.length-1]),t=tm(n)?im(n):n;return t&&fc(t,1)}}function wc(e,n,t,r){var a=Fc(e,xc(n,e.typeParameters,Sc(e.typeParameters),t));if(r){var i=My(kc(a));if(i){var o=os(i);o.typeParameters=r;var s=os(a);return s.resolvedReturnType=jc(o),s}}return a}function Fc(n,t){var r=n.instantiations||(n.instantiations=new e.Map),a=Yc(t),i=r.get(a);return i||r.set(a,i=Gc(n,t)),i}function Gc(e,n){return Sd(e,function(e,n){return pd(e.typeParameters,n)}(e,n),!0)}function Bc(e){return e.typeParameters?e.erasedSignatureCache||(e.erasedSignatureCache=function(e){return Sd(e,yd(e.typeParameters),!0)}(e)):e}function Uc(n){return n.typeParameters?n.canonicalSignatureCache||(n.canonicalSignatureCache=function(n){return wc(n,e.map(n.typeParameters,(function(e){return e.target&&!Us(e.target)?e.target:e})),e.isInJSFile(n.declaration))}(n)):n}function Vc(n){var t=n.typeParameters;if(t){var r=yd(t);return Sd(n,pd(t,e.map(t,(function(e){return kd(Ws(e),r)||De}))),!0)}return n}function jc(n){if(!n.isolatedSignatureType){var t=n.declaration?n.declaration.kind:0,r=166===t||170===t||175===t,a=wa(16);a.members=L,a.properties=e.emptyArray,a.callSignatures=r?e.emptyArray:[n],a.constructSignatures=r?[n]:e.emptyArray,n.isolatedSignatureType=a}return n.isolatedSignatureType}function Hc(e){return e.members.get("__index")}function Kc(n,t){var r=1===t?144:147,a=Hc(n);if(a)for(var i=0,o=a.declarations;i<o.length;i++){var s=o[i],c=e.cast(s,e.isIndexSignatureDeclaration);if(1===c.parameters.length){var l=c.parameters[0];if(l.type&&l.type.kind===r)return c}}}function Wc(e,n,t){return{type:e,isReadonly:n,declaration:t}}function zc(n,t){var r=Kc(n,t);if(r)return Wc(r.type?sd(r.type):Ee,e.hasEffectiveModifier(r,64),r)}function qc(n){return e.mapDefined(e.filter(n.symbol&&n.symbol.declarations,e.isTypeParameterDeclaration),e.getEffectiveConstraintOfTypeParameter)[0]}function Jc(n){if(!n.constraint)if(n.target){var t=Us(n.target);n.constraint=t?kd(t,n.mapper):cn}else{var r=qc(n);if(r){var a=sd(r);1&a.flags&&a!==xe&&(a=190===r.parent.parent.kind?Ye:De),n.constraint=a}else n.constraint=function(n){var t;if(n.symbol)for(var r=0,a=n.symbol.declarations;r<a.length;r++){var i=a[r];if(185===i.parent.kind){var o=e.walkUpParenthesizedTypesAndGetParentAndChild(i.parent.parent),s=o[0],c=void 0===s?i.parent:s,l=o[1];if(173===l.kind){var u=l,d=tb(u);if(d){var p=u.typeArguments.indexOf(c);if(p<d.length){var m=Us(d[p]);if(m){var f=kd(m,pd(d,eb(u,d)));f!==n&&(t=e.append(t,f))}}}}else 160===l.kind&&l.dotDotDotToken||181===l.kind||192===l.kind&&l.dotDotDotToken?t=e.append(t,Nl(De)):194===l.kind&&(t=e.append(t,Oe))}}return t&&tu(t)}(n)||cn}return n.constraint===cn?void 0:n.constraint}function Xc(n){var t=e.getDeclarationOfKind(n.symbol,159),r=e.isJSDocTemplateTag(t.parent)?e.getHostSignatureFromJSDoc(t.parent):t.parent;return r&&xa(r)}function Yc(e){var n="";if(e)for(var t=e.length,r=0;r<t;){for(var a=e[r].id,i=1;r+i<t&&e[r+i].id===a+i;)i++;n.length&&(n+=","),n+=a,i>1&&(n+=":"+i),r+=i}return n}function Qc(n,t){for(var r=0,a=0,i=n;a<i.length;a++){var o=i[a];o.flags&t||(r|=e.getObjectFlags(o))}return 3670016&r}function Zc(e,n){var t=Yc(n),r=e.instantiations.get(t);return r||(r=wa(4,e.symbol),e.instantiations.set(t,r),r.objectFlags|=n?Qc(n,0):0,r.target=e,r.resolvedTypeArguments=n),r}function $c(e){var n=Oa(e.flags);return n.symbol=e.symbol,n.objectFlags=e.objectFlags,n.target=e.target,n.resolvedTypeArguments=e.resolvedTypeArguments,n}function el(e,n,t){var r=Hu(n),a=Ku(r),i=wa(4,e.symbol);return i.target=e,i.node=n,i.mapper=t,i.aliasSymbol=r,i.aliasTypeArguments=t?ud(a,t):a,i}function nl(n){var t,r;if(!n.resolvedTypeArguments){if(!hi(n,6))return(null===(t=n.target.localTypeParameters)||void 0===t?void 0:t.map((function(){return xe})))||e.emptyArray;var a=n.node,i=a?173===a.kind?e.concatenate(n.target.outerTypeParameters,eb(a,n.target.localTypeParameters)):178===a.kind?[sd(a.elementType)]:e.map(a.elements,sd):e.emptyArray;Ti()?n.resolvedTypeArguments=n.mapper?ud(i,n.mapper):i:(n.resolvedTypeArguments=(null===(r=n.target.localTypeParameters)||void 0===r?void 0:r.map((function(){return xe})))||e.emptyArray,lr(n.node||c,n.target.symbol?e.Diagnostics.Type_arguments_for_0_circularly_reference_themselves:e.Diagnostics.Tuple_type_arguments_circularly_reference_themselves,n.target.symbol&&ri(n.target.symbol)))}return n.resolvedTypeArguments}function tl(n){return e.length(n.target.typeParameters)}function rl(n,t){var r=Fo(Sa(t)),a=r.localTypeParameters;if(a){var i=e.length(n.typeArguments),o=Sc(a),s=e.isInJSFile(n);if(!(!X&&s)&&(i<o||i>a.length)){var c=s&&e.isExpressionWithTypeArguments(n)&&!e.isJSDocAugmentsTag(n.parent);if(lr(n,o===a.length?c?e.Diagnostics.Expected_0_type_arguments_provide_these_with_an_extends_tag:e.Diagnostics.Generic_type_0_requires_1_type_argument_s:c?e.Diagnostics.Expected_0_1_type_arguments_provide_these_with_an_extends_tag:e.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments,ii(r,void 0,2),o,a.length),!s)return xe}return 173===n.kind&&Pl(n,e.length(n.typeArguments)!==a.length)?el(r,n,void 0):Zc(r,e.concatenate(r.outerTypeParameters,xc(gl(n),a,o,s)))}return ml(n,t)?r:xe}function al(n,t){var r=Fo(n);if(r===Ae&&C.has(n.escapedName)&&t&&1===t.length)return fu(n,t[0]);var a=Tr(n),i=a.typeParameters,o=Yc(t),s=a.instantiations.get(o);return s||a.instantiations.set(o,s=kd(r,pd(i,xc(t,i,Sc(i),e.isInJSFile(n.valueDeclaration))))),s}function il(n){switch(n.kind){case 173:return n.typeName;case 223:var t=n.expression;if(e.isEntityNameExpression(t))return t}}function ol(e,n,t){return e&&ca(e,n,t)||he}function sl(n,t){if(t===he)return xe;if(96&(t=function(n){var t=n.valueDeclaration;if(t&&e.isInJSFile(t)&&!(524288&n.flags)&&!e.getExpandoInitializer(t,!1)){var r=e.isVariableDeclaration(t)?e.getDeclaredExpandoInitializer(t):e.getAssignedExpandoInitializer(t);if(r){var a=xa(r);if(a)return _v(a,n)}}}(t)||t).flags)return rl(n,t);if(524288&t.flags)return function(n,t){var r=Fo(t),a=Tr(t).typeParameters;if(a){var i=e.length(n.typeArguments),o=Sc(a);return i<o||i>a.length?(lr(n,o===a.length?e.Diagnostics.Generic_type_0_requires_1_type_argument_s:e.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments,ri(t),o,a.length),xe):al(t,gl(n))}return ml(n,t)?r:xe}(n,t);var r=Go(t);if(r)return ml(n,t)?ed(r):xe;if(111551&t.flags&&pl(n)){var a=function(e,n){var t=Sr(e);if(!t.resolvedJSDocType){var r=lo(n),a=r;if(n.valueDeclaration){var i=195===e.kind&&e.qualifier;r.symbol&&r.symbol!==n&&i&&(a=sl(e,r.symbol))}t.resolvedJSDocType=a}return t.resolvedJSDocType}(n,t);return a||(ol(il(n),788968),lo(t))}return xe}function cl(e,n){if(3&n.flags||n===e)return e;var t=Hl(e)+">"+Hl(n),r=ge.get(t);if(r)return r;var a=Oa(33554432);return a.baseType=e,a.substitute=n,ge.set(t,a),a}function ll(e){return 179===e.kind&&1===e.elements.length}function ul(e,n,t){return ll(n)&&ll(t)?ul(e,n.elements[0],t.elements[0]):Pu(sd(n))===e?sd(t):void 0}function dl(n,t){for(var r;t&&!e.isStatement(t)&&311!==t.kind;){var a=t.parent;if(184===a.kind&&t===a.trueType){var i=ul(n,a.checkType,a.extendsType);i&&(r=e.append(r,i))}t=a}return r?cl(n,tu(e.append(r,n))):n}function pl(e){return!!(4194304&e.flags)&&(173===e.kind||195===e.kind)}function ml(n,t){return!n.typeArguments||(lr(n,e.Diagnostics.Type_0_is_not_generic,t?ri(t):n.typeName?e.declarationNameToString(n.typeName):o),!1)}function fl(n){if(e.isIdentifier(n.typeName)){var t=n.typeArguments;switch(n.typeName.escapedText){case"String":return ml(n),Oe;case"Number":return ml(n),Pe;case"Boolean":return ml(n),Ue;case"Void":return ml(n),je;case"Undefined":return ml(n),Ce;case"Null":return ml(n),Re;case"Function":case"function":return ml(n),vn;case"array":return t&&t.length||X?void 0:Cn;case"promise":return t&&t.length||X?void 0:zv(Ee);case"Object":if(t&&2===t.length){if(e.isJSDocIndexSignature(n)){var r=sd(t[0]),a=Wc(sd(t[1]),!1);return Va(void 0,L,e.emptyArray,e.emptyArray,r===Oe?a:void 0,r===Pe?a:void 0)}return Ee}return ml(n),X?void 0:Ee}}}function _l(n){var t=Sr(n);if(!t.resolvedType){if(e.isConstTypeReference(n)&&e.isAssertionExpression(n.parent))return t.resolvedSymbol=he,t.resolvedType=Lh(n.parent.expression);var r=void 0,a=void 0,i=788968;pl(n)&&((a=fl(n))||((r=ol(il(n),i,!0))===he?r=ol(il(n),900095):ol(il(n),i),a=sl(n,r))),a||(a=sl(n,r=ol(il(n),i))),t.resolvedSymbol=r,t.resolvedType=a}return t.resolvedType}function gl(n){return e.map(n.typeArguments,sd)}function yl(e){var n=Sr(e);return n.resolvedType||(n.resolvedType=ed(Im(Hh(e.exprName)))),n.resolvedType}function vl(n,t){function r(e){for(var n=0,t=e.declarations;n<t.length;n++){var r=t[n];switch(r.kind){case 252:case 253:case 255:return r}}}if(!n)return t?on:nn;var a=Fo(n);return 524288&a.flags?e.length(a.typeParameters)!==t?(lr(r(n),e.Diagnostics.Global_type_0_must_have_1_type_parameter_s,e.symbolName(n),t),t?on:nn):a:(lr(r(n),e.Diagnostics.Global_type_0_must_be_a_class_or_interface_type,e.symbolName(n)),t?on:nn)}function hl(n,t){return bl(n,111551,t?e.Diagnostics.Cannot_find_global_value_0:void 0)}function bl(e,n,t){return Cr(void 0,e,n,t,e,!1)}function El(n,t,r){var a=function(n,t){return bl(n,788968,t?e.Diagnostics.Cannot_find_global_type_0:void 0)}(n,r);return a||r?vl(a,t):void 0}function Tl(e){return kn||(kn=hl("Symbol",e))}function Sl(e){return On||(On=El("Symbol",0,e))||nn}function xl(e){return Mn||(Mn=El("Promise",1,e))||on}function Ll(e){return Fn||(Fn=hl("Promise",e))}function Al(e){return Bn||(Bn=El("Iterable",1,e))||on}function Dl(e,n){void 0===n&&(n=0);var t=bl(e,788968,void 0);return t&&vl(t,n)}function Cl(e,n){return e!==on?Zc(e,n):nn}function Il(e){return Cl(Pn||(Pn=El("TypedPropertyDescriptor",1,!0))||on,[e])}function Nl(e,n){return Cl(n?Tn:En,[e])}function Rl(e){switch(e.kind){case 180:return 2;case 181:return kl(e);case 192:return e.questionToken?2:e.dotDotDotToken?kl(e):1;default:return 1}}function kl(e){return od(e.type)?4:8}function Ol(n){var t=function(n){return e.isTypeOperatorNode(n)&&142===n.operator}(n.parent);return od(n)?t?Tn:En:Gl(e.map(n.elements,Rl),t,e.some(n.elements,(function(e){return 192!==e.kind}))?void 0:n.elements)}function Pl(n,t){return!!Hu(n)||Ml(n)&&(178===n.kind?wl(n.elementType):179===n.kind?e.some(n.elements,wl):t||e.some(n.typeArguments,wl))}function Ml(e){var n=e.parent;switch(n.kind){case 186:case 192:case 173:case 182:case 183:case 189:case 184:case 188:case 178:case 179:return Ml(n);case 254:return!0}return!1}function wl(n){switch(n.kind){case 173:return pl(n)||!!(524288&ol(n.typeName,788968).flags);case 176:return!0;case 188:return 151!==n.operator&&wl(n.type);case 186:case 180:case 192:case 307:case 305:case 306:case 301:return wl(n.type);case 181:return 178!==n.type.kind||wl(n.type.elementType);case 182:case 183:return e.some(n.types,wl);case 189:return wl(n.objectType)||wl(n.indexType);case 184:return wl(n.checkType)||wl(n.extendsType)||wl(n.trueType)||wl(n.falseType)}return!1}function Fl(n,t,r,a){void 0===r&&(r=!1);var i=Gl(t||e.map(n,(function(e){return 1})),r,a);return i===on?nn:n.length?Bl(i,n):i}function Gl(n,t,r){if(1===n.length&&4&n[0])return t?Tn:En;var a=e.map(n,(function(e){return 1&e?"#":2&e?"?":4&e?".":"*"})).join()+(t?"R":"")+(r&&r.length?","+e.map(r,R).join(","):""),i=le.get(a);return i||le.set(a,i=function(n,t,r){var a,i=n.length,o=e.findLastIndex(n,(function(e){return!!(9&e)}))+1,s=[],c=0;if(i){a=new Array(i);for(var l=0;l<i;l++){var u=a[l]=Fa(),d=n[l];if(!(12&(c|=d))){var p=mr(4|(2&d?16777216:0),""+l,t?8:0);p.tupleLabelDeclaration=null==r?void 0:r[l],p.type=u,s.push(p)}}}var m=s.length,f=mr(4,"length");if(12&c)f.type=Pe;else{var _=[];for(l=o;l<=i;l++)_.push(td(l));f.type=Xl(_)}s.push(f);var g=wa(12);return g.typeParameters=a,g.outerTypeParameters=void 0,g.localTypeParameters=a,g.instantiations=new e.Map,g.instantiations.set(Yc(g.typeParameters),g),g.target=g,g.resolvedTypeArguments=g.typeParameters,g.thisType=Fa(),g.thisType.isThisType=!0,g.thisType.constraint=g,g.declaredProperties=s,g.declaredCallSignatures=e.emptyArray,g.declaredConstructSignatures=e.emptyArray,g.declaredStringIndexInfo=void 0,g.declaredNumberIndexInfo=void 0,g.elementFlags=n,g.minLength=o,g.fixedLength=m,g.hasRestElement=!!(12&c),g.combinedFlags=c,g.readonly=t,g.labeledElementDeclarations=r,g}(n,t,r)),i}function Bl(e,n){return 8&e.objectFlags&&8&e.combinedFlags?Ul(e,n):Zc(e,n)}function Ul(n,t){var r,a,i,o=e.findIndex(t,(function(e,t){return!!(8&n.elementFlags[t]&&1179648&e.flags)}));if(o>=0)return ru(e.map(t,(function(e,t){return 8&n.elementFlags[t]?e:De})))?Yf(t[o],(function(r){return Ul(n,e.replaceElement(t,o,r))})):xe;if(e.findIndex(t,(function(e,t){return!(!(8&n.elementFlags[t])||58982400&e.flags||Os(e))}))<0)return Zc(n,t);for(var s,c=[],l=[],u=[],d=-1,p=function(o){var s=t[o],c=n.elementFlags[o];8&c?58982400&s.flags||Os(s)?_(s,8,null===(r=n.labeledElementDeclarations)||void 0===r?void 0:r[o]):tm(s)?e.forEach(nl(s),(function(e,n){var t;return _(e,s.target.elementFlags[n],null===(t=s.target.labeledElementDeclarations)||void 0===t?void 0:t[n])})):_(Kp(s)&&fc(s,1)||xe,4,null===(a=n.labeledElementDeclarations)||void 0===a?void 0:a[o]):_(s,c,null===(i=n.labeledElementDeclarations)||void 0===i?void 0:i[o])},m=0;m<t.length;m++)p(m);s&&(c[c.length-1]=Xl(s));var f=Gl(l,n.readonly,u);return f===on?nn:l.length?Zc(f,c):f;function _(e,n,t){if(s)s.push(8&n?Iu(e,Pe):e);else{if(1&n&&d>=0){for(var r=d;r<l.length;r++)2&l[r]&&(l[r]=1);d=-1}else 2&n&&d<0?d=l.length:4&n&&(s=[e]);c.push(e),l.push(n),u&&t?u.push(t):u=void 0}}}function Vl(n,t,r){void 0===r&&(r=0);var a=n.target,i=tl(n)-r;return t>a.fixedLength?function(e){var n=im(e);return n&&Nl(n)}(n)||Fl(e.emptyArray):Fl(nl(n).slice(t,i),a.elementFlags.slice(t,i),!1,a.labeledElementDeclarations&&a.labeledElementDeclarations.slice(t,i))}function jl(n){return Xl(e.append(e.arrayOf(n.target.fixedLength,(function(e){return td(""+e)})),uu(n.target.readonly?Tn:En)))}function Hl(e){return e.id}function Kl(n,t){return e.binarySearch(n,t,Hl,e.compareValues)>=0}function Wl(n,t){var r=e.binarySearch(n,t,Hl,e.compareValues);return r<0&&(n.splice(~r,0,t),!0)}function zl(n,t,r){var a=r.flags;if(1048576&a)return ql(n,t,r.types);if(!(131072&a))if(t|=205258751&a,469499904&a&&(t|=262144),r===Se&&(t|=8388608),!W&&98304&a)524288&e.getObjectFlags(r)||(t|=4194304);else{var i=n.length,o=i&&r.id>n[i-1].id?~i:e.binarySearch(n,r,Hl,e.compareValues);o<0&&n.splice(~o,0,r)}return t}function ql(e,n,t){for(var r=0,a=t;r<a.length;r++){n=zl(e,n,a[r])}return n}function Jl(n,t){var r=n.length;if(0===r||function(e){var n=e[0];if(1024&n.flags){for(var t=La(n.symbol),r=1;r<e.length;r++){var a=e[r];if(!(1024&a.flags)||t!==La(a.symbol))return!1}return!0}return!1}(n))return!0;for(var a=r,i=0;a>0;)for(var o=n[--a],s=0,l=n;s<l.length;s++){var u=l[s];if(o!==u){if(1e5===i)if(i/(r-a)*r>(t?25e6:1e6))return e.tracing.instant("check","removeSubtypes_DepthLimit",{typeIds:n.map((function(e){return e.id}))}),lr(c,e.Diagnostics.Expression_produces_a_union_type_that_is_too_complex_to_represent),!1;if(i++,_p(o,u,nr)&&(!(1&e.getObjectFlags(po(o)))||!(1&e.getObjectFlags(po(u)))||qd(o,u))){e.orderedRemoveItemAt(n,a);break}}}return!0}function Xl(n,t,r,a){if(void 0===t&&(t=1),0===n.length)return He;if(1===n.length)return n[0];var i=[],o=ql(i,0,n);if(0!==t){if(3&o)return 1&o?8388608&o?Se:Ee:De;switch(t){case 1:11136&o&&function(n,t){for(var r=n.length;r>0;){var a=n[--r];(128&a.flags&&4&t||256&a.flags&&8&t||2048&a.flags&&64&t||8192&a.flags&&4096&t||nd(a)&&Kl(n,a.regularType))&&e.orderedRemoveItemAt(n,r)}}(i,o),128&o&&134217728&o&&function(n){var t=e.filter(n,Eu);if(t.length)for(var r=n.length,a=function(){r--;var a=n[r];128&a.flags&&e.some(t,(function(e){return Wd(a,e)}))&&e.orderedRemoveItemAt(n,r)};r>0;)a()}(i);break;case 2:if(!Jl(i,!(262144&o)))return xe}if(0===i.length)return 65536&o?4194304&o?Re:ke:32768&o?4194304&o?Ce:Ie:He}return Ql(i,(469647395&o?0:262144)|(2097152&o?268435456:0),r,a)}function Yl(e,n){return e.kind===n.kind&&e.parameterIndex===n.parameterIndex}function Ql(e,n,t,r){if(0===e.length)return He;if(1===e.length)return e[0];var a=Yc(e),i=ue.get(a);return i||(i=Oa(1048576),ue.set(a,i),i.objectFlags=n|Qc(e,98304),i.types=e,i.aliasSymbol=t,i.aliasTypeArguments=r),i}function Zl(e,n,t){var r=t.flags;return 2097152&r?$l(e,n,t.types):(dp(t)?16777216&n||(n|=16777216,e.set(t.id.toString(),t)):(3&r?t===Se&&(n|=8388608):!W&&98304&r||e.has(t.id.toString())||(109440&t.flags&&109440&n&&(n|=67108864),e.set(t.id.toString(),t)),n|=205258751&r),n)}function $l(e,n,t){for(var r=0,a=t;r<a.length;r++){n=Zl(e,n,ed(a[r]))}return n}function eu(e,n){for(var t=0,r=e;t<r.length;t++){var a=r[t];if(!Kl(a.types,n)){var i=128&n.flags?Oe:256&n.flags?Pe:2048&n.flags?Me:8192&n.flags?Ve:void 0;if(!i||!Kl(a.types,i))return!1}}return!0}function nu(n,t){if(e.every(n,(function(n){return!!(1048576&n.flags)&&e.some(n.types,(function(e){return!!(e.flags&t)}))}))){for(var r=0;r<n.length;r++)n[r]=Jf(n[r],(function(e){return!(e.flags&t)}));return!0}return!1}function tu(n,t,r){var a=new e.Map,i=$l(a,0,n),o=e.arrayFrom(a.values());if(131072&i||W&&98304&i&&84410368&i||67108864&i&&402783228&i||402653316&i&&67238776&i||296&i&&469891796&i||2112&i&&469889980&i||12288&i&&469879804&i||49152&i&&469842940&i)return He;if(134217728&i&&128&i&&function(n){for(var t=n.length,r=e.filter(n,(function(e){return!!(128&e.flags)}));t>0;){var a=n[--t];if(134217728&a.flags)for(var i=0,o=r;i<o.length;i++){if(Wd(o[i],a)){e.orderedRemoveItemAt(n,t);break}if(Eu(a))return!0}}return!1}(o))return He;if(1&i)return 8388608&i?Se:Ee;if(!W&&98304&i)return 32768&i?Ce:Re;if((4&i&&128&i||8&i&&256&i||64&i&&2048&i||4096&i&&8192&i)&&function(n,t){for(var r=n.length;r>0;){var a=n[--r];(4&a.flags&&128&t||8&a.flags&&256&t||64&a.flags&&2048&t||4096&a.flags&&8192&t)&&e.orderedRemoveItemAt(n,r)}}(o,i),16777216&i&&524288&i&&e.orderedRemoveItemAt(o,e.findIndex(o,dp)),0===o.length)return De;if(1===o.length)return o[0];var s=Yc(o),c=de.get(s);if(!c){if(1048576&i)if(function(n){var t,r=e.findIndex(n,(function(n){return!!(262144&e.getObjectFlags(n))}));if(r<0)return!1;for(var a=r+1;a<n.length;){var i=n[a];262144&e.getObjectFlags(i)?((t||(t=[n[r]])).push(i),e.orderedRemoveItemAt(n,a)):a++}if(!t)return!1;for(var o=[],s=[],c=0,l=t;c<l.length;c++)for(var u=0,d=l[c].types;u<d.length;u++)Wl(o,i=d[u])&&eu(t,i)&&Wl(s,i);return n[r]=Ql(s,262144),!0}(o))c=tu(o,t,r);else if(nu(o,32768))c=Xl([tu(o),Ce],1,t,r);else if(nu(o,65536))c=Xl([tu(o),Re],1,t,r);else{if(!ru(o))return xe;var l=e.findIndex(o,(function(e){return 0!=(1048576&e.flags)})),u=o[l];c=Xl(e.map(u.types,(function(n){return tu(e.replaceElement(o,l,n))})),1,t,r)}else c=function(e,n,t){var r=Oa(2097152);return r.objectFlags=Qc(e,98304),r.types=e,r.aliasSymbol=n,r.aliasTypeArguments=t,r}(o,t,r);de.set(s,c)}return c}function ru(n){var t=e.reduceLeft(n,(function(e,n){return e*(1048576&n.flags?n.types.length:131072&n.flags?0:1)}),1);return!(t>=1e5)||(e.tracing.instant("check","checkCrossProductUnion_DepthLimit",{typeIds:n.map((function(e){return e.id})),size:t}),lr(c,e.Diagnostics.Expression_produces_a_union_type_that_is_too_complex_to_represent),!1)}function au(e,n){var t=Oa(4194304);return t.type=e,t.stringsOnly=n,t}function iu(n){return!(!n||!(16777216&n.flags&&!n.root.isDistributive||137363456&n.flags&&e.some(n.types,iu)||272629760&n.flags&&iu(n.type)||8388608&n.flags&&iu(n.indexType)||33554432&n.flags&&iu(n.substitute)))}function ou(n){return e.isPrivateIdentifier(n)?He:e.isIdentifier(n)?td(e.unescapeLeadingUnderscores(n.escapedText)):ed(e.isComputedPropertyName(n)?vg(n):Hh(n))}function su(n,t){if(!(24&e.getDeclarationModifierFlagsFromSymbol(n))){var r=Tr(ts(n)).nameType;if(!r&&!e.isKnownSymbol(n))if("default"===n.escapedName)r=td("default");else{var a=n.valueDeclaration&&e.getNameOfDeclaration(n.valueDeclaration);r=a&&ou(a)||td(e.symbolName(n))}if(r&&r.flags&t)return r}return He}function cu(n,t){return Xl(e.map(Gs(n),(function(e){return su(e,t)})))}function lu(e){var n=mc(e,1);return n!==pt?n:void 0}function uu(n,t,r){return void 0===t&&(t=Q),1048576&(n=rc(n)).flags?tu(e.map(n.types,(function(e){return uu(e,t,r)}))):2097152&n.flags?Xl(e.map(n.types,(function(e){return uu(e,t,r)}))):58982400&n.flags||rm(n)||Os(n)&&iu(Ls(n))?function(e,n){return n?e.resolvedStringIndexType||(e.resolvedStringIndexType=au(e,!0)):e.resolvedIndexType||(e.resolvedIndexType=au(e,!1))}(n,t):32&e.getObjectFlags(n)?function(e,n){var t=Jf(xs(e),(function(e){return!(n&&5&e.flags)})),r=e.declaration.nameType&&sd(e.declaration.nameType);return r?Yf(t,(function(n){return kd(r,Ed(e.mapper,Ss(e),n))})):t}(n,r):n===Se?Se:2&n.flags?He:131073&n.flags?Ye:t?!r&&mc(n,0)?Oe:cu(n,128):!r&&mc(n,0)?Xl([Oe,Pe,cu(n,8192)]):lu(n)?Xl([Pe,cu(n,8320)]):cu(n,8576)}function du(n){if(Q)return n;var t=Qn||(Qn=bl("Extract",524288,e.Diagnostics.Cannot_find_global_type_0));return t?al(t,[n,Oe]):Oe}function pu(n,t){var r=e.findIndex(t,(function(e){return!!(1179648&e.flags)}));if(r>=0)return ru(t)?Yf(t[r],(function(a){return pu(n,e.replaceElement(t,r,a))})):xe;if(e.contains(t,Se))return Se;var a=[],i=[],o=n[0];if(!function e(n,t){for(var r=0;r<t.length;r++){var s=t[r];if(101248&s.flags)o+=mu(s)||"",o+=n[r+1];else if(134217728&s.flags){if(o+=s.texts[0],!e(s.texts,s.types))return!1;o+=n[r+1]}else{if(!Su(s)&&!bu(s))return!1;a.push(s),i.push(o),o=n[r+1]}}return!0}(n,t))return Oe;if(0===a.length)return td(o);i.push(o);var s=Yc(a)+"|"+e.map(i,(function(e){return e.length})).join(",")+"|"+i.join(""),c=fe.get(s);return c||fe.set(s,c=function(e,n){var t=Oa(134217728);return t.texts=e,t.types=n,t}(i,a)),c}function mu(n){return 128&n.flags?n.value:256&n.flags?""+n.value:2048&n.flags?e.pseudoBigIntToString(n.value):512&n.flags?n.intrinsicName:65536&n.flags?"null":32768&n.flags?"undefined":void 0}function fu(e,n){return 1179648&n.flags?Yf(n,(function(n){return fu(e,n)})):Su(n)?function(e,n){var t=k(e)+","+Hl(n),r=_e.get(t);r||_e.set(t,r=function(e,n){var t=Oa(268435456);return t.symbol=e,t.type=n,t}(e,n));return r}(e,n):128&n.flags?td(function(e,n){switch(C.get(e.escapedName)){case 0:return n.toUpperCase();case 1:return n.toLowerCase();case 2:return n.charAt(0).toUpperCase()+n.slice(1);case 3:return n.charAt(0).toLowerCase()+n.slice(1)}return n}(e,n.value)):n}function _u(n){return!X&&(!!(16384&e.getObjectFlags(n))||(1048576&n.flags?e.every(n.types,_u):2097152&n.flags?e.some(n.types,_u):!!(465829888&n.flags)&&_u(Js(n))))}function gu(n,t){var r=t&&202===t.kind?t:void 0;return qo(n)?Zo(n):r&&xy(r.argumentExpression,n,!1)?e.getPropertyNameForKnownSymbolName(e.idText(r.argumentExpression.name)):t&&e.isPropertyName(t)?e.getPropertyNameForPropertyNameNode(t):void 0}function yu(n,t){return!(8208&t.flags)||!e.isCallLikeExpression(e.findAncestor(n,(function(n){return!e.isAccessExpression(n)}))||n.parent)&&e.every(t.declarations,(function(n){return!e.isFunctionLike(n)||!!(134217728&e.getCombinedNodeFlags(n))}))}function vu(n,t,r,a,i,o,s,c,l){var u,d=o&&202===o.kind?o:void 0,p=o&&e.isPrivateIdentifier(o)?void 0:gu(r,o);if(void 0!==p){var m=cc(t,p);if(m){if(l&&o&&134217728&Wg(m)&&yu(o,m))dr(!1,null!==(u=null==d?void 0:d.argumentExpression)&&void 0!==u?u:e.isIndexedAccessTypeNode(o)?o.indexType:o,e.Diagnostics._0_is_deprecated,p);if(d){if(hy(m,d,107===d.expression.kind),sh(d,m,e.getAssignmentTargetKind(d)))return void lr(d.argumentExpression,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,ri(m));if(4&s&&(Sr(o).resolvedSymbol=m),cy(d,m))return Te}var f=lo(m);return d&&1!==e.getAssignmentTargetKind(d)?y_(d,f):f}if(qf(t,tm)&&yg(p)&&+p>=0){if(o&&qf(t,(function(e){return!e.target.hasRestElement}))&&!(8&s)){var _=hu(o);tm(t)?lr(_,e.Diagnostics.Tuple_type_0_of_length_1_has_no_element_at_index_2,ii(t),tl(t),e.unescapeLeadingUnderscores(p)):lr(_,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.unescapeLeadingUnderscores(p),ii(t))}return T(mc(t,1)),Yf(t,(function(e){var n=im(e)||Ce;return c?Xl([n,Ce]):n}))}}if(!(98304&r.flags)&&ph(r,402665900)){if(131073&t.flags)return t;var g=mc(t,0),y=ph(r,296)&&mc(t,1)||g;if(y)return 1&s&&y===g?void(d&&lr(d,e.Diagnostics.Type_0_cannot_be_used_to_index_type_1,ii(r),ii(n))):o&&!ph(r,12)?(lr(_=hu(o),e.Diagnostics.Type_0_cannot_be_used_as_an_index_type,ii(r)),c?Xl([y.type,Ce]):y.type):(T(y),c?Xl([y.type,Ce]):y.type);if(131072&r.flags)return He;if(_u(t))return Ee;if(d&&!fh(t)){if(t.symbol===re&&void 0!==p&&re.exports.has(p)&&418&re.exports.get(p).flags)lr(d,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.unescapeLeadingUnderscores(p),ii(t));else if(X&&!V.suppressImplicitAnyIndexErrors&&!i)if(void 0!==p&&py(p,t))lr(d,e.Diagnostics.Property_0_is_a_static_member_of_type_1,p,ii(t));else if(fc(t,1))lr(d.argumentExpression,e.Diagnostics.Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number);else{var v=void 0;if(void 0!==p&&(v=_y(p,t)))void 0!==v&&lr(d.argumentExpression,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2,p,ii(t),v);else{var h=function(n,t,r){function a(e){var t=ws(n,e);if(t){var a=Py(lo(t));return!!a&&Fv(a)>=1&&zd(r,Ov(a,0))}return!1}var i=e.isAssignmentTarget(t)?"set":"get";if(!a(i))return;var o=e.tryGetPropertyAccessOrIdentifierToString(t.expression);void 0===o?o=i:o+="."+i;return o}(t,d,r);if(void 0!==h)lr(d,e.Diagnostics.Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1,ii(t),h);else{var b=void 0;if(1024&r.flags)b=e.chainDiagnosticMessages(void 0,e.Diagnostics.Property_0_does_not_exist_on_type_1,"["+ii(r)+"]",ii(t));else if(8192&r.flags){var E=sa(r.symbol,d);b=e.chainDiagnosticMessages(void 0,e.Diagnostics.Property_0_does_not_exist_on_type_1,"["+E+"]",ii(t))}else 128&r.flags||256&r.flags?b=e.chainDiagnosticMessages(void 0,e.Diagnostics.Property_0_does_not_exist_on_type_1,r.value,ii(t)):12&r.flags&&(b=e.chainDiagnosticMessages(void 0,e.Diagnostics.No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1,ii(r),ii(t)));b=e.chainDiagnosticMessages(b,e.Diagnostics.Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1,ii(a),ii(t)),Yt.add(e.createDiagnosticForNodeFromMessageChain(d,b))}}}return}}if(_u(t))return Ee;if(o){_=hu(o);384&r.flags?lr(_,e.Diagnostics.Property_0_does_not_exist_on_type_1,""+r.value,ii(t)):12&r.flags?lr(_,e.Diagnostics.Type_0_has_no_matching_index_signature_for_type_1,ii(t),ii(r)):lr(_,e.Diagnostics.Type_0_cannot_be_used_as_an_index_type,ii(r))}return Li(r)?r:void 0;function T(n){n&&n.isReadonly&&d&&(e.isAssignmentTarget(d)||e.isDeleteTarget(d))&&lr(d,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,ii(t))}}function hu(e){return 202===e.kind?e.argumentExpression:189===e.kind?e.indexType:158===e.kind?e.expression:e}function bu(e){return-1!==Ze.types.indexOf(e)||!!(1&e.flags)}function Eu(n){return!!(134217728&n.flags)&&e.every(n.types,bu)}function Tu(n){return 3145728&n.flags?(4194304&n.objectFlags||(n.objectFlags|=4194304|(e.some(n.types,Tu)?8388608:0)),!!(8388608&n.objectFlags)):!!(58982400&n.flags)||Os(n)||rm(n)}function Su(n){return 3145728&n.flags?(16777216&n.objectFlags||(n.objectFlags|=16777216|(e.some(n.types,Su)?33554432:0)),!!(33554432&n.objectFlags)):!!(465829888&n.flags)&&!Eu(n)}function xu(e){return!!(262144&e.flags&&e.isThisType)}function Lu(n,t){return 8388608&n.flags?function(n,t){var r=t?"simplifiedForWriting":"simplifiedForReading";if(n[r])return n[r]===ln?n:n[r];n[r]=ln;var a=function(e){if(33554432&e.flags)return e.substitute;return e}(Lu(n.objectType,t)),i=Lu(n.indexType,t),o=function(n,t,r){if(1048576&t.flags){var a=e.map(t.types,(function(e){return Lu(Iu(n,e),r)}));return r?tu(a):Xl(a)}}(a,i,t);if(o)return n[r]=o;if(!(465829888&i.flags)){var s=Au(a,i,t);if(s)return n[r]=s}if(rm(a)&&296&i.flags){var c=sm(a,8&i.flags?0:a.target.fixedLength,0,t);if(c)return n[r]=c}if(Os(a))return n[r]=Yf(Cu(a,n.indexType),(function(e){return Lu(e,t)}));return n[r]=n}(n,t):16777216&n.flags?function(e,n){var t=e.checkType,r=e.extendsType,a=wu(e),i=Fu(e);if(131072&i.flags&&Pu(a)===Pu(t)){if(1&t.flags||zd(Pd(t),Pd(r)))return Lu(a,n);if(Du(t,r))return He}else if(131072&a.flags&&Pu(i)===Pu(t)){if(!(1&t.flags)&&zd(Pd(t),Pd(r)))return He;if(1&t.flags||Du(t,r))return Lu(i,n)}return e}(n,t):n}function Au(n,t,r){if(3145728&n.flags){var a=e.map(n.types,(function(e){return Lu(Iu(e,t),r)}));return 2097152&n.flags||r?tu(a):Xl(a)}}function Du(e,n){return!!(131072&Xl([fs(e,n),He]).flags)}function Cu(e,n){var t=pd([Ss(e)],[n]),r=vd(e.mapper,t);return kd(As(e),r)}function Iu(e,n,t,r,a,i,o){return void 0===o&&(o=0),Ru(e,n,t,r,o,a,i)||(r?xe:De)}function Nu(e,n){return qf(e,(function(e){if(384&e.flags){var t=Zo(e);if(yg(t)){var r=+t;return r>=0&&r<n}}return!1}))}function Ru(e,n,t,r,a,i,o){if(void 0===a&&(a=0),e===Se||n===Se)return Se;var s=t||!!V.noUncheckedIndexedAccess&&16==(18&a);if(!pp(e)||98304&n.flags||!ph(n,12)||(n=Oe),Su(n)||(r&&189!==r.kind?rm(e)&&!Nu(n,e.target.fixedLength):Tu(e)&&(!tm(e)||!Nu(n,e.target.fixedLength)))){if(3&e.flags)return e;var c=e.id+","+n.id+(s?"?":""),l=me.get(c);return l||me.set(c,l=function(e,n,t,r,a){var i=Oa(8388608);return i.objectType=e,i.indexType=n,i.aliasSymbol=t,i.aliasTypeArguments=r,i.noUncheckedIndexedAccessCandidate=a,i}(e,n,i,o,s)),l}var u=$s(e);if(1048576&n.flags&&!(16&n.flags)){for(var d=[],p=!1,m=0,f=n.types;m<f.length;m++){var _=vu(e,u,f[m],n,p,r,a,s);if(_)d.push(_);else{if(!r)return;p=!0}}if(p)return;return 2&a?tu(d,i,o):Xl(d,1,i,o)}return vu(e,u,n,n,!1,r,4|a,s,!0)}function ku(e){var n=Sr(e);if(!n.resolvedType){var t=sd(e.objectType),r=sd(e.indexType),a=Hu(e),i=Iu(t,r,void 0,e,a,Ku(a));n.resolvedType=8388608&i.flags&&i.objectType===t&&i.indexType===r?dl(i,e):i}return n.resolvedType}function Ou(e){var n=Sr(e);if(!n.resolvedType){var t=wa(32,e.symbol);t.declaration=e,t.aliasSymbol=Hu(e),t.aliasTypeArguments=Ku(t.aliasSymbol),n.resolvedType=t,xs(t)}return n.resolvedType}function Pu(e){return 33554432&e.flags?e.baseType:8388608&e.flags&&(33554432&e.objectType.flags||33554432&e.indexType.flags)?Iu(Pu(e.objectType),Pu(e.indexType)):e}function Mu(n,t){for(var r,a,i=function(){var i=kd(n.checkType,t),o=Tu(i)||Su(i),s=kd(n.extendsType,t);if(i===Se||s===Se)return{value:Se};var c=void 0;if(n.inferTypeParameters){var l=wm(n.inferTypeParameters,void 0,0);o&&e.some(n.inferTypeParameters,(function(e){return e===s}))||tf(l.inferences,i,s,768),c=hd(t,l.mapper)}var u=c?kd(n.extendsType,c):s;if(!o&&!Tu(u)&&!Su(u)){if(!(3&u.flags)&&(1&i.flags||!zd(Od(i),Od(u)))){1&i.flags&&(a||(a=[])).push(kd(sd(n.node.trueType),c||t));var d=sd(n.node.falseType);if(16777216&d.flags){var p=d.root;if(p.node.parent===n.node&&(!p.isDistributive||p.checkType===n.checkType))return n=p,"continue"}return r=kd(d,t),"break"}if(3&u.flags||zd(Pd(i),Pd(u)))return r=kd(sd(n.node.trueType),c||t),"break"}var m=Pu(i);return(r=Oa(16777216)).root=n,r.checkType=m,r.extendsType=s,r.mapper=t,r.combinedMapper=c,r.aliasSymbol=n.aliasSymbol,r.aliasTypeArguments=ud(n.aliasTypeArguments,t),"break"};;){var o=i();if("object"==typeof o)return o.value;if("break"===o)break}return a?Xl(e.append(a,r)):r}function wu(e){return e.resolvedTrueType||(e.resolvedTrueType=kd(sd(e.root.node.trueType),e.mapper))}function Fu(e){return e.resolvedFalseType||(e.resolvedFalseType=kd(sd(e.root.node.falseType),e.mapper))}function Gu(n){var t;return n.locals&&n.locals.forEach((function(n){262144&n.flags&&(t=e.append(t,Fo(n)))})),t}function Bu(n){return e.isIdentifier(n)?[n]:e.append(Bu(n.left),n.right)}function Uu(n){var t=Sr(n);if(!t.resolvedType){if(n.isTypeOf&&n.typeArguments)return lr(n,e.Diagnostics.Type_arguments_cannot_be_used_here),t.resolvedSymbol=he,t.resolvedType=xe;if(!e.isLiteralImportTypeNode(n))return lr(n.argument,e.Diagnostics.String_literal_expected),t.resolvedSymbol=he,t.resolvedType=xe;var r=n.isTypeOf?111551:4194304&n.flags?900095:788968,a=ua(n,n.argument.literal);if(!a)return t.resolvedSymbol=he,t.resolvedType=xe;var i=fa(a,!1);if(e.nodeIsMissing(n.qualifier)){if(i.flags&r)t.resolvedType=Vu(n,t,i,r);else lr(n,111551===r?e.Diagnostics.Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here:e.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0,n.argument.literal.text),t.resolvedSymbol=he,t.resolvedType=xe}else{for(var o=Bu(n.qualifier),s=i,c=void 0;c=o.shift();){var l=o.length?1920:r,u=Sa($r(s)),d=n.isTypeOf?cc(lo(u),c.escapedText):Lr(ha(u),c.escapedText,l);if(!d)return lr(c,e.Diagnostics.Namespace_0_has_no_exported_member_1,sa(s),e.declarationNameToString(c)),t.resolvedType=xe;Sr(c).resolvedSymbol=d,Sr(c.parent).resolvedSymbol=d,s=d}t.resolvedType=Vu(n,t,s,r)}}return t.resolvedType}function Vu(e,n,t,r){var a=$r(t);return n.resolvedSymbol=a,111551===r?lo(t):sl(e,a)}function ju(n){var t=Sr(n);if(!t.resolvedType){var r=Hu(n);if(0!==ns(n.symbol).size||r){var a=wa(16,n.symbol);a.aliasSymbol=r,a.aliasTypeArguments=Ku(r),e.isJSDocTypeLiteral(n)&&n.isArrayType&&(a=Nl(a)),t.resolvedType=a}else t.resolvedType=an}return t.resolvedType}function Hu(n){for(var t=n.parent;e.isParenthesizedTypeNode(t)||e.isJSDocTypeExpression(t)||e.isTypeOperatorNode(t)&&142===t.operator;)t=t.parent;return e.isTypeAlias(t)?xa(t):void 0}function Ku(e){return e?yo(e):void 0}function Wu(e){return!!(524288&e.flags)&&!Os(e)}function zu(e){return up(e)||!!(474058748&e.flags)}function qu(n,t){if(2===n.types.length){var r=n.types[0],a=n.types[1];if(e.every(n.types,zu))return up(r)?r:up(a)?a:nn;if(zu(r))return i(a);if(zu(a))return i(r)}function i(n){for(var r=e.createSymbolTable(),a=0,i=Gs(n);a<i.length;a++){var o=i[a];if(24&e.getDeclarationModifierFlagsFromSymbol(o));else if(Xu(o)){var s=65536&o.flags&&!(32768&o.flags),c=mr(16777220,o.escapedName,t?8:0);c.type=s?Ce:lo(o),c.declarations=o.declarations,c.nameType=Tr(o).nameType,c.syntheticOrigin=o,r.set(o.escapedName,c)}}var l=Va(n.symbol,r,e.emptyArray,e.emptyArray,mc(n,0),mc(n,1));return l.objectFlags|=1048704,l}}function Ju(n,t,r,a,i){if(1&n.flags||1&t.flags)return Ee;if(2&n.flags||2&t.flags)return De;if(131072&n.flags)return t;if(131072&t.flags)return n;var o;if(1048576&n.flags)return(o=qu(n,i))?Ju(o,t,r,a,i):ru([n,t])?Yf(n,(function(e){return Ju(e,t,r,a,i)})):xe;if(1048576&t.flags)return(o=qu(t,i))?Ju(n,o,r,a,i):ru([n,t])?Yf(t,(function(e){return Ju(n,e,r,a,i)})):xe;if(473960444&t.flags)return n;if(Tu(n)||Tu(t)){if(up(n))return t;if(2097152&n.flags){var s=n.types,c=s[s.length-1];if(Wu(c)&&Wu(t))return tu(e.concatenate(s.slice(0,s.length-1),[Ju(c,t,r,a,i)]))}return tu([n,t])}var l,u,d=e.createSymbolTable(),p=new e.Set;n===nn?(l=mc(t,0),u=mc(t,1)):(l=gs(mc(n,0),mc(t,0)),u=gs(mc(n,1),mc(t,1)));for(var m=0,f=Gs(t);m<f.length;m++){var _=f[m];24&e.getDeclarationModifierFlagsFromSymbol(_)?p.add(_.escapedName):Xu(_)&&d.set(_.escapedName,Yu(_,i))}for(var g=0,y=Gs(n);g<y.length;g++){var v=y[g];if(!p.has(v.escapedName)&&Xu(v))if(d.has(v.escapedName)){var h=lo(_=d.get(v.escapedName));if(16777216&_.flags){var b=e.concatenate(v.declarations,_.declarations),E=mr(4|16777216&v.flags,v.escapedName);E.type=Xl([lo(v),If(h,524288)]),E.leftSpread=v,E.rightSpread=_,E.declarations=b,E.nameType=Tr(v).nameType,d.set(v.escapedName,E)}}else d.set(v.escapedName,Yu(v,i))}var T=Va(r,d,e.emptyArray,e.emptyArray,Qu(l,i),Qu(u,i));return T.objectFlags|=1049728|a,T}function Xu(n){return!(e.some(n.declarations,e.isPrivateIdentifierPropertyDeclaration)||106496&n.flags&&n.declarations.some((function(n){return e.isClassLike(n.parent)})))}function Yu(e,n){var t=65536&e.flags&&!(32768&e.flags);if(!t&&n===oh(e))return e;var r=mr(4|16777216&e.flags,e.escapedName,n?8:0);return r.type=t?Ce:lo(e),r.declarations=e.declarations,r.nameType=Tr(e).nameType,r.syntheticOrigin=e,r}function Qu(e,n){return e&&e.isReadonly!==n?Wc(e.type,n,e.declaration):e}function Zu(e,n,t){var r=Oa(e);return r.symbol=t,r.value=n,r}function $u(e){if(2944&e.flags){if(!e.freshType){var n=Zu(e.flags,e.value,e.symbol);n.regularType=e,n.freshType=n,e.freshType=n}return e.freshType}return e}function ed(n){return 2944&n.flags?n.regularType:1048576&n.flags?n.regularType||(n.regularType=Xl(e.sameMap(n.types,ed))):n}function nd(e){return!!(2944&e.flags)&&e.freshType===e}function td(n,t,r){var a=(t||"")+("number"==typeof n?"#":"string"==typeof n?"@":"n")+("object"==typeof n?e.pseudoBigIntToString(n):n),i=pe.get(a);if(!i){var o=("number"==typeof n?256:"string"==typeof n?128:2048)|(t?1024:0);pe.set(a,i=Zu(o,n,r)),i.regularType=i}return i}function rd(n){if(e.isValidESSymbolDeclaration(n)){var t=xa(n),r=Tr(t);return r.uniqueESSymbolType||(r.uniqueESSymbolType=function(e){var n=Oa(8192);return n.symbol=e,n.escapedName="__@"+n.symbol.escapedName+"@"+k(n.symbol),n}(t))}return Ve}function ad(n){var t=Sr(n);return t.resolvedType||(t.resolvedType=function(n){var t=e.getThisContainer(n,!1),r=t&&t.parent;if(r&&(e.isClassLike(r)||253===r.kind)&&!e.hasSyntacticModifier(t,32)&&(!e.isConstructorDeclaration(t)||e.isNodeDescendantOf(n,t.body)))return Co(xa(r)).thisType;if(r&&e.isObjectLiteralExpression(r)&&e.isBinaryExpression(r.parent)&&6===e.getAssignmentDeclarationKind(r.parent))return Co(xa(r.parent.left).parent).thisType;var a=4194304&n.flags?e.getHostSignatureFromJSDoc(n):void 0;return a&&e.isFunctionExpression(a)&&e.isBinaryExpression(a.parent)&&3===e.getAssignmentDeclarationKind(a.parent)?Co(xa(a.parent.left).parent).thisType:fv(t)&&e.isNodeDescendantOf(n,t.body)?Co(xa(t)).thisType:(lr(n,e.Diagnostics.A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface),xe)}(n)),t.resolvedType}function id(e){return sd(od(e.type)||e.type)}function od(e){switch(e.kind){case 186:return od(e.type);case 179:if(1===e.elements.length&&(181===(e=e.elements[0]).kind||192===e.kind&&e.dotDotDotToken))return od(e.type);break;case 178:return e.elementType}}function sd(e){return dl(cd(e),e)}function cd(n){switch(n.kind){case 128:case 303:case 304:return Ee;case 152:return De;case 147:return Oe;case 144:return Pe;case 155:return Me;case 131:return Ue;case 148:return Ve;case 113:return je;case 150:return Ce;case 103:return Re;case 141:return He;case 145:return 131072&n.flags&&!X?Ee:Je;case 136:return Ae;case 187:case 107:return ad(n);case 191:return function(e){if(103===e.literal.kind)return Re;var n=Sr(e);return n.resolvedType||(n.resolvedType=ed(Hh(e.literal))),n.resolvedType}(n);case 173:return _l(n);case 172:return n.assertsModifier?je:Ue;case 223:return _l(n);case 176:return yl(n);case 178:case 179:return function(n){var t=Sr(n);if(!t.resolvedType){var r=Ol(n);if(r===on)t.resolvedType=nn;else if(179===n.kind&&e.some(n.elements,(function(e){return!!(8&Rl(e))}))||!Pl(n)){var a=178===n.kind?[sd(n.elementType)]:e.map(n.elements,sd);t.resolvedType=Bl(r,a)}else t.resolvedType=179===n.kind&&0===n.elements.length?r:el(r,n,void 0)}return t.resolvedType}(n);case 180:return function(e){var n=sd(e.type);return W?fm(n):n}(n);case 182:return function(n){var t=Sr(n);if(!t.resolvedType){var r=Hu(n);t.resolvedType=Xl(e.map(n.types,sd),1,r,Ku(r))}return t.resolvedType}(n);case 183:return function(n){var t=Sr(n);if(!t.resolvedType){var r=Hu(n);t.resolvedType=tu(e.map(n.types,sd),r,Ku(r))}return t.resolvedType}(n);case 305:return function(e){var n=sd(e.type);return W?mm(n,65536):n}(n);case 307:return Pi(sd(n.type));case 192:return function(e){var n=Sr(e);return n.resolvedType||(n.resolvedType=e.dotDotDotToken?id(e):e.questionToken&&W?fm(sd(e.type)):sd(e.type))}(n);case 186:case 306:case 301:return sd(n.type);case 181:return id(n);case 309:return function(n){var t=sd(n.type),r=n.parent,a=n.parent.parent;if(e.isJSDocTypeExpression(n.parent)&&e.isJSDocParameterTag(a)){var i=e.getHostSignatureFromJSDoc(a);if(i){var o=e.lastOrUndefined(i.parameters),s=e.getParameterSymbolFromJSDoc(a);if(!o||s&&o.symbol===s&&e.isRestParameter(o))return Nl(t)}}if(e.isParameter(r)&&e.isJSDocFunctionType(r.parent))return Nl(t);return Pi(t)}(n);case 174:case 175:case 177:case 312:case 308:case 313:return ju(n);case 188:return function(n){var t=Sr(n);if(!t.resolvedType)switch(n.operator){case 138:t.resolvedType=uu(sd(n.type));break;case 151:t.resolvedType=148===n.type.kind?rd(e.walkUpParenthesizedTypes(n.parent)):xe;break;case 142:t.resolvedType=sd(n.type);break;default:throw e.Debug.assertNever(n.operator)}return t.resolvedType}(n);case 189:return ku(n);case 190:return Ou(n);case 184:return function(n){var t=Sr(n);if(!t.resolvedType){var r=sd(n.checkType),a=Hu(n),i=Ku(a),o=_o(n,!0),s=i?o:e.filter(o,(function(e){return Ld(e,n)})),c={node:n,checkType:r,extendsType:sd(n.extendsType),isDistributive:!!(262144&r.flags),inferTypeParameters:Gu(n),outerTypeParameters:s,instantiations:void 0,aliasSymbol:a,aliasTypeArguments:i};t.resolvedType=Mu(c,void 0),s&&(c.instantiations=new e.Map,c.instantiations.set(Yc(s),t.resolvedType))}return t.resolvedType}(n);case 185:return function(e){var n=Sr(e);return n.resolvedType||(n.resolvedType=wo(xa(e.typeParameter))),n.resolvedType}(n);case 193:return function(n){var t=Sr(n);return t.resolvedType||(t.resolvedType=pu(__spreadArrays([n.head.text],e.map(n.templateSpans,(function(e){return e.literal.text}))),e.map(n.templateSpans,(function(e){return sd(e.type)})))),t.resolvedType}(n);case 195:return Uu(n);case 78:case 157:case 201:var t=bT(n);return t?Fo(t):xe;default:return xe}}function ld(e,n,t){if(e&&e.length)for(var r=0;r<e.length;r++){var a=e[r],i=t(a,n);if(a!==i){var o=0===r?[]:e.slice(0,r);for(o.push(i),r++;r<e.length;r++)o.push(t(e[r],n));return o}}return e}function ud(e,n){return ld(e,n,kd)}function dd(e,n){return ld(e,n,Sd)}function pd(e,n){return 1===e.length?fd(e[0],n?n[0]:Ee):function(e,n){return{kind:1,sources:e,targets:n}}(e,n)}function md(e,n){switch(n.kind){case 0:return e===n.source?n.target:e;case 1:for(var t=n.sources,r=n.targets,a=0;a<t.length;a++)if(e===t[a])return r?r[a]:Ee;return e;case 2:return n.func(e);case 3:case 4:var i=md(e,n.mapper1);return i!==e&&3===n.kind?kd(i,n.mapper2):md(i,n.mapper2)}}function fd(e,n){return{kind:0,source:e,target:n}}function _d(e){return{kind:2,func:e}}function gd(e,n,t){return{kind:e,mapper1:n,mapper2:t}}function yd(e){return pd(e,void 0)}function vd(e,n){return e?gd(3,e,n):n}function hd(e,n){return e?gd(4,e,n):n}function bd(e,n,t){return t?gd(4,fd(e,n),t):fd(e,n)}function Ed(e,n,t){return e?gd(4,e,fd(n,t)):fd(n,t)}function Td(e){var n=Fa(e.symbol);return n.target=e,n}function Sd(n,t,r){var a;if(n.typeParameters&&!r){a=e.map(n.typeParameters,Td),t=vd(pd(n.typeParameters,a),t);for(var i=0,o=a;i<o.length;i++){o[i].mapper=t}}var s=is(n.declaration,a,n.thisParameter&&xd(n.thisParameter,t),ld(n.parameters,t,xd),void 0,void 0,n.minArgumentCount,19&n.flags);return s.target=n,s.mapper=t,s}function xd(n,t){var r=Tr(n);if(r.type&&!Hm(r.type))return n;1&e.getCheckFlags(n)&&(n=r.target,t=vd(r.mapper,t));var a=mr(n.flags,n.escapedName,1|53256&e.getCheckFlags(n));return a.declarations=n.declarations,a.parent=n.parent,a.target=n,a.mapper=t,n.valueDeclaration&&(a.valueDeclaration=n.valueDeclaration),r.nameType&&(a.nameType=r.nameType),a}function Ld(n,t){if(n.symbol&&n.symbol.declarations&&1===n.symbol.declarations.length){for(var r=n.symbol.declarations[0].parent,a=t;a!==r;a=a.parent)if(!a||230===a.kind||184===a.kind&&e.forEachChild(a.extendsType,i))return!0;return!!e.forEachChild(t,i)}return!0;function i(t){switch(t.kind){case 187:return!!n.isThisType;case 78:return!n.isThisType&&e.isPartOfTypeNode(t)&&function(e){return!(157===e.kind||173===e.parent.kind&&e.parent.typeArguments&&e===e.parent.typeName||195===e.parent.kind&&e.parent.typeArguments&&e===e.parent.qualifier)}(t)&&cd(t)===n;case 176:return!0}return!!e.forEachChild(t,i)}}function Ad(e){var n=xs(e);if(4194304&n.flags){var t=Pu(n.type);if(262144&t.flags)return t}}function Dd(n,t){var r=Ad(n);if(r){var a=kd(r,t);if(r!==a)return Yf(rc(a),(function(a){if(61603843&a.flags&&a!==Se&&a!==xe){if(!n.declaration.nameType){if(Up(a))return function(e,n,t){var r=Id(n,Pe,!0,t);return r===xe?xe:Nl(r,Cd(Vp(e),Ns(n)))}(a,n,bd(r,a,t));if(rm(a))return function(n,t,r,a){var i=n.target.elementFlags,o=e.map(nl(n),(function(e,n){var o=8&i[n]?e:4&i[n]?Nl(e):Fl([e],[i[n]]);return Dd(t,bd(r,o,a))})),s=Cd(n.target.readonly,Ns(t));return Fl(o,e.map(o,(function(e){return 8})),s)}(a,n,r,t);if(tm(a))return function(n,t,r){var a=n.target.elementFlags,i=e.map(nl(n),(function(e,n){return Id(t,td(""+n),!!(2&a[n]),r)})),o=Ns(t),s=4&o?e.map(a,(function(e){return 1&e?2:e})):8&o?e.map(a,(function(e){return 2&e?1:e})):a,c=Cd(n.target.readonly,o);return e.contains(i,xe)?xe:Fl(i,s,c,n.target.labeledElementDeclarations)}(a,n,bd(r,a,t))}return Nd(n,bd(r,a,t))}return a}))}return Nd(n,t)}function Cd(e,n){return!!(1&n)||!(2&n)&&e}function Id(e,n,t,r){var a=Ed(r,Ss(e),n),i=kd(As(e.target||e),a),o=Ns(e);return W&&4&o&&!dh(i,49152)?fm(i):W&&8&o&&t?If(i,524288):i}function Nd(e,n){var t=wa(64|e.objectFlags,e.symbol);if(32&e.objectFlags){t.declaration=e.declaration;var r=Ss(e),a=Td(r);t.typeParameter=a,n=vd(fd(r,a),n),a.mapper=n}return t.target=e,t.mapper=n,t.aliasSymbol=e.aliasSymbol,t.aliasTypeArguments=ud(e.aliasTypeArguments,n),t}function Rd(n,t){var r=n.root;if(r.outerTypeParameters){var a=e.map(r.outerTypeParameters,(function(e){return md(e,t)})),i=Yc(a),o=r.instantiations.get(i);if(!o)o=function(e,n){if(e.isDistributive){var t=e.checkType,r=md(t,n);if(t!==r&&1179648&r.flags)return Yf(r,(function(r){return Mu(e,bd(t,r,n))}))}return Mu(e,n)}(r,pd(r.outerTypeParameters,a)),r.instantiations.set(i,o);return o}return n}function kd(n,t){if(!(n&&t&&Hm(n)))return n;if(50===T||h>=5e6)return e.tracing.instant("check","instantiateType_DepthLimit",{typeId:n.id,instantiationDepth:T,instantiationCount:h}),lr(c,e.Diagnostics.Type_instantiation_is_excessively_deep_and_possibly_infinite),xe;v++,h++,T++;var r=function(n,t){var r=n.flags;if(262144&r)return md(n,t);if(524288&r){var a=n.objectFlags;if(52&a){if(4&a&&!n.node){var i=n.resolvedTypeArguments,o=ud(i,t);return o!==i?Bl(n.target,o):n}return function(n,t){var r=4&n.objectFlags?n.node:n.symbol.declarations[0],a=Sr(r),i=4&n.objectFlags?a.resolvedType:64&n.objectFlags?n.target:n,o=a.outerTypeParameters;if(!o){var s=_o(r,!0);if(fv(r)){var c=gc(r);s=e.addRange(s,c)}o=s||e.emptyArray,o=(4&i.objectFlags||2048&i.symbol.flags)&&!i.aliasTypeArguments?e.filter(o,(function(e){return Ld(e,r)})):o,a.outerTypeParameters=o}if(o.length){var l=vd(n.mapper,t),u=e.map(o,(function(e){return md(e,l)})),d=Yc(u);i.instantiations||(i.instantiations=new e.Map,i.instantiations.set(Yc(o),i));var p=i.instantiations.get(d);if(!p){var m=pd(o,u);p=4&i.objectFlags?el(n.target,n.node,m):32&i.objectFlags?Dd(i,m):Nd(i,m),i.instantiations.set(d,p)}return p}return n}(n,t)}return n}if(3145728&r){var s=n.types,c=ud(s,t);return c===s?n:2097152&r?tu(c,n.aliasSymbol,ud(n.aliasTypeArguments,t)):Xl(c,1,n.aliasSymbol,ud(n.aliasTypeArguments,t))}if(4194304&r)return uu(kd(n.type,t));if(134217728&r)return pu(n.texts,ud(n.types,t));if(268435456&r)return fu(n.symbol,kd(n.type,t));if(8388608&r)return Iu(kd(n.objectType,t),kd(n.indexType,t),n.noUncheckedIndexedAccessCandidate,void 0,n.aliasSymbol,ud(n.aliasTypeArguments,t));if(16777216&r)return Rd(n,vd(n.mapper,t));if(33554432&r){var l=kd(n.baseType,t);if(8650752&l.flags)return cl(l,kd(n.substitute,t));var u=kd(n.substitute,t);return 3&u.flags||zd(Pd(l),Pd(u))?l:u}return n}(n,t);return T--,r}function Od(e){return 262143&e.flags?e:e.permissiveInstantiation||(e.permissiveInstantiation=kd(e,en))}function Pd(e){return 262143&e.flags?e:(e.restrictiveInstantiation||(e.restrictiveInstantiation=kd(e,$e),e.restrictiveInstantiation.restrictiveInstantiation=e.restrictiveInstantiation),e.restrictiveInstantiation)}function Md(e,n){return e&&Wc(kd(e.type,n),e.isReadonly,e.declaration)}function wd(n){switch(e.Debug.assert(165!==n.kind||e.isObjectLiteralMethod(n)),n.kind){case 208:case 209:case 165:case 251:return Fd(n);case 200:return e.some(n.properties,wd);case 199:return e.some(n.elements,wd);case 217:return wd(n.whenTrue)||wd(n.whenFalse);case 216:return(56===n.operatorToken.kind||60===n.operatorToken.kind)&&(wd(n.left)||wd(n.right));case 288:return wd(n.initializer);case 207:return wd(n.expression);case 281:return e.some(n.properties,wd)||e.isJsxOpeningElement(n.parent)&&e.some(n.parent.parent.children,wd);case 280:var t=n.initializer;return!!t&&wd(t);case 283:var r=n.expression;return!!r&&wd(r)}return!1}function Fd(n){return(!e.isFunctionDeclaration(n)||e.isInJSFile(n)&&!!ki(n))&&(Gd(n)||function(n){return!n.typeParameters&&!e.getEffectiveReturnTypeNode(n)&&!!n.body&&230!==n.body.kind&&wd(n.body)}(n))}function Gd(n){if(!n.typeParameters){if(e.some(n.parameters,(function(n){return!e.getEffectiveTypeAnnotationNode(n)})))return!0;if(209!==n.kind){var t=e.firstOrUndefined(n.parameters);if(!t||!e.parameterIsThisKeyword(t))return!0}}return!1}function Bd(n){return(e.isInJSFile(n)&&e.isFunctionDeclaration(n)||lg(n)||e.isObjectLiteralMethod(n))&&Fd(n)}function Ud(n){if(524288&n.flags){var t=Ps(n);if(t.constructSignatures.length||t.callSignatures.length){var r=wa(16,n.symbol);return r.members=t.members,r.properties=t.properties,r.callSignatures=e.emptyArray,r.constructSignatures=e.emptyArray,r}}else if(2097152&n.flags)return tu(e.map(n.types,Ud));return n}function Vd(e,n){return _p(e,n,ar)}function jd(e,n){return _p(e,n,ar)?-1:0}function Hd(e,n){return _p(e,n,tr)?-1:0}function Kd(e,n){return _p(e,n,er)?-1:0}function Wd(e,n){return _p(e,n,er)}function zd(e,n){return _p(e,n,tr)}function qd(n,t){return 1048576&n.flags?e.every(n.types,(function(e){return qd(e,t)})):1048576&t.flags?e.some(t.types,(function(e){return qd(n,e)})):58982400&n.flags?qd(Ws(n)||De,t):t===yn?!!(67633152&n.flags):t===vn?!!(524288&n.flags)&&Df(n):mo(n,po(t))||Up(t)&&!Vp(t)&&qd(n,Tn)}function Jd(e,n){return _p(e,n,rr)}function Xd(e,n){return Jd(e,n)||Jd(n,e)}function Yd(e,n,t,r,a,i){return vp(e,n,tr,t,r,a,i)}function Qd(e,n,t,r,a,i){return Zd(e,n,tr,t,r,a,i,void 0)}function Zd(e,n,t,r,a,i,o,s){return!!_p(e,n,t)||(!r||!ep(a,e,n,t,i,o,s))&&vp(e,n,t,r,i,o,s)}function $d(n){return!!(16777216&n.flags||2097152&n.flags&&e.some(n.types,$d))}function ep(n,t,r,a,i,o,s){if(!n||$d(r))return!1;if(!vp(t,r,a,void 0)&&function(n,t,r,a,i,o,s){for(var c=uc(t,0),l=uc(t,1),u=0,d=[l,c];u<d.length;u++){var p=d[u];if(e.some(p,(function(e){var n=kc(e);return!(131073&n.flags)&&vp(n,r,a,void 0)}))){var m=s||{};Yd(t,r,n,i,o,m);var f=m.errors[m.errors.length-1];return e.addRelatedInfo(f,e.createDiagnosticForNode(n,p===l?e.Diagnostics.Did_you_mean_to_use_new_with_this_expression:e.Diagnostics.Did_you_mean_to_call_this_expression)),!0}}return!1}(n,t,r,a,i,o,s))return!0;switch(n.kind){case 283:case 207:return ep(n.expression,t,r,a,i,o,s);case 216:switch(n.operatorToken.kind){case 62:case 27:return ep(n.right,t,r,a,i,o,s)}break;case 200:return function(n,t,r,a,i,o){return!(131068&r.flags)&&rp(function(n){var t,r,a,i;return __generator(this,(function(o){switch(o.label){case 0:if(!e.length(n.properties))return[2];t=0,r=n.properties,o.label=1;case 1:if(!(t<r.length))return[3,8];if(a=r[t],e.isSpreadAssignment(a))return[3,7];if(!(i=su(xa(a),8576))||131072&i.flags)return[3,7];switch(a.kind){case 168:case 167:case 165:case 289:return[3,2];case 288:return[3,4]}return[3,6];case 2:return[4,{errorNode:a.name,innerExpression:void 0,nameType:i}];case 3:return o.sent(),[3,7];case 4:return[4,{errorNode:a.name,innerExpression:a.initializer,nameType:i,errorMessage:e.isComputedNonLiteralName(a.name)?e.Diagnostics.Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1:void 0}];case 5:return o.sent(),[3,7];case 6:e.Debug.assertNever(a),o.label=7;case 7:return t++,[3,1];case 8:return[2]}}))}(n),t,r,a,i,o)}(n,t,r,a,o,s);case 199:return function(e,n,t,r,a,i){if(131068&t.flags)return!1;if(zp(n))return rp(ip(e,t),n,t,r,a,i);var o=e.contextualType;e.contextualType=t;try{var s=mg(e,1,!0);return e.contextualType=o,!!zp(s)&&rp(ip(e,t),s,t,r,a,i)}finally{e.contextualType=o}}(n,t,r,a,o,s);case 281:return function(n,t,r,a,i,o){var s,c=rp(function(n){var t,r,a;return __generator(this,(function(i){switch(i.label){case 0:if(!e.length(n.properties))return[2];t=0,r=n.properties,i.label=1;case 1:return t<r.length?(a=r[t],e.isJsxSpreadAttribute(a)?[3,3]:[4,{errorNode:a.name,innerExpression:a.initializer,nameType:td(e.idText(a.name))}]):[3,4];case 2:i.sent(),i.label=3;case 3:return t++,[3,1];case 4:return[2]}}))}(n),t,r,a,i,o);if(e.isJsxOpeningElement(n.parent)&&e.isJsxElement(n.parent.parent)){var l=n.parent.parent,u=Pg(kg(n)),d=void 0===u?"children":e.unescapeLeadingUnderscores(u),p=td(d),m=Iu(r,p),f=e.getSemanticJsxChildren(l.children);if(!e.length(f))return c;var _=e.length(f)>1,g=Jf(m,qp),y=Jf(m,(function(e){return!qp(e)}));if(_){if(g!==He){var v=Fl(Ag(l,0));c=rp(function(n,t){var r,a,i,o,s;return __generator(this,(function(c){switch(c.label){case 0:if(!e.length(n.children))return[2];r=0,a=0,c.label=1;case 1:return a<n.children.length?(i=n.children[a],o=td(a-r),(s=ap(i,o,t))?[4,s]:[3,3]):[3,5];case 2:return c.sent(),[3,4];case 3:r++,c.label=4;case 4:return a++,[3,1];case 5:return[2]}}))}(l,E),v,g,a,i,o)||c}else if(!_p(Iu(t,p),m,a)){c=!0;var h=lr(l.openingElement.tagName,e.Diagnostics.This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided,d,ii(m));o&&o.skipLogging&&(o.errors||(o.errors=[])).push(h)}}else if(y!==He){var b=ap(f[0],p,E);b&&(c=rp(function(){return __generator(this,(function(e){switch(e.label){case 0:return[4,b];case 1:return e.sent(),[2]}}))}(),t,r,a,i,o)||c)}else if(!_p(Iu(t,p),m,a)){c=!0;h=lr(l.openingElement.tagName,e.Diagnostics.This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided,d,ii(m));o&&o.skipLogging&&(o.errors||(o.errors=[])).push(h)}}return c;function E(){if(!s){var t=e.getTextOfNode(n.parent.tagName),a=Pg(kg(n)),i=void 0===a?"children":e.unescapeLeadingUnderscores(a),o=Iu(r,td(i)),c=e.Diagnostics._0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2;s=__assign(__assign({},c),{key:"!!ALREADY FORMATTED!!",message:e.formatMessage(void 0,c,t,i,ii(o))})}return s}}(n,t,r,a,o,s);case 209:return function(n,t,r,a,i,o){if(e.isBlock(n.body))return!1;if(e.some(n.parameters,e.hasType))return!1;var s=Py(t);if(!s)return!1;var c=uc(r,0);if(!e.length(c))return!1;var l=n.body,u=kc(s),d=Xl(e.map(c,kc));if(!vp(u,d,a,void 0)){var p=l&&ep(l,u,d,a,void 0,i,o);if(p)return p;var m=o||{};if(vp(u,d,a,l,void 0,i,m),m.errors)return r.symbol&&e.length(r.symbol.declarations)&&e.addRelatedInfo(m.errors[m.errors.length-1],e.createDiagnosticForNode(r.symbol.declarations[0],e.Diagnostics.The_expected_type_comes_from_the_return_type_of_this_signature)),0==(2&e.getFunctionFlags(n))&&!xi(u,"then")&&vp(zv(u),d,a,void 0)&&e.addRelatedInfo(m.errors[m.errors.length-1],e.createDiagnosticForNode(n,e.Diagnostics.Did_you_mean_to_mark_this_function_as_async)),!0}return!1}(n,t,r,a,o,s)}return!1}function np(e,n,t){var r=Ru(n,t);if(r)return r;if(1048576&n.flags){var a=bp(e,n);if(a)return Ru(a,t)}}function tp(e,n){e.contextualType=n;try{return Rh(e,1,n)}finally{e.contextualType=void 0}}function rp(n,t,r,a,i,o){for(var s=!1,c=n.next();!c.done;c=n.next()){var l=c.value,u=l.errorNode,d=l.innerExpression,p=l.nameType,m=l.errorMessage,f=np(t,r,p);if(f&&!(8388608&f.flags)){var _=Ru(t,p);if(_&&!vp(_,f,a,void 0))if(d&&ep(d,_,f,a,void 0,i,o))s=!0;else{var g=o||{},y=d?tp(d,_):_;if(vp(y,f,a,u,m,i,g)&&y!==_&&vp(_,f,a,u,m,i,g),g.errors){var v=g.errors[g.errors.length-1],h=qo(p)?Zo(p):void 0,b=void 0!==h?cc(r,h):void 0,E=!1;if(!b){var T=ph(p,296)&&mc(r,1)||mc(r,0)||void 0;T&&T.declaration&&!e.getSourceFileOfNode(T.declaration).hasNoDefaultLib&&(E=!0,e.addRelatedInfo(v,e.createDiagnosticForNode(T.declaration,e.Diagnostics.The_expected_type_comes_from_this_index_signature)))}if(!E&&(b&&e.length(b.declarations)||r.symbol&&e.length(r.symbol.declarations))){var S=b&&e.length(b.declarations)?b.declarations[0]:r.symbol.declarations[0];e.getSourceFileOfNode(S).hasNoDefaultLib||e.addRelatedInfo(v,e.createDiagnosticForNode(S,e.Diagnostics.The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1,!h||8192&p.flags?ii(p):e.unescapeLeadingUnderscores(h),ii(r)))}}s=!0}}}return s}function ap(n,t,r){switch(n.kind){case 283:return{errorNode:n,innerExpression:n.expression,nameType:t};case 11:if(n.containsOnlyTriviaWhiteSpaces)break;return{errorNode:n,innerExpression:void 0,nameType:t,errorMessage:r()};case 273:case 274:case 277:return{errorNode:n,innerExpression:n,nameType:t};default:return e.Debug.assertNever(n,"Found invalid jsx child")}}function ip(n,t){var r,a,i,o;return __generator(this,(function(s){switch(s.label){case 0:if(!(r=e.length(n.elements)))return[2];a=0,s.label=1;case 1:return a<r?zp(t)&&!cc(t,""+a)?[3,3]:(i=n.elements[a],e.isOmittedExpression(i)?[3,3]:(o=td(a),[4,{errorNode:i,innerExpression:i,nameType:o}])):[3,4];case 2:s.sent(),s.label=3;case 3:return a++,[3,1];case 4:return[2]}}))}function op(e,n,t,r,a){return vp(e,n,rr,t,r,a)}function sp(n,t,r,a,i,o,s,c){if(n===t)return-1;if(!(l=t).typeParameters&&(!l.thisParameter||Li(Cv(l.thisParameter)))&&1===l.parameters.length&&G(l)&&(Cv(l.parameters[0])===Cn||Li(Cv(l.parameters[0])))&&Li(kc(l)))return-1;var l,u=wv(t);if(!Gv(t)&&(8&r?Gv(n)||wv(n)>u:Fv(n)>u))return 0;n.typeParameters&&n.typeParameters!==t.typeParameters&&(n=Fy(n,t=Uc(t),void 0,s));var d=wv(n),p=Uv(n),m=Uv(t);if((p||m)&&kd(p||m,c),p&&m&&d!==u)return 0;var f=t.declaration?t.declaration.kind:0,_=!(3&r)&&z&&165!==f&&164!==f&&166!==f,g=-1,y=Nc(n);if(y&&y!==je){var v=Nc(t);if(v){if(!(x=!_&&s(y,v,!1)||s(v,y,a)))return a&&i(e.Diagnostics.The_this_types_of_each_signature_are_incompatible),0;g&=x}}for(var h=p||m?Math.min(d,u):Math.max(d,u),b=p||m?h-1:-1,E=0;E<h;E++){var T=E===b?Mv(n,E):Pv(n,E),S=E===b?Mv(t,E):Pv(t,E);if(T&&S){var x,L=3&r?void 0:Py(_m(T)),A=3&r?void 0:Py(_m(S));if((x=L&&A&&!Rc(L)&&!Rc(A)&&(98304&um(T))==(98304&um(S))?sp(A,L,8&r|(_?2:1),a,i,o,s,c):!(3&r)&&!_&&s(T,S,!1)||s(S,T,a))&&8&r&&E>=Fv(n)&&E<Fv(t)&&s(T,S,!1)&&(x=0),!x)return a&&i(e.Diagnostics.Types_of_parameters_0_and_1_are_incompatible,e.unescapeLeadingUnderscores(Nv(n,E)),e.unescapeLeadingUnderscores(Nv(t,E))),0;g&=x}}if(!(4&r)){var D=Pc(t)?Ee:t.declaration&&fv(t.declaration)?Co(Sa(t.declaration.symbol)):kc(t);if(D===je)return g;var C=Pc(n)?Ee:n.declaration&&fv(n.declaration)?Co(Sa(n.declaration.symbol)):kc(n),I=Rc(t);if(I){var N=Rc(n);if(N)g&=function(n,t,r,a,i){if(n.kind!==t.kind)return r&&(a(e.Diagnostics.A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard),a(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,di(n),di(t))),0;if((1===n.kind||3===n.kind)&&n.parameterIndex!==t.parameterIndex)return r&&(a(e.Diagnostics.Parameter_0_is_not_in_the_same_position_as_parameter_1,n.parameterName,t.parameterName),a(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,di(n),di(t))),0;var o=n.type===t.type?-1:n.type&&t.type?i(n.type,t.type,r):0;0===o&&r&&a(e.Diagnostics.Type_predicate_0_is_not_assignable_to_1,di(n),di(t));return o}(N,I,a,i,s);else if(e.isIdentifierTypePredicate(I))return a&&i(e.Diagnostics.Signature_0_must_be_a_type_predicate,ai(n)),0}else!(g&=1&r&&s(D,C,!1)||s(C,D,a))&&a&&o&&o(C,D)}return g}function cp(e,n){var t=Bc(e),r=Bc(n),a=kc(t),i=kc(r);return!(i!==je&&!_p(i,a,tr)&&!_p(a,i,tr))&&0!==sp(t,r,!0?4:0,!1,void 0,void 0,Hd,void 0)}function lp(e){return e!==sn&&0===e.properties.length&&0===e.callSignatures.length&&0===e.constructSignatures.length&&!e.stringIndexInfo&&!e.numberIndexInfo}function up(n){return 524288&n.flags?!Os(n)&&lp(Ps(n)):!!(67108864&n.flags)||(1048576&n.flags?e.some(n.types,up):!!(2097152&n.flags)&&e.every(n.types,up))}function dp(n){return!!(16&e.getObjectFlags(n)&&(n.members&&lp(n)||n.symbol&&2048&n.symbol.flags&&0===ns(n.symbol).size))}function pp(n){return 524288&n.flags&&!Os(n)&&0===Gs(n).length&&mc(n,0)&&!mc(n,1)||3145728&n.flags&&e.every(n.types,pp)||!1}function mp(n,t,r){if(n===t)return!0;var a=k(n)+","+k(t),i=ir.get(a);if(void 0!==i&&(4&i||!(2&i)||!r))return!!(1&i);if(!(n.escapedName===t.escapedName&&256&n.flags&&256&t.flags))return ir.set(a,6),!1;for(var o=lo(t),s=0,c=Gs(lo(n));s<c.length;s++){var l=c[s];if(8&l.flags){var u=cc(o,l.escapedName);if(!(u&&8&u.flags))return r?(r(e.Diagnostics.Property_0_is_missing_in_type_1,e.symbolName(l),ii(Fo(t),void 0,64)),ir.set(a,6)):ir.set(a,2),!1}}return ir.set(a,1),!0}function fp(e,n,t,r){var a=e.flags,i=n.flags;if(3&i||131072&a||e===Se)return!0;if(131072&i)return!1;if(402653316&a&&4&i)return!0;if(128&a&&1024&a&&128&i&&!(1024&i)&&e.value===n.value)return!0;if(296&a&&8&i)return!0;if(256&a&&1024&a&&256&i&&!(1024&i)&&e.value===n.value)return!0;if(2112&a&&64&i)return!0;if(528&a&&16&i)return!0;if(12288&a&&4096&i)return!0;if(32&a&&32&i&&mp(e.symbol,n.symbol,r))return!0;if(1024&a&&1024&i){if(1048576&a&&1048576&i&&mp(e.symbol,n.symbol,r))return!0;if(2944&a&&2944&i&&e.value===n.value&&mp(La(e.symbol),La(n.symbol),r))return!0}if(32768&a&&(!W||49152&i))return!0;if(65536&a&&(!W||65536&i))return!0;if(524288&a&&67108864&i)return!0;if(t===tr||t===rr){if(1&a)return!0;if(264&a&&!(1024&a)&&(32&i||256&i&&1024&i))return!0}return!1}function _p(e,n,t){if(nd(e)&&(e=e.regularType),nd(n)&&(n=n.regularType),e===n)return!0;if(t!==ar){if(t===rr&&!(131072&n.flags)&&fp(n,e,t)||fp(e,n,t))return!0}else if(!(3145728&e.flags||3145728&n.flags||e.flags===n.flags||469237760&e.flags))return!1;if(524288&e.flags&&524288&n.flags){var r=t.get(Np(e,n,0,t));if(void 0!==r)return!!(1&r)}return!!(469499904&e.flags||469499904&n.flags)&&vp(e,n,t,void 0)}function gp(n,t){return 4096&e.getObjectFlags(n)&&!Sg(t.escapedName)}function yp(n,t){for(;;){var r=nd(n)?n.regularType:4&e.getObjectFlags(n)&&n.node?Zc(n.target,nl(n)):3145728&n.flags?rc(n):33554432&n.flags?t?n.baseType:n.substitute:25165824&n.flags?Lu(n,t):n;if(r===n)break;n=r}return n}function vp(n,t,r,a,i,s,l){var u,d,p,m,f,_,g=0,y=0,v=0,h=!1,b=0,E=[],T=!1;e.Debug.assert(r!==ar||!a,"no error reporting in identity checking");var S=F(n,t,!!a,i);if(E.length&&O(),h){e.tracing.instant("check","checkTypeRelatedTo_DepthLimit",{sourceId:n.id,targetId:t.id,depth:y});var x=lr(a||c,e.Diagnostics.Excessive_stack_depth_comparing_types_0_and_1,ii(n),ii(t));l&&(l.errors||(l.errors=[])).push(x)}else if(u){if(s){var L=s();L&&(e.concatenateDiagnosticMessageChains(L,u),u=L)}var A=void 0;if(i&&a&&!S&&n.symbol){var C=Tr(n.symbol);if(C.originatingImport&&!e.isImportCall(C.originatingImport))if(vp(lo(C.target),t,r,void 0)){var I=e.createDiagnosticForNode(C.originatingImport,e.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead);A=e.append(A,I)}}x=e.createDiagnosticForNodeFromMessageChain(a,u,A);d&&e.addRelatedInfo.apply(void 0,__spreadArrays([x],d)),l&&(l.errors||(l.errors=[])).push(x),l&&l.skipLogging||Yt.add(x)}return a&&l&&l.skipLogging&&0===S&&e.Debug.assert(!!l.errors,"missed opportunity to interact with error."),0!==S;function N(e){u=e.errorInfo,_=e.lastSkippedInfo,E=e.incompatibleStack,b=e.overrideNextErrorInfo,d=e.relatedInfo}function R(){return{errorInfo:u,lastSkippedInfo:_,incompatibleStack:E.slice(),overrideNextErrorInfo:b,relatedInfo:d?d.slice():void 0}}function k(e,n,t,r,a){b++,_=void 0,E.push([e,n,t,r,a])}function O(){var n=E;E=[];var t=_;if(_=void 0,1===n.length)return P.apply(void 0,n[0]),void(t&&M.apply(void 0,__spreadArrays([void 0],t)));for(var r="",a=[];n.length;){var i=n.pop(),o=i[0],s=i.slice(1);switch(o.code){case e.Diagnostics.Types_of_property_0_are_incompatible.code:0===r.indexOf("new ")&&(r="("+r+")");var c=""+s[0];r=0===r.length?""+c:e.isIdentifierText(c,V.target)?r+"."+c:"["===c[0]&&"]"===c[c.length-1]?""+r+c:r+"["+c+"]";break;case e.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible.code:case e.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible.code:case e.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code:case e.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code:if(0===r.length){var l=o;o.code===e.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code?l=e.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible:o.code===e.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code&&(l=e.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible),a.unshift([l,s[0],s[1]])}else{r=""+(o.code===e.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible.code||o.code===e.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code?"new ":"")+r+"("+(o.code===e.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code||o.code===e.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code?"":"...")+")"}break;default:return e.Debug.fail("Unhandled Diagnostic: "+o.code)}}r?P(")"===r[r.length-1]?e.Diagnostics.The_types_returned_by_0_are_incompatible_between_these_types:e.Diagnostics.The_types_of_0_are_incompatible_between_these_types,r):a.shift();for(var u=0,d=a;u<d.length;u++){var p=d[u],m=(o=p[0],s=p.slice(1),o.elidedInCompatabilityPyramid);o.elidedInCompatabilityPyramid=!1,P.apply(void 0,__spreadArrays([o],s)),o.elidedInCompatabilityPyramid=m}t&&M.apply(void 0,__spreadArrays([void 0],t))}function P(n,t,r,i,o){e.Debug.assert(!!a),E.length&&O(),n.elidedInCompatabilityPyramid||(u=e.chainDiagnosticMessages(u,n,t,r,i,o))}function M(n,t,a){E.length&&O();var i=oi(t,a),o=i[0],s=i[1],c=t,l=o;if(Yp(t)&&!hp(a)&&(c=Qp(t),e.Debug.assert(!zd(c,a),"generalized source shouldn't be assignable"),l=si(c)),262144&a.flags){var u=Ws(a),d=void 0;u&&(zd(c,u)||(d=zd(t,u)))?P(e.Diagnostics._0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_constraint_2,d?o:l,s,ii(u)):P(e.Diagnostics._0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1,s,l)}n||(n=r===rr?e.Diagnostics.Type_0_is_not_comparable_to_type_1:o===s?e.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated:e.Diagnostics.Type_0_is_not_assignable_to_type_1),P(n,l,s)}function w(n,t,r){return tm(n)?n.target.readonly&&jp(t)?(r&&P(e.Diagnostics.The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1,ii(n),ii(t)),!1):tm(t)||Up(t):Vp(n)&&jp(t)?(r&&P(e.Diagnostics.The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1,ii(n),ii(t)),!1):!tm(t)||Up(n)}function F(n,t,i,o,s){if(void 0===i&&(i=!1),void 0===s&&(s=0),524288&n.flags&&131068&t.flags)return fp(n,t,r,i?P:void 0)?-1:(S(n,t,0,!!(4096&e.getObjectFlags(n))),0);var c=yp(n,!1),l=yp(t,!0);if(c===l)return-1;if(r===ar)return function(e,n){var t=e.flags&n.flags;if(!(469237760&t))return 0;if(3145728&t){var r=G(e,n);return r&&(r&=G(n,e)),r}return j(e,n,!1,0)}(c,l);if(262144&c.flags&&Bs(c)===l)return-1;if(1048576&l.flags&&524288&c.flags&&l.types.length<=3&&dh(l,98304)){var d=Qf(l,-98305);if(!(1179648&d.flags)){if(c===d)return-1;l=d}}if(r===rr&&!(131072&l.flags)&&fp(l,c,r)||fp(c,l,r,i?P:void 0))return-1;var p=!!(4096&e.getObjectFlags(c)),m=!(2&s)&&of(c)&&32768&e.getObjectFlags(c);if(m&&function(n,t,i){if(!Hg(t)||!X&&16384&e.getObjectFlags(t))return!1;var o=!!(4096&e.getObjectFlags(n));if((r===tr||r===rr)&&(Wf(yn,t)||!o&&up(t)))return!1;var s,c=t;1048576&t.flags&&(c=VS(n,t,F)||function(e){if(dh(e,67108864)){var n=Jf(e,(function(e){return!(131068&e.flags)}));if(!(131072&n.flags))return n}return e}(t),s=1048576&c.flags?c.types:[c]);for(var l=function(t){if(function(e,n){return e.valueDeclaration&&n.valueDeclaration&&e.valueDeclaration.parent===n.valueDeclaration}(t,n.symbol)&&!gp(n,t)){if(!jg(c,t.escapedName,o)){if(i){var r=Jf(c,Hg);if(!a)return{value:e.Debug.fail()};if(e.isJsxAttributes(a)||e.isJsxOpeningLikeElement(a)||e.isJsxOpeningLikeElement(a.parent)){t.valueDeclaration&&e.isJsxAttribute(t.valueDeclaration)&&e.getSourceFileOfNode(a)===e.getSourceFileOfNode(t.valueDeclaration.name)&&(a=t.valueDeclaration.name);var l=ri(t),u=fy(l,r);(p=u?ri(u):void 0)?P(e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2,l,ii(r),p):P(e.Diagnostics.Property_0_does_not_exist_on_type_1,l,ii(r))}else{var d=n.symbol&&e.firstOrUndefined(n.symbol.declarations),p=void 0;if(t.valueDeclaration&&e.findAncestor(t.valueDeclaration,(function(e){return e===d}))&&e.getSourceFileOfNode(d)===e.getSourceFileOfNode(a)){var m=t.valueDeclaration;e.Debug.assertNode(m,e.isObjectLiteralElementLike),a=m;var f=m.name;e.isIdentifier(f)&&(p=_y(f,r))}void 0!==p?P(e.Diagnostics.Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2,ri(t),ii(r),p):P(e.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1,ri(t),ii(r))}}return{value:!0}}if(s&&!F(lo(t),function(n,t){var r=function(n,r){var a=3145728&(r=Zs(r)).flags?tc(r,t):ws(r,t),i=a&&lo(a)||yg(t)&&fc(r,1)||fc(r,0)||Ce;return e.append(n,i)};return Xl(e.reduceLeft(n,r,void 0)||e.emptyArray)}(s,t.escapedName),i))return i&&k(e.Diagnostics.Types_of_property_0_are_incompatible,ri(t)),{value:!0}}},u=0,d=Gs(n);u<d.length;u++){var p=l(d[u]);if("object"==typeof p)return p.value}return!1}(c,l,i))return i&&M(o,c,l),0;var f=r!==rr&&!(2&s)&&2752508&c.flags&&c!==yn&&2621440&l.flags&&Tp(l)&&(Gs(c).length>0||DT(c));if(f&&!function(e,n,t){for(var r=0,a=Gs(e);r<a.length;r++){if(jg(n,a[r].escapedName,t))return!0}return!1}(c,l,p)){if(i){var g=uc(c,0),y=uc(c,1);g.length>0&&F(kc(g[0]),l,!1)||y.length>0&&F(kc(y[0]),l,!1)?P(e.Diagnostics.Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it,ii(c),ii(l)):P(e.Diagnostics.Type_0_has_no_properties_in_common_with_type_1,ii(c),ii(l))}return 0}var v=0,h=R();if(1048576&c.flags?v=r===rr?U(c,l,i&&!(131068&c.flags),s):function(e,n,t,r){for(var a=-1,i=e.types,o=0;o<i.length;o++){var s=i[o];if(1048576&n.flags&&n.types.length===i.length){var c=F(s,n.types[o],!1,void 0,r);if(c){a&=c;continue}}var l=F(s,n,t,void 0,r);if(!l)return 0;a&=l}return a}(c,l,i&&!(131068&c.flags),s):(1048576&l.flags?v=B(Sm(c),l,i&&!(131068&c.flags)&&!(131068&l.flags)):2097152&l.flags?v=function(e,n,t,r){for(var a=-1,i=n.types,o=0,s=i;o<s.length;o++){var c=F(e,s[o],t,void 0,r);if(!c)return 0;a&=c}return a}(Sm(c),l,i,2):2097152&c.flags&&(v=U(c,l,!1,1)),!v&&(469499904&c.flags||469499904&l.flags)&&(v=j(c,l,i,s))&&N(h)),!v&&2359296&c.flags){var E=function(n,t){for(var r,a=!1,i=0,o=n;i<o.length;i++)if(465829888&(u=o[i]).flags){for(var s=Bs(u);s&&21233664&s.flags;)s=Bs(s);s&&(r=e.append(r,s),t&&(r=e.append(r,u)))}else 469892092&u.flags&&(a=!0);if(r&&(t||a)){if(a)for(var c=0,l=n;c<l.length;c++){var u;469892092&(u=l[c]).flags&&(r=e.append(r,u))}return tu(r)}}(2097152&c.flags?c.types:[c],!!(1048576&l.flags));E&&(2097152&c.flags||1048576&l.flags)&&qf(E,(function(e){return e!==c}))&&(v=F(E,l,!1,void 0,s))&&N(h)}return v&&!T&&(2097152&l.flags&&(m||f)||Wu(l)&&!Up(l)&&!tm(l)&&2097152&c.flags&&3670016&Zs(c).flags&&!e.some(c.types,(function(n){return!!(2097152&e.getObjectFlags(n))})))&&(T=!0,v&=j(c,l,i,4),T=!1),S(c,l,v,p),v;function S(r,s,c,l){if(!c&&i){r=n.aliasSymbol?n:r,s=t.aliasSymbol?t:s;var d=b>0;if(d&&b--,524288&r.flags&&524288&s.flags){var p=u;w(r,s,i),u!==p&&(d=!!u)}if(524288&r.flags&&131068&s.flags)!function(n,t){var r=ci(n.symbol)?ii(n,n.symbol.valueDeclaration):ii(n),a=ci(t.symbol)?ii(t,t.symbol.valueDeclaration):ii(t);(Sn===n&&Oe===t||xn===n&&Pe===t||Ln===n&&Ue===t||Sl(!1)===n&&Ve===t)&&P(e.Diagnostics._0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible,a,r)}(r,s);else if(r.symbol&&524288&r.flags&&yn===r)P(e.Diagnostics.The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead);else if(l&&2097152&s.flags){var m=s.types,f=Ig(D.IntrinsicAttributes,a),g=Ig(D.IntrinsicClassAttributes,a);if(f!==xe&&g!==xe&&(e.contains(m,f)||e.contains(m,g)))return c}else u=sc(u,t);if(!o&&d)return _=[r,s],c;M(o,r,s)}}}function G(e,n){for(var t=-1,r=0,a=e.types;r<a.length;r++){var i=B(a[r],n,!1);if(!i)return 0;t&=i}return t}function B(e,n,t){var r=n.types;if(1048576&n.flags&&Kl(r,e))return-1;for(var a=0,i=r;a<i.length;a++){var o=F(e,i[a],!1);if(o)return o}t&&F(e,bp(e,n,F)||r[r.length-1],!0);return 0}function U(e,n,t,r){var a=e.types;if(1048576&e.flags&&Kl(a,n))return-1;for(var i=a.length,o=0;o<i;o++){var s=F(a[o],n,t&&o===i-1,void 0,r);if(s)return s}return 0}function j(n,t,a,i){if(h)return 0;var o=Np(n,t,i|(T?8:0),r),s=r.get(o);if(void 0!==s&&(!(a&&2&s)||4&s)){if(it){var c=24&s;8&c&&kd(n,_d(H)),16&c&&kd(n,_d(K))}return 1&s?-1:0}if(p){for(var l=0;l<g;l++)if(o===p[l])return 3;if(100===y)return h=!0,0}else p=[],m=[],f=[];var d=g;p[g]=o,g++,m[y]=n,f[y]=t,y++;var _,b=v;1&v||!Pp(n,m,y)||(v|=1),2&v||!Pp(t,f,y)||(v|=2);var E=0;it&&(_=it,it=function(e){return E|=e?16:8,_(e)}),3===v&&e.tracing.instant("check","recursiveTypeRelatedTo_DepthLimit",{sourceId:n.id,sourceIdStack:m.map((function(e){return e.id})),targetId:t.id,targetIdStack:f.map((function(e){return e.id})),depth:y});var S=3!==v?function(n,t,a,i){e.tracing.push("check","structuredTypeRelatedTo",{sourceId:n.id,targetId:t.id});var o=function(n,t,a,i){if(4&i)return Y(n,t,a,void 0,0);var o,s,c=n.flags&t.flags;if(r===ar&&!(524288&c)){if(4194304&c)return F(n.type,t.type,!1);var l=0;return 8388608&c&&(l=F(n.objectType,t.objectType,!1))&&(l&=F(n.indexType,t.indexType,!1))||16777216&c&&n.root.isDistributive===t.root.isDistributive&&(l=F(n.checkType,t.checkType,!1))&&(l&=F(n.extendsType,t.extendsType,!1))&&(l&=F(wu(n),wu(t),!1))&&(l&=F(Fu(n),Fu(t),!1))?l:33554432&c?F(n.substitute,t.substitute,!1):0}var d=!1,p=R();if(17301504&n.flags&&n.aliasSymbol&&n.aliasTypeArguments&&n.aliasSymbol===t.aliasSymbol&&!n.aliasTypeArgumentsContainsMarker&&!t.aliasTypeArgumentsContainsMarker){if((V=xp(n.aliasSymbol))===e.emptyArray)return 1;if(void 0!==(j=$(n.aliasTypeArguments,t.aliasTypeArguments,V,i)))return j}if(am(n)&&!n.target.readonly&&(o=F(nl(n)[0],t))||am(t)&&(t.target.readonly||jp(Ws(n)||n))&&(o=F(n,nl(t)[0])))return o;if(262144&t.flags){if(32&e.getObjectFlags(n)&&!n.declaration.nameType&&F(uu(t),xs(n))&&!(4&Ns(n))){var m=As(n),f=Iu(t,Ss(n));if(o=F(m,f,a))return o}}else if(4194304&t.flags){var _=t.type;if(4194304&n.flags&&(o=F(_,n.type,!1)))return o;if(tm(_)){if(o=F(n,jl(_),a))return o}else if((k=Vs(_))&&-1===F(n,uu(k,t.stringsOnly),a))return-1}else if(8388608&t.flags){if(r===tr||r===rr){var g=t.objectType,y=t.indexType,v=Ws(g)||g,h=Ws(y)||y;if(!Tu(v)&&!Su(h)){var b=2|(v!==g?1:0);if((k=Ru(v,h,t.noUncheckedIndexedAccessCandidate,void 0,b))&&(o=F(n,k,a)))return o}}}else if(Os(t)&&!t.declaration.nameType){var E=As(t),T=Ns(t);if(!(8&T)){if(8388608&E.flags&&E.objectType===n&&E.indexType===Ss(t))return-1;if(!Os(n)){var S=xs(t),x=uu(n,void 0,!0),L=4&T,A=L?fs(S,x):void 0;if(L?!(131072&A.flags):F(S,x)){m=As(t);var D=Ss(t),C=Qf(m,-98305);if(8388608&C.flags&&C.indexType===D){if(o=F(n,C.objectType,a))return o}else{f=Iu(n,A?tu([A,D]):D);if(o=F(f,m,a))return o}}s=u,N(p)}}}else if(134217728&t.flags&&128&n.flags&&Eu(t)){var I=nf(n,t);if(I&&e.every(I,(function(e,n){return ef(e,t.types[n])})))return-1}if(8650752&n.flags){if(8388608&n.flags&&8388608&t.flags){if((o=F(n.objectType,t.objectType,a))&&(o&=F(n.indexType,t.indexType,a)),o)return N(p),o}else if(!(k=Bs(n))||262144&n.flags&&1&k.flags){if(o=F(nn,Qf(t,-67108865)))return N(p),o}else{if(o=F(k,t,!1,void 0,i))return N(p),o;if(o=F(rs(k,n),t,a,void 0,i))return N(p),o}}else if(4194304&n.flags){if(o=F(Ye,t,a))return N(p),o}else if(134217728&n.flags){if(134217728&t.flags&&n.texts.length===t.texts.length&&n.types.length===t.types.length&&e.every(n.texts,(function(e,n){return e===t.texts[n]}))&&e.every(kd(n,_d(K)).types,(function(e,n){return!!(5&t.types[n].flags)||!!F(e,t.types[n],!1)})))return-1;if((k=Ws(n))&&k!==n&&(o=F(k,t,a)))return N(p),o}else if(268435456&n.flags){var k;if(268435456&t.flags&&n.symbol===t.symbol){if(o=F(n.type,t.type,a))return N(p),o}else if((k=Ws(n))&&(o=F(k,t,a)))return N(p),o}else if(16777216&n.flags){if(16777216&t.flags){var O=n.root.inferTypeParameters,P=n.extendsType,M=void 0;if(O){var w=wm(O,void 0,0,F);tf(w.inferences,t.extendsType,P,768),P=kd(P,w.mapper),M=w.mapper}if(Vd(P,t.extendsType)&&(F(n.checkType,t.checkType)||F(t.checkType,n.checkType))&&((o=F(kd(wu(n),M),wu(t),a))&&(o&=F(Fu(n),Fu(t),a)),o))return N(p),o}else{var G=Hs(n);if(G&&(o=F(G,t,a)))return N(p),o}var B=js(n);if(B&&(o=F(B,t,a)))return N(p),o}else{if(r!==er&&r!==nr&&(Z=t,32&e.getObjectFlags(Z)&&4&Ns(Z))&&up(n))return-1;if(Os(t))return Os(n)&&(o=function(e,n,t){if(r===rr||(r===ar?Ns(e)===Ns(n):ks(e)<=ks(n))){var a;if(a=F(xs(n),kd(xs(e),_d(ks(e)<0?H:K)),t)){var i=pd([Ss(e)],[Ss(n)]);if(kd(Ls(e),i)===kd(Ls(n),i))return a&F(kd(As(e),i),As(n),t)}}return 0}(n,t,a))?(N(p),o):0;var U=!!(131068&n.flags);if(r!==ar)n=Zs(n);else if(Os(n))return 0;if(4&e.getObjectFlags(n)&&4&e.getObjectFlags(t)&&n.target===t.target&&!(8192&e.getObjectFlags(n)||8192&e.getObjectFlags(t))){var V,j;if((V=Ap(n.target))===e.emptyArray)return 1;if(void 0!==(j=$(nl(n),nl(t),V,i)))return j}else{if(Vp(t)?Up(n)||tm(n):Up(t)&&tm(n)&&!n.target.readonly)return r!==ar?F(fc(n,1)||Ee,fc(t,1)||Ee,a):0;if((r===er||r===nr)&&up(t)&&32768&e.getObjectFlags(t)&&!up(n))return 0}if(2621440&n.flags&&524288&t.flags){var z=a&&u===p.errorInfo&&!U;if((o=Y(n,t,z,void 0,i))&&(o&=Q(n,t,0,z))&&(o&=Q(n,t,1,z))&&(o&=te(n,t,0,U,z,i))&&(o&=te(n,t,1,U,z,i)),d&&o)u=s||u||p.errorInfo;else if(o)return o}if(2621440&n.flags&&1048576&t.flags){var J=Qf(t,36175872);if(1048576&J.flags){var X=function(n,t){var a=Sf(Gs(n),t);if(!a)return 0;for(var i=1,o=0,s=a;o<s.length;o++){if((i*=Xf(lo(d=s[o])))>25)return e.tracing.instant("check","typeRelatedToDiscriminatedType_DepthLimit",{sourceId:n.id,targetId:t.id,numCombinations:i}),0}for(var c=new Array(a.length),l=new e.Set,u=0;u<a.length;u++){var d,p=lo(d=a[u]);c[u]=1048576&p.flags?p.types:[p],l.add(d.escapedName)}for(var m=e.cartesianProduct(c),f=[],_=function(i){var o=!1;e:for(var s=0,c=t.types;s<c.length;s++){for(var l=c[s],u=function(e){var o=a[e],s=cc(l,o.escapedName);return s?o===s?"continue":q(n,t,o,s,(function(n){return i[e]}),!1,0,W||r===rr)?void 0:"continue-outer":"continue-outer"},d=0;d<a.length;d++){switch(u(d)){case"continue-outer":continue e}}e.pushIfUnique(f,l,e.equateValues),o=!0}if(!o)return{value:0}},g=0,y=m;g<y.length;g++){var v=_(y[g]);if("object"==typeof v)return v.value}for(var h=-1,b=0,E=f;b<E.length;b++){var T=E[b];if((h&=Y(n,T,!1,l,0))&&(h&=Q(n,T,0,!1))&&(h&=Q(n,T,1,!1))&&(!(h&=te(n,T,0,!1,!1,0))||tm(n)&&tm(T)||(h&=te(n,T,1,!1,!1,0))),!h)return h}return h}(n,J);if(X)return X}}}var Z;return 0;function $(n,t,i,c){if(o=function(n,t,a,i,o){if(void 0===n&&(n=e.emptyArray),void 0===t&&(t=e.emptyArray),void 0===a&&(a=e.emptyArray),n.length!==t.length&&r===ar)return 0;for(var s=n.length<=t.length?n.length:t.length,c=-1,l=0;l<s;l++){var u=l<a.length?a[l]:1,d=7&u;if(4!==d){var p=n[l],m=t[l],f=-1;if(8&u?f=r===ar?F(p,m,!1):jd(p,m):1===d?f=F(p,m,i,void 0,o):2===d?f=F(m,p,i,void 0,o):3===d?(f=F(m,p,!1))||(f=F(p,m,i,void 0,o)):(f=F(p,m,i,void 0,o))&&(f&=F(m,p,i,void 0,o)),!f)return 0;c&=f}}return c}(n,t,i,a,c))return o;if(e.some(i,(function(e){return!!(24&e)})))return s=void 0,void N(p);var l=t&&function(e,n){for(var t=0;t<n.length;t++)if(1==(7&n[t])&&16384&e[t].flags)return!0;return!1}(t,i);if(d=!l,i!==e.emptyArray&&!l){if(d&&(!a||!e.some(i,(function(e){return 0==(7&e)}))))return 0;s=u,N(p)}}}(n,t,a,i);return e.tracing.pop(),o}(n,t,a,i):3;if(it&&(it=_),v=b,y--,S){if(-1===S||0===y){if(-1===S||3===S)for(l=d;l<g;l++)r.set(p[l],1|E);g=d}}else r.set(o,2|(a?4:0)|E),g=d;return S}function H(e){return!it||e!==dn&&e!==pn&&e!==ot||it(!1),e}function K(e){return!it||e!==dn&&e!==pn&&e!==ot||it(!0),e}function z(e,n){if(!n||0===e.length)return e;for(var t,r=0;r<e.length;r++)n.has(e[r].escapedName)?t||(t=e.slice(0,r)):t&&t.push(e[r]);return t||e}function q(n,t,r,a,i,o,s,c){var l=e.getDeclarationModifierFlagsFromSymbol(r),u=e.getDeclarationModifierFlagsFromSymbol(a);if(8&l||8&u){if(r.valueDeclaration!==a.valueDeclaration)return o&&(8&l&&8&u?P(e.Diagnostics.Types_have_separate_declarations_of_a_private_property_0,ri(a)):P(e.Diagnostics.Property_0_is_private_in_type_1_but_not_in_type_2,ri(a),ii(8&l?n:t),ii(8&l?t:n))),0}else if(16&u){if(!function(n,t){return!Rp(t,(function(t){return!!(16&e.getDeclarationModifierFlagsFromSymbol(t))&&(r=n,a=kp(t),!Rp(r,(function(e){var n=kp(e);return!!n&&mo(n,a)})));var r,a}))}(r,a))return o&&P(e.Diagnostics.Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2,ri(a),ii(kp(r)||n),ii(kp(a)||t)),0}else if(16&l)return o&&P(e.Diagnostics.Property_0_is_protected_in_type_1_but_public_in_type_2,ri(a),ii(n),ii(t)),0;var d=function(n,t,r,a,i){var o=W&&!!(48&e.getCheckFlags(t)),s=r(n);if(65536&e.getCheckFlags(t)&&!Tr(t).type){var c=Tr(t);e.Debug.assertIsDefined(c.deferralParent),e.Debug.assertIsDefined(c.deferralConstituents);for(var l=!!(1048576&c.deferralParent.flags),u=l?0:-1,d=0,p=c.deferralConstituents;d<p.length;d++){var m=F(s,p[d],!1,void 0,l?0:2);if(l){if(m)return m}else{if(!m)return F(s,Pi(lo(t),o),a);u&=m}}return l&&!u&&o&&(u=F(s,Ce)),l&&!u&&a?F(s,Pi(lo(t),o),a):u}return F(s,Pi(lo(t),o),a,void 0,i)}(r,a,i,o,s);return d?c||!(16777216&r.flags)||16777216&a.flags?d:(o&&P(e.Diagnostics.Property_0_is_optional_in_type_1_but_required_in_type_2,ri(a),ii(n),ii(t)),0):(o&&k(e.Diagnostics.Types_of_property_0_are_incompatible,ri(a)),0)}function J(n,t,r,a){var s=!1;if(r.valueDeclaration&&e.isNamedDeclaration(r.valueDeclaration)&&e.isPrivateIdentifier(r.valueDeclaration.name)&&n.symbol&&32&n.symbol.flags){var c=r.valueDeclaration.name.escapedText,l=e.getSymbolNameForPrivateIdentifier(n.symbol,c);if(l&&cc(n,l)){var p=e.factory.getDeclarationName(n.symbol.valueDeclaration),m=e.factory.getDeclarationName(t.symbol.valueDeclaration);return void P(e.Diagnostics.Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2,kr(c),kr(""===p.escapedText?o:p),kr(""===m.escapedText?o:m))}}var f,_=e.arrayFrom(Xm(n,t,a,!1));if((!i||i.code!==e.Diagnostics.Class_0_incorrectly_implements_interface_1.code&&i.code!==e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code)&&(s=!0),1===_.length){var g=ri(r);P.apply(void 0,__spreadArrays([e.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2,g],oi(n,t))),e.length(r.declarations)&&(f=e.createDiagnosticForNode(r.declarations[0],e.Diagnostics._0_is_declared_here,g),e.Debug.assert(!!u),d?d.push(f):d=[f]),s&&u&&b++}else w(n,t,!1)&&(_.length>5?P(e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more,ii(n),ii(t),e.map(_.slice(0,4),(function(e){return ri(e)})).join(", "),_.length-4):P(e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2,ii(n),ii(t),e.map(_,(function(e){return ri(e)})).join(", ")),s&&u&&b++)}function Y(n,t,a,i,o){if(r===ar)return function(e,n,t){if(!(524288&e.flags&&524288&n.flags))return 0;var r=z(Ms(e),t),a=z(Ms(n),t);if(r.length!==a.length)return 0;for(var i=-1,o=0,s=r;o<s.length;o++){var c=s[o],l=ws(n,c.escapedName);if(!l)return 0;var u=Fp(c,l,F);if(!u)return 0;i&=u}return i}(n,t,i);var s=-1;if(tm(t)){if(Up(n)||tm(n)){if(!t.target.readonly&&(Vp(n)||tm(n)&&n.target.readonly))return 0;var c=tl(n),l=tl(t),u=tm(n)?4&n.target.combinedFlags:4,d=4&t.target.combinedFlags,p=tm(n)?n.target.minLength:0,m=t.target.minLength;if(!u&&c<m)return a&&P(e.Diagnostics.Source_has_0_element_s_but_target_requires_1,c,m),0;if(!d&&l<p)return a&&P(e.Diagnostics.Source_has_0_element_s_but_target_allows_only_1,p,l),0;if(!d&&u)return a&&(p<m?P(e.Diagnostics.Target_requires_0_element_s_but_source_may_have_fewer,m):P(e.Diagnostics.Target_allows_only_0_element_s_but_source_may_have_more,l)),0;for(var f=Math.max(c,l),_=0;_<f;_++){var g=_<l?t.target.elementFlags[_]:d,y=tm(n)&&_<c?n.target.elementFlags[_]:u,v=!!i;if(y&&g){if(8&g&&!(8&y)||8&y&&!(12&g))return a&&P(e.Diagnostics.Element_at_index_0_is_variadic_in_one_type_but_not_in_the_other,_),0;if(1&g&&!(1&y))return a&&P(e.Diagnostics.Property_0_is_optional_in_type_1_but_required_in_type_2,_,ii(n),ii(t)),0;if(v&&((12&y||12&g)&&(v=!1),v&&(null==i?void 0:i.has(""+_))))continue;var h=nl(n)[Math.min(_,c-1)],b=nl(t)[Math.min(_,l-1)];if(!(O=F(h,8&y&&4&g?Nl(b):b,a,void 0,o)))return a&&k(e.Diagnostics.Types_of_property_0_are_incompatible,_),0;s&=O}}return s}if(12&t.target.combinedFlags)return 0}var E=!(r!==er&&r!==nr||of(n)||Wp(n)||tm(n)),T=Ym(n,t,E,!1);if(T)return a&&J(n,t,T,E),0;if(of(t))for(var S=0,x=z(Gs(n),i);S<x.length;S++){if(!ws(t,(I=x[S]).escapedName))if((h=lo(I))!==Ce&&h!==Ie&&h!==Ne)return a&&P(e.Diagnostics.Property_0_does_not_exist_on_type_1,ri(I),ii(t)),0}for(var L=Gs(t),A=tm(n)&&tm(t),D=0,C=z(L,i);D<C.length;D++){var I,N=C[D],R=N.escapedName;if(!(4194304&N.flags)&&(!A||yg(R)||"length"===R))if((I=cc(n,R))&&I!==N){var O;if(!(O=q(n,t,I,N,lo,a,o,r===rr)))return 0;s&=O}}return s}function Q(n,t,a,i){var o,s;if(r===ar)return function(e,n,t){var r=uc(e,t),a=uc(n,t);if(r.length!==a.length)return 0;for(var i=-1,o=0;o<r.length;o++){var s=Gp(r[o],a[o],!1,!1,!1,F);if(!s)return 0;i&=s}return i}(n,t,a);if(t===sn||n===sn)return-1;var c=n.symbol&&fv(n.symbol.valueDeclaration),l=t.symbol&&fv(t.symbol.valueDeclaration),u=uc(n,c&&1===a?0:a),d=uc(t,l&&1===a?0:a);if(1===a&&u.length&&d.length){if(e.isAbstractConstructorType(n)&&!e.isAbstractConstructorType(t))return i&&P(e.Diagnostics.Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type),0;if(!function(n,t,r){if(!n.declaration||!t.declaration)return!0;var a=e.getSelectedEffectiveModifierFlags(n.declaration,24),i=e.getSelectedEffectiveModifierFlags(t.declaration,24);if(8===i)return!0;if(16===i&&8!==a)return!0;if(16!==i&&!a)return!0;r&&P(e.Diagnostics.Cannot_assign_a_0_constructor_type_to_a_1_constructor_type,pi(a),pi(i));return!1}(u[0],d[0],i))return 0}var p=-1,m=R(),f=1===a?$:Z,_=e.getObjectFlags(n),g=e.getObjectFlags(t);if(64&_&&64&g&&n.symbol===t.symbol)for(var y=0;y<d.length;y++){if(!(C=ee(u[y],d[y],!0,i,f(u[y],d[y]))))return 0;p&=C}else if(1===u.length&&1===d.length){var v=r===rr||!!V.noStrictGenericChecks,h=e.first(u),b=e.first(d);if(!(p=ee(h,b,v,i,f(h,b)))&&i&&1===a&&_&g&&(166===(null===(o=b.declaration)||void 0===o?void 0:o.kind)||166===(null===(s=h.declaration)||void 0===s?void 0:s.kind))){var E=function(e){return ai(e,void 0,262144,a)};return P(e.Diagnostics.Type_0_is_not_assignable_to_type_1,E(h),E(b)),P(e.Diagnostics.Types_of_construct_signatures_are_incompatible),p}}else e:for(var T=0,S=d;T<S.length;T++){for(var x=S[T],L=i,A=0,D=u;A<D.length;A++){var C,I=D[A];if(C=ee(I,x,!0,L,f(I,x))){p&=C,N(m);continue e}L=!1}return L&&P(e.Diagnostics.Type_0_provides_no_match_for_the_signature_1,ii(n),ai(x,void 0,void 0,a)),0}return p}function Z(n,t){return 0===n.parameters.length&&0===t.parameters.length?function(n,t){return k(e.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1,ii(n),ii(t))}:function(n,t){return k(e.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible,ii(n),ii(t))}}function $(n,t){return 0===n.parameters.length&&0===t.parameters.length?function(n,t){return k(e.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1,ii(n),ii(t))}:function(n,t){return k(e.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible,ii(n),ii(t))}}function ee(e,n,t,a,i){return sp(t?Bc(e):e,t?Bc(n):n,r===nr?8:0,a,P,i,F,_d(K))}function ne(n,t,r){var a=F(n,t,r);return!a&&r&&P(e.Diagnostics.Index_signatures_are_incompatible),a}function te(n,t,a,i,o,s){if(r===ar)return function(e,n,t){var r=mc(n,t),a=mc(e,t);if(!a&&!r)return-1;if(a&&r&&a.isReadonly===r.isReadonly)return F(a.type,r.type);return 0}(n,t,a);var c=fc(t,a);if(!c||1&c.flags&&!i)return-1;if(Os(n))return fc(t,0)?F(As(n),c,o):0;var l=fc(n,a)||1===a&&fc(n,0);if(l)return ne(l,c,o);if(!(1&s)&&Em(n)){var u=function(n,t,r,a){for(var i=-1,o=0,s=2097152&n.flags?Fs(n):Ms(n);o<s.length;o++){var c=s[o];if(!gp(n,c)){var l=Tr(c).nameType;if(!(l&&8192&l.flags)&&(0===r||yg(c.escapedName))){var u=F(lo(c),t,a);if(!u)return a&&P(e.Diagnostics.Property_0_is_incompatible_with_index_signature,ri(c)),0;i&=u}}}return i}(n,c,a,o);if(u&&0===a){var d=fc(n,1);d&&(u&=ne(d,c,o))}return u}return o&&P(e.Diagnostics.Index_signature_is_missing_in_type_0,ii(n)),0}}function hp(n){if(16&n.flags)return!1;if(3145728&n.flags)return!!e.forEach(n.types,hp);if(465829888&n.flags){var t=Bs(n);if(t&&t!==n)return hp(t)}return Xp(n)||!!(134217728&n.flags)}function bp(n,t,r){return void 0===r&&(r=Hd),VS(n,t,r,!0)||function(n,t){var r=e.getObjectFlags(n);if(20&r&&1048576&t.flags)return e.find(t.types,(function(t){if(524288&t.flags){var a=r&e.getObjectFlags(t);if(4&a)return n.target===t.target;if(16&a)return!!n.aliasSymbol&&n.aliasSymbol===t.aliasSymbol}return!1}))}(n,t)||function(n,t){if(128&e.getObjectFlags(n)&&zf(t,Kp))return e.find(t.types,(function(e){return!Kp(e)}))}(n,t)||function(n,t){var r=0;if(uc(n,r).length>0||uc(n,r=1).length>0)return e.find(t.types,(function(e){return uc(e,r).length>0}))}(n,t)||function(n,t){for(var r,a=0,i=0,o=t.types;i<o.length;i++){var s=o[i],c=tu([uu(n),uu(s)]);if(4194304&c.flags)r=s,a=1/0;else if(1048576&c.flags){var l=e.length(e.filter(c.types,Xp));l>=a&&(r=s,a=l)}else Xp(c)&&1>=a&&(r=s,a=1)}return r}(n,t)}function Ep(n,t,r,a,i){for(var o=n.types.map((function(e){})),s=0,c=t;s<c.length;s++){var l=c[s],u=l[0],d=l[1],p=nc(n,d);if(!(i&&p&&16&e.getCheckFlags(p)))for(var m=0,f=0,_=n.types;f<_.length;f++){var g=xi(_[f],d);g&&r(u(),g)?o[m]=void 0===o[m]||o[m]:o[m]=!1,m++}}var y=o.indexOf(!0);if(-1===y)return a;for(var v=o.indexOf(!0,y+1);-1!==v;){if(!Vd(n.types[y],n.types[v]))return a;v=o.indexOf(!0,v+1)}return n.types[y]}function Tp(n){if(524288&n.flags){var t=Ps(n);return 0===t.callSignatures.length&&0===t.constructSignatures.length&&!t.stringIndexInfo&&!t.numberIndexInfo&&t.properties.length>0&&e.every(t.properties,(function(e){return!!(16777216&e.flags)}))}return!!(2097152&n.flags)&&e.every(n.types,Tp)}function Sp(n,t,r){var a=Zc(n,e.map(n.typeParameters,(function(e){return e===t?r:e})));return a.objectFlags|=8192,a}function xp(e){var n=Tr(e);return Lp(n.typeParameters,n,(function(t,r,a){var i=al(e,ud(n.typeParameters,fd(r,a)));return i.aliasTypeArgumentsContainsMarker=!0,i}))}function Lp(n,t,r){var a,i,o;void 0===n&&(n=e.emptyArray);var s=t.variances;if(!s){e.tracing.push("check","getVariancesWorker",{arity:n.length,id:null!==(o=null!==(a=t.id)&&void 0!==a?a:null===(i=t.declaredType)||void 0===i?void 0:i.id)&&void 0!==o?o:-1}),t.variances=e.emptyArray,s=[];for(var c=function(e){var n=!1,a=!1,i=it;it=function(e){return e?a=!0:n=!0};var o=r(t,e,dn),c=r(t,e,pn),l=(zd(c,o)?1:0)|(zd(o,c)?2:0);3===l&&zd(r(t,e,ot),o)&&(l=4),it=i,(n||a)&&(n&&(l|=8),a&&(l|=16)),s.push(l)},l=0,u=n;l<u.length;l++){c(u[l])}t.variances=s,e.tracing.pop()}return s}function Ap(e){return e===En||e===Tn||8&e.objectFlags?U:Lp(e.typeParameters,e,Sp)}function Dp(e){return 262144&e.flags&&!Us(e)}function Cp(n){return function(n){return!!(4&e.getObjectFlags(n))&&!n.node}(n)&&e.some(nl(n),(function(e){return Dp(e)||Cp(e)}))}function Ip(e,n,t){void 0===t&&(t=0);for(var r=""+e.target.id,a=0,i=nl(e);a<i.length;a++){var o=i[a];if(Dp(o)){var s=n.indexOf(o);s<0&&(s=n.length,n.push(o)),r+="="+s}else t<4&&Cp(o)?r+="<"+Ip(o,n,t+1)+">":r+="-"+o.id}return r}function Np(e,n,t,r){if(r===ar&&e.id>n.id){var a=e;e=n,n=a}var i=t?":"+t:"";if(Cp(e)&&Cp(n)){var o=[];return Ip(e,o)+","+Ip(n,o)+i}return e.id+","+n.id+i}function Rp(n,t){if(!(6&e.getCheckFlags(n)))return t(n);for(var r=0,a=n.containingType.types;r<a.length;r++){var i=cc(a[r],n.escapedName),o=i&&Rp(i,t);if(o)return o}}function kp(e){return e.parent&&32&e.parent.flags?Fo(La(e)):void 0}function Op(e){var n=kp(e),t=n&&Lo(n)[0];return t&&xi(t,e.escapedName)}function Pp(e,n,t){if(t>=5){var r=Mp(e);if(r)for(var a=0,i=0;i<t;i++)if(Mp(n[i])===r&&++a>=5)return!0}return!1}function Mp(n){if(524288&n.flags&&!sf(n)){if(e.getObjectFlags(n)&&n.node)return n.node;if(n.symbol&&!(16&e.getObjectFlags(n)&&32&n.symbol.flags))return n.symbol;if(tm(n))return n.target}if(8388608&n.flags){do{n=n.objectType}while(8388608&n.flags);return n}if(16777216&n.flags)return n.root}function wp(e,n){return 0!==Fp(e,n,jd)}function Fp(n,t,r){if(n===t)return-1;var a=24&e.getDeclarationModifierFlagsFromSymbol(n);if(a!==(24&e.getDeclarationModifierFlagsFromSymbol(t)))return 0;if(a){if(UE(n)!==UE(t))return 0}else if((16777216&n.flags)!=(16777216&t.flags))return 0;return oh(n)!==oh(t)?0:r(lo(n),lo(t))}function Gp(n,t,r,a,i,o){if(n===t)return-1;if(!function(e,n,t){var r=wv(e),a=wv(n),i=Fv(e),o=Fv(n),s=Gv(e),c=Gv(n);return r===a&&i===o&&s===c||!!(t&&i<=o)}(n,t,r))return 0;if(e.length(n.typeParameters)!==e.length(t.typeParameters))return 0;if(t.typeParameters){for(var s=pd(n.typeParameters,t.typeParameters),c=0;c<t.typeParameters.length;c++){if(!((_=n.typeParameters[c])===(m=t.typeParameters[c])||o(kd(Jc(_),s)||De,Jc(m)||De)&&o(kd(Ys(_),s)||De,Ys(m)||De)))return 0}n=Sd(n,s,!0)}var l=-1;if(!a){var u=Nc(n);if(u){var d=Nc(t);if(d){if(!(f=o(u,d)))return 0;l&=f}}}var p=wv(t);for(c=0;c<p;c++){var m,f,_=Ov(n,c);if(!(f=o(m=Ov(t,c),_)))return 0;l&=f}if(!i){var g=Rc(n),y=Rc(t);l&=g||y?function(e,n,t){return e&&n&&Yl(e,n)?e.type===n.type?-1:e.type&&n.type?t(e.type,n.type):0:0}(g,y,o):o(kc(n),kc(t))}return l}function Bp(n){return function(e){for(var n,t=0,r=e;t<r.length;t++){var a=r[t],i=Qp(a);if(n||(n=i),i===a||i!==n)return!1}return!0}(n)?Xl(n):e.reduceLeft(n,(function(e,n){return Wd(e,n)?n:e}))}function Up(n){return!!(4&e.getObjectFlags(n))&&(n.target===En||n.target===Tn)}function Vp(n){return!!(4&e.getObjectFlags(n))&&n.target===Tn}function jp(e){return Up(e)&&!Vp(e)||tm(e)&&!e.target.readonly}function Hp(e){return Up(e)?nl(e)[0]:void 0}function Kp(e){return Up(e)||!(98304&e.flags)&&zd(e,Nn)}function Wp(e){var n=Up(e)?nl(e)[0]:void 0;return n===Ie||n===ze}function zp(e){return tm(e)||!!cc(e,"0")}function qp(e){return Kp(e)||zp(e)}function Jp(e){return!(240512&e.flags)}function Xp(e){return!!(109440&e.flags)}function Yp(n){return!!(16&n.flags)||(1048576&n.flags?!!(1024&n.flags)||e.every(n.types,Xp):Xp(n))}function Qp(n){return 1024&n.flags?Oo(n):128&n.flags?Oe:256&n.flags?Pe:2048&n.flags?Me:512&n.flags?Ue:1048576&n.flags?Xl(e.sameMap(n.types,Qp)):n}function Zp(n){return 1024&n.flags&&nd(n)?Oo(n):128&n.flags&&nd(n)?Oe:256&n.flags&&nd(n)?Pe:2048&n.flags&&nd(n)?Me:512&n.flags&&nd(n)?Ue:1048576&n.flags?Xl(e.sameMap(n.types,Zp)):n}function $p(n){return 8192&n.flags?Ve:1048576&n.flags?Xl(e.sameMap(n.types,$p)):n}function em(e,n){return Ih(e,n)||(e=$p(Zp(e))),e}function nm(e,n,t,r){e&&Xp(e)&&(e=em(e,n?DE(t,n,r):void 0));return e}function tm(n){return!!(4&e.getObjectFlags(n)&&8&n.target.objectFlags)}function rm(e){return tm(e)&&!!(8&e.target.combinedFlags)}function am(e){return rm(e)&&1===e.target.elementFlags.length}function im(e){return sm(e,e.target.fixedLength)}function om(n){return tm(n)?tl(n)-e.findLastIndex(n.target.elementFlags,(function(e){return!(3&e)}))-1:0}function sm(e,n,t,r){void 0===t&&(t=0),void 0===r&&(r=!1);var a=tl(e)-t;if(n<a){for(var i=nl(e),o=[],s=n;s<a;s++){var c=i[s];o.push(8&e.target.elementFlags[s]?Iu(c,Pe):c)}return r?tu(o):Xl(o)}}function cm(e){return"0"===e.value.base10Value}function lm(e){for(var n=0,t=0,r=e;t<r.length;t++){n|=um(r[t])}return n}function um(e){return 1048576&e.flags?lm(e.types):128&e.flags?""===e.value?128:0:256&e.flags?0===e.value?256:0:2048&e.flags?cm(e)?2048:0:512&e.flags?e===we||e===Fe?512:0:117724&e.flags}function dm(e){return 117632&um(e)?Jf(e,(function(e){return!(117632&um(e))})):e}function pm(e){return 4&e.flags?Ct:8&e.flags?It:64&e.flags?Nt:e===Fe||e===we||114691&e.flags||128&e.flags&&""===e.value||256&e.flags&&0===e.value||2048&e.flags&&cm(e)?e:He}function mm(e,n){var t=n&~e.flags&98304;return 0===t?e:Xl(32768===t?[e,Ce]:65536===t?[e,Re]:[e,Ce,Re])}function fm(n){return e.Debug.assert(W),32768&n.flags?n:Xl([n,Ce])}function _m(e){return W?function(e){return Rn||(Rn=bl("NonNullable",524288,void 0)||he),Rn!==he?al(Rn,[e]):If(e,2097152)}(e):e}function gm(e){return W?Xl([e,Ne]):e}function ym(e){return e!==Ne}function vm(e){return W?Jf(e,ym):e}function hm(n,t,r){return r?e.isOutermostOptionalChain(t)?fm(n):gm(n):n}function bm(n,t){return e.isExpressionOfOptionalChainRoot(t)?_m(n):e.isOptionalChain(t)?vm(n):n}function Em(n){return 2097152&n.flags?e.every(n.types,Em):!(!n.symbol||0==(7040&n.symbol.flags)||DT(n))||!!(2048&e.getObjectFlags(n)&&Em(n.source))}function Tm(n,t){var r=mr(n.flags,n.escapedName,8&e.getCheckFlags(n));r.declarations=n.declarations,r.parent=n.parent,r.type=t,r.target=n,n.valueDeclaration&&(r.valueDeclaration=n.valueDeclaration);var a=Tr(n).nameType;return a&&(r.nameType=a),r}function Sm(n){if(!(of(n)&&32768&e.getObjectFlags(n)))return n;var t=n.regularType;if(t)return t;var r=n,a=function(n,t){for(var r=e.createSymbolTable(),a=0,i=Ms(n);a<i.length;a++){var o=i[a],s=lo(o),c=t(s);r.set(o.escapedName,c===s?o:Tm(o,c))}return r}(n,Sm),i=Va(r.symbol,a,r.callSignatures,r.constructSignatures,r.stringIndexInfo,r.numberIndexInfo);return i.flags=r.flags,i.objectFlags|=-32769&r.objectFlags,n.regularType=i,i}function xm(e,n,t){return{parent:e,propertyName:n,siblings:t,resolvedProperties:void 0}}function Lm(e){if(!e.siblings){for(var n=[],t=0,r=Lm(e.parent);t<r.length;t++){var a=r[t];if(of(a)){var i=ws(a,e.propertyName);i&&zf(lo(i),(function(e){n.push(e)}))}}e.siblings=n}return e.siblings}function Am(n){if(!n.resolvedProperties){for(var t=new e.Map,r=0,a=Lm(n);r<a.length;r++){var i=a[r];if(of(i)&&!(1024&e.getObjectFlags(i)))for(var o=0,s=Gs(i);o<s.length;o++){var c=s[o];t.set(c.escapedName,c)}}n.resolvedProperties=e.arrayFrom(t.values())}return n.resolvedProperties}function Dm(e,n){if(!(4&e.flags))return e;var t=lo(e),r=Nm(t,n&&xm(n,e.escapedName,void 0));return r===t?e:Tm(e,r)}function Cm(e){var n=ve.get(e.escapedName);if(n)return n;var t=Tm(e,Ce);return t.flags|=16777216,ve.set(e.escapedName,t),t}function Im(e){return Nm(e,void 0)}function Nm(n,t){if(1572864&e.getObjectFlags(n)){if(void 0===t&&n.widened)return n.widened;var r=void 0;if(98305&n.flags)r=Ee;else if(of(n))r=function(n,t){for(var r=e.createSymbolTable(),a=0,i=Ms(n);a<i.length;a++){var o=i[a];r.set(o.escapedName,Dm(o,t))}if(t)for(var s=0,c=Am(t);s<c.length;s++)o=c[s],r.has(o.escapedName)||r.set(o.escapedName,Cm(o));var l=mc(n,0),u=mc(n,1),d=Va(n.symbol,r,e.emptyArray,e.emptyArray,l&&Wc(Im(l.type),l.isReadonly),u&&Wc(Im(u.type),u.isReadonly));return d.objectFlags|=2113536&e.getObjectFlags(n),d}(n,t);else if(1048576&n.flags){var a=t||xm(void 0,void 0,n.types),i=e.sameMap(n.types,(function(e){return 98304&e.flags?e:Nm(e,a)}));r=Xl(i,e.some(i,up)?2:1)}else 2097152&n.flags?r=tu(e.sameMap(n.types,Im)):(Up(n)||tm(n))&&(r=Zc(n.target,e.sameMap(nl(n),Im)));return r&&void 0===t&&(n.widened=r),r||n}return n}function Rm(n){var t=!1;if(524288&e.getObjectFlags(n)){if(1048576&n.flags)if(e.some(n.types,up))t=!0;else for(var r=0,a=n.types;r<a.length;r++){Rm(u=a[r])&&(t=!0)}if(Up(n)||tm(n))for(var i=0,o=nl(n);i<o.length;i++){Rm(u=o[i])&&(t=!0)}if(of(n))for(var s=0,c=Ms(n);s<c.length;s++){var l=c[s],u=lo(l);524288&e.getObjectFlags(u)&&(Rm(u)||lr(l.valueDeclaration,e.Diagnostics.Object_literal_s_property_0_implicitly_has_an_1_type,ri(l),ii(Im(u))),t=!0)}}return t}function km(n,t,r){var a=ii(Im(t));if(!e.isInJSFile(n)||e.isCheckJsEnabledForFile(e.getSourceFileOfNode(n),V)){var i;switch(n.kind){case 216:case 163:case 162:i=X?e.Diagnostics.Member_0_implicitly_has_an_1_type:e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;break;case 160:var o=n;if(e.isIdentifier(o.name)&&(e.isCallSignatureDeclaration(o.parent)||e.isMethodSignature(o.parent)||e.isFunctionTypeNode(o.parent))&&o.parent.parameters.indexOf(o)>-1&&(Cr(o,o.name.escapedText,788968,void 0,o.name.escapedText,!0)||o.name.originalKeywordKind&&e.isTypeNodeKind(o.name.originalKeywordKind))){var s="arg"+o.parent.parameters.indexOf(o);return void dr(X,n,e.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1,s,e.declarationNameToString(o.name))}i=n.dotDotDotToken?X?e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type:e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage:X?e.Diagnostics.Parameter_0_implicitly_has_an_1_type:e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;break;case 198:if(i=e.Diagnostics.Binding_element_0_implicitly_has_an_1_type,!X)return;break;case 308:return void lr(n,e.Diagnostics.Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type,a);case 251:case 165:case 164:case 167:case 168:case 208:case 209:if(X&&!n.name)return void lr(n,3===r?e.Diagnostics.Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type_annotation:e.Diagnostics.Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type,a);i=X?3===r?e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type:e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type:e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage;break;case 190:return void(X&&lr(n,e.Diagnostics.Mapped_object_type_implicitly_has_an_any_template_type));default:i=X?e.Diagnostics.Variable_0_implicitly_has_an_1_type:e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage}dr(X,n,i,e.declarationNameToString(e.getNameOfDeclaration(n)),a)}}function Om(n,r,a){!(t&&X&&524288&e.getObjectFlags(r))||a&&ug(n)||Rm(r)||km(n,r,a)}function Pm(e,n,t){var r=wv(e),a=wv(n),i=Bv(e),o=Bv(n),s=o?a-1:a,c=i?s:Math.min(r,s),l=Nc(e);if(l){var u=Nc(n);u&&t(l,u)}for(var d=0;d<c;d++)t(Ov(e,d),Ov(n,d));o&&t(Mv(e,c),o)}function Mm(e,n,t){var r=Rc(e),a=Rc(n);r&&a&&Yl(r,a)&&r.type&&a.type?t(r.type,a.type):t(kc(e),kc(n))}function wm(e,n,t,r){return Fm(e.map(Um),n,t,r||Hd)}function Fm(e,n,t,r){var a={inferences:e,signature:n,flags:t,compareTypes:r,mapper:_d((function(e){return Gm(a,e,!0)})),nonFixingMapper:_d((function(e){return Gm(a,e,!1)}))};return a}function Gm(e,n,t){for(var r=e.inferences,a=0;a<r.length;a++){var i=r[a];if(n===i.typeParameter)return t&&!i.isFixed&&(Bm(r),i.isFixed=!0),uf(e,a)}return n}function Bm(e){for(var n=0,t=e;n<t.length;n++){var r=t[n];r.isFixed||(r.inferredType=void 0)}}function Um(e){return{typeParameter:e,candidates:void 0,contraCandidates:void 0,inferredType:void 0,priority:void 0,topLevel:!0,isFixed:!1,impliedArity:void 0}}function Vm(e){return{typeParameter:e.typeParameter,candidates:e.candidates&&e.candidates.slice(),contraCandidates:e.contraCandidates&&e.contraCandidates.slice(),inferredType:e.inferredType,priority:e.priority,topLevel:e.topLevel,isFixed:e.isFixed,impliedArity:e.impliedArity}}function jm(e){return e&&e.mapper}function Hm(n){var t=e.getObjectFlags(n);if(67108864&t)return!!(134217728&t);var r=!!(465829888&n.flags||524288&n.flags&&!Km(n)&&(4&t&&(n.node||e.forEach(nl(n),Hm))||16&t&&n.symbol&&14384&n.symbol.flags&&n.symbol.declarations||131104&t)||3145728&n.flags&&!(1024&n.flags)&&!Km(n)&&e.some(n.types,Hm));return 3899393&n.flags&&(n.objectFlags|=67108864|(r?134217728:0)),r}function Km(n){if(n.aliasSymbol&&!n.aliasTypeArguments){var t=e.getDeclarationOfKind(n.aliasSymbol,254);return!(!t||!e.findAncestor(t.parent,(function(e){return 297===e.kind||256!==e.kind&&"quit"})))}return!1}function Wm(n,t){return!!(n===t||3145728&n.flags&&e.some(n.types,(function(e){return Wm(e,t)}))||16777216&n.flags&&(wu(n)===t||Fu(n)===t))}function zm(n,t,r){if(!Et){var a=n.id+","+t.id+","+r.id;if(bt.has(a))return bt.get(a);Et=!0;var i=function(n,t,r){if(!(mc(n,0)||0!==Gs(n).length&&qm(n)))return;if(Up(n))return Nl(Jm(nl(n)[0],t,r),Vp(n));if(tm(n)){return Fl(e.map(nl(n),(function(e){return Jm(e,t,r)})),4&Ns(t)?e.sameMap(n.target.elementFlags,(function(e){return 2&e?1:e})):n.target.elementFlags,n.target.readonly,n.target.labeledElementDeclarations)}var a=wa(2064,void 0);return a.source=n,a.mappedType=t,a.constraintType=r,a}(n,t,r);return Et=!1,bt.set(a,i),i}}function qm(n){return!(2097152&e.getObjectFlags(n))||of(n)&&e.some(Gs(n),(function(e){return qm(lo(e))}))||tm(n)&&e.some(nl(n),qm)}function Jm(e,n,t){var r=Iu(t.type,Ss(n)),a=As(n),i=Um(r);return tf([i],e,a),Qm(i)||De}function Xm(n,t,r,a){var i,o,s,c,l,u,d;return __generator(this,(function(p){switch(p.label){case 0:i=Gs(t),o=0,s=i,p.label=1;case 1:return o<s.length?Wo(c=s[o])||!r&&(16777216&c.flags||48&e.getCheckFlags(c))?[3,5]:(l=cc(n,c.escapedName))?[3,3]:[4,c]:[3,6];case 2:return p.sent(),[3,5];case 3:return a&&109440&(u=lo(c)).flags?1&(d=lo(l)).flags||ed(d)===ed(u)?[3,5]:[4,c]:[3,5];case 4:p.sent(),p.label=5;case 5:return o++,[3,1];case 6:return[2]}}))}function Ym(e,n,t,r){var a=Xm(e,n,t,r).next();if(!a.done)return a.value}function Qm(e){return e.candidates?Xl(e.candidates,2):e.contraCandidates?tu(e.contraCandidates):void 0}function Zm(e){return!!Sr(e).skipDirectInference}function $m(n){return!(!n.symbol||!e.some(n.symbol.declarations,Zm))}function ef(n,t){if(1048576&t.flags)return!!zf(t,(function(e){return ef(n,e)}));switch(t){case Oe:return!0;case Pe:return""!==n.value&&isFinite(+n.value);case Me:return""!==n.value&&function(n){var t=e.createScanner(99,!1),r=!0;t.setOnError((function(){return r=!1})),t.setText(n+"n");var a=t.scan();40===a&&(a=t.scan());var i=t.getTokenFlags();return r&&9===a&&t.getTextPos()===n.length+1&&!(512&i)}(n.value);case Ge:return"true"===n.value;case we:return"false"===n.value;case Ce:return"undefined"===n.value;case Re:return"null"===n.value;default:return!!(1&t.flags)}}function nf(e,n){var t=e.value,r=n.texts,a=r.length-1,i=r[0],o=r[a];if(t.startsWith(i)&&t.slice(i.length).endsWith(o)){for(var s=[],c=t.slice(i.length,t.length-o.length),l=0,u=1;u<a;u++){var d=r[u],p=d.length>0?c.indexOf(d,l):l<c.length?l+1:-1;if(p<0)return;s.push(td(c.slice(l,p))),l=p+d.length}return s.push(td(c.slice(l))),s}}function tf(n,t,r,a,i){void 0===a&&(a=0),void 0===i&&(i=!1);var o,s,c,l,u=!1,d=1024,p=!0,m=0;function f(t,s){if(Hm(s)){if(t===Se){var c=o;return o=t,f(s,s),void(o=c)}if(t.aliasSymbol&&t.aliasTypeArguments&&t.aliasSymbol===s.aliasSymbol)v(t.aliasTypeArguments,s.aliasTypeArguments,xp(t.aliasSymbol));else if(t===s&&3145728&t.flags)for(var l=0,m=t.types;l<m.length;l++){var h=m[l];f(h,h)}else{if(1048576&s.flags){var T=y(1048576&t.flags?t.types:[t],s.types,rf),L=y(T[0],T[1],af),A=L[0];if(0===(C=L[1]).length)return;if(s=Xl(C),0===A.length)return void _(t,s,1);t=Xl(A)}else if(2097152&s.flags&&e.some(s.types,(function(e){return!!b(e)||Os(e)&&!!b(Ad(e)||He)}))){if(!(1048576&t.flags)){var D=y(2097152&t.flags?t.types:[t],s.types,Vd),C=(A=D[0],D[1]);if(0===A.length||0===C.length)return;t=tu(A),s=tu(C)}}else 41943040&s.flags&&(s=Pu(s));if(8650752&s.flags){if(2097152&e.getObjectFlags(t)||t===Le||t===Ke||64&a&&(t===Te||t===In)||$m(t))return;var I=b(s);if(I){if(!I.isFixed){if((void 0===I.priority||a<I.priority)&&(I.candidates=void 0,I.contraCandidates=void 0,I.topLevel=!0,I.priority=a),a===I.priority){var N=o||t;i&&!u?e.contains(I.contraCandidates,N)||(I.contraCandidates=e.append(I.contraCandidates,N),Bm(n)):e.contains(I.candidates,N)||(I.candidates=e.append(I.candidates,N),Bm(n))}!(64&a)&&262144&s.flags&&I.topLevel&&!Wm(r,s)&&(I.topLevel=!1,Bm(n))}return void(d=Math.min(d,a))}if((F=Lu(s,!1))!==s)g(t,F,f);else if(8388608&s.flags){var R=Lu(s.indexType,!1);if(465829888&R.flags){var k=Au(Lu(s.objectType,!1),R,!1);k&&k!==s&&g(t,k,f)}}}if(!(4&e.getObjectFlags(t)&&4&e.getObjectFlags(s)&&(t.target===s.target||Up(t)&&Up(s)))||t.node&&s.node)if(4194304&t.flags&&4194304&s.flags)i=!i,f(t.type,s.type),i=!i;else if((Yp(t)||4&t.flags)&&4194304&s.flags){var O=function(n){var t=e.createSymbolTable();zf(n,(function(n){if(128&n.flags){var r=e.escapeLeadingUnderscores(n.value),a=mr(4,r);a.type=Ee,n.symbol&&(a.declarations=n.symbol.declarations,a.valueDeclaration=n.symbol.valueDeclaration),t.set(r,a)}}));var r=4&n.flags?Wc(nn,!1):void 0;return Va(void 0,t,e.emptyArray,e.emptyArray,r,void 0)}(t);i=!i,_(O,s.type,128),i=!i}else if(8388608&t.flags&&8388608&s.flags)f(t.objectType,s.objectType),f(t.indexType,s.indexType);else if(268435456&t.flags&&268435456&s.flags)t.symbol===s.symbol&&f(t.type,s.type);else if(16777216&s.flags)g(t,s,S);else if(3145728&s.flags)E(t,s.types,s.flags);else if(1048576&t.flags)for(var P=0,M=t.types;P<M.length;P++){f(M[P],s)}else if(134217728&s.flags)!function(n,t){for(var r=128&n.flags?nf(n,t):134217728&n.flags&&e.arraysEqual(n.texts,t.texts)?n.types:void 0,a=t.types,i=0;i<a.length;i++)f(r?r[i]:He,a[i])}(t,s);else{if(t=rc(t),!(256&a&&467927040&t.flags)){var w=Zs(t);if(w!==t&&p&&!(2621440&w.flags))return p=!1,f(w,s);t=w}2621440&t.flags&&g(t,s,x)}else v(nl(t),nl(s),Ap(t.target));var F;if(25165824&t.flags)(F=Lu(t,i))!==t&&f(F,s)}}}function _(e,n,t){var r=a;a|=t,f(e,n),a=r}function g(n,t,r){var a=n.id+","+t.id,i=s&&s.get(a);if(void 0===i){(s||(s=new e.Map)).set(a,-1);var o=d;d=1024;var u=m,p=Mp(n)||n,f=Mp(t)||t;p&&e.contains(c,p)&&(m|=1),f&&e.contains(l,f)&&(m|=2),3!==m?(p&&(c||(c=[])).push(p),f&&(l||(l=[])).push(f),r(n,t),f&&l.pop(),p&&c.pop()):d=-1,m=u,s.set(a,d),d=Math.min(d,o)}else d=Math.min(d,i)}function y(n,t,r){for(var a,i,o=0,s=t;o<s.length;o++)for(var c=s[o],l=0,u=n;l<u.length;l++){var d=u[l];r(d,c)&&(f(d,c),a=e.appendIfUnique(a,d),i=e.appendIfUnique(i,c))}return[a?e.filter(n,(function(n){return!e.contains(a,n)})):n,i?e.filter(t,(function(n){return!e.contains(i,n)})):t]}function v(e,n,t){for(var r=e.length<n.length?e.length:n.length,a=0;a<r;a++)a<t.length&&2==(7&t[a])?h(e[a],n[a]):f(e[a],n[a])}function h(e,n){z||512&a?(i=!i,f(e,n),i=!i):f(e,n)}function b(e){if(8650752&e.flags)for(var t=0,r=n;t<r.length;t++){var a=r[t];if(e===a.typeParameter)return a}}function E(n,t,r){var i=0;if(1048576&r){for(var o=void 0,s=1048576&n.flags?n.types:[n],c=new Array(s.length),l=!1,u=0,p=t;u<p.length;u++){if(b(x=p[u]))o=x,i++;else for(var m=0;m<s.length;m++){var g=d;d=1024,f(s[m],x),d===a&&(c[m]=!0),l=l||-1===d,d=Math.min(d,g)}}if(0===i){var y=function(n){for(var t,r=0,a=n;r<a.length;r++){var i=a[r],o=2097152&i.flags&&e.find(i.types,(function(e){return!!b(e)}));if(!o||t&&o!==t)return;t=o}return t}(t);return void(y&&_(n,y,1))}if(1===i&&!l){var v=e.flatMap(s,(function(e,n){return c[n]?void 0:e}));if(v.length)return void f(Xl(v),o)}}else for(var h=0,E=t;h<E.length;h++){b(x=E[h])?i++:f(n,x)}if(2097152&r?1===i:i>0)for(var T=0,S=t;T<S.length;T++){var x;b(x=S[T])&&_(n,x,1)}}function T(n,t,r){if(1048576&r.flags){for(var a=!1,i=0,o=r.types;i<o.length;i++){a=T(n,t,o[i])||a}return a}if(4194304&r.flags){var s=b(r.type);if(s&&!s.isFixed&&!$m(n)){var c=zm(n,t,r);c&&_(c,s.typeParameter,2097152&e.getObjectFlags(n)?8:4)}return!0}if(262144&r.flags){_(uu(n),r,16);var l=Bs(r);if(l&&T(n,t,l))return!0;var u=e.map(Gs(n),lo),d=fc(n,0),p=lu(n),m=p&&p.type;return f(Xl(e.append(e.append(u,d),m)),As(t)),!0}return!1}function S(e,n){if(16777216&e.flags)f(e.checkType,n.checkType),f(e.extendsType,n.extendsType),f(wu(e),wu(n)),f(Fu(e),Fu(n));else{var t=a;a|=i?32:0,E(e,[wu(n),Fu(n)],n.flags),a=t}}function x(n,t){if(4&e.getObjectFlags(n)&&4&e.getObjectFlags(t)&&(n.target===t.target||Up(n)&&Up(t)))v(nl(n),nl(t),Ap(n.target));else{if(Os(n)&&Os(t)){f(xs(n),xs(t)),f(As(n),As(t));var r=Ls(n),a=Ls(t);r&&a&&f(r,a)}var i,o;if(32&e.getObjectFlags(t)&&!t.declaration.nameType)if(T(n,t,xs(t)))return;if(!function(e,n){return tm(e)&&tm(n)?function(e,n){return!(8&n.target.combinedFlags)&&n.target.minLength>e.target.minLength||!n.target.hasRestElement&&(e.target.hasRestElement||n.target.fixedLength<e.target.fixedLength)}(e,n):!!Ym(e,n,!1,!0)&&!!Ym(n,e,!1,!0)}(n,t)){if(Up(n)||tm(n)){if(tm(t)){var s=tl(n),c=tl(t),l=nl(t),u=t.target.elementFlags;if(tm(n)&&(o=t,tl(i=n)===tl(o)&&e.every(i.target.elementFlags,(function(e,n){return(12&e)==(12&o.target.elementFlags[n])})))){for(var d=0;d<c;d++)f(nl(n)[d],l[d]);return}var p=tm(n)?Math.min(n.target.fixedLength,t.target.fixedLength):0,m=!tm(n)||s>0&&4&n.target.elementFlags[s-1]?nl(n)[s-1]:void 0,g=12&t.target.combinedFlags?m?om(t):Math.min(om(n),om(t)):0,y=m?0:g;for(d=0;d<p;d++)f(nl(n)[d],l[d]);if(m&&s-p==1)for(d=p;d<c-g;d++)f(8&u[d]?Nl(m):m,l[d]);else{var h=c-p-g;if(2===h&&u[p]&u[p+1]&8&&tm(n)){var E=b(l[p]);E&&void 0!==E.impliedArity&&(f(Vl(n,p,y+s-E.impliedArity),l[p]),f(Vl(n,p+E.impliedArity,y),l[p+1]))}else if(1===h&&8&u[p]){var S=2&t.target.elementFlags[c-1];_(tm(n)?Vl(n,p,y):Nl(m),l[p],S?2:0)}else if(1===h&&4&u[p]){var x=tm(n)?sm(n,p,y):m;x&&f(x,l[p])}}for(d=0;d<g;d++)f(m||nl(n)[s-d-1],l[c-d-1]);return}if(Up(t))return void D(n,t)}!function(e,n){for(var t=Ms(n),r=0,a=t;r<a.length;r++){var i=a[r],o=cc(e,i.escapedName);o&&f(lo(o),lo(i))}}(n,t),L(n,t,0),L(n,t,1),D(n,t)}}}function L(n,t,r){for(var a=uc(n,r),i=uc(t,r),o=a.length,s=i.length,c=o<s?o:s,l=!!(2097152&e.getObjectFlags(n)),u=0;u<c;u++)A(Vc(a[o-c+u]),Bc(i[s-c+u]),l)}function A(e,n,t){if(!t){var r=u,a=n.declaration?n.declaration.kind:0;u=u||165===a||164===a||166===a,Pm(e,n,h),u=r}Mm(e,n,f)}function D(e,n){var t=fc(n,0);t&&((r=fc(e,0)||_c(e,0))&&f(r,t));var r,a=fc(n,1);a&&((r=fc(e,1)||fc(e,0)||_c(e,1))&&f(r,a))}f(t,r)}function rf(e,n){return Vd(e,n)||!!(4&n.flags&&128&e.flags||8&n.flags&&256&e.flags)}function af(e,n){return!!(524288&e.flags&&524288&n.flags&&e.symbol&&e.symbol===n.symbol||e.aliasSymbol&&e.aliasTypeArguments&&e.aliasSymbol===n.aliasSymbol)}function of(n){return!!(128&e.getObjectFlags(n))}function sf(n){return!!(65664&e.getObjectFlags(n))}function cf(n){return 208&n.priority?tu(n.contraCandidates):(t=n.contraCandidates,e.reduceLeft(t,(function(e,n){return Wd(n,e)?n:e})));var t}function lf(n,t){var r,a,i=function(n){if(n.length>1){var t=e.filter(n,sf);if(t.length){var r=Xl(t,2);return e.concatenate(e.filter(n,(function(e){return!sf(e)})),[r])}}return n}(n.candidates),o=(r=n.typeParameter,!!(a=Us(r))&&dh(16777216&a.flags?js(a):a,406978556)),s=!o&&n.topLevel&&(n.isFixed||!Wm(kc(t),n.typeParameter)),c=o?e.sameMap(i,ed):s?e.sameMap(i,Zp):i;return Im(208&n.priority?Xl(c,2):function(n){if(!W)return Bp(n);var t=e.filter(n,(function(e){return!(98304&e.flags)}));return t.length?mm(Bp(t),98304&lm(n)):Xl(n,2)}(c))}function uf(n,t){var r=n.inferences[t];if(!r.inferredType){var a=void 0,i=n.signature;if(i){var o=r.candidates?lf(r,i):void 0;if(r.contraCandidates){var s=cf(r);a=!o||131072&o.flags||!Wd(o,s)?s:o}else if(o)a=o;else if(1&n.flags)a=Ke;else{var c=Ys(r.typeParameter);c&&(a=kd(c,hd(function(n,t){return _d((function(r){return e.findIndex(n.inferences,(function(e){return e.typeParameter===r}))>=t?De:r}))}(n,t),n.nonFixingMapper)))}}else a=Qm(r);r.inferredType=a||df(!!(2&n.flags));var l=Us(r.typeParameter);if(l){var u=kd(l,n.nonFixingMapper);a&&n.compareTypes(a,rs(u,a))||(r.inferredType=a=u)}}return r.inferredType}function df(e){return e?Ee:De}function pf(e){for(var n=[],t=0;t<e.inferences.length;t++)n.push(uf(e,t));return n}function mf(n){switch(n.escapedText){case"document":case"console":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom;case"$":return V.types?e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig:e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery;case"describe":case"suite":case"it":case"test":return V.types?e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig:e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha;case"process":case"require":case"Buffer":case"module":return V.types?e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig:e.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode;case"Map":case"Set":case"Promise":case"Symbol":case"WeakMap":case"WeakSet":case"Iterator":case"AsyncIterator":case"SharedArrayBuffer":case"Atomics":case"AsyncIterable":case"AsyncIterableIterator":case"AsyncGenerator":case"AsyncGeneratorFunction":case"BigInt":case"Reflect":case"BigInt64Array":case"BigUint64Array":return e.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_1_or_later;default:return 289===n.parent.kind?e.Diagnostics.No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer:e.Diagnostics.Cannot_find_name_0}}function ff(n){var t=Sr(n);return t.resolvedSymbol||(t.resolvedSymbol=!e.nodeIsMissing(n)&&Cr(n,n.escapedText,1160127,mf(n),n,!e.isWriteOnlyAccess(n),!1,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1)||he),t.resolvedSymbol}function _f(n){return!!e.findAncestor(n,(function(e){return 176===e.kind||78!==e.kind&&157!==e.kind&&"quit"}))}function gf(e,n,t,r){switch(e.kind){case 78:var a=ff(e);return a!==he?(r?R(r):"-1")+"|"+Hl(n)+"|"+Hl(t)+"|"+(T_(e)?"@":"")+k(a):void 0;case 107:return"0|"+(r?R(r):"-1")+"|"+Hl(n)+"|"+Hl(t);case 225:case 207:return gf(e.expression,n,t,r);case 201:case 202:var i=hf(e);if(void 0!==i){var o=gf(e.expression,n,t,r);return o&&o+"."+i}}}function yf(n,t){switch(t.kind){case 207:case 225:return yf(n,t.expression);case 216:return e.isAssignmentExpression(t)&&yf(n,t.left)||e.isBinaryExpression(t)&&27===t.operatorToken.kind&&yf(n,t.right)}switch(n.kind){case 78:case 79:return 78===t.kind&&ff(n)===ff(t)||(249===t.kind||198===t.kind)&&Na(ff(n))===xa(t);case 107:return 107===t.kind;case 105:return 105===t.kind;case 225:case 207:return yf(n.expression,t);case 201:case 202:return e.isAccessExpression(t)&&hf(n)===hf(t)&&yf(n.expression,t.expression)}return!1}function vf(e,n){return yf(e,n)||216===n.kind&&55===n.operatorToken.kind&&(vf(e,n.left)||vf(e,n.right))}function hf(n){return 201===n.kind?n.name.escapedText:e.isStringOrNumericLiteralLike(n.argumentExpression)?e.escapeLeadingUnderscores(n.argumentExpression.text):void 0}function bf(n,t){for(;e.isAccessExpression(n);)if(yf(n=n.expression,t))return!0;return!1}function Ef(n,t){for(;e.isOptionalChain(n);)if(yf(n=n.expression,t))return!0;return!1}function Tf(n,t){if(n&&1048576&n.flags){var r=nc(n,t);if(r&&2&e.getCheckFlags(r))return void 0===r.isDiscriminantProperty&&(r.isDiscriminantProperty=192==(192&r.checkFlags)&&!dh(lo(r),465829888)),!!r.isDiscriminantProperty}return!1}function Sf(e,n){for(var t,r=0,a=e;r<a.length;r++){var i=a[r];if(Tf(n,i.escapedName)){if(t){t.push(i);continue}t=[i]}}return t}function xf(e,n){return yf(e,n)||bf(e,n)}function Lf(e){return(!e.id||e.id<0)&&(e.id=u,u++),e.id}function Af(e,n){if(e!==n){if(131072&n.flags)return n;var t=Jf(e,(function(e){return function(e,n){if(!(1048576&e.flags))return zd(e,n);for(var t=0,r=e.types;t<r.length;t++)if(zd(r[t],n))return!0;return!1}(n,e)}));if(512&n.flags&&nd(n)&&(t=Yf(t,$u)),zd(n,t))return t}return e}function Df(e){var n=Ps(e);return!!(n.callSignatures.length||n.constructSignatures.length||n.members.get("bind")&&Wd(e,vn))}function Cf(n){var t=n.flags;if(4&t)return W?16317953:16776705;if(128&t){var r=""===n.value;return W?r?12123649:7929345:r?12582401:16776705}if(40&t)return W?16317698:16776450;if(256&t){var a=0===n.value;return W?a?12123394:7929090:a?12582146:16776450}if(64&t)return W?16317188:16775940;if(2048&t){a=cm(n);return W?a?12122884:7928580:a?12581636:16775940}return 16&t?W?16316168:16774920:528&t?W?n===we||n===Fe?12121864:7927560:n===we||n===Fe?12580616:16774920:524288&t?16&e.getObjectFlags(n)&&up(n)?W?16318463:16777215:Df(n)?W?7880640:16728e3:W?7888800:16736160:49152&t?9830144:65536&t?9363232:12288&t?W?7925520:16772880:67108864&t?W?7888800:16736160:131072&t?0:465829888&t?Cf(Ws(n)||De):3145728&t?function(e){for(var n=0,t=0,r=e;t<r.length;t++)n|=Cf(r[t]);return n}(n.types):16777215}function If(e,n){return Jf(e,(function(e){return 0!=(Cf(e)&n)}))}function Nf(e,n){if(n){var t=Uh(n);return Xl([If(e,524288),t])}return e}function Rf(e,n){var t=ou(n);if(!qo(t))return xe;var r=Zo(t);return x_(xi(e,r),n)||yg(r)&&Of(fc(e,1))||Of(fc(e,0))||xe}function kf(e,n){return qf(e,zp)&&function(e,n){return xi(e,""+n)||(qf(e,tm)?Yf(e,(function(e){return im(e)||Ce})):void 0)}(e,n)||Of(aE(65,e,Ce,void 0))||xe}function Of(e){return e&&V.noUncheckedIndexedAccess?Xl([e,Ce]):e}function Pf(e){return Nl(aE(65,e,Ce,void 0)||xe)}function Mf(e){return 216===e.parent.kind&&e.parent.left===e||239===e.parent.kind&&e.parent.initializer===e}function wf(e){return Rf(Ff(e.parent),e.name)}function Ff(e){var n=e.parent;switch(n.kind){case 238:return Oe;case 239:return rE(n)||xe;case 216:return function(e){return 199===e.parent.kind&&Mf(e.parent)||288===e.parent.kind&&Mf(e.parent.parent)?Nf(Ff(e),e.right):Uh(e.right)}(n);case 210:return Ce;case 199:return function(e,n){return kf(Ff(e),e.elements.indexOf(n))}(n,e);case 220:return function(e){return Pf(Ff(e.parent))}(n);case 288:return wf(n);case 289:return function(e){return Nf(wf(e),e.objectAssignmentInitializer)}(n)}return xe}function Gf(e){return Sr(e).resolvedType||Uh(e)}function Bf(e){return 249===e.kind?function(e){return e.initializer?Gf(e.initializer):238===e.parent.parent.kind?Oe:239===e.parent.parent.kind&&rE(e.parent.parent)||xe}(e):function(e){var n=e.parent,t=Bf(n.parent);return Nf(196===n.kind?Rf(t,e.propertyName||e.name):e.dotDotDotToken?Pf(t):kf(t,n.elements.indexOf(e)),e.initializer)}(e)}function Uf(e){switch(e.kind){case 207:return Uf(e.expression);case 216:switch(e.operatorToken.kind){case 62:case 74:case 75:case 76:return Uf(e.left);case 27:return Uf(e.right)}}return e}function Vf(e){var n=e.parent;return 207===n.kind||216===n.kind&&62===n.operatorToken.kind&&n.left===e||216===n.kind&&27===n.operatorToken.kind&&n.right===e?Vf(n):e}function jf(e){return 284===e.kind?ed(Uh(e.expression)):He}function Hf(e){var n=Sr(e);if(!n.switchTypes){n.switchTypes=[];for(var t=0,r=e.caseBlock.clauses;t<r.length;t++){var a=r[t];n.switchTypes.push(jf(a))}}return n.switchTypes}function Kf(n,t){for(var r=[],a=0,i=n.caseBlock.clauses;a<i.length;a++){var o=i[a];if(284===o.kind){if(e.isStringLiteralLike(o.expression)){r.push(o.expression.text);continue}return e.emptyArray}t&&r.push(void 0)}return r}function Wf(e,n){return e===n||1048576&n.flags&&function(e,n){if(1048576&e.flags){for(var t=0,r=e.types;t<r.length;t++){var a=r[t];if(!Kl(n.types,a))return!1}return!0}if(1024&e.flags&&Oo(e)===n)return!0;return Kl(n.types,e)}(e,n)}function zf(n,t){return 1048576&n.flags?e.forEach(n.types,t):t(n)}function qf(n,t){return 1048576&n.flags?e.every(n.types,t):t(n)}function Jf(n,t){if(1048576&n.flags){var r=n.types,a=e.filter(r,t);return a===r?n:Ql(a,n.objectFlags)}return 131072&n.flags||t(n)?n:He}function Xf(e){return 1048576&e.flags?e.types.length:1}function Yf(e,n,t){if(131072&e.flags)return e;if(!(1048576&e.flags))return n(e);for(var r,a=0,i=e.types;a<i.length;a++){var o=n(i[a]);o&&(r?r.push(o):r=[o])}return r&&Xl(r,t?0:1)}function Qf(e,n){return Jf(e,(function(e){return 0!=(e.flags&n)}))}function Zf(e,n){return Wf(Oe,e)&&dh(n,128)||Wf(Pe,e)&&dh(n,256)||Wf(Me,e)&&dh(n,2048)?Yf(e,(function(e){return 4&e.flags?Qf(n,132):8&e.flags?Qf(n,264):64&e.flags?Qf(n,2112):e})):e}function $f(e){return 0===e.flags}function e_(e){return 0===e.flags?e.type:e}function n_(e,n){return n?{flags:0,type:131072&e.flags?Ke:e}:e}function t_(e){return ye[e.id]||(ye[e.id]=function(e){var n=wa(256);return n.elementType=e,n}(e))}function r_(e,n){var t=Sm(Qp(jh(n)));return Wf(t,e.elementType)?e:t_(Xl([e.elementType,t]))}function a_(e){return e.finalArrayType||(e.finalArrayType=131072&(n=e.elementType).flags?In:Nl(1048576&n.flags?Xl(n.types,2):n));var n}function i_(n){return 256&e.getObjectFlags(n)?a_(n):n}function o_(n){return 256&e.getObjectFlags(n)?n.elementType:He}function s_(n,t){return function(n){for(var t=!1,r=0,a=n;r<a.length;r++){var i=a[r];if(!(131072&i.flags)){if(!(256&e.getObjectFlags(i)))return!1;t=!0}}return t}(n)?t_(Xl(e.map(n,o_))):Xl(e.sameMap(n,i_),t)}function c_(n){var t=Vf(n),r=t.parent,a=e.isPropertyAccessExpression(r)&&("length"===r.name.escapedText||203===r.parent.kind&&e.isIdentifier(r.name)&&e.isPushOrUnshiftIdentifier(r.name)),i=202===r.kind&&r.expression===t&&216===r.parent.kind&&62===r.parent.operatorToken.kind&&r.parent.left===r&&!e.isAssignmentTarget(r.parent)&&ph(Uh(r.argumentExpression),296);return a||i}function l_(n,t){if(8752&n.flags)return lo(n);if(7&n.flags){if(262144&e.getCheckFlags(n)){var r=n.syntheticOrigin;if(r&&l_(r))return lo(n)}var a=n.valueDeclaration;if(a){if(function(n){return(249===n.kind||160===n.kind||163===n.kind||162===n.kind)&&!!e.getEffectiveTypeAnnotationNode(n)}(a))return lo(n);if(e.isVariableDeclaration(a)&&239===a.parent.parent.kind){var i=a.parent.parent,o=u_(i.expression,void 0);if(o)return aE(i.awaitModifier?15:13,o,Ce,void 0)}t&&e.addRelatedInfo(t,e.createDiagnosticForNode(a,e.Diagnostics._0_needs_an_explicit_type_annotation,ri(n)))}}}function u_(n,t){if(!(16777216&n.flags))switch(n.kind){case 78:var r=Na(ff(n));return l_(2097152&r.flags?ea(r):r,t);case 107:return function(n){var t=e.getThisContainer(n,!1);if(e.isFunctionLike(t)){var r=Lc(t);if(r.thisParameter)return l_(r.thisParameter)}if(e.isClassLike(t.parent)){var a=xa(t.parent);return e.hasSyntacticModifier(t,32)?lo(a):Fo(a).thisType}}(n);case 105:return P_(n);case 201:var a=u_(n.expression,t),i=a&&cc(a,n.name.escapedText);return i&&l_(i,t);case 207:return u_(n.expression,t)}}function d_(n){var t=Sr(n),r=t.effectsSignature;if(void 0===r){var a=void 0;233===n.parent.kind?a=u_(n.expression,void 0):105!==n.expression.kind&&(a=e.isOptionalChain(n)?ny(bm(Hh(n.expression),n.expression),n.expression):Xg(n.expression));var i=uc(a&&Zs(a)||De,0),o=1!==i.length||i[0].typeParameters?e.some(i,p_)?mv(n):void 0:i[0];r=t.effectsSignature=o&&p_(o)?o:lt}return r===lt?void 0:r}function p_(e){return!!(Rc(e)||e.declaration&&131072&(Oc(e.declaration)||De).flags)}function m_(e){var n=__(e,!1);return et=e,nt=n,n}function f_(n){var t=e.skipParentheses(n);return 94===t.kind||216===t.kind&&(55===t.operatorToken.kind&&(f_(t.left)||f_(t.right))||56===t.operatorToken.kind&&f_(t.left)&&f_(t.right))}function __(n,t){for(;;){if(n===et)return nt;var r=n.flags;if(4096&r){if(!t){var a=Lf(n),i=Kt[a];return void 0!==i?i:Kt[a]=__(n,!0)}t=!1}if(368&r)n=n.antecedent;else if(512&r){var o=d_(n.node);if(o){var s=Rc(o);if(s&&3===s.kind&&!s.type){var c=n.node.arguments[s.parameterIndex];if(c&&f_(c))return!1}if(131072&kc(o).flags)return!1}n=n.antecedent}else{if(4&r)return e.some(n.antecedents,(function(e){return __(e,!1)}));if(8&r)n=n.antecedents[0];else{if(!(128&r)){if(1024&r){et=void 0;var l=n.target,u=l.antecedents;l.antecedents=n.antecedents;var d=__(n.antecedent,!1);return l.antecedents=u,d}return!(1&r)}if(n.clauseStart===n.clauseEnd&&$v(n.switchStatement))return!1;n=n.antecedent}}}}function g_(n,t){for(;;){var r=n.flags;if(4096&r){if(!t){var a=Lf(n),i=Wt[a];return void 0!==i?i:Wt[a]=g_(n,!0)}t=!1}if(496&r)n=n.antecedent;else if(512&r){if(105===n.node.expression.kind)return!0;n=n.antecedent}else{if(4&r)return e.every(n.antecedents,(function(e){return g_(e,!1)}));if(!(8&r)){if(1024&r){var o=n.target,s=o.antecedents;o.antecedents=n.antecedents;var c=g_(n.antecedent,!1);return o.antecedents=s,c}return!!(1&r)}n=n.antecedents[0]}}}function y_(n,t,r,a,i){var o;void 0===r&&(r=t);var s=!1,c=0;if(At)return xe;if(!n.flowNode||!i&&!(536624127&t.flags))return t;Dt++;var l=Lt,u=e_(m(n.flowNode));Lt=l;var d=256&e.getObjectFlags(u)&&c_(n)?In:i_(u);return d===qe||n.parent&&225===n.parent.kind&&131072&If(d,2097152).flags?t:d;function p(){return s?o:(s=!0,o=gf(n,t,r,a))}function m(i){if(2e3===c)return e.tracing.instant("check","getTypeAtFlowNode_DepthLimit",{flowId:i.id}),At=!0,o=n,s=e.findAncestor(o,e.isFunctionOrModuleBlock),u=e.getSourceFileOfNode(o),d=e.getSpanOfTokenAtPosition(u,s.statements.pos),Yt.add(e.createFileDiagnostic(u,d.start,d.length,e.Diagnostics.The_containing_function_or_module_body_is_too_large_for_control_flow_analysis)),xe;var o,s,u,d,p;for(c++;;){var f=i.flags;if(4096&f){for(var g=l;g<Lt;g++)if(jt[g]===i)return c--,Ht[g];p=i}var b=void 0;if(16&f){if(!(b=_(i))){i=i.antecedent;continue}}else if(512&f){if(!(b=y(i))){i=i.antecedent;continue}}else if(96&f)b=h(i);else if(128&f)b=T(i);else if(12&f){if(1===i.antecedents.length){i=i.antecedents[0];continue}b=4&f?S(i):x(i)}else if(256&f){if(!(b=v(i))){i=i.antecedent;continue}}else if(1024&f){var E=i.target,L=E.antecedents;E.antecedents=i.antecedents,b=m(i.antecedent),E.antecedents=L}else if(2&f){var A=i.node;if(A&&A!==a&&201!==n.kind&&202!==n.kind&&107!==n.kind){i=A.flowNode;continue}b=r}else b=Wb(t);return p&&(jt[Lt]=p,Ht[Lt]=b,Lt++),c--,b}}function f(e){var t=e.node;return x_(249===t.kind||198===t.kind?Bf(t):Ff(t),n)}function _(r){var a=r.node;if(yf(n,a)){if(!m_(r))return qe;if(2===e.getAssignmentTargetKind(a)){var i=m(r.antecedent);return n_(Qp(e_(i)),$f(i))}if(t===Te||t===In){if(function(e){return 249===e.kind&&e.initializer&&Oi(e.initializer)||198!==e.kind&&216===e.parent.kind&&Oi(e.parent.right)}(a))return t_(He);var o=Zp(f(r));return zd(o,t)?o:Cn}return 1048576&t.flags?Af(t,f(r)):t}if(bf(n,a)){if(!m_(r))return qe;if(e.isVariableDeclaration(a)&&(e.isInJSFile(a)||e.isVarConst(a))){var s=e.getDeclaredExpandoInitializer(a);if(s&&(208===s.kind||209===s.kind))return m(r.antecedent)}return t}if(e.isVariableDeclaration(a)&&238===a.parent.parent.kind&&yf(n,a.parent.parent.expression))return Qg(e_(m(r.antecedent)))}function g(n,t){var r=e.skipParentheses(t);if(94===r.kind)return qe;if(216===r.kind){if(55===r.operatorToken.kind)return g(g(n,r.left),r.right);if(56===r.operatorToken.kind)return Xl([g(n,r.left),g(n,r.right)])}return j(n,r,!0)}function y(e){var n=d_(e.node);if(n){var t=Rc(n);if(t&&(2===t.kind||3===t.kind)){var r=m(e.antecedent),a=i_(e_(r)),i=t.type?V(a,t,e.node,!0):3===t.kind&&t.parameterIndex>=0&&t.parameterIndex<e.node.arguments.length?g(a,e.node.arguments[t.parameterIndex]):a;return i===a?r:n_(i,$f(r))}if(131072&kc(n).flags)return qe}}function v(r){if(t===Te||t===In){var a=r.node,i=203===a.kind?a.expression.expression:a.left.expression;if(yf(n,Uf(i))){var o=m(r.antecedent),s=e_(o);if(256&e.getObjectFlags(s)){var c=s;if(203===a.kind)for(var l=0,u=a.arguments;l<u.length;l++){c=r_(c,u[l])}else ph(jh(a.left.argumentExpression),296)&&(c=r_(c,a.right));return c===s?o:n_(c,$f(o))}return o}}}function h(e){var n=m(e.antecedent),t=e_(n);if(131072&t.flags)return n;var r=0!=(32&e.flags),a=i_(t),i=j(a,e.node,r);return i===a?n:n_(i,$f(n))}function T(t){var r=t.switchStatement.expression,a=m(t.antecedent),i=e_(a);return yf(n,r)?i=P(i,t.switchStatement,t.clauseStart,t.clauseEnd):211===r.kind&&yf(n,r.expression)?i=function(n,t,r,a){var i=Kf(t,!0);if(!i.length)return n;var o,s,c=e.findIndex(i,(function(e){return void 0===e})),l=r===a||c>=r&&c<a;if(c>-1){var u=i.filter((function(e){return void 0!==e})),d=c<r?r-1:r,p=c<a?a-1:a;o=u.slice(d,p),s=Zv(d,p,u,l)}else o=i.slice(r,a),s=Zv(r,a,i,l);if(l)return Jf(n,(function(e){return(Cf(e)&s)===s}));var m=If(Xl(o.map((function(e){return M(n,e)||n}))),s);return If(Yf(n,w(m)),s)}(i,t.switchStatement,t.clauseStart,t.clauseEnd):(W&&(Ef(r,n)?i=O(i,t.switchStatement,t.clauseStart,t.clauseEnd,(function(e){return!(163840&e.flags)})):211===r.kind&&Ef(r.expression,n)&&(i=O(i,t.switchStatement,t.clauseStart,t.clauseEnd,(function(e){return!(131072&e.flags||128&e.flags&&"undefined"===e.value)})))),L(r,i)&&(i=A(i,r,(function(e){return P(e,t.switchStatement,t.clauseStart,t.clauseEnd)})))),n_(i,$f(a))}function S(n){for(var a,i=[],o=!1,s=!1,c=0,l=n.antecedents;c<l.length;c++){var u=l[c];if(!a&&128&u.flags&&u.clauseStart===u.clauseEnd)a=u;else{if((p=e_(d=m(u)))===t&&t===r)return p;e.pushIfUnique(i,p),Wf(p,t)||(o=!0),$f(d)&&(s=!0)}}if(a){var d,p=e_(d=m(a));if(!e.contains(i,p)&&!$v(a.switchStatement)){if(p===t&&t===r)return p;i.push(p),Wf(p,t)||(o=!0),$f(d)&&(s=!0)}}return n_(s_(i,o?2:1),s)}function x(n){var r=Lf(n),a=Gt[r]||(Gt[r]=new e.Map),i=p();if(!i)return t;var o=a.get(i);if(o)return o;for(var s=St;s<xt;s++)if(Bt[s]===n&&Ut[s]===i&&Vt[s].length)return n_(s_(Vt[s],1),!0);for(var c,l=[],u=!1,d=0,f=n.antecedents;d<f.length;d++){var _=f[d],g=void 0;if(c){Bt[xt]=n,Ut[xt]=i,Vt[xt]=l,xt++;var y=tt;tt=void 0,g=m(_),tt=y,xt--;var v=a.get(i);if(v)return v}else g=c=m(_);var h=e_(g);if(e.pushIfUnique(l,h),Wf(h,t)||(u=!0),h===t)break}var b=s_(l,u?2:1);return $f(c)?n_(b,!0):(a.set(i,b),b)}function L(r,a){var i=1048576&t.flags?t:a;if(!(1048576&i.flags&&e.isAccessExpression(r)))return!1;var o=hf(r);return void 0!==o&&(yf(n,r.expression)&&Tf(i,o))}function A(e,n,t){var r=hf(n);if(void 0===r)return e;var a=xi(e,r);if(!a)return e;var i=t(a);return Jf(e,(function(e){var n=function(e,n){return xi(e,n)||yg(n)&&fc(e,1)||fc(e,0)||De}(e,r);return!(131072&n.flags)&&Jd(n,i)}))}function D(e,t,r){return yf(n,t)?If(e,r?4194304:8388608):(W&&r&&Ef(t,n)&&(e=If(e,2097152)),L(t,e)?A(e,t,(function(e){return If(e,r?4194304:8388608)})):e)}function C(n,t,r){if(1572864&n.flags||xu(n)||2097152&n.flags&&e.every(n.types,(function(e){return e.symbol!==re}))){var a=e.escapeLeadingUnderscores(t.text);return Jf(n,(function(e){return function(e,n,t){if(mc(e,0))return!0;var r=cc(e,n);return r?!!(16777216&r.flags)||t:!t}(e,a,r)}))}return n}function I(t,r,a){switch(r.operatorToken.kind){case 62:case 74:case 75:case 76:return D(j(t,r.right,a),r.left,a);case 34:case 35:case 36:case 37:var i=r.operatorToken.kind,o=Uf(r.left),s=Uf(r.right);if(211===o.kind&&e.isStringLiteralLike(s))return k(t,o,i,s,a);if(211===s.kind&&e.isStringLiteralLike(o))return k(t,s,i,o,a);if(yf(n,o))return R(t,i,s,a);if(yf(n,s))return R(t,i,o,a);if(W&&(Ef(o,n)?t=N(t,i,s,a):Ef(s,n)&&(t=N(t,i,o,a))),L(o,t))return A(t,o,(function(e){return R(e,i,s,a)}));if(L(s,t))return A(t,s,(function(e){return R(e,i,o,a)}));if(F(o))return G(t,i,s,a);if(F(s))return G(t,i,o,a);break;case 101:return function(t,r,a){var i=Uf(r.left);if(!yf(n,i))return a&&W&&Ef(i,n)?If(t,2097152):t;var o,s=Uh(r.right);if(!qd(s,vn))return t;var c=cc(s,"prototype");if(c){var l=lo(c);Li(l)||(o=l)}if(Li(t)&&(o===yn||o===vn))return t;if(!o){var u=uc(s,1);o=u.length?Xl(e.map(u,(function(e){return kc(Bc(e))}))):nn}if(!a&&1048576&s.flags){if(!e.find(s.types,(function(e){return!ho(e)})))return t}return B(t,o,a,qd)}(t,r,a);case 100:var c=Uf(r.right);if(e.isStringLiteralLike(r.left)&&yf(n,c))return C(t,r.left,a);break;case 27:return j(t,r.right,a)}return t}function N(e,n,t,r){var a=34===n||36===n,i=34===n||35===n?98304:32768,o=Uh(t);return a!==r&&qf(o,(function(e){return!!(e.flags&i)}))||a===r&&qf(o,(function(e){return!(e.flags&(3|i))}))?If(e,2097152):e}function R(e,n,t,r){if(1&e.flags)return e;35!==n&&37!==n||(r=!r);var a=Uh(t);if(2&e.flags&&r&&(36===n||37===n))return 67239932&a.flags?a:524288&a.flags?Je:e;if(98304&a.flags)return W?If(e,34===n||35===n?r?262144:2097152:65536&a.flags?r?131072:1048576:r?65536:524288):e;if(r)return Zf(Jf(e,34===n?function(e){return Xd(e,a)||(n=a,0!=(524&e.flags)&&0!=(28&n.flags));var n}:function(e){return Xd(e,a)}),a);if(Xp(a)){var i=ed(a);return Jf(e,(function(e){return Xp(e)?!Xd(e,a):ed(e)!==i}))}return e}function k(e,t,r,a,i){35!==r&&37!==r||(i=!i);var o=Uf(t.expression);if(!yf(n,o))return W&&Ef(o,n)&&i===("undefined"!==a.text)?If(e,2097152):e;if(1&e.flags&&"function"===a.text)return e;if(i&&2&e.flags&&"object"===a.text){if(216===t.parent.parent.kind){var s=t.parent.parent;if(55===s.operatorToken.kind&&s.right===t.parent&&vf(n,s.left))return Je}return Xl([Je,Re])}var c=i?b.get(a.text)||128:E.get(a.text)||32768,l=M(e,a.text);return If(i&&l?Yf(e,w(l)):e,c)}function O(n,t,r,a,i){return r!==a&&e.every(Hf(t).slice(r,a),i)?If(n,2097152):n}function P(n,t,r,a){var i=Hf(t);if(!i.length)return n;var o=i.slice(r,a),s=r===a||e.contains(o,He);if(2&n.flags&&!s){for(var c=void 0,l=0;l<o.length;l+=1){var u=o[l];if(67239932&u.flags)void 0!==c&&c.push(u);else{if(!(524288&u.flags))return n;void 0===c&&(c=o.slice(0,l)),c.push(Je)}}return Xl(void 0===c?o:c)}var d=Xl(o),p=131072&d.flags?He:Zf(Jf(n,(function(e){return Xd(d,e)})),d);if(!s)return p;var m=Jf(n,(function(n){return!(Xp(n)&&e.contains(i,ed(n)))}));return 131072&p.flags?m:Xl([p,m])}function M(e,n){switch(n){case"function":return 1&e.flags?e:vn;case"object":return 2&e.flags?Xl([Je,Re]):e;default:return Zt.get(n)}}function w(e){return function(n){if(Wd(n,e))return n;if(Wd(e,n))return e;if(465829888&n.flags){var t=Ws(n)||Ee;if(Wd(e,t))return tu([n,e])}return n}}function F(t){return(e.isPropertyAccessExpression(t)&&"constructor"===e.idText(t.name)||e.isElementAccessExpression(t)&&e.isStringLiteralLike(t.argumentExpression)&&"constructor"===t.argumentExpression.text)&&yf(n,t.expression)}function G(n,t,r,a){if(a?34!==t&&36!==t:35!==t&&37!==t)return n;var i=Uh(r);if(!XT(i)&&!ho(i))return n;var o=cc(i,"prototype");if(!o)return n;var s=lo(o),c=Li(s)?void 0:s;return c&&c!==yn&&c!==vn?Li(n)?c:Jf(n,(function(n){return function(n,t){if(524288&n.flags&&1&e.getObjectFlags(n)||524288&t.flags&&1&e.getObjectFlags(t))return n.symbol===t.symbol;return Wd(n,t)}(n,c)})):n}function B(e,n,t,r){if(!t)return Jf(e,(function(e){return!r(e,n)}));if(1048576&e.flags){var a=Jf(e,(function(e){return r(e,n)}));if(!(131072&a.flags))return a}return Wd(n,e)?n:Wd(e,n)?e:tu([e,n])}function U(t,r,a){if(function(e,n){if(e.arguments)for(var t=0,r=e.arguments;t<r.length;t++)if(xf(n,r[t]))return!0;return!(201!==e.expression.kind||!xf(n,e.expression.expression))}(r,n)){var i=a||!e.isCallChain(r)?d_(r):void 0,o=i&&Rc(i);if(o&&(0===o.kind||1===o.kind))return V(t,o,r,a)}return t}function V(t,r,a,i){if(r.type&&(!Li(t)||r.type!==yn&&r.type!==vn)){var o=function(n,t){if(1===n.kind||3===n.kind)return t.arguments[n.parameterIndex];var r=e.skipParentheses(t.expression);return e.isAccessExpression(r)?e.skipParentheses(r.expression):void 0}(r,a);if(o){if(yf(n,o))return B(t,r.type,i,Wd);if(W&&i&&Ef(o,n)&&!(65536&Cf(r.type))&&(t=If(t,2097152)),L(o,t))return A(t,o,(function(e){return B(e,r.type,i,Wd)}))}}return t}function j(t,r,a){if(e.isExpressionOfOptionalChainRoot(r)||e.isBinaryExpression(r.parent)&&60===r.parent.operatorToken.kind&&r.parent.left===r)return function(e,t,r){if(yf(n,t))return If(e,r?2097152:262144);if(L(t,e))return A(e,t,(function(e){return If(e,r?2097152:262144)}));return e}(t,r,a);switch(r.kind){case 78:case 107:case 105:case 201:case 202:return D(t,r,a);case 203:return U(t,r,a);case 207:case 225:return j(t,r.expression,a);case 216:return I(t,r,a);case 214:if(53===r.operator)return j(t,r.operand,!a)}return t}}function v_(n){return e.findAncestor(n.parent,(function(n){return e.isFunctionLike(n)&&!e.getImmediatelyInvokedFunctionExpression(n)||257===n.kind||297===n.kind||163===n.kind}))}function h_(n){var t,r=e.getRootDeclaration(n.valueDeclaration).parent,a=Sr(r);return 8388608&a.flags||(a.flags|=8388608,t=r,e.findAncestor(t.parent,(function(n){return e.isFunctionLike(n)&&!!(8388608&Sr(n).flags)}))||b_(r)),n.isAssigned||!1}function b_(n){if(78===n.kind){if(e.isAssignmentTarget(n)){var t=ff(n);t.valueDeclaration&&160===e.getRootDeclaration(t.valueDeclaration).kind&&(t.isAssigned=!0)}}else e.forEachChild(n,b_)}function E_(e){return 3&e.flags&&0!=(2&Wg(e))&&lo(e)!==In}function T_(e){var n=e.parent;return 201===n.kind||203===n.kind&&n.expression===e||202===n.kind&&n.expression===e||198===n.kind&&n.name===e&&!!n.initializer}function S_(e){return 58982400&e.flags&&dh(Ws(e)||De,98304)}function x_(e,n){return e&&T_(n)&&zf(e,S_)?Yf(Im(e),zs):e}function L_(n,t){!Zr(n,111551)||_f(t)||ra(n)||(V.preserveConstEnums&&function(n){return!!e.findAncestor(n,(function(n){return n.parent&&e.isExportAssignment(n.parent)&&n.parent.expression===n&&e.isEntityNameExpression(n)}))}(t)||!GT(ea(n))?ia(n):function(e){var n=Tr(e);n.constEnumReferenced||(n.constEnumReferenced=!0)}(n))}function A_(n){var t=ff(n);if(t===he)return xe;if(t===ie){var r=e.getContainingFunction(n);return j<2&&(209===r.kind?lr(n,e.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression):e.hasSyntacticModifier(r,256)&&lr(n,e.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method)),Sr(r).flags|=8192,lo(t)}n.parent&&e.isPropertyAccessExpression(n.parent)&&n.parent.expression===n||L_(t,n);var a=Na(t),i=2097152&a.flags?ea(a):a;134217728&Wg(i)&&yu(n.parent,i)&&dr(!1,n,e.Diagnostics._0_is_deprecated,n.escapedText);var o=a.valueDeclaration;if(32&a.flags)if(252===o.kind&&e.nodeIsDecorated(o))for(r=e.getContainingClass(n);void 0!==r;){if(r===o&&r.name!==n){Sr(o).flags|=16777216,Sr(n).flags|=33554432;break}r=e.getContainingClass(r)}else if(221===o.kind)for(r=e.getThisContainer(n,!1);297!==r.kind;){if(r.parent===o){163===r.kind&&e.hasSyntacticModifier(r,32)&&(Sr(o).flags|=16777216,Sr(n).flags|=33554432);break}r=e.getThisContainer(r,!1)}!function(n,t){if(j>=2||0==(34&t.flags)||e.isSourceFile(t.valueDeclaration)||287===t.valueDeclaration.parent.kind)return;var r=e.getEnclosingBlockScopeContainer(t.valueDeclaration),a=function(n,t){return!!e.findAncestor(n,(function(n){return n===t?"quit":e.isFunctionLike(n)}))}(n.parent,r),i=r,o=!1;for(;i&&!e.nodeStartsNewLexicalEnvironment(i);){if(e.isIterationStatement(i,!1)){o=!0;break}i=i.parent}if(o){if(a){var s=!0;if(e.isForStatement(r))if((d=e.getAncestor(t.valueDeclaration,250))&&d.parent===r){var c=function(n,t){return e.findAncestor(n,(function(e){return e===t?"quit":e===t.initializer||e===t.condition||e===t.incrementor||e===t.statement}))}(n.parent,r);if(c){var l=Sr(c);l.flags|=131072;var u=l.capturedBlockScopeBindings||(l.capturedBlockScopeBindings=[]);e.pushIfUnique(u,t),c===r.initializer&&(s=!1)}}s&&(Sr(i).flags|=65536)}var d;if(e.isForStatement(r))(d=e.getAncestor(t.valueDeclaration,250))&&d.parent===r&&function(n,t){var r=n;for(;207===r.parent.kind;)r=r.parent;var a=!1;if(e.isAssignmentTarget(r))a=!0;else if(214===r.parent.kind||215===r.parent.kind){var i=r.parent;a=45===i.operator||46===i.operator}if(!a)return!1;return!!e.findAncestor(r,(function(e){return e===t?"quit":e===t.statement}))}(n,r)&&(Sr(t.valueDeclaration).flags|=4194304);Sr(t.valueDeclaration).flags|=524288}a&&(Sr(t.valueDeclaration).flags|=262144)}(n,t);var s=x_(lo(a),n),c=e.getAssignmentTargetKind(n);if(c){if(!(3&a.flags||e.isInJSFile(n)&&512&a.flags))return lr(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_not_a_variable,ri(t)),xe;if(oh(a))return 3&a.flags?lr(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_constant,ri(t)):lr(n,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,ri(t)),xe}var l=2097152&a.flags;if(3&a.flags){if(1===c)return s}else{if(!l)return s;o=e.find(t.declarations,w)}if(!o)return s;for(var u=160===e.getRootDeclaration(o).kind,d=v_(o),p=v_(n),m=p!==d,f=n.parent&&n.parent.parent&&e.isSpreadAssignment(n.parent)&&Mf(n.parent.parent),_=134217728&t.flags;p!==d&&(208===p.kind||209===p.kind||e.isObjectLiteralOrClassExpressionMethod(p))&&(E_(a)||u&&!h_(a));)p=v_(p);var g=u||l||m||f||_||e.isBindingElement(o)||s!==Te&&s!==In&&(!W||0!=(16387&s.flags)||_f(n)||270===n.parent.kind)||225===n.parent.kind||249===o.kind&&o.exclamationToken||8388608&o.flags,y=y_(n,s,g?u?function(e,n){if(hi(n.symbol,2)){var t=W&&160===n.kind&&n.initializer&&32768&um(e)&&!(32768&um(Hh(n.initializer)));return Ti(),t?If(e,524288):e}return co(n.symbol),e}(s,o):s:s===Te||s===In?Ce:fm(s),p,!g);if(c_(n)||s!==Te&&s!==In){if(!g&&!(32768&um(s))&&32768&um(y))return lr(n,e.Diagnostics.Variable_0_is_used_before_being_assigned,ri(t)),s}else if(y===Te||y===In)return X&&(lr(e.getNameOfDeclaration(o),e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined,ri(t),ii(y)),lr(n,e.Diagnostics.Variable_0_implicitly_has_an_1_type,ri(t),ii(y))),Wb(y);return c?Qp(y):y}function D_(e,n){(Sr(e).flags|=2,163===n.kind||166===n.kind)?Sr(n.parent).flags|=4:Sr(n).flags|=4}function C_(n){return e.isSuperCall(n)?n:e.isFunctionLike(n)?void 0:e.forEachChild(n,C_)}function I_(e){return So(Fo(xa(e)))===ke}function N_(n,t,r){var a=t.parent;e.getClassExtendsHeritageElement(a)&&!I_(a)&&n.flowNode&&!g_(n.flowNode,!1)&&lr(n,r)}function R_(n){var t=e.getThisContainer(n,!0),r=!1;switch(166===t.kind&&N_(n,t,e.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class),209===t.kind&&(t=e.getThisContainer(t,!1),r=!0),t.kind){case 256:lr(n,e.Diagnostics.this_cannot_be_referenced_in_a_module_or_namespace_body);break;case 255:lr(n,e.Diagnostics.this_cannot_be_referenced_in_current_location);break;case 166:O_(n,t)&&lr(n,e.Diagnostics.this_cannot_be_referenced_in_constructor_arguments);break;case 163:case 162:!e.hasSyntacticModifier(t,32)||99===V.target&&V.useDefineForClassFields||lr(n,e.Diagnostics.this_cannot_be_referenced_in_a_static_property_initializer);break;case 158:lr(n,e.Diagnostics.this_cannot_be_referenced_in_a_computed_property_name)}r&&j<2&&D_(n,t);var a=k_(n,!0,t);if(Y){var i=lo(re);if(a===i&&r)lr(n,e.Diagnostics.The_containing_arrow_function_captures_the_global_value_of_this);else if(!a){var o=lr(n,e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation);if(!e.isSourceFile(t)){var s=k_(t);s&&s!==i&&e.addRelatedInfo(o,e.createDiagnosticForNode(t,e.Diagnostics.An_outer_value_of_this_is_shadowed_by_this_container))}}}return a||Ee}function k_(n,t,r){void 0===t&&(t=!0),void 0===r&&(r=e.getThisContainer(n,!1));var a=e.isInJSFile(n);if(e.isFunctionLike(r)&&(!U_(n)||e.getThisParameter(r))){var i=to(r)||a&&function(n){var t=e.getJSDocType(n);if(t&&308===t.kind){var r=t;if(r.parameters.length>0&&r.parameters[0].name&&"this"===r.parameters[0].name.escapedText)return sd(r.parameters[0].type)}var a=e.getJSDocThisTag(n);if(a&&a.typeExpression)return sd(a.typeExpression)}(r);if(!i){var o=function(n){if(208===n.kind&&e.isBinaryExpression(n.parent)&&3===e.getAssignmentDeclarationKind(n.parent))return n.parent.left.expression.expression;if(165===n.kind&&200===n.parent.kind&&e.isBinaryExpression(n.parent.parent)&&6===e.getAssignmentDeclarationKind(n.parent.parent))return n.parent.parent.left.expression;if(208===n.kind&&288===n.parent.kind&&200===n.parent.parent.kind&&e.isBinaryExpression(n.parent.parent.parent)&&6===e.getAssignmentDeclarationKind(n.parent.parent.parent))return n.parent.parent.parent.left.expression;if(208===n.kind&&e.isPropertyAssignment(n.parent)&&e.isIdentifier(n.parent.name)&&("value"===n.parent.name.escapedText||"get"===n.parent.name.escapedText||"set"===n.parent.name.escapedText)&&e.isObjectLiteralExpression(n.parent.parent)&&e.isCallExpression(n.parent.parent.parent)&&n.parent.parent.parent.arguments[2]===n.parent.parent&&9===e.getAssignmentDeclarationKind(n.parent.parent.parent))return n.parent.parent.parent.arguments[0].expression;if(e.isMethodDeclaration(n)&&e.isIdentifier(n.name)&&("value"===n.name.escapedText||"get"===n.name.escapedText||"set"===n.name.escapedText)&&e.isObjectLiteralExpression(n.parent)&&e.isCallExpression(n.parent.parent)&&n.parent.parent.arguments[2]===n.parent&&9===e.getAssignmentDeclarationKind(n.parent.parent))return n.parent.parent.arguments[0].expression}(r);if(a&&o){var s=Hh(o).symbol;s&&s.members&&16&s.flags&&(i=Fo(s).thisType)}else fv(r)&&(i=Fo(Sa(r.symbol)).thisType);i||(i=F_(r))}if(i)return y_(n,i)}if(e.isClassLike(r.parent)){var c=xa(r.parent);return y_(n,e.hasSyntacticModifier(r,32)?lo(c):Fo(c).thisType)}if(e.isSourceFile(r)){if(r.commonJsModuleIndicator){var l=xa(r);return l&&lo(l)}if(r.externalModuleIndicator)return Ce;if(t)return lo(re)}}function O_(n,t){return!!e.findAncestor(n,(function(n){return e.isFunctionLikeDeclaration(n)?"quit":160===n.kind&&n.parent===t}))}function P_(n){var t=203===n.parent.kind&&n.parent.expression===n,r=e.getSuperContainer(n,!0),a=r,i=!1;if(!t)for(;a&&209===a.kind;)a=e.getSuperContainer(a,!0),i=j<2;var o=0;if(!function(n){if(!n)return!1;if(t)return 166===n.kind;if(e.isClassLike(n.parent)||200===n.parent.kind)return e.hasSyntacticModifier(n,32)?165===n.kind||164===n.kind||167===n.kind||168===n.kind:165===n.kind||164===n.kind||167===n.kind||168===n.kind||163===n.kind||162===n.kind||166===n.kind;return!1}(a)){var s=e.findAncestor(n,(function(e){return e===a?"quit":158===e.kind}));return s&&158===s.kind?lr(n,e.Diagnostics.super_cannot_be_referenced_in_a_computed_property_name):t?lr(n,e.Diagnostics.Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors):a&&a.parent&&(e.isClassLike(a.parent)||200===a.parent.kind)?lr(n,e.Diagnostics.super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class):lr(n,e.Diagnostics.super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions),xe}if(t||166!==r.kind||N_(n,a,e.Diagnostics.super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class),o=e.hasSyntacticModifier(a,32)||t?512:256,Sr(n).flags|=o,165===a.kind&&e.hasSyntacticModifier(a,256)&&(e.isSuperProperty(n.parent)&&e.isAssignmentTarget(n.parent)?Sr(a).flags|=4096:Sr(a).flags|=2048),i&&D_(n.parent,a),200===a.parent.kind)return j<2?(lr(n,e.Diagnostics.super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher),xe):Ee;var c=a.parent;if(!e.getClassExtendsHeritageElement(c))return lr(n,e.Diagnostics.super_can_only_be_referenced_in_a_derived_class),xe;var l=Fo(xa(c)),u=l&&Lo(l)[0];return u?166===a.kind&&O_(n,a)?(lr(n,e.Diagnostics.super_cannot_be_referenced_in_constructor_arguments),xe):512===o?So(l):rs(u,l.thisType):xe}function M_(n){return 4&e.getObjectFlags(n)&&n.target===Dn?nl(n)[0]:void 0}function w_(n){return Yf(n,(function(n){return 2097152&n.flags?e.forEach(n.types,M_):M_(n)}))}function F_(n){if(209!==n.kind){if(Bd(n)){var t=dg(n);if(t){var r=t.thisParameter;if(r)return lo(r)}}var a=e.isInJSFile(n);if(Y||a){var i=function(e){return 165!==e.kind&&167!==e.kind&&168!==e.kind||200!==e.parent.kind?208===e.kind&&288===e.parent.kind?e.parent.parent:void 0:e.parent}(n);if(i){for(var o=ng(i),s=i,c=o;c;){var l=w_(c);if(l)return kd(l,jm(ig(i)));if(288!==s.parent.kind)break;c=ng(s=s.parent.parent)}return Im(o?_m(o):Lh(i))}var u=e.walkUpParenthesizedExpressions(n.parent);if(216===u.kind&&62===u.operatorToken.kind){var d=u.left;if(e.isAccessExpression(d)){var p=d.expression;if(a&&e.isIdentifier(p)){var m=e.getSourceFileOfNode(u);if(m.commonJsModuleIndicator&&ff(p)===m.symbol)return}return Im(Lh(p))}}}}}function G_(n){var t=n.parent;if(Bd(t)){var r=e.getImmediatelyInvokedFunctionExpression(t);if(r&&r.arguments){var a=zy(r),i=t.parameters.indexOf(n);if(n.dotDotDotToken)return Uy(a,i,a.length,Ee,void 0,0);var o=Sr(r),s=o.resolvedSignature;o.resolvedSignature=ct;var c=i<a.length?Zp(Hh(a[i])):n.initializer?void 0:Ie;return o.resolvedSignature=s,c}var l=dg(t);if(l){var u=t.parameters.indexOf(n)-(e.getThisParameter(t)?1:0);return n.dotDotDotToken&&e.lastOrUndefined(t.parameters)===n?Mv(l,u):Pv(l,u)}}}function B_(n){var t=e.getEffectiveTypeAnnotationNode(n);if(t)return sd(t);switch(n.kind){case 160:return G_(n);case 198:return function(n){var t=n.parent.parent,r=n.propertyName||n.name,a=B_(t)||198!==t.kind&&t.initializer&&Dh(t);if(!a||e.isBindingPattern(r)||e.isComputedNonLiteralName(r))return;if(197===t.name.kind){var i=e.indexOfNode(n.parent.elements,n);if(i<0)return;return Q_(a,i)}var o=ou(r);if(qo(o)){return xi(a,Zo(o))}}(n);case 163:if(e.hasSyntacticModifier(n,32))return function(n){var t=e.isExpression(n.parent)&&ag(n.parent);return t?J_(t,xa(n).escapedName):void 0}(n)}}function U_(n){for(var t=!1;n.parent&&!e.isFunctionLike(n.parent);){if(e.isParameter(n.parent)&&(t||n.parent.initializer===n))return!0;e.isBindingElement(n.parent)&&n.parent.initializer===n&&(t=!0),n=n.parent}return!1}function V_(n,t){var r=!!(2&e.getFunctionFlags(t)),a=j_(t);if(a)return DE(n,a,r)||void 0}function j_(e){var n=Oc(e);if(n)return n;var t=ug(e);return t&&!Pc(t)?kc(t):void 0}function H_(e,n){var t=zy(e).indexOf(n);return-1===t?void 0:K_(e,t)}function K_(n,t){var r=Sr(n).resolvedSignature===ut?ut:mv(n);return e.isJsxOpeningLikeElement(n)&&0===t?og(r,n):Ov(r,t)}function W_(n,t){var r=n.parent,a=r.left,i=r.operatorToken,o=r.right;switch(i.kind){case 62:case 75:case 74:case 76:return n===o?function(n){var t=e.getAssignmentDeclarationKind(n);switch(t){case 0:return Uh(n.left);case 5:case 1:case 6:case 3:if(z_(n,t))return q_(n,t);if(n.left.symbol){var r=n.left.symbol.valueDeclaration;if(!r)return;var a=e.cast(n.left,e.isAccessExpression),i=e.getEffectiveTypeAnnotationNode(r);if(i)return sd(i);if(e.isIdentifier(a.expression)){var o=a.expression,s=Cr(o,o.escapedText,111551,void 0,o.escapedText,!0);if(s){var c=s.valueDeclaration&&e.getEffectiveTypeAnnotationNode(s.valueDeclaration);if(c){var l=e.getElementOrPropertyAccessName(a);if(void 0!==l)return J_(sd(c),l)}return}}return e.isInJSFile(r)?void 0:Uh(n.left)}return Uh(n.left);case 2:case 4:return q_(n,t);case 7:case 8:case 9:return e.Debug.fail("Does not apply");default:return e.Debug.assertNever(t)}}(r):void 0;case 56:case 60:var s=ag(r,t);return n===o&&(s&&s.pattern||!s&&!e.isDefaultedExpandoInitializer(r))?Uh(a):s;case 55:case 27:return n===o?ag(r,t):void 0;default:return}}function z_(n,t){if(void 0===t&&(t=e.getAssignmentDeclarationKind(n)),4===t)return!0;if(!e.isInJSFile(n)||5!==t||!e.isIdentifier(n.left.expression))return!1;var r=n.left.expression.escapedText,a=Cr(n.left,r,111551,void 0,void 0,!0,!0);return e.isThisInitializedDeclaration(null==a?void 0:a.valueDeclaration)}function q_(n,t){if(!n.symbol)return Uh(n.left);if(n.symbol.valueDeclaration){var r=e.getEffectiveTypeAnnotationNode(n.symbol.valueDeclaration);if(r){var a=sd(r);if(a)return a}}if(2!==t){var i=e.cast(n.left,e.isAccessExpression);if(e.isObjectLiteralMethod(e.getThisContainer(i.expression,!1))){var o=R_(i.expression),s=e.getElementOrPropertyAccessName(i);return void 0!==s&&J_(o,s)||void 0}}}function J_(n,t){return Yf(n,(function(n){if(Os(n)){var r=xs(n),a=Ws(r)||r,i=td(e.unescapeLeadingUnderscores(t));if(zd(i,a))return Cu(n,i)}else if(3670016&n.flags){var o=cc(n,t);if(o)return c=o,262144&e.getCheckFlags(c)&&!c.type&&bi(c,0)>=0?void 0:lo(o);if(tm(n)){var s=im(n);if(s&&yg(t)&&+t>=0)return s}return yg(t)&&X_(n,1)||X_(n,0)}var c}),!0)}function X_(e,n){return Yf(e,(function(e){return pc(e,n)}),!0)}function Y_(e,n){var t=ng(e.parent,n);if(t){if(!Qo(e)){var r=J_(t,xa(e).escapedName);if(r)return r}return _g(e.name)&&X_(t,1)||X_(t,0)}}function Q_(e,n){return e&&(J_(e,""+n)||Yf(e,(function(e){return iE(1,e,Ce,void 0,!1)}),!0))}function Z_(n){var t=n.parent;return e.isJsxAttributeLike(t)?ag(n):e.isJsxElement(t)?function(n,t){var r=ng(n.openingElement.tagName),a=Pg(kg(n));if(r&&!Li(r)&&a&&""!==a){var i=e.getSemanticJsxChildren(n.children),o=i.indexOf(t),s=J_(r,a);return s&&(1===i.length?s:Yf(s,(function(e){return Kp(e)?Iu(e,td(o)):e}),!0))}}(t,n):void 0}function $_(n){if(e.isJsxAttribute(n)){var t=ng(n.parent);if(!t||Li(t))return;return J_(t,n.name.escapedText)}return ag(n.parent)}function eg(e){switch(e.kind){case 10:case 8:case 9:case 14:case 109:case 94:case 103:case 78:case 150:return!0;case 201:case 207:return eg(e.expression);case 283:return!e.expression||eg(e.expression)}return!1}function ng(n,t){var r=tg(e.isObjectLiteralMethod(n)?function(n,t){if(e.Debug.assert(e.isObjectLiteralMethod(n)),!(16777216&n.flags))return Y_(n,t)}(n,t):ag(n,t),n,t);if(r&&!(t&&2&t&&8650752&r.flags)){var a=Yf(r,Zs,!0);if(1048576&a.flags){if(e.isObjectLiteralExpression(n))return function(n,t){return Ep(t,e.map(e.filter(n.properties,(function(e){return!!e.symbol&&288===e.kind&&eg(e.initializer)&&Tf(t,e.symbol.escapedName)})),(function(e){return[function(){return Hh(e.initializer)},e.symbol.escapedName]})),zd,t)}(n,a);if(e.isJsxAttributes(n))return function(n,t){return Ep(t,e.map(e.filter(n.properties,(function(e){return!!e.symbol&&280===e.kind&&Tf(t,e.symbol.escapedName)&&(!e.initializer||eg(e.initializer))})),(function(e){return[e.initializer?function(){return Hh(e.initializer)}:function(){return Ge},e.symbol.escapedName]})),zd,t)}(n,a)}return a}}function tg(n,t,r){if(n&&dh(n,465829888)){var a=ig(t);if(a&&e.some(a.inferences,wh)){if(r&&1&r)return rg(n,a.nonFixingMapper);if(a.returnMapper)return rg(n,a.returnMapper)}}return n}function rg(n,t){return 465829888&n.flags?kd(n,t):1048576&n.flags?Xl(e.map(n.types,(function(e){return rg(e,t)})),0):2097152&n.flags?tu(e.map(n.types,(function(e){return rg(e,t)}))):n}function ag(n,t){if(16777216&n.flags);else{if(n.contextualType)return n.contextualType;var r=n.parent;switch(r.kind){case 249:case 160:case 163:case 162:case 198:return function(n,t){var r=n.parent;if(e.hasInitializer(r)&&n===r.initializer){var a=B_(r);if(a)return a;if(!(8&t)&&e.isBindingPattern(r.name))return Ji(r.name,!0,!1)}}(n,t);case 209:case 242:return function(n){var t=e.getContainingFunction(n);if(t){var r=j_(t);if(r){var a=e.getFunctionFlags(t);if(1&a){var i=dE(r,2&a?2:1,void 0);if(!i)return;r=i.returnType}if(2&a){var o=Yf(r,mb);return o&&Xl([o,qv(o)])}return r}}}(n);case 219:return function(n){var t=e.getContainingFunction(n);if(t){var r=e.getFunctionFlags(t),a=j_(t);if(a)return n.asteriskToken?a:DE(0,a,0!=(2&r))}}(r);case 213:return function(e,n){var t=ag(e,n);if(t){var r=mb(t);return r&&Xl([r,qv(r)])}}(r,t);case 203:if(99===r.expression.kind)return Oe;case 204:return H_(r,n);case 206:case 224:return e.isConstTypeReference(r.type)?function(n){if(e.isCallLikeExpression(n.parent))return H_(n.parent,n);return}(r):sd(r.type);case 216:return W_(n,t);case 288:case 289:return Y_(r,t);case 290:return ng(r.parent,t);case 199:var a=r;return Q_(ng(a,t),e.indexOfNode(a.elements,n));case 217:return function(e,n){var t=e.parent;return e===t.whenTrue||e===t.whenFalse?ag(t,n):void 0}(n,t);case 228:return e.Debug.assert(218===r.parent.kind),function(e,n){if(205===e.parent.kind)return H_(e.parent,n)}(r.parent,n);case 207:var i=e.isInJSFile(r)?e.getJSDocTypeTag(r):void 0;return i?sd(i.typeExpression.type):ag(r,t);case 283:return Z_(r);case 280:case 282:return $_(r);case 275:case 274:return function(n,t){if(e.isJsxOpeningElement(n)&&n.parent.contextualType&&4!==t)return n.parent.contextualType;return K_(n,0)}(r,t)}}}function ig(n){var t=e.findAncestor(n,(function(e){return!!e.inferenceContext}));return t&&t.inferenceContext}function og(n,t){return 0!==jy(t)?function(e,n){var t=jv(e,De);t=sg(n,kg(n),t);var r=Ig(D.IntrinsicAttributes,n);r!==xe&&(t=fs(r,t));return t}(n,t):function(n,t){var r=kg(t),a=(o=r,Og(D.ElementAttributesPropertyNameContainer,o)),i=void 0===a?jv(n,De):""===a?kc(n):function(e,n){if(e.unionSignatures){for(var t=[],r=0,a=e.unionSignatures;r<a.length;r++){var i=kc(a[r]);if(Li(i))return i;var o=xi(i,n);if(!o)return;t.push(o)}return tu(t)}var s=kc(e);return Li(s)?s:xi(s,n)}(n,a);var o;if(!i)return a&&e.length(t.attributes.properties)&&lr(t,e.Diagnostics.JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property,e.unescapeLeadingUnderscores(a)),De;if(Li(i=sg(t,r,i)))return i;var s=i,c=Ig(D.IntrinsicClassAttributes,t);if(c!==xe){var l=yo(c.symbol),u=kc(n);s=fs(l?Zc(c,xc([u],l,Sc(l),e.isInJSFile(t))):c,s)}var d=Ig(D.IntrinsicAttributes,t);return d!==xe&&(s=fs(d,s)),s}(n,t)}function sg(n,t,r){var a,i=(a=t)&&Lr(a.exports,D.LibraryManagedAttributes,788968);if(i){var o=Fo(i),s=function(e){if(xg(e.tagName))return jc(uv(e,n=Fg(e)));var n,t=Lh(e.tagName);return 128&t.flags?(n=wg(t,e))?jc(uv(e,n)):xe:t}(n);if(e.length(o.typeParameters)>=2)return Zc(o,c=xc([s,r],o.typeParameters,2,e.isInJSFile(n)));if(e.length(o.aliasTypeArguments)>=2){var c=xc([s,r],o.aliasTypeArguments,2,e.isInJSFile(n));return al(o.aliasSymbol,c)}}return r}function cg(n,t){var r=uc(n,0);if(1===r.length){var a=r[0];if(!function(n,t){for(var r=0;r<t.parameters.length;r++){var a=t.parameters[r];if(a.initializer||a.questionToken||a.dotDotDotToken||vc(a))break}t.parameters.length&&e.parameterIsThisKeyword(t.parameters[0])&&r--;return!Gv(n)&&wv(n)<r}(a,t))return a}}function lg(e){return 208===e.kind||209===e.kind}function ug(n){return lg(n)||e.isObjectLiteralMethod(n)?dg(n):void 0}function dg(n){e.Debug.assert(165!==n.kind||e.isObjectLiteralMethod(n));var t=Ac(n);if(t)return t;var r=ng(n,1);if(r){if(!(1048576&r.flags))return cg(r,n);for(var a,i=0,o=r.types;i<o.length;i++){var s=cg(o[i],n);if(s)if(a){if(!Gp(a[0],s,!1,!0,!0,jd))return;a.push(s)}else a=[s]}return a?1===a.length?a[0]:ss(a[0],a):void 0}}function pg(e){return 198===e.kind&&!!e.initializer||216===e.kind&&62===e.operatorToken.kind}function mg(n,t,r){for(var a=n.elements,i=a.length,o=[],s=[],c=ng(n),l=e.isAssignmentTarget(n),u=Nh(n),d=0;d<i;d++){var p=a[d];if(220===p.kind){j<2&&cS(p,V.downlevelIteration?1536:2048);var m=Hh(p.expression,t,r);if(Kp(m))o.push(m),s.push(8);else if(l){var f=fc(m,1)||iE(65,m,Ce,void 0,!1)||De;o.push(f),s.push(4)}else o.push(aE(33,m,Ce,p.expression)),s.push(4)}else{var _=Rh(p,t,Q_(c,o.length),r);o.push(_),s.push(1)}}return l?Fl(o,s):r||u||c&&zf(c,zp)?fg(Fl(o,s,u)):fg(Nl(o.length?Xl(e.sameMap(o,(function(e,n){return 8&s[n]?Ru(e,Pe)||Ee:e})),2):W?ze:Ie,u))}function fg(n){if(!(4&e.getObjectFlags(n)))return n;var t=n.literalType;return t||((t=n.literalType=$c(n)).objectFlags|=1114112),t}function _g(e){switch(e.kind){case 158:return function(e){return ph(vg(e),296)}(e);case 78:return yg(e.escapedText);case 8:case 10:return yg(e.text);default:return!1}}function gg(e){return"Infinity"===e||"-Infinity"===e||"NaN"===e}function yg(e){return(+e).toString()===e}function vg(n){var t=Sr(n.expression);return t.resolvedType||(t.resolvedType=Hh(n.expression),98304&t.resolvedType.flags||!ph(t.resolvedType,402665900)&&!zd(t.resolvedType,Xe)?lr(n,e.Diagnostics.A_computed_property_name_must_be_of_type_string_number_symbol_or_any):xy(n.expression,t.resolvedType,!0)),t.resolvedType}function hg(n,t,r,a){for(var i,o,s,c=[],l=t;l<r.length;l++)(0===a||(i=r[l],o=void 0,s=void 0,s=null===(o=i.declarations)||void 0===o?void 0:o[0],yg(i.escapedName)||s&&e.isNamedDeclaration(s)&&_g(s.name)))&&c.push(lo(r[l]));return Wc(c.length?Xl(c,2):Ce,Nh(n))}function bg(n){e.Debug.assert(0!=(2097152&n.flags),"Should only get Alias here.");var t=Tr(n);if(!t.immediateTarget){var r=Br(n);if(!r)return e.Debug.fail();t.immediateTarget=Qr(r,!0)}return t.immediateTarget}function Eg(n,t){var r=e.isAssignmentTarget(n);!function(n,t){for(var r=new e.Map,a=0,i=n.properties;a<i.length;a++){var o=i[a];if(290!==o.kind){var s=o.name;if(158===s.kind&&bS(s),289===o.kind&&!t&&o.objectAssignmentInitializer)return FS(o.equalsToken,e.Diagnostics.Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern);if(79===s.kind)return FS(s,e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);if(o.modifiers)for(var c=0,l=o.modifiers;c<l.length;c++){var u=l[c];129===u.kind&&165===o.kind||FS(u,e.Diagnostics._0_modifier_cannot_be_used_here,e.getTextOfNode(u))}var d=void 0;switch(o.kind){case 289:SS(o.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context);case 288:TS(o.questionToken,e.Diagnostics.An_object_member_cannot_be_declared_optional),8===s.kind&&US(s),d=4;break;case 165:d=8;break;case 167:d=1;break;case 168:d=2;break;default:throw e.Debug.assertNever(o,"Unexpected syntax kind:"+o.kind)}if(!t){var p=e.getPropertyNameForPropertyNameNode(s);if(void 0===p)continue;var m=r.get(p);if(m)if(12&d&&12&m)FS(s,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(s));else{if(!(3&d&&3&m))return FS(s,e.Diagnostics.An_object_literal_cannot_have_property_and_accessor_with_the_same_name);if(3===m||d===m)return FS(s,e.Diagnostics.An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name);r.set(p,d|m)}else r.set(p,d)}}else if(t){var f=e.skipParentheses(o.expression);if(e.isArrayLiteralExpression(f)||e.isObjectLiteralExpression(f))return FS(o.expression,e.Diagnostics.A_rest_element_cannot_contain_a_binding_pattern)}}}(n,r);for(var a=W?e.createSymbolTable():void 0,i=e.createSymbolTable(),o=[],s=nn,c=ng(n),l=c&&c.pattern&&(196===c.pattern.kind||200===c.pattern.kind),u=Nh(n),d=u?8:0,p=e.isInJSFile(n)&&!e.isInJsonFile(n),m=e.getJSDocEnumTag(n),f=!c&&p&&!m,_=Z,g=!1,y=!1,v=!1,h=0,b=n.properties;h<b.length;h++){var E=b[h];E.name&&e.isComputedPropertyName(E.name)&&!e.isWellKnownSymbolSyntactically(E.name)&&vg(E.name)}for(var T=0,S=0,x=n.properties;S<x.length;S++){var L=x[S],A=xa(L),D=L.name&&158===L.name.kind&&!e.isWellKnownSymbolSyntactically(L.name.expression)?vg(L.name):void 0;if(288===L.kind||289===L.kind||e.isObjectLiteralMethod(L)){var C=288===L.kind?kh(L,t):289===L.kind?Rh(!r&&L.objectAssignmentInitializer?L.objectAssignmentInitializer:L.name,t):Oh(L,t);if(p){var I=ki(L);I?(Yd(C,I,L),C=I):m&&m.typeExpression&&Yd(C,sd(m.typeExpression),L)}_|=3670016&e.getObjectFlags(C);var N=D&&qo(D)?D:void 0,R=N?mr(4|A.flags,Zo(N),4096|d):mr(4|A.flags,A.escapedName,d);if(N&&(R.nameType=N),r)(288===L.kind&&pg(L.initializer)||289===L.kind&&L.objectAssignmentInitializer)&&(R.flags|=16777216);else if(l&&!(512&e.getObjectFlags(c))){var k=cc(c,A.escapedName);k?R.flags|=16777216&k.flags:V.suppressExcessPropertyErrors||mc(c,0)||lr(L.name,e.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1,ri(A),ii(c))}R.declarations=A.declarations,R.parent=A.parent,A.valueDeclaration&&(R.valueDeclaration=A.valueDeclaration),R.type=C,R.target=A,A=R,null==a||a.set(R.escapedName,R)}else{if(290===L.kind){if(j<2&&cS(L,2),o.length>0&&(s=Ju(s,M(),n.symbol,_,u),o=[],i=e.createSymbolTable(),y=!1,v=!1),!Tg(C=rc(Hh(L.expression))))return lr(L,e.Diagnostics.Spread_types_may_only_be_created_from_object_types),xe;a&&Dg(C,a,L),s=Ju(s,C,n.symbol,_,u),T=o.length;continue}e.Debug.assert(167===L.kind||168===L.kind),sT(L)}!D||8576&D.flags?i.set(A.escapedName,A):zd(D,Xe)&&(zd(D,Pe)?v=!0:y=!0,r&&(g=!0)),o.push(A)}if(l&&290!==n.parent.kind)for(var O=0,P=Gs(c);O<P.length;O++){R=P[O];i.get(R.escapedName)||cc(s,R.escapedName)||(16777216&R.flags||lr(R.valueDeclaration||R.bindingElement,e.Diagnostics.Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value),i.set(R.escapedName,R),o.push(R))}return s!==nn?(o.length>0&&(s=Ju(s,M(),n.symbol,_,u),o=[],i=e.createSymbolTable(),y=!1,v=!1),Yf(s,(function(e){return e===nn?M():e}))):M();function M(){var t=y?hg(n,T,o,0):void 0,a=v?hg(n,T,o,1):void 0,s=Va(n.symbol,i,e.emptyArray,e.emptyArray,t,a);return s.objectFlags|=1048704|_,f&&(s.objectFlags|=16384),g&&(s.objectFlags|=512),r&&(s.pattern=n),s}}function Tg(n){if(465829888&n.flags){var t=Ws(n);if(void 0!==t)return Tg(t)}return!!(126615553&n.flags||117632&um(n)&&Tg(dm(n))||3145728&n.flags&&e.every(n.types,Tg))}function Sg(n){return!e.stringContains(n,"-")}function xg(n){return 78===n.kind&&e.isIntrinsicJsxName(n.escapedText)}function Lg(e,n){return e.initializer?Rh(e.initializer,n):Ge}function Ag(e,n){for(var t=[],r=0,a=e.children;r<a.length;r++){var i=a[r];if(11===i.kind)i.containsOnlyTriviaWhiteSpaces||t.push(Oe);else{if(283===i.kind&&!i.expression)continue;t.push(Rh(i,n))}}return t}function Dg(n,t,r){for(var a=0,i=Gs(n);a<i.length;a++){var o=i[a],s=t.get(o.escapedName),c=lo(o);if(s&&!dh(c,98304)&&!(dh(c,3)&&16777216&o.flags)){var l=lr(s.valueDeclaration,e.Diagnostics._0_is_specified_more_than_once_so_this_usage_will_be_overwritten,e.unescapeLeadingUnderscores(s.escapedName));e.addRelatedInfo(l,e.createDiagnosticForNode(r,e.Diagnostics.This_spread_always_overwrites_this_property))}}}function Cg(n,t){return function(n,t){for(var r,a=n.attributes,i=W?e.createSymbolTable():void 0,o=e.createSymbolTable(),s=tn,c=!1,l=!1,u=4096,d=Pg(kg(n)),p=0,m=a.properties;p<m.length;p++){var f=m[p],_=f.symbol;if(e.isJsxAttribute(f)){var g=Lg(f,t);u|=3670016&e.getObjectFlags(g);var y=mr(4|_.flags,_.escapedName);y.declarations=_.declarations,y.parent=_.parent,_.valueDeclaration&&(y.valueDeclaration=_.valueDeclaration),y.type=g,y.target=_,o.set(y.escapedName,y),null==i||i.set(y.escapedName,y),f.name.escapedText===d&&(l=!0)}else e.Debug.assert(282===f.kind),o.size>0&&(s=Ju(s,x(),a.symbol,u,!1),o=e.createSymbolTable()),Li(g=rc(Lh(f.expression,t)))&&(c=!0),Tg(g)?(s=Ju(s,g,a.symbol,u,!1),i&&Dg(g,i,f)):r=r?tu([r,g]):g}c||o.size>0&&(s=Ju(s,x(),a.symbol,u,!1));var v=273===n.parent.kind?n.parent:void 0;if(v&&v.openingElement===n&&v.children.length>0){var h=Ag(v,t);if(!c&&d&&""!==d){l&&lr(a,e.Diagnostics._0_are_specified_twice_The_attribute_named_0_will_be_overwritten,e.unescapeLeadingUnderscores(d));var b=ng(n.attributes),E=b&&J_(b,d),T=mr(4,d);T.type=1===h.length?h[0]:E&&zf(E,zp)?Fl(h):Nl(Xl(h)),T.valueDeclaration=e.factory.createPropertySignature(void 0,e.unescapeLeadingUnderscores(d),void 0,void 0),e.setParent(T.valueDeclaration,a),T.valueDeclaration.symbol=T;var S=e.createSymbolTable();S.set(d,T),s=Ju(s,Va(a.symbol,S,e.emptyArray,e.emptyArray,void 0,void 0),a.symbol,u,!1)}}return c?Ee:r&&s!==tn?tu([r,s]):r||(s===tn?x():s);function x(){u|=Z;var n=Va(a.symbol,o,e.emptyArray,e.emptyArray,void 0,void 0);return n.objectFlags|=1048704|u,n}}(n.parent,t)}function Ig(e,n){var t=kg(n),r=t&&ha(t),a=r&&Lr(r,e,788968);return a?Fo(a):xe}function Ng(n){var t=Sr(n);if(!t.resolvedSymbol){var r=Ig(D.IntrinsicElements,n);if(r!==xe){if(!e.isIdentifier(n.tagName))return e.Debug.fail();var a=cc(r,n.tagName.escapedText);return a?(t.jsxFlags|=1,t.resolvedSymbol=a):fc(r,0)?(t.jsxFlags|=2,t.resolvedSymbol=r.symbol):(lr(n,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.idText(n.tagName),"JSX."+D.IntrinsicElements),t.resolvedSymbol=he)}return X&&lr(n,e.Diagnostics.JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists,e.unescapeLeadingUnderscores(D.IntrinsicElements)),t.resolvedSymbol=he}return t.resolvedSymbol}function Rg(n){var t=n&&e.getSourceFileOfNode(n),r=t&&Sr(t);if(!r||!1!==r.jsxImplicitImportContainer){if(r&&r.jsxImplicitImportContainer)return r.jsxImplicitImportContainer;var a=e.getJSXRuntimeImport(e.getJSXImplicitImportBase(V,t),V);if(a){var i=pa(n,a,e.getEmitModuleResolutionKind(V)===e.ModuleResolutionKind.Classic?e.Diagnostics.Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option:e.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations,n),o=i&&i!==he?Sa($r(i)):void 0;return r&&(r.jsxImplicitImportContainer=o||!1),o}}}function kg(e){var n=e&&Sr(e);if(n&&n.jsxNamespace)return n.jsxNamespace;if(!n||!1!==n.jsxNamespace){var t=Rg(e);if(!t||t===he){var r=sr(e);t=Cr(e,r,1920,void 0,r,!1)}if(t){var a=$r(Lr(ha($r(t)),D.JSX,1920));if(a&&a!==he)return n&&(n.jsxNamespace=a),a}n&&(n.jsxNamespace=!1)}var i=$r(bl(D.JSX,1920,void 0));return i!==he?i:void 0}function Og(n,t){var r=t&&Lr(t.exports,n,788968),a=r&&Fo(r),i=a&&Gs(a);if(i){if(0===i.length)return"";if(1===i.length)return i[0].escapedName;i.length>1&&lr(r.declarations[0],e.Diagnostics.The_global_type_JSX_0_may_not_have_more_than_one_property,e.unescapeLeadingUnderscores(n))}}function Pg(e){return Og(D.ElementChildrenAttributeNameContainer,e)}function Mg(n,t){if(4&n.flags)return[ct];if(128&n.flags){var r=wg(n,t);return r?[uv(t,r)]:(lr(t,e.Diagnostics.Property_0_does_not_exist_on_type_1,n.value,"JSX."+D.IntrinsicElements),e.emptyArray)}var a=Zs(n),i=uc(a,1);return 0===i.length&&(i=uc(a,0)),0===i.length&&1048576&a.flags&&(i=ps(e.map(a.types,(function(e){return Mg(e,t)})))),i}function wg(n,t){var r=Ig(D.IntrinsicElements,t);if(r!==xe){var a=n.value,i=cc(r,e.escapeLeadingUnderscores(a));if(i)return lo(i);var o=fc(r,0);return o||void 0}return Ee}function Fg(n){e.Debug.assert(xg(n.tagName));var t=Sr(n);if(!t.resolvedJsxElementAttributesType){var r=Ng(n);return 1&t.jsxFlags?t.resolvedJsxElementAttributesType=lo(r):2&t.jsxFlags?t.resolvedJsxElementAttributesType=fc(Fo(r),0):t.resolvedJsxElementAttributesType=xe}return t.resolvedJsxElementAttributesType}function Gg(e){var n=Ig(D.ElementClass,e);if(n!==xe)return n}function Bg(e){return Ig(D.Element,e)}function Ug(e){var n=Bg(e);if(n)return Xl([n,Re])}function Vg(n){var t,r=e.isJsxOpeningLikeElement(n);if(r&&function(n){gS(n,n.typeArguments);for(var t=new e.Map,r=0,a=n.attributes.properties;r<a.length;r++){var i=a[r];if(282!==i.kind){var o=i.name,s=i.initializer;if(t.get(o.escapedText))return FS(o,e.Diagnostics.JSX_elements_cannot_have_multiple_attributes_with_the_same_name);if(t.set(o.escapedText,!0),s&&283===s.kind&&!s.expression)return FS(s,e.Diagnostics.JSX_attributes_must_only_be_assigned_a_non_empty_expression)}}}(n),t=n,0===(V.jsx||0)&&lr(t,e.Diagnostics.Cannot_use_JSX_unless_the_jsx_flag_is_provided),void 0===Bg(t)&&X&&lr(t,e.Diagnostics.JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist),!Rg(n)){var a=Yt&&2===V.jsx?e.Diagnostics.Cannot_find_name_0:void 0,i=sr(n),o=r?n.tagName:n,s=void 0;e.isJsxOpeningFragment(n)&&"null"===i||(s=Cr(o,i,111551,a,i,!0)),s&&(s.isReferenced=67108863,2097152&s.flags&&!ra(s)&&ia(s))}if(r){var c=n,l=mv(c);yv(l,n),function(n,t,r){if(1===n)(a=Ug(r))&&vp(t,a,tr,r.tagName,e.Diagnostics.Its_return_type_0_is_not_a_valid_JSX_element,o);else if(0===n)(i=Gg(r))&&vp(t,i,tr,r.tagName,e.Diagnostics.Its_instance_type_0_is_not_a_valid_JSX_element,o);else{var a=Ug(r),i=Gg(r);if(!a||!i)return;vp(t,Xl([a,i]),tr,r.tagName,e.Diagnostics.Its_element_type_0_is_not_a_valid_JSX_element,o)}function o(){var n=e.getTextOfNode(r.tagName);return e.chainDiagnosticMessages(void 0,e.Diagnostics._0_cannot_be_used_as_a_JSX_component,n)}}(jy(c),kc(l),c)}}function jg(e,n,t){if(524288&e.flags){var r=Ps(e);if(r.stringIndexInfo||r.numberIndexInfo&&yg(n)||ws(e,n)||t&&!Sg(n))return!0}else if(3145728&e.flags&&Hg(e))for(var a=0,i=e.types;a<i.length;a++){if(jg(i[a],n,t))return!0}return!1}function Hg(n){return!!(524288&n.flags&&!(512&e.getObjectFlags(n))||67108864&n.flags||1048576&n.flags&&e.some(n.types,Hg)||2097152&n.flags&&e.every(n.types,Hg))}function Kg(n,t){if(function(n){if(n.expression&&e.isCommaSequence(n.expression))FS(n.expression,e.Diagnostics.JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array)}(n),n.expression){var r=Hh(n.expression,t);return n.dotDotDotToken&&r!==Ee&&!Up(r)&&lr(n,e.Diagnostics.JSX_spread_child_must_be_an_array_type),r}return xe}function Wg(n){return n.valueDeclaration?e.getCombinedNodeFlags(n.valueDeclaration):0}function zg(n){if(8192&n.flags||4&e.getCheckFlags(n))return!0;if(e.isInJSFile(n.valueDeclaration)){var t=n.valueDeclaration.parent;return t&&e.isBinaryExpression(t)&&3===e.getAssignmentDeclarationKind(t)}}function qg(n,t,r,a){var i,o=e.getDeclarationModifierFlagsFromSymbol(a),s=157===n.kind?n.right:195===n.kind?n:n.name;if(t){if(j<2&&Jg(a))return lr(s,e.Diagnostics.Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword),!1;if(128&o)return lr(s,e.Diagnostics.Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression,ri(a),ii(kp(a))),!1}if(128&o&&e.isThisProperty(n)&&Jg(a)&&((i=e.getClassLikeDeclarationOfSymbol(La(a)))&&function(n){return!!e.findAncestor(n,(function(n){return!!(e.isConstructorDeclaration(n)&&e.nodeIsPresent(n.body)||e.isPropertyDeclaration(n))||!(!e.isClassLike(n)&&!e.isFunctionLikeDeclaration(n))&&"quit"}))}(n)))return lr(s,e.Diagnostics.Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor,ri(a),e.getTextOfIdentifierOrLiteral(i.name)),!1;if(e.isPropertyAccessExpression(n)&&e.isPrivateIdentifier(n.name))return!!e.getContainingClass(n)||(lr(s,e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies),!1);if(!(24&o))return!0;if(8&o)return!!yT(n,i=e.getClassLikeDeclarationOfSymbol(La(a)))||(lr(s,e.Diagnostics.Property_0_is_private_and_only_accessible_within_class_1,ri(a),ii(kp(a))),!1);if(t)return!0;var c=gT(n,(function(n){var t=Fo(xa(n));return function(n,t){return Rp(t,(function(t){return!!(16&e.getDeclarationModifierFlagsFromSymbol(t))&&!mo(n,kp(t))}))?void 0:n}(t,a)?t:void 0}));if(!c){var l=void 0;if(32&o||!(l=function(n){var t=e.getThisContainer(n,!1);return t&&e.isFunctionLike(t)?e.getThisParameter(t):void 0}(n))||!l.type)return lr(s,e.Diagnostics.Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses,ri(a),ii(kp(a)||r)),!1;var u=sd(l.type);c=(262144&u.flags?Us(u):u).target}return!!(32&o)||(262144&r.flags&&(r=r.isThisType?Us(r):Ws(r)),!(!r||!mo(r,c))||(lr(s,e.Diagnostics.Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1,ri(a),ii(c)),!1))}function Jg(e){return!!Rp(e,(function(e){return!(8192&e.flags)}))}function Xg(e){return ny(Hh(e),e)}function Yg(e){return!!(98304&(W?um(e):e.flags))}function Qg(e){return Yg(e)?_m(e):e}function Zg(n,t){lr(n,32768&t?65536&t?e.Diagnostics.Object_is_possibly_null_or_undefined:e.Diagnostics.Object_is_possibly_undefined:e.Diagnostics.Object_is_possibly_null)}function $g(n,t){lr(n,32768&t?65536&t?e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null_or_undefined:e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_undefined:e.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null)}function ey(n,t,r){if(W&&2&n.flags)return lr(t,e.Diagnostics.Object_is_of_type_unknown),xe;var a=98304&(W?um(n):n.flags);if(a){r(t,a);var i=_m(n);return 229376&i.flags?xe:i}return n}function ny(e,n){return ey(e,n,Zg)}function ty(n,t){var r=ny(n,t);return r!==xe&&16384&r.flags&&lr(t,e.Diagnostics.Object_is_possibly_undefined),r}function ry(e){return 32&e.flags?function(e){var n=Hh(e.expression),t=bm(n,e.expression);return hm(ly(e,e.expression,ny(t,e.expression),e.name),e,t!==n)}(e):ly(e,e.expression,Xg(e.expression),e.name)}function ay(e){return ly(e,e.left,Xg(e.left),e.right)}function iy(n){for(;207===n.parent.kind;)n=n.parent;return e.isCallOrNewExpression(n.parent)&&n.parent.expression===n}function oy(n,t){for(var r=e.getContainingClass(t);r;r=e.getContainingClass(r)){var a=r.symbol,i=e.getSymbolNameForPrivateIdentifier(a,n),o=a.members&&a.members.get(i)||a.exports&&a.exports.get(i);if(o)return o}}function sy(e,n){return cc(e,n.escapedName)}function cy(n,t){return(wi(t)||e.isThisProperty(n)&&Fi(t))&&e.getThisContainer(n,!0)===Gi(t)}function ly(n,t,r,a){var i=Sr(t).resolvedSymbol,s=e.getAssignmentTargetKind(n),c=Zs(0!==s||iy(n)?Im(r):r);e.isPrivateIdentifier(a)&&cS(n,1048576);var l,u,d=Li(c)||c===Ke;if(e.isPrivateIdentifier(a)){var p=oy(a.escapedText,a);if(d){if(p)return c;if(!e.getContainingClass(a))return FS(a,e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies),Ee}if(!(l=p?sy(r,p):void 0)&&function(n,t,r){var a,i=Gs(n);i&&e.forEach(i,(function(n){var r=n.valueDeclaration;if(r&&e.isNamedDeclaration(r)&&e.isPrivateIdentifier(r.name)&&r.name.escapedText===t.escapedText)return a=n,!0}));var s=kr(t);if(a){var c=a.valueDeclaration,l=e.getContainingClass(c);if(e.Debug.assert(!!l),r){var u=r.valueDeclaration,d=e.getContainingClass(u);if(e.Debug.assert(!!d),e.findAncestor(d,(function(e){return l===e}))){var p=lr(t,e.Diagnostics.The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_private_identifier_with_the_same_spelling,s,ii(n));return e.addRelatedInfo(p,e.createDiagnosticForNode(u,e.Diagnostics.The_shadowing_declaration_of_0_is_defined_here,s),e.createDiagnosticForNode(c,e.Diagnostics.The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here,s)),!0}}return lr(t,e.Diagnostics.Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier,s,kr(l.name||o)),!0}return!1}(r,a,p))return xe}else{if(d)return e.isIdentifier(t)&&i&&L_(i,n),c;l=cc(c,a.escapedText)}if(!e.isIdentifier(t)||!i||l&&GT(l)||L_(i,n),l){if(134217728&Wg(l)&&yu(n,l)&&dr(!1,a,e.Diagnostics._0_is_deprecated,a.escapedText),function(n,t,r){var a,i=n.valueDeclaration;if(!i||e.getSourceFileOfNode(t).isDeclarationFile)return;var o=e.idText(r);!function(n){return!!e.findAncestor(n,(function(n){switch(n.kind){case 163:return!0;case 288:case 165:case 167:case 168:case 290:case 158:case 228:case 283:case 280:case 281:case 282:case 275:case 223:case 286:return!1;default:return!e.isExpressionNode(n)&&"quit"}}))}(t)||e.isAccessExpression(t)&&e.isAccessExpression(t.expression)||Ar(i,r)||function(e){if(!(32&e.parent.flags))return!1;var n=lo(e.parent);for(;;){if(!(n=n.symbol&&dy(n)))return!1;var t=cc(n,e.escapedName);if(t&&t.valueDeclaration)return!0}}(n)?252!==i.kind||173===t.parent.kind||8388608&i.flags||Ar(i,r)||(a=lr(r,e.Diagnostics.Class_0_used_before_its_declaration,o)):a=lr(r,e.Diagnostics.Property_0_is_used_before_its_initialization,o);a&&e.addRelatedInfo(a,e.createDiagnosticForNode(i,e.Diagnostics._0_is_declared_here,o))}(l,n,a),hy(l,n,107===t.kind),Sr(n).resolvedSymbol=l,qg(n,105===t.kind,c,l),sh(n,l,s))return lr(a,e.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property,e.idText(a)),xe;u=cy(n,l)?Te:x_(lo(l),n)}else{var m=e.isPrivateIdentifier(a)||0!==s&&Tu(r)&&!xu(r)?void 0:mc(c,0);if(!m||!m.type)return _u(r)?Ee:r.symbol===re?(re.exports.has(a.escapedText)&&418&re.exports.get(a.escapedText).flags?lr(a,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.unescapeLeadingUnderscores(a.escapedText),ii(r)):X&&lr(a,e.Diagnostics.Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature,ii(r)),Ee):(a.escapedText&&!Pr(n)&&function(n,t){var r,a;if(!e.isPrivateIdentifier(n)&&1048576&t.flags&&!(131068&t.flags))for(var i=0,o=t.types;i<o.length;i++){var s=o[i];if(!cc(s,n.escapedText)&&!mc(s,0)){r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.declarationNameToString(n),ii(s));break}}if(py(n.escapedText,t))r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_is_a_static_member_of_type_1,e.declarationNameToString(n),ii(t));else{var c=db(t);if(c&&cc(c,n.escapedText))r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1,e.declarationNameToString(n),ii(t)),a=e.createDiagnosticForNode(n,e.Diagnostics.Did_you_forget_to_use_await);else{var l=e.declarationNameToString(n),u=ii(t),d=function(n,t){var r=Zs(t).symbol;if(!r)return;for(var a=e.getScriptTargetFeatures(),i=e.getOwnKeys(a),o=0,s=i;o<s.length;o++){var c=s[o],l=a[c][e.symbolName(r)];if(void 0!==l&&e.contains(l,n))return c}}(l,t);if(void 0!==d)r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2_or_later,l,u,d);else{var p=my(n,t);if(void 0!==p){var m=e.symbolName(p);r=e.chainDiagnosticMessages(r,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2,l,u,m),a=p.valueDeclaration&&e.createDiagnosticForNode(p.valueDeclaration,e.Diagnostics._0_is_declared_here,m)}else r=e.chainDiagnosticMessages(sc(r,t),e.Diagnostics.Property_0_does_not_exist_on_type_1,l,u)}}}var f=e.createDiagnosticForNodeFromMessageChain(n,r);a&&e.addRelatedInfo(f,a);Yt.add(f)}(a,xu(r)?c:r),xe);m.isReadonly&&(e.isAssignmentTarget(n)||e.isDeleteTarget(n))&&lr(n,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,ii(c)),u=V.noUncheckedIndexedAccess&&!e.isAssignmentTarget(n)?Xl([m.type,Ce]):m.type}return uy(n,l,u,a)}function uy(n,t,r,a){var i=e.getAssignmentTargetKind(n);if(!e.isAccessExpression(n)||1===i||t&&!(98311&t.flags)&&!(8192&t.flags&&1048576&r.flags))return r;if(r===Te)return Ui(n,t);var o=!1;if(W&&J&&107===n.expression.kind){var s=t&&t.valueDeclaration;if(s&&VE(s)){var c=v_(n);166!==c.kind||c.parent!==s.parent||8388608&s.flags||(o=!0)}}else W&&t&&t.valueDeclaration&&e.isPropertyAccessExpression(t.valueDeclaration)&&e.getAssignmentDeclarationPropertyAccessKind(t.valueDeclaration)&&v_(n)===v_(t.valueDeclaration)&&(o=!0);var l=y_(n,r,o?fm(r):r);return o&&!(32768&um(r))&&32768&um(l)?(lr(a,e.Diagnostics.Property_0_is_used_before_being_assigned,ri(t)),r):i?Qp(l):l}function dy(e){var n=Lo(e);if(0!==n.length)return tu(n)}function py(n,t){var r=t.symbol&&cc(lo(t.symbol),n);return void 0!==r&&r.valueDeclaration&&e.hasSyntacticModifier(r.valueDeclaration,32)}function my(n,t){return vy(e.isString(n)?n:e.idText(n),Gs(t),111551)}function fy(n,t){var r=e.isString(n)?n:e.idText(n),a=Gs(t),i="for"===r?e.find(a,(function(n){return"htmlFor"===e.symbolName(n)})):"class"===r?e.find(a,(function(n){return"className"===e.symbolName(n)})):void 0;return null!=i?i:vy(r,a,111551)}function _y(n,t){var r=my(n,t);return r&&e.symbolName(r)}function gy(n,t,r){return e.Debug.assert(void 0!==t,"outername should always be defined"),Ir(n,t,r,void 0,t,!1,!1,(function(n,r,a){return e.Debug.assertEqual(t,r,"name should equal outerName"),Lr(n,r,a)||vy(e.unescapeLeadingUnderscores(r),e.arrayFrom(n.values()),a)}))}function yy(n,t){return t.exports&&vy(e.idText(n),ya(t),2623475)}function vy(n,t,r){return e.getSpellingSuggestion(n,t,(function(n){var t=e.symbolName(n);if(e.startsWith(t,'"'))return;if(n.flags&r)return t;if(2097152&n.flags){var a=function(e){if(Tr(e).target!==be)return ea(e)}(n);if(a&&a.flags&r)return t}return}))}function hy(n,t,r){var a=n&&106500&n.flags&&n.valueDeclaration;if(a){var i=e.hasEffectiveModifier(a,8),o=e.isNamedDeclaration(n.valueDeclaration)&&e.isPrivateIdentifier(n.valueDeclaration.name);if((i||o)&&(!t||!e.isWriteOnlyAccess(t)||65536&n.flags)){if(r){var s=e.findAncestor(t,e.isFunctionLikeDeclaration);if(s&&s.symbol===n)return}(1&e.getCheckFlags(n)?Tr(n).target:n).isReferenced=67108863}}}function by(n,t,r,a){if(a===xe||Li(a))return!0;var i=cc(a,r);if(i){if(e.isPropertyAccessExpression(n)&&i.valueDeclaration&&e.isPrivateIdentifierPropertyDeclaration(i.valueDeclaration)){var o=e.getContainingClass(i.valueDeclaration);return!e.isOptionalChain(n)&&!!e.findAncestor(n,(function(e){return e===o}))}return qg(n,t,a,i)}return e.isInJSFile(n)&&0!=(1048576&a.flags)&&a.types.some((function(e){return by(n,t,r,e)}))}function Ey(n){var t=n.initializer;if(250===t.kind){var r=t.declarations[0];if(r&&!e.isBindingPattern(r.name))return xa(r)}else if(78===t.kind)return ff(t)}function Ty(e){return 32&e.flags?function(e){var n=Hh(e.expression),t=bm(n,e.expression);return hm(Sy(e,ny(t,e.expression)),e,t!==n)}(e):Sy(e,Xg(e.expression))}function Sy(n,t){var r=0!==e.getAssignmentTargetKind(n)||iy(n)?Im(t):t,a=n.argumentExpression,i=Hh(a);if(r===xe||r===Ke)return r;if(fh(r)&&!e.isStringLiteralLike(a))return lr(a,e.Diagnostics.A_const_enum_member_can_only_be_accessed_using_a_string_literal),xe;var o=Ru(r,function(n){var t,r=e.skipParentheses(n);if(78===r.kind){var a=ff(r);if(3&a.flags)for(var i=n,o=n.parent;o;){if(238===o.kind&&i===o.statement&&Ey(o)===a&&fc(t=Uh(o.expression),1)&&!fc(t,0))return!0;i=o,o=o.parent}}return!1}(a)?Pe:i,void 0,n,16|(e.isAssignmentTarget(n)?2|(Tu(r)&&!xu(r)?1:0):0))||xe;return ab(uy(n,o.symbol,o,a),n)}function xy(n,t,r){if(t===xe)return!1;if(!e.isWellKnownSymbolSyntactically(n))return!1;if(0==(12288&t.flags))return r&&lr(n,e.Diagnostics.A_computed_property_name_of_the_form_0_must_be_of_type_symbol,e.getTextOfNode(n)),!1;var a=n.expression,i=ff(a);if(!i)return!1;var o=Tl(!0);return!!o&&(i===o||(r&&lr(a,e.Diagnostics.Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object),!1))}function Ly(n){return e.isCallOrNewExpression(n)||e.isTaggedTemplateExpression(n)||e.isJsxOpeningLikeElement(n)}function Ay(n){return Ly(n)&&e.forEach(n.typeArguments,iT),205===n.kind?Hh(n.template):e.isJsxOpeningLikeElement(n)?Hh(n.attributes):161!==n.kind&&e.forEach(n.arguments,(function(e){Hh(e)})),ct}function Dy(e){return Ay(e),lt}function Cy(e){return!!e&&(220===e.kind||227===e.kind&&e.isSpread)}function Iy(n){return e.findIndex(n,Cy)}function Ny(e){return!!(16384&e.flags)}function Ry(e){return!!(49155&e.flags)}function ky(n,t,r,a){var i;void 0===a&&(a=!1);var o=!1,s=wv(r),c=Fv(r);if(205===n.kind)if(i=t.length,218===n.template.kind){var l=e.last(n.template.templateSpans);o=e.nodeIsMissing(l.literal)||!!l.literal.isUnterminated}else{var u=n.template;e.Debug.assert(14===u.kind),o=!!u.isUnterminated}else if(161===n.kind)i=qy(n,r);else if(e.isJsxOpeningLikeElement(n)){if(o=n.attributes.end===n.end)return!0;i=0===c?t.length:1,s=0===t.length?s:1,c=Math.min(c,1)}else{if(!n.arguments)return e.Debug.assert(204===n.kind),0===Fv(r);i=a?t.length+1:t.length,o=n.arguments.end===n.end;var d=Iy(t);if(d>=0)return d>=Fv(r)&&(Gv(r)||d<wv(r))}if(!Gv(r)&&i>s)return!1;if(o||i>=c)return!0;for(var p=i;p<c;p++){if(131072&Jf(Ov(r,p),e.isInJSFile(n)&&!W?Ry:Ny).flags)return!1}return!0}function Oy(n,t){var r=e.length(n.typeParameters),a=Sc(n.typeParameters);return!e.some(t)||t.length>=a&&t.length<=r}function Py(e){return wy(e,0,!1)}function My(e){return wy(e,0,!1)||wy(e,1,!1)}function wy(e,n,t){if(524288&e.flags){var r=Ps(e);if(t||0===r.properties.length&&!r.stringIndexInfo&&!r.numberIndexInfo){if(0===n&&1===r.callSignatures.length&&0===r.constructSignatures.length)return r.callSignatures[0];if(1===n&&1===r.constructSignatures.length&&0===r.callSignatures.length)return r.constructSignatures[0]}}}function Fy(n,t,r,a){var i=wm(n.typeParameters,n,0,a),o=Bv(t),s=r&&(o&&262144&o.flags?r.nonFixingMapper:r.mapper);return Pm(s?Sd(t,s):t,n,(function(e,n){tf(i.inferences,e,n)})),r||Mm(t,n,(function(e,n){tf(i.inferences,e,n,64)})),wc(n,pf(i),e.isInJSFile(t.declaration))}function Gy(n,t,r,a,i){if(e.isJsxOpeningLikeElement(n))return function(e,n,t,r){var a=og(n,e),i=xh(e.attributes,a,r,t);return tf(r.inferences,i,a),pf(r)}(n,t,a,i);if(161!==n.kind){var o=ag(n,e.every(t.typeParameters,(function(e){return!!Ys(e)}))?8:0);if(o){var s=ig(n),c=kd(o,jm(function(n,t){return void 0===t&&(t=0),n&&Fm(e.map(n.inferences,Vm),n.signature,n.flags|t,n.compareTypes)}(s,1))),l=Py(c),u=l&&l.typeParameters?jc(Fc(l,l.typeParameters)):c,d=kc(t);tf(i.inferences,u,d,64);var p=wm(t.typeParameters,t,i.flags),m=kd(o,s&&s.returnMapper);tf(p.inferences,m,d),i.returnMapper=e.some(p.inferences,wh)?jm(function(n){var t=e.filter(n.inferences,wh);return t.length?Fm(e.map(t,Vm),n.signature,n.flags,n.compareTypes):void 0}(p)):void 0}}var f=Uv(t),_=f?Math.min(wv(t)-1,r.length):r.length;if(f&&262144&f.flags){var g=e.find(i.inferences,(function(e){return e.typeParameter===f}));g&&(g.impliedArity=e.findIndex(r,Cy,_)<0?r.length-_:void 0)}var y=Nc(t);if(y){var v=Ky(n),h=v?Hh(v):je;tf(i.inferences,h,y)}for(var b=0;b<_;b++){var E=r[b];if(222!==E.kind){var T=Ov(t,b),S=xh(E,T,i,a);tf(i.inferences,S,T)}}if(f){var x=Uy(r,_,r.length,f,i,a);tf(i.inferences,x,f)}return pf(i)}function By(e){return 1048576&e.flags?Yf(e,By):1&e.flags||jp(Ws(e)||e)?e:tm(e)?Fl(nl(e),e.target.elementFlags,!1,e.target.labeledElementDeclarations):Fl([e],[8])}function Uy(n,t,r,a,i,o){if(t>=r-1&&Cy(d=n[r-1]))return By(227===d.kind?d.type:xh(d.expression,a,i,o));for(var s=[],c=[],l=[],u=t;u<r;u++){var d;if(Cy(d=n[u])){var p=227===d.kind?d.type:Hh(d.expression);Kp(p)?(s.push(p),c.push(8)):(s.push(aE(33,p,Ce,220===d.kind?d.expression:d)),c.push(4))}else{var m=Iu(a,td(u-t)),f=xh(d,m,i,o),_=dh(m,406978556);s.push(_?ed(f):Zp(f)),c.push(1)}227===d.kind&&d.tupleNameSource&&l.push(d.tupleNameSource)}return Fl(s,c,!1,e.length(l)===e.length(s)?l:void 0)}function Vy(n,t,r,a){for(var i,o=e.isInJSFile(n.declaration),s=n.typeParameters,c=xc(e.map(t,sd),s,Sc(s),o),l=0;l<t.length;l++){e.Debug.assert(void 0!==s[l],"Should not call checkTypeArguments with too many type arguments");var u=Us(s[l]);if(u){var d=r&&a?function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1)}:void 0,p=a||e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1;i||(i=pd(s,c));var m=c[l];if(!Yd(m,rs(kd(u,i),m),r?t[l]:void 0,p,d))return}}return c}function jy(n){if(xg(n.tagName))return 2;var t=Zs(Hh(n.tagName));return e.length(uc(t,1))?0:e.length(uc(t,0))?1:2}function Hy(n,t,r,a,i,o,s){var c={errors:void 0,skipLogging:!0};if(e.isJsxOpeningLikeElement(n))return function(n,t,r,a,i,o,s){var c=og(t,n),l=xh(n.attributes,c,void 0,a);return function(){var t,r=e.isJsxOpeningElement(n)||e.isJsxSelfClosingElement(n)&&!xg(n.tagName)?Hh(n.tagName):void 0;if(!r)return!0;var a=uc(r,0);if(!e.length(a))return!0;var o=iS(n);if(!o)return!0;var c=ca(o,111551,!0,!1,n);if(!c)return!0;var l=uc(lo(c),0);if(!e.length(l))return!0;for(var u=!1,d=0,p=0,m=l;p<m.length;p++){var f=uc(Ov(m[p],0),0);if(e.length(f))for(var _=0,g=f;_<g.length;_++){var y=g[_];if(u=!0,Gv(y))return!0;var v=wv(y);v>d&&(d=v)}}if(!u)return!0;for(var h=1/0,b=0,E=a;b<E.length;b++){var T=Fv(E[b]);T<h&&(h=T)}if(h<=d)return!0;if(i){var S=e.createDiagnosticForNode(n.tagName,e.Diagnostics.Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3,e.entityNameToString(n.tagName),h,e.entityNameToString(o),d),x=null===(t=bT(n.tagName))||void 0===t?void 0:t.valueDeclaration;x&&e.addRelatedInfo(S,e.createDiagnosticForNode(x,e.Diagnostics._0_is_declared_here,e.entityNameToString(n.tagName))),s&&s.skipLogging&&(s.errors||(s.errors=[])).push(S),s.skipLogging||Yt.add(S)}return!1}()&&Zd(l,c,r,i?n.tagName:void 0,n.attributes,void 0,o,s)}(n,r,a,i,o,s,c)?void 0:(e.Debug.assert(!o||!!c.errors,"jsx should have errors when reporting errors"),c.errors||e.emptyArray);var l=Nc(r);if(l&&l!==je&&204!==n.kind){var u=Ky(n),d=void 0;u?(d=Hh(u),e.isOptionalChainRoot(u.parent)?d=_m(d):e.isOptionalChain(u.parent)&&(d=vm(d))):d=je;var p=o?u||n:void 0,m=e.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1;if(!vp(d,l,a,p,m,s,c))return e.Debug.assert(!o||!!c.errors,"this parameter should have errors when reporting errors"),c.errors||e.emptyArray}for(var f=e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1,_=Uv(r),g=_?Math.min(wv(r)-1,t.length):t.length,y=0;y<g;y++){var v=t[y];if(222!==v.kind){var h=Ov(r,y),b=xh(v,h,void 0,i),E=4&i?Sm(b):b;if(!Zd(E,h,a,o?v:void 0,v,f,s,c))return e.Debug.assert(!o||!!c.errors,"parameter should have errors when reporting errors"),S(v,E,h),c.errors||e.emptyArray}}if(_){var T=Uy(t,g,t.length,_,void 0,i);p=o?g<t.length?t[g]:n:void 0;if(!vp(T,_,a,p,f,void 0,c))return e.Debug.assert(!o||!!c.errors,"rest parameter should have errors when reporting errors"),S(p,T,_),c.errors||e.emptyArray}return;function S(n,t,r){if(n&&o&&c.errors&&c.errors.length){if(ub(r))return;var i=ub(t);i&&_p(i,r,a)&&e.addRelatedInfo(c.errors[0],e.createDiagnosticForNode(n,e.Diagnostics.Did_you_forget_to_use_await))}}}function Ky(n){if(203===n.kind){var t=e.skipOuterExpressions(n.expression);if(e.isAccessExpression(t))return t.expression}}function Wy(n,t,r,a){var i=e.parseNodeFactory.createSyntheticExpression(t,r,a);return e.setTextRange(i,n),e.setParent(i,n),i}function zy(n){if(205===n.kind){var t=n.template,r=[Wy(t,Xn||(Xn=El("TemplateStringsArray",0,!0))||nn)];return 218===t.kind&&e.forEach(t.templateSpans,(function(e){r.push(e.expression)})),r}if(161===n.kind)return function(n){var t=n.parent,r=n.expression;switch(t.kind){case 252:case 221:return[Wy(r,lo(xa(t)))];case 160:var a=t.parent;return[Wy(r,166===t.parent.kind?lo(xa(a)):xe),Wy(r,Ee),Wy(r,Pe)];case 163:case 165:case 167:case 168:var i=163!==t.kind&&0!==j;return[Wy(r,xT(t)),Wy(r,LT(t)),Wy(r,i?Il(ET(t)):Ee)]}return e.Debug.fail()}(n);if(e.isJsxOpeningLikeElement(n))return n.attributes.properties.length>0||e.isJsxOpeningElement(n)&&n.parent.children.length>0?[n.attributes]:e.emptyArray;var a=n.arguments||e.emptyArray,i=Iy(a);if(i>=0){for(var o=a.slice(0,i),s=function(n){var t=a[n],r=220===t.kind&&(xt?Hh(t.expression):Lh(t.expression));r&&tm(r)?e.forEach(nl(r),(function(e,n){var a,i=r.target.elementFlags[n],s=Wy(t,4&i?Nl(e):e,!!(12&i),null===(a=r.target.labeledElementDeclarations)||void 0===a?void 0:a[n]);o.push(s)})):o.push(t)},c=i;c<a.length;c++)s(c);return o}return a}function qy(n,t){switch(n.parent.kind){case 252:case 221:return 1;case 163:return 2;case 165:case 167:case 168:return 0===j||t.parameters.length<=2?2:3;case 160:return 3;default:return e.Debug.fail()}}function Jy(n,t){var r,a,i=e.getSourceFileOfNode(n);if(e.isPropertyAccessExpression(n.expression)){var o=e.getErrorSpanForNode(i,n.expression.name);r=o.start,a=t?o.length:n.end-r}else{var s=e.getErrorSpanForNode(i,n.expression);r=s.start,a=t?s.length:n.end-r}return{start:r,length:a,sourceFile:i}}function Xy(n,t,r,a,i,o){if(e.isCallExpression(n)){var s=Jy(n),c=s.sourceFile,l=s.start,u=s.length;return e.createFileDiagnostic(c,l,u,t,r,a,i,o)}return e.createDiagnosticForNode(n,t,r,a,i,o)}function Yy(n,t,r){for(var a,i=Number.POSITIVE_INFINITY,o=Number.NEGATIVE_INFINITY,s=Number.NEGATIVE_INFINITY,c=Number.POSITIVE_INFINITY,l=r.length,u=0,d=t;u<d.length;u++){var p=d[u],m=Fv(p),f=wv(p);m<l&&m>s&&(s=m),l<f&&f<c&&(c=f),m<i&&(i=m,a=p),o=Math.max(o,f)}var _,g,y=e.some(t,Gv),v=y?i:i<o?i+"-"+o:i,h=Iy(r)>-1;l<=o&&h&&l--;var b=y||h?y&&h?e.Diagnostics.Expected_at_least_0_arguments_but_got_1_or_more:y?e.Diagnostics.Expected_at_least_0_arguments_but_got_1:e.Diagnostics.Expected_0_arguments_but_got_1_or_more:1===v&&0===l&&function(n){if(!e.isCallExpression(n)||!e.isIdentifier(n.expression))return!1;var t=Cr(n.expression,n.expression.escapedText,111551,void 0,void 0,!1),r=null==t?void 0:t.valueDeclaration;if(!(r&&e.isParameter(r)&&lg(r.parent)&&e.isNewExpression(r.parent.parent)&&e.isIdentifier(r.parent.parent.expression)))return!1;var a=Ll(!1);return!!a&&bT(r.parent.parent.expression,!0)===a}(n)?e.Diagnostics.Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise:e.Diagnostics.Expected_0_arguments_but_got_1;if(a&&Fv(a)>l&&a.declaration){var E=a.declaration.parameters[a.thisParameter?l+1:l];E&&(g=e.createDiagnosticForNode(E,e.isBindingPattern(E.name)?e.Diagnostics.An_argument_matching_this_binding_pattern_was_not_provided:e.isRestParameter(E)?e.Diagnostics.Arguments_for_the_rest_parameter_0_were_not_provided:e.Diagnostics.An_argument_for_0_was_not_provided,E.name?e.isBindingPattern(E.name)?void 0:e.idText(e.getFirstIdentifier(E.name)):l))}if(i<l&&l<o)return Xy(n,e.Diagnostics.No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments,l,s,c);if(!h&&l<i){var T=Xy(n,b,v,l);return g?e.addRelatedInfo(T,g):T}if(y||h){if(_=e.factory.createNodeArray(r),h&&l){var S=e.elementAt(r,Iy(r)+1)||void 0;_=e.factory.createNodeArray(r.slice(o>l&&S?r.indexOf(S):Math.min(o,r.length-1)))}}else _=e.factory.createNodeArray(r.slice(o));var x=e.first(_).pos,L=e.last(_).end;L===x&&L++,e.setTextRangePosEnd(_,x,L);var A=e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),_,b,v,l);return g?e.addRelatedInfo(A,g):A}function Qy(n,r,a,i,o,s){var c,l=205===n.kind,u=161===n.kind,d=e.isJsxOpeningLikeElement(n),p=!a&&t;u||(c=n.typeArguments,(l||d||105!==n.expression.kind)&&e.forEach(c,iT));var m=a||[];if(function(n,t,r){var a,i,o,s,c=0,l=-1;e.Debug.assert(!t.length);for(var u=0,d=n;u<d.length;u++){var p=d[u],m=p.declaration&&xa(p.declaration),f=p.declaration&&p.declaration.parent;i&&m!==i?(o=c=t.length,a=f):a&&f===a?o+=1:(a=f,o=c),i=m,B(p)?(s=++l,c++):s=o,t.splice(s,0,r?cs(p,r):p)}}(r,m,o),!m.length)return p&&Yt.add(Xy(n,e.Diagnostics.Call_target_does_not_contain_any_signatures)),Dy(n);var f,_,g,y,v=zy(n),h=1===m.length&&!m[0].typeParameters,b=u||h||!e.some(v,wd)?0:4,E=!!(16&i)&&203===n.kind&&n.arguments.hasTrailingComma;if(m.length>1&&(y=W(m,er,h,E)),y||(y=W(m,tr,h,E)),y)return y;if(p)if(f)if(1===f.length||f.length>3){var T,S=f[f.length-1];f.length>3&&(T=e.chainDiagnosticMessages(T,e.Diagnostics.The_last_overload_gave_the_following_error),T=e.chainDiagnosticMessages(T,e.Diagnostics.No_overload_matches_this_call));var x=Hy(n,v,S,tr,0,!0,(function(){return T}));if(x)for(var L=0,A=x;L<A.length;L++){var D=A[L];S.declaration&&f.length>3&&e.addRelatedInfo(D,e.createDiagnosticForNode(S.declaration,e.Diagnostics.The_last_overload_is_declared_here)),K(S,D),Yt.add(D)}else e.Debug.fail("No error for last overload signature")}else{for(var C=[],I=0,N=Number.MAX_VALUE,R=0,k=0,O=function(t){var r=Hy(n,v,t,tr,0,!0,(function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.Overload_0_of_1_2_gave_the_following_error,k+1,m.length,ai(t))}));r?(r.length<=N&&(N=r.length,R=k),I=Math.max(I,r.length),C.push(r)):e.Debug.fail("No error for 3 or fewer overload signatures"),k++},P=0,M=f;P<M.length;P++){O(M[P])}var w=I>1?C[R]:e.flatten(C);e.Debug.assert(w.length>0,"No errors reported for 3 or fewer overload signatures");var F=e.chainDiagnosticMessages(e.map(w,(function(e){return"string"==typeof e.messageText?e:e.messageText})),e.Diagnostics.No_overload_matches_this_call),U=__spreadArrays(e.flatMap(w,(function(e){return e.relatedInformation}))),V=void 0;if(e.every(w,(function(e){return e.start===w[0].start&&e.length===w[0].length&&e.file===w[0].file}))){var j=w[0];V={file:j.file,start:j.start,length:j.length,code:F.code,category:F.category,messageText:F,relatedInformation:U}}else V=e.createDiagnosticForNodeFromMessageChain(n,F,U);K(f[0],V),Yt.add(V)}else if(_)Yt.add(Yy(n,[_],v));else if(g)Vy(g,n.typeArguments,!0,s);else{var H=e.filter(r,(function(e){return Oy(e,c)}));0===H.length?Yt.add(function(n,t,r){var a=r.length;if(1===t.length){var i=Sc((d=t[0]).typeParameters),o=e.length(d.typeParameters);return e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.Expected_0_type_arguments_but_got_1,i<o?i+"-"+o:i,a)}for(var s=-1/0,c=1/0,l=0,u=t;l<u.length;l++){var d,p=Sc((d=u[l]).typeParameters);o=e.length(d.typeParameters),p>a?c=Math.min(c,p):o<a&&(s=Math.max(s,o))}return s!==-1/0&&c!==1/0?e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments,a,s,c):e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),r,e.Diagnostics.Expected_0_type_arguments_but_got_1,s===-1/0?c:s,a)}(n,r,c)):u?s&&Yt.add(Xy(n,s)):Yt.add(Yy(n,H,v))}return function(n,t,r,a){return e.Debug.assert(t.length>0),sT(n),a||1===t.length||t.some((function(e){return!!e.typeParameters}))?function(n,t,r){var a=function(e,n){for(var t=-1,r=-1,a=0;a<e.length;a++){var i=e[a],o=wv(i);if(Gv(i)||o>=n)return a;o>r&&(r=o,t=a)}return t}(t,void 0===ae?r.length:ae),i=t[a],o=i.typeParameters;if(!o)return i;var s=Ly(n)?n.typeArguments:void 0,c=s?Gc(i,function(e,n,t){var r=e.map(ET);for(;r.length>n.length;)r.pop();for(;r.length<n.length;)r.push(Us(n[r.length])||df(t));return r}(s,o,e.isInJSFile(n))):function(n,t,r,a){var i=wm(t,r,e.isInJSFile(n)?2:0),o=Gy(n,r,a,12,i);return Gc(r,o)}(n,o,i,r);return t[a]=c,c}(n,t,r):function(n){var t,r=e.mapDefined(n,(function(e){return e.thisParameter}));r.length&&(t=$y(r,r.map(Cv)));for(var a=e.minAndMax(n,Zy),i=a.min,o=a.max,s=[],c=function(t){var r=e.mapDefined(n,(function(n){return G(n)?t<n.parameters.length-1?n.parameters[t]:e.last(n.parameters):t<n.parameters.length?n.parameters[t]:void 0}));e.Debug.assert(0!==r.length),s.push($y(r,e.mapDefined(n,(function(e){return Pv(e,t)}))))},l=0;l<o;l++)c(l);var u=e.mapDefined(n,(function(n){return G(n)?e.last(n.parameters):void 0})),d=0;if(0!==u.length){var p=Nl(Xl(e.mapDefined(n,Mc),2));s.push(ev(u,p)),d|=1}n.some(B)&&(d|=2);return is(n[0].declaration,void 0,t,s,tu(n.map(kc)),void 0,i,d)}(t)}(n,m,v,!!a);function K(n,t){var r,a,i=f,o=_,s=g,c=(null===(a=null===(r=n.declaration)||void 0===r?void 0:r.symbol)||void 0===a?void 0:a.declarations)||e.emptyArray,l=c.length>1?e.find(c,(function(n){return e.isFunctionLikeDeclaration(n)&&e.nodeIsPresent(n.body)})):void 0;if(l){var u=Lc(l),d=!u.typeParameters;W([u],tr,d)&&e.addRelatedInfo(t,e.createDiagnosticForNode(l,e.Diagnostics.The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_are_not_externally_visible))}f=i,_=o,g=s}function W(t,r,a,i){if(void 0===i&&(i=!1),f=void 0,_=void 0,g=void 0,a){var o=t[0];if(e.some(c)||!ky(n,v,o,i))return;return Hy(n,v,o,r,0,!1,void 0)?void(f=[o]):o}for(var s=0;s<t.length;s++){if(Oy(o=t[s],c)&&ky(n,v,o,i)){var l=void 0,u=void 0;if(o.typeParameters){var d=void 0;if(e.some(c)){if(!(d=Vy(o,c,!1))){g=o;continue}}else u=wm(o.typeParameters,o,e.isInJSFile(n)?2:0),d=Gy(n,o,v,8|b,u),b|=4&u.flags?8:0;if(l=wc(o,d,e.isInJSFile(o.declaration),u&&u.inferredTypeParameters),Uv(o)&&!ky(n,v,l,i)){_=l;continue}}else l=o;if(!Hy(n,v,l,r,b,!1,void 0)){if(b){if(b=0,u)if(l=wc(o,d=Gy(n,o,v,b,u),e.isInJSFile(o.declaration),u&&u.inferredTypeParameters),Uv(o)&&!ky(n,v,l,i)){_=l;continue}if(Hy(n,v,l,r,b,!1,void 0)){(f||(f=[])).push(l);continue}}return t[s]=l,l}(f||(f=[])).push(l)}}}}function Zy(e){var n=e.parameters.length;return G(e)?n-1:n}function $y(e,n){return ev(e,Xl(n,2))}function ev(n,t){return Tm(e.first(n),t)}function nv(e){return!(!e.typeParameters||!XT(kc(e)))}function tv(e,n,t,r){return Li(e)||Li(n)&&!!(262144&e.flags)||!t&&!r&&!(1179648&n.flags)&&zd(e,vn)}function rv(n,t,r){if(n.arguments&&j<1){var a=Iy(n.arguments);a>=0&&lr(n.arguments[a],e.Diagnostics.Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher)}var i=Xg(n.expression);if(i===Ke)return dt;if((i=Zs(i))===xe)return Dy(n);if(Li(i))return n.typeArguments&&lr(n,e.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments),Ay(n);var o=uc(i,1);if(o.length){if(!function(n,t){if(!t||!t.declaration)return!0;var r=t.declaration,a=e.getSelectedEffectiveModifierFlags(r,24);if(!a||166!==r.kind)return!0;var i=e.getClassLikeDeclarationOfSymbol(r.parent.symbol),o=Fo(r.parent.symbol);if(!yT(n,i)){var s=e.getContainingClass(n);if(s&&16&a){var c=ET(s);if(av(r.parent.symbol,c))return!0}return 8&a&&lr(n,e.Diagnostics.Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration,ii(o)),16&a&&lr(n,e.Diagnostics.Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration,ii(o)),!1}return!0}(n,o[0]))return Dy(n);var s=i.symbol&&e.getClassLikeDeclarationOfSymbol(i.symbol);return s&&e.hasSyntacticModifier(s,128)?(lr(n,e.Diagnostics.Cannot_create_an_instance_of_an_abstract_class),Dy(n)):Qy(n,o,t,r,0)}var c=uc(i,0);if(c.length){var l=Qy(n,c,t,r,0);return X||(l.declaration&&!fv(l.declaration)&&kc(l)!==je&&lr(n,e.Diagnostics.Only_a_void_function_can_be_called_with_the_new_keyword),Nc(l)===je&&lr(n,e.Diagnostics.A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void)),l}return ov(n.expression,i,1),Dy(n)}function av(n,t){var r=Lo(t);if(!e.length(r))return!1;var a=r[0];if(2097152&a.flags){for(var i=ys(a.types),o=0,s=0,c=a.types;s<c.length;s++){var l=c[s];if(!i[o]&&3&e.getObjectFlags(l)){if(l.symbol===n)return!0;if(av(n,l))return!0}o++}return!1}return a.symbol===n||av(n,a)}function iv(n,t,r){var a,i=0===r,o=mb(t),s=o&&uc(o,r).length>0;if(1048576&t.flags){for(var c=!1,l=0,u=t.types;l<u.length;l++){var d=u[l];if(0!==uc(d,r).length){if(c=!0,a)break}else if(a||(a=e.chainDiagnosticMessages(a,i?e.Diagnostics.Type_0_has_no_call_signatures:e.Diagnostics.Type_0_has_no_construct_signatures,ii(d)),a=e.chainDiagnosticMessages(a,i?e.Diagnostics.Not_all_constituents_of_type_0_are_callable:e.Diagnostics.Not_all_constituents_of_type_0_are_constructable,ii(t))),c)break}c||(a=e.chainDiagnosticMessages(void 0,i?e.Diagnostics.No_constituent_of_type_0_is_callable:e.Diagnostics.No_constituent_of_type_0_is_constructable,ii(t))),a||(a=e.chainDiagnosticMessages(a,i?e.Diagnostics.Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_other:e.Diagnostics.Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_with_each_other,ii(t)))}else a=e.chainDiagnosticMessages(a,i?e.Diagnostics.Type_0_has_no_call_signatures:e.Diagnostics.Type_0_has_no_construct_signatures,ii(t));var p=i?e.Diagnostics.This_expression_is_not_callable:e.Diagnostics.This_expression_is_not_constructable;if(e.isCallExpression(n.parent)&&0===n.parent.arguments.length){var m=Sr(n).resolvedSymbol;m&&32768&m.flags&&(p=e.Diagnostics.This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without)}return{messageChain:e.chainDiagnosticMessages(a,p),relatedMessage:s?e.Diagnostics.Did_you_forget_to_use_await:void 0}}function ov(n,t,r,a){var i=iv(n,t,r),o=i.messageChain,s=i.relatedMessage,c=e.createDiagnosticForNodeFromMessageChain(n,o);if(s&&e.addRelatedInfo(c,e.createDiagnosticForNode(n,s)),e.isCallExpression(n.parent)){var l=Jy(n.parent,!0),u=l.start,d=l.length;c.start=u,c.length=d}Yt.add(c),sv(t,r,a?e.addRelatedInfo(c,a):c)}function sv(n,t,r){if(n.symbol){var a=Tr(n.symbol).originatingImport;if(a&&!e.isImportCall(a)){var i=uc(lo(Tr(n.symbol).target),t);if(!i||!i.length)return;e.addRelatedInfo(r,e.createDiagnosticForNode(a,e.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead))}}}function cv(n){switch(n.parent.kind){case 252:case 221:return e.Diagnostics.Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression;case 160:return e.Diagnostics.Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression;case 163:return e.Diagnostics.Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression;case 165:case 167:case 168:return e.Diagnostics.Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression;default:return e.Debug.fail()}}function lv(n,t,r){var a=Hh(n.expression),i=Zs(a);if(i===xe)return Dy(n);var o,s,c=uc(i,0),l=uc(i,1).length;if(tv(a,i,c.length,l))return Ay(n);if(o=n,(s=c).length&&e.every(s,(function(e){return 0===e.minArgumentCount&&!G(e)&&e.parameters.length<qy(o,e)}))){var u=e.getTextOfNode(n.expression,!1);return lr(n,e.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0,u),Dy(n)}var d=cv(n);if(!c.length){var p=iv(n.expression,i,0),m=e.chainDiagnosticMessages(p.messageChain,d),f=e.createDiagnosticForNodeFromMessageChain(n.expression,m);return p.relatedMessage&&e.addRelatedInfo(f,e.createDiagnosticForNode(n.expression,p.relatedMessage)),Yt.add(f),sv(i,0,f),Dy(n)}return Qy(n,c,t,r,0,d)}function uv(n,t){var r=kg(n),a=r&&ha(r),i=a&&Lr(a,D.Element,788968),o=i&&ee.symbolToEntityName(i,788968,n),s=e.factory.createFunctionTypeNode(void 0,[e.factory.createParameterDeclaration(void 0,void 0,void 0,"props",void 0,ee.typeToTypeNode(t,n))],o?e.factory.createTypeReferenceNode(o,void 0):e.factory.createKeywordTypeNode(128)),c=mr(1,"props");return c.type=t,is(s,void 0,void 0,[c],i?Fo(i):xe,void 0,1,0)}function dv(n,t,r){if(xg(n.tagName)){var a=Fg(n),i=uv(n,a);return Qd(xh(n.attributes,og(i,n),void 0,0),a,n.tagName,n.attributes),e.length(n.typeArguments)&&(e.forEach(n.typeArguments,iT),Yt.add(e.createDiagnosticForNodeArray(e.getSourceFileOfNode(n),n.typeArguments,e.Diagnostics.Expected_0_type_arguments_but_got_1,0,e.length(n.typeArguments)))),i}var o=Hh(n.tagName),s=Zs(o);if(s===xe)return Dy(n);var c=Mg(o,n);return tv(o,s,c.length,0)?Ay(n):0===c.length?(lr(n.tagName,e.Diagnostics.JSX_element_type_0_does_not_have_any_construct_or_call_signatures,e.getTextOfNode(n.tagName)),Dy(n)):Qy(n,c,t,r,0)}function pv(n,t,r){switch(n.kind){case 203:return function(n,t,r){if(105===n.expression.kind){var a=P_(n.expression);if(Li(a)){for(var i=0,o=n.arguments;i<o.length;i++)Hh(o[i]);return ct}if(a!==xe){var s=e.getEffectiveBaseTypeNode(e.getContainingClass(n));if(s)return Qy(n,To(a,s.typeArguments,s),t,r,0)}return Ay(n)}var c,l=Hh(n.expression);if(e.isCallChain(n)){var u=bm(l,n.expression);c=u===l?0:e.isOutermostOptionalChain(n)?8:4,l=u}else c=0;if((l=ey(l,n.expression,$g))===Ke)return dt;var d=Zs(l);if(d===xe)return Dy(n);var p=uc(d,0),m=uc(d,1).length;if(tv(l,d,p.length,m))return l!==xe&&n.typeArguments&&lr(n,e.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments),Ay(n);if(!p.length){if(m)lr(n,e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new,ii(l));else{var f=void 0;if(1===n.arguments.length){var _=e.getSourceFileOfNode(n).text;e.isLineBreak(_.charCodeAt(e.skipTrivia(_,n.expression.end,!0)-1))&&(f=e.createDiagnosticForNode(n.expression,e.Diagnostics.Are_you_missing_a_semicolon))}ov(n.expression,d,0,f)}return Dy(n)}return 8&r&&!n.typeArguments&&p.some(nv)?(Mh(n,r),ut):p.some((function(n){return e.isInJSFile(n.declaration)&&!!e.getJSDocClassTag(n.declaration)}))?(lr(n,e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new,ii(l)),Dy(n)):Qy(n,p,t,r,c)}(n,t,r);case 204:return rv(n,t,r);case 205:return function(n,t,r){var a=Hh(n.tag),i=Zs(a);if(i===xe)return Dy(n);var o=uc(i,0),s=uc(i,1).length;if(tv(a,i,o.length,s))return Ay(n);if(!o.length){if(e.isArrayLiteralExpression(n.parent)){var c=e.createDiagnosticForNode(n.tag,e.Diagnostics.It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tagged_template_expression_which_cannot_be_invoked);return Yt.add(c),Dy(n)}return ov(n.tag,i,0),Dy(n)}return Qy(n,o,t,r,0)}(n,t,r);case 161:return lv(n,t,r);case 275:case 274:return dv(n,t,r)}throw e.Debug.assertNever(n,"Branch in 'resolveSignature' should be unreachable.")}function mv(e,n,t){var r=Sr(e),a=r.resolvedSignature;if(a&&a!==ut&&!n)return a;r.resolvedSignature=ut;var i=pv(e,n,t||0);return i!==ut&&(r.resolvedSignature=St===xt?i:a),i}function fv(n){var t;if(!n||!e.isInJSFile(n))return!1;var r=e.isFunctionDeclaration(n)||e.isFunctionExpression(n)?n:e.isVariableDeclaration(n)&&n.initializer&&e.isFunctionExpression(n.initializer)?n.initializer:void 0;if(r){if(e.getJSDocClassTag(n))return!0;var a=xa(r);return!!(null===(t=null==a?void 0:a.members)||void 0===t?void 0:t.size)}return!1}function _v(n,t){var r,a;if(t){var i=Tr(t);if(!i.inferredClassSymbol||!i.inferredClassSymbol.has(k(n))){var o=e.isTransientSymbol(n)?n:gr(n);return o.exports=o.exports||e.createSymbolTable(),o.members=o.members||e.createSymbolTable(),o.flags|=32&t.flags,(null===(r=t.exports)||void 0===r?void 0:r.size)&&br(o.exports,t.exports),(null===(a=t.members)||void 0===a?void 0:a.size)&&br(o.members,t.members),(i.inferredClassSymbol||(i.inferredClassSymbol=new e.Map)).set(k(o),o),o}return i.inferredClassSymbol.get(k(n))}}function gv(n,t){if(n.parent){var r,a;if(e.isVariableDeclaration(n.parent)&&n.parent.initializer===n){if(!(e.isInJSFile(n)||e.isVarConst(n.parent)&&e.isFunctionLikeDeclaration(n)))return;r=n.parent.name,a=n.parent}else if(e.isBinaryExpression(n.parent)){var i=n.parent,o=n.parent.operatorToken.kind;if(62!==o||!t&&i.right!==n){if(!(56!==o&&60!==o||(e.isVariableDeclaration(i.parent)&&i.parent.initializer===i?(r=i.parent.name,a=i.parent):e.isBinaryExpression(i.parent)&&62===i.parent.operatorToken.kind&&(t||i.parent.right===i)&&(a=r=i.parent.left),r&&e.isBindableStaticNameExpression(r)&&e.isSameEntityName(r,i.left))))return}else a=r=i.left}else t&&e.isFunctionDeclaration(n)&&(r=n.name,a=n);if(a&&r&&(t||e.getExpandoInitializer(n,e.isPrototypeAccess(r))))return xa(a)}}function yv(n,t){n.declaration&&134217728&n.declaration.flags&&dr(!1,vv(t),e.Diagnostics._0_is_deprecated,ai(n))}function vv(n){switch((n=e.skipParentheses(n)).kind){case 203:case 161:case 204:return vv(n.expression);case 205:return vv(n.tag);case 275:case 274:return vv(n.tagName);case 202:return n.argumentExpression;case 201:return n.name;case 173:var t=n;return e.isQualifiedName(t.typeName)?t.typeName.right:t;default:return n}}function hv(n){if(!e.isCallExpression(n))return!1;var t=n.expression;if(e.isPropertyAccessExpression(t)&&"for"===t.name.escapedText&&(t=t.expression),!e.isIdentifier(t)||"Symbol"!==t.escapedText)return!1;var r=Tl(!1);return!!r&&r===Cr(t,"Symbol",111551,void 0,void 0,!1)}function bv(n){if(yS(n.arguments)||function(n){if(H===e.ModuleKind.ES2015)return FS(n,e.Diagnostics.Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system_or_umd);if(n.typeArguments)return FS(n,e.Diagnostics.Dynamic_import_cannot_have_type_arguments);var t=n.arguments;if(1!==t.length)return FS(n,e.Diagnostics.Dynamic_import_must_have_one_specifier_as_an_argument);if(pS(t),e.isSpreadElement(t[0]))return FS(t[0],e.Diagnostics.Specifier_of_dynamic_import_cannot_be_spread_element)}(n),0===n.arguments.length)return Jv(n,Ee);for(var t=n.arguments[0],r=Lh(t),a=1;a<n.arguments.length;++a)Lh(n.arguments[a]);(32768&r.flags||65536&r.flags||!zd(r,Oe))&&lr(t,e.Diagnostics.Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0,ii(r));var i=ua(n,t);if(i){var o=_a(i,t,!0,!1);if(o)return Jv(n,Ev(lo(o),o,i))}return Jv(n,Ee)}function Ev(n,t,r){if(K&&n&&n!==xe){var a=n;if(!a.syntheticType)if(Wr(e.find(r.declarations,e.isSourceFile),r,!1)){var i=e.createSymbolTable(),o=mr(2097152,"default");o.parent=r,o.nameType=td("default"),o.target=$r(t),i.set("default",o);var s=mr(2048,"__type"),c=Va(s,i,e.emptyArray,e.emptyArray,void 0,void 0);s.type=c,a.syntheticType=Tg(n)?Ju(n,c,s,0,!1):c}else a.syntheticType=n;return a.syntheticType}return n}function Tv(n){if(!e.isRequireCall(n,!0))return!1;if(!e.isIdentifier(n.expression))return e.Debug.fail();var t=Cr(n.expression,n.expression.escapedText,111551,void 0,void 0,!0);if(t===oe)return!0;if(2097152&t.flags)return!1;var r=16&t.flags?251:3&t.flags?249:0;if(0!==r){var a=e.getDeclarationOfKind(t,r);return!!a&&!!(8388608&a.flags)}return!1}function Sv(n){(function(n){if(n.questionDotToken||32&n.flags)return FS(n.template,e.Diagnostics.Tagged_template_expressions_are_not_permitted_in_an_optional_chain);return!1})(n)||gS(n,n.typeArguments),j<2&&cS(n,524288);var t=mv(n);return yv(t,n),kc(t)}function xv(n){switch(n.kind){case 10:case 14:case 8:case 9:case 109:case 94:case 199:case 200:case 218:return!0;case 207:return xv(n.expression);case 214:var t=n.operator,r=n.operand;return 40===t&&(8===r.kind||9===r.kind)||39===t&&8===r.kind;case 201:case 202:var a=n.expression;if(e.isIdentifier(a)){var i=bT(a);return i&&2097152&i.flags&&(i=ea(i)),!!(i&&384&i.flags&&1===ko(i))}}return!1}function Lv(n,r,a,i){var o=Hh(a,i);if(e.isConstTypeReference(r))return xv(a)||lr(a,e.Diagnostics.A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array_or_object_literals),ed(o);iT(r),o=Sm(Qp(o));var s=sd(r);t&&s!==xe&&(Jd(s,Im(o))||op(o,s,n,e.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first));return s}function Av(e){return 32&e.flags?function(e){var n=Hh(e.expression),t=bm(n,e.expression);return hm(_m(t),e,t!==n)}(e):_m(Hh(e.expression))}function Dv(n){return function(n){var t=n.name.escapedText;switch(n.keywordToken){case 102:if("target"!==t)return FS(n.name,e.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2,n.name.escapedText,e.tokenToString(n.keywordToken),"target");break;case 99:if("meta"!==t)FS(n.name,e.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2,n.name.escapedText,e.tokenToString(n.keywordToken),"meta")}}(n),102===n.keywordToken?function(n){var t=e.getNewTargetContainer(n);return t?166===t.kind?lo(xa(t.parent)):lo(xa(t)):(lr(n,e.Diagnostics.Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor,"new.target"),xe)}(n):99===n.keywordToken?function(n){H!==e.ModuleKind.ES2020&&H!==e.ModuleKind.ESNext&&H!==e.ModuleKind.System&&lr(n,e.Diagnostics.The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system);var t=e.getSourceFileOfNode(n);return e.Debug.assert(!!(2097152&t.flags),"Containing file is missing import meta node flag."),e.Debug.assert(!!t.externalModuleIndicator,"Containing file should be a module."),"meta"===n.name.escapedText?function(){return Yn||(Yn=El("ImportMeta",0,!0))||nn}():xe}(n):e.Debug.assertNever(n.keywordToken)}function Cv(n){var t=lo(n);if(W){var r=n.valueDeclaration;if(r&&e.hasInitializer(r))return fm(t)}return t}function Iv(n){return e.Debug.assert(e.isIdentifier(n.name)),n.name.escapedText}function Nv(e,n,t){var r=e.parameters.length-(G(e)?1:0);if(n<r)return e.parameters[n].escapedName;var a=e.parameters[r]||he,i=t||lo(a);if(tm(i)){var o=i.target.labeledElementDeclarations,s=n-r;return o&&Iv(o[s])||a.escapedName+"_"+s}return a.escapedName}function Rv(n){return 192===n.kind||e.isParameter(n)&&n.name&&e.isIdentifier(n.name)}function kv(e,n){var t=e.parameters.length-(G(e)?1:0);if(n<t){var r=e.parameters[n].valueDeclaration;return r&&Rv(r)?r:void 0}var a=e.parameters[t]||he,i=lo(a);if(tm(i)){var o=i.target.labeledElementDeclarations;return o&&o[n-t]}return a.valueDeclaration&&Rv(a.valueDeclaration)?a.valueDeclaration:void 0}function Ov(e,n){return Pv(e,n)||Ee}function Pv(e,n){var t=e.parameters.length-(G(e)?1:0);if(n<t)return Cv(e.parameters[n]);if(G(e)){var r=lo(e.parameters[t]),a=n-t;if(!tm(r)||r.target.hasRestElement||a<r.target.fixedLength)return Iu(r,td(a))}}function Mv(n,t){var r=wv(n),a=Fv(n),i=Bv(n);if(i&&t>=r-1)return t===r-1?i:Nl(Iu(i,Pe));for(var o=[],s=[],c=[],l=t;l<r;l++){!i||l<r-1?(o.push(Ov(n,l)),s.push(l<a?1:2)):(o.push(i),s.push(8));var u=kv(n,l);u&&c.push(u)}return Fl(o,s,!1,e.length(c)===e.length(o)?c:void 0)}function wv(e){var n=e.parameters.length;if(G(e)){var t=lo(e.parameters[n-1]);if(tm(t))return n+t.target.fixedLength-(t.target.hasRestElement?0:1)}return n}function Fv(n,t){var r=1&t,a=2&t;if(a||void 0===n.resolvedMinArgumentCount){var i=void 0;if(G(n)){var o=lo(n.parameters[n.parameters.length-1]);if(tm(o)){var s=e.findIndex(o.target.elementFlags,(function(e){return!(1&e)})),c=s<0?o.target.fixedLength:s;c>0&&(i=n.parameters.length-1+c)}}if(void 0===i){if(!r&&16&n.flags)return 0;i=n.minArgumentCount}if(a)return i;for(var l=i-1;l>=0;l--){if(131072&Jf(Ov(n,l),Ny).flags)break;i=l}n.resolvedMinArgumentCount=i}return n.resolvedMinArgumentCount}function Gv(e){if(G(e)){var n=lo(e.parameters[e.parameters.length-1]);return!tm(n)||n.target.hasRestElement}return!1}function Bv(e){if(G(e)){var n=lo(e.parameters[e.parameters.length-1]);if(!tm(n))return n;if(n.target.hasRestElement)return Vl(n,n.target.fixedLength)}}function Uv(e){var n=Bv(e);return!n||Up(n)||Li(n)||0!=(131072&rc(n).flags)?void 0:n}function Vv(e){return jv(e,He)}function jv(e,n){return e.parameters.length>0?Ov(e,0):n}function Hv(n,t){(n.typeParameters=t.typeParameters,t.thisParameter)&&((!(i=n.thisParameter)||i.valueDeclaration&&!i.valueDeclaration.type)&&(i||(n.thisParameter=Tm(t.thisParameter,void 0)),Kv(n.thisParameter,lo(t.thisParameter))));for(var r=n.parameters.length-(G(n)?1:0),a=0;a<r;a++){var i=n.parameters[a];if(!e.getEffectiveTypeAnnotationNode(i.valueDeclaration))Kv(i,Pv(t,a))}if(G(n)){i=e.last(n.parameters);if(e.isTransientSymbol(i)||!e.getEffectiveTypeAnnotationNode(i.valueDeclaration))Kv(i,Mv(t,r))}}function Kv(e,n){var t=Tr(e);if(!t.type){var r=e.valueDeclaration;t.type=n||Xi(r,!0),78!==r.name.kind&&(t.type===De&&(t.type=Ji(r.name)),Wv(r.name))}}function Wv(n){for(var t=0,r=n.elements;t<r.length;t++){var a=r[t];e.isOmittedExpression(a)||(78===a.name.kind?Tr(xa(a)).type=Ri(a):Wv(a.name))}}function zv(e){var n=xl(!0);return n!==on?Zc(n,[e=mb(e)||De]):De}function qv(e){var n,t=(n=!0,wn||(wn=El("PromiseLike",1,n))||on);return t!==on?Zc(t,[e=mb(e)||De]):De}function Jv(n,t){var r=zv(t);return r===De?(lr(n,e.isImportCall(n)?e.Diagnostics.A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option:e.Diagnostics.An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option),xe):(Ll(!0)||lr(n,e.isImportCall(n)?e.Diagnostics.A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option:e.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option),r)}function Xv(n,t){if(!n.body)return xe;var r,a,i,o=e.getFunctionFlags(n),s=0!=(2&o),c=0!=(1&o),l=je;if(230!==n.body.kind)r=Lh(n.body,t&&-9&t),s&&(r=pb(r,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member));else if(c){var u=nh(n,t);u?u.length>0&&(r=Xl(u,2)):l=He;var d=function(n,t){var r=[],a=[],i=0!=(2&e.getFunctionFlags(n));return e.forEachYieldExpression(n.body,(function(n){var o,s=n.expression?Hh(n.expression,t):Ie;if(e.pushIfUnique(r,Qv(n,s,Ee,i)),n.asteriskToken){var c=dE(s,i?19:17,n.expression);o=c&&c.nextType}else o=ag(n);o&&e.pushIfUnique(a,o)})),{yieldTypes:r,nextTypes:a}}(n,t),p=d.yieldTypes,m=d.nextTypes;a=e.some(p)?Xl(p,2):void 0,i=e.some(m)?tu(m):void 0}else{var f=nh(n,t);if(!f)return 2&o?Jv(n,He):He;if(0===f.length)return 2&o?Jv(n,je):je;r=Xl(f,2)}if(r||a||i){if(a&&Om(n,a,3),r&&Om(n,r,1),i&&Om(n,i,2),r&&Xp(r)||a&&Xp(a)||i&&Xp(i)){var _=ug(n),g=_?_===Lc(n)?c?void 0:r:tg(kc(_),n):void 0;c?(a=nm(a,g,0,s),r=nm(r,g,1,s),i=nm(i,g,2,s)):r=function(e,n,t){return e&&Xp(e)&&(e=em(e,n?t?db(n):n:void 0)),e}(r,g,s)}a&&(a=Im(a)),r&&(r=Im(r)),i&&(i=Im(i))}return c?Yv(a||He,r||l,i||V_(2,n)||De,s):s?zv(r||l):r||l}function Yv(e,n,t,r){var a=r?vt:ht,i=a.getGlobalGeneratorType(!1);if(e=a.resolveIterationType(e,void 0)||De,n=a.resolveIterationType(n,void 0)||De,t=a.resolveIterationType(t,void 0)||De,i===on){var o=a.getGlobalIterableIteratorType(!1),s=o!==on?_E(o,a):void 0,c=s?s.returnType:Ee,l=s?s.nextType:Ce;return zd(n,c)&&zd(l,t)?o!==on?Cl(o,[e]):(a.getGlobalIterableIteratorType(!0),nn):(a.getGlobalGeneratorType(!0),nn)}return Cl(i,[e,n,t])}function Qv(n,t,r,a){var i=n.expression||n,o=n.asteriskToken?aE(a?19:17,t,r,i):t;return a?mb(o,i,n.asteriskToken?e.Diagnostics.Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member:e.Diagnostics.Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member):o}function Zv(e,n,t,r){var a=0;if(r){for(var i=n;i<t.length;i++)a|=E.get(t[i])||32768;for(i=e;i<n;i++)a&=~(E.get(t[i])||0);for(i=0;i<e;i++)a|=E.get(t[i])||32768}else{for(i=e;i<n;i++)a|=b.get(t[i])||128;for(i=0;i<e;i++)a&=~(b.get(t[i])||0)}return a}function $v(n){var t=Sr(n);return void 0!==t.isExhaustive?t.isExhaustive:t.isExhaustive=function(n){if(211===n.expression.kind){var t=Uh(n.expression.expression),r=Zv(0,0,Kf(n,!1),!0),a=Ws(t)||t;return 3&a.flags?556800==(556800&r):!!(131072&Jf(a,(function(e){return(Cf(e)&r)===r})).flags)}var i=Uh(n.expression);if(!Yp(i))return!1;var o=Hf(n);if(!o.length||e.some(o,Jp))return!1;return s=Yf(i,ed),c=o,1048576&s.flags?!e.forEach(s.types,(function(n){return!e.contains(c,n)})):e.contains(c,s);var s,c}(n)}function eh(e){return e.endFlowNode&&m_(e.endFlowNode)}function nh(n,t){var r=e.getFunctionFlags(n),a=[],i=eh(n),o=!1;if(e.forEachReturnStatement(n.body,(function(s){var c=s.expression;if(c){var l=Lh(c,t&&-9&t);2&r&&(l=pb(l,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)),131072&l.flags&&(o=!0),e.pushIfUnique(a,l)}else i=!0})),0!==a.length||i||!o&&!function(e){switch(e.kind){case 208:case 209:return!0;case 165:return 200===e.parent.kind;default:return!1}}(n))return!(W&&a.length&&i)||fv(n)&&a.some((function(e){return e.symbol===n.symbol}))||e.pushIfUnique(a,Ce),a}function th(n,r){if(t){var a=e.getFunctionFlags(n),i=r&&NE(r,a);if((!i||!dh(i,16385))&&164!==n.kind&&!e.nodeIsMissing(n.body)&&230===n.body.kind&&eh(n)){var o=512&n.flags;if(i&&131072&i.flags)lr(e.getEffectiveReturnTypeNode(n),e.Diagnostics.A_function_returning_never_cannot_have_a_reachable_end_point);else if(i&&!o)lr(e.getEffectiveReturnTypeNode(n),e.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value);else if(i&&W&&!zd(Ce,i))lr(e.getEffectiveReturnTypeNode(n)||n,e.Diagnostics.Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined);else if(V.noImplicitReturns){if(!i){if(!o)return;if(RE(n,kc(Lc(n))))return}lr(e.getEffectiveReturnTypeNode(n)||n,e.Diagnostics.Not_all_code_paths_return_a_value)}}}}function rh(n,t){if(e.Debug.assert(165!==n.kind||e.isObjectLiteralMethod(n)),sT(n),t&&4&t&&wd(n)){if(!e.getEffectiveReturnTypeNode(n)&&!Gd(n)){var r=dg(n);if(r&&Hm(kc(r))){var a=Sr(n);if(a.contextFreeType)return a.contextFreeType;var i=Xv(n,t),o=is(void 0,void 0,void 0,e.emptyArray,i,void 0,0,0),s=Va(n.symbol,L,[o],e.emptyArray,void 0,void 0);return s.objectFlags|=2097152,a.contextFreeType=s}}return sn}return _S(n)||208!==n.kind||ES(n),function(n,t){var r=Sr(n);if(!(1024&r.flags)){var a=dg(n);if(!(1024&r.flags)){r.flags|=1024;var i=e.firstOrUndefined(uc(lo(xa(n)),0));if(!i)return;if(wd(n))if(a){var o=ig(n);t&&2&t&&function(n,t,r){for(var a=n.parameters.length-(G(n)?1:0),i=0;i<a;i++){var o=n.parameters[i].valueDeclaration;if(o.type){var s=e.getEffectiveTypeAnnotationNode(o);s&&tf(r.inferences,sd(s),Ov(t,i))}}var c=Bv(t);if(c&&262144&c.flags){Hv(n,Sd(t,r.nonFixingMapper));var l=wv(t)-1;tf(r.inferences,Mv(n,l),c)}}(i,a,o),Hv(i,o?Sd(a,o.mapper):a)}else!function(e){e.thisParameter&&Kv(e.thisParameter);for(var n=0,t=e.parameters;n<t.length;n++)Kv(t[n])}(i);if(a&&!Oc(n)&&!i.resolvedReturnType){var s=Xv(n,t);i.resolvedReturnType||(i.resolvedReturnType=s)}qh(n)}}}(n,t),lo(xa(n))}function ah(e,n,t,r){if(void 0===r&&(r=!1),!zd(n,Qe)){var a=r&&ub(n);return pr(e,!!a&&zd(a,Qe),t),!1}return!0}function ih(n){if(!e.isCallExpression(n))return!1;if(!e.isBindableObjectDefinePropertyCall(n))return!1;var t=Lh(n.arguments[2]);if(xi(t,"value")){var r=cc(t,"writable"),a=r&&lo(r);if(!a||a===we||a===Fe)return!0;if(r&&r.valueDeclaration&&e.isPropertyAssignment(r.valueDeclaration)){var i=Hh(r.valueDeclaration.initializer);if(i===we||i===Fe)return!0}return!1}return!cc(t,"set")}function oh(n){return!!(8&e.getCheckFlags(n)||4&n.flags&&64&e.getDeclarationModifierFlagsFromSymbol(n)||3&n.flags&&2&Wg(n)||98304&n.flags&&!(65536&n.flags)||8&n.flags||e.some(n.declarations,ih))}function sh(n,t,r){var a,i;if(0===r)return!1;if(oh(t)){if(4&t.flags&&e.isAccessExpression(n)&&107===n.expression.kind){var o=e.getContainingFunction(n);if(!o||166!==o.kind&&!fv(o))return!0;if(t.valueDeclaration){var s=e.isBinaryExpression(t.valueDeclaration),c=o.parent===t.valueDeclaration.parent,l=o===t.valueDeclaration.parent,u=s&&(null===(a=t.parent)||void 0===a?void 0:a.valueDeclaration)===o.parent,d=s&&(null===(i=t.parent)||void 0===i?void 0:i.valueDeclaration)===o;return!(c||l||u||d)}}return!0}if(e.isAccessExpression(n)){var p=e.skipParentheses(n.expression);if(78===p.kind){var m=Sr(p).resolvedSymbol;if(2097152&m.flags){var f=Br(m);return!!f&&263===f.kind}}}return!1}function ch(n,t,r){var a=e.skipOuterExpressions(n,7);return 78===a.kind||e.isAccessExpression(a)?!(32&a.flags)||(lr(n,r),!1):(lr(n,t),!1)}function lh(n){Hh(n.expression);var t=e.skipParentheses(n.expression);if(!e.isAccessExpression(t))return lr(t,e.Diagnostics.The_operand_of_a_delete_operator_must_be_a_property_reference),Ue;e.isPropertyAccessExpression(t)&&e.isPrivateIdentifier(t.name)&&lr(t,e.Diagnostics.The_operand_of_a_delete_operator_cannot_be_a_private_identifier);var r=Na(Sr(t).resolvedSymbol);return r&&(oh(r)&&lr(t,e.Diagnostics.The_operand_of_a_delete_operator_cannot_be_a_read_only_property),function(n,t){var r=131075;!W||t.flags&r||32768&um(t)||lr(n,e.Diagnostics.The_operand_of_a_delete_operator_must_be_optional)}(t,lo(r))),Ue}function uh(e){return dh(e,2112)?ph(e,3)||dh(e,296)?Qe:Me:Pe}function dh(e,n){if(e.flags&n)return!0;if(3145728&e.flags)for(var t=0,r=e.types;t<r.length;t++){if(dh(r[t],n))return!0}return!1}function ph(e,n,t){return!!(e.flags&n)||!(t&&114691&e.flags)&&(!!(296&n)&&zd(e,Pe)||!!(2112&n)&&zd(e,Me)||!!(402653316&n)&&zd(e,Oe)||!!(528&n)&&zd(e,Ue)||!!(16384&n)&&zd(e,je)||!!(131072&n)&&zd(e,He)||!!(65536&n)&&zd(e,Re)||!!(32768&n)&&zd(e,Ce)||!!(4096&n)&&zd(e,Ve)||!!(67108864&n)&&zd(e,Je))}function mh(n,t,r){return 1048576&n.flags?e.every(n.types,(function(e){return mh(e,t,r)})):ph(n,t,r)}function fh(n){return!!(16&e.getObjectFlags(n))&&!!n.symbol&&_h(n.symbol)}function _h(e){return 0!=(128&e.flags)}function gh(n,t,r,a,i){void 0===i&&(i=!1);var o=n.properties,s=o[r];if(288===s.kind||289===s.kind){var c=s.name,l=ou(c);if(qo(l)){var u=cc(t,Zo(l));u&&(hy(u,s,i),qg(s,!1,t,u))}var d=Ci(s,Iu(t,l,void 0,c,void 0,void 0,16));return vh(289===s.kind?s:s.initializer,d)}if(290===s.kind){if(!(r<o.length-1)){j<99&&cS(s,4);var p=[];if(a)for(var m=0,f=a;m<f.length;m++){var _=f[m];e.isSpreadAssignment(_)||p.push(_.name)}d=Di(t,p,t.symbol);return pS(a,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),vh(s.expression,d)}lr(s,e.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern)}else lr(s,e.Diagnostics.Property_assignment_expected)}function yh(n,t,r,a,i){var o=n.elements,s=o[r];if(222!==s.kind){if(220!==s.kind){var c=td(r);if(Kp(t)){var l=16|(pg(s)?8:0),u=Ru(t,c,void 0,Wy(s,c),l)||xe;return vh(s,Ci(s,pg(s)?If(u,524288):u),i)}return vh(s,a,i)}if(r<o.length-1)lr(s,e.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);else{var d=s.expression;if(216!==d.kind||62!==d.operatorToken.kind)return pS(n.elements,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),vh(d,qf(t,tm)?Yf(t,(function(e){return Vl(e,r)})):Nl(a),i);lr(d.operatorToken,e.Diagnostics.A_rest_element_cannot_have_an_initializer)}}}function vh(n,t,r,a){var i;if(289===n.kind){var o=n;o.objectAssignmentInitializer&&(!W||32768&um(Hh(o.objectAssignmentInitializer))||(t=If(t,524288)),function(e,n,t,r,a){var i,o=n.kind;if(62===o&&(200===e.kind||199===e.kind))return vh(e,Hh(t,r),r,107===t.kind);i=55===o||56===o||60===o?eE(e,r):Hh(e,r);var s=Hh(t,r);Sh(e,n,t,i,s,a)}(o.name,o.equalsToken,o.objectAssignmentInitializer,r)),i=n.name}else i=n;return 216===i.kind&&62===i.operatorToken.kind&&(Eh(i,r),i=i.left),200===i.kind?function(e,n,t){var r=e.properties;if(W&&0===r.length)return ny(n,e);for(var a=0;a<r.length;a++)gh(e,n,a,r,t);return n}(i,t,a):199===i.kind?function(e,n,t){var r=e.elements;j<2&&V.downlevelIteration&&cS(e,512);for(var a=aE(193,n,Ce,e)||xe,i=V.noUncheckedIndexedAccess?void 0:a,o=0;o<r.length;o++){var s=a;220===e.elements[o].kind&&(s=i=null!=i?i:aE(65,n,Ce,e)||xe),yh(e,n,o,s,t)}return n}(i,t,r):function(n,t,r){var a=Hh(n,r),i=290===n.parent.kind?e.Diagnostics.The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access:e.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access,o=290===n.parent.kind?e.Diagnostics.The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access:e.Diagnostics.The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access;ch(n,i,o)&&Qd(t,a,n,n);e.isPrivateIdentifierPropertyAccessExpression(n)&&cS(n.parent,2097152);return t}(i,t,r)}function hh(n){switch((n=e.skipParentheses(n)).kind){case 78:case 10:case 13:case 205:case 218:case 14:case 8:case 9:case 109:case 94:case 103:case 150:case 208:case 221:case 209:case 199:case 200:case 211:case 225:case 274:case 273:return!0;case 217:return hh(n.whenTrue)&&hh(n.whenFalse);case 216:return!e.isAssignmentOperator(n.operatorToken.kind)&&(hh(n.left)&&hh(n.right));case 214:case 215:switch(n.operator){case 53:case 39:case 40:case 54:return!0}return!1;case 212:case 206:case 224:default:return!1}}function bh(e,n){return 0!=(98304&n.flags)||Jd(e,n)}function Eh(n,t){for(var r,a={expr:[n],state:[0],leftType:[void 0]},i=0;i>=0;)switch(n=a.expr[i],a.state[i]){case 0:if(e.isInJSFile(n)&&e.getAssignedExpandoInitializer(n)){l(Hh(n.right,t));break}if(Th(n),62===(o=n.operatorToken.kind)&&(200===n.left.kind||199===n.left.kind)){l(vh(n.left,Hh(n.right,t),t,107===n.right.kind));break}u(1),d(n.left);break;case 1:var o,s=r;a.leftType[i]=s,55!==(o=n.operatorToken.kind)&&56!==o&&60!==o||$b(s,n.left),u(2),d(n.right);break;case 2:s=a.leftType[i];var c=r;l(Sh(n.left,n.operatorToken,n.right,s,c,n));break;default:return e.Debug.fail("Invalid state "+a.state[i]+" for checkBinaryExpression")}return r;function l(e){r=e,i--}function u(e){a.state[i]=e}function d(n){e.isBinaryExpression(n)?(i++,a.expr[i]=n,a.state[i]=0,a.leftType[i]=void 0):r=Hh(n,t)}}function Th(n){var t=n.left,r=n.operatorToken,a=n.right;60===r.kind&&(!e.isBinaryExpression(t)||56!==t.operatorToken.kind&&55!==t.operatorToken.kind||FS(t,e.Diagnostics._0_and_1_operations_cannot_be_mixed_without_parentheses,e.tokenToString(t.operatorToken.kind),e.tokenToString(r.kind)),!e.isBinaryExpression(a)||56!==a.operatorToken.kind&&55!==a.operatorToken.kind||FS(a,e.Diagnostics._0_and_1_operations_cannot_be_mixed_without_parentheses,e.tokenToString(a.operatorToken.kind),e.tokenToString(r.kind)))}function Sh(n,r,a,i,o,s){var c,l,u=r.kind;switch(u){case 41:case 42:case 65:case 66:case 43:case 67:case 44:case 68:case 40:case 64:case 47:case 69:case 48:case 70:case 49:case 71:case 51:case 73:case 52:case 77:case 50:case 72:if(i===Ke||o===Ke)return Ke;i=ny(i,n),o=ny(o,a);var d=void 0;if(528&i.flags&&528&o.flags&&void 0!==(d=function(e){switch(e){case 51:case 73:return 56;case 52:case 77:return 37;case 50:case 72:return 55;default:return}}(r.kind)))return lr(s||r,e.Diagnostics.The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead,e.tokenToString(r.kind),e.tokenToString(d)),Pe;var p,m=ah(n,i,e.Diagnostics.The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type,!0),f=ah(a,o,e.Diagnostics.The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type,!0);if(ph(i,3)&&ph(o,3)||!dh(i,2112)&&!dh(o,2112))p=Pe;else if(x(i,o)){switch(u){case 49:case 71:C();break;case 42:case 66:j<3&&lr(s,e.Diagnostics.Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_later)}p=Me}else C(x),p=xe;return m&&f&&A(p),p;case 39:case 63:if(i===Ke||o===Ke)return Ke;ph(i,402653316)||ph(o,402653316)||(i=ny(i,n),o=ny(o,a));var _=void 0;if(ph(i,296,!0)&&ph(o,296,!0)?_=Pe:ph(i,2112,!0)&&ph(o,2112,!0)?_=Me:ph(i,402653316,!0)||ph(o,402653316,!0)?_=Oe:(Li(i)||Li(o))&&(_=i===xe||o===xe?xe:Ee),_&&!L(u))return _;if(!_){var g=402655727;return C((function(e,n){return ph(e,g)&&ph(n,g)})),Ee}return 63===u&&A(_),_;case 29:case 31:case 32:case 33:return L(u)&&(i=Qp(ny(i,n)),o=Qp(ny(o,a)),D((function(e,n){return Jd(e,n)||Jd(n,e)||zd(e,Qe)&&zd(n,Qe)}))),Ue;case 34:case 35:case 36:case 37:return D((function(e,n){return bh(e,n)||bh(n,e)})),Ue;case 101:return function(n,t,r,a){return r===Ke||a===Ke?Ke:(!Li(r)&&mh(r,131068)&&lr(n,e.Diagnostics.The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter),Li(a)||DT(a)||Wd(a,vn)||lr(t,e.Diagnostics.The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type),Ue)}(n,a,i,o);case 100:return function(n,t,r,a){return r===Ke||a===Ke?Ke:(r=ny(r,n),a=ny(a,t),mh(r,402665900)||ph(r,407109632)||lr(n,e.Diagnostics.The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol),mh(a,126091264)||lr(t,e.Diagnostics.The_right_hand_side_of_an_in_expression_must_be_of_type_any_an_object_type_or_a_type_parameter),Ue)}(n,a,i,o);case 55:case 75:var y=4194304&Cf(i)?Xl([(l=W?i:Qp(o),Yf(l,pm)),o]):i;return 75===u&&A(o),y;case 56:case 74:var v=8388608&Cf(i)?Xl([dm(i),o],2):i;return 74===u&&A(o),v;case 60:case 76:var h=262144&Cf(i)?Xl([_m(i),o],2):i;return 76===u&&A(o),h;case 62:var b=e.isBinaryExpression(n.parent)?e.getAssignmentDeclarationKind(n.parent):0;return function(n,t){if(2===n)for(var r=0,a=Ms(t);r<a.length;r++){var i=a[r],o=lo(i);if(o.symbol&&32&o.symbol.flags){var s=i.escapedName,c=Cr(i.valueDeclaration,s,788968,void 0,s,!1);c&&c.declarations.some(e.isJSDocTypedefTag)&&(vr(c,e.Diagnostics.Duplicate_identifier_0,e.unescapeLeadingUnderscores(s),i),vr(i,e.Diagnostics.Duplicate_identifier_0,e.unescapeLeadingUnderscores(s),c))}}}(b,o),function(t){var r;switch(t){case 2:return!0;case 1:case 5:case 6:case 3:case 4:var i=xa(n),o=e.getAssignedExpandoInitializer(a);return!!o&&e.isObjectLiteralExpression(o)&&!!(null===(r=null==i?void 0:i.exports)||void 0===r?void 0:r.size);default:return!1}}(b)?(524288&o.flags&&(2===b||6===b||up(o)||Df(o)||1&e.getObjectFlags(o))||A(o),i):(A(o),Sm(o));case 27:if(!V.allowUnreachableCode&&hh(n)&&(78!==(c=a).kind||"eval"!==c.escapedText)){var E=e.getSourceFileOfNode(n),T=E.text,S=e.skipTrivia(T,n.pos);E.parseDiagnostics.some((function(n){return n.code===e.Diagnostics.JSX_expressions_must_have_one_parent_element.code&&e.textSpanContainsPosition(n,S)}))||lr(n,e.Diagnostics.Left_side_of_comma_operator_is_unused_and_has_no_side_effects)}return o;default:return e.Debug.fail()}function x(e,n){return ph(e,2112)&&ph(n,2112)}function L(t){var r=dh(i,12288)?n:dh(o,12288)?a:void 0;return!r||(lr(r,e.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol,e.tokenToString(t)),!1)}function A(r){t&&e.isAssignmentOperator(u)&&(!ch(n,e.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access,e.Diagnostics.The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access)||e.isIdentifier(n)&&"exports"===e.unescapeLeadingUnderscores(n.escapedText)||Qd(r,i,n,a))}function D(e){return!e(i,o)&&(C(e),!0)}function C(n){var t,a=!1,c=s||r;if(n){var l=mb(i),u=mb(o);a=!(l===i&&u===o)&&!(!l||!u)&&n(l,u)}var d=i,p=o;!a&&n&&(t=function(e,n,t){var r=e,a=n,i=Qp(e),o=Qp(n);t(i,o)||(r=i,a=o);return[r,a]}(i,o,n),d=t[0],p=t[1]);var m=oi(d,p),f=m[0],_=m[1];(function(n,t,a,i){var o;switch(r.kind){case 36:case 34:o="false";break;case 37:case 35:o="true"}if(o)return pr(n,t,e.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap,o,a,i);return})(c,a,f,_)||pr(c,a,e.Diagnostics.Operator_0_cannot_be_applied_to_types_1_and_2,e.tokenToString(r.kind),f,_)}}function xh(n,t,r,a){var i=function(n){return 281!==n.kind||e.isJsxSelfClosingElement(n.parent)?n:n.parent.parent}(n),o=i.contextualType,s=i.inferenceContext;try{i.contextualType=t,i.inferenceContext=r;var c=Hh(n,1|a|(r?2:0));return dh(c,2944)&&Ih(c,tg(t,n))?ed(c):c}finally{i.contextualType=o,i.inferenceContext=s}}function Lh(e,n){var t=Sr(e);if(!t.resolvedType){if(n&&0!==n)return Hh(e,n);var r=St,a=tt;St=xt,tt=void 0,t.resolvedType=Hh(e,n),tt=a,St=r}return t.resolvedType}function Ah(n){return 206===(n=e.skipParentheses(n)).kind||224===n.kind}function Dh(n,t){var r=e.getEffectiveInitializer(n),a=Vh(r)||(t?xh(r,t,void 0,0):Lh(r));return e.isParameter(n)&&197===n.name.kind&&tm(a)&&!a.target.hasRestElement&&tl(a)<n.name.elements.length?function(n,t){for(var r=t.elements,a=nl(n).slice(),i=n.target.elementFlags.slice(),o=tl(n);o<r.length;o++){var s=r[o];(o<r.length-1||198!==s.kind||!s.dotDotDotToken)&&(a.push(!e.isOmittedExpression(s)&&pg(s)?zi(s,!1,!1):Ee),i.push(2),e.isOmittedExpression(s)||pg(s)||km(s,Ee))}return Fl(a,i,n.target.readonly)}(a,n.name):a}function Ch(n,t){var r=2&e.getCombinedNodeFlags(n)||e.isDeclarationReadonly(n)?t:Zp(t);if(e.isInJSFile(n)){if(98304&r.flags)return km(n,Ee),Ee;if(Wp(r))return km(n,Cn),Cn}return r}function Ih(n,t){if(t){if(3145728&t.flags){var r=t.types;return e.some(r,(function(e){return Ih(n,e)}))}if(58982400&t.flags){var a=Ws(t)||De;return dh(a,4)&&dh(n,128)||dh(a,8)&&dh(n,256)||dh(a,64)&&dh(n,2048)||dh(a,4096)&&dh(n,8192)||Ih(n,a)}return!!(406847616&t.flags&&dh(n,128)||256&t.flags&&dh(n,256)||2048&t.flags&&dh(n,2048)||512&t.flags&&dh(n,512)||8192&t.flags&&dh(n,8192))}return!1}function Nh(n){var t=n.parent;return e.isAssertionExpression(t)&&e.isConstTypeReference(t.type)||(e.isParenthesizedExpression(t)||e.isArrayLiteralExpression(t)||e.isSpreadElement(t))&&Nh(t)||(e.isPropertyAssignment(t)||e.isShorthandPropertyAssignment(t)||e.isTemplateSpan(t))&&Nh(t.parent)}function Rh(e,n,t,r){var a=Hh(e,n,r);return Nh(e)?ed(a):Ah(e)?a:em(a,tg(2===arguments.length?ag(e):t,e))}function kh(e,n){return 158===e.name.kind&&vg(e.name),Rh(e.initializer,n)}function Oh(e,n){return DS(e),158===e.name.kind&&vg(e.name),Ph(e,rh(e,n),n)}function Ph(n,t,r){if(r&&10&r){var a=wy(t,0,!0),i=wy(t,1,!0),o=a||i;if(o&&o.typeParameters){var s=ng(n,2);if(s){var c=wy(_m(s),a?0:1,!1);if(c&&!c.typeParameters){if(8&r)return Mh(n,r),sn;var l=ig(n),u=l.signature&&kc(l.signature),d=u&&My(u);if(d&&!d.typeParameters&&!e.every(l.inferences,wh)){var p=function(n,t){for(var r,a,i=[],o=0,s=t;o<s.length;o++){var c=(m=s[o]).symbol.escapedName;if(Fh(n.inferredTypeParameters,c)||Fh(i,c)){var l=Fa(mr(262144,Gh(e.concatenate(n.inferredTypeParameters,i),c)));l.target=m,r=e.append(r,m),a=e.append(a,l),i.push(l)}else i.push(m)}if(a)for(var u=pd(r,a),d=0,p=a;d<p.length;d++){var m;(m=p[d]).mapper=u}return i}(l,o.typeParameters),m=Fc(o,p),f=e.map(l.inferences,(function(e){return Um(e.typeParameter)}));if(Pm(m,c,(function(e,n){tf(f,e,n,0,!0)})),e.some(f,wh)&&(Mm(m,c,(function(e,n){tf(f,e,n)})),!function(e,n){for(var t=0;t<e.length;t++)if(wh(e[t])&&wh(n[t]))return!0;return!1}(l.inferences,f)))return function(e,n){for(var t=0;t<e.length;t++)!wh(e[t])&&wh(n[t])&&(e[t]=n[t])}(l.inferences,f),l.inferredTypeParameters=e.concatenate(l.inferredTypeParameters,p),jc(m)}return jc(Fy(o,c,l))}}}}return t}function Mh(e,n){2&n&&(ig(e).flags|=4)}function wh(e){return!(!e.candidates&&!e.contraCandidates)}function Fh(n,t){return e.some(n,(function(e){return e.symbol.escapedName===t}))}function Gh(e,n){for(var t=n.length;t>1&&n.charCodeAt(t-1)>=48&&n.charCodeAt(t-1)<=57;)t--;for(var r=n.slice(0,t),a=1;;a++){var i=r+a;if(!Fh(e,i))return i}}function Bh(e){var n=Py(e);if(n&&!n.typeParameters)return kc(n)}function Uh(n){var t=Vh(n);if(t)return t;if(67108864&n.flags&&tt){var r=tt[R(n)];if(r)return r}var a=Dt,i=Hh(n);Dt!==a&&((tt||(tt=[]))[R(n)]=i,e.setNodeFlags(n,67108864|n.flags));return i}function Vh(n){var t=e.skipParentheses(n);if(!e.isCallExpression(t)||105===t.expression.kind||e.isRequireCall(t,!0)||hv(t)){if(e.isAssertionExpression(t)&&!e.isConstTypeReference(t.type))return sd(t.type);if(8===n.kind||10===n.kind||109===n.kind||94===n.kind)return Hh(n)}else{var r=e.isCallChain(t)?function(e){var n=Hh(e.expression),t=bm(n,e.expression),r=Bh(n);return r&&hm(r,e,t!==n)}(t):Bh(Xg(t.expression));if(r)return r}}function jh(e){var n=Sr(e);if(n.contextFreeType)return n.contextFreeType;var t=e.contextualType;e.contextualType=Ee;try{return n.contextFreeType=Hh(e,4)}finally{e.contextualType=t}}function Hh(n,a,i){e.tracing.push("check","checkExpression",{kind:n.kind,pos:n.pos,end:n.end});var o=c;c=n,h=0;var s=Ph(n,function(n,a,i){var o=n.kind;if(r)switch(o){case 221:case 208:case 209:r.throwIfCancellationRequested()}switch(o){case 78:return A_(n);case 107:return R_(n);case 105:return P_(n);case 103:return ke;case 14:case 10:return $u(td(n.text));case 8:return US(n),$u(td(+n.text));case 9:return function(n){if(!(e.isLiteralTypeNode(n.parent)||e.isPrefixUnaryExpression(n.parent)&&e.isLiteralTypeNode(n.parent.parent))&&j<7&&FS(n,e.Diagnostics.BigInt_literals_are_not_available_when_targeting_lower_than_ES2020))return!0}(n),$u(function(n){return td({negative:!1,base10Value:e.parsePseudoBigInt(n.text)})}(n));case 109:return Ge;case 94:return we;case 218:return function(n){for(var t=[n.head.text],r=[],a=0,i=n.templateSpans;a<i.length;a++){var o=i[a],s=Hh(o.expression);dh(s,12288)&&lr(o.expression,e.Diagnostics.Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String),t.push(o.literal.text),r.push(zd(s,Ze)?s:Oe)}return Nh(n)?pu(t,r):Oe}(n);case 13:return An;case 199:return mg(n,a,i);case 200:return Eg(n,a);case 201:return ry(n);case 157:return ay(n);case 202:return Ty(n);case 203:if(99===n.expression.kind)return bv(n);case 204:return function(n,t){var r;gS(n,n.typeArguments)||yS(n.arguments);var a=mv(n,void 0,t);if(a===ut)return We;if(yv(a,n),105===n.expression.kind)return je;if(204===n.kind){var i=a.declaration;if(i&&166!==i.kind&&170!==i.kind&&175!==i.kind&&!e.isJSDocConstructSignature(i)&&!fv(i))return X&&lr(n,e.Diagnostics.new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type),Ee}if(e.isInJSFile(n)&&Tv(n))return Ic(n.arguments[0]);var o=kc(a);if(12288&o.flags&&hv(n))return rd(e.walkUpParenthesizedExpressions(n.parent));if(203===n.kind&&233===n.parent.kind&&16384&o.flags&&Rc(a))if(e.isDottedName(n.expression)){if(!d_(n)){var s=lr(n.expression,e.Diagnostics.Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation);u_(n.expression,s)}}else lr(n.expression,e.Diagnostics.Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name);if(e.isInJSFile(n)){var c=gv(n,!1);if(null===(r=null==c?void 0:c.exports)||void 0===r?void 0:r.size){var l=Va(c,c.exports,e.emptyArray,e.emptyArray,void 0,void 0);return l.objectFlags|=16384,tu([o,l])}}return o}(n,a);case 205:return Sv(n);case 207:return function(n,t){var r=e.isInJSFile(n)?e.getJSDocTypeTag(n):void 0;if(r)return Lv(r,r.typeExpression.type,n.expression,t);return Hh(n.expression,t)}(n,a);case 221:return function(e){return GE(e),sT(e),lo(xa(e))}(n);case 208:case 209:return rh(n,a);case 211:return function(e){return Hh(e.expression),$t}(n);case 206:case 224:return function(e){return Lv(e,e.type,e.expression)}(n);case 225:return Av(n);case 226:return Dv(n);case 210:return lh(n);case 212:return function(e){return Hh(e.expression),Ie}(n);case 213:return function(n){if(t){var r;if(!(32768&n.flags))if(e.isInTopLevelContext(n)){if(!PS(r=e.getSourceFileOfNode(n))){var a=void 0;if(!e.isEffectiveExternalModule(r,V)){a||(a=e.getSpanOfTokenAtPosition(r,n.pos));var i=e.createFileDiagnostic(r,a.start,a.length,e.Diagnostics.await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module);Yt.add(i)}(H!==e.ModuleKind.ESNext&&H!==e.ModuleKind.System||j<4)&&(a=e.getSpanOfTokenAtPosition(r,n.pos),i=e.createFileDiagnostic(r,a.start,a.length,e.Diagnostics.Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher),Yt.add(i))}}else if(!PS(r=e.getSourceFileOfNode(n))){a=e.getSpanOfTokenAtPosition(r,n.pos),i=e.createFileDiagnostic(r,a.start,a.length,e.Diagnostics.await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules);var o=e.getContainingFunction(n);if(o&&166!==o.kind&&0==(2&e.getFunctionFlags(o))){var s=e.createDiagnosticForNode(o,e.Diagnostics.Did_you_mean_to_mark_this_function_as_async);e.addRelatedInfo(i,s)}Yt.add(i)}U_(n)&&lr(n,e.Diagnostics.await_expressions_cannot_be_used_in_a_parameter_initializer)}var c=Hh(n.expression),l=pb(c,n,e.Diagnostics.Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);return l!==c||l===xe||3&c.flags||ur(!1,e.createDiagnosticForNode(n,e.Diagnostics.await_has_no_effect_on_the_type_of_this_expression)),l}(n);case 214:return function(n){var t=Hh(n.operand);if(t===Ke)return Ke;switch(n.operand.kind){case 8:switch(n.operator){case 40:return $u(td(-n.operand.text));case 39:return $u(td(+n.operand.text))}break;case 9:if(40===n.operator)return $u(td({negative:!0,base10Value:e.parsePseudoBigInt(n.operand.text)}))}switch(n.operator){case 39:case 40:case 54:return ny(t,n.operand),dh(t,12288)&&lr(n.operand,e.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol,e.tokenToString(n.operator)),39===n.operator?(dh(t,2112)&&lr(n.operand,e.Diagnostics.Operator_0_cannot_be_applied_to_type_1,e.tokenToString(n.operator),ii(Qp(t))),Pe):uh(t);case 53:eE(n.operand);var r=12582912&Cf(t);return 4194304===r?we:8388608===r?Ge:Ue;case 45:case 46:return ah(n.operand,ny(t,n.operand),e.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type)&&ch(n.operand,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access),uh(t)}return xe}(n);case 215:return function(n){var t=Hh(n.operand);return t===Ke?Ke:(ah(n.operand,ny(t,n.operand),e.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type)&&ch(n.operand,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access,e.Diagnostics.The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access),uh(t))}(n);case 216:return Eh(n,a);case 217:return function(e,n){var t=eE(e.condition);return Zb(e.condition,e.whenTrue,t),Xl([Hh(e.whenTrue,n),Hh(e.whenFalse,n)],2)}(n,a);case 220:return function(e,n){return j<2&&cS(e,V.downlevelIteration?1536:2048),aE(33,Hh(e.expression,n),Ce,e.expression)}(n,a);case 222:return Ie;case 219:return function(n){t&&(8192&n.flags||MS(n,e.Diagnostics.A_yield_expression_is_only_allowed_in_a_generator_body),U_(n)&&lr(n,e.Diagnostics.yield_expressions_cannot_be_used_in_a_parameter_initializer));var r=e.getContainingFunction(n);if(!r)return Ee;var a=e.getFunctionFlags(r);if(!(1&a))return Ee;var i=0!=(2&a);n.asteriskToken&&(i&&j<99&&cS(n,53248),!i&&j<2&&V.downlevelIteration&&cS(n,256));var o=Oc(r),s=o&&CE(o,i),c=s&&s.yieldType||Ee,l=s&&s.nextType||Ee,u=i?mb(l)||Ee:l,d=n.expression?Hh(n.expression):Ie,p=Qv(n,d,u,i);return o&&p&&Qd(p,c,n.expression||n,n.expression),n.asteriskToken?oE(i?19:17,1,d,n.expression)||Ee:o?DE(2,o,i)||Ee:V_(2,r)||Ee}(n);case 227:return function(e){return e.isSpread?Iu(e.type,Pe):e.type}(n);case 283:return Kg(n,a);case 273:case 274:return function(e,n){return sT(e),Bg(e)||Ee}(n);case 277:return function(n){Vg(n.openingFragment);var t=e.getSourceFileOfNode(n);return!e.getJSXTransformEnabled(V)||!V.jsxFactory&&!t.pragmas.has("jsx")||V.jsxFragmentFactory||t.pragmas.has("jsxfrag")||lr(n,V.jsxFactory?e.Diagnostics.The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_compiler_option:e.Diagnostics.An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments),Ag(n),Bg(n)||Ee}(n);case 281:return Cg(n,a);case 275:e.Debug.fail("Shouldn't ever directly check a JsxOpeningElement")}return xe}(n,a,i),a);return fh(s)&&function(n,t){201===n.parent.kind&&n.parent.expression===n||202===n.parent.kind&&n.parent.expression===n||(78===n.kind||157===n.kind)&&vT(n)||176===n.parent.kind&&n.parent.exprName===n||270===n.parent.kind||lr(n,e.Diagnostics.const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query);if(V.isolatedModules){e.Debug.assert(!!(128&t.symbol.flags)),8388608&t.symbol.valueDeclaration.flags&&lr(n,e.Diagnostics.Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided)}}(n,s),c=o,e.tracing.pop(),s}function Kh(n){n.expression&&MS(n.expression,e.Diagnostics.Type_expected),iT(n.constraint),iT(n.default);var r=wo(xa(n));Ws(r),function(e){return Xs(e)!==ln}(r)||lr(n.default,e.Diagnostics.Type_parameter_0_has_a_circular_default,ii(r));var a=Us(r),i=Ys(r);a&&i&&Yd(i,rs(kd(a,fd(r,i)),i),n.default,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1),t&&PE(n.name,e.Diagnostics.Type_parameter_name_cannot_be_0)}function Wh(n){uS(n),zb(n);var t=e.getContainingFunction(n);e.hasSyntacticModifier(n,92)&&(166===t.kind&&e.nodeIsPresent(t.body)||lr(n,e.Diagnostics.A_parameter_property_is_only_allowed_in_a_constructor_implementation),166===t.kind&&e.isIdentifier(n.name)&&"constructor"===n.name.escapedText&&lr(n.name,e.Diagnostics.constructor_cannot_be_used_as_a_parameter_property_name)),n.questionToken&&e.isBindingPattern(n.name)&&t.body&&lr(n,e.Diagnostics.A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature),n.name&&e.isIdentifier(n.name)&&("this"===n.name.escapedText||"new"===n.name.escapedText)&&(0!==t.parameters.indexOf(n)&&lr(n,e.Diagnostics.A_0_parameter_must_be_the_first_parameter,n.name.escapedText),166!==t.kind&&170!==t.kind&&175!==t.kind||lr(n,e.Diagnostics.A_constructor_cannot_have_a_this_parameter),209===t.kind&&lr(n,e.Diagnostics.An_arrow_function_cannot_have_a_this_parameter),167!==t.kind&&168!==t.kind||lr(n,e.Diagnostics.get_and_set_accessors_cannot_declare_this_parameters)),!n.dotDotDotToken||e.isBindingPattern(n.name)||zd(rc(lo(n.symbol)),Nn)||lr(n,e.Diagnostics.A_rest_parameter_must_be_of_an_array_type)}function zh(n,t,r){for(var a=0,i=n.elements;a<i.length;a++){var o=i[a];if(!e.isOmittedExpression(o)){var s=o.name;if(78===s.kind&&s.escapedText===r)return lr(t,e.Diagnostics.A_type_predicate_cannot_reference_element_0_in_a_binding_pattern,r),!0;if((197===s.kind||196===s.kind)&&zh(s,t,r))return!0}}}function qh(n){171===n.kind?function(n){uS(n)||function(n){var t=n.parameters[0];if(1!==n.parameters.length)return FS(t?t.name:n,e.Diagnostics.An_index_signature_must_have_exactly_one_parameter);if(pS(n.parameters,e.Diagnostics.An_index_signature_cannot_have_a_trailing_comma),t.dotDotDotToken)return FS(t.dotDotDotToken,e.Diagnostics.An_index_signature_cannot_have_a_rest_parameter);if(e.hasEffectiveModifiers(t))return FS(t.name,e.Diagnostics.An_index_signature_parameter_cannot_have_an_accessibility_modifier);if(t.questionToken)return FS(t.questionToken,e.Diagnostics.An_index_signature_parameter_cannot_have_a_question_mark);if(t.initializer)return FS(t.name,e.Diagnostics.An_index_signature_parameter_cannot_have_an_initializer);if(!t.type)return FS(t.name,e.Diagnostics.An_index_signature_parameter_must_have_a_type_annotation);if(147!==t.type.kind&&144!==t.type.kind){var r=sd(t.type);return 4&r.flags||8&r.flags?FS(t.name,e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_type_alias_Consider_writing_0_Colon_1_Colon_2_instead,e.getTextOfNode(t.name),ii(r),ii(n.type?sd(n.type):Ee)):1048576&r.flags&&mh(r,384,!0)?FS(t.name,e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead):FS(t.name,e.Diagnostics.An_index_signature_parameter_type_must_be_either_string_or_number)}if(!n.type)return FS(n,e.Diagnostics.An_index_signature_must_have_a_type_annotation)}(n)}(n):174!==n.kind&&251!==n.kind&&175!==n.kind&&169!==n.kind&&166!==n.kind&&170!==n.kind||_S(n);var r=e.getFunctionFlags(n);if(4&r||(3==(3&r)&&j<99&&cS(n,12288),2==(3&r)&&j<4&&cS(n,64),0!=(3&r)&&j<2&&cS(n,128)),ME(n.typeParameters),e.forEach(n.parameters,Wh),n.type&&iT(n.type),t){!function(n){if(j>=2||!e.hasRestParameter(n)||8388608&n.flags||e.nodeIsMissing(n.body))return;e.forEach(n.parameters,(function(n){n.name&&!e.isBindingPattern(n.name)&&n.name.escapedText===ie.escapedName&&cr("noEmit",n,e.Diagnostics.Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters)}))}(n);var a=e.getEffectiveReturnTypeNode(n);if(X&&!a)switch(n.kind){case 170:lr(n,e.Diagnostics.Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type);break;case 169:lr(n,e.Diagnostics.Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type)}if(a){var i=e.getFunctionFlags(n);if(1==(5&i)){var o=sd(a);if(o===je)lr(a,e.Diagnostics.A_generator_cannot_have_a_void_type_annotation);else{var s=DE(0,o,0!=(2&i))||Ee;Yd(Yv(s,DE(1,o,0!=(2&i))||s,DE(2,o,0!=(2&i))||De,!!(2&i)),o,a)}}else 2==(3&i)&&function(n,t){var r=sd(t);if(j>=2){if(r===xe)return;var a=xl(!0);if(a!==on&&!uo(r,a))return void lr(t,e.Diagnostics.The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0,ii(mb(r)||je))}else{if(function(n){gb(n&&e.getEntityNameFromTypeNode(n))}(t),r===xe)return;var i=e.getEntityNameFromTypeNode(t);if(void 0===i)return void lr(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,ii(r));var o=ca(i,111551,!0),s=o?lo(o):xe;if(s===xe)return void(78===i.kind&&"Promise"===i.escapedText&&po(r)===xl(!1)?lr(t,e.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option):lr(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,e.entityNameToString(i)));var c=(d=!0,Gn||(Gn=El("PromiseConstructorLike",0,d))||nn);if(c===nn)return void lr(t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value,e.entityNameToString(i));if(!Yd(s,c,t,e.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value))return;var l=i&&e.getFirstIdentifier(i),u=Lr(n.locals,l.escapedText,111551);if(u)return void lr(u.valueDeclaration,e.Diagnostics.Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions,e.idText(l),e.entityNameToString(i))}var d;pb(r,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)}(n,a)}171!==n.kind&&308!==n.kind&&xb(n)}}function Jh(n){for(var t=new e.Map,r=0,a=n.members;r<a.length;r++){var i=a[r];if(162===i.kind){var o=void 0,s=i.name;switch(s.kind){case 10:case 8:o=s.text;break;case 78:o=e.idText(s);break;default:continue}t.get(o)?(lr(e.getNameOfDeclaration(i.symbol.valueDeclaration),e.Diagnostics.Duplicate_identifier_0,o),lr(i.name,e.Diagnostics.Duplicate_identifier_0,o)):t.set(o,!0)}}}function Xh(n){if(253===n.kind){var t=xa(n);if(t.declarations.length>0&&t.declarations[0]!==n)return}var r=Hc(xa(n));if(r)for(var a=!1,i=!1,o=0,s=r.declarations;o<s.length;o++){var c=s[o];if(1===c.parameters.length&&c.parameters[0].type)switch(c.parameters[0].type.kind){case 147:i?lr(c,e.Diagnostics.Duplicate_string_index_signature):i=!0;break;case 144:a?lr(c,e.Diagnostics.Duplicate_number_index_signature):a=!0}}}function Yh(n){if(uS(n)||function(n){if(e.isClassLike(n.parent)){if(e.isStringLiteral(n.name)&&"constructor"===n.name.text)return FS(n.name,e.Diagnostics.Classes_may_not_have_a_field_named_constructor);if(AS(n.name,e.Diagnostics.A_computed_property_name_in_a_class_property_declaration_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0;if(j<2&&e.isPrivateIdentifier(n.name))return FS(n.name,e.Diagnostics.Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher)}else if(253===n.parent.kind){if(AS(n.name,e.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0;if(n.initializer)return FS(n.initializer,e.Diagnostics.An_interface_property_cannot_have_an_initializer)}else if(177===n.parent.kind){if(AS(n.name,e.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type))return!0;if(n.initializer)return FS(n.initializer,e.Diagnostics.A_type_literal_property_cannot_have_an_initializer)}8388608&n.flags&&IS(n);if(e.isPropertyDeclaration(n)&&n.exclamationToken&&(!e.isClassLike(n.parent)||!n.type||n.initializer||8388608&n.flags||e.hasSyntacticModifier(n,160))){var t=n.initializer?e.Diagnostics.Declarations_with_initializers_cannot_also_have_definite_assignment_assertions:n.type?e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context:e.Diagnostics.Declarations_with_definite_assignment_assertions_must_also_have_type_annotations;return FS(n.exclamationToken,t)}}(n)||bS(n.name),zb(n),e.isPrivateIdentifier(n.name)&&j<99)for(var t=e.getEnclosingBlockScopeContainer(n);t;t=e.getEnclosingBlockScopeContainer(t))Sr(t).flags|=67108864}function Qh(n){qh(n),function(n){var t=e.isInJSFile(n)?e.getJSDocTypeParameterDeclarations(n):void 0,r=n.typeParameters||t&&e.firstOrUndefined(t);if(r){var a=r.pos===r.end?r.pos:e.skipTrivia(e.getSourceFileOfNode(n).text,r.pos);return wS(n,a,r.end-a,e.Diagnostics.Type_parameters_cannot_appear_on_a_constructor_declaration)}}(n)||function(n){var t=e.getEffectiveReturnTypeNode(n);if(t)FS(t,e.Diagnostics.Type_annotation_cannot_appear_on_a_constructor_declaration)}(n),iT(n.body);var r=xa(n);if(n===e.getDeclarationOfKind(r,n.kind)&&cb(r),!e.nodeIsMissing(n.body)&&t){var a=n.parent;if(e.getClassExtendsHeritageElement(a)){D_(n.parent,a);var i=I_(a),o=C_(n.body);if(o){if(i&&lr(o,e.Diagnostics.A_constructor_cannot_contain_a_super_call_when_its_class_extends_null),(99!==V.target||!V.useDefineForClassFields)&&(e.some(n.parent.members,(function(n){return!!e.isPrivateIdentifierPropertyDeclaration(n)||163===n.kind&&!e.hasSyntacticModifier(n,32)&&!!n.initializer}))||e.some(n.parameters,(function(n){return e.hasSyntacticModifier(n,92)})))){for(var s=void 0,c=0,l=n.body.statements;c<l.length;c++){var u=l[c];if(233===u.kind&&e.isSuperCall(u.expression)){s=u;break}if(!e.isPrologueDirective(u))break}s||lr(n,e.Diagnostics.A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_parameter_properties_or_private_identifiers)}}else i||lr(n,e.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call)}}}function Zh(n){if(t){if(_S(n)||function(n){if(!(8388608&n.flags)){if(j<1)return FS(n.name,e.Diagnostics.Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher);if(void 0===n.body&&!e.hasSyntacticModifier(n,128))return wS(n,n.end-1,";".length,e.Diagnostics._0_expected,"{")}if(n.body&&e.hasSyntacticModifier(n,128))return FS(n,e.Diagnostics.An_abstract_accessor_cannot_have_an_implementation);if(n.typeParameters)return FS(n.name,e.Diagnostics.An_accessor_cannot_have_type_parameters);if(!function(e){return LS(e)||e.parameters.length===(167===e.kind?0:1)}(n))return FS(n.name,167===n.kind?e.Diagnostics.A_get_accessor_cannot_have_parameters:e.Diagnostics.A_set_accessor_must_have_exactly_one_parameter);if(168===n.kind){if(n.type)return FS(n.name,e.Diagnostics.A_set_accessor_cannot_have_a_return_type_annotation);var t=e.Debug.checkDefined(e.getSetAccessorValueParameter(n),"Return value does not match parameter count assertion.");if(t.dotDotDotToken)return FS(t.dotDotDotToken,e.Diagnostics.A_set_accessor_cannot_have_rest_parameter);if(t.questionToken)return FS(t.questionToken,e.Diagnostics.A_set_accessor_cannot_have_an_optional_parameter);if(t.initializer)return FS(n.name,e.Diagnostics.A_set_accessor_parameter_cannot_have_an_initializer)}return!1}(n)||bS(n.name),Eb(n),qh(n),167===n.kind&&!(8388608&n.flags)&&e.nodeIsPresent(n.body)&&256&n.flags&&(512&n.flags||lr(n.name,e.Diagnostics.A_get_accessor_must_return_a_value)),158===n.name.kind&&vg(n.name),e.isPrivateIdentifier(n.name)&&lr(n.name,e.Diagnostics.An_accessor_cannot_be_named_with_a_private_identifier),!Qo(n)){var r=167===n.kind?168:167,a=e.getDeclarationOfKind(xa(n),r);if(a){var i=e.getEffectiveModifierFlags(n),o=e.getEffectiveModifierFlags(a);(28&i)!=(28&o)&&lr(n.name,e.Diagnostics.Getter_and_setter_accessors_do_not_agree_in_visibility),(128&i)!=(128&o)&&lr(n.name,e.Diagnostics.Accessors_must_both_be_abstract_or_non_abstract),$h(n,a,no,e.Diagnostics.get_and_set_accessor_must_have_the_same_type),$h(n,a,to,e.Diagnostics.get_and_set_accessor_must_have_the_same_this_type)}}var s=ro(xa(n));167===n.kind&&th(n,s)}iT(n.body)}function $h(e,n,t,r){var a=t(e),i=t(n);a&&i&&!Vd(a,i)&&lr(e,r)}function eb(n,t){return xc(e.map(n.typeArguments,sd),t,Sc(t),e.isInJSFile(n))}function nb(n,t){for(var r,a,i=!0,o=0;o<t.length;o++){var s=Us(t[o]);s&&(r||(a=pd(t,r=eb(n,t))),i=i&&Yd(r[o],kd(s,a),n.typeArguments[o],e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1))}return i}function tb(n){var t=_l(n);if(t!==xe){var r=Sr(n).resolvedSymbol;if(r)return 524288&r.flags&&Tr(r).typeParameters||(4&e.getObjectFlags(t)?t.target.localTypeParameters:void 0)}}function rb(n){gS(n,n.typeArguments),173!==n.kind||void 0===n.typeName.jsdocDotPos||e.isInJSFile(n)||e.isInJSDoc(n)||wS(n,n.typeName.jsdocDotPos,1,e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments),e.forEach(n.typeArguments,iT);var r=_l(n);if(r!==xe){if(n.typeArguments&&t){var a=tb(n);a&&nb(n,a)}var i=Sr(n).resolvedSymbol;i&&(e.some(i.declarations,(function(e){return fT(e)&&!!(134217728&e.flags)}))&&dr(!1,vv(n),e.Diagnostics._0_is_deprecated,i.escapedName),32&r.flags&&8&i.flags&&lr(n,e.Diagnostics.Enum_type_0_has_members_with_initializers_that_are_not_literals,ii(r)))}}function ab(n,t){if(!(8388608&n.flags))return n;var r=n.objectType,a=n.indexType;if(zd(a,uu(r,!1)))return 202===t.kind&&e.isAssignmentTarget(t)&&32&e.getObjectFlags(r)&&1&Ns(r)&&lr(t,e.Diagnostics.Index_signature_in_type_0_only_permits_reading,ii(r)),n;var i=Zs(r);if(mc(i,1)&&ph(a,296))return n;if(Tu(r)){var o=gu(a,t);if(o){var s=zf(i,(function(e){return cc(e,o)}));if(s&&24&e.getDeclarationModifierFlagsFromSymbol(s))return lr(t,e.Diagnostics.Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter,e.unescapeLeadingUnderscores(o)),xe}}return lr(t,e.Diagnostics.Type_0_cannot_be_used_to_index_type_1,ii(a),ii(r)),xe}function ib(n){!function(n){if(151===n.operator){if(148!==n.type.kind)return FS(n.type,e.Diagnostics._0_expected,e.tokenToString(148));var t=e.walkUpParenthesizedTypes(n.parent);switch(e.isInJSFile(t)&&e.isJSDocTypeExpression(t)&&(t=t.parent,e.isJSDocTypeTag(t)&&(t=t.parent.parent)),t.kind){case 249:var r=t;if(78!==r.name.kind)return FS(n,e.Diagnostics.unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name);if(!e.isVariableDeclarationInVariableStatement(r))return FS(n,e.Diagnostics.unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement);if(!(2&r.parent.flags))return FS(t.name,e.Diagnostics.A_variable_whose_type_is_a_unique_symbol_type_must_be_const);break;case 163:if(!e.hasSyntacticModifier(t,32)||!e.hasEffectiveModifier(t,64))return FS(t.name,e.Diagnostics.A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly);break;case 162:if(!e.hasSyntacticModifier(t,64))return FS(t.name,e.Diagnostics.A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly);break;default:FS(n,e.Diagnostics.unique_symbol_types_are_not_allowed_here)}}else if(142===n.operator&&178!==n.type.kind&&179!==n.type.kind)MS(n,e.Diagnostics.readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types,e.tokenToString(148))}(n),iT(n.type)}function ob(n){return(e.hasEffectiveModifier(n,8)||e.isPrivateIdentifierPropertyDeclaration(n))&&!!(8388608&n.flags)}function sb(n,t){var r=e.getCombinedModifierFlags(n);return 253!==n.parent.kind&&252!==n.parent.kind&&221!==n.parent.kind&&8388608&n.flags&&(2&r||e.isModuleBlock(n.parent)&&e.isModuleDeclaration(n.parent.parent)&&e.isGlobalScopeAugmentation(n.parent.parent)||(r|=1),r|=2),r&t}function cb(n){if(t){for(var r,a,i,o=0,s=155,c=!1,l=!0,u=!1,d=n.declarations,p=0!=(16384&n.flags),m=!1,f=!1,_=!1,g=[],y=0,v=d;y<v.length;y++){var h=v[y],b=8388608&h.flags,E=h.parent&&(253===h.parent.kind||177===h.parent.kind)||b;if(E&&(i=void 0),252!==h.kind&&221!==h.kind||b||(_=!0),251===h.kind||165===h.kind||164===h.kind||166===h.kind){g.push(h);var T=sb(h,155);o|=T,s&=T,c=c||e.hasQuestionToken(h),l=l&&e.hasQuestionToken(h);var S=e.nodeIsPresent(h.body);S&&r?p?f=!0:m=!0:(null==i?void 0:i.parent)===h.parent&&i.end!==h.pos&&N(i),S?r||(r=h):u=!0,i=h,E||(a=h)}}if(f&&e.forEach(g,(function(n){lr(n,e.Diagnostics.Multiple_constructor_implementations_are_not_allowed)})),m&&e.forEach(g,(function(n){lr(e.getNameOfDeclaration(n)||n,e.Diagnostics.Duplicate_function_implementation)})),_&&!p&&16&n.flags&&e.forEach(d,(function(t){hr(t,e.Diagnostics.Duplicate_identifier_0,e.symbolName(n),d)})),!a||a.body||e.hasSyntacticModifier(a,128)||a.questionToken||N(a),u&&(function(n,t,r,a,i){if(0!=(a^i)){var o=sb(I(n,t),r);e.forEach(n,(function(n){var t=sb(n,r)^o;1&t?lr(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_exported_or_non_exported):2&t?lr(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_ambient_or_non_ambient):24&t?lr(e.getNameOfDeclaration(n)||n,e.Diagnostics.Overload_signatures_must_all_be_public_private_or_protected):128&t&&lr(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_abstract_or_non_abstract)}))}}(d,r,155,o,s),function(n,t,r,a){if(r!==a){var i=e.hasQuestionToken(I(n,t));e.forEach(n,(function(n){e.hasQuestionToken(n)!==i&&lr(e.getNameOfDeclaration(n),e.Diagnostics.Overload_signatures_must_all_be_optional_or_required)}))}}(d,r,c,l),r))for(var x=Cc(n),L=Lc(r),A=0,D=x;A<D.length;A++){var C=D[A];if(!cp(L,C)){e.addRelatedInfo(lr(C.declaration,e.Diagnostics.This_overload_signature_is_not_compatible_with_its_implementation_signature),e.createDiagnosticForNode(r,e.Diagnostics.The_implementation_signature_is_declared_here));break}}}function I(e,n){return void 0!==n&&n.parent===e[0].parent?n:e[0]}function N(n){if(!n.name||!e.nodeIsMissing(n.name)){var t=!1,r=e.forEachChild(n.parent,(function(e){if(t)return e;t=e===n}));if(r&&r.pos===n.end&&r.kind===n.kind){var a=r.name||r,i=r.name;if(n.name&&i&&(e.isPrivateIdentifier(n.name)&&e.isPrivateIdentifier(i)&&n.name.escapedText===i.escapedText||e.isComputedPropertyName(n.name)&&e.isComputedPropertyName(i)||e.isPropertyNameLiteral(n.name)&&e.isPropertyNameLiteral(i)&&e.getEscapedTextOfIdentifierOrLiteral(n.name)===e.getEscapedTextOfIdentifierOrLiteral(i))){if((165===n.kind||164===n.kind)&&e.hasSyntacticModifier(n,32)!==e.hasSyntacticModifier(r,32))lr(a,e.hasSyntacticModifier(n,32)?e.Diagnostics.Function_overload_must_be_static:e.Diagnostics.Function_overload_must_not_be_static);return}if(e.nodeIsPresent(r.body))return void lr(a,e.Diagnostics.Function_implementation_name_must_be_0,e.declarationNameToString(n.name))}var o=n.name||n;p?lr(o,e.Diagnostics.Constructor_implementation_is_missing):e.hasSyntacticModifier(n,128)?lr(o,e.Diagnostics.All_declarations_of_an_abstract_method_must_be_consecutive):lr(o,e.Diagnostics.Function_implementation_is_missing_or_not_immediately_following_the_declaration)}}}function lb(n){if(t){var r=n.localSymbol;if((r||(r=xa(n)).exportSymbol)&&e.getDeclarationOfKind(r,n.kind)===n){for(var a=0,i=0,o=0,s=0,c=r.declarations;s<c.length;s++){var l=y(_=c[s]),u=sb(_,513);1&u?512&u?o|=l:a|=l:i|=l}var d=a&i,p=o&(a|i);if(d||p)for(var m=0,f=r.declarations;m<f.length;m++){l=y(_=f[m]);var _,g=e.getNameOfDeclaration(_);l&p?lr(g,e.Diagnostics.Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead,e.declarationNameToString(g)):l&d&&lr(g,e.Diagnostics.Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local,e.declarationNameToString(g))}}}function y(n){var t=n;switch(t.kind){case 253:case 254:case 331:case 324:case 325:return 2;case 256:return e.isAmbientModule(t)||0!==e.getModuleInstanceState(t)?5:4;case 252:case 255:case 291:return 3;case 297:return 7;case 266:if(!e.isEntityNameExpression(t.expression))return 1;t=t.expression;case 260:case 263:case 262:var r=0,a=ea(xa(t));return e.forEach(a.declarations,(function(e){r|=y(e)})),r;case 249:case 198:case 251:case 265:case 78:return 1;default:return e.Debug.failBadSyntaxKind(t)}}}function ub(e,n,t,r){var a=db(e,n);return a&&mb(a,n,t,r)}function db(n,t){if(!Li(n)){var r=n;if(r.promisedTypeOfPromise)return r.promisedTypeOfPromise;if(uo(n,xl(!1)))return r.promisedTypeOfPromise=nl(n)[0];var a=xi(n,"then");if(!Li(a)){var i=a?uc(a,0):e.emptyArray;if(0!==i.length){var o=If(Xl(e.map(i,Vv)),2097152);if(!Li(o)){var s=uc(o,0);if(0!==s.length)return r.promisedTypeOfPromise=Xl(e.map(s,Vv),2);t&&lr(t,e.Diagnostics.The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback)}}else t&&lr(t,e.Diagnostics.A_promise_must_have_a_then_method)}}}function pb(e,n,t,r){return mb(e,n,t,r)||xe}function mb(e,n,t,r){if(Li(e))return e;var a=e;return a.awaitedTypeOfType?a.awaitedTypeOfType:a.awaitedTypeOfType=Yf(e,n?function(e){return fb(e,n,t,r)}:fb)}function fb(n,t,r,a){var i=n;if(i.awaitedTypeOfType)return i.awaitedTypeOfType;var o=db(n);if(o){if(n.id===o.id||Xt.lastIndexOf(o.id)>=0)return void(t&&lr(t,e.Diagnostics.Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method));Xt.push(n.id);var s=mb(o,t,r,a);if(Xt.pop(),!s)return;return i.awaitedTypeOfType=s}if(!function(e){var n=xi(e,"then");return!!n&&uc(If(n,2097152),0).length>0}(n))return i.awaitedTypeOfType=n;if(t){if(!r)return e.Debug.fail();lr(t,r,a)}}function _b(n){var t=mv(n);yv(t,n);var r=kc(t);if(!(1&r.flags)){var a,i,o=cv(n);switch(n.parent.kind){case 252:a=Xl([lo(xa(n.parent)),je]);break;case 160:a=je,i=e.chainDiagnosticMessages(void 0,e.Diagnostics.The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any);break;case 163:a=je,i=e.chainDiagnosticMessages(void 0,e.Diagnostics.The_return_type_of_a_property_decorator_function_must_be_either_void_or_any);break;case 165:case 167:case 168:a=Xl([Il(ET(n.parent)),je]);break;default:return e.Debug.fail()}Yd(r,a,n,o,(function(){return i}))}}function gb(n){if(n){var t=e.getFirstIdentifier(n),r=2097152|(78===n.kind?788968:1920),a=Cr(t,t.escapedText,r,void 0,void 0,!0);a&&2097152&a.flags&&Ra(a)&&!GT(ea(a))&&!ra(a)&&ia(a)}}function yb(n){var t=vb(n);t&&e.isEntityName(t)&&gb(t)}function vb(e){if(e)switch(e.kind){case 183:case 182:return hb(e.types);case 184:return hb([e.trueType,e.falseType]);case 186:case 192:return vb(e.type);case 173:return e.typeName}}function hb(n){for(var t,r=0,a=n;r<a.length;r++){for(var i=a[r];186===i.kind||192===i.kind;)i=i.type;if(141!==i.kind&&(W||(191!==i.kind||103!==i.literal.kind)&&150!==i.kind)){var o=vb(i);if(!o)return;if(t){if(!e.isIdentifier(t)||!e.isIdentifier(o)||t.escapedText!==o.escapedText)return}else t=o}}return t}function bb(n){var t=e.getEffectiveTypeAnnotationNode(n);return e.isRestParameter(n)?e.getRestParameterElementType(t):t}function Eb(n){if(n.decorators&&e.nodeCanBeDecorated(n,n.parent,n.parent.parent)){V.experimentalDecorators||lr(n,e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning);var t=n.decorators[0];if(cS(t,8),160===n.kind&&cS(t,32),V.emitDecoratorMetadata)switch(cS(t,16),n.kind){case 252:var r=e.getFirstConstructorWithBody(n);if(r)for(var a=0,i=r.parameters;a<i.length;a++){yb(bb(i[a]))}break;case 167:case 168:var o=167===n.kind?168:167,s=e.getDeclarationOfKind(xa(n),o);yb(eo(n)||s&&eo(s));break;case 165:for(var c=0,l=n.parameters;c<l.length;c++){yb(bb(l[c]))}yb(e.getEffectiveReturnTypeNode(n));break;case 163:yb(e.getEffectiveTypeAnnotationNode(n));break;case 160:yb(bb(n));for(var u=0,d=n.parent.parameters;u<d.length;u++){yb(bb(d[u]))}}e.forEach(n.decorators,_b)}}function Tb(e){switch(e.kind){case 78:return e;case 201:return e.name;default:return}}function Sb(n){Eb(n),qh(n);var r=e.getFunctionFlags(n);if(n.name&&158===n.name.kind&&vg(n.name),!Qo(n)){var a=xa(n),i=n.localSymbol||a,o=e.find(i.declarations,(function(e){return e.kind===n.kind&&!(131072&e.flags)}));n===o&&cb(i),a.parent&&cb(a)}var s=164===n.kind?void 0:n.body;if(iT(s),th(n,Oc(n)),t&&!e.getEffectiveReturnTypeNode(n)&&(e.nodeIsMissing(s)&&!ob(n)&&km(n,Ee),1&r&&e.nodeIsPresent(s)&&kc(Lc(n))),e.isInJSFile(n)){var c=e.getJSDocTypeTag(n);c&&c.typeExpression&&!cg(sd(c.typeExpression),n)&&lr(c,e.Diagnostics.The_type_of_a_function_declaration_must_match_the_function_s_signature)}}function xb(n){if(t){var r=e.getSourceFileOfNode(n),a=Tt.get(r.path);a||(a=[],Tt.set(r.path,a)),a.push(n)}}function Lb(n,t){for(var r=0,a=n;r<a.length;r++){var i=a[r];switch(i.kind){case 252:case 221:Cb(i,t),Nb(i,t);break;case 297:case 256:case 230:case 258:case 237:case 238:case 239:Mb(i,t);break;case 166:case 208:case 251:case 209:case 165:case 167:case 168:i.body&&Mb(i,t),Nb(i,t);break;case 164:case 169:case 170:case 174:case 175:case 254:case 253:Nb(i,t);break;case 185:Ib(i,t);break;default:e.Debug.assertNever(i,"Node should not have been registered for unused identifiers check")}}}function Ab(n,t,r){var a=e.getNameOfDeclaration(n)||n,i=fT(n)?e.Diagnostics._0_is_declared_but_never_used:e.Diagnostics._0_is_declared_but_its_value_is_never_read;r(n,0,e.createDiagnosticForNode(a,i,t))}function Db(n){return e.isIdentifier(n)&&95===e.idText(n).charCodeAt(0)}function Cb(n,t){for(var r=0,a=n.members;r<a.length;r++){var i=a[r];switch(i.kind){case 165:case 163:case 167:case 168:if(168===i.kind&&32768&i.symbol.flags)break;var o=xa(i);o.isReferenced||!(e.hasEffectiveModifier(i,8)||e.isNamedDeclaration(i)&&e.isPrivateIdentifier(i.name))||8388608&i.flags||t(i,0,e.createDiagnosticForNode(i.name,e.Diagnostics._0_is_declared_but_its_value_is_never_read,ri(o)));break;case 166:for(var s=0,c=i.parameters;s<c.length;s++){var l=c[s];!l.symbol.isReferenced&&e.hasSyntacticModifier(l,8)&&t(l,0,e.createDiagnosticForNode(l.name,e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read,e.symbolName(l.symbol)))}break;case 171:case 229:break;default:e.Debug.fail()}}}function Ib(n,t){var r=n.typeParameter;Rb(r)&&t(n,1,e.createDiagnosticForNode(n,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.idText(r.name)))}function Nb(n,t){if(e.last(xa(n).declarations)===n)for(var r=e.getEffectiveTypeParameterDeclarations(n),a=new e.Set,i=0,o=r;i<o.length;i++){var s=o[i];if(Rb(s)){var c=e.idText(s.name),l=s.parent;if(185!==l.kind&&l.typeParameters.every(Rb)){if(e.tryAddToSet(a,l)){var u=e.isJSDocTemplateTag(l)?e.rangeOfNode(l):e.rangeOfTypeParameters(l.typeParameters),d=1===l.typeParameters.length,p=d?e.Diagnostics._0_is_declared_but_its_value_is_never_read:e.Diagnostics.All_type_parameters_are_unused,m=d?c:void 0;t(s,1,e.createFileDiagnostic(e.getSourceFileOfNode(l),u.pos,u.end-u.pos,p,m))}}else t(s,1,e.createDiagnosticForNode(s,e.Diagnostics._0_is_declared_but_its_value_is_never_read,c))}}}function Rb(e){return!(262144&Sa(e.symbol).isReferenced||Db(e.name))}function kb(e,n,t,r){var a=String(r(n)),i=e.get(a);i?i[1].push(t):e.set(a,[n,[t]])}function Ob(n){return e.tryCast(e.getRootDeclaration(n),e.isParameter)}function Pb(n){return e.isBindingElement(n)&&Db(n.name)?!!e.findAncestor(n.parent,(function(n){return!(e.isArrayBindingPattern(n)||e.isVariableDeclaration(n)||e.isVariableDeclarationList(n))&&(!!e.isForOfStatement(n)||"quit")})):e.isAmbientModule(n)||(e.isVariableDeclaration(n)&&e.isForInOrOfStatement(n.parent.parent)||Fb(n))&&Db(n.name)}function Mb(n,t){var r=new e.Map,a=new e.Map,i=new e.Map;n.locals.forEach((function(n){var o;if(!(262144&n.flags?!(3&n.flags)||3&n.isReferenced:n.isReferenced||n.exportSymbol))for(var s=0,c=n.declarations;s<c.length;s++){var l=c[s];if(!Pb(l))if(Fb(l))kb(r,262===(o=l).kind?o:263===o.kind?o.parent:o.parent.parent,l,R);else if(e.isBindingElement(l)&&e.isObjectBindingPattern(l.parent)){l!==e.last(l.parent.elements)&&e.last(l.parent.elements).dotDotDotToken||kb(a,l.parent,l,R)}else if(e.isVariableDeclaration(l))kb(i,l.parent,l,R);else{var u=n.valueDeclaration&&Ob(n.valueDeclaration),d=n.valueDeclaration&&e.getNameOfDeclaration(n.valueDeclaration);u&&d?e.isParameterPropertyDeclaration(u,u.parent)||e.parameterIsThisKeyword(u)||Db(d)||t(u,1,e.createDiagnosticForNode(d,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.symbolName(n))):Ab(l,e.symbolName(n),t)}}})),r.forEach((function(n){var r=n[0],a=n[1],i=r.parent;if((r.name?1:0)+(r.namedBindings?263===r.namedBindings.kind?1:r.namedBindings.elements.length:0)===a.length)t(i,0,1===a.length?e.createDiagnosticForNode(i,e.Diagnostics._0_is_declared_but_its_value_is_never_read,e.idText(e.first(a).name)):e.createDiagnosticForNode(i,e.Diagnostics.All_imports_in_import_declaration_are_unused));else for(var o=0,s=a;o<s.length;o++){var c=s[o];Ab(c,e.idText(c.name),t)}})),a.forEach((function(n){var r=n[0],a=n[1],o=Ob(r.parent)?1:0;if(r.elements.length===a.length)1===a.length&&249===r.parent.kind&&250===r.parent.parent.kind?kb(i,r.parent.parent,r.parent,R):t(r,o,1===a.length?e.createDiagnosticForNode(r,e.Diagnostics._0_is_declared_but_its_value_is_never_read,wb(e.first(a).name)):e.createDiagnosticForNode(r,e.Diagnostics.All_destructured_elements_are_unused));else for(var s=0,c=a;s<c.length;s++){var l=c[s];t(l,o,e.createDiagnosticForNode(l,e.Diagnostics._0_is_declared_but_its_value_is_never_read,wb(l.name)))}})),i.forEach((function(n){var r=n[0],a=n[1];if(r.declarations.length===a.length)t(r,0,1===a.length?e.createDiagnosticForNode(e.first(a).name,e.Diagnostics._0_is_declared_but_its_value_is_never_read,wb(e.first(a).name)):e.createDiagnosticForNode(232===r.parent.kind?r.parent:r,e.Diagnostics.All_variables_are_unused));else for(var i=0,o=a;i<o.length;i++){var s=o[i];t(s,0,e.createDiagnosticForNode(s,e.Diagnostics._0_is_declared_but_its_value_is_never_read,wb(s.name)))}}))}function wb(n){switch(n.kind){case 78:return e.idText(n);case 197:case 196:return wb(e.cast(e.first(n.elements),e.isBindingElement).name);default:return e.Debug.assertNever(n)}}function Fb(e){return 262===e.kind||265===e.kind||263===e.kind}function Gb(n){if(230===n.kind&&BS(n),e.isFunctionOrModuleBlock(n)){var t=At;e.forEach(n.statements,iT),At=t}else e.forEach(n.statements,iT);n.locals&&xb(n)}function Bb(n,t,r){if(!t||t.escapedText!==r)return!1;if(163===n.kind||162===n.kind||165===n.kind||164===n.kind||167===n.kind||168===n.kind)return!1;if(8388608&n.flags)return!1;var a=e.getRootDeclaration(n);return 160!==a.kind||!e.nodeIsMissing(a.parent.body)}function Ub(n){e.findAncestor(n,(function(t){return!!(4&WT(t))&&(78!==n.kind?lr(e.getNameOfDeclaration(n),e.Diagnostics.Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference):lr(n,e.Diagnostics.Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference),!0)}))}function Vb(n){e.findAncestor(n,(function(t){return!!(8&WT(t))&&(78!==n.kind?lr(e.getNameOfDeclaration(n),e.Diagnostics.Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference):lr(n,e.Diagnostics.Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference),!0)}))}function jb(n){67108864&WT(e.getEnclosingBlockScopeContainer(n))&&cr("noEmit",n,e.Diagnostics.Compiler_reserves_name_0_when_emitting_private_identifier_downlevel,"WeakMap")}function Hb(n,t){if(!(H>=e.ModuleKind.ES2015)&&(Bb(n,t,"require")||Bb(n,t,"exports"))&&(!e.isModuleDeclaration(n)||1===e.getModuleInstanceState(n))){var r=Si(n);297===r.kind&&e.isExternalOrCommonJsModule(r)&&cr("noEmit",t,e.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module,e.declarationNameToString(t),e.declarationNameToString(t))}}function Kb(n,t){if(!(j>=4)&&Bb(n,t,"Promise")&&(!e.isModuleDeclaration(n)||1===e.getModuleInstanceState(n))){var r=Si(n);297===r.kind&&e.isExternalOrCommonJsModule(r)&&2048&r.flags&&cr("noEmit",t,e.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions,e.declarationNameToString(t),e.declarationNameToString(t))}}function Wb(e){return e===Te?Ee:e===In?Cn:e}function zb(n){var t;if(Eb(n),e.isBindingElement(n)||iT(n.type),n.name){if(158===n.name.kind&&(vg(n.name),n.initializer&&Lh(n.initializer)),198===n.kind){196===n.parent.kind&&j<99&&cS(n,4),n.propertyName&&158===n.propertyName.kind&&vg(n.propertyName);var r=n.parent.parent,a=Ai(r),i=n.propertyName||n.name;if(a&&!e.isBindingPattern(i)){var o=ou(i);if(qo(o)){var s=cc(a,Zo(o));s&&(hy(s,void 0,!1),qg(r,!!r.initializer&&105===r.initializer.kind,a,s))}}}if(e.isBindingPattern(n.name)&&(197===n.name.kind&&j<2&&V.downlevelIteration&&cS(n,512),e.forEach(n.name.elements,iT)),n.initializer&&e.isParameterDeclaration(n)&&e.nodeIsMissing(e.getContainingFunction(n).body))lr(n,e.Diagnostics.A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation);else if(e.isBindingPattern(n.name)){var c=n.initializer&&238!==n.parent.parent.kind,l=0===n.name.elements.length;if(c||l){var u=Xi(n);if(c){var d=Lh(n.initializer);W&&l?ty(d,n):Qd(d,Xi(n),n,n.initializer)}l&&(e.isArrayBindingPattern(n.name)?aE(65,u,Ce,n):W&&ty(u,n))}}else{var p=xa(n);if(2097152&p.flags&&e.isRequireVariableDeclaration(n,!0))QE(n);else{var m=Wb(lo(p));if(n===p.valueDeclaration){var f=e.getEffectiveInitializer(n);if(f)e.isInJSFile(n)&&e.isObjectLiteralExpression(f)&&(0===f.properties.length||e.isPrototypeAccess(n.name))&&!!(null===(t=p.exports)||void 0===t?void 0:t.size)||238===n.parent.parent.kind||Qd(Lh(f),m,n,f,void 0);p.declarations.length>1&&e.some(p.declarations,(function(t){return t!==n&&e.isVariableLike(t)&&!Jb(t,n)}))&&lr(n.name,e.Diagnostics.All_declarations_of_0_must_have_identical_modifiers,e.declarationNameToString(n.name))}else{var _=Wb(Xi(n));m===xe||_===xe||Vd(m,_)||67108864&p.flags||qb(p.valueDeclaration,m,n,_),n.initializer&&Qd(Lh(n.initializer),_,n,n.initializer,void 0),Jb(n,p.valueDeclaration)||lr(n.name,e.Diagnostics.All_declarations_of_0_must_have_identical_modifiers,e.declarationNameToString(n.name))}163!==n.kind&&162!==n.kind&&(lb(n),249!==n.kind&&198!==n.kind||function(n){if(0==(3&e.getCombinedNodeFlags(n))&&!e.isParameterDeclaration(n)&&(249!==n.kind||n.initializer)){var t=xa(n);if(1&t.flags){if(!e.isIdentifier(n.name))return e.Debug.fail();var r=Cr(n,n.name.escapedText,3,void 0,void 0,!1);if(r&&r!==t&&2&r.flags&&3&Wg(r)){var a=e.getAncestor(r.valueDeclaration,250),i=232===a.parent.kind&&a.parent.parent?a.parent.parent:void 0;if(!i||!(230===i.kind&&e.isFunctionLike(i.parent)||257===i.kind||256===i.kind||297===i.kind)){var o=ri(r);lr(n,e.Diagnostics.Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1,o,o)}}}}}(n),Hb(n,n.name),Kb(n,n.name),j<99&&Bb(n,n.name,"WeakMap")&&Jt.push(n))}}}}function qb(n,t,r,a){var i=e.getNameOfDeclaration(r),o=163===r.kind||162===r.kind?e.Diagnostics.Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2:e.Diagnostics.Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2,s=e.declarationNameToString(i),c=lr(i,o,s,ii(t),ii(a));n&&e.addRelatedInfo(c,e.createDiagnosticForNode(n,e.Diagnostics._0_was_also_declared_here,s))}function Jb(n,t){if(160===n.kind&&249===t.kind||249===n.kind&&160===t.kind)return!0;if(e.hasQuestionToken(n)!==e.hasQuestionToken(t))return!1;return e.getSelectedEffectiveModifierFlags(n,504)===e.getSelectedEffectiveModifierFlags(t,504)}function Xb(n){e.tracing.push("check","checkVariableDeclaration",{kind:n.kind,pos:n.pos,end:n.end}),function(n){if(238!==n.parent.parent.kind&&239!==n.parent.parent.kind)if(8388608&n.flags)IS(n);else if(!n.initializer){if(e.isBindingPattern(n.name)&&!e.isBindingPattern(n.parent))return FS(n,e.Diagnostics.A_destructuring_declaration_must_have_an_initializer);if(e.isVarConst(n))return FS(n,e.Diagnostics.const_declarations_must_be_initialized)}if(n.exclamationToken&&(232!==n.parent.parent.kind||!n.type||n.initializer||8388608&n.flags)){var t=n.initializer?e.Diagnostics.Declarations_with_initializers_cannot_also_have_definite_assignment_assertions:e.Diagnostics.Declarations_with_definite_assignment_assertions_must_also_have_type_annotations;return FS(n.exclamationToken,t)}var r=e.getEmitModuleKind(V);r<e.ModuleKind.ES2015&&r!==e.ModuleKind.System&&!(8388608&n.parent.parent.flags)&&e.hasSyntacticModifier(n.parent.parent,1)&&NS(n.name);(e.isLet(n)||e.isVarConst(n))&&RS(n.name)}(n),zb(n),e.tracing.pop()}function Yb(n){return function(n){if(n.dotDotDotToken){var t=n.parent.elements;if(n!==e.last(t))return FS(n,e.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);if(pS(t,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),n.propertyName)return FS(n.name,e.Diagnostics.A_rest_element_cannot_have_a_property_name)}if(n.dotDotDotToken&&n.initializer)wS(n,n.initializer.pos-1,1,e.Diagnostics.A_rest_element_cannot_have_an_initializer)}(n),zb(n)}function Qb(n){uS(n)||kS(n.declarationList)||function(n){if(!OS(n.parent)){if(e.isLet(n.declarationList))return FS(n,e.Diagnostics.let_declarations_can_only_be_declared_inside_a_block);if(e.isVarConst(n.declarationList))FS(n,e.Diagnostics.const_declarations_can_only_be_declared_inside_a_block)}}(n),e.forEach(n.declarationList.declarations,iT)}function Zb(n,t,r){if(W){var a=e.isIdentifier(n)?n:e.isPropertyAccessExpression(n)?n.name:void 0;if(a)if(!um(r))if(0!==uc(r,0).length){var i=bT(a);if(i)e.forEachChild(t,(function t(r){if(e.isIdentifier(r)){var o=bT(r);if(o&&o===i){if(e.isIdentifier(n))return!0;for(var s=a.parent,c=r.parent;s&&c;){if(e.isIdentifier(s)&&e.isIdentifier(c)||107===s.kind&&107===c.kind)return bT(s)===bT(c);if(!e.isPropertyAccessExpression(s)||!e.isPropertyAccessExpression(c))return!1;if(bT(s.name)!==bT(c.name))return!1;c=c.expression,s=s.expression}}}return e.forEachChild(r,t)}))||lr(n,e.Diagnostics.This_condition_will_always_return_true_since_the_function_is_always_defined_Did_you_mean_to_call_it_instead)}}}function $b(n,t){return 16384&n.flags&&lr(t,e.Diagnostics.An_expression_of_type_void_cannot_be_tested_for_truthiness),n}function eE(e,n){return $b(Hh(e,n),e)}function nE(n){xS(n);var t,r=Qg(Hh(n.expression));if(250===n.initializer.kind){var a=n.initializer.declarations[0];a&&e.isBindingPattern(a.name)&&lr(a.name,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern),tE(n)}else{var i=n.initializer,o=Hh(i);199===i.kind||200===i.kind?lr(i,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern):zd(131072&(t=du(uu(r))).flags?Oe:t,o)?ch(i,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access):lr(i,e.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any)}r!==He&&ph(r,126091264)||lr(n.expression,e.Diagnostics.The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0,ii(r)),iT(n.statement),n.locals&&xb(n)}function tE(e){var n=e.initializer;n.declarations.length>=1&&Xb(n.declarations[0])}function rE(e){return aE(e.awaitModifier?15:13,Xg(e.expression),Ce,e.expression)}function aE(e,n,t,r){return Li(n)?n:iE(e,n,t,r,!0)||Ee}function iE(n,t,r,a,i){var o=0!=(2&n);if(t!==He){var s=j>=2,c=!s&&V.downlevelIteration,l=V.noUncheckedIndexedAccess&&!!(128&n);if(s||c||o){var u=dE(t,n,s?a:void 0);if(i&&u){var d=8&n?e.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_send_0:32&n?e.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_always_send_0:64&n?e.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring_will_always_send_0:16&n?e.Diagnostics.Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_containing_generator_will_always_send_0:void 0;d&&Yd(r,u.nextType,a,d)}if(u||s)return l?Of(u&&u.yieldType):u&&u.yieldType}var p=t,m=!1,f=!1;if(4&n){if(1048576&p.flags){var _=t.types,g=e.filter(_,(function(e){return!(402653316&e.flags)}));g!==_&&(p=Xl(g,2))}else 402653316&p.flags&&(p=He);if((f=p!==t)&&(j<1&&a&&(lr(a,e.Diagnostics.Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher),m=!0),131072&p.flags))return l?Of(Oe):Oe}if(!Kp(p)){if(a&&!m){var y=oE(n,0,t,void 0),v=4&n&&!f?c?[e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator,!0]:y?[e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators,!1]:[e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type,!0]:c?[e.Diagnostics.Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator,!0]:y?[e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators,!1]:[e.Diagnostics.Type_0_is_not_an_array_type,!0],h=v[0];pr(a,v[1]&&!!ub(p),h,ii(p))}return f?l?Of(Oe):Oe:void 0}var b=fc(p,1);return f&&b?402653316&b.flags&&!V.noUncheckedIndexedAccess?Oe:Xl(l?[b,Oe,Ce]:[b,Oe],2):128&n?Of(b):b}vE(a,t,o)}function oE(e,n,t,r){if(!Li(t)){var a=dE(t,e,r);return a&&a[F(n)]}}function sE(e,n,t){if(void 0===e&&(e=He),void 0===n&&(n=He),void 0===t&&(t=De),67359327&e.flags&&180227&n.flags&&180227&t.flags){var r=Yc([e,n,t]),a=mt.get(r);return a||(a={yieldType:e,returnType:n,nextType:t},mt.set(r,a)),a}return{yieldType:e,returnType:n,nextType:t}}function cE(n){for(var t,r,a,i=0,o=n;i<o.length;i++){var s=o[i];if(void 0!==s&&s!==ft){if(s===_t)return _t;t=e.append(t,s.yieldType),r=e.append(r,s.returnType),a=e.append(a,s.nextType)}}return t||r||a?sE(t&&Xl(t),r&&Xl(r),a&&tu(a)):ft}function lE(e,n){return e[n]}function uE(e,n,t){return e[n]=t}function dE(n,t,r){if(Li(n))return _t;if(!(1048576&n.flags)){var a=mE(n,t,r);return a===ft?void(r&&vE(r,n,!!(2&t))):a}var i,o=2&t?"iterationTypesOfAsyncIterable":"iterationTypesOfIterable",s=lE(n,o);if(s)return s===ft?void 0:s;for(var c=0,l=n.types;c<l.length;c++){var u=mE(l[c],t,r);if(u===ft)return r&&vE(r,n,!!(2&t)),void uE(n,o,ft);i=e.append(i,u)}var d=i?cE(i):ft;return uE(n,o,d),d===ft?void 0:d}function pE(e,n){if(e===ft)return ft;if(e===_t)return _t;var t=e.yieldType,r=e.returnType,a=e.nextType;return sE(mb(t,n)||Ee,mb(r,n)||Ee,a)}function mE(e,n,t){if(Li(e))return _t;var r;if(2&n&&(r=fE(e,vt)||gE(e,vt)))return r;if(1&n&&(r=fE(e,ht)||gE(e,ht))){if(!(2&n))return r;if(r!==ft)return uE(e,"iterationTypesOfAsyncIterable",pE(r,t))}if(2&n&&(r=yE(e,vt,t))!==ft)return r;if(1&n&&(r=yE(e,ht,t))!==ft)return 2&n?uE(e,"iterationTypesOfAsyncIterable",r?pE(r,t):ft):r;return ft}function fE(e,n){return lE(e,n.iterableCacheKey)}function _E(e,n){var t=fE(e,n)||yE(e,n,void 0);return t===ft?yt:t}function gE(e,n){var t;if(uo(e,t=n.getGlobalIterableType(!1))||uo(e,t=n.getGlobalIterableIteratorType(!1))){var r=nl(e)[0],a=_E(t,n),i=a.returnType,o=a.nextType;return uE(e,n.iterableCacheKey,sE(r,i,o))}if(uo(e,n.getGlobalGeneratorType(!1))){var s=nl(e);r=s[0],i=s[1],o=s[2];return uE(e,n.iterableCacheKey,sE(r,i,o))}}function yE(n,t,r){var a,i=cc(n,e.getPropertyNameForKnownSymbolName(t.iteratorSymbolName)),o=!i||16777216&i.flags?void 0:lo(i);if(Li(o))return uE(n,t.iterableCacheKey,_t);var s=o?uc(o,0):void 0;if(!e.some(s))return uE(n,t.iterableCacheKey,ft);var c=null!==(a=hE(tu(e.map(s,kc)),t,r))&&void 0!==a?a:ft;return uE(n,t.iterableCacheKey,c)}function vE(n,t,r){var a=r?e.Diagnostics.Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator:e.Diagnostics.Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator;pr(n,!!ub(t),a,ii(t))}function hE(e,n,t){if(Li(e))return _t;var r=bE(e,n)||function(e,n){var t=n.getGlobalIterableIteratorType(!1);if(uo(e,t)){var r=nl(e)[0],a=bE(t,n)||AE(t,n,void 0),i=a===ft?yt:a,o=i.returnType,s=i.nextType;return uE(e,n.iteratorCacheKey,sE(r,o,s))}if(uo(e,n.getGlobalIteratorType(!1))||uo(e,n.getGlobalGeneratorType(!1))){var c=nl(e);r=c[0],o=c[1],s=c[2];return uE(e,n.iteratorCacheKey,sE(r,o,s))}}(e,n)||AE(e,n,t);return r===ft?void 0:r}function bE(e,n){return lE(e,n.iteratorCacheKey)}function EE(e,n){var t=xi(e,"done")||we;return zd(0===n?we:Ge,t)}function TE(e){return EE(e,0)}function SE(e){return EE(e,1)}function xE(e){if(Li(e))return _t;var n,t=lE(e,"iterationTypesOfIteratorResult");if(t)return t;if(uo(e,(n=!1,Hn||(Hn=El("IteratorYieldResult",1,n))||on)))return uE(e,"iterationTypesOfIteratorResult",sE(nl(e)[0],void 0,void 0));if(uo(e,function(e){return Kn||(Kn=El("IteratorReturnResult",1,e))||on}(!1)))return uE(e,"iterationTypesOfIteratorResult",sE(void 0,nl(e)[0],void 0));var r=Jf(e,TE),a=r!==He?xi(r,"value"):void 0,i=Jf(e,SE),o=i!==He?xi(i,"value"):void 0;return uE(e,"iterationTypesOfIteratorResult",a||o?sE(a,o||je,void 0):ft)}function LE(n,t,r,a){var i,o,s,c,l=cc(n,r);if(l||"next"===r){var u=!l||"next"===r&&16777216&l.flags?void 0:"next"===r?lo(l):If(lo(l),2097152);if(Li(u))return"next"===r?_t:gt;var d,p,m,f,_,g=u?uc(u,0):e.emptyArray;if(0===g.length){if(a)lr(a,"next"===r?t.mustHaveANextMethodDiagnostic:t.mustBeAMethodDiagnostic,r);return"next"===r?_t:void 0}if((null==u?void 0:u.symbol)&&1===g.length){var y=t.getGlobalGeneratorType(!1),v=t.getGlobalIteratorType(!1),h=(null===(o=null===(i=y.symbol)||void 0===i?void 0:i.members)||void 0===o?void 0:o.get(r))===u.symbol,b=!h&&(null===(c=null===(s=v.symbol)||void 0===s?void 0:s.members)||void 0===c?void 0:c.get(r))===u.symbol;if(h||b){var E=h?y:v,T=u.mapper;return sE(md(E.typeParameters[0],T),md(E.typeParameters[1],T),"next"===r?md(E.typeParameters[2],T):void 0)}}for(var S=0,x=g;S<x.length;S++){var L=x[S];"throw"!==r&&e.some(L.parameters)&&(d=e.append(d,Ov(L,0))),p=e.append(p,kc(L))}if("throw"!==r){var A=d?Xl(d):De;if("next"===r)f=A;else if("return"===r){var D=t.resolveIterationType(A,a)||Ee;m=e.append(m,D)}}var C=p?tu(p):He,I=xE(t.resolveIterationType(C,a)||Ee);return I===ft?(a&&lr(a,t.mustHaveAValueDiagnostic,r),_=Ee,m=e.append(m,Ee)):(_=I.yieldType,m=e.append(m,I.returnType)),sE(_,Xl(m),f)}}function AE(e,n,t){var r=cE([LE(e,n,"next",t),LE(e,n,"return",t),LE(e,n,"throw",t)]);return uE(e,n.iteratorCacheKey,r)}function DE(e,n,t){if(!Li(n)){var r=CE(n,t);return r&&r[F(e)]}}function CE(e,n){if(Li(e))return _t;var t=n?vt:ht;return dE(e,n?2:1,void 0)||hE(e,t,void 0)}function IE(n){BS(n)||function(n){var t=n;for(;t;){if(e.isFunctionLike(t))return FS(n,e.Diagnostics.Jump_target_cannot_cross_function_boundary);switch(t.kind){case 245:if(n.label&&t.label.escapedText===n.label.escapedText)return!!(240===n.kind&&!e.isIterationStatement(t.statement,!0))&&FS(n,e.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement);break;case 244:if(241===n.kind&&!n.label)return!1;break;default:if(e.isIterationStatement(t,!1)&&!n.label)return!1}t=t.parent}n.label?FS(n,241===n.kind?e.Diagnostics.A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement:e.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement):FS(n,241===n.kind?e.Diagnostics.A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement:e.Diagnostics.A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement)}(n)}function NE(e,n){var t,r,a=!!(2&n);return!!(1&n)?null!==(t=DE(1,e,a))&&void 0!==t?t:xe:a?null!==(r=mb(e))&&void 0!==r?r:xe:e}function RE(n,t){var r=NE(t,e.getFunctionFlags(n));return!!r&&dh(r,16387)}function kE(n){BS(n)||e.isIdentifier(n.expression)&&!n.expression.escapedText&&function(n,t,r,a,i){var o=e.getSourceFileOfNode(n);if(!PS(o)){var s=e.getSpanOfTokenAtPosition(o,n.pos);return Yt.add(e.createFileDiagnostic(o,e.textSpanEnd(s),0,t,r,a,i)),!0}}(n,e.Diagnostics.Line_break_not_permitted_here),n.expression&&Hh(n.expression)}function OE(n){var t,r=Kc(n.symbol,1),a=Kc(n.symbol,0),i=fc(n,0),o=fc(n,1);if(i||o){e.forEach(Ms(n),(function(e){var t=lo(e);m(e,t,n,a,i,0),m(e,t,n,r,o,1)}));var s=n.symbol.valueDeclaration;if(1&e.getObjectFlags(n)&&e.isClassLike(s))for(var c=0,l=s.members;c<l.length;c++){var u=l[c];if(!e.hasSyntacticModifier(u,32)&&Qo(u)){var d=xa(u),p=lo(d);m(d,p,n,a,i,0),m(d,p,n,r,o,1)}}}i&&o&&(!(t=r||a)&&2&e.getObjectFlags(n)&&(t=e.forEach(Lo(n),(function(e){return fc(e,0)&&fc(e,1)}))?void 0:n.symbol.declarations[0]));function m(n,t,r,a,i,o){if(i&&!e.isKnownSymbol(n)){var s=n.valueDeclaration,c=s&&e.getNameOfDeclaration(s);if((!c||!e.isPrivateIdentifier(c))&&(1!==o||(c?_g(c):yg(n.escapedName)))){var l;if(s&&c&&(216===s.kind||158===c.kind||n.parent===r.symbol))l=s;else if(a)l=a;else if(2&e.getObjectFlags(r)){l=e.forEach(Lo(r),(function(e){return ws(e,n.escapedName)&&fc(e,o)}))?void 0:r.symbol.declarations[0]}if(l&&!zd(t,i))lr(l,0===o?e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_string_index_type_2:e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2,ri(n),ii(t),ii(i))}}}t&&!zd(o,i)&&lr(t,e.Diagnostics.Numeric_index_type_0_is_not_assignable_to_string_index_type_1,ii(o),ii(i))}function PE(e,n){switch(e.escapedText){case"any":case"unknown":case"number":case"bigint":case"boolean":case"string":case"symbol":case"void":case"object":lr(e,n,e.escapedText)}}function ME(n){if(n)for(var r=!1,a=0;a<n.length;a++){var i=n[a];if(Kh(i),t){i.default?(r=!0,wE(i.default,n,a)):r&&lr(i,e.Diagnostics.Required_type_parameters_may_not_follow_optional_type_parameters);for(var o=0;o<a;o++)n[o].symbol===i.symbol&&lr(i.name,e.Diagnostics.Duplicate_identifier_0,e.declarationNameToString(i.name))}}}function wE(n,t,r){!function n(a){if(173===a.kind){var i=_l(a);if(262144&i.flags)for(var o=r;o<t.length;o++)i.symbol===xa(t[o])&&lr(a,e.Diagnostics.Type_parameter_defaults_can_only_reference_previously_declared_type_parameters)}e.forEachChild(a,n)}(n)}function FE(n){if(1!==n.declarations.length){var t=Tr(n);if(!t.typeParametersChecked){t.typeParametersChecked=!0;var r=function(n){return e.filter(n.declarations,(function(e){return 252===e.kind||253===e.kind}))}(n);if(r.length<=1)return;if(!function(n,t){for(var r=e.length(t),a=Sc(t),i=0,o=n;i<o.length;i++){var s=o[i],c=e.getEffectiveTypeParameterDeclarations(s),l=c.length;if(l<a||l>r)return!1;for(var u=0;u<l;u++){var d=c[u],p=t[u];if(d.name.escapedText!==p.symbol.escapedName)return!1;var m=e.getEffectiveConstraintOfTypeParameter(d),f=m&&sd(m),_=Us(p);if(f&&_&&!Vd(f,_))return!1;var g=d.default&&sd(d.default),y=Ys(p);if(g&&y&&!Vd(g,y))return!1}}return!0}(r,Fo(n).localTypeParameters))for(var a=ri(n),i=0,o=r;i<o.length;i++){lr(o[i].name,e.Diagnostics.All_declarations_of_0_must_have_identical_type_parameters,a)}}}}function GE(n){var r;!function(n){var t=e.getSourceFileOfNode(n);(function(n){var t=!1,r=!1;if(!uS(n)&&n.heritageClauses)for(var a=0,i=n.heritageClauses;a<i.length;a++){var o=i[a];if(93===o.token){if(t)return MS(o,e.Diagnostics.extends_clause_already_seen);if(r)return MS(o,e.Diagnostics.extends_clause_must_precede_implements_clause);if(o.types.length>1)return MS(o.types[1],e.Diagnostics.Classes_can_only_extend_a_single_class);t=!0}else{if(e.Debug.assert(116===o.token),r)return MS(o,e.Diagnostics.implements_clause_already_seen);r=!0}vS(o)}})(n)||mS(n.typeParameters,t)}(n),Eb(n),n.name&&(PE(n.name,e.Diagnostics.Class_name_cannot_be_0),Hb(n,n.name),Kb(n,n.name),8388608&n.flags||(r=n.name,1===j&&"Object"===r.escapedText&&H<e.ModuleKind.ES2015&&lr(r,e.Diagnostics.Class_name_cannot_be_Object_when_targeting_ES5_with_module_0,e.ModuleKind[H]))),ME(e.getEffectiveTypeParameterDeclarations(n)),lb(n);var a=xa(n),i=Fo(a),o=rs(i),s=lo(a);FE(a),cb(a),function(n){for(var t=new e.Map,r=new e.Map,a=new e.Map,i=0,o=n.members;i<o.length;i++){var s=o[i];if(166===s.kind)for(var c=0,l=s.parameters;c<l.length;c++){var u=l[c];e.isParameterPropertyDeclaration(u,s)&&!e.isBindingPattern(u.name)&&_(t,u.name,u.name.escapedText,3)}else{var d=e.hasSyntacticModifier(s,32),p=s.name;if(!p)return;var m=e.isPrivateIdentifier(p)?a:d?r:t,f=p&&e.getPropertyNameForPropertyNameNode(p);if(f)switch(s.kind){case 167:_(m,p,f,1);break;case 168:_(m,p,f,2);break;case 163:_(m,p,f,3);break;case 165:_(m,p,f,8)}}}function _(n,t,r,a){var i=n.get(r);i?8&i?8!==a&&lr(t,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(t)):i&a?lr(t,e.Diagnostics.Duplicate_identifier_0,e.getTextOfNode(t)):n.set(r,i|a):n.set(r,a)}}(n),8388608&n.flags||function(n){for(var t=0,r=n.members;t<r.length;t++){var a=r[t],i=a.name;if(e.hasSyntacticModifier(a,32)&&i){var o=e.getPropertyNameForPropertyNameNode(i);switch(o){case"name":case"length":case"caller":case"arguments":case"prototype":lr(i,e.Diagnostics.Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1,o,gi(xa(n)))}}}}(n);var c=e.getEffectiveBaseTypeNode(n);if(c){e.forEach(c.typeArguments,iT),j<2&&cS(c.parent,1);var l=e.getClassExtendsHeritageElement(n);l&&l!==c&&Hh(l.expression);var u=Lo(i);if(u.length&&t){var d=u[0],p=So(i),m=Zs(p);if(function(n,t){var r=uc(n,1);if(r.length){var a=r[0].declaration;if(a&&e.hasEffectiveModifier(a,8))yT(t,e.getClassLikeDeclarationOfSymbol(n.symbol))||lr(t,e.Diagnostics.Cannot_extend_a_class_0_Class_constructor_is_marked_as_private,sa(n.symbol))}}(m,c),iT(c.expression),e.some(c.typeArguments)){e.forEach(c.typeArguments,iT);for(var f=0,_=Eo(m,c.typeArguments,c);f<_.length;f++){if(!nb(c,_[f].typeParameters))break}}if(Yd(o,T=rs(d,i.thisType),void 0)?Yd(s,Ud(m),n.name||n,e.Diagnostics.Class_static_side_0_incorrectly_extends_base_class_static_side_1):BE(n,o,T,e.Diagnostics.Class_0_incorrectly_extends_base_class_1),8650752&p.flags&&!vo(s)&&lr(n.name||n,e.Diagnostics.A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any),!(m.symbol&&32&m.symbol.flags||8650752&p.flags)){var g=To(m,c.typeArguments,c);e.forEach(g,(function(e){return!fv(e.declaration)&&!Vd(kc(e),d)}))&&lr(c.expression,e.Diagnostics.Base_constructors_must_all_have_the_same_return_type)}!function(n,t){var r=Gs(t);e:for(var a=0,i=r;a<i.length;a++){var o=i[a],s=UE(o);if(!(4194304&s.flags)){var c=ws(n,s.escapedName);if(c){var l=UE(c),u=e.getDeclarationModifierFlagsFromSymbol(s);if(e.Debug.assert(!!l,"derived should point to something, even if it is the base class' declaration."),l===s){var d=e.getClassLikeDeclarationOfSymbol(n.symbol);if(128&u&&(!d||!e.hasSyntacticModifier(d,128))){for(var p=0,m=Lo(n);p<m.length;p++){var f=m[p];if(f!==t){var _=ws(f,s.escapedName),g=_&&UE(_);if(g&&g!==s)continue e}}221===d.kind?lr(d,e.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1,ri(o),ii(t)):lr(d,e.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2,ii(n),ri(o),ii(t))}}else{var y=e.getDeclarationModifierFlagsFromSymbol(l);if(8&u||8&y)continue;var v=void 0,h=98308&s.flags,b=98308&l.flags;if(h&&b){if(128&u&&!(s.valueDeclaration&&e.isPropertyDeclaration(s.valueDeclaration)&&s.valueDeclaration.initializer)||s.valueDeclaration&&253===s.valueDeclaration.parent.kind||l.valueDeclaration&&e.isBinaryExpression(l.valueDeclaration))continue;var E=4!==h&&4===b;if(E||4===h&&4!==b){var T=E?e.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property:e.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor;lr(e.getNameOfDeclaration(l.valueDeclaration)||l.valueDeclaration,T,ri(s),ii(t),ii(n))}else if(V.useDefineForClassFields){var S=e.find(l.declarations,(function(e){return 163===e.kind&&!e.initializer}));if(S&&!(33554432&l.flags)&&!(128&u)&&!(128&y)&&!l.declarations.some((function(e){return!!(8388608&e.flags)}))){var x=ka(e.getClassLikeDeclarationOfSymbol(n.symbol)),L=S.name;if(S.exclamationToken||!x||!e.isIdentifier(L)||!W||!jE(L,n,x)){var A=e.Diagnostics.Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration;lr(e.getNameOfDeclaration(l.valueDeclaration)||l.valueDeclaration,A,ri(s),ii(t))}}}continue}if(zg(s)){if(zg(l)||4&l.flags)continue;e.Debug.assert(!!(98304&l.flags)),v=e.Diagnostics.Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor}else v=98304&s.flags?e.Diagnostics.Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function:e.Diagnostics.Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function;lr(e.getNameOfDeclaration(l.valueDeclaration)||l.valueDeclaration,v,ii(t),ri(s),ii(n))}}}}}(i,d)}}var y=e.getEffectiveImplementsTypeNodes(n);if(y)for(var v=0,h=y;v<h.length;v++){var b=h[v];if(e.isEntityNameExpression(b.expression)||lr(b.expression,e.Diagnostics.A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments),rb(b),t){var E=rc(sd(b));if(E!==xe)if(Do(E)){var T,S=E.symbol&&32&E.symbol.flags?e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass:e.Diagnostics.Class_0_incorrectly_implements_interface_1;Yd(o,T=rs(E,i.thisType),void 0)||BE(n,o,T,S)}else lr(b,e.Diagnostics.A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members)}}t&&(OE(i),Xh(n),function(n){if(!W||!J||8388608&n.flags)return;for(var t=ka(n),r=0,a=n.members;r<a.length;r++){var i=a[r];if(!(2&e.getEffectiveModifierFlags(i))&&VE(i)){var o=i.name;if(e.isIdentifier(o)||e.isPrivateIdentifier(o)){var s=lo(xa(i));3&s.flags||32768&um(s)||t&&jE(o,s,t)||lr(i.name,e.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor,e.declarationNameToString(o))}}}}(n))}function BE(n,t,r,a){for(var i=!1,o=function(n){if(e.hasStaticModifier(n))return"continue";var a=n.name&&bT(n.name)||bT(n);if(a){var o=cc(t,a.escapedName),s=cc(r,a.escapedName);if(o&&s){Yd(lo(o),lo(s),n.name||n,void 0,(function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2,ri(a),ii(t),ii(r))}))||(i=!0)}}},s=0,c=n.members;s<c.length;s++){o(c[s])}i||Yd(t,r,n.name||n,a)}function UE(n){return 1&e.getCheckFlags(n)?n.target:n}function VE(n){return 163===n.kind&&!e.hasSyntacticModifier(n,160)&&!n.exclamationToken&&!n.initializer}function jE(n,t,r){var a=e.factory.createPropertyAccessExpression(e.factory.createThis(),n);return e.setParent(a.expression,a),e.setParent(a,r),a.flowNode=r.returnFlowNode,!(32768&um(y_(a,t,fm(t))))}function HE(n){if(uS(n)||function(n){var t=!1;if(n.heritageClauses)for(var r=0,a=n.heritageClauses;r<a.length;r++){var i=a[r];if(93!==i.token)return e.Debug.assert(116===i.token),MS(i,e.Diagnostics.Interface_declaration_cannot_have_implements_clause);if(t)return MS(i,e.Diagnostics.extends_clause_already_seen);t=!0,vS(i)}}(n),ME(n.typeParameters),t){PE(n.name,e.Diagnostics.Interface_name_cannot_be_0),lb(n);var r=xa(n);if(FE(r),n===e.getDeclarationOfKind(r,253)){var a=Fo(r),i=rs(a);if(function(n,t){var r=Lo(n);if(r.length<2)return!0;var a=new e.Map;e.forEach(zo(n).declaredProperties,(function(e){a.set(e.escapedName,{prop:e,containingType:n})}));for(var i=!0,o=0,s=r;o<s.length;o++)for(var c=s[o],l=0,u=Gs(rs(c,n.thisType));l<u.length;l++){var d=u[l],p=a.get(d.escapedName);if(p){if(p.containingType!==n&&!wp(p.prop,d)){i=!1;var m=ii(p.containingType),f=ii(c),_=e.chainDiagnosticMessages(void 0,e.Diagnostics.Named_property_0_of_types_1_and_2_are_not_identical,ri(d),m,f);_=e.chainDiagnosticMessages(_,e.Diagnostics.Interface_0_cannot_simultaneously_extend_types_1_and_2,ii(n),m,f),Yt.add(e.createDiagnosticForNodeFromMessageChain(t,_))}}else a.set(d.escapedName,{prop:d,containingType:c})}return i}(a,n.name)){for(var o=0,s=Lo(a);o<s.length;o++){Yd(i,rs(s[o],a.thisType),n.name,e.Diagnostics.Interface_0_incorrectly_extends_interface_1)}OE(a)}}Jh(n)}e.forEach(e.getInterfaceBaseTypeNodes(n),(function(n){e.isEntityNameExpression(n.expression)||lr(n.expression,e.Diagnostics.An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments),rb(n)})),e.forEach(n.members,iT),t&&(Xh(n),xb(n))}function KE(e){var n=Sr(e);if(!(16384&n.flags)){n.flags|=16384;for(var t=0,r=0,a=e.members;r<a.length;r++){var i=a[r],o=WE(i,t);Sr(i).enumMemberValue=o,t="number"==typeof o?o+1:void 0}}}function WE(n,t){if(e.isComputedNonLiteralName(n.name))lr(n.name,e.Diagnostics.Computed_property_names_are_not_allowed_in_enums);else{var r=e.getTextOfPropertyName(n.name);yg(r)&&!gg(r)&&lr(n.name,e.Diagnostics.An_enum_member_cannot_have_a_numeric_name)}return n.initializer?function(n){var t=ko(xa(n.parent)),r=e.isEnumConst(n.parent),a=n.initializer,i=1!==t||Ro(n)?s(a):void 0;if(void 0!==i)r&&"number"==typeof i&&!isFinite(i)&&lr(a,isNaN(i)?e.Diagnostics.const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN:e.Diagnostics.const_enum_member_initializer_was_evaluated_to_a_non_finite_value);else{if(1===t)return lr(a,e.Diagnostics.Computed_values_are_not_permitted_in_an_enum_with_string_valued_members),0;if(r)lr(a,e.Diagnostics.const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values);else if(8388608&n.parent.flags)lr(a,e.Diagnostics.In_ambient_enum_declarations_member_initializer_must_be_constant_expression);else{var o=Hh(a);ph(o,296)?Yd(o,Fo(xa(n.parent)),a,void 0):lr(a,e.Diagnostics.Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhaustiveness_checks_consider_using_an_object_literal_instead,ii(o))}}return i;function s(t){switch(t.kind){case 214:var r=s(t.operand);if("number"==typeof r)switch(t.operator){case 39:return r;case 40:return-r;case 54:return~r}break;case 216:var a=s(t.left),i=s(t.right);if("number"==typeof a&&"number"==typeof i)switch(t.operatorToken.kind){case 51:return a|i;case 50:return a&i;case 48:return a>>i;case 49:return a>>>i;case 47:return a<<i;case 52:return a^i;case 41:return a*i;case 43:return a/i;case 39:return a+i;case 40:return a-i;case 44:return a%i;case 42:return Math.pow(a,i)}else if("string"==typeof a&&"string"==typeof i&&39===t.operatorToken.kind)return a+i;break;case 10:case 14:return t.text;case 8:return US(t),+t.text;case 207:return s(t.expression);case 78:var o=t;return gg(o.escapedText)?+o.escapedText:e.nodeIsMissing(t)?0:c(t,xa(n.parent),o.escapedText);case 202:case 201:var l=t;if(zE(l)){var u=Uh(l.expression);if(u.symbol&&384&u.symbol.flags){var d=void 0;return d=201===l.kind?l.name.escapedText:e.escapeLeadingUnderscores(e.cast(l.argumentExpression,e.isLiteralExpression).text),c(t,u.symbol,d)}}}}function c(t,r,a){var i=r.exports.get(a);if(i){var o=i.valueDeclaration;if(o!==n)return Ar(o,n)?zT(o):(lr(t,e.Diagnostics.A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums),0);lr(t,e.Diagnostics.Property_0_is_used_before_being_assigned,ri(i))}}}(n):8388608&n.parent.flags&&!e.isEnumConst(n.parent)&&0===ko(xa(n.parent))?void 0:void 0!==t?t:void lr(n.name,e.Diagnostics.Enum_member_must_have_initializer)}function zE(n){return 78===n.kind||201===n.kind&&zE(n.expression)||202===n.kind&&zE(n.expression)&&e.isStringLiteralLike(n.argumentExpression)}function qE(n){e.isPrivateIdentifier(n.name)&&lr(n,e.Diagnostics.An_enum_member_cannot_be_named_with_a_private_identifier)}function JE(n){if(t){var r=e.isGlobalScopeAugmentation(n),a=8388608&n.flags;r&&!a&&lr(n.name,e.Diagnostics.Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context);var i=e.isAmbientModule(n);if(nT(n,i?e.Diagnostics.An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file:e.Diagnostics.A_namespace_declaration_is_only_allowed_in_a_namespace_or_module))return;uS(n)||a||10!==n.name.kind||FS(n.name,e.Diagnostics.Only_ambient_modules_can_use_quoted_names),e.isIdentifier(n.name)&&(Hb(n,n.name),Kb(n,n.name)),lb(n);var o=xa(n);if(512&o.flags&&!a&&o.declarations.length>1&&O(n,!!V.preserveConstEnums||!!V.isolatedModules)){var s=function(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t];if((252===a.kind||251===a.kind&&e.nodeIsPresent(a.body))&&!(8388608&a.flags))return a}}(o);s&&(e.getSourceFileOfNode(n)!==e.getSourceFileOfNode(s)?lr(n.name,e.Diagnostics.A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged):n.pos<s.pos&&lr(n.name,e.Diagnostics.A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged));var c=e.getDeclarationOfKind(o,252);c&&(d=n,p=c,m=e.getEnclosingBlockScopeContainer(d),f=e.getEnclosingBlockScopeContainer(p),xr(m)?xr(f):!xr(f)&&m===f)&&(Sr(n).flags|=32768)}if(i)if(e.isExternalModuleAugmentation(n)){if((r||33554432&xa(n).flags)&&n.body)for(var l=0,u=n.body.statements;l<u.length;l++){XE(u[l],r)}}else xr(n.parent)?r?lr(n.name,e.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations):e.isExternalModuleNameRelative(e.getTextOfIdentifierOrLiteral(n.name))&&lr(n.name,e.Diagnostics.Ambient_module_declaration_cannot_specify_relative_module_name):lr(n.name,r?e.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations:e.Diagnostics.Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces)}var d,p,m,f;n.body&&(iT(n.body),e.isGlobalScopeAugmentation(n)||xb(n))}function XE(n,t){switch(n.kind){case 232:for(var r=0,a=n.declarationList.declarations;r<a.length;r++){XE(a[r],t)}break;case 266:case 267:MS(n,e.Diagnostics.Exports_and_export_assignments_are_not_permitted_in_module_augmentations);break;case 260:case 261:MS(n,e.Diagnostics.Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module);break;case 198:case 249:var i=n.name;if(e.isBindingPattern(i)){for(var o=0,s=i.elements;o<s.length;o++){XE(s[o],t)}break}case 252:case 255:case 251:case 253:case 256:case 254:if(t)return;var c=xa(n);if(c){var l=!(33554432&c.flags);l||(l=!!c.parent&&e.isExternalModuleAugmentation(c.parent.declarations[0]))}}}function YE(n){var t=e.getExternalModuleName(n);if(!t||e.nodeIsMissing(t))return!1;if(!e.isStringLiteral(t))return lr(t,e.Diagnostics.String_literal_expected),!1;var r=257===n.parent.kind&&e.isAmbientModule(n.parent.parent);return 297===n.parent.kind||r?!(r&&e.isExternalModuleNameRelative(t.text)&&!mi(n))||(lr(n,e.Diagnostics.Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name),!1):(lr(t,267===n.kind?e.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace:e.Diagnostics.Import_declarations_in_a_namespace_cannot_reference_a_module),!1)}function QE(n){var t=xa(n),r=ea(t);if(r!==he){var a=(1160127&(t=Sa(t.exportSymbol||t)).flags?111551:0)|(788968&t.flags?788968:0)|(1920&t.flags?1920:0);if(r.flags&a)lr(n,270===n.kind?e.Diagnostics.Export_declaration_conflicts_with_exported_declaration_of_0:e.Diagnostics.Import_declaration_conflicts_with_local_declaration_of_0,ri(t));!V.isolatedModules||270!==n.kind||n.parent.parent.isTypeOnly||111551&r.flags||8388608&n.flags||lr(n,e.Diagnostics.Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type),e.isImportSpecifier(n)&&e.every(r.declarations,(function(n){return!!(134217728&e.getCombinedNodeFlags(n))}))&&dr(!1,n.name,e.Diagnostics._0_is_deprecated,t.escapedName)}}function ZE(n){Hb(n,n.name),Kb(n,n.name),QE(n),265===n.kind&&"default"===e.idText(n.propertyName||n.name)&&V.esModuleInterop&&H!==e.ModuleKind.System&&H<e.ModuleKind.ES2015&&cS(n,262144)}function $E(n){if(!nT(n,e.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)&&(!uS(n)&&e.hasEffectiveModifiers(n)&&MS(n,e.Diagnostics.An_import_declaration_cannot_have_modifiers),YE(n))){var t=n.importClause;if(t&&!function(n){if(n.isTypeOnly&&n.name&&n.namedBindings)return FS(n,e.Diagnostics.A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both);return!1}(t))if(t.name&&ZE(t),t.namedBindings)if(263===t.namedBindings.kind)ZE(t.namedBindings),H!==e.ModuleKind.System&&H<e.ModuleKind.ES2015&&V.esModuleInterop&&cS(n,131072);else ua(n,n.moduleSpecifier)&&e.forEach(t.namedBindings.elements,ZE)}}function eT(n){if(!nT(n,e.Diagnostics.An_export_declaration_can_only_be_used_in_a_module)&&(!uS(n)&&e.hasEffectiveModifiers(n)&&MS(n,e.Diagnostics.An_export_declaration_cannot_have_modifiers),n.moduleSpecifier&&n.exportClause&&e.isNamedExports(n.exportClause)&&e.length(n.exportClause.elements)&&0===j&&cS(n,4194304),function(n){var t,r=n.isTypeOnly&&268!==(null===(t=n.exportClause)||void 0===t?void 0:t.kind);r&&FS(n,e.Diagnostics.Only_named_exports_may_use_export_type)}(n),!n.moduleSpecifier||YE(n)))if(n.exportClause&&!e.isNamespaceExport(n.exportClause)){e.forEach(n.exportClause.elements,rT);var t=257===n.parent.kind&&e.isAmbientModule(n.parent.parent),r=!t&&257===n.parent.kind&&!n.moduleSpecifier&&8388608&n.flags;297===n.parent.kind||t||r||lr(n,e.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace)}else{var a=ua(n,n.moduleSpecifier);a&&ga(a)?lr(n.moduleSpecifier,e.Diagnostics.Module_0_uses_export_and_cannot_be_used_with_export_Asterisk,ri(a)):n.exportClause&&QE(n.exportClause),H!==e.ModuleKind.System&&H<e.ModuleKind.ES2015&&(n.exportClause?V.esModuleInterop&&cS(n,131072):cS(n,65536))}}function nT(e,n){var t=297===e.parent.kind||257===e.parent.kind||256===e.parent.kind;return t||MS(e,n),!t}function tT(n){return e.forEachImportClauseDeclaration(n,(function(e){return!!Tr(xa(e)).constEnumReferenced}))}function rT(n){if(QE(n),e.getEmitDeclarations(V)&&vi(n.propertyName||n.name,!0),n.parent.parent.moduleSpecifier)V.esModuleInterop&&H!==e.ModuleKind.System&&H<e.ModuleKind.ES2015&&"default"===e.idText(n.propertyName||n.name)&&cS(n,262144);else{var t=n.propertyName||n.name,r=Cr(t,t.escapedText,2998271,void 0,void 0,!0);if(r&&(r===te||r===re||xr(Si(r.declarations[0]))))lr(t,e.Diagnostics.Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module,e.idText(t));else{aa(n);var a=r&&(2097152&r.flags?ea(r):r);(!a||a===he||111551&a.flags)&&Lh(n.propertyName||n.name)}}}function aT(n){var t=xa(n),r=Tr(t);if(!r.exportsChecked){var a=t.exports.get("export=");if(a&&function(n){return e.forEachEntry(n.exports,(function(e,n){return"export="!==n}))}(t)){var i=Br(a)||a.valueDeclaration;mi(i)||e.isInJSFile(i)||lr(i,e.Diagnostics.An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements)}var o=ba(t);o&&o.forEach((function(n,t){var r=n.declarations,a=n.flags;if("__export"!==t&&!(1984&a)){var i=e.countWhere(r,A);if(!(524288&a&&i<=2)&&i>1)for(var o=0,s=r;o<s.length;o++){var c=s[o];P(c)&&Yt.add(e.createDiagnosticForNode(c,e.Diagnostics.Cannot_redeclare_exported_variable_0,e.unescapeLeadingUnderscores(t)))}}})),r.exportsChecked=!0}}function iT(n){if(n){var a=c;c=n,h=0,function(n){e.isInJSFile(n)&&e.forEach(n.jsDoc,(function(n){var t=n.tags;return e.forEach(t,iT)}));var a=n.kind;if(r)switch(a){case 256:case 252:case 253:case 251:r.throwIfCancellationRequested()}a>=232&&a<=248&&n.flowNode&&!m_(n.flowNode)&&dr(!1===V.allowUnreachableCode,n,e.Diagnostics.Unreachable_code_detected);switch(a){case 159:return Kh(n);case 160:return Wh(n);case 163:return Yh(n);case 162:return function(n){return e.isPrivateIdentifier(n.name)&&lr(n,e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies),Yh(n)}(n);case 174:case 175:case 169:case 170:case 171:return qh(n);case 165:case 164:return function(n){DS(n)||bS(n.name),e.isPrivateIdentifier(n.name)&&lr(n,e.Diagnostics.A_method_cannot_be_named_with_a_private_identifier),Sb(n),e.hasSyntacticModifier(n,128)&&165===n.kind&&n.body&&lr(n,e.Diagnostics.Method_0_cannot_have_an_implementation_because_it_is_marked_abstract,e.declarationNameToString(n.name))}(n);case 166:return Qh(n);case 167:case 168:return Zh(n);case 173:return rb(n);case 172:return function(n){var t=function(e){switch(e.parent.kind){case 209:case 169:case 251:case 208:case 174:case 165:case 164:var n=e.parent;if(e===n.type)return n}}(n);if(t){var r=Lc(t),a=Rc(r);if(a){iT(n.type);var i=n.parameterName;if(0===a.kind||2===a.kind)ad(i);else if(a.parameterIndex>=0)G(r)&&a.parameterIndex===r.parameters.length-1?lr(i,e.Diagnostics.A_type_predicate_cannot_reference_a_rest_parameter):a.type&&Yd(a.type,lo(r.parameters[a.parameterIndex]),n.type,void 0,(function(){return e.chainDiagnosticMessages(void 0,e.Diagnostics.A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type)}));else if(i){for(var o=!1,s=0,c=t.parameters;s<c.length;s++){var l=c[s].name;if(e.isBindingPattern(l)&&zh(l,i,a.parameterName)){o=!0;break}}o||lr(n.parameterName,e.Diagnostics.Cannot_find_parameter_0,a.parameterName)}}}else lr(n,e.Diagnostics.A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods)}(n);case 176:return function(e){yl(e)}(n);case 177:return function(n){e.forEach(n.members,iT),t&&(OE(ju(n)),Xh(n),Jh(n))}(n);case 178:return function(e){iT(e.elementType)}(n);case 179:return function(n){for(var t=n.elements,r=!1,a=!1,i=e.some(t,e.isNamedTupleMember),o=0;o<t.length;o++){var s=t[o];if(192!==s.kind&&i){FS(s,e.Diagnostics.Tuple_members_must_all_have_names_or_all_not_have_names);break}var c=Rl(s);if(8&c){var l=sd(s.type);if(!Kp(l)){lr(s,e.Diagnostics.A_rest_element_type_must_be_an_array_type);break}(Up(l)||tm(l)&&4&l.target.combinedFlags)&&(a=!0)}else if(4&c)a=!0;else if(2&c)r=!0;else if(r){FS(s,e.Diagnostics.A_required_element_cannot_follow_an_optional_element);break}if(a&&o!==t.length-1){FS(s,e.Diagnostics.A_rest_element_must_be_last_in_a_tuple_type);break}}e.forEach(n.elements,iT),sd(n)}(n);case 182:case 183:return function(n){e.forEach(n.types,iT),sd(n)}(n);case 186:case 180:case 181:return iT(n.type);case 187:return function(e){ad(e)}(n);case 188:return ib(n);case 184:return function(n){e.forEachChild(n,iT)}(n);case 185:return function(n){e.findAncestor(n,(function(e){return e.parent&&184===e.parent.kind&&e.parent.extendsType===e}))||FS(n,e.Diagnostics.infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type),iT(n.typeParameter),xb(n)}(n);case 193:return function(e){for(var n=0,t=e.templateSpans;n<t.length;n++){var r=t[n];iT(r.type),Yd(sd(r.type),Ze,r.type)}sd(e)}(n);case 195:return function(e){iT(e.argument),sd(e)}(n);case 192:return function(n){n.dotDotDotToken&&n.questionToken&&FS(n,e.Diagnostics.A_tuple_member_cannot_be_both_optional_and_rest),180===n.type.kind&&FS(n.type,e.Diagnostics.A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type),181===n.type.kind&&FS(n.type,e.Diagnostics.A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type),iT(n.type),sd(n)}(n);case 315:return function(n){var t=e.getEffectiveJSDocHost(n);if(t&&(e.isClassDeclaration(t)||e.isClassExpression(t))){var r=e.getJSDocTags(t).filter(e.isJSDocAugmentsTag);e.Debug.assert(r.length>0),r.length>1&&lr(r[1],e.Diagnostics.Class_declarations_cannot_have_more_than_one_augments_or_extends_tag);var a=Tb(n.class.expression),i=e.getClassExtendsHeritageElement(t);if(i){var o=Tb(i.expression);o&&a.escapedText!==o.escapedText&&lr(a,e.Diagnostics.JSDoc_0_1_does_not_match_the_extends_2_clause,e.idText(n.tagName),e.idText(a),e.idText(o))}}else lr(t,e.Diagnostics.JSDoc_0_is_not_attached_to_a_class,e.idText(n.tagName))}(n);case 316:return function(n){var t=e.getEffectiveJSDocHost(n);t&&(e.isClassDeclaration(t)||e.isClassExpression(t))||lr(t,e.Diagnostics.JSDoc_0_is_not_attached_to_a_class,e.idText(n.tagName))}(n);case 331:case 324:case 325:return function(n){n.typeExpression||lr(n.name,e.Diagnostics.JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags),n.name&&PE(n.name,e.Diagnostics.Type_alias_name_cannot_be_0),iT(n.typeExpression)}(n);case 330:return function(e){iT(e.constraint);for(var n=0,t=e.typeParameters;n<t.length;n++)iT(t[n])}(n);case 329:return function(e){iT(e.typeExpression)}(n);case 326:return function(n){if(iT(n.typeExpression),!e.getParameterSymbolFromJSDoc(n)){var t=e.getHostSignatureFromJSDoc(n);if(t){var r=e.getJSDocTags(t).filter(e.isJSDocParameterTag).indexOf(n);if(r>-1&&r<t.parameters.length&&e.isBindingPattern(t.parameters[r].name))return;Dc(t)?e.findLast(e.getJSDocTags(t),e.isJSDocParameterTag)===n&&n.typeExpression&&n.typeExpression.type&&!Up(sd(n.typeExpression.type))&&lr(n.name,e.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type,e.idText(157===n.name.kind?n.name.right:n.name)):e.isQualifiedName(n.name)?lr(n.name,e.Diagnostics.Qualified_name_0_is_not_allowed_without_a_leading_param_object_1,e.entityNameToString(n.name),e.entityNameToString(n.name.left)):lr(n.name,e.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name,e.idText(n.name))}}}(n);case 333:return function(e){iT(e.typeExpression)}(n);case 308:!function(n){!t||n.type||e.isJSDocConstructSignature(n)||km(n,Ee),qh(n)}(n);case 306:case 305:case 303:case 304:case 312:return oT(n),void e.forEachChild(n,iT);case 309:return void function(n){oT(n),iT(n.type);var t=n.parent;if(e.isParameter(t)&&e.isJSDocFunctionType(t.parent))return void(e.last(t.parent.parameters)!==t&&lr(n,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list));e.isJSDocTypeExpression(t)||lr(n,e.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature);var r=n.parent.parent;if(!e.isJSDocParameterTag(r))return void lr(n,e.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature);var a=e.getParameterSymbolFromJSDoc(r);if(!a)return;var i=e.getHostSignatureFromJSDoc(r);i&&e.last(i.parameters).symbol===a||lr(n,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list)}(n);case 301:return iT(n.type);case 189:return function(e){iT(e.objectType),iT(e.indexType),ab(ku(e),e)}(n);case 190:return function(n){iT(n.typeParameter),iT(n.nameType),iT(n.type),n.type||km(n,Ee);var t=Ou(n),r=Ls(t);r?Yd(r,Ye,n.nameType):Yd(xs(t),Ye,e.getEffectiveConstraintOfTypeParameter(n.typeParameter))}(n);case 251:return function(e){t&&(Sb(e),ES(e),Hb(e,e.name),Kb(e,e.name))}(n);case 230:case 257:return Gb(n);case 232:return Qb(n);case 233:return function(e){BS(e),Hh(e.expression)}(n);case 234:return function(n){BS(n);var t=eE(n.expression);Zb(n.expression,n.thenStatement,t),iT(n.thenStatement),231===n.thenStatement.kind&&lr(n.thenStatement,e.Diagnostics.The_body_of_an_if_statement_cannot_be_the_empty_statement),iT(n.elseStatement)}(n);case 235:return function(e){BS(e),iT(e.statement),eE(e.expression)}(n);case 236:return function(e){BS(e),eE(e.expression),iT(e.statement)}(n);case 237:return function(n){BS(n)||n.initializer&&250===n.initializer.kind&&kS(n.initializer),n.initializer&&(250===n.initializer.kind?e.forEach(n.initializer.declarations,Xb):Hh(n.initializer)),n.condition&&eE(n.condition),n.incrementor&&Hh(n.incrementor),iT(n.statement),n.locals&&xb(n)}(n);case 238:return nE(n);case 239:return function(n){if(xS(n),n.awaitModifier?2==(6&e.getFunctionFlags(e.getContainingFunction(n)))&&j<99&&cS(n,32768):V.downlevelIteration&&j<2&&cS(n,256),250===n.initializer.kind)tE(n);else{var t=n.initializer,r=rE(n);if(199===t.kind||200===t.kind)vh(t,r||xe);else{var a=Hh(t);ch(t,e.Diagnostics.The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access,e.Diagnostics.The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access),r&&Qd(r,a,t,n.expression)}}iT(n.statement),n.locals&&xb(n)}(n);case 240:case 241:return IE(n);case 242:return function(n){var t;if(!BS(n)){var r=e.getContainingFunction(n);if(r){var a=kc(Lc(r)),i=e.getFunctionFlags(r);if(W||n.expression||131072&a.flags){var o=n.expression?Lh(n.expression):Ce;if(168===r.kind)n.expression&&lr(n,e.Diagnostics.Setters_cannot_return_a_value);else if(166===r.kind)n.expression&&!Qd(o,a,n,n.expression)&&lr(n,e.Diagnostics.Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class);else if(Oc(r)){var s=null!==(t=NE(a,i))&&void 0!==t?t:a,c=2&i?pb(o,n,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member):o;s&&Qd(c,s,n,n.expression)}}else 166!==r.kind&&V.noImplicitReturns&&!RE(r,a)&&lr(n,e.Diagnostics.Not_all_code_paths_return_a_value)}else MS(n,e.Diagnostics.A_return_statement_can_only_be_used_within_a_function_body)}}(n);case 243:return function(n){BS(n)||32768&n.flags&&MS(n,e.Diagnostics.with_statements_are_not_allowed_in_an_async_function_block),Hh(n.expression);var t=e.getSourceFileOfNode(n);if(!PS(t)){var r=e.getSpanOfTokenAtPosition(t,n.pos).start;wS(t,r,n.statement.pos-r,e.Diagnostics.The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any)}}(n);case 244:return function(n){var r;BS(n);var a=!1,i=Hh(n.expression),o=Yp(i);e.forEach(n.caseBlock.clauses,(function(n){if(285!==n.kind||a||(void 0===r?r=n:(FS(n,e.Diagnostics.A_default_clause_cannot_appear_more_than_once_in_a_switch_statement),a=!0)),t&&284===n.kind){var s=Hh(n.expression),c=Yp(s),l=i;c&&o||(s=c?Qp(s):s,l=Qp(i)),bh(l,s)||op(s,l,n.expression,void 0)}e.forEach(n.statements,iT),V.noFallthroughCasesInSwitch&&n.fallthroughFlowNode&&m_(n.fallthroughFlowNode)&&lr(n,e.Diagnostics.Fallthrough_case_in_switch)})),n.caseBlock.locals&&xb(n.caseBlock)}(n);case 245:return function(n){BS(n)||e.findAncestor(n.parent,(function(t){return e.isFunctionLike(t)?"quit":245===t.kind&&t.label.escapedText===n.label.escapedText&&(FS(n.label,e.Diagnostics.Duplicate_label_0,e.getTextOfNode(n.label)),!0)})),iT(n.statement)}(n);case 246:return kE(n);case 247:return function(n){BS(n),Gb(n.tryBlock);var t=n.catchClause;if(t){if(t.variableDeclaration){var r=t.variableDeclaration;if(r.type){var a=Mi(r,!1);!a||3&a.flags||MS(r.type,e.Diagnostics.Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified)}else if(r.initializer)MS(r.initializer,e.Diagnostics.Catch_clause_variable_cannot_have_an_initializer);else{var i=t.block.locals;i&&e.forEachKey(t.locals,(function(n){var t=i.get(n);t&&0!=(2&t.flags)&&FS(t.valueDeclaration,e.Diagnostics.Cannot_redeclare_identifier_0_in_catch_clause,n)}))}}Gb(t.block)}n.finallyBlock&&Gb(n.finallyBlock)}(n);case 249:return Xb(n);case 198:return Yb(n);case 252:return function(n){n.name||e.hasSyntacticModifier(n,512)||MS(n,e.Diagnostics.A_class_declaration_without_the_default_modifier_must_have_a_name),GE(n),e.forEach(n.members,iT),xb(n)}(n);case 253:return HE(n);case 254:return function(n){uS(n),PE(n.name,e.Diagnostics.Type_alias_name_cannot_be_0),lb(n),ME(n.typeParameters),136===n.type.kind?C.has(n.name.escapedText)&&1===e.length(n.typeParameters)||lr(n.type,e.Diagnostics.The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types):(iT(n.type),xb(n))}(n);case 255:return function(n){if(t){uS(n),PE(n.name,e.Diagnostics.Enum_name_cannot_be_0),Hb(n,n.name),Kb(n,n.name),lb(n),n.members.forEach(qE),KE(n);var r=xa(n);if(n===e.getDeclarationOfKind(r,n.kind)){if(r.declarations.length>1){var a=e.isEnumConst(n);e.forEach(r.declarations,(function(n){e.isEnumDeclaration(n)&&e.isEnumConst(n)!==a&&lr(e.getNameOfDeclaration(n),e.Diagnostics.Enum_declarations_must_all_be_const_or_non_const)}))}var i=!1;e.forEach(r.declarations,(function(n){if(255!==n.kind)return!1;var t=n;if(!t.members.length)return!1;var r=t.members[0];r.initializer||(i?lr(r.name,e.Diagnostics.In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element):i=!0)}))}}}(n);case 256:return JE(n);case 261:return $E(n);case 260:return function(n){if(!nT(n,e.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)&&(uS(n),e.isInternalModuleImportEqualsDeclaration(n)||YE(n)))if(ZE(n),e.hasSyntacticModifier(n,1)&&aa(n),272!==n.moduleReference.kind){var t=ea(xa(n));if(t!==he){if(111551&t.flags){var r=e.getFirstIdentifier(n.moduleReference);1920&ca(r,112575).flags||lr(r,e.Diagnostics.Module_0_is_hidden_by_a_local_declaration_with_the_same_name,e.declarationNameToString(r))}788968&t.flags&&PE(n.name,e.Diagnostics.Import_name_cannot_be_0)}}else H>=e.ModuleKind.ES2015&&!(8388608&n.flags)&&FS(n,e.Diagnostics.Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead)}(n);case 267:return eT(n);case 266:return function(n){if(!nT(n,e.Diagnostics.An_export_assignment_can_only_be_used_in_a_module)){var t=297===n.parent.kind?n.parent:n.parent.parent;if(256!==t.kind||e.isAmbientModule(t)){if(!uS(n)&&e.hasEffectiveModifiers(n)&&MS(n,e.Diagnostics.An_export_assignment_cannot_have_modifiers),78===n.expression.kind){var r=n.expression,a=ca(r,67108863,!0,!0,n);if(a){L_(a,r);var i=2097152&a.flags?ea(a):a;(i===he||111551&i.flags)&&Lh(n.expression)}else Lh(n.expression);e.getEmitDeclarations(V)&&vi(n.expression,!0)}else Lh(n.expression);aT(t),8388608&n.flags&&!e.isEntityNameExpression(n.expression)&&FS(n.expression,e.Diagnostics.The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context),!n.isExportEquals||8388608&n.flags||(H>=e.ModuleKind.ES2015?FS(n,e.Diagnostics.Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead):H===e.ModuleKind.System&&FS(n,e.Diagnostics.Export_assignment_is_not_supported_when_module_flag_is_system))}else n.isExportEquals?lr(n,e.Diagnostics.An_export_assignment_cannot_be_used_in_a_namespace):lr(n,e.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module)}}(n);case 231:case 248:return void BS(n);case 271:(function(e){Eb(e)})(n)}}(n),c=a}}function oT(n){e.isInJSFile(n)||FS(n,e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments)}function sT(n){var t=Sr(e.getSourceFileOfNode(n));if(!(1&t.flags)){t.deferredNodes=t.deferredNodes||new e.Map;var r=R(n);t.deferredNodes.set(r,n)}}function cT(n){var t=c;switch(c=n,h=0,n.kind){case 203:case 204:case 205:case 161:case 275:Ay(n);break;case 208:case 209:case 165:case 164:!function(n){e.Debug.assert(165!==n.kind||e.isObjectLiteralMethod(n));var t=e.getFunctionFlags(n),r=Oc(n);if(th(n,r),n.body)if(e.getEffectiveReturnTypeNode(n)||kc(Lc(n)),230===n.body.kind)iT(n.body);else{var a=Hh(n.body),i=r&&NE(r,t);i&&Qd(2==(3&t)?pb(a,n.body,e.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member):a,i,n.body,n.body)}}(n);break;case 167:case 168:Zh(n);break;case 221:!function(n){e.forEach(n.members,iT),xb(n)}(n);break;case 274:!function(e){Vg(e)}(n);break;case 273:!function(e){Vg(e.openingElement),xg(e.closingElement.tagName)?Ng(e.closingElement):Hh(e.closingElement.tagName),Ag(e)}(n)}c=t}function lT(t){var r=["check","checkSourceFile",{path:t.path}];e.tracing.begin.apply(e.tracing,r),e.performance.mark("beforeCheck"),function(t){var r=Sr(t);if(!(1&r.flags)){if(e.skipTypeChecking(t,V,n))return;!function(n){!!(8388608&n.flags)&&function(n){for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];if((e.isDeclaration(a)||232===a.kind)&&GS(a))return!0}}(n)}(t),e.clear(zt),e.clear(qt),e.clear(Jt),e.forEach(t.statements,iT),iT(t.endOfFileToken),function(e){var n=Sr(e);n.deferredNodes&&n.deferredNodes.forEach(cT)}(t),e.isExternalOrCommonJsModule(t)&&xb(t),t.isDeclarationFile||!V.noUnusedLocals&&!V.noUnusedParameters||Lb(dT(t),(function(n,t,r){!e.containsParseError(n)&&uT(t,!!(8388608&n.flags))&&Yt.add(r)})),2===V.importsNotUsedAsValues&&!t.isDeclarationFile&&e.isExternalModule(t)&&function(n){for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];e.isImportDeclaration(a)&&a.importClause&&!a.importClause.isTypeOnly&&(i=a.importClause,e.forEachImportClauseDeclaration(i,(function(e){return!!xa(e).isReferenced})))&&!BT(a.importClause,!0)&&!tT(a.importClause)&&lr(a,e.Diagnostics.This_import_is_never_used_as_a_value_and_must_use_import_type_because_the_importsNotUsedAsValues_is_set_to_error)}var i}(t),e.isExternalOrCommonJsModule(t)&&aT(t),zt.length&&(e.forEach(zt,Ub),e.clear(zt)),qt.length&&(e.forEach(qt,Vb),e.clear(qt)),Jt.length&&(e.forEach(Jt,jb),e.clear(Jt)),r.flags|=1}}(t),e.performance.mark("afterCheck"),e.performance.measure("Check","beforeCheck","afterCheck"),e.tracing.end.apply(e.tracing,r)}function uT(n,t){if(t)return!1;switch(n){case 0:return!!V.noUnusedLocals;case 1:return!!V.noUnusedParameters;default:return e.Debug.assertNever(n)}}function dT(n){return Tt.get(n.path)||e.emptyArray}function pT(t,a){try{return r=a,function(t){if(mT(),t){var r=Yt.getGlobalDiagnostics(),a=r.length;lT(t);var i=Yt.getDiagnostics(t.fileName),o=Yt.getGlobalDiagnostics();if(o!==r){var s=e.relativeComplement(r,o,e.compareDiagnostics);return e.concatenate(s,i)}return 0===a&&o.length>0?e.concatenate(o,i):i}return e.forEach(n.getSourceFiles(),lT),Yt.getDiagnostics()}(t)}finally{r=void 0}}function mT(){if(!t)throw new Error("Trying to get diagnostics from a type checker that does not produce them.")}function fT(e){switch(e.kind){case 159:case 252:case 253:case 254:case 255:case 331:case 324:case 325:return!0;case 262:return e.isTypeOnly;case 265:case 270:return e.parent.parent.isTypeOnly;default:return!1}}function _T(e){for(;157===e.parent.kind;)e=e.parent;return 173===e.parent.kind}function gT(n,t){for(var r;(n=e.getContainingClass(n))&&!(r=t(n)););return r}function yT(e,n){return!!gT(e,(function(e){return e===n}))}function vT(e){return void 0!==function(e){for(;157===e.parent.kind;)e=e.parent;return 260===e.parent.kind?e.parent.moduleReference===e?e.parent:void 0:266===e.parent.kind&&e.parent.expression===e?e.parent:void 0}(e)}function hT(n){if(e.isDeclarationName(n))return xa(n.parent);if(e.isInJSFile(n)&&201===n.parent.kind&&n.parent===n.parent.parent.left&&!e.isPrivateIdentifier(n)){var t=function(n){switch(e.getAssignmentDeclarationKind(n.parent.parent)){case 1:case 3:return xa(n.parent);case 4:case 2:case 5:return xa(n.parent.parent)}}(n);if(t)return t}if(266===n.parent.kind&&e.isEntityNameExpression(n)){var r=ca(n,2998271,!0);if(r&&r!==he)return r}else if(!e.isPropertyAccessExpression(n)&&!e.isPrivateIdentifier(n)&&vT(n)){var a=e.getAncestor(n,260);return e.Debug.assert(void 0!==a),oa(n,!0)}if(!e.isPropertyAccessExpression(n)&&!e.isPrivateIdentifier(n)){var i=function(n){for(var t=n.parent;e.isQualifiedName(t);)n=t,t=t.parent;if(t&&195===t.kind&&t.qualifier===n)return t}(n);if(i){sd(i);var o=Sr(n).resolvedSymbol;return o===he?void 0:o}}for(;e.isRightSideOfQualifiedNameOrPropertyAccess(n);)n=n.parent;if(function(e){for(;201===e.parent.kind;)e=e.parent;return 223===e.parent.kind}(n)){var s=0;223===n.parent.kind?(s=788968,e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)&&(s|=111551)):s=1920,s|=2097152;var c=e.isEntityNameExpression(n)?ca(n,s):void 0;if(c)return c}if(326===n.parent.kind)return e.getParameterSymbolFromJSDoc(n.parent);if(159===n.parent.kind&&330===n.parent.parent.kind){e.Debug.assert(!e.isInJSFile(n));var l=e.getTypeParameterFromJsDoc(n.parent);return l&&l.symbol}if(e.isExpressionNode(n)){if(e.nodeIsMissing(n))return;if(78===n.kind){if(e.isJSXTagName(n)&&xg(n)){var u=Ng(n.parent);return u===he?void 0:u}return ca(n,111551,!1,!0)}if(201===n.kind||157===n.kind){var d=Sr(n);return d.resolvedSymbol||(201===n.kind?ry(n):ay(n)),d.resolvedSymbol}}else{if(_T(n))return ca(n,s=173===n.parent.kind?788968:1920,!1,!0);if(function(e){for(;157===e.parent.kind;)e=e.parent;for(;201===e.parent.kind;)e=e.parent;return 302===e.parent.kind}(n))return ca(n,s=901119,!1,!0,e.getHostSignatureFromJSDoc(n))}return 172===n.parent.kind?ca(n,1):void 0}function bT(n,t){if(297===n.kind)return e.isExternalModule(n)?Sa(n.symbol):void 0;var r=n.parent,a=r.parent;if(!(16777216&n.flags)){if(M(n)){var i=xa(r);return e.isImportOrExportSpecifier(n.parent)&&n.parent.propertyName===n?bg(i):i}if(e.isLiteralComputedPropertyDeclarationName(n))return xa(r.parent);if(78===n.kind){if(vT(n))return hT(n);if(198===r.kind&&196===a.kind&&n===r.propertyName){var o=cc(ET(a),n.escapedText);if(o)return o}}switch(n.kind){case 78:case 79:case 201:case 157:return hT(n);case 107:var s=e.getThisContainer(n,!1);if(e.isFunctionLike(s)){var c=Lc(s);if(c.thisParameter)return c.thisParameter}if(e.isInExpressionContext(n))return Hh(n).symbol;case 187:return ad(n).symbol;case 105:return Hh(n).symbol;case 132:var l=n.parent;return l&&166===l.kind?l.parent.symbol:void 0;case 10:case 14:if(e.isExternalModuleImportEqualsDeclaration(n.parent.parent)&&e.getExternalModuleImportEqualsDeclarationExpression(n.parent.parent)===n||(261===n.parent.kind||267===n.parent.kind)&&n.parent.moduleSpecifier===n||e.isInJSFile(n)&&e.isRequireCall(n.parent,!1)||e.isImportCall(n.parent)||e.isLiteralTypeNode(n.parent)&&e.isLiteralImportTypeNode(n.parent.parent)&&n.parent.parent.argument===n.parent)return ua(n,n,t);if(e.isCallExpression(r)&&e.isBindableObjectDefinePropertyCall(r)&&r.arguments[1]===n)return xa(r);case 8:var u=e.isElementAccessExpression(r)?r.argumentExpression===n?Uh(r.expression):void 0:e.isLiteralTypeNode(r)&&e.isIndexedAccessTypeNode(a)?sd(a.objectType):void 0;return u&&cc(u,e.escapeLeadingUnderscores(n.text));case 87:case 97:case 38:case 83:return xa(n.parent);case 195:return e.isLiteralImportTypeNode(n)?bT(n.argument.literal,t):void 0;case 92:return e.isExportAssignment(n.parent)?e.Debug.checkDefined(n.parent.symbol):void 0;default:return}}}function ET(n){if(e.isSourceFile(n)&&!e.isExternalModule(n))return xe;if(16777216&n.flags)return xe;var t,r,a=e.tryGetClassImplementingOrExtendingExpressionWithTypeArguments(n),i=a&&Co(xa(a.class));if(e.isPartOfTypeNode(n)){var o=sd(n);return i?rs(o,i.thisType):o}if(e.isExpressionNode(n))return ST(n);if(i&&!a.isImplements){var s=e.firstOrUndefined(Lo(i));return s?rs(s,i.thisType):xe}if(fT(n))return Fo(r=xa(n));if(78===(t=n).kind&&fT(t.parent)&&e.getNameOfDeclaration(t.parent)===t)return(r=bT(n))?Fo(r):xe;if(e.isDeclaration(n))return lo(r=xa(n));if(M(n))return(r=bT(n))?lo(r):xe;if(e.isBindingPattern(n))return Mi(n.parent,!0)||xe;if(vT(n)&&(r=bT(n))){var c=Fo(r);return c!==xe?c:lo(r)}return xe}function TT(n){if(e.Debug.assert(200===n.kind||199===n.kind),239===n.parent.kind)return vh(n,rE(n.parent)||xe);if(216===n.parent.kind)return vh(n,Uh(n.parent.right)||xe);if(288===n.parent.kind){var t=e.cast(n.parent.parent,e.isObjectLiteralExpression);return gh(t,TT(t)||xe,e.indexOfNode(t.properties,n.parent))}var r=e.cast(n.parent,e.isArrayLiteralExpression),a=TT(r)||xe,i=aE(65,a,Ce,n.parent)||xe;return yh(r,a,r.elements.indexOf(n),i)}function ST(n){return e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent),ed(Uh(n))}function xT(n){var t=xa(n.parent);return e.hasSyntacticModifier(n,32)?lo(t):Fo(t)}function LT(n){var t=n.name;switch(t.kind){case 78:return td(e.idText(t));case 8:case 10:return td(t.text);case 158:var r=vg(t);return ph(r,12288)?r:Oe;default:return e.Debug.fail("Unsupported property name.")}}function AT(n){n=Zs(n);var t=e.createSymbolTable(Gs(n)),r=uc(n,0).length?hn:uc(n,1).length?bn:void 0;return r&&e.forEach(Gs(r),(function(e){t.has(e.escapedName)||t.set(e.escapedName,e)})),Ba(t)}function DT(n){return e.typeHasCallOrConstructSignatures(n,se)}function CT(n){if(e.isGeneratedIdentifier(n))return!1;var t=e.getParseTreeNode(n,e.isIdentifier);if(!t)return!1;var r=t.parent;return!!r&&(!((e.isPropertyAccessExpression(r)||e.isPropertyAssignment(r))&&r.name===t)&&nS(t)===ie)}function IT(n){var t=ua(n.parent,n);if(!t||e.isShorthandAmbientModuleSymbol(t))return!0;var r=ga(t),a=Tr(t=fa(t));return void 0===a.exportsSomeValue&&(a.exportsSomeValue=r?!!(111551&t.flags):e.forEachEntry(ba(t),(function(e){return(e=$r(e))&&!!(111551&e.flags)}))),a.exportsSomeValue}function NT(n,t){var r=e.getParseTreeNode(n,e.isIdentifier);if(r){var a=nS(r,function(n){return e.isModuleOrEnumDeclaration(n.parent)&&n===n.parent.name}(r));if(a){if(1048576&a.flags){var i=Sa(a.exportSymbol);if(!t&&944&i.flags&&!(3&i.flags))return;a=i}var o=La(a);if(o){if(512&o.flags&&297===o.valueDeclaration.kind){var s=o.valueDeclaration;return s!==e.getSourceFileOfNode(r)?void 0:s}return e.findAncestor(r.parent,(function(n){return e.isModuleOrEnumDeclaration(n)&&xa(n)===o}))}}}}function RT(n){if(n.generatedImportReference)return n.generatedImportReference;var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=nS(t);if(Zr(r,111551)&&!ra(r))return Br(r)}}function kT(n){if(418&n.flags&&!e.isSourceFile(n.valueDeclaration)){var t=Tr(n);if(void 0===t.isDeclarationWithCollidingName){var r=e.getEnclosingBlockScopeContainer(n.valueDeclaration);if(e.isStatementWithLocals(r)||function(n){return e.isBindingElement(n.valueDeclaration)&&287===e.walkUpBindingElementsAndPatterns(n.valueDeclaration).parent.kind}(n)){var a=Sr(n.valueDeclaration);if(Cr(r.parent,n.escapedName,111551,void 0,void 0,!1))t.isDeclarationWithCollidingName=!0;else if(262144&a.flags){var i=524288&a.flags,o=e.isIterationStatement(r,!1),s=230===r.kind&&e.isIterationStatement(r.parent,!1);t.isDeclarationWithCollidingName=!(e.isBlockScopedContainerTopLevel(r)||i&&(o||s))}else t.isDeclarationWithCollidingName=!1}}return t.isDeclarationWithCollidingName}return!1}function OT(n){if(!e.isGeneratedIdentifier(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=nS(t);if(r&&kT(r))return r.valueDeclaration}}}function PT(n){var t=e.getParseTreeNode(n,e.isDeclaration);if(t){var r=xa(t);if(r)return kT(r)}return!1}function MT(n){switch(n.kind){case 260:return FT(xa(n)||he);case 262:case 263:case 265:case 270:var t=xa(n)||he;return FT(t)&&!ra(t);case 267:var r=n.exportClause;return!!r&&(e.isNamespaceExport(r)||e.some(r.elements,MT));case 266:return!n.expression||78!==n.expression.kind||FT(xa(n)||he)}return!1}function wT(n){var t=e.getParseTreeNode(n,e.isImportEqualsDeclaration);return!(void 0===t||297!==t.parent.kind||!e.isInternalModuleImportEqualsDeclaration(t))&&(FT(xa(t))&&t.moduleReference&&!e.nodeIsMissing(t.moduleReference))}function FT(e){var n=ea(e);return n===he||!!(111551&n.flags)&&(V.preserveConstEnums||!GT(n))}function GT(e){return _h(e)||!!e.constEnumOnlyModule}function BT(n,t){if(Ur(n)){var r=xa(n);if(r&&Tr(r).referenced)return!0;var a=Tr(r).target;if(a&&1&e.getEffectiveModifierFlags(n)&&111551&a.flags&&(V.preserveConstEnums||!GT(a)))return!0}return!!t&&!!e.forEachChild(n,(function(e){return BT(e,t)}))}function UT(n){if(e.nodeIsPresent(n.body)){if(e.isGetAccessor(n)||e.isSetAccessor(n))return!1;var t=Cc(xa(n));return t.length>1||1===t.length&&t[0].declaration!==n}return!1}function VT(n){return!(!W||bc(n)||e.isJSDocParameterTag(n)||!n.initializer||e.hasSyntacticModifier(n,92))}function jT(n){return W&&bc(n)&&!n.initializer&&e.hasSyntacticModifier(n,92)}function HT(n){var t=e.getParseTreeNode(n,e.isFunctionDeclaration);if(!t)return!1;var r=xa(t);return!!(r&&16&r.flags)&&!!e.forEachEntry(ha(r),(function(n){return 111551&n.flags&&n.valueDeclaration&&e.isPropertyAccessExpression(n.valueDeclaration)}))}function KT(n){var t=e.getParseTreeNode(n,e.isFunctionDeclaration);if(!t)return e.emptyArray;var r=xa(t);return r&&Gs(lo(r))||e.emptyArray}function WT(e){return Sr(e).flags||0}function zT(e){return KE(e.parent),Sr(e).enumMemberValue}function qT(e){switch(e.kind){case 291:case 201:case 202:return!0}return!1}function JT(n){if(291===n.kind)return zT(n);var t=Sr(n).resolvedSymbol;if(t&&8&t.flags){var r=t.valueDeclaration;if(e.isEnumConst(r.parent))return zT(r)}}function XT(e){return!!(524288&e.flags)&&uc(e,0).length>0}function YT(n,t){var r,a=e.getParseTreeNode(n,e.isEntityName);if(!a)return e.TypeReferenceSerializationKind.Unknown;if(t&&!(t=e.getParseTreeNode(t)))return e.TypeReferenceSerializationKind.Unknown;var i=ca(a,111551,!0,!0,t),o=(null===(r=null==i?void 0:i.declarations)||void 0===r?void 0:r.every(e.isTypeOnlyImportOrExportDeclaration))||!1,s=i&&2097152&i.flags?ea(i):i,c=ca(a,788968,!0,!1,t);if(s&&s===c){var l=Ll(!1);if(l&&s===l)return e.TypeReferenceSerializationKind.Promise;var u=lo(s);if(u&&ho(u))return o?e.TypeReferenceSerializationKind.TypeWithCallSignature:e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue}if(!c)return o?e.TypeReferenceSerializationKind.ObjectType:e.TypeReferenceSerializationKind.Unknown;var d=Fo(c);return d===xe?o?e.TypeReferenceSerializationKind.ObjectType:e.TypeReferenceSerializationKind.Unknown:3&d.flags?e.TypeReferenceSerializationKind.ObjectType:ph(d,245760)?e.TypeReferenceSerializationKind.VoidNullableOrNeverType:ph(d,528)?e.TypeReferenceSerializationKind.BooleanType:ph(d,296)?e.TypeReferenceSerializationKind.NumberLikeType:ph(d,2112)?e.TypeReferenceSerializationKind.BigIntLikeType:ph(d,402653316)?e.TypeReferenceSerializationKind.StringLikeType:tm(d)?e.TypeReferenceSerializationKind.ArrayLikeType:ph(d,12288)?e.TypeReferenceSerializationKind.ESSymbolType:XT(d)?e.TypeReferenceSerializationKind.TypeWithCallSignature:Up(d)?e.TypeReferenceSerializationKind.ArrayLikeType:e.TypeReferenceSerializationKind.ObjectType}function QT(n,t,r,a,i){var o=e.getParseTreeNode(n,e.isVariableLikeOrAccessor);if(!o)return e.factory.createToken(128);var s=xa(o),c=!s||133120&s.flags?xe:Zp(lo(s));return 8192&c.flags&&c.symbol===s&&(r|=1048576),i&&(c=fm(c)),ee.typeToTypeNode(c,t,1024|r,a)}function ZT(n,t,r,a){var i=e.getParseTreeNode(n,e.isFunctionLike);if(!i)return e.factory.createToken(128);var o=Lc(i);return ee.typeToTypeNode(kc(o),t,1024|r,a)}function $T(n,t,r,a){var i=e.getParseTreeNode(n,e.isExpression);if(!i)return e.factory.createToken(128);var o=Im(ST(i));return ee.typeToTypeNode(o,t,1024|r,a)}function eS(n){return ne.has(e.escapeLeadingUnderscores(n))}function nS(n,t){var r=Sr(n).resolvedSymbol;if(r)return r;var a=n;if(t){var i=n.parent;e.isDeclaration(i)&&n===i.name&&(a=Si(i))}return Cr(a,n.escapedText,3257279,void 0,void 0,!0)}function tS(n){if(!e.isGeneratedIdentifier(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t){var r=nS(t);if(r)return Na(r).valueDeclaration}}}function rS(n){return!!(e.isDeclarationReadonly(n)||e.isVariableDeclaration(n)&&e.isVarConst(n))&&nd(lo(xa(n)))}function aS(n,t){return function(n,t,r){var a=1024&n.flags?ee.symbolToExpression(n.symbol,111551,t,void 0,r):n===Ge?e.factory.createTrue():n===we&&e.factory.createFalse();if(a)return a;var i=n.value;return"object"==typeof i?e.factory.createBigIntLiteral(i):"number"==typeof i?e.factory.createNumericLiteral(i):e.factory.createStringLiteral(i)}(lo(xa(n)),n,t)}function iS(n){return n?(sr(n),e.getSourceFileOfNode(n).localJsxFactory||at):at}function oS(n){if(n){var t=e.getSourceFileOfNode(n);if(t){if(t.localJsxFragmentFactory)return t.localJsxFragmentFactory;var r=t.pragmas.get("jsxfrag"),a=e.isArray(r)?r[0]:r;if(a)return t.localJsxFragmentFactory=e.parseIsolatedEntityName(a.arguments.factory,j),t.localJsxFragmentFactory}}if(V.jsxFragmentFactory)return e.parseIsolatedEntityName(V.jsxFragmentFactory,j)}function sS(n){var t=256===n.kind?e.tryCast(n.name,e.isStringLiteral):e.getExternalModuleName(n),r=da(t,t,void 0);if(r)return e.getDeclarationOfKind(r,297)}function cS(n,t){if((a&t)!==t&&V.importHelpers){var r=e.getSourceFileOfNode(n);if(e.isEffectiveExternalModule(r,V)&&!(8388608&n.flags)){var i=function(n,t){s||(s=pa(n,e.externalHelpersModuleNameText,e.Diagnostics.This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found,t)||he);return s}(r,n);if(i!==he)for(var o=t&~a,c=1;c<=4194304;c<<=1)if(o&c){var l=lS(c);Lr(i.exports,e.escapeLeadingUnderscores(l),111551)||lr(n,e.Diagnostics.This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_version_of_0,e.externalHelpersModuleNameText,l)}a|=t}}}function lS(n){switch(n){case 1:return"__extends";case 2:return"__assign";case 4:return"__rest";case 8:return"__decorate";case 16:return"__metadata";case 32:return"__param";case 64:return"__awaiter";case 128:return"__generator";case 256:return"__values";case 512:return"__read";case 1024:return"__spread";case 2048:return"__spreadArrays";case 4096:return"__await";case 8192:return"__asyncGenerator";case 16384:return"__asyncDelegator";case 32768:return"__asyncValues";case 65536:return"__exportStar";case 131072:return"__importStar";case 262144:return"__importDefault";case 524288:return"__makeTemplateObject";case 1048576:return"__classPrivateFieldGet";case 2097152:return"__classPrivateFieldSet";case 4194304:return"__createBinding";default:return e.Debug.fail("Unrecognized helper")}}function uS(n){return function(n){if(!n.decorators)return!1;if(!e.nodeCanBeDecorated(n,n.parent,n.parent.parent))return 165!==n.kind||e.nodeIsPresent(n.body)?MS(n,e.Diagnostics.Decorators_are_not_valid_here):MS(n,e.Diagnostics.A_decorator_can_only_decorate_a_method_implementation_not_an_overload);if(167===n.kind||168===n.kind){var t=e.getAllAccessorDeclarations(n.parent.members,n);if(t.firstAccessor.decorators&&n===t.secondAccessor)return MS(n,e.Diagnostics.Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name)}return!1}(n)||function(n){var t,r,a,i,o=function(n){return!!n.modifiers&&(function(n){switch(n.kind){case 167:case 168:case 166:case 163:case 162:case 165:case 164:case 171:case 256:case 261:case 260:case 267:case 266:case 208:case 209:case 160:return!1;default:if(257===n.parent.kind||297===n.parent.kind)return!1;switch(n.kind){case 251:return dS(n,129);case 252:return dS(n,125);case 253:case 232:case 254:return!0;case 255:return dS(n,84);default:return e.Debug.fail(),!1}}}(n)?MS(n,e.Diagnostics.Modifiers_cannot_appear_here):void 0)}(n);if(void 0!==o)return o;for(var s=0,c=0,l=n.modifiers;c<l.length;c++){var u=l[c];if(142!==u.kind){if(162===n.kind||164===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_type_member,e.tokenToString(u.kind));if(171===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_an_index_signature,e.tokenToString(u.kind))}switch(u.kind){case 84:if(255!==n.kind)return FS(n,e.Diagnostics.A_class_member_cannot_have_the_0_keyword,e.tokenToString(84));break;case 122:case 121:case 120:var d=pi(e.modifierToFlag(u.kind));if(28&s)return FS(u,e.Diagnostics.Accessibility_modifier_already_seen);if(32&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"static");if(64&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"readonly");if(256&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"async");if(257===n.parent.kind||297===n.parent.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element,d);if(128&s)return 120===u.kind?FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,d,"abstract"):FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,d,"abstract");if(e.isPrivateIdentifierPropertyDeclaration(n))return FS(u,e.Diagnostics.An_accessibility_modifier_cannot_be_used_with_a_private_identifier);s|=e.modifierToFlag(u.kind);break;case 123:if(32&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"static");if(64&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"static","readonly");if(256&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"static","async");if(257===n.parent.kind||297===n.parent.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element,"static");if(160===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"static");if(128&s)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"static","abstract");if(e.isPrivateIdentifierPropertyDeclaration(n))return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_a_private_identifier,"static");s|=32,t=u;break;case 142:if(64&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"readonly");if(163!==n.kind&&162!==n.kind&&171!==n.kind&&160!==n.kind)return FS(u,e.Diagnostics.readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature);s|=64,i=u;break;case 92:if(1&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"export");if(2&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","declare");if(128&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","abstract");if(256&s)return FS(u,e.Diagnostics._0_modifier_must_precede_1_modifier,"export","async");if(e.isClassLike(n.parent))return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_class_elements_of_this_kind,"export");if(160===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"export");s|=1;break;case 87:var p=297===n.parent.kind?n.parent:n.parent.parent;if(256===p.kind&&!e.isAmbientModule(p))return FS(u,e.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module);s|=512;break;case 133:if(2&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"declare");if(256&s)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context,"async");if(e.isClassLike(n.parent)&&!e.isPropertyDeclaration(n))return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_class_elements_of_this_kind,"declare");if(160===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"declare");if(8388608&n.parent.flags&&257===n.parent.kind)return FS(u,e.Diagnostics.A_declare_modifier_cannot_be_used_in_an_already_ambient_context);if(e.isPrivateIdentifierPropertyDeclaration(n))return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_a_private_identifier,"declare");s|=2,r=u;break;case 125:if(128&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"abstract");if(252!==n.kind){if(165!==n.kind&&163!==n.kind&&167!==n.kind&&168!==n.kind)return FS(u,e.Diagnostics.abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration);if(252!==n.parent.kind||!e.hasSyntacticModifier(n.parent,128))return FS(u,e.Diagnostics.Abstract_methods_can_only_appear_within_an_abstract_class);if(32&s)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"static","abstract");if(8&s)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"private","abstract");if(256&s&&a)return FS(a,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"async","abstract")}if(e.isNamedDeclaration(n)&&79===n.name.kind)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_a_private_identifier,"abstract");s|=128;break;case 129:if(256&s)return FS(u,e.Diagnostics._0_modifier_already_seen,"async");if(2&s||8388608&n.parent.flags)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context,"async");if(160===n.kind)return FS(u,e.Diagnostics._0_modifier_cannot_appear_on_a_parameter,"async");if(128&s)return FS(u,e.Diagnostics._0_modifier_cannot_be_used_with_1_modifier,"async","abstract");s|=256,a=u}}if(166===n.kind)return 32&s?FS(t,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"static"):128&s?FS(t,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"abstract"):256&s?FS(a,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"async"):!!(64&s)&&FS(i,e.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration,"readonly");if((261===n.kind||260===n.kind)&&2&s)return FS(r,e.Diagnostics.A_0_modifier_cannot_be_used_with_an_import_declaration,"declare");if(160===n.kind&&92&s&&e.isBindingPattern(n.name))return FS(n,e.Diagnostics.A_parameter_property_may_not_be_declared_using_a_binding_pattern);if(160===n.kind&&92&s&&n.dotDotDotToken)return FS(n,e.Diagnostics.A_parameter_property_cannot_be_declared_using_a_rest_parameter);if(256&s)return function(n,t){switch(n.kind){case 165:case 251:case 208:case 209:return!1}return FS(t,e.Diagnostics._0_modifier_cannot_be_used_here,"async")}(n,a);return!1}(n)}function dS(e,n){return e.modifiers.length>1||e.modifiers[0].kind!==n}function pS(n,t){return void 0===t&&(t=e.Diagnostics.Trailing_comma_not_allowed),!(!n||!n.hasTrailingComma)&&wS(n[0],n.end-",".length,",".length,t)}function mS(n,t){if(n&&0===n.length){var r=n.pos-"<".length;return wS(t,r,e.skipTrivia(t.text,n.end)+">".length-r,e.Diagnostics.Type_parameter_list_cannot_be_empty)}return!1}function fS(n){if(j>=3){var t=n.body&&e.isBlock(n.body)&&e.findUseStrictPrologue(n.body.statements);if(t){var r=(i=n.parameters,e.filter(i,(function(n){return!!n.initializer||e.isBindingPattern(n.name)||e.isRestParameter(n)})));if(e.length(r)){e.forEach(r,(function(n){e.addRelatedInfo(lr(n,e.Diagnostics.This_parameter_is_not_allowed_with_use_strict_directive),e.createDiagnosticForNode(t,e.Diagnostics.use_strict_directive_used_here))}));var a=r.map((function(n,t){return 0===t?e.createDiagnosticForNode(n,e.Diagnostics.Non_simple_parameter_declared_here):e.createDiagnosticForNode(n,e.Diagnostics.and_here)}));return e.addRelatedInfo.apply(void 0,__spreadArrays([lr(t,e.Diagnostics.use_strict_directive_cannot_be_used_with_non_simple_parameter_list)],a)),!0}}}var i;return!1}function _S(n){var t=e.getSourceFileOfNode(n);return uS(n)||mS(n.typeParameters,t)||function(n){for(var t=!1,r=n.length,a=0;a<r;a++){var i=n[a];if(i.dotDotDotToken){if(a!==r-1)return FS(i.dotDotDotToken,e.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);if(8388608&i.flags||pS(n,e.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma),i.questionToken)return FS(i.questionToken,e.Diagnostics.A_rest_parameter_cannot_be_optional);if(i.initializer)return FS(i.name,e.Diagnostics.A_rest_parameter_cannot_have_an_initializer)}else if(bc(i)){if(t=!0,i.questionToken&&i.initializer)return FS(i.name,e.Diagnostics.Parameter_cannot_have_question_mark_and_initializer)}else if(t&&!i.initializer)return FS(i.name,e.Diagnostics.A_required_parameter_cannot_follow_an_optional_parameter)}}(n.parameters)||function(n,t){if(!e.isArrowFunction(n))return!1;var r=n.equalsGreaterThanToken,a=e.getLineAndCharacterOfPosition(t,r.pos).line,i=e.getLineAndCharacterOfPosition(t,r.end).line;return a!==i&&FS(r,e.Diagnostics.Line_terminator_not_permitted_before_arrow)}(n,t)||e.isFunctionLikeDeclaration(n)&&fS(n)}function gS(n,t){return pS(t)||function(n,t){if(t&&0===t.length){var r=e.getSourceFileOfNode(n),a=t.pos-"<".length;return wS(r,a,e.skipTrivia(r.text,t.end)+">".length-a,e.Diagnostics.Type_argument_list_cannot_be_empty)}return!1}(n,t)}function yS(n){return function(n){if(n)for(var t=0,r=n;t<r.length;t++){var a=r[t];if(222===a.kind)return wS(a,a.pos,0,e.Diagnostics.Argument_expression_expected)}return!1}(n)}function vS(n){var t=n.types;if(pS(t))return!0;if(t&&0===t.length){var r=e.tokenToString(n.token);return wS(n,t.pos,0,e.Diagnostics._0_list_cannot_be_empty,r)}return e.some(t,hS)}function hS(e){return gS(e,e.typeArguments)}function bS(n){if(158!==n.kind)return!1;var t=n;return 216===t.expression.kind&&27===t.expression.operatorToken.kind&&FS(t.expression,e.Diagnostics.A_comma_expression_is_not_allowed_in_a_computed_property_name)}function ES(n){if(n.asteriskToken){if(e.Debug.assert(251===n.kind||208===n.kind||165===n.kind),8388608&n.flags)return FS(n.asteriskToken,e.Diagnostics.Generators_are_not_allowed_in_an_ambient_context);if(!n.body)return FS(n.asteriskToken,e.Diagnostics.An_overload_signature_cannot_be_declared_as_a_generator)}}function TS(e,n){return!!e&&FS(e,n)}function SS(e,n){return!!e&&FS(e,n)}function xS(n){if(BS(n))return!0;if(239===n.kind&&n.awaitModifier&&0==(32768&n.flags)){if(!PS(e.getSourceFileOfNode(n))){var t=e.createDiagnosticForNode(n.awaitModifier,e.Diagnostics.A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator),r=e.getContainingFunction(n);if(r&&166!==r.kind){e.Debug.assert(0==(2&e.getFunctionFlags(r)),"Enclosing function should never be an async function.");var a=e.createDiagnosticForNode(r,e.Diagnostics.Did_you_mean_to_mark_this_function_as_async);e.addRelatedInfo(t,a)}return Yt.add(t),!0}return!1}if(250===n.initializer.kind){var i=n.initializer;if(!kS(i)){var o=i.declarations;if(!o.length)return!1;if(o.length>1){t=238===n.kind?e.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement:e.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement;return MS(i.declarations[1],t)}var s=o[0];if(s.initializer){var t=238===n.kind?e.Diagnostics.The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer:e.Diagnostics.The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer;return FS(s.name,t)}if(s.type)return FS(s,t=238===n.kind?e.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation:e.Diagnostics.The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation)}}return!1}function LS(n){if(n.parameters.length===(167===n.kind?1:2))return e.getThisParameter(n)}function AS(n,t){if(function(n){return e.isDynamicName(n)&&!Jo(n)}(n))return FS(n,t)}function DS(n){if(_S(n))return!0;if(165===n.kind){if(200===n.parent.kind){if(n.modifiers&&(1!==n.modifiers.length||129!==e.first(n.modifiers).kind))return MS(n,e.Diagnostics.Modifiers_cannot_appear_here);if(TS(n.questionToken,e.Diagnostics.An_object_member_cannot_be_declared_optional))return!0;if(SS(n.exclamationToken,e.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context))return!0;if(void 0===n.body)return wS(n,n.end-1,";".length,e.Diagnostics._0_expected,"{")}if(ES(n))return!0}if(e.isClassLike(n.parent)){if(8388608&n.flags)return AS(n.name,e.Diagnostics.A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);if(165===n.kind&&!n.body)return AS(n.name,e.Diagnostics.A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)}else{if(253===n.parent.kind)return AS(n.name,e.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);if(177===n.parent.kind)return AS(n.name,e.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)}}function CS(n){return e.isStringOrNumericLiteralLike(n)||214===n.kind&&40===n.operator&&8===n.operand.kind}function IS(n){var t,r=n.initializer;if(r){var a=!(CS(r)||function(n){if((e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n)&&CS(n.argumentExpression))&&e.isEntityNameExpression(n.expression))return!!(1024&Lh(n).flags)}(r)||109===r.kind||94===r.kind||(t=r,9===t.kind||214===t.kind&&40===t.operator&&9===t.operand.kind)),i=e.isDeclarationReadonly(n)||e.isVariableDeclaration(n)&&e.isVarConst(n);if(!i||n.type)return FS(r,e.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts);if(a)return FS(r,e.Diagnostics.A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference);if(!i||a)return FS(r,e.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts)}}function NS(n){if(78===n.kind){if("__esModule"===e.idText(n))return function(n,t,r,a,i,o){if(!PS(e.getSourceFileOfNode(t)))return cr(n,t,r,a,i,o),!0;return!1}("noEmit",n,e.Diagnostics.Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules)}else for(var t=0,r=n.elements;t<r.length;t++){var a=r[t];if(!e.isOmittedExpression(a))return NS(a.name)}return!1}function RS(n){if(78===n.kind){if(118===n.originalKeywordKind)return FS(n,e.Diagnostics.let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations)}else for(var t=0,r=n.elements;t<r.length;t++){var a=r[t];e.isOmittedExpression(a)||RS(a.name)}return!1}function kS(n){var t=n.declarations;return!!pS(n.declarations)||!n.declarations.length&&wS(n,t.pos,t.end-t.pos,e.Diagnostics.Variable_declaration_list_cannot_be_empty)}function OS(e){switch(e.kind){case 234:case 235:case 236:case 243:case 237:case 238:case 239:return!1;case 245:return OS(e.parent)}return!0}function PS(e){return e.parseDiagnostics.length>0}function MS(n,t,r,a,i){var o=e.getSourceFileOfNode(n);if(!PS(o)){var s=e.getSpanOfTokenAtPosition(o,n.pos);return Yt.add(e.createFileDiagnostic(o,s.start,s.length,t,r,a,i)),!0}return!1}function wS(n,t,r,a,i,o,s){var c=e.getSourceFileOfNode(n);return!PS(c)&&(Yt.add(e.createFileDiagnostic(c,t,r,a,i,o,s)),!0)}function FS(n,t,r,a,i){return!PS(e.getSourceFileOfNode(n))&&(Yt.add(e.createDiagnosticForNode(n,t,r,a,i)),!0)}function GS(n){return 253!==n.kind&&254!==n.kind&&261!==n.kind&&260!==n.kind&&267!==n.kind&&266!==n.kind&&259!==n.kind&&!e.hasSyntacticModifier(n,515)&&MS(n,e.Diagnostics.Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier)}function BS(n){if(8388608&n.flags){if(!Sr(n).hasReportedStatementInAmbientContext&&(e.isFunctionLike(n.parent)||e.isAccessor(n.parent)))return Sr(n).hasReportedStatementInAmbientContext=MS(n,e.Diagnostics.An_implementation_cannot_be_declared_in_ambient_contexts);if(230===n.parent.kind||257===n.parent.kind||297===n.parent.kind){var t=Sr(n.parent);if(!t.hasReportedStatementInAmbientContext)return t.hasReportedStatementInAmbientContext=MS(n,e.Diagnostics.Statements_are_not_allowed_in_ambient_contexts)}}return!1}function US(n){if(32&n.numericLiteralFlags){var t=void 0;if(j>=1?t=e.Diagnostics.Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0:e.isChildOfNodeWithKind(n,191)?t=e.Diagnostics.Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0:e.isChildOfNodeWithKind(n,291)&&(t=e.Diagnostics.Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0),t){var r=e.isPrefixUnaryExpression(n.parent)&&40===n.parent.operator,a=(r?"-":"")+"0o"+n.text;return FS(r?n.parent:n,t,a)}}return function(n){if(16&n.numericLiteralFlags||n.text.length<=15||-1!==n.text.indexOf("."))return;var t=+e.getTextOfNode(n);if(t<=Math.pow(2,53)-1&&t+1>t)return;ur(!1,e.createDiagnosticForNode(n,e.Diagnostics.Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers))}(n),!1}function VS(n,t,r,a){if(1048576&t.flags&&2621440&n.flags){var i=Gs(n);if(i){var o=Sf(i,t);if(o)return Ep(t,e.map(o,(function(e){return[function(){return lo(e)},e.escapedName]})),r,void 0,a)}}}},function(e){e.JSX="JSX",e.IntrinsicElements="IntrinsicElements",e.ElementClass="ElementClass",e.ElementAttributesPropertyNameContainer="ElementAttributesProperty",e.ElementChildrenAttributeNameContainer="ElementChildrenAttribute",e.Element="Element",e.IntrinsicAttributes="IntrinsicAttributes",e.IntrinsicClassAttributes="IntrinsicClassAttributes",e.LibraryManagedAttributes="LibraryManagedAttributes"}(D||(D={})),e.signatureHasRestParameter=G,e.signatureHasLiteralTypes=B}(ts||(ts={})),function(e){var n=e.or(e.isTypeNode,e.isTypeParameterDeclaration);function t(n,t,r,a){if(void 0===n||void 0===t)return n;var i,o=t(n);return o===n?n:void 0!==o?(i=e.isArray(o)?(a||c)(o):o,e.Debug.assertNode(i,r),i):void 0}function r(n,t,r,a,i){if(void 0===n||void 0===t)return n;var o,s,c=n.length;(void 0===a||a<0)&&(a=0),(void 0===i||i>c-a)&&(i=c-a);var l=-1,u=-1;(a>0||i<c)&&(o=[],s=n.hasTrailingComma&&a+i===c);for(var d=0;d<i;d++){var p=n[d+a],m=void 0!==p?t(p):void 0;if((void 0!==o||void 0===m||m!==p)&&(void 0===o&&(o=n.slice(0,d),s=n.hasTrailingComma,l=n.pos,u=n.end),m))if(e.isArray(m))for(var f=0,_=m;f<_.length;f++){var g=_[f];e.Debug.assertNode(g,r),o.push(g)}else e.Debug.assertNode(m,r),o.push(m)}if(o){var y=e.factory.createNodeArray(o,s);return e.setTextRangePosEnd(y,l,u),y}return n}function a(n,t,a,i,o,s){return void 0===s&&(s=r),a.startLexicalEnvironment(),n=s(n,t,e.isStatement,i),o&&(n=a.factory.ensureUseStrict(n)),e.factory.mergeLexicalEnvironment(n,a.endLexicalEnvironment())}function i(n,t,a,i){var s;return void 0===i&&(i=r),a.startLexicalEnvironment(),n&&(a.setLexicalEnvironmentFlags(1,!0),s=i(n,t,e.isParameterDeclaration),2&a.getLexicalEnvironmentFlags()&&e.getEmitScriptTarget(a.getCompilerOptions())>=2&&(s=function(n,t){for(var r,a=0;a<n.length;a++){var i=n[a],s=o(i,t);(r||s!==i)&&(r||(r=n.slice(0,a)),r[a]=s)}if(r)return e.setTextRange(t.factory.createNodeArray(r,n.hasTrailingComma),n);return n}(s,a)),a.setLexicalEnvironmentFlags(1,!1)),a.suspendLexicalEnvironment(),s}function o(n,t){return n.dotDotDotToken?n:e.isBindingPattern(n.name)?function(e,n){var t=n.factory;return n.addInitializationStatement(t.createVariableStatement(void 0,t.createVariableDeclarationList([t.createVariableDeclaration(e.name,void 0,e.type,e.initializer?t.createConditionalExpression(t.createStrictEquality(t.getGeneratedNameForNode(e),t.createVoidZero()),void 0,e.initializer,void 0,t.getGeneratedNameForNode(e)):t.getGeneratedNameForNode(e))]))),t.updateParameterDeclaration(e,e.decorators,e.modifiers,e.dotDotDotToken,t.getGeneratedNameForNode(e),e.questionToken,e.type,void 0)}(n,t):n.initializer?function(n,t,r,a){var i=a.factory;return a.addInitializationStatement(i.createIfStatement(i.createTypeCheck(i.cloneNode(t),"undefined"),e.setEmitFlags(e.setTextRange(i.createBlock([i.createExpressionStatement(e.setEmitFlags(e.setTextRange(i.createAssignment(e.setEmitFlags(i.cloneNode(t),48),e.setEmitFlags(r,1584|e.getEmitFlags(r))),n),1536))]),n),1953))),i.updateParameterDeclaration(n,n.decorators,n.modifiers,n.dotDotDotToken,n.name,n.questionToken,n.type,void 0)}(n,n.name,n.initializer,t):n}function s(n,r,a,i){void 0===i&&(i=t),a.resumeLexicalEnvironment();var o=i(n,r,e.isConciseBody),s=a.endLexicalEnvironment();if(e.some(s)){if(!o)return a.factory.createBlock(s);var c=a.factory.converters.convertToFunctionBlock(o),l=e.factory.mergeLexicalEnvironment(c.statements,s);return a.factory.updateBlock(c,l)}return o}function c(n){return e.Debug.assert(n.length<=1,"Too many nodes written to output."),e.singleOrUndefined(n)}e.visitNode=t,e.visitNodes=r,e.visitLexicalEnvironment=a,e.visitParameterList=i,e.visitFunctionBody=s,e.visitEachChild=function(o,c,l,u,d,p){if(void 0===u&&(u=r),void 0===p&&(p=t),void 0!==o){var m=o.kind;if(m>0&&m<=156||187===m)return o;var f=l.factory;switch(m){case 78:return f.updateIdentifier(o,u(o.typeArguments,c,n));case 157:return f.updateQualifiedName(o,p(o.left,c,e.isEntityName),p(o.right,c,e.isIdentifier));case 158:return f.updateComputedPropertyName(o,p(o.expression,c,e.isExpression));case 159:return f.updateTypeParameterDeclaration(o,p(o.name,c,e.isIdentifier),p(o.constraint,c,e.isTypeNode),p(o.default,c,e.isTypeNode));case 160:return f.updateParameterDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.dotDotDotToken,d,e.isToken),p(o.name,c,e.isBindingName),p(o.questionToken,d,e.isToken),p(o.type,c,e.isTypeNode),p(o.initializer,c,e.isExpression));case 161:return f.updateDecorator(o,p(o.expression,c,e.isExpression));case 162:return f.updatePropertySignature(o,u(o.modifiers,c,e.isToken),p(o.name,c,e.isPropertyName),p(o.questionToken,d,e.isToken),p(o.type,c,e.isTypeNode));case 163:return f.updatePropertyDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isPropertyName),p(o.questionToken||o.exclamationToken,d,e.isToken),p(o.type,c,e.isTypeNode),p(o.initializer,c,e.isExpression));case 164:return f.updateMethodSignature(o,u(o.modifiers,c,e.isModifier),p(o.name,c,e.isPropertyName),p(o.questionToken,d,e.isToken),u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 165:return f.updateMethodDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.asteriskToken,d,e.isToken),p(o.name,c,e.isPropertyName),p(o.questionToken,d,e.isToken),u(o.typeParameters,c,e.isTypeParameterDeclaration),i(o.parameters,c,l,u),p(o.type,c,e.isTypeNode),s(o.body,c,l,p));case 166:return f.updateConstructorDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),i(o.parameters,c,l,u),s(o.body,c,l,p));case 167:return f.updateGetAccessorDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isPropertyName),i(o.parameters,c,l,u),p(o.type,c,e.isTypeNode),s(o.body,c,l,p));case 168:return f.updateSetAccessorDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isPropertyName),i(o.parameters,c,l,u),s(o.body,c,l,p));case 169:return f.updateCallSignature(o,u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 170:return f.updateConstructSignature(o,u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 171:return f.updateIndexSignature(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 172:return f.updateTypePredicateNode(o,p(o.assertsModifier,c),p(o.parameterName,c),p(o.type,c,e.isTypeNode));case 173:return f.updateTypeReferenceNode(o,p(o.typeName,c,e.isEntityName),u(o.typeArguments,c,e.isTypeNode));case 174:return f.updateFunctionTypeNode(o,u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 175:return f.updateConstructorTypeNode(o,u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.parameters,c,e.isParameterDeclaration),p(o.type,c,e.isTypeNode));case 176:return f.updateTypeQueryNode(o,p(o.exprName,c,e.isEntityName));case 177:return f.updateTypeLiteralNode(o,u(o.members,c,e.isTypeElement));case 178:return f.updateArrayTypeNode(o,p(o.elementType,c,e.isTypeNode));case 179:return f.updateTupleTypeNode(o,u(o.elements,c,e.isTypeNode));case 180:return f.updateOptionalTypeNode(o,p(o.type,c,e.isTypeNode));case 181:return f.updateRestTypeNode(o,p(o.type,c,e.isTypeNode));case 182:return f.updateUnionTypeNode(o,u(o.types,c,e.isTypeNode));case 183:return f.updateIntersectionTypeNode(o,u(o.types,c,e.isTypeNode));case 184:return f.updateConditionalTypeNode(o,p(o.checkType,c,e.isTypeNode),p(o.extendsType,c,e.isTypeNode),p(o.trueType,c,e.isTypeNode),p(o.falseType,c,e.isTypeNode));case 185:return f.updateInferTypeNode(o,p(o.typeParameter,c,e.isTypeParameterDeclaration));case 195:return f.updateImportTypeNode(o,p(o.argument,c,e.isTypeNode),p(o.qualifier,c,e.isEntityName),r(o.typeArguments,c,e.isTypeNode),o.isTypeOf);case 192:return f.updateNamedTupleMember(o,t(o.dotDotDotToken,c,e.isToken),t(o.name,c,e.isIdentifier),t(o.questionToken,c,e.isToken),t(o.type,c,e.isTypeNode));case 186:return f.updateParenthesizedType(o,p(o.type,c,e.isTypeNode));case 188:return f.updateTypeOperatorNode(o,p(o.type,c,e.isTypeNode));case 189:return f.updateIndexedAccessTypeNode(o,p(o.objectType,c,e.isTypeNode),p(o.indexType,c,e.isTypeNode));case 190:return f.updateMappedTypeNode(o,p(o.readonlyToken,d,e.isToken),p(o.typeParameter,c,e.isTypeParameterDeclaration),p(o.nameType,c,e.isTypeNode),p(o.questionToken,d,e.isToken),p(o.type,c,e.isTypeNode));case 191:return f.updateLiteralTypeNode(o,p(o.literal,c,e.isExpression));case 193:return f.updateTemplateLiteralType(o,p(o.head,c,e.isTemplateHead),u(o.templateSpans,c,e.isTemplateLiteralTypeSpan));case 194:return f.updateTemplateLiteralTypeSpan(o,p(o.type,c,e.isTypeNode),p(o.literal,c,e.isTemplateMiddleOrTemplateTail));case 196:return f.updateObjectBindingPattern(o,u(o.elements,c,e.isBindingElement));case 197:return f.updateArrayBindingPattern(o,u(o.elements,c,e.isArrayBindingElement));case 198:return f.updateBindingElement(o,p(o.dotDotDotToken,d,e.isToken),p(o.propertyName,c,e.isPropertyName),p(o.name,c,e.isBindingName),p(o.initializer,c,e.isExpression));case 199:return f.updateArrayLiteralExpression(o,u(o.elements,c,e.isExpression));case 200:return f.updateObjectLiteralExpression(o,u(o.properties,c,e.isObjectLiteralElementLike));case 201:return 32&o.flags?f.updatePropertyAccessChain(o,p(o.expression,c,e.isExpression),p(o.questionDotToken,d,e.isToken),p(o.name,c,e.isIdentifier)):f.updatePropertyAccessExpression(o,p(o.expression,c,e.isExpression),p(o.name,c,e.isIdentifierOrPrivateIdentifier));case 202:return 32&o.flags?f.updateElementAccessChain(o,p(o.expression,c,e.isExpression),p(o.questionDotToken,d,e.isToken),p(o.argumentExpression,c,e.isExpression)):f.updateElementAccessExpression(o,p(o.expression,c,e.isExpression),p(o.argumentExpression,c,e.isExpression));case 203:return 32&o.flags?f.updateCallChain(o,p(o.expression,c,e.isExpression),p(o.questionDotToken,d,e.isToken),u(o.typeArguments,c,e.isTypeNode),u(o.arguments,c,e.isExpression)):f.updateCallExpression(o,p(o.expression,c,e.isExpression),u(o.typeArguments,c,e.isTypeNode),u(o.arguments,c,e.isExpression));case 204:return f.updateNewExpression(o,p(o.expression,c,e.isExpression),u(o.typeArguments,c,e.isTypeNode),u(o.arguments,c,e.isExpression));case 205:return f.updateTaggedTemplateExpression(o,p(o.tag,c,e.isExpression),r(o.typeArguments,c,e.isExpression),p(o.template,c,e.isTemplateLiteral));case 206:return f.updateTypeAssertion(o,p(o.type,c,e.isTypeNode),p(o.expression,c,e.isExpression));case 207:return f.updateParenthesizedExpression(o,p(o.expression,c,e.isExpression));case 208:return f.updateFunctionExpression(o,u(o.modifiers,c,e.isModifier),p(o.asteriskToken,d,e.isToken),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),i(o.parameters,c,l,u),p(o.type,c,e.isTypeNode),s(o.body,c,l,p));case 209:return f.updateArrowFunction(o,u(o.modifiers,c,e.isModifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),i(o.parameters,c,l,u),p(o.type,c,e.isTypeNode),p(o.equalsGreaterThanToken,d,e.isToken),s(o.body,c,l,p));case 210:return f.updateDeleteExpression(o,p(o.expression,c,e.isExpression));case 211:return f.updateTypeOfExpression(o,p(o.expression,c,e.isExpression));case 212:return f.updateVoidExpression(o,p(o.expression,c,e.isExpression));case 213:return f.updateAwaitExpression(o,p(o.expression,c,e.isExpression));case 214:return f.updatePrefixUnaryExpression(o,p(o.operand,c,e.isExpression));case 215:return f.updatePostfixUnaryExpression(o,p(o.operand,c,e.isExpression));case 216:return f.updateBinaryExpression(o,p(o.left,c,e.isExpression),p(o.operatorToken,d,e.isToken),p(o.right,c,e.isExpression));case 217:return f.updateConditionalExpression(o,p(o.condition,c,e.isExpression),p(o.questionToken,d,e.isToken),p(o.whenTrue,c,e.isExpression),p(o.colonToken,d,e.isToken),p(o.whenFalse,c,e.isExpression));case 218:return f.updateTemplateExpression(o,p(o.head,c,e.isTemplateHead),u(o.templateSpans,c,e.isTemplateSpan));case 219:return f.updateYieldExpression(o,p(o.asteriskToken,d,e.isToken),p(o.expression,c,e.isExpression));case 220:return f.updateSpreadElement(o,p(o.expression,c,e.isExpression));case 221:return f.updateClassExpression(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.heritageClauses,c,e.isHeritageClause),u(o.members,c,e.isClassElement));case 223:return f.updateExpressionWithTypeArguments(o,p(o.expression,c,e.isExpression),u(o.typeArguments,c,e.isTypeNode));case 224:return f.updateAsExpression(o,p(o.expression,c,e.isExpression),p(o.type,c,e.isTypeNode));case 225:return 32&o.flags?f.updateNonNullChain(o,p(o.expression,c,e.isExpression)):f.updateNonNullExpression(o,p(o.expression,c,e.isExpression));case 226:return f.updateMetaProperty(o,p(o.name,c,e.isIdentifier));case 228:return f.updateTemplateSpan(o,p(o.expression,c,e.isExpression),p(o.literal,c,e.isTemplateMiddleOrTemplateTail));case 230:return f.updateBlock(o,u(o.statements,c,e.isStatement));case 232:return f.updateVariableStatement(o,u(o.modifiers,c,e.isModifier),p(o.declarationList,c,e.isVariableDeclarationList));case 233:return f.updateExpressionStatement(o,p(o.expression,c,e.isExpression));case 234:return f.updateIfStatement(o,p(o.expression,c,e.isExpression),p(o.thenStatement,c,e.isStatement,f.liftToBlock),p(o.elseStatement,c,e.isStatement,f.liftToBlock));case 235:return f.updateDoStatement(o,p(o.statement,c,e.isStatement,f.liftToBlock),p(o.expression,c,e.isExpression));case 236:return f.updateWhileStatement(o,p(o.expression,c,e.isExpression),p(o.statement,c,e.isStatement,f.liftToBlock));case 237:return f.updateForStatement(o,p(o.initializer,c,e.isForInitializer),p(o.condition,c,e.isExpression),p(o.incrementor,c,e.isExpression),p(o.statement,c,e.isStatement,f.liftToBlock));case 238:return f.updateForInStatement(o,p(o.initializer,c,e.isForInitializer),p(o.expression,c,e.isExpression),p(o.statement,c,e.isStatement,f.liftToBlock));case 239:return f.updateForOfStatement(o,p(o.awaitModifier,d,e.isToken),p(o.initializer,c,e.isForInitializer),p(o.expression,c,e.isExpression),p(o.statement,c,e.isStatement,f.liftToBlock));case 240:return f.updateContinueStatement(o,p(o.label,c,e.isIdentifier));case 241:return f.updateBreakStatement(o,p(o.label,c,e.isIdentifier));case 242:return f.updateReturnStatement(o,p(o.expression,c,e.isExpression));case 243:return f.updateWithStatement(o,p(o.expression,c,e.isExpression),p(o.statement,c,e.isStatement,f.liftToBlock));case 244:return f.updateSwitchStatement(o,p(o.expression,c,e.isExpression),p(o.caseBlock,c,e.isCaseBlock));case 245:return f.updateLabeledStatement(o,p(o.label,c,e.isIdentifier),p(o.statement,c,e.isStatement,f.liftToBlock));case 246:return f.updateThrowStatement(o,p(o.expression,c,e.isExpression));case 247:return f.updateTryStatement(o,p(o.tryBlock,c,e.isBlock),p(o.catchClause,c,e.isCatchClause),p(o.finallyBlock,c,e.isBlock));case 249:return f.updateVariableDeclaration(o,p(o.name,c,e.isBindingName),p(o.exclamationToken,d,e.isToken),p(o.type,c,e.isTypeNode),p(o.initializer,c,e.isExpression));case 250:return f.updateVariableDeclarationList(o,u(o.declarations,c,e.isVariableDeclaration));case 251:return f.updateFunctionDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.asteriskToken,d,e.isToken),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),i(o.parameters,c,l,u),p(o.type,c,e.isTypeNode),s(o.body,c,l,p));case 252:return f.updateClassDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.heritageClauses,c,e.isHeritageClause),u(o.members,c,e.isClassElement));case 253:return f.updateInterfaceDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),u(o.heritageClauses,c,e.isHeritageClause),u(o.members,c,e.isTypeElement));case 254:return f.updateTypeAliasDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),u(o.typeParameters,c,e.isTypeParameterDeclaration),p(o.type,c,e.isTypeNode));case 255:return f.updateEnumDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),u(o.members,c,e.isEnumMember));case 256:return f.updateModuleDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),p(o.body,c,e.isModuleBody));case 257:return f.updateModuleBlock(o,u(o.statements,c,e.isStatement));case 258:return f.updateCaseBlock(o,u(o.clauses,c,e.isCaseOrDefaultClause));case 259:return f.updateNamespaceExportDeclaration(o,p(o.name,c,e.isIdentifier));case 260:return f.updateImportEqualsDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.name,c,e.isIdentifier),p(o.moduleReference,c,e.isModuleReference));case 261:return f.updateImportDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.importClause,c,e.isImportClause),p(o.moduleSpecifier,c,e.isExpression));case 262:return f.updateImportClause(o,o.isTypeOnly,p(o.name,c,e.isIdentifier),p(o.namedBindings,c,e.isNamedImportBindings));case 263:return f.updateNamespaceImport(o,p(o.name,c,e.isIdentifier));case 269:return f.updateNamespaceExport(o,p(o.name,c,e.isIdentifier));case 264:return f.updateNamedImports(o,u(o.elements,c,e.isImportSpecifier));case 265:return f.updateImportSpecifier(o,p(o.propertyName,c,e.isIdentifier),p(o.name,c,e.isIdentifier));case 266:return f.updateExportAssignment(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),p(o.expression,c,e.isExpression));case 267:return f.updateExportDeclaration(o,u(o.decorators,c,e.isDecorator),u(o.modifiers,c,e.isModifier),o.isTypeOnly,p(o.exportClause,c,e.isNamedExportBindings),p(o.moduleSpecifier,c,e.isExpression));case 268:return f.updateNamedExports(o,u(o.elements,c,e.isExportSpecifier));case 270:return f.updateExportSpecifier(o,p(o.propertyName,c,e.isIdentifier),p(o.name,c,e.isIdentifier));case 272:return f.updateExternalModuleReference(o,p(o.expression,c,e.isExpression));case 273:return f.updateJsxElement(o,p(o.openingElement,c,e.isJsxOpeningElement),u(o.children,c,e.isJsxChild),p(o.closingElement,c,e.isJsxClosingElement));case 274:return f.updateJsxSelfClosingElement(o,p(o.tagName,c,e.isJsxTagNameExpression),u(o.typeArguments,c,e.isTypeNode),p(o.attributes,c,e.isJsxAttributes));case 275:return f.updateJsxOpeningElement(o,p(o.tagName,c,e.isJsxTagNameExpression),u(o.typeArguments,c,e.isTypeNode),p(o.attributes,c,e.isJsxAttributes));case 276:return f.updateJsxClosingElement(o,p(o.tagName,c,e.isJsxTagNameExpression));case 277:return f.updateJsxFragment(o,p(o.openingFragment,c,e.isJsxOpeningFragment),u(o.children,c,e.isJsxChild),p(o.closingFragment,c,e.isJsxClosingFragment));case 280:return f.updateJsxAttribute(o,p(o.name,c,e.isIdentifier),p(o.initializer,c,e.isStringLiteralOrJsxExpression));case 281:return f.updateJsxAttributes(o,u(o.properties,c,e.isJsxAttributeLike));case 282:return f.updateJsxSpreadAttribute(o,p(o.expression,c,e.isExpression));case 283:return f.updateJsxExpression(o,p(o.expression,c,e.isExpression));case 284:return f.updateCaseClause(o,p(o.expression,c,e.isExpression),u(o.statements,c,e.isStatement));case 285:return f.updateDefaultClause(o,u(o.statements,c,e.isStatement));case 286:return f.updateHeritageClause(o,u(o.types,c,e.isExpressionWithTypeArguments));case 287:return f.updateCatchClause(o,p(o.variableDeclaration,c,e.isVariableDeclaration),p(o.block,c,e.isBlock));case 288:return f.updatePropertyAssignment(o,p(o.name,c,e.isPropertyName),p(o.initializer,c,e.isExpression));case 289:return f.updateShorthandPropertyAssignment(o,p(o.name,c,e.isIdentifier),p(o.objectAssignmentInitializer,c,e.isExpression));case 290:return f.updateSpreadAssignment(o,p(o.expression,c,e.isExpression));case 291:return f.updateEnumMember(o,p(o.name,c,e.isPropertyName),p(o.initializer,c,e.isExpression));case 297:return f.updateSourceFile(o,a(o.statements,c,l));case 336:return f.updatePartiallyEmittedExpression(o,p(o.expression,c,e.isExpression));case 337:return f.updateCommaListExpression(o,u(o.elements,c,e.isExpression));default:return o}}}}(ts||(ts={})),function(e){e.createSourceMapGenerator=function(n,t,r,a,o){var c,l,u=o.extendedDiagnostics?e.performance.createTimer("Source Map","beforeSourcemap","afterSourcemap"):e.performance.nullTimer,d=u.enter,p=u.exit,m=[],f=[],_=new e.Map,g=[],y="",v=0,h=0,b=0,E=0,T=0,S=0,x=!1,L=0,A=0,D=0,C=0,I=0,N=0,R=!1,k=!1,O=!1;return{getSources:function(){return m},addSource:P,setSourceContent:M,addName:w,addMapping:F,appendSourceMap:function(n,t,r,a,o,s){e.Debug.assert(n>=L,"generatedLine cannot backtrack"),e.Debug.assert(t>=0,"generatedCharacter cannot be negative"),d();for(var c,l=[],u=i(r.mappings),m=u.next();!m.done;m=u.next()){var f=m.value;if(s&&(f.generatedLine>s.line||f.generatedLine===s.line&&f.generatedCharacter>s.character))break;if(!o||!(f.generatedLine<o.line||o.line===f.generatedLine&&f.generatedCharacter<o.character)){var _=void 0,g=void 0,y=void 0,v=void 0;if(void 0!==f.sourceIndex){if(void 0===(_=l[f.sourceIndex])){var h=r.sources[f.sourceIndex],b=r.sourceRoot?e.combinePaths(r.sourceRoot,h):h,E=e.combinePaths(e.getDirectoryPath(a),b);l[f.sourceIndex]=_=P(E),r.sourcesContent&&"string"==typeof r.sourcesContent[f.sourceIndex]&&M(_,r.sourcesContent[f.sourceIndex])}g=f.sourceLine,y=f.sourceCharacter,r.names&&void 0!==f.nameIndex&&(c||(c=[]),void 0===(v=c[f.nameIndex])&&(c[f.nameIndex]=v=w(r.names[f.nameIndex])))}var T=f.generatedLine-(o?o.line:0),S=T+n,x=o&&o.line===f.generatedLine?f.generatedCharacter-o.character:f.generatedCharacter;F(S,0===T?x+t:x,_,g,y,v)}}p()},toJSON:B,toString:function(){return JSON.stringify(B())}};function P(t){d();var r=e.getRelativePathToDirectoryOrUrl(a,t,n.getCurrentDirectory(),n.getCanonicalFileName,!0),i=_.get(r);return void 0===i&&(i=f.length,f.push(r),m.push(t),_.set(r,i)),p(),i}function M(e,n){if(d(),null!==n){for(c||(c=[]);c.length<e;)c.push(null);c[e]=n}p()}function w(n){d(),l||(l=new e.Map);var t=l.get(n);return void 0===t&&(t=g.length,g.push(n),l.set(n,t)),p(),t}function F(n,t,r,a,i,o){e.Debug.assert(n>=L,"generatedLine cannot backtrack"),e.Debug.assert(t>=0,"generatedCharacter cannot be negative"),e.Debug.assert(void 0===r||r>=0,"sourceIndex cannot be negative"),e.Debug.assert(void 0===a||a>=0,"sourceLine cannot be negative"),e.Debug.assert(void 0===i||i>=0,"sourceCharacter cannot be negative"),d(),(function(e,n){return!R||L!==e||A!==n}(n,t)||function(e,n,t){return void 0!==e&&void 0!==n&&void 0!==t&&D===e&&(C>n||C===n&&I>t)}(r,a,i))&&(G(),L=n,A=t,k=!1,O=!1,R=!0),void 0!==r&&void 0!==a&&void 0!==i&&(D=r,C=a,I=i,k=!0,void 0!==o&&(N=o,O=!0)),p()}function G(){if(R&&(!x||v!==L||h!==A||b!==D||E!==C||T!==I||S!==N)){if(d(),v<L)do{y+=";",v++,h=0}while(v<L);else e.Debug.assertEqual(v,L,"generatedLine cannot backtrack"),x&&(y+=",");y+=s(A-h),h=A,k&&(y+=s(D-b),b=D,y+=s(C-E),E=C,y+=s(I-T),T=I,O&&(y+=s(N-S),S=N)),x=!0,p()}}function B(){return G(),{version:3,file:t,sourceRoot:r,sources:f,names:g,mappings:y,sourcesContent:c}}};var n=/^\/\/[@#] source[M]appingURL=(.+)\s*$/,t=/^\s*(\/\/[@#] .*)?$/;function r(e){return"string"==typeof e||null===e}function a(n){return null!==n&&"object"==typeof n&&3===n.version&&"string"==typeof n.file&&"string"==typeof n.mappings&&e.isArray(n.sources)&&e.every(n.sources,e.isString)&&(void 0===n.sourceRoot||null===n.sourceRoot||"string"==typeof n.sourceRoot)&&(void 0===n.sourcesContent||null===n.sourcesContent||e.isArray(n.sourcesContent)&&e.every(n.sourcesContent,r))&&(void 0===n.names||null===n.names||e.isArray(n.names)&&e.every(n.names,e.isString))}function i(e){var n,t=!1,r=0,a=0,i=0,o=0,s=0,c=0,l=0;return{get pos(){return r},get error(){return n},get state(){return u(!0,!0)},next:function(){for(;!t&&r<e.length;){var n=e.charCodeAt(r);if(59!==n){if(44!==n){var p=!1,y=!1;if(i+=g(),f())return d();if(i<0)return m("Invalid generatedCharacter found");if(!_()){if(p=!0,o+=g(),f())return d();if(o<0)return m("Invalid sourceIndex found");if(_())return m("Unsupported Format: No entries after sourceIndex");if(s+=g(),f())return d();if(s<0)return m("Invalid sourceLine found");if(_())return m("Unsupported Format: No entries after sourceLine");if(c+=g(),f())return d();if(c<0)return m("Invalid sourceCharacter found");if(!_()){if(y=!0,l+=g(),f())return d();if(l<0)return m("Invalid nameIndex found");if(!_())return m("Unsupported Error Format: Entries after nameIndex")}}return{value:u(p,y),done:t}}r++}else a++,i=0,r++}return d()}};function u(e,n){return{generatedLine:a,generatedCharacter:i,sourceIndex:e?o:void 0,sourceLine:e?s:void 0,sourceCharacter:e?c:void 0,nameIndex:n?l:void 0}}function d(){return t=!0,{value:void 0,done:!0}}function p(e){void 0===n&&(n=e)}function m(e){return p(e),d()}function f(){return void 0!==n}function _(){return r===e.length||44===e.charCodeAt(r)||59===e.charCodeAt(r)}function g(){for(var n,t=!0,a=0,i=0;t;r++){if(r>=e.length)return p("Error in decoding base64VLQFormatDecode, past the mapping string"),-1;var o=(n=e.charCodeAt(r))>=65&&n<=90?n-65:n>=97&&n<=122?n-97+26:n>=48&&n<=57?n-48+52:43===n?62:47===n?63:-1;if(-1===o)return p("Invalid character in VLQ"),-1;t=0!=(32&o),i|=(31&o)<<a,a+=5}return 0==(1&i)?i>>=1:i=-(i>>=1),i}}function o(e){return void 0!==e.sourceIndex&&void 0!==e.sourceLine&&void 0!==e.sourceCharacter}function s(n){n<0?n=1+(-n<<1):n<<=1;var t,r="";do{var a=31&n;(n>>=5)>0&&(a|=32),r+=String.fromCharCode((t=a)>=0&&t<26?65+t:t>=26&&t<52?97+t-26:t>=52&&t<62?48+t-52:62===t?43:63===t?47:e.Debug.fail(t+": not a base64 value"))}while(n>0);return r}function c(e){return void 0!==e.sourceIndex&&void 0!==e.sourcePosition}function l(e,n){return e.generatedPosition===n.generatedPosition&&e.sourceIndex===n.sourceIndex&&e.sourcePosition===n.sourcePosition}function u(n,t){return e.Debug.assert(n.sourceIndex===t.sourceIndex),e.compareValues(n.sourcePosition,t.sourcePosition)}function d(n,t){return e.compareValues(n.generatedPosition,t.generatedPosition)}function p(e){return e.sourcePosition}function m(e){return e.generatedPosition}e.getLineInfo=function(e,n){return{getLineCount:function(){return n.length},getLineText:function(t){return e.substring(n[t],n[t+1])}}},e.tryGetSourceMappingURL=function(e){for(var r=e.getLineCount()-1;r>=0;r--){var a=e.getLineText(r),i=n.exec(a);if(i)return i[1];if(!a.match(t))break}},e.isRawSourceMap=a,e.tryParseRawSourceMap=function(e){try{var n=JSON.parse(e);if(a(n))return n}catch(e){}},e.decodeMappings=i,e.sameMapping=function(e,n){return e===n||e.generatedLine===n.generatedLine&&e.generatedCharacter===n.generatedCharacter&&e.sourceIndex===n.sourceIndex&&e.sourceLine===n.sourceLine&&e.sourceCharacter===n.sourceCharacter&&e.nameIndex===n.nameIndex},e.isSourceMapping=o,e.createDocumentPositionMapper=function(n,t,r){var a,s,f,_=e.getDirectoryPath(r),g=t.sourceRoot?e.getNormalizedAbsolutePath(t.sourceRoot,_):_,y=e.getNormalizedAbsolutePath(t.file,_),v=n.getSourceFileLike(y),h=t.sources.map((function(n){return e.getNormalizedAbsolutePath(n,g)})),b=new e.Map(h.map((function(e,t){return[n.getCanonicalFileName(e),t]})));return{getSourcePosition:function(n){var t=x();if(!e.some(t))return n;var r=e.binarySearchKey(t,n.pos,m,e.compareValues);r<0&&(r=~r);var a=t[r];if(void 0===a||!c(a))return n;return{fileName:h[a.sourceIndex],pos:a.sourcePosition}},getGeneratedPosition:function(t){var r=b.get(n.getCanonicalFileName(t.fileName));if(void 0===r)return t;var a=S(r);if(!e.some(a))return t;var i=e.binarySearchKey(a,t.pos,p,e.compareValues);i<0&&(i=~i);var o=a[i];if(void 0===o||o.sourceIndex!==r)return t;return{fileName:y,pos:o.generatedPosition}}};function E(r){var a,i,s=void 0!==v?e.getPositionOfLineAndCharacter(v,r.generatedLine,r.generatedCharacter,!0):-1;if(o(r)){var c=n.getSourceFileLike(h[r.sourceIndex]);a=t.sources[r.sourceIndex],i=void 0!==c?e.getPositionOfLineAndCharacter(c,r.sourceLine,r.sourceCharacter,!0):-1}return{generatedPosition:s,source:a,sourceIndex:r.sourceIndex,sourcePosition:i,nameIndex:r.nameIndex}}function T(){if(void 0===a){var r=i(t.mappings),o=e.arrayFrom(r,E);void 0!==r.error?(n.log&&n.log("Encountered error while decoding sourcemap: "+r.error),a=e.emptyArray):a=o}return a}function S(n){if(void 0===f){for(var t=[],r=0,a=T();r<a.length;r++){var i=a[r];if(c(i)){var o=t[i.sourceIndex];o||(t[i.sourceIndex]=o=[]),o.push(i)}}f=t.map((function(n){return e.sortAndDeduplicate(n,u,l)}))}return f[n]}function x(){if(void 0===s){for(var n=[],t=0,r=T();t<r.length;t++){var a=r[t];n.push(a)}s=e.sortAndDeduplicate(n,d,l)}return s}},e.identitySourceMapConsumer={getSourcePosition:e.identity,getGeneratedPosition:e.identity}}(ts||(ts={})),function(e){function n(n){return(n=e.getOriginalNode(n))?e.getNodeId(n):0}function t(e){return void 0!==e.propertyName&&"default"===e.propertyName.escapedText}function r(n){if(e.getNamespaceDeclarationNode(n))return!0;var r=n.importClause&&n.importClause.namedBindings;if(!r)return!1;if(!e.isNamedImports(r))return!1;for(var a=0,i=0,o=r.elements;i<o.length;i++){t(o[i])&&a++}return a>0&&a!==r.elements.length||!!(r.elements.length-a)&&e.isDefaultImport(n)}function a(n){return!r(n)&&(e.isDefaultImport(n)||!!n.importClause&&e.isNamedImports(n.importClause.namedBindings)&&function(n){return!!n&&!!e.isNamedImports(n)&&e.some(n.elements,t)}(n.importClause.namedBindings))}function i(n,t,r){if(e.isBindingPattern(n.name))for(var a=0,o=n.name.elements;a<o.length;a++){var s=o[a];e.isOmittedExpression(s)||(r=i(s,t,r))}else if(!e.isGeneratedIdentifier(n.name)){var c=e.idText(n.name);t.get(c)||(t.set(c,!0),r=e.append(r,n.name))}return r}function o(e,n,t){var r=e[n];return r?r.push(t):e[n]=r=[t],r}function s(n){return e.isStringLiteralLike(n)||8===n.kind||e.isKeyword(n.kind)||e.isIdentifier(n)}e.getOriginalNodeId=n,e.chainBundle=function(n,t){return function(r){return 297===r.kind?t(r):function(r){return n.factory.createBundle(e.map(r.sourceFiles,t),r.prepends)}(r)}},e.getExportNeedsImportStarHelper=function(n){return!!e.getNamespaceDeclarationNode(n)},e.getImportNeedsImportStarHelper=r,e.getImportNeedsImportDefaultHelper=a,e.collectExternalModuleInfo=function(t,s,c,l){for(var u,d,p=[],m=e.createMultiMap(),f=[],_=new e.Map,g=!1,y=!1,v=!1,h=!1,b=0,E=s.statements;b<E.length;b++){var T=E[b];switch(T.kind){case 261:p.push(T),!v&&r(T)&&(v=!0),!h&&a(T)&&(h=!0);break;case 260:272===T.moduleReference.kind&&p.push(T);break;case 267:if(T.moduleSpecifier)if(T.exportClause)if(p.push(T),e.isNamedExports(T.exportClause))C(T);else{var S=T.exportClause.name;_.get(e.idText(S))||(o(f,n(T),S),_.set(e.idText(S),!0),u=e.append(u,S)),v=!0}else p.push(T),y=!0;else C(T);break;case 266:T.isExportEquals&&!d&&(d=T);break;case 232:if(e.hasSyntacticModifier(T,1))for(var x=0,L=T.declarationList.declarations;x<L.length;x++){var A=L[x];u=i(A,_,u)}break;case 251:if(e.hasSyntacticModifier(T,1))if(e.hasSyntacticModifier(T,512))g||(o(f,n(T),t.factory.getDeclarationName(T)),g=!0);else{S=T.name;_.get(e.idText(S))||(o(f,n(T),S),_.set(e.idText(S),!0),u=e.append(u,S))}break;case 252:if(e.hasSyntacticModifier(T,1))if(e.hasSyntacticModifier(T,512))g||(o(f,n(T),t.factory.getDeclarationName(T)),g=!0);else(S=T.name)&&!_.get(e.idText(S))&&(o(f,n(T),S),_.set(e.idText(S),!0),u=e.append(u,S))}}var D=e.createExternalHelpersImportDeclarationIfNeeded(t.factory,t.getEmitHelperFactory(),s,l,y,v,h);return D&&p.unshift(D),{externalImports:p,exportSpecifiers:m,exportEquals:d,hasExportStarsToExportValues:y,exportedBindings:f,exportedNames:u,externalHelpersImportDeclaration:D};function C(t){for(var r=0,a=e.cast(t.exportClause,e.isNamedExports).elements;r<a.length;r++){var i=a[r];if(!_.get(e.idText(i.name))){var s=i.propertyName||i.name;t.moduleSpecifier||m.add(e.idText(s),i);var l=c.getReferencedImportDeclaration(s)||c.getReferencedValueDeclaration(s);l&&o(f,n(l),i.name),_.set(e.idText(i.name),!0),u=e.append(u,i.name)}}}},e.isSimpleCopiableExpression=s,e.isSimpleInlineableExpression=function(n){return!e.isIdentifier(n)&&s(n)||e.isWellKnownSymbolSyntactically(n)},e.isCompoundAssignment=function(e){return e>=63&&e<=77},e.getNonAssignmentOperatorForCompoundAssignment=function(e){switch(e){case 63:return 39;case 64:return 40;case 65:return 41;case 66:return 42;case 67:return 43;case 68:return 44;case 69:return 47;case 70:return 48;case 71:return 49;case 72:return 50;case 73:return 51;case 77:return 52;case 74:return 56;case 75:return 55;case 76:return 60}},e.addPrologueDirectivesAndInitialSuperCall=function(n,t,r,a){if(t.body){var i=t.body.statements,o=n.copyPrologue(i,r,!1,a);if(o===i.length)return o;var s=e.findIndex(i,(function(n){return e.isExpressionStatement(n)&&e.isSuperCall(n.expression)}),o);if(s>-1){for(var c=o;c<=s;c++)r.push(e.visitNode(i[c],a,e.isStatement));return s+1}return o}return 0},e.getProperties=function(n,t,r){return e.filter(n.members,(function(n){return function(n,t,r){return e.isPropertyDeclaration(n)&&(!!n.initializer||!t)&&e.hasStaticModifier(n)===r}(n,t,r)}))},e.isInitializedProperty=function(e){return 163===e.kind&&void 0!==e.initializer}}(ts||(ts={})),function(e){function n(t,r){var a=e.getTargetOfBindingOrAssignmentElement(t);return e.isBindingOrAssignmentPattern(a)?function(t,r){for(var a=e.getElementsOfBindingOrAssignmentPattern(t),i=0,o=a;i<o.length;i++){if(n(o[i],r))return!0}return!1}(a,r):!!e.isIdentifier(a)&&a.escapedText===r}function t(n){var r=e.tryGetPropertyNameOfBindingOrAssignmentElement(n);if(r&&e.isComputedPropertyName(r)&&!e.isLiteralExpression(r.expression))return!0;var a,i=e.getTargetOfBindingOrAssignmentElement(n);return!!i&&e.isBindingOrAssignmentPattern(i)&&(a=i,!!e.forEach(e.getElementsOfBindingOrAssignmentPattern(a),t))}function r(n,t,s,c,l){var u=e.getTargetOfBindingOrAssignmentElement(t);if(!l){var d=e.visitNode(e.getInitializerOfBindingOrAssignmentElement(t),n.visitor,e.isExpression);d?s?(s=function(e,n,t,r){return n=o(e,n,!0,r),e.context.factory.createConditionalExpression(e.context.factory.createTypeCheck(n,"undefined"),void 0,t,void 0,n)}(n,s,d,c),!e.isSimpleInlineableExpression(d)&&e.isBindingOrAssignmentPattern(u)&&(s=o(n,s,!0,c))):s=d:s||(s=n.context.factory.createVoidZero())}e.isObjectBindingOrAssignmentPattern(u)?function(n,t,a,s,c){var l,u,d=e.getElementsOfBindingOrAssignmentPattern(a),p=d.length;if(1!==p){s=o(n,s,!e.isDeclarationBindingElement(t)||0!==p,c)}for(var m=0;m<p;m++){var f=d[m];if(e.getRestIndicatorOfBindingOrAssignmentElement(f)){if(m===p-1){l&&(n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(l),s,c,a),l=void 0);g=n.context.getEmitHelperFactory().createRestHelper(s,d,u,a);r(n,f,g,f)}}else{var _=e.getPropertyNameOfBindingOrAssignmentElement(f);if(!(n.level>=1)||24576&f.transformFlags||24576&e.getTargetOfBindingOrAssignmentElement(f).transformFlags||e.isComputedPropertyName(_)){l&&(n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(l),s,c,a),l=void 0);var g=i(n,s,_);e.isComputedPropertyName(_)&&(u=e.append(u,g.argumentExpression)),r(n,f,g,f)}else l=e.append(l,e.visitNode(f,n.visitor))}}l&&n.emitBindingOrAssignment(n.createObjectBindingOrAssignmentPattern(l),s,c,a)}(n,t,u,s,c):e.isArrayBindingOrAssignmentPattern(u)?function(n,t,i,s,c){var l,u,d=e.getElementsOfBindingOrAssignmentPattern(i),p=d.length;if(n.level<1&&n.downlevelIteration)s=o(n,e.setTextRange(n.context.getEmitHelperFactory().createReadHelper(s,p>0&&e.getRestIndicatorOfBindingOrAssignmentElement(d[p-1])?void 0:p),c),!1,c);else if(1!==p&&(n.level<1||0===p)||e.every(d,e.isOmittedExpression)){s=o(n,s,!e.isDeclarationBindingElement(t)||0!==p,c)}for(var m=0;m<p;m++){var f=d[m];if(n.level>=1)if(16384&f.transformFlags||n.hasTransformedPriorElement&&!a(f)){n.hasTransformedPriorElement=!0;var _=n.context.factory.createTempVariable(void 0);n.hoistTempVariables&&n.context.hoistVariableDeclaration(_),u=e.append(u,[_,f]),l=e.append(l,n.createArrayBindingOrAssignmentElement(_))}else l=e.append(l,f);else{if(e.isOmittedExpression(f))continue;if(e.getRestIndicatorOfBindingOrAssignmentElement(f)){if(m===p-1){g=n.context.factory.createArraySliceCall(s,m);r(n,f,g,f)}}else{var g=n.context.factory.createElementAccessExpression(s,m);r(n,f,g,f)}}}l&&n.emitBindingOrAssignment(n.createArrayBindingOrAssignmentPattern(l),s,c,i);if(u)for(var y=0,v=u;y<v.length;y++){var h=v[y],b=h[0];r(n,f=h[1],b,f)}}(n,t,u,s,c):n.emitBindingOrAssignment(u,s,c,t)}function a(n){var t=e.getTargetOfBindingOrAssignmentElement(n);if(!t||e.isOmittedExpression(t))return!0;var r=e.tryGetPropertyNameOfBindingOrAssignmentElement(n);if(r&&!e.isPropertyNameLiteral(r))return!1;var i=e.getInitializerOfBindingOrAssignmentElement(n);return!(i&&!e.isSimpleInlineableExpression(i))&&(e.isBindingOrAssignmentPattern(t)?e.every(e.getElementsOfBindingOrAssignmentPattern(t),a):e.isIdentifier(t))}function i(n,t,r){if(e.isComputedPropertyName(r)){var a=o(n,e.visitNode(r.expression,n.visitor),!1,r);return n.context.factory.createElementAccessExpression(t,a)}if(e.isStringOrNumericLiteralLike(r)){a=e.factory.cloneNode(r);return n.context.factory.createElementAccessExpression(t,a)}var i=n.context.factory.createIdentifier(e.idText(r));return n.context.factory.createPropertyAccessExpression(t,i)}function o(n,t,r,a){if(e.isIdentifier(t)&&r)return t;var i=n.context.factory.createTempVariable(void 0);return n.hoistTempVariables?(n.context.hoistVariableDeclaration(i),n.emitExpression(e.setTextRange(n.context.factory.createAssignment(i,t),a))):n.emitBindingOrAssignment(i,t,a,void 0),i}function s(e){return e}!function(e){e[e.All=0]="All",e[e.ObjectRest=1]="ObjectRest"}(e.FlattenLevel||(e.FlattenLevel={})),e.flattenDestructuringAssignment=function(a,i,c,l,u,d){var p,m,f=a;if(e.isDestructuringAssignment(a))for(p=a.right;e.isEmptyArrayLiteral(a.left)||e.isEmptyObjectLiteral(a.left);){if(!e.isDestructuringAssignment(p))return e.visitNode(p,i,e.isExpression);f=a=p,p=a.right}var _={context:c,level:l,downlevelIteration:!!c.getCompilerOptions().downlevelIteration,hoistTempVariables:!0,emitExpression:g,emitBindingOrAssignment:function(n,t,r,a){e.Debug.assertNode(n,d?e.isIdentifier:e.isExpression);var o=d?d(n,t,r):e.setTextRange(c.factory.createAssignment(e.visitNode(n,i,e.isExpression),t),r);o.original=a,g(o)},createArrayBindingOrAssignmentPattern:function(n){return function(n,t){return n.createArrayLiteralExpression(e.map(t,n.converters.convertToArrayAssignmentElement))}(c.factory,n)},createObjectBindingOrAssignmentPattern:function(n){return function(n,t){return n.createObjectLiteralExpression(e.map(t,n.converters.convertToObjectAssignmentElement))}(c.factory,n)},createArrayBindingOrAssignmentElement:s,visitor:i};if(p&&(p=e.visitNode(p,i,e.isExpression),e.isIdentifier(p)&&n(a,p.escapedText)||t(a)?p=o(_,p,!1,f):u?p=o(_,p,!0,f):e.nodeIsSynthesized(a)&&(f=p)),r(_,a,p,f,e.isDestructuringAssignment(a)),p&&u){if(!e.some(m))return p;m.push(p)}return c.factory.inlineExpressions(m)||c.factory.createOmittedExpression();function g(n){m=e.append(m,n)}},e.flattenDestructuringBinding=function(a,i,s,c,l,u,d){var p;void 0===u&&(u=!1);var m=[],f=[],_={context:s,level:c,downlevelIteration:!!s.getCompilerOptions().downlevelIteration,hoistTempVariables:u,emitExpression:function(n){p=e.append(p,n)},emitBindingOrAssignment:C,createArrayBindingOrAssignmentPattern:function(n){return function(n,t){return e.Debug.assertEachNode(t,e.isArrayBindingElement),n.createArrayBindingPattern(t)}(s.factory,n)},createObjectBindingOrAssignmentPattern:function(n){return function(n,t){return e.Debug.assertEachNode(t,e.isBindingElement),n.createObjectBindingPattern(t)}(s.factory,n)},createArrayBindingOrAssignmentElement:function(e){return function(e,n){return e.createBindingElement(void 0,void 0,n)}(s.factory,e)},visitor:i};if(e.isVariableDeclaration(a)){var g=e.getInitializerOfBindingOrAssignmentElement(a);g&&(e.isIdentifier(g)&&n(a,g.escapedText)||t(a))&&(g=o(_,g,!1,g),a=s.factory.updateVariableDeclaration(a,a.name,void 0,void 0,g))}if(r(_,a,l,a,d),p){var y=s.factory.createTempVariable(void 0);if(u){var v=s.factory.inlineExpressions(p);p=void 0,C(y,v,void 0,void 0)}else{s.hoistVariableDeclaration(y);var h=e.last(m);h.pendingExpressions=e.append(h.pendingExpressions,s.factory.createAssignment(y,h.value)),e.addRange(h.pendingExpressions,p),h.value=y}}for(var b=0,E=m;b<E.length;b++){var T=E[b],S=T.pendingExpressions,x=T.name,L=(v=T.value,T.location),A=T.original,D=s.factory.createVariableDeclaration(x,void 0,void 0,S?s.factory.inlineExpressions(e.append(S,v)):v);D.original=A,e.setTextRange(D,L),f.push(D)}return f;function C(n,t,r,a){e.Debug.assertNode(n,e.isBindingName),p&&(t=s.factory.inlineExpressions(e.append(p,t)),p=void 0),m.push({pendingExpressions:p,name:n,value:t,location:r,original:a})}}}(ts||(ts={})),function(e){var n;function t(n){return n.templateFlags?e.factory.createVoidZero():e.factory.createStringLiteral(n.text)}function r(n,t){var r=n.rawText;if(void 0===r){r=e.getSourceTextOfNodeFromSourceFile(t,n);var a=14===n.kind||17===n.kind;r=r.substring(1,r.length-(a?1:2))}return r=r.replace(/\r\n?/g,"\n"),e.setTextRange(e.factory.createStringLiteral(r),n)}!function(e){e[e.LiftRestriction=0]="LiftRestriction",e[e.All=1]="All"}(n=e.ProcessLevel||(e.ProcessLevel={})),e.processTaggedTemplateExpression=function(a,i,o,s,c,l){var u=e.visitNode(i.tag,o,e.isExpression),d=[void 0],p=[],m=[],f=i.template;if(l===n.LiftRestriction&&!e.hasInvalidEscape(f))return e.visitEachChild(i,o,a);if(e.isNoSubstitutionTemplateLiteral(f))p.push(t(f)),m.push(r(f,s));else{p.push(t(f.head)),m.push(r(f.head,s));for(var _=0,g=f.templateSpans;_<g.length;_++){var y=g[_];p.push(t(y.literal)),m.push(r(y.literal,s)),d.push(e.visitNode(y.expression,o,e.isExpression))}}var v=a.getEmitHelperFactory().createTemplateObjectHelper(e.factory.createArrayLiteralExpression(p),e.factory.createArrayLiteralExpression(m));if(e.isExternalModule(s)){var h=e.factory.createUniqueName("templateObject");c(h),d[0]=e.factory.createLogicalOr(h,e.factory.createAssignment(h,v))}else d[0]=v;return e.factory.createCallExpression(u,void 0,d)}}(ts||(ts={})),function(e){var n,t;!function(e){e[e.ClassAliases=1]="ClassAliases",e[e.NamespaceExports=2]="NamespaceExports",e[e.NonQualifiedEnumMembers=8]="NonQualifiedEnumMembers"}(n||(n={})),function(e){e[e.None=0]="None",e[e.HasStaticInitializedProperties=1]="HasStaticInitializedProperties",e[e.HasConstructorDecorators=2]="HasConstructorDecorators",e[e.HasMemberDecorators=4]="HasMemberDecorators",e[e.IsExportOfNamespace=8]="IsExportOfNamespace",e[e.IsNamedExternalExport=16]="IsNamedExternalExport",e[e.IsDefaultExternalExport=32]="IsDefaultExternalExport",e[e.IsDerivedClass=64]="IsDerivedClass",e[e.UseImmediatelyInvokedFunctionExpression=128]="UseImmediatelyInvokedFunctionExpression",e[e.HasAnyDecorators=6]="HasAnyDecorators",e[e.NeedsName=5]="NeedsName",e[e.MayNeedImmediatelyInvokedFunctionExpression=7]="MayNeedImmediatelyInvokedFunctionExpression",e[e.IsExported=56]="IsExported"}(t||(t={})),e.transformTypeScript=function(n){var t,r,a,i,o,s,c,l,u,d,p=n.factory,m=n.getEmitHelperFactory,f=n.startLexicalEnvironment,_=n.resumeLexicalEnvironment,g=n.endLexicalEnvironment,y=n.hoistVariableDeclaration,v=n.getEmitResolver(),h=n.getCompilerOptions(),b=e.getStrictOptionValue(h,"strictNullChecks"),E=e.getEmitScriptTarget(h),T=e.getEmitModuleKind(h),S=n.onEmitNode,x=n.onSubstituteNode;return n.onEmitNode=function(n,r,a){var i=d,o=t;e.isSourceFile(r)&&(t=r);2&l&&function(n){return 256===e.getOriginalNode(n).kind}(r)&&(d|=2);8&l&&function(n){return 255===e.getOriginalNode(n).kind}(r)&&(d|=8);S(n,r,a),d=i,t=o},n.onSubstituteNode=function(n,t){if(t=x(n,t),1===n)return function(n){switch(n.kind){case 78:return function(n){return function(n){if(1&l&&33554432&v.getNodeCheckFlags(n)){var t=v.getReferencedValueDeclaration(n);if(t){var r=u[t.id];if(r){var a=p.cloneNode(r);return e.setSourceMapRange(a,n),e.setCommentRange(a,n),a}}}return}(n)||Ue(n)||n}(n);case 201:case 202:return function(e){return Ve(e)}(n)}return n}(t);if(e.isShorthandPropertyAssignment(t))return function(n){if(2&l){var t=n.name,r=Ue(t);if(r){if(n.objectAssignmentInitializer){var a=p.createAssignment(r,n.objectAssignmentInitializer);return e.setTextRange(p.createPropertyAssignment(t,a),n)}return e.setTextRange(p.createPropertyAssignment(t,r),n)}}return n}(t);return t},n.enableSubstitution(201),n.enableSubstitution(202),function(n){if(298===n.kind)return function(n){return p.createBundle(n.sourceFiles.map(L),e.mapDefined(n.prepends,(function(n){return 300===n.kind?e.createUnparsedSourceFile(n,"js"):n})))}(n);return L(n)};function L(r){if(r.isDeclarationFile)return r;t=r;var a=A(r,F);return e.addEmitHelpers(a,n.readEmitHelpers()),t=void 0,a}function A(n,t){var r=i,a=o,l=s,u=c;!function(n){switch(n.kind){case 297:case 258:case 257:case 230:i=n,o=void 0,s=void 0;break;case 252:case 251:if(e.hasSyntacticModifier(n,2))break;n.name?be(n):e.Debug.assert(252===n.kind||e.hasSyntacticModifier(n,512)),e.isClassDeclaration(n)&&(o=n)}}(n);var d=t(n);return i!==r&&(s=l),i=r,o=a,c=u,d}function D(e){return A(e,C)}function C(e){return 1&e.transformFlags?w(e):e}function I(e){return A(e,N)}function N(t){switch(t.kind){case 261:case 260:case 266:case 267:return function(t){if(e.getParseTreeNode(t)!==t)return 1&t.transformFlags?e.visitEachChild(t,D,n):t;switch(t.kind){case 261:return function(n){if(!n.importClause)return n;if(n.importClause.isTypeOnly)return;var t=e.visitNode(n.importClause,Le,e.isImportClause);return t||1===h.importsNotUsedAsValues||2===h.importsNotUsedAsValues?p.updateImportDeclaration(n,void 0,void 0,t,n.moduleSpecifier):void 0}(t);case 260:return Ne(t);case 266:return function(t){return v.isValueAliasDeclaration(t)?e.visitEachChild(t,D,n):void 0}(t);case 267:return function(n){if(n.isTypeOnly)return;if(!n.exportClause||e.isNamespaceExport(n.exportClause))return n;if(!v.isValueAliasDeclaration(n))return;var t=e.visitNode(n.exportClause,Ce,e.isNamedExportBindings);return t?p.updateExportDeclaration(n,void 0,void 0,n.isTypeOnly,t,n.moduleSpecifier):void 0}(t);default:e.Debug.fail("Unhandled ellided statement")}}(t);default:return C(t)}}function R(e){return A(e,k)}function k(n){if(267!==n.kind&&261!==n.kind&&262!==n.kind&&(260!==n.kind||272!==n.moduleReference.kind))return 1&n.transformFlags||e.hasSyntacticModifier(n,1)?w(n):n}function O(e){return A(e,P)}function P(n){switch(n.kind){case 166:return fe(n);case 163:return me(n);case 171:case 167:case 168:case 165:return C(n);case 229:return n;default:return e.Debug.failBadSyntaxKind(n)}}function M(n){if(!(2270&e.modifierToFlag(n.kind)||r&&92===n.kind))return n}function w(o){if(e.isStatement(o)&&e.hasSyntacticModifier(o,2))return p.createNotEmittedStatement(o);switch(o.kind){case 92:case 87:return r?void 0:o;case 122:case 120:case 121:case 125:case 84:case 133:case 142:case 178:case 179:case 180:case 181:case 177:case 172:case 159:case 128:case 152:case 131:case 147:case 144:case 141:case 113:case 148:case 175:case 174:case 176:case 173:case 182:case 183:case 184:case 186:case 187:case 188:case 189:case 190:case 191:case 171:case 161:case 254:return;case 163:return me(o);case 259:return;case 166:return fe(o);case 253:return p.createNotEmittedStatement(o);case 252:return function(a){if(!(U(a)||r&&e.hasSyntacticModifier(a,1)))return e.visitEachChild(a,D,n);var i=e.getProperties(a,!0,!0),o=function(n,t){var r=0;e.some(t)&&(r|=1);var a=e.getEffectiveBaseTypeNode(n);a&&103!==e.skipOuterExpressions(a.expression).kind&&(r|=64);(function(n){if(n.decorators&&n.decorators.length>0)return!0;var t=e.getFirstConstructorWithBody(n);if(t)return e.forEach(t.parameters,G);return!1})(n)&&(r|=2);e.childIsDecorated(n)&&(r|=4);Re(n)?r|=8:!function(n){return ke(n)&&e.hasSyntacticModifier(n,512)}(n)?Oe(n)&&(r|=16):r|=32;E<=1&&7&r&&(r|=128);return r}(a,i);128&o&&n.startLexicalEnvironment();var s=a.name||(5&o?p.getGeneratedNameForNode(a):void 0),c=2&o?function(t,r){var a=e.moveRangePastDecorators(t),i=function(t){if(16777216&v.getNodeCheckFlags(t)){0==(1&l)&&(l|=1,n.enableSubstitution(78),u=[]);var r=p.createUniqueName(t.name&&!e.isGeneratedIdentifier(t.name)?e.idText(t.name):"default");return u[e.getOriginalNodeId(t)]=r,y(r),r}}(t),o=p.getLocalName(t,!1,!0),s=e.visitNodes(t.heritageClauses,D,e.isHeritageClause),c=V(t),d=p.createClassExpression(void 0,void 0,r,void 0,s,c);e.setOriginalNode(d,t),e.setTextRange(d,a);var m=p.createVariableStatement(void 0,p.createVariableDeclarationList([p.createVariableDeclaration(o,void 0,void 0,i?p.createAssignment(i,d):d)],1));return e.setOriginalNode(m,t),e.setTextRange(m,a),e.setCommentRange(m,t),m}(a,s):function(n,t,r){var a=128&r?void 0:e.visitNodes(n.modifiers,M,e.isModifier),i=p.createClassDeclaration(void 0,a,t,void 0,e.visitNodes(n.heritageClauses,D,e.isHeritageClause),V(n)),o=e.getEmitFlags(n);1&r&&(o|=32);return e.setTextRange(i,n),e.setOriginalNode(i,n),e.setEmitFlags(i,o),i}(a,s,o),d=[c];if(q(d,a,!1),q(d,a,!0),function(n,t){var r=function(n){var t=function(n){var t=n.decorators,r=K(e.getFirstConstructorWithBody(n));if(!t&&!r)return;return{decorators:t,parameters:r}}(n),r=z(n,n,t);if(!r)return;var a=u&&u[e.getOriginalNodeId(n)],i=p.getLocalName(n,!1,!0),o=m().createDecorateHelper(r,i),s=p.createAssignment(i,a?p.createAssignment(a,o):o);return e.setEmitFlags(s,1536),e.setSourceMapRange(s,e.moveRangePastDecorators(n)),s}(t);r&&n.push(e.setOriginalNode(p.createExpressionStatement(r),t))}(d,a),128&o){var f=e.createTokenRange(e.skipTrivia(t.text,a.members.end),19),_=p.getInternalName(a),g=p.createPartiallyEmittedExpression(_);e.setTextRangeEnd(g,f.end),e.setEmitFlags(g,1536);var h=p.createReturnStatement(g);e.setTextRangePos(h,f.pos),e.setEmitFlags(h,1920),d.push(h),e.insertStatementsAfterStandardPrologue(d,n.endLexicalEnvironment());var b=p.createImmediatelyInvokedArrowFunction(d);e.setEmitFlags(b,33554432);var T=p.createVariableStatement(void 0,p.createVariableDeclarationList([p.createVariableDeclaration(p.getLocalName(a,!1,!1),void 0,void 0,b)]));e.setOriginalNode(T,a),e.setCommentRange(T,a),e.setSourceMapRange(T,e.moveRangePastDecorators(a)),e.startOnNewLine(T),d=[T]}8&o?Me(d,a):(128&o||2&o)&&(32&o?d.push(p.createExportDefault(p.getLocalName(a,!1,!0))):16&o&&d.push(p.createExternalModuleExport(p.getLocalName(a,!1,!0))));d.length>1&&(d.push(p.createEndOfDeclarationMarker(a)),e.setEmitFlags(c,4194304|e.getEmitFlags(c)));return e.singleOrMany(d)}(o);case 221:return function(t){if(!U(t))return e.visitEachChild(t,D,n);var r=p.createClassExpression(void 0,void 0,t.name,void 0,e.visitNodes(t.heritageClauses,D,e.isHeritageClause),V(t));return e.setOriginalNode(r,t),e.setTextRange(r,t),r}(o);case 286:return function(t){if(116===t.token)return;return e.visitEachChild(t,D,n)}(o);case 223:return function(n){return p.updateExpressionWithTypeArguments(n,e.visitNode(n.expression,D,e.isLeftHandSideExpression),void 0)}(o);case 165:return function(t){if(!pe(t))return;var r=p.updateMethodDeclaration(t,void 0,e.visitNodes(t.modifiers,M,e.isModifier),t.asteriskToken,de(t),void 0,void 0,e.visitParameterList(t.parameters,D,n),void 0,e.visitFunctionBody(t.body,D,n));r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t)));return r}(o);case 167:return function(t){if(!ge(t))return;var r=p.updateGetAccessorDeclaration(t,void 0,e.visitNodes(t.modifiers,M,e.isModifier),de(t),e.visitParameterList(t.parameters,D,n),void 0,e.visitFunctionBody(t.body,D,n)||p.createBlock([]));r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t)));return r}(o);case 168:return function(t){if(!ge(t))return;var r=p.updateSetAccessorDeclaration(t,void 0,e.visitNodes(t.modifiers,M,e.isModifier),de(t),e.visitParameterList(t.parameters,D,n),e.visitFunctionBody(t.body,D,n)||p.createBlock([]));r!==t&&(e.setCommentRange(r,t),e.setSourceMapRange(r,e.moveRangePastDecorators(t)));return r}(o);case 251:return function(t){if(!pe(t))return p.createNotEmittedStatement(t);var r=p.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,M,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,D,n),void 0,e.visitFunctionBody(t.body,D,n)||p.createBlock([]));if(Re(t)){var a=[r];return Me(a,t),a}return r}(o);case 208:return function(t){if(!pe(t))return p.createOmittedExpression();return p.updateFunctionExpression(t,e.visitNodes(t.modifiers,M,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,D,n),void 0,e.visitFunctionBody(t.body,D,n)||p.createBlock([]))}(o);case 209:return function(t){return p.updateArrowFunction(t,e.visitNodes(t.modifiers,M,e.isModifier),void 0,e.visitParameterList(t.parameters,D,n),void 0,t.equalsGreaterThanToken,e.visitFunctionBody(t.body,D,n))}(o);case 160:return function(n){if(e.parameterIsThisKeyword(n))return;var t=p.updateParameterDeclaration(n,void 0,void 0,n.dotDotDotToken,e.visitNode(n.name,D,e.isBindingName),void 0,void 0,e.visitNode(n.initializer,D,e.isExpression));t!==n&&(e.setCommentRange(t,n),e.setTextRange(t,e.moveRangePastModifiers(n)),e.setSourceMapRange(t,e.moveRangePastModifiers(n)),e.setEmitFlags(t.name,32));return t}(o);case 207:return function(r){var a=e.skipOuterExpressions(r.expression,-7);if(e.isAssertionExpression(a)){var i=e.visitNode(r.expression,D,e.isExpression);return e.length(e.getLeadingCommentRangesOfNode(i,t))?p.updateParenthesizedExpression(r,i):p.createPartiallyEmittedExpression(i,r)}return e.visitEachChild(r,D,n)}(o);case 206:case 224:return function(n){var t=e.visitNode(n.expression,D,e.isExpression);return p.createPartiallyEmittedExpression(t,n)}(o);case 203:return function(n){return p.updateCallExpression(n,e.visitNode(n.expression,D,e.isExpression),void 0,e.visitNodes(n.arguments,D,e.isExpression))}(o);case 204:return function(n){return p.updateNewExpression(n,e.visitNode(n.expression,D,e.isExpression),void 0,e.visitNodes(n.arguments,D,e.isExpression))}(o);case 205:return function(n){return p.updateTaggedTemplateExpression(n,e.visitNode(n.tag,D,e.isExpression),void 0,e.visitNode(n.template,D,e.isExpression))}(o);case 225:return function(n){var t=e.visitNode(n.expression,D,e.isLeftHandSideExpression);return p.createPartiallyEmittedExpression(t,n)}(o);case 255:return function(n){if(!function(n){return!e.isEnumConst(n)||h.preserveConstEnums||h.isolatedModules}(n))return p.createNotEmittedStatement(n);var r=[],o=2,s=Te(r,n);s&&(T===e.ModuleKind.System&&i===t||(o|=512));var c=Ge(n),l=Be(n),u=e.hasSyntacticModifier(n,1)?p.getExternalModuleOrNamespaceExportName(a,n,!1,!0):p.getLocalName(n,!1,!0),d=p.createLogicalOr(u,p.createAssignment(u,p.createObjectLiteralExpression()));if(he(n)){var m=p.getLocalName(n,!1,!0);d=p.createAssignment(m,d)}var _=p.createExpressionStatement(p.createCallExpression(p.createFunctionExpression(void 0,void 0,void 0,void 0,[p.createParameterDeclaration(void 0,void 0,void 0,c)],void 0,function(n,t){var r=a;a=t;var i=[];f();var o=e.map(n.members,ve);return e.insertStatementsAfterStandardPrologue(i,g()),e.addRange(i,o),a=r,p.createBlock(e.setTextRange(p.createNodeArray(i),n.members),!0)}(n,l)),void 0,[d]));e.setOriginalNode(_,n),s&&(e.setSyntheticLeadingComments(_,void 0),e.setSyntheticTrailingComments(_,void 0));return e.setTextRange(_,n),e.addEmitFlags(_,o),r.push(_),r.push(p.createEndOfDeclarationMarker(n)),r}(o);case 232:return function(t){if(Re(t)){var r=e.getInitializedVariables(t.declarationList);if(0===r.length)return;return e.setTextRange(p.createExpressionStatement(p.inlineExpressions(e.map(r,ye))),t)}return e.visitEachChild(t,D,n)}(o);case 249:return function(n){return p.updateVariableDeclaration(n,e.visitNode(n.name,D,e.isBindingName),void 0,void 0,e.visitNode(n.initializer,D,e.isExpression))}(o);case 256:return Se(o);case 260:return Ne(o);case 274:return function(n){return p.updateJsxSelfClosingElement(n,e.visitNode(n.tagName,D,e.isJsxTagNameExpression),void 0,e.visitNode(n.attributes,D,e.isJsxAttributes))}(o);case 275:return function(n){return p.updateJsxOpeningElement(n,e.visitNode(n.tagName,D,e.isJsxTagNameExpression),void 0,e.visitNode(n.attributes,D,e.isJsxAttributes))}(o);default:return e.visitEachChild(o,D,n)}}function F(t){var r=e.getStrictOptionValue(h,"alwaysStrict")&&!(e.isExternalModule(t)&&T>=e.ModuleKind.ES2015)&&!e.isJsonSourceFile(t);return p.updateSourceFile(t,e.visitLexicalEnvironment(t.statements,I,n,0,r))}function G(e){return void 0!==e.decorators&&e.decorators.length>0}function B(e){return!!(2048&e.transformFlags)}function U(n){return e.some(n.decorators)||e.some(n.typeParameters)||e.some(n.heritageClauses,B)||e.some(n.members,B)}function V(n){var t=[],r=e.getFirstConstructorWithBody(n),a=r&&e.filter(r.parameters,(function(n){return e.isParameterPropertyDeclaration(n,r)}));if(a)for(var i=0,o=a;i<o.length;i++){var s=o[i];e.isIdentifier(s.name)&&t.push(e.setOriginalNode(p.createPropertyDeclaration(void 0,void 0,s.name,void 0,void 0,void 0),s))}return e.addRange(t,e.visitNodes(n.members,O,e.isClassElement)),e.setTextRange(p.createNodeArray(t),n.members)}function j(n,t){return e.filter(n.members,t?function(e){return H(e,!0,n)}:function(e){return H(e,!1,n)})}function H(n,t,r){return e.nodeOrChildIsDecorated(n,r)&&t===e.hasSyntacticModifier(n,32)}function K(n){var t;if(n)for(var r=n.parameters,a=r.length>0&&e.parameterIsThisKeyword(r[0]),i=a?1:0,o=a?r.length-1:r.length,s=0;s<o;s++){var c=r[s+i];(t||c.decorators)&&(t||(t=new Array(o)),t[s]=c.decorators)}return t}function W(n,t){switch(t.kind){case 167:case 168:return function(n,t){if(!t.body)return;var r=e.getAllAccessorDeclarations(n.members,t),a=r.firstAccessor,i=r.secondAccessor,o=r.setAccessor,s=a.decorators?a:i&&i.decorators?i:void 0;if(!s||t!==s)return;var c=s.decorators,l=K(o);if(!c&&!l)return;return{decorators:c,parameters:l}}(n,t);case 165:return function(e){if(!e.body)return;var n=e.decorators,t=K(e);if(!n&&!t)return;return{decorators:n,parameters:t}}(t);case 163:return function(e){var n=e.decorators;if(!n)return;return{decorators:n}}(t);default:return}}function z(n,t,r){if(r){var a=[];return e.addRange(a,e.map(r.decorators,X)),e.addRange(a,e.flatMap(r.parameters,Y)),function(e,n,t){(function(e,n,t){h.emitDecoratorMetadata&&(Q(e)&&t.push(m().createMetadataHelper("design:type",ee(e))),$(e)&&t.push(m().createMetadataHelper("design:paramtypes",ne(e,n))),Z(e)&&t.push(m().createMetadataHelper("design:returntype",te(e))))})(e,n,t)}(n,t,a),a}}function q(n,t,r){e.addRange(n,e.map(function(e,n){for(var t,r=j(e,n),a=0,i=r;a<i.length;a++){var o=J(e,i[a]);o&&(t?t.push(o):t=[o])}return t}(t,r),Pe))}function J(n,t){var r=z(t,n,W(n,t));if(r){var a=function(n,t){return e.hasSyntacticModifier(t,32)?p.getDeclarationName(n):function(e){return p.createPropertyAccessExpression(p.getDeclarationName(e),"prototype")}(n)}(n,t),i=ue(t,!0),o=E>0?163===t.kind?p.createVoidZero():p.createNull():void 0,s=m().createDecorateHelper(r,a,i,o);return e.setTextRange(s,e.moveRangePastDecorators(t)),e.setEmitFlags(s,1536),s}}function X(n){return e.visitNode(n.expression,D,e.isExpression)}function Y(n,t){var r;if(n){r=[];for(var a=0,i=n;a<i.length;a++){var o=i[a],s=m().createParamHelper(X(o),t);e.setTextRange(s,o.expression),e.setEmitFlags(s,1536),r.push(s)}}return r}function Q(e){var n=e.kind;return 165===n||167===n||168===n||163===n}function Z(e){return 165===e.kind}function $(n){switch(n.kind){case 252:case 221:return void 0!==e.getFirstConstructorWithBody(n);case 165:case 167:case 168:return!0}return!1}function ee(n){switch(n.kind){case 163:case 160:return re(n.type);case 168:case 167:return re(function(n){var t=v.getAllAccessorDeclarations(n);return t.setAccessor&&e.getSetAccessorTypeAnnotationNode(t.setAccessor)||t.getAccessor&&e.getEffectiveReturnTypeNode(t.getAccessor)}(n));case 252:case 221:case 165:return p.createIdentifier("Function");default:return p.createVoidZero()}}function ne(n,t){var r=e.isClassLike(n)?e.getFirstConstructorWithBody(n):e.isFunctionLike(n)&&e.nodeIsPresent(n.body)?n:void 0,a=[];if(r)for(var i=function(n,t){if(t&&167===n.kind){var r=e.getAllAccessorDeclarations(t.members,n).setAccessor;if(r)return r.parameters}return n.parameters}(r,t),o=i.length,s=0;s<o;s++){var c=i[s];0===s&&e.isIdentifier(c.name)&&"this"===c.name.escapedText||(c.dotDotDotToken?a.push(re(e.getRestParameterElementType(c.type))):a.push(ee(c)))}return p.createArrayLiteralExpression(a)}function te(n){return e.isFunctionLike(n)&&n.type?re(n.type):e.isAsyncFunction(n)?p.createIdentifier("Promise"):p.createVoidZero()}function re(n){if(void 0===n)return p.createIdentifier("Object");switch(n.kind){case 113:case 150:case 141:return p.createVoidZero();case 186:return re(n.type);case 174:case 175:return p.createIdentifier("Function");case 178:case 179:return p.createIdentifier("Array");case 172:case 131:return p.createIdentifier("Boolean");case 147:return p.createIdentifier("String");case 145:return p.createIdentifier("Object");case 191:switch(n.literal.kind){case 10:case 14:return p.createIdentifier("String");case 214:case 8:return p.createIdentifier("Number");case 9:return le();case 109:case 94:return p.createIdentifier("Boolean");case 103:return p.createVoidZero();default:return e.Debug.failBadSyntaxKind(n.literal)}case 144:return p.createIdentifier("Number");case 155:return le();case 148:return E<2?ce():p.createIdentifier("Symbol");case 173:return function(n){var t=v.getTypeReferenceSerializationKind(n.typeName,o||i);switch(t){case e.TypeReferenceSerializationKind.Unknown:if(e.findAncestor(n,(function(n){return n.parent&&e.isConditionalTypeNode(n.parent)&&(n.parent.trueType===n||n.parent.falseType===n)})))return p.createIdentifier("Object");var r=oe(n.typeName),a=p.createTempVariable(y);return p.createConditionalExpression(p.createTypeCheck(p.createAssignment(a,r),"function"),void 0,a,void 0,p.createIdentifier("Object"));case e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue:return se(n.typeName);case e.TypeReferenceSerializationKind.VoidNullableOrNeverType:return p.createVoidZero();case e.TypeReferenceSerializationKind.BigIntLikeType:return le();case e.TypeReferenceSerializationKind.BooleanType:return p.createIdentifier("Boolean");case e.TypeReferenceSerializationKind.NumberLikeType:return p.createIdentifier("Number");case e.TypeReferenceSerializationKind.StringLikeType:return p.createIdentifier("String");case e.TypeReferenceSerializationKind.ArrayLikeType:return p.createIdentifier("Array");case e.TypeReferenceSerializationKind.ESSymbolType:return E<2?ce():p.createIdentifier("Symbol");case e.TypeReferenceSerializationKind.TypeWithCallSignature:return p.createIdentifier("Function");case e.TypeReferenceSerializationKind.Promise:return p.createIdentifier("Promise");case e.TypeReferenceSerializationKind.ObjectType:return p.createIdentifier("Object");default:return e.Debug.assertNever(t)}}(n);case 183:case 182:return ae(n.types);case 184:return ae([n.trueType,n.falseType]);case 188:if(142===n.operator)return re(n.type);break;case 176:case 189:case 190:case 177:case 128:case 152:case 187:case 195:break;case 303:case 304:case 308:case 309:case 310:break;case 305:case 306:case 307:return re(n.type);default:return e.Debug.failBadSyntaxKind(n)}return p.createIdentifier("Object")}function ae(n){for(var t,r=0,a=n;r<a.length;r++){for(var i=a[r];186===i.kind;)i=i.type;if(141!==i.kind&&(b||(191!==i.kind||103!==i.literal.kind)&&150!==i.kind)){var o=re(i);if(e.isIdentifier(o)&&"Object"===o.escapedText)return o;if(t){if(!e.isIdentifier(t)||!e.isIdentifier(o)||t.escapedText!==o.escapedText)return p.createIdentifier("Object")}else t=o}}return t||p.createVoidZero()}function ie(e,n){return p.createLogicalAnd(p.createStrictInequality(p.createTypeOfExpression(e),p.createStringLiteral("undefined")),n)}function oe(e){if(78===e.kind){var n=se(e);return ie(n,n)}if(78===e.left.kind)return ie(se(e.left),se(e));var t=oe(e.left),r=p.createTempVariable(y);return p.createLogicalAnd(p.createLogicalAnd(t.left,p.createStrictInequality(p.createAssignment(r,t.right),p.createVoidZero())),p.createPropertyAccessExpression(r,e.right))}function se(n){switch(n.kind){case 78:var t=e.setParent(e.setTextRange(e.parseNodeFactory.cloneNode(n),n),n.parent);return t.original=void 0,e.setParent(t,e.getParseTreeNode(i)),t;case 157:return function(e){return p.createPropertyAccessExpression(se(e.left),e.right)}(n)}}function ce(){return p.createConditionalExpression(p.createTypeCheck(p.createIdentifier("Symbol"),"function"),void 0,p.createIdentifier("Symbol"),void 0,p.createIdentifier("Object"))}function le(){return E<99?p.createConditionalExpression(p.createTypeCheck(p.createIdentifier("BigInt"),"function"),void 0,p.createIdentifier("BigInt"),void 0,p.createIdentifier("Object")):p.createIdentifier("BigInt")}function ue(n,t){var r=n.name;return e.isPrivateIdentifier(r)?p.createIdentifier(""):e.isComputedPropertyName(r)?t&&!e.isSimpleInlineableExpression(r.expression)?p.getGeneratedNameForNode(r):r.expression:e.isIdentifier(r)?p.createStringLiteral(e.idText(r)):p.cloneNode(r)}function de(n){var t=n.name;if(e.isComputedPropertyName(t)&&(!e.hasStaticModifier(n)&&c||e.some(n.decorators))){var r=e.visitNode(t.expression,D,e.isExpression),a=e.skipPartiallyEmittedExpressions(r);if(!e.isSimpleInlineableExpression(a)){var i=p.getGeneratedNameForNode(t);return y(i),p.updateComputedPropertyName(t,p.createAssignment(i,r))}}return e.visitNode(t,D,e.isPropertyName)}function pe(n){return!e.nodeIsMissing(n.body)}function me(n){if(!(8388608&n.flags)){var t=p.updatePropertyDeclaration(n,void 0,e.visitNodes(n.modifiers,D,e.isModifier),de(n),void 0,void 0,e.visitNode(n.initializer,D));return t!==n&&(e.setCommentRange(t,n),e.setSourceMapRange(t,e.moveRangePastDecorators(n))),t}}function fe(t){if(pe(t))return p.updateConstructorDeclaration(t,void 0,void 0,e.visitParameterList(t.parameters,D,n),function(t,r){var a=r&&e.filter(r.parameters,(function(n){return e.isParameterPropertyDeclaration(n,r)}));if(!e.some(a))return e.visitFunctionBody(t,D,n);var i=[],o=0;_(),o=e.addPrologueDirectivesAndInitialSuperCall(p,r,i,D),e.addRange(i,e.map(a,_e)),e.addRange(i,e.visitNodes(t.statements,D,e.isStatement,o)),i=p.mergeLexicalEnvironment(i,g());var s=p.createBlock(e.setTextRange(p.createNodeArray(i),t.statements),!0);return e.setTextRange(s,t),e.setOriginalNode(s,t),s}(t.body,t))}function _e(n){var t=n.name;if(e.isIdentifier(t)){var r=e.setParent(e.setTextRange(p.cloneNode(t),t),t.parent);e.setEmitFlags(r,1584);var a=e.setParent(e.setTextRange(p.cloneNode(t),t),t.parent);return e.setEmitFlags(a,1536),e.startOnNewLine(e.removeAllComments(e.setTextRange(e.setOriginalNode(p.createExpressionStatement(p.createAssignment(e.setTextRange(p.createPropertyAccessExpression(p.createThis(),r),n.name),a)),n),e.moveRangePos(n,-1))))}}function ge(n){return!(e.nodeIsMissing(n.body)&&e.hasSyntacticModifier(n,128))}function ye(t){var r=t.name;return e.isBindingPattern(r)?e.flattenDestructuringAssignment(t,D,n,0,!1,we):e.setTextRange(p.createAssignment(Fe(r),e.visitNode(t.initializer,D,e.isExpression)),t)}function ve(t){var r=ue(t,!1),i=function(t){var r=v.getConstantValue(t);return void 0!==r?"string"==typeof r?p.createStringLiteral(r):p.createNumericLiteral(r):(0==(8&l)&&(l|=8,n.enableSubstitution(78)),t.initializer?e.visitNode(t.initializer,D,e.isExpression):p.createVoidZero())}(t),o=p.createAssignment(p.createElementAccessExpression(a,r),i),s=10===i.kind?o:p.createAssignment(p.createElementAccessExpression(a,o),r);return e.setTextRange(p.createExpressionStatement(e.setTextRange(s,t)),t)}function he(n){return Re(n)||ke(n)&&T!==e.ModuleKind.ES2015&&T!==e.ModuleKind.ES2020&&T!==e.ModuleKind.ESNext&&T!==e.ModuleKind.System}function be(n){s||(s=new e.Map);var t=Ee(n);s.has(t)||s.set(t,n)}function Ee(n){return e.Debug.assertNode(n.name,e.isIdentifier),n.name.escapedText}function Te(n,t){var r=p.createVariableStatement(e.visitNodes(t.modifiers,M,e.isModifier),p.createVariableDeclarationList([p.createVariableDeclaration(p.getLocalName(t,!1,!0))],297===i.kind?0:1));if(e.setOriginalNode(r,t),be(t),function(e){if(s){var n=Ee(e);return s.get(n)===e}return!0}(t))return 255===t.kind?e.setSourceMapRange(r.declarationList,t):e.setSourceMapRange(r,t),e.setCommentRange(r,t),e.addEmitFlags(r,4195328),n.push(r),!0;var a=p.createMergeDeclarationMarker(r);return e.setEmitFlags(a,4195840),n.push(a),!1}function Se(o){if(!function(n){var t=e.getParseTreeNode(n,e.isModuleDeclaration);return!t||e.isInstantiatedModule(t,!!h.preserveConstEnums||!!h.isolatedModules)}(o))return p.createNotEmittedStatement(o);e.Debug.assertNode(o.name,e.isIdentifier,"A TypeScript namespace should have an Identifier name."),0==(2&l)&&(l|=2,n.enableSubstitution(78),n.enableSubstitution(289),n.enableEmitNotification(256));var c=[],u=2,d=Te(c,o);d&&(T===e.ModuleKind.System&&i===t||(u|=512));var m=Ge(o),_=Be(o),y=e.hasSyntacticModifier(o,1)?p.getExternalModuleOrNamespaceExportName(a,o,!1,!0):p.getLocalName(o,!1,!0),v=p.createLogicalOr(y,p.createAssignment(y,p.createObjectLiteralExpression()));if(he(o)){var b=p.getLocalName(o,!1,!0);v=p.createAssignment(b,v)}var E=p.createExpressionStatement(p.createCallExpression(p.createFunctionExpression(void 0,void 0,void 0,void 0,[p.createParameterDeclaration(void 0,void 0,void 0,m)],void 0,function(n,t){var i=a,o=r,c=s;a=t,r=n,s=void 0;var l,u,d=[];if(f(),n.body)if(257===n.body.kind)A(n.body,(function(n){return e.addRange(d,e.visitNodes(n.statements,R,e.isStatement))})),l=n.body.statements,u=n.body;else{var m=Se(n.body);m&&(e.isArray(m)?e.addRange(d,m):d.push(m));var _=xe(n).body;l=e.moveRangePos(_.statements,-1)}e.insertStatementsAfterStandardPrologue(d,g()),a=i,r=o,s=c;var y=p.createBlock(e.setTextRange(p.createNodeArray(d),l),!0);e.setTextRange(y,u),n.body&&257===n.body.kind||e.setEmitFlags(y,1536|e.getEmitFlags(y));return y}(o,_)),void 0,[v]));return e.setOriginalNode(E,o),d&&(e.setSyntheticLeadingComments(E,void 0),e.setSyntheticTrailingComments(E,void 0)),e.setTextRange(E,o),e.addEmitFlags(E,u),c.push(E),c.push(p.createEndOfDeclarationMarker(o)),c}function xe(e){if(256===e.body.kind)return xe(e.body)||e.body}function Le(n){if(!n.isTypeOnly){var t=v.isReferencedAliasDeclaration(n)?n.name:void 0,r=e.visitNode(n.namedBindings,Ae,e.isNamedImportBindings);return t||r?p.updateImportClause(n,!1,t,r):void 0}}function Ae(n){if(263===n.kind)return v.isReferencedAliasDeclaration(n)?n:void 0;var t=e.visitNodes(n.elements,De,e.isImportSpecifier);return e.some(t)?p.updateNamedImports(n,t):void 0}function De(e){return v.isReferencedAliasDeclaration(e)?e:void 0}function Ce(n){return e.isNamespaceExport(n)?function(n){return p.updateNamespaceExport(n,e.visitNode(n.name,D,e.isIdentifier))}(n):function(n){var t=e.visitNodes(n.elements,Ie,e.isExportSpecifier);return e.some(t)?p.updateNamedExports(n,t):void 0}(n)}function Ie(e){return v.isValueAliasDeclaration(e)?e:void 0}function Ne(r){if(e.isExternalModuleImportEqualsDeclaration(r)){var i=v.isReferencedAliasDeclaration(r);return i||1!==h.importsNotUsedAsValues?i?e.visitEachChild(r,D,n):void 0:e.setOriginalNode(e.setTextRange(p.createImportDeclaration(void 0,void 0,void 0,r.moduleReference.expression),r),r)}if(function(n){return v.isReferencedAliasDeclaration(n)||!e.isExternalModule(t)&&v.isTopLevelValueImportEqualsWithEntityName(n)}(r)){var o,s,c,l=e.createExpressionFromEntityName(p,r.moduleReference);return e.setEmitFlags(l,3584),Oe(r)||!Re(r)?e.setOriginalNode(e.setTextRange(p.createVariableStatement(e.visitNodes(r.modifiers,M,e.isModifier),p.createVariableDeclarationList([e.setOriginalNode(p.createVariableDeclaration(r.name,void 0,void 0,l),r)])),r),r):e.setOriginalNode((o=r.name,s=l,c=r,e.setTextRange(p.createExpressionStatement(p.createAssignment(p.getNamespaceMemberName(a,o,!1,!0),s)),c)),r)}}function Re(n){return void 0!==r&&e.hasSyntacticModifier(n,1)}function ke(n){return void 0===r&&e.hasSyntacticModifier(n,1)}function Oe(n){return ke(n)&&!e.hasSyntacticModifier(n,512)}function Pe(e){return p.createExpressionStatement(e)}function Me(n,t){var r=p.createAssignment(p.getExternalModuleOrNamespaceExportName(a,t,!1,!0),p.getLocalName(t));e.setSourceMapRange(r,e.createRange(t.name?t.name.pos:t.pos,t.end));var i=p.createExpressionStatement(r);e.setSourceMapRange(i,e.createRange(-1,t.end)),n.push(i)}function we(n,t,r){return e.setTextRange(p.createAssignment(Fe(n),t),r)}function Fe(e){return p.getNamespaceMemberName(a,e,!1,!0)}function Ge(n){var t=p.getGeneratedNameForNode(n);return e.setSourceMapRange(t,n.name),t}function Be(e){return p.getGeneratedNameForNode(e)}function Ue(n){if(l&d&&!e.isGeneratedIdentifier(n)&&!e.isLocalName(n)){var t=v.getReferencedExportContainer(n,!1);if(t&&297!==t.kind)if(2&d&&256===t.kind||8&d&&255===t.kind)return e.setTextRange(p.createPropertyAccessExpression(p.getGeneratedNameForNode(t),n),n)}}function Ve(n){var t=function(n){if(h.isolatedModules)return;return e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n)?v.getConstantValue(n):void 0}(n);if(void 0!==t){e.setConstantValue(n,t);var r="string"==typeof t?p.createStringLiteral(t):p.createNumericLiteral(t);if(!h.removeComments){var a=e.getOriginalNode(n,e.isAccessExpression),i=e.isPropertyAccessExpression(a)?e.declarationNameToString(a.name):e.getTextOfNode(a.argumentExpression);e.addSyntheticTrailingComment(r,3," "+i+" ")}return r}return n}}}(ts||(ts={})),function(e){var n,t;!function(e){e[e.ClassAliases=1]="ClassAliases"}(n||(n={})),function(e){e[e.InstanceField=0]="InstanceField"}(t||(t={})),e.transformClassFields=function(n){var t,r,a,i,o=n.factory,s=n.hoistVariableDeclaration,c=n.endLexicalEnvironment,l=n.resumeLexicalEnvironment,u=n.getEmitResolver(),d=n.getCompilerOptions(),p=e.getEmitScriptTarget(d),m=p<99,f=n.onSubstituteNode;n.onSubstituteNode=function(n,a){if(a=f(n,a),1===n)return function(n){switch(n.kind){case 78:return function(n){return function(n){if(1&t&&33554432&u.getNodeCheckFlags(n)){var a=u.getReferencedValueDeclaration(n);if(a){var i=r[a.id];if(i){var s=o.cloneNode(i);return e.setSourceMapRange(s,n),e.setCommentRange(s,n),s}}}return}(n)||n}(n)}return n}(a);return a};var _,g=[];return e.chainBundle(n,(function(t){var r=n.getCompilerOptions();if(t.isDeclarationFile||r.useDefineForClassFields&&99===r.target)return t;var a=e.visitEachChild(t,y,n);return e.addEmitHelpers(a,n.readEmitHelpers()),a}));function y(c){if(!(4194304&c.transformFlags))return c;switch(c.kind){case 221:case 252:return function(c){var l=a;a=void 0,m&&(g.push(_),_=void 0);var d=e.isClassDeclaration(c)?function(t){if(!e.forEach(t.members,A))return e.visitEachChild(t,y,n);var r=e.getEffectiveBaseTypeNode(t),i=!(!r||103===e.skipOuterExpressions(r.expression).kind),s=[o.updateClassDeclaration(t,void 0,t.modifiers,t.name,void 0,e.visitNodes(t.heritageClauses,y,e.isHeritageClause),D(t,i))];e.some(a)&&s.push(o.createExpressionStatement(o.inlineExpressions(a)));var c=e.getProperties(t,!0,!0);e.some(c)&&I(s,c,o.getInternalName(t));return s}(c):function(c){if(!e.forEach(c.members,A))return e.visitEachChild(c,y,n);var l=e.isClassDeclaration(e.getOriginalNode(c)),d=e.getProperties(c,!0,!0),p=e.getEffectiveBaseTypeNode(c),m=!(!p||103===e.skipOuterExpressions(p.expression).kind),f=o.updateClassExpression(c,e.visitNodes(c.decorators,y,e.isDecorator),c.modifiers,c.name,void 0,e.visitNodes(c.heritageClauses,y,e.isHeritageClause),D(c,m));if(e.some(d)||e.some(a)){if(l)return e.Debug.assertIsDefined(i,"Decorated classes transformed by TypeScript are expected to be within a variable declaration."),i&&a&&e.some(a)&&i.push(o.createExpressionStatement(o.inlineExpressions(a))),i&&e.some(d)&&I(i,d,o.getInternalName(c)),f;var _=[],g=16777216&u.getNodeCheckFlags(c),v=o.createTempVariable(s,!!g);if(g){0==(1&t)&&(t|=1,n.enableSubstitution(78),r=[]);var h=o.cloneNode(v);h.autoGenerateFlags&=-9,r[e.getOriginalNodeId(c)]=h}return e.setEmitFlags(f,65536|e.getEmitFlags(f)),_.push(e.startOnNewLine(o.createAssignment(v,f))),e.addRange(_,e.map(a,e.startOnNewLine)),e.addRange(_,function(n,t){for(var r=[],a=0,i=n;a<i.length;a++){var o=i[a],s=N(o,t);s&&(e.startOnNewLine(s),e.setSourceMapRange(s,e.moveRangePastModifiers(o)),e.setCommentRange(s,o),e.setOriginalNode(s,o),r.push(s))}return r}(d,v)),_.push(e.startOnNewLine(v)),o.inlineExpressions(_)}return f}(c);m&&(_=g.pop());return a=l,d}(c);case 163:return E(c);case 232:return function(t){var r=i;i=[];var a=e.visitEachChild(t,y,n),o=e.some(i)?__spreadArrays([a],i):a;return i=r,o}(c);case 158:return b(c);case 201:return function(t){if(m&&e.isPrivateIdentifier(t.name)){var r=O(t.name);if(r)return e.setOriginalNode(T(r,t.expression),t)}return e.visitEachChild(t,y,n)}(c);case 214:return function(t){if(m&&e.isPrivateIdentifierPropertyAccessExpression(t.operand)){var r=45===t.operator?39:46===t.operator?40:void 0,a=void 0;if(r&&(a=O(t.operand.name))){var i=x(e.visitNode(t.operand.expression,y,e.isExpression)),s=i.readExpression,c=i.initializeExpression,l=o.createPrefixUnaryExpression(39,T(a,s));return e.setOriginalNode(L(a,c||s,o.createBinaryExpression(l,r,o.createNumericLiteral(1)),62),t)}}return e.visitEachChild(t,y,n)}(c);case 215:return S(c,!1);case 203:return function(t){if(m&&e.isPrivateIdentifierPropertyAccessExpression(t.expression)){var r=o.createCallBinding(t.expression,s,p),a=r.thisArg,i=r.target;return o.updateCallExpression(t,o.createPropertyAccessExpression(e.visitNode(i,y),"call"),void 0,__spreadArrays([e.visitNode(a,y,e.isExpression)],e.visitNodes(t.arguments,y,e.isExpression)))}return e.visitEachChild(t,y,n)}(c);case 216:return function(t){if(m){if(e.isDestructuringAssignment(t)){var r=a;a=void 0,t=o.updateBinaryExpression(t,e.visitNode(t.left,v),t.operatorToken,e.visitNode(t.right,y));var i=e.some(a)?o.inlineExpressions(e.compact(__spreadArrays(a,[t]))):t;return a=r,i}if(e.isAssignmentExpression(t)&&e.isPrivateIdentifierPropertyAccessExpression(t.left)){var s=O(t.left.name);if(s)return e.setOriginalNode(L(s,t.left.expression,t.right,t.operatorToken.kind),t)}}return e.visitEachChild(t,y,n)}(c);case 79:return function(n){if(!m)return n;return e.setOriginalNode(o.createIdentifier(""),n)}(c);case 233:return function(t){if(e.isPostfixUnaryExpression(t.expression))return o.updateExpressionStatement(t,S(t.expression,!0));return e.visitEachChild(t,y,n)}(c);case 237:return function(t){if(t.incrementor&&e.isPostfixUnaryExpression(t.incrementor))return o.updateForStatement(t,e.visitNode(t.initializer,y,e.isForInitializer),e.visitNode(t.condition,y,e.isExpression),S(t.incrementor,!0),e.visitNode(t.statement,y,e.isStatement));return e.visitEachChild(t,y,n)}(c);case 205:return function(t){if(m&&e.isPrivateIdentifierPropertyAccessExpression(t.tag)){var r=o.createCallBinding(t.tag,s,p),a=r.thisArg,i=r.target;return o.updateTaggedTemplateExpression(t,o.createCallExpression(o.createPropertyAccessExpression(e.visitNode(i,y),"bind"),void 0,[e.visitNode(a,y,e.isExpression)]),void 0,e.visitNode(t.template,y,e.isTemplateLiteral))}return e.visitEachChild(t,y,n)}(c)}return e.visitEachChild(c,y,n)}function v(n){switch(n.kind){case 200:case 199:return function(n){return e.isArrayLiteralExpression(n)?o.updateArrayLiteralExpression(n,e.visitNodes(n.elements,M,e.isExpression)):o.updateObjectLiteralExpression(n,e.visitNodes(n.properties,w,e.isObjectLiteralElementLike))}(n);default:return y(n)}}function h(t){switch(t.kind){case 166:return;case 167:case 168:case 165:return e.visitEachChild(t,h,n);case 163:return E(t);case 158:return b(t);case 229:return t;default:return y(t)}}function b(t){var r=e.visitEachChild(t,y,n);if(e.some(a)){var i=a;i.push(t.expression),a=[],r=o.updateComputedPropertyName(r,o.inlineExpressions(i))}return r}function E(t){if(e.Debug.assert(!e.some(t.decorators)),!m&&e.isPrivateIdentifier(t.name))return o.updatePropertyDeclaration(t,void 0,e.visitNodes(t.modifiers,y,e.isModifier),t.name,void 0,void 0,void 0);var r=function(n,t){if(e.isComputedPropertyName(n)){var r=e.visitNode(n.expression,y,e.isExpression),a=e.skipPartiallyEmittedExpressions(r),i=e.isSimpleInlineableExpression(a);if(!(e.isAssignmentExpression(a)&&e.isGeneratedIdentifier(a.left))&&!i&&t){var c=o.getGeneratedNameForNode(n);return s(c),o.createAssignment(c,r)}return i||e.isIdentifier(a)?void 0:r}}(t.name,!!t.initializer||!!n.getCompilerOptions().useDefineForClassFields);r&&!e.isSimpleInlineableExpression(r)&&R().push(r)}function T(t,r){switch(r=e.visitNode(r,y,e.isExpression),t.placement){case 0:return n.getEmitHelperFactory().createClassPrivateFieldGetHelper(e.nodeIsSynthesized(r)?r:o.cloneNode(r),t.weakMapName);default:return e.Debug.fail("Unexpected private identifier placement")}}function S(t,r){if(m&&e.isPrivateIdentifierPropertyAccessExpression(t.operand)){var a=45===t.operator?39:46===t.operator?40:void 0,i=void 0;if(a&&(i=O(t.operand.name))){var c=x(e.visitNode(t.operand.expression,y,e.isExpression)),l=c.readExpression,u=c.initializeExpression,d=o.createPrefixUnaryExpression(39,T(i,l)),p=r?void 0:o.createTempVariable(s);return e.setOriginalNode(o.inlineExpressions(e.compact([L(i,u||l,o.createBinaryExpression(p?o.createAssignment(p,d):d,a,o.createNumericLiteral(1)),62),p])),t)}}return e.visitEachChild(t,y,n)}function x(n){var t=e.nodeIsSynthesized(n)?n:o.cloneNode(n);if(e.isSimpleInlineableExpression(n))return{readExpression:t,initializeExpression:void 0};var r=o.createTempVariable(s);return{readExpression:r,initializeExpression:o.createAssignment(r,t)}}function L(t,r,a,i){switch(t.placement){case 0:return function(t,r,a,i){if(r=e.visitNode(r,y,e.isExpression),a=e.visitNode(a,y,e.isExpression),e.isCompoundAssignment(i)){var s=x(r),c=s.readExpression,l=s.initializeExpression;return n.getEmitHelperFactory().createClassPrivateFieldSetHelper(l||c,t.weakMapName,o.createBinaryExpression(n.getEmitHelperFactory().createClassPrivateFieldGetHelper(c,t.weakMapName),e.getNonAssignmentOperatorForCompoundAssignment(i),a))}return n.getEmitHelperFactory().createClassPrivateFieldSetHelper(r,t.weakMapName,a)}(t,r,a,i);default:return e.Debug.fail("Unexpected private identifier placement")}}function A(n){return e.isPropertyDeclaration(n)||m&&n.name&&e.isPrivateIdentifier(n.name)}function D(t,r){if(m)for(var a=0,i=t.members;a<i.length;a++){var s=i[a];e.isPrivateIdentifierPropertyDeclaration(s)&&k(s.name)}var u=[],d=function(t,r){var a=e.visitNode(e.getFirstConstructorWithBody(t),y,e.isConstructorDeclaration),i=t.members.filter(C);if(!e.some(i))return a;var s=e.visitParameterList(a?a.parameters:void 0,y,n),u=function(t,r,a){var i=n.getCompilerOptions().useDefineForClassFields,s=e.getProperties(t,!1,!1);i||(s=e.filter(s,(function(n){return!!n.initializer||e.isPrivateIdentifier(n.name)})));if(!r&&!e.some(s))return e.visitFunctionBody(void 0,y,n);l();var u=0,d=[];!r&&a&&d.push(o.createExpressionStatement(o.createCallExpression(o.createSuper(),void 0,[o.createSpreadElement(o.createIdentifier("arguments"))])));r&&(u=e.addPrologueDirectivesAndInitialSuperCall(o,r,d,y));if(null==r?void 0:r.body){var p=e.findIndex(r.body.statements,(function(n){return!e.isParameterPropertyDeclaration(e.getOriginalNode(n),r)}),u);-1===p&&(p=r.body.statements.length),p>u&&(i||e.addRange(d,e.visitNodes(r.body.statements,y,e.isStatement,u,p-u)),u=p)}I(d,s,o.createThis()),r&&e.addRange(d,e.visitNodes(r.body.statements,y,e.isStatement,u));return d=o.mergeLexicalEnvironment(d,c()),e.setTextRange(o.createBlock(e.setTextRange(o.createNodeArray(d),r?r.body.statements:t.members),!0),r?r.body:void 0)}(t,a,r);if(!u)return;return e.startOnNewLine(e.setOriginalNode(e.setTextRange(o.createConstructorDeclaration(void 0,void 0,null!=s?s:[],u),a||t),a))}(t,r);return d&&u.push(d),e.addRange(u,e.visitNodes(t.members,h,e.isClassElement)),e.setTextRange(o.createNodeArray(u),t.members)}function C(t){return!(!e.isPropertyDeclaration(t)||e.hasStaticModifier(t))&&(n.getCompilerOptions().useDefineForClassFields?p<99:e.isInitializedProperty(t)||m&&e.isPrivateIdentifierPropertyDeclaration(t))}function I(n,t,r){for(var a=0,i=t;a<i.length;a++){var s=i[a],c=N(s,r);if(c){var l=o.createExpressionStatement(c);e.setSourceMapRange(l,e.moveRangePastModifiers(s)),e.setCommentRange(l,s),e.setOriginalNode(l,s),n.push(l)}}}function N(t,r){var a,i=!n.getCompilerOptions().useDefineForClassFields,s=e.isComputedPropertyName(t.name)&&!e.isSimpleInlineableExpression(t.name.expression)?o.updateComputedPropertyName(t.name,o.getGeneratedNameForNode(t.name)):t.name;if(m&&e.isPrivateIdentifier(s)){var c=O(s);if(c)switch(c.placement){case 0:return function(n,t,r){return e.factory.createCallExpression(e.factory.createPropertyAccessExpression(r,"set"),void 0,[n,t||e.factory.createVoidZero()])}(r,e.visitNode(t.initializer,y,e.isExpression),c.weakMapName)}else e.Debug.fail("Undeclared private name for property declaration.")}if((!e.isPrivateIdentifier(s)||t.initializer)&&(!e.isPrivateIdentifier(s)||t.initializer)){var l=e.getOriginalNode(t),u=t.initializer||i?null!==(a=e.visitNode(t.initializer,y,e.isExpression))&&void 0!==a?a:o.createVoidZero():e.isParameterPropertyDeclaration(l,l.parent)&&e.isIdentifier(s)?s:o.createVoidZero();if(i||e.isPrivateIdentifier(s)){var d=e.createMemberAccessForPropertyName(o,r,s,s);return o.createAssignment(d,u)}var p=e.isComputedPropertyName(s)?s.expression:e.isIdentifier(s)?o.createStringLiteral(e.unescapeLeadingUnderscores(s.escapedText)):s,f=o.createPropertyDescriptor({value:u,configurable:!0,writable:!0,enumerable:!0});return o.createObjectDefinePropertyCall(r,p,f)}}function R(){return a||(a=[])}function k(n){var t=e.getTextOfPropertyName(n),r=o.createUniqueName("_"+t.substring(1),24);s(r),(_||(_=new e.Map)).set(n.escapedText,{placement:0,weakMapName:r}),R().push(o.createAssignment(r,o.createNewExpression(o.createIdentifier("WeakMap"),void 0,[])))}function O(e){if(_&&(t=_.get(e.escapedText)))return t;for(var n=g.length-1;n>=0;--n){var t,r=g[n];if(r)if(t=r.get(e.escapedText))return t}}function P(t){var r=o.getGeneratedNameForNode(t),a=O(t.name);if(!a)return e.visitEachChild(t,y,n);var i=t.expression;return(e.isThisProperty(t)||e.isSuperProperty(t)||!e.isSimpleCopiableExpression(t.expression))&&(i=o.createTempVariable(s,!0),R().push(o.createBinaryExpression(i,62,t.expression))),o.createPropertyAccessExpression(o.createParenthesizedExpression(o.createObjectLiteralExpression([o.createSetAccessorDeclaration(void 0,void 0,"value",[o.createParameterDeclaration(void 0,void 0,void 0,r,void 0,void 0,void 0)],o.createBlock([o.createExpressionStatement(L(a,i,r,62))]))])),"value")}function M(n){var t=e.getTargetOfBindingOrAssignmentElement(n);if(t&&e.isPrivateIdentifierPropertyAccessExpression(t)){var r=P(t);return e.isAssignmentExpression(n)?o.updateBinaryExpression(n,r,n.operatorToken,e.visitNode(n.right,y,e.isExpression)):e.isSpreadElement(n)?o.updateSpreadElement(n,r):r}return e.visitNode(n,v)}function w(n){if(e.isPropertyAssignment(n)){var t=e.getTargetOfBindingOrAssignmentElement(n);if(t&&e.isPrivateIdentifierPropertyAccessExpression(t)){var r=e.getInitializerOfBindingOrAssignmentElement(n),a=P(t);return o.updatePropertyAssignment(n,e.visitNode(n.name,y),r?o.createAssignment(a,e.visitNode(r,y)):a)}return o.updatePropertyAssignment(n,e.visitNode(n.name,y),e.visitNode(n.initializer,v))}return e.visitNode(n,y)}}}(ts||(ts={})),function(e){var n,t;function r(n,t,r,a){var i=0!=(4096&t.getNodeCheckFlags(r)),o=[];return a.forEach((function(t,r){var a=e.unescapeLeadingUnderscores(r),s=[];s.push(n.createPropertyAssignment("get",n.createArrowFunction(void 0,void 0,[],void 0,void 0,e.setEmitFlags(n.createPropertyAccessExpression(e.setEmitFlags(n.createSuper(),4),a),4)))),i&&s.push(n.createPropertyAssignment("set",n.createArrowFunction(void 0,void 0,[n.createParameterDeclaration(void 0,void 0,void 0,"v",void 0,void 0,void 0)],void 0,void 0,n.createAssignment(e.setEmitFlags(n.createPropertyAccessExpression(e.setEmitFlags(n.createSuper(),4),a),4),n.createIdentifier("v"))))),o.push(n.createPropertyAssignment(a,n.createObjectLiteralExpression(s)))})),n.createVariableStatement(void 0,n.createVariableDeclarationList([n.createVariableDeclaration(n.createUniqueName("_super",48),void 0,void 0,n.createCallExpression(n.createPropertyAccessExpression(n.createIdentifier("Object"),"create"),void 0,[n.createNull(),n.createObjectLiteralExpression(o,!0)]))],2))}!function(e){e[e.AsyncMethodsWithSuper=1]="AsyncMethodsWithSuper"}(n||(n={})),function(e){e[e.NonTopLevel=1]="NonTopLevel",e[e.HasLexicalThis=2]="HasLexicalThis"}(t||(t={})),e.transformES2017=function(n){var t,a,i,o,s=n.factory,c=n.getEmitHelperFactory,l=n.resumeLexicalEnvironment,u=n.endLexicalEnvironment,d=n.hoistVariableDeclaration,p=n.getEmitResolver(),m=n.getCompilerOptions(),f=e.getEmitScriptTarget(m),_=0,g=[],y=0,v=n.onEmitNode,h=n.onSubstituteNode;return n.onEmitNode=function(n,r,a){if(1&t&&function(e){var n=e.kind;return 252===n||166===n||165===n||167===n||168===n}(r)){var i=6144&p.getNodeCheckFlags(r);if(i!==_){var o=_;return _=i,v(n,r,a),void(_=o)}}else if(t&&g[e.getNodeId(r)]){o=_;return _=0,v(n,r,a),void(_=o)}v(n,r,a)},n.onSubstituteNode=function(n,t){if(t=h(n,t),1===n&&_)return function(n){switch(n.kind){case 201:return B(n);case 202:return U(n);case 203:return function(n){var t=n.expression;if(e.isSuperProperty(t)){var r=e.isPropertyAccessExpression(t)?B(t):U(t);return s.createCallExpression(s.createPropertyAccessExpression(r,"call"),void 0,__spreadArrays([s.createThis()],n.arguments))}return n}(n)}return n}(t);return t},e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;b(1,!1),b(2,!e.isEffectiveStrictModeSourceFile(t,m));var r=e.visitEachChild(t,L,n);return e.addEmitHelpers(r,n.readEmitHelpers()),r}));function b(e,n){y=n?y|e:y&~e}function E(e){return 0!=(y&e)}function T(){return E(2)}function S(e,n,t){var r=e&~y;if(r){b(r,!0);var a=n(t);return b(r,!1),a}return n(t)}function x(t){return e.visitEachChild(t,L,n)}function L(t){if(0==(64&t.transformFlags))return t;switch(t.kind){case 129:return;case 213:return function(t){if(!E(1))return e.visitEachChild(t,L,n);return e.setOriginalNode(e.setTextRange(s.createYieldExpression(void 0,e.visitNode(t.expression,L,e.isExpression)),t),t)}(t);case 165:return S(3,D,t);case 251:return S(3,C,t);case 208:return S(3,I,t);case 209:return S(1,N,t);case 201:return i&&e.isPropertyAccessExpression(t)&&105===t.expression.kind&&i.add(t.name.escapedText),e.visitEachChild(t,L,n);case 202:return i&&105===t.expression.kind&&(o=!0),e.visitEachChild(t,L,n);case 167:case 168:case 166:case 252:case 221:return S(3,x,t);default:return e.visitEachChild(t,L,n)}}function A(t){if(e.isNodeWithPossibleHoistedDeclaration(t))switch(t.kind){case 232:return function(t){if(k(t.declarationList)){var r=O(t.declarationList,!1);return r?s.createExpressionStatement(r):void 0}return e.visitEachChild(t,L,n)}(t);case 237:return function(n){var t=n.initializer;return s.updateForStatement(n,k(t)?O(t,!1):e.visitNode(n.initializer,L,e.isForInitializer),e.visitNode(n.condition,L,e.isExpression),e.visitNode(n.incrementor,L,e.isExpression),e.visitNode(n.statement,A,e.isStatement,s.liftToBlock))}(t);case 238:return function(n){return s.updateForInStatement(n,k(n.initializer)?O(n.initializer,!0):e.visitNode(n.initializer,L,e.isForInitializer),e.visitNode(n.expression,L,e.isExpression),e.visitNode(n.statement,A,e.isStatement,s.liftToBlock))}(t);case 239:return function(n){return s.updateForOfStatement(n,e.visitNode(n.awaitModifier,L,e.isToken),k(n.initializer)?O(n.initializer,!0):e.visitNode(n.initializer,L,e.isForInitializer),e.visitNode(n.expression,L,e.isExpression),e.visitNode(n.statement,A,e.isStatement,s.liftToBlock))}(t);case 287:return function(t){var r,i=new e.Set;if(R(t.variableDeclaration,i),i.forEach((function(n,t){a.has(t)&&(r||(r=new e.Set(a)),r.delete(t))})),r){var o=a;a=r;var s=e.visitEachChild(t,A,n);return a=o,s}return e.visitEachChild(t,A,n)}(t);case 230:case 244:case 258:case 284:case 285:case 247:case 235:case 236:case 234:case 243:case 245:return e.visitEachChild(t,A,n);default:return e.Debug.assertNever(t,"Unhandled node.")}return L(t)}function D(t){return s.updateMethodDeclaration(t,void 0,e.visitNodes(t.modifiers,L,e.isModifier),t.asteriskToken,t.name,void 0,void 0,e.visitParameterList(t.parameters,L,n),void 0,2&e.getFunctionFlags(t)?F(t):e.visitFunctionBody(t.body,L,n))}function C(t){return s.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,L,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,L,n),void 0,2&e.getFunctionFlags(t)?F(t):e.visitFunctionBody(t.body,L,n))}function I(t){return s.updateFunctionExpression(t,e.visitNodes(t.modifiers,L,e.isModifier),t.asteriskToken,t.name,void 0,e.visitParameterList(t.parameters,L,n),void 0,2&e.getFunctionFlags(t)?F(t):e.visitFunctionBody(t.body,L,n))}function N(t){return s.updateArrowFunction(t,e.visitNodes(t.modifiers,L,e.isModifier),void 0,e.visitParameterList(t.parameters,L,n),void 0,t.equalsGreaterThanToken,2&e.getFunctionFlags(t)?F(t):e.visitFunctionBody(t.body,L,n))}function R(n,t){var r=n.name;if(e.isIdentifier(r))t.add(r.escapedText);else for(var a=0,i=r.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||R(o,t)}}function k(n){return!!n&&e.isVariableDeclarationList(n)&&!(3&n.flags)&&n.declarations.some(w)}function O(n,t){!function(n){e.forEach(n.declarations,P)}(n);var r=e.getInitializedVariables(n);return 0===r.length?t?e.visitNode(s.converters.convertToAssignmentElementTarget(n.declarations[0].name),L,e.isExpression):void 0:s.inlineExpressions(e.map(r,M))}function P(n){var t=n.name;if(e.isIdentifier(t))d(t);else for(var r=0,a=t.elements;r<a.length;r++){var i=a[r];e.isOmittedExpression(i)||P(i)}}function M(n){var t=e.setSourceMapRange(s.createAssignment(s.converters.convertToAssignmentElementTarget(n.name),n.initializer),n);return e.visitNode(t,L,e.isExpression)}function w(n){var t=n.name;if(e.isIdentifier(t))return a.has(t.escapedText);for(var r=0,i=t.elements;r<i.length;r++){var o=i[r];if(!e.isOmittedExpression(o)&&w(o))return!0}return!1}function F(d){l();var m=e.getOriginalNode(d,e.isFunctionLike).type,_=f<2?function(n){var t=n&&e.getEntityNameFromTypeNode(n);if(t&&e.isEntityName(t)){var r=p.getTypeReferenceSerializationKind(t);if(r===e.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue||r===e.TypeReferenceSerializationKind.Unknown)return t}return}(m):void 0,y=209===d.kind,v=0!=(8192&p.getNodeCheckFlags(d)),h=a;a=new e.Set;for(var b=0,E=d.parameters;b<E.length;b++){R(E[b],a)}var S,x=i,A=o;if(y||(i=new e.Set,o=!1),y){var D=c().createAwaiterHelper(T(),v,_,G(d.body)),C=u();if(e.some(C)){P=s.converters.convertToFunctionBlock(D);S=s.updateBlock(P,e.setTextRange(s.createNodeArray(e.concatenate(C,P.statements)),P.statements))}else S=D}else{var I=[],N=s.copyPrologue(d.body.statements,I,!1,L);I.push(s.createReturnStatement(c().createAwaiterHelper(T(),v,_,G(d.body,N)))),e.insertStatementsAfterStandardPrologue(I,u());var k=f>=2&&6144&p.getNodeCheckFlags(d);if(k&&(0==(1&t)&&(t|=1,n.enableSubstitution(203),n.enableSubstitution(201),n.enableSubstitution(202),n.enableEmitNotification(252),n.enableEmitNotification(165),n.enableEmitNotification(167),n.enableEmitNotification(168),n.enableEmitNotification(166),n.enableEmitNotification(232)),i.size)){var O=r(s,p,d,i);g[e.getNodeId(O)]=!0,e.insertStatementsAfterStandardPrologue(I,[O])}var P=s.createBlock(I,!0);e.setTextRange(P,d.body),k&&o&&(4096&p.getNodeCheckFlags(d)?e.addEmitHelper(P,e.advancedAsyncSuperHelper):2048&p.getNodeCheckFlags(d)&&e.addEmitHelper(P,e.asyncSuperHelper)),S=P}return a=h,y||(i=x,o=A),S}function G(n,t){return e.isBlock(n)?s.updateBlock(n,e.visitNodes(n.statements,A,e.isStatement,t)):s.converters.convertToFunctionBlock(e.visitNode(n,A,e.isConciseBody))}function B(n){return 105===n.expression.kind?e.setTextRange(s.createPropertyAccessExpression(s.createUniqueName("_super",48),n.name),n):n}function U(n){return 105===n.expression.kind?(t=n.argumentExpression,r=n,4096&_?e.setTextRange(s.createPropertyAccessExpression(s.createCallExpression(s.createUniqueName("_superIndex",48),void 0,[t]),"value"),r):e.setTextRange(s.createCallExpression(s.createUniqueName("_superIndex",48),void 0,[t]),r)):n;var t,r}},e.createSuperAccessVariableStatement=r}(ts||(ts={})),function(e){var n,t;!function(e){e[e.AsyncMethodsWithSuper=1]="AsyncMethodsWithSuper"}(n||(n={})),function(e){e[e.None=0]="None",e[e.HasLexicalThis=1]="HasLexicalThis",e[e.IterationContainer=2]="IterationContainer",e[e.AncestorFactsMask=3]="AncestorFactsMask",e[e.SourceFileIncludes=1]="SourceFileIncludes",e[e.SourceFileExcludes=2]="SourceFileExcludes",e[e.StrictModeSourceFileIncludes=0]="StrictModeSourceFileIncludes",e[e.ClassOrFunctionIncludes=1]="ClassOrFunctionIncludes",e[e.ClassOrFunctionExcludes=2]="ClassOrFunctionExcludes",e[e.ArrowFunctionIncludes=0]="ArrowFunctionIncludes",e[e.ArrowFunctionExcludes=2]="ArrowFunctionExcludes",e[e.IterationStatementIncludes=2]="IterationStatementIncludes",e[e.IterationStatementExcludes=0]="IterationStatementExcludes"}(t||(t={})),e.transformES2018=function(n){var t=n.factory,r=n.getEmitHelperFactory,a=n.resumeLexicalEnvironment,i=n.endLexicalEnvironment,o=n.hoistVariableDeclaration,s=n.getEmitResolver(),c=n.getCompilerOptions(),l=e.getEmitScriptTarget(c),u=n.onEmitNode;n.onEmitNode=function(n,t,r){if(1&p&&function(e){var n=e.kind;return 252===n||166===n||165===n||167===n||168===n}(t)){var a=6144&s.getNodeCheckFlags(t);if(a!==h){var i=h;return h=a,u(n,t,r),void(h=i)}}else if(p&&E[e.getNodeId(t)]){i=h;return h=0,u(n,t,r),void(h=i)}u(n,t,r)};var d=n.onSubstituteNode;n.onSubstituteNode=function(n,r){if(r=d(n,r),1===n&&h)return function(n){switch(n.kind){case 201:return W(n);case 202:return z(n);case 203:return function(n){var r=n.expression;if(e.isSuperProperty(r)){var a=e.isPropertyAccessExpression(r)?W(r):z(r);return t.createCallExpression(t.createPropertyAccessExpression(a,"call"),void 0,__spreadArrays([t.createThis()],n.arguments))}return n}(n)}return n}(r);return r};var p,m,f,_,g,y,v=!1,h=0,b=0,E=[];return e.chainBundle(n,(function(r){if(r.isDeclarationFile)return r;f=r;var a=function(r){var a=T(2,e.isEffectiveStrictModeSourceFile(r,c)?0:1);v=!1;var i=e.visitEachChild(r,L,n),o=e.concatenate(i.statements,_&&[t.createVariableStatement(void 0,t.createVariableDeclarationList(_))]),s=t.updateSourceFile(i,e.setTextRange(t.createNodeArray(o),r.statements));return S(a),s}(r);return e.addEmitHelpers(a,n.readEmitHelpers()),f=void 0,_=void 0,a}));function T(e,n){var t=b;return b=3&(b&~e|n),t}function S(e){b=e}function x(n){_=e.append(_,t.createVariableDeclaration(n))}function L(e){return N(e,!1)}function A(e){return N(e,!0)}function D(e){if(129!==e.kind)return e}function C(e,n,t,r){if(function(e,n){return b!==(b&~e|n)}(t,r)){var a=T(t,r),i=e(n);return S(a),i}return e(n)}function I(t){return e.visitEachChild(t,L,n)}function N(a,i){if(0==(32&a.transformFlags))return a;switch(a.kind){case 213:return function(a){if(2&m&&1&m)return e.setOriginalNode(e.setTextRange(t.createYieldExpression(void 0,r().createAwaitHelper(e.visitNode(a.expression,L,e.isExpression))),a),a);return e.visitEachChild(a,L,n)}(a);case 219:return function(a){if(2&m&&1&m){if(a.asteriskToken){var i=e.visitNode(e.Debug.assertDefined(a.expression),L,e.isExpression);return e.setOriginalNode(e.setTextRange(t.createYieldExpression(void 0,r().createAwaitHelper(t.updateYieldExpression(a,a.asteriskToken,e.setTextRange(r().createAsyncDelegatorHelper(e.setTextRange(r().createAsyncValuesHelper(i),i)),i)))),a),a)}return e.setOriginalNode(e.setTextRange(t.createYieldExpression(void 0,P(a.expression?e.visitNode(a.expression,L,e.isExpression):t.createVoidZero())),a),a)}return e.visitEachChild(a,L,n)}(a);case 242:return function(r){if(2&m&&1&m)return t.updateReturnStatement(r,P(r.expression?e.visitNode(r.expression,L,e.isExpression):t.createVoidZero()));return e.visitEachChild(r,L,n)}(a);case 245:return function(r){if(2&m){var a=e.unwrapInnermostStatementOfLabel(r);return 239===a.kind&&a.awaitModifier?O(a,r):t.restoreEnclosingLabel(e.visitNode(a,L,e.isStatement,t.liftToBlock),r)}return e.visitEachChild(r,L,n)}(a);case 200:return function(a){if(16384&a.transformFlags){var i=function(n){for(var r,a=[],i=0,o=n;i<o.length;i++){var s=o[i];if(290===s.kind){r&&(a.push(t.createObjectLiteralExpression(r)),r=void 0);var c=s.expression;a.push(e.visitNode(c,L,e.isExpression))}else r=e.append(r,288===s.kind?t.createPropertyAssignment(s.name,e.visitNode(s.initializer,L,e.isExpression)):e.visitNode(s,L,e.isObjectLiteralElementLike))}r&&a.push(t.createObjectLiteralExpression(r));return a}(a.properties);i.length&&200!==i[0].kind&&i.unshift(t.createObjectLiteralExpression());var o=i[0];if(i.length>1){for(var s=1;s<i.length;s++)o=r().createAssignHelper([o,i[s]]);return o}return r().createAssignHelper(i)}return e.visitEachChild(a,L,n)}(a);case 216:return function(r,a){if(e.isDestructuringAssignment(r)&&16384&r.left.transformFlags)return e.flattenDestructuringAssignment(r,L,n,1,!a);if(27===r.operatorToken.kind)return t.updateBinaryExpression(r,e.visitNode(r.left,A,e.isExpression),r.operatorToken,e.visitNode(r.right,a?A:L,e.isExpression));return e.visitEachChild(r,L,n)}(a,i);case 337:return function(r,a){if(a)return e.visitEachChild(r,A,n);for(var i,o=0;o<r.elements.length;o++){var s=r.elements[o],c=e.visitNode(s,o<r.elements.length-1?A:L,e.isExpression);(i||c!==s)&&(i||(i=r.elements.slice(0,o)),i.push(c))}var l=i?e.setTextRange(t.createNodeArray(i),r.elements):r.elements;return t.updateCommaListExpression(r,l)}(a,i);case 287:return function(r){if(r.variableDeclaration&&e.isBindingPattern(r.variableDeclaration.name)&&16384&r.variableDeclaration.name.transformFlags){var a=t.getGeneratedNameForNode(r.variableDeclaration.name),i=t.updateVariableDeclaration(r.variableDeclaration,r.variableDeclaration.name,void 0,void 0,a),o=e.flattenDestructuringBinding(i,L,n,1),s=e.visitNode(r.block,L,e.isBlock);return e.some(o)&&(s=t.updateBlock(s,__spreadArrays([t.createVariableStatement(void 0,o)],s.statements))),t.updateCatchClause(r,t.updateVariableDeclaration(r.variableDeclaration,a,void 0,void 0,void 0),s)}return e.visitEachChild(r,L,n)}(a);case 232:return function(t){if(e.hasSyntacticModifier(t,1)){var r=v;v=!0;var a=e.visitEachChild(t,L,n);return v=r,a}return e.visitEachChild(t,L,n)}(a);case 249:return function(e){if(v){var n=v;v=!1;var t=R(e,!0);return v=n,t}return R(e,!1)}(a);case 235:case 236:case 238:return C(I,a,0,2);case 239:return O(a,void 0);case 237:return C(k,a,0,2);case 212:return function(t){return e.visitEachChild(t,A,n)}(a);case 166:return C(M,a,2,1);case 165:return C(G,a,2,1);case 167:return C(w,a,2,1);case 168:return C(F,a,2,1);case 251:return C(B,a,2,1);case 208:return C(V,a,2,1);case 209:return C(U,a,2,0);case 160:return function(r){if(16384&r.transformFlags)return t.updateParameterDeclaration(r,void 0,void 0,r.dotDotDotToken,t.getGeneratedNameForNode(r),void 0,void 0,e.visitNode(r.initializer,L,e.isExpression));return e.visitEachChild(r,L,n)}(a);case 233:return function(t){return e.visitEachChild(t,A,n)}(a);case 207:return function(t,r){return e.visitEachChild(t,r?A:L,n)}(a,i);case 205:return function(t){return e.processTaggedTemplateExpression(n,t,L,f,x,e.ProcessLevel.LiftRestriction)}(a);case 201:return g&&e.isPropertyAccessExpression(a)&&105===a.expression.kind&&g.add(a.name.escapedText),e.visitEachChild(a,L,n);case 202:return g&&105===a.expression.kind&&(y=!0),e.visitEachChild(a,L,n);case 252:case 221:return C(I,a,2,1);default:return e.visitEachChild(a,L,n)}}function R(t,r){return e.isBindingPattern(t.name)&&16384&t.name.transformFlags?e.flattenDestructuringBinding(t,L,n,1,void 0,r):e.visitEachChild(t,L,n)}function k(n){return t.updateForStatement(n,e.visitNode(n.initializer,A,e.isForInitializer),e.visitNode(n.condition,L,e.isExpression),e.visitNode(n.incrementor,A,e.isExpression),e.visitNode(n.statement,L,e.isStatement))}function O(a,i){var s=T(0,2);16384&a.initializer.transformFlags&&(a=function(n){var r=e.skipParentheses(n.initializer);if(e.isVariableDeclarationList(r)||e.isAssignmentPattern(r)){var a=void 0,i=void 0,o=t.createTempVariable(void 0),s=[e.createForOfBindingStatement(t,r,o)];return e.isBlock(n.statement)?(e.addRange(s,n.statement.statements),a=n.statement,i=n.statement.statements):n.statement&&(e.append(s,n.statement),a=n.statement,i=n.statement),t.updateForOfStatement(n,n.awaitModifier,e.setTextRange(t.createVariableDeclarationList([e.setTextRange(t.createVariableDeclaration(o),n.initializer)],1),n.initializer),n.expression,e.setTextRange(t.createBlock(e.setTextRange(t.createNodeArray(s),i),!0),a))}return n}(a));var c=a.awaitModifier?function(n,a,i){var s=e.visitNode(n.expression,L,e.isExpression),c=e.isIdentifier(s)?t.getGeneratedNameForNode(s):t.createTempVariable(void 0),l=e.isIdentifier(s)?t.getGeneratedNameForNode(c):t.createTempVariable(void 0),u=t.createUniqueName("e"),d=t.getGeneratedNameForNode(u),p=t.createTempVariable(void 0),m=e.setTextRange(r().createAsyncValuesHelper(s),n.expression),f=t.createCallExpression(t.createPropertyAccessExpression(c,"next"),void 0,[]),_=t.createPropertyAccessExpression(l,"done"),g=t.createPropertyAccessExpression(l,"value"),y=t.createFunctionCallCall(p,c,[]);o(u),o(p);var v=2&i?t.inlineExpressions([t.createAssignment(u,t.createVoidZero()),m]):m,h=e.setEmitFlags(e.setTextRange(t.createForStatement(e.setEmitFlags(e.setTextRange(t.createVariableDeclarationList([e.setTextRange(t.createVariableDeclaration(c,void 0,void 0,v),n.expression),t.createVariableDeclaration(l)]),n.expression),2097152),t.createComma(t.createAssignment(l,P(f)),t.createLogicalNot(_)),void 0,function(n,r){var a,i,o=e.createForOfBindingStatement(t,n.initializer,r),s=[e.visitNode(o,L,e.isStatement)],c=e.visitNode(n.statement,L,e.isStatement);e.isBlock(c)?(e.addRange(s,c.statements),a=c,i=c.statements):s.push(c);return e.setEmitFlags(e.setTextRange(t.createBlock(e.setTextRange(t.createNodeArray(s),i),!0),a),432)}(n,g)),n),256);return t.createTryStatement(t.createBlock([t.restoreEnclosingLabel(h,a)]),t.createCatchClause(t.createVariableDeclaration(d),e.setEmitFlags(t.createBlock([t.createExpressionStatement(t.createAssignment(u,t.createObjectLiteralExpression([t.createPropertyAssignment("error",d)])))]),1)),t.createBlock([t.createTryStatement(t.createBlock([e.setEmitFlags(t.createIfStatement(t.createLogicalAnd(t.createLogicalAnd(l,t.createLogicalNot(_)),t.createAssignment(p,t.createPropertyAccessExpression(c,"return"))),t.createExpressionStatement(P(y))),1)]),void 0,e.setEmitFlags(t.createBlock([e.setEmitFlags(t.createIfStatement(u,t.createThrowStatement(t.createPropertyAccessExpression(u,"error"))),1)]),1))]))}(a,i,s):t.restoreEnclosingLabel(e.visitEachChild(a,L,n),i);return S(s),c}function P(e){return 1&m?t.createYieldExpression(void 0,r().createAwaitHelper(e)):t.createAwaitExpression(e)}function M(r){var a=m;m=0;var i=t.updateConstructorDeclaration(r,void 0,r.modifiers,e.visitParameterList(r.parameters,L,n),H(r));return m=a,i}function w(r){var a=m;m=0;var i=t.updateGetAccessorDeclaration(r,void 0,r.modifiers,e.visitNode(r.name,L,e.isPropertyName),e.visitParameterList(r.parameters,L,n),void 0,H(r));return m=a,i}function F(r){var a=m;m=0;var i=t.updateSetAccessorDeclaration(r,void 0,r.modifiers,e.visitNode(r.name,L,e.isPropertyName),e.visitParameterList(r.parameters,L,n),H(r));return m=a,i}function G(r){var a=m;m=e.getFunctionFlags(r);var i=t.updateMethodDeclaration(r,void 0,1&m?e.visitNodes(r.modifiers,D,e.isModifier):r.modifiers,2&m?void 0:r.asteriskToken,e.visitNode(r.name,L,e.isPropertyName),e.visitNode(void 0,L,e.isToken),void 0,e.visitParameterList(r.parameters,L,n),void 0,2&m&&1&m?j(r):H(r));return m=a,i}function B(r){var a=m;m=e.getFunctionFlags(r);var i=t.updateFunctionDeclaration(r,void 0,1&m?e.visitNodes(r.modifiers,D,e.isModifier):r.modifiers,2&m?void 0:r.asteriskToken,r.name,void 0,e.visitParameterList(r.parameters,L,n),void 0,2&m&&1&m?j(r):H(r));return m=a,i}function U(r){var a=m;m=e.getFunctionFlags(r);var i=t.updateArrowFunction(r,r.modifiers,void 0,e.visitParameterList(r.parameters,L,n),void 0,r.equalsGreaterThanToken,H(r));return m=a,i}function V(r){var a=m;m=e.getFunctionFlags(r);var i=t.updateFunctionExpression(r,1&m?e.visitNodes(r.modifiers,D,e.isModifier):r.modifiers,2&m?void 0:r.asteriskToken,r.name,void 0,e.visitParameterList(r.parameters,L,n),void 0,2&m&&1&m?j(r):H(r));return m=a,i}function j(o){a();var c=[],u=t.copyPrologue(o.body.statements,c,!1,L);K(c,o);var d=g,m=y;g=new e.Set,y=!1;var f=t.createReturnStatement(r().createAsyncGeneratorHelper(t.createFunctionExpression(void 0,t.createToken(41),o.name&&t.getGeneratedNameForNode(o.name),void 0,[],void 0,t.updateBlock(o.body,e.visitLexicalEnvironment(o.body.statements,L,n,u))),!!(1&b))),_=l>=2&&6144&s.getNodeCheckFlags(o);if(_){0==(1&p)&&(p|=1,n.enableSubstitution(203),n.enableSubstitution(201),n.enableSubstitution(202),n.enableEmitNotification(252),n.enableEmitNotification(165),n.enableEmitNotification(167),n.enableEmitNotification(168),n.enableEmitNotification(166),n.enableEmitNotification(232));var v=e.createSuperAccessVariableStatement(t,s,o,g);E[e.getNodeId(v)]=!0,e.insertStatementsAfterStandardPrologue(c,[v])}c.push(f),e.insertStatementsAfterStandardPrologue(c,i());var h=t.updateBlock(o.body,c);return _&&y&&(4096&s.getNodeCheckFlags(o)?e.addEmitHelper(h,e.advancedAsyncSuperHelper):2048&s.getNodeCheckFlags(o)&&e.addEmitHelper(h,e.asyncSuperHelper)),g=d,y=m,h}function H(n){var r;a();var o=0,s=[],c=null!==(r=e.visitNode(n.body,L,e.isConciseBody))&&void 0!==r?r:t.createBlock([]);e.isBlock(c)&&(o=t.copyPrologue(c.statements,s,!1,L)),e.addRange(s,K(void 0,n));var l=i();if(o>0||e.some(s)||e.some(l)){var u=t.converters.convertToFunctionBlock(c,!0);return e.insertStatementsAfterStandardPrologue(s,l),e.addRange(s,u.statements.slice(o)),t.updateBlock(u,e.setTextRange(t.createNodeArray(s),u.statements))}return c}function K(r,a){for(var i=0,o=a.parameters;i<o.length;i++){var s=o[i];if(16384&s.transformFlags){var c=t.getGeneratedNameForNode(s),l=e.flattenDestructuringBinding(s,L,n,1,c,!1,!0);if(e.some(l)){var u=t.createVariableStatement(void 0,t.createVariableDeclarationList(l));e.setEmitFlags(u,1048576),r=e.append(r,u)}}}return r}function W(n){return 105===n.expression.kind?e.setTextRange(t.createPropertyAccessExpression(t.createUniqueName("_super",48),n.name),n):n}function z(n){return 105===n.expression.kind?(r=n.argumentExpression,a=n,4096&h?e.setTextRange(t.createPropertyAccessExpression(t.createCallExpression(t.createIdentifier("_superIndex"),void 0,[r]),"value"),a):e.setTextRange(t.createCallExpression(t.createIdentifier("_superIndex"),void 0,[r]),a)):n;var r,a}}}(ts||(ts={})),function(e){e.transformES2019=function(n){var t=n.factory;return e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;return e.visitEachChild(t,r,n)}));function r(a){if(0==(16&a.transformFlags))return a;switch(a.kind){case 287:return function(a){if(!a.variableDeclaration)return t.updateCatchClause(a,t.createVariableDeclaration(t.createTempVariable(void 0)),e.visitNode(a.block,r,e.isBlock));return e.visitEachChild(a,r,n)}(a);default:return e.visitEachChild(a,r,n)}}}}(ts||(ts={})),function(e){e.transformES2020=function(n){var t=n.factory,r=n.hoistVariableDeclaration;return e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;return e.visitEachChild(t,a,n)}));function a(c){if(0==(8&c.transformFlags))return c;switch(c.kind){case 201:case 202:case 203:if(32&c.flags){var l=o(c,!1,!1);return e.Debug.assertNotNode(l,e.isSyntheticReference),l}return e.visitEachChild(c,a,n);case 216:return 60===c.operatorToken.kind?function(n){var i=e.visitNode(n.left,a,e.isExpression),o=i;e.isSimpleCopiableExpression(i)||(o=t.createTempVariable(r),i=t.createAssignment(o,i));return t.createConditionalExpression(s(i,o),void 0,o,void 0,e.visitNode(n.right,a,e.isExpression))}(c):e.visitEachChild(c,a,n);case 210:return function(n){return e.isOptionalChain(e.skipParentheses(n.expression))?e.setOriginalNode(i(n.expression,!1,!0),n):t.updateDeleteExpression(n,e.visitNode(n.expression,a,e.isExpression))}(c);default:return e.visitEachChild(c,a,n)}}function i(s,c,l){switch(s.kind){case 207:return function(n,r,a){var o=i(n.expression,r,a);return e.isSyntheticReference(o)?t.createSyntheticReferenceExpression(t.updateParenthesizedExpression(n,o.expression),o.thisArg):t.updateParenthesizedExpression(n,o)}(s,c,l);case 201:case 202:return function(n,i,s){if(e.isOptionalChain(n))return o(n,i,s);var c,l=e.visitNode(n.expression,a,e.isExpression);return e.Debug.assertNotNode(l,e.isSyntheticReference),i&&(e.isSimpleCopiableExpression(l)?c=l:(c=t.createTempVariable(r),l=t.createAssignment(c,l))),l=201===n.kind?t.updatePropertyAccessExpression(n,l,e.visitNode(n.name,a,e.isIdentifier)):t.updateElementAccessExpression(n,l,e.visitNode(n.argumentExpression,a,e.isExpression)),c?t.createSyntheticReferenceExpression(l,c):l}(s,c,l);case 203:return function(t,r){return e.isOptionalChain(t)?o(t,r,!1):e.visitEachChild(t,a,n)}(s,c);default:return e.visitNode(s,a,e.isExpression)}}function o(n,o,c){var l=function(n){e.Debug.assertNotNode(n,e.isNonNullChain);for(var t=[n];!n.questionDotToken&&!e.isTaggedTemplateExpression(n);)n=e.cast(e.skipPartiallyEmittedExpressions(n.expression),e.isOptionalChain),e.Debug.assertNotNode(n,e.isNonNullChain),t.unshift(n);return{expression:n.expression,chain:t}}(n),u=l.expression,d=l.chain,p=i(u,e.isCallChain(d[0]),!1),m=e.isSyntheticReference(p)?p.thisArg:void 0,f=e.isSyntheticReference(p)?p.expression:p,_=f;e.isSimpleCopiableExpression(f)||(_=t.createTempVariable(r),f=t.createAssignment(_,f));for(var g,y=_,v=0;v<d.length;v++){var h=d[v];switch(h.kind){case 201:case 202:v===d.length-1&&o&&(e.isSimpleCopiableExpression(y)?g=y:(g=t.createTempVariable(r),y=t.createAssignment(g,y))),y=201===h.kind?t.createPropertyAccessExpression(y,e.visitNode(h.name,a,e.isIdentifier)):t.createElementAccessExpression(y,e.visitNode(h.argumentExpression,a,e.isExpression));break;case 203:y=0===v&&m?t.createFunctionCallCall(y,105===m.kind?t.createThis():m,e.visitNodes(h.arguments,a,e.isExpression)):t.createCallExpression(y,void 0,e.visitNodes(h.arguments,a,e.isExpression))}e.setOriginalNode(y,h)}var b=c?t.createConditionalExpression(s(f,_,!0),void 0,t.createTrue(),void 0,t.createDeleteExpression(y)):t.createConditionalExpression(s(f,_,!0),void 0,t.createVoidZero(),void 0,y);return g?t.createSyntheticReferenceExpression(b,g):b}function s(e,n,r){return t.createBinaryExpression(t.createBinaryExpression(e,t.createToken(r?36:37),t.createNull()),t.createToken(r?56:55),t.createBinaryExpression(n,t.createToken(r?36:37),t.createVoidZero()))}}}(ts||(ts={})),function(e){e.transformESNext=function(n){var t=n.hoistVariableDeclaration,r=n.factory;return e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;return e.visitEachChild(t,a,n)}));function a(i){if(0==(4&i.transformFlags))return i;switch(i.kind){case 216:var o=i;if(e.isLogicalOrCoalescingAssignmentExpression(o))return function(n){var i=n.operatorToken,o=e.getNonAssignmentOperatorForCompoundAssignment(i.kind),s=e.skipParentheses(e.visitNode(n.left,a,e.isLeftHandSideExpression)),c=s,l=e.skipParentheses(e.visitNode(n.right,a,e.isExpression));if(e.isAccessExpression(s)){var u=e.isSimpleCopiableExpression(s.expression),d=u?s.expression:r.createTempVariable(t),p=u?s.expression:r.createAssignment(d,s.expression);if(e.isPropertyAccessExpression(s))c=r.createPropertyAccessExpression(d,s.name),s=r.createPropertyAccessExpression(p,s.name);else{var m=e.isSimpleCopiableExpression(s.argumentExpression),f=m?s.argumentExpression:r.createTempVariable(t);c=r.createElementAccessExpression(d,f),s=r.createElementAccessExpression(p,m?s.argumentExpression:r.createAssignment(f,s.argumentExpression))}}return r.createBinaryExpression(s,o,r.createParenthesizedExpression(r.createAssignment(c,l)))}(o);default:return e.visitEachChild(i,a,n)}}}}(ts||(ts={})),function(e){e.transformJsx=function(t){var r,a,i=t.factory,o=t.getEmitHelperFactory,s=t.getCompilerOptions();return e.chainBundle(t,(function(n){if(n.isDeclarationFile)return n;r=n,(a={}).importSpecifier=e.getJSXImplicitImportBase(s,n);var o=e.visitEachChild(n,d,t);e.addEmitHelpers(o,t.readEmitHelpers());var c=o.statements;a.filenameDeclaration&&(c=e.insertStatementAfterCustomPrologue(c.slice(),i.createVariableStatement(void 0,i.createVariableDeclarationList([a.filenameDeclaration],2))));if(a.utilizedImplicitRuntimeImports)for(var l=0,u=e.arrayFrom(a.utilizedImplicitRuntimeImports.entries());l<u.length;l++){var p=u[l],m=p[0],f=p[1];if(e.isExternalModule(n)){var _=i.createImportDeclaration(void 0,void 0,i.createImportClause(!1,void 0,i.createNamedImports(e.arrayFrom(f.values()))),i.createStringLiteral(m));e.setParentRecursive(_,!1),c=e.insertStatementAfterCustomPrologue(c.slice(),_)}else if(e.isExternalOrCommonJsModule(n)){var g=i.createVariableStatement(void 0,i.createVariableDeclarationList([i.createVariableDeclaration(i.createObjectBindingPattern(e.map(e.arrayFrom(f.values()),(function(e){return i.createBindingElement(void 0,e.propertyName,e.name)}))),void 0,void 0,i.createCallExpression(i.createIdentifier("require"),void 0,[i.createStringLiteral(m)]))],2));e.setParentRecursive(g,!1),c=e.insertStatementAfterCustomPrologue(c.slice(),g)}}c!==o.statements&&(o=i.updateSourceFile(o,c));return a=void 0,o}));function c(){if(a.filenameDeclaration)return a.filenameDeclaration.name;var e=i.createVariableDeclaration(i.createUniqueName("_jsxFileName",48),void 0,void 0,i.createStringLiteral(r.fileName));return a.filenameDeclaration=e,a.filenameDeclaration.name}function l(e){return u(function(e){return 5===s.jsx?"jsxDEV":e>1?"jsxs":"jsx"}(e))}function u(n){var t,r,o="createElement"===n?a.importSpecifier:e.getJSXRuntimeImport(a.importSpecifier,s),c=null===(r=null===(t=a.utilizedImplicitRuntimeImports)||void 0===t?void 0:t.get(o))||void 0===r?void 0:r.get(n);if(c)return c.name;a.utilizedImplicitRuntimeImports||(a.utilizedImplicitRuntimeImports=e.createMap());var l=a.utilizedImplicitRuntimeImports.get(o);l||(l=e.createMap(),a.utilizedImplicitRuntimeImports.set(o,l));var u=i.createUniqueName("_"+n,112),d=i.createImportSpecifier(i.createIdentifier(n),u);return u.generatedImportReference=d,l.set(n,d),u}function d(n){return 2&n.transformFlags?function(n){switch(n.kind){case 273:return f(n,!1);case 274:return _(n,!1);case 277:return g(n,!1);case 283:return I(n);default:return e.visitEachChild(n,d,t)}}(n):n}function p(n){switch(n.kind){case 11:return function(n){var t=function(n){for(var t,r=0,a=-1,i=0;i<n.length;i++){var o=n.charCodeAt(i);e.isLineBreak(o)?(-1!==r&&-1!==a&&(t=A(t,n.substr(r,a-r+1))),r=-1):e.isWhiteSpaceSingleLine(o)||(a=i,-1===r&&(r=i))}return-1!==r?A(t,n.substr(r)):t}(n.text);return void 0===t?void 0:i.createStringLiteral(t)}(n);case 283:return I(n);case 273:return f(n,!0);case 274:return _(n,!0);case 277:return g(n,!0);default:return e.Debug.failBadSyntaxKind(n)}}function m(n){return void 0===a.importSpecifier||function(n){for(var t=!1,r=0,a=n.attributes.properties;r<a.length;r++){var i=a[r];if(e.isJsxSpreadAttribute(i))t=!0;else if(t&&e.isJsxAttribute(i)&&"key"===i.name.escapedText)return!0}return!1}(n)}function f(e,n){return(m(e.openingElement)?b:v)(e.openingElement,e.children,n,e)}function _(e,n){return(m(e)?b:v)(e,void 0,n,e)}function g(e,n){return(void 0===a.importSpecifier?T:E)(e.openingFragment,e.children,n,e)}function y(n){var t=e.getSemanticJsxChildren(n);if(1===e.length(t)){var r=p(t[0]);return r&&i.createObjectLiteralExpression([i.createPropertyAssignment("children",r)])}var a=e.mapDefined(n,p);return a.length?i.createObjectLiteralExpression([i.createPropertyAssignment("children",i.createArrayLiteralExpression(a))]):void 0}function v(n,t,r,a){var s=C(n),c=e.find(n.attributes.properties,(function(n){return!!n.name&&e.isIdentifier(n.name)&&"key"===n.name.escapedText})),l=c?e.filter(n.attributes.properties,(function(e){return e!==c})):n.attributes.properties,u=[];if(l.length&&(u=e.flatten(e.spanMap(l,e.isJsxSpreadAttribute,(function(n,t){return t?e.map(n,S):i.createObjectLiteralExpression(e.map(n,x))}))),e.isJsxSpreadAttribute(l[0])&&u.unshift(i.createObjectLiteralExpression())),t&&t.length){var d=y(t);d&&u.push(d)}return h(s,0===u.length?i.createObjectLiteralExpression([]):e.singleOrUndefined(u)||o().createAssignHelper(u),c,e.length(e.getSemanticJsxChildren(t||e.emptyArray)),r,a)}function h(n,t,a,o,u,d){var p=[n,t,a?L(a.initializer):i.createVoidZero()];if(5===s.jsx){var m=e.getOriginalNode(r);if(m&&e.isSourceFile(m)){p.push(o>1?i.createTrue():i.createFalse());var f=e.getLineAndCharacterOfPosition(m,d.pos);p.push(i.createObjectLiteralExpression([i.createPropertyAssignment("fileName",c()),i.createPropertyAssignment("lineNumber",i.createNumericLiteral(f.line+1)),i.createPropertyAssignment("columnNumber",i.createNumericLiteral(f.character+1))])),p.push(i.createThis())}}var _=e.setTextRange(i.createCallExpression(l(o),void 0,p),d);return u&&e.startOnNewLine(_),_}function b(n,c,l,d){var m,f=C(n),_=n.attributes.properties;if(0===_.length)m=i.createNull();else{var g=e.flatten(e.spanMap(_,e.isJsxSpreadAttribute,(function(n,t){return t?e.map(n,S):i.createObjectLiteralExpression(e.map(n,x))})));e.isJsxSpreadAttribute(_[0])&&g.unshift(i.createObjectLiteralExpression()),(m=e.singleOrUndefined(g))||(m=o().createAssignHelper(g))}var y=void 0===a.importSpecifier?e.createJsxFactoryExpression(i,t.getEmitResolver().getJsxFactoryEntity(r),s.reactNamespace,n):u("createElement"),v=e.createExpressionForJsxElement(i,y,f,m,e.mapDefined(c,p),d);return l&&e.startOnNewLine(v),v}function E(n,t,r,a){var o;if(t&&t.length){var s=y(t);s&&(o=s)}return h(u("Fragment"),o||i.createObjectLiteralExpression([]),void 0,e.length(e.getSemanticJsxChildren(t)),r,a)}function T(n,a,o,c){var l=e.createExpressionForJsxFragment(i,t.getEmitResolver().getJsxFactoryEntity(r),t.getEmitResolver().getJsxFragmentFactoryEntity(r),s.reactNamespace,e.mapDefined(a,p),n,c);return o&&e.startOnNewLine(l),l}function S(n){return e.visitNode(n.expression,d,e.isExpression)}function x(n){var t=function(n){var t=n.name,r=e.idText(t);return/^[A-Za-z_]\w*$/.test(r)?t:i.createStringLiteral(r)}(n),r=L(n.initializer);return i.createPropertyAssignment(t,r)}function L(n){if(void 0===n)return i.createTrue();if(10===n.kind){var t=void 0!==n.singleQuote?n.singleQuote:!e.isStringDoubleQuoted(n,r),a=i.createStringLiteral((o=n.text,((s=D(o))===o?void 0:s)||n.text),t);return e.setTextRange(a,n)}return 283===n.kind?void 0===n.expression?i.createTrue():e.visitNode(n.expression,d,e.isExpression):e.Debug.failBadSyntaxKind(n);var o,s}function A(e,n){var t=D(n);return void 0===e?t:e+" "+t}function D(t){return t.replace(/&((#((\d+)|x([\da-fA-F]+)))|(\w+));/g,(function(t,r,a,i,o,s,c){if(o)return e.utf16EncodeAsString(parseInt(o,10));if(s)return e.utf16EncodeAsString(parseInt(s,16));var l=n.get(c);return l?e.utf16EncodeAsString(l):t}))}function C(n){if(273===n.kind)return C(n.openingElement);var t=n.tagName;return e.isIdentifier(t)&&e.isIntrinsicJsxName(t.escapedText)?i.createStringLiteral(e.idText(t)):e.createExpressionFromEntityName(i,t)}function I(n){return e.visitNode(n.expression,d,e.isExpression)}};var n=new e.Map(e.getEntries({quot:34,amp:38,apos:39,lt:60,gt:62,nbsp:160,iexcl:161,cent:162,pound:163,curren:164,yen:165,brvbar:166,sect:167,uml:168,copy:169,ordf:170,laquo:171,not:172,shy:173,reg:174,macr:175,deg:176,plusmn:177,sup2:178,sup3:179,acute:180,micro:181,para:182,middot:183,cedil:184,sup1:185,ordm:186,raquo:187,frac14:188,frac12:189,frac34:190,iquest:191,Agrave:192,Aacute:193,Acirc:194,Atilde:195,Auml:196,Aring:197,AElig:198,Ccedil:199,Egrave:200,Eacute:201,Ecirc:202,Euml:203,Igrave:204,Iacute:205,Icirc:206,Iuml:207,ETH:208,Ntilde:209,Ograve:210,Oacute:211,Ocirc:212,Otilde:213,Ouml:214,times:215,Oslash:216,Ugrave:217,Uacute:218,Ucirc:219,Uuml:220,Yacute:221,THORN:222,szlig:223,agrave:224,aacute:225,acirc:226,atilde:227,auml:228,aring:229,aelig:230,ccedil:231,egrave:232,eacute:233,ecirc:234,euml:235,igrave:236,iacute:237,icirc:238,iuml:239,eth:240,ntilde:241,ograve:242,oacute:243,ocirc:244,otilde:245,ouml:246,divide:247,oslash:248,ugrave:249,uacute:250,ucirc:251,uuml:252,yacute:253,thorn:254,yuml:255,OElig:338,oelig:339,Scaron:352,scaron:353,Yuml:376,fnof:402,circ:710,tilde:732,Alpha:913,Beta:914,Gamma:915,Delta:916,Epsilon:917,Zeta:918,Eta:919,Theta:920,Iota:921,Kappa:922,Lambda:923,Mu:924,Nu:925,Xi:926,Omicron:927,Pi:928,Rho:929,Sigma:931,Tau:932,Upsilon:933,Phi:934,Chi:935,Psi:936,Omega:937,alpha:945,beta:946,gamma:947,delta:948,epsilon:949,zeta:950,eta:951,theta:952,iota:953,kappa:954,lambda:955,mu:956,nu:957,xi:958,omicron:959,pi:960,rho:961,sigmaf:962,sigma:963,tau:964,upsilon:965,phi:966,chi:967,psi:968,omega:969,thetasym:977,upsih:978,piv:982,ensp:8194,emsp:8195,thinsp:8201,zwnj:8204,zwj:8205,lrm:8206,rlm:8207,ndash:8211,mdash:8212,lsquo:8216,rsquo:8217,sbquo:8218,ldquo:8220,rdquo:8221,bdquo:8222,dagger:8224,Dagger:8225,bull:8226,hellip:8230,permil:8240,prime:8242,Prime:8243,lsaquo:8249,rsaquo:8250,oline:8254,frasl:8260,euro:8364,image:8465,weierp:8472,real:8476,trade:8482,alefsym:8501,larr:8592,uarr:8593,rarr:8594,darr:8595,harr:8596,crarr:8629,lArr:8656,uArr:8657,rArr:8658,dArr:8659,hArr:8660,forall:8704,part:8706,exist:8707,empty:8709,nabla:8711,isin:8712,notin:8713,ni:8715,prod:8719,sum:8721,minus:8722,lowast:8727,radic:8730,prop:8733,infin:8734,ang:8736,and:8743,or:8744,cap:8745,cup:8746,int:8747,there4:8756,sim:8764,cong:8773,asymp:8776,ne:8800,equiv:8801,le:8804,ge:8805,sub:8834,sup:8835,nsub:8836,sube:8838,supe:8839,oplus:8853,otimes:8855,perp:8869,sdot:8901,lceil:8968,rceil:8969,lfloor:8970,rfloor:8971,lang:9001,rang:9002,loz:9674,spades:9824,clubs:9827,hearts:9829,diams:9830}))}(ts||(ts={})),function(e){e.transformES2016=function(n){var t=n.factory,r=n.hoistVariableDeclaration;return e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;return e.visitEachChild(t,a,n)}));function a(i){if(0==(128&i.transformFlags))return i;switch(i.kind){case 216:return function(i){switch(i.operatorToken.kind){case 66:return function(n){var i,o,s=e.visitNode(n.left,a,e.isExpression),c=e.visitNode(n.right,a,e.isExpression);if(e.isElementAccessExpression(s)){var l=t.createTempVariable(r),u=t.createTempVariable(r);i=e.setTextRange(t.createElementAccessExpression(e.setTextRange(t.createAssignment(l,s.expression),s.expression),e.setTextRange(t.createAssignment(u,s.argumentExpression),s.argumentExpression)),s),o=e.setTextRange(t.createElementAccessExpression(l,u),s)}else if(e.isPropertyAccessExpression(s)){l=t.createTempVariable(r);i=e.setTextRange(t.createPropertyAccessExpression(e.setTextRange(t.createAssignment(l,s.expression),s.expression),s.name),s),o=e.setTextRange(t.createPropertyAccessExpression(l,s.name),s)}else i=s,o=s;return e.setTextRange(t.createAssignment(i,e.setTextRange(t.createGlobalMethodCall("Math","pow",[o,c]),n)),n)}(i);case 42:return function(n){var r=e.visitNode(n.left,a,e.isExpression),i=e.visitNode(n.right,a,e.isExpression);return e.setTextRange(t.createGlobalMethodCall("Math","pow",[r,i]),n)}(i);default:return e.visitEachChild(i,a,n)}}(i);default:return e.visitEachChild(i,a,n)}}}}(ts||(ts={})),function(e){var n,t,r,a,i;!function(e){e[e.CapturedThis=1]="CapturedThis",e[e.BlockScopedBindings=2]="BlockScopedBindings"}(n||(n={})),function(e){e[e.Body=1]="Body",e[e.Initializer=2]="Initializer"}(t||(t={})),function(e){e[e.ToOriginal=0]="ToOriginal",e[e.ToOutParameter=1]="ToOutParameter"}(r||(r={})),function(e){e[e.Break=2]="Break",e[e.Continue=4]="Continue",e[e.Return=8]="Return"}(a||(a={})),function(e){e[e.None=0]="None",e[e.Function=1]="Function",e[e.ArrowFunction=2]="ArrowFunction",e[e.AsyncFunctionBody=4]="AsyncFunctionBody",e[e.NonStaticClassElement=8]="NonStaticClassElement",e[e.CapturesThis=16]="CapturesThis",e[e.ExportedVariableStatement=32]="ExportedVariableStatement",e[e.TopLevel=64]="TopLevel",e[e.Block=128]="Block",e[e.IterationStatement=256]="IterationStatement",e[e.IterationStatementBlock=512]="IterationStatementBlock",e[e.IterationContainer=1024]="IterationContainer",e[e.ForStatement=2048]="ForStatement",e[e.ForInOrForOfStatement=4096]="ForInOrForOfStatement",e[e.ConstructorWithCapturedSuper=8192]="ConstructorWithCapturedSuper",e[e.AncestorFactsMask=16383]="AncestorFactsMask",e[e.BlockScopeIncludes=0]="BlockScopeIncludes",e[e.BlockScopeExcludes=7104]="BlockScopeExcludes",e[e.SourceFileIncludes=64]="SourceFileIncludes",e[e.SourceFileExcludes=8064]="SourceFileExcludes",e[e.FunctionIncludes=65]="FunctionIncludes",e[e.FunctionExcludes=16286]="FunctionExcludes",e[e.AsyncFunctionBodyIncludes=69]="AsyncFunctionBodyIncludes",e[e.AsyncFunctionBodyExcludes=16278]="AsyncFunctionBodyExcludes",e[e.ArrowFunctionIncludes=66]="ArrowFunctionIncludes",e[e.ArrowFunctionExcludes=15232]="ArrowFunctionExcludes",e[e.ConstructorIncludes=73]="ConstructorIncludes",e[e.ConstructorExcludes=16278]="ConstructorExcludes",e[e.DoOrWhileStatementIncludes=1280]="DoOrWhileStatementIncludes",e[e.DoOrWhileStatementExcludes=0]="DoOrWhileStatementExcludes",e[e.ForStatementIncludes=3328]="ForStatementIncludes",e[e.ForStatementExcludes=5056]="ForStatementExcludes",e[e.ForInOrForOfStatementIncludes=5376]="ForInOrForOfStatementIncludes",e[e.ForInOrForOfStatementExcludes=3008]="ForInOrForOfStatementExcludes",e[e.BlockIncludes=128]="BlockIncludes",e[e.BlockExcludes=6976]="BlockExcludes",e[e.IterationStatementBlockIncludes=512]="IterationStatementBlockIncludes",e[e.IterationStatementBlockExcludes=7104]="IterationStatementBlockExcludes",e[e.NewTarget=16384]="NewTarget",e[e.CapturedLexicalThis=32768]="CapturedLexicalThis",e[e.SubtreeFactsMask=-16384]="SubtreeFactsMask",e[e.ArrowFunctionSubtreeExcludes=0]="ArrowFunctionSubtreeExcludes",e[e.FunctionSubtreeExcludes=49152]="FunctionSubtreeExcludes"}(i||(i={})),e.transformES2015=function(n){var t,r,a,i,o,s,c=n.factory,l=n.getEmitHelperFactory,u=n.startLexicalEnvironment,d=n.resumeLexicalEnvironment,p=n.endLexicalEnvironment,m=n.hoistVariableDeclaration,f=n.getCompilerOptions(),_=n.getEmitResolver(),g=n.onSubstituteNode,y=n.onEmitNode;function v(n){i=e.append(i,c.createVariableDeclaration(n))}return n.onEmitNode=function(n,t,r){if(1&s&&e.isFunctionLike(t)){var a=h(16286,8&e.getEmitFlags(t)?81:65);return y(n,t,r),void b(a,0,0)}y(n,t,r)},n.onSubstituteNode=function(n,t){if(t=g(n,t),1===n)return function(n){switch(n.kind){case 78:return function(n){if(2&s&&!e.isInternalName(n)){var t=_.getReferencedDeclarationWithCollidingName(n);if(t&&(!e.isClassLike(t)||!function(n,t){var r=e.getParseTreeNode(t);if(!r||r===n||r.end<=n.pos||r.pos>=n.end)return!1;var a=e.getEnclosingBlockScopeContainer(n);for(;r;){if(r===a||r===n)return!1;if(e.isClassElement(r)&&r.parent===n)return!0;r=r.parent}return!1}(t,n)))return e.setTextRange(c.getGeneratedNameForNode(e.getNameOfDeclaration(t)),n)}return n}(n);case 107:return function(n){if(1&s&&16&a)return e.setTextRange(c.createUniqueName("_this",48),n);return n}(n)}return n}(t);if(e.isIdentifier(t))return function(n){if(2&s&&!e.isInternalName(n)){var t=e.getParseTreeNode(n,e.isIdentifier);if(t&&function(e){switch(e.parent.kind){case 198:case 252:case 255:case 249:return e.parent.name===e&&_.isDeclarationWithCollidingName(e.parent)}return!1}(t))return e.setTextRange(c.getGeneratedNameForNode(t),n)}return n}(t);return t},e.chainBundle(n,(function(o){if(o.isDeclarationFile)return o;t=o,r=o.text;var s=function(n){var t=h(8064,64),r=[],a=[];u();var o=c.copyPrologue(n.statements,r,!1,S);e.addRange(a,e.visitNodes(n.statements,S,e.isStatement,o)),i&&a.push(c.createVariableStatement(void 0,c.createVariableDeclarationList(i)));return c.mergeLexicalEnvironment(r,p()),G(r,n),b(t,0,0),c.updateSourceFile(n,e.setTextRange(c.createNodeArray(e.concatenate(r,a)),n.statements))}(o);return e.addEmitHelpers(s,n.readEmitHelpers()),t=void 0,r=void 0,i=void 0,a=0,s}));function h(e,n){var t=a;return a=16383&(a&~e|n),t}function b(e,n,t){a=-16384&(a&~n|t)|e}function E(e){return 0!=(8192&a)&&242===e.kind&&!e.expression}function T(n){return 0!=(256&n.transformFlags)||void 0!==o||8192&a&&function(n){return 1048576&n.transformFlags&&(e.isReturnStatement(n)||e.isIfStatement(n)||e.isWithStatement(n)||e.isSwitchStatement(n)||e.isCaseBlock(n)||e.isCaseClause(n)||e.isDefaultClause(n)||e.isTryStatement(n)||e.isCatchClause(n)||e.isLabeledStatement(n)||e.isIterationStatement(n,!1)||e.isBlock(n))}(n)||e.isIterationStatement(n,!1)&&ue(n)||0!=(33554432&e.getEmitFlags(n))}function S(e){return T(e)?A(e,!1):e}function x(e){return T(e)?A(e,!0):e}function L(e){return 105===e.kind?Re(!0):S(e)}function A(r,i){switch(r.kind){case 123:return;case 252:return function(n){var t=c.createVariableDeclaration(c.getLocalName(n,!0),void 0,void 0,I(n));e.setOriginalNode(t,n);var r=[],a=c.createVariableStatement(void 0,c.createVariableDeclarationList([t]));if(e.setOriginalNode(a,n),e.setTextRange(a,n),e.startOnNewLine(a),r.push(a),e.hasSyntacticModifier(n,1)){var i=e.hasSyntacticModifier(n,512)?c.createExportDefault(c.getLocalName(n)):c.createExternalModuleExport(c.getLocalName(n));e.setOriginalNode(i,a),r.push(i)}var o=e.getEmitFlags(n);0==(4194304&o)&&(r.push(c.createEndOfDeclarationMarker(n)),e.setEmitFlags(a,4194304|o));return e.singleOrMany(r)}(r);case 221:return function(e){return I(e)}(r);case 160:return function(n){return n.dotDotDotToken?void 0:e.isBindingPattern(n.name)?e.setOriginalNode(e.setTextRange(c.createParameterDeclaration(void 0,void 0,void 0,c.getGeneratedNameForNode(n),void 0,void 0,void 0),n),n):n.initializer?e.setOriginalNode(e.setTextRange(c.createParameterDeclaration(void 0,void 0,void 0,n.name,void 0,void 0,void 0),n),n):n}(r);case 251:return function(t){var r=o;o=void 0;var i=h(16286,65),s=e.visitParameterList(t.parameters,S,n),l=z(t),u=16384&a?c.getLocalName(t):t.name;return b(i,49152,0),o=r,c.updateFunctionDeclaration(t,void 0,e.visitNodes(t.modifiers,S,e.isModifier),t.asteriskToken,u,void 0,s,void 0,l)}(r);case 209:return function(t){4096&t.transformFlags&&(a|=32768);var r=o;o=void 0;var i=h(15232,66),s=c.createFunctionExpression(void 0,void 0,void 0,void 0,e.visitParameterList(t.parameters,S,n),void 0,z(t));e.setTextRange(s,t),e.setOriginalNode(s,t),e.setEmitFlags(s,8),32768&a&&Oe();return b(i,0,0),o=r,s}(r);case 208:return function(t){var r=262144&e.getEmitFlags(t)?h(16278,69):h(16286,65),i=o;o=void 0;var s=e.visitParameterList(t.parameters,S,n),l=z(t),u=16384&a?c.getLocalName(t):t.name;return b(r,49152,0),o=i,c.updateFunctionExpression(t,void 0,t.asteriskToken,u,void 0,s,void 0,l)}(r);case 249:return X(r);case 78:return C(r);case 250:return function(t){if(3&t.flags||131072&t.transformFlags){3&t.flags&&ke();var r=e.flatMap(t.declarations,1&t.flags?J:X),a=c.createVariableDeclarationList(r);return e.setOriginalNode(a,t),e.setTextRange(a,t),e.setCommentRange(a,t),131072&t.transformFlags&&(e.isBindingPattern(t.declarations[0].name)||e.isBindingPattern(e.last(t.declarations).name))&&e.setSourceMapRange(a,function(n){for(var t=-1,r=-1,a=0,i=n;a<i.length;a++){var o=i[a];t=-1===t?o.pos:-1===o.pos?t:Math.min(t,o.pos),r=Math.max(r,o.end)}return e.createRange(t,r)}(r)),a}return e.visitEachChild(t,S,n)}(r);case 244:return function(t){if(void 0!==o){var r=o.allowedNonLabeledJumps;o.allowedNonLabeledJumps|=2;var a=e.visitEachChild(t,S,n);return o.allowedNonLabeledJumps=r,a}return e.visitEachChild(t,S,n)}(r);case 258:return function(t){var r=h(7104,0),a=e.visitEachChild(t,S,n);return b(r,0,0),a}(r);case 230:return function(t,r){if(r)return e.visitEachChild(t,S,n);var i=256&a?h(7104,512):h(6976,128),o=e.visitEachChild(t,S,n);return b(i,0,0),o}(r,!1);case 241:case 240:return function(t){if(o){var r=241===t.kind?2:4;if(!(t.label&&o.labels&&o.labels.get(e.idText(t.label))||!t.label&&o.allowedNonLabeledJumps&r)){var a=void 0,i=t.label;i?241===t.kind?(a="break-"+i.escapedText,ye(o,!0,e.idText(i),a)):(a="continue-"+i.escapedText,ye(o,!1,e.idText(i),a)):241===t.kind?(o.nonLocalJumps|=2,a="break"):(o.nonLocalJumps|=4,a="continue");var s=c.createStringLiteral(a);if(o.loopOutParameters.length){for(var l=o.loopOutParameters,u=void 0,d=0;d<l.length;d++){var p=_e(l[d],1);u=0===d?p:c.createBinaryExpression(u,27,p)}s=c.createBinaryExpression(u,27,s)}return c.createReturnStatement(s)}}return e.visitEachChild(t,S,n)}(r);case 245:return function(n){o&&!o.labels&&(o.labels=new e.Map);var t=e.unwrapInnermostStatementOfLabel(n,o&&Y);return e.isIterationStatement(t,!1)?function(e,n){switch(e.kind){case 235:case 236:return $(e,n);case 237:return ee(e,n);case 238:return ne(e,n);case 239:return te(e,n)}}(t,n):c.restoreEnclosingLabel(e.visitNode(t,S,e.isStatement,c.liftToBlock),n,o&&Q)}(r);case 235:case 236:return $(r,void 0);case 237:return ee(r,void 0);case 238:return ne(r,void 0);case 239:return te(r,void 0);case 233:return function(t){return e.visitEachChild(t,x,n)}(r);case 200:return function(t){for(var r=t.properties,i=-1,o=!1,s=0;s<r.length;s++){var l=r[s];if(262144&l.transformFlags&&4&a||(o=158===e.Debug.checkDefined(l.name).kind)){i=s;break}}if(i<0)return e.visitEachChild(t,S,n);var u=c.createTempVariable(m),d=[],p=c.createAssignment(u,e.setEmitFlags(c.createObjectLiteralExpression(e.visitNodes(r,S,e.isObjectLiteralElementLike,0,i),t.multiLine),o?65536:0));t.multiLine&&e.startOnNewLine(p);return d.push(p),function(n,t,r,a){for(var i=t.properties,o=i.length,s=a;s<o;s++){var c=i[s];switch(c.kind){case 167:case 168:var l=e.getAllAccessorDeclarations(t.properties,c);c===l.firstAccessor&&n.push(K(r,l,t,!!t.multiLine));break;case 165:n.push(Te(c,r,t,t.multiLine));break;case 288:n.push(be(c,r,t.multiLine));break;case 289:n.push(Ee(c,r,t.multiLine));break;default:e.Debug.failBadSyntaxKind(t)}}}(d,t,u,i),d.push(t.multiLine?e.startOnNewLine(e.setParent(e.setTextRange(c.cloneNode(u),u),u.parent)):u),c.inlineExpressions(d)}(r);case 287:return function(t){var r,a=h(7104,0);if(e.Debug.assert(!!t.variableDeclaration,"Catch clause variable should always be present when downleveling ES2015."),e.isBindingPattern(t.variableDeclaration.name)){var i=c.createTempVariable(void 0),o=c.createVariableDeclaration(i);e.setTextRange(o,t.variableDeclaration);var s=e.flattenDestructuringBinding(t.variableDeclaration,S,n,0,i),l=c.createVariableDeclarationList(s);e.setTextRange(l,t.variableDeclaration);var u=c.createVariableStatement(void 0,l);r=c.updateCatchClause(t,o,(d=t.block,p=u,m=e.visitNodes(d.statements,S,e.isStatement),c.updateBlock(d,__spreadArrays([p],m))))}else r=e.visitEachChild(t,S,n);var d,p,m;return b(a,0,0),r}(r);case 289:return function(n){return e.setTextRange(c.createPropertyAssignment(n.name,C(c.cloneNode(n.name))),n)}(r);case 158:return function(t){return e.visitEachChild(t,S,n)}(r);case 199:return function(t){if(e.some(t.elements,e.isSpreadElement))return xe(t.elements,!0,!!t.multiLine,!!t.elements.hasTrailingComma);return e.visitEachChild(t,S,n)}(r);case 203:return function(n){if(33554432&e.getEmitFlags(n))return function(n){var t=e.cast(e.cast(e.skipOuterExpressions(n.expression),e.isArrowFunction).body,e.isBlock),r=function(n){return e.isVariableStatement(n)&&!!e.first(n.declarationList.declarations).initializer},a=o;o=void 0;var i=e.visitNodes(t.statements,S,e.isStatement);o=a;var s=e.filter(i,r),l=e.filter(i,(function(e){return!r(e)})),u=e.cast(e.first(s),e.isVariableStatement).declarationList.declarations[0],d=e.skipOuterExpressions(u.initializer),p=e.tryCast(d,e.isAssignmentExpression),m=e.cast(p?e.skipOuterExpressions(p.right):d,e.isCallExpression),f=e.cast(e.skipOuterExpressions(m.expression),e.isFunctionExpression),_=f.body.statements,g=0,y=-1,v=[];if(p){var h=e.tryCast(_[g],e.isExpressionStatement);h&&(v.push(h),g++),v.push(_[g]),g++,v.push(c.createExpressionStatement(c.createAssignment(p.left,e.cast(u.name,e.isIdentifier))))}for(;!e.isReturnStatement(e.elementAt(_,y));)y--;e.addRange(v,_,g,y),y<-1&&e.addRange(v,_,y+1);return e.addRange(v,l),e.addRange(v,s,1),c.restoreOuterExpressions(n.expression,c.restoreOuterExpressions(u.initializer,c.restoreOuterExpressions(p&&p.right,c.updateCallExpression(m,c.restoreOuterExpressions(m.expression,c.updateFunctionExpression(f,void 0,void 0,void 0,void 0,f.parameters,void 0,c.updateBlock(f.body,v))),void 0,m.arguments))))}(n);var t=e.skipOuterExpressions(n.expression);if(105===t.kind||e.isSuperProperty(t)||e.some(n.arguments,e.isSpreadElement))return Se(n,!0);return c.updateCallExpression(n,e.visitNode(n.expression,L,e.isExpression),void 0,e.visitNodes(n.arguments,S,e.isExpression))}(r);case 204:return function(t){if(e.some(t.arguments,e.isSpreadElement)){var r=c.createCallBinding(c.createPropertyAccessExpression(t.expression,"bind"),m),a=r.target,i=r.thisArg;return c.createNewExpression(c.createFunctionApplyCall(e.visitNode(a,S,e.isExpression),i,xe(c.createNodeArray(__spreadArrays([c.createVoidZero()],t.arguments)),!1,!1,!1)),void 0,[])}return e.visitEachChild(t,S,n)}(r);case 207:return function(t,r){return e.visitEachChild(t,r?x:S,n)}(r,i);case 216:return q(r,i);case 337:return function(t,r){if(r)return e.visitEachChild(t,x,n);for(var a,i=0;i<t.elements.length;i++){var o=t.elements[i],s=e.visitNode(o,i<t.elements.length-1?x:S,e.isExpression);(a||s!==o)&&(a||(a=t.elements.slice(0,i)),a.push(s))}var l=a?e.setTextRange(c.createNodeArray(a),t.elements):t.elements;return c.updateCommaListExpression(t,l)}(r,i);case 14:case 15:case 16:case 17:return function(n){return e.setTextRange(c.createStringLiteral(n.text),n)}(r);case 10:return function(n){if(n.hasExtendedUnicodeEscape)return e.setTextRange(c.createStringLiteral(n.text),n);return n}(r);case 8:return function(n){if(384&n.numericLiteralFlags)return e.setTextRange(c.createNumericLiteral(n.text),n);return n}(r);case 205:return function(r){return e.processTaggedTemplateExpression(n,r,S,t,v,e.ProcessLevel.All)}(r);case 218:return function(n){var t=[];(function(n,t){if(!function(n){return e.Debug.assert(0!==n.templateSpans.length),0!==n.head.text.length||0===n.templateSpans[0].literal.text.length}(t))return;n.push(c.createStringLiteral(t.head.text))})(t,n),function(n,t){for(var r=0,a=t.templateSpans;r<a.length;r++){var i=a[r];n.push(e.visitNode(i.expression,S,e.isExpression)),0!==i.literal.text.length&&n.push(c.createStringLiteral(i.literal.text))}}(t,n);var r=e.reduceLeft(t,c.createAdd);e.nodeIsSynthesized(r)&&e.setTextRange(r,n);return r}(r);case 219:return function(t){return e.visitEachChild(t,S,n)}(r);case 220:return function(n){return e.visitNode(n.expression,S,e.isExpression)}(r);case 105:return Re(!1);case 107:return function(e){2&a&&(a|=32768);if(o)return 2&a?(o.containsLexicalThis=!0,e):o.thisName||(o.thisName=c.createUniqueName("this"));return e}(r);case 226:return function(e){if(102===e.keywordToken&&"target"===e.name.escapedText)return a|=16384,c.createUniqueName("_newTarget",48);return e}(r);case 165:return function(n){e.Debug.assert(!e.isComputedPropertyName(n.name));var t=W(n,e.moveRangePos(n,-1),void 0,void 0);return e.setEmitFlags(t,512|e.getEmitFlags(t)),e.setTextRange(c.createPropertyAssignment(n.name,t),n)}(r);case 167:case 168:return function(t){e.Debug.assert(!e.isComputedPropertyName(t.name));var r=o;o=void 0;var a,i=h(16286,65),s=e.visitParameterList(t.parameters,S,n),l=z(t);a=167===t.kind?c.updateGetAccessorDeclaration(t,t.decorators,t.modifiers,t.name,s,t.type,l):c.updateSetAccessorDeclaration(t,t.decorators,t.modifiers,t.name,s,l);return b(i,49152,0),o=r,a}(r);case 232:return function(t){var r,a=h(0,e.hasSyntacticModifier(t,1)?32:0);if(o&&0==(3&t.declarationList.flags)&&!function(n){return 1===n.declarationList.declarations.length&&!!n.declarationList.declarations[0].initializer&&!!(33554432&e.getEmitFlags(n.declarationList.declarations[0].initializer))}(t)){for(var i=void 0,s=0,l=t.declarationList.declarations;s<l.length;s++){var u=l[s];if(pe(o,u),u.initializer){var d=void 0;e.isBindingPattern(u.name)?d=e.flattenDestructuringAssignment(u,S,n,0):(d=c.createBinaryExpression(u.name,62,e.visitNode(u.initializer,S,e.isExpression)),e.setTextRange(d,u)),i=e.append(i,d)}}r=i?e.setTextRange(c.createExpressionStatement(c.inlineExpressions(i)),t):void 0}else r=e.visitEachChild(t,S,n);return b(a,0,0),r}(r);case 242:return function(t){if(o)return o.nonLocalJumps|=8,E(t)&&(t=D(t)),c.createReturnStatement(c.createObjectLiteralExpression([c.createPropertyAssignment(c.createIdentifier("value"),t.expression?e.visitNode(t.expression,S,e.isExpression):c.createVoidZero())]));if(E(t))return D(t);return e.visitEachChild(t,S,n)}(r);case 212:return function(t){return e.visitEachChild(t,x,n)}(r);default:return e.visitEachChild(r,S,n)}}function D(n){return e.setOriginalNode(c.createReturnStatement(c.createUniqueName("_this",48)),n)}function C(e){return o&&_.isArgumentsLocalBinding(e)?o.argumentsName||(o.argumentsName=c.createUniqueName("arguments")):e}function I(i){i.name&&ke();var s=e.getClassExtendsHeritageElement(i),m=c.createFunctionExpression(void 0,void 0,void 0,void 0,s?[c.createParameterDeclaration(void 0,void 0,void 0,c.createUniqueName("_super",48))]:[],void 0,function(i,s){var m=[],f=c.getInternalName(i),_=e.isIdentifierANonContextualKeyword(f)?c.getGeneratedNameForNode(f):f;u(),function(n,t,r){r&&n.push(e.setTextRange(c.createExpressionStatement(l().createExtendsHelper(c.getInternalName(t))),r))}(m,i,s),function(t,r,i,s){var l=o;o=void 0;var u=h(16278,73),m=e.getFirstConstructorWithBody(r),f=function(n,t){if(!n||!t)return!1;if(e.some(n.parameters))return!1;var r=e.firstOrUndefined(n.body.statements);if(!r||!e.nodeIsSynthesized(r)||233!==r.kind)return!1;var a=r.expression;if(!e.nodeIsSynthesized(a)||203!==a.kind)return!1;var i=a.expression;if(!e.nodeIsSynthesized(i)||105!==i.kind)return!1;var o=e.singleOrUndefined(a.arguments);if(!o||!e.nodeIsSynthesized(o)||220!==o.kind)return!1;var s=o.expression;return e.isIdentifier(s)&&"arguments"===s.escapedText}(m,void 0!==s),_=c.createFunctionDeclaration(void 0,void 0,void 0,i,void 0,function(t,r){return e.visitParameterList(t&&!r?t.parameters:void 0,S,n)||[]}(m,f),void 0,function(n,t,r,i){var o=!!r&&103!==e.skipOuterExpressions(r.expression).kind;if(!n)return function(n,t){var r=[];d(),c.mergeLexicalEnvironment(r,p()),t&&r.push(c.createReturnStatement(k()));var a=c.createNodeArray(r);e.setTextRange(a,n.members);var i=c.createBlock(a,!0);return e.setTextRange(i,n),e.setEmitFlags(i,1536),i}(t,o);var s=[],l=[];d();var u,m=0;i||(m=c.copyStandardPrologue(n.body.statements,s,!1));P(l,n),F(l,n,i),i||(m=c.copyCustomPrologue(n.body.statements,l,m,S));if(i)u=k();else if(o&&m<n.body.statements.length){var f=n.body.statements[m];e.isExpressionStatement(f)&&e.isSuperCall(f.expression)&&(u=function(e){return Se(e,!1)}(f.expression))}u&&(a|=8192,m++);if(e.addRange(l,e.visitNodes(n.body.statements,S,e.isStatement,m)),c.mergeLexicalEnvironment(s,p()),U(s,n,!1),o)if(!u||m!==n.body.statements.length||4096&n.body.transformFlags)B(l,n,u||R()),N(n.body)||l.push(c.createReturnStatement(c.createUniqueName("_this",48)));else{var _=e.cast(e.cast(u,e.isBinaryExpression).left,e.isCallExpression),g=c.createReturnStatement(u);e.setCommentRange(g,e.getCommentRange(_)),e.setEmitFlags(_,1536),l.push(g)}else G(s,n);var y=c.createBlock(e.setTextRange(c.createNodeArray(e.concatenate(s,l)),n.body.statements),!0);return e.setTextRange(y,n.body),y}(m,r,s,f));e.setTextRange(_,m||r),s&&e.setEmitFlags(_,8);t.push(_),b(u,49152,0),o=l}(m,i,_,s),function(n,r){for(var a=0,i=r.members;a<i.length;a++){var o=i[a];switch(o.kind){case 229:n.push(V(o));break;case 165:n.push(j(Pe(r,o),o,r));break;case 167:case 168:var s=e.getAllAccessorDeclarations(r.members,o);o===s.firstAccessor&&n.push(H(Pe(r,o),s,r));break;case 166:break;default:e.Debug.failBadSyntaxKind(o,t&&t.fileName)}}}(m,i);var g=e.createTokenRange(e.skipTrivia(r,i.members.end),19),y=c.createPartiallyEmittedExpression(_);e.setTextRangeEnd(y,g.end),e.setEmitFlags(y,1536);var v=c.createReturnStatement(y);e.setTextRangePos(v,g.pos),e.setEmitFlags(v,1920),m.push(v),e.insertStatementsAfterStandardPrologue(m,p());var E=c.createBlock(e.setTextRange(c.createNodeArray(m),i.members),!0);return e.setEmitFlags(E,1536),E}(i,s));e.setEmitFlags(m,65536&e.getEmitFlags(i)|524288);var f=c.createPartiallyEmittedExpression(m);e.setTextRangeEnd(f,i.end),e.setEmitFlags(f,1536);var _=c.createPartiallyEmittedExpression(f);e.setTextRangeEnd(_,e.skipTrivia(r,i.pos)),e.setEmitFlags(_,1536);var g=c.createParenthesizedExpression(c.createCallExpression(_,void 0,s?[e.visitNode(s.expression,S,e.isExpression)]:[]));return e.addSyntheticLeadingComment(g,3,"* @class "),g}function N(n){if(242===n.kind)return!0;if(234===n.kind){var t=n;if(t.elseStatement)return N(t.thenStatement)&&N(t.elseStatement)}else if(230===n.kind){var r=e.lastOrUndefined(n.statements);if(r&&N(r))return!0}return!1}function R(){return e.setEmitFlags(c.createThis(),4)}function k(){return c.createLogicalOr(c.createLogicalAnd(c.createStrictInequality(c.createUniqueName("_super",48),c.createNull()),c.createFunctionApplyCall(c.createUniqueName("_super",48),R(),c.createIdentifier("arguments"))),R())}function O(n){return void 0!==n.initializer||e.isBindingPattern(n.name)}function P(n,t){if(!e.some(t.parameters,O))return!1;for(var r=!1,a=0,i=t.parameters;a<i.length;a++){var o=i[a],s=o.name,c=o.initializer;o.dotDotDotToken||(e.isBindingPattern(s)?r=M(n,o,s,c)||r:c&&(w(n,o,s,c),r=!0))}return r}function M(t,r,a,i){return a.elements.length>0?(e.insertStatementAfterCustomPrologue(t,e.setEmitFlags(c.createVariableStatement(void 0,c.createVariableDeclarationList(e.flattenDestructuringBinding(r,S,n,0,c.getGeneratedNameForNode(r)))),1048576)),!0):!!i&&(e.insertStatementAfterCustomPrologue(t,e.setEmitFlags(c.createExpressionStatement(c.createAssignment(c.getGeneratedNameForNode(r),e.visitNode(i,S,e.isExpression))),1048576)),!0)}function w(n,t,r,a){a=e.visitNode(a,S,e.isExpression);var i=c.createIfStatement(c.createTypeCheck(c.cloneNode(r),"undefined"),e.setEmitFlags(e.setTextRange(c.createBlock([c.createExpressionStatement(e.setEmitFlags(e.setTextRange(c.createAssignment(e.setEmitFlags(e.setParent(e.setTextRange(c.cloneNode(r),r),r.parent),48),e.setEmitFlags(a,1584|e.getEmitFlags(a))),t),1536))]),t),1953));e.startOnNewLine(i),e.setTextRange(i,t),e.setEmitFlags(i,1050528),e.insertStatementAfterCustomPrologue(n,i)}function F(t,r,a){var i=[],o=e.lastOrUndefined(r.parameters);if(!function(e,n){return!(!e||!e.dotDotDotToken||n)}(o,a))return!1;var s=78===o.name.kind?e.setParent(e.setTextRange(c.cloneNode(o.name),o.name),o.name.parent):c.createTempVariable(void 0);e.setEmitFlags(s,48);var l=78===o.name.kind?c.cloneNode(o.name):s,u=r.parameters.length-1,d=c.createLoopVariable();i.push(e.setEmitFlags(e.setTextRange(c.createVariableStatement(void 0,c.createVariableDeclarationList([c.createVariableDeclaration(s,void 0,void 0,c.createArrayLiteralExpression([]))])),o),1048576));var p=c.createForStatement(e.setTextRange(c.createVariableDeclarationList([c.createVariableDeclaration(d,void 0,void 0,c.createNumericLiteral(u))]),o),e.setTextRange(c.createLessThan(d,c.createPropertyAccessExpression(c.createIdentifier("arguments"),"length")),o),e.setTextRange(c.createPostfixIncrement(d),o),c.createBlock([e.startOnNewLine(e.setTextRange(c.createExpressionStatement(c.createAssignment(c.createElementAccessExpression(l,0===u?d:c.createSubtract(d,c.createNumericLiteral(u))),c.createElementAccessExpression(c.createIdentifier("arguments"),d))),o))]));return e.setEmitFlags(p,1048576),e.startOnNewLine(p),i.push(p),78!==o.name.kind&&i.push(e.setEmitFlags(e.setTextRange(c.createVariableStatement(void 0,c.createVariableDeclarationList(e.flattenDestructuringBinding(o,S,n,0,l))),o),1048576)),e.insertStatementsAfterCustomPrologue(t,i),!0}function G(e,n){return!!(32768&a&&209!==n.kind)&&(B(e,n,c.createThis()),!0)}function B(n,t,r){Oe();var a=c.createVariableStatement(void 0,c.createVariableDeclarationList([c.createVariableDeclaration(c.createUniqueName("_this",48),void 0,void 0,r)]));e.setEmitFlags(a,1050112),e.setSourceMapRange(a,t),e.insertStatementAfterCustomPrologue(n,a)}function U(n,t,r){if(16384&a){var i=void 0;switch(t.kind){case 209:return n;case 165:case 167:case 168:i=c.createVoidZero();break;case 166:i=c.createPropertyAccessExpression(e.setEmitFlags(c.createThis(),4),"constructor");break;case 251:case 208:i=c.createConditionalExpression(c.createLogicalAnd(e.setEmitFlags(c.createThis(),4),c.createBinaryExpression(e.setEmitFlags(c.createThis(),4),101,c.getLocalName(t))),void 0,c.createPropertyAccessExpression(e.setEmitFlags(c.createThis(),4),"constructor"),void 0,c.createVoidZero());break;default:return e.Debug.failBadSyntaxKind(t)}var o=c.createVariableStatement(void 0,c.createVariableDeclarationList([c.createVariableDeclaration(c.createUniqueName("_newTarget",48),void 0,void 0,i)]));e.setEmitFlags(o,1050112),r&&(n=n.slice()),e.insertStatementAfterCustomPrologue(n,o)}return n}function V(n){return e.setTextRange(c.createEmptyStatement(),n)}function j(t,r,a){var i,o=e.getCommentRange(r),s=e.getSourceMapRange(r),l=W(r,r,void 0,a),u=e.visitNode(r.name,S,e.isPropertyName);if(!e.isPrivateIdentifier(u)&&n.getCompilerOptions().useDefineForClassFields){var d=e.isComputedPropertyName(u)?u.expression:e.isIdentifier(u)?c.createStringLiteral(e.unescapeLeadingUnderscores(u.escapedText)):u;i=c.createObjectDefinePropertyCall(t,d,c.createPropertyDescriptor({value:l,enumerable:!1,writable:!0,configurable:!0}))}else{var p=e.createMemberAccessForPropertyName(c,t,u,r.name);i=c.createAssignment(p,l)}e.setEmitFlags(l,1536),e.setSourceMapRange(l,s);var m=e.setTextRange(c.createExpressionStatement(i),r);return e.setOriginalNode(m,r),e.setCommentRange(m,o),e.setEmitFlags(m,48),m}function H(n,t,r){var a=c.createExpressionStatement(K(n,t,r,!1));return e.setEmitFlags(a,1536),e.setSourceMapRange(a,e.getSourceMapRange(t.firstAccessor)),a}function K(n,t,r,a){var i=t.firstAccessor,o=t.getAccessor,s=t.setAccessor,l=e.setParent(e.setTextRange(c.cloneNode(n),n),n.parent);e.setEmitFlags(l,1568),e.setSourceMapRange(l,i.name);var u=e.visitNode(i.name,S,e.isPropertyName);if(e.isPrivateIdentifier(u))return e.Debug.failBadSyntaxKind(u,"Encountered unhandled private identifier while transforming ES2015.");var d=e.createExpressionForPropertyName(c,u);e.setEmitFlags(d,1552),e.setSourceMapRange(d,i.name);var p=[];if(o){var m=W(o,void 0,void 0,r);e.setSourceMapRange(m,e.getSourceMapRange(o)),e.setEmitFlags(m,512);var f=c.createPropertyAssignment("get",m);e.setCommentRange(f,e.getCommentRange(o)),p.push(f)}if(s){var _=W(s,void 0,void 0,r);e.setSourceMapRange(_,e.getSourceMapRange(s)),e.setEmitFlags(_,512);var g=c.createPropertyAssignment("set",_);e.setCommentRange(g,e.getCommentRange(s)),p.push(g)}p.push(c.createPropertyAssignment("enumerable",o||s?c.createFalse():c.createTrue()),c.createPropertyAssignment("configurable",c.createTrue()));var y=c.createCallExpression(c.createPropertyAccessExpression(c.createIdentifier("Object"),"defineProperty"),void 0,[l,d,c.createObjectLiteralExpression(p,!0)]);return a&&e.startOnNewLine(y),y}function W(t,r,i,s){var l=o;o=void 0;var u=s&&e.isClassLike(s)&&!e.hasSyntacticModifier(t,32)?h(16286,73):h(16286,65),d=e.visitParameterList(t.parameters,S,n),p=z(t);return 16384&a&&!i&&(251===t.kind||208===t.kind)&&(i=c.getGeneratedNameForNode(t)),b(u,49152,0),o=l,e.setOriginalNode(e.setTextRange(c.createFunctionExpression(void 0,t.asteriskToken,i,void 0,d,void 0,p),r),t)}function z(n){var r,a,i,o=!1,s=!1,l=[],u=[],m=n.body;if(d(),e.isBlock(m)&&(i=c.copyStandardPrologue(m.statements,l,!1),i=c.copyCustomPrologue(m.statements,u,i,S,e.isHoistedFunction),i=c.copyCustomPrologue(m.statements,u,i,S,e.isHoistedVariableStatement)),o=P(u,n)||o,o=F(u,n,!1)||o,e.isBlock(m))i=c.copyCustomPrologue(m.statements,u,i,S),r=m.statements,e.addRange(u,e.visitNodes(m.statements,S,e.isStatement,i)),!o&&m.multiLine&&(o=!0);else{e.Debug.assert(209===n.kind),r=e.moveRangeEnd(m,-1);var f=n.equalsGreaterThanToken;e.nodeIsSynthesized(f)||e.nodeIsSynthesized(m)||(e.rangeEndIsOnSameLineAsRangeStart(f,m,t)?s=!0:o=!0);var _=e.visitNode(m,S,e.isExpression),g=c.createReturnStatement(_);e.setTextRange(g,m),e.moveSyntheticComments(g,m),e.setEmitFlags(g,1440),u.push(g),a=m}if(c.mergeLexicalEnvironment(l,p()),U(l,n,!1),G(l,n),e.some(l)&&(o=!0),u.unshift.apply(u,l),e.isBlock(m)&&e.arrayIsEqualTo(u,m.statements))return m;var y=c.createBlock(e.setTextRange(c.createNodeArray(u),r),o);return e.setTextRange(y,n.body),!o&&s&&e.setEmitFlags(y,1),a&&e.setTokenSourceMapRange(y,19,a),e.setOriginalNode(y,n.body),y}function q(t,r){return e.isDestructuringAssignment(t)?e.flattenDestructuringAssignment(t,S,n,0,!r):27===t.operatorToken.kind?c.updateBinaryExpression(t,e.visitNode(t.left,x,e.isExpression),t.operatorToken,e.visitNode(t.right,r?x:S,e.isExpression)):e.visitEachChild(t,S,n)}function J(t){var r=t.name;return e.isBindingPattern(r)?X(t):!t.initializer&&function(e){var n=_.getNodeCheckFlags(e),t=262144&n,r=524288&n;return!(0!=(64&a)||t&&r&&0!=(512&a))&&0==(4096&a)&&(!_.isDeclarationWithCollidingName(e)||r&&!t&&0==(6144&a))}(t)?c.updateVariableDeclaration(t,t.name,void 0,void 0,c.createVoidZero()):e.visitEachChild(t,S,n)}function X(t){var r,a=h(32,0);return r=e.isBindingPattern(t.name)?e.flattenDestructuringBinding(t,S,n,0,void 0,0!=(32&a)):e.visitEachChild(t,S,n),b(a,0,0),r}function Y(n){o.labels.set(e.idText(n.label),!0)}function Q(n){o.labels.set(e.idText(n.label),!1)}function Z(t,r,i,s,l){var d=h(t,r),m=function(t,r,i,s){if(!ue(t)){var l=void 0;o&&(l=o.allowedNonLabeledJumps,o.allowedNonLabeledJumps=6);var d=s?s(t,r,void 0,i):c.restoreEnclosingLabel(e.isForStatement(t)?function(n){return c.updateForStatement(n,e.visitNode(n.initializer,x,e.isForInitializer),e.visitNode(n.condition,S,e.isExpression),e.visitNode(n.incrementor,x,e.isExpression),e.visitNode(n.statement,S,e.isStatement,c.liftToBlock))}(t):e.visitEachChild(t,S,n),r,o&&Q);return o&&(o.allowedNonLabeledJumps=l),d}var m=function(n){var t;switch(n.kind){case 237:case 238:case 239:var r=n.initializer;r&&250===r.kind&&(t=r)}var a=[],i=[];if(t&&3&e.getCombinedNodeFlags(t))for(var s=ce(n),c=0,l=t.declarations;c<l.length;c++){he(n,l[c],a,i,s)}var u={loopParameters:a,loopOutParameters:i};o&&(o.argumentsName&&(u.argumentsName=o.argumentsName),o.thisName&&(u.thisName=o.thisName),o.hoistedLocalVariables&&(u.hoistedLocalVariables=o.hoistedLocalVariables));return u}(t),f=[],_=o;o=m;var g,y=ce(t)?function(n,t){var r=c.createUniqueName("_loop_init"),i=0!=(262144&n.initializer.transformFlags),o=0;t.containsLexicalThis&&(o|=8);i&&4&a&&(o|=262144);var s=[];s.push(c.createVariableStatement(void 0,n.initializer)),ge(t.loopOutParameters,2,1,s);var l=c.createVariableStatement(void 0,e.setEmitFlags(c.createVariableDeclarationList([c.createVariableDeclaration(r,void 0,void 0,e.setEmitFlags(c.createFunctionExpression(void 0,i?c.createToken(41):void 0,void 0,void 0,void 0,void 0,e.visitNode(c.createBlock(s,!0),S,e.isBlock)),o))]),2097152)),u=c.createVariableDeclarationList(e.map(t.loopOutParameters,fe));return{functionName:r,containsYield:i,functionDeclaration:l,part:u}}(t,m):void 0,v=de(t)?function(n,t,r){var i=c.createUniqueName("_loop");u();var o=e.visitNode(n.statement,S,e.isStatement,c.liftToBlock),s=p(),l=[];(le(n)||function(n){return e.isForStatement(n)&&!!n.incrementor&&se(n.incrementor)}(n))&&(t.conditionVariable=c.createUniqueName("inc"),n.incrementor?l.push(c.createIfStatement(t.conditionVariable,c.createExpressionStatement(e.visitNode(n.incrementor,S,e.isExpression)),c.createExpressionStatement(c.createAssignment(t.conditionVariable,c.createTrue())))):l.push(c.createIfStatement(c.createLogicalNot(t.conditionVariable),c.createExpressionStatement(c.createAssignment(t.conditionVariable,c.createTrue())))),le(n)&&l.push(c.createIfStatement(c.createPrefixUnaryExpression(53,e.visitNode(n.condition,S,e.isExpression)),e.visitNode(c.createBreakStatement(),S,e.isStatement))));e.isBlock(o)?e.addRange(l,o.statements):l.push(o);ge(t.loopOutParameters,1,1,l),e.insertStatementsAfterStandardPrologue(l,s);var d=c.createBlock(l,!0);e.isBlock(o)&&e.setOriginalNode(d,o);var m=0!=(262144&n.statement.transformFlags),f=0;t.containsLexicalThis&&(f|=8);m&&0!=(4&a)&&(f|=262144);var _=c.createVariableStatement(void 0,e.setEmitFlags(c.createVariableDeclarationList([c.createVariableDeclaration(i,void 0,void 0,e.setEmitFlags(c.createFunctionExpression(void 0,m?c.createToken(41):void 0,void 0,void 0,t.loopParameters,void 0,d),f))]),2097152)),g=function(n,t,r,a){var i=[],o=!(-5&t.nonLocalJumps||t.labeledNonLocalBreaks||t.labeledNonLocalContinues),s=c.createCallExpression(n,void 0,e.map(t.loopParameters,(function(e){return e.name}))),l=a?c.createYieldExpression(c.createToken(41),e.setEmitFlags(s,8388608)):s;if(o)i.push(c.createExpressionStatement(l)),ge(t.loopOutParameters,1,0,i);else{var u=c.createUniqueName("state"),d=c.createVariableStatement(void 0,c.createVariableDeclarationList([c.createVariableDeclaration(u,void 0,void 0,l)]));if(i.push(d),ge(t.loopOutParameters,1,0,i),8&t.nonLocalJumps){var p=void 0;r?(r.nonLocalJumps|=8,p=c.createReturnStatement(u)):p=c.createReturnStatement(c.createPropertyAccessExpression(u,"value")),i.push(c.createIfStatement(c.createTypeCheck(u,"object"),p))}if(2&t.nonLocalJumps&&i.push(c.createIfStatement(c.createStrictEquality(u,c.createStringLiteral("break")),c.createBreakStatement())),t.labeledNonLocalBreaks||t.labeledNonLocalContinues){var m=[];ve(t.labeledNonLocalBreaks,!0,u,r,m),ve(t.labeledNonLocalContinues,!1,u,r,m),i.push(c.createSwitchStatement(u,c.createCaseBlock(m)))}}return i}(i,t,r,m);return{functionName:i,containsYield:m,functionDeclaration:_,part:g}}(t,m,_):void 0;o=_,y&&f.push(y.functionDeclaration);v&&f.push(v.functionDeclaration);(function(e,n,t){var r;n.argumentsName&&(t?t.argumentsName=n.argumentsName:(r||(r=[])).push(c.createVariableDeclaration(n.argumentsName,void 0,void 0,c.createIdentifier("arguments"))));n.thisName&&(t?t.thisName=n.thisName:(r||(r=[])).push(c.createVariableDeclaration(n.thisName,void 0,void 0,c.createIdentifier("this"))));if(n.hoistedLocalVariables)if(t)t.hoistedLocalVariables=n.hoistedLocalVariables;else{r||(r=[]);for(var a=0,i=n.hoistedLocalVariables;a<i.length;a++){var o=i[a];r.push(c.createVariableDeclaration(o))}}if(n.loopOutParameters.length){r||(r=[]);for(var s=0,l=n.loopOutParameters;s<l.length;s++){var u=l[s];r.push(c.createVariableDeclaration(u.outParamName))}}n.conditionVariable&&(r||(r=[]),r.push(c.createVariableDeclaration(n.conditionVariable,void 0,void 0,c.createFalse())));r&&e.push(c.createVariableStatement(void 0,c.createVariableDeclarationList(r)))})(f,m,_),y&&f.push((h=y.functionName,b=y.containsYield,E=c.createCallExpression(h,void 0,[]),T=b?c.createYieldExpression(c.createToken(41),e.setEmitFlags(E,8388608)):E,c.createExpressionStatement(T)));var h,b,E,T;if(v)if(s)g=s(t,r,v.part,i);else{var L=me(t,y,c.createBlock(v.part,!0));g=c.restoreEnclosingLabel(L,r,o&&Q)}else{var A=me(t,y,e.visitNode(t.statement,S,e.isStatement,c.liftToBlock));g=c.restoreEnclosingLabel(A,r,o&&Q)}return f.push(g),f}(i,s,d,l);return b(d,0,0),m}function $(e,n){return Z(0,1280,e,n)}function ee(e,n){return Z(5056,3328,e,n)}function ne(e,n){return Z(3008,5376,e,n)}function te(e,n){return Z(3008,5376,e,n,f.downlevelIteration?oe:ie)}function re(t,r,a){var i=[],o=t.initializer;if(e.isVariableDeclarationList(o)){3&t.initializer.flags&&ke();var s=e.firstOrUndefined(o.declarations);if(s&&e.isBindingPattern(s.name)){var l=e.flattenDestructuringBinding(s,S,n,0,r),u=e.setTextRange(c.createVariableDeclarationList(l),t.initializer);e.setOriginalNode(u,t.initializer),e.setSourceMapRange(u,e.createRange(l[0].pos,e.last(l).end)),i.push(c.createVariableStatement(void 0,u))}else i.push(e.setTextRange(c.createVariableStatement(void 0,e.setOriginalNode(e.setTextRange(c.createVariableDeclarationList([c.createVariableDeclaration(s?s.name:c.createTempVariable(void 0),void 0,void 0,r)]),e.moveRangePos(o,-1)),o)),e.moveRangeEnd(o,-1)))}else{var d=c.createAssignment(o,r);e.isDestructuringAssignment(d)?i.push(c.createExpressionStatement(q(d,!0))):(e.setTextRangeEnd(d,o.end),i.push(e.setTextRange(c.createExpressionStatement(e.visitNode(d,S,e.isExpression)),e.moveRangeEnd(o,-1))))}if(a)return ae(e.addRange(i,a));var p=e.visitNode(t.statement,S,e.isStatement,c.liftToBlock);return e.isBlock(p)?c.updateBlock(p,e.setTextRange(c.createNodeArray(e.concatenate(i,p.statements)),p.statements)):(i.push(p),ae(i))}function ae(n){return e.setEmitFlags(c.createBlock(c.createNodeArray(n),!0),432)}function ie(n,t,r){var a=e.visitNode(n.expression,S,e.isExpression),i=c.createLoopVariable(),s=e.isIdentifier(a)?c.getGeneratedNameForNode(a):c.createTempVariable(void 0);e.setEmitFlags(a,48|e.getEmitFlags(a));var l=e.setTextRange(c.createForStatement(e.setEmitFlags(e.setTextRange(c.createVariableDeclarationList([e.setTextRange(c.createVariableDeclaration(i,void 0,void 0,c.createNumericLiteral(0)),e.moveRangePos(n.expression,-1)),e.setTextRange(c.createVariableDeclaration(s,void 0,void 0,a),n.expression)]),n.expression),2097152),e.setTextRange(c.createLessThan(i,c.createPropertyAccessExpression(s,"length")),n.expression),e.setTextRange(c.createPostfixIncrement(i),n.expression),re(n,c.createElementAccessExpression(s,i),r)),n);return e.setEmitFlags(l,256),e.setTextRange(l,n),c.restoreEnclosingLabel(l,t,o&&Q)}function oe(n,t,r,a){var i=e.visitNode(n.expression,S,e.isExpression),s=e.isIdentifier(i)?c.getGeneratedNameForNode(i):c.createTempVariable(void 0),u=e.isIdentifier(i)?c.getGeneratedNameForNode(s):c.createTempVariable(void 0),d=c.createUniqueName("e"),p=c.getGeneratedNameForNode(d),f=c.createTempVariable(void 0),_=e.setTextRange(l().createValuesHelper(i),n.expression),g=c.createCallExpression(c.createPropertyAccessExpression(s,"next"),void 0,[]);m(d),m(f);var y=1024&a?c.inlineExpressions([c.createAssignment(d,c.createVoidZero()),_]):_,v=e.setEmitFlags(e.setTextRange(c.createForStatement(e.setEmitFlags(e.setTextRange(c.createVariableDeclarationList([e.setTextRange(c.createVariableDeclaration(s,void 0,void 0,y),n.expression),c.createVariableDeclaration(u,void 0,void 0,g)]),n.expression),2097152),c.createLogicalNot(c.createPropertyAccessExpression(u,"done")),c.createAssignment(u,g),re(n,c.createPropertyAccessExpression(u,"value"),r)),n),256);return c.createTryStatement(c.createBlock([c.restoreEnclosingLabel(v,t,o&&Q)]),c.createCatchClause(c.createVariableDeclaration(p),e.setEmitFlags(c.createBlock([c.createExpressionStatement(c.createAssignment(d,c.createObjectLiteralExpression([c.createPropertyAssignment("error",p)])))]),1)),c.createBlock([c.createTryStatement(c.createBlock([e.setEmitFlags(c.createIfStatement(c.createLogicalAnd(c.createLogicalAnd(u,c.createLogicalNot(c.createPropertyAccessExpression(u,"done"))),c.createAssignment(f,c.createPropertyAccessExpression(s,"return"))),c.createExpressionStatement(c.createFunctionCallCall(f,s,[]))),1)]),void 0,e.setEmitFlags(c.createBlock([e.setEmitFlags(c.createIfStatement(d,c.createThrowStatement(c.createPropertyAccessExpression(d,"error"))),1)]),1))]))}function se(e){return 0!=(131072&_.getNodeCheckFlags(e))}function ce(n){return e.isForStatement(n)&&!!n.initializer&&se(n.initializer)}function le(n){return e.isForStatement(n)&&!!n.condition&&se(n.condition)}function ue(e){return de(e)||ce(e)}function de(e){return 0!=(65536&_.getNodeCheckFlags(e))}function pe(n,t){n.hoistedLocalVariables||(n.hoistedLocalVariables=[]),function t(r){if(78===r.kind)n.hoistedLocalVariables.push(r);else for(var a=0,i=r.elements;a<i.length;a++){var o=i[a];e.isOmittedExpression(o)||t(o.name)}}(t.name)}function me(n,t,r){switch(n.kind){case 237:return function(n,t,r){var a=n.condition&&se(n.condition),i=a||n.incrementor&&se(n.incrementor);return c.updateForStatement(n,e.visitNode(t?t.part:n.initializer,x,e.isForInitializer),e.visitNode(a?void 0:n.condition,S,e.isExpression),e.visitNode(i?void 0:n.incrementor,x,e.isExpression),r)}(n,t,r);case 238:return function(n,t){return c.updateForInStatement(n,e.visitNode(n.initializer,S,e.isForInitializer),e.visitNode(n.expression,S,e.isExpression),t)}(n,r);case 239:return function(n,t){return c.updateForOfStatement(n,void 0,e.visitNode(n.initializer,S,e.isForInitializer),e.visitNode(n.expression,S,e.isExpression),t)}(n,r);case 235:return function(n,t){return c.updateDoStatement(n,t,e.visitNode(n.expression,S,e.isExpression))}(n,r);case 236:return function(n,t){return c.updateWhileStatement(n,e.visitNode(n.expression,S,e.isExpression),t)}(n,r);default:return e.Debug.failBadSyntaxKind(n,"IterationStatement expected")}}function fe(e){return c.createVariableDeclaration(e.originalName,void 0,void 0,e.outParamName)}function _e(e,n){var t=0===n?e.outParamName:e.originalName,r=0===n?e.originalName:e.outParamName;return c.createBinaryExpression(r,62,t)}function ge(e,n,t,r){for(var a=0,i=e;a<i.length;a++){var o=i[a];o.flags&n&&r.push(c.createExpressionStatement(_e(o,t)))}}function ye(n,t,r,a){t?(n.labeledNonLocalBreaks||(n.labeledNonLocalBreaks=new e.Map),n.labeledNonLocalBreaks.set(r,a)):(n.labeledNonLocalContinues||(n.labeledNonLocalContinues=new e.Map),n.labeledNonLocalContinues.set(r,a))}function ve(e,n,t,r,a){e&&e.forEach((function(e,i){var o=[];if(!r||r.labels&&r.labels.get(i)){var s=c.createIdentifier(i);o.push(n?c.createBreakStatement(s):c.createContinueStatement(s))}else ye(r,n,i,e),o.push(c.createReturnStatement(t));a.push(c.createCaseClause(c.createStringLiteral(e),o))}))}function he(n,t,r,a,i){var o=t.name;if(e.isBindingPattern(o))for(var s=0,l=o.elements;s<l.length;s++){var u=l[s];e.isOmittedExpression(u)||he(n,u,r,a,i)}else{r.push(c.createParameterDeclaration(void 0,void 0,void 0,o));var d=_.getNodeCheckFlags(t);if(4194304&d||i){var p=c.createUniqueName("out_"+e.idText(o)),m=0;4194304&d&&(m|=1),e.isForStatement(n)&&n.initializer&&_.isBindingCapturedByNode(n.initializer,t)&&(m|=2),a.push({flags:m,originalName:o,outParamName:p})}}}function be(n,t,r){var a=c.createAssignment(e.createMemberAccessForPropertyName(c,t,e.visitNode(n.name,S,e.isPropertyName)),e.visitNode(n.initializer,S,e.isExpression));return e.setTextRange(a,n),r&&e.startOnNewLine(a),a}function Ee(n,t,r){var a=c.createAssignment(e.createMemberAccessForPropertyName(c,t,e.visitNode(n.name,S,e.isPropertyName)),c.cloneNode(n.name));return e.setTextRange(a,n),r&&e.startOnNewLine(a),a}function Te(n,t,r,a){var i=c.createAssignment(e.createMemberAccessForPropertyName(c,t,e.visitNode(n.name,S,e.isPropertyName)),W(n,n,void 0,r));return e.setTextRange(i,n),a&&e.startOnNewLine(i),i}function Se(t,r){if(8192&t.transformFlags||105===t.expression.kind||e.isSuperProperty(e.skipOuterExpressions(t.expression))){var a=c.createCallBinding(t.expression,m),i=a.target,o=a.thisArg;105===t.expression.kind&&e.setEmitFlags(o,4);var s=void 0;if(s=8192&t.transformFlags?c.createFunctionApplyCall(e.visitNode(i,L,e.isExpression),105===t.expression.kind?o:e.visitNode(o,S,e.isExpression),xe(t.arguments,!1,!1,!1)):e.setTextRange(c.createFunctionCallCall(e.visitNode(i,L,e.isExpression),105===t.expression.kind?o:e.visitNode(o,S,e.isExpression),e.visitNodes(t.arguments,S,e.isExpression)),t),105===t.expression.kind){var l=c.createLogicalOr(s,R());s=r?c.createAssignment(c.createUniqueName("_this",48),l):l}return e.setOriginalNode(s,t)}return e.visitEachChild(t,S,n)}function xe(n,t,r,a){var i,o=n.length,s=e.flatten(e.spanMap(n,De,(function(e,n,t,i){return n(e,r,a&&i===o)})));if(f.downlevelIteration){if(1===s.length)if(Ae(c=s[0],"___spread"))return s[0];return l().createSpreadHelper(s)}if(1===s.length){var c=s[0];if(!t||(i=c,e.isArrayLiteralExpression(i)&&e.every(i.elements,Le))||Ae(c,"___spreadArrays"))return s[0]}return l().createSpreadArraysHelper(s)}function Le(n){return!e.isOmittedExpression(n)}function Ae(n,t){return e.isCallExpression(n)&&e.isIdentifier(n.expression)&&4096&e.getEmitFlags(n.expression)&&n.expression.escapedText===t}function De(n){return e.isSpreadElement(n)?Ce:Ie}function Ce(n){return e.map(n,Ne)}function Ie(n,t,r){return c.createArrayLiteralExpression(e.visitNodes(c.createNodeArray(n,r),S,e.isExpression),t)}function Ne(n){return e.visitNode(n.expression,S,e.isExpression)}function Re(e){return 8&a&&!e?c.createPropertyAccessExpression(c.createUniqueName("_super",48),"prototype"):c.createUniqueName("_super",48)}function ke(){0==(2&s)&&(s|=2,n.enableSubstitution(78))}function Oe(){0==(1&s)&&(s|=1,n.enableSubstitution(107),n.enableEmitNotification(166),n.enableEmitNotification(165),n.enableEmitNotification(167),n.enableEmitNotification(168),n.enableEmitNotification(209),n.enableEmitNotification(208),n.enableEmitNotification(251))}function Pe(n,t){return e.hasSyntacticModifier(t,32)?c.getInternalName(n):c.createPropertyAccessExpression(c.getInternalName(n),"prototype")}}}(ts||(ts={})),function(e){e.transformES5=function(n){var t,r,a=n.factory,i=n.getCompilerOptions();1!==i.jsx&&3!==i.jsx||(t=n.onEmitNode,n.onEmitNode=function(n,a,i){switch(a.kind){case 275:case 276:case 274:var o=a.tagName;r[e.getOriginalNodeId(o)]=!0}t(n,a,i)},n.enableEmitNotification(275),n.enableEmitNotification(276),n.enableEmitNotification(274),r=[]);var o=n.onSubstituteNode;return n.onSubstituteNode=function(n,t){if(t.id&&r&&r[t.id])return o(n,t);if(t=o(n,t),e.isPropertyAccessExpression(t))return function(n){if(e.isPrivateIdentifier(n.name))return n;var t=s(n.name);if(t)return e.setTextRange(a.createElementAccessExpression(n.expression,t),n);return n}(t);if(e.isPropertyAssignment(t))return function(n){var t=e.isIdentifier(n.name)&&s(n.name);if(t)return a.updatePropertyAssignment(n,t,n.initializer);return n}(t);return t},n.enableSubstitution(201),n.enableSubstitution(288),e.chainBundle(n,(function(e){return e}));function s(n){var t=n.originalKeywordKind||(e.nodeIsSynthesized(n)?e.stringToToken(e.idText(n)):void 0);if(void 0!==t&&t>=80&&t<=115)return e.setTextRange(a.createStringLiteralFromNode(n),n)}}}(ts||(ts={})),function(e){var n,t,r,a,i;!function(e){e[e.Nop=0]="Nop",e[e.Statement=1]="Statement",e[e.Assign=2]="Assign",e[e.Break=3]="Break",e[e.BreakWhenTrue=4]="BreakWhenTrue",e[e.BreakWhenFalse=5]="BreakWhenFalse",e[e.Yield=6]="Yield",e[e.YieldStar=7]="YieldStar",e[e.Return=8]="Return",e[e.Throw=9]="Throw",e[e.Endfinally=10]="Endfinally"}(n||(n={})),function(e){e[e.Open=0]="Open",e[e.Close=1]="Close"}(t||(t={})),function(e){e[e.Exception=0]="Exception",e[e.With=1]="With",e[e.Switch=2]="Switch",e[e.Loop=3]="Loop",e[e.Labeled=4]="Labeled"}(r||(r={})),function(e){e[e.Try=0]="Try",e[e.Catch=1]="Catch",e[e.Finally=2]="Finally",e[e.Done=3]="Done"}(a||(a={})),function(e){e[e.Next=0]="Next",e[e.Throw=1]="Throw",e[e.Return=2]="Return",e[e.Break=3]="Break",e[e.Yield=4]="Yield",e[e.YieldStar=5]="YieldStar",e[e.Catch=6]="Catch",e[e.Endfinally=7]="Endfinally"}(i||(i={})),e.transformGenerators=function(n){var t,r,a,i,o,s,c,l,u,d,p=n.factory,m=n.getEmitHelperFactory,f=n.resumeLexicalEnvironment,_=n.endLexicalEnvironment,g=n.hoistFunctionDeclaration,y=n.hoistVariableDeclaration,v=n.getCompilerOptions(),h=e.getEmitScriptTarget(v),b=n.getEmitResolver(),E=n.onSubstituteNode;n.onSubstituteNode=function(n,a){if(a=E(n,a),1===n)return function(n){if(e.isIdentifier(n))return function(n){if(!e.isGeneratedIdentifier(n)&&t&&t.has(e.idText(n))){var a=e.getOriginalNode(n);if(e.isIdentifier(a)&&a.parent){var i=b.getReferencedValueDeclaration(a);if(i){var o=r[e.getOriginalNodeId(i)];if(o){var s=e.setParent(e.setTextRange(p.cloneNode(o),o),o.parent);return e.setSourceMapRange(s,n),e.setCommentRange(s,n),s}}}}return n}(n);return n}(a);return a};var T,S,x,L,A,D,C,I,N,R,k,O,P=1,M=0,w=0;return e.chainBundle(n,(function(t){if(t.isDeclarationFile||0==(512&t.transformFlags))return t;var r=e.visitEachChild(t,F,n);return e.addEmitHelpers(r,n.readEmitHelpers()),r}));function F(t){var r=t.transformFlags;return i?function(t){switch(t.kind){case 235:case 236:return function(t){return i?(ie(),t=e.visitEachChild(t,F,n),se(),t):e.visitEachChild(t,F,n)}(t);case 244:return function(t){i&&ne({kind:2,isScript:!0,breakLabel:-1});t=e.visitEachChild(t,F,n),i&&ce();return t}(t);case 245:return function(t){i&&ne({kind:4,isScript:!0,labelText:e.idText(t.label),breakLabel:-1});t=e.visitEachChild(t,F,n),i&&le();return t}(t);default:return G(t)}}(t):a?G(t):e.isFunctionLikeDeclaration(t)&&t.asteriskToken?function(n){switch(n.kind){case 251:return B(n);case 208:return U(n);default:return e.Debug.failBadSyntaxKind(n)}}(t):512&r?e.visitEachChild(t,F,n):t}function G(t){switch(t.kind){case 251:return B(t);case 208:return U(t);case 167:case 168:return function(t){var r=a,o=i;return a=!1,i=!1,t=e.visitEachChild(t,F,n),a=r,i=o,t}(t);case 232:return function(n){if(262144&n.transformFlags)return void q(n.declarationList);if(1048576&e.getEmitFlags(n))return n;for(var t=0,r=n.declarationList.declarations;t<r.length;t++){var a=r[t];y(a.name)}var i=e.getInitializedVariables(n.declarationList);if(0===i.length)return;return e.setSourceMapRange(p.createExpressionStatement(p.inlineExpressions(e.map(i,J))),n)}(t);case 237:return function(t){i&&ie();var r=t.initializer;if(r&&e.isVariableDeclarationList(r)){for(var a=0,o=r.declarations;a<o.length;a++){var s=o[a];y(s.name)}var c=e.getInitializedVariables(r);t=p.updateForStatement(t,c.length>0?p.inlineExpressions(e.map(c,J)):void 0,e.visitNode(t.condition,F,e.isExpression),e.visitNode(t.incrementor,F,e.isExpression),e.visitNode(t.statement,F,e.isStatement,p.liftToBlock))}else t=e.visitEachChild(t,F,n);i&&se();return t}(t);case 238:return function(t){i&&ie();var r=t.initializer;if(e.isVariableDeclarationList(r)){for(var a=0,o=r.declarations;a<o.length;a++){var s=o[a];y(s.name)}t=p.updateForInStatement(t,r.declarations[0].name,e.visitNode(t.expression,F,e.isExpression),e.visitNode(t.statement,F,e.isStatement,p.liftToBlock))}else t=e.visitEachChild(t,F,n);i&&se();return t}(t);case 241:return function(t){if(i){var r=fe(t.label&&e.idText(t.label));if(r>0)return ve(r,t)}return e.visitEachChild(t,F,n)}(t);case 240:return function(t){if(i){var r=_e(t.label&&e.idText(t.label));if(r>0)return ve(r,t)}return e.visitEachChild(t,F,n)}(t);case 242:return function(n){return t=e.visitNode(n.expression,F,e.isExpression),r=n,e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression(t?[ye(2),t]:[ye(2)])),r);var t,r}(t);default:return 262144&t.transformFlags?function(t){switch(t.kind){case 216:return function(t){var r=e.getExpressionAssociativity(t);switch(r){case 0:return function(t){if(X(t.right))return e.isLogicalOperator(t.operatorToken.kind)?function(n){var t=$(),r=Z();Ee(r,e.visitNode(n.left,F,e.isExpression),n.left),55===n.operatorToken.kind?xe(t,r,n.left):Se(t,r,n.left);return Ee(r,e.visitNode(n.right,F,e.isExpression),n.right),ee(t),r}(t):27===t.operatorToken.kind?j(t):p.updateBinaryExpression(t,Q(e.visitNode(t.left,F,e.isExpression)),t.operatorToken,e.visitNode(t.right,F,e.isExpression));return e.visitEachChild(t,F,n)}(t);case 1:return function(t){var r=t.left,a=t.right;if(X(a)){var i=void 0;switch(r.kind){case 201:i=p.updatePropertyAccessExpression(r,Q(e.visitNode(r.expression,F,e.isLeftHandSideExpression)),r.name);break;case 202:i=p.updateElementAccessExpression(r,Q(e.visitNode(r.expression,F,e.isLeftHandSideExpression)),Q(e.visitNode(r.argumentExpression,F,e.isExpression)));break;default:i=e.visitNode(r,F,e.isExpression)}var o=t.operatorToken.kind;return e.isCompoundAssignment(o)?e.setTextRange(p.createAssignment(i,e.setTextRange(p.createBinaryExpression(Q(i),e.getNonAssignmentOperatorForCompoundAssignment(o),e.visitNode(a,F,e.isExpression)),t)),t):p.updateBinaryExpression(t,i,t.operatorToken,e.visitNode(a,F,e.isExpression))}return e.visitEachChild(t,F,n)}(t);default:return e.Debug.assertNever(r)}}(t);case 337:return function(n){for(var t=[],r=0,a=n.elements;r<a.length;r++){var i=a[r];e.isBinaryExpression(i)&&27===i.operatorToken.kind?t.push(j(i)):(X(i)&&t.length>0&&(Le(1,[p.createExpressionStatement(p.inlineExpressions(t))]),t=[]),t.push(e.visitNode(i,F,e.isExpression)))}return p.inlineExpressions(t)}(t);case 217:return function(t){if(X(t.whenTrue)||X(t.whenFalse)){var r=$(),a=$(),i=Z();return xe(r,e.visitNode(t.condition,F,e.isExpression),t.condition),Ee(i,e.visitNode(t.whenTrue,F,e.isExpression),t.whenTrue),Te(a),ee(r),Ee(i,e.visitNode(t.whenFalse,F,e.isExpression),t.whenFalse),ee(a),i}return e.visitEachChild(t,F,n)}(t);case 219:return function(n){var t=$(),r=e.visitNode(n.expression,F,e.isExpression);if(n.asteriskToken){!function(e,n){Le(7,[e],n)}(0==(8388608&e.getEmitFlags(n.expression))?e.setTextRange(m().createValuesHelper(r),n):r,n)}else!function(e,n){Le(6,[e],n)}(r,n);return ee(t),function(n){return e.setTextRange(p.createCallExpression(p.createPropertyAccessExpression(L,"sent"),void 0,[]),n)}(n)}(t);case 199:return function(e){return H(e.elements,void 0,void 0,e.multiLine)}(t);case 200:return function(n){var t=n.properties,r=n.multiLine,a=Y(t),i=Z();Ee(i,p.createObjectLiteralExpression(e.visitNodes(t,F,e.isObjectLiteralElementLike,0,a),r));var o=e.reduceLeft(t,s,[],a);return o.push(r?e.startOnNewLine(e.setParent(e.setTextRange(p.cloneNode(i),i),i.parent)):i),p.inlineExpressions(o);function s(t,a){X(a)&&t.length>0&&(be(p.createExpressionStatement(p.inlineExpressions(t))),t=[]);var o=e.createExpressionForObjectLiteralElementLike(p,n,a,i),s=e.visitNode(o,F,e.isExpression);return s&&(r&&e.startOnNewLine(s),t.push(s)),t}}(t);case 202:return function(t){if(X(t.argumentExpression))return p.updateElementAccessExpression(t,Q(e.visitNode(t.expression,F,e.isLeftHandSideExpression)),e.visitNode(t.argumentExpression,F,e.isExpression));return e.visitEachChild(t,F,n)}(t);case 203:return function(t){if(!e.isImportCall(t)&&e.forEach(t.arguments,X)){var r=p.createCallBinding(t.expression,y,h,!0),a=r.target,i=r.thisArg;return e.setOriginalNode(e.setTextRange(p.createFunctionApplyCall(Q(e.visitNode(a,F,e.isLeftHandSideExpression)),i,H(t.arguments)),t),t)}return e.visitEachChild(t,F,n)}(t);case 204:return function(t){if(e.forEach(t.arguments,X)){var r=p.createCallBinding(p.createPropertyAccessExpression(t.expression,"bind"),y),a=r.target,i=r.thisArg;return e.setOriginalNode(e.setTextRange(p.createNewExpression(p.createFunctionApplyCall(Q(e.visitNode(a,F,e.isExpression)),i,H(t.arguments,p.createVoidZero())),void 0,[]),t),t)}return e.visitEachChild(t,F,n)}(t);default:return e.visitEachChild(t,F,n)}}(t):1049088&t.transformFlags?e.visitEachChild(t,F,n):t}}function B(t){if(t.asteriskToken)t=e.setOriginalNode(e.setTextRange(p.createFunctionDeclaration(void 0,t.modifiers,void 0,t.name,void 0,e.visitParameterList(t.parameters,F,n),void 0,V(t.body)),t),t);else{var r=a,o=i;a=!1,i=!1,t=e.visitEachChild(t,F,n),a=r,i=o}return a?void g(t):t}function U(t){if(t.asteriskToken)t=e.setOriginalNode(e.setTextRange(p.createFunctionExpression(void 0,void 0,t.name,void 0,e.visitParameterList(t.parameters,F,n),void 0,V(t.body)),t),t);else{var r=a,o=i;a=!1,i=!1,t=e.visitEachChild(t,F,n),a=r,i=o}return t}function V(n){var t=[],r=a,m=i,g=o,y=s,v=c,h=l,b=u,E=d,A=P,D=T,C=S,I=x,N=L;a=!0,i=!1,o=void 0,s=void 0,c=void 0,l=void 0,u=void 0,d=void 0,P=1,T=void 0,S=void 0,x=void 0,L=p.createTempVariable(void 0),f();var R=p.copyPrologue(n.statements,t,!1,F);K(n.statements,R);var k=Ae();return e.insertStatementsAfterStandardPrologue(t,_()),t.push(p.createReturnStatement(k)),a=r,i=m,o=g,s=y,c=v,l=h,u=b,d=E,P=A,T=D,S=C,x=I,L=N,e.setTextRange(p.createBlock(t,n.multiLine),n)}function j(n){var t=[];return r(n.left),r(n.right),p.inlineExpressions(t);function r(n){e.isBinaryExpression(n)&&27===n.operatorToken.kind?(r(n.left),r(n.right)):(X(n)&&t.length>0&&(Le(1,[p.createExpressionStatement(p.inlineExpressions(t))]),t=[]),t.push(e.visitNode(n,F,e.isExpression)))}}function H(n,t,r,a){var i,o=Y(n);if(o>0){i=Z();var s=e.visitNodes(n,F,e.isExpression,0,o);Ee(i,p.createArrayLiteralExpression(t?__spreadArrays([t],s):s)),t=void 0}var c=e.reduceLeft(n,(function(n,r){if(X(r)&&n.length>0){var o=void 0!==i;i||(i=Z()),Ee(i,o?p.createArrayConcatCall(i,[p.createArrayLiteralExpression(n,a)]):p.createArrayLiteralExpression(t?__spreadArrays([t],n):n,a)),t=void 0,n=[]}return n.push(e.visitNode(r,F,e.isExpression)),n}),[],o);return i?p.createArrayConcatCall(i,[p.createArrayLiteralExpression(c,a)]):e.setTextRange(p.createArrayLiteralExpression(t?__spreadArrays([t],c):c,a),r)}function K(e,n){void 0===n&&(n=0);for(var t=e.length,r=n;r<t;r++)z(e[r])}function W(n){e.isBlock(n)?K(n.statements):z(n)}function z(a){var o=i;i||(i=X(a)),function(a){switch(a.kind){case 230:return function(n){X(n)?K(n.statements):be(e.visitNode(n,F,e.isStatement))}(a);case 233:return function(n){be(e.visitNode(n,F,e.isStatement))}(a);case 234:return function(n){if(X(n))if(X(n.thenStatement)||X(n.elseStatement)){var t=$(),r=n.elseStatement?$():void 0;xe(n.elseStatement?r:t,e.visitNode(n.expression,F,e.isExpression),n.expression),W(n.thenStatement),n.elseStatement&&(Te(t),ee(r),W(n.elseStatement)),ee(t)}else be(e.visitNode(n,F,e.isStatement));else be(e.visitNode(n,F,e.isStatement))}(a);case 235:return function(n){if(X(n)){var t=$(),r=$();oe(t),ee(r),W(n.statement),ee(t),Se(r,e.visitNode(n.expression,F,e.isExpression)),se()}else be(e.visitNode(n,F,e.isStatement))}(a);case 236:return function(n){if(X(n)){var t=$(),r=oe(t);ee(t),xe(r,e.visitNode(n.expression,F,e.isExpression)),W(n.statement),Te(t),se()}else be(e.visitNode(n,F,e.isStatement))}(a);case 237:return function(n){if(X(n)){var t=$(),r=$(),a=oe(r);if(n.initializer){var i=n.initializer;e.isVariableDeclarationList(i)?q(i):be(e.setTextRange(p.createExpressionStatement(e.visitNode(i,F,e.isExpression)),i))}ee(t),n.condition&&xe(a,e.visitNode(n.condition,F,e.isExpression)),W(n.statement),ee(r),n.incrementor&&be(e.setTextRange(p.createExpressionStatement(e.visitNode(n.incrementor,F,e.isExpression)),n.incrementor)),Te(t),se()}else be(e.visitNode(n,F,e.isStatement))}(a);case 238:return function(n){if(X(n)){var t=Z(),r=Z(),a=p.createLoopVariable(),i=n.initializer;y(a),Ee(t,p.createArrayLiteralExpression()),be(p.createForInStatement(r,e.visitNode(n.expression,F,e.isExpression),p.createExpressionStatement(p.createCallExpression(p.createPropertyAccessExpression(t,"push"),void 0,[r])))),Ee(a,p.createNumericLiteral(0));var o=$(),s=$(),c=oe(s);ee(o),xe(c,p.createLessThan(a,p.createPropertyAccessExpression(t,"length")));var l=void 0;if(e.isVariableDeclarationList(i)){for(var u=0,d=i.declarations;u<d.length;u++){var m=d[u];y(m.name)}l=p.cloneNode(i.declarations[0].name)}else l=e.visitNode(i,F,e.isExpression),e.Debug.assert(e.isLeftHandSideExpression(l));Ee(l,p.createElementAccessExpression(t,a)),W(n.statement),ee(s),be(p.createExpressionStatement(p.createPostfixIncrement(a))),Te(o),se()}else be(e.visitNode(n,F,e.isStatement))}(a);case 240:return function(n){var t=_e(n.label?e.idText(n.label):void 0);t>0?Te(t,n):be(n)}(a);case 241:return function(n){var t=fe(n.label?e.idText(n.label):void 0);t>0?Te(t,n):be(n)}(a);case 242:return function(n){t=e.visitNode(n.expression,F,e.isExpression),r=n,Le(8,[t],r);var t,r}(a);case 243:return function(n){X(n)?(t=Q(e.visitNode(n.expression,F,e.isExpression)),r=$(),a=$(),ee(r),ne({kind:1,expression:t,startLabel:r,endLabel:a}),W(n.statement),e.Debug.assert(1===ae()),ee(te().endLabel)):be(e.visitNode(n,F,e.isStatement));var t,r,a}(a);case 244:return function(n){if(X(n.caseBlock)){for(var t=n.caseBlock,r=t.clauses.length,a=(ne({kind:2,isScript:!1,breakLabel:f=$()}),f),i=Q(e.visitNode(n.expression,F,e.isExpression)),o=[],s=-1,c=0;c<r;c++){var l=t.clauses[c];o.push($()),285===l.kind&&-1===s&&(s=c)}for(var u=0,d=[];u<r;){var m=0;for(c=u;c<r;c++){if(284===(l=t.clauses[c]).kind){if(X(l.expression)&&d.length>0)break;d.push(p.createCaseClause(e.visitNode(l.expression,F,e.isExpression),[ve(o[c],l.expression)]))}else m++}d.length&&(be(p.createSwitchStatement(i,p.createCaseBlock(d))),u+=d.length,d=[]),m>0&&(u+=m,m=0)}Te(s>=0?o[s]:a);for(c=0;c<r;c++)ee(o[c]),K(t.clauses[c].statements);ce()}else be(e.visitNode(n,F,e.isStatement));var f}(a);case 245:return function(n){X(n)?(t=e.idText(n.label),r=$(),ne({kind:4,isScript:!1,labelText:t,breakLabel:r}),W(n.statement),le()):be(e.visitNode(n,F,e.isStatement));var t,r}(a);case 246:return function(n){var t;r=e.visitNode(null!==(t=n.expression)&&void 0!==t?t:p.createVoidZero(),F,e.isExpression),a=n,Le(9,[r],a);var r,a}(a);case 247:return function(a){X(a)?(i=$(),o=$(),ee(i),ne({kind:0,state:0,startLabel:i,endLabel:o}),he(),W(a.tryBlock),a.catchClause&&(!function(a){var i;if(e.Debug.assert(0===ae()),e.isGeneratedIdentifier(a.name))i=a.name,y(a.name);else{var o=e.idText(a.name);i=Z(o),t||(t=new e.Map,r=[],n.enableSubstitution(78)),t.set(o,!0),r[e.getOriginalNodeId(a)]=i}var s=re();e.Debug.assert(s.state<1),Te(s.endLabel);var c=$();ee(c),s.state=1,s.catchVariable=i,s.catchLabel=c,Ee(i,p.createCallExpression(p.createPropertyAccessExpression(L,"sent"),void 0,[])),he()}(a.catchClause.variableDeclaration),W(a.catchClause.block)),a.finallyBlock&&(!function(){e.Debug.assert(0===ae());var n=re();e.Debug.assert(n.state<2),Te(n.endLabel);var t=$();ee(t),n.state=2,n.finallyLabel=t}(),W(a.finallyBlock)),function(){e.Debug.assert(0===ae());var n=te();n.state<2?Te(n.endLabel):Le(10);ee(n.endLabel),he(),n.state=3}()):be(e.visitEachChild(a,F,n));var i,o}(a);default:be(e.visitNode(a,F,e.isStatement))}}(a),i=o}function q(n){for(var t=0,r=n.declarations;t<r.length;t++){var a=r[t],i=p.cloneNode(a.name);e.setCommentRange(i,a.name),y(i)}for(var o=e.getInitializedVariables(n),s=o.length,c=0,l=[];c<s;){for(var u=c;u<s;u++){if(X((a=o[u]).initializer)&&l.length>0)break;l.push(J(a))}l.length&&(be(p.createExpressionStatement(p.inlineExpressions(l))),c+=l.length,l=[])}}function J(n){return e.setSourceMapRange(p.createAssignment(e.setSourceMapRange(p.cloneNode(n.name),n.name),e.visitNode(n.initializer,F,e.isExpression)),n)}function X(e){return!!e&&0!=(262144&e.transformFlags)}function Y(e){for(var n=e.length,t=0;t<n;t++)if(X(e[t]))return t;return-1}function Q(n){if(e.isGeneratedIdentifier(n)||4096&e.getEmitFlags(n))return n;var t=p.createTempVariable(y);return Ee(t,n,n),t}function Z(e){var n=e?p.createUniqueName(e):p.createTempVariable(void 0);return y(n),n}function $(){u||(u=[]);var e=P;return P++,u[e]=-1,e}function ee(n){e.Debug.assert(void 0!==u,"No labels were defined."),u[n]=T?T.length:0}function ne(e){o||(o=[],c=[],s=[],l=[]);var n=c.length;return c[n]=0,s[n]=T?T.length:0,o[n]=e,l.push(e),n}function te(){var n=re();if(void 0===n)return e.Debug.fail("beginBlock was never called.");var t=c.length;return c[t]=1,s[t]=T?T.length:0,o[t]=n,l.pop(),n}function re(){return e.lastOrUndefined(l)}function ae(){var e=re();return e&&e.kind}function ie(){ne({kind:3,isScript:!0,breakLabel:-1,continueLabel:-1})}function oe(e){var n=$();return ne({kind:3,isScript:!1,breakLabel:n,continueLabel:e}),n}function se(){e.Debug.assert(3===ae());var n=te(),t=n.breakLabel;n.isScript||ee(t)}function ce(){e.Debug.assert(2===ae());var n=te(),t=n.breakLabel;n.isScript||ee(t)}function le(){e.Debug.assert(4===ae());var n=te();n.isScript||ee(n.breakLabel)}function ue(e){return 2===e.kind||3===e.kind}function de(e){return 4===e.kind}function pe(e){return 3===e.kind}function me(e,n){for(var t=n;t>=0;t--){var r=l[t];if(!de(r))break;if(r.labelText===e)return!0}return!1}function fe(e){if(l)if(e)for(var n=l.length-1;n>=0;n--){if(de(t=l[n])&&t.labelText===e)return t.breakLabel;if(ue(t)&&me(e,n-1))return t.breakLabel}else for(n=l.length-1;n>=0;n--){var t;if(ue(t=l[n]))return t.breakLabel}return 0}function _e(e){if(l)if(e)for(var n=l.length-1;n>=0;n--){if(pe(t=l[n])&&me(e,n-1))return t.continueLabel}else for(n=l.length-1;n>=0;n--){var t;if(pe(t=l[n]))return t.continueLabel}return 0}function ge(e){if(void 0!==e&&e>0){void 0===d&&(d=[]);var n=p.createNumericLiteral(-1);return void 0===d[e]?d[e]=[n]:d[e].push(n),n}return p.createOmittedExpression()}function ye(n){var t=p.createNumericLiteral(n);return e.addSyntheticTrailingComment(t,3,function(e){switch(e){case 2:return"return";case 3:return"break";case 4:return"yield";case 5:return"yield*";case 7:return"endfinally";default:return}}(n)),t}function ve(n,t){return e.Debug.assertLessThan(0,n,"Invalid label"),e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression([ye(3),ge(n)])),t)}function he(){Le(0)}function be(e){e?Le(1,[e]):he()}function Ee(e,n,t){Le(2,[e,n],t)}function Te(e,n){Le(3,[e],n)}function Se(e,n,t){Le(4,[e,n],t)}function xe(e,n,t){Le(5,[e,n],t)}function Le(e,n,t){void 0===T&&(T=[],S=[],x=[]),void 0===u&&ee($());var r=T.length;T[r]=e,S[r]=n,x[r]=t}function Ae(){M=0,w=0,A=void 0,D=!1,C=!1,I=void 0,N=void 0,R=void 0,k=void 0,O=void 0;var n=function(){if(T){for(var n=0;n<T.length;n++)Ne(n);De(T.length)}else De(0);if(I){var t=p.createPropertyAccessExpression(L,"label"),r=p.createSwitchStatement(t,p.createCaseBlock(I));return[e.startOnNewLine(r)]}if(N)return N;return[]}();return m().createGeneratorHelper(e.setEmitFlags(p.createFunctionExpression(void 0,void 0,void 0,void 0,[p.createParameterDeclaration(void 0,void 0,void 0,L)],void 0,p.createBlock(n,n.length>0)),524288))}function De(e){(function(e){if(!C)return!0;if(!u||!d)return!1;for(var n=0;n<u.length;n++)if(u[n]===e&&d[n])return!0;return!1})(e)&&(Ie(e),O=void 0,ke(void 0,void 0)),N&&I&&Ce(!1),function(){if(void 0!==d&&void 0!==A)for(var e=0;e<A.length;e++){var n=A[e];if(void 0!==n)for(var t=0,r=n;t<r.length;t++){var a=r[t],i=d[a];if(void 0!==i)for(var o=0,s=i;o<s.length;o++){s[o].text=String(e)}}}}()}function Ce(e){if(I||(I=[]),N){if(O)for(var n=O.length-1;n>=0;n--){var t=O[n];N=[p.createWithStatement(t.expression,p.createBlock(N))]}if(k){var r=k.startLabel,a=k.catchLabel,i=k.finallyLabel,o=k.endLabel;N.unshift(p.createExpressionStatement(p.createCallExpression(p.createPropertyAccessExpression(p.createPropertyAccessExpression(L,"trys"),"push"),void 0,[p.createArrayLiteralExpression([ge(r),ge(a),ge(i),ge(o)])]))),k=void 0}e&&N.push(p.createExpressionStatement(p.createAssignment(p.createPropertyAccessExpression(L,"label"),p.createNumericLiteral(w+1))))}I.push(p.createCaseClause(p.createNumericLiteral(w),N||[])),N=void 0}function Ie(e){if(u)for(var n=0;n<u.length;n++)u[n]===e&&(N&&(Ce(!D),D=!1,C=!1,w++),void 0===A&&(A=[]),void 0===A[w]?A[w]=[n]:A[w].push(n))}function Ne(n){if(Ie(n),function(e){if(o)for(;M<c.length&&s[M]<=e;M++){var n=o[M],t=c[M];switch(n.kind){case 0:0===t?(R||(R=[]),N||(N=[]),R.push(k),k=n):1===t&&(k=R.pop());break;case 1:0===t?(O||(O=[]),O.push(n)):1===t&&O.pop()}}}(n),!D){D=!1,C=!1;var t=T[n];if(0!==t){if(10===t)return D=!0,void Re(p.createReturnStatement(p.createArrayLiteralExpression([ye(7)])));var r=S[n];if(1===t)return Re(r[0]);var a,i,l,u=x[n];switch(t){case 2:return a=r[0],i=r[1],l=u,void Re(e.setTextRange(p.createExpressionStatement(p.createAssignment(a,i)),l));case 3:return function(n,t){D=!0,Re(e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression([ye(3),ge(n)])),t),384))}(r[0],u);case 4:return function(n,t,r){Re(e.setEmitFlags(p.createIfStatement(t,e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression([ye(3),ge(n)])),r),384)),1))}(r[0],r[1],u);case 5:return function(n,t,r){Re(e.setEmitFlags(p.createIfStatement(p.createLogicalNot(t),e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression([ye(3),ge(n)])),r),384)),1))}(r[0],r[1],u);case 6:return function(n,t){D=!0,Re(e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression(n?[ye(4),n]:[ye(4)])),t),384))}(r[0],u);case 7:return function(n,t){D=!0,Re(e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression([ye(5),n])),t),384))}(r[0],u);case 8:return ke(r[0],u);case 9:return function(n,t){D=!0,C=!0,Re(e.setTextRange(p.createThrowStatement(n),t))}(r[0],u)}}}}function Re(e){e&&(N?N.push(e):N=[e])}function ke(n,t){D=!0,C=!0,Re(e.setEmitFlags(e.setTextRange(p.createReturnStatement(p.createArrayLiteralExpression(n?[ye(2),n]:[ye(2)])),t),384))}}}(ts||(ts={})),function(e){e.transformModule=function(t){var r=t.factory,a=t.getEmitHelperFactory,i=t.startLexicalEnvironment,o=t.endLexicalEnvironment,s=t.hoistVariableDeclaration,c=t.getCompilerOptions(),l=t.getEmitResolver(),u=t.getEmitHost(),d=e.getEmitScriptTarget(c),p=e.getEmitModuleKind(c),m=t.onSubstituteNode,f=t.onEmitNode;t.onSubstituteNode=function(n,t){if((t=m(n,t)).id&&y[t.id])return t;if(1===n)return function(n){switch(n.kind){case 78:return X(n);case 216:return function(n){if(e.isAssignmentOperator(n.operatorToken.kind)&&e.isIdentifier(n.left)&&!e.isGeneratedIdentifier(n.left)&&!e.isLocalName(n.left)&&!e.isDeclarationNameOfEnumOrNamespace(n.left)){var t=Y(n.left);if(t){for(var r=n,a=0,i=t;a<i.length;a++){var o=i[a];y[e.getNodeId(r)]=!0,r=q(o,r,n)}return r}}return n}(n);case 215:case 214:return function(n){if((45===n.operator||46===n.operator)&&e.isIdentifier(n.operand)&&!e.isGeneratedIdentifier(n.operand)&&!e.isLocalName(n.operand)&&!e.isDeclarationNameOfEnumOrNamespace(n.operand)){var t=Y(n.operand);if(t){for(var a=215===n.kind?e.setTextRange(r.createBinaryExpression(n.operand,r.createToken(45===n.operator?63:64),r.createNumericLiteral(1)),n):n,i=0,o=t;i<o.length;i++){var s=o[i];y[e.getNodeId(a)]=!0,a=q(s,a)}return a}}return n}(n)}return n}(t);if(e.isShorthandPropertyAssignment(t))return function(n){var t=n.name,a=X(t);if(a!==t){if(n.objectAssignmentInitializer){var i=r.createAssignment(a,n.objectAssignmentInitializer);return e.setTextRange(r.createPropertyAssignment(t,i),n)}return e.setTextRange(r.createPropertyAssignment(t,a),n)}return n}(t);return t},t.onEmitNode=function(n,t,r){297===t.kind?(_=t,g=h[e.getOriginalNodeId(_)],y=[],f(n,t,r),_=void 0,g=void 0,y=void 0):f(n,t,r)},t.enableSubstitution(78),t.enableSubstitution(216),t.enableSubstitution(214),t.enableSubstitution(215),t.enableSubstitution(289),t.enableEmitNotification(297);var _,g,y,v,h=[],b=[];return e.chainBundle(t,(function(n){if(n.isDeclarationFile||!(e.isEffectiveExternalModule(n,c)||2097152&n.transformFlags||e.isJsonSourceFile(n)&&e.hasJsonModuleEmitEnabled(c)&&e.outFile(c)))return n;_=n,g=e.collectExternalModuleInfo(t,n,l,c),h[e.getOriginalNodeId(n)]=g;var r=function(n){switch(n){case e.ModuleKind.AMD:return S;case e.ModuleKind.UMD:return x;default:return T}}(p)(n);return _=void 0,g=void 0,v=!1,r}));function E(){return!(g.exportEquals||!e.isExternalModule(_))}function T(n){i();var a=[],s=e.getStrictOptionValue(c,"alwaysStrict")||!c.noImplicitUseStrict&&e.isExternalModule(_),l=r.copyPrologue(n.statements,a,s&&!e.isJsonSourceFile(n),I);if(E()&&e.append(a,z()),e.length(g.exportedNames))for(var u=0;u<g.exportedNames.length;u+=50)e.append(a,r.createExpressionStatement(e.reduceLeft(g.exportedNames.slice(u,u+50),(function(n,t){return r.createAssignment(r.createPropertyAccessExpression(r.createIdentifier("exports"),r.createIdentifier(e.idText(t))),n)}),r.createVoidZero())));e.append(a,e.visitNode(g.externalHelpersImportDeclaration,I,e.isStatement)),e.addRange(a,e.visitNodes(n.statements,I,e.isStatement,l)),C(a,!1),e.insertStatementsAfterStandardPrologue(a,o());var d=r.updateSourceFile(n,e.setTextRange(r.createNodeArray(a),n.statements));return e.addEmitHelpers(d,t.readEmitHelpers()),d}function S(n){var a=r.createIdentifier("define"),i=e.tryGetModuleNameFromFile(r,n,u,c),o=e.isJsonSourceFile(n)&&n,s=L(n,!0),l=s.aliasedModuleNames,d=s.unaliasedModuleNames,p=s.importAliasNames,m=r.updateSourceFile(n,e.setTextRange(r.createNodeArray([r.createExpressionStatement(r.createCallExpression(a,void 0,__spreadArrays(i?[i]:[],[r.createArrayLiteralExpression(o?e.emptyArray:__spreadArrays([r.createStringLiteral("require"),r.createStringLiteral("exports")],l,d)),o?o.statements.length?o.statements[0].expression:r.createObjectLiteralExpression():r.createFunctionExpression(void 0,void 0,void 0,void 0,__spreadArrays([r.createParameterDeclaration(void 0,void 0,void 0,"require"),r.createParameterDeclaration(void 0,void 0,void 0,"exports")],p),void 0,D(n))])))]),n.statements));return e.addEmitHelpers(m,t.readEmitHelpers()),m}function x(n){var a=L(n,!1),i=a.aliasedModuleNames,o=a.unaliasedModuleNames,s=a.importAliasNames,l=e.tryGetModuleNameFromFile(r,n,u,c),d=r.createFunctionExpression(void 0,void 0,void 0,void 0,[r.createParameterDeclaration(void 0,void 0,void 0,"factory")],void 0,e.setTextRange(r.createBlock([r.createIfStatement(r.createLogicalAnd(r.createTypeCheck(r.createIdentifier("module"),"object"),r.createTypeCheck(r.createPropertyAccessExpression(r.createIdentifier("module"),"exports"),"object")),r.createBlock([r.createVariableStatement(void 0,[r.createVariableDeclaration("v",void 0,void 0,r.createCallExpression(r.createIdentifier("factory"),void 0,[r.createIdentifier("require"),r.createIdentifier("exports")]))]),e.setEmitFlags(r.createIfStatement(r.createStrictInequality(r.createIdentifier("v"),r.createIdentifier("undefined")),r.createExpressionStatement(r.createAssignment(r.createPropertyAccessExpression(r.createIdentifier("module"),"exports"),r.createIdentifier("v")))),1)]),r.createIfStatement(r.createLogicalAnd(r.createTypeCheck(r.createIdentifier("define"),"function"),r.createPropertyAccessExpression(r.createIdentifier("define"),"amd")),r.createBlock([r.createExpressionStatement(r.createCallExpression(r.createIdentifier("define"),void 0,__spreadArrays(l?[l]:[],[r.createArrayLiteralExpression(__spreadArrays([r.createStringLiteral("require"),r.createStringLiteral("exports")],i,o)),r.createIdentifier("factory")])))])))],!0),void 0)),p=r.updateSourceFile(n,e.setTextRange(r.createNodeArray([r.createExpressionStatement(r.createCallExpression(d,void 0,[r.createFunctionExpression(void 0,void 0,void 0,void 0,__spreadArrays([r.createParameterDeclaration(void 0,void 0,void 0,"require"),r.createParameterDeclaration(void 0,void 0,void 0,"exports")],s),void 0,D(n))]))]),n.statements));return e.addEmitHelpers(p,t.readEmitHelpers()),p}function L(n,t){for(var a=[],i=[],o=[],s=0,d=n.amdDependencies;s<d.length;s++){var p=d[s];p.name?(a.push(r.createStringLiteral(p.path)),o.push(r.createParameterDeclaration(void 0,void 0,void 0,p.name))):i.push(r.createStringLiteral(p.path))}for(var m=0,f=g.externalImports;m<f.length;m++){var y=f[m],v=e.getExternalModuleNameLiteral(r,y,_,u,l,c),h=e.getLocalNameForExternalImport(r,y,_);v&&(t&&h?(e.setEmitFlags(h,4),a.push(v),o.push(r.createParameterDeclaration(void 0,void 0,void 0,h))):i.push(v))}return{aliasedModuleNames:a,unaliasedModuleNames:i,importAliasNames:o}}function A(n){if(!e.isImportEqualsDeclaration(n)&&!e.isExportDeclaration(n)&&e.getExternalModuleNameLiteral(r,n,_,u,l,c)){var t=e.getLocalNameForExternalImport(r,n,_),a=P(n,t);if(a!==t)return r.createExpressionStatement(r.createAssignment(t,a))}}function D(t){i();var a=[],s=r.copyPrologue(t.statements,a,!c.noImplicitUseStrict,I);E()&&e.append(a,z()),e.length(g.exportedNames)&&e.append(a,r.createExpressionStatement(e.reduceLeft(g.exportedNames,(function(n,t){return r.createAssignment(r.createPropertyAccessExpression(r.createIdentifier("exports"),r.createIdentifier(e.idText(t))),n)}),r.createVoidZero()))),e.append(a,e.visitNode(g.externalHelpersImportDeclaration,I,e.isStatement)),p===e.ModuleKind.AMD&&e.addRange(a,e.mapDefined(g.externalImports,A)),e.addRange(a,e.visitNodes(t.statements,I,e.isStatement,s)),C(a,!0),e.insertStatementsAfterStandardPrologue(a,o());var l=r.createBlock(a,!0);return v&&e.addEmitHelper(l,n),l}function C(n,t){if(g.exportEquals){var a=e.visitNode(g.exportEquals.expression,N);if(a)if(t){var i=r.createReturnStatement(a);e.setTextRange(i,g.exportEquals),e.setEmitFlags(i,1920),n.push(i)}else{i=r.createExpressionStatement(r.createAssignment(r.createPropertyAccessExpression(r.createIdentifier("module"),"exports"),a));e.setTextRange(i,g.exportEquals),e.setEmitFlags(i,1536),n.push(i)}}}function I(n){switch(n.kind){case 261:return function(n){var t,a=e.getNamespaceDeclarationNode(n);if(p!==e.ModuleKind.AMD){if(!n.importClause)return e.setOriginalNode(e.setTextRange(r.createExpressionStatement(M(n)),n),n);var i=[];a&&!e.isDefaultImport(n)?i.push(r.createVariableDeclaration(r.cloneNode(a.name),void 0,void 0,P(n,M(n)))):(i.push(r.createVariableDeclaration(r.getGeneratedNameForNode(n),void 0,void 0,P(n,M(n)))),a&&e.isDefaultImport(n)&&i.push(r.createVariableDeclaration(r.cloneNode(a.name),void 0,void 0,r.getGeneratedNameForNode(n)))),t=e.append(t,e.setOriginalNode(e.setTextRange(r.createVariableStatement(void 0,r.createVariableDeclarationList(i,d>=2?2:0)),n),n))}else a&&e.isDefaultImport(n)&&(t=e.append(t,r.createVariableStatement(void 0,r.createVariableDeclarationList([e.setOriginalNode(e.setTextRange(r.createVariableDeclaration(r.cloneNode(a.name),void 0,void 0,r.getGeneratedNameForNode(n)),n),n)],d>=2?2:0))));if(G(n)){var o=e.getOriginalNodeId(n);b[o]=B(b[o],n)}else t=B(t,n);return e.singleOrMany(t)}(n);case 260:return function(n){var t;e.Debug.assert(e.isExternalModuleImportEqualsDeclaration(n),"import= for internal module references should be handled in an earlier transformer."),p!==e.ModuleKind.AMD?t=e.hasSyntacticModifier(n,1)?e.append(t,e.setOriginalNode(e.setTextRange(r.createExpressionStatement(q(n.name,M(n))),n),n)):e.append(t,e.setOriginalNode(e.setTextRange(r.createVariableStatement(void 0,r.createVariableDeclarationList([r.createVariableDeclaration(r.cloneNode(n.name),void 0,void 0,M(n))],d>=2?2:0)),n),n)):e.hasSyntacticModifier(n,1)&&(t=e.append(t,e.setOriginalNode(e.setTextRange(r.createExpressionStatement(q(r.getExportName(n),r.getLocalName(n))),n),n)));if(G(n)){var a=e.getOriginalNodeId(n);b[a]=U(b[a],n)}else t=U(t,n);return e.singleOrMany(t)}(n);case 267:return function(n){if(!n.moduleSpecifier)return;var t=r.getGeneratedNameForNode(n);if(n.exportClause&&e.isNamedExports(n.exportClause)){var i=[];p!==e.ModuleKind.AMD&&i.push(e.setOriginalNode(e.setTextRange(r.createVariableStatement(void 0,r.createVariableDeclarationList([r.createVariableDeclaration(t,void 0,void 0,M(n))])),n),n));for(var o=0,s=n.exportClause.elements;o<s.length;o++){var l=s[o];if(0===d)i.push(e.setOriginalNode(e.setTextRange(r.createExpressionStatement(a().createCreateBindingHelper(t,r.createStringLiteralFromNode(l.propertyName||l.name),l.propertyName?r.createStringLiteralFromNode(l.name):void 0)),l),l));else{var u=!(!c.esModuleInterop||67108864&e.getEmitFlags(n)||"default"!==e.idText(l.propertyName||l.name)),m=r.createPropertyAccessExpression(u?a().createImportDefaultHelper(t):t,l.propertyName||l.name);i.push(e.setOriginalNode(e.setTextRange(r.createExpressionStatement(q(r.getExportName(l),m,void 0,!0)),l),l))}}return e.singleOrMany(i)}return n.exportClause?((i=[]).push(e.setOriginalNode(e.setTextRange(r.createExpressionStatement(q(r.cloneNode(n.exportClause.name),function(n,t){if(!c.esModuleInterop||67108864&e.getEmitFlags(n))return t;if(e.getExportNeedsImportStarHelper(n))return a().createImportStarHelper(t);return t}(n,p!==e.ModuleKind.AMD?M(n):e.isExportNamespaceAsDefaultDeclaration(n)?t:r.createIdentifier(e.idText(n.exportClause.name))))),n),n)),e.singleOrMany(i)):e.setOriginalNode(e.setTextRange(r.createExpressionStatement(a().createExportStarHelper(p!==e.ModuleKind.AMD?M(n):t)),n),n)}(n);case 266:return function(n){if(n.isExportEquals)return;var t,a=n.original;if(a&&G(a)){var i=e.getOriginalNodeId(n);b[i]=W(b[i],r.createIdentifier("default"),e.visitNode(n.expression,N),n,!0)}else t=W(t,r.createIdentifier("default"),e.visitNode(n.expression,N),n,!0);return e.singleOrMany(t)}(n);case 232:return function(n){var a,i,o;if(e.hasSyntacticModifier(n,1)){for(var s=void 0,c=!1,l=0,u=n.declarationList.declarations;l<u.length;l++){var d=u[l];if(e.isIdentifier(d.name)&&e.isLocalName(d.name))s||(s=e.visitNodes(n.modifiers,J,e.isModifier)),i=e.append(i,d);else if(d.initializer)if(!e.isBindingPattern(d.name)&&(e.isArrowFunction(d.initializer)||e.isFunctionExpression(d.initializer)||e.isClassExpression(d.initializer))){var p=r.createAssignment(e.setTextRange(r.createPropertyAccessExpression(r.createIdentifier("exports"),d.name),d.name),r.createIdentifier(e.getTextOfIdentifierOrLiteral(d.name))),m=r.createVariableDeclaration(d.name,d.exclamationToken,d.type,e.visitNode(d.initializer,N));i=e.append(i,m),o=e.append(o,p),c=!0}else o=e.append(o,F(d))}if(i&&(a=e.append(a,r.updateVariableStatement(n,s,r.updateVariableDeclarationList(n.declarationList,i)))),o){var f=e.setOriginalNode(e.setTextRange(r.createExpressionStatement(r.inlineExpressions(o)),n),n);c&&e.removeAllComments(f),a=e.append(a,f)}}else a=e.append(a,e.visitEachChild(n,N,t));if(G(n)){var _=e.getOriginalNodeId(n);b[_]=V(b[_],n)}else a=V(a,n);return e.singleOrMany(a)}(n);case 251:return function(n){var a;a=e.hasSyntacticModifier(n,1)?e.append(a,e.setOriginalNode(e.setTextRange(r.createFunctionDeclaration(void 0,e.visitNodes(n.modifiers,J,e.isModifier),n.asteriskToken,r.getDeclarationName(n,!0,!0),void 0,e.visitNodes(n.parameters,N),void 0,e.visitEachChild(n.body,N,t)),n),n)):e.append(a,e.visitEachChild(n,N,t));if(G(n)){var i=e.getOriginalNodeId(n);b[i]=H(b[i],n)}else a=H(a,n);return e.singleOrMany(a)}(n);case 252:return function(n){var a;a=e.hasSyntacticModifier(n,1)?e.append(a,e.setOriginalNode(e.setTextRange(r.createClassDeclaration(void 0,e.visitNodes(n.modifiers,J,e.isModifier),r.getDeclarationName(n,!0,!0),void 0,e.visitNodes(n.heritageClauses,N),e.visitNodes(n.members,N)),n),n)):e.append(a,e.visitEachChild(n,N,t));if(G(n)){var i=e.getOriginalNodeId(n);b[i]=H(b[i],n)}else a=H(a,n);return e.singleOrMany(a)}(n);case 338:return function(n){if(G(n)&&232===n.original.kind){var t=e.getOriginalNodeId(n);b[t]=V(b[t],n.original)}return n}(n);case 339:return function(n){var t=e.getOriginalNodeId(n),r=b[t];if(r)return delete b[t],e.append(r,n);return n}(n);default:return e.visitEachChild(n,N,t)}}function N(n){return 2097152&n.transformFlags||1024&n.transformFlags?e.isImportCall(n)?function(n){var t=e.visitNode(e.firstOrUndefined(n.arguments),N),a=!!(4096&n.transformFlags);switch(c.module){case e.ModuleKind.AMD:return k(t,a);case e.ModuleKind.UMD:return function(n,t){if(v=!0,e.isSimpleCopiableExpression(n)){var a=e.isGeneratedIdentifier(n)?n:e.isStringLiteral(n)?r.createStringLiteralFromNode(n):e.setEmitFlags(e.setTextRange(r.cloneNode(n),n),1536);return r.createConditionalExpression(r.createIdentifier("__syncRequire"),void 0,O(n,t),void 0,k(a,t))}var i=r.createTempVariable(s);return r.createComma(r.createAssignment(i,n),r.createConditionalExpression(r.createIdentifier("__syncRequire"),void 0,O(i,t),void 0,k(i,t)))}(null!=t?t:r.createVoidZero(),a);case e.ModuleKind.CommonJS:default:return O(t,a)}}(n):e.isDestructuringAssignment(n)?function(n){if(R(n.left))return e.flattenDestructuringAssignment(n,N,t,0,!1,w);return e.visitEachChild(n,N,t)}(n):e.visitEachChild(n,N,t):n}function R(n){if(e.isObjectLiteralExpression(n))for(var t=0,r=n.properties;t<r.length;t++){switch((o=r[t]).kind){case 288:if(R(o.initializer))return!0;break;case 289:if(R(o.name))return!0;break;case 290:if(R(o.expression))return!0;break;case 165:case 167:case 168:return!1;default:e.Debug.assertNever(o,"Unhandled object member kind")}}else if(e.isArrayLiteralExpression(n))for(var a=0,i=n.elements;a<i.length;a++){var o=i[a];if(e.isSpreadElement(o)){if(R(o.expression))return!0}else if(R(o))return!0}else if(e.isIdentifier(n))return e.length(Y(n))>(e.isExportName(n)?1:0);return!1}function k(n,t){var i,o=r.createUniqueName("resolve"),s=r.createUniqueName("reject"),l=[r.createParameterDeclaration(void 0,void 0,void 0,o),r.createParameterDeclaration(void 0,void 0,void 0,s)],u=r.createBlock([r.createExpressionStatement(r.createCallExpression(r.createIdentifier("require"),void 0,[r.createArrayLiteralExpression([n||r.createOmittedExpression()]),o,s]))]);d>=2?i=r.createArrowFunction(void 0,void 0,l,void 0,void 0,u):(i=r.createFunctionExpression(void 0,void 0,void 0,void 0,l,void 0,u),t&&e.setEmitFlags(i,8));var p=r.createNewExpression(r.createIdentifier("Promise"),void 0,[i]);return c.esModuleInterop?r.createCallExpression(r.createPropertyAccessExpression(p,r.createIdentifier("then")),void 0,[a().createImportStarCallbackHelper()]):p}function O(n,t){var i,o=r.createCallExpression(r.createPropertyAccessExpression(r.createIdentifier("Promise"),"resolve"),void 0,[]),s=r.createCallExpression(r.createIdentifier("require"),void 0,n?[n]:[]);return c.esModuleInterop&&(s=a().createImportStarHelper(s)),d>=2?i=r.createArrowFunction(void 0,void 0,[],void 0,void 0,s):(i=r.createFunctionExpression(void 0,void 0,void 0,void 0,[],void 0,r.createBlock([r.createReturnStatement(s)])),t&&e.setEmitFlags(i,8)),r.createCallExpression(r.createPropertyAccessExpression(o,"then"),void 0,[i])}function P(n,t){return!c.esModuleInterop||67108864&e.getEmitFlags(n)?t:e.getImportNeedsImportStarHelper(n)?a().createImportStarHelper(t):e.getImportNeedsImportDefaultHelper(n)?a().createImportDefaultHelper(t):t}function M(n){var t=e.getExternalModuleNameLiteral(r,n,_,u,l,c),a=[];return t&&a.push(t),r.createCallExpression(r.createIdentifier("require"),void 0,a)}function w(n,t,a){var i=Y(n);if(i){for(var o=e.isExportName(n)?t:r.createAssignment(n,t),s=0,c=i;s<c.length;s++){var l=c[s];e.setEmitFlags(o,4),o=q(l,o,a)}return o}return r.createAssignment(n,t)}function F(n){return e.isBindingPattern(n.name)?e.flattenDestructuringAssignment(e.visitNode(n,N),void 0,t,0,!1,w):r.createAssignment(e.setTextRange(r.createPropertyAccessExpression(r.createIdentifier("exports"),n.name),n.name),n.initializer?e.visitNode(n.initializer,N):r.createVoidZero())}function G(n){return 0!=(4194304&e.getEmitFlags(n))}function B(e,n){if(g.exportEquals)return e;var t=n.importClause;if(!t)return e;t.name&&(e=K(e,t));var r=t.namedBindings;if(r)switch(r.kind){case 263:e=K(e,r);break;case 264:for(var a=0,i=r.elements;a<i.length;a++){e=K(e,i[a],!0)}}return e}function U(e,n){return g.exportEquals?e:K(e,n)}function V(e,n){if(g.exportEquals)return e;for(var t=0,r=n.declarationList.declarations;t<r.length;t++){e=j(e,r[t])}return e}function j(n,t){if(g.exportEquals)return n;if(e.isBindingPattern(t.name))for(var r=0,a=t.name.elements;r<a.length;r++){var i=a[r];e.isOmittedExpression(i)||(n=j(n,i))}else e.isGeneratedIdentifier(t.name)||(n=K(n,t));return n}function H(n,t){if(g.exportEquals)return n;e.hasSyntacticModifier(t,1)&&(n=W(n,e.hasSyntacticModifier(t,512)?r.createIdentifier("default"):r.getDeclarationName(t),r.getLocalName(t),t));return t.name&&(n=K(n,t)),n}function K(n,t,a){var i=r.getDeclarationName(t),o=g.exportSpecifiers.get(e.idText(i));if(o)for(var s=0,c=o;s<c.length;s++){var l=c[s];n=W(n,l.name,i,l.name,void 0,a)}return n}function W(n,t,a,i,o,s){return n=e.append(n,function(n,t,a,i,o){var s=e.setTextRange(r.createExpressionStatement(q(n,t,void 0,o)),a);e.startOnNewLine(s),i||e.setEmitFlags(s,1536);return s}(t,a,i,o,s))}function z(){var n;return n=0===d?r.createExpressionStatement(q(r.createIdentifier("__esModule"),r.createTrue())):r.createExpressionStatement(r.createCallExpression(r.createPropertyAccessExpression(r.createIdentifier("Object"),"defineProperty"),void 0,[r.createIdentifier("exports"),r.createStringLiteral("__esModule"),r.createObjectLiteralExpression([r.createPropertyAssignment("value",r.createTrue())])])),e.setEmitFlags(n,1048576),n}function q(n,t,a,i){return e.setTextRange(i&&0!==d?r.createCallExpression(r.createPropertyAccessExpression(r.createIdentifier("Object"),"defineProperty"),void 0,[r.createIdentifier("exports"),r.createStringLiteralFromNode(n),r.createObjectLiteralExpression([r.createPropertyAssignment("enumerable",r.createTrue()),r.createPropertyAssignment("get",r.createFunctionExpression(void 0,void 0,void 0,void 0,[],void 0,r.createBlock([r.createReturnStatement(t)])))])]):r.createAssignment(r.createPropertyAccessExpression(r.createIdentifier("exports"),r.cloneNode(n)),t),a)}function J(e){switch(e.kind){case 92:case 87:return}return e}function X(n){if(4096&e.getEmitFlags(n)){var t=e.getExternalHelpersModuleName(_);return t?r.createPropertyAccessExpression(t,n):n}if((!e.isGeneratedIdentifier(n)||64&n.autoGenerateFlags)&&!e.isLocalName(n)){var a=l.getReferencedExportContainer(n,e.isExportName(n));if(a&&297===a.kind)return e.setTextRange(r.createPropertyAccessExpression(r.createIdentifier("exports"),r.cloneNode(n)),n);var i=l.getReferencedImportDeclaration(n);if(i){if(e.isImportClause(i))return e.setTextRange(r.createPropertyAccessExpression(r.getGeneratedNameForNode(i.parent),r.createIdentifier("default")),n);if(e.isImportSpecifier(i)){var o=i.propertyName||i.name;return e.setTextRange(r.createPropertyAccessExpression(r.getGeneratedNameForNode(i.parent.parent.parent),r.cloneNode(o)),n)}}}return n}function Y(n){if(!e.isGeneratedIdentifier(n)){var t=l.getReferencedImportDeclaration(n)||l.getReferencedValueDeclaration(n);if(t)return g&&g.exportedBindings[e.getOriginalNodeId(t)]}}};var n={name:"typescript:dynamicimport-sync-require",scoped:!0,text:'\n            var __syncRequire = typeof module === "object" && typeof module.exports === "object";'}}(ts||(ts={})),function(e){e.transformSystemModule=function(n){var t=n.factory,r=n.startLexicalEnvironment,a=n.endLexicalEnvironment,i=n.hoistVariableDeclaration,o=n.getCompilerOptions(),s=n.getEmitResolver(),c=n.getEmitHost(),l=n.onSubstituteNode,u=n.onEmitNode;n.onSubstituteNode=function(n,r){if(function(e){return y&&e.id&&y[e.id]}(r=l(n,r)))return r;if(1===n)return function(n){switch(n.kind){case 78:return function(n){if(4096&e.getEmitFlags(n)){var r=e.getExternalHelpersModuleName(d);return r?t.createPropertyAccessExpression(r,n):n}if(!e.isGeneratedIdentifier(n)&&!e.isLocalName(n)){var a=s.getReferencedImportDeclaration(n);if(a){if(e.isImportClause(a))return e.setTextRange(t.createPropertyAccessExpression(t.getGeneratedNameForNode(a.parent),t.createIdentifier("default")),n);if(e.isImportSpecifier(a))return e.setTextRange(t.createPropertyAccessExpression(t.getGeneratedNameForNode(a.parent.parent.parent),t.cloneNode(a.propertyName||a.name)),n)}}return n}(n);case 216:return function(n){if(e.isAssignmentOperator(n.operatorToken.kind)&&e.isIdentifier(n.left)&&!e.isGeneratedIdentifier(n.left)&&!e.isLocalName(n.left)&&!e.isDeclarationNameOfEnumOrNamespace(n.left)){var t=q(n.left);if(t){for(var r=n,a=0,i=t;a<i.length;a++){r=V(i[a],J(r))}return r}}return n}(n);case 214:case 215:return function(n){if((45===n.operator||46===n.operator)&&e.isIdentifier(n.operand)&&!e.isGeneratedIdentifier(n.operand)&&!e.isLocalName(n.operand)&&!e.isDeclarationNameOfEnumOrNamespace(n.operand)){var r=q(n.operand);if(r){for(var a=215===n.kind?e.setTextRange(t.createPrefixUnaryExpression(n.operator,n.operand),n):n,i=0,o=r;i<o.length;i++){a=V(o[i],J(a))}return 215===n.kind&&(a=45===n.operator?t.createSubtract(J(a),t.createNumericLiteral(1)):t.createAdd(J(a),t.createNumericLiteral(1))),a}}return n}(n);case 226:return function(n){if(e.isImportMeta(n))return t.createPropertyAccessExpression(f,t.createIdentifier("meta"));return n}(n)}return n}(r);if(4===n)return function(n){switch(n.kind){case 289:return function(n){var r=n.name;if(!e.isGeneratedIdentifier(r)&&!e.isLocalName(r)){var a=s.getReferencedImportDeclaration(r);if(a){if(e.isImportClause(a))return e.setTextRange(t.createPropertyAssignment(t.cloneNode(r),t.createPropertyAccessExpression(t.getGeneratedNameForNode(a.parent),t.createIdentifier("default"))),n);if(e.isImportSpecifier(a))return e.setTextRange(t.createPropertyAssignment(t.cloneNode(r),t.createPropertyAccessExpression(t.getGeneratedNameForNode(a.parent.parent.parent),t.cloneNode(a.propertyName||a.name))),n)}}return n}(n)}return n}(r);return r},n.onEmitNode=function(n,t,r){if(297===t.kind){var a=e.getOriginalNodeId(t);d=t,p=v[a],m=b[a],y=E[a],f=T[a],y&&delete E[a],u(n,t,r),d=void 0,p=void 0,m=void 0,f=void 0,y=void 0}else u(n,t,r)},n.enableSubstitution(78),n.enableSubstitution(289),n.enableSubstitution(216),n.enableSubstitution(214),n.enableSubstitution(215),n.enableSubstitution(226),n.enableEmitNotification(297);var d,p,m,f,_,g,y,v=[],h=[],b=[],E=[],T=[];return e.chainBundle(n,(function(i){if(i.isDeclarationFile||!(e.isEffectiveExternalModule(i,o)||2097152&i.transformFlags))return i;var l=e.getOriginalNodeId(i);d=i,g=i,p=v[l]=e.collectExternalModuleInfo(n,i,s,o),m=t.createUniqueName("exports"),b[l]=m,f=T[l]=t.createUniqueName("context");var u=function(n){for(var r=new e.Map,a=[],i=0,l=n;i<l.length;i++){var u=l[i],p=e.getExternalModuleNameLiteral(t,u,d,c,s,o);if(p){var m=p.text,f=r.get(m);void 0!==f?a[f].externalImports.push(u):(r.set(m,a.length),a.push({name:p,externalImports:[u]}))}}return a}(p.externalImports),h=function(n,i){var s=[];r();var c=e.getStrictOptionValue(o,"alwaysStrict")||!o.noImplicitUseStrict&&e.isExternalModule(d),l=t.copyPrologue(n.statements,s,c,L);s.push(t.createVariableStatement(void 0,t.createVariableDeclarationList([t.createVariableDeclaration("__moduleName",void 0,void 0,t.createLogicalAnd(f,t.createPropertyAccessExpression(f,"id")))]))),e.visitNode(p.externalHelpersImportDeclaration,L,e.isStatement);var u=e.visitNodes(n.statements,L,e.isStatement,l);e.addRange(s,_),e.insertStatementsAfterStandardPrologue(s,a());var m=function(e){if(!p.hasExportStarsToExportValues)return;if(!p.exportedNames&&0===p.exportSpecifiers.size){for(var n=!1,r=0,a=p.externalImports;r<a.length;r++){var i=a[r];if(267===i.kind&&i.exportClause){n=!0;break}}if(!n){var o=S(void 0);return e.push(o),o.name}}var s=[];if(p.exportedNames)for(var c=0,l=p.exportedNames;c<l.length;c++){var u=l[c];"default"!==u.escapedText&&s.push(t.createPropertyAssignment(t.createStringLiteralFromNode(u),t.createTrue()))}var d=t.createUniqueName("exportedNames");e.push(t.createVariableStatement(void 0,t.createVariableDeclarationList([t.createVariableDeclaration(d,void 0,void 0,t.createObjectLiteralExpression(s,!0))])));var m=S(d);return e.push(m),m.name}(s),g=524288&n.transformFlags?t.createModifiersFromModifierFlags(256):void 0,y=t.createObjectLiteralExpression([t.createPropertyAssignment("setters",x(m,i)),t.createPropertyAssignment("execute",t.createFunctionExpression(g,void 0,void 0,void 0,[],void 0,t.createBlock(u,!0)))],!0);return s.push(t.createReturnStatement(y)),t.createBlock(s,!0)}(i,u),A=t.createFunctionExpression(void 0,void 0,void 0,void 0,[t.createParameterDeclaration(void 0,void 0,void 0,m),t.createParameterDeclaration(void 0,void 0,void 0,f)],void 0,h),D=e.tryGetModuleNameFromFile(t,i,c,o),C=t.createArrayLiteralExpression(e.map(u,(function(e){return e.name}))),I=e.setEmitFlags(t.updateSourceFile(i,e.setTextRange(t.createNodeArray([t.createExpressionStatement(t.createCallExpression(t.createPropertyAccessExpression(t.createIdentifier("System"),"register"),void 0,D?[D,C,A]:[C,A]))]),i.statements)),1024);e.outFile(o)||e.moveEmitHelpers(I,h,(function(e){return!e.scoped}));y&&(E[l]=y,y=void 0);return d=void 0,p=void 0,m=void 0,f=void 0,_=void 0,g=void 0,I}));function S(n){var r=t.createUniqueName("exportStar"),a=t.createIdentifier("m"),i=t.createIdentifier("n"),o=t.createIdentifier("exports"),s=t.createStrictInequality(i,t.createStringLiteral("default"));return n&&(s=t.createLogicalAnd(s,t.createLogicalNot(t.createCallExpression(t.createPropertyAccessExpression(n,"hasOwnProperty"),void 0,[i])))),t.createFunctionDeclaration(void 0,void 0,void 0,r,void 0,[t.createParameterDeclaration(void 0,void 0,void 0,a)],void 0,t.createBlock([t.createVariableStatement(void 0,t.createVariableDeclarationList([t.createVariableDeclaration(o,void 0,void 0,t.createObjectLiteralExpression([]))])),t.createForInStatement(t.createVariableDeclarationList([t.createVariableDeclaration(i)]),a,t.createBlock([e.setEmitFlags(t.createIfStatement(s,t.createExpressionStatement(t.createAssignment(t.createElementAccessExpression(o,i),t.createElementAccessExpression(a,i)))),1)])),t.createExpressionStatement(t.createCallExpression(m,void 0,[o]))],!0))}function x(n,r){for(var a=[],i=0,o=r;i<o.length;i++){for(var s=o[i],c=e.forEach(s.externalImports,(function(n){return e.getLocalNameForExternalImport(t,n,d)})),l=c?t.getGeneratedNameForNode(c):t.createUniqueName(""),u=[],p=0,f=s.externalImports;p<f.length;p++){var _=f[p],g=e.getLocalNameForExternalImport(t,_,d);switch(_.kind){case 261:if(!_.importClause)break;case 260:e.Debug.assert(void 0!==g),u.push(t.createExpressionStatement(t.createAssignment(g,l)));break;case 267:if(e.Debug.assert(void 0!==g),_.exportClause)if(e.isNamedExports(_.exportClause)){for(var y=[],v=0,h=_.exportClause.elements;v<h.length;v++){var b=h[v];y.push(t.createPropertyAssignment(t.createStringLiteral(e.idText(b.name)),t.createElementAccessExpression(l,t.createStringLiteral(e.idText(b.propertyName||b.name)))))}u.push(t.createExpressionStatement(t.createCallExpression(m,void 0,[t.createObjectLiteralExpression(y,!0)])))}else u.push(t.createExpressionStatement(t.createCallExpression(m,void 0,[t.createStringLiteral(e.idText(_.exportClause.name)),l])));else u.push(t.createExpressionStatement(t.createCallExpression(n,void 0,[l])))}}a.push(t.createFunctionExpression(void 0,void 0,void 0,void 0,[t.createParameterDeclaration(void 0,void 0,void 0,l)],void 0,t.createBlock(u,!0)))}return t.createArrayLiteralExpression(a,!0)}function L(n){switch(n.kind){case 261:return function(n){var r;n.importClause&&i(e.getLocalNameForExternalImport(t,n,d));if(k(n)){var a=e.getOriginalNodeId(n);h[a]=O(h[a],n)}else r=O(r,n);return e.singleOrMany(r)}(n);case 260:return function(n){var r;if(e.Debug.assert(e.isExternalModuleImportEqualsDeclaration(n),"import= for internal module references should be handled in an earlier transformer."),i(e.getLocalNameForExternalImport(t,n,d)),k(n)){var a=e.getOriginalNodeId(n);h[a]=P(h[a],n)}else r=P(r,n);return e.singleOrMany(r)}(n);case 267:return function(n){return void e.Debug.assertIsDefined(n)}(n);case 266:return function(n){if(n.isExportEquals)return;var r=e.visitNode(n.expression,K,e.isExpression),a=n.original;if(!a||!k(a))return U(t.createIdentifier("default"),r,!0);var i=e.getOriginalNodeId(n);h[i]=B(h[i],t.createIdentifier("default"),r,!0)}(n);default:return j(n)}}function A(n){if(e.isBindingPattern(n.name))for(var r=0,a=n.name.elements;r<a.length;r++){var o=a[r];e.isOmittedExpression(o)||A(o)}else i(t.cloneNode(n.name))}function D(n){return 0==(2097152&e.getEmitFlags(n))&&(297===g.kind||0==(3&e.getOriginalNode(n).flags))}function C(t,r){var a=r?I:N;return e.isBindingPattern(t.name)?e.flattenDestructuringAssignment(t,K,n,0,!1,a):t.initializer?a(t.name,e.visitNode(t.initializer,K,e.isExpression)):t.name}function I(e,n,t){return R(e,n,t,!0)}function N(e,n,t){return R(e,n,t,!1)}function R(n,r,a,o){return i(t.cloneNode(n)),o?V(n,J(e.setTextRange(t.createAssignment(n,r),a))):J(e.setTextRange(t.createAssignment(n,r),a))}function k(n){return 0!=(4194304&e.getEmitFlags(n))}function O(e,n){if(p.exportEquals)return e;var t=n.importClause;if(!t)return e;t.name&&(e=G(e,t));var r=t.namedBindings;if(r)switch(r.kind){case 263:e=G(e,r);break;case 264:for(var a=0,i=r.elements;a<i.length;a++){e=G(e,i[a])}}return e}function P(e,n){return p.exportEquals?e:G(e,n)}function M(e,n,t){if(p.exportEquals)return e;for(var r=0,a=n.declarationList.declarations;r<a.length;r++){var i=a[r];(i.initializer||t)&&(e=w(e,i,t))}return e}function w(n,r,a){if(p.exportEquals)return n;if(e.isBindingPattern(r.name))for(var i=0,o=r.name.elements;i<o.length;i++){var s=o[i];e.isOmittedExpression(s)||(n=w(n,s,a))}else if(!e.isGeneratedIdentifier(r.name)){var c=void 0;a&&(n=B(n,r.name,t.getLocalName(r)),c=e.idText(r.name)),n=G(n,r,c)}return n}function F(n,r){if(p.exportEquals)return n;var a;if(e.hasSyntacticModifier(r,1)){var i=e.hasSyntacticModifier(r,512)?t.createStringLiteral("default"):r.name;n=B(n,i,t.getLocalName(r)),a=e.getTextOfIdentifierOrLiteral(i)}return r.name&&(n=G(n,r,a)),n}function G(n,r,a){if(p.exportEquals)return n;var i=t.getDeclarationName(r),o=p.exportSpecifiers.get(e.idText(i));if(o)for(var s=0,c=o;s<c.length;s++){var l=c[s];l.name.escapedText!==a&&(n=B(n,l.name,i))}return n}function B(n,t,r,a){return n=e.append(n,U(t,r,a))}function U(n,r,a){var i=t.createExpressionStatement(V(n,r));return e.startOnNewLine(i),a||e.setEmitFlags(i,1536),i}function V(n,r){var a=e.isIdentifier(n)?t.createStringLiteralFromNode(n):n;return e.setEmitFlags(r,1536|e.getEmitFlags(r)),e.setCommentRange(t.createCallExpression(m,void 0,[a,r]),r)}function j(r){switch(r.kind){case 232:return function(n){if(!D(n.declarationList))return e.visitNode(n,K,e.isStatement);for(var r,a,i=e.hasSyntacticModifier(n,1),o=k(n),s=0,c=n.declarationList.declarations;s<c.length;s++){var l=c[s];l.initializer?r=e.append(r,C(l,i&&!o)):A(l)}if(r&&(a=e.append(a,e.setTextRange(t.createExpressionStatement(t.inlineExpressions(r)),n))),o){var u=e.getOriginalNodeId(n);h[u]=M(h[u],n,i)}else a=M(a,n,!1);return e.singleOrMany(a)}(r);case 251:return function(r){if(_=e.hasSyntacticModifier(r,1)?e.append(_,t.updateFunctionDeclaration(r,r.decorators,e.visitNodes(r.modifiers,z,e.isModifier),r.asteriskToken,t.getDeclarationName(r,!0,!0),void 0,e.visitNodes(r.parameters,K,e.isParameterDeclaration),void 0,e.visitNode(r.body,K,e.isBlock))):e.append(_,e.visitEachChild(r,K,n)),k(r)){var a=e.getOriginalNodeId(r);h[a]=F(h[a],r)}else _=F(_,r)}(r);case 252:return function(n){var r,a=t.getLocalName(n);if(i(a),r=e.append(r,e.setTextRange(t.createExpressionStatement(t.createAssignment(a,e.setTextRange(t.createClassExpression(e.visitNodes(n.decorators,K,e.isDecorator),void 0,n.name,void 0,e.visitNodes(n.heritageClauses,K,e.isHeritageClause),e.visitNodes(n.members,K,e.isClassElement)),n))),n)),k(n)){var o=e.getOriginalNodeId(n);h[o]=F(h[o],n)}else r=F(r,n);return e.singleOrMany(r)}(r);case 237:return function(n){var r=g;return g=n,n=t.updateForStatement(n,n.initializer&&H(n.initializer),e.visitNode(n.condition,K,e.isExpression),e.visitNode(n.incrementor,K,e.isExpression),e.visitNode(n.statement,j,e.isStatement)),g=r,n}(r);case 238:return function(n){var r=g;return g=n,n=t.updateForInStatement(n,H(n.initializer),e.visitNode(n.expression,K,e.isExpression),e.visitNode(n.statement,j,e.isStatement,t.liftToBlock)),g=r,n}(r);case 239:return function(n){var r=g;return g=n,n=t.updateForOfStatement(n,n.awaitModifier,H(n.initializer),e.visitNode(n.expression,K,e.isExpression),e.visitNode(n.statement,j,e.isStatement,t.liftToBlock)),g=r,n}(r);case 235:return function(n){return t.updateDoStatement(n,e.visitNode(n.statement,j,e.isStatement,t.liftToBlock),e.visitNode(n.expression,K,e.isExpression))}(r);case 236:return function(n){return t.updateWhileStatement(n,e.visitNode(n.expression,K,e.isExpression),e.visitNode(n.statement,j,e.isStatement,t.liftToBlock))}(r);case 245:return function(n){return t.updateLabeledStatement(n,n.label,e.visitNode(n.statement,j,e.isStatement,t.liftToBlock))}(r);case 243:return function(n){return t.updateWithStatement(n,e.visitNode(n.expression,K,e.isExpression),e.visitNode(n.statement,j,e.isStatement,t.liftToBlock))}(r);case 244:return function(n){return t.updateSwitchStatement(n,e.visitNode(n.expression,K,e.isExpression),e.visitNode(n.caseBlock,j,e.isCaseBlock))}(r);case 258:return function(n){var r=g;return g=n,n=t.updateCaseBlock(n,e.visitNodes(n.clauses,j,e.isCaseOrDefaultClause)),g=r,n}(r);case 284:return function(n){return t.updateCaseClause(n,e.visitNode(n.expression,K,e.isExpression),e.visitNodes(n.statements,j,e.isStatement))}(r);case 285:case 247:return function(t){return e.visitEachChild(t,j,n)}(r);case 287:return function(n){var r=g;return g=n,n=t.updateCatchClause(n,n.variableDeclaration,e.visitNode(n.block,j,e.isBlock)),g=r,n}(r);case 230:return function(t){var r=g;return g=t,t=e.visitEachChild(t,j,n),g=r,t}(r);case 338:return function(n){if(k(n)&&232===n.original.kind){var t=e.getOriginalNodeId(n),r=e.hasSyntacticModifier(n.original,1);h[t]=M(h[t],n.original,r)}return n}(r);case 339:return function(n){var t=e.getOriginalNodeId(n),r=h[t];if(r)return delete h[t],e.append(r,n);var a=e.getOriginalNode(n);return e.isModuleOrEnumDeclaration(a)?e.append(G(r,a),n):n}(r);default:return K(r)}}function H(r){if(function(n){return e.isVariableDeclarationList(n)&&D(n)}(r)){for(var a=void 0,i=0,o=r.declarations;i<o.length;i++){var s=o[i];a=e.append(a,C(s,!1)),s.initializer||A(s)}return a?t.inlineExpressions(a):t.createOmittedExpression()}return e.visitEachChild(r,j,n)}function K(r){return e.isDestructuringAssignment(r)?function(t){if(W(t.left))return e.flattenDestructuringAssignment(t,K,n,0,!0);return e.visitEachChild(t,K,n)}(r):e.isImportCall(r)?function(n){return t.createCallExpression(t.createPropertyAccessExpression(f,t.createIdentifier("import")),void 0,e.some(n.arguments)?[e.visitNode(n.arguments[0],K)]:[])}(r):1024&r.transformFlags||2097152&r.transformFlags?e.visitEachChild(r,K,n):r}function W(n){if(e.isAssignmentExpression(n,!0))return W(n.left);if(e.isSpreadElement(n))return W(n.expression);if(e.isObjectLiteralExpression(n))return e.some(n.properties,W);if(e.isArrayLiteralExpression(n))return e.some(n.elements,W);if(e.isShorthandPropertyAssignment(n))return W(n.name);if(e.isPropertyAssignment(n))return W(n.initializer);if(e.isIdentifier(n)){var t=s.getReferencedExportContainer(n);return void 0!==t&&297===t.kind}return!1}function z(e){switch(e.kind){case 92:case 87:return}return e}function q(n){var r;if(!e.isGeneratedIdentifier(n)){var a=s.getReferencedImportDeclaration(n)||s.getReferencedValueDeclaration(n);if(a){var i=s.getReferencedExportContainer(n,!1);i&&297===i.kind&&(r=e.append(r,t.getDeclarationName(a))),r=e.addRange(r,p&&p.exportedBindings[e.getOriginalNodeId(a)])}}return r}function J(n){return void 0===y&&(y=[]),y[e.getNodeId(n)]=!0,n}}}(ts||(ts={})),function(e){e.transformECMAScriptModule=function(n){var t,r=n.factory,a=n.getEmitHelperFactory,i=n.getCompilerOptions(),o=n.onEmitNode,s=n.onSubstituteNode;return n.onEmitNode=function(n,r,a){e.isSourceFile(r)?((e.isExternalModule(r)||i.isolatedModules)&&i.importHelpers&&(t=new e.Map),o(n,r,a),t=void 0):o(n,r,a)},n.onSubstituteNode=function(n,a){if(a=s(n,a),t&&e.isIdentifier(a)&&4096&e.getEmitFlags(a))return function(n){var a=e.idText(n),i=t.get(a);i||t.set(a,i=r.createUniqueName(a,48));return i}(a);return a},n.enableEmitNotification(297),n.enableSubstitution(78),e.chainBundle(n,(function(t){if(t.isDeclarationFile)return t;if(e.isExternalModule(t)||i.isolatedModules){var o=function(t){var o=e.createExternalHelpersImportDeclarationIfNeeded(r,a(),t,i);if(o){var s=[],l=r.copyPrologue(t.statements,s);return e.append(s,o),e.addRange(s,e.visitNodes(t.statements,c,e.isStatement,l)),r.updateSourceFile(t,e.setTextRange(r.createNodeArray(s),t.statements))}return e.visitEachChild(t,c,n)}(t);return!e.isExternalModule(t)||e.some(o.statements,e.isExternalModuleIndicator)?o:r.updateSourceFile(o,e.setTextRange(r.createNodeArray(__spreadArrays(o.statements,[e.createEmptyExports(r)])),o.statements))}return t}));function c(n){switch(n.kind){case 260:return;case 266:return function(e){return e.isExportEquals?void 0:e}(n);case 267:return function(n){if(void 0!==i.module&&i.module>e.ModuleKind.ES2015)return n;if(!n.exportClause||!e.isNamespaceExport(n.exportClause)||!n.moduleSpecifier)return n;var t=n.exportClause.name,a=r.getGeneratedNameForNode(t),o=r.createImportDeclaration(void 0,void 0,r.createImportClause(!1,void 0,r.createNamespaceImport(a)),n.moduleSpecifier);e.setOriginalNode(o,n.exportClause);var s=e.isExportNamespaceAsDefaultDeclaration(n)?r.createExportDefault(a):r.createExportDeclaration(void 0,void 0,!1,r.createNamedExports([r.createExportSpecifier(a,t)]));return e.setOriginalNode(s,n),[o,s]}(n)}return n}}}(ts||(ts={})),function(e){function n(n){return e.isVariableDeclaration(n)||e.isPropertyDeclaration(n)||e.isPropertySignature(n)||e.isPropertyAccessExpression(n)||e.isBindingElement(n)||e.isConstructorDeclaration(n)?t:e.isSetAccessor(n)||e.isGetAccessor(n)?function(t){var r;r=168===n.kind?e.hasSyntacticModifier(n,32)?t.errorModuleName?e.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1:e.hasSyntacticModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1;return{diagnosticMessage:r,errorNode:n.name,typeName:n.name}}:e.isConstructSignatureDeclaration(n)||e.isCallSignatureDeclaration(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isFunctionDeclaration(n)||e.isIndexSignatureDeclaration(n)?function(t){var r;switch(n.kind){case 170:r=t.errorModuleName?e.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 169:r=t.errorModuleName?e.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 171:r=t.errorModuleName?e.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0;break;case 165:case 164:r=e.hasSyntacticModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0:252===n.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0:t.errorModuleName?e.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0;break;case 251:r=t.errorModuleName?2===t.accessibility?e.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named:e.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1:e.Diagnostics.Return_type_of_exported_function_has_or_is_using_private_name_0;break;default:return e.Debug.fail("This is unknown kind for signature: "+n.kind)}return{diagnosticMessage:r,errorNode:n.name||n}}:e.isParameter(n)?e.isParameterPropertyDeclaration(n,n.parent)&&e.hasSyntacticModifier(n.parent,8)?t:function(t){var r=function(t){switch(n.parent.kind){case 166:return t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1;case 170:case 175:return t.errorModuleName?e.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;case 169:return t.errorModuleName?e.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;case 171:return t.errorModuleName?e.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1;case 165:case 164:return e.hasSyntacticModifier(n.parent,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:252===n.parent.parent.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;case 251:case 174:return t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_private_name_1;case 168:case 167:return t.errorModuleName?2===t.accessibility?e.Diagnostics.Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Parameter_0_of_accessor_has_or_is_using_private_name_1;default:return e.Debug.fail("Unknown parent for parameter: "+e.SyntaxKind[n.parent.kind])}}(t);return void 0!==r?{diagnosticMessage:r,errorNode:n,typeName:n.name}:void 0}:e.isTypeParameterDeclaration(n)?function(){var t;switch(n.parent.kind){case 252:t=e.Diagnostics.Type_parameter_0_of_exported_class_has_or_is_using_private_name_1;break;case 253:t=e.Diagnostics.Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1;break;case 190:t=e.Diagnostics.Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1;break;case 175:case 170:t=e.Diagnostics.Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;break;case 169:t=e.Diagnostics.Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;break;case 165:case 164:t=e.hasSyntacticModifier(n.parent,32)?e.Diagnostics.Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1:252===n.parent.parent.kind?e.Diagnostics.Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1:e.Diagnostics.Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;break;case 174:case 251:t=e.Diagnostics.Type_parameter_0_of_exported_function_has_or_is_using_private_name_1;break;case 254:t=e.Diagnostics.Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1;break;default:return e.Debug.fail("This is unknown parent for type parameter: "+n.parent.kind)}return{diagnosticMessage:t,errorNode:n,typeName:n.name}}:e.isExpressionWithTypeArguments(n)?function(){var t;t=e.isClassDeclaration(n.parent.parent)?e.isHeritageClause(n.parent)&&116===n.parent.token?e.Diagnostics.Implements_clause_of_exported_class_0_has_or_is_using_private_name_1:n.parent.parent.name?e.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1:e.Diagnostics.extends_clause_of_exported_class_has_or_is_using_private_name_0:e.Diagnostics.extends_clause_of_exported_interface_0_has_or_is_using_private_name_1;return{diagnosticMessage:t,errorNode:n,typeName:e.getNameOfDeclaration(n.parent.parent)}}:e.isImportEqualsDeclaration(n)?function(){return{diagnosticMessage:e.Diagnostics.Import_declaration_0_is_using_private_name_1,errorNode:n,typeName:n.name}}:e.isTypeAliasDeclaration(n)?function(){return{diagnosticMessage:e.Diagnostics.Exported_type_alias_0_has_or_is_using_private_name_1,errorNode:n.type,typeName:n.name}}:e.Debug.assertNever(n,"Attempted to set a declaration diagnostic context for unhandled node kind: "+e.SyntaxKind[n.kind]);function t(t){var r=function(t){return 249===n.kind||198===n.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Exported_variable_0_has_or_is_using_private_name_1:163===n.kind||201===n.kind||162===n.kind||160===n.kind&&e.hasSyntacticModifier(n.parent,8)?e.hasSyntacticModifier(n,32)?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:252===n.parent.kind||160===n.kind?t.errorModuleName?2===t.accessibility?e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1:t.errorModuleName?e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1:void 0}(t);return void 0!==r?{diagnosticMessage:r,errorNode:n,typeName:n.name}:void 0}}e.canProduceDiagnostics=function(n){return e.isVariableDeclaration(n)||e.isPropertyDeclaration(n)||e.isPropertySignature(n)||e.isBindingElement(n)||e.isSetAccessor(n)||e.isGetAccessor(n)||e.isConstructSignatureDeclaration(n)||e.isCallSignatureDeclaration(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isFunctionDeclaration(n)||e.isParameter(n)||e.isTypeParameterDeclaration(n)||e.isExpressionWithTypeArguments(n)||e.isImportEqualsDeclaration(n)||e.isTypeAliasDeclaration(n)||e.isConstructorDeclaration(n)||e.isIndexSignatureDeclaration(n)||e.isPropertyAccessExpression(n)},e.createGetSymbolAccessibilityDiagnosticForNodeName=function(t){return e.isSetAccessor(t)||e.isGetAccessor(t)?function(n){var r=function(n){return e.hasSyntacticModifier(t,32)?n.errorModuleName?2===n.accessibility?e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1:252===t.parent.kind?n.errorModuleName?2===n.accessibility?e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1:n.errorModuleName?e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1}(n);return void 0!==r?{diagnosticMessage:r,errorNode:t,typeName:t.name}:void 0}:e.isMethodSignature(t)||e.isMethodDeclaration(t)?function(n){var r=function(n){return e.hasSyntacticModifier(t,32)?n.errorModuleName?2===n.accessibility?e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_private_name_1:252===t.parent.kind?n.errorModuleName?2===n.accessibility?e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named:e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_private_name_1:n.errorModuleName?e.Diagnostics.Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2:e.Diagnostics.Method_0_of_exported_interface_has_or_is_using_private_name_1}(n);return void 0!==r?{diagnosticMessage:r,errorNode:t,typeName:t.name}:void 0}:n(t)},e.createGetSymbolAccessibilityDiagnosticForNode=n}(ts||(ts={})),function(e){function n(n,t){var r=t.text.substring(n.pos,n.end);return e.stringContains(r,"@internal")}function t(t,r){var a=e.getParseTreeNode(t);if(a&&160===a.kind){var i=a.parent.parameters.indexOf(a),o=i>0?a.parent.parameters[i-1]:void 0,s=r.text,c=o?e.concatenate(e.getTrailingCommentRanges(s,e.skipTrivia(s,o.end+1,!1,!0)),e.getLeadingCommentRanges(s,t.pos)):e.getTrailingCommentRanges(s,e.skipTrivia(s,t.pos,!1,!0));return c&&c.length&&n(e.last(c),r)}var l=a&&e.getLeadingCommentRangesOfNode(a,r);return!!e.forEach(l,(function(e){return n(e,r)}))}e.getDeclarationDiagnostics=function(n,t,r){var i=n.getCompilerOptions();return e.transformNodes(t,n,e.factory,i,r?[r]:e.filter(n.getSourceFiles(),e.isSourceFileNotJson),[a],!1).diagnostics},e.isInternalDeclaration=t;var r=531469;function a(n){var a,s,c,l,u,d,p,m,f,_,g,y=function(){return e.Debug.fail("Diagnostic emitted without context")},v=y,h=!0,b=!1,E=!1,T=!1,S=!1,x=n.factory,L=n.getEmitHost(),A={trackSymbol:function(e,n,t){if(262144&e.flags)return;k(D.isSymbolAccessible(e,n,t,!0)),R(D.getTypeReferenceDirectivesForSymbol(e,t))},reportInaccessibleThisError:function(){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary,e.declarationNameToString(p),"this"))},reportInaccessibleUniqueSymbolError:function(){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary,e.declarationNameToString(p),"unique symbol"))},reportCyclicStructureError:function(){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialized_A_type_annotation_is_necessary,e.declarationNameToString(p)))},reportPrivateInBaseOfClassExpression:function(t){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.Property_0_of_exported_class_expression_may_not_be_private_or_protected,t))},reportLikelyUnsafeImportRequiredError:function(t){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary,e.declarationNameToString(p),t))},reportTruncationError:function(){p&&n.addDiagnostic(e.createDiagnosticForNode(p,e.Diagnostics.The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_type_annotation_is_needed))},moduleResolverHost:L,trackReferencedAmbientModule:function(n,t){var r=D.getTypeReferenceDirectivesForSymbol(t,67108863);if(e.length(r))return R(r);var a=e.getSourceFileOfNode(n);f.set(e.getOriginalNodeId(a),a)},trackExternalModuleSymbolOfImportTypeNode:function(e){b||(d||(d=[])).push(e)},reportNonlocalAugmentation:function(t,r,a){for(var i=e.find(r.declarations,(function(n){return e.getSourceFileOfNode(n)===t})),o=e.filter(a.declarations,(function(n){return e.getSourceFileOfNode(n)!==t})),s=0,c=o;s<c.length;s++){var l=c[s];n.addDiagnostic(e.addRelatedInfo(e.createDiagnosticForNode(l,e.Diagnostics.Declaration_augments_declaration_in_another_file_This_cannot_be_serialized),e.createDiagnosticForNode(i,e.Diagnostics.This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_file)))}}},D=n.getEmitResolver(),C=n.getCompilerOptions(),I=C.noResolve,N=C.stripInternal;return function(t){if(297===t.kind&&t.isDeclarationFile)return t;if(298===t.kind){b=!0,f=new e.Map,_=new e.Map;var r=!1,i=x.createBundle(e.map(t.sourceFiles,(function(t){if(!t.isDeclarationFile){if(r=r||t.hasNoDefaultLib,m=t,a=t,c=void 0,u=!1,l=new e.Map,v=y,T=!1,S=!1,P(t,f),M(t,_),e.isExternalOrCommonJsModule(t)||e.isJsonSourceFile(t)){E=!1,h=!1;var i=e.isSourceFileJS(t)?x.createNodeArray(O(t,!0)):e.visitNodes(t.statements,Z);return x.updateSourceFile(t,[x.createModuleDeclaration([],[x.createModifier(133)],x.createStringLiteral(e.getResolvedExternalModuleName(n.getEmitHost(),t)),x.createModuleBlock(e.setTextRange(x.createNodeArray(Y(i)),t.statements)))],!0,[],[],!1,[])}h=!0;var o=e.isSourceFileJS(t)?x.createNodeArray(O(t)):e.visitNodes(t.statements,Z);return x.updateSourceFile(t,Y(o),!0,[],[],!1,[])}})),e.mapDefined(t.prepends,(function(n){if(300===n.kind){var t=e.createUnparsedSourceFile(n,"dts",N);return r=r||!!t.hasNoDefaultLib,P(t,f),R(t.typeReferenceDirectives),M(t,_),t}return n})));i.syntheticFileReferences=[],i.syntheticTypeReferences=B(),i.syntheticLibReferences=G(),i.hasNoDefaultLib=r;var o=e.getDirectoryPath(e.normalizeSlashes(e.getOutputPathsFor(t,L,!0).declarationFilePath)),p=V(i.syntheticFileReferences,o);return f.forEach(p),i}h=!0,T=!1,S=!1,a=t,m=t,v=y,b=!1,E=!1,u=!1,c=void 0,l=new e.Map,s=void 0,f=P(m,new e.Map),_=M(m,new e.Map);var A,D=[],I=e.getDirectoryPath(e.normalizeSlashes(e.getOutputPathsFor(t,L,!0).declarationFilePath)),k=V(D,I);if(e.isSourceFileJS(m))A=x.createNodeArray(O(t)),f.forEach(k),g=e.filter(A,e.isAnyImportSyntax);else{var w=e.visitNodes(t.statements,Z);A=e.setTextRange(x.createNodeArray(Y(w)),t.statements),f.forEach(k),g=e.filter(A,e.isAnyImportSyntax),e.isExternalModule(t)&&(!E||T&&!S)&&(A=e.setTextRange(x.createNodeArray(__spreadArrays(A,[e.createEmptyExports(x)])),A))}var F=x.updateSourceFile(t,A,!0,D,B(),t.hasNoDefaultLib,G());return F.exportedModulesFromDeclarationEmit=d,F;function G(){return e.map(e.arrayFrom(_.keys()),(function(e){return{fileName:e,pos:-1,end:-1}}))}function B(){return s?e.mapDefined(e.arrayFrom(s.keys()),U):[]}function U(n){if(g)for(var t=0,r=g;t<r.length;t++){var a=r[t];if(e.isImportEqualsDeclaration(a)&&e.isExternalModuleReference(a.moduleReference)){var i=a.moduleReference.expression;if(e.isStringLiteralLike(i)&&i.text===n)return}else if(e.isImportDeclaration(a)&&e.isStringLiteral(a.moduleSpecifier)&&a.moduleSpecifier.text===n)return}return{fileName:n,pos:-1,end:-1}}function V(n,r){return function(a){var i;if(a.isDeclarationFile)i=a.fileName;else{if(b&&e.contains(t.sourceFiles,a))return;var o=e.getOutputPathsFor(a,L,!0);i=o.declarationFilePath||o.jsFilePath||a.fileName}if(i){var s=e.moduleSpecifiers.getModuleSpecifier(__assign(__assign({},C),{baseUrl:C.baseUrl&&e.toPath(C.baseUrl,L.getCurrentDirectory(),L.getCanonicalFileName)}),m,e.toPath(r,L.getCurrentDirectory(),L.getCanonicalFileName),e.toPath(i,L.getCurrentDirectory(),L.getCanonicalFileName),L,void 0);if(!e.pathIsRelative(s))return void R([s]);var c=e.getRelativePathToDirectoryOrUrl(r,i,L.getCurrentDirectory(),L.getCanonicalFileName,!1);if(e.startsWith(c,"./")&&e.hasExtension(c)&&(c=c.substring(2)),e.startsWith(c,"node_modules/")||e.pathContainsNodeModules(c))return;n.push({pos:-1,end:-1,fileName:c})}}}};function R(n){if(n){s=s||new e.Set;for(var t=0,r=n;t<r.length;t++){var a=r[t];s.add(a)}}}function k(t){if(0===t.accessibility){if(t&&t.aliasesToMakeVisible)if(c)for(var r=0,a=t.aliasesToMakeVisible;r<a.length;r++){var i=a[r];e.pushIfUnique(c,i)}else c=t.aliasesToMakeVisible}else{var o=v(t);o&&(o.typeName?n.addDiagnostic(e.createDiagnosticForNode(t.errorNode||o.errorNode,o.diagnosticMessage,e.getTextOfNode(o.typeName),t.errorSymbolName,t.errorModuleName)):n.addDiagnostic(e.createDiagnosticForNode(t.errorNode||o.errorNode,o.diagnosticMessage,t.errorSymbolName,t.errorModuleName)))}}function O(n,t){var a=v;v=function(t){return{diagnosticMessage:t.errorModuleName?e.Diagnostics.Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotation_may_unblock_declaration_emit:e.Diagnostics.Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_declaration_emit,errorNode:t.errorNode||n}};var i=D.getDeclarationStatementsForSourceFile(n,r,A,t);return v=a,i}function P(n,t){return I||!e.isUnparsedSource(n)&&e.isSourceFileJS(n)||e.forEach(n.referencedFiles,(function(r){var a=L.getSourceFileFromReference(n,r);a&&t.set(e.getOriginalNodeId(a),a)})),t}function M(n,t){return e.forEach(n.libReferenceDirectives,(function(n){L.getLibFileFromReference(n)&&t.set(e.toFileNameLowerCase(n.fileName),!0)})),t}function w(n){return 78===n.kind?n:197===n.kind?x.updateArrayBindingPattern(n,e.visitNodes(n.elements,t)):x.updateObjectBindingPattern(n,e.visitNodes(n.elements,t));function t(e){return 222===e.kind?e:x.updateBindingElement(e,e.dotDotDotToken,e.propertyName,w(e.name),G(e)?e.initializer:void 0)}}function F(n,t,r){var a;u||(a=v,v=e.createGetSymbolAccessibilityDiagnosticForNode(n));var o=x.updateParameterDeclaration(n,void 0,function(n,t,r){return e.factory.createModifiersFromModifierFlags(i(n,t,r))}(n,t),n.dotDotDotToken,w(n.name),D.isOptionalParameter(n)?n.questionToken||x.createToken(57):void 0,U(n,r||n.type,!0),B(n));return u||(v=a),o}function G(n){return function(n){switch(n.kind){case 163:case 162:return!e.hasEffectiveModifier(n,8);case 160:case 249:return!0}return!1}(n)&&D.isLiteralConstDeclaration(e.getParseTreeNode(n))}function B(n){if(G(n))return D.createLiteralConstValue(e.getParseTreeNode(n),A)}function U(n,t,i){if((i||!e.hasEffectiveModifier(n,8))&&!G(n)){var o,s=160===n.kind&&(D.isRequiredInitializedParameter(n)||D.isOptionalUninitializedParameterProperty(n));return t&&!s?e.visitNode(t,Q):e.getParseTreeNode(n)?168===n.kind?x.createKeywordTypeNode(128):(p=n.name,u||(o=v,v=e.createGetSymbolAccessibilityDiagnosticForNode(n)),249===n.kind||198===n.kind?c(D.createTypeOfDeclaration(n,a,r,A)):160===n.kind||163===n.kind||162===n.kind?n.initializer?c(D.createTypeOfDeclaration(n,a,r,A,s)||D.createTypeOfExpression(n.initializer,a,r,A)):c(D.createTypeOfDeclaration(n,a,r,A,s)):c(D.createReturnTypeOfSignatureDeclaration(n,a,r,A))):t?e.visitNode(t,Q):x.createKeywordTypeNode(128)}function c(e){return p=void 0,u||(v=o),e||x.createKeywordTypeNode(128)}}function V(n){switch((n=e.getParseTreeNode(n)).kind){case 251:case 256:case 253:case 252:case 254:case 255:return!D.isDeclarationVisible(n);case 249:return!j(n);case 260:case 261:case 267:case 266:return!1}return!1}function j(n){return!e.isOmittedExpression(n)&&(e.isBindingPattern(n.name)?e.some(n.name.elements,j):D.isDeclarationVisible(n))}function H(n,t,r){if(!e.hasEffectiveModifier(n,8)){var a=e.map(t,(function(e){return F(e,r)}));if(a)return x.createNodeArray(a,t.hasTrailingComma)}}function K(n,t){var r;if(!t){var a=e.getThisParameter(n);a&&(r=[F(a)])}if(e.isSetAccessorDeclaration(n)){var i=void 0;if(!t){var o=e.getSetAccessorValueParameter(n);if(o)i=F(o,void 0,ie(n,D.getAllAccessorDeclarations(n)))}i||(i=x.createParameterDeclaration(void 0,void 0,void 0,"value")),r=e.append(r,i)}return x.createNodeArray(r||e.emptyArray)}function W(n,t){return e.hasEffectiveModifier(n,8)?void 0:e.visitNodes(t,Q)}function z(n){return e.isSourceFile(n)||e.isTypeAliasDeclaration(n)||e.isModuleDeclaration(n)||e.isClassDeclaration(n)||e.isInterfaceDeclaration(n)||e.isFunctionLike(n)||e.isIndexSignatureDeclaration(n)||e.isMappedTypeNode(n)}function q(e,n){k(D.isEntityNameVisible(e,n)),R(D.getTypeReferenceDirectivesForEntityName(e))}function J(n,t){return e.hasJSDocNodes(n)&&e.hasJSDocNodes(t)&&(n.jsDoc=t.jsDoc),e.setCommentRange(n,e.getCommentRange(t))}function X(t,r){if(r){if(E=E||256!==t.kind&&195!==t.kind,e.isStringLiteralLike(r))if(b){var a=e.getExternalModuleNameFromDeclaration(n.getEmitHost(),D,t);if(a)return x.createStringLiteral(a)}else{var i=D.getSymbolOfExternalModuleSpecifier(r);i&&(d||(d=[])).push(i)}return r}}function Y(n){for(;e.length(c);){var t=c.shift();if(!e.isLateVisibilityPaintedStatement(t))return e.Debug.fail("Late replaced statement was found which is not handled by the declaration transformer!: "+(e.SyntaxKind?e.SyntaxKind[t.kind]:t.kind));var r=h;h=t.parent&&e.isSourceFile(t.parent)&&!(e.isExternalModule(t.parent)&&b);var a=ee(t);h=r,l.set(e.getOriginalNodeId(t),a)}return e.visitNodes(n,(function(n){if(e.isLateVisibilityPaintedStatement(n)){var t=e.getOriginalNodeId(n);if(l.has(t)){var r=l.get(t);return l.delete(t),r&&((e.isArray(r)?e.some(r,e.needsScopeMarker):e.needsScopeMarker(r))&&(T=!0),e.isSourceFile(n.parent)&&(e.isArray(r)?e.some(r,e.isExternalModuleIndicator):e.isExternalModuleIndicator(r))&&(E=!0)),r}}return n}))}function Q(t){if(!te(t)){if(e.isDeclaration(t)){if(V(t))return;if(e.hasDynamicName(t)&&!D.isLateBound(e.getParseTreeNode(t)))return}if(!(e.isFunctionLike(t)&&D.isImplementationOfOverload(t)||e.isSemicolonClassElement(t))){var r;z(t)&&(r=a,a=t);var i=v,o=e.canProduceDiagnostics(t),s=u,c=(177===t.kind||190===t.kind)&&254!==t.parent.kind;if((e.isMethodDeclaration(t)||e.isMethodSignature(t))&&e.hasEffectiveModifier(t,8)){if(t.symbol&&t.symbol.declarations&&t.symbol.declarations[0]!==t)return;return b(x.createPropertyDeclaration(void 0,ae(t),t.name,void 0,void 0,void 0))}if(o&&!u&&(v=e.createGetSymbolAccessibilityDiagnosticForNode(t)),e.isTypeQueryNode(t)&&q(t.exprName,a),c&&(u=!0),function(e){switch(e.kind){case 170:case 166:case 165:case 167:case 168:case 163:case 162:case 164:case 169:case 171:case 249:case 159:case 223:case 173:case 184:case 174:case 175:case 195:return!0}return!1}(t))switch(t.kind){case 223:(e.isEntityName(t.expression)||e.isEntityNameExpression(t.expression))&&q(t.expression,a);var l=e.visitEachChild(t,Q,n);return b(x.updateExpressionWithTypeArguments(l,l.expression,l.typeArguments));case 173:q(t.typeName,a);l=e.visitEachChild(t,Q,n);return b(x.updateTypeReferenceNode(l,l.typeName,l.typeArguments));case 170:return b(x.updateConstructSignature(t,W(t,t.typeParameters),H(t,t.parameters),U(t,t.type)));case 166:return b(x.createConstructorDeclaration(void 0,ae(t),H(t,t.parameters,0),void 0));case 165:return e.isPrivateIdentifier(t.name)?b(void 0):b(x.createMethodDeclaration(void 0,ae(t),void 0,t.name,t.questionToken,W(t,t.typeParameters),H(t,t.parameters),U(t,t.type),void 0));case 167:if(e.isPrivateIdentifier(t.name))return b(void 0);var d=ie(t,D.getAllAccessorDeclarations(t));return b(x.updateGetAccessorDeclaration(t,void 0,ae(t),t.name,K(t,e.hasEffectiveModifier(t,8)),U(t,d),void 0));case 168:return e.isPrivateIdentifier(t.name)?b(void 0):b(x.updateSetAccessorDeclaration(t,void 0,ae(t),t.name,K(t,e.hasEffectiveModifier(t,8)),void 0));case 163:return e.isPrivateIdentifier(t.name)?b(void 0):b(x.updatePropertyDeclaration(t,void 0,ae(t),t.name,t.questionToken,U(t,t.type),B(t)));case 162:return e.isPrivateIdentifier(t.name)?b(void 0):b(x.updatePropertySignature(t,ae(t),t.name,t.questionToken,U(t,t.type)));case 164:return e.isPrivateIdentifier(t.name)?b(void 0):b(x.updateMethodSignature(t,ae(t),t.name,t.questionToken,W(t,t.typeParameters),H(t,t.parameters),U(t,t.type)));case 169:return b(x.updateCallSignature(t,W(t,t.typeParameters),H(t,t.parameters),U(t,t.type)));case 171:return b(x.updateIndexSignature(t,void 0,ae(t),H(t,t.parameters),e.visitNode(t.type,Q)||x.createKeywordTypeNode(128)));case 249:return e.isBindingPattern(t.name)?ne(t.name):(c=!0,u=!0,b(x.updateVariableDeclaration(t,t.name,void 0,U(t,t.type),B(t))));case 159:return function(n){return 165===n.parent.kind&&e.hasEffectiveModifier(n.parent,8)}(t)&&(t.default||t.constraint)?b(x.updateTypeParameterDeclaration(t,t.name,void 0,void 0)):b(e.visitEachChild(t,Q,n));case 184:var f=e.visitNode(t.checkType,Q),_=e.visitNode(t.extendsType,Q),g=a;a=t.trueType;var y=e.visitNode(t.trueType,Q);a=g;var h=e.visitNode(t.falseType,Q);return b(x.updateConditionalTypeNode(t,f,_,y,h));case 174:return b(x.updateFunctionTypeNode(t,e.visitNodes(t.typeParameters,Q),H(t,t.parameters),e.visitNode(t.type,Q)));case 175:return b(x.updateConstructorTypeNode(t,e.visitNodes(t.typeParameters,Q),H(t,t.parameters),e.visitNode(t.type,Q)));case 195:return e.isLiteralImportTypeNode(t)?b(x.updateImportTypeNode(t,x.updateLiteralTypeNode(t.argument,X(t,t.argument.literal)),t.qualifier,e.visitNodes(t.typeArguments,Q,e.isTypeNode),t.isTypeOf)):b(t);default:e.Debug.assertNever(t,"Attempted to process unhandled node kind: "+e.SyntaxKind[t.kind])}return e.isTupleTypeNode(t)&&e.getLineAndCharacterOfPosition(m,t.pos).line===e.getLineAndCharacterOfPosition(m,t.end).line&&e.setEmitFlags(t,1),b(e.visitEachChild(t,Q,n))}}function b(n){return n&&o&&e.hasDynamicName(t)&&function(n){var t;u||(t=v,v=e.createGetSymbolAccessibilityDiagnosticForNodeName(n));p=n.name,e.Debug.assert(D.isLateBound(e.getParseTreeNode(n))),q(n.name.expression,a),u||(v=t);p=void 0}(t),z(t)&&(a=r),o&&!u&&(v=i),c&&(u=s),n===t?n:n&&e.setOriginalNode(J(n,t),t)}}function Z(n){if(function(e){switch(e.kind){case 251:case 256:case 260:case 253:case 252:case 254:case 255:case 232:case 261:case 267:case 266:return!0}return!1}(n)&&!te(n)){switch(n.kind){case 267:return e.isSourceFile(n.parent)&&(E=!0),S=!0,x.updateExportDeclaration(n,void 0,n.modifiers,n.isTypeOnly,n.exportClause,X(n,n.moduleSpecifier));case 266:if(e.isSourceFile(n.parent)&&(E=!0),S=!0,78===n.expression.kind)return n;var t=x.createUniqueName("_default",16);v=function(){return{diagnosticMessage:e.Diagnostics.Default_export_of_the_module_has_or_is_using_private_name_0,errorNode:n}};var a=x.createVariableDeclaration(t,void 0,D.createTypeOfExpression(n.expression,n,r,A),void 0);return[x.createVariableStatement(h?[x.createModifier(133)]:[],x.createVariableDeclarationList([a],2)),x.updateExportAssignment(n,n.decorators,n.modifiers,t)]}var i=ee(n);return l.set(e.getOriginalNodeId(n),i),n}}function $(n){if(e.isImportEqualsDeclaration(n)||e.hasEffectiveModifier(n,512)||!e.canHaveModifiers(n))return n;var t=x.createModifiersFromModifierFlags(11262&e.getEffectiveModifierFlags(n));return x.updateModifiers(n,t)}function ee(n){if(!te(n)){switch(n.kind){case 260:return function(n){if(D.isDeclarationVisible(n)){if(272===n.moduleReference.kind){var t=e.getExternalModuleImportEqualsDeclarationExpression(n);return x.updateImportEqualsDeclaration(n,void 0,n.modifiers,n.name,x.updateExternalModuleReference(n.moduleReference,X(n,t)))}var r=v;return v=e.createGetSymbolAccessibilityDiagnosticForNode(n),q(n.moduleReference,a),v=r,n}}(n);case 261:return function(n){if(!n.importClause)return x.updateImportDeclaration(n,void 0,n.modifiers,n.importClause,X(n,n.moduleSpecifier));var t=n.importClause&&n.importClause.name&&D.isDeclarationVisible(n.importClause)?n.importClause.name:void 0;if(!n.importClause.namedBindings)return t&&x.updateImportDeclaration(n,void 0,n.modifiers,x.updateImportClause(n.importClause,n.importClause.isTypeOnly,t,void 0),X(n,n.moduleSpecifier));if(263===n.importClause.namedBindings.kind){var r=D.isDeclarationVisible(n.importClause.namedBindings)?n.importClause.namedBindings:void 0;return t||r?x.updateImportDeclaration(n,void 0,n.modifiers,x.updateImportClause(n.importClause,n.importClause.isTypeOnly,t,r),X(n,n.moduleSpecifier)):void 0}var a=e.mapDefined(n.importClause.namedBindings.elements,(function(e){return D.isDeclarationVisible(e)?e:void 0}));return a&&a.length||t?x.updateImportDeclaration(n,void 0,n.modifiers,x.updateImportClause(n.importClause,n.importClause.isTypeOnly,t,a&&a.length?x.updateNamedImports(n.importClause.namedBindings,a):void 0),X(n,n.moduleSpecifier)):D.isImportRequiredByAugmentation(n)?x.updateImportDeclaration(n,void 0,n.modifiers,void 0,X(n,n.moduleSpecifier)):void 0}(n)}if(!(e.isDeclaration(n)&&V(n)||e.isFunctionLike(n)&&D.isImplementationOfOverload(n))){var t;z(n)&&(t=a,a=n);var i=e.canProduceDiagnostics(n),o=v;i&&(v=e.createGetSymbolAccessibilityDiagnosticForNode(n));var s=h;switch(n.kind){case 254:return de(x.updateTypeAliasDeclaration(n,void 0,ae(n),n.name,e.visitNodes(n.typeParameters,Q,e.isTypeParameterDeclaration),e.visitNode(n.type,Q,e.isTypeNode)));case 253:return de(x.updateInterfaceDeclaration(n,void 0,ae(n),n.name,W(n,n.typeParameters),oe(n.heritageClauses),e.visitNodes(n.members,Q)));case 251:var c=de(x.updateFunctionDeclaration(n,void 0,ae(n),void 0,n.name,W(n,n.typeParameters),H(n,n.parameters),U(n,n.type),void 0));if(c&&D.isExpandoFunctionDeclaration(n)){var u=D.getPropertiesOfContainerFunction(n),d=e.parseNodeFactory.createModuleDeclaration(void 0,void 0,c.name||x.createIdentifier("_default"),x.createModuleBlock([]),16);e.setParent(d,a),d.locals=e.createSymbolTable(u),d.symbol=u[0].parent;var p=[],m=e.mapDefined(u,(function(n){if(e.isPropertyAccessExpression(n.valueDeclaration)){v=e.createGetSymbolAccessibilityDiagnosticForNode(n.valueDeclaration);var t=D.createTypeOfDeclaration(n.valueDeclaration,d,r,A);v=o;var a=e.unescapeLeadingUnderscores(n.escapedName),i=e.isStringANonContextualKeyword(a),s=i?x.getGeneratedNameForNode(n.valueDeclaration):x.createIdentifier(a);i&&p.push([s,a]);var c=x.createVariableDeclaration(s,void 0,t,void 0);return x.createVariableStatement(i?void 0:[x.createToken(92)],x.createVariableDeclarationList([c]))}}));p.length?m.push(x.createExportDeclaration(void 0,void 0,!1,x.createNamedExports(e.map(p,(function(e){var n=e[0],t=e[1];return x.createExportSpecifier(n,t)}))))):m=e.mapDefined(m,(function(e){return x.updateModifiers(e,0)}));var f=x.createModuleDeclaration(void 0,ae(n),n.name,x.createModuleBlock(m),16);if(!e.hasEffectiveModifier(c,512))return[c,f];var _=x.createModifiersFromModifierFlags(-514&e.getEffectiveModifierFlags(c)|2),g=x.updateFunctionDeclaration(c,void 0,_,void 0,c.name,c.typeParameters,c.parameters,c.type,void 0),y=x.updateModuleDeclaration(f,void 0,_,f.name,f.body),b=x.createExportAssignment(void 0,void 0,!1,f.name);return e.isSourceFile(n.parent)&&(E=!0),S=!0,[g,y,b]}return c;case 256:h=!1;var L=n.body;if(L&&257===L.kind){var C=T,I=S;S=!1,T=!1;var N=Y(e.visitNodes(L.statements,Z));8388608&n.flags&&(T=!1),e.isGlobalScopeAugmentation(n)||function(n){return e.some(n,re)}(N)||S||(N=T?x.createNodeArray(__spreadArrays(N,[e.createEmptyExports(x)])):e.visitNodes(N,$));var R=x.updateModuleBlock(L,N);h=s,T=C,S=I;var k=ae(n);return de(x.updateModuleDeclaration(n,void 0,k,e.isExternalModuleAugmentation(n)?X(n,n.name):n.name,R))}h=s;k=ae(n);h=!1,e.visitNode(L,Z);var O=e.getOriginalNodeId(L);R=l.get(O);return l.delete(O),de(x.updateModuleDeclaration(n,void 0,k,n.name,R));case 252:_=x.createNodeArray(ae(n));var P=W(n,n.typeParameters),M=e.getFirstConstructorWithBody(n),w=void 0;if(M){var F=v;w=e.compact(e.flatMap(M.parameters,(function(n){if(e.hasSyntacticModifier(n,92)&&!te(n))return v=e.createGetSymbolAccessibilityDiagnosticForNode(n),78===n.name.kind?J(x.createPropertyDeclaration(void 0,ae(n),n.name,n.questionToken,U(n,n.type),B(n)),n):function t(r){for(var a,i=0,o=r.elements;i<o.length;i++){var s=o[i];e.isOmittedExpression(s)||(e.isBindingPattern(s.name)&&(a=e.concatenate(a,t(s.name))),(a=a||[]).push(x.createPropertyDeclaration(void 0,ae(n),s.name,void 0,U(s,void 0),void 0)))}return a}(n.name)}))),v=F}var G=e.some(n.members,(function(n){return!!n.name&&e.isPrivateIdentifier(n.name)}))?[x.createPropertyDeclaration(void 0,void 0,x.createPrivateIdentifier("#private"),void 0,void 0,void 0)]:void 0,K=e.concatenate(e.concatenate(G,w),e.visitNodes(n.members,Q)),ee=x.createNodeArray(K),ne=e.getEffectiveBaseTypeNode(n);if(ne&&!e.isEntityNameExpression(ne.expression)&&103!==ne.expression.kind){var ie=n.name?e.unescapeLeadingUnderscores(n.name.escapedText):"default",se=x.createUniqueName(ie+"_base",16);v=function(){return{diagnosticMessage:e.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1,errorNode:ne,typeName:n.name}};var ce=x.createVariableDeclaration(se,void 0,D.createTypeOfExpression(ne.expression,n,r,A),void 0),le=x.createVariableStatement(h?[x.createModifier(133)]:[],x.createVariableDeclarationList([ce],2)),ue=x.createNodeArray(e.map(n.heritageClauses,(function(n){if(93===n.token){var t=v;v=e.createGetSymbolAccessibilityDiagnosticForNode(n.types[0]);var r=x.updateHeritageClause(n,e.map(n.types,(function(n){return x.updateExpressionWithTypeArguments(n,se,e.visitNodes(n.typeArguments,Q))})));return v=t,r}return x.updateHeritageClause(n,e.visitNodes(x.createNodeArray(e.filter(n.types,(function(n){return e.isEntityNameExpression(n.expression)||103===n.expression.kind}))),Q))})));return[le,de(x.updateClassDeclaration(n,void 0,_,n.name,P,ue,ee))]}ue=oe(n.heritageClauses);return de(x.updateClassDeclaration(n,void 0,_,n.name,P,ue,ee));case 232:return de(function(n){if(!e.forEach(n.declarationList.declarations,j))return;var t=e.visitNodes(n.declarationList.declarations,Q);if(!e.length(t))return;return x.updateVariableStatement(n,x.createNodeArray(ae(n)),x.updateVariableDeclarationList(n.declarationList,t))}(n));case 255:return de(x.updateEnumDeclaration(n,void 0,x.createNodeArray(ae(n)),n.name,x.createNodeArray(e.mapDefined(n.members,(function(e){if(!te(e)){var n=D.getConstantValue(e);return J(x.updateEnumMember(e,e.name,void 0!==n?"string"==typeof n?x.createStringLiteral(n):x.createNumericLiteral(n):void 0),e)}})))))}return e.Debug.assertNever(n,"Unhandled top-level node in declaration emit: "+e.SyntaxKind[n.kind])}}function de(r){return z(n)&&(a=t),i&&(v=o),256===n.kind&&(h=s),r===n?r:r&&e.setOriginalNode(J(r,n),n)}}function ne(n){return e.flatten(e.mapDefined(n.elements,(function(n){return function(n){if(222===n.kind)return;if(n.name){if(!j(n))return;return e.isBindingPattern(n.name)?ne(n.name):x.createVariableDeclaration(n.name,void 0,U(n,void 0),void 0)}}(n)})))}function te(e){return!!N&&!!e&&t(e,m)}function re(n){return e.isExportAssignment(n)||e.isExportDeclaration(n)}function ae(n){var t=e.getEffectiveModifierFlags(n),r=function(n){var t=11003,r=h&&!function(e){if(253===e.kind)return!0;return!1}(n)?2:0,a=297===n.parent.kind;(!a||b&&a&&e.isExternalModule(n.parent))&&(t^=2,r=0);return i(n,t,r)}(n);return t===r?n.modifiers:x.createModifiersFromModifierFlags(r)}function ie(n,t){var r=o(n);return r||n===t.firstAccessor||(r=o(t.firstAccessor),v=e.createGetSymbolAccessibilityDiagnosticForNode(t.firstAccessor)),!r&&t.secondAccessor&&n!==t.secondAccessor&&(r=o(t.secondAccessor),v=e.createGetSymbolAccessibilityDiagnosticForNode(t.secondAccessor)),r}function oe(n){return x.createNodeArray(e.filter(e.map(n,(function(n){return x.updateHeritageClause(n,e.visitNodes(x.createNodeArray(e.filter(n.types,(function(t){return e.isEntityNameExpression(t.expression)||93===n.token&&103===t.expression.kind}))),Q))})),(function(e){return e.types&&!!e.types.length})))}}function i(n,t,r){void 0===t&&(t=11259),void 0===r&&(r=0);var a=e.getEffectiveModifierFlags(n)&t|r;return 512&a&&!(1&a)&&(a^=1),512&a&&2&a&&(a^=2),a}function o(e){if(e)return 167===e.kind?e.type:e.parameters.length>0?e.parameters[0].type:void 0}e.transformDeclarations=a}(ts||(ts={})),function(e){var n,t;function r(n,t,r){if(r)return e.emptyArray;var a=e.getEmitScriptTarget(n),i=e.getEmitModuleKind(n),s=[];return e.addRange(s,t&&e.map(t.before,o)),s.push(e.transformTypeScript),s.push(e.transformClassFields),e.getJSXTransformEnabled(n)&&s.push(e.transformJsx),a<99&&s.push(e.transformESNext),a<7&&s.push(e.transformES2020),a<6&&s.push(e.transformES2019),a<5&&s.push(e.transformES2018),a<4&&s.push(e.transformES2017),a<3&&s.push(e.transformES2016),a<2&&(s.push(e.transformES2015),s.push(e.transformGenerators)),s.push(function(n){switch(n){case e.ModuleKind.ESNext:case e.ModuleKind.ES2020:case e.ModuleKind.ES2015:return e.transformECMAScriptModule;case e.ModuleKind.System:return e.transformSystemModule;default:return e.transformModule}}(i)),a<1&&s.push(e.transformES5),e.addRange(s,t&&e.map(t.after,o)),s}function a(n){var t=[];return t.push(e.transformDeclarations),e.addRange(t,n&&e.map(n.afterDeclarations,s)),t}function i(n,t){return function(r){var a=n(r);return"function"==typeof a?t(r,a):function(n){return function(t){return e.isBundle(t)?n.transformBundle(t):n.transformSourceFile(t)}}(a)}}function o(n){return i(n,e.chainBundle)}function s(e){return i(e,(function(e,n){return n}))}function c(e,n){return n}function l(e,n,t){t(e,n)}!function(e){e[e.Uninitialized=0]="Uninitialized",e[e.Initialized=1]="Initialized",e[e.Completed=2]="Completed",e[e.Disposed=3]="Disposed"}(n||(n={})),function(e){e[e.Substitution=1]="Substitution",e[e.EmitNotifications=2]="EmitNotifications"}(t||(t={})),e.noTransformers={scriptTransformers:e.emptyArray,declarationTransformers:e.emptyArray},e.getTransformers=function(e,n,t){return{scriptTransformers:r(e,n,t),declarationTransformers:a(n)}},e.noEmitSubstitution=c,e.noEmitNotification=l,e.transformNodes=function(n,t,r,a,i,o,s){for(var u,d,p,m,f=new Array(341),_=0,g=[],y=[],v=[],h=[],b=0,E=!1,T=c,S=l,x=0,L=[],A={factory:r,getCompilerOptions:function(){return a},getEmitResolver:function(){return n},getEmitHost:function(){return t},getEmitHelperFactory:e.memoize((function(){return e.createEmitHelperFactory(A)})),startLexicalEnvironment:function(){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!E,"Lexical environment is suspended."),g[b]=u,y[b]=d,v[b]=p,h[b]=_,b++,u=void 0,d=void 0,p=void 0,_=0},suspendLexicalEnvironment:function(){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!E,"Lexical environment is already suspended."),E=!0},resumeLexicalEnvironment:function(){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(E,"Lexical environment is not suspended."),E=!1},endLexicalEnvironment:function(){var n;if(e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.Debug.assert(!E,"Lexical environment is suspended."),u||d||p){if(d&&(n=__spreadArrays(d)),u){var t=r.createVariableStatement(void 0,r.createVariableDeclarationList(u));e.setEmitFlags(t,1048576),n?n.push(t):n=[t]}p&&(n=n?__spreadArrays(n,p):__spreadArrays(p))}b--,u=g[b],d=y[b],p=v[b],_=h[b],0===b&&(g=[],y=[],v=[],h=[]);return n},setLexicalEnvironmentFlags:function(e,n){_=n?_|e:_&~e},getLexicalEnvironmentFlags:function(){return _},hoistVariableDeclaration:function(n){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed.");var t=e.setEmitFlags(r.createVariableDeclaration(n),64);u?u.push(t):u=[t];1&_&&(_|=2)},hoistFunctionDeclaration:function(n){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.setEmitFlags(n,1048576),d?d.push(n):d=[n]},addInitializationStatement:function(n){e.Debug.assert(x>0,"Cannot modify the lexical environment during initialization."),e.Debug.assert(x<2,"Cannot modify the lexical environment after transformation has completed."),e.setEmitFlags(n,1048576),p?p.push(n):p=[n]},requestEmitHelper:function n(t){if(e.Debug.assert(x>0,"Cannot modify the transformation context during initialization."),e.Debug.assert(x<2,"Cannot modify the transformation context after transformation has completed."),e.Debug.assert(!t.scoped,"Cannot request a scoped emit helper."),t.dependencies)for(var r=0,a=t.dependencies;r<a.length;r++){var i=a[r];n(i)}m=e.append(m,t)},readEmitHelpers:function(){e.Debug.assert(x>0,"Cannot modify the transformation context during initialization."),e.Debug.assert(x<2,"Cannot modify the transformation context after transformation has completed.");var n=m;return m=void 0,n},enableSubstitution:function(n){e.Debug.assert(x<2,"Cannot modify the transformation context after transformation has completed."),f[n]|=1},enableEmitNotification:function(n){e.Debug.assert(x<2,"Cannot modify the transformation context after transformation has completed."),f[n]|=2},isSubstitutionEnabled:w,isEmitNotificationEnabled:F,get onSubstituteNode(){return T},set onSubstituteNode(n){e.Debug.assert(x<1,"Cannot modify transformation hooks after initialization has completed."),e.Debug.assert(void 0!==n,"Value must not be 'undefined'"),T=n},get onEmitNode(){return S},set onEmitNode(n){e.Debug.assert(x<1,"Cannot modify transformation hooks after initialization has completed."),e.Debug.assert(void 0!==n,"Value must not be 'undefined'"),S=n},addDiagnostic:function(e){L.push(e)}},D=0,C=i;D<C.length;D++){var I=C[D];e.disposeEmitNodes(e.getSourceFileOfNode(e.getParseTreeNode(I)))}e.performance.mark("beforeTransform");var N=o.map((function(e){return e(A)})),R=function(e){for(var n=0,t=N;n<t.length;n++){e=(0,t[n])(e)}return e};x=1;for(var k=[],O=0,P=i;O<P.length;O++){I=P[O];e.tracing.push("emit","transformNodes",297===I.kind?{path:I.path}:{kind:I.kind,pos:I.pos,end:I.end}),k.push((s?R:M)(I)),e.tracing.pop()}return x=2,e.performance.mark("afterTransform"),e.performance.measure("transformTime","beforeTransform","afterTransform"),{transformed:k,substituteNode:function(n,t){return e.Debug.assert(x<3,"Cannot substitute a node after the result is disposed."),t&&w(t)&&T(n,t)||t},emitNodeWithNotification:function(n,t,r){e.Debug.assert(x<3,"Cannot invoke TransformationResult callbacks after the result is disposed."),t&&(F(t)?S(n,t,r):r(n,t))},isEmitNotificationEnabled:F,dispose:function(){if(x<3){for(var n=0,t=i;n<t.length;n++){var r=t[n];e.disposeEmitNodes(e.getSourceFileOfNode(e.getParseTreeNode(r)))}u=void 0,g=void 0,d=void 0,y=void 0,T=void 0,S=void 0,m=void 0,x=3}},diagnostics:L};function M(n){return!n||e.isSourceFile(n)&&n.isDeclarationFile?n:R(n)}function w(n){return 0!=(1&f[n.kind])&&0==(4&e.getEmitFlags(n))}function F(n){return 0!=(2&f[n.kind])||0!=(2&e.getEmitFlags(n))}},e.nullTransformationContext={get factory(){return e.factory},enableEmitNotification:e.noop,enableSubstitution:e.noop,endLexicalEnvironment:e.returnUndefined,getCompilerOptions:function(){return{}},getEmitHost:e.notImplemented,getEmitResolver:e.notImplemented,getEmitHelperFactory:e.notImplemented,setLexicalEnvironmentFlags:e.noop,getLexicalEnvironmentFlags:function(){return 0},hoistFunctionDeclaration:e.noop,hoistVariableDeclaration:e.noop,addInitializationStatement:e.noop,isEmitNotificationEnabled:e.notImplemented,isSubstitutionEnabled:e.notImplemented,onEmitNode:e.noop,onSubstituteNode:e.notImplemented,readEmitHelpers:e.notImplemented,requestEmitHelper:e.noop,resumeLexicalEnvironment:e.noop,startLexicalEnvironment:e.noop,suspendLexicalEnvironment:e.noop,addDiagnostic:e.noop}}(ts||(ts={})),function(e){var n,t,r=function(){var e=[];return e[1024]=["{","}"],e[2048]=["(",")"],e[4096]=["<",">"],e[8192]=["[","]"],e}(),a={pos:-1,end:-1};function i(n,t,r,a,i,s){void 0===a&&(a=!1);var l=e.isArray(r)?r:e.getSourceFilesToEmit(n,r,a),u=n.getCompilerOptions();if(e.outFile(u)){var d=n.getPrependNodes();if(l.length||d.length){var p=e.factory.createBundle(l,d);if(_=t(c(p,n,a),p))return _}}else{if(!i)for(var m=0,f=l;m<f.length;m++){var _,g=f[m];if(_=t(c(g,n,a),g))return _}if(s){var y=o(u);if(y)return t({buildInfoPath:y},void 0)}}}function o(n){var t=n.configFilePath;if(e.isIncrementalCompilation(n)){if(n.tsBuildInfoFile)return n.tsBuildInfoFile;var r,a=e.outFile(n);if(a)r=e.removeFileExtension(a);else{if(!t)return;var i=e.removeFileExtension(t);r=n.outDir?n.rootDir?e.resolvePath(n.outDir,e.getRelativePathFromDirectory(n.rootDir,i,!0)):e.combinePaths(n.outDir,e.getBaseFileName(i)):i}return r+".tsbuildinfo"}}function s(n,t){var r=e.outFile(n),a=n.emitDeclarationOnly?void 0:r,i=a&&l(a,n),s=t||e.getEmitDeclarations(n)?e.removeFileExtension(r)+".d.ts":void 0;return{jsFilePath:a,sourceMapFilePath:i,declarationFilePath:s,declarationMapPath:s&&e.getAreDeclarationMapsEnabled(n)?s+".map":void 0,buildInfoPath:o(n)}}function c(n,t,r){var a=t.getCompilerOptions();if(298===n.kind)return s(a,r);var i=e.getOwnEmitOutputFilePath(n.fileName,t,u(n,a)),o=e.isJsonSourceFile(n),c=o&&0===e.comparePaths(n.fileName,i,t.getCurrentDirectory(),!t.useCaseSensitiveFileNames()),d=a.emitDeclarationOnly||c?void 0:i,p=!d||e.isJsonSourceFile(n)?void 0:l(d,a),m=r||e.getEmitDeclarations(a)&&!o?e.getDeclarationEmitOutputFilePath(n.fileName,t):void 0;return{jsFilePath:d,sourceMapFilePath:p,declarationFilePath:m,declarationMapPath:m&&e.getAreDeclarationMapsEnabled(a)?m+".map":void 0,buildInfoPath:void 0}}function l(e,n){return n.sourceMap&&!n.inlineSourceMap?e+".map":void 0}function u(n,t){if(e.isJsonSourceFile(n))return".json";if(1===t.jsx)if(e.isSourceFileJS(n)){if(e.fileExtensionIs(n.fileName,".jsx"))return".jsx"}else if(1===n.languageVariant)return".jsx";return".js"}function d(n,t,r,a){return a?e.resolvePath(a,e.getRelativePathFromDirectory(function(n){return n.options.rootDir||e.getDirectoryPath(e.Debug.checkDefined(n.options.configFilePath))}(t),n,r)):n}function p(n,t,r){return e.Debug.assert(!e.fileExtensionIs(n,".d.ts")&&!e.fileExtensionIs(n,".json")),e.changeExtension(d(n,t,r,t.options.declarationDir||t.options.outDir),".d.ts")}function m(n,t,r){if(!t.options.emitDeclarationOnly){var a=e.fileExtensionIs(n,".json"),i=e.changeExtension(d(n,t,r,t.options.outDir),a?".json":e.fileExtensionIs(n,".tsx")&&1===t.options.jsx?".jsx":".js");return a&&0===e.comparePaths(n,i,e.Debug.checkDefined(t.options.configFilePath),r)?void 0:i}}function f(){var n;return{addOutput:function(e){e&&(n||(n=[])).push(e)},getOutputs:function(){return n||e.emptyArray}}}function _(e,n){var t=s(e.options,!1),r=t.jsFilePath,a=t.sourceMapFilePath,i=t.declarationFilePath,o=t.declarationMapPath,c=t.buildInfoPath;n(r),n(a),n(i),n(o),n(c)}function g(n,t,r,a){if(!e.fileExtensionIs(t,".d.ts")){var i=m(t,n,r);if(a(i),!e.fileExtensionIs(t,".json")&&(i&&n.options.sourceMap&&a(i+".map"),e.getEmitDeclarations(n.options))){var o=p(t,n,r);a(o),n.options.declarationMap&&a(o+".map")}}}function y(n,t,r,a,o,s,c){var l,u,d=a.scriptTransformers,p=a.declarationTransformers,m=t.getCompilerOptions(),f=m.sourceMap||m.inlineSourceMap||e.getAreDeclarationMapsEnabled(m)?[]:void 0,_=m.listEmittedFiles?[]:void 0,g=e.createDiagnosticCollection(),y=e.getNewLineCharacter(m,(function(){return t.getNewLine()})),h=e.createTextWriter(y),E=e.performance.createTimer("printTime","beforePrint","afterPrint"),T=E.enter,S=E.exit,x=!1;return T(),i(t,(function(a,i){var s,f=a.jsFilePath,y=a.sourceMapFilePath,h=a.declarationFilePath,E=a.declarationMapPath,T=a.buildInfoPath;T&&i&&e.isBundle(i)&&(s=e.getDirectoryPath(e.getNormalizedAbsolutePath(T,t.getCurrentDirectory())),l={commonSourceDirectory:S(t.getCommonSourceDirectory()),sourceFiles:i.sourceFiles.map((function(n){return S(e.getNormalizedAbsolutePath(n.fileName,t.getCurrentDirectory()))}))});e.tracing.push("emit","emitJsFileOrBundle",{jsFilePath:f}),function(r,a,i,s){if(!r||o||!a)return;if(a&&t.isEmitBlocked(a)||m.noEmit)return void(x=!0);var c=e.transformNodes(n,t,e.factory,m,[r],d,!1),u=b({removeComments:m.removeComments,newLine:m.newLine,noEmitHelpers:m.noEmitHelpers,module:m.module,target:m.target,sourceMap:m.sourceMap,inlineSourceMap:m.inlineSourceMap,inlineSources:m.inlineSources,extendedDiagnostics:m.extendedDiagnostics,writeBundleFileInfo:!!l,relativeToBuildInfo:s},{hasGlobalName:n.hasGlobalName,onEmitNode:c.emitNodeWithNotification,isEmitNotificationEnabled:c.isEmitNotificationEnabled,substituteNode:c.substituteNode});e.Debug.assert(1===c.transformed.length,"Should only see one output from the transform"),A(a,i,c.transformed[0],u,m),c.dispose(),l&&(l.js=u.bundleFileInfo)}(i,f,y,S),e.tracing.pop(),e.tracing.push("emit","emitDeclarationFileOrBundle",{declarationFilePath:h}),function(r,a,i,s){if(!r)return;if(!a)return void((o||m.emitDeclarationOnly)&&(x=!0));var d=e.isSourceFile(r)?[r]:r.sourceFiles,f=c?d:e.filter(d,e.isSourceFileNotJson),_=e.outFile(m)?[e.factory.createBundle(f,e.isSourceFile(r)?void 0:r.prepends)]:f;o&&!e.getEmitDeclarations(m)&&f.forEach(L);var y=e.transformNodes(n,t,e.factory,m,_,p,!1);if(e.length(y.diagnostics))for(var v=0,h=y.diagnostics;v<h.length;v++){var E=h[v];g.add(E)}var T=b({removeComments:m.removeComments,newLine:m.newLine,noEmitHelpers:!0,module:m.module,target:m.target,sourceMap:m.sourceMap,inlineSourceMap:m.inlineSourceMap,extendedDiagnostics:m.extendedDiagnostics,onlyPrintJsDocStyle:!0,writeBundleFileInfo:!!l,recordInternalSection:!!l,relativeToBuildInfo:s},{hasGlobalName:n.hasGlobalName,onEmitNode:y.emitNodeWithNotification,isEmitNotificationEnabled:y.isEmitNotificationEnabled,substituteNode:y.substituteNode}),S=!!y.diagnostics&&!!y.diagnostics.length||!!t.isEmitBlocked(a)||!!m.noEmit;if(x=x||S,(!S||c)&&(e.Debug.assert(1===y.transformed.length,"Should only see one output from the decl transform"),A(a,i,y.transformed[0],T,{sourceMap:m.declarationMap,sourceRoot:m.sourceRoot,mapRoot:m.mapRoot,extendedDiagnostics:m.extendedDiagnostics}),c&&297===y.transformed[0].kind)){var D=y.transformed[0];u=D.exportedModulesFromDeclarationEmit}y.dispose(),l&&(l.dts=T.bundleFileInfo)}(i,h,E,S),e.tracing.pop(),e.tracing.push("emit","emitBuildInfo",{buildInfoPath:T}),function(n,a){if(!a||r||x)return;var i=t.getProgramBuildInfo();if(t.isEmitBlocked(a))return void(x=!0);var o=e.version;e.writeFile(t,g,a,v({bundle:n,program:i,version:o}),!1)}(l,T),e.tracing.pop(),!x&&_&&(o||(f&&_.push(f),y&&_.push(y),T&&_.push(T)),h&&_.push(h),E&&_.push(E));function S(n){return e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(s,n,t.getCanonicalFileName))}}),e.getSourceFilesToEmit(t,r,c),c,s,!r),S(),{emitSkipped:x,diagnostics:g.getDiagnostics(),emittedFiles:_,sourceMaps:f,exportedModulesFromDeclarationEmit:u};function L(t){e.isExportAssignment(t)?78===t.expression.kind&&n.collectLinkedAliases(t.expression,!0):e.isExportSpecifier(t)?n.collectLinkedAliases(t.propertyName||t.name,!0):e.forEachChild(t,L)}function A(n,r,a,i,o){var s,c=298===a.kind?a:void 0,l=297===a.kind?a:void 0,u=c?c.sourceFiles:[l];if(function(n,t){return(n.sourceMap||n.inlineSourceMap)&&(297!==t.kind||!e.fileExtensionIs(t.fileName,".json"))}(o,a)&&(s=e.createSourceMapGenerator(t,e.getBaseFileName(e.normalizeSlashes(n)),function(n){var t=e.normalizeSlashes(n.sourceRoot||"");return t?e.ensureTrailingDirectorySeparator(t):t}(o),function(n,r,a){if(n.sourceRoot)return t.getCommonSourceDirectory();if(n.mapRoot){var i=e.normalizeSlashes(n.mapRoot);return a&&(i=e.getDirectoryPath(e.getSourceFilePathInNewDir(a.fileName,t,i))),0===e.getRootLength(i)&&(i=e.combinePaths(t.getCommonSourceDirectory(),i)),i}return e.getDirectoryPath(e.normalizePath(r))}(o,n,l),o)),c?i.writeBundle(c,h,s):i.writeFile(l,h,s),s){f&&f.push({inputSourceFileNames:s.getSources(),sourceMap:s.toJSON()});var d=function(n,r,a,i,o){if(n.inlineSourceMap){var s=r.toString();return"data:application/json;base64,"+e.base64encode(e.sys,s)}var c=e.getBaseFileName(e.normalizeSlashes(e.Debug.checkDefined(i)));if(n.mapRoot){var l=e.normalizeSlashes(n.mapRoot);return o&&(l=e.getDirectoryPath(e.getSourceFilePathInNewDir(o.fileName,t,l))),0===e.getRootLength(l)?(l=e.combinePaths(t.getCommonSourceDirectory(),l),encodeURI(e.getRelativePathToDirectoryOrUrl(e.getDirectoryPath(e.normalizePath(a)),e.combinePaths(l,c),t.getCurrentDirectory(),t.getCanonicalFileName,!0))):encodeURI(e.combinePaths(l,c))}return encodeURI(c)}(o,s,n,r,l);if(d&&(h.isAtStartOfLine()||h.rawWrite(y),h.writeComment("//# sourceMappingURL="+d)),r){var p=s.toString();e.writeFile(t,g,r,p,!1,u)}}else h.writeLine();e.writeFile(t,g,n,h.getText(),!!m.emitBOM,u),h.clear()}}function v(e){return JSON.stringify(e,void 0,2)}function h(e){return JSON.parse(e)}function b(n,t){void 0===n&&(n={}),void 0===t&&(t={});var i,o,s,c,l,u,d,p,m,f,_,g,y,v,h,b,E,T,S,x=t.hasGlobalName,L=t.onEmitNode,A=void 0===L?e.noEmitNotification:L,D=t.isEmitNotificationEnabled,C=t.substituteNode,I=void 0===C?e.noEmitSubstitution:C,N=t.onBeforeEmitNodeArray,R=t.onAfterEmitNodeArray,k=t.onBeforeEmitToken,O=t.onAfterEmitToken,P=!!n.extendedDiagnostics,M=e.getNewLineCharacter(n),w=e.getEmitModuleKind(n),F=new e.Map,G=n.preserveSourceNewlines,B=function(e){f.write(e)},U=n.writeBundleFileInfo?{sections:[]}:void 0,V=U?e.Debug.checkDefined(n.relativeToBuildInfo):void 0,j=n.recordInternalSection,H=0,K="text",W=!0,z=-1,q=-1,J=-1,X=-1,Y=-1,Q=!1,Z=!!n.removeComments,$=e.performance.createTimerIf(P,"commentTime","beforeComment","afterComment"),ee=$.enter,ne=$.exit;return ve(),{printNode:function(n,t,r){switch(n){case 0:e.Debug.assert(e.isSourceFile(t),"Expected a SourceFile node.");break;case 2:e.Debug.assert(e.isIdentifier(t),"Expected an Identifier node.");break;case 1:e.Debug.assert(e.isExpression(t),"Expected an Expression node.")}switch(t.kind){case 297:return re(t);case 298:return te(t);case 299:return function(e,n){var t=f;ye(n,void 0),_e(4,e,void 0),ve(),f=t}(t,me()),fe()}return ae(n,t,r,me()),fe()},printList:function(e,n,t){return ie(e,n,t,me()),fe()},printFile:re,printBundle:te,writeNode:ae,writeList:ie,writeFile:pe,writeBundle:de,bundleFileInfo:U};function te(e){return de(e,me(),void 0),fe()}function re(e){return pe(e,me(),void 0),fe()}function ae(e,n,t,r){var a=f;ye(r,void 0),_e(e,n,t),ve(),f=a}function ie(e,n,t,r){var i=f;ye(r,void 0),t&&ge(t),Sn(a,n,e),ve(),f=i}function oe(){return f.getTextPosWithWriteLine?f.getTextPosWithWriteLine():f.getTextPos()}function se(n,t,r){var a=e.lastOrUndefined(U.sections);a&&a.kind===r?a.end=t:U.sections.push({pos:n,end:t,kind:r})}function ce(n){if(j&&U&&i&&(e.isDeclaration(n)||e.isVariableStatement(n))&&e.isInternalDeclaration(n,i)&&"internal"!==K){var t=K;return ue(f.getTextPos()),H=oe(),K="internal",t}}function le(e){e&&(ue(f.getTextPos()),H=oe(),K=e)}function ue(e){return H<e&&(se(H,e,K),!0)}function de(t,r,a){var i;g=!1;var o=f;ye(r,a),un(t),ln(t),Ne(t),function(n){an(!!n.hasNoDefaultLib,n.syntheticFileReferences||[],n.syntheticTypeReferences||[],n.syntheticLibReferences||[]);for(var t=0,r=n.prepends;t<r.length;t++){var a=r[t];if(e.isUnparsedSource(a)&&a.syntheticReferences)for(var i=0,o=a.syntheticReferences;i<o.length;i++){be(o[i]),wn()}}}(t);for(var s=0,c=t.prepends;s<c.length;s++){var l=c[s];wn();var u=f.getTextPos(),d=U&&U.sections;if(d&&(U.sections=[]),_e(4,l,void 0),U){var p=U.sections;U.sections=d,l.oldFileOfCurrentEmit?(i=U.sections).push.apply(i,p):(p.forEach((function(n){return e.Debug.assert(e.isBundleFileTextLike(n))})),U.sections.push({pos:u,end:f.getTextPos(),kind:"prepend",data:V(l.fileName),texts:p}))}}H=oe();for(var m=0,_=t.sourceFiles;m<_.length;m++){var y=_[m];_e(0,y,y)}if(U&&t.sourceFiles.length&&ue(f.getTextPos())){var v=function(n){for(var t,r=new e.Set,a=0;a<n.sourceFiles.length;a++){for(var i=n.sourceFiles[a],o=void 0,s=0,c=0,l=i.statements;c<l.length;c++){var u=l[c];if(!e.isPrologueDirective(u))break;r.has(u.expression.text)||(r.add(u.expression.text),(o||(o=[])).push({pos:u.pos,end:u.end,expression:{pos:u.expression.pos,end:u.expression.end,text:u.expression.text}}),s=s<u.end?u.end:s)}o&&(t||(t=[])).push({file:a,text:i.text.substring(0,s),directives:o})}return t}(t);v&&(U.sources||(U.sources={}),U.sources.prologues=v);var h=function(t){var r;if(w===e.ModuleKind.None||n.noEmitHelpers)return;for(var a=new e.Map,i=0,o=t.sourceFiles;i<o.length;i++){var s=o[i],c=void 0!==e.getExternalHelpersModuleName(s),l=Re(s);if(l)for(var u=0,d=l;u<d.length;u++){var p=d[u];p.scoped||c||a.get(p.name)||(a.set(p.name,!0),(r||(r=[])).push(p.name))}}return r}(t);h&&(U.sources||(U.sources={}),U.sources.helpers=h)}ve(),f=o}function pe(e,n,t){g=!0;var r=f;ye(n,t),un(e),ln(e),_e(0,e,e),ve(),f=r}function me(){return _||(_=e.createTextWriter(M))}function fe(){var e=_.getText();return _.clear(),e}function _e(e,n,t){t&&ge(t),xe(e,n)}function ge(e){i=e,b=void 0,E=void 0,e&&Bt(e)}function ye(t,r){t&&n.omitTrailingSemicolon&&(t=e.getTrailingSemicolonDeferringWriter(t)),y=r,W=!(f=t)||!y}function ve(){o=[],s=[],c=new e.Set,l=[],u=0,d=[],i=void 0,b=void 0,E=void 0,T=void 0,S=void 0,ye(void 0,void 0)}function he(){return b||(b=e.getLineStarts(i))}function be(e){if(void 0!==e){var n=ce(e),t=xe(4,e);return le(n),t}}function Ee(e){if(void 0!==e)return xe(2,e)}function Te(e){if(void 0!==e)return xe(1,e)}function Se(n){return xe(e.isStringLiteral(n)?6:4,n)}function xe(n,t){var r=T,a=S,i=G;T=t,S=void 0,G&&134217728&e.getEmitFlags(t)&&(G=!1),Le(0,n,t)(n,t),e.Debug.assert(T===t);var o=S;return T=r,S=a,G=i,o||t}function Le(n,t,r){switch(n){case 0:if(A!==e.noEmitNotification&&(!D||D(r)))return De;case 1:if(I!==e.noEmitSubstitution&&(S=I(t,r))!==r)return Ie;case 2:if(!Z&&297!==r.kind)return yt;case 3:if(!W&&297!==r.kind&&!e.isInJsonFile(r))return Mt;case 4:return Ce;default:return e.Debug.assertNever(n)}}function Ae(e,n,t){return Le(e+1,n,t)}function De(n,t){e.Debug.assert(T===t);var r=Ae(0,n,t);A(n,t,r),e.Debug.assert(T===t)}function Ce(n,t){if(e.Debug.assert(T===t||S===t),0===n)return function(n){wn();var t=n.statements;if(Et){if(0===t.length||!e.isPrologueDirective(t[0])||e.nodeIsSynthesized(t[0]))return void Et(n,t,on)}on(n)}(e.cast(t,e.isSourceFile));if(2===n)return Pe(e.cast(t,e.isIdentifier));if(6===n)return ke(e.cast(t,e.isStringLiteral),!0);if(3===n)return function(e){be(e.name),Pn(),Nn("in"),Pn(),be(e.constraint)}(e.cast(t,e.isTypeParameterDeclaration));if(5===n)return e.Debug.assertNode(t,e.isEmptyStatement),Fe(!0);if(4===n){if(e.isKeyword(t.kind))return Un(t,Nn);switch(t.kind){case 15:case 16:case 17:return ke(t,!1);case 299:case 293:return function(e){for(var n=0,t=e.texts;n<t.length;n++){var r=t[n];wn(),be(r)}}(t);case 292:return Oe(t);case 294:case 295:return o=t,s=oe(),Oe(o),void(U&&se(s,f.getTextPos(),294===o.kind?"text":"internal"));case 296:return function(n){var t=oe();if(Oe(n),U){var r=e.clone(n.section);r.pos=t,r.end=f.getTextPos(),U.sections.push(r)}}(t);case 78:return Pe(t);case 79:return function(e){(e.symbol?Dn:B)(tt(e,!1),e.symbol)}(t);case 157:return function(e){(function(e){78===e.kind?Te(e):be(e)})(e.left),Cn("."),be(e.right)}(t);case 158:return function(e){Cn("["),Te(e.expression),Cn("]")}(t);case 159:return function(e){be(e.name),e.constraint&&(Pn(),Nn("extends"),Pn(),be(e.constraint));e.default&&(Pn(),Rn("="),Pn(),be(e.default))}(t);case 160:return function(e){vn(e,e.decorators),pn(e,e.modifiers),be(e.dotDotDotToken),dn(e.name,kn),be(e.questionToken),e.parent&&308===e.parent.kind&&!e.name?be(e.type):mn(e.type);fn(e.initializer,e.type?e.type.end:e.questionToken?e.questionToken.end:e.name?e.name.end:e.modifiers?e.modifiers.end:e.decorators?e.decorators.end:e.pos,e)}(t);case 161:return a=t,Cn("@"),void Te(a.expression);case 162:return function(e){vn(e,e.decorators),pn(e,e.modifiers),dn(e.name,Mn),be(e.questionToken),mn(e.type),In()}(t);case 163:return function(e){vn(e,e.decorators),pn(e,e.modifiers),be(e.name),be(e.questionToken),be(e.exclamationToken),mn(e.type),fn(e.initializer,e.type?e.type.end:e.questionToken?e.questionToken.end:e.name.end,e),In()}(t);case 164:return function(e){at(e),vn(e,e.decorators),pn(e,e.modifiers),be(e.name),be(e.questionToken),bn(e,e.typeParameters),En(e,e.parameters),mn(e.type),In(),it(e)}(t);case 165:return function(e){vn(e,e.decorators),pn(e,e.modifiers),be(e.asteriskToken),be(e.name),be(e.questionToken),He(e,Ke)}(t);case 166:return function(e){pn(e,e.modifiers),Nn("constructor"),He(e,Ke)}(t);case 167:case 168:return function(e){vn(e,e.decorators),pn(e,e.modifiers),Nn(167===e.kind?"get":"set"),Pn(),be(e.name),He(e,Ke)}(t);case 169:return function(e){at(e),vn(e,e.decorators),pn(e,e.modifiers),bn(e,e.typeParameters),En(e,e.parameters),mn(e.type),In(),it(e)}(t);case 170:return function(e){at(e),vn(e,e.decorators),pn(e,e.modifiers),Nn("new"),Pn(),bn(e,e.typeParameters),En(e,e.parameters),mn(e.type),In(),it(e)}(t);case 171:return function(e){vn(e,e.decorators),pn(e,e.modifiers),n=e,t=e.parameters,Sn(n,t,8848),mn(e.type),In();var n,t}(t);case 194:return function(e){be(e.type),be(e.literal)}(t);case 172:return function(e){e.assertsModifier&&(be(e.assertsModifier),Pn());be(e.parameterName),e.type&&(Pn(),Nn("is"),Pn(),be(e.type))}(t);case 173:return function(e){be(e.typeName),hn(e,e.typeArguments)}(t);case 174:return function(e){at(e),bn(e,e.typeParameters),Tn(e,e.parameters),Pn(),Cn("=>"),Pn(),be(e.type),it(e)}(t);case 308:return function(e){Nn("function"),En(e,e.parameters),Cn(":"),be(e.type)}(t);case 175:return function(e){at(e),Nn("new"),Pn(),bn(e,e.typeParameters),En(e,e.parameters),Pn(),Cn("=>"),Pn(),be(e.type),it(e)}(t);case 176:return function(e){Nn("typeof"),Pn(),be(e.exprName)}(t);case 177:return function(n){Cn("{");var t=1&e.getEmitFlags(n)?768:32897;Sn(n,n.members,524288|t),Cn("}")}(t);case 178:return function(e){be(e.elementType),Cn("["),Cn("]")}(t);case 179:return function(n){Ue(22,n.pos,Cn,n);var t=1&e.getEmitFlags(n)?528:657;Sn(n,n.elements,524288|t),Ue(23,n.elements.end,Cn,n)}(t);case 180:return function(e){be(e.type),Cn("?")}(t);case 182:return function(e){Sn(e,e.types,516)}(t);case 183:return function(e){Sn(e,e.types,520)}(t);case 184:return function(e){be(e.checkType),Pn(),Nn("extends"),Pn(),be(e.extendsType),Pn(),Cn("?"),Pn(),be(e.trueType),Pn(),Cn(":"),Pn(),be(e.falseType)}(t);case 185:return function(e){Nn("infer"),Pn(),be(e.typeParameter)}(t);case 186:return function(e){Cn("("),be(e.type),Cn(")")}(t);case 223:return function(e){Te(e.expression),hn(e,e.typeArguments)}(t);case 187:return void Nn("this");case 188:return function(e){Vn(e.operator,Nn),Pn(),be(e.type)}(t);case 189:return function(e){be(e.objectType),Cn("["),be(e.indexType),Cn("]")}(t);case 190:return function(n){var t=e.getEmitFlags(n);Cn("{"),1&t?Pn():(wn(),Fn());n.readonlyToken&&(be(n.readonlyToken),142!==n.readonlyToken.kind&&Nn("readonly"),Pn());Cn("["),xe(3,n.typeParameter),n.nameType&&(Pn(),Nn("as"),Pn(),be(n.nameType));Cn("]"),n.questionToken&&(be(n.questionToken),57!==n.questionToken.kind&&Cn("?"));Cn(":"),Pn(),be(n.type),In(),1&t?Pn():(wn(),Gn());Cn("}")}(t);case 191:return function(e){Te(e.literal)}(t);case 193:return function(e){be(e.head),Sn(e,e.templateSpans,262144)}(t);case 195:return function(e){e.isTypeOf&&(Nn("typeof"),Pn());Nn("import"),Cn("("),be(e.argument),Cn(")"),e.qualifier&&(Cn("."),be(e.qualifier));hn(e,e.typeArguments)}(t);case 303:return void Cn("*");case 304:return void Cn("?");case 305:return function(e){Cn("?"),be(e.type)}(t);case 306:return function(e){Cn("!"),be(e.type)}(t);case 307:return function(e){be(e.type),Cn("=")}(t);case 181:case 309:return function(e){Cn("..."),be(e.type)}(t);case 192:return function(e){be(e.dotDotDotToken),be(e.name),be(e.questionToken),Ue(58,e.name.end,Cn,e),Pn(),be(e.type)}(t);case 196:return function(e){Cn("{"),Sn(e,e.elements,525136),Cn("}")}(t);case 197:return function(e){Cn("["),Sn(e,e.elements,524880),Cn("]")}(t);case 198:return function(e){be(e.dotDotDotToken),e.propertyName&&(be(e.propertyName),Cn(":"),Pn());be(e.name),fn(e.initializer,e.name.end,e)}(t);case 228:return function(e){Te(e.expression),be(e.literal)}(t);case 229:return void In();case 230:return function(e){we(e,!e.multiLine&&et(e))}(t);case 232:return function(e){pn(e,e.modifiers),be(e.declarationList),In()}(t);case 231:return Fe(!1);case 233:return function(n){Te(n.expression),(!e.isJsonSourceFile(i)||e.nodeIsSynthesized(n.expression))&&In()}(t);case 234:return function(e){var n=Ue(98,e.pos,Nn,e);Pn(),Ue(20,n,Cn,e),Te(e.expression),Ue(21,e.expression.end,Cn,e),yn(e,e.thenStatement),e.elseStatement&&(jn(e,e.thenStatement,e.elseStatement),Ue(90,e.thenStatement.end,Nn,e),234===e.elseStatement.kind?(Pn(),be(e.elseStatement)):yn(e,e.elseStatement))}(t);case 235:return function(n){Ue(89,n.pos,Nn,n),yn(n,n.statement),e.isBlock(n.statement)&&!G?Pn():jn(n,n.statement,n.expression);Ge(n,n.statement.end),In()}(t);case 236:return function(e){Ge(e,e.pos),yn(e,e.statement)}(t);case 237:return function(e){var n=Ue(96,e.pos,Nn,e);Pn();var t=Ue(20,n,Cn,e);Be(e.initializer),t=Ue(26,e.initializer?e.initializer.end:t,Cn,e),gn(e.condition),t=Ue(26,e.condition?e.condition.end:t,Cn,e),gn(e.incrementor),Ue(21,e.incrementor?e.incrementor.end:t,Cn,e),yn(e,e.statement)}(t);case 238:return function(e){var n=Ue(96,e.pos,Nn,e);Pn(),Ue(20,n,Cn,e),Be(e.initializer),Pn(),Ue(100,e.initializer.end,Nn,e),Pn(),Te(e.expression),Ue(21,e.expression.end,Cn,e),yn(e,e.statement)}(t);case 239:return function(e){var n=Ue(96,e.pos,Nn,e);Pn(),function(e){e&&(be(e),Pn())}(e.awaitModifier),Ue(20,n,Cn,e),Be(e.initializer),Pn(),Ue(156,e.initializer.end,Nn,e),Pn(),Te(e.expression),Ue(21,e.expression.end,Cn,e),yn(e,e.statement)}(t);case 240:return function(e){Ue(85,e.pos,Nn,e),_n(e.label),In()}(t);case 241:return function(e){Ue(80,e.pos,Nn,e),_n(e.label),In()}(t);case 242:return function(e){Ue(104,e.pos,Nn,e),gn(e.expression),In()}(t);case 243:return function(e){var n=Ue(115,e.pos,Nn,e);Pn(),Ue(20,n,Cn,e),Te(e.expression),Ue(21,e.expression.end,Cn,e),yn(e,e.statement)}(t);case 244:return function(e){var n=Ue(106,e.pos,Nn,e);Pn(),Ue(20,n,Cn,e),Te(e.expression),Ue(21,e.expression.end,Cn,e),Pn(),be(e.caseBlock)}(t);case 245:return function(e){be(e.label),Ue(58,e.label.end,Cn,e),Pn(),be(e.statement)}(t);case 246:return function(e){Ue(108,e.pos,Nn,e),gn(e.expression),In()}(t);case 247:return function(e){Ue(110,e.pos,Nn,e),Pn(),be(e.tryBlock),e.catchClause&&(jn(e,e.tryBlock,e.catchClause),be(e.catchClause));e.finallyBlock&&(jn(e,e.catchClause||e.tryBlock,e.finallyBlock),Ue(95,(e.catchClause||e.tryBlock).end,Nn,e),Pn(),be(e.finallyBlock))}(t);case 248:return function(e){Bn(86,e.pos,Nn),In()}(t);case 249:return function(e){be(e.name),be(e.exclamationToken),mn(e.type),fn(e.initializer,e.type?e.type.end:e.name.end,e)}(t);case 250:return function(n){Nn(e.isLet(n)?"let":e.isVarConst(n)?"const":"var"),Pn(),Sn(n,n.declarations,528)}(t);case 251:return function(e){Ve(e)}(t);case 252:return function(e){Je(e)}(t);case 253:return function(e){vn(e,e.decorators),pn(e,e.modifiers),Nn("interface"),Pn(),be(e.name),bn(e,e.typeParameters),Sn(e,e.heritageClauses,512),Pn(),Cn("{"),Sn(e,e.members,129),Cn("}")}(t);case 254:return function(e){vn(e,e.decorators),pn(e,e.modifiers),Nn("type"),Pn(),be(e.name),bn(e,e.typeParameters),Pn(),Cn("="),Pn(),be(e.type),In()}(t);case 255:return function(e){pn(e,e.modifiers),Nn("enum"),Pn(),be(e.name),Pn(),Cn("{"),Sn(e,e.members,145),Cn("}")}(t);case 256:return function(e){pn(e,e.modifiers),1024&~e.flags&&(Nn(16&e.flags?"namespace":"module"),Pn());be(e.name);var n=e.body;if(!n)return In();for(;256===n.kind;)Cn("."),be(n.name),n=n.body;Pn(),be(n)}(t);case 257:return function(n){at(n),e.forEach(n.statements,st),we(n,et(n)),it(n)}(t);case 258:return function(e){Ue(18,e.pos,Cn,e),Sn(e,e.clauses,129),Ue(19,e.clauses.end,Cn,e,!0)}(t);case 259:return function(e){var n=Ue(92,e.pos,Nn,e);Pn(),n=Ue(126,n,Nn,e),Pn(),n=Ue(140,n,Nn,e),Pn(),be(e.name),In()}(t);case 260:return function(e){pn(e,e.modifiers),Ue(99,e.modifiers?e.modifiers.end:e.pos,Nn,e),Pn(),be(e.name),Pn(),Ue(62,e.name.end,Cn,e),Pn(),function(e){78===e.kind?Te(e):be(e)}(e.moduleReference),In()}(t);case 261:return function(e){pn(e,e.modifiers),Ue(99,e.modifiers?e.modifiers.end:e.pos,Nn,e),Pn(),e.importClause&&(be(e.importClause),Pn(),Ue(153,e.importClause.end,Nn,e),Pn());Te(e.moduleSpecifier),In()}(t);case 262:return function(e){e.isTypeOnly&&(Ue(149,e.pos,Nn,e),Pn());be(e.name),e.name&&e.namedBindings&&(Ue(27,e.name.end,Cn,e),Pn());be(e.namedBindings)}(t);case 263:return function(e){var n=Ue(41,e.pos,Cn,e);Pn(),Ue(126,n,Nn,e),Pn(),be(e.name)}(t);case 269:return function(e){var n=Ue(41,e.pos,Cn,e);Pn(),Ue(126,n,Nn,e),Pn(),be(e.name)}(t);case 264:return function(e){Xe(e)}(t);case 265:return function(e){Ye(e)}(t);case 266:return function(e){var n=Ue(92,e.pos,Nn,e);Pn(),e.isExportEquals?Ue(62,n,Rn,e):Ue(87,n,Nn,e);Pn(),Te(e.expression),In()}(t);case 267:return function(e){var n=Ue(92,e.pos,Nn,e);Pn(),e.isTypeOnly&&(n=Ue(149,n,Nn,e),Pn());e.exportClause?be(e.exportClause):n=Ue(41,n,Cn,e);if(e.moduleSpecifier){Pn(),Ue(153,e.exportClause?e.exportClause.end:n,Nn,e),Pn(),Te(e.moduleSpecifier)}In()}(t);case 268:return function(e){Xe(e)}(t);case 270:return function(e){Ye(e)}(t);case 271:return;case 272:return function(e){Nn("require"),Cn("("),Te(e.expression),Cn(")")}(t);case 11:return function(e){f.writeLiteral(e.text)}(t);case 275:case 278:return function(n){if(Cn("<"),e.isJsxOpeningElement(n)){var t=Yn(n.tagName,n);Qe(n.tagName),hn(n,n.typeArguments),n.attributes.properties&&n.attributes.properties.length>0&&Pn(),be(n.attributes),Qn(n.attributes,n),Wn(t)}Cn(">")}(t);case 276:case 279:return function(n){Cn("</"),e.isJsxClosingElement(n)&&Qe(n.tagName);Cn(">")}(t);case 280:return function(e){be(e.name),function(e,n,t,r){t&&(n(e),r(t))}("=",Cn,e.initializer,Se)}(t);case 281:return function(e){Sn(e,e.properties,262656)}(t);case 282:return function(e){Cn("{..."),Te(e.expression),Cn("}")}(t);case 283:return function(e){e.expression&&(Cn("{"),be(e.dotDotDotToken),Te(e.expression),Cn("}"))}(t);case 284:return function(e){Ue(81,e.pos,Nn,e),Pn(),Te(e.expression),Ze(e,e.statements,e.expression.end)}(t);case 285:return function(e){var n=Ue(87,e.pos,Nn,e);Ze(e,e.statements,n)}(t);case 286:return function(e){Pn(),Vn(e.token,Nn),Pn(),Sn(e,e.types,528)}(t);case 287:return function(e){var n=Ue(82,e.pos,Nn,e);Pn(),e.variableDeclaration&&(Ue(20,n,Cn,e),be(e.variableDeclaration),Ue(21,e.variableDeclaration.end,Cn,e),Pn());be(e.block)}(t);case 288:return function(n){be(n.name),Cn(":"),Pn();var t=n.initializer;if(It&&0==(512&e.getEmitFlags(t))){It(e.getCommentRange(t).pos)}Te(t)}(t);case 289:return function(e){be(e.name),e.objectAssignmentInitializer&&(Pn(),Cn("="),Pn(),Te(e.objectAssignmentInitializer))}(t);case 290:return function(e){e.expression&&(Ue(25,e.pos,Cn,e),Te(e.expression))}(t);case 291:return function(e){be(e.name),fn(e.initializer,e.name.end,e)}(t);case 326:case 333:return function(e){nn(e.tagName),rn(e.typeExpression),Pn(),e.isBracketed&&Cn("[");be(e.name),e.isBracketed&&Cn("]");tn(e.comment)}(t);case 327:case 329:case 328:case 325:return nn((r=t).tagName),rn(r.typeExpression),void tn(r.comment);case 316:case 315:return function(e){nn(e.tagName),Pn(),Cn("{"),be(e.class),Cn("}"),tn(e.comment)}(t);case 330:return function(e){nn(e.tagName),rn(e.constraint),Pn(),Sn(e,e.typeParameters,528),tn(e.comment)}(t);case 331:return function(e){nn(e.tagName),e.typeExpression&&(301===e.typeExpression.kind?rn(e.typeExpression):(Pn(),Cn("{"),B("Object"),e.typeExpression.isArrayType&&(Cn("["),Cn("]")),Cn("}")));e.fullName&&(Pn(),be(e.fullName));tn(e.comment),e.typeExpression&&312===e.typeExpression.kind&&$e(e.typeExpression)}(t);case 324:return function(e){nn(e.tagName),e.name&&(Pn(),be(e.name));tn(e.comment),en(e.typeExpression)}(t);case 313:return en(t);case 312:return $e(t);case 319:case 314:return function(e){nn(e.tagName),tn(e.comment)}(t);case 332:return function(e){nn(e.tagName),be(e.name),tn(e.comment)}(t);case 302:return function(e){Pn(),Cn("{"),be(e.name),Cn("}")}(t);case 311:return function(e){if(B("/**"),e.comment)for(var n=0,t=e.comment.split(/\r\n?|\n/g);n<t.length;n++){var r=t[n];wn(),Pn(),Cn("*"),Pn(),B(r)}e.tags&&(1!==e.tags.length||329!==e.tags[0].kind||e.comment?Sn(e,e.tags,33):(Pn(),be(e.tags[0])));Pn(),B("*/")}(t)}if(e.isExpression(t))n=1,I!==e.noEmitSubstitution&&(S=t=I(n,t));else if(e.isToken(t))return Un(t,Cn)}var r,a,o,s;if(1===n)switch(t.kind){case 8:case 9:return function(e){ke(e,!1)}(t);case 10:case 13:case 14:return ke(t,!1);case 78:return Pe(t);case 94:case 103:case 105:case 109:case 107:case 99:return void Un(t,Nn);case 199:return function(e){var n=e.elements,t=e.multiLine?65536:0;xn(e,n,8914|t)}(t);case 200:return function(n){e.forEach(n.properties,ct);var t=65536&e.getEmitFlags(n);t&&Fn();var r=n.multiLine?65536:0,a=i.languageVersion>=1&&!e.isJsonSourceFile(i)?64:0;Sn(n,n.properties,526226|a|r),t&&Gn()}(t);case 201:return function(n){var t=e.cast(Te(n.expression),e.isExpression),r=n.questionDotToken||e.setTextRangePosEnd(e.factory.createToken(24),n.expression.end,n.name.pos),a=$n(n,n.expression,r),i=$n(n,r,n.name);Kn(a,!1),28===r.kind||!function(n){if(n=e.skipPartiallyEmittedExpressions(n),e.isNumericLiteral(n)){var t=rt(n,!0,!1);return!n.numericLiteralFlags&&!e.stringContains(t,e.tokenToString(24))}if(e.isAccessExpression(n)){var r=e.getConstantValue(n);return"number"==typeof r&&isFinite(r)&&Math.floor(r)===r}}(t)||f.hasTrailingComment()||f.hasTrailingWhitespace()||Cn(".");n.questionDotToken?be(r):Ue(r.kind,n.expression.end,Cn,n);Kn(i,!1),be(n.name),Wn(a,i)}(t);case 202:return function(e){Te(e.expression),be(e.questionDotToken),Ue(22,e.expression.end,Cn,e),Te(e.argumentExpression),Ue(23,e.argumentExpression.end,Cn,e)}(t);case 203:return function(e){Te(e.expression),be(e.questionDotToken),hn(e,e.typeArguments),xn(e,e.arguments,2576)}(t);case 204:return function(e){Ue(102,e.pos,Nn,e),Pn(),Te(e.expression),hn(e,e.typeArguments),xn(e,e.arguments,18960)}(t);case 205:return function(e){Te(e.tag),hn(e,e.typeArguments),Pn(),Te(e.template)}(t);case 206:return function(e){Cn("<"),be(e.type),Cn(">"),Te(e.expression)}(t);case 207:return function(e){var n=Ue(20,e.pos,Cn,e),t=Yn(e.expression,e);Te(e.expression),Qn(e.expression,e),Wn(t),Ue(21,e.expression?e.expression.end:n,Cn,e)}(t);case 208:return function(e){lt(e.name),Ve(e)}(t);case 209:return function(e){vn(e,e.decorators),pn(e,e.modifiers),He(e,Me)}(t);case 210:return function(e){Ue(88,e.pos,Nn,e),Pn(),Te(e.expression)}(t);case 211:return function(e){Ue(111,e.pos,Nn,e),Pn(),Te(e.expression)}(t);case 212:return function(e){Ue(113,e.pos,Nn,e),Pn(),Te(e.expression)}(t);case 213:return function(e){Ue(130,e.pos,Nn,e),Pn(),Te(e.expression)}(t);case 214:return function(e){Vn(e.operator,Rn),function(e){var n=e.operand;return 214===n.kind&&(39===e.operator&&(39===n.operator||45===n.operator)||40===e.operator&&(40===n.operator||46===n.operator))}(e)&&Pn();Te(e.operand)}(t);case 215:return function(e){Te(e.operand),Vn(e.operator,Rn)}(t);case 216:return function(n){var t=[n],r=[0],a=0;for(;a>=0;)switch(n=t[a],r[a]){case 0:c(n.left);break;case 1:var i=27!==n.operatorToken.kind,o=$n(n,n.left,n.operatorToken),s=$n(n,n.operatorToken,n.right);Kn(o,i),Dt(n.operatorToken.pos),Un(n.operatorToken,100===n.operatorToken.kind?Nn:Rn),It(n.operatorToken.end,!0),Kn(s,!0),c(n.right);break;case 2:Wn(o=$n(n,n.left,n.operatorToken),s=$n(n,n.operatorToken,n.right)),a--;break;default:return e.Debug.fail("Invalid state "+r[a]+" for emitBinaryExpressionWorker")}function c(n){r[a]++;var i=T,o=S;T=n,S=void 0;var s=Le(0,1,n);s===Ce&&e.isBinaryExpression(n)?(a++,r[a]=0,t[a]=n):s(1,n),e.Debug.assert(T===n),T=i,S=o}}(t);case 217:return function(e){var n=$n(e,e.condition,e.questionToken),t=$n(e,e.questionToken,e.whenTrue),r=$n(e,e.whenTrue,e.colonToken),a=$n(e,e.colonToken,e.whenFalse);Te(e.condition),Kn(n,!0),be(e.questionToken),Kn(t,!0),Te(e.whenTrue),Wn(n,t),Kn(r,!0),be(e.colonToken),Kn(a,!0),Te(e.whenFalse),Wn(r,a)}(t);case 218:return function(e){be(e.head),Sn(e,e.templateSpans,262144)}(t);case 219:return function(e){Ue(124,e.pos,Nn,e),be(e.asteriskToken),gn(e.expression)}(t);case 220:return function(e){Ue(25,e.pos,Cn,e),Te(e.expression)}(t);case 221:return function(e){lt(e.name),Je(e)}(t);case 222:return;case 224:return function(e){Te(e.expression),e.type&&(Pn(),Nn("as"),Pn(),be(e.type))}(t);case 225:return function(e){Te(e.expression),Rn("!")}(t);case 226:return function(e){Bn(e.keywordToken,e.pos,Cn),Cn("."),be(e.name)}(t);case 273:return function(e){be(e.openingElement),Sn(e,e.children,262144),be(e.closingElement)}(t);case 274:return function(e){Cn("<"),Qe(e.tagName),hn(e,e.typeArguments),Pn(),be(e.attributes),Cn("/>")}(t);case 277:return function(e){be(e.openingFragment),Sn(e,e.children,262144),be(e.closingFragment)}(t);case 336:return function(e){Te(e.expression)}(t);case 337:return function(e){xn(e,e.elements,528)}(t)}}function Ie(n,t){e.Debug.assert(T===t||S===t),Ae(1,n,t)(n,S),e.Debug.assert(T===t||S===t)}function Ne(t){var r=!1,a=298===t.kind?t:void 0;if(!a||w!==e.ModuleKind.None){for(var o=a?a.prepends.length:0,s=a?a.sourceFiles.length+o:1,c=0;c<s;c++){var l=a?c<o?a.prepends[c]:a.sourceFiles[c-o]:t,u=e.isSourceFile(l)?l:e.isUnparsedSource(l)?void 0:i,d=n.noEmitHelpers||!!u&&e.hasRecordedExternalHelpers(u),p=(e.isSourceFile(l)||e.isUnparsedSource(l))&&!g,m=e.isUnparsedSource(l)?l.helpers:Re(l);if(m)for(var _=0,y=m;_<y.length;_++){var v=y[_];if(v.scoped){if(a)continue}else{if(d)continue;if(p){if(F.get(v.name))continue;F.set(v.name,!0)}}var h=oe();"string"==typeof v.text?Hn(v.text):Hn(v.text(gt)),U&&U.sections.push({pos:h,end:f.getTextPos(),kind:"emitHelpers",data:v.name}),r=!0}}return r}}function Re(n){var t=e.getEmitHelpers(n);return t&&e.stableSort(t,e.compareEmitHelpers)}function ke(t,r){var a,i=rt(t,n.neverAsciiEscape,r);!n.sourceMap&&!n.inlineSourceMap||10!==t.kind&&!e.isTemplateLiteralKind(t.kind)?function(e){f.writeStringLiteral(e)}(i):(a=i,f.writeLiteral(a))}function Oe(e){f.rawWrite(e.parent.text.substring(e.pos,e.end))}function Pe(e){(e.symbol?Dn:B)(tt(e,!1),e.symbol),Sn(e,e.typeArguments,53776)}function Me(e){bn(e,e.typeParameters),Tn(e,e.parameters),mn(e.type),Pn(),be(e.equalsGreaterThanToken)}function we(n,t){Ue(18,n.pos,Cn,n);var r=t||1&e.getEmitFlags(n)?768:129;Sn(n,n.statements,r),Ue(19,n.statements.end,Cn,n,!!(1&r))}function Fe(e){e?Cn(";"):In()}function Ge(e,n){var t=Ue(114,n,Nn,e);Pn(),Ue(20,t,Cn,e),Te(e.expression),Ue(21,e.expression.end,Cn,e)}function Be(e){void 0!==e&&(250===e.kind?be(e):Te(e))}function Ue(n,t,r,a,o){var s=e.getParseTreeNode(a),c=s&&s.kind===a.kind,l=t;if(c&&i&&(t=e.skipTrivia(i.text,t)),Dt&&c&&a.pos!==l){var u=o&&i&&!e.positionsAreOnSameLine(l,t,i);u&&Fn(),Dt(l),u&&Gn()}return t=Vn(n,r,t),It&&c&&a.end!==t&&It(t,!0),t}function Ve(e){vn(e,e.decorators),pn(e,e.modifiers),Nn("function"),be(e.asteriskToken),Pn(),Ee(e.name),He(e,Ke)}function je(e,n){We(n)}function He(n,t){var r=n.body;if(r)if(e.isBlock(r)){var a=65536&e.getEmitFlags(n);a&&Fn(),at(n),e.forEach(n.parameters,st),st(n.body),t(n),A?A(4,r,je):We(r),it(n),a&&Gn()}else t(n),Pn(),Te(r);else t(n),In()}function Ke(e){bn(e,e.typeParameters),En(e,e.parameters),mn(e.type)}function We(n){Pn(),Cn("{"),Fn();var t=function(n){if(1&e.getEmitFlags(n))return!0;if(n.multiLine)return!1;if(!e.nodeIsSynthesized(n)&&!e.rangeIsOnSingleLine(n,i))return!1;if(zn(n,n.statements,2)||Jn(n,n.statements,2))return!1;for(var t,r=0,a=n.statements;r<a.length;r++){var o=a[r];if(qn(t,o,2)>0)return!1;t=o}return!0}(n)?ze:qe;Et?Et(n,n.statements,t):t(n),Gn(),Bn(19,n.statements.end,Cn,n)}function ze(e){qe(e,!0)}function qe(e,n){var t=sn(e.statements),r=f.getTextPos();Ne(e),0===t&&r===f.getTextPos()&&n?(Gn(),Sn(e,e.statements,768),Fn()):Sn(e,e.statements,1,t)}function Je(n){e.forEach(n.members,ct),vn(n,n.decorators),pn(n,n.modifiers),Nn("class"),n.name&&(Pn(),Ee(n.name));var t=65536&e.getEmitFlags(n);t&&Fn(),bn(n,n.typeParameters),Sn(n,n.heritageClauses,0),Pn(),Cn("{"),Sn(n,n.members,129),Cn("}"),t&&Gn()}function Xe(e){Cn("{"),Sn(e,e.elements,525136),Cn("}")}function Ye(e){e.propertyName&&(be(e.propertyName),Pn(),Ue(126,e.propertyName.end,Nn,e),Pn()),be(e.name)}function Qe(e){78===e.kind?Te(e):be(e)}function Ze(n,t,r){var a=163969;1===t.length&&(e.nodeIsSynthesized(n)||e.nodeIsSynthesized(t[0])||e.rangeStartPositionsAreOnSameLine(n,t[0],i))?(Bn(58,r,Cn,n),Pn(),a&=-130):Ue(58,r,Cn,n),Sn(n,t,a)}function $e(n){Sn(n,e.factory.createNodeArray(n.jsDocPropertyTags),33)}function en(n){n.typeParameters&&Sn(n,e.factory.createNodeArray(n.typeParameters),33),n.parameters&&Sn(n,e.factory.createNodeArray(n.parameters),33),n.type&&(wn(),Pn(),Cn("*"),Pn(),be(n.type))}function nn(e){Cn("@"),be(e)}function tn(e){e&&(Pn(),B(e))}function rn(e){e&&(Pn(),Cn("{"),be(e.type),Cn("}"))}function an(e,n,t,r){if(e){var a=f.getTextPos();On('/// <reference no-default-lib="true"/>'),U&&U.sections.push({pos:a,end:f.getTextPos(),kind:"no-default-lib"}),wn()}if(i&&i.moduleName&&(On('/// <amd-module name="'+i.moduleName+'" />'),wn()),i&&i.amdDependencies)for(var o=0,s=i.amdDependencies;o<s.length;o++){var c=s[o];c.name?On('/// <amd-dependency name="'+c.name+'" path="'+c.path+'" />'):On('/// <amd-dependency path="'+c.path+'" />'),wn()}for(var l=0,u=n;l<u.length;l++){var d=u[l];a=f.getTextPos();On('/// <reference path="'+d.fileName+'" />'),U&&U.sections.push({pos:a,end:f.getTextPos(),kind:"reference",data:d.fileName}),wn()}for(var p=0,m=t;p<m.length;p++){d=m[p],a=f.getTextPos();On('/// <reference types="'+d.fileName+'" />'),U&&U.sections.push({pos:a,end:f.getTextPos(),kind:"type",data:d.fileName}),wn()}for(var _=0,g=r;_<g.length;_++){d=g[_],a=f.getTextPos();On('/// <reference lib="'+d.fileName+'" />'),U&&U.sections.push({pos:a,end:f.getTextPos(),kind:"lib",data:d.fileName}),wn()}}function on(n){var t=n.statements;at(n),e.forEach(n.statements,st),Ne(n);var r=e.findIndex(t,(function(n){return!e.isPrologueDirective(n)}));!function(e){e.isDeclarationFile&&an(e.hasNoDefaultLib,e.referencedFiles,e.typeReferenceDirectives,e.libReferenceDirectives)}(n),Sn(n,t,1,-1===r?t.length:r),it(n)}function sn(n,t,r,a){for(var i=!!t,o=0;o<n.length;o++){var s=n[o];if(!e.isPrologueDirective(s))return o;if(!r||!r.has(s.expression.text)){i&&(i=!1,ge(t)),wn();var c=f.getTextPos();be(s),a&&U&&U.sections.push({pos:c,end:f.getTextPos(),kind:"prologue",data:s.expression.text}),r&&r.add(s.expression.text)}}return n.length}function cn(e,n){for(var t=0,r=e;t<r.length;t++){var a=r[t];if(!n.has(a.data)){wn();var i=f.getTextPos();be(a),U&&U.sections.push({pos:i,end:f.getTextPos(),kind:"prologue",data:a.data}),n&&n.add(a.data)}}}function ln(n){if(e.isSourceFile(n))sn(n.statements,n);else{for(var t=new e.Set,r=0,a=n.prepends;r<a.length;r++){cn(a[r].prologues,t)}for(var i=0,o=n.sourceFiles;i<o.length;i++){var s=o[i];sn(s.statements,s,t,!0)}ge(void 0)}}function un(n){if(e.isSourceFile(n)||e.isUnparsedSource(n)){var t=e.getShebang(n.text);if(t)return On(t),wn(),!0}else{for(var r=0,a=n.prepends;r<a.length;r++){var i=a[r];if(e.Debug.assertNode(i,e.isUnparsedSource),un(i))return!0}for(var o=0,s=n.sourceFiles;o<s.length;o++){if(un(s[o]))return!0}}}function dn(e,n){if(e){var t=B;B=n,be(e),B=t}}function pn(e,n){n&&n.length&&(Sn(e,n,262656),Pn())}function mn(e){e&&(Cn(":"),Pn(),be(e))}function fn(e,n,t){e&&(Pn(),Ue(62,n,Rn,t),Pn(),Te(e))}function _n(e){e&&(Pn(),be(e))}function gn(e){e&&(Pn(),Te(e))}function yn(n,t){e.isBlock(t)||1&e.getEmitFlags(n)?(Pn(),be(t)):(wn(),Fn(),e.isEmptyStatement(t)?xe(5,t):be(t),Gn())}function vn(e,n){Sn(e,n,2146305)}function hn(e,n){Sn(e,n,53776)}function bn(n,t){if(e.isFunctionLike(n)&&n.typeArguments)return hn(n,n.typeArguments);Sn(n,t,53776)}function En(e,n){Sn(e,n,2576)}function Tn(n,t){!function(n,t){var r=e.singleOrUndefined(t);return r&&r.pos===n.pos&&e.isArrowFunction(n)&&!n.type&&!e.some(n.decorators)&&!e.some(n.modifiers)&&!e.some(n.typeParameters)&&!e.some(r.decorators)&&!e.some(r.modifiers)&&!r.dotDotDotToken&&!r.questionToken&&!r.type&&!r.initializer&&e.isIdentifier(r.name)}(n,t)?En(n,t):Sn(n,t,528)}function Sn(e,n,t,r,a){An(be,e,n,t,r,a)}function xn(e,n,t,r,a){An(Te,e,n,t,r,a)}function Ln(e){switch(60&e){case 0:break;case 16:Cn(",");break;case 4:Pn(),Cn("|");break;case 32:Pn(),Cn("*"),Pn();break;case 8:Pn(),Cn("&")}}function An(n,t,a,o,s,c){void 0===s&&(s=0),void 0===c&&(c=a?a.length-s:0);var l=void 0===a;if(!(l&&16384&o)){var u=void 0===a||s>=a.length||0===c;if(u&&32768&o)return N&&N(a),void(R&&R(a));if(15360&o&&(Cn(function(e){return r[15360&e][0]}(o)),u&&!l&&It(a.pos,!0)),N&&N(a),u)!(1&o)||G&&e.rangeIsOnSingleLine(t,i)?256&o&&!(524288&o)&&Pn():wn();else{var d=0==(262144&o),p=d,f=zn(t,a,o);f?(wn(f),p=!1):256&o&&Pn(),128&o&&Fn();for(var _=void 0,g=void 0,y=!1,v=0;v<c;v++){var h=a[s+v];if(32&o)wn(),Ln(o);else if(_){60&o&&_.end!==t.end&&Dt(_.end),Ln(o),le(g);var b=qn(_,h,o);b>0?(0==(131&o)&&(Fn(),y=!0),wn(b),p=!1):_&&512&o&&Pn()}if(g=ce(h),p){if(It)It(e.getCommentRange(h).pos)}else p=d;m=h.pos,n(h),y&&(Gn(),y=!1),_=h}var E=64&o&&a.hasTrailingComma;16&o&&E&&Cn(","),_&&60&o&&_.end!==t.end&&!(1024&e.getEmitFlags(_))&&Dt(_.end),128&o&&Gn(),le(g);var T=Jn(t,a,o);T?wn(T):2097408&o&&Pn()}R&&R(a),15360&o&&(u&&!l&&Dt(a.end),Cn(function(e){return r[15360&e][1]}(o)))}}function Dn(e,n){f.writeSymbol(e,n)}function Cn(e){f.writePunctuation(e)}function In(){f.writeTrailingSemicolon(";")}function Nn(e){f.writeKeyword(e)}function Rn(e){f.writeOperator(e)}function kn(e){f.writeParameter(e)}function On(e){f.writeComment(e)}function Pn(){f.writeSpace(" ")}function Mn(e){f.writeProperty(e)}function wn(e){void 0===e&&(e=1);for(var n=0;n<e;n++)f.writeLine(n>0)}function Fn(){f.increaseIndent()}function Gn(){f.decreaseIndent()}function Bn(n,t,r,a){return W?Vn(n,r,t):function(n,t,r,a,i){if(W||n&&e.isInJsonFile(n))return i(t,r,a);var o=n&&n.emitNode,s=o&&o.flags||0,c=o&&o.tokenSourceMapRanges&&o.tokenSourceMapRanges[t],l=c&&c.source||v;a=wt(l,c?c.pos:a),0==(128&s)&&a>=0&&Gt(l,a);a=i(t,r,a),c&&(a=c.end);0==(256&s)&&a>=0&&Gt(l,a);return a}(a,n,r,t,Vn)}function Un(n,t){k&&k(n),t(e.tokenToString(n.kind)),O&&O(n)}function Vn(n,t,r){var a=e.tokenToString(n);return t(a),r<0?r:r+a.length}function jn(n,t,r){if(1&e.getEmitFlags(n))Pn();else if(G){var a=$n(n,t,r);a?wn(a):Pn()}else wn()}function Hn(n){for(var t=n.split(/\r\n?|\n/g),r=e.guessIndentation(t),a=0,i=t;a<i.length;a++){var o=i[a],s=r?o.slice(r):o;s.length&&(wn(),B(s))}}function Kn(e,n){e?(Fn(),wn(e)):n&&Pn()}function Wn(e,n){e&&Gn(),n&&Gn()}function zn(n,t,r){if(2&r||G){if(65536&r)return 1;var a=t[0];if(void 0===a)return e.rangeIsOnSingleLine(n,i)?0:1;if(a.pos===m)return 0;if(11===a.kind)return 0;if(!(e.positionIsSynthesized(n.pos)||e.nodeIsSynthesized(a)||a.parent&&e.getOriginalNode(a.parent)!==e.getOriginalNode(n)))return G?Xn((function(t){return e.getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter(a.pos,n.pos,i,t)})):e.rangeStartPositionsAreOnSameLine(n,a,i)?0:1;if(Zn(a,r))return 1}return 1&r?1:0}function qn(n,t,r){if(2&r||G){if(void 0===n||void 0===t)return 0;if(11===t.kind)return 0;if(!e.nodeIsSynthesized(n)&&!e.nodeIsSynthesized(t)&&n.parent===t.parent)return G?Xn((function(r){return e.getLinesBetweenRangeEndAndRangeStart(n,t,i,r)})):e.rangeEndIsOnSameLineAsRangeStart(n,t,i)?0:1;if(Zn(n,r)||Zn(t,r))return 1}else if(e.getStartsOnNewLine(t))return 1;return 1&r?1:0}function Jn(n,t,r){if(2&r||G){if(65536&r)return 1;var a=e.lastOrUndefined(t);if(void 0===a)return e.rangeIsOnSingleLine(n,i)?0:1;if(!(e.positionIsSynthesized(n.pos)||e.nodeIsSynthesized(a)||a.parent&&a.parent!==n)){if(G){var o=e.isNodeArray(t)&&!e.positionIsSynthesized(t.end)?t.end:a.end;return Xn((function(t){return e.getLinesBetweenPositionAndNextNonWhitespaceCharacter(o,n.end,i,t)}))}return e.rangeEndPositionsAreOnSameLine(n,a,i)?0:1}if(Zn(a,r))return 1}return 1&r&&!(131072&r)?1:0}function Xn(n){e.Debug.assert(!!G);var t=n(!0);return 0===t?n(!1):t}function Yn(e,n){var t=G&&zn(n,[e],0);return t&&Kn(t,!1),!!t}function Qn(e,n){var t=G&&Jn(n,[e],0);t&&wn(t)}function Zn(n,t){if(e.nodeIsSynthesized(n)){var r=e.getStartsOnNewLine(n);return void 0===r?0!=(65536&t):r}return 0!=(65536&t)}function $n(n,t,r){return 131072&e.getEmitFlags(n)?0:(n=nt(n),t=nt(t),r=nt(r),e.getStartsOnNewLine(r)?1:e.nodeIsSynthesized(n)||e.nodeIsSynthesized(t)||e.nodeIsSynthesized(r)?0:G?Xn((function(n){return e.getLinesBetweenRangeEndAndRangeStart(t,r,i,n)})):e.rangeEndIsOnSameLineAsRangeStart(t,r,i)?0:1)}function et(n){return 0===n.statements.length&&e.rangeEndIsOnSameLineAsRangeStart(n,n,i)}function nt(n){for(;207===n.kind&&e.nodeIsSynthesized(n);)n=n.expression;return n}function tt(n,t){return e.isGeneratedIdentifier(n)?ut(n):(e.isIdentifier(n)||e.isPrivateIdentifier(n))&&(e.nodeIsSynthesized(n)||!n.parent||!i||n.parent&&i&&e.getSourceFileOfNode(n)!==e.getOriginalNode(i))?e.idText(n):10===n.kind&&n.textSourceNode?tt(n.textSourceNode,t):!e.isLiteralExpression(n)||!e.nodeIsSynthesized(n)&&n.parent?e.getSourceTextOfNodeFromSourceFile(i,n,t):n.text}function rt(t,r,a){if(10===t.kind&&t.textSourceNode){var o=t.textSourceNode;if(e.isIdentifier(o)||e.isNumericLiteral(o)){var s=e.isNumericLiteral(o)?o.text:tt(o);return a?'"'+e.escapeJsxAttributeString(s)+'"':r||16777216&e.getEmitFlags(t)?'"'+e.escapeString(s)+'"':'"'+e.escapeNonAsciiString(s)+'"'}return rt(o,r,a)}var c=(r?1:0)|(a?2:0)|(n.terminateUnterminatedLiterals?4:0);return e.getLiteralText(t,i,c)}function at(n){n&&524288&e.getEmitFlags(n)||(l.push(u),u=0,d.push(p))}function it(n){n&&524288&e.getEmitFlags(n)||(u=l.pop(),p=d.pop())}function ot(n){p&&p!==e.lastOrUndefined(d)||(p=new e.Set),p.add(n)}function st(n){if(n)switch(n.kind){case 230:e.forEach(n.statements,st);break;case 245:case 243:case 235:case 236:st(n.statement);break;case 234:st(n.thenStatement),st(n.elseStatement);break;case 237:case 239:case 238:st(n.initializer),st(n.statement);break;case 244:st(n.caseBlock);break;case 258:e.forEach(n.clauses,st);break;case 284:case 285:e.forEach(n.statements,st);break;case 247:st(n.tryBlock),st(n.catchClause),st(n.finallyBlock);break;case 287:st(n.variableDeclaration),st(n.block);break;case 232:st(n.declarationList);break;case 250:e.forEach(n.declarations,st);break;case 249:case 160:case 198:case 252:lt(n.name);break;case 251:lt(n.name),524288&e.getEmitFlags(n)&&(e.forEach(n.parameters,st),st(n.body));break;case 196:case 197:e.forEach(n.elements,st);break;case 261:st(n.importClause);break;case 262:lt(n.name),st(n.namedBindings);break;case 263:case 269:lt(n.name);break;case 264:e.forEach(n.elements,st);break;case 265:lt(n.propertyName||n.name)}}function ct(e){if(e)switch(e.kind){case 288:case 289:case 163:case 165:case 167:case 168:lt(e.name)}}function lt(n){n&&(e.isGeneratedIdentifier(n)?ut(n):e.isBindingPattern(n)&&st(n))}function ut(n){if(4==(7&n.autoGenerateFlags))return dt(function(n){var t=n.autoGenerateId,r=n,a=r.original;for(;a&&(r=a,!(e.isIdentifier(r)&&4&r.autoGenerateFlags&&r.autoGenerateId!==t));)a=r.original;return r}(n),n.autoGenerateFlags);var t=n.autoGenerateId;return s[t]||(s[t]=function(n){switch(7&n.autoGenerateFlags){case 1:return ft(0,!!(8&n.autoGenerateFlags));case 2:return ft(268435456,!!(8&n.autoGenerateFlags));case 3:return _t(e.idText(n),32&n.autoGenerateFlags?mt:pt,!!(16&n.autoGenerateFlags),!!(8&n.autoGenerateFlags))}return e.Debug.fail("Unsupported GeneratedIdentifierKind.")}(n))}function dt(n,t){var r=e.getNodeId(n);return o[r]||(o[r]=function(n,t){switch(n.kind){case 78:return _t(tt(n),pt,!!(16&t),!!(8&t));case 256:case 255:return function(n){var t=tt(n.name);return function(n,t){for(var r=t;e.isNodeDescendantOf(r,t);r=r.nextContainer)if(r.locals){var a=r.locals.get(e.escapeLeadingUnderscores(n));if(a&&3257279&a.flags)return!1}return!0}(t,n)?t:_t(t)}(n);case 261:case 267:return function(n){var t=e.getExternalModuleName(n);return _t(e.isStringLiteral(t)?e.makeIdentifierFromModuleName(t.text):"module")}(n);case 251:case 252:case 266:return _t("default");case 221:return _t("class");case 165:case 167:case 168:return function(n){if(e.isIdentifier(n.name))return dt(n.name);return ft(0)}(n);case 158:return ft(0,!0);default:return ft(0)}}(n,t))}function pt(e){return mt(e)&&!c.has(e)&&!(p&&p.has(e))}function mt(n){return!i||e.isFileLevelUniqueName(i,n,x)}function ft(e,n){if(e&&!(u&e)&&pt(t=268435456===e?"_i":"_n"))return u|=e,n&&ot(t),t;for(;;){var t,r=268435455&u;if(u++,8!==r&&13!==r)if(pt(t=r<26?"_"+String.fromCharCode(97+r):"_"+(r-26)))return n&&ot(t),t}}function _t(e,n,t,r){if(void 0===n&&(n=pt),t&&n(e))return r?ot(e):c.add(e),e;95!==e.charCodeAt(e.length-1)&&(e+="_");for(var a=1;;){var i=e+a;if(n(i))return r?ot(i):c.add(i),i;a++}}function gt(e){return _t(e,mt,!0)}function yt(n,t){e.Debug.assert(T===t||S===t),ee(),Q=!1;var r=e.getEmitFlags(t),a=e.getCommentRange(t),i=a.pos,o=a.end,s=335!==t.kind,c=i<0||0!=(512&r)||11===t.kind,l=o<0||0!=(1024&r)||11===t.kind,u=J,d=X,p=Y;(i>0||o>0)&&i!==o&&(c||Tt(i,s),(!c||i>=0&&0!=(512&r))&&(J=i),(!l||o>=0&&0!=(1024&r))&&(X=o,250===t.kind&&(Y=o))),e.forEach(e.getSyntheticLeadingComments(t),vt),ne();var m=Ae(2,n,t);2048&r?(Z=!0,m(n,t),Z=!1):m(n,t),ee(),e.forEach(e.getSyntheticTrailingComments(t),ht),(i>0||o>0)&&i!==o&&(J=u,X=d,Y=p,!l&&s&&function(e){kt(e,Ct)}(o)),ne(),e.Debug.assert(T===t||S===t)}function vt(e){(e.hasLeadingNewline||2===e.kind)&&f.writeLine(),bt(e),e.hasTrailingNewLine||2===e.kind?f.writeLine():f.writeSpace(" ")}function ht(e){f.isAtStartOfLine()||f.writeSpace(" "),bt(e),e.hasTrailingNewLine&&f.writeLine()}function bt(n){var t=function(e){return 3===e.kind?"/*"+e.text+"*/":"//"+e.text}(n),r=3===n.kind?e.computeLineStarts(t):void 0;e.writeCommentRange(t,r,f,0,t.length,M)}function Et(n,t,r){ee();var a,o,s=t.pos,c=t.end,l=e.getEmitFlags(n),u=Z||c<0||0!=(1024&l);s<0||0!=(512&l)||(a=t,(o=e.emitDetachedComments(i.text,he(),f,Ot,a,M,Z))&&(E?E.push(o):E=[o])),ne(),2048&l&&!Z?(Z=!0,r(n),Z=!1):r(n),ee(),u||(Tt(t.end,!0),Q&&!f.isAtStartOfLine()&&f.writeLine()),ne()}function Tt(e,n){Q=!1,n?0===e&&(null==i?void 0:i.isDeclarationFile)?Rt(e,xt):Rt(e,At):0===e&&Rt(e,St)}function St(e,n,t,r,a){Pt(e,n)&&At(e,n,t,r,a)}function xt(e,n,t,r,a){Pt(e,n)||At(e,n,t,r,a)}function Lt(t,r){return!n.onlyPrintJsDocStyle||(e.isJSDocLikeText(t,r)||e.isPinnedComment(t,r))}function At(n,t,r,a,o){Lt(i.text,n)&&(Q||(e.emitNewLineBeforeLeadingCommentOfPosition(he(),f,o,n),Q=!0),Ft(n),e.writeCommentRange(i.text,he(),f,n,t,M),Ft(t),a?f.writeLine():3===r&&f.writeSpace(" "))}function Dt(e){Z||-1===e||Tt(e,!0)}function Ct(n,t,r,a){Lt(i.text,n)&&(f.isAtStartOfLine()||f.writeSpace(" "),Ft(n),e.writeCommentRange(i.text,he(),f,n,t,M),Ft(t),a&&f.writeLine())}function It(e,n){Z||(ee(),kt(e,n?Ct:Nt),ne())}function Nt(n,t,r,a){Ft(n),e.writeCommentRange(i.text,he(),f,n,t,M),Ft(t),a?f.writeLine():f.writeSpace(" ")}function Rt(n,t){!i||-1!==J&&n===J||(function(n){return void 0!==E&&e.last(E).nodePos===n}(n)?function(n){var t=e.last(E).detachedCommentEndPos;E.length-1?E.pop():E=void 0;e.forEachLeadingCommentRange(i.text,t,n,t)}(t):e.forEachLeadingCommentRange(i.text,n,t,n))}function kt(n,t){i&&(-1===X||n!==X&&n!==Y)&&e.forEachTrailingCommentRange(i.text,n,t)}function Ot(n,t,r,a,o,s){Lt(i.text,a)&&(Ft(a),e.writeCommentRange(n,t,r,a,o,s),Ft(o))}function Pt(n,t){return e.isRecognizedTripleSlashComment(i.text,n,t)}function Mt(n,t){e.Debug.assert(T===t||S===t);var r=Ae(3,n,t);if(e.isUnparsedSource(t)||e.isUnparsedPrepend(t))r(n,t);else if(e.isUnparsedNode(t)){var a=function(n){return void 0===n.parsedSourceMap&&void 0!==n.sourceMapText&&(n.parsedSourceMap=e.tryParseRawSourceMap(n.sourceMapText)||!1),n.parsedSourceMap||void 0}(t.parent);a&&y&&y.appendSourceMap(f.getLine(),f.getColumn(),a,t.parent.sourceMapPath,t.parent.getLineAndCharacterOfPosition(t.pos),t.parent.getLineAndCharacterOfPosition(t.end)),r(n,t)}else{var i=e.getSourceMapRange(t),o=i.pos,s=i.end,c=i.source,l=void 0===c?v:c,u=e.getEmitFlags(t);335!==t.kind&&0==(16&u)&&o>=0&&Gt(l,wt(l,o)),64&u?(W=!0,r(n,t),W=!1):r(n,t),335!==t.kind&&0==(32&u)&&s>=0&&Gt(l,s)}e.Debug.assert(T===t||S===t)}function wt(n,t){return n.skipTrivia?n.skipTrivia(t):e.skipTrivia(n.text,t)}function Ft(n){if(!(W||e.positionIsSynthesized(n)||Ut(v))){var t=e.getLineAndCharacterOfPosition(v,n),r=t.line,a=t.character;y.addMapping(f.getLine(),f.getColumn(),z,r,a,void 0)}}function Gt(e,n){if(e!==v){var t=v,r=z;Bt(e),Ft(n),function(e,n){v=e,z=n}(t,r)}else Ft(n)}function Bt(e){W||(v=e,e!==h?Ut(e)||(z=y.addSource(e.fileName),n.inlineSources&&y.setSourceContent(z,e.text),h=e,q=z):z=q)}function Ut(n){return e.fileExtensionIs(n.fileName,".json")}}e.isBuildInfoFile=function(n){return e.fileExtensionIs(n,".tsbuildinfo")},e.forEachEmittedFile=i,e.getTsBuildInfoEmitOutputFilePath=o,e.getOutputPathsForBundle=s,e.getOutputPathsFor=c,e.getOutputExtension=u,e.getOutputDeclarationFileName=p,e.getAllProjectOutputs=function(n,t){var r=f(),a=r.addOutput,i=r.getOutputs;if(e.outFile(n.options))_(n,a);else{for(var s=0,c=n.fileNames;s<c.length;s++){g(n,c[s],t,a)}a(o(n.options))}return i()},e.getOutputFileNames=function(n,t,r){t=e.normalizePath(t),e.Debug.assert(e.contains(n.fileNames,t),"Expected fileName to be present in command line");var a=f(),i=a.addOutput,o=a.getOutputs;return e.outFile(n.options)?_(n,i):g(n,t,r,i),o()},e.getFirstProjectOutput=function(n,t){if(e.outFile(n.options)){var r=s(n.options,!1).jsFilePath;return e.Debug.checkDefined(r,"project "+n.options.configFilePath+" expected to have at least one output")}for(var a=0,i=n.fileNames;a<i.length;a++){var c=i[a];if(!e.fileExtensionIs(c,".d.ts")){if(r=m(c,n,t))return r;if(!e.fileExtensionIs(c,".json")&&e.getEmitDeclarations(n.options))return p(c,n,t)}}var l=o(n.options);return l||e.Debug.fail("project "+n.options.configFilePath+" expected to have at least one output")},e.emitFiles=y,e.getBuildInfoText=v,e.getBuildInfo=h,e.notImplementedResolver={hasGlobalName:e.notImplemented,getReferencedExportContainer:e.notImplemented,getReferencedImportDeclaration:e.notImplemented,getReferencedDeclarationWithCollidingName:e.notImplemented,isDeclarationWithCollidingName:e.notImplemented,isValueAliasDeclaration:e.notImplemented,isReferencedAliasDeclaration:e.notImplemented,isTopLevelValueImportEqualsWithEntityName:e.notImplemented,getNodeCheckFlags:e.notImplemented,isDeclarationVisible:e.notImplemented,isLateBound:function(e){return!1},collectLinkedAliases:e.notImplemented,isImplementationOfOverload:e.notImplemented,isRequiredInitializedParameter:e.notImplemented,isOptionalUninitializedParameterProperty:e.notImplemented,isExpandoFunctionDeclaration:e.notImplemented,getPropertiesOfContainerFunction:e.notImplemented,createTypeOfDeclaration:e.notImplemented,createReturnTypeOfSignatureDeclaration:e.notImplemented,createTypeOfExpression:e.notImplemented,createLiteralConstValue:e.notImplemented,isSymbolAccessible:e.notImplemented,isEntityNameVisible:e.notImplemented,getConstantValue:e.notImplemented,getReferencedValueDeclaration:e.notImplemented,getTypeReferenceSerializationKind:e.notImplemented,isOptionalParameter:e.notImplemented,moduleExportsSomeValue:e.notImplemented,isArgumentsLocalBinding:e.notImplemented,getExternalModuleFileFromDeclaration:e.notImplemented,getTypeReferenceDirectivesForEntityName:e.notImplemented,getTypeReferenceDirectivesForSymbol:e.notImplemented,isLiteralConstDeclaration:e.notImplemented,getJsxFactoryEntity:e.notImplemented,getJsxFragmentFactoryEntity:e.notImplemented,getAllAccessorDeclarations:e.notImplemented,getSymbolOfExternalModuleSpecifier:e.notImplemented,isBindingCapturedByNode:e.notImplemented,getDeclarationStatementsForSourceFile:e.notImplemented,isImportRequiredByAugmentation:e.notImplemented},e.emitUsingBuildInfo=function(n,t,r,a){var i=s(n.options,!1),o=i.buildInfoPath,c=i.jsFilePath,l=i.sourceMapFilePath,u=i.declarationFilePath,d=i.declarationMapPath,p=t.readFile(e.Debug.checkDefined(o));if(!p)return o;var m=t.readFile(e.Debug.checkDefined(c));if(!m)return c;var f=l&&t.readFile(l);if(l&&!f||n.options.inlineSourceMap)return l||"inline sourcemap decoding";var _=u&&t.readFile(u);if(u&&!_)return u;var g=d&&t.readFile(d);if(d&&!g||n.options.inlineSourceMap)return d||"inline sourcemap decoding";var b=h(p);if(!b.bundle||!b.bundle.js||_&&!b.bundle.dts)return o;var E=e.getDirectoryPath(e.getNormalizedAbsolutePath(o,t.getCurrentDirectory())),T=e.createInputFiles(m,_,l,f,d,g,c,u,o,b,!0),S=[],x=e.createPrependNodes(n.projectReferences,r,(function(e){return t.readFile(e)})),L=function(n,t,r){var a,i=e.Debug.checkDefined(n.js),o=(null===(a=i.sources)||void 0===a?void 0:a.prologues)&&e.arrayToMap(i.sources.prologues,(function(e){return e.file}));return n.sourceFiles.map((function(n,a){var i,s,c=null==o?void 0:o.get(a),l=null==c?void 0:c.directives.map((function(n){var t=e.setTextRange(e.factory.createStringLiteral(n.expression.text),n.expression),r=e.setTextRange(e.factory.createExpressionStatement(t),n);return e.setParent(t,r),r})),u=e.factory.createToken(1),d=e.factory.createSourceFile(null!=l?l:[],u,0);return d.fileName=e.getRelativePathFromDirectory(r.getCurrentDirectory(),e.getNormalizedAbsolutePath(n,t),!r.useCaseSensitiveFileNames()),d.text=null!==(i=null==c?void 0:c.text)&&void 0!==i?i:"",e.setTextRangePosWidth(d,0,null!==(s=null==c?void 0:c.text.length)&&void 0!==s?s:0),e.setEachParent(d.statements,d),e.setTextRangePosWidth(u,d.end,0),e.setParent(u,d),d}))}(b.bundle,E,t),A={getPrependNodes:e.memoize((function(){return __spreadArrays(x,[T])})),getCanonicalFileName:t.getCanonicalFileName,getCommonSourceDirectory:function(){return e.getNormalizedAbsolutePath(b.bundle.commonSourceDirectory,E)},getCompilerOptions:function(){return n.options},getCurrentDirectory:function(){return t.getCurrentDirectory()},getNewLine:function(){return t.getNewLine()},getSourceFile:e.returnUndefined,getSourceFileByPath:e.returnUndefined,getSourceFiles:function(){return L},getLibFileFromReference:e.notImplemented,isSourceFileFromExternalLibrary:e.returnFalse,getResolvedProjectReferenceToRedirect:e.returnUndefined,getProjectReferenceRedirect:e.returnUndefined,isSourceOfProjectReferenceRedirect:e.returnFalse,writeFile:function(n,t,r){switch(n){case c:if(m===t)return;break;case l:if(f===t)return;break;case o:var a=h(t);a.program=b.program;var i=b.bundle,s=i.js,p=i.dts,y=i.sourceFiles;return a.bundle.js.sources=s.sources,p&&(a.bundle.dts.sources=p.sources),a.bundle.sourceFiles=y,void S.push({name:n,text:v(a),writeByteOrderMark:r});case u:if(_===t)return;break;case d:if(g===t)return;break;default:e.Debug.fail("Unexpected path: "+n)}S.push({name:n,text:t,writeByteOrderMark:r})},isEmitBlocked:e.returnFalse,readFile:function(e){return t.readFile(e)},fileExists:function(e){return t.fileExists(e)},useCaseSensitiveFileNames:function(){return t.useCaseSensitiveFileNames()},getProgramBuildInfo:e.returnUndefined,getSourceFileFromReference:e.returnUndefined,redirectTargetsMap:e.createMultiMap()};return y(e.notImplementedResolver,A,void 0,e.getTransformers(n.options,a)),S},function(e){e[e.Notification=0]="Notification",e[e.Substitution=1]="Substitution",e[e.Comments=2]="Comments",e[e.SourceMaps=3]="SourceMaps",e[e.Emit=4]="Emit"}(n||(n={})),e.createPrinter=b,function(e){e[e.Auto=0]="Auto",e[e.CountMask=268435455]="CountMask",e[e._i=268435456]="_i"}(t||(t={}))}(ts||(ts={})),function(e){var n;function t(e,n,t,r,a){return e.watchFile(n,t,r,a)}function r(e,n,r,a,i,o){return t(e,n,(function(e,n){return r(e,n,o)}),a,i)}function a(e,n,t,r,a){return e.watchDirectory(n,t,0!=(1&r),a)}function i(e,t){switch(e){case n.None:return t;case n.TriggerOnly:return c;case n.Verbose:return t===a?s:o}}function o(e,n,t,r,a,i,o,s,u,d,p,m){d(p+":: Added:: "+l(n,r,a,o,s,m));var f=c(e,n,t,r,a,i,o,s,u,d,p,m);return{close:function(){d(p+":: Close:: "+l(n,r,a,o,s,m)),f.close()}}}function s(n,t,r,a,i,o,s,u,d,p,m,f){var _=m+":: Added:: "+l(t,a,i,s,u,f);p(_);var g=e.timestamp(),y=c(n,t,r,a,i,o,s,u,d,p,m,f),v=e.timestamp()-g;return p("Elapsed:: "+v+"ms "+_),{close:function(){var n=m+":: Close:: "+l(t,a,i,s,u,f);p(n);var r=e.timestamp();y.close();var o=e.timestamp()-r;p("Elapsed:: "+o+"ms "+n)}}}function c(n,t,r,a,i,o,s,c,u,d,p,m){return u(n,t,(function(n,u){var f=p+":: Triggered with "+n+" "+(void 0!==u?u:"")+":: "+l(t,a,i,s,c,m);d(f);var _=e.timestamp();r(n,u,o);var g=e.timestamp()-_;d("Elapsed:: "+g+"ms "+f)}),a,i)}function l(e,n,t,r,a,i){return"WatchInfo: "+e+" "+n+" "+JSON.stringify(t)+" "+(i?i(r,a):void 0===a?r:r+" "+a)}function u(e){e.watcher.close()}e.createCachedDirectoryStructureHost=function(n,t,r){if(n.getDirectories&&n.readDirectory){var a=new e.Map,i=e.createGetCanonicalFileName(r);return{useCaseSensitiveFileNames:r,fileExists:function(e){var t=c(o(e));return t&&p(t.files,l(e))||n.fileExists(e)},readFile:function(e,t){return n.readFile(e,t)},directoryExists:n.directoryExists&&function(t){var r=o(t);return a.has(e.ensureTrailingDirectorySeparator(r))||n.directoryExists(t)},getDirectories:function(e){var t=o(e),r=u(e,t);if(r)return r.directories.slice();return n.getDirectories(e)},readDirectory:function(a,i,s,c,l){var d=o(a),p=u(a,d);if(p)return e.matchFiles(a,i,s,c,r,t,l,(function(n){var t=o(n);if(t===d)return p;return u(n,t)||e.emptyFileSystemEntries}),f);return n.readDirectory(a,i,s,c,l)},createDirectory:n.createDirectory&&function(e){var t=c(o(e)),r=l(e);t&&m(t.directories,r,!0);n.createDirectory(e)},writeFile:n.writeFile&&function(e,t,r){var a=c(o(e));a&&_(a,l(e),!0);return n.writeFile(e,t,r)},addOrDeleteFileOrDirectory:function(e,t){if(s(t))return void g();var r=c(t);if(!r)return;if(!n.directoryExists)return void g();var a=l(e),i={fileExists:n.fileExists(t),directoryExists:n.directoryExists(t)};i.directoryExists||p(r.directories,a)?g():_(r,a,i.fileExists);return i},addOrDeleteFile:function(n,t,r){if(r===e.FileWatcherEventKind.Changed)return;var a=c(t);a&&_(a,l(n),r===e.FileWatcherEventKind.Created)},clearCache:g,realpath:n.realpath&&f}}function o(n){return e.toPath(n,t,i)}function s(n){return a.get(e.ensureTrailingDirectorySeparator(n))}function c(n){return s(e.getDirectoryPath(n))}function l(n){return e.getBaseFileName(e.normalizePath(n))}function u(t,r){var i=s(r=e.ensureTrailingDirectorySeparator(r));if(i)return i;try{return function(t,r){var i={files:e.map(n.readDirectory(t,void 0,void 0,["*.*"]),l)||[],directories:n.getDirectories(t)||[]};return a.set(e.ensureTrailingDirectorySeparator(r),i),i}(t,r)}catch(n){return void e.Debug.assert(!a.has(e.ensureTrailingDirectorySeparator(r)))}}function d(e,n){return i(e)===i(n)}function p(n,t){return e.some(n,(function(e){return d(e,t)}))}function m(n,t,r){if(p(n,t)){if(!r)return e.filterMutate(n,(function(e){return!d(e,t)}))}else if(r)return n.push(t)}function f(e){return n.realpath?n.realpath(e):e}function _(e,n,t){m(e.files,n,t)}function g(){a.clear()}},function(e){e[e.None=0]="None",e[e.Partial=1]="Partial",e[e.Full=2]="Full"}(e.ConfigFileProgramReloadLevel||(e.ConfigFileProgramReloadLevel={})),e.updateMissingFilePathsWatch=function(n,t,r){var a=n.getMissingFilePaths(),i=e.arrayToMap(a,e.identity,e.returnTrue);e.mutateMap(t,i,{createNewValue:r,onDeleteValue:e.closeFileWatcher})},e.updateWatchingWildcardDirectories=function(n,t,r){function a(e,n){return{watcher:r(e,n),flags:n}}e.mutateMap(n,t,{createNewValue:a,onDeleteValue:u,onExistingValue:function(e,t,r){if(e.flags===t)return;e.watcher.close(),n.set(r,a(r,t))}})},e.isIgnoredFileFromWildCardWatching=function(n){var t=n.watchedDirPath,r=n.fileOrDirectory,a=n.fileOrDirectoryPath,i=n.configFileName,o=n.options,s=n.configFileSpecs,c=n.program,l=n.extraFileExtensions,u=n.currentDirectory,d=n.useCaseSensitiveFileNames,p=n.writeLog,m=e.removeIgnoredPath(a);if(!m)return p("Project: "+i+" Detected ignored path: "+r),!0;if((a=m)===t)return!1;if(e.hasExtension(a)&&!e.isSupportedSourceFileName(r,o,l))return p("Project: "+i+" Detected file add/remove of non supported extension: "+r),!0;if(e.isExcludedFile(r,s,e.getNormalizedAbsolutePath(e.getDirectoryPath(i),u),d,u))return p("Project: "+i+" Detected excluded file: "+r),!0;if(!c)return!1;if(o.outFile||o.outDir)return!1;if(e.fileExtensionIs(a,".d.ts")){if(o.declarationDir)return!1}else if(!e.fileExtensionIsOneOf(a,e.supportedJSExtensions))return!1;var f=e.removeFileExtension(a),_=function(e){return!!e.getState}(c)?c.getProgramOrUndefined():c;return!(!g(f+".ts")&&!g(f+".tsx"))&&(p("Project: "+i+" Detected output file: "+r),!0);function g(e){return _?!!_.getSourceFileByPath(e):c.getState().fileInfos.has(e)}},e.isEmittedFileOfProgram=function(e,n){return!!e&&e.isEmittedFile(n)},function(e){e[e.None=0]="None",e[e.TriggerOnly=1]="TriggerOnly",e[e.Verbose=2]="Verbose"}(n=e.WatchLogLevel||(e.WatchLogLevel={})),e.getWatchFactory=function(o,s,c){return function(t,a,o,s,c){var l=i(t,s),u=t===n.None?r:l,d=i(t,c);t===n.Verbose&&e.sysLog===e.noop&&e.setSysLog((function(e){return a(e)}));return{watchFile:function(e,n,t,r,i,c,u){return l(e,n,t,r,i,void 0,c,u,s,a,"FileWatcher",o)},watchFilePath:function(e,n,t,r,i,c,l,d){return u(e,n,t,r,i,c,l,d,s,a,"FileWatcher",o)},watchDirectory:function(e,n,t,r,i,s,l){return d(e,n,t,r,i,void 0,s,l,c,a,"DirectoryWatcher",o)}}}(o,s,c,t,a)},e.getFallbackOptions=function(n){var t=null==n?void 0:n.fallbackPolling;return{watchFile:void 0!==t?t:e.WatchFileKind.PriorityPollingInterval}},e.closeFileWatcherOf=u}(ts||(ts={})),function(e){function n(n,t){var r=e.getDirectoryPath(t),a=e.isRootedDiskPath(n)?n:e.combinePaths(r,n);return e.normalizePath(a)}function t(n,t,r){var a;return e.forEach(n,(function(n){var i=e.getNormalizedPathComponents(n,t);if(i.pop(),a){for(var o=Math.min(a.length,i.length),s=0;s<o;s++)if(r(a[s])!==r(i[s])){if(0===s)return!0;a.length=s;break}i.length<a.length&&(a.length=i.length)}else a=i}))?"":a?e.getPathFromPathComponents(a):t}function r(e,n){return a(e,n)}function a(n,t,r){void 0===r&&(r=e.sys);var a,i=new e.Map,o=e.createGetCanonicalFileName(r.useCaseSensitiveFileNames),s=e.maybeBind(r,r.createHash)||e.generateDjb2Hash;function c(){return e.getDirectoryPath(e.normalizePath(r.getExecutingFilePath()))}var l=e.getNewLineCharacter(n,(function(){return r.newLine})),u=r.realpath&&function(e){return r.realpath(e)},d={getSourceFile:function(n,r,a){var i;try{e.performance.mark("beforeIORead"),i=d.readFile(n),e.performance.mark("afterIORead"),e.performance.measure("I/O Read","beforeIORead","afterIORead")}catch(e){a&&a(e.message),i=""}return void 0!==i?e.createSourceFile(n,i,r,t):void 0},getDefaultLibLocation:c,getDefaultLibFileName:function(n){return e.combinePaths(c(),e.getDefaultLibFileName(n))},writeFile:function(t,o,c,l){try{e.performance.mark("beforeIOWrite"),e.writeFileEnsuringDirectories(t,o,c,(function(t,i,o){return function(t,i,o){if(!e.isWatchSet(n)||!r.getModifiedTime)return void r.writeFile(t,i,o);a||(a=new e.Map);var c=s(i),l=r.getModifiedTime(t);if(l){var u=a.get(t);if(u&&u.byteOrderMark===o&&u.hash===c&&u.mtime.getTime()===l.getTime())return}r.writeFile(t,i,o);var d=r.getModifiedTime(t)||e.missingFileModifiedTime;a.set(t,{hash:c,byteOrderMark:o,mtime:d})}(t,i,o)}),(function(e){return(d.createDirectory||r.createDirectory)(e)}),(function(e){return n=e,!!i.has(n)||!!(d.directoryExists||r.directoryExists)(n)&&(i.set(n,!0),!0);var n})),e.performance.mark("afterIOWrite"),e.performance.measure("I/O Write","beforeIOWrite","afterIOWrite")}catch(e){l&&l(e.message)}},getCurrentDirectory:e.memoize((function(){return r.getCurrentDirectory()})),useCaseSensitiveFileNames:function(){return r.useCaseSensitiveFileNames},getCanonicalFileName:o,getNewLine:function(){return l},fileExists:function(e){return r.fileExists(e)},readFile:function(e){return r.readFile(e)},trace:function(e){return r.write(e+l)},directoryExists:function(e){return r.directoryExists(e)},getEnvironmentVariable:function(e){return r.getEnvironmentVariable?r.getEnvironmentVariable(e):""},getDirectories:function(e){return r.getDirectories(e)},realpath:u,readDirectory:function(e,n,t,a,i){return r.readDirectory(e,n,t,a,i)},createDirectory:function(e){return r.createDirectory(e)},createHash:e.maybeBind(r,r.createHash)};return d}function i(n,t){var r=e.diagnosticCategoryName(n)+" TS"+n.code+": "+f(n.messageText,t.getNewLine())+t.getNewLine();if(n.file){var a=e.getLineAndCharacterOfPosition(n.file,n.start),i=a.line,o=a.character,s=n.file.fileName;return e.convertToRelativePath(s,t.getCurrentDirectory(),(function(e){return t.getCanonicalFileName(e)}))+"("+(i+1)+","+(o+1)+"): "+r}return r}var o;e.findConfigFile=function(n,t,r){return void 0===r&&(r="tsconfig.json"),e.forEachAncestorDirectory(n,(function(n){var a=e.combinePaths(n,r);return t(a)?a:void 0}))},e.resolveTripleslashReference=n,e.computeCommonSourceDirectoryOfFilenames=t,e.createCompilerHost=r,e.createCompilerHostWorker=a,e.changeCompilerHostLikeToUseCache=function(n,t,r){var a=n.readFile,i=n.fileExists,o=n.directoryExists,s=n.createDirectory,c=n.writeFile,l=new e.Map,u=new e.Map,d=new e.Map,p=new e.Map,m=function(e,t){var r=a.call(n,t);return l.set(e,void 0!==r&&r),r};n.readFile=function(r){var i=t(r),o=l.get(i);return void 0!==o?!1!==o?o:void 0:e.fileExtensionIs(r,".json")||e.isBuildInfoFile(r)?m(i,r):a.call(n,r)};var f=r?function(n,a,i,o){var s=t(n),c=p.get(s);if(c)return c;var l=r(n,a,i,o);return l&&(e.isDeclarationFileName(n)||e.fileExtensionIs(n,".json"))&&p.set(s,l),l}:void 0;return n.fileExists=function(e){var r=t(e),a=u.get(r);if(void 0!==a)return a;var o=i.call(n,e);return u.set(r,!!o),o},c&&(n.writeFile=function(e,r,a,i,o){var s=t(e);u.delete(s);var d=l.get(s);if(void 0!==d&&d!==r)l.delete(s),p.delete(s);else if(f){var m=p.get(s);m&&m.text!==r&&p.delete(s)}c.call(n,e,r,a,i,o)}),o&&s&&(n.directoryExists=function(e){var r=t(e),a=d.get(r);if(void 0!==a)return a;var i=o.call(n,e);return d.set(r,!!i),i},n.createDirectory=function(e){var r=t(e);d.delete(r),s.call(n,e)}),{originalReadFile:a,originalFileExists:i,originalDirectoryExists:o,originalCreateDirectory:s,originalWriteFile:c,getSourceFileWithCache:f,readFileWithCache:function(e){var n=t(e),r=l.get(n);return void 0!==r?!1!==r?r:void 0:m(n,e)}}},e.getPreEmitDiagnostics=function(n,t,r){var a;return a=e.addRange(a,n.getConfigFileParsingDiagnostics()),a=e.addRange(a,n.getOptionsDiagnostics(r)),a=e.addRange(a,n.getSyntacticDiagnostics(t,r)),a=e.addRange(a,n.getGlobalDiagnostics(r)),a=e.addRange(a,n.getSemanticDiagnostics(t,r)),e.getEmitDeclarations(n.getCompilerOptions())&&(a=e.addRange(a,n.getDeclarationDiagnostics(t,r))),e.sortAndDeduplicateDiagnostics(a||e.emptyArray)},e.formatDiagnostics=function(e,n){for(var t="",r=0,a=e;r<a.length;r++){t+=i(a[r],n)}return t},e.formatDiagnostic=i,function(e){e.Grey="[90m",e.Red="[91m",e.Yellow="[93m",e.Blue="[94m",e.Cyan="[96m"}(o=e.ForegroundColorEscapeSequences||(e.ForegroundColorEscapeSequences={}));var s="[7m",c="[0m",l="    ";function u(n){switch(n){case e.DiagnosticCategory.Error:return o.Red;case e.DiagnosticCategory.Warning:return o.Yellow;case e.DiagnosticCategory.Suggestion:return e.Debug.fail("Should never get an Info diagnostic on the command line.");case e.DiagnosticCategory.Message:return o.Blue}}function d(e,n){return n+e+c}function p(n,t,r,a,i,o){var l=e.getLineAndCharacterOfPosition(n,t),u=l.line,p=l.character,m=e.getLineAndCharacterOfPosition(n,t+r),f=m.line,_=m.character,g=e.getLineAndCharacterOfPosition(n,n.text.length).line,y=f-u>=4,v=(f+1+"").length;y&&(v=Math.max("...".length,v));for(var h="",b=u;b<=f;b++){h+=o.getNewLine(),y&&u+1<b&&b<f-1&&(h+=a+d(e.padLeft("...",v),s)+" "+o.getNewLine(),b=f-1);var E=e.getPositionOfLineAndCharacter(n,b,0),T=b<g?e.getPositionOfLineAndCharacter(n,b+1,0):n.text.length,S=n.text.slice(E,T);if(S=(S=S.replace(/\s+$/g,"")).replace("\t"," "),h+=a+d(e.padLeft(b+1+"",v),s)+" ",h+=S+o.getNewLine(),h+=a+d(e.padLeft("",v),s)+" ",h+=i,b===u){var x=b===f?_:void 0;h+=S.slice(0,p).replace(/\S/g," "),h+=S.slice(p,x).replace(/./g,"~")}else h+=b===f?S.slice(0,_).replace(/./g,"~"):S.replace(/./g,"~");h+=c}return h}function m(n,t,r,a){void 0===a&&(a=d);var i=e.getLineAndCharacterOfPosition(n,t),s=i.line,c=i.character,l="";return l+=a(r?e.convertToRelativePath(n.fileName,r.getCurrentDirectory(),(function(e){return r.getCanonicalFileName(e)})):n.fileName,o.Cyan),l+=":",l+=a(""+(s+1),o.Yellow),l+=":",l+=a(""+(c+1),o.Yellow)}function f(n,t,r){if(void 0===r&&(r=0),e.isString(n))return n;if(void 0===n)return"";var a="";if(r){a+=t;for(var i=0;i<r;i++)a+="  "}if(a+=n.messageText,r++,n.next)for(var o=0,s=n.next;o<s.length;o++){a+=f(s[o],t,r)}return a}function _(n,t,r,a){if(0===n.length)return[];for(var i=[],o=new e.Map,s=0,c=n;s<c.length;s++){var l=c[s],u=void 0;o.has(l)?u=o.get(l):o.set(l,u=a(l,t,r)),i.push(u)}return i}function g(n,t,r,a){var i;return function n(t,o,s){if(a){var c=a(t,s);if(c)return c}return e.forEach(o,(function(t,a){if(!t||!(null==i?void 0:i.has(t.sourceFile.path))){var o=r(t,s,a);return o||!t?o:((i||(i=new e.Set)).add(t.sourceFile.path),n(t.commandLine.projectReferences,t.references,t))}}))}(n,t,void 0)}function y(n,t,r,a){var i=n.getCompilerOptions();if(i.noEmit)return n.getSemanticDiagnostics(t,a),t||e.outFile(i)?e.emitSkippedWithNoDiagnostics:n.emitBuildInfo(r,a);if(i.noEmitOnError){var o=__spreadArrays(n.getOptionsDiagnostics(a),n.getSyntacticDiagnostics(t,a),n.getGlobalDiagnostics(a),n.getSemanticDiagnostics(t,a));if(0===o.length&&e.getEmitDeclarations(n.getCompilerOptions())&&(o=n.getDeclarationDiagnostics(void 0,a)),o.length){var s;if(!t&&!e.outFile(i)){var c=n.emitBuildInfo(r,a);c.diagnostics&&(o=__spreadArrays(o,c.diagnostics)),s=c.emittedFiles}return{diagnostics:o,sourceMaps:void 0,emittedFiles:s,emitSkipped:!0}}}}function v(n,t){return e.filter(n,(function(e){return!e.skippedOn||!t[e.skippedOn]}))}function h(n,t){return void 0===t&&(t=n),{fileExists:function(e){return t.fileExists(e)},readDirectory:function(n,r,a,i,o){return e.Debug.assertIsDefined(t.readDirectory,"'CompilerHost.readDirectory' must be implemented to correctly process 'projectReferences'"),t.readDirectory(n,r,a,i,o)},readFile:function(e){return t.readFile(e)},useCaseSensitiveFileNames:n.useCaseSensitiveFileNames(),getCurrentDirectory:function(){return n.getCurrentDirectory()},onUnRecoverableConfigFileDiagnostic:n.onUnRecoverableConfigFileDiagnostic||e.returnUndefined,trace:n.trace?function(e){return n.trace(e)}:void 0}}function b(n,t,r){if(!n)return e.emptyArray;for(var a,i=0;i<n.length;i++){var o=n[i],s=t(o,i);if(o.prepend&&s&&s.options){if(!e.outFile(s.options))continue;var c=e.getOutputPathsForBundle(s.options,!0),l=c.jsFilePath,u=c.sourceMapFilePath,d=c.declarationFilePath,p=c.declarationMapPath,m=c.buildInfoPath,f=e.createInputFiles(r,l,u,d,p,m);(a||(a=[])).push(f)}}return a||e.emptyArray}function E(n,t){var r=t||n;return e.resolveConfigFileProjectName(r.path)}function T(n,t){switch(t.extension){case".ts":case".d.ts":return;case".tsx":return r();case".jsx":return r()||a();case".js":return a();case".json":return n.resolveJsonModule?void 0:e.Diagnostics.Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used}function r(){return n.jsx?void 0:e.Diagnostics.Module_0_was_resolved_to_1_but_jsx_is_not_set}function a(){return e.getAllowJSCompilerOption(n)||!e.getStrictOptionValue(n,"noImplicitAny")?void 0:e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type}}function S(e){for(var n=e.imports,t=e.moduleAugmentations,r=n.map((function(e){return e.text})),a=0,i=t;a<i.length;a++){var o=i[a];10===o.kind&&r.push(o.text)}return r}e.formatColorAndReset=d,e.formatLocation=m,e.formatDiagnosticsWithColorAndContext=function(n,t){for(var r="",a=0,i=n;a<i.length;a++){var s=i[a];if(s.file)r+=m(y=s.file,v=s.start,t),r+=" - ";if(r+=d(e.diagnosticCategoryName(s),u(s.category)),r+=d(" TS"+s.code+": ",o.Grey),r+=f(s.messageText,t.getNewLine()),s.file&&(r+=t.getNewLine(),r+=p(s.file,s.start,s.length,"",u(s.category),t),s.relatedInformation)){r+=t.getNewLine();for(var c=0,_=s.relatedInformation;c<_.length;c++){var g=_[c],y=g.file,v=g.start,h=g.length,b=g.messageText;y&&(r+=t.getNewLine(),r+="  "+m(y,v,t),r+=p(y,v,h,l,o.Cyan,t)),r+=t.getNewLine(),r+=l+f(b,t.getNewLine())}}r+=t.getNewLine()}return r},e.flattenDiagnosticMessageText=f,e.loadWithLocalCache=_,e.forEachResolvedProjectReference=function(e,n){return g(void 0,e,(function(e,t){return e&&n(e,t)}))},e.inferredTypesContainingFile="__inferred type names__.ts",e.isProgramUptoDate=function(n,t,r,a,i,o,s,c){if(!n||(null==s?void 0:s()))return!1;if(!e.arrayIsEqualTo(n.getRootFileNames(),t))return!1;var l;if(!e.arrayIsEqualTo(n.getProjectReferences(),c,(function(t,r,a){if(!e.projectReferenceIsEqualTo(t,r))return!1;return p(n.getResolvedProjectReferences()[a],t)})))return!1;if(n.getSourceFiles().some((function(e){return!d(e)||o(e.path)})))return!1;if(n.getMissingFilePaths().some(i))return!1;var u=n.getCompilerOptions();return!!e.compareDataObjects(u,r)&&(!u.configFile||!r.configFile||u.configFile.text===r.configFile.text);function d(e){return e.version===a(e.resolvedPath,e.fileName)}function p(n,t){return n?!!e.contains(l,n)||!!d(n.sourceFile)&&((l||(l=[])).push(n),!e.forEach(n.references,(function(e,t){return!p(e,n.commandLine.projectReferences[t])}))):!i(E(t))}},e.getConfigFileParsingDiagnostics=function(e){return e.options.configFile?__spreadArrays(e.options.configFile.parseDiagnostics,e.errors):e.errors},e.createProgram=function(a,i,o,s,c){var l,u,d,p,m,f,x,L,A,D,C,I=e.isArray(a)?function(e,n,t,r,a){return{rootNames:e,options:n,host:t,oldProgram:r,configFileParsingDiagnostics:a}}(a,i,o,s,c):a,N=I.rootNames,R=I.options,k=I.configFileParsingDiagnostics,O=I.projectReferences,P=I.oldProgram,M=new e.Map,w={},F={},G=new e.Map,B=e.createDiagnosticCollection(),U="number"==typeof R.maxNodeModuleJsDepth?R.maxNodeModuleJsDepth:0,V=0,j=new e.Map,H=new e.Map,K=["program","createProgram"];e.tracing.begin.apply(e.tracing,K),e.performance.mark("beforeProgram");var W,z,q,J,X=I.host||r(R),Y=h(X),Q=R.noLib,Z=e.memoize((function(){return X.getDefaultLibFileName(R)})),$=X.getDefaultLibLocation?X.getDefaultLibLocation():e.getDirectoryPath(Z()),ee=e.createDiagnosticCollection(),ne=X.getCurrentDirectory(),te=e.getSupportedExtensions(R),re=e.getSuppoertedExtensionsWithJsonIfResolveJsonModule(R,te),ae=new e.Map,ie=X.hasInvalidatedResolution||e.returnFalse;if(X.resolveModuleNames)q=function(n,t,r,a){return X.resolveModuleNames(e.Debug.checkEachDefined(n),t,r,a,R).map((function(n){if(!n||void 0!==n.extension)return n;var t=e.clone(n);return t.extension=e.extensionFromPath(n.resolvedFileName),t}))};else{z=e.createModuleResolutionCache(ne,(function(e){return X.getCanonicalFileName(e)}),R);var oe=function(n,t,r){return e.resolveModuleName(n,t,R,X,z,r).resolvedModule};q=function(n,t,r,a){return _(e.Debug.checkEachDefined(n),t,a,oe)}}if(X.resolveTypeReferenceDirectives)J=function(n,t,r){return X.resolveTypeReferenceDirectives(e.Debug.checkEachDefined(n),t,r,R)};else{var se=function(n,t,r){return e.resolveTypeReferenceDirective(n,t,R,X,r).resolvedTypeReferenceDirective};J=function(n,t,r){return _(e.Debug.checkEachDefined(n),t,r,se)}}var ce,le,ue,de,pe,me=new e.Map,fe=new e.Map,_e=e.createMultiMap(),ge=new e.Map,ye=X.useCaseSensitiveFileNames()?new e.Map:void 0,ve=!!(null===(l=X.useSourceOfProjectReferenceRedirect)||void 0===l?void 0:l.call(X))&&!R.disableSourceOfProjectReferenceRedirect,he=function(n){var t,r,a=n.compilerHost.fileExists,i=n.compilerHost.directoryExists,o=n.compilerHost.getDirectories,s=n.compilerHost.realpath;if(!n.useSourceOfProjectReferenceRedirect)return{onProgramCreateComplete:e.noop,fileExists:l};n.compilerHost.fileExists=l,i&&(r=n.compilerHost.directoryExists=function(r){return i.call(n.compilerHost,r)?(p(r),!0):!!n.getResolvedProjectReferences()&&(t||(t=new e.Set,n.forEachResolvedProjectReference((function(r){var a=e.outFile(r.commandLine.options);if(a)t.add(e.getDirectoryPath(n.toPath(a)));else{var i=r.commandLine.options.declarationDir||r.commandLine.options.outDir;i&&t.add(n.toPath(i))}}))),m(r,!1))});o&&(n.compilerHost.getDirectories=function(e){return!n.getResolvedProjectReferences()||i&&i.call(n.compilerHost,e)?o.call(n.compilerHost,e):[]});s&&(n.compilerHost.realpath=function(e){var t;return(null===(t=n.getSymlinkCache().getSymlinkedFiles())||void 0===t?void 0:t.get(n.toPath(e)))||s.call(n.compilerHost,e)});return{onProgramCreateComplete:c,fileExists:l,directoryExists:r};function c(){n.compilerHost.fileExists=a,n.compilerHost.directoryExists=i,n.compilerHost.getDirectories=o}function l(t){return!!a.call(n.compilerHost,t)||!!n.getResolvedProjectReferences()&&(!!e.isDeclarationFileName(t)&&m(t,!0))}function u(t){var r=n.getSourceOfProjectReferenceRedirect(t);return void 0!==r?!e.isString(r)||a.call(n.compilerHost,r):void 0}function d(r){var a=n.toPath(r),i=""+a+e.directorySeparator;return e.forEachKey(t,(function(n){return a===n||e.startsWith(n,i)||e.startsWith(a,n+"/")}))}function p(t){var r;if(n.getResolvedProjectReferences()&&s&&e.stringContains(t,e.nodeModulesPathPart)){var a=n.getSymlinkCache(),i=e.ensureTrailingDirectorySeparator(n.toPath(t));if(!(null===(r=a.getSymlinkedDirectories())||void 0===r?void 0:r.has(i))){var o,c=e.normalizePath(s.call(n.compilerHost,t));c!==t&&(o=e.ensureTrailingDirectorySeparator(n.toPath(c)))!==i?a.setSymlinkedDirectory(i,{real:e.ensureTrailingDirectorySeparator(c),realPath:o}):a.setSymlinkedDirectory(i,!1)}}}function m(t,r){var a,i=r?function(e){return u(e)}:function(e){return d(e)},o=i(t);if(void 0!==o)return o;var s=n.getSymlinkCache(),c=s.getSymlinkedDirectories();if(!c)return!1;var l=n.toPath(t);return!!e.stringContains(l,e.nodeModulesPathPart)&&(!(!r||!(null===(a=s.getSymlinkedFiles())||void 0===a?void 0:a.has(l)))||(e.firstDefinedIterator(c.entries(),(function(a){var o=a[0],c=a[1];if(c&&e.startsWith(l,o)){var u=i(l.replace(o,c.realPath));if(r&&u){var d=e.getNormalizedAbsolutePath(t,n.compilerHost.getCurrentDirectory());s.setSymlinkedFile(l,""+c.real+d.replace(new RegExp(o,"i"),""))}return u}}))||!1))}}({compilerHost:X,getSymlinkCache:st,useSourceOfProjectReferenceRedirect:ve,toPath:Xe,getResolvedProjectReferences:$e,getSourceOfProjectReferenceRedirect:wn,forEachResolvedProjectReference:Mn}),be=he.onProgramCreateComplete,Ee=he.fileExists,Te=he.directoryExists;e.tracing.push("program","shouldProgramCreateNewSourceFiles",{hasOldProgram:!!P});var Se,xe=function(n,t){if(!n)return!1;var r=n.getCompilerOptions();return!!e.sourceFileAffectingCompilerOptions.some((function(n){return!e.isJsonEqual(e.getCompilerOptionValue(r,n),e.getCompilerOptionValue(t,n))}))}(P,R);if(e.tracing.pop(),e.tracing.push("program","tryReuseStructureFromOldProgram",{}),Se=function(){var n;if(!P)return 0;var t=P.getCompilerOptions();if(e.changesAffectModuleResolution(t,R))return 0;var r=P.getRootFileNames();if(!e.arrayIsEqualTo(r,N))return 0;if(!e.arrayIsEqualTo(R.types,t.types))return 0;if(g(P.getProjectReferences(),P.getResolvedProjectReferences(),(function(e,n,t){var r=qn((n?n.commandLine.projectReferences:O)[t]);return e?!r||r.sourceFile!==e.sourceFile:void 0!==r}),(function(n,t){var r=t?Gn(t.sourceFile.path).commandLine.projectReferences:O;return!e.arrayIsEqualTo(n,r,e.projectReferenceIsEqualTo)})))return 0;O&&(le=O.map(qn));var a=[],i=[];if(Se=2,P.getMissingFilePaths().some((function(e){return X.fileExists(e)})))return 0;var o,s=P.getSourceFiles();!function(e){e[e.Exists=0]="Exists",e[e.Modified=1]="Modified"}(o||(o={}));for(var c=new e.Map,l=0,u=s;l<u.length;l++){var d=u[l];if(!(K=X.getSourceFileByPath?X.getSourceFileByPath(d.fileName,d.resolvedPath,R.target,void 0,xe):X.getSourceFile(d.fileName,R.target,void 0,xe)))return 0;e.Debug.assert(!K.redirectInfo,"Host should not return a redirect source file from `getSourceFile`");var p=void 0;if(d.redirectInfo){if(K!==d.redirectInfo.unredirected)return 0;p=!1,K=d}else if(P.redirectTargetsMap.has(d.path)){if(K!==d)return 0;p=!1}else p=K!==d;K.path=d.path,K.originalFileName=d.originalFileName,K.resolvedPath=d.resolvedPath,K.fileName=d.fileName;var f=P.sourceFileToPackageName.get(d.path);if(void 0!==f){var _=c.get(f),y=p?1:0;if(void 0!==_&&1===y||1===_)return 0;c.set(f,y)}if(p){if(!e.arrayIsEqualTo(d.libReferenceDirectives,K.libReferenceDirectives,En))return 0;d.hasNoDefaultLib!==K.hasNoDefaultLib&&(Se=1),e.arrayIsEqualTo(d.referencedFiles,K.referencedFiles,En)||(Se=1),xn(K),e.arrayIsEqualTo(d.imports,K.imports,Tn)||(Se=1),e.arrayIsEqualTo(d.moduleAugmentations,K.moduleAugmentations,Tn)||(Se=1),(3145728&d.flags)!=(3145728&K.flags)&&(Se=1),e.arrayIsEqualTo(d.typeReferenceDirectives,K.typeReferenceDirectives,En)||(Se=1),i.push({oldFile:d,newFile:K})}else ie(d.path)&&(Se=1,i.push({oldFile:d,newFile:K}));a.push(K)}if(2!==Se)return Se;for(var v=i.map((function(e){return e.oldFile})),h=0,b=s;h<b.length;h++){var E=b[h];if(!e.contains(v,E))for(var T=0,x=E.ambientModuleNames;T<x.length;T++){var L=x[T];M.set(L,E.fileName)}}for(var A=0,D=i;A<D.length;A++){var I=D[A],k=(d=I.oldFile,S(K=I.newFile)),w=Qe(k,K);e.hasChangesInResolutions(k,w,d.resolvedModules,e.moduleResolutionIsEqualTo)?(Se=1,K.resolvedModules=e.zipToMap(k,w)):K.resolvedModules=d.resolvedModules;var F=e.map(K.typeReferenceDirectives,(function(n){return e.toFileNameLowerCase(n.fileName)})),U=We(F,K);e.hasChangesInResolutions(F,U,d.resolvedTypeReferenceDirectiveNames,e.typeDirectiveIsEqualTo)?(Se=1,K.resolvedTypeReferenceDirectiveNames=e.zipToMap(F,U)):K.resolvedTypeReferenceDirectiveNames=d.resolvedTypeReferenceDirectiveNames}if(2!==Se)return Se;if(null===(n=X.hasChangedAutomaticTypeDirectiveNames)||void 0===n?void 0:n.call(X))return 1;ce=P.getMissingFilePaths(),C=P.getRefFileMap(),e.Debug.assert(a.length===P.getSourceFiles().length);for(var V=0,j=a;V<j.length;V++){var K=j[V];ge.set(K.path,K)}P.getFilesByNameMap().forEach((function(e,n){e?e.path!==n?ge.set(n,ge.get(e.path)):P.isSourceFileFromExternalLibrary(e)&&H.set(e.path,!0):ge.set(n,e)})),m=a,B=P.getFileProcessingDiagnostics();for(var W=0,z=i;W<z.length;W++){var q=z[W];B.reattachFileDiagnostics(q.newFile)}return G=P.getResolvedTypeReferenceDirectives(),fe=P.sourceFileToPackageName,_e=P.redirectTargetsMap,2}(),e.tracing.pop(),2!==Se){if(d=[],p=[],O&&(le||(le=O.map(qn)),N.length))for(var Le=0,Ae=le;Le<Ae.length;Le++){var De=Ae[Le];if(De){var Ce=e.outFile(De.commandLine.options);if(ve){if(Ce||e.getEmitModuleKind(De.commandLine.options)===e.ModuleKind.None)for(var Ie=0,Ne=De.commandLine.fileNames;Ie<Ne.length;Ie++){An(Oe=Ne[Ie],!1,!1,void 0)}}else if(Ce)An(e.changeExtension(Ce,".d.ts"),!1,!1,void 0);else if(e.getEmitModuleKind(De.commandLine.options)===e.ModuleKind.None)for(var Re=0,ke=De.commandLine.fileNames;Re<ke.length;Re++){var Oe=ke[Re];e.fileExtensionIs(Oe,".d.ts")||e.fileExtensionIs(Oe,".json")||An(e.getOutputDeclarationFileName(Oe,De.commandLine,!X.useCaseSensitiveFileNames()),!1,!1,void 0)}}}e.tracing.push("program","processRootFiles",{count:N.length}),e.forEach(N,(function(e){return bn(e,!1,!1)})),e.tracing.pop();var Pe=N.length?e.getAutomaticTypeDirectiveNames(R,X):e.emptyArray;if(Pe.length){e.tracing.push("program","processTypeReferences",{count:Pe.length});for(var Me=R.configFilePath?e.getDirectoryPath(R.configFilePath):X.getCurrentDirectory(),we=We(Pe,e.combinePaths(Me,e.inferredTypesContainingFile)),Fe=0;Fe<Pe.length;Fe++)Vn(Pe[Fe],we[Fe]);e.tracing.pop()}if(N.length&&!Q){var Ge=Z();!R.lib&&Ge?bn(Ge,!0,!1):e.forEach(R.lib,(function(n){bn(e.combinePaths($,n),!0,!1)}))}ce=e.arrayFrom(e.mapDefinedIterator(ge.entries(),(function(e){var n=e[0];return void 0===e[1]?n:void 0}))),m=e.stableSort(d,(function(n,t){return e.compareValues(Je(n),Je(t))})).concat(p),d=void 0,p=void 0}if(e.Debug.assert(!!ce),P&&X.onReleaseOldSourceFile){for(var Be=0,Ue=P.getSourceFiles();Be<Ue.length;Be++){var Ve=Ue[Be],je=sn(Ve.resolvedPath);(xe||!je||Ve.resolvedPath===Ve.path&&je.resolvedPath!==Ve.path)&&X.onReleaseOldSourceFile(Ve,P.getCompilerOptions(),!!sn(Ve.path))}P.forEachResolvedProjectReference((function(e){Gn(e.sourceFile.path)||X.onReleaseOldSourceFile(e.sourceFile,P.getCompilerOptions(),!1)}))}P=void 0;var He={getRootFileNames:function(){return N},getSourceFile:on,getSourceFileByPath:sn,getSourceFiles:function(){return m},getMissingFilePaths:function(){return ce},getRefFileMap:function(){return C},getFilesByNameMap:function(){return ge},getCompilerOptions:function(){return R},getSyntacticDiagnostics:function(e,n){return cn(e,un,n)},getOptionsDiagnostics:function(){return e.sortAndDeduplicateDiagnostics(e.concatenate(B.getGlobalDiagnostics(),e.concatenate(ee.getGlobalDiagnostics(),function(){if(!R.configFile)return e.emptyArray;var n=ee.getDiagnostics(R.configFile.fileName);return Mn((function(t){n=e.concatenate(n,ee.getDiagnostics(t.sourceFile.fileName))})),n}())))},getGlobalDiagnostics:function(){return N.length?e.sortAndDeduplicateDiagnostics(tn().getGlobalDiagnostics().slice()):e.emptyArray},getSemanticDiagnostics:function(e,n){return cn(e,pn,n)},getCachedSemanticDiagnostics:function(e){var n;return e?null===(n=w.perFile)||void 0===n?void 0:n.get(e.path):w.allDiagnostics},getSuggestionDiagnostics:function(e,n){return dn((function(){return tn().getSuggestionDiagnostics(e,n)}))},getDeclarationDiagnostics:function(n,t){var r=He.getCompilerOptions();return!n||e.outFile(r)?gn(n,t):cn(n,hn,t)},getBindAndCheckDiagnostics:function(e,n){return mn(e,n)},getProgramDiagnostics:ln,getTypeChecker:rn,getClassifiableNames:function(){var n;if(!D){rn(),D=new e.Set;for(var t=0,r=m;t<r.length;t++){var a=r[t];null===(n=a.classifiableNames)||void 0===n||n.forEach((function(e){return D.add(e)}))}}return D},getDiagnosticsProducingTypeChecker:tn,getCommonSourceDirectory:Ye,emit:function(n,t,r,a,i,o){var s=["emit","emit",{path:null==n?void 0:n.path}];e.tracing.begin.apply(e.tracing,s);var c=dn((function(){return function(n,t,r,a,i,o,s){if(!s){var c=y(n,t,r,a);if(c)return c}var l=tn().getEmitResolver(e.outFile(R)?void 0:t,a);e.performance.mark("beforeEmit");var u=e.emitFiles(l,Ze(r),t,e.getTransformers(R,o,i),i,!1,s);return e.performance.mark("afterEmit"),e.performance.measure("Emit","beforeEmit","afterEmit"),u}(He,n,t,r,a,i,o)}));return e.tracing.end.apply(e.tracing,s),c},getCurrentDirectory:function(){return ne},getNodeCount:function(){return tn().getNodeCount()},getIdentifierCount:function(){return tn().getIdentifierCount()},getSymbolCount:function(){return tn().getSymbolCount()},getTypeCatalog:function(){return tn().getTypeCatalog()},getTypeCount:function(){return tn().getTypeCount()},getInstantiationCount:function(){return tn().getInstantiationCount()},getRelationCacheSizes:function(){return tn().getRelationCacheSizes()},getFileProcessingDiagnostics:function(){return B},getResolvedTypeReferenceDirectives:function(){return G},isSourceFileFromExternalLibrary:nn,isSourceFileDefaultLibrary:function(n){if(n.hasNoDefaultLib)return!0;if(!R.noLib)return!1;var t=X.useCaseSensitiveFileNames()?e.equateStringsCaseSensitive:e.equateStringsCaseInsensitive;return R.lib?e.some(R.lib,(function(r){return t(n.fileName,e.combinePaths($,r))})):t(n.fileName,Z())},dropDiagnosticsProducingTypeChecker:function(){L=void 0},getSourceFileFromReference:function(e,t){return Ln(n(t.fileName,e.fileName),(function(e){return ge.get(Xe(e))||void 0}))},getLibFileFromReference:function(n){var t=e.toFileNameLowerCase(n.fileName),r=e.libMap.get(t);if(r)return on(e.combinePaths($,r))},sourceFileToPackageName:fe,redirectTargetsMap:_e,isEmittedFile:function(n){if(R.noEmit)return!1;var t=Xe(n);if(sn(t))return!1;var r=e.outFile(R);if(r)return ot(t,r)||ot(t,e.removeFileExtension(r)+".d.ts");if(R.declarationDir&&e.containsPath(R.declarationDir,t,ne,!X.useCaseSensitiveFileNames()))return!0;if(R.outDir)return e.containsPath(R.outDir,t,ne,!X.useCaseSensitiveFileNames());if(e.fileExtensionIsOneOf(t,e.supportedJSExtensions)||e.fileExtensionIs(t,".d.ts")){var a=e.removeFileExtension(t);return!!sn(a+".ts")||!!sn(a+".tsx")}return!1},getConfigFileParsingDiagnostics:function(){return k||e.emptyArray},getResolvedModuleWithFailedLookupLocationsFromCache:function(n,t){return z&&e.resolveModuleNameFromCache(n,t,z)},getProjectReferences:function(){return O},getResolvedProjectReferences:$e,getProjectReferenceRedirect:Rn,getResolvedProjectReferenceToRedirect:Pn,getResolvedProjectReferenceByPath:Gn,forEachResolvedProjectReference:Mn,isSourceOfProjectReferenceRedirect:Fn,emitBuildInfo:function(n){e.Debug.assert(!e.outFile(R));var t=["emit","emitBuildInfo"];e.tracing.begin.apply(e.tracing,t),e.performance.mark("beforeEmit");var r=e.emitFiles(e.notImplementedResolver,Ze(n),void 0,e.noTransformers,!1,!0);return e.performance.mark("afterEmit"),e.performance.measure("Emit","beforeEmit","afterEmit"),e.tracing.end.apply(e.tracing,t),r},fileExists:Ee,directoryExists:Te,getSymlinkCache:st,realpath:null===(u=X.realpath)||void 0===u?void 0:u.bind(X),useCaseSensitiveFileNames:function(){return X.useCaseSensitiveFileNames()},structureIsReused:Se};return be(),function(){R.strictPropertyInitialization&&!e.getStrictOptionValue(R,"strictNullChecks")&&$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"strictPropertyInitialization","strictNullChecks");R.isolatedModules&&(R.out&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"out","isolatedModules"),R.outFile&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"outFile","isolatedModules"));R.inlineSourceMap&&(R.sourceMap&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"sourceMap","inlineSourceMap"),R.mapRoot&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"mapRoot","inlineSourceMap"));R.composite&&(!1===R.declaration&&$n(e.Diagnostics.Composite_projects_may_not_disable_declaration_emit,"declaration"),!1===R.incremental&&$n(e.Diagnostics.Composite_projects_may_not_disable_incremental_compilation,"declaration"));var n=e.outFile(R);R.tsBuildInfoFile?e.isIncrementalCompilation(R)||$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"tsBuildInfoFile","incremental","composite"):!R.incremental||n||R.configFilePath||ee.add(e.createCompilerDiagnostic(e.Diagnostics.Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBuildInfoFile_is_specified));if(function(){var n=R.suppressOutputPathCheck?void 0:e.getTsBuildInfoEmitOutputFilePath(R);g(O,le,(function(t,r,a){var i=(r?r.commandLine.projectReferences:O)[a],o=r&&r.sourceFile;if(t){var s=t.commandLine.options;if(!s.composite||s.noEmit)(r?r.commandLine.fileNames:N).length&&(s.composite||nt(o,a,e.Diagnostics.Referenced_project_0_must_have_setting_composite_Colon_true,i.path),s.noEmit&&nt(o,a,e.Diagnostics.Referenced_project_0_may_not_disable_emit,i.path));if(i.prepend){var c=e.outFile(s);c?X.fileExists(c)||nt(o,a,e.Diagnostics.Output_file_0_from_project_1_does_not_exist,c,i.path):nt(o,a,e.Diagnostics.Cannot_prepend_project_0_because_it_does_not_have_outFile_set,i.path)}!r&&n&&n===e.getTsBuildInfoEmitOutputFilePath(s)&&(nt(o,a,e.Diagnostics.Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1,n,i.path),ae.set(Xe(n),!0))}else nt(o,a,e.Diagnostics.File_0_not_found,i.path)}))}(),R.composite)for(var t=new e.Set(N.map(Xe)),r=0,a=m;r<a.length;r++){var i=a[r];e.sourceFileMayBeEmitted(i,He)&&!t.has(i.path)&&Xn(i,t,e.Diagnostics.File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_include_pattern,i.fileName,R.configFilePath||"")}if(R.paths)for(var o in R.paths)if(e.hasProperty(R.paths,o))if(e.hasZeroOrOneAsteriskCharacter(o)||Qn(!0,o,e.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character,o),e.isArray(R.paths[o])){var s=R.paths[o].length;0===s&&Qn(!1,o,e.Diagnostics.Substitutions_for_pattern_0_shouldn_t_be_an_empty_array,o);for(var c=0;c<s;c++){var l=R.paths[o][c],u=typeof l;"string"===u?(e.hasZeroOrOneAsteriskCharacter(l)||Yn(o,c,e.Diagnostics.Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character,l,o),R.baseUrl||e.pathIsRelative(l)||e.pathIsAbsolute(l)||Yn(o,c,e.Diagnostics.Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash)):Yn(o,c,e.Diagnostics.Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2,l,o,u)}}else Qn(!1,o,e.Diagnostics.Substitutions_for_pattern_0_should_be_an_array,o);R.sourceMap||R.inlineSourceMap||(R.inlineSources&&$n(e.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided,"inlineSources"),R.sourceRoot&&$n(e.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided,"sourceRoot"));R.out&&R.outFile&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"out","outFile");!R.mapRoot||R.sourceMap||R.declarationMap||$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"mapRoot","sourceMap","declarationMap");R.declarationDir&&(e.getEmitDeclarations(R)||$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"declarationDir","declaration","composite"),n&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"declarationDir",R.out?"out":"outFile"));R.declarationMap&&!e.getEmitDeclarations(R)&&$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"declarationMap","declaration","composite");R.lib&&R.noLib&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"lib","noLib");R.noImplicitUseStrict&&e.getStrictOptionValue(R,"alwaysStrict")&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"noImplicitUseStrict","alwaysStrict");var d=R.target||0,p=e.find(m,(function(n){return e.isExternalModule(n)&&!n.isDeclarationFile}));if(R.isolatedModules){R.module===e.ModuleKind.None&&d<2&&$n(e.Diagnostics.Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher,"isolatedModules","target");var f=e.find(m,(function(n){return!e.isExternalModule(n)&&!e.isSourceFileJS(n)&&!n.isDeclarationFile&&6!==n.scriptKind}));if(f){var _=e.getErrorSpanForNode(f,f);ee.add(e.createFileDiagnostic(f,_.start,_.length,e.Diagnostics._0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_import_export_or_an_empty_export_statement_to_make_it_a_module,e.getBaseFileName(f.fileName)))}}else if(p&&d<2&&R.module===e.ModuleKind.None){_=e.getErrorSpanForNode(p,p.externalModuleIndicator);ee.add(e.createFileDiagnostic(p,_.start,_.length,e.Diagnostics.Cannot_use_imports_exports_or_module_augmentations_when_module_is_none))}if(n&&!R.emitDeclarationOnly)if(R.module&&R.module!==e.ModuleKind.AMD&&R.module!==e.ModuleKind.System)$n(e.Diagnostics.Only_amd_and_system_modules_are_supported_alongside_0,R.out?"out":"outFile","module");else if(void 0===R.module&&p){_=e.getErrorSpanForNode(p,p.externalModuleIndicator);ee.add(e.createFileDiagnostic(p,_.start,_.length,e.Diagnostics.Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system,R.out?"out":"outFile"))}R.resolveJsonModule&&(e.getEmitModuleResolutionKind(R)!==e.ModuleResolutionKind.NodeJs?$n(e.Diagnostics.Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy,"resolveJsonModule"):e.hasJsonModuleEmitEnabled(R)||$n(e.Diagnostics.Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext,"resolveJsonModule","module"));if(R.outDir||R.sourceRoot||R.mapRoot){var y=Ye();R.outDir&&""===y&&m.some((function(n){return e.getRootLength(n.fileName)>1}))&&$n(e.Diagnostics.Cannot_find_the_common_subdirectory_path_for_the_input_files,"outDir")}R.useDefineForClassFields&&0===d&&$n(e.Diagnostics.Option_0_cannot_be_specified_when_option_target_is_ES3,"useDefineForClassFields");R.checkJs&&!e.getAllowJSCompilerOption(R)&&ee.add(e.createCompilerDiagnostic(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"checkJs","allowJs"));R.emitDeclarationOnly&&(e.getEmitDeclarations(R)||$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2,"emitDeclarationOnly","declaration","composite"),R.noEmit&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"emitDeclarationOnly","noEmit"));R.emitDecoratorMetadata&&!R.experimentalDecorators&&$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"emitDecoratorMetadata","experimentalDecorators");R.jsxFactory?(R.reactNamespace&&$n(e.Diagnostics.Option_0_cannot_be_specified_with_option_1,"reactNamespace","jsxFactory"),4!==R.jsx&&5!==R.jsx||$n(e.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1,"jsxFactory",e.inverseJsxOptionMap.get(""+R.jsx)),e.parseIsolatedEntityName(R.jsxFactory,d)||et("jsxFactory",e.Diagnostics.Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name,R.jsxFactory)):R.reactNamespace&&!e.isIdentifierText(R.reactNamespace,d)&&et("reactNamespace",e.Diagnostics.Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier,R.reactNamespace);R.jsxFragmentFactory&&(R.jsxFactory||$n(e.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1,"jsxFragmentFactory","jsxFactory"),4!==R.jsx&&5!==R.jsx||$n(e.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1,"jsxFragmentFactory",e.inverseJsxOptionMap.get(""+R.jsx)),e.parseIsolatedEntityName(R.jsxFragmentFactory,d)||et("jsxFragmentFactory",e.Diagnostics.Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name,R.jsxFragmentFactory));R.reactNamespace&&(4!==R.jsx&&5!==R.jsx||$n(e.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1,"reactNamespace",e.inverseJsxOptionMap.get(""+R.jsx)));R.jsxImportSource&&2===R.jsx&&$n(e.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1,"jsxImportSource",e.inverseJsxOptionMap.get(""+R.jsx));if(!R.noEmit&&!R.suppressOutputPathCheck){var v=Ze(),h=new e.Set;e.forEachEmittedFile(v,(function(e){R.emitDeclarationOnly||b(e.jsFilePath,h),b(e.declarationFilePath,h)}))}function b(n,t){if(n){var r=Xe(n);if(ge.has(r)){var a=void 0;R.configFilePath||(a=e.chainDiagnosticMessages(void 0,e.Diagnostics.Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig)),a=e.chainDiagnosticMessages(a,e.Diagnostics.Cannot_write_file_0_because_it_would_overwrite_input_file,n),it(n,e.createCompilerDiagnosticFromMessageChain(a))}var i=X.useCaseSensitiveFileNames()?r:e.toFileNameLowerCase(r);t.has(i)?it(n,e.createCompilerDiagnostic(e.Diagnostics.Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files,n)):t.add(i)}}}(),e.performance.mark("afterProgram"),e.performance.measure("Program","beforeProgram","afterProgram"),e.tracing.end.apply(e.tracing,K),He;function Ke(n,t,r){if(!n.length)return e.emptyArray;var a=e.getNormalizedAbsolutePath(t.originalFileName,ne),i=ze(t);e.tracing.push("program","resolveModuleNamesWorker",{containingFileName:a}),e.performance.mark("beforeResolveModule");var o=q(n,a,r,i);return e.performance.mark("afterResolveModule"),e.performance.measure("ResolveModule","beforeResolveModule","afterResolveModule"),e.tracing.pop(),o}function We(n,t){if(!n.length)return[];var r=e.isString(t)?t:e.getNormalizedAbsolutePath(t.originalFileName,ne),a=e.isString(t)?void 0:ze(t);e.tracing.push("program","resolveTypeReferenceDirectiveNamesWorker",{containingFileName:r}),e.performance.mark("beforeResolveTypeReference");var i=J(n,r,a);return e.performance.mark("afterResolveTypeReference"),e.performance.measure("ResolveTypeReference","beforeResolveTypeReference","afterResolveTypeReference"),e.tracing.pop(),i}function ze(n){var t=Pn(n.originalFileName);if(t||!e.fileExtensionIs(n.originalFileName,".d.ts"))return t;var r=qe(n.originalFileName,n.path);if(r)return r;if(X.realpath&&R.preserveSymlinks&&e.stringContains(n.originalFileName,e.nodeModulesPathPart)){var a=X.realpath(n.originalFileName),i=Xe(a);return i===n.path?void 0:qe(a,i)}}function qe(n,t){var r=wn(n);return e.isString(r)?Pn(r):r?Mn((function(n){var r=e.outFile(n.commandLine.options);if(r)return Xe(r)===t?n:void 0})):void 0}function Je(n){if(e.containsPath($,n.fileName,!1)){var t=e.getBaseFileName(n.fileName);if("lib.d.ts"===t||"lib.es6.d.ts"===t)return 0;var r=e.removeSuffix(e.removePrefix(t,"lib."),".d.ts"),a=e.libs.indexOf(r);if(-1!==a)return a+1}return e.libs.length+2}function Xe(n){return e.toPath(n,ne,Kn)}function Ye(){if(void 0===x){var n=e.filter(m,(function(n){return e.sourceFileMayBeEmitted(n,He)}));R.rootDir&&zn(n,R.rootDir)?x=e.getNormalizedAbsolutePath(R.rootDir,ne):R.composite&&R.configFilePath?zn(n,x=e.getDirectoryPath(e.normalizeSlashes(R.configFilePath))):(r=n,x=t(e.mapDefined(r,(function(e){return e.isDeclarationFile?void 0:e.fileName})),ne,Kn)),x&&x[x.length-1]!==e.directorySeparator&&(x+=e.directorySeparator)}var r;return x}function Qe(n,t){if(0===Se&&!t.ambientModuleNames.length)return Ke(n,t,void 0);var r,a,i,o=P&&P.getSourceFile(t.fileName);if(o!==t&&t.resolvedModules){for(var s=[],c=0,l=n;c<l.length;c++){var u=l[c],d=t.resolvedModules.get(u);s.push(d)}return s}for(var p={},m=0;m<n.length;m++){u=n[m];if(t===o&&!ie(o.path)){var f=e.getResolvedModule(o,u);if(f){e.isTraceEnabled(R,X)&&e.trace(X,e.Diagnostics.Reusing_resolution_of_module_0_to_file_1_from_old_program,u,e.getNormalizedAbsolutePath(t.originalFileName,ne)),(a||(a=new Array(n.length)))[m]=f,(i||(i=[])).push(u);continue}}var _=!1;e.contains(t.ambientModuleNames,u)?(_=!0,e.isTraceEnabled(R,X)&&e.trace(X,e.Diagnostics.Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1,u,e.getNormalizedAbsolutePath(t.originalFileName,ne))):_=v(u),_?(a||(a=new Array(n.length)))[m]=p:(r||(r=[])).push(u)}var g=r&&r.length?Ke(r,t,i):e.emptyArray;if(!a)return e.Debug.assert(g.length===n.length),g;var y=0;for(m=0;m<a.length;m++)a[m]?a[m]===p&&(a[m]=void 0):(a[m]=g[y],y++);return e.Debug.assert(y===g.length),a;function v(n){var t=e.getResolvedModule(o,n),r=t&&P.getSourceFile(t.resolvedFileName);if(t&&r)return!1;var a=M.get(n);return!!a&&(e.isTraceEnabled(R,X)&&e.trace(X,e.Diagnostics.Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified,n,a),!0)}}function Ze(n){return{getPrependNodes:en,getCanonicalFileName:Kn,getCommonSourceDirectory:He.getCommonSourceDirectory,getCompilerOptions:He.getCompilerOptions,getCurrentDirectory:function(){return ne},getNewLine:function(){return X.getNewLine()},getSourceFile:He.getSourceFile,getSourceFileByPath:He.getSourceFileByPath,getSourceFiles:He.getSourceFiles,getLibFileFromReference:He.getLibFileFromReference,isSourceFileFromExternalLibrary:nn,getResolvedProjectReferenceToRedirect:Pn,getProjectReferenceRedirect:Rn,isSourceOfProjectReferenceRedirect:Fn,getSymlinkCache:st,writeFile:n||function(e,n,t,r,a){return X.writeFile(e,n,t,r,a)},isEmitBlocked:an,readFile:function(e){return X.readFile(e)},fileExists:function(n){var t=Xe(n);return!!sn(t)||!e.contains(ce,t)&&X.fileExists(n)},useCaseSensitiveFileNames:function(){return X.useCaseSensitiveFileNames()},getProgramBuildInfo:function(){return He.getProgramBuildInfo&&He.getProgramBuildInfo()},getSourceFileFromReference:function(e,n){return He.getSourceFileFromReference(e,n)},redirectTargetsMap:_e}}function $e(){return le}function en(){return b(O,(function(e,n){var t;return null===(t=le[n])||void 0===t?void 0:t.commandLine}),(function(e){var n=Xe(e),t=sn(n);return t?t.text:ge.has(n)?void 0:X.readFile(n)}))}function nn(e){return!!H.get(e.path)}function tn(){return L||(L=e.createTypeChecker(He,!0))}function rn(){return A||(A=e.createTypeChecker(He,!1))}function an(e){return ae.has(Xe(e))}function on(e){return sn(Xe(e))}function sn(e){return ge.get(e)||void 0}function cn(n,t,r){return n?t(n,r):e.sortAndDeduplicateDiagnostics(e.flatMap(He.getSourceFiles(),(function(e){return r&&r.throwIfCancellationRequested(),t(e,r)})))}function ln(n){return e.skipTypeChecking(n,R,He)?e.emptyArray:function(n){for(var t,r=[],a=1;a<arguments.length;a++)r[a-1]=arguments[a];var i=e.flatten(r);if(!(null===(t=n.commentDirectives)||void 0===t?void 0:t.length))return i;return _n(n,n.commentDirectives,i).diagnostics}(n,B.getDiagnostics(n.fileName),ee.getDiagnostics(n.fileName))}function un(n){return e.isSourceFileJS(n)?(n.additionalSyntacticDiagnostics||(n.additionalSyntacticDiagnostics=function(n){return dn((function(){var t=[];return r(n,n),e.forEachChildRecursively(n,r,a),t;function r(n,r){switch(r.kind){case 160:case 163:case 165:if(r.questionToken===n)return t.push(s(n,e.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files,"?")),"skip";case 164:case 166:case 167:case 168:case 208:case 251:case 209:case 249:if(r.type===n)return t.push(s(n,e.Diagnostics.Type_annotations_can_only_be_used_in_TypeScript_files)),"skip"}switch(n.kind){case 262:if(n.isTypeOnly)return t.push(s(r,e.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files,"import type")),"skip";break;case 267:if(n.isTypeOnly)return t.push(s(n,e.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files,"export type")),"skip";break;case 260:return t.push(s(n,e.Diagnostics.import_can_only_be_used_in_TypeScript_files)),"skip";case 266:if(n.isExportEquals)return t.push(s(n,e.Diagnostics.export_can_only_be_used_in_TypeScript_files)),"skip";break;case 286:if(116===n.token)return t.push(s(n,e.Diagnostics.implements_clauses_can_only_be_used_in_TypeScript_files)),"skip";break;case 253:var a=e.tokenToString(117);return e.Debug.assertIsDefined(a),t.push(s(n,e.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files,a)),"skip";case 256:var i=16&n.flags?e.tokenToString(140):e.tokenToString(139);return e.Debug.assertIsDefined(i),t.push(s(n,e.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files,i)),"skip";case 254:return t.push(s(n,e.Diagnostics.Type_aliases_can_only_be_used_in_TypeScript_files)),"skip";case 255:var o=e.Debug.checkDefined(e.tokenToString(91));return t.push(s(n,e.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files,o)),"skip";case 225:return t.push(s(n,e.Diagnostics.Non_null_assertions_can_only_be_used_in_TypeScript_files)),"skip";case 224:return t.push(s(n.type,e.Diagnostics.Type_assertion_expressions_can_only_be_used_in_TypeScript_files)),"skip";case 206:e.Debug.fail()}}function a(n,r){switch(r.decorators!==n||R.experimentalDecorators||t.push(s(r,e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning)),r.kind){case 252:case 221:case 165:case 166:case 167:case 168:case 208:case 251:case 209:if(n===r.typeParameters)return t.push(o(n,e.Diagnostics.Type_parameter_declarations_can_only_be_used_in_TypeScript_files)),"skip";case 232:if(n===r.modifiers)return i(r.modifiers,232===r.kind),"skip";break;case 163:if(n===r.modifiers){for(var a=0,c=n;a<c.length;a++){var l=c[a];123!==l.kind&&t.push(s(l,e.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files,e.tokenToString(l.kind)))}return"skip"}break;case 160:if(n===r.modifiers)return t.push(o(n,e.Diagnostics.Parameter_modifiers_can_only_be_used_in_TypeScript_files)),"skip";break;case 203:case 204:case 223:case 274:case 275:case 205:if(n===r.typeArguments)return t.push(o(n,e.Diagnostics.Type_arguments_can_only_be_used_in_TypeScript_files)),"skip"}}function i(n,r){for(var a=0,i=n;a<i.length;a++){var o=i[a];switch(o.kind){case 84:if(r)continue;case 122:case 120:case 121:case 142:case 133:case 125:t.push(s(o,e.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files,e.tokenToString(o.kind)))}}}function o(t,r,a,i,o){var s=t.pos;return e.createFileDiagnostic(n,s,t.end-s,r,a,i,o)}function s(t,r,a,i,o){return e.createDiagnosticForNodeInSourceFile(n,t,r,a,i,o)}}))}(n)),e.concatenate(n.additionalSyntacticDiagnostics,n.parseDiagnostics)):n.parseDiagnostics}function dn(n){try{return n()}catch(n){throw n instanceof e.OperationCanceledException&&(A=void 0,L=void 0),n}}function pn(n,t){return e.concatenate(v(mn(n,t),R),ln(n))}function mn(e,n){return vn(e,n,w,fn)}function fn(n,t){return dn((function(){if(e.skipTypeChecking(n,R,He))return e.emptyArray;var r=tn();e.Debug.assert(!!n.bindDiagnostics);var a=e.isCheckJsEnabledForFile(n,R),i=!(!!n.checkJsDirective&&!1===n.checkJsDirective.enabled)&&(3===n.scriptKind||4===n.scriptKind||5===n.scriptKind||a||7===n.scriptKind),o=i?n.bindDiagnostics:e.emptyArray,s=i?r.getDiagnostics(n,t):e.emptyArray;return function(n,t){for(var r,a=[],i=2;i<arguments.length;i++)a[i-2]=arguments[i];var o=e.flatten(a);if(!t||!(null===(r=n.commentDirectives)||void 0===r?void 0:r.length))return o;for(var s=_n(n,n.commentDirectives,o),c=s.diagnostics,l=s.directives,u=0,d=l.getUnusedExpectations();u<d.length;u++){var p=d[u];c.push(e.createDiagnosticForRange(n,p.range,e.Diagnostics.Unused_ts_expect_error_directive))}return c}(n,i,o,s,a?n.jsDocDiagnostics:void 0)}))}function _n(n,t,r){var a=e.createCommentDirectivesMap(n,t);return{diagnostics:r.filter((function(n){return-1===function(n,t){var r=n.file,a=n.start;if(!r)return-1;var i=e.getLineStarts(r),o=e.computeLineAndCharacterOfPosition(i,a).line-1;for(;o>=0;){if(t.markUsed(o))return o;var s=r.text.slice(i[o],i[o+1]).trim();if(""!==s&&!/^(\s*)\/\/(.*)$/.test(s))return-1;o--}return-1}(n,a)})),directives:a}}function gn(e,n){return vn(e,n,F,yn)}function yn(n,t){return dn((function(){var r=tn().getEmitResolver(n,t);return e.getDeclarationDiagnostics(Ze(e.noop),r,n)||e.emptyArray}))}function vn(n,t,r,a){var i,o=n?null===(i=r.perFile)||void 0===i?void 0:i.get(n.path):r.allDiagnostics;if(o)return o;var s=a(n,t);return n?(r.perFile||(r.perFile=new e.Map)).set(n.path,s):r.allDiagnostics=s,s}function hn(e,n){return e.isDeclarationFile?[]:gn(e,n)}function bn(n,t,r){An(e.normalizePath(n),t,r,void 0)}function En(e,n){return e.fileName===n.fileName}function Tn(e,n){return 78===e.kind?78===n.kind&&e.escapedText===n.escapedText:10===n.kind&&e.text===n.text}function Sn(n,t){var r=e.factory.createStringLiteral(n),a=e.factory.createImportDeclaration(void 0,void 0,void 0,r);return e.addEmitFlags(a,67108864),e.setParent(r,a),e.setParent(a,t),r.flags&=-9,a.flags&=-9,r}function xn(n){if(!n.imports){var t,r,a,i=e.isSourceFileJS(n),o=e.isExternalModule(n);if((R.isolatedModules||o)&&!n.isDeclarationFile){R.importHelpers&&(t=[Sn(e.externalHelpersModuleNameText,n)]);var s=e.getJSXRuntimeImport(e.getJSXImplicitImportBase(R,n),R);s&&(t||(t=[])).push(Sn(s,n))}for(var c=0,l=n.statements;c<l.length;c++){u(l[c],!1)}return(1048576&n.flags||i)&&function(n){var r=/import|require/g;for(;null!==r.exec(n.text);){var a=d(n,r.lastIndex);i&&e.isRequireCall(a,!0)||e.isImportCall(a)&&1===a.arguments.length&&e.isStringLiteralLike(a.arguments[0])?t=e.append(t,a.arguments[0]):e.isLiteralImportTypeNode(a)&&(t=e.append(t,a.argument.literal))}}(n),n.imports=t||e.emptyArray,n.moduleAugmentations=r||e.emptyArray,void(n.ambientModuleNames=a||e.emptyArray)}function u(i,s){if(e.isAnyImportOrReExport(i)){var c=e.getExternalModuleName(i);!(c&&e.isStringLiteral(c)&&c.text)||s&&e.isExternalModuleNameRelative(c.text)||(t=e.append(t,c))}else if(e.isModuleDeclaration(i)&&e.isAmbientModule(i)&&(s||e.hasSyntacticModifier(i,2)||n.isDeclarationFile)){var l=e.getTextOfIdentifierOrLiteral(i.name);if(o||s&&!e.isExternalModuleNameRelative(l))(r||(r=[])).push(i.name);else if(!s){n.isDeclarationFile&&(a||(a=[])).push(l);var d=i.body;if(d)for(var p=0,m=d.statements;p<m.length;p++){u(m[p],!0)}}}}function d(n,t){for(var r=n,a=function(e){if(e.pos<=t&&(t<e.end||t===e.end&&1===e.kind))return e};;){var o=i&&e.hasJSDocNodes(r)&&e.forEach(r.jsDoc,a)||e.forEachChild(r,a);if(!o)return r;r=o}}}function Ln(n,t,r,a){if(e.hasExtension(n)){var i=X.getCanonicalFileName(n);if(!R.allowNonTsExtensions&&!e.forEach(re,(function(n){return e.fileExtensionIs(i,n)})))return void(r&&(e.hasJSFileExtension(i)?r(e.Diagnostics.File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option,n):r(e.Diagnostics.File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1,n,"'"+te.join("', '")+"'")));var o=t(n);if(r)if(o)a&&i===X.getCanonicalFileName(a.fileName)&&r(e.Diagnostics.A_file_cannot_have_a_reference_to_itself);else{var s=Rn(n);s?r(e.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1,s,n):r(e.Diagnostics.File_0_not_found,n)}return o}var c=R.allowNonTsExtensions&&t(n);if(c)return c;if(!r||!R.allowNonTsExtensions){var l=e.forEach(te,(function(e){return t(n+e)}));return r&&!l&&r(e.Diagnostics.Could_not_resolve_the_path_0_with_the_extensions_Colon_1,n,"'"+te.join("', '")+"'"),l}r(e.Diagnostics.File_0_not_found,n)}function An(e,n,t,r,a){Ln(e,(function(e){return Cn(e,Xe(e),n,t,a,r)}),(function(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return B.add(Hn.apply(void 0,__spreadArrays([a,e],n)))}),a&&a.file)}function Dn(n,t,r){var a=r?void 0:C&&C.get(t.path),i=a&&e.find(a,(function(e){return e.referencedFileName===t.fileName}));B.add(i?Jn(i,e.Diagnostics.Already_included_file_name_0_differs_from_file_name_1_only_in_casing,t.fileName,n):Hn(r,e.Diagnostics.File_name_0_differs_from_already_included_file_name_1_only_in_casing,n,t.fileName))}function Cn(n,t,r,a,i,o){e.tracing.push("program","findSourceFile",{fileName:n,isDefaultLib:r||void 0,refKind:i?e.RefFileKind[i.kind]:void 0});var s=function(n,t,r,a,i,o){if(ve){var s=wn(n);if(!s&&X.realpath&&R.preserveSymlinks&&e.isDeclarationFileName(n)&&e.stringContains(n,e.nodeModulesPathPart)){var c=X.realpath(n);c!==n&&(s=wn(c))}if(s){var l=e.isString(s)?Cn(s,Xe(s),r,a,i,o):void 0;return l&&Nn(l,t,void 0),l}}var u,m=n;if(ge.has(t)){var f=ge.get(t);if(In(n,f||void 0,i),f&&R.forceConsistentCasingInFileNames){var _=f.fileName;Xe(_)!==Xe(n)&&(n=Rn(n)||n),e.getNormalizedAbsolutePathWithoutRoot(_,ne)!==e.getNormalizedAbsolutePathWithoutRoot(n,ne)&&Dn(n,f,i)}return f&&H.get(f.path)&&0===V?(H.set(f.path,!1),R.noResolve||(Bn(f,r),Un(f)),R.noLib||jn(f),j.set(f.path,!1),Wn(f)):f&&j.get(f.path)&&V<U&&(j.set(f.path,!1),Wn(f)),f||void 0}if(i&&!ve){var g=kn(n);if(g){if(e.outFile(g.commandLine.options))return;var y=On(g,n);n=y,u=Xe(y)}}var v=X.getSourceFile(n,R.target,(function(t){return B.add(Hn(i,e.Diagnostics.Cannot_read_file_0_Colon_1,n,t))}),xe);if(o){var h=e.packageIdToString(o),b=me.get(h);if(b){var E=function(e,n,t,r,a,i){var o=Object.create(e);return o.fileName=t,o.path=r,o.resolvedPath=a,o.originalFileName=i,o.redirectInfo={redirectTarget:e,unredirected:n},H.set(r,V>0),Object.defineProperties(o,{id:{get:function(){return this.redirectInfo.redirectTarget.id},set:function(e){this.redirectInfo.redirectTarget.id=e}},symbol:{get:function(){return this.redirectInfo.redirectTarget.symbol},set:function(e){this.redirectInfo.redirectTarget.symbol=e}}}),o}(b,v,n,t,Xe(n),m);return _e.add(b.path,n),Nn(E,t,u),fe.set(t,o.name),p.push(E),E}v&&(me.set(h,v),fe.set(t,o.name))}if(Nn(v,t,u),v){if(H.set(t,V>0),v.fileName=n,v.path=t,v.resolvedPath=Xe(n),v.originalFileName=m,In(n,v,i),X.useCaseSensitiveFileNames()){var T=e.toFileNameLowerCase(t),S=ye.get(T);S?Dn(n,S,i):ye.set(T,v)}Q=Q||v.hasNoDefaultLib&&!a,R.noResolve||(Bn(v,r),Un(v)),R.noLib||jn(v),Wn(v),r?d.push(v):p.push(v)}return v}(n,t,r,a,i,o);return e.tracing.pop(),s}function In(n,t,r){r&&t&&(C||(C=e.createMultiMap())).add(t.path,{referencedFileName:n,kind:r.kind,index:r.index,file:r.file.path})}function Nn(e,n,t){t?(ge.set(t,e),ge.set(n,e||!1)):ge.set(n,e)}function Rn(e){var n=kn(e);return n&&On(n,e)}function kn(n){if(le&&le.length&&!e.fileExtensionIs(n,".d.ts")&&!e.fileExtensionIs(n,".json"))return Pn(n)}function On(n,t){var r=e.outFile(n.commandLine.options);return r?e.changeExtension(r,".d.ts"):e.getOutputDeclarationFileName(t,n.commandLine,!X.useCaseSensitiveFileNames())}function Pn(n){void 0===de&&(de=new e.Map,Mn((function(e){Xe(R.configFilePath)!==e.sourceFile.path&&e.commandLine.fileNames.forEach((function(n){return de.set(Xe(n),e.sourceFile.path)}))})));var t=de.get(Xe(n));return t&&Gn(t)}function Mn(n){return e.forEachResolvedProjectReference(le,n)}function wn(n){if(e.isDeclarationFileName(n))return void 0===pe&&(pe=new e.Map,Mn((function(n){var t=e.outFile(n.commandLine.options);if(t){var r=e.changeExtension(t,".d.ts");pe.set(Xe(r),!0)}else e.forEach(n.commandLine.fileNames,(function(t){if(!e.fileExtensionIs(t,".d.ts")&&!e.fileExtensionIs(t,".json")){var r=e.getOutputDeclarationFileName(t,n.commandLine,X.useCaseSensitiveFileNames());pe.set(Xe(r),t)}}))}))),pe.get(Xe(n))}function Fn(e){return ve&&!!Pn(e)}function Gn(e){if(ue)return ue.get(e)||void 0}function Bn(t,r){e.forEach(t.referencedFiles,(function(a,i){An(n(a.fileName,t.fileName),r,!1,void 0,{kind:e.RefFileKind.ReferenceFile,index:i,file:t,pos:a.pos,end:a.end})}))}function Un(n){var t=e.map(n.typeReferenceDirectives,(function(n){return e.toFileNameLowerCase(n.fileName)}));if(t)for(var r=We(t,n),a=0;a<t.length;a++){var i=n.typeReferenceDirectives[a],o=r[a],s=e.toFileNameLowerCase(i.fileName);e.setResolvedTypeReferenceDirective(n,s,o),Vn(s,o,{kind:e.RefFileKind.TypeReferenceDirective,index:a,file:n,pos:i.pos,end:i.end})}}function Vn(n,t,r){e.tracing.push("program","processTypeReferenceDirective",{directive:n,hasResolved:!!Qe,refKind:null==r?void 0:r.kind,refPath:null==r?void 0:r.file.path}),function(n,t,r){var a=G.get(n);if(a&&a.primary)return;var i=!0;if(t){if(t.isExternalLibraryImport&&V++,t.primary)An(t.resolvedFileName,!1,!1,t.packageId,r);else if(a){if(t.resolvedFileName!==a.resolvedFileName){var o=X.readFile(t.resolvedFileName),s=on(a.resolvedFileName);if(o!==s.text){var c=r?void 0:C&&C.get(s.path),l=c&&e.find(c,(function(e){return e.referencedFileName===s.fileName}));B.add(l?Jn(l,e.Diagnostics.Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict,n,t.resolvedFileName,a.resolvedFileName):Hn(r,e.Diagnostics.Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict,n,t.resolvedFileName,a.resolvedFileName))}}i=!1}else An(t.resolvedFileName,!1,!1,t.packageId,r);t.isExternalLibraryImport&&V--}else B.add(Hn(r,e.Diagnostics.Cannot_find_type_definition_file_for_0,n));i&&G.set(n,t)}(n,t,r),e.tracing.pop()}function jn(n){e.forEach(n.libReferenceDirectives,(function(t){var r=e.toFileNameLowerCase(t.fileName),a=e.libMap.get(r);if(a)bn(e.combinePaths($,a),!0,!0);else{var i=e.removeSuffix(e.removePrefix(r,"lib."),".d.ts"),o=e.getSpellingSuggestion(i,e.libs,e.identity),s=o?e.Diagnostics.Cannot_find_lib_definition_for_0_Did_you_mean_1:e.Diagnostics.Cannot_find_lib_definition_for_0;B.add(e.createFileDiagnostic(n,t.pos,t.end-t.pos,s,r,o))}}))}function Hn(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];return n?e.createFileDiagnostic.apply(void 0,__spreadArrays([n.file,n.pos,n.end-n.pos,t],r)):e.createCompilerDiagnostic.apply(void 0,__spreadArrays([t],r))}function Kn(e){return X.getCanonicalFileName(e)}function Wn(n){if(xn(n),n.imports.length||n.moduleAugmentations.length){var t=S(n),r=Qe(t,n);e.Debug.assert(r.length===t.length);for(var a=0;a<t.length;a++){var i=r[a];if(e.setResolvedModule(n,t[a],i),i){var o=i.isExternalLibraryImport,s=!e.resolutionExtensionIsTSOrJson(i.extension),c=o&&s,l=i.resolvedFileName;o&&V++;var u=c&&V>U,d=l&&!T(R,i)&&!R.noResolve&&a<n.imports.length&&!u&&!(s&&!e.getAllowJSCompilerOption(R))&&(e.isInJSFile(n.imports[a])||!(4194304&n.imports[a].flags));if(u)j.set(n.path,!0);else if(d){var p=Xe(l),m=e.skipTrivia(n.text,n.imports[a].pos);Cn(l,p,!1,!1,{kind:e.RefFileKind.Import,index:a,file:n,pos:m,end:n.imports[a].end},i.packageId)}o&&V--}}}else n.resolvedModules=void 0}function zn(n,t){for(var r,a=!0,i=X.getCanonicalFileName(e.getNormalizedAbsolutePath(t,ne)),o=0,s=n;o<s.length;o++){var c=s[o];if(!c.isDeclarationFile)0!==X.getCanonicalFileName(e.getNormalizedAbsolutePath(c.fileName,ne)).indexOf(i)&&(r||(r=new e.Set(N.map(Xe))),Xn(c,r,e.Diagnostics.File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files,c.fileName,t),a=!1)}return a}function qn(n){ue||(ue=new e.Map);var t,r,a=E(n),i=Xe(a),o=ue.get(i);if(void 0!==o)return o||void 0;if(X.getParsedCommandLine){if(!(t=X.getParsedCommandLine(a)))return Nn(void 0,i,void 0),void ue.set(i,!1);r=e.Debug.checkDefined(t.options.configFile),e.Debug.assert(!r.path||r.path===i),Nn(r,i,void 0)}else{var s=e.getNormalizedAbsolutePath(e.getDirectoryPath(a),X.getCurrentDirectory());if(Nn(r=X.getSourceFile(a,100),i,void 0),void 0===r)return void ue.set(i,!1);t=e.parseJsonSourceFileConfigFileContent(r,Y,s,void 0,a)}r.fileName=a,r.path=i,r.resolvedPath=i,r.originalFileName=a;var c={commandLine:t,sourceFile:r};return ue.set(i,c),t.projectReferences&&(c.references=t.projectReferences.map(qn)),c}function Jn(n,t){for(var r,a,i=[],o=2;o<arguments.length;o++)i[o-2]=arguments[o];var s,c,l=e.Debug.checkDefined(sn(n.file)),u=n.kind,d=n.index;switch(u){case e.RefFileKind.Import:s=e.skipTrivia(l.text,l.imports[d].pos),c=l.imports[d].end;break;case e.RefFileKind.ReferenceFile:s=(r=l.referencedFiles[d]).pos,c=r.end;break;case e.RefFileKind.TypeReferenceDirective:s=(a=l.typeReferenceDirectives[d]).pos,c=a.end;break;default:return e.Debug.assertNever(u)}return e.createFileDiagnostic.apply(void 0,__spreadArrays([l,s,c-s,t],i))}function Xn(n,t,r){for(var a=[],i=3;i<arguments.length;i++)a[i-3]=arguments[i];var o=null==C?void 0:C.get(n.path),s=e.forEach(o,(function(e){return t.has(e.file)?e:void 0}))||e.elementAt(o,0);ee.add(s?Jn.apply(void 0,__spreadArrays([s,r],a)):e.createCompilerDiagnostic.apply(void 0,__spreadArrays([r],a)))}function Yn(n,t,r,a,i,o){for(var s=!0,c=0,l=Zn();c<l.length;c++){var u=l[c];if(e.isObjectLiteralExpression(u.initializer))for(var d=0,p=e.getPropertyAssignment(u.initializer,n);d<p.length;d++){var m=p[d].initializer;e.isArrayLiteralExpression(m)&&m.elements.length>t&&(ee.add(e.createDiagnosticForNodeInSourceFile(R.configFile,m.elements[t],r,a,i,o)),s=!1)}}s&&ee.add(e.createCompilerDiagnostic(r,a,i,o))}function Qn(n,t,r,a){for(var i=!0,o=0,s=Zn();o<s.length;o++){var c=s[o];e.isObjectLiteralExpression(c.initializer)&&at(c.initializer,n,t,void 0,r,a)&&(i=!1)}i&&ee.add(e.createCompilerDiagnostic(r,a))}function Zn(){return function(n){var t=rt();if(t)return e.getPropertyAssignment(t,n)}("paths")||e.emptyArray}function $n(e,n,t,r){tt(!0,n,t,e,n,t,r)}function et(e,n,t){tt(!1,e,void 0,n,t)}function nt(n,t,r,a,i){var o=e.firstDefined(e.getTsConfigPropArray(n||R.configFile,"references"),(function(n){return e.isArrayLiteralExpression(n.initializer)?n.initializer:void 0}));o&&o.elements.length>t?ee.add(e.createDiagnosticForNodeInSourceFile(n||R.configFile,o.elements[t],r,a,i)):ee.add(e.createCompilerDiagnostic(r,a,i))}function tt(n,t,r,a,i,o,s){var c=rt();(!c||!at(c,n,t,r,a,i,o,s))&&ee.add(e.createCompilerDiagnostic(a,i,o,s))}function rt(){if(void 0===W){W=null;var n=e.getTsConfigObjectLiteralExpression(R.configFile);if(n)for(var t=0,r=e.getPropertyAssignment(n,"compilerOptions");t<r.length;t++){var a=r[t];if(e.isObjectLiteralExpression(a.initializer)){W=a.initializer;break}}}return W}function at(n,t,r,a,i,o,s,c){for(var l=e.getPropertyAssignment(n,r,a),u=0,d=l;u<d.length;u++){var p=d[u];ee.add(e.createDiagnosticForNodeInSourceFile(R.configFile,t?p.name:p.initializer,i,o,s,c))}return!!l.length}function it(e,n){ae.set(Xe(e),!0),ee.add(n)}function ot(n,t){return 0===e.comparePaths(n,t,ne,!X.useCaseSensitiveFileNames())}function st(){return X.getSymlinkCache?X.getSymlinkCache():f||(f=e.discoverProbableSymlinks(m,Kn,X.getCurrentDirectory()))}},e.emitSkippedWithNoDiagnostics={diagnostics:e.emptyArray,sourceMaps:void 0,emittedFiles:void 0,emitSkipped:!0},e.handleNoEmitOptions=y,e.filterSemanticDiagnotics=v,e.parseConfigHostFromCompilerHostLike=h,e.createPrependNodes=b,e.resolveProjectReferencePath=E,e.getResolutionDiagnostic=T}(ts||(ts={})),function(e){function n(e,n,t,r,a,i){var o=[],s=e.emit(n,(function(e,n,t){o.push({name:e,writeByteOrderMark:t,text:n})}),r,t,a,i),c=s.emitSkipped,l=s.diagnostics,u=s.exportedModulesFromDeclarationEmit;return{outputFiles:o,emitSkipped:c,diagnostics:l,exportedModulesFromDeclarationEmit:u}}e.getFileEmitOutput=n,function(t){function r(n){if(n.declarations&&n.declarations[0]){var t=e.getSourceFileOfNode(n.declarations[0]);return t&&t.resolvedPath}}function a(e,n){var t=e.getSymbolAtLocation(n);return t&&r(t)}function i(n,t,r,a){return e.toPath(n.getProjectReferenceRedirect(t)||t,r,a)}function o(n,t,r){var o;if(t.imports&&t.imports.length>0)for(var s=n.getTypeChecker(),c=0,l=t.imports;c<l.length;c++){var u=a(s,l[c]);u&&S(u)}var d=e.getDirectoryPath(t.resolvedPath);if(t.referencedFiles&&t.referencedFiles.length>0)for(var p=0,m=t.referencedFiles;p<m.length;p++){var f=m[p];S(i(n,f.fileName,d,r))}if(t.resolvedTypeReferenceDirectiveNames&&t.resolvedTypeReferenceDirectiveNames.forEach((function(e){if(e){var t=e.resolvedFileName;S(i(n,t,d,r))}})),t.moduleAugmentations.length){s=n.getTypeChecker();for(var _=0,g=t.moduleAugmentations;_<g.length;_++){var y=g[_];if(e.isStringLiteral(y)){var v=s.getSymbolAtLocation(y);v&&T(v)}}}for(var h=0,b=n.getTypeChecker().getAmbientModules();h<b.length;h++){var E=b[h];E.declarations.length>1&&T(E)}return o;function T(n){for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r],o=e.getSourceFileOfNode(i);o&&o!==t&&S(o.resolvedPath)}}function S(n){(o||(o=new e.Set)).add(n)}}function s(e,n){return n&&!n.referencedMap==!e}function c(e,n){n.forEach((function(n,t){return l(e,n,t)}))}function l(e,n,t){e.fileInfos.get(t).signature=n,e.hasCalledUpdateShapeSignature.add(t)}function u(t,a,i,o,s,c,l){if(e.Debug.assert(!!i),e.Debug.assert(!l||!!t.exportedModulesMap,"Compute visible to outside map only if visibleToOutsideReferencedMap present in the state"),t.hasCalledUpdateShapeSignature.has(i.resolvedPath)||o.has(i.resolvedPath))return!1;var u=t.fileInfos.get(i.resolvedPath);if(!u)return e.Debug.fail();var d,p=u.signature;if(i.isDeclarationFile){if(d=i.version,l&&d!==p){var m=t.referencedMap?t.referencedMap.get(i.resolvedPath):void 0;l.set(i.resolvedPath,m||!1)}}else{var f=n(a,i,!0,s,void 0,!0),_=f.outputFiles&&a.getCompilerOptions().declarationMap?f.outputFiles.length>1?f.outputFiles[1]:void 0:f.outputFiles.length>0?f.outputFiles[0]:void 0;_?(e.Debug.assert(e.fileExtensionIs(_.name,".d.ts"),"File extension for signature expected to be dts",(function(){return"Found: "+e.getAnyExtensionFromPath(_.name)+" for "+_.name+":: All output files: "+JSON.stringify(f.outputFiles.map((function(e){return e.name})))})),d=(c||e.generateDjb2Hash)(_.text),l&&d!==p&&function(n,t,a){if(!t)return void a.set(n.resolvedPath,!1);var i;function o(n){n&&(i||(i=new e.Set),i.add(n))}t.forEach((function(e){return o(r(e))})),a.set(n.resolvedPath,i||!1)}(i,f.exportedModulesFromDeclarationEmit,l)):d=p}return o.set(i.resolvedPath,d),!p||d!==p}function d(n,t){if(!n.allFileNames){var r=t.getSourceFiles();n.allFileNames=r===e.emptyArray?e.emptyArray:r.map((function(e){return e.fileName}))}return n.allFileNames}function p(n,t){return e.arrayFrom(e.mapDefinedIterator(n.referencedMap.entries(),(function(e){var n=e[0];return e[1].has(t)?n:void 0})))}function m(n){return function(n){return e.some(n.moduleAugmentations,(function(n){return e.isGlobalScopeAugmentation(n.parent)}))}(n)||!e.isExternalModule(n)&&!function(n){for(var t=0,r=n.statements;t<r.length;t++){var a=r[t];if(!e.isModuleWithStringLiteralName(a))return!1}return!0}(n)}function f(n,t,r){if(n.allFilesExcludingDefaultLibraryFile)return n.allFilesExcludingDefaultLibraryFile;var a;r&&c(r);for(var i=0,o=t.getSourceFiles();i<o.length;i++){var s=o[i];s!==r&&c(s)}return n.allFilesExcludingDefaultLibraryFile=a||e.emptyArray,n.allFilesExcludingDefaultLibraryFile;function c(e){t.isSourceFileDefaultLibrary(e)||(a||(a=[])).push(e)}}function _(n,t,r){var a=t.getCompilerOptions();return a&&e.outFile(a)?[r]:f(n,t,r)}function g(n,t,r,a,i,o,s){if(m(r))return f(n,t,r);var c=t.getCompilerOptions();if(c&&(c.isolatedModules||e.outFile(c)))return[r];var l=new e.Map;l.set(r.resolvedPath,r);for(var d=p(n,r.resolvedPath);d.length>0;){var _=d.pop();if(!l.has(_)){var g=t.getSourceFileByPath(_);l.set(_,g),g&&u(n,t,g,a,i,o,s)&&d.push.apply(d,p(n,g.resolvedPath))}}return e.arrayFrom(e.mapDefinedIterator(l.values(),(function(e){return e})))}t.canReuseOldState=s,t.create=function(n,t,r){var a=new e.Map,i=n.getCompilerOptions().module!==e.ModuleKind.None?new e.Map:void 0,c=i?new e.Map:void 0,l=new e.Set,u=s(i,r);n.getTypeChecker();for(var d=0,p=n.getSourceFiles();d<p.length;d++){var f=p[d],_=e.Debug.checkDefined(f.version,"Program intended to be used with Builder should have source files with versions set"),g=u?r.fileInfos.get(f.resolvedPath):void 0;if(i){var y=o(n,f,t);if(y&&i.set(f.resolvedPath,y),u){var v=r.exportedModulesMap.get(f.resolvedPath);v&&c.set(f.resolvedPath,v)}}a.set(f.resolvedPath,{version:_,signature:g&&g.signature,affectsGlobalScope:m(f)})}return{fileInfos:a,referencedMap:i,exportedModulesMap:c,hasCalledUpdateShapeSignature:l}},t.releaseCache=function(e){e.allFilesExcludingDefaultLibraryFile=void 0,e.allFileNames=void 0},t.clone=function(n){return{fileInfos:new e.Map(n.fileInfos),referencedMap:n.referencedMap&&new e.Map(n.referencedMap),exportedModulesMap:n.exportedModulesMap&&new e.Map(n.exportedModulesMap),hasCalledUpdateShapeSignature:new e.Set(n.hasCalledUpdateShapeSignature)}},t.getFilesAffectedBy=function(n,t,r,a,i,o,s){var l=o||new e.Map,d=t.getSourceFileByPath(r);if(!d)return e.emptyArray;if(!u(n,t,d,l,a,i,s))return[d];var p=(n.referencedMap?g:_)(n,t,d,l,a,i,s);return o||c(n,l),p},t.updateSignaturesFromCache=c,t.updateSignatureOfFile=l,t.updateShapeSignature=u,t.updateExportedFilesMapFromCache=function(n,t){t&&(e.Debug.assert(!!n.exportedModulesMap),t.forEach((function(e,t){e?n.exportedModulesMap.set(t,e):n.exportedModulesMap.delete(t)})))},t.getAllDependencies=function(n,t,r){var a=t.getCompilerOptions();if(e.outFile(a))return d(n,t);if(!n.referencedMap||m(r))return d(n,t);for(var i=new e.Set,o=[r.resolvedPath];o.length;){var s=o.pop();if(!i.has(s)){i.add(s);var c=n.referencedMap.get(s);if(c)for(var l=c.keys(),u=l.next();!u.done;u=l.next())o.push(u.value)}}return e.arrayFrom(e.mapDefinedIterator(i.keys(),(function(e){var n,r;return null!==(r=null===(n=t.getSourceFileByPath(e))||void 0===n?void 0:n.fileName)&&void 0!==r?r:e})))},t.getReferencedByPaths=p,t.getAllFilesExcludingDefaultLibraryFile=f}(e.BuilderState||(e.BuilderState={}))}(ts||(ts={})),function(e){var n;function t(n,t,a){var i=e.BuilderState.create(n,t,a);i.program=n;var o=n.getCompilerOptions();i.compilerOptions=o,e.outFile(o)||(i.semanticDiagnosticsPerFile=new e.Map),i.changedFilesSet=new e.Set;var s=e.BuilderState.canReuseOldState(i.referencedMap,a),c=s?a.compilerOptions:void 0,l=s&&a.semanticDiagnosticsPerFile&&!!i.semanticDiagnosticsPerFile&&!e.compilerOptionsAffectSemanticDiagnostics(o,c);if(s){if(!a.currentChangedFilePath){var u=a.currentAffectedFilesSignatures;e.Debug.assert(!(a.affectedFiles||u&&u.size),"Cannot reuse if only few affected files of currentChangedFile were iterated")}var d=a.changedFilesSet;l&&e.Debug.assert(!d||!e.forEachKey(d,(function(e){return a.semanticDiagnosticsPerFile.has(e)})),"Semantic diagnostics shouldnt be available for changed files"),null==d||d.forEach((function(e){return i.changedFilesSet.add(e)})),!e.outFile(o)&&a.affectedFilesPendingEmit&&(i.affectedFilesPendingEmit=a.affectedFilesPendingEmit.slice(),i.affectedFilesPendingEmitKind=a.affectedFilesPendingEmitKind&&new e.Map(a.affectedFilesPendingEmitKind),i.affectedFilesPendingEmitIndex=a.affectedFilesPendingEmitIndex,i.seenAffectedFiles=new e.Set)}var p=i.referencedMap,m=s?a.referencedMap:void 0,f=l&&!o.skipLibCheck==!c.skipLibCheck,_=f&&!o.skipDefaultLibCheck==!c.skipDefaultLibCheck;return i.fileInfos.forEach((function(o,c){var u,d,g,y;if(!s||!(u=a.fileInfos.get(c))||u.version!==o.version||(g=d=p&&p.get(c),y=m&&m.get(c),g!==y&&(void 0===g||void 0===y||g.size!==y.size||e.forEachKey(g,(function(e){return!y.has(e)}))))||d&&e.forEachKey(d,(function(e){return!i.fileInfos.has(e)&&a.fileInfos.has(e)})))i.changedFilesSet.add(c);else if(l){var v=n.getSourceFileByPath(c);if(v.isDeclarationFile&&!f)return;if(v.hasNoDefaultLib&&!_)return;var h=a.semanticDiagnosticsPerFile.get(c);h&&(i.semanticDiagnosticsPerFile.set(c,a.hasReusableDiagnostic?function(n,t,a){if(!n.length)return e.emptyArray;var i=e.getDirectoryPath(e.getNormalizedAbsolutePath(e.getTsBuildInfoEmitOutputFilePath(t.getCompilerOptions()),t.getCurrentDirectory()));return n.map((function(e){var n=r(e,t,o);n.reportsUnnecessary=e.reportsUnnecessary,n.reportsDeprecated=e.reportDeprecated,n.source=e.source,n.skippedOn=e.skippedOn;var a=e.relatedInformation;return n.relatedInformation=a?a.length?a.map((function(e){return r(e,t,o)})):[]:void 0,n}));function o(n){return e.toPath(n,i,a)}}(h,n,t):h),i.semanticDiagnosticsFromOldState||(i.semanticDiagnosticsFromOldState=new e.Set),i.semanticDiagnosticsFromOldState.add(c))}})),s&&e.forEachEntry(a.fileInfos,(function(e,n){return e.affectsGlobalScope&&!i.fileInfos.has(n)}))?e.BuilderState.getAllFilesExcludingDefaultLibraryFile(i,n,void 0).forEach((function(e){return i.changedFilesSet.add(e.resolvedPath)})):c&&!e.outFile(o)&&e.compilerOptionsAffectEmit(o,c)&&(n.getSourceFiles().forEach((function(e){return h(i,e.resolvedPath,1)})),e.Debug.assert(!i.seenAffectedFiles||!i.seenAffectedFiles.size),i.seenAffectedFiles=i.seenAffectedFiles||new e.Set),i.buildInfoEmitPending=!!i.changedFilesSet.size,i}function r(e,n,t){var r=e.file;return __assign(__assign({},e),{file:r?n.getSourceFileByPath(t(r)):void 0})}function a(n,t){e.Debug.assert(!t||!n.affectedFiles||n.affectedFiles[n.affectedFilesIndex-1]!==t||!n.semanticDiagnosticsPerFile.has(t.resolvedPath))}function i(n,t,r){for(;;){var a=n.affectedFiles;if(a){for(var i=n.seenAffectedFiles,s=n.affectedFilesIndex;s<a.length;){var c=a[s];if(!i.has(c.resolvedPath))return n.affectedFilesIndex=s,o(n,c,t,r),c;s++}n.changedFilesSet.delete(n.currentChangedFilePath),n.currentChangedFilePath=void 0,e.BuilderState.updateSignaturesFromCache(n,n.currentAffectedFilesSignatures),n.currentAffectedFilesSignatures.clear(),e.BuilderState.updateExportedFilesMapFromCache(n,n.currentAffectedFilesExportedModulesMap),n.affectedFiles=void 0}var l=n.changedFilesSet.keys().next();if(l.done)return;var u=e.Debug.checkDefined(n.program),d=u.getCompilerOptions();if(e.outFile(d))return e.Debug.assert(!n.semanticDiagnosticsPerFile),u;n.currentAffectedFilesSignatures||(n.currentAffectedFilesSignatures=new e.Map),n.exportedModulesMap&&(n.currentAffectedFilesExportedModulesMap||(n.currentAffectedFilesExportedModulesMap=new e.Map)),n.affectedFiles=e.BuilderState.getFilesAffectedBy(n,u,l.value,t,r,n.currentAffectedFilesSignatures,n.currentAffectedFilesExportedModulesMap),n.currentChangedFilePath=l.value,n.affectedFilesIndex=0,n.seenAffectedFiles||(n.seenAffectedFiles=new e.Set)}}function o(n,t,r,a){if(s(n,t.resolvedPath),n.allFilesExcludingDefaultLibraryFile!==n.affectedFiles)n.compilerOptions.assumeChangesOnlyAffectDirectDependencies||function(n,t,r){if(!n.exportedModulesMap||!n.changedFilesSet.has(t.resolvedPath))return;if(!c(n,t.resolvedPath))return;if(n.compilerOptions.isolatedModules){var a=new e.Map;a.set(t.resolvedPath,!0);for(var i=e.BuilderState.getReferencedByPaths(n,t.resolvedPath);i.length>0;){var o=i.pop();if(!a.has(o))if(a.set(o,!0),r(n,o)&&c(n,o)){var s=e.Debug.checkDefined(n.program).getSourceFileByPath(o);i.push.apply(i,e.BuilderState.getReferencedByPaths(n,s.resolvedPath))}}}e.Debug.assert(!!n.currentAffectedFilesExportedModulesMap);var u=new e.Set;if(e.forEachEntry(n.currentAffectedFilesExportedModulesMap,(function(e,a){return e&&e.has(t.resolvedPath)&&l(n,a,u,r)})))return;e.forEachEntry(n.exportedModulesMap,(function(e,a){return!n.currentAffectedFilesExportedModulesMap.has(a)&&e.has(t.resolvedPath)&&l(n,a,u,r)}))}(n,t,(function(n,t){return function(n,t,r,a){if(s(n,t),!n.changedFilesSet.has(t)){var i=e.Debug.checkDefined(n.program),o=i.getSourceFileByPath(t);o&&(e.BuilderState.updateShapeSignature(n,i,o,e.Debug.checkDefined(n.currentAffectedFilesSignatures),r,a,n.currentAffectedFilesExportedModulesMap),e.getEmitDeclarations(n.compilerOptions)&&h(n,t,0))}return!1}(n,t,r,a)}));else if(!n.cleanedDiagnosticsOfLibFiles){n.cleanedDiagnosticsOfLibFiles=!0;var i=e.Debug.checkDefined(n.program),o=i.getCompilerOptions();e.forEach(i.getSourceFiles(),(function(t){return i.isSourceFileDefaultLibrary(t)&&!e.skipTypeChecking(t,o,i)&&s(n,t.resolvedPath)}))}}function s(e,n){return!e.semanticDiagnosticsFromOldState||(e.semanticDiagnosticsFromOldState.delete(n),e.semanticDiagnosticsPerFile.delete(n),!e.semanticDiagnosticsFromOldState.size)}function c(n,t){return e.Debug.checkDefined(n.currentAffectedFilesSignatures).get(t)!==e.Debug.checkDefined(n.fileInfos.get(t)).signature}function l(n,t,r,a){return e.forEachEntry(n.referencedMap,(function(e,i){return e.has(t)&&u(n,i,r,a)}))}function u(n,t,r,a){return!!e.tryAddToSet(r,t)&&(!!a(n,t)||(e.Debug.assert(!!n.currentAffectedFilesExportedModulesMap),!!e.forEachEntry(n.currentAffectedFilesExportedModulesMap,(function(e,i){return e&&e.has(t)&&u(n,i,r,a)}))||(!!e.forEachEntry(n.exportedModulesMap,(function(e,i){return!n.currentAffectedFilesExportedModulesMap.has(i)&&e.has(t)&&u(n,i,r,a)}))||!!e.forEachEntry(n.referencedMap,(function(e,i){return e.has(t)&&!r.has(i)&&a(n,i)})))))}function d(n,t,r,a,i){i?n.buildInfoEmitPending=!1:t===n.program?(n.changedFilesSet.clear(),n.programEmitComplete=!0):(n.seenAffectedFiles.add(t.resolvedPath),void 0!==r&&(n.seenEmittedFiles||(n.seenEmittedFiles=new e.Map)).set(t.resolvedPath,r),a?(n.affectedFilesPendingEmitIndex++,n.buildInfoEmitPending=!0):n.affectedFilesIndex++)}function p(e,n,t){return d(e,t),{result:n,affected:t}}function m(e,n,t,r,a,i){return d(e,t,r,a,i),{result:n,affected:t}}function f(n,t,r){return e.concatenate(function(n,t,r){var a=t.resolvedPath;if(n.semanticDiagnosticsPerFile){var i=n.semanticDiagnosticsPerFile.get(a);if(i)return e.filterSemanticDiagnotics(i,n.compilerOptions)}var o=e.Debug.checkDefined(n.program).getBindAndCheckDiagnostics(t,r);n.semanticDiagnosticsPerFile&&n.semanticDiagnosticsPerFile.set(a,o);return e.filterSemanticDiagnotics(o,n.compilerOptions)}(n,t,r),e.Debug.checkDefined(n.program).getProgramDiagnostics(t))}function _(n,t){var r={},a=e.getOptionsNameMap().optionsNameMap;for(var i in n)e.hasProperty(n,i)&&(r[i]=g(a.get(i.toLowerCase()),n[i],t));return r.configFilePath&&(r.configFilePath=t(r.configFilePath)),r}function g(e,n,t){if(e)if("list"===e.type){var r=n;if(e.element.isFilePath&&r.length)return r.map(t)}else if(e.isFilePath)return t(n);return n}function y(n,t){return e.Debug.assert(!!n.length),n.map((function(e){var n=v(e,t);n.reportsUnnecessary=e.reportsUnnecessary,n.reportDeprecated=e.reportsDeprecated,n.source=e.source,n.skippedOn=e.skippedOn;var r=e.relatedInformation;return n.relatedInformation=r?r.length?r.map((function(e){return v(e,t)})):[]:void 0,n}))}function v(e,n){var t=e.file;return __assign(__assign({},e),{file:t?n(t.resolvedPath):void 0})}function h(n,t,r){n.affectedFilesPendingEmit||(n.affectedFilesPendingEmit=[]),n.affectedFilesPendingEmitKind||(n.affectedFilesPendingEmitKind=new e.Map);var a=n.affectedFilesPendingEmitKind.get(t);n.affectedFilesPendingEmit.push(t),n.affectedFilesPendingEmitKind.set(t,a||r),void 0===n.affectedFilesPendingEmitIndex&&(n.affectedFilesPendingEmitIndex=0)}function b(n,t){if(n){var r=new e.Map;for(var a in n)e.hasProperty(n,a)&&r.set(t(a),new e.Set(n[a].map(t)));return r}}function E(n,t){return{getState:e.notImplemented,backupState:e.noop,restoreState:e.noop,getProgram:r,getProgramOrUndefined:function(){return n.program},releaseProgram:function(){return n.program=void 0},getCompilerOptions:function(){return n.compilerOptions},getSourceFile:function(e){return r().getSourceFile(e)},getSourceFiles:function(){return r().getSourceFiles()},getOptionsDiagnostics:function(e){return r().getOptionsDiagnostics(e)},getGlobalDiagnostics:function(e){return r().getGlobalDiagnostics(e)},getConfigFileParsingDiagnostics:function(){return t},getSyntacticDiagnostics:function(e,n){return r().getSyntacticDiagnostics(e,n)},getDeclarationDiagnostics:function(e,n){return r().getDeclarationDiagnostics(e,n)},getSemanticDiagnostics:function(e,n){return r().getSemanticDiagnostics(e,n)},emit:function(e,n,t,a,i){return r().emit(e,n,t,a,i)},emitBuildInfo:function(e,n){return r().emitBuildInfo(e,n)},getAllDependencies:e.notImplemented,getCurrentDirectory:function(){return r().getCurrentDirectory()},close:e.noop};function r(){return e.Debug.checkDefined(n.program)}}!function(e){e[e.DtsOnly=0]="DtsOnly",e[e.Full=1]="Full"}(e.BuilderFileEmit||(e.BuilderFileEmit={})),function(e){e[e.SemanticDiagnosticsBuilderProgram=0]="SemanticDiagnosticsBuilderProgram",e[e.EmitAndSemanticDiagnosticsBuilderProgram=1]="EmitAndSemanticDiagnosticsBuilderProgram"}(n=e.BuilderProgramKind||(e.BuilderProgramKind={})),e.getBuilderCreationParameters=function(n,t,r,a,i,o){var s,c,l;return void 0===n?(e.Debug.assert(void 0===t),s=r,l=a,e.Debug.assert(!!l),c=l.getProgram()):e.isArray(n)?(l=a,c=e.createProgram({rootNames:n,options:t,host:r,oldProgram:l&&l.getProgramOrUndefined(),configFileParsingDiagnostics:i,projectReferences:o}),s=r):(c=n,s=t,l=r,i=a),{host:s,newProgram:c,oldProgram:l,configFileParsingDiagnostics:i||e.emptyArray}},e.createBuilderProgram=function(r,o){var s=o.newProgram,c=o.host,l=o.oldProgram,u=o.configFileParsingDiagnostics,g=l&&l.getState();if(g&&s===g.program&&u===s.getConfigFileParsingDiagnostics())return s=void 0,g=void 0,l;var v,b=e.createGetCanonicalFileName(c.useCaseSensitiveFileNames()),T=e.maybeBind(c,c.createHash),S=t(s,b,g);s.getProgramBuildInfo=function(){return function(n,t){if(!e.outFile(n.compilerOptions)){var r=e.Debug.checkDefined(n.program).getCurrentDirectory(),a=e.getDirectoryPath(e.getNormalizedAbsolutePath(e.getTsBuildInfoEmitOutputFilePath(n.compilerOptions),r)),i={};n.fileInfos.forEach((function(e,t){var r=n.currentAffectedFilesSignatures&&n.currentAffectedFilesSignatures.get(t);i[A(t)]=void 0===r?e:{version:e.version,signature:r,affectsGlobalScope:e.affectsGlobalScope}}));var o={fileInfos:i,options:_(n.compilerOptions,(function(n){return A(e.getNormalizedAbsolutePath(n,r))}))};if(n.referencedMap){for(var s={},c=0,l=e.arrayFrom(n.referencedMap.keys()).sort(e.compareStringsCaseSensitive);c<l.length;c++)s[A(m=l[c])]=e.arrayFrom(n.referencedMap.get(m).keys(),A).sort(e.compareStringsCaseSensitive);o.referencedMap=s}if(n.exportedModulesMap){for(var u={},d=0,p=e.arrayFrom(n.exportedModulesMap.keys()).sort(e.compareStringsCaseSensitive);d<p.length;d++){var m=p[d],f=n.currentAffectedFilesExportedModulesMap&&n.currentAffectedFilesExportedModulesMap.get(m);void 0===f?u[A(m)]=e.arrayFrom(n.exportedModulesMap.get(m).keys(),A).sort(e.compareStringsCaseSensitive):f&&(u[A(m)]=e.arrayFrom(f.keys(),A).sort(e.compareStringsCaseSensitive))}o.exportedModulesMap=u}if(n.semanticDiagnosticsPerFile){for(var g=[],v=0,h=e.arrayFrom(n.semanticDiagnosticsPerFile.keys()).sort(e.compareStringsCaseSensitive);v<h.length;v++){m=h[v];var b=n.semanticDiagnosticsPerFile.get(m);g.push(b.length?[A(m),n.hasReusableDiagnostic?b:y(b,A)]:A(m))}o.semanticDiagnosticsPerFile=g}if(n.affectedFilesPendingEmit){for(var E=[],T=new e.Set,S=0,x=n.affectedFilesPendingEmit.slice(n.affectedFilesPendingEmitIndex).sort(e.compareStringsCaseSensitive);S<x.length;S++){var L=x[S];e.tryAddToSet(T,L)&&E.push([A(L),n.affectedFilesPendingEmitKind.get(L)])}o.affectedFilesPendingEmit=E}return o}function A(n){return e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(a,n,t))}}(S,b)},s=void 0,l=void 0,g=void 0;var x=E(S,u);return x.getState=function(){return S},x.backupState=function(){e.Debug.assert(void 0===v),v=function(n){var t=e.BuilderState.clone(n);return t.semanticDiagnosticsPerFile=n.semanticDiagnosticsPerFile&&new e.Map(n.semanticDiagnosticsPerFile),t.changedFilesSet=new e.Set(n.changedFilesSet),t.affectedFiles=n.affectedFiles,t.affectedFilesIndex=n.affectedFilesIndex,t.currentChangedFilePath=n.currentChangedFilePath,t.currentAffectedFilesSignatures=n.currentAffectedFilesSignatures&&new e.Map(n.currentAffectedFilesSignatures),t.currentAffectedFilesExportedModulesMap=n.currentAffectedFilesExportedModulesMap&&new e.Map(n.currentAffectedFilesExportedModulesMap),t.seenAffectedFiles=n.seenAffectedFiles&&new e.Set(n.seenAffectedFiles),t.cleanedDiagnosticsOfLibFiles=n.cleanedDiagnosticsOfLibFiles,t.semanticDiagnosticsFromOldState=n.semanticDiagnosticsFromOldState&&new e.Set(n.semanticDiagnosticsFromOldState),t.program=n.program,t.compilerOptions=n.compilerOptions,t.affectedFilesPendingEmit=n.affectedFilesPendingEmit&&n.affectedFilesPendingEmit.slice(),t.affectedFilesPendingEmitKind=n.affectedFilesPendingEmitKind&&new e.Map(n.affectedFilesPendingEmitKind),t.affectedFilesPendingEmitIndex=n.affectedFilesPendingEmitIndex,t.seenEmittedFiles=n.seenEmittedFiles&&new e.Map(n.seenEmittedFiles),t.programEmitComplete=n.programEmitComplete,t}(S)},x.restoreState=function(){S=e.Debug.checkDefined(v),v=void 0},x.getAllDependencies=function(n){return e.BuilderState.getAllDependencies(S,e.Debug.checkDefined(S.program),n)},x.getSemanticDiagnostics=function(n,t){a(S,n);var r,i=e.Debug.checkDefined(S.program).getCompilerOptions();if(e.outFile(i))return e.Debug.assert(!S.semanticDiagnosticsPerFile),e.Debug.checkDefined(S.program).getSemanticDiagnostics(n,t);if(n)return f(S,n,t);for(;A(t););for(var o=0,s=e.Debug.checkDefined(S.program).getSourceFiles();o<s.length;o++){var c=s[o];r=e.addRange(r,f(S,c,t))}return r||e.emptyArray},x.emit=function(t,i,o,s,l){var u,d,p,m=!1;r===n.EmitAndSemanticDiagnosticsBuilderProgram||t||e.outFile(S.compilerOptions)||S.compilerOptions.noEmit||!S.compilerOptions.noEmitOnError||(m=!0,u=S.affectedFilesPendingEmit&&S.affectedFilesPendingEmit.slice(),d=S.affectedFilesPendingEmitKind&&new e.Map(S.affectedFilesPendingEmitKind),p=S.affectedFilesPendingEmitIndex);r===n.EmitAndSemanticDiagnosticsBuilderProgram&&a(S,t);var f=e.handleNoEmitOptions(x,t,i,o);if(f)return f;m&&(S.affectedFilesPendingEmit=u,S.affectedFilesPendingEmitKind=d,S.affectedFilesPendingEmitIndex=p);if(!t&&r===n.EmitAndSemanticDiagnosticsBuilderProgram){for(var _=[],g=!1,y=void 0,v=[],h=void 0;h=L(i,o,s,l);)g=g||h.result.emitSkipped,y=e.addRange(y,h.result.diagnostics),v=e.addRange(v,h.result.emittedFiles),_=e.addRange(_,h.result.sourceMaps);return{emitSkipped:g,diagnostics:y||e.emptyArray,emittedFiles:v,sourceMaps:_}}return e.Debug.checkDefined(S.program).emit(t,i||e.maybeBind(c,c.writeFile),o,s,l)},x.releaseProgram=function(){!function(n){e.BuilderState.releaseCache(n),n.program=void 0}(S),v=void 0},r===n.SemanticDiagnosticsBuilderProgram?x.getSemanticDiagnosticsOfNextAffectedFile=A:r===n.EmitAndSemanticDiagnosticsBuilderProgram?(x.getSemanticDiagnosticsOfNextAffectedFile=A,x.emitNextAffectedFile=L,x.emitBuildInfo=function(n,t){if(S.buildInfoEmitPending){var r=e.Debug.checkDefined(S.program).emitBuildInfo(n||e.maybeBind(c,c.writeFile),t);return S.buildInfoEmitPending=!1,r}return e.emitSkippedWithNoDiagnostics}):e.notImplemented(),x;function L(n,t,r,a){var o=i(S,t,T),s=1,l=!1;if(!o)if(e.outFile(S.compilerOptions)){var u=e.Debug.checkDefined(S.program);if(S.programEmitComplete)return;o=u}else{var d=function(n){var t=n.affectedFilesPendingEmit;if(t){for(var r=n.seenEmittedFiles||(n.seenEmittedFiles=new e.Map),a=n.affectedFilesPendingEmitIndex;a<t.length;a++){var i=e.Debug.checkDefined(n.program).getSourceFileByPath(t[a]);if(i){var o=r.get(i.resolvedPath),s=e.Debug.checkDefined(e.Debug.checkDefined(n.affectedFilesPendingEmitKind).get(i.resolvedPath));if(void 0===o||o<s)return n.affectedFilesPendingEmitIndex=a,{affectedFile:i,emitKind:s}}}n.affectedFilesPendingEmit=void 0,n.affectedFilesPendingEmitKind=void 0,n.affectedFilesPendingEmitIndex=void 0}}(S);if(!d){if(!S.buildInfoEmitPending)return;var p=e.Debug.checkDefined(S.program);return m(S,p.emitBuildInfo(n||e.maybeBind(c,c.writeFile),t),p,1,!1,!0)}o=d.affectedFile,s=d.emitKind,l=!0}return m(S,e.Debug.checkDefined(S.program).emit(o===S.program?void 0:o,n||e.maybeBind(c,c.writeFile),t,r||0===s,a),o,s,l)}function A(e,t){for(;;){var a=i(S,e,T);if(!a)return;if(a===S.program)return p(S,S.program.getSemanticDiagnostics(void 0,e),a);if((r===n.EmitAndSemanticDiagnosticsBuilderProgram||S.compilerOptions.noEmit||S.compilerOptions.noEmitOnError)&&h(S,a.resolvedPath,1),!t||!t(a))return p(S,f(S,a,e),a);d(S,a)}}},e.createBuildProgramUsingProgramBuildInfo=function(n,t,r){var a=e.getDirectoryPath(e.getNormalizedAbsolutePath(t,r.getCurrentDirectory())),i=e.createGetCanonicalFileName(r.useCaseSensitiveFileNames()),o=new e.Map;for(var s in n.fileInfos)e.hasProperty(n.fileInfos,s)&&o.set(l(s),n.fileInfos[s]);var c={fileInfos:o,compilerOptions:e.convertToOptionsWithAbsolutePaths(n.options,(function(n){return e.getNormalizedAbsolutePath(n,a)})),referencedMap:b(n.referencedMap,l),exportedModulesMap:b(n.exportedModulesMap,l),semanticDiagnosticsPerFile:n.semanticDiagnosticsPerFile&&e.arrayToMap(n.semanticDiagnosticsPerFile,(function(n){return l(e.isString(n)?n:n[0])}),(function(n){return e.isString(n)?e.emptyArray:n[1]})),hasReusableDiagnostic:!0,affectedFilesPendingEmit:e.map(n.affectedFilesPendingEmit,(function(e){return l(e[0])})),affectedFilesPendingEmitKind:n.affectedFilesPendingEmit&&e.arrayToMap(n.affectedFilesPendingEmit,(function(e){return l(e[0])}),(function(e){return e[1]})),affectedFilesPendingEmitIndex:n.affectedFilesPendingEmit&&0};return{getState:function(){return c},backupState:e.noop,restoreState:e.noop,getProgram:e.notImplemented,getProgramOrUndefined:e.returnUndefined,releaseProgram:e.noop,getCompilerOptions:function(){return c.compilerOptions},getSourceFile:e.notImplemented,getSourceFiles:e.notImplemented,getOptionsDiagnostics:e.notImplemented,getGlobalDiagnostics:e.notImplemented,getConfigFileParsingDiagnostics:e.notImplemented,getSyntacticDiagnostics:e.notImplemented,getDeclarationDiagnostics:e.notImplemented,getSemanticDiagnostics:e.notImplemented,emit:e.notImplemented,getAllDependencies:e.notImplemented,getCurrentDirectory:e.notImplemented,emitNextAffectedFile:e.notImplemented,getSemanticDiagnosticsOfNextAffectedFile:e.notImplemented,emitBuildInfo:e.notImplemented,close:e.noop};function l(n){return e.toPath(n,a,i)}},e.createRedirectedBuilderProgram=E}(ts||(ts={})),function(e){e.createSemanticDiagnosticsBuilderProgram=function(n,t,r,a,i,o){return e.createBuilderProgram(e.BuilderProgramKind.SemanticDiagnosticsBuilderProgram,e.getBuilderCreationParameters(n,t,r,a,i,o))},e.createEmitAndSemanticDiagnosticsBuilderProgram=function(n,t,r,a,i,o){return e.createBuilderProgram(e.BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram,e.getBuilderCreationParameters(n,t,r,a,i,o))},e.createAbstractBuilder=function(n,t,r,a,i,o){var s=e.getBuilderCreationParameters(n,t,r,a,i,o),c=s.newProgram,l=s.configFileParsingDiagnostics;return e.createRedirectedBuilderProgram({program:c,compilerOptions:c.getCompilerOptions()},l)}}(ts||(ts={})),function(e){function n(n){return e.endsWith(n,"/node_modules/.staging")?e.removeSuffix(n,"/.staging"):e.some(e.ignoredPaths,(function(t){return e.stringContains(n,t)}))?void 0:n}function t(n){var t=e.getRootLength(n);if(n.length===t)return!1;var r=n.indexOf(e.directorySeparator,t);if(-1===r)return!1;var a=n.substring(t,r+1),i=t>1||47!==n.charCodeAt(0);if(i&&0!==n.search(/[a-zA-Z]:/)&&0===a.search(/[a-zA-z]\$\//)){if(-1===(r=n.indexOf(e.directorySeparator,r+1)))return!1;a=n.substring(t+a.length,r+1)}if(i&&0!==a.search(/users\//i))return!0;for(var o=r+1,s=2;s>0;s--)if(0===(o=n.indexOf(e.directorySeparator,o)+1))return!1;return!0}e.removeIgnoredPath=n,e.canWatchDirectory=t,e.createResolutionCache=function(r,a,i){var o,s,c,l=e.createMultiMap(),u=[],d=e.createMultiMap(),p=!1,m=[],f=[],_=[],g=e.memoize((function(){return r.getCurrentDirectory()})),y=r.getCachedDirectoryStructureHost(),v=new e.Map,h=e.createCacheWithRedirects(),b=e.createCacheWithRedirects(),E=e.createModuleResolutionCacheWithMaps(h,b,g(),r.getCanonicalFileName),T=new e.Map,S=e.createCacheWithRedirects(),x=[".ts",".tsx",".js",".jsx",".json"],L=new e.Map,A=new e.Map,D=a&&e.removeTrailingDirectorySeparator(e.getNormalizedAbsolutePath(a,g())),C=D&&r.toPath(D),I=void 0!==C?C.split(e.directorySeparator).length:0,N=new e.Map;return{startRecordingFilesWithChangedResolutions:function(){o=[]},finishRecordingFilesWithChangedResolutions:function(){var e=o;return o=void 0,e},startCachingPerDirectoryResolution:M,finishCachingPerDirectoryResolution:function(){c=void 0,M(),A.forEach((function(e,n){0===e.refCount&&(A.delete(n),e.watcher.close())})),p=!1},resolveModuleNames:function(n,t,r,a){return F({names:n,containingFile:t,redirectedReference:a,cache:v,perDirectoryCacheWithRedirects:h,loader:w,getResolutionWithResolvedFileName:R,shouldRetryResolution:function(n){return!n.resolvedModule||!e.resolutionExtensionIsTSOrJson(n.resolvedModule.extension)},reusedNames:r,logChanges:i})},getResolvedModuleWithFailedLookupLocationsFromCache:function(e,n){var t=v.get(r.toPath(n));return t&&t.get(e)},resolveTypeReferenceDirectives:function(n,t,r){return F({names:n,containingFile:t,redirectedReference:r,cache:T,perDirectoryCacheWithRedirects:S,loader:e.resolveTypeReferenceDirective,getResolutionWithResolvedFileName:k,shouldRetryResolution:function(e){return void 0===e.resolvedTypeReferenceDirective}})},removeResolutionsFromProjectReferenceRedirects:function(n){if(!e.fileExtensionIs(n,".json"))return;var t=r.getCurrentProgram();if(!t)return;var a=t.getResolvedProjectReferenceByPath(n);if(!a)return;a.commandLine.fileNames.forEach((function(e){return Y(r.toPath(e))}))},removeResolutionsOfFile:Y,hasChangedAutomaticTypeDirectiveNames:function(){return p},invalidateResolutionOfFile:function(n){Y(n);var t=p;Q(d.get(n),e.returnTrue)&&p&&!t&&r.onChangedAutomaticTypeDirectiveNames()},invalidateResolutionsOfFailedLookupLocations:$,setFilesWithInvalidatedNonRelativeUnresolvedImports:function(n){e.Debug.assert(c===n||void 0===c),c=n},createHasInvalidatedResolution:function(n){if($(),n)return s=void 0,e.returnTrue;var t=s;return s=void 0,function(e){return!!t&&t.has(e)||P(e)}},updateTypeRootsWatch:function(){var n=r.getCompilationSettings();if(n.types)return void ne();var t=e.getEffectiveTypeRoots(n,{directoryExists:re,getCurrentDirectory:g});t?e.mutateMap(N,e.arrayToMap(t,(function(e){return r.toPath(e)})),{createNewValue:te,onDeleteValue:e.closeFileWatcher}):ne()},closeTypeRootsWatch:ne,clear:function(){e.clearMap(A,e.closeFileWatcherOf),L.clear(),l.clear(),ne(),v.clear(),T.clear(),d.clear(),u.length=0,m.length=0,f.length=0,_.length=0,M(),p=!1}};function R(e){return e.resolvedModule}function k(e){return e.resolvedTypeReferenceDirective}function O(n,t){return!(void 0===n||t.length<=n.length)&&(e.startsWith(t,n)&&t[n.length]===e.directorySeparator)}function P(e){if(!c)return!1;var n=c.get(e);return!!n&&!!n.length}function M(){h.clear(),b.clear(),S.clear(),l.forEach(K),l.clear()}function w(n,t,a,i,o){var s,c=e.resolveModuleName(n,t,a,i,E,o);if(!r.getGlobalCache)return c;var l=r.getGlobalCache();if(!(void 0===l||e.isExternalModuleNameRelative(n)||c.resolvedModule&&e.extensionIsTS(c.resolvedModule.extension))){var u=e.loadModuleFromGlobalCache(e.Debug.checkDefined(r.globalCacheResolutionModuleName)(n),r.projectName,a,i,l),d=u.resolvedModule,p=u.failedLookupLocations;if(d)return c.resolvedModule=d,(s=c.failedLookupLocations).push.apply(s,p),c}return c}function F(n){var t,a=n.names,i=n.containingFile,s=n.redirectedReference,c=n.cache,l=n.perDirectoryCacheWithRedirects,u=n.loader,d=n.getResolutionWithResolvedFileName,p=n.shouldRetryResolution,m=n.reusedNames,f=n.logChanges,_=r.toPath(i),g=c.get(_)||c.set(_,new e.Map).get(_),y=e.getDirectoryPath(_),v=l.getOrCreateMapOfCacheRedirects(s),h=v.get(y);h||(h=new e.Map,v.set(y,h));for(var b=[],E=r.getCompilationSettings(),T=f&&P(_),S=r.getCurrentProgram(),x=S&&S.getResolvedProjectReferenceToRedirect(i),L=x?!s||s.sourceFile.path!==x.sourceFile.path:!!s,A=new e.Map,D=0,C=a;D<C.length;D++){var I=C[D],N=g.get(I);if(!A.has(I)&&L||!N||N.isInvalidated||T&&!e.isExternalModuleNameRelative(I)&&p(N)){var R=N,k=h.get(I);k?N=k:(N=u(I,i,E,(null===(t=r.getCompilerHost)||void 0===t?void 0:t.call(r))||r,s),h.set(I,N)),g.set(I,N),j(I,N,_,d),R&&z(R,_,d),f&&o&&!O(R,N)&&(o.push(_),f=!1)}e.Debug.assert(void 0!==N&&!N.isInvalidated),A.set(I,!0),b.push(d(N))}return g.forEach((function(n,t){A.has(t)||e.contains(m,t)||(z(n,_,d),g.delete(t))})),b;function O(e,n){if(e===n)return!0;if(!e||!n)return!1;var t=d(e),r=d(n);return t===r||!(!t||!r)&&t.resolvedFileName===r.resolvedFileName}}function G(n){return e.endsWith(n,"/node_modules/@types")}function B(n,t){if(O(C,t)){n=e.isRootedDiskPath(n)?e.normalizePath(n):e.getNormalizedAbsolutePath(n,g());var r=t.split(e.directorySeparator),a=n.split(e.directorySeparator);return e.Debug.assert(a.length===r.length,"FailedLookup: "+n+" failedLookupLocationPath: "+t),r.length>I+1?{dir:a.slice(0,I+1).join(e.directorySeparator),dirPath:r.slice(0,I+1).join(e.directorySeparator)}:{dir:D,dirPath:C,nonRecursive:!1}}return U(e.getDirectoryPath(e.getNormalizedAbsolutePath(n,g())),e.getDirectoryPath(t))}function U(n,r){for(;e.pathContainsNodeModules(r);)n=e.getDirectoryPath(n),r=e.getDirectoryPath(r);if(e.isNodeModulesDirectory(r))return t(e.getDirectoryPath(r))?{dir:n,dirPath:r}:void 0;var a,i,o=!0;if(void 0!==C)for(;!O(r,C);){var s=e.getDirectoryPath(r);if(s===r)break;o=!1,a=r,i=n,r=s,n=e.getDirectoryPath(n)}return t(r)?{dir:i||n,dirPath:a||r,nonRecursive:o}:void 0}function V(n){return e.fileExtensionIsOneOf(n,x)}function j(n,t,a,i){if(t.refCount)t.refCount++,e.Debug.assertDefined(t.files);else{t.refCount=1,e.Debug.assert(0===e.length(t.files)),e.isExternalModuleNameRelative(n)?H(t):l.add(n,t);var o=i(t);o&&o.resolvedFileName&&d.add(r.toPath(o.resolvedFileName),t)}(t.files||(t.files=[])).push(a)}function H(n){e.Debug.assert(!!n.refCount);var t=n.failedLookupLocations;if(t.length){u.push(n);for(var a=!1,i=0,o=t;i<o.length;i++){var s=o[i],c=r.toPath(s),l=B(s,c);if(l){var d=l.dir,p=l.dirPath,m=l.nonRecursive;if(!V(c)){var f=L.get(c)||0;L.set(c,f+1)}p===C?(e.Debug.assert(!m),a=!0):W(d,p,m)}}a&&W(D,C,!0)}}function K(e,n){var t=r.getCurrentProgram();t&&t.getTypeChecker().tryFindAmbientModuleWithoutAugmentations(n)||e.forEach(H)}function W(n,t,r){var a=A.get(t);a?(e.Debug.assert(!!r==!!a.nonRecursive),a.refCount++):A.set(t,{watcher:J(n,t,r),refCount:1,nonRecursive:r})}function z(n,t,a){if(e.unorderedRemoveItem(e.Debug.assertDefined(n.files),t),n.refCount--,!n.refCount){var i=a(n);if(i&&i.resolvedFileName&&d.remove(r.toPath(i.resolvedFileName),n),e.unorderedRemoveItem(u,n)){for(var o=!1,s=0,c=n.failedLookupLocations;s<c.length;s++){var l=c[s],p=r.toPath(l),m=B(l,p);if(m){var f=m.dirPath,_=L.get(p);_&&(1===_?L.delete(p):(e.Debug.assert(_>1),L.set(p,_-1))),f===C?o=!0:q(f)}}o&&q(C)}}}function q(e){A.get(e).refCount--}function J(e,n,t){return r.watchDirectoryOfFailedLookupLocation(e,(function(e){var t=r.toPath(e);y&&y.addOrDeleteFileOrDirectory(e,t),Z(t,n===t)}),t?0:1)}function X(e,n,t){var r=e.get(n);r&&(r.forEach((function(e){return z(e,n,t)})),e.delete(n))}function Y(e){X(v,e,R),X(T,e,k)}function Q(n,t){if(!n)return!1;for(var r=!1,a=0,i=n;a<i.length;a++){var o=i[a];if(!o.isInvalidated&&t(o)){o.isInvalidated=r=!0;for(var c=0,l=e.Debug.assertDefined(o.files);c<l.length;c++){var u=l[c];(s||(s=new e.Set)).add(u),p=p||e.endsWith(u,e.inferredTypesContainingFile)}}}return r}function Z(t,a){if(a)_.push(t);else{var i=n(t);if(!i)return!1;if(t=i,r.fileIsOpen(t))return!1;var o=e.getDirectoryPath(t);if(G(t)||e.isNodeModulesDirectory(t)||G(o)||e.isNodeModulesDirectory(o))m.push(t),f.push(t);else{if(!V(t)&&!L.has(t))return!1;if(e.isEmittedFileOfProgram(r.getCurrentProgram(),t))return!1;m.push(t)}}r.scheduleInvalidateResolutionsOfFailedLookupLocations()}function $(){if(!m.length&&!f.length&&!_.length)return!1;var e=Q(u,ee);return m.length=0,f.length=0,_.length=0,e}function ee(n){return n.failedLookupLocations.some((function(n){var t=r.toPath(n);return e.contains(m,t)||f.some((function(n){return e.startsWith(t,n)}))||_.some((function(e){return O(e,t)}))}))}function ne(){e.clearMap(N,e.closeFileWatcher)}function te(e,n){return r.watchTypeRootsDirectory(n,(function(t){var a=r.toPath(t);y&&y.addOrDeleteFileOrDirectory(t,a),p=!0,r.onChangedAutomaticTypeDirectiveNames();var i=function(e,n){if(O(C,n))return C;var t=U(e,n);return t&&A.has(t.dirPath)?t.dirPath:void 0}(n,e);i&&Z(a,i===a)}),1)}function re(n){var a=e.getDirectoryPath(e.getDirectoryPath(n)),i=r.toPath(a);return i===C||t(i)}}}(ts||(ts={})),function(e){!function(n){var t,r;function a(n,t,r){var a=n.importModuleSpecifierPreference,i=n.importModuleSpecifierEnding;return{relativePreference:"relative"===a?0:"non-relative"===a?1:2,ending:function(){switch(i){case"minimal":return 0;case"index":return 1;case"js":return 2;default:return function(n){var t=n.imports;return e.firstDefined(t,(function(n){var t=n.text;return e.pathIsRelative(t)?e.hasJSFileExtension(t):void 0}))||!1}(r)?2:e.getEmitModuleResolutionKind(t)!==e.ModuleResolutionKind.NodeJs?1:0}}()}}function i(n,t,r,a,i){var c=o(t,a),l=p(t,r,a);return e.firstDefined(l,(function(e){return f(e,c,a,n)}))||s(r,c,n,a,i)}function o(n,t){return{getCanonicalFileName:e.createGetCanonicalFileName(!t.useCaseSensitiveFileNames||t.useCaseSensitiveFileNames()),sourceDirectory:e.getDirectoryPath(n)}}function s(n,t,r,a,i){var o=t.getCanonicalFileName,s=t.sourceDirectory,l=i.ending,u=i.relativePreference,d=r.baseUrl,p=r.paths,f=r.rootDirs,h=f&&function(n,t,r,a,i,o){var s=_(t,n,a);if(void 0===s)return;var c=_(r,n,a),l=void 0!==c?e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(c,s,a)):s;return e.getEmitModuleResolutionKind(o)===e.ModuleResolutionKind.NodeJs?g(l,i,o):e.removeFileExtension(l)}(f,n,s,o,l,r)||g(e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(s,n,o)),l,r);if(!d&&!p||0===u)return h;var b=y(n,e.getPathsBasePath(r,a)||d,o);if(!b)return h;var E=g(b,l,r),T=p&&m(e.removeFileExtension(b),E,p),S=void 0===T&&void 0!==d?E:T;return S?1===u?S:(2!==u&&e.Debug.assertNever(u),v(S)||c(h)<c(S)?h:S):h}function c(n){for(var t=0,r=e.startsWith(n,"./")?2:0;r<n.length;r++)47===n.charCodeAt(r)&&t++;return t}function l(e){var n=e.match(/\//g);return n?n.length:0}function u(n,t){return e.compareBooleans(t.isRedirect,n.isRedirect)||e.compareValues(l(n.path),l(t.path))}function d(n,t,r,a,i){var o=e.hostGetCanonicalFileName(r),s=r.getCurrentDirectory(),c=r.isSourceOfProjectReferenceRedirect(t)?r.getProjectReferenceRedirect(t):void 0,l=r.redirectTargetsMap.get(e.toPath(t,s,o))||e.emptyArray,u=__spreadArrays(c?[c]:e.emptyArray,[t],l).map((function(n){return e.getNormalizedAbsolutePath(n,s)}));if(!a){var d=e.forEach(u,(function(e){return i(e,c===e)}));if(d)return d}var p=(r.getSymlinkCache?r.getSymlinkCache():e.discoverProbableSymlinks(r.getSourceFiles(),o,s)).getSymlinkedDirectories(),m=!r.useCaseSensitiveFileNames||r.useCaseSensitiveFileNames();return p&&e.forEachEntry(p,(function(t,a){if(!1!==t&&!e.startsWithDirectory(n,t.realPath,o))return e.forEach(u,(function(n){if(e.containsPath(t.real,n,!m)){var s=e.getRelativePathFromDirectory(t.real,n,o),l=e.resolvePath(a,s);if(!r.fileExists||r.fileExists(l)){var u=i(l,n===c);if(u)return u}}}))}))||(a?e.forEach(u,(function(e){return i(e,e===c)})):void 0)}function p(n,t,r){var a=r.getCurrentDirectory(),i=e.hostGetCanonicalFileName(r),o=new e.Map,s=!1;d(n,t,r,!0,(function(n,t){var r=e.pathContainsNodeModules(n);o.set(n,{path:i(n),isRedirect:t,isInNodeModules:r}),s=s||r}));for(var c,l=[],p=function(n){var t,r=e.ensureTrailingDirectorySeparator(n);o.forEach((function(n,a){var i=n.path,s=n.isRedirect,c=n.isInNodeModules;e.startsWith(i,r)&&((t||(t=[])).push({path:a,isRedirect:s,isInNodeModules:c}),o.delete(a))})),t&&(t.length>1&&t.sort(u),l.push.apply(l,t));var a=e.getDirectoryPath(n);if(a===n)return c=n,"break";c=n=a},m=e.getDirectoryPath(e.toPath(n,a,i));0!==o.size;){var f=p(m);if(m=c,"break"===f)break}if(o.size){var _=e.arrayFrom(o.values());_.length>1&&_.sort(u),l.push.apply(l,_)}return l}function m(n,t,r){for(var a in r)for(var i=0,o=r[a];i<o.length;i++){var s=o[i],c=e.removeFileExtension(e.normalizePath(s)),l=c.indexOf("*");if(-1!==l){var u=c.substr(0,l),d=c.substr(l+1);if(t.length>=u.length+d.length&&e.startsWith(t,u)&&e.endsWith(t,d)||!d&&t===e.removeTrailingDirectorySeparator(u)){var p=t.substr(u.length,t.length-d.length);return a.replace("*",p)}}else if(c===t||c===n)return a}}function f(n,t,r,a,i){var o=n.path,s=n.isRedirect,c=t.getCanonicalFileName,l=t.sourceDirectory;if(r.fileExists&&r.readFile){var u=function(n){var t,r=0,a=0,i=0,o=0;!function(e){e[e.BeforeNodeModules=0]="BeforeNodeModules",e[e.NodeModules=1]="NodeModules",e[e.Scope=2]="Scope",e[e.PackageContent=3]="PackageContent"}(t||(t={}));var s=0,c=0,l=0;for(;c>=0;)switch(s=c,c=n.indexOf("/",s+1),l){case 0:n.indexOf(e.nodeModulesPathPart,s)===s&&(r=s,a=c,l=1);break;case 1:case 2:1===l&&"@"===n.charAt(s+1)?l=2:(i=c,l=3);break;case 3:l=n.indexOf(e.nodeModulesPathPart,s)===s?1:3}return o=s,l>1?{topLevelNodeModulesIndex:r,topLevelPackageNameIndex:a,packageRootIndex:i,fileNameIndex:o}:void 0}(o);if(u){var d=o,p=!1;if(!i)for(var f=u.packageRootIndex,_=void 0;;){var y=x(f),v=y.moduleFileToTry,h=y.packageRootPath;if(h){d=h,p=!0;break}if(_||(_=v),-1===(f=o.indexOf(e.directorySeparator,f+1))){d=L(_);break}}if(!s||p){var b=r.getGlobalTypingsCacheLocation&&r.getGlobalTypingsCacheLocation(),E=c(d.substring(0,u.topLevelNodeModulesIndex));if(e.startsWith(l,E)||b&&e.startsWith(c(b),E)){var T=d.substring(u.topLevelPackageNameIndex+1),S=e.getPackageNameFromTypesPackageName(T);return e.getEmitModuleResolutionKind(a)!==e.ModuleResolutionKind.NodeJs&&S===T?void 0:S}}}}function x(n){var t=o.substring(0,n),i=e.combinePaths(t,"package.json"),s=o;if(r.fileExists(i)){var l=JSON.parse(r.readFile(i)),u=l.typesVersions?e.getPackageJsonTypesVersionsPaths(l.typesVersions):void 0;if(u){var d=o.slice(t.length+1),p=m(e.removeFileExtension(d),g(d,0,a),u.paths);void 0!==p&&(s=e.combinePaths(t,p))}var f=l.typings||l.types||l.main;if(e.isString(f)){var _=e.toPath(f,t,c);if(e.removeFileExtension(_)===e.removeFileExtension(c(s)))return{packageRootPath:t,moduleFileToTry:s}}}return{moduleFileToTry:s}}function L(n){var t=e.removeFileExtension(n);return"/index"!==c(t.substring(u.fileNameIndex))||function(n,t){if(!n.fileExists)return;for(var r=e.getSupportedExtensions({allowJs:!0},[{extension:"node",isMixedContent:!1},{extension:"json",isMixedContent:!1,scriptKind:6}]),a=0,i=r;a<i.length;a++){var o=t+i[a];if(n.fileExists(o))return o}}(r,t.substring(0,u.fileNameIndex))?t:t.substring(0,u.fileNameIndex)}}function _(n,t,r){return e.firstDefined(t,(function(e){var t=y(n,e,r);return v(t)?void 0:t}))}function g(n,t,r){if(e.fileExtensionIs(n,".json"))return n;var a=e.removeFileExtension(n);switch(t){case 0:return e.removeSuffix(a,"/index");case 1:return a;case 2:return a+function(n,t){var r=e.extensionFromPath(n);switch(r){case".ts":case".d.ts":return".js";case".tsx":return 1===t.jsx?".jsx":".js";case".js":case".jsx":case".json":return r;case".tsbuildinfo":return e.Debug.fail("Extension .tsbuildinfo is unsupported:: FileName:: "+n);default:return e.Debug.assertNever(r)}}(n,r);default:return e.Debug.assertNever(t)}}function y(n,t,r){var a=e.getRelativePathToDirectoryOrUrl(t,n,t,r,!1);return e.isRootedDiskPath(a)?void 0:a}function v(n){return e.startsWith(n,"..")}!function(e){e[e.Relative=0]="Relative",e[e.NonRelative=1]="NonRelative",e[e.Auto=2]="Auto"}(t||(t={})),function(e){e[e.Minimal=0]="Minimal",e[e.Index=1]="Index",e[e.JsExtension=2]="JsExtension"}(r||(r={})),n.updateModuleSpecifier=function(n,t,r,a,o){var s=i(n,t,r,a,function(n,t){return{relativePreference:e.isExternalModuleNameRelative(t)?0:1,ending:e.hasJSFileExtension(t)?2:e.getEmitModuleResolutionKind(n)!==e.ModuleResolutionKind.NodeJs||e.endsWith(t,"index")?1:0}}(n,o));if(s!==o)return s},n.getModuleSpecifier=function(e,n,t,r,o,s){return void 0===s&&(s={}),i(e,t,r,o,a(s,e,n))},n.getNodeModulesPackageName=function(n,t,r,a){var i=o(t,a),s=p(t,r,a);return e.firstDefined(s,(function(e){return f(e,i,a,n,!0)}))},n.getModuleSpecifiers=function(n,t,r,i,c){var l=function(n){var t=e.find(n.declarations,(function(n){return e.isNonGlobalAmbientModule(n)&&(!e.isExternalModuleAugmentation(n)||!e.isExternalModuleNameRelative(e.getTextOfIdentifierOrLiteral(n.name)))}));if(t)return t.name.text}(n);if(l)return[l];for(var u,d,m,_=o(r.path,i),g=e.getSourceFileOfNode(n.valueDeclaration||e.getNonAugmentationDeclaration(n)),y=p(r.path,g.originalFileName,i),v=a(c,t,r),h=e.some(y,(function(e){return e.isInNodeModules})),b=0,E=y;b<E.length;b++){var T=E[b],S=f(T,_,i,t);if(u=e.append(u,S),S&&T.isRedirect)return u;if(!S&&!T.isRedirect){var x=s(T.path,_,t,i,v);e.pathIsBareSpecifier(x)?d=e.append(d,x):h&&!T.isInNodeModules||(m=e.append(m,x))}}return(null==d?void 0:d.length)?d:(null==u?void 0:u.length)?u:e.Debug.checkDefined(m)},n.countPathComponents=c,n.forEachFileNameOfModule=d}(e.moduleSpecifiers||(e.moduleSpecifiers={}))}(ts||(ts={})),function(e){var n=e.sys?{getCurrentDirectory:function(){return e.sys.getCurrentDirectory()},getNewLine:function(){return e.sys.newLine},getCanonicalFileName:e.createGetCanonicalFileName(e.sys.useCaseSensitiveFileNames)}:void 0;function t(t,r){var a=t===e.sys?n:{getCurrentDirectory:function(){return t.getCurrentDirectory()},getNewLine:function(){return t.newLine},getCanonicalFileName:e.createGetCanonicalFileName(t.useCaseSensitiveFileNames)};if(!r)return function(n){return t.write(e.formatDiagnostic(n,a))};var i=new Array(1);return function(n){i[0]=n,t.write(e.formatDiagnosticsWithColorAndContext(i,a)+a.getNewLine()),i[0]=void 0}}function r(n,t,r){return!(!n.clearScreen||r.preserveWatchOutput||r.extendedDiagnostics||r.diagnostics||!e.contains(e.screenStartingMessageCodes,t.code))&&(n.clearScreen(),!0)}function a(e){return e.now?e.now().toLocaleTimeString("en-US",{timeZone:"UTC"}):(new Date).toLocaleTimeString()}function i(n,t){return t?function(t,i,o){r(n,t,o);var s="["+e.formatColorAndReset(a(n),e.ForegroundColorEscapeSequences.Grey)+"] ";s+=""+e.flattenDiagnosticMessageText(t.messageText,n.newLine)+(i+i),n.write(s)}:function(t,i,o){var s="";r(n,t,o)||(s+=i),s+=a(n)+" - ",s+=""+e.flattenDiagnosticMessageText(t.messageText,n.newLine)+function(n,t){return e.contains(e.screenStartingMessageCodes,n.code)?t+t:t}(t,i),n.write(s)}}function o(n){return e.countWhere(n,(function(n){return n.category===e.DiagnosticCategory.Error}))}function s(n){return 1===n?e.Diagnostics.Found_1_error_Watching_for_file_changes:e.Diagnostics.Found_0_errors_Watching_for_file_changes}function c(n,t){if(0===n)return"";var r=e.createCompilerDiagnostic(1===n?e.Diagnostics.Found_1_error:e.Diagnostics.Found_0_errors,n);return""+t+e.flattenDiagnosticMessageText(r.messageText,t)+t+t}function l(n,t){(n.getCompilerOptions().listFiles||n.getCompilerOptions().listFilesOnly)&&e.forEach(n.getSourceFiles(),(function(e){t(e.fileName)}))}function u(n,t,r,a,i,s,c,u){var d=!!n.getCompilerOptions().listFilesOnly,p=n.getConfigFileParsingDiagnostics().slice(),m=p.length;e.addRange(p,n.getSyntacticDiagnostics(void 0,s)),p.length===m&&(e.addRange(p,n.getOptionsDiagnostics(s)),d||(e.addRange(p,n.getGlobalDiagnostics(s)),p.length===m&&e.addRange(p,n.getSemanticDiagnostics(void 0,s))));var f=d?{emitSkipped:!0,diagnostics:e.emptyArray}:n.emit(void 0,i,s,c,u),_=f.emittedFiles,g=f.diagnostics;e.addRange(p,g);var y=e.sortAndDeduplicateDiagnostics(p);if(y.forEach(t),r){var v=n.getCurrentDirectory();e.forEach(_,(function(n){var t=e.getNormalizedAbsolutePath(n,v);r("TSFILE: "+t)})),l(n,r)}return a&&a(o(y)),{emitResult:f,diagnostics:y}}function d(n,t,r,a,i,o,s,c){var l=u(n,t,r,a,i,o,s,c),d=l.emitResult,p=l.diagnostics;return d.emitSkipped&&p.length>0?e.ExitStatus.DiagnosticsPresent_OutputsSkipped:p.length>0?e.ExitStatus.DiagnosticsPresent_OutputsGenerated:e.ExitStatus.Success}function p(n,t){return void 0===n&&(n=e.sys),{onWatchStatusChange:t||i(n),watchFile:e.maybeBind(n,n.watchFile)||e.returnNoopFileWatcher,watchDirectory:e.maybeBind(n,n.watchDirectory)||e.returnNoopFileWatcher,setTimeout:e.maybeBind(n,n.setTimeout)||e.noop,clearTimeout:e.maybeBind(n,n.clearTimeout)||e.noop}}function m(n,t){var r=e.memoize((function(){return e.getDirectoryPath(e.normalizePath(n.getExecutingFilePath()))}));return{useCaseSensitiveFileNames:function(){return n.useCaseSensitiveFileNames},getNewLine:function(){return n.newLine},getCurrentDirectory:e.memoize((function(){return n.getCurrentDirectory()})),getDefaultLibLocation:r,getDefaultLibFileName:function(n){return e.combinePaths(r(),e.getDefaultLibFileName(n))},fileExists:function(e){return n.fileExists(e)},readFile:function(e,t){return n.readFile(e,t)},directoryExists:function(e){return n.directoryExists(e)},getDirectories:function(e){return n.getDirectories(e)},readDirectory:function(e,t,r,a,i){return n.readDirectory(e,t,r,a,i)},realpath:e.maybeBind(n,n.realpath),getEnvironmentVariable:e.maybeBind(n,n.getEnvironmentVariable),trace:function(e){return n.write(e+n.newLine)},createDirectory:function(e){return n.createDirectory(e)},writeFile:function(e,t,r){return n.writeFile(e,t,r)},createHash:e.maybeBind(n,n.createHash),createProgram:t||e.createEmitAndSemanticDiagnosticsBuilderProgram}}function f(n,t,r,a){void 0===n&&(n=e.sys);var i=function(e){return n.write(e+n.newLine)},o=m(n,t);return e.copyProperties(o,p(n,a)),o.afterProgramCreate=function(t){var a=t.getCompilerOptions(),c=e.getNewLineCharacter(a,(function(){return n.newLine}));u(t,r,i,(function(n){return o.onWatchStatusChange(e.createCompilerDiagnostic(s(n),n),c,a,n)}))},o}function _(n,t,r){t(r),n.exit(e.ExitStatus.DiagnosticsPresent_OutputsSkipped)}e.createDiagnosticReporter=t,e.screenStartingMessageCodes=[e.Diagnostics.Starting_compilation_in_watch_mode.code,e.Diagnostics.File_change_detected_Starting_incremental_compilation.code],e.getLocaleTimeString=a,e.createWatchStatusReporter=i,e.parseConfigFileWithSystem=function(n,t,r,a,i){var o=a;o.onUnRecoverableConfigFileDiagnostic=function(e){return _(a,i,e)};var s=e.getParsedCommandLineOfConfigFile(n,t,o,void 0,r);return o.onUnRecoverableConfigFileDiagnostic=void 0,s},e.getErrorCountForSummary=o,e.getWatchErrorSummaryDiagnosticMessage=s,e.getErrorSummaryText=c,e.listFiles=l,e.emitFilesAndReportErrors=u,e.emitFilesAndReportErrorsAndGetExitStatus=d,e.noopFileWatcher={close:e.noop},e.returnNoopFileWatcher=function(){return e.noopFileWatcher},e.createWatchHost=p,e.WatchType={ConfigFile:"Config file",SourceFile:"Source file",MissingFile:"Missing file",WildcardDirectory:"Wild card directory",FailedLookupLocations:"Failed Lookup Locations",TypeRoots:"Type roots"},e.createWatchFactory=function(n,t){var r=n.trace?t.extendedDiagnostics?e.WatchLogLevel.Verbose:t.diagnostics?e.WatchLogLevel.TriggerOnly:e.WatchLogLevel.None:e.WatchLogLevel.None,a=r!==e.WatchLogLevel.None?function(e){return n.trace(e)}:e.noop,i=e.getWatchFactory(r,a);return i.writeLog=a,i},e.createCompilerHostFromProgramHost=function(n,t,r){void 0===r&&(r=n);var a=n.useCaseSensitiveFileNames(),i=e.memoize((function(){return n.getNewLine()}));return{getSourceFile:function(r,a,i){var o;try{e.performance.mark("beforeIORead"),o=n.readFile(r,t().charset),e.performance.mark("afterIORead"),e.performance.measure("I/O Read","beforeIORead","afterIORead")}catch(e){i&&i(e.message),o=""}return void 0!==o?e.createSourceFile(r,o,a):void 0},getDefaultLibLocation:e.maybeBind(n,n.getDefaultLibLocation),getDefaultLibFileName:function(e){return n.getDefaultLibFileName(e)},writeFile:function(t,r,a,i){try{e.performance.mark("beforeIOWrite"),e.writeFileEnsuringDirectories(t,r,a,(function(e,t,r){return n.writeFile(e,t,r)}),(function(e){return n.createDirectory(e)}),(function(e){return n.directoryExists(e)})),e.performance.mark("afterIOWrite"),e.performance.measure("I/O Write","beforeIOWrite","afterIOWrite")}catch(e){i&&i(e.message)}},getCurrentDirectory:e.memoize((function(){return n.getCurrentDirectory()})),useCaseSensitiveFileNames:function(){return a},getCanonicalFileName:e.createGetCanonicalFileName(a),getNewLine:function(){return e.getNewLineCharacter(t(),i)},fileExists:function(e){return n.fileExists(e)},readFile:function(e){return n.readFile(e)},trace:e.maybeBind(n,n.trace),directoryExists:e.maybeBind(r,r.directoryExists),getDirectories:e.maybeBind(r,r.getDirectories),realpath:e.maybeBind(n,n.realpath),getEnvironmentVariable:e.maybeBind(n,n.getEnvironmentVariable)||function(){return""},createHash:e.maybeBind(n,n.createHash),readDirectory:e.maybeBind(n,n.readDirectory)}},e.setGetSourceFileAsHashVersioned=function(n,t){var r=n.getSourceFile,a=e.maybeBind(t,t.createHash)||e.generateDjb2Hash;n.getSourceFile=function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var i=r.call.apply(r,__spreadArrays([n],e));return i&&(i.version=a(i.text)),i}},e.createProgramHost=m,e.createWatchCompilerHostOfConfigFile=function(e){var n=e.configFileName,r=e.optionsToExtend,a=e.watchOptionsToExtend,i=e.extraFileExtensions,o=e.system,s=e.createProgram,c=e.reportDiagnostic,l=e.reportWatchStatus,u=c||t(o),d=f(o,s,u,l);return d.onUnRecoverableConfigFileDiagnostic=function(e){return _(o,u,e)},d.configFileName=n,d.optionsToExtend=r,d.watchOptionsToExtend=a,d.extraFileExtensions=i,d},e.createWatchCompilerHostOfFilesAndCompilerOptions=function(e){var n=e.rootFiles,r=e.options,a=e.watchOptions,i=e.projectReferences,o=e.system,s=e.createProgram,c=e.reportDiagnostic,l=e.reportWatchStatus,u=f(o,s,c||t(o),l);return u.rootFiles=n,u.options=r,u.watchOptions=a,u.projectReferences=i,u},e.performIncrementalCompilation=function(n){var r=n.system||e.sys,a=n.host||(n.host=e.createIncrementalCompilerHost(n.options,r)),i=e.createIncrementalProgram(n),o=d(i,n.reportDiagnostic||t(r),(function(e){return a.trace&&a.trace(e)}),n.reportErrorSummary||n.options.pretty?function(e){return r.write(c(e,r.newLine))}:void 0);return n.afterProgramEmitAndDiagnostics&&n.afterProgramEmitAndDiagnostics(i),o}}(ts||(ts={})),function(e){function n(n,t){if(!e.outFile(n)){var r=e.getTsBuildInfoEmitOutputFilePath(n);if(r){var a=t.readFile(r);if(a){var i=e.getBuildInfo(a);if(i.version===e.version&&i.program)return e.createBuildProgramUsingProgramBuildInfo(i.program,r,t)}}}}function t(n,t){void 0===t&&(t=e.sys);var r=e.createCompilerHostWorker(n,void 0,t);return r.createHash=e.maybeBind(t,t.createHash),e.setGetSourceFileAsHashVersioned(r,t),e.changeCompilerHostLikeToUseCache(r,(function(n){return e.toPath(n,r.getCurrentDirectory(),r.getCanonicalFileName)})),r}e.readBuilderProgram=n,e.createIncrementalCompilerHost=t,e.createIncrementalProgram=function(r){var a=r.rootNames,i=r.options,o=r.configFileParsingDiagnostics,s=r.projectReferences,c=r.host,l=r.createProgram;return c=c||t(i),(l=l||e.createEmitAndSemanticDiagnosticsBuilderProgram)(a,i,c,n(i,c),o,s)},e.createWatchCompilerHost=function(n,t,r,a,i,o,s,c){return e.isArray(n)?e.createWatchCompilerHostOfFilesAndCompilerOptions({rootFiles:n,options:t,watchOptions:c,projectReferences:s,system:r,createProgram:a,reportDiagnostic:i,reportWatchStatus:o}):e.createWatchCompilerHostOfConfigFile({configFileName:n,optionsToExtend:t,watchOptionsToExtend:s,extraFileExtensions:c,system:r,createProgram:a,reportDiagnostic:i,reportWatchStatus:o})},e.createWatchProgram=function(t){var r,a,i,o,s,c,l,u,d,p=new e.Map,m=!1,f=t.useCaseSensitiveFileNames(),_=t.getCurrentDirectory(),g=t.configFileName,y=t.optionsToExtend,v=void 0===y?{}:y,h=t.watchOptionsToExtend,b=t.extraFileExtensions,E=t.createProgram,T=t.rootFiles,S=t.options,x=t.watchOptions,L=t.projectReferences,A=!1,D=!1,C=void 0===g?void 0:e.createCachedDirectoryStructureHost(t,_,f),I=C||t,N=e.parseConfigHostFromCompilerHostLike(t,I),R=q();g&&t.configFileParsingResult&&(ce(t.configFileParsingResult),R=q()),ee(e.Diagnostics.Starting_compilation_in_watch_mode),g&&!t.configFileParsingResult&&(R=e.getNewLineCharacter(v,(function(){return t.getNewLine()})),e.Debug.assert(!T),se(),R=q());var k,O=e.createWatchFactory(t,S),P=O.watchFile,M=O.watchFilePath,w=O.watchDirectory,F=O.writeLog,G=e.createGetCanonicalFileName(f);F("Current directory: "+_+" CaseSensitiveFileNames: "+f),g&&(k=P(t,g,(function(){e.Debug.assert(!!g),a=e.ConfigFileProgramReloadLevel.Full,ae()}),e.PollingInterval.High,x,e.WatchType.ConfigFile));var B=e.createCompilerHostFromProgramHost(t,(function(){return S}),I);e.setGetSourceFileAsHashVersioned(B,t);var U=B.getSourceFile;B.getSourceFile=function(e){for(var n=[],t=1;t<arguments.length;t++)n[t-1]=arguments[t];return Q.apply(void 0,__spreadArrays([e,J(e)],n))},B.getSourceFileByPath=Q,B.getNewLine=function(){return R},B.fileExists=Y,B.onReleaseOldSourceFile=function(e,n,t){var r=p.get(e.resolvedPath);void 0!==r&&(X(r)?(l||(l=[])).push(e.path):r.sourceFile===e&&(r.fileWatcher&&r.fileWatcher.close(),p.delete(e.resolvedPath),t||V.removeResolutionsOfFile(e.path)))},B.toPath=J,B.getCompilationSettings=function(){return S},B.useSourceOfProjectReferenceRedirect=e.maybeBind(t,t.useSourceOfProjectReferenceRedirect),B.watchDirectoryOfFailedLookupLocation=function(n,r,a){return w(t,n,r,a,x,e.WatchType.FailedLookupLocations)},B.watchTypeRootsDirectory=function(n,r,a){return w(t,n,r,a,x,e.WatchType.TypeRoots)},B.getCachedDirectoryStructureHost=function(){return C},B.scheduleInvalidateResolutionsOfFailedLookupLocations=function(){if(!t.setTimeout||!t.clearTimeout)return V.invalidateResolutionsOfFailedLookupLocations();var e=te();F("Scheduling invalidateFailedLookup"+(e?", Cancelled earlier one":"")),c=t.setTimeout(re,250)},B.onInvalidatedResolution=ae,B.onChangedAutomaticTypeDirectiveNames=ae,B.fileIsOpen=e.returnFalse,B.getCurrentProgram=W,B.writeLog=F;var V=e.createResolutionCache(B,g?e.getDirectoryPath(e.getNormalizedAbsolutePath(g,_)):_,!1);B.resolveModuleNames=t.resolveModuleNames?function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return t.resolveModuleNames.apply(t,e)}:function(e,n,t,r){return V.resolveModuleNames(e,n,t,r)},B.resolveTypeReferenceDirectives=t.resolveTypeReferenceDirectives?function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return t.resolveTypeReferenceDirectives.apply(t,e)}:function(e,n,t){return V.resolveTypeReferenceDirectives(e,n,t)};var j=!!t.resolveModuleNames||!!t.resolveTypeReferenceDirectives;return r=n(S,B),z(),me(),g?{getCurrentProgram:K,getProgram:oe,close:H}:{getCurrentProgram:K,getProgram:oe,updateRootFileNames:function(n){e.Debug.assert(!g,"Cannot update root file names with config file watch mode"),T=n,ae()},close:H};function H(){te(),V.clear(),e.clearMap(p,(function(e){e&&e.fileWatcher&&(e.fileWatcher.close(),e.fileWatcher=void 0)})),k&&(k.close(),k=void 0),o&&(e.clearMap(o,e.closeFileWatcherOf),o=void 0),i&&(e.clearMap(i,e.closeFileWatcher),i=void 0)}function K(){return r}function W(){return r&&r.getProgramOrUndefined()}function z(){F("Synchronizing program"),te();var n=K();m&&(R=q(),n&&e.changesAffectModuleResolution(n.getCompilerOptions(),S)&&V.clear());var a=V.createHasInvalidatedResolution(j);return e.isProgramUptoDate(W(),T,S,$,Y,a,ne,L)?D&&(r=E(void 0,void 0,B,r,d,L),D=!1):function(n){F("CreatingProgramWith::"),F("  roots: "+JSON.stringify(T)),F("  options: "+JSON.stringify(S));var t=m||!W();m=!1,D=!1,V.startCachingPerDirectoryResolution(),B.hasInvalidatedResolution=n,B.hasChangedAutomaticTypeDirectiveNames=ne,r=E(T,S,B,r,d,L),V.finishCachingPerDirectoryResolution(),e.updateMissingFilePathsWatch(r.getProgram(),i||(i=new e.Map),de),t&&V.updateTypeRootsWatch();if(l){for(var a=0,o=l;a<o.length;a++){var s=o[a];i.has(s)||p.delete(s)}l=void 0}}(a),t.afterProgramCreate&&n!==r&&t.afterProgramCreate(r),r}function q(){return e.getNewLineCharacter(S||v,(function(){return t.getNewLine()}))}function J(n){return e.toPath(n,_,G)}function X(e){return"boolean"==typeof e}function Y(e){var n=J(e);return!X(p.get(n))&&I.fileExists(e)}function Q(n,r,a,i,o){var s=p.get(r);if(!X(s)){if(void 0===s||o||function(e){return"boolean"==typeof e.version}(s)){var c=U(n,a,i);if(s)c?(s.sourceFile=c,s.version=c.version,s.fileWatcher||(s.fileWatcher=M(t,n,le,e.PollingInterval.Low,x,r,e.WatchType.SourceFile))):(s.fileWatcher&&s.fileWatcher.close(),p.set(r,!1));else if(c){var l=M(t,n,le,e.PollingInterval.Low,x,r,e.WatchType.SourceFile);p.set(r,{sourceFile:c,version:c.version,fileWatcher:l})}else p.set(r,!1);return c}return s.sourceFile}}function Z(e){var n=p.get(e);void 0!==n&&(X(n)?p.set(e,{version:!1}):n.version=!1)}function $(e){var n=p.get(e);return n&&n.version?n.version:void 0}function ee(n){t.onWatchStatusChange&&t.onWatchStatusChange(e.createCompilerDiagnostic(n),R,S||v)}function ne(){return V.hasChangedAutomaticTypeDirectiveNames()}function te(){return!!c&&(t.clearTimeout(c),c=void 0,!0)}function re(){c=void 0,V.invalidateResolutionsOfFailedLookupLocations()&&ae()}function ae(){t.setTimeout&&t.clearTimeout&&(s&&t.clearTimeout(s),F("Scheduling update"),s=t.setTimeout(ie,250))}function ie(){s=void 0,ee(e.Diagnostics.File_change_detected_Starting_incremental_compilation),oe()}function oe(){switch(a){case e.ConfigFileProgramReloadLevel.Partial:e.perfLogger.logStartUpdateProgram("PartialConfigReload"),function(){F("Reloading new file names and options");var n=e.getFileNamesFromConfigSpecs(u,e.getNormalizedAbsolutePath(e.getDirectoryPath(g),_),S,N,b);e.updateErrorForNoInputFiles(n,e.getNormalizedAbsolutePath(g,_),u,d,A)&&(D=!0);T=n.fileNames,z()}();break;case e.ConfigFileProgramReloadLevel.Full:e.perfLogger.logStartUpdateProgram("FullConfigReload"),function(){F("Reloading config file: "+g),a=e.ConfigFileProgramReloadLevel.None,C&&C.clearCache();se(),m=!0,z(),me()}();break;default:e.perfLogger.logStartUpdateProgram("SynchronizeProgram"),z()}return e.perfLogger.logStopUpdateProgram("Done"),K()}function se(){ce(e.getParsedCommandLineOfConfigFile(g,v,N,void 0,h,b))}function ce(n){T=n.fileNames,S=n.options,x=n.watchOptions,u=n.configFileSpecs,L=n.projectReferences,d=e.getConfigFileParsingDiagnostics(n).slice(),A=e.canJsonReportNoInputFiles(n.raw),D=!0}function le(n,t,r){ue(n,r,t),t===e.FileWatcherEventKind.Deleted&&p.has(r)&&V.invalidateResolutionOfFile(r),V.removeResolutionsFromProjectReferenceRedirects(r),Z(r),ae()}function ue(e,n,t){C&&C.addOrDeleteFile(e,n,t)}function de(n){return M(t,n,pe,e.PollingInterval.Medium,x,n,e.WatchType.MissingFile)}function pe(n,t,r){ue(n,r,t),t===e.FileWatcherEventKind.Created&&i.has(r)&&(i.get(r).close(),i.delete(r),Z(r),ae())}function me(){u?e.updateWatchingWildcardDirectories(o||(o=new e.Map),new e.Map(e.getEntries(u.wildcardDirectories)),fe):o&&e.clearMap(o,e.closeFileWatcherOf)}function fe(n,r){return w(t,n,(function(t){e.Debug.assert(!!g);var r=J(t);C&&C.addOrDeleteFileOrDirectory(t,r),Z(r),e.isIgnoredFileFromWildCardWatching({watchedDirPath:J(n),fileOrDirectory:t,fileOrDirectoryPath:r,configFileName:g,configFileSpecs:u,extraFileExtensions:b,options:S,program:K(),currentDirectory:_,useCaseSensitiveFileNames:f,writeLog:F})||a!==e.ConfigFileProgramReloadLevel.Full&&(a=e.ConfigFileProgramReloadLevel.Partial,ae())}),r,x,e.WatchType.WildcardDirectory)}}}(ts||(ts={})),function(e){!function(e){e[e.Unbuildable=0]="Unbuildable",e[e.UpToDate=1]="UpToDate",e[e.UpToDateWithUpstreamTypes=2]="UpToDateWithUpstreamTypes",e[e.OutOfDateWithPrepend=3]="OutOfDateWithPrepend",e[e.OutputMissing=4]="OutputMissing",e[e.OutOfDateWithSelf=5]="OutOfDateWithSelf",e[e.OutOfDateWithUpstream=6]="OutOfDateWithUpstream",e[e.UpstreamOutOfDate=7]="UpstreamOutOfDate",e[e.UpstreamBlocked=8]="UpstreamBlocked",e[e.ComputingUpstream=9]="ComputingUpstream",e[e.TsVersionOutputOfDate=10]="TsVersionOutputOfDate",e[e.ContainerOnly=11]="ContainerOnly"}(e.UpToDateStatusType||(e.UpToDateStatusType={})),e.resolveConfigFileProjectName=function(n){return e.fileExtensionIs(n,".json")?n:e.combinePaths(n,"tsconfig.json")}}(ts||(ts={})),function(e){var n,t,r,a=new Date(-864e13),i=new Date(864e13);function o(n,t){return function(e,n,t){var r,a=e.get(n);return a||(r=t(),e.set(n,r)),a||r}(n,t,(function(){return new e.Map}))}function s(e,n){return n>e?n:e}function c(n){return e.fileExtensionIs(n,".d.ts")}function l(e){return!!e&&!!e.buildOrder}function u(e){return l(e)?e.buildOrder:e}function d(n,t){return function(r){var a=t?"["+e.formatColorAndReset(e.getLocaleTimeString(n),e.ForegroundColorEscapeSequences.Grey)+"] ":e.getLocaleTimeString(n)+" - ";a+=""+e.flattenDiagnosticMessageText(r.messageText,n.newLine)+(n.newLine+n.newLine),n.write(a)}}function p(n,t,r,a){var i=e.createProgramHost(n,t);return i.getModifiedTime=n.getModifiedTime?function(e){return n.getModifiedTime(e)}:e.returnUndefined,i.setModifiedTime=n.setModifiedTime?function(e,t){return n.setModifiedTime(e,t)}:e.noop,i.deleteFile=n.deleteFile?function(e){return n.deleteFile(e)}:e.noop,i.reportDiagnostic=r||e.createDiagnosticReporter(n),i.reportSolutionBuilderStatus=a||d(n),i.now=e.maybeBind(n,n.now),i}function m(n,t,r,a,i){var o,s,c=t,l=t,u=c.getCurrentDirectory(),d=e.createGetCanonicalFileName(c.useCaseSensitiveFileNames()),p=(o=a,s={},e.commonOptionsWithBuild.forEach((function(n){e.hasProperty(o,n.name)&&(s[n.name]=o[n.name])})),s),m=e.createCompilerHostFromProgramHost(c,(function(){return S.projectCompilerOptions}));e.setGetSourceFileAsHashVersioned(m,c),m.getParsedCommandLine=function(e){return y(S,e,_(S,e))},m.resolveModuleNames=e.maybeBind(c,c.resolveModuleNames),m.resolveTypeReferenceDirectives=e.maybeBind(c,c.resolveTypeReferenceDirectives);var f=m.resolveModuleNames?void 0:e.createModuleResolutionCache(u,d);if(!m.resolveModuleNames){var g=function(n,t,r){return e.resolveModuleName(n,t,S.projectCompilerOptions,m,f,r).resolvedModule};m.resolveModuleNames=function(n,t,r,a){return e.loadWithLocalCache(e.Debug.checkEachDefined(n),t,a,g)}}var v=e.createWatchFactory(l,a),h=v.watchFile,b=v.watchFilePath,E=v.watchDirectory,T=v.writeLog,S={host:c,hostWithWatch:l,currentDirectory:u,getCanonicalFileName:d,parseConfigFileHost:e.parseConfigHostFromCompilerHostLike(c),writeFileName:c.trace?function(e){return c.trace(e)}:void 0,options:a,baseCompilerOptions:p,rootNames:r,baseWatchOptions:i,resolvedConfigFilePaths:new e.Map,configFileCache:new e.Map,projectStatus:new e.Map,buildInfoChecked:new e.Map,extendedConfigCache:new e.Map,builderPrograms:new e.Map,diagnostics:new e.Map,projectPendingBuild:new e.Map,projectErrorsReported:new e.Map,compilerHost:m,moduleResolutionCache:f,buildOrder:void 0,readFileWithCache:function(e){return c.readFile(e)},projectCompilerOptions:p,cache:void 0,allProjectBuildPending:!0,needsSummary:!0,watchAllProjectsPending:n,currentInvalidatedProject:void 0,watch:n,allWatchedWildcardDirectories:new e.Map,allWatchedInputFiles:new e.Map,allWatchedConfigFiles:new e.Map,timerToBuildInvalidatedProject:void 0,reportFileChangeDetected:!1,watchFile:h,watchFilePath:b,watchDirectory:E,writeLog:T};return S}function f(n,t){return e.toPath(t,n.currentDirectory,n.getCanonicalFileName)}function _(e,n){var t=e.resolvedConfigFilePaths,r=t.get(n);if(void 0!==r)return r;var a=f(e,n);return t.set(n,a),a}function g(e){return!!e.options}function y(n,t,r){var a,i=n.configFileCache,o=i.get(r);if(o)return g(o)?o:void 0;var s,c=n.parseConfigFileHost,l=n.baseCompilerOptions,u=n.baseWatchOptions,d=n.extendedConfigCache,p=n.host;return p.getParsedCommandLine?(s=p.getParsedCommandLine(t))||(a=e.createCompilerDiagnostic(e.Diagnostics.File_0_not_found,t)):(c.onUnRecoverableConfigFileDiagnostic=function(e){return a=e},s=e.getParsedCommandLineOfConfigFile(t,l,c,d,u),c.onUnRecoverableConfigFileDiagnostic=e.noop),i.set(r,s||a),s}function v(n,t){return e.resolveConfigFileProjectName(e.resolvePath(n.currentDirectory,t))}function h(n,t){for(var r,a,i=new e.Map,o=new e.Map,s=[],c=0,l=t;c<l.length;c++){u(l[c])}return a?{buildOrder:r||e.emptyArray,circularDiagnostics:a}:r||e.emptyArray;function u(t,c){var l=_(n,t);if(!o.has(l))if(i.has(l))c||(a||(a=[])).push(e.createCompilerDiagnostic(e.Diagnostics.Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0,s.join("\r\n")));else{i.set(l,!0),s.push(t);var d=y(n,t,l);if(d&&d.projectReferences)for(var p=0,m=d.projectReferences;p<m.length;p++){var f=m[p];u(v(n,f.path),c||f.circular)}s.pop(),o.set(l,!0),(r||(r=[])).push(t)}}}function b(n){return n.buildOrder||function(n){var t=h(n,n.rootNames.map((function(e){return v(n,e)})));n.resolvedConfigFilePaths.clear();var r=new e.Map(u(t).map((function(e){return[_(n,e),!0]}))),a={onDeleteValue:e.noop};e.mutateMapSkippingNewValues(n.configFileCache,r,a),e.mutateMapSkippingNewValues(n.projectStatus,r,a),e.mutateMapSkippingNewValues(n.buildInfoChecked,r,a),e.mutateMapSkippingNewValues(n.builderPrograms,r,a),e.mutateMapSkippingNewValues(n.diagnostics,r,a),e.mutateMapSkippingNewValues(n.projectPendingBuild,r,a),e.mutateMapSkippingNewValues(n.projectErrorsReported,r,a),n.watch&&(e.mutateMapSkippingNewValues(n.allWatchedConfigFiles,r,{onDeleteValue:e.closeFileWatcher}),e.mutateMapSkippingNewValues(n.allWatchedWildcardDirectories,r,{onDeleteValue:function(n){return n.forEach(e.closeFileWatcherOf)}}),e.mutateMapSkippingNewValues(n.allWatchedInputFiles,r,{onDeleteValue:function(n){return n.forEach(e.closeFileWatcher)}}));return n.buildOrder=t}(n)}function E(n,t,r){var a=t&&v(n,t),i=b(n);if(l(i))return i;if(a){var o=_(n,a);if(-1===e.findIndex(i,(function(e){return _(n,e)===o})))return}var s=a?h(n,[a]):i;return e.Debug.assert(!l(s)),e.Debug.assert(!r||void 0!==a),e.Debug.assert(!r||s[s.length-1]===a),r?s.slice(0,s.length-1):s}function T(n){n.cache&&S(n);var t=n.compilerHost,r=n.host,a=n.readFileWithCache,i=t.getSourceFile,o=e.changeCompilerHostLikeToUseCache(r,(function(e){return f(n,e)}),(function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return i.call.apply(i,__spreadArrays([t],e))})),s=o.originalReadFile,c=o.originalFileExists,l=o.originalDirectoryExists,u=o.originalCreateDirectory,d=o.originalWriteFile,p=o.getSourceFileWithCache,m=o.readFileWithCache;n.readFileWithCache=m,t.getSourceFile=p,n.cache={originalReadFile:s,originalFileExists:c,originalDirectoryExists:l,originalCreateDirectory:u,originalWriteFile:d,originalReadFileWithCache:a,originalGetSourceFile:i}}function S(e){if(e.cache){var n=e.cache,t=e.host,r=e.compilerHost,a=e.extendedConfigCache,i=e.moduleResolutionCache;t.readFile=n.originalReadFile,t.fileExists=n.originalFileExists,t.directoryExists=n.originalDirectoryExists,t.createDirectory=n.originalCreateDirectory,t.writeFile=n.originalWriteFile,r.getSourceFile=n.originalGetSourceFile,e.readFileWithCache=n.originalReadFileWithCache,a.clear(),i&&(i.directoryToModuleNameMap.clear(),i.moduleNameToDirectoryMap.clear()),e.cache=void 0}}function x(e,n){e.projectStatus.delete(n),e.diagnostics.delete(n)}function L(e,n,t){var r=e.projectPendingBuild,a=r.get(n);(void 0===a||a<t)&&r.set(n,t)}function A(n,t){n.allProjectBuildPending&&(n.allProjectBuildPending=!1,n.options.watch&&Z(n,e.Diagnostics.Starting_compilation_in_watch_mode),T(n),u(b(n)).forEach((function(t){return n.projectPendingBuild.set(_(n,t),e.ConfigFileProgramReloadLevel.None)})),t&&t.throwIfCancellationRequested())}function D(n,t){return n.projectPendingBuild.delete(t),n.currentInvalidatedProject=void 0,n.diagnostics.has(t)?e.ExitStatus.DiagnosticsPresent_OutputsSkipped:e.ExitStatus.Success}function C(e,n,r,a,i){var o=!0;return{kind:t.UpdateOutputFileStamps,project:n,projectPath:r,buildOrder:i,getCompilerOptions:function(){return a.options},getCurrentDirectory:function(){return e.currentDirectory},updateOutputFileStatmps:function(){G(e,a,r),o=!1},done:function(){return o&&G(e,a,r),D(e,r)}}}function I(o,l,u,d,p,m,g){var h,b,E,T=o===t.Build?r.CreateProgram:r.EmitBundle;return o===t.Build?{kind:o,project:u,projectPath:d,buildOrder:g,getCompilerOptions:function(){return m.options},getCurrentDirectory:function(){return l.currentDirectory},getBuilderProgram:function(){return x(e.identity)},getProgram:function(){return x((function(e){return e.getProgramOrUndefined()}))},getSourceFile:function(e){return x((function(n){return n.getSourceFile(e)}))},getSourceFiles:function(){return L((function(e){return e.getSourceFiles()}))},getOptionsDiagnostics:function(e){return L((function(n){return n.getOptionsDiagnostics(e)}))},getGlobalDiagnostics:function(e){return L((function(n){return n.getGlobalDiagnostics(e)}))},getConfigFileParsingDiagnostics:function(){return L((function(e){return e.getConfigFileParsingDiagnostics()}))},getSyntacticDiagnostics:function(e,n){return L((function(t){return t.getSyntacticDiagnostics(e,n)}))},getAllDependencies:function(e){return L((function(n){return n.getAllDependencies(e)}))},getSemanticDiagnostics:function(e,n){return L((function(t){return t.getSemanticDiagnostics(e,n)}))},getSemanticDiagnosticsOfNextAffectedFile:function(e,n){return x((function(t){return t.getSemanticDiagnosticsOfNextAffectedFile&&t.getSemanticDiagnosticsOfNextAffectedFile(e,n)}))},emit:function(e,n,t,a,i){return e||a?x((function(r){return r.emit(e,n,t,a,i)})):(V(r.SemanticDiagnostics,t),T===r.EmitBuildInfo?w(n,t):T===r.Emit?M(n,t,i):void 0)},done:S}:{kind:o,project:u,projectPath:d,buildOrder:g,getCompilerOptions:function(){return m.options},getCurrentDirectory:function(){return l.currentDirectory},emit:function(e,n){return T!==r.EmitBundle?E:U(e,n)},done:S};function S(e,n,t){return V(r.Done,e,n,t),D(l,d)}function x(e){return V(r.CreateProgram),h&&e(h)}function L(n){return x(n)||e.emptyArray}function A(){if(e.Debug.assert(void 0===h),l.options.dry)return Q(l,e.Diagnostics.A_non_dry_build_would_build_project_0,u),b=n.Success,void(T=r.QueueReferencingProjects);if(l.options.verbose&&Q(l,e.Diagnostics.Building_project_0,u),0===m.fileNames.length)return ee(l,d,e.getConfigFileParsingDiagnostics(m)),b=n.None,void(T=r.QueueReferencingProjects);var t=l.host,a=l.compilerHost;l.projectCompilerOptions=m.options,function(n,t,r){if(!n.moduleResolutionCache)return;var a=n.moduleResolutionCache,i=f(n,t);if(0===a.directoryToModuleNameMap.redirectsMap.size)e.Debug.assert(0===a.moduleNameToDirectoryMap.redirectsMap.size),a.directoryToModuleNameMap.redirectsMap.set(i,a.directoryToModuleNameMap.ownMap),a.moduleNameToDirectoryMap.redirectsMap.set(i,a.moduleNameToDirectoryMap.ownMap);else{e.Debug.assert(a.moduleNameToDirectoryMap.redirectsMap.size>0);var o={sourceFile:r.options.configFile,commandLine:r};a.directoryToModuleNameMap.setOwnMap(a.directoryToModuleNameMap.getOrCreateMapOfCacheRedirects(o)),a.moduleNameToDirectoryMap.setOwnMap(a.moduleNameToDirectoryMap.getOrCreateMapOfCacheRedirects(o))}a.directoryToModuleNameMap.setOwnOptions(r.options),a.moduleNameToDirectoryMap.setOwnOptions(r.options)}(l,u,m),h=t.createProgram(m.fileNames,m.options,a,function(n,t,r){var a=n.options,i=n.builderPrograms,o=n.compilerHost;if(a.force)return;var s=i.get(t);return s||e.readBuilderProgram(r.options,o)}(l,d,m),e.getConfigFileParsingDiagnostics(m),m.projectReferences),l.watch&&l.builderPrograms.set(d,h),T++}function C(e,n,t){var r;e.length?(r=P(l,d,h,m,e,n,t),b=r.buildResult,T=r.step):T++}function N(t){e.Debug.assertIsDefined(h),C(__spreadArrays(h.getConfigFileParsingDiagnostics(),h.getOptionsDiagnostics(t),h.getGlobalDiagnostics(t),h.getSyntacticDiagnostics(void 0,t)),n.SyntaxErrors,"Syntactic")}function R(t){C(e.Debug.checkDefined(h).getSemanticDiagnostics(void 0,t),n.TypeErrors,"Semantic")}function M(t,i,o){var u,p;e.Debug.assertIsDefined(h),e.Debug.assert(T===r.Emit),h.backupState();var _=[],g=e.emitFilesAndReportErrors(h,(function(e){return(p||(p=[])).push(e)}),void 0,void 0,(function(e,n,t){return _.push({name:e,text:n,writeByteOrderMark:t})}),i,!1,o).emitResult;if(p)return h.restoreState(),u=P(l,d,h,m,p,n.DeclarationEmitErrors,"Declaration file"),b=u.buildResult,T=u.step,{emitSkipped:!0,diagnostics:g.diagnostics};var y=l.host,v=l.compilerHost,E=n.DeclarationOutputUnchanged,S=a,x=!1,L=e.createDiagnosticCollection(),A=new e.Map;return _.forEach((function(r){var a,i=r.name,o=r.text,u=r.writeByteOrderMark;!x&&c(i)&&(y.fileExists(i)&&l.readFileWithCache(i)===o?a=y.getModifiedTime(i):(E&=~n.DeclarationOutputUnchanged,x=!0)),A.set(f(l,i),i),e.writeFile(t?{writeFile:t}:v,L,i,o,u),void 0!==a&&(S=s(a,S))})),G(L,A,S,x,_.length?_[0].name:e.getFirstProjectOutput(m,!y.useCaseSensitiveFileNames()),E),g}function w(t,a){e.Debug.assertIsDefined(h),e.Debug.assert(T===r.EmitBuildInfo);var i=h.emitBuildInfo(t,a);return i.diagnostics.length&&($(l,i.diagnostics),l.diagnostics.set(d,__spreadArrays(l.diagnostics.get(d),i.diagnostics)),b=n.EmitErrors&b),i.emittedFiles&&l.writeFileName&&i.emittedFiles.forEach((function(e){return k(l,m,e)})),O(l,h,m),T=r.QueueReferencingProjects,i}function G(t,a,o,s,c,u){var p,f=t.getDiagnostics();if(f.length)return p=P(l,d,h,m,f,n.EmitErrors,"Emit"),b=p.buildResult,T=p.step,f;l.writeFileName&&a.forEach((function(e){return k(l,m,e)}));var _=F(l,m,o,e.Diagnostics.Updating_unchanged_output_timestamps_of_project_0,a);return l.diagnostics.delete(d),l.projectStatus.set(d,{type:e.UpToDateStatusType.UpToDate,newestDeclarationFileContentChangedTime:s?i:_,oldestOutputFileName:c}),O(l,h,m),T=r.QueueReferencingProjects,b=u,f}function U(i,s){if(e.Debug.assert(o===t.UpdateBundle),l.options.dry)return Q(l,e.Diagnostics.A_non_dry_build_would_update_output_of_project_0,u),b=n.Success,void(T=r.QueueReferencingProjects);l.options.verbose&&Q(l,e.Diagnostics.Updating_output_of_project_0,u);var c=l.compilerHost;l.projectCompilerOptions=m.options;var h=e.emitUsingBuildInfo(m,c,(function(e){var n=v(l,e.path);return y(l,n,_(l,n))}),s);if(e.isString(h))return Q(l,e.Diagnostics.Cannot_update_output_of_project_0_because_there_was_error_reading_file_1,u,Y(l,h)),T=r.BuildInvalidatedProjectOfBundle,E=I(t.Build,l,u,d,p,m,g);e.Debug.assert(!!h.length);var S=e.createDiagnosticCollection(),x=new e.Map;return h.forEach((function(n){var t=n.name,r=n.text,a=n.writeByteOrderMark;x.set(f(l,t),t),e.writeFile(i?{writeFile:i}:c,S,t,r,a)})),{emitSkipped:!1,diagnostics:G(S,x,a,!1,h[0].name,n.DeclarationOutputUnchanged)}}function V(n,t,a,i){for(;T<=n&&T<r.Done;){var o=T;switch(T){case r.CreateProgram:A();break;case r.SyntaxDiagnostics:N(t);break;case r.SemanticDiagnostics:R(t);break;case r.Emit:M(a,t,i);break;case r.EmitBuildInfo:w(a,t);break;case r.EmitBundle:U(a,i);break;case r.BuildInvalidatedProjectOfBundle:e.Debug.checkDefined(E).done(t),T=r.Done;break;case r.QueueReferencingProjects:B(l,u,d,p,m,g,e.Debug.checkDefined(b)),T++;break;case r.Done:default:e.assertType(T)}e.Debug.assert(T>o)}}}function N(n,t,r){var a=n.options;return!(t.type===e.UpToDateStatusType.OutOfDateWithPrepend&&!a.force)||(0===r.fileNames.length||!!e.getConfigFileParsingDiagnostics(r).length||!e.isIncrementalCompilation(r.options))}function R(n,r,a){if(n.projectPendingBuild.size&&!l(r)){if(n.currentInvalidatedProject)return e.arrayIsEqualTo(n.currentInvalidatedProject.buildOrder,r)?n.currentInvalidatedProject:void 0;for(var i=n.options,o=n.projectPendingBuild,s=0;s<r.length;s++){var c=r[s],u=_(n,c),d=n.projectPendingBuild.get(u);if(void 0!==d){a&&(a=!1,re(n,r));var p=y(n,c,u);if(p){if(d===e.ConfigFileProgramReloadLevel.Full)z(n,c,u,p),q(n,c,u,p),J(n,c,u,p);else if(d===e.ConfigFileProgramReloadLevel.Partial){var m=e.getFileNamesFromConfigSpecs(p.configFileSpecs,e.getDirectoryPath(c),p.options,n.parseConfigFileHost);e.updateErrorForNoInputFiles(m,c,p.configFileSpecs,p.errors,e.canJsonReportNoInputFiles(p.raw)),p.fileNames=m.fileNames,J(n,c,u,p)}var f=w(n,p,u);if(ae(n,c,f),!i.force){if(f.type===e.UpToDateStatusType.UpToDate){ee(n,u,e.getConfigFileParsingDiagnostics(p)),o.delete(u),i.dry&&Q(n,e.Diagnostics.Project_0_is_up_to_date,c);continue}if(f.type===e.UpToDateStatusType.UpToDateWithUpstreamTypes)return ee(n,u,e.getConfigFileParsingDiagnostics(p)),C(n,c,u,p,r)}if(f.type!==e.UpToDateStatusType.UpstreamBlocked){if(f.type!==e.UpToDateStatusType.ContainerOnly)return I(N(n,f,p)?t.Build:t.UpdateBundle,n,c,u,s,p,r);ee(n,u,e.getConfigFileParsingDiagnostics(p)),o.delete(u)}else ee(n,u,e.getConfigFileParsingDiagnostics(p)),o.delete(u),i.verbose&&Q(n,f.upstreamProjectBlocked?e.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_was_not_built:e.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_has_errors,c,f.upstreamProjectName)}else ne(n,u),o.delete(u)}}}}function k(e,n,t){var r=e.writeFileName;r&&n.options.listEmittedFiles&&r("TSFILE: "+t)}function O(n,t,r){t?(t&&n.writeFileName&&e.listFiles(t,n.writeFileName),n.host.afterProgramEmitAndDiagnostics&&n.host.afterProgramEmitAndDiagnostics(t),t.releaseProgram()):n.host.afterEmitBundle&&n.host.afterEmitBundle(r),n.projectCompilerOptions=n.baseCompilerOptions}function P(t,a,i,o,s,c,l){var u=!(c&n.SyntaxErrors)&&i&&!e.outFile(i.getCompilerOptions());return ee(t,a,s),t.projectStatus.set(a,{type:e.UpToDateStatusType.Unbuildable,reason:l+" errors"}),u?{buildResult:c,step:r.EmitBuildInfo}:(O(t,i,o),{buildResult:c,step:r.QueueReferencingProjects})}function M(n,t,r,a){if(r<(n.host.getModifiedTime(t)||e.missingFileModifiedTime))return{type:e.UpToDateStatusType.OutOfDateWithSelf,outOfDateOutputFileName:a,newerInputFileName:t}}function w(n,t,r){if(void 0===t)return{type:e.UpToDateStatusType.Unbuildable,reason:"File deleted mid-build"};var o=n.projectStatus.get(r);if(void 0!==o)return o;var l=function(n,t,r){for(var o=void 0,l=a,u=n.host,d=0,p=t.fileNames;d<p.length;d++){var m=p[d];if(!u.fileExists(m))return{type:e.UpToDateStatusType.Unbuildable,reason:m+" does not exist"};var f=u.getModifiedTime(m)||e.missingFileModifiedTime;f>l&&(o=m,l=f)}if(!t.fileNames.length&&!e.canJsonReportNoInputFiles(t.raw))return{type:e.UpToDateStatusType.ContainerOnly};for(var g,v=e.getAllProjectOutputs(t,!u.useCaseSensitiveFileNames()),h="(none)",b=i,E="(none)",T=a,S=a,x=!1,L=0,A=v;L<A.length;L++){var D=A[L];if(!u.fileExists(D)){g=D;break}var C=u.getModifiedTime(D)||e.missingFileModifiedTime;if(C<b&&(b=C,h=D),C<l){x=!0;break}C>T&&(T=C,E=D),c(D)&&(S=s(S,u.getModifiedTime(D)||e.missingFileModifiedTime))}var I,N=!1,R=!1;if(t.projectReferences){n.projectStatus.set(r,{type:e.UpToDateStatusType.ComputingUpstream});for(var k=0,O=t.projectReferences;k<O.length;k++){var P=O[k];R=R||!!P.prepend;var F=e.resolveProjectReferencePath(P),G=_(n,F),B=w(n,y(n,F,G),G);if(B.type!==e.UpToDateStatusType.ComputingUpstream&&B.type!==e.UpToDateStatusType.ContainerOnly){if(B.type===e.UpToDateStatusType.Unbuildable||B.type===e.UpToDateStatusType.UpstreamBlocked)return{type:e.UpToDateStatusType.UpstreamBlocked,upstreamProjectName:P.path,upstreamProjectBlocked:B.type===e.UpToDateStatusType.UpstreamBlocked};if(B.type!==e.UpToDateStatusType.UpToDate)return{type:e.UpToDateStatusType.UpstreamOutOfDate,upstreamProjectName:P.path};if(!g){if(B.newestInputFileTime&&B.newestInputFileTime<=b)continue;if(B.newestDeclarationFileContentChangedTime&&B.newestDeclarationFileContentChangedTime<=b){N=!0,I=P.path;continue}return e.Debug.assert(void 0!==h,"Should have an oldest output filename here"),{type:e.UpToDateStatusType.OutOfDateWithUpstream,outOfDateOutputFileName:h,newerProjectName:P.path}}}}}if(void 0!==g)return{type:e.UpToDateStatusType.OutputMissing,missingOutputFileName:g};if(x)return{type:e.UpToDateStatusType.OutOfDateWithSelf,outOfDateOutputFileName:h,newerInputFileName:o};var U=M(n,t.options.configFilePath,b,h);if(U)return U;var V=e.forEach(t.options.configFile.extendedSourceFiles||e.emptyArray,(function(e){return M(n,e,b,h)}));if(V)return V;if(!n.buildInfoChecked.has(r)){n.buildInfoChecked.set(r,!0);var j=e.getTsBuildInfoEmitOutputFilePath(t.options);if(j){var H=n.readFileWithCache(j),K=H&&e.getBuildInfo(H);if(K&&(K.bundle||K.program)&&K.version!==e.version)return{type:e.UpToDateStatusType.TsVersionOutputOfDate,version:K.version}}}return R&&N?{type:e.UpToDateStatusType.OutOfDateWithPrepend,outOfDateOutputFileName:h,newerProjectName:I}:{type:N?e.UpToDateStatusType.UpToDateWithUpstreamTypes:e.UpToDateStatusType.UpToDate,newestDeclarationFileContentChangedTime:S,newestInputFileTime:l,newestOutputFileTime:T,newestInputFileName:o,newestOutputFileName:E,oldestOutputFileName:h}}(n,t,r);return n.projectStatus.set(r,l),l}function F(n,t,r,a,i){var o=n.host,l=e.getAllProjectOutputs(t,!o.useCaseSensitiveFileNames());if(!i||l.length!==i.size)for(var u=!!n.options.verbose,d=o.now?o.now():new Date,p=0,m=l;p<m.length;p++){var _=m[p];i&&i.has(f(n,_))||(u&&(u=!1,Q(n,a,t.options.configFilePath)),c(_)&&(r=s(r,o.getModifiedTime(_)||e.missingFileModifiedTime)),o.setModifiedTime(_,d))}return r}function G(n,t,r){if(n.options.dry)return Q(n,e.Diagnostics.A_non_dry_build_would_update_timestamps_for_output_of_project_0,t.options.configFilePath);var i=F(n,t,a,e.Diagnostics.Updating_output_timestamps_of_project_0);n.projectStatus.set(r,{type:e.UpToDateStatusType.UpToDate,newestDeclarationFileContentChangedTime:i,oldestOutputFileName:e.getFirstProjectOutput(t,!n.host.useCaseSensitiveFileNames())})}function B(t,r,a,i,o,s,c){if(!(c&n.AnyErrors)&&o.options.composite)for(var l=i+1;l<s.length;l++){var u=s[l],d=_(t,u);if(!t.projectPendingBuild.has(d)){var p=y(t,u,d);if(p&&p.projectReferences)for(var m=0,f=p.projectReferences;m<f.length;m++){var g=f[m];if(_(t,v(t,g.path))===a){var h=t.projectStatus.get(d);if(h)switch(h.type){case e.UpToDateStatusType.UpToDate:if(c&n.DeclarationOutputUnchanged){g.prepend?t.projectStatus.set(d,{type:e.UpToDateStatusType.OutOfDateWithPrepend,outOfDateOutputFileName:h.oldestOutputFileName,newerProjectName:r}):h.type=e.UpToDateStatusType.UpToDateWithUpstreamTypes;break}case e.UpToDateStatusType.UpToDateWithUpstreamTypes:case e.UpToDateStatusType.OutOfDateWithPrepend:c&n.DeclarationOutputUnchanged||t.projectStatus.set(d,{type:e.UpToDateStatusType.OutOfDateWithUpstream,outOfDateOutputFileName:h.type===e.UpToDateStatusType.OutOfDateWithPrepend?h.outOfDateOutputFileName:h.oldestOutputFileName,newerProjectName:r});break;case e.UpToDateStatusType.UpstreamBlocked:_(t,v(t,h.upstreamProjectName))===a&&x(t,d)}L(t,d,e.ConfigFileProgramReloadLevel.None);break}}}}}function U(n,t,r,a){var i=E(n,t,a);if(!i)return e.ExitStatus.InvalidProject_OutputsSkipped;A(n,r);for(var o=!0,s=0;;){var c=R(n,i,o);if(!c)break;o=!1,c.done(r),n.diagnostics.has(c.projectPath)||s++}return S(n),te(n,i),function(e,n){if(!e.watchAllProjectsPending)return;e.watchAllProjectsPending=!1;for(var t=0,r=u(n);t<r.length;t++){var a=r[t],i=_(e,a),o=y(e,a,i);z(e,a,i,o),o&&(q(e,a,i,o),J(e,a,i,o))}}(n,i),l(i)?e.ExitStatus.ProjectReferenceCycle_OutputsSkipped:i.some((function(e){return n.diagnostics.has(_(n,e))}))?s?e.ExitStatus.DiagnosticsPresent_OutputsGenerated:e.ExitStatus.DiagnosticsPresent_OutputsSkipped:e.ExitStatus.Success}function V(n,t,r){var a=E(n,t,r);if(!a)return e.ExitStatus.InvalidProject_OutputsSkipped;if(l(a))return $(n,a.circularDiagnostics),e.ExitStatus.ProjectReferenceCycle_OutputsSkipped;for(var i=n.options,o=n.host,s=i.dry?[]:void 0,c=0,u=a;c<u.length;c++){var d=u[c],p=_(n,d),m=y(n,d,p);if(void 0!==m)for(var f=0,g=e.getAllProjectOutputs(m,!o.useCaseSensitiveFileNames());f<g.length;f++){var v=g[f];o.fileExists(v)&&(s?s.push(v):(o.deleteFile(v),j(n,p,e.ConfigFileProgramReloadLevel.None)))}else ne(n,p)}return s&&Q(n,e.Diagnostics.A_non_dry_build_would_delete_the_following_files_Colon_0,s.map((function(e){return"\r\n * "+e})).join("")),e.ExitStatus.Success}function j(n,t,r){n.host.getParsedCommandLine&&r===e.ConfigFileProgramReloadLevel.Partial&&(r=e.ConfigFileProgramReloadLevel.Full),r===e.ConfigFileProgramReloadLevel.Full&&(n.configFileCache.delete(t),n.buildOrder=void 0),n.needsSummary=!0,x(n,t),L(n,t,r),T(n)}function H(e,n,t){e.reportFileChangeDetected=!0,j(e,n,t),K(e)}function K(e){var n=e.hostWithWatch;n.setTimeout&&n.clearTimeout&&(e.timerToBuildInvalidatedProject&&n.clearTimeout(e.timerToBuildInvalidatedProject),e.timerToBuildInvalidatedProject=n.setTimeout(W,250,e))}function W(n){n.timerToBuildInvalidatedProject=void 0,n.reportFileChangeDetected&&(n.reportFileChangeDetected=!1,n.projectErrorsReported.clear(),Z(n,e.Diagnostics.File_change_detected_Starting_incremental_compilation));var t=b(n),r=R(n,t,!1);r&&(r.done(),n.projectPendingBuild.size)?n.watch&&!n.timerToBuildInvalidatedProject&&K(n):(S(n),te(n,t))}function z(n,t,r,a){n.watch&&!n.allWatchedConfigFiles.has(r)&&n.allWatchedConfigFiles.set(r,n.watchFile(n.hostWithWatch,t,(function(){H(n,r,e.ConfigFileProgramReloadLevel.Full)}),e.PollingInterval.High,null==a?void 0:a.watchOptions,e.WatchType.ConfigFile,t))}function q(n,t,r,a){n.watch&&e.updateWatchingWildcardDirectories(o(n.allWatchedWildcardDirectories,r),new e.Map(e.getEntries(a.configFileSpecs.wildcardDirectories)),(function(i,o){return n.watchDirectory(n.hostWithWatch,i,(function(o){e.isIgnoredFileFromWildCardWatching({watchedDirPath:f(n,i),fileOrDirectory:o,fileOrDirectoryPath:f(n,o),configFileName:t,configFileSpecs:a.configFileSpecs,currentDirectory:n.currentDirectory,options:a.options,program:n.builderPrograms.get(r),useCaseSensitiveFileNames:n.parseConfigFileHost.useCaseSensitiveFileNames,writeLog:function(e){return n.writeLog(e)}})||H(n,r,e.ConfigFileProgramReloadLevel.Partial)}),o,null==a?void 0:a.watchOptions,e.WatchType.WildcardDirectory,t)}))}function J(n,t,r,a){n.watch&&e.mutateMap(o(n.allWatchedInputFiles,r),e.arrayToMap(a.fileNames,(function(e){return f(n,e)})),{createNewValue:function(i,o){return n.watchFilePath(n.hostWithWatch,o,(function(){return H(n,r,e.ConfigFileProgramReloadLevel.None)}),e.PollingInterval.Low,null==a?void 0:a.watchOptions,i,e.WatchType.SourceFile,t)},onDeleteValue:e.closeFileWatcher})}function X(n,t,r,a,i){var o=m(n,t,r,a,i);return{build:function(e,n){return U(o,e,n)},clean:function(e){return V(o,e)},buildReferences:function(e,n){return U(o,e,n,!0)},cleanReferences:function(e){return V(o,e,!0)},getNextInvalidatedProject:function(e){return A(o,e),R(o,b(o),!1)},getBuildOrder:function(){return b(o)},getUpToDateStatusOfProject:function(e){var n=v(o,e),t=_(o,n);return w(o,y(o,n,t),t)},invalidateProject:function(n,t){return j(o,n,t||e.ConfigFileProgramReloadLevel.None)},buildNextInvalidatedProject:function(){return W(o)},getAllParsedConfigs:function(){return e.arrayFrom(e.mapDefinedIterator(o.configFileCache.values(),(function(e){return g(e)?e:void 0})))},close:function(){return function(n){e.clearMap(n.allWatchedConfigFiles,e.closeFileWatcher),e.clearMap(n.allWatchedWildcardDirectories,(function(n){return e.clearMap(n,e.closeFileWatcherOf)})),e.clearMap(n.allWatchedInputFiles,(function(n){return e.clearMap(n,e.closeFileWatcher)}))}(o)}}}function Y(n,t){return e.convertToRelativePath(t,n.currentDirectory,(function(e){return n.getCanonicalFileName(e)}))}function Q(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];n.host.reportSolutionBuilderStatus(e.createCompilerDiagnostic.apply(void 0,__spreadArrays([t],r)))}function Z(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];n.hostWithWatch.onWatchStatusChange&&n.hostWithWatch.onWatchStatusChange(e.createCompilerDiagnostic.apply(void 0,__spreadArrays([t],r)),n.host.getNewLine(),n.baseCompilerOptions)}function $(e,n){var t=e.host;n.forEach((function(e){return t.reportDiagnostic(e)}))}function ee(e,n,t){$(e,t),e.projectErrorsReported.set(n,!0),t.length&&e.diagnostics.set(n,t)}function ne(e,n){ee(e,n,[e.configFileCache.get(n)])}function te(n,t){if(n.needsSummary){n.needsSummary=!1;var r=n.watch||!!n.host.reportErrorSummary,a=n.diagnostics,i=0;l(t)?(re(n,t.buildOrder),$(n,t.circularDiagnostics),r&&(i+=e.getErrorCountForSummary(t.circularDiagnostics))):(t.forEach((function(t){var r=_(n,t);n.projectErrorsReported.has(r)||$(n,a.get(r)||e.emptyArray)})),r&&a.forEach((function(n){return i+=e.getErrorCountForSummary(n)}))),n.watch?Z(n,e.getWatchErrorSummaryDiagnosticMessage(i),i):n.host.reportErrorSummary&&n.host.reportErrorSummary(i)}}function re(n,t){n.options.verbose&&Q(n,e.Diagnostics.Projects_in_this_build_Colon_0,t.map((function(e){return"\r\n    * "+Y(n,e)})).join(""))}function ae(n,t,r){n.options.verbose&&function(n,t,r){switch(r.type){case e.UpToDateStatusType.OutOfDateWithSelf:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2,Y(n,t),Y(n,r.outOfDateOutputFileName),Y(n,r.newerInputFileName));case e.UpToDateStatusType.OutOfDateWithUpstream:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2,Y(n,t),Y(n,r.outOfDateOutputFileName),Y(n,r.newerProjectName));case e.UpToDateStatusType.OutputMissing:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_output_file_1_does_not_exist,Y(n,t),Y(n,r.missingOutputFileName));case e.UpToDateStatusType.UpToDate:if(void 0!==r.newestInputFileTime)return Q(n,e.Diagnostics.Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2,Y(n,t),Y(n,r.newestInputFileName||""),Y(n,r.oldestOutputFileName||""));break;case e.UpToDateStatusType.OutOfDateWithPrepend:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed,Y(n,t),Y(n,r.newerProjectName));case e.UpToDateStatusType.UpToDateWithUpstreamTypes:return Q(n,e.Diagnostics.Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies,Y(n,t));case e.UpToDateStatusType.UpstreamOutOfDate:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date,Y(n,t),Y(n,r.upstreamProjectName));case e.UpToDateStatusType.UpstreamBlocked:return Q(n,r.upstreamProjectBlocked?e.Diagnostics.Project_0_can_t_be_built_because_its_dependency_1_was_not_built:e.Diagnostics.Project_0_can_t_be_built_because_its_dependency_1_has_errors,Y(n,t),Y(n,r.upstreamProjectName));case e.UpToDateStatusType.Unbuildable:return Q(n,e.Diagnostics.Failed_to_parse_file_0_Colon_1,Y(n,t),r.reason);case e.UpToDateStatusType.TsVersionOutputOfDate:return Q(n,e.Diagnostics.Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_current_version_2,Y(n,t),r.version,e.version);case e.UpToDateStatusType.ContainerOnly:case e.UpToDateStatusType.ComputingUpstream:break;default:e.assertType(r)}}(n,t,r)}!function(e){e[e.None=0]="None",e[e.Success=1]="Success",e[e.DeclarationOutputUnchanged=2]="DeclarationOutputUnchanged",e[e.ConfigFileErrors=4]="ConfigFileErrors",e[e.SyntaxErrors=8]="SyntaxErrors",e[e.TypeErrors=16]="TypeErrors",e[e.DeclarationEmitErrors=32]="DeclarationEmitErrors",e[e.EmitErrors=64]="EmitErrors",e[e.AnyErrors=124]="AnyErrors"}(n||(n={})),e.isCircularBuildOrder=l,e.getBuildOrderFromAnyBuildOrder=u,e.createBuilderStatusReporter=d,e.createSolutionBuilderHost=function(n,t,r,a,i){void 0===n&&(n=e.sys);var o=p(n,t,r,a);return o.reportErrorSummary=i,o},e.createSolutionBuilderWithWatchHost=function(n,t,r,a,i){void 0===n&&(n=e.sys);var o=p(n,t,r,a),s=e.createWatchHost(n,i);return e.copyProperties(o,s),o},e.createSolutionBuilder=function(e,n,t){return X(!1,e,n,t)},e.createSolutionBuilderWithWatch=function(e,n,t,r){return X(!0,e,n,t,r)},function(e){e[e.Build=0]="Build",e[e.UpdateBundle=1]="UpdateBundle",e[e.UpdateOutputFileStamps=2]="UpdateOutputFileStamps"}(t=e.InvalidatedProjectKind||(e.InvalidatedProjectKind={})),function(e){e[e.CreateProgram=0]="CreateProgram",e[e.SyntaxDiagnostics=1]="SyntaxDiagnostics",e[e.SemanticDiagnostics=2]="SemanticDiagnostics",e[e.Emit=3]="Emit",e[e.EmitBundle=4]="EmitBundle",e[e.EmitBuildInfo=5]="EmitBuildInfo",e[e.BuildInvalidatedProjectOfBundle=6]="BuildInvalidatedProjectOfBundle",e[e.QueueReferencingProjects=7]="QueueReferencingProjects",e[e.Done=8]="Done"}(r||(r={}))}(ts||(ts={})),function(e){!function(n){n.ActionSet="action::set",n.ActionInvalidate="action::invalidate",n.ActionPackageInstalled="action::packageInstalled",n.EventTypesRegistry="event::typesRegistry",n.EventBeginInstallTypes="event::beginInstallTypes",n.EventEndInstallTypes="event::endInstallTypes",n.EventInitializationFailed="event::initializationFailed",function(e){e.GlobalCacheLocation="--globalTypingsCacheLocation",e.LogFile="--logFile",e.EnableTelemetry="--enableTelemetry",e.TypingSafeListLocation="--typingSafeListLocation",e.TypesMapLocation="--typesMapLocation",e.NpmLocation="--npmLocation",e.ValidateDefaultNpmLocation="--validateDefaultNpmLocation"}(n.Arguments||(n.Arguments={})),n.hasArgument=function(n){return e.sys.args.indexOf(n)>=0},n.findArgument=function(n){var t=e.sys.args.indexOf(n);return t>=0&&t<e.sys.args.length-1?e.sys.args[t+1]:void 0},n.nowString=function(){var n=new Date;return e.padLeft(n.getHours().toString(),2,"0")+":"+e.padLeft(n.getMinutes().toString(),2,"0")+":"+e.padLeft(n.getSeconds().toString(),2,"0")+"."+e.padLeft(n.getMilliseconds().toString(),3,"0")}}(e.server||(e.server={}))}(ts||(ts={})),function(e){!function(n){function t(n,t){return new e.Version(e.getProperty(t,"ts"+e.versionMajorMinor)||e.getProperty(t,"latest")).compareTo(n.version)<=0}function r(e){return n.nodeCoreModules.has(e)?"node":e}n.isTypingUpToDate=t,n.nodeCoreModuleList=["assert","async_hooks","buffer","child_process","cluster","console","constants","crypto","dgram","dns","domain","events","fs","http","https","http2","inspector","net","os","path","perf_hooks","process","punycode","querystring","readline","repl","stream","string_decoder","timers","tls","tty","url","util","v8","vm","zlib"],n.nodeCoreModules=new e.Set(n.nodeCoreModuleList),n.nonRelativeModuleNameForTypingCache=r,n.loadSafeList=function(n,t){var r=e.readConfigFile(t,(function(e){return n.readFile(e)}));return new e.Map(e.getEntries(r.config))},n.loadTypesMap=function(n,t){var r=e.readConfigFile(t,(function(e){return n.readFile(e)}));if(r.config)return new e.Map(e.getEntries(r.config.simpleMap))},n.discoverTypings=function(n,a,i,o,s,c,l,u,d){if(!l||!l.enable)return{cachedTypingPaths:[],newTypingNames:[],filesToWatch:[]};var p=new e.Map;i=e.mapDefined(i,(function(n){var t=e.normalizePath(n);if(e.hasJSFileExtension(t))return t}));var m=[];l.include&&S(l.include,"Explicitly included types");var f=l.exclude||[],_=new e.Set(i.map(e.getDirectoryPath));_.add(o),_.forEach((function(n){x(e.combinePaths(n,"package.json"),m),x(e.combinePaths(n,"bower.json"),m),L(e.combinePaths(n,"bower_components"),m),L(e.combinePaths(n,"node_modules"),m)})),l.disableFilenameBasedTypeAcquisition||function(n){var t=e.mapDefined(n,(function(n){if(e.hasJSFileExtension(n)){var t=e.removeFileExtension(e.getBaseFileName(n.toLowerCase())),r=e.removeMinAndVersionNumbers(t);return s.get(r)}}));t.length&&S(t,"Inferred typings from file names");e.some(n,(function(n){return e.fileExtensionIs(n,".jsx")}))&&(a&&a("Inferred 'react' typings due to presence of '.jsx' extension"),T("react"))}(i),u&&S(e.deduplicate(u.map(r),e.equateStringsCaseSensitive,e.compareStringsCaseSensitive),"Inferred typings from unresolved imports"),c.forEach((function(e,n){var r=d.get(n);p.has(n)&&void 0===p.get(n)&&void 0!==r&&t(e,r)&&p.set(n,e.typingLocation)}));for(var g=0,y=f;g<y.length;g++){var v=y[g];p.delete(v)&&a&&a("Typing for "+v+" is in exclude list, will be ignored.")}var h=[],b=[];p.forEach((function(e,n){void 0!==e?b.push(e):h.push(n)}));var E={cachedTypingPaths:b,newTypingNames:h,filesToWatch:m};return a&&a("Result: "+JSON.stringify(E)),E;function T(e){p.has(e)||p.set(e,void 0)}function S(n,t){a&&a(t+": "+JSON.stringify(n)),e.forEach(n,T)}function x(t,r){if(n.fileExists(t)){r.push(t);var a=e.readConfigFile(t,(function(e){return n.readFile(e)})).config;S(e.flatMap([a.dependencies,a.devDependencies,a.optionalDependencies,a.peerDependencies],e.getOwnKeys),"Typing names in '"+t+"' dependencies")}}function L(t,r){if(r.push(t),n.directoryExists(t)){var i=n.readDirectory(t,[".json"],void 0,void 0,2);a&&a("Searching for typing names in "+t+"; all files: "+JSON.stringify(i));for(var o=[],s=0,c=i;s<c.length;s++){var l=c[s],u=e.normalizePath(l),d=e.getBaseFileName(u);if("package.json"===d||"bower.json"===d){var m=e.readConfigFile(u,(function(e){return n.readFile(e)})).config;if(("package.json"!==d||!m._requiredBy||0!==e.filter(m._requiredBy,(function(e){return"#"===e[0]||"/"===e})).length)&&m.name){var f=m.types||m.typings;if(f){var _=e.getNormalizedAbsolutePath(f,e.getDirectoryPath(u));a&&a("    Package '"+m.name+"' provides its own types."),p.set(m.name,_)}else o.push(m.name)}}}S(o,"    Found package names")}}},function(e){e[e.Ok=0]="Ok",e[e.EmptyName=1]="EmptyName",e[e.NameTooLong=2]="NameTooLong",e[e.NameStartsWithDot=3]="NameStartsWithDot",e[e.NameStartsWithUnderscore=4]="NameStartsWithUnderscore",e[e.NameContainsNonURISafeCharacters=5]="NameContainsNonURISafeCharacters"}(n.NameValidationResult||(n.NameValidationResult={}));function a(e,n){if(!e)return 1;if(e.length>214)return 2;if(46===e.charCodeAt(0))return 3;if(95===e.charCodeAt(0))return 4;if(n){var t=/^@([^/]+)\/([^/]+)$/.exec(e);if(t){var r=a(t[1],!1);if(0!==r)return{name:t[1],isScopeName:!0,result:r};var i=a(t[2],!1);return 0!==i?{name:t[2],isScopeName:!1,result:i}:0}}return encodeURIComponent(e)!==e?5:0}function i(n,t,r,a){var i=a?"Scope":"Package";switch(t){case 1:return"'"+n+"':: "+i+" name '"+r+"' cannot be empty";case 2:return"'"+n+"':: "+i+" name '"+r+"' should be less than 214 characters";case 3:return"'"+n+"':: "+i+" name '"+r+"' cannot start with '.'";case 4:return"'"+n+"':: "+i+" name '"+r+"' cannot start with '_'";case 5:return"'"+n+"':: "+i+" name '"+r+"' contains non URI safe characters";case 0:return e.Debug.fail();default:throw e.Debug.assertNever(t)}}n.validatePackageName=function(e){return a(e,!0)},n.renderPackageNameValidationFailure=function(e,n){return"object"==typeof e?i(n,e.result,e.name,e.isScopeName):i(n,e,n,!1)}}(e.JsTyping||(e.JsTyping={}))}(ts||(ts={})),function(e){var n,t;function r(e){return{indentSize:4,tabSize:4,newLineCharacter:e||"\n",convertTabsToSpaces:!0,indentStyle:n.Smart,insertSpaceAfterConstructor:!1,insertSpaceAfterCommaDelimiter:!0,insertSpaceAfterSemicolonInForStatements:!0,insertSpaceBeforeAndAfterBinaryOperators:!0,insertSpaceAfterKeywordsInControlFlowStatements:!0,insertSpaceAfterFunctionKeywordForAnonymousFunctions:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets:!1,insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces:!0,insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces:!1,insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces:!1,insertSpaceBeforeFunctionParenthesis:!1,placeOpenBraceOnNewLineForFunctions:!1,placeOpenBraceOnNewLineForControlBlocks:!1,semicolons:t.Ignore,trimTrailingWhitespace:!0}}!function(e){var n=function(){function e(e){this.text=e}return e.prototype.getText=function(e,n){return 0===e&&n===this.text.length?this.text:this.text.substring(e,n)},e.prototype.getLength=function(){return this.text.length},e.prototype.getChangeRange=function(){},e}();e.fromString=function(e){return new n(e)}}(e.ScriptSnapshot||(e.ScriptSnapshot={})),function(e){e[e.Dependencies=1]="Dependencies",e[e.DevDependencies=2]="DevDependencies",e[e.PeerDependencies=4]="PeerDependencies",e[e.OptionalDependencies=8]="OptionalDependencies",e[e.All=15]="All"}(e.PackageJsonDependencyGroup||(e.PackageJsonDependencyGroup={})),function(e){e[e.Off=0]="Off",e[e.On=1]="On",e[e.Auto=2]="Auto"}(e.PackageJsonAutoImportPreference||(e.PackageJsonAutoImportPreference={})),function(e){e[e.Semantic=0]="Semantic",e[e.PartialSemantic=1]="PartialSemantic",e[e.Syntactic=2]="Syntactic"}(e.LanguageServiceMode||(e.LanguageServiceMode={})),e.emptyOptions={},function(e){e.Original="original",e.TwentyTwenty="2020"}(e.SemanticClassificationFormat||(e.SemanticClassificationFormat={})),function(e){e.none="none",e.definition="definition",e.reference="reference",e.writtenReference="writtenReference"}(e.HighlightSpanKind||(e.HighlightSpanKind={})),function(e){e[e.None=0]="None",e[e.Block=1]="Block",e[e.Smart=2]="Smart"}(n=e.IndentStyle||(e.IndentStyle={})),function(e){e.Ignore="ignore",e.Insert="insert",e.Remove="remove"}(t=e.SemicolonPreference||(e.SemicolonPreference={})),e.getDefaultFormatCodeSettings=r,e.testFormatSettings=r("\n"),function(e){e[e.aliasName=0]="aliasName",e[e.className=1]="className",e[e.enumName=2]="enumName",e[e.fieldName=3]="fieldName",e[e.interfaceName=4]="interfaceName",e[e.keyword=5]="keyword",e[e.lineBreak=6]="lineBreak",e[e.numericLiteral=7]="numericLiteral",e[e.stringLiteral=8]="stringLiteral",e[e.localName=9]="localName",e[e.methodName=10]="methodName",e[e.moduleName=11]="moduleName",e[e.operator=12]="operator",e[e.parameterName=13]="parameterName",e[e.propertyName=14]="propertyName",e[e.punctuation=15]="punctuation",e[e.space=16]="space",e[e.text=17]="text",e[e.typeParameterName=18]="typeParameterName",e[e.enumMemberName=19]="enumMemberName",e[e.functionName=20]="functionName",e[e.regularExpressionLiteral=21]="regularExpressionLiteral"}(e.SymbolDisplayPartKind||(e.SymbolDisplayPartKind={})),function(e){e.Comment="comment",e.Region="region",e.Code="code",e.Imports="imports"}(e.OutliningSpanKind||(e.OutliningSpanKind={})),function(e){e[e.JavaScript=0]="JavaScript",e[e.SourceMap=1]="SourceMap",e[e.Declaration=2]="Declaration"}(e.OutputFileType||(e.OutputFileType={})),function(e){e[e.None=0]="None",e[e.InMultiLineCommentTrivia=1]="InMultiLineCommentTrivia",e[e.InSingleQuoteStringLiteral=2]="InSingleQuoteStringLiteral",e[e.InDoubleQuoteStringLiteral=3]="InDoubleQuoteStringLiteral",e[e.InTemplateHeadOrNoSubstitutionTemplate=4]="InTemplateHeadOrNoSubstitutionTemplate",e[e.InTemplateMiddleOrTail=5]="InTemplateMiddleOrTail",e[e.InTemplateSubstitutionPosition=6]="InTemplateSubstitutionPosition"}(e.EndOfLineState||(e.EndOfLineState={})),function(e){e[e.Punctuation=0]="Punctuation",e[e.Keyword=1]="Keyword",e[e.Operator=2]="Operator",e[e.Comment=3]="Comment",e[e.Whitespace=4]="Whitespace",e[e.Identifier=5]="Identifier",e[e.NumberLiteral=6]="NumberLiteral",e[e.BigIntLiteral=7]="BigIntLiteral",e[e.StringLiteral=8]="StringLiteral",e[e.RegExpLiteral=9]="RegExpLiteral"}(e.TokenClass||(e.TokenClass={})),function(e){e.unknown="",e.warning="warning",e.keyword="keyword",e.scriptElement="script",e.moduleElement="module",e.classElement="class",e.localClassElement="local class",e.interfaceElement="interface",e.typeElement="type",e.enumElement="enum",e.enumMemberElement="enum member",e.variableElement="var",e.localVariableElement="local var",e.functionElement="function",e.localFunctionElement="local function",e.memberFunctionElement="method",e.memberGetAccessorElement="getter",e.memberSetAccessorElement="setter",e.memberVariableElement="property",e.constructorImplementationElement="constructor",e.callSignatureElement="call",e.indexSignatureElement="index",e.constructSignatureElement="construct",e.parameterElement="parameter",e.typeParameterElement="type parameter",e.primitiveType="primitive type",e.label="label",e.alias="alias",e.constElement="const",e.letElement="let",e.directory="directory",e.externalModuleName="external module name",e.jsxAttribute="JSX attribute",e.string="string"}(e.ScriptElementKind||(e.ScriptElementKind={})),function(e){e.none="",e.publicMemberModifier="public",e.privateMemberModifier="private",e.protectedMemberModifier="protected",e.exportedModifier="export",e.ambientModifier="declare",e.staticModifier="static",e.abstractModifier="abstract",e.optionalModifier="optional",e.deprecatedModifier="deprecated",e.dtsModifier=".d.ts",e.tsModifier=".ts",e.tsxModifier=".tsx",e.jsModifier=".js",e.jsxModifier=".jsx",e.jsonModifier=".json"}(e.ScriptElementKindModifier||(e.ScriptElementKindModifier={})),function(e){e.comment="comment",e.identifier="identifier",e.keyword="keyword",e.numericLiteral="number",e.bigintLiteral="bigint",e.operator="operator",e.stringLiteral="string",e.whiteSpace="whitespace",e.text="text",e.punctuation="punctuation",e.className="class name",e.enumName="enum name",e.interfaceName="interface name",e.moduleName="module name",e.typeParameterName="type parameter name",e.typeAliasName="type alias name",e.parameterName="parameter name",e.docCommentTagName="doc comment tag name",e.jsxOpenTagName="jsx open tag name",e.jsxCloseTagName="jsx close tag name",e.jsxSelfClosingTagName="jsx self closing tag name",e.jsxAttribute="jsx attribute",e.jsxText="jsx text",e.jsxAttributeStringLiteralValue="jsx attribute string literal value"}(e.ClassificationTypeNames||(e.ClassificationTypeNames={})),function(e){e[e.comment=1]="comment",e[e.identifier=2]="identifier",e[e.keyword=3]="keyword",e[e.numericLiteral=4]="numericLiteral",e[e.operator=5]="operator",e[e.stringLiteral=6]="stringLiteral",e[e.regularExpressionLiteral=7]="regularExpressionLiteral",e[e.whiteSpace=8]="whiteSpace",e[e.text=9]="text",e[e.punctuation=10]="punctuation",e[e.className=11]="className",e[e.enumName=12]="enumName",e[e.interfaceName=13]="interfaceName",e[e.moduleName=14]="moduleName",e[e.typeParameterName=15]="typeParameterName",e[e.typeAliasName=16]="typeAliasName",e[e.parameterName=17]="parameterName",e[e.docCommentTagName=18]="docCommentTagName",e[e.jsxOpenTagName=19]="jsxOpenTagName",e[e.jsxCloseTagName=20]="jsxCloseTagName",e[e.jsxSelfClosingTagName=21]="jsxSelfClosingTagName",e[e.jsxAttribute=22]="jsxAttribute",e[e.jsxText=23]="jsxText",e[e.jsxAttributeStringLiteralValue=24]="jsxAttributeStringLiteralValue",e[e.bigintLiteral=25]="bigintLiteral"}(e.ClassificationType||(e.ClassificationType={}))}(ts||(ts={})),function(e){function n(n){switch(n.kind){case 249:return e.isInJSFile(n)&&e.getJSDocEnumTag(n)?7:1;case 160:case 198:case 163:case 162:case 288:case 289:case 165:case 164:case 166:case 167:case 168:case 251:case 208:case 209:case 287:case 280:return 1;case 159:case 253:case 254:case 177:return 2;case 331:return void 0===n.name?3:2;case 291:case 252:return 3;case 256:return e.isAmbientModule(n)||1===e.getModuleInstanceState(n)?5:4;case 255:case 264:case 265:case 260:case 261:case 266:case 267:return 7;case 297:return 5}return 7}function t(n){for(;157===n.parent.kind;)n=n.parent;return e.isInternalModuleImportEqualsDeclaration(n.parent)&&n.parent.moduleReference===n}function r(e){return e.expression}function a(e){return e.tag}function i(e){return e.tagName}function o(n,t,r,a,i){var o=a?c(n):s(n);return i&&(o=e.skipOuterExpressions(o)),!!o&&!!o.parent&&t(o.parent)&&r(o.parent)===o}function s(e){return d(e)?e.parent:e}function c(e){return d(e)||p(e)?e.parent:e}function l(n){var t;return e.isIdentifier(n)&&(null===(t=e.tryCast(n.parent,e.isBreakOrContinueStatement))||void 0===t?void 0:t.label)===n}function u(n){var t;return e.isIdentifier(n)&&(null===(t=e.tryCast(n.parent,e.isLabeledStatement))||void 0===t?void 0:t.label)===n}function d(n){var t;return(null===(t=e.tryCast(n.parent,e.isPropertyAccessExpression))||void 0===t?void 0:t.name)===n}function p(n){var t;return(null===(t=e.tryCast(n.parent,e.isElementAccessExpression))||void 0===t?void 0:t.argumentExpression)===n}e.scanner=e.createScanner(99,!0),function(e){e[e.None=0]="None",e[e.Value=1]="Value",e[e.Type=2]="Type",e[e.Namespace=4]="Namespace",e[e.All=7]="All"}(e.SemanticMeaning||(e.SemanticMeaning={})),e.getMeaningFromDeclaration=n,e.getMeaningFromLocation=function(r){return 297===(r=N(r)).kind?1:266===r.parent.kind||272===r.parent.kind||265===r.parent.kind||262===r.parent.kind||e.isImportEqualsDeclaration(r.parent)&&r===r.parent.name?7:t(r)?function(n){var t=157===n.kind?n:e.isQualifiedName(n.parent)&&n.parent.right===n?n.parent:void 0;return t&&260===t.parent.kind?7:4}(r):e.isDeclarationName(r)?n(r.parent):e.isEntityName(r)&&e.isJSDocNameReference(r.parent)?7:function(n){e.isRightSideOfQualifiedNameOrPropertyAccess(n)&&(n=n.parent);switch(n.kind){case 107:return!e.isExpressionNode(n);case 187:return!0}switch(n.parent.kind){case 173:return!0;case 195:return!n.parent.isTypeOf;case 223:return!e.isExpressionWithTypeArgumentsInClassExtendsClause(n.parent)}return!1}(r)?2:function(e){return function(e){var n=e,t=!0;if(157===n.parent.kind){for(;n.parent&&157===n.parent.kind;)n=n.parent;t=n.right===e}return 173===n.parent.kind&&!t}(e)||function(e){var n=e,t=!0;if(201===n.parent.kind){for(;n.parent&&201===n.parent.kind;)n=n.parent;t=n.name===e}if(!t&&223===n.parent.kind&&286===n.parent.parent.kind){var r=n.parent.parent.parent;return 252===r.kind&&116===n.parent.parent.token||253===r.kind&&93===n.parent.parent.token}return!1}(e)}(r)?4:e.isTypeParameterDeclaration(r.parent)?(e.Debug.assert(e.isJSDocTemplateTag(r.parent.parent)),2):e.isLiteralTypeNode(r.parent)?3:1},e.isInRightSideOfInternalImportEqualsDeclaration=t,e.isCallExpressionTarget=function(n,t,a){return void 0===t&&(t=!1),void 0===a&&(a=!1),o(n,e.isCallExpression,r,t,a)},e.isNewExpressionTarget=function(n,t,a){return void 0===t&&(t=!1),void 0===a&&(a=!1),o(n,e.isNewExpression,r,t,a)},e.isCallOrNewExpressionTarget=function(n,t,a){return void 0===t&&(t=!1),void 0===a&&(a=!1),o(n,e.isCallOrNewExpression,r,t,a)},e.isTaggedTemplateTag=function(n,t,r){return void 0===t&&(t=!1),void 0===r&&(r=!1),o(n,e.isTaggedTemplateExpression,a,t,r)},e.isDecoratorTarget=function(n,t,a){return void 0===t&&(t=!1),void 0===a&&(a=!1),o(n,e.isDecorator,r,t,a)},e.isJsxOpeningLikeElementTagName=function(n,t,r){return void 0===t&&(t=!1),void 0===r&&(r=!1),o(n,e.isJsxOpeningLikeElement,i,t,r)},e.climbPastPropertyAccess=s,e.climbPastPropertyOrElementAccess=c,e.getTargetLabel=function(e,n){for(;e;){if(245===e.kind&&e.label.escapedText===n)return e.label;e=e.parent}},e.hasPropertyAccessExpressionWithName=function(n,t){return!!e.isPropertyAccessExpression(n.expression)&&n.expression.name.text===t},e.isJumpStatementTarget=l,e.isLabelOfLabeledStatement=u,e.isLabelName=function(e){return u(e)||l(e)},e.isTagName=function(n){var t;return(null===(t=e.tryCast(n.parent,e.isJSDocTag))||void 0===t?void 0:t.tagName)===n},e.isRightSideOfQualifiedName=function(n){var t;return(null===(t=e.tryCast(n.parent,e.isQualifiedName))||void 0===t?void 0:t.right)===n},e.isRightSideOfPropertyAccess=d,e.isArgumentExpressionOfElementAccess=p,e.isNameOfModuleDeclaration=function(n){var t;return(null===(t=e.tryCast(n.parent,e.isModuleDeclaration))||void 0===t?void 0:t.name)===n},e.isNameOfFunctionDeclaration=function(n){var t;return e.isIdentifier(n)&&(null===(t=e.tryCast(n.parent,e.isFunctionLike))||void 0===t?void 0:t.name)===n},e.isLiteralNameOfPropertyDeclarationOrIndexAccess=function(n){switch(n.parent.kind){case 163:case 162:case 288:case 291:case 165:case 164:case 167:case 168:case 256:return e.getNameOfDeclaration(n.parent)===n;case 202:return n.parent.argumentExpression===n;case 158:return!0;case 191:return 189===n.parent.parent.kind;default:return!1}},e.isExpressionOfExternalModuleImportEqualsDeclaration=function(n){return e.isExternalModuleImportEqualsDeclaration(n.parent.parent)&&e.getExternalModuleImportEqualsDeclarationExpression(n.parent.parent)===n},e.getContainerNode=function(n){for(e.isJSDocTypeAlias(n)&&(n=n.parent.parent);;){if(!(n=n.parent))return;switch(n.kind){case 297:case 165:case 164:case 251:case 208:case 167:case 168:case 252:case 253:case 255:case 256:return n}}},e.getNodeKind=function n(t){switch(t.kind){case 297:return e.isExternalModule(t)?"module":"script";case 256:return"module";case 252:case 221:return"class";case 253:return"interface";case 254:case 324:case 331:return"type";case 255:return"enum";case 249:return c(t);case 198:return c(e.getRootDeclaration(t));case 209:case 251:case 208:return"function";case 167:return"getter";case 168:return"setter";case 165:case 164:return"method";case 288:var r=t.initializer;return e.isFunctionLike(r)?"method":"property";case 163:case 162:case 289:case 290:return"property";case 171:return"index";case 170:return"construct";case 169:return"call";case 166:return"constructor";case 159:return"type parameter";case 291:return"enum member";case 160:return e.hasSyntacticModifier(t,92)?"property":"parameter";case 260:case 265:case 270:case 263:case 269:return"alias";case 216:var a=e.getAssignmentDeclarationKind(t),i=t.right;switch(a){case 7:case 8:case 9:case 0:return"";case 1:case 2:var o=n(i);return""===o?"const":o;case 3:return e.isFunctionExpression(i)?"method":"property";case 4:return"property";case 5:return e.isFunctionExpression(i)?"method":"property";case 6:return"local class";default:return e.assertType(a),""}case 78:return e.isImportClause(t.parent)?"alias":"";case 266:var s=n(t.expression);return""===s?"const":s;default:return""}function c(n){return e.isVarConst(n)?"const":e.isLet(n)?"let":"var"}},e.isThis=function(n){switch(n.kind){case 107:return!0;case 78:return e.identifierIsThisKeyword(n)&&160===n.parent.kind;default:return!1}};var m=/^\/\/\/\s*</;function f(e,n){return g(e.pos,e.end,n)}function _(e,n){return e.pos<n&&n<e.end}function g(e,n,t){return e<=t.pos&&n>=t.end}function y(e,n,t,r){return Math.max(e,t)<Math.min(n,r)}function v(n,t){if(void 0===n||e.nodeIsMissing(n))return!1;switch(n.kind){case 252:case 253:case 255:case 200:case 196:case 177:case 230:case 257:case 258:case 264:case 268:return h(n,19,t);case 287:return v(n.block,t);case 204:if(!n.arguments)return!0;case 203:case 207:case 186:return h(n,21,t);case 174:case 175:return v(n.type,t);case 166:case 167:case 168:case 251:case 208:case 165:case 164:case 170:case 169:case 209:return n.body?v(n.body,t):n.type?v(n.type,t):b(n,21,t);case 256:return!!n.body&&v(n.body,t);case 234:return n.elseStatement?v(n.elseStatement,t):v(n.thenStatement,t);case 233:return v(n.expression,t)||b(n,26,t);case 199:case 197:case 202:case 158:case 179:return h(n,23,t);case 171:return n.type?v(n.type,t):b(n,23,t);case 284:case 285:return!1;case 237:case 238:case 239:case 236:return v(n.statement,t);case 235:return b(n,114,t)?h(n,21,t):v(n.statement,t);case 176:return v(n.exprName,t);case 211:case 210:case 212:case 219:case 220:return v(n.expression,t);case 205:return v(n.template,t);case 218:return v(e.lastOrUndefined(n.templateSpans),t);case 228:return e.nodeIsPresent(n.literal);case 267:case 261:return e.nodeIsPresent(n.moduleSpecifier);case 214:return v(n.operand,t);case 216:return v(n.right,t);case 217:return v(n.whenFalse,t);default:return!0}}function h(n,t,r){var a=n.getChildren(r);if(a.length){var i=e.last(a);if(i.kind===t)return!0;if(26===i.kind&&1!==a.length)return a[a.length-2].kind===t}return!1}function b(e,n,t){return!!E(e,n,t)}function E(n,t,r){return e.find(n.getChildren(r),(function(e){return e.kind===t}))}function T(n){var t=e.find(n.parent.getChildren(),(function(t){return e.isSyntaxList(t)&&f(t,n)}));return e.Debug.assert(!t||e.contains(t.getChildren(),n)),t}function S(e){return 87===e.kind}function x(e){return 83===e.kind}function L(e){return 97===e.kind}function A(n,t){if(!t)switch(n.kind){case 252:case 221:return function(n){if(e.isNamedDeclaration(n))return n.name;if(e.isClassDeclaration(n)){var t=e.find(n.modifiers,S);if(t)return t}if(e.isClassExpression(n)){var r=e.find(n.getChildren(),x);if(r)return r}}(n);case 251:case 208:return function(n){if(e.isNamedDeclaration(n))return n.name;if(e.isFunctionDeclaration(n)){var t=e.find(n.modifiers,S);if(t)return t}if(e.isFunctionExpression(n)){var r=e.find(n.getChildren(),L);if(r)return r}}(n)}if(e.isNamedDeclaration(n))return n.name}function D(n,t){if(n.importClause){if(n.importClause.name&&n.importClause.namedBindings)return;if(n.importClause.name)return n.importClause.name;if(n.importClause.namedBindings){if(e.isNamedImports(n.importClause.namedBindings)){var r=e.singleOrUndefined(n.importClause.namedBindings.elements);if(!r)return;return r.name}if(e.isNamespaceImport(n.importClause.namedBindings))return n.importClause.namedBindings.name}}if(!t)return n.moduleSpecifier}function C(n,t){if(n.exportClause){if(e.isNamedExports(n.exportClause)){if(!e.singleOrUndefined(n.exportClause.elements))return;return n.exportClause.elements[0].name}if(e.isNamespaceExport(n.exportClause))return n.exportClause.name}if(!t)return n.moduleSpecifier}function I(n,t){var r=n.parent;if((e.isModifier(n)&&(t||87!==n.kind)?e.contains(r.modifiers,n):83===n.kind?e.isClassDeclaration(r)||e.isClassExpression(n):97===n.kind?e.isFunctionDeclaration(r)||e.isFunctionExpression(n):117===n.kind?e.isInterfaceDeclaration(r):91===n.kind?e.isEnumDeclaration(r):149===n.kind?e.isTypeAliasDeclaration(r):140===n.kind||139===n.kind?e.isModuleDeclaration(r):99===n.kind?e.isImportEqualsDeclaration(r):134===n.kind?e.isGetAccessorDeclaration(r):146===n.kind&&e.isSetAccessorDeclaration(r))&&(i=A(r,t)))return i;if((112===n.kind||84===n.kind||118===n.kind)&&e.isVariableDeclarationList(r)&&1===r.declarations.length){var a=r.declarations[0];if(e.isIdentifier(a.name))return a.name}if(149===n.kind){if(e.isImportClause(r)&&r.isTypeOnly)if(i=D(r.parent,t))return i;if(e.isExportDeclaration(r)&&r.isTypeOnly)if(i=C(r,t))return i}if(126===n.kind){if(e.isImportSpecifier(r)&&r.propertyName||e.isExportSpecifier(r)&&r.propertyName||e.isNamespaceImport(r)||e.isNamespaceExport(r))return r.name;if(e.isExportDeclaration(r)&&r.exportClause&&e.isNamespaceExport(r.exportClause))return r.exportClause.name}if(99===n.kind&&e.isImportDeclaration(r)&&(i=D(r,t)))return i;if(92===n.kind){var i;if(e.isExportDeclaration(r))if(i=C(r,t))return i;if(e.isExportAssignment(r))return e.skipOuterExpressions(r.expression)}if(143===n.kind&&e.isExternalModuleReference(r))return r.expression;if(153===n.kind&&(e.isImportDeclaration(r)||e.isExportDeclaration(r))&&r.moduleSpecifier)return r.moduleSpecifier;if((93===n.kind||116===n.kind)&&e.isHeritageClause(r)&&r.token===n.kind&&(i=function(e){if(1===e.types.length)return e.types[0].expression}(r)))return i;if(93===n.kind){if(e.isTypeParameterDeclaration(r)&&r.constraint&&e.isTypeReferenceNode(r.constraint))return r.constraint.typeName;if(e.isConditionalTypeNode(r)&&e.isTypeReferenceNode(r.extendsType))return r.extendsType.typeName}if(135===n.kind&&e.isInferTypeNode(r))return r.typeParameter.name;if(100===n.kind&&e.isTypeParameterDeclaration(r)&&e.isMappedTypeNode(r.parent))return r.name;if(138===n.kind&&e.isTypeOperatorNode(r)&&138===r.operator&&e.isTypeReferenceNode(r.type))return r.type.typeName;if(142===n.kind&&e.isTypeOperatorNode(r)&&142===r.operator&&e.isArrayTypeNode(r.type)&&e.isTypeReferenceNode(r.type.elementType))return r.type.elementType.typeName;if(!t){if((102===n.kind&&e.isNewExpression(r)||113===n.kind&&e.isVoidExpression(r)||111===n.kind&&e.isTypeOfExpression(r)||130===n.kind&&e.isAwaitExpression(r)||124===n.kind&&e.isYieldExpression(r)||88===n.kind&&e.isDeleteExpression(r))&&r.expression)return e.skipOuterExpressions(r.expression);if((100===n.kind||101===n.kind)&&e.isBinaryExpression(r)&&r.operatorToken===n)return e.skipOuterExpressions(r.right);if(126===n.kind&&e.isAsExpression(r)&&e.isTypeReferenceNode(r.type))return r.type.typeName;if(100===n.kind&&e.isForInStatement(r)||156===n.kind&&e.isForOfStatement(r))return e.skipOuterExpressions(r.expression)}return n}function N(e){return I(e,!1)}function R(e,n,t){return O(e,n,!1,t,!1)}function k(e,n){return O(e,n,!0,void 0,!1)}function O(e,n,t,r,a){var i=e;e:for(;;){for(var o=0,s=i.getChildren(e);o<s.length;o++){var c=s[o];if((t?c.getFullStart():c.getStart(e,!0))>n)break;var l=c.getEnd();if(n<l||n===l&&(1===c.kind||a)){i=c;continue e}if(r&&l===n){var u=M(n,e,c);if(u&&r(u))return u}}return i}}function P(n,t,r){return function t(a){if(e.isToken(a)&&a.pos===n.end)return a;return e.firstDefined(a.getChildren(r),(function(e){return(e.pos<=n.pos&&e.end>n.end||e.pos===n.end)&&W(e,r)?t(e):void 0}))}(t)}function M(n,t,r,a){var i=function i(o){if(w(o)&&1!==o.kind)return o;var s=o.getChildren(t),c=e.binarySearchKey(s,n,(function(e,n){return n}),(function(e,t){return n<s[e].end?!s[e-1]||n>=s[e-1].end?0:1:-1}));if(c>=0&&s[c]){var l=s[c];if(n<l.end){if(l.getStart(t,!a)>=n||!W(l,t)||B(l)){var u=G(s,c,t);return u&&F(u,t)}return i(l)}}e.Debug.assert(void 0!==r||297===o.kind||1===o.kind||e.isJSDocCommentContainingNode(o));var d=G(s,s.length,t);return d&&F(d,t)}(r||t);return e.Debug.assert(!(i&&B(i))),i}function w(n){return e.isToken(n)&&!B(n)}function F(e,n){if(w(e))return e;var t=e.getChildren(n);if(0===t.length)return e;var r=G(t,t.length,n);return r&&F(r,n)}function G(n,t,r){for(var a=t-1;a>=0;a--){if(B(n[a]))e.Debug.assert(a>0,"`JsxText` tokens should not be the first child of `JsxElement | JsxSelfClosingElement`");else if(W(n[a],r))return n[a]}}function B(n){return e.isJsxText(n)&&n.containsOnlyTriviaWhiteSpaces}function U(e,n,t){for(var r=e.kind,a=0;;){var i=M(e.getFullStart(),t);if(!i)return;if((e=i).kind===n){if(0===a)return e;a--}else e.kind===r&&a++}}function V(e,n,t){return n?e.getNonNullableType():t?e.getNonOptionalType():e}function j(n,t,r){var a=r.getTypeAtLocation(n);return e.isOptionalChain(n.parent)&&(a=V(a,e.isOptionalChainRoot(n.parent),!0)),(e.isNewExpression(n.parent)?a.getConstructSignatures():a.getCallSignatures()).filter((function(e){return!!e.typeParameters&&e.typeParameters.length>=t}))}function H(n,t){for(var r=n,a=0,i=0;r;){switch(r.kind){case 29:if((r=M(r.getFullStart(),t))&&28===r.kind&&(r=M(r.getFullStart(),t)),!r||!e.isIdentifier(r))return;if(!a)return e.isDeclarationName(r)?void 0:{called:r,nTypeArguments:i};a--;break;case 49:a=3;break;case 48:a=2;break;case 31:a++;break;case 19:if(!(r=U(r,18,t)))return;break;case 21:if(!(r=U(r,20,t)))return;break;case 23:if(!(r=U(r,22,t)))return;break;case 27:i++;break;case 38:case 78:case 10:case 8:case 9:case 109:case 94:case 111:case 93:case 138:case 24:case 51:case 57:case 58:break;default:if(e.isTypeNode(r))break;return}r=M(r.getFullStart(),t)}}function K(n,t,r){return e.formatting.getRangeOfEnclosingComment(n,t,void 0,r)}function W(e,n){return 1===e.kind?!!e.jsDoc:0!==e.getWidth(n)}function z(e,n,t){var r=K(e,n,void 0);return!!r&&t===m.test(e.text.substring(r.pos,r.end))}function q(n,t,r){return e.createTextSpanFromBounds(n.getStart(t),(r||n).getEnd())}function J(n){if(!n.isUnterminated)return e.createTextSpanFromBounds(n.getStart()+1,n.getEnd()-1)}function X(e,n){return{span:e,newText:n}}function Y(e){return 149===e.kind}function Q(n,t){return{fileExists:function(e){return n.fileExists(e)},getCurrentDirectory:function(){return t.getCurrentDirectory()},readFile:e.maybeBind(t,t.readFile),useCaseSensitiveFileNames:e.maybeBind(t,t.useCaseSensitiveFileNames),getSymlinkCache:e.maybeBind(t,t.getSymlinkCache)||n.getSymlinkCache,getGlobalTypingsCacheLocation:e.maybeBind(t,t.getGlobalTypingsCacheLocation),getSourceFiles:function(){return n.getSourceFiles()},redirectTargetsMap:n.redirectTargetsMap,getProjectReferenceRedirect:function(e){return n.getProjectReferenceRedirect(e)},isSourceOfProjectReferenceRedirect:function(e){return n.isSourceOfProjectReferenceRedirect(e)}}}function Z(e,n){return __assign(__assign({},Q(e,n)),{getCommonSourceDirectory:function(){return e.getCommonSourceDirectory()}})}function $(n,t,r,a,i){return e.factory.createImportDeclaration(void 0,void 0,n||t?e.factory.createImportClause(!!i,n,t&&t.length?e.factory.createNamedImports(t):void 0):void 0,"string"==typeof r?ee(r,a):r)}function ee(n,t){return e.factory.createStringLiteral(n,0===t)}function ne(n,t){return e.isStringDoubleQuoted(n,t)?1:0}function te(n,t){if(t.quotePreference&&"auto"!==t.quotePreference)return"single"===t.quotePreference?0:1;var r=n.imports&&e.find(n.imports,(function(n){return e.isStringLiteral(n)&&!e.nodeIsSynthesized(n.parent)}));return r?ne(r,n):1}function re(n){return"default"!==n.escapedName?n.escapedName:e.firstDefined(n.declarations,(function(n){var t=e.getNameOfDeclaration(n);return t&&78===t.kind?t.escapedText:void 0}))}function ae(n,t,r,a){var i=new e.Map;return function n(o){if(!(96&o.flags&&e.addToSeen(i,e.getSymbolId(o))))return;return e.firstDefined(o.declarations,(function(i){return e.firstDefined(e.getAllSuperTypeNodes(i),(function(i){var o=r.getTypeAtLocation(i),s=o&&o.symbol&&r.getPropertyOfType(o,t);return o&&s&&(e.firstDefined(r.getRootSymbols(s),a)||n(o.symbol))}))}))}(n)}function ie(n,t,r){return e.textSpanContainsPosition(n,t.getStart(r))&&t.getEnd()<=e.textSpanEnd(n)}function oe(e,n){return!!e&&!!n&&e.start===n.start&&e.length===n.length}function se(e){return e.declarations&&e.declarations.length>0&&160===e.declarations[0].kind}e.getLineStartPositionForPosition=function(n,t){return e.getLineStarts(t)[t.getLineAndCharacterOfPosition(n).line]},e.rangeContainsRange=f,e.rangeContainsRangeExclusive=function(e,n){return _(e,n.pos)&&_(e,n.end)},e.rangeContainsPosition=function(e,n){return e.pos<=n&&n<=e.end},e.rangeContainsPositionExclusive=_,e.startEndContainsRange=g,e.rangeContainsStartEnd=function(e,n,t){return e.pos<=n&&e.end>=t},e.rangeOverlapsWithStartEnd=function(e,n,t){return y(e.pos,e.end,n,t)},e.nodeOverlapsWithStartEnd=function(e,n,t,r){return y(e.getStart(n),e.end,t,r)},e.startEndOverlapsWithStartEnd=y,e.positionBelongsToNode=function(n,t,r){return e.Debug.assert(n.pos<=t),t<n.end||!v(n,r)},e.findListItemInfo=function(n){var t=T(n);if(t){var r=t.getChildren();return{listItemIndex:e.indexOfNode(r,n),list:t}}},e.hasChildOfKind=b,e.findChildOfKind=E,e.findContainingList=T,e.getAdjustedReferenceLocation=N,e.getAdjustedRenameLocation=function(e){return I(e,!0)},e.getTouchingPropertyName=function(n,t){return R(n,t,(function(n){return e.isPropertyNameLiteral(n)||e.isKeyword(n.kind)||e.isPrivateIdentifier(n)}))},e.getTouchingToken=R,e.getTokenAtPosition=k,e.findTokenOnLeftOfPosition=function(n,t){var r=k(n,t);return e.isToken(r)&&t>r.getStart(n)&&t<r.getEnd()?r:M(t,n)},e.findNextToken=P,e.findPrecedingToken=M,e.isInString=function(n,t,r){if(void 0===r&&(r=M(t,n)),r&&e.isStringTextContainingNode(r)){var a=r.getStart(n),i=r.getEnd();if(a<t&&t<i)return!0;if(t===i)return!!r.isUnterminated}return!1},e.isInsideJsxElementOrAttribute=function(e,n){var t=k(e,n);return!!t&&(11===t.kind||(29===t.kind&&11===t.parent.kind||(29===t.kind&&283===t.parent.kind||(!(!t||19!==t.kind||283!==t.parent.kind)||29===t.kind&&276===t.parent.kind))))},e.isInTemplateString=function(n,t){var r=k(n,t);return e.isTemplateLiteralKind(r.kind)&&t>r.getStart(n)},e.isInJSXText=function(n,t){var r=k(n,t);return!!e.isJsxText(r)||(!(18!==r.kind||!e.isJsxExpression(r.parent)||!e.isJsxElement(r.parent.parent))||!(29!==r.kind||!e.isJsxOpeningLikeElement(r.parent)||!e.isJsxElement(r.parent.parent)))},e.isInsideJsxElement=function(e,n){return function(t){for(;t;)if(t.kind>=274&&t.kind<=283||11===t.kind||29===t.kind||31===t.kind||78===t.kind||19===t.kind||18===t.kind||43===t.kind)t=t.parent;else{if(273!==t.kind)return!1;if(n>t.getStart(e))return!0;t=t.parent}return!1}(k(e,n))},e.findPrecedingMatchingToken=U,e.removeOptionality=V,e.isPossiblyTypeArgumentPosition=function n(t,r,a){var i=H(t,r);return void 0!==i&&(e.isPartOfTypeNode(i.called)||0!==j(i.called,i.nTypeArguments,a).length||n(i.called,r,a))},e.getPossibleGenericSignatures=j,e.getPossibleTypeArgumentsInfo=H,e.isInComment=K,e.hasDocComment=function(n,t){var r=k(n,t);return!!e.findAncestor(r,e.isJSDoc)},e.getNodeModifiers=function(n){var t=e.isDeclaration(n)?e.getCombinedNodeFlagsAlwaysIncludeJSDoc(n):0,r=[];return 8&t&&r.push("private"),16&t&&r.push("protected"),4&t&&r.push("public"),32&t&&r.push("static"),128&t&&r.push("abstract"),1&t&&r.push("export"),8192&t&&r.push("deprecated"),8388608&n.flags&&r.push("declare"),266===n.kind&&r.push("export"),r.length>0?r.join(","):""},e.getTypeArgumentOrTypeParameterList=function(n){return 173===n.kind||203===n.kind?n.typeArguments:e.isFunctionLike(n)||252===n.kind||253===n.kind?n.typeParameters:void 0},e.isComment=function(e){return 2===e||3===e},e.isStringOrRegularExpressionOrTemplateLiteral=function(n){return!(10!==n&&13!==n&&!e.isTemplateLiteralKind(n))},e.isPunctuation=function(e){return 18<=e&&e<=77},e.isInsideTemplateLiteral=function(n,t,r){return e.isTemplateLiteralKind(n.kind)&&n.getStart(r)<t&&t<n.end||!!n.isUnterminated&&t===n.end},e.isAccessibilityModifier=function(e){switch(e){case 122:case 120:case 121:return!0}return!1},e.cloneCompilerOptions=function(n){var t=e.clone(n);return e.setConfigFileInOptions(t,n&&n.configFile),t},e.isArrayLiteralOrObjectLiteralDestructuringPattern=function e(n){if(199===n.kind||200===n.kind){if(216===n.parent.kind&&n.parent.left===n&&62===n.parent.operatorToken.kind)return!0;if(239===n.parent.kind&&n.parent.initializer===n)return!0;if(e(288===n.parent.kind?n.parent.parent:n.parent))return!0}return!1},e.isInReferenceComment=function(e,n){return z(e,n,!0)},e.isInNonReferenceComment=function(e,n){return z(e,n,!1)},e.getReplacementSpanForContextToken=function(e){if(e)switch(e.kind){case 10:case 14:return J(e);default:return q(e)}},e.createTextSpanFromNode=q,e.createTextSpanFromStringLiteralLikeContent=J,e.createTextRangeFromNode=function(n,t){return e.createRange(n.getStart(t),n.end)},e.createTextSpanFromRange=function(n){return e.createTextSpanFromBounds(n.pos,n.end)},e.createTextRangeFromSpan=function(n){return e.createRange(n.start,n.start+n.length)},e.createTextChangeFromStartLength=function(n,t,r){return X(e.createTextSpan(n,t),r)},e.createTextChange=X,e.typeKeywords=[128,127,155,131,94,138,141,103,144,145,142,147,148,109,113,150,151,152],e.isTypeKeyword=function(n){return e.contains(e.typeKeywords,n)},e.isTypeKeywordToken=Y,e.isExternalModuleSymbol=function(e){return!!(1536&e.flags)&&34===e.name.charCodeAt(0)},e.nodeSeenTracker=function(){var n=[];return function(t){var r=e.getNodeId(t);return!n[r]&&(n[r]=!0)}},e.getSnapshotText=function(e){return e.getText(0,e.getLength())},e.repeatString=function(e,n){for(var t="",r=0;r<n;r++)t+=e;return t},e.skipConstraint=function(e){return e.isTypeParameter()&&e.getConstraint()||e},e.getNameFromPropertyName=function(n){return 158===n.kind?e.isStringOrNumericLiteralLike(n.expression)?n.expression.text:void 0:e.isPrivateIdentifier(n)?e.idText(n):e.getTextOfIdentifierOrLiteral(n)},e.programContainsModules=function(e){return e.getSourceFiles().some((function(n){return!(n.isDeclarationFile||e.isSourceFileFromExternalLibrary(n)||!n.externalModuleIndicator&&!n.commonJsModuleIndicator)}))},e.programContainsEs6Modules=function(e){return e.getSourceFiles().some((function(n){return!n.isDeclarationFile&&!e.isSourceFileFromExternalLibrary(n)&&!!n.externalModuleIndicator}))},e.compilerOptionsIndicateEs6Modules=function(e){return!!e.module||e.target>=2||!!e.noEmit},e.createModuleSpecifierResolutionHost=Q,e.getModuleSpecifierResolverHost=Z,e.makeImportIfNecessary=function(e,n,t,r){return e||n&&n.length?$(e,n,t,r):void 0},e.makeImport=$,e.makeStringLiteral=ee,function(e){e[e.Single=0]="Single",e[e.Double=1]="Double"}(e.QuotePreference||(e.QuotePreference={})),e.quotePreferenceFromString=ne,e.getQuotePreference=te,e.getQuoteFromPreference=function(n){switch(n){case 0:return"'";case 1:return'"';default:return e.Debug.assertNever(n)}},e.symbolNameNoDefault=function(n){var t=re(n);return void 0===t?void 0:e.unescapeLeadingUnderscores(t)},e.symbolEscapedNameNoDefault=re,e.isObjectBindingElementWithoutPropertyName=function(n){return e.isBindingElement(n)&&e.isObjectBindingPattern(n.parent)&&e.isIdentifier(n.name)&&!n.propertyName},e.getPropertySymbolFromBindingElement=function(e,n){var t=e.getTypeAtLocation(n.parent);return t&&e.getPropertyOfType(t,n.name.text)},e.getPropertySymbolsFromBaseTypes=ae,e.isMemberSymbolInBaseType=function(e,n){return ae(e.parent,e.name,n,(function(e){return!0}))||!1},e.getParentNodeInSpan=function(n,t,r){if(n)for(;n.parent;){if(e.isSourceFile(n.parent)||!ie(r,n.parent,t))return n;n=n.parent}},e.findModifier=function(n,t){return n.modifiers&&e.find(n.modifiers,(function(e){return e.kind===t}))},e.insertImports=function(n,t,r,a){var i=232===(e.isArray(r)?r[0]:r).kind?e.isRequireVariableStatement:e.isAnyImportSyntax,o=e.filter(t.statements,i),s=e.isArray(r)?e.stableSort(r,e.OrganizeImports.compareImportsOrRequireStatements):[r];if(o.length)if(o&&e.OrganizeImports.importsAreSorted(o))for(var c=0,l=s;c<l.length;c++){var u=l[c],d=e.OrganizeImports.getImportDeclarationInsertionIndex(o,u);if(0===d){var p=o[0]===t.statements[0]?{leadingTriviaOption:e.textChanges.LeadingTriviaOption.Exclude}:{};n.insertNodeBefore(t,o[0],u,!1,p)}else{var m=o[d-1];n.insertNodeAfter(t,m,u)}}else{var f=e.lastOrUndefined(o);f?n.insertNodesAfter(t,f,s):n.insertNodesAtTopOfFile(t,s,a)}else n.insertNodesAtTopOfFile(t,s,a)},e.getTypeKeywordOfTypeOnlyImport=function(n,t){return e.Debug.assert(n.isTypeOnly),e.cast(n.getChildAt(0,t),Y)},e.textSpansEqual=oe,e.documentSpansEqual=function(e,n){return e.fileName===n.fileName&&oe(e.textSpan,n.textSpan)},e.forEachUnique=function(e,n){if(e)for(var t=0;t<e.length;t++)if(e.indexOf(e[t])===t){var r=n(e[t],t);if(r)return r}},e.isTextWhiteSpaceLike=function(n,t,r){for(var a=t;a<r;a++)if(!e.isWhiteSpaceLike(n.charCodeAt(a)))return!1;return!0},e.isFirstDeclarationOfSymbolParameter=se;var ce=function(){var n,t,r,a,i=10*e.defaultMaximumTruncationLength;l();var o=function(n){return c(n,e.SymbolDisplayPartKind.text)};return{displayParts:function(){var t=n.length&&n[n.length-1].text;return a>i&&t&&"..."!==t&&(e.isWhiteSpaceLike(t.charCodeAt(t.length-1))||n.push(ue(" ",e.SymbolDisplayPartKind.space)),n.push(ue("...",e.SymbolDisplayPartKind.punctuation))),n},writeKeyword:function(n){return c(n,e.SymbolDisplayPartKind.keyword)},writeOperator:function(n){return c(n,e.SymbolDisplayPartKind.operator)},writePunctuation:function(n){return c(n,e.SymbolDisplayPartKind.punctuation)},writeTrailingSemicolon:function(n){return c(n,e.SymbolDisplayPartKind.punctuation)},writeSpace:function(n){return c(n,e.SymbolDisplayPartKind.space)},writeStringLiteral:function(n){return c(n,e.SymbolDisplayPartKind.stringLiteral)},writeParameter:function(n){return c(n,e.SymbolDisplayPartKind.parameterName)},writeProperty:function(n){return c(n,e.SymbolDisplayPartKind.propertyName)},writeLiteral:function(n){return c(n,e.SymbolDisplayPartKind.stringLiteral)},writeSymbol:function(e,t){if(a>i)return;s(),a+=e.length,n.push(le(e,t))},writeLine:function(){if(a>i)return;a+=1,n.push(me()),t=!0},write:o,writeComment:o,getText:function(){return""},getTextPos:function(){return 0},getColumn:function(){return 0},getLine:function(){return 0},isAtStartOfLine:function(){return!1},hasTrailingWhitespace:function(){return!1},hasTrailingComment:function(){return!1},rawWrite:e.notImplemented,getIndent:function(){return r},increaseIndent:function(){r++},decreaseIndent:function(){r--},clear:l,trackSymbol:e.noop,reportInaccessibleThisError:e.noop,reportInaccessibleUniqueSymbolError:e.noop,reportPrivateInBaseOfClassExpression:e.noop};function s(){if(!(a>i)&&t){var o=e.getIndentString(r);o&&(a+=o.length,n.push(ue(o,e.SymbolDisplayPartKind.space))),t=!1}}function c(e,t){a>i||(s(),a+=e.length,n.push(ue(e,t)))}function l(){n=[],t=!0,r=0,a=0}}();function le(n,t){return ue(n,function(n){var t=n.flags;if(3&t)return se(n)?e.SymbolDisplayPartKind.parameterName:e.SymbolDisplayPartKind.localName;if(4&t)return e.SymbolDisplayPartKind.propertyName;if(32768&t)return e.SymbolDisplayPartKind.propertyName;if(65536&t)return e.SymbolDisplayPartKind.propertyName;if(8&t)return e.SymbolDisplayPartKind.enumMemberName;if(16&t)return e.SymbolDisplayPartKind.functionName;if(32&t)return e.SymbolDisplayPartKind.className;if(64&t)return e.SymbolDisplayPartKind.interfaceName;if(384&t)return e.SymbolDisplayPartKind.enumName;if(1536&t)return e.SymbolDisplayPartKind.moduleName;if(8192&t)return e.SymbolDisplayPartKind.methodName;if(262144&t)return e.SymbolDisplayPartKind.typeParameterName;if(524288&t)return e.SymbolDisplayPartKind.aliasName;if(2097152&t)return e.SymbolDisplayPartKind.aliasName;return e.SymbolDisplayPartKind.text}(t))}function ue(n,t){return{text:n,kind:e.SymbolDisplayPartKind[t]}}function de(n){return ue(e.tokenToString(n),e.SymbolDisplayPartKind.keyword)}function pe(n){return ue(n,e.SymbolDisplayPartKind.text)}e.symbolPart=le,e.displayPart=ue,e.spacePart=function(){return ue(" ",e.SymbolDisplayPartKind.space)},e.keywordPart=de,e.punctuationPart=function(n){return ue(e.tokenToString(n),e.SymbolDisplayPartKind.punctuation)},e.operatorPart=function(n){return ue(e.tokenToString(n),e.SymbolDisplayPartKind.operator)},e.textOrKeywordPart=function(n){var t=e.stringToToken(n);return void 0===t?pe(n):de(t)},e.textPart=pe;function me(){return ue("\n",e.SymbolDisplayPartKind.lineBreak)}function fe(e){try{return e(ce),ce.displayParts()}finally{ce.clear()}}function _e(e){return 0!=(33554432&e.flags)}function ge(e){return 0!=(2097152&e.flags)}function ye(e,n){void 0===n&&(n=!0);var t=e&&he(e);return t&&!n&&be(t),t}function ve(n,t,r){var a=r(n);return a?e.setOriginalNode(a,n):a=he(n,r),a&&!t&&be(a),a}function he(n,t){var r=t?e.visitEachChild(n,(function(e){return ve(e,!0,t)}),e.nullTransformationContext):e.visitEachChild(n,ye,e.nullTransformationContext);if(r===n){var a=e.isStringLiteral(n)?e.setOriginalNode(e.factory.createStringLiteralFromNode(n),n):e.isNumericLiteral(n)?e.setOriginalNode(e.factory.createNumericLiteral(n.text,n.numericLiteralFlags),n):e.factory.cloneNode(n);return e.setTextRange(a,n)}return r.parent=void 0,r}function be(e){Ee(e),Te(e)}function Ee(e){Se(e,512,xe)}function Te(n){Se(n,1024,e.getLastChild)}function Se(n,t,r){e.addEmitFlags(n,t);var a=r(n);a&&Se(a,t,r)}function xe(e){return e.forEachChild((function(e){return e}))}function Le(n,t,r,a,i){e.forEachLeadingCommentRange(r.text,n.pos,Ce(t,r,a,i,e.addSyntheticLeadingComment))}function Ae(n,t,r,a,i){e.forEachTrailingCommentRange(r.text,n.end,Ce(t,r,a,i,e.addSyntheticTrailingComment))}function De(n,t,r,a,i){e.forEachTrailingCommentRange(r.text,n.pos,Ce(t,r,a,i,e.addSyntheticLeadingComment))}function Ce(e,n,t,r,a){return function(i,o,s,c){3===s?(i+=2,o-=2):i+=2,a(e,t||s,n.text.slice(i,o),void 0!==r?r:c)}}function Ie(n,t){if(e.startsWith(n,t))return 0;var r=n.indexOf(" "+t);return-1===r&&(r=n.indexOf("."+t)),-1===r&&(r=n.indexOf('"'+t)),-1===r?-1:r+1}function Ne(e){switch(e){case 36:case 34:case 37:case 35:return!0;default:return!1}}function Re(e,n){return n.getTypeAtLocation(e.parent.parent.expression)}function ke(e){return 169===e||170===e||171===e||162===e||164===e}function Oe(e){return 251===e||166===e||165===e||167===e||168===e}function Pe(e){return 256===e}function Me(e){return 232===e||233===e||235===e||240===e||241===e||242===e||246===e||248===e||163===e||254===e||261===e||260===e||267===e||259===e||266===e}function we(e,n){return Ge(e,e.fileExists,n)}function Fe(e){try{return e()}catch(e){return}}function Ge(e,n){for(var t=[],r=2;r<arguments.length;r++)t[r-2]=arguments[r];return Fe((function(){return n&&n.apply(e,t)}))}function Be(n,t){if(t.readFile){var r=function(e){try{return JSON.parse(e)}catch(e){return}}(t.readFile(n)||""),a={};if(r)for(var i=0,o=["dependencies","devDependencies","optionalDependencies","peerDependencies"];i<o.length;i++){var s=o[i],c=r[s];if(c){var l=new e.Map;for(var u in c)l.set(u,c[u]);a[s]=l}}var d=[[1,a.dependencies],[2,a.devDependencies],[8,a.optionalDependencies],[4,a.peerDependencies]];return __assign(__assign({},a),{parseable:!!r,fileName:n,get:p,has:function(e,n){return!!p(e,n)}})}function p(e,n){void 0===n&&(n=15);for(var t=0,r=d;t<r.length;t++){var a=r[t],i=a[0],o=a[1];if(o&&n&i){var s=o.get(e);if(void 0!==s)return s}}}}function Ue(e){return void 0!==e.file&&void 0!==e.start&&void 0!==e.length}function Ve(n){var t=n.getSourceFile();return!(!t.externalModuleIndicator&&!t.commonJsModuleIndicator)&&(e.isInJSFile(n)||!e.findAncestor(n,e.isGlobalScopeAugmentation))}e.getNewLineOrDefaultFromHost=function(e,n){var t;return(null==n?void 0:n.newLineCharacter)||(null===(t=e.getNewLine)||void 0===t?void 0:t.call(e))||"\r\n"},e.lineBreakPart=me,e.mapToDisplayParts=fe,e.typeToDisplayParts=function(e,n,t,r){return void 0===r&&(r=0),fe((function(a){e.writeType(n,t,17408|r,a)}))},e.symbolToDisplayParts=function(e,n,t,r,a){return void 0===a&&(a=0),fe((function(i){e.writeSymbol(n,t,r,8|a,i)}))},e.signatureToDisplayParts=function(e,n,t,r){return void 0===r&&(r=0),r|=25632,fe((function(a){e.writeSignature(n,t,r,void 0,a)}))},e.isImportOrExportSpecifierName=function(n){return!!n.parent&&e.isImportOrExportSpecifier(n.parent)&&n.parent.propertyName===n},e.getScriptKind=function(n,t){return e.ensureScriptKind(n,t.getScriptKind&&t.getScriptKind(n))},e.getSymbolTarget=function(n,t){for(var r=n;ge(r)||_e(r)&&r.target;)r=_e(r)&&r.target?r.target:e.skipAlias(r,t);return r},e.getUniqueSymbolId=function(n,t){return e.getSymbolId(e.skipAlias(n,t))},e.getFirstNonSpaceCharacterPosition=function(n,t){for(;e.isWhiteSpaceLike(n.charCodeAt(t));)t+=1;return t},e.getPrecedingNonSpaceCharacterPosition=function(n,t){for(;t>-1&&e.isWhiteSpaceSingleLine(n.charCodeAt(t));)t-=1;return t+1},e.getSynthesizedDeepClone=ye,e.getSynthesizedDeepCloneWithReplacements=ve,e.getSynthesizedDeepClones=function(n,t){return void 0===t&&(t=!0),n&&e.factory.createNodeArray(n.map((function(e){return ye(e,t)})),n.hasTrailingComma)},e.getSynthesizedDeepClonesWithReplacements=function(n,t,r){return e.factory.createNodeArray(n.map((function(e){return ve(e,t,r)})),n.hasTrailingComma)},e.suppressLeadingAndTrailingTrivia=be,e.suppressLeadingTrivia=Ee,e.suppressTrailingTrivia=Te,e.copyComments=function(e,n){var t=e.getSourceFile();!function(e,n){for(var t=e.getFullStart(),r=e.getStart(),a=t;a<r;a++)if(10===n.charCodeAt(a))return!0;return!1}(e,t.text)?De(e,n,t):Le(e,n,t),Ae(e,n,t)},e.getUniqueName=function(n,t){for(var r=n,a=1;!e.isFileLevelUniqueName(t,r);a++)r=n+"_"+a;return r},e.getRenameLocation=function(n,t,r,a){for(var i=0,o=-1,s=0,c=n;s<c.length;s++){var l=c[s],u=l.fileName,d=l.textChanges;e.Debug.assert(u===t);for(var p=0,m=d;p<m.length;p++){var f=m[p],_=f.span,g=f.newText,y=Ie(g,r);if(-1!==y&&(o=_.start+i+y,!a))return o;i+=g.length-_.length}}return e.Debug.assert(a),e.Debug.assert(o>=0),o},e.copyLeadingComments=Le,e.copyTrailingComments=Ae,e.copyTrailingAsLeadingComments=De,e.needsParentheses=function(n){return e.isBinaryExpression(n)&&27===n.operatorToken.kind||e.isObjectLiteralExpression(n)},e.getContextualTypeFromParent=function(e,n){var t=e.parent;switch(t.kind){case 204:return n.getContextualType(t);case 216:var r=t,a=r.left,i=r.operatorToken,o=r.right;return Ne(i.kind)?n.getTypeAtLocation(e===o?a:o):n.getContextualType(e);case 284:return t.expression===e?Re(t,n):void 0;default:return n.getContextualType(e)}},e.quote=function(n,t,r){var a=te(n,t),i=JSON.stringify(r);return 0===a?"'"+e.stripQuotes(i).replace("'","\\'").replace('\\"','"')+"'":i},e.isEqualityOperatorKind=Ne,e.isStringLiteralOrTemplate=function(e){switch(e.kind){case 10:case 14:case 218:case 205:return!0;default:return!1}},e.hasIndexSignature=function(e){return!!e.getStringIndexType()||!!e.getNumberIndexType()},e.getSwitchedType=Re,e.ANONYMOUS="anonymous function",e.getTypeNodeIfAccessible=function(e,n,t,r){var a=t.getTypeChecker(),i=!0,o=function(){i=!1},s=a.typeToTypeNode(e,n,1,{trackSymbol:function(e,n,t){i=i&&0===a.isSymbolAccessible(e,n,t,!1).accessibility},reportInaccessibleThisError:o,reportPrivateInBaseOfClassExpression:o,reportInaccessibleUniqueSymbolError:o,moduleResolverHost:Z(t,r)});return i?s:void 0},e.syntaxRequiresTrailingCommaOrSemicolonOrASI=ke,e.syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI=Oe,e.syntaxRequiresTrailingModuleBlockOrSemicolonOrASI=Pe,e.syntaxRequiresTrailingSemicolonOrASI=Me,e.syntaxMayBeASICandidate=e.or(ke,Oe,Pe,Me),e.positionIsASICandidate=function(n,t,r){var a=e.findAncestor(t,(function(t){return t.end!==n?"quit":e.syntaxMayBeASICandidate(t.kind)}));return!!a&&function(n,t){var r=n.getLastToken(t);if(r&&26===r.kind)return!1;if(ke(n.kind)){if(r&&27===r.kind)return!1}else if(Pe(n.kind)){if((a=e.last(n.getChildren(t)))&&e.isModuleBlock(a))return!1}else if(Oe(n.kind)){var a;if((a=e.last(n.getChildren(t)))&&e.isFunctionBlock(a))return!1}else if(!Me(n.kind))return!1;if(235===n.kind)return!0;var i=P(n,e.findAncestor(n,(function(e){return!e.parent})),t);return!i||19===i.kind||t.getLineAndCharacterOfPosition(n.getEnd()).line!==t.getLineAndCharacterOfPosition(i.getStart(t)).line}(a,r)},e.probablyUsesSemicolons=function(n){var t=0,r=0;return e.forEachChild(n,(function a(i){if(Me(i.kind)){var o=i.getLastToken(n);o&&26===o.kind?t++:r++}return t+r>=5||e.forEachChild(i,a)})),0===t&&r<=1||t/r>.2},e.tryGetDirectories=function(e,n){return Ge(e,e.getDirectories,n)||[]},e.tryReadDirectory=function(n,t,r,a,i){return Ge(n,n.readDirectory,t,r,a,i)||e.emptyArray},e.tryFileExists=we,e.tryDirectoryExists=function(n,t){return Fe((function(){return e.directoryProbablyExists(t,n)}))||!1},e.tryAndIgnoreErrors=Fe,e.tryIOAndConsumeErrors=Ge,e.findPackageJsons=function(n,t,r){var a=[];return e.forEachAncestorDirectory(n,(function(n){if(n===r)return!0;var i=e.combinePaths(n,"package.json");we(t,i)&&a.push(i)})),a},e.findPackageJson=function(n,t){var r;return e.forEachAncestorDirectory(n,(function(n){return"node_modules"===n||(!!(r=e.findConfigFile(n,(function(e){return we(t,e)}),"package.json"))||void 0)})),r},e.getPackageJsonsVisibleToFile=function(n,t){if(!t.fileExists)return[];var r=[];return e.forEachAncestorDirectory(e.getDirectoryPath(n),(function(n){var a=e.combinePaths(n,"package.json");if(t.fileExists(a)){var i=Be(a,t);i&&r.push(i)}})),r},e.createPackageJsonInfo=Be,e.consumesNodeCoreModules=function(n){return e.some(n.imports,(function(n){var t=n.text;return e.JsTyping.nodeCoreModules.has(t)}))},e.isInsideNodeModules=function(n){return e.contains(e.getPathComponents(n),"node_modules")},e.isDiagnosticWithLocation=Ue,e.findDiagnosticForNode=function(n,t){var r=q(n),a=e.binarySearchKey(t,r,e.identity,e.compareTextSpans);if(a>=0){var i=t[a];return e.Debug.assertEqual(i.file,n.getSourceFile(),"Diagnostics proided to 'findDiagnosticForNode' must be from a single SourceFile"),e.cast(i,Ue)}},e.getDiagnosticsWithinSpan=function(n,t){var r,a=e.binarySearchKey(t,n.start,(function(e){return e.start}),e.compareValues);for(a<0&&(a=~a);(null===(r=t[a-1])||void 0===r?void 0:r.start)===n.start;)a--;for(var i=[],o=e.textSpanEnd(n);;){var s=e.tryCast(t[a],Ue);if(!s||s.start>o)break;e.textSpanContainsTextSpan(n,s)&&i.push(s),a++}return i},e.getRefactorContextSpan=function(n){var t=n.startPosition,r=n.endPosition;return e.createTextSpanFromBounds(t,void 0===r?t:r)},e.mapOneOrMany=function(n,t,r){return void 0===r&&(r=e.identity),n?e.isArray(n)?r(e.map(n,t)):t(n,0):void 0},e.firstOrOnly=function(n){return e.isArray(n)?e.first(n):n},e.getNameForExportedSymbol=function(n,t){return"export="===n.escapedName||"default"===n.escapedName?e.firstDefined(n.declarations,(function(n){return e.isExportAssignment(n)&&e.isIdentifier(n.expression)?n.expression.text:void 0}))||e.codefix.moduleSymbolToValidIdentifier(function(n){var t;return e.Debug.checkDefined(n.parent,"Symbol parent was undefined. Flags: "+e.Debug.formatSymbolFlags(n.flags)+". Declarations: "+(null===(t=n.declarations)||void 0===t?void 0:t.map((function(n){var t=e.Debug.formatSyntaxKind(n.kind),r=e.isInJSFile(n),a=n.expression;return(r?"[JS]":"")+t+(a?" (expression: "+e.Debug.formatSyntaxKind(a.kind)+")":"")})).join(", "))+".")}(n),t):n.name},e.stringContainsAt=function(e,n,t){var r=n.length;if(r+t>e.length)return!1;for(var a=0;a<r;a++)if(n.charCodeAt(a)!==e.charCodeAt(a+t))return!1;return!0},e.startsWithUnderscore=function(e){return 95===e.charCodeAt(0)},e.isGlobalDeclaration=function(e){return!Ve(e)},e.isNonGlobalDeclaration=Ve}(ts||(ts={})),function(e){e.createClassifier=function(){var o=e.createScanner(99,!1);function s(a,s,c){var l=0,u=0,d=[],p=function(n){switch(n){case 3:return{prefix:'"\\\n'};case 2:return{prefix:"'\\\n"};case 1:return{prefix:"/*\n"};case 4:return{prefix:"`\n"};case 5:return{prefix:"}\n",pushTemplate:!0};case 6:return{prefix:"",pushTemplate:!0};case 0:return{prefix:""};default:return e.Debug.assertNever(n)}}(s),m=p.prefix,f=p.pushTemplate;a=m+a;var _=m.length;f&&d.push(15),o.setText(a);var g=0,y=[],v=0;do{l=o.scan(),e.isTrivia(l)||(E(),u=l);var h=o.getTextPos();if(r(o.getTokenPos(),h,_,i(l),y),h>=a.length){var b=t(o,l,e.lastOrUndefined(d));void 0!==b&&(g=b)}}while(1!==l);function E(){switch(l){case 43:case 67:n[u]||13!==o.reScanSlashToken()||(l=13);break;case 29:78===u&&v++;break;case 31:v>0&&v--;break;case 128:case 147:case 144:case 131:case 148:v>0&&!c&&(l=78);break;case 15:d.push(l);break;case 18:d.length>0&&d.push(l);break;case 19:if(d.length>0){var t=e.lastOrUndefined(d);15===t?17===(l=o.reScanTemplateToken(!1))?d.pop():e.Debug.assertEqual(l,16,"Should have been a template middle."):(e.Debug.assertEqual(t,18,"Should have been an open brace"),d.pop())}break;default:if(!e.isKeyword(l))break;(24===u||e.isKeyword(u)&&e.isKeyword(l)&&!function(n,t){if(!e.isAccessibilityModifier(n))return!0;switch(t){case 134:case 146:case 132:case 123:return!0;default:return!1}}(u,l))&&(l=78)}}return{endOfLineState:g,spans:y}}return{getClassificationsForLine:function(n,t,r){return function(n,t){for(var r=[],i=n.spans,o=0,s=0;s<i.length;s+=3){var c=i[s],l=i[s+1],u=i[s+2];if(o>=0){var d=c-o;d>0&&r.push({length:d,classification:e.TokenClass.Whitespace})}r.push({length:l,classification:a(u)}),o=c+l}var p=t.length-o;p>0&&r.push({length:p,classification:e.TokenClass.Whitespace});return{entries:r,finalLexState:n.endOfLineState}}(s(n,t,r),n)},getEncodedLexicalClassifications:s}};var n=e.arrayToNumericMap([78,10,8,9,13,107,45,46,21,23,19,109,94],(function(e){return e}),(function(){return!0}));function t(n,t,r){switch(t){case 10:if(!n.isUnterminated())return;for(var a=n.getTokenText(),i=a.length-1,o=0;92===a.charCodeAt(i-o);)o++;if(0==(1&o))return;return 34===a.charCodeAt(0)?3:2;case 3:return n.isUnterminated()?1:void 0;default:if(e.isTemplateLiteralKind(t)){if(!n.isUnterminated())return;switch(t){case 17:return 5;case 14:return 4;default:return e.Debug.fail("Only 'NoSubstitutionTemplateLiteral's and 'TemplateTail's can be unterminated; got SyntaxKind #"+t)}}return 15===r?6:void 0}}function r(e,n,t,r,a){if(8!==r){0===e&&t>0&&(e+=t);var i=n-e;i>0&&a.push(e-t,i,r)}}function a(n){switch(n){case 1:return e.TokenClass.Comment;case 3:return e.TokenClass.Keyword;case 4:return e.TokenClass.NumberLiteral;case 25:return e.TokenClass.BigIntLiteral;case 5:return e.TokenClass.Operator;case 6:return e.TokenClass.StringLiteral;case 8:return e.TokenClass.Whitespace;case 10:return e.TokenClass.Punctuation;case 2:case 11:case 12:case 13:case 14:case 15:case 16:case 9:case 17:return e.TokenClass.Identifier;default:return}}function i(n){if(e.isKeyword(n))return 3;if(function(e){switch(e){case 41:case 43:case 44:case 39:case 40:case 47:case 48:case 49:case 29:case 31:case 32:case 33:case 101:case 100:case 126:case 34:case 35:case 36:case 37:case 50:case 52:case 51:case 55:case 56:case 73:case 72:case 77:case 69:case 70:case 71:case 63:case 64:case 65:case 67:case 68:case 62:case 27:case 60:case 74:case 75:case 76:return!0;default:return!1}}(n)||function(e){switch(e){case 39:case 40:case 54:case 53:case 45:case 46:return!0;default:return!1}}(n))return 5;if(n>=18&&n<=77)return 10;switch(n){case 8:return 4;case 9:return 25;case 10:return 6;case 13:return 7;case 7:case 3:case 2:return 1;case 5:case 4:return 8;case 78:default:return e.isTemplateLiteralKind(n)?6:2}}function o(e,n){switch(n){case 256:case 252:case 253:case 251:case 221:case 208:case 209:e.throwIfCancellationRequested()}}function s(n,t,r,a,i){var s=[];return r.forEachChild((function l(u){if(u&&e.textSpanIntersectsWith(i,u.pos,u.getFullWidth())){if(o(t,u.kind),e.isIdentifier(u)&&!e.nodeIsMissing(u)&&a.has(u.escapedText)){var d=n.getSymbolAtLocation(u),p=d&&c(d,e.getMeaningFromLocation(u),n);p&&function(n,t,r){var a=t-n;e.Debug.assert(a>0,"Classification had non-positive length of "+a),s.push(n),s.push(a),s.push(r)}(u.getStart(r),u.getEnd(),p)}u.forEachChild(l)}})),{spans:s,endOfLineState:0}}function c(n,t,r){var a=n.getFlags();return 0==(2885600&a)?void 0:32&a?11:384&a?12:524288&a?16:1536&a?4&t||1&t&&function(n){return e.some(n.declarations,(function(n){return e.isModuleDeclaration(n)&&1===e.getModuleInstanceState(n)}))}(n)?14:void 0:2097152&a?c(r.getAliasedSymbol(n),t,r):2&t?64&a?13:262144&a?15:void 0:void 0}function l(e){switch(e){case 1:return"comment";case 2:return"identifier";case 3:return"keyword";case 4:return"number";case 25:return"bigint";case 5:return"operator";case 6:return"string";case 8:return"whitespace";case 9:return"text";case 10:return"punctuation";case 11:return"class name";case 12:return"enum name";case 13:return"interface name";case 14:return"module name";case 15:return"type parameter name";case 16:return"type alias name";case 17:return"parameter name";case 18:return"doc comment tag name";case 19:return"jsx open tag name";case 20:return"jsx close tag name";case 21:return"jsx self closing tag name";case 22:return"jsx attribute";case 23:return"jsx text";case 24:return"jsx attribute string literal value";default:return}}function u(n){e.Debug.assert(n.spans.length%3==0);for(var t=n.spans,r=[],a=0;a<t.length;a+=3)r.push({textSpan:e.createTextSpan(t[a],t[a+1]),classificationType:l(t[a+2])});return r}function d(n,t,r){var a=r.start,i=r.length,s=e.createScanner(99,!1,t.languageVariant,t.text),c=e.createScanner(99,!1,t.languageVariant,t.text),l=[];return v(t),{spans:l,endOfLineState:0};function u(e,n,t){l.push(e),l.push(n),l.push(t)}function d(n,r,a,i){if(3===r){var o=e.parseIsolatedJSDocComment(t.text,a,i);if(o&&o.jsDoc)return e.setParent(o.jsDoc,n),void function(e){var n=e.pos;if(e.tags)for(var t=0,r=e.tags;t<r.length;t++){var a=r[t];switch(a.pos!==n&&p(n,a.pos-n),u(a.pos,1,10),u(a.tagName.pos,a.tagName.end-a.tagName.pos,18),n=a.tagName.end,a.kind){case 326:i(a);break;case 330:m(a),n=a.end;break;case 329:case 327:v(a.typeExpression),n=a.end}}n!==e.end&&p(n,e.end-n);return;function i(e){e.isNameFirst&&(p(n,e.name.pos-n),u(e.name.pos,e.name.end-e.name.pos,17),n=e.name.end),e.typeExpression&&(p(n,e.typeExpression.pos-n),v(e.typeExpression),n=e.typeExpression.end),e.isNameFirst||(p(n,e.name.pos-n),u(e.name.pos,e.name.end-e.name.pos,17),n=e.name.end)}}(o.jsDoc)}else if(2===r&&function(n,r){var a=/^(\/\/\/\s*)(<)(?:(\S+)((?:[^/]|\/[^>])*)(\/>)?)?/im,i=/(\S+)(\s*)(=)(\s*)('[^']+'|"[^"]+")/gim,o=t.text.substr(n,r),s=a.exec(o);if(!s)return!1;if(!s[3]||!(s[3]in e.commentPragmas))return!1;var c=n;p(c,s[1].length),u(c+=s[1].length,s[2].length,10),u(c+=s[2].length,s[3].length,21),c+=s[3].length;var l=s[4],d=c;for(;;){var m=i.exec(l);if(!m)break;var f=c+m.index;f>d&&(p(d,f-d),d=f),u(d,m[1].length,22),d+=m[1].length,m[2].length&&(p(d,m[2].length),d+=m[2].length),u(d,m[3].length,5),d+=m[3].length,m[4].length&&(p(d,m[4].length),d+=m[4].length),u(d,m[5].length,24),d+=m[5].length}(c+=s[4].length)>d&&p(d,c-d);s[5]&&(u(c,s[5].length,10),c+=s[5].length);var _=n+r;c<_&&p(c,_-c);return!0}(a,i))return;p(a,i)}function p(e,n){u(e,n,1)}function m(e){for(var n=0,t=e.getChildren();n<t.length;n++){v(t[n])}}function f(n,t,r){var a;for(a=t;a<r&&!e.isLineBreak(n.charCodeAt(a));a++);for(u(t,a-t,1),c.setTextPos(a);c.getTextPos()<r;)_()}function _(){var e=c.getTextPos(),n=c.scan(),t=c.getTextPos(),r=y(n);r&&u(e,t-e,r)}function g(n){if(e.isJSDoc(n))return!0;if(e.nodeIsMissing(n))return!0;var r=function(e){switch(e.parent&&e.parent.kind){case 275:if(e.parent.tagName===e)return 19;break;case 276:if(e.parent.tagName===e)return 20;break;case 274:if(e.parent.tagName===e)return 21;break;case 280:if(e.parent.name===e)return 22}return}(n);if(!e.isToken(n)&&11!==n.kind&&void 0===r)return!1;var a=11===n.kind?n.pos:function(n){for(s.setTextPos(n.pos);;){var r=s.getTextPos();if(!e.couldStartTrivia(t.text,r))return r;var a=s.scan(),i=s.getTextPos(),o=i-r;if(!e.isTrivia(a))return r;switch(a){case 4:case 5:continue;case 2:case 3:d(n,a,r,o),s.setTextPos(i);continue;case 7:var c=t.text,l=c.charCodeAt(r);if(60===l||62===l){u(r,o,1);continue}e.Debug.assert(124===l||61===l),f(c,r,i);break;case 6:break;default:e.Debug.assertNever(a)}}}(n),i=n.end-a;if(e.Debug.assert(i>=0),i>0){var o=r||y(n.kind,n);o&&u(a,i,o)}return!0}function y(n,t){if(e.isKeyword(n))return 3;if((29===n||31===n)&&t&&e.getTypeArgumentOrTypeParameterList(t.parent))return 10;if(e.isPunctuation(n)){if(t){var r=t.parent;if(62===n&&(249===r.kind||163===r.kind||160===r.kind||280===r.kind))return 5;if(216===r.kind||214===r.kind||215===r.kind||217===r.kind)return 5}return 10}if(8===n)return 4;if(9===n)return 25;if(10===n)return t&&280===t.parent.kind?24:6;if(13===n)return 6;if(e.isTemplateLiteralKind(n))return 6;if(11===n)return 23;if(78===n){if(t)switch(t.parent.kind){case 252:return t.parent.name===t?11:void 0;case 159:return t.parent.name===t?15:void 0;case 253:return t.parent.name===t?13:void 0;case 255:return t.parent.name===t?12:void 0;case 256:return t.parent.name===t?14:void 0;case 160:return t.parent.name===t?e.isThisIdentifier(t)?3:17:void 0}return 2}}function v(r){if(r&&e.decodedTextSpanIntersectsWith(a,i,r.pos,r.getFullWidth())){o(n,r.kind);for(var s=0,c=r.getChildren(t);s<c.length;s++){var l=c[s];g(l)||v(l)}}}}e.getSemanticClassifications=function(e,n,t,r,a){return u(s(e,n,t,r,a))},e.getEncodedSemanticClassifications=s,e.getSyntacticClassifications=function(e,n,t){return u(d(e,n,t))},e.getEncodedSyntacticClassifications=d}(ts||(ts={})),function(e){!function(n){!function(n){function t(e,n,t,a){return{spans:r(e,t,a,n),endOfLineState:0}}function r(n,t,r,s){var c=[];return n&&t&&function(n,t,r,s,c){var l=n.getTypeChecker(),u=!1;function d(p){switch(p.kind){case 256:case 252:case 253:case 251:case 221:case 208:case 209:c.throwIfCancellationRequested()}if(p&&e.textSpanIntersectsWith(r,p.pos,p.getFullWidth())&&0!==p.getFullWidth()){var m=u;if((e.isJsxElement(p)||e.isJsxSelfClosingElement(p))&&(u=!0),e.isJsxExpression(p)&&(u=!1),e.isIdentifier(p)&&!u&&!function(n){var t=n.parent;return t&&(e.isImportClause(t)||e.isImportSpecifier(t)||e.isNamespaceImport(t))}(p)){var f=l.getSymbolAtLocation(p);if(f){2097152&f.flags&&(f=l.getAliasedSymbol(f));var _=function(n,t){var r=n.getFlags();if(32&r)return 0;if(384&r)return 1;if(524288&r)return 5;if(64&r){if(2&t)return 2}else if(262144&r)return 4;var i=n.valueDeclaration||n.declarations&&n.declarations[0];i&&e.isBindingElement(i)&&(i=a(i));return i&&o.get(i.kind)}(f,e.getMeaningFromLocation(p));if(void 0!==_){var g=0;if(p.parent)(e.isBindingElement(p.parent)||o.get(p.parent.kind)===_)&&p.parent.name===p&&(g=1);6===_&&i(p)&&(_=9),_=function(n,t,r){if(7===r||9===r||6===r){var a=n.getTypeAtLocation(t);if(a){var o=function(e){return e(a)||a.isUnion()&&a.types.some(e)};if(6!==r&&o((function(e){return e.getConstructSignatures().length>0})))return 0;if(o((function(e){return e.getCallSignatures().length>0}))&&!o((function(e){return e.getProperties().length>0}))||function(n){for(;i(n);)n=n.parent;return e.isCallExpression(n.parent)&&n.parent.expression===n}(t))return 9===r?11:10}}return r}(l,p,_);var y=f.valueDeclaration;if(y){var v=e.getCombinedModifierFlags(y),h=e.getCombinedNodeFlags(y);32&v&&(g|=2),256&v&&(g|=4),0!==_&&2!==_&&(64&v||2&h||8&f.getFlags())&&(g|=8),7!==_&&10!==_||!function(n,t){e.isBindingElement(n)&&(n=a(n));if(e.isVariableDeclaration(n))return(!e.isSourceFile(n.parent.parent.parent)||e.isCatchClause(n.parent))&&n.getSourceFile()===t;if(e.isFunctionDeclaration(n))return!e.isSourceFile(n.parent)&&n.getSourceFile()===t;return!1}(y,t)||(g|=32),n.isSourceFileDefaultLibrary(y.getSourceFile())&&(g|=16)}else f.declarations&&f.declarations.some((function(e){return n.isSourceFileDefaultLibrary(e.getSourceFile())}))&&(g|=16);s(p,_,g)}}}e.forEachChild(p,d),u=m}}d(t)}(n,t,r,(function(e,n,r){c.push(e.getStart(t),e.getWidth(t),(n+1<<8)+r)}),s),c}function a(n){for(;;){if(!e.isBindingElement(n.parent.parent))return n.parent.parent;n=n.parent.parent}}function i(n){return e.isQualifiedName(n.parent)&&n.parent.right===n||e.isPropertyAccessExpression(n.parent)&&n.parent.name===n}!function(e){e[e.typeOffset=8]="typeOffset",e[e.modifierMask=255]="modifierMask"}(n.TokenEncodingConsts||(n.TokenEncodingConsts={})),function(e){e[e.class=0]="class",e[e.enum=1]="enum",e[e.interface=2]="interface",e[e.namespace=3]="namespace",e[e.typeParameter=4]="typeParameter",e[e.type=5]="type",e[e.parameter=6]="parameter",e[e.variable=7]="variable",e[e.enumMember=8]="enumMember",e[e.property=9]="property",e[e.function=10]="function",e[e.member=11]="member"}(n.TokenType||(n.TokenType={})),function(e){e[e.declaration=0]="declaration",e[e.static=1]="static",e[e.async=2]="async",e[e.readonly=3]="readonly",e[e.defaultLibrary=4]="defaultLibrary",e[e.local=5]="local"}(n.TokenModifier||(n.TokenModifier={})),n.getSemanticClassifications=function(n,r,a,i){var o=t(n,r,a,i);e.Debug.assert(o.spans.length%3==0);for(var s=o.spans,c=[],l=0;l<s.length;l+=3)c.push({textSpan:e.createTextSpan(s[l],s[l+1]),classificationType:s[l+2]});return c},n.getEncodedSemanticClassifications=t;var o=new e.Map([[249,7],[160,6],[163,9],[256,3],[255,1],[291,8],[252,0],[165,11],[251,10],[208,10],[164,11],[167,9],[168,9],[162,9],[253,2],[254,5],[159,4],[288,9],[289,9]])}(n.v2020||(n.v2020={}))}(e.classifier||(e.classifier={}))}(ts||(ts={})),function(e){!function(n){!function(t){function r(e){return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:!0,entries:e.map((function(e){var t=e.name,r=e.kind,i=e.span;return{name:t,kind:r,kindModifiers:a(e.extension),sortText:n.SortText.LocationPriority,replacementSpan:i}}))}}function a(n){switch(n){case".d.ts":return".d.ts";case".js":return".js";case".json":return".json";case".jsx":return".jsx";case".ts":return".ts";case".tsx":return".tsx";case".tsbuildinfo":return e.Debug.fail("Extension .tsbuildinfo is unsupported.");case void 0:return"";default:return e.Debug.assertNever(n)}}var i;function o(n,t,r,a,i,o){var u,d,p=s(t.parent);switch(p.kind){case 191:var f=s(p.parent);switch(f.kind){case 173:var _=f,g=e.findAncestor(p,(function(e){return e.parent===_}));return g?{kind:2,types:l(a.getTypeArgumentConstraint(g)),isNewIdentifier:!1}:void 0;case 189:var y=f,v=y.indexType,h=y.objectType;if(!e.rangeContainsPosition(v,r))return;return c(a.getTypeFromTypeNode(h));case 195:return{kind:0,paths:m(n,t,i,o,a)};case 182:if(!e.isTypeReferenceNode(f.parent))return;var b=(u=f,d=p,e.mapDefined(u.types,(function(n){return n!==d&&e.isLiteralTypeNode(n)&&e.isStringLiteral(n.literal)?n.literal.text:void 0})));return{kind:2,types:l(a.getTypeArgumentConstraint(f)).filter((function(n){return!e.contains(b,n.value)})),isNewIdentifier:!1};default:return}case 288:return e.isObjectLiteralExpression(p.parent)&&p.name===t?c(a.getContextualType(p.parent)):L();case 202:var E=p,T=E.expression,S=E.argumentExpression;return t===e.skipParentheses(S)?c(a.getTypeAtLocation(T)):void 0;case 203:case 204:if(!e.isRequireCall(p,!1)&&!e.isImportCall(p)){var x=e.SignatureHelp.getArgumentInfoForCompletions(t,r,n);return x?function(n,t){var r=!1,a=new e.Map,i=[];return t.getResolvedSignature(n.invocation,i,n.argumentCount),{kind:2,types:e.flatMap(i,(function(i){if(e.signatureHasRestParameter(i)||!(n.argumentCount>i.parameters.length)){var o=t.getParameterType(i,n.argumentIndex);return r=r||!!(4&o.flags),l(o,a)}})),isNewIdentifier:r}}(x,a):L()}case 261:case 267:case 272:return{kind:0,paths:m(n,t,i,o,a)};default:return L()}function L(){return{kind:2,types:l(e.getContextualTypeFromParent(t,a)),isNewIdentifier:!1}}}function s(n){switch(n.kind){case 186:return e.walkUpParenthesizedTypes(n);case 207:return e.walkUpParenthesizedExpressions(n);default:return n}}function c(n){return n&&{kind:1,symbols:e.filter(n.getApparentProperties(),(function(n){return!(n.valueDeclaration&&e.isPrivateIdentifierPropertyDeclaration(n.valueDeclaration))})),hasIndexSignature:e.hasIndexSignature(n)}}function l(n,t){return void 0===t&&(t=new e.Map),n?(n=e.skipConstraint(n)).isUnion()?e.flatMap(n.types,(function(e){return l(e,t)})):!n.isStringLiteral()||1024&n.flags||!e.addToSeen(t,n.value)?e.emptyArray:[n]:e.emptyArray}function u(e,n,t){return{name:e,kind:n,extension:t}}function d(e){return u(e,"directory",void 0)}function p(n,t,r){var a=function(n,t){var r=Math.max(n.lastIndexOf(e.directorySeparator),n.lastIndexOf("\\")),a=-1!==r?r+1:0,i=n.length-a;return 0===i||e.isIdentifierText(n.substr(a,i),99)?void 0:e.createTextSpan(t+a,i)}(n,t);return r.map((function(e){return{name:e.name,kind:e.kind,extension:e.extension,span:a}}))}function m(n,t,r,a,i){return p(t.text,t.getStart(n)+1,function(n,t,r,a,i){var o=e.normalizeSlashes(t.text),s=n.path,c=e.getDirectoryPath(s);return function(e){if(e&&e.length>=2&&46===e.charCodeAt(0)){var n=e.length>=3&&46===e.charCodeAt(1)?2:1,t=e.charCodeAt(n);return 47===t||92===t}return!1}(o)||!r.baseUrl&&(e.isRootedDiskPath(o)||e.isUrl(o))?function(n,t,r,a,i){var o=f(r);return r.rootDirs?function(n,t,r,a,i,o,s){var c=i.project||o.getCurrentDirectory(),l=!(o.useCaseSensitiveFileNames&&o.useCaseSensitiveFileNames()),u=function(n,t,r,a){n=n.map((function(n){return e.normalizePath(e.isRootedDiskPath(n)?n:e.combinePaths(t,n))}));var i=e.firstDefined(n,(function(n){return e.containsPath(n,r,t,a)?r.substr(n.length):void 0}));return e.deduplicate(__spreadArrays(n.map((function(n){return e.combinePaths(n,i)})),[r]),e.equateStringsCaseSensitive,e.compareStringsCaseSensitive)}(n,c,r,l);return e.flatMap(u,(function(e){return g(t,e,a,o,s)}))}(r.rootDirs,n,t,o,r,a,i):g(n,t,o,a,i)}(o,c,r,a,s):function(n,t,r,a,i){var o=r.baseUrl,s=r.paths,c=[],l=f(r);if(o){var d=r.project||a.getCurrentDirectory(),p=e.normalizePath(e.combinePaths(d,o));g(n,p,l,a,void 0,c),s&&y(c,n,p,l.extensions,s,a)}for(var m=v(n),_=0,h=function(n,t,r){var a=r.getAmbientModules().map((function(n){return e.stripQuotes(n.name)})).filter((function(t){return e.startsWith(t,n)}));if(void 0!==t){var i=e.ensureTrailingDirectorySeparator(t);return a.map((function(n){return e.removePrefix(n,i)}))}return a}(n,m,i);_<h.length;_++){var E=h[_];c.push(u(E,"external module name",void 0))}if(b(a,r,t,m,l,c),e.getEmitModuleResolutionKind(r)===e.ModuleResolutionKind.NodeJs){var S=!1;if(void 0===m)for(var x=function(e){c.some((function(n){return n.name===e}))||(S=!0,c.push(u(e,"external module name",void 0)))},L=0,A=function(n,t){if(!n.readFile||!n.fileExists)return e.emptyArray;for(var r=[],a=0,i=e.findPackageJsons(t,n);a<i.length;a++)for(var o=i[a],s=e.readJson(o,n),c=0,l=T;c<l.length;c++){var u=s[l[c]];if(u)for(var d in u)u.hasOwnProperty(d)&&!e.startsWith(d,"@types/")&&r.push(d)}return r}(a,t);L<A.length;L++){x(A[L])}S||e.forEachAncestorDirectory(t,(function(t){var r=e.combinePaths(t,"node_modules");e.tryDirectoryExists(a,r)&&g(n,r,l,a,void 0,c)}))}return c}(o,c,r,a,i)}(n,t,r,a,i))}function f(e,n){return void 0===n&&(n=!1),{extensions:_(e),includeExtensions:n}}function _(n){var t=e.getSupportedExtensions(n);return n.resolveJsonModule&&e.getEmitModuleResolutionKind(n)===e.ModuleResolutionKind.NodeJs?t.concat(".json"):t}function g(n,t,r,a,i,o){var s=r.extensions,c=r.includeExtensions;void 0===o&&(o=[]),void 0===n&&(n=""),n=e.normalizeSlashes(n),e.hasTrailingDirectorySeparator(n)||(n=e.getDirectoryPath(n)),""===n&&(n="."+e.directorySeparator),n=e.ensureTrailingDirectorySeparator(n);var l=e.resolvePath(t,n),p=e.hasTrailingDirectorySeparator(l)?l:e.getDirectoryPath(l),m=!(a.useCaseSensitiveFileNames&&a.useCaseSensitiveFileNames());if(!e.tryDirectoryExists(a,p))return o;var f=e.tryReadDirectory(a,p,s,void 0,["./*"]);if(f){for(var _=new e.Map,g=0,v=f;g<v.length;g++){var h=v[g];if(h=e.normalizePath(h),!i||0!==e.comparePaths(h,i,t,m)){var b=c||e.fileExtensionIs(h,".json")?e.getBaseFileName(h):e.removeFileExtension(e.getBaseFileName(h));_.set(b,e.tryGetExtensionFromPath(h))}}_.forEach((function(e,n){o.push(u(n,"script",e))}))}var E=e.tryGetDirectories(a,p);if(E)for(var T=0,S=E;T<S.length;T++){var x=S[T],L=e.getBaseFileName(e.normalizePath(x));"@types"!==L&&o.push(d(L))}var A=e.findPackageJson(p,a);if(A){var D=e.readJson(A,a).typesVersions;if("object"==typeof D){var C=e.getPackageJsonTypesVersionsPaths(D),I=C&&C.paths,N=l.slice(e.ensureTrailingDirectorySeparator(p).length);I&&y(o,N,p,s,I,a)}}return o}function y(n,t,r,a,i,o){for(var s in i)if(e.hasProperty(i,s)){var c=i[s];if(c)for(var l=function(e,t,r){n.some((function(n){return n.name===e}))||n.push(u(e,t,r))},d=0,p=h(s,c,t,r,a,o);d<p.length;d++){var m=p[d];l(m.name,m.kind,m.extension)}}}function v(n){return S(n)?e.hasTrailingDirectorySeparator(n)?n:e.getDirectoryPath(n):void 0}function h(n,t,r,a,i,o){if(!e.endsWith(n,"*"))return e.stringContains(n,"*")?e.emptyArray:l(n);var s=n.slice(0,n.length-1),c=e.tryRemovePrefix(r,s);return void 0===c?l(s):e.flatMap(t,(function(n){return function(n,t,r,a,i){if(!i.readDirectory)return;var o=e.hasZeroOrOneAsteriskCharacter(r)?e.tryParsePattern(r):void 0;if(!o)return;var s=e.resolvePath(o.prefix),c=e.hasTrailingDirectorySeparator(o.prefix)?s:e.getDirectoryPath(s),l=e.hasTrailingDirectorySeparator(o.prefix)?"":e.getBaseFileName(s),p=S(n),m=p?e.hasTrailingDirectorySeparator(n)?n:e.getDirectoryPath(n):void 0,f=p?e.combinePaths(c,l+m):c,_=e.normalizePath(o.suffix),g=e.normalizePath(e.combinePaths(t,f)),y=p?g:e.ensureTrailingDirectorySeparator(g)+l,v=_?"**/*":"./*",h=e.mapDefined(e.tryReadDirectory(i,g,a,void 0,[v]),(function(n){var t=e.tryGetExtensionFromPath(n),r=E(n);return void 0===r?void 0:u(e.removeFileExtension(r),"script",t)})),b=e.mapDefined(e.tryGetDirectories(i,g).map((function(n){return e.combinePaths(g,n)})),(function(e){var n=E(e);return void 0===n?void 0:d(n)}));return __spreadArrays(h,b);function E(n){var t,r,a,i=(t=e.normalizePath(n),r=y,a=_,e.startsWith(t,r)&&e.endsWith(t,a)?t.slice(r.length,t.length-a.length):void 0);return void 0===i?void 0:function(n){return n[0]===e.directorySeparator?n.slice(1):n}(i)}}(c,a,n,i,o)}));function l(n){return e.startsWith(n,r)?[d(n)]:e.emptyArray}}function b(n,t,r,a,i,o){void 0===o&&(o=[]);for(var s=new e.Map,c=0,l=e.tryAndIgnoreErrors((function(){return e.getEffectiveTypeRoots(t,n)}))||e.emptyArray;c<l.length;c++){f(l[c])}for(var d=0,p=e.findPackageJsons(r,n);d<p.length;d++){var m=p[d];f(e.combinePaths(e.getDirectoryPath(m),"node_modules/@types"))}return o;function f(r){if(e.tryDirectoryExists(n,r))for(var c=0,l=e.tryGetDirectories(n,r);c<l.length;c++){var d=l[c],p=e.unmangleScopedPackageName(d);if(!t.types||e.contains(t.types,p))if(void 0===a)s.has(p)||(o.push(u(p,"external module name",void 0)),s.set(p,!0));else{var m=e.combinePaths(r,d),f=e.tryRemoveDirectoryPrefix(a,p,e.hostGetCanonicalFileName(n));void 0!==f&&g(f,m,i,n,void 0,o)}}}}t.getStringLiteralCompletions=function(t,a,i,s,c,l,u,d){if(e.isInReferenceComment(t,a))return(m=function(n,t,r,a){var i=e.getTokenAtPosition(n,t),o=e.getLeadingCommentRanges(n.text,i.pos),s=o&&e.find(o,(function(e){return t>=e.pos&&t<=e.end}));if(!s)return;var c=n.text.slice(s.pos,t),l=E.exec(c);if(!l)return;var u=l[1],d=l[2],m=l[3],_=e.getDirectoryPath(n.path),y="path"===d?g(m,_,f(r,!0),a,n.path):"types"===d?b(a,r,_,v(m),f(r)):e.Debug.fail();return p(m,s.pos+u.length,y)}(t,a,c,l))&&r(m);if(e.isInString(t,a,i)){if(!i||!e.isStringLiteralLike(i))return;var m;return function(t,a,i,o,s,c){if(void 0===t)return;var l=e.createTextSpanFromStringLiteralLikeContent(a);switch(t.kind){case 0:return r(t.paths);case 1:var u=[];return n.getCompletionEntriesFromSymbols(t.symbols,u,a,i,i,o,99,s,4,c),{isGlobalCompletion:!1,isMemberCompletion:!0,isNewIdentifierLocation:t.hasIndexSignature,optionalReplacementSpan:l,entries:u};case 2:u=t.types.map((function(t){return{name:t.value,kindModifiers:"",kind:"string",sortText:n.SortText.LocationPriority,replacementSpan:e.getReplacementSpanForContextToken(a)}}));return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:t.isNewIdentifier,optionalReplacementSpan:l,entries:u};default:return e.Debug.assertNever(t)}}(m=o(t,i,a,s,c,l),i,t,s,u,d)}},t.getStringLiteralCompletionDetails=function(t,r,i,s,c,l,u,d){if(s&&e.isStringLiteralLike(s)){var p=o(r,s,i,c,l,u);return p&&function(t,r,i,o,s,c){switch(i.kind){case 0:return(l=e.find(i.paths,(function(e){return e.name===t})))&&n.createCompletionDetails(t,a(l.extension),l.kind,[e.textPart(t)]);case 1:var l;return(l=e.find(i.symbols,(function(e){return e.name===t})))&&n.createCompletionDetailsForSymbol(l,s,o,r,c);case 2:return e.find(i.types,(function(e){return e.value===t}))?n.createCompletionDetails(t,"","type",[e.textPart(t)]):void 0;default:return e.Debug.assertNever(i)}}(t,s,p,r,c,d)}},function(e){e[e.Paths=0]="Paths",e[e.Properties=1]="Properties",e[e.Types=2]="Types"}(i||(i={}));var E=/^(\/\/\/\s*<reference\s+(path|types)\s*=\s*(?:'|"))([^\3"]*)$/,T=["dependencies","devDependencies","peerDependencies","optionalDependencies"];function S(n){return e.stringContains(n,e.directorySeparator)}}(n.StringCompletions||(n.StringCompletions={}))}(e.Completions||(e.Completions={}))}(ts||(ts={})),function(e){!function(n){var t,r,a,i,o,s;function c(e){return!!(e&&4&e.kind)}function l(e){return c(e)&&!!e.isFromPackageJson}function u(e){return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:!1,entries:e}}function d(n){return 78===(null==n?void 0:n.kind)?e.createTextSpanFromNode(n):void 0}function p(n,t){return e.isSourceFileJS(n)&&!e.isCheckJsEnabledForFile(n,t)}function m(e){switch(e){case 0:case 3:case 2:return!0;default:return!1}}function f(n,t,r){return"object"==typeof r?e.pseudoBigIntToString(r)+"n":e.isString(r)?e.quote(n,t,r):JSON.stringify(r)}function _(e,n,r){return{name:f(e,n,r),kind:"string",kindModifiers:"",sortText:t.LocationPriority}}function g(n,t,r,a,i,o,s,u,d,p,m,f,_){var g,b=e.getReplacementSpanForContextToken(r),E=d&&function(e){return!!(16&e.kind)}(d),T=d&&function(e){return!!(2&e.kind)}(d)||u;if(d&&function(e){return!!(1&e.kind)}(d))g=u?"this"+(E?"?.":"")+"["+y(i,_,s)+"]":"this"+(E?"?.":".")+s;else if((T||E)&&m){g=T?u?"["+y(i,_,s)+"]":"["+s+"]":s,(E||m.questionDotToken)&&(g="?."+g);var S=e.findChildOfKind(m,24,i)||e.findChildOfKind(m,28,i);if(!S)return;var x=e.startsWith(s,m.name.text)?m.name.end:S.end;b=e.createTextSpanFromBounds(S.getStart(i),x)}if(f&&(void 0===g&&(g=s),g="{"+g+"}","boolean"!=typeof f&&(b=e.createTextSpanFromNode(f,i))),d&&function(e){return!!(8&e.kind)}(d)&&m){void 0===g&&(g=s);var L=e.findPrecedingToken(m.pos,i),A="";L&&e.positionIsASICandidate(L.end,L.parent,i)&&(A=";"),A+="(await "+m.expression.getText()+")",g=u?""+A+g:A+(E?"?.":".")+g,b=e.createTextSpanFromBounds(m.getStart(i),m.end)}if(void 0===g||_.includeCompletionsWithInsertText)return{name:s,kind:e.SymbolDisplay.getSymbolKind(o,n,a),kindModifiers:e.SymbolDisplay.getSymbolModifiers(n),sortText:t,source:h(d),hasAction:d&&c(d)||void 0,isRecommended:v(n,p,o)||void 0,insertText:g,replacementSpan:b,isPackageJsonImport:l(d)||void 0}}function y(n,t,r){return/^\d+$/.test(r)?r:e.quote(n,t,r)}function v(e,n,t){return e===n||!!(1048576&e.flags)&&t.getExportSymbolOfSymbol(e)===n}function h(n){return c(n)?e.stripQuotes(n.moduleSymbol.name):1===(null==n?void 0:n.kind)?r.ThisProperty:void 0}function b(n,r,a,i,o,s,c,l,u,d,p,m,f,_,y,v){for(var h=e.timestamp(),b=new e.Map,E=0,T=n;E<T.length;E++){var S=T[E],x=y?y[e.getSymbolId(S)]:void 0,L=D(S,c,x,u,!!m);if(L){var A=L.name,C=L.needsConvertPropertyAccess;if(!b.get(A)){var I=g(S,v&&v[e.getSymbolId(S)]||t.LocationPriority,a,i,o,s,A,C,x,_,p,f,d);if(I){var N=!(x||void 0===S.parent&&!e.some(S.declarations,(function(e){return e.getSourceFile()===i.getSourceFile()})));b.set(A,N),r.push(I)}}}}return l("getCompletionsAtPosition: getCompletionEntriesFromSymbols: "+(e.timestamp()-h)),{has:function(e){return b.has(e)},add:function(e){return b.set(e,!0)}}}function E(n,t,r,a,i,o,s){var c=n.getCompilerOptions(),l=A(n,t,r,p(r,c),a,{includeCompletionsForModuleExports:!0,includeCompletionsWithInsertText:!0},i,o);if(!l)return{type:"none"};if(0!==l.kind)return{type:"request",request:l};var u=l.symbols,d=l.literals,m=l.location,_=l.completionKind,g=l.symbolToOriginInfoMap,y=l.previousToken,v=l.isJsxInitializer,b=l.isTypeOnlyLocation,E=e.find(d,(function(e){return f(r,s,e)===i.name}));return void 0!==E?{type:"literal",literal:E}:e.firstDefined(u,(function(n){var t=g[e.getSymbolId(n)],r=D(n,c.target,t,_,l.isJsxIdentifierExpected);return r&&r.name===i.name&&h(t)===i.source?{type:"symbol",symbol:n,location:m,symbolToOriginInfoMap:g,previousToken:y,isJsxInitializer:v,isTypeOnlyLocation:b}:void 0}))||{type:"none"}}function T(n,t,r){return x(n,"",t,[e.displayPart(n,r)])}function S(n,t,r,a,i,o,s){var c=t.runWithCancellationToken(i,(function(t){return e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(t,n,r,a,a,7)})),l=c.displayParts,u=c.documentation,d=c.symbolKind,p=c.tags;return x(n.name,e.SymbolDisplay.getSymbolModifiers(n),d,l,u,p,o,s)}function x(e,n,t,r,a,i,o,s){return{name:e,kindModifiers:n,kind:t,displayParts:r,documentation:a,tags:i,codeActions:o,source:s}}function L(n,t,r){var a=r.getAccessibleSymbolChain(n,t,67108863,!1);return a?e.first(a):n.parent&&(function(e){return e.declarations.some((function(e){return 297===e.kind}))}(n.parent)?n:L(n.parent,t,r))}function A(n,r,a,i,o,s,c,l){var u=n.getTypeChecker(),d=e.timestamp(),p=e.getTokenAtPosition(a,o);r("getCompletionData: Get current token: "+(e.timestamp()-d)),d=e.timestamp();var m=e.isInComment(a,o,p);r("getCompletionData: Is inside comment: "+(e.timestamp()-d));var f=!1,_=!1;if(m){if(e.hasDocComment(a,o)){if(64===a.text.charCodeAt(o-1))return{kind:1};var g=e.getLineStartPositionForPosition(o,a);if(!/[^\*|\s(/)]/.test(a.text.substring(g,o)))return{kind:2}}var y=function(n,t){var r=e.findAncestor(n,e.isJSDoc);return r&&r.tags&&(e.rangeContainsPosition(r,t)?e.findLast(r.tags,(function(e){return e.pos<t})):void 0)}(p,o);if(y){if(y.tagName.pos<=o&&o<=y.tagName.end)return{kind:1};if(function(e){switch(e.kind){case 326:case 333:case 327:case 329:case 331:return!0;default:return!1}}(y)&&y.typeExpression&&301===y.typeExpression.kind&&((p=e.getTokenAtPosition(a,o))&&(e.isDeclarationName(p)||333===p.parent.kind&&p.parent.name===p)||(f=pe(y.typeExpression))),!f&&e.isJSDocParameterTag(y)&&(e.nodeIsMissing(y.name)||y.name.pos<=o&&o<=y.name.end))return{kind:3,tag:y}}if(!f)return void r("Returning an empty list because completion was inside a regular comment or plain text part of a JsDoc comment.")}d=e.timestamp();var v=e.findPrecedingToken(o,a,void 0);r("getCompletionData: Get previous token 1: "+(e.timestamp()-d));var h=v;if(h&&o<=h.end&&(e.isIdentifierOrPrivateIdentifier(h)||e.isKeyword(h.kind))){var b=e.timestamp();h=e.findPrecedingToken(h.getFullStart(),a,void 0),r("getCompletionData: Get previous token 2: "+(e.timestamp()-b))}var E,T=p,S=!1,x=!1,A=!1,D=!1,C=!1,I=!1,N=e.getTouchingPropertyName(a,o);if(h){if(function(n){var t=e.timestamp(),i=function(n){return(e.isRegularExpressionLiteral(n)||e.isStringTextContainingNode(n))&&(e.rangeContainsPositionExclusive(e.createTextRangeFromSpan(e.createTextSpanFromNode(n)),o)||o===n.end&&(!!n.isUnterminated||e.isRegularExpressionLiteral(n)))}(n)||function(n){var t=n.parent,r=t.kind;switch(n.kind){case 27:return 249===r||function(n){return 250===n.parent.kind&&!e.isPossiblyTypeArgumentPosition(n,a,u)}(n)||232===r||255===r||ce(r)||253===r||197===r||254===r||e.isClassLike(t)&&!!t.typeParameters&&t.typeParameters.end>=n.pos;case 24:return 197===r;case 58:return 198===r;case 22:return 197===r;case 20:return 287===r||ce(r);case 18:return 255===r;case 29:return 252===r||221===r||253===r||254===r||e.isFunctionLikeKind(r);case 123:return 163===r&&!e.isClassLike(t.parent);case 25:return 160===r||!!t.parent&&197===t.parent.kind;case 122:case 120:case 121:return 160===r&&!e.isConstructorDeclaration(t.parent);case 126:return 265===r||270===r||263===r;case 134:case 146:return!F(n);case 83:case 91:case 117:case 97:case 112:case 99:case 118:case 84:case 149:return!0;case 41:return e.isFunctionLike(n.parent)&&!e.isMethodDeclaration(n.parent)}if(k(P(n))&&F(n))return!1;if(se(n)&&(!e.isIdentifier(n)||e.isParameterPropertyModifier(P(n))||pe(n)))return!1;switch(P(n)){case 125:case 83:case 84:case 133:case 91:case 97:case 117:case 118:case 120:case 121:case 122:case 123:case 112:return!0;case 129:return e.isPropertyDeclaration(n.parent)}return e.isDeclarationName(n)&&!e.isJsxAttribute(n.parent)&&!(e.isClassLike(n.parent)&&(n!==v||o>v.end))}(n)||function(e){if(8===e.kind){var n=e.getFullText();return"."===n.charAt(n.length-1)}return!1}(n)||function(e){if(11===e.kind)return!0;if(31===e.kind&&e.parent){if(275===e.parent.kind)return 275!==N.parent.kind;if(276===e.parent.kind||274===e.parent.kind)return!!e.parent.parent&&273===e.parent.parent.kind}return!1}(n);return r("getCompletionsAtPosition: isCompletionListBlocker: "+(e.timestamp()-t)),i}(h))return void r("Returning an empty list because completion was requested in an invalid position.");var O=h.parent;if(24===h.kind||28===h.kind)switch(S=24===h.kind,x=28===h.kind,O.kind){case 201:if(T=(E=O).expression,(e.isCallExpression(T)||e.isFunctionLike(T))&&T.end===h.pos&&T.getChildCount(a)&&21!==e.last(T.getChildren(a)).kind)return;break;case 157:T=O.left;break;case 256:T=O.name;break;case 195:case 226:T=O;break;default:return}else if(1===a.languageVariant){if(O&&201===O.kind&&(h=O,O=O.parent),p.parent===N)switch(p.kind){case 31:273!==p.parent.kind&&275!==p.parent.kind||(N=p);break;case 43:274===p.parent.kind&&(N=p)}switch(O.kind){case 276:43===h.kind&&(D=!0,N=h);break;case 216:if(!G(O))break;case 274:case 273:case 275:I=!0,29===h.kind&&(A=!0,N=h);break;case 283:19===v.kind&&31===p.kind&&(I=!0);break;case 280:if(O.initializer===v&&v.end<o){I=!0;break}switch(v.kind){case 62:C=!0;break;case 78:I=!0,O!==v.parent&&!O.initializer&&e.findChildOfKind(O,62,a)&&(C=v)}}}}var B=e.timestamp(),U=5,V=!1,j=0,H=[],K=[],W=[],z=l.getImportSuggestionsCache&&l.getImportSuggestionsCache(),q=ae();if(S||x)!function(){U=2;var n=e.isLiteralImportTypeNode(T),t=f||n&&!T.isTypeOf||e.isPartOfTypeNode(T.parent)||e.isPossiblyTypeArgumentPosition(h,a,u),r=e.isInRightSideOfInternalImportEqualsDeclaration(T);if(e.isEntityName(T)||n||e.isPropertyAccessExpression(T)){var i=e.isModuleDeclaration(T.parent);i&&(V=!0);var o=u.getSymbolAtLocation(T);if(o&&1920&(o=e.skipAlias(o,u)).flags){var c=u.getExportsOfModule(o);e.Debug.assertEachIsDefined(c,"getExportsOfModule() should all be defined");for(var l=function(e){return u.isValidPropertyAccess(n?T:T.parent,e.name)},d=function(e){return ie(e)},p=i?function(e){return!!(1920&e.flags)&&!e.declarations.every((function(e){return e.parent===T.parent}))}:r?function(e){return d(e)||l(e)}:t?d:l,m=0,_=c;m<_.length;m++){var g=_[m];p(g)&&H.push(g)}if(!t&&o.declarations&&o.declarations.some((function(e){return 297!==e.kind&&256!==e.kind&&255!==e.kind}))){var y=!1;if((b=u.getTypeOfSymbolAtLocation(o,T).getNonOptionalType()).isNullableType())((E=S&&!x&&!1!==s.includeAutomaticOptionalChainCompletions)||x)&&(b=b.getNonNullableType(),E&&(y=!0));ee(b,!!(32768&T.flags),y)}return}}if(e.isMetaProperty(T)&&(102===T.keywordToken||99===T.keywordToken)&&h===T.getChildAt(1)){var v=102===T.keywordToken?"target":"meta";return void H.push(u.createSymbol(4,e.escapeLeadingUnderscores(v)))}if(!t){var b,E;y=!1;if((b=u.getTypeAtLocation(T).getNonOptionalType()).isNullableType())((E=S&&!x&&!1!==s.includeAutomaticOptionalChainCompletions)||x)&&(b=b.getNonNullableType(),E&&(y=!0));ee(b,!!(32768&T.flags),y)}}();else if(A){var J=u.getJsxIntrinsicTagNamesAt(N);e.Debug.assertEachIsDefined(J,"getJsxIntrinsicTagNames() should all be defined"),re(),H=J.concat(H),U=3,j=0}else if(D){var X=h.parent.parent.openingElement.tagName,Y=u.getSymbolAtLocation(X);Y&&(H=[Y]),U=3,j=0}else if(!re())return;r("getCompletionData: Semantic work: "+(e.timestamp()-B));var Q=v&&function(n,t,r,a){var i=n.parent;switch(n.kind){case 78:return e.getContextualTypeFromParent(n,a);case 62:switch(i.kind){case 249:return a.getContextualType(i.initializer);case 216:return a.getTypeAtLocation(i.left);case 280:return a.getContextualTypeForJsxAttribute(i);default:return}case 102:return a.getContextualType(i);case 81:return e.getSwitchedType(e.cast(i,e.isCaseClause),a);case 18:return e.isJsxExpression(i)&&273!==i.parent.kind?a.getContextualTypeForJsxAttribute(i.parent):void 0;default:var o=e.SignatureHelp.getArgumentInfoForCompletions(n,t,r);return o?a.getContextualTypeForArgumentAtIndex(o.invocation,o.argumentIndex+(27===n.kind?1:0)):e.isEqualityOperatorKind(n.kind)&&e.isBinaryExpression(i)&&e.isEqualityOperatorKind(i.operatorToken.kind)?a.getTypeAtLocation(i.left):a.getContextualType(n)}}(v,o,a,u),Z=e.mapDefined(Q&&(Q.isUnion()?Q.types:[Q]),(function(e){return e.isLiteral()?e.value:void 0})),$=v&&Q&&function(n,t,r){return e.firstDefined(t&&(t.isUnion()?t.types:[t]),(function(t){var a=t&&t.symbol;return a&&424&a.flags&&!e.isAbstractConstructorSymbol(a)?L(a,n,r):void 0}))}(v,Q,u);return{kind:0,symbols:H,completionKind:U,isInSnippetScope:_,propertyAccessToConvert:E,isNewIdentifierLocation:V,location:N,keywordFilters:j,literals:Z,symbolToOriginInfoMap:K,recommendedCompletion:$,previousToken:v,isJsxInitializer:C,insideJsDocTagTypeExpression:f,symbolToSortTextMap:W,isTypeOnlyLocation:q,isJsxIdentifierExpected:I};function ee(n,t,r){V=!!n.getStringIndexType(),x&&e.some(n.getCallSignatures())&&(V=!0);var a=195===T.kind?T:T.parent;if(i)H.push.apply(H,e.filter(w(n,u),(function(e){return u.isValidPropertyAccessForCompletions(a,n,e)})));else for(var o=0,c=n.getApparentProperties();o<c.length;o++){var l=c[o];u.isValidPropertyAccessForCompletions(a,n,l)&&ne(l,!1,r)}if(t&&s.includeCompletionsWithInsertText){var d=u.getPromisedTypeOfPromise(n);if(d)for(var p=0,m=d.getApparentProperties();p<m.length;p++){l=m[p];u.isValidPropertyAccessForCompletions(a,d,l)&&ne(l,!0,r)}}}function ne(n,r,a){var i=e.firstDefined(n.declarations,(function(n){return e.tryCast(e.getNameOfDeclaration(n),e.isComputedPropertyName)}));if(i){var o=te(i.expression),c=o&&u.getSymbolAtLocation(o),l=c&&L(c,h,u);if(l&&!K[e.getSymbolId(l)]){H.push(l);var d=l.parent;K[e.getSymbolId(l)]=d&&e.isExternalModuleSymbol(d)?{kind:f(6),moduleSymbol:d,isDefaultExport:!1}:{kind:f(2)}}else s.includeCompletionsWithInsertText&&(m(n),p(n),H.push(n))}else m(n),p(n),H.push(n);function p(n){(function(n){return!!(n.valueDeclaration&&32&e.getEffectiveModifierFlags(n.valueDeclaration)&&e.isClassLike(n.valueDeclaration.parent))})(n)&&(W[e.getSymbolId(n)]=t.LocalDeclarationPriority)}function m(n){s.includeCompletionsWithInsertText&&(r&&!K[e.getSymbolId(n)]?K[e.getSymbolId(n)]={kind:f(8)}:a&&(K[e.getSymbolId(n)]={kind:16}))}function f(e){return a?16|e:e}}function te(n){return e.isIdentifier(n)?n:e.isPropertyAccessExpression(n)?te(n.expression):void 0}function re(){return 1===(function(){var n,t,r=function(n){if(n){var t=n.parent;switch(n.kind){case 18:case 27:if(e.isObjectLiteralExpression(t)||e.isObjectBindingPattern(t))return t;break;case 41:return e.isMethodDeclaration(t)?e.tryCast(t.parent,e.isObjectLiteralExpression):void 0;case 78:return"async"===n.text&&e.isShorthandPropertyAssignment(n.parent)?n.parent.parent:void 0}}return}(h);if(!r)return 0;if(U=0,200===r.kind){var a=function(n,t){var r=t.getContextualType(n);if(r)return r;if(e.isBinaryExpression(n.parent)&&62===n.parent.operatorToken.kind)return t.getTypeAtLocation(n.parent);return}(r,u);if(void 0===a)return 2;var i=u.getContextualType(r,4);V=e.hasIndexSignature(i||a),n=M(a,i,r,u),t=r.properties}else{e.Debug.assert(196===r.kind),V=!1;var o=e.getRootDeclaration(r.parent);if(!e.isVariableLike(o))return e.Debug.fail("Root declaration is not variable-like.");var s=e.hasInitializer(o)||e.hasType(o)||239===o.parent.parent.kind;if(s||160!==o.kind||(e.isExpression(o.parent)?s=!!u.getContextualType(o.parent):165!==o.parent.kind&&168!==o.parent.kind||(s=e.isExpression(o.parent.parent)&&!!u.getContextualType(o.parent.parent))),s){var c=u.getTypeAtLocation(r);if(!c)return 2;var l=e.getContainingClass(r);n=u.getPropertiesOfType(c).filter((function(n){return!(24&e.getDeclarationModifierFlagsFromSymbol(n))||l&&e.contains(c.symbol.declarations,l)})),t=r.elements}}n&&n.length>0&&(H=function(n,t){if(0===t.length)return n;for(var r=new e.Set,a=new e.Set,i=0,o=t;i<o.length;i++){var s=o[i];if((288===s.kind||289===s.kind||198===s.kind||165===s.kind||167===s.kind||168===s.kind||290===s.kind)&&!pe(s)){var c=void 0;if(e.isSpreadAssignment(s))le(s,r);else if(e.isBindingElement(s)&&s.propertyName)78===s.propertyName.kind&&(c=s.propertyName.escapedText);else{var l=e.getNameOfDeclaration(s);c=l&&e.isPropertyNameLiteral(l)?e.getEscapedTextOfIdentifierOrLiteral(l):void 0}void 0!==c&&a.add(c)}}var u=n.filter((function(e){return!a.has(e.escapedName)}));return de(r,u),u}(n,e.Debug.checkDefined(t)));return ue(),1}()||function(){var n=!h||18!==h.kind&&27!==h.kind?void 0:e.tryCast(h.parent,e.isNamedImportsOrExports);if(!n)return 0;var t=(264===n.kind?n.parent.parent:n.parent).moduleSpecifier;if(!t)return 264===n.kind?2:0;var r=u.getSymbolAtLocation(t);if(!r)return 2;U=3,V=!1;var a=u.getExportsAndPropertiesOfModule(r),i=new e.Set(n.elements.filter((function(e){return!pe(e)})).map((function(e){return(e.propertyName||e.name).escapedText})));return H=a.filter((function(e){return"default"!==e.escapedName&&!i.has(e.escapedName)})),1}()||function(){var n,r=!h||18!==h.kind&&27!==h.kind?void 0:e.tryCast(h.parent,e.isNamedExports);if(!r)return 0;var a=e.findAncestor(r,e.or(e.isSourceFile,e.isModuleDeclaration));return U=5,V=!1,null===(n=a.locals)||void 0===n||n.forEach((function(n,r){var i,o;H.push(n),(null===(o=null===(i=a.symbol)||void 0===i?void 0:i.exports)||void 0===o?void 0:o.has(r))&&(W[e.getSymbolId(n)]=t.OptionalMember)})),1}()||(function(n){if(n){var t=n.parent;switch(n.kind){case 20:case 27:return e.isConstructorDeclaration(n.parent)?n.parent:void 0;default:if(se(n))return t.parent}}}(h)?(U=5,V=!0,j=4,1):0)||function(){var n=function(n,t,r,a){switch(r.kind){case 334:return e.tryCast(r.parent,e.isObjectTypeDeclaration);case 1:var i=e.tryCast(e.lastOrUndefined(e.cast(r.parent,e.isSourceFile).statements),e.isObjectTypeDeclaration);if(i&&!e.findChildOfKind(i,19,n))return i;break;case 78:if(e.isPropertyDeclaration(r.parent)&&r.parent.initializer===r)return;if(F(r))return e.findAncestor(r,e.isObjectTypeDeclaration)}if(!t)return;switch(t.kind){case 62:return;case 26:case 19:return F(r)&&r.parent.name===r?r.parent.parent:e.tryCast(r,e.isObjectTypeDeclaration);case 18:case 27:return e.tryCast(t.parent,e.isObjectTypeDeclaration);default:if(!F(t))return e.getLineAndCharacterOfPosition(n,t.getEnd()).line!==e.getLineAndCharacterOfPosition(n,a).line&&e.isObjectTypeDeclaration(r)?r:void 0;var o=e.isClassLike(t.parent.parent)?k:R;return o(t.kind)||41===t.kind||e.isIdentifier(t)&&o(e.stringToToken(t.text))?t.parent.parent:void 0}}(a,h,N,o);if(!n)return 0;if(U=3,V=!0,j=41===h.kind?0:e.isClassLike(n)?2:3,!e.isClassLike(n))return 1;var t=26===h.kind?h.parent.parent:h.parent,r=e.isClassElement(t)?e.getEffectiveModifierFlags(t):0;if(78===h.kind&&!pe(h))switch(h.getText()){case"private":r|=8;break;case"static":r|=32}if(!(8&r)){var i=e.flatMap(e.getAllSuperTypeNodes(n),(function(e){var t=u.getTypeAtLocation(e);return 32&r?(null==t?void 0:t.symbol)&&u.getPropertiesOfType(u.getTypeOfSymbolAtLocation(t.symbol,n)):t&&u.getPropertiesOfType(t)}));H=function(n,t,r){for(var a=new e.Set,i=0,o=t;i<o.length;i++){var s=o[i];if((163===s.kind||165===s.kind||167===s.kind||168===s.kind)&&(!pe(s)&&!e.hasEffectiveModifier(s,8)&&e.hasEffectiveModifier(s,32)===!!(32&r))){var c=e.getPropertyNameForPropertyNameNode(s.name);c&&a.add(c)}}return n.filter((function(n){return!(a.has(n.escapedName)||!n.declarations||8&e.getDeclarationModifierFlagsFromSymbol(n)||n.valueDeclaration&&e.isPrivateIdentifierPropertyDeclaration(n.valueDeclaration))}))}(i,n.members,r)}return 1}()||function(){var n=function(n){if(n){var t=n.parent;switch(n.kind){case 31:case 30:case 43:case 78:case 201:case 281:case 280:case 282:if(t&&(274===t.kind||275===t.kind)){if(31===n.kind){var r=e.findPrecedingToken(n.pos,a,void 0);if(!t.typeArguments||r&&43===r.kind)break}return t}if(280===t.kind)return t.parent.parent;break;case 10:if(t&&(280===t.kind||282===t.kind))return t.parent.parent;break;case 19:if(t&&283===t.kind&&t.parent&&280===t.parent.kind)return t.parent.parent.parent;if(t&&282===t.kind)return t.parent.parent}}return}(h),t=n&&u.getContextualType(n.attributes);if(!t)return 0;var r=n&&u.getContextualType(n.attributes,4);return H=function(n,t){for(var r=new e.Set,a=new e.Set,i=0,o=t;i<o.length;i++){var s=o[i];pe(s)||(280===s.kind?r.add(s.name.escapedText):e.isJsxSpreadAttribute(s)&&le(s,a))}var c=n.filter((function(e){return!r.has(e.escapedName)}));return de(a,c),c}(M(t,r,n.attributes,u),n.attributes.properties),ue(),U=3,V=!1,1}()||(function(){j=function(n){if(n){var t,r=e.findAncestor(n.parent,(function(n){return e.isClassLike(n)?"quit":!(!e.isFunctionLikeDeclaration(n)||t!==n.body)||(t=n,!1)}));return r&&r}}(h)?5:1,U=1,V=function(e){if(e){var n=e.parent.kind;switch(P(e)){case 27:return 203===n||166===n||204===n||199===n||216===n||174===n;case 20:return 203===n||166===n||204===n||207===n||186===n;case 22:return 199===n||171===n||158===n;case 139:case 140:return!0;case 24:return 256===n;case 18:return 252===n;case 62:return 249===n||216===n;case 15:return 218===n;case 16:return 228===n;case 122:case 120:case 121:return 163===n}}return!1}(h),v!==h&&e.Debug.assert(!!v,"Expected 'contextToken' to be defined when different from 'previousToken'.");var i=v!==h?v.getStart():o,d=function(n,t,r){var a=n;for(;a&&!e.positionBelongsToNode(a,t,r);)a=a.parent;return a}(h,i,a)||a;_=function(n){switch(n.kind){case 297:case 218:case 283:case 230:return!0;default:return e.isStatement(n)}}(d);var p=2887656|(q?0:111551);H=u.getSymbolsInScope(d,p),e.Debug.assertEachIsDefined(H,"getSymbolsInScope() should all be defined");for(var m=0,f=H;m<f.length;m++){var g=f[m];u.isArgumentsSymbol(g)||e.some(g.declarations,(function(e){return e.getSourceFile()===a}))||(W[e.getSymbolId(g)]=t.GlobalsOrKeywords)}if(s.includeCompletionsWithInsertText&&297!==d.kind){var y=u.tryGetThisTypeAt(d,!1);if(y&&!function(e,n,t){var r=t.resolveName("self",void 0,111551,!1);if(r&&t.getTypeOfSymbolAtLocation(r,n)===e)return!0;var a=t.resolveName("global",void 0,111551,!1);if(a&&t.getTypeOfSymbolAtLocation(a,n)===e)return!0;var i=t.resolveName("globalThis",void 0,111551,!1);if(i&&t.getTypeOfSymbolAtLocation(i,n)===e)return!0;return!1}(y,a,u))for(var b=0,E=w(y,u);b<E.length;b++){g=E[b];K[e.getSymbolId(g)]={kind:1},H.push(g),W[e.getSymbolId(g)]=t.SuggestedClassMembers}}if(s.includeCompletionsForModuleExports&&(a.externalModuleIndicator||a.commonJsModuleIndicator||e.compilerOptionsIndicateEs6Modules(n.getCompilerOptions())||e.programContainsModules(n))){var T=v&&e.isIdentifier(v)?v.text.toLowerCase():"",S=function(t,i){var o=z&&z.get(a.fileName,u,c&&i.getProjectVersion?i.getProjectVersion():void 0);if(o)return r("getSymbolsFromOtherSourceFileExports: Using cached list"),o;var s=e.timestamp();r("getSymbolsFromOtherSourceFileExports: Recomputing list"+(c?" for details entry":""));var l=new e.Map,d=new e.Map,p=new e.Map,m=new e.Map,f=[],_=new e.Map;return e.codefix.forEachExternalModuleToImportFrom(n,i,a,!c,!0,(function(n,t,r,a){if(!c||!c.source||e.stripQuotes(n.name)===c.source){var i=r.getTypeChecker(),o=i.resolveExternalModuleSymbol(n);if(e.addToSeen(l,e.getSymbolId(o))){o!==n&&e.every(o.declarations,e.isNonGlobalDeclaration)&&g(o,n,a,!0);for(var s=0,u=i.getExportsAndPropertiesOfModule(n);s<u.length;s++){var f=u[s],y=e.getSymbolId(f).toString();if(e.addToSeen(d,y)&&!e.some(f.declarations,(function(n){return e.isExportSpecifier(n)&&!!n.propertyName&&e.isIdentifierANonContextualKeyword(n.name)}))){var v=i.getMergedSymbol(f.parent)!==o;if(v||e.some(f.declarations,(function(n){return e.isExportSpecifier(n)&&!n.propertyName&&!!n.parent.parent.moduleSpecifier}))){var h=v?f:oe(f);if(!h)continue;var b=e.getSymbolId(h).toString();_.has(b)||p.has(b)?e.addToSeen(p,y):(m.set(b,{alias:f,moduleSymbol:n,isFromPackageJson:a}),p.set(y,!0))}else m.delete(y),g(f,n,a,!1)}}}}})),m.forEach((function(e){return g(e.alias,e.moduleSymbol,e.isFromPackageJson,!1)})),r("getSymbolsFromOtherSourceFileExports: "+(e.timestamp()-s)),f;function g(n,r,a,i){var o="default"===n.escapedName;if(o&&(n=e.getLocalSymbolForExportDefault(n)||n),!u.isUndefinedSymbol(n)){e.addToSeen(_,e.getSymbolId(n));var s={kind:4,moduleSymbol:r,isDefaultExport:o,isFromPackageJson:a};f.push({symbol:n,symbolName:e.getNameForExportedSymbol(n,t),origin:s,skipFilter:i})}}}(n.getCompilerOptions().target,l);!c&&z&&z.set(a.fileName,S,l.getProjectVersion&&l.getProjectVersion()),S.forEach((function(n){var r=n.symbol,a=n.symbolName,i=n.skipFilter,o=n.origin;if(c){if(c.source&&e.stripQuotes(o.moduleSymbol.name)!==c.source)return}else if(!i&&!function(e,n){if(0===n.length)return!0;for(var t=0,r=0;r<e.length;r++)if(e.charCodeAt(r)===n.charCodeAt(t)&&++t===n.length)return!0;return!1}(a.toLowerCase(),T))return;var s=e.getSymbolId(r);H.push(r),K[s]=o,W[s]=t.AutoImportSuggestions}))}!function(n){var t=ae();t&&(j=h&&e.isAssertionExpression(h.parent)?6:7);e.filterMutate(n,(function(n){if(!e.isSourceFile(N)){if(e.isExportAssignment(N.parent))return!0;if(n=e.skipAlias(n,u),e.isInRightSideOfInternalImportEqualsDeclaration(N))return!!(1920&n.flags);if(t)return ie(n)}return!!(111551&e.getCombinedLocalAndExportSymbolFlags(n))}))}(H)}(),1))}function ae(){return f||!function(n){return n&&111===n.kind&&(176===n.parent.kind||e.isTypeOfExpression(n.parent))}(h)&&(e.isPossiblyTypeArgumentPosition(h,a,u)||e.isPartOfTypeNode(N)||function(n){if(n){var t=n.parent.kind;switch(n.kind){case 58:return 163===t||162===t||160===t||249===t||e.isFunctionLikeKind(t);case 62:return 254===t;case 126:return 224===t;case 29:return 173===t||206===t;case 93:return 159===t}}return!1}(h))}function ie(n,t){void 0===t&&(t=new e.Map);var r=e.skipAlias(n.exportSymbol||n,u);return!!(788968&r.flags)||!!(1536&r.flags)&&e.addToSeen(t,e.getSymbolId(r))&&u.getExportsOfModule(r).some((function(e){return ie(e,t)}))}function oe(n){return function(e,n,t){var r=n;for(;2097152&r.flags&&(r=e.getImmediateAliasedSymbol(r));)if(t(r))return r}(u,n,(function(n){return e.some(n.declarations,(function(n){return e.isExportSpecifier(n)||!!n.localSymbol}))}))}function se(n){return!!n.parent&&e.isParameter(n.parent)&&e.isConstructorDeclaration(n.parent.parent)&&(e.isParameterPropertyModifier(n.kind)||e.isDeclarationName(n))}function ce(n){return e.isFunctionLikeKind(n)&&166!==n}function le(e,n){var t=e.expression,r=u.getSymbolAtLocation(t),a=r&&u.getTypeOfSymbolAtLocation(r,t),i=a&&a.properties;i&&i.forEach((function(e){n.add(e.name)}))}function ue(){H.forEach((function(n){16777216&n.flags&&(W[e.getSymbolId(n)]=W[e.getSymbolId(n)]||t.OptionalMember)}))}function de(n,r){if(0!==n.size)for(var a=0,i=r;a<i.length;a++){var o=i[a];n.has(o.name)&&(W[e.getSymbolId(o)]=t.MemberDeclaredBySpreadAssignment)}}function pe(e){return e.getStart(a)<=o&&o<=e.getEnd()}}function D(n,t,r,a,i){var o=c(r)?e.getNameForExportedSymbol(n,t):n.name;if(!(void 0===o||1536&n.flags&&e.isSingleOrDoubleQuote(o.charCodeAt(0))||e.isKnownSymbol(n))){var s={name:o,needsConvertPropertyAccess:!1};if(e.isIdentifierText(o,t,i?1:0)||n.valueDeclaration&&e.isPrivateIdentifierPropertyDeclaration(n.valueDeclaration))return s;switch(a){case 3:return;case 0:return{name:JSON.stringify(o),needsConvertPropertyAccess:!1};case 2:case 1:return 32===o.charCodeAt(0)?void 0:{name:o,needsConvertPropertyAccess:!0};case 5:case 4:return s;default:e.Debug.assertNever(a)}}}!function(e){e.LocalDeclarationPriority="0",e.LocationPriority="1",e.OptionalMember="2",e.MemberDeclaredBySpreadAssignment="3",e.SuggestedClassMembers="4",e.GlobalsOrKeywords="5",e.AutoImportSuggestions="6",e.JavascriptIdentifiers="7"}(t=n.SortText||(n.SortText={})),function(e){e.ThisProperty="ThisProperty/"}(r=n.CompletionSource||(n.CompletionSource={})),function(e){e[e.ThisType=1]="ThisType",e[e.SymbolMember=2]="SymbolMember",e[e.Export=4]="Export",e[e.Promise=8]="Promise",e[e.Nullable=16]="Nullable",e[e.SymbolMemberNoExport=2]="SymbolMemberNoExport",e[e.SymbolMemberExport=6]="SymbolMemberExport"}(a||(a={})),function(e){e[e.None=0]="None",e[e.All=1]="All",e[e.ClassElementKeywords=2]="ClassElementKeywords",e[e.InterfaceElementKeywords=3]="InterfaceElementKeywords",e[e.ConstructorParameterKeywords=4]="ConstructorParameterKeywords",e[e.FunctionLikeBodyKeywords=5]="FunctionLikeBodyKeywords",e[e.TypeAssertionKeywords=6]="TypeAssertionKeywords",e[e.TypeKeywords=7]="TypeKeywords",e[e.Last=7]="Last"}(i||(i={})),function(e){e[e.Continue=0]="Continue",e[e.Success=1]="Success",e[e.Fail=2]="Fail"}(o||(o={})),n.createImportSuggestionsForFileCache=function(){var n,t,r;return{isEmpty:function(){return!n},clear:function(){n=void 0,r=void 0,t=void 0},set:function(e,a,i){n=a,r=e,i&&(t=i)},get:function(a,i,o){if(a===r)return o?t===o?n:void 0:(e.forEach(n,(function(e){var n,t,r;(null===(n=e.symbol.declarations)||void 0===n?void 0:n.length)&&(e.symbol=i.getMergedSymbol(e.origin.isDefaultExport&&null!==(t=e.symbol.declarations[0].localSymbol)&&void 0!==t?t:e.symbol.declarations[0].symbol)),(null===(r=e.origin.moduleSymbol.declarations)||void 0===r?void 0:r.length)&&(e.origin.moduleSymbol=i.getMergedSymbol(e.origin.moduleSymbol.declarations[0].symbol))})),n)}}},n.getCompletionsAtPosition=function(r,a,i,o,s,c,l){var f=a.getTypeChecker(),g=a.getCompilerOptions(),y=e.findPrecedingToken(s,o);if(!l||e.isInString(o,s,y)||function(n,t,r,a){switch(t){case".":case"@":return!0;case'"':case"'":case"`":return!!r&&e.isStringLiteralOrTemplate(r)&&a===r.getStart(n)+1;case"#":return!!r&&e.isPrivateIdentifier(r)&&!!e.getContainingClass(r);case"<":return!!r&&29===r.kind&&(!e.isBinaryExpression(r.parent)||G(r.parent));case"/":return!!r&&(e.isStringLiteralLike(r)?!!e.tryGetImportFromModuleSpecifier(r):43===r.kind&&e.isJsxClosingElement(r.parent));default:return e.Debug.assertNever(t)}}(o,l,y,s)){var v=n.StringCompletions.getStringLiteralCompletions(o,s,y,f,g,r,i,c);if(v)return v;if(y&&e.isBreakOrContinueStatement(y.parent)&&(80===y.kind||85===y.kind||78===y.kind))return function(n){var r=function(n){var r=[],a=new e.Map,i=n;for(;i&&!e.isFunctionLike(i);){if(e.isLabeledStatement(i)){var o=i.label.text;a.has(o)||(a.set(o,!0),r.push({name:o,kindModifiers:"",kind:"label",sortText:t.LocationPriority}))}i=i.parent}return r}(n);if(r.length)return{isGlobalCompletion:!1,isMemberCompletion:!1,isNewIdentifierLocation:!1,entries:r}}(y.parent);var h=A(a,i,o,p(o,g),s,c,void 0,r);if(h)switch(h.kind){case 0:return function(n,r,a,i,o,s){var c=o.symbols,l=o.completionKind,u=o.isInSnippetScope,f=o.isNewIdentifierLocation,g=o.location,y=o.propertyAccessToConvert,v=o.keywordFilters,h=o.literals,E=o.symbolToOriginInfoMap,T=o.recommendedCompletion,S=o.isJsxInitializer,x=o.insideJsDocTagTypeExpression,L=o.symbolToSortTextMap;if(g&&g.parent&&e.isJsxClosingElement(g.parent)){var A=g.parent.parent.openingElement.tagName,D=!!e.findChildOfKind(g.parent,31,n),I={name:A.getFullText(n)+(D?"":">"),kind:"class",kindModifiers:void 0,sortText:t.LocationPriority};return{isGlobalCompletion:!1,isMemberCompletion:!0,isNewIdentifierLocation:!1,optionalReplacementSpan:d(g),entries:[I]}}var R=[];if(p(n,a)){var k=b(c,R,void 0,g,n,r,a.target,i,l,s,y,o.isJsxIdentifierExpected,S,T,E,L);!function(n,r,a,i,o){e.getNameTable(n).forEach((function(n,s){if(n!==r){var c=e.unescapeLeadingUnderscores(s);!a.has(c)&&e.isIdentifierText(c,i)&&(a.add(c),o.push({name:c,kind:"warning",kindModifiers:"",sortText:t.JavascriptIdentifiers,isFromUncheckedFile:!0}))}}))}(n,g.pos,k,a.target,R)}else{if(!(f||c&&0!==c.length||0!==v))return;b(c,R,void 0,g,n,r,a.target,i,l,s,y,o.isJsxIdentifierExpected,S,T,E,L)}if(0!==v)for(var O=new e.Set(R.map((function(e){return e.name}))),P=0,M=function(n,t){if(!t)return N(n);var r=n+7+1;return C[r]||(C[r]=N(n).filter((function(n){return!function(e){switch(e){case 125:case 128:case 155:case 131:case 133:case 91:case 154:case 116:case 135:case 117:case 137:case 138:case 139:case 140:case 141:case 144:case 145:case 120:case 121:case 122:case 142:case 147:case 148:case 149:case 151:case 152:return!0;default:return!1}}(e.stringToToken(n.name))})))}(v,!x&&e.isSourceFileJS(n));P<M.length;P++){var w=M[P];O.has(w.name)||R.push(w)}for(var F=0,G=h;F<G.length;F++){var B=G[F];R.push(_(n,s,B))}return{isGlobalCompletion:u,isMemberCompletion:m(l),isNewIdentifierLocation:f,optionalReplacementSpan:d(g),entries:R}}(o,f,g,i,h,c);case 1:return u(e.JsDoc.getJSDocTagNameCompletions());case 2:return u(e.JsDoc.getJSDocTagCompletions());case 3:return u(e.JsDoc.getJSDocParameterNameCompletions(h.tag));default:return e.Debug.assertNever(h)}}},n.getCompletionEntriesFromSymbols=b,n.getCompletionEntryDetails=function(t,r,a,i,o,s,l,u,d){var p=t.getTypeChecker(),m=t.getCompilerOptions(),_=o.name,g=e.findPrecedingToken(i,a);if(e.isInString(a,i,g))return n.StringCompletions.getStringLiteralCompletionDetails(_,a,i,g,p,m,s,d);var y=E(t,r,a,i,o,s,u);switch(y.type){case"request":var v=y.request;switch(v.kind){case 1:return e.JsDoc.getJSDocTagNameCompletionDetails(_);case 2:return e.JsDoc.getJSDocTagCompletionDetails(_);case 3:return e.JsDoc.getJSDocParameterNameCompletionDetails(_);default:return e.Debug.assertNever(v)}case"symbol":var h=y.symbol,b=y.location,x=function(n,t,r,a,i,o,s,l,u,d,p){var m=n[e.getSymbolId(t)];if(!m||!c(m))return{codeActions:void 0,sourceDisplay:void 0};var f=m.moduleSymbol,_=a.getMergedSymbol(e.skipAlias(t.exportSymbol||t,a)),g=e.codefix.getImportCompletionAction(_,f,s,e.getNameForExportedSymbol(t,o.target),i,r,d,u&&e.isIdentifier(u)?u.getStart(s):l,p),y=g.moduleSpecifier,v=g.codeAction;return{sourceDisplay:[e.textPart(y)],codeActions:[v]}}(y.symbolToOriginInfoMap,h,t,p,s,m,a,i,y.previousToken,l,u);return S(h,p,a,b,d,x.codeActions,x.sourceDisplay);case"literal":var L=y.literal;return T(f(a,u,L),"string","string"==typeof L?e.SymbolDisplayPartKind.stringLiteral:e.SymbolDisplayPartKind.numericLiteral);case"none":return I().some((function(e){return e.name===_}))?T(_,"keyword",e.SymbolDisplayPartKind.keyword):void 0;default:e.Debug.assertNever(y)}},n.createCompletionDetailsForSymbol=S,n.createCompletionDetails=x,n.getCompletionEntrySymbol=function(e,n,t,r,a,i,o){var s=E(e,n,t,r,a,i,o);return"symbol"===s.type?s.symbol:void 0},function(e){e[e.Data=0]="Data",e[e.JsDocTagName=1]="JsDocTagName",e[e.JsDocTag=2]="JsDocTag",e[e.JsDocParameterName=3]="JsDocParameterName"}(s||(s={})),function(e){e[e.ObjectPropertyDeclaration=0]="ObjectPropertyDeclaration",e[e.Global=1]="Global",e[e.PropertyAccess=2]="PropertyAccess",e[e.MemberLike=3]="MemberLike",e[e.String=4]="String",e[e.None=5]="None"}(n.CompletionKind||(n.CompletionKind={}));var C=[],I=e.memoize((function(){for(var n=[],r=80;r<=156;r++)n.push({name:e.tokenToString(r),kind:"keyword",kindModifiers:"",sortText:t.GlobalsOrKeywords});return n}));function N(n){return C[n]||(C[n]=I().filter((function(t){var r=e.stringToToken(t.name);switch(n){case 0:return!1;case 1:return O(r)||133===r||139===r||149===r||140===r||e.isTypeKeyword(r)&&150!==r;case 5:return O(r);case 2:return k(r);case 3:return R(r);case 4:return e.isParameterPropertyModifier(r);case 6:return e.isTypeKeyword(r)||84===r;case 7:return e.isTypeKeyword(r);default:return e.Debug.assertNever(n)}})))}function R(e){return 142===e}function k(n){switch(n){case 125:case 132:case 134:case 146:case 129:case 133:return!0;default:return e.isClassMemberModifier(n)}}function O(n){return 129===n||130===n||126===n||!e.isContextualKeyword(n)&&!k(n)}function P(n){return e.isIdentifier(n)?n.originalKeywordKind||0:n.kind}function M(n,t,r,a){var i=t&&t!==n,o=!i||3&t.flags?n:a.getUnionType([n,t]),s=o.isUnion()?a.getAllPossiblePropertiesOfTypes(o.types.filter((function(n){return!(131068&n.flags||a.isArrayLikeType(n)||e.typeHasCallOrConstructSignatures(n,a)||a.isTypeInvalidDueToUnionDiscriminant(n,r))}))):o.getApparentProperties();return i?s.filter((function(n){return e.some(n.declarations,(function(e){return e.parent!==r}))})):s}function w(n,t){return n.isUnion()?e.Debug.checkEachDefined(t.getAllPossiblePropertiesOfTypes(n.types),"getAllPossiblePropertiesOfTypes() should all be defined"):e.Debug.checkEachDefined(n.getApparentProperties(),"getApparentProperties() should all be defined")}function F(n){return n.parent&&e.isClassOrTypeElement(n.parent)&&e.isObjectTypeDeclaration(n.parent.parent)}function G(n){var t=n.left;return e.nodeIsMissing(t)}}(e.Completions||(e.Completions={}))}(ts||(ts={})),function(e){!function(n){function t(n,t){return{fileName:t.fileName,textSpan:e.createTextSpanFromNode(n,t),kind:"none"}}function r(n){return e.isThrowStatement(n)?[n]:e.isTryStatement(n)?e.concatenate(n.catchClause?r(n.catchClause):n.tryBlock&&r(n.tryBlock),n.finallyBlock&&r(n.finallyBlock)):e.isFunctionLike(n)?void 0:i(n,r)}function a(n){return e.isBreakOrContinueStatement(n)?[n]:e.isFunctionLike(n)?void 0:i(n,a)}function i(n,t){var r=[];return n.forEachChild((function(n){var a=t(n);void 0!==a&&r.push.apply(r,e.toArray(a))})),r}function o(e,n){var t=s(n);return!!t&&t===e}function s(n){return e.findAncestor(n,(function(t){switch(t.kind){case 244:if(240===n.kind)return!1;case 237:case 238:case 239:case 236:case 235:return!n.label||function(n,t){return!!e.findAncestor(n.parent,(function(n){return e.isLabeledStatement(n)?n.label.escapedText===t:"quit"}))}(t,n.label.escapedText);default:return e.isFunctionLike(t)&&"quit"}}))}function c(n,t){for(var r=[],a=2;a<arguments.length;a++)r[a-2]=arguments[a];return!(!t||!e.contains(r,t.kind))&&(n.push(t),!0)}function l(n){var t=[];if(c(t,n.getFirstToken(),96,114,89)&&235===n.kind)for(var r=n.getChildren(),i=r.length-1;i>=0&&!c(t,r[i],114);i--);return e.forEach(a(n.statement),(function(e){o(n,e)&&c(t,e.getFirstToken(),80,85)})),t}function u(e){var n=s(e);if(n)switch(n.kind){case 237:case 238:case 239:case 235:case 236:return l(n);case 244:return d(n)}}function d(n){var t=[];return c(t,n.getFirstToken(),106),e.forEach(n.caseBlock.clauses,(function(r){c(t,r.getFirstToken(),81,87),e.forEach(a(r),(function(e){o(n,e)&&c(t,e.getFirstToken(),80)}))})),t}function p(n,t){var r=[];(c(r,n.getFirstToken(),110),n.catchClause&&c(r,n.catchClause.getFirstToken(),82),n.finallyBlock)&&c(r,e.findChildOfKind(n,95,t),95);return r}function m(n,t){var a=function(n){for(var t=n;t.parent;){var r=t.parent;if(e.isFunctionBlock(r)||297===r.kind)return r;if(e.isTryStatement(r)&&r.tryBlock===t&&r.catchClause)return t;t=r}}(n);if(a){var i=[];return e.forEach(r(a),(function(n){i.push(e.findChildOfKind(n,108,t))})),e.isFunctionBlock(a)&&e.forEachReturnStatement(a,(function(n){i.push(e.findChildOfKind(n,104,t))})),i}}function f(n,t){var a=e.getContainingFunction(n);if(a){var i=[];return e.forEachReturnStatement(e.cast(a.body,e.isBlock),(function(n){i.push(e.findChildOfKind(n,104,t))})),e.forEach(r(a.body),(function(n){i.push(e.findChildOfKind(n,108,t))})),i}}function _(n){var t=e.getContainingFunction(n);if(t){var r=[];return t.modifiers&&t.modifiers.forEach((function(e){c(r,e,129)})),e.forEachChild(t,(function(n){g(n,(function(n){e.isAwaitExpression(n)&&c(r,n.getFirstToken(),130)}))})),r}}function g(n,t){t(n),e.isFunctionLike(n)||e.isClassLike(n)||e.isInterfaceDeclaration(n)||e.isModuleDeclaration(n)||e.isTypeAliasDeclaration(n)||e.isTypeNode(n)||e.forEachChild(n,(function(e){return g(e,t)}))}n.getDocumentHighlights=function(n,r,a,i,o){var s=e.getTouchingPropertyName(a,i);if(s.parent&&(e.isJsxOpeningElement(s.parent)&&s.parent.tagName===s||e.isJsxClosingElement(s.parent))){var y=s.parent.parent,v=[y.openingElement,y.closingElement].map((function(e){return t(e.tagName,a)}));return[{fileName:a.fileName,highlightSpans:v}]}return function(n,t,r,a,i){var o=new e.Set(i.map((function(e){return e.fileName}))),s=e.FindAllReferences.getReferenceEntriesForNode(n,t,r,i,a,void 0,o);if(!s)return;var c=e.arrayToMultiMap(s.map(e.FindAllReferences.toHighlightSpan),(function(e){return e.fileName}),(function(e){return e.span}));return e.arrayFrom(c.entries(),(function(n){var t=n[0],a=n[1];if(!o.has(t)){e.Debug.assert(r.redirectTargetsMap.has(t));var s=r.getSourceFile(t);t=e.find(i,(function(e){return!!e.redirectInfo&&e.redirectInfo.redirectTarget===s})).fileName,e.Debug.assert(o.has(t))}return{fileName:t,highlightSpans:a}}))}(i,s,n,r,o)||function(n,r){var a=function(n,r){switch(n.kind){case 98:case 90:return e.isIfStatement(n.parent)?function(n,r){for(var a=function(n,t){var r=[];for(;e.isIfStatement(n.parent)&&n.parent.elseStatement===n;)n=n.parent;for(;;){var a=n.getChildren(t);c(r,a[0],98);for(var i=a.length-1;i>=0&&!c(r,a[i],90);i--);if(!n.elseStatement||!e.isIfStatement(n.elseStatement))break;n=n.elseStatement}return r}(n,r),i=[],o=0;o<a.length;o++){if(90===a[o].kind&&o<a.length-1){for(var s=a[o],l=a[o+1],u=!0,d=l.getStart(r)-1;d>=s.end;d--)if(!e.isWhiteSpaceSingleLine(r.text.charCodeAt(d))){u=!1;break}if(u){i.push({fileName:r.fileName,textSpan:e.createTextSpanFromBounds(s.getStart(),l.end),kind:"reference"}),o++;continue}}i.push(t(a[o],r))}return i}(n.parent,r):void 0;case 104:return s(n.parent,e.isReturnStatement,f);case 108:return s(n.parent,e.isThrowStatement,m);case 110:case 82:case 95:return s(82===n.kind?n.parent.parent:n.parent,e.isTryStatement,p);case 106:return s(n.parent,e.isSwitchStatement,d);case 81:case 87:return e.isDefaultClause(n.parent)||e.isCaseClause(n.parent)?s(n.parent.parent.parent,e.isSwitchStatement,d):void 0;case 80:case 85:return s(n.parent,e.isBreakOrContinueStatement,u);case 96:case 114:case 89:return s(n.parent,(function(n){return e.isIterationStatement(n,!0)}),l);case 132:return o(e.isConstructorDeclaration,[132]);case 134:case 146:return o(e.isAccessor,[134,146]);case 130:return s(n.parent,e.isAwaitExpression,_);case 129:return y(_(n));case 124:return y(function(n){var t=e.getContainingFunction(n);if(!t)return;var r=[];return e.forEachChild(t,(function(n){g(n,(function(n){e.isYieldExpression(n)&&c(r,n.getFirstToken(),124)}))})),r}(n));default:return e.isModifierKind(n.kind)&&(e.isDeclaration(n.parent)||e.isVariableStatement(n.parent))?y((a=n.kind,i=n.parent,e.mapDefined(function(n,t){var r=n.parent;switch(r.kind){case 257:case 297:case 230:case 284:case 285:return 128&t&&e.isClassDeclaration(n)?__spreadArrays(n.members,[n]):r.statements;case 166:case 165:case 251:return __spreadArrays(r.parameters,e.isClassLike(r.parent)?r.parent.members:[]);case 252:case 221:case 253:case 177:var a=r.members;if(92&t){var i=e.find(r.members,e.isConstructorDeclaration);if(i)return __spreadArrays(a,i.parameters)}else if(128&t)return __spreadArrays(a,[r]);return a;case 200:return;default:e.Debug.assertNever(r,"Invalid container kind.")}}(i,e.modifierToFlag(a)),(function(n){return e.findModifier(n,a)})))):void 0}var a,i;function o(t,a){return s(n.parent,t,(function(n){return e.mapDefined(n.symbol.declarations,(function(n){return t(n)?e.find(n.getChildren(r),(function(n){return e.contains(a,n.kind)})):void 0}))}))}function s(e,n,t){return n(e)?y(t(e,r)):void 0}function y(e){return e&&e.map((function(e){return t(e,r)}))}}(n,r);return a&&[{fileName:r.fileName,highlightSpans:a}]}(s,a)}}(e.DocumentHighlights||(e.DocumentHighlights={}))}(ts||(ts={})),function(e){function n(n,r,a){void 0===r&&(r="");var i=new e.Map,o=e.createGetCanonicalFileName(!!n);function s(e,n,t,r,a,i,o){return l(e,n,t,r,a,i,!0,o)}function c(e,n,t,r,a,i,o){return l(e,n,t,r,a,i,!1,o)}function l(n,t,r,o,s,c,l,u){var d=e.getOrUpdate(i,o,(function(){return new e.Map})),p=d.get(t),m=6===u?100:r.target||1;!p&&a&&((f=a.getDocument(o,t))&&(e.Debug.assert(l),p={sourceFile:f,languageServiceRefCount:0},d.set(t,p)));if(p)p.sourceFile.version!==c&&(p.sourceFile=e.updateLanguageServiceSourceFile(p.sourceFile,s,c,s.getChangeRange(p.sourceFile.scriptSnapshot)),a&&a.setDocument(o,t,p.sourceFile)),l&&p.languageServiceRefCount++;else{var f=e.createLanguageServiceSourceFile(n,s,m,c,!1,u);a&&a.setDocument(o,t,f),p={sourceFile:f,languageServiceRefCount:1},d.set(t,p)}return e.Debug.assert(0!==p.languageServiceRefCount),p.sourceFile}function u(n,t){var r=e.Debug.checkDefined(i.get(t)),a=r.get(n);a.languageServiceRefCount--,e.Debug.assert(a.languageServiceRefCount>=0),0===a.languageServiceRefCount&&r.delete(n)}return{acquireDocument:function(n,a,i,c,l){return s(n,e.toPath(n,r,o),a,t(a),i,c,l)},acquireDocumentWithKey:s,updateDocument:function(n,a,i,s,l){return c(n,e.toPath(n,r,o),a,t(a),i,s,l)},updateDocumentWithKey:c,releaseDocument:function(n,a){return u(e.toPath(n,r,o),t(a))},releaseDocumentWithKey:u,getLanguageServiceRefCounts:function(n){return e.arrayFrom(i.entries(),(function(e){var t=e[0],r=e[1].get(n);return[t,r&&r.languageServiceRefCount]}))},reportStats:function(){var n=e.arrayFrom(i.keys()).filter((function(e){return e&&"_"===e.charAt(0)})).map((function(e){var n=i.get(e),t=[];return n.forEach((function(e,n){t.push({name:n,refCount:e.languageServiceRefCount})})),t.sort((function(e,n){return n.refCount-e.refCount})),{bucket:e,sourceFiles:t}}));return JSON.stringify(n,void 0,2)},getKeyForCompilationSettings:t}}function t(n){return e.sourceFileAffectingCompilerOptions.map((function(t){return e.getCompilerOptionValue(n,t)})).join("|")}e.createDocumentRegistry=function(e,t){return n(e,t)},e.createDocumentRegistryInternal=n}(ts||(ts={})),function(e){!function(n){function t(n,t){return e.forEach(297===n.kind?n.statements:n.body.statements,(function(n){return t(n)||s(n)&&e.forEach(n.body&&n.body.statements,t)}))}function r(n,r){if(n.externalModuleIndicator||void 0!==n.imports)for(var a=0,i=n.imports;a<i.length;a++){var o=i[a];r(e.importFromModuleSpecifier(o),o)}else t(n,(function(n){switch(n.kind){case 267:case 261:(t=n).moduleSpecifier&&e.isStringLiteral(t.moduleSpecifier)&&r(t,t.moduleSpecifier);break;case 260:var t;c(t=n)&&r(t,t.moduleReference.expression)}}))}function a(n,t,r){var a=n.parent;if(a){var i=r.getMergedSymbol(a);return e.isExternalModuleSymbol(i)?{exportingModuleSymbol:i,exportKind:t}:void 0}}function i(e,n){return n.getMergedSymbol(o(e).symbol)}function o(n){if(203===n.kind)return n.getSourceFile();var t=n.parent;return 297===t.kind?t:(e.Debug.assert(257===t.kind),e.cast(t.parent,s))}function s(e){return 256===e.kind&&10===e.name.kind}function c(e){return 272===e.moduleReference.kind&&10===e.moduleReference.expression.kind}n.createImportTracker=function(n,a,s,l){var u=function(n,t,a){for(var i=new e.Map,o=0,s=n;o<s.length;o++){var c=s[o];a&&a.throwIfCancellationRequested(),r(c,(function(n,r){var a=t.getSymbolAtLocation(r);if(a){var o=e.getSymbolId(a).toString(),s=i.get(o);s||i.set(o,s=[]),s.push(n)}}))}return i}(n,s,l);return function(r,d,p){var m=function(n,r,a,s,c,l){var u=s.exportingModuleSymbol,d=s.exportKind,p=e.nodeSeenTracker(),m=e.nodeSeenTracker(),f=[],_=!!u.globalExports,g=_?void 0:[];return v(u),{directImports:f,indirectUsers:y()};function y(){if(_)return n;for(var t=0,a=u.declarations;t<a.length;t++){var i=a[t];e.isExternalModuleAugmentation(i)&&r.has(i.getSourceFile().fileName)&&b(i)}return g.map(e.getSourceFileOfNode)}function v(n){var t=T(n);if(t)for(var r=0,a=t;r<a.length;r++){var s=a[r];if(p(s))switch(l&&l.throwIfCancellationRequested(),s.kind){case 203:if(!_){var u=s.parent;if(2===d&&249===u.kind){var m=u.name;if(78===m.kind){f.push(m);break}}}break;case 78:break;case 260:h(s,s.name,e.hasSyntacticModifier(s,1),!1);break;case 261:f.push(s);var g=s.importClause&&s.importClause.namedBindings;g&&263===g.kind?h(s,g.name,!1,!0):!_&&e.isDefaultImport(s)&&b(o(s));break;case 267:s.exportClause?269===s.exportClause.kind?E(o(s)):f.push(s):v(i(s,c));break;case 195:f.push(s);break;default:e.Debug.failBadSyntaxKind(s,"Unexpected import kind.")}}}function h(n,r,a,i){if(2===d)i||f.push(n);else if(!_){var s=o(n);e.Debug.assert(297===s.kind||256===s.kind),a||function(n,r,a){var i=a.getSymbolAtLocation(r);return!!t(n,(function(n){if(e.isExportDeclaration(n)){var t=n.exportClause;return!n.moduleSpecifier&&t&&e.isNamedExports(t)&&t.elements.some((function(e){return a.getExportSpecifierLocalTargetSymbol(e)===i}))}}))}(s,r,c)?E(s):b(s)}}function b(n){e.Debug.assert(!_);var t=m(n);return t&&g.push(n),t}function E(n){if(b(n)){var t=c.getMergedSymbol(n.symbol);e.Debug.assert(!!(1536&t.flags));var r=T(t);if(r)for(var a=0,i=r;a<i.length;a++){var s=i[a];e.isImportTypeNode(s)||E(o(s))}}}function T(n){return a.get(e.getSymbolId(n).toString())}}(n,a,u,d,s,l),f=m.directImports,_=m.indirectUsers;return __assign({indirectUsers:_},function(n,t,r,a,i){var o=[],s=[];function l(e,n){o.push([e,n])}if(n)for(var u=0,d=n;u<d.length;u++){p(d[u])}return{importSearches:o,singleReferences:s};function p(n){if(260!==n.kind)if(78!==n.kind)if(195!==n.kind){if(10===n.moduleSpecifier.kind)if(267!==n.kind){var o=n.importClause||{name:void 0,namedBindings:void 0},u=o.name,d=o.namedBindings;if(d)switch(d.kind){case 263:m(d.name);break;case 264:0!==r&&1!==r||f(d);break;default:e.Debug.assertNever(d)}if(u&&(1===r||2===r)&&(!i||u.escapedText===e.symbolEscapedNameNoDefault(t)))l(u,a.getSymbolAtLocation(u))}else n.exportClause&&e.isNamedExports(n.exportClause)&&f(n.exportClause)}else if(n.qualifier){var p=e.getFirstIdentifier(n.qualifier);p.escapedText===e.symbolName(t)&&s.push(p)}else 2===r&&s.push(n.argument.literal);else m(n);else c(n)&&m(n.name)}function m(e){2!==r||i&&!_(e.escapedText)||l(e,a.getSymbolAtLocation(e))}function f(e){if(e)for(var n=0,r=e.elements;n<r.length;n++){var o=r[n],c=o.name,u=o.propertyName;if(_((u||c).escapedText))if(u)s.push(u),i&&c.escapedText!==t.escapedName||l(c,a.getSymbolAtLocation(c));else l(c,270===o.kind&&o.propertyName?a.getExportSpecifierLocalTargetSymbol(o):a.getSymbolAtLocation(c))}}function _(e){return e===t.escapedName||0!==r&&"default"===e}}(f,r,d.exportKind,s,p))}},function(e){e[e.Named=0]="Named",e[e.Default=1]="Default",e[e.ExportEquals=2]="ExportEquals"}(n.ExportKind||(n.ExportKind={})),function(e){e[e.Import=0]="Import",e[e.Export=1]="Export"}(n.ImportExport||(n.ImportExport={})),n.findModuleReferences=function(e,n,t){for(var a=[],i=e.getTypeChecker(),o=0,s=n;o<s.length;o++){var c=s[o],l=t.valueDeclaration;if(297===l.kind){for(var u=0,d=c.referencedFiles;u<d.length;u++){var p=d[u];e.getSourceFileFromReference(c,p)===l&&a.push({kind:"reference",referencingFile:c,ref:p})}for(var m=0,f=c.typeReferenceDirectives;m<f.length;m++){p=f[m];var _=e.getResolvedTypeReferenceDirectives().get(p.fileName);void 0!==_&&_.resolvedFileName===l.fileName&&a.push({kind:"reference",referencingFile:c,ref:p})}}r(c,(function(e,n){i.getSymbolAtLocation(n)===t&&a.push({kind:"import",literal:n})}))}return a},n.getImportOrExportSymbol=function(n,t,r,i){return i?o():o()||function(){if(!function(n){var t=n.parent;switch(t.kind){case 260:return t.name===n&&c(t);case 265:return!t.propertyName;case 262:case 263:return e.Debug.assert(t.name===n),!0;case 198:return e.isInJSFile(n)&&e.isRequireVariableDeclaration(t,!0);default:return!1}}(n))return;var a=r.getImmediateAliasedSymbol(t);if(!a)return;"export="===(a=function(n,t){if(n.declarations)for(var r=0,a=n.declarations;r<a.length;r++){var i=a[r];if(e.isExportSpecifier(i)&&!i.propertyName&&!i.parent.parent.moduleSpecifier)return t.getExportSpecifierLocalTargetSymbol(i);if(e.isPropertyAccessExpression(i)&&e.isModuleExportsAccessExpression(i.expression)&&!e.isPrivateIdentifier(i.name))return t.getExportSpecifierLocalTargetSymbol(i.name);if(e.isShorthandPropertyAssignment(i)&&e.isBinaryExpression(i.parent.parent)&&2===e.getAssignmentDeclarationKind(i.parent.parent))return t.getExportSpecifierLocalTargetSymbol(i.name)}return n}(a,r)).escapedName&&(a=function(n,t){if(2097152&n.flags)return e.Debug.checkDefined(t.getImmediateAliasedSymbol(n));var r=n.valueDeclaration;if(e.isExportAssignment(r))return e.Debug.checkDefined(r.expression.symbol);if(e.isBinaryExpression(r))return e.Debug.checkDefined(r.right.symbol);if(e.isSourceFile(r))return e.Debug.checkDefined(r.symbol);return e.Debug.fail()}(a,r));var i=e.symbolEscapedNameNoDefault(a);if(void 0===i||"default"===i||i===t.escapedName)return{kind:0,symbol:a}}();function o(){var a=n.parent,o=a.parent;if(t.exportSymbol)return 201===a.kind?t.declarations.some((function(e){return e===a}))&&e.isBinaryExpression(o)?d(o,!1):void 0:s(t.exportSymbol,l(a));var c=function(n,t){var r=e.isVariableDeclaration(n)?n:e.isBindingElement(n)?e.walkUpBindingElementsAndPatterns(n):void 0;return r?n.name!==t||e.isCatchClause(r.parent)?void 0:e.isVariableStatement(r.parent.parent)?r.parent.parent:void 0:n}(a,n);if(c&&e.hasSyntacticModifier(c,1)){if(e.isImportEqualsDeclaration(c)&&c.moduleReference===n){if(i)return;return{kind:0,symbol:r.getSymbolAtLocation(c.name)}}return s(t,l(c))}if(e.isNamespaceExport(a))return s(t,0);if(e.isExportAssignment(a))return u(a);if(e.isExportAssignment(o))return u(o);if(e.isBinaryExpression(a))return d(a,!0);if(e.isBinaryExpression(o))return d(o,!0);if(e.isJSDocTypedefTag(a))return s(t,0);function u(n){var r=e.Debug.checkDefined(n.symbol.parent,"Expected export symbol to have a parent"),a=n.isExportEquals?2:1;return{kind:1,symbol:t,exportInfo:{exportingModuleSymbol:r,exportKind:a}}}function d(n,a){var i;switch(e.getAssignmentDeclarationKind(n)){case 1:i=0;break;case 2:i=2;break;default:return}var o=a?r.getSymbolAtLocation(e.getNameOfAccessExpression(e.cast(n.left,e.isAccessExpression))):t;return o&&s(o,i)}}function s(e,n){var t=a(e,n,r);return t&&{kind:1,symbol:e,exportInfo:t}}function l(n){return e.hasSyntacticModifier(n,512)?1:0}},n.getExportInfo=a}(e.FindAllReferences||(e.FindAllReferences={}))}(ts||(ts={})),function(e){!function(n){var t;function r(e,n){return void 0===n&&(n=1),{kind:n,node:e.name||e,context:i(e)}}function a(e){return e&&void 0===e.kind}function i(n){if(e.isDeclaration(n))return o(n);if(n.parent){if(!e.isDeclaration(n.parent)&&!e.isExportAssignment(n.parent)){if(e.isInJSFile(n)){var t=e.isBinaryExpression(n.parent)?n.parent:e.isAccessExpression(n.parent)&&e.isBinaryExpression(n.parent.parent)&&n.parent.parent.left===n.parent?n.parent.parent:void 0;if(t&&0!==e.getAssignmentDeclarationKind(t))return o(t)}if(e.isJsxOpeningElement(n.parent)||e.isJsxClosingElement(n.parent))return n.parent.parent;if(e.isJsxSelfClosingElement(n.parent)||e.isLabeledStatement(n.parent)||e.isBreakOrContinueStatement(n.parent))return n.parent;if(e.isStringLiteralLike(n)){var r=e.tryGetImportFromModuleSpecifier(n);if(r){var a=e.findAncestor(r,(function(n){return e.isDeclaration(n)||e.isStatement(n)||e.isJSDocTag(n)}));return e.isDeclaration(a)?o(a):a}}var i=e.findAncestor(n,e.isComputedPropertyName);return i?o(i.parent):void 0}return n.parent.name===n||e.isConstructorDeclaration(n.parent)||e.isExportAssignment(n.parent)||(e.isImportOrExportSpecifier(n.parent)||e.isBindingElement(n.parent))&&n.parent.propertyName===n||87===n.kind&&e.hasSyntacticModifier(n.parent,513)?o(n.parent):void 0}}function o(n){if(n)switch(n.kind){case 249:return e.isVariableDeclarationList(n.parent)&&1===n.parent.declarations.length?e.isVariableStatement(n.parent.parent)?n.parent.parent:e.isForInOrOfStatement(n.parent.parent)?o(n.parent.parent):n.parent:n;case 198:return o(n.parent.parent);case 265:return n.parent.parent.parent;case 270:case 263:return n.parent.parent;case 262:case 269:return n.parent;case 216:return e.isExpressionStatement(n.parent)?n.parent:n;case 239:case 238:return{start:n.initializer,end:n.expression};case 288:case 289:return e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent)?o(e.findAncestor(n.parent,(function(n){return e.isBinaryExpression(n)||e.isForInOrOfStatement(n)}))):n;default:return n}}function s(e,n,t){if(t){var r=a(t)?f(t.start,n,t.end):f(t,n);return r.start!==e.start||r.length!==e.length?{contextSpan:r}:void 0}}function c(n,a,i,o,s){if(297!==o.kind){var c=n.getTypeChecker();if(289===o.parent.kind){var u=[];return t.getReferenceEntriesForShorthandPropertyAssignment(o,c,(function(e){return u.push(r(e))})),u}if(105===o.kind||e.isSuperProperty(o.parent)){var d=c.getSymbolAtLocation(o);return d.valueDeclaration&&[r(d.valueDeclaration)]}return l(s,o,n,i,a,{implementations:!0,use:1})}}function l(n,r,a,i,o,s,c){return void 0===s&&(s={}),void 0===c&&(c=new e.Set(i.map((function(e){return e.fileName})))),u(t.getReferencedSymbolsForNode(n,r,a,i,o,s,c))}function u(n){return n&&e.flatMap(n,(function(e){return e.references}))}function d(n,r,a){var i=t.getIntersectingMeaningFromDeclarations(a,n),o=n.declarations&&e.firstOrUndefined(n.declarations)||a,s=e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(r,n,o.getSourceFile(),o,o,i);return{displayParts:s.displayParts,kind:s.symbolKind}}function p(e){var n=m(e);if(0===e.kind)return __assign(__assign({},n),{isWriteAccess:!1,isDefinition:!1});var t=e.kind,r=e.node;return __assign(__assign({},n),{isWriteAccess:g(r),isDefinition:y(r),isInString:2===t||void 0})}function m(e){if(0===e.kind)return{textSpan:e.textSpan,fileName:e.fileName};var n=e.node.getSourceFile(),t=f(e.node,n);return __assign({textSpan:t,fileName:n.fileName},s(t,n,e.context))}function f(n,t,r){var a=n.getStart(t),i=(r||n).getEnd();return e.isStringLiteralLike(n)&&(e.Debug.assert(void 0===r),a+=1,i-=1),e.createTextSpanFromBounds(a,i)}function _(e){return 0===e.kind?e.textSpan:f(e.node,e.node.getSourceFile())}function g(n){var t=e.getDeclarationFromName(n);return!!t&&function(n){if(8388608&n.flags)return!0;switch(n.kind){case 216:case 198:case 252:case 221:case 87:case 255:case 291:case 270:case 262:case 260:case 265:case 253:case 324:case 331:case 280:case 256:case 259:case 263:case 269:case 160:case 289:case 254:case 159:return!0;case 288:return!e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent);case 251:case 208:case 166:case 165:case 167:case 168:return!!n.body;case 249:case 163:return!!n.initializer||e.isCatchClause(n.parent);case 164:case 162:case 333:case 326:return!1;default:return e.Debug.failBadSyntaxKind(n)}}(t)||87===n.kind||e.isWriteAccess(n)}function y(n){return 87===n.kind||!!e.getDeclarationFromName(n)||e.isLiteralComputedPropertyDeclarationName(n)||132===n.kind&&e.isConstructorDeclaration(n.parent)}!function(e){e[e.Symbol=0]="Symbol",e[e.Label=1]="Label",e[e.Keyword=2]="Keyword",e[e.This=3]="This",e[e.String=4]="String"}(n.DefinitionKind||(n.DefinitionKind={})),function(e){e[e.Span=0]="Span",e[e.Node=1]="Node",e[e.StringLiteral=2]="StringLiteral",e[e.SearchedLocalFoundProperty=3]="SearchedLocalFoundProperty",e[e.SearchedPropertyFoundLocal=4]="SearchedPropertyFoundLocal"}(n.EntryKind||(n.EntryKind={})),n.nodeEntry=r,n.isContextWithStartAndEndNode=a,n.getContextNode=o,n.toContextSpan=s,function(e){e[e.Other=0]="Other",e[e.References=1]="References",e[e.Rename=2]="Rename"}(n.FindReferencesUse||(n.FindReferencesUse={})),n.findReferencedSymbols=function(n,r,a,i,c){var l=e.getTouchingPropertyName(i,c),u=t.getReferencedSymbolsForNode(c,l,n,a,r,{use:1}),m=n.getTypeChecker();return u&&u.length?e.mapDefined(u,(function(n){var t=n.definition,a=n.references;return t&&{definition:m.runWithCancellationToken(r,(function(n){return function(n,t,r){var a=function(){switch(n.type){case 0:var a=d(f=n.symbol,t,r),i=a.displayParts,s=a.kind,c=i.map((function(e){return e.text})).join(""),l=f.declarations&&e.firstOrUndefined(f.declarations);return{node:l?e.getNameOfDeclaration(l)||l:r,name:c,kind:s,displayParts:i,context:o(l)};case 1:var u=n.node;return{node:u,name:u.text,kind:"label",displayParts:[e.displayPart(u.text,e.SymbolDisplayPartKind.text)]};case 2:var p=n.node,m=e.tokenToString(p.kind);return{node:p,name:m,kind:"keyword",displayParts:[{text:m,kind:"keyword"}]};case 3:var f,_=n.node,g=(f=t.getSymbolAtLocation(_))&&e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(t,f,_.getSourceFile(),e.getContainerNode(_),_).displayParts||[e.textPart("this")];return{node:_,name:"this",kind:"var",displayParts:g};case 4:var y=n.node;return{node:y,name:y.text,kind:"var",displayParts:[e.displayPart(e.getTextOfNode(y),e.SymbolDisplayPartKind.stringLiteral)]};default:return e.Debug.assertNever(n)}}(),i=a.node,c=a.name,l=a.kind,u=a.displayParts,p=a.context,m=i.getSourceFile(),_=f(e.isComputedPropertyName(i)?i.expression:i,m);return __assign({containerKind:"",containerName:"",fileName:m.fileName,kind:l,name:c,textSpan:_,displayParts:u},s(_,m,p))}(t,n,l)})),references:a.map(p)}})):void 0},n.getImplementationsAtPosition=function(n,t,r,a,i){var o,s=e.getTouchingPropertyName(a,i),l=c(n,t,r,s,i);if(201===s.parent.kind||198===s.parent.kind||202===s.parent.kind||105===s.kind)o=l&&__spreadArrays(l);else for(var u=l&&__spreadArrays(l),p=new e.Map;u&&u.length;){var f=u.shift();if(e.addToSeen(p,e.getNodeId(f.node))){o=e.append(o,f);var _=c(n,t,r,f.node,f.node.pos);_&&u.push.apply(u,_)}}var g=n.getTypeChecker();return e.map(o,(function(n){return function(n,t){var r=m(n);if(0!==n.kind){var a=n.node;return __assign(__assign({},r),function(n,t){var r=t.getSymbolAtLocation(e.isDeclaration(n)&&n.name?n.name:n);return r?d(r,t,n):200===n.kind?{kind:"interface",displayParts:[e.punctuationPart(20),e.textPart("object literal"),e.punctuationPart(21)]}:221===n.kind?{kind:"local class",displayParts:[e.punctuationPart(20),e.textPart("anonymous local class"),e.punctuationPart(21)]}:{kind:e.getNodeKind(n),displayParts:[]}}(a,t))}return __assign(__assign({},r),{kind:"",displayParts:[]})}(n,g)}))},n.findReferenceOrRenameEntries=function(n,r,a,i,o,s,c){return e.map(u(t.getReferencedSymbolsForNode(o,i,n,a,r,s)),(function(e){return c(e,i,n.getTypeChecker())}))},n.getReferenceEntriesForNode=l,n.toRenameLocation=function(n,t,r,a){return __assign(__assign({},m(n)),a&&function(n,t,r){if(0!==n.kind&&e.isIdentifier(t)){var a=n.node,i=n.kind,o=a.parent,s=t.text,c=e.isShorthandPropertyAssignment(o);if(c||e.isObjectBindingElementWithoutPropertyName(o)&&o.name===a&&void 0===o.dotDotDotToken){var l={prefixText:s+": "},u={suffixText:": "+s};if(3===i)return l;if(4===i)return u;if(c){var d=o.parent;return e.isObjectLiteralExpression(d)&&e.isBinaryExpression(d.parent)&&e.isModuleExportsAccessExpression(d.parent.left)?l:u}return l}if(e.isImportSpecifier(o)&&!o.propertyName){var p=e.isExportSpecifier(t.parent)?r.getExportSpecifierLocalTargetSymbol(t.parent):r.getSymbolAtLocation(t);return e.contains(p.declarations,o)?{prefixText:s+" as "}:e.emptyOptions}if(e.isExportSpecifier(o)&&!o.propertyName)return t===n.node||r.getSymbolAtLocation(t)===r.getSymbolAtLocation(n.node)?{prefixText:s+" as "}:{suffixText:" as "+s}}return e.emptyOptions}(n,t,r))},n.toReferenceEntry=p,n.toHighlightSpan=function(e){var n=m(e);if(0===e.kind)return{fileName:n.fileName,span:{textSpan:n.textSpan,kind:"reference"}};var t=g(e.node),r=__assign({textSpan:n.textSpan,kind:t?"writtenReference":"reference",isInString:2===e.kind||void 0},n.contextSpan&&{contextSpan:n.contextSpan});return{fileName:n.fileName,span:r}},n.getTextSpanOfEntry=_,function(t){function a(n,t,r){if(n.parent&&e.isNamespaceExportDeclaration(n.parent)){var a=r.getAliasedSymbol(t),i=r.getMergedSymbol(a);if(a!==i)return i}}function i(n,t,r,a,i,s){var l=1536&n.flags&&n.declarations&&e.find(n.declarations,e.isSourceFile);if(l){var d=n.exports.get("export="),p=c(t,n,!!d,r,s);if(!d||!s.has(l.fileName))return p;var m=t.getTypeChecker();return o(t,p,u(n=e.skipAlias(d,m),void 0,r,s,m,a,i))}}function o(n){for(var t,r=[],a=1;a<arguments.length;a++)r[a-1]=arguments[a];for(var i=0,o=r;i<o.length;i++){var c=o[i];if(c&&c.length)if(t)for(var l=function(r){if(!r.definition||0!==r.definition.type)return t.push(r),"continue";var a=r.definition.symbol,i=e.findIndex(t,(function(e){return!!e.definition&&0===e.definition.type&&e.definition.symbol===a}));if(-1===i)return t.push(r),"continue";var o=t[i];t[i]={definition:o.definition,references:o.references.concat(r.references).sort((function(t,r){var a=s(n,t),i=s(n,r);if(a!==i)return e.compareValues(a,i);var o=_(t),c=_(r);return o.start!==c.start?e.compareValues(o.start,c.start):e.compareValues(o.length,c.length)}))}},u=0,d=c;u<d.length;u++){var p=d[u];l(p)}else t=c}return t}function s(e,n){var t=0===n.kind?e.getSourceFile(n.fileName):n.node.getSourceFile();return e.getSourceFiles().indexOf(t)}function c(t,a,i,o,s){e.Debug.assert(!!a.valueDeclaration);var c=e.mapDefined(n.findModuleReferences(t,o,a),(function(n){if("import"===n.kind){var t=n.literal.parent;if(e.isLiteralTypeNode(t)){var a=e.cast(t.parent,e.isImportTypeNode);if(i&&!a.qualifier)return}return r(n.literal)}return{kind:0,fileName:n.referencingFile.fileName,textSpan:e.createTextSpanFromRange(n.ref)}}));if(a.declarations)for(var l=0,u=a.declarations;l<u.length;l++){switch((f=u[l]).kind){case 297:break;case 256:s.has(f.getSourceFile().fileName)&&c.push(r(f.name));break;default:e.Debug.assert(!!(33554432&a.flags),"Expected a module symbol to be declared by a SourceFile or ModuleDeclaration.")}}var d=a.exports.get("export=");if(d)for(var p=0,m=d.declarations;p<m.length;p++){var f,_=(f=m[p]).getSourceFile();if(s.has(_.fileName)){var g=e.isBinaryExpression(f)&&e.isPropertyAccessExpression(f.left)?f.left.expression:e.isExportAssignment(f)?e.Debug.checkDefined(e.findChildOfKind(f,92,_)):e.getNameOfDeclaration(f)||f;c.push(r(g))}}return c.length?[{definition:{type:0,symbol:a},references:c}]:e.emptyArray}function l(n){return 142===n.kind&&e.isTypeOperatorNode(n.parent)&&142===n.parent.operator}function u(n,t,r,a,i,o,s){var c=t&&function(n,t,r,a){var i=t.parent;if(e.isExportSpecifier(i)&&a)return C(t,n,i,r);return e.firstDefined(n.declarations,(function(a){if(!a.parent){if(33554432&n.flags)return;e.Debug.fail("Unexpected symbol at "+e.Debug.formatSyntaxKind(t.kind)+": "+e.Debug.formatSymbol(n))}return e.isTypeLiteralNode(a.parent)&&e.isUnionTypeNode(a.parent.parent)?r.getPropertyOfType(r.getTypeFromTypeNode(a.parent.parent),n.name):void 0}))}(n,t,i,!V(s))||n,l=t?G(t,c):7,u=[],p=new m(r,a,t?function(n){switch(n.kind){case 132:return 1;case 78:if(e.isClassLike(n.parent))return e.Debug.assert(n.parent.name===n),2;default:return 0}}(t):0,i,o,l,s,u),_=V(s)?e.find(c.declarations,e.isExportSpecifier):void 0;if(_)D(_.name,c,_,p.createSearch(t,n,void 0),p,!0,!0);else if(t&&87===t.kind)I(t,c,p),f(t,c,{exportingModuleSymbol:e.Debug.checkDefined(c.parent,"Expected export symbol to have a parent"),exportKind:1},p);else{var g=p.createSearch(t,c,void 0,{allSearchSymbols:t?M(c,t,i,2===s.use,!!s.providePrefixAndSuffixTextForRename,!!s.implementations):[c]});d(c,p,g)}return u}function d(n,t,r){var a=function(n){var t=n.declarations,r=n.flags,a=n.parent,i=n.valueDeclaration;if(i&&(208===i.kind||221===i.kind))return i;if(!t)return;if(8196&r){var o=e.find(t,(function(n){return e.hasEffectiveModifier(n,8)||e.isPrivateIdentifierPropertyDeclaration(n)}));return o?e.getAncestor(o,252):void 0}if(t.some(e.isObjectBindingElementWithoutPropertyName))return;var s,c=a&&!(262144&n.flags);if(c&&(!e.isExternalModuleSymbol(a)||a.globalExports))return;for(var l=0,u=t;l<u.length;l++){var d=u[l],p=e.getContainerNode(d);if(s&&s!==p)return;if(!p||297===p.kind&&!e.isExternalOrCommonJsModule(p))return;s=p}return c?s.getSourceFile():s}(n);if(a)x(a,a.getSourceFile(),r,t,!(e.isSourceFile(a)&&!e.contains(t.sourceFiles,a)));else for(var i=0,o=t.sourceFiles;i<o.length;i++){var s=o[i];t.cancellationToken.throwIfCancellationRequested(),v(s,r,t)}}var p;t.getReferencedSymbolsForNode=function(n,t,s,d,p,m,f){if(void 0===m&&(m={}),void 0===f&&(f=new e.Set(d.map((function(e){return e.fileName})))),1===m.use?t=e.getAdjustedReferenceLocation(t):2===m.use&&(t=e.getAdjustedRenameLocation(t)),e.isSourceFile(t)){var _=e.GoToDefinition.getReferenceAtPosition(t,n,s),g=_&&s.getTypeChecker().getMergedSymbol(_.file.symbol);return g&&c(s,g,!1,d,f)}if(!m.implementations){var y=function(n,t,a){if(e.isTypeKeyword(n.kind)){if(113===n.kind&&e.isVoidExpression(n.parent))return;if(142===n.kind&&!l(n))return;return function(n,t,a,i){var o=e.flatMap(n,(function(n){return a.throwIfCancellationRequested(),e.mapDefined(b(n,e.tokenToString(t),n),(function(e){if(e.kind===t&&(!i||i(e)))return r(e)}))}));return o.length?[{definition:{type:2,node:o[0].node},references:o}]:void 0}(t,n.kind,a,142===n.kind?l:void 0)}if(e.isJumpStatementTarget(n)){var i=e.getTargetLabel(n.parent,n.text);return i&&T(i.parent,i)}if(e.isLabelOfLabeledStatement(n))return T(n.parent,n);if(e.isThis(n))return function(n,t,a){var i=e.getThisContainer(n,!1),o=32;switch(i.kind){case 165:case 164:if(e.isObjectLiteralMethod(i))break;case 163:case 162:case 166:case 167:case 168:o&=e.getSyntacticModifierFlags(i),i=i.parent;break;case 297:if(e.isExternalModule(i)||P(n))return;case 251:case 208:break;default:return}var s=e.flatMap(297===i.kind?t:[i.getSourceFile()],(function(n){return a.throwIfCancellationRequested(),b(n,"this",e.isSourceFile(i)?n:i).filter((function(n){if(!e.isThis(n))return!1;var t=e.getThisContainer(n,!1);switch(i.kind){case 208:case 251:return i.symbol===t.symbol;case 165:case 164:return e.isObjectLiteralMethod(i)&&i.symbol===t.symbol;case 221:case 252:return t.parent&&i.symbol===t.parent.symbol&&(32&e.getSyntacticModifierFlags(t))===o;case 297:return 297===t.kind&&!e.isExternalModule(t)&&!P(n)}}))})).map((function(e){return r(e)}));return[{definition:{type:3,node:e.firstDefined(s,(function(n){return e.isParameter(n.node.parent)?n.node:void 0}))||n},references:s}]}(n,t,a);if(105===n.kind)return function(n){var t=e.getSuperContainer(n,!1);if(!t)return;var a=32;switch(t.kind){case 163:case 162:case 165:case 164:case 166:case 167:case 168:a&=e.getSyntacticModifierFlags(t),t=t.parent;break;default:return}var i=t.getSourceFile(),o=e.mapDefined(b(i,"super",t),(function(n){if(105===n.kind){var i=e.getSuperContainer(n,!1);return i&&(32&e.getSyntacticModifierFlags(i))===a&&i.parent.symbol===t.symbol?r(n):void 0}}));return[{definition:{type:0,symbol:t.symbol},references:o}]}(n);return}(t,d,p);if(y)return y}var v=s.getTypeChecker(),h=v.getSymbolAtLocation(t);if(!h)return!m.implementations&&e.isStringLiteral(t)?function(n,t,a){var i=e.flatMap(t,(function(t){return a.throwIfCancellationRequested(),e.mapDefined(b(t,n.text),(function(t){return e.isStringLiteral(t)&&t.text===n.text?r(t,2):void 0}))}));return[{definition:{type:4,node:n},references:i}]}(t,d,p):void 0;if("export="===h.escapedName)return c(s,h.parent,!1,d,f);var E=i(h,s,d,p,m,f);if(E&&!(33554432&h.flags))return E;var S=a(t,h,v),x=S&&i(S,s,d,p,m,f);return o(s,E,u(h,t,d,f,v,p,m),x)},function(e){e[e.None=0]="None",e[e.Constructor=1]="Constructor",e[e.Class=2]="Class"}(p||(p={}));var m=function(){function t(n,t,r,a,i,o,s,c){this.sourceFiles=n,this.sourceFilesSet=t,this.specialSearchKind=r,this.checker=a,this.cancellationToken=i,this.searchMeaning=o,this.options=s,this.result=c,this.inheritsFromCache=new e.Map,this.markSeenContainingTypeReference=e.nodeSeenTracker(),this.markSeenReExportRHS=e.nodeSeenTracker(),this.symbolIdToReferences=[],this.sourceFileToSeenSymbols=[]}return t.prototype.includesSourceFile=function(e){return this.sourceFilesSet.has(e.fileName)},t.prototype.getImportSearches=function(e,t){return this.importTracker||(this.importTracker=n.createImportTracker(this.sourceFiles,this.sourceFilesSet,this.checker,this.cancellationToken)),this.importTracker(e,t,2===this.options.use)},t.prototype.createSearch=function(n,t,r,a){void 0===a&&(a={});var i=a.text,o=void 0===i?e.stripQuotes(e.symbolName(e.getLocalSymbolForExportDefault(t)||function(n){if(33555968&n.flags){var t=n.declarations&&e.find(n.declarations,(function(n){return!e.isSourceFile(n)&&!e.isModuleDeclaration(n)}));return t&&t.symbol}}(t)||t)):i,s=a.allSearchSymbols,c=void 0===s?[t]:s;return{symbol:t,comingFrom:r,text:o,escapedText:e.escapeLeadingUnderscores(o),parents:this.options.implementations&&n?function(n,t,r){var a=e.isRightSideOfPropertyAccess(n)?n.parent:void 0,i=a&&r.getTypeAtLocation(a.expression),o=e.mapDefined(i&&(i.isUnionOrIntersection()?i.types:i.symbol===t.parent?void 0:[i]),(function(e){return e.symbol&&96&e.symbol.flags?e.symbol:void 0}));return 0===o.length?void 0:o}(n,t,this.checker):void 0,allSearchSymbols:c,includes:function(n){return e.contains(c,n)}}},t.prototype.referenceAdder=function(n){var t=e.getSymbolId(n),a=this.symbolIdToReferences[t];return a||(a=this.symbolIdToReferences[t]=[],this.result.push({definition:{type:0,symbol:n},references:a})),function(e,n){return a.push(r(e,n))}},t.prototype.addStringOrCommentReference=function(e,n){this.result.push({definition:void 0,references:[{kind:0,fileName:e,textSpan:n}]})},t.prototype.markSearchedSymbols=function(n,t){for(var r=e.getNodeId(n),a=this.sourceFileToSeenSymbols[r]||(this.sourceFileToSeenSymbols[r]=new e.Set),i=!1,o=0,s=t;o<s.length;o++){var c=s[o];i=e.tryAddToSet(a,e.getSymbolId(c))||i}return i},t}();function f(e,n,t,r){var a=r.getImportSearches(n,t),i=a.importSearches,o=a.singleReferences,s=a.indirectUsers;if(o.length)for(var c=r.referenceAdder(n),l=0,u=o;l<u.length;l++){var d=u[l];g(d,r)&&c(d)}for(var p=0,m=i;p<m.length;p++){var f=m[p],_=f[0],y=f[1];S(_.getSourceFile(),r.createSearch(_,y,1),r)}if(s.length){var h=void 0;switch(t.exportKind){case 0:h=r.createSearch(e,n,1);break;case 1:h=2===r.options.use?void 0:r.createSearch(e,n,1,{text:"default"})}if(h)for(var b=0,E=s;b<E.length;b++){v(E[b],h,r)}}}function g(n,t){return!!L(n,t)&&(2!==t.options.use||!!e.isIdentifier(n)&&!(e.isImportOrExportSpecifier(n.parent)&&"default"===n.escapedText))}function y(e,n){if(e.declarations)for(var t=0,r=e.declarations;t<r.length;t++){var a=r[t],i=a.getSourceFile();S(i,n.createSearch(a,e,0),n,n.includesSourceFile(i))}}function v(n,t,r){void 0!==e.getNameTable(n).get(t.escapedText)&&S(n,t,r)}function h(n,t,r,a,i){void 0===i&&(i=r);var o=e.isParameterPropertyDeclaration(n.parent,n.parent.parent)?e.first(t.getSymbolsOfParameterPropertyDeclaration(n.parent,n.text)):t.getSymbolAtLocation(n);if(o)for(var s=0,c=b(r,o.name,i);s<c.length;s++){var l=c[s];if(e.isIdentifier(l)&&l!==n&&l.escapedText===n.escapedText){var u=t.getSymbolAtLocation(l);if(u===o||t.getShorthandAssignmentValueSymbol(l.parent)===o||e.isExportSpecifier(l.parent)&&C(l,u,l.parent,t)===o){var d=a(l);if(d)return d}}}}function b(n,t,r){return void 0===r&&(r=n),E(n,t,r).map((function(t){return e.getTouchingPropertyName(n,t)}))}function E(n,t,r){void 0===r&&(r=n);var a=[];if(!t||!t.length)return a;for(var i=n.text,o=i.length,s=t.length,c=i.indexOf(t,r.pos);c>=0&&!(c>r.end);){var l=c+s;0!==c&&e.isIdentifierPart(i.charCodeAt(c-1),99)||l!==o&&e.isIdentifierPart(i.charCodeAt(l),99)||a.push(c),c=i.indexOf(t,c+s+1)}return a}function T(n,t){var a=n.getSourceFile(),i=t.text,o=e.mapDefined(b(a,i,n),(function(n){return n===t||e.isJumpStatementTarget(n)&&e.getTargetLabel(n,i)===t?r(n):void 0}));return[{definition:{type:1,node:t},references:o}]}function S(e,n,t,r){return void 0===r&&(r=!0),t.cancellationToken.throwIfCancellationRequested(),x(e,e,n,t,r)}function x(e,n,t,r,a){if(r.markSearchedSymbols(n,t.allSearchSymbols))for(var i=0,o=E(n,t.text,e);i<o.length;i++){A(n,o[i],t,r,a)}}function L(n,t){return!!(e.getMeaningFromLocation(n)&t.searchMeaning)}function A(t,r,a,i,o){var s=e.getTouchingPropertyName(t,r);if(function(n,t){switch(n.kind){case 79:case 78:return n.text.length===t.length;case 14:case 10:var r=n;return(e.isLiteralNameOfPropertyDeclarationOrIndexAccess(r)||e.isNameOfModuleDeclaration(n)||e.isExpressionOfExternalModuleImportEqualsDeclaration(n)||e.isCallExpression(n.parent)&&e.isBindableObjectDefinePropertyCall(n.parent)&&n.parent.arguments[1]===n)&&r.text.length===t.length;case 8:return e.isLiteralNameOfPropertyDeclarationOrIndexAccess(n)&&n.text.length===t.length;case 87:return"default".length===t.length;default:return!1}}(s,a.text)){if(L(s,i)){var c=i.checker.getSymbolAtLocation(s);if(c){var l=s.parent;if(!e.isImportSpecifier(l)||l.propertyName!==s){if(e.isExportSpecifier(l))return e.Debug.assert(78===s.kind),void D(s,c,l,a,i,o);var u=function(n,t,r,a){var i=a.checker;return w(t,r,i,!1,2!==a.options.use||!!a.options.providePrefixAndSuffixTextForRename,(function(r,a,i,o){return i&&F(t)!==F(i)&&(i=void 0),n.includes(i||a||r)?{symbol:!a||6&e.getCheckFlags(r)?r:a,kind:o}:void 0}),(function(e){return!(n.parents&&!n.parents.some((function(n){return O(e.parent,n,a.inheritsFromCache,i)})))}))}(a,c,s,i);if(u){switch(i.specialSearchKind){case 0:o&&I(s,u,i);break;case 1:!function(n,t,r,a){e.isNewExpressionTarget(n)&&I(n,r.symbol,a);var i=function(){return a.referenceAdder(r.symbol)};if(e.isClassLike(n.parent))e.Debug.assert(87===n.kind||n.parent.name===n),function(n,t,r){var a=N(n);if(a&&a.declarations)for(var i=0,o=a.declarations;i<o.length;i++){var s=o[i],c=e.findChildOfKind(s,132,t);e.Debug.assert(166===s.kind&&!!c),r(c)}n.exports&&n.exports.forEach((function(n){var t=n.valueDeclaration;if(t&&165===t.kind){var a=t.body;a&&U(a,107,(function(n){e.isNewExpressionTarget(n)&&r(n)}))}}))}(r.symbol,t,i());else{var o=(s=n,e.tryGetClassExtendingExpressionWithTypeArguments(e.climbPastPropertyAccess(s).parent));o&&(function(n,t){var r=N(n.symbol);if(!r||!r.declarations)return;for(var a=0,i=r.declarations;a<i.length;a++){var o=i[a];e.Debug.assert(166===o.kind);var s=o.body;s&&U(s,105,(function(n){e.isCallExpressionTarget(n)&&t(n)}))}}(o,i()),function(e,n){if(function(e){return!!N(e.symbol)}(e))return;var t=e.symbol,r=n.createSearch(void 0,t,void 0);d(t,n,r)}(o,a))}var s}(s,t,a,i);break;case 2:!function(n,t,r){I(n,t.symbol,r);var a=n.parent;if(2===r.options.use||!e.isClassLike(a))return;e.Debug.assert(a.name===n);for(var i=r.referenceAdder(t.symbol),o=0,s=a.members;o<s.length;o++){var c=s[o];e.isMethodOrAccessor(c)&&e.hasSyntacticModifier(c,32)&&(c.body&&c.body.forEachChild((function n(t){107===t.kind?i(t):e.isFunctionLike(t)||e.isClassLike(t)||t.forEachChild(n)})))}}(s,a,i);break;default:e.Debug.assertNever(i.specialSearchKind)}!function(e,t,r,a){var i=n.getImportOrExportSymbol(e,t,a.checker,1===r.comingFrom);if(!i)return;var o=i.symbol;0===i.kind?V(a.options)||y(o,a):f(e,o,i.exportInfo,a)}(s,c,a,i)}else!function(n,t,r){var a=n.flags,i=n.valueDeclaration,o=r.checker.getShorthandAssignmentValueSymbol(i),s=i&&e.getNameOfDeclaration(i);33554432&a||!s||!t.includes(o)||I(s,o,r)}(c,a,i)}}}}else!i.options.implementations&&(i.options.findInStrings&&e.isInString(t,r)||i.options.findInComments&&e.isInNonReferenceComment(t,r))&&i.addStringOrCommentReference(t.fileName,e.createTextSpan(r,a.text.length))}function D(t,r,a,i,o,s,c){e.Debug.assert(!c||!!o.options.providePrefixAndSuffixTextForRename,"If alwaysGetReferences is true, then prefix/suffix text must be enabled");var l=a.parent,u=a.propertyName,d=a.name,p=l.parent,m=C(t,r,a,o.checker);if(c||i.includes(m)){if(u?t===u?(p.moduleSpecifier||b(),s&&2!==o.options.use&&o.markSeenReExportRHS(d)&&I(d,e.Debug.checkDefined(a.symbol),o)):o.markSeenReExportRHS(t)&&b():2===o.options.use&&"default"===d.escapedText||b(),!V(o.options)||c){var _=87===t.originalKeywordKind||87===a.name.originalKeywordKind?1:0,g=e.Debug.checkDefined(a.symbol),v=n.getExportInfo(g,_,o.checker);v&&f(t,g,v,o)}if(1!==i.comingFrom&&p.moduleSpecifier&&!u&&!V(o.options)){var h=o.checker.getExportSpecifierLocalTargetSymbol(a);h&&y(h,o)}}function b(){s&&I(t,m,o)}}function C(n,t,r,a){return function(n,t){var r=t.parent,a=t.propertyName,i=t.name;return e.Debug.assert(a===n||i===n),a?a===n:!r.parent.moduleSpecifier}(n,r)&&a.getExportSpecifierLocalTargetSymbol(r)||t}function I(n,t,r){var a="kind"in t?t:{kind:void 0,symbol:t},i=a.kind,o=a.symbol,s=r.referenceAdder(o);r.options.implementations?function(n,t,r){if(e.isDeclarationName(n)&&(a=n.parent,8388608&a.flags?!e.isInterfaceDeclaration(a)&&!e.isTypeAliasDeclaration(a):e.isVariableLike(a)?e.hasInitializer(a):e.isFunctionLikeDeclaration(a)?a.body:e.isClassLike(a)||e.isModuleOrEnumDeclaration(a)))return void t(n);var a;if(78!==n.kind)return;289===n.parent.kind&&B(n,r.checker,t);var i=R(n);if(i)return void t(i);var o=e.findAncestor(n,(function(n){return!e.isQualifiedName(n.parent)&&!e.isTypeNode(n.parent)&&!e.isTypeElement(n.parent)})),s=o.parent;if(e.hasType(s)&&s.type===o&&r.markSeenContainingTypeReference(s))if(e.hasInitializer(s))l(s.initializer);else if(e.isFunctionLike(s)&&s.body){var c=s.body;230===c.kind?e.forEachReturnStatement(c,(function(e){e.expression&&l(e.expression)})):l(c)}else e.isAssertionExpression(s)&&l(s.expression);function l(e){k(e)&&t(e)}}(n,s,r):s(n,i)}function N(e){return e.members&&e.members.get("__constructor")}function R(n){return e.isIdentifier(n)||e.isPropertyAccessExpression(n)?R(n.parent):e.isExpressionWithTypeArguments(n)?e.tryCast(n.parent.parent,e.isClassLike):void 0}function k(e){switch(e.kind){case 207:return k(e.expression);case 209:case 208:case 200:case 221:case 199:return!0;default:return!1}}function O(n,t,r,a){if(n===t)return!0;var i=e.getSymbolId(n)+","+e.getSymbolId(t),o=r.get(i);if(void 0!==o)return o;r.set(i,!1);var s=!!n.declarations&&n.declarations.some((function(n){return e.getAllSuperTypeNodes(n).some((function(e){var n=a.getTypeAtLocation(e);return!!n&&!!n.symbol&&O(n.symbol,t,r,a)}))}));return r.set(i,s),s}function P(e){return 78===e.kind&&160===e.parent.kind&&e.parent.name===e}function M(e,n,t,r,a,i){var o=[];return w(e,n,t,r,!(r&&a),(function(n,t,r){r&&F(e)!==F(r)&&(r=void 0),o.push(r||t||n)}),(function(){return!i})),o}function w(n,t,r,i,o,s,c){var l=e.getContainingObjectLiteralElement(t);if(l){var u=r.getShorthandAssignmentValueSymbol(t.parent);if(u&&i)return s(u,void 0,void 0,3);var d=r.getContextualType(l.parent),p=d&&e.firstDefined(e.getPropertySymbolsFromContextualType(l,r,d,!0),(function(e){return x(e,4)}));if(p)return p;var m=function(n,t){return e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent.parent)?t.getPropertySymbolOfDestructuringAssignment(n):void 0}(t,r),f=m&&s(m,void 0,void 0,4);if(f)return f;var _=u&&s(u,void 0,void 0,3);if(_)return _}var g=a(t,n,r);if(g){var y=s(g,void 0,void 0,1);if(y)return y}var v=x(n);if(v)return v;if(n.valueDeclaration&&e.isParameterPropertyDeclaration(n.valueDeclaration,n.valueDeclaration.parent)){var h=r.getSymbolsOfParameterPropertyDeclaration(e.cast(n.valueDeclaration,e.isParameter),n.name);return e.Debug.assert(2===h.length&&!!(1&h[0].flags)&&!!(4&h[1].flags)),x(1&n.flags?h[1]:h[0])}var b=e.getDeclarationOfKind(n,270);if(!i||b&&!b.propertyName){var E=b&&r.getExportSpecifierLocalTargetSymbol(b);if(E){var T=s(E,void 0,void 0,1);if(T)return T}}if(!i){var S=void 0;return(S=o?e.isObjectBindingElementWithoutPropertyName(t.parent)?e.getPropertySymbolFromBindingElement(r,t.parent):void 0:L(n,r))&&x(S,4)}if(e.Debug.assert(i),o)return(S=L(n,r))&&x(S,4);function x(n,t){return e.firstDefined(r.getRootSymbols(n),(function(a){return s(n,a,void 0,t)||(a.parent&&96&a.parent.flags&&c(a)?e.getPropertySymbolsFromBaseTypes(a.parent,a.name,r,(function(e){return s(n,a,e,t)})):void 0)}))}function L(n,t){var r=e.getDeclarationOfKind(n,198);if(r&&e.isObjectBindingElementWithoutPropertyName(r))return e.getPropertySymbolFromBindingElement(t,r)}}function F(n){return!!n.valueDeclaration&&!!(32&e.getEffectiveModifierFlags(n.valueDeclaration))}function G(n,t){var r=e.getMeaningFromLocation(n),a=t.declarations;if(a){var i=void 0;do{i=r;for(var o=0,s=a;o<s.length;o++){var c=s[o],l=e.getMeaningFromDeclaration(c);l&r&&(r|=l)}}while(r!==i)}return r}function B(n,t,r){var a=t.getSymbolAtLocation(n),i=t.getShorthandAssignmentValueSymbol(a.valueDeclaration);if(i)for(var o=0,s=i.getDeclarations();o<s.length;o++){var c=s[o];1&e.getMeaningFromDeclaration(c)&&r(c)}}function U(n,t,r){e.forEachChild(n,(function(e){e.kind===t&&r(e),U(e,t,r)}))}function V(e){return 2===e.use&&e.providePrefixAndSuffixTextForRename}t.eachExportReference=function(t,r,a,i,o,s,c,l){for(var u=n.createImportTracker(t,new e.Set(t.map((function(e){return e.fileName}))),r,a)(i,{exportKind:c?1:0,exportingModuleSymbol:o},!1),d=u.importSearches,p=u.indirectUsers,m=0,f=d;m<f.length;m++){l(f[m][0])}for(var _=0,g=p;_<g.length;_++)for(var y=0,v=b(g[_],c?"default":s);y<v.length;y++){var h=v[y];e.isIdentifier(h)&&!e.isImportOrExportSpecifier(h.parent)&&r.getSymbolAtLocation(h)===i&&l(h)}},t.isSymbolReferencedInFile=function(e,n,t,r){return void 0===r&&(r=t),h(e,n,t,(function(){return!0}),r)||!1},t.eachSymbolReferenceInFile=h,t.eachSignatureCall=function(n,t,r,a){if(n.name&&e.isIdentifier(n.name))for(var i=e.Debug.checkDefined(r.getSymbolAtLocation(n.name)),o=0,s=t;o<s.length;o++)for(var c=0,l=b(s[o],i.name);c<l.length;c++){var u=l[c];if(e.isIdentifier(u)&&u!==n.name&&u.escapedText===n.name.escapedText){var d=e.climbPastPropertyAccess(u),p=d.parent;if(e.isCallExpression(p)&&p.expression===d){var m=r.getSymbolAtLocation(u);m&&r.getRootSymbols(m).some((function(e){return e===i}))&&a(p)}}}},t.getIntersectingMeaningFromDeclarations=G,t.getReferenceEntriesForShorthandPropertyAssignment=B}(t=n.Core||(n.Core={}))}(e.FindAllReferences||(e.FindAllReferences={}))}(ts||(ts={})),function(e){!function(n){function t(n){return(e.isFunctionExpression(n)||e.isArrowFunction(n)||e.isClassExpression(n))&&e.isVariableDeclaration(n.parent)&&n===n.parent.initializer&&e.isIdentifier(n.parent.name)&&!!(2&e.getCombinedNodeFlags(n.parent))}function r(n){return e.isSourceFile(n)||e.isModuleDeclaration(n)||e.isFunctionDeclaration(n)||e.isFunctionExpression(n)||e.isClassDeclaration(n)||e.isClassExpression(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isGetAccessorDeclaration(n)||e.isSetAccessorDeclaration(n)}function a(n){return e.isSourceFile(n)||e.isModuleDeclaration(n)&&e.isIdentifier(n.name)||e.isFunctionDeclaration(n)||e.isClassDeclaration(n)||e.isMethodDeclaration(n)||e.isMethodSignature(n)||e.isGetAccessorDeclaration(n)||e.isSetAccessorDeclaration(n)||function(n){return(e.isFunctionExpression(n)||e.isClassExpression(n))&&e.isNamedDeclaration(n)}(n)||t(n)}function i(n){return e.isSourceFile(n)?n:e.isNamedDeclaration(n)?n.name:t(n)?n.parent.name:e.Debug.checkDefined(n.modifiers&&e.find(n.modifiers,o))}function o(e){return 87===e.kind}function s(e,n){var t=i(n);return t&&e.getSymbolAtLocation(t)}function c(n,t){if(t.body)return t;if(e.isConstructorDeclaration(t))return e.getFirstConstructorWithBody(t.parent);if(e.isFunctionDeclaration(t)||e.isMethodDeclaration(t)){var r=s(n,t);return r&&r.valueDeclaration&&e.isFunctionLikeDeclaration(r.valueDeclaration)&&r.valueDeclaration.body?r.valueDeclaration:void 0}return t}function l(n,t){var r,i=s(n,t);if(i&&i.declarations){var o=e.indicesOf(i.declarations),c=e.map(i.declarations,(function(e){return{file:e.getSourceFile().fileName,pos:e.pos}}));o.sort((function(n,t){return e.compareStringsCaseSensitive(c[n].file,c[t].file)||c[n].pos-c[t].pos}));for(var l=void 0,u=0,d=e.map(o,(function(e){return i.declarations[e]}));u<d.length;u++){var p=d[u];a(p)&&(l&&l.parent===p.parent&&l.end===p.pos||(r=e.append(r,p)),l=p)}}return r}function u(n,t){var r,a,i;return e.isFunctionLikeDeclaration(t)?null!==(a=null!==(r=c(n,t))&&void 0!==r?r:l(n,t))&&void 0!==a?a:t:null!==(i=l(n,t))&&void 0!==i?i:t}function d(n,i){for(var o=n.getTypeChecker(),s=!1;;){if(a(i))return u(o,i);var c;if(r(i))return(c=e.findAncestor(i,a))&&u(o,c);if(e.isDeclarationName(i))return a(i.parent)?u(o,i.parent):r(i.parent)?(c=e.findAncestor(i.parent,a))&&u(o,c):e.isVariableDeclaration(i.parent)&&i.parent.initializer&&t(i.parent.initializer)?i.parent.initializer:void 0;if(e.isConstructorDeclaration(i))return a(i.parent)?i.parent:void 0;if(e.isVariableDeclaration(i)&&i.initializer&&t(i.initializer))return i.initializer;if(!s){var l=o.getSymbolAtLocation(i);if(l&&(2097152&l.flags&&(l=o.getAliasedSymbol(l)),l.valueDeclaration)){s=!0,i=l.valueDeclaration;continue}}return}}function p(n,r){var a=r.getSourceFile(),i=function(n,r){if(e.isSourceFile(r))return{text:r.fileName,pos:0,end:0};if((e.isFunctionDeclaration(r)||e.isClassDeclaration(r))&&!e.isNamedDeclaration(r)){var a=r.modifiers&&e.find(r.modifiers,o);if(a)return{text:"default",pos:a.getStart(),end:a.getEnd()}}var i=t(r)?r.parent.name:e.Debug.checkDefined(e.getNameOfDeclaration(r),"Expected call hierarchy item to have a name"),s=e.isIdentifier(i)?e.idText(i):e.isStringOrNumericLiteralLike(i)?i.text:e.isComputedPropertyName(i)&&e.isStringOrNumericLiteralLike(i.expression)?i.expression.text:void 0;if(void 0===s){var c=n.getTypeChecker(),l=c.getSymbolAtLocation(i);l&&(s=c.symbolToString(l,r))}if(void 0===s){var u=e.createPrinter({removeComments:!0,omitTrailingSemicolon:!0});s=e.usingSingleLineStringWriter((function(e){return u.writeNode(4,r,r.getSourceFile(),e)}))}return{text:s,pos:i.getStart(),end:i.getEnd()}}(n,r),s=function(n){var r,a;if(t(n))return e.isModuleBlock(n.parent.parent.parent.parent)&&e.isIdentifier(n.parent.parent.parent.parent.parent.name)?n.parent.parent.parent.parent.parent.name.getText():void 0;switch(n.kind){case 167:case 168:case 165:return 200===n.parent.kind?null===(r=e.getAssignedName(n.parent))||void 0===r?void 0:r.getText():null===(a=e.getNameOfDeclaration(n.parent))||void 0===a?void 0:a.getText();case 251:case 252:case 256:if(e.isModuleBlock(n.parent)&&e.isIdentifier(n.parent.parent.name))return n.parent.parent.name.getText()}}(r),c=e.getNodeKind(r),l=e.getNodeModifiers(r),u=e.createTextSpanFromBounds(e.skipTrivia(a.text,r.getFullStart(),!1,!0),r.getEnd()),d=e.createTextSpanFromBounds(i.pos,i.end);return{file:a.fileName,kind:c,kindModifiers:l,name:i.text,containerName:s,span:u,selectionSpan:d}}function m(e){return void 0!==e}function f(n){if(1===n.kind){var t=n.node;if(e.isCallOrNewExpressionTarget(t,!0,!0)||e.isTaggedTemplateTag(t,!0,!0)||e.isDecoratorTarget(t,!0,!0)||e.isJsxOpeningLikeElementTagName(t,!0,!0)||e.isRightSideOfPropertyAccess(t)||e.isArgumentExpressionOfElementAccess(t)){var r=t.getSourceFile();return{declaration:e.findAncestor(t,a)||r,range:e.createTextRangeFromNode(t,r)}}}}function _(n){return e.getNodeId(n.declaration)}function g(n,t){var r=[],i=function(n,t){function r(r){var a=e.isTaggedTemplateExpression(r)?r.tag:e.isJsxOpeningLikeElement(r)?r.tagName:e.isAccessExpression(r)?r:r.expression,i=d(n,a);if(i){var o=e.createTextRangeFromNode(a,r.getSourceFile());if(e.isArray(i))for(var s=0,c=i;s<c.length;s++){var l=c[s];t.push({declaration:l,range:o})}else t.push({declaration:i,range:o})}}return function n(t){if(t&&!(8388608&t.flags))if(a(t)){if(e.isClassLike(t))for(var i=0,o=t.members;i<o.length;i++){var s=o[i];s.name&&e.isComputedPropertyName(s.name)&&n(s.name.expression)}}else{switch(t.kind){case 78:case 260:case 261:case 267:case 253:case 254:return;case 206:case 224:return void n(t.expression);case 249:case 160:return n(t.name),void n(t.initializer);case 203:case 204:return r(t),n(t.expression),void e.forEach(t.arguments,n);case 205:return r(t),n(t.tag),void n(t.template);case 275:case 274:return r(t),n(t.tagName),void n(t.attributes);case 161:return r(t),void n(t.expression);case 201:case 202:r(t),e.forEachChild(t,n)}e.isPartOfTypeNode(t)||e.forEachChild(t,n)}}}(n,r);switch(t.kind){case 297:!function(n,t){e.forEach(n.statements,t)}(t,i);break;case 256:!function(n,t){!e.hasSyntacticModifier(n,2)&&n.body&&e.isModuleBlock(n.body)&&e.forEach(n.body.statements,t)}(t,i);break;case 251:case 208:case 209:case 165:case 167:case 168:!function(n,t,r){var a=c(n,t);a&&(e.forEach(a.parameters,r),r(a.body))}(n.getTypeChecker(),t,i);break;case 252:case 221:!function(n,t){e.forEach(n.decorators,t);var r=e.getClassExtendsHeritageElement(n);r&&t(r.expression);for(var a=0,i=n.members;a<i.length;a++){var o=i[a];e.forEach(o.decorators,t),e.isPropertyDeclaration(o)?t(o.initializer):e.isConstructorDeclaration(o)&&o.body&&(e.forEach(o.parameters,t),t(o.body))}}(t,i);break;default:e.Debug.assertNever(t)}return r}n.resolveCallHierarchyDeclaration=d,n.createCallHierarchyItem=p,n.getIncomingCalls=function(n,t,r){if(e.isSourceFile(t)||e.isModuleDeclaration(t))return[];var a=i(t),o=e.filter(e.FindAllReferences.findReferenceOrRenameEntries(n,r,n.getSourceFiles(),a,0,{use:1},f),m);return o?e.group(o,_,(function(t){return function(n,t){return r=p(n,t[0].declaration),a=e.map(t,(function(n){return e.createTextSpanFromRange(n.range)})),{from:r,fromSpans:a};var r,a}(n,t)})):[]},n.getOutgoingCalls=function(n,t){return 8388608&t.flags||e.isMethodSignature(t)?[]:e.group(g(n,t),_,(function(t){return function(n,t){return r=p(n,t[0].declaration),a=e.map(t,(function(n){return e.createTextSpanFromRange(n.range)})),{to:r,fromSpans:a};var r,a}(n,t)}))}}(e.CallHierarchy||(e.CallHierarchy={}))}(ts||(ts={})),function(e){function n(n,r,a,i){var o=a(n);return function(n){var s=i&&i.tryGetSourcePosition({fileName:n,pos:0}),c=function(n){if(a(n)===o)return r;var t=e.tryRemoveDirectoryPrefix(n,o,a);return void 0===t?void 0:r+"/"+t}(s?s.fileName:n);return s?void 0===c?void 0:function(n,r,a,i){var o=e.getRelativePathFromFile(n,r,i);return t(e.getDirectoryPath(a),o)}(s.fileName,c,n,a):c}}function t(n,t){return e.ensurePathIsNonModuleName(function(n,t){return e.normalizePath(e.combinePaths(n,t))}(n,t))}function r(n,t,r){if(n){if(n.resolvedModule){var a=s(n.resolvedModule.resolvedFileName);if(a)return a}var i=e.forEach(n.failedLookupLocations,(function(n){var a=t(n);return a&&e.find(r,(function(e){return e.fileName===a}))?o(n):void 0}))||e.forEach(n.failedLookupLocations,o);return i||n.resolvedModule&&{newFileName:n.resolvedModule.resolvedFileName,updated:!1}}function o(n){return e.endsWith(n,"/package.json")?void 0:s(n)}function s(e){var n=t(e);return n&&{newFileName:n,updated:!0}}}function a(n,t){return e.createRange(n.getStart(t)+1,n.end-1)}function i(n,t){if(e.isObjectLiteralExpression(n))for(var r=0,a=n.properties;r<a.length;r++){var i=a[r];e.isPropertyAssignment(i)&&e.isStringLiteral(i.name)&&t(i,i.name.text)}}e.getEditsForFileRename=function(o,s,c,l,u,d,p){var m=e.hostUsesCaseSensitiveFileNames(l),f=e.createGetCanonicalFileName(m),_=n(s,c,f,p),g=n(c,s,f,p);return e.textChanges.ChangeTracker.with({host:l,formatContext:u,preferences:d},(function(n){!function(n,r,o,s,c,l,u){var d=n.getCompilerOptions().configFile;if(!d)return;var p=e.getDirectoryPath(d.fileName),m=e.getTsConfigObjectLiteralExpression(d);if(!m)return;function f(n){for(var t=!1,r=0,a=e.isArrayLiteralExpression(n.initializer)?n.initializer.elements:[n.initializer];r<a.length;r++){t=_(a[r])||t}return t}function _(n){if(!e.isStringLiteral(n))return!1;var i=t(p,n.text),s=o(i);return void 0!==s&&(r.replaceRangeWithText(d,a(n,d),g(s)),!0)}function g(n){return e.getRelativePathFromDirectory(p,n,!u)}i(m,(function(n,t){switch(t){case"files":case"include":case"exclude":if(!f(n)&&"include"===t&&e.isArrayLiteralExpression(n.initializer)){var a=e.mapDefined(n.initializer.elements,(function(n){return e.isStringLiteral(n)?n.text:void 0})),o=e.getFileMatcherPatterns(p,[],a,u,l);e.getRegexFromPattern(e.Debug.checkDefined(o.includeFilePattern),u).test(s)&&!e.getRegexFromPattern(e.Debug.checkDefined(o.includeFilePattern),u).test(c)&&r.insertNodeAfter(d,e.last(n.initializer.elements),e.factory.createStringLiteral(g(c)))}break;case"compilerOptions":i(n.initializer,(function(n,t){var r=e.getOptionFromName(t);r&&(r.isFilePath||"list"===r.type&&r.element.isFilePath)?f(n):"paths"===t&&i(n.initializer,(function(n){if(e.isArrayLiteralExpression(n.initializer))for(var t=0,r=n.initializer.elements;t<r.length;t++){_(r[t])}}))}))}}))}(o,n,_,s,c,l.getCurrentDirectory(),m),function(n,i,o,s,c,l){for(var u=n.getSourceFiles(),d=function(d){var p=o(d.fileName),m=null!=p?p:d.fileName,f=e.getDirectoryPath(m),_=s(d.fileName),g=_||d.fileName,y=e.getDirectoryPath(g),v=void 0!==p||void 0!==_;!function(n,t,r,i){for(var o=0,s=n.referencedFiles||e.emptyArray;o<s.length;o++){var c=s[o];void 0!==(d=r(c.fileName))&&d!==n.text.slice(c.pos,c.end)&&t.replaceRangeWithText(n,c,d)}for(var l=0,u=n.imports;l<u.length;l++){var d,p=u[l];void 0!==(d=i(p))&&d!==p.text&&t.replaceRangeWithText(n,a(p,n),d)}}(d,i,(function(n){if(e.pathIsRelative(n)){var r=t(y,n),a=o(r);return void 0===a?void 0:e.ensurePathIsNonModuleName(e.getRelativePathFromDirectory(f,a,l))}}),(function(t){var a=n.getTypeChecker().getSymbolAtLocation(t);if(!a||!a.declarations.some((function(n){return e.isAmbientModule(n)}))){var i=void 0!==_?r(e.resolveModuleName(t.text,g,n.getCompilerOptions(),c),o,u):function(n,t,a,i,o,s){if(n){var c=e.find(n.declarations,e.isSourceFile).fileName,l=s(c);return void 0===l?{newFileName:c,updated:!1}:{newFileName:l,updated:!0}}return r(o.resolveModuleNames?o.getResolvedModuleWithFailedLookupLocationsFromCache&&o.getResolvedModuleWithFailedLookupLocationsFromCache(t.text,a.fileName):i.getResolvedModuleWithFailedLookupLocationsFromCache(t.text,a.fileName),s,i.getSourceFiles())}(a,t,d,n,c,o);return void 0!==i&&(i.updated||v&&e.pathIsRelative(t.text))?e.moduleSpecifiers.updateModuleSpecifier(n.getCompilerOptions(),l(m),i.newFileName,e.createModuleSpecifierResolutionHost(n,c),t.text):void 0}}))},p=0,m=u;p<m.length;p++){d(m[p])}}(o,n,_,g,l,f)}))},e.getPathUpdater=n}(ts||(ts={})),function(e){!function(n){function t(n,t,i){var d,p,m=a(t,i,n);if(m)return[(d=m.fileName,p=m.file.fileName,{fileName:p,textSpan:e.createTextSpanFromBounds(0,0),kind:"script",name:d,containerName:void 0,containerKind:void 0})];var f=e.getTouchingPropertyName(t,i);if(f!==t){var _=f.parent,g=n.getTypeChecker();if(e.isJumpStatementTarget(f)){var y=e.getTargetLabel(f.parent,f.text);return y?[l(g,y,"label",f.text,void 0)]:void 0}var v=function(n,t){var a=t.getSymbolAtLocation(n);for(;a;)if(2097152&a.flags&&o(n,a.declarations[0])){var i=t.getAliasedSymbol(a);if(!i.declarations)break;a=i}else{if(!r(a))break;var s=t.resolveName(a.name,a.valueDeclaration,111551,!1);if(!e.some(null==s?void 0:s.declarations))break;a=s}return a}(f,g);if(!v)return function(n,t){if(!e.isPropertyAccessExpression(n.parent)||n.parent.name!==n)return;var r=t.getTypeAtLocation(n.parent.expression);return e.mapDefined(r.isUnionOrIntersection()?r.types:[r],(function(e){var n=t.getIndexInfoOfType(e,0);return n&&n.declaration&&u(t,n.declaration)}))}(f,g);var h=function(n,t){var r=function(n){var t=e.findAncestor(n,(function(n){return!e.isRightSideOfPropertyAccess(n)})),r=null==t?void 0:t.parent;return r&&e.isCallLikeExpression(r)&&e.getInvokedExpression(r)===t?r:void 0}(t),a=r&&n.getResolvedSignature(r);return e.tryCast(a&&a.declaration,(function(n){return e.isFunctionLike(n)&&!e.isFunctionTypeNode(n)}))}(g,f);if(h&&(!e.isJsxOpeningLikeElement(f.parent)||!function(e){switch(e.kind){case 166:case 175:case 170:return!0;default:return!1}}(h))){var b=u(g,h);if(g.getRootSymbols(v).some((function(n){return function(n,t){return n===t.symbol||n===t.symbol.parent||e.isAssignmentExpression(t.parent)||!e.isCallLikeExpression(t.parent)&&n===t.parent.symbol}(n,h)})))return[b];var E=s(g,v,f,h)||e.emptyArray;return 105===f.kind?__spreadArrays([b],E):__spreadArrays(E,[b])}if(289===f.parent.kind){var T=g.getShorthandAssignmentValueSymbol(v.valueDeclaration);return T?T.declarations.map((function(e){return c(e,g,T,f)})):[]}if(e.isPropertyName(f)&&e.isBindingElement(_)&&e.isObjectBindingPattern(_.parent)&&f===(_.propertyName||_.name)){var S=e.getNameFromPropertyName(f),x=g.getTypeAtLocation(_.parent);return void 0===S?e.emptyArray:e.flatMap(x.isUnion()?x.types:[x],(function(e){var n=e.getProperty(S);return n&&s(g,n,f)}))}var L=e.getContainingObjectLiteralElement(f);if(L){var A=L&&g.getContextualType(L.parent);if(A)return e.flatMap(e.getPropertySymbolsFromContextualType(L,g,A,!1),(function(e){return s(g,e,f)}))}return s(g,v,f)}}function r(n){var t=e.tryCast(n.valueDeclaration,e.isShorthandPropertyAssignment),r=e.tryCast(null==t?void 0:t.parent.parent,e.isAssignmentExpression);return!!r&&2===e.getAssignmentDeclarationKind(r)}function a(e,n,t){var r=d(e.referencedFiles,n);if(r)return(o=t.getSourceFileFromReference(e,r))&&{fileName:r.fileName,file:o};var a=d(e.typeReferenceDirectives,n);if(a){var i=t.getResolvedTypeReferenceDirectives().get(a.fileName);return(o=i&&t.getSourceFile(i.resolvedFileName))&&{fileName:a.fileName,file:o}}var o,s=d(e.libReferenceDirectives,n);return s?(o=t.getLibFileFromReference(s))&&{fileName:s.fileName,file:o}:void 0}function i(n,t,r){return e.flatMap(!n.isUnion()||32&n.flags?[n]:n.types,(function(e){return e.symbol&&s(t,e.symbol,r)}))}function o(n,t){if(78!==n.kind)return!1;if(n.parent===t)return!0;switch(t.kind){case 262:case 260:return!0;case 265:return 264===t.parent.kind;case 198:case 249:return e.isInJSFile(t)&&e.isRequireVariableDeclaration(t,!0);default:return!1}}function s(n,t,r,a){var i=e.filter(t.declarations,(function(n){return n!==a&&(!e.isAssignmentDeclaration(n)||n===t.valueDeclaration)}))||void 0;return function(){if(32&t.flags&&!(19&t.flags)&&(e.isNewExpressionTarget(r)||132===r.kind)){return o((e.find(i,e.isClassLike)||e.Debug.fail("Expected declaration to have at least one class-like declaration")).members,!0)}}()||(e.isCallOrNewExpressionTarget(r)||e.isNameOfFunctionDeclaration(r)?o(i,!1):void 0)||e.map(i,(function(e){return c(e,n,t,r)}));function o(a,i){if(a){var o=a.filter(i?e.isConstructorDeclaration:e.isFunctionLike),s=o.filter((function(e){return!!e.body}));return o.length?0!==s.length?s.map((function(e){return c(e,n,t,r)})):[c(e.last(o),n,t,r)]:void 0}}}function c(n,t,r,a){var i=t.symbolToString(r),o=e.SymbolDisplay.getSymbolKind(t,r,a),s=r.parent?t.symbolToString(r.parent,a):"";return l(t,n,o,i,s)}function l(n,t,r,a,i){var o=e.getNameOfDeclaration(t)||t,s=o.getSourceFile(),c=e.createTextSpanFromNode(o,s);return __assign(__assign({fileName:s.fileName,textSpan:c,kind:r,name:a,containerKind:void 0,containerName:i},e.FindAllReferences.toContextSpan(c,s,e.FindAllReferences.getContextNode(t))),{isLocal:!n.isDeclarationVisible(t)})}function u(e,n){return c(n,e,n.symbol,n)}function d(n,t){return e.find(n,(function(n){return e.textRangeContainsPositionInclusive(n,t)}))}n.getDefinitionAtPosition=t,n.getReferenceAtPosition=a,n.getTypeDefinitionAtPosition=function(n,t,r){var a=e.getTouchingPropertyName(t,r);if(a!==t){var o=n.getSymbolAtLocation(a);if(o){var s=n.getTypeOfSymbolAtLocation(o,a),c=function(n,t,r){if(t.symbol===n||n.valueDeclaration&&t.symbol&&e.isVariableDeclaration(n.valueDeclaration)&&n.valueDeclaration.initializer===t.symbol.valueDeclaration){var a=t.getCallSignatures();if(1===a.length)return r.getReturnTypeOfSignature(e.first(a))}return}(o,s,n),l=c&&i(c,n,a);return l&&0!==l.length?l:i(s,n,a)}}},n.getDefinitionAndBoundSpan=function(n,r,a){var i=t(n,r,a);if(i&&0!==i.length){var o=d(r.referencedFiles,a)||d(r.typeReferenceDirectives,a)||d(r.libReferenceDirectives,a);if(o)return{definitions:i,textSpan:e.createTextSpanFromRange(o)};var s=e.getTouchingPropertyName(r,a);return{definitions:i,textSpan:e.createTextSpan(s.getStart(),s.getWidth())}}},n.findReferenceInPosition=d}(e.GoToDefinition||(e.GoToDefinition={}))}(ts||(ts={})),function(e){!function(n){var t,r,a=["abstract","access","alias","argument","async","augments","author","borrows","callback","class","classdesc","constant","constructor","constructs","copyright","default","deprecated","description","emits","enum","event","example","exports","extends","external","field","file","fileoverview","fires","function","generator","global","hideconstructor","host","ignore","implements","inheritdoc","inner","instance","interface","kind","lends","license","listens","member","memberof","method","mixes","module","name","namespace","override","package","param","private","property","protected","public","readonly","requires","returns","see","since","static","summary","template","this","throws","todo","tutorial","type","typedef","var","variation","version","virtual","yields"];function i(e){var n=e.comment;switch(e.kind){case 316:case 315:return r(e.class);case 330:return a(e.typeParameters.map((function(e){return e.getText()})).join(", "));case 329:return r(e.typeExpression);case 331:case 324:case 333:case 326:case 332:var t=e.name;return t?r(t):n;default:return n}function r(e){return a(e.getText())}function a(e){return void 0===n?e:e+" "+n}}function o(n){return{name:n,kind:"",kindModifiers:"",displayParts:[e.textPart(n)],documentation:e.emptyArray,tags:void 0,codeActions:void 0}}function s(n){switch(n.kind){case 251:case 208:case 165:case 166:case 164:case 209:return{commentOwner:n,parameters:n.parameters};case 288:return s(n.initializer);case 252:case 253:case 162:case 255:case 291:case 254:return{commentOwner:n};case 232:var t=n.declarationList.declarations;return{commentOwner:n,parameters:1===t.length&&t[0].initializer?function(n){for(;207===n.kind;)n=n.expression;switch(n.kind){case 208:case 209:return n.parameters;case 221:var t=e.find(n.members,e.isConstructorDeclaration);return t?t.parameters:e.emptyArray}return e.emptyArray}(t[0].initializer):void 0};case 297:return"quit";case 256:return 256===n.parent.kind?void 0:{commentOwner:n};case 233:return s(n.expression);case 216:var r=n;return 0===e.getAssignmentDeclarationKind(r)?"quit":{commentOwner:n,parameters:e.isFunctionLike(r.right)?r.right.parameters:e.emptyArray};case 163:var a=n.initializer;if(a&&(e.isFunctionExpression(a)||e.isArrowFunction(a)))return{commentOwner:n,parameters:a.parameters}}}n.getJsDocCommentsFromDeclarations=function(n){var t=[];return e.forEachUnique(n,(function(n){for(var r=0,a=function(n){switch(n.kind){case 326:case 333:return[n];case 324:case 331:return[n,n.parent];default:return e.getJSDocCommentsAndTags(n)}}(n);r<a.length;r++){var i=a[r].comment;void 0!==i&&e.pushIfUnique(t,i)}})),e.intersperse(e.map(t,e.textPart),e.lineBreakPart())},n.getJsDocTagsFromDeclarations=function(n){var t=[];return e.forEachUnique(n,(function(n){for(var r=0,a=e.getJSDocTags(n);r<a.length;r++){var o=a[r];t.push({name:o.tagName.text,text:i(o)})}})),t},n.getJSDocTagNameCompletions=function(){return t||(t=e.map(a,(function(n){return{name:n,kind:"keyword",kindModifiers:"",sortText:e.Completions.SortText.LocationPriority}})))},n.getJSDocTagNameCompletionDetails=o,n.getJSDocTagCompletions=function(){return r||(r=e.map(a,(function(n){return{name:"@"+n,kind:"keyword",kindModifiers:"",sortText:e.Completions.SortText.LocationPriority}})))},n.getJSDocTagCompletionDetails=o,n.getJSDocParameterNameCompletions=function(n){if(!e.isIdentifier(n.name))return e.emptyArray;var t=n.name.text,r=n.parent,a=r.parent;return e.isFunctionLike(a)?e.mapDefined(a.parameters,(function(a){if(e.isIdentifier(a.name)){var i=a.name.text;if(!r.tags.some((function(t){return t!==n&&e.isJSDocParameterTag(t)&&e.isIdentifier(t.name)&&t.name.escapedText===i}))&&(void 0===t||e.startsWith(i,t)))return{name:i,kind:"parameter",kindModifiers:"",sortText:e.Completions.SortText.LocationPriority}}})):[]},n.getJSDocParameterNameCompletionDetails=function(n){return{name:n,kind:"parameter",kindModifiers:"",displayParts:[e.textPart(n)],documentation:e.emptyArray,tags:void 0,codeActions:void 0}},n.getDocCommentTemplateAtPosition=function(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.findAncestor(a,e.isJSDoc);if(!i||void 0===i.comment&&!e.length(i.tags)){var o=a.getStart(t);if(i||!(o<r)){var c=function(n){return e.forEachAncestor(n,s)}(a);if(c){var l=c.commentOwner,u=c.parameters;if(!(l.getStart(t)<r)){if(!u||0===u.length){return{newText:"/** */",caretOffset:3}}var d=function(n,t){for(var r=n.text,a=e.getLineStartPositionForPosition(t,n),i=a;i<=t&&e.isWhiteSpaceSingleLine(r.charCodeAt(i));i++);return r.slice(a,i)}(t,r),p="/**"+n+d+" * ";return{newText:p+n+function(e,n,t,r){return e.map((function(e,a){var i=e.name,o=e.dotDotDotToken,s=78===i.kind?i.text:"param"+a;return t+" * @param "+(n?o?"{...any} ":"{any} ":"")+s+r})).join("")}(u,e.hasJSFileExtension(t.fileName),d,n)+d+" */"+(o===r?n+d:""),caretOffset:p.length}}}}}}}(e.JsDoc||(e.JsDoc={}))}(ts||(ts={})),function(e){!function(n){function t(e,n){switch(e.kind){case 262:case 265:case 260:var t=n.getSymbolAtLocation(e.name),r=n.getAliasedSymbol(t);return t.escapedName!==r.escapedName;default:return!0}}function r(n,t){var r=e.getNameOfDeclaration(n);return!!r&&(i(r,t)||158===r.kind&&a(r.expression,t))}function a(n,t){return i(n,t)||e.isPropertyAccessExpression(n)&&(t.push(n.name.text),!0)&&a(n.expression,t)}function i(n,t){return e.isPropertyNameLiteral(n)&&(t.push(e.getTextOfIdentifierOrLiteral(n)),!0)}function o(n){var t=[],i=e.getNameOfDeclaration(n);if(i&&158===i.kind&&!a(i.expression,t))return e.emptyArray;t.shift();for(var o=e.getContainerNode(n);o;){if(!r(o,t))return e.emptyArray;o=e.getContainerNode(o)}return t.reverse()}function s(n,t){return e.compareValues(n.matchKind,t.matchKind)||e.compareStringsCaseSensitiveUI(n.name,t.name)}function c(n){var t=n.declaration,r=e.getContainerNode(t),a=r&&e.getNameOfDeclaration(r);return{name:n.name,kind:e.getNodeKind(t),kindModifiers:e.getNodeModifiers(t),matchKind:e.PatternMatchKind[n.matchKind],isCaseSensitive:n.isCaseSensitive,fileName:n.fileName,textSpan:e.createTextSpanFromNode(t),containerName:a?a.text:"",containerKind:a?e.getNodeKind(r):""}}n.getNavigateToItems=function(n,r,a,i,l,u){var d=e.createPatternMatcher(i);if(!d)return e.emptyArray;for(var p=[],m=function(e){if(a.throwIfCancellationRequested(),u&&e.isDeclarationFile)return"continue";e.getNamedDeclarations().forEach((function(n,a){!function(e,n,r,a,i,s){var c=e.getMatchForLastSegmentOfPattern(n);if(!c)return;for(var l=0,u=r;l<u.length;l++){var d=u[l];if(t(d,a))if(e.patternContainsDots){var p=e.getFullMatch(o(d),n);p&&s.push({name:n,fileName:i,matchKind:p.kind,isCaseSensitive:p.isCaseSensitive,declaration:d})}else s.push({name:n,fileName:i,matchKind:c.kind,isCaseSensitive:c.isCaseSensitive,declaration:d})}}(d,a,n,r,e.fileName,p)}))},f=0,_=n;f<_.length;f++){m(_[f])}return p.sort(s),(void 0===l?p:p.slice(0,l)).map(c)}}(e.NavigateTo||(e.NavigateTo={}))}(ts||(ts={})),function(e){!function(n){var t,r,a,i,o,s=/\s+/g,c=[],l=[],u=[];function d(){a=void 0,r=void 0,c=[],i=void 0,u=[]}function p(e){return K(e.getText(a))}function m(e){return e.node.kind}function f(e,n){e.children?e.children.push(n):e.children=[n]}function _(n){e.Debug.assert(!c.length);var t={node:n,name:void 0,additionalNodes:void 0,parent:void 0,children:void 0,indent:0};i=t;for(var r=0,a=n.statements;r<a.length;r++){x(a[r])}return T(),e.Debug.assert(!i&&!c.length),t}function g(e,n){f(i,y(e,n))}function y(n,t){return{node:n,name:t||(e.isDeclaration(n)||e.isExpression(n)?e.getNameOfDeclaration(n):void 0),additionalNodes:void 0,parent:i,children:void 0,indent:i.indent+1}}function v(n){o||(o=new e.Map),o.set(n,!0)}function h(e){for(var n=0;n<e;n++)T()}function b(n,t){for(var r=[];!e.isPropertyNameLiteral(t);){var a=e.getNameOrArgument(t),i=e.getElementOrPropertyAccessName(t);t=t.expression,"prototype"===i||e.isPrivateIdentifier(a)||r.push(a)}r.push(t);for(var o=r.length-1;o>0;o--){E(n,a=r[o])}return[r.length-1,r[0]]}function E(e,n){var t=y(e,n);f(i,t),c.push(i),l.push(o),o=void 0,i=t}function T(){i.children&&(L(i.children,i),N(i.children)),i=c.pop(),o=l.pop()}function S(e,n,t){E(e,t),x(n),T()}function x(n){var t;if(r.throwIfCancellationRequested(),n&&!e.isToken(n))switch(n.kind){case 166:var a=n;S(a,a.body);for(var i=0,s=a.parameters;i<s.length;i++){var c=s[i];e.isParameterPropertyDeclaration(c,a)&&g(c)}break;case 165:case 167:case 168:case 164:e.hasDynamicName(n)||S(n,n.body);break;case 163:case 162:e.hasDynamicName(n)||g(n);break;case 262:var l=n;l.name&&g(l.name);var u=l.namedBindings;if(u)if(263===u.kind)g(u);else for(var d=0,p=u.elements;d<p.length;d++){g(p[d])}break;case 289:S(n,n.name);break;case 290:var m=n.expression;e.isIdentifier(m)?g(n,m):g(n);break;case 198:case 288:case 249:var f=n,_=f.name,y=f.initializer;e.isBindingPattern(_)?x(_):y&&function(e){switch(e.kind){case 209:case 208:case 221:return!0;default:return!1}}(y)?(E(n),e.forEachChild(y,x),T()):S(n,y);break;case 251:var L=n.name;L&&e.isIdentifier(L)&&v(L.text),S(n,n.body);break;case 209:case 208:S(n,n.body);break;case 255:E(n);for(var A=0,D=n.members;A<D.length;A++){B(N=D[A])||g(N)}T();break;case 252:case 221:case 253:E(n);for(var C=0,I=n.members;C<I.length;C++){var N;x(N=I[C])}T();break;case 256:S(n,G(n).body);break;case 266:var R=n.expression,k=e.isObjectLiteralExpression(R)||e.isCallExpression(R)?R:e.isArrowFunction(R)||e.isFunctionExpression(R)?R.body:void 0;k?(E(n),x(k),T()):g(n);break;case 270:case 260:case 171:case 169:case 170:case 254:g(n);break;case 203:case 216:var O=e.getAssignmentDeclarationKind(n);switch(O){case 1:case 2:return void S(n,n.right);case 6:case 3:var P=(H=n).left,M=3===O?P.expression:P,w=0,F=void 0;return e.isIdentifier(M.expression)?(v(M.expression.text),F=M.expression):(w=(t=b(H,M.expression))[0],F=t[1]),6===O?e.isObjectLiteralExpression(H.right)&&H.right.properties.length>0&&(E(H,F),e.forEachChild(H.right,x),T()):e.isFunctionExpression(H.right)||e.isArrowFunction(H.right)?S(n,H.right,F):(E(H,F),S(n,H.right,P.name),T()),void h(w);case 7:case 9:var U=n,V=(F=7===O?U.arguments[0]:U.arguments[0].expression,U.arguments[1]),j=b(n,F);w=j[0];return E(n,j[1]),E(n,e.setTextRange(e.factory.createIdentifier(V.text),V)),x(n.arguments[2]),T(),T(),void h(w);case 5:var H,K=(P=(H=n).left).expression;if(e.isIdentifier(K)&&"prototype"!==e.getElementOrPropertyAccessName(P)&&o&&o.has(K.text))return void(e.isFunctionExpression(H.right)||e.isArrowFunction(H.right)?S(n,H.right,K):e.isBindableStaticAccessExpression(P)&&(E(H,K),S(H.left,H.right,e.getNameOrArgument(P)),T()));break;case 4:case 0:case 8:break;default:e.Debug.assertNever(O)}default:e.hasJSDocNodes(n)&&e.forEach(n.jsDoc,(function(n){e.forEach(n.tags,(function(n){e.isJSDocTypeAlias(n)&&g(n)}))})),e.forEachChild(n,x)}}function L(n,t){var r=new e.Map;e.filterMutate(n,(function(n,a){var i=n.name||e.getNameOfDeclaration(n.node),o=i&&p(i);if(!o)return!0;var s=r.get(o);if(!s)return r.set(o,n),!0;if(s instanceof Array){for(var c=0,l=s;c<l.length;c++){var u;if(D(u=l[c],n,a,t))return!1}return s.push(n),!0}return!D(u=s,n,a,t)&&(r.set(o,[u,n]),!0)}))}n.getNavigationBarItems=function(n,t){r=t,a=n;try{return e.map(function(e){var n=[];function t(e){if(r(e)&&(n.push(e),e.children))for(var a=0,i=e.children;a<i.length;a++){t(i[a])}}return t(e),n;function r(e){if(e.children)return!0;switch(m(e)){case 252:case 221:case 255:case 253:case 256:case 297:case 254:case 331:case 324:return!0;case 209:case 251:case 208:return n(e);default:return!1}function n(e){if(!e.node.body)return!1;switch(m(e.parent)){case 257:case 297:case 165:case 166:return!0;default:return!1}}}}(_(n)),M)}finally{d()}},n.getNavigationTree=function(e,n){r=n,a=e;try{return P(_(e))}finally{d()}};var A=((t={})[5]=!0,t[3]=!0,t[7]=!0,t[9]=!0,t[0]=!1,t[1]=!1,t[2]=!1,t[8]=!1,t[6]=!0,t[4]=!1,t);function D(n,t,r,a){return!!function(n,t,r,a){function i(n){return e.isFunctionExpression(n)||e.isFunctionDeclaration(n)||e.isVariableDeclaration(n)}var o=e.isBinaryExpression(t.node)||e.isCallExpression(t.node)?e.getAssignmentDeclarationKind(t.node):0,s=e.isBinaryExpression(n.node)||e.isCallExpression(n.node)?e.getAssignmentDeclarationKind(n.node):0;if(A[o]&&A[s]||i(n.node)&&A[o]||i(t.node)&&A[s]||e.isClassDeclaration(n.node)&&A[o]||e.isClassDeclaration(t.node)&&A[s]||e.isClassDeclaration(n.node)&&i(t.node)||e.isClassDeclaration(t.node)&&i(n.node)){var c=n.additionalNodes&&e.lastOrUndefined(n.additionalNodes)||n.node;if(!e.isClassDeclaration(n.node)&&!e.isClassDeclaration(t.node)||i(n.node)||i(t.node)){var l=i(n.node)?n.node:i(t.node)?t.node:void 0;if(void 0!==l){var u=y(e.setTextRange(e.factory.createConstructorDeclaration(void 0,void 0,[],void 0),l));u.indent=n.indent+1,u.children=n.node===l?n.children:t.children,n.children=n.node===l?e.concatenate([u],t.children||[t]):e.concatenate(n.children||[n],[u])}else(n.children||t.children)&&(n.children=e.concatenate(n.children||[n],t.children||[t]),n.children&&(L(n.children,n),N(n.children)));c=n.node=e.setTextRange(e.factory.createClassDeclaration(void 0,void 0,n.name||e.factory.createIdentifier("__class__"),void 0,void 0,[]),n.node)}else n.children=e.concatenate(n.children,t.children),n.children&&L(n.children,n);var d=t.node;return a.children[r-1].node.end===c.end?e.setTextRange(c,{pos:c.pos,end:d.end}):(n.additionalNodes||(n.additionalNodes=[]),n.additionalNodes.push(e.setTextRange(e.factory.createClassDeclaration(void 0,void 0,n.name||e.factory.createIdentifier("__class__"),void 0,void 0,[]),t.node))),!0}return 0!==o}(n,t,r,a)||!!function(n,t,r){if(n.kind!==t.kind||n.parent!==t.parent&&(!C(n,r)||!C(t,r)))return!1;switch(n.kind){case 163:case 165:case 167:case 168:return e.hasSyntacticModifier(n,32)===e.hasSyntacticModifier(t,32);case 256:return I(n,t);default:return!0}}(n.node,t.node,a)&&(function(n,t){var r;n.additionalNodes=n.additionalNodes||[],n.additionalNodes.push(t.node),t.additionalNodes&&(r=n.additionalNodes).push.apply(r,t.additionalNodes);n.children=e.concatenate(n.children,t.children),n.children&&(L(n.children,n),N(n.children))}(n,t),!0)}function C(n,t){var r=e.isModuleBlock(n.parent)?n.parent.parent:n.parent;return r===t.node||e.contains(t.additionalNodes,r)}function I(e,n){return e.body.kind===n.body.kind&&(256!==e.body.kind||I(e.body,n.body))}function N(e){e.sort(R)}function R(n,t){return e.compareStringsCaseSensitiveUI(k(n.node),k(t.node))||e.compareValues(m(n),m(t))}function k(n){if(256===n.kind)return F(n);var t=e.getNameOfDeclaration(n);if(t&&e.isPropertyName(t)){var r=e.getPropertyNameForPropertyNameNode(t);return r&&e.unescapeLeadingUnderscores(r)}switch(n.kind){case 208:case 209:case 221:return j(n);default:return}}function O(n,t){if(256===n.kind)return K(F(n));if(t){var r=e.isIdentifier(t)?t.text:e.isElementAccessExpression(t)?"["+p(t.argumentExpression)+"]":p(t);if(r.length>0)return K(r)}switch(n.kind){case 297:var a=n;return e.isExternalModule(a)?'"'+e.escapeString(e.getBaseFileName(e.removeFileExtension(e.normalizePath(a.fileName))))+'"':"<global>";case 266:return e.isExportAssignment(n)&&n.isExportEquals?"export=":"default";case 209:case 251:case 208:case 252:case 221:return 512&e.getSyntacticModifierFlags(n)?"default":j(n);case 166:return"constructor";case 170:return"new()";case 169:return"()";case 171:return"[]";default:return"<unknown>"}}function P(n){return{text:O(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:V(n.node),spans:w(n),nameSpan:n.name&&U(n.name),childItems:e.map(n.children,P)}}function M(n){return{text:O(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:V(n.node),spans:w(n),childItems:e.map(n.children,(function(n){return{text:O(n.node,n.name),kind:e.getNodeKind(n.node),kindModifiers:e.getNodeModifiers(n.node),spans:w(n),childItems:u,indent:0,bolded:!1,grayed:!1}}))||u,indent:n.indent,bolded:!1,grayed:!1}}function w(e){var n=[U(e.node)];if(e.additionalNodes)for(var t=0,r=e.additionalNodes;t<r.length;t++){var a=r[t];n.push(U(a))}return n}function F(n){if(e.isAmbientModule(n))return e.getTextOfNode(n.name);for(var t=[e.getTextOfIdentifierOrLiteral(n.name)];n.body&&256===n.body.kind;)n=n.body,t.push(e.getTextOfIdentifierOrLiteral(n.name));return t.join(".")}function G(n){return n.body&&e.isModuleDeclaration(n.body)?G(n.body):n}function B(e){return!e.name||158===e.name.kind}function U(n){return 297===n.kind?e.createTextSpanFromRange(n):e.createTextSpanFromNode(n,a)}function V(n){return n.parent&&249===n.parent.kind&&(n=n.parent),e.getNodeModifiers(n)}function j(n){var t=n.parent;if(n.name&&e.getFullWidth(n.name)>0)return K(e.declarationNameToString(n.name));if(e.isVariableDeclaration(t))return K(e.declarationNameToString(t.name));if(e.isBinaryExpression(t)&&62===t.operatorToken.kind)return p(t.left).replace(s,"");if(e.isPropertyAssignment(t))return p(t.name);if(512&e.getSyntacticModifierFlags(n))return"default";if(e.isClassLike(n))return"<class>";if(e.isCallExpression(t)){var r=H(t.expression);if(void 0!==r)return(r=K(r)).length>150?r+" callback":r+"("+K(e.mapDefined(t.arguments,(function(n){return e.isStringLiteralLike(n)?n.getText(a):void 0})).join(", "))+") callback"}return"<function>"}function H(n){if(e.isIdentifier(n))return n.text;if(e.isPropertyAccessExpression(n)){var t=H(n.expression),r=n.name.text;return void 0===t?r:t+"."+r}}function K(e){return(e=e.length>150?e.substring(0,150)+"...":e).replace(/\\?(\r?\n|\r|\u2028|\u2029)/g,"")}}(e.NavigationBar||(e.NavigationBar={}))}(ts||(ts={})),function(e){!function(n){function t(n,t){var r=e.isStringLiteral(t)&&t.text;return e.isString(r)&&e.some(n.moduleAugmentations,(function(n){return e.isStringLiteral(n)&&n.text===r}))}function r(n){return void 0!==n&&e.isStringLiteralLike(n)?n.text:void 0}function a(n){var t;if(0===n.length)return n;var r=function(n){for(var t,r={defaultImports:[],namespaceImports:[],namedImports:[]},a={defaultImports:[],namespaceImports:[],namedImports:[]},i=0,o=n;i<o.length;i++){var s=o[i];if(void 0!==s.importClause){var c=s.importClause.isTypeOnly?r:a,l=s.importClause,u=l.name,d=l.namedBindings;u&&c.defaultImports.push(s),d&&(e.isNamespaceImport(d)?c.namespaceImports.push(s):c.namedImports.push(s))}else t=t||s}return{importWithoutClause:t,typeOnlyImports:r,regularImports:a}}(n),a=r.importWithoutClause,i=r.typeOnlyImports,c=r.regularImports,l=[];a&&l.push(a);for(var d=0,p=[c,i];d<p.length;d++){var m=p[d],f=m===i,_=m.defaultImports,g=m.namespaceImports,y=m.namedImports;if(f||1!==_.length||1!==g.length||0!==y.length){for(var v=0,h=e.stableSort(g,(function(e,n){return u(e.importClause.namedBindings.name,n.importClause.namedBindings.name)}));v<h.length;v++){var b=h[v];l.push(o(b,void 0,b.importClause.namedBindings))}if(0!==_.length||0!==y.length){var E=void 0,T=[];if(1===_.length)E=_[0].importClause.name;else for(var S=0,x=_;S<x.length;S++){C=x[S];T.push(e.factory.createImportSpecifier(e.factory.createIdentifier("default"),C.importClause.name))}T.push.apply(T,e.flatMap(y,(function(e){return e.importClause.namedBindings.elements})));var L=s(T),A=_.length>0?_[0]:y[0],D=0===L.length?E?void 0:e.factory.createNamedImports(e.emptyArray):0===y.length?e.factory.createNamedImports(L):e.factory.updateNamedImports(y[0].importClause.namedBindings,L);f&&E&&D?(l.push(o(A,E,void 0)),l.push(o(null!==(t=y[0])&&void 0!==t?t:A,void 0,D))):l.push(o(A,E,D))}}else{var C=_[0];l.push(o(C,C.importClause.name,g[0].importClause.namedBindings))}}return l}function i(n){if(0===n.length)return n;var t=function(e){for(var n,t=[],r=[],a=0,i=e;a<i.length;a++){var o=i[a];void 0===o.exportClause?n=n||o:o.isTypeOnly?r.push(o):t.push(o)}return{exportWithoutClause:n,namedExports:t,typeOnlyExports:r}}(n),r=t.exportWithoutClause,a=t.namedExports,i=t.typeOnlyExports,o=[];r&&o.push(r);for(var c=0,l=[a,i];c<l.length;c++){var u=l[c];if(0!==u.length){var d=[];d.push.apply(d,e.flatMap(u,(function(n){return n.exportClause&&e.isNamedExports(n.exportClause)?n.exportClause.elements:e.emptyArray})));var p=s(d),m=u[0];o.push(e.factory.updateExportDeclaration(m,m.decorators,m.modifiers,m.isTypeOnly,m.exportClause&&(e.isNamedExports(m.exportClause)?e.factory.updateNamedExports(m.exportClause,p):e.factory.updateNamespaceExport(m.exportClause,m.exportClause.name)),m.moduleSpecifier))}}return o}function o(n,t,r){return e.factory.updateImportDeclaration(n,n.decorators,n.modifiers,e.factory.updateImportClause(n.importClause,n.importClause.isTypeOnly,t,r),n.moduleSpecifier)}function s(n){return e.stableSort(n,c)}function c(e,n){return u(e.propertyName||e.name,n.propertyName||n.name)||u(e.name,n.name)}function l(n,t){var a=void 0===n?void 0:r(n),i=void 0===t?void 0:r(t);return e.compareBooleans(void 0===a,void 0===i)||e.compareBooleans(e.isExternalModuleNameRelative(a),e.isExternalModuleNameRelative(i))||e.compareStringsCaseInsensitive(a,i)}function u(n,t){return e.compareStringsCaseInsensitive(n.text,t.text)}function d(n){var t;switch(n.kind){case 260:return null===(t=e.tryCast(n.moduleReference,e.isExternalModuleReference))||void 0===t?void 0:t.expression;case 261:return n.moduleSpecifier;case 232:return n.declarationList.declarations[0].initializer.arguments[0]}}function p(n,t){return l(d(n),d(t))||function(n,t){return e.compareValues(m(n),m(t))}(n,t)}function m(e){var n;switch(e.kind){case 261:return e.importClause?e.importClause.isTypeOnly?1:263===(null===(n=e.importClause.namedBindings)||void 0===n?void 0:n.kind)?2:e.importClause.name?3:4:0;case 260:return 5;case 232:return 6}}n.organizeImports=function(n,s,c,u,d){var m=e.textChanges.ChangeTracker.fromContext({host:c,formatContext:s,preferences:d}),f=function(r){return e.stableSort(a(function(n,r,a){for(var i=a.getTypeChecker(),s=i.getJsxNamespace(r),c=!!(2&r.transformFlags),l=[],u=0,d=n;u<d.length;u++){var p=d[u],m=p.importClause,f=p.moduleSpecifier;if(m){var _=m.name,g=m.namedBindings;if(_&&!v(_)&&(_=void 0),g)if(e.isNamespaceImport(g))v(g.name)||(g=void 0);else{var y=g.elements.filter((function(e){return v(e.name)}));y.length<g.elements.length&&(g=y.length?e.factory.updateNamedImports(g,y):void 0)}_||g?l.push(o(p,_,g)):t(r,f)&&(r.isDeclarationFile?l.push(e.factory.createImportDeclaration(p.decorators,p.modifiers,void 0,f)):l.push(p))}else l.push(p)}return l;function v(n){return c&&n.text===s||e.FindAllReferences.Core.isSymbolReferencedInFile(n,i,r)}}(r,n,u)),(function(e,n){return p(e,n)}))};v(n.statements.filter(e.isImportDeclaration),f),v(n.statements.filter(e.isExportDeclaration),i);for(var _=0,g=n.statements.filter(e.isAmbientModule);_<g.length;_++){var y=g[_];if(y.body)v(y.body.statements.filter(e.isImportDeclaration),f),v(y.body.statements.filter(e.isExportDeclaration),i)}return m.getChanges();function v(t,a){if(0!==e.length(t)){e.suppressLeadingTrivia(t[0]);var i=e.group(t,(function(e){return r(e.moduleSpecifier)})),o=e.stableSort(i,(function(e,n){return l(e[0].moduleSpecifier,n[0].moduleSpecifier)})),u=e.flatMap(o,(function(e){return r(e[0].moduleSpecifier)?a(e):e}));0===u.length?m.delete(n,t[0]):m.replaceNodeWithNodes(n,t[0],u,{leadingTriviaOption:e.textChanges.LeadingTriviaOption.Exclude,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Include,suffix:e.getNewLineOrDefaultFromHost(c,s.options)});for(var d=1;d<t.length;d++)m.deleteNode(n,t[d])}}},n.coalesceImports=a,n.coalesceExports=i,n.compareImportOrExportSpecifiers=c,n.compareModuleSpecifiers=l,n.importsAreSorted=function(n){return e.arrayIsSorted(n,p)},n.importSpecifiersAreSorted=function(n){return e.arrayIsSorted(n,c)},n.getImportDeclarationInsertionIndex=function(n,t){var r=e.binarySearch(n,t,e.identity,p);return r<0?~r:r},n.getImportSpecifierInsertionIndex=function(n,t){var r=e.binarySearch(n,t,e.identity,c);return r<0?~r:r},n.compareImportsOrRequireStatements=p}(e.OrganizeImports||(e.OrganizeImports={}))}(ts||(ts={})),function(e){!function(n){n.collectElements=function(n,t){var c=[];return function(n,t,r){var c=40,l=0,u=__spreadArrays(n.statements,[n.endOfFileToken]),d=u.length;for(;l<d;){for(;l<d&&!e.isAnyImportSyntax(u[l]);)f(u[l]),l++;if(l===d)break;for(var p=l;l<d&&e.isAnyImportSyntax(u[l]);)a(u[l],n,t,r),l++;var m=l-1;m!==p&&r.push(i(e.findChildOfKind(u[p],99,n).getStart(n),u[m].getEnd(),"imports"))}function f(l){var u;if(0!==c){t.throwIfCancellationRequested(),(e.isDeclaration(l)||e.isVariableStatement(l)||1===l.kind)&&a(l,n,t,r),e.isFunctionLike(l)&&e.isBinaryExpression(l.parent)&&e.isPropertyAccessExpression(l.parent.left)&&a(l.parent.left,n,t,r);var d=function(n,t){switch(n.kind){case 230:if(e.isFunctionLike(n.parent))return function(n,t,r){var a=function(n,t,r){if(e.isNodeArrayMultiLine(n.parameters,r)){var a=e.findChildOfKind(n,20,r);if(a)return a}return e.findChildOfKind(t,18,r)}(n,t,r),i=e.findChildOfKind(t,19,r);return a&&i&&o(a,i,n,r,209!==n.kind)}(n.parent,n,t);switch(n.parent.kind){case 235:case 238:case 239:case 237:case 234:case 236:case 243:case 287:return _(n.parent);case 247:var r=n.parent;if(r.tryBlock===n)return _(n.parent);if(r.finallyBlock===n){var a=e.findChildOfKind(r,95,t);if(a)return _(a)}default:return s(e.createTextSpanFromNode(n,t),"code")}case 257:return _(n.parent);case 252:case 221:case 253:case 255:case 258:case 177:case 196:return _(n);case 179:return _(n,!1,!e.isTupleTypeNode(n.parent),22);case 284:case 285:return g(n.statements);case 200:return f(n);case 199:return f(n,22);case 273:return u(n);case 277:return d(n);case 274:case 275:return p(n.attributes);case 218:case 14:return m(n);case 197:return _(n,!1,!e.isBindingElement(n.parent),22);case 209:return l(n);case 203:return c(n)}function c(n){if(n.arguments.length){var r=e.findChildOfKind(n,20,t),a=e.findChildOfKind(n,21,t);if(r&&a&&!e.positionsAreOnSameLine(r.pos,a.pos,t))return o(r,a,n,t,!1,!0)}}function l(n){if(!e.isBlock(n.body)&&!e.positionsAreOnSameLine(n.body.getFullStart(),n.body.getEnd(),t))return s(e.createTextSpanFromBounds(n.body.getFullStart(),n.body.getEnd()),"code",e.createTextSpanFromNode(n))}function u(n){var r=e.createTextSpanFromBounds(n.openingElement.getStart(t),n.closingElement.getEnd()),a=n.openingElement.tagName.getText(t);return s(r,"code",r,!1,"<"+a+">...</"+a+">")}function d(n){var r=e.createTextSpanFromBounds(n.openingFragment.getStart(t),n.closingFragment.getEnd());return s(r,"code",r,!1,"<>...</>")}function p(e){if(0!==e.properties.length)return i(e.getStart(t),e.getEnd(),"code")}function m(e){if(14!==e.kind||0!==e.text.length)return i(e.getStart(t),e.getEnd(),"code")}function f(n,t){return void 0===t&&(t=18),_(n,!1,!e.isArrayLiteralExpression(n.parent)&&!e.isCallExpression(n.parent),t)}function _(r,a,i,s,c){void 0===a&&(a=!1),void 0===i&&(i=!0),void 0===s&&(s=18),void 0===c&&(c=18===s?19:23);var l=e.findChildOfKind(n,s,t),u=e.findChildOfKind(n,c,t);return l&&u&&o(l,u,r,t,a,i)}function g(n){return n.length?s(e.createTextSpanFromRange(n),"code"):void 0}}(l,n);d&&r.push(d),c--,e.isCallExpression(l)?(c++,f(l.expression),c--,l.arguments.forEach(f),null===(u=l.typeArguments)||void 0===u||u.forEach(f)):e.isIfStatement(l)&&l.elseStatement&&e.isIfStatement(l.elseStatement)?(f(l.expression),f(l.thenStatement),c++,f(l.elseStatement),c--):l.forEachChild(f),c++}}}(n,t,c),function(n,t){for(var a=[],i=n.getLineStarts(),o=0,c=i;o<c.length;o++){var l=c[o],u=n.getLineEndOfPosition(l),d=r(n.text.substring(l,u));if(d&&!e.isInComment(n,l))if(d[1]){var p=a.pop();p&&(p.textSpan.length=u-p.textSpan.start,p.hintSpan.length=u-p.textSpan.start,t.push(p))}else{var m=e.createTextSpanFromBounds(n.text.indexOf("//",l),u);a.push(s(m,"region",m,!1,d[2]||"#region"))}}}(n,c),c.sort((function(e,n){return e.textSpan.start-n.textSpan.start}))};var t=/^\s*\/\/\s*#(end)?region(?:\s+(.*))?(?:\r)?$/;function r(e){return t.exec(e)}function a(n,t,a,o){var s=e.getLeadingCommentRangesOfNode(n,t);if(s){for(var c=-1,l=-1,u=0,d=t.getFullText(),p=0,m=s;p<m.length;p++){var f=m[p],_=f.kind,g=f.pos,y=f.end;switch(a.throwIfCancellationRequested(),_){case 2:if(r(d.slice(g,y))){v(),u=0;break}0===u&&(c=g),l=y,u++;break;case 3:v(),o.push(i(g,y,"comment")),u=0;break;default:e.Debug.assertNever(_)}}v()}function v(){u>1&&o.push(i(c,l,"comment"))}}function i(n,t,r){return s(e.createTextSpanFromBounds(n,t),r)}function o(n,t,r,a,i,o){return void 0===i&&(i=!1),void 0===o&&(o=!0),s(e.createTextSpanFromBounds(o?n.getFullStart():n.getStart(a),t.getEnd()),"code",e.createTextSpanFromNode(r,a),i)}function s(e,n,t,r,a){return void 0===t&&(t=e),void 0===r&&(r=!1),void 0===a&&(a="..."),{textSpan:e,kind:n,hintSpan:t,bannerText:a,autoCollapse:r}}}(e.OutliningElementsCollector||(e.OutliningElementsCollector={}))}(ts||(ts={})),function(e){var n;function t(e,n){return{kind:e,isCaseSensitive:n}}function r(e,n){var t=n.get(e);return t||n.set(e,t=v(e)),t}function a(a,i,o){var s=function(e,n){for(var t=e.length-n.length,r=function(t){if(L(n,(function(n,r){return p(e.charCodeAt(r+t))===n})))return{value:t}},a=0;a<=t;a++){var i=r(a);if("object"==typeof i)return i.value}return-1}(a,i.textLowerCase);if(0===s)return t(i.text.length===a.length?n.exact:n.prefix,e.startsWith(a,i.text));if(i.isLowerCase){if(-1===s)return;for(var d=0,m=r(a,o);d<m.length;d++){var f=m[d];if(c(a,f,i.text,!0))return t(n.substring,c(a,f,i.text,!1))}if(i.text.length<a.length&&u(a.charCodeAt(s)))return t(n.substring,!1)}else{if(a.indexOf(i.text)>0)return t(n.substring,!0);if(i.characterSpans.length>0){var _=r(a,o),g=!!l(a,_,i,!1)||!l(a,_,i,!0)&&void 0;if(void 0!==g)return t(n.camelCase,g)}}}function i(e,n,t){if(L(n.totalTextChunk.text,(function(e){return 32!==e&&42!==e}))){var r=a(e,n.totalTextChunk,t);if(r)return r}for(var i,s=0,c=n.subWordTextChunks;s<c.length;s++){i=o(i,a(e,c[s],t))}return i}function o(n,t){return e.min(n,t,s)}function s(n,t){return void 0===n?1:void 0===t?-1:e.compareValues(n.kind,t.kind)||e.compareBooleans(!n.isCaseSensitive,!t.isCaseSensitive)}function c(e,n,t,r,a){return void 0===a&&(a={start:0,length:t.length}),a.length<=n.length&&x(0,a.length,(function(i){return function(e,n,t){return t?p(e)===p(n):e===n}(t.charCodeAt(a.start+i),e.charCodeAt(n.start+i),r)}))}function l(n,t,r,a){for(var i,o,s=r.characterSpans,l=0,d=0;;){if(d===s.length)return!0;if(l===t.length)return!1;for(var p=t[l],m=!1;d<s.length;d++){var f=s[d];if(m&&(!u(r.text.charCodeAt(s[d-1].start))||!u(r.text.charCodeAt(s[d].start))))break;if(!c(n,p,r.text,a,f))break;m=!0,i=void 0===i?l:i,o=void 0===o||o,p=e.createTextSpan(p.start+f.length,p.length-f.length)}m||void 0===o||(o=!1),l++}}function u(n){if(n>=65&&n<=90)return!0;if(n<127||!e.isUnicodeIdentifierStart(n,99))return!1;var t=String.fromCharCode(n);return t===t.toUpperCase()}function d(n){if(n>=97&&n<=122)return!0;if(n<127||!e.isUnicodeIdentifierStart(n,99))return!1;var t=String.fromCharCode(n);return t===t.toLowerCase()}function p(e){return e>=65&&e<=90?e-65+97:e<127?e:String.fromCharCode(e).toLowerCase().charCodeAt(0)}function m(e){return e>=48&&e<=57}function f(e){return u(e)||d(e)||m(e)||95===e||36===e}function _(e){for(var n=[],t=0,r=0,a=0;a<e.length;a++){f(e.charCodeAt(a))?(0===r&&(t=a),r++):r>0&&(n.push(g(e.substr(t,r))),r=0)}return r>0&&n.push(g(e.substr(t,r))),n}function g(e){var n=e.toLowerCase();return{text:e,textLowerCase:n,isLowerCase:e===n,characterSpans:y(e)}}function y(e){return h(e,!1)}function v(e){return h(e,!0)}function h(n,t){for(var r=[],a=0,i=1;i<n.length;i++){var o=m(n.charCodeAt(i-1)),s=m(n.charCodeAt(i)),c=S(n,t,i),l=t&&T(n,i,a);(b(n.charCodeAt(i-1))||b(n.charCodeAt(i))||o!==s||c||l)&&(E(n,a,i)||r.push(e.createTextSpan(a,i-a)),a=i)}return E(n,a,n.length)||r.push(e.createTextSpan(a,n.length-a)),r}function b(e){switch(e){case 33:case 34:case 35:case 37:case 38:case 39:case 40:case 41:case 42:case 44:case 45:case 46:case 47:case 58:case 59:case 63:case 64:case 91:case 92:case 93:case 95:case 123:case 125:return!0}return!1}function E(e,n,t){return L(e,(function(e){return b(e)&&95!==e}),n,t)}function T(e,n,t){return n!==t&&n+1<e.length&&u(e.charCodeAt(n))&&d(e.charCodeAt(n+1))&&L(e,u,t,n)}function S(e,n,t){var r=u(e.charCodeAt(t-1));return u(e.charCodeAt(t))&&(!n||!r)}function x(e,n,t){for(var r=e;r<n;r++)if(!t(r))return!1;return!0}function L(e,n,t,r){return void 0===t&&(t=0),void 0===r&&(r=e.length),x(t,r,(function(t){return n(e.charCodeAt(t),t)}))}!function(e){e[e.exact=0]="exact",e[e.prefix=1]="prefix",e[e.substring=2]="substring",e[e.camelCase=3]="camelCase"}(n=e.PatternMatchKind||(e.PatternMatchKind={})),e.createPatternMatcher=function(n){var t=new e.Map,r=n.trim().split(".").map((function(e){return{totalTextChunk:g(n=e.trim()),subWordTextChunks:_(n)};var n}));if(!r.some((function(e){return!e.subWordTextChunks.length})))return{getFullMatch:function(n,a){return function(n,t,r,a){var s;if(!i(t,e.last(r),a))return;if(r.length-1>n.length)return;for(var c=r.length-2,l=n.length-1;c>=0;c-=1,l-=1)s=o(s,i(n[l],r[c],a));return s}(n,a,r,t)},getMatchForLastSegmentOfPattern:function(n){return i(n,e.last(r),t)},patternContainsDots:r.length>1}},e.breakIntoCharacterSpans=y,e.breakIntoWordSpans=v}(ts||(ts={})),function(e){e.preProcessFile=function(n,t,r){void 0===t&&(t=!0),void 0===r&&(r=!1);var a,i,o,s={languageVersion:1,pragmas:void 0,checkJsDirective:void 0,referencedFiles:[],typeReferenceDirectives:[],libReferenceDirectives:[],amdDependencies:[],hasNoDefaultLib:void 0,moduleName:void 0},c=[],l=0,u=!1;function d(){return i=o,18===(o=e.scanner.scan())?l++:19===o&&l--,o}function p(){var n=e.scanner.getTokenValue(),t=e.scanner.getTokenPos();return{fileName:n,pos:t,end:t+n.length}}function m(){c.push(p()),f()}function f(){0===l&&(u=!0)}function _(){var n=e.scanner.getToken();return 133===n&&(139===(n=d())&&10===(n=d())&&(a||(a=[]),a.push({ref:p(),depth:l})),!0)}function g(){if(24===i)return!1;var n=e.scanner.getToken();if(99===n){if(20===(n=d())){if(10===(n=d())||14===n)return m(),!0}else{if(10===n)return m(),!0;149===n&&e.scanner.lookAhead((function(){var n=e.scanner.scan();return 153!==n&&(41===n||18===n||78===n||e.isKeyword(n))}))&&(n=d());if(78===n||e.isKeyword(n))if(153===(n=d())){if(10===(n=d()))return m(),!0}else if(62===n){if(v(!0))return!0}else{if(27!==n)return!0;n=d()}if(18===n){for(n=d();19!==n&&1!==n;)n=d();19===n&&153===(n=d())&&10===(n=d())&&m()}else 41===n&&126===(n=d())&&(78===(n=d())||e.isKeyword(n))&&153===(n=d())&&10===(n=d())&&m()}return!0}return!1}function y(){var n=e.scanner.getToken();if(92===n){if(f(),149===(n=d()))e.scanner.lookAhead((function(){var n=e.scanner.scan();return 41===n||18===n}))&&(n=d());if(18===n){for(n=d();19!==n&&1!==n;)n=d();19===n&&153===(n=d())&&10===(n=d())&&m()}else if(41===n)153===(n=d())&&10===(n=d())&&m();else if(99===n){if(149===(n=d()))e.scanner.lookAhead((function(){var n=e.scanner.scan();return 78===n||e.isKeyword(n)}))&&(n=d());if((78===n||e.isKeyword(n))&&62===(n=d())&&v(!0))return!0}return!0}return!1}function v(n,t){void 0===t&&(t=!1);var r=n?d():e.scanner.getToken();return 143===r&&(20===(r=d())&&(10===(r=d())||t&&14===r)&&m(),!0)}function h(){var n=e.scanner.getToken();if(78===n&&"define"===e.scanner.getTokenValue()){if(20!==(n=d()))return!0;if(10===(n=d())||14===n){if(27!==(n=d()))return!0;n=d()}if(22!==n)return!0;for(n=d();23!==n&&1!==n;)10!==n&&14!==n||m(),n=d();return!0}return!1}if(t&&function(){for(e.scanner.setText(n),d();1!==e.scanner.getToken();)_()||g()||y()||r&&(v(!1,!0)||h())||d();e.scanner.setText(void 0)}(),e.processCommentPragmas(s,n),e.processPragmasIntoFields(s,e.noop),u){if(a)for(var b=0,E=a;b<E.length;b++){var T=E[b];c.push(T.ref)}return{referencedFiles:s.referencedFiles,typeReferenceDirectives:s.typeReferenceDirectives,libReferenceDirectives:s.libReferenceDirectives,importedFiles:c,isLibFile:!!s.hasNoDefaultLib,ambientExternalModules:void 0}}var S=void 0;if(a)for(var x=0,L=a;x<L.length;x++){0===(T=L[x]).depth?(S||(S=[]),S.push(T.ref.fileName)):c.push(T.ref)}return{referencedFiles:s.referencedFiles,typeReferenceDirectives:s.typeReferenceDirectives,libReferenceDirectives:s.libReferenceDirectives,importedFiles:c,isLibFile:!!s.hasNoDefaultLib,ambientExternalModules:S}}}(ts||(ts={})),function(e){!function(n){function t(e,n,t,r,i,o){return{canRename:!0,fileToRename:void 0,kind:t,displayName:e,fullDisplayName:n,kindModifiers:r,triggerSpan:a(i,o)}}function r(n){return{canRename:!1,localizedErrorMessage:e.getLocaleSpecificMessage(n)}}function a(n,t){var r=n.getStart(t),a=n.getWidth(t);return e.isStringLiteralLike(n)&&(r+=1,a-=2),e.createTextSpan(r,a)}n.getRenameInfo=function(n,a,i,o){var s=e.getAdjustedRenameLocation(e.getTouchingPropertyName(a,i));if(function(n){switch(n.kind){case 78:case 79:case 10:case 14:case 107:return!0;case 8:return e.isLiteralNameOfPropertyDeclarationOrIndexAccess(n);default:return!1}}(s)){var c=function(n,a,i,o,s){var c=a.getSymbolAtLocation(n);if(!c){if(e.isLabelName(n)){var l=e.getTextOfNode(n);return t(l,l,"label","",n,i)}return}var u=c.declarations;if(!u||0===u.length)return;if(u.some(o))return r(e.Diagnostics.You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library);if(e.isIdentifier(n)&&87===n.originalKeywordKind&&c.parent&&1536&c.parent.flags)return;if(e.isStringLiteralLike(n)&&e.tryGetImportFromModuleSpecifier(n))return s&&s.allowRenameOfImportPath?function(n,t,a){if(!e.isExternalModuleNameRelative(n.text))return r(e.Diagnostics.You_cannot_rename_a_module_via_a_global_import);var i=e.find(a.declarations,e.isSourceFile);if(!i)return;var o=e.endsWith(n.text,"/index")||e.endsWith(n.text,"/index.js")?void 0:e.tryRemoveSuffix(e.removeFileExtension(i.fileName),"/index"),s=void 0===o?i.fileName:o,c=void 0===o?"module":"directory",l=n.text.lastIndexOf("/")+1,u=e.createTextSpan(n.getStart(t)+1+l,n.text.length-l);return{canRename:!0,fileToRename:s,kind:c,displayName:s,fullDisplayName:s,kindModifiers:"",triggerSpan:u}}(n,i,c):void 0;var d=e.SymbolDisplay.getSymbolKind(a,c,n),p=e.isImportOrExportSpecifierName(n)||e.isStringOrNumericLiteralLike(n)&&158===n.parent.kind?e.stripQuotes(e.getTextOfIdentifierOrLiteral(n)):void 0,m=p||a.symbolToString(c),f=p||a.getFullyQualifiedName(c);return t(m,f,d,e.SymbolDisplay.getSymbolModifiers(c),n,i)}(s,n.getTypeChecker(),a,(function(e){return n.isSourceFileDefaultLibrary(e.getSourceFile())}),o);if(c)return c}return r(e.Diagnostics.You_cannot_rename_this_element)}}(e.Rename||(e.Rename={}))}(ts||(ts={})),function(e){!function(n){function t(n,t,r){return e.Debug.assert(r.pos<=t),t<r.end||r.getEnd()===t&&e.getTouchingPropertyName(n,t).pos<r.end}n.getSmartSelectionRange=function(n,r){var i,o,s,u={textSpan:e.createTextSpanFromBounds(r.getFullStart(),r.getEnd())},d=r;e:for(;;){var p=a(d);if(!p.length)break;for(var m=0;m<p.length;m++){var f=p[m-1],_=p[m],g=p[m+1];if(e.getTokenPosOfNode(_,r,!0)>n)break e;if(t(r,n,_)){if(e.isBlock(_)||e.isTemplateSpan(_)||e.isTemplateHead(_)||e.isTemplateTail(_)||f&&e.isTemplateHead(f)||e.isVariableDeclarationList(_)&&e.isVariableStatement(d)||e.isSyntaxList(_)&&e.isVariableDeclarationList(d)||e.isVariableDeclaration(_)&&e.isSyntaxList(d)&&1===p.length||e.isJSDocTypeExpression(_)||e.isJSDocSignature(_)||e.isJSDocTypeLiteral(_)){d=_;break}if(e.isTemplateSpan(d)&&g&&e.isTemplateMiddleOrTemplateTail(g))b(_.getFullStart()-"${".length,g.getStart()+"}".length);var y=e.isSyntaxList(_)&&(s=void 0,18===(s=(o=f)&&o.kind)||22===s||20===s||275===s)&&c(g)&&!e.positionsAreOnSameLine(f.getStart(),g.getStart(),r),v=y?f.getEnd():_.getStart(),h=y?g.getStart():l(r,_);e.hasJSDocNodes(_)&&(null===(i=_.jsDoc)||void 0===i?void 0:i.length)&&b(e.first(_.jsDoc).getStart(),h),b(v,h),(e.isStringLiteral(_)||e.isTemplateLiteral(_))&&b(v+1,h-1),d=_;break}if(m===p.length-1)break e}}return u;function b(t,r){if(t!==r){var a=e.createTextSpanFromBounds(t,r);(!u||!e.textSpansEqual(a,u.textSpan)&&e.textSpanIntersectsWithPosition(a,n))&&(u=__assign({textSpan:a},u&&{parent:u}))}}};var r=e.or(e.isImportDeclaration,e.isImportEqualsDeclaration);function a(n){if(e.isSourceFile(n))return i(n.getChildAt(0).getChildren(),r);if(e.isMappedTypeNode(n)){var t=n.getChildren(),a=t[0],c=t.slice(1),l=e.Debug.checkDefined(c.pop());e.Debug.assertEqual(a.kind,18),e.Debug.assertEqual(l.kind,19);var u=i(c,(function(e){return e===n.readonlyToken||142===e.kind||e===n.questionToken||57===e.kind}));return[a,s(o(i(u,(function(e){var n=e.kind;return 22===n||159===n||23===n})),(function(e){return 58===e.kind}))),l]}if(e.isPropertySignature(n))return o(c=i(n.getChildren(),(function(t){return t===n.name||e.contains(n.modifiers,t)})),(function(e){return 58===e.kind}));if(e.isParameter(n)){var d=i(n.getChildren(),(function(e){return e===n.dotDotDotToken||e===n.name}));return o(i(d,(function(e){return e===d[0]||e===n.questionToken})),(function(e){return 62===e.kind}))}return e.isBindingElement(n)?o(n.getChildren(),(function(e){return 62===e.kind})):n.getChildren()}function i(e,n){for(var t,r=[],a=0,i=e;a<i.length;a++){var o=i[a];n(o)?(t=t||[]).push(o):(t&&(r.push(s(t)),t=void 0),r.push(o))}return t&&r.push(s(t)),r}function o(n,t,r){if(void 0===r&&(r=!0),n.length<2)return n;var a=e.findIndex(n,t);if(-1===a)return n;var i=n.slice(0,a),o=n[a],c=e.last(n),l=r&&26===c.kind,u=n.slice(a+1,l?n.length-1:void 0),d=e.compact([i.length?s(i):void 0,o,u.length?s(u):void 0]);return l?d.concat(c):d}function s(n){return e.Debug.assertGreaterThanOrEqual(n.length,1),e.setTextRangePosEnd(e.parseNodeFactory.createSyntaxList(n),n[0].pos,e.last(n).end)}function c(e){var n=e&&e.kind;return 19===n||23===n||21===n||276===n}function l(e,n){switch(n.kind){case 326:case 324:case 333:case 331:case 328:return e.getLineEndOfPosition(n.getStart());default:return n.getEnd()}}}(e.SmartSelectionRange||(e.SmartSelectionRange={}))}(ts||(ts={})),function(e){!function(n){var t,r;function a(n,t,r){for(var a=n.getFullStart(),i=n.parent;i;){var o=e.findPrecedingToken(a,t,i,!0);if(o)return e.rangeContainsRange(r,o);i=i.parent}return e.Debug.fail("Could not find preceding token")}function i(n,t){var r=function(n,t){if(29===n.kind||20===n.kind)return{list:m(n.parent,n,t),argumentIndex:0};var r=e.findContainingList(n);return r&&{list:r,argumentIndex:u(r,n)}}(n,t);if(r){var a=r.list,i=r.argumentIndex,o=function(n){var t=n.getChildren(),r=e.countWhere(t,(function(e){return 27!==e.kind}));t.length>0&&27===e.last(t).kind&&r++;return r}(a);return 0!==i&&e.Debug.assertLessThan(i,o),{list:a,argumentIndex:i,argumentCount:o,argumentsSpan:function(n,t){var r=n.getFullStart(),a=e.skipTrivia(t.text,n.getEnd(),!1);return e.createTextSpan(r,a-r)}(a,t)}}}function o(n,t,r){var a=n.parent;if(e.isCallOrNewExpression(a)){var o=a,s=i(n,r);if(!s)return;var c=s.list,l=s.argumentIndex,u=s.argumentCount,p=s.argumentsSpan;return{isTypeParameterList:!!a.typeArguments&&a.typeArguments.pos===c.pos,invocation:{kind:0,node:o},argumentsSpan:p,argumentIndex:l,argumentCount:u}}if(e.isNoSubstitutionTemplateLiteral(n)&&e.isTaggedTemplateExpression(a))return e.isInsideTemplateLiteral(n,t,r)?d(a,0,r):void 0;if(e.isTemplateHead(n)&&205===a.parent.kind){var m=a,f=m.parent;return e.Debug.assert(218===m.kind),d(f,l=e.isInsideTemplateLiteral(n,t,r)?0:1,r)}if(e.isTemplateSpan(a)&&e.isTaggedTemplateExpression(a.parent.parent)){var _=a;f=a.parent.parent;if(e.isTemplateTail(n)&&!e.isInsideTemplateLiteral(n,t,r))return;return d(f,l=function(n,t,r,a){if(e.Debug.assert(r>=t.getStart(),"Assumed 'position' could not occur before node."),e.isTemplateLiteralToken(t))return e.isInsideTemplateLiteral(t,r,a)?0:n+2;return n+1}(_.parent.templateSpans.indexOf(_),n,t,r),r)}if(e.isJsxOpeningLikeElement(a)){var g=a.attributes.pos,y=e.skipTrivia(r.text,a.attributes.end,!1);return{isTypeParameterList:!1,invocation:{kind:0,node:a},argumentsSpan:e.createTextSpan(g,y-g),argumentIndex:0,argumentCount:1}}var v=e.getPossibleTypeArgumentsInfo(n,r);if(v){var h=v.called,b=v.nTypeArguments;return{isTypeParameterList:!0,invocation:o={kind:1,called:h},argumentsSpan:p=e.createTextSpanFromBounds(h.getStart(r),n.end),argumentIndex:b,argumentCount:b+1}}}function s(n){return e.isBinaryExpression(n.parent)?s(n.parent):n}function c(n){return e.isBinaryExpression(n.left)?c(n.left)+1:2}function l(n){return"__type"===n.name&&e.firstDefined(n.declarations,(function(n){return e.isFunctionTypeNode(n)?n.parent.symbol:void 0}))||n}function u(e,n){for(var t=0,r=0,a=e.getChildren();r<a.length;r++){var i=a[r];if(i===n)break;27!==i.kind&&t++}return t}function d(n,t,r){var a=e.isNoSubstitutionTemplateLiteral(n.template)?1:n.template.templateSpans.length+1;return 0!==t&&e.Debug.assertLessThan(t,a),{isTypeParameterList:!1,invocation:{kind:0,node:n},argumentsSpan:p(n,r),argumentIndex:t,argumentCount:a}}function p(n,t){var r=n.template,a=r.getStart(),i=r.getEnd();218===r.kind&&(0===e.last(r.templateSpans).literal.getFullWidth()&&(i=e.skipTrivia(t.text,i,!1)));return e.createTextSpan(a,i-a)}function m(n,t,r){var a=n.getChildren(r),i=a.indexOf(t);return e.Debug.assert(i>=0&&a.length>i+1),a[i+1]}function f(n){return 0===n.kind?e.getInvokedExpression(n.node):n.called}function _(e){return 0===e.kind?e.node:1===e.kind?e.called:e.node}!function(e){e[e.Call=0]="Call",e[e.TypeArgs=1]="TypeArgs",e[e.Contextual=2]="Contextual"}(t||(t={})),n.getSignatureHelpItems=function(n,t,r,u,d){var p=n.getTypeChecker(),m=e.findTokenOnLeftOfPosition(t,r);if(m){var g=!!u&&"characterTyped"===u.kind;if(!g||!e.isInString(t,r,m)&&!e.isInComment(t,r)){var h=!!u&&"invoked"===u.kind,b=function(n,t,r,a,u){for(var d=function(n){e.Debug.assert(e.rangeContainsRange(n.parent,n),"Not a subspan",(function(){return"Child: "+e.Debug.formatSyntaxKind(n.kind)+", parent: "+e.Debug.formatSyntaxKind(n.parent.kind)}));var u=function(n,t,r,a){return function(n,t,r,a){var o=function(n,t,r){if(20!==n.kind&&27!==n.kind)return;var a=n.parent;switch(a.kind){case 207:case 165:case 208:case 209:var o=i(n,t);if(!o)return;var l=o.argumentIndex,u=o.argumentCount,d=o.argumentsSpan,p=e.isMethodDeclaration(a)?r.getContextualTypeForObjectLiteralElement(a):r.getContextualType(a);return p&&{contextualType:p,argumentIndex:l,argumentCount:u,argumentsSpan:d};case 216:var m=s(a),f=r.getContextualType(m),_=20===n.kind?0:c(a)-1,g=c(m);return f&&{contextualType:f,argumentIndex:_,argumentCount:g,argumentsSpan:e.createTextSpanFromNode(a)};default:return}}(n,r,a);if(!o)return;var u=o.contextualType,d=o.argumentIndex,p=o.argumentCount,m=o.argumentsSpan,f=u.getNonNullableType(),_=f.getCallSignatures();return 1!==_.length?void 0:{isTypeParameterList:!1,invocation:{kind:2,signature:e.first(_),node:n,symbol:l(f.symbol)},argumentsSpan:m,argumentIndex:d,argumentCount:p}}(n,0,r,a)||o(n,t,r)}(n,t,r,a);if(u)return{value:u}},p=n;!e.isSourceFile(p)&&(u||!e.isBlock(p));p=p.parent){var m=d(p);if("object"==typeof m)return m.value}return}(m,r,t,p,h);if(b){d.throwIfCancellationRequested();var E=function(n,t,r,i,o){var s=n.invocation,c=n.argumentCount;switch(s.kind){case 0:if(o&&!function(n,t,r){if(!e.isCallOrNewExpression(t))return!1;var i=t.getChildren(r);switch(n.kind){case 20:return e.contains(i,n);case 27:var o=e.findContainingList(n);return!!o&&e.contains(i,o);case 29:return a(n,r,t.expression);default:return!1}}(i,s.node,r))return;var l=[],u=t.getResolvedSignatureForSignatureHelp(s.node,l,c);return 0===l.length?void 0:{kind:0,candidates:l,resolvedSignature:u};case 1:var d=s.called;if(o&&!a(i,r,e.isIdentifier(d)?d.parent:d))return;if(0!==(l=e.getPossibleGenericSignatures(d,c,t)).length)return{kind:0,candidates:l,resolvedSignature:e.first(l)};var p=t.getSymbolAtLocation(d);return p&&{kind:1,symbol:p};case 2:return{kind:0,candidates:[s.signature],resolvedSignature:s.signature};default:return e.Debug.assertNever(s)}}(b,p,t,m,g);return d.throwIfCancellationRequested(),E?p.runWithCancellationToken(d,(function(e){return 0===E.kind?y(E.candidates,E.resolvedSignature,b,t,e):function(e,n,t,r){var a=n.argumentCount,i=n.argumentsSpan,o=n.invocation,s=n.argumentIndex,c=r.getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(e);return c?{items:[v(e,c,r,_(o),t)],applicableSpan:i,selectedItemIndex:0,argumentIndex:s,argumentCount:a}:void 0}(E.symbol,b,t,e)})):e.isSourceFileJS(t)?function(n,t,r){if(2===n.invocation.kind)return;var a=f(n.invocation),i=e.isPropertyAccessExpression(a)?a.name.text:void 0,o=t.getTypeChecker();return void 0===i?void 0:e.firstDefined(t.getSourceFiles(),(function(t){return e.firstDefined(t.getNamedDeclarations().get(i),(function(e){var a=e.symbol&&o.getTypeOfSymbolAtLocation(e.symbol,e),i=a&&a.getCallSignatures();if(i&&i.length)return o.runWithCancellationToken(r,(function(e){return y(i,i[0],n,t,e,!0)}))}))}))}(b,n,d):void 0}}}},function(e){e[e.Candidate=0]="Candidate",e[e.Type=1]="Type"}(r||(r={})),n.getArgumentInfoForCompletions=function(e,n,t){var r=o(e,n,t);return!r||r.isTypeParameterList||0!==r.invocation.kind?void 0:{invocation:r.invocation.node,argumentCount:r.argumentCount,argumentIndex:r.argumentIndex}};var g=70246400;function y(n,t,r,a,i,o){var s,c=r.isTypeParameterList,l=r.argumentCount,u=r.argumentsSpan,d=r.invocation,p=r.argumentIndex,m=_(d),g=2===d.kind?d.symbol:i.getSymbolAtLocation(f(d))||o&&(null===(s=t.declaration)||void 0===s?void 0:s.symbol),y=g?e.symbolToDisplayParts(i,g,o?a:void 0,void 0):e.emptyArray,v=e.map(n,(function(n){return function(n,t,r,a,i,o){var s=(r?b:E)(n,a,i,o);return e.map(s,(function(r){var o=r.isVariadic,s=r.parameters,c=r.prefix,l=r.suffix,u=__spreadArrays(t,c),d=__spreadArrays(l,function(n,t,r){return e.mapToDisplayParts((function(e){e.writePunctuation(":"),e.writeSpace(" ");var a=r.getTypePredicateOfSignature(n);a?r.writeTypePredicate(a,t,void 0,e):r.writeType(r.getReturnTypeOfSignature(n),t,void 0,e)}))}(n,i,a)),p=n.getDocumentationComment(a),m=n.getJsDocTags();return{isVariadic:o,prefixDisplayParts:u,suffixDisplayParts:d,separatorDisplayParts:h,parameters:s,documentation:p,tags:m}}))}(n,y,c,i,m,a)}));0!==p&&e.Debug.assertLessThan(p,l);for(var T=0,S=0,x=0;x<v.length;x++){var L=v[x];if(n[x]===t&&(T=S,L.length>1))for(var A=0,D=0,C=L;D<C.length;D++){var I=C[D];if(I.isVariadic||I.parameters.length>=l){T=S+A;break}A++}S+=L.length}return e.Debug.assert(-1!==T),{items:e.flatMapToMutable(v,e.identity),applicableSpan:u,selectedItemIndex:T,argumentIndex:p,argumentCount:l}}function v(n,t,r,a,i){var o=e.symbolToDisplayParts(r,n),s=e.createPrinter({removeComments:!0}),c=t.map((function(e){return T(e,r,a,i,s)})),l=n.getDocumentationComment(r),u=n.getJsDocTags();return{isVariadic:!1,prefixDisplayParts:__spreadArrays(o,[e.punctuationPart(29)]),suffixDisplayParts:[e.punctuationPart(31)],separatorDisplayParts:h,parameters:c,documentation:l,tags:u}}var h=[e.punctuationPart(27),e.spacePart()];function b(n,t,r,a){var i=(n.target||n).typeParameters,o=e.createPrinter({removeComments:!0}),s=(i||e.emptyArray).map((function(e){return T(e,t,r,a,o)})),c=n.thisParameter?[t.symbolToParameterDeclaration(n.thisParameter,r,g)]:[];return t.getExpandedParameters(n).map((function(n){var i=e.factory.createNodeArray(__spreadArrays(c,e.map(n,(function(e){return t.symbolToParameterDeclaration(e,r,g)})))),l=e.mapToDisplayParts((function(e){o.writeList(2576,i,a,e)}));return{isVariadic:!1,parameters:s,prefix:[e.punctuationPart(29)],suffix:__spreadArrays([e.punctuationPart(31)],l)}}))}function E(n,t,r,a){var i=t.hasEffectiveRestParameter(n),o=e.createPrinter({removeComments:!0}),s=e.mapToDisplayParts((function(i){if(n.typeParameters&&n.typeParameters.length){var s=e.factory.createNodeArray(n.typeParameters.map((function(e){return t.typeParameterToDeclaration(e,r,g)})));o.writeList(53776,s,a,i)}})),c=t.getExpandedParameters(n);return c.map((function(n){return{isVariadic:i&&(1===c.length||!!(32768&n[n.length-1].checkFlags)),parameters:n.map((function(n){return function(n,t,r,a,i){var o=e.mapToDisplayParts((function(e){var o=t.symbolToParameterDeclaration(n,r,g);i.writeNode(4,o,a,e)})),s=t.isOptionalParameter(n.valueDeclaration);return{name:n.name,documentation:n.getDocumentationComment(t),displayParts:o,isOptional:s}}(n,t,r,a,o)})),prefix:__spreadArrays(s,[e.punctuationPart(20)]),suffix:[e.punctuationPart(21)]}}))}function T(n,t,r,a,i){var o=e.mapToDisplayParts((function(e){var o=t.typeParameterToDeclaration(n,r,g);i.writeNode(4,o,a,e)}));return{name:n.symbol.name,documentation:n.symbol.getDocumentationComment(t),displayParts:o,isOptional:!1}}}(e.SignatureHelp||(e.SignatureHelp={}))}(ts||(ts={})),function(e){var n=/^data:(?:application\/json(?:;charset=[uU][tT][fF]-8);base64,([A-Za-z0-9+\/=]+)$)?/;function t(n,t,r){var a=e.tryParseRawSourceMap(t);if(a&&a.sources&&a.file&&a.mappings&&(!a.sourcesContent||!a.sourcesContent.some(e.isString)))return e.createDocumentPositionMapper(n,a,r)}e.getSourceMapper=function(n){var t=e.createGetCanonicalFileName(n.useCaseSensitiveFileNames()),r=n.getCurrentDirectory(),a=new e.Map,i=new e.Map;return{tryGetSourcePosition:function n(t){if(!e.isDeclarationFileName(t.fileName))return;if(!c(t.fileName))return;var r=s(t.fileName).getSourcePosition(t);return r&&r!==t?n(r)||r:void 0},tryGetGeneratedPosition:function(a){if(e.isDeclarationFileName(a.fileName))return;var i=c(a.fileName);if(!i)return;var o=n.getProgram();if(o.isSourceOfProjectReferenceRedirect(i.fileName))return;var l=o.getCompilerOptions(),u=e.outFile(l),d=u?e.removeFileExtension(u)+".d.ts":e.getDeclarationEmitOutputFilePathWorker(a.fileName,o.getCompilerOptions(),r,o.getCommonSourceDirectory(),t);if(void 0===d)return;var p=s(d,a.fileName).getGeneratedPosition(a);return p===a?void 0:p},toLineColumnOffset:function(e,n){return u(e).getLineAndCharacterOfPosition(n)},clearCache:function(){a.clear(),i.clear()}};function o(n){return e.toPath(n,r,t)}function s(r,a){var s,c=o(r),l=i.get(c);if(l)return l;if(n.getDocumentPositionMapper)s=n.getDocumentPositionMapper(r,a);else if(n.readFile){var d=u(r);s=d&&e.getDocumentPositionMapper({getSourceFileLike:u,getCanonicalFileName:t,log:function(e){return n.log(e)}},r,e.getLineInfo(d.text,e.getLineStarts(d)),(function(e){return!n.fileExists||n.fileExists(e)?n.readFile(e):void 0}))}return i.set(c,s||e.identitySourceMapConsumer),s||e.identitySourceMapConsumer}function c(e){var t=n.getProgram();if(t){var r=o(e),a=t.getSourceFileByPath(r);return a&&a.resolvedPath===r?a:void 0}}function l(t){var r=o(t),i=a.get(r);if(void 0!==i)return i||void 0;if(n.readFile&&(!n.fileExists||n.fileExists(r))){var s=n.readFile(r),c=!!s&&function(n,t){return{text:n,lineMap:t,getLineAndCharacterOfPosition:function(n){return e.computeLineAndCharacterOfPosition(e.getLineStarts(this),n)}}}(s);return a.set(r,c),c||void 0}a.set(r,!1)}function u(e){return n.getSourceFileLike?n.getSourceFileLike(e):c(e)||l(e)}},e.getDocumentPositionMapper=function(r,a,i,o){var s=e.tryGetSourceMappingURL(i);if(s){var c=n.exec(s);if(c){if(c[1]){var l=c[1];return t(r,e.base64decode(e.sys,l),a)}s=void 0}}var u=[];s&&u.push(s),u.push(a+".map");for(var d=s&&e.getNormalizedAbsolutePath(s,e.getDirectoryPath(a)),p=0,m=u;p<m.length;p++){var f=m[p],_=e.getNormalizedAbsolutePath(f,e.getDirectoryPath(a)),g=o(_,d);if(e.isString(g))return t(r,g,_);if(void 0!==g)return g||void 0}}}(ts||(ts={})),function(e){var n=new e.Map;function t(n){return e.isPropertyAccessExpression(n)?t(n.expression):n}function r(n){switch(n.kind){case 261:var t=n.importClause,r=n.moduleSpecifier;return t&&!t.name&&t.namedBindings&&263===t.namedBindings.kind&&e.isStringLiteral(r)?t.namedBindings.name:void 0;case 260:return n.name;default:return}}function a(n){return e.isReturnStatement(n)&&!!n.expression&&i(n.expression)}function i(n){if(!o(n)||!n.arguments.every(s))return!1;for(var t=n.expression;o(t)||e.isPropertyAccessExpression(t);){if(e.isCallExpression(t)&&!t.arguments.every(s))return!1;t=t.expression}return!0}function o(n){return e.isCallExpression(n)&&(e.hasPropertyAccessExpressionWithName(n,"then")&&function(n){return!(n.arguments.length>2)&&(n.arguments.length<2||e.some(n.arguments,(function(n){return 103===n.kind||e.isIdentifier(n)&&"undefined"===n.text})))}(n)||e.hasPropertyAccessExpressionWithName(n,"catch"))}function s(e){switch(e.kind){case 251:case 208:case 209:n.set(c(e),!0);case 103:case 78:return!0;default:return!1}}function c(e){return e.pos.toString()+":"+e.end.toString()}e.computeSuggestionDiagnostics=function(i,o,s){o.getSemanticDiagnostics(i,s);var l,u=[],d=o.getTypeChecker();i.commonJsModuleIndicator&&(e.programContainsEs6Modules(o)||e.compilerOptionsIndicateEs6Modules(o.getCompilerOptions()))&&function(n){return n.statements.some((function(n){switch(n.kind){case 232:return n.declarationList.declarations.some((function(n){return!!n.initializer&&e.isRequireCall(t(n.initializer),!0)}));case 233:var r=n.expression;if(!e.isBinaryExpression(r))return e.isRequireCall(r,!0);var a=e.getAssignmentDeclarationKind(r);return 1===a||2===a;default:return!1}}))}(i)&&u.push(e.createDiagnosticForNode((l=i.commonJsModuleIndicator,e.isBinaryExpression(l)?l.left:l),e.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module));var p=e.isSourceFileJS(i);if(n.clear(),function t(r){if(p)(function(n,t){var r,a,i,o;if(208===n.kind){if(e.isVariableDeclaration(n.parent)&&(null===(r=n.symbol.members)||void 0===r?void 0:r.size))return!0;var s=t.getSymbolOfExpando(n,!1);return!(!s||!(null===(a=s.exports)||void 0===a?void 0:a.size)&&!(null===(i=s.members)||void 0===i?void 0:i.size))}if(251===n.kind)return!!(null===(o=n.symbol.members)||void 0===o?void 0:o.size);return!1})(r,d)&&u.push(e.createDiagnosticForNode(e.isVariableDeclaration(r.parent)?r.parent.name:r,e.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration));else{if(e.isVariableStatement(r)&&r.parent===i&&2&r.declarationList.flags&&1===r.declarationList.declarations.length){var o=r.declarationList.declarations[0].initializer;o&&e.isRequireCall(o,!0)&&u.push(e.createDiagnosticForNode(o,e.Diagnostics.require_call_may_be_converted_to_an_import))}e.codefix.parameterShouldGetTypeFromJSDoc(r)&&u.push(e.createDiagnosticForNode(r.name||r,e.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types))}e.isFunctionLikeDeclaration(r)&&function(t,r,i){(function(n,t){return!e.isAsyncFunction(n)&&n.body&&e.isBlock(n.body)&&(r=n.body,!!e.forEachReturnStatement(r,a))&&function(e,n){var t=n.getTypeAtLocation(e),r=n.getSignaturesOfType(t,0),a=r.length?n.getReturnTypeOfSignature(r[0]):void 0;return!!a&&!!n.getPromisedTypeOfPromise(a)}(n,t);var r})(t,r)&&!n.has(c(t))&&i.push(e.createDiagnosticForNode(!t.name&&e.isVariableDeclaration(t.parent)&&e.isIdentifier(t.parent.name)?t.parent.name:t,e.Diagnostics.This_may_be_converted_to_an_async_function))}(r,d,u);r.forEachChild(t)}(i),e.getAllowSyntheticDefaultImports(o.getCompilerOptions()))for(var m=0,f=i.imports;m<f.length;m++){var _=f[m],g=r(e.importFromModuleSpecifier(_));if(g){var y=e.getResolvedModule(i,_.text),v=y&&o.getSourceFile(y.resolvedFileName);v&&v.externalModuleIndicator&&e.isExportAssignment(v.externalModuleIndicator)&&v.externalModuleIndicator.isExportEquals&&u.push(e.createDiagnosticForNode(g,e.Diagnostics.Import_may_be_converted_to_a_default_import))}}return e.addRange(u,i.bindSuggestionDiagnostics),e.addRange(u,o.getSuggestionDiagnostics(i,s)),u.sort((function(e,n){return e.start-n.start}))},e.isReturnStatementWithFixablePromiseHandler=a,e.isFixablePromiseHandler=i}(ts||(ts={})),function(e){!function(n){var t=70246400;function r(n,t,r){var i=a(n,t,r);if(""!==i)return i;var o=e.getCombinedLocalAndExportSymbolFlags(t);return 32&o?e.getDeclarationOfKind(t,221)?"local class":"class":384&o?"enum":524288&o?"type":64&o?"interface":262144&o?"type parameter":8&o?"enum member":2097152&o?"alias":1536&o?"module":i}function a(n,t,r){var a=n.getRootSymbols(t);if(1===a.length&&8192&e.first(a).flags&&0!==n.getTypeOfSymbolAtLocation(t,r).getNonNullableType().getCallSignatures().length)return"method";if(n.isUndefinedSymbol(t))return"var";if(n.isArgumentsSymbol(t))return"local var";if(107===r.kind&&e.isExpression(r))return"parameter";var o=e.getCombinedLocalAndExportSymbolFlags(t);if(3&o)return e.isFirstDeclarationOfSymbolParameter(t)?"parameter":t.valueDeclaration&&e.isVarConst(t.valueDeclaration)?"const":e.forEach(t.declarations,e.isLet)?"let":i(t)?"local var":"var";if(16&o)return i(t)?"local function":"function";if(32768&o)return"getter";if(65536&o)return"setter";if(8192&o)return"method";if(16384&o)return"constructor";if(4&o){if(33554432&o&&6&t.checkFlags){var s=e.forEach(n.getRootSymbols(t),(function(n){var t=n.getFlags();if(98311&t)return"property";e.Debug.assert(!!(8208&t))}));return s||(n.getTypeOfSymbolAtLocation(t,r).getCallSignatures().length?"method":"property")}switch(r.parent&&r.parent.kind){case 275:case 273:case 274:return 78===r.kind?"property":"JSX attribute";case 280:return"JSX attribute";default:return"property"}}return""}function i(n){return!n.parent&&e.forEach(n.declarations,(function(n){if(208===n.kind)return!0;if(249!==n.kind&&251!==n.kind)return!1;for(var t=n.parent;!e.isFunctionBlock(t);t=t.parent)if(297===t.kind||257===t.kind)return!1;return!0}))}n.getSymbolKind=r,n.getSymbolModifiers=function(n){var t=n&&n.declarations&&n.declarations.length>0?e.getNodeModifiers(n.declarations[0]):"",r=n&&16777216&n.flags?"optional":"";return t&&r?t+","+r:t||r},n.getSymbolDisplayPartsDocumentationAndSymbolKind=function n(i,o,s,c,l,u,d){void 0===u&&(u=e.getMeaningFromLocation(l));var p,m,f,_,g=[],y=[],v=[],h=e.getCombinedLocalAndExportSymbolFlags(o),b=1&u?a(i,o,l):"",E=!1,T=107===l.kind&&e.isInExpressionContext(l),S=!1;if(107===l.kind&&!T)return{displayParts:[e.keywordPart(107)],documentation:[],symbolKind:"primitive type",tags:void 0};if(""!==b||32&h||2097152&h){"getter"!==b&&"setter"!==b||(b="property");var x=void 0;if(p=T?i.getTypeAtLocation(l):i.getTypeOfSymbolAtLocation(o.exportSymbol||o,l),l.parent&&201===l.parent.kind){var L=l.parent.name;(L===l||L&&0===L.getFullWidth())&&(l=l.parent)}var A=void 0;if(e.isCallOrNewExpression(l)?A=l:(e.isCallExpressionTarget(l)||e.isNewExpressionTarget(l)||l.parent&&(e.isJsxOpeningLikeElement(l.parent)||e.isTaggedTemplateExpression(l.parent))&&e.isFunctionLike(o.valueDeclaration))&&(A=l.parent),A){x=i.getResolvedSignature(A);var D=204===A.kind||e.isCallExpression(A)&&105===A.expression.kind,C=D?p.getConstructSignatures():p.getCallSignatures();if(e.contains(C,x.target)||e.contains(C,x)||(x=C.length?C[0]:void 0),x){switch(D&&32&h?(b="constructor",Y(p.symbol,b)):2097152&h?(Q(b="alias"),g.push(e.spacePart()),D&&(g.push(e.keywordPart(102)),g.push(e.spacePart())),X(o)):Y(o,b),b){case"JSX attribute":case"property":case"var":case"const":case"let":case"parameter":case"local var":g.push(e.punctuationPart(58)),g.push(e.spacePart()),16&e.getObjectFlags(p)||!p.symbol||(e.addRange(g,e.symbolToDisplayParts(i,p.symbol,c,void 0,5)),g.push(e.lineBreakPart())),D&&(g.push(e.keywordPart(102)),g.push(e.spacePart())),Z(x,C,262144);break;default:Z(x,C)}E=!0,S=C.length>1}}else if(e.isNameOfFunctionDeclaration(l)&&!(98304&h)||132===l.kind&&166===l.parent.kind){var I=l.parent;if(o.declarations&&e.find(o.declarations,(function(e){return e===(132===l.kind?I.parent:I)}))){C=166===I.kind?p.getNonNullableType().getConstructSignatures():p.getNonNullableType().getCallSignatures();x=i.isImplementationOfOverload(I)?C[0]:i.getSignatureFromDeclaration(I),166===I.kind?(b="constructor",Y(p.symbol,b)):Y(169!==I.kind||2048&p.symbol.flags||4096&p.symbol.flags?o:p.symbol,b),Z(x,C),E=!0,S=C.length>1}}}if(32&h&&!E&&!T&&(q(),e.getDeclarationOfKind(o,221)?Q("local class"):g.push(e.keywordPart(83)),g.push(e.spacePart()),X(o),$(o,s)),64&h&&2&u&&(z(),g.push(e.keywordPart(117)),g.push(e.spacePart()),X(o),$(o,s)),524288&h&&2&u&&(z(),g.push(e.keywordPart(149)),g.push(e.spacePart()),X(o),$(o,s),g.push(e.spacePart()),g.push(e.operatorPart(62)),g.push(e.spacePart()),e.addRange(g,e.typeToDisplayParts(i,i.getDeclaredTypeOfSymbol(o),c,8388608))),384&h&&(z(),e.some(o.declarations,(function(n){return e.isEnumDeclaration(n)&&e.isEnumConst(n)}))&&(g.push(e.keywordPart(84)),g.push(e.spacePart())),g.push(e.keywordPart(91)),g.push(e.spacePart()),X(o)),1536&h&&!T){z();var N=(H=e.getDeclarationOfKind(o,256))&&H.name&&78===H.name.kind;g.push(e.keywordPart(N?140:139)),g.push(e.spacePart()),X(o)}if(262144&h&&2&u)if(z(),g.push(e.punctuationPart(20)),g.push(e.textPart("type parameter")),g.push(e.punctuationPart(21)),g.push(e.spacePart()),X(o),o.parent)J(),X(o.parent,c),$(o.parent,c);else{var R=e.getDeclarationOfKind(o,159);if(void 0===R)return e.Debug.fail();if(H=R.parent)if(e.isFunctionLikeKind(H.kind)){J();x=i.getSignatureFromDeclaration(H);170===H.kind?(g.push(e.keywordPart(102)),g.push(e.spacePart())):169!==H.kind&&H.name&&X(H.symbol),e.addRange(g,e.signatureToDisplayParts(i,x,s,32))}else 254===H.kind&&(J(),g.push(e.keywordPart(149)),g.push(e.spacePart()),X(H.symbol),$(H.symbol,s))}if(8&h&&(b="enum member",Y(o,"enum member"),291===(H=o.declarations[0]).kind)){var k=i.getConstantValue(H);void 0!==k&&(g.push(e.spacePart()),g.push(e.operatorPart(62)),g.push(e.spacePart()),g.push(e.displayPart(e.getTextOfConstantValue(k),"number"==typeof k?e.SymbolDisplayPartKind.numericLiteral:e.SymbolDisplayPartKind.stringLiteral)))}if(2097152&o.flags){if(z(),!E){var O=i.getAliasedSymbol(o);if(O!==o&&O.declarations&&O.declarations.length>0){var P=O.declarations[0],M=e.getNameOfDeclaration(P);if(M){var w=e.isModuleWithStringLiteralName(P)&&e.hasSyntacticModifier(P,2),F="default"!==o.name&&!w,G=n(i,O,e.getSourceFileOfNode(P),P,M,u,F?o:O);g.push.apply(g,G.displayParts),g.push(e.lineBreakPart()),f=G.documentation,_=G.tags}}}switch(o.declarations[0].kind){case 259:g.push(e.keywordPart(92)),g.push(e.spacePart()),g.push(e.keywordPart(140));break;case 266:g.push(e.keywordPart(92)),g.push(e.spacePart()),g.push(e.keywordPart(o.declarations[0].isExportEquals?62:87));break;case 270:g.push(e.keywordPart(92));break;default:g.push(e.keywordPart(99))}g.push(e.spacePart()),X(o),e.forEach(o.declarations,(function(n){if(260===n.kind){var t=n;if(e.isExternalModuleImportEqualsDeclaration(t))g.push(e.spacePart()),g.push(e.operatorPart(62)),g.push(e.spacePart()),g.push(e.keywordPart(143)),g.push(e.punctuationPart(20)),g.push(e.displayPart(e.getTextOfNode(e.getExternalModuleImportEqualsDeclarationExpression(t)),e.SymbolDisplayPartKind.stringLiteral)),g.push(e.punctuationPart(21));else{var r=i.getSymbolAtLocation(t.moduleReference);r&&(g.push(e.spacePart()),g.push(e.operatorPart(62)),g.push(e.spacePart()),X(r,c))}return!0}}))}if(!E)if(""!==b){if(p)if(T?(z(),g.push(e.keywordPart(107))):Y(o,b),"property"===b||"JSX attribute"===b||3&h||"local var"===b||T){if(g.push(e.punctuationPart(58)),g.push(e.spacePart()),p.symbol&&262144&p.symbol.flags){var B=e.mapToDisplayParts((function(n){var r=i.typeParameterToDeclaration(p,c,t);W().writeNode(4,r,e.getSourceFileOfNode(e.getParseTreeNode(c)),n)}));e.addRange(g,B)}else e.addRange(g,e.typeToDisplayParts(i,p,c));if(o.target&&o.target.tupleLabelDeclaration){var U=o.target.tupleLabelDeclaration;e.Debug.assertNode(U.name,e.isIdentifier),g.push(e.spacePart()),g.push(e.punctuationPart(20)),g.push(e.textPart(e.idText(U.name))),g.push(e.punctuationPart(21))}}else if(16&h||8192&h||16384&h||131072&h||98304&h||"method"===b){(C=p.getNonNullableType().getCallSignatures()).length&&(Z(C[0],C),S=C.length>1)}}else b=r(i,o,l);if(0!==y.length||S||(y=o.getContextualDocumentationComment(c,i)),0===y.length&&4&h&&o.parent&&e.forEach(o.parent.declarations,(function(e){return 297===e.kind})))for(var V=0,j=o.declarations;V<j.length;V++){var H;if((H=j[V]).parent&&216===H.parent.kind){var K=i.getSymbolAtLocation(H.parent.right);if(K&&(y=K.getDocumentationComment(i),v=K.getJsDocTags(),y.length>0))break}}return 0!==v.length||S||(v=o.getJsDocTags()),0===y.length&&f&&(y=f),0===v.length&&_&&(v=_),{displayParts:g,documentation:y,symbolKind:b,tags:0===v.length?void 0:v};function W(){return m||(m=e.createPrinter({removeComments:!0})),m}function z(){g.length&&g.push(e.lineBreakPart()),q()}function q(){d&&(Q("alias"),g.push(e.spacePart()))}function J(){g.push(e.spacePart()),g.push(e.keywordPart(100)),g.push(e.spacePart())}function X(n,t){d&&n===o&&(n=d);var r=e.symbolToDisplayParts(i,n,t||s,void 0,7);e.addRange(g,r),16777216&o.flags&&g.push(e.punctuationPart(57))}function Y(n,t){z(),t&&(Q(t),n&&!e.some(n.declarations,(function(n){return e.isArrowFunction(n)||(e.isFunctionExpression(n)||e.isClassExpression(n))&&!n.name}))&&(g.push(e.spacePart()),X(n)))}function Q(n){switch(n){case"var":case"function":case"let":case"const":case"constructor":return void g.push(e.textOrKeywordPart(n));default:return g.push(e.punctuationPart(20)),g.push(e.textOrKeywordPart(n)),void g.push(e.punctuationPart(21))}}function Z(n,t,r){void 0===r&&(r=0),e.addRange(g,e.signatureToDisplayParts(i,n,c,32|r)),t.length>1&&(g.push(e.spacePart()),g.push(e.punctuationPart(20)),g.push(e.operatorPart(39)),g.push(e.displayPart((t.length-1).toString(),e.SymbolDisplayPartKind.numericLiteral)),g.push(e.spacePart()),g.push(e.textPart(2===t.length?"overload":"overloads")),g.push(e.punctuationPart(21))),y=n.getDocumentationComment(i),v=n.getJsDocTags(),t.length>1&&0===y.length&&0===v.length&&(y=t[0].getDocumentationComment(i),v=t[0].getJsDocTags())}function $(n,r){var a=e.mapToDisplayParts((function(a){var o=i.symbolToTypeParameterDeclarations(n,r,t);W().writeList(53776,o,e.getSourceFileOfNode(e.getParseTreeNode(r)),a)}));e.addRange(g,a)}}}(e.SymbolDisplay||(e.SymbolDisplay={}))}(ts||(ts={})),function(e){function n(n,t){var a=[],i=t.compilerOptions?r(t.compilerOptions,a):{},o=e.getDefaultCompilerOptions();for(var s in o)e.hasProperty(o,s)&&void 0===i[s]&&(i[s]=o[s]);for(var c=0,l=e.transpileOptionValueCompilerOptions;c<l.length;c++){var u=l[c];i[u.name]=u.transpileOptionValue}i.suppressOutputPathCheck=!0,i.allowNonTsExtensions=!0;var d=t.fileName||(t.compilerOptions&&t.compilerOptions.jsx?"module.tsx":"module.ts"),p=e.createSourceFile(d,n,i.target);t.moduleName&&(p.moduleName=t.moduleName),t.renamedDependencies&&(p.renamedDependencies=new e.Map(e.getEntries(t.renamedDependencies)));var m,f,_=e.getNewLineCharacter(i),g={getSourceFile:function(n){return n===e.normalizePath(d)?p:void 0},writeFile:function(n,t){e.fileExtensionIs(n,".map")?(e.Debug.assertEqual(f,void 0,"Unexpected multiple source map outputs, file:",n),f=t):(e.Debug.assertEqual(m,void 0,"Unexpected multiple outputs, file:",n),m=t)},getDefaultLibFileName:function(){return"lib.d.ts"},useCaseSensitiveFileNames:function(){return!1},getCanonicalFileName:function(e){return e},getCurrentDirectory:function(){return""},getNewLine:function(){return _},fileExists:function(e){return e===d},readFile:function(){return""},directoryExists:function(){return!0},getDirectories:function(){return[]}},y=e.createProgram([d],i,g);return t.reportDiagnostics&&(e.addRange(a,y.getSyntacticDiagnostics(p)),e.addRange(a,y.getOptionsDiagnostics())),y.emit(void 0,void 0,void 0,void 0,t.transformers),void 0===m?e.Debug.fail("Output generation failed"):{outputText:m,diagnostics:a,sourceMapText:f}}var t;function r(n,r){t=t||e.filter(e.optionDeclarations,(function(n){return"object"==typeof n.type&&!e.forEachEntry(n.type,(function(e){return"number"!=typeof e}))})),n=e.cloneCompilerOptions(n);for(var a=function(t){if(!e.hasProperty(n,t.name))return"continue";var a=n[t.name];e.isString(a)?n[t.name]=e.parseCustomTypeOption(t,a,r):e.forEachEntry(t.type,(function(e){return e===a}))||r.push(e.createCompilerDiagnosticForInvalidCustomType(t))},i=0,o=t;i<o.length;i++){a(o[i])}return n}e.transpileModule=n,e.transpile=function(t,r,a,i,o){var s=n(t,{compilerOptions:r,fileName:a,reportDiagnostics:!!i,moduleName:o});return e.addRange(i,s.diagnostics),s.outputText},e.fixupCompilerOptions=r}(ts||(ts={})),function(e){!function(n){!function(e){e[e.FormatDocument=0]="FormatDocument",e[e.FormatSelection=1]="FormatSelection",e[e.FormatOnEnter=2]="FormatOnEnter",e[e.FormatOnSemicolon=3]="FormatOnSemicolon",e[e.FormatOnOpeningCurlyBrace=4]="FormatOnOpeningCurlyBrace",e[e.FormatOnClosingCurlyBrace=5]="FormatOnClosingCurlyBrace"}(n.FormattingRequestKind||(n.FormattingRequestKind={}));var t=function(){function n(e,n,t){this.sourceFile=e,this.formattingRequestKind=n,this.options=t}return n.prototype.updateContext=function(n,t,r,a,i){this.currentTokenSpan=e.Debug.checkDefined(n),this.currentTokenParent=e.Debug.checkDefined(t),this.nextTokenSpan=e.Debug.checkDefined(r),this.nextTokenParent=e.Debug.checkDefined(a),this.contextNode=e.Debug.checkDefined(i),this.contextNodeAllOnSameLine=void 0,this.nextNodeAllOnSameLine=void 0,this.tokensAreOnSameLine=void 0,this.contextNodeBlockIsOnOneLine=void 0,this.nextNodeBlockIsOnOneLine=void 0},n.prototype.ContextNodeAllOnSameLine=function(){return void 0===this.contextNodeAllOnSameLine&&(this.contextNodeAllOnSameLine=this.NodeIsOnOneLine(this.contextNode)),this.contextNodeAllOnSameLine},n.prototype.NextNodeAllOnSameLine=function(){return void 0===this.nextNodeAllOnSameLine&&(this.nextNodeAllOnSameLine=this.NodeIsOnOneLine(this.nextTokenParent)),this.nextNodeAllOnSameLine},n.prototype.TokensAreOnSameLine=function(){if(void 0===this.tokensAreOnSameLine){var e=this.sourceFile.getLineAndCharacterOfPosition(this.currentTokenSpan.pos).line,n=this.sourceFile.getLineAndCharacterOfPosition(this.nextTokenSpan.pos).line;this.tokensAreOnSameLine=e===n}return this.tokensAreOnSameLine},n.prototype.ContextNodeBlockIsOnOneLine=function(){return void 0===this.contextNodeBlockIsOnOneLine&&(this.contextNodeBlockIsOnOneLine=this.BlockIsOnOneLine(this.contextNode)),this.contextNodeBlockIsOnOneLine},n.prototype.NextNodeBlockIsOnOneLine=function(){return void 0===this.nextNodeBlockIsOnOneLine&&(this.nextNodeBlockIsOnOneLine=this.BlockIsOnOneLine(this.nextTokenParent)),this.nextNodeBlockIsOnOneLine},n.prototype.NodeIsOnOneLine=function(e){return this.sourceFile.getLineAndCharacterOfPosition(e.getStart(this.sourceFile)).line===this.sourceFile.getLineAndCharacterOfPosition(e.getEnd()).line},n.prototype.BlockIsOnOneLine=function(n){var t=e.findChildOfKind(n,18,this.sourceFile),r=e.findChildOfKind(n,19,this.sourceFile);return!(!t||!r)&&this.sourceFile.getLineAndCharacterOfPosition(t.getEnd()).line===this.sourceFile.getLineAndCharacterOfPosition(r.getStart(this.sourceFile)).line},n}();n.FormattingContext=t}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){var t,r=e.createScanner(99,!1,0),a=e.createScanner(99,!1,1);!function(e){e[e.Scan=0]="Scan",e[e.RescanGreaterThanToken=1]="RescanGreaterThanToken",e[e.RescanSlashToken=2]="RescanSlashToken",e[e.RescanTemplateToken=3]="RescanTemplateToken",e[e.RescanJsxIdentifier=4]="RescanJsxIdentifier",e[e.RescanJsxText=5]="RescanJsxText",e[e.RescanJsxAttributeValue=6]="RescanJsxAttributeValue"}(t||(t={})),n.getFormattingScanner=function(t,i,o,s,c){var l=1===i?a:r;l.setText(t),l.setTextPos(o);var u,d,p,m,f,_=!0,g=c({advance:function(){f=void 0,l.getStartPos()!==o?_=!!d&&4===e.last(d).kind:l.scan();u=void 0,d=void 0;var n=l.getStartPos();for(;n<s;){var t=l.getToken();if(!e.isTrivia(t))break;l.scan();var r={pos:n,end:l.getStartPos(),kind:t};n=l.getStartPos(),u=e.append(u,r)}p=l.getStartPos()},readTokenInfo:function(t){e.Debug.assert(y());var r=function(e){switch(e.kind){case 33:case 70:case 71:case 49:case 48:return!0}return!1}(t)?1:(i=t,13===i.kind?2:function(e){return 16===e.kind||17===e.kind}(t)?3:function(n){if(n.parent)switch(n.parent.kind){case 280:case 275:case 276:case 274:return e.isKeyword(n.kind)||78===n.kind}return!1}(t)?4:function(n){if(e.isJsxText(n)){var t=e.findAncestor(n.parent,(function(n){return e.isJsxElement(n)}));return!!t&&!e.isParenthesizedExpression(t.parent)}return!1}(t)?5:(a=t,a.parent&&e.isJsxAttribute(a.parent)&&a.parent.initializer===a?6:0));var a;var i;if(f&&r===m)return h(f,t);l.getStartPos()!==p&&(e.Debug.assert(void 0!==f),l.setTextPos(p),l.scan());var o=function(n,t){var r=l.getToken();switch(m=0,t){case 1:if(31===r){m=1;var a=l.reScanGreaterToken();return e.Debug.assert(n.kind===a),a}break;case 2:if(43===(i=r)||67===i){m=2;a=l.reScanSlashToken();return e.Debug.assert(n.kind===a),a}break;case 3:if(19===r)return m=3,l.reScanTemplateToken(!1);break;case 4:return m=4,l.scanJsxIdentifier();case 5:return m=5,l.reScanJsxToken();case 6:return m=6,l.reScanJsxAttributeValue();case 0:break;default:e.Debug.assertNever(t)}var i;return r}(t,r),c=n.createTextRangeWithKind(l.getStartPos(),l.getTextPos(),o);d&&(d=void 0);for(;l.getStartPos()<s&&(o=l.scan(),e.isTrivia(o));){var _=n.createTextRangeWithKind(l.getStartPos(),l.getTextPos(),o);if(d||(d=[]),d.push(_),4===o){l.scan();break}}return h(f={leadingTrivia:u,trailingTrivia:d,token:c},t)},readEOFTokenRange:function(){return e.Debug.assert(v()),n.createTextRangeWithKind(l.getStartPos(),l.getTextPos(),1)},isOnToken:y,isOnEOF:v,getCurrentLeadingTrivia:function(){return u},lastTrailingTriviaWasNewLine:function(){return _},skipToEndOf:function(e){l.setTextPos(e.end),p=l.getStartPos(),m=void 0,f=void 0,_=!1,u=void 0,d=void 0},skipToStartOf:function(e){l.setTextPos(e.pos),p=l.getStartPos(),m=void 0,f=void 0,_=!1,u=void 0,d=void 0}});return f=void 0,l.setText(void 0),g;function y(){var n=f?f.token.kind:l.getToken();return(f?f.token.pos:l.getStartPos())<s&&1!==n&&!e.isTrivia(n)}function v(){return 1===(f?f.token.kind:l.getToken())}function h(n,t){return e.isToken(t)&&n.token.kind!==t.kind&&(n.token.kind=t.kind),n}}}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){n.anyContext=e.emptyArray,function(e){e[e.StopProcessingSpaceActions=1]="StopProcessingSpaceActions",e[e.StopProcessingTokenActions=2]="StopProcessingTokenActions",e[e.InsertSpace=4]="InsertSpace",e[e.InsertNewLine=8]="InsertNewLine",e[e.DeleteSpace=16]="DeleteSpace",e[e.DeleteToken=32]="DeleteToken",e[e.InsertTrailingSemicolon=64]="InsertTrailingSemicolon",e[e.StopAction=3]="StopAction",e[e.ModifySpaceAction=28]="ModifySpaceAction",e[e.ModifyTokenAction=96]="ModifyTokenAction"}(n.RuleAction||(n.RuleAction={})),function(e){e[e.None=0]="None",e[e.CanDeleteNewLines=1]="CanDeleteNewLines"}(n.RuleFlags||(n.RuleFlags={}))}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){function t(e,n,t,r,i,o){return void 0===o&&(o=0),{leftTokenRange:a(n),rightTokenRange:a(t),rule:{debugName:e,context:r,action:i,flags:o}}}function r(e){return{tokens:e,isSpecific:!0}}function a(n){return"number"==typeof n?r([n]):e.isArray(n)?r(n):n}function i(n,t,a){void 0===a&&(a=[]);for(var i=[],o=n;o<=t;o++)e.contains(a,o)||i.push(o);return r(i)}function o(e,n){return function(t){return t.options&&t.options[e]===n}}function s(e){return function(n){return n.options&&n.options.hasOwnProperty(e)&&!!n.options[e]}}function c(e){return function(n){return n.options&&n.options.hasOwnProperty(e)&&!n.options[e]}}function l(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!n.options[e]}}function u(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!n.options[e]||n.TokensAreOnSameLine()}}function d(e){return function(n){return!n.options||!n.options.hasOwnProperty(e)||!!n.options[e]}}function p(e){return 237===e.contextNode.kind}function m(e){return!p(e)}function f(e){switch(e.contextNode.kind){case 216:return 27!==e.contextNode.operatorToken.kind;case 217:case 184:case 224:case 270:case 265:case 172:case 182:case 183:return!0;case 198:case 254:case 260:case 249:case 160:case 291:case 163:case 162:return 62===e.currentTokenSpan.kind||62===e.nextTokenSpan.kind;case 238:case 159:return 100===e.currentTokenSpan.kind||100===e.nextTokenSpan.kind||62===e.currentTokenSpan.kind||62===e.nextTokenSpan.kind;case 239:return 156===e.currentTokenSpan.kind||156===e.nextTokenSpan.kind}return!1}function _(e){return!f(e)}function g(e){return!y(e)}function y(n){var t=n.contextNode.kind;return 163===t||162===t||160===t||249===t||e.isFunctionLikeKind(t)}function v(e){return 217===e.contextNode.kind||184===e.contextNode.kind}function h(e){return e.TokensAreOnSameLine()||x(e)}function b(e){return 196===e.contextNode.kind||190===e.contextNode.kind||function(e){return S(e)&&(e.ContextNodeAllOnSameLine()||e.ContextNodeBlockIsOnOneLine())}(e)}function E(e){return x(e)&&!(e.NextNodeAllOnSameLine()||e.NextNodeBlockIsOnOneLine())}function T(e){return S(e)&&!(e.ContextNodeAllOnSameLine()||e.ContextNodeBlockIsOnOneLine())}function S(e){return L(e.contextNode)}function x(e){return L(e.nextTokenParent)}function L(e){if(N(e))return!0;switch(e.kind){case 230:case 258:case 200:case 257:return!0}return!1}function A(e){switch(e.contextNode.kind){case 251:case 165:case 164:case 167:case 168:case 169:case 208:case 166:case 209:case 253:return!0}return!1}function D(e){return!A(e)}function C(e){return 251===e.contextNode.kind||208===e.contextNode.kind}function I(e){return N(e.contextNode)}function N(e){switch(e.kind){case 252:case 221:case 253:case 255:case 177:case 256:case 267:case 268:case 261:case 264:return!0}return!1}function R(e){switch(e.currentTokenParent.kind){case 252:case 256:case 255:case 287:case 257:case 244:return!0;case 230:var n=e.currentTokenParent.parent;if(!n||209!==n.kind&&208!==n.kind)return!0}return!1}function k(e){switch(e.contextNode.kind){case 234:case 244:case 237:case 238:case 239:case 236:case 247:case 235:case 243:case 287:return!0;default:return!1}}function O(e){return 200===e.contextNode.kind}function P(e){return function(e){return 203===e.contextNode.kind}(e)||function(e){return 204===e.contextNode.kind}(e)}function M(e){return 27!==e.currentTokenSpan.kind}function w(e){return 23!==e.nextTokenSpan.kind}function F(e){return 21!==e.nextTokenSpan.kind}function G(e){return 209===e.contextNode.kind}function B(e){return 195===e.contextNode.kind}function U(e){return e.TokensAreOnSameLine()&&11!==e.contextNode.kind}function V(e){return 11!==e.contextNode.kind}function j(e){return 273!==e.contextNode.kind&&277!==e.contextNode.kind}function H(e){return 283===e.contextNode.kind||282===e.contextNode.kind}function K(e){return 280===e.nextTokenParent.kind}function W(e){return 280===e.contextNode.kind}function z(e){return 274===e.contextNode.kind}function q(e){return!A(e)&&!x(e)}function J(e){return e.TokensAreOnSameLine()&&!!e.contextNode.decorators&&X(e.currentTokenParent)&&!X(e.nextTokenParent)}function X(n){for(;e.isExpressionNode(n);)n=n.parent;return 161===n.kind}function Y(e){return 250===e.currentTokenParent.kind&&e.currentTokenParent.getStart(e.sourceFile)===e.currentTokenSpan.pos}function Q(e){return 2!==e.formattingRequestKind}function Z(e){return 256===e.contextNode.kind}function $(e){return 177===e.contextNode.kind}function ee(e){return 170===e.contextNode.kind}function ne(e,n){if(29!==e.kind&&31!==e.kind)return!1;switch(n.kind){case 173:case 206:case 254:case 252:case 221:case 253:case 251:case 208:case 209:case 165:case 164:case 169:case 170:case 203:case 204:case 223:return!0;default:return!1}}function te(e){return ne(e.currentTokenSpan,e.currentTokenParent)||ne(e.nextTokenSpan,e.nextTokenParent)}function re(e){return 206===e.contextNode.kind}function ae(e){return 113===e.currentTokenSpan.kind&&212===e.currentTokenParent.kind}function ie(e){return 219===e.contextNode.kind&&void 0!==e.contextNode.expression}function oe(e){return 225===e.contextNode.kind}function se(e){return!function(e){switch(e.contextNode.kind){case 234:case 237:case 238:case 239:case 235:case 236:return!0;default:return!1}}(e)}function ce(n){var t=n.nextTokenSpan.kind,r=n.nextTokenSpan.pos;if(e.isTrivia(t)){var a=n.nextTokenParent===n.currentTokenParent?e.findNextToken(n.currentTokenParent,e.findAncestor(n.currentTokenParent,(function(e){return!e.parent})),n.sourceFile):n.nextTokenParent.getFirstToken(n.sourceFile);if(!a)return!0;t=a.kind,r=a.getStart(n.sourceFile)}return n.sourceFile.getLineAndCharacterOfPosition(n.currentTokenSpan.pos).line===n.sourceFile.getLineAndCharacterOfPosition(r).line?19===t||1===t:229!==t&&26!==t&&(253===n.contextNode.kind||254===n.contextNode.kind?!e.isPropertySignature(n.currentTokenParent)||!!n.currentTokenParent.type||20!==t:e.isPropertyDeclaration(n.currentTokenParent)?!n.currentTokenParent.initializer:237!==n.currentTokenParent.kind&&231!==n.currentTokenParent.kind&&229!==n.currentTokenParent.kind&&22!==t&&20!==t&&39!==t&&40!==t&&43!==t&&13!==t&&27!==t&&218!==t&&15!==t&&14!==t&&24!==t)}function le(n){return e.positionIsASICandidate(n.currentTokenSpan.end,n.currentTokenParent,n.sourceFile)}n.getAllRules=function(){for(var a=[],S=0;S<=156;S++)1!==S&&a.push(S);function L(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return{tokens:a.filter((function(n){return!e.some((function(e){return e===n}))})),isSpecific:!1}}var N={tokens:a,isSpecific:!1},X=r(__spreadArrays(a,[3])),ne=r(__spreadArrays(a,[1])),ue=i(80,156),de=i(29,77),pe=[100,101,156,126,137],me=__spreadArrays([78],e.typeKeywords),fe=X,_e=r([78,3,83,92,99]),ge=r([21,3,89,110,95,90]),ye=[t("IgnoreBeforeComment",N,[2,3],n.anyContext,1),t("IgnoreAfterLineComment",2,N,n.anyContext,1),t("NotSpaceBeforeColon",N,58,[U,_,g],16),t("SpaceAfterColon",58,N,[U,_],4),t("NoSpaceBeforeQuestionMark",N,57,[U,_,g],16),t("SpaceAfterQuestionMarkInConditionalOperator",57,N,[U,v],4),t("NoSpaceAfterQuestionMark",57,N,[U],16),t("NoSpaceBeforeDot",N,[24,28],[U],16),t("NoSpaceAfterDot",[24,28],N,[U],16),t("NoSpaceBetweenImportParenInImportType",99,20,[U,B],16),t("NoSpaceAfterUnaryPrefixOperator",[45,46,54,53],[8,9,78,20,22,18,107,102],[U,_],16),t("NoSpaceAfterUnaryPreincrementOperator",45,[78,20,107,102],[U],16),t("NoSpaceAfterUnaryPredecrementOperator",46,[78,20,107,102],[U],16),t("NoSpaceBeforeUnaryPostincrementOperator",[78,21,23,102],45,[U,se],16),t("NoSpaceBeforeUnaryPostdecrementOperator",[78,21,23,102],46,[U,se],16),t("SpaceAfterPostincrementWhenFollowedByAdd",45,39,[U,f],4),t("SpaceAfterAddWhenFollowedByUnaryPlus",39,39,[U,f],4),t("SpaceAfterAddWhenFollowedByPreincrement",39,45,[U,f],4),t("SpaceAfterPostdecrementWhenFollowedBySubtract",46,40,[U,f],4),t("SpaceAfterSubtractWhenFollowedByUnaryMinus",40,40,[U,f],4),t("SpaceAfterSubtractWhenFollowedByPredecrement",40,46,[U,f],4),t("NoSpaceAfterCloseBrace",19,[27,26],[U],16),t("NewLineBeforeCloseBraceInBlockContext",X,19,[T],8),t("SpaceAfterCloseBrace",19,L(21),[U,R],4),t("SpaceBetweenCloseBraceAndElse",19,90,[U],4),t("SpaceBetweenCloseBraceAndWhile",19,114,[U],4),t("NoSpaceBetweenEmptyBraceBrackets",18,19,[U,O],16),t("SpaceAfterConditionalClosingParen",21,22,[k],4),t("NoSpaceBetweenFunctionKeywordAndStar",97,41,[C],16),t("SpaceAfterStarInGeneratorDeclaration",41,78,[C],4),t("SpaceAfterFunctionInFuncDecl",97,N,[A],4),t("NewLineAfterOpenBraceInBlockContext",18,N,[T],8),t("SpaceAfterGetSetInMember",[134,146],78,[A],4),t("NoSpaceBetweenYieldKeywordAndStar",124,41,[U,ie],16),t("SpaceBetweenYieldOrYieldStarAndOperand",[124,41],N,[U,ie],4),t("NoSpaceBetweenReturnAndSemicolon",104,26,[U],16),t("SpaceAfterCertainKeywords",[112,108,102,88,104,111,130],N,[U],4),t("SpaceAfterLetConstInVariableDeclaration",[118,84],N,[U,Y],4),t("NoSpaceBeforeOpenParenInFuncCall",N,20,[U,P,M],16),t("SpaceBeforeBinaryKeywordOperator",N,pe,[U,f],4),t("SpaceAfterBinaryKeywordOperator",pe,N,[U,f],4),t("SpaceAfterVoidOperator",113,N,[U,ae],4),t("SpaceBetweenAsyncAndOpenParen",129,20,[G,U],4),t("SpaceBetweenAsyncAndFunctionKeyword",129,[97,78],[U],4),t("NoSpaceBetweenTagAndTemplateString",[78,21],[14,15],[U],16),t("SpaceBeforeJsxAttribute",N,78,[K,U],4),t("SpaceBeforeSlashInJsxOpeningElement",N,43,[z,U],4),t("NoSpaceBeforeGreaterThanTokenInJsxOpeningElement",43,31,[z,U],16),t("NoSpaceBeforeEqualInJsxAttribute",N,62,[W,U],16),t("NoSpaceAfterEqualInJsxAttribute",62,N,[W,U],16),t("NoSpaceAfterModuleImport",[139,143],20,[U],16),t("SpaceAfterCertainTypeScriptKeywords",[125,83,133,87,91,92,93,134,116,99,117,139,140,120,122,121,142,146,123,149,153,138,135],N,[U],4),t("SpaceBeforeCertainTypeScriptKeywords",N,[93,116,153],[U],4),t("SpaceAfterModuleName",10,18,[Z],4),t("SpaceBeforeArrow",N,38,[U],4),t("SpaceAfterArrow",38,N,[U],4),t("NoSpaceAfterEllipsis",25,78,[U],16),t("NoSpaceAfterOptionalParameters",57,[21,27],[U,_],16),t("NoSpaceBetweenEmptyInterfaceBraceBrackets",18,19,[U,$],16),t("NoSpaceBeforeOpenAngularBracket",me,29,[U,te],16),t("NoSpaceBetweenCloseParenAndAngularBracket",21,29,[U,te],16),t("NoSpaceAfterOpenAngularBracket",29,N,[U,te],16),t("NoSpaceBeforeCloseAngularBracket",N,31,[U,te],16),t("NoSpaceAfterCloseAngularBracket",31,[20,22,31,27],[U,te,D],16),t("SpaceBeforeAt",[21,78],59,[U],4),t("NoSpaceAfterAt",59,N,[U],16),t("SpaceAfterDecorator",N,[125,78,92,87,83,123,122,120,121,134,146,22,41],[J],4),t("NoSpaceBeforeNonNullAssertionOperator",N,53,[U,oe],16),t("NoSpaceAfterNewKeywordOnConstructorSignature",102,20,[U,ee],16),t("SpaceLessThanAndNonJSXTypeAnnotation",29,29,[U],4)],ve=[t("SpaceAfterConstructor",132,20,[s("insertSpaceAfterConstructor"),U],4),t("NoSpaceAfterConstructor",132,20,[l("insertSpaceAfterConstructor"),U],16),t("SpaceAfterComma",27,N,[s("insertSpaceAfterCommaDelimiter"),U,j,w,F],4),t("NoSpaceAfterComma",27,N,[l("insertSpaceAfterCommaDelimiter"),U,j],16),t("SpaceAfterAnonymousFunctionKeyword",[97,41],20,[s("insertSpaceAfterFunctionKeywordForAnonymousFunctions"),A],4),t("NoSpaceAfterAnonymousFunctionKeyword",[97,41],20,[l("insertSpaceAfterFunctionKeywordForAnonymousFunctions"),A],16),t("SpaceAfterKeywordInControl",ue,20,[s("insertSpaceAfterKeywordsInControlFlowStatements"),k],4),t("NoSpaceAfterKeywordInControl",ue,20,[l("insertSpaceAfterKeywordsInControlFlowStatements"),k],16),t("SpaceAfterOpenParen",20,N,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),U],4),t("SpaceBeforeCloseParen",N,21,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),U],4),t("SpaceBetweenOpenParens",20,20,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),U],4),t("NoSpaceBetweenParens",20,21,[U],16),t("NoSpaceAfterOpenParen",20,N,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),U],16),t("NoSpaceBeforeCloseParen",N,21,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"),U],16),t("SpaceAfterOpenBracket",22,N,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),U],4),t("SpaceBeforeCloseBracket",N,23,[s("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),U],4),t("NoSpaceBetweenBrackets",22,23,[U],16),t("NoSpaceAfterOpenBracket",22,N,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),U],16),t("NoSpaceBeforeCloseBracket",N,23,[l("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"),U],16),t("SpaceAfterOpenBrace",18,N,[d("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),b],4),t("SpaceBeforeCloseBrace",N,19,[d("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),b],4),t("NoSpaceBetweenEmptyBraceBrackets",18,19,[U,O],16),t("NoSpaceAfterOpenBrace",18,N,[c("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),U],16),t("NoSpaceBeforeCloseBrace",N,19,[c("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"),U],16),t("SpaceBetweenEmptyBraceBrackets",18,19,[s("insertSpaceAfterOpeningAndBeforeClosingEmptyBraces")],4),t("NoSpaceBetweenEmptyBraceBrackets",18,19,[c("insertSpaceAfterOpeningAndBeforeClosingEmptyBraces"),U],16),t("SpaceAfterTemplateHeadAndMiddle",[15,16],N,[s("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),V],4,1),t("SpaceBeforeTemplateMiddleAndTail",N,[16,17],[s("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),U],4),t("NoSpaceAfterTemplateHeadAndMiddle",[15,16],N,[l("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),V],16,1),t("NoSpaceBeforeTemplateMiddleAndTail",N,[16,17],[l("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"),U],16),t("SpaceAfterOpenBraceInJsxExpression",18,N,[s("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),U,H],4),t("SpaceBeforeCloseBraceInJsxExpression",N,19,[s("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),U,H],4),t("NoSpaceAfterOpenBraceInJsxExpression",18,N,[l("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),U,H],16),t("NoSpaceBeforeCloseBraceInJsxExpression",N,19,[l("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"),U,H],16),t("SpaceAfterSemicolonInFor",26,N,[s("insertSpaceAfterSemicolonInForStatements"),U,p],4),t("NoSpaceAfterSemicolonInFor",26,N,[l("insertSpaceAfterSemicolonInForStatements"),U,p],16),t("SpaceBeforeBinaryOperator",N,de,[s("insertSpaceBeforeAndAfterBinaryOperators"),U,f],4),t("SpaceAfterBinaryOperator",de,N,[s("insertSpaceBeforeAndAfterBinaryOperators"),U,f],4),t("NoSpaceBeforeBinaryOperator",N,de,[l("insertSpaceBeforeAndAfterBinaryOperators"),U,f],16),t("NoSpaceAfterBinaryOperator",de,N,[l("insertSpaceBeforeAndAfterBinaryOperators"),U,f],16),t("SpaceBeforeOpenParenInFuncDecl",N,20,[s("insertSpaceBeforeFunctionParenthesis"),U,A],4),t("NoSpaceBeforeOpenParenInFuncDecl",N,20,[l("insertSpaceBeforeFunctionParenthesis"),U,A],16),t("NewLineBeforeOpenBraceInControl",ge,18,[s("placeOpenBraceOnNewLineForControlBlocks"),k,E],8,1),t("NewLineBeforeOpenBraceInFunction",fe,18,[s("placeOpenBraceOnNewLineForFunctions"),A,E],8,1),t("NewLineBeforeOpenBraceInTypeScriptDeclWithBlock",_e,18,[s("placeOpenBraceOnNewLineForFunctions"),I,E],8,1),t("SpaceAfterTypeAssertion",31,N,[s("insertSpaceAfterTypeAssertion"),U,re],4),t("NoSpaceAfterTypeAssertion",31,N,[l("insertSpaceAfterTypeAssertion"),U,re],16),t("SpaceBeforeTypeAnnotation",N,[57,58],[s("insertSpaceBeforeTypeAnnotation"),U,y],4),t("NoSpaceBeforeTypeAnnotation",N,[57,58],[l("insertSpaceBeforeTypeAnnotation"),U,y],16),t("NoOptionalSemicolon",26,ne,[o("semicolons",e.SemicolonPreference.Remove),ce],32),t("OptionalSemicolon",N,ne,[o("semicolons",e.SemicolonPreference.Insert),le],64)],he=[t("NoSpaceBeforeSemicolon",N,26,[U],16),t("SpaceBeforeOpenBraceInControl",ge,18,[u("placeOpenBraceOnNewLineForControlBlocks"),k,Q,h],4,1),t("SpaceBeforeOpenBraceInFunction",fe,18,[u("placeOpenBraceOnNewLineForFunctions"),A,x,Q,h],4,1),t("SpaceBeforeOpenBraceInTypeScriptDeclWithBlock",_e,18,[u("placeOpenBraceOnNewLineForFunctions"),I,Q,h],4,1),t("NoSpaceBeforeComma",N,27,[U],16),t("NoSpaceBeforeOpenBracket",L(129,81),22,[U],16),t("NoSpaceAfterCloseBracket",23,N,[U,q],16),t("SpaceAfterSemicolon",26,N,[U],4),t("SpaceBetweenForAndAwaitKeyword",96,130,[U],4),t("SpaceBetweenStatements",[21,89,90,81],N,[U,j,m],4),t("SpaceAfterTryCatchFinally",[110,82,95],18,[U],4)];return __spreadArrays(ye,ve,he)}}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){var t;function r(){var r,o;return void 0===t&&(r=n.getAllRules(),o=function(e){for(var n=new Array(c*c),t=new Array(n.length),r=0,a=e;r<a.length;r++)for(var o=a[r],s=o.leftTokenRange.isSpecific&&o.rightTokenRange.isSpecific,u=0,d=o.leftTokenRange.tokens;u<d.length;u++)for(var p=d[u],m=0,f=o.rightTokenRange.tokens;m<f.length;m++){var _=i(p,f[m]),g=n[_];void 0===g&&(g=n[_]=[]),l(g,o.rule,s,t,_)}return n}(r),t=function(n){var t=o[i(n.currentTokenSpan.kind,n.nextTokenSpan.kind)];if(t){for(var r=[],s=0,c=0,l=t;c<l.length;c++){var u=l[c],d=~a(s);u.action&d&&e.every(u.context,(function(e){return e(n)}))&&(r.push(u),s|=u.action)}if(r.length)return r}}),t}function a(e){var n=0;return 1&e&&(n|=28),2&e&&(n|=96),28&e&&(n|=28),96&e&&(n|=96),n}function i(n,t){return e.Debug.assert(n<=156&&t<=156,"Must compute formatting context from tokens"),n*c+t}n.getFormatContext=function(e,n){return{options:e,getRules:r(),host:n}};var o,s=31,c=157;function l(t,r,a,i,c){var l,u,d,p=3&r.action?a?o.StopRulesSpecific:o.StopRulesAny:r.context!==n.anyContext?a?o.ContextRulesSpecific:o.ContextRulesAny:a?o.NoContextRulesSpecific:o.NoContextRulesAny,m=i[c]||0;t.splice(function(e,n){for(var t=0,r=0;r<=n;r+=5)t+=31&e,e>>=5;return t}(m,p),0,r),i[c]=(d=1+((l=m)>>(u=p)&s),e.Debug.assert((d&s)===d,"Adding more rules into the sub-bucket than allowed. Maximum allowed is 32 rules."),l&~(s<<u)|d<<u)}!function(e){e[e.StopRulesSpecific=0]="StopRulesSpecific",e[e.StopRulesAny=5]="StopRulesAny",e[e.ContextRulesSpecific=10]="ContextRulesSpecific",e[e.ContextRulesAny=15]="ContextRulesAny",e[e.NoContextRulesSpecific=20]="NoContextRulesSpecific",e[e.NoContextRulesAny=25]="NoContextRulesAny"}(o||(o={}))}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){var t,r,a,i,o;function s(n,t,r){var a=e.findPrecedingToken(n,r);return a&&a.kind===t&&n===a.getEnd()?a:void 0}function c(e){for(var n=e;n&&n.parent&&n.parent.end===e.end&&!l(n.parent,n);)n=n.parent;return n}function l(n,t){switch(n.kind){case 252:case 253:return e.rangeContainsRange(n.members,t);case 256:var r=n.body;return!!r&&257===r.kind&&e.rangeContainsRange(r.statements,t);case 297:case 230:case 257:return e.rangeContainsRange(n.statements,t);case 287:return e.rangeContainsRange(n.block.statements,t)}return!1}function u(n,t,r,a){return n?d({pos:e.getLineStartPositionForPosition(n.getStart(t),t),end:n.end},t,r,a):[]}function d(t,r,a,i){var o=function(n,t){return function r(a){var i=e.forEachChild(a,(function(r){return e.startEndContainsRange(r.getStart(t),r.end,n)&&r}));if(i){var o=r(i);if(o)return o}return a}(t)}(t,r);return n.getFormattingScanner(r.text,r.languageVariant,function(n,t,r){var a=n.getStart(r);if(a===t.pos&&n.end===t.end)return a;var i=e.findPrecedingToken(t.pos,r);return i?i.end>=t.pos?n.pos:i.end:n.pos}(o,t,r),t.end,(function(s){return p(t,o,n.SmartIndenter.getIndentationForNode(o,t,r,a.options),function(e,t,r){for(var a,i=-1;e;){var o=r.getLineAndCharacterOfPosition(e.getStart(r)).line;if(-1!==i&&o!==i)break;if(n.SmartIndenter.shouldIndentChildNode(t,e,a,r))return t.indentSize;i=o,a=e,e=e.parent}return 0}(o,a.options,r),s,a,i,function(n,t){if(!n.length)return i;var r=n.filter((function(n){return e.rangeOverlapsWithStartEnd(t,n.start,n.start+n.length)})).sort((function(e,n){return e.start-n.start}));if(!r.length)return i;var a=0;return function(n){for(;;){if(a>=r.length)return!1;var t=r[a];if(n.end<=t.start)return!1;if(e.startEndOverlapsWithStartEnd(n.pos,n.end,t.start,t.start+t.length))return!0;a++}};function i(){return!1}}(r.parseDiagnostics,t),r)}))}function p(t,r,a,i,o,s,c,l,u){var d,p,f,_,g=s.options,y=s.getRules,v=s.host,h=new n.FormattingContext(u,c,g),b=-1,E=[];if(o.advance(),o.isOnToken()){var T=u.getLineAndCharacterOfPosition(r.getStart(u)).line,S=T;r.decorators&&(S=u.getLineAndCharacterOfPosition(e.getNonDecoratorTokenPosOfNode(r,u)).line),function r(a,i,s,c,m,y){if(!e.rangeOverlapsWithStartEnd(t,a.getStart(u),a.getEnd()))return;var v=L(a,s,m,y),h=i;e.forEachChild(a,(function(e){S(e,-1,a,v,s,c,!1)}),(function(e){x(e,a,s,v)}));for(;o.isOnToken();){var E=o.readTokenInfo(a);if(E.token.end>a.end)break;11!==a.kind?k(E,a,v,a):o.advance()}if(!a.parent&&o.isOnEOF()){var T=o.readEOFTokenRange();T.end<=a.end&&d&&I(T,u.getLineAndCharacterOfPosition(T.pos).line,a,d,f,p,i,v)}function S(i,s,c,l,d,p,m,f){var y=i.getStart(u),v=u.getLineAndCharacterOfPosition(y).line,E=v;i.decorators&&(E=u.getLineAndCharacterOfPosition(e.getNonDecoratorTokenPosOfNode(i,u)).line);var T=-1;if(m&&e.rangeContainsRange(t,c)&&-1!==(T=function(t,r,a,i,o){if(e.rangeOverlapsWithStartEnd(i,t,r)||e.rangeContainsStartEnd(i,t,r)){if(-1!==o)return o}else{var s=u.getLineAndCharacterOfPosition(t).line,c=e.getLineStartPositionForPosition(t,u),l=n.SmartIndenter.findFirstNonWhitespaceColumn(c,t,u,g);if(s!==a||t===l){var d=n.SmartIndenter.getBaseIndentation(g);return d>l?d:l}}return-1}(y,i.end,d,t,s))&&(s=T),!e.rangeOverlapsWithStartEnd(t,i.pos,i.end))return i.end<t.pos&&o.skipToEndOf(i),s;if(0===i.getFullWidth())return s;for(;o.isOnToken();){if((S=o.readTokenInfo(a)).token.end>y){S.token.pos>y&&o.skipToStartOf(i);break}k(S,a,l,a)}if(!o.isOnToken())return s;if(e.isToken(i)){var S=o.readTokenInfo(i);if(11!==i.kind)return e.Debug.assert(S.token.end===i.end,"Token end is child end"),k(S,a,l,i),s}var x=161===i.kind?v:p,L=function(e,t,r,a,i,o){var s=n.SmartIndenter.shouldIndentChildNode(g,e)?g.indentSize:0;return o===t?{indentation:t===_?b:i.getIndentation(),delta:Math.min(g.indentSize,i.getDelta(e)+s)}:-1===r?20===e.kind&&t===_?{indentation:b,delta:i.getDelta(e)}:n.SmartIndenter.childStartsOnTheSameLineWithElseInIfStatement(a,e,t,u)||n.SmartIndenter.argumentStartsOnSameLineAsPreviousArgument(a,e,t,u)?{indentation:i.getIndentation(),delta:s}:{indentation:i.getIndentation()+i.getDelta(e),delta:s}:{indentation:r,delta:s}}(i,v,T,a,l,x);if(r(i,h,v,E,L.indentation,L.delta),11===i.kind){var A={pos:i.getStart(),end:i.getEnd()};if(A.pos!==A.end){var D=c.getChildren(u),C=D[e.findIndex(D,(function(e){return e.pos===i.pos}))-1];if(C&&u.getLineAndCharacterOfPosition(A.end).line!==u.getLineAndCharacterOfPosition(C.end).line){var I=u.getLineAndCharacterOfPosition(A.pos).line===u.getLineAndCharacterOfPosition(C.end).line;R(A,L.indentation,I,!1,!0)}}}return h=a,f&&199===c.kind&&-1===s&&(s=L.indentation),s}function x(t,r,i,s){e.Debug.assert(e.isNodeArray(t));var c=function(e,n){switch(e.kind){case 166:case 251:case 208:case 165:case 164:case 209:if(e.typeParameters===n)return 29;if(e.parameters===n)return 20;break;case 203:case 204:if(e.typeArguments===n)return 29;if(e.arguments===n)return 20;break;case 173:if(e.typeArguments===n)return 29;break;case 177:return 18}return 0}(r,t),l=s,d=i;if(0!==c)for(;o.isOnToken();){if((v=o.readTokenInfo(r)).token.end>t.pos)break;if(v.token.kind===c){d=u.getLineAndCharacterOfPosition(v.token.pos).line,k(v,r,s,r);var p=void 0;if(-1!==b)p=b;else{var m=e.getLineStartPositionForPosition(v.token.pos,u);p=n.SmartIndenter.findFirstNonWhitespaceColumn(m,v.token.pos,u,g)}l=L(r,i,p,g.indentSize)}else k(v,r,s,r)}for(var f=-1,_=0;_<t.length;_++){f=S(t[_],f,a,l,d,d,!0,0===_)}var y=function(e){switch(e){case 20:return 21;case 29:return 31;case 18:return 19}return 0}(c);if(0!==y&&o.isOnToken()){var v;if(27===(v=o.readTokenInfo(r)).token.kind&&e.isCallLikeExpression(r))d!==u.getLineAndCharacterOfPosition(v.token.pos).line&&(o.advance(),v=o.isOnToken()?o.readTokenInfo(r):void 0);v&&v.token.kind===y&&e.rangeContainsRange(r,v.token)&&k(v,r,l,r,!0)}}function k(n,r,a,i,s){e.Debug.assert(e.rangeContainsRange(r,n.token));var c=o.lastTrailingTriviaWasNewLine(),p=!1;n.leadingTrivia&&D(n.leadingTrivia,r,h,a);var m=0,f=e.rangeContainsRange(t,n.token),g=u.getLineAndCharacterOfPosition(n.token.pos);if(f){var y=l(n.token),v=d;if(m=C(n.token,g,r,h,a),!y)if(0===m){var E=v&&u.getLineAndCharacterOfPosition(v.end).line;p=c&&g.line!==E}else p=1===m}if(n.trailingTrivia&&D(n.trailingTrivia,r,h,a),p){var T=f&&!l(n.token)?a.getIndentationForToken(g.line,n.token.kind,i,!!s):-1,S=!0;if(n.leadingTrivia){var x=a.getIndentationForComment(n.token.kind,T,i);S=A(n.leadingTrivia,x,S,(function(e){return N(e.pos,x,!1)}))}-1!==T&&S&&(N(n.token.pos,T,1===m),_=g.line,b=T)}o.advance(),h=r}}(r,r,T,S,a,i)}if(!o.isOnToken()){var x=o.getCurrentLeadingTrivia();x&&(A(x,a,!1,(function(e){return C(e,u.getLineAndCharacterOfPosition(e.pos),r,r,void 0)})),!1!==g.trimTrailingWhitespace&&function(){var e=d?d.end:t.pos,n=u.getLineAndCharacterOfPosition(e).line,r=u.getLineAndCharacterOfPosition(t.end).line;k(n,r+1,d)}())}return E;function L(t,r,a,i){return{getIndentationForComment:function(e,n,t){switch(e){case 19:case 23:case 21:return a+o(t)}return-1!==n?n:a},getIndentationForToken:function(n,i,s,c){return!c&&function(n,a,i){switch(a){case 18:case 19:case 21:case 90:case 114:case 59:return!1;case 43:case 31:switch(i.kind){case 275:case 276:case 274:return!1}break;case 22:case 23:if(190!==i.kind)return!1}return r!==n&&!(t.decorators&&a===function(n){if(n.modifiers&&n.modifiers.length)return n.modifiers[0].kind;switch(n.kind){case 252:return 83;case 253:return 117;case 251:return 97;case 255:return 255;case 167:return 134;case 168:return 146;case 165:if(n.asteriskToken)return 41;case 163:case 160:var t=e.getNameOfDeclaration(n);if(t)return t.kind}}(t))}(n,i,s)?a+o(s):a},getIndentation:function(){return a},getDelta:o,recomputeIndentation:function(e,r){n.SmartIndenter.shouldIndentChildNode(g,r,t,u)&&(a+=e?g.indentSize:-g.indentSize,i=n.SmartIndenter.shouldIndentChildNode(g,t)?g.indentSize:0)}};function o(e){return n.SmartIndenter.nodeWillIndentChild(g,t,e,u,!0)?i:0}}function A(n,r,a,i){for(var o=0,s=n;o<s.length;o++){var c=s[o],l=e.rangeContainsRange(t,c);switch(c.kind){case 3:l&&R(c,r,!a),a=!1;break;case 2:a&&l&&i(c),a=!1;break;case 4:a=!0}}return a}function D(n,r,a,i){for(var o=0,s=n;o<s.length;o++){var c=s[o];if(e.isComment(c.kind)&&e.rangeContainsRange(t,c))C(c,u.getLineAndCharacterOfPosition(c.pos),r,a,i)}}function C(e,n,r,a,i){var o=0;l(e)||(d?o=I(e,n.line,r,d,f,p,a,i):k(u.getLineAndCharacterOfPosition(t.pos).line,n.line));return d=e,p=r,f=n.line,o}function I(n,t,r,a,i,o,s,c){h.updateContext(a,o,n,r,s);var l=y(h),d=!1!==h.options.trimTrailingWhitespace,p=0;return l?e.forEachRight(l,(function(o){switch(p=function(n,t,r,a,i){var o=i!==r;switch(n.action){case 1:return 0;case 16:if(t.end!==a.pos)return P(t.end,a.pos-t.end),o?2:0;break;case 32:P(t.pos,t.end-t.pos);break;case 8:if(1!==n.flags&&r!==i)return 0;if(1!==i-r)return M(t.end,a.pos-t.end,e.getNewLineOrDefaultFromHost(v,g)),o?0:1;break;case 4:if(1!==n.flags&&r!==i)return 0;if(1!==a.pos-t.end||32!==u.text.charCodeAt(t.end))return M(t.end,a.pos-t.end," "),o?2:0;break;case 64:s=t.end,(c=";")&&E.push(e.createTextChangeFromStartLength(s,0,c))}var s,c;return 0}(o,a,i,n,t)){case 2:r.getStart(u)===n.pos&&c.recomputeIndentation(!1,s);break;case 1:r.getStart(u)===n.pos&&c.recomputeIndentation(!0,s);break;default:e.Debug.assert(0===p)}d=d&&!(16&o.action)&&1!==o.flags})):d=d&&1!==n.kind,t!==i&&d&&k(i,t,a),p}function N(n,t,r){var a=m(t,g);if(r)M(n,0,a);else{var i=u.getLineAndCharacterOfPosition(n),o=e.getStartPositionOfLine(i.line,u);(t!==function(e,n){for(var t=0,r=0;r<n;r++)9===u.text.charCodeAt(e+r)?t+=g.tabSize-t%g.tabSize:t++;return t}(o,i.character)||function(e,n){return e!==u.text.substr(n,e.length)}(a,o))&&M(o,i.character,a)}}function R(t,r,a,i,o){void 0===i&&(i=!0);var s=u.getLineAndCharacterOfPosition(t.pos).line,c=u.getLineAndCharacterOfPosition(t.end).line;if(s!==c){for(var l=[],d=t.pos,p=s;p<c;p++){var f=e.getEndLinePosition(p,u);l.push({pos:d,end:f}),d=e.getStartPositionOfLine(p+1,u)}if(i&&l.push({pos:d,end:t.end}),0!==l.length){var _=e.getStartPositionOfLine(s,u),y=n.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(_,l[0].pos,u,g),v=0;a&&(v=1,s++);for(var h=r-y.column,b=v;b<l.length;b++,s++){var E=e.getStartPositionOfLine(s,u),T=0===b?y:n.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(l[b].pos,l[b].end,u,g);if(o){if(e.isLineBreak(u.text.charCodeAt(e.getStartPositionOfLine(s,u))))continue;h=r-T.column}var S=T.column+h;if(S>0){var x=m(S,g);M(E,T.character,x)}else P(E,T.character)}}}else a||N(t.pos,r,!1)}function k(n,t,r){for(var a=n;a<t;a++){var i=e.getStartPositionOfLine(a,u),o=e.getEndLinePosition(a,u);if(!(r&&(e.isComment(r.kind)||e.isStringOrRegularExpressionOrTemplateLiteral(r.kind))&&r.pos<=o&&r.end>o)){var s=O(i,o);-1!==s&&(e.Debug.assert(s===i||!e.isWhiteSpaceSingleLine(u.text.charCodeAt(s-1))),P(s,o+1-s))}}}function O(n,t){for(var r=t;r>=n&&e.isWhiteSpaceSingleLine(u.text.charCodeAt(r));)r--;return r!==t?r+1:-1}function P(n,t){t&&E.push(e.createTextChangeFromStartLength(n,t,""))}function M(n,t,r){(t||r)&&E.push(e.createTextChangeFromStartLength(n,t,r))}}function m(n,t){if((!a||a.tabSize!==t.tabSize||a.indentSize!==t.indentSize)&&(a={tabSize:t.tabSize,indentSize:t.indentSize},i=o=void 0),t.convertTabsToSpaces){var r=void 0,s=Math.floor(n/t.indentSize),c=n%t.indentSize;return o||(o=[]),void 0===o[s]?(r=e.repeatString(" ",t.indentSize*s),o[s]=r):r=o[s],c?r+e.repeatString(" ",c):r}var l=Math.floor(n/t.tabSize),u=n-l*t.tabSize,d=void 0;return i||(i=[]),void 0===i[l]?i[l]=d=e.repeatString("\t",l):d=i[l],u?d+e.repeatString(" ",u):d}n.createTextRangeWithKind=function(n,t,r){var a={pos:n,end:t,kind:r};return e.Debug.isDebugging&&Object.defineProperty(a,"__debugKind",{get:function(){return e.Debug.formatSyntaxKind(r)}}),a},function(e){e[e.Unknown=-1]="Unknown"}(t||(t={})),n.formatOnEnter=function(n,t,r){var a=t.getLineAndCharacterOfPosition(n).line;if(0===a)return[];for(var i=e.getEndLinePosition(a,t);e.isWhiteSpaceSingleLine(t.text.charCodeAt(i));)i--;return e.isLineBreak(t.text.charCodeAt(i))&&i--,d({pos:e.getStartPositionOfLine(a-1,t),end:i+1},t,r,2)},n.formatOnSemicolon=function(e,n,t){return u(c(s(e,26,n)),n,t,3)},n.formatOnOpeningCurly=function(n,t,r){var a=s(n,18,t);if(!a)return[];var i=c(a.parent);return d({pos:e.getLineStartPositionForPosition(i.getStart(t),t),end:n},t,r,4)},n.formatOnClosingCurly=function(e,n,t){return u(c(s(e,19,n)),n,t,5)},n.formatDocument=function(e,n){return d({pos:0,end:e.text.length},e,n,0)},n.formatSelection=function(n,t,r,a){return d({pos:e.getLineStartPositionForPosition(n,r),end:t},r,a,1)},n.formatNodeGivenIndentation=function(e,t,r,a,i,o){var s={pos:0,end:t.text.length};return n.getFormattingScanner(t.text,r,s.pos,s.end,(function(n){return p(s,e,a,i,n,o,1,(function(e){return!1}),t)}))},function(e){e[e.None=0]="None",e[e.LineAdded=1]="LineAdded",e[e.LineRemoved=2]="LineRemoved"}(r||(r={})),n.getRangeOfEnclosingComment=function(n,t,r,a){void 0===a&&(a=e.getTokenAtPosition(n,t));var i=e.findAncestor(a,e.isJSDoc);if(i&&(a=i.parent),!(a.getStart(n)<=t&&t<a.getEnd())){var o=(r=null===r?void 0:void 0===r?e.findPrecedingToken(t,n):r)&&e.getTrailingCommentRanges(n.text,r.end),s=e.getLeadingCommentRangesOfNode(a,n),c=e.concatenate(o,s);return c&&e.find(c,(function(r){return e.rangeContainsPositionExclusive(r,t)||t===r.end&&(2===r.kind||t===n.getFullWidth())}))}},n.getIndentationString=m}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r,a;function i(e){return e.baseIndentSize||0}function o(e,n,t,r,a,o,l){for(var u=e.parent;u;){var m=!0;if(t){var f=e.getStart(a);m=f<t.pos||f>t.end}var _=s(u,e,a),y=_.line===n.line||p(u,e,n.line,a);if(m){var v=g(e,a,l,!y);if(-1!==v)return v+r;if(-1!==(v=c(e,u,n,y,a,l)))return v+r}T(l,u,e,a,o)&&!y&&(r+=l.indentSize);var h=d(u,e,n.line,a);u=(e=u).parent,n=h?a.getLineAndCharacterOfPosition(e.getStart(a)):_}return r+i(l)}function s(e,n,t){var r=m(n,t),a=r?r.pos:e.getStart(t);return t.getLineAndCharacterOfPosition(a)}function c(n,t,r,a,i,o){return(e.isDeclaration(n)||e.isStatementButNotDeclaration(n))&&(297===t.kind||!a)?v(r,i,o):-1}function l(n,t,r,a){var i=e.findNextToken(n,t,a);return i?18===i.kind?1:19===i.kind&&r===u(i,a).line?2:0:0}function u(e,n){return n.getLineAndCharacterOfPosition(e.getStart(n))}function d(n,t,r,a){if(!e.isCallExpression(n)||!e.contains(n.arguments,t))return!1;var i=n.expression.getEnd();return e.getLineAndCharacterOfPosition(a,i).line===r}function p(n,t,r,a){if(234===n.kind&&n.elseStatement===t){var i=e.findChildOfKind(n,90,a);return e.Debug.assert(void 0!==i),u(i,a).line===r}return!1}function m(e,n){return e.parent&&f(e.getStart(n),e.getEnd(),e.parent,n)}function f(n,t,r,a){switch(r.kind){case 173:return i(r.typeArguments);case 200:return i(r.properties);case 199:return i(r.elements);case 177:return i(r.members);case 251:case 208:case 209:case 165:case 164:case 169:case 166:case 175:case 170:return i(r.typeParameters)||i(r.parameters);case 252:case 221:case 253:case 254:case 330:return i(r.typeParameters);case 204:case 203:return i(r.typeArguments)||i(r.arguments);case 250:return i(r.declarations);case 264:case 268:return i(r.elements);case 196:case 197:return i(r.elements)}function i(i){return i&&e.rangeContainsStartEnd(function(e,n,t){for(var r=e.getChildren(t),a=1;a<r.length-1;a++)if(r[a].pos===n.pos&&r[a].end===n.end)return{pos:r[a-1].end,end:r[a+1].getStart(t)};return n}(r,i,a),n,t)?i:void 0}}function _(e,n,t){return e?v(n.getLineAndCharacterOfPosition(e.pos),n,t):-1}function g(e,n,t,r){if(e.parent&&250===e.parent.kind)return-1;var a=m(e,n);if(a){var i=a.indexOf(e);if(-1!==i){var o=y(a,i,n,t);if(-1!==o)return o}return _(a,n,t)+(r?t.indentSize:0)}return-1}function y(n,t,r,a){e.Debug.assert(t>=0&&t<n.length);for(var i=u(n[t],r),o=t-1;o>=0;o--)if(27!==n[o].kind){if(r.getLineAndCharacterOfPosition(n[o].end).line!==i.line)return v(i,r,a);i=u(n[o],r)}return-1}function v(e,n,t){var r=n.getPositionOfLineAndCharacter(e.line,0);return b(r,r+e.character,n,t)}function h(n,t,r,a){for(var i=0,o=0,s=n;s<t;s++){var c=r.text.charCodeAt(s);if(!e.isWhiteSpaceSingleLine(c))break;9===c?o+=a.tabSize+o%a.tabSize:o++,i++}return{column:o,character:i}}function b(e,n,t,r){return h(e,n,t,r).column}function E(e,n,t,r,a){var i=t?t.kind:0;switch(n.kind){case 233:case 252:case 221:case 253:case 255:case 254:case 199:case 230:case 257:case 200:case 177:case 190:case 179:case 258:case 285:case 284:case 207:case 201:case 203:case 204:case 232:case 266:case 242:case 217:case 197:case 196:case 275:case 278:case 274:case 283:case 164:case 169:case 170:case 160:case 174:case 175:case 186:case 205:case 213:case 268:case 264:case 270:case 265:case 163:return!0;case 249:case 288:case 216:if(!e.indentMultiLineObjectLiteralBeginningOnBlankLine&&r&&200===i)return S(r,t);if(216!==n.kind)return!0;break;case 235:case 236:case 238:case 239:case 237:case 234:case 251:case 208:case 165:case 166:case 167:case 168:return 230!==i;case 209:return r&&207===i?S(r,t):230!==i;case 267:return 268!==i;case 261:return 262!==i||!!t.namedBindings&&264!==t.namedBindings.kind;case 273:return 276!==i;case 277:return 279!==i;case 183:case 182:if(177===i||179===i)return!1}return a}function T(e,n,t,r,a){return void 0===a&&(a=!1),E(e,n,t,r,!1)&&!(a&&t&&function(e,n){switch(e){case 242:case 246:case 240:case 241:return 230!==n.kind;default:return!1}}(t.kind,n))}function S(n,t){var r=e.skipTrivia(n.text,t.pos);return n.getLineAndCharacterOfPosition(r).line===n.getLineAndCharacterOfPosition(t.end).line}!function(e){e[e.Unknown=-1]="Unknown"}(r||(r={})),t.getIndentation=function(t,r,a,s){if(void 0===s&&(s=!1),t>r.text.length)return i(a);if(a.indentStyle===e.IndentStyle.None)return 0;var c=e.findPrecedingToken(t,r,void 0,!0),d=n.getRangeOfEnclosingComment(r,t,c||null);if(d&&3===d.kind)return function(n,t,r,a){var i=e.getLineAndCharacterOfPosition(n,t).line-1,o=e.getLineAndCharacterOfPosition(n,a.pos).line;if(e.Debug.assert(o>=0),i<=o)return b(e.getStartPositionOfLine(o,n),t,n,r);var s=e.getStartPositionOfLine(i,n),c=h(s,t,n,r),l=c.column,u=c.character;if(0===l)return l;return 42===n.text.charCodeAt(s+u)?l-1:l}(r,t,a,d);if(!c)return i(a);if(e.isStringOrRegularExpressionOrTemplateLiteral(c.kind)&&c.getStart(r)<=t&&t<c.end)return 0;var p=r.getLineAndCharacterOfPosition(t).line;if(a.indentStyle===e.IndentStyle.Block)return function(n,t,r){var a=t;for(;a>0;){var i=n.text.charCodeAt(a);if(!e.isWhiteSpaceLike(i))break;a--}return b(e.getLineStartPositionForPosition(a,n),a,n,r)}(r,t,a);if(27===c.kind&&216!==c.parent.kind){var m=function(n,t,r){var a=e.findListItemInfo(n);return a&&a.listItemIndex>0?y(a.list.getChildren(),a.listItemIndex-1,t,r):-1}(c,r,a);if(-1!==m)return m}var v=function(e,n,t){return n&&f(e,e,n,t)}(t,c.parent,r);return v&&!e.rangeContainsRange(v,c)?_(v,r,a)+a.indentSize:function(n,t,r,a,s,c){var d,p=r;for(;p;){if(e.positionBelongsToNode(p,t,n)&&T(c,p,d,n,!0)){var m=u(p,n),f=l(r,p,a,n);return o(p,m,void 0,0!==f?s&&2===f?c.indentSize:0:a!==m.line?c.indentSize:0,n,!0,c)}var _=g(p,n,c,!0);if(-1!==_)return _;d=p,p=p.parent}return i(c)}(r,t,c,p,s,a)},t.getIndentationForNode=function(e,n,t,r){var a=t.getLineAndCharacterOfPosition(e.getStart(t));return o(e,a,n,0,t,!1,r)},t.getBaseIndentation=i,function(e){e[e.Unknown=0]="Unknown",e[e.OpenBrace=1]="OpenBrace",e[e.CloseBrace=2]="CloseBrace"}(a||(a={})),t.isArgumentAndStartLineOverlapsExpressionBeingCalled=d,t.childStartsOnTheSameLineWithElseInIfStatement=p,t.argumentStartsOnSameLineAsPreviousArgument=function(n,t,r,a){if(e.isCallOrNewExpression(n)){if(!n.arguments)return!1;var i=e.find(n.arguments,(function(e){return e.pos===t.pos}));if(!i)return!1;var o=n.arguments.indexOf(i);if(0===o)return!1;var s=n.arguments[o-1];if(r===e.getLineAndCharacterOfPosition(a,s.getEnd()).line)return!0}return!1},t.getContainingList=m,t.findFirstNonWhitespaceCharacterAndColumn=h,t.findFirstNonWhitespaceColumn=b,t.nodeWillIndentChild=E,t.shouldIndentChildNode=T}(n.SmartIndenter||(n.SmartIndenter={}))}(e.formatting||(e.formatting={}))}(ts||(ts={})),function(e){!function(n){function t(n){var t=n.__pos;return e.Debug.assert("number"==typeof t),t}function r(n,t){e.Debug.assert("number"==typeof t),n.__pos=t}function a(n){var t=n.__end;return e.Debug.assert("number"==typeof t),t}function i(n,t){e.Debug.assert("number"==typeof t),n.__end=t}var o,s;function c(n,t){return e.skipTrivia(n,t,!1,!0)}!function(e){e[e.Exclude=0]="Exclude",e[e.IncludeAll=1]="IncludeAll",e[e.JSDoc=2]="JSDoc",e[e.StartLine=3]="StartLine"}(o=n.LeadingTriviaOption||(n.LeadingTriviaOption={})),function(e){e[e.Exclude=0]="Exclude",e[e.ExcludeWhitespace=1]="ExcludeWhitespace",e[e.Include=2]="Include"}(s=n.TrailingTriviaOption||(n.TrailingTriviaOption={}));var l,u={leadingTriviaOption:o.Exclude,trailingTriviaOption:s.Exclude};function d(e,n,t,r){return{pos:p(e,n,r),end:m(e,t,r)}}function p(n,t,r){var a=r.leadingTriviaOption;if(a===o.Exclude)return t.getStart(n);if(a===o.StartLine)return e.getLineStartPositionForPosition(t.getStart(n),n);if(a===o.JSDoc){var i=e.getJSDocCommentRanges(t,n.text);if(null==i?void 0:i.length)return e.getLineStartPositionForPosition(i[0].pos,n)}var s=t.getFullStart(),l=t.getStart(n);if(s===l)return l;var u=e.getLineStartPositionForPosition(s,n);if(e.getLineStartPositionForPosition(l,n)===u)return a===o.IncludeAll?s:l;var d=s>0?1:0,p=e.getStartPositionOfLine(e.getLineOfLocalPosition(n,u)+d,n);return p=c(n.text,p),e.getStartPositionOfLine(e.getLineOfLocalPosition(n,p),n)}function m(n,t,r){var a,i=t.end,o=r.trailingTriviaOption;if(o===s.Exclude)return i;if(o===s.ExcludeWhitespace){var c=e.concatenate(e.getTrailingCommentRanges(n.text,i),e.getLeadingCommentRanges(n.text,i)),l=null===(a=null==c?void 0:c[c.length-1])||void 0===a?void 0:a.end;return l||i}var u=e.skipTrivia(n.text,i,!0);return u===i||o!==s.Include&&!e.isLineBreak(n.text.charCodeAt(u-1))?i:u}function f(e,n){return!!n&&!!e.parent&&(27===n.kind||26===n.kind&&200===e.parent.kind)}!function(e){e[e.Remove=0]="Remove",e[e.ReplaceWithSingleNode=1]="ReplaceWithSingleNode",e[e.ReplaceWithMultipleNodes=2]="ReplaceWithMultipleNodes",e[e.Text=3]="Text"}(l||(l={})),n.isThisTypeAnnotatable=function(n){return e.isFunctionExpression(n)||e.isFunctionDeclaration(n)};var _,g,y=function(){function n(n,t){this.newLineCharacter=n,this.formatContext=t,this.changes=[],this.newFiles=[],this.classesWithNodesInsertedAtStart=new e.Map,this.deletedNodes=[]}return n.fromContext=function(t){return new n(e.getNewLineOrDefaultFromHost(t.host,t.formatContext.options),t.formatContext)},n.with=function(e,t){var r=n.fromContext(e);return t(r),r.getChanges()},n.prototype.pushRaw=function(n,t){e.Debug.assertEqual(n.fileName,t.fileName);for(var r=0,a=t.textChanges;r<a.length;r++){var i=a[r];this.changes.push({kind:l.Text,sourceFile:n,text:i.newText,range:e.createTextRangeFromSpan(i.span)})}},n.prototype.deleteRange=function(e,n){this.changes.push({kind:l.Remove,sourceFile:e,range:n})},n.prototype.delete=function(e,n){this.deletedNodes.push({sourceFile:e,node:n})},n.prototype.deleteNode=function(e,n,t){void 0===t&&(t={leadingTriviaOption:o.IncludeAll}),this.deleteRange(e,d(e,n,n,t))},n.prototype.deleteModifier=function(n,t){this.deleteRange(n,{pos:t.getStart(n),end:e.skipTrivia(n.text,t.end,!0)})},n.prototype.deleteNodeRange=function(e,n,t,r){void 0===r&&(r={leadingTriviaOption:o.IncludeAll});var a=p(e,n,r),i=m(e,t,r);this.deleteRange(e,{pos:a,end:i})},n.prototype.deleteNodeRangeExcludingEnd=function(e,n,t,r){void 0===r&&(r={leadingTriviaOption:o.IncludeAll});var a=p(e,n,r),i=void 0===t?e.text.length:p(e,t,r);this.deleteRange(e,{pos:a,end:i})},n.prototype.replaceRange=function(e,n,t,r){void 0===r&&(r={}),this.changes.push({kind:l.ReplaceWithSingleNode,sourceFile:e,range:n,options:r,node:t})},n.prototype.replaceNode=function(e,n,t,r){void 0===r&&(r=u),this.replaceRange(e,d(e,n,n,r),t,r)},n.prototype.replaceNodeRange=function(e,n,t,r,a){void 0===a&&(a=u),this.replaceRange(e,d(e,n,t,a),r,a)},n.prototype.replaceRangeWithNodes=function(e,n,t,r){void 0===r&&(r={}),this.changes.push({kind:l.ReplaceWithMultipleNodes,sourceFile:e,range:n,options:r,nodes:t})},n.prototype.replaceNodeWithNodes=function(e,n,t,r){void 0===r&&(r=u),this.replaceRangeWithNodes(e,d(e,n,n,r),t,r)},n.prototype.replaceNodeWithText=function(e,n,t){this.replaceRangeWithText(e,d(e,n,n,u),t)},n.prototype.replaceNodeRangeWithNodes=function(e,n,t,r,a){void 0===a&&(a=u),this.replaceRangeWithNodes(e,d(e,n,t,a),r,a)},n.prototype.nextCommaToken=function(n,t){var r=e.findNextToken(t,t.parent,n);return r&&27===r.kind?r:void 0},n.prototype.replacePropertyAssignment=function(e,n,t){var r=this.nextCommaToken(e,n)?"":","+this.newLineCharacter;this.replaceNode(e,n,t,{suffix:r})},n.prototype.insertNodeAt=function(n,t,r,a){void 0===a&&(a={}),this.replaceRange(n,e.createRange(t),r,a)},n.prototype.insertNodesAt=function(n,t,r,a){void 0===a&&(a={}),this.replaceRangeWithNodes(n,e.createRange(t),r,a)},n.prototype.insertNodeAtTopOfFile=function(e,n,t){this.insertAtTopOfFile(e,n,t)},n.prototype.insertNodesAtTopOfFile=function(e,n,t){this.insertAtTopOfFile(e,n,t)},n.prototype.insertAtTopOfFile=function(n,t,r){var a=function(n){for(var t,r=0,a=n.statements;r<a.length;r++){var i=a[r];if(!e.isPrologueDirective(i))break;t=i}var o=0,s=n.text;if(t)return o=t.end,_(),o;var c=e.getShebang(s);void 0!==c&&(o=c.length,_());var l,u,d=e.getLeadingCommentRanges(s,o);if(!d)return o;for(var p=0,m=d;p<m.length;p++){var f=m[p];if(3===f.kind){if(e.isPinnedComment(s,f.pos)){l={range:f,pinnedOrTripleSlash:!0};continue}}else if(e.isRecognizedTripleSlashComment(s,f.pos,f.end)){l={range:f,pinnedOrTripleSlash:!0};continue}if(l){if(l.pinnedOrTripleSlash)break;if(n.getLineAndCharacterOfPosition(f.pos).line>=n.getLineAndCharacterOfPosition(l.range.end).line+2)break}if(n.statements.length)if(void 0===u&&(u=n.getLineAndCharacterOfPosition(n.statements[0].getStart()).line),u<n.getLineAndCharacterOfPosition(f.end).line+2)break;l={range:f,pinnedOrTripleSlash:!1}}l&&(o=l.range.end,_());return o;function _(){if(o<s.length){var n=s.charCodeAt(o);e.isLineBreak(n)&&++o<s.length&&13===n&&10===s.charCodeAt(o)&&o++}}}(n),i={prefix:0===a?void 0:this.newLineCharacter,suffix:(e.isLineBreak(n.text.charCodeAt(a))?"":this.newLineCharacter)+(r?this.newLineCharacter:"")};e.isArray(t)?this.insertNodesAt(n,a,t,i):this.insertNodeAt(n,a,t,i)},n.prototype.insertFirstParameter=function(n,t,r){var a=e.firstOrUndefined(t);a?this.insertNodeBefore(n,a,r):this.insertNodeAt(n,t.pos,r)},n.prototype.insertNodeBefore=function(e,n,t,r,a){void 0===r&&(r=!1),void 0===a&&(a={}),this.insertNodeAt(e,p(e,n,a),t,this.getOptionsForInsertNodeBefore(n,t,r))},n.prototype.insertModifierBefore=function(n,t,r){var a=r.getStart(n);this.insertNodeAt(n,a,e.factory.createToken(t),{suffix:" "})},n.prototype.insertLastModifierBefore=function(n,t,r){if(r.modifiers){var a=r.modifiers.end;this.insertNodeAt(n,a,e.factory.createToken(t),{prefix:" "})}else this.insertModifierBefore(n,t,r)},n.prototype.insertCommentBeforeLine=function(n,t,r,a){var i=e.getStartPositionOfLine(t,n),o=e.getFirstNonSpaceCharacterPosition(n.text,i),s=S(n,o),c=e.getTouchingToken(n,s?o:r),l=n.text.slice(i,o),u=(s?"":this.newLineCharacter)+"//"+a+this.newLineCharacter+l;this.insertText(n,c.getStart(n),u)},n.prototype.insertJsdocCommentBefore=function(n,t,r){var a=t.getStart(n);if(t.jsDoc)for(var i=0,o=t.jsDoc;i<o.length;i++){var s=o[i];this.deleteRange(n,{pos:e.getLineStartPositionForPosition(s.getStart(n),n),end:m(n,s,{})})}var c=e.getPrecedingNonSpaceCharacterPosition(n.text,a-1),l=n.text.slice(c,a);this.insertNodeAt(n,a,r,{preserveLeadingWhitespace:!1,suffix:this.newLineCharacter+l})},n.prototype.replaceRangeWithText=function(e,n,t){this.changes.push({kind:l.Text,sourceFile:e,range:n,text:t})},n.prototype.insertText=function(n,t,r){this.replaceRangeWithText(n,e.createRange(t),r)},n.prototype.tryInsertTypeAnnotation=function(n,t,r){var a,i;if(e.isFunctionLike(t)){if(!(i=e.findChildOfKind(t,21,n))){if(!e.isArrowFunction(t))return!1;i=e.first(t.parameters)}}else i=null!==(a=249===t.kind?t.exclamationToken:t.questionToken)&&void 0!==a?a:t.name;return this.insertNodeAt(n,i.end,r,{prefix:": "}),!0},n.prototype.tryInsertThisTypeAnnotation=function(n,t,r){var a=e.findChildOfKind(t,20,n).getStart(n)+1,i=t.parameters.length?", ":"";this.insertNodeAt(n,a,r,{prefix:"this: ",suffix:i})},n.prototype.insertTypeParameters=function(n,t,r){var a=(e.findChildOfKind(t,20,n)||e.first(t.parameters)).getStart(n);this.insertNodesAt(n,a,r,{prefix:"<",suffix:">",joiner:", "})},n.prototype.getOptionsForInsertNodeBefore=function(n,t,r){return e.isStatement(n)||e.isClassElement(n)?{suffix:r?this.newLineCharacter+this.newLineCharacter:this.newLineCharacter}:e.isVariableDeclaration(n)?{suffix:", "}:e.isParameter(n)?e.isParameter(t)?{suffix:", "}:{}:e.isStringLiteral(n)&&e.isImportDeclaration(n.parent)||e.isNamedImports(n)?{suffix:", "}:e.isImportSpecifier(n)?{suffix:","+(r?this.newLineCharacter:" ")}:e.Debug.failBadSyntaxKind(n)},n.prototype.insertNodeAtConstructorStart=function(n,t,r){var a=e.firstOrUndefined(t.body.statements);a&&t.body.multiLine?this.insertNodeBefore(n,a,r):this.replaceConstructorBody(n,t,__spreadArrays([r],t.body.statements))},n.prototype.insertNodeAtConstructorEnd=function(n,t,r){var a=e.lastOrUndefined(t.body.statements);a&&t.body.multiLine?this.insertNodeAfter(n,a,r):this.replaceConstructorBody(n,t,__spreadArrays(t.body.statements,[r]))},n.prototype.replaceConstructorBody=function(n,t,r){this.replaceNode(n,t.body,e.factory.createBlock(r,!0))},n.prototype.insertNodeAtEndOfScope=function(n,t,r){var a=p(n,t.getLastToken(),{});this.insertNodeAt(n,a,r,{prefix:e.isLineBreak(n.text.charCodeAt(t.getLastToken().pos))?this.newLineCharacter:this.newLineCharacter+this.newLineCharacter,suffix:this.newLineCharacter})},n.prototype.insertNodeAtClassStart=function(e,n,t){this.insertNodeAtStartWorker(e,n,t)},n.prototype.insertNodeAtObjectStart=function(e,n,t){this.insertNodeAtStartWorker(e,n,t)},n.prototype.insertNodeAtStartWorker=function(e,n,t){var r,a=null!==(r=this.guessIndentationFromExistingMembers(e,n))&&void 0!==r?r:this.computeIndentationForNewMember(e,n);this.insertNodeAt(e,h(n).pos,t,this.getInsertNodeAtStartInsertOptions(e,n,a))},n.prototype.guessIndentationFromExistingMembers=function(n,t){for(var r,a=t,i=0,o=h(t);i<o.length;i++){var s=o[i];if(e.rangeStartPositionsAreOnSameLine(a,s,n))return;var c=s.getStart(n),l=e.formatting.SmartIndenter.findFirstNonWhitespaceColumn(e.getLineStartPositionForPosition(c,n),c,n,this.formatContext.options);if(void 0===r)r=l;else if(l!==r)return;a=s}return r},n.prototype.computeIndentationForNewMember=function(n,t){var r,a=t.getStart(n);return e.formatting.SmartIndenter.findFirstNonWhitespaceColumn(e.getLineStartPositionForPosition(a,n),a,n,this.formatContext.options)+(null!==(r=this.formatContext.options.indentSize)&&void 0!==r?r:4)},n.prototype.getInsertNodeAtStartInsertOptions=function(n,t,r){var a=0===h(t).length,i=e.addToSeen(this.classesWithNodesInsertedAtStart,e.getNodeId(t),{node:t,sourceFile:n}),o=e.isObjectLiteralExpression(t)&&(!e.isJsonSourceFile(n)||!a);return{indentation:r,prefix:(e.isObjectLiteralExpression(t)&&e.isJsonSourceFile(n)&&a&&!i?",":"")+this.newLineCharacter,suffix:o?",":""}},n.prototype.insertNodeAfterComma=function(e,n,t){var r=this.insertNodeAfterWorker(e,this.nextCommaToken(e,n)||n,t);this.insertNodeAt(e,r,t,this.getInsertNodeAfterOptions(e,n))},n.prototype.insertNodeAfter=function(e,n,t){var r=this.insertNodeAfterWorker(e,n,t);this.insertNodeAt(e,r,t,this.getInsertNodeAfterOptions(e,n))},n.prototype.insertNodeAtEndOfList=function(e,n,t){this.insertNodeAt(e,n.end,t,{prefix:", "})},n.prototype.insertNodesAfter=function(n,t,r){var a=this.insertNodeAfterWorker(n,t,e.first(r));this.insertNodesAt(n,a,r,this.getInsertNodeAfterOptions(n,t))},n.prototype.insertNodeAfterWorker=function(n,t,r){var a,i;return a=t,i=r,((e.isPropertySignature(a)||e.isPropertyDeclaration(a))&&e.isClassOrTypeElement(i)&&158===i.name.kind||e.isStatementButNotDeclaration(a)&&e.isStatementButNotDeclaration(i))&&59!==n.text.charCodeAt(t.end-1)&&this.replaceRange(n,e.createRange(t.end),e.factory.createToken(26)),m(n,t,{})},n.prototype.getInsertNodeAfterOptions=function(n,t){var r=this.getInsertNodeAfterOptionsWorker(t);return __assign(__assign({},r),{prefix:t.end===n.end&&e.isStatement(t)?r.prefix?"\n"+r.prefix:"\n":r.prefix})},n.prototype.getInsertNodeAfterOptionsWorker=function(n){switch(n.kind){case 252:case 256:return{prefix:this.newLineCharacter,suffix:this.newLineCharacter};case 249:case 10:case 78:return{prefix:", "};case 288:return{suffix:","+this.newLineCharacter};case 92:return{prefix:" "};case 160:return{};default:return e.Debug.assert(e.isStatement(n)||e.isClassOrTypeElement(n)),{suffix:this.newLineCharacter}}},n.prototype.insertName=function(n,t,r){if(e.Debug.assert(!t.name),209===t.kind){var a=e.findChildOfKind(t,38,n),i=e.findChildOfKind(t,20,n);i?(this.insertNodesAt(n,i.getStart(n),[e.factory.createToken(97),e.factory.createIdentifier(r)],{joiner:" "}),x(this,n,a)):(this.insertText(n,e.first(t.parameters).getStart(n),"function "+r+"("),this.replaceRange(n,a,e.factory.createToken(21))),230!==t.body.kind&&(this.insertNodesAt(n,t.body.getStart(n),[e.factory.createToken(18),e.factory.createToken(104)],{joiner:" ",suffix:" "}),this.insertNodesAt(n,t.body.end,[e.factory.createToken(26),e.factory.createToken(19)],{joiner:" "}))}else{var o=e.findChildOfKind(t,208===t.kind?97:83,n).end;this.insertNodeAt(n,o,e.factory.createIdentifier(r),{prefix:" "})}},n.prototype.insertExportModifier=function(e,n){this.insertText(e,n.getStart(e),"export ")},n.prototype.insertNodeInListAfter=function(n,t,r,a){if(void 0===a&&(a=e.formatting.SmartIndenter.getContainingList(t,n)),a){var i=e.indexOfNode(a,t);if(!(i<0)){var o=t.getEnd();if(i!==a.length-1){var s=e.getTokenAtPosition(n,t.end);if(s&&f(t,s)){var l=e.getLineAndCharacterOfPosition(n,c(n.text,a[i+1].getFullStart())),u=e.getLineAndCharacterOfPosition(n,s.end),d=void 0,p=void 0;u.line===l.line?(p=s.end,d=function(e){for(var n="",t=0;t<e;t++)n+=" ";return n}(l.character-u.character)):p=e.getStartPositionOfLine(l.line,n);var m=""+e.tokenToString(s.kind)+n.text.substring(s.end,a[i+1].getStart(n));this.replaceRange(n,e.createRange(p,a[i+1].getStart(n)),r,{prefix:d,suffix:m})}}else{var _=t.getStart(n),g=e.getLineStartPositionForPosition(_,n),y=void 0,v=!1;if(1===a.length)y=27;else{var h=e.findPrecedingToken(t.pos,n);y=f(t,h)?h.kind:27,v=e.getLineStartPositionForPosition(a[i-1].getStart(n),n)!==g}if(function(n,t){for(var r=t;r<n.length;){var a=n.charCodeAt(r);if(!e.isWhiteSpaceSingleLine(a))return 47===a;r++}return!1}(n.text,t.end)&&(v=!0),v){this.replaceRange(n,e.createRange(o),e.factory.createToken(y));for(var b=e.formatting.SmartIndenter.findFirstNonWhitespaceColumn(g,_,n,this.formatContext.options),E=e.skipTrivia(n.text,o,!0,!1);E!==o&&e.isLineBreak(n.text.charCodeAt(E-1));)E--;this.replaceRange(n,e.createRange(E),r,{indentation:b,prefix:this.newLineCharacter})}else this.replaceRange(n,e.createRange(o),r,{prefix:e.tokenToString(y)+" "})}}}else e.Debug.fail("node is not a list element")},n.prototype.parenthesizeExpression=function(n,t){this.replaceRange(n,e.rangeOfNode(t),e.factory.createParenthesizedExpression(t))},n.prototype.finishClassesWithNodesInsertedAtStart=function(){var n=this;this.classesWithNodesInsertedAtStart.forEach((function(t){var r=t.node,a=t.sourceFile,i=function(n,t){var r=e.findChildOfKind(n,18,t),a=e.findChildOfKind(n,19,t);return[null==r?void 0:r.end,null==a?void 0:a.end]}(r,a),o=i[0],s=i[1];if(void 0!==o&&void 0!==s){var c=0===h(r).length,l=e.positionsAreOnSameLine(o,s,a);c&&l&&o!==s-1&&n.deleteRange(a,e.createRange(o,s-1)),l&&n.insertText(a,s-1,n.newLineCharacter)}}))},n.prototype.finishDeleteDeclarations=function(){for(var n=this,t=new e.Set,r=function(n,r){a.deletedNodes.some((function(t){return t.sourceFile===n&&e.rangeContainsRangeExclusive(t.node,r)}))||(e.isArray(r)?a.deleteRange(n,e.rangeOfTypeParameters(r)):g.deleteDeclaration(a,t,n,r))},a=this,i=0,o=this.deletedNodes;i<o.length;i++){var s=o[i];r(s.sourceFile,s.node)}t.forEach((function(r){var a=r.getSourceFile(),i=e.formatting.SmartIndenter.getContainingList(r,a);if(r===e.last(i)){var o=e.findLastIndex(i,(function(e){return!t.has(e)}),i.length-2);-1!==o&&n.deleteRange(a,{pos:i[o].end,end:v(a,i[o+1])})}}))},n.prototype.getChanges=function(e){this.finishDeleteDeclarations(),this.finishClassesWithNodesInsertedAtStart();for(var n=_.getTextChangesFromChanges(this.changes,this.newLineCharacter,this.formatContext,e),t=0,r=this.newFiles;t<r.length;t++){var a=r[t],i=a.oldFile,o=a.fileName,s=a.statements;n.push(_.newFileChanges(i,o,s,this.newLineCharacter,this.formatContext))}return n},n.prototype.createNewFile=function(e,n,t){this.newFiles.push({oldFile:e,fileName:n,statements:t})},n}();function v(n,t){return e.skipTrivia(n.text,p(n,t,{leadingTriviaOption:o.IncludeAll}),!1,!0)}function h(n){return e.isObjectLiteralExpression(n)?n.properties:n.members}function b(n,t){for(var r=t.length-1;r>=0;r--){var a=t[r],i=a.span,o=a.newText;n=""+n.substring(0,i.start)+o+n.substring(e.textSpanEnd(i))}return n}function E(n){var r=e.visitEachChild(n,E,e.nullTransformationContext,T,E),i=e.nodeIsSynthesized(r)?r:Object.create(r);return e.setTextRangePosEnd(i,t(n),a(n)),i}function T(n,r,i,o,s){var c=e.visitNodes(n,r,i,o,s);if(!c)return c;var l=c===n?e.factory.createNodeArray(c.slice(0)):c;return e.setTextRangePosEnd(l,t(n),a(n)),l}function S(n,t){return!(e.isInComment(n,t)||e.isInString(n,t)||e.isInTemplateString(n,t)||e.isInJSXText(n,t))}function x(e,n,t,r){void 0===r&&(r={leadingTriviaOption:o.IncludeAll});var a=p(n,t,r),i=m(n,t,r);e.deleteRange(n,{pos:a,end:i})}function L(n,t,r,a){var i=e.Debug.checkDefined(e.formatting.SmartIndenter.getContainingList(a,r)),o=e.indexOfNode(i,a);e.Debug.assert(-1!==o),1!==i.length?(e.Debug.assert(!t.has(a),"Deleting a node twice"),t.add(a),n.deleteRange(r,{pos:v(r,a),end:o===i.length-1?m(r,a,{}):v(r,i[o+1])})):x(n,r,a)}n.ChangeTracker=y,n.getNewFileText=function(e,n,t,r){return _.newFileChangesWorker(void 0,n,e,t,r)},function(n){function t(n,t,r,i,o){var s=r.map((function(e){return 4===e?"":a(e,n,i).text})).join(i),c=e.createSourceFile("any file name",s,99,!0,t);return b(s,e.formatting.formatDocument(c,o))+i}function a(n,t,a){var o=function(n){var t=0,a=e.createTextWriter(n);function o(n,r){if(r||!function(n){return e.skipTrivia(n,0)===n.length}(n)){t=a.getTextPos();for(var i=0;e.isWhiteSpaceLike(n.charCodeAt(n.length-i-1));)i++;t-=i}}function s(e){a.write(e),o(e,!1)}function c(e){a.writeComment(e)}function l(e){a.writeKeyword(e),o(e,!1)}function u(e){a.writeOperator(e),o(e,!1)}function d(e){a.writePunctuation(e),o(e,!1)}function p(e){a.writeTrailingSemicolon(e),o(e,!1)}function m(e){a.writeParameter(e),o(e,!1)}function f(e){a.writeProperty(e),o(e,!1)}function _(e){a.writeSpace(e),o(e,!1)}function g(e){a.writeStringLiteral(e),o(e,!1)}function y(e,n){a.writeSymbol(e,n),o(e,!1)}function v(e){a.writeLine(e)}function h(){a.increaseIndent()}function b(){a.decreaseIndent()}function E(){return a.getText()}function T(e){a.rawWrite(e),o(e,!1)}function S(e){a.writeLiteral(e),o(e,!0)}function x(){return a.getTextPos()}function L(){return a.getLine()}function A(){return a.getColumn()}function D(){return a.getIndent()}function C(){return a.isAtStartOfLine()}function I(){a.clear(),t=0}return{onEmitNode:function(e,n,a){n&&r(n,t),a(e,n),n&&i(n,t)},onBeforeEmitNodeArray:function(e){e&&r(e,t)},onAfterEmitNodeArray:function(e){e&&i(e,t)},onBeforeEmitToken:function(e){e&&r(e,t)},onAfterEmitToken:function(e){e&&i(e,t)},write:s,writeComment:c,writeKeyword:l,writeOperator:u,writePunctuation:d,writeTrailingSemicolon:p,writeParameter:m,writeProperty:f,writeSpace:_,writeStringLiteral:g,writeSymbol:y,writeLine:v,increaseIndent:h,decreaseIndent:b,getText:E,rawWrite:T,writeLiteral:S,getTextPos:x,getLine:L,getColumn:A,getIndent:D,isAtStartOfLine:C,hasTrailingComment:function(){return a.hasTrailingComment()},hasTrailingWhitespace:function(){return a.hasTrailingWhitespace()},clear:I}}(a),s="\n"===a?1:0;return e.createPrinter({newLine:s,neverAsciiEscape:!0,preserveSourceNewlines:!0,terminateUnterminatedLiterals:!0},o).writeNode(4,n,t,o),{text:o.getText(),node:E(n)}}n.getTextChangesFromChanges=function(n,t,r,i){return e.mapDefined(e.group(n,(function(e){return e.sourceFile.path})),(function(n){for(var o=n[0].sourceFile,s=e.stableSort(n,(function(e,n){return e.range.pos-n.range.pos||e.range.end-n.range.end})),c=function(n){e.Debug.assert(s[n].range.end<=s[n+1].range.pos,"Changes overlap",(function(){return JSON.stringify(s[n].range)+" and "+JSON.stringify(s[n+1].range)}))},u=0;u<s.length-1;u++)c(u);var d=e.mapDefined(s,(function(n){var s=e.createTextSpanFromRange(n.range),c=function(n,t,r,i,o){if(n.kind===l.Remove)return"";if(n.kind===l.Text)return n.text;var s=n.options,c=void 0===s?{}:s,u=n.range.pos,d=function(n){return function(n,t,r,i,o,s,c){var l=i.indentation,u=i.prefix,d=i.delta,p=a(n,t,o),m=p.node,f=p.text;c&&c(m,f);var _=function(n,t){var r=n.options,a=!r.semicolons||r.semicolons===e.SemicolonPreference.Ignore,i=r.semicolons===e.SemicolonPreference.Remove||a&&!e.probablyUsesSemicolons(t);return __assign(__assign({},r),{semicolons:i?e.SemicolonPreference.Remove:e.SemicolonPreference.Ignore})}(s,t),g=void 0!==l?l:e.formatting.SmartIndenter.getIndentation(r,t,_,u===o||e.getLineStartPositionForPosition(r,t)===r);void 0===d&&(d=e.formatting.SmartIndenter.shouldIndentChildNode(_,n)&&_.indentSize||0);var y={text:f,getLineAndCharacterOfPosition:function(n){return e.getLineAndCharacterOfPosition(this,n)}},v=e.formatting.formatNodeGivenIndentation(m,y,t.languageVariant,g,d,__assign(__assign({},s),{options:_}));return b(f,v)}(n,t,u,c,r,i,o)},p=n.kind===l.ReplaceWithMultipleNodes?n.nodes.map((function(n){return e.removeSuffix(d(n),r)})).join(n.options.joiner||r):d(n.node),m=c.preserveLeadingWhitespace||void 0!==c.indentation||e.getLineStartPositionForPosition(u,t)===u?p:p.replace(/^\s+/,"");return(c.prefix||"")+m+(!c.suffix||e.endsWith(m,c.suffix)?"":c.suffix)}(n,o,t,r,i);if(s.length!==c.length||!e.stringContainsAt(o.text,c,s.start))return e.createTextChange(s,c)}));return d.length>0?{fileName:o.fileName,textChanges:d}:void 0}))},n.newFileChanges=function(n,r,a,i,o){var s=t(n,e.getScriptKindFromFileName(r),a,i,o);return{fileName:r,textChanges:[e.createTextChange(e.createTextSpan(0,0),s)],isNewFile:!0}},n.newFileChangesWorker=t,n.getNonformattedText=a}(_||(_={})),n.applyChanges=b,n.isValidLocationToAddComment=S,function(n){function t(n,t,r){if(r.parent.name){var a=e.Debug.checkDefined(e.getTokenAtPosition(t,r.pos-1));n.deleteRange(t,{pos:a.getStart(t),end:r.end})}else{x(n,t,e.getAncestor(r,261))}}n.deleteDeclaration=function(n,r,a,i){switch(i.kind){case 160:var c=i.parent;e.isArrowFunction(c)&&1===c.parameters.length&&!e.findChildOfKind(c,20,a)?n.replaceNodeWithText(a,i,"()"):L(n,r,a,i);break;case 261:case 260:x(n,a,i,{leadingTriviaOption:a.imports.length&&i===e.first(a.imports).parent||i===e.find(a.statements,e.isAnyImportSyntax)?o.Exclude:e.hasJSDocNodes(i)?o.JSDoc:o.StartLine});break;case 198:var l=i.parent;197===l.kind&&i!==e.last(l.elements)?x(n,a,i):L(n,r,a,i);break;case 249:!function(n,t,r,a){var i=a.parent;if(287===i.kind)return void n.deleteNodeRange(r,e.findChildOfKind(i,20,r),e.findChildOfKind(i,21,r));if(1!==i.declarations.length)return void L(n,t,r,a);var s=i.parent;switch(s.kind){case 239:case 238:n.replaceNode(r,a,e.factory.createObjectLiteralExpression());break;case 237:x(n,r,i);break;case 232:x(n,r,s,{leadingTriviaOption:e.hasJSDocNodes(s)?o.JSDoc:o.StartLine});break;default:e.Debug.assertNever(s)}}(n,r,a,i);break;case 159:L(n,r,a,i);break;case 265:var u=i.parent;1===u.elements.length?t(n,a,u):L(n,r,a,i);break;case 263:t(n,a,i);break;case 26:x(n,a,i,{trailingTriviaOption:s.Exclude});break;case 97:x(n,a,i,{leadingTriviaOption:o.Exclude});break;case 252:case 251:x(n,a,i,{leadingTriviaOption:e.hasJSDocNodes(i)?o.JSDoc:o.StartLine});break;default:e.isImportClause(i.parent)&&i.parent.name===i?function(n,t,r){if(r.namedBindings){var a=r.name.getStart(t),i=e.getTokenAtPosition(t,r.name.end);if(i&&27===i.kind){var o=e.skipTrivia(t.text,i.end,!1,!0);n.deleteRange(t,{pos:a,end:o})}else x(n,t,r.name)}else x(n,t,r.parent)}(n,a,i.parent):e.isCallExpression(i.parent)&&e.contains(i.parent.arguments,i)?L(n,r,a,i):x(n,a,i)}}}(g||(g={})),n.deleteNode=x}(e.textChanges||(e.textChanges={}))}(ts||(ts={})),function(e){!function(n){var t=e.createMultiMap(),r=new e.Map;function a(n){return e.isArray(n)?e.formatStringFromArgs(e.getLocaleSpecificMessage(n[0]),n.slice(1)):e.getLocaleSpecificMessage(n)}function i(e,n,t,r,a,i){return{fixName:e,description:n,changes:t,fixId:r,fixAllDescription:a,commands:i?[i]:void 0}}function o(e,n){return{changes:e,commands:n}}function s(n,t,r){for(var a=0,i=c(n);a<i.length;a++){var o=i[a];e.contains(t,o.code)&&r(o)}}function c(n){var t=n.program,r=n.sourceFile,a=n.cancellationToken;return __spreadArrays(t.getSemanticDiagnostics(r,a),t.getSyntacticDiagnostics(r,a),e.computeSuggestionDiagnostics(r,t,a))}n.createCodeFixActionWithoutFixAll=function(e,n,t){return i(e,a(t),n,void 0,void 0)},n.createCodeFixAction=function(e,n,t,r,o,s){return i(e,a(t),n,r,a(o),s)},n.registerCodeFix=function(n){for(var a=0,i=n.errorCodes;a<i.length;a++){var o=i[a];t.add(String(o),n)}if(n.fixIds)for(var s=0,c=n.fixIds;s<c.length;s++){var l=c[s];e.Debug.assert(!r.has(l)),r.set(l,n)}},n.getSupportedErrorCodes=function(){return e.arrayFrom(t.keys())},n.getFixes=function(n){var r=c(n),a=t.get(String(n.errorCode));return e.flatMap(a,(function(t){return e.map(t.getCodeActions(n),function(n,t){for(var r=n.errorCodes,a=0,i=0,o=t;i<o.length;i++){var s=o[i];if(e.contains(r,s.code)&&a++,a>1)break}var c=a<2;return function(e){var n=e.fixId,t=e.fixAllDescription,r=__rest(e,["fixId","fixAllDescription"]);return c?r:__assign(__assign({},r),{fixId:n,fixAllDescription:t})}}(t,r))}))},n.getAllFixes=function(n){return r.get(e.cast(n.fixId,e.isString)).getAllCodeActions(n)},n.createCombinedCodeActions=o,n.createFileTextChanges=function(e,n){return{fileName:e,textChanges:n}},n.codeFixAll=function(n,t,r){var a=[];return o(e.textChanges.ChangeTracker.with(n,(function(e){return s(n,t,(function(n){return r(e,n,a)}))})),0===a.length?void 0:a)},n.eachDiagnostic=s}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){var n,t;n=e.refactor||(e.refactor={}),t=new e.Map,n.registerRefactor=function(e,n){t.set(e,n)},n.getApplicableRefactors=function(n){return e.arrayFrom(e.flatMapIterator(t.values(),(function(e){return n.cancellationToken&&n.cancellationToken.isCancellationRequested()?void 0:e.getAvailableActions(n)})))},n.getEditsForRefactor=function(e,n,r){var a=t.get(n);return a&&a.getEditsForAction(e,r)}}(ts||(ts={})),function(e){!function(n){var t="addConvertToUnknownForNonOverlappingTypes",r=[e.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.Debug.checkDefined(e.findAncestor(a,(function(n){return e.isAsExpression(n)||e.isTypeAssertionExpression(n)})),"Expected to find an assertion expression"),o=e.isAsExpression(i)?e.factory.createAsExpression(i.expression,e.factory.createKeywordTypeNode(152)):e.factory.createTypeAssertion(e.factory.createKeywordTypeNode(152),i.expression);n.replaceNode(t,i.expression,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start)}));return[n.createCodeFixAction(t,i,e.Diagnostics.Add_unknown_conversion_for_non_overlapping_types,t,e.Diagnostics.Add_unknown_to_all_conversions_of_non_overlapping_types)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n.start)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){n.registerCodeFix({errorCodes:[e.Diagnostics.await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module.code],getCodeActions:function(t){var r=t.sourceFile,a=e.textChanges.ChangeTracker.with(t,(function(n){var t=e.factory.createExportDeclaration(void 0,void 0,!1,e.factory.createNamedExports([]),void 0);n.insertNodeAtEndOfScope(r,r,t)}));return[n.createCodeFixActionWithoutFixAll("addEmptyExportDeclaration",a,e.Diagnostics.Add_export_to_make_this_file_into_a_module)]}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingAsync",r=[e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1.code,e.Diagnostics.Type_0_is_not_comparable_to_type_1.code];function a(r,a,i,o){var s=i((function(n){return function(n,t,r,a){if(a&&a.has(e.getNodeId(r)))return;null==a||a.add(e.getNodeId(r));var i=e.factory.updateModifiers(e.getSynthesizedDeepClone(r,!0),e.factory.createNodeArray(e.factory.createModifiersFromModifierFlags(256|e.getSyntacticModifierFlags(r))));n.replaceNode(t,r,i)}(n,r.sourceFile,a,o)}));return n.createCodeFixAction(t,s,e.Diagnostics.Add_async_modifier_to_containing_function,t,e.Diagnostics.Add_all_missing_async_modifiers)}function i(n,t){if(t){var r=e.getTokenAtPosition(n,t.start);return e.findAncestor(r,(function(r){return r.getStart(n)<t.start||r.getEnd()>e.textSpanEnd(t)?"quit":(e.isArrowFunction(r)||e.isMethodDeclaration(r)||e.isFunctionExpression(r)||e.isFunctionDeclaration(r))&&e.textSpansEqual(t,e.createTextSpanFromNode(r,n))}))}}n.registerCodeFix({fixIds:[t],errorCodes:r,getCodeActions:function(n){var t=n.sourceFile,r=n.errorCode,o=n.cancellationToken,s=n.program,c=n.span,l=e.find(s.getDiagnosticsProducingTypeChecker().getDiagnostics(t,o),function(n,t){return function(r){var a=r.start,i=r.length,o=r.relatedInformation,s=r.code;return e.isNumber(a)&&e.isNumber(i)&&e.textSpansEqual({start:a,length:i},n)&&s===t&&!!o&&e.some(o,(function(n){return n.code===e.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code}))}}(c,r)),u=i(t,l&&l.relatedInformation&&e.find(l.relatedInformation,(function(n){return n.code===e.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code})));if(u){return[a(n,u,(function(t){return e.textChanges.ChangeTracker.with(n,t)}))]}},getAllCodeActions:function(t){var o=t.sourceFile,s=new e.Set;return n.codeFixAll(t,r,(function(n,r){var c=r.relatedInformation&&e.find(r.relatedInformation,(function(n){return n.code===e.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code})),l=i(o,c);if(l){return a(t,l,(function(e){return e(n),[]}),s)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingAwait",r=e.Diagnostics.Property_0_does_not_exist_on_type_1.code,a=[e.Diagnostics.This_expression_is_not_callable.code,e.Diagnostics.This_expression_is_not_constructable.code],i=__spreadArrays([e.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type.code,e.Diagnostics.The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type.code,e.Diagnostics.The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type.code,e.Diagnostics.Operator_0_cannot_be_applied_to_type_1.code,e.Diagnostics.Operator_0_cannot_be_applied_to_types_1_and_2.code,e.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap.code,e.Diagnostics.Type_0_is_not_an_array_type.code,e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type.code,e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators.code,e.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator.code,e.Diagnostics.Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator.code,e.Diagnostics.Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator.code,e.Diagnostics.Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator.code,e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,r],a);function o(t,r,a,o,s,c){var d=t.sourceFile,p=t.program,m=t.cancellationToken,f=function(n,t,r,a,o){var s=function(n,t){if(e.isPropertyAccessExpression(n.parent)&&e.isIdentifier(n.parent.expression))return{identifiers:[n.parent.expression],isCompleteFix:!0};if(e.isIdentifier(n))return{identifiers:[n],isCompleteFix:!0};if(e.isBinaryExpression(n)){for(var r=void 0,a=!0,i=0,o=[n.left,n.right];i<o.length;i++){var s=o[i],c=t.getTypeAtLocation(s);if(t.getPromisedTypeOfPromise(c)){if(!e.isIdentifier(s)){a=!1;continue}(r||(r=[])).push(s)}}return r&&{identifiers:r,isCompleteFix:a}}}(n,o);if(!s)return;for(var c,u=s.isCompleteFix,d=function(n){var s=o.getSymbolAtLocation(n);if(!s)return"continue";var d=e.tryCast(s.valueDeclaration,e.isVariableDeclaration),p=d&&e.tryCast(d.name,e.isIdentifier),m=e.getAncestor(d,232);if(!d||!m||d.type||!d.initializer||m.getSourceFile()!==t||e.hasSyntacticModifier(m,1)||!p||!l(d.initializer))return u=!1,"continue";var f=a.getSemanticDiagnostics(t,r);if(e.FindAllReferences.Core.eachSymbolReferenceInFile(p,o,t,(function(r){return n!==r&&!function(n,t,r,a){var o=e.isPropertyAccessExpression(n.parent)?n.parent.name:e.isBinaryExpression(n.parent)?n.parent:n,s=e.find(t,(function(e){return e.start===o.getStart(r)&&e.start+e.length===o.getEnd()}));return s&&e.contains(i,s.code)||1&a.getTypeAtLocation(o).flags}(r,f,t,o)})))return u=!1,"continue";(c||(c=[])).push({expression:d.initializer,declarationSymbol:s})},p=0,m=s.identifiers;p<m.length;p++){d(m[p])}return c&&{initializers:c,needsSecondPassForFixAll:!u}}(r,d,m,p,o);if(f){var _=s((function(n){e.forEach(f.initializers,(function(e){var t=e.expression;return u(n,a,d,o,t,c)})),c&&f.needsSecondPassForFixAll&&u(n,a,d,o,r,c)}));return n.createCodeFixActionWithoutFixAll("addMissingAwaitToInitializer",_,1===f.initializers.length?[e.Diagnostics.Add_await_to_initializer_for_0,f.initializers[0].declarationSymbol.name]:e.Diagnostics.Add_await_to_initializers)}}function s(r,a,i,o,s,c){var l=s((function(e){return u(e,i,r.sourceFile,o,a,c)}));return n.createCodeFixAction(t,l,e.Diagnostics.Add_await,t,e.Diagnostics.Fix_all_expressions_possibly_missing_await)}function c(n,t,r,a,i){var o=e.getTokenAtPosition(n,r.start),s=e.findAncestor(o,(function(t){return t.getStart(n)<r.start||t.getEnd()>e.textSpanEnd(r)?"quit":e.isExpression(t)&&e.textSpansEqual(r,e.createTextSpanFromNode(t,n))}));return s&&function(n,t,r,a,i){var o=i.getDiagnosticsProducingTypeChecker().getDiagnostics(n,a);return e.some(o,(function(n){var a=n.start,i=n.length,o=n.relatedInformation,s=n.code;return e.isNumber(a)&&e.isNumber(i)&&e.textSpansEqual({start:a,length:i},r)&&s===t&&!!o&&e.some(o,(function(n){return n.code===e.Diagnostics.Did_you_forget_to_use_await.code}))}))}(n,t,r,a,i)&&l(s)?s:void 0}function l(n){return 32768&n.kind||!!e.findAncestor(n,(function(n){return n.parent&&e.isArrowFunction(n.parent)&&n.parent.body===n||e.isBlock(n)&&(251===n.parent.kind||208===n.parent.kind||209===n.parent.kind||165===n.parent.kind)}))}function u(n,t,i,o,s,c){if(e.isBinaryExpression(s))for(var l=0,u=[s.left,s.right];l<u.length;l++){var p=u[l];if(c&&e.isIdentifier(p))if((_=o.getSymbolAtLocation(p))&&c.has(e.getSymbolId(_)))continue;var m=o.getTypeAtLocation(p),f=o.getPromisedTypeOfPromise(m)?e.factory.createAwaitExpression(p):p;n.replaceNode(i,p,f)}else if(t===r&&e.isPropertyAccessExpression(s.parent)){if(c&&e.isIdentifier(s.parent.expression))if((_=o.getSymbolAtLocation(s.parent.expression))&&c.has(e.getSymbolId(_)))return;n.replaceNode(i,s.parent.expression,e.factory.createParenthesizedExpression(e.factory.createAwaitExpression(s.parent.expression))),d(n,s.parent.expression,i)}else if(e.contains(a,t)&&e.isCallOrNewExpression(s.parent)){if(c&&e.isIdentifier(s))if((_=o.getSymbolAtLocation(s))&&c.has(e.getSymbolId(_)))return;n.replaceNode(i,s,e.factory.createParenthesizedExpression(e.factory.createAwaitExpression(s))),d(n,s,i)}else{var _;if(c&&e.isVariableDeclaration(s.parent)&&e.isIdentifier(s.parent.name))if((_=o.getSymbolAtLocation(s.parent.name))&&!e.tryAddToSet(c,e.getSymbolId(_)))return;n.replaceNode(i,s,e.factory.createAwaitExpression(s))}}function d(n,t,r){var a=e.findPrecedingToken(t.pos,r);a&&e.positionIsASICandidate(a.end,a.parent,r)&&n.insertText(r,t.getStart(r),";")}n.registerCodeFix({fixIds:[t],errorCodes:i,getCodeActions:function(n){var t=n.sourceFile,r=n.errorCode,a=c(t,r,n.span,n.cancellationToken,n.program);if(a){var i=n.program.getTypeChecker(),l=function(t){return e.textChanges.ChangeTracker.with(n,t)};return e.compact([o(n,a,r,i,l),s(n,a,r,i,l)])}},getAllCodeActions:function(t){var r=t.sourceFile,a=t.program,l=t.cancellationToken,u=t.program.getTypeChecker(),d=new e.Set;return n.codeFixAll(t,i,(function(e,n){var i=c(r,n.code,n,l,a);if(i){var p=function(n){return n(e),[]};return o(t,i,n.code,u,p,d)||s(t,i,n.code,u,p,d)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingConst",r=[e.Diagnostics.Cannot_find_name_0.code,e.Diagnostics.No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer.code];function a(n,t,r,a,s){var c=e.getTokenAtPosition(t,r),l=e.findAncestor(c,(function(n){return e.isForInOrOfStatement(n.parent)?n.parent.initializer===n:!function(e){switch(e.kind){case 78:case 199:case 200:case 288:case 289:return!0;default:return!1}}(n)&&"quit"}));if(l)return i(n,l,t,s);var u=c.parent;if(e.isBinaryExpression(u)&&62===u.operatorToken.kind&&e.isExpressionStatement(u.parent))return i(n,c,t,s);if(e.isArrayLiteralExpression(u)){var d=a.getTypeChecker();if(!e.every(u.elements,(function(n){return function(n,t){var r=e.isIdentifier(n)?n:e.isAssignmentExpression(n,!0)&&e.isIdentifier(n.left)?n.left:void 0;return!!r&&!t.getSymbolAtLocation(r)}(n,d)})))return;return i(n,u,t,s)}var p=e.findAncestor(c,(function(n){return!!e.isExpressionStatement(n.parent)||!function(e){switch(e.kind){case 78:case 216:case 27:return!0;default:return!1}}(n)&&"quit"}));if(p){if(!o(p,a.getTypeChecker()))return;return i(n,p,t,s)}}function i(n,t,r,a){a&&!e.tryAddToSet(a,t)||n.insertModifierBefore(r,84,t)}function o(n,t){return!!e.isBinaryExpression(n)&&(27===n.operatorToken.kind?e.every([n.left,n.right],(function(e){return o(e,t)})):62===n.operatorToken.kind&&e.isIdentifier(n.left)&&!t.getSymbolAtLocation(n.left))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start,r.program)}));if(i.length>0)return[n.createCodeFixAction(t,i,e.Diagnostics.Add_const_to_unresolved_variable,t,e.Diagnostics.Add_const_to_all_unresolved_variables)]},fixIds:[t],getAllCodeActions:function(t){var i=new e.Set;return n.codeFixAll(t,r,(function(e,n){return a(e,n.file,n.start,t.program,i)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingDeclareProperty",r=[e.Diagnostics.Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration.code];function a(n,t,r,a){var i=e.getTokenAtPosition(t,r);if(e.isIdentifier(i)){var o=i.parent;163!==o.kind||a&&!e.tryAddToSet(a,o)||n.insertModifierBefore(t,133,o)}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start)}));if(i.length>0)return[n.createCodeFixAction(t,i,e.Diagnostics.Prefix_with_declare,t,e.Diagnostics.Prefix_all_incorrect_property_declarations_with_declare)]},fixIds:[t],getAllCodeActions:function(t){var i=new e.Set;return n.codeFixAll(t,r,(function(e,n){return a(e,n.file,n.start,i)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingInvocationForDecorator",r=[e.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.findAncestor(a,e.isDecorator);e.Debug.assert(!!i,"Expected position to be owned by a decorator.");var o=e.factory.createCallExpression(i.expression,void 0,void 0);n.replaceNode(t,i.expression,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start)}));return[n.createCodeFixAction(t,i,e.Diagnostics.Call_decorator_expression,t,e.Diagnostics.Add_to_all_uncalled_decorators)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n.start)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addNameToNamelessParameter",r=[e.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r);if(!e.isIdentifier(a))return e.Debug.fail("add-name-to-nameless-parameter operates on identifiers, but got a "+e.Debug.formatSyntaxKind(a.kind));var i=a.parent;if(!e.isParameter(i))return e.Debug.fail("Tried to add a parameter name to a non-parameter: "+e.Debug.formatSyntaxKind(a.kind));var o=i.parent.parameters.indexOf(i);e.Debug.assert(!i.type,"Tried to add a parameter name to a parameter that already had one."),e.Debug.assert(o>-1,"Parameter not found in parent parameter list.");var s=e.factory.createParameterDeclaration(void 0,i.modifiers,i.dotDotDotToken,"arg"+o,i.questionToken,e.factory.createTypeReferenceNode(a,void 0),i.initializer);n.replaceNode(t,a,s)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start)}));return[n.createCodeFixAction(t,i,e.Diagnostics.Add_parameter_name,t,e.Diagnostics.Add_names_to_all_parameters_without_names)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n.start)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="annotateWithTypeFromJSDoc",r=[e.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types.code];function a(n,t){var r=e.getTokenAtPosition(n,t);return e.tryCast(e.isParameter(r.parent)?r.parent.parent:r.parent,i)}function i(n){return function(n){return e.isFunctionLikeDeclaration(n)||249===n.kind||162===n.kind||163===n.kind}(n)&&o(n)}function o(n){return e.isFunctionLikeDeclaration(n)?n.parameters.some(o)||!n.type&&!!e.getJSDocReturnType(n):!n.type&&!!e.getJSDocType(n)}function s(n,t,r){if(e.isFunctionLikeDeclaration(r)&&(e.getJSDocReturnType(r)||r.parameters.some((function(n){return!!e.getJSDocType(n)})))){if(!r.typeParameters){var a=e.getJSDocTypeParameterDeclarations(r);a.length&&n.insertTypeParameters(t,r,a)}var i=e.isArrowFunction(r)&&!e.findChildOfKind(r,20,t);i&&n.insertNodeBefore(t,e.first(r.parameters),e.factory.createToken(20));for(var o=0,s=r.parameters;o<s.length;o++){var l=s[o];if(!l.type){var u=e.getJSDocType(l);u&&n.tryInsertTypeAnnotation(t,l,c(u))}}if(i&&n.insertNodeAfter(t,e.last(r.parameters),e.factory.createToken(21)),!r.type){var d=e.getJSDocReturnType(r);d&&n.tryInsertTypeAnnotation(t,r,c(d))}}else{var p=e.Debug.checkDefined(e.getJSDocType(r),"A JSDocType for this declaration should exist");e.Debug.assert(!r.type,"The JSDocType decl should have a type"),n.tryInsertTypeAnnotation(t,r,c(p))}}function c(n){switch(n.kind){case 303:case 304:return e.factory.createTypeReferenceNode("any",e.emptyArray);case 307:return function(n){return e.factory.createUnionTypeNode([e.visitNode(n.type,c),e.factory.createTypeReferenceNode("undefined",e.emptyArray)])}(n);case 306:return c(n.type);case 305:return function(n){return e.factory.createUnionTypeNode([e.visitNode(n.type,c),e.factory.createTypeReferenceNode("null",e.emptyArray)])}(n);case 309:return function(n){return e.factory.createArrayTypeNode(e.visitNode(n.type,c))}(n);case 308:return function(n){var t;return e.factory.createFunctionTypeNode(e.emptyArray,n.parameters.map(l),null!==(t=n.type)&&void 0!==t?t:e.factory.createKeywordTypeNode(128))}(n);case 173:return function(n){var t=n.typeName,r=n.typeArguments;if(e.isIdentifier(n.typeName)){if(e.isJSDocIndexSignature(n))return function(n){var t=e.factory.createParameterDeclaration(void 0,void 0,void 0,144===n.typeArguments[0].kind?"n":"s",void 0,e.factory.createTypeReferenceNode(144===n.typeArguments[0].kind?"number":"string",[]),void 0),r=e.factory.createTypeLiteralNode([e.factory.createIndexSignature(void 0,void 0,[t],n.typeArguments[1])]);return e.setEmitFlags(r,1),r}(n);var a=n.typeName.text;switch(n.typeName.text){case"String":case"Boolean":case"Object":case"Number":a=a.toLowerCase();break;case"array":case"date":case"promise":a=a[0].toUpperCase()+a.slice(1)}t=e.factory.createIdentifier(a),r="Array"!==a&&"Promise"!==a||n.typeArguments?e.visitNodes(n.typeArguments,c):e.factory.createNodeArray([e.factory.createTypeReferenceNode("any",e.emptyArray)])}return e.factory.createTypeReferenceNode(t,r)}(n);default:var t=e.visitEachChild(n,c,e.nullTransformationContext);return e.setEmitFlags(t,1),t}}function l(n){var t=n.parent.parameters.indexOf(n),r=309===n.type.kind&&t===n.parent.parameters.length-1,a=n.name||(r?"rest":"arg"+t),i=r?e.factory.createToken(25):n.dotDotDotToken;return e.factory.createParameterDeclaration(n.decorators,n.modifiers,i,a,n.questionToken,e.visitNode(n.type,c),n.initializer)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=a(r.sourceFile,r.span.start);if(i){var o=e.textChanges.ChangeTracker.with(r,(function(e){return s(e,r.sourceFile,i)}));return[n.createCodeFixAction(t,o,e.Diagnostics.Annotate_with_type_from_JSDoc,t,e.Diagnostics.Annotate_everything_with_types_from_JSDoc)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=a(n.file,n.start);t&&s(e,n.file,t)}))}}),n.parameterShouldGetTypeFromJSDoc=i}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="convertFunctionToEs6Class",r=[e.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration.code];function a(n,t,r,a){var s=a.getSymbolAtLocation(e.getTokenAtPosition(t,r));if(s&&19&s.flags){var c=s.valueDeclaration;if(e.isFunctionDeclaration(c))n.replaceNode(t,c,function(n){var t=d(s);n.body&&t.unshift(e.factory.createConstructorDeclaration(void 0,void 0,n.parameters,n.body));var r=i(n,92);return e.factory.createClassDeclaration(void 0,r,n.name,void 0,void 0,t)}(c));else if(e.isVariableDeclaration(c)){var l=function(n){var t=n.initializer;if(!t||!e.isFunctionExpression(t)||!e.isIdentifier(n.name))return;var r=d(n.symbol);t.body&&r.unshift(e.factory.createConstructorDeclaration(void 0,void 0,t.parameters,t.body));var a=i(n.parent.parent,92);return e.factory.createClassDeclaration(void 0,a,n.name,void 0,void 0,r)}(c);if(!l)return;var u=c.parent.parent;e.isVariableDeclarationList(c.parent)&&c.parent.declarations.length>1?(n.delete(t,c),n.insertNodeAfter(t,u,l)):n.replaceNode(t,u,l)}}function d(r){var a=[];return r.members&&r.members.forEach((function(e,r){if("constructor"!==r){var i=s(e,void 0);i&&a.push.apply(a,i)}else n.delete(t,e.valueDeclaration.parent)})),r.exports&&r.exports.forEach((function(n){if("prototype"===n.name){var t=n.declarations[0];if(1===n.declarations.length&&e.isPropertyAccessExpression(t)&&e.isBinaryExpression(t.parent)&&62===t.parent.operatorToken.kind&&e.isObjectLiteralExpression(t.parent.right))(r=s(t.parent.right.symbol,void 0))&&a.push.apply(a,r)}else{var r;(r=s(n,[e.factory.createToken(123)]))&&a.push.apply(a,r)}})),a;function s(r,a){var s=[];if(!(8192&r.flags||4096&r.flags))return s;var c,l,u=r.valueDeclaration,d=u.parent,p=d.right;if(c=u,l=p,!(e.isPropertyAccessExpression(c)?o(c)||e.isFunctionLike(l):e.every(c.properties,(function(n){return!!(e.isMethodDeclaration(n)||e.isGetOrSetAccessorDeclaration(n)||e.isPropertyAssignment(n)&&e.isFunctionExpression(n.initializer)&&n.name||o(n))}))))return s;var m=d.parent&&233===d.parent.kind?d.parent:d;if(n.delete(t,m),!p)return s.push(e.factory.createPropertyDeclaration([],a,r.name,void 0,void 0,void 0)),s;if(e.isPropertyAccessExpression(u)&&(e.isFunctionExpression(p)||e.isArrowFunction(p)))return _(s,p,u.name);if(e.isObjectLiteralExpression(p))return e.flatMap(p.properties,(function(n){return e.isMethodDeclaration(n)||e.isGetOrSetAccessorDeclaration(n)?s.concat(n):e.isPropertyAssignment(n)&&e.isFunctionExpression(n.initializer)?_(s,n.initializer,n.name):o(n)?s:[]}));if(e.isSourceFileJS(t))return s;if(!e.isPropertyAccessExpression(u))return s;var f=e.factory.createPropertyDeclaration(void 0,a,u.name,void 0,void 0,p);return e.copyLeadingComments(d.parent,f,t),s.push(f),s;function _(n,r,o){return e.isFunctionExpression(r)?function(n,r,o){var s=e.concatenate(a,i(r,129)),c=e.factory.createMethodDeclaration(void 0,s,void 0,o,void 0,void 0,r.parameters,void 0,r.body);return e.copyLeadingComments(d,c,t),n.concat(c)}(n,r,o):function(n,r,o){var s,c=r.body;s=230===c.kind?c:e.factory.createBlock([e.factory.createReturnStatement(c)]);var l=e.concatenate(a,i(r,129)),u=e.factory.createMethodDeclaration(void 0,l,void 0,o,void 0,void 0,r.parameters,void 0,s);return e.copyLeadingComments(d,u,t),n.concat(u)}(n,r,o)}}}}function i(n,t){return e.filter(n.modifiers,(function(e){return e.kind===t}))}function o(n){return!!n.name&&!(!e.isIdentifier(n.name)||"constructor"!==n.name.text)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start,r.program.getTypeChecker())}));return[n.createCodeFixAction(t,i,e.Diagnostics.Convert_function_to_an_ES2015_class,t,e.Diagnostics.Convert_all_constructor_functions_to_classes)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){return a(n,t.file,t.start,e.program.getTypeChecker())}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t,r="convertToAsyncFunction",a=[e.Diagnostics.This_may_be_converted_to_an_async_function.code],i=!0;function o(n,t,r,a,i){var o,u=e.getTokenAtPosition(t,r);if(o=e.isIdentifier(u)&&e.isVariableDeclaration(u.parent)&&u.parent.initializer&&e.isFunctionLikeDeclaration(u.parent.initializer)?u.parent.initializer:e.tryCast(e.getContainingFunction(e.getTokenAtPosition(t,r)),e.isFunctionLikeDeclaration)){var p,m,f=new e.Map,_=e.isInJSFile(o),y=function(n,t){if(!n.body)return new e.Set;var r=new e.Set;return e.forEachChild(n.body,(function n(a){s(a,t,"then")?(r.add(e.getNodeId(a)),e.forEach(a.arguments,n)):s(a,t,"catch")?(r.add(e.getNodeId(a)),e.forEachChild(a,n)):c(a,t)?r.add(e.getNodeId(a)):e.forEachChild(a,n)})),r}(o,a),v=function(n,t,r,a){var i=new e.Map,o=e.createMultiMap();return e.forEachChild(n,(function n(s){if(e.isIdentifier(s)){var c=t.getSymbolAtLocation(s),u=c&&function(e,n){return e.valueDeclaration&&e.valueDeclaration.getSourceFile()===n}(c,a);if(c&&u){var d=g(t.getTypeAtLocation(s),t),p=e.getSymbolId(c).toString();if(!d||e.isParameter(s.parent)||e.isFunctionLikeDeclaration(s.parent)||r.has(p)){if(s.parent&&(e.isParameter(s.parent)||e.isVariableDeclaration(s.parent)||e.isBindingElement(s.parent))){var m=s.text,f=o.get(m);if(f&&f.some((function(e){return e!==c}))){var _=l(s,o);i.set(p,_.identifier),r.set(p,_),o.add(m,c)}else{var y=e.getSynthesizedDeepClone(s);r.set(p,E(y)),o.add(m,c)}}}else{var v=e.firstOrUndefined(d.parameters),h=v&&e.isParameter(v.valueDeclaration)&&e.tryCast(v.valueDeclaration.name,e.isIdentifier)||e.factory.createUniqueName("result",16),b=l(h,o);r.set(p,b),o.add(h.text,c)}}}else e.forEachChild(s,n)})),e.getSynthesizedDeepCloneWithReplacements(n,!0,(function(n){if(e.isBindingElement(n)&&e.isIdentifier(n.name)&&e.isObjectBindingPattern(n.parent)){if((a=(r=t.getSymbolAtLocation(n.name))&&i.get(String(e.getSymbolId(r))))&&a.text!==(n.name||n.propertyName).getText())return e.factory.createBindingElement(n.dotDotDotToken,n.propertyName||n.name,a,n.initializer)}else if(e.isIdentifier(n)){var r,a;if(a=(r=t.getSymbolAtLocation(n))&&i.get(String(e.getSymbolId(r))))return e.factory.createIdentifier(a.text)}}))}(o,a,f,i.sourceFile),h=v.body&&e.isBlock(v.body)?(p=v.body,m=[],e.forEachReturnStatement(p,(function(n){e.isReturnStatementWithFixablePromiseHandler(n)&&m.push(n)})),m):e.emptyArray,b={checker:a,synthNamesMap:f,setOfExpressionsToReturn:y,isInJSFile:_};if(h.length){n.insertLastModifierBefore(t,129,o);for(var T=function(r){e.forEachChild(r,(function a(i){if(e.isCallExpression(i)){var o=d(i,b);n.replaceNodeWithNodes(t,r,o)}else e.isFunctionLike(i)||e.forEachChild(i,a)}))},S=0,x=h;S<x.length;S++){T(x[S])}}}}function s(n,t,r){if(!e.isCallExpression(n))return!1;var a=e.hasPropertyAccessExpressionWithName(n,r)&&t.getTypeAtLocation(n);return!(!a||!t.getPromisedTypeOfPromise(a))}function c(n,t){return!!e.isExpression(n)&&!!t.getPromisedTypeOfPromise(t.getTypeAtLocation(n))}function l(n,t){var r=(t.get(n.text)||e.emptyArray).length;return E(0===r?n:e.factory.createIdentifier(n.text+"_"+r))}function u(){return i=!1,e.emptyArray}function d(n,t,r){if(s(n,t.checker,"then"))return 0===n.arguments.length?u():function(n,t,r){var a=n.arguments,i=a[0],o=a[1],s=v(i,t),c=f(i,r,s,n,t);if(o){var l=v(o,t),u=e.factory.createBlock(d(n.expression,t,s).concat(c)),p=f(o,r,l,n,t),m=l?T(l)?l.identifier.text:l.bindingPattern:"e",_=e.factory.createVariableDeclaration(m),g=e.factory.createCatchClause(_,e.factory.createBlock(p));return[e.factory.createTryStatement(u,g,void 0)]}return d(n.expression,t,s).concat(c)}(n,t,r);if(s(n,t.checker,"catch"))return 0===n.arguments.length?u():function(n,t,r){var a,i=n.arguments[0],o=v(i,t);r&&!S(n,t)&&(T(r)?(a=r,t.synthNamesMap.forEach((function(n,a){if(n.identifier.text===r.identifier.text){var i=function(n){return E(e.factory.createUniqueName(n.identifier.text,16))}(r);t.synthNamesMap.set(a,i)}}))):a=E(e.factory.createUniqueName("result",16),r.types),a.hasBeenDeclared=!0);var s,c,l=e.factory.createBlock(d(n.expression,t,a)),u=f(i,a,o,n,t),p=o?T(o)?o.identifier.text:o.bindingPattern:"e",m=e.factory.createVariableDeclaration(p),_=e.factory.createCatchClause(m,e.factory.createBlock(u));if(a&&!S(n,t)){c=e.getSynthesizedDeepClone(a.identifier);var g=a.types,y=t.checker.getUnionType(g,2),h=t.isInJSFile?void 0:t.checker.typeToTypeNode(y,void 0,void 0),b=[e.factory.createVariableDeclaration(c,void 0,h)];s=e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList(b,1))}var x=e.factory.createTryStatement(l,_,void 0),L=r&&c&&(A=r,1===A.kind)&&e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(e.getSynthesizedDeepClone(r.bindingPattern),void 0,void 0,c)],2));var A;return e.compact([s,x,L])}(n,t,r);if(e.isPropertyAccessExpression(n))return d(n.expression,t,r);var a=t.checker.getTypeAtLocation(n);return a&&t.checker.getPromisedTypeOfPromise(a)?(e.Debug.assertNode(n.original.parent,e.isPropertyAccessExpression),function(n,t,r){if(S(n,t))return[e.factory.createReturnStatement(e.getSynthesizedDeepClone(n))];return p(r,e.factory.createAwaitExpression(n),void 0)}(n,t,r)):u()}function p(n,t,r){return!n||h(n)?[e.factory.createExpressionStatement(t)]:T(n)&&n.hasBeenDeclared?[e.factory.createExpressionStatement(e.factory.createAssignment(e.getSynthesizedDeepClone(n.identifier),t))]:[e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(e.getSynthesizedDeepClone(b(n)),void 0,r,t)],2))]}function m(n,t){if(t&&n){var r=e.factory.createUniqueName("result",16);return __spreadArrays(p(E(r),n,t),[e.factory.createReturnStatement(r)])}return[e.factory.createReturnStatement(n)]}function f(n,t,r,a,i){var o,s,l,d,f;switch(n.kind){case 103:break;case 78:if(!r)break;var v=e.factory.createCallExpression(e.getSynthesizedDeepClone(n),void 0,T(r)?[r.identifier]:[]);if(S(a,i))return m(v,null===(o=a.typeArguments)||void 0===o?void 0:o[0]);var h=i.checker.getTypeAtLocation(n),E=i.checker.getSignaturesOfType(h,0);if(!E.length)return u();var x=E[0].getReturnType(),L=p(t,e.factory.createAwaitExpression(v),null===(s=a.typeArguments)||void 0===s?void 0:s[0]);return t&&t.types.push(x),L;case 208:case 209:var A=n.body,D=null===(l=g(i.checker.getTypeAtLocation(n),i.checker))||void 0===l?void 0:l.getReturnType();if(e.isBlock(A)){for(var C=[],I=!1,N=0,R=A.statements;N<R.length;N++){var k=R[N];if(e.isReturnStatement(k))if(I=!0,e.isReturnStatementWithFixablePromiseHandler(k))C=C.concat(y(i,[k],t));else{var O=D&&k.expression?_(i.checker,D,k.expression):k.expression;C.push.apply(C,m(O,null===(d=a.typeArguments)||void 0===d?void 0:d[0]))}else C.push(k)}return S(a,i)?C.map((function(n){return e.getSynthesizedDeepClone(n)})):function(n,t,r,a){for(var i=[],o=0,s=n;o<s.length;o++){var l=s[o];if(e.isReturnStatement(l)){if(l.expression){var u=c(l.expression,r.checker)?e.factory.createAwaitExpression(l.expression):l.expression;void 0===t?i.push(e.factory.createExpressionStatement(u)):i.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(b(t),void 0,void 0,u)],2)))}}else i.push(e.getSynthesizedDeepClone(l))}a||void 0===t||i.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(b(t),void 0,void 0,e.factory.createIdentifier("undefined"))],2)));return i}(C,t,i,I)}var P=y(i,e.isFixablePromiseHandler(A)?[e.factory.createReturnStatement(A)]:e.emptyArray,t);if(P.length>0)return P;if(D){O=_(i.checker,D,A);if(S(a,i))return m(O,null===(f=a.typeArguments)||void 0===f?void 0:f[0]);var M=p(t,O,void 0);return t&&t.types.push(D),M}return u();default:return u()}return e.emptyArray}function _(n,t,r){var a=e.getSynthesizedDeepClone(r);return n.getPromisedTypeOfPromise(t)?e.factory.createAwaitExpression(a):a}function g(n,t){var r=t.getSignaturesOfType(n,0);return e.lastOrUndefined(r)}function y(n,t,r){for(var a=[],i=0,o=t;i<o.length;i++){var s=o[i];e.forEachChild(s,(function t(i){if(e.isCallExpression(i)){var o=d(i,n,r);if((a=a.concat(o)).length>0)return}else e.isFunctionLike(i)||e.forEachChild(i,t)}))}return a}function v(n,t){var r,a=[];e.isFunctionLikeDeclaration(n)?n.parameters.length>0&&(r=function n(t){if(e.isIdentifier(t))return i(t);var r=e.flatMap(t.elements,(function(t){return e.isOmittedExpression(t)?[]:[n(t.name)]}));return function(n,t,r){void 0===t&&(t=e.emptyArray);void 0===r&&(r=[]);return{kind:1,bindingPattern:n,elements:t,types:r}}(t,r)}(n.parameters[0].name)):e.isIdentifier(n)&&(r=i(n));if(r&&(!("identifier"in r)||"undefined"!==r.identifier.text))return r;function i(n){var r,i=function(e){return e.symbol?e.symbol:t.checker.getSymbolAtLocation(e)}((r=n).original?r.original:r);return i&&t.synthNamesMap.get(e.getSymbolId(i).toString())||E(n,a)}}function h(n){return!n||(T(n)?!n.identifier.text:e.every(n.elements,h))}function b(e){return T(e)?e.identifier:e.bindingPattern}function E(e,n){return void 0===n&&(n=[]),{kind:0,identifier:e,types:n,hasBeenDeclared:!1}}function T(e){return 0===e.kind}function S(n,t){return!!n.original&&t.setOfExpressionsToReturn.has(e.getNodeId(n.original))}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){i=!0;var a=e.textChanges.ChangeTracker.with(t,(function(e){return o(e,t.sourceFile,t.span.start,t.program.getTypeChecker(),t)}));return i?[n.createCodeFixAction(r,a,e.Diagnostics.Convert_to_async_function,r,e.Diagnostics.Convert_all_to_async_functions)]:[]},fixIds:[r],getAllCodeActions:function(e){return n.codeFixAll(e,a,(function(n,t){return o(n,t.file,t.start,e.program.getTypeChecker(),e)}))}}),function(e){e[e.Identifier=0]="Identifier",e[e.BindingPattern=1]="BindingPattern"}(t||(t={}))}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){function t(n,t,r,a){for(var i=0,o=n.imports;i<o.length;i++){var s=o[i],c=e.getResolvedModule(n,s.text);if(c&&c.resolvedFileName===t.fileName){var l=e.importFromModuleSpecifier(s);switch(l.kind){case 260:r.replaceNode(n,l,e.makeImport(l.name,void 0,s,a));break;case 203:e.isRequireCall(l,!1)&&r.replaceNode(n,l,e.factory.createPropertyAccessExpression(e.getSynthesizedDeepClone(l),"default"))}}}}function r(n,t){n.forEachChild((function r(a){if(e.isPropertyAccessExpression(a)&&e.isExportsOrModuleExportsOrAlias(n,a.expression)&&e.isIdentifier(a.name)){var i=a.parent;t(a,e.isBinaryExpression(i)&&i.left===a&&62===i.operatorToken.kind)}a.forEachChild(r)}))}function a(n,t,r,a,l,u,d,m,f){switch(t.kind){case 232:return i(n,t,a,r,l,u,f),!1;case 233:var y=t.expression;switch(y.kind){case 203:return e.isRequireCall(y,!0)&&a.replaceNode(n,t,e.makeImport(void 0,void 0,y.arguments[0],f)),!1;case 216:return 62===y.operatorToken.kind&&function(n,t,r,a,i,l){var u=r.left,d=r.right;if(!e.isPropertyAccessExpression(u))return!1;if(e.isExportsOrModuleExportsOrAlias(n,u)){if(!e.isExportsOrModuleExportsOrAlias(n,d)){var m=e.isObjectLiteralExpression(d)?function(n,t){var r=e.mapAllOrFail(n.properties,(function(n){switch(n.kind){case 167:case 168:case 289:case 290:return;case 288:return e.isIdentifier(n.name)?function(n,t,r){var a=[e.factory.createToken(92)];switch(t.kind){case 208:var i=t.name;if(i&&i.text!==n)return o();case 209:return p(n,a,t,r);case 221:return function(n,t,r,a){return e.factory.createClassDeclaration(e.getSynthesizedDeepClones(r.decorators),e.concatenate(t,e.getSynthesizedDeepClones(r.modifiers)),n,e.getSynthesizedDeepClones(r.typeParameters),e.getSynthesizedDeepClones(r.heritageClauses),c(r.members,a))}(n,a,t,r);default:return o()}function o(){return _(a,e.factory.createIdentifier(n),c(t,r))}}(n.name.text,n.initializer,t):void 0;case 165:return e.isIdentifier(n.name)?p(n.name.text,[e.factory.createToken(92)],n,t):void 0;default:e.Debug.assertNever(n,"Convert to ES6 got invalid prop kind "+n.kind)}}));return r&&[r,!1]}(d,l):e.isRequireCall(d,!0)?function(n,t){var r=n.text,a=t.getSymbolAtLocation(n),i=a?a.exports:e.emptyMap;return i.has("export=")?[[s(r)],!0]:i.has("default")?i.size>1?[[o(r),s(r)],!0]:[[s(r)],!0]:[[o(r)],!1]}(d.arguments[0],t):void 0;return m?(a.replaceNodeWithNodes(n,r.parent,m[0]),m[1]):(a.replaceRangeWithText(n,e.createRange(u.getStart(n),d.pos),"export default"),!0)}a.delete(n,r.parent)}else e.isExportsOrModuleExportsOrAlias(n,u.expression)&&function(n,t,r,a){var i=t.left.name.text,o=a.get(i);if(void 0!==o){var s=[_(void 0,o,t.right),g([e.factory.createExportSpecifier(o,i)])];r.replaceNodeWithNodes(n,t.parent,s)}else!function(n,t,r){var a=n.left,i=n.right,o=n.parent,s=a.name.text;if(!(e.isFunctionExpression(i)||e.isArrowFunction(i)||e.isClassExpression(i))||i.name&&i.name.text!==s)r.replaceNodeRangeWithNodes(t,a.expression,e.findChildOfKind(a,24,t),[e.factory.createToken(92),e.factory.createToken(84)],{joiner:" ",suffix:" "});else{r.replaceRange(t,{pos:a.getStart(t),end:i.getStart(t)},e.factory.createToken(92),{suffix:" "}),i.name||r.insertName(t,i,s);var c=e.findChildOfKind(o,26,t);c&&r.delete(t,c)}}(t,n,r)}(n,r,a,i);return!1}(n,r,y,a,d,m)}default:return!1}}function i(t,r,a,i,o,s,c){var u,d=r.declarationList,p=!1,g=e.map(d.declarations,(function(r){var a=r.name,u=r.initializer;if(u){if(e.isExportsOrModuleExportsOrAlias(t,u))return p=!0,y([]);if(e.isRequireCall(u,!0))return p=!0,function(t,r,a,i,o,s){switch(t.kind){case 196:var c=e.mapAllOrFail(t.elements,(function(n){return n.dotDotDotToken||n.initializer||n.propertyName&&!e.isIdentifier(n.propertyName)||!e.isIdentifier(n.name)?void 0:f(n.propertyName&&n.propertyName.text,n.name.text)}));if(c)return y([e.makeImport(void 0,c,r,s)]);case 197:var u=l(n.moduleSpecifierToValidIdentifier(r.text,o),i);return y([e.makeImport(e.factory.createIdentifier(u),void 0,r,s),_(void 0,e.getSynthesizedDeepClone(t),e.factory.createIdentifier(u))]);case 78:return function(n,t,r,a,i){for(var o,s=r.getSymbolAtLocation(n),c=new e.Map,u=!1,d=0,p=a.original.get(n.text);d<p.length;d++){var m=p[d];if(r.getSymbolAtLocation(m)===s&&m!==n){var f=m.parent;if(e.isPropertyAccessExpression(f)){var _=f.expression,g=f.name.text;e.Debug.assert(_===m,"Didn't expect expression === use");var v=c.get(g);void 0===v&&(v=l(g,a),c.set(g,v)),(null!=o?o:o=new e.Map).set(f,e.factory.createIdentifier(v))}else u=!0}}var h=0===c.size?void 0:e.arrayFrom(e.mapIterator(c.entries(),(function(n){var t=n[0],r=n[1];return e.factory.createImportSpecifier(t===r?void 0:e.factory.createIdentifier(t),e.factory.createIdentifier(r))})));h||(u=!0);return y([e.makeImport(u?e.getSynthesizedDeepClone(n):void 0,h,t,i)],o)}(t,r,a,i,s);default:return e.Debug.assertNever(t,"Convert to ES6 module got invalid name kind "+t.kind)}}(a,u.arguments[0],i,o,s,c);if(e.isPropertyAccessExpression(u)&&e.isRequireCall(u.expression,!0))return p=!0,function(n,t,r,a,i){switch(n.kind){case 196:case 197:var o=l(t,a);return y([m(o,t,r,i),_(void 0,n,e.factory.createIdentifier(o))]);case 78:return y([m(n.text,t,r,i)]);default:return e.Debug.assertNever(n,"Convert to ES6 module got invalid syntax form "+n.kind)}}(a,u.name.text,u.expression.arguments[0],o,c)}return y([e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([r],d.flags))])}));if(p)return a.replaceNodeWithNodes(t,r,e.flatMap(g,(function(e){return e.newImports}))),e.forEach(g,(function(n){n.useSitesToUnqualify&&e.copyEntries(n.useSitesToUnqualify,null!=u?u:u=new e.Map)})),u}function o(e){return g(void 0,e)}function s(n){return g([e.factory.createExportSpecifier(void 0,"default")],n)}function c(n,t){return t&&e.some(e.arrayFrom(t.keys()),(function(t){return e.rangeContainsRange(n,t)}))?e.isArray(n)?e.getSynthesizedDeepClonesWithReplacements(n,!0,r):e.getSynthesizedDeepCloneWithReplacements(n,!0,r):n;function r(e){if(201===e.kind){var n=t.get(e);return t.delete(e),n}}}function l(e,n){for(;n.original.has(e)||n.additional.has(e);)e="_"+e;return n.additional.add(e),e}function u(n){var t=e.createMultiMap();return d(n,(function(e){return t.add(e.text,e)})),t}function d(n,t){e.isIdentifier(n)&&function(e){var n=e.parent;switch(n.kind){case 201:return n.name!==e;case 198:case 265:return n.propertyName!==e;default:return!0}}(n)&&t(n),n.forEachChild((function(e){return d(e,t)}))}function p(n,t,r,a){return e.factory.createFunctionDeclaration(e.getSynthesizedDeepClones(r.decorators),e.concatenate(t,e.getSynthesizedDeepClones(r.modifiers)),e.getSynthesizedDeepClone(r.asteriskToken),n,e.getSynthesizedDeepClones(r.typeParameters),e.getSynthesizedDeepClones(r.parameters),e.getSynthesizedDeepClone(r.type),e.factory.converters.convertToFunctionBlock(c(r.body,a)))}function m(n,t,r,a){return"default"===t?e.makeImport(e.factory.createIdentifier(n),void 0,r,a):e.makeImport(void 0,[f(t,n)],r,a)}function f(n,t){return e.factory.createImportSpecifier(void 0!==n&&n!==t?e.factory.createIdentifier(n):void 0,e.factory.createIdentifier(t))}function _(n,t,r){return e.factory.createVariableStatement(n,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(t,void 0,void 0,r)],2))}function g(n,t){return e.factory.createExportDeclaration(void 0,void 0,!1,n&&e.factory.createNamedExports(n),void 0===t?void 0:e.factory.createStringLiteral(t))}function y(e,n){return{newImports:e,useSitesToUnqualify:n}}n.registerCodeFix({errorCodes:[e.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module.code],getCodeActions:function(o){var s=o.sourceFile,c=o.program,d=o.preferences,p=e.textChanges.ChangeTracker.with(o,(function(n){if(function(n,t,o,s,c){var d={original:u(n),additional:new e.Set},p=function(n,t,a){var i=new e.Map;return r(n,(function(n){var r=n.name,o=r.text,s=r.originalKeywordKind;!i.has(o)&&(void 0!==s&&e.isNonContextualKeyword(s)||t.resolveName(o,n,111551,!0))&&i.set(o,l("_"+o,a))})),i}(n,t,d);!function(n,t,a){r(n,(function(r,i){if(!i){var o=r.name.text;a.replaceNode(n,r,e.factory.createIdentifier(t.get(o)||o))}}))}(n,p,o);for(var m,f=!1,_=0,g=e.filter(n.statements,e.isVariableStatement);_<g.length;_++){var y=g[_],v=i(n,y,o,t,d,s,c);v&&e.copyEntries(v,null!=m?m:m=new e.Map)}for(var h=0,b=e.filter(n.statements,(function(n){return!e.isVariableStatement(n)}));h<b.length;h++){y=b[h];var E=a(n,y,t,o,d,s,p,m,c);f=f||E}return null==m||m.forEach((function(e,t){o.replaceNode(n,t,e)})),f}(s,c.getTypeChecker(),n,c.getCompilerOptions().target,e.getQuotePreference(s,d)))for(var o=0,p=c.getSourceFiles();o<p.length;o++){var m=p[o];t(m,s,n,e.getQuotePreference(m,d))}}));return[n.createCodeFixActionWithoutFixAll("convertToEs6Module",p,e.Diagnostics.Convert_to_ES6_module)]}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="correctQualifiedNameToIndexedAccessType",r=[e.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1.code];function a(n,t){var r=e.findAncestor(e.getTokenAtPosition(n,t),e.isQualifiedName);return e.Debug.assert(!!r,"Expected position to be owned by a qualified name."),e.isIdentifier(r.left)?r:void 0}function i(n,t,r){var a=r.right.text,i=e.factory.createIndexedAccessTypeNode(e.factory.createTypeReferenceNode(r.left,void 0),e.factory.createLiteralTypeNode(e.factory.createStringLiteral(a)));n.replaceNode(t,r,i)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=a(r.sourceFile,r.span.start);if(o){var s=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,r.sourceFile,o)})),c=o.left.text+'["'+o.right.text+'"]';return[n.createCodeFixAction(t,s,[e.Diagnostics.Rewrite_as_the_indexed_access_type_0,c],t,e.Diagnostics.Rewrite_all_as_indexed_access_types)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics.Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type.code],r="convertToTypeOnlyExport";function a(n,t){return e.tryCast(e.getTokenAtPosition(t,n.start).parent,e.isExportSpecifier)}function i(n,r,a){if(r){var i=r.parent,o=i.parent,s=function(n,r){var a=n.parent;if(1===a.elements.length)return a.elements;var i=e.getDiagnosticsWithinSpan(e.createTextSpanFromNode(a),r.program.getSemanticDiagnostics(r.sourceFile,r.cancellationToken));return e.filter(a.elements,(function(r){var a;return r===n||(null===(a=e.findDiagnosticForNode(r,i))||void 0===a?void 0:a.code)===t[0]}))}(r,a);if(s.length===i.elements.length)n.insertModifierBefore(a.sourceFile,149,i);else{var c=e.factory.updateExportDeclaration(o,o.decorators,o.modifiers,!1,e.factory.updateNamedExports(i,e.filter(i.elements,(function(n){return!e.contains(s,n)}))),o.moduleSpecifier),l=e.factory.createExportDeclaration(void 0,void 0,!0,e.factory.createNamedExports(s),o.moduleSpecifier);n.replaceNode(a.sourceFile,o,c,{leadingTriviaOption:e.textChanges.LeadingTriviaOption.IncludeAll,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Exclude}),n.insertNodeAfter(a.sourceFile,o,l)}}}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var o=e.textChanges.ChangeTracker.with(t,(function(e){return i(e,a(t.span,t.sourceFile),t)}));if(o.length)return[n.createCodeFixAction(r,o,e.Diagnostics.Convert_to_type_only_export,r,e.Diagnostics.Convert_all_re_exported_types_to_type_only_exports)]},fixIds:[r],getAllCodeActions:function(r){var o=new e.Map;return n.codeFixAll(r,t,(function(n,t){var s=a(t,r.sourceFile);s&&e.addToSeen(o,e.getNodeId(s.parent.parent))&&i(n,s,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics.This_import_is_never_used_as_a_value_and_must_use_import_type_because_the_importsNotUsedAsValues_is_set_to_error.code],r="convertToTypeOnlyImport";function a(n,t){return e.tryCast(e.getTokenAtPosition(t,n.start).parent,e.isImportDeclaration)}function i(n,t,r){if(null==t?void 0:t.importClause){var a=t.importClause;n.insertText(r.sourceFile,t.getStart()+"import".length," type"),a.name&&a.namedBindings&&(n.deleteNodeRangeExcludingEnd(r.sourceFile,a.name,t.importClause.namedBindings),n.insertNodeBefore(r.sourceFile,t,e.factory.updateImportDeclaration(t,void 0,void 0,e.factory.createImportClause(!0,a.name,void 0),t.moduleSpecifier)))}}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var o=e.textChanges.ChangeTracker.with(t,(function(e){i(e,a(t.span,t.sourceFile),t)}));if(o.length)return[n.createCodeFixAction(r,o,e.Diagnostics.Convert_to_type_only_import,r,e.Diagnostics.Convert_all_imports_not_used_as_a_value_to_type_only_imports)]},fixIds:[r],getAllCodeActions:function(e){return n.codeFixAll(e,t,(function(n,t){i(n,a(t,e.sourceFile),e)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="convertLiteralTypeToMappedType",r=[e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0.code];function a(n,t){var r=e.getTokenAtPosition(n,t);if(e.isIdentifier(r)){var a=e.cast(r.parent.parent,e.isPropertySignature),i=r.getText(n);return{container:e.cast(a.parent,e.isTypeLiteralNode),typeNode:a.type,constraint:i,name:"K"===i?"P":"K"}}}function i(n,t,r){var a=r.container,i=r.typeNode,o=r.constraint,s=r.name;n.replaceNode(t,a,e.factory.createMappedTypeNode(void 0,e.factory.createTypeParameterDeclaration(s,e.factory.createTypeReferenceNode(o)),void 0,void 0,i))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start);if(c){var l=c.name,u=c.constraint,d=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c)}));return[n.createCodeFixAction(t,d,[e.Diagnostics.Convert_0_to_1_in_0,u,l],t,e.Diagnostics.Convert_all_type_literals_to_mapped_type)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics.Class_0_incorrectly_implements_interface_1.code,e.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code],r="fixClassIncorrectlyImplementsInterface";function a(n,t){return e.Debug.checkDefined(e.getContainingClass(e.getTokenAtPosition(n,t)),"There should be a containing class")}function i(n){return!(n.valueDeclaration&&8&e.getEffectiveModifierFlags(n.valueDeclaration))}function o(t,r,a,o,s,c){var l=t.program.getTypeChecker(),u=function(n,t){var r=e.getEffectiveBaseTypeNode(n);if(!r)return e.createSymbolTable();var a=t.getTypeAtLocation(r),o=t.getPropertiesOfType(a);return e.createSymbolTable(o.filter(i))}(o,l),d=l.getTypeAtLocation(r),p=l.getPropertiesOfType(d).filter(e.and(i,(function(e){return!u.has(e.escapedName)}))),m=l.getTypeAtLocation(o),f=e.find(o.members,(function(n){return e.isConstructorDeclaration(n)}));m.getNumberIndexType()||g(d,1),m.getStringIndexType()||g(d,0);var _=n.createImportAdder(a,t.program,c,t.host);function g(e,r){var i=l.getIndexInfoOfType(e,r);i&&y(a,o,l.indexInfoToIndexSignatureDeclaration(i,r,o,void 0,n.getNoopSymbolTrackerWithResolver(t)))}function y(e,n,t){f?s.insertNodeAfter(e,f,t):s.insertNodeAtClassStart(e,n,t)}n.createMissingMemberNodes(o,p,a,t,c,_,(function(e){return y(a,o,e)})),_.writeFixes(s)}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var i=t.sourceFile,s=t.span,c=a(i,s.start);return e.mapDefined(e.getEffectiveImplementsTypeNodes(c),(function(a){var s=e.textChanges.ChangeTracker.with(t,(function(e){return o(t,a,i,c,e,t.preferences)}));return 0===s.length?void 0:n.createCodeFixAction(r,s,[e.Diagnostics.Implement_interface_0,a.getText(i)],r,e.Diagnostics.Implement_all_unimplemented_interfaces)}))},fixIds:[r],getAllCodeActions:function(r){var i=new e.Map;return n.codeFixAll(r,t,(function(n,t){var s=a(t.file,t.start);if(e.addToSeen(i,e.getNodeId(s)))for(var c=0,l=e.getEffectiveImplementsTypeNodes(s);c<l.length;c++){var u=l[c];o(r,u,t.file,s,n,r.preferences)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){n.importFixName="import";var t,r,a="fixMissingImport",i=[e.Diagnostics.Cannot_find_name_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,e.Diagnostics.Cannot_find_namespace_0.code,e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code,e.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here.code];function o(n,t,r,a,i){var o=t.getCompilerOptions(),l=[],u=[],p=new e.Map,f=new e.Map;return{addImportFromDiagnostic:function(e,n){var t=m(n,e.code,e.start,r);if(!t||!t.fixes.length)return;_(t)},addImportFromExportedSymbol:function(l,u){var p=e.Debug.checkDefined(l.parent),m=e.getNameForExportedSymbol(l,e.getEmitScriptTarget(o)),f=t.getTypeChecker(),g=f.getMergedSymbol(e.skipAlias(l,f)),y=c(n,g,p,m,i,t,r),v=!!u&&2===o.importsNotUsedAsValues,h=d(n,t);_({fixes:[s(n,y,p,m,t,void 0,v,h,i,a)],symbolName:m})},writeFixes:function(t){for(var r,i=e.getQuotePreference(n,a),o=0,s=l;o<s.length;o++){var c=s[o];b(t,n,c)}for(var d=0,m=u;d<m.length;d++){c=m[d];E(t,n,c,i)}p.forEach((function(e){var r=e.importClauseOrBindingPattern,a=e.defaultImport,i=e.namedImports,o=e.canUseTypeOnlyImport;h(t,n,r,a,i,o)})),f.forEach((function(n,t){var a=n.useRequire,o=__rest(n,["useRequire"]),s=a?x:S;r=e.combine(r,s(t,i,o))})),r&&e.insertImports(t,n,r,!0)}};function _(n){var t=n.fixes,r=n.symbolName,a=e.first(t);switch(a.kind){case 0:l.push(a);break;case 1:u.push(a);break;case 2:var i=a.importClauseOrBindingPattern,o=a.importKind,s=a.canUseTypeOnlyImport,c=String(e.getNodeId(i));(d=p.get(c))||p.set(c,d={importClauseOrBindingPattern:i,defaultImport:void 0,namedImports:[],canUseTypeOnlyImport:s}),0===o?e.pushIfUnique(d.namedImports,r):(e.Debug.assert(void 0===d.defaultImport||d.defaultImport===r,"(Add to Existing) Default import should be missing or match symbolName"),d.defaultImport=r);break;case 3:var d,m=a.moduleSpecifier,_=(o=a.importKind,a.useRequire),g=a.typeOnly;switch((d=f.get(m))?d.typeOnly=d.typeOnly&&g:f.set(m,d={namedImports:[],namespaceLikeImport:void 0,typeOnly:g,useRequire:_}),o){case 1:e.Debug.assert(void 0===d.defaultImport||d.defaultImport===r,"(Add new) Default import should be missing or match symbolName"),d.defaultImport=r;break;case 0:e.pushIfUnique(d.namedImports||(d.namedImports=[]),r);break;case 3:case 2:e.Debug.assert(void 0===d.namespaceLikeImport||d.namespaceLikeImport.name===r,"Namespacelike import shoudl be missing or match symbolName"),d.namespaceLikeImport={importKind:o,name:r}}break;default:e.Debug.assertNever(a,"fix wasn't never - got kind "+a.kind)}}}function s(n,t,r,a,i,o,s,c,l,d){return e.Debug.assert(t.some((function(e){return e.moduleSymbol===r})),"Some exportInfo should match the specified moduleSymbol"),e.first(u(t,a,o,s,c,i,n,l,d))}function c(n,t,r,a,i,o,s){var c=[],u=o.getCompilerOptions();return D(o,i,n,!1,s,(function(i,o,s){var d=s.getTypeChecker();if(!o||i===r||!e.startsWith(n.fileName,e.getDirectoryPath(o.fileName))){var p=_(n,i,d,u);!p||p.name!==a&&I(i,u.target)!==a||e.skipAlias(p.symbol,d)!==t||c.push({moduleSymbol:i,importKind:p.kind,exportedSymbolIsTypeOnly:l(p.symbol,d)});for(var m=0,f=d.getExportsAndPropertiesOfModule(i);m<f.length;m++){var g=f[m];g.name===a&&e.skipAlias(g,d)===t&&c.push({moduleSymbol:i,importKind:0,exportedSymbolIsTypeOnly:l(g,d)})}}})),c}function l(n,t){return!(111551&e.skipAlias(n,t).flags)}function u(n,t,r,a,i,o,s,c,l){var u=o.getTypeChecker(),d=e.flatMap(n,(function(n){return function(n,t,r){var a=n.moduleSymbol,i=n.importKind;return n.exportedSymbolIsTypeOnly&&e.isSourceFileJS(r)?e.emptyArray:e.mapDefined(r.imports,(function(n){var r=e.importFromModuleSpecifier(n);return e.isRequireVariableDeclaration(r.parent,!0)?t.resolveExternalModuleName(n)===a?{declaration:r.parent,importKind:i}:void 0:(261===r.kind||260===r.kind)&&t.getSymbolAtLocation(n)===a?{declaration:r,importKind:i}:void 0}))}(n,u,s)})),m=void 0===r?void 0:function(n,t,r,a){return e.firstDefined(n,(function(n){var i=n.declaration,o=function(n){var t,r,a;switch(n.kind){case 249:return null===(t=e.tryCast(n.name,e.isIdentifier))||void 0===t?void 0:t.text;case 260:return n.name.text;case 261:return null===(a=e.tryCast(null===(r=n.importClause)||void 0===r?void 0:r.namedBindings,e.isNamespaceImport))||void 0===a?void 0:a.name.text;default:return e.Debug.assertNever(n)}}(i);if(o){var s=function(n,t){var r;switch(n.kind){case 249:return t.resolveExternalModuleName(n.initializer.arguments[0]);case 260:return t.getAliasedSymbol(n.symbol);case 261:var a=e.tryCast(null===(r=n.importClause)||void 0===r?void 0:r.namedBindings,e.isNamespaceImport);return a&&t.getAliasedSymbol(a.symbol);default:return e.Debug.assertNever(n)}}(i,a);if(s&&s.exports.has(e.escapeLeadingUnderscores(t)))return{kind:0,namespacePrefix:o,position:r}}}))}(d,t,r,u),f=function(n,t){return e.firstDefined(n,(function(e){var n=e.declaration,r=e.importKind;if(260!==n.kind){if(249===n.kind)return 0!==r&&1!==r||196!==n.name.kind?void 0:{kind:2,importClauseOrBindingPattern:n.name,importKind:r,moduleSpecifier:n.initializer.arguments[0].text,canUseTypeOnlyImport:!1};var a=n.importClause;if(a){var i=a.name,o=a.namedBindings;return 1===r&&!i||0===r&&(!o||264===o.kind)?{kind:2,importClauseOrBindingPattern:a,importKind:r,moduleSpecifier:n.moduleSpecifier.getText(),canUseTypeOnlyImport:t}:void 0}}}))}(d,void 0!==r&&function(n,t){return e.isValidTypeOnlyAliasUseSite(e.getTokenAtPosition(n,t))}(s,r)),_=f?[f]:function(n,t,r,a,i,o,s,c,l){var u=e.firstDefined(t,(function(n){return function(n,t,r){var a=n.declaration,i=n.importKind,o=261===a.kind?a.moduleSpecifier:249===a.kind?a.initializer.arguments[0]:272===a.moduleReference.kind?a.moduleReference.expression:void 0;return o&&e.isStringLiteral(o)?{kind:3,moduleSpecifier:o.text,importKind:i,typeOnly:t,useRequire:r}:void 0}(n,o,s)}));return u?[u]:p(r,a,i,o,s,n,c,l)}(n,d,o,s,r,a,i,c,l);return __spreadArrays(m?[m]:e.emptyArray,_)}function d(n,t){if(!e.isSourceFileJS(n))return!1;if(n.commonJsModuleIndicator&&!n.externalModuleIndicator)return!0;if(n.externalModuleIndicator&&!n.commonJsModuleIndicator)return!1;var r=t.getCompilerOptions();if(r.configFile)return e.getEmitModuleKind(r)<e.ModuleKind.ES2015;for(var a=0,i=t.getSourceFiles();a<i.length;a++){var o=i[a];if(o!==n&&e.isSourceFileJS(o)&&!t.isSourceFileFromExternalLibrary(o)){if(o.commonJsModuleIndicator&&!o.externalModuleIndicator)return!0;if(o.externalModuleIndicator&&!o.commonJsModuleIndicator)return!1}}return!0}function p(n,t,r,a,i,o,s,c){var l=e.isSourceFileJS(t),u=n.getCompilerOptions(),d=R(t,n,s).allowsImportingSpecifier,p=e.flatMap(o,(function(o){var d=o.moduleSymbol,p=o.importKind,m=o.exportedSymbolIsTypeOnly;return e.moduleSpecifiers.getModuleSpecifiers(d,u,t,e.createModuleSpecifierResolutionHost(n,s),c).map((function(n){return m&&l?{kind:1,moduleSpecifier:n,position:e.Debug.checkDefined(r,"position should be defined")}:{kind:3,moduleSpecifier:n,importKind:p,useRequire:i,typeOnly:a}}))}));return e.sort(p,(function(e,n){var t=d(e.moduleSpecifier),r=d(n.moduleSpecifier);return t&&!r?-1:r&&!t?1:e.moduleSpecifier.length-n.moduleSpecifier.length}))}function m(n,t,r,a){var i=e.getTokenAtPosition(n.sourceFile,r),o=t===e.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code?function(n,t){var r=n.sourceFile,a=n.program,i=n.host,o=n.preferences,s=a.getTypeChecker(),c=function(n,t){var r=e.isIdentifier(n)?t.getSymbolAtLocation(n):void 0;if(e.isUMDExportSymbol(r))return r;var a=n.parent;return e.isJsxOpeningLikeElement(a)&&a.tagName===n||e.isJsxOpeningFragment(a)?e.tryCast(t.resolveName(t.getJsxNamespace(a),e.isJsxOpeningLikeElement(a)?n:a,111551,!1),e.isUMDExportSymbol):void 0}(t,s);if(!c)return;var l=s.getAliasedSymbol(c),p=c.name,m=[{moduleSymbol:l,importKind:f(r,a.getCompilerOptions()),exportedSymbolIsTypeOnly:!1}],_=d(r,a);return{fixes:u(m,p,e.isIdentifier(t)?t.getStart(r):void 0,!1,_,a,r,i,o),symbolName:p}}(n,i):e.isIdentifier(i)?function(n,t,r){var a=n.sourceFile,i=n.program,o=n.cancellationToken,s=n.host,c=n.preferences,p=i.getTypeChecker(),m=function(n,t,r){var a=r.parent;if((e.isJsxOpeningLikeElement(a)||e.isJsxClosingElement(a))&&a.tagName===r){var i=t.getJsxNamespace(n);if(e.isIntrinsicJsxName(r.text)||!t.resolveName(i,a,111551,!0))return i}return r.text}(a,p,t);e.Debug.assert("default"!==m,"'default' isn't a legal identifier and couldn't occur here");var f=2===i.getCompilerOptions().importsNotUsedAsValues&&e.isValidTypeOnlyAliasUseSite(t),g=d(a,i),y=function(n,t,r,a,i,o,s){var c=e.createMultiMap();function u(n,t,r,a){c.add(e.getUniqueSymbolId(t,a).toString(),{moduleSymbol:n,importKind:r,exportedSymbolIsTypeOnly:l(t,a)})}return D(i,s,a,!0,o,(function(e,i,o){var s=o.getTypeChecker();r.throwIfCancellationRequested();var c=o.getCompilerOptions(),l=_(a,e,s,c);l&&(l.name===n||I(e,c.target)===n)&&A(l.symbolForMeaning,t)&&u(e,l.symbol,l.kind,s);var d=s.tryGetMemberInModuleExportsAndProperties(n,e);d&&A(d,t)&&u(e,d,0,s)})),c}(m,e.getMeaningFromLocation(t),o,a,i,r,s);return{fixes:e.arrayFrom(e.flatMapIterator(y.entries(),(function(e){e[0];return u(e[1],m,t.getStart(a),f,g,i,a,s,c)}))),symbolName:m}}(n,i,a):void 0;return o&&__assign(__assign({},o),{fixes:e.sort(o.fixes,(function(e,n){return e.kind-n.kind}))})}function f(n,t){if(e.getAllowSyntheticDefaultImports(t))return 1;var r=e.getEmitModuleKind(t);switch(r){case e.ModuleKind.AMD:case e.ModuleKind.CommonJS:case e.ModuleKind.UMD:return e.isInJSFile(n)&&e.isExternalModule(n)?2:3;case e.ModuleKind.System:case e.ModuleKind.ES2015:case e.ModuleKind.ES2020:case e.ModuleKind.ESNext:case e.ModuleKind.None:return 2;default:return e.Debug.assertNever(r,"Unexpected moduleKind "+r)}}function _(e,n,t,r){var a=function(e,n,t,r){var a=t.tryGetMemberInModuleExports("default",n);if(a)return{symbol:a,kind:1};var i=t.resolveExternalModuleSymbol(n);return i===n?void 0:{symbol:i,kind:g(e,r)}}(e,n,t,r);if(a){var i=a.symbol,o=a.kind,s=y(i,n,t,r);return s&&__assign({symbol:i,kind:o},s)}}function g(n,t){var r=e.getAllowSyntheticDefaultImports(t);if(e.getEmitModuleKind(t)>=e.ModuleKind.ES2015)return r?1:2;if(e.isInJSFile(n))return e.isExternalModule(n)?1:3;for(var a=0,i=n.statements;a<i.length;a++){var o=i[a];if(e.isImportEqualsDeclaration(o))return 3}return r?1:3}function y(n,t,r,a){var i=e.getLocalSymbolForExportDefault(n);if(i)return{symbolForMeaning:i,name:i.name};var o,s=(o=n).declarations&&e.firstDefined(o.declarations,(function(n){if(e.isExportAssignment(n)){if(e.isIdentifier(n.expression))return n.expression.text}else if(e.isExportSpecifier(n))return e.Debug.assert("default"===n.name.text,"Expected the specifier to be a default export"),n.propertyName&&n.propertyName.text}));if(void 0!==s)return{symbolForMeaning:n,name:s};if(2097152&n.flags){var c=r.getImmediateAliasedSymbol(n);if(c&&c.parent)return y(c,c.parent,r,a)}return"default"!==n.escapedName&&"export="!==n.escapedName?{symbolForMeaning:n,name:n.getName()}:{symbolForMeaning:n,name:I(t,a.target)}}function v(t,r,i,o,s){var c,l=e.textChanges.ChangeTracker.with(t,(function(n){c=function(n,t,r,a,i){switch(a.kind){case 0:return b(n,t,a),[e.Diagnostics.Change_0_to_1,r,a.namespacePrefix+"."+r];case 1:return E(n,t,a,i),[e.Diagnostics.Change_0_to_1,r,T(a.moduleSpecifier,i)+r];case 2:var o=a.importClauseOrBindingPattern,s=a.importKind,c=a.canUseTypeOnlyImport,l=a.moduleSpecifier;h(n,t,o,1===s?r:void 0,0===s?[r]:e.emptyArray,c);var u=e.stripQuotes(l);return[1===s?e.Diagnostics.Add_default_import_0_to_existing_import_declaration_from_1:e.Diagnostics.Add_0_to_existing_import_declaration_from_1,r,u];case 3:s=a.importKind,l=a.moduleSpecifier;var d=a.typeOnly,p=a.useRequire?x:S,m=1===s?{defaultImport:r,typeOnly:d}:0===s?{namedImports:[r],typeOnly:d}:{namespaceLikeImport:{importKind:s,name:r},typeOnly:d};return e.insertImports(n,t,p(l,i,m),!0),[1===s?e.Diagnostics.Import_default_0_from_module_1:e.Diagnostics.Import_0_from_module_1,r,l];default:return e.Debug.assertNever(a,"Unexpected fix kind "+a.kind)}}(n,r,i,o,s)}));return n.createCodeFixAction(n.importFixName,l,c,a,e.Diagnostics.Add_all_missing_imports)}function h(n,t,r,a,i,o){if(196!==r.kind){var s=!o&&r.isTypeOnly;if(a&&(e.Debug.assert(!r.name,"Cannot add a default import to an import clause that already has one"),n.insertNodeAt(t,r.getStart(t),e.factory.createIdentifier(a),{suffix:", "})),i.length){var c=r.namedBindings&&e.cast(r.namedBindings,e.isNamedImports).elements,l=e.stableSort(i.map((function(n){return e.factory.createImportSpecifier(void 0,e.factory.createIdentifier(n))})),e.OrganizeImports.compareImportOrExportSpecifiers);if((null==c?void 0:c.length)&&e.OrganizeImports.importSpecifiersAreSorted(c))for(var u=0,d=l;u<d.length;u++){var p=d[u],m=e.OrganizeImports.getImportSpecifierInsertionIndex(c,p),f=r.namedBindings.elements[m-1];f?n.insertNodeInListAfter(t,f,p):n.insertNodeBefore(t,c[0],p,!e.positionsAreOnSameLine(c[0].getStart(),r.parent.getStart(),t))}else if(null==c?void 0:c.length)for(var _=0,g=l;_<g.length;_++){p=g[_];n.insertNodeInListAfter(t,e.last(c),p,c)}else if(l.length){var y=e.factory.createNamedImports(l);r.namedBindings?n.replaceNode(t,r.namedBindings,y):n.insertNodeAfter(t,e.Debug.checkDefined(r.name,"Import clause must have either named imports or a default import"),y)}}s&&n.delete(t,e.getTypeKeywordOfTypeOnlyImport(r,t))}else{a&&b(r,a,"default");for(var v=0,h=i;v<h.length;v++){b(r,h[v],void 0)}}function b(r,a,i){var o=e.factory.createBindingElement(void 0,i,a);r.elements.length?n.insertNodeInListAfter(t,e.last(r.elements),o):n.replaceNode(t,r,e.factory.createObjectBindingPattern([o]))}}function b(e,n,t){var r=t.namespacePrefix,a=t.position;e.insertText(n,a,r+".")}function E(e,n,t,r){var a=t.moduleSpecifier,i=t.position;e.insertText(n,i,T(a,r))}function T(n,t){var r=e.getQuoteFromPreference(t);return"import("+r+n+r+")."}function S(n,t,r){var a,i,o,s=e.makeStringLiteral(n,t);(void 0!==r.defaultImport||(null===(a=r.namedImports)||void 0===a?void 0:a.length))&&(o=e.combine(o,e.makeImport(void 0===r.defaultImport?void 0:e.factory.createIdentifier(r.defaultImport),null===(i=r.namedImports)||void 0===i?void 0:i.map((function(n){return e.factory.createImportSpecifier(void 0,e.factory.createIdentifier(n))})),n,t,r.typeOnly)));var c=r.namespaceLikeImport,l=r.typeOnly;if(c){var u=3===c.importKind?e.factory.createImportEqualsDeclaration(void 0,void 0,e.factory.createIdentifier(c.name),e.factory.createExternalModuleReference(s)):e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(l,void 0,e.factory.createNamespaceImport(e.factory.createIdentifier(c.name))),s);o=e.combine(o,u)}return e.Debug.checkDefined(o)}function x(n,t,r){var a,i,o,s=e.makeStringLiteral(n,t);if(r.defaultImport||(null===(a=r.namedImports)||void 0===a?void 0:a.length)){var c=(null===(i=r.namedImports)||void 0===i?void 0:i.map((function(n){return e.factory.createBindingElement(void 0,void 0,n)})))||[];r.defaultImport&&c.unshift(e.factory.createBindingElement(void 0,"default",r.defaultImport));var l=L(e.factory.createObjectBindingPattern(c),s);o=e.combine(o,l)}if(r.namespaceLikeImport){l=L(r.namespaceLikeImport.name,s);o=e.combine(o,l)}return e.Debug.checkDefined(o)}function L(n,t){return e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration("string"==typeof n?e.factory.createIdentifier(n):n,void 0,void 0,e.factory.createCallExpression(e.factory.createIdentifier("require"),void 0,[t]))],2))}function A(n,t){var r=n.declarations;return e.some(r,(function(n){return!!(e.getMeaningFromDeclaration(n)&t)}))}function D(n,t,r,a,i,o){var s,c;C(n,t,r,a,(function(e,t){return o(e,t,n,!1)}));var l=i&&(null===(s=t.getPackageJsonAutoImportProvider)||void 0===s?void 0:s.call(t));if(l){var u=e.timestamp();C(l,t,r,a,(function(e,n){return o(e,n,l,!0)})),null===(c=t.log)||void 0===c||c.call(t,"forEachExternalModuleToImportFrom autoImportProvider: "+(e.timestamp()-u))}}function C(n,t,r,a,i){var o,s=0,c=e.createModuleSpecifierResolutionHost(n,t),l=a&&R(r,n,t,c);!function(n,t,r){for(var a=0,i=n.getAmbientModules();a<i.length;a++){r(i[a],void 0)}for(var o=0,s=t;o<s.length;o++){var c=s[o];e.isExternalOrCommonJsModule(c)&&r(n.getMergedSymbol(c.symbol),c)}}(n.getTypeChecker(),n.getSourceFiles(),(function(t,a){void 0===a?!l||l.allowsImportingAmbientModule(t)?i(t,a):l&&s++:a&&a!==r&&function(n,t,r,a){var i,o=e.hostGetCanonicalFileName(a),s=null===(i=a.getGlobalTypingsCacheLocation)||void 0===i?void 0:i.call(a);return!!e.moduleSpecifiers.forEachFileNameOfModule(t.fileName,r.fileName,a,!1,(function(a){var i=n.getSourceFile(a);return(i===r||!i)&&function(n,t,r,a){var i=e.forEachAncestorDirectory(t,(function(n){return"node_modules"===e.getBaseFileName(n)?n:void 0})),o=i&&e.getDirectoryPath(r(i));return void 0===o||e.startsWith(r(n),o)||!!a&&e.startsWith(r(a),o)}(t.fileName,a,o,s)}))}(n,r,a,c)&&(!l||l.allowsImportingSourceFile(a)?i(t,a):l&&s++)})),null===(o=t.log)||void 0===o||o.call(t,"forEachExternalModuleToImportFrom: filtered out "+s+" modules by package.json contents")}function I(n,t){return N(e.removeFileExtension(e.stripQuotes(n.name)),t)}function N(n,t){var r=e.getBaseFileName(e.removeSuffix(n,"/index")),a="",i=!0,o=r.charCodeAt(0);e.isIdentifierStart(o,t)?a+=String.fromCharCode(o):i=!1;for(var s=1;s<r.length;s++){var c=r.charCodeAt(s),l=e.isIdentifierPart(c,t);if(l){var u=String.fromCharCode(c);i||(u=u.toUpperCase()),a+=u}i=l}return e.isStringANonContextualKeyword(a)?"_"+a:a||"_"}function R(n,t,r,a){void 0===a&&(a=e.createModuleSpecifierResolutionHost(t,r));var i,o=(r.getPackageJsonsVisibleToFile&&r.getPackageJsonsVisibleToFile(n.fileName)||e.getPackageJsonsVisibleToFile(n.fileName,r)).filter((function(e){return e.parseable}));return{allowsImportingAmbientModule:function(n){if(!o.length)return!0;var t=l(n.valueDeclaration.getSourceFile().fileName);if(void 0===t)return!0;var r=e.stripQuotes(n.getName());if(c(r))return!0;return s(t)||s(r)},allowsImportingSourceFile:function(e){if(!o.length)return!0;var n=l(e.fileName);if(!n)return!0;return s(n)},allowsImportingSpecifier:function(n){if(!o.length||c(n))return!0;if(e.pathIsRelative(n)||e.isRootedDiskPath(n))return!0;return s(n)},moduleSpecifierResolutionHost:a};function s(n){for(var t=u(n),r=0,a=o;r<a.length;r++){var i=a[r];if(i.has(t)||i.has(e.getTypesPackageName(t)))return!0}return!1}function c(t){return!!(e.isSourceFileJS(n)&&e.JsTyping.nodeCoreModules.has(t)&&(void 0===i&&(i=e.consumesNodeCoreModules(n)),i))}function l(t){if(e.stringContains(t,"node_modules")){var i=e.moduleSpecifiers.getNodeModulesPackageName(r.getCompilationSettings(),n.path,t,a);if(i)return e.pathIsRelative(i)||e.isRootedDiskPath(i)?void 0:u(i)}}function u(n){var t=e.getPathComponents(e.getPackageNameFromTypesPackageName(n)).slice(1);return e.startsWith(t[0],"@")?t[0]+"/"+t[1]:t[0]}}n.registerCodeFix({errorCodes:i,getCodeActions:function(n){var t=n.errorCode,r=n.preferences,a=n.sourceFile,i=n.span,o=m(n,t,i.start,!0);if(o){var s=o.fixes,c=o.symbolName,l=e.getQuotePreference(a,r);return s.map((function(e){return v(n,a,c,e,l)}))}},fixIds:[a],getAllCodeActions:function(t){var r=o(t.sourceFile,t.program,!0,t.preferences,t.host);return n.eachDiagnostic(t,i,(function(e){return r.addImportFromDiagnostic(e,t)})),n.createCombinedCodeActions(e.textChanges.ChangeTracker.with(t,r.writeFixes))}}),n.createImportAdder=function(e,n,t,r){return o(e,n,!1,t,r)},function(e){e[e.UseNamespace=0]="UseNamespace",e[e.ImportType=1]="ImportType",e[e.AddToExisting=2]="AddToExisting",e[e.AddNew=3]="AddNew"}(t||(t={})),function(e){e[e.Named=0]="Named",e[e.Default=1]="Default",e[e.Namespace=2]="Namespace",e[e.CommonJS=3]="CommonJS"}(r||(r={})),n.getImportCompletionAction=function(n,t,r,a,i,o,l,u,m){var f,_,g,y,h=o.getCompilerOptions(),b=c(r,n,t,a,i,o,!0),E=d(r,o),T=2===h.importsNotUsedAsValues&&!e.isSourceFileJS(r)&&e.isValidTypeOnlyAliasUseSite(e.getTokenAtPosition(r,u)),S=e.first(p(o,r,u,T,E,b,i,m)).moduleSpecifier,x=s(r,b,t,a,o,u,T,E,i,m);return{moduleSpecifier:S,codeAction:(f=v({host:i,formatContext:l,preferences:m},r,a,x,e.getQuotePreference(r,m)),_=f.description,g=f.changes,y=f.commands,{description:_,changes:g,commands:y})}},n.forEachExternalModuleToImportFrom=D,n.moduleSymbolToValidIdentifier=I,n.moduleSpecifierToValidIdentifier=N}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixImplicitThis",r=[e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code];function a(n,t,r,a){var i=e.getTokenAtPosition(t,r);e.Debug.assert(107===i.kind);var o=e.getThisContainer(i,!1);if((e.isFunctionDeclaration(o)||e.isFunctionExpression(o))&&!e.isSourceFile(e.getThisContainer(o,!1))){var s=e.Debug.assertDefined(e.findChildOfKind(o,97,t)),c=o.name,l=e.Debug.assertDefined(o.body);if(e.isFunctionExpression(o)){if(c&&e.FindAllReferences.Core.isSymbolReferencedInFile(c,a,t,l))return;return n.delete(t,s),c&&n.delete(t,c),n.insertText(t,l.pos," =>"),[e.Diagnostics.Convert_function_expression_0_to_arrow_function,c?c.text:e.ANONYMOUS]}return n.replaceNode(t,s,e.factory.createToken(84)),n.insertText(t,c.end," = "),n.insertText(t,l.pos," =>"),[e.Diagnostics.Convert_function_declaration_0_to_arrow_function,c.text]}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i,o=r.sourceFile,s=r.program,c=r.span,l=e.textChanges.ChangeTracker.with(r,(function(e){i=a(e,o,c.start,s.getTypeChecker())}));return i?[n.createCodeFixAction(t,l,i,t,e.Diagnostics.Fix_all_implicit_this_errors)]:e.emptyArray},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){a(n,t.file,t.start,e.program.getTypeChecker())}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixIncorrectNamedTupleSyntax",r=[e.Diagnostics.A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type.code,e.Diagnostics.A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type.code];n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var a=r.sourceFile,i=r.span,o=function(n,t){var r=e.getTokenAtPosition(n,t);return e.findAncestor(r,(function(e){return 192===e.kind}))}(a,i.start),s=e.textChanges.ChangeTracker.with(r,(function(n){return function(n,t,r){if(!r)return;var a=r.type,i=!1,o=!1;for(;180===a.kind||181===a.kind||186===a.kind;)180===a.kind?i=!0:181===a.kind&&(o=!0),a=a.type;var s=e.factory.updateNamedTupleMember(r,r.dotDotDotToken||(o?e.factory.createToken(25):void 0),r.name,r.questionToken||(i?e.factory.createToken(57):void 0),a);if(s===r)return;n.replaceNode(t,r,s)}(n,a,o)}));return[n.createCodeFixAction(t,s,e.Diagnostics.Move_labeled_tuple_element_modifiers_to_labels,t,e.Diagnostics.Move_labeled_tuple_element_modifiers_to_labels)]},fixIds:[t]})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixSpelling",r=[e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,e.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2.code,e.Diagnostics.No_overload_matches_this_call.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1.code];function a(n,t,r,a){var i=e.getTokenAtPosition(n,t),o=i.parent;if(a!==e.Diagnostics.No_overload_matches_this_call.code&&a!==e.Diagnostics.Type_0_is_not_assignable_to_type_1.code||e.isJsxAttribute(o)){var s,c=r.program.getTypeChecker();if(e.isPropertyAccessExpression(o)&&o.name===i){e.Debug.assert(e.isIdentifierOrPrivateIdentifier(i),"Expected an identifier for spelling (property access)");var l=c.getTypeAtLocation(o.expression);32&o.flags&&(l=c.getNonNullableType(l)),s=c.getSuggestedSymbolForNonexistentProperty(i,l)}else if(e.isQualifiedName(o)&&o.right===i){var u=c.getSymbolAtLocation(o.left);u&&1536&u.flags&&(s=c.getSuggestedSymbolForNonexistentModule(o.right,u))}else if(e.isImportSpecifier(o)&&o.name===i){e.Debug.assertNode(i,e.isIdentifier,"Expected an identifier for spelling (import)");var d=function(n,t,r){if(!r||!e.isStringLiteralLike(r.moduleSpecifier))return;var a=e.getResolvedModule(n,r.moduleSpecifier.text);return a?t.program.getSourceFile(a.resolvedFileName):void 0}(n,r,e.findAncestor(i,e.isImportDeclaration));d&&d.symbol&&(s=c.getSuggestedSymbolForNonexistentModule(i,d.symbol))}else if(e.isJsxAttribute(o)&&o.name===i){e.Debug.assertNode(i,e.isIdentifier,"Expected an identifier for JSX attribute");var p=e.findAncestor(i,e.isJsxOpeningLikeElement),m=c.getContextualTypeForArgumentAtIndex(p,0);s=c.getSuggestedSymbolForNonexistentJSXAttribute(i,m)}else{var f=e.getMeaningFromLocation(i),_=e.getTextOfNode(i);e.Debug.assert(void 0!==_,"name should be defined"),s=c.getSuggestedSymbolForNonexistentSymbol(i,_,function(e){var n=0;4&e&&(n|=1920);2&e&&(n|=788968);1&e&&(n|=111551);return n}(f))}return void 0===s?void 0:{node:i,suggestedSymbol:s}}}function i(n,t,r,a,i){var o=e.symbolName(a);if(!e.isIdentifierText(o,i)&&e.isPropertyAccessExpression(r.parent)){var s=a.valueDeclaration;e.isNamedDeclaration(s)&&e.isPrivateIdentifier(s.name)?n.replaceNode(t,r,e.factory.createIdentifier(o)):n.replaceNode(t,r.parent,e.factory.createElementAccessExpression(r.parent.expression,e.factory.createStringLiteral(o)))}else n.replaceNode(t,r,e.factory.createIdentifier(o))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.errorCode,c=a(o,r.span.start,r,s);if(c){var l=c.node,u=c.suggestedSymbol,d=r.host.getCompilationSettings().target,p=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,l,u,d)}));return[n.createCodeFixAction("spelling",p,[e.Diagnostics.Change_spelling_to_0,e.symbolName(u)],t,e.Diagnostics.Fix_all_detected_spelling_errors)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=a(t.file,t.start,e,t.code),o=e.host.getCompilationSettings().target;r&&i(n,e.sourceFile,r.node,r.suggestedSymbol,o)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t,r="returnValueCorrect",a="fixAddReturnStatement",i="fixRemoveBracesFromArrowFunctionBody",o="fixWrapTheBlockWithParen",s=[e.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1.code,e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code];function c(n,t,r){var a=n.createSymbol(4,t.escapedText);a.type=n.getTypeAtLocation(r);var i=e.createSymbolTable([a]);return n.createAnonymousType(void 0,i,[],[],void 0,void 0)}function l(n,r,a,i){if(r.body&&e.isBlock(r.body)&&1===e.length(r.body.statements)){var o=e.first(r.body.statements);if(e.isExpressionStatement(o)&&u(n,r,n.getTypeAtLocation(o.expression),a,i))return{declaration:r,kind:t.MissingReturnStatement,expression:o.expression,statement:o,commentSource:o.expression};if(e.isLabeledStatement(o)&&e.isExpressionStatement(o.statement)){var s=e.factory.createObjectLiteralExpression([e.factory.createPropertyAssignment(o.label,o.statement.expression)]);if(u(n,r,c(n,o.label,o.statement.expression),a,i))return e.isArrowFunction(r)?{declaration:r,kind:t.MissingParentheses,expression:s,statement:o,commentSource:o.statement.expression}:{declaration:r,kind:t.MissingReturnStatement,expression:s,statement:o,commentSource:o.statement.expression}}else if(e.isBlock(o)&&1===e.length(o.statements)){var l=e.first(o.statements);if(e.isLabeledStatement(l)&&e.isExpressionStatement(l.statement)){s=e.factory.createObjectLiteralExpression([e.factory.createPropertyAssignment(l.label,l.statement.expression)]);if(u(n,r,c(n,l.label,l.statement.expression),a,i))return{declaration:r,kind:t.MissingReturnStatement,expression:s,statement:o,commentSource:l}}}}}function u(n,t,r,a,i){if(i){var o=n.getSignatureFromDeclaration(t);if(o){e.hasSyntacticModifier(t,256)&&(r=n.createPromiseType(r));var s=n.createSignature(t,o.typeParameters,o.thisParameter,o.parameters,r,void 0,o.minArgumentCount,o.flags);r=n.createAnonymousType(void 0,e.createSymbolTable(),[s],[],void 0,void 0)}else r=n.getAnyType()}return n.isTypeAssignableTo(r,a)}function d(n,t,r,a){var i=e.getTokenAtPosition(t,r);if(i.parent){var o=e.findAncestor(i.parent,e.isFunctionLikeDeclaration);switch(a){case e.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value.code:if(!(o&&o.body&&o.type&&e.rangeContainsRange(o.type,i)))return;return l(n,o,n.getTypeFromTypeNode(o.type),!1);case e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code:if(!o||!e.isCallExpression(o.parent)||!o.body)return;var s=o.parent.arguments.indexOf(o),c=n.getContextualTypeForArgumentAtIndex(o.parent,s);if(!c)return;return l(n,o,c,!0);case e.Diagnostics.Type_0_is_not_assignable_to_type_1.code:if(!e.isDeclarationName(i)||!e.isVariableLike(i.parent)&&!e.isJsxAttribute(i.parent))return;var u=function(n){switch(n.kind){case 249:case 160:case 198:case 163:case 288:return n.initializer;case 280:return n.initializer&&(e.isJsxExpression(n.initializer)?n.initializer.expression:void 0);case 289:case 162:case 291:case 333:case 326:return}}(i.parent);if(!u||!e.isFunctionLikeDeclaration(u)||!u.body)return;return l(n,u,n.getTypeAtLocation(i.parent),!0)}}}function p(n,t,r,a){e.suppressLeadingAndTrailingTrivia(r);var i=e.probablyUsesSemicolons(t);n.replaceNode(t,a,e.factory.createReturnStatement(r),{leadingTriviaOption:e.textChanges.LeadingTriviaOption.Exclude,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Exclude,suffix:i?";":void 0})}function m(n,t,r,a,i,o){var s=o||e.needsParentheses(a)?e.factory.createParenthesizedExpression(a):a;e.suppressLeadingAndTrailingTrivia(i),e.copyComments(i,s),n.replaceNode(t,r.body,s)}function f(n,t,r,a){n.replaceNode(t,r.body,e.factory.createParenthesizedExpression(a))}function _(t,i,o){var s=e.textChanges.ChangeTracker.with(t,(function(e){return p(e,t.sourceFile,i,o)}));return n.createCodeFixAction(r,s,e.Diagnostics.Add_a_return_statement,a,e.Diagnostics.Add_all_missing_return_statement)}function g(t,a,i){var s=e.textChanges.ChangeTracker.with(t,(function(e){return f(e,t.sourceFile,a,i)}));return n.createCodeFixAction(r,s,e.Diagnostics.Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal,o,e.Diagnostics.Wrap_all_object_literal_with_parentheses)}!function(e){e[e.MissingReturnStatement=0]="MissingReturnStatement",e[e.MissingParentheses=1]="MissingParentheses"}(t||(t={})),n.registerCodeFix({errorCodes:s,fixIds:[a,i,o],getCodeActions:function(a){var o=a.program,s=a.sourceFile,c=a.span.start,l=a.errorCode,u=d(o.getTypeChecker(),s,c,l);if(u)return u.kind===t.MissingReturnStatement?e.append([_(a,u.expression,u.statement)],e.isArrowFunction(u.declaration)?function(t,a,o,s){var c=e.textChanges.ChangeTracker.with(t,(function(e){return m(e,t.sourceFile,a,o,s,!1)}));return n.createCodeFixAction(r,c,e.Diagnostics.Remove_braces_from_arrow_function_body,i,e.Diagnostics.Remove_braces_from_all_arrow_function_bodies_with_relevant_issues)}(a,u.declaration,u.expression,u.commentSource):void 0):[g(a,u.declaration,u.expression)]},getAllCodeActions:function(t){return n.codeFixAll(t,s,(function(n,r){var s=d(t.program.getTypeChecker(),r.file,r.start,r.code);if(s)switch(t.fixId){case a:p(n,r.file,s.expression,s.statement);break;case i:if(!e.isArrowFunction(s.declaration))return;m(n,r.file,s.declaration,s.expression,s.commentSource,!1);break;case o:if(!e.isArrowFunction(s.declaration))return;f(n,r.file,s.declaration,s.expression);break;default:e.Debug.fail(JSON.stringify(t.fixId))}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t,r="addMissingMember",a=[e.Diagnostics.Property_0_does_not_exist_on_type_1.code,e.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,e.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2.code,e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2.code,e.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more.code],i="addMissingMember";function o(n,t,r,a){var i=e.getTokenAtPosition(n,t);if(e.isIdentifier(i)||e.isPrivateIdentifier(i)){var o=i.parent;if(e.isPropertyAccessExpression(o)){var s=e.skipConstraint(r.getTypeAtLocation(o.expression)),c=s.symbol;if(c&&c.declarations){var l=e.find(c.declarations,e.isClassLike);if(l||!e.isPrivateIdentifier(i)){var u=l||e.find(c.declarations,e.isInterfaceDeclaration);if(u&&!a.isSourceFileFromExternalLibrary(u.getSourceFile())){var d=(s.target||s)!==r.getDeclaredTypeOfSymbol(c);if(d&&(e.isPrivateIdentifier(i)||e.isInterfaceDeclaration(u)))return;var p=u.getSourceFile(),m=(d?32:0)|(e.startsWithUnderscore(i.text)?8:0),f=e.isSourceFileJS(p);return{kind:1,token:i,call:e.tryCast(o.parent,e.isCallExpression),modifierFlags:m,parentDeclaration:u,declSourceFile:p,isJSFile:f}}var _=e.find(c.declarations,e.isEnumDeclaration);return!_||e.isPrivateIdentifier(i)||a.isSourceFileFromExternalLibrary(_.getSourceFile())?void 0:{kind:0,token:i,parentDeclaration:_}}}}}}function s(n,t,r,a,i){var o=a.text;if(i){if(221===r.kind)return;var s=r.name.getText(),l=c(e.factory.createIdentifier(s),o);n.insertNodeAfter(t,r,l)}else if(e.isPrivateIdentifier(a)){var u=e.factory.createPropertyDeclaration(void 0,void 0,o,void 0,void 0,void 0),p=d(r);p?n.insertNodeAfter(t,p,u):n.insertNodeAtClassStart(t,r,u)}else{var m=e.getFirstConstructorWithBody(r);if(!m)return;var f=c(e.factory.createThis(),o);n.insertNodeAtConstructorEnd(t,m,f)}}function c(n,t){return e.factory.createExpressionStatement(e.factory.createAssignment(e.factory.createPropertyAccessExpression(n,t),e.factory.createIdentifier("undefined")))}function l(n,t,r){var a;if(216===r.parent.parent.kind){var i=r.parent.parent,o=r.parent===i.left?i.right:i.left,s=n.getWidenedType(n.getBaseTypeOfLiteralType(n.getTypeAtLocation(o)));a=n.typeToTypeNode(s,t,void 0)}else{var c=n.getContextualType(r.parent);a=c?n.typeToTypeNode(c,void 0,void 0):void 0}return a||e.factory.createKeywordTypeNode(128)}function u(n,t,r,a,i,o){var s=e.factory.createPropertyDeclaration(void 0,o?e.factory.createNodeArray(e.factory.createModifiersFromModifierFlags(o)):void 0,a,void 0,i,void 0),c=d(r);c?n.insertNodeAfter(t,c,s):n.insertNodeAtClassStart(t,r,s)}function d(n){for(var t,r=0,a=n.members;r<a.length;r++){var i=a[r];if(!e.isPropertyDeclaration(i))break;t=i}return t}function p(t,r,a,i,o,s,c,l){var u=n.createImportAdder(c,t.program,t.preferences,t.host),d=n.createMethodFromCallExpression(t,u,a,i,o,s,l),p=e.findAncestor(a,(function(n){return e.isMethodDeclaration(n)||e.isConstructorDeclaration(n)}));p&&p.parent===s?r.insertNodeAfter(c,p,d):r.insertNodeAtClassStart(c,s,d),u.writeFixes(r)}function m(n,t,r,a){var i=e.some(a.members,(function(e){var n=t.getTypeAtLocation(e);return!!(n&&402653316&n.flags)})),o=e.factory.createEnumMember(r,i?e.factory.createStringLiteral(r.text):void 0);n.replaceNode(a.getSourceFile(),a,e.factory.updateEnumDeclaration(a,a.decorators,a.modifiers,a.name,e.concatenate(a.members,e.singleElementArray(o))),{leadingTriviaOption:e.textChanges.LeadingTriviaOption.IncludeAll,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Exclude})}n.registerCodeFix({errorCodes:a,getCodeActions:function(t){var a=o(t.sourceFile,t.span.start,t.program.getTypeChecker(),t.program);if(a){if(0===a.kind){var c=a.token,d=a.parentDeclaration,f=e.textChanges.ChangeTracker.with(t,(function(e){return m(e,t.program.getTypeChecker(),c,d)}));return[n.createCodeFixAction(r,f,[e.Diagnostics.Add_missing_enum_member_0,c.text],i,e.Diagnostics.Add_all_missing_members)]}return e.concatenate(function(t,a){var o=a.parentDeclaration,s=a.declSourceFile,c=a.modifierFlags,l=a.token,u=a.call,d=a.isJSFile;if(void 0===u)return;if(e.isPrivateIdentifier(l))return;var m=l.text,f=function(n){return e.textChanges.ChangeTracker.with(t,(function(e){return p(t,e,u,m,n,o,s,d)}))},_=[n.createCodeFixAction(r,f(32&c),[32&c?e.Diagnostics.Declare_static_method_0:e.Diagnostics.Declare_method_0,m],i,e.Diagnostics.Add_all_missing_members)];8&c&&_.unshift(n.createCodeFixActionWithoutFixAll(r,f(8),[e.Diagnostics.Declare_private_method_0,m]));return _}(t,a),function(t,a){return a.isJSFile?e.singleElementArray(function(t,a){var o=a.parentDeclaration,c=a.declSourceFile,l=a.modifierFlags,u=a.token;if(e.isInterfaceDeclaration(o))return;var d=e.textChanges.ChangeTracker.with(t,(function(e){return s(e,c,o,u,!!(32&l))}));if(0===d.length)return;var p=32&l?e.Diagnostics.Initialize_static_property_0:e.isPrivateIdentifier(u)?e.Diagnostics.Declare_a_private_field_named_0:e.Diagnostics.Initialize_property_0_in_the_constructor;return n.createCodeFixAction(r,d,[p,u.text],i,e.Diagnostics.Add_all_missing_members)}(t,a)):function(t,a){var o=a.parentDeclaration,s=a.declSourceFile,c=a.modifierFlags,d=a.token,p=d.text,m=32&c,f=l(t.program.getTypeChecker(),o,d),_=function(n){return e.textChanges.ChangeTracker.with(t,(function(e){return u(e,s,o,p,f,n)}))},g=[n.createCodeFixAction(r,_(32&c),[m?e.Diagnostics.Declare_static_property_0:e.Diagnostics.Declare_property_0,p],i,e.Diagnostics.Add_all_missing_members)];if(m||e.isPrivateIdentifier(d))return g;8&c&&g.unshift(n.createCodeFixActionWithoutFixAll(r,_(8),[e.Diagnostics.Declare_private_property_0,p]));return g.push(function(t,a,i,o,s){var c=e.factory.createKeywordTypeNode(147),l=e.factory.createParameterDeclaration(void 0,void 0,void 0,"x",void 0,c,void 0),u=e.factory.createIndexSignature(void 0,void 0,[l],s),d=e.textChanges.ChangeTracker.with(t,(function(e){return e.insertNodeAtClassStart(a,i,u)}));return n.createCodeFixActionWithoutFixAll(r,d,[e.Diagnostics.Add_index_signature_for_property_0,o])}(t,s,o,d.text,f)),g}(t,a)}(t,a))}},fixIds:[i],getAllCodeActions:function(t){var r=t.program,i=r.getTypeChecker(),c=new e.Map,d=new e.Map;return n.createCombinedCodeActions(e.textChanges.ChangeTracker.with(t,(function(f){n.eachDiagnostic(t,a,(function(n){var r=o(n.file,n.start,i,t.program);if(r&&e.addToSeen(c,e.getNodeId(r.parentDeclaration)+"#"+r.token.text))if(0===r.kind){var a=r.token,s=r.parentDeclaration;m(f,i,a,s)}else{s=r.parentDeclaration;var l=r.token,u=e.getOrUpdate(d,s,(function(){return[]}));u.some((function(e){return e.token.text===l.text}))||u.push(r)}})),d.forEach((function(a,o){for(var c=n.getAllSupers(o,i),m=function(n){if(c.some((function(e){var t=d.get(e);return!!t&&t.some((function(e){return e.token.text===n.token.text}))})))return"continue";var a=n.parentDeclaration,i=n.declSourceFile,o=n.modifierFlags,m=n.token,_=n.call,g=n.isJSFile;if(_&&!e.isPrivateIdentifier(m))p(t,f,_,m.text,32&o,a,i,g);else if(g&&!e.isInterfaceDeclaration(a))s(f,i,a,m,!!(32&o));else{var y=l(r.getTypeChecker(),a,m);u(f,i,a,m.text,y,32&o)}},_=0,g=a;_<g.length;_++){m(g[_])}}))})))}}),function(e){e[e.Enum=0]="Enum",e[e.ClassOrInterface=1]="ClassOrInterface"}(t||(t={}))}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addMissingNewOperator",r=[e.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new.code];function a(n,t,r){var a=e.cast(function(n,t){var r=e.getTokenAtPosition(n,t.start),a=e.textSpanEnd(t);for(;r.end<a;)r=r.parent;return r}(t,r),e.isCallExpression),i=e.factory.createNewExpression(a.expression,a.typeArguments,a.arguments);n.replaceNode(t,a,i)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=r.sourceFile,o=r.span,s=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,i,o)}));return[n.createCodeFixAction(t,s,e.Diagnostics.Add_missing_new_operator_to_call,t,e.Diagnostics.Add_missing_new_operator_to_all_calls)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){return a(n,e.sourceFile,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="installTypesPackage",r=e.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations.code,a=[r,e.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type.code];function i(e,n){return{type:"install package",file:e,packageName:n}}function o(n,t){var r=e.cast(e.getTokenAtPosition(n,t),e.isStringLiteral).text,a=e.parsePackageName(r).packageName;return e.isExternalModuleNameRelative(a)?void 0:a}function s(n,t,a){return a===r?e.JsTyping.nodeCoreModules.has(n)?"@types/node":void 0:t.isKnownTypesPackageName(n)?e.getTypesPackageName(n):void 0}n.registerCodeFix({errorCodes:a,getCodeActions:function(r){var a=r.host,c=r.sourceFile,l=o(c,r.span.start);if(void 0!==l){var u=s(l,a,r.errorCode);return void 0===u?[]:[n.createCodeFixAction("fixCannotFindModule",[],[e.Diagnostics.Install_0,u],t,e.Diagnostics.Install_all_missing_types_packages,i(c.fileName,u))]}},fixIds:[t],getAllCodeActions:function(r){return n.codeFixAll(r,a,(function(n,a,c){var l=o(a.file,a.start);if(void 0!==l)switch(r.fixId){case t:var u=s(l,r.host,a.code);u&&c.push(i(a.file.fileName,u));break;default:e.Debug.fail("Bad fixId: "+r.fixId)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2.code,e.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1.code],r="fixClassDoesntImplementInheritedAbstractMember";function a(n,t){var r=e.getTokenAtPosition(n,t);return e.cast(r.parent,e.isClassLike)}function i(t,r,a,i,s){var c=e.getEffectiveBaseTypeNode(t),l=a.program.getTypeChecker(),u=l.getTypeAtLocation(c),d=l.getPropertiesOfType(u).filter(o),p=n.createImportAdder(r,a.program,s,a.host);n.createMissingMemberNodes(t,d,r,a,s,p,(function(e){return i.insertNodeAtClassStart(r,t,e)})),p.writeFixes(i)}function o(n){var t=e.getSyntacticModifierFlags(e.first(n.getDeclarations()));return!(8&t||!(128&t))}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var o=t.sourceFile,s=t.span,c=e.textChanges.ChangeTracker.with(t,(function(e){return i(a(o,s.start),o,t,e,t.preferences)}));return 0===c.length?void 0:[n.createCodeFixAction(r,c,e.Diagnostics.Implement_inherited_abstract_class,r,e.Diagnostics.Implement_all_inherited_abstract_classes)]},fixIds:[r],getAllCodeActions:function(r){var o=new e.Map;return n.codeFixAll(r,t,(function(n,t){var s=a(t.file,t.start);e.addToSeen(o,e.getNodeId(s))&&i(s,r.sourceFile,r,n,r.preferences)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="classSuperMustPrecedeThisAccess",r=[e.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class.code];function a(e,n,t,r){e.insertNodeAtConstructorStart(n,t,r),e.delete(n,r)}function i(n,t){var r=e.getTokenAtPosition(n,t);if(107===r.kind){var a=e.getContainingFunction(r),i=o(a.body);return i&&!i.expression.arguments.some((function(n){return e.isPropertyAccessExpression(n)&&n.expression===r}))?{constructor:a,superCall:i}:void 0}}function o(n){return e.isExpressionStatement(n)&&e.isSuperCall(n.expression)?n:e.isFunctionLike(n)?void 0:e.forEachChild(n,o)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=i(o,s.start);if(c){var l=c.constructor,u=c.superCall,d=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,o,l,u)}));return[n.createCodeFixAction(t,d,e.Diagnostics.Make_super_call_the_first_statement_in_the_constructor,t,e.Diagnostics.Make_all_super_calls_the_first_statement_in_their_constructor)]}},fixIds:[t],getAllCodeActions:function(t){var o=t.sourceFile,s=new e.Map;return n.codeFixAll(t,r,(function(n,t){var r=i(t.file,t.start);if(r){var c=r.constructor,l=r.superCall;e.addToSeen(s,e.getNodeId(c.parent))&&a(n,o,c,l)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="constructorForDerivedNeedSuperCall",r=[e.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call.code];function a(n,t){var r=e.getTokenAtPosition(n,t);return e.Debug.assert(e.isConstructorDeclaration(r.parent),"token should be at the constructor declaration"),r.parent}function i(n,t,r){var a=e.factory.createExpressionStatement(e.factory.createCallExpression(e.factory.createSuper(),void 0,e.emptyArray));n.insertNodeAtConstructorStart(t,r,a)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start),l=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c)}));return[n.createCodeFixAction(t,l,e.Diagnostics.Add_missing_super_call,t,e.Diagnostics.Add_all_missing_super_calls)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){return i(n,e.sourceFile,a(t.file,t.start))}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="enableExperimentalDecorators",r=[e.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning.code];function a(t,r){n.setJsonCompilerOptionValue(t,r,"experimentalDecorators",e.factory.createTrue())}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=r.program.getCompilerOptions().configFile;if(void 0!==i){var o=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,i)}));return[n.createCodeFixActionWithoutFixAll(t,o,e.Diagnostics.Enable_the_experimentalDecorators_option_in_your_configuration_file)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n){var t=e.program.getCompilerOptions().configFile;void 0!==t&&a(n,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixEnableJsxFlag",r=[e.Diagnostics.Cannot_use_JSX_unless_the_jsx_flag_is_provided.code];function a(t,r){n.setJsonCompilerOptionValue(t,r,"jsx",e.factory.createStringLiteral("react"))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=r.program.getCompilerOptions().configFile;if(void 0!==i){var o=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,i)}));return[n.createCodeFixActionWithoutFixAll(t,o,e.Diagnostics.Enable_the_jsx_flag_in_your_configuration_file)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n){var t=e.program.getCompilerOptions().configFile;void 0!==t&&a(n,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){n.registerCodeFix({errorCodes:[e.Diagnostics.Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher.code],getCodeActions:function(t){var r=t.program.getCompilerOptions(),a=r.configFile;if(void 0!==a){var i=[],o=e.getEmitModuleKind(r);if(o>=e.ModuleKind.ES2015&&o<e.ModuleKind.ESNext){var s=e.textChanges.ChangeTracker.with(t,(function(t){n.setJsonCompilerOptionValue(t,a,"module",e.factory.createStringLiteral("esnext"))}));i.push(n.createCodeFixActionWithoutFixAll("fixModuleOption",s,[e.Diagnostics.Set_the_module_option_in_your_configuration_file_to_0,"esnext"]))}var c=e.getEmitScriptTarget(r);if(c<4||c>99){s=e.textChanges.ChangeTracker.with(t,(function(t){if(e.getTsConfigObjectLiteralExpression(a)){var r=[["target",e.factory.createStringLiteral("es2017")]];o===e.ModuleKind.CommonJS&&r.push(["module",e.factory.createStringLiteral("commonjs")]),n.setJsonCompilerOptionValues(t,a,r)}}));i.push(n.createCodeFixActionWithoutFixAll("fixTargetOption",s,[e.Diagnostics.Set_the_target_option_in_your_configuration_file_to_0,"es2017"]))}return i.length?i:void 0}}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixPropertyAssignment",r=[e.Diagnostics.Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern.code];function a(n,t,r){n.replaceNode(t,r,e.factory.createPropertyAssignment(r.name,r.objectAssignmentInitializer))}function i(n,t){return e.cast(e.getTokenAtPosition(n,t).parent,e.isShorthandPropertyAssignment)}n.registerCodeFix({errorCodes:r,fixIds:[t],getCodeActions:function(r){var o=i(r.sourceFile,r.span.start),s=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,o)}));return[n.createCodeFixAction(t,s,[e.Diagnostics.Change_0_to_1,"=",":"],t,[e.Diagnostics.Switch_each_misused_0_to_1,"=",":"])]},getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,i(n.file,n.start))}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="extendsInterfaceBecomesImplements",r=[e.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements.code];function a(n,t){var r=e.getTokenAtPosition(n,t),a=e.getContainingClass(r).heritageClauses,i=a[0].getFirstToken();return 93===i.kind?{extendsToken:i,heritageClauses:a}:void 0}function i(n,t,r,a){if(n.replaceNode(t,r,e.factory.createToken(116)),2===a.length&&93===a[0].token&&116===a[1].token){var i=a[1].getFirstToken(),o=i.getFullStart();n.replaceRange(t,{pos:o,end:o},e.factory.createToken(27));for(var s=t.text,c=i.end;c<s.length&&e.isWhiteSpaceSingleLine(s.charCodeAt(c));)c++;n.deleteRange(t,{pos:i.getStart(),end:c})}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=a(o,r.span.start);if(s){var c=s.extendsToken,l=s.heritageClauses,u=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c,l)}));return[n.createCodeFixAction(t,u,e.Diagnostics.Change_extends_to_implements,t,e.Diagnostics.Change_all_extended_interfaces_to_implements)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t.extendsToken,t.heritageClauses)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="forgottenThisPropertyAccess",r=e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,a=[e.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,e.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies.code,r];function i(n,t,a){var i=e.getTokenAtPosition(n,t);if(e.isIdentifier(i))return{node:i,className:a===r?e.getContainingClass(i).name.text:void 0}}function o(n,t,r){var a=r.node,i=r.className;e.suppressLeadingAndTrailingTrivia(a),n.replaceNode(t,a,e.factory.createPropertyAccessExpression(i?e.factory.createIdentifier(i):e.factory.createThis(),a))}n.registerCodeFix({errorCodes:a,getCodeActions:function(r){var a=r.sourceFile,s=i(a,r.span.start,r.errorCode);if(s){var c=e.textChanges.ChangeTracker.with(r,(function(e){return o(e,a,s)}));return[n.createCodeFixAction(t,c,[e.Diagnostics.Add_0_to_unresolved_variable,s.className||"this"],t,e.Diagnostics.Add_qualifier_to_all_unresolved_variables_matching_a_member_name)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,a,(function(n,t){var r=i(t.file,t.start,t.code);r&&o(n,e.sourceFile,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixInvalidJsxCharacters_expression",r="fixInvalidJsxCharacters_htmlEntity",a=[e.Diagnostics.Unexpected_token_Did_you_mean_or_gt.code,e.Diagnostics.Unexpected_token_Did_you_mean_or_rbrace.code];n.registerCodeFix({errorCodes:a,fixIds:[t,r],getCodeActions:function(a){var i=a.sourceFile,s=a.preferences,c=a.span,l=e.textChanges.ChangeTracker.with(a,(function(e){return o(e,s,i,c.start,!1)})),u=e.textChanges.ChangeTracker.with(a,(function(e){return o(e,s,i,c.start,!0)}));return[n.createCodeFixAction(t,l,e.Diagnostics.Wrap_invalid_character_in_an_expression_container,t,e.Diagnostics.Wrap_all_invalid_characters_in_an_expression_container),n.createCodeFixAction(r,u,e.Diagnostics.Convert_invalid_character_to_its_html_entity_code,r,e.Diagnostics.Convert_all_invalid_characters_to_HTML_entity_code)]},getAllCodeActions:function(e){return n.codeFixAll(e,a,(function(n,t){return o(n,e.preferences,t.file,t.start,e.fixId===r)}))}});var i={">":"&gt;","}":"&rbrace;"};function o(n,t,r,a,o){var s=r.getText()[a];if(function(n){return e.hasProperty(i,n)}(s)){var c=o?i[s]:"{"+e.quote(r,t,s)+"}";n.replaceRangeWithText(r,{pos:a,end:a+1},c)}}}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="unusedIdentifier",r="unusedIdentifier_prefix",a="unusedIdentifier_delete",i="unusedIdentifier_infer",o=[e.Diagnostics._0_is_declared_but_its_value_is_never_read.code,e.Diagnostics._0_is_declared_but_never_used.code,e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code,e.Diagnostics.All_imports_in_import_declaration_are_unused.code,e.Diagnostics.All_destructured_elements_are_unused.code,e.Diagnostics.All_variables_are_unused.code,e.Diagnostics.All_type_parameters_are_unused.code];function s(n,t,r){n.replaceNode(t,r.parent,e.factory.createKeywordTypeNode(152))}function c(r,i){return n.createCodeFixAction(t,r,i,a,e.Diagnostics.Delete_all_unused_declarations)}function l(n,t,r){n.delete(t,e.Debug.checkDefined(e.cast(r.parent,e.isDeclarationWithTypeParameterChildren).typeParameters,"The type parameter to delete should exist"))}function u(n){return 99===n.kind?e.tryCast(n.parent,e.isImportDeclaration):void 0}function d(n,t){return e.isVariableDeclarationList(t.parent)&&e.first(t.parent.getChildren(n))===t}function p(e,n,t){e.delete(n,232===t.parent.kind?t.parent:t)}function m(n,t,r){e.forEach(r.elements,(function(e){return n.delete(t,e)}))}function f(n,t,r,a){t!==e.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code&&(135===a.kind&&(a=e.cast(a.parent,e.isInferTypeNode).typeParameter.name),e.isIdentifier(a)&&function(e){switch(e.parent.kind){case 160:case 159:return!0;case 249:switch(e.parent.parent.parent.kind){case 239:case 238:return!0}}return!1}(a)&&(n.replaceNode(r,a,e.factory.createIdentifier("_"+a.text)),e.isParameter(a.parent)&&e.getJSDocParameterTags(a.parent).forEach((function(t){e.isIdentifier(t.name)&&n.replaceNode(r,t.name,e.factory.createIdentifier("_"+t.name.text))}))))}function _(n,t,r,a,i,o){!function(n,t,r,a,i,o){var s=n.parent;e.isParameter(s)?function(n,t,r,a,i,o){void 0===o&&(o=!1);(function(n,t,r,a){var i=r.parent;switch(i.kind){case 165:var o=n.getSymbolAtLocation(i.name);if(e.isMemberSymbolInBaseType(o,n))return!1;case 166:return!0;case 251:return!i.name||!function(n,t,r){return!!e.FindAllReferences.Core.eachSymbolReferenceInFile(r,n,t,(function(n){return e.isIdentifier(n)&&e.isCallExpression(n.parent)&&n.parent.arguments.indexOf(n)>=0}))}(n,t,i.name)||g(i,r,a);case 208:case 209:return g(i,r,a);case 168:return!1;default:return e.Debug.failBadSyntaxKind(i)}})(a,t,r,o)&&(r.modifiers&&r.modifiers.length>0&&(!e.isIdentifier(r.name)||e.FindAllReferences.Core.isSymbolReferencedInFile(r.name,a,t))?r.modifiers.forEach((function(e){return n.deleteModifier(t,e)})):(n.delete(t,r),function(n,t,r,a,i){e.FindAllReferences.Core.eachSignatureCall(r.parent,a,i,(function(e){var a=r.parent.parameters.indexOf(r);e.arguments.length>a&&n.delete(t,e.arguments[a])}))}(n,t,r,i,a)))}(t,r,s,a,i,o):t.delete(r,e.isImportClause(s)?n:e.isComputedPropertyName(s)?s.parent:s)}(t,r,n,a,i,o),e.isIdentifier(t)&&function(n,t,r,a){e.FindAllReferences.Core.eachSymbolReferenceInFile(r,a,t,(function(r){e.isPropertyAccessExpression(r.parent)&&r.parent.name===r&&(r=r.parent),e.isBinaryExpression(r.parent)&&e.isExpressionStatement(r.parent.parent)&&r.parent.left===r&&n.delete(t,r.parent.parent)}))}(r,n,t,a)}function g(n,t,r){var a=n.parameters,i=a.indexOf(t);return e.Debug.assert(-1!==i,"The parameter should already be in the list"),r?a.slice(i+1).every((function(n){return e.isIdentifier(n.name)&&!n.symbol.isReferenced})):i===a.length-1}n.registerCodeFix({errorCodes:o,getCodeActions:function(a){var o=a.errorCode,g=a.sourceFile,y=a.program,v=y.getTypeChecker(),h=y.getSourceFiles(),b=e.getTokenAtPosition(g,a.span.start);if(e.isJSDocTemplateTag(b))return[c(e.textChanges.ChangeTracker.with(a,(function(e){return e.delete(g,b)})),e.Diagnostics.Remove_template_tag)];if(29===b.kind)return[c(L=e.textChanges.ChangeTracker.with(a,(function(e){return l(e,g,b)})),e.Diagnostics.Remove_type_parameters)];var E=u(b);if(E)return[c(L=e.textChanges.ChangeTracker.with(a,(function(e){return e.delete(g,E)})),[e.Diagnostics.Remove_import_from_0,e.showModuleSpecifier(E)])];if(e.isObjectBindingPattern(b.parent)){if(e.isParameter(b.parent.parent)){var T=b.parent.elements,S=[T.length>1?e.Diagnostics.Remove_unused_declarations_for_Colon_0:e.Diagnostics.Remove_unused_declaration_for_Colon_0,e.map(T,(function(e){return e.getText(g)})).join(", ")];return[c(e.textChanges.ChangeTracker.with(a,(function(e){return m(e,g,b.parent)})),S)]}return[c(e.textChanges.ChangeTracker.with(a,(function(e){return e.delete(g,b.parent.parent)})),e.Diagnostics.Remove_unused_destructuring_declaration)]}if(d(g,b))return[c(e.textChanges.ChangeTracker.with(a,(function(e){return p(e,g,b.parent)})),e.Diagnostics.Remove_variable_statement)];var x=[];if(135===b.kind){var L=e.textChanges.ChangeTracker.with(a,(function(e){return s(e,g,b)})),A=e.cast(b.parent,e.isInferTypeNode).typeParameter.name.text;x.push(n.createCodeFixAction(t,L,[e.Diagnostics.Replace_infer_0_with_unknown,A],i,e.Diagnostics.Replace_all_unused_infer_with_unknown))}else{var D=e.textChanges.ChangeTracker.with(a,(function(e){return _(g,b,e,v,h,!1)}));if(D.length){A=e.isComputedPropertyName(b.parent)?b.parent:b;x.push(c(D,[e.Diagnostics.Remove_unused_declaration_for_Colon_0,A.getText(g)]))}}var C=e.textChanges.ChangeTracker.with(a,(function(e){return f(e,o,g,b)}));return C.length&&x.push(n.createCodeFixAction(t,C,[e.Diagnostics.Prefix_0_with_an_underscore,b.getText(g)],r,e.Diagnostics.Prefix_all_unused_declarations_with_where_possible)),x},fixIds:[r,a,i],getAllCodeActions:function(t){var c=t.sourceFile,g=t.program,y=g.getTypeChecker(),v=g.getSourceFiles();return n.codeFixAll(t,o,(function(n,o){var g=e.getTokenAtPosition(c,o.start);switch(t.fixId){case r:f(n,o.code,c,g);break;case a:if(135===g.kind)break;var h=u(g);h?n.delete(c,h):e.isJSDocTemplateTag(g)?n.delete(c,g):29===g.kind?l(n,c,g):e.isObjectBindingPattern(g.parent)?e.isParameter(g.parent.parent)?m(n,c,g.parent):n.delete(c,g.parent.parent):d(c,g)?p(n,c,g.parent):_(c,g,n,y,v,!0);break;case i:135===g.kind&&s(n,c,g);break;default:e.Debug.fail(JSON.stringify(t.fixId))}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixUnreachableCode",r=[e.Diagnostics.Unreachable_code_detected.code];function a(n,t,r,a,i){var o=e.getTokenAtPosition(t,r),s=e.findAncestor(o,e.isStatement);if(s.getStart(t)!==o.getStart(t)){var c=JSON.stringify({statementKind:e.Debug.formatSyntaxKind(s.kind),tokenKind:e.Debug.formatSyntaxKind(o.kind),errorCode:i,start:r,length:a});e.Debug.fail("Token and statement should start at the same point. "+c)}var l=(e.isBlock(s.parent)?s.parent:s).parent;if(!e.isBlock(s.parent)||s===e.first(s.parent.statements))switch(l.kind){case 234:if(l.elseStatement){if(e.isBlock(s.parent))break;return void n.replaceNode(t,s,e.factory.createBlock(e.emptyArray))}case 236:case 237:return void n.delete(t,l)}if(e.isBlock(s.parent)){var u=r+a,d=e.Debug.checkDefined(function(e,n){for(var t,r=0,a=e;r<a.length;r++){var i=a[r];if(!n(i))break;t=i}return t}(e.sliceAfter(s.parent.statements,s),(function(e){return e.pos<u})),"Some statement should be last");n.deleteNodeRange(t,s,d)}else n.delete(t,s)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start,r.span.length,r.errorCode)}));return[n.createCodeFixAction(t,i,e.Diagnostics.Remove_unreachable_code,t,e.Diagnostics.Remove_all_unreachable_code)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n.start,n.length,n.code)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixUnusedLabel",r=[e.Diagnostics.Unused_label.code];function a(n,t,r){var a=e.getTokenAtPosition(t,r),i=e.cast(a.parent,e.isLabeledStatement),o=a.getStart(t),s=i.statement.getStart(t),c=e.positionsAreOnSameLine(o,s,t)?s:e.skipTrivia(t.text,e.findChildOfKind(i,58,t).end,!0);n.deleteRange(t,{pos:o,end:c})}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span.start)}));return[n.createCodeFixAction(t,i,e.Diagnostics.Remove_unused_label,t,e.Diagnostics.Remove_all_unused_labels)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n.start)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixJSDocTypes_plain",r="fixJSDocTypes_nullable",a=[e.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments.code];function i(e,n,t,r,a){e.replaceNode(n,t,a.typeToTypeNode(r,t,void 0))}function o(n,t,r){var a=e.findAncestor(e.getTokenAtPosition(n,t),s),i=a&&a.type;return i&&{typeNode:i,type:r.getTypeFromTypeNode(i)}}function s(e){switch(e.kind){case 224:case 169:case 170:case 251:case 167:case 171:case 190:case 165:case 164:case 160:case 163:case 162:case 168:case 254:case 206:case 249:return!0;default:return!1}}n.registerCodeFix({errorCodes:a,getCodeActions:function(a){var s=a.sourceFile,c=a.program.getTypeChecker(),l=o(s,a.span.start,c);if(l){var u=l.typeNode,d=l.type,p=u.getText(s),m=[f(d,t,e.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript)];return 305===u.kind&&m.push(f(c.getNullableType(d,32768),r,e.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types)),m}function f(t,r,o){var l=e.textChanges.ChangeTracker.with(a,(function(e){return i(e,s,u,t,c)}));return n.createCodeFixAction("jdocTypes",l,[e.Diagnostics.Change_0_to_1,p,c.typeToString(t)],r,o)}},fixIds:[t,r],getAllCodeActions:function(e){var t=e.fixId,s=e.program,c=e.sourceFile,l=s.getTypeChecker();return n.codeFixAll(e,a,(function(e,n){var a=o(n.file,n.start,l);if(a){var s=a.typeNode,u=a.type,d=305===s.kind&&t===r?l.getNullableType(u,32768):u;i(e,c,s,d,l)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixMissingCallParentheses",r=[e.Diagnostics.This_condition_will_always_return_true_since_the_function_is_always_defined_Did_you_mean_to_call_it_instead.code];function a(e,n,t){e.replaceNodeWithText(n,t,t.text+"()")}function i(n,t){var r=e.getTokenAtPosition(n,t);if(e.isPropertyAccessExpression(r.parent)){for(var a=r.parent;e.isPropertyAccessExpression(a.parent);)a=a.parent;return a.name}if(e.isIdentifier(r))return r}n.registerCodeFix({errorCodes:r,fixIds:[t],getCodeActions:function(r){var o=i(r.sourceFile,r.span.start);if(o){var s=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,o)}));return[n.createCodeFixAction(t,s,e.Diagnostics.Add_missing_call_parentheses,t,e.Diagnostics.Add_all_missing_call_parentheses)]}},getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=i(n.file,n.start);t&&a(e,n.file,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixAwaitInSyncFunction",r=[e.Diagnostics.await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules.code,e.Diagnostics.A_for_await_of_statement_is_only_allowed_within_an_async_function_or_async_generator.code];function a(n,t){var r=e.getTokenAtPosition(n,t),a=e.getContainingFunction(r);if(a){var i,o;switch(a.kind){case 165:i=a.name;break;case 251:case 208:i=e.findChildOfKind(a,97,n);break;case 209:i=e.findChildOfKind(a,20,n)||e.first(a.parameters);break;default:return}return i&&{insertBefore:i,returnType:(o=a,o.type?o.type:e.isVariableDeclaration(o.parent)&&o.parent.type&&e.isFunctionTypeNode(o.parent.type)?o.parent.type.type:void 0)}}}function i(n,t,r){var a=r.insertBefore,i=r.returnType;if(i){var o=e.getEntityNameFromTypeNode(i);o&&78===o.kind&&"Promise"===o.text||n.replaceNode(t,i,e.factory.createTypeReferenceNode("Promise",e.factory.createNodeArray([i])))}n.insertModifierBefore(t,129,a)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start);if(c){var l=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c)}));return[n.createCodeFixAction(t,l,e.Diagnostics.Add_async_modifier_to_containing_function,t,e.Diagnostics.Add_all_missing_async_modifiers)]}},fixIds:[t],getAllCodeActions:function(t){var o=new e.Map;return n.codeFixAll(t,r,(function(n,r){var s=a(r.file,r.start);s&&e.addToSeen(o,e.getNodeId(s.insertBefore))&&i(n,t.sourceFile,s)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property.code,e.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor.code],r="fixPropertyOverrideAccessor";function a(t,r,a,i,o){var s,c;if(i===e.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property.code)s=r,c=r+a;else if(i===e.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor.code){var l=o.program.getTypeChecker(),u=e.getTokenAtPosition(t,r).parent;e.Debug.assert(e.isAccessor(u),"error span of fixPropertyOverrideAccessor should only be on an accessor");var d=u.parent;e.Debug.assert(e.isClassLike(d),"erroneous accessors should only be inside classes");var p=e.singleOrUndefined(n.getAllSupers(d,l));if(!p)return[];var m=e.unescapeLeadingUnderscores(e.getTextOfPropertyName(u.name)),f=l.getPropertyOfType(l.getTypeAtLocation(p),m);if(!f||!f.valueDeclaration)return[];s=f.valueDeclaration.pos,c=f.valueDeclaration.end,t=e.getSourceFileOfNode(f.valueDeclaration)}else e.Debug.fail("fixPropertyOverrideAccessor codefix got unexpected error code "+i);return n.generateAccessorFromProperty(t,o.program,s,c,o,e.Diagnostics.Generate_get_and_set_accessors.message)}n.registerCodeFix({errorCodes:t,getCodeActions:function(t){var i=a(t.sourceFile,t.span.start,t.span.length,t.errorCode,t);if(i)return[n.createCodeFixAction(r,i,e.Diagnostics.Generate_get_and_set_accessors,r,e.Diagnostics.Generate_get_and_set_accessors_for_all_overriding_properties)]},fixIds:[r],getAllCodeActions:function(e){return n.codeFixAll(e,t,(function(n,t){var r=a(t.file,t.start,t.length,t.code,e);if(r)for(var i=0,o=r;i<o.length;i++){var s=o[i];n.pushRaw(e.sourceFile,s)}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="inferFromUsage",r=[e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code,e.Diagnostics.Variable_0_implicitly_has_an_1_type.code,e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code,e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code,e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code,e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code,e.Diagnostics.Member_0_implicitly_has_an_1_type.code,e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code,e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code,e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code];function a(n,t){switch(n){case e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:case e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.isSetAccessorDeclaration(e.getContainingFunction(t))?e.Diagnostics.Infer_type_of_0_from_usage:e.Diagnostics.Infer_parameter_types_from_usage;case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Infer_parameter_types_from_usage;case e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code:return e.Diagnostics.Infer_this_type_of_0_from_usage;default:return e.Diagnostics.Infer_type_of_0_from_usage}}function i(t,r,a,i,d,_,g,y,v){if(e.isParameterPropertyModifier(a.kind)||78===a.kind||25===a.kind||107===a.kind){var h=a.parent,b=n.createImportAdder(r,d,v,y);switch(i=function(n){switch(n){case e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code;case e.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Variable_0_implicitly_has_an_1_type.code;case e.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code;case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code;case e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code:return e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code;case e.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code;case e.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code:return e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code;case e.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:return e.Diagnostics.Member_0_implicitly_has_an_1_type.code}return n}(i)){case e.Diagnostics.Member_0_implicitly_has_an_1_type.code:case e.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code:if(e.isVariableDeclaration(h)&&g(h)||e.isPropertyDeclaration(h)||e.isPropertySignature(h))return o(t,b,r,h,d,y,_),b.writeFixes(t),h;if(e.isPropertyAccessExpression(h)){var E=p(h.name,d,_),T=e.getTypeNodeIfAccessible(E,h,d,y);if(T){var S=e.factory.createJSDocTypeTag(void 0,e.factory.createJSDocTypeExpression(T),"");u(t,r,e.cast(h.parent.parent,e.isExpressionStatement),[S])}return b.writeFixes(t),h}return;case e.Diagnostics.Variable_0_implicitly_has_an_1_type.code:var x=d.getTypeChecker().getSymbolAtLocation(a);return x&&x.valueDeclaration&&e.isVariableDeclaration(x.valueDeclaration)&&g(x.valueDeclaration)?(o(t,b,r,x.valueDeclaration,d,y,_),b.writeFixes(t),x.valueDeclaration):void 0}var L=e.getContainingFunction(a);if(void 0!==L){var A;switch(i){case e.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:if(e.isSetAccessorDeclaration(L)){s(t,b,r,L,d,y,_),A=L;break}case e.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:if(g(L)){var D=e.cast(h,e.isParameter);!function(n,t,r,a,i,o,s,u){if(!e.isIdentifier(a.name))return;var d=function(n,t,r,a){var i=m(n,t,r,a);return i&&f(r,i,a).parameters(n)||n.parameters.map((function(n){return{declaration:n,type:e.isIdentifier(n.name)?p(n.name,r,a):r.getTypeChecker().getAnyType()}}))}(i,r,o,u);if(e.Debug.assert(i.parameters.length===d.length,"Parameter count and inference count should match"),e.isInJSFile(i))l(n,r,d,o,s);else{var _=e.isArrowFunction(i)&&!e.findChildOfKind(i,20,r);_&&n.insertNodeBefore(r,e.first(i.parameters),e.factory.createToken(20));for(var g=0,y=d;g<y.length;g++){var v=y[g],h=v.declaration,b=v.type;!h||h.type||h.initializer||c(n,t,r,h,b,o,s)}_&&n.insertNodeAfter(r,e.last(i.parameters),e.factory.createToken(21))}}(t,b,r,D,L,d,y,_),A=D}break;case e.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code:case e.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code:e.isGetAccessorDeclaration(L)&&e.isIdentifier(L.name)&&(c(t,b,r,L,p(L.name,d,_),d,y),A=L);break;case e.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code:e.isSetAccessorDeclaration(L)&&(s(t,b,r,L,d,y,_),A=L);break;case e.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code:e.textChanges.isThisTypeAnnotatable(L)&&g(L)&&(!function(n,t,r,a,i,o){var s=m(r,t,a,o);if(!s||!s.length)return;var c=f(a,s,o).thisParameter(),l=e.getTypeNodeIfAccessible(c,r,a,i);if(!l)return;e.isInJSFile(r)?function(n,t,r,a){u(n,t,r,[e.factory.createJSDocThisTag(void 0,e.factory.createJSDocTypeExpression(a))])}(n,t,r,l):n.tryInsertThisTypeAnnotation(t,r,l)}(t,r,L,d,y,_),A=L);break;default:return e.Debug.fail(String(i))}return b.writeFixes(t),A}}}function o(n,t,r,a,i,o,s){e.isIdentifier(a.name)&&c(n,t,r,a,p(a.name,i,s),i,o)}function s(n,t,r,a,i,o,s){var u=e.firstOrUndefined(a.parameters);if(u&&e.isIdentifier(a.name)&&e.isIdentifier(u.name)){var d=p(a.name,i,s);d===i.getTypeChecker().getAnyType()&&(d=p(u.name,i,s)),e.isInJSFile(a)?l(n,r,[{declaration:u,type:d}],i,o):c(n,t,r,u,d,i,o)}}function c(t,r,a,i,o,s,c){var l=e.getTypeNodeIfAccessible(o,i,s,c);if(l)if(e.isInJSFile(a)&&162!==i.kind){var d=e.isVariableDeclaration(i)?e.tryCast(i.parent.parent,e.isVariableStatement):i;if(!d)return;var p=e.factory.createJSDocTypeExpression(l);u(t,a,d,[e.isGetAccessorDeclaration(i)?e.factory.createJSDocReturnTag(void 0,p,""):e.factory.createJSDocTypeTag(void 0,p,"")])}else(function(t,r,a,i,o,s){var c=n.tryGetAutoImportableReferenceFromTypeNode(t,s);if(c&&i.tryInsertTypeAnnotation(a,r,c.typeNode))return e.forEach(c.symbols,(function(e){return o.addImportFromExportedSymbol(e,!0)})),!0;return!1})(l,i,a,t,r,e.getEmitScriptTarget(s.getCompilerOptions()))||t.tryInsertTypeAnnotation(a,i,l)}function l(n,t,r,a,i){var o=r.length&&r[0].declaration.parent;o&&u(n,t,o,e.mapDefined(r,(function(n){var t=n.declaration;if(!t.initializer&&!e.getJSDocType(t)&&e.isIdentifier(t.name)){var r=n.type&&e.getTypeNodeIfAccessible(n.type,t,a,i),o=e.factory.cloneNode(t.name);return e.setEmitFlags(o,3584),r&&e.factory.createJSDocParameterTag(void 0,o,!!n.isOptional,e.factory.createJSDocTypeExpression(r),!1,"")}})))}function u(n,t,r,a){var i=e.mapDefined(r.jsDoc,(function(e){return e.comment})),o=e.flatMapToMutable(r.jsDoc,(function(e){return e.tags})),s=a.filter((function(n){return!o||!o.some((function(t,r){var a=function(n,t){if(n.kind!==t.kind)return;switch(n.kind){case 326:var r=n,a=t;return e.isIdentifier(r.name)&&e.isIdentifier(a.name)&&r.name.escapedText===a.name.escapedText?e.factory.createJSDocParameterTag(void 0,a.name,!1,a.typeExpression,a.isNameFirst,r.comment):void 0;case 327:return e.factory.createJSDocReturnTag(void 0,t.typeExpression,n.comment)}}(t,n);return a&&(o[r]=a),!!a}))})),c=e.factory.createJSDocComment(i.join("\n"),e.factory.createNodeArray(__spreadArrays(o||e.emptyArray,s))),l=209===r.kind?function(e){if(163===e.parent.kind)return e.parent;return e.parent.parent}(r):r;l.jsDoc=r.jsDoc,l.jsDocCache=r.jsDocCache,n.insertJsdocCommentBefore(t,l,c)}function d(n,t,r){return e.mapDefined(e.FindAllReferences.getReferenceEntriesForNode(-1,n,t,t.getSourceFiles(),r),(function(n){return 0!==n.kind?e.tryCast(n.node,e.isIdentifier):void 0}))}function p(e,n,t){return f(n,d(e,n,t),t).single()}function m(n,t,r,a){var i;switch(n.kind){case 166:i=e.findChildOfKind(n,132,t);break;case 209:case 208:var o=n.parent;i=e.isVariableDeclaration(o)&&e.isIdentifier(o.name)?o.name:n.name;break;case 251:case 165:i=n.name}if(i)return d(i,r,a)}function f(n,t,r){var a=n.getTypeChecker(),i={string:function(){return a.getStringType()},number:function(){return a.getNumberType()},Array:function(e){return a.createArrayType(e)},Promise:function(e){return a.createPromiseType(e)}},o=[a.getStringType(),a.getNumberType(),a.createArrayType(a.getAnyType()),a.createPromiseType(a.getAnyType())];return{single:function(){return f(l(t))},parameters:function(i){if(0===t.length||!i.parameters)return;for(var o=s(),c=0,p=t;c<p.length;c++){var m=p[c];r.throwIfCancellationRequested(),u(m,o)}var _=__spreadArrays(o.constructs||[],o.calls||[]);return i.parameters.map((function(t,o){for(var s=[],c=e.isRestParameter(t),u=!1,p=0,m=_;p<m.length;p++){var g=m[p];if(g.argumentTypes.length<=o)u=e.isInJSFile(i),s.push(a.getUndefinedType());else if(c)for(var y=o;y<g.argumentTypes.length;y++)s.push(a.getBaseTypeOfLiteralType(g.argumentTypes[y]));else s.push(a.getBaseTypeOfLiteralType(g.argumentTypes[o]))}if(e.isIdentifier(t.name)){var v=l(d(t.name,n,r));s.push.apply(s,c?e.mapDefined(v,a.getElementTypeOfArrayType):v)}var h=f(s);return{type:c?a.createArrayType(h):h,isOptional:u&&!c,declaration:t}}))},thisParameter:function(){for(var n=s(),a=0,i=t;a<i.length;a++){var o=i[a];r.throwIfCancellationRequested(),u(o,n)}return f(n.candidateThisTypes||e.emptyArray)}};function s(){return{isNumber:void 0,isString:void 0,isNumberOrString:void 0,candidateTypes:void 0,properties:void 0,calls:void 0,constructs:void 0,numberIndex:void 0,stringIndex:void 0,candidateThisTypes:void 0,inferredTypes:void 0}}function c(n){for(var t=new e.Map,r=0,a=n;r<a.length;r++){var i=a[r];i.properties&&i.properties.forEach((function(e,n){t.has(n)||t.set(n,[]),t.get(n).push(e)}))}var o=new e.Map;return t.forEach((function(e,n){o.set(n,c(e))})),{isNumber:n.some((function(e){return e.isNumber})),isString:n.some((function(e){return e.isString})),isNumberOrString:n.some((function(e){return e.isNumberOrString})),candidateTypes:e.flatMap(n,(function(e){return e.candidateTypes})),properties:o,calls:e.flatMap(n,(function(e){return e.calls})),constructs:e.flatMap(n,(function(e){return e.constructs})),numberIndex:e.forEach(n,(function(e){return e.numberIndex})),stringIndex:e.forEach(n,(function(e){return e.stringIndex})),candidateThisTypes:e.flatMap(n,(function(e){return e.candidateThisTypes})),inferredTypes:void 0}}function l(e){for(var n={isNumber:void 0,isString:void 0,isNumberOrString:void 0,candidateTypes:void 0,properties:void 0,calls:void 0,constructs:void 0,numberIndex:void 0,stringIndex:void 0,candidateThisTypes:void 0,inferredTypes:void 0},t=0,a=e;t<a.length;t++){var i=a[t];r.throwIfCancellationRequested(),u(i,n)}return _(n)}function u(n,t){for(;e.isRightSideOfQualifiedNameOrPropertyAccess(n);)n=n.parent;switch(n.parent.kind){case 233:!function(n,t){v(t,e.isCallExpression(n)?a.getVoidType():a.getAnyType())}(n,t);break;case 215:t.isNumber=!0;break;case 214:!function(e,n){switch(e.operator){case 45:case 46:case 40:case 54:n.isNumber=!0;break;case 39:n.isNumberOrString=!0}}(n.parent,t);break;case 216:!function(n,t,r){switch(t.operatorToken.kind){case 42:case 41:case 43:case 44:case 47:case 48:case 49:case 50:case 51:case 52:case 64:case 66:case 65:case 67:case 68:case 72:case 73:case 77:case 69:case 71:case 70:case 40:case 29:case 32:case 31:case 33:var i=a.getTypeAtLocation(t.left===n?t.right:t.left);1056&i.flags?v(r,i):r.isNumber=!0;break;case 63:case 39:var o=a.getTypeAtLocation(t.left===n?t.right:t.left);1056&o.flags?v(r,o):296&o.flags?r.isNumber=!0:402653316&o.flags?r.isString=!0:1&o.flags||(r.isNumberOrString=!0);break;case 62:case 34:case 36:case 37:case 35:v(r,a.getTypeAtLocation(t.left===n?t.right:t.left));break;case 100:n===t.left&&(r.isString=!0);break;case 56:case 60:n!==t.left||249!==n.parent.parent.kind&&!e.isAssignmentExpression(n.parent.parent,!0)||v(r,a.getTypeAtLocation(t.right))}}(n,n.parent,t);break;case 284:case 285:!function(e,n){v(n,a.getTypeAtLocation(e.parent.parent.expression))}(n.parent,t);break;case 203:case 204:n.parent.expression===n?function(e,n){var t={argumentTypes:[],return_:{isNumber:void 0,isString:void 0,isNumberOrString:void 0,candidateTypes:void 0,properties:void 0,calls:void 0,constructs:void 0,numberIndex:void 0,stringIndex:void 0,candidateThisTypes:void 0,inferredTypes:void 0}};if(e.arguments)for(var r=0,i=e.arguments;r<i.length;r++){var o=i[r];t.argumentTypes.push(a.getTypeAtLocation(o))}u(e,t.return_),203===e.kind?(n.calls||(n.calls=[])).push(t):(n.constructs||(n.constructs=[])).push(t)}(n.parent,t):p(n,t);break;case 201:!function(n,t){var r=e.escapeLeadingUnderscores(n.name.text);t.properties||(t.properties=new e.Map);var a=t.properties.get(r)||{isNumber:void 0,isString:void 0,isNumberOrString:void 0,candidateTypes:void 0,properties:void 0,calls:void 0,constructs:void 0,numberIndex:void 0,stringIndex:void 0,candidateThisTypes:void 0,inferredTypes:void 0};u(n,a),t.properties.set(r,a)}(n.parent,t);break;case 202:!function(e,n,t){if(n===e.argumentExpression)return void(t.isNumberOrString=!0);var r=a.getTypeAtLocation(e.argumentExpression),i={isNumber:void 0,isString:void 0,isNumberOrString:void 0,candidateTypes:void 0,properties:void 0,calls:void 0,constructs:void 0,numberIndex:void 0,stringIndex:void 0,candidateThisTypes:void 0,inferredTypes:void 0};u(e,i),296&r.flags?t.numberIndex=i:t.stringIndex=i}(n.parent,n,t);break;case 288:case 289:!function(n,t){var r=e.isVariableDeclaration(n.parent.parent)?n.parent.parent:n.parent;h(t,a.getTypeAtLocation(r))}(n.parent,t);break;case 163:!function(e,n){h(n,a.getTypeAtLocation(e.parent))}(n.parent,t);break;case 249:var r=n.parent,i=r.name,o=r.initializer;if(n===i){o&&v(t,a.getTypeAtLocation(o));break}default:return p(n,t)}}function p(n,t){e.isExpressionNode(n)&&v(t,a.getContextualType(n))}function m(e){return f(_(e))}function f(n){if(!n.length)return a.getAnyType();var t=a.getUnionType([a.getStringType(),a.getNumberType()]),r=function(n,t){for(var r=[],a=0,i=n;a<i.length;a++)for(var o=i[a],s=0,c=t;s<c.length;s++){var l=c[s],u=l.high,d=l.low;u(o)&&(e.Debug.assert(!d(o),"Priority can't have both low and high"),r.push(d))}return n.filter((function(e){return r.every((function(n){return!n(e)}))}))}(n,[{high:function(e){return e===a.getStringType()||e===a.getNumberType()},low:function(e){return e===t}},{high:function(e){return!(16385&e.flags)},low:function(e){return!!(16385&e.flags)}},{high:function(n){return!(114689&n.flags||16&e.getObjectFlags(n))},low:function(n){return!!(16&e.getObjectFlags(n))}}]),i=r.filter((function(n){return 16&e.getObjectFlags(n)}));return i.length&&(r=r.filter((function(n){return!(16&e.getObjectFlags(n))}))).push(function(n){if(1===n.length)return n[0];for(var t=[],r=[],i=[],o=[],s=!1,c=!1,l=e.createMultiMap(),u=0,d=n;u<d.length;u++){for(var p=d[u],m=0,f=a.getPropertiesOfType(p);m<f.length;m++){var _=f[m];l.add(_.name,a.getTypeOfSymbolAtLocation(_,_.valueDeclaration))}t.push.apply(t,a.getSignaturesOfType(p,0)),r.push.apply(r,a.getSignaturesOfType(p,1)),p.stringIndexInfo&&(i.push(p.stringIndexInfo.type),s=s||p.stringIndexInfo.isReadonly),p.numberIndexInfo&&(o.push(p.numberIndexInfo.type),c=c||p.numberIndexInfo.isReadonly)}var g=e.mapEntries(l,(function(e,t){var r=t.length<n.length?16777216:0,i=a.createSymbol(4|r,e);return i.type=a.getUnionType(t),[e,i]}));return a.createAnonymousType(n[0].symbol,g,t,r,i.length?a.createIndexInfo(a.getUnionType(i),s):void 0,o.length?a.createIndexInfo(a.getUnionType(o),c):void 0)}(i)),a.getWidenedType(a.getUnionType(r.map(a.getBaseTypeOfLiteralType),2))}function _(n){var t,r,s,c=[];return n.isNumber&&c.push(a.getNumberType()),n.isString&&c.push(a.getStringType()),n.isNumberOrString&&c.push(a.getUnionType([a.getStringType(),a.getNumberType()])),n.numberIndex&&c.push(a.createArrayType(m(n.numberIndex))),((null===(t=n.properties)||void 0===t?void 0:t.size)||(null===(r=n.calls)||void 0===r?void 0:r.length)||(null===(s=n.constructs)||void 0===s?void 0:s.length)||n.stringIndex)&&c.push(function(n){var t=new e.Map;n.properties&&n.properties.forEach((function(e,n){var r=a.createSymbol(4,n);r.type=m(e),t.set(n,r)}));var r=n.calls?[y(n.calls)]:[],i=n.constructs?[y(n.constructs)]:[],o=n.stringIndex&&a.createIndexInfo(m(n.stringIndex),!1);return a.createAnonymousType(void 0,t,r,i,o,void 0)}(n)),c.push.apply(c,(n.candidateTypes||[]).map((function(e){return a.getBaseTypeOfLiteralType(e)}))),c.push.apply(c,function(n){if(!n.properties||!n.properties.size)return[];var t=o.filter((function(t){return function(n,t){return!!t.properties&&!e.forEachEntry(t.properties,(function(t,r){var i,o=a.getTypeOfPropertyOfType(n,r);return!o||(t.calls?!a.getSignaturesOfType(o,0).length||!a.isTypeAssignableTo(o,(i=t.calls,a.createAnonymousType(void 0,e.createSymbolTable(),[y(i)],e.emptyArray,void 0,void 0))):!a.isTypeAssignableTo(o,m(t)))}))}(t,n)}));if(0<t.length&&t.length<3)return t.map((function(t){return function(n,t){if(!(4&e.getObjectFlags(n)&&t.properties))return n;var r=n.target,o=e.singleOrUndefined(r.typeParameters);if(!o)return n;var s=[];return t.properties.forEach((function(n,t){var i=a.getTypeOfPropertyOfType(r,t);e.Debug.assert(!!i,"generic should have all the properties of its reference."),s.push.apply(s,g(i,m(n),o))})),i[n.symbol.escapedName](f(s))}(t,n)}));return[]}(n)),c}function g(n,t,r){if(n===r)return[t];if(3145728&n.flags)return e.flatMap(n.types,(function(e){return g(e,t,r)}));if(4&e.getObjectFlags(n)&&4&e.getObjectFlags(t)){var i=a.getTypeArguments(n),o=a.getTypeArguments(t),s=[];if(i&&o)for(var c=0;c<i.length;c++)o[c]&&s.push.apply(s,g(i[c],o[c],r));return s}var l=a.getSignaturesOfType(n,0),u=a.getSignaturesOfType(t,0);return 1===l.length&&1===u.length?function(n,t,r){for(var i=[],o=0;o<n.parameters.length;o++){var s=n.parameters[o],c=t.parameters[o],l=n.declaration&&e.isRestParameter(n.declaration.parameters[o]);if(!c)break;var u=a.getTypeOfSymbolAtLocation(s,s.valueDeclaration),d=l&&a.getElementTypeOfArrayType(u);d&&(u=d);var p=c.type||a.getTypeOfSymbolAtLocation(c,c.valueDeclaration);i.push.apply(i,g(u,p,r))}var m=a.getReturnTypeOfSignature(n),f=a.getReturnTypeOfSignature(t);return i.push.apply(i,g(m,f,r)),i}(l[0],u[0],r):[]}function y(n){for(var t=[],r=Math.max.apply(Math,n.map((function(e){return e.argumentTypes.length}))),i=function(r){var i=a.createSymbol(1,e.escapeLeadingUnderscores("arg"+r));i.type=f(n.map((function(e){return e.argumentTypes[r]||a.getUndefinedType()}))),n.some((function(e){return void 0===e.argumentTypes[r]}))&&(i.flags|=16777216),t.push(i)},o=0;o<r;o++)i(o);var s=m(c(n.map((function(e){return e.return_}))));return a.createSignature(void 0,void 0,void 0,t,s,void 0,r,0)}function v(e,n){!n||1&n.flags||131072&n.flags||(e.candidateTypes||(e.candidateTypes=[])).push(n)}function h(e,n){!n||1&n.flags||131072&n.flags||(e.candidateThisTypes||(e.candidateThisTypes=[])).push(n)}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o,s=r.sourceFile,c=r.program,l=r.span.start,u=r.errorCode,d=r.cancellationToken,p=r.host,m=r.preferences,f=e.getTokenAtPosition(s,l),_=e.textChanges.ChangeTracker.with(r,(function(n){o=i(n,s,f,u,c,d,e.returnTrue,p,m)})),g=o&&e.getNameOfDeclaration(o);return g&&0!==_.length?[n.createCodeFixAction(t,_,[a(u,f),g.getText(s)],t,e.Diagnostics.Infer_all_types_from_usage)]:void 0},fixIds:[t],getAllCodeActions:function(t){var a=t.sourceFile,o=t.program,s=t.cancellationToken,c=t.host,l=t.preferences,u=e.nodeSeenTracker();return n.codeFixAll(t,r,(function(n,t){i(n,a,e.getTokenAtPosition(t.file,t.start),t.code,o,s,u,c,l)}))}}),n.addJSDocTags=u}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixReturnTypeInAsyncFunction",r=[e.Diagnostics.The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0.code];function a(n,t,r){if(!e.isInJSFile(n)){var a=e.getTokenAtPosition(n,r),i=e.findAncestor(a,e.isFunctionLikeDeclaration),o=null==i?void 0:i.type;if(o){var s=t.getTypeFromTypeNode(o),c=t.getAwaitedType(s)||t.getVoidType(),l=t.typeToTypeNode(c,o,void 0);return l?{returnTypeNode:o,returnType:s,promisedTypeNode:l,promisedType:c}:void 0}}}function i(n,t,r,a){n.replaceNode(t,r,e.factory.createTypeReferenceNode("Promise",[a]))}n.registerCodeFix({errorCodes:r,fixIds:[t],getCodeActions:function(r){var o=r.sourceFile,s=r.program,c=r.span,l=s.getTypeChecker(),u=a(o,s.getTypeChecker(),c.start);if(u){var d=u.returnTypeNode,p=u.returnType,m=u.promisedTypeNode,f=u.promisedType,_=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,d,m)}));return[n.createCodeFixAction(t,_,[e.Diagnostics.Replace_0_with_Promise_1,l.typeToString(p),l.typeToString(f)],t,e.Diagnostics.Fix_all_incorrect_return_type_of_an_async_functions)]}},getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=a(t.file,e.program.getTypeChecker(),t.start);r&&i(n,t.file,r.returnTypeNode,r.promisedTypeNode)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="disableJsDiagnostics",r="disableJsDiagnostics",a=e.mapDefined(Object.keys(e.Diagnostics),(function(n){var t=e.Diagnostics[n];return t.category===e.DiagnosticCategory.Error?t.code:void 0}));function i(n,t,r,a){var i=e.getLineAndCharacterOfPosition(t,r).line;a&&!e.tryAddToSet(a,i)||n.insertCommentBeforeLine(t,i,r," @ts-ignore")}n.registerCodeFix({errorCodes:a,getCodeActions:function(a){var o=a.sourceFile,s=a.program,c=a.span,l=a.host,u=a.formatContext;if(e.isInJSFile(o)&&e.isCheckJsEnabledForFile(o,s.getCompilerOptions())){var d=o.checkJsDirective?"":e.getNewLineOrDefaultFromHost(l,u.options),p=[n.createCodeFixActionWithoutFixAll(t,[n.createFileTextChanges(o.fileName,[e.createTextChange(o.checkJsDirective?e.createTextSpanFromBounds(o.checkJsDirective.pos,o.checkJsDirective.end):e.createTextSpan(0,0),"// @ts-nocheck"+d)])],e.Diagnostics.Disable_checking_for_this_file)];return e.textChanges.isValidLocationToAddComment(o,c.start)&&p.unshift(n.createCodeFixAction(t,e.textChanges.ChangeTracker.with(a,(function(e){return i(e,o,c.start)})),e.Diagnostics.Ignore_this_error_message,r,e.Diagnostics.Add_ts_ignore_to_all_error_messages)),p}},fixIds:[r],getAllCodeActions:function(t){var r=new e.Set;return n.codeFixAll(t,a,(function(n,t){e.textChanges.isValidLocationToAddComment(t.file,t.start)&&i(n,t.file,t.start,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){function t(n){return{trackSymbol:e.noop,moduleResolverHost:e.getModuleSpecifierResolverHost(n.program,n.host)}}function r(n,r,a,s,c,l,d){var m=n.getDeclarations();if(m&&m.length){var f=s.program.getTypeChecker(),_=e.getEmitScriptTarget(s.program.getCompilerOptions()),g=m[0],y=e.getSynthesizedDeepClone(e.getNameOfDeclaration(g),!1),v=function(n){if(4&n)return e.factory.createToken(122);if(16&n)return e.factory.createToken(121);return}(e.getEffectiveModifierFlags(g)),h=v?e.factory.createNodeArray([v]):void 0,b=f.getWidenedType(f.getTypeOfSymbolAtLocation(n,r)),E=!!(16777216&n.flags),T=!!(8388608&r.flags),S=e.getQuotePreference(a,c);switch(g.kind){case 162:case 163:var x=0===S?268435456:void 0,L=f.typeToTypeNode(b,r,x,t(s));if(l)(A=u(L,_))&&(L=A.typeNode,p(l,A.symbols));d(e.factory.createPropertyDeclaration(void 0,h,y,E?e.factory.createToken(57):void 0,L,void 0));break;case 167:case 168:var A,D=f.typeToTypeNode(b,r,void 0,t(s)),C=e.getAllAccessorDeclarations(m,g),I=C.secondAccessor?[C.firstAccessor,C.secondAccessor]:[C.firstAccessor];if(l)(A=u(D,_))&&(D=A.typeNode,p(l,A.symbols));for(var N=0,R=I;N<R.length;N++){var k=R[N];if(e.isGetAccessorDeclaration(k))d(e.factory.createGetAccessorDeclaration(void 0,h,y,e.emptyArray,D,T?void 0:o(S)));else{e.Debug.assertNode(k,e.isSetAccessorDeclaration,"The counterpart to a getter should be a setter");var O=e.getSetAccessorValueParameter(k),P=O&&e.isIdentifier(O.name)?e.idText(O.name):void 0;d(e.factory.createSetAccessorDeclaration(void 0,h,y,i(1,[P],[D],1,!1),T?void 0:o(S)))}}break;case 164:case 165:var M=f.getSignaturesOfType(b,0);if(!e.some(M))break;if(1===m.length){e.Debug.assert(1===M.length,"One declaration implies one signature"),G(S,M[0],h,y,T?void 0:o(S));break}for(var w=0,F=M;w<F.length;w++){G(S,F[w],e.getSynthesizedDeepClones(h,!1),e.getSynthesizedDeepClone(y,!1))}if(!T)if(m.length>M.length)G(S,f.getSignatureFromDeclaration(m[m.length-1]),h,y,o(S));else e.Debug.assert(m.length===M.length,"Declarations and signatures should match count"),d(function(n,t,r,a,s){for(var c=n[0],l=n[0].minArgumentCount,u=!1,d=0,p=n;d<p.length;d++){var m=p[d];l=Math.min(m.minArgumentCount,l),e.signatureHasRestParameter(m)&&(u=!0),m.parameters.length>=c.parameters.length&&(!e.signatureHasRestParameter(m)||e.signatureHasRestParameter(c))&&(c=m)}var f=c.parameters.length-(e.signatureHasRestParameter(c)?1:0),_=c.parameters.map((function(e){return e.name})),g=i(f,_,void 0,l,!1);if(u){var y=e.factory.createArrayTypeNode(e.factory.createKeywordTypeNode(128)),v=e.factory.createParameterDeclaration(void 0,void 0,e.factory.createToken(25),_[f]||"rest",f>=l?e.factory.createToken(57):void 0,y,void 0);g.push(v)}return function(n,t,r,a,i,s,c){return e.factory.createMethodDeclaration(void 0,n,void 0,t,r?e.factory.createToken(57):void 0,a,i,s,o(c))}(a,t,r,void 0,g,void 0,s)}(M,y,E,h,S))}}function G(n,a,i,o,c){var m=function(n,r,a,i,o,s,c,l,d){var m=n.program,f=m.getTypeChecker(),_=e.getEmitScriptTarget(m.getCompilerOptions()),g=1073742081|(0===r?268435456:0),y=f.signatureToSignatureDeclaration(a,165,i,g,t(n));if(!y)return;var v=y.typeParameters,h=y.parameters,b=y.type;if(d){if(v){var E=e.sameMap(v,(function(n){var t,r=n.constraint,a=n.default;r&&((t=u(r,_))&&(r=t.typeNode,p(d,t.symbols)));a&&((t=u(a,_))&&(a=t.typeNode,p(d,t.symbols)));return e.factory.updateTypeParameterDeclaration(n,n.name,r,a)}));v!==E&&(v=e.setTextRange(e.factory.createNodeArray(E,v.hasTrailingComma),v))}var T=e.sameMap(h,(function(n){var t=u(n.type,_),r=n.type;return t&&(r=t.typeNode,p(d,t.symbols)),e.factory.updateParameterDeclaration(n,n.decorators,n.modifiers,n.dotDotDotToken,n.name,n.questionToken,r,n.initializer)}));if(h!==T&&(h=e.setTextRange(e.factory.createNodeArray(T,h.hasTrailingComma),h)),b){var S=u(b,_);S&&(b=S.typeNode,p(d,S.symbols))}}return e.factory.updateMethodDeclaration(y,void 0,o,y.asteriskToken,s,c?e.factory.createToken(57):void 0,v,h,b,l)}(s,n,a,r,i,o,E,c,l);m&&d(m)}}function a(n,t,r,a,i,o,s){var c=n.typeToTypeNode(r,a,o,s);if(c&&e.isImportTypeNode(c)){var l=u(c,i);if(l)return p(t,l.symbols),l.typeNode}return c}function i(n,t,r,a,i){for(var o=[],s=0;s<n;s++){var c=e.factory.createParameterDeclaration(void 0,void 0,void 0,t&&t[s]||"arg"+s,void 0!==a&&s>=a?e.factory.createToken(57):void 0,i?void 0:r&&r[s]||e.factory.createKeywordTypeNode(128),void 0);o.push(c)}return o}function o(n){return e.factory.createBlock([e.factory.createThrowStatement(e.factory.createNewExpression(e.factory.createIdentifier("Error"),void 0,[e.factory.createStringLiteral("Method not implemented.",0===n)]))],!0)}function s(n,t,r){var a=e.getTsConfigObjectLiteralExpression(t);if(a){var i=l(a,"compilerOptions");if(void 0!==i){var o=i.initializer;if(e.isObjectLiteralExpression(o))for(var s=0,u=r;s<u.length;s++){var d=u[s],p=d[0],m=d[1],f=l(o,p);void 0===f?n.insertNodeAtObjectStart(t,o,c(p,m)):n.replaceNode(t,f.initializer,m)}}else n.insertNodeAtObjectStart(t,a,c("compilerOptions",e.factory.createObjectLiteralExpression(r.map((function(e){return c(e[0],e[1])})),!0)))}}function c(n,t){return e.factory.createPropertyAssignment(e.factory.createStringLiteral(n),t)}function l(n,t){return e.find(n.properties,(function(n){return e.isPropertyAssignment(n)&&!!n.name&&e.isStringLiteral(n.name)&&n.name.text===t}))}function u(n,t){var r,a=e.visitNode(n,(function n(a){var i;if(e.isLiteralImportTypeNode(a)&&a.qualifier){var o=e.getFirstIdentifier(a.qualifier),s=e.getNameForExportedSymbol(o.symbol,t),c=s!==o.text?d(a.qualifier,e.factory.createIdentifier(s)):a.qualifier;r=e.append(r,o.symbol);var l=null===(i=a.typeArguments)||void 0===i?void 0:i.map(n);return e.factory.createTypeReferenceNode(c,l)}return e.visitEachChild(a,n,e.nullTransformationContext)}));if(r&&a)return{typeNode:a,symbols:r}}function d(n,t){return 78===n.kind?t:e.factory.createQualifiedName(d(n.left,t),n.right)}function p(e,n){n.forEach((function(n){return e.addImportFromExportedSymbol(n,!0)}))}n.createMissingMemberNodes=function(e,n,t,a,i,o,s){for(var c=e.symbol.members,l=0,u=n;l<u.length;l++){var d=u[l];c.has(d.escapedName)||r(d,e,t,a,i,o,s)}},n.getNoopSymbolTrackerWithResolver=t,n.createMethodFromCallExpression=function(n,r,s,c,l,u,d){var p=!e.isInterfaceDeclaration(u),m=s.typeArguments,f=s.arguments,_=s.parent,g=e.getEmitScriptTarget(n.program.getCompilerOptions()),y=n.program.getTypeChecker(),v=t(n),h=e.map(f,(function(e){return a(y,r,y.getBaseTypeOfLiteralType(y.getTypeAtLocation(e)),u,g,void 0,v)})),b=e.map(f,(function(n){return e.isIdentifier(n)?n.text:e.isPropertyAccessExpression(n)&&e.isIdentifier(n.name)?n.name.text:void 0})),E=y.getContextualType(s),T=d||!E?void 0:y.typeToTypeNode(E,u,void 0,v),S=e.getQuotePreference(n.sourceFile,n.preferences);return e.factory.createMethodDeclaration(void 0,l?e.factory.createNodeArray(e.factory.createModifiersFromModifierFlags(l)):void 0,e.isYieldExpression(_)?e.factory.createToken(41):void 0,c,void 0,d?void 0:e.map(m,(function(n,t){return e.factory.createTypeParameterDeclaration(84+m.length-1<=90?String.fromCharCode(84+t):"T"+t)})),i(f.length,b,h,void 0,d),T,p?o(S):void 0)},n.typeToAutoImportableTypeNode=a,n.setJsonCompilerOptionValues=s,n.setJsonCompilerOptionValue=function(e,n,t,r){s(e,n,[[t,r]])},n.createJsonPropertyAssignment=c,n.findJsonProperty=l,n.tryGetAutoImportableReferenceFromTypeNode=u,n.importSymbols=p}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){function t(n){return e.isParameterPropertyDeclaration(n,n.parent)||e.isPropertyDeclaration(n)||e.isPropertyAssignment(n)}function r(n,t){return e.isIdentifier(t)?e.factory.createIdentifier(n):e.factory.createStringLiteral(n)}function a(n,t,r){var a=t?r.name:e.factory.createThis();return e.isIdentifier(n)?e.factory.createPropertyAccessExpression(a,n):e.factory.createElementAccessExpression(a,e.factory.createStringLiteralFromNode(n))}function i(n){return n?e.factory.createNodeArray(e.factory.createModifiersFromModifierFlags(n)):void 0}function o(n,a,i,o,s){void 0===s&&(s=!0);var l=e.getTokenAtPosition(n,i),u=i===o&&s,d=e.findAncestor(l.parent,t);if(!d||!e.nodeOverlapsWithStartEnd(d.name,n,i,o)&&!u)return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_property_for_which_to_generate_accessor)};if(!function(n){return e.isIdentifier(n)||e.isStringLiteral(n)}(d.name))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Name_is_not_valid)};if(124!=(124|e.getEffectiveModifierFlags(d)))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Can_only_convert_property_with_modifier)};var p=d.name.text,m=e.startsWithUnderscore(p),f=r(m?p:e.getUniqueName("_"+p,n),d.name),_=r(m?e.getUniqueName(p.substring(1),n):p,d.name);return{info:{isStatic:e.hasStaticModifier(d),isReadonly:e.hasEffectiveReadonlyModifier(d),type:c(d,a),container:160===d.kind?d.parent.parent:d.parent,originalName:d.name.text,declaration:d,fieldName:f,accessorName:_,renameAccessor:m}}}function s(n,t,r,a,i){e.isParameterPropertyDeclaration(a,a.parent)?n.insertNodeAtClassStart(t,i,r):e.isPropertyAssignment(a)?n.insertNodeAfterComma(t,a,r):n.insertNodeAfter(t,a,r)}function c(n,t){var r=e.getTypeAnnotationNode(n);if(e.isPropertyDeclaration(n)&&r&&n.questionToken){var a=t.getTypeChecker(),i=a.getTypeFromTypeNode(r);if(!a.isTypeAssignableTo(a.getUndefinedType(),i)){var o=e.isUnionTypeNode(r)?r.types:[r];return e.factory.createUnionTypeNode(__spreadArrays(o,[e.factory.createKeywordTypeNode(150)]))}}return r}n.generateAccessorFromProperty=function(n,t,r,c,l,u){var d=o(n,t,r,c);if(d&&d.info){var p,m,f=e.textChanges.ChangeTracker.fromContext(l),_=d.info,g=_.isStatic,y=_.isReadonly,v=_.fieldName,h=_.accessorName,b=_.originalName,E=_.type,T=_.container,S=_.declaration;if(e.suppressLeadingAndTrailingTrivia(v),e.suppressLeadingAndTrailingTrivia(h),e.suppressLeadingAndTrailingTrivia(S),e.suppressLeadingAndTrailingTrivia(T),e.isClassLike(T)){var x=e.getEffectiveModifierFlags(S);if(e.isSourceFileJS(n)){var L=i(x);p=L,m=L}else p=i(function(e){e&=-65,16&(e&=-9)||(e|=4);return e}(x)),m=i(function(e){return e&=-5,e&=-17,e|=8}(x))}!function(n,t,r,a,i,o){e.isPropertyDeclaration(r)?function(n,t,r,a,i,o){var s=e.factory.updatePropertyDeclaration(r,r.decorators,o,i,r.questionToken||r.exclamationToken,a,r.initializer);n.replaceNode(t,r,s)}(n,t,r,a,i,o):e.isPropertyAssignment(r)?function(n,t,r,a){var i=e.factory.updatePropertyAssignment(r,a,r.initializer);n.replacePropertyAssignment(t,r,i)}(n,t,r,i):n.replaceNode(t,r,e.factory.updateParameterDeclaration(r,r.decorators,o,r.dotDotDotToken,e.cast(i,e.isIdentifier),r.questionToken,r.type,r.initializer))}(f,n,S,E,v,m);var A=function(n,t,r,i,o,s){return e.factory.createGetAccessorDeclaration(void 0,i,t,void 0,r,e.factory.createBlock([e.factory.createReturnStatement(a(n,o,s))],!0))}(v,h,E,p,g,T);if(e.suppressLeadingAndTrailingTrivia(A),s(f,n,A,S,T),y){var D=e.getFirstConstructorWithBody(T);D&&function(n,t,r,a,i){if(!r.body)return;r.body.forEachChild((function r(o){e.isElementAccessExpression(o)&&107===o.expression.kind&&e.isStringLiteral(o.argumentExpression)&&o.argumentExpression.text===i&&e.isWriteAccess(o)&&n.replaceNode(t,o.argumentExpression,e.factory.createStringLiteral(a)),e.isPropertyAccessExpression(o)&&107===o.expression.kind&&o.name.text===i&&e.isWriteAccess(o)&&n.replaceNode(t,o.name,e.factory.createIdentifier(a)),e.isFunctionLike(o)||e.isClassLike(o)||o.forEachChild(r)}))}(f,n,D,v.text,b)}else{var C=function(n,t,r,i,o,s){return e.factory.createSetAccessorDeclaration(void 0,i,t,[e.factory.createParameterDeclaration(void 0,void 0,void 0,e.factory.createIdentifier("value"),void 0,r)],e.factory.createBlock([e.factory.createExpressionStatement(e.factory.createAssignment(a(n,o,s),e.factory.createIdentifier("value")))],!0))}(v,h,E,p,g,T);e.suppressLeadingAndTrailingTrivia(C),s(f,n,C,S,T)}return f.getChanges()}},n.getAccessorConvertiblePropertyAtPosition=o,n.getAllSupers=function(n,t){for(var r=[];n;){var a=e.getClassExtendsHeritageElement(n),i=a&&t.getSymbolAtLocation(a.expression);if(!i)break;var o=2097152&i.flags?t.getAliasedSymbol(i):i,s=e.find(o.declarations,e.isClassLike);if(!s)break;r.push(s),n=s}return r}}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="invalidImportSyntax";function r(r,a,i,o){var s=e.textChanges.ChangeTracker.with(r,(function(e){return e.replaceNode(a,i,o)}));return n.createCodeFixActionWithoutFixAll(t,s,[e.Diagnostics.Replace_import_with_0,s[0].textChanges[0].newText])}function a(a,i){var o=a.program.getTypeChecker().getTypeAtLocation(i);if(!o.symbol||!o.symbol.originatingImport)return[];var s=[],c=o.symbol.originatingImport;if(e.isImportCall(c)||e.addRange(s,function(n,t){var a=e.getSourceFileOfNode(t),i=e.getNamespaceDeclarationNode(t),o=n.program.getCompilerOptions(),s=[];return s.push(r(n,a,t,e.makeImport(i.name,void 0,t.moduleSpecifier,e.getQuotePreference(a,n.preferences)))),e.getEmitModuleKind(o)===e.ModuleKind.CommonJS&&s.push(r(n,a,t,e.factory.createImportEqualsDeclaration(void 0,void 0,i.name,e.factory.createExternalModuleReference(t.moduleSpecifier)))),s}(a,c)),e.isExpression(i)&&(!e.isNamedDeclaration(i.parent)||i.parent.name!==i)){var l=a.sourceFile,u=e.textChanges.ChangeTracker.with(a,(function(n){return n.replaceNode(l,i,e.factory.createPropertyAccessExpression(i,"default"),{})}));s.push(n.createCodeFixActionWithoutFixAll(t,u,e.Diagnostics.Use_synthetic_default_member))}return s}n.registerCodeFix({errorCodes:[e.Diagnostics.This_expression_is_not_callable.code,e.Diagnostics.This_expression_is_not_constructable.code],getCodeActions:function(n){var t=n.sourceFile,r=e.Diagnostics.This_expression_is_not_callable.code===n.errorCode?203:204,i=e.findAncestor(e.getTokenAtPosition(t,n.span.start),(function(e){return e.kind===r}));if(!i)return[];var o=i.expression;return a(n,o)}}),n.registerCodeFix({errorCodes:[e.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,e.Diagnostics.Type_0_does_not_satisfy_the_constraint_1.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1.code,e.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated.code,e.Diagnostics.Type_predicate_0_is_not_assignable_to_1.code,e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_string_index_type_2.code,e.Diagnostics.Property_0_of_type_1_is_not_assignable_to_numeric_index_type_2.code,e.Diagnostics.Numeric_index_type_0_is_not_assignable_to_string_index_type_1.code,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2.code,e.Diagnostics.Property_0_in_type_1_is_not_assignable_to_type_2.code,e.Diagnostics.Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property.code,e.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1.code],getCodeActions:function(n){var t=n.sourceFile,r=e.findAncestor(e.getTokenAtPosition(t,n.span.start),(function(e){return e.getStart()===n.span.start&&e.getEnd()===n.span.start+n.span.length}));if(!r)return[];return a(n,r)}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="strictClassInitialization",r="addMissingPropertyDefiniteAssignmentAssertions",a="addMissingPropertyUndefinedType",i="addMissingPropertyInitializer",o=[e.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor.code];function s(n,t){var r=e.getTokenAtPosition(n,t);return e.isIdentifier(r)?e.cast(r.parent,e.isPropertyDeclaration):void 0}function c(a,i){var o=e.textChanges.ChangeTracker.with(a,(function(e){return l(e,a.sourceFile,i)}));return n.createCodeFixAction(t,o,[e.Diagnostics.Add_definite_assignment_assertion_to_property_0,i.getText()],r,e.Diagnostics.Add_definite_assignment_assertions_to_all_uninitialized_properties)}function l(n,t,r){var a=e.factory.updatePropertyDeclaration(r,r.decorators,r.modifiers,r.name,e.factory.createToken(53),r.type,r.initializer);n.replaceNode(t,r,a)}function u(r,i){var o=e.textChanges.ChangeTracker.with(r,(function(e){return d(e,r.sourceFile,i)}));return n.createCodeFixAction(t,o,[e.Diagnostics.Add_undefined_type_to_property_0,i.name.getText()],a,e.Diagnostics.Add_undefined_type_to_all_uninitialized_properties)}function d(n,t,r){var a=e.factory.createKeywordTypeNode(150),i=r.type,o=e.isUnionTypeNode(i)?i.types.concat(a):[i,a];n.replaceNode(t,i,e.factory.createUnionTypeNode(o))}function p(n,t,r,a){var i=e.factory.updatePropertyDeclaration(r,r.decorators,r.modifiers,r.name,r.questionToken,r.type,a);n.replaceNode(t,r,i)}function m(e,n){return f(e,e.getTypeFromTypeNode(n.type))}function f(n,t){if(512&t.flags)return t===n.getFalseType()||t===n.getFalseType(!0)?e.factory.createFalse():e.factory.createTrue();if(t.isStringLiteral())return e.factory.createStringLiteral(t.value);if(t.isNumberLiteral())return e.factory.createNumericLiteral(t.value);if(2048&t.flags)return e.factory.createBigIntLiteral(t.value);if(t.isUnion())return e.firstDefined(t.types,(function(e){return f(n,e)}));if(t.isClass()){var r=e.getClassLikeDeclarationOfSymbol(t.symbol);if(!r||e.hasSyntacticModifier(r,128))return;var a=e.getFirstConstructorWithBody(r);if(a&&a.parameters.length)return;return e.factory.createNewExpression(e.factory.createIdentifier(t.symbol.name),void 0,void 0)}return n.isArrayLikeType(t)?e.factory.createArrayLiteralExpression():void 0}n.registerCodeFix({errorCodes:o,getCodeActions:function(r){var a=s(r.sourceFile,r.span.start);if(a){var o=[u(r,a),c(r,a)];return e.append(o,function(r,a){var o=m(r.program.getTypeChecker(),a);if(!o)return;var s=e.textChanges.ChangeTracker.with(r,(function(e){return p(e,r.sourceFile,a,o)}));return n.createCodeFixAction(t,s,[e.Diagnostics.Add_initializer_to_property_0,a.name.getText()],i,e.Diagnostics.Add_initializers_to_all_uninitialized_properties)}(r,a)),o}},fixIds:[r,a,i],getAllCodeActions:function(t){return n.codeFixAll(t,o,(function(n,o){var c=s(o.file,o.start);if(c)switch(t.fixId){case r:l(n,o.file,c);break;case a:d(n,o.file,c);break;case i:var u=m(t.program.getTypeChecker(),c);if(!u)return;p(n,o.file,c,u);break;default:e.Debug.fail(JSON.stringify(t.fixId))}}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="requireInTs",r=[e.Diagnostics.require_call_may_be_converted_to_an_import.code];function a(n,t,r){var a=r.allowSyntheticDefaults,i=r.defaultImportName,o=r.namedImports,s=r.statement,c=r.required;n.replaceNode(t,s,i&&!a?e.factory.createImportEqualsDeclaration(void 0,void 0,i,e.factory.createExternalModuleReference(c)):e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,i,o),c))}function i(n,t,r){var a=e.getTokenAtPosition(n,r).parent;if(!e.isRequireCall(a,!0))throw e.Debug.failBadSyntaxKind(a);var i=e.cast(a.parent,e.isVariableDeclaration),o=e.tryCast(i.name,e.isIdentifier),s=e.isObjectBindingPattern(i.name)?function(n){for(var t=[],r=0,a=n.elements;r<a.length;r++){var i=a[r];if(!e.isIdentifier(i.name)||i.initializer)return;t.push(e.factory.createImportSpecifier(e.tryCast(i.propertyName,e.isIdentifier),i.name))}if(t.length)return e.factory.createNamedImports(t)}(i.name):void 0;if(o||s)return{allowSyntheticDefaults:e.getAllowSyntheticDefaultImports(t.getCompilerOptions()),defaultImportName:o,namedImports:s,statement:e.cast(i.parent.parent,e.isVariableStatement),required:e.first(a.arguments)}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=i(r.sourceFile,r.program,r.span.start);if(o){var s=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,o)}));return[n.createCodeFixAction(t,s,e.Diagnostics.Convert_require_to_import,t,e.Diagnostics.Convert_all_require_to_import)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=i(t.file,e.program,t.start);r&&a(n,e.sourceFile,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="useDefaultImport",r=[e.Diagnostics.Import_may_be_converted_to_a_default_import.code];function a(n,t){var r=e.getTokenAtPosition(n,t);if(e.isIdentifier(r)){var a=r.parent;if(e.isImportEqualsDeclaration(a)&&e.isExternalModuleReference(a.moduleReference))return{importNode:a,name:r,moduleSpecifier:a.moduleReference.expression};if(e.isNamespaceImport(a)){var i=a.parent.parent;return{importNode:i,name:r,moduleSpecifier:i.moduleSpecifier}}}}function i(n,t,r,a){n.replaceNode(t,r.importNode,e.makeImport(r.name,void 0,r.moduleSpecifier,e.getQuotePreference(t,a)))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span.start,c=a(o,s);if(c){var l=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c,r.preferences)}));return[n.createCodeFixAction(t,l,e.Diagnostics.Convert_to_default_import,t,e.Diagnostics.Convert_all_to_default_imports)]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=a(t.file,t.start);r&&i(n,t.file,r,e.preferences)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="useBigintLiteral",r=[e.Diagnostics.Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers.code];function a(n,t,r){var a=e.tryCast(e.getTokenAtPosition(t,r.start),e.isNumericLiteral);if(a){var i=a.getText(t)+"n";n.replaceNode(t,a,e.factory.createBigIntLiteral(i))}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span)}));if(i.length>0)return[n.createCodeFixAction(t,i,e.Diagnostics.Convert_to_a_bigint_numeric_literal,t,e.Diagnostics.Convert_all_to_bigint_numeric_literals)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixAddModuleReferTypeMissingTypeof",r=[e.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0.code];function a(n,t){var r=e.getTokenAtPosition(n,t);return e.Debug.assert(99===r.kind,"This token should be an ImportKeyword"),e.Debug.assert(195===r.parent.kind,"Token parent should be an ImportType"),r.parent}function i(n,t,r){var a=e.factory.updateImportTypeNode(r,r.argument,r.qualifier,r.typeArguments,!0);n.replaceNode(t,r,a)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start),l=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c)}));return[n.createCodeFixAction(t,l,e.Diagnostics.Add_missing_typeof,t,e.Diagnostics.Add_missing_typeof)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){return i(n,e.sourceFile,a(t.file,t.start))}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="wrapJsxInFragment",r=[e.Diagnostics.JSX_expressions_must_have_one_parent_element.code];function a(n,t){var r=e.getTokenAtPosition(n,t).parent.parent;if((e.isBinaryExpression(r)||(r=r.parent,e.isBinaryExpression(r)))&&e.nodeIsMissing(r.operatorToken))return r}function i(n,t,r){var a=function(n){var t=[],r=n;for(;;){if(e.isBinaryExpression(r)&&e.nodeIsMissing(r.operatorToken)&&27===r.operatorToken.kind){if(t.push(r.left),e.isJsxChild(r.right))return t.push(r.right),t;if(e.isBinaryExpression(r.right)){r=r.right;continue}return}return}}(r);a&&n.replaceNode(t,r,e.factory.createJsxFragment(e.factory.createJsxOpeningFragment(),a,e.factory.createJsxJsxClosingFragment()))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.program.getCompilerOptions().jsx;if(2===o||3===o){var s=r.sourceFile,c=r.span,l=a(s,c.start);if(l){var u=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,s,l)}));return[n.createCodeFixAction(t,u,e.Diagnostics.Wrap_in_JSX_fragment,t,e.Diagnostics.Wrap_all_unparented_JSX_in_JSX_fragment)]}}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=a(e.sourceFile,t.start);r&&i(n,e.sourceFile,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixConvertToMappedObjectType",r=[e.Diagnostics.An_index_signature_parameter_type_cannot_be_a_union_type_Consider_using_a_mapped_object_type_instead.code];function a(n,t){var r=e.getTokenAtPosition(n,t),a=e.cast(r.parent.parent,e.isIndexSignatureDeclaration);if(!e.isClassDeclaration(a.parent))return{indexSignature:a,container:e.isInterfaceDeclaration(a.parent)?a.parent:e.cast(a.parent.parent,e.isTypeAliasDeclaration)}}function i(n,t,r){var a,i,o=r.indexSignature,s=r.container,c=(e.isInterfaceDeclaration(s)?s.members:s.type.members).filter((function(n){return!e.isIndexSignatureDeclaration(n)})),l=e.first(o.parameters),u=e.factory.createTypeParameterDeclaration(e.cast(l.name,e.isIdentifier),l.type),d=e.factory.createMappedTypeNode(e.hasEffectiveReadonlyModifier(o)?e.factory.createModifier(142):void 0,u,void 0,o.questionToken,o.type),p=e.factory.createIntersectionTypeNode(__spreadArrays(e.getAllSuperTypeNodes(s),[d],c.length?[e.factory.createTypeLiteralNode(c)]:e.emptyArray));n.replaceNode(t,s,(a=s,i=p,e.factory.createTypeAliasDeclaration(a.decorators,a.modifiers,a.name,a.typeParameters,i)))}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=r.span,c=a(o,s.start);if(c){var l=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,c)})),u=e.idText(c.container.name);return[n.createCodeFixAction(t,l,[e.Diagnostics.Convert_0_to_mapped_object_type,u],t,[e.Diagnostics.Convert_0_to_mapped_object_type,u])]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){var t=a(n.file,n.start);t&&i(e,n.file,t)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="removeAccidentalCallParentheses",r=[e.Diagnostics.This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without.code];n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var a=e.findAncestor(e.getTokenAtPosition(r.sourceFile,r.span.start),e.isCallExpression);if(a){var i=e.textChanges.ChangeTracker.with(r,(function(e){e.deleteRange(r.sourceFile,{pos:a.expression.end,end:a.end})}));return[n.createCodeFixActionWithoutFixAll(t,i,e.Diagnostics.Remove_parentheses)]}},fixIds:[t]})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="removeUnnecessaryAwait",r=[e.Diagnostics.await_has_no_effect_on_the_type_of_this_expression.code];function a(n,t,r){var a=e.tryCast(e.getTokenAtPosition(t,r.start),(function(e){return 130===e.kind})),i=a&&e.tryCast(a.parent,e.isAwaitExpression);if(i){var o=i;if(e.isParenthesizedExpression(i.parent)){var s=e.getLeftmostExpression(i.expression,!1);if(e.isIdentifier(s)){var c=e.findPrecedingToken(i.parent.pos,t);c&&102!==c.kind&&(o=i.parent)}}n.replaceNode(t,o,i.expression)}}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span)}));if(i.length>0)return[n.createCodeFixAction(t,i,e.Diagnostics.Remove_unnecessary_await,t,e.Diagnostics.Remove_all_unnecessary_uses_of_await)]},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(e,n){return a(e,n.file,n)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t=[e.Diagnostics.A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both.code],r="splitTypeOnlyImport";function a(n,t){return e.findAncestor(e.getTokenAtPosition(n,t.start),e.isImportDeclaration)}function i(n,t,r){if(t){var a=e.Debug.checkDefined(t.importClause);n.replaceNode(r.sourceFile,t,e.factory.updateImportDeclaration(t,t.decorators,t.modifiers,e.factory.updateImportClause(a,a.isTypeOnly,a.name,void 0),t.moduleSpecifier)),n.insertNodeAfter(r.sourceFile,t,e.factory.createImportDeclaration(void 0,void 0,e.factory.updateImportClause(a,a.isTypeOnly,void 0,a.namedBindings),t.moduleSpecifier))}}n.registerCodeFix({errorCodes:t,fixIds:[r],getCodeActions:function(t){var o=e.textChanges.ChangeTracker.with(t,(function(e){return i(e,a(t.sourceFile,t.span),t)}));if(o.length)return[n.createCodeFixAction(r,o,e.Diagnostics.Split_into_two_separate_import_declarations,r,e.Diagnostics.Split_all_invalid_type_only_imports)]},getAllCodeActions:function(e){return n.codeFixAll(e,t,(function(n,t){i(n,a(e.sourceFile,t),e)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixConvertConstToLet",r=[e.Diagnostics.Cannot_assign_to_0_because_it_is_a_constant.code];n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var a=r.sourceFile,i=r.span,o=r.program,s=function(n,t,r){var a=e.getTokenAtPosition(n,t),i=r.getTypeChecker().getSymbolAtLocation(a);if(i)return i.valueDeclaration.parent.parent}(a,i.start,o),c=e.textChanges.ChangeTracker.with(r,(function(e){return function(e,n,t){if(!t)return;var r=t.getStart();e.replaceRangeWithText(n,{pos:r,end:r+5},"let")}(e,a,s)}));return[n.createCodeFixAction(t,c,e.Diagnostics.Convert_const_to_let,t,e.Diagnostics.Convert_const_to_let)]},fixIds:[t]})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="fixExpectedComma",r=[e.Diagnostics._0_expected.code];function a(n,t,r){var a=e.getTokenAtPosition(n,t);return 26===a.kind&&a.parent&&(e.isObjectLiteralExpression(a.parent)||e.isArrayLiteralExpression(a.parent))?{node:a}:void 0}function i(n,t,r){var a=r.node,i=e.factory.createToken(27);n.replaceNode(t,a,i)}n.registerCodeFix({errorCodes:r,getCodeActions:function(r){var o=r.sourceFile,s=a(o,r.span.start,r.errorCode);if(s){var c=e.textChanges.ChangeTracker.with(r,(function(e){return i(e,o,s)}));return[n.createCodeFixAction(t,c,[e.Diagnostics.Change_0_to_1,";",","],t,[e.Diagnostics.Change_0_to_1,";",","])]}},fixIds:[t],getAllCodeActions:function(e){return n.codeFixAll(e,r,(function(n,t){var r=a(t.file,t.start,t.code);r&&i(n,e.sourceFile,r)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="addVoidToPromise",r=[e.Diagnostics.Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise.code];function a(n,t,r,a,i){var o=e.getTokenAtPosition(t,r.start);if(e.isIdentifier(o)&&e.isCallExpression(o.parent)&&o.parent.expression===o&&0===o.parent.arguments.length){var s=a.getTypeChecker(),c=s.getSymbolAtLocation(o),l=null==c?void 0:c.valueDeclaration;if(l&&e.isParameter(l)&&e.isNewExpression(l.parent.parent)&&!(null==i?void 0:i.has(l))){null==i||i.add(l);var u=function(n){var t;if(!e.isInJSFile(n))return n.typeArguments;if(e.isParenthesizedExpression(n.parent)){var r=null===(t=e.getJSDocTypeTag(n.parent))||void 0===t?void 0:t.typeExpression.type;if(r&&e.isTypeReferenceNode(r)&&e.isIdentifier(r.typeName)&&"Promise"===e.idText(r.typeName))return r.typeArguments}}(l.parent.parent);if(e.some(u)){var d=u[0],p=!e.isUnionTypeNode(d)&&!e.isParenthesizedTypeNode(d)&&e.isParenthesizedTypeNode(e.factory.createUnionTypeNode([d,e.factory.createKeywordTypeNode(113)]).types[0]);p&&n.insertText(t,d.pos,"("),n.insertText(t,d.end,p?") | void":" | void")}else{var m=s.getResolvedSignature(o.parent),f=null==m?void 0:m.parameters[0],_=f&&s.getTypeOfSymbolAtLocation(f,l.parent.parent);e.isInJSFile(l)?(!_||3&_.flags)&&(n.insertText(t,l.parent.parent.end,")"),n.insertText(t,e.skipTrivia(t.text,l.parent.parent.pos),"/** @type {Promise<void>} */(")):(!_||2&_.flags)&&n.insertText(t,l.parent.parent.expression.end,"<void>")}}}}n.registerCodeFix({errorCodes:r,fixIds:[t],getCodeActions:function(r){var i=e.textChanges.ChangeTracker.with(r,(function(e){return a(e,r.sourceFile,r.span,r.program)}));if(i.length>0)return[n.createCodeFixAction("addVoidToPromise",i,e.Diagnostics.Add_void_to_Promise_resolved_without_a_value,t,e.Diagnostics.Add_void_to_all_Promises_resolved_without_a_value)]},getAllCodeActions:function(t){return n.codeFixAll(t,r,(function(n,r){return a(n,r.file,r,t.program,new e.Set)}))}})}(e.codefix||(e.codefix={}))}(ts||(ts={})),function(e){!function(n){var t="Convert export",r="Convert default export to named export",a="Convert named export to default export";function i(n,t){void 0===t&&(t=!0);var r=n.file,a=e.getRefactorContextSpan(n),i=e.getTokenAtPosition(r,a.start),o=i.parent&&1&e.getSyntacticModifierFlags(i.parent)&&t?i.parent:e.getParentNodeInSpan(i,r,a);if(!(o&&(e.isSourceFile(o.parent)||e.isModuleBlock(o.parent)&&e.isAmbientModule(o.parent.parent))))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_export_statement)};var s=e.isSourceFile(o.parent)?o.parent.symbol:o.parent.parent.symbol,c=e.getSyntacticModifierFlags(o),l=!!(512&c);if(!(1&c)||!l&&s.exports.has("default"))return{error:e.getLocaleSpecificMessage(e.Diagnostics.This_file_already_has_a_default_export)};switch(o.kind){case 251:case 252:case 253:case 255:case 254:case 256:var u=o;return u.name&&e.isIdentifier(u.name)?{info:{exportNode:u,exportName:u.name,wasDefault:l,exportingModuleSymbol:s}}:void 0;case 232:var d=o;if(!(2&d.declarationList.flags)||1!==d.declarationList.declarations.length)return;var p=e.first(d.declarationList.declarations);if(!p.initializer)return;return e.Debug.assert(!l,"Can't have a default flag here"),e.isIdentifier(p.name)?{info:{exportNode:d,exportName:p.name,wasDefault:l,exportingModuleSymbol:s}}:void 0;default:return}}function o(n,t){return e.factory.createImportSpecifier(n===t?void 0:e.factory.createIdentifier(n),e.factory.createIdentifier(t))}n.registerRefactor(t,{getAvailableActions:function(n){var o=i(n,"invoked"===n.triggerReason);if(!o)return e.emptyArray;if(void 0===o.error){var s=o.info.wasDefault?e.Diagnostics.Convert_default_export_to_named_export.message:e.Diagnostics.Convert_named_export_to_default_export.message,c=o.info.wasDefault?r:a;return[{name:t,description:s,actions:[{name:c,description:s}]}]}return n.preferences.provideRefactorNotApplicableReason?[{name:t,description:e.Diagnostics.Convert_default_export_to_named_export.message,actions:[{name:r,description:e.Diagnostics.Convert_default_export_to_named_export.message,notApplicableReason:o.error}]},{name:t,description:e.Diagnostics.Convert_named_export_to_default_export.message,actions:[{name:a,description:e.Diagnostics.Convert_named_export_to_default_export.message,notApplicableReason:o.error}]}]:e.emptyArray},getEditsForAction:function(n,t){return e.Debug.assert(t===r||t===a,"Unexpected action name"),{edits:e.textChanges.ChangeTracker.with(n,(function(t){var r,a,s,c,l,u;return a=n.file,s=n.program,c=e.Debug.checkDefined(null===(r=i(n))||void 0===r?void 0:r.info,"context must have info"),l=t,u=n.cancellationToken,function(n,t,r,a){var i=t.wasDefault,o=t.exportNode,s=t.exportName;if(i)r.delete(n,e.Debug.checkDefined(e.findModifier(o,87),"Should find a default keyword in modifier list"));else{var c=e.Debug.checkDefined(e.findModifier(o,92),"Should find an export keyword in modifier list");switch(o.kind){case 251:case 252:case 253:r.insertNodeAfter(n,c,e.factory.createToken(87));break;case 232:var l=e.first(o.declarationList.declarations);if(!e.FindAllReferences.Core.isSymbolReferencedInFile(s,a,n)&&!l.type){r.replaceNode(n,o,e.factory.createExportDefault(e.Debug.checkDefined(l.initializer,"Initializer was previously known to be present")));break}case 255:case 254:case 256:r.deleteModifier(n,c),r.insertNodeAfter(n,o,e.factory.createExportDefault(e.factory.createIdentifier(s.text)));break;default:e.Debug.assertNever(o,"Unexpected exportNode kind "+o.kind)}}}(a,c,l,s.getTypeChecker()),void function(n,t,r,a){var i=t.wasDefault,s=t.exportName,c=t.exportingModuleSymbol,l=n.getTypeChecker(),u=e.Debug.checkDefined(l.getSymbolAtLocation(s),"Export name should resolve to a symbol");e.FindAllReferences.Core.eachExportReference(n.getSourceFiles(),l,a,u,c,s.text,i,(function(n){var t=n.getSourceFile();i?function(n,t,r,a){var i=t.parent;switch(i.kind){case 201:r.replaceNode(n,t,e.factory.createIdentifier(a));break;case 265:case 270:var s=i;r.replaceNode(n,s,o(a,s.name.text));break;case 262:var c=i;e.Debug.assert(c.name===t,"Import clause name should match provided ref"),s=o(a,t.text);var l=c.namedBindings;if(l)if(263===l.kind){r.deleteRange(n,{pos:t.getStart(n),end:l.getStart(n)});var u=e.isStringLiteral(c.parent.moduleSpecifier)?e.quotePreferenceFromString(c.parent.moduleSpecifier,n):1,d=e.makeImport(void 0,[o(a,t.text)],c.parent.moduleSpecifier,u);r.insertNodeAfter(n,c.parent,d)}else r.delete(n,t),r.insertNodeAtEndOfList(n,l.elements,s);else r.replaceNode(n,t,e.factory.createNamedImports([s]));break;default:e.Debug.failBadSyntaxKind(i)}}(t,n,r,s.text):function(n,t,r){var a,i,o=t.parent;switch(o.kind){case 201:r.replaceNode(n,t,e.factory.createIdentifier("default"));break;case 265:var s=e.factory.createIdentifier(o.name.text);1===o.parent.elements.length?r.replaceNode(n,o.parent,s):(r.delete(n,o),r.insertNodeBefore(n,o.parent,s));break;case 270:r.replaceNode(n,o,(a="default",i=o.name.text,e.factory.createExportSpecifier(a===i?void 0:e.factory.createIdentifier(a),e.factory.createIdentifier(i))));break;default:e.Debug.assertNever(o,"Unexpected parent kind "+o.kind)}}(t,n,r)}))}(s,c,l,u)})),renameFilename:void 0,renameLocation:void 0}}})}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){var t="Convert import",r="Convert namespace import to named imports",a="Convert named imports to namespace import";function i(n,t){void 0===t&&(t=!0);var r=n.file,a=e.getRefactorContextSpan(n),i=e.getTokenAtPosition(r,a.start),o=t?e.findAncestor(i,e.isImportDeclaration):e.getParentNodeInSpan(i,r,a);if(!o||!e.isImportDeclaration(o))return{error:"Selection is not an import declaration."};if(!(o.getEnd()<a.start+a.length)){var s=o.importClause;return s?s.namedBindings?{info:s.namedBindings}:{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_namespace_import_or_named_imports)}:{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_import_clause)}}}function o(n,t,r){return e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,t,r&&r.length?e.factory.createNamedImports(r):void 0),n.moduleSpecifier)}n.registerRefactor(t,{getAvailableActions:function(n){var o=i(n,"invoked"===n.triggerReason);if(!o)return e.emptyArray;if(void 0===o.error){var s=263===o.info.kind?e.Diagnostics.Convert_namespace_import_to_named_imports.message:e.Diagnostics.Convert_named_imports_to_namespace_import.message,c=263===o.info.kind?r:a;return[{name:t,description:s,actions:[{name:c,description:s}]}]}return n.preferences.provideRefactorNotApplicableReason?[{name:t,description:e.Diagnostics.Convert_namespace_import_to_named_imports.message,actions:[{name:r,description:e.Diagnostics.Convert_namespace_import_to_named_imports.message,notApplicableReason:o.error}]},{name:t,description:e.Diagnostics.Convert_named_imports_to_namespace_import.message,actions:[{name:a,description:e.Diagnostics.Convert_named_imports_to_namespace_import.message,notApplicableReason:o.error}]}]:e.emptyArray},getEditsForAction:function(n,t){return e.Debug.assert(t===r||t===a,"Unexpected action name"),{edits:e.textChanges.ChangeTracker.with(n,(function(t){var r,a,s,c,l,u;return a=n.file,s=n.program,c=t,l=e.Debug.checkDefined(null===(r=i(n))||void 0===r?void 0:r.info,"Context must provide an import to convert"),u=s.getTypeChecker(),void(263===l.kind?function(n,t,r,a,i){var s=!1,c=[],l=new e.Map;e.FindAllReferences.Core.eachSymbolReferenceInFile(a.name,t,n,(function(n){if(e.isPropertyAccessExpression(n.parent)){var r=e.cast(n.parent,e.isPropertyAccessExpression),a=r.name.text;t.resolveName(a,n,67108863,!0)&&l.set(a,!0),e.Debug.assert(r.expression===n,"Parent expression should match id"),c.push(r)}else s=!0}));for(var u=new e.Map,d=0,p=c;d<p.length;d++){var m=p[d],f=m.name.text,_=u.get(f);void 0===_&&u.set(f,_=l.has(f)?e.getUniqueName(f,n):f),r.replaceNode(n,m,e.factory.createIdentifier(_))}var g=[];u.forEach((function(n,t){g.push(e.factory.createImportSpecifier(n===t?void 0:e.factory.createIdentifier(t),e.factory.createIdentifier(n)))}));var y=a.parent.parent;s&&!i?r.insertNodeAfter(n,y,o(y,void 0,g)):r.replaceNode(n,y,o(y,s?e.factory.createIdentifier(a.name.text):void 0,g))}(a,u,c,l,e.getAllowSyntheticDefaultImports(s.getCompilerOptions())):function(n,t,r,a){for(var i=a.parent.parent,s=i.moduleSpecifier,c=s&&e.isStringLiteral(s)?e.codefix.moduleSpecifierToValidIdentifier(s.text,99):"module",l=a.elements.some((function(r){return e.FindAllReferences.Core.eachSymbolReferenceInFile(r.name,t,n,(function(e){return!!t.resolveName(c,e,67108863,!0)}))||!1}))?e.getUniqueName(c,n):c,u=[],d=function(a){var i=(a.propertyName||a.name).text;e.FindAllReferences.Core.eachSymbolReferenceInFile(a.name,t,n,(function(t){var o=e.factory.createPropertyAccessExpression(e.factory.createIdentifier(l),i);e.isShorthandPropertyAssignment(t.parent)?r.replaceNode(n,t.parent,e.factory.createPropertyAssignment(t.text,o)):e.isExportSpecifier(t.parent)&&!t.parent.propertyName?u.some((function(e){return e.name===a.name}))||u.push(e.factory.createImportSpecifier(a.propertyName&&e.factory.createIdentifier(a.propertyName.text),e.factory.createIdentifier(a.name.text))):r.replaceNode(n,t,o)}))},p=0,m=a.elements;p<m.length;p++)d(m[p]);r.replaceNode(n,a,e.factory.createNamespaceImport(e.factory.createIdentifier(l))),u.length&&r.insertNodeAfter(n,a.parent.parent,o(i,void 0,u))}(a,u,c,l))})),renameFilename:void 0,renameLocation:void 0}}})}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Convert to optional chain expression",a=e.getLocaleSpecificMessage(e.Diagnostics.Convert_to_optional_chain_expression);function i(n){return e.isBinaryExpression(n)||e.isConditionalExpression(n)}function o(n){return i(n)||function(n){return e.isExpressionStatement(n)||e.isReturnStatement(n)||e.isVariableStatement(n)}(n)}function s(n,t){void 0===t&&(t=!0);var r=n.file,a=n.program,s=e.getRefactorContextSpan(n),u=0===s.length;if(!u||t){var p=e.getTokenAtPosition(r,s.start),m=e.findTokenOnLeftOfPosition(r,s.start+s.length),f=e.createTextSpanFromBounds(p.pos,m&&m.end>=p.pos?m.getEnd():p.getEnd()),_=u?function(e){for(;e.parent;){if(o(e)&&!o(e.parent))return e;e=e.parent}return}(p):function(e,n){for(;e.parent;){if(o(e)&&0!==n.length&&e.end>=n.start+n.length)return e;e=e.parent}return}(p,f),g=_&&o(_)?function(n){if(i(n))return n;if(e.isVariableStatement(n)){var t=e.getSingleVariableOfVariableStatement(n),r=null==t?void 0:t.initializer;return r&&i(r)?r:void 0}return n.expression&&i(n.expression)?n.expression:void 0}(_):void 0;if(!g)return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_convertible_access_expression)};var y=a.getTypeChecker();return e.isConditionalExpression(g)?function(n,t){var r=n.condition,a=d(n.whenTrue);if(!a||t.isNullableType(t.getTypeAtLocation(a)))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_convertible_access_expression)};if((e.isPropertyAccessExpression(r)||e.isIdentifier(r))&&l(r,a.expression))return{info:{finalExpression:a,occurrences:[r],expression:n}};if(e.isBinaryExpression(r)){var i=c(a.expression,r);return i?{info:{finalExpression:a,occurrences:i,expression:n}}:{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_matching_access_expressions)}}}(g,y):function(n){if(55!==n.operatorToken.kind)return{error:e.getLocaleSpecificMessage(e.Diagnostics.Can_only_convert_logical_AND_access_chains)};var t=d(n.right);if(!t)return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_convertible_access_expression)};var r=c(t.expression,n.left);return r?{info:{finalExpression:t,occurrences:r,expression:n}}:{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_matching_access_expressions)}}(g)}}function c(n,t){for(var r=[];e.isBinaryExpression(t)&&55===t.operatorToken.kind;){var a=l(e.skipParentheses(n),e.skipParentheses(t.right));if(!a)break;r.push(a),n=a,t=t.left}var i=l(n,t);return i&&r.push(i),r.length>0?r:void 0}function l(n,t){if(e.isIdentifier(t)||e.isPropertyAccessExpression(t)||e.isElementAccessExpression(t))return function(n,t){for(;(e.isCallExpression(n)||e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n))&&u(n)!==u(t);)n=n.expression;for(;e.isPropertyAccessExpression(n)&&e.isPropertyAccessExpression(t)||e.isElementAccessExpression(n)&&e.isElementAccessExpression(t);){if(u(n)!==u(t))return!1;n=n.expression,t=t.expression}return e.isIdentifier(n)&&e.isIdentifier(t)&&n.getText()===t.getText()}(n,t)?t:void 0}function u(n){return e.isIdentifier(n)||e.isStringOrNumericLiteralLike(n)?n.getText():e.isPropertyAccessExpression(n)?u(n.name):e.isElementAccessExpression(n)?u(n.argumentExpression):void 0}function d(n){return n=e.skipParentheses(n),e.isBinaryExpression(n)?d(n.left):(e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n)||e.isCallExpression(n))&&!e.isOptionalChain(n)?n:void 0}function p(n,t,r){if(e.isPropertyAccessExpression(t)||e.isElementAccessExpression(t)||e.isCallExpression(t)){var a=p(n,t.expression,r),i=r.length>0?r[r.length-1]:void 0,o=(null==i?void 0:i.getText())===t.expression.getText();if(o&&r.pop(),e.isCallExpression(t))return o?e.factory.createCallChain(a,e.factory.createToken(28),t.typeArguments,t.arguments):e.factory.createCallChain(a,t.questionDotToken,t.typeArguments,t.arguments);if(e.isPropertyAccessExpression(t))return o?e.factory.createPropertyAccessChain(a,e.factory.createToken(28),t.name):e.factory.createPropertyAccessChain(a,t.questionDotToken,t.name);if(e.isElementAccessExpression(t))return o?e.factory.createElementAccessChain(a,e.factory.createToken(28),t.argumentExpression):e.factory.createElementAccessChain(a,t.questionDotToken,t.argumentExpression)}return t}n.registerRefactor(r,{getAvailableActions:function(n){var t=s(n,"invoked"===n.triggerReason);if(!t)return e.emptyArray;if(!t.error)return[{name:r,description:a,actions:[{name:r,description:a}]}];if(n.preferences.provideRefactorNotApplicableReason)return[{name:r,description:a,actions:[{name:r,description:a,notApplicableReason:t.error}]}];return e.emptyArray},getEditsForAction:function(n,t){var r=s(n);return r&&r.info?{edits:e.textChanges.ChangeTracker.with(n,(function(t){return function(n,t,r,a,i){var o=a.finalExpression,s=a.occurrences,c=a.expression,l=s[s.length-1],u=p(t,o,s);u&&(e.isPropertyAccessExpression(u)||e.isElementAccessExpression(u)||e.isCallExpression(u))&&(e.isBinaryExpression(c)?r.replaceNodeRange(n,l,o,u):e.isConditionalExpression(c)&&r.replaceNode(n,c,e.factory.createBinaryExpression(u,e.factory.createToken(60),c.whenFalse)))}(n.file,n.program.getTypeChecker(),t,e.Debug.checkDefined(r.info,"context must have info"))})),renameFilename:void 0,renameLocation:void 0}:void 0}})}(n.convertToOptionalChainExpression||(n.convertToOptionalChainExpression={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Convert overload list to single signature",a=e.Diagnostics.Convert_overload_list_to_single_signature.message;function i(e){switch(e.kind){case 164:case 165:case 169:case 166:case 170:case 251:return!0}return!1}function o(n,t,r){var a=e.getTokenAtPosition(n,t),o=e.findAncestor(a,i);if(o){var s=r.getTypeChecker(),c=o.symbol;if(c){var l=c.declarations;if(!(e.length(l)<=1)&&e.every(l,(function(t){return e.getSourceFileOfNode(t)===n}))&&i(l[0])){var u=l[0].kind;if(e.every(l,(function(e){return e.kind===u}))){var d=l;if(!e.some(d,(function(n){return!!n.typeParameters||e.some(n.parameters,(function(n){return!!n.decorators||!!n.modifiers||!e.isIdentifier(n.name)}))}))){var p=e.mapDefined(d,(function(e){return s.getSignatureFromDeclaration(e)}));if(e.length(p)===e.length(l)){var m=s.getReturnTypeOfSignature(p[0]);if(e.every(p,(function(e){return s.getReturnTypeOfSignature(e)===m})))return d}}}}}}}n.registerRefactor(r,{getEditsForAction:function(n){var t=n.file,r=n.startPosition,a=n.program,i=o(t,r,a);if(!i)return;var s=a.getTypeChecker(),c=i[i.length-1],l=c;switch(c.kind){case 164:l=e.factory.updateMethodSignature(c,c.modifiers,c.name,c.questionToken,c.typeParameters,d(i),c.type);break;case 165:l=e.factory.updateMethodDeclaration(c,c.decorators,c.modifiers,c.asteriskToken,c.name,c.questionToken,c.typeParameters,d(i),c.type,c.body);break;case 169:l=e.factory.updateCallSignature(c,c.typeParameters,d(i),c.type);break;case 166:l=e.factory.updateConstructorDeclaration(c,c.decorators,c.modifiers,d(i),c.body);break;case 170:l=e.factory.updateConstructSignature(c,c.typeParameters,d(i),c.type);break;case 251:l=e.factory.updateFunctionDeclaration(c,c.decorators,c.modifiers,c.asteriskToken,c.name,c.typeParameters,d(i),c.type,c.body);break;default:return e.Debug.failBadSyntaxKind(c,"Unhandled signature kind in overload list conversion refactoring")}if(l===c)return;var u=e.textChanges.ChangeTracker.with(n,(function(e){e.replaceNodeRange(t,i[0],i[i.length-1],l)}));return{renameFilename:void 0,renameLocation:void 0,edits:u};function d(n){var t=n[n.length-1];return e.isFunctionLikeDeclaration(t)&&t.body&&(n=n.slice(0,n.length-1)),e.factory.createNodeArray([e.factory.createParameterDeclaration(void 0,void 0,e.factory.createToken(25),"args",void 0,e.factory.createUnionTypeNode(e.map(n,p)))])}function p(n){var t=e.map(n.parameters,m);return e.setEmitFlags(e.factory.createTupleTypeNode(t),e.some(t,(function(n){return!!e.length(e.getSyntheticLeadingComments(n))}))?0:1)}function m(n){e.Debug.assert(e.isIdentifier(n.name));var t=e.setTextRange(e.factory.createNamedTupleMember(n.dotDotDotToken,n.name,n.questionToken,n.type||e.factory.createKeywordTypeNode(128)),n),r=n.symbol&&n.symbol.getDocumentationComment(s);if(r){var a=e.displayPartsToString(r);a.length&&e.setSyntheticLeadingComments(t,[{text:"*\n"+a.split("\n").map((function(e){return" * "+e})).join("\n")+"\n ",kind:3,pos:-1,end:-1,hasTrailingNewLine:!0,hasLeadingNewline:!0}])}return t}},getAvailableActions:function(n){var t=n.file,i=n.startPosition,s=n.program;return o(t,i,s)?[{name:r,description:a,actions:[{name:r,description:a}]}]:e.emptyArray}})}(n.addOrRemoveBracesToArrowFunction||(n.addOrRemoveBracesToArrowFunction={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r,a,i,o,s="Extract Symbol";function c(n){var t=u(n.file,e.getRefactorContextSpan(n),"invoked"===n.triggerReason),r=t.targetRange;if(void 0===r)return t.errors&&0!==t.errors.length&&n.preferences.provideRefactorNotApplicableReason?[{name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_function),actions:[{description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_function),name:"function_extract_error",notApplicableReason:S(t.errors)}]},{name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_constant),actions:[{description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_constant),name:"constant_extract_error",notApplicableReason:S(t.errors)}]}]:e.emptyArray;var a=function(n,t){var r=m(n,t),a=r.scopes,i=r.readsAndWrites,o=i.functionErrorsPerScope,s=i.constantErrorsPerScope;return a.map((function(n,t){var r,a,i=function(n){return e.isFunctionLikeDeclaration(n)?"inner function":e.isClassLike(n)?"method":"function"}(n),c=function(n){return e.isClassLike(n)?"readonly field":"constant"}(n),l=e.isFunctionLikeDeclaration(n)?function(n){switch(n.kind){case 166:return"constructor";case 208:case 251:return n.name?"function '"+n.name.text+"'":e.ANONYMOUS;case 209:return"arrow function";case 165:return"method '"+n.name.getText()+"'";case 167:return"'get "+n.name.getText()+"'";case 168:return"'set "+n.name.getText()+"'";default:throw e.Debug.assertNever(n,"Unexpected scope kind "+n.kind)}}(n):e.isClassLike(n)?function(e){return 252===e.kind?e.name?"class '"+e.name.text+"'":"anonymous class declaration":e.name?"class expression '"+e.name.text+"'":"anonymous class expression"}(n):function(e){return 257===e.kind?"namespace '"+e.parent.name.getText()+"'":e.externalModuleIndicator?0:1}(n);return 1===l?(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[i,"global"]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[c,"global"])):0===l?(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[i,"module"]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1_scope),[c,"module"])):(r=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1),[i,l]),a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_1),[c,l])),0!==t||e.isClassLike(n)||(a=e.formatStringFromArgs(e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_0_in_enclosing_scope),[c])),{functionExtraction:{description:r,errors:o[t]},constantExtraction:{description:a,errors:s[t]}}}))}(r,n);if(void 0===a)return e.emptyArray;for(var i,o,c=[],l=new e.Map,d=[],p=new e.Map,f=0,_=0,g=a;_<g.length;_++){var y=g[_],v=y.functionExtraction,h=y.constantExtraction,b=v.description;if(0===v.errors.length?l.has(b)||(l.set(b,!0),c.push({description:b,name:"function_scope_"+f})):i||(i={description:b,name:"function_scope_"+f,notApplicableReason:S(v.errors)}),0===h.errors.length){var E=h.description;p.has(E)||(p.set(E,!0),d.push({description:E,name:"constant_scope_"+f}))}else o||(o={description:b,name:"constant_scope_"+f,notApplicableReason:S(h.errors)});f++}var T=[];return c.length?T.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_function),actions:c}):n.preferences.provideRefactorNotApplicableReason&&i&&T.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_function),actions:[i]}),d.length?T.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_constant),actions:d}):n.preferences.provideRefactorNotApplicableReason&&o&&T.push({name:s,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_constant),actions:[o]}),T.length?T:e.emptyArray;function S(e){var n=e[0].messageText;return"string"!=typeof n&&(n=n.messageText),n}}function l(n,t){var r=u(n.file,e.getRefactorContextSpan(n)).targetRange,i=/^function_scope_(\d+)$/.exec(t);if(i){var o=+i[1];return e.Debug.assert(isFinite(o),"Expected to parse a finite number from the function scope index"),function(n,t,r){var i=m(n,t),o=i.scopes,s=i.readsAndWrites,c=s.target,l=s.usagesPerScope,u=s.functionErrorsPerScope,d=s.exposedVariableDeclarations;return e.Debug.assert(!u[r].length,"The extraction went missing? How?"),t.cancellationToken.throwIfCancellationRequested(),function(n,t,r,i,o,s){var c,l,u=r.usages,d=r.typeParameterUsages,p=r.substitutions,m=s.program.getTypeChecker(),v=e.getEmitScriptTarget(s.program.getCompilerOptions()),h=e.codefix.createImportAdder(s.file,s.program,s.preferences,s.host),E=t.getSourceFile(),T=e.getUniqueName(e.isClassLike(t)?"newMethod":"newFunction",E),S=e.isInJSFile(t),x=e.factory.createIdentifier(T),L=[],A=[];u.forEach((function(n,r){var a;if(!S){var i=m.getTypeOfSymbolAtLocation(n.symbol,n.node);i=m.getBaseTypeOfLiteralType(i),a=e.codefix.typeToAutoImportableTypeNode(m,h,i,t,v,1)}var o=e.factory.createParameterDeclaration(void 0,void 0,void 0,r,void 0,a);L.push(o),2===n.usage&&(l||(l=[])).push(n),A.push(e.factory.createIdentifier(r))}));var D=e.arrayFrom(d.values()).map((function(e){return{type:e,declaration:f(e)}})).sort(_),C=0===D.length?void 0:D.map((function(e){return e.declaration})),I=void 0!==C?C.map((function(n){return e.factory.createTypeReferenceNode(n.name,void 0)})):void 0;if(e.isExpression(n)&&!S){var N=m.getContextualType(n);c=m.typeToTypeNode(N,t,1)}var R,k=function(n,t,r,a,i){var o,s=void 0!==r||t.length>0;if(e.isBlock(n)&&!s&&0===a.size)return{body:e.factory.createBlock(n.statements,!0),returnValueProperty:void 0};var c=!1,l=e.factory.createNodeArray(e.isBlock(n)?n.statements.slice(0):[e.isStatement(n)?n:e.factory.createReturnStatement(n)]);if(s||a.size){var u=e.visitNodes(l,p).slice();if(s&&!i&&e.isStatement(n)){var d=g(t,r);1===d.length?u.push(e.factory.createReturnStatement(d[0].name)):u.push(e.factory.createReturnStatement(e.factory.createObjectLiteralExpression(d)))}return{body:e.factory.createBlock(u,!0),returnValueProperty:o}}return{body:e.factory.createBlock(l,!0),returnValueProperty:void 0};function p(n){if(!c&&e.isReturnStatement(n)&&s){var i=g(t,r);return n.expression&&(o||(o="__return"),i.unshift(e.factory.createPropertyAssignment(o,e.visitNode(n.expression,p)))),1===i.length?e.factory.createReturnStatement(i[0].name):e.factory.createReturnStatement(e.factory.createObjectLiteralExpression(i))}var l=c;c=c||e.isFunctionLikeDeclaration(n)||e.isClassLike(n);var u=a.get(e.getNodeId(n).toString()),d=u?e.getSynthesizedDeepClone(u):e.visitEachChild(n,p,e.nullTransformationContext);return c=l,d}}(n,i,l,p,!!(o.facts&a.HasReturn)),O=k.body,P=k.returnValueProperty;if(e.suppressLeadingAndTrailingTrivia(O),e.isClassLike(t)){var M=S?[]:[e.factory.createModifier(120)];o.facts&a.InStaticRegion&&M.push(e.factory.createModifier(123)),o.facts&a.IsAsyncFunction&&M.push(e.factory.createModifier(129)),R=e.factory.createMethodDeclaration(void 0,M.length?M:void 0,o.facts&a.IsGenerator?e.factory.createToken(41):void 0,x,void 0,C,L,c,O)}else R=e.factory.createFunctionDeclaration(void 0,o.facts&a.IsAsyncFunction?[e.factory.createToken(129)]:void 0,o.facts&a.IsGenerator?e.factory.createToken(41):void 0,x,C,L,c,O);var w=e.textChanges.ChangeTracker.fromContext(s),F=function(n,t){return e.find(function(n){if(e.isFunctionLikeDeclaration(n)){var t=n.body;if(e.isBlock(t))return t.statements}else{if(e.isModuleBlock(n)||e.isSourceFile(n))return n.statements;if(e.isClassLike(n))return n.members;e.assertType(n)}return e.emptyArray}(t),(function(t){return t.pos>=n&&e.isFunctionLikeDeclaration(t)&&!e.isConstructorDeclaration(t)}))}((y(o.range)?e.last(o.range):o.range).end,t);F?w.insertNodeBefore(s.file,F,R,!0):w.insertNodeAtEndOfScope(s.file,t,R);h.writeFixes(w);var G=[],B=function(n,t,r){var i=e.factory.createIdentifier(r);if(e.isClassLike(n)){var o=t.facts&a.InStaticRegion?e.factory.createIdentifier(n.name.text):e.factory.createThis();return e.factory.createPropertyAccessExpression(o,i)}return i}(t,o,T),U=e.factory.createCallExpression(B,I,A);o.facts&a.IsGenerator&&(U=e.factory.createYieldExpression(e.factory.createToken(41),U));o.facts&a.IsAsyncFunction&&(U=e.factory.createAwaitExpression(U));b(n)&&(U=e.factory.createJsxExpression(void 0,U));if(i.length&&!l)if(e.Debug.assert(!P,"Expected no returnValueProperty"),e.Debug.assert(!(o.facts&a.HasReturn),"Expected RangeFacts.HasReturn flag to be unset"),1===i.length){var V=i[0];G.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(e.getSynthesizedDeepClone(V.name),void 0,e.getSynthesizedDeepClone(V.type),U)],V.parent.flags)))}else{for(var j=[],H=[],K=i[0].parent.flags,W=!1,z=0,q=i;z<q.length;z++){V=q[z];j.push(e.factory.createBindingElement(void 0,void 0,e.getSynthesizedDeepClone(V.name)));var J=m.typeToTypeNode(m.getBaseTypeOfLiteralType(m.getTypeAtLocation(V)),t,1);H.push(e.factory.createPropertySignature(void 0,V.symbol.name,void 0,J)),W=W||void 0!==V.type,K&=V.parent.flags}var X=W?e.factory.createTypeLiteralNode(H):void 0;X&&e.setEmitFlags(X,1),G.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(e.factory.createObjectBindingPattern(j),void 0,X,U)],K)))}else if(i.length||l){if(i.length)for(var Y=0,Q=i;Y<Q.length;Y++){var Z=(V=Q[Y]).parent.flags;2&Z&&(Z=-3&Z|1),G.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(V.symbol.name,void 0,re(V.type))],Z)))}P&&G.push(e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(P,void 0,re(c))],1)));var $=g(i,l);P&&$.unshift(e.factory.createShorthandPropertyAssignment(P)),1===$.length?(e.Debug.assert(!P,"Shouldn't have returnValueProperty here"),G.push(e.factory.createExpressionStatement(e.factory.createAssignment($[0].name,U))),o.facts&a.HasReturn&&G.push(e.factory.createReturnStatement())):(G.push(e.factory.createExpressionStatement(e.factory.createAssignment(e.factory.createObjectLiteralExpression($),U))),P&&G.push(e.factory.createReturnStatement(e.factory.createIdentifier(P))))}else o.facts&a.HasReturn?G.push(e.factory.createReturnStatement(U)):y(o.range)?G.push(e.factory.createExpressionStatement(U)):G.push(U);y(o.range)?w.replaceNodeRangeWithNodes(s.file,e.first(o.range),e.last(o.range),G):w.replaceNodeWithNodes(s.file,o.range,G);var ee=w.getChanges(),ne=(y(o.range)?e.first(o.range):o.range).getSourceFile().fileName,te=e.getRenameLocation(ee,ne,T,!1);return{renameFilename:ne,renameLocation:te,edits:ee};function re(n){if(void 0!==n){for(var t=e.getSynthesizedDeepClone(n),r=t;e.isParenthesizedTypeNode(r);)r=r.type;return e.isUnionTypeNode(r)&&e.find(r.types,(function(e){return 150===e.kind}))?t:e.factory.createUnionTypeNode([t,e.factory.createKeywordTypeNode(150)])}}}(c,o[r],l[r],d,n,t)}(r,n,o)}var s=/^constant_scope_(\d+)$/.exec(t);if(s){o=+s[1];return e.Debug.assert(isFinite(o),"Expected to parse a finite number from the constant scope index"),function(n,t,r){var i=m(n,t),o=i.scopes,s=i.readsAndWrites,c=s.target,l=s.usagesPerScope,u=s.constantErrorsPerScope,d=s.exposedVariableDeclarations;return e.Debug.assert(!u[r].length,"The extraction went missing? How?"),e.Debug.assert(0===d.length,"Extract constant accepted a range containing a variable declaration?"),t.cancellationToken.throwIfCancellationRequested(),function(n,t,r,i,o){var s,c=r.substitutions,l=o.program.getTypeChecker(),u=t.getSourceFile(),d=e.getUniqueName(e.isClassLike(t)?"newProperty":"newLocal",u),m=e.isInJSFile(t),f=m||!l.isContextSensitive(n)?void 0:l.typeToTypeNode(l.getContextualType(n),t,1),_=function(n,t){return t.size?r(n):n;function r(n){var a=t.get(e.getNodeId(n).toString());return a?e.getSynthesizedDeepClone(a):e.visitEachChild(n,r,e.nullTransformationContext)}}(n,c);s=I(f,_),f=s.variableType,_=s.initializer,e.suppressLeadingAndTrailingTrivia(_);var g=e.textChanges.ChangeTracker.fromContext(o);if(e.isClassLike(t)){e.Debug.assert(!m,"Cannot extract to a JS class");var y=[];y.push(e.factory.createModifier(120)),i&a.InStaticRegion&&y.push(e.factory.createModifier(123)),y.push(e.factory.createModifier(142));var v=e.factory.createPropertyDeclaration(void 0,y,d,void 0,f,_),E=e.factory.createPropertyAccessExpression(i&a.InStaticRegion?e.factory.createIdentifier(t.name.getText()):e.factory.createThis(),e.factory.createIdentifier(d));b(n)&&(E=e.factory.createJsxExpression(void 0,E));var T=function(n,t){var r,a=t.members;e.Debug.assert(a.length>0,"Found no members");for(var i=!0,o=0,s=a;o<s.length;o++){var c=s[o];if(c.pos>n)return r||a[0];if(i&&!e.isPropertyDeclaration(c)){if(void 0!==r)return c;i=!1}r=c}return void 0===r?e.Debug.fail():r}(n.pos,t);g.insertNodeBefore(o.file,T,v,!0),g.replaceNode(o.file,n,E)}else{var S=e.factory.createVariableDeclaration(d,void 0,f,_),x=function(n,t){var r;for(;void 0!==n&&n!==t;){if(e.isVariableDeclaration(n)&&n.initializer===r&&e.isVariableDeclarationList(n.parent)&&n.parent.declarations.length>1)return n;r=n,n=n.parent}}(n,t);if(x){g.insertNodeBefore(o.file,x,S);E=e.factory.createIdentifier(d);g.replaceNode(o.file,n,E)}else if(233===n.parent.kind&&t===e.findAncestor(n,p)){var L=e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([S],2));g.replaceNode(o.file,n.parent,L)}else{L=e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([S],2));if(0===(T=function(n,t){var r;e.Debug.assert(!e.isClassLike(t));for(var a=n;a!==t;a=a.parent)p(a)&&(r=a);for(a=(r||n).parent;;a=a.parent){if(h(a)){for(var i=void 0,o=0,s=a.statements;o<s.length;o++){var c=s[o];if(c.pos>n.pos)break;i=c}return!i&&e.isCaseClause(a)?(e.Debug.assert(e.isSwitchStatement(a.parent.parent),"Grandparent isn't a switch statement"),a.parent.parent):e.Debug.checkDefined(i,"prevStatement failed to get set")}e.Debug.assert(a!==t,"Didn't encounter a block-like before encountering scope")}}(n,t)).pos?g.insertNodeAtTopOfFile(o.file,L,!1):g.insertNodeBefore(o.file,T,L,!1),233===n.parent.kind)g.delete(o.file,n.parent);else{E=e.factory.createIdentifier(d);b(n)&&(E=e.factory.createJsxExpression(void 0,E)),g.replaceNode(o.file,n,E)}}}var A=g.getChanges(),D=n.getSourceFile().fileName,C=e.getRenameLocation(A,D,d,!0);return{renameFilename:D,renameLocation:C,edits:A};function I(r,a){if(void 0===r)return{variableType:r,initializer:a};if(!e.isFunctionExpression(a)&&!e.isArrowFunction(a)||a.typeParameters)return{variableType:r,initializer:a};var i=l.getTypeAtLocation(n),o=e.singleOrUndefined(l.getSignaturesOfType(i,0));if(!o)return{variableType:r,initializer:a};if(o.getTypeParameters())return{variableType:r,initializer:a};for(var s=[],c=!1,u=0,d=a.parameters;u<d.length;u++){var p=d[u];if(p.type)s.push(p);else{var m=l.getTypeAtLocation(p);m===l.getAnyType()&&(c=!0),s.push(e.factory.updateParameterDeclaration(p,p.decorators,p.modifiers,p.dotDotDotToken,p.name,p.questionToken,p.type||l.typeToTypeNode(m,t,1),p.initializer))}}if(c)return{variableType:r,initializer:a};if(r=void 0,e.isArrowFunction(a))a=e.factory.updateArrowFunction(a,n.modifiers,a.typeParameters,s,a.type||l.typeToTypeNode(o.getReturnType(),t,1),a.equalsGreaterThanToken,a.body);else{if(o&&o.thisParameter){var f=e.firstOrUndefined(s);if(!f||e.isIdentifier(f.name)&&"this"!==f.name.escapedText){var _=l.getTypeOfSymbolAtLocation(o.thisParameter,n);s.splice(0,0,e.factory.createParameterDeclaration(void 0,void 0,void 0,"this",void 0,l.typeToTypeNode(_,t,1)))}}a=e.factory.updateFunctionExpression(a,n.modifiers,a.asteriskToken,a.name,a.typeParameters,s,a.type||l.typeToTypeNode(o.getReturnType(),t,1),a.body)}return{variableType:r,initializer:a}}}(e.isExpression(c)?c:c.statements[0].expression,o[r],l[r],n.facts,t)}(r,n,o)}e.Debug.fail("Unrecognized action name")}function u(n,t,i){void 0===i&&(i=!0);var o=t.length;if(0===o&&!i)return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractEmpty)]};var s=0===o&&i,c=e.getTokenAtPosition(n,t.start),l=s?function(n){return e.findAncestor(n,(function(n){return n.parent&&v(n)&&!e.isBinaryExpression(n.parent)}))}(c):e.getParentNodeInSpan(c,n,t),u=e.findTokenOnLeftOfPosition(n,e.textSpanEnd(t)),p=s?l:e.getParentNodeInSpan(u,n,t),m=[],f=a.None;if(!l||!p)return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractRange)]};if(l.parent!==p.parent)return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractRange)]};if(l!==p){if(!h(l.parent))return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractRange)]};for(var _=[],g=0,y=l.parent.statements;g<y.length;g++){var b=y[g];if(b===l||_.length){var E=x(b);if(E)return{errors:E};_.push(b)}if(b===p)break}return _.length?{targetRange:{range:_,facts:f,declarations:m}}:{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractRange)]}}if(e.isJSDoc(l))return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractJSDoc)]};if(e.isReturnStatement(l)&&!l.expression)return{errors:[e.createFileDiagnostic(n,t.start,o,r.cannotExtractRange)]};var T=function(n){if(e.isReturnStatement(n)){if(n.expression)return n.expression}else if(e.isVariableStatement(n)){for(var t=0,r=void 0,a=0,i=n.declarationList.declarations;a<i.length;a++){var o=i[a];o.initializer&&(t++,r=o.initializer)}if(1===t)return r}else if(e.isVariableDeclaration(n)&&n.initializer)return n.initializer;return n}(l),S=function(n){if(e.isIdentifier(e.isExpressionStatement(n)?n.expression:n))return[e.createDiagnosticForNode(n,r.cannotExtractIdentifier)];return}(T)||x(T);return S?{errors:S}:{targetRange:{range:d(T),facts:f,declarations:m}};function x(n){var i;if(function(e){e[e.None=0]="None",e[e.Break=1]="Break",e[e.Continue=2]="Continue",e[e.Return=4]="Return"}(i||(i={})),e.Debug.assert(n.pos<=n.end,"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809 (1)"),e.Debug.assert(!e.positionIsSynthesized(n.pos),"This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809 (2)"),!(e.isStatement(n)||e.isExpressionNode(n)&&v(n)))return[e.createDiagnosticForNode(n,r.statementOrExpressionExpected)];if(8388608&n.flags)return[e.createDiagnosticForNode(n,r.cannotExtractAmbientBlock)];var o,s=e.getContainingClass(n);s&&function(n,t){for(var r=n;r!==t;){if(163===r.kind){e.hasSyntacticModifier(r,32)&&(f|=a.InStaticRegion);break}if(160===r.kind){166===e.getContainingFunction(r).kind&&(f|=a.InStaticRegion);break}165===r.kind&&e.hasSyntacticModifier(r,32)&&(f|=a.InStaticRegion),r=r.parent}}(n,s);var c,l=4;return function n(i){if(o)return!0;if(e.isDeclaration(i)){var s=249===i.kind?i.parent.parent:i;if(e.hasSyntacticModifier(s,1))return(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractExportedEntity)),!0;m.push(i.symbol)}switch(i.kind){case 261:return(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractImport)),!0;case 266:return(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractExportedEntity)),!0;case 105:if(203===i.parent.kind){var u=e.getContainingClass(i);if(u.pos<t.start||u.end>=t.start+t.length)return(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractSuper)),!0}else f|=a.UsesThis;break;case 209:e.forEachChild(i,(function n(t){if(e.isThis(t))f|=a.UsesThis;else{if(e.isClassLike(t)||e.isFunctionLike(t)&&!e.isArrowFunction(t))return!1;e.forEachChild(t,n)}}));case 252:case 251:e.isSourceFile(i.parent)&&void 0===i.parent.externalModuleIndicator&&(o||(o=[])).push(e.createDiagnosticForNode(i,r.functionWillNotBeVisibleInTheNewScope));case 221:case 208:case 165:case 166:case 167:case 168:return!1}var d=l;switch(i.kind){case 234:case 247:l=0;break;case 230:i.parent&&247===i.parent.kind&&i.parent.finallyBlock===i&&(l=4);break;case 285:case 284:l|=1;break;default:e.isIterationStatement(i,!1)&&(l|=3)}switch(i.kind){case 187:case 107:f|=a.UsesThis;break;case 245:var p=i.label;(c||(c=[])).push(p.escapedText),e.forEachChild(i,n),c.pop();break;case 241:case 240:(p=i.label)?e.contains(c,p.escapedText)||(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange)):l&(241===i.kind?1:2)||(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractRangeContainingConditionalBreakOrContinueStatements));break;case 213:f|=a.IsAsyncFunction;break;case 219:f|=a.IsGenerator;break;case 242:4&l?f|=a.HasReturn:(o||(o=[])).push(e.createDiagnosticForNode(i,r.cannotExtractRangeContainingConditionalReturnStatement));break;default:e.forEachChild(i,n)}l=d}(n),o}}function d(n){return e.isStatement(n)?[n]:e.isExpressionNode(n)?e.isExpressionStatement(n.parent)?[n.parent]:n:void 0}function p(n){return e.isFunctionLikeDeclaration(n)||e.isSourceFile(n)||e.isModuleBlock(n)||e.isClassLike(n)}function m(n,t){var i=t.file,o=function(n){var t=y(n.range)?e.first(n.range):n.range;if(n.facts&a.UsesThis){var r=e.getContainingClass(t);if(r){var i=e.findAncestor(t,e.isFunctionLikeDeclaration);return i?[i,r]:[r]}}for(var o=[];;)if(160===(t=t.parent).kind&&(t=e.findAncestor(t,(function(n){return e.isFunctionLikeDeclaration(n)})).parent),p(t)&&(o.push(t),297===t.kind))return o}(n);return{scopes:o,readsAndWrites:function(n,t,i,o,s,c){var l,u,d=new e.Map,p=[],m=[],f=[],_=[],g=[],v=new e.Map,h=[],b=y(n.range)?1===n.range.length&&e.isExpressionStatement(n.range[0])?n.range[0].expression:void 0:n.range;if(void 0===b){var E=n.range,T=e.first(E).getStart(),S=e.last(E).end;u=e.createFileDiagnostic(o,T,S-T,r.expressionExpected)}else 147456&s.getTypeAtLocation(b).flags&&(u=e.createDiagnosticForNode(b,r.uselessConstantType));for(var x=0,L=t;x<L.length;x++){var A=L[x];p.push({usages:new e.Map,typeParameterUsages:new e.Map,substitutions:new e.Map}),m.push(new e.Map),f.push(e.isFunctionLikeDeclaration(A)&&251!==A.kind?[e.createDiagnosticForNode(A,r.cannotExtractToOtherFunctionLike)]:[]);var D=[];u&&D.push(u),e.isClassLike(A)&&e.isInJSFile(A)&&D.push(e.createDiagnosticForNode(A,r.cannotExtractToJSClass)),e.isArrowFunction(A)&&!e.isBlock(A.body)&&D.push(e.createDiagnosticForNode(A,r.cannotExtractToExpressionArrowFunction)),_.push(D)}var C=new e.Map,I=y(n.range)?e.factory.createBlock(n.range):n.range,N=y(n.range)?e.first(n.range):n.range,R=j(N);if(K(I),R&&!y(n.range)){H(s.getContextualType(n.range))}if(d.size>0){for(var k=new e.Map,O=0,P=N;void 0!==P&&O<t.length;P=P.parent)if(P===t[O]&&(k.forEach((function(e,n){p[O].typeParameterUsages.set(n,e)})),O++),e.isDeclarationWithTypeParameters(P))for(var M=0,w=e.getEffectiveTypeParameterDeclarations(P);M<w.length;M++){var F=w[M],G=s.getTypeAtLocation(F);d.has(G.id.toString())&&k.set(G.id.toString(),G)}e.Debug.assert(O===t.length,"Should have iterated all scopes")}if(g.length){var B=e.isBlockScope(t[0],t[0].parent)?t[0]:e.getEnclosingBlockScopeContainer(t[0]);e.forEachChild(B,q)}for(var U=function(t){var a=p[t];if(t>0&&(a.usages.size>0||a.typeParameterUsages.size>0)){var i=y(n.range)?n.range[0]:n.range;_[t].push(e.createDiagnosticForNode(i,r.cannotAccessVariablesFromNestedScopes))}var o,s=!1;if(p[t].usages.forEach((function(n){2===n.usage&&(s=!0,106500&n.symbol.flags&&n.symbol.valueDeclaration&&e.hasEffectiveModifier(n.symbol.valueDeclaration,64)&&(o=n.symbol.valueDeclaration))})),e.Debug.assert(y(n.range)||0===h.length,"No variable declarations expected if something was extracted"),s&&!y(n.range)){var c=e.createDiagnosticForNode(n.range,r.cannotWriteInExpression);f[t].push(c),_[t].push(c)}else if(o&&t>0){c=e.createDiagnosticForNode(o,r.cannotExtractReadonlyPropertyInitializerOutsideConstructor);f[t].push(c),_[t].push(c)}else if(l){c=e.createDiagnosticForNode(l,r.cannotExtractExportedEntity);f[t].push(c),_[t].push(c)}},V=0;V<t.length;V++)U(V);return{target:I,usagesPerScope:p,functionErrorsPerScope:f,constantErrorsPerScope:_,exposedVariableDeclarations:h};function j(n){return!!e.findAncestor(n,(function(n){return e.isDeclarationWithTypeParameters(n)&&0!==e.getEffectiveTypeParameterDeclarations(n).length}))}function H(e){for(var n=0,t=s.getSymbolWalker((function(){return c.throwIfCancellationRequested(),!0})).walkType(e).visitedTypes;n<t.length;n++){var r=t[n];r.isTypeParameter()&&d.set(r.id.toString(),r)}}function K(n,t){(void 0===t&&(t=1),R)&&H(s.getTypeAtLocation(n));if(e.isDeclaration(n)&&n.symbol&&g.push(n),e.isAssignmentExpression(n))K(n.left,2),K(n.right);else if(e.isUnaryExpressionWithWrite(n))K(n.operand,2);else if(e.isPropertyAccessExpression(n)||e.isElementAccessExpression(n))e.forEachChild(n,K);else if(e.isIdentifier(n)){if(!n.parent)return;if(e.isQualifiedName(n.parent)&&n!==n.parent.left)return;if(e.isPropertyAccessExpression(n.parent)&&n!==n.parent.expression)return;W(n,t,e.isPartOfTypeNode(n))}else e.forEachChild(n,K)}function W(n,r,a){var i=z(n,r,a);if(i)for(var o=0;o<t.length;o++){var s=m[o].get(i);s&&p[o].substitutions.set(e.getNodeId(n).toString(),s)}}function z(c,l,u){var d=J(c);if(d){var g=e.getSymbolId(d).toString(),y=C.get(g);if(y&&y>=l)return g;if(C.set(g,l),y){for(var v=0,h=p;v<h.length;v++){var b=h[v];b.usages.get(c.text)&&b.usages.set(c.text,{usage:l,symbol:d,node:c})}return g}var E=d.getDeclarations(),T=E&&e.find(E,(function(e){return e.getSourceFile()===o}));if(T&&!e.rangeContainsStartEnd(i,T.getStart(),T.end)){if(n.facts&a.IsGenerator&&2===l){for(var S=e.createDiagnosticForNode(c,r.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators),x=0,L=f;x<L.length;x++){L[x].push(S)}for(var A=0,D=_;A<D.length;A++){D[A].push(S)}}for(var I=0;I<t.length;I++){var N=t[I];if(s.resolveName(d.name,N,d.flags,!1)!==d&&!m[I].has(g)){var R=X(d.exportSymbol||d,N,u);if(R)m[I].set(g,R);else if(u){if(!(262144&d.flags)){S=e.createDiagnosticForNode(c,r.typeWillNotBeVisibleInTheNewScope);f[I].push(S),_[I].push(S)}}else p[I].usages.set(c.text,{usage:l,symbol:d,node:c})}}return g}}}function q(t){if(!(t===n.range||y(n.range)&&n.range.indexOf(t)>=0)){var r=e.isIdentifier(t)?J(t):s.getSymbolAtLocation(t);if(r){var a=e.find(g,(function(e){return e.symbol===r}));if(a)if(e.isVariableDeclaration(a)){var i=a.symbol.id.toString();v.has(i)||(h.push(a),v.set(i,!0))}else l=l||a}e.forEachChild(t,q)}}function J(n){return n.parent&&e.isShorthandPropertyAssignment(n.parent)&&n.parent.name===n?s.getShorthandAssignmentValueSymbol(n.parent):s.getSymbolAtLocation(n)}function X(n,t,r){if(n){var a=n.getDeclarations();if(a&&a.some((function(e){return e.parent===t})))return e.factory.createIdentifier(n.name);var i=X(n.parent,t,r);if(void 0!==i)return r?e.factory.createQualifiedName(i,e.factory.createIdentifier(n.name)):e.factory.createPropertyAccessExpression(i,n.name)}}}(n,o,function(n,t){return y(n.range)?{pos:e.first(n.range).getStart(t),end:e.last(n.range).getEnd()}:n.range}(n,i),i,t.program.getTypeChecker(),t.cancellationToken)}}function f(e){var n,t=e.symbol;if(t&&t.declarations)for(var r=0,a=t.declarations;r<a.length;r++){var i=a[r];(void 0===n||i.pos<n.pos)&&(n=i)}return n}function _(n,t){var r=n.type,a=n.declaration,i=t.type,o=t.declaration;return e.compareProperties(a,o,"pos",e.compareValues)||e.compareStringsCaseSensitive(r.symbol?r.symbol.getName():"",i.symbol?i.symbol.getName():"")||e.compareValues(r.id,i.id)}function g(n,t){var r=e.map(n,(function(n){return e.factory.createShorthandPropertyAssignment(n.symbol.name)})),a=e.map(t,(function(n){return e.factory.createShorthandPropertyAssignment(n.symbol.name)}));return void 0===r?a:void 0===a?r:r.concat(a)}function y(n){return e.isArray(n)}function v(e){var n=e.parent;switch(n.kind){case 291:return!1}switch(e.kind){case 10:return 261!==n.kind&&265!==n.kind;case 220:case 196:case 198:return!1;case 78:return 198!==n.kind&&265!==n.kind&&270!==n.kind}return!0}function h(e){switch(e.kind){case 230:case 297:case 257:case 284:return!0;default:return!1}}function b(n){return(e.isJsxElement(n)||e.isJsxSelfClosingElement(n)||e.isJsxFragment(n))&&e.isJsxElement(n.parent)}n.registerRefactor(s,{getAvailableActions:c,getEditsForAction:l}),t.getAvailableActions=c,t.getEditsForAction=l,function(n){function t(n){return{message:n,code:0,category:e.DiagnosticCategory.Message,key:n}}n.cannotExtractRange=t("Cannot extract range."),n.cannotExtractImport=t("Cannot extract import statement."),n.cannotExtractSuper=t("Cannot extract super call."),n.cannotExtractJSDoc=t("Cannot extract JSDoc."),n.cannotExtractEmpty=t("Cannot extract empty range."),n.expressionExpected=t("expression expected."),n.uselessConstantType=t("No reason to extract constant of type."),n.statementOrExpressionExpected=t("Statement or expression expected."),n.cannotExtractRangeContainingConditionalBreakOrContinueStatements=t("Cannot extract range containing conditional break or continue statements."),n.cannotExtractRangeContainingConditionalReturnStatement=t("Cannot extract range containing conditional return statement."),n.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange=t("Cannot extract range containing labeled break or continue with target outside of the range."),n.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators=t("Cannot extract range containing writes to references located outside of the target range in generators."),n.typeWillNotBeVisibleInTheNewScope=t("Type will not visible in the new scope."),n.functionWillNotBeVisibleInTheNewScope=t("Function will not visible in the new scope."),n.cannotExtractIdentifier=t("Select more than a single identifier."),n.cannotExtractExportedEntity=t("Cannot extract exported declaration"),n.cannotWriteInExpression=t("Cannot write back side-effects when extracting an expression"),n.cannotExtractReadonlyPropertyInitializerOutsideConstructor=t("Cannot move initialization of read-only class property outside of the constructor"),n.cannotExtractAmbientBlock=t("Cannot extract code from ambient contexts"),n.cannotAccessVariablesFromNestedScopes=t("Cannot access variables from nested scopes"),n.cannotExtractToOtherFunctionLike=t("Cannot extract method to a function-like scope that is not a function"),n.cannotExtractToJSClass=t("Cannot extract constant to a class scope in JS"),n.cannotExtractToExpressionArrowFunction=t("Cannot extract constant to an arrow function without a block")}(r=t.Messages||(t.Messages={})),function(e){e[e.None=0]="None",e[e.HasReturn=1]="HasReturn",e[e.IsGenerator=2]="IsGenerator",e[e.IsAsyncFunction=4]="IsAsyncFunction",e[e.UsesThis=8]="UsesThis",e[e.InStaticRegion=16]="InStaticRegion"}(a||(a={})),t.getRangeToExtract=u,function(e){e[e.Module=0]="Module",e[e.Global=1]="Global"}(i||(i={})),function(e){e[e.Read=1]="Read",e[e.Write=2]="Write"}(o||(o={}))}(n.extractSymbol||(n.extractSymbol={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){var t="Extract type",r="Extract to type alias",a="Extract to interface",i="Extract to typedef";function o(n,t){void 0===t&&(t=!0);var r=n.file,a=n.startPosition,i=e.isSourceFileJS(r),o=e.getTokenAtPosition(r,a),l=e.createTextRangeFromSpan(e.getRefactorContextSpan(n)),u=l.pos===l.end&&t,d=e.findAncestor(o,(function(n){return n.parent&&e.isTypeNode(n)&&!c(l,n.parent,r)&&(u||e.nodeOverlapsWithStartEnd(o,r,l.pos,l.end))}));if(!d||!e.isTypeNode(d))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Selection_is_not_a_valid_type_node)};var p=n.program.getTypeChecker(),m=e.Debug.checkDefined(e.findAncestor(d,e.isStatement),"Should find a statement"),f=function(n,t,r,a){var i=[];return o(t)?void 0:i;function o(s){if(e.isTypeReferenceNode(s)){if(e.isIdentifier(s.typeName)&&(p=n.resolveName(s.typeName.text,s.typeName,262144,!0))){var l=e.cast(e.first(p.declarations),e.isTypeParameterDeclaration);c(r,l,a)&&!c(t,l,a)&&e.pushIfUnique(i,l)}}else if(e.isInferTypeNode(s)){var u=e.findAncestor(s,(function(n){return e.isConditionalTypeNode(n)&&c(n.extendsType,s,a)}));if(!u||!c(t,u,a))return!0}else if(e.isTypePredicateNode(s)||e.isThisTypeNode(s)){var d=e.findAncestor(s.parent,e.isFunctionLike);if(d&&d.type&&c(d.type,s,a)&&!c(t,d,a))return!0}else if(e.isTypeQueryNode(s)){var p;if(e.isIdentifier(s.exprName)){if((p=n.resolveName(s.exprName.text,s.exprName,111551,!1))&&c(r,p.valueDeclaration,a)&&!c(t,p.valueDeclaration,a))return!0}else if(e.isThisIdentifier(s.exprName.left)&&!c(t,s.parent,a))return!0}return a&&e.isTupleTypeNode(s)&&e.getLineAndCharacterOfPosition(a,s.pos).line===e.getLineAndCharacterOfPosition(a,s.end).line&&e.setEmitFlags(s,1),e.forEachChild(s,o)}}(p,d,m,r);return f?{info:{isJS:i,selection:d,firstStatement:m,typeParameters:f,typeElements:s(p,d)}}:{error:e.getLocaleSpecificMessage(e.Diagnostics.No_type_could_be_extracted_from_this_type_node)}}function s(n,t){if(t){if(e.isIntersectionTypeNode(t)){for(var r=[],a=new e.Map,i=0,o=t.types;i<o.length;i++){var c=s(n,o[i]);if(!c||!c.every((function(n){return n.name&&e.addToSeen(a,e.getNameFromPropertyName(n.name))})))return;e.addRange(r,c)}return r}return e.isParenthesizedTypeNode(t)?s(n,t.type):e.isTypeLiteralNode(t)?t.members:void 0}}function c(n,t,r){return e.rangeContainsStartEnd(n,e.skipTrivia(r.text,t.pos),t.end)}n.registerRefactor(t,{getAvailableActions:function(n){var s=o(n,"invoked"===n.triggerReason);return s?void 0===s.error?[{name:t,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_type),actions:s.info.isJS?[{name:i,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_typedef)}]:e.append([{name:r,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_type_alias)}],s.info.typeElements&&{name:a,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_interface)})}]:n.preferences.provideRefactorNotApplicableReason?[{name:t,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_type),actions:[{name:i,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_typedef),notApplicableReason:s.error},{name:r,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_type_alias),notApplicableReason:s.error},{name:a,description:e.getLocaleSpecificMessage(e.Diagnostics.Extract_to_interface),notApplicableReason:s.error}]}]:e.emptyArray:e.emptyArray},getEditsForAction:function(n,t){var s,c=n.file,l=e.Debug.checkDefined(null===(s=o(n))||void 0===s?void 0:s.info,"Expected to find a range to extract"),u=e.getUniqueName("NewType",c),d=e.textChanges.ChangeTracker.with(n,(function(n){switch(t){case r:return e.Debug.assert(!l.isJS,"Invalid actionName/JS combo"),function(n,t,r,a){var i=a.firstStatement,o=a.selection,s=a.typeParameters,c=e.factory.createTypeAliasDeclaration(void 0,void 0,r,s.map((function(n){return e.factory.updateTypeParameterDeclaration(n,n.name,n.constraint,void 0)})),o);n.insertNodeBefore(t,i,e.ignoreSourceNewlines(c),!0),n.replaceNode(t,o,e.factory.createTypeReferenceNode(r,s.map((function(n){return e.factory.createTypeReferenceNode(n.name,void 0)}))),{leadingTriviaOption:e.textChanges.LeadingTriviaOption.Exclude,trailingTriviaOption:e.textChanges.TrailingTriviaOption.ExcludeWhitespace})}(n,c,u,l);case i:return e.Debug.assert(l.isJS,"Invalid actionName/JS combo"),function(n,t,r,a){var i=a.firstStatement,o=a.selection,s=a.typeParameters,c=e.factory.createJSDocTypedefTag(e.factory.createIdentifier("typedef"),e.factory.createJSDocTypeExpression(o),e.factory.createIdentifier(r)),l=[];e.forEach(s,(function(n){var t=e.getEffectiveConstraintOfTypeParameter(n),r=e.factory.createTypeParameterDeclaration(n.name),a=e.factory.createJSDocTemplateTag(e.factory.createIdentifier("template"),t&&e.cast(t,e.isJSDocTypeExpression),[r]);l.push(a)})),n.insertNodeBefore(t,i,e.factory.createJSDocComment(void 0,e.factory.createNodeArray(e.concatenate(l,[c]))),!0),n.replaceNode(t,o,e.factory.createTypeReferenceNode(r,s.map((function(n){return e.factory.createTypeReferenceNode(n.name,void 0)}))))}(n,c,u,l);case a:return e.Debug.assert(!l.isJS&&!!l.typeElements,"Invalid actionName/JS combo"),function(n,t,r,a){var i,o=a.firstStatement,s=a.selection,c=a.typeParameters,l=a.typeElements,u=e.factory.createInterfaceDeclaration(void 0,void 0,r,c,void 0,l);e.setTextRange(u,null===(i=l[0])||void 0===i?void 0:i.parent),n.insertNodeBefore(t,o,e.ignoreSourceNewlines(u),!0),n.replaceNode(t,s,e.factory.createTypeReferenceNode(r,c.map((function(n){return e.factory.createTypeReferenceNode(n.name,void 0)}))),{leadingTriviaOption:e.textChanges.LeadingTriviaOption.Exclude,trailingTriviaOption:e.textChanges.TrailingTriviaOption.ExcludeWhitespace})}(n,c,u,l);default:e.Debug.fail("Unexpected action name")}})),p=c.fileName;return{edits:d,renameFilename:p,renameLocation:e.getRenameLocation(d,p,u,!1)}}})}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){var t,r;n.generateGetAccessorAndSetAccessor||(n.generateGetAccessorAndSetAccessor={}),t="Generate 'get' and 'set' accessors",r=e.Diagnostics.Generate_get_and_set_accessors.message,n.registerRefactor(t,{getEditsForAction:function(n,t){if(n.endPosition){var r=e.codefix.getAccessorConvertiblePropertyAtPosition(n.file,n.program,n.startPosition,n.endPosition);if(r&&r.info){var a=e.codefix.generateAccessorFromProperty(n.file,n.program,n.startPosition,n.endPosition,n,t);if(a){var i=n.file.fileName,o=r.info.renameAccessor?r.info.accessorName:r.info.fieldName;return{renameFilename:i,renameLocation:(e.isIdentifier(o)?0:-1)+e.getRenameLocation(a,i,o.text,e.isParameter(r.info.declaration)),edits:a}}}}},getAvailableActions:function(n){if(!n.endPosition)return e.emptyArray;var a=e.codefix.getAccessorConvertiblePropertyAtPosition(n.file,n.program,n.startPosition,n.endPosition,"invoked"===n.triggerReason);return a?a.error?n.preferences.provideRefactorNotApplicableReason?[{name:t,description:r,actions:[{name:t,description:r,notApplicableReason:a.error}]}]:e.emptyArray:[{name:t,description:r,actions:[{name:t,description:r}]}]:e.emptyArray}})}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){var t="Move to a new file";function r(n){var t=function(n){var t=n.file,r=e.createTextRangeFromSpan(e.getRefactorContextSpan(n)),a=t.statements,i=e.findIndex(a,(function(e){return e.end>r.pos}));if(-1!==i){var o=a[i];if(e.isNamedDeclaration(o)&&o.name&&e.rangeContainsRange(o.name,r))return{toMove:[a[i]],afterLast:a[i+1]};if(!(r.pos>o.getStart(t))){var s=e.findIndex(a,(function(e){return e.end>r.end}),i);if(-1===s||!(0===s||a[s].getStart(t)<r.end))return{toMove:a.slice(i,-1===s?a.length:s),afterLast:-1===s?void 0:a[s]}}}}(n);if(void 0!==t){var r=[],i=[],o=t.toMove,s=t.afterLast;return e.getRangesWhere(o,a,(function(e,n){for(var t=e;t<n;t++)r.push(o[t]);i.push({first:o[e],afterLast:s})})),0===r.length?void 0:{all:r,ranges:i}}}function a(n){return!function(n){switch(n.kind){case 261:return!0;case 260:return!e.hasSyntacticModifier(n,1);case 232:return n.declarationList.declarations.every((function(n){return!!n.initializer&&e.isRequireCall(n.initializer,!0)}));default:return!1}}(n)&&!e.isPrologueDirective(n)}function i(e,n,t){for(var r=0,a=n;r<a.length;r++){var i=a[r],o=i.first,s=i.afterLast;t.deleteNodeRangeExcludingEnd(e,o,s)}}function o(e){return 261===e.kind?e.moduleSpecifier:260===e.kind?e.moduleReference.expression:e.initializer.arguments[0]}function s(n,t){if(e.isImportDeclaration(n))e.isStringLiteral(n.moduleSpecifier)&&t(n);else if(e.isImportEqualsDeclaration(n))e.isExternalModuleReference(n.moduleReference)&&e.isStringLiteralLike(n.moduleReference.expression)&&t(n);else if(e.isVariableStatement(n))for(var r=0,a=n.declarationList.declarations;r<a.length;r++){var i=a[r];i.initializer&&e.isRequireCall(i.initializer,!0)&&t(i)}}function c(n,t,r,a,i){if(r=e.ensurePathIsNonModuleName(r),a){var o=t.map((function(n){return e.factory.createImportSpecifier(void 0,e.factory.createIdentifier(n))}));return e.makeImportIfNecessary(n,o,r,i)}e.Debug.assert(!n,"No default import should exist");var s=t.map((function(n){return e.factory.createBindingElement(void 0,void 0,n)}));return s.length?l(e.factory.createObjectBindingPattern(s),void 0,u(e.factory.createStringLiteral(r))):void 0}function l(n,t,r,a){return void 0===a&&(a=2),e.factory.createVariableStatement(void 0,e.factory.createVariableDeclarationList([e.factory.createVariableDeclaration(n,void 0,t,r)],a))}function u(n){return e.factory.createCallExpression(e.factory.createIdentifier("require"),void 0,[n])}function d(n,t,r,a){switch(t.kind){case 261:!function(n,t,r,a){if(!t.importClause)return;var i=t.importClause,o=i.name,s=i.namedBindings,c=!o||a(o),l=!s||(263===s.kind?a(s.name):0!==s.elements.length&&s.elements.every((function(e){return a(e.name)})));if(c&&l)r.delete(n,t);else if(o&&c&&r.delete(n,o),s)if(l)r.replaceNode(n,t.importClause,e.factory.updateImportClause(t.importClause,t.importClause.isTypeOnly,o,void 0));else if(264===s.kind)for(var u=0,d=s.elements;u<d.length;u++){var p=d[u];a(p.name)&&r.delete(n,p)}}(n,t,r,a);break;case 260:a(t.name)&&r.delete(n,t);break;case 249:!function(n,t,r,a){var i=t.name;switch(i.kind){case 78:a(i)&&r.delete(n,i);break;case 197:break;case 196:if(i.elements.every((function(n){return e.isIdentifier(n.name)&&a(n.name)})))r.delete(n,e.isVariableDeclarationList(t.parent)&&1===t.parent.declarations.length?t.parent.parent:t);else for(var o=0,s=i.elements;o<s.length;o++){var c=s[o];e.isIdentifier(c.name)&&a(c.name)&&r.delete(n,c.name)}}}(n,t,r,a);break;default:e.Debug.assertNever(t,"Unexpected import decl kind "+t.kind)}}function p(n){switch(n.kind){case 260:case 265:case 262:case 263:return!0;case 249:return m(n);case 198:return e.isVariableDeclaration(n.parent.parent)&&m(n.parent.parent);default:return!1}}function m(n){return e.isSourceFile(n.parent.parent.parent)&&!!n.initializer&&e.isRequireCall(n.initializer,!0)}function f(n,t,r){switch(n.kind){case 261:var a=n.importClause;if(!a)return;var i=a.name&&r(a.name)?a.name:void 0,o=a.namedBindings&&function(n,t){if(263===n.kind)return t(n.name)?n:void 0;var r=n.elements.filter((function(e){return t(e.name)}));return r.length?e.factory.createNamedImports(r):void 0}(a.namedBindings,r);return i||o?e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,i,o),t):void 0;case 260:return r(n.name)?n:void 0;case 249:var s=function(n,t){switch(n.kind){case 78:return t(n)?n:void 0;case 197:return n;case 196:var r=n.elements.filter((function(n){return n.propertyName||!e.isIdentifier(n.name)||t(n.name)}));return r.length?e.factory.createObjectBindingPattern(r):void 0}}(n.name,r);return s?l(s,n.type,u(t),n.parent.flags):void 0;default:return e.Debug.assertNever(n,"Unexpected import kind "+n.kind)}}function _(n,t,r){n.forEachChild((function n(a){if(e.isIdentifier(a)&&!e.isDeclarationName(a)){var i=t.getSymbolAtLocation(a);i&&r(i)}else a.forEachChild(n)}))}n.registerRefactor(t,{getAvailableActions:function(n){if(!n.preferences.allowTextChangesInNewFiles||void 0===r(n))return e.emptyArray;var a=e.getLocaleSpecificMessage(e.Diagnostics.Move_to_a_new_file);return[{name:t,description:a,actions:[{name:t,description:a}]}]},getEditsForAction:function(n,a){e.Debug.assert(a===t,"Wrong refactor invoked");var l=e.Debug.checkDefined(r(n));return{edits:e.textChanges.ChangeTracker.with(n,(function(t){return r=n.file,a=n.program,m=l,E=t,I=n.host,N=n.preferences,R=a.getTypeChecker(),k=function(n,t,r){var a=new g,i=new g,o=new g,s=S(e.find(t,(function(e){return!!(2&e.transformFlags)})));s&&i.add(s);for(var c=0,l=t;c<l.length;c++)b(T=l[c],(function(n){a.add(e.Debug.checkDefined(e.isExpressionStatement(n)?r.getSymbolAtLocation(n.expression.left):n.symbol,"Need a symbol here"))}));for(var u=0,d=t;u<d.length;u++)_(T=d[u],r,(function(e){if(e.declarations)for(var t=0,r=e.declarations;t<r.length;t++){var s=r[t];p(s)?i.add(e):y(s)&&v(s)===n&&!a.has(e)&&o.add(e)}}));for(var m=i.clone(),f=new g,h=0,E=n.statements;h<E.length;h++){var T=E[h];e.contains(t,T)||(s&&2&T.transformFlags&&m.delete(s),_(T,r,(function(e){a.has(e)&&f.add(e),m.delete(e)})))}return{movedSymbols:a,newFileImportsFromOldFile:o,oldFileImportsFromNewFile:f,oldImportsNeededByNewFile:i,unusedImportsFromOldFile:m};function S(n){if(void 0!==n){var t=r.getJsxNamespace(n),a=r.resolveName(t,n,1920,!0);return a&&e.some(a.declarations,p)?a:void 0}}}(r,m.all,R),O=e.getDirectoryPath(r.fileName),P=e.extensionFromPath(r.fileName),M=function(n,t,r,a){for(var i=n,o=1;;o++){var s=e.combinePaths(r,i+t);if(!a.fileExists(s))return i;i=n+"."+o}}(k.movedSymbols.forEachEntry(e.symbolNameNoDefault)||"newFile",P,O,I),w=M+P,E.createNewFile(r,e.combinePaths(O,w),function(n,t,r,a,l,p,m){var _=l.getTypeChecker(),g=e.takeWhile(n.statements,e.isPrologueDirective);if(!n.externalModuleIndicator&&!n.commonJsModuleIndicator)return i(n,a.ranges,r),__spreadArrays(g,a.all);var v=!!n.externalModuleIndicator,E=e.getQuotePreference(n,m),I=function(n,t,r,a){var i,o=[];return n.forEach((function(n){"default"===n.escapedName?i=e.factory.createIdentifier(e.symbolNameNoDefault(n)):o.push(n.name)})),c(i,o,t,r,a)}(t.oldFileImportsFromNewFile,p,v,E);I&&e.insertImports(r,n,I,!0),function(n,t,r,a,i){for(var o=0,c=n.statements;o<c.length;o++){var l=c[o];e.contains(t,l)||s(l,(function(e){return d(n,e,r,(function(e){return a.has(i.getSymbolAtLocation(e))}))}))}}(n,a.all,r,t.unusedImportsFromOldFile,_),i(n,a.ranges,r),function(n,t,r,a,i){for(var c=t.getTypeChecker(),l=function(t){if(t===r)return"continue";for(var l=function(l){s(l,(function(s){if(c.getSymbolAtLocation(o(s))===r.symbol){var p=function(n){var t=e.isBindingElement(n.parent)?e.getPropertySymbolFromBindingElement(c,n.parent):e.skipAlias(c.getSymbolAtLocation(n),c);return!!t&&a.has(t)};d(t,s,n,p);var m=e.combinePaths(e.getDirectoryPath(o(s).text),i),_=f(s,e.factory.createStringLiteral(m),p);_&&n.insertNodeAfter(t,l,_);var g=function(n){switch(n.kind){case 261:return n.importClause&&n.importClause.namedBindings&&263===n.importClause.namedBindings.kind?n.importClause.namedBindings.name:void 0;case 260:return n.name;case 249:return e.tryCast(n.name,e.isIdentifier);default:return e.Debug.assertNever(n,"Unexpected node kind "+n.kind)}}(s);g&&function(n,t,r,a,i,o,s,c){var l=e.codefix.moduleSpecifierToValidIdentifier(i,99),d=!1,p=[];if(e.FindAllReferences.Core.eachSymbolReferenceInFile(s,r,t,(function(n){e.isPropertyAccessExpression(n.parent)&&(d=d||!!r.resolveName(l,n,67108863,!0),a.has(r.getSymbolAtLocation(n.parent.name))&&p.push(n))})),p.length){for(var m=d?e.getUniqueName(l,t):l,f=0,_=p;f<_.length;f++){var g=_[f];n.replaceNode(t,g,e.factory.createIdentifier(m))}n.insertNodeAfter(t,c,function(n,t,r){var a=e.factory.createIdentifier(t),i=e.factory.createStringLiteral(r);switch(n.kind){case 261:return e.factory.createImportDeclaration(void 0,void 0,e.factory.createImportClause(!1,void 0,e.factory.createNamespaceImport(a)),i);case 260:return e.factory.createImportEqualsDeclaration(void 0,void 0,a,e.factory.createExternalModuleReference(i));case 249:return e.factory.createVariableDeclaration(a,void 0,void 0,u(i));default:return e.Debug.assertNever(n,"Unexpected node kind "+n.kind)}}(c,i,o))}}(n,t,c,a,i,m,g,s)}}))},p=0,m=t.statements;p<m.length;p++)l(m[p])},p=0,m=t.getSourceFiles();p<m.length;p++)l(m[p])}(r,l,n,t.movedSymbols,p);var N=function(n,t,r,a,i,l,u){for(var d,p=[],m=0,_=n.statements;m<_.length;m++)s(_[m],(function(n){e.append(p,f(n,o(n),(function(e){return t.has(i.getSymbolAtLocation(e))})))}));var g=[],v=e.nodeSeenTracker();return r.forEach((function(t){for(var r=0,i=t.declarations;r<i.length;r++){var o=i[r];if(y(o)){var s=T(o);if(s){var c=S(o);v(c)&&x(n,c,a,l),e.hasSyntacticModifier(o,512)?d=s:g.push(s.text)}}}})),e.append(p,c(d,g,e.removeFileExtension(e.getBaseFileName(n.fileName)),l,u)),p}(n,t.oldImportsNeededByNewFile,t.newFileImportsFromOldFile,r,_,v,E),R=function(n,t,r,a){return e.flatMap(t,(function(t){if(o=t,e.Debug.assert(e.isSourceFile(o.parent),"Node parent should be a SourceFile"),(h(o)||e.isVariableStatement(o))&&!L(n,t,a)&&b(t,(function(n){return r.has(e.Debug.checkDefined(n.symbol))}))){var i=function(e,n){return n?[A(e)]:function(e){return __spreadArrays([e],D(e).map(C))}(e)}(t,a);if(i)return i}var o;return t}))}(n,a.all,t.oldFileImportsFromNewFile,v);return N.length&&R.length?__spreadArrays(g,N,[4],R):__spreadArrays(g,N,R)}(r,k,E,m,a,M,N)),void function(n,t,r,a,i){var o=n.getCompilerOptions().configFile;if(o){var s=e.normalizePath(e.combinePaths(r,"..",a)),c=e.getRelativePathFromFile(o.fileName,s,i),l=o.statements[0]&&e.tryCast(o.statements[0].expression,e.isObjectLiteralExpression),u=l&&e.find(l.properties,(function(n){return e.isPropertyAssignment(n)&&e.isStringLiteral(n.name)&&"files"===n.name.text}));u&&e.isArrayLiteralExpression(u.initializer)&&t.insertNodeInListAfter(o,e.last(u.initializer.elements),e.factory.createStringLiteral(c),u.initializer.elements)}}(a,E,r.fileName,w,e.hostGetCanonicalFileName(I));var r,a,m,E,I,N,R,k,O,P,M,w})),renameFilename:void 0,renameLocation:void 0}}});var g=function(){function n(){this.map=new e.Map}return n.prototype.add=function(n){this.map.set(String(e.getSymbolId(n)),n)},n.prototype.has=function(n){return this.map.has(String(e.getSymbolId(n)))},n.prototype.delete=function(n){this.map.delete(String(e.getSymbolId(n)))},n.prototype.forEach=function(e){this.map.forEach(e)},n.prototype.forEachEntry=function(n){return e.forEachEntry(this.map,n)},n.prototype.clone=function(){var t=new n;return e.copyEntries(this.map,t.map),t},n}();function y(n){return h(n)&&e.isSourceFile(n.parent)||e.isVariableDeclaration(n)&&e.isSourceFile(n.parent.parent.parent)}function v(n){return e.isVariableDeclaration(n)?n.parent.parent.parent:n.parent}function h(e){switch(e.kind){case 251:case 252:case 256:case 255:case 254:case 253:case 260:return!0;default:return!1}}function b(n,t){switch(n.kind){case 251:case 252:case 256:case 255:case 254:case 253:case 260:return t(n);case 232:return e.firstDefined(n.declarationList.declarations,(function(e){return E(e.name,t)}));case 233:var r=n.expression;return e.isBinaryExpression(r)&&1===e.getAssignmentDeclarationKind(r)?t(n):void 0}}function E(n,t){switch(n.kind){case 78:return t(e.cast(n.parent,(function(n){return e.isVariableDeclaration(n)||e.isBindingElement(n)})));case 197:case 196:return e.firstDefined(n.elements,(function(n){return e.isOmittedExpression(n)?void 0:E(n.name,t)}));default:return e.Debug.assertNever(n,"Unexpected name kind "+n.kind)}}function T(n){return e.isExpressionStatement(n)?e.tryCast(n.expression.left.name,e.isIdentifier):e.tryCast(n.name,e.isIdentifier)}function S(n){switch(n.kind){case 249:return n.parent.parent;case 198:return S(e.cast(n.parent.parent,(function(n){return e.isVariableDeclaration(n)||e.isBindingElement(n)})));default:return n}}function x(n,t,r,a){if(!L(n,t,a))if(a)e.isExpressionStatement(t)||r.insertExportModifier(n,t);else{var i=D(t);0!==i.length&&r.insertNodesAfter(n,t,i.map(C))}}function L(n,t,r){return r?!e.isExpressionStatement(t)&&e.hasSyntacticModifier(t,1):D(t).some((function(t){return n.symbol.exports.has(e.escapeLeadingUnderscores(t))}))}function A(n){var t=e.concatenate([e.factory.createModifier(92)],n.modifiers);switch(n.kind){case 251:return e.factory.updateFunctionDeclaration(n,n.decorators,t,n.asteriskToken,n.name,n.typeParameters,n.parameters,n.type,n.body);case 252:return e.factory.updateClassDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members);case 232:return e.factory.updateVariableStatement(n,t,n.declarationList);case 256:return e.factory.updateModuleDeclaration(n,n.decorators,t,n.name,n.body);case 255:return e.factory.updateEnumDeclaration(n,n.decorators,t,n.name,n.members);case 254:return e.factory.updateTypeAliasDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.type);case 253:return e.factory.updateInterfaceDeclaration(n,n.decorators,t,n.name,n.typeParameters,n.heritageClauses,n.members);case 260:return e.factory.updateImportEqualsDeclaration(n,n.decorators,t,n.name,n.moduleReference);case 233:return e.Debug.fail();default:return e.Debug.assertNever(n,"Unexpected declaration kind "+n.kind)}}function D(n){switch(n.kind){case 251:case 252:return[n.name.text];case 232:return e.mapDefined(n.declarationList.declarations,(function(n){return e.isIdentifier(n.name)?n.name.text:void 0}));case 256:case 255:case 254:case 253:case 260:return e.emptyArray;case 233:return e.Debug.fail("Can't export an ExpressionStatement");default:return e.Debug.assertNever(n,"Unexpected decl kind "+n.kind)}}function C(n){return e.factory.createExpressionStatement(e.factory.createBinaryExpression(e.factory.createPropertyAccessExpression(e.factory.createIdentifier("exports"),e.factory.createIdentifier(n)),62,e.factory.createIdentifier(n)))}}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Add or remove braces in an arrow function",a=e.Diagnostics.Add_or_remove_braces_in_an_arrow_function.message,i="Add braces to arrow function",o="Remove braces from arrow function",s=e.Diagnostics.Add_braces_to_arrow_function.message,c=e.Diagnostics.Remove_braces_from_arrow_function.message;function l(n,t,r){void 0===r&&(r=!0);var a=e.getTokenAtPosition(n,t),i=e.getContainingFunction(a);if(!i)return{error:e.getLocaleSpecificMessage(e.Diagnostics.Could_not_find_a_containing_arrow_function)};if(!e.isArrowFunction(i))return{error:e.getLocaleSpecificMessage(e.Diagnostics.Containing_function_is_not_an_arrow_function)};if(e.rangeContainsRange(i,a)&&(!e.rangeContainsRange(i.body,a)||r)){if(e.isExpression(i.body))return{info:{func:i,addBraces:!0,expression:i.body}};if(1===i.body.statements.length){var o=e.first(i.body.statements);if(e.isReturnStatement(o))return{info:{func:i,addBraces:!1,expression:o.expression,returnStatement:o}}}}}n.registerRefactor(r,{getEditsForAction:function(n,t){var r=n.file,a=n.startPosition,s=l(r,a);if(!s||!s.info)return;var c,u=s.info,d=u.expression,p=u.returnStatement,m=u.func;if(t===i){var f=e.factory.createReturnStatement(d);c=e.factory.createBlock([f],!0),e.suppressLeadingAndTrailingTrivia(c),e.copyLeadingComments(d,f,r,3,!0)}else if(t===o&&p){var _=d||e.factory.createVoidZero();c=e.needsParentheses(_)?e.factory.createParenthesizedExpression(_):_,e.suppressLeadingAndTrailingTrivia(c),e.copyTrailingAsLeadingComments(p,c,r,3,!1),e.copyLeadingComments(p,c,r,3,!1),e.copyTrailingComments(p,c,r,3,!1)}else e.Debug.fail("invalid action");var g=e.textChanges.ChangeTracker.with(n,(function(e){e.replaceNode(r,m.body,c)}));return{renameFilename:void 0,renameLocation:void 0,edits:g}},getAvailableActions:function(n){var t=n.file,u=n.startPosition,d=n.triggerReason,p=l(t,u,"invoked"===d);if(!p)return e.emptyArray;if(void 0===p.error)return[{name:r,description:a,actions:[p.info.addBraces?{name:i,description:s}:{name:o,description:c}]}];if(n.preferences.provideRefactorNotApplicableReason)return[{name:r,description:a,actions:[{name:i,description:s,notApplicableReason:p.error},{name:o,description:c,notApplicableReason:p.error}]}];return e.emptyArray}})}(n.addOrRemoveBracesToArrowFunction||(n.addOrRemoveBracesToArrowFunction={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Convert parameters to destructured object";function a(n){var t=n.node;return e.isImportSpecifier(t.parent)||e.isImportClause(t.parent)||e.isImportEqualsDeclaration(t.parent)||e.isNamespaceImport(t.parent)||e.isExportSpecifier(t.parent)||e.isExportAssignment(t.parent)?t:void 0}function i(n){if(e.isDeclaration(n.node.parent))return n.node}function o(n){if(n.node.parent){var t=n.node,r=t.parent;switch(r.kind){case 203:case 204:var a=e.tryCast(r,e.isCallOrNewExpression);if(a&&a.expression===t)return a;break;case 201:var i=e.tryCast(r,e.isPropertyAccessExpression);if(i&&i.parent&&i.name===t){var o=e.tryCast(i.parent,e.isCallOrNewExpression);if(o&&o.expression===i)return o}break;case 202:var s=e.tryCast(r,e.isElementAccessExpression);if(s&&s.parent&&s.argumentExpression===t){var c=e.tryCast(s.parent,e.isCallOrNewExpression);if(c&&c.expression===s)return c}}}}function s(n){if(n.node.parent){var t=n.node,r=t.parent;switch(r.kind){case 201:var a=e.tryCast(r,e.isPropertyAccessExpression);if(a&&a.expression===t)return a;break;case 202:var i=e.tryCast(r,e.isElementAccessExpression);if(i&&i.expression===t)return i}}}function c(n){var t=n.node;if(2===e.getMeaningFromLocation(t)||e.isExpressionWithTypeArgumentsInClassExtendsClause(t.parent))return t}function l(n,t,r){var a=e.getTouchingToken(n,t),i=e.getContainingFunctionDeclaration(a);if(!function(n){var t=e.findAncestor(n,e.isJSDocNode);if(t){var r=e.findAncestor(t,(function(n){return!e.isJSDocNode(n)}));return!!r&&e.isFunctionLikeDeclaration(r)}return!1}(a))return!(i&&function(n,t){if(!function(n,t){return function(e){if(m(e))return e.length-1;return e.length}(n)>=2&&e.every(n,(function(n){return function(n,t){if(e.isRestParameter(n)){var r=t.getTypeAtLocation(n);if(!t.isArrayType(r)&&!t.isTupleType(r))return!1}return!n.modifiers&&!n.decorators&&e.isIdentifier(n.name)}(n,t)}))}(n.parameters,t))return!1;switch(n.kind){case 251:return d(n)&&u(n,t);case 165:return u(n,t);case 166:return e.isClassDeclaration(n.parent)?d(n.parent)&&u(n,t):p(n.parent.parent)&&u(n,t);case 208:case 209:return p(n.parent)}return!1}(i,r)&&e.rangeContainsRange(i,a))||i.body&&e.rangeContainsRange(i.body,a)?void 0:i}function u(e,n){return!!e.body&&!n.isImplementationOfOverload(e)}function d(n){return!!n.name||!!e.findModifier(n,87)}function p(n){return e.isVariableDeclaration(n)&&e.isVarConst(n)&&e.isIdentifier(n.name)&&!n.type}function m(n){return n.length>0&&e.isThis(n[0].name)}function f(n){return m(n)&&(n=e.factory.createNodeArray(n.slice(1),n.hasTrailingComma)),n}function _(n,t){var r=f(n.parameters),a=e.isRestParameter(e.last(r)),i=a?t.slice(0,r.length-1):t,o=e.map(i,(function(n,t){var a,i,o=g(r[t]),s=(a=o,i=n,e.isIdentifier(i)&&e.getTextOfIdentifierOrLiteral(i)===a?e.factory.createShorthandPropertyAssignment(a):e.factory.createPropertyAssignment(a,i));return e.suppressLeadingAndTrailingTrivia(s.name),e.isPropertyAssignment(s)&&e.suppressLeadingAndTrailingTrivia(s.initializer),e.copyComments(n,s),s}));if(a&&t.length>=r.length){var s=t.slice(r.length-1),c=e.factory.createPropertyAssignment(g(e.last(r)),e.factory.createArrayLiteralExpression(s));o.push(c)}return e.factory.createObjectLiteralExpression(o,!1)}function g(n){return e.getTextOfIdentifierOrLiteral(n.name)}n.registerRefactor(r,{getEditsForAction:function(n,t){e.Debug.assert(t===r,"Unexpected action name");var u=n.file,d=n.startPosition,p=n.program,y=n.cancellationToken,v=n.host,h=l(u,d,p.getTypeChecker());if(!h||!y)return;var b=function(n,t,r){var l=function(n){switch(n.kind){case 251:return n.name?[n.name]:[e.Debug.checkDefined(e.findModifier(n,87),"Nameless function declaration should be a default export")];case 165:return[n.name];case 166:var t=e.Debug.checkDefined(e.findChildOfKind(n,132,n.getSourceFile()),"Constructor declaration should have constructor keyword");return 221===n.parent.kind?[n.parent.parent.name,t]:[t];case 209:return[n.parent.name];case 208:return n.name?[n.name,n.parent.name]:[n.parent.name];default:return e.Debug.assertNever(n,"Unexpected function declaration kind "+n.kind)}}(n),u=e.isConstructorDeclaration(n)?function(n){switch(n.parent.kind){case 252:var t=n.parent;return t.name?[t.name]:[e.Debug.checkDefined(e.findModifier(t,87),"Nameless class declaration should be a default export")];case 221:var r=n.parent,a=n.parent.parent,i=r.name;return i?[i,a.name]:[a.name]}}(n):[],d=e.deduplicate(__spreadArrays(l,u),e.equateValues),p=t.getTypeChecker(),m=f(e.flatMap(d,(function(n){return e.FindAllReferences.getReferenceEntriesForNode(-1,n,t,t.getSourceFiles(),r)})));e.every(m.declarations,(function(n){return e.contains(d,n)}))||(m.valid=!1);return m;function f(t){for(var r={accessExpressions:[],typeUsages:[]},d={functionCalls:[],declarations:[],classReferences:r,valid:!0},p=e.map(l,_),m=e.map(u,_),f=e.isConstructorDeclaration(n),g=0,y=t;g<y.length;g++){var v=y[g];if(1===v.kind){if(e.contains(p,_(v.node))||e.isNewExpressionTarget(v.node)){if(a(v))continue;if(b=i(v)){d.declarations.push(b);continue}var h=o(v);if(h){d.functionCalls.push(h);continue}}if(f&&e.contains(m,_(v.node))){var b;if(a(v))continue;if(b=i(v)){d.declarations.push(b);continue}var E=s(v);if(E){r.accessExpressions.push(E);continue}if(e.isClassDeclaration(n.parent)){var T=c(v);if(T){r.typeUsages.push(T);continue}}}d.valid=!1}else d.valid=!1}return d}function _(n){var t=p.getSymbolAtLocation(n);return t&&e.getSymbolTarget(t,p)}}(h,p,y);if(b.valid){return{renameFilename:void 0,renameLocation:void 0,edits:e.textChanges.ChangeTracker.with(n,(function(n){return function(n,t,r,a,i,o){var s=e.map(function(n,t,r){var a,i=t.getTypeChecker(),o=f(n.parameters),s=e.map(o,_),c=e.factory.createObjectBindingPattern(s),l=y(o);e.every(o,b)&&(a=e.factory.createObjectLiteralExpression());var u=e.factory.createParameterDeclaration(void 0,void 0,void 0,c,void 0,l,a);if(m(n.parameters)){var d=n.parameters[0],p=e.factory.createParameterDeclaration(void 0,void 0,void 0,d.name,void 0,d.type);return e.suppressLeadingAndTrailingTrivia(p.name),e.copyComments(d.name,p.name),d.type&&(e.suppressLeadingAndTrailingTrivia(p.type),e.copyComments(d.type,p.type)),e.factory.createNodeArray([p,u])}return e.factory.createNodeArray([u]);function _(n){var t=e.factory.createBindingElement(void 0,void 0,g(n),e.isRestParameter(n)&&b(n)?e.factory.createArrayLiteralExpression():n.initializer);return e.suppressLeadingAndTrailingTrivia(t),n.initializer&&t.initializer&&e.copyComments(n.initializer,t.initializer),t}function y(n){var t=e.map(n,v);return e.addEmitFlags(e.factory.createTypeLiteralNode(t),1)}function v(n){var t=n.type;t||!n.initializer&&!e.isRestParameter(n)||(t=h(n));var r=e.factory.createPropertySignature(void 0,g(n),b(n)?e.factory.createToken(57):n.questionToken,t);return e.suppressLeadingAndTrailingTrivia(r),e.copyComments(n.name,r.name),n.type&&r.type&&e.copyComments(n.type,r.type),r}function h(n){var a=i.getTypeAtLocation(n);return e.getTypeNodeIfAccessible(a,n,t,r)}function b(n){if(e.isRestParameter(n)){var t=i.getTypeAtLocation(n);return!i.isTupleType(t)}return i.isOptionalParameter(n)}}(i,t,r),(function(n){return e.getSynthesizedDeepClone(n)}));a.replaceNodeRangeWithNodes(n,e.first(i.parameters),e.last(i.parameters),s,{joiner:", ",indentation:0,leadingTriviaOption:e.textChanges.LeadingTriviaOption.IncludeAll,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Include});for(var c=e.sortAndDeduplicate(o.functionCalls,(function(n,t){return e.compareValues(n.pos,t.pos)})),l=0,u=c;l<u.length;l++){var d=u[l];if(d.arguments&&d.arguments.length){var p=e.getSynthesizedDeepClone(_(i,d.arguments),!0);a.replaceNodeRange(e.getSourceFileOfNode(d),e.first(d.arguments),e.last(d.arguments),p,{leadingTriviaOption:e.textChanges.LeadingTriviaOption.IncludeAll,trailingTriviaOption:e.textChanges.TrailingTriviaOption.Include})}}}(u,p,v,n,h,b)}))}}return{edits:[]}},getAvailableActions:function(n){var t=n.file,a=n.startPosition;if(e.isSourceFileJS(t))return e.emptyArray;if(!l(t,a,n.program.getTypeChecker()))return e.emptyArray;var i=e.getLocaleSpecificMessage(e.Diagnostics.Convert_parameters_to_destructured_object);return[{name:r,description:i,actions:[{name:r,description:i}]}]}})}(n.convertParamsToDestructuredObject||(n.convertParamsToDestructuredObject={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Convert to template string",a=e.getLocaleSpecificMessage(e.Diagnostics.Convert_to_template_string);function i(n,t){var r=e.getTokenAtPosition(n,t),a=s(r);return!c(a)&&e.isParenthesizedExpression(a.parent)&&e.isBinaryExpression(a.parent.parent)?a.parent.parent:r}function o(n,t){var r=s(t),a=n.file,i=function(n,t){var r=n.nodes,a=n.operators,i=function(n,t){return function(r,a){r<n.length&&e.copyTrailingComments(n[r],a,t,3,!1)}}(a,t),o=function(n,t,r){return function(a,i){for(;a.length>0;){var o=a.shift();e.copyTrailingComments(n[o],i,t,3,!1),r(o,i)}}}(r,t,i),s=u(0,r),c=s[0],l=s[1],p=s[2];if(c===r.length){var m=e.factory.createNoSubstitutionTemplateLiteral(l);return o(p,m),m}var f=[],_=e.factory.createTemplateHead(l);o(p,_);for(var g=c;g<r.length;g++){var y=d(r[g]);i(g,y);var v=u(g+1,r),h=v[0],b=v[1],E=v[2],T=(g=h-1)===r.length-1?e.factory.createTemplateTail(b):e.factory.createTemplateMiddle(b);o(E,T),f.push(e.factory.createTemplateSpan(y,T))}return e.factory.createTemplateExpression(_,f)}(l(r),a),o=e.getTrailingCommentRanges(a.text,r.end);if(o){var c=o[o.length-1],p={pos:o[0].pos,end:c.end};return e.textChanges.ChangeTracker.with(n,(function(e){e.deleteRange(a,p),e.replaceNode(a,r,i)}))}return e.textChanges.ChangeTracker.with(n,(function(e){return e.replaceNode(a,r,i)}))}function s(n){return e.findAncestor(n.parent,(function(n){switch(n.kind){case 201:case 202:return!1;case 216:return!(e.isBinaryExpression(n.parent)&&(t=n.parent,62!==t.operatorToken.kind));default:return"quit"}var t}))||n}function c(e){var n=l(e),t=n.containsString,r=n.areOperatorsValid;return t&&r}function l(n){if(e.isBinaryExpression(n)){var t=l(n.left),r=t.nodes,a=t.operators,i=t.containsString,o=t.areOperatorsValid;if(!i&&!e.isStringLiteral(n.right))return{nodes:[n],operators:[],containsString:!1,areOperatorsValid:!0};var s=39===n.operatorToken.kind,c=o&&s;return r.push(n.right),a.push(n.operatorToken),{nodes:r,operators:a,containsString:!0,areOperatorsValid:c}}return{nodes:[n],operators:[],containsString:e.isStringLiteral(n),areOperatorsValid:!0}}n.registerRefactor(r,{getEditsForAction:function(n,t){var r=n.file,s=n.startPosition,c=i(r,s);switch(t){case a:return{edits:o(n,c)};default:return e.Debug.fail("invalid action")}},getAvailableActions:function(n){var t=n.file,o=n.startPosition,l=s(i(t,o)),u={name:r,description:a,actions:[]};if(e.isBinaryExpression(l)&&c(l))return u.actions.push({name:r,description:a}),[u];return e.emptyArray}});function u(n,t){for(var r="",a=[];n<t.length&&e.isStringLiteral(t[n]);){r+=t[n].text,a.push(n),n++}return[n,r,a]}function d(n){return e.isParenthesizedExpression(n)&&(!function(n){var t=n.getSourceFile();e.copyTrailingComments(n,n.expression,t,3,!1),e.copyTrailingAsLeadingComments(n.expression,n.expression,t,3,!1)}(n),n=n.expression),n}}(n.convertStringOrTemplateLiteral||(n.convertStringOrTemplateLiteral={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){!function(n){!function(t){var r="Convert arrow function or function expression",a=e.getLocaleSpecificMessage(e.Diagnostics.Convert_arrow_function_or_function_expression),i="Convert to anonymous function",o="Convert to named function",s="Convert to arrow function",c=e.getLocaleSpecificMessage(e.Diagnostics.Convert_to_anonymous_function),l=e.getLocaleSpecificMessage(e.Diagnostics.Convert_to_named_function),u=e.getLocaleSpecificMessage(e.Diagnostics.Convert_to_arrow_function);function d(n){var t=!1;return n.forEachChild((function n(r){e.isThis(r)?t=!0:e.isClassLike(r)||e.isFunctionDeclaration(r)||e.isFunctionExpression(r)||e.forEachChild(r,n)})),t}function p(n,t,r){var a=e.getTokenAtPosition(n,t),i=function(n){if(!function(n){return e.isVariableDeclaration(n)||e.isVariableDeclarationList(n)&&1===n.declarations.length}(n))return;var t=(e.isVariableDeclaration(n)?n:n.declarations[0]).initializer;return t&&e.isArrowFunction(t)?t:void 0}(a.parent);if(i&&!d(i.body))return{selectedVariableDeclaration:!0,func:i};var o=e.getContainingFunction(a),s=r.getTypeChecker();if(o&&(e.isFunctionExpression(o)||e.isArrowFunction(o))&&!e.rangeContainsRange(o.body,a)&&!d(o.body)){if(e.isFunctionExpression(o)&&o.name&&e.FindAllReferences.Core.isSymbolReferencedInFile(o.name,s,n))return;return{selectedVariableDeclaration:!1,func:o}}}function m(n){return e.isExpression(n)?e.factory.createBlock([e.factory.createReturnStatement(n)],!0):n}n.registerRefactor(r,{getEditsForAction:function(n,t){var r=n.file,a=n.startPosition,c=n.program,l=p(r,a,c);if(!l)return;var u=l.func,d=[];switch(t){case i:d.push.apply(d,function(n,t){var r=n.file,a=m(t.body),i=e.factory.createFunctionExpression(t.modifiers,t.asteriskToken,void 0,t.typeParameters,t.parameters,t.type,a);return e.textChanges.ChangeTracker.with(n,(function(e){return e.replaceNode(r,t,i)}))}(n,u));break;case o:var f=function(n){var t=n.parent;if(!e.isVariableDeclaration(t)||!e.isVariableDeclarationInVariableStatement(t))return;var r=t.parent,a=r.parent;return e.isVariableDeclarationList(r)&&e.isVariableStatement(a)&&e.isIdentifier(t.name)?{variableDeclaration:t,variableDeclarationList:r,statement:a,name:t.name}:void 0}(u);if(!f)return;d.push.apply(d,function(n,t,r){var a=n.file,i=m(t.body),o=r.variableDeclaration,s=r.variableDeclarationList,c=r.statement,l=r.name;e.suppressLeadingTrivia(c);var u=e.factory.createFunctionDeclaration(t.decorators,c.modifiers,t.asteriskToken,l,t.typeParameters,t.parameters,t.type,i);return 1===s.declarations.length?e.textChanges.ChangeTracker.with(n,(function(e){return e.replaceNode(a,c,u)})):e.textChanges.ChangeTracker.with(n,(function(e){e.delete(a,o),e.insertNodeAfter(a,c,u)}))}(n,u,f));break;case s:if(!e.isFunctionExpression(u))return;d.push.apply(d,function(n,t){var r,a=n.file,i=t.body.statements[0];!function(n,t){return 1===n.statements.length&&e.isReturnStatement(t)&&!!t.expression}(t.body,i)?r=t.body:(r=i.expression,e.suppressLeadingAndTrailingTrivia(r),e.copyComments(i,r));var o=e.factory.createArrowFunction(t.modifiers,t.typeParameters,t.parameters,t.type,e.factory.createToken(38),r);return e.textChanges.ChangeTracker.with(n,(function(e){return e.replaceNode(a,t,o)}))}(n,u));break;default:return e.Debug.fail("invalid action")}return{renameFilename:void 0,renameLocation:void 0,edits:d}},getAvailableActions:function(n){var t=n.file,d=n.startPosition,m=n.program,f=p(t,d,m);if(!f)return e.emptyArray;var _=f.selectedVariableDeclaration,g=f.func,y=[];(_||e.isArrowFunction(g)&&e.isVariableDeclaration(g.parent))&&y.push({name:o,description:l});!_&&e.isArrowFunction(g)&&y.push({name:i,description:c});e.isFunctionExpression(g)&&y.push({name:s,description:u});return[{name:r,description:a,actions:y}]}})}(n.convertArrowFunctionOrFunctionExpression||(n.convertArrowFunctionOrFunctionExpression={}))}(e.refactor||(e.refactor={}))}(ts||(ts={})),function(e){function n(n,r,a,i){var l=e.isNodeKind(n)?new t(n,r,a):78===n?new s(78,r,a):79===n?new c(79,r,a):new o(n,r,a);return l.parent=i,l.flags=25358336&i.flags,l}e.servicesVersion="0.8";var t=function(){function t(e,n,t){this.pos=n,this.end=t,this.flags=0,this.modifierFlagsCache=0,this.transformFlags=0,this.parent=void 0,this.kind=e}return t.prototype.assertHasRealPosition=function(n){e.Debug.assert(!e.positionIsSynthesized(this.pos)&&!e.positionIsSynthesized(this.end),n||"Node must have a real position for this operation")},t.prototype.getSourceFile=function(){return e.getSourceFileOfNode(this)},t.prototype.getStart=function(n,t){return this.assertHasRealPosition(),e.getTokenPosOfNode(this,n,t)},t.prototype.getFullStart=function(){return this.assertHasRealPosition(),this.pos},t.prototype.getEnd=function(){return this.assertHasRealPosition(),this.end},t.prototype.getWidth=function(e){return this.assertHasRealPosition(),this.getEnd()-this.getStart(e)},t.prototype.getFullWidth=function(){return this.assertHasRealPosition(),this.end-this.pos},t.prototype.getLeadingTriviaWidth=function(e){return this.assertHasRealPosition(),this.getStart(e)-this.pos},t.prototype.getFullText=function(e){return this.assertHasRealPosition(),(e||this.getSourceFile()).text.substring(this.pos,this.end)},t.prototype.getText=function(e){return this.assertHasRealPosition(),e||(e=this.getSourceFile()),e.text.substring(this.getStart(e),this.getEnd())},t.prototype.getChildCount=function(e){return this.getChildren(e).length},t.prototype.getChildAt=function(e,n){return this.getChildren(n)[e]},t.prototype.getChildren=function(t){return this.assertHasRealPosition("Node without a real position cannot be scanned and thus has no token nodes - use forEachChild and collect the result if that's fine"),this._children||(this._children=function(t,a){if(!e.isNodeKind(t.kind))return e.emptyArray;var i=[];if(e.isJSDocCommentContainingNode(t))return t.forEachChild((function(e){i.push(e)})),i;e.scanner.setText((a||t.getSourceFile()).text);var o=t.pos,s=function(e){r(i,o,e.pos,t),i.push(e),o=e.end},c=function(e){r(i,o,e.pos,t),i.push(function(e,t){var a=n(334,e.pos,e.end,t);a._children=[];for(var i=e.pos,o=0,s=e;o<s.length;o++){var c=s[o];r(a._children,i,c.pos,t),a._children.push(c),i=c.end}return r(a._children,i,e.end,t),a}(e,t)),o=e.end};return e.forEach(t.jsDoc,s),o=t.pos,t.forEachChild(s,c),r(i,o,t.end,t),e.scanner.setText(void 0),i}(this,t))},t.prototype.getFirstToken=function(n){this.assertHasRealPosition();var t=this.getChildren(n);if(t.length){var r=e.find(t,(function(e){return e.kind<301||e.kind>333}));return r.kind<157?r:r.getFirstToken(n)}},t.prototype.getLastToken=function(n){this.assertHasRealPosition();var t=this.getChildren(n),r=e.lastOrUndefined(t);if(r)return r.kind<157?r:r.getLastToken(n)},t.prototype.forEachChild=function(n,t){return e.forEachChild(this,n,t)},t}();function r(t,r,a,i){for(e.scanner.setTextPos(r);r<a;){var o=e.scanner.scan(),s=e.scanner.getTextPos();if(s<=a&&(78===o&&e.Debug.fail("Did not expect "+e.Debug.formatSyntaxKind(i.kind)+" to have an Identifier in its trivia"),t.push(n(o,r,s,i))),r=s,1===o)break}}var a=function(){function n(e,n){this.pos=e,this.end=n,this.flags=0,this.modifierFlagsCache=0,this.transformFlags=0,this.parent=void 0}return n.prototype.getSourceFile=function(){return e.getSourceFileOfNode(this)},n.prototype.getStart=function(n,t){return e.getTokenPosOfNode(this,n,t)},n.prototype.getFullStart=function(){return this.pos},n.prototype.getEnd=function(){return this.end},n.prototype.getWidth=function(e){return this.getEnd()-this.getStart(e)},n.prototype.getFullWidth=function(){return this.end-this.pos},n.prototype.getLeadingTriviaWidth=function(e){return this.getStart(e)-this.pos},n.prototype.getFullText=function(e){return(e||this.getSourceFile()).text.substring(this.pos,this.end)},n.prototype.getText=function(e){return e||(e=this.getSourceFile()),e.text.substring(this.getStart(e),this.getEnd())},n.prototype.getChildCount=function(){return 0},n.prototype.getChildAt=function(){},n.prototype.getChildren=function(){return 1===this.kind&&this.jsDoc||e.emptyArray},n.prototype.getFirstToken=function(){},n.prototype.getLastToken=function(){},n.prototype.forEachChild=function(){},n}(),i=function(){function n(e,n){this.flags=e,this.escapedName=n}return n.prototype.getFlags=function(){return this.flags},Object.defineProperty(n.prototype,"name",{get:function(){return e.symbolName(this)},enumerable:!1,configurable:!0}),n.prototype.getEscapedName=function(){return this.escapedName},n.prototype.getName=function(){return this.name},n.prototype.getDeclarations=function(){return this.declarations},n.prototype.getDocumentationComment=function(n){if(!this.documentationComment)if(this.documentationComment=e.emptyArray,!this.declarations&&this.target&&this.target.tupleLabelDeclaration){var t=this.target.tupleLabelDeclaration;this.documentationComment=p([t],n)}else this.documentationComment=p(this.declarations,n);return this.documentationComment},n.prototype.getContextualDocumentationComment=function(n,t){switch(null==n?void 0:n.kind){case 167:return this.contextualGetAccessorDocumentationComment||(this.contextualGetAccessorDocumentationComment=e.emptyArray,this.contextualGetAccessorDocumentationComment=p(e.filter(this.declarations,e.isGetAccessor),t)),this.contextualGetAccessorDocumentationComment;case 168:return this.contextualSetAccessorDocumentationComment||(this.contextualSetAccessorDocumentationComment=e.emptyArray,this.contextualSetAccessorDocumentationComment=p(e.filter(this.declarations,e.isSetAccessor),t)),this.contextualSetAccessorDocumentationComment;default:return this.getDocumentationComment(t)}},n.prototype.getJsDocTags=function(){return void 0===this.tags&&(this.tags=e.JsDoc.getJsDocTagsFromDeclarations(this.declarations)),this.tags},n}(),o=function(e){function n(n,t,r){var a=e.call(this,t,r)||this;return a.kind=n,a}return __extends(n,e),n}(a),s=function(n){function t(e,t,r){var a=n.call(this,t,r)||this;return a.kind=78,a}return __extends(t,n),Object.defineProperty(t.prototype,"text",{get:function(){return e.idText(this)},enumerable:!1,configurable:!0}),t}(a);s.prototype.kind=78;var c=function(n){function t(e,t,r){return n.call(this,t,r)||this}return __extends(t,n),Object.defineProperty(t.prototype,"text",{get:function(){return e.idText(this)},enumerable:!1,configurable:!0}),t}(a);c.prototype.kind=79;var l=function(){function n(e,n){this.checker=e,this.flags=n}return n.prototype.getFlags=function(){return this.flags},n.prototype.getSymbol=function(){return this.symbol},n.prototype.getProperties=function(){return this.checker.getPropertiesOfType(this)},n.prototype.getProperty=function(e){return this.checker.getPropertyOfType(this,e)},n.prototype.getApparentProperties=function(){return this.checker.getAugmentedPropertiesOfType(this)},n.prototype.getCallSignatures=function(){return this.checker.getSignaturesOfType(this,0)},n.prototype.getConstructSignatures=function(){return this.checker.getSignaturesOfType(this,1)},n.prototype.getStringIndexType=function(){return this.checker.getIndexTypeOfType(this,0)},n.prototype.getNumberIndexType=function(){return this.checker.getIndexTypeOfType(this,1)},n.prototype.getBaseTypes=function(){return this.isClassOrInterface()?this.checker.getBaseTypes(this):void 0},n.prototype.isNullableType=function(){return this.checker.isNullableType(this)},n.prototype.getNonNullableType=function(){return this.checker.getNonNullableType(this)},n.prototype.getNonOptionalType=function(){return this.checker.getNonOptionalType(this)},n.prototype.getConstraint=function(){return this.checker.getBaseConstraintOfType(this)},n.prototype.getDefault=function(){return this.checker.getDefaultFromTypeParameter(this)},n.prototype.isUnion=function(){return!!(1048576&this.flags)},n.prototype.isIntersection=function(){return!!(2097152&this.flags)},n.prototype.isUnionOrIntersection=function(){return!!(3145728&this.flags)},n.prototype.isLiteral=function(){return!!(384&this.flags)},n.prototype.isStringLiteral=function(){return!!(128&this.flags)},n.prototype.isNumberLiteral=function(){return!!(256&this.flags)},n.prototype.isTypeParameter=function(){return!!(262144&this.flags)},n.prototype.isClassOrInterface=function(){return!!(3&e.getObjectFlags(this))},n.prototype.isClass=function(){return!!(1&e.getObjectFlags(this))},Object.defineProperty(n.prototype,"typeArguments",{get:function(){if(4&e.getObjectFlags(this))return this.checker.getTypeArguments(this)},enumerable:!1,configurable:!0}),n}(),u=function(){function n(e,n){this.checker=e,this.flags=n}return n.prototype.getDeclaration=function(){return this.declaration},n.prototype.getTypeParameters=function(){return this.typeParameters},n.prototype.getParameters=function(){return this.parameters},n.prototype.getReturnType=function(){return this.checker.getReturnTypeOfSignature(this)},n.prototype.getDocumentationComment=function(){return this.documentationComment||(this.documentationComment=p(e.singleElementArray(this.declaration),this.checker))},n.prototype.getJsDocTags=function(){return void 0===this.jsDocTags&&(this.jsDocTags=this.declaration?e.JsDoc.getJsDocTagsFromDeclarations([this.declaration]):[]),this.jsDocTags},n}();function d(n){return e.getJSDocTags(n).some((function(e){return"inheritDoc"===e.tagName.text}))}function p(n,t){if(!n)return e.emptyArray;var r=e.JsDoc.getJsDocCommentsFromDeclarations(n);return(0===r.length||n.some(d))&&e.forEachUnique(n,(function(n){var a=function(n,t,r){return e.firstDefined(n.parent?e.getAllSuperTypeNodes(n.parent):e.emptyArray,(function(e){var n=r.getTypeAtLocation(e),a=n&&r.getPropertyOfType(n,t),i=a&&a.getDocumentationComment(r);return i&&i.length?i:void 0}))}(n,n.symbol.name,t);a&&(r=0===r.length?a.slice():a.concat(e.lineBreakPart(),r))})),r}var m=function(n){function t(e,t,r){var a=n.call(this,e,t,r)||this;return a.kind=297,a}return __extends(t,n),t.prototype.update=function(n,t){return e.updateSourceFile(this,n,t)},t.prototype.getLineAndCharacterOfPosition=function(n){return e.getLineAndCharacterOfPosition(this,n)},t.prototype.getLineStarts=function(){return e.getLineStarts(this)},t.prototype.getPositionOfLineAndCharacter=function(n,t,r){return e.computePositionOfLineAndCharacter(e.getLineStarts(this),n,t,this.text,r)},t.prototype.getLineEndOfPosition=function(e){var n,t=this.getLineAndCharacterOfPosition(e).line,r=this.getLineStarts();t+1>=r.length&&(n=this.getEnd()),n||(n=r[t+1]-1);var a=this.getFullText();return"\n"===a[n]&&"\r"===a[n-1]?n-1:n},t.prototype.getNamedDeclarations=function(){return this.namedDeclarations||(this.namedDeclarations=this.computeNamedDeclarations()),this.namedDeclarations},t.prototype.computeNamedDeclarations=function(){var n=e.createMultiMap();return this.forEachChild((function a(i){switch(i.kind){case 251:case 208:case 165:case 164:var o=i,s=r(o);if(s){var c=function(e){var t=n.get(e);t||n.set(e,t=[]);return t}(s),l=e.lastOrUndefined(c);l&&o.parent===l.parent&&o.symbol===l.symbol?o.body&&!l.body&&(c[c.length-1]=o):c.push(o)}e.forEachChild(i,a);break;case 252:case 221:case 253:case 254:case 255:case 256:case 260:case 270:case 265:case 262:case 263:case 167:case 168:case 177:t(i),e.forEachChild(i,a);break;case 160:if(!e.hasSyntacticModifier(i,92))break;case 249:case 198:var u=i;if(e.isBindingPattern(u.name)){e.forEachChild(u.name,a);break}u.initializer&&a(u.initializer);case 291:case 163:case 162:t(i);break;case 267:var d=i;d.exportClause&&(e.isNamedExports(d.exportClause)?e.forEach(d.exportClause.elements,a):a(d.exportClause.name));break;case 261:var p=i.importClause;p&&(p.name&&t(p.name),p.namedBindings&&(263===p.namedBindings.kind?t(p.namedBindings):e.forEach(p.namedBindings.elements,a)));break;case 216:0!==e.getAssignmentDeclarationKind(i)&&t(i);default:e.forEachChild(i,a)}})),n;function t(e){var t=r(e);t&&n.add(t,e)}function r(n){var t=e.getNonAssignedNameOfDeclaration(n);return t&&(e.isComputedPropertyName(t)&&e.isPropertyAccessExpression(t.expression)?t.expression.name.text:e.isPropertyName(t)?e.getNameFromPropertyName(t):void 0)}},t}(t),f=function(){function n(e,n,t){this.fileName=e,this.text=n,this.skipTrivia=t}return n.prototype.getLineAndCharacterOfPosition=function(n){return e.getLineAndCharacterOfPosition(this,n)},n}();function _(n){var t=!0;for(var r in n)if(e.hasProperty(n,r)&&!g(r)){t=!1;break}if(t)return n;var a={};for(var r in n){if(e.hasProperty(n,r))a[g(r)?r:r.charAt(0).toLowerCase()+r.substr(1)]=n[r]}return a}function g(e){return!e.length||e.charAt(0)===e.charAt(0).toLowerCase()}function y(){return{target:1,jsx:1}}e.toEditorSettings=_,e.displayPartsToString=function(n){return n?e.map(n,(function(e){return e.text})).join(""):""},e.getDefaultCompilerOptions=y,e.getSupportedCodeFixes=function(){return e.codefix.getSupportedErrorCodes()};var v=function(){function n(n,t){this.host=n,this.currentDirectory=n.getCurrentDirectory(),this.fileNameToEntry=new e.Map;for(var r=0,a=n.getScriptFileNames();r<a.length;r++){var i=a[r];this.createEntry(i,e.toPath(i,this.currentDirectory,t))}this._compilationSettings=n.getCompilationSettings()||{target:1,jsx:1}}return n.prototype.compilationSettings=function(){return this._compilationSettings},n.prototype.getProjectReferences=function(){return this.host.getProjectReferences&&this.host.getProjectReferences()},n.prototype.createEntry=function(n,t){var r,a=this.host.getScriptSnapshot(n);return r=a?{hostFileName:n,version:this.host.getScriptVersion(n),scriptSnapshot:a,scriptKind:e.getScriptKind(n,this.host)}:n,this.fileNameToEntry.set(t,r),r},n.prototype.getEntryByPath=function(e){return this.fileNameToEntry.get(e)},n.prototype.getHostFileInformation=function(n){var t=this.fileNameToEntry.get(n);return e.isString(t)?void 0:t},n.prototype.getOrCreateEntryByPath=function(n,t){var r=this.getEntryByPath(t)||this.createEntry(n,t);return e.isString(r)?void 0:r},n.prototype.getRootFileNames=function(){var n=[];return this.fileNameToEntry.forEach((function(t){e.isString(t)?n.push(t):n.push(t.hostFileName)})),n},n.prototype.getScriptSnapshot=function(e){var n=this.getHostFileInformation(e);return n&&n.scriptSnapshot},n}(),h=function(){function n(e){this.host=e}return n.prototype.getCurrentSourceFile=function(n){var t=this.host.getScriptSnapshot(n);if(!t)throw new Error("Could not find file: '"+n+"'.");var r,a=e.getScriptKind(n,this.host),i=this.host.getScriptVersion(n);if(this.currentFileName!==n)r=E(n,t,99,i,!0,a);else if(this.currentFileVersion!==i){var o=t.getChangeRange(this.currentFileScriptSnapshot);r=T(this.currentSourceFile,t,i,o)}return r&&(this.currentFileVersion=i,this.currentFileName=n,this.currentFileScriptSnapshot=t,this.currentSourceFile=r),this.currentSourceFile},n}();function b(e,n,t){e.version=t,e.scriptSnapshot=n}function E(n,t,r,a,i,o){var s=e.createSourceFile(n,e.getSnapshotText(t),r,i,o);return b(s,t,a),s}function T(n,t,r,a,i){if(a&&r!==n.version){var o=void 0,s=0!==a.span.start?n.text.substr(0,a.span.start):"",c=e.textSpanEnd(a.span)!==n.text.length?n.text.substr(e.textSpanEnd(a.span)):"";if(0===a.newLength)o=s&&c?s+c:s||c;else{var l=t.getText(a.span.start,a.span.start+a.newLength);o=s&&c?s+l+c:s?s+l:l+c}var u=e.updateSourceFile(n,o,a,i);return b(u,t,r),u.nameTable=void 0,n!==u&&n.scriptSnapshot&&(n.scriptSnapshot.dispose&&n.scriptSnapshot.dispose(),n.scriptSnapshot=void 0),u}return E(n.fileName,t,n.languageVersion,r,!0,n.scriptKind)}e.createLanguageServiceSourceFile=E,e.updateLanguageServiceSourceFile=T;var S=function(){function n(e){this.cancellationToken=e}return n.prototype.isCancellationRequested=function(){return!!this.cancellationToken&&this.cancellationToken.isCancellationRequested()},n.prototype.throwIfCancellationRequested=function(){if(this.isCancellationRequested())throw new e.OperationCanceledException},n}(),x=function(){function n(e,n){void 0===n&&(n=20),this.hostCancellationToken=e,this.throttleWaitMilliseconds=n,this.lastCancellationCheckTime=0}return n.prototype.isCancellationRequested=function(){var n=e.timestamp();return Math.abs(n-this.lastCancellationCheckTime)>=this.throttleWaitMilliseconds&&(this.lastCancellationCheckTime=n,this.hostCancellationToken.isCancellationRequested())},n.prototype.throwIfCancellationRequested=function(){if(this.isCancellationRequested())throw new e.OperationCanceledException},n}();e.ThrottledCancellationToken=x;var L=["getSyntacticDiagnostics","getSemanticDiagnostics","getSuggestionDiagnostics","getCompilerOptionsDiagnostics","getSemanticClassifications","getEncodedSemanticClassifications","getCodeFixesAtPosition","getCombinedCodeFix","applyCodeActionCommand","organizeImports","getEditsForFileRename","getEmitOutput","getApplicableRefactors","getEditsForRefactor","prepareCallHierarchy","provideCallHierarchyIncomingCalls","provideCallHierarchyOutgoingCalls"],A=__spreadArrays(L,["getCompletionsAtPosition","getCompletionEntryDetails","getCompletionEntrySymbol","getSignatureHelpItems","getQuickInfoAtPosition","getDefinitionAtPosition","getDefinitionAndBoundSpan","getImplementationAtPosition","getTypeDefinitionAtPosition","getReferencesAtPosition","findReferences","getOccurrencesAtPosition","getDocumentHighlights","getNavigateToItems","getRenameInfo","findRenameLocations","getApplicableRefactors"]);function D(n){var t=function(n){switch(n.kind){case 10:case 14:case 8:if(158===n.parent.kind)return e.isObjectLiteralElement(n.parent.parent)?n.parent.parent:void 0;case 78:return!e.isObjectLiteralElement(n.parent)||200!==n.parent.parent.kind&&281!==n.parent.parent.kind||n.parent.name!==n?void 0:n.parent}return}(n);return t&&(e.isObjectLiteralExpression(t.parent)||e.isJsxAttributes(t.parent))?t:void 0}function C(n,t,r,a){var i=e.getNameFromPropertyName(n.name);if(!i)return e.emptyArray;if(!r.isUnion())return(o=r.getProperty(i))?[o]:e.emptyArray;var o,s=e.mapDefined(r.types,(function(r){return e.isObjectLiteralExpression(n.parent)&&t.isTypeInvalidDueToUnionDiscriminant(r,n.parent)?void 0:r.getProperty(i)}));if(a&&(0===s.length||s.length===r.types.length)&&(o=r.getProperty(i)))return[o];return 0===s.length?e.mapDefined(r.types,(function(e){return e.getProperty(i)})):s}e.createLanguageService=function(n,t,r){var a,i;void 0===t&&(t=e.createDocumentRegistry(n.useCaseSensitiveFileNames&&n.useCaseSensitiveFileNames(),n.getCurrentDirectory())),i=void 0===r?e.LanguageServiceMode.Semantic:"boolean"==typeof r?r?e.LanguageServiceMode.Syntactic:e.LanguageServiceMode.Semantic:r;var o,s,c=new h(n),l=0,u=new S(n.getCancellationToken&&n.getCancellationToken()),d=n.getCurrentDirectory();function p(e){n.log&&n.log(e)}!e.localizedDiagnosticMessages&&n.getLocalizedDiagnosticMessages&&e.setLocalizedDiagnosticMessages(n.getLocalizedDiagnosticMessages());var m=e.hostUsesCaseSensitiveFileNames(n),f=e.createGetCanonicalFileName(m),g=e.getSourceMapper({useCaseSensitiveFileNames:function(){return m},getCurrentDirectory:function(){return d},getProgram:E,fileExists:e.maybeBind(n,n.fileExists),readFile:e.maybeBind(n,n.readFile),getDocumentPositionMapper:e.maybeBind(n,n.getDocumentPositionMapper),getSourceFileLike:e.maybeBind(n,n.getSourceFileLike),log:p});function y(e){var n=o.getSourceFile(e);if(!n){var t=new Error("Could not find source file: '"+e+"'.");throw t.ProgramFiles=o.getSourceFiles().map((function(e){return e.fileName})),t}return n}function b(){var r,a;if(e.Debug.assert(i!==e.LanguageServiceMode.Syntactic),n.getProjectVersion){var c=n.getProjectVersion();if(c){if(s===c&&!(null===(r=n.hasChangedAutomaticTypeDirectiveNames)||void 0===r?void 0:r.call(n)))return;s=c}}var _=n.getTypeRootsVersion?n.getTypeRootsVersion():0;l!==_&&(p("TypeRoots version has changed; provide new program"),o=void 0,l=_);var y=new v(n,f),h=y.getRootFileNames(),b=n.hasInvalidatedResolution||e.returnFalse,E=e.maybeBind(n,n.hasChangedAutomaticTypeDirectiveNames),T=y.getProjectReferences();if(!e.isProgramUptoDate(o,h,y.compilationSettings(),(function(e,t){return n.getScriptVersion(t)}),D,b,E,T)){var S=y.compilationSettings(),x={getSourceFile:function(n,t,r,a){return C(n,e.toPath(n,d,f),t,r,a)},getSourceFileByPath:C,getCancellationToken:function(){return u},getCanonicalFileName:f,useCaseSensitiveFileNames:function(){return m},getNewLine:function(){return e.getNewLineCharacter(S,(function(){return e.getNewLineOrDefaultFromHost(n)}))},getDefaultLibFileName:function(e){return n.getDefaultLibFileName(e)},writeFile:e.noop,getCurrentDirectory:function(){return d},fileExists:D,readFile:function(t){var r=e.toPath(t,d,f),a=y&&y.getEntryByPath(r);if(a)return e.isString(a)?void 0:e.getSnapshotText(a.scriptSnapshot);return n.readFile&&n.readFile(t)},getSymlinkCache:e.maybeBind(n,n.getSymlinkCache),realpath:e.maybeBind(n,n.realpath),directoryExists:function(t){return e.directoryProbablyExists(t,n)},getDirectories:function(e){return n.getDirectories?n.getDirectories(e):[]},readDirectory:function(t,r,a,i,o){return e.Debug.checkDefined(n.readDirectory,"'LanguageServiceHost.readDirectory' must be implemented to correctly process 'projectReferences'"),n.readDirectory(t,r,a,i,o)},onReleaseOldSourceFile:function(e,n){var r=t.getKeyForCompilationSettings(n);t.releaseDocumentWithKey(e.resolvedPath,r)},hasInvalidatedResolution:b,hasChangedAutomaticTypeDirectiveNames:E,trace:e.maybeBind(n,n.trace),resolveModuleNames:e.maybeBind(n,n.resolveModuleNames),resolveTypeReferenceDirectives:e.maybeBind(n,n.resolveTypeReferenceDirectives),useSourceOfProjectReferenceRedirect:e.maybeBind(n,n.useSourceOfProjectReferenceRedirect)};null===(a=n.setCompilerHost)||void 0===a||a.call(n,x);var L=t.getKeyForCompilationSettings(S),A={rootNames:h,options:S,host:x,oldProgram:o,projectReferences:T};return o=e.createProgram(A),y=void 0,g.clearCache(),void o.getTypeChecker()}function D(t){var r=e.toPath(t,d,f),a=y&&y.getEntryByPath(r);return a?!e.isString(a):!!n.fileExists&&n.fileExists(t)}function C(n,r,a,i,s){e.Debug.assert(void 0!==y,"getOrCreateSourceFileByPath called after typical CompilerHost lifetime, check the callstack something with a reference to an old host.");var c=y&&y.getOrCreateEntryByPath(n,r);if(c){if(!s){var l=o&&o.getSourceFileByPath(r);if(l)return e.Debug.assertEqual(c.scriptKind,l.scriptKind,"Registered script kind should match new script kind."),t.updateDocumentWithKey(n,r,S,L,c.scriptSnapshot,c.version,c.scriptKind)}return t.acquireDocumentWithKey(n,r,S,L,c.scriptSnapshot,c.version,c.scriptKind)}}}function E(){if(i!==e.LanguageServiceMode.Syntactic)return b(),o;e.Debug.assert(void 0===o)}function T(n,t,r){var a=e.normalizePath(n);e.Debug.assert(r.some((function(n){return e.normalizePath(n)===a}))),b();var i=e.mapDefined(r,(function(e){return o.getSourceFile(e)})),s=y(n);return e.DocumentHighlights.getDocumentHighlights(o,u,s,t,i)}function x(n,t,r,a){b();var i=r&&2===r.use?o.getSourceFiles().filter((function(e){return!o.isSourceFileDefaultLibrary(e)})):o.getSourceFiles();return e.FindAllReferences.findReferenceOrRenameEntries(o,u,i,n,t,r,a)}function I(t){var r=e.getScriptKind(t,n);return 3===r||4===r}var N=new e.Map(e.getEntries(((a={})[18]=19,a[20]=21,a[22]=23,a[31]=29,a)));function R(t){var r;return e.Debug.assertEqual(t.type,"install package"),n.installPackage?n.installPackage({fileName:(r=t.file,e.toPath(r,d,f)),packageName:t.packageName}):Promise.reject("Host does not implement `installPackage`")}function k(e,n){return{lineStarts:e.getLineStarts(),firstLine:e.getLineAndCharacterOfPosition(n.pos).line,lastLine:e.getLineAndCharacterOfPosition(n.end).line}}function O(n,t,r){for(var a=c.getCurrentSourceFile(n),i=[],o=k(a,t),s=o.lineStarts,l=o.firstLine,u=o.lastLine,d=r||!1,p=Number.MAX_VALUE,m=new e.Map,f=new RegExp(/\S/),_=e.isInsideJsxElement(a,s[l]),g=_?"{/*":"//",y=l;y<=u;y++){var v=a.text.substring(s[y],a.getLineEndOfPosition(s[y])),h=f.exec(v);h&&(p=Math.min(p,h.index),m.set(y.toString(),h.index),v.substr(h.index,g.length)!==g&&(d=void 0===r||r))}for(y=l;y<=u;y++)if(l===u||s[y]!==t.end){var b=m.get(y.toString());void 0!==b&&(_?i.push.apply(i,P(n,{pos:s[y]+p,end:a.getLineEndOfPosition(s[y])},d,_)):d?i.push({newText:g,span:{length:0,start:s[y]+p}}):a.text.substr(s[y]+b,g.length)===g&&i.push({newText:"",span:{length:g.length,start:s[y]+b}}))}return i}function P(n,t,r,a){for(var i,o=c.getCurrentSourceFile(n),s=[],l=o.text,u=!1,d=r||!1,p=[],m=t.pos,f=void 0!==a?a:e.isInsideJsxElement(o,m),_=f?"{/*":"/*",g=f?"*/}":"*/",y=f?"\\{\\/\\*":"\\/\\*",v=f?"\\*\\/\\}":"\\*\\/";m<=t.end;){var h=l.substr(m,_.length)===_?_.length:0,b=e.isInComment(o,m+h);if(b)f&&(b.pos--,b.end++),p.push(b.pos),3===b.kind&&p.push(b.end),u=!0,m=b.end+1;else{var E=l.substring(m,t.end).search("("+y+")|("+v+")");d=void 0!==r?r:d||!e.isTextWhiteSpaceLike(l,m,-1===E?t.end:m+E),m=-1===E?t.end+1:m+E+g.length}}if(d||!u){2!==(null===(i=e.isInComment(o,t.pos))||void 0===i?void 0:i.kind)&&e.insertSorted(p,t.pos,e.compareValues),e.insertSorted(p,t.end,e.compareValues);var T=p[0];l.substr(T,_.length)!==_&&s.push({newText:_,span:{length:0,start:T}});for(var S=1;S<p.length-1;S++)l.substr(p[S]-g.length,g.length)!==g&&s.push({newText:g,span:{length:0,start:p[S]}}),l.substr(p[S],_.length)!==_&&s.push({newText:_,span:{length:0,start:p[S]}});s.length%2!=0&&s.push({newText:g,span:{length:0,start:p[p.length-1]}})}else for(var x=0,L=p;x<L.length;x++){var A=L[x],D=A-g.length>0?A-g.length:0;h=l.substr(D,g.length)===g?g.length:0;s.push({newText:"",span:{length:_.length,start:A-h}})}return s}function M(n){var t=n.openingElement,r=n.closingElement,a=n.parent;return!e.tagNamesAreEquivalent(t.tagName,r.tagName)||e.isJsxElement(a)&&e.tagNamesAreEquivalent(t.tagName,a.openingElement.tagName)&&M(a)}function w(t,r,a,i,o){var s="number"==typeof r?[r,void 0]:[r.pos,r.end];return{file:t,startPosition:s[0],endPosition:s[1],program:E(),host:n,formatContext:e.formatting.getFormatContext(i,n),cancellationToken:u,preferences:a,triggerReason:o}}N.forEach((function(e,n){return N.set(e.toString(),Number(n))}));var F={dispose:function(){if(o){var r=t.getKeyForCompilationSettings(o.getCompilerOptions());e.forEach(o.getSourceFiles(),(function(e){return t.releaseDocumentWithKey(e.resolvedPath,r)})),o=void 0}n=void 0},cleanupSemanticCache:function(){o=void 0},getSyntacticDiagnostics:function(e){return b(),o.getSyntacticDiagnostics(y(e),u).slice()},getSemanticDiagnostics:function(n){b();var t=y(n),r=o.getSemanticDiagnostics(t,u);if(!e.getEmitDeclarations(o.getCompilerOptions()))return r.slice();var a=o.getDeclarationDiagnostics(t,u);return __spreadArrays(r,a)},getSuggestionDiagnostics:function(n){return b(),e.computeSuggestionDiagnostics(y(n),o,u)},getCompilerOptionsDiagnostics:function(){return b(),__spreadArrays(o.getOptionsDiagnostics(u),o.getGlobalDiagnostics(u))},getSyntacticClassifications:function(n,t){return e.getSyntacticClassifications(u,c.getCurrentSourceFile(n),t)},getSemanticClassifications:function(n,t,r){return I(n)?(b(),"2020"===(r||"original")?e.classifier.v2020.getSemanticClassifications(o,u,y(n),t):e.getSemanticClassifications(o.getTypeChecker(),u,y(n),o.getClassifiableNames(),t)):[]},getEncodedSyntacticClassifications:function(n,t){return e.getEncodedSyntacticClassifications(u,c.getCurrentSourceFile(n),t)},getEncodedSemanticClassifications:function(n,t,r){return I(n)?(b(),"original"===(r||"original")?e.getEncodedSemanticClassifications(o.getTypeChecker(),u,y(n),o.getClassifiableNames(),t):e.classifier.v2020.getEncodedSemanticClassifications(o,u,y(n),t)):{spans:[],endOfLineState:0}},getCompletionsAtPosition:function(t,r,a){void 0===a&&(a=e.emptyOptions);var i=__assign(__assign({},e.identity(a)),{includeCompletionsForModuleExports:a.includeCompletionsForModuleExports||a.includeExternalModuleExports,includeCompletionsWithInsertText:a.includeCompletionsWithInsertText||a.includeInsertTextCompletions});return b(),e.Completions.getCompletionsAtPosition(n,o,p,y(t),r,i,a.triggerCharacter)},getCompletionEntryDetails:function(t,r,a,i,s,c){return void 0===c&&(c=e.emptyOptions),b(),e.Completions.getCompletionEntryDetails(o,p,y(t),r,{name:a,source:s},n,i&&e.formatting.getFormatContext(i,n),c,u)},getCompletionEntrySymbol:function(t,r,a,i,s){return void 0===s&&(s=e.emptyOptions),b(),e.Completions.getCompletionEntrySymbol(o,p,y(t),r,{name:a,source:i},n,s)},getSignatureHelpItems:function(n,t,r){var a=(void 0===r?e.emptyOptions:r).triggerReason;b();var i=y(n);return e.SignatureHelp.getSignatureHelpItems(o,i,t,a,u)},getQuickInfoAtPosition:function(n,t){b();var r=y(n),a=e.getTouchingPropertyName(r,t);if(a!==r){var i=o.getTypeChecker(),s=function(n){if(e.isNewExpression(n.parent)&&n.pos===n.parent.pos)return n.parent.expression;return n}(a),c=function(n,t){var r=D(n);if(r){var a=t.getContextualType(r.parent),i=a&&C(r,t,a,!1);if(i&&1===i.length)return e.first(i)}return t.getSymbolAtLocation(n)}(s,i);if(!c||i.isUnknownSymbol(c)){var l=function(n,t,r){switch(t.kind){case 78:return!e.isLabelName(t)&&!e.isTagName(t);case 201:case 157:return!e.isInComment(n,r);case 107:case 187:case 105:return!0;default:return!1}}(r,s,t)?i.getTypeAtLocation(s):void 0;return l&&{kind:"",kindModifiers:"",textSpan:e.createTextSpanFromNode(s,r),displayParts:i.runWithCancellationToken(u,(function(n){return e.typeToDisplayParts(n,l,e.getContainerNode(s))})),documentation:l.symbol?l.symbol.getDocumentationComment(i):void 0,tags:l.symbol?l.symbol.getJsDocTags():void 0}}var d=i.runWithCancellationToken(u,(function(n){return e.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(n,c,r,e.getContainerNode(s),s)})),p=d.symbolKind,m=d.displayParts,f=d.documentation,_=d.tags;return{kind:p,kindModifiers:e.SymbolDisplay.getSymbolModifiers(c),textSpan:e.createTextSpanFromNode(s,r),displayParts:m,documentation:f,tags:_}}},getDefinitionAtPosition:function(n,t){return b(),e.GoToDefinition.getDefinitionAtPosition(o,y(n),t)},getDefinitionAndBoundSpan:function(n,t){return b(),e.GoToDefinition.getDefinitionAndBoundSpan(o,y(n),t)},getImplementationAtPosition:function(n,t){return b(),e.FindAllReferences.getImplementationsAtPosition(o,u,o.getSourceFiles(),y(n),t)},getTypeDefinitionAtPosition:function(n,t){return b(),e.GoToDefinition.getTypeDefinitionAtPosition(o.getTypeChecker(),y(n),t)},getReferencesAtPosition:function(n,t){return b(),x(e.getTouchingPropertyName(y(n),t),t,{use:1},e.FindAllReferences.toReferenceEntry)},findReferences:function(n,t){return b(),e.FindAllReferences.findReferencedSymbols(o,u,o.getSourceFiles(),y(n),t)},getOccurrencesAtPosition:function(n,t){return e.flatMap(T(n,t,[n]),(function(e){return e.highlightSpans.map((function(n){return __assign(__assign({fileName:e.fileName,textSpan:n.textSpan,isWriteAccess:"writtenReference"===n.kind,isDefinition:!1},n.isInString&&{isInString:!0}),n.contextSpan&&{contextSpan:n.contextSpan})}))}))},getDocumentHighlights:T,getNameOrDottedNameSpan:function(n,t,r){var a=c.getCurrentSourceFile(n),i=e.getTouchingPropertyName(a,t);if(i!==a){switch(i.kind){case 201:case 157:case 10:case 94:case 109:case 103:case 105:case 107:case 187:case 78:break;default:return}for(var o=i;;)if(e.isRightSideOfPropertyAccess(o)||e.isRightSideOfQualifiedName(o))o=o.parent;else{if(!e.isNameOfModuleDeclaration(o))break;if(256!==o.parent.parent.kind||o.parent.parent.body!==o.parent)break;o=o.parent.parent.name}return e.createTextSpanFromBounds(o.getStart(),i.getEnd())}},getBreakpointStatementAtPosition:function(n,t){var r=c.getCurrentSourceFile(n);return e.BreakpointResolver.spanInSourceFileAtLocation(r,t)},getNavigateToItems:function(n,t,r,a){void 0===a&&(a=!1),b();var i=r?[y(r)]:o.getSourceFiles();return e.NavigateTo.getNavigateToItems(i,o.getTypeChecker(),u,n,t,a)},getRenameInfo:function(n,t,r){return b(),e.Rename.getRenameInfo(o,y(n),t,r)},getSmartSelectionRange:function(n,t){return e.SmartSelectionRange.getSmartSelectionRange(t,c.getCurrentSourceFile(n))},findRenameLocations:function(n,t,r,a,i){b();var o=y(n),s=e.getAdjustedRenameLocation(e.getTouchingPropertyName(o,t));if(e.isIdentifier(s)&&(e.isJsxOpeningElement(s.parent)||e.isJsxClosingElement(s.parent))&&e.isIntrinsicJsxName(s.escapedText)){var c=s.parent.parent;return[c.openingElement,c.closingElement].map((function(n){var t=e.createTextSpanFromNode(n.tagName,o);return __assign({fileName:o.fileName,textSpan:t},e.FindAllReferences.toContextSpan(t,o,n.parent))}))}return x(s,t,{findInStrings:r,findInComments:a,providePrefixAndSuffixTextForRename:i,use:2},(function(n,t,r){return e.FindAllReferences.toRenameLocation(n,t,r,i||!1)}))},getNavigationBarItems:function(n){return e.NavigationBar.getNavigationBarItems(c.getCurrentSourceFile(n),u)},getNavigationTree:function(n){return e.NavigationBar.getNavigationTree(c.getCurrentSourceFile(n),u)},getOutliningSpans:function(n){var t=c.getCurrentSourceFile(n);return e.OutliningElementsCollector.collectElements(t,u)},getTodoComments:function(n,t){b();var r=y(n);u.throwIfCancellationRequested();var a,i,o=r.text,s=[];if(t.length>0&&(i=r.fileName,!e.stringContains(i,"/node_modules/")))for(var c=function(){var n="("+/(?:^(?:\s|\*)*)/.source+"|"+/(?:\/\/+\s*)/.source+"|"+/(?:\/\*+\s*)/.source+")",r="(?:"+e.map(t,(function(e){return"("+(e.text.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g,"\\$&")+")")})).join("|")+")";return new RegExp(n+"("+r+/(?:.*?)/.source+")"+/(?:$|\*\/)/.source,"gim")}(),l=void 0;l=c.exec(o);){u.throwIfCancellationRequested();e.Debug.assert(l.length===t.length+3);var d=l[1],p=l.index+d.length;if(e.isInComment(r,p)){for(var m=void 0,f=0;f<t.length;f++)l[f+3]&&(m=t[f]);if(void 0===m)return e.Debug.fail();if(!((a=o.charCodeAt(p+m.text.length))>=97&&a<=122||a>=65&&a<=90||a>=48&&a<=57)){var _=l[2];s.push({descriptor:m,message:_,position:p})}}}return s},getBraceMatchingAtPosition:function(n,t){var r=c.getCurrentSourceFile(n),a=e.getTouchingToken(r,t),i=a.getStart(r)===t?N.get(a.kind.toString()):void 0,o=i&&e.findChildOfKind(a.parent,i,r);return o?[e.createTextSpanFromNode(a,r),e.createTextSpanFromNode(o,r)].sort((function(e,n){return e.start-n.start})):e.emptyArray},getIndentationAtPosition:function(n,t,r){var a=e.timestamp(),i=_(r),o=c.getCurrentSourceFile(n);p("getIndentationAtPosition: getCurrentSourceFile: "+(e.timestamp()-a)),a=e.timestamp();var s=e.formatting.SmartIndenter.getIndentation(t,o,i);return p("getIndentationAtPosition: computeIndentation  : "+(e.timestamp()-a)),s},getFormattingEditsForRange:function(t,r,a,i){var o=c.getCurrentSourceFile(t);return e.formatting.formatSelection(r,a,o,e.formatting.getFormatContext(_(i),n))},getFormattingEditsForDocument:function(t,r){return e.formatting.formatDocument(c.getCurrentSourceFile(t),e.formatting.getFormatContext(_(r),n))},getFormattingEditsAfterKeystroke:function(t,r,a,i){var o=c.getCurrentSourceFile(t),s=e.formatting.getFormatContext(_(i),n);if(!e.isInComment(o,r))switch(a){case"{":return e.formatting.formatOnOpeningCurly(r,o,s);case"}":return e.formatting.formatOnClosingCurly(r,o,s);case";":return e.formatting.formatOnSemicolon(r,o,s);case"\n":return e.formatting.formatOnEnter(r,o,s)}return[]},getDocCommentTemplateAtPosition:function(t,r){return e.JsDoc.getDocCommentTemplateAtPosition(e.getNewLineOrDefaultFromHost(n),c.getCurrentSourceFile(t),r)},isValidBraceCompletionAtPosition:function(n,t,r){if(60===r)return!1;var a=c.getCurrentSourceFile(n);if(e.isInString(a,t))return!1;if(e.isInsideJsxElementOrAttribute(a,t))return 123===r;if(e.isInTemplateString(a,t))return!1;switch(r){case 39:case 34:case 96:return!e.isInComment(a,t)}return!0},getJsxClosingTagAtPosition:function(n,t){var r=c.getCurrentSourceFile(n),a=e.findPrecedingToken(t,r);if(a){var i=31===a.kind&&e.isJsxOpeningElement(a.parent)?a.parent.parent:e.isJsxText(a)?a.parent:void 0;return i&&M(i)?{newText:"</"+i.openingElement.tagName.getText(r)+">"}:void 0}},getSpanOfEnclosingComment:function(n,t,r){var a=c.getCurrentSourceFile(n),i=e.formatting.getRangeOfEnclosingComment(a,t);return!i||r&&3!==i.kind?void 0:e.createTextSpanFromRange(i)},getCodeFixesAtPosition:function(t,r,a,i,s,c){void 0===c&&(c=e.emptyOptions),b();var l=y(t),d=e.createTextSpanFromBounds(r,a),p=e.formatting.getFormatContext(s,n);return e.flatMap(e.deduplicate(i,e.equateValues,e.compareValues),(function(t){return u.throwIfCancellationRequested(),e.codefix.getFixes({errorCode:t,sourceFile:l,span:d,program:o,host:n,cancellationToken:u,formatContext:p,preferences:c})}))},getCombinedCodeFix:function(t,r,a,i){void 0===i&&(i=e.emptyOptions),b(),e.Debug.assert("file"===t.type);var s=y(t.fileName),c=e.formatting.getFormatContext(a,n);return e.codefix.getAllFixes({fixId:r,sourceFile:s,program:o,host:n,cancellationToken:u,formatContext:c,preferences:i})},applyCodeActionCommand:function(n,t){var r="string"==typeof n?t:n;return e.isArray(r)?Promise.all(r.map((function(e){return R(e)}))):R(r)},organizeImports:function(t,r,a){void 0===a&&(a=e.emptyOptions),b(),e.Debug.assert("file"===t.type);var i=y(t.fileName),s=e.formatting.getFormatContext(r,n);return e.OrganizeImports.organizeImports(i,s,n,o,a)},getEditsForFileRename:function(t,r,a,i){return void 0===i&&(i=e.emptyOptions),e.getEditsForFileRename(E(),t,r,n,e.formatting.getFormatContext(a,n),i,g)},getEmitOutput:function(t,r,a){b();var i=y(t),s=n.getCustomTransformers&&n.getCustomTransformers();return e.getFileEmitOutput(o,i,!!r,u,s,a)},getNonBoundSourceFile:function(e){return c.getCurrentSourceFile(e)},getProgram:E,getAutoImportProvider:function(){var e;return null===(e=n.getPackageJsonAutoImportProvider)||void 0===e?void 0:e.call(n)},getApplicableRefactors:function(n,t,r,a){void 0===r&&(r=e.emptyOptions),b();var i=y(n);return e.refactor.getApplicableRefactors(w(i,t,r,e.emptyOptions,a))},getEditsForRefactor:function(n,t,r,a,i,o){void 0===o&&(o=e.emptyOptions),b();var s=y(n);return e.refactor.getEditsForRefactor(w(s,r,o,t),a,i)},toLineColumnOffset:g.toLineColumnOffset,getSourceMapper:function(){return g},clearSourceMapperCache:function(){return g.clearCache()},prepareCallHierarchy:function(n,t){b();var r=e.CallHierarchy.resolveCallHierarchyDeclaration(o,e.getTouchingPropertyName(y(n),t));return r&&e.mapOneOrMany(r,(function(n){return e.CallHierarchy.createCallHierarchyItem(o,n)}))},provideCallHierarchyIncomingCalls:function(n,t){b();var r=y(n),a=e.firstOrOnly(e.CallHierarchy.resolveCallHierarchyDeclaration(o,0===t?r:e.getTouchingPropertyName(r,t)));return a?e.CallHierarchy.getIncomingCalls(o,a,u):[]},provideCallHierarchyOutgoingCalls:function(n,t){b();var r=y(n),a=e.firstOrOnly(e.CallHierarchy.resolveCallHierarchyDeclaration(o,0===t?r:e.getTouchingPropertyName(r,t)));return a?e.CallHierarchy.getOutgoingCalls(o,a):[]},toggleLineComment:O,toggleMultilineComment:P,commentSelection:function(e,n){var t=k(c.getCurrentSourceFile(e),n);return t.firstLine===t.lastLine&&n.pos!==n.end?P(e,n,!0):O(e,n,!0)},uncommentSelection:function(n,t){var r=c.getCurrentSourceFile(n),a=[],i=t.pos,o=t.end;i===o&&(o+=e.isInsideJsxElement(r,i)?2:1);for(var s=i;s<=o;s++){var l=e.isInComment(r,s);if(l){switch(l.kind){case 2:a.push.apply(a,O(n,{end:l.end,pos:l.pos+1},!1));break;case 3:a.push.apply(a,P(n,{end:l.end,pos:l.pos+1},!1))}s=l.end+1}}return a}};switch(i){case e.LanguageServiceMode.Semantic:break;case e.LanguageServiceMode.PartialSemantic:L.forEach((function(e){return F[e]=function(){throw new Error("LanguageService Operation: "+e+" not allowed in LanguageServiceMode.PartialSemantic")}}));break;case e.LanguageServiceMode.Syntactic:A.forEach((function(e){return F[e]=function(){throw new Error("LanguageService Operation: "+e+" not allowed in LanguageServiceMode.Syntactic")}}));break;default:e.Debug.assertNever(i)}return F},e.getNameTable=function(n){return n.nameTable||function(n){var t=n.nameTable=new e.Map;n.forEachChild((function n(r){if(e.isIdentifier(r)&&!e.isTagName(r)&&r.escapedText||e.isStringOrNumericLiteralLike(r)&&function(n){return e.isDeclarationName(n)||272===n.parent.kind||function(e){return e&&e.parent&&202===e.parent.kind&&e.parent.argumentExpression===e}(n)||e.isLiteralComputedPropertyDeclarationName(n)}(r)){var a=e.getEscapedTextOfIdentifierOrLiteral(r);t.set(a,void 0===t.get(a)?r.pos:-1)}else if(e.isPrivateIdentifier(r)){a=r.escapedText;t.set(a,void 0===t.get(a)?r.pos:-1)}if(e.forEachChild(r,n),e.hasJSDocNodes(r))for(var i=0,o=r.jsDoc;i<o.length;i++){var s=o[i];e.forEachChild(s,n)}}))}(n),n.nameTable},e.getContainingObjectLiteralElement=D,e.getPropertySymbolsFromContextualType=C,e.getDefaultLibFilePath=function(n){if("undefined"!=typeof __dirname)return __dirname+e.directorySeparator+e.getDefaultLibFileName(n);throw new Error("getDefaultLibFilePath is only supported when consumed as a node module. ")},e.setObjectAllocator({getNodeConstructor:function(){return t},getTokenConstructor:function(){return o},getIdentifierConstructor:function(){return s},getPrivateIdentifierConstructor:function(){return c},getSourceFileConstructor:function(){return m},getSymbolConstructor:function(){return i},getTypeConstructor:function(){return l},getSignatureConstructor:function(){return u},getSourceMapSourceConstructor:function(){return f}})}(ts||(ts={})),function(e){!function(n){n.spanInSourceFileAtLocation=function(n,t){if(!n.isDeclarationFile){var r=e.getTokenAtPosition(n,t),a=n.getLineAndCharacterOfPosition(t).line;if(n.getLineAndCharacterOfPosition(r.getStart(n)).line>a){var i=e.findPrecedingToken(r.pos,n);if(!i||n.getLineAndCharacterOfPosition(i.getEnd()).line!==a)return;r=i}if(!(8388608&r.flags))return d(r)}function o(t,r){var a=t.decorators?e.skipTrivia(n.text,t.decorators.end):t.getStart(n);return e.createTextSpanFromBounds(a,(r||t).getEnd())}function s(t,r){return o(t,e.findNextToken(r,r.parent,n))}function c(e,t){return e&&a===n.getLineAndCharacterOfPosition(e.getStart(n)).line?d(e):d(t)}function l(t){return d(e.findPrecedingToken(t.pos,n))}function u(t){return d(e.findNextToken(t,t.parent,n))}function d(t){if(t){var r=t.parent;switch(t.kind){case 232:return v(t.declarationList.declarations[0]);case 249:case 163:case 162:return v(t);case 160:return function n(t){if(e.isBindingPattern(t.name))return T(t.name);if(function(n){return!!n.initializer||void 0!==n.dotDotDotToken||e.hasSyntacticModifier(n,12)}(t))return o(t);var r=t.parent,a=r.parameters.indexOf(t);return e.Debug.assert(-1!==a),0!==a?n(r.parameters[a-1]):d(r.body)}(t);case 251:case 165:case 164:case 167:case 168:case 166:case 208:case 209:return function(e){if(!e.body)return;if(h(e))return o(e);return d(e.body)}(t);case 230:if(e.isFunctionBlock(t))return function(e){var n=e.statements.length?e.statements[0]:e.getLastToken();if(h(e.parent))return c(e.parent,n);return d(n)}(t);case 257:return b(t);case 287:return b(t.block);case 233:return o(t.expression);case 242:return o(t.getChildAt(0),t.expression);case 236:return s(t,t.expression);case 235:return d(t.statement);case 248:return o(t.getChildAt(0));case 234:return s(t,t.expression);case 245:return d(t.statement);case 241:case 240:return o(t.getChildAt(0),t.label);case 237:return function(e){if(e.initializer)return E(e);if(e.condition)return o(e.condition);if(e.incrementor)return o(e.incrementor)}(t);case 238:return s(t,t.expression);case 239:return E(t);case 244:return s(t,t.expression);case 284:case 285:return d(t.statements[0]);case 247:return b(t.tryBlock);case 246:case 266:return o(t,t.expression);case 260:return o(t,t.moduleReference);case 261:case 267:return o(t,t.moduleSpecifier);case 256:if(1!==e.getModuleInstanceState(t))return;case 252:case 255:case 291:case 198:return o(t);case 243:return d(t.statement);case 161:return g=r.decorators,e.createTextSpanFromBounds(e.skipTrivia(n.text,g.pos),g.end);case 196:case 197:return T(t);case 253:case 254:return;case 26:case 1:return c(e.findPrecedingToken(t.pos,n));case 27:return l(t);case 18:return function(t){switch(t.parent.kind){case 255:var r=t.parent;return c(e.findPrecedingToken(t.pos,n,t.parent),r.members.length?r.members[0]:r.getLastToken(n));case 252:var a=t.parent;return c(e.findPrecedingToken(t.pos,n,t.parent),a.members.length?a.members[0]:a.getLastToken(n));case 258:return c(t.parent.parent,t.parent.clauses[0])}return d(t.parent)}(t);case 19:return function(n){switch(n.parent.kind){case 257:if(1!==e.getModuleInstanceState(n.parent.parent))return;case 255:case 252:return o(n);case 230:if(e.isFunctionBlock(n.parent))return o(n);case 287:return d(e.lastOrUndefined(n.parent.statements));case 258:var t=n.parent,r=e.lastOrUndefined(t.clauses);return r?d(e.lastOrUndefined(r.statements)):void 0;case 196:var a=n.parent;return d(e.lastOrUndefined(a.elements)||a);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent)){var i=n.parent;return o(e.lastOrUndefined(i.properties)||i)}return d(n.parent)}}(t);case 23:return function(n){switch(n.parent.kind){case 197:var t=n.parent;return o(e.lastOrUndefined(t.elements)||t);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(n.parent)){var r=n.parent;return o(e.lastOrUndefined(r.elements)||r)}return d(n.parent)}}(t);case 20:return function(e){if(235===e.parent.kind||203===e.parent.kind||204===e.parent.kind)return l(e);if(207===e.parent.kind)return u(e);return d(e.parent)}(t);case 21:return function(e){switch(e.parent.kind){case 208:case 251:case 209:case 165:case 164:case 167:case 168:case 166:case 236:case 235:case 237:case 239:case 203:case 204:case 207:return l(e);default:return d(e.parent)}}(t);case 58:return function(n){if(e.isFunctionLike(n.parent)||288===n.parent.kind||160===n.parent.kind)return l(n);return d(n.parent)}(t);case 31:case 29:return function(e){if(206===e.parent.kind)return u(e);return d(e.parent)}(t);case 114:return function(e){if(235===e.parent.kind)return s(e,e.parent.expression);return d(e.parent)}(t);case 90:case 82:case 95:return u(t);case 156:return function(e){if(239===e.parent.kind)return u(e);return d(e.parent)}(t);default:if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(t))return S(t);if((78===t.kind||220===t.kind||288===t.kind||289===t.kind)&&e.isArrayLiteralOrObjectLiteralDestructuringPattern(r))return o(t);if(216===t.kind){var a=t,i=a.left,p=a.operatorToken;if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(i))return S(i);if(62===p.kind&&e.isArrayLiteralOrObjectLiteralDestructuringPattern(t.parent))return o(t);if(27===p.kind)return d(i)}if(e.isExpressionNode(t))switch(r.kind){case 235:return l(t);case 161:return d(t.parent);case 237:case 239:return o(t);case 216:if(27===t.parent.operatorToken.kind)return o(t);break;case 209:if(t.parent.body===t)return o(t)}switch(t.parent.kind){case 288:if(t.parent.name===t&&!e.isArrayLiteralOrObjectLiteralDestructuringPattern(t.parent.parent))return d(t.parent.initializer);break;case 206:if(t.parent.type===t)return u(t.parent.type);break;case 249:case 160:var m=t.parent,f=m.initializer,_=m.type;if(f===t||_===t||e.isAssignmentOperator(t.kind))return l(t);break;case 216:i=t.parent.left;if(e.isArrayLiteralOrObjectLiteralDestructuringPattern(i)&&t!==i)return l(t);break;default:if(e.isFunctionLike(t.parent)&&t.parent.type===t)return l(t)}return d(t.parent)}}var g;function y(t){return e.isVariableDeclarationList(t.parent)&&t.parent.declarations[0]===t?o(e.findPrecedingToken(t.pos,n,t.parent),t):o(t)}function v(t){if(238===t.parent.parent.kind)return d(t.parent.parent);var r=t.parent;return e.isBindingPattern(t.name)?T(t.name):t.initializer||e.hasSyntacticModifier(t,1)||239===r.parent.kind?y(t):e.isVariableDeclarationList(t.parent)&&t.parent.declarations[0]!==t?d(e.findPrecedingToken(t.pos,n,t.parent)):void 0}function h(n){return e.hasSyntacticModifier(n,1)||252===n.parent.kind&&166!==n.kind}function b(t){switch(t.parent.kind){case 256:if(1!==e.getModuleInstanceState(t.parent))return;case 236:case 234:case 238:return c(t.parent,t.statements[0]);case 237:case 239:return c(e.findPrecedingToken(t.pos,n,t.parent),t.statements[0])}return d(t.statements[0])}function E(e){if(250!==e.initializer.kind)return d(e.initializer);var n=e.initializer;return n.declarations.length>0?d(n.declarations[0]):void 0}function T(n){var t=e.forEach(n.elements,(function(e){return 222!==e.kind?e:void 0}));return t?d(t):198===n.parent.kind?o(n.parent):y(n.parent)}function S(n){e.Debug.assert(197!==n.kind&&196!==n.kind);var t=199===n.kind?n.elements:n.properties,r=e.forEach(t,(function(e){return 222!==e.kind?e:void 0}));return r?d(r):o(216===n.parent.kind?n.parent:n)}}}}(e.BreakpointResolver||(e.BreakpointResolver={}))}(ts||(ts={})),function(e){e.transform=function(n,t,r){var a=[];r=e.fixupCompilerOptions(r,a);var i=e.isArray(n)?n:[n],o=e.transformNodes(void 0,void 0,e.factory,r,i,t,!0);return o.diagnostics=e.concatenate(o.diagnostics,a),o}}(ts||(ts={}));var ts,debugObjectHost=function(){return this}();!function(e){function n(e,n){e&&e.log("*INTERNAL ERROR* - Exception in typescript services: "+n.message)}var t=function(){function n(e){this.scriptSnapshotShim=e}return n.prototype.getText=function(e,n){return this.scriptSnapshotShim.getText(e,n)},n.prototype.getLength=function(){return this.scriptSnapshotShim.getLength()},n.prototype.getChangeRange=function(n){var t=n,r=this.scriptSnapshotShim.getChangeRange(t.scriptSnapshotShim);if(null===r)return null;var a=JSON.parse(r);return e.createTextChangeRange(e.createTextSpan(a.span.start,a.span.length),a.newLength)},n.prototype.dispose=function(){"dispose"in this.scriptSnapshotShim&&this.scriptSnapshotShim.dispose()},n}(),r=function(){function n(n){var t=this;this.shimHost=n,this.loggingEnabled=!1,this.tracingEnabled=!1,"getModuleResolutionsForFile"in this.shimHost&&(this.resolveModuleNames=function(n,r){var a=JSON.parse(t.shimHost.getModuleResolutionsForFile(r));return e.map(n,(function(n){var t=e.getProperty(a,n);return t?{resolvedFileName:t,extension:e.extensionFromPath(t),isExternalLibraryImport:!1}:void 0}))}),"directoryExists"in this.shimHost&&(this.directoryExists=function(e){return t.shimHost.directoryExists(e)}),"getTypeReferenceDirectiveResolutionsForFile"in this.shimHost&&(this.resolveTypeReferenceDirectives=function(n,r){var a=JSON.parse(t.shimHost.getTypeReferenceDirectiveResolutionsForFile(r));return e.map(n,(function(n){return e.getProperty(a,n)}))})}return n.prototype.log=function(e){this.loggingEnabled&&this.shimHost.log(e)},n.prototype.trace=function(e){this.tracingEnabled&&this.shimHost.trace(e)},n.prototype.error=function(e){this.shimHost.error(e)},n.prototype.getProjectVersion=function(){if(this.shimHost.getProjectVersion)return this.shimHost.getProjectVersion()},n.prototype.getTypeRootsVersion=function(){return this.shimHost.getTypeRootsVersion?this.shimHost.getTypeRootsVersion():0},n.prototype.useCaseSensitiveFileNames=function(){return!!this.shimHost.useCaseSensitiveFileNames&&this.shimHost.useCaseSensitiveFileNames()},n.prototype.getCompilationSettings=function(){var e=this.shimHost.getCompilationSettings();if(null===e||""===e)throw Error("LanguageServiceShimHostAdapter.getCompilationSettings: empty compilationSettings");var n=JSON.parse(e);return n.allowNonTsExtensions=!0,n},n.prototype.getScriptFileNames=function(){var e=this.shimHost.getScriptFileNames();return JSON.parse(e)},n.prototype.getScriptSnapshot=function(e){var n=this.shimHost.getScriptSnapshot(e);return n&&new t(n)},n.prototype.getScriptKind=function(e){return"getScriptKind"in this.shimHost?this.shimHost.getScriptKind(e):0},n.prototype.getScriptVersion=function(e){return this.shimHost.getScriptVersion(e)},n.prototype.getLocalizedDiagnosticMessages=function(){var e=this.shimHost.getLocalizedDiagnosticMessages();if(null===e||""===e)return null;try{return JSON.parse(e)}catch(e){return this.log(e.description||"diagnosticMessages.generated.json has invalid JSON format"),null}},n.prototype.getCancellationToken=function(){var n=this.shimHost.getCancellationToken();return new e.ThrottledCancellationToken(n)},n.prototype.getCurrentDirectory=function(){return this.shimHost.getCurrentDirectory()},n.prototype.getDirectories=function(e){return JSON.parse(this.shimHost.getDirectories(e))},n.prototype.getDefaultLibFileName=function(e){return this.shimHost.getDefaultLibFileName(JSON.stringify(e))},n.prototype.readDirectory=function(n,t,r,a,i){var o=e.getFileMatcherPatterns(n,r,a,this.shimHost.useCaseSensitiveFileNames(),this.shimHost.getCurrentDirectory());return JSON.parse(this.shimHost.readDirectory(n,JSON.stringify(t),JSON.stringify(o.basePaths),o.excludePattern,o.includeFilePattern,o.includeDirectoryPattern,i))},n.prototype.readFile=function(e,n){return this.shimHost.readFile(e,n)},n.prototype.fileExists=function(e){return this.shimHost.fileExists(e)},n}();e.LanguageServiceShimHostAdapter=r;var a=function(){function n(e){var n=this;this.shimHost=e,this.useCaseSensitiveFileNames=!!this.shimHost.useCaseSensitiveFileNames&&this.shimHost.useCaseSensitiveFileNames(),"directoryExists"in this.shimHost?this.directoryExists=function(e){return n.shimHost.directoryExists(e)}:this.directoryExists=void 0,"realpath"in this.shimHost?this.realpath=function(e){return n.shimHost.realpath(e)}:this.realpath=void 0}return n.prototype.readDirectory=function(n,t,r,a,i){var o=e.getFileMatcherPatterns(n,r,a,this.shimHost.useCaseSensitiveFileNames(),this.shimHost.getCurrentDirectory());return JSON.parse(this.shimHost.readDirectory(n,JSON.stringify(t),JSON.stringify(o.basePaths),o.excludePattern,o.includeFilePattern,o.includeDirectoryPattern,i))},n.prototype.fileExists=function(e){return this.shimHost.fileExists(e)},n.prototype.readFile=function(e){return this.shimHost.readFile(e)},n.prototype.getDirectories=function(e){return JSON.parse(this.shimHost.getDirectories(e))},n}();function i(e,n,t,r){return o(e,n,!0,t,r)}function o(t,r,a,i,o){try{var s=function(n,t,r,a){var i;a&&(n.log(t),i=e.timestamp());var o=r();if(a){var s=e.timestamp();if(n.log(t+" completed in "+(s-i)+" msec"),e.isString(o)){var c=o;c.length>128&&(c=c.substring(0,128)+"..."),n.log("  result.length="+c.length+", result='"+JSON.stringify(c)+"'")}}return o}(t,r,i,o);return a?JSON.stringify({result:s}):s}catch(a){return a instanceof e.OperationCanceledException?JSON.stringify({canceled:!0}):(n(t,a),a.description=r,JSON.stringify({error:a}))}}e.CoreServicesShimHostAdapter=a;var s=function(){function e(e){this.factory=e,e.registerShim(this)}return e.prototype.dispose=function(e){this.factory.unregisterShim(this)},e}();function c(n,t){return n.map((function(n){return function(n,t){return{message:e.flattenDiagnosticMessageText(n.messageText,t),start:n.start,length:n.length,category:e.diagnosticCategoryName(n),code:n.code,reportsUnnecessary:n.reportsUnnecessary,reportsDeprecated:n.reportsDeprecated}}(n,t)}))}e.realizeDiagnostics=c;var l=function(n){function t(e,t,r){var a=n.call(this,e)||this;return a.host=t,a.languageService=r,a.logPerformance=!1,a.logger=a.host,a}return __extends(t,n),t.prototype.forwardJSONCall=function(e,n){return i(this.logger,e,n,this.logPerformance)},t.prototype.dispose=function(e){this.logger.log("dispose()"),this.languageService.dispose(),this.languageService=null,debugObjectHost&&debugObjectHost.CollectGarbage&&(debugObjectHost.CollectGarbage(),this.logger.log("CollectGarbage()")),this.logger=null,n.prototype.dispose.call(this,e)},t.prototype.refresh=function(e){this.forwardJSONCall("refresh("+e+")",(function(){return null}))},t.prototype.cleanupSemanticCache=function(){var e=this;this.forwardJSONCall("cleanupSemanticCache()",(function(){return e.languageService.cleanupSemanticCache(),null}))},t.prototype.realizeDiagnostics=function(n){return c(n,e.getNewLineOrDefaultFromHost(this.host))},t.prototype.getSyntacticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getSyntacticClassifications('"+n+"', "+t+", "+r+")",(function(){return a.languageService.getSyntacticClassifications(n,e.createTextSpan(t,r))}))},t.prototype.getSemanticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getSemanticClassifications('"+n+"', "+t+", "+r+")",(function(){return a.languageService.getSemanticClassifications(n,e.createTextSpan(t,r))}))},t.prototype.getEncodedSyntacticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getEncodedSyntacticClassifications('"+n+"', "+t+", "+r+")",(function(){return u(a.languageService.getEncodedSyntacticClassifications(n,e.createTextSpan(t,r)))}))},t.prototype.getEncodedSemanticClassifications=function(n,t,r){var a=this;return this.forwardJSONCall("getEncodedSemanticClassifications('"+n+"', "+t+", "+r+")",(function(){return u(a.languageService.getEncodedSemanticClassifications(n,e.createTextSpan(t,r)))}))},t.prototype.getSyntacticDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSyntacticDiagnostics('"+e+"')",(function(){var t=n.languageService.getSyntacticDiagnostics(e);return n.realizeDiagnostics(t)}))},t.prototype.getSemanticDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSemanticDiagnostics('"+e+"')",(function(){var t=n.languageService.getSemanticDiagnostics(e);return n.realizeDiagnostics(t)}))},t.prototype.getSuggestionDiagnostics=function(e){var n=this;return this.forwardJSONCall("getSuggestionDiagnostics('"+e+"')",(function(){return n.realizeDiagnostics(n.languageService.getSuggestionDiagnostics(e))}))},t.prototype.getCompilerOptionsDiagnostics=function(){var e=this;return this.forwardJSONCall("getCompilerOptionsDiagnostics()",(function(){var n=e.languageService.getCompilerOptionsDiagnostics();return e.realizeDiagnostics(n)}))},t.prototype.getQuickInfoAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getQuickInfoAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getQuickInfoAtPosition(e,n)}))},t.prototype.getNameOrDottedNameSpan=function(e,n,t){var r=this;return this.forwardJSONCall("getNameOrDottedNameSpan('"+e+"', "+n+", "+t+")",(function(){return r.languageService.getNameOrDottedNameSpan(e,n,t)}))},t.prototype.getBreakpointStatementAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getBreakpointStatementAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getBreakpointStatementAtPosition(e,n)}))},t.prototype.getSignatureHelpItems=function(e,n,t){var r=this;return this.forwardJSONCall("getSignatureHelpItems('"+e+"', "+n+")",(function(){return r.languageService.getSignatureHelpItems(e,n,t)}))},t.prototype.getDefinitionAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getDefinitionAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getDefinitionAtPosition(e,n)}))},t.prototype.getDefinitionAndBoundSpan=function(e,n){var t=this;return this.forwardJSONCall("getDefinitionAndBoundSpan('"+e+"', "+n+")",(function(){return t.languageService.getDefinitionAndBoundSpan(e,n)}))},t.prototype.getTypeDefinitionAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getTypeDefinitionAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getTypeDefinitionAtPosition(e,n)}))},t.prototype.getImplementationAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getImplementationAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getImplementationAtPosition(e,n)}))},t.prototype.getRenameInfo=function(e,n,t){var r=this;return this.forwardJSONCall("getRenameInfo('"+e+"', "+n+")",(function(){return r.languageService.getRenameInfo(e,n,t)}))},t.prototype.getSmartSelectionRange=function(e,n){var t=this;return this.forwardJSONCall("getSmartSelectionRange('"+e+"', "+n+")",(function(){return t.languageService.getSmartSelectionRange(e,n)}))},t.prototype.findRenameLocations=function(e,n,t,r,a){var i=this;return this.forwardJSONCall("findRenameLocations('"+e+"', "+n+", "+t+", "+r+", "+a+")",(function(){return i.languageService.findRenameLocations(e,n,t,r,a)}))},t.prototype.getBraceMatchingAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getBraceMatchingAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getBraceMatchingAtPosition(e,n)}))},t.prototype.isValidBraceCompletionAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("isValidBraceCompletionAtPosition('"+e+"', "+n+", "+t+")",(function(){return r.languageService.isValidBraceCompletionAtPosition(e,n,t)}))},t.prototype.getSpanOfEnclosingComment=function(e,n,t){var r=this;return this.forwardJSONCall("getSpanOfEnclosingComment('"+e+"', "+n+")",(function(){return r.languageService.getSpanOfEnclosingComment(e,n,t)}))},t.prototype.getIndentationAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("getIndentationAtPosition('"+e+"', "+n+")",(function(){var a=JSON.parse(t);return r.languageService.getIndentationAtPosition(e,n,a)}))},t.prototype.getReferencesAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getReferencesAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getReferencesAtPosition(e,n)}))},t.prototype.findReferences=function(e,n){var t=this;return this.forwardJSONCall("findReferences('"+e+"', "+n+")",(function(){return t.languageService.findReferences(e,n)}))},t.prototype.getOccurrencesAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getOccurrencesAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getOccurrencesAtPosition(e,n)}))},t.prototype.getDocumentHighlights=function(n,t,r){var a=this;return this.forwardJSONCall("getDocumentHighlights('"+n+"', "+t+")",(function(){var i=a.languageService.getDocumentHighlights(n,t,JSON.parse(r)),o=e.toFileNameLowerCase(e.normalizeSlashes(n));return e.filter(i,(function(n){return e.toFileNameLowerCase(e.normalizeSlashes(n.fileName))===o}))}))},t.prototype.getCompletionsAtPosition=function(e,n,t){var r=this;return this.forwardJSONCall("getCompletionsAtPosition('"+e+"', "+n+", "+t+")",(function(){return r.languageService.getCompletionsAtPosition(e,n,t)}))},t.prototype.getCompletionEntryDetails=function(e,n,t,r,a,i){var o=this;return this.forwardJSONCall("getCompletionEntryDetails('"+e+"', "+n+", '"+t+"')",(function(){var s=void 0===r?void 0:JSON.parse(r);return o.languageService.getCompletionEntryDetails(e,n,t,s,a,i)}))},t.prototype.getFormattingEditsForRange=function(e,n,t,r){var a=this;return this.forwardJSONCall("getFormattingEditsForRange('"+e+"', "+n+", "+t+")",(function(){var i=JSON.parse(r);return a.languageService.getFormattingEditsForRange(e,n,t,i)}))},t.prototype.getFormattingEditsForDocument=function(e,n){var t=this;return this.forwardJSONCall("getFormattingEditsForDocument('"+e+"')",(function(){var r=JSON.parse(n);return t.languageService.getFormattingEditsForDocument(e,r)}))},t.prototype.getFormattingEditsAfterKeystroke=function(e,n,t,r){var a=this;return this.forwardJSONCall("getFormattingEditsAfterKeystroke('"+e+"', "+n+", '"+t+"')",(function(){var i=JSON.parse(r);return a.languageService.getFormattingEditsAfterKeystroke(e,n,t,i)}))},t.prototype.getDocCommentTemplateAtPosition=function(e,n){var t=this;return this.forwardJSONCall("getDocCommentTemplateAtPosition('"+e+"', "+n+")",(function(){return t.languageService.getDocCommentTemplateAtPosition(e,n)}))},t.prototype.getNavigateToItems=function(e,n,t){var r=this;return this.forwardJSONCall("getNavigateToItems('"+e+"', "+n+", "+t+")",(function(){return r.languageService.getNavigateToItems(e,n,t)}))},t.prototype.getNavigationBarItems=function(e){var n=this;return this.forwardJSONCall("getNavigationBarItems('"+e+"')",(function(){return n.languageService.getNavigationBarItems(e)}))},t.prototype.getNavigationTree=function(e){var n=this;return this.forwardJSONCall("getNavigationTree('"+e+"')",(function(){return n.languageService.getNavigationTree(e)}))},t.prototype.getOutliningSpans=function(e){var n=this;return this.forwardJSONCall("getOutliningSpans('"+e+"')",(function(){return n.languageService.getOutliningSpans(e)}))},t.prototype.getTodoComments=function(e,n){var t=this;return this.forwardJSONCall("getTodoComments('"+e+"')",(function(){return t.languageService.getTodoComments(e,JSON.parse(n))}))},t.prototype.prepareCallHierarchy=function(e,n){var t=this;return this.forwardJSONCall("prepareCallHierarchy('"+e+"', "+n+")",(function(){return t.languageService.prepareCallHierarchy(e,n)}))},t.prototype.provideCallHierarchyIncomingCalls=function(e,n){var t=this;return this.forwardJSONCall("provideCallHierarchyIncomingCalls('"+e+"', "+n+")",(function(){return t.languageService.provideCallHierarchyIncomingCalls(e,n)}))},t.prototype.provideCallHierarchyOutgoingCalls=function(e,n){var t=this;return this.forwardJSONCall("provideCallHierarchyOutgoingCalls('"+e+"', "+n+")",(function(){return t.languageService.provideCallHierarchyOutgoingCalls(e,n)}))},t.prototype.getEmitOutput=function(e){var n=this;return this.forwardJSONCall("getEmitOutput('"+e+"')",(function(){var t=n.languageService.getEmitOutput(e),r=t.diagnostics,a=__rest(t,["diagnostics"]);return __assign(__assign({},a),{diagnostics:n.realizeDiagnostics(r)})}))},t.prototype.getEmitOutputObject=function(e){var n=this;return o(this.logger,"getEmitOutput('"+e+"')",!1,(function(){return n.languageService.getEmitOutput(e)}),this.logPerformance)},t.prototype.toggleLineComment=function(e,n){var t=this;return this.forwardJSONCall("toggleLineComment('"+e+"', '"+JSON.stringify(n)+"')",(function(){return t.languageService.toggleLineComment(e,n)}))},t.prototype.toggleMultilineComment=function(e,n){var t=this;return this.forwardJSONCall("toggleMultilineComment('"+e+"', '"+JSON.stringify(n)+"')",(function(){return t.languageService.toggleMultilineComment(e,n)}))},t.prototype.commentSelection=function(e,n){var t=this;return this.forwardJSONCall("commentSelection('"+e+"', '"+JSON.stringify(n)+"')",(function(){return t.languageService.commentSelection(e,n)}))},t.prototype.uncommentSelection=function(e,n){var t=this;return this.forwardJSONCall("uncommentSelection('"+e+"', '"+JSON.stringify(n)+"')",(function(){return t.languageService.uncommentSelection(e,n)}))},t}(s);function u(e){return{spans:e.spans.join(","),endOfLineState:e.endOfLineState}}var d=function(n){function t(t,r){var a=n.call(this,t)||this;return a.logger=r,a.logPerformance=!1,a.classifier=e.createClassifier(),a}return __extends(t,n),t.prototype.getEncodedLexicalClassifications=function(e,n,t){var r=this;return void 0===t&&(t=!1),i(this.logger,"getEncodedLexicalClassifications",(function(){return u(r.classifier.getEncodedLexicalClassifications(e,n,t))}),this.logPerformance)},t.prototype.getClassificationsForLine=function(e,n,t){void 0===t&&(t=!1);for(var r=this.classifier.getClassificationsForLine(e,n,t),a="",i=0,o=r.entries;i<o.length;i++){var s=o[i];a+=s.length+"\n",a+=s.classification+"\n"}return a+=r.finalLexState},t}(s),p=function(n){function t(e,t,r){var a=n.call(this,e)||this;return a.logger=t,a.host=r,a.logPerformance=!1,a}return __extends(t,n),t.prototype.forwardJSONCall=function(e,n){return i(this.logger,e,n,this.logPerformance)},t.prototype.resolveModuleName=function(n,t,r){var a=this;return this.forwardJSONCall("resolveModuleName('"+n+"')",(function(){var i=JSON.parse(r),o=e.resolveModuleName(t,e.normalizeSlashes(n),i,a.host),s=o.resolvedModule?o.resolvedModule.resolvedFileName:void 0;return o.resolvedModule&&".ts"!==o.resolvedModule.extension&&".tsx"!==o.resolvedModule.extension&&".d.ts"!==o.resolvedModule.extension&&(s=void 0),{resolvedFileName:s,failedLookupLocations:o.failedLookupLocations}}))},t.prototype.resolveTypeReferenceDirective=function(n,t,r){var a=this;return this.forwardJSONCall("resolveTypeReferenceDirective("+n+")",(function(){var i=JSON.parse(r),o=e.resolveTypeReferenceDirective(t,e.normalizeSlashes(n),i,a.host);return{resolvedFileName:o.resolvedTypeReferenceDirective?o.resolvedTypeReferenceDirective.resolvedFileName:void 0,primary:!o.resolvedTypeReferenceDirective||o.resolvedTypeReferenceDirective.primary,failedLookupLocations:o.failedLookupLocations}}))},t.prototype.getPreProcessedFileInfo=function(n,t){var r=this;return this.forwardJSONCall("getPreProcessedFileInfo('"+n+"')",(function(){var n=e.preProcessFile(e.getSnapshotText(t),!0,!0);return{referencedFiles:r.convertFileReferences(n.referencedFiles),importedFiles:r.convertFileReferences(n.importedFiles),ambientExternalModules:n.ambientExternalModules,isLibFile:n.isLibFile,typeReferenceDirectives:r.convertFileReferences(n.typeReferenceDirectives),libReferenceDirectives:r.convertFileReferences(n.libReferenceDirectives)}}))},t.prototype.getAutomaticTypeDirectiveNames=function(n){var t=this;return this.forwardJSONCall("getAutomaticTypeDirectiveNames('"+n+"')",(function(){var r=JSON.parse(n);return e.getAutomaticTypeDirectiveNames(r,t.host)}))},t.prototype.convertFileReferences=function(n){if(n){for(var t=[],r=0,a=n;r<a.length;r++){var i=a[r];t.push({path:e.normalizeSlashes(i.fileName),position:i.pos,length:i.end-i.pos})}return t}},t.prototype.getTSConfigFileInfo=function(n,t){var r=this;return this.forwardJSONCall("getTSConfigFileInfo('"+n+"')",(function(){var a=e.parseJsonText(n,e.getSnapshotText(t)),i=e.normalizeSlashes(n),o=e.parseJsonSourceFileConfigFileContent(a,r.host,e.getDirectoryPath(i),{},i);return{options:o.options,typeAcquisition:o.typeAcquisition,files:o.fileNames,raw:o.raw,errors:c(__spreadArrays(a.parseDiagnostics,o.errors),"\r\n")}}))},t.prototype.getDefaultCompilationSettings=function(){return this.forwardJSONCall("getDefaultCompilationSettings()",(function(){return e.getDefaultCompilerOptions()}))},t.prototype.discoverTypings=function(n){var t=this,r=e.createGetCanonicalFileName(!1);return this.forwardJSONCall("discoverTypings()",(function(){var a=JSON.parse(n);return void 0===t.safeList&&(t.safeList=e.JsTyping.loadSafeList(t.host,e.toPath(a.safeListPath,a.safeListPath,r))),e.JsTyping.discoverTypings(t.host,(function(e){return t.logger.log(e)}),a.fileNames,e.toPath(a.projectRootPath,a.projectRootPath,r),t.safeList,a.packageNameToTypingLocation,a.typeAcquisition,a.unresolvedImports,a.typesRegistry)}))},t}(s),m=function(){function t(){this._shims=[]}return t.prototype.getServicesVersion=function(){return e.servicesVersion},t.prototype.createLanguageServiceShim=function(t){try{void 0===this.documentRegistry&&(this.documentRegistry=e.createDocumentRegistry(t.useCaseSensitiveFileNames&&t.useCaseSensitiveFileNames(),t.getCurrentDirectory()));var a=new r(t),i=e.createLanguageService(a,this.documentRegistry,!1);return new l(this,t,i)}catch(e){throw n(t,e),e}},t.prototype.createClassifierShim=function(e){try{return new d(this,e)}catch(t){throw n(e,t),t}},t.prototype.createCoreServicesShim=function(e){try{var t=new a(e);return new p(this,e,t)}catch(t){throw n(e,t),t}},t.prototype.close=function(){e.clear(this._shims),this.documentRegistry=void 0},t.prototype.registerShim=function(e){this._shims.push(e)},t.prototype.unregisterShim=function(e){for(var n=0;n<this._shims.length;n++)if(this._shims[n]===e)return void delete this._shims[n];throw new Error("Invalid operation")},t}();e.TypeScriptServicesFactory=m}(ts||(ts={})),function(){if("object"!=typeof globalThis)try{Object.defineProperty(Object.prototype,"__magic__",{get:function(){return this},configurable:!0}),__magic__.globalThis=__magic__,"undefined"==typeof globalThis&&(window.globalThis=window),delete Object.prototype.__magic__}catch(e){window.globalThis=window}}(),("undefined"==typeof process||process.browser)&&(globalThis.TypeScript=globalThis.TypeScript||{},globalThis.TypeScript.Services=globalThis.TypeScript.Services||{},globalThis.TypeScript.Services.TypeScriptServicesFactory=ts.TypeScriptServicesFactory,globalThis.toolsVersion=ts.versionMajorMinor),"undefined"!=typeof module&&module.exports&&(module.exports=ts),function(e){var n={since:"4.0",warnAfter:"4.1",message:"Use the appropriate method on 'ts.factory' or the 'factory' supplied by your transformation context instead."};e.createNodeArray=e.Debug.deprecate(e.factory.createNodeArray,n),e.createNumericLiteral=e.Debug.deprecate(e.factory.createNumericLiteral,n),e.createBigIntLiteral=e.Debug.deprecate(e.factory.createBigIntLiteral,n),e.createStringLiteral=e.Debug.deprecate(e.factory.createStringLiteral,n),e.createStringLiteralFromNode=e.Debug.deprecate(e.factory.createStringLiteralFromNode,n),e.createRegularExpressionLiteral=e.Debug.deprecate(e.factory.createRegularExpressionLiteral,n),e.createLoopVariable=e.Debug.deprecate(e.factory.createLoopVariable,n),e.createUniqueName=e.Debug.deprecate(e.factory.createUniqueName,n),e.createPrivateIdentifier=e.Debug.deprecate(e.factory.createPrivateIdentifier,n),e.createSuper=e.Debug.deprecate(e.factory.createSuper,n),e.createThis=e.Debug.deprecate(e.factory.createThis,n),e.createNull=e.Debug.deprecate(e.factory.createNull,n),e.createTrue=e.Debug.deprecate(e.factory.createTrue,n),e.createFalse=e.Debug.deprecate(e.factory.createFalse,n),e.createModifier=e.Debug.deprecate(e.factory.createModifier,n),e.createModifiersFromModifierFlags=e.Debug.deprecate(e.factory.createModifiersFromModifierFlags,n),e.createQualifiedName=e.Debug.deprecate(e.factory.createQualifiedName,n),e.updateQualifiedName=e.Debug.deprecate(e.factory.updateQualifiedName,n),e.createComputedPropertyName=e.Debug.deprecate(e.factory.createComputedPropertyName,n),e.updateComputedPropertyName=e.Debug.deprecate(e.factory.updateComputedPropertyName,n),e.createTypeParameterDeclaration=e.Debug.deprecate(e.factory.createTypeParameterDeclaration,n),e.updateTypeParameterDeclaration=e.Debug.deprecate(e.factory.updateTypeParameterDeclaration,n),e.createParameter=e.Debug.deprecate(e.factory.createParameterDeclaration,n),e.updateParameter=e.Debug.deprecate(e.factory.updateParameterDeclaration,n),e.createDecorator=e.Debug.deprecate(e.factory.createDecorator,n),e.updateDecorator=e.Debug.deprecate(e.factory.updateDecorator,n),e.createProperty=e.Debug.deprecate(e.factory.createPropertyDeclaration,n),e.updateProperty=e.Debug.deprecate(e.factory.updatePropertyDeclaration,n),e.createMethod=e.Debug.deprecate(e.factory.createMethodDeclaration,n),e.updateMethod=e.Debug.deprecate(e.factory.updateMethodDeclaration,n),e.createConstructor=e.Debug.deprecate(e.factory.createConstructorDeclaration,n),e.updateConstructor=e.Debug.deprecate(e.factory.updateConstructorDeclaration,n),e.createGetAccessor=e.Debug.deprecate(e.factory.createGetAccessorDeclaration,n),e.updateGetAccessor=e.Debug.deprecate(e.factory.updateGetAccessorDeclaration,n),e.createSetAccessor=e.Debug.deprecate(e.factory.createSetAccessorDeclaration,n),e.updateSetAccessor=e.Debug.deprecate(e.factory.updateSetAccessorDeclaration,n),e.createCallSignature=e.Debug.deprecate(e.factory.createCallSignature,n),e.updateCallSignature=e.Debug.deprecate(e.factory.updateCallSignature,n),e.createConstructSignature=e.Debug.deprecate(e.factory.createConstructSignature,n),e.updateConstructSignature=e.Debug.deprecate(e.factory.updateConstructSignature,n),e.updateIndexSignature=e.Debug.deprecate(e.factory.updateIndexSignature,n),e.createKeywordTypeNode=e.Debug.deprecate(e.factory.createKeywordTypeNode,n),e.createTypePredicateNodeWithModifier=e.Debug.deprecate(e.factory.createTypePredicateNode,n),e.updateTypePredicateNodeWithModifier=e.Debug.deprecate(e.factory.updateTypePredicateNode,n),e.createTypeReferenceNode=e.Debug.deprecate(e.factory.createTypeReferenceNode,n),e.updateTypeReferenceNode=e.Debug.deprecate(e.factory.updateTypeReferenceNode,n),e.createFunctionTypeNode=e.Debug.deprecate(e.factory.createFunctionTypeNode,n),e.updateFunctionTypeNode=e.Debug.deprecate(e.factory.updateFunctionTypeNode,n),e.createConstructorTypeNode=e.Debug.deprecate(e.factory.createConstructorTypeNode,n),e.updateConstructorTypeNode=e.Debug.deprecate(e.factory.updateConstructorTypeNode,n),e.createTypeQueryNode=e.Debug.deprecate(e.factory.createTypeQueryNode,n),e.updateTypeQueryNode=e.Debug.deprecate(e.factory.updateTypeQueryNode,n),e.createTypeLiteralNode=e.Debug.deprecate(e.factory.createTypeLiteralNode,n),e.updateTypeLiteralNode=e.Debug.deprecate(e.factory.updateTypeLiteralNode,n),e.createArrayTypeNode=e.Debug.deprecate(e.factory.createArrayTypeNode,n),e.updateArrayTypeNode=e.Debug.deprecate(e.factory.updateArrayTypeNode,n),e.createTupleTypeNode=e.Debug.deprecate(e.factory.createTupleTypeNode,n),e.updateTupleTypeNode=e.Debug.deprecate(e.factory.updateTupleTypeNode,n),e.createOptionalTypeNode=e.Debug.deprecate(e.factory.createOptionalTypeNode,n),e.updateOptionalTypeNode=e.Debug.deprecate(e.factory.updateOptionalTypeNode,n),e.createRestTypeNode=e.Debug.deprecate(e.factory.createRestTypeNode,n),e.updateRestTypeNode=e.Debug.deprecate(e.factory.updateRestTypeNode,n),e.createUnionTypeNode=e.Debug.deprecate(e.factory.createUnionTypeNode,n),e.updateUnionTypeNode=e.Debug.deprecate(e.factory.updateUnionTypeNode,n),e.createIntersectionTypeNode=e.Debug.deprecate(e.factory.createIntersectionTypeNode,n),e.updateIntersectionTypeNode=e.Debug.deprecate(e.factory.updateIntersectionTypeNode,n),e.createConditionalTypeNode=e.Debug.deprecate(e.factory.createConditionalTypeNode,n),e.updateConditionalTypeNode=e.Debug.deprecate(e.factory.updateConditionalTypeNode,n),e.createInferTypeNode=e.Debug.deprecate(e.factory.createInferTypeNode,n),e.updateInferTypeNode=e.Debug.deprecate(e.factory.updateInferTypeNode,n),e.createImportTypeNode=e.Debug.deprecate(e.factory.createImportTypeNode,n),e.updateImportTypeNode=e.Debug.deprecate(e.factory.updateImportTypeNode,n),e.createParenthesizedType=e.Debug.deprecate(e.factory.createParenthesizedType,n),e.updateParenthesizedType=e.Debug.deprecate(e.factory.updateParenthesizedType,n),e.createThisTypeNode=e.Debug.deprecate(e.factory.createThisTypeNode,n),e.updateTypeOperatorNode=e.Debug.deprecate(e.factory.updateTypeOperatorNode,n),e.createIndexedAccessTypeNode=e.Debug.deprecate(e.factory.createIndexedAccessTypeNode,n),e.updateIndexedAccessTypeNode=e.Debug.deprecate(e.factory.updateIndexedAccessTypeNode,n),e.createMappedTypeNode=e.Debug.deprecate(e.factory.createMappedTypeNode,n),e.updateMappedTypeNode=e.Debug.deprecate(e.factory.updateMappedTypeNode,n),e.createLiteralTypeNode=e.Debug.deprecate(e.factory.createLiteralTypeNode,n),e.updateLiteralTypeNode=e.Debug.deprecate(e.factory.updateLiteralTypeNode,n),e.createObjectBindingPattern=e.Debug.deprecate(e.factory.createObjectBindingPattern,n),e.updateObjectBindingPattern=e.Debug.deprecate(e.factory.updateObjectBindingPattern,n),e.createArrayBindingPattern=e.Debug.deprecate(e.factory.createArrayBindingPattern,n),e.updateArrayBindingPattern=e.Debug.deprecate(e.factory.updateArrayBindingPattern,n),e.createBindingElement=e.Debug.deprecate(e.factory.createBindingElement,n),e.updateBindingElement=e.Debug.deprecate(e.factory.updateBindingElement,n),e.createArrayLiteral=e.Debug.deprecate(e.factory.createArrayLiteralExpression,n),e.updateArrayLiteral=e.Debug.deprecate(e.factory.updateArrayLiteralExpression,n),e.createObjectLiteral=e.Debug.deprecate(e.factory.createObjectLiteralExpression,n),e.updateObjectLiteral=e.Debug.deprecate(e.factory.updateObjectLiteralExpression,n),e.createPropertyAccess=e.Debug.deprecate(e.factory.createPropertyAccessExpression,n),e.updatePropertyAccess=e.Debug.deprecate(e.factory.updatePropertyAccessExpression,n),e.createPropertyAccessChain=e.Debug.deprecate(e.factory.createPropertyAccessChain,n),e.updatePropertyAccessChain=e.Debug.deprecate(e.factory.updatePropertyAccessChain,n),e.createElementAccess=e.Debug.deprecate(e.factory.createElementAccessExpression,n),e.updateElementAccess=e.Debug.deprecate(e.factory.updateElementAccessExpression,n),e.createElementAccessChain=e.Debug.deprecate(e.factory.createElementAccessChain,n),e.updateElementAccessChain=e.Debug.deprecate(e.factory.updateElementAccessChain,n),e.createCall=e.Debug.deprecate(e.factory.createCallExpression,n),e.updateCall=e.Debug.deprecate(e.factory.updateCallExpression,n),e.createCallChain=e.Debug.deprecate(e.factory.createCallChain,n),e.updateCallChain=e.Debug.deprecate(e.factory.updateCallChain,n),e.createNew=e.Debug.deprecate(e.factory.createNewExpression,n),e.updateNew=e.Debug.deprecate(e.factory.updateNewExpression,n),e.createTypeAssertion=e.Debug.deprecate(e.factory.createTypeAssertion,n),e.updateTypeAssertion=e.Debug.deprecate(e.factory.updateTypeAssertion,n),e.createParen=e.Debug.deprecate(e.factory.createParenthesizedExpression,n),e.updateParen=e.Debug.deprecate(e.factory.updateParenthesizedExpression,n),e.createFunctionExpression=e.Debug.deprecate(e.factory.createFunctionExpression,n),e.updateFunctionExpression=e.Debug.deprecate(e.factory.updateFunctionExpression,n),e.createDelete=e.Debug.deprecate(e.factory.createDeleteExpression,n),e.updateDelete=e.Debug.deprecate(e.factory.updateDeleteExpression,n),e.createTypeOf=e.Debug.deprecate(e.factory.createTypeOfExpression,n),e.updateTypeOf=e.Debug.deprecate(e.factory.updateTypeOfExpression,n),e.createVoid=e.Debug.deprecate(e.factory.createVoidExpression,n),e.updateVoid=e.Debug.deprecate(e.factory.updateVoidExpression,n),e.createAwait=e.Debug.deprecate(e.factory.createAwaitExpression,n),e.updateAwait=e.Debug.deprecate(e.factory.updateAwaitExpression,n),e.createPrefix=e.Debug.deprecate(e.factory.createPrefixUnaryExpression,n),e.updatePrefix=e.Debug.deprecate(e.factory.updatePrefixUnaryExpression,n),e.createPostfix=e.Debug.deprecate(e.factory.createPostfixUnaryExpression,n),e.updatePostfix=e.Debug.deprecate(e.factory.updatePostfixUnaryExpression,n),e.createBinary=e.Debug.deprecate(e.factory.createBinaryExpression,n),e.updateConditional=e.Debug.deprecate(e.factory.updateConditionalExpression,n),e.createTemplateExpression=e.Debug.deprecate(e.factory.createTemplateExpression,n),e.updateTemplateExpression=e.Debug.deprecate(e.factory.updateTemplateExpression,n),e.createTemplateHead=e.Debug.deprecate(e.factory.createTemplateHead,n),e.createTemplateMiddle=e.Debug.deprecate(e.factory.createTemplateMiddle,n),e.createTemplateTail=e.Debug.deprecate(e.factory.createTemplateTail,n),e.createNoSubstitutionTemplateLiteral=e.Debug.deprecate(e.factory.createNoSubstitutionTemplateLiteral,n),e.updateYield=e.Debug.deprecate(e.factory.updateYieldExpression,n),e.createSpread=e.Debug.deprecate(e.factory.createSpreadElement,n),e.updateSpread=e.Debug.deprecate(e.factory.updateSpreadElement,n),e.createOmittedExpression=e.Debug.deprecate(e.factory.createOmittedExpression,n),e.createAsExpression=e.Debug.deprecate(e.factory.createAsExpression,n),e.updateAsExpression=e.Debug.deprecate(e.factory.updateAsExpression,n),e.createNonNullExpression=e.Debug.deprecate(e.factory.createNonNullExpression,n),e.updateNonNullExpression=e.Debug.deprecate(e.factory.updateNonNullExpression,n),e.createNonNullChain=e.Debug.deprecate(e.factory.createNonNullChain,n),e.updateNonNullChain=e.Debug.deprecate(e.factory.updateNonNullChain,n),e.createMetaProperty=e.Debug.deprecate(e.factory.createMetaProperty,n),e.updateMetaProperty=e.Debug.deprecate(e.factory.updateMetaProperty,n),e.createTemplateSpan=e.Debug.deprecate(e.factory.createTemplateSpan,n),e.updateTemplateSpan=e.Debug.deprecate(e.factory.updateTemplateSpan,n),e.createSemicolonClassElement=e.Debug.deprecate(e.factory.createSemicolonClassElement,n),e.createBlock=e.Debug.deprecate(e.factory.createBlock,n),e.updateBlock=e.Debug.deprecate(e.factory.updateBlock,n),e.createVariableStatement=e.Debug.deprecate(e.factory.createVariableStatement,n),e.updateVariableStatement=e.Debug.deprecate(e.factory.updateVariableStatement,n),e.createEmptyStatement=e.Debug.deprecate(e.factory.createEmptyStatement,n),e.createExpressionStatement=e.Debug.deprecate(e.factory.createExpressionStatement,n),e.updateExpressionStatement=e.Debug.deprecate(e.factory.updateExpressionStatement,n),e.createStatement=e.Debug.deprecate(e.factory.createExpressionStatement,n),e.updateStatement=e.Debug.deprecate(e.factory.updateExpressionStatement,n),e.createIf=e.Debug.deprecate(e.factory.createIfStatement,n),e.updateIf=e.Debug.deprecate(e.factory.updateIfStatement,n),e.createDo=e.Debug.deprecate(e.factory.createDoStatement,n),e.updateDo=e.Debug.deprecate(e.factory.updateDoStatement,n),e.createWhile=e.Debug.deprecate(e.factory.createWhileStatement,n),e.updateWhile=e.Debug.deprecate(e.factory.updateWhileStatement,n),e.createFor=e.Debug.deprecate(e.factory.createForStatement,n),e.updateFor=e.Debug.deprecate(e.factory.updateForStatement,n),e.createForIn=e.Debug.deprecate(e.factory.createForInStatement,n),e.updateForIn=e.Debug.deprecate(e.factory.updateForInStatement,n),e.createForOf=e.Debug.deprecate(e.factory.createForOfStatement,n),e.updateForOf=e.Debug.deprecate(e.factory.updateForOfStatement,n),e.createContinue=e.Debug.deprecate(e.factory.createContinueStatement,n),e.updateContinue=e.Debug.deprecate(e.factory.updateContinueStatement,n),e.createBreak=e.Debug.deprecate(e.factory.createBreakStatement,n),e.updateBreak=e.Debug.deprecate(e.factory.updateBreakStatement,n),e.createReturn=e.Debug.deprecate(e.factory.createReturnStatement,n),e.updateReturn=e.Debug.deprecate(e.factory.updateReturnStatement,n),e.createWith=e.Debug.deprecate(e.factory.createWithStatement,n),e.updateWith=e.Debug.deprecate(e.factory.updateWithStatement,n),e.createSwitch=e.Debug.deprecate(e.factory.createSwitchStatement,n),e.updateSwitch=e.Debug.deprecate(e.factory.updateSwitchStatement,n),e.createLabel=e.Debug.deprecate(e.factory.createLabeledStatement,n),e.updateLabel=e.Debug.deprecate(e.factory.updateLabeledStatement,n),e.createThrow=e.Debug.deprecate(e.factory.createThrowStatement,n),e.updateThrow=e.Debug.deprecate(e.factory.updateThrowStatement,n),e.createTry=e.Debug.deprecate(e.factory.createTryStatement,n),e.updateTry=e.Debug.deprecate(e.factory.updateTryStatement,n),e.createDebuggerStatement=e.Debug.deprecate(e.factory.createDebuggerStatement,n),e.createVariableDeclarationList=e.Debug.deprecate(e.factory.createVariableDeclarationList,n),e.updateVariableDeclarationList=e.Debug.deprecate(e.factory.updateVariableDeclarationList,n),e.createFunctionDeclaration=e.Debug.deprecate(e.factory.createFunctionDeclaration,n),e.updateFunctionDeclaration=e.Debug.deprecate(e.factory.updateFunctionDeclaration,n),e.createClassDeclaration=e.Debug.deprecate(e.factory.createClassDeclaration,n),e.updateClassDeclaration=e.Debug.deprecate(e.factory.updateClassDeclaration,n),e.createInterfaceDeclaration=e.Debug.deprecate(e.factory.createInterfaceDeclaration,n),e.updateInterfaceDeclaration=e.Debug.deprecate(e.factory.updateInterfaceDeclaration,n),e.createTypeAliasDeclaration=e.Debug.deprecate(e.factory.createTypeAliasDeclaration,n),e.updateTypeAliasDeclaration=e.Debug.deprecate(e.factory.updateTypeAliasDeclaration,n),e.createEnumDeclaration=e.Debug.deprecate(e.factory.createEnumDeclaration,n),e.updateEnumDeclaration=e.Debug.deprecate(e.factory.updateEnumDeclaration,n),e.createModuleDeclaration=e.Debug.deprecate(e.factory.createModuleDeclaration,n),e.updateModuleDeclaration=e.Debug.deprecate(e.factory.updateModuleDeclaration,n),e.createModuleBlock=e.Debug.deprecate(e.factory.createModuleBlock,n),e.updateModuleBlock=e.Debug.deprecate(e.factory.updateModuleBlock,n),e.createCaseBlock=e.Debug.deprecate(e.factory.createCaseBlock,n),e.updateCaseBlock=e.Debug.deprecate(e.factory.updateCaseBlock,n),e.createNamespaceExportDeclaration=e.Debug.deprecate(e.factory.createNamespaceExportDeclaration,n),e.updateNamespaceExportDeclaration=e.Debug.deprecate(e.factory.updateNamespaceExportDeclaration,n),e.createImportEqualsDeclaration=e.Debug.deprecate(e.factory.createImportEqualsDeclaration,n),e.updateImportEqualsDeclaration=e.Debug.deprecate(e.factory.updateImportEqualsDeclaration,n),e.createImportDeclaration=e.Debug.deprecate(e.factory.createImportDeclaration,n),e.updateImportDeclaration=e.Debug.deprecate(e.factory.updateImportDeclaration,n),e.createNamespaceImport=e.Debug.deprecate(e.factory.createNamespaceImport,n),e.updateNamespaceImport=e.Debug.deprecate(e.factory.updateNamespaceImport,n),e.createNamedImports=e.Debug.deprecate(e.factory.createNamedImports,n),e.updateNamedImports=e.Debug.deprecate(e.factory.updateNamedImports,n),e.createImportSpecifier=e.Debug.deprecate(e.factory.createImportSpecifier,n),e.updateImportSpecifier=e.Debug.deprecate(e.factory.updateImportSpecifier,n),e.createExportAssignment=e.Debug.deprecate(e.factory.createExportAssignment,n),e.updateExportAssignment=e.Debug.deprecate(e.factory.updateExportAssignment,n),e.createNamedExports=e.Debug.deprecate(e.factory.createNamedExports,n),e.updateNamedExports=e.Debug.deprecate(e.factory.updateNamedExports,n),e.createExportSpecifier=e.Debug.deprecate(e.factory.createExportSpecifier,n),e.updateExportSpecifier=e.Debug.deprecate(e.factory.updateExportSpecifier,n),e.createExternalModuleReference=e.Debug.deprecate(e.factory.createExternalModuleReference,n),e.updateExternalModuleReference=e.Debug.deprecate(e.factory.updateExternalModuleReference,n),e.createJSDocTypeExpression=e.Debug.deprecate(e.factory.createJSDocTypeExpression,n),e.createJSDocTypeTag=e.Debug.deprecate(e.factory.createJSDocTypeTag,n),e.createJSDocReturnTag=e.Debug.deprecate(e.factory.createJSDocReturnTag,n),e.createJSDocThisTag=e.Debug.deprecate(e.factory.createJSDocThisTag,n),e.createJSDocComment=e.Debug.deprecate(e.factory.createJSDocComment,n),e.createJSDocParameterTag=e.Debug.deprecate(e.factory.createJSDocParameterTag,n),e.createJSDocClassTag=e.Debug.deprecate(e.factory.createJSDocClassTag,n),e.createJSDocAugmentsTag=e.Debug.deprecate(e.factory.createJSDocAugmentsTag,n),e.createJSDocEnumTag=e.Debug.deprecate(e.factory.createJSDocEnumTag,n),e.createJSDocTemplateTag=e.Debug.deprecate(e.factory.createJSDocTemplateTag,n),e.createJSDocTypedefTag=e.Debug.deprecate(e.factory.createJSDocTypedefTag,n),e.createJSDocCallbackTag=e.Debug.deprecate(e.factory.createJSDocCallbackTag,n),e.createJSDocSignature=e.Debug.deprecate(e.factory.createJSDocSignature,n),e.createJSDocPropertyTag=e.Debug.deprecate(e.factory.createJSDocPropertyTag,n),e.createJSDocTypeLiteral=e.Debug.deprecate(e.factory.createJSDocTypeLiteral,n),e.createJSDocImplementsTag=e.Debug.deprecate(e.factory.createJSDocImplementsTag,n),e.createJSDocAuthorTag=e.Debug.deprecate(e.factory.createJSDocAuthorTag,n),e.createJSDocPublicTag=e.Debug.deprecate(e.factory.createJSDocPublicTag,n),e.createJSDocPrivateTag=e.Debug.deprecate(e.factory.createJSDocPrivateTag,n),e.createJSDocProtectedTag=e.Debug.deprecate(e.factory.createJSDocProtectedTag,n),e.createJSDocReadonlyTag=e.Debug.deprecate(e.factory.createJSDocReadonlyTag,n),e.createJSDocTag=e.Debug.deprecate(e.factory.createJSDocUnknownTag,n),e.createJsxElement=e.Debug.deprecate(e.factory.createJsxElement,n),e.updateJsxElement=e.Debug.deprecate(e.factory.updateJsxElement,n),e.createJsxSelfClosingElement=e.Debug.deprecate(e.factory.createJsxSelfClosingElement,n),e.updateJsxSelfClosingElement=e.Debug.deprecate(e.factory.updateJsxSelfClosingElement,n),e.createJsxOpeningElement=e.Debug.deprecate(e.factory.createJsxOpeningElement,n),e.updateJsxOpeningElement=e.Debug.deprecate(e.factory.updateJsxOpeningElement,n),e.createJsxClosingElement=e.Debug.deprecate(e.factory.createJsxClosingElement,n),e.updateJsxClosingElement=e.Debug.deprecate(e.factory.updateJsxClosingElement,n),e.createJsxFragment=e.Debug.deprecate(e.factory.createJsxFragment,n),e.createJsxText=e.Debug.deprecate(e.factory.createJsxText,n),e.updateJsxText=e.Debug.deprecate(e.factory.updateJsxText,n),e.createJsxOpeningFragment=e.Debug.deprecate(e.factory.createJsxOpeningFragment,n),e.createJsxJsxClosingFragment=e.Debug.deprecate(e.factory.createJsxJsxClosingFragment,n),e.updateJsxFragment=e.Debug.deprecate(e.factory.updateJsxFragment,n),e.createJsxAttribute=e.Debug.deprecate(e.factory.createJsxAttribute,n),e.updateJsxAttribute=e.Debug.deprecate(e.factory.updateJsxAttribute,n),e.createJsxAttributes=e.Debug.deprecate(e.factory.createJsxAttributes,n),e.updateJsxAttributes=e.Debug.deprecate(e.factory.updateJsxAttributes,n),e.createJsxSpreadAttribute=e.Debug.deprecate(e.factory.createJsxSpreadAttribute,n),e.updateJsxSpreadAttribute=e.Debug.deprecate(e.factory.updateJsxSpreadAttribute,n),e.createJsxExpression=e.Debug.deprecate(e.factory.createJsxExpression,n),e.updateJsxExpression=e.Debug.deprecate(e.factory.updateJsxExpression,n),e.createCaseClause=e.Debug.deprecate(e.factory.createCaseClause,n),e.updateCaseClause=e.Debug.deprecate(e.factory.updateCaseClause,n),e.createDefaultClause=e.Debug.deprecate(e.factory.createDefaultClause,n),e.updateDefaultClause=e.Debug.deprecate(e.factory.updateDefaultClause,n),e.createHeritageClause=e.Debug.deprecate(e.factory.createHeritageClause,n),e.updateHeritageClause=e.Debug.deprecate(e.factory.updateHeritageClause,n),e.createCatchClause=e.Debug.deprecate(e.factory.createCatchClause,n),e.updateCatchClause=e.Debug.deprecate(e.factory.updateCatchClause,n),e.createPropertyAssignment=e.Debug.deprecate(e.factory.createPropertyAssignment,n),e.updatePropertyAssignment=e.Debug.deprecate(e.factory.updatePropertyAssignment,n),e.createShorthandPropertyAssignment=e.Debug.deprecate(e.factory.createShorthandPropertyAssignment,n),e.updateShorthandPropertyAssignment=e.Debug.deprecate(e.factory.updateShorthandPropertyAssignment,n),e.createSpreadAssignment=e.Debug.deprecate(e.factory.createSpreadAssignment,n),e.updateSpreadAssignment=e.Debug.deprecate(e.factory.updateSpreadAssignment,n),e.createEnumMember=e.Debug.deprecate(e.factory.createEnumMember,n),e.updateEnumMember=e.Debug.deprecate(e.factory.updateEnumMember,n),e.updateSourceFileNode=e.Debug.deprecate(e.factory.updateSourceFile,n),e.createNotEmittedStatement=e.Debug.deprecate(e.factory.createNotEmittedStatement,n),e.createPartiallyEmittedExpression=e.Debug.deprecate(e.factory.createPartiallyEmittedExpression,n),e.updatePartiallyEmittedExpression=e.Debug.deprecate(e.factory.updatePartiallyEmittedExpression,n),e.createCommaList=e.Debug.deprecate(e.factory.createCommaListExpression,n),e.updateCommaList=e.Debug.deprecate(e.factory.updateCommaListExpression,n),e.createBundle=e.Debug.deprecate(e.factory.createBundle,n),e.updateBundle=e.Debug.deprecate(e.factory.updateBundle,n),e.createImmediatelyInvokedFunctionExpression=e.Debug.deprecate(e.factory.createImmediatelyInvokedFunctionExpression,n),e.createImmediatelyInvokedArrowFunction=e.Debug.deprecate(e.factory.createImmediatelyInvokedArrowFunction,n),e.createVoidZero=e.Debug.deprecate(e.factory.createVoidZero,n),e.createExportDefault=e.Debug.deprecate(e.factory.createExportDefault,n),e.createExternalModuleExport=e.Debug.deprecate(e.factory.createExternalModuleExport,n),e.createNamespaceExport=e.Debug.deprecate(e.factory.createNamespaceExport,n),e.updateNamespaceExport=e.Debug.deprecate(e.factory.updateNamespaceExport,n),e.createToken=e.Debug.deprecate((function(n){return e.factory.createToken(n)}),n),e.createIdentifier=e.Debug.deprecate((function(n){return e.factory.createIdentifier(n,void 0,void 0)}),n),e.createTempVariable=e.Debug.deprecate((function(n){return e.factory.createTempVariable(n,void 0)}),n),e.getGeneratedNameForNode=e.Debug.deprecate((function(n){return e.factory.getGeneratedNameForNode(n,void 0)}),n),e.createOptimisticUniqueName=e.Debug.deprecate((function(n){return e.factory.createUniqueName(n,16)}),n),e.createFileLevelUniqueName=e.Debug.deprecate((function(n){return e.factory.createUniqueName(n,48)}),n),e.createIndexSignature=e.Debug.deprecate((function(n,t,r,a){return e.factory.createIndexSignature(n,t,r,a)}),n),e.createTypePredicateNode=e.Debug.deprecate((function(n,t){return e.factory.createTypePredicateNode(void 0,n,t)}),n),e.updateTypePredicateNode=e.Debug.deprecate((function(n,t,r){return e.factory.updateTypePredicateNode(n,void 0,t,r)}),n),e.createLiteral=e.Debug.deprecate((function(n){return"number"==typeof n?e.factory.createNumericLiteral(n):"object"==typeof n&&"base10Value"in n?e.factory.createBigIntLiteral(n):"boolean"==typeof n?n?e.factory.createTrue():e.factory.createFalse():"string"==typeof n?e.factory.createStringLiteral(n,void 0):e.factory.createStringLiteralFromNode(n)}),{since:"4.0",warnAfter:"4.1",message:"Use `factory.createStringLiteral`, `factory.createStringLiteralFromNode`, `factory.createNumericLiteral`, `factory.createBigIntLiteral`, `factory.createTrue`, `factory.createFalse`, or the factory supplied by your transformation context instead."}),e.createMethodSignature=e.Debug.deprecate((function(n,t,r,a,i){return e.factory.createMethodSignature(void 0,a,i,n,t,r)}),n),e.updateMethodSignature=e.Debug.deprecate((function(n,t,r,a,i,o){return e.factory.updateMethodSignature(n,n.modifiers,i,o,t,r,a)}),n),e.createTypeOperatorNode=e.Debug.deprecate((function(n,t){var r;return t?r=n:(t=n,r=138),e.factory.createTypeOperatorNode(r,t)}),n),e.createTaggedTemplate=e.Debug.deprecate((function(n,t,r){var a;return r?a=t:r=t,e.factory.createTaggedTemplateExpression(n,a,r)}),n),e.updateTaggedTemplate=e.Debug.deprecate((function(n,t,r,a){var i;return a?i=r:a=r,e.factory.updateTaggedTemplateExpression(n,t,i,a)}),n),e.updateBinary=e.Debug.deprecate((function(n,t,r,a){return void 0===a&&(a=n.operatorToken),"number"==typeof a&&(a=a===n.operatorToken.kind?n.operatorToken:e.factory.createToken(a)),e.factory.updateBinaryExpression(n,t,a,r)}),n),e.createConditional=e.Debug.deprecate((function(n,t,r,a,i){return 5===arguments.length?e.factory.createConditionalExpression(n,t,r,a,i):3===arguments.length?e.factory.createConditionalExpression(n,e.factory.createToken(57),t,e.factory.createToken(58),r):e.Debug.fail("Argument count mismatch")}),n),e.createYield=e.Debug.deprecate((function(n,t){var r;return t?r=n:t=n,e.factory.createYieldExpression(r,t)}),n),e.createClassExpression=e.Debug.deprecate((function(n,t,r,a,i){return e.factory.createClassExpression(void 0,n,t,r,a,i)}),n),e.updateClassExpression=e.Debug.deprecate((function(n,t,r,a,i,o){return e.factory.updateClassExpression(n,void 0,t,r,a,i,o)}),n),e.createPropertySignature=e.Debug.deprecate((function(n,t,r,a,i){var o=e.factory.createPropertySignature(n,t,r,a);return o.initializer=i,o}),n),e.updatePropertySignature=e.Debug.deprecate((function(n,t,r,a,i,o){var s=e.factory.updatePropertySignature(n,t,r,a,i);return n.initializer!==o&&(s===n&&(s=e.factory.cloneNode(n)),s.initializer=o),s}),n),e.createExpressionWithTypeArguments=e.Debug.deprecate((function(n,t){return e.factory.createExpressionWithTypeArguments(t,n)}),n),e.updateExpressionWithTypeArguments=e.Debug.deprecate((function(n,t,r){return e.factory.updateExpressionWithTypeArguments(n,r,t)}),n),e.createArrowFunction=e.Debug.deprecate((function(n,t,r,a,i,o){return 6===arguments.length?e.factory.createArrowFunction(n,t,r,a,i,o):5===arguments.length?e.factory.createArrowFunction(n,t,r,a,void 0,i):e.Debug.fail("Argument count mismatch")}),n),e.updateArrowFunction=e.Debug.deprecate((function(n,t,r,a,i,o,s){return 7===arguments.length?e.factory.updateArrowFunction(n,t,r,a,i,o,s):6===arguments.length?e.factory.updateArrowFunction(n,t,r,a,i,n.equalsGreaterThanToken,o):e.Debug.fail("Argument count mismatch")}),n),e.createVariableDeclaration=e.Debug.deprecate((function(n,t,r,a){return 4===arguments.length?e.factory.createVariableDeclaration(n,t,r,a):arguments.length>=1&&arguments.length<=3?e.factory.createVariableDeclaration(n,void 0,t,r):e.Debug.fail("Argument count mismatch")}),n),e.updateVariableDeclaration=e.Debug.deprecate((function(n,t,r,a,i){return 5===arguments.length?e.factory.updateVariableDeclaration(n,t,r,a,i):4===arguments.length?e.factory.updateVariableDeclaration(n,t,n.exclamationToken,r,a):e.Debug.fail("Argument count mismatch")}),n),e.createImportClause=e.Debug.deprecate((function(n,t,r){return void 0===r&&(r=!1),e.factory.createImportClause(r,n,t)}),n),e.updateImportClause=e.Debug.deprecate((function(n,t,r,a){return e.factory.updateImportClause(n,a,t,r)}),n),e.createExportDeclaration=e.Debug.deprecate((function(n,t,r,a,i){return void 0===i&&(i=!1),e.factory.createExportDeclaration(n,t,i,r,a)}),n),e.updateExportDeclaration=e.Debug.deprecate((function(n,t,r,a,i,o){return e.factory.updateExportDeclaration(n,t,r,o,a,i)}),n),e.createJSDocParamTag=e.Debug.deprecate((function(n,t,r,a){return e.factory.createJSDocParameterTag(void 0,n,t,r,!1,a)}),n),e.createComma=e.Debug.deprecate((function(n,t){return e.factory.createComma(n,t)}),n),e.createLessThan=e.Debug.deprecate((function(n,t){return e.factory.createLessThan(n,t)}),n),e.createAssignment=e.Debug.deprecate((function(n,t){return e.factory.createAssignment(n,t)}),n),e.createStrictEquality=e.Debug.deprecate((function(n,t){return e.factory.createStrictEquality(n,t)}),n),e.createStrictInequality=e.Debug.deprecate((function(n,t){return e.factory.createStrictInequality(n,t)}),n),e.createAdd=e.Debug.deprecate((function(n,t){return e.factory.createAdd(n,t)}),n),e.createSubtract=e.Debug.deprecate((function(n,t){return e.factory.createSubtract(n,t)}),n),e.createLogicalAnd=e.Debug.deprecate((function(n,t){return e.factory.createLogicalAnd(n,t)}),n),e.createLogicalOr=e.Debug.deprecate((function(n,t){return e.factory.createLogicalOr(n,t)}),n),e.createPostfixIncrement=e.Debug.deprecate((function(n){return e.factory.createPostfixIncrement(n)}),n),e.createLogicalNot=e.Debug.deprecate((function(n){return e.factory.createLogicalNot(n)}),n),e.createNode=e.Debug.deprecate((function(n,t,r){return void 0===t&&(t=0),void 0===r&&(r=0),e.setTextRangePosEnd(297===n?e.parseBaseNodeFactory.createBaseSourceFileNode(n):78===n?e.parseBaseNodeFactory.createBaseIdentifierNode(n):79===n?e.parseBaseNodeFactory.createBasePrivateIdentifierNode(n):e.isNodeKind(n)?e.parseBaseNodeFactory.createBaseNode(n):e.parseBaseNodeFactory.createBaseTokenNode(n),t,r)}),{since:"4.0",warnAfter:"4.1",message:"Use an appropriate `factory` method instead."}),e.getMutableClone=e.Debug.deprecate((function(n){var t=e.factory.cloneNode(n);return e.setTextRange(t,n),e.setParent(t,n.parent),t}),{since:"4.0",warnAfter:"4.1",message:"Use `factory.cloneNode` instead and use `setCommentRange` or `setSourceMapRange` and avoid setting `parent`."}),e.isTypeAssertion=e.Debug.deprecate((function(e){return 206===e.kind}),{since:"4.0",warnAfter:"4.1",message:"Use `isTypeAssertionExpression` instead."})}(ts||(ts={})),define("vs/language/typescript/lib/typescriptServices",[],(function(){return ts})),define("vs/language/typescript/lib/lib",["require","exports"],(function(e,n){"use strict";Object.defineProperty(n,"__esModule",{value:!0}),n.libFileMap=void 0,n.libFileMap={},n.libFileMap["lib.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es5" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n',n.libFileMap["lib.dom.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// DOM APIs\n/////////////////////////////\n\ninterface Account {\n    displayName: string;\n    id: string;\n    imageURL?: string;\n    name?: string;\n    rpDisplayName: string;\n}\n\ninterface AddEventListenerOptions extends EventListenerOptions {\n    once?: boolean;\n    passive?: boolean;\n}\n\ninterface AesCbcParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCtrParams extends Algorithm {\n    counter: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    length: number;\n}\n\ninterface AesDerivedKeyParams extends Algorithm {\n    length: number;\n}\n\ninterface AesGcmParams extends Algorithm {\n    additionalData?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    tagLength?: number;\n}\n\ninterface AesKeyAlgorithm extends KeyAlgorithm {\n    length: number;\n}\n\ninterface AesKeyGenParams extends Algorithm {\n    length: number;\n}\n\ninterface Algorithm {\n    name: string;\n}\n\ninterface AnalyserOptions extends AudioNodeOptions {\n    fftSize?: number;\n    maxDecibels?: number;\n    minDecibels?: number;\n    smoothingTimeConstant?: number;\n}\n\ninterface AnimationEventInit extends EventInit {\n    animationName?: string;\n    elapsedTime?: number;\n    pseudoElement?: string;\n}\n\ninterface AnimationPlaybackEventInit extends EventInit {\n    currentTime?: number | null;\n    timelineTime?: number | null;\n}\n\ninterface AssertionOptions {\n    allowList?: ScopedCredentialDescriptor[];\n    extensions?: WebAuthnExtensions;\n    rpId?: string;\n    timeoutSeconds?: number;\n}\n\ninterface AssignedNodesOptions {\n    flatten?: boolean;\n}\n\ninterface AudioBufferOptions {\n    length: number;\n    numberOfChannels?: number;\n    sampleRate: number;\n}\n\ninterface AudioBufferSourceOptions {\n    buffer?: AudioBuffer | null;\n    detune?: number;\n    loop?: boolean;\n    loopEnd?: number;\n    loopStart?: number;\n    playbackRate?: number;\n}\n\ninterface AudioContextInfo {\n    currentTime?: number;\n    sampleRate?: number;\n}\n\ninterface AudioContextOptions {\n    latencyHint?: AudioContextLatencyCategory | number;\n    sampleRate?: number;\n}\n\ninterface AudioNodeOptions {\n    channelCount?: number;\n    channelCountMode?: ChannelCountMode;\n    channelInterpretation?: ChannelInterpretation;\n}\n\ninterface AudioParamDescriptor {\n    automationRate?: AutomationRate;\n    defaultValue?: number;\n    maxValue?: number;\n    minValue?: number;\n    name: string;\n}\n\ninterface AudioProcessingEventInit extends EventInit {\n    inputBuffer: AudioBuffer;\n    outputBuffer: AudioBuffer;\n    playbackTime: number;\n}\n\ninterface AudioTimestamp {\n    contextTime?: number;\n    performanceTime?: number;\n}\n\ninterface AudioWorkletNodeOptions extends AudioNodeOptions {\n    numberOfInputs?: number;\n    numberOfOutputs?: number;\n    outputChannelCount?: number[];\n    parameterData?: Record<string, number>;\n    processorOptions?: any;\n}\n\ninterface AuthenticationExtensionsClientInputs {\n    appid?: string;\n    appidExclude?: string;\n    credProps?: boolean;\n    uvm?: boolean;\n}\n\ninterface AuthenticationExtensionsClientOutputs {\n    appid?: boolean;\n    credProps?: CredentialPropertiesOutput;\n    uvm?: UvmEntries;\n}\n\ninterface AuthenticatorSelectionCriteria {\n    authenticatorAttachment?: AuthenticatorAttachment;\n    requireResidentKey?: boolean;\n    residentKey?: ResidentKeyRequirement;\n    userVerification?: UserVerificationRequirement;\n}\n\ninterface BiquadFilterOptions extends AudioNodeOptions {\n    Q?: number;\n    detune?: number;\n    frequency?: number;\n    gain?: number;\n    type?: BiquadFilterType;\n}\n\ninterface BlobPropertyBag {\n    endings?: EndingType;\n    type?: string;\n}\n\ninterface ByteLengthChunk {\n    byteLength?: number;\n}\n\ninterface CacheQueryOptions {\n    ignoreMethod?: boolean;\n    ignoreSearch?: boolean;\n    ignoreVary?: boolean;\n}\n\ninterface CanvasRenderingContext2DSettings {\n    alpha?: boolean;\n    desynchronized?: boolean;\n}\n\ninterface ChannelMergerOptions extends AudioNodeOptions {\n    numberOfInputs?: number;\n}\n\ninterface ChannelSplitterOptions extends AudioNodeOptions {\n    numberOfOutputs?: number;\n}\n\ninterface ClientData {\n    challenge: string;\n    extensions?: WebAuthnExtensions;\n    hashAlg: string | Algorithm;\n    origin: string;\n    rpId: string;\n    tokenBinding?: string;\n}\n\ninterface ClientQueryOptions {\n    includeUncontrolled?: boolean;\n    type?: ClientTypes;\n}\n\ninterface ClipboardEventInit extends EventInit {\n    clipboardData?: DataTransfer | null;\n}\n\ninterface CloseEventInit extends EventInit {\n    code?: number;\n    reason?: string;\n    wasClean?: boolean;\n}\n\ninterface CompositionEventInit extends UIEventInit {\n    data?: string;\n}\n\ninterface ComputedEffectTiming extends EffectTiming {\n    activeDuration?: number;\n    currentIteration?: number | null;\n    endTime?: number;\n    localTime?: number | null;\n    progress?: number | null;\n}\n\ninterface ComputedKeyframe {\n    composite: CompositeOperationOrAuto;\n    computedOffset: number;\n    easing: string;\n    offset: number | null;\n    [property: string]: string | number | null | undefined;\n}\n\ninterface ConfirmSiteSpecificExceptionsInformation extends ExceptionInformation {\n    arrayOfDomainStrings?: string[];\n}\n\ninterface ConstantSourceOptions {\n    offset?: number;\n}\n\ninterface ConstrainBooleanParameters {\n    exact?: boolean;\n    ideal?: boolean;\n}\n\ninterface ConstrainDOMStringParameters {\n    exact?: string | string[];\n    ideal?: string | string[];\n}\n\ninterface ConstrainDoubleRange extends DoubleRange {\n    exact?: number;\n    ideal?: number;\n}\n\ninterface ConstrainULongRange extends ULongRange {\n    exact?: number;\n    ideal?: number;\n}\n\ninterface ConstrainVideoFacingModeParameters {\n    exact?: VideoFacingModeEnum | VideoFacingModeEnum[];\n    ideal?: VideoFacingModeEnum | VideoFacingModeEnum[];\n}\n\ninterface ConvolverOptions extends AudioNodeOptions {\n    buffer?: AudioBuffer | null;\n    disableNormalization?: boolean;\n}\n\ninterface CredentialCreationOptions {\n    publicKey?: PublicKeyCredentialCreationOptions;\n    signal?: AbortSignal;\n}\n\ninterface CredentialPropertiesOutput {\n    rk?: boolean;\n}\n\ninterface CredentialRequestOptions {\n    mediation?: CredentialMediationRequirement;\n    publicKey?: PublicKeyCredentialRequestOptions;\n    signal?: AbortSignal;\n}\n\ninterface CustomEventInit<T = any> extends EventInit {\n    detail?: T;\n}\n\ninterface DOMMatrix2DInit {\n    a?: number;\n    b?: number;\n    c?: number;\n    d?: number;\n    e?: number;\n    f?: number;\n    m11?: number;\n    m12?: number;\n    m21?: number;\n    m22?: number;\n    m41?: number;\n    m42?: number;\n}\n\ninterface DOMMatrixInit extends DOMMatrix2DInit {\n    is2D?: boolean;\n    m13?: number;\n    m14?: number;\n    m23?: number;\n    m24?: number;\n    m31?: number;\n    m32?: number;\n    m33?: number;\n    m34?: number;\n    m43?: number;\n    m44?: number;\n}\n\ninterface DOMPointInit {\n    w?: number;\n    x?: number;\n    y?: number;\n    z?: number;\n}\n\ninterface DOMQuadInit {\n    p1?: DOMPointInit;\n    p2?: DOMPointInit;\n    p3?: DOMPointInit;\n    p4?: DOMPointInit;\n}\n\ninterface DOMRectInit {\n    height?: number;\n    width?: number;\n    x?: number;\n    y?: number;\n}\n\ninterface DelayOptions extends AudioNodeOptions {\n    delayTime?: number;\n    maxDelayTime?: number;\n}\n\ninterface DeviceLightEventInit extends EventInit {\n    value?: number;\n}\n\ninterface DeviceMotionEventAccelerationInit {\n    x?: number | null;\n    y?: number | null;\n    z?: number | null;\n}\n\ninterface DeviceMotionEventInit extends EventInit {\n    acceleration?: DeviceMotionEventAccelerationInit;\n    accelerationIncludingGravity?: DeviceMotionEventAccelerationInit;\n    interval?: number;\n    rotationRate?: DeviceMotionEventRotationRateInit;\n}\n\ninterface DeviceMotionEventRotationRateInit {\n    alpha?: number | null;\n    beta?: number | null;\n    gamma?: number | null;\n}\n\ninterface DeviceOrientationEventInit extends EventInit {\n    absolute?: boolean;\n    alpha?: number | null;\n    beta?: number | null;\n    gamma?: number | null;\n}\n\ninterface DevicePermissionDescriptor extends PermissionDescriptor {\n    deviceId?: string;\n    name: "camera" | "microphone" | "speaker";\n}\n\ninterface DocumentTimelineOptions {\n    originTime?: number;\n}\n\ninterface DoubleRange {\n    max?: number;\n    min?: number;\n}\n\ninterface DragEventInit extends MouseEventInit {\n    dataTransfer?: DataTransfer | null;\n}\n\ninterface DynamicsCompressorOptions extends AudioNodeOptions {\n    attack?: number;\n    knee?: number;\n    ratio?: number;\n    release?: number;\n    threshold?: number;\n}\n\ninterface EcKeyAlgorithm extends KeyAlgorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcKeyGenParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcKeyImportParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcdhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface EcdsaParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface EffectTiming {\n    delay?: number;\n    direction?: PlaybackDirection;\n    duration?: number | string;\n    easing?: string;\n    endDelay?: number;\n    fill?: FillMode;\n    iterationStart?: number;\n    iterations?: number;\n}\n\ninterface ElementCreationOptions {\n    is?: string;\n}\n\ninterface ElementDefinitionOptions {\n    extends?: string;\n}\n\ninterface ErrorEventInit extends EventInit {\n    colno?: number;\n    error?: any;\n    filename?: string;\n    lineno?: number;\n    message?: string;\n}\n\ninterface EventInit {\n    bubbles?: boolean;\n    cancelable?: boolean;\n    composed?: boolean;\n}\n\ninterface EventListenerOptions {\n    capture?: boolean;\n}\n\ninterface EventModifierInit extends UIEventInit {\n    altKey?: boolean;\n    ctrlKey?: boolean;\n    metaKey?: boolean;\n    modifierAltGraph?: boolean;\n    modifierCapsLock?: boolean;\n    modifierFn?: boolean;\n    modifierFnLock?: boolean;\n    modifierHyper?: boolean;\n    modifierNumLock?: boolean;\n    modifierScrollLock?: boolean;\n    modifierSuper?: boolean;\n    modifierSymbol?: boolean;\n    modifierSymbolLock?: boolean;\n    shiftKey?: boolean;\n}\n\ninterface EventSourceInit {\n    withCredentials?: boolean;\n}\n\ninterface ExceptionInformation {\n    domain?: string | null;\n}\n\ninterface FilePropertyBag extends BlobPropertyBag {\n    lastModified?: number;\n}\n\ninterface FocusEventInit extends UIEventInit {\n    relatedTarget?: EventTarget | null;\n}\n\ninterface FocusNavigationEventInit extends EventInit {\n    navigationReason?: string | null;\n    originHeight?: number;\n    originLeft?: number;\n    originTop?: number;\n    originWidth?: number;\n}\n\ninterface FocusNavigationOrigin {\n    originHeight?: number;\n    originLeft?: number;\n    originTop?: number;\n    originWidth?: number;\n}\n\ninterface FocusOptions {\n    preventScroll?: boolean;\n}\n\ninterface FullscreenOptions {\n    navigationUI?: FullscreenNavigationUI;\n}\n\ninterface GainOptions extends AudioNodeOptions {\n    gain?: number;\n}\n\ninterface GamepadEventInit extends EventInit {\n    gamepad: Gamepad;\n}\n\ninterface GetNotificationOptions {\n    tag?: string;\n}\n\ninterface GetRootNodeOptions {\n    composed?: boolean;\n}\n\ninterface HashChangeEventInit extends EventInit {\n    newURL?: string;\n    oldURL?: string;\n}\n\ninterface HkdfParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    info: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface HmacImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface HmacKeyAlgorithm extends KeyAlgorithm {\n    hash: KeyAlgorithm;\n    length: number;\n}\n\ninterface HmacKeyGenParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface IDBIndexParameters {\n    multiEntry?: boolean;\n    unique?: boolean;\n}\n\ninterface IDBObjectStoreParameters {\n    autoIncrement?: boolean;\n    keyPath?: string | string[] | null;\n}\n\ninterface IDBVersionChangeEventInit extends EventInit {\n    newVersion?: number | null;\n    oldVersion?: number;\n}\n\ninterface IIRFilterOptions extends AudioNodeOptions {\n    feedback: number[];\n    feedforward: number[];\n}\n\ninterface ImageBitmapOptions {\n    colorSpaceConversion?: ColorSpaceConversion;\n    imageOrientation?: ImageOrientation;\n    premultiplyAlpha?: PremultiplyAlpha;\n    resizeHeight?: number;\n    resizeQuality?: ResizeQuality;\n    resizeWidth?: number;\n}\n\ninterface ImageBitmapRenderingContextSettings {\n    alpha?: boolean;\n}\n\ninterface ImageEncodeOptions {\n    quality?: number;\n    type?: string;\n}\n\ninterface InputEventInit extends UIEventInit {\n    data?: string | null;\n    inputType?: string;\n    isComposing?: boolean;\n}\n\ninterface IntersectionObserverEntryInit {\n    boundingClientRect: DOMRectInit;\n    intersectionRatio: number;\n    intersectionRect: DOMRectInit;\n    isIntersecting: boolean;\n    rootBounds: DOMRectInit | null;\n    target: Element;\n    time: number;\n}\n\ninterface IntersectionObserverInit {\n    root?: Element | null;\n    rootMargin?: string;\n    threshold?: number | number[];\n}\n\ninterface JsonWebKey {\n    alg?: string;\n    crv?: string;\n    d?: string;\n    dp?: string;\n    dq?: string;\n    e?: string;\n    ext?: boolean;\n    k?: string;\n    key_ops?: string[];\n    kty?: string;\n    n?: string;\n    oth?: RsaOtherPrimesInfo[];\n    p?: string;\n    q?: string;\n    qi?: string;\n    use?: string;\n    x?: string;\n    y?: string;\n}\n\ninterface KeyAlgorithm {\n    name: string;\n}\n\ninterface KeyboardEventInit extends EventModifierInit {\n    code?: string;\n    isComposing?: boolean;\n    key?: string;\n    /** @deprecated */\n    keyCode?: number;\n    location?: number;\n    repeat?: boolean;\n}\n\ninterface Keyframe {\n    composite?: CompositeOperationOrAuto;\n    easing?: string;\n    offset?: number | null;\n    [property: string]: string | number | null | undefined;\n}\n\ninterface KeyframeAnimationOptions extends KeyframeEffectOptions {\n    id?: string;\n}\n\ninterface KeyframeEffectOptions extends EffectTiming {\n    composite?: CompositeOperation;\n    iterationComposite?: IterationCompositeOperation;\n}\n\ninterface MediaElementAudioSourceOptions {\n    mediaElement: HTMLMediaElement;\n}\n\ninterface MediaEncryptedEventInit extends EventInit {\n    initData?: ArrayBuffer | null;\n    initDataType?: string;\n}\n\ninterface MediaKeyMessageEventInit extends EventInit {\n    message: ArrayBuffer;\n    messageType: MediaKeyMessageType;\n}\n\ninterface MediaKeySystemConfiguration {\n    audioCapabilities?: MediaKeySystemMediaCapability[];\n    distinctiveIdentifier?: MediaKeysRequirement;\n    initDataTypes?: string[];\n    label?: string;\n    persistentState?: MediaKeysRequirement;\n    sessionTypes?: string[];\n    videoCapabilities?: MediaKeySystemMediaCapability[];\n}\n\ninterface MediaKeySystemMediaCapability {\n    contentType?: string;\n    robustness?: string;\n}\n\ninterface MediaQueryListEventInit extends EventInit {\n    matches?: boolean;\n    media?: string;\n}\n\ninterface MediaStreamAudioSourceOptions {\n    mediaStream: MediaStream;\n}\n\ninterface MediaStreamConstraints {\n    audio?: boolean | MediaTrackConstraints;\n    peerIdentity?: string;\n    video?: boolean | MediaTrackConstraints;\n}\n\ninterface MediaStreamErrorEventInit extends EventInit {\n    error?: MediaStreamError | null;\n}\n\ninterface MediaStreamEventInit extends EventInit {\n    stream?: MediaStream;\n}\n\ninterface MediaStreamTrackAudioSourceOptions {\n    mediaStreamTrack: MediaStreamTrack;\n}\n\ninterface MediaStreamTrackEventInit extends EventInit {\n    track: MediaStreamTrack;\n}\n\ninterface MediaTrackCapabilities {\n    aspectRatio?: DoubleRange;\n    autoGainControl?: boolean[];\n    channelCount?: ULongRange;\n    deviceId?: string;\n    echoCancellation?: boolean[];\n    facingMode?: string[];\n    frameRate?: DoubleRange;\n    groupId?: string;\n    height?: ULongRange;\n    latency?: DoubleRange;\n    noiseSuppression?: boolean[];\n    resizeMode?: string[];\n    sampleRate?: ULongRange;\n    sampleSize?: ULongRange;\n    width?: ULongRange;\n}\n\ninterface MediaTrackConstraintSet {\n    aspectRatio?: ConstrainDouble;\n    autoGainControl?: ConstrainBoolean;\n    channelCount?: ConstrainULong;\n    deviceId?: ConstrainDOMString;\n    echoCancellation?: ConstrainBoolean;\n    facingMode?: ConstrainDOMString;\n    frameRate?: ConstrainDouble;\n    groupId?: ConstrainDOMString;\n    height?: ConstrainULong;\n    latency?: ConstrainDouble;\n    noiseSuppression?: ConstrainBoolean;\n    resizeMode?: ConstrainDOMString;\n    sampleRate?: ConstrainULong;\n    sampleSize?: ConstrainULong;\n    width?: ConstrainULong;\n}\n\ninterface MediaTrackConstraints extends MediaTrackConstraintSet {\n    advanced?: MediaTrackConstraintSet[];\n}\n\ninterface MediaTrackSettings {\n    aspectRatio?: number;\n    autoGainControl?: boolean;\n    channelCount?: number;\n    deviceId?: string;\n    echoCancellation?: boolean;\n    facingMode?: string;\n    frameRate?: number;\n    groupId?: string;\n    height?: number;\n    latency?: number;\n    noiseSuppression?: boolean;\n    resizeMode?: string;\n    sampleRate?: number;\n    sampleSize?: number;\n    width?: number;\n}\n\ninterface MediaTrackSupportedConstraints {\n    aspectRatio?: boolean;\n    autoGainControl?: boolean;\n    channelCount?: boolean;\n    deviceId?: boolean;\n    echoCancellation?: boolean;\n    facingMode?: boolean;\n    frameRate?: boolean;\n    groupId?: boolean;\n    height?: boolean;\n    latency?: boolean;\n    noiseSuppression?: boolean;\n    resizeMode?: boolean;\n    sampleRate?: boolean;\n    sampleSize?: boolean;\n    width?: boolean;\n}\n\ninterface MessageEventInit<T = any> extends EventInit {\n    data?: T;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[];\n    source?: MessageEventSource | null;\n}\n\ninterface MidiPermissionDescriptor extends PermissionDescriptor {\n    name: "midi";\n    sysex?: boolean;\n}\n\ninterface MouseEventInit extends EventModifierInit {\n    button?: number;\n    buttons?: number;\n    clientX?: number;\n    clientY?: number;\n    movementX?: number;\n    movementY?: number;\n    relatedTarget?: EventTarget | null;\n    screenX?: number;\n    screenY?: number;\n}\n\ninterface MultiCacheQueryOptions extends CacheQueryOptions {\n    cacheName?: string;\n}\n\ninterface MutationObserverInit {\n    /**\n     * Set to a list of attribute local names (without namespace) if not all attribute mutations need to be observed and attributes is true or omitted.\n     */\n    attributeFilter?: string[];\n    /**\n     * Set to true if attributes is true or omitted and target\'s attribute value before the mutation needs to be recorded.\n     */\n    attributeOldValue?: boolean;\n    /**\n     * Set to true if mutations to target\'s attributes are to be observed. Can be omitted if attributeOldValue or attributeFilter is specified.\n     */\n    attributes?: boolean;\n    /**\n     * Set to true if mutations to target\'s data are to be observed. Can be omitted if characterDataOldValue is specified.\n     */\n    characterData?: boolean;\n    /**\n     * Set to true if characterData is set to true or omitted and target\'s data before the mutation needs to be recorded.\n     */\n    characterDataOldValue?: boolean;\n    /**\n     * Set to true if mutations to target\'s children are to be observed.\n     */\n    childList?: boolean;\n    /**\n     * Set to true if mutations to not just target, but also target\'s descendants are to be observed.\n     */\n    subtree?: boolean;\n}\n\ninterface NavigationPreloadState {\n    enabled?: boolean;\n    headerValue?: string;\n}\n\ninterface NotificationAction {\n    action: string;\n    icon?: string;\n    title: string;\n}\n\ninterface NotificationOptions {\n    actions?: NotificationAction[];\n    badge?: string;\n    body?: string;\n    data?: any;\n    dir?: NotificationDirection;\n    icon?: string;\n    image?: string;\n    lang?: string;\n    renotify?: boolean;\n    requireInteraction?: boolean;\n    silent?: boolean;\n    tag?: string;\n    timestamp?: number;\n    vibrate?: VibratePattern;\n}\n\ninterface OfflineAudioCompletionEventInit extends EventInit {\n    renderedBuffer: AudioBuffer;\n}\n\ninterface OfflineAudioContextOptions {\n    length: number;\n    numberOfChannels?: number;\n    sampleRate: number;\n}\n\ninterface OptionalEffectTiming {\n    delay?: number;\n    direction?: PlaybackDirection;\n    duration?: number | string;\n    easing?: string;\n    endDelay?: number;\n    fill?: FillMode;\n    iterationStart?: number;\n    iterations?: number;\n}\n\ninterface OscillatorOptions extends AudioNodeOptions {\n    detune?: number;\n    frequency?: number;\n    periodicWave?: PeriodicWave;\n    type?: OscillatorType;\n}\n\ninterface PageTransitionEventInit extends EventInit {\n    persisted?: boolean;\n}\n\ninterface PannerOptions extends AudioNodeOptions {\n    coneInnerAngle?: number;\n    coneOuterAngle?: number;\n    coneOuterGain?: number;\n    distanceModel?: DistanceModelType;\n    maxDistance?: number;\n    orientationX?: number;\n    orientationY?: number;\n    orientationZ?: number;\n    panningModel?: PanningModelType;\n    positionX?: number;\n    positionY?: number;\n    positionZ?: number;\n    refDistance?: number;\n    rolloffFactor?: number;\n}\n\ninterface PaymentCurrencyAmount {\n    currency: string;\n    currencySystem?: string;\n    value: string;\n}\n\ninterface PaymentDetailsBase {\n    displayItems?: PaymentItem[];\n    modifiers?: PaymentDetailsModifier[];\n    shippingOptions?: PaymentShippingOption[];\n}\n\ninterface PaymentDetailsInit extends PaymentDetailsBase {\n    id?: string;\n    total: PaymentItem;\n}\n\ninterface PaymentDetailsModifier {\n    additionalDisplayItems?: PaymentItem[];\n    data?: any;\n    supportedMethods: string | string[];\n    total?: PaymentItem;\n}\n\ninterface PaymentDetailsUpdate extends PaymentDetailsBase {\n    error?: string;\n    total?: PaymentItem;\n}\n\ninterface PaymentItem {\n    amount: PaymentCurrencyAmount;\n    label: string;\n    pending?: boolean;\n}\n\ninterface PaymentMethodData {\n    data?: any;\n    supportedMethods: string | string[];\n}\n\ninterface PaymentOptions {\n    requestPayerEmail?: boolean;\n    requestPayerName?: boolean;\n    requestPayerPhone?: boolean;\n    requestShipping?: boolean;\n    shippingType?: string;\n}\n\ninterface PaymentRequestUpdateEventInit extends EventInit {\n}\n\ninterface PaymentShippingOption {\n    amount: PaymentCurrencyAmount;\n    id: string;\n    label: string;\n    selected?: boolean;\n}\n\ninterface Pbkdf2Params extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    iterations: number;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface PerformanceObserverInit {\n    buffered?: boolean;\n    entryTypes?: string[];\n    type?: string;\n}\n\ninterface PeriodicWaveConstraints {\n    disableNormalization?: boolean;\n}\n\ninterface PeriodicWaveOptions extends PeriodicWaveConstraints {\n    imag?: number[] | Float32Array;\n    real?: number[] | Float32Array;\n}\n\ninterface PermissionDescriptor {\n    name: PermissionName;\n}\n\ninterface PipeOptions {\n    preventAbort?: boolean;\n    preventCancel?: boolean;\n    preventClose?: boolean;\n    signal?: AbortSignal;\n}\n\ninterface PointerEventInit extends MouseEventInit {\n    height?: number;\n    isPrimary?: boolean;\n    pointerId?: number;\n    pointerType?: string;\n    pressure?: number;\n    tangentialPressure?: number;\n    tiltX?: number;\n    tiltY?: number;\n    twist?: number;\n    width?: number;\n}\n\ninterface PopStateEventInit extends EventInit {\n    state?: any;\n}\n\ninterface PositionOptions {\n    enableHighAccuracy?: boolean;\n    maximumAge?: number;\n    timeout?: number;\n}\n\ninterface PostMessageOptions {\n    transfer?: any[];\n}\n\ninterface ProgressEventInit extends EventInit {\n    lengthComputable?: boolean;\n    loaded?: number;\n    total?: number;\n}\n\ninterface PromiseRejectionEventInit extends EventInit {\n    promise: Promise<any>;\n    reason?: any;\n}\n\ninterface PropertyIndexedKeyframes {\n    composite?: CompositeOperationOrAuto | CompositeOperationOrAuto[];\n    easing?: string | string[];\n    offset?: number | (number | null)[];\n    [property: string]: string | string[] | number | null | (number | null)[] | undefined;\n}\n\ninterface PublicKeyCredentialCreationOptions {\n    attestation?: AttestationConveyancePreference;\n    authenticatorSelection?: AuthenticatorSelectionCriteria;\n    challenge: BufferSource;\n    excludeCredentials?: PublicKeyCredentialDescriptor[];\n    extensions?: AuthenticationExtensionsClientInputs;\n    pubKeyCredParams: PublicKeyCredentialParameters[];\n    rp: PublicKeyCredentialRpEntity;\n    timeout?: number;\n    user: PublicKeyCredentialUserEntity;\n}\n\ninterface PublicKeyCredentialDescriptor {\n    id: BufferSource;\n    transports?: AuthenticatorTransport[];\n    type: PublicKeyCredentialType;\n}\n\ninterface PublicKeyCredentialEntity {\n    name: string;\n}\n\ninterface PublicKeyCredentialParameters {\n    alg: COSEAlgorithmIdentifier;\n    type: PublicKeyCredentialType;\n}\n\ninterface PublicKeyCredentialRequestOptions {\n    allowCredentials?: PublicKeyCredentialDescriptor[];\n    challenge: BufferSource;\n    extensions?: AuthenticationExtensionsClientInputs;\n    rpId?: string;\n    timeout?: number;\n    userVerification?: UserVerificationRequirement;\n}\n\ninterface PublicKeyCredentialRpEntity extends PublicKeyCredentialEntity {\n    id?: string;\n}\n\ninterface PublicKeyCredentialUserEntity extends PublicKeyCredentialEntity {\n    displayName: string;\n    id: BufferSource;\n}\n\ninterface PushPermissionDescriptor extends PermissionDescriptor {\n    name: "push";\n    userVisibleOnly?: boolean;\n}\n\ninterface PushSubscriptionJSON {\n    endpoint?: string;\n    expirationTime?: number | null;\n    keys?: Record<string, string>;\n}\n\ninterface PushSubscriptionOptionsInit {\n    applicationServerKey?: BufferSource | string | null;\n    userVisibleOnly?: boolean;\n}\n\ninterface QueuingStrategy<T = any> {\n    highWaterMark?: number;\n    size?: QueuingStrategySizeCallback<T>;\n}\n\ninterface RTCAnswerOptions extends RTCOfferAnswerOptions {\n}\n\ninterface RTCCertificateExpiration {\n    expires?: number;\n}\n\ninterface RTCConfiguration {\n    bundlePolicy?: RTCBundlePolicy;\n    certificates?: RTCCertificate[];\n    iceCandidatePoolSize?: number;\n    iceServers?: RTCIceServer[];\n    iceTransportPolicy?: RTCIceTransportPolicy;\n    peerIdentity?: string;\n    rtcpMuxPolicy?: RTCRtcpMuxPolicy;\n}\n\ninterface RTCDTMFToneChangeEventInit extends EventInit {\n    tone: string;\n}\n\ninterface RTCDataChannelEventInit extends EventInit {\n    channel: RTCDataChannel;\n}\n\ninterface RTCDataChannelInit {\n    id?: number;\n    maxPacketLifeTime?: number;\n    maxRetransmits?: number;\n    negotiated?: boolean;\n    ordered?: boolean;\n    priority?: RTCPriorityType;\n    protocol?: string;\n}\n\ninterface RTCDtlsFingerprint {\n    algorithm?: string;\n    value?: string;\n}\n\ninterface RTCDtlsParameters {\n    fingerprints?: RTCDtlsFingerprint[];\n    role?: RTCDtlsRole;\n}\n\ninterface RTCErrorEventInit extends EventInit {\n    error: RTCError;\n}\n\ninterface RTCErrorInit {\n    errorDetail: RTCErrorDetailType;\n    httpRequestStatusCode?: number;\n    receivedAlert?: number;\n    sctpCauseCode?: number;\n    sdpLineNumber?: number;\n    sentAlert?: number;\n}\n\ninterface RTCIceCandidateAttributes extends RTCStats {\n    addressSourceUrl?: string;\n    candidateType?: RTCStatsIceCandidateType;\n    ipAddress?: string;\n    portNumber?: number;\n    priority?: number;\n    transport?: string;\n}\n\ninterface RTCIceCandidateComplete {\n}\n\ninterface RTCIceCandidateDictionary {\n    foundation?: string;\n    ip?: string;\n    msMTurnSessionId?: string;\n    port?: number;\n    priority?: number;\n    protocol?: RTCIceProtocol;\n    relatedAddress?: string;\n    relatedPort?: number;\n    tcpType?: RTCIceTcpCandidateType;\n    type?: RTCIceCandidateType;\n}\n\ninterface RTCIceCandidateInit {\n    candidate?: string;\n    sdpMLineIndex?: number | null;\n    sdpMid?: string | null;\n    usernameFragment?: string | null;\n}\n\ninterface RTCIceCandidatePair {\n    local?: RTCIceCandidate;\n    remote?: RTCIceCandidate;\n}\n\ninterface RTCIceCandidatePairStats extends RTCStats {\n    availableIncomingBitrate?: number;\n    availableOutgoingBitrate?: number;\n    bytesReceived?: number;\n    bytesSent?: number;\n    localCandidateId?: string;\n    nominated?: boolean;\n    priority?: number;\n    readable?: boolean;\n    remoteCandidateId?: string;\n    roundTripTime?: number;\n    state?: RTCStatsIceCandidatePairState;\n    transportId?: string;\n    writable?: boolean;\n}\n\ninterface RTCIceGatherOptions {\n    gatherPolicy?: RTCIceGatherPolicy;\n    iceservers?: RTCIceServer[];\n}\n\ninterface RTCIceParameters {\n    password?: string;\n    usernameFragment?: string;\n}\n\ninterface RTCIceServer {\n    credential?: string | RTCOAuthCredential;\n    credentialType?: RTCIceCredentialType;\n    urls: string | string[];\n    username?: string;\n}\n\ninterface RTCIdentityProviderOptions {\n    peerIdentity?: string;\n    protocol?: string;\n    usernameHint?: string;\n}\n\ninterface RTCInboundRTPStreamStats extends RTCRTPStreamStats {\n    bytesReceived?: number;\n    fractionLost?: number;\n    jitter?: number;\n    packetsLost?: number;\n    packetsReceived?: number;\n}\n\ninterface RTCMediaStreamTrackStats extends RTCStats {\n    audioLevel?: number;\n    echoReturnLoss?: number;\n    echoReturnLossEnhancement?: number;\n    frameHeight?: number;\n    frameWidth?: number;\n    framesCorrupted?: number;\n    framesDecoded?: number;\n    framesDropped?: number;\n    framesPerSecond?: number;\n    framesReceived?: number;\n    framesSent?: number;\n    remoteSource?: boolean;\n    ssrcIds?: string[];\n    trackIdentifier?: string;\n}\n\ninterface RTCOAuthCredential {\n    accessToken: string;\n    macKey: string;\n}\n\ninterface RTCOfferAnswerOptions {\n    voiceActivityDetection?: boolean;\n}\n\ninterface RTCOfferOptions extends RTCOfferAnswerOptions {\n    iceRestart?: boolean;\n    offerToReceiveAudio?: boolean;\n    offerToReceiveVideo?: boolean;\n}\n\ninterface RTCOutboundRTPStreamStats extends RTCRTPStreamStats {\n    bytesSent?: number;\n    packetsSent?: number;\n    roundTripTime?: number;\n    targetBitrate?: number;\n}\n\ninterface RTCPeerConnectionIceErrorEventInit extends EventInit {\n    errorCode: number;\n    hostCandidate?: string;\n    statusText?: string;\n    url?: string;\n}\n\ninterface RTCPeerConnectionIceEventInit extends EventInit {\n    candidate?: RTCIceCandidate | null;\n    url?: string | null;\n}\n\ninterface RTCRTPStreamStats extends RTCStats {\n    associateStatsId?: string;\n    codecId?: string;\n    firCount?: number;\n    isRemote?: boolean;\n    mediaTrackId?: string;\n    mediaType?: string;\n    nackCount?: number;\n    pliCount?: number;\n    sliCount?: number;\n    ssrc?: string;\n    transportId?: string;\n}\n\ninterface RTCRtcpFeedback {\n    parameter?: string;\n    type?: string;\n}\n\ninterface RTCRtcpParameters {\n    cname?: string;\n    reducedSize?: boolean;\n}\n\ninterface RTCRtpCapabilities {\n    codecs: RTCRtpCodecCapability[];\n    headerExtensions: RTCRtpHeaderExtensionCapability[];\n}\n\ninterface RTCRtpCodecCapability {\n    channels?: number;\n    clockRate: number;\n    mimeType: string;\n    sdpFmtpLine?: string;\n}\n\ninterface RTCRtpCodecParameters {\n    channels?: number;\n    clockRate: number;\n    mimeType: string;\n    payloadType: number;\n    sdpFmtpLine?: string;\n}\n\ninterface RTCRtpCodingParameters {\n    rid?: string;\n}\n\ninterface RTCRtpContributingSource {\n    audioLevel?: number;\n    rtpTimestamp: number;\n    source: number;\n    timestamp: number;\n}\n\ninterface RTCRtpDecodingParameters extends RTCRtpCodingParameters {\n}\n\ninterface RTCRtpEncodingParameters extends RTCRtpCodingParameters {\n    active?: boolean;\n    codecPayloadType?: number;\n    dtx?: RTCDtxStatus;\n    maxBitrate?: number;\n    maxFramerate?: number;\n    ptime?: number;\n    scaleResolutionDownBy?: number;\n}\n\ninterface RTCRtpFecParameters {\n    mechanism?: string;\n    ssrc?: number;\n}\n\ninterface RTCRtpHeaderExtension {\n    kind?: string;\n    preferredEncrypt?: boolean;\n    preferredId?: number;\n    uri?: string;\n}\n\ninterface RTCRtpHeaderExtensionCapability {\n    uri?: string;\n}\n\ninterface RTCRtpHeaderExtensionParameters {\n    encrypted?: boolean;\n    id: number;\n    uri: string;\n}\n\ninterface RTCRtpParameters {\n    codecs: RTCRtpCodecParameters[];\n    headerExtensions: RTCRtpHeaderExtensionParameters[];\n    rtcp: RTCRtcpParameters;\n}\n\ninterface RTCRtpReceiveParameters extends RTCRtpParameters {\n    encodings: RTCRtpDecodingParameters[];\n}\n\ninterface RTCRtpRtxParameters {\n    ssrc?: number;\n}\n\ninterface RTCRtpSendParameters extends RTCRtpParameters {\n    degradationPreference?: RTCDegradationPreference;\n    encodings: RTCRtpEncodingParameters[];\n    priority?: RTCPriorityType;\n    transactionId: string;\n}\n\ninterface RTCRtpSynchronizationSource extends RTCRtpContributingSource {\n    voiceActivityFlag?: boolean;\n}\n\ninterface RTCRtpTransceiverInit {\n    direction?: RTCRtpTransceiverDirection;\n    sendEncodings?: RTCRtpEncodingParameters[];\n    streams?: MediaStream[];\n}\n\ninterface RTCRtpUnhandled {\n    muxId?: string;\n    payloadType?: number;\n    ssrc?: number;\n}\n\ninterface RTCSessionDescriptionInit {\n    sdp?: string;\n    type?: RTCSdpType;\n}\n\ninterface RTCSrtpKeyParam {\n    keyMethod?: string;\n    keySalt?: string;\n    lifetime?: string;\n    mkiLength?: number;\n    mkiValue?: number;\n}\n\ninterface RTCSrtpSdesParameters {\n    cryptoSuite?: string;\n    keyParams?: RTCSrtpKeyParam[];\n    sessionParams?: string[];\n    tag?: number;\n}\n\ninterface RTCSsrcRange {\n    max?: number;\n    min?: number;\n}\n\ninterface RTCStats {\n    id: string;\n    timestamp: number;\n    type: RTCStatsType;\n}\n\ninterface RTCStatsEventInit extends EventInit {\n    report: RTCStatsReport;\n}\n\ninterface RTCStatsReport {\n}\n\ninterface RTCTrackEventInit extends EventInit {\n    receiver: RTCRtpReceiver;\n    streams?: MediaStream[];\n    track: MediaStreamTrack;\n    transceiver: RTCRtpTransceiver;\n}\n\ninterface RTCTransportStats extends RTCStats {\n    activeConnection?: boolean;\n    bytesReceived?: number;\n    bytesSent?: number;\n    localCertificateId?: string;\n    remoteCertificateId?: string;\n    rtcpTransportStatsId?: string;\n    selectedCandidatePairId?: string;\n}\n\ninterface ReadableStreamReadDoneResult<T> {\n    done: true;\n    value?: T;\n}\n\ninterface ReadableStreamReadValueResult<T> {\n    done: false;\n    value: T;\n}\n\ninterface RegistrationOptions {\n    scope?: string;\n    type?: WorkerType;\n    updateViaCache?: ServiceWorkerUpdateViaCache;\n}\n\ninterface RequestInit {\n    /**\n     * A BodyInit object or null to set request\'s body.\n     */\n    body?: BodyInit | null;\n    /**\n     * A string indicating how the request will interact with the browser\'s cache to set request\'s cache.\n     */\n    cache?: RequestCache;\n    /**\n     * A string indicating whether credentials will be sent with the request always, never, or only when sent to a same-origin URL. Sets request\'s credentials.\n     */\n    credentials?: RequestCredentials;\n    /**\n     * A Headers object, an object literal, or an array of two-item arrays to set request\'s headers.\n     */\n    headers?: HeadersInit;\n    /**\n     * A cryptographic hash of the resource to be fetched by request. Sets request\'s integrity.\n     */\n    integrity?: string;\n    /**\n     * A boolean to set request\'s keepalive.\n     */\n    keepalive?: boolean;\n    /**\n     * A string to set request\'s method.\n     */\n    method?: string;\n    /**\n     * A string to indicate whether the request will use CORS, or will be restricted to same-origin URLs. Sets request\'s mode.\n     */\n    mode?: RequestMode;\n    /**\n     * A string indicating whether request follows redirects, results in an error upon encountering a redirect, or returns the redirect (in an opaque fashion). Sets request\'s redirect.\n     */\n    redirect?: RequestRedirect;\n    /**\n     * A string whose value is a same-origin URL, "about:client", or the empty string, to set request\'s referrer.\n     */\n    referrer?: string;\n    /**\n     * A referrer policy to set request\'s referrerPolicy.\n     */\n    referrerPolicy?: ReferrerPolicy;\n    /**\n     * An AbortSignal to set request\'s signal.\n     */\n    signal?: AbortSignal | null;\n    /**\n     * Can only be null. Used to disassociate request from any Window.\n     */\n    window?: any;\n}\n\ninterface ResponseInit {\n    headers?: HeadersInit;\n    status?: number;\n    statusText?: string;\n}\n\ninterface RsaHashedImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaHashedKeyAlgorithm extends RsaKeyAlgorithm {\n    hash: KeyAlgorithm;\n}\n\ninterface RsaHashedKeyGenParams extends RsaKeyGenParams {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaKeyAlgorithm extends KeyAlgorithm {\n    modulusLength: number;\n    publicExponent: BigInteger;\n}\n\ninterface RsaKeyGenParams extends Algorithm {\n    modulusLength: number;\n    publicExponent: BigInteger;\n}\n\ninterface RsaOaepParams extends Algorithm {\n    label?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface RsaOtherPrimesInfo {\n    d?: string;\n    r?: string;\n    t?: string;\n}\n\ninterface RsaPssParams extends Algorithm {\n    saltLength: number;\n}\n\ninterface SVGBoundingBoxOptions {\n    clipped?: boolean;\n    fill?: boolean;\n    markers?: boolean;\n    stroke?: boolean;\n}\n\ninterface ScopedCredentialDescriptor {\n    id: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null;\n    transports?: Transport[];\n    type: ScopedCredentialType;\n}\n\ninterface ScopedCredentialOptions {\n    excludeList?: ScopedCredentialDescriptor[];\n    extensions?: WebAuthnExtensions;\n    rpId?: string;\n    timeoutSeconds?: number;\n}\n\ninterface ScopedCredentialParameters {\n    algorithm: string | Algorithm;\n    type: ScopedCredentialType;\n}\n\ninterface ScrollIntoViewOptions extends ScrollOptions {\n    block?: ScrollLogicalPosition;\n    inline?: ScrollLogicalPosition;\n}\n\ninterface ScrollOptions {\n    behavior?: ScrollBehavior;\n}\n\ninterface ScrollToOptions extends ScrollOptions {\n    left?: number;\n    top?: number;\n}\n\ninterface SecurityPolicyViolationEventInit extends EventInit {\n    blockedURI?: string;\n    columnNumber?: number;\n    documentURI?: string;\n    effectiveDirective?: string;\n    lineNumber?: number;\n    originalPolicy?: string;\n    referrer?: string;\n    sourceFile?: string;\n    statusCode?: number;\n    violatedDirective?: string;\n}\n\ninterface ServiceWorkerMessageEventInit extends EventInit {\n    data?: any;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[] | null;\n    source?: ServiceWorker | MessagePort | null;\n}\n\ninterface ShadowRootInit {\n    delegatesFocus?: boolean;\n    mode: ShadowRootMode;\n}\n\ninterface ShareData {\n    text?: string;\n    title?: string;\n    url?: string;\n}\n\ninterface SpeechSynthesisErrorEventInit extends SpeechSynthesisEventInit {\n    error: SpeechSynthesisErrorCode;\n}\n\ninterface SpeechSynthesisEventInit extends EventInit {\n    charIndex?: number;\n    charLength?: number;\n    elapsedTime?: number;\n    name?: string;\n    utterance: SpeechSynthesisUtterance;\n}\n\ninterface StaticRangeInit {\n    endContainer: Node;\n    endOffset: number;\n    startContainer: Node;\n    startOffset: number;\n}\n\ninterface StereoPannerOptions extends AudioNodeOptions {\n    pan?: number;\n}\n\ninterface StorageEstimate {\n    quota?: number;\n    usage?: number;\n}\n\ninterface StorageEventInit extends EventInit {\n    key?: string | null;\n    newValue?: string | null;\n    oldValue?: string | null;\n    storageArea?: Storage | null;\n    url?: string;\n}\n\ninterface StoreExceptionsInformation extends ExceptionInformation {\n    detailURI?: string | null;\n    explanationString?: string | null;\n    siteName?: string | null;\n}\n\ninterface StoreSiteSpecificExceptionsInformation extends StoreExceptionsInformation {\n    arrayOfDomainStrings?: string[];\n}\n\ninterface TextDecodeOptions {\n    stream?: boolean;\n}\n\ninterface TextDecoderOptions {\n    fatal?: boolean;\n    ignoreBOM?: boolean;\n}\n\ninterface TextEncoderEncodeIntoResult {\n    read?: number;\n    written?: number;\n}\n\ninterface TouchEventInit extends EventModifierInit {\n    changedTouches?: Touch[];\n    targetTouches?: Touch[];\n    touches?: Touch[];\n}\n\ninterface TouchInit {\n    altitudeAngle?: number;\n    azimuthAngle?: number;\n    clientX?: number;\n    clientY?: number;\n    force?: number;\n    identifier: number;\n    pageX?: number;\n    pageY?: number;\n    radiusX?: number;\n    radiusY?: number;\n    rotationAngle?: number;\n    screenX?: number;\n    screenY?: number;\n    target: EventTarget;\n    touchType?: TouchType;\n}\n\ninterface TrackEventInit extends EventInit {\n    track?: TextTrack | null;\n}\n\ninterface Transformer<I = any, O = any> {\n    flush?: TransformStreamDefaultControllerCallback<O>;\n    readableType?: undefined;\n    start?: TransformStreamDefaultControllerCallback<O>;\n    transform?: TransformStreamDefaultControllerTransformCallback<I, O>;\n    writableType?: undefined;\n}\n\ninterface TransitionEventInit extends EventInit {\n    elapsedTime?: number;\n    propertyName?: string;\n    pseudoElement?: string;\n}\n\ninterface UIEventInit extends EventInit {\n    detail?: number;\n    view?: Window | null;\n}\n\ninterface ULongRange {\n    max?: number;\n    min?: number;\n}\n\ninterface UnderlyingByteSource {\n    autoAllocateChunkSize?: number;\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableByteStreamControllerCallback;\n    start?: ReadableByteStreamControllerCallback;\n    type: "bytes";\n}\n\ninterface UnderlyingSink<W = any> {\n    abort?: WritableStreamErrorCallback;\n    close?: WritableStreamDefaultControllerCloseCallback;\n    start?: WritableStreamDefaultControllerStartCallback;\n    type?: undefined;\n    write?: WritableStreamDefaultControllerWriteCallback<W>;\n}\n\ninterface UnderlyingSource<R = any> {\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableStreamDefaultControllerCallback<R>;\n    start?: ReadableStreamDefaultControllerCallback<R>;\n    type?: undefined;\n}\n\ninterface VRDisplayEventInit extends EventInit {\n    display: VRDisplay;\n    reason?: VRDisplayEventReason;\n}\n\ninterface VRLayer {\n    leftBounds?: number[] | Float32Array | null;\n    rightBounds?: number[] | Float32Array | null;\n    source?: HTMLCanvasElement | null;\n}\n\ninterface VRStageParameters {\n    sittingToStandingTransform?: Float32Array;\n    sizeX?: number;\n    sizeY?: number;\n}\n\ninterface WaveShaperOptions extends AudioNodeOptions {\n    curve?: number[] | Float32Array;\n    oversample?: OverSampleType;\n}\n\ninterface WebAuthnExtensions {\n}\n\ninterface WebGLContextAttributes {\n    alpha?: boolean;\n    antialias?: boolean;\n    depth?: boolean;\n    desynchronized?: boolean;\n    failIfMajorPerformanceCaveat?: boolean;\n    powerPreference?: WebGLPowerPreference;\n    premultipliedAlpha?: boolean;\n    preserveDrawingBuffer?: boolean;\n    stencil?: boolean;\n}\n\ninterface WebGLContextEventInit extends EventInit {\n    statusMessage?: string;\n}\n\ninterface WheelEventInit extends MouseEventInit {\n    deltaMode?: number;\n    deltaX?: number;\n    deltaY?: number;\n    deltaZ?: number;\n}\n\ninterface WorkerOptions {\n    credentials?: RequestCredentials;\n    name?: string;\n    type?: WorkerType;\n}\n\ninterface WorkletOptions {\n    credentials?: RequestCredentials;\n}\n\ninterface EventListener {\n    (evt: Event): void;\n}\n\ntype XPathNSResolver = ((prefix: string | null) => string | null) | { lookupNamespaceURI(prefix: string | null): string | null; };\n\n/** The ANGLE_instanced_arrays extension is part of the WebGL API and allows to draw the same object, or groups of similar objects multiple times, if they share the same vertex data, primitive count and type. */\ninterface ANGLE_instanced_arrays {\n    drawArraysInstancedANGLE(mode: GLenum, first: GLint, count: GLsizei, primcount: GLsizei): void;\n    drawElementsInstancedANGLE(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr, primcount: GLsizei): void;\n    vertexAttribDivisorANGLE(index: GLuint, divisor: GLuint): void;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR_ANGLE: GLenum;\n}\n\n/** A controller object that allows you to abort one or more DOM requests as and when desired. */\ninterface AbortController {\n    /**\n     * Returns the AbortSignal object associated with this object.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Invoking this method will set this object\'s AbortSignal\'s aborted flag and signal to any observers that the associated activity is to be aborted.\n     */\n    abort(): void;\n}\n\ndeclare var AbortController: {\n    prototype: AbortController;\n    new(): AbortController;\n};\n\ninterface AbortSignalEventMap {\n    "abort": Event;\n}\n\n/** A signal object that allows you to communicate with a DOM request (such as a Fetch) and abort it if required via an AbortController object. */\ninterface AbortSignal extends EventTarget {\n    /**\n     * Returns true if this AbortSignal\'s AbortController has signaled to abort, and false otherwise.\n     */\n    readonly aborted: boolean;\n    onabort: ((this: AbortSignal, ev: Event) => any) | null;\n    addEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AbortSignal: {\n    prototype: AbortSignal;\n    new(): AbortSignal;\n};\n\ninterface AbstractRange {\n    /**\n     * Returns true if range is collapsed, and false otherwise.\n     */\n    readonly collapsed: boolean;\n    /**\n     * Returns range\'s end node.\n     */\n    readonly endContainer: Node;\n    /**\n     * Returns range\'s end offset.\n     */\n    readonly endOffset: number;\n    /**\n     * Returns range\'s start node.\n     */\n    readonly startContainer: Node;\n    /**\n     * Returns range\'s start offset.\n     */\n    readonly startOffset: number;\n}\n\ndeclare var AbstractRange: {\n    prototype: AbstractRange;\n    new(): AbstractRange;\n};\n\ninterface AbstractWorkerEventMap {\n    "error": ErrorEvent;\n}\n\ninterface AbstractWorker {\n    onerror: ((this: AbstractWorker, ev: ErrorEvent) => any) | null;\n    addEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface AesCfbParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCmacParams extends Algorithm {\n    length: number;\n}\n\n/** A node able to provide real-time frequency and time-domain analysis information. It is an AudioNode that passes the audio stream unchanged from the input to the output, but allows you to take the generated data, process it, and create audio visualizations. */\ninterface AnalyserNode extends AudioNode {\n    fftSize: number;\n    readonly frequencyBinCount: number;\n    maxDecibels: number;\n    minDecibels: number;\n    smoothingTimeConstant: number;\n    getByteFrequencyData(array: Uint8Array): void;\n    getByteTimeDomainData(array: Uint8Array): void;\n    getFloatFrequencyData(array: Float32Array): void;\n    getFloatTimeDomainData(array: Float32Array): void;\n}\n\ndeclare var AnalyserNode: {\n    prototype: AnalyserNode;\n    new(context: BaseAudioContext, options?: AnalyserOptions): AnalyserNode;\n};\n\ninterface Animatable {\n    animate(keyframes: Keyframe[] | PropertyIndexedKeyframes | null, options?: number | KeyframeAnimationOptions): Animation;\n    getAnimations(): Animation[];\n}\n\ninterface AnimationEventMap {\n    "cancel": AnimationPlaybackEvent;\n    "finish": AnimationPlaybackEvent;\n}\n\ninterface Animation extends EventTarget {\n    currentTime: number | null;\n    effect: AnimationEffect | null;\n    readonly finished: Promise<Animation>;\n    id: string;\n    oncancel: ((this: Animation, ev: AnimationPlaybackEvent) => any) | null;\n    onfinish: ((this: Animation, ev: AnimationPlaybackEvent) => any) | null;\n    readonly pending: boolean;\n    readonly playState: AnimationPlayState;\n    playbackRate: number;\n    readonly ready: Promise<Animation>;\n    startTime: number | null;\n    timeline: AnimationTimeline | null;\n    cancel(): void;\n    finish(): void;\n    pause(): void;\n    play(): void;\n    reverse(): void;\n    updatePlaybackRate(playbackRate: number): void;\n    addEventListener<K extends keyof AnimationEventMap>(type: K, listener: (this: Animation, ev: AnimationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AnimationEventMap>(type: K, listener: (this: Animation, ev: AnimationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Animation: {\n    prototype: Animation;\n    new(effect?: AnimationEffect | null, timeline?: AnimationTimeline | null): Animation;\n};\n\ninterface AnimationEffect {\n    getComputedTiming(): ComputedEffectTiming;\n    getTiming(): EffectTiming;\n    updateTiming(timing?: OptionalEffectTiming): void;\n}\n\ndeclare var AnimationEffect: {\n    prototype: AnimationEffect;\n    new(): AnimationEffect;\n};\n\n/** Events providing information related to animations. */\ninterface AnimationEvent extends Event {\n    readonly animationName: string;\n    readonly elapsedTime: number;\n    readonly pseudoElement: string;\n}\n\ndeclare var AnimationEvent: {\n    prototype: AnimationEvent;\n    new(type: string, animationEventInitDict?: AnimationEventInit): AnimationEvent;\n};\n\ninterface AnimationFrameProvider {\n    cancelAnimationFrame(handle: number): void;\n    requestAnimationFrame(callback: FrameRequestCallback): number;\n}\n\ninterface AnimationPlaybackEvent extends Event {\n    readonly currentTime: number | null;\n    readonly timelineTime: number | null;\n}\n\ndeclare var AnimationPlaybackEvent: {\n    prototype: AnimationPlaybackEvent;\n    new(type: string, eventInitDict?: AnimationPlaybackEventInit): AnimationPlaybackEvent;\n};\n\ninterface AnimationTimeline {\n    readonly currentTime: number | null;\n}\n\ndeclare var AnimationTimeline: {\n    prototype: AnimationTimeline;\n    new(): AnimationTimeline;\n};\n\ninterface ApplicationCacheEventMap {\n    "cached": Event;\n    "checking": Event;\n    "downloading": Event;\n    "error": Event;\n    "noupdate": Event;\n    "obsolete": Event;\n    "progress": ProgressEvent<ApplicationCache>;\n    "updateready": Event;\n}\n\ninterface ApplicationCache extends EventTarget {\n    /** @deprecated */\n    oncached: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onchecking: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    ondownloading: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onerror: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onnoupdate: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onobsolete: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    onprogress: ((this: ApplicationCache, ev: ProgressEvent<ApplicationCache>) => any) | null;\n    /** @deprecated */\n    onupdateready: ((this: ApplicationCache, ev: Event) => any) | null;\n    /** @deprecated */\n    readonly status: number;\n    /** @deprecated */\n    abort(): void;\n    /** @deprecated */\n    swapCache(): void;\n    /** @deprecated */\n    update(): void;\n    readonly CHECKING: number;\n    readonly DOWNLOADING: number;\n    readonly IDLE: number;\n    readonly OBSOLETE: number;\n    readonly UNCACHED: number;\n    readonly UPDATEREADY: number;\n    addEventListener<K extends keyof ApplicationCacheEventMap>(type: K, listener: (this: ApplicationCache, ev: ApplicationCacheEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ApplicationCacheEventMap>(type: K, listener: (this: ApplicationCache, ev: ApplicationCacheEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ApplicationCache: {\n    prototype: ApplicationCache;\n    new(): ApplicationCache;\n    readonly CHECKING: number;\n    readonly DOWNLOADING: number;\n    readonly IDLE: number;\n    readonly OBSOLETE: number;\n    readonly UNCACHED: number;\n    readonly UPDATEREADY: number;\n};\n\n/** A DOM element\'s attribute as an object. In most DOM methods, you will probably directly retrieve the attribute as a string (e.g., Element.getAttribute(), but certain functions (e.g., Element.getAttributeNode()) or means of iterating give Attr types. */\ninterface Attr extends Node {\n    readonly localName: string;\n    readonly name: string;\n    readonly namespaceURI: string | null;\n    readonly ownerDocument: Document;\n    readonly ownerElement: Element | null;\n    readonly prefix: string | null;\n    readonly specified: boolean;\n    value: string;\n}\n\ndeclare var Attr: {\n    prototype: Attr;\n    new(): Attr;\n};\n\n/** A short audio asset residing in memory, created from an audio file using the AudioContext.decodeAudioData() method, or from raw data using AudioContext.createBuffer(). Once put into an AudioBuffer, the audio can then be played by being passed into an AudioBufferSourceNode. */\ninterface AudioBuffer {\n    readonly duration: number;\n    readonly length: number;\n    readonly numberOfChannels: number;\n    readonly sampleRate: number;\n    copyFromChannel(destination: Float32Array, channelNumber: number, bufferOffset?: number): void;\n    copyToChannel(source: Float32Array, channelNumber: number, bufferOffset?: number): void;\n    getChannelData(channel: number): Float32Array;\n}\n\ndeclare var AudioBuffer: {\n    prototype: AudioBuffer;\n    new(options: AudioBufferOptions): AudioBuffer;\n};\n\n/** An AudioScheduledSourceNode which represents an audio source consisting of in-memory audio data, stored in an AudioBuffer. It\'s especially useful for playing back audio which has particularly stringent timing accuracy requirements, such as for sounds that must match a specific rhythm and can be kept in memory rather than being played from disk or the network. */\ninterface AudioBufferSourceNode extends AudioScheduledSourceNode {\n    buffer: AudioBuffer | null;\n    readonly detune: AudioParam;\n    loop: boolean;\n    loopEnd: number;\n    loopStart: number;\n    readonly playbackRate: AudioParam;\n    start(when?: number, offset?: number, duration?: number): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioBufferSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioBufferSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioBufferSourceNode: {\n    prototype: AudioBufferSourceNode;\n    new(context: BaseAudioContext, options?: AudioBufferSourceOptions): AudioBufferSourceNode;\n};\n\n/** An audio-processing graph built from audio modules linked together, each represented by an AudioNode. */\ninterface AudioContext extends BaseAudioContext {\n    readonly baseLatency: number;\n    readonly outputLatency: number;\n    close(): Promise<void>;\n    createMediaElementSource(mediaElement: HTMLMediaElement): MediaElementAudioSourceNode;\n    createMediaStreamDestination(): MediaStreamAudioDestinationNode;\n    createMediaStreamSource(mediaStream: MediaStream): MediaStreamAudioSourceNode;\n    createMediaStreamTrackSource(mediaStreamTrack: MediaStreamTrack): MediaStreamTrackAudioSourceNode;\n    getOutputTimestamp(): AudioTimestamp;\n    resume(): Promise<void>;\n    suspend(): Promise<void>;\n    addEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: AudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: AudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioContext: {\n    prototype: AudioContext;\n    new(contextOptions?: AudioContextOptions): AudioContext;\n};\n\n/** AudioDestinationNode has no output (as it is the output, no more AudioNode can be linked after it in the audio graph) and one input. The number of channels in the input must be between 0 and the maxChannelCount value or an exception is raised. */\ninterface AudioDestinationNode extends AudioNode {\n    readonly maxChannelCount: number;\n}\n\ndeclare var AudioDestinationNode: {\n    prototype: AudioDestinationNode;\n    new(): AudioDestinationNode;\n};\n\n/** The position and orientation of the unique person listening to the audio scene, and is used in audio spatialization. All PannerNodes spatialize in relation to the AudioListener stored in the BaseAudioContext.listener attribute. */\ninterface AudioListener {\n    readonly forwardX: AudioParam;\n    readonly forwardY: AudioParam;\n    readonly forwardZ: AudioParam;\n    readonly positionX: AudioParam;\n    readonly positionY: AudioParam;\n    readonly positionZ: AudioParam;\n    readonly upX: AudioParam;\n    readonly upY: AudioParam;\n    readonly upZ: AudioParam;\n    /** @deprecated */\n    setOrientation(x: number, y: number, z: number, xUp: number, yUp: number, zUp: number): void;\n    /** @deprecated */\n    setPosition(x: number, y: number, z: number): void;\n}\n\ndeclare var AudioListener: {\n    prototype: AudioListener;\n    new(): AudioListener;\n};\n\n/** A generic interface for representing an audio processing module. Examples include: */\ninterface AudioNode extends EventTarget {\n    channelCount: number;\n    channelCountMode: ChannelCountMode;\n    channelInterpretation: ChannelInterpretation;\n    readonly context: BaseAudioContext;\n    readonly numberOfInputs: number;\n    readonly numberOfOutputs: number;\n    connect(destinationNode: AudioNode, output?: number, input?: number): AudioNode;\n    connect(destinationParam: AudioParam, output?: number): void;\n    disconnect(): void;\n    disconnect(output: number): void;\n    disconnect(destinationNode: AudioNode): void;\n    disconnect(destinationNode: AudioNode, output: number): void;\n    disconnect(destinationNode: AudioNode, output: number, input: number): void;\n    disconnect(destinationParam: AudioParam): void;\n    disconnect(destinationParam: AudioParam, output: number): void;\n}\n\ndeclare var AudioNode: {\n    prototype: AudioNode;\n    new(): AudioNode;\n};\n\n/** The Web Audio API\'s AudioParam interface represents an audio-related parameter, usually a parameter of an AudioNode (such as GainNode.gain). */\ninterface AudioParam {\n    automationRate: AutomationRate;\n    readonly defaultValue: number;\n    readonly maxValue: number;\n    readonly minValue: number;\n    value: number;\n    cancelAndHoldAtTime(cancelTime: number): AudioParam;\n    cancelScheduledValues(cancelTime: number): AudioParam;\n    exponentialRampToValueAtTime(value: number, endTime: number): AudioParam;\n    linearRampToValueAtTime(value: number, endTime: number): AudioParam;\n    setTargetAtTime(target: number, startTime: number, timeConstant: number): AudioParam;\n    setValueAtTime(value: number, startTime: number): AudioParam;\n    setValueCurveAtTime(values: number[] | Float32Array, startTime: number, duration: number): AudioParam;\n}\n\ndeclare var AudioParam: {\n    prototype: AudioParam;\n    new(): AudioParam;\n};\n\ninterface AudioParamMap {\n    forEach(callbackfn: (value: AudioParam, key: string, parent: AudioParamMap) => void, thisArg?: any): void;\n}\n\ndeclare var AudioParamMap: {\n    prototype: AudioParamMap;\n    new(): AudioParamMap;\n};\n\n/** The Web Audio API events that occur when a ScriptProcessorNode input buffer is ready to be processed. */\ninterface AudioProcessingEvent extends Event {\n    readonly inputBuffer: AudioBuffer;\n    readonly outputBuffer: AudioBuffer;\n    readonly playbackTime: number;\n}\n\ndeclare var AudioProcessingEvent: {\n    prototype: AudioProcessingEvent;\n    new(type: string, eventInitDict: AudioProcessingEventInit): AudioProcessingEvent;\n};\n\ninterface AudioScheduledSourceNodeEventMap {\n    "ended": Event;\n}\n\ninterface AudioScheduledSourceNode extends AudioNode {\n    onended: ((this: AudioScheduledSourceNode, ev: Event) => any) | null;\n    start(when?: number): void;\n    stop(when?: number): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioScheduledSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: AudioScheduledSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioScheduledSourceNode: {\n    prototype: AudioScheduledSourceNode;\n    new(): AudioScheduledSourceNode;\n};\n\ninterface AudioWorklet extends Worklet {\n}\n\ndeclare var AudioWorklet: {\n    prototype: AudioWorklet;\n    new(): AudioWorklet;\n};\n\ninterface AudioWorkletNodeEventMap {\n    "processorerror": Event;\n}\n\ninterface AudioWorkletNode extends AudioNode {\n    onprocessorerror: ((this: AudioWorkletNode, ev: Event) => any) | null;\n    readonly parameters: AudioParamMap;\n    readonly port: MessagePort;\n    addEventListener<K extends keyof AudioWorkletNodeEventMap>(type: K, listener: (this: AudioWorkletNode, ev: AudioWorkletNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioWorkletNodeEventMap>(type: K, listener: (this: AudioWorkletNode, ev: AudioWorkletNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AudioWorkletNode: {\n    prototype: AudioWorkletNode;\n    new(context: BaseAudioContext, name: string, options?: AudioWorkletNodeOptions): AudioWorkletNode;\n};\n\ninterface AuthenticatorAssertionResponse extends AuthenticatorResponse {\n    readonly authenticatorData: ArrayBuffer;\n    readonly signature: ArrayBuffer;\n    readonly userHandle: ArrayBuffer | null;\n}\n\ndeclare var AuthenticatorAssertionResponse: {\n    prototype: AuthenticatorAssertionResponse;\n    new(): AuthenticatorAssertionResponse;\n};\n\ninterface AuthenticatorAttestationResponse extends AuthenticatorResponse {\n    readonly attestationObject: ArrayBuffer;\n}\n\ndeclare var AuthenticatorAttestationResponse: {\n    prototype: AuthenticatorAttestationResponse;\n    new(): AuthenticatorAttestationResponse;\n};\n\ninterface AuthenticatorResponse {\n    readonly clientDataJSON: ArrayBuffer;\n}\n\ndeclare var AuthenticatorResponse: {\n    prototype: AuthenticatorResponse;\n    new(): AuthenticatorResponse;\n};\n\ninterface BarProp {\n    readonly visible: boolean;\n}\n\ndeclare var BarProp: {\n    prototype: BarProp;\n    new(): BarProp;\n};\n\ninterface BaseAudioContextEventMap {\n    "statechange": Event;\n}\n\ninterface BaseAudioContext extends EventTarget {\n    readonly audioWorklet: AudioWorklet;\n    readonly currentTime: number;\n    readonly destination: AudioDestinationNode;\n    readonly listener: AudioListener;\n    onstatechange: ((this: BaseAudioContext, ev: Event) => any) | null;\n    readonly sampleRate: number;\n    readonly state: AudioContextState;\n    createAnalyser(): AnalyserNode;\n    createBiquadFilter(): BiquadFilterNode;\n    createBuffer(numberOfChannels: number, length: number, sampleRate: number): AudioBuffer;\n    createBufferSource(): AudioBufferSourceNode;\n    createChannelMerger(numberOfInputs?: number): ChannelMergerNode;\n    createChannelSplitter(numberOfOutputs?: number): ChannelSplitterNode;\n    createConstantSource(): ConstantSourceNode;\n    createConvolver(): ConvolverNode;\n    createDelay(maxDelayTime?: number): DelayNode;\n    createDynamicsCompressor(): DynamicsCompressorNode;\n    createGain(): GainNode;\n    createIIRFilter(feedforward: number[], feedback: number[]): IIRFilterNode;\n    createOscillator(): OscillatorNode;\n    createPanner(): PannerNode;\n    createPeriodicWave(real: number[] | Float32Array, imag: number[] | Float32Array, constraints?: PeriodicWaveConstraints): PeriodicWave;\n    createScriptProcessor(bufferSize?: number, numberOfInputChannels?: number, numberOfOutputChannels?: number): ScriptProcessorNode;\n    createStereoPanner(): StereoPannerNode;\n    createWaveShaper(): WaveShaperNode;\n    decodeAudioData(audioData: ArrayBuffer, successCallback?: DecodeSuccessCallback | null, errorCallback?: DecodeErrorCallback | null): Promise<AudioBuffer>;\n    addEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: BaseAudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BaseAudioContextEventMap>(type: K, listener: (this: BaseAudioContext, ev: BaseAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var BaseAudioContext: {\n    prototype: BaseAudioContext;\n    new(): BaseAudioContext;\n};\n\n/** The beforeunload event is fired when the window, the document and its resources are about to be unloaded. */\ninterface BeforeUnloadEvent extends Event {\n    returnValue: any;\n}\n\ndeclare var BeforeUnloadEvent: {\n    prototype: BeforeUnloadEvent;\n    new(): BeforeUnloadEvent;\n};\n\ninterface BhxBrowser {\n    readonly lastError: DOMException;\n    checkMatchesGlobExpression(pattern: string, value: string): boolean;\n    checkMatchesUriExpression(pattern: string, value: string): boolean;\n    clearLastError(): void;\n    currentWindowId(): number;\n    fireExtensionApiTelemetry(functionName: string, isSucceeded: boolean, isSupported: boolean, errorString: string): void;\n    genericFunction(functionId: number, destination: any, parameters?: string, callbackId?: number): void;\n    genericSynchronousFunction(functionId: number, parameters?: string): string;\n    getExtensionId(): string;\n    getThisAddress(): any;\n    registerGenericFunctionCallbackHandler(callbackHandler: Function): void;\n    registerGenericListenerHandler(eventHandler: Function): void;\n    setLastError(parameters: string): void;\n    webPlatformGenericFunction(destination: any, parameters?: string, callbackId?: number): void;\n}\n\ndeclare var BhxBrowser: {\n    prototype: BhxBrowser;\n    new(): BhxBrowser;\n};\n\n/** A simple low-order filter, and is created using the AudioContext.createBiquadFilter() method. It is an AudioNode that can represent different kinds of filters, tone control devices, and graphic equalizers. */\ninterface BiquadFilterNode extends AudioNode {\n    readonly Q: AudioParam;\n    readonly detune: AudioParam;\n    readonly frequency: AudioParam;\n    readonly gain: AudioParam;\n    type: BiquadFilterType;\n    getFrequencyResponse(frequencyHz: Float32Array, magResponse: Float32Array, phaseResponse: Float32Array): void;\n}\n\ndeclare var BiquadFilterNode: {\n    prototype: BiquadFilterNode;\n    new(context: BaseAudioContext, options?: BiquadFilterOptions): BiquadFilterNode;\n};\n\n/** A file-like object of immutable, raw data. Blobs represent data that isn\'t necessarily in a JavaScript-native format. The File interface is based on Blob, inheriting blob functionality and expanding it to support files on the user\'s system. */\ninterface Blob {\n    readonly size: number;\n    readonly type: string;\n    arrayBuffer(): Promise<ArrayBuffer>;\n    slice(start?: number, end?: number, contentType?: string): Blob;\n    stream(): ReadableStream;\n    text(): Promise<string>;\n}\n\ndeclare var Blob: {\n    prototype: Blob;\n    new(blobParts?: BlobPart[], options?: BlobPropertyBag): Blob;\n};\n\ninterface Body {\n    readonly body: ReadableStream<Uint8Array> | null;\n    readonly bodyUsed: boolean;\n    arrayBuffer(): Promise<ArrayBuffer>;\n    blob(): Promise<Blob>;\n    formData(): Promise<FormData>;\n    json(): Promise<any>;\n    text(): Promise<string>;\n}\n\ninterface BroadcastChannelEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\ninterface BroadcastChannel extends EventTarget {\n    /**\n     * Returns the channel name (as passed to the constructor).\n     */\n    readonly name: string;\n    onmessage: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    /**\n     * Closes the BroadcastChannel object, opening it up to garbage collection.\n     */\n    close(): void;\n    /**\n     * Sends the given message to other BroadcastChannel objects set up for this channel. Messages can be structured objects, e.g. nested objects and arrays.\n     */\n    postMessage(message: any): void;\n    addEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var BroadcastChannel: {\n    prototype: BroadcastChannel;\n    new(name: string): BroadcastChannel;\n};\n\n/** This Streams API interface provides a built-in byte length queuing strategy that can be used when constructing streams. */\ninterface ByteLengthQueuingStrategy extends QueuingStrategy<ArrayBufferView> {\n    highWaterMark: number;\n    size(chunk: ArrayBufferView): number;\n}\n\ndeclare var ByteLengthQueuingStrategy: {\n    prototype: ByteLengthQueuingStrategy;\n    new(options: { highWaterMark: number }): ByteLengthQueuingStrategy;\n};\n\n/** A CDATA section that can be used within XML to include extended portions of unescaped text. The symbols < and & don’t need escaping as they normally do when inside a CDATA section. */\ninterface CDATASection extends Text {\n}\n\ndeclare var CDATASection: {\n    prototype: CDATASection;\n    new(): CDATASection;\n};\n\n/** A single condition CSS at-rule, which consists of a condition and a statement block. It is a child of CSSGroupingRule. */\ninterface CSSConditionRule extends CSSGroupingRule {\n    conditionText: string;\n}\n\ndeclare var CSSConditionRule: {\n    prototype: CSSConditionRule;\n    new(): CSSConditionRule;\n};\n\ninterface CSSFontFaceRule extends CSSRule {\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSFontFaceRule: {\n    prototype: CSSFontFaceRule;\n    new(): CSSFontFaceRule;\n};\n\n/** Any CSS at-rule that contains other rules nested within it. */\ninterface CSSGroupingRule extends CSSRule {\n    readonly cssRules: CSSRuleList;\n    deleteRule(index: number): void;\n    insertRule(rule: string, index?: number): number;\n}\n\ndeclare var CSSGroupingRule: {\n    prototype: CSSGroupingRule;\n    new(): CSSGroupingRule;\n};\n\ninterface CSSImportRule extends CSSRule {\n    readonly href: string;\n    readonly media: MediaList;\n    readonly styleSheet: CSSStyleSheet;\n}\n\ndeclare var CSSImportRule: {\n    prototype: CSSImportRule;\n    new(): CSSImportRule;\n};\n\n/** An object representing a set of style for a given keyframe. It corresponds to the contains of a single keyframe of a @keyframes at-rule. It implements the CSSRule interface with a type value of 8 (CSSRule.KEYFRAME_RULE). */\ninterface CSSKeyframeRule extends CSSRule {\n    keyText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSKeyframeRule: {\n    prototype: CSSKeyframeRule;\n    new(): CSSKeyframeRule;\n};\n\n/** An object representing a complete set of keyframes for a CSS animation. It corresponds to the contains of a whole @keyframes at-rule. It implements the CSSRule interface with a type value of 7 (CSSRule.KEYFRAMES_RULE). */\ninterface CSSKeyframesRule extends CSSRule {\n    readonly cssRules: CSSRuleList;\n    name: string;\n    appendRule(rule: string): void;\n    deleteRule(select: string): void;\n    findRule(select: string): CSSKeyframeRule | null;\n}\n\ndeclare var CSSKeyframesRule: {\n    prototype: CSSKeyframesRule;\n    new(): CSSKeyframesRule;\n};\n\n/** A single CSS @media rule. It implements the CSSConditionRule interface, and therefore the CSSGroupingRule and the CSSRule interface with a type value of 4 (CSSRule.MEDIA_RULE). */\ninterface CSSMediaRule extends CSSConditionRule {\n    readonly media: MediaList;\n}\n\ndeclare var CSSMediaRule: {\n    prototype: CSSMediaRule;\n    new(): CSSMediaRule;\n};\n\n/** An object representing a single CSS @namespace at-rule. It implements the CSSRule interface, with a type value of 10 (CSSRule.NAMESPACE_RULE). */\ninterface CSSNamespaceRule extends CSSRule {\n    readonly namespaceURI: string;\n    readonly prefix: string;\n}\n\ndeclare var CSSNamespaceRule: {\n    prototype: CSSNamespaceRule;\n    new(): CSSNamespaceRule;\n};\n\n/** CSSPageRule is an interface representing a single CSS @page rule. It implements the CSSRule interface with a type value of 6 (CSSRule.PAGE_RULE). */\ninterface CSSPageRule extends CSSGroupingRule {\n    selectorText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSPageRule: {\n    prototype: CSSPageRule;\n    new(): CSSPageRule;\n};\n\n/** A single CSS rule. There are several types of rules, listed in the Type constants section below. */\ninterface CSSRule {\n    cssText: string;\n    readonly parentRule: CSSRule | null;\n    readonly parentStyleSheet: CSSStyleSheet | null;\n    readonly type: number;\n    readonly CHARSET_RULE: number;\n    readonly FONT_FACE_RULE: number;\n    readonly IMPORT_RULE: number;\n    readonly KEYFRAMES_RULE: number;\n    readonly KEYFRAME_RULE: number;\n    readonly MEDIA_RULE: number;\n    readonly NAMESPACE_RULE: number;\n    readonly PAGE_RULE: number;\n    readonly STYLE_RULE: number;\n    readonly SUPPORTS_RULE: number;\n}\n\ndeclare var CSSRule: {\n    prototype: CSSRule;\n    new(): CSSRule;\n    readonly CHARSET_RULE: number;\n    readonly FONT_FACE_RULE: number;\n    readonly IMPORT_RULE: number;\n    readonly KEYFRAMES_RULE: number;\n    readonly KEYFRAME_RULE: number;\n    readonly MEDIA_RULE: number;\n    readonly NAMESPACE_RULE: number;\n    readonly PAGE_RULE: number;\n    readonly STYLE_RULE: number;\n    readonly SUPPORTS_RULE: number;\n};\n\n/** A CSSRuleList is an (indirect-modify only) array-like object containing an ordered collection of CSSRule objects. */\ninterface CSSRuleList {\n    readonly length: number;\n    item(index: number): CSSRule | null;\n    [index: number]: CSSRule;\n}\n\ndeclare var CSSRuleList: {\n    prototype: CSSRuleList;\n    new(): CSSRuleList;\n};\n\n/** An object that is a CSS declaration block, and exposes style information and various style-related methods and properties. */\ninterface CSSStyleDeclaration {\n    alignContent: string;\n    alignItems: string;\n    alignSelf: string;\n    alignmentBaseline: string;\n    all: string;\n    animation: string;\n    animationDelay: string;\n    animationDirection: string;\n    animationDuration: string;\n    animationFillMode: string;\n    animationIterationCount: string;\n    animationName: string;\n    animationPlayState: string;\n    animationTimingFunction: string;\n    backfaceVisibility: string;\n    background: string;\n    backgroundAttachment: string;\n    backgroundClip: string;\n    backgroundColor: string;\n    backgroundImage: string;\n    backgroundOrigin: string;\n    backgroundPosition: string;\n    backgroundPositionX: string;\n    backgroundPositionY: string;\n    backgroundRepeat: string;\n    backgroundSize: string;\n    baselineShift: string;\n    blockSize: string;\n    border: string;\n    borderBlockEnd: string;\n    borderBlockEndColor: string;\n    borderBlockEndStyle: string;\n    borderBlockEndWidth: string;\n    borderBlockStart: string;\n    borderBlockStartColor: string;\n    borderBlockStartStyle: string;\n    borderBlockStartWidth: string;\n    borderBottom: string;\n    borderBottomColor: string;\n    borderBottomLeftRadius: string;\n    borderBottomRightRadius: string;\n    borderBottomStyle: string;\n    borderBottomWidth: string;\n    borderCollapse: string;\n    borderColor: string;\n    borderImage: string;\n    borderImageOutset: string;\n    borderImageRepeat: string;\n    borderImageSlice: string;\n    borderImageSource: string;\n    borderImageWidth: string;\n    borderInlineEnd: string;\n    borderInlineEndColor: string;\n    borderInlineEndStyle: string;\n    borderInlineEndWidth: string;\n    borderInlineStart: string;\n    borderInlineStartColor: string;\n    borderInlineStartStyle: string;\n    borderInlineStartWidth: string;\n    borderLeft: string;\n    borderLeftColor: string;\n    borderLeftStyle: string;\n    borderLeftWidth: string;\n    borderRadius: string;\n    borderRight: string;\n    borderRightColor: string;\n    borderRightStyle: string;\n    borderRightWidth: string;\n    borderSpacing: string;\n    borderStyle: string;\n    borderTop: string;\n    borderTopColor: string;\n    borderTopLeftRadius: string;\n    borderTopRightRadius: string;\n    borderTopStyle: string;\n    borderTopWidth: string;\n    borderWidth: string;\n    bottom: string;\n    boxShadow: string;\n    boxSizing: string;\n    breakAfter: string;\n    breakBefore: string;\n    breakInside: string;\n    captionSide: string;\n    caretColor: string;\n    clear: string;\n    clip: string;\n    clipPath: string;\n    clipRule: string;\n    color: string;\n    colorInterpolation: string;\n    colorInterpolationFilters: string;\n    columnCount: string;\n    columnFill: string;\n    columnGap: string;\n    columnRule: string;\n    columnRuleColor: string;\n    columnRuleStyle: string;\n    columnRuleWidth: string;\n    columnSpan: string;\n    columnWidth: string;\n    columns: string;\n    content: string;\n    counterIncrement: string;\n    counterReset: string;\n    cssFloat: string;\n    cssText: string;\n    cursor: string;\n    direction: string;\n    display: string;\n    dominantBaseline: string;\n    emptyCells: string;\n    fill: string;\n    fillOpacity: string;\n    fillRule: string;\n    filter: string;\n    flex: string;\n    flexBasis: string;\n    flexDirection: string;\n    flexFlow: string;\n    flexGrow: string;\n    flexShrink: string;\n    flexWrap: string;\n    float: string;\n    floodColor: string;\n    floodOpacity: string;\n    font: string;\n    fontFamily: string;\n    fontFeatureSettings: string;\n    fontKerning: string;\n    fontSize: string;\n    fontSizeAdjust: string;\n    fontStretch: string;\n    fontStyle: string;\n    fontSynthesis: string;\n    fontVariant: string;\n    fontVariantCaps: string;\n    fontVariantEastAsian: string;\n    fontVariantLigatures: string;\n    fontVariantNumeric: string;\n    fontVariantPosition: string;\n    fontWeight: string;\n    gap: string;\n    glyphOrientationVertical: string;\n    grid: string;\n    gridArea: string;\n    gridAutoColumns: string;\n    gridAutoFlow: string;\n    gridAutoRows: string;\n    gridColumn: string;\n    gridColumnEnd: string;\n    gridColumnGap: string;\n    gridColumnStart: string;\n    gridGap: string;\n    gridRow: string;\n    gridRowEnd: string;\n    gridRowGap: string;\n    gridRowStart: string;\n    gridTemplate: string;\n    gridTemplateAreas: string;\n    gridTemplateColumns: string;\n    gridTemplateRows: string;\n    height: string;\n    hyphens: string;\n    imageOrientation: string;\n    imageRendering: string;\n    inlineSize: string;\n    justifyContent: string;\n    justifyItems: string;\n    justifySelf: string;\n    left: string;\n    readonly length: number;\n    letterSpacing: string;\n    lightingColor: string;\n    lineBreak: string;\n    lineHeight: string;\n    listStyle: string;\n    listStyleImage: string;\n    listStylePosition: string;\n    listStyleType: string;\n    margin: string;\n    marginBlockEnd: string;\n    marginBlockStart: string;\n    marginBottom: string;\n    marginInlineEnd: string;\n    marginInlineStart: string;\n    marginLeft: string;\n    marginRight: string;\n    marginTop: string;\n    marker: string;\n    markerEnd: string;\n    markerMid: string;\n    markerStart: string;\n    mask: string;\n    maskComposite: string;\n    maskImage: string;\n    maskPosition: string;\n    maskRepeat: string;\n    maskSize: string;\n    maskType: string;\n    maxBlockSize: string;\n    maxHeight: string;\n    maxInlineSize: string;\n    maxWidth: string;\n    minBlockSize: string;\n    minHeight: string;\n    minInlineSize: string;\n    minWidth: string;\n    objectFit: string;\n    objectPosition: string;\n    opacity: string;\n    order: string;\n    orphans: string;\n    outline: string;\n    outlineColor: string;\n    outlineOffset: string;\n    outlineStyle: string;\n    outlineWidth: string;\n    overflow: string;\n    overflowAnchor: string;\n    overflowWrap: string;\n    overflowX: string;\n    overflowY: string;\n    overscrollBehavior: string;\n    overscrollBehaviorBlock: string;\n    overscrollBehaviorInline: string;\n    overscrollBehaviorX: string;\n    overscrollBehaviorY: string;\n    padding: string;\n    paddingBlockEnd: string;\n    paddingBlockStart: string;\n    paddingBottom: string;\n    paddingInlineEnd: string;\n    paddingInlineStart: string;\n    paddingLeft: string;\n    paddingRight: string;\n    paddingTop: string;\n    pageBreakAfter: string;\n    pageBreakBefore: string;\n    pageBreakInside: string;\n    paintOrder: string;\n    readonly parentRule: CSSRule | null;\n    perspective: string;\n    perspectiveOrigin: string;\n    placeContent: string;\n    placeItems: string;\n    placeSelf: string;\n    pointerEvents: string;\n    position: string;\n    quotes: string;\n    resize: string;\n    right: string;\n    rotate: string;\n    rowGap: string;\n    rubyAlign: string;\n    rubyPosition: string;\n    scale: string;\n    scrollBehavior: string;\n    shapeRendering: string;\n    stopColor: string;\n    stopOpacity: string;\n    stroke: string;\n    strokeDasharray: string;\n    strokeDashoffset: string;\n    strokeLinecap: string;\n    strokeLinejoin: string;\n    strokeMiterlimit: string;\n    strokeOpacity: string;\n    strokeWidth: string;\n    tabSize: string;\n    tableLayout: string;\n    textAlign: string;\n    textAlignLast: string;\n    textAnchor: string;\n    textCombineUpright: string;\n    textDecoration: string;\n    textDecorationColor: string;\n    textDecorationLine: string;\n    textDecorationStyle: string;\n    textEmphasis: string;\n    textEmphasisColor: string;\n    textEmphasisPosition: string;\n    textEmphasisStyle: string;\n    textIndent: string;\n    textJustify: string;\n    textOrientation: string;\n    textOverflow: string;\n    textRendering: string;\n    textShadow: string;\n    textTransform: string;\n    textUnderlinePosition: string;\n    top: string;\n    touchAction: string;\n    transform: string;\n    transformBox: string;\n    transformOrigin: string;\n    transformStyle: string;\n    transition: string;\n    transitionDelay: string;\n    transitionDuration: string;\n    transitionProperty: string;\n    transitionTimingFunction: string;\n    translate: string;\n    unicodeBidi: string;\n    userSelect: string;\n    verticalAlign: string;\n    visibility: string;\n    /** @deprecated */\n    webkitAlignContent: string;\n    /** @deprecated */\n    webkitAlignItems: string;\n    /** @deprecated */\n    webkitAlignSelf: string;\n    /** @deprecated */\n    webkitAnimation: string;\n    /** @deprecated */\n    webkitAnimationDelay: string;\n    /** @deprecated */\n    webkitAnimationDirection: string;\n    /** @deprecated */\n    webkitAnimationDuration: string;\n    /** @deprecated */\n    webkitAnimationFillMode: string;\n    /** @deprecated */\n    webkitAnimationIterationCount: string;\n    /** @deprecated */\n    webkitAnimationName: string;\n    /** @deprecated */\n    webkitAnimationPlayState: string;\n    /** @deprecated */\n    webkitAnimationTimingFunction: string;\n    /** @deprecated */\n    webkitAppearance: string;\n    /** @deprecated */\n    webkitBackfaceVisibility: string;\n    /** @deprecated */\n    webkitBackgroundClip: string;\n    /** @deprecated */\n    webkitBackgroundOrigin: string;\n    /** @deprecated */\n    webkitBackgroundSize: string;\n    /** @deprecated */\n    webkitBorderBottomLeftRadius: string;\n    /** @deprecated */\n    webkitBorderBottomRightRadius: string;\n    /** @deprecated */\n    webkitBorderRadius: string;\n    /** @deprecated */\n    webkitBorderTopLeftRadius: string;\n    /** @deprecated */\n    webkitBorderTopRightRadius: string;\n    /** @deprecated */\n    webkitBoxAlign: string;\n    /** @deprecated */\n    webkitBoxFlex: string;\n    /** @deprecated */\n    webkitBoxOrdinalGroup: string;\n    /** @deprecated */\n    webkitBoxOrient: string;\n    /** @deprecated */\n    webkitBoxPack: string;\n    /** @deprecated */\n    webkitBoxShadow: string;\n    /** @deprecated */\n    webkitBoxSizing: string;\n    /** @deprecated */\n    webkitFilter: string;\n    /** @deprecated */\n    webkitFlex: string;\n    /** @deprecated */\n    webkitFlexBasis: string;\n    /** @deprecated */\n    webkitFlexDirection: string;\n    /** @deprecated */\n    webkitFlexFlow: string;\n    /** @deprecated */\n    webkitFlexGrow: string;\n    /** @deprecated */\n    webkitFlexShrink: string;\n    /** @deprecated */\n    webkitFlexWrap: string;\n    /** @deprecated */\n    webkitJustifyContent: string;\n    webkitLineClamp: string;\n    /** @deprecated */\n    webkitMask: string;\n    /** @deprecated */\n    webkitMaskBoxImage: string;\n    /** @deprecated */\n    webkitMaskBoxImageOutset: string;\n    /** @deprecated */\n    webkitMaskBoxImageRepeat: string;\n    /** @deprecated */\n    webkitMaskBoxImageSlice: string;\n    /** @deprecated */\n    webkitMaskBoxImageSource: string;\n    /** @deprecated */\n    webkitMaskBoxImageWidth: string;\n    /** @deprecated */\n    webkitMaskClip: string;\n    /** @deprecated */\n    webkitMaskComposite: string;\n    /** @deprecated */\n    webkitMaskImage: string;\n    /** @deprecated */\n    webkitMaskOrigin: string;\n    /** @deprecated */\n    webkitMaskPosition: string;\n    /** @deprecated */\n    webkitMaskRepeat: string;\n    /** @deprecated */\n    webkitMaskSize: string;\n    /** @deprecated */\n    webkitOrder: string;\n    /** @deprecated */\n    webkitPerspective: string;\n    /** @deprecated */\n    webkitPerspectiveOrigin: string;\n    webkitTapHighlightColor: string;\n    /** @deprecated */\n    webkitTextFillColor: string;\n    /** @deprecated */\n    webkitTextSizeAdjust: string;\n    /** @deprecated */\n    webkitTextStroke: string;\n    /** @deprecated */\n    webkitTextStrokeColor: string;\n    /** @deprecated */\n    webkitTextStrokeWidth: string;\n    /** @deprecated */\n    webkitTransform: string;\n    /** @deprecated */\n    webkitTransformOrigin: string;\n    /** @deprecated */\n    webkitTransformStyle: string;\n    /** @deprecated */\n    webkitTransition: string;\n    /** @deprecated */\n    webkitTransitionDelay: string;\n    /** @deprecated */\n    webkitTransitionDuration: string;\n    /** @deprecated */\n    webkitTransitionProperty: string;\n    /** @deprecated */\n    webkitTransitionTimingFunction: string;\n    /** @deprecated */\n    webkitUserSelect: string;\n    whiteSpace: string;\n    widows: string;\n    width: string;\n    willChange: string;\n    wordBreak: string;\n    wordSpacing: string;\n    wordWrap: string;\n    writingMode: string;\n    zIndex: string;\n    /** @deprecated */\n    zoom: string;\n    getPropertyPriority(property: string): string;\n    getPropertyValue(property: string): string;\n    item(index: number): string;\n    removeProperty(property: string): string;\n    setProperty(property: string, value: string | null, priority?: string): void;\n    [index: number]: string;\n}\n\ndeclare var CSSStyleDeclaration: {\n    prototype: CSSStyleDeclaration;\n    new(): CSSStyleDeclaration;\n};\n\n/** CSSStyleRule represents a single CSS style rule. It implements the CSSRule interface with a type value of 1 (CSSRule.STYLE_RULE). */\ninterface CSSStyleRule extends CSSRule {\n    selectorText: string;\n    readonly style: CSSStyleDeclaration;\n}\n\ndeclare var CSSStyleRule: {\n    prototype: CSSStyleRule;\n    new(): CSSStyleRule;\n};\n\n/** A single CSS style sheet. It inherits properties and methods from its parent, StyleSheet. */\ninterface CSSStyleSheet extends StyleSheet {\n    readonly cssRules: CSSRuleList;\n    readonly ownerRule: CSSRule | null;\n    readonly rules: CSSRuleList;\n    addRule(selector?: string, style?: string, index?: number): number;\n    deleteRule(index: number): void;\n    insertRule(rule: string, index?: number): number;\n    removeRule(index?: number): void;\n}\n\ndeclare var CSSStyleSheet: {\n    prototype: CSSStyleSheet;\n    new(): CSSStyleSheet;\n};\n\n/** An object representing a single CSS @supports at-rule. It implements the CSSConditionRule interface, and therefore the CSSRule and CSSGroupingRule interfaces with a type value of 12 (CSSRule.SUPPORTS_RULE). */\ninterface CSSSupportsRule extends CSSConditionRule {\n}\n\ndeclare var CSSSupportsRule: {\n    prototype: CSSSupportsRule;\n    new(): CSSSupportsRule;\n};\n\n/** Provides a storage mechanism for Request / Response object pairs that are cached, for example as part of the ServiceWorker life cycle. Note that the Cache interface is exposed to windowed scopes as well as workers. You don\'t have to use it in conjunction with service workers, even though it is defined in the service worker spec. */\ninterface Cache {\n    add(request: RequestInfo): Promise<void>;\n    addAll(requests: RequestInfo[]): Promise<void>;\n    delete(request: RequestInfo, options?: CacheQueryOptions): Promise<boolean>;\n    keys(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Request>>;\n    match(request: RequestInfo, options?: CacheQueryOptions): Promise<Response | undefined>;\n    matchAll(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Response>>;\n    put(request: RequestInfo, response: Response): Promise<void>;\n}\n\ndeclare var Cache: {\n    prototype: Cache;\n    new(): Cache;\n};\n\n/** The storage for Cache objects. */\ninterface CacheStorage {\n    delete(cacheName: string): Promise<boolean>;\n    has(cacheName: string): Promise<boolean>;\n    keys(): Promise<string[]>;\n    match(request: RequestInfo, options?: MultiCacheQueryOptions): Promise<Response | undefined>;\n    open(cacheName: string): Promise<Cache>;\n}\n\ndeclare var CacheStorage: {\n    prototype: CacheStorage;\n    new(): CacheStorage;\n};\n\ninterface CanvasCompositing {\n    globalAlpha: number;\n    globalCompositeOperation: string;\n}\n\ninterface CanvasDrawImage {\n    drawImage(image: CanvasImageSource, dx: number, dy: number): void;\n    drawImage(image: CanvasImageSource, dx: number, dy: number, dw: number, dh: number): void;\n    drawImage(image: CanvasImageSource, sx: number, sy: number, sw: number, sh: number, dx: number, dy: number, dw: number, dh: number): void;\n}\n\ninterface CanvasDrawPath {\n    beginPath(): void;\n    clip(fillRule?: CanvasFillRule): void;\n    clip(path: Path2D, fillRule?: CanvasFillRule): void;\n    fill(fillRule?: CanvasFillRule): void;\n    fill(path: Path2D, fillRule?: CanvasFillRule): void;\n    isPointInPath(x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInPath(path: Path2D, x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInStroke(x: number, y: number): boolean;\n    isPointInStroke(path: Path2D, x: number, y: number): boolean;\n    stroke(): void;\n    stroke(path: Path2D): void;\n}\n\ninterface CanvasFillStrokeStyles {\n    fillStyle: string | CanvasGradient | CanvasPattern;\n    strokeStyle: string | CanvasGradient | CanvasPattern;\n    createLinearGradient(x0: number, y0: number, x1: number, y1: number): CanvasGradient;\n    createPattern(image: CanvasImageSource, repetition: string | null): CanvasPattern | null;\n    createRadialGradient(x0: number, y0: number, r0: number, x1: number, y1: number, r1: number): CanvasGradient;\n}\n\ninterface CanvasFilters {\n    filter: string;\n}\n\n/** An opaque object describing a gradient. It is returned by the methods CanvasRenderingContext2D.createLinearGradient() or CanvasRenderingContext2D.createRadialGradient(). */\ninterface CanvasGradient {\n    /**\n     * Adds a color stop with the given color to the gradient at the given offset. 0.0 is the offset at one end of the gradient, 1.0 is the offset at the other end.\n     * \n     * Throws an "IndexSizeError" DOMException if the offset is out of range. Throws a "SyntaxError" DOMException if the color cannot be parsed.\n     */\n    addColorStop(offset: number, color: string): void;\n}\n\ndeclare var CanvasGradient: {\n    prototype: CanvasGradient;\n    new(): CanvasGradient;\n};\n\ninterface CanvasImageData {\n    createImageData(sw: number, sh: number): ImageData;\n    createImageData(imagedata: ImageData): ImageData;\n    getImageData(sx: number, sy: number, sw: number, sh: number): ImageData;\n    putImageData(imagedata: ImageData, dx: number, dy: number): void;\n    putImageData(imagedata: ImageData, dx: number, dy: number, dirtyX: number, dirtyY: number, dirtyWidth: number, dirtyHeight: number): void;\n}\n\ninterface CanvasImageSmoothing {\n    imageSmoothingEnabled: boolean;\n    imageSmoothingQuality: ImageSmoothingQuality;\n}\n\ninterface CanvasPath {\n    arc(x: number, y: number, radius: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    arcTo(x1: number, y1: number, x2: number, y2: number, radius: number): void;\n    bezierCurveTo(cp1x: number, cp1y: number, cp2x: number, cp2y: number, x: number, y: number): void;\n    closePath(): void;\n    ellipse(x: number, y: number, radiusX: number, radiusY: number, rotation: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    lineTo(x: number, y: number): void;\n    moveTo(x: number, y: number): void;\n    quadraticCurveTo(cpx: number, cpy: number, x: number, y: number): void;\n    rect(x: number, y: number, w: number, h: number): void;\n}\n\ninterface CanvasPathDrawingStyles {\n    lineCap: CanvasLineCap;\n    lineDashOffset: number;\n    lineJoin: CanvasLineJoin;\n    lineWidth: number;\n    miterLimit: number;\n    getLineDash(): number[];\n    setLineDash(segments: number[]): void;\n}\n\n/** An opaque object describing a pattern, based on an image, a canvas, or a video, created by the CanvasRenderingContext2D.createPattern() method. */\ninterface CanvasPattern {\n    /**\n     * Sets the transformation matrix that will be used when rendering the pattern during a fill or stroke painting operation.\n     */\n    setTransform(transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var CanvasPattern: {\n    prototype: CanvasPattern;\n    new(): CanvasPattern;\n};\n\ninterface CanvasRect {\n    clearRect(x: number, y: number, w: number, h: number): void;\n    fillRect(x: number, y: number, w: number, h: number): void;\n    strokeRect(x: number, y: number, w: number, h: number): void;\n}\n\n/** The CanvasRenderingContext2D interface, part of the Canvas API, provides the 2D rendering context for the drawing surface of a <canvas> element. It is used for drawing shapes, text, images, and other objects. */\ninterface CanvasRenderingContext2D extends CanvasCompositing, CanvasDrawImage, CanvasDrawPath, CanvasFillStrokeStyles, CanvasFilters, CanvasImageData, CanvasImageSmoothing, CanvasPath, CanvasPathDrawingStyles, CanvasRect, CanvasShadowStyles, CanvasState, CanvasText, CanvasTextDrawingStyles, CanvasTransform, CanvasUserInterface {\n    readonly canvas: HTMLCanvasElement;\n}\n\ndeclare var CanvasRenderingContext2D: {\n    prototype: CanvasRenderingContext2D;\n    new(): CanvasRenderingContext2D;\n};\n\ninterface CanvasShadowStyles {\n    shadowBlur: number;\n    shadowColor: string;\n    shadowOffsetX: number;\n    shadowOffsetY: number;\n}\n\ninterface CanvasState {\n    restore(): void;\n    save(): void;\n}\n\ninterface CanvasText {\n    fillText(text: string, x: number, y: number, maxWidth?: number): void;\n    measureText(text: string): TextMetrics;\n    strokeText(text: string, x: number, y: number, maxWidth?: number): void;\n}\n\ninterface CanvasTextDrawingStyles {\n    direction: CanvasDirection;\n    font: string;\n    textAlign: CanvasTextAlign;\n    textBaseline: CanvasTextBaseline;\n}\n\ninterface CanvasTransform {\n    getTransform(): DOMMatrix;\n    resetTransform(): void;\n    rotate(angle: number): void;\n    scale(x: number, y: number): void;\n    setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    setTransform(transform?: DOMMatrix2DInit): void;\n    transform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    translate(x: number, y: number): void;\n}\n\ninterface CanvasUserInterface {\n    drawFocusIfNeeded(element: Element): void;\n    drawFocusIfNeeded(path: Path2D, element: Element): void;\n    scrollPathIntoView(): void;\n    scrollPathIntoView(path: Path2D): void;\n}\n\ninterface CaretPosition {\n    readonly offset: number;\n    readonly offsetNode: Node;\n    getClientRect(): DOMRect | null;\n}\n\ndeclare var CaretPosition: {\n    prototype: CaretPosition;\n    new(): CaretPosition;\n};\n\n/** The ChannelMergerNode interface, often used in conjunction with its opposite, ChannelSplitterNode, reunites different mono inputs into a single output. Each input is used to fill a channel of the output. This is useful for accessing each channels separately, e.g. for performing channel mixing where gain must be separately controlled on each channel. */\ninterface ChannelMergerNode extends AudioNode {\n}\n\ndeclare var ChannelMergerNode: {\n    prototype: ChannelMergerNode;\n    new(context: BaseAudioContext, options?: ChannelMergerOptions): ChannelMergerNode;\n};\n\n/** The ChannelSplitterNode interface, often used in conjunction with its opposite, ChannelMergerNode, separates the different channels of an audio source into a set of mono outputs. This is useful for accessing each channel separately, e.g. for performing channel mixing where gain must be separately controlled on each channel. */\ninterface ChannelSplitterNode extends AudioNode {\n}\n\ndeclare var ChannelSplitterNode: {\n    prototype: ChannelSplitterNode;\n    new(context: BaseAudioContext, options?: ChannelSplitterOptions): ChannelSplitterNode;\n};\n\n/** The CharacterData abstract interface represents a Node object that contains characters. This is an abstract interface, meaning there aren\'t any object of type CharacterData: it is implemented by other interfaces, like Text, Comment, or ProcessingInstruction which aren\'t abstract. */\ninterface CharacterData extends Node, ChildNode, NonDocumentTypeChildNode {\n    data: string;\n    readonly length: number;\n    readonly ownerDocument: Document;\n    appendData(data: string): void;\n    deleteData(offset: number, count: number): void;\n    insertData(offset: number, data: string): void;\n    replaceData(offset: number, count: number, data: string): void;\n    substringData(offset: number, count: number): string;\n}\n\ndeclare var CharacterData: {\n    prototype: CharacterData;\n    new(): CharacterData;\n};\n\ninterface ChildNode extends Node {\n    /**\n     * Inserts nodes just after node, while replacing strings in nodes with equivalent Text nodes.\n     * \n     * Throws a "HierarchyRequestError" DOMException if the constraints of the node tree are violated.\n     */\n    after(...nodes: (Node | string)[]): void;\n    /**\n     * Inserts nodes just before node, while replacing strings in nodes with equivalent Text nodes.\n     * \n     * Throws a "HierarchyRequestError" DOMException if the constraints of the node tree are violated.\n     */\n    before(...nodes: (Node | string)[]): void;\n    /**\n     * Removes node.\n     */\n    remove(): void;\n    /**\n     * Replaces node with nodes, while replacing strings in nodes with equivalent Text nodes.\n     * \n     * Throws a "HierarchyRequestError" DOMException if the constraints of the node tree are violated.\n     */\n    replaceWith(...nodes: (Node | string)[]): void;\n}\n\ninterface ClientRect {\n    bottom: number;\n    readonly height: number;\n    left: number;\n    right: number;\n    top: number;\n    readonly width: number;\n}\n\ndeclare var ClientRect: {\n    prototype: ClientRect;\n    new(): ClientRect;\n};\n\ninterface ClientRectList {\n    readonly length: number;\n    item(index: number): ClientRect;\n    [index: number]: ClientRect;\n}\n\ndeclare var ClientRectList: {\n    prototype: ClientRectList;\n    new(): ClientRectList;\n};\n\ninterface Clipboard extends EventTarget {\n    readText(): Promise<string>;\n    writeText(data: string): Promise<void>;\n}\n\ndeclare var Clipboard: {\n    prototype: Clipboard;\n    new(): Clipboard;\n};\n\n/** Events providing information related to modification of the clipboard, that is cut, copy, and paste events. */\ninterface ClipboardEvent extends Event {\n    readonly clipboardData: DataTransfer | null;\n}\n\ndeclare var ClipboardEvent: {\n    prototype: ClipboardEvent;\n    new(type: string, eventInitDict?: ClipboardEventInit): ClipboardEvent;\n};\n\n/** A CloseEvent is sent to clients using WebSockets when the connection is closed. This is delivered to the listener indicated by the WebSocket object\'s onclose attribute. */\ninterface CloseEvent extends Event {\n    /**\n     * Returns the WebSocket connection close code provided by the server.\n     */\n    readonly code: number;\n    /**\n     * Returns the WebSocket connection close reason provided by the server.\n     */\n    readonly reason: string;\n    /**\n     * Returns true if the connection closed cleanly; false otherwise.\n     */\n    readonly wasClean: boolean;\n}\n\ndeclare var CloseEvent: {\n    prototype: CloseEvent;\n    new(type: string, eventInitDict?: CloseEventInit): CloseEvent;\n};\n\n/** Textual notations within markup; although it is generally not visually shown, such comments are available to be read in the source view. */\ninterface Comment extends CharacterData {\n}\n\ndeclare var Comment: {\n    prototype: Comment;\n    new(data?: string): Comment;\n};\n\n/** The DOM CompositionEvent represents events that occur due to the user indirectly entering text. */\ninterface CompositionEvent extends UIEvent {\n    readonly data: string;\n}\n\ndeclare var CompositionEvent: {\n    prototype: CompositionEvent;\n    new(type: string, eventInitDict?: CompositionEventInit): CompositionEvent;\n};\n\ninterface ConcatParams extends Algorithm {\n    algorithmId: Uint8Array;\n    hash?: string | Algorithm;\n    partyUInfo: Uint8Array;\n    partyVInfo: Uint8Array;\n    privateInfo?: Uint8Array;\n    publicInfo?: Uint8Array;\n}\n\ninterface ConstantSourceNode extends AudioScheduledSourceNode {\n    readonly offset: AudioParam;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: ConstantSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: ConstantSourceNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ConstantSourceNode: {\n    prototype: ConstantSourceNode;\n    new(context: BaseAudioContext, options?: ConstantSourceOptions): ConstantSourceNode;\n};\n\n/** An AudioNode that performs a Linear Convolution on a given AudioBuffer, often used to achieve a reverb effect. A ConvolverNode always has exactly one input and one output. */\ninterface ConvolverNode extends AudioNode {\n    buffer: AudioBuffer | null;\n    normalize: boolean;\n}\n\ndeclare var ConvolverNode: {\n    prototype: ConvolverNode;\n    new(context: BaseAudioContext, options?: ConvolverOptions): ConvolverNode;\n};\n\n/** This Streams API interface provides a built-in byte length queuing strategy that can be used when constructing streams. */\ninterface CountQueuingStrategy extends QueuingStrategy {\n    highWaterMark: number;\n    size(chunk: any): 1;\n}\n\ndeclare var CountQueuingStrategy: {\n    prototype: CountQueuingStrategy;\n    new(options: { highWaterMark: number }): CountQueuingStrategy;\n};\n\ninterface Credential {\n    readonly id: string;\n    readonly type: string;\n}\n\ndeclare var Credential: {\n    prototype: Credential;\n    new(): Credential;\n};\n\ninterface CredentialsContainer {\n    create(options?: CredentialCreationOptions): Promise<Credential | null>;\n    get(options?: CredentialRequestOptions): Promise<Credential | null>;\n    preventSilentAccess(): Promise<void>;\n    store(credential: Credential): Promise<Credential>;\n}\n\ndeclare var CredentialsContainer: {\n    prototype: CredentialsContainer;\n    new(): CredentialsContainer;\n};\n\n/** Basic cryptography features available in the current context. It allows access to a cryptographically strong random number generator and to cryptographic primitives. */\ninterface Crypto {\n    readonly subtle: SubtleCrypto;\n    getRandomValues<T extends Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | null>(array: T): T;\n}\n\ndeclare var Crypto: {\n    prototype: Crypto;\n    new(): Crypto;\n};\n\n/** The CryptoKey dictionary of the Web Crypto API represents a cryptographic key. */\ninterface CryptoKey {\n    readonly algorithm: KeyAlgorithm;\n    readonly extractable: boolean;\n    readonly type: KeyType;\n    readonly usages: KeyUsage[];\n}\n\ndeclare var CryptoKey: {\n    prototype: CryptoKey;\n    new(): CryptoKey;\n};\n\n/** The CryptoKeyPair dictionary of the Web Crypto API represents a key pair for an asymmetric cryptography algorithm, also known as a public-key algorithm. */\ninterface CryptoKeyPair {\n    privateKey: CryptoKey;\n    publicKey: CryptoKey;\n}\n\ndeclare var CryptoKeyPair: {\n    prototype: CryptoKeyPair;\n    new(): CryptoKeyPair;\n};\n\ninterface CustomElementRegistry {\n    define(name: string, constructor: CustomElementConstructor, options?: ElementDefinitionOptions): void;\n    get(name: string): any;\n    upgrade(root: Node): void;\n    whenDefined(name: string): Promise<void>;\n}\n\ndeclare var CustomElementRegistry: {\n    prototype: CustomElementRegistry;\n    new(): CustomElementRegistry;\n};\n\ninterface CustomEvent<T = any> extends Event {\n    /**\n     * Returns any custom data event was created with. Typically used for synthetic events.\n     */\n    readonly detail: T;\n    initCustomEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, detailArg: T): void;\n}\n\ndeclare var CustomEvent: {\n    prototype: CustomEvent;\n    new<T>(typeArg: string, eventInitDict?: CustomEventInit<T>): CustomEvent<T>;\n};\n\n/** An error object that contains an error name. */\ninterface DOMError {\n    readonly name: string;\n    toString(): string;\n}\n\ndeclare var DOMError: {\n    prototype: DOMError;\n    new(): DOMError;\n};\n\n/** An abnormal event (called an exception) which occurs as a result of calling a method or accessing a property of a web API. */\ninterface DOMException {\n    readonly code: number;\n    readonly message: string;\n    readonly name: string;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n}\n\ndeclare var DOMException: {\n    prototype: DOMException;\n    new(message?: string, name?: string): DOMException;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n};\n\n/** An object providing methods which are not dependent on any particular document. Such an object is returned by the Document.implementation property. */\ninterface DOMImplementation {\n    createDocument(namespace: string | null, qualifiedName: string | null, doctype?: DocumentType | null): XMLDocument;\n    createDocumentType(qualifiedName: string, publicId: string, systemId: string): DocumentType;\n    createHTMLDocument(title?: string): Document;\n    /** @deprecated */\n    hasFeature(...args: any[]): true;\n}\n\ndeclare var DOMImplementation: {\n    prototype: DOMImplementation;\n    new(): DOMImplementation;\n};\n\ninterface DOML2DeprecatedColorProperty {\n    color: string;\n}\n\ninterface DOMMatrix extends DOMMatrixReadOnly {\n    a: number;\n    b: number;\n    c: number;\n    d: number;\n    e: number;\n    f: number;\n    m11: number;\n    m12: number;\n    m13: number;\n    m14: number;\n    m21: number;\n    m22: number;\n    m23: number;\n    m24: number;\n    m31: number;\n    m32: number;\n    m33: number;\n    m34: number;\n    m41: number;\n    m42: number;\n    m43: number;\n    m44: number;\n    invertSelf(): DOMMatrix;\n    multiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    preMultiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    rotateAxisAngleSelf(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVectorSelf(x?: number, y?: number): DOMMatrix;\n    rotateSelf(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    scale3dSelf(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scaleSelf(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    setMatrixValue(transformList: string): DOMMatrix;\n    skewXSelf(sx?: number): DOMMatrix;\n    skewYSelf(sy?: number): DOMMatrix;\n    translateSelf(tx?: number, ty?: number, tz?: number): DOMMatrix;\n}\n\ndeclare var DOMMatrix: {\n    prototype: DOMMatrix;\n    new(init?: string | number[]): DOMMatrix;\n    fromFloat32Array(array32: Float32Array): DOMMatrix;\n    fromFloat64Array(array64: Float64Array): DOMMatrix;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrix;\n};\n\ntype SVGMatrix = DOMMatrix;\ndeclare var SVGMatrix: typeof DOMMatrix;\n\ntype WebKitCSSMatrix = DOMMatrix;\ndeclare var WebKitCSSMatrix: typeof DOMMatrix;\n\ninterface DOMMatrixReadOnly {\n    readonly a: number;\n    readonly b: number;\n    readonly c: number;\n    readonly d: number;\n    readonly e: number;\n    readonly f: number;\n    readonly is2D: boolean;\n    readonly isIdentity: boolean;\n    readonly m11: number;\n    readonly m12: number;\n    readonly m13: number;\n    readonly m14: number;\n    readonly m21: number;\n    readonly m22: number;\n    readonly m23: number;\n    readonly m24: number;\n    readonly m31: number;\n    readonly m32: number;\n    readonly m33: number;\n    readonly m34: number;\n    readonly m41: number;\n    readonly m42: number;\n    readonly m43: number;\n    readonly m44: number;\n    flipX(): DOMMatrix;\n    flipY(): DOMMatrix;\n    inverse(): DOMMatrix;\n    multiply(other?: DOMMatrixInit): DOMMatrix;\n    rotate(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    rotateAxisAngle(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVector(x?: number, y?: number): DOMMatrix;\n    scale(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scale3d(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    /** @deprecated */\n    scaleNonUniform(scaleX?: number, scaleY?: number): DOMMatrix;\n    skewX(sx?: number): DOMMatrix;\n    skewY(sy?: number): DOMMatrix;\n    toFloat32Array(): Float32Array;\n    toFloat64Array(): Float64Array;\n    toJSON(): any;\n    transformPoint(point?: DOMPointInit): DOMPoint;\n    translate(tx?: number, ty?: number, tz?: number): DOMMatrix;\n    toString(): string;\n}\n\ndeclare var DOMMatrixReadOnly: {\n    prototype: DOMMatrixReadOnly;\n    new(init?: string | number[]): DOMMatrixReadOnly;\n    fromFloat32Array(array32: Float32Array): DOMMatrixReadOnly;\n    fromFloat64Array(array64: Float64Array): DOMMatrixReadOnly;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrixReadOnly;\n    toString(): string;\n};\n\n/** Provides the ability to parse XML or HTML source code from a string into a DOM Document. */\ninterface DOMParser {\n    /**\n     * Parses string using either the HTML or XML parser, according to type, and returns the resulting Document. type can be "text/html" (which will invoke the HTML parser), or any of "text/xml", "application/xml", "application/xhtml+xml", or "image/svg+xml" (which will invoke the XML parser).\n     * \n     * For the XML parser, if string cannot be parsed, then the returned Document will contain elements describing the resulting error.\n     * \n     * Note that script elements are not evaluated during parsing, and the resulting document\'s encoding will always be UTF-8.\n     * \n     * Values other than the above for type will cause a TypeError exception to be thrown.\n     */\n    parseFromString(string: string, type: DOMParserSupportedType): Document;\n}\n\ndeclare var DOMParser: {\n    prototype: DOMParser;\n    new(): DOMParser;\n};\n\ninterface DOMPoint extends DOMPointReadOnly {\n    w: number;\n    x: number;\n    y: number;\n    z: number;\n}\n\ndeclare var DOMPoint: {\n    prototype: DOMPoint;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPoint;\n    fromPoint(other?: DOMPointInit): DOMPoint;\n};\n\ntype SVGPoint = DOMPoint;\ndeclare var SVGPoint: typeof DOMPoint;\n\ninterface DOMPointReadOnly {\n    readonly w: number;\n    readonly x: number;\n    readonly y: number;\n    readonly z: number;\n    matrixTransform(matrix?: DOMMatrixInit): DOMPoint;\n    toJSON(): any;\n}\n\ndeclare var DOMPointReadOnly: {\n    prototype: DOMPointReadOnly;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPointReadOnly;\n    fromPoint(other?: DOMPointInit): DOMPointReadOnly;\n};\n\ninterface DOMQuad {\n    readonly p1: DOMPoint;\n    readonly p2: DOMPoint;\n    readonly p3: DOMPoint;\n    readonly p4: DOMPoint;\n    getBounds(): DOMRect;\n    toJSON(): any;\n}\n\ndeclare var DOMQuad: {\n    prototype: DOMQuad;\n    new(p1?: DOMPointInit, p2?: DOMPointInit, p3?: DOMPointInit, p4?: DOMPointInit): DOMQuad;\n    fromQuad(other?: DOMQuadInit): DOMQuad;\n    fromRect(other?: DOMRectInit): DOMQuad;\n};\n\ninterface DOMRect extends DOMRectReadOnly {\n    height: number;\n    width: number;\n    x: number;\n    y: number;\n}\n\ndeclare var DOMRect: {\n    prototype: DOMRect;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRect;\n    fromRect(other?: DOMRectInit): DOMRect;\n};\n\ntype SVGRect = DOMRect;\ndeclare var SVGRect: typeof DOMRect;\n\ninterface DOMRectList {\n    readonly length: number;\n    item(index: number): DOMRect | null;\n    [index: number]: DOMRect;\n}\n\ndeclare var DOMRectList: {\n    prototype: DOMRectList;\n    new(): DOMRectList;\n};\n\ninterface DOMRectReadOnly {\n    readonly bottom: number;\n    readonly height: number;\n    readonly left: number;\n    readonly right: number;\n    readonly top: number;\n    readonly width: number;\n    readonly x: number;\n    readonly y: number;\n    toJSON(): any;\n}\n\ndeclare var DOMRectReadOnly: {\n    prototype: DOMRectReadOnly;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRectReadOnly;\n    fromRect(other?: DOMRectInit): DOMRectReadOnly;\n};\n\ninterface DOMSettableTokenList extends DOMTokenList {\n    value: string;\n}\n\ndeclare var DOMSettableTokenList: {\n    prototype: DOMSettableTokenList;\n    new(): DOMSettableTokenList;\n};\n\n/** A type returned by some APIs which contains a list of DOMString (strings). */\ninterface DOMStringList {\n    /**\n     * Returns the number of strings in strings.\n     */\n    readonly length: number;\n    /**\n     * Returns true if strings contains string, and false otherwise.\n     */\n    contains(string: string): boolean;\n    /**\n     * Returns the string with index index from strings.\n     */\n    item(index: number): string | null;\n    [index: number]: string;\n}\n\ndeclare var DOMStringList: {\n    prototype: DOMStringList;\n    new(): DOMStringList;\n};\n\n/** Used by the dataset HTML attribute to represent data for custom attributes added to elements. */\ninterface DOMStringMap {\n    [name: string]: string | undefined;\n}\n\ndeclare var DOMStringMap: {\n    prototype: DOMStringMap;\n    new(): DOMStringMap;\n};\n\n/** A set of space-separated tokens. Such a set is returned by Element.classList, HTMLLinkElement.relList, HTMLAnchorElement.relList, HTMLAreaElement.relList, HTMLIframeElement.sandbox, or HTMLOutputElement.htmlFor. It is indexed beginning with 0 as with JavaScript Array objects. DOMTokenList is always case-sensitive. */\ninterface DOMTokenList {\n    /**\n     * Returns the number of tokens.\n     */\n    readonly length: number;\n    /**\n     * Returns the associated set as string.\n     * \n     * Can be set, to change the associated attribute.\n     */\n    value: string;\n    toString(): string;\n    /**\n     * Adds all arguments passed, except those already present.\n     * \n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty string.\n     * \n     * Throws an "InvalidCharacterError" DOMException if one of the arguments contains any ASCII whitespace.\n     */\n    add(...tokens: string[]): void;\n    /**\n     * Returns true if token is present, and false otherwise.\n     */\n    contains(token: string): boolean;\n    /**\n     * Returns the token with index index.\n     */\n    item(index: number): string | null;\n    /**\n     * Removes arguments passed, if they are present.\n     * \n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty string.\n     * \n     * Throws an "InvalidCharacterError" DOMException if one of the arguments contains any ASCII whitespace.\n     */\n    remove(...tokens: string[]): void;\n    /**\n     * Replaces token with newToken.\n     * \n     * Returns true if token was replaced with newToken, and false otherwise.\n     * \n     * Throws a "SyntaxError" DOMException if one of the arguments is the empty string.\n     * \n     * Throws an "InvalidCharacterError" DOMException if one of the arguments contains any ASCII whitespace.\n     */\n    replace(oldToken: string, newToken: string): void;\n    /**\n     * Returns true if token is in the associated attribute\'s supported tokens. Returns false otherwise.\n     * \n     * Throws a TypeError if the associated attribute has no supported tokens defined.\n     */\n    supports(token: string): boolean;\n    /**\n     * If force is not given, "toggles" token, removing it if it\'s present and adding it if it\'s not present. If force is true, adds token (same as add()). If force is false, removes token (same as remove()).\n     * \n     * Returns true if token is now present, and false otherwise.\n     * \n     * Throws a "SyntaxError" DOMException if token is empty.\n     * \n     * Throws an "InvalidCharacterError" DOMException if token contains any spaces.\n     */\n    toggle(token: string, force?: boolean): boolean;\n    forEach(callbackfn: (value: string, key: number, parent: DOMTokenList) => void, thisArg?: any): void;\n    [index: number]: string;\n}\n\ndeclare var DOMTokenList: {\n    prototype: DOMTokenList;\n    new(): DOMTokenList;\n};\n\ninterface DataCue extends TextTrackCue {\n    data: ArrayBuffer;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: DataCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: DataCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var DataCue: {\n    prototype: DataCue;\n    new(): DataCue;\n};\n\n/** Used to hold the data that is being dragged during a drag and drop operation. It may hold one or more data items, each of one or more data types. For more information about drag and drop, see HTML Drag and Drop API. */\ninterface DataTransfer {\n    /**\n     * Returns the kind of operation that is currently selected. If the kind of operation isn\'t one of those that is allowed by the effectAllowed attribute, then the operation will fail.\n     * \n     * Can be set, to change the selected operation.\n     * \n     * The possible values are "none", "copy", "link", and "move".\n     */\n    dropEffect: "none" | "copy" | "link" | "move";\n    /**\n     * Returns the kinds of operations that are to be allowed.\n     * \n     * Can be set (during the dragstart event), to change the allowed operations.\n     * \n     * The possible values are "none", "copy", "copyLink", "copyMove", "link", "linkMove", "move", "all", and "uninitialized",\n     */\n    effectAllowed: "none" | "copy" | "copyLink" | "copyMove" | "link" | "linkMove" | "move" | "all" | "uninitialized";\n    /**\n     * Returns a FileList of the files being dragged, if any.\n     */\n    readonly files: FileList;\n    /**\n     * Returns a DataTransferItemList object, with the drag data.\n     */\n    readonly items: DataTransferItemList;\n    /**\n     * Returns a frozen array listing the formats that were set in the dragstart event. In addition, if any files are being dragged, then one of the types will be the string "Files".\n     */\n    readonly types: ReadonlyArray<string>;\n    /**\n     * Removes the data of the specified formats. Removes all data if the argument is omitted.\n     */\n    clearData(format?: string): void;\n    /**\n     * Returns the specified data. If there is no such data, returns the empty string.\n     */\n    getData(format: string): string;\n    /**\n     * Adds the specified data.\n     */\n    setData(format: string, data: string): void;\n    /**\n     * Uses the given element to update the drag feedback, replacing any previously specified feedback.\n     */\n    setDragImage(image: Element, x: number, y: number): void;\n}\n\ndeclare var DataTransfer: {\n    prototype: DataTransfer;\n    new(): DataTransfer;\n};\n\n/** One drag data item. During a drag operation, each drag event has a dataTransfer property which contains a list of drag data items. Each item in the list is a DataTransferItem object. */\ninterface DataTransferItem {\n    /**\n     * Returns the drag data item kind, one of: "string", "file".\n     */\n    readonly kind: string;\n    /**\n     * Returns the drag data item type string.\n     */\n    readonly type: string;\n    /**\n     * Returns a File object, if the drag data item kind is File.\n     */\n    getAsFile(): File | null;\n    /**\n     * Invokes the callback with the string data as the argument, if the drag data item kind is text.\n     */\n    getAsString(callback: FunctionStringCallback | null): void;\n    webkitGetAsEntry(): any;\n}\n\ndeclare var DataTransferItem: {\n    prototype: DataTransferItem;\n    new(): DataTransferItem;\n};\n\n/** A list of DataTransferItem objects representing items being dragged. During a drag operation, each DragEvent has a dataTransfer property and that property is a DataTransferItemList. */\ninterface DataTransferItemList {\n    /**\n     * Returns the number of items in the drag data store.\n     */\n    readonly length: number;\n    /**\n     * Adds a new entry for the given data to the drag data store. If the data is plain text then a type string has to be provided also.\n     */\n    add(data: string, type: string): DataTransferItem | null;\n    add(data: File): DataTransferItem | null;\n    /**\n     * Removes all the entries in the drag data store.\n     */\n    clear(): void;\n    item(index: number): DataTransferItem;\n    /**\n     * Removes the indexth entry in the drag data store.\n     */\n    remove(index: number): void;\n    [name: number]: DataTransferItem;\n}\n\ndeclare var DataTransferItemList: {\n    prototype: DataTransferItemList;\n    new(): DataTransferItemList;\n};\n\ninterface DeferredPermissionRequest {\n    readonly id: number;\n    readonly type: MSWebViewPermissionType;\n    readonly uri: string;\n    allow(): void;\n    deny(): void;\n}\n\ndeclare var DeferredPermissionRequest: {\n    prototype: DeferredPermissionRequest;\n    new(): DeferredPermissionRequest;\n};\n\n/** A delay-line; an AudioNode audio-processing module that causes a delay between the arrival of an input data and its propagation to the output. */\ninterface DelayNode extends AudioNode {\n    readonly delayTime: AudioParam;\n}\n\ndeclare var DelayNode: {\n    prototype: DelayNode;\n    new(context: BaseAudioContext, options?: DelayOptions): DelayNode;\n};\n\n/** Provides information about the amount of acceleration the device is experiencing along all three axes. */\ninterface DeviceAcceleration {\n    readonly x: number | null;\n    readonly y: number | null;\n    readonly z: number | null;\n}\n\ndeclare var DeviceAcceleration: {\n    prototype: DeviceAcceleration;\n    new(): DeviceAcceleration;\n};\n\n/** The DeviceLightEvent provides web developers with information from photo sensors or similiar detectors about ambient light levels near the device. For example this may be useful to adjust the screen\'s brightness based on the current ambient light level in order to save energy or provide better readability. */\ninterface DeviceLightEvent extends Event {\n    readonly value: number;\n}\n\ndeclare var DeviceLightEvent: {\n    prototype: DeviceLightEvent;\n    new(typeArg: string, eventInitDict?: DeviceLightEventInit): DeviceLightEvent;\n};\n\n/** The DeviceMotionEvent provides web developers with information about the speed of changes for the device\'s position and orientation. */\ninterface DeviceMotionEvent extends Event {\n    readonly acceleration: DeviceMotionEventAcceleration | null;\n    readonly accelerationIncludingGravity: DeviceMotionEventAcceleration | null;\n    readonly interval: number;\n    readonly rotationRate: DeviceMotionEventRotationRate | null;\n}\n\ndeclare var DeviceMotionEvent: {\n    prototype: DeviceMotionEvent;\n    new(type: string, eventInitDict?: DeviceMotionEventInit): DeviceMotionEvent;\n    requestPermission(): Promise<PermissionState>;\n};\n\ninterface DeviceMotionEventAcceleration {\n    readonly x: number | null;\n    readonly y: number | null;\n    readonly z: number | null;\n}\n\ninterface DeviceMotionEventRotationRate {\n    readonly alpha: number | null;\n    readonly beta: number | null;\n    readonly gamma: number | null;\n}\n\n/** The DeviceOrientationEvent provides web developers with information from the physical orientation of the device running the web page. */\ninterface DeviceOrientationEvent extends Event {\n    readonly absolute: boolean;\n    readonly alpha: number | null;\n    readonly beta: number | null;\n    readonly gamma: number | null;\n}\n\ndeclare var DeviceOrientationEvent: {\n    prototype: DeviceOrientationEvent;\n    new(type: string, eventInitDict?: DeviceOrientationEventInit): DeviceOrientationEvent;\n    requestPermission(): Promise<PermissionState>;\n};\n\n/** Provides information about the rate at which the device is rotating around all three axes. */\ninterface DeviceRotationRate {\n    readonly alpha: number | null;\n    readonly beta: number | null;\n    readonly gamma: number | null;\n}\n\ndeclare var DeviceRotationRate: {\n    prototype: DeviceRotationRate;\n    new(): DeviceRotationRate;\n};\n\ninterface DhImportKeyParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyAlgorithm extends KeyAlgorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface DhKeyGenParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DocumentEventMap extends GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n    "fullscreenchange": Event;\n    "fullscreenerror": Event;\n    "pointerlockchange": Event;\n    "pointerlockerror": Event;\n    "readystatechange": Event;\n    "visibilitychange": Event;\n}\n\n/** Any web page loaded in the browser and serves as an entry point into the web page\'s content, which is the DOM tree. */\ninterface Document extends Node, DocumentAndElementEventHandlers, DocumentOrShadowRoot, GlobalEventHandlers, NonElementParentNode, ParentNode, XPathEvaluatorBase {\n    /**\n     * Sets or gets the URL for the current document.\n     */\n    readonly URL: string;\n    /**\n     * Sets or gets the color of all active links in the document.\n     */\n    /** @deprecated */\n    alinkColor: string;\n    /**\n     * Returns a reference to the collection of elements contained by the object.\n     */\n    /** @deprecated */\n    readonly all: HTMLAllCollection;\n    /**\n     * Retrieves a collection of all a objects that have a name and/or id property. Objects in this collection are in HTML source order.\n     */\n    /** @deprecated */\n    readonly anchors: HTMLCollectionOf<HTMLAnchorElement>;\n    /**\n     * Retrieves a collection of all applet objects in the document.\n     */\n    /** @deprecated */\n    readonly applets: HTMLCollectionOf<HTMLAppletElement>;\n    /**\n     * Deprecated. Sets or retrieves a value that indicates the background color behind the object.\n     */\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Specifies the beginning and end of the document body.\n     */\n    body: HTMLElement;\n    /**\n     * Returns document\'s encoding.\n     */\n    readonly characterSet: string;\n    /**\n     * Gets or sets the character set used to encode the object.\n     */\n    readonly charset: string;\n    /**\n     * Gets a value that indicates whether standards-compliant mode is switched on for the object.\n     */\n    readonly compatMode: string;\n    /**\n     * Returns document\'s content type.\n     */\n    readonly contentType: string;\n    /**\n     * Returns the HTTP cookies that apply to the Document. If there are no cookies or cookies can\'t be applied to this resource, the empty string will be returned.\n     * \n     * Can be set, to add a new cookie to the element\'s set of HTTP cookies.\n     * \n     * If the contents are sandboxed into a unique origin (e.g. in an iframe with the sandbox attribute), a "SecurityError" DOMException will be thrown on getting and setting.\n     */\n    cookie: string;\n    /**\n     * Returns the script element, or the SVG script element, that is currently executing, as long as the element represents a classic script. In the case of reentrant script execution, returns the one that most recently started executing amongst those that have not yet finished executing.\n     * \n     * Returns null if the Document is not currently executing a script or SVG script element (e.g., because the running script is an event handler, or a timeout), or if the currently executing script or SVG script element represents a module script.\n     */\n    readonly currentScript: HTMLOrSVGScriptElement | null;\n    readonly defaultView: (WindowProxy & typeof globalThis) | null;\n    /**\n     * Sets or gets a value that indicates whether the document can be edited.\n     */\n    designMode: string;\n    /**\n     * Sets or retrieves a value that indicates the reading order of the object.\n     */\n    dir: string;\n    /**\n     * Gets an object representing the document type declaration associated with the current document.\n     */\n    readonly doctype: DocumentType | null;\n    /**\n     * Gets a reference to the root node of the document.\n     */\n    readonly documentElement: HTMLElement;\n    /**\n     * Returns document\'s URL.\n     */\n    readonly documentURI: string;\n    /**\n     * Sets or gets the security domain of the document.\n     */\n    domain: string;\n    /**\n     * Retrieves a collection of all embed objects in the document.\n     */\n    readonly embeds: HTMLCollectionOf<HTMLEmbedElement>;\n    /**\n     * Sets or gets the foreground (text) color of the document.\n     */\n    /** @deprecated */\n    fgColor: string;\n    /**\n     * Retrieves a collection, in source order, of all form objects in the document.\n     */\n    readonly forms: HTMLCollectionOf<HTMLFormElement>;\n    /** @deprecated */\n    readonly fullscreen: boolean;\n    /**\n     * Returns true if document has the ability to display elements fullscreen and fullscreen is supported, or false otherwise.\n     */\n    readonly fullscreenEnabled: boolean;\n    /**\n     * Returns the head element.\n     */\n    readonly head: HTMLHeadElement;\n    readonly hidden: boolean;\n    /**\n     * Retrieves a collection, in source order, of img objects in the document.\n     */\n    readonly images: HTMLCollectionOf<HTMLImageElement>;\n    /**\n     * Gets the implementation object of the current document.\n     */\n    readonly implementation: DOMImplementation;\n    /**\n     * Returns the character encoding used to create the webpage that is loaded into the document object.\n     */\n    readonly inputEncoding: string;\n    /**\n     * Gets the date that the page was last modified, if the page supplies one.\n     */\n    readonly lastModified: string;\n    /**\n     * Sets or gets the color of the document links.\n     */\n    /** @deprecated */\n    linkColor: string;\n    /**\n     * Retrieves a collection of all a objects that specify the href property and all area objects in the document.\n     */\n    readonly links: HTMLCollectionOf<HTMLAnchorElement | HTMLAreaElement>;\n    /**\n     * Contains information about the current URL.\n     */\n    location: Location;\n    onfullscreenchange: ((this: Document, ev: Event) => any) | null;\n    onfullscreenerror: ((this: Document, ev: Event) => any) | null;\n    onpointerlockchange: ((this: Document, ev: Event) => any) | null;\n    onpointerlockerror: ((this: Document, ev: Event) => any) | null;\n    /**\n     * Fires when the state of the object has changed.\n     * @param ev The event\n     */\n    onreadystatechange: ((this: Document, ev: Event) => any) | null;\n    onvisibilitychange: ((this: Document, ev: Event) => any) | null;\n    readonly ownerDocument: null;\n    /**\n     * Return an HTMLCollection of the embed elements in the Document.\n     */\n    readonly plugins: HTMLCollectionOf<HTMLEmbedElement>;\n    /**\n     * Retrieves a value that indicates the current state of the object.\n     */\n    readonly readyState: DocumentReadyState;\n    /**\n     * Gets the URL of the location that referred the user to the current page.\n     */\n    readonly referrer: string;\n    /**\n     * Retrieves a collection of all script objects in the document.\n     */\n    readonly scripts: HTMLCollectionOf<HTMLScriptElement>;\n    readonly scrollingElement: Element | null;\n    readonly timeline: DocumentTimeline;\n    /**\n     * Contains the title of the document.\n     */\n    title: string;\n    readonly visibilityState: VisibilityState;\n    /**\n     * Sets or gets the color of the links that the user has visited.\n     */\n    /** @deprecated */\n    vlinkColor: string;\n    /**\n     * Moves node from another document and returns it.\n     * \n     * If node is a document, throws a "NotSupportedError" DOMException or, if node is a shadow root, throws a "HierarchyRequestError" DOMException.\n     */\n    adoptNode<T extends Node>(source: T): T;\n    /** @deprecated */\n    captureEvents(): void;\n    caretPositionFromPoint(x: number, y: number): CaretPosition | null;\n    /** @deprecated */\n    caretRangeFromPoint(x: number, y: number): Range;\n    /** @deprecated */\n    clear(): void;\n    /**\n     * Closes an output stream and forces the sent data to display.\n     */\n    close(): void;\n    /**\n     * Creates an attribute object with a specified name.\n     * @param name String that sets the attribute object\'s name.\n     */\n    createAttribute(localName: string): Attr;\n    createAttributeNS(namespace: string | null, qualifiedName: string): Attr;\n    /**\n     * Returns a CDATASection node whose data is data.\n     */\n    createCDATASection(data: string): CDATASection;\n    /**\n     * Creates a comment object with the specified data.\n     * @param data Sets the comment object\'s data.\n     */\n    createComment(data: string): Comment;\n    /**\n     * Creates a new document.\n     */\n    createDocumentFragment(): DocumentFragment;\n    /**\n     * Creates an instance of the element for the specified tag.\n     * @param tagName The name of an element.\n     */\n    createElement<K extends keyof HTMLElementTagNameMap>(tagName: K, options?: ElementCreationOptions): HTMLElementTagNameMap[K];\n    /** @deprecated */\n    createElement<K extends keyof HTMLElementDeprecatedTagNameMap>(tagName: K, options?: ElementCreationOptions): HTMLElementDeprecatedTagNameMap[K];\n    createElement(tagName: string, options?: ElementCreationOptions): HTMLElement;\n    /**\n     * Returns an element with namespace namespace. Its namespace prefix will be everything before ":" (U+003E) in qualifiedName or null. Its local name will be everything after ":" (U+003E) in qualifiedName or qualifiedName.\n     * \n     * If localName does not match the Name production an "InvalidCharacterError" DOMException will be thrown.\n     * \n     * If one of the following conditions is true a "NamespaceError" DOMException will be thrown:\n     * \n     * localName does not match the QName production.\n     * Namespace prefix is not null and namespace is the empty string.\n     * Namespace prefix is "xml" and namespace is not the XML namespace.\n     * qualifiedName or namespace prefix is "xmlns" and namespace is not the XMLNS namespace.\n     * namespace is the XMLNS namespace and neither qualifiedName nor namespace prefix is "xmlns".\n     * \n     * When supplied, options\'s is can be used to create a customized built-in element.\n     */\n    createElementNS(namespaceURI: "http://www.w3.org/1999/xhtml", qualifiedName: string): HTMLElement;\n    createElementNS<K extends keyof SVGElementTagNameMap>(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: K): SVGElementTagNameMap[K];\n    createElementNS(namespaceURI: "http://www.w3.org/2000/svg", qualifiedName: string): SVGElement;\n    createElementNS(namespaceURI: string | null, qualifiedName: string, options?: ElementCreationOptions): Element;\n    createElementNS(namespace: string | null, qualifiedName: string, options?: string | ElementCreationOptions): Element;\n    createEvent(eventInterface: "AnimationEvent"): AnimationEvent;\n    createEvent(eventInterface: "AnimationPlaybackEvent"): AnimationPlaybackEvent;\n    createEvent(eventInterface: "AudioProcessingEvent"): AudioProcessingEvent;\n    createEvent(eventInterface: "BeforeUnloadEvent"): BeforeUnloadEvent;\n    createEvent(eventInterface: "ClipboardEvent"): ClipboardEvent;\n    createEvent(eventInterface: "CloseEvent"): CloseEvent;\n    createEvent(eventInterface: "CompositionEvent"): CompositionEvent;\n    createEvent(eventInterface: "CustomEvent"): CustomEvent;\n    createEvent(eventInterface: "DeviceLightEvent"): DeviceLightEvent;\n    createEvent(eventInterface: "DeviceMotionEvent"): DeviceMotionEvent;\n    createEvent(eventInterface: "DeviceOrientationEvent"): DeviceOrientationEvent;\n    createEvent(eventInterface: "DragEvent"): DragEvent;\n    createEvent(eventInterface: "ErrorEvent"): ErrorEvent;\n    createEvent(eventInterface: "Event"): Event;\n    createEvent(eventInterface: "Events"): Event;\n    createEvent(eventInterface: "FocusEvent"): FocusEvent;\n    createEvent(eventInterface: "FocusNavigationEvent"): FocusNavigationEvent;\n    createEvent(eventInterface: "GamepadEvent"): GamepadEvent;\n    createEvent(eventInterface: "HashChangeEvent"): HashChangeEvent;\n    createEvent(eventInterface: "IDBVersionChangeEvent"): IDBVersionChangeEvent;\n    createEvent(eventInterface: "InputEvent"): InputEvent;\n    createEvent(eventInterface: "KeyboardEvent"): KeyboardEvent;\n    createEvent(eventInterface: "ListeningStateChangedEvent"): ListeningStateChangedEvent;\n    createEvent(eventInterface: "MSGestureEvent"): MSGestureEvent;\n    createEvent(eventInterface: "MSMediaKeyMessageEvent"): MSMediaKeyMessageEvent;\n    createEvent(eventInterface: "MSMediaKeyNeededEvent"): MSMediaKeyNeededEvent;\n    createEvent(eventInterface: "MSPointerEvent"): MSPointerEvent;\n    createEvent(eventInterface: "MediaEncryptedEvent"): MediaEncryptedEvent;\n    createEvent(eventInterface: "MediaKeyMessageEvent"): MediaKeyMessageEvent;\n    createEvent(eventInterface: "MediaQueryListEvent"): MediaQueryListEvent;\n    createEvent(eventInterface: "MediaStreamErrorEvent"): MediaStreamErrorEvent;\n    createEvent(eventInterface: "MediaStreamEvent"): MediaStreamEvent;\n    createEvent(eventInterface: "MediaStreamTrackEvent"): MediaStreamTrackEvent;\n    createEvent(eventInterface: "MessageEvent"): MessageEvent;\n    createEvent(eventInterface: "MouseEvent"): MouseEvent;\n    createEvent(eventInterface: "MouseEvents"): MouseEvent;\n    createEvent(eventInterface: "MutationEvent"): MutationEvent;\n    createEvent(eventInterface: "MutationEvents"): MutationEvent;\n    createEvent(eventInterface: "OfflineAudioCompletionEvent"): OfflineAudioCompletionEvent;\n    createEvent(eventInterface: "OverflowEvent"): OverflowEvent;\n    createEvent(eventInterface: "PageTransitionEvent"): PageTransitionEvent;\n    createEvent(eventInterface: "PaymentRequestUpdateEvent"): PaymentRequestUpdateEvent;\n    createEvent(eventInterface: "PermissionRequestedEvent"): PermissionRequestedEvent;\n    createEvent(eventInterface: "PointerEvent"): PointerEvent;\n    createEvent(eventInterface: "PopStateEvent"): PopStateEvent;\n    createEvent(eventInterface: "ProgressEvent"): ProgressEvent;\n    createEvent(eventInterface: "PromiseRejectionEvent"): PromiseRejectionEvent;\n    createEvent(eventInterface: "RTCDTMFToneChangeEvent"): RTCDTMFToneChangeEvent;\n    createEvent(eventInterface: "RTCDataChannelEvent"): RTCDataChannelEvent;\n    createEvent(eventInterface: "RTCDtlsTransportStateChangedEvent"): RTCDtlsTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCErrorEvent"): RTCErrorEvent;\n    createEvent(eventInterface: "RTCIceCandidatePairChangedEvent"): RTCIceCandidatePairChangedEvent;\n    createEvent(eventInterface: "RTCIceGathererEvent"): RTCIceGathererEvent;\n    createEvent(eventInterface: "RTCIceTransportStateChangedEvent"): RTCIceTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceErrorEvent"): RTCPeerConnectionIceErrorEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceEvent"): RTCPeerConnectionIceEvent;\n    createEvent(eventInterface: "RTCSsrcConflictEvent"): RTCSsrcConflictEvent;\n    createEvent(eventInterface: "RTCStatsEvent"): RTCStatsEvent;\n    createEvent(eventInterface: "RTCTrackEvent"): RTCTrackEvent;\n    createEvent(eventInterface: "SVGZoomEvent"): SVGZoomEvent;\n    createEvent(eventInterface: "SVGZoomEvents"): SVGZoomEvent;\n    createEvent(eventInterface: "SecurityPolicyViolationEvent"): SecurityPolicyViolationEvent;\n    createEvent(eventInterface: "ServiceWorkerMessageEvent"): ServiceWorkerMessageEvent;\n    createEvent(eventInterface: "SpeechRecognitionEvent"): SpeechRecognitionEvent;\n    createEvent(eventInterface: "SpeechSynthesisErrorEvent"): SpeechSynthesisErrorEvent;\n    createEvent(eventInterface: "SpeechSynthesisEvent"): SpeechSynthesisEvent;\n    createEvent(eventInterface: "StorageEvent"): StorageEvent;\n    createEvent(eventInterface: "TextEvent"): TextEvent;\n    createEvent(eventInterface: "TouchEvent"): TouchEvent;\n    createEvent(eventInterface: "TrackEvent"): TrackEvent;\n    createEvent(eventInterface: "TransitionEvent"): TransitionEvent;\n    createEvent(eventInterface: "UIEvent"): UIEvent;\n    createEvent(eventInterface: "UIEvents"): UIEvent;\n    createEvent(eventInterface: "VRDisplayEvent"): VRDisplayEvent;\n    createEvent(eventInterface: "VRDisplayEvent "): VRDisplayEvent ;\n    createEvent(eventInterface: "WebGLContextEvent"): WebGLContextEvent;\n    createEvent(eventInterface: "WheelEvent"): WheelEvent;\n    createEvent(eventInterface: string): Event;\n    /**\n     * Creates a NodeIterator object that you can use to traverse filtered lists of nodes or elements in a document.\n     * @param root The root element or node to start traversing on.\n     * @param whatToShow The type of nodes or elements to appear in the node list\n     * @param filter A custom NodeFilter function to use. For more information, see filter. Use null for no filter.\n     * @param entityReferenceExpansion A flag that specifies whether entity reference nodes are expanded.\n     */\n    createNodeIterator(root: Node, whatToShow?: number, filter?: NodeFilter | null): NodeIterator;\n    /**\n     * Returns a ProcessingInstruction node whose target is target and data is data. If target does not match the Name production an "InvalidCharacterError" DOMException will be thrown. If data contains "?>" an "InvalidCharacterError" DOMException will be thrown.\n     */\n    createProcessingInstruction(target: string, data: string): ProcessingInstruction;\n    /**\n     *  Returns an empty range object that has both of its boundary points positioned at the beginning of the document.\n     */\n    createRange(): Range;\n    /**\n     * Creates a text string from the specified value.\n     * @param data String that specifies the nodeValue property of the text node.\n     */\n    createTextNode(data: string): Text;\n    /**\n     * Creates a TreeWalker object that you can use to traverse filtered lists of nodes or elements in a document.\n     * @param root The root element or node to start traversing on.\n     * @param whatToShow The type of nodes or elements to appear in the node list. For more information, see whatToShow.\n     * @param filter A custom NodeFilter function to use.\n     * @param entityReferenceExpansion A flag that specifies whether entity reference nodes are expanded.\n     */\n    createTreeWalker(root: Node, whatToShow?: number, filter?: NodeFilter | null): TreeWalker;\n    /** @deprecated */\n    createTreeWalker(root: Node, whatToShow: number, filter: NodeFilter | null, entityReferenceExpansion?: boolean): TreeWalker;\n    /**\n     * Returns the element for the specified x coordinate and the specified y coordinate.\n     * @param x The x-offset\n     * @param y The y-offset\n     */\n    elementFromPoint(x: number, y: number): Element | null;\n    elementsFromPoint(x: number, y: number): Element[];\n    /**\n     * Executes a command on the current document, current selection, or the given range.\n     * @param commandId String that specifies the command to execute. This command can be any of the command identifiers that can be executed in script.\n     * @param showUI Display the user interface, defaults to false.\n     * @param value Value to assign.\n     */\n    execCommand(commandId: string, showUI?: boolean, value?: string): boolean;\n    /**\n     * Stops document\'s fullscreen element from being displayed fullscreen and resolves promise when done.\n     */\n    exitFullscreen(): Promise<void>;\n    exitPointerLock(): void;\n    getAnimations(): Animation[];\n    /**\n     * Returns a reference to the first object with the specified value of the ID or NAME attribute.\n     * @param elementId String that specifies the ID value. Case-insensitive.\n     */\n    getElementById(elementId: string): HTMLElement | null;\n    /**\n     * Returns a HTMLCollection of the elements in the object on which the method was invoked (a document or an element) that have all the classes given by classNames. The classNames argument is interpreted as a space-separated list of classes.\n     */\n    getElementsByClassName(classNames: string): HTMLCollectionOf<Element>;\n    /**\n     * Gets a collection of objects based on the value of the NAME or ID attribute.\n     * @param elementName Gets a collection of objects based on the value of the NAME or ID attribute.\n     */\n    getElementsByName(elementName: string): NodeListOf<HTMLElement>;\n    /**\n     * Retrieves a collection of objects based on the specified element name.\n     * @param name Specifies the name of an element.\n     */\n    getElementsByTagName<K extends keyof HTMLElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<HTMLElementTagNameMap[K]>;\n    getElementsByTagName<K extends keyof SVGElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<SVGElementTagNameMap[K]>;\n    getElementsByTagName(qualifiedName: string): HTMLCollectionOf<Element>;\n    /**\n     * If namespace and localName are "*" returns a HTMLCollection of all descendant elements.\n     * \n     * If only namespace is "*" returns a HTMLCollection of all descendant elements whose local name is localName.\n     * \n     * If only localName is "*" returns a HTMLCollection of all descendant elements whose namespace is namespace.\n     * \n     * Otherwise, returns a HTMLCollection of all descendant elements whose namespace is namespace and local name is localName.\n     */\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/1999/xhtml", localName: string): HTMLCollectionOf<HTMLElement>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/2000/svg", localName: string): HTMLCollectionOf<SVGElement>;\n    getElementsByTagNameNS(namespaceURI: string, localName: string): HTMLCollectionOf<Element>;\n    /**\n     * Returns an object representing the current selection of the document that is loaded into the object displaying a webpage.\n     */\n    getSelection(): Selection | null;\n    /**\n     * Gets a value indicating whether the object currently has focus.\n     */\n    hasFocus(): boolean;\n    /**\n     * Returns a copy of node. If deep is true, the copy also includes the node\'s descendants.\n     * \n     * If node is a document or a shadow root, throws a "NotSupportedError" DOMException.\n     */\n    importNode<T extends Node>(importedNode: T, deep: boolean): T;\n    /**\n     * Opens a new window and loads a document specified by a given URL. Also, opens a new window that uses the url parameter and the name parameter to collect the output of the write method and the writeln method.\n     * @param url Specifies a MIME type for the document.\n     * @param name Specifies the name of the window. This name is used as the value for the TARGET attribute on a form or an anchor element.\n     * @param features Contains a list of items separated by commas. Each item consists of an option and a value, separated by an equals sign (for example, "fullscreen=yes, toolbar=yes"). The following values are supported.\n     * @param replace Specifies whether the existing entry for the document is replaced in the history list.\n     */\n    open(url?: string, name?: string, features?: string, replace?: boolean): Document;\n    /**\n     * Returns a Boolean value that indicates whether a specified command can be successfully executed using execCommand, given the current state of the document.\n     * @param commandId Specifies a command identifier.\n     */\n    queryCommandEnabled(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates whether the specified command is in the indeterminate state.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandIndeterm(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates the current state of the command.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandState(commandId: string): boolean;\n    /**\n     * Returns a Boolean value that indicates whether the current command is supported on the current range.\n     * @param commandId Specifies a command identifier.\n     */\n    queryCommandSupported(commandId: string): boolean;\n    /**\n     * Returns the current value of the document, range, or current selection for the given command.\n     * @param commandId String that specifies a command identifier.\n     */\n    queryCommandValue(commandId: string): string;\n    /** @deprecated */\n    releaseEvents(): void;\n    /**\n     * Writes one or more HTML expressions to a document in the specified window.\n     * @param content Specifies the text and HTML tags to write.\n     */\n    write(...text: string[]): void;\n    /**\n     * Writes one or more HTML expressions, followed by a carriage return, to a document in the specified window.\n     * @param content The text and HTML tags to write.\n     */\n    writeln(...text: string[]): void;\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: Document, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: Document, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Document: {\n    prototype: Document;\n    new(): Document;\n};\n\ninterface DocumentAndElementEventHandlersEventMap {\n    "copy": ClipboardEvent;\n    "cut": ClipboardEvent;\n    "paste": ClipboardEvent;\n}\n\ninterface DocumentAndElementEventHandlers {\n    oncopy: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    oncut: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    onpaste: ((this: DocumentAndElementEventHandlers, ev: ClipboardEvent) => any) | null;\n    addEventListener<K extends keyof DocumentAndElementEventHandlersEventMap>(type: K, listener: (this: DocumentAndElementEventHandlers, ev: DocumentAndElementEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentAndElementEventHandlersEventMap>(type: K, listener: (this: DocumentAndElementEventHandlers, ev: DocumentAndElementEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface DocumentEvent {\n    createEvent(eventInterface: "AnimationEvent"): AnimationEvent;\n    createEvent(eventInterface: "AnimationPlaybackEvent"): AnimationPlaybackEvent;\n    createEvent(eventInterface: "AudioProcessingEvent"): AudioProcessingEvent;\n    createEvent(eventInterface: "BeforeUnloadEvent"): BeforeUnloadEvent;\n    createEvent(eventInterface: "ClipboardEvent"): ClipboardEvent;\n    createEvent(eventInterface: "CloseEvent"): CloseEvent;\n    createEvent(eventInterface: "CompositionEvent"): CompositionEvent;\n    createEvent(eventInterface: "CustomEvent"): CustomEvent;\n    createEvent(eventInterface: "DeviceLightEvent"): DeviceLightEvent;\n    createEvent(eventInterface: "DeviceMotionEvent"): DeviceMotionEvent;\n    createEvent(eventInterface: "DeviceOrientationEvent"): DeviceOrientationEvent;\n    createEvent(eventInterface: "DragEvent"): DragEvent;\n    createEvent(eventInterface: "ErrorEvent"): ErrorEvent;\n    createEvent(eventInterface: "Event"): Event;\n    createEvent(eventInterface: "Events"): Event;\n    createEvent(eventInterface: "FocusEvent"): FocusEvent;\n    createEvent(eventInterface: "FocusNavigationEvent"): FocusNavigationEvent;\n    createEvent(eventInterface: "GamepadEvent"): GamepadEvent;\n    createEvent(eventInterface: "HashChangeEvent"): HashChangeEvent;\n    createEvent(eventInterface: "IDBVersionChangeEvent"): IDBVersionChangeEvent;\n    createEvent(eventInterface: "InputEvent"): InputEvent;\n    createEvent(eventInterface: "KeyboardEvent"): KeyboardEvent;\n    createEvent(eventInterface: "ListeningStateChangedEvent"): ListeningStateChangedEvent;\n    createEvent(eventInterface: "MSGestureEvent"): MSGestureEvent;\n    createEvent(eventInterface: "MSMediaKeyMessageEvent"): MSMediaKeyMessageEvent;\n    createEvent(eventInterface: "MSMediaKeyNeededEvent"): MSMediaKeyNeededEvent;\n    createEvent(eventInterface: "MSPointerEvent"): MSPointerEvent;\n    createEvent(eventInterface: "MediaEncryptedEvent"): MediaEncryptedEvent;\n    createEvent(eventInterface: "MediaKeyMessageEvent"): MediaKeyMessageEvent;\n    createEvent(eventInterface: "MediaQueryListEvent"): MediaQueryListEvent;\n    createEvent(eventInterface: "MediaStreamErrorEvent"): MediaStreamErrorEvent;\n    createEvent(eventInterface: "MediaStreamEvent"): MediaStreamEvent;\n    createEvent(eventInterface: "MediaStreamTrackEvent"): MediaStreamTrackEvent;\n    createEvent(eventInterface: "MessageEvent"): MessageEvent;\n    createEvent(eventInterface: "MouseEvent"): MouseEvent;\n    createEvent(eventInterface: "MouseEvents"): MouseEvent;\n    createEvent(eventInterface: "MutationEvent"): MutationEvent;\n    createEvent(eventInterface: "MutationEvents"): MutationEvent;\n    createEvent(eventInterface: "OfflineAudioCompletionEvent"): OfflineAudioCompletionEvent;\n    createEvent(eventInterface: "OverflowEvent"): OverflowEvent;\n    createEvent(eventInterface: "PageTransitionEvent"): PageTransitionEvent;\n    createEvent(eventInterface: "PaymentRequestUpdateEvent"): PaymentRequestUpdateEvent;\n    createEvent(eventInterface: "PermissionRequestedEvent"): PermissionRequestedEvent;\n    createEvent(eventInterface: "PointerEvent"): PointerEvent;\n    createEvent(eventInterface: "PopStateEvent"): PopStateEvent;\n    createEvent(eventInterface: "ProgressEvent"): ProgressEvent;\n    createEvent(eventInterface: "PromiseRejectionEvent"): PromiseRejectionEvent;\n    createEvent(eventInterface: "RTCDTMFToneChangeEvent"): RTCDTMFToneChangeEvent;\n    createEvent(eventInterface: "RTCDataChannelEvent"): RTCDataChannelEvent;\n    createEvent(eventInterface: "RTCDtlsTransportStateChangedEvent"): RTCDtlsTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCErrorEvent"): RTCErrorEvent;\n    createEvent(eventInterface: "RTCIceCandidatePairChangedEvent"): RTCIceCandidatePairChangedEvent;\n    createEvent(eventInterface: "RTCIceGathererEvent"): RTCIceGathererEvent;\n    createEvent(eventInterface: "RTCIceTransportStateChangedEvent"): RTCIceTransportStateChangedEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceErrorEvent"): RTCPeerConnectionIceErrorEvent;\n    createEvent(eventInterface: "RTCPeerConnectionIceEvent"): RTCPeerConnectionIceEvent;\n    createEvent(eventInterface: "RTCSsrcConflictEvent"): RTCSsrcConflictEvent;\n    createEvent(eventInterface: "RTCStatsEvent"): RTCStatsEvent;\n    createEvent(eventInterface: "RTCTrackEvent"): RTCTrackEvent;\n    createEvent(eventInterface: "SVGZoomEvent"): SVGZoomEvent;\n    createEvent(eventInterface: "SVGZoomEvents"): SVGZoomEvent;\n    createEvent(eventInterface: "SecurityPolicyViolationEvent"): SecurityPolicyViolationEvent;\n    createEvent(eventInterface: "ServiceWorkerMessageEvent"): ServiceWorkerMessageEvent;\n    createEvent(eventInterface: "SpeechRecognitionEvent"): SpeechRecognitionEvent;\n    createEvent(eventInterface: "SpeechSynthesisErrorEvent"): SpeechSynthesisErrorEvent;\n    createEvent(eventInterface: "SpeechSynthesisEvent"): SpeechSynthesisEvent;\n    createEvent(eventInterface: "StorageEvent"): StorageEvent;\n    createEvent(eventInterface: "TextEvent"): TextEvent;\n    createEvent(eventInterface: "TouchEvent"): TouchEvent;\n    createEvent(eventInterface: "TrackEvent"): TrackEvent;\n    createEvent(eventInterface: "TransitionEvent"): TransitionEvent;\n    createEvent(eventInterface: "UIEvent"): UIEvent;\n    createEvent(eventInterface: "UIEvents"): UIEvent;\n    createEvent(eventInterface: "VRDisplayEvent"): VRDisplayEvent;\n    createEvent(eventInterface: "VRDisplayEvent "): VRDisplayEvent ;\n    createEvent(eventInterface: "WebGLContextEvent"): WebGLContextEvent;\n    createEvent(eventInterface: "WheelEvent"): WheelEvent;\n    createEvent(eventInterface: string): Event;\n}\n\n/** A minimal document object that has no parent. It is used as a lightweight version of Document that stores a segment of a document structure comprised of nodes just like a standard document. The key difference is that because the document fragment isn\'t part of the active document tree structure, changes made to the fragment don\'t affect the document, cause reflow, or incur any performance impact that can occur when changes are made. */\ninterface DocumentFragment extends Node, NonElementParentNode, ParentNode {\n    readonly ownerDocument: Document;\n    getElementById(elementId: string): HTMLElement | null;\n}\n\ndeclare var DocumentFragment: {\n    prototype: DocumentFragment;\n    new(): DocumentFragment;\n};\n\ninterface DocumentOrShadowRoot {\n    readonly activeElement: Element | null;\n    /**\n     * Returns document\'s fullscreen element.\n     */\n    readonly fullscreenElement: Element | null;\n    readonly pointerLockElement: Element | null;\n    /**\n     * Retrieves a collection of styleSheet objects representing the style sheets that correspond to each instance of a link or style object in the document.\n     */\n    readonly styleSheets: StyleSheetList;\n    caretPositionFromPoint(x: number, y: number): CaretPosition | null;\n    /** @deprecated */\n    caretRangeFromPoint(x: number, y: number): Range;\n    elementFromPoint(x: number, y: number): Element | null;\n    elementsFromPoint(x: number, y: number): Element[];\n    getSelection(): Selection | null;\n}\n\ninterface DocumentTimeline extends AnimationTimeline {\n}\n\ndeclare var DocumentTimeline: {\n    prototype: DocumentTimeline;\n    new(options?: DocumentTimelineOptions): DocumentTimeline;\n};\n\n/** A Node containing a doctype. */\ninterface DocumentType extends Node, ChildNode {\n    readonly name: string;\n    readonly ownerDocument: Document;\n    readonly publicId: string;\n    readonly systemId: string;\n}\n\ndeclare var DocumentType: {\n    prototype: DocumentType;\n    new(): DocumentType;\n};\n\n/** A DOM event that represents a drag and drop interaction. The user initiates a drag by placing a pointer device (such as a mouse) on the touch surface and then dragging the pointer to a new location (such as another DOM element). Applications are free to interpret a drag and drop interaction in an application-specific way. */\ninterface DragEvent extends MouseEvent {\n    /**\n     * Returns the DataTransfer object for the event.\n     */\n    readonly dataTransfer: DataTransfer | null;\n}\n\ndeclare var DragEvent: {\n    prototype: DragEvent;\n    new(type: string, eventInitDict?: DragEventInit): DragEvent;\n};\n\n/** Inherits properties from its parent, AudioNode. */\ninterface DynamicsCompressorNode extends AudioNode {\n    readonly attack: AudioParam;\n    readonly knee: AudioParam;\n    readonly ratio: AudioParam;\n    readonly reduction: number;\n    readonly release: AudioParam;\n    readonly threshold: AudioParam;\n}\n\ndeclare var DynamicsCompressorNode: {\n    prototype: DynamicsCompressorNode;\n    new(context: BaseAudioContext, options?: DynamicsCompressorOptions): DynamicsCompressorNode;\n};\n\ninterface EXT_blend_minmax {\n    readonly MAX_EXT: GLenum;\n    readonly MIN_EXT: GLenum;\n}\n\n/** The EXT_frag_depth extension is part of the WebGL API and enables to set a depth value of a fragment from within the fragment shader. */\ninterface EXT_frag_depth {\n}\n\ninterface EXT_sRGB {\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING_EXT: GLenum;\n    readonly SRGB8_ALPHA8_EXT: GLenum;\n    readonly SRGB_ALPHA_EXT: GLenum;\n    readonly SRGB_EXT: GLenum;\n}\n\ninterface EXT_shader_texture_lod {\n}\n\n/** The EXT_texture_filter_anisotropic extension is part of the WebGL API and exposes two constants for anisotropic filtering (AF). */\ninterface EXT_texture_filter_anisotropic {\n    readonly MAX_TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n    readonly TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n}\n\ninterface ElementEventMap {\n    "fullscreenchange": Event;\n    "fullscreenerror": Event;\n}\n\n/** Element is the most general base class from which all objects in a Document inherit. It only has methods and properties common to all kinds of elements. More specific classes inherit from Element. */\ninterface Element extends Node, Animatable, ChildNode, InnerHTML, NonDocumentTypeChildNode, ParentNode, Slottable {\n    readonly assignedSlot: HTMLSlotElement | null;\n    readonly attributes: NamedNodeMap;\n    /**\n     * Allows for manipulation of element\'s class content attribute as a set of whitespace-separated tokens through a DOMTokenList object.\n     */\n    readonly classList: DOMTokenList;\n    /**\n     * Returns the value of element\'s class content attribute. Can be set to change it.\n     */\n    className: string;\n    readonly clientHeight: number;\n    readonly clientLeft: number;\n    readonly clientTop: number;\n    readonly clientWidth: number;\n    /**\n     * Returns the value of element\'s id content attribute. Can be set to change it.\n     */\n    id: string;\n    /**\n     * Returns the local name.\n     */\n    readonly localName: string;\n    /**\n     * Returns the namespace.\n     */\n    readonly namespaceURI: string | null;\n    onfullscreenchange: ((this: Element, ev: Event) => any) | null;\n    onfullscreenerror: ((this: Element, ev: Event) => any) | null;\n    outerHTML: string;\n    readonly ownerDocument: Document;\n    /**\n     * Returns the namespace prefix.\n     */\n    readonly prefix: string | null;\n    readonly scrollHeight: number;\n    scrollLeft: number;\n    scrollTop: number;\n    readonly scrollWidth: number;\n    /**\n     * Returns element\'s shadow root, if any, and if shadow root\'s mode is "open", and null otherwise.\n     */\n    readonly shadowRoot: ShadowRoot | null;\n    /**\n     * Returns the value of element\'s slot content attribute. Can be set to change it.\n     */\n    slot: string;\n    /**\n     * Returns the HTML-uppercased qualified name.\n     */\n    readonly tagName: string;\n    /**\n     * Creates a shadow root for element and returns it.\n     */\n    attachShadow(init: ShadowRootInit): ShadowRoot;\n    /**\n     * Returns the first (starting at element) inclusive ancestor that matches selectors, and null otherwise.\n     */\n    closest<K extends keyof HTMLElementTagNameMap>(selector: K): HTMLElementTagNameMap[K] | null;\n    closest<K extends keyof SVGElementTagNameMap>(selector: K): SVGElementTagNameMap[K] | null;\n    closest<E extends Element = Element>(selector: string): E | null;\n    /**\n     * Returns element\'s first attribute whose qualified name is qualifiedName, and null if there is no such attribute otherwise.\n     */\n    getAttribute(qualifiedName: string): string | null;\n    /**\n     * Returns element\'s attribute whose namespace is namespace and local name is localName, and null if there is no such attribute otherwise.\n     */\n    getAttributeNS(namespace: string | null, localName: string): string | null;\n    /**\n     * Returns the qualified names of all element\'s attributes. Can contain duplicates.\n     */\n    getAttributeNames(): string[];\n    getAttributeNode(qualifiedName: string): Attr | null;\n    getAttributeNodeNS(namespace: string | null, localName: string): Attr | null;\n    getBoundingClientRect(): DOMRect;\n    getClientRects(): DOMRectList;\n    /**\n     * Returns a HTMLCollection of the elements in the object on which the method was invoked (a document or an element) that have all the classes given by classNames. The classNames argument is interpreted as a space-separated list of classes.\n     */\n    getElementsByClassName(classNames: string): HTMLCollectionOf<Element>;\n    getElementsByTagName<K extends keyof HTMLElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<HTMLElementTagNameMap[K]>;\n    getElementsByTagName<K extends keyof SVGElementTagNameMap>(qualifiedName: K): HTMLCollectionOf<SVGElementTagNameMap[K]>;\n    getElementsByTagName(qualifiedName: string): HTMLCollectionOf<Element>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/1999/xhtml", localName: string): HTMLCollectionOf<HTMLElement>;\n    getElementsByTagNameNS(namespaceURI: "http://www.w3.org/2000/svg", localName: string): HTMLCollectionOf<SVGElement>;\n    getElementsByTagNameNS(namespaceURI: string, localName: string): HTMLCollectionOf<Element>;\n    /**\n     * Returns true if element has an attribute whose qualified name is qualifiedName, and false otherwise.\n     */\n    hasAttribute(qualifiedName: string): boolean;\n    /**\n     * Returns true if element has an attribute whose namespace is namespace and local name is localName.\n     */\n    hasAttributeNS(namespace: string | null, localName: string): boolean;\n    /**\n     * Returns true if element has attributes, and false otherwise.\n     */\n    hasAttributes(): boolean;\n    hasPointerCapture(pointerId: number): boolean;\n    insertAdjacentElement(position: InsertPosition, insertedElement: Element): Element | null;\n    insertAdjacentHTML(where: InsertPosition, html: string): void;\n    insertAdjacentText(where: InsertPosition, text: string): void;\n    /**\n     * Returns true if matching selectors against element\'s root yields element, and false otherwise.\n     */\n    matches(selectors: string): boolean;\n    msGetRegionContent(): any;\n    releasePointerCapture(pointerId: number): void;\n    /**\n     * Removes element\'s first attribute whose qualified name is qualifiedName.\n     */\n    removeAttribute(qualifiedName: string): void;\n    /**\n     * Removes element\'s attribute whose namespace is namespace and local name is localName.\n     */\n    removeAttributeNS(namespace: string | null, localName: string): void;\n    removeAttributeNode(attr: Attr): Attr;\n    /**\n     * Displays element fullscreen and resolves promise when done.\n     * \n     * When supplied, options\'s navigationUI member indicates whether showing navigation UI while in fullscreen is preferred or not. If set to "show", navigation simplicity is preferred over screen space, and if set to "hide", more screen space is preferred. User agents are always free to honor user preference over the application\'s. The default value "auto" indicates no application preference.\n     */\n    requestFullscreen(options?: FullscreenOptions): Promise<void>;\n    requestPointerLock(): void;\n    scroll(options?: ScrollToOptions): void;\n    scroll(x: number, y: number): void;\n    scrollBy(options?: ScrollToOptions): void;\n    scrollBy(x: number, y: number): void;\n    scrollIntoView(arg?: boolean | ScrollIntoViewOptions): void;\n    scrollTo(options?: ScrollToOptions): void;\n    scrollTo(x: number, y: number): void;\n    /**\n     * Sets the value of element\'s first attribute whose qualified name is qualifiedName to value.\n     */\n    setAttribute(qualifiedName: string, value: string): void;\n    /**\n     * Sets the value of element\'s attribute whose namespace is namespace and local name is localName to value.\n     */\n    setAttributeNS(namespace: string | null, qualifiedName: string, value: string): void;\n    setAttributeNode(attr: Attr): Attr | null;\n    setAttributeNodeNS(attr: Attr): Attr | null;\n    setPointerCapture(pointerId: number): void;\n    /**\n     * If force is not given, "toggles" qualifiedName, removing it if it is present and adding it if it is not present. If force is true, adds qualifiedName. If force is false, removes qualifiedName.\n     * \n     * Returns true if qualifiedName is now present, and false otherwise.\n     */\n    toggleAttribute(qualifiedName: string, force?: boolean): boolean;\n    webkitMatchesSelector(selectors: string): boolean;\n    addEventListener<K extends keyof ElementEventMap>(type: K, listener: (this: Element, ev: ElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ElementEventMap>(type: K, listener: (this: Element, ev: ElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Element: {\n    prototype: Element;\n    new(): Element;\n};\n\ninterface ElementCSSInlineStyle {\n    readonly style: CSSStyleDeclaration;\n}\n\ninterface ElementContentEditable {\n    contentEditable: string;\n    enterKeyHint: string;\n    inputMode: string;\n    readonly isContentEditable: boolean;\n}\n\n/** Events providing information related to errors in scripts or in files. */\ninterface ErrorEvent extends Event {\n    readonly colno: number;\n    readonly error: any;\n    readonly filename: string;\n    readonly lineno: number;\n    readonly message: string;\n}\n\ndeclare var ErrorEvent: {\n    prototype: ErrorEvent;\n    new(type: string, eventInitDict?: ErrorEventInit): ErrorEvent;\n};\n\n/** An event which takes place in the DOM. */\ninterface Event {\n    /**\n     * Returns true or false depending on how event was initialized. True if event goes through its target\'s ancestors in reverse tree order, and false otherwise.\n     */\n    readonly bubbles: boolean;\n    cancelBubble: boolean;\n    /**\n     * Returns true or false depending on how event was initialized. Its return value does not always carry meaning, but true can indicate that part of the operation during which event was dispatched, can be canceled by invoking the preventDefault() method.\n     */\n    readonly cancelable: boolean;\n    /**\n     * Returns true or false depending on how event was initialized. True if event invokes listeners past a ShadowRoot node that is the root of its target, and false otherwise.\n     */\n    readonly composed: boolean;\n    /**\n     * Returns the object whose event listener\'s callback is currently being invoked.\n     */\n    readonly currentTarget: EventTarget | null;\n    /**\n     * Returns true if preventDefault() was invoked successfully to indicate cancelation, and false otherwise.\n     */\n    readonly defaultPrevented: boolean;\n    /**\n     * Returns the event\'s phase, which is one of NONE, CAPTURING_PHASE, AT_TARGET, and BUBBLING_PHASE.\n     */\n    readonly eventPhase: number;\n    /**\n     * Returns true if event was dispatched by the user agent, and false otherwise.\n     */\n    readonly isTrusted: boolean;\n    returnValue: boolean;\n    /** @deprecated */\n    readonly srcElement: EventTarget | null;\n    /**\n     * Returns the object to which event is dispatched (its target).\n     */\n    readonly target: EventTarget | null;\n    /**\n     * Returns the event\'s timestamp as the number of milliseconds measured relative to the time origin.\n     */\n    readonly timeStamp: number;\n    /**\n     * Returns the type of event, e.g. "click", "hashchange", or "submit".\n     */\n    readonly type: string;\n    /**\n     * Returns the invocation target objects of event\'s path (objects on which listeners will be invoked), except for any nodes in shadow trees of which the shadow root\'s mode is "closed" that are not reachable from event\'s currentTarget.\n     */\n    composedPath(): EventTarget[];\n    initEvent(type: string, bubbles?: boolean, cancelable?: boolean): void;\n    /**\n     * If invoked when the cancelable attribute value is true, and while executing a listener for the event with passive set to false, signals to the operation that caused event to be dispatched that it needs to be canceled.\n     */\n    preventDefault(): void;\n    /**\n     * Invoking this method prevents event from reaching any registered event listeners after the current one finishes running and, when dispatched in a tree, also prevents event from reaching any other objects.\n     */\n    stopImmediatePropagation(): void;\n    /**\n     * When dispatched in a tree, invoking this method prevents event from reaching any objects other than the current object.\n     */\n    stopPropagation(): void;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n}\n\ndeclare var Event: {\n    prototype: Event;\n    new(type: string, eventInitDict?: EventInit): Event;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n};\n\ninterface EventListenerObject {\n    handleEvent(evt: Event): void;\n}\n\ninterface EventSourceEventMap {\n    "error": Event;\n    "message": MessageEvent;\n    "open": Event;\n}\n\ninterface EventSource extends EventTarget {\n    onerror: ((this: EventSource, ev: Event) => any) | null;\n    onmessage: ((this: EventSource, ev: MessageEvent) => any) | null;\n    onopen: ((this: EventSource, ev: Event) => any) | null;\n    /**\n     * Returns the state of this EventSource object\'s connection. It can have the values described below.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the URL providing the event stream.\n     */\n    readonly url: string;\n    /**\n     * Returns true if the credentials mode for connection requests to the URL providing the event stream is set to "include", and false otherwise.\n     */\n    readonly withCredentials: boolean;\n    /**\n     * Aborts any instances of the fetch algorithm started for this EventSource object, and sets the readyState attribute to CLOSED.\n     */\n    close(): void;\n    readonly CLOSED: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    addEventListener<K extends keyof EventSourceEventMap>(type: K, listener: (this: EventSource, ev: EventSourceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof EventSourceEventMap>(type: K, listener: (this: EventSource, ev: EventSourceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var EventSource: {\n    prototype: EventSource;\n    new(url: string, eventSourceInitDict?: EventSourceInit): EventSource;\n    readonly CLOSED: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n};\n\n/** EventTarget is a DOM interface implemented by objects that can receive events and may have listeners for them. */\ninterface EventTarget {\n    /**\n     * Appends an event listener for events whose type attribute value is type. The callback argument sets the callback that will be invoked when the event is dispatched.\n     * \n     * The options argument sets listener-specific options. For compatibility this can be a boolean, in which case the method behaves exactly as if the value was specified as options\'s capture.\n     * \n     * When set to true, options\'s capture prevents callback from being invoked when the event\'s eventPhase attribute value is BUBBLING_PHASE. When false (or not present), callback will not be invoked when event\'s eventPhase attribute value is CAPTURING_PHASE. Either way, callback will be invoked if event\'s eventPhase attribute value is AT_TARGET.\n     * \n     * When set to true, options\'s passive indicates that the callback will not cancel the event by invoking preventDefault(). This is used to enable performance optimizations described in § 2.8 Observing event listeners.\n     * \n     * When set to true, options\'s once indicates that the callback will only be invoked once after which the event listener will be removed.\n     * \n     * The event listener is appended to target\'s event listener list and is not appended if it has the same type, callback, and capture.\n     */\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject | null, options?: boolean | AddEventListenerOptions): void;\n    /**\n     * Dispatches a synthetic event event to target and returns true if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n     */\n    dispatchEvent(event: Event): boolean;\n    /**\n     * Removes the event listener in target\'s event listener list with the same type, callback, and options.\n     */\n    removeEventListener(type: string, callback: EventListenerOrEventListenerObject | null, options?: EventListenerOptions | boolean): void;\n}\n\ndeclare var EventTarget: {\n    prototype: EventTarget;\n    new(): EventTarget;\n};\n\ninterface ExtensionScriptApis {\n    extensionIdToShortId(extensionId: string): number;\n    fireExtensionApiTelemetry(functionName: string, isSucceeded: boolean, isSupported: boolean, errorString: string): void;\n    genericFunction(routerAddress: any, parameters?: string, callbackId?: number): void;\n    genericSynchronousFunction(functionId: number, parameters?: string): string;\n    genericWebRuntimeCallout(to: any, from: any, payload: string): void;\n    getExtensionId(): string;\n    registerGenericFunctionCallbackHandler(callbackHandler: Function): void;\n    registerGenericPersistentCallbackHandler(callbackHandler: Function): void;\n    registerWebRuntimeCallbackHandler(handler: Function): any;\n}\n\ndeclare var ExtensionScriptApis: {\n    prototype: ExtensionScriptApis;\n    new(): ExtensionScriptApis;\n};\n\ninterface External {\n    /** @deprecated */\n    AddSearchProvider(): void;\n    /** @deprecated */\n    IsSearchProviderInstalled(): void;\n}\n\ndeclare var External: {\n    prototype: External;\n    new(): External;\n};\n\n/** Provides information about files and allows JavaScript in a web page to access their content. */\ninterface File extends Blob {\n    readonly lastModified: number;\n    readonly name: string;\n}\n\ndeclare var File: {\n    prototype: File;\n    new(fileBits: BlobPart[], fileName: string, options?: FilePropertyBag): File;\n};\n\n/** An object of this type is returned by the files property of the HTML <input> element; this lets you access the list of files selected with the <input type="file"> element. It\'s also used for a list of files dropped into web content when using the drag and drop API; see the DataTransfer object for details on this usage. */\ninterface FileList {\n    readonly length: number;\n    item(index: number): File | null;\n    [index: number]: File;\n}\n\ndeclare var FileList: {\n    prototype: FileList;\n    new(): FileList;\n};\n\ninterface FileReaderEventMap {\n    "abort": ProgressEvent<FileReader>;\n    "error": ProgressEvent<FileReader>;\n    "load": ProgressEvent<FileReader>;\n    "loadend": ProgressEvent<FileReader>;\n    "loadstart": ProgressEvent<FileReader>;\n    "progress": ProgressEvent<FileReader>;\n}\n\n/** Lets web applications asynchronously read the contents of files (or raw data buffers) stored on the user\'s computer, using File or Blob objects to specify the file or data to read. */\ninterface FileReader extends EventTarget {\n    readonly error: DOMException | null;\n    onabort: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onerror: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onload: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onloadend: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onloadstart: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onprogress: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    readonly readyState: number;\n    readonly result: string | ArrayBuffer | null;\n    abort(): void;\n    readAsArrayBuffer(blob: Blob): void;\n    readAsBinaryString(blob: Blob): void;\n    readAsDataURL(blob: Blob): void;\n    readAsText(blob: Blob, encoding?: string): void;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n    addEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var FileReader: {\n    prototype: FileReader;\n    new(): FileReader;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n};\n\n/** Focus-related events like focus, blur, focusin, or focusout. */\ninterface FocusEvent extends UIEvent {\n    readonly relatedTarget: EventTarget | null;\n}\n\ndeclare var FocusEvent: {\n    prototype: FocusEvent;\n    new(type: string, eventInitDict?: FocusEventInit): FocusEvent;\n};\n\ninterface FocusNavigationEvent extends Event {\n    readonly navigationReason: NavigationReason;\n    readonly originHeight: number;\n    readonly originLeft: number;\n    readonly originTop: number;\n    readonly originWidth: number;\n    requestFocus(): void;\n}\n\ndeclare var FocusNavigationEvent: {\n    prototype: FocusNavigationEvent;\n    new(type: string, eventInitDict?: FocusNavigationEventInit): FocusNavigationEvent;\n};\n\n/** Provides a way to easily construct a set of key/value pairs representing form fields and their values, which can then be easily sent using the XMLHttpRequest.send() method. It uses the same format a form would use if the encoding type were set to "multipart/form-data". */\ninterface FormData {\n    append(name: string, value: string | Blob, fileName?: string): void;\n    delete(name: string): void;\n    get(name: string): FormDataEntryValue | null;\n    getAll(name: string): FormDataEntryValue[];\n    has(name: string): boolean;\n    set(name: string, value: string | Blob, fileName?: string): void;\n    forEach(callbackfn: (value: FormDataEntryValue, key: string, parent: FormData) => void, thisArg?: any): void;\n}\n\ndeclare var FormData: {\n    prototype: FormData;\n    new(form?: HTMLFormElement): FormData;\n};\n\n/** A change in volume. It is an AudioNode audio-processing module that causes a given gain to be applied to the input data before its propagation to the output. A GainNode always has exactly one input and one output, both with the same number of channels. */\ninterface GainNode extends AudioNode {\n    readonly gain: AudioParam;\n}\n\ndeclare var GainNode: {\n    prototype: GainNode;\n    new(context: BaseAudioContext, options?: GainOptions): GainNode;\n};\n\n/** This Gamepad API interface defines an individual gamepad or other controller, allowing access to information such as button presses, axis positions, and id. */\ninterface Gamepad {\n    readonly axes: ReadonlyArray<number>;\n    readonly buttons: ReadonlyArray<GamepadButton>;\n    readonly connected: boolean;\n    readonly hand: GamepadHand;\n    readonly hapticActuators: ReadonlyArray<GamepadHapticActuator>;\n    readonly id: string;\n    readonly index: number;\n    readonly mapping: GamepadMappingType;\n    readonly pose: GamepadPose | null;\n    readonly timestamp: number;\n}\n\ndeclare var Gamepad: {\n    prototype: Gamepad;\n    new(): Gamepad;\n};\n\n/** An individual button of a gamepad or other controller, allowing access to the current state of different types of buttons available on the control device. */\ninterface GamepadButton {\n    readonly pressed: boolean;\n    readonly touched: boolean;\n    readonly value: number;\n}\n\ndeclare var GamepadButton: {\n    prototype: GamepadButton;\n    new(): GamepadButton;\n};\n\n/** This Gamepad API interface contains references to gamepads connected to the system, which is what the gamepad events Window.gamepadconnected and Window.gamepaddisconnected are fired in response to. */\ninterface GamepadEvent extends Event {\n    readonly gamepad: Gamepad;\n}\n\ndeclare var GamepadEvent: {\n    prototype: GamepadEvent;\n    new(type: string, eventInitDict: GamepadEventInit): GamepadEvent;\n};\n\n/** This Gamepad API interface represents hardware in the controller designed to provide haptic feedback to the user (if available), most commonly vibration hardware. */\ninterface GamepadHapticActuator {\n    readonly type: GamepadHapticActuatorType;\n    pulse(value: number, duration: number): Promise<boolean>;\n}\n\ndeclare var GamepadHapticActuator: {\n    prototype: GamepadHapticActuator;\n    new(): GamepadHapticActuator;\n};\n\n/** This Gamepad API interface represents the pose of a WebVR controller at a given timestamp (which includes orientation, position, velocity, and acceleration information.) */\ninterface GamepadPose {\n    readonly angularAcceleration: Float32Array | null;\n    readonly angularVelocity: Float32Array | null;\n    readonly hasOrientation: boolean;\n    readonly hasPosition: boolean;\n    readonly linearAcceleration: Float32Array | null;\n    readonly linearVelocity: Float32Array | null;\n    readonly orientation: Float32Array | null;\n    readonly position: Float32Array | null;\n}\n\ndeclare var GamepadPose: {\n    prototype: GamepadPose;\n    new(): GamepadPose;\n};\n\ninterface GenericTransformStream {\n    /**\n     * Returns a readable stream whose chunks are strings resulting from running encoding\'s decoder on the chunks written to writable.\n     */\n    readonly readable: ReadableStream;\n    /**\n     * Returns a writable stream which accepts [AllowShared] BufferSource chunks and runs them through encoding\'s decoder before making them available to readable.\n     * \n     * Typically this will be used via the pipeThrough() method on a ReadableStream source.\n     * \n     * ```\n     * var decoder = new TextDecoderStream(encoding);\n     * byteReadable\n     *   .pipeThrough(decoder)\n     *   .pipeTo(textWritable);\n     * ```\n     * \n     * If the error mode is "fatal" and encoding\'s decoder returns error, both readable and writable will be errored with a TypeError.\n     */\n    readonly writable: WritableStream;\n}\n\n/** An object able to programmatically obtain the position of the device. It gives Web content access to the location of the device. This allows a Web site or app to offer customized results based on the user\'s location. */\ninterface Geolocation {\n    clearWatch(watchId: number): void;\n    getCurrentPosition(successCallback: PositionCallback, errorCallback?: PositionErrorCallback, options?: PositionOptions): void;\n    watchPosition(successCallback: PositionCallback, errorCallback?: PositionErrorCallback, options?: PositionOptions): number;\n}\n\ndeclare var Geolocation: {\n    prototype: Geolocation;\n    new(): Geolocation;\n};\n\ninterface GeolocationCoordinates {\n    readonly accuracy: number;\n    readonly altitude: number | null;\n    readonly altitudeAccuracy: number | null;\n    readonly heading: number | null;\n    readonly latitude: number;\n    readonly longitude: number;\n    readonly speed: number | null;\n}\n\ndeclare var GeolocationCoordinates: {\n    prototype: GeolocationCoordinates;\n    new(): GeolocationCoordinates;\n};\n\ninterface GeolocationPosition {\n    readonly coords: GeolocationCoordinates;\n    readonly timestamp: number;\n}\n\ndeclare var GeolocationPosition: {\n    prototype: GeolocationPosition;\n    new(): GeolocationPosition;\n};\n\ninterface GeolocationPositionError {\n    readonly code: number;\n    readonly message: string;\n    readonly PERMISSION_DENIED: number;\n    readonly POSITION_UNAVAILABLE: number;\n    readonly TIMEOUT: number;\n}\n\ndeclare var GeolocationPositionError: {\n    prototype: GeolocationPositionError;\n    new(): GeolocationPositionError;\n    readonly PERMISSION_DENIED: number;\n    readonly POSITION_UNAVAILABLE: number;\n    readonly TIMEOUT: number;\n};\n\ninterface GlobalEventHandlersEventMap {\n    "abort": UIEvent;\n    "animationcancel": AnimationEvent;\n    "animationend": AnimationEvent;\n    "animationiteration": AnimationEvent;\n    "animationstart": AnimationEvent;\n    "auxclick": MouseEvent;\n    "blur": FocusEvent;\n    "cancel": Event;\n    "canplay": Event;\n    "canplaythrough": Event;\n    "change": Event;\n    "click": MouseEvent;\n    "close": Event;\n    "contextmenu": MouseEvent;\n    "cuechange": Event;\n    "dblclick": MouseEvent;\n    "drag": DragEvent;\n    "dragend": DragEvent;\n    "dragenter": DragEvent;\n    "dragexit": Event;\n    "dragleave": DragEvent;\n    "dragover": DragEvent;\n    "dragstart": DragEvent;\n    "drop": DragEvent;\n    "durationchange": Event;\n    "emptied": Event;\n    "ended": Event;\n    "error": ErrorEvent;\n    "focus": FocusEvent;\n    "focusin": FocusEvent;\n    "focusout": FocusEvent;\n    "gotpointercapture": PointerEvent;\n    "input": Event;\n    "invalid": Event;\n    "keydown": KeyboardEvent;\n    "keypress": KeyboardEvent;\n    "keyup": KeyboardEvent;\n    "load": Event;\n    "loadeddata": Event;\n    "loadedmetadata": Event;\n    "loadstart": Event;\n    "lostpointercapture": PointerEvent;\n    "mousedown": MouseEvent;\n    "mouseenter": MouseEvent;\n    "mouseleave": MouseEvent;\n    "mousemove": MouseEvent;\n    "mouseout": MouseEvent;\n    "mouseover": MouseEvent;\n    "mouseup": MouseEvent;\n    "pause": Event;\n    "play": Event;\n    "playing": Event;\n    "pointercancel": PointerEvent;\n    "pointerdown": PointerEvent;\n    "pointerenter": PointerEvent;\n    "pointerleave": PointerEvent;\n    "pointermove": PointerEvent;\n    "pointerout": PointerEvent;\n    "pointerover": PointerEvent;\n    "pointerup": PointerEvent;\n    "progress": ProgressEvent;\n    "ratechange": Event;\n    "reset": Event;\n    "resize": UIEvent;\n    "scroll": Event;\n    "securitypolicyviolation": SecurityPolicyViolationEvent;\n    "seeked": Event;\n    "seeking": Event;\n    "select": Event;\n    "selectionchange": Event;\n    "selectstart": Event;\n    "stalled": Event;\n    "submit": Event;\n    "suspend": Event;\n    "timeupdate": Event;\n    "toggle": Event;\n    "touchcancel": TouchEvent;\n    "touchend": TouchEvent;\n    "touchmove": TouchEvent;\n    "touchstart": TouchEvent;\n    "transitioncancel": TransitionEvent;\n    "transitionend": TransitionEvent;\n    "transitionrun": TransitionEvent;\n    "transitionstart": TransitionEvent;\n    "volumechange": Event;\n    "waiting": Event;\n    "wheel": WheelEvent;\n}\n\ninterface GlobalEventHandlers {\n    /**\n     * Fires when the user aborts the download.\n     * @param ev The event.\n     */\n    onabort: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    onanimationcancel: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationend: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationiteration: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onanimationstart: ((this: GlobalEventHandlers, ev: AnimationEvent) => any) | null;\n    onauxclick: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the object loses the input focus.\n     * @param ev The focus event.\n     */\n    onblur: ((this: GlobalEventHandlers, ev: FocusEvent) => any) | null;\n    oncancel: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when playback is possible, but would require further buffering.\n     * @param ev The event.\n     */\n    oncanplay: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    oncanplaythrough: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the contents of the object or selection have changed.\n     * @param ev The event.\n     */\n    onchange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user clicks the left mouse button on the object\n     * @param ev The mouse event.\n     */\n    onclick: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onclose: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user clicks the right mouse button in the client area, opening the context menu.\n     * @param ev The mouse event.\n     */\n    oncontextmenu: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    oncuechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user double-clicks the object.\n     * @param ev The mouse event.\n     */\n    ondblclick: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires on the source object continuously during a drag operation.\n     * @param ev The event.\n     */\n    ondrag: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the source object when the user releases the mouse at the close of a drag operation.\n     * @param ev The event.\n     */\n    ondragend: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the target element when the user drags the object to a valid drop target.\n     * @param ev The drag event.\n     */\n    ondragenter: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    ondragexit: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires on the target object when the user moves the mouse out of a valid drop target during a drag operation.\n     * @param ev The drag event.\n     */\n    ondragleave: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the target element continuously while the user drags the object over a valid drop target.\n     * @param ev The event.\n     */\n    ondragover: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Fires on the source object when the user starts to drag a text selection or selected object.\n     * @param ev The event.\n     */\n    ondragstart: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    ondrop: ((this: GlobalEventHandlers, ev: DragEvent) => any) | null;\n    /**\n     * Occurs when the duration attribute is updated.\n     * @param ev The event.\n     */\n    ondurationchange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the media element is reset to its initial state.\n     * @param ev The event.\n     */\n    onemptied: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the end of playback is reached.\n     * @param ev The event\n     */\n    onended: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when an error occurs during object loading.\n     * @param ev The event.\n     */\n    onerror: OnErrorEventHandler;\n    /**\n     * Fires when the object receives focus.\n     * @param ev The event.\n     */\n    onfocus: ((this: GlobalEventHandlers, ev: FocusEvent) => any) | null;\n    ongotpointercapture: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    oninput: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    oninvalid: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user presses a key.\n     * @param ev The keyboard event\n     */\n    onkeydown: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires when the user presses an alphanumeric key.\n     * @param ev The event.\n     */\n    onkeypress: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires when the user releases a key.\n     * @param ev The keyboard event\n     */\n    onkeyup: ((this: GlobalEventHandlers, ev: KeyboardEvent) => any) | null;\n    /**\n     * Fires immediately after the browser loads the object.\n     * @param ev The event.\n     */\n    onload: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when media data is loaded at the current playback position.\n     * @param ev The event.\n     */\n    onloadeddata: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the duration and dimensions of the media have been determined.\n     * @param ev The event.\n     */\n    onloadedmetadata: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when Internet Explorer begins looking for media data.\n     * @param ev The event.\n     */\n    onloadstart: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onlostpointercapture: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    /**\n     * Fires when the user clicks the object with either mouse button.\n     * @param ev The mouse event.\n     */\n    onmousedown: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onmouseenter: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    onmouseleave: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse over the object.\n     * @param ev The mouse event.\n     */\n    onmousemove: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse pointer outside the boundaries of the object.\n     * @param ev The mouse event.\n     */\n    onmouseout: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user moves the mouse pointer into the object.\n     * @param ev The mouse event.\n     */\n    onmouseover: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Fires when the user releases a mouse button while the mouse is over the object.\n     * @param ev The mouse event.\n     */\n    onmouseup: ((this: GlobalEventHandlers, ev: MouseEvent) => any) | null;\n    /**\n     * Occurs when playback is paused.\n     * @param ev The event.\n     */\n    onpause: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the play method is requested.\n     * @param ev The event.\n     */\n    onplay: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the audio or video has started playing.\n     * @param ev The event.\n     */\n    onplaying: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onpointercancel: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerdown: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerenter: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerleave: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointermove: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerout: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerover: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    onpointerup: ((this: GlobalEventHandlers, ev: PointerEvent) => any) | null;\n    /**\n     * Occurs to indicate progress while downloading media data.\n     * @param ev The event.\n     */\n    onprogress: ((this: GlobalEventHandlers, ev: ProgressEvent) => any) | null;\n    /**\n     * Occurs when the playback rate is increased or decreased.\n     * @param ev The event.\n     */\n    onratechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the user resets a form.\n     * @param ev The event.\n     */\n    onreset: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onresize: ((this: GlobalEventHandlers, ev: UIEvent) => any) | null;\n    /**\n     * Fires when the user repositions the scroll box in the scroll bar on the object.\n     * @param ev The event.\n     */\n    onscroll: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onsecuritypolicyviolation: ((this: GlobalEventHandlers, ev: SecurityPolicyViolationEvent) => any) | null;\n    /**\n     * Occurs when the seek operation ends.\n     * @param ev The event.\n     */\n    onseeked: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the current playback position is moved.\n     * @param ev The event.\n     */\n    onseeking: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Fires when the current selection changes.\n     * @param ev The event.\n     */\n    onselect: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onselectionchange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onselectstart: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when the download has stopped.\n     * @param ev The event.\n     */\n    onstalled: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onsubmit: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs if the load operation has been intentionally halted.\n     * @param ev The event.\n     */\n    onsuspend: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs to indicate the current playback position.\n     * @param ev The event.\n     */\n    ontimeupdate: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    ontoggle: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    ontouchcancel?: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchend?: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchmove?: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontouchstart?: ((this: GlobalEventHandlers, ev: TouchEvent) => any) | null;\n    ontransitioncancel: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionend: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionrun: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    ontransitionstart: ((this: GlobalEventHandlers, ev: TransitionEvent) => any) | null;\n    /**\n     * Occurs when the volume is changed, or playback is muted or unmuted.\n     * @param ev The event.\n     */\n    onvolumechange: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    /**\n     * Occurs when playback stops because the next frame of a video resource is not available.\n     * @param ev The event.\n     */\n    onwaiting: ((this: GlobalEventHandlers, ev: Event) => any) | null;\n    onwheel: ((this: GlobalEventHandlers, ev: WheelEvent) => any) | null;\n    addEventListener<K extends keyof GlobalEventHandlersEventMap>(type: K, listener: (this: GlobalEventHandlers, ev: GlobalEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof GlobalEventHandlersEventMap>(type: K, listener: (this: GlobalEventHandlers, ev: GlobalEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface HTMLAllCollection {\n    /**\n     * Returns the number of elements in the collection.\n     */\n    readonly length: number;\n    /**\n     * Returns the item with index index from the collection (determined by tree order).\n     */\n    item(nameOrIndex?: string): HTMLCollection | Element | null;\n    /**\n     * Returns the item with ID or name name from the collection.\n     * \n     * If there are multiple matching items, then an HTMLCollection object containing all those elements is returned.\n     * \n     * Only button, form, iframe, input, map, meta, object, select, and textarea elements can have a name for the purpose of this method; their name is given by the value of their name attribute.\n     */\n    namedItem(name: string): HTMLCollection | Element | null;\n    [index: number]: Element;\n}\n\ndeclare var HTMLAllCollection: {\n    prototype: HTMLAllCollection;\n    new(): HTMLAllCollection;\n};\n\n/** Hyperlink elements and provides special properties and methods (beyond those of the regular HTMLElement object interface that they inherit from) for manipulating the layout and presentation of such elements. */\ninterface HTMLAnchorElement extends HTMLElement, HTMLHyperlinkElementUtils {\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    /**\n     * Sets or retrieves the coordinates of the object.\n     */\n    /** @deprecated */\n    coords: string;\n    download: string;\n    /**\n     * Sets or retrieves the language code of the object.\n     */\n    hreflang: string;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    name: string;\n    ping: string;\n    referrerPolicy: string;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    /** @deprecated */\n    rev: string;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    shape: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAnchorElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAnchorElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAnchorElement: {\n    prototype: HTMLAnchorElement;\n    new(): HTMLAnchorElement;\n};\n\ninterface HTMLAppletElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    /** @deprecated */\n    alt: string;\n    /**\n     * Sets or retrieves a character string that can be used to implement your own archive functionality for the object.\n     */\n    /** @deprecated */\n    archive: string;\n    /** @deprecated */\n    code: string;\n    /**\n     * Sets or retrieves the URL of the component.\n     */\n    /** @deprecated */\n    codeBase: string;\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    /** @deprecated */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /** @deprecated */\n    object: string;\n    /** @deprecated */\n    vspace: number;\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAppletElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAppletElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAppletElement: {\n    prototype: HTMLAppletElement;\n    new(): HTMLAppletElement;\n};\n\n/** Provides special properties and methods (beyond those of the regular object HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of <area> elements. */\ninterface HTMLAreaElement extends HTMLElement, HTMLHyperlinkElementUtils {\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Sets or retrieves the coordinates of the object.\n     */\n    coords: string;\n    download: string;\n    /**\n     * Sets or gets whether clicks in this region cause action.\n     */\n    /** @deprecated */\n    noHref: boolean;\n    ping: string;\n    referrerPolicy: string;\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the shape of the object.\n     */\n    shape: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAreaElement: {\n    prototype: HTMLAreaElement;\n    new(): HTMLAreaElement;\n};\n\n/** Provides access to the properties of <audio> elements, as well as methods to manipulate them. It derives from the HTMLMediaElement interface. */\ninterface HTMLAudioElement extends HTMLMediaElement {\n    addEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLAudioElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLAudioElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLAudioElement: {\n    prototype: HTMLAudioElement;\n    new(): HTMLAudioElement;\n};\n\n/** A HTML line break element (<br>). It inherits from HTMLElement. */\ninterface HTMLBRElement extends HTMLElement {\n    /**\n     * Sets or retrieves the side on which floating objects are not to be positioned when any IHTMLBlockElement is inserted into the document.\n     */\n    /** @deprecated */\n    clear: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBRElement: {\n    prototype: HTMLBRElement;\n    new(): HTMLBRElement;\n};\n\n/** Contains the base URI for a document. This object inherits all of the properties and methods as described in the HTMLElement interface. */\ninterface HTMLBaseElement extends HTMLElement {\n    /**\n     * Gets or sets the baseline URL on which relative links are based.\n     */\n    href: string;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBaseElement: {\n    prototype: HTMLBaseElement;\n    new(): HTMLBaseElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <basefont> elements. */\ninterface HTMLBaseFontElement extends HTMLElement, DOML2DeprecatedColorProperty {\n    /**\n     * Sets or retrieves the current typeface family.\n     */\n    /** @deprecated */\n    face: string;\n    /**\n     * Sets or retrieves the font size of the object.\n     */\n    /** @deprecated */\n    size: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLBaseFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBaseFontElement: {\n    prototype: HTMLBaseFontElement;\n    new(): HTMLBaseFontElement;\n};\n\ninterface HTMLBodyElementEventMap extends HTMLElementEventMap, WindowEventHandlersEventMap {\n    "orientationchange": Event;\n}\n\n/** Provides special properties (beyond those inherited from the regular HTMLElement interface) for manipulating <body> elements. */\ninterface HTMLBodyElement extends HTMLElement, WindowEventHandlers {\n    /** @deprecated */\n    aLink: string;\n    /** @deprecated */\n    background: string;\n    /** @deprecated */\n    bgColor: string;\n    /** @deprecated */\n    link: string;\n    /** @deprecated */\n    onorientationchange: ((this: HTMLBodyElement, ev: Event) => any) | null;\n    /** @deprecated */\n    text: string;\n    /** @deprecated */\n    vLink: string;\n    addEventListener<K extends keyof HTMLBodyElementEventMap>(type: K, listener: (this: HTMLBodyElement, ev: HTMLBodyElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLBodyElementEventMap>(type: K, listener: (this: HTMLBodyElement, ev: HTMLBodyElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLBodyElement: {\n    prototype: HTMLBodyElement;\n    new(): HTMLBodyElement;\n};\n\n/** Provides properties and methods (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <button> elements. */\ninterface HTMLButtonElement extends HTMLElement {\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Overrides the action attribute (where the data on a form is sent) on the parent form element.\n     */\n    formAction: string;\n    /**\n     * Used to override the encoding (formEnctype attribute) specified on the form element.\n     */\n    formEnctype: string;\n    /**\n     * Overrides the submit method attribute previously specified on a form element.\n     */\n    formMethod: string;\n    /**\n     * Overrides any validation or required attributes on a form or form elements to allow it to be submitted without validation. This can be used to create a "save draft"-type submit option.\n     */\n    formNoValidate: boolean;\n    /**\n     * Overrides the target attribute on a form element.\n     */\n    formTarget: string;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets the classification and default behavior of the button.\n     */\n    type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Sets or retrieves the default or selected value of the control.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLButtonElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLButtonElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLButtonElement: {\n    prototype: HTMLButtonElement;\n    new(): HTMLButtonElement;\n};\n\n/** Provides properties and methods for manipulating the layout and presentation of <canvas> elements. The HTMLCanvasElement interface also inherits the properties and methods of the HTMLElement interface. */\ninterface HTMLCanvasElement extends HTMLElement {\n    /**\n     * Gets or sets the height of a canvas element on a document.\n     */\n    height: number;\n    /**\n     * Gets or sets the width of a canvas element on a document.\n     */\n    width: number;\n    /**\n     * Returns an object that provides methods and properties for drawing and manipulating images and graphics on a canvas element in a document. A context object includes information about colors, line widths, fonts, and other graphic parameters that can be drawn on a canvas.\n     * @param contextId The identifier (ID) of the type of canvas to create. Internet Explorer 9 and Internet Explorer 10 support only a 2-D context using canvas.getContext("2d"); IE11 Preview also supports 3-D or WebGL context using canvas.getContext("experimental-webgl");\n     */\n    getContext(contextId: "2d", options?: CanvasRenderingContext2DSettings): CanvasRenderingContext2D | null;\n    getContext(contextId: "bitmaprenderer", options?: ImageBitmapRenderingContextSettings): ImageBitmapRenderingContext | null;\n    getContext(contextId: "webgl", options?: WebGLContextAttributes): WebGLRenderingContext | null;\n    getContext(contextId: "webgl2", options?: WebGLContextAttributes): WebGL2RenderingContext | null;\n    getContext(contextId: string, options?: any): RenderingContext | null;\n    toBlob(callback: BlobCallback, type?: string, quality?: any): void;\n    /**\n     * Returns the content of the current canvas as an image that you can use as a source for another canvas or an HTML element.\n     * @param type The standard MIME type for the image format to return. If you do not specify this parameter, the default value is a PNG format image.\n     */\n    toDataURL(type?: string, quality?: any): string;\n    transferControlToOffscreen(): OffscreenCanvas;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLCanvasElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLCanvasElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLCanvasElement: {\n    prototype: HTMLCanvasElement;\n    new(): HTMLCanvasElement;\n};\n\n/** A generic collection (array-like object similar to arguments) of elements (in document order) and offers methods and properties for selecting from the list. */\ninterface HTMLCollectionBase {\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    readonly length: number;\n    /**\n     * Retrieves an object from various collections.\n     */\n    item(index: number): Element | null;\n    [index: number]: Element;\n}\n\ninterface HTMLCollection extends HTMLCollectionBase {\n    /**\n     * Retrieves a select object or an object from an options collection.\n     */\n    namedItem(name: string): Element | null;\n}\n\ndeclare var HTMLCollection: {\n    prototype: HTMLCollection;\n    new(): HTMLCollection;\n};\n\ninterface HTMLCollectionOf<T extends Element> extends HTMLCollectionBase {\n    item(index: number): T | null;\n    namedItem(name: string): T | null;\n    [index: number]: T;\n}\n\n/** Provides special properties (beyond those of the regular HTMLElement interface it also has available to it by inheritance) for manipulating definition list (<dl>) elements. */\ninterface HTMLDListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDListElement: {\n    prototype: HTMLDListElement;\n    new(): HTMLDListElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <data> elements. */\ninterface HTMLDataElement extends HTMLElement {\n    value: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDataElement: {\n    prototype: HTMLDataElement;\n    new(): HTMLDataElement;\n};\n\n/** Provides special properties (beyond the HTMLElement object interface it also has available to it by inheritance) to manipulate <datalist> elements and their content. */\ninterface HTMLDataListElement extends HTMLElement {\n    readonly options: HTMLCollectionOf<HTMLOptionElement>;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDataListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDataListElement: {\n    prototype: HTMLDataListElement;\n    new(): HTMLDataListElement;\n};\n\ninterface HTMLDetailsElement extends HTMLElement {\n    open: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDetailsElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDetailsElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDetailsElement: {\n    prototype: HTMLDetailsElement;\n    new(): HTMLDetailsElement;\n};\n\ninterface HTMLDialogElement extends HTMLElement {\n    open: boolean;\n    returnValue: string;\n    close(returnValue?: string): void;\n    show(): void;\n    showModal(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDialogElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDialogElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDialogElement: {\n    prototype: HTMLDialogElement;\n    new(): HTMLDialogElement;\n};\n\ninterface HTMLDirectoryElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDirectoryElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDirectoryElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDirectoryElement: {\n    prototype: HTMLDirectoryElement;\n    new(): HTMLDirectoryElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <div> elements. */\ninterface HTMLDivElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDivElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLDivElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDivElement: {\n    prototype: HTMLDivElement;\n    new(): HTMLDivElement;\n};\n\n/** The HTMLDocument property of Window objects is an alias that browsers expose for the Document interface object. */\ninterface HTMLDocument extends Document {\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: HTMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: HTMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLDocument: {\n    prototype: HTMLDocument;\n    new(): HTMLDocument;\n};\n\ninterface HTMLElementEventMap extends ElementEventMap, GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n}\n\n/** Any HTML element. Some elements directly implement this interface, while others implement it via an interface that inherits it. */\ninterface HTMLElement extends Element, DocumentAndElementEventHandlers, ElementCSSInlineStyle, ElementCSSInlineStyle, ElementContentEditable, GlobalEventHandlers, HTMLOrSVGElement {\n    accessKey: string;\n    readonly accessKeyLabel: string;\n    autocapitalize: string;\n    dir: string;\n    draggable: boolean;\n    hidden: boolean;\n    innerText: string;\n    lang: string;\n    readonly offsetHeight: number;\n    readonly offsetLeft: number;\n    readonly offsetParent: Element | null;\n    readonly offsetTop: number;\n    readonly offsetWidth: number;\n    spellcheck: boolean;\n    title: string;\n    translate: boolean;\n    click(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLElement: {\n    prototype: HTMLElement;\n    new(): HTMLElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <embed> elements. */\ninterface HTMLEmbedElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    src: string;\n    type: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    getSVGDocument(): Document | null;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLEmbedElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLEmbedElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLEmbedElement: {\n    prototype: HTMLEmbedElement;\n    new(): HTMLEmbedElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of <fieldset> elements. */\ninterface HTMLFieldSetElement extends HTMLElement {\n    disabled: boolean;\n    readonly elements: HTMLCollection;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    name: string;\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFieldSetElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFieldSetElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFieldSetElement: {\n    prototype: HTMLFieldSetElement;\n    new(): HTMLFieldSetElement;\n};\n\n/** Implements the document object model (DOM) representation of the font element. The HTML Font Element <font> defines the font size, font face and color of text. */\ninterface HTMLFontElement extends HTMLElement {\n    /** @deprecated */\n    color: string;\n    /**\n     * Sets or retrieves the current typeface family.\n     */\n    /** @deprecated */\n    face: string;\n    /** @deprecated */\n    size: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFontElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFontElement: {\n    prototype: HTMLFontElement;\n    new(): HTMLFontElement;\n};\n\n/** A collection of HTML form control elements.  */\ninterface HTMLFormControlsCollection extends HTMLCollectionBase {\n    /**\n     * Returns the item with ID or name name from the collection.\n     * \n     * If there are multiple matching items, then a RadioNodeList object containing all those elements is returned.\n     */\n    namedItem(name: string): RadioNodeList | Element | null;\n}\n\ndeclare var HTMLFormControlsCollection: {\n    prototype: HTMLFormControlsCollection;\n    new(): HTMLFormControlsCollection;\n};\n\n/** A <form> element in the DOM; it allows access to and in some cases modification of aspects of the form, as well as access to its component elements. */\ninterface HTMLFormElement extends HTMLElement {\n    /**\n     * Sets or retrieves a list of character encodings for input data that must be accepted by the server processing the form.\n     */\n    acceptCharset: string;\n    /**\n     * Sets or retrieves the URL to which the form content is sent for processing.\n     */\n    action: string;\n    /**\n     * Specifies whether autocomplete is applied to an editable text field.\n     */\n    autocomplete: string;\n    /**\n     * Retrieves a collection, in source order, of all controls in a given form.\n     */\n    readonly elements: HTMLFormControlsCollection;\n    /**\n     * Sets or retrieves the MIME encoding for the form.\n     */\n    encoding: string;\n    /**\n     * Sets or retrieves the encoding type for the form.\n     */\n    enctype: string;\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    readonly length: number;\n    /**\n     * Sets or retrieves how to send the form data to the server.\n     */\n    method: string;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Designates a form that is not validated when submitted.\n     */\n    noValidate: boolean;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    target: string;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Fires when the user resets a form.\n     */\n    reset(): void;\n    /**\n     * Fires when a FORM is about to be submitted.\n     */\n    submit(): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFormElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFormElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: Element;\n    [name: string]: any;\n}\n\ndeclare var HTMLFormElement: {\n    prototype: HTMLFormElement;\n    new(): HTMLFormElement;\n};\n\ninterface HTMLFrameElement extends HTMLElement {\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    /** @deprecated */\n    readonly contentDocument: Document | null;\n    /**\n     * Retrieves the object of the specified.\n     */\n    /** @deprecated */\n    readonly contentWindow: WindowProxy | null;\n    /**\n     * Sets or retrieves whether to display a border for the frame.\n     */\n    /** @deprecated */\n    frameBorder: string;\n    /**\n     * Sets or retrieves a URI to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /**\n     * Sets or retrieves the top and bottom margin heights before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginHeight: string;\n    /**\n     * Sets or retrieves the left and right margin widths before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginWidth: string;\n    /**\n     * Sets or retrieves the frame name.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * Sets or retrieves whether the user can resize the frame.\n     */\n    /** @deprecated */\n    noResize: boolean;\n    /**\n     * Sets or retrieves whether the frame can be scrolled.\n     */\n    /** @deprecated */\n    scrolling: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    /** @deprecated */\n    src: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFrameElement: {\n    prototype: HTMLFrameElement;\n    new(): HTMLFrameElement;\n};\n\ninterface HTMLFrameSetElementEventMap extends HTMLElementEventMap, WindowEventHandlersEventMap {\n}\n\n/** Provides special properties (beyond those of the regular HTMLElement interface they also inherit) for manipulating <frameset> elements. */\ninterface HTMLFrameSetElement extends HTMLElement, WindowEventHandlers {\n    /**\n     * Sets or retrieves the frame widths of the object.\n     */\n    /** @deprecated */\n    cols: string;\n    /**\n     * Sets or retrieves the frame heights of the object.\n     */\n    /** @deprecated */\n    rows: string;\n    addEventListener<K extends keyof HTMLFrameSetElementEventMap>(type: K, listener: (this: HTMLFrameSetElement, ev: HTMLFrameSetElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLFrameSetElementEventMap>(type: K, listener: (this: HTMLFrameSetElement, ev: HTMLFrameSetElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLFrameSetElement: {\n    prototype: HTMLFrameSetElement;\n    new(): HTMLFrameSetElement;\n};\n\n/** Provides special properties (beyond those of the HTMLElement interface it also has available to it by inheritance) for manipulating <hr> elements. */\ninterface HTMLHRElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    color: string;\n    /**\n     * Sets or retrieves whether the horizontal rule is drawn with 3-D shading.\n     */\n    /** @deprecated */\n    noShade: boolean;\n    /** @deprecated */\n    size: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHRElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHRElement: {\n    prototype: HTMLHRElement;\n    new(): HTMLHRElement;\n};\n\n/** Contains the descriptive information, or metadata, for a document. This object inherits all of the properties and methods described in the HTMLElement interface. */\ninterface HTMLHeadElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHeadElement: {\n    prototype: HTMLHeadElement;\n    new(): HTMLHeadElement;\n};\n\n/** The different heading elements. It inherits methods and properties from the HTMLElement interface. */\ninterface HTMLHeadingElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadingElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHeadingElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHeadingElement: {\n    prototype: HTMLHeadingElement;\n    new(): HTMLHeadingElement;\n};\n\n/** Serves as the root node for a given HTML document. This object inherits the properties and methods described in the HTMLElement interface. */\ninterface HTMLHtmlElement extends HTMLElement {\n    /**\n     * Sets or retrieves the DTD version that governs the current document.\n     */\n    /** @deprecated */\n    version: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHtmlElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLHtmlElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLHtmlElement: {\n    prototype: HTMLHtmlElement;\n    new(): HTMLHtmlElement;\n};\n\ninterface HTMLHyperlinkElementUtils {\n    hash: string;\n    host: string;\n    hostname: string;\n    href: string;\n    toString(): string;\n    readonly origin: string;\n    password: string;\n    pathname: string;\n    port: string;\n    protocol: string;\n    search: string;\n    username: string;\n}\n\n/** Provides special properties and methods (beyond those of the HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of inline frame elements. */\ninterface HTMLIFrameElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    allow: string;\n    allowFullscreen: boolean;\n    allowPaymentRequest: boolean;\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    readonly contentDocument: Document | null;\n    /**\n     * Retrieves the object of the specified.\n     */\n    readonly contentWindow: WindowProxy | null;\n    /**\n     * Sets or retrieves whether to display a border for the frame.\n     */\n    /** @deprecated */\n    frameBorder: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    /**\n     * Sets or retrieves a URI to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /**\n     * Sets or retrieves the top and bottom margin heights before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginHeight: string;\n    /**\n     * Sets or retrieves the left and right margin widths before displaying the text in a frame.\n     */\n    /** @deprecated */\n    marginWidth: string;\n    /**\n     * Sets or retrieves the frame name.\n     */\n    name: string;\n    referrerPolicy: ReferrerPolicy;\n    readonly sandbox: DOMTokenList;\n    /**\n     * Sets or retrieves whether the frame can be scrolled.\n     */\n    /** @deprecated */\n    scrolling: string;\n    /**\n     * Sets or retrieves a URL to be loaded by the object.\n     */\n    src: string;\n    /**\n     * Sets or retrives the content of the page that is to contain.\n     */\n    srcdoc: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    getSVGDocument(): Document | null;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLIFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLIFrameElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLIFrameElement: {\n    prototype: HTMLIFrameElement;\n    new(): HTMLIFrameElement;\n};\n\n/** Provides special properties and methods for manipulating <img> elements. */\ninterface HTMLImageElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Specifies the properties of a border drawn around an object.\n     */\n    /** @deprecated */\n    border: string;\n    /**\n     * Retrieves whether the object is fully loaded.\n     */\n    readonly complete: boolean;\n    crossOrigin: string | null;\n    readonly currentSrc: string;\n    decoding: "async" | "sync" | "auto";\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: number;\n    /**\n     * Sets or retrieves the width of the border to draw around the object.\n     */\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Sets or retrieves whether the image is a server-side image map.\n     */\n    isMap: boolean;\n    loading: string;\n    /**\n     * Sets or retrieves a Uniform Resource Identifier (URI) to a long description of the object.\n     */\n    /** @deprecated */\n    longDesc: string;\n    /** @deprecated */\n    lowsrc: string;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    /** @deprecated */\n    name: string;\n    /**\n     * The original height of the image resource before sizing.\n     */\n    readonly naturalHeight: number;\n    /**\n     * The original width of the image resource before sizing.\n     */\n    readonly naturalWidth: number;\n    referrerPolicy: string;\n    sizes: string;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcset: string;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    useMap: string;\n    /**\n     * Sets or retrieves the vertical margin for the object.\n     */\n    /** @deprecated */\n    vspace: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: number;\n    readonly x: number;\n    readonly y: number;\n    decode(): Promise<void>;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLImageElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLImageElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLImageElement: {\n    prototype: HTMLImageElement;\n    new(): HTMLImageElement;\n};\n\n/** Provides special properties and methods for manipulating the options, layout, and presentation of <input> elements. */\ninterface HTMLInputElement extends HTMLElement {\n    /**\n     * Sets or retrieves a comma-separated list of content types.\n     */\n    accept: string;\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a text alternative to the graphic.\n     */\n    alt: string;\n    /**\n     * Specifies whether autocomplete is applied to an editable text field.\n     */\n    autocomplete: string;\n    /**\n     * Sets or retrieves the state of the check box or radio button.\n     */\n    checked: boolean;\n    /**\n     * Sets or retrieves the state of the check box or radio button.\n     */\n    defaultChecked: boolean;\n    /**\n     * Sets or retrieves the initial contents of the object.\n     */\n    defaultValue: string;\n    dirName: string;\n    disabled: boolean;\n    /**\n     * Returns a FileList object on a file type input object.\n     */\n    files: FileList | null;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Overrides the action attribute (where the data on a form is sent) on the parent form element.\n     */\n    formAction: string;\n    /**\n     * Used to override the encoding (formEnctype attribute) specified on the form element.\n     */\n    formEnctype: string;\n    /**\n     * Overrides the submit method attribute previously specified on a form element.\n     */\n    formMethod: string;\n    /**\n     * Overrides any validation or required attributes on a form or form elements to allow it to be submitted without validation. This can be used to create a "save draft"-type submit option.\n     */\n    formNoValidate: boolean;\n    /**\n     * Overrides the target attribute on a form element.\n     */\n    formTarget: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: number;\n    indeterminate: boolean;\n    readonly labels: NodeListOf<HTMLLabelElement> | null;\n    /**\n     * Specifies the ID of a pre-defined datalist of options for an input element.\n     */\n    readonly list: HTMLElement | null;\n    /**\n     * Defines the maximum acceptable value for an input element with type="number".When used with the min and step attributes, lets you control the range and increment (such as only even numbers) that the user can enter into an input field.\n     */\n    max: string;\n    /**\n     * Sets or retrieves the maximum number of characters that the user can enter in a text control.\n     */\n    maxLength: number;\n    /**\n     * Defines the minimum acceptable value for an input element with type="number". When used with the max and step attributes, lets you control the range and increment (such as even numbers only) that the user can enter into an input field.\n     */\n    min: string;\n    minLength: number;\n    /**\n     * Sets or retrieves the Boolean value indicating whether multiple items can be selected from a list.\n     */\n    multiple: boolean;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets or sets a string containing a regular expression that the user\'s input must match.\n     */\n    pattern: string;\n    /**\n     * Gets or sets a text string that is displayed in an input field as a hint or prompt to users as the format or type of information they need to enter.The text appears in an input field until the user puts focus on the field.\n     */\n    placeholder: string;\n    readOnly: boolean;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    selectionDirection: "forward" | "backward" | "none" | null;\n    /**\n     * Gets or sets the end position or offset of a text selection.\n     */\n    selectionEnd: number | null;\n    /**\n     * Gets or sets the starting position or offset of a text selection.\n     */\n    selectionStart: number | null;\n    size: number;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    /**\n     * Defines an increment or jump between values that you want to allow the user to enter. When used with the max and min attributes, lets you control the range and increment (for example, allow only even numbers) that the user can enter into an input field.\n     */\n    step: string;\n    /**\n     * Returns the content type of the object.\n     */\n    type: string;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    /** @deprecated */\n    useMap: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Returns the value of the data at the cursor\'s current position.\n     */\n    value: string;\n    /**\n     * Returns a Date object representing the form control\'s value, if applicable; otherwise, returns null. Can be set, to change the value. Throws an "InvalidStateError" DOMException if the control isn\'t date- or time-based.\n     */\n    valueAsDate: Date | null;\n    /**\n     * Returns the input field value as a number.\n     */\n    valueAsNumber: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: number;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Makes the selection equal to the current object.\n     */\n    select(): void;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    setRangeText(replacement: string): void;\n    setRangeText(replacement: string, start: number, end: number, selectionMode?: SelectionMode): void;\n    /**\n     * Sets the start and end positions of a selection in a text field.\n     * @param start The offset into the text field for the start of the selection.\n     * @param end The offset into the text field for the end of the selection.\n     * @param direction The direction in which the selection is performed.\n     */\n    setSelectionRange(start: number, end: number, direction?: "forward" | "backward" | "none"): void;\n    /**\n     * Decrements a range input control\'s value by the value given by the Step attribute. If the optional parameter is used, it will decrement the input control\'s step value multiplied by the parameter\'s value.\n     * @param n Value to decrement the value by.\n     */\n    stepDown(n?: number): void;\n    /**\n     * Increments a range input control\'s value by the value given by the Step attribute. If the optional parameter is used, will increment the input control\'s value by that value.\n     * @param n Value to increment the value by.\n     */\n    stepUp(n?: number): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLInputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLInputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLInputElement: {\n    prototype: HTMLInputElement;\n    new(): HTMLInputElement;\n};\n\n/** Exposes specific properties and methods (beyond those defined by regular HTMLElement interface it also has available to it by inheritance) for manipulating list elements. */\ninterface HTMLLIElement extends HTMLElement {\n    /** @deprecated */\n    type: string;\n    /**\n     * Sets or retrieves the value of a list item.\n     */\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLIElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLIElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLIElement: {\n    prototype: HTMLLIElement;\n    new(): HTMLLIElement;\n};\n\n/** Gives access to properties specific to <label> elements. It inherits methods and properties from the base HTMLElement interface. */\ninterface HTMLLabelElement extends HTMLElement {\n    readonly control: HTMLElement | null;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the object to which the given label object is assigned.\n     */\n    htmlFor: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLabelElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLabelElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLabelElement: {\n    prototype: HTMLLabelElement;\n    new(): HTMLLabelElement;\n};\n\n/** The HTMLLegendElement is an interface allowing to access properties of the <legend> elements. It inherits properties and methods from the HTMLElement interface. */\ninterface HTMLLegendElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLegendElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLegendElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLegendElement: {\n    prototype: HTMLLegendElement;\n    new(): HTMLLegendElement;\n};\n\n/** Reference information for external resources and the relationship of those resources to a document and vice-versa. This object inherits all of the properties and methods of the HTMLElement interface. */\ninterface HTMLLinkElement extends HTMLElement, LinkStyle {\n    as: string;\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    crossOrigin: string | null;\n    disabled: boolean;\n    /**\n     * Sets or retrieves a destination URL or an anchor point.\n     */\n    href: string;\n    /**\n     * Sets or retrieves the language code of the object.\n     */\n    hreflang: string;\n    imageSizes: string;\n    imageSrcset: string;\n    integrity: string;\n    /**\n     * Sets or retrieves the media type.\n     */\n    media: string;\n    referrerPolicy: string;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    rel: string;\n    readonly relList: DOMTokenList;\n    /**\n     * Sets or retrieves the relationship between the object and the destination of the link.\n     */\n    /** @deprecated */\n    rev: string;\n    readonly sizes: DOMTokenList;\n    /**\n     * Sets or retrieves the window or frame at which to target content.\n     */\n    /** @deprecated */\n    target: string;\n    /**\n     * Sets or retrieves the MIME type of the object.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLinkElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLLinkElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLLinkElement: {\n    prototype: HTMLLinkElement;\n    new(): HTMLLinkElement;\n};\n\n/** Provides special properties and methods (beyond those of the regular object HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of map elements. */\ninterface HTMLMapElement extends HTMLElement {\n    /**\n     * Retrieves a collection of the area objects defined for the given map object.\n     */\n    readonly areas: HTMLCollection;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMapElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMapElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMapElement: {\n    prototype: HTMLMapElement;\n    new(): HTMLMapElement;\n};\n\ninterface HTMLMarqueeElementEventMap extends HTMLElementEventMap {\n    "bounce": Event;\n    "finish": Event;\n    "start": Event;\n}\n\n/** Provides methods to manipulate <marquee> elements. */\ninterface HTMLMarqueeElement extends HTMLElement {\n    /** @deprecated */\n    behavior: string;\n    /** @deprecated */\n    bgColor: string;\n    /** @deprecated */\n    direction: string;\n    /** @deprecated */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /** @deprecated */\n    loop: number;\n    /** @deprecated */\n    onbounce: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    onfinish: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    onstart: ((this: HTMLMarqueeElement, ev: Event) => any) | null;\n    /** @deprecated */\n    scrollAmount: number;\n    /** @deprecated */\n    scrollDelay: number;\n    /** @deprecated */\n    trueSpeed: boolean;\n    /** @deprecated */\n    vspace: number;\n    /** @deprecated */\n    width: string;\n    /** @deprecated */\n    start(): void;\n    /** @deprecated */\n    stop(): void;\n    addEventListener<K extends keyof HTMLMarqueeElementEventMap>(type: K, listener: (this: HTMLMarqueeElement, ev: HTMLMarqueeElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMarqueeElementEventMap>(type: K, listener: (this: HTMLMarqueeElement, ev: HTMLMarqueeElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMarqueeElement: {\n    prototype: HTMLMarqueeElement;\n    new(): HTMLMarqueeElement;\n};\n\ninterface HTMLMediaElementEventMap extends HTMLElementEventMap {\n    "encrypted": MediaEncryptedEvent;\n    "waitingforkey": Event;\n}\n\n/** Adds to HTMLElement the properties and methods needed to support basic media-related capabilities that are common to audio and video. */\ninterface HTMLMediaElement extends HTMLElement {\n    /**\n     * Gets or sets a value that indicates whether to start playing the media automatically.\n     */\n    autoplay: boolean;\n    /**\n     * Gets a collection of buffered time ranges.\n     */\n    readonly buffered: TimeRanges;\n    /**\n     * Gets or sets a flag that indicates whether the client provides a set of controls for the media (in case the developer does not include controls for the player).\n     */\n    controls: boolean;\n    crossOrigin: string | null;\n    /**\n     * Gets the address or URL of the current media resource that is selected by IHTMLMediaElement.\n     */\n    readonly currentSrc: string;\n    /**\n     * Gets or sets the current playback position, in seconds.\n     */\n    currentTime: number;\n    defaultMuted: boolean;\n    /**\n     * Gets or sets the default playback rate when the user is not using fast forward or reverse for a video or audio resource.\n     */\n    defaultPlaybackRate: number;\n    /**\n     * Returns the duration in seconds of the current media resource. A NaN value is returned if duration is not available, or Infinity if the media resource is streaming.\n     */\n    readonly duration: number;\n    /**\n     * Gets information about whether the playback has ended or not.\n     */\n    readonly ended: boolean;\n    /**\n     * Returns an object representing the current error state of the audio or video element.\n     */\n    readonly error: MediaError | null;\n    /**\n     * Gets or sets a flag to specify whether playback should restart after it completes.\n     */\n    loop: boolean;\n    readonly mediaKeys: MediaKeys | null;\n    /**\n     * Gets or sets a flag that indicates whether the audio (either audio or the audio track on video media) is muted.\n     */\n    muted: boolean;\n    /**\n     * Gets the current network activity for the element.\n     */\n    readonly networkState: number;\n    onencrypted: ((this: HTMLMediaElement, ev: MediaEncryptedEvent) => any) | null;\n    onwaitingforkey: ((this: HTMLMediaElement, ev: Event) => any) | null;\n    /**\n     * Gets a flag that specifies whether playback is paused.\n     */\n    readonly paused: boolean;\n    /**\n     * Gets or sets the current rate of speed for the media resource to play. This speed is expressed as a multiple of the normal speed of the media resource.\n     */\n    playbackRate: number;\n    /**\n     * Gets TimeRanges for the current media resource that has been played.\n     */\n    readonly played: TimeRanges;\n    /**\n     * Gets or sets the current playback position, in seconds.\n     */\n    preload: string;\n    readonly readyState: number;\n    /**\n     * Returns a TimeRanges object that represents the ranges of the current media resource that can be seeked.\n     */\n    readonly seekable: TimeRanges;\n    /**\n     * Gets a flag that indicates whether the client is currently moving to a new playback position in the media resource.\n     */\n    readonly seeking: boolean;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcObject: MediaProvider | null;\n    readonly textTracks: TextTrackList;\n    /**\n     * Gets or sets the volume level for audio portions of the media element.\n     */\n    volume: number;\n    addTextTrack(kind: TextTrackKind, label?: string, language?: string): TextTrack;\n    /**\n     * Returns a string that specifies whether the client can play a given media resource type.\n     */\n    canPlayType(type: string): CanPlayTypeResult;\n    fastSeek(time: number): void;\n    /**\n     * Resets the audio or video object and loads a new media resource.\n     */\n    load(): void;\n    /**\n     * Pauses the current playback and sets paused to TRUE. This can be used to test whether the media is playing or paused. You can also use the pause or play events to tell whether the media is playing or not.\n     */\n    pause(): void;\n    /**\n     * Loads and starts playback of a media resource.\n     */\n    play(): Promise<void>;\n    setMediaKeys(mediaKeys: MediaKeys | null): Promise<void>;\n    readonly HAVE_CURRENT_DATA: number;\n    readonly HAVE_ENOUGH_DATA: number;\n    readonly HAVE_FUTURE_DATA: number;\n    readonly HAVE_METADATA: number;\n    readonly HAVE_NOTHING: number;\n    readonly NETWORK_EMPTY: number;\n    readonly NETWORK_IDLE: number;\n    readonly NETWORK_LOADING: number;\n    readonly NETWORK_NO_SOURCE: number;\n    addEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLMediaElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLMediaElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMediaElement: {\n    prototype: HTMLMediaElement;\n    new(): HTMLMediaElement;\n    readonly HAVE_CURRENT_DATA: number;\n    readonly HAVE_ENOUGH_DATA: number;\n    readonly HAVE_FUTURE_DATA: number;\n    readonly HAVE_METADATA: number;\n    readonly HAVE_NOTHING: number;\n    readonly NETWORK_EMPTY: number;\n    readonly NETWORK_IDLE: number;\n    readonly NETWORK_LOADING: number;\n    readonly NETWORK_NO_SOURCE: number;\n};\n\ninterface HTMLMenuElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMenuElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMenuElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMenuElement: {\n    prototype: HTMLMenuElement;\n    new(): HTMLMenuElement;\n};\n\n/** Contains descriptive metadata about a document. It inherits all of the properties and methods described in the HTMLElement interface. */\ninterface HTMLMetaElement extends HTMLElement {\n    /**\n     * Gets or sets meta-information to associate with httpEquiv or name.\n     */\n    content: string;\n    /**\n     * Gets or sets information used to bind the value of a content attribute of a meta element to an HTTP response header.\n     */\n    httpEquiv: string;\n    /**\n     * Sets or retrieves the value specified in the content attribute of the meta object.\n     */\n    name: string;\n    /**\n     * Sets or retrieves a scheme to be used in interpreting the value of a property specified for the object.\n     */\n    /** @deprecated */\n    scheme: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMetaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMetaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMetaElement: {\n    prototype: HTMLMetaElement;\n    new(): HTMLMetaElement;\n};\n\n/** The HTML <meter> elements expose the HTMLMeterElement interface, which provides special properties and methods (beyond the HTMLElement object interface they also have available to them by inheritance) for manipulating the layout and presentation of <meter> elements. */\ninterface HTMLMeterElement extends HTMLElement {\n    high: number;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    low: number;\n    max: number;\n    min: number;\n    optimum: number;\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMeterElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLMeterElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLMeterElement: {\n    prototype: HTMLMeterElement;\n    new(): HTMLMeterElement;\n};\n\n/** Provides special properties (beyond the regular methods and properties available through the HTMLElement interface they also have available to them by inheritance) for manipulating modification elements, that is <del> and <ins>. */\ninterface HTMLModElement extends HTMLElement {\n    /**\n     * Sets or retrieves reference information about the object.\n     */\n    cite: string;\n    /**\n     * Sets or retrieves the date and time of a modification to the object.\n     */\n    dateTime: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLModElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLModElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLModElement: {\n    prototype: HTMLModElement;\n    new(): HTMLModElement;\n};\n\n/** Provides special properties (beyond those defined on the regular HTMLElement interface it also has available to it by inheritance) for manipulating ordered list elements. */\ninterface HTMLOListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    reversed: boolean;\n    /**\n     * The starting number.\n     */\n    start: number;\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOListElement: {\n    prototype: HTMLOListElement;\n    new(): HTMLOListElement;\n};\n\n/** Provides special properties and methods (beyond those on the HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of <object> element, representing external resources. */\ninterface HTMLObjectElement extends HTMLElement {\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a character string that can be used to implement your own archive functionality for the object.\n     */\n    /** @deprecated */\n    archive: string;\n    /** @deprecated */\n    border: string;\n    /**\n     * Sets or retrieves the URL of the file containing the compiled Java class.\n     */\n    /** @deprecated */\n    code: string;\n    /**\n     * Sets or retrieves the URL of the component.\n     */\n    /** @deprecated */\n    codeBase: string;\n    /**\n     * Sets or retrieves the Internet media type for the code associated with the object.\n     */\n    /** @deprecated */\n    codeType: string;\n    /**\n     * Retrieves the document object of the page or frame.\n     */\n    readonly contentDocument: Document | null;\n    readonly contentWindow: WindowProxy | null;\n    /**\n     * Sets or retrieves the URL that references the data of the object.\n     */\n    data: string;\n    /** @deprecated */\n    declare: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    height: string;\n    /** @deprecated */\n    hspace: number;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Sets or retrieves a message to be displayed while an object is loading.\n     */\n    /** @deprecated */\n    standby: string;\n    /**\n     * Sets or retrieves the MIME type of the object.\n     */\n    type: string;\n    /**\n     * Sets or retrieves the URL, often with a bookmark extension (#name), to use as a client-side image map.\n     */\n    useMap: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /** @deprecated */\n    vspace: number;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    width: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    getSVGDocument(): Document | null;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLObjectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLObjectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLObjectElement: {\n    prototype: HTMLObjectElement;\n    new(): HTMLObjectElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement object interface they also have available to them by inheritance) for manipulating the layout and presentation of <optgroup> elements. */\ninterface HTMLOptGroupElement extends HTMLElement {\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves a value that you can use to implement your own label functionality for the object.\n     */\n    label: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptGroupElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptGroupElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOptGroupElement: {\n    prototype: HTMLOptGroupElement;\n    new(): HTMLOptGroupElement;\n};\n\n/** <option> elements and inherits all classes and methods of the HTMLElement interface. */\ninterface HTMLOptionElement extends HTMLElement {\n    /**\n     * Sets or retrieves the status of an option.\n     */\n    defaultSelected: boolean;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    /**\n     * Sets or retrieves the ordinal position of an option in a list box.\n     */\n    readonly index: number;\n    /**\n     * Sets or retrieves a value that you can use to implement your own label functionality for the object.\n     */\n    label: string;\n    /**\n     * Sets or retrieves whether the option in the list box is the default item.\n     */\n    selected: boolean;\n    /**\n     * Sets or retrieves the text string specified by the option tag.\n     */\n    text: string;\n    /**\n     * Sets or retrieves the value which is returned to the server when the form control is submitted.\n     */\n    value: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOptionElement: {\n    prototype: HTMLOptionElement;\n    new(): HTMLOptionElement;\n};\n\n/** HTMLOptionsCollection is an interface representing a collection of HTML option elements (in document order) and offers methods and properties for traversing the list as well as optionally altering its items. This type is returned solely by the "options" property of select. */\ninterface HTMLOptionsCollection extends HTMLCollectionOf<HTMLOptionElement> {\n    /**\n     * Returns the number of elements in the collection.\n     * \n     * When set to a smaller number, truncates the number of option elements in the corresponding container.\n     * \n     * When set to a greater number, adds new blank option elements to that container.\n     */\n    length: number;\n    /**\n     * Returns the index of the first selected item, if any, or −1 if there is no selected item.\n     * \n     * Can be set, to change the selection.\n     */\n    selectedIndex: number;\n    /**\n     * Inserts element before the node given by before.\n     * \n     * The before argument can be a number, in which case element is inserted before the item with that number, or an element from the collection, in which case element is inserted before that element.\n     * \n     * If before is omitted, null, or a number out of range, then element will be added at the end of the list.\n     * \n     * This method will throw a "HierarchyRequestError" DOMException if element is an ancestor of the element into which it is to be inserted.\n     */\n    add(element: HTMLOptionElement | HTMLOptGroupElement, before?: HTMLElement | number | null): void;\n    /**\n     * Removes the item with index index from the collection.\n     */\n    remove(index: number): void;\n}\n\ndeclare var HTMLOptionsCollection: {\n    prototype: HTMLOptionsCollection;\n    new(): HTMLOptionsCollection;\n};\n\ninterface HTMLOrSVGElement {\n    autofocus: boolean;\n    readonly dataset: DOMStringMap;\n    nonce?: string;\n    tabIndex: number;\n    blur(): void;\n    focus(options?: FocusOptions): void;\n}\n\n/** Provides properties and methods (beyond those inherited from HTMLElement) for manipulating the layout and presentation of <output> elements. */\ninterface HTMLOutputElement extends HTMLElement {\n    defaultValue: string;\n    readonly form: HTMLFormElement | null;\n    readonly htmlFor: DOMTokenList;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    name: string;\n    readonly type: string;\n    readonly validationMessage: string;\n    readonly validity: ValidityState;\n    value: string;\n    readonly willValidate: boolean;\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOutputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLOutputElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLOutputElement: {\n    prototype: HTMLOutputElement;\n    new(): HTMLOutputElement;\n};\n\n/** Provides special properties (beyond those of the regular HTMLElement object interface it inherits) for manipulating <p> elements. */\ninterface HTMLParagraphElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParagraphElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParagraphElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLParagraphElement: {\n    prototype: HTMLParagraphElement;\n    new(): HTMLParagraphElement;\n};\n\n/** Provides special properties (beyond those of the regular HTMLElement object interface it inherits) for manipulating <param> elements, representing a pair of a key and a value that acts as a parameter for an <object> element. */\ninterface HTMLParamElement extends HTMLElement {\n    /**\n     * Sets or retrieves the name of an input parameter for an element.\n     */\n    name: string;\n    /**\n     * Sets or retrieves the content type of the resource designated by the value attribute.\n     */\n    /** @deprecated */\n    type: string;\n    /**\n     * Sets or retrieves the value of an input parameter for an element.\n     */\n    value: string;\n    /**\n     * Sets or retrieves the data type of the value attribute.\n     */\n    /** @deprecated */\n    valueType: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParamElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLParamElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLParamElement: {\n    prototype: HTMLParamElement;\n    new(): HTMLParamElement;\n};\n\n/** A <picture> HTML element. It doesn\'t implement specific properties or methods. */\ninterface HTMLPictureElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPictureElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPictureElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLPictureElement: {\n    prototype: HTMLPictureElement;\n    new(): HTMLPictureElement;\n};\n\n/** Exposes specific properties and methods (beyond those of the HTMLElement interface it also has available to it by inheritance) for manipulating a block of preformatted text (<pre>). */\ninterface HTMLPreElement extends HTMLElement {\n    /**\n     * Sets or gets a value that you can use to implement your own width functionality for the object.\n     */\n    /** @deprecated */\n    width: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPreElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLPreElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLPreElement: {\n    prototype: HTMLPreElement;\n    new(): HTMLPreElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of <progress> elements. */\ninterface HTMLProgressElement extends HTMLElement {\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Defines the maximum, or "done" value for a progress element.\n     */\n    max: number;\n    /**\n     * Returns the quotient of value/max when the value attribute is set (determinate progress bar), or -1 when the value attribute is missing (indeterminate progress bar).\n     */\n    readonly position: number;\n    /**\n     * Sets or gets the current value of a progress element. The value must be a non-negative number between 0 and the max value.\n     */\n    value: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLProgressElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLProgressElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLProgressElement: {\n    prototype: HTMLProgressElement;\n    new(): HTMLProgressElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating quoting elements, like <blockquote> and <q>, but not the <cite> element. */\ninterface HTMLQuoteElement extends HTMLElement {\n    /**\n     * Sets or retrieves reference information about the object.\n     */\n    cite: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLQuoteElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLQuoteElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLQuoteElement: {\n    prototype: HTMLQuoteElement;\n    new(): HTMLQuoteElement;\n};\n\n/** HTML <script> elements expose the HTMLScriptElement interface, which provides special properties and methods for manipulating the behavior and execution of <script> elements (beyond the inherited HTMLElement interface). */\ninterface HTMLScriptElement extends HTMLElement {\n    async: boolean;\n    /**\n     * Sets or retrieves the character set used to encode the object.\n     */\n    /** @deprecated */\n    charset: string;\n    crossOrigin: string | null;\n    /**\n     * Sets or retrieves the status of the script.\n     */\n    defer: boolean;\n    /**\n     * Sets or retrieves the event for which the script is written.\n     */\n    /** @deprecated */\n    event: string;\n    /**\n     * Sets or retrieves the object that is bound to the event script.\n     */\n    /** @deprecated */\n    htmlFor: string;\n    integrity: string;\n    noModule: boolean;\n    referrerPolicy: string;\n    /**\n     * Retrieves the URL to an external file that contains the source code or data.\n     */\n    src: string;\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    /**\n     * Sets or retrieves the MIME type for the associated scripting engine.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLScriptElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLScriptElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLScriptElement: {\n    prototype: HTMLScriptElement;\n    new(): HTMLScriptElement;\n};\n\n/** A <select> HTML Element. These elements also share all of the properties and methods of other HTML elements via the HTMLElement interface. */\ninterface HTMLSelectElement extends HTMLElement {\n    autocomplete: string;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the number of objects in a collection.\n     */\n    length: number;\n    /**\n     * Sets or retrieves the Boolean value indicating whether multiple items can be selected from a list.\n     */\n    multiple: boolean;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    readonly options: HTMLOptionsCollection;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    /**\n     * Sets or retrieves the index of the selected option in a select object.\n     */\n    selectedIndex: number;\n    readonly selectedOptions: HTMLCollectionOf<HTMLOptionElement>;\n    /**\n     * Sets or retrieves the number of rows in the list box.\n     */\n    size: number;\n    /**\n     * Retrieves the type of select control based on the value of the MULTIPLE attribute.\n     */\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Sets or retrieves the value which is returned to the server when the form control is submitted.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Adds an element to the areas, controlRange, or options collection.\n     * @param element Variant of type Number that specifies the index position in the collection where the element is placed. If no value is given, the method places the element at the end of the collection.\n     * @param before Variant of type Object that specifies an element to insert before, or null to append the object to the collection.\n     */\n    add(element: HTMLOptionElement | HTMLOptGroupElement, before?: HTMLElement | number | null): void;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    /**\n     * Retrieves a select object or an object from an options collection.\n     * @param name Variant of type Number or String that specifies the object or collection to retrieve. If this parameter is an integer, it is the zero-based index of the object. If this parameter is a string, all objects with matching name or id properties are retrieved, and a collection is returned if more than one match is made.\n     * @param index Variant of type Number that specifies the zero-based index of the object to retrieve when a collection is returned.\n     */\n    item(index: number): Element | null;\n    /**\n     * Retrieves a select object or an object from an options collection.\n     * @param namedItem A String that specifies the name or id property of the object to retrieve. A collection is returned if more than one match is made.\n     */\n    namedItem(name: string): HTMLOptionElement | null;\n    /**\n     * Removes an element from the collection.\n     * @param index Number that specifies the zero-based index of the element to remove from the collection.\n     */\n    remove(): void;\n    remove(index: number): void;\n    reportValidity(): boolean;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSelectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSelectElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [name: number]: HTMLOptionElement | HTMLOptGroupElement;\n}\n\ndeclare var HTMLSelectElement: {\n    prototype: HTMLSelectElement;\n    new(): HTMLSelectElement;\n};\n\ninterface HTMLSlotElement extends HTMLElement {\n    name: string;\n    assignedElements(options?: AssignedNodesOptions): Element[];\n    assignedNodes(options?: AssignedNodesOptions): Node[];\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSlotElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSlotElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSlotElement: {\n    prototype: HTMLSlotElement;\n    new(): HTMLSlotElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement object interface it also has available to it by inheritance) for manipulating <source> elements. */\ninterface HTMLSourceElement extends HTMLElement {\n    /**\n     * Gets or sets the intended media type of the media source.\n     */\n    media: string;\n    sizes: string;\n    /**\n     * The address or URL of the a media resource that is to be considered.\n     */\n    src: string;\n    srcset: string;\n    /**\n     * Gets or sets the MIME type of a media resource.\n     */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSourceElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSourceElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSourceElement: {\n    prototype: HTMLSourceElement;\n    new(): HTMLSourceElement;\n};\n\n/** A <span> element and derives from the HTMLElement interface, but without implementing any additional properties or methods. */\ninterface HTMLSpanElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSpanElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLSpanElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLSpanElement: {\n    prototype: HTMLSpanElement;\n    new(): HTMLSpanElement;\n};\n\n/** A <style> element. It inherits properties and methods from its parent, HTMLElement, and from LinkStyle. */\ninterface HTMLStyleElement extends HTMLElement, LinkStyle {\n    /**\n     * Sets or retrieves the media type.\n     */\n    media: string;\n    /**\n     * Retrieves the CSS language in which the style sheet is written.\n     */\n    /** @deprecated */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLStyleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLStyleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLStyleElement: {\n    prototype: HTMLStyleElement;\n    new(): HTMLStyleElement;\n};\n\n/** Special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating table caption elements. */\ninterface HTMLTableCaptionElement extends HTMLElement {\n    /**\n     * Sets or retrieves the alignment of the caption or legend.\n     */\n    /** @deprecated */\n    align: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCaptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCaptionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableCaptionElement: {\n    prototype: HTMLTableCaptionElement;\n    new(): HTMLTableCaptionElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of table cells, either header or data cells, in an HTML document. */\ninterface HTMLTableCellElement extends HTMLElement {\n    /**\n     * Sets or retrieves abbreviated text for the object.\n     */\n    abbr: string;\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /**\n     * Sets or retrieves a comma-delimited list of conceptual categories associated with the object.\n     */\n    /** @deprecated */\n    axis: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Retrieves the position of the object in the cells collection of a row.\n     */\n    readonly cellIndex: number;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number columns in the table that the object should span.\n     */\n    colSpan: number;\n    /**\n     * Sets or retrieves a list of header cells that provide information for the object.\n     */\n    headers: string;\n    /**\n     * Sets or retrieves the height of the object.\n     */\n    /** @deprecated */\n    height: string;\n    /**\n     * Sets or retrieves whether the browser automatically performs wordwrap.\n     */\n    /** @deprecated */\n    noWrap: boolean;\n    /**\n     * Sets or retrieves how many rows in a table the cell should span.\n     */\n    rowSpan: number;\n    /**\n     * Sets or retrieves the group of cells in a table to which the object\'s information applies.\n     */\n    scope: string;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableCellElement: {\n    prototype: HTMLTableCellElement;\n    new(): HTMLTableCellElement;\n};\n\n/** Provides special properties (beyond the HTMLElement interface it also has available to it inheritance) for manipulating single or grouped table column elements. */\ninterface HTMLTableColElement extends HTMLElement {\n    /**\n     * Sets or retrieves the alignment of the object relative to the display or table.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number of columns in the group.\n     */\n    span: number;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableColElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableColElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableColElement: {\n    prototype: HTMLTableColElement;\n    new(): HTMLTableColElement;\n};\n\ninterface HTMLTableDataCellElement extends HTMLTableCellElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableDataCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableDataCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableDataCellElement: {\n    prototype: HTMLTableDataCellElement;\n    new(): HTMLTableDataCellElement;\n};\n\n/** Provides special properties and methods (beyond the regular HTMLElement object interface it also has available to it by inheritance) for manipulating the layout and presentation of tables in an HTML document. */\ninterface HTMLTableElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Sets or retrieves the width of the border to draw around the object.\n     */\n    /** @deprecated */\n    border: string;\n    /**\n     * Retrieves the caption object of a table.\n     */\n    caption: HTMLTableCaptionElement | null;\n    /**\n     * Sets or retrieves the amount of space between the border of the cell and the content of the cell.\n     */\n    /** @deprecated */\n    cellPadding: string;\n    /**\n     * Sets or retrieves the amount of space between cells in a table.\n     */\n    /** @deprecated */\n    cellSpacing: string;\n    /**\n     * Sets or retrieves the way the border frame around the table is displayed.\n     */\n    /** @deprecated */\n    frame: string;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    readonly rows: HTMLCollectionOf<HTMLTableRowElement>;\n    /**\n     * Sets or retrieves which dividing lines (inner borders) are displayed.\n     */\n    /** @deprecated */\n    rules: string;\n    /**\n     * Sets or retrieves a description and/or structure of the object.\n     */\n    /** @deprecated */\n    summary: string;\n    /**\n     * Retrieves a collection of all tBody objects in the table. Objects in this collection are in source order.\n     */\n    readonly tBodies: HTMLCollectionOf<HTMLTableSectionElement>;\n    /**\n     * Retrieves the tFoot object of the table.\n     */\n    tFoot: HTMLTableSectionElement | null;\n    /**\n     * Retrieves the tHead object of the table.\n     */\n    tHead: HTMLTableSectionElement | null;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    /** @deprecated */\n    width: string;\n    /**\n     * Creates an empty caption element in the table.\n     */\n    createCaption(): HTMLTableCaptionElement;\n    /**\n     * Creates an empty tBody element in the table.\n     */\n    createTBody(): HTMLTableSectionElement;\n    /**\n     * Creates an empty tFoot element in the table.\n     */\n    createTFoot(): HTMLTableSectionElement;\n    /**\n     * Returns the tHead element object if successful, or null otherwise.\n     */\n    createTHead(): HTMLTableSectionElement;\n    /**\n     * Deletes the caption element and its contents from the table.\n     */\n    deleteCaption(): void;\n    /**\n     * Removes the specified row (tr) from the element and from the rows collection.\n     * @param index Number that specifies the zero-based position in the rows collection of the row to remove.\n     */\n    deleteRow(index: number): void;\n    /**\n     * Deletes the tFoot element and its contents from the table.\n     */\n    deleteTFoot(): void;\n    /**\n     * Deletes the tHead element and its contents from the table.\n     */\n    deleteTHead(): void;\n    /**\n     * Creates a new row (tr) in the table, and adds the row to the rows collection.\n     * @param index Number that specifies where to insert the row in the rows collection. The default value is -1, which appends the new row to the end of the rows collection.\n     */\n    insertRow(index?: number): HTMLTableRowElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableElement: {\n    prototype: HTMLTableElement;\n    new(): HTMLTableElement;\n};\n\ninterface HTMLTableHeaderCellElement extends HTMLTableCellElement {\n    scope: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableHeaderCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableHeaderCellElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableHeaderCellElement: {\n    prototype: HTMLTableHeaderCellElement;\n    new(): HTMLTableHeaderCellElement;\n};\n\n/** Provides special properties and methods (beyond the HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of rows in an HTML table. */\ninterface HTMLTableRowElement extends HTMLElement {\n    /**\n     * Sets or retrieves how the object is aligned with adjacent text.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    bgColor: string;\n    /**\n     * Retrieves a collection of all cells in the table row.\n     */\n    readonly cells: HTMLCollectionOf<HTMLTableDataCellElement | HTMLTableHeaderCellElement>;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Retrieves the position of the object in the rows collection for the table.\n     */\n    readonly rowIndex: number;\n    /**\n     * Retrieves the position of the object in the collection.\n     */\n    readonly sectionRowIndex: number;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Removes the specified cell from the table row, as well as from the cells collection.\n     * @param index Number that specifies the zero-based position of the cell to remove from the table row. If no value is provided, the last cell in the cells collection is deleted.\n     */\n    deleteCell(index: number): void;\n    /**\n     * Creates a new cell in the table row, and adds the cell to the cells collection.\n     * @param index Number that specifies where to insert the cell in the tr. The default value is -1, which appends the new cell to the end of the cells collection.\n     */\n    insertCell(index?: number): HTMLTableDataCellElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableRowElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableRowElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableRowElement: {\n    prototype: HTMLTableRowElement;\n    new(): HTMLTableRowElement;\n};\n\n/** Provides special properties and methods (beyond the HTMLElement interface it also has available to it by inheritance) for manipulating the layout and presentation of sections, that is headers, footers and bodies, in an HTML table. */\ninterface HTMLTableSectionElement extends HTMLElement {\n    /**\n     * Sets or retrieves a value that indicates the table alignment.\n     */\n    /** @deprecated */\n    align: string;\n    /** @deprecated */\n    ch: string;\n    /** @deprecated */\n    chOff: string;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    readonly rows: HTMLCollectionOf<HTMLTableRowElement>;\n    /** @deprecated */\n    vAlign: string;\n    /**\n     * Removes the specified row (tr) from the element and from the rows collection.\n     * @param index Number that specifies the zero-based position in the rows collection of the row to remove.\n     */\n    deleteRow(index: number): void;\n    /**\n     * Creates a new row (tr) in the table, and adds the row to the rows collection.\n     * @param index Number that specifies where to insert the row in the rows collection. The default value is -1, which appends the new row to the end of the rows collection.\n     */\n    insertRow(index?: number): HTMLTableRowElement;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableSectionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTableSectionElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTableSectionElement: {\n    prototype: HTMLTableSectionElement;\n    new(): HTMLTableSectionElement;\n};\n\n/** Enables access to the contents of an HTML <template> element. */\ninterface HTMLTemplateElement extends HTMLElement {\n    readonly content: DocumentFragment;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTemplateElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTemplateElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTemplateElement: {\n    prototype: HTMLTemplateElement;\n    new(): HTMLTemplateElement;\n};\n\n/** Provides special properties and methods for manipulating the layout and presentation of <textarea> elements. */\ninterface HTMLTextAreaElement extends HTMLElement {\n    autocomplete: string;\n    /**\n     * Sets or retrieves the width of the object.\n     */\n    cols: number;\n    /**\n     * Sets or retrieves the initial contents of the object.\n     */\n    defaultValue: string;\n    dirName: string;\n    disabled: boolean;\n    /**\n     * Retrieves a reference to the form that the object is embedded in.\n     */\n    readonly form: HTMLFormElement | null;\n    readonly labels: NodeListOf<HTMLLabelElement>;\n    /**\n     * Sets or retrieves the maximum number of characters that the user can enter in a text control.\n     */\n    maxLength: number;\n    minLength: number;\n    /**\n     * Sets or retrieves the name of the object.\n     */\n    name: string;\n    /**\n     * Gets or sets a text string that is displayed in an input field as a hint or prompt to users as the format or type of information they need to enter.The text appears in an input field until the user puts focus on the field.\n     */\n    placeholder: string;\n    /**\n     * Sets or retrieves the value indicated whether the content of the object is read-only.\n     */\n    readOnly: boolean;\n    /**\n     * When present, marks an element that can\'t be submitted without a value.\n     */\n    required: boolean;\n    /**\n     * Sets or retrieves the number of horizontal rows contained in the object.\n     */\n    rows: number;\n    selectionDirection: "forward" | "backward" | "none";\n    /**\n     * Gets or sets the end position or offset of a text selection.\n     */\n    selectionEnd: number;\n    /**\n     * Gets or sets the starting position or offset of a text selection.\n     */\n    selectionStart: number;\n    readonly textLength: number;\n    /**\n     * Retrieves the type of control.\n     */\n    readonly type: string;\n    /**\n     * Returns the error message that would be displayed if the user submits the form, or an empty string if no error message. It also triggers the standard error message, such as "this is a required field". The result is that the user sees validation messages without actually submitting.\n     */\n    readonly validationMessage: string;\n    /**\n     * Returns a  ValidityState object that represents the validity states of an element.\n     */\n    readonly validity: ValidityState;\n    /**\n     * Retrieves or sets the text in the entry field of the textArea element.\n     */\n    value: string;\n    /**\n     * Returns whether an element will successfully validate based on forms validation rules and constraints.\n     */\n    readonly willValidate: boolean;\n    /**\n     * Sets or retrieves how to handle wordwrapping in the object.\n     */\n    wrap: string;\n    /**\n     * Returns whether a form will validate when it is submitted, without having to submit it.\n     */\n    checkValidity(): boolean;\n    reportValidity(): boolean;\n    /**\n     * Highlights the input area of a form element.\n     */\n    select(): void;\n    /**\n     * Sets a custom error message that is displayed when a form is submitted.\n     * @param error Sets a custom error message that is displayed when a form is submitted.\n     */\n    setCustomValidity(error: string): void;\n    setRangeText(replacement: string): void;\n    setRangeText(replacement: string, start: number, end: number, selectionMode?: SelectionMode): void;\n    /**\n     * Sets the start and end positions of a selection in a text field.\n     * @param start The offset into the text field for the start of the selection.\n     * @param end The offset into the text field for the end of the selection.\n     * @param direction The direction in which the selection is performed.\n     */\n    setSelectionRange(start: number, end: number, direction?: "forward" | "backward" | "none"): void;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTextAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTextAreaElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTextAreaElement: {\n    prototype: HTMLTextAreaElement;\n    new(): HTMLTextAreaElement;\n};\n\n/** Provides special properties (beyond the regular HTMLElement interface it also has available to it by inheritance) for manipulating <time> elements. */\ninterface HTMLTimeElement extends HTMLElement {\n    dateTime: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTimeElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTimeElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTimeElement: {\n    prototype: HTMLTimeElement;\n    new(): HTMLTimeElement;\n};\n\n/** Contains the title for a document. This element inherits all of the properties and methods of the HTMLElement interface. */\ninterface HTMLTitleElement extends HTMLElement {\n    /**\n     * Retrieves or sets the text of the object as a string.\n     */\n    text: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTitleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTitleElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTitleElement: {\n    prototype: HTMLTitleElement;\n    new(): HTMLTitleElement;\n};\n\n/** The HTMLTrackElement */\ninterface HTMLTrackElement extends HTMLElement {\n    default: boolean;\n    kind: string;\n    label: string;\n    readonly readyState: number;\n    src: string;\n    srclang: string;\n    readonly track: TextTrack;\n    readonly ERROR: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTrackElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLTrackElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLTrackElement: {\n    prototype: HTMLTrackElement;\n    new(): HTMLTrackElement;\n    readonly ERROR: number;\n    readonly LOADED: number;\n    readonly LOADING: number;\n    readonly NONE: number;\n};\n\n/** Provides special properties (beyond those defined on the regular HTMLElement interface it also has available to it by inheritance) for manipulating unordered list elements. */\ninterface HTMLUListElement extends HTMLElement {\n    /** @deprecated */\n    compact: boolean;\n    /** @deprecated */\n    type: string;\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUListElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLUListElement: {\n    prototype: HTMLUListElement;\n    new(): HTMLUListElement;\n};\n\n/** An invalid HTML element and derives from the HTMLElement interface, but without implementing any additional properties or methods. */\ninterface HTMLUnknownElement extends HTMLElement {\n    addEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUnknownElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLElementEventMap>(type: K, listener: (this: HTMLUnknownElement, ev: HTMLElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLUnknownElement: {\n    prototype: HTMLUnknownElement;\n    new(): HTMLUnknownElement;\n};\n\n/** Provides special properties and methods for manipulating video objects. It also inherits properties and methods of HTMLMediaElement and HTMLElement. */\ninterface HTMLVideoElement extends HTMLMediaElement {\n    /**\n     * Gets or sets the height of the video element.\n     */\n    height: number;\n    /**\n     * Gets or sets the playsinline of the video element. for example, On iPhone, video elements will now be allowed to play inline, and will not automatically enter fullscreen mode when playback begins.\n     */\n    playsInline: boolean;\n    /**\n     * Gets or sets a URL of an image to display, for example, like a movie poster. This can be a still frame from the video, or another image if no video data is available.\n     */\n    poster: string;\n    /**\n     * Gets the intrinsic height of a video in CSS pixels, or zero if the dimensions are not known.\n     */\n    readonly videoHeight: number;\n    /**\n     * Gets the intrinsic width of a video in CSS pixels, or zero if the dimensions are not known.\n     */\n    readonly videoWidth: number;\n    /**\n     * Gets or sets the width of the video element.\n     */\n    width: number;\n    getVideoPlaybackQuality(): VideoPlaybackQuality;\n    addEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLVideoElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof HTMLMediaElementEventMap>(type: K, listener: (this: HTMLVideoElement, ev: HTMLMediaElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var HTMLVideoElement: {\n    prototype: HTMLVideoElement;\n    new(): HTMLVideoElement;\n};\n\n/** Events that fire when the fragment identifier of the URL has changed. */\ninterface HashChangeEvent extends Event {\n    /**\n     * Returns the URL of the session history entry that is now current.\n     */\n    readonly newURL: string;\n    /**\n     * Returns the URL of the session history entry that was previously current.\n     */\n    readonly oldURL: string;\n}\n\ndeclare var HashChangeEvent: {\n    prototype: HashChangeEvent;\n    new(type: string, eventInitDict?: HashChangeEventInit): HashChangeEvent;\n};\n\n/** This Fetch API interface allows you to perform various actions on HTTP request and response headers. These actions include retrieving, setting, adding to, and removing. A Headers object has an associated header list, which is initially empty and consists of zero or more name and value pairs.  You can add to this using methods like append() (see Examples.) In all methods of this interface, header names are matched by case-insensitive byte sequence. */\ninterface Headers {\n    append(name: string, value: string): void;\n    delete(name: string): void;\n    get(name: string): string | null;\n    has(name: string): boolean;\n    set(name: string, value: string): void;\n    forEach(callbackfn: (value: string, key: string, parent: Headers) => void, thisArg?: any): void;\n}\n\ndeclare var Headers: {\n    prototype: Headers;\n    new(init?: HeadersInit): Headers;\n};\n\n/** Allows manipulation of the browser session history, that is the pages visited in the tab or frame that the current page is loaded in. */\ninterface History {\n    readonly length: number;\n    scrollRestoration: ScrollRestoration;\n    readonly state: any;\n    back(): void;\n    forward(): void;\n    go(delta?: number): void;\n    pushState(data: any, title: string, url?: string | null): void;\n    replaceState(data: any, title: string, url?: string | null): void;\n}\n\ndeclare var History: {\n    prototype: History;\n    new(): History;\n};\n\ninterface IDBArrayKey extends Array<IDBValidKey> {\n}\n\n/** This IndexedDB API interface represents a cursor for traversing or iterating over multiple records in a database. */\ninterface IDBCursor {\n    /**\n     * Returns the direction ("next", "nextunique", "prev" or "prevunique") of the cursor.\n     */\n    readonly direction: IDBCursorDirection;\n    /**\n     * Returns the key of the cursor. Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly key: IDBValidKey;\n    /**\n     * Returns the effective key of the cursor. Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly primaryKey: IDBValidKey;\n    /**\n     * Returns the IDBObjectStore or IDBIndex the cursor was opened from.\n     */\n    readonly source: IDBObjectStore | IDBIndex;\n    /**\n     * Advances the cursor through the next count records in range.\n     */\n    advance(count: number): void;\n    /**\n     * Advances the cursor to the next record in range.\n     */\n    continue(key?: IDBValidKey): void;\n    /**\n     * Advances the cursor to the next record in range matching or after key and primaryKey. Throws an "InvalidAccessError" DOMException if the source is not an index.\n     */\n    continuePrimaryKey(key: IDBValidKey, primaryKey: IDBValidKey): void;\n    /**\n     * Delete the record pointed at by the cursor with a new value.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    delete(): IDBRequest<undefined>;\n    /**\n     * Updated the record pointed at by the cursor with a new value.\n     * \n     * Throws a "DataError" DOMException if the effective object store uses in-line keys and the key would have changed.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    update(value: any): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBCursor: {\n    prototype: IDBCursor;\n    new(): IDBCursor;\n};\n\n/** This IndexedDB API interface represents a cursor for traversing or iterating over multiple records in a database. It is the same as the IDBCursor, except that it includes the value property. */\ninterface IDBCursorWithValue extends IDBCursor {\n    /**\n     * Returns the cursor\'s current value.\n     */\n    readonly value: any;\n}\n\ndeclare var IDBCursorWithValue: {\n    prototype: IDBCursorWithValue;\n    new(): IDBCursorWithValue;\n};\n\ninterface IDBDatabaseEventMap {\n    "abort": Event;\n    "close": Event;\n    "error": Event;\n    "versionchange": IDBVersionChangeEvent;\n}\n\n/** This IndexedDB API interface provides a connection to a database; you can use an IDBDatabase object to open a transaction on your database then create, manipulate, and delete objects (data) in that database. The interface provides the only way to get and manage versions of the database. */\ninterface IDBDatabase extends EventTarget {\n    /**\n     * Returns the name of the database.\n     */\n    readonly name: string;\n    /**\n     * Returns a list of the names of object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBDatabase, ev: Event) => any) | null;\n    onclose: ((this: IDBDatabase, ev: Event) => any) | null;\n    onerror: ((this: IDBDatabase, ev: Event) => any) | null;\n    onversionchange: ((this: IDBDatabase, ev: IDBVersionChangeEvent) => any) | null;\n    /**\n     * Returns the version of the database.\n     */\n    readonly version: number;\n    /**\n     * Closes the connection once all running transactions have finished.\n     */\n    close(): void;\n    /**\n     * Creates a new object store with the given name and options and returns a new IDBObjectStore.\n     * \n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createObjectStore(name: string, options?: IDBObjectStoreParameters): IDBObjectStore;\n    /**\n     * Deletes the object store with the given name.\n     * \n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    deleteObjectStore(name: string): void;\n    /**\n     * Returns a new transaction with the given mode ("readonly" or "readwrite") and scope which can be a single object store name or an array of names.\n     */\n    transaction(storeNames: string | string[], mode?: IDBTransactionMode): IDBTransaction;\n    addEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBDatabase: {\n    prototype: IDBDatabase;\n    new(): IDBDatabase;\n};\n\n/** In the following code snippet, we make a request to open a database, and include handlers for the success and error cases. For a full working example, see our To-do Notifications app (view example live.) */\ninterface IDBFactory {\n    /**\n     * Compares two values as keys. Returns -1 if key1 precedes key2, 1 if key2 precedes key1, and 0 if the keys are equal.\n     * \n     * Throws a "DataError" DOMException if either input is not a valid key.\n     */\n    cmp(first: any, second: any): number;\n    /**\n     * Attempts to delete the named database. If the database already exists and there are open connections that don\'t close in response to a versionchange event, the request will be blocked until all they close. If the request is successful request\'s result will be null.\n     */\n    deleteDatabase(name: string): IDBOpenDBRequest;\n    /**\n     * Attempts to open a connection to the named database with the current version, or 1 if it does not already exist. If the request is successful request\'s result will be the connection.\n     */\n    open(name: string, version?: number): IDBOpenDBRequest;\n}\n\ndeclare var IDBFactory: {\n    prototype: IDBFactory;\n    new(): IDBFactory;\n};\n\n/** IDBIndex interface of the IndexedDB API provides asynchronous access to an index in a database. An index is a kind of object store for looking up records in another object store, called the referenced object store. You use this interface to retrieve data. */\ninterface IDBIndex {\n    readonly keyPath: string | string[];\n    readonly multiEntry: boolean;\n    /**\n     * Returns the name of the index.\n     */\n    name: string;\n    /**\n     * Returns the IDBObjectStore the index belongs to.\n     */\n    readonly objectStore: IDBObjectStore;\n    readonly unique: boolean;\n    /**\n     * Retrieves the number of records matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Retrieves the value of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(key: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(key: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    /**\n     * Opens a cursor over the records matching query, ordered by direction. If query is null, all records in index are matched.\n     * \n     * If successful, request\'s result will be an IDBCursorWithValue, or null if there were no matching records.\n     */\n    openCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in index are matched.\n     * \n     * If successful, request\'s result will be an IDBCursor, or null if there were no matching records.\n     */\n    openKeyCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n}\n\ndeclare var IDBIndex: {\n    prototype: IDBIndex;\n    new(): IDBIndex;\n};\n\n/** A key range can be a single value or a range with upper and lower bounds or endpoints. If the key range has both upper and lower bounds, then it is bounded; if it has no bounds, it is unbounded. A bounded key range can either be open (the endpoints are excluded) or closed (the endpoints are included). To retrieve all keys within a certain range, you can use the following code constructs: */\ninterface IDBKeyRange {\n    /**\n     * Returns lower bound, or undefined if none.\n     */\n    readonly lower: any;\n    /**\n     * Returns true if the lower open flag is set, and false otherwise.\n     */\n    readonly lowerOpen: boolean;\n    /**\n     * Returns upper bound, or undefined if none.\n     */\n    readonly upper: any;\n    /**\n     * Returns true if the upper open flag is set, and false otherwise.\n     */\n    readonly upperOpen: boolean;\n    /**\n     * Returns true if key is included in the range, and false otherwise.\n     */\n    includes(key: any): boolean;\n}\n\ndeclare var IDBKeyRange: {\n    prototype: IDBKeyRange;\n    new(): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning from lower to upper. If lowerOpen is true, lower is not included in the range. If upperOpen is true, upper is not included in the range.\n     */\n    bound(lower: any, upper: any, lowerOpen?: boolean, upperOpen?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange starting at key with no upper bound. If open is true, key is not included in the range.\n     */\n    lowerBound(lower: any, open?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning only key.\n     */\n    only(value: any): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange with no lower bound and ending at key. If open is true, key is not included in the range.\n     */\n    upperBound(upper: any, open?: boolean): IDBKeyRange;\n};\n\n/** This example shows a variety of different uses of object stores, from updating the data structure with IDBObjectStore.createIndex inside an onupgradeneeded function, to adding a new item to our object store with IDBObjectStore.add. For a full working example, see our To-do Notifications app (view example live.) */\ninterface IDBObjectStore {\n    /**\n     * Returns true if the store has a key generator, and false otherwise.\n     */\n    readonly autoIncrement: boolean;\n    /**\n     * Returns a list of the names of indexes in the store.\n     */\n    readonly indexNames: DOMStringList;\n    /**\n     * Returns the key path of the store, or null if none.\n     */\n    readonly keyPath: string | string[];\n    /**\n     * Returns the name of the store.\n     */\n    name: string;\n    /**\n     * Returns the associated transaction.\n     */\n    readonly transaction: IDBTransaction;\n    /**\n     * Adds or updates a record in store with the given value and key.\n     * \n     * If the store uses in-line keys and key is specified a "DataError" DOMException will be thrown.\n     * \n     * If put() is used, any existing record with the key will be replaced. If add() is used, and if a record with the key already exists the request will fail, with request\'s error set to a "ConstraintError" DOMException.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    add(value: any, key?: IDBValidKey): IDBRequest<IDBValidKey>;\n    /**\n     * Deletes all records in store.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    clear(): IDBRequest<undefined>;\n    /**\n     * Retrieves the number of records matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Creates a new index in store with the given name, keyPath and options and returns a new IDBIndex. If the keyPath and options define constraints that cannot be satisfied with the data already in store the upgrade transaction will abort with a "ConstraintError" DOMException.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createIndex(name: string, keyPath: string | string[], options?: IDBIndexParameters): IDBIndex;\n    /**\n     * Deletes records in store with the given key or in the given key range in query.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    delete(key: IDBValidKey | IDBKeyRange): IDBRequest<undefined>;\n    /**\n     * Deletes the index in store with the given name.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    deleteIndex(name: string): void;\n    /**\n     * Retrieves the value of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(query: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(query: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    index(name: string): IDBIndex;\n    /**\n     * Opens a cursor over the records matching query, ordered by direction. If query is null, all records in store are matched.\n     * \n     * If successful, request\'s result will be an IDBCursorWithValue pointing at the first matching record, or null if there were no matching records.\n     */\n    openCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in store are matched.\n     * \n     * If successful, request\'s result will be an IDBCursor pointing at the first matching record, or null if there were no matching records.\n     */\n    openKeyCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n    /**\n     * Adds or updates a record in store with the given value and key.\n     * \n     * If the store uses in-line keys and key is specified a "DataError" DOMException will be thrown.\n     * \n     * If put() is used, any existing record with the key will be replaced. If add() is used, and if a record with the key already exists the request will fail, with request\'s error set to a "ConstraintError" DOMException.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    put(value: any, key?: IDBValidKey): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBObjectStore: {\n    prototype: IDBObjectStore;\n    new(): IDBObjectStore;\n};\n\ninterface IDBOpenDBRequestEventMap extends IDBRequestEventMap {\n    "blocked": Event;\n    "upgradeneeded": IDBVersionChangeEvent;\n}\n\n/** Also inherits methods from its parents IDBRequest and EventTarget. */\ninterface IDBOpenDBRequest extends IDBRequest<IDBDatabase> {\n    onblocked: ((this: IDBOpenDBRequest, ev: Event) => any) | null;\n    onupgradeneeded: ((this: IDBOpenDBRequest, ev: IDBVersionChangeEvent) => any) | null;\n    addEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBOpenDBRequest: {\n    prototype: IDBOpenDBRequest;\n    new(): IDBOpenDBRequest;\n};\n\ninterface IDBRequestEventMap {\n    "error": Event;\n    "success": Event;\n}\n\n/** The request object does not initially contain any information about the result of the operation, but once information becomes available, an event is fired on the request, and the information becomes available through the properties of the IDBRequest instance. */\ninterface IDBRequest<T = any> extends EventTarget {\n    /**\n     * When a request is completed, returns the error (a DOMException), or null if the request succeeded. Throws a "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly error: DOMException | null;\n    onerror: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    onsuccess: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    /**\n     * Returns "pending" until a request is complete, then returns "done".\n     */\n    readonly readyState: IDBRequestReadyState;\n    /**\n     * When a request is completed, returns the result, or undefined if the request failed. Throws a "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly result: T;\n    /**\n     * Returns the IDBObjectStore, IDBIndex, or IDBCursor the request was made against, or null if is was an open request.\n     */\n    readonly source: IDBObjectStore | IDBIndex | IDBCursor;\n    /**\n     * Returns the IDBTransaction the request was made within. If this as an open request, then it returns an upgrade transaction while it is running, or null otherwise.\n     */\n    readonly transaction: IDBTransaction | null;\n    addEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBRequest: {\n    prototype: IDBRequest;\n    new(): IDBRequest;\n};\n\ninterface IDBTransactionEventMap {\n    "abort": Event;\n    "complete": Event;\n    "error": Event;\n}\n\ninterface IDBTransaction extends EventTarget {\n    /**\n     * Returns the transaction\'s connection.\n     */\n    readonly db: IDBDatabase;\n    /**\n     * If the transaction was aborted, returns the error (a DOMException) providing the reason.\n     */\n    readonly error: DOMException;\n    /**\n     * Returns the mode the transaction was created with ("readonly" or "readwrite"), or "versionchange" for an upgrade transaction.\n     */\n    readonly mode: IDBTransactionMode;\n    /**\n     * Returns a list of the names of object stores in the transaction\'s scope. For an upgrade transaction this is all object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBTransaction, ev: Event) => any) | null;\n    oncomplete: ((this: IDBTransaction, ev: Event) => any) | null;\n    onerror: ((this: IDBTransaction, ev: Event) => any) | null;\n    /**\n     * Aborts the transaction. All pending requests will fail with a "AbortError" DOMException and all changes made to the database will be reverted.\n     */\n    abort(): void;\n    /**\n     * Returns an IDBObjectStore in the transaction\'s scope.\n     */\n    objectStore(name: string): IDBObjectStore;\n    addEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBTransaction: {\n    prototype: IDBTransaction;\n    new(): IDBTransaction;\n};\n\n/** This IndexedDB API interface indicates that the version of the database has changed, as the result of an IDBOpenDBRequest.onupgradeneeded event handler function. */\ninterface IDBVersionChangeEvent extends Event {\n    readonly newVersion: number | null;\n    readonly oldVersion: number;\n}\n\ndeclare var IDBVersionChangeEvent: {\n    prototype: IDBVersionChangeEvent;\n    new(type: string, eventInitDict?: IDBVersionChangeEventInit): IDBVersionChangeEvent;\n};\n\n/** The IIRFilterNode interface of the Web Audio API is a AudioNode processor which implements a general infinite impulse response (IIR)  filter; this type of filter can be used to implement tone control devices and graphic equalizers as well. It lets the parameters of the filter response be specified, so that it can be tuned as needed. */\ninterface IIRFilterNode extends AudioNode {\n    getFrequencyResponse(frequencyHz: Float32Array, magResponse: Float32Array, phaseResponse: Float32Array): void;\n}\n\ndeclare var IIRFilterNode: {\n    prototype: IIRFilterNode;\n    new(context: BaseAudioContext, options: IIRFilterOptions): IIRFilterNode;\n};\n\ninterface ImageBitmap {\n    /**\n     * Returns the intrinsic height of the image, in CSS pixels.\n     */\n    readonly height: number;\n    /**\n     * Returns the intrinsic width of the image, in CSS pixels.\n     */\n    readonly width: number;\n    /**\n     * Releases imageBitmap\'s underlying bitmap data.\n     */\n    close(): void;\n}\n\ndeclare var ImageBitmap: {\n    prototype: ImageBitmap;\n    new(): ImageBitmap;\n};\n\ninterface ImageBitmapRenderingContext {\n    /**\n     * Returns the canvas element that the context is bound to.\n     */\n    readonly canvas: HTMLCanvasElement | OffscreenCanvas;\n    /**\n     * Transfers the underlying bitmap data from imageBitmap to context, and the bitmap becomes the contents of the canvas element to which context is bound.\n     */\n    transferFromImageBitmap(bitmap: ImageBitmap | null): void;\n}\n\ndeclare var ImageBitmapRenderingContext: {\n    prototype: ImageBitmapRenderingContext;\n    new(): ImageBitmapRenderingContext;\n};\n\n/** The underlying pixel data of an area of a <canvas> element. It is created using the ImageData() constructor or creator methods on the CanvasRenderingContext2D object associated with a canvas: createImageData() and getImageData(). It can also be used to set a part of the canvas by using putImageData(). */\ninterface ImageData {\n    /**\n     * Returns the one-dimensional array containing the data in RGBA order, as integers in the range 0 to 255.\n     */\n    readonly data: Uint8ClampedArray;\n    /**\n     * Returns the actual dimensions of the data in the ImageData object, in pixels.\n     */\n    readonly height: number;\n    /**\n     * Returns the actual dimensions of the data in the ImageData object, in pixels.\n     */\n    readonly width: number;\n}\n\ndeclare var ImageData: {\n    prototype: ImageData;\n    new(sw: number, sh: number): ImageData;\n    new(data: Uint8ClampedArray, sw: number, sh?: number): ImageData;\n};\n\ninterface InnerHTML {\n    innerHTML: string;\n}\n\ninterface InputDeviceInfo extends MediaDeviceInfo {\n    getCapabilities(): MediaTrackCapabilities;\n}\n\ndeclare var InputDeviceInfo: {\n    prototype: InputDeviceInfo;\n    new(): InputDeviceInfo;\n};\n\ninterface InputEvent extends UIEvent {\n    readonly data: string | null;\n    readonly inputType: string;\n    readonly isComposing: boolean;\n}\n\ndeclare var InputEvent: {\n    prototype: InputEvent;\n    new(type: string, eventInitDict?: InputEventInit): InputEvent;\n};\n\n/** provides a way to asynchronously observe changes in the intersection of a target element with an ancestor element or with a top-level document\'s viewport. */\ninterface IntersectionObserver {\n    readonly root: Element | null;\n    readonly rootMargin: string;\n    readonly thresholds: ReadonlyArray<number>;\n    disconnect(): void;\n    observe(target: Element): void;\n    takeRecords(): IntersectionObserverEntry[];\n    unobserve(target: Element): void;\n}\n\ndeclare var IntersectionObserver: {\n    prototype: IntersectionObserver;\n    new(callback: IntersectionObserverCallback, options?: IntersectionObserverInit): IntersectionObserver;\n};\n\n/** This Intersection Observer API interface describes the intersection between the target element and its root container at a specific moment of transition. */\ninterface IntersectionObserverEntry {\n    readonly boundingClientRect: DOMRectReadOnly;\n    readonly intersectionRatio: number;\n    readonly intersectionRect: DOMRectReadOnly;\n    readonly isIntersecting: boolean;\n    readonly rootBounds: DOMRectReadOnly | null;\n    readonly target: Element;\n    readonly time: number;\n}\n\ndeclare var IntersectionObserverEntry: {\n    prototype: IntersectionObserverEntry;\n    new(intersectionObserverEntryInit: IntersectionObserverEntryInit): IntersectionObserverEntry;\n};\n\n/** KeyboardEvent objects describe a user interaction with the keyboard; each event describes a single interaction between the user and a key (or combination of a key with modifier keys) on the keyboard. */\ninterface KeyboardEvent extends UIEvent {\n    readonly altKey: boolean;\n    /** @deprecated */\n    char: string;\n    /** @deprecated */\n    readonly charCode: number;\n    readonly code: string;\n    readonly ctrlKey: boolean;\n    readonly isComposing: boolean;\n    readonly key: string;\n    /** @deprecated */\n    readonly keyCode: number;\n    readonly location: number;\n    readonly metaKey: boolean;\n    readonly repeat: boolean;\n    readonly shiftKey: boolean;\n    getModifierState(keyArg: string): boolean;\n    readonly DOM_KEY_LOCATION_LEFT: number;\n    readonly DOM_KEY_LOCATION_NUMPAD: number;\n    readonly DOM_KEY_LOCATION_RIGHT: number;\n    readonly DOM_KEY_LOCATION_STANDARD: number;\n}\n\ndeclare var KeyboardEvent: {\n    prototype: KeyboardEvent;\n    new(type: string, eventInitDict?: KeyboardEventInit): KeyboardEvent;\n    readonly DOM_KEY_LOCATION_LEFT: number;\n    readonly DOM_KEY_LOCATION_NUMPAD: number;\n    readonly DOM_KEY_LOCATION_RIGHT: number;\n    readonly DOM_KEY_LOCATION_STANDARD: number;\n};\n\ninterface KeyframeEffect extends AnimationEffect {\n    composite: CompositeOperation;\n    iterationComposite: IterationCompositeOperation;\n    target: Element | null;\n    getKeyframes(): ComputedKeyframe[];\n    setKeyframes(keyframes: Keyframe[] | PropertyIndexedKeyframes | null): void;\n}\n\ndeclare var KeyframeEffect: {\n    prototype: KeyframeEffect;\n    new(target: Element | null, keyframes: Keyframe[] | PropertyIndexedKeyframes | null, options?: number | KeyframeEffectOptions): KeyframeEffect;\n    new(source: KeyframeEffect): KeyframeEffect;\n};\n\ninterface LinkStyle {\n    readonly sheet: CSSStyleSheet | null;\n}\n\ninterface ListeningStateChangedEvent extends Event {\n    readonly label: string;\n    readonly state: ListeningState;\n}\n\ndeclare var ListeningStateChangedEvent: {\n    prototype: ListeningStateChangedEvent;\n    new(): ListeningStateChangedEvent;\n};\n\n/** The location (URL) of the object it is linked to. Changes done on it are reflected on the object it relates to. Both the Document and Window interface have such a linked Location, accessible via Document.location and Window.location respectively. */\ninterface Location {\n    /**\n     * Returns a DOMStringList object listing the origins of the ancestor browsing contexts, from the parent browsing context to the top-level browsing context.\n     */\n    readonly ancestorOrigins: DOMStringList;\n    /**\n     * Returns the Location object\'s URL\'s fragment (includes leading "#" if non-empty).\n     * \n     * Can be set, to navigate to the same URL with a changed fragment (ignores leading "#").\n     */\n    hash: string;\n    /**\n     * Returns the Location object\'s URL\'s host and port (if different from the default port for the scheme).\n     * \n     * Can be set, to navigate to the same URL with a changed host and port.\n     */\n    host: string;\n    /**\n     * Returns the Location object\'s URL\'s host.\n     * \n     * Can be set, to navigate to the same URL with a changed host.\n     */\n    hostname: string;\n    /**\n     * Returns the Location object\'s URL.\n     * \n     * Can be set, to navigate to the given URL.\n     */\n    href: string;\n    toString(): string;\n    /**\n     * Returns the Location object\'s URL\'s origin.\n     */\n    readonly origin: string;\n    /**\n     * Returns the Location object\'s URL\'s path.\n     * \n     * Can be set, to navigate to the same URL with a changed path.\n     */\n    pathname: string;\n    /**\n     * Returns the Location object\'s URL\'s port.\n     * \n     * Can be set, to navigate to the same URL with a changed port.\n     */\n    port: string;\n    /**\n     * Returns the Location object\'s URL\'s scheme.\n     * \n     * Can be set, to navigate to the same URL with a changed scheme.\n     */\n    protocol: string;\n    /**\n     * Returns the Location object\'s URL\'s query (includes leading "?" if non-empty).\n     * \n     * Can be set, to navigate to the same URL with a changed query (ignores leading "?").\n     */\n    search: string;\n    /**\n     * Navigates to the given URL.\n     */\n    assign(url: string): void;\n    /**\n     * Reloads the current page.\n     */\n    reload(): void;\n    /** @deprecated */\n    reload(forcedReload: boolean): void;\n    /**\n     * Removes the current page from the session history and navigates to the given URL.\n     */\n    replace(url: string): void;\n}\n\ndeclare var Location: {\n    prototype: Location;\n    new(): Location;\n};\n\ninterface MSAssertion {\n    readonly id: string;\n    readonly type: MSCredentialType;\n}\n\ndeclare var MSAssertion: {\n    prototype: MSAssertion;\n    new(): MSAssertion;\n};\n\ninterface MSBlobBuilder {\n    append(data: any, endings?: string): void;\n    getBlob(contentType?: string): Blob;\n}\n\ndeclare var MSBlobBuilder: {\n    prototype: MSBlobBuilder;\n    new(): MSBlobBuilder;\n};\n\ninterface MSFIDOCredentialAssertion extends MSAssertion {\n    readonly algorithm: string | Algorithm;\n    readonly attestation: any;\n    readonly publicKey: string;\n    readonly transportHints: MSTransportType[];\n}\n\ndeclare var MSFIDOCredentialAssertion: {\n    prototype: MSFIDOCredentialAssertion;\n    new(): MSFIDOCredentialAssertion;\n};\n\ninterface MSFIDOSignature {\n    readonly authnrData: string;\n    readonly clientData: string;\n    readonly signature: string;\n}\n\ndeclare var MSFIDOSignature: {\n    prototype: MSFIDOSignature;\n    new(): MSFIDOSignature;\n};\n\ninterface MSFIDOSignatureAssertion extends MSAssertion {\n    readonly signature: MSFIDOSignature;\n}\n\ndeclare var MSFIDOSignatureAssertion: {\n    prototype: MSFIDOSignatureAssertion;\n    new(): MSFIDOSignatureAssertion;\n};\n\ninterface MSFileSaver {\n    msSaveBlob(blob: any, defaultName?: string): boolean;\n    msSaveOrOpenBlob(blob: any, defaultName?: string): boolean;\n}\n\ninterface MSGesture {\n    target: Element;\n    addPointer(pointerId: number): void;\n    stop(): void;\n}\n\ndeclare var MSGesture: {\n    prototype: MSGesture;\n    new(): MSGesture;\n};\n\n/** The MSGestureEvent is a proprietary interface specific to Internet Explorer and Microsoft Edge which represents events that occur due to touch gestures. Events using this interface include MSGestureStart, MSGestureEnd, MSGestureTap, MSGestureHold, MSGestureChange, and MSInertiaStart. */\ninterface MSGestureEvent extends UIEvent {\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly expansion: number;\n    readonly gestureObject: any;\n    readonly hwTimestamp: number;\n    readonly offsetX: number;\n    readonly offsetY: number;\n    readonly rotation: number;\n    readonly scale: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly translationX: number;\n    readonly translationY: number;\n    readonly velocityAngular: number;\n    readonly velocityExpansion: number;\n    readonly velocityX: number;\n    readonly velocityY: number;\n    initGestureEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, offsetXArg: number, offsetYArg: number, translationXArg: number, translationYArg: number, scaleArg: number, expansionArg: number, rotationArg: number, velocityXArg: number, velocityYArg: number, velocityExpansionArg: number, velocityAngularArg: number, hwTimestampArg: number): void;\n    readonly MSGESTURE_FLAG_BEGIN: number;\n    readonly MSGESTURE_FLAG_CANCEL: number;\n    readonly MSGESTURE_FLAG_END: number;\n    readonly MSGESTURE_FLAG_INERTIA: number;\n    readonly MSGESTURE_FLAG_NONE: number;\n}\n\ndeclare var MSGestureEvent: {\n    prototype: MSGestureEvent;\n    new(): MSGestureEvent;\n    readonly MSGESTURE_FLAG_BEGIN: number;\n    readonly MSGESTURE_FLAG_CANCEL: number;\n    readonly MSGESTURE_FLAG_END: number;\n    readonly MSGESTURE_FLAG_INERTIA: number;\n    readonly MSGESTURE_FLAG_NONE: number;\n};\n\n/** The msGraphicsTrust() constructor returns an object that provides properties for info on protected video playback. */\ninterface MSGraphicsTrust {\n    readonly constrictionActive: boolean;\n    readonly status: string;\n}\n\ndeclare var MSGraphicsTrust: {\n    prototype: MSGraphicsTrust;\n    new(): MSGraphicsTrust;\n};\n\ninterface MSInputMethodContextEventMap {\n    "MSCandidateWindowHide": Event;\n    "MSCandidateWindowShow": Event;\n    "MSCandidateWindowUpdate": Event;\n}\n\ninterface MSInputMethodContext extends EventTarget {\n    readonly compositionEndOffset: number;\n    readonly compositionStartOffset: number;\n    oncandidatewindowhide: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    oncandidatewindowshow: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    oncandidatewindowupdate: ((this: MSInputMethodContext, ev: Event) => any) | null;\n    readonly target: HTMLElement;\n    getCandidateWindowClientRect(): ClientRect;\n    getCompositionAlternatives(): string[];\n    hasComposition(): boolean;\n    isCandidateWindowVisible(): boolean;\n    addEventListener<K extends keyof MSInputMethodContextEventMap>(type: K, listener: (this: MSInputMethodContext, ev: MSInputMethodContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MSInputMethodContextEventMap>(type: K, listener: (this: MSInputMethodContext, ev: MSInputMethodContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MSInputMethodContext: {\n    prototype: MSInputMethodContext;\n    new(): MSInputMethodContext;\n};\n\ninterface MSMediaKeyError {\n    readonly code: number;\n    readonly systemCode: number;\n    readonly MS_MEDIA_KEYERR_CLIENT: number;\n    readonly MS_MEDIA_KEYERR_DOMAIN: number;\n    readonly MS_MEDIA_KEYERR_HARDWARECHANGE: number;\n    readonly MS_MEDIA_KEYERR_OUTPUT: number;\n    readonly MS_MEDIA_KEYERR_SERVICE: number;\n    readonly MS_MEDIA_KEYERR_UNKNOWN: number;\n}\n\ndeclare var MSMediaKeyError: {\n    prototype: MSMediaKeyError;\n    new(): MSMediaKeyError;\n    readonly MS_MEDIA_KEYERR_CLIENT: number;\n    readonly MS_MEDIA_KEYERR_DOMAIN: number;\n    readonly MS_MEDIA_KEYERR_HARDWARECHANGE: number;\n    readonly MS_MEDIA_KEYERR_OUTPUT: number;\n    readonly MS_MEDIA_KEYERR_SERVICE: number;\n    readonly MS_MEDIA_KEYERR_UNKNOWN: number;\n};\n\ninterface MSMediaKeyMessageEvent extends Event {\n    readonly destinationURL: string | null;\n    readonly message: Uint8Array;\n}\n\ndeclare var MSMediaKeyMessageEvent: {\n    prototype: MSMediaKeyMessageEvent;\n    new(): MSMediaKeyMessageEvent;\n};\n\ninterface MSMediaKeyNeededEvent extends Event {\n    readonly initData: Uint8Array | null;\n}\n\ndeclare var MSMediaKeyNeededEvent: {\n    prototype: MSMediaKeyNeededEvent;\n    new(): MSMediaKeyNeededEvent;\n};\n\ninterface MSMediaKeySession extends EventTarget {\n    readonly error: MSMediaKeyError | null;\n    readonly keySystem: string;\n    readonly sessionId: string;\n    close(): void;\n    update(key: Uint8Array): void;\n}\n\ndeclare var MSMediaKeySession: {\n    prototype: MSMediaKeySession;\n    new(): MSMediaKeySession;\n};\n\ninterface MSMediaKeys {\n    readonly keySystem: string;\n    createSession(type: string, initData: Uint8Array, cdmData?: Uint8Array | null): MSMediaKeySession;\n}\n\ndeclare var MSMediaKeys: {\n    prototype: MSMediaKeys;\n    new(keySystem: string): MSMediaKeys;\n    isTypeSupported(keySystem: string, type?: string | null): boolean;\n    isTypeSupportedWithFeatures(keySystem: string, type?: string | null): string;\n};\n\ninterface MSNavigatorDoNotTrack {\n    confirmSiteSpecificTrackingException(args: ConfirmSiteSpecificExceptionsInformation): boolean;\n    confirmWebWideTrackingException(args: ExceptionInformation): boolean;\n    removeSiteSpecificTrackingException(args: ExceptionInformation): void;\n    removeWebWideTrackingException(args: ExceptionInformation): void;\n    storeSiteSpecificTrackingException(args: StoreSiteSpecificExceptionsInformation): void;\n    storeWebWideTrackingException(args: StoreExceptionsInformation): void;\n}\n\ninterface MSPointerEvent extends MouseEvent {\n    readonly currentPoint: any;\n    readonly height: number;\n    readonly hwTimestamp: number;\n    readonly intermediatePoints: any;\n    readonly isPrimary: boolean;\n    readonly pointerId: number;\n    readonly pointerType: any;\n    readonly pressure: number;\n    readonly rotation: number;\n    readonly tiltX: number;\n    readonly tiltY: number;\n    readonly width: number;\n    getCurrentPoint(element: Element): void;\n    getIntermediatePoints(element: Element): void;\n    initPointerEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, ctrlKeyArg: boolean, altKeyArg: boolean, shiftKeyArg: boolean, metaKeyArg: boolean, buttonArg: number, relatedTargetArg: EventTarget, offsetXArg: number, offsetYArg: number, widthArg: number, heightArg: number, pressure: number, rotation: number, tiltX: number, tiltY: number, pointerIdArg: number, pointerType: any, hwTimestampArg: number, isPrimary: boolean): void;\n}\n\ndeclare var MSPointerEvent: {\n    prototype: MSPointerEvent;\n    new(typeArg: string, eventInitDict?: PointerEventInit): MSPointerEvent;\n};\n\ninterface MSStream {\n    readonly type: string;\n    msClose(): void;\n    msDetachStream(): any;\n}\n\ndeclare var MSStream: {\n    prototype: MSStream;\n    new(): MSStream;\n};\n\n/** The MediaDevicesInfo interface contains information that describes a single media input or output device. */\ninterface MediaDeviceInfo {\n    readonly deviceId: string;\n    readonly groupId: string;\n    readonly kind: MediaDeviceKind;\n    readonly label: string;\n    toJSON(): any;\n}\n\ndeclare var MediaDeviceInfo: {\n    prototype: MediaDeviceInfo;\n    new(): MediaDeviceInfo;\n};\n\ninterface MediaDevicesEventMap {\n    "devicechange": Event;\n}\n\n/** Provides access to connected media input devices like cameras and microphones, as well as screen sharing. In essence, it lets you obtain access to any hardware source of media data. */\ninterface MediaDevices extends EventTarget {\n    ondevicechange: ((this: MediaDevices, ev: Event) => any) | null;\n    enumerateDevices(): Promise<MediaDeviceInfo[]>;\n    getSupportedConstraints(): MediaTrackSupportedConstraints;\n    getUserMedia(constraints?: MediaStreamConstraints): Promise<MediaStream>;\n    addEventListener<K extends keyof MediaDevicesEventMap>(type: K, listener: (this: MediaDevices, ev: MediaDevicesEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaDevicesEventMap>(type: K, listener: (this: MediaDevices, ev: MediaDevicesEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaDevices: {\n    prototype: MediaDevices;\n    new(): MediaDevices;\n};\n\n/** A MediaElementSourceNode has no inputs and exactly one output, and is created using the AudioContext.createMediaElementSource method. The amount of channels in the output equals the number of channels of the audio referenced by the HTMLMediaElement used in the creation of the node, or is 1 if the HTMLMediaElement has no audio. */\ninterface MediaElementAudioSourceNode extends AudioNode {\n    readonly mediaElement: HTMLMediaElement;\n}\n\ndeclare var MediaElementAudioSourceNode: {\n    prototype: MediaElementAudioSourceNode;\n    new(context: AudioContext, options: MediaElementAudioSourceOptions): MediaElementAudioSourceNode;\n};\n\ninterface MediaEncryptedEvent extends Event {\n    readonly initData: ArrayBuffer | null;\n    readonly initDataType: string;\n}\n\ndeclare var MediaEncryptedEvent: {\n    prototype: MediaEncryptedEvent;\n    new(type: string, eventInitDict?: MediaEncryptedEventInit): MediaEncryptedEvent;\n};\n\n/** An error which occurred while handling media in an HTML media element based on HTMLMediaElement, such as <audio> or <video>. */\ninterface MediaError {\n    readonly code: number;\n    readonly message: string;\n    readonly MEDIA_ERR_ABORTED: number;\n    readonly MEDIA_ERR_DECODE: number;\n    readonly MEDIA_ERR_NETWORK: number;\n    readonly MEDIA_ERR_SRC_NOT_SUPPORTED: number;\n}\n\ndeclare var MediaError: {\n    prototype: MediaError;\n    new(): MediaError;\n    readonly MEDIA_ERR_ABORTED: number;\n    readonly MEDIA_ERR_DECODE: number;\n    readonly MEDIA_ERR_NETWORK: number;\n    readonly MEDIA_ERR_SRC_NOT_SUPPORTED: number;\n};\n\n/** This EncryptedMediaExtensions API interface contains the content and related data when the content decryption module generates a message for the session. */\ninterface MediaKeyMessageEvent extends Event {\n    readonly message: ArrayBuffer;\n    readonly messageType: MediaKeyMessageType;\n}\n\ndeclare var MediaKeyMessageEvent: {\n    prototype: MediaKeyMessageEvent;\n    new(type: string, eventInitDict: MediaKeyMessageEventInit): MediaKeyMessageEvent;\n};\n\ninterface MediaKeySessionEventMap {\n    "keystatuseschange": Event;\n    "message": MediaKeyMessageEvent;\n}\n\n/** This EncryptedMediaExtensions API interface represents a context for message exchange with a content decryption module (CDM). */\ninterface MediaKeySession extends EventTarget {\n    readonly closed: Promise<void>;\n    readonly expiration: number;\n    readonly keyStatuses: MediaKeyStatusMap;\n    onkeystatuseschange: ((this: MediaKeySession, ev: Event) => any) | null;\n    onmessage: ((this: MediaKeySession, ev: MediaKeyMessageEvent) => any) | null;\n    readonly sessionId: string;\n    close(): Promise<void>;\n    generateRequest(initDataType: string, initData: BufferSource): Promise<void>;\n    load(sessionId: string): Promise<boolean>;\n    remove(): Promise<void>;\n    update(response: BufferSource): Promise<void>;\n    addEventListener<K extends keyof MediaKeySessionEventMap>(type: K, listener: (this: MediaKeySession, ev: MediaKeySessionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaKeySessionEventMap>(type: K, listener: (this: MediaKeySession, ev: MediaKeySessionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaKeySession: {\n    prototype: MediaKeySession;\n    new(): MediaKeySession;\n};\n\n/** This EncryptedMediaExtensions API interface is a read-only map of media key statuses by key IDs. */\ninterface MediaKeyStatusMap {\n    readonly size: number;\n    get(keyId: BufferSource): MediaKeyStatus | undefined;\n    has(keyId: BufferSource): boolean;\n    forEach(callbackfn: (value: MediaKeyStatus, key: BufferSource, parent: MediaKeyStatusMap) => void, thisArg?: any): void;\n}\n\ndeclare var MediaKeyStatusMap: {\n    prototype: MediaKeyStatusMap;\n    new(): MediaKeyStatusMap;\n};\n\n/** This EncryptedMediaExtensions API interface provides access to a Key System for decryption and/or a content protection provider. You can request an instance of this object using the Navigator.requestMediaKeySystemAccess method. */\ninterface MediaKeySystemAccess {\n    readonly keySystem: string;\n    createMediaKeys(): Promise<MediaKeys>;\n    getConfiguration(): MediaKeySystemConfiguration;\n}\n\ndeclare var MediaKeySystemAccess: {\n    prototype: MediaKeySystemAccess;\n    new(): MediaKeySystemAccess;\n};\n\n/** This EncryptedMediaExtensions API interface the represents a set of keys that an associated HTMLMediaElement can use for decryption of media data during playback. */\ninterface MediaKeys {\n    createSession(sessionType?: MediaKeySessionType): MediaKeySession;\n    setServerCertificate(serverCertificate: BufferSource): Promise<boolean>;\n}\n\ndeclare var MediaKeys: {\n    prototype: MediaKeys;\n    new(): MediaKeys;\n};\n\ninterface MediaList {\n    readonly length: number;\n    mediaText: string;\n    toString(): string;\n    appendMedium(medium: string): void;\n    deleteMedium(medium: string): void;\n    item(index: number): string | null;\n    [index: number]: string;\n}\n\ndeclare var MediaList: {\n    prototype: MediaList;\n    new(): MediaList;\n};\n\ninterface MediaQueryListEventMap {\n    "change": MediaQueryListEvent;\n}\n\n/** Stores information on a media query applied to a document, and handles sending notifications to listeners when the media query state change (i.e. when the media query test starts or stops evaluating to true). */\ninterface MediaQueryList extends EventTarget {\n    readonly matches: boolean;\n    readonly media: string;\n    onchange: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null;\n    /** @deprecated */\n    addListener(listener: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null): void;\n    /** @deprecated */\n    removeListener(listener: ((this: MediaQueryList, ev: MediaQueryListEvent) => any) | null): void;\n    addEventListener<K extends keyof MediaQueryListEventMap>(type: K, listener: (this: MediaQueryList, ev: MediaQueryListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaQueryListEventMap>(type: K, listener: (this: MediaQueryList, ev: MediaQueryListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaQueryList: {\n    prototype: MediaQueryList;\n    new(): MediaQueryList;\n};\n\ninterface MediaQueryListEvent extends Event {\n    readonly matches: boolean;\n    readonly media: string;\n}\n\ndeclare var MediaQueryListEvent: {\n    prototype: MediaQueryListEvent;\n    new(type: string, eventInitDict?: MediaQueryListEventInit): MediaQueryListEvent;\n};\n\ninterface MediaSourceEventMap {\n    "sourceclose": Event;\n    "sourceended": Event;\n    "sourceopen": Event;\n}\n\n/** This Media Source Extensions API interface represents a source of media data for an HTMLMediaElement object. A MediaSource object can be attached to a HTMLMediaElement to be played in the user agent. */\ninterface MediaSource extends EventTarget {\n    readonly activeSourceBuffers: SourceBufferList;\n    duration: number;\n    onsourceclose: ((this: MediaSource, ev: Event) => any) | null;\n    onsourceended: ((this: MediaSource, ev: Event) => any) | null;\n    onsourceopen: ((this: MediaSource, ev: Event) => any) | null;\n    readonly readyState: ReadyState;\n    readonly sourceBuffers: SourceBufferList;\n    addSourceBuffer(type: string): SourceBuffer;\n    clearLiveSeekableRange(): void;\n    endOfStream(error?: EndOfStreamError): void;\n    removeSourceBuffer(sourceBuffer: SourceBuffer): void;\n    setLiveSeekableRange(start: number, end: number): void;\n    addEventListener<K extends keyof MediaSourceEventMap>(type: K, listener: (this: MediaSource, ev: MediaSourceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaSourceEventMap>(type: K, listener: (this: MediaSource, ev: MediaSourceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaSource: {\n    prototype: MediaSource;\n    new(): MediaSource;\n    isTypeSupported(type: string): boolean;\n};\n\ninterface MediaStreamEventMap {\n    "addtrack": MediaStreamTrackEvent;\n    "removetrack": MediaStreamTrackEvent;\n}\n\n/** A stream of media content. A stream consists of several tracks such as video or audio tracks. Each track is specified as an instance of MediaStreamTrack. */\ninterface MediaStream extends EventTarget {\n    readonly active: boolean;\n    readonly id: string;\n    onaddtrack: ((this: MediaStream, ev: MediaStreamTrackEvent) => any) | null;\n    onremovetrack: ((this: MediaStream, ev: MediaStreamTrackEvent) => any) | null;\n    addTrack(track: MediaStreamTrack): void;\n    clone(): MediaStream;\n    getAudioTracks(): MediaStreamTrack[];\n    getTrackById(trackId: string): MediaStreamTrack | null;\n    getTracks(): MediaStreamTrack[];\n    getVideoTracks(): MediaStreamTrack[];\n    removeTrack(track: MediaStreamTrack): void;\n    addEventListener<K extends keyof MediaStreamEventMap>(type: K, listener: (this: MediaStream, ev: MediaStreamEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaStreamEventMap>(type: K, listener: (this: MediaStream, ev: MediaStreamEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaStream: {\n    prototype: MediaStream;\n    new(): MediaStream;\n    new(stream: MediaStream): MediaStream;\n    new(tracks: MediaStreamTrack[]): MediaStream;\n};\n\ninterface MediaStreamAudioDestinationNode extends AudioNode {\n    readonly stream: MediaStream;\n}\n\ndeclare var MediaStreamAudioDestinationNode: {\n    prototype: MediaStreamAudioDestinationNode;\n    new(context: AudioContext, options?: AudioNodeOptions): MediaStreamAudioDestinationNode;\n};\n\n/** A type of AudioNode which operates as an audio source whose media is received from a MediaStream obtained using the WebRTC or Media Capture and Streams APIs. */\ninterface MediaStreamAudioSourceNode extends AudioNode {\n    readonly mediaStream: MediaStream;\n}\n\ndeclare var MediaStreamAudioSourceNode: {\n    prototype: MediaStreamAudioSourceNode;\n    new(context: AudioContext, options: MediaStreamAudioSourceOptions): MediaStreamAudioSourceNode;\n};\n\ninterface MediaStreamError {\n    readonly constraintName: string | null;\n    readonly message: string | null;\n    readonly name: string;\n}\n\ndeclare var MediaStreamError: {\n    prototype: MediaStreamError;\n    new(): MediaStreamError;\n};\n\ninterface MediaStreamErrorEvent extends Event {\n    readonly error: MediaStreamError | null;\n}\n\ndeclare var MediaStreamErrorEvent: {\n    prototype: MediaStreamErrorEvent;\n    new(typeArg: string, eventInitDict?: MediaStreamErrorEventInit): MediaStreamErrorEvent;\n};\n\n/** Events that occurs in relation to a MediaStream. Two events of this type can be thrown: addstream and removestream. */\ninterface MediaStreamEvent extends Event {\n    readonly stream: MediaStream | null;\n}\n\ndeclare var MediaStreamEvent: {\n    prototype: MediaStreamEvent;\n    new(type: string, eventInitDict: MediaStreamEventInit): MediaStreamEvent;\n};\n\ninterface MediaStreamTrackEventMap {\n    "ended": Event;\n    "isolationchange": Event;\n    "mute": Event;\n    "unmute": Event;\n}\n\n/** A single media track within a stream; typically, these are audio or video tracks, but other track types may exist as well. */\ninterface MediaStreamTrack extends EventTarget {\n    enabled: boolean;\n    readonly id: string;\n    readonly isolated: boolean;\n    readonly kind: string;\n    readonly label: string;\n    readonly muted: boolean;\n    onended: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    onisolationchange: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    onmute: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    onunmute: ((this: MediaStreamTrack, ev: Event) => any) | null;\n    readonly readyState: MediaStreamTrackState;\n    applyConstraints(constraints?: MediaTrackConstraints): Promise<void>;\n    clone(): MediaStreamTrack;\n    getCapabilities(): MediaTrackCapabilities;\n    getConstraints(): MediaTrackConstraints;\n    getSettings(): MediaTrackSettings;\n    stop(): void;\n    addEventListener<K extends keyof MediaStreamTrackEventMap>(type: K, listener: (this: MediaStreamTrack, ev: MediaStreamTrackEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MediaStreamTrackEventMap>(type: K, listener: (this: MediaStreamTrack, ev: MediaStreamTrackEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MediaStreamTrack: {\n    prototype: MediaStreamTrack;\n    new(): MediaStreamTrack;\n};\n\ninterface MediaStreamTrackAudioSourceNode extends AudioNode {\n}\n\ndeclare var MediaStreamTrackAudioSourceNode: {\n    prototype: MediaStreamTrackAudioSourceNode;\n    new(context: AudioContext, options: MediaStreamTrackAudioSourceOptions): MediaStreamTrackAudioSourceNode;\n};\n\n/** Events which indicate that a MediaStream has had tracks added to or removed from the stream through calls to Media Stream API methods. These events are sent to the stream when these changes occur. */\ninterface MediaStreamTrackEvent extends Event {\n    readonly track: MediaStreamTrack;\n}\n\ndeclare var MediaStreamTrackEvent: {\n    prototype: MediaStreamTrackEvent;\n    new(type: string, eventInitDict: MediaStreamTrackEventInit): MediaStreamTrackEvent;\n};\n\n/** This Channel Messaging API interface allows us to create a new message channel and send data through it via its two MessagePort properties. */\ninterface MessageChannel {\n    /**\n     * Returns the first MessagePort object.\n     */\n    readonly port1: MessagePort;\n    /**\n     * Returns the second MessagePort object.\n     */\n    readonly port2: MessagePort;\n}\n\ndeclare var MessageChannel: {\n    prototype: MessageChannel;\n    new(): MessageChannel;\n};\n\n/** A message received by a target object. */\ninterface MessageEvent<T = any> extends Event {\n    /**\n     * Returns the data of the message.\n     */\n    readonly data: T;\n    /**\n     * Returns the last event ID string, for server-sent events.\n     */\n    readonly lastEventId: string;\n    /**\n     * Returns the origin of the message, for server-sent events and cross-document messaging.\n     */\n    readonly origin: string;\n    /**\n     * Returns the MessagePort array sent with the message, for cross-document messaging and channel messaging.\n     */\n    readonly ports: ReadonlyArray<MessagePort>;\n    /**\n     * Returns the WindowProxy of the source window, for cross-document messaging, and the MessagePort being attached, in the connect event fired at SharedWorkerGlobalScope objects.\n     */\n    readonly source: MessageEventSource | null;\n}\n\ndeclare var MessageEvent: {\n    prototype: MessageEvent;\n    new<T>(type: string, eventInitDict?: MessageEventInit<T>): MessageEvent<T>;\n};\n\ninterface MessagePortEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** This Channel Messaging API interface represents one of the two ports of a MessageChannel, allowing messages to be sent from one port and listening out for them arriving at the other. */\ninterface MessagePort extends EventTarget {\n    onmessage: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    /**\n     * Disconnects the port, so that it is no longer active.\n     */\n    close(): void;\n    /**\n     * Posts a message through the channel. Objects listed in transfer are transferred, not just cloned, meaning that they are no longer usable on the sending side.\n     * \n     * Throws a "DataCloneError" DOMException if transfer contains duplicate objects or port, or if message could not be cloned.\n     */\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    /**\n     * Begins dispatching messages received on the port.\n     */\n    start(): void;\n    addEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MessagePort: {\n    prototype: MessagePort;\n    new(): MessagePort;\n};\n\n/** Provides contains information about a MIME type associated with a particular plugin. NavigatorPlugins.mimeTypes returns an array of this object. */\ninterface MimeType {\n    /**\n     * Returns the MIME type\'s description.\n     */\n    readonly description: string;\n    /**\n     * Returns the Plugin object that implements this MIME type.\n     */\n    readonly enabledPlugin: Plugin;\n    /**\n     * Returns the MIME type\'s typical file extensions, in a comma-separated list.\n     */\n    readonly suffixes: string;\n    /**\n     * Returns the MIME type.\n     */\n    readonly type: string;\n}\n\ndeclare var MimeType: {\n    prototype: MimeType;\n    new(): MimeType;\n};\n\n/** Returns an array of MimeType instances, each of which contains information about a supported browser plugins. This object is returned by NavigatorPlugins.mimeTypes. */\ninterface MimeTypeArray {\n    readonly length: number;\n    item(index: number): MimeType | null;\n    namedItem(name: string): MimeType | null;\n    [index: number]: MimeType;\n}\n\ndeclare var MimeTypeArray: {\n    prototype: MimeTypeArray;\n    new(): MimeTypeArray;\n};\n\n/** Events that occur due to the user interacting with a pointing device (such as a mouse). Common events using this interface include click, dblclick, mouseup, mousedown. */\ninterface MouseEvent extends UIEvent {\n    readonly altKey: boolean;\n    readonly button: number;\n    readonly buttons: number;\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly ctrlKey: boolean;\n    readonly metaKey: boolean;\n    readonly movementX: number;\n    readonly movementY: number;\n    readonly offsetX: number;\n    readonly offsetY: number;\n    readonly pageX: number;\n    readonly pageY: number;\n    readonly relatedTarget: EventTarget | null;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly shiftKey: boolean;\n    readonly x: number;\n    readonly y: number;\n    getModifierState(keyArg: string): boolean;\n    initMouseEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, detailArg: number, screenXArg: number, screenYArg: number, clientXArg: number, clientYArg: number, ctrlKeyArg: boolean, altKeyArg: boolean, shiftKeyArg: boolean, metaKeyArg: boolean, buttonArg: number, relatedTargetArg: EventTarget | null): void;\n}\n\ndeclare var MouseEvent: {\n    prototype: MouseEvent;\n    new(type: string, eventInitDict?: MouseEventInit): MouseEvent;\n};\n\n/** Provides event properties that are specific to modifications to the Document Object Model (DOM) hierarchy and nodes. */\ninterface MutationEvent extends Event {\n    readonly attrChange: number;\n    readonly attrName: string;\n    readonly newValue: string;\n    readonly prevValue: string;\n    readonly relatedNode: Node;\n    initMutationEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, relatedNodeArg: Node, prevValueArg: string, newValueArg: string, attrNameArg: string, attrChangeArg: number): void;\n    readonly ADDITION: number;\n    readonly MODIFICATION: number;\n    readonly REMOVAL: number;\n}\n\ndeclare var MutationEvent: {\n    prototype: MutationEvent;\n    new(): MutationEvent;\n    readonly ADDITION: number;\n    readonly MODIFICATION: number;\n    readonly REMOVAL: number;\n};\n\n/** Provides the ability to watch for changes being made to the DOM tree. It is designed as a replacement for the older Mutation Events feature which was part of the DOM3 Events specification. */\ninterface MutationObserver {\n    /**\n     * Stops observer from observing any mutations. Until the observe() method is used again, observer\'s callback will not be invoked.\n     */\n    disconnect(): void;\n    /**\n     * Instructs the user agent to observe a given target (a node) and report any mutations based on the criteria given by options (an object).\n     * \n     * The options argument allows for setting mutation observation options via object members.\n     */\n    observe(target: Node, options?: MutationObserverInit): void;\n    /**\n     * Empties the record queue and returns what was in there.\n     */\n    takeRecords(): MutationRecord[];\n}\n\ndeclare var MutationObserver: {\n    prototype: MutationObserver;\n    new(callback: MutationCallback): MutationObserver;\n};\n\n/** A MutationRecord represents an individual DOM mutation. It is the object that is passed to MutationObserver\'s callback. */\ninterface MutationRecord {\n    /**\n     * Return the nodes added and removed respectively.\n     */\n    readonly addedNodes: NodeList;\n    /**\n     * Returns the local name of the changed attribute, and null otherwise.\n     */\n    readonly attributeName: string | null;\n    /**\n     * Returns the namespace of the changed attribute, and null otherwise.\n     */\n    readonly attributeNamespace: string | null;\n    /**\n     * Return the previous and next sibling respectively of the added or removed nodes, and null otherwise.\n     */\n    readonly nextSibling: Node | null;\n    /**\n     * The return value depends on type. For "attributes", it is the value of the changed attribute before the change. For "characterData", it is the data of the changed node before the change. For "childList", it is null.\n     */\n    readonly oldValue: string | null;\n    /**\n     * Return the previous and next sibling respectively of the added or removed nodes, and null otherwise.\n     */\n    readonly previousSibling: Node | null;\n    /**\n     * Return the nodes added and removed respectively.\n     */\n    readonly removedNodes: NodeList;\n    /**\n     * Returns the node the mutation affected, depending on the type. For "attributes", it is the element whose attribute changed. For "characterData", it is the CharacterData node. For "childList", it is the node whose children changed.\n     */\n    readonly target: Node;\n    /**\n     * Returns "attributes" if it was an attribute mutation. "characterData" if it was a mutation to a CharacterData node. And "childList" if it was a mutation to the tree of nodes.\n     */\n    readonly type: MutationRecordType;\n}\n\ndeclare var MutationRecord: {\n    prototype: MutationRecord;\n    new(): MutationRecord;\n};\n\n/** A collection of Attr objects. Objects inside a NamedNodeMap are not in any particular order, unlike NodeList, although they may be accessed by an index as in an array. */\ninterface NamedNodeMap {\n    readonly length: number;\n    getNamedItem(qualifiedName: string): Attr | null;\n    getNamedItemNS(namespace: string | null, localName: string): Attr | null;\n    item(index: number): Attr | null;\n    removeNamedItem(qualifiedName: string): Attr;\n    removeNamedItemNS(namespace: string | null, localName: string): Attr;\n    setNamedItem(attr: Attr): Attr | null;\n    setNamedItemNS(attr: Attr): Attr | null;\n    [index: number]: Attr;\n}\n\ndeclare var NamedNodeMap: {\n    prototype: NamedNodeMap;\n    new(): NamedNodeMap;\n};\n\ninterface NavigationPreloadManager {\n    disable(): Promise<void>;\n    enable(): Promise<void>;\n    getState(): Promise<NavigationPreloadState>;\n    setHeaderValue(value: string): Promise<void>;\n}\n\ndeclare var NavigationPreloadManager: {\n    prototype: NavigationPreloadManager;\n    new(): NavigationPreloadManager;\n};\n\n/** The state and the identity of the user agent. It allows scripts to query it and to register themselves to carry on some activities. */\ninterface Navigator extends MSFileSaver, MSNavigatorDoNotTrack, NavigatorAutomationInformation, NavigatorBeacon, NavigatorConcurrentHardware, NavigatorContentUtils, NavigatorCookies, NavigatorID, NavigatorLanguage, NavigatorOnLine, NavigatorPlugins, NavigatorStorage {\n    readonly activeVRDisplays: ReadonlyArray<VRDisplay>;\n    readonly clipboard: Clipboard;\n    readonly credentials: CredentialsContainer;\n    readonly doNotTrack: string | null;\n    readonly geolocation: Geolocation;\n    readonly maxTouchPoints: number;\n    readonly mediaDevices: MediaDevices;\n    readonly msManipulationViewsEnabled: boolean;\n    readonly msMaxTouchPoints: number;\n    readonly msPointerEnabled: boolean;\n    readonly permissions: Permissions;\n    readonly pointerEnabled: boolean;\n    readonly serviceWorker: ServiceWorkerContainer;\n    getGamepads(): (Gamepad | null)[];\n    getUserMedia(constraints: MediaStreamConstraints, successCallback: NavigatorUserMediaSuccessCallback, errorCallback: NavigatorUserMediaErrorCallback): void;\n    getVRDisplays(): Promise<VRDisplay[]>;\n    msLaunchUri(uri: string, successCallback?: MSLaunchUriCallback, noHandlerCallback?: MSLaunchUriCallback): void;\n    requestMediaKeySystemAccess(keySystem: string, supportedConfigurations: MediaKeySystemConfiguration[]): Promise<MediaKeySystemAccess>;\n    sendBeacon(url: string, data?: BodyInit | null): boolean;\n    share(data?: ShareData): Promise<void>;\n    vibrate(pattern: number | number[]): boolean;\n}\n\ndeclare var Navigator: {\n    prototype: Navigator;\n    new(): Navigator;\n};\n\ninterface NavigatorAutomationInformation {\n    readonly webdriver: boolean;\n}\n\ninterface NavigatorBeacon {\n    sendBeacon(url: string, data?: Blob | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | FormData | string | null): boolean;\n}\n\ninterface NavigatorConcurrentHardware {\n    readonly hardwareConcurrency: number;\n}\n\ninterface NavigatorContentUtils {\n    registerProtocolHandler(scheme: string, url: string, title: string): void;\n    unregisterProtocolHandler(scheme: string, url: string): void;\n}\n\ninterface NavigatorCookies {\n    readonly cookieEnabled: boolean;\n}\n\ninterface NavigatorID {\n    readonly appCodeName: string;\n    readonly appName: string;\n    readonly appVersion: string;\n    readonly platform: string;\n    readonly product: string;\n    readonly productSub: string;\n    readonly userAgent: string;\n    readonly vendor: string;\n    readonly vendorSub: string;\n}\n\ninterface NavigatorLanguage {\n    readonly language: string;\n    readonly languages: ReadonlyArray<string>;\n}\n\ninterface NavigatorOnLine {\n    readonly onLine: boolean;\n}\n\ninterface NavigatorPlugins {\n    readonly mimeTypes: MimeTypeArray;\n    readonly plugins: PluginArray;\n    javaEnabled(): boolean;\n}\n\ninterface NavigatorStorage {\n    readonly storage: StorageManager;\n}\n\n/** Node is an interface from which a number of DOM API object types inherit. It allows those types to be treated similarly; for example, inheriting the same set of methods, or being tested in the same way. */\ninterface Node extends EventTarget {\n    /**\n     * Returns node\'s node document\'s document base URL.\n     */\n    readonly baseURI: string;\n    /**\n     * Returns the children.\n     */\n    readonly childNodes: NodeListOf<ChildNode>;\n    /**\n     * Returns the first child.\n     */\n    readonly firstChild: ChildNode | null;\n    /**\n     * Returns true if node is connected and false otherwise.\n     */\n    readonly isConnected: boolean;\n    /**\n     * Returns the last child.\n     */\n    readonly lastChild: ChildNode | null;\n    /** @deprecated */\n    readonly namespaceURI: string | null;\n    /**\n     * Returns the next sibling.\n     */\n    readonly nextSibling: ChildNode | null;\n    /**\n     * Returns a string appropriate for the type of node.\n     */\n    readonly nodeName: string;\n    /**\n     * Returns the type of node.\n     */\n    readonly nodeType: number;\n    nodeValue: string | null;\n    /**\n     * Returns the node document. Returns null for documents.\n     */\n    readonly ownerDocument: Document | null;\n    /**\n     * Returns the parent element.\n     */\n    readonly parentElement: HTMLElement | null;\n    /**\n     * Returns the parent.\n     */\n    readonly parentNode: Node & ParentNode | null;\n    /**\n     * Returns the previous sibling.\n     */\n    readonly previousSibling: ChildNode | null;\n    textContent: string | null;\n    appendChild<T extends Node>(newChild: T): T;\n    /**\n     * Returns a copy of node. If deep is true, the copy also includes the node\'s descendants.\n     */\n    cloneNode(deep?: boolean): Node;\n    /**\n     * Returns a bitmask indicating the position of other relative to node.\n     */\n    compareDocumentPosition(other: Node): number;\n    /**\n     * Returns true if other is an inclusive descendant of node, and false otherwise.\n     */\n    contains(other: Node | null): boolean;\n    /**\n     * Returns node\'s root.\n     */\n    getRootNode(options?: GetRootNodeOptions): Node;\n    /**\n     * Returns whether node has children.\n     */\n    hasChildNodes(): boolean;\n    insertBefore<T extends Node>(newChild: T, refChild: Node | null): T;\n    isDefaultNamespace(namespace: string | null): boolean;\n    /**\n     * Returns whether node and otherNode have the same properties.\n     */\n    isEqualNode(otherNode: Node | null): boolean;\n    isSameNode(otherNode: Node | null): boolean;\n    lookupNamespaceURI(prefix: string | null): string | null;\n    lookupPrefix(namespace: string | null): string | null;\n    /**\n     * Removes empty exclusive Text nodes and concatenates the data of remaining contiguous exclusive Text nodes into the first of their nodes.\n     */\n    normalize(): void;\n    removeChild<T extends Node>(oldChild: T): T;\n    replaceChild<T extends Node>(newChild: Node, oldChild: T): T;\n    readonly ATTRIBUTE_NODE: number;\n    /**\n     * node is a CDATASection node.\n     */\n    readonly CDATA_SECTION_NODE: number;\n    /**\n     * node is a Comment node.\n     */\n    readonly COMMENT_NODE: number;\n    /**\n     * node is a DocumentFragment node.\n     */\n    readonly DOCUMENT_FRAGMENT_NODE: number;\n    /**\n     * node is a document.\n     */\n    readonly DOCUMENT_NODE: number;\n    /**\n     * Set when other is a descendant of node.\n     */\n    readonly DOCUMENT_POSITION_CONTAINED_BY: number;\n    /**\n     * Set when other is an ancestor of node.\n     */\n    readonly DOCUMENT_POSITION_CONTAINS: number;\n    /**\n     * Set when node and other are not in the same tree.\n     */\n    readonly DOCUMENT_POSITION_DISCONNECTED: number;\n    /**\n     * Set when other is following node.\n     */\n    readonly DOCUMENT_POSITION_FOLLOWING: number;\n    readonly DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC: number;\n    /**\n     * Set when other is preceding node.\n     */\n    readonly DOCUMENT_POSITION_PRECEDING: number;\n    /**\n     * node is a doctype.\n     */\n    readonly DOCUMENT_TYPE_NODE: number;\n    /**\n     * node is an element.\n     */\n    readonly ELEMENT_NODE: number;\n    readonly ENTITY_NODE: number;\n    readonly ENTITY_REFERENCE_NODE: number;\n    readonly NOTATION_NODE: number;\n    /**\n     * node is a ProcessingInstruction node.\n     */\n    readonly PROCESSING_INSTRUCTION_NODE: number;\n    /**\n     * node is a Text node.\n     */\n    readonly TEXT_NODE: number;\n}\n\ndeclare var Node: {\n    prototype: Node;\n    new(): Node;\n    readonly ATTRIBUTE_NODE: number;\n    /**\n     * node is a CDATASection node.\n     */\n    readonly CDATA_SECTION_NODE: number;\n    /**\n     * node is a Comment node.\n     */\n    readonly COMMENT_NODE: number;\n    /**\n     * node is a DocumentFragment node.\n     */\n    readonly DOCUMENT_FRAGMENT_NODE: number;\n    /**\n     * node is a document.\n     */\n    readonly DOCUMENT_NODE: number;\n    /**\n     * Set when other is a descendant of node.\n     */\n    readonly DOCUMENT_POSITION_CONTAINED_BY: number;\n    /**\n     * Set when other is an ancestor of node.\n     */\n    readonly DOCUMENT_POSITION_CONTAINS: number;\n    /**\n     * Set when node and other are not in the same tree.\n     */\n    readonly DOCUMENT_POSITION_DISCONNECTED: number;\n    /**\n     * Set when other is following node.\n     */\n    readonly DOCUMENT_POSITION_FOLLOWING: number;\n    readonly DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC: number;\n    /**\n     * Set when other is preceding node.\n     */\n    readonly DOCUMENT_POSITION_PRECEDING: number;\n    /**\n     * node is a doctype.\n     */\n    readonly DOCUMENT_TYPE_NODE: number;\n    /**\n     * node is an element.\n     */\n    readonly ELEMENT_NODE: number;\n    readonly ENTITY_NODE: number;\n    readonly ENTITY_REFERENCE_NODE: number;\n    readonly NOTATION_NODE: number;\n    /**\n     * node is a ProcessingInstruction node.\n     */\n    readonly PROCESSING_INSTRUCTION_NODE: number;\n    /**\n     * node is a Text node.\n     */\n    readonly TEXT_NODE: number;\n};\n\n/** An object used to filter the nodes in a NodeIterator or TreeWalker. They don\'t know anything about the DOM or how to traverse nodes; they just know how to evaluate a single node against the provided filter. */\ninterface NodeFilter {\n    acceptNode(node: Node): number;\n}\n\ndeclare var NodeFilter: {\n    readonly FILTER_ACCEPT: number;\n    readonly FILTER_REJECT: number;\n    readonly FILTER_SKIP: number;\n    readonly SHOW_ALL: number;\n    readonly SHOW_ATTRIBUTE: number;\n    readonly SHOW_CDATA_SECTION: number;\n    readonly SHOW_COMMENT: number;\n    readonly SHOW_DOCUMENT: number;\n    readonly SHOW_DOCUMENT_FRAGMENT: number;\n    readonly SHOW_DOCUMENT_TYPE: number;\n    readonly SHOW_ELEMENT: number;\n    readonly SHOW_ENTITY: number;\n    readonly SHOW_ENTITY_REFERENCE: number;\n    readonly SHOW_NOTATION: number;\n    readonly SHOW_PROCESSING_INSTRUCTION: number;\n    readonly SHOW_TEXT: number;\n};\n\n/** An iterator over the members of a list of the nodes in a subtree of the DOM. The nodes will be returned in document order. */\ninterface NodeIterator {\n    readonly filter: NodeFilter | null;\n    readonly pointerBeforeReferenceNode: boolean;\n    readonly referenceNode: Node;\n    readonly root: Node;\n    readonly whatToShow: number;\n    detach(): void;\n    nextNode(): Node | null;\n    previousNode(): Node | null;\n}\n\ndeclare var NodeIterator: {\n    prototype: NodeIterator;\n    new(): NodeIterator;\n};\n\n/** NodeList objects are collections of nodes, usually returned by properties such as Node.childNodes and methods such as document.querySelectorAll(). */\ninterface NodeList {\n    /**\n     * Returns the number of nodes in the collection.\n     */\n    readonly length: number;\n    /**\n     * Returns the node with index index from the collection. The nodes are sorted in tree order.\n     */\n    item(index: number): Node | null;\n    /**\n     * Performs the specified action for each node in an list.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the list.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: Node, key: number, parent: NodeList) => void, thisArg?: any): void;\n    [index: number]: Node;\n}\n\ndeclare var NodeList: {\n    prototype: NodeList;\n    new(): NodeList;\n};\n\ninterface NodeListOf<TNode extends Node> extends NodeList {\n    length: number;\n    item(index: number): TNode;\n    /**\n     * Performs the specified action for each node in an list.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the list.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: TNode, key: number, parent: NodeListOf<TNode>) => void, thisArg?: any): void;\n    [index: number]: TNode;\n}\n\ninterface NonDocumentTypeChildNode {\n    /**\n     * Returns the first following sibling that is an element, and null otherwise.\n     */\n    readonly nextElementSibling: Element | null;\n    /**\n     * Returns the first preceding sibling that is an element, and null otherwise.\n     */\n    readonly previousElementSibling: Element | null;\n}\n\ninterface NonElementParentNode {\n    /**\n     * Returns the first element within node\'s descendants whose ID is elementId.\n     */\n    getElementById(elementId: string): Element | null;\n}\n\ninterface NotificationEventMap {\n    "click": Event;\n    "close": Event;\n    "error": Event;\n    "show": Event;\n}\n\n/** This Notifications API interface is used to configure and display desktop notifications to the user. */\ninterface Notification extends EventTarget {\n    readonly actions: ReadonlyArray<NotificationAction>;\n    readonly badge: string;\n    readonly body: string;\n    readonly data: any;\n    readonly dir: NotificationDirection;\n    readonly icon: string;\n    readonly image: string;\n    readonly lang: string;\n    onclick: ((this: Notification, ev: Event) => any) | null;\n    onclose: ((this: Notification, ev: Event) => any) | null;\n    onerror: ((this: Notification, ev: Event) => any) | null;\n    onshow: ((this: Notification, ev: Event) => any) | null;\n    readonly renotify: boolean;\n    readonly requireInteraction: boolean;\n    readonly silent: boolean;\n    readonly tag: string;\n    readonly timestamp: number;\n    readonly title: string;\n    readonly vibrate: ReadonlyArray<number>;\n    close(): void;\n    addEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Notification: {\n    prototype: Notification;\n    new(title: string, options?: NotificationOptions): Notification;\n    readonly maxActions: number;\n    readonly permission: NotificationPermission;\n    requestPermission(deprecatedCallback?: NotificationPermissionCallback): Promise<NotificationPermission>;\n};\n\n/** The OES_element_index_uint extension is part of the WebGL API and adds support for gl.UNSIGNED_INT types to WebGLRenderingContext.drawElements(). */\ninterface OES_element_index_uint {\n}\n\n/** The OES_standard_derivatives extension is part of the WebGL API and adds the GLSL derivative functions dFdx, dFdy, and fwidth. */\ninterface OES_standard_derivatives {\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT_OES: GLenum;\n}\n\n/** The OES_texture_float extension is part of the WebGL API and exposes floating-point pixel types for textures. */\ninterface OES_texture_float {\n}\n\n/** The OES_texture_float_linear extension is part of the WebGL API and allows linear filtering with floating-point pixel types for textures. */\ninterface OES_texture_float_linear {\n}\n\n/** The OES_texture_half_float extension is part of the WebGL API and adds texture formats with 16- (aka half float) and 32-bit floating-point components. */\ninterface OES_texture_half_float {\n    readonly HALF_FLOAT_OES: GLenum;\n}\n\n/** The OES_texture_half_float_linear extension is part of the WebGL API and allows linear filtering with half floating-point pixel types for textures. */\ninterface OES_texture_half_float_linear {\n}\n\ninterface OES_vertex_array_object {\n    bindVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    createVertexArrayOES(): WebGLVertexArrayObjectOES | null;\n    deleteVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    isVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): GLboolean;\n    readonly VERTEX_ARRAY_BINDING_OES: GLenum;\n}\n\n/** The Web Audio API OfflineAudioCompletionEvent interface represents events that occur when the processing of an OfflineAudioContext is terminated. The complete event implements this interface. */\ninterface OfflineAudioCompletionEvent extends Event {\n    readonly renderedBuffer: AudioBuffer;\n}\n\ndeclare var OfflineAudioCompletionEvent: {\n    prototype: OfflineAudioCompletionEvent;\n    new(type: string, eventInitDict: OfflineAudioCompletionEventInit): OfflineAudioCompletionEvent;\n};\n\ninterface OfflineAudioContextEventMap extends BaseAudioContextEventMap {\n    "complete": OfflineAudioCompletionEvent;\n}\n\n/** An AudioContext interface representing an audio-processing graph built from linked together AudioNodes. In contrast with a standard AudioContext, an OfflineAudioContext doesn\'t render the audio to the device hardware; instead, it generates it, as fast as it can, and outputs the result to an AudioBuffer. */\ninterface OfflineAudioContext extends BaseAudioContext {\n    readonly length: number;\n    oncomplete: ((this: OfflineAudioContext, ev: OfflineAudioCompletionEvent) => any) | null;\n    resume(): Promise<void>;\n    startRendering(): Promise<AudioBuffer>;\n    suspend(suspendTime: number): Promise<void>;\n    addEventListener<K extends keyof OfflineAudioContextEventMap>(type: K, listener: (this: OfflineAudioContext, ev: OfflineAudioContextEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof OfflineAudioContextEventMap>(type: K, listener: (this: OfflineAudioContext, ev: OfflineAudioContextEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var OfflineAudioContext: {\n    prototype: OfflineAudioContext;\n    new(contextOptions: OfflineAudioContextOptions): OfflineAudioContext;\n    new(numberOfChannels: number, length: number, sampleRate: number): OfflineAudioContext;\n};\n\ninterface OffscreenCanvas extends EventTarget {\n    /**\n     * These attributes return the dimensions of the OffscreenCanvas object\'s bitmap.\n     * \n     * They can be set, to replace the bitmap with a new, transparent black bitmap of the specified dimensions (effectively resizing it).\n     */\n    height: number;\n    /**\n     * These attributes return the dimensions of the OffscreenCanvas object\'s bitmap.\n     * \n     * They can be set, to replace the bitmap with a new, transparent black bitmap of the specified dimensions (effectively resizing it).\n     */\n    width: number;\n    /**\n     * Returns a promise that will fulfill with a new Blob object representing a file containing the image in the OffscreenCanvas object.\n     * \n     * The argument, if provided, is a dictionary that controls the encoding options of the image file to be created. The type field specifies the file format and has a default value of "image/png"; that type is also used if the requested type isn\'t supported. If the image format supports variable quality (such as "image/jpeg"), then the quality field is a number in the range 0.0 to 1.0 inclusive indicating the desired quality level for the resulting image.\n     */\n    convertToBlob(options?: ImageEncodeOptions): Promise<Blob>;\n    /**\n     * Returns an object that exposes an API for drawing on the OffscreenCanvas object. contextId specifies the desired API: "2d", "bitmaprenderer", "webgl", or "webgl2". options is handled by that API.\n     * \n     * This specification defines the "2d" context below, which is similar but distinct from the "2d" context that is created from a canvas element. The WebGL specifications define the "webgl" and "webgl2" contexts. [WEBGL]\n     * \n     * Returns null if the canvas has already been initialized with another context type (e.g., trying to get a "2d" context after getting a "webgl" context).\n     */\n    getContext(contextId: "2d", options?: CanvasRenderingContext2DSettings): OffscreenCanvasRenderingContext2D | null;\n    getContext(contextId: "bitmaprenderer", options?: ImageBitmapRenderingContextSettings): ImageBitmapRenderingContext | null;\n    getContext(contextId: "webgl", options?: WebGLContextAttributes): WebGLRenderingContext | null;\n    getContext(contextId: "webgl2", options?: WebGLContextAttributes): WebGL2RenderingContext | null;\n    getContext(contextId: OffscreenRenderingContextId, options?: any): OffscreenRenderingContext | null;\n    /**\n     * Returns a newly created ImageBitmap object with the image in the OffscreenCanvas object. The image in the OffscreenCanvas object is replaced with a new blank image.\n     */\n    transferToImageBitmap(): ImageBitmap;\n}\n\ndeclare var OffscreenCanvas: {\n    prototype: OffscreenCanvas;\n    new(width: number, height: number): OffscreenCanvas;\n};\n\ninterface OffscreenCanvasRenderingContext2D extends CanvasCompositing, CanvasDrawImage, CanvasDrawPath, CanvasFillStrokeStyles, CanvasFilters, CanvasImageData, CanvasImageSmoothing, CanvasPath, CanvasPathDrawingStyles, CanvasRect, CanvasShadowStyles, CanvasState, CanvasText, CanvasTextDrawingStyles, CanvasTransform {\n    readonly canvas: OffscreenCanvas;\n    commit(): void;\n}\n\ndeclare var OffscreenCanvasRenderingContext2D: {\n    prototype: OffscreenCanvasRenderingContext2D;\n    new(): OffscreenCanvasRenderingContext2D;\n};\n\n/** The OscillatorNode interface represents a periodic waveform, such as a sine wave. It is an AudioScheduledSourceNode audio-processing module that causes a specified frequency of a given wave to be created—in effect, a constant tone. */\ninterface OscillatorNode extends AudioScheduledSourceNode {\n    readonly detune: AudioParam;\n    readonly frequency: AudioParam;\n    type: OscillatorType;\n    setPeriodicWave(periodicWave: PeriodicWave): void;\n    addEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: OscillatorNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AudioScheduledSourceNodeEventMap>(type: K, listener: (this: OscillatorNode, ev: AudioScheduledSourceNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var OscillatorNode: {\n    prototype: OscillatorNode;\n    new(context: BaseAudioContext, options?: OscillatorOptions): OscillatorNode;\n};\n\ninterface OverconstrainedError extends Error {\n    constraint: string;\n}\n\ndeclare var OverconstrainedError: {\n    prototype: OverconstrainedError;\n    new(): OverconstrainedError;\n};\n\ninterface OverflowEvent extends UIEvent {\n    readonly horizontalOverflow: boolean;\n    readonly orient: number;\n    readonly verticalOverflow: boolean;\n    readonly BOTH: number;\n    readonly HORIZONTAL: number;\n    readonly VERTICAL: number;\n}\n\ndeclare var OverflowEvent: {\n    prototype: OverflowEvent;\n    new(): OverflowEvent;\n    readonly BOTH: number;\n    readonly HORIZONTAL: number;\n    readonly VERTICAL: number;\n};\n\n/** The PageTransitionEvent is fired when a document is being loaded or unloaded. */\ninterface PageTransitionEvent extends Event {\n    /**\n     * For the pageshow event, returns false if the page is newly being loaded (and the load event will fire). Otherwise, returns true.\n     * \n     * For the pagehide event, returns false if the page is going away for the last time. Otherwise, returns true, meaning that (if nothing conspires to make the page unsalvageable) the page might be reused if the user navigates back to this page.\n     * \n     * Things that can cause the page to be unsalvageable include:\n     * \n     * Listening for beforeunload events\n     * Listening for unload events\n     * Having iframes that are not salvageable\n     * Active WebSocket objects\n     * Aborting a Document\n     */\n    readonly persisted: boolean;\n}\n\ndeclare var PageTransitionEvent: {\n    prototype: PageTransitionEvent;\n    new(type: string, eventInitDict?: PageTransitionEventInit): PageTransitionEvent;\n};\n\n/** A PannerNode always has exactly one input and one output: the input can be mono or stereo but the output is always stereo (2 channels); you can\'t have panning effects without at least two audio channels! */\ninterface PannerNode extends AudioNode {\n    coneInnerAngle: number;\n    coneOuterAngle: number;\n    coneOuterGain: number;\n    distanceModel: DistanceModelType;\n    maxDistance: number;\n    readonly orientationX: AudioParam;\n    readonly orientationY: AudioParam;\n    readonly orientationZ: AudioParam;\n    panningModel: PanningModelType;\n    readonly positionX: AudioParam;\n    readonly positionY: AudioParam;\n    readonly positionZ: AudioParam;\n    refDistance: number;\n    rolloffFactor: number;\n    /** @deprecated */\n    setOrientation(x: number, y: number, z: number): void;\n    /** @deprecated */\n    setPosition(x: number, y: number, z: number): void;\n}\n\ndeclare var PannerNode: {\n    prototype: PannerNode;\n    new(context: BaseAudioContext, options?: PannerOptions): PannerNode;\n};\n\ninterface ParentNode {\n    readonly childElementCount: number;\n    /**\n     * Returns the child elements.\n     */\n    readonly children: HTMLCollection;\n    /**\n     * Returns the first child that is an element, and null otherwise.\n     */\n    readonly firstElementChild: Element | null;\n    /**\n     * Returns the last child that is an element, and null otherwise.\n     */\n    readonly lastElementChild: Element | null;\n    /**\n     * Inserts nodes after the last child of node, while replacing strings in nodes with equivalent Text nodes.\n     * \n     * Throws a "HierarchyRequestError" DOMException if the constraints of the node tree are violated.\n     */\n    append(...nodes: (Node | string)[]): void;\n    /**\n     * Inserts nodes before the first child of node, while replacing strings in nodes with equivalent Text nodes.\n     * \n     * Throws a "HierarchyRequestError" DOMException if the constraints of the node tree are violated.\n     */\n    prepend(...nodes: (Node | string)[]): void;\n    /**\n     * Returns the first element that is a descendant of node that matches selectors.\n     */\n    querySelector<K extends keyof HTMLElementTagNameMap>(selectors: K): HTMLElementTagNameMap[K] | null;\n    querySelector<K extends keyof SVGElementTagNameMap>(selectors: K): SVGElementTagNameMap[K] | null;\n    querySelector<E extends Element = Element>(selectors: string): E | null;\n    /**\n     * Returns all element descendants of node that match selectors.\n     */\n    querySelectorAll<K extends keyof HTMLElementTagNameMap>(selectors: K): NodeListOf<HTMLElementTagNameMap[K]>;\n    querySelectorAll<K extends keyof SVGElementTagNameMap>(selectors: K): NodeListOf<SVGElementTagNameMap[K]>;\n    querySelectorAll<E extends Element = Element>(selectors: string): NodeListOf<E>;\n}\n\n/** This Canvas 2D API interface is used to declare a path that can then be used on a CanvasRenderingContext2D object. The path methods of the CanvasRenderingContext2D interface are also present on this interface, which gives you the convenience of being able to retain and replay your path whenever desired. */\ninterface Path2D extends CanvasPath {\n    /**\n     * Adds to the path the path given by the argument.\n     */\n    addPath(path: Path2D, transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var Path2D: {\n    prototype: Path2D;\n    new(path?: Path2D | string): Path2D;\n};\n\n/** This Payment Request API interface is used to store shipping or payment address information. */\ninterface PaymentAddress {\n    readonly addressLine: string[];\n    readonly city: string;\n    readonly country: string;\n    readonly dependentLocality: string;\n    readonly languageCode: string;\n    readonly organization: string;\n    readonly phone: string;\n    readonly postalCode: string;\n    readonly recipient: string;\n    readonly region: string;\n    readonly sortingCode: string;\n    toJSON(): any;\n}\n\ndeclare var PaymentAddress: {\n    prototype: PaymentAddress;\n    new(): PaymentAddress;\n};\n\ninterface PaymentRequestEventMap {\n    "shippingaddresschange": Event;\n    "shippingoptionchange": Event;\n}\n\n/** This Payment Request API interface is the primary access point into the API, and lets web content and apps accept payments from the end user. */\ninterface PaymentRequest extends EventTarget {\n    readonly id: string;\n    onshippingaddresschange: ((this: PaymentRequest, ev: Event) => any) | null;\n    onshippingoptionchange: ((this: PaymentRequest, ev: Event) => any) | null;\n    readonly shippingAddress: PaymentAddress | null;\n    readonly shippingOption: string | null;\n    readonly shippingType: PaymentShippingType | null;\n    abort(): Promise<void>;\n    canMakePayment(): Promise<boolean>;\n    show(): Promise<PaymentResponse>;\n    addEventListener<K extends keyof PaymentRequestEventMap>(type: K, listener: (this: PaymentRequest, ev: PaymentRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PaymentRequestEventMap>(type: K, listener: (this: PaymentRequest, ev: PaymentRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var PaymentRequest: {\n    prototype: PaymentRequest;\n    new(methodData: PaymentMethodData[], details: PaymentDetailsInit, options?: PaymentOptions): PaymentRequest;\n};\n\n/** This Payment Request API interface enables a web page to update the details of a PaymentRequest in response to a user action. */\ninterface PaymentRequestUpdateEvent extends Event {\n    updateWith(detailsPromise: PaymentDetailsUpdate | Promise<PaymentDetailsUpdate>): void;\n}\n\ndeclare var PaymentRequestUpdateEvent: {\n    prototype: PaymentRequestUpdateEvent;\n    new(type: string, eventInitDict?: PaymentRequestUpdateEventInit): PaymentRequestUpdateEvent;\n};\n\n/** This Payment Request API interface is returned after a user selects a payment method and approves a payment request. */\ninterface PaymentResponse {\n    readonly details: any;\n    readonly methodName: string;\n    readonly payerEmail: string | null;\n    readonly payerName: string | null;\n    readonly payerPhone: string | null;\n    readonly requestId: string;\n    readonly shippingAddress: PaymentAddress | null;\n    readonly shippingOption: string | null;\n    complete(result?: PaymentComplete): Promise<void>;\n    toJSON(): any;\n}\n\ndeclare var PaymentResponse: {\n    prototype: PaymentResponse;\n    new(): PaymentResponse;\n};\n\ninterface PerfWidgetExternal {\n    readonly activeNetworkRequestCount: number;\n    readonly averageFrameTime: number;\n    readonly averagePaintTime: number;\n    readonly extraInformationEnabled: boolean;\n    readonly independentRenderingEnabled: boolean;\n    readonly irDisablingContentString: string;\n    readonly irStatusAvailable: boolean;\n    readonly maxCpuSpeed: number;\n    readonly paintRequestsPerSecond: number;\n    readonly performanceCounter: number;\n    readonly performanceCounterFrequency: number;\n    addEventListener(eventType: string, callback: Function): void;\n    getMemoryUsage(): number;\n    getProcessCpuUsage(): number;\n    getRecentCpuUsage(last: number | null): any;\n    getRecentFrames(last: number | null): any;\n    getRecentMemoryUsage(last: number | null): any;\n    getRecentPaintRequests(last: number | null): any;\n    removeEventListener(eventType: string, callback: Function): void;\n    repositionWindow(x: number, y: number): void;\n    resizeWindow(width: number, height: number): void;\n}\n\ndeclare var PerfWidgetExternal: {\n    prototype: PerfWidgetExternal;\n    new(): PerfWidgetExternal;\n};\n\ninterface PerformanceEventMap {\n    "resourcetimingbufferfull": Event;\n}\n\n/** Provides access to performance-related information for the current page. It\'s part of the High Resolution Time API, but is enhanced by the Performance Timeline API, the Navigation Timing API, the User Timing API, and the Resource Timing API. */\ninterface Performance extends EventTarget {\n    /** @deprecated */\n    readonly navigation: PerformanceNavigation;\n    onresourcetimingbufferfull: ((this: Performance, ev: Event) => any) | null;\n    readonly timeOrigin: number;\n    /** @deprecated */\n    readonly timing: PerformanceTiming;\n    clearMarks(markName?: string): void;\n    clearMeasures(measureName?: string): void;\n    clearResourceTimings(): void;\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n    mark(markName: string): void;\n    measure(measureName: string, startMark?: string, endMark?: string): void;\n    now(): number;\n    setResourceTimingBufferSize(maxSize: number): void;\n    toJSON(): any;\n    addEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Performance: {\n    prototype: Performance;\n    new(): Performance;\n};\n\n/** Encapsulates a single performance metric that is part of the performance timeline. A performance entry can be directly created by making a performance mark or measure (for example by calling the mark() method) at an explicit point in an application. Performance entries are also created in indirect ways such as loading a resource (such as an image). */\ninterface PerformanceEntry {\n    readonly duration: number;\n    readonly entryType: string;\n    readonly name: string;\n    readonly startTime: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceEntry: {\n    prototype: PerformanceEntry;\n    new(): PerformanceEntry;\n};\n\n/** PerformanceMark is an abstract interface for PerformanceEntry objects with an entryType of "mark". Entries of this type are created by calling performance.mark() to add a named DOMHighResTimeStamp (the mark) to the browser\'s performance timeline. */\ninterface PerformanceMark extends PerformanceEntry {\n}\n\ndeclare var PerformanceMark: {\n    prototype: PerformanceMark;\n    new(): PerformanceMark;\n};\n\n/** PerformanceMeasure is an abstract interface for PerformanceEntry objects with an entryType of "measure". Entries of this type are created by calling performance.measure() to add a named DOMHighResTimeStamp (the measure) between two marks to the browser\'s performance timeline. */\ninterface PerformanceMeasure extends PerformanceEntry {\n}\n\ndeclare var PerformanceMeasure: {\n    prototype: PerformanceMeasure;\n    new(): PerformanceMeasure;\n};\n\n/** The legacy PerformanceNavigation interface represents information about how the navigation to the current document was done. */\ninterface PerformanceNavigation {\n    readonly redirectCount: number;\n    readonly type: number;\n    toJSON(): any;\n    readonly TYPE_BACK_FORWARD: number;\n    readonly TYPE_NAVIGATE: number;\n    readonly TYPE_RELOAD: number;\n    readonly TYPE_RESERVED: number;\n}\n\ndeclare var PerformanceNavigation: {\n    prototype: PerformanceNavigation;\n    new(): PerformanceNavigation;\n    readonly TYPE_BACK_FORWARD: number;\n    readonly TYPE_NAVIGATE: number;\n    readonly TYPE_RELOAD: number;\n    readonly TYPE_RESERVED: number;\n};\n\n/** Provides methods and properties to store and retrieve metrics regarding the browser\'s document navigation events. For example, this interface can be used to determine how much time it takes to load or unload a document. */\ninterface PerformanceNavigationTiming extends PerformanceResourceTiming {\n    readonly domComplete: number;\n    readonly domContentLoadedEventEnd: number;\n    readonly domContentLoadedEventStart: number;\n    readonly domInteractive: number;\n    readonly loadEventEnd: number;\n    readonly loadEventStart: number;\n    readonly redirectCount: number;\n    readonly type: NavigationType;\n    readonly unloadEventEnd: number;\n    readonly unloadEventStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceNavigationTiming: {\n    prototype: PerformanceNavigationTiming;\n    new(): PerformanceNavigationTiming;\n};\n\ninterface PerformanceObserver {\n    disconnect(): void;\n    observe(options?: PerformanceObserverInit): void;\n    takeRecords(): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserver: {\n    prototype: PerformanceObserver;\n    new(callback: PerformanceObserverCallback): PerformanceObserver;\n    readonly supportedEntryTypes: ReadonlyArray<string>;\n};\n\ninterface PerformanceObserverEntryList {\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserverEntryList: {\n    prototype: PerformanceObserverEntryList;\n    new(): PerformanceObserverEntryList;\n};\n\n/** Enables retrieval and analysis of detailed network timing data regarding the loading of an application\'s resources. An application can use the timing metrics to determine, for example, the length of time it takes to fetch a specific resource, such as an XMLHttpRequest, <SVG>, image, or script. */\ninterface PerformanceResourceTiming extends PerformanceEntry {\n    readonly connectEnd: number;\n    readonly connectStart: number;\n    readonly decodedBodySize: number;\n    readonly domainLookupEnd: number;\n    readonly domainLookupStart: number;\n    readonly encodedBodySize: number;\n    readonly fetchStart: number;\n    readonly initiatorType: string;\n    readonly nextHopProtocol: string;\n    readonly redirectEnd: number;\n    readonly redirectStart: number;\n    readonly requestStart: number;\n    readonly responseEnd: number;\n    readonly responseStart: number;\n    readonly secureConnectionStart: number;\n    readonly transferSize: number;\n    readonly workerStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceResourceTiming: {\n    prototype: PerformanceResourceTiming;\n    new(): PerformanceResourceTiming;\n};\n\n/** A legacy interface kept for backwards compatibility and contains properties that offer performance timing information for various events which occur during the loading and use of the current page. You get a PerformanceTiming object describing your page using the window.performance.timing property. */\ninterface PerformanceTiming {\n    readonly connectEnd: number;\n    readonly connectStart: number;\n    readonly domComplete: number;\n    readonly domContentLoadedEventEnd: number;\n    readonly domContentLoadedEventStart: number;\n    readonly domInteractive: number;\n    readonly domLoading: number;\n    readonly domainLookupEnd: number;\n    readonly domainLookupStart: number;\n    readonly fetchStart: number;\n    readonly loadEventEnd: number;\n    readonly loadEventStart: number;\n    readonly navigationStart: number;\n    readonly redirectEnd: number;\n    readonly redirectStart: number;\n    readonly requestStart: number;\n    readonly responseEnd: number;\n    readonly responseStart: number;\n    readonly secureConnectionStart: number;\n    readonly unloadEventEnd: number;\n    readonly unloadEventStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceTiming: {\n    prototype: PerformanceTiming;\n    new(): PerformanceTiming;\n};\n\n/** PeriodicWave has no inputs or outputs; it is used to define custom oscillators when calling OscillatorNode.setPeriodicWave(). The PeriodicWave itself is created/returned by AudioContext.createPeriodicWave(). */\ninterface PeriodicWave {\n}\n\ndeclare var PeriodicWave: {\n    prototype: PeriodicWave;\n    new(context: BaseAudioContext, options?: PeriodicWaveOptions): PeriodicWave;\n};\n\ninterface PermissionRequest extends DeferredPermissionRequest {\n    readonly state: MSWebViewPermissionState;\n    defer(): void;\n}\n\ndeclare var PermissionRequest: {\n    prototype: PermissionRequest;\n    new(): PermissionRequest;\n};\n\ninterface PermissionRequestedEvent extends Event {\n    readonly permissionRequest: PermissionRequest;\n}\n\ndeclare var PermissionRequestedEvent: {\n    prototype: PermissionRequestedEvent;\n    new(): PermissionRequestedEvent;\n};\n\ninterface PermissionStatusEventMap {\n    "change": Event;\n}\n\ninterface PermissionStatus extends EventTarget {\n    onchange: ((this: PermissionStatus, ev: Event) => any) | null;\n    readonly state: PermissionState;\n    addEventListener<K extends keyof PermissionStatusEventMap>(type: K, listener: (this: PermissionStatus, ev: PermissionStatusEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PermissionStatusEventMap>(type: K, listener: (this: PermissionStatus, ev: PermissionStatusEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var PermissionStatus: {\n    prototype: PermissionStatus;\n    new(): PermissionStatus;\n};\n\ninterface Permissions {\n    query(permissionDesc: PermissionDescriptor | DevicePermissionDescriptor | MidiPermissionDescriptor | PushPermissionDescriptor): Promise<PermissionStatus>;\n}\n\ndeclare var Permissions: {\n    prototype: Permissions;\n    new(): Permissions;\n};\n\n/** Provides information about a browser plugin. */\ninterface Plugin {\n    /**\n     * Returns the plugin\'s description.\n     */\n    readonly description: string;\n    /**\n     * Returns the plugin library\'s filename, if applicable on the current platform.\n     */\n    readonly filename: string;\n    /**\n     * Returns the number of MIME types, represented by MimeType objects, supported by the plugin.\n     */\n    readonly length: number;\n    /**\n     * Returns the plugin\'s name.\n     */\n    readonly name: string;\n    /**\n     * Returns the specified MimeType object.\n     */\n    item(index: number): MimeType | null;\n    namedItem(name: string): MimeType | null;\n    [index: number]: MimeType;\n}\n\ndeclare var Plugin: {\n    prototype: Plugin;\n    new(): Plugin;\n};\n\n/** Used to store a list of Plugin objects describing the available plugins; it\'s returned by the window.navigator.plugins property. The PluginArray is not a JavaScript array, but has the length property and supports accessing individual items using bracket notation (plugins[2]), as well as via item(index) and namedItem("name") methods. */\ninterface PluginArray {\n    readonly length: number;\n    item(index: number): Plugin | null;\n    namedItem(name: string): Plugin | null;\n    refresh(reload?: boolean): void;\n    [index: number]: Plugin;\n}\n\ndeclare var PluginArray: {\n    prototype: PluginArray;\n    new(): PluginArray;\n};\n\n/** The state of a DOM event produced by a pointer such as the geometry of the contact point, the device type that generated the event, the amount of pressure that was applied on the contact surface, etc. */\ninterface PointerEvent extends MouseEvent {\n    readonly height: number;\n    readonly isPrimary: boolean;\n    readonly pointerId: number;\n    readonly pointerType: string;\n    readonly pressure: number;\n    readonly tangentialPressure: number;\n    readonly tiltX: number;\n    readonly tiltY: number;\n    readonly twist: number;\n    readonly width: number;\n}\n\ndeclare var PointerEvent: {\n    prototype: PointerEvent;\n    new(type: string, eventInitDict?: PointerEventInit): PointerEvent;\n};\n\n/** PopStateEvent is an event handler for the popstate event on the window. */\ninterface PopStateEvent extends Event {\n    /**\n     * Returns a copy of the information that was provided to pushState() or replaceState().\n     */\n    readonly state: any;\n}\n\ndeclare var PopStateEvent: {\n    prototype: PopStateEvent;\n    new(type: string, eventInitDict?: PopStateEventInit): PopStateEvent;\n};\n\n/** A processing instruction embeds application-specific instructions in XML which can be ignored by other applications that don\'t recognize them. */\ninterface ProcessingInstruction extends CharacterData, LinkStyle {\n    readonly ownerDocument: Document;\n    readonly target: string;\n}\n\ndeclare var ProcessingInstruction: {\n    prototype: ProcessingInstruction;\n    new(): ProcessingInstruction;\n};\n\n/** Events measuring progress of an underlying process, like an HTTP request (for an XMLHttpRequest, or the loading of the underlying resource of an <img>, <audio>, <video>, <style> or <link>). */\ninterface ProgressEvent<T extends EventTarget = EventTarget> extends Event {\n    readonly lengthComputable: boolean;\n    readonly loaded: number;\n    readonly target: T | null;\n    readonly total: number;\n}\n\ndeclare var ProgressEvent: {\n    prototype: ProgressEvent;\n    new(type: string, eventInitDict?: ProgressEventInit): ProgressEvent;\n};\n\ninterface PromiseRejectionEvent extends Event {\n    readonly promise: Promise<any>;\n    readonly reason: any;\n}\n\ndeclare var PromiseRejectionEvent: {\n    prototype: PromiseRejectionEvent;\n    new(type: string, eventInitDict: PromiseRejectionEventInit): PromiseRejectionEvent;\n};\n\ninterface PublicKeyCredential extends Credential {\n    readonly rawId: ArrayBuffer;\n    readonly response: AuthenticatorResponse;\n    getClientExtensionResults(): AuthenticationExtensionsClientOutputs;\n}\n\ndeclare var PublicKeyCredential: {\n    prototype: PublicKeyCredential;\n    new(): PublicKeyCredential;\n    isUserVerifyingPlatformAuthenticatorAvailable(): Promise<boolean>;\n};\n\n/** This Push API interface provides a way to receive notifications from third-party servers as well as request URLs for push notifications. */\ninterface PushManager {\n    getSubscription(): Promise<PushSubscription | null>;\n    permissionState(options?: PushSubscriptionOptionsInit): Promise<PushPermissionState>;\n    subscribe(options?: PushSubscriptionOptionsInit): Promise<PushSubscription>;\n}\n\ndeclare var PushManager: {\n    prototype: PushManager;\n    new(): PushManager;\n    readonly supportedContentEncodings: ReadonlyArray<string>;\n};\n\n/** This Push API interface provides a subcription\'s URL endpoint and allows unsubscription from a push service. */\ninterface PushSubscription {\n    readonly endpoint: string;\n    readonly expirationTime: number | null;\n    readonly options: PushSubscriptionOptions;\n    getKey(name: PushEncryptionKeyName): ArrayBuffer | null;\n    toJSON(): PushSubscriptionJSON;\n    unsubscribe(): Promise<boolean>;\n}\n\ndeclare var PushSubscription: {\n    prototype: PushSubscription;\n    new(): PushSubscription;\n};\n\ninterface PushSubscriptionOptions {\n    readonly applicationServerKey: ArrayBuffer | null;\n    readonly userVisibleOnly: boolean;\n}\n\ndeclare var PushSubscriptionOptions: {\n    prototype: PushSubscriptionOptions;\n    new(): PushSubscriptionOptions;\n};\n\ninterface RTCCertificate {\n    readonly expires: number;\n    getFingerprints(): RTCDtlsFingerprint[];\n}\n\ndeclare var RTCCertificate: {\n    prototype: RTCCertificate;\n    new(): RTCCertificate;\n    getSupportedAlgorithms(): AlgorithmIdentifier[];\n};\n\ninterface RTCDTMFSenderEventMap {\n    "tonechange": RTCDTMFToneChangeEvent;\n}\n\ninterface RTCDTMFSender extends EventTarget {\n    readonly canInsertDTMF: boolean;\n    ontonechange: ((this: RTCDTMFSender, ev: RTCDTMFToneChangeEvent) => any) | null;\n    readonly toneBuffer: string;\n    insertDTMF(tones: string, duration?: number, interToneGap?: number): void;\n    addEventListener<K extends keyof RTCDTMFSenderEventMap>(type: K, listener: (this: RTCDTMFSender, ev: RTCDTMFSenderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDTMFSenderEventMap>(type: K, listener: (this: RTCDTMFSender, ev: RTCDTMFSenderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDTMFSender: {\n    prototype: RTCDTMFSender;\n    new(): RTCDTMFSender;\n};\n\n/** Events sent to indicate that DTMF tones have started or finished playing. This interface is used by the tonechange event. */\ninterface RTCDTMFToneChangeEvent extends Event {\n    readonly tone: string;\n}\n\ndeclare var RTCDTMFToneChangeEvent: {\n    prototype: RTCDTMFToneChangeEvent;\n    new(type: string, eventInitDict: RTCDTMFToneChangeEventInit): RTCDTMFToneChangeEvent;\n};\n\ninterface RTCDataChannelEventMap {\n    "bufferedamountlow": Event;\n    "close": Event;\n    "error": RTCErrorEvent;\n    "message": MessageEvent;\n    "open": Event;\n}\n\ninterface RTCDataChannel extends EventTarget {\n    binaryType: string;\n    readonly bufferedAmount: number;\n    bufferedAmountLowThreshold: number;\n    readonly id: number | null;\n    readonly label: string;\n    readonly maxPacketLifeTime: number | null;\n    readonly maxRetransmits: number | null;\n    readonly negotiated: boolean;\n    onbufferedamountlow: ((this: RTCDataChannel, ev: Event) => any) | null;\n    onclose: ((this: RTCDataChannel, ev: Event) => any) | null;\n    onerror: ((this: RTCDataChannel, ev: RTCErrorEvent) => any) | null;\n    onmessage: ((this: RTCDataChannel, ev: MessageEvent) => any) | null;\n    onopen: ((this: RTCDataChannel, ev: Event) => any) | null;\n    readonly ordered: boolean;\n    readonly priority: RTCPriorityType;\n    readonly protocol: string;\n    readonly readyState: RTCDataChannelState;\n    close(): void;\n    send(data: string): void;\n    send(data: Blob): void;\n    send(data: ArrayBuffer): void;\n    send(data: ArrayBufferView): void;\n    addEventListener<K extends keyof RTCDataChannelEventMap>(type: K, listener: (this: RTCDataChannel, ev: RTCDataChannelEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDataChannelEventMap>(type: K, listener: (this: RTCDataChannel, ev: RTCDataChannelEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDataChannel: {\n    prototype: RTCDataChannel;\n    new(): RTCDataChannel;\n};\n\ninterface RTCDataChannelEvent extends Event {\n    readonly channel: RTCDataChannel;\n}\n\ndeclare var RTCDataChannelEvent: {\n    prototype: RTCDataChannelEvent;\n    new(type: string, eventInitDict: RTCDataChannelEventInit): RTCDataChannelEvent;\n};\n\ninterface RTCDtlsTransportEventMap {\n    "error": RTCErrorEvent;\n    "statechange": Event;\n}\n\ninterface RTCDtlsTransport extends EventTarget {\n    readonly iceTransport: RTCIceTransport;\n    onerror: ((this: RTCDtlsTransport, ev: RTCErrorEvent) => any) | null;\n    onstatechange: ((this: RTCDtlsTransport, ev: Event) => any) | null;\n    readonly state: RTCDtlsTransportState;\n    getRemoteCertificates(): ArrayBuffer[];\n    addEventListener<K extends keyof RTCDtlsTransportEventMap>(type: K, listener: (this: RTCDtlsTransport, ev: RTCDtlsTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDtlsTransportEventMap>(type: K, listener: (this: RTCDtlsTransport, ev: RTCDtlsTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDtlsTransport: {\n    prototype: RTCDtlsTransport;\n    new(): RTCDtlsTransport;\n};\n\ninterface RTCDtlsTransportStateChangedEvent extends Event {\n    readonly state: RTCDtlsTransportState;\n}\n\ndeclare var RTCDtlsTransportStateChangedEvent: {\n    prototype: RTCDtlsTransportStateChangedEvent;\n    new(): RTCDtlsTransportStateChangedEvent;\n};\n\ninterface RTCDtmfSenderEventMap {\n    "tonechange": RTCDTMFToneChangeEvent;\n}\n\ninterface RTCDtmfSender extends EventTarget {\n    readonly canInsertDTMF: boolean;\n    readonly duration: number;\n    readonly interToneGap: number;\n    ontonechange: ((this: RTCDtmfSender, ev: RTCDTMFToneChangeEvent) => any) | null;\n    readonly sender: RTCRtpSender;\n    readonly toneBuffer: string;\n    insertDTMF(tones: string, duration?: number, interToneGap?: number): void;\n    addEventListener<K extends keyof RTCDtmfSenderEventMap>(type: K, listener: (this: RTCDtmfSender, ev: RTCDtmfSenderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCDtmfSenderEventMap>(type: K, listener: (this: RTCDtmfSender, ev: RTCDtmfSenderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCDtmfSender: {\n    prototype: RTCDtmfSender;\n    new(sender: RTCRtpSender): RTCDtmfSender;\n};\n\ninterface RTCError extends DOMException {\n    readonly errorDetail: RTCErrorDetailType;\n    readonly httpRequestStatusCode: number | null;\n    readonly receivedAlert: number | null;\n    readonly sctpCauseCode: number | null;\n    readonly sdpLineNumber: number | null;\n    readonly sentAlert: number | null;\n}\n\ndeclare var RTCError: {\n    prototype: RTCError;\n    new(init: RTCErrorInit, message?: string): RTCError;\n};\n\ninterface RTCErrorEvent extends Event {\n    readonly error: RTCError;\n}\n\ndeclare var RTCErrorEvent: {\n    prototype: RTCErrorEvent;\n    new(type: string, eventInitDict: RTCErrorEventInit): RTCErrorEvent;\n};\n\n/** The RTCIceCandidate interface—part of the WebRTC API—represents a candidate Internet Connectivity Establishment (ICE) configuration which may be used to establish an RTCPeerConnection. */\ninterface RTCIceCandidate {\n    readonly candidate: string;\n    readonly component: RTCIceComponent | null;\n    readonly foundation: string | null;\n    readonly port: number | null;\n    readonly priority: number | null;\n    readonly protocol: RTCIceProtocol | null;\n    readonly relatedAddress: string | null;\n    readonly relatedPort: number | null;\n    readonly sdpMLineIndex: number | null;\n    readonly sdpMid: string | null;\n    readonly tcpType: RTCIceTcpCandidateType | null;\n    readonly type: RTCIceCandidateType | null;\n    readonly usernameFragment: string | null;\n    toJSON(): RTCIceCandidateInit;\n}\n\ndeclare var RTCIceCandidate: {\n    prototype: RTCIceCandidate;\n    new(candidateInitDict?: RTCIceCandidateInit): RTCIceCandidate;\n};\n\ninterface RTCIceCandidatePairChangedEvent extends Event {\n    readonly pair: RTCIceCandidatePair;\n}\n\ndeclare var RTCIceCandidatePairChangedEvent: {\n    prototype: RTCIceCandidatePairChangedEvent;\n    new(): RTCIceCandidatePairChangedEvent;\n};\n\ninterface RTCIceGathererEventMap {\n    "error": Event;\n    "localcandidate": RTCIceGathererEvent;\n}\n\ninterface RTCIceGatherer extends RTCStatsProvider {\n    readonly component: RTCIceComponent;\n    onerror: ((this: RTCIceGatherer, ev: Event) => any) | null;\n    onlocalcandidate: ((this: RTCIceGatherer, ev: RTCIceGathererEvent) => any) | null;\n    createAssociatedGatherer(): RTCIceGatherer;\n    getLocalCandidates(): RTCIceCandidateDictionary[];\n    getLocalParameters(): RTCIceParameters;\n    addEventListener<K extends keyof RTCIceGathererEventMap>(type: K, listener: (this: RTCIceGatherer, ev: RTCIceGathererEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCIceGathererEventMap>(type: K, listener: (this: RTCIceGatherer, ev: RTCIceGathererEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCIceGatherer: {\n    prototype: RTCIceGatherer;\n    new(options: RTCIceGatherOptions): RTCIceGatherer;\n};\n\ninterface RTCIceGathererEvent extends Event {\n    readonly candidate: RTCIceCandidateDictionary | RTCIceCandidateComplete;\n}\n\ndeclare var RTCIceGathererEvent: {\n    prototype: RTCIceGathererEvent;\n    new(): RTCIceGathererEvent;\n};\n\ninterface RTCIceTransportEventMap {\n    "gatheringstatechange": Event;\n    "selectedcandidatepairchange": Event;\n    "statechange": Event;\n}\n\n/** Provides access to information about the ICE transport layer over which the data is being sent and received. */\ninterface RTCIceTransport extends EventTarget {\n    readonly component: RTCIceComponent;\n    readonly gatheringState: RTCIceGathererState;\n    ongatheringstatechange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    onselectedcandidatepairchange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    onstatechange: ((this: RTCIceTransport, ev: Event) => any) | null;\n    readonly role: RTCIceRole;\n    readonly state: RTCIceTransportState;\n    getLocalCandidates(): RTCIceCandidate[];\n    getLocalParameters(): RTCIceParameters | null;\n    getRemoteCandidates(): RTCIceCandidate[];\n    getRemoteParameters(): RTCIceParameters | null;\n    getSelectedCandidatePair(): RTCIceCandidatePair | null;\n    addEventListener<K extends keyof RTCIceTransportEventMap>(type: K, listener: (this: RTCIceTransport, ev: RTCIceTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCIceTransportEventMap>(type: K, listener: (this: RTCIceTransport, ev: RTCIceTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCIceTransport: {\n    prototype: RTCIceTransport;\n    new(): RTCIceTransport;\n};\n\ninterface RTCIceTransportStateChangedEvent extends Event {\n    readonly state: RTCIceTransportState;\n}\n\ndeclare var RTCIceTransportStateChangedEvent: {\n    prototype: RTCIceTransportStateChangedEvent;\n    new(): RTCIceTransportStateChangedEvent;\n};\n\ninterface RTCIdentityAssertion {\n    idp: string;\n    name: string;\n}\n\ndeclare var RTCIdentityAssertion: {\n    prototype: RTCIdentityAssertion;\n    new(idp: string, name: string): RTCIdentityAssertion;\n};\n\ninterface RTCPeerConnectionEventMap {\n    "connectionstatechange": Event;\n    "datachannel": RTCDataChannelEvent;\n    "icecandidate": RTCPeerConnectionIceEvent;\n    "icecandidateerror": RTCPeerConnectionIceErrorEvent;\n    "iceconnectionstatechange": Event;\n    "icegatheringstatechange": Event;\n    "negotiationneeded": Event;\n    "signalingstatechange": Event;\n    "statsended": RTCStatsEvent;\n    "track": RTCTrackEvent;\n}\n\n/** A WebRTC connection between the local computer and a remote peer. It provides methods to connect to a remote peer, maintain and monitor the connection, and close the connection once it\'s no longer needed. */\ninterface RTCPeerConnection extends EventTarget {\n    readonly canTrickleIceCandidates: boolean | null;\n    readonly connectionState: RTCPeerConnectionState;\n    readonly currentLocalDescription: RTCSessionDescription | null;\n    readonly currentRemoteDescription: RTCSessionDescription | null;\n    readonly iceConnectionState: RTCIceConnectionState;\n    readonly iceGatheringState: RTCIceGatheringState;\n    readonly idpErrorInfo: string | null;\n    readonly idpLoginUrl: string | null;\n    readonly localDescription: RTCSessionDescription | null;\n    onconnectionstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    ondatachannel: ((this: RTCPeerConnection, ev: RTCDataChannelEvent) => any) | null;\n    onicecandidate: ((this: RTCPeerConnection, ev: RTCPeerConnectionIceEvent) => any) | null;\n    onicecandidateerror: ((this: RTCPeerConnection, ev: RTCPeerConnectionIceErrorEvent) => any) | null;\n    oniceconnectionstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onicegatheringstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onnegotiationneeded: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onsignalingstatechange: ((this: RTCPeerConnection, ev: Event) => any) | null;\n    onstatsended: ((this: RTCPeerConnection, ev: RTCStatsEvent) => any) | null;\n    ontrack: ((this: RTCPeerConnection, ev: RTCTrackEvent) => any) | null;\n    readonly peerIdentity: Promise<RTCIdentityAssertion>;\n    readonly pendingLocalDescription: RTCSessionDescription | null;\n    readonly pendingRemoteDescription: RTCSessionDescription | null;\n    readonly remoteDescription: RTCSessionDescription | null;\n    readonly sctp: RTCSctpTransport | null;\n    readonly signalingState: RTCSignalingState;\n    addIceCandidate(candidate: RTCIceCandidateInit | RTCIceCandidate): Promise<void>;\n    addTrack(track: MediaStreamTrack, ...streams: MediaStream[]): RTCRtpSender;\n    addTransceiver(trackOrKind: MediaStreamTrack | string, init?: RTCRtpTransceiverInit): RTCRtpTransceiver;\n    close(): void;\n    createAnswer(options?: RTCOfferOptions): Promise<RTCSessionDescriptionInit>;\n    createDataChannel(label: string, dataChannelDict?: RTCDataChannelInit): RTCDataChannel;\n    createOffer(options?: RTCOfferOptions): Promise<RTCSessionDescriptionInit>;\n    getConfiguration(): RTCConfiguration;\n    getIdentityAssertion(): Promise<string>;\n    getReceivers(): RTCRtpReceiver[];\n    getSenders(): RTCRtpSender[];\n    getStats(selector?: MediaStreamTrack | null): Promise<RTCStatsReport>;\n    getTransceivers(): RTCRtpTransceiver[];\n    removeTrack(sender: RTCRtpSender): void;\n    setConfiguration(configuration: RTCConfiguration): void;\n    setIdentityProvider(provider: string, options?: RTCIdentityProviderOptions): void;\n    setLocalDescription(description: RTCSessionDescriptionInit): Promise<void>;\n    setRemoteDescription(description: RTCSessionDescriptionInit): Promise<void>;\n    addEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: RTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: RTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCPeerConnection: {\n    prototype: RTCPeerConnection;\n    new(configuration?: RTCConfiguration): RTCPeerConnection;\n    generateCertificate(keygenAlgorithm: AlgorithmIdentifier): Promise<RTCCertificate>;\n    getDefaultIceServers(): RTCIceServer[];\n};\n\ninterface RTCPeerConnectionIceErrorEvent extends Event {\n    readonly errorCode: number;\n    readonly errorText: string;\n    readonly hostCandidate: string;\n    readonly url: string;\n}\n\ndeclare var RTCPeerConnectionIceErrorEvent: {\n    prototype: RTCPeerConnectionIceErrorEvent;\n    new(type: string, eventInitDict: RTCPeerConnectionIceErrorEventInit): RTCPeerConnectionIceErrorEvent;\n};\n\n/** Events that occurs in relation to ICE candidates with the target, usually an RTCPeerConnection. Only one event is of this type: icecandidate. */\ninterface RTCPeerConnectionIceEvent extends Event {\n    readonly candidate: RTCIceCandidate | null;\n    readonly url: string | null;\n}\n\ndeclare var RTCPeerConnectionIceEvent: {\n    prototype: RTCPeerConnectionIceEvent;\n    new(type: string, eventInitDict?: RTCPeerConnectionIceEventInit): RTCPeerConnectionIceEvent;\n};\n\n/** This WebRTC API interface manages the reception and decoding of data for a MediaStreamTrack on an RTCPeerConnection. */\ninterface RTCRtpReceiver {\n    readonly rtcpTransport: RTCDtlsTransport | null;\n    readonly track: MediaStreamTrack;\n    readonly transport: RTCDtlsTransport | null;\n    getContributingSources(): RTCRtpContributingSource[];\n    getParameters(): RTCRtpReceiveParameters;\n    getStats(): Promise<RTCStatsReport>;\n    getSynchronizationSources(): RTCRtpSynchronizationSource[];\n}\n\ndeclare var RTCRtpReceiver: {\n    prototype: RTCRtpReceiver;\n    new(): RTCRtpReceiver;\n    getCapabilities(kind: string): RTCRtpCapabilities | null;\n};\n\n/** Provides the ability to control and obtain details about how a particular MediaStreamTrack is encoded and sent to a remote peer. */\ninterface RTCRtpSender {\n    readonly dtmf: RTCDTMFSender | null;\n    readonly rtcpTransport: RTCDtlsTransport | null;\n    readonly track: MediaStreamTrack | null;\n    readonly transport: RTCDtlsTransport | null;\n    getParameters(): RTCRtpSendParameters;\n    getStats(): Promise<RTCStatsReport>;\n    replaceTrack(withTrack: MediaStreamTrack | null): Promise<void>;\n    setParameters(parameters: RTCRtpSendParameters): Promise<void>;\n    setStreams(...streams: MediaStream[]): void;\n}\n\ndeclare var RTCRtpSender: {\n    prototype: RTCRtpSender;\n    new(): RTCRtpSender;\n    getCapabilities(kind: string): RTCRtpCapabilities | null;\n};\n\ninterface RTCRtpTransceiver {\n    readonly currentDirection: RTCRtpTransceiverDirection | null;\n    direction: RTCRtpTransceiverDirection;\n    readonly mid: string | null;\n    readonly receiver: RTCRtpReceiver;\n    readonly sender: RTCRtpSender;\n    setCodecPreferences(codecs: RTCRtpCodecCapability[]): void;\n    stop(): void;\n}\n\ndeclare var RTCRtpTransceiver: {\n    prototype: RTCRtpTransceiver;\n    new(): RTCRtpTransceiver;\n};\n\ninterface RTCSctpTransportEventMap {\n    "statechange": Event;\n}\n\ninterface RTCSctpTransport extends EventTarget {\n    readonly maxChannels: number | null;\n    readonly maxMessageSize: number;\n    onstatechange: ((this: RTCSctpTransport, ev: Event) => any) | null;\n    readonly state: RTCSctpTransportState;\n    readonly transport: RTCDtlsTransport;\n    addEventListener<K extends keyof RTCSctpTransportEventMap>(type: K, listener: (this: RTCSctpTransport, ev: RTCSctpTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCSctpTransportEventMap>(type: K, listener: (this: RTCSctpTransport, ev: RTCSctpTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCSctpTransport: {\n    prototype: RTCSctpTransport;\n    new(): RTCSctpTransport;\n};\n\n/** One end of a connection—or potential connection—and how it\'s configured. Each RTCSessionDescription consists of a description type indicating which part of the offer/answer negotiation process it describes and of the SDP descriptor of the session. */\ninterface RTCSessionDescription {\n    readonly sdp: string;\n    readonly type: RTCSdpType;\n    toJSON(): any;\n}\n\ndeclare var RTCSessionDescription: {\n    prototype: RTCSessionDescription;\n    new(descriptionInitDict?: RTCSessionDescriptionInit): RTCSessionDescription;\n};\n\ninterface RTCSrtpSdesTransportEventMap {\n    "error": Event;\n}\n\ninterface RTCSrtpSdesTransport extends EventTarget {\n    onerror: ((this: RTCSrtpSdesTransport, ev: Event) => any) | null;\n    readonly transport: RTCIceTransport;\n    addEventListener<K extends keyof RTCSrtpSdesTransportEventMap>(type: K, listener: (this: RTCSrtpSdesTransport, ev: RTCSrtpSdesTransportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCSrtpSdesTransportEventMap>(type: K, listener: (this: RTCSrtpSdesTransport, ev: RTCSrtpSdesTransportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var RTCSrtpSdesTransport: {\n    prototype: RTCSrtpSdesTransport;\n    new(transport: RTCIceTransport, encryptParameters: RTCSrtpSdesParameters, decryptParameters: RTCSrtpSdesParameters): RTCSrtpSdesTransport;\n    getLocalParameters(): RTCSrtpSdesParameters[];\n};\n\ninterface RTCSsrcConflictEvent extends Event {\n    readonly ssrc: number;\n}\n\ndeclare var RTCSsrcConflictEvent: {\n    prototype: RTCSsrcConflictEvent;\n    new(): RTCSsrcConflictEvent;\n};\n\ninterface RTCStatsEvent extends Event {\n    readonly report: RTCStatsReport;\n}\n\ndeclare var RTCStatsEvent: {\n    prototype: RTCStatsEvent;\n    new(type: string, eventInitDict: RTCStatsEventInit): RTCStatsEvent;\n};\n\ninterface RTCStatsProvider extends EventTarget {\n    getStats(): Promise<RTCStatsReport>;\n    msGetStats(): Promise<RTCStatsReport>;\n}\n\ndeclare var RTCStatsProvider: {\n    prototype: RTCStatsProvider;\n    new(): RTCStatsProvider;\n};\n\ninterface RTCStatsReport {\n    forEach(callbackfn: (value: any, key: string, parent: RTCStatsReport) => void, thisArg?: any): void;\n}\n\ndeclare var RTCStatsReport: {\n    prototype: RTCStatsReport;\n    new(): RTCStatsReport;\n};\n\ninterface RTCTrackEvent extends Event {\n    readonly receiver: RTCRtpReceiver;\n    readonly streams: ReadonlyArray<MediaStream>;\n    readonly track: MediaStreamTrack;\n    readonly transceiver: RTCRtpTransceiver;\n}\n\ndeclare var RTCTrackEvent: {\n    prototype: RTCTrackEvent;\n    new(type: string, eventInitDict: RTCTrackEventInit): RTCTrackEvent;\n};\n\ninterface RadioNodeList extends NodeList {\n    value: string;\n}\n\ndeclare var RadioNodeList: {\n    prototype: RadioNodeList;\n    new(): RadioNodeList;\n};\n\ninterface RandomSource {\n    getRandomValues<T extends Int8Array | Uint8ClampedArray | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array>(array: T): T;\n}\n\ndeclare var RandomSource: {\n    prototype: RandomSource;\n    new(): RandomSource;\n};\n\n/** A fragment of a document that can contain nodes and parts of text nodes. */\ninterface Range extends AbstractRange {\n    /**\n     * Returns the node, furthest away from the document, that is an ancestor of both range\'s start node and end node.\n     */\n    readonly commonAncestorContainer: Node;\n    cloneContents(): DocumentFragment;\n    cloneRange(): Range;\n    collapse(toStart?: boolean): void;\n    compareBoundaryPoints(how: number, sourceRange: Range): number;\n    /**\n     * Returns −1 if the point is before the range, 0 if the point is in the range, and 1 if the point is after the range.\n     */\n    comparePoint(node: Node, offset: number): number;\n    createContextualFragment(fragment: string): DocumentFragment;\n    deleteContents(): void;\n    detach(): void;\n    extractContents(): DocumentFragment;\n    getBoundingClientRect(): DOMRect;\n    getClientRects(): DOMRectList;\n    insertNode(node: Node): void;\n    /**\n     * Returns whether range intersects node.\n     */\n    intersectsNode(node: Node): boolean;\n    isPointInRange(node: Node, offset: number): boolean;\n    selectNode(node: Node): void;\n    selectNodeContents(node: Node): void;\n    setEnd(node: Node, offset: number): void;\n    setEndAfter(node: Node): void;\n    setEndBefore(node: Node): void;\n    setStart(node: Node, offset: number): void;\n    setStartAfter(node: Node): void;\n    setStartBefore(node: Node): void;\n    surroundContents(newParent: Node): void;\n    toString(): string;\n    readonly END_TO_END: number;\n    readonly END_TO_START: number;\n    readonly START_TO_END: number;\n    readonly START_TO_START: number;\n}\n\ndeclare var Range: {\n    prototype: Range;\n    new(): Range;\n    readonly END_TO_END: number;\n    readonly END_TO_START: number;\n    readonly START_TO_END: number;\n    readonly START_TO_START: number;\n    toString(): string;\n};\n\ninterface ReadableByteStreamController {\n    readonly byobRequest: ReadableStreamBYOBRequest | undefined;\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: ArrayBufferView): void;\n    error(error?: any): void;\n}\n\ndeclare var ReadableByteStreamController: {\n    prototype: ReadableByteStreamController;\n    new(): ReadableByteStreamController;\n};\n\n/** This Streams API interface represents a readable stream of byte data. The Fetch API offers a concrete instance of a ReadableStream through the body property of a Response object. */\ninterface ReadableStream<R = any> {\n    readonly locked: boolean;\n    cancel(reason?: any): Promise<void>;\n    getReader(options: { mode: "byob" }): ReadableStreamBYOBReader;\n    getReader(): ReadableStreamDefaultReader<R>;\n    pipeThrough<T>({ writable, readable }: { writable: WritableStream<R>, readable: ReadableStream<T> }, options?: PipeOptions): ReadableStream<T>;\n    pipeTo(dest: WritableStream<R>, options?: PipeOptions): Promise<void>;\n    tee(): [ReadableStream<R>, ReadableStream<R>];\n}\n\ndeclare var ReadableStream: {\n    prototype: ReadableStream;\n    new(underlyingSource: UnderlyingByteSource, strategy?: { highWaterMark?: number, size?: undefined }): ReadableStream<Uint8Array>;\n    new<R = any>(underlyingSource?: UnderlyingSource<R>, strategy?: QueuingStrategy<R>): ReadableStream<R>;\n};\n\ninterface ReadableStreamBYOBReader {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read<T extends ArrayBufferView>(view: T): Promise<ReadableStreamReadResult<T>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamBYOBReader: {\n    prototype: ReadableStreamBYOBReader;\n    new(): ReadableStreamBYOBReader;\n};\n\ninterface ReadableStreamBYOBRequest {\n    readonly view: ArrayBufferView;\n    respond(bytesWritten: number): void;\n    respondWithNewView(view: ArrayBufferView): void;\n}\n\ndeclare var ReadableStreamBYOBRequest: {\n    prototype: ReadableStreamBYOBRequest;\n    new(): ReadableStreamBYOBRequest;\n};\n\ninterface ReadableStreamDefaultController<R = any> {\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: R): void;\n    error(error?: any): void;\n}\n\ndeclare var ReadableStreamDefaultController: {\n    prototype: ReadableStreamDefaultController;\n    new(): ReadableStreamDefaultController;\n};\n\ninterface ReadableStreamDefaultReader<R = any> {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamDefaultReader: {\n    prototype: ReadableStreamDefaultReader;\n    new(): ReadableStreamDefaultReader;\n};\n\ninterface ReadableStreamReader<R = any> {\n    cancel(): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamReader: {\n    prototype: ReadableStreamReader;\n    new(): ReadableStreamReader;\n};\n\n/** This Fetch API interface represents a resource request. */\ninterface Request extends Body {\n    /**\n     * Returns the cache mode associated with request, which is a string indicating how the request will interact with the browser\'s cache when fetching.\n     */\n    readonly cache: RequestCache;\n    /**\n     * Returns the credentials mode associated with request, which is a string indicating whether credentials will be sent with the request always, never, or only when sent to a same-origin URL.\n     */\n    readonly credentials: RequestCredentials;\n    /**\n     * Returns the kind of resource requested by request, e.g., "document" or "script".\n     */\n    readonly destination: RequestDestination;\n    /**\n     * Returns a Headers object consisting of the headers associated with request. Note that headers added in the network layer by the user agent will not be accounted for in this object, e.g., the "Host" header.\n     */\n    readonly headers: Headers;\n    /**\n     * Returns request\'s subresource integrity metadata, which is a cryptographic hash of the resource being fetched. Its value consists of multiple hashes separated by whitespace. [SRI]\n     */\n    readonly integrity: string;\n    /**\n     * Returns a boolean indicating whether or not request is for a history navigation (a.k.a. back-foward navigation).\n     */\n    readonly isHistoryNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request is for a reload navigation.\n     */\n    readonly isReloadNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request can outlive the global in which it was created.\n     */\n    readonly keepalive: boolean;\n    /**\n     * Returns request\'s HTTP method, which is "GET" by default.\n     */\n    readonly method: string;\n    /**\n     * Returns the mode associated with request, which is a string indicating whether the request will use CORS, or will be restricted to same-origin URLs.\n     */\n    readonly mode: RequestMode;\n    /**\n     * Returns the redirect mode associated with request, which is a string indicating how redirects for the request will be handled during fetching. A request will follow redirects by default.\n     */\n    readonly redirect: RequestRedirect;\n    /**\n     * Returns the referrer of request. Its value can be a same-origin URL if explicitly set in init, the empty string to indicate no referrer, and "about:client" when defaulting to the global\'s default. This is used during fetching to determine the value of the `Referer` header of the request being made.\n     */\n    readonly referrer: string;\n    /**\n     * Returns the referrer policy associated with request. This is used during fetching to compute the value of the request\'s referrer.\n     */\n    readonly referrerPolicy: ReferrerPolicy;\n    /**\n     * Returns the signal associated with request, which is an AbortSignal object indicating whether or not request has been aborted, and its abort event handler.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Returns the URL of request as a string.\n     */\n    readonly url: string;\n    clone(): Request;\n}\n\ndeclare var Request: {\n    prototype: Request;\n    new(input: RequestInfo, init?: RequestInit): Request;\n};\n\n/** This Fetch API interface represents the response to a request. */\ninterface Response extends Body {\n    readonly headers: Headers;\n    readonly ok: boolean;\n    readonly redirected: boolean;\n    readonly status: number;\n    readonly statusText: string;\n    readonly trailer: Promise<Headers>;\n    readonly type: ResponseType;\n    readonly url: string;\n    clone(): Response;\n}\n\ndeclare var Response: {\n    prototype: Response;\n    new(body?: BodyInit | null, init?: ResponseInit): Response;\n    error(): Response;\n    redirect(url: string, status?: number): Response;\n};\n\n/** Provides access to the properties of <a> element, as well as methods to manipulate them. */\ninterface SVGAElement extends SVGGraphicsElement, SVGURIReference {\n    readonly target: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAElement: {\n    prototype: SVGAElement;\n    new(): SVGAElement;\n};\n\n/** Used to represent a value that can be an <angle> or <number> value. An SVGAngle reflected through the animVal attribute is always read only. */\ninterface SVGAngle {\n    readonly unitType: number;\n    value: number;\n    valueAsString: string;\n    valueInSpecifiedUnits: number;\n    convertToSpecifiedUnits(unitType: number): void;\n    newValueSpecifiedUnits(unitType: number, valueInSpecifiedUnits: number): void;\n    readonly SVG_ANGLETYPE_DEG: number;\n    readonly SVG_ANGLETYPE_GRAD: number;\n    readonly SVG_ANGLETYPE_RAD: number;\n    readonly SVG_ANGLETYPE_UNKNOWN: number;\n    readonly SVG_ANGLETYPE_UNSPECIFIED: number;\n}\n\ndeclare var SVGAngle: {\n    prototype: SVGAngle;\n    new(): SVGAngle;\n    readonly SVG_ANGLETYPE_DEG: number;\n    readonly SVG_ANGLETYPE_GRAD: number;\n    readonly SVG_ANGLETYPE_RAD: number;\n    readonly SVG_ANGLETYPE_UNKNOWN: number;\n    readonly SVG_ANGLETYPE_UNSPECIFIED: number;\n};\n\ninterface SVGAnimateElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateElement: {\n    prototype: SVGAnimateElement;\n    new(): SVGAnimateElement;\n};\n\ninterface SVGAnimateMotionElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateMotionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateMotionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateMotionElement: {\n    prototype: SVGAnimateMotionElement;\n    new(): SVGAnimateMotionElement;\n};\n\ninterface SVGAnimateTransformElement extends SVGAnimationElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateTransformElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimateTransformElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimateTransformElement: {\n    prototype: SVGAnimateTransformElement;\n    new(): SVGAnimateTransformElement;\n};\n\n/** Used for attributes of basic type <angle> which can be animated. */\ninterface SVGAnimatedAngle {\n    readonly animVal: SVGAngle;\n    readonly baseVal: SVGAngle;\n}\n\ndeclare var SVGAnimatedAngle: {\n    prototype: SVGAnimatedAngle;\n    new(): SVGAnimatedAngle;\n};\n\n/** Used for attributes of type boolean which can be animated. */\ninterface SVGAnimatedBoolean {\n    readonly animVal: boolean;\n    baseVal: boolean;\n}\n\ndeclare var SVGAnimatedBoolean: {\n    prototype: SVGAnimatedBoolean;\n    new(): SVGAnimatedBoolean;\n};\n\n/** Used for attributes whose value must be a constant from a particular enumeration and which can be animated. */\ninterface SVGAnimatedEnumeration {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedEnumeration: {\n    prototype: SVGAnimatedEnumeration;\n    new(): SVGAnimatedEnumeration;\n};\n\n/** Used for attributes of basic type <integer> which can be animated. */\ninterface SVGAnimatedInteger {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedInteger: {\n    prototype: SVGAnimatedInteger;\n    new(): SVGAnimatedInteger;\n};\n\n/** Used for attributes of basic type <length> which can be animated. */\ninterface SVGAnimatedLength {\n    readonly animVal: SVGLength;\n    readonly baseVal: SVGLength;\n}\n\ndeclare var SVGAnimatedLength: {\n    prototype: SVGAnimatedLength;\n    new(): SVGAnimatedLength;\n};\n\n/** Used for attributes of type SVGLengthList which can be animated. */\ninterface SVGAnimatedLengthList {\n    readonly animVal: SVGLengthList;\n    readonly baseVal: SVGLengthList;\n}\n\ndeclare var SVGAnimatedLengthList: {\n    prototype: SVGAnimatedLengthList;\n    new(): SVGAnimatedLengthList;\n};\n\n/** Used for attributes of basic type <Number> which can be animated. */\ninterface SVGAnimatedNumber {\n    readonly animVal: number;\n    baseVal: number;\n}\n\ndeclare var SVGAnimatedNumber: {\n    prototype: SVGAnimatedNumber;\n    new(): SVGAnimatedNumber;\n};\n\n/** The SVGAnimatedNumber interface is used for attributes which take a list of numbers and which can be animated. */\ninterface SVGAnimatedNumberList {\n    readonly animVal: SVGNumberList;\n    readonly baseVal: SVGNumberList;\n}\n\ndeclare var SVGAnimatedNumberList: {\n    prototype: SVGAnimatedNumberList;\n    new(): SVGAnimatedNumberList;\n};\n\ninterface SVGAnimatedPoints {\n    readonly animatedPoints: SVGPointList;\n    readonly points: SVGPointList;\n}\n\n/** Used for attributes of type SVGPreserveAspectRatio which can be animated. */\ninterface SVGAnimatedPreserveAspectRatio {\n    readonly animVal: SVGPreserveAspectRatio;\n    readonly baseVal: SVGPreserveAspectRatio;\n}\n\ndeclare var SVGAnimatedPreserveAspectRatio: {\n    prototype: SVGAnimatedPreserveAspectRatio;\n    new(): SVGAnimatedPreserveAspectRatio;\n};\n\n/** Used for attributes of basic SVGRect which can be animated. */\ninterface SVGAnimatedRect {\n    readonly animVal: DOMRectReadOnly;\n    readonly baseVal: DOMRect;\n}\n\ndeclare var SVGAnimatedRect: {\n    prototype: SVGAnimatedRect;\n    new(): SVGAnimatedRect;\n};\n\n/** The SVGAnimatedString interface represents string attributes which can be animated from each SVG declaration. You need to create SVG attribute before doing anything else, everything should be declared inside this. */\ninterface SVGAnimatedString {\n    readonly animVal: string;\n    baseVal: string;\n}\n\ndeclare var SVGAnimatedString: {\n    prototype: SVGAnimatedString;\n    new(): SVGAnimatedString;\n};\n\n/** Used for attributes which take a list of numbers and which can be animated. */\ninterface SVGAnimatedTransformList {\n    readonly animVal: SVGTransformList;\n    readonly baseVal: SVGTransformList;\n}\n\ndeclare var SVGAnimatedTransformList: {\n    prototype: SVGAnimatedTransformList;\n    new(): SVGAnimatedTransformList;\n};\n\ninterface SVGAnimationElement extends SVGElement {\n    readonly targetElement: SVGElement;\n    getCurrentTime(): number;\n    getSimpleDuration(): number;\n    getStartTime(): number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimationElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGAnimationElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGAnimationElement: {\n    prototype: SVGAnimationElement;\n    new(): SVGAnimationElement;\n};\n\n/** An interface for the <circle> element. The circle element is defined by the cx and cy attributes that denote the coordinates of the centre of the circle. */\ninterface SVGCircleElement extends SVGGeometryElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly r: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCircleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCircleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGCircleElement: {\n    prototype: SVGCircleElement;\n    new(): SVGCircleElement;\n};\n\n/** Provides access to the properties of <clipPath> elements, as well as methods to manipulate them. */\ninterface SVGClipPathElement extends SVGElement {\n    readonly clipPathUnits: SVGAnimatedEnumeration;\n    readonly transform: SVGAnimatedTransformList;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGClipPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGClipPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGClipPathElement: {\n    prototype: SVGClipPathElement;\n    new(): SVGClipPathElement;\n};\n\n/** A base interface used by the component transfer function interfaces. */\ninterface SVGComponentTransferFunctionElement extends SVGElement {\n    readonly amplitude: SVGAnimatedNumber;\n    readonly exponent: SVGAnimatedNumber;\n    readonly intercept: SVGAnimatedNumber;\n    readonly offset: SVGAnimatedNumber;\n    readonly slope: SVGAnimatedNumber;\n    readonly tableValues: SVGAnimatedNumberList;\n    readonly type: SVGAnimatedEnumeration;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_DISCRETE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_GAMMA: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_IDENTITY: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_LINEAR: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_TABLE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGComponentTransferFunctionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGComponentTransferFunctionElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGComponentTransferFunctionElement: {\n    prototype: SVGComponentTransferFunctionElement;\n    new(): SVGComponentTransferFunctionElement;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_DISCRETE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_GAMMA: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_IDENTITY: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_LINEAR: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_TABLE: number;\n    readonly SVG_FECOMPONENTTRANSFER_TYPE_UNKNOWN: number;\n};\n\ninterface SVGCursorElement extends SVGElement {\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCursorElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGCursorElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGCursorElement: {\n    prototype: SVGCursorElement;\n    new(): SVGCursorElement;\n};\n\n/** Corresponds to the <defs> element. */\ninterface SVGDefsElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDefsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDefsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGDefsElement: {\n    prototype: SVGDefsElement;\n    new(): SVGDefsElement;\n};\n\n/** Corresponds to the <desc> element. */\ninterface SVGDescElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDescElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGDescElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGDescElement: {\n    prototype: SVGDescElement;\n    new(): SVGDescElement;\n};\n\ninterface SVGElementEventMap extends ElementEventMap, GlobalEventHandlersEventMap, DocumentAndElementEventHandlersEventMap {\n}\n\n/** All of the SVG DOM interfaces that correspond directly to elements in the SVG language derive from the SVGElement interface. */\ninterface SVGElement extends Element, DocumentAndElementEventHandlers, DocumentAndElementEventHandlers, ElementCSSInlineStyle, GlobalEventHandlers, GlobalEventHandlers, HTMLOrSVGElement, SVGElementInstance {\n    /** @deprecated */\n    readonly className: any;\n    readonly ownerSVGElement: SVGSVGElement | null;\n    readonly viewportElement: SVGElement | null;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGElement: {\n    prototype: SVGElement;\n    new(): SVGElement;\n};\n\ninterface SVGElementInstance extends EventTarget {\n    readonly correspondingElement: SVGElement;\n    readonly correspondingUseElement: SVGUseElement;\n}\n\ndeclare var SVGElementInstance: {\n    prototype: SVGElementInstance;\n    new(): SVGElementInstance;\n};\n\ninterface SVGElementInstanceList {\n    /** @deprecated */\n    readonly length: number;\n    /** @deprecated */\n    item(index: number): SVGElementInstance;\n}\n\ndeclare var SVGElementInstanceList: {\n    prototype: SVGElementInstanceList;\n    new(): SVGElementInstanceList;\n};\n\n/** Provides access to the properties of <ellipse> elements. */\ninterface SVGEllipseElement extends SVGGeometryElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly rx: SVGAnimatedLength;\n    readonly ry: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGEllipseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGEllipseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGEllipseElement: {\n    prototype: SVGEllipseElement;\n    new(): SVGEllipseElement;\n};\n\n/** Corresponds to the <feBlend> element. */\ninterface SVGFEBlendElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly mode: SVGAnimatedEnumeration;\n    readonly SVG_FEBLEND_MODE_COLOR: number;\n    readonly SVG_FEBLEND_MODE_COLOR_BURN: number;\n    readonly SVG_FEBLEND_MODE_COLOR_DODGE: number;\n    readonly SVG_FEBLEND_MODE_DARKEN: number;\n    readonly SVG_FEBLEND_MODE_DIFFERENCE: number;\n    readonly SVG_FEBLEND_MODE_EXCLUSION: number;\n    readonly SVG_FEBLEND_MODE_HARD_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_HUE: number;\n    readonly SVG_FEBLEND_MODE_LIGHTEN: number;\n    readonly SVG_FEBLEND_MODE_LUMINOSITY: number;\n    readonly SVG_FEBLEND_MODE_MULTIPLY: number;\n    readonly SVG_FEBLEND_MODE_NORMAL: number;\n    readonly SVG_FEBLEND_MODE_OVERLAY: number;\n    readonly SVG_FEBLEND_MODE_SATURATION: number;\n    readonly SVG_FEBLEND_MODE_SCREEN: number;\n    readonly SVG_FEBLEND_MODE_SOFT_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEBlendElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEBlendElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEBlendElement: {\n    prototype: SVGFEBlendElement;\n    new(): SVGFEBlendElement;\n    readonly SVG_FEBLEND_MODE_COLOR: number;\n    readonly SVG_FEBLEND_MODE_COLOR_BURN: number;\n    readonly SVG_FEBLEND_MODE_COLOR_DODGE: number;\n    readonly SVG_FEBLEND_MODE_DARKEN: number;\n    readonly SVG_FEBLEND_MODE_DIFFERENCE: number;\n    readonly SVG_FEBLEND_MODE_EXCLUSION: number;\n    readonly SVG_FEBLEND_MODE_HARD_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_HUE: number;\n    readonly SVG_FEBLEND_MODE_LIGHTEN: number;\n    readonly SVG_FEBLEND_MODE_LUMINOSITY: number;\n    readonly SVG_FEBLEND_MODE_MULTIPLY: number;\n    readonly SVG_FEBLEND_MODE_NORMAL: number;\n    readonly SVG_FEBLEND_MODE_OVERLAY: number;\n    readonly SVG_FEBLEND_MODE_SATURATION: number;\n    readonly SVG_FEBLEND_MODE_SCREEN: number;\n    readonly SVG_FEBLEND_MODE_SOFT_LIGHT: number;\n    readonly SVG_FEBLEND_MODE_UNKNOWN: number;\n};\n\n/** Corresponds to the <feColorMatrix> element. */\ninterface SVGFEColorMatrixElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly type: SVGAnimatedEnumeration;\n    readonly values: SVGAnimatedNumberList;\n    readonly SVG_FECOLORMATRIX_TYPE_HUEROTATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_LUMINANCETOALPHA: number;\n    readonly SVG_FECOLORMATRIX_TYPE_MATRIX: number;\n    readonly SVG_FECOLORMATRIX_TYPE_SATURATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEColorMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEColorMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEColorMatrixElement: {\n    prototype: SVGFEColorMatrixElement;\n    new(): SVGFEColorMatrixElement;\n    readonly SVG_FECOLORMATRIX_TYPE_HUEROTATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_LUMINANCETOALPHA: number;\n    readonly SVG_FECOLORMATRIX_TYPE_MATRIX: number;\n    readonly SVG_FECOLORMATRIX_TYPE_SATURATE: number;\n    readonly SVG_FECOLORMATRIX_TYPE_UNKNOWN: number;\n};\n\n/** Corresponds to the <feComponentTransfer> element. */\ninterface SVGFEComponentTransferElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEComponentTransferElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEComponentTransferElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEComponentTransferElement: {\n    prototype: SVGFEComponentTransferElement;\n    new(): SVGFEComponentTransferElement;\n};\n\n/** Corresponds to the <feComposite> element. */\ninterface SVGFECompositeElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly k1: SVGAnimatedNumber;\n    readonly k2: SVGAnimatedNumber;\n    readonly k3: SVGAnimatedNumber;\n    readonly k4: SVGAnimatedNumber;\n    readonly operator: SVGAnimatedEnumeration;\n    readonly SVG_FECOMPOSITE_OPERATOR_ARITHMETIC: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_ATOP: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_IN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OUT: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OVER: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_UNKNOWN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_XOR: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFECompositeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFECompositeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFECompositeElement: {\n    prototype: SVGFECompositeElement;\n    new(): SVGFECompositeElement;\n    readonly SVG_FECOMPOSITE_OPERATOR_ARITHMETIC: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_ATOP: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_IN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OUT: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_OVER: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_UNKNOWN: number;\n    readonly SVG_FECOMPOSITE_OPERATOR_XOR: number;\n};\n\n/** Corresponds to the <feConvolveMatrix> element. */\ninterface SVGFEConvolveMatrixElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly bias: SVGAnimatedNumber;\n    readonly divisor: SVGAnimatedNumber;\n    readonly edgeMode: SVGAnimatedEnumeration;\n    readonly in1: SVGAnimatedString;\n    readonly kernelMatrix: SVGAnimatedNumberList;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly orderX: SVGAnimatedInteger;\n    readonly orderY: SVGAnimatedInteger;\n    readonly preserveAlpha: SVGAnimatedBoolean;\n    readonly targetX: SVGAnimatedInteger;\n    readonly targetY: SVGAnimatedInteger;\n    readonly SVG_EDGEMODE_DUPLICATE: number;\n    readonly SVG_EDGEMODE_NONE: number;\n    readonly SVG_EDGEMODE_UNKNOWN: number;\n    readonly SVG_EDGEMODE_WRAP: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEConvolveMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEConvolveMatrixElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEConvolveMatrixElement: {\n    prototype: SVGFEConvolveMatrixElement;\n    new(): SVGFEConvolveMatrixElement;\n    readonly SVG_EDGEMODE_DUPLICATE: number;\n    readonly SVG_EDGEMODE_NONE: number;\n    readonly SVG_EDGEMODE_UNKNOWN: number;\n    readonly SVG_EDGEMODE_WRAP: number;\n};\n\n/** Corresponds to the <feDiffuseLighting> element. */\ninterface SVGFEDiffuseLightingElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly diffuseConstant: SVGAnimatedNumber;\n    readonly in1: SVGAnimatedString;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly surfaceScale: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDiffuseLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDiffuseLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDiffuseLightingElement: {\n    prototype: SVGFEDiffuseLightingElement;\n    new(): SVGFEDiffuseLightingElement;\n};\n\n/** Corresponds to the <feDisplacementMap> element. */\ninterface SVGFEDisplacementMapElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly in2: SVGAnimatedString;\n    readonly scale: SVGAnimatedNumber;\n    readonly xChannelSelector: SVGAnimatedEnumeration;\n    readonly yChannelSelector: SVGAnimatedEnumeration;\n    readonly SVG_CHANNEL_A: number;\n    readonly SVG_CHANNEL_B: number;\n    readonly SVG_CHANNEL_G: number;\n    readonly SVG_CHANNEL_R: number;\n    readonly SVG_CHANNEL_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDisplacementMapElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDisplacementMapElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDisplacementMapElement: {\n    prototype: SVGFEDisplacementMapElement;\n    new(): SVGFEDisplacementMapElement;\n    readonly SVG_CHANNEL_A: number;\n    readonly SVG_CHANNEL_B: number;\n    readonly SVG_CHANNEL_G: number;\n    readonly SVG_CHANNEL_R: number;\n    readonly SVG_CHANNEL_UNKNOWN: number;\n};\n\n/** Corresponds to the <feDistantLight> element. */\ninterface SVGFEDistantLightElement extends SVGElement {\n    readonly azimuth: SVGAnimatedNumber;\n    readonly elevation: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDistantLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDistantLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDistantLightElement: {\n    prototype: SVGFEDistantLightElement;\n    new(): SVGFEDistantLightElement;\n};\n\ninterface SVGFEDropShadowElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly dx: SVGAnimatedNumber;\n    readonly dy: SVGAnimatedNumber;\n    readonly in1: SVGAnimatedString;\n    readonly stdDeviationX: SVGAnimatedNumber;\n    readonly stdDeviationY: SVGAnimatedNumber;\n    setStdDeviation(stdDeviationX: number, stdDeviationY: number): void;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDropShadowElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEDropShadowElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEDropShadowElement: {\n    prototype: SVGFEDropShadowElement;\n    new(): SVGFEDropShadowElement;\n};\n\n/** Corresponds to the <feFlood> element. */\ninterface SVGFEFloodElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFloodElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFloodElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFloodElement: {\n    prototype: SVGFEFloodElement;\n    new(): SVGFEFloodElement;\n};\n\n/** Corresponds to the <feFuncA> element. */\ninterface SVGFEFuncAElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncAElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncAElement: {\n    prototype: SVGFEFuncAElement;\n    new(): SVGFEFuncAElement;\n};\n\n/** Corresponds to the <feFuncB> element. */\ninterface SVGFEFuncBElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncBElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncBElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncBElement: {\n    prototype: SVGFEFuncBElement;\n    new(): SVGFEFuncBElement;\n};\n\n/** Corresponds to the <feFuncG> element. */\ninterface SVGFEFuncGElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncGElement: {\n    prototype: SVGFEFuncGElement;\n    new(): SVGFEFuncGElement;\n};\n\n/** Corresponds to the <feFuncR> element. */\ninterface SVGFEFuncRElement extends SVGComponentTransferFunctionElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncRElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEFuncRElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEFuncRElement: {\n    prototype: SVGFEFuncRElement;\n    new(): SVGFEFuncRElement;\n};\n\n/** Corresponds to the <feGaussianBlur> element. */\ninterface SVGFEGaussianBlurElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly stdDeviationX: SVGAnimatedNumber;\n    readonly stdDeviationY: SVGAnimatedNumber;\n    setStdDeviation(stdDeviationX: number, stdDeviationY: number): void;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEGaussianBlurElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEGaussianBlurElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEGaussianBlurElement: {\n    prototype: SVGFEGaussianBlurElement;\n    new(): SVGFEGaussianBlurElement;\n};\n\n/** Corresponds to the <feImage> element. */\ninterface SVGFEImageElement extends SVGElement, SVGFilterPrimitiveStandardAttributes, SVGURIReference {\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEImageElement: {\n    prototype: SVGFEImageElement;\n    new(): SVGFEImageElement;\n};\n\n/** Corresponds to the <feMerge> element. */\ninterface SVGFEMergeElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMergeElement: {\n    prototype: SVGFEMergeElement;\n    new(): SVGFEMergeElement;\n};\n\n/** Corresponds to the <feMergeNode> element. */\ninterface SVGFEMergeNodeElement extends SVGElement {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeNodeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMergeNodeElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMergeNodeElement: {\n    prototype: SVGFEMergeNodeElement;\n    new(): SVGFEMergeNodeElement;\n};\n\n/** Corresponds to the <feMorphology> element. */\ninterface SVGFEMorphologyElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly operator: SVGAnimatedEnumeration;\n    readonly radiusX: SVGAnimatedNumber;\n    readonly radiusY: SVGAnimatedNumber;\n    readonly SVG_MORPHOLOGY_OPERATOR_DILATE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_ERODE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMorphologyElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEMorphologyElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEMorphologyElement: {\n    prototype: SVGFEMorphologyElement;\n    new(): SVGFEMorphologyElement;\n    readonly SVG_MORPHOLOGY_OPERATOR_DILATE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_ERODE: number;\n    readonly SVG_MORPHOLOGY_OPERATOR_UNKNOWN: number;\n};\n\n/** Corresponds to the <feOffset> element. */\ninterface SVGFEOffsetElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly dx: SVGAnimatedNumber;\n    readonly dy: SVGAnimatedNumber;\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEOffsetElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEOffsetElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEOffsetElement: {\n    prototype: SVGFEOffsetElement;\n    new(): SVGFEOffsetElement;\n};\n\n/** Corresponds to the <fePointLight> element. */\ninterface SVGFEPointLightElement extends SVGElement {\n    readonly x: SVGAnimatedNumber;\n    readonly y: SVGAnimatedNumber;\n    readonly z: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEPointLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFEPointLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFEPointLightElement: {\n    prototype: SVGFEPointLightElement;\n    new(): SVGFEPointLightElement;\n};\n\n/** Corresponds to the <feSpecularLighting> element. */\ninterface SVGFESpecularLightingElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    readonly kernelUnitLengthX: SVGAnimatedNumber;\n    readonly kernelUnitLengthY: SVGAnimatedNumber;\n    readonly specularConstant: SVGAnimatedNumber;\n    readonly specularExponent: SVGAnimatedNumber;\n    readonly surfaceScale: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpecularLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpecularLightingElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFESpecularLightingElement: {\n    prototype: SVGFESpecularLightingElement;\n    new(): SVGFESpecularLightingElement;\n};\n\n/** Corresponds to the <feSpotLight> element. */\ninterface SVGFESpotLightElement extends SVGElement {\n    readonly limitingConeAngle: SVGAnimatedNumber;\n    readonly pointsAtX: SVGAnimatedNumber;\n    readonly pointsAtY: SVGAnimatedNumber;\n    readonly pointsAtZ: SVGAnimatedNumber;\n    readonly specularExponent: SVGAnimatedNumber;\n    readonly x: SVGAnimatedNumber;\n    readonly y: SVGAnimatedNumber;\n    readonly z: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpotLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFESpotLightElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFESpotLightElement: {\n    prototype: SVGFESpotLightElement;\n    new(): SVGFESpotLightElement;\n};\n\n/** Corresponds to the <feTile> element. */\ninterface SVGFETileElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly in1: SVGAnimatedString;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETileElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETileElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFETileElement: {\n    prototype: SVGFETileElement;\n    new(): SVGFETileElement;\n};\n\n/** Corresponds to the <feTurbulence> element. */\ninterface SVGFETurbulenceElement extends SVGElement, SVGFilterPrimitiveStandardAttributes {\n    readonly baseFrequencyX: SVGAnimatedNumber;\n    readonly baseFrequencyY: SVGAnimatedNumber;\n    readonly numOctaves: SVGAnimatedInteger;\n    readonly seed: SVGAnimatedNumber;\n    readonly stitchTiles: SVGAnimatedEnumeration;\n    readonly type: SVGAnimatedEnumeration;\n    readonly SVG_STITCHTYPE_NOSTITCH: number;\n    readonly SVG_STITCHTYPE_STITCH: number;\n    readonly SVG_STITCHTYPE_UNKNOWN: number;\n    readonly SVG_TURBULENCE_TYPE_FRACTALNOISE: number;\n    readonly SVG_TURBULENCE_TYPE_TURBULENCE: number;\n    readonly SVG_TURBULENCE_TYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETurbulenceElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFETurbulenceElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFETurbulenceElement: {\n    prototype: SVGFETurbulenceElement;\n    new(): SVGFETurbulenceElement;\n    readonly SVG_STITCHTYPE_NOSTITCH: number;\n    readonly SVG_STITCHTYPE_STITCH: number;\n    readonly SVG_STITCHTYPE_UNKNOWN: number;\n    readonly SVG_TURBULENCE_TYPE_FRACTALNOISE: number;\n    readonly SVG_TURBULENCE_TYPE_TURBULENCE: number;\n    readonly SVG_TURBULENCE_TYPE_UNKNOWN: number;\n};\n\n/** Provides access to the properties of <filter> elements, as well as methods to manipulate them. */\ninterface SVGFilterElement extends SVGElement, SVGURIReference {\n    readonly filterUnits: SVGAnimatedEnumeration;\n    readonly height: SVGAnimatedLength;\n    readonly primitiveUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFilterElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGFilterElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGFilterElement: {\n    prototype: SVGFilterElement;\n    new(): SVGFilterElement;\n};\n\ninterface SVGFilterPrimitiveStandardAttributes {\n    readonly height: SVGAnimatedLength;\n    readonly result: SVGAnimatedString;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n}\n\ninterface SVGFitToViewBox {\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    readonly viewBox: SVGAnimatedRect;\n}\n\n/** Provides access to the properties of <foreignObject> elements, as well as methods to manipulate them. */\ninterface SVGForeignObjectElement extends SVGGraphicsElement {\n    readonly height: SVGAnimatedLength;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGForeignObjectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGForeignObjectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGForeignObjectElement: {\n    prototype: SVGForeignObjectElement;\n    new(): SVGForeignObjectElement;\n};\n\n/** Corresponds to the <g> element. */\ninterface SVGGElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGElement: {\n    prototype: SVGGElement;\n    new(): SVGGElement;\n};\n\ninterface SVGGeometryElement extends SVGGraphicsElement {\n    readonly pathLength: SVGAnimatedNumber;\n    getPointAtLength(distance: number): DOMPoint;\n    getTotalLength(): number;\n    isPointInFill(point?: DOMPointInit): boolean;\n    isPointInStroke(point?: DOMPointInit): boolean;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGeometryElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGeometryElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGeometryElement: {\n    prototype: SVGGeometryElement;\n    new(): SVGGeometryElement;\n};\n\n/** The SVGGradient interface is a base interface used by SVGLinearGradientElement and SVGRadialGradientElement. */\ninterface SVGGradientElement extends SVGElement, SVGURIReference {\n    readonly gradientTransform: SVGAnimatedTransformList;\n    readonly gradientUnits: SVGAnimatedEnumeration;\n    readonly spreadMethod: SVGAnimatedEnumeration;\n    readonly SVG_SPREADMETHOD_PAD: number;\n    readonly SVG_SPREADMETHOD_REFLECT: number;\n    readonly SVG_SPREADMETHOD_REPEAT: number;\n    readonly SVG_SPREADMETHOD_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGradientElement: {\n    prototype: SVGGradientElement;\n    new(): SVGGradientElement;\n    readonly SVG_SPREADMETHOD_PAD: number;\n    readonly SVG_SPREADMETHOD_REFLECT: number;\n    readonly SVG_SPREADMETHOD_REPEAT: number;\n    readonly SVG_SPREADMETHOD_UNKNOWN: number;\n};\n\n/** SVG elements whose primary purpose is to directly render graphics into a group. */\ninterface SVGGraphicsElement extends SVGElement, SVGTests {\n    readonly transform: SVGAnimatedTransformList;\n    getBBox(options?: SVGBoundingBoxOptions): DOMRect;\n    getCTM(): DOMMatrix | null;\n    getScreenCTM(): DOMMatrix | null;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGraphicsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGGraphicsElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGGraphicsElement: {\n    prototype: SVGGraphicsElement;\n    new(): SVGGraphicsElement;\n};\n\n/** Corresponds to the <image> element. */\ninterface SVGImageElement extends SVGGraphicsElement, SVGURIReference {\n    readonly height: SVGAnimatedLength;\n    readonly preserveAspectRatio: SVGAnimatedPreserveAspectRatio;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGImageElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGImageElement: {\n    prototype: SVGImageElement;\n    new(): SVGImageElement;\n};\n\n/** Correspond to the <length> basic data type. */\ninterface SVGLength {\n    readonly unitType: number;\n    value: number;\n    valueAsString: string;\n    valueInSpecifiedUnits: number;\n    convertToSpecifiedUnits(unitType: number): void;\n    newValueSpecifiedUnits(unitType: number, valueInSpecifiedUnits: number): void;\n    readonly SVG_LENGTHTYPE_CM: number;\n    readonly SVG_LENGTHTYPE_EMS: number;\n    readonly SVG_LENGTHTYPE_EXS: number;\n    readonly SVG_LENGTHTYPE_IN: number;\n    readonly SVG_LENGTHTYPE_MM: number;\n    readonly SVG_LENGTHTYPE_NUMBER: number;\n    readonly SVG_LENGTHTYPE_PC: number;\n    readonly SVG_LENGTHTYPE_PERCENTAGE: number;\n    readonly SVG_LENGTHTYPE_PT: number;\n    readonly SVG_LENGTHTYPE_PX: number;\n    readonly SVG_LENGTHTYPE_UNKNOWN: number;\n}\n\ndeclare var SVGLength: {\n    prototype: SVGLength;\n    new(): SVGLength;\n    readonly SVG_LENGTHTYPE_CM: number;\n    readonly SVG_LENGTHTYPE_EMS: number;\n    readonly SVG_LENGTHTYPE_EXS: number;\n    readonly SVG_LENGTHTYPE_IN: number;\n    readonly SVG_LENGTHTYPE_MM: number;\n    readonly SVG_LENGTHTYPE_NUMBER: number;\n    readonly SVG_LENGTHTYPE_PC: number;\n    readonly SVG_LENGTHTYPE_PERCENTAGE: number;\n    readonly SVG_LENGTHTYPE_PT: number;\n    readonly SVG_LENGTHTYPE_PX: number;\n    readonly SVG_LENGTHTYPE_UNKNOWN: number;\n};\n\n/** The SVGLengthList defines a list of SVGLength objects. */\ninterface SVGLengthList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGLength): SVGLength;\n    clear(): void;\n    getItem(index: number): SVGLength;\n    initialize(newItem: SVGLength): SVGLength;\n    insertItemBefore(newItem: SVGLength, index: number): SVGLength;\n    removeItem(index: number): SVGLength;\n    replaceItem(newItem: SVGLength, index: number): SVGLength;\n    [index: number]: SVGLength;\n}\n\ndeclare var SVGLengthList: {\n    prototype: SVGLengthList;\n    new(): SVGLengthList;\n};\n\n/** Provides access to the properties of <line> elements, as well as methods to manipulate them. */\ninterface SVGLineElement extends SVGGeometryElement {\n    readonly x1: SVGAnimatedLength;\n    readonly x2: SVGAnimatedLength;\n    readonly y1: SVGAnimatedLength;\n    readonly y2: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGLineElement: {\n    prototype: SVGLineElement;\n    new(): SVGLineElement;\n};\n\n/** Corresponds to the <linearGradient> element. */\ninterface SVGLinearGradientElement extends SVGGradientElement {\n    readonly x1: SVGAnimatedLength;\n    readonly x2: SVGAnimatedLength;\n    readonly y1: SVGAnimatedLength;\n    readonly y2: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLinearGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGLinearGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGLinearGradientElement: {\n    prototype: SVGLinearGradientElement;\n    new(): SVGLinearGradientElement;\n};\n\ninterface SVGMarkerElement extends SVGElement, SVGFitToViewBox {\n    readonly markerHeight: SVGAnimatedLength;\n    readonly markerUnits: SVGAnimatedEnumeration;\n    readonly markerWidth: SVGAnimatedLength;\n    readonly orientAngle: SVGAnimatedAngle;\n    readonly orientType: SVGAnimatedEnumeration;\n    readonly refX: SVGAnimatedLength;\n    readonly refY: SVGAnimatedLength;\n    setOrientToAngle(angle: SVGAngle): void;\n    setOrientToAuto(): void;\n    readonly SVG_MARKERUNITS_STROKEWIDTH: number;\n    readonly SVG_MARKERUNITS_UNKNOWN: number;\n    readonly SVG_MARKERUNITS_USERSPACEONUSE: number;\n    readonly SVG_MARKER_ORIENT_ANGLE: number;\n    readonly SVG_MARKER_ORIENT_AUTO: number;\n    readonly SVG_MARKER_ORIENT_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMarkerElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMarkerElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMarkerElement: {\n    prototype: SVGMarkerElement;\n    new(): SVGMarkerElement;\n    readonly SVG_MARKERUNITS_STROKEWIDTH: number;\n    readonly SVG_MARKERUNITS_UNKNOWN: number;\n    readonly SVG_MARKERUNITS_USERSPACEONUSE: number;\n    readonly SVG_MARKER_ORIENT_ANGLE: number;\n    readonly SVG_MARKER_ORIENT_AUTO: number;\n    readonly SVG_MARKER_ORIENT_UNKNOWN: number;\n};\n\n/** Provides access to the properties of <mask> elements, as well as methods to manipulate them. */\ninterface SVGMaskElement extends SVGElement {\n    readonly height: SVGAnimatedLength;\n    readonly maskContentUnits: SVGAnimatedEnumeration;\n    readonly maskUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMaskElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMaskElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMaskElement: {\n    prototype: SVGMaskElement;\n    new(): SVGMaskElement;\n};\n\n/** Corresponds to the <metadata> element. */\ninterface SVGMetadataElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMetadataElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGMetadataElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGMetadataElement: {\n    prototype: SVGMetadataElement;\n    new(): SVGMetadataElement;\n};\n\n/** Corresponds to the <number> basic data type. */\ninterface SVGNumber {\n    value: number;\n}\n\ndeclare var SVGNumber: {\n    prototype: SVGNumber;\n    new(): SVGNumber;\n};\n\n/** The SVGNumberList defines a list of SVGNumber objects. */\ninterface SVGNumberList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGNumber): SVGNumber;\n    clear(): void;\n    getItem(index: number): SVGNumber;\n    initialize(newItem: SVGNumber): SVGNumber;\n    insertItemBefore(newItem: SVGNumber, index: number): SVGNumber;\n    removeItem(index: number): SVGNumber;\n    replaceItem(newItem: SVGNumber, index: number): SVGNumber;\n    [index: number]: SVGNumber;\n}\n\ndeclare var SVGNumberList: {\n    prototype: SVGNumberList;\n    new(): SVGNumberList;\n};\n\n/** Corresponds to the <path> element. */\ninterface SVGPathElement extends SVGGraphicsElement {\n    /** @deprecated */\n    readonly pathSegList: SVGPathSegList;\n    /** @deprecated */\n    createSVGPathSegArcAbs(x: number, y: number, r1: number, r2: number, angle: number, largeArcFlag: boolean, sweepFlag: boolean): SVGPathSegArcAbs;\n    /** @deprecated */\n    createSVGPathSegArcRel(x: number, y: number, r1: number, r2: number, angle: number, largeArcFlag: boolean, sweepFlag: boolean): SVGPathSegArcRel;\n    /** @deprecated */\n    createSVGPathSegClosePath(): SVGPathSegClosePath;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicAbs(x: number, y: number, x1: number, y1: number, x2: number, y2: number): SVGPathSegCurvetoCubicAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicRel(x: number, y: number, x1: number, y1: number, x2: number, y2: number): SVGPathSegCurvetoCubicRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicSmoothAbs(x: number, y: number, x2: number, y2: number): SVGPathSegCurvetoCubicSmoothAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoCubicSmoothRel(x: number, y: number, x2: number, y2: number): SVGPathSegCurvetoCubicSmoothRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticAbs(x: number, y: number, x1: number, y1: number): SVGPathSegCurvetoQuadraticAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticRel(x: number, y: number, x1: number, y1: number): SVGPathSegCurvetoQuadraticRel;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticSmoothAbs(x: number, y: number): SVGPathSegCurvetoQuadraticSmoothAbs;\n    /** @deprecated */\n    createSVGPathSegCurvetoQuadraticSmoothRel(x: number, y: number): SVGPathSegCurvetoQuadraticSmoothRel;\n    /** @deprecated */\n    createSVGPathSegLinetoAbs(x: number, y: number): SVGPathSegLinetoAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoHorizontalAbs(x: number): SVGPathSegLinetoHorizontalAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoHorizontalRel(x: number): SVGPathSegLinetoHorizontalRel;\n    /** @deprecated */\n    createSVGPathSegLinetoRel(x: number, y: number): SVGPathSegLinetoRel;\n    /** @deprecated */\n    createSVGPathSegLinetoVerticalAbs(y: number): SVGPathSegLinetoVerticalAbs;\n    /** @deprecated */\n    createSVGPathSegLinetoVerticalRel(y: number): SVGPathSegLinetoVerticalRel;\n    /** @deprecated */\n    createSVGPathSegMovetoAbs(x: number, y: number): SVGPathSegMovetoAbs;\n    /** @deprecated */\n    createSVGPathSegMovetoRel(x: number, y: number): SVGPathSegMovetoRel;\n    /** @deprecated */\n    getPathSegAtLength(distance: number): number;\n    getPointAtLength(distance: number): SVGPoint;\n    getTotalLength(): number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPathElement: {\n    prototype: SVGPathElement;\n    new(): SVGPathElement;\n};\n\ninterface SVGPathSeg {\n    readonly pathSegType: number;\n    readonly pathSegTypeAsLetter: string;\n    readonly PATHSEG_ARC_ABS: number;\n    readonly PATHSEG_ARC_REL: number;\n    readonly PATHSEG_CLOSEPATH: number;\n    readonly PATHSEG_CURVETO_CUBIC_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_REL: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL: number;\n    readonly PATHSEG_LINETO_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_REL: number;\n    readonly PATHSEG_LINETO_REL: number;\n    readonly PATHSEG_LINETO_VERTICAL_ABS: number;\n    readonly PATHSEG_LINETO_VERTICAL_REL: number;\n    readonly PATHSEG_MOVETO_ABS: number;\n    readonly PATHSEG_MOVETO_REL: number;\n    readonly PATHSEG_UNKNOWN: number;\n}\n\ndeclare var SVGPathSeg: {\n    prototype: SVGPathSeg;\n    new(): SVGPathSeg;\n    readonly PATHSEG_ARC_ABS: number;\n    readonly PATHSEG_ARC_REL: number;\n    readonly PATHSEG_CLOSEPATH: number;\n    readonly PATHSEG_CURVETO_CUBIC_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_REL: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_CUBIC_SMOOTH_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_REL: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS: number;\n    readonly PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL: number;\n    readonly PATHSEG_LINETO_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_ABS: number;\n    readonly PATHSEG_LINETO_HORIZONTAL_REL: number;\n    readonly PATHSEG_LINETO_REL: number;\n    readonly PATHSEG_LINETO_VERTICAL_ABS: number;\n    readonly PATHSEG_LINETO_VERTICAL_REL: number;\n    readonly PATHSEG_MOVETO_ABS: number;\n    readonly PATHSEG_MOVETO_REL: number;\n    readonly PATHSEG_UNKNOWN: number;\n};\n\ninterface SVGPathSegArcAbs extends SVGPathSeg {\n    angle: number;\n    largeArcFlag: boolean;\n    r1: number;\n    r2: number;\n    sweepFlag: boolean;\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegArcAbs: {\n    prototype: SVGPathSegArcAbs;\n    new(): SVGPathSegArcAbs;\n};\n\ninterface SVGPathSegArcRel extends SVGPathSeg {\n    angle: number;\n    largeArcFlag: boolean;\n    r1: number;\n    r2: number;\n    sweepFlag: boolean;\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegArcRel: {\n    prototype: SVGPathSegArcRel;\n    new(): SVGPathSegArcRel;\n};\n\ninterface SVGPathSegClosePath extends SVGPathSeg {\n}\n\ndeclare var SVGPathSegClosePath: {\n    prototype: SVGPathSegClosePath;\n    new(): SVGPathSegClosePath;\n};\n\ninterface SVGPathSegCurvetoCubicAbs extends SVGPathSeg {\n    x: number;\n    x1: number;\n    x2: number;\n    y: number;\n    y1: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicAbs: {\n    prototype: SVGPathSegCurvetoCubicAbs;\n    new(): SVGPathSegCurvetoCubicAbs;\n};\n\ninterface SVGPathSegCurvetoCubicRel extends SVGPathSeg {\n    x: number;\n    x1: number;\n    x2: number;\n    y: number;\n    y1: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicRel: {\n    prototype: SVGPathSegCurvetoCubicRel;\n    new(): SVGPathSegCurvetoCubicRel;\n};\n\ninterface SVGPathSegCurvetoCubicSmoothAbs extends SVGPathSeg {\n    x: number;\n    x2: number;\n    y: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicSmoothAbs: {\n    prototype: SVGPathSegCurvetoCubicSmoothAbs;\n    new(): SVGPathSegCurvetoCubicSmoothAbs;\n};\n\ninterface SVGPathSegCurvetoCubicSmoothRel extends SVGPathSeg {\n    x: number;\n    x2: number;\n    y: number;\n    y2: number;\n}\n\ndeclare var SVGPathSegCurvetoCubicSmoothRel: {\n    prototype: SVGPathSegCurvetoCubicSmoothRel;\n    new(): SVGPathSegCurvetoCubicSmoothRel;\n};\n\ninterface SVGPathSegCurvetoQuadraticAbs extends SVGPathSeg {\n    x: number;\n    x1: number;\n    y: number;\n    y1: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticAbs: {\n    prototype: SVGPathSegCurvetoQuadraticAbs;\n    new(): SVGPathSegCurvetoQuadraticAbs;\n};\n\ninterface SVGPathSegCurvetoQuadraticRel extends SVGPathSeg {\n    x: number;\n    x1: number;\n    y: number;\n    y1: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticRel: {\n    prototype: SVGPathSegCurvetoQuadraticRel;\n    new(): SVGPathSegCurvetoQuadraticRel;\n};\n\ninterface SVGPathSegCurvetoQuadraticSmoothAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticSmoothAbs: {\n    prototype: SVGPathSegCurvetoQuadraticSmoothAbs;\n    new(): SVGPathSegCurvetoQuadraticSmoothAbs;\n};\n\ninterface SVGPathSegCurvetoQuadraticSmoothRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegCurvetoQuadraticSmoothRel: {\n    prototype: SVGPathSegCurvetoQuadraticSmoothRel;\n    new(): SVGPathSegCurvetoQuadraticSmoothRel;\n};\n\ninterface SVGPathSegLinetoAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoAbs: {\n    prototype: SVGPathSegLinetoAbs;\n    new(): SVGPathSegLinetoAbs;\n};\n\ninterface SVGPathSegLinetoHorizontalAbs extends SVGPathSeg {\n    x: number;\n}\n\ndeclare var SVGPathSegLinetoHorizontalAbs: {\n    prototype: SVGPathSegLinetoHorizontalAbs;\n    new(): SVGPathSegLinetoHorizontalAbs;\n};\n\ninterface SVGPathSegLinetoHorizontalRel extends SVGPathSeg {\n    x: number;\n}\n\ndeclare var SVGPathSegLinetoHorizontalRel: {\n    prototype: SVGPathSegLinetoHorizontalRel;\n    new(): SVGPathSegLinetoHorizontalRel;\n};\n\ninterface SVGPathSegLinetoRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoRel: {\n    prototype: SVGPathSegLinetoRel;\n    new(): SVGPathSegLinetoRel;\n};\n\ninterface SVGPathSegLinetoVerticalAbs extends SVGPathSeg {\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoVerticalAbs: {\n    prototype: SVGPathSegLinetoVerticalAbs;\n    new(): SVGPathSegLinetoVerticalAbs;\n};\n\ninterface SVGPathSegLinetoVerticalRel extends SVGPathSeg {\n    y: number;\n}\n\ndeclare var SVGPathSegLinetoVerticalRel: {\n    prototype: SVGPathSegLinetoVerticalRel;\n    new(): SVGPathSegLinetoVerticalRel;\n};\n\ninterface SVGPathSegList {\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGPathSeg): SVGPathSeg;\n    clear(): void;\n    getItem(index: number): SVGPathSeg;\n    initialize(newItem: SVGPathSeg): SVGPathSeg;\n    insertItemBefore(newItem: SVGPathSeg, index: number): SVGPathSeg;\n    removeItem(index: number): SVGPathSeg;\n    replaceItem(newItem: SVGPathSeg, index: number): SVGPathSeg;\n}\n\ndeclare var SVGPathSegList: {\n    prototype: SVGPathSegList;\n    new(): SVGPathSegList;\n};\n\ninterface SVGPathSegMovetoAbs extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegMovetoAbs: {\n    prototype: SVGPathSegMovetoAbs;\n    new(): SVGPathSegMovetoAbs;\n};\n\ninterface SVGPathSegMovetoRel extends SVGPathSeg {\n    x: number;\n    y: number;\n}\n\ndeclare var SVGPathSegMovetoRel: {\n    prototype: SVGPathSegMovetoRel;\n    new(): SVGPathSegMovetoRel;\n};\n\n/** Corresponds to the <pattern> element. */\ninterface SVGPatternElement extends SVGElement, SVGFitToViewBox, SVGURIReference {\n    readonly height: SVGAnimatedLength;\n    readonly patternContentUnits: SVGAnimatedEnumeration;\n    readonly patternTransform: SVGAnimatedTransformList;\n    readonly patternUnits: SVGAnimatedEnumeration;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPatternElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPatternElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPatternElement: {\n    prototype: SVGPatternElement;\n    new(): SVGPatternElement;\n};\n\ninterface SVGPointList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: DOMPoint): DOMPoint;\n    clear(): void;\n    getItem(index: number): DOMPoint;\n    initialize(newItem: DOMPoint): DOMPoint;\n    insertItemBefore(newItem: DOMPoint, index: number): DOMPoint;\n    removeItem(index: number): DOMPoint;\n    replaceItem(newItem: DOMPoint, index: number): DOMPoint;\n    [index: number]: DOMPoint;\n}\n\ndeclare var SVGPointList: {\n    prototype: SVGPointList;\n    new(): SVGPointList;\n};\n\n/** Provides access to the properties of <polygon> elements, as well as methods to manipulate them. */\ninterface SVGPolygonElement extends SVGGeometryElement, SVGAnimatedPoints {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolygonElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolygonElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPolygonElement: {\n    prototype: SVGPolygonElement;\n    new(): SVGPolygonElement;\n};\n\n/** Provides access to the properties of <polyline> elements, as well as methods to manipulate them. */\ninterface SVGPolylineElement extends SVGGeometryElement, SVGAnimatedPoints {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolylineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGPolylineElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGPolylineElement: {\n    prototype: SVGPolylineElement;\n    new(): SVGPolylineElement;\n};\n\n/** Corresponds to the preserveAspectRatio attribute, which is available for some of SVG\'s elements. */\ninterface SVGPreserveAspectRatio {\n    align: number;\n    meetOrSlice: number;\n    readonly SVG_MEETORSLICE_MEET: number;\n    readonly SVG_MEETORSLICE_SLICE: number;\n    readonly SVG_MEETORSLICE_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_NONE: number;\n    readonly SVG_PRESERVEASPECTRATIO_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMIN: number;\n}\n\ndeclare var SVGPreserveAspectRatio: {\n    prototype: SVGPreserveAspectRatio;\n    new(): SVGPreserveAspectRatio;\n    readonly SVG_MEETORSLICE_MEET: number;\n    readonly SVG_MEETORSLICE_SLICE: number;\n    readonly SVG_MEETORSLICE_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_NONE: number;\n    readonly SVG_PRESERVEASPECTRATIO_UNKNOWN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMAXYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMIDYMIN: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMAX: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMID: number;\n    readonly SVG_PRESERVEASPECTRATIO_XMINYMIN: number;\n};\n\n/** Corresponds to the <RadialGradient> element. */\ninterface SVGRadialGradientElement extends SVGGradientElement {\n    readonly cx: SVGAnimatedLength;\n    readonly cy: SVGAnimatedLength;\n    readonly fr: SVGAnimatedLength;\n    readonly fx: SVGAnimatedLength;\n    readonly fy: SVGAnimatedLength;\n    readonly r: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRadialGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRadialGradientElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGRadialGradientElement: {\n    prototype: SVGRadialGradientElement;\n    new(): SVGRadialGradientElement;\n};\n\n/** Provides access to the properties of <rect> elements, as well as methods to manipulate them. */\ninterface SVGRectElement extends SVGGeometryElement {\n    readonly height: SVGAnimatedLength;\n    readonly rx: SVGAnimatedLength;\n    readonly ry: SVGAnimatedLength;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGRectElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGRectElement: {\n    prototype: SVGRectElement;\n    new(): SVGRectElement;\n};\n\ninterface SVGSVGElementEventMap extends SVGElementEventMap {\n    "SVGUnload": Event;\n    "SVGZoom": SVGZoomEvent;\n}\n\n/** Provides access to the properties of <svg> elements, as well as methods to manipulate them. This interface contains also various miscellaneous commonly-used utility methods, such as matrix operations and the ability to control the time of redraw on visual rendering devices. */\ninterface SVGSVGElement extends SVGGraphicsElement, DocumentEvent, SVGFitToViewBox, SVGZoomAndPan {\n    /** @deprecated */\n    contentScriptType: string;\n    /** @deprecated */\n    contentStyleType: string;\n    currentScale: number;\n    readonly currentTranslate: SVGPoint;\n    readonly height: SVGAnimatedLength;\n    onunload: ((this: SVGSVGElement, ev: Event) => any) | null;\n    onzoom: ((this: SVGSVGElement, ev: SVGZoomEvent) => any) | null;\n    /** @deprecated */\n    readonly pixelUnitToMillimeterX: number;\n    /** @deprecated */\n    readonly pixelUnitToMillimeterY: number;\n    /** @deprecated */\n    readonly screenPixelToMillimeterX: number;\n    /** @deprecated */\n    readonly screenPixelToMillimeterY: number;\n    /** @deprecated */\n    readonly viewport: SVGRect;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    animationsPaused(): boolean;\n    checkEnclosure(element: SVGElement, rect: SVGRect): boolean;\n    checkIntersection(element: SVGElement, rect: SVGRect): boolean;\n    createSVGAngle(): SVGAngle;\n    createSVGLength(): SVGLength;\n    createSVGMatrix(): SVGMatrix;\n    createSVGNumber(): SVGNumber;\n    createSVGPoint(): SVGPoint;\n    createSVGRect(): SVGRect;\n    createSVGTransform(): SVGTransform;\n    createSVGTransformFromMatrix(matrix: SVGMatrix): SVGTransform;\n    deselectAll(): void;\n    /** @deprecated */\n    forceRedraw(): void;\n    getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\n    getCurrentTime(): number;\n    getElementById(elementId: string): Element;\n    getEnclosureList(rect: SVGRect, referenceElement: SVGElement | null): NodeListOf<SVGCircleElement | SVGEllipseElement | SVGImageElement | SVGLineElement | SVGPathElement | SVGPolygonElement | SVGPolylineElement | SVGRectElement | SVGTextElement | SVGUseElement>;\n    getIntersectionList(rect: SVGRect, referenceElement: SVGElement | null): NodeListOf<SVGCircleElement | SVGEllipseElement | SVGImageElement | SVGLineElement | SVGPathElement | SVGPolygonElement | SVGPolylineElement | SVGRectElement | SVGTextElement | SVGUseElement>;\n    pauseAnimations(): void;\n    setCurrentTime(seconds: number): void;\n    /** @deprecated */\n    suspendRedraw(maxWaitMilliseconds: number): number;\n    unpauseAnimations(): void;\n    /** @deprecated */\n    unsuspendRedraw(suspendHandleID: number): void;\n    /** @deprecated */\n    unsuspendRedrawAll(): void;\n    addEventListener<K extends keyof SVGSVGElementEventMap>(type: K, listener: (this: SVGSVGElement, ev: SVGSVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGSVGElementEventMap>(type: K, listener: (this: SVGSVGElement, ev: SVGSVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSVGElement: {\n    prototype: SVGSVGElement;\n    new(): SVGSVGElement;\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\n/** Corresponds to the SVG <script> element. */\ninterface SVGScriptElement extends SVGElement, SVGURIReference {\n    type: string;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGScriptElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGScriptElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGScriptElement: {\n    prototype: SVGScriptElement;\n    new(): SVGScriptElement;\n};\n\n/** Corresponds to the <stop> element. */\ninterface SVGStopElement extends SVGElement {\n    readonly offset: SVGAnimatedNumber;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStopElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStopElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGStopElement: {\n    prototype: SVGStopElement;\n    new(): SVGStopElement;\n};\n\n/** The SVGStringList defines a list of DOMString objects. */\ninterface SVGStringList {\n    readonly length: number;\n    readonly numberOfItems: number;\n    appendItem(newItem: string): string;\n    clear(): void;\n    getItem(index: number): string;\n    initialize(newItem: string): string;\n    insertItemBefore(newItem: string, index: number): string;\n    removeItem(index: number): string;\n    replaceItem(newItem: string, index: number): string;\n    [index: number]: string;\n}\n\ndeclare var SVGStringList: {\n    prototype: SVGStringList;\n    new(): SVGStringList;\n};\n\n/** Corresponds to the SVG <style> element. */\ninterface SVGStyleElement extends SVGElement {\n    disabled: boolean;\n    media: string;\n    title: string;\n    type: string;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStyleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGStyleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGStyleElement: {\n    prototype: SVGStyleElement;\n    new(): SVGStyleElement;\n};\n\n/** Corresponds to the <switch> element. */\ninterface SVGSwitchElement extends SVGGraphicsElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSwitchElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSwitchElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSwitchElement: {\n    prototype: SVGSwitchElement;\n    new(): SVGSwitchElement;\n};\n\n/** Corresponds to the <symbol> element. */\ninterface SVGSymbolElement extends SVGElement, SVGFitToViewBox {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSymbolElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGSymbolElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGSymbolElement: {\n    prototype: SVGSymbolElement;\n    new(): SVGSymbolElement;\n};\n\n/** A <tspan> element. */\ninterface SVGTSpanElement extends SVGTextPositioningElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTSpanElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTSpanElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTSpanElement: {\n    prototype: SVGTSpanElement;\n    new(): SVGTSpanElement;\n};\n\ninterface SVGTests {\n    readonly requiredExtensions: SVGStringList;\n    readonly systemLanguage: SVGStringList;\n}\n\n/** Implemented by elements that support rendering child text content. It is inherited by various text-related interfaces, such as SVGTextElement, SVGTSpanElement, SVGTRefElement, SVGAltGlyphElement and SVGTextPathElement. */\ninterface SVGTextContentElement extends SVGGraphicsElement {\n    readonly lengthAdjust: SVGAnimatedEnumeration;\n    readonly textLength: SVGAnimatedLength;\n    getCharNumAtPosition(point?: DOMPointInit): number;\n    getComputedTextLength(): number;\n    getEndPositionOfChar(charnum: number): DOMPoint;\n    getExtentOfChar(charnum: number): DOMRect;\n    getNumberOfChars(): number;\n    getRotationOfChar(charnum: number): number;\n    getStartPositionOfChar(charnum: number): DOMPoint;\n    getSubStringLength(charnum: number, nchars: number): number;\n    selectSubString(charnum: number, nchars: number): void;\n    readonly LENGTHADJUST_SPACING: number;\n    readonly LENGTHADJUST_SPACINGANDGLYPHS: number;\n    readonly LENGTHADJUST_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextContentElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextContentElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextContentElement: {\n    prototype: SVGTextContentElement;\n    new(): SVGTextContentElement;\n    readonly LENGTHADJUST_SPACING: number;\n    readonly LENGTHADJUST_SPACINGANDGLYPHS: number;\n    readonly LENGTHADJUST_UNKNOWN: number;\n};\n\n/** Corresponds to the <text> elements. */\ninterface SVGTextElement extends SVGTextPositioningElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextElement: {\n    prototype: SVGTextElement;\n    new(): SVGTextElement;\n};\n\n/** Corresponds to the <textPath> element. */\ninterface SVGTextPathElement extends SVGTextContentElement, SVGURIReference {\n    readonly method: SVGAnimatedEnumeration;\n    readonly spacing: SVGAnimatedEnumeration;\n    readonly startOffset: SVGAnimatedLength;\n    readonly TEXTPATH_METHODTYPE_ALIGN: number;\n    readonly TEXTPATH_METHODTYPE_STRETCH: number;\n    readonly TEXTPATH_METHODTYPE_UNKNOWN: number;\n    readonly TEXTPATH_SPACINGTYPE_AUTO: number;\n    readonly TEXTPATH_SPACINGTYPE_EXACT: number;\n    readonly TEXTPATH_SPACINGTYPE_UNKNOWN: number;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPathElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextPathElement: {\n    prototype: SVGTextPathElement;\n    new(): SVGTextPathElement;\n    readonly TEXTPATH_METHODTYPE_ALIGN: number;\n    readonly TEXTPATH_METHODTYPE_STRETCH: number;\n    readonly TEXTPATH_METHODTYPE_UNKNOWN: number;\n    readonly TEXTPATH_SPACINGTYPE_AUTO: number;\n    readonly TEXTPATH_SPACINGTYPE_EXACT: number;\n    readonly TEXTPATH_SPACINGTYPE_UNKNOWN: number;\n};\n\n/** Implemented by elements that support attributes that position individual text glyphs. It is inherited by SVGTextElement, SVGTSpanElement, SVGTRefElement and SVGAltGlyphElement. */\ninterface SVGTextPositioningElement extends SVGTextContentElement {\n    readonly dx: SVGAnimatedLengthList;\n    readonly dy: SVGAnimatedLengthList;\n    readonly rotate: SVGAnimatedNumberList;\n    readonly x: SVGAnimatedLengthList;\n    readonly y: SVGAnimatedLengthList;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPositioningElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTextPositioningElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTextPositioningElement: {\n    prototype: SVGTextPositioningElement;\n    new(): SVGTextPositioningElement;\n};\n\n/** Corresponds to the <title> element. */\ninterface SVGTitleElement extends SVGElement {\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTitleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGTitleElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGTitleElement: {\n    prototype: SVGTitleElement;\n    new(): SVGTitleElement;\n};\n\n/** SVGTransform is the interface for one of the component transformations within an SVGTransformList; thus, an SVGTransform object corresponds to a single component (e.g., scale(…) or matrix(…)) within a transform attribute. */\ninterface SVGTransform {\n    readonly angle: number;\n    readonly matrix: SVGMatrix;\n    readonly type: number;\n    setMatrix(matrix: SVGMatrix): void;\n    setRotate(angle: number, cx: number, cy: number): void;\n    setScale(sx: number, sy: number): void;\n    setSkewX(angle: number): void;\n    setSkewY(angle: number): void;\n    setTranslate(tx: number, ty: number): void;\n    readonly SVG_TRANSFORM_MATRIX: number;\n    readonly SVG_TRANSFORM_ROTATE: number;\n    readonly SVG_TRANSFORM_SCALE: number;\n    readonly SVG_TRANSFORM_SKEWX: number;\n    readonly SVG_TRANSFORM_SKEWY: number;\n    readonly SVG_TRANSFORM_TRANSLATE: number;\n    readonly SVG_TRANSFORM_UNKNOWN: number;\n}\n\ndeclare var SVGTransform: {\n    prototype: SVGTransform;\n    new(): SVGTransform;\n    readonly SVG_TRANSFORM_MATRIX: number;\n    readonly SVG_TRANSFORM_ROTATE: number;\n    readonly SVG_TRANSFORM_SCALE: number;\n    readonly SVG_TRANSFORM_SKEWX: number;\n    readonly SVG_TRANSFORM_SKEWY: number;\n    readonly SVG_TRANSFORM_TRANSLATE: number;\n    readonly SVG_TRANSFORM_UNKNOWN: number;\n};\n\n/** The SVGTransformList defines a list of SVGTransform objects. */\ninterface SVGTransformList {\n    readonly numberOfItems: number;\n    appendItem(newItem: SVGTransform): SVGTransform;\n    clear(): void;\n    consolidate(): SVGTransform;\n    createSVGTransformFromMatrix(matrix: SVGMatrix): SVGTransform;\n    getItem(index: number): SVGTransform;\n    initialize(newItem: SVGTransform): SVGTransform;\n    insertItemBefore(newItem: SVGTransform, index: number): SVGTransform;\n    removeItem(index: number): SVGTransform;\n    replaceItem(newItem: SVGTransform, index: number): SVGTransform;\n}\n\ndeclare var SVGTransformList: {\n    prototype: SVGTransformList;\n    new(): SVGTransformList;\n};\n\ninterface SVGURIReference {\n    readonly href: SVGAnimatedString;\n}\n\n/** A commonly used set of constants used for reflecting gradientUnits, patternContentUnits and other similar attributes. */\ninterface SVGUnitTypes {\n    readonly SVG_UNIT_TYPE_OBJECTBOUNDINGBOX: number;\n    readonly SVG_UNIT_TYPE_UNKNOWN: number;\n    readonly SVG_UNIT_TYPE_USERSPACEONUSE: number;\n}\n\ndeclare var SVGUnitTypes: {\n    prototype: SVGUnitTypes;\n    new(): SVGUnitTypes;\n    readonly SVG_UNIT_TYPE_OBJECTBOUNDINGBOX: number;\n    readonly SVG_UNIT_TYPE_UNKNOWN: number;\n    readonly SVG_UNIT_TYPE_USERSPACEONUSE: number;\n};\n\n/** Corresponds to the <use> element. */\ninterface SVGUseElement extends SVGGraphicsElement, SVGURIReference {\n    readonly animatedInstanceRoot: SVGElementInstance | null;\n    readonly height: SVGAnimatedLength;\n    readonly instanceRoot: SVGElementInstance | null;\n    readonly width: SVGAnimatedLength;\n    readonly x: SVGAnimatedLength;\n    readonly y: SVGAnimatedLength;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGUseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGUseElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGUseElement: {\n    prototype: SVGUseElement;\n    new(): SVGUseElement;\n};\n\n/** Provides access to the properties of <view> elements, as well as methods to manipulate them. */\ninterface SVGViewElement extends SVGElement, SVGFitToViewBox, SVGZoomAndPan {\n    /** @deprecated */\n    readonly viewTarget: SVGStringList;\n    addEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGViewElement, ev: SVGElementEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SVGElementEventMap>(type: K, listener: (this: SVGViewElement, ev: SVGElementEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SVGViewElement: {\n    prototype: SVGViewElement;\n    new(): SVGViewElement;\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\n/** Used to reflect the zoomAndPan attribute, and is mixed in to other interfaces for elements that support this attribute. */\ninterface SVGZoomAndPan {\n    readonly zoomAndPan: number;\n}\n\ndeclare var SVGZoomAndPan: {\n    readonly SVG_ZOOMANDPAN_DISABLE: number;\n    readonly SVG_ZOOMANDPAN_MAGNIFY: number;\n    readonly SVG_ZOOMANDPAN_UNKNOWN: number;\n};\n\ninterface SVGZoomEvent extends UIEvent {\n    readonly newScale: number;\n    readonly newTranslate: SVGPoint;\n    readonly previousScale: number;\n    readonly previousTranslate: SVGPoint;\n    readonly zoomRectScreen: SVGRect;\n}\n\ndeclare var SVGZoomEvent: {\n    prototype: SVGZoomEvent;\n    new(): SVGZoomEvent;\n};\n\ninterface ScopedCredential {\n    readonly id: ArrayBuffer;\n    readonly type: ScopedCredentialType;\n}\n\ndeclare var ScopedCredential: {\n    prototype: ScopedCredential;\n    new(): ScopedCredential;\n};\n\ninterface ScopedCredentialInfo {\n    readonly credential: ScopedCredential;\n    readonly publicKey: CryptoKey;\n}\n\ndeclare var ScopedCredentialInfo: {\n    prototype: ScopedCredentialInfo;\n    new(): ScopedCredentialInfo;\n};\n\n/** A screen, usually the one on which the current window is being rendered, and is obtained using window.screen. */\ninterface Screen {\n    readonly availHeight: number;\n    readonly availWidth: number;\n    readonly colorDepth: number;\n    readonly height: number;\n    readonly orientation: ScreenOrientation;\n    readonly pixelDepth: number;\n    readonly width: number;\n}\n\ndeclare var Screen: {\n    prototype: Screen;\n    new(): Screen;\n};\n\ninterface ScreenOrientationEventMap {\n    "change": Event;\n}\n\ninterface ScreenOrientation extends EventTarget {\n    readonly angle: number;\n    onchange: ((this: ScreenOrientation, ev: Event) => any) | null;\n    readonly type: OrientationType;\n    lock(orientation: OrientationLockType): Promise<void>;\n    unlock(): void;\n    addEventListener<K extends keyof ScreenOrientationEventMap>(type: K, listener: (this: ScreenOrientation, ev: ScreenOrientationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ScreenOrientationEventMap>(type: K, listener: (this: ScreenOrientation, ev: ScreenOrientationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ScreenOrientation: {\n    prototype: ScreenOrientation;\n    new(): ScreenOrientation;\n};\n\ninterface ScriptProcessorNodeEventMap {\n    "audioprocess": AudioProcessingEvent;\n}\n\n/** Allows the generation, processing, or analyzing of audio using JavaScript. */\ninterface ScriptProcessorNode extends AudioNode {\n    /** @deprecated */\n    readonly bufferSize: number;\n    /** @deprecated */\n    onaudioprocess: ((this: ScriptProcessorNode, ev: AudioProcessingEvent) => any) | null;\n    addEventListener<K extends keyof ScriptProcessorNodeEventMap>(type: K, listener: (this: ScriptProcessorNode, ev: ScriptProcessorNodeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ScriptProcessorNodeEventMap>(type: K, listener: (this: ScriptProcessorNode, ev: ScriptProcessorNodeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ScriptProcessorNode: {\n    prototype: ScriptProcessorNode;\n    new(): ScriptProcessorNode;\n};\n\n/** Inherits from Event, and represents the event object of an event sent on a document or worker when its content security policy is violated. */\ninterface SecurityPolicyViolationEvent extends Event {\n    readonly blockedURI: string;\n    readonly columnNumber: number;\n    readonly documentURI: string;\n    readonly effectiveDirective: string;\n    readonly lineNumber: number;\n    readonly originalPolicy: string;\n    readonly referrer: string;\n    readonly sourceFile: string;\n    readonly statusCode: number;\n    readonly violatedDirective: string;\n}\n\ndeclare var SecurityPolicyViolationEvent: {\n    prototype: SecurityPolicyViolationEvent;\n    new(type: string, eventInitDict?: SecurityPolicyViolationEventInit): SecurityPolicyViolationEvent;\n};\n\n/** A Selection object represents the range of text selected by the user or the current position of the caret. To obtain a Selection object for examination or modification, call Window.getSelection(). */\ninterface Selection {\n    readonly anchorNode: Node | null;\n    readonly anchorOffset: number;\n    readonly focusNode: Node | null;\n    readonly focusOffset: number;\n    readonly isCollapsed: boolean;\n    readonly rangeCount: number;\n    readonly type: string;\n    addRange(range: Range): void;\n    collapse(node: Node | null, offset?: number): void;\n    collapseToEnd(): void;\n    collapseToStart(): void;\n    containsNode(node: Node, allowPartialContainment?: boolean): boolean;\n    deleteFromDocument(): void;\n    empty(): void;\n    extend(node: Node, offset?: number): void;\n    getRangeAt(index: number): Range;\n    removeAllRanges(): void;\n    removeRange(range: Range): void;\n    selectAllChildren(node: Node): void;\n    setBaseAndExtent(anchorNode: Node, anchorOffset: number, focusNode: Node, focusOffset: number): void;\n    setPosition(node: Node | null, offset?: number): void;\n    toString(): string;\n}\n\ndeclare var Selection: {\n    prototype: Selection;\n    new(): Selection;\n    toString(): string;\n};\n\ninterface ServiceUIFrameContext {\n    getCachedFrameMessage(key: string): string;\n    postFrameMessage(key: string, data: string): void;\n}\ndeclare var ServiceUIFrameContext: ServiceUIFrameContext;\n\ninterface ServiceWorkerEventMap extends AbstractWorkerEventMap {\n    "statechange": Event;\n}\n\n/** This ServiceWorker API interface provides a reference to a service worker. Multiple browsing contexts (e.g. pages, workers, etc.) can be associated with the same service worker, each through a unique ServiceWorker object. */\ninterface ServiceWorker extends EventTarget, AbstractWorker {\n    onstatechange: ((this: ServiceWorker, ev: Event) => any) | null;\n    readonly scriptURL: string;\n    readonly state: ServiceWorkerState;\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    addEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorker: {\n    prototype: ServiceWorker;\n    new(): ServiceWorker;\n};\n\ninterface ServiceWorkerContainerEventMap {\n    "controllerchange": Event;\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** The ServiceWorkerContainer interface of the ServiceWorker API provides an object representing the service worker as an overall unit in the network ecosystem, including facilities to register, unregister and update service workers, and access the state of service workers and their registrations. */\ninterface ServiceWorkerContainer extends EventTarget {\n    readonly controller: ServiceWorker | null;\n    oncontrollerchange: ((this: ServiceWorkerContainer, ev: Event) => any) | null;\n    onmessage: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    readonly ready: Promise<ServiceWorkerRegistration>;\n    getRegistration(clientURL?: string): Promise<ServiceWorkerRegistration | undefined>;\n    getRegistrations(): Promise<ReadonlyArray<ServiceWorkerRegistration>>;\n    register(scriptURL: string, options?: RegistrationOptions): Promise<ServiceWorkerRegistration>;\n    startMessages(): void;\n    addEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerContainer: {\n    prototype: ServiceWorkerContainer;\n    new(): ServiceWorkerContainer;\n};\n\n/** This ServiceWorker API interface contains information about an event sent to a ServiceWorkerContainer target. This extends the default message event to allow setting a ServiceWorker object as the source of a message. The event object is accessed via the handler function of a message event, when fired by a message received from a service worker. */\ninterface ServiceWorkerMessageEvent extends Event {\n    readonly data: any;\n    readonly lastEventId: string;\n    readonly origin: string;\n    readonly ports: ReadonlyArray<MessagePort> | null;\n    readonly source: ServiceWorker | MessagePort | null;\n}\n\ndeclare var ServiceWorkerMessageEvent: {\n    prototype: ServiceWorkerMessageEvent;\n    new(type: string, eventInitDict?: ServiceWorkerMessageEventInit): ServiceWorkerMessageEvent;\n};\n\ninterface ServiceWorkerRegistrationEventMap {\n    "updatefound": Event;\n}\n\n/** This ServiceWorker API interface represents the service worker registration. You register a service worker to control one or more pages that share the same origin. */\ninterface ServiceWorkerRegistration extends EventTarget {\n    readonly active: ServiceWorker | null;\n    readonly installing: ServiceWorker | null;\n    readonly navigationPreload: NavigationPreloadManager;\n    onupdatefound: ((this: ServiceWorkerRegistration, ev: Event) => any) | null;\n    readonly pushManager: PushManager;\n    readonly scope: string;\n    readonly sync: SyncManager;\n    readonly updateViaCache: ServiceWorkerUpdateViaCache;\n    readonly waiting: ServiceWorker | null;\n    getNotifications(filter?: GetNotificationOptions): Promise<Notification[]>;\n    showNotification(title: string, options?: NotificationOptions): Promise<void>;\n    unregister(): Promise<boolean>;\n    update(): Promise<void>;\n    addEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerRegistration: {\n    prototype: ServiceWorkerRegistration;\n    new(): ServiceWorkerRegistration;\n};\n\ninterface ShadowRoot extends DocumentFragment, DocumentOrShadowRoot, InnerHTML {\n    readonly host: Element;\n    readonly mode: ShadowRootMode;\n    /**\n     * Throws a "NotSupportedError" DOMException if context object is a shadow root.\n     */\n}\n\ndeclare var ShadowRoot: {\n    prototype: ShadowRoot;\n    new(): ShadowRoot;\n};\n\ninterface SharedWorker extends EventTarget, AbstractWorker {\n    /**\n     * Returns sharedWorker\'s MessagePort object which can be used to communicate with the global environment.\n     */\n    readonly port: MessagePort;\n    addEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: SharedWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: SharedWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SharedWorker: {\n    prototype: SharedWorker;\n    new(scriptURL: string, options?: string | WorkerOptions): SharedWorker;\n};\n\ninterface Slottable {\n    readonly assignedSlot: HTMLSlotElement | null;\n}\n\ninterface SourceBufferEventMap {\n    "abort": Event;\n    "error": Event;\n    "update": Event;\n    "updateend": Event;\n    "updatestart": Event;\n}\n\n/** A chunk of media to be passed into an HTMLMediaElement and played, via a MediaSource object. This can be made up of one or several media segments. */\ninterface SourceBuffer extends EventTarget {\n    appendWindowEnd: number;\n    appendWindowStart: number;\n    readonly buffered: TimeRanges;\n    mode: AppendMode;\n    onabort: ((this: SourceBuffer, ev: Event) => any) | null;\n    onerror: ((this: SourceBuffer, ev: Event) => any) | null;\n    onupdate: ((this: SourceBuffer, ev: Event) => any) | null;\n    onupdateend: ((this: SourceBuffer, ev: Event) => any) | null;\n    onupdatestart: ((this: SourceBuffer, ev: Event) => any) | null;\n    timestampOffset: number;\n    readonly updating: boolean;\n    abort(): void;\n    appendBuffer(data: BufferSource): void;\n    remove(start: number, end: number): void;\n    addEventListener<K extends keyof SourceBufferEventMap>(type: K, listener: (this: SourceBuffer, ev: SourceBufferEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SourceBufferEventMap>(type: K, listener: (this: SourceBuffer, ev: SourceBufferEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SourceBuffer: {\n    prototype: SourceBuffer;\n    new(): SourceBuffer;\n};\n\ninterface SourceBufferListEventMap {\n    "addsourcebuffer": Event;\n    "removesourcebuffer": Event;\n}\n\n/** A simple container list for multiple SourceBuffer objects. */\ninterface SourceBufferList extends EventTarget {\n    readonly length: number;\n    onaddsourcebuffer: ((this: SourceBufferList, ev: Event) => any) | null;\n    onremovesourcebuffer: ((this: SourceBufferList, ev: Event) => any) | null;\n    addEventListener<K extends keyof SourceBufferListEventMap>(type: K, listener: (this: SourceBufferList, ev: SourceBufferListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SourceBufferListEventMap>(type: K, listener: (this: SourceBufferList, ev: SourceBufferListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: SourceBuffer;\n}\n\ndeclare var SourceBufferList: {\n    prototype: SourceBufferList;\n    new(): SourceBufferList;\n};\n\ninterface SpeechGrammar {\n    src: string;\n    weight: number;\n}\n\ndeclare var SpeechGrammar: {\n    prototype: SpeechGrammar;\n    new(): SpeechGrammar;\n};\n\ninterface SpeechGrammarList {\n    readonly length: number;\n    addFromString(string: string, weight?: number): void;\n    addFromURI(src: string, weight?: number): void;\n    item(index: number): SpeechGrammar;\n    [index: number]: SpeechGrammar;\n}\n\ndeclare var SpeechGrammarList: {\n    prototype: SpeechGrammarList;\n    new(): SpeechGrammarList;\n};\n\ninterface SpeechRecognitionEventMap {\n    "audioend": Event;\n    "audiostart": Event;\n    "end": Event;\n    "error": ErrorEvent;\n    "nomatch": SpeechRecognitionEvent;\n    "result": SpeechRecognitionEvent;\n    "soundend": Event;\n    "soundstart": Event;\n    "speechend": Event;\n    "speechstart": Event;\n    "start": Event;\n}\n\ninterface SpeechRecognition extends EventTarget {\n    continuous: boolean;\n    grammars: SpeechGrammarList;\n    interimResults: boolean;\n    lang: string;\n    maxAlternatives: number;\n    onaudioend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onaudiostart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onerror: ((this: SpeechRecognition, ev: ErrorEvent) => any) | null;\n    onnomatch: ((this: SpeechRecognition, ev: SpeechRecognitionEvent) => any) | null;\n    onresult: ((this: SpeechRecognition, ev: SpeechRecognitionEvent) => any) | null;\n    onsoundend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onsoundstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onspeechend: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onspeechstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    onstart: ((this: SpeechRecognition, ev: Event) => any) | null;\n    abort(): void;\n    start(): void;\n    stop(): void;\n    addEventListener<K extends keyof SpeechRecognitionEventMap>(type: K, listener: (this: SpeechRecognition, ev: SpeechRecognitionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechRecognitionEventMap>(type: K, listener: (this: SpeechRecognition, ev: SpeechRecognitionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechRecognition: {\n    prototype: SpeechRecognition;\n    new(): SpeechRecognition;\n};\n\ninterface SpeechRecognitionAlternative {\n    readonly confidence: number;\n    readonly transcript: string;\n}\n\ndeclare var SpeechRecognitionAlternative: {\n    prototype: SpeechRecognitionAlternative;\n    new(): SpeechRecognitionAlternative;\n};\n\ninterface SpeechRecognitionEvent extends Event {\n    readonly resultIndex: number;\n    readonly results: SpeechRecognitionResultList;\n}\n\ndeclare var SpeechRecognitionEvent: {\n    prototype: SpeechRecognitionEvent;\n    new(): SpeechRecognitionEvent;\n};\n\ninterface SpeechRecognitionResult {\n    readonly isFinal: boolean;\n    readonly length: number;\n    item(index: number): SpeechRecognitionAlternative;\n    [index: number]: SpeechRecognitionAlternative;\n}\n\ndeclare var SpeechRecognitionResult: {\n    prototype: SpeechRecognitionResult;\n    new(): SpeechRecognitionResult;\n};\n\ninterface SpeechRecognitionResultList {\n    readonly length: number;\n    item(index: number): SpeechRecognitionResult;\n    [index: number]: SpeechRecognitionResult;\n}\n\ndeclare var SpeechRecognitionResultList: {\n    prototype: SpeechRecognitionResultList;\n    new(): SpeechRecognitionResultList;\n};\n\ninterface SpeechSynthesisEventMap {\n    "voiceschanged": Event;\n}\n\n/** This Web Speech API interface is the controller interface for the speech service; this can be used to retrieve information about the synthesis voices available on the device, start and pause speech, and other commands besides. */\ninterface SpeechSynthesis extends EventTarget {\n    onvoiceschanged: ((this: SpeechSynthesis, ev: Event) => any) | null;\n    readonly paused: boolean;\n    readonly pending: boolean;\n    readonly speaking: boolean;\n    cancel(): void;\n    getVoices(): SpeechSynthesisVoice[];\n    pause(): void;\n    resume(): void;\n    speak(utterance: SpeechSynthesisUtterance): void;\n    addEventListener<K extends keyof SpeechSynthesisEventMap>(type: K, listener: (this: SpeechSynthesis, ev: SpeechSynthesisEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechSynthesisEventMap>(type: K, listener: (this: SpeechSynthesis, ev: SpeechSynthesisEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechSynthesis: {\n    prototype: SpeechSynthesis;\n    new(): SpeechSynthesis;\n};\n\ninterface SpeechSynthesisErrorEvent extends SpeechSynthesisEvent {\n    readonly error: SpeechSynthesisErrorCode;\n}\n\ndeclare var SpeechSynthesisErrorEvent: {\n    prototype: SpeechSynthesisErrorEvent;\n    new(type: string, eventInitDict: SpeechSynthesisErrorEventInit): SpeechSynthesisErrorEvent;\n};\n\n/** This Web Speech API interface contains information about the current state of SpeechSynthesisUtterance objects that have been processed in the speech service. */\ninterface SpeechSynthesisEvent extends Event {\n    readonly charIndex: number;\n    readonly charLength: number;\n    readonly elapsedTime: number;\n    readonly name: string;\n    readonly utterance: SpeechSynthesisUtterance;\n}\n\ndeclare var SpeechSynthesisEvent: {\n    prototype: SpeechSynthesisEvent;\n    new(type: string, eventInitDict: SpeechSynthesisEventInit): SpeechSynthesisEvent;\n};\n\ninterface SpeechSynthesisUtteranceEventMap {\n    "boundary": SpeechSynthesisEvent;\n    "end": SpeechSynthesisEvent;\n    "error": SpeechSynthesisErrorEvent;\n    "mark": SpeechSynthesisEvent;\n    "pause": SpeechSynthesisEvent;\n    "resume": SpeechSynthesisEvent;\n    "start": SpeechSynthesisEvent;\n}\n\n/** This Web Speech API interface represents a speech request. It contains the content the speech service should read and information about how to read it (e.g. language, pitch and volume.) */\ninterface SpeechSynthesisUtterance extends EventTarget {\n    lang: string;\n    onboundary: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onend: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onerror: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisErrorEvent) => any) | null;\n    onmark: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onpause: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onresume: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    onstart: ((this: SpeechSynthesisUtterance, ev: SpeechSynthesisEvent) => any) | null;\n    pitch: number;\n    rate: number;\n    text: string;\n    voice: SpeechSynthesisVoice | null;\n    volume: number;\n    addEventListener<K extends keyof SpeechSynthesisUtteranceEventMap>(type: K, listener: (this: SpeechSynthesisUtterance, ev: SpeechSynthesisUtteranceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SpeechSynthesisUtteranceEventMap>(type: K, listener: (this: SpeechSynthesisUtterance, ev: SpeechSynthesisUtteranceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SpeechSynthesisUtterance: {\n    prototype: SpeechSynthesisUtterance;\n    new(text?: string): SpeechSynthesisUtterance;\n};\n\n/** This Web Speech API interface represents a voice that the system supports. Every SpeechSynthesisVoice has its own relative speech service including information about language, name and URI. */\ninterface SpeechSynthesisVoice {\n    readonly default: boolean;\n    readonly lang: string;\n    readonly localService: boolean;\n    readonly name: string;\n    readonly voiceURI: string;\n}\n\ndeclare var SpeechSynthesisVoice: {\n    prototype: SpeechSynthesisVoice;\n    new(): SpeechSynthesisVoice;\n};\n\ninterface StaticRange extends AbstractRange {\n}\n\ndeclare var StaticRange: {\n    prototype: StaticRange;\n    new(init: StaticRangeInit): StaticRange;\n};\n\n/** The pan property takes a unitless value between -1 (full left pan) and 1 (full right pan). This interface was introduced as a much simpler way to apply a simple panning effect than having to use a full PannerNode. */\ninterface StereoPannerNode extends AudioNode {\n    readonly pan: AudioParam;\n}\n\ndeclare var StereoPannerNode: {\n    prototype: StereoPannerNode;\n    new(context: BaseAudioContext, options?: StereoPannerOptions): StereoPannerNode;\n};\n\n/** This Web Storage API interface provides access to a particular domain\'s session or local storage. It allows, for example, the addition, modification, or deletion of stored data items. */\ninterface Storage {\n    /**\n     * Returns the number of key/value pairs currently present in the list associated with the object.\n     */\n    readonly length: number;\n    /**\n     * Empties the list associated with the object of all key/value pairs, if there are any.\n     */\n    clear(): void;\n    /**\n     * Returns the current value associated with the given key, or null if the given key does not exist in the list associated with the object.\n     */\n    getItem(key: string): string | null;\n    /**\n     * Returns the name of the nth key in the list, or null if n is greater than or equal to the number of key/value pairs in the object.\n     */\n    key(index: number): string | null;\n    /**\n     * Removes the key/value pair with the given key from the list associated with the object, if a key/value pair with the given key exists.\n     */\n    removeItem(key: string): void;\n    /**\n     * Sets the value of the pair identified by key to value, creating a new key/value pair if none existed for key previously.\n     * \n     * Throws a "QuotaExceededError" DOMException exception if the new value couldn\'t be set. (Setting could fail if, e.g., the user has disabled storage for the site, or if the quota has been exceeded.)\n     */\n    setItem(key: string, value: string): void;\n    [name: string]: any;\n}\n\ndeclare var Storage: {\n    prototype: Storage;\n    new(): Storage;\n};\n\n/** A StorageEvent is sent to a window when a storage area it has access to is changed within the context of another document. */\ninterface StorageEvent extends Event {\n    /**\n     * Returns the key of the storage item being changed.\n     */\n    readonly key: string | null;\n    /**\n     * Returns the new value of the key of the storage item whose value is being changed.\n     */\n    readonly newValue: string | null;\n    /**\n     * Returns the old value of the key of the storage item whose value is being changed.\n     */\n    readonly oldValue: string | null;\n    /**\n     * Returns the Storage object that was affected.\n     */\n    readonly storageArea: Storage | null;\n    /**\n     * Returns the URL of the document whose storage item changed.\n     */\n    readonly url: string;\n}\n\ndeclare var StorageEvent: {\n    prototype: StorageEvent;\n    new(type: string, eventInitDict?: StorageEventInit): StorageEvent;\n};\n\ninterface StorageManager {\n    estimate(): Promise<StorageEstimate>;\n    persist(): Promise<boolean>;\n    persisted(): Promise<boolean>;\n}\n\ndeclare var StorageManager: {\n    prototype: StorageManager;\n    new(): StorageManager;\n};\n\ninterface StyleMedia {\n    readonly type: string;\n    matchMedium(mediaquery: string): boolean;\n}\n\ndeclare var StyleMedia: {\n    prototype: StyleMedia;\n    new(): StyleMedia;\n};\n\n/** A single style sheet. CSS style sheets will further implement the more specialized CSSStyleSheet interface. */\ninterface StyleSheet {\n    disabled: boolean;\n    readonly href: string | null;\n    readonly media: MediaList;\n    readonly ownerNode: Element | ProcessingInstruction | null;\n    readonly parentStyleSheet: CSSStyleSheet | null;\n    readonly title: string | null;\n    readonly type: string;\n}\n\ndeclare var StyleSheet: {\n    prototype: StyleSheet;\n    new(): StyleSheet;\n};\n\n/** A list of StyleSheet. */\ninterface StyleSheetList {\n    readonly length: number;\n    item(index: number): CSSStyleSheet | null;\n    [index: number]: CSSStyleSheet;\n}\n\ndeclare var StyleSheetList: {\n    prototype: StyleSheetList;\n    new(): StyleSheetList;\n};\n\n/** This Web Crypto API interface provides a number of low-level cryptographic functions. It is accessed via the Crypto.subtle properties available in a window context (via Window.crypto). */\ninterface SubtleCrypto {\n    decrypt(algorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    deriveBits(algorithm: AlgorithmIdentifier | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfParams | Pbkdf2Params, baseKey: CryptoKey, length: number): Promise<ArrayBuffer>;\n    deriveKey(algorithm: AlgorithmIdentifier | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfParams | Pbkdf2Params, baseKey: CryptoKey, derivedKeyType: string | AesDerivedKeyParams | HmacImportParams | ConcatParams | HkdfParams | Pbkdf2Params, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    digest(algorithm: AlgorithmIdentifier, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    encrypt(algorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    exportKey(format: "jwk", key: CryptoKey): Promise<JsonWebKey>;\n    exportKey(format: "raw" | "pkcs8" | "spki", key: CryptoKey): Promise<ArrayBuffer>;\n    exportKey(format: string, key: CryptoKey): Promise<JsonWebKey | ArrayBuffer>;\n    generateKey(algorithm: RsaHashedKeyGenParams | EcKeyGenParams | DhKeyGenParams, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKeyPair>;\n    generateKey(algorithm: AesKeyGenParams | HmacKeyGenParams | Pbkdf2Params, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    generateKey(algorithm: AlgorithmIdentifier, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKeyPair | CryptoKey>;\n    importKey(format: "jwk", keyData: JsonWebKey, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    importKey(format: "raw" | "pkcs8" | "spki", keyData: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    importKey(format: string, keyData: JsonWebKey | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    sign(algorithm: AlgorithmIdentifier | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    unwrapKey(format: "raw" | "pkcs8" | "spki" | "jwk" | string, wrappedKey: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, unwrappingKey: CryptoKey, unwrapAlgorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, unwrappedKeyAlgorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    verify(algorithm: AlgorithmIdentifier | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, signature: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<boolean>;\n    wrapKey(format: "raw" | "pkcs8" | "spki" | "jwk" | string, key: CryptoKey, wrappingKey: CryptoKey, wrapAlgorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams): Promise<ArrayBuffer>;\n}\n\ndeclare var SubtleCrypto: {\n    prototype: SubtleCrypto;\n    new(): SubtleCrypto;\n};\n\n/** This ServiceWorker API interface provides an interface for registering and listing sync registrations. */\ninterface SyncManager {\n    getTags(): Promise<string[]>;\n    register(tag: string): Promise<void>;\n}\n\ndeclare var SyncManager: {\n    prototype: SyncManager;\n    new(): SyncManager;\n};\n\n/** The textual content of Element or Attr. If an element has no markup within its content, it has a single child implementing Text that contains the element\'s text. However, if the element contains markup, it is parsed into information items and Text nodes that form its children. */\ninterface Text extends CharacterData, Slottable {\n    readonly assignedSlot: HTMLSlotElement | null;\n    /**\n     * Returns the combined data of all direct Text node siblings.\n     */\n    readonly wholeText: string;\n    /**\n     * Splits data at the given offset and returns the remainder as Text node.\n     */\n    splitText(offset: number): Text;\n}\n\ndeclare var Text: {\n    prototype: Text;\n    new(data?: string): Text;\n};\n\n/** A decoder for a specific method, that is a specific character encoding, like utf-8, iso-8859-2, koi8, cp1261, gbk, etc. A decoder takes a stream of bytes as input and emits a stream of code points. For a more scalable, non-native library, see StringView – a C-like representation of strings based on typed arrays. */\ninterface TextDecoder extends TextDecoderCommon {\n    /**\n     * Returns the result of running encoding\'s decoder. The method can be invoked zero or more times with options\'s stream set to true, and then once without options\'s stream (or set to false), to process a fragmented stream. If the invocation without options\'s stream (or set to false) has no input, it\'s clearest to omit both arguments.\n     * \n     * ```\n     * var string = "", decoder = new TextDecoder(encoding), buffer;\n     * while(buffer = next_chunk()) {\n     *   string += decoder.decode(buffer, {stream:true});\n     * }\n     * string += decoder.decode(); // end-of-stream\n     * ```\n     * \n     * If the error mode is "fatal" and encoding\'s decoder returns error, throws a TypeError.\n     */\n    decode(input?: BufferSource, options?: TextDecodeOptions): string;\n}\n\ndeclare var TextDecoder: {\n    prototype: TextDecoder;\n    new(label?: string, options?: TextDecoderOptions): TextDecoder;\n};\n\ninterface TextDecoderCommon {\n    /**\n     * Returns encoding\'s name, lowercased.\n     */\n    readonly encoding: string;\n    /**\n     * Returns true if error mode is "fatal", and false otherwise.\n     */\n    readonly fatal: boolean;\n    /**\n     * Returns true if ignore BOM flag is set, and false otherwise.\n     */\n    readonly ignoreBOM: boolean;\n}\n\ninterface TextDecoderStream extends GenericTransformStream, TextDecoderCommon {\n    readonly readable: ReadableStream<string>;\n    readonly writable: WritableStream<BufferSource>;\n}\n\ndeclare var TextDecoderStream: {\n    prototype: TextDecoderStream;\n    new(label?: string, options?: TextDecoderOptions): TextDecoderStream;\n};\n\n/** TextEncoder takes a stream of code points as input and emits a stream of bytes. For a more scalable, non-native library, see StringView – a C-like representation of strings based on typed arrays. */\ninterface TextEncoder extends TextEncoderCommon {\n    /**\n     * Returns the result of running UTF-8\'s encoder.\n     */\n    encode(input?: string): Uint8Array;\n    /**\n     * Runs the UTF-8 encoder on source, stores the result of that operation into destination, and returns the progress made as a dictionary whereby read is the number of converted code units of source and written is the number of bytes modified in destination.\n     */\n    encodeInto(source: string, destination: Uint8Array): TextEncoderEncodeIntoResult;\n}\n\ndeclare var TextEncoder: {\n    prototype: TextEncoder;\n    new(): TextEncoder;\n};\n\ninterface TextEncoderCommon {\n    /**\n     * Returns "utf-8".\n     */\n    readonly encoding: string;\n}\n\ninterface TextEncoderStream extends GenericTransformStream, TextEncoderCommon {\n    readonly readable: ReadableStream<Uint8Array>;\n    readonly writable: WritableStream<string>;\n}\n\ndeclare var TextEncoderStream: {\n    prototype: TextEncoderStream;\n    new(): TextEncoderStream;\n};\n\ninterface TextEvent extends UIEvent {\n    readonly data: string;\n    initTextEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, viewArg: Window, dataArg: string, inputMethod: number, locale: string): void;\n    readonly DOM_INPUT_METHOD_DROP: number;\n    readonly DOM_INPUT_METHOD_HANDWRITING: number;\n    readonly DOM_INPUT_METHOD_IME: number;\n    readonly DOM_INPUT_METHOD_KEYBOARD: number;\n    readonly DOM_INPUT_METHOD_MULTIMODAL: number;\n    readonly DOM_INPUT_METHOD_OPTION: number;\n    readonly DOM_INPUT_METHOD_PASTE: number;\n    readonly DOM_INPUT_METHOD_SCRIPT: number;\n    readonly DOM_INPUT_METHOD_UNKNOWN: number;\n    readonly DOM_INPUT_METHOD_VOICE: number;\n}\n\ndeclare var TextEvent: {\n    prototype: TextEvent;\n    new(): TextEvent;\n    readonly DOM_INPUT_METHOD_DROP: number;\n    readonly DOM_INPUT_METHOD_HANDWRITING: number;\n    readonly DOM_INPUT_METHOD_IME: number;\n    readonly DOM_INPUT_METHOD_KEYBOARD: number;\n    readonly DOM_INPUT_METHOD_MULTIMODAL: number;\n    readonly DOM_INPUT_METHOD_OPTION: number;\n    readonly DOM_INPUT_METHOD_PASTE: number;\n    readonly DOM_INPUT_METHOD_SCRIPT: number;\n    readonly DOM_INPUT_METHOD_UNKNOWN: number;\n    readonly DOM_INPUT_METHOD_VOICE: number;\n};\n\n/** The dimensions of a piece of text in the canvas, as created by the CanvasRenderingContext2D.measureText() method. */\ninterface TextMetrics {\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxLeft: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxRight: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly alphabeticBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly emHeightAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly emHeightDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly fontBoundingBoxAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly fontBoundingBoxDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly hangingBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly ideographicBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly width: number;\n}\n\ndeclare var TextMetrics: {\n    prototype: TextMetrics;\n    new(): TextMetrics;\n};\n\ninterface TextTrackEventMap {\n    "cuechange": Event;\n}\n\n/** This interface also inherits properties from EventTarget. */\ninterface TextTrack extends EventTarget {\n    /**\n     * Returns the text track cues from the text track list of cues that are currently active (i.e. that start before the current playback position and end after it), as a TextTrackCueList object.\n     */\n    readonly activeCues: TextTrackCueList | null;\n    /**\n     * Returns the text track list of cues, as a TextTrackCueList object.\n     */\n    readonly cues: TextTrackCueList | null;\n    /**\n     * Returns the ID of the given track.\n     * \n     * For in-band tracks, this is the ID that can be used with a fragment if the format supports media fragment syntax, and that can be used with the getTrackById() method.\n     * \n     * For TextTrack objects corresponding to track elements, this is the ID of the track element.\n     */\n    readonly id: string;\n    /**\n     * Returns the text track in-band metadata track dispatch type string.\n     */\n    readonly inBandMetadataTrackDispatchType: string;\n    /**\n     * Returns the text track kind string.\n     */\n    readonly kind: TextTrackKind;\n    /**\n     * Returns the text track label, if there is one, or the empty string otherwise (indicating that a custom label probably needs to be generated from the other attributes of the object if the object is exposed to the user).\n     */\n    readonly label: string;\n    /**\n     * Returns the text track language string.\n     */\n    readonly language: string;\n    /**\n     * Returns the text track mode, represented by a string from the following list:\n     * \n     * Can be set, to change the mode.\n     */\n    mode: TextTrackMode;\n    oncuechange: ((this: TextTrack, ev: Event) => any) | null;\n    readonly sourceBuffer: SourceBuffer | null;\n    /**\n     * Adds the given cue to textTrack\'s text track list of cues.\n     */\n    addCue(cue: TextTrackCue): void;\n    /**\n     * Removes the given cue from textTrack\'s text track list of cues.\n     */\n    removeCue(cue: TextTrackCue): void;\n    addEventListener<K extends keyof TextTrackEventMap>(type: K, listener: (this: TextTrack, ev: TextTrackEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackEventMap>(type: K, listener: (this: TextTrack, ev: TextTrackEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var TextTrack: {\n    prototype: TextTrack;\n    new(): TextTrack;\n};\n\ninterface TextTrackCueEventMap {\n    "enter": Event;\n    "exit": Event;\n}\n\n/** TextTrackCues represent a string of text that will be displayed for some duration of time on a TextTrack. This includes the start and end times that the cue will be displayed. A TextTrackCue cannot be used directly, instead one of the derived types (e.g. VTTCue) must be used. */\ninterface TextTrackCue extends EventTarget {\n    /**\n     * Returns the text track cue end time, in seconds.\n     * \n     * Can be set.\n     */\n    endTime: number;\n    /**\n     * Returns the text track cue identifier.\n     * \n     * Can be set.\n     */\n    id: string;\n    onenter: ((this: TextTrackCue, ev: Event) => any) | null;\n    onexit: ((this: TextTrackCue, ev: Event) => any) | null;\n    /**\n     * Returns true if the text track cue pause-on-exit flag is set, false otherwise.\n     * \n     * Can be set.\n     */\n    pauseOnExit: boolean;\n    /**\n     * Returns the text track cue start time, in seconds.\n     * \n     * Can be set.\n     */\n    startTime: number;\n    /**\n     * Returns the TextTrack object to which this text track cue belongs, if any, or null otherwise.\n     */\n    readonly track: TextTrack | null;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: TextTrackCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: TextTrackCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var TextTrackCue: {\n    prototype: TextTrackCue;\n    new(): TextTrackCue;\n};\n\ninterface TextTrackCueList {\n    /**\n     * Returns the number of cues in the list.\n     */\n    readonly length: number;\n    /**\n     * Returns the first text track cue (in text track cue order) with text track cue identifier id.\n     * \n     * Returns null if none of the cues have the given identifier or if the argument is the empty string.\n     */\n    getCueById(id: string): TextTrackCue | null;\n    [index: number]: TextTrackCue;\n}\n\ndeclare var TextTrackCueList: {\n    prototype: TextTrackCueList;\n    new(): TextTrackCueList;\n};\n\ninterface TextTrackListEventMap {\n    "addtrack": TrackEvent;\n    "change": Event;\n    "removetrack": TrackEvent;\n}\n\ninterface TextTrackList extends EventTarget {\n    readonly length: number;\n    onaddtrack: ((this: TextTrackList, ev: TrackEvent) => any) | null;\n    onchange: ((this: TextTrackList, ev: Event) => any) | null;\n    onremovetrack: ((this: TextTrackList, ev: TrackEvent) => any) | null;\n    getTrackById(id: string): TextTrack | null;\n    addEventListener<K extends keyof TextTrackListEventMap>(type: K, listener: (this: TextTrackList, ev: TextTrackListEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackListEventMap>(type: K, listener: (this: TextTrackList, ev: TextTrackListEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: TextTrack;\n}\n\ndeclare var TextTrackList: {\n    prototype: TextTrackList;\n    new(): TextTrackList;\n};\n\n/** Used to represent a set of time ranges, primarily for the purpose of tracking which portions of media have been buffered when loading it for use by the <audio> and <video> elements. */\ninterface TimeRanges {\n    /**\n     * Returns the number of ranges in the object.\n     */\n    readonly length: number;\n    /**\n     * Returns the time for the end of the range with the given index.\n     * \n     * Throws an "IndexSizeError" DOMException if the index is out of range.\n     */\n    end(index: number): number;\n    /**\n     * Returns the time for the start of the range with the given index.\n     * \n     * Throws an "IndexSizeError" DOMException if the index is out of range.\n     */\n    start(index: number): number;\n}\n\ndeclare var TimeRanges: {\n    prototype: TimeRanges;\n    new(): TimeRanges;\n};\n\n/** A single contact point on a touch-sensitive device. The contact point is commonly a finger or stylus and the device may be a touchscreen or trackpad. */\ninterface Touch {\n    readonly altitudeAngle: number;\n    readonly azimuthAngle: number;\n    readonly clientX: number;\n    readonly clientY: number;\n    readonly force: number;\n    readonly identifier: number;\n    readonly pageX: number;\n    readonly pageY: number;\n    readonly radiusX: number;\n    readonly radiusY: number;\n    readonly rotationAngle: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly target: EventTarget;\n    readonly touchType: TouchType;\n}\n\ndeclare var Touch: {\n    prototype: Touch;\n    new(touchInitDict: TouchInit): Touch;\n};\n\n/** An event sent when the state of contacts with a touch-sensitive surface changes. This surface can be a touch screen or trackpad, for example. The event can describe one or more points of contact with the screen and includes support for detecting movement, addition and removal of contact points, and so forth. */\ninterface TouchEvent extends UIEvent {\n    readonly altKey: boolean;\n    readonly changedTouches: TouchList;\n    readonly ctrlKey: boolean;\n    readonly metaKey: boolean;\n    readonly shiftKey: boolean;\n    readonly targetTouches: TouchList;\n    readonly touches: TouchList;\n}\n\ndeclare var TouchEvent: {\n    prototype: TouchEvent;\n    new(type: string, eventInitDict?: TouchEventInit): TouchEvent;\n};\n\n/** A list of contact points on a touch surface. For example, if the user has three fingers on the touch surface (such as a screen or trackpad), the corresponding TouchList object would have one Touch object for each finger, for a total of three entries. */\ninterface TouchList {\n    readonly length: number;\n    item(index: number): Touch | null;\n    [index: number]: Touch;\n}\n\ndeclare var TouchList: {\n    prototype: TouchList;\n    new(): TouchList;\n};\n\n/** The TrackEvent interface, part of the HTML DOM specification, is used for events which represent changes to the set of available tracks on an HTML media element; these events are addtrack and removetrack. */\ninterface TrackEvent extends Event {\n    /**\n     * Returns the track object (TextTrack, AudioTrack, or VideoTrack) to which the event relates.\n     */\n    readonly track: TextTrack | null;\n}\n\ndeclare var TrackEvent: {\n    prototype: TrackEvent;\n    new(type: string, eventInitDict?: TrackEventInit): TrackEvent;\n};\n\ninterface TransformStream<I = any, O = any> {\n    readonly readable: ReadableStream<O>;\n    readonly writable: WritableStream<I>;\n}\n\ndeclare var TransformStream: {\n    prototype: TransformStream;\n    new<I = any, O = any>(transformer?: Transformer<I, O>, writableStrategy?: QueuingStrategy<I>, readableStrategy?: QueuingStrategy<O>): TransformStream<I, O>;\n};\n\ninterface TransformStreamDefaultController<O = any> {\n    readonly desiredSize: number | null;\n    enqueue(chunk: O): void;\n    error(reason?: any): void;\n    terminate(): void;\n}\n\ndeclare var TransformStreamDefaultController: {\n    prototype: TransformStreamDefaultController;\n    new(): TransformStreamDefaultController;\n};\n\n/** Events providing information related to transitions. */\ninterface TransitionEvent extends Event {\n    readonly elapsedTime: number;\n    readonly propertyName: string;\n    readonly pseudoElement: string;\n}\n\ndeclare var TransitionEvent: {\n    prototype: TransitionEvent;\n    new(type: string, transitionEventInitDict?: TransitionEventInit): TransitionEvent;\n};\n\n/** The nodes of a document subtree and a position within them. */\ninterface TreeWalker {\n    currentNode: Node;\n    readonly filter: NodeFilter | null;\n    readonly root: Node;\n    readonly whatToShow: number;\n    firstChild(): Node | null;\n    lastChild(): Node | null;\n    nextNode(): Node | null;\n    nextSibling(): Node | null;\n    parentNode(): Node | null;\n    previousNode(): Node | null;\n    previousSibling(): Node | null;\n}\n\ndeclare var TreeWalker: {\n    prototype: TreeWalker;\n    new(): TreeWalker;\n};\n\n/** Simple user interface events. */\ninterface UIEvent extends Event {\n    readonly detail: number;\n    readonly view: Window | null;\n    /** @deprecated */\n    readonly which: number;\n}\n\ndeclare var UIEvent: {\n    prototype: UIEvent;\n    new(type: string, eventInitDict?: UIEventInit): UIEvent;\n};\n\n/** The URL interface represents an object providing static methods used for creating object URLs. */\ninterface URL {\n    hash: string;\n    host: string;\n    hostname: string;\n    href: string;\n    toString(): string;\n    readonly origin: string;\n    password: string;\n    pathname: string;\n    port: string;\n    protocol: string;\n    search: string;\n    readonly searchParams: URLSearchParams;\n    username: string;\n    toJSON(): string;\n}\n\ndeclare var URL: {\n    prototype: URL;\n    new(url: string, base?: string | URL): URL;\n    createObjectURL(object: any): string;\n    revokeObjectURL(url: string): void;\n};\n\ntype webkitURL = URL;\ndeclare var webkitURL: typeof URL;\n\ninterface URLSearchParams {\n    /**\n     * Appends a specified key/value pair as a new search parameter.\n     */\n    append(name: string, value: string): void;\n    /**\n     * Deletes the given search parameter, and its associated value, from the list of all search parameters.\n     */\n    delete(name: string): void;\n    /**\n     * Returns the first value associated to the given search parameter.\n     */\n    get(name: string): string | null;\n    /**\n     * Returns all the values association with a given search parameter.\n     */\n    getAll(name: string): string[];\n    /**\n     * Returns a Boolean indicating if such a search parameter exists.\n     */\n    has(name: string): boolean;\n    /**\n     * Sets the value associated to a given search parameter to the given value. If there were several values, delete the others.\n     */\n    set(name: string, value: string): void;\n    sort(): void;\n    /**\n     * Returns a string containing a query string suitable for use in a URL. Does not include the question mark.\n     */\n    toString(): string;\n    forEach(callbackfn: (value: string, key: string, parent: URLSearchParams) => void, thisArg?: any): void;\n}\n\ndeclare var URLSearchParams: {\n    prototype: URLSearchParams;\n    new(init?: string[][] | Record<string, string> | string | URLSearchParams): URLSearchParams;\n    toString(): string;\n};\n\n/** This WebVR API interface represents any VR device supported by this API. It includes generic information such as device IDs and descriptions, as well as methods for starting to present a VR scene, retrieving eye parameters and display capabilities, and other important functionality. */\ninterface VRDisplay extends EventTarget {\n    readonly capabilities: VRDisplayCapabilities;\n    depthFar: number;\n    depthNear: number;\n    readonly displayId: number;\n    readonly displayName: string;\n    readonly isConnected: boolean;\n    readonly isPresenting: boolean;\n    readonly stageParameters: VRStageParameters | null;\n    cancelAnimationFrame(handle: number): void;\n    exitPresent(): Promise<void>;\n    getEyeParameters(whichEye: string): VREyeParameters;\n    getFrameData(frameData: VRFrameData): boolean;\n    getLayers(): VRLayer[];\n    /** @deprecated */\n    getPose(): VRPose;\n    requestAnimationFrame(callback: FrameRequestCallback): number;\n    requestPresent(layers: VRLayer[]): Promise<void>;\n    resetPose(): void;\n    submitFrame(pose?: VRPose): void;\n}\n\ndeclare var VRDisplay: {\n    prototype: VRDisplay;\n    new(): VRDisplay;\n};\n\n/** This WebVR API interface describes the capabilities of a VRDisplay — its features can be used to perform VR device capability tests, for example can it return position information. */\ninterface VRDisplayCapabilities {\n    readonly canPresent: boolean;\n    readonly hasExternalDisplay: boolean;\n    readonly hasOrientation: boolean;\n    readonly hasPosition: boolean;\n    readonly maxLayers: number;\n}\n\ndeclare var VRDisplayCapabilities: {\n    prototype: VRDisplayCapabilities;\n    new(): VRDisplayCapabilities;\n};\n\n/** This WebVR API interface represents represents the event object of WebVR-related events (see the list of WebVR window extensions). */\ninterface VRDisplayEvent extends Event {\n    readonly display: VRDisplay;\n    readonly reason: VRDisplayEventReason | null;\n}\n\ndeclare var VRDisplayEvent: {\n    prototype: VRDisplayEvent;\n    new(type: string, eventInitDict: VRDisplayEventInit): VRDisplayEvent;\n};\n\n/** This WebVR API interface represents all the information required to correctly render a scene for a given eye, including field of view information. */\ninterface VREyeParameters {\n    /** @deprecated */\n    readonly fieldOfView: VRFieldOfView;\n    readonly offset: Float32Array;\n    readonly renderHeight: number;\n    readonly renderWidth: number;\n}\n\ndeclare var VREyeParameters: {\n    prototype: VREyeParameters;\n    new(): VREyeParameters;\n};\n\n/** This WebVR API interface represents a field of view defined by 4 different degree values describing the view from a center point. */\ninterface VRFieldOfView {\n    readonly downDegrees: number;\n    readonly leftDegrees: number;\n    readonly rightDegrees: number;\n    readonly upDegrees: number;\n}\n\ndeclare var VRFieldOfView: {\n    prototype: VRFieldOfView;\n    new(): VRFieldOfView;\n};\n\n/** This WebVR API interface represents all the information needed to render a single frame of a VR scene; constructed by VRDisplay.getFrameData(). */\ninterface VRFrameData {\n    readonly leftProjectionMatrix: Float32Array;\n    readonly leftViewMatrix: Float32Array;\n    readonly pose: VRPose;\n    readonly rightProjectionMatrix: Float32Array;\n    readonly rightViewMatrix: Float32Array;\n    readonly timestamp: number;\n}\n\ndeclare var VRFrameData: {\n    prototype: VRFrameData;\n    new(): VRFrameData;\n};\n\n/** This WebVR API interface represents the state of a VR sensor at a given timestamp (which includes orientation, position, velocity, and acceleration information.) */\ninterface VRPose {\n    readonly angularAcceleration: Float32Array | null;\n    readonly angularVelocity: Float32Array | null;\n    readonly linearAcceleration: Float32Array | null;\n    readonly linearVelocity: Float32Array | null;\n    readonly orientation: Float32Array | null;\n    readonly position: Float32Array | null;\n    readonly timestamp: number;\n}\n\ndeclare var VRPose: {\n    prototype: VRPose;\n    new(): VRPose;\n};\n\ninterface VTTCue extends TextTrackCue {\n    align: AlignSetting;\n    line: LineAndPositionSetting;\n    lineAlign: LineAlignSetting;\n    position: LineAndPositionSetting;\n    positionAlign: PositionAlignSetting;\n    region: VTTRegion | null;\n    size: number;\n    snapToLines: boolean;\n    text: string;\n    vertical: DirectionSetting;\n    getCueAsHTML(): DocumentFragment;\n    addEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: VTTCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof TextTrackCueEventMap>(type: K, listener: (this: VTTCue, ev: TextTrackCueEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var VTTCue: {\n    prototype: VTTCue;\n    new(startTime: number, endTime: number, text: string): VTTCue;\n};\n\ninterface VTTRegion {\n    id: string;\n    lines: number;\n    regionAnchorX: number;\n    regionAnchorY: number;\n    scroll: ScrollSetting;\n    viewportAnchorX: number;\n    viewportAnchorY: number;\n    width: number;\n}\n\ndeclare var VTTRegion: {\n    prototype: VTTRegion;\n    new(): VTTRegion;\n};\n\n/** The validity states that an element can be in, with respect to constraint validation. Together, they help explain why an element\'s value fails to validate, if it\'s not valid. */\ninterface ValidityState {\n    readonly badInput: boolean;\n    readonly customError: boolean;\n    readonly patternMismatch: boolean;\n    readonly rangeOverflow: boolean;\n    readonly rangeUnderflow: boolean;\n    readonly stepMismatch: boolean;\n    readonly tooLong: boolean;\n    readonly tooShort: boolean;\n    readonly typeMismatch: boolean;\n    readonly valid: boolean;\n    readonly valueMissing: boolean;\n}\n\ndeclare var ValidityState: {\n    prototype: ValidityState;\n    new(): ValidityState;\n};\n\n/** Returned by the HTMLVideoElement.getVideoPlaybackQuality() method and contains metrics that can be used to determine the playback quality of a video. */\ninterface VideoPlaybackQuality {\n    readonly creationTime: number;\n    readonly droppedVideoFrames: number;\n    readonly totalVideoFrames: number;\n}\n\ndeclare var VideoPlaybackQuality: {\n    prototype: VideoPlaybackQuality;\n    new(): VideoPlaybackQuality;\n};\n\ninterface VisualViewportEventMap {\n    "resize": UIEvent;\n    "scroll": Event;\n}\n\ninterface VisualViewport extends EventTarget {\n    readonly height: number;\n    readonly offsetLeft: number;\n    readonly offsetTop: number;\n    onresize: ((this: VisualViewport, ev: UIEvent) => any) | null;\n    onscroll: ((this: VisualViewport, ev: Event) => any) | null;\n    readonly pageLeft: number;\n    readonly pageTop: number;\n    readonly scale: number;\n    readonly width: number;\n    addEventListener<K extends keyof VisualViewportEventMap>(type: K, listener: (this: VisualViewport, ev: VisualViewportEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof VisualViewportEventMap>(type: K, listener: (this: VisualViewport, ev: VisualViewportEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var VisualViewport: {\n    prototype: VisualViewport;\n    new(): VisualViewport;\n};\n\ninterface WEBGL_color_buffer_float {\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE_EXT: GLenum;\n    readonly RGBA32F_EXT: GLenum;\n    readonly UNSIGNED_NORMALIZED_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_astc {\n    getSupportedProfiles(): string[];\n    readonly COMPRESSED_RGBA_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x8_KHR: GLenum;\n}\n\n/** The WEBGL_compressed_texture_s3tc extension is part of the WebGL API and exposes four S3TC compressed texture formats. */\ninterface WEBGL_compressed_texture_s3tc {\n    readonly COMPRESSED_RGBA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_RGB_S3TC_DXT1_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_s3tc_srgb {\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_SRGB_S3TC_DXT1_EXT: GLenum;\n}\n\n/** The WEBGL_debug_renderer_info extension is part of the WebGL API and exposes two constants with information about the graphics driver for debugging purposes. */\ninterface WEBGL_debug_renderer_info {\n    readonly UNMASKED_RENDERER_WEBGL: GLenum;\n    readonly UNMASKED_VENDOR_WEBGL: GLenum;\n}\n\ninterface WEBGL_debug_shaders {\n    getTranslatedShaderSource(shader: WebGLShader): string;\n}\n\n/** The WEBGL_depth_texture extension is part of the WebGL API and defines 2D depth and depth-stencil textures. */\ninterface WEBGL_depth_texture {\n    readonly UNSIGNED_INT_24_8_WEBGL: GLenum;\n}\n\ninterface WEBGL_draw_buffers {\n    drawBuffersWEBGL(buffers: GLenum[]): void;\n    readonly COLOR_ATTACHMENT0_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT10_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT11_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT12_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT13_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT14_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT15_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT1_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT2_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT3_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT4_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT5_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT6_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT7_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT8_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT9_WEBGL: GLenum;\n    readonly DRAW_BUFFER0_WEBGL: GLenum;\n    readonly DRAW_BUFFER10_WEBGL: GLenum;\n    readonly DRAW_BUFFER11_WEBGL: GLenum;\n    readonly DRAW_BUFFER12_WEBGL: GLenum;\n    readonly DRAW_BUFFER13_WEBGL: GLenum;\n    readonly DRAW_BUFFER14_WEBGL: GLenum;\n    readonly DRAW_BUFFER15_WEBGL: GLenum;\n    readonly DRAW_BUFFER1_WEBGL: GLenum;\n    readonly DRAW_BUFFER2_WEBGL: GLenum;\n    readonly DRAW_BUFFER3_WEBGL: GLenum;\n    readonly DRAW_BUFFER4_WEBGL: GLenum;\n    readonly DRAW_BUFFER5_WEBGL: GLenum;\n    readonly DRAW_BUFFER6_WEBGL: GLenum;\n    readonly DRAW_BUFFER7_WEBGL: GLenum;\n    readonly DRAW_BUFFER8_WEBGL: GLenum;\n    readonly DRAW_BUFFER9_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS_WEBGL: GLenum;\n    readonly MAX_DRAW_BUFFERS_WEBGL: GLenum;\n}\n\ninterface WEBGL_lose_context {\n    loseContext(): void;\n    restoreContext(): void;\n}\n\n/** A WaveShaperNode always has exactly one input and one output. */\ninterface WaveShaperNode extends AudioNode {\n    curve: Float32Array | null;\n    oversample: OverSampleType;\n}\n\ndeclare var WaveShaperNode: {\n    prototype: WaveShaperNode;\n    new(context: BaseAudioContext, options?: WaveShaperOptions): WaveShaperNode;\n};\n\ninterface WebAuthentication {\n    getAssertion(assertionChallenge: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, options?: AssertionOptions): Promise<WebAuthnAssertion>;\n    makeCredential(accountInformation: Account, cryptoParameters: ScopedCredentialParameters[], attestationChallenge: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, options?: ScopedCredentialOptions): Promise<ScopedCredentialInfo>;\n}\n\ndeclare var WebAuthentication: {\n    prototype: WebAuthentication;\n    new(): WebAuthentication;\n};\n\ninterface WebAuthnAssertion {\n    readonly authenticatorData: ArrayBuffer;\n    readonly clientData: ArrayBuffer;\n    readonly credential: ScopedCredential;\n    readonly signature: ArrayBuffer;\n}\n\ndeclare var WebAuthnAssertion: {\n    prototype: WebAuthnAssertion;\n    new(): WebAuthnAssertion;\n};\n\ninterface WebGL2RenderingContext extends WebGL2RenderingContextBase, WebGL2RenderingContextOverloads, WebGLRenderingContextBase {\n}\n\ndeclare var WebGL2RenderingContext: {\n    prototype: WebGL2RenderingContext;\n    new(): WebGL2RenderingContext;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n    readonly ACTIVE_UNIFORM_BLOCKS: GLenum;\n    readonly ALREADY_SIGNALED: GLenum;\n    readonly ANY_SAMPLES_PASSED: GLenum;\n    readonly ANY_SAMPLES_PASSED_CONSERVATIVE: GLenum;\n    readonly COLOR: GLenum;\n    readonly COLOR_ATTACHMENT1: GLenum;\n    readonly COLOR_ATTACHMENT10: GLenum;\n    readonly COLOR_ATTACHMENT11: GLenum;\n    readonly COLOR_ATTACHMENT12: GLenum;\n    readonly COLOR_ATTACHMENT13: GLenum;\n    readonly COLOR_ATTACHMENT14: GLenum;\n    readonly COLOR_ATTACHMENT15: GLenum;\n    readonly COLOR_ATTACHMENT2: GLenum;\n    readonly COLOR_ATTACHMENT3: GLenum;\n    readonly COLOR_ATTACHMENT4: GLenum;\n    readonly COLOR_ATTACHMENT5: GLenum;\n    readonly COLOR_ATTACHMENT6: GLenum;\n    readonly COLOR_ATTACHMENT7: GLenum;\n    readonly COLOR_ATTACHMENT8: GLenum;\n    readonly COLOR_ATTACHMENT9: GLenum;\n    readonly COMPARE_REF_TO_TEXTURE: GLenum;\n    readonly CONDITION_SATISFIED: GLenum;\n    readonly COPY_READ_BUFFER: GLenum;\n    readonly COPY_READ_BUFFER_BINDING: GLenum;\n    readonly COPY_WRITE_BUFFER: GLenum;\n    readonly COPY_WRITE_BUFFER_BINDING: GLenum;\n    readonly CURRENT_QUERY: GLenum;\n    readonly DEPTH: GLenum;\n    readonly DEPTH24_STENCIL8: GLenum;\n    readonly DEPTH32F_STENCIL8: GLenum;\n    readonly DEPTH_COMPONENT24: GLenum;\n    readonly DEPTH_COMPONENT32F: GLenum;\n    readonly DRAW_BUFFER0: GLenum;\n    readonly DRAW_BUFFER1: GLenum;\n    readonly DRAW_BUFFER10: GLenum;\n    readonly DRAW_BUFFER11: GLenum;\n    readonly DRAW_BUFFER12: GLenum;\n    readonly DRAW_BUFFER13: GLenum;\n    readonly DRAW_BUFFER14: GLenum;\n    readonly DRAW_BUFFER15: GLenum;\n    readonly DRAW_BUFFER2: GLenum;\n    readonly DRAW_BUFFER3: GLenum;\n    readonly DRAW_BUFFER4: GLenum;\n    readonly DRAW_BUFFER5: GLenum;\n    readonly DRAW_BUFFER6: GLenum;\n    readonly DRAW_BUFFER7: GLenum;\n    readonly DRAW_BUFFER8: GLenum;\n    readonly DRAW_BUFFER9: GLenum;\n    readonly DRAW_FRAMEBUFFER: GLenum;\n    readonly DRAW_FRAMEBUFFER_BINDING: GLenum;\n    readonly DYNAMIC_COPY: GLenum;\n    readonly DYNAMIC_READ: GLenum;\n    readonly FLOAT_32_UNSIGNED_INT_24_8_REV: GLenum;\n    readonly FLOAT_MAT2x3: GLenum;\n    readonly FLOAT_MAT2x4: GLenum;\n    readonly FLOAT_MAT3x2: GLenum;\n    readonly FLOAT_MAT3x4: GLenum;\n    readonly FLOAT_MAT4x2: GLenum;\n    readonly FLOAT_MAT4x3: GLenum;\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_BLUE_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_GREEN_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_RED_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER: GLenum;\n    readonly FRAMEBUFFER_DEFAULT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: GLenum;\n    readonly HALF_FLOAT: GLenum;\n    readonly INTERLEAVED_ATTRIBS: GLenum;\n    readonly INT_2_10_10_10_REV: GLenum;\n    readonly INT_SAMPLER_2D: GLenum;\n    readonly INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly INT_SAMPLER_3D: GLenum;\n    readonly INT_SAMPLER_CUBE: GLenum;\n    readonly INVALID_INDEX: GLenum;\n    readonly MAX: GLenum;\n    readonly MAX_3D_TEXTURE_SIZE: GLenum;\n    readonly MAX_ARRAY_TEXTURE_LAYERS: GLenum;\n    readonly MAX_CLIENT_WAIT_TIMEOUT_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS: GLenum;\n    readonly MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_COMBINED_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_DRAW_BUFFERS: GLenum;\n    readonly MAX_ELEMENTS_INDICES: GLenum;\n    readonly MAX_ELEMENTS_VERTICES: GLenum;\n    readonly MAX_ELEMENT_INDEX: GLenum;\n    readonly MAX_FRAGMENT_INPUT_COMPONENTS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly MAX_SAMPLES: GLenum;\n    readonly MAX_SERVER_WAIT_TIMEOUT: GLenum;\n    readonly MAX_TEXTURE_LOD_BIAS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS: GLenum;\n    readonly MAX_UNIFORM_BLOCK_SIZE: GLenum;\n    readonly MAX_UNIFORM_BUFFER_BINDINGS: GLenum;\n    readonly MAX_VARYING_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_OUTPUT_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MIN: GLenum;\n    readonly MIN_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly OBJECT_TYPE: GLenum;\n    readonly PACK_ROW_LENGTH: GLenum;\n    readonly PACK_SKIP_PIXELS: GLenum;\n    readonly PACK_SKIP_ROWS: GLenum;\n    readonly PIXEL_PACK_BUFFER: GLenum;\n    readonly PIXEL_PACK_BUFFER_BINDING: GLenum;\n    readonly PIXEL_UNPACK_BUFFER: GLenum;\n    readonly PIXEL_UNPACK_BUFFER_BINDING: GLenum;\n    readonly QUERY_RESULT: GLenum;\n    readonly QUERY_RESULT_AVAILABLE: GLenum;\n    readonly R11F_G11F_B10F: GLenum;\n    readonly R16F: GLenum;\n    readonly R16I: GLenum;\n    readonly R16UI: GLenum;\n    readonly R32F: GLenum;\n    readonly R32I: GLenum;\n    readonly R32UI: GLenum;\n    readonly R8: GLenum;\n    readonly R8I: GLenum;\n    readonly R8UI: GLenum;\n    readonly R8_SNORM: GLenum;\n    readonly RASTERIZER_DISCARD: GLenum;\n    readonly READ_BUFFER: GLenum;\n    readonly READ_FRAMEBUFFER: GLenum;\n    readonly READ_FRAMEBUFFER_BINDING: GLenum;\n    readonly RED: GLenum;\n    readonly RED_INTEGER: GLenum;\n    readonly RENDERBUFFER_SAMPLES: GLenum;\n    readonly RG: GLenum;\n    readonly RG16F: GLenum;\n    readonly RG16I: GLenum;\n    readonly RG16UI: GLenum;\n    readonly RG32F: GLenum;\n    readonly RG32I: GLenum;\n    readonly RG32UI: GLenum;\n    readonly RG8: GLenum;\n    readonly RG8I: GLenum;\n    readonly RG8UI: GLenum;\n    readonly RG8_SNORM: GLenum;\n    readonly RGB10_A2: GLenum;\n    readonly RGB10_A2UI: GLenum;\n    readonly RGB16F: GLenum;\n    readonly RGB16I: GLenum;\n    readonly RGB16UI: GLenum;\n    readonly RGB32F: GLenum;\n    readonly RGB32I: GLenum;\n    readonly RGB32UI: GLenum;\n    readonly RGB8: GLenum;\n    readonly RGB8I: GLenum;\n    readonly RGB8UI: GLenum;\n    readonly RGB8_SNORM: GLenum;\n    readonly RGB9_E5: GLenum;\n    readonly RGBA16F: GLenum;\n    readonly RGBA16I: GLenum;\n    readonly RGBA16UI: GLenum;\n    readonly RGBA32F: GLenum;\n    readonly RGBA32I: GLenum;\n    readonly RGBA32UI: GLenum;\n    readonly RGBA8: GLenum;\n    readonly RGBA8I: GLenum;\n    readonly RGBA8UI: GLenum;\n    readonly RGBA8_SNORM: GLenum;\n    readonly RGBA_INTEGER: GLenum;\n    readonly RGB_INTEGER: GLenum;\n    readonly RG_INTEGER: GLenum;\n    readonly SAMPLER_2D_ARRAY: GLenum;\n    readonly SAMPLER_2D_ARRAY_SHADOW: GLenum;\n    readonly SAMPLER_2D_SHADOW: GLenum;\n    readonly SAMPLER_3D: GLenum;\n    readonly SAMPLER_BINDING: GLenum;\n    readonly SAMPLER_CUBE_SHADOW: GLenum;\n    readonly SEPARATE_ATTRIBS: GLenum;\n    readonly SIGNALED: GLenum;\n    readonly SIGNED_NORMALIZED: GLenum;\n    readonly SRGB: GLenum;\n    readonly SRGB8: GLenum;\n    readonly SRGB8_ALPHA8: GLenum;\n    readonly STATIC_COPY: GLenum;\n    readonly STATIC_READ: GLenum;\n    readonly STENCIL: GLenum;\n    readonly STREAM_COPY: GLenum;\n    readonly STREAM_READ: GLenum;\n    readonly SYNC_CONDITION: GLenum;\n    readonly SYNC_FENCE: GLenum;\n    readonly SYNC_FLAGS: GLenum;\n    readonly SYNC_FLUSH_COMMANDS_BIT: GLenum;\n    readonly SYNC_GPU_COMMANDS_COMPLETE: GLenum;\n    readonly SYNC_STATUS: GLenum;\n    readonly TEXTURE_2D_ARRAY: GLenum;\n    readonly TEXTURE_3D: GLenum;\n    readonly TEXTURE_BASE_LEVEL: GLenum;\n    readonly TEXTURE_BINDING_2D_ARRAY: GLenum;\n    readonly TEXTURE_BINDING_3D: GLenum;\n    readonly TEXTURE_COMPARE_FUNC: GLenum;\n    readonly TEXTURE_COMPARE_MODE: GLenum;\n    readonly TEXTURE_IMMUTABLE_FORMAT: GLenum;\n    readonly TEXTURE_IMMUTABLE_LEVELS: GLenum;\n    readonly TEXTURE_MAX_LEVEL: GLenum;\n    readonly TEXTURE_MAX_LOD: GLenum;\n    readonly TEXTURE_MIN_LOD: GLenum;\n    readonly TEXTURE_WRAP_R: GLenum;\n    readonly TIMEOUT_EXPIRED: GLenum;\n    readonly TIMEOUT_IGNORED: GLint64;\n    readonly TRANSFORM_FEEDBACK: GLenum;\n    readonly TRANSFORM_FEEDBACK_ACTIVE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_MODE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_SIZE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_START: GLenum;\n    readonly TRANSFORM_FEEDBACK_PAUSED: GLenum;\n    readonly TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN: GLenum;\n    readonly TRANSFORM_FEEDBACK_VARYINGS: GLenum;\n    readonly UNIFORM_ARRAY_STRIDE: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORMS: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES: GLenum;\n    readonly UNIFORM_BLOCK_BINDING: GLenum;\n    readonly UNIFORM_BLOCK_DATA_SIZE: GLenum;\n    readonly UNIFORM_BLOCK_INDEX: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER: GLenum;\n    readonly UNIFORM_BUFFER: GLenum;\n    readonly UNIFORM_BUFFER_BINDING: GLenum;\n    readonly UNIFORM_BUFFER_OFFSET_ALIGNMENT: GLenum;\n    readonly UNIFORM_BUFFER_SIZE: GLenum;\n    readonly UNIFORM_BUFFER_START: GLenum;\n    readonly UNIFORM_IS_ROW_MAJOR: GLenum;\n    readonly UNIFORM_MATRIX_STRIDE: GLenum;\n    readonly UNIFORM_OFFSET: GLenum;\n    readonly UNIFORM_SIZE: GLenum;\n    readonly UNIFORM_TYPE: GLenum;\n    readonly UNPACK_IMAGE_HEIGHT: GLenum;\n    readonly UNPACK_ROW_LENGTH: GLenum;\n    readonly UNPACK_SKIP_IMAGES: GLenum;\n    readonly UNPACK_SKIP_PIXELS: GLenum;\n    readonly UNPACK_SKIP_ROWS: GLenum;\n    readonly UNSIGNALED: GLenum;\n    readonly UNSIGNED_INT_10F_11F_11F_REV: GLenum;\n    readonly UNSIGNED_INT_24_8: GLenum;\n    readonly UNSIGNED_INT_2_10_10_10_REV: GLenum;\n    readonly UNSIGNED_INT_5_9_9_9_REV: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_3D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_CUBE: GLenum;\n    readonly UNSIGNED_INT_VEC2: GLenum;\n    readonly UNSIGNED_INT_VEC3: GLenum;\n    readonly UNSIGNED_INT_VEC4: GLenum;\n    readonly UNSIGNED_NORMALIZED: GLenum;\n    readonly VERTEX_ARRAY_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_INTEGER: GLenum;\n    readonly WAIT_FAILED: GLenum;\n};\n\ninterface WebGL2RenderingContextBase {\n    beginQuery(target: GLenum, query: WebGLQuery): void;\n    beginTransformFeedback(primitiveMode: GLenum): void;\n    bindBufferBase(target: GLenum, index: GLuint, buffer: WebGLBuffer | null): void;\n    bindBufferRange(target: GLenum, index: GLuint, buffer: WebGLBuffer | null, offset: GLintptr, size: GLsizeiptr): void;\n    bindSampler(unit: GLuint, sampler: WebGLSampler | null): void;\n    bindTransformFeedback(target: GLenum, tf: WebGLTransformFeedback | null): void;\n    bindVertexArray(array: WebGLVertexArrayObject | null): void;\n    blitFramebuffer(srcX0: GLint, srcY0: GLint, srcX1: GLint, srcY1: GLint, dstX0: GLint, dstY0: GLint, dstX1: GLint, dstY1: GLint, mask: GLbitfield, filter: GLenum): void;\n    clearBufferfi(buffer: GLenum, drawbuffer: GLint, depth: GLfloat, stencil: GLint): void;\n    clearBufferfv(buffer: GLenum, drawbuffer: GLint, values: Float32List, srcOffset?: GLuint): void;\n    clearBufferiv(buffer: GLenum, drawbuffer: GLint, values: Int32List, srcOffset?: GLuint): void;\n    clearBufferuiv(buffer: GLenum, drawbuffer: GLint, values: Uint32List, srcOffset?: GLuint): void;\n    clientWaitSync(sync: WebGLSync, flags: GLbitfield, timeout: GLuint64): GLenum;\n    compressedTexImage3D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexImage3D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    compressedTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    copyBufferSubData(readTarget: GLenum, writeTarget: GLenum, readOffset: GLintptr, writeOffset: GLintptr, size: GLsizeiptr): void;\n    copyTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    createQuery(): WebGLQuery | null;\n    createSampler(): WebGLSampler | null;\n    createTransformFeedback(): WebGLTransformFeedback | null;\n    createVertexArray(): WebGLVertexArrayObject | null;\n    deleteQuery(query: WebGLQuery | null): void;\n    deleteSampler(sampler: WebGLSampler | null): void;\n    deleteSync(sync: WebGLSync | null): void;\n    deleteTransformFeedback(tf: WebGLTransformFeedback | null): void;\n    deleteVertexArray(vertexArray: WebGLVertexArrayObject | null): void;\n    drawArraysInstanced(mode: GLenum, first: GLint, count: GLsizei, instanceCount: GLsizei): void;\n    drawBuffers(buffers: GLenum[]): void;\n    drawElementsInstanced(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr, instanceCount: GLsizei): void;\n    drawRangeElements(mode: GLenum, start: GLuint, end: GLuint, count: GLsizei, type: GLenum, offset: GLintptr): void;\n    endQuery(target: GLenum): void;\n    endTransformFeedback(): void;\n    fenceSync(condition: GLenum, flags: GLbitfield): WebGLSync | null;\n    framebufferTextureLayer(target: GLenum, attachment: GLenum, texture: WebGLTexture | null, level: GLint, layer: GLint): void;\n    getActiveUniformBlockName(program: WebGLProgram, uniformBlockIndex: GLuint): string | null;\n    getActiveUniformBlockParameter(program: WebGLProgram, uniformBlockIndex: GLuint, pname: GLenum): any;\n    getActiveUniforms(program: WebGLProgram, uniformIndices: GLuint[], pname: GLenum): any;\n    getBufferSubData(target: GLenum, srcByteOffset: GLintptr, dstBuffer: ArrayBufferView, dstOffset?: GLuint, length?: GLuint): void;\n    getFragDataLocation(program: WebGLProgram, name: string): GLint;\n    getIndexedParameter(target: GLenum, index: GLuint): any;\n    getInternalformatParameter(target: GLenum, internalformat: GLenum, pname: GLenum): any;\n    getQuery(target: GLenum, pname: GLenum): WebGLQuery | null;\n    getQueryParameter(query: WebGLQuery, pname: GLenum): any;\n    getSamplerParameter(sampler: WebGLSampler, pname: GLenum): any;\n    getSyncParameter(sync: WebGLSync, pname: GLenum): any;\n    getTransformFeedbackVarying(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getUniformBlockIndex(program: WebGLProgram, uniformBlockName: string): GLuint;\n    getUniformIndices(program: WebGLProgram, uniformNames: string[]): GLuint[] | null;\n    invalidateFramebuffer(target: GLenum, attachments: GLenum[]): void;\n    invalidateSubFramebuffer(target: GLenum, attachments: GLenum[], x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    isQuery(query: WebGLQuery | null): GLboolean;\n    isSampler(sampler: WebGLSampler | null): GLboolean;\n    isSync(sync: WebGLSync | null): GLboolean;\n    isTransformFeedback(tf: WebGLTransformFeedback | null): GLboolean;\n    isVertexArray(vertexArray: WebGLVertexArrayObject | null): GLboolean;\n    pauseTransformFeedback(): void;\n    readBuffer(src: GLenum): void;\n    renderbufferStorageMultisample(target: GLenum, samples: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    resumeTransformFeedback(): void;\n    samplerParameterf(sampler: WebGLSampler, pname: GLenum, param: GLfloat): void;\n    samplerParameteri(sampler: WebGLSampler, pname: GLenum, param: GLint): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView | null): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    texStorage2D(target: GLenum, levels: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    texStorage3D(target: GLenum, levels: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, srcData: ArrayBufferView | null, srcOffset?: GLuint): void;\n    transformFeedbackVaryings(program: WebGLProgram, varyings: string[], bufferMode: GLenum): void;\n    uniform1ui(location: WebGLUniformLocation | null, v0: GLuint): void;\n    uniform1uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint): void;\n    uniform2uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint, v2: GLuint): void;\n    uniform3uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint, v2: GLuint, v3: GLuint): void;\n    uniform4uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformBlockBinding(program: WebGLProgram, uniformBlockIndex: GLuint, uniformBlockBinding: GLuint): void;\n    uniformMatrix2x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    vertexAttribDivisor(index: GLuint, divisor: GLuint): void;\n    vertexAttribI4i(index: GLuint, x: GLint, y: GLint, z: GLint, w: GLint): void;\n    vertexAttribI4iv(index: GLuint, values: Int32List): void;\n    vertexAttribI4ui(index: GLuint, x: GLuint, y: GLuint, z: GLuint, w: GLuint): void;\n    vertexAttribI4uiv(index: GLuint, values: Uint32List): void;\n    vertexAttribIPointer(index: GLuint, size: GLint, type: GLenum, stride: GLsizei, offset: GLintptr): void;\n    waitSync(sync: WebGLSync, flags: GLbitfield, timeout: GLint64): void;\n    readonly ACTIVE_UNIFORM_BLOCKS: GLenum;\n    readonly ALREADY_SIGNALED: GLenum;\n    readonly ANY_SAMPLES_PASSED: GLenum;\n    readonly ANY_SAMPLES_PASSED_CONSERVATIVE: GLenum;\n    readonly COLOR: GLenum;\n    readonly COLOR_ATTACHMENT1: GLenum;\n    readonly COLOR_ATTACHMENT10: GLenum;\n    readonly COLOR_ATTACHMENT11: GLenum;\n    readonly COLOR_ATTACHMENT12: GLenum;\n    readonly COLOR_ATTACHMENT13: GLenum;\n    readonly COLOR_ATTACHMENT14: GLenum;\n    readonly COLOR_ATTACHMENT15: GLenum;\n    readonly COLOR_ATTACHMENT2: GLenum;\n    readonly COLOR_ATTACHMENT3: GLenum;\n    readonly COLOR_ATTACHMENT4: GLenum;\n    readonly COLOR_ATTACHMENT5: GLenum;\n    readonly COLOR_ATTACHMENT6: GLenum;\n    readonly COLOR_ATTACHMENT7: GLenum;\n    readonly COLOR_ATTACHMENT8: GLenum;\n    readonly COLOR_ATTACHMENT9: GLenum;\n    readonly COMPARE_REF_TO_TEXTURE: GLenum;\n    readonly CONDITION_SATISFIED: GLenum;\n    readonly COPY_READ_BUFFER: GLenum;\n    readonly COPY_READ_BUFFER_BINDING: GLenum;\n    readonly COPY_WRITE_BUFFER: GLenum;\n    readonly COPY_WRITE_BUFFER_BINDING: GLenum;\n    readonly CURRENT_QUERY: GLenum;\n    readonly DEPTH: GLenum;\n    readonly DEPTH24_STENCIL8: GLenum;\n    readonly DEPTH32F_STENCIL8: GLenum;\n    readonly DEPTH_COMPONENT24: GLenum;\n    readonly DEPTH_COMPONENT32F: GLenum;\n    readonly DRAW_BUFFER0: GLenum;\n    readonly DRAW_BUFFER1: GLenum;\n    readonly DRAW_BUFFER10: GLenum;\n    readonly DRAW_BUFFER11: GLenum;\n    readonly DRAW_BUFFER12: GLenum;\n    readonly DRAW_BUFFER13: GLenum;\n    readonly DRAW_BUFFER14: GLenum;\n    readonly DRAW_BUFFER15: GLenum;\n    readonly DRAW_BUFFER2: GLenum;\n    readonly DRAW_BUFFER3: GLenum;\n    readonly DRAW_BUFFER4: GLenum;\n    readonly DRAW_BUFFER5: GLenum;\n    readonly DRAW_BUFFER6: GLenum;\n    readonly DRAW_BUFFER7: GLenum;\n    readonly DRAW_BUFFER8: GLenum;\n    readonly DRAW_BUFFER9: GLenum;\n    readonly DRAW_FRAMEBUFFER: GLenum;\n    readonly DRAW_FRAMEBUFFER_BINDING: GLenum;\n    readonly DYNAMIC_COPY: GLenum;\n    readonly DYNAMIC_READ: GLenum;\n    readonly FLOAT_32_UNSIGNED_INT_24_8_REV: GLenum;\n    readonly FLOAT_MAT2x3: GLenum;\n    readonly FLOAT_MAT2x4: GLenum;\n    readonly FLOAT_MAT3x2: GLenum;\n    readonly FLOAT_MAT3x4: GLenum;\n    readonly FLOAT_MAT4x2: GLenum;\n    readonly FLOAT_MAT4x3: GLenum;\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_BLUE_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_GREEN_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_RED_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER: GLenum;\n    readonly FRAMEBUFFER_DEFAULT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: GLenum;\n    readonly HALF_FLOAT: GLenum;\n    readonly INTERLEAVED_ATTRIBS: GLenum;\n    readonly INT_2_10_10_10_REV: GLenum;\n    readonly INT_SAMPLER_2D: GLenum;\n    readonly INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly INT_SAMPLER_3D: GLenum;\n    readonly INT_SAMPLER_CUBE: GLenum;\n    readonly INVALID_INDEX: GLenum;\n    readonly MAX: GLenum;\n    readonly MAX_3D_TEXTURE_SIZE: GLenum;\n    readonly MAX_ARRAY_TEXTURE_LAYERS: GLenum;\n    readonly MAX_CLIENT_WAIT_TIMEOUT_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS: GLenum;\n    readonly MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_COMBINED_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_DRAW_BUFFERS: GLenum;\n    readonly MAX_ELEMENTS_INDICES: GLenum;\n    readonly MAX_ELEMENTS_VERTICES: GLenum;\n    readonly MAX_ELEMENT_INDEX: GLenum;\n    readonly MAX_FRAGMENT_INPUT_COMPONENTS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly MAX_SAMPLES: GLenum;\n    readonly MAX_SERVER_WAIT_TIMEOUT: GLenum;\n    readonly MAX_TEXTURE_LOD_BIAS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS: GLenum;\n    readonly MAX_UNIFORM_BLOCK_SIZE: GLenum;\n    readonly MAX_UNIFORM_BUFFER_BINDINGS: GLenum;\n    readonly MAX_VARYING_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_OUTPUT_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MIN: GLenum;\n    readonly MIN_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly OBJECT_TYPE: GLenum;\n    readonly PACK_ROW_LENGTH: GLenum;\n    readonly PACK_SKIP_PIXELS: GLenum;\n    readonly PACK_SKIP_ROWS: GLenum;\n    readonly PIXEL_PACK_BUFFER: GLenum;\n    readonly PIXEL_PACK_BUFFER_BINDING: GLenum;\n    readonly PIXEL_UNPACK_BUFFER: GLenum;\n    readonly PIXEL_UNPACK_BUFFER_BINDING: GLenum;\n    readonly QUERY_RESULT: GLenum;\n    readonly QUERY_RESULT_AVAILABLE: GLenum;\n    readonly R11F_G11F_B10F: GLenum;\n    readonly R16F: GLenum;\n    readonly R16I: GLenum;\n    readonly R16UI: GLenum;\n    readonly R32F: GLenum;\n    readonly R32I: GLenum;\n    readonly R32UI: GLenum;\n    readonly R8: GLenum;\n    readonly R8I: GLenum;\n    readonly R8UI: GLenum;\n    readonly R8_SNORM: GLenum;\n    readonly RASTERIZER_DISCARD: GLenum;\n    readonly READ_BUFFER: GLenum;\n    readonly READ_FRAMEBUFFER: GLenum;\n    readonly READ_FRAMEBUFFER_BINDING: GLenum;\n    readonly RED: GLenum;\n    readonly RED_INTEGER: GLenum;\n    readonly RENDERBUFFER_SAMPLES: GLenum;\n    readonly RG: GLenum;\n    readonly RG16F: GLenum;\n    readonly RG16I: GLenum;\n    readonly RG16UI: GLenum;\n    readonly RG32F: GLenum;\n    readonly RG32I: GLenum;\n    readonly RG32UI: GLenum;\n    readonly RG8: GLenum;\n    readonly RG8I: GLenum;\n    readonly RG8UI: GLenum;\n    readonly RG8_SNORM: GLenum;\n    readonly RGB10_A2: GLenum;\n    readonly RGB10_A2UI: GLenum;\n    readonly RGB16F: GLenum;\n    readonly RGB16I: GLenum;\n    readonly RGB16UI: GLenum;\n    readonly RGB32F: GLenum;\n    readonly RGB32I: GLenum;\n    readonly RGB32UI: GLenum;\n    readonly RGB8: GLenum;\n    readonly RGB8I: GLenum;\n    readonly RGB8UI: GLenum;\n    readonly RGB8_SNORM: GLenum;\n    readonly RGB9_E5: GLenum;\n    readonly RGBA16F: GLenum;\n    readonly RGBA16I: GLenum;\n    readonly RGBA16UI: GLenum;\n    readonly RGBA32F: GLenum;\n    readonly RGBA32I: GLenum;\n    readonly RGBA32UI: GLenum;\n    readonly RGBA8: GLenum;\n    readonly RGBA8I: GLenum;\n    readonly RGBA8UI: GLenum;\n    readonly RGBA8_SNORM: GLenum;\n    readonly RGBA_INTEGER: GLenum;\n    readonly RGB_INTEGER: GLenum;\n    readonly RG_INTEGER: GLenum;\n    readonly SAMPLER_2D_ARRAY: GLenum;\n    readonly SAMPLER_2D_ARRAY_SHADOW: GLenum;\n    readonly SAMPLER_2D_SHADOW: GLenum;\n    readonly SAMPLER_3D: GLenum;\n    readonly SAMPLER_BINDING: GLenum;\n    readonly SAMPLER_CUBE_SHADOW: GLenum;\n    readonly SEPARATE_ATTRIBS: GLenum;\n    readonly SIGNALED: GLenum;\n    readonly SIGNED_NORMALIZED: GLenum;\n    readonly SRGB: GLenum;\n    readonly SRGB8: GLenum;\n    readonly SRGB8_ALPHA8: GLenum;\n    readonly STATIC_COPY: GLenum;\n    readonly STATIC_READ: GLenum;\n    readonly STENCIL: GLenum;\n    readonly STREAM_COPY: GLenum;\n    readonly STREAM_READ: GLenum;\n    readonly SYNC_CONDITION: GLenum;\n    readonly SYNC_FENCE: GLenum;\n    readonly SYNC_FLAGS: GLenum;\n    readonly SYNC_FLUSH_COMMANDS_BIT: GLenum;\n    readonly SYNC_GPU_COMMANDS_COMPLETE: GLenum;\n    readonly SYNC_STATUS: GLenum;\n    readonly TEXTURE_2D_ARRAY: GLenum;\n    readonly TEXTURE_3D: GLenum;\n    readonly TEXTURE_BASE_LEVEL: GLenum;\n    readonly TEXTURE_BINDING_2D_ARRAY: GLenum;\n    readonly TEXTURE_BINDING_3D: GLenum;\n    readonly TEXTURE_COMPARE_FUNC: GLenum;\n    readonly TEXTURE_COMPARE_MODE: GLenum;\n    readonly TEXTURE_IMMUTABLE_FORMAT: GLenum;\n    readonly TEXTURE_IMMUTABLE_LEVELS: GLenum;\n    readonly TEXTURE_MAX_LEVEL: GLenum;\n    readonly TEXTURE_MAX_LOD: GLenum;\n    readonly TEXTURE_MIN_LOD: GLenum;\n    readonly TEXTURE_WRAP_R: GLenum;\n    readonly TIMEOUT_EXPIRED: GLenum;\n    readonly TIMEOUT_IGNORED: GLint64;\n    readonly TRANSFORM_FEEDBACK: GLenum;\n    readonly TRANSFORM_FEEDBACK_ACTIVE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_MODE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_SIZE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_START: GLenum;\n    readonly TRANSFORM_FEEDBACK_PAUSED: GLenum;\n    readonly TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN: GLenum;\n    readonly TRANSFORM_FEEDBACK_VARYINGS: GLenum;\n    readonly UNIFORM_ARRAY_STRIDE: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORMS: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES: GLenum;\n    readonly UNIFORM_BLOCK_BINDING: GLenum;\n    readonly UNIFORM_BLOCK_DATA_SIZE: GLenum;\n    readonly UNIFORM_BLOCK_INDEX: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER: GLenum;\n    readonly UNIFORM_BUFFER: GLenum;\n    readonly UNIFORM_BUFFER_BINDING: GLenum;\n    readonly UNIFORM_BUFFER_OFFSET_ALIGNMENT: GLenum;\n    readonly UNIFORM_BUFFER_SIZE: GLenum;\n    readonly UNIFORM_BUFFER_START: GLenum;\n    readonly UNIFORM_IS_ROW_MAJOR: GLenum;\n    readonly UNIFORM_MATRIX_STRIDE: GLenum;\n    readonly UNIFORM_OFFSET: GLenum;\n    readonly UNIFORM_SIZE: GLenum;\n    readonly UNIFORM_TYPE: GLenum;\n    readonly UNPACK_IMAGE_HEIGHT: GLenum;\n    readonly UNPACK_ROW_LENGTH: GLenum;\n    readonly UNPACK_SKIP_IMAGES: GLenum;\n    readonly UNPACK_SKIP_PIXELS: GLenum;\n    readonly UNPACK_SKIP_ROWS: GLenum;\n    readonly UNSIGNALED: GLenum;\n    readonly UNSIGNED_INT_10F_11F_11F_REV: GLenum;\n    readonly UNSIGNED_INT_24_8: GLenum;\n    readonly UNSIGNED_INT_2_10_10_10_REV: GLenum;\n    readonly UNSIGNED_INT_5_9_9_9_REV: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_3D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_CUBE: GLenum;\n    readonly UNSIGNED_INT_VEC2: GLenum;\n    readonly UNSIGNED_INT_VEC3: GLenum;\n    readonly UNSIGNED_INT_VEC4: GLenum;\n    readonly UNSIGNED_NORMALIZED: GLenum;\n    readonly VERTEX_ARRAY_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_INTEGER: GLenum;\n    readonly WAIT_FAILED: GLenum;\n}\n\ninterface WebGL2RenderingContextOverloads {\n    bufferData(target: GLenum, size: GLsizeiptr, usage: GLenum): void;\n    bufferData(target: GLenum, srcData: BufferSource | null, usage: GLenum): void;\n    bufferData(target: GLenum, srcData: ArrayBufferView, usage: GLenum, srcOffset: GLuint, length?: GLuint): void;\n    bufferSubData(target: GLenum, dstByteOffset: GLintptr, srcData: BufferSource): void;\n    bufferSubData(target: GLenum, dstByteOffset: GLintptr, srcData: ArrayBufferView, srcOffset: GLuint, length?: GLuint): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, dstData: ArrayBufferView | null): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, offset: GLintptr): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, dstData: ArrayBufferView, dstOffset: GLuint): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    uniform1fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform1iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n}\n\n/** Part of the WebGL API and represents the information returned by calling the WebGLRenderingContext.getActiveAttrib() and WebGLRenderingContext.getActiveUniform() methods. */\ninterface WebGLActiveInfo {\n    readonly name: string;\n    readonly size: GLint;\n    readonly type: GLenum;\n}\n\ndeclare var WebGLActiveInfo: {\n    prototype: WebGLActiveInfo;\n    new(): WebGLActiveInfo;\n};\n\n/** Part of the WebGL API and represents an opaque buffer object storing data such as vertices or colors. */\ninterface WebGLBuffer extends WebGLObject {\n}\n\ndeclare var WebGLBuffer: {\n    prototype: WebGLBuffer;\n    new(): WebGLBuffer;\n};\n\n/** The WebContextEvent interface is part of the WebGL API and is an interface for an event that is generated in response to a status change to the WebGL rendering context. */\ninterface WebGLContextEvent extends Event {\n    readonly statusMessage: string;\n}\n\ndeclare var WebGLContextEvent: {\n    prototype: WebGLContextEvent;\n    new(type: string, eventInit?: WebGLContextEventInit): WebGLContextEvent;\n};\n\n/** Part of the WebGL API and represents a collection of buffers that serve as a rendering destination. */\ninterface WebGLFramebuffer extends WebGLObject {\n}\n\ndeclare var WebGLFramebuffer: {\n    prototype: WebGLFramebuffer;\n    new(): WebGLFramebuffer;\n};\n\ninterface WebGLObject {\n}\n\ndeclare var WebGLObject: {\n    prototype: WebGLObject;\n    new(): WebGLObject;\n};\n\n/** The WebGLProgram is part of the WebGL API and is a combination of two compiled WebGLShaders consisting of a vertex shader and a fragment shader (both written in GLSL). */\ninterface WebGLProgram extends WebGLObject {\n}\n\ndeclare var WebGLProgram: {\n    prototype: WebGLProgram;\n    new(): WebGLProgram;\n};\n\ninterface WebGLQuery extends WebGLObject {\n}\n\ndeclare var WebGLQuery: {\n    prototype: WebGLQuery;\n    new(): WebGLQuery;\n};\n\n/** Part of the WebGL API and represents a buffer that can contain an image, or can be source or target of an rendering operation. */\ninterface WebGLRenderbuffer extends WebGLObject {\n}\n\ndeclare var WebGLRenderbuffer: {\n    prototype: WebGLRenderbuffer;\n    new(): WebGLRenderbuffer;\n};\n\n/** Provides an interface to the OpenGL ES 2.0 graphics rendering context for the drawing surface of an HTML <canvas> element. */\ninterface WebGLRenderingContext extends WebGLRenderingContextBase, WebGLRenderingContextOverloads {\n}\n\ndeclare var WebGLRenderingContext: {\n    prototype: WebGLRenderingContext;\n    new(): WebGLRenderingContext;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n};\n\ninterface WebGLRenderingContextBase {\n    readonly canvas: HTMLCanvasElement | OffscreenCanvas;\n    readonly drawingBufferHeight: GLsizei;\n    readonly drawingBufferWidth: GLsizei;\n    activeTexture(texture: GLenum): void;\n    attachShader(program: WebGLProgram, shader: WebGLShader): void;\n    bindAttribLocation(program: WebGLProgram, index: GLuint, name: string): void;\n    bindBuffer(target: GLenum, buffer: WebGLBuffer | null): void;\n    bindFramebuffer(target: GLenum, framebuffer: WebGLFramebuffer | null): void;\n    bindRenderbuffer(target: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    bindTexture(target: GLenum, texture: WebGLTexture | null): void;\n    blendColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    blendEquation(mode: GLenum): void;\n    blendEquationSeparate(modeRGB: GLenum, modeAlpha: GLenum): void;\n    blendFunc(sfactor: GLenum, dfactor: GLenum): void;\n    blendFuncSeparate(srcRGB: GLenum, dstRGB: GLenum, srcAlpha: GLenum, dstAlpha: GLenum): void;\n    checkFramebufferStatus(target: GLenum): GLenum;\n    clear(mask: GLbitfield): void;\n    clearColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    clearDepth(depth: GLclampf): void;\n    clearStencil(s: GLint): void;\n    colorMask(red: GLboolean, green: GLboolean, blue: GLboolean, alpha: GLboolean): void;\n    compileShader(shader: WebGLShader): void;\n    copyTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, x: GLint, y: GLint, width: GLsizei, height: GLsizei, border: GLint): void;\n    copyTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    createBuffer(): WebGLBuffer | null;\n    createFramebuffer(): WebGLFramebuffer | null;\n    createProgram(): WebGLProgram | null;\n    createRenderbuffer(): WebGLRenderbuffer | null;\n    createShader(type: GLenum): WebGLShader | null;\n    createTexture(): WebGLTexture | null;\n    cullFace(mode: GLenum): void;\n    deleteBuffer(buffer: WebGLBuffer | null): void;\n    deleteFramebuffer(framebuffer: WebGLFramebuffer | null): void;\n    deleteProgram(program: WebGLProgram | null): void;\n    deleteRenderbuffer(renderbuffer: WebGLRenderbuffer | null): void;\n    deleteShader(shader: WebGLShader | null): void;\n    deleteTexture(texture: WebGLTexture | null): void;\n    depthFunc(func: GLenum): void;\n    depthMask(flag: GLboolean): void;\n    depthRange(zNear: GLclampf, zFar: GLclampf): void;\n    detachShader(program: WebGLProgram, shader: WebGLShader): void;\n    disable(cap: GLenum): void;\n    disableVertexAttribArray(index: GLuint): void;\n    drawArrays(mode: GLenum, first: GLint, count: GLsizei): void;\n    drawElements(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr): void;\n    enable(cap: GLenum): void;\n    enableVertexAttribArray(index: GLuint): void;\n    finish(): void;\n    flush(): void;\n    framebufferRenderbuffer(target: GLenum, attachment: GLenum, renderbuffertarget: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    framebufferTexture2D(target: GLenum, attachment: GLenum, textarget: GLenum, texture: WebGLTexture | null, level: GLint): void;\n    frontFace(mode: GLenum): void;\n    generateMipmap(target: GLenum): void;\n    getActiveAttrib(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getActiveUniform(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getAttachedShaders(program: WebGLProgram): WebGLShader[] | null;\n    getAttribLocation(program: WebGLProgram, name: string): GLint;\n    getBufferParameter(target: GLenum, pname: GLenum): any;\n    getContextAttributes(): WebGLContextAttributes | null;\n    getError(): GLenum;\n    getExtension(extensionName: "EXT_blend_minmax"): EXT_blend_minmax | null;\n    getExtension(extensionName: "EXT_texture_filter_anisotropic"): EXT_texture_filter_anisotropic | null;\n    getExtension(extensionName: "EXT_frag_depth"): EXT_frag_depth | null;\n    getExtension(extensionName: "EXT_shader_texture_lod"): EXT_shader_texture_lod | null;\n    getExtension(extensionName: "EXT_sRGB"): EXT_sRGB | null;\n    getExtension(extensionName: "OES_vertex_array_object"): OES_vertex_array_object | null;\n    getExtension(extensionName: "WEBGL_color_buffer_float"): WEBGL_color_buffer_float | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_astc"): WEBGL_compressed_texture_astc | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc_srgb"): WEBGL_compressed_texture_s3tc_srgb | null;\n    getExtension(extensionName: "WEBGL_debug_shaders"): WEBGL_debug_shaders | null;\n    getExtension(extensionName: "WEBGL_draw_buffers"): WEBGL_draw_buffers | null;\n    getExtension(extensionName: "WEBGL_lose_context"): WEBGL_lose_context | null;\n    getExtension(extensionName: "WEBGL_depth_texture"): WEBGL_depth_texture | null;\n    getExtension(extensionName: "WEBGL_debug_renderer_info"): WEBGL_debug_renderer_info | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc"): WEBGL_compressed_texture_s3tc | null;\n    getExtension(extensionName: "OES_texture_half_float_linear"): OES_texture_half_float_linear | null;\n    getExtension(extensionName: "OES_texture_half_float"): OES_texture_half_float | null;\n    getExtension(extensionName: "OES_texture_float_linear"): OES_texture_float_linear | null;\n    getExtension(extensionName: "OES_texture_float"): OES_texture_float | null;\n    getExtension(extensionName: "OES_standard_derivatives"): OES_standard_derivatives | null;\n    getExtension(extensionName: "OES_element_index_uint"): OES_element_index_uint | null;\n    getExtension(extensionName: "ANGLE_instanced_arrays"): ANGLE_instanced_arrays | null;\n    getExtension(extensionName: string): any;\n    getFramebufferAttachmentParameter(target: GLenum, attachment: GLenum, pname: GLenum): any;\n    getParameter(pname: GLenum): any;\n    getProgramInfoLog(program: WebGLProgram): string | null;\n    getProgramParameter(program: WebGLProgram, pname: GLenum): any;\n    getRenderbufferParameter(target: GLenum, pname: GLenum): any;\n    getShaderInfoLog(shader: WebGLShader): string | null;\n    getShaderParameter(shader: WebGLShader, pname: GLenum): any;\n    getShaderPrecisionFormat(shadertype: GLenum, precisiontype: GLenum): WebGLShaderPrecisionFormat | null;\n    getShaderSource(shader: WebGLShader): string | null;\n    getSupportedExtensions(): string[] | null;\n    getTexParameter(target: GLenum, pname: GLenum): any;\n    getUniform(program: WebGLProgram, location: WebGLUniformLocation): any;\n    getUniformLocation(program: WebGLProgram, name: string): WebGLUniformLocation | null;\n    getVertexAttrib(index: GLuint, pname: GLenum): any;\n    getVertexAttribOffset(index: GLuint, pname: GLenum): GLintptr;\n    hint(target: GLenum, mode: GLenum): void;\n    isBuffer(buffer: WebGLBuffer | null): GLboolean;\n    isContextLost(): boolean;\n    isEnabled(cap: GLenum): GLboolean;\n    isFramebuffer(framebuffer: WebGLFramebuffer | null): GLboolean;\n    isProgram(program: WebGLProgram | null): GLboolean;\n    isRenderbuffer(renderbuffer: WebGLRenderbuffer | null): GLboolean;\n    isShader(shader: WebGLShader | null): GLboolean;\n    isTexture(texture: WebGLTexture | null): GLboolean;\n    lineWidth(width: GLfloat): void;\n    linkProgram(program: WebGLProgram): void;\n    pixelStorei(pname: GLenum, param: GLint | GLboolean): void;\n    polygonOffset(factor: GLfloat, units: GLfloat): void;\n    renderbufferStorage(target: GLenum, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    sampleCoverage(value: GLclampf, invert: GLboolean): void;\n    scissor(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    shaderSource(shader: WebGLShader, source: string): void;\n    stencilFunc(func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilFuncSeparate(face: GLenum, func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilMask(mask: GLuint): void;\n    stencilMaskSeparate(face: GLenum, mask: GLuint): void;\n    stencilOp(fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    stencilOpSeparate(face: GLenum, fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    texParameterf(target: GLenum, pname: GLenum, param: GLfloat): void;\n    texParameteri(target: GLenum, pname: GLenum, param: GLint): void;\n    uniform1f(location: WebGLUniformLocation | null, x: GLfloat): void;\n    uniform1i(location: WebGLUniformLocation | null, x: GLint): void;\n    uniform2f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat): void;\n    uniform2i(location: WebGLUniformLocation | null, x: GLint, y: GLint): void;\n    uniform3f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    uniform3i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint): void;\n    uniform4f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    uniform4i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint, w: GLint): void;\n    useProgram(program: WebGLProgram | null): void;\n    validateProgram(program: WebGLProgram): void;\n    vertexAttrib1f(index: GLuint, x: GLfloat): void;\n    vertexAttrib1fv(index: GLuint, values: Float32List): void;\n    vertexAttrib2f(index: GLuint, x: GLfloat, y: GLfloat): void;\n    vertexAttrib2fv(index: GLuint, values: Float32List): void;\n    vertexAttrib3f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    vertexAttrib3fv(index: GLuint, values: Float32List): void;\n    vertexAttrib4f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    vertexAttrib4fv(index: GLuint, values: Float32List): void;\n    vertexAttribPointer(index: GLuint, size: GLint, type: GLenum, normalized: GLboolean, stride: GLsizei, offset: GLintptr): void;\n    viewport(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n}\n\ninterface WebGLRenderingContextOverloads {\n    bufferData(target: GLenum, size: GLsizeiptr, usage: GLenum): void;\n    bufferData(target: GLenum, data: BufferSource | null, usage: GLenum): void;\n    bufferSubData(target: GLenum, offset: GLintptr, data: BufferSource): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, data: ArrayBufferView): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, data: ArrayBufferView): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    uniform1fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform1iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform2fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform2iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform3fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform3iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform4fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform4iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n}\n\ninterface WebGLSampler extends WebGLObject {\n}\n\ndeclare var WebGLSampler: {\n    prototype: WebGLSampler;\n    new(): WebGLSampler;\n};\n\n/** The WebGLShader is part of the WebGL API and can either be a vertex or a fragment shader. A WebGLProgram requires both types of shaders. */\ninterface WebGLShader extends WebGLObject {\n}\n\ndeclare var WebGLShader: {\n    prototype: WebGLShader;\n    new(): WebGLShader;\n};\n\n/** Part of the WebGL API and represents the information returned by calling the WebGLRenderingContext.getShaderPrecisionFormat() method. */\ninterface WebGLShaderPrecisionFormat {\n    readonly precision: GLint;\n    readonly rangeMax: GLint;\n    readonly rangeMin: GLint;\n}\n\ndeclare var WebGLShaderPrecisionFormat: {\n    prototype: WebGLShaderPrecisionFormat;\n    new(): WebGLShaderPrecisionFormat;\n};\n\ninterface WebGLSync extends WebGLObject {\n}\n\ndeclare var WebGLSync: {\n    prototype: WebGLSync;\n    new(): WebGLSync;\n};\n\n/** Part of the WebGL API and represents an opaque texture object providing storage and state for texturing operations. */\ninterface WebGLTexture extends WebGLObject {\n}\n\ndeclare var WebGLTexture: {\n    prototype: WebGLTexture;\n    new(): WebGLTexture;\n};\n\ninterface WebGLTransformFeedback extends WebGLObject {\n}\n\ndeclare var WebGLTransformFeedback: {\n    prototype: WebGLTransformFeedback;\n    new(): WebGLTransformFeedback;\n};\n\n/** Part of the WebGL API and represents the location of a uniform variable in a shader program. */\ninterface WebGLUniformLocation {\n}\n\ndeclare var WebGLUniformLocation: {\n    prototype: WebGLUniformLocation;\n    new(): WebGLUniformLocation;\n};\n\ninterface WebGLVertexArrayObject extends WebGLObject {\n}\n\ndeclare var WebGLVertexArrayObject: {\n    prototype: WebGLVertexArrayObject;\n    new(): WebGLVertexArrayObject;\n};\n\ninterface WebGLVertexArrayObjectOES extends WebGLObject {\n}\n\ninterface WebKitPoint {\n    x: number;\n    y: number;\n}\n\ndeclare var WebKitPoint: {\n    prototype: WebKitPoint;\n    new(x?: number, y?: number): WebKitPoint;\n};\n\ninterface WebSocketEventMap {\n    "close": CloseEvent;\n    "error": Event;\n    "message": MessageEvent;\n    "open": Event;\n}\n\n/** Provides the API for creating and managing a WebSocket connection to a server, as well as for sending and receiving data on the connection. */\ninterface WebSocket extends EventTarget {\n    /**\n     * Returns a string that indicates how binary data from the WebSocket object is exposed to scripts:\n     * \n     * Can be set, to change how binary data is returned. The default is "blob".\n     */\n    binaryType: BinaryType;\n    /**\n     * Returns the number of bytes of application data (UTF-8 text and binary data) that have been queued using send() but not yet been transmitted to the network.\n     * \n     * If the WebSocket connection is closed, this attribute\'s value will only increase with each call to the send() method. (The number does not reset to zero once the connection closes.)\n     */\n    readonly bufferedAmount: number;\n    /**\n     * Returns the extensions selected by the server, if any.\n     */\n    readonly extensions: string;\n    onclose: ((this: WebSocket, ev: CloseEvent) => any) | null;\n    onerror: ((this: WebSocket, ev: Event) => any) | null;\n    onmessage: ((this: WebSocket, ev: MessageEvent) => any) | null;\n    onopen: ((this: WebSocket, ev: Event) => any) | null;\n    /**\n     * Returns the subprotocol selected by the server, if any. It can be used in conjunction with the array form of the constructor\'s second argument to perform subprotocol negotiation.\n     */\n    readonly protocol: string;\n    /**\n     * Returns the state of the WebSocket object\'s connection. It can have the values described below.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the URL that was used to establish the WebSocket connection.\n     */\n    readonly url: string;\n    /**\n     * Closes the WebSocket connection, optionally using code as the the WebSocket connection close code and reason as the the WebSocket connection close reason.\n     */\n    close(code?: number, reason?: string): void;\n    /**\n     * Transmits data using the WebSocket connection. data can be a string, a Blob, an ArrayBuffer, or an ArrayBufferView.\n     */\n    send(data: string | ArrayBufferLike | Blob | ArrayBufferView): void;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    addEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var WebSocket: {\n    prototype: WebSocket;\n    new(url: string, protocols?: string | string[]): WebSocket;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n};\n\n/** Events that occur due to the user moving a mouse wheel or similar input device. */\ninterface WheelEvent extends MouseEvent {\n    readonly deltaMode: number;\n    readonly deltaX: number;\n    readonly deltaY: number;\n    readonly deltaZ: number;\n    readonly DOM_DELTA_LINE: number;\n    readonly DOM_DELTA_PAGE: number;\n    readonly DOM_DELTA_PIXEL: number;\n}\n\ndeclare var WheelEvent: {\n    prototype: WheelEvent;\n    new(type: string, eventInitDict?: WheelEventInit): WheelEvent;\n    readonly DOM_DELTA_LINE: number;\n    readonly DOM_DELTA_PAGE: number;\n    readonly DOM_DELTA_PIXEL: number;\n};\n\ninterface WindowEventMap extends GlobalEventHandlersEventMap, WindowEventHandlersEventMap {\n    "abort": UIEvent;\n    "afterprint": Event;\n    "beforeprint": Event;\n    "beforeunload": BeforeUnloadEvent;\n    "blur": FocusEvent;\n    "canplay": Event;\n    "canplaythrough": Event;\n    "change": Event;\n    "click": MouseEvent;\n    "compassneedscalibration": Event;\n    "contextmenu": MouseEvent;\n    "dblclick": MouseEvent;\n    "devicelight": DeviceLightEvent;\n    "devicemotion": DeviceMotionEvent;\n    "deviceorientation": DeviceOrientationEvent;\n    "deviceorientationabsolute": DeviceOrientationEvent;\n    "drag": DragEvent;\n    "dragend": DragEvent;\n    "dragenter": DragEvent;\n    "dragleave": DragEvent;\n    "dragover": DragEvent;\n    "dragstart": DragEvent;\n    "drop": DragEvent;\n    "durationchange": Event;\n    "emptied": Event;\n    "ended": Event;\n    "error": ErrorEvent;\n    "focus": FocusEvent;\n    "hashchange": HashChangeEvent;\n    "input": Event;\n    "invalid": Event;\n    "keydown": KeyboardEvent;\n    "keypress": KeyboardEvent;\n    "keyup": KeyboardEvent;\n    "load": Event;\n    "loadeddata": Event;\n    "loadedmetadata": Event;\n    "loadstart": Event;\n    "message": MessageEvent;\n    "mousedown": MouseEvent;\n    "mouseenter": MouseEvent;\n    "mouseleave": MouseEvent;\n    "mousemove": MouseEvent;\n    "mouseout": MouseEvent;\n    "mouseover": MouseEvent;\n    "mouseup": MouseEvent;\n    "mousewheel": Event;\n    "MSGestureChange": Event;\n    "MSGestureDoubleTap": Event;\n    "MSGestureEnd": Event;\n    "MSGestureHold": Event;\n    "MSGestureStart": Event;\n    "MSGestureTap": Event;\n    "MSInertiaStart": Event;\n    "MSPointerCancel": Event;\n    "MSPointerDown": Event;\n    "MSPointerEnter": Event;\n    "MSPointerLeave": Event;\n    "MSPointerMove": Event;\n    "MSPointerOut": Event;\n    "MSPointerOver": Event;\n    "MSPointerUp": Event;\n    "offline": Event;\n    "online": Event;\n    "orientationchange": Event;\n    "pagehide": PageTransitionEvent;\n    "pageshow": PageTransitionEvent;\n    "pause": Event;\n    "play": Event;\n    "playing": Event;\n    "popstate": PopStateEvent;\n    "progress": ProgressEvent<Window>;\n    "ratechange": Event;\n    "readystatechange": ProgressEvent<Window>;\n    "reset": Event;\n    "resize": UIEvent;\n    "scroll": Event;\n    "seeked": Event;\n    "seeking": Event;\n    "select": Event;\n    "stalled": Event;\n    "storage": StorageEvent;\n    "submit": Event;\n    "suspend": Event;\n    "timeupdate": Event;\n    "unload": Event;\n    "volumechange": Event;\n    "vrdisplayactivate": Event;\n    "vrdisplayblur": Event;\n    "vrdisplayconnect": Event;\n    "vrdisplaydeactivate": Event;\n    "vrdisplaydisconnect": Event;\n    "vrdisplayfocus": Event;\n    "vrdisplaypointerrestricted": Event;\n    "vrdisplaypointerunrestricted": Event;\n    "vrdisplaypresentchange": Event;\n    "waiting": Event;\n}\n\n/** A window containing a DOM document; the document property points to the DOM document loaded in that window. */\ninterface Window extends EventTarget, AnimationFrameProvider, GlobalEventHandlers, WindowEventHandlers, WindowLocalStorage, WindowOrWorkerGlobalScope, WindowSessionStorage {\n    readonly applicationCache: ApplicationCache;\n    readonly clientInformation: Navigator;\n    readonly closed: boolean;\n    customElements: CustomElementRegistry;\n    defaultStatus: string;\n    readonly devicePixelRatio: number;\n    readonly doNotTrack: string;\n    readonly document: Document;\n    /** @deprecated */\n    readonly event: Event | undefined;\n    /** @deprecated */\n    readonly external: External;\n    readonly frameElement: Element;\n    readonly frames: Window;\n    readonly history: History;\n    readonly innerHeight: number;\n    readonly innerWidth: number;\n    readonly length: number;\n    location: Location;\n    readonly locationbar: BarProp;\n    readonly menubar: BarProp;\n    readonly msContentScript: ExtensionScriptApis;\n    name: string;\n    readonly navigator: Navigator;\n    offscreenBuffering: string | boolean;\n    oncompassneedscalibration: ((this: Window, ev: Event) => any) | null;\n    ondevicelight: ((this: Window, ev: DeviceLightEvent) => any) | null;\n    ondevicemotion: ((this: Window, ev: DeviceMotionEvent) => any) | null;\n    ondeviceorientation: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\n    ondeviceorientationabsolute: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\n    onmousewheel: ((this: Window, ev: Event) => any) | null;\n    onmsgesturechange: ((this: Window, ev: Event) => any) | null;\n    onmsgesturedoubletap: ((this: Window, ev: Event) => any) | null;\n    onmsgestureend: ((this: Window, ev: Event) => any) | null;\n    onmsgesturehold: ((this: Window, ev: Event) => any) | null;\n    onmsgesturestart: ((this: Window, ev: Event) => any) | null;\n    onmsgesturetap: ((this: Window, ev: Event) => any) | null;\n    onmsinertiastart: ((this: Window, ev: Event) => any) | null;\n    onmspointercancel: ((this: Window, ev: Event) => any) | null;\n    onmspointerdown: ((this: Window, ev: Event) => any) | null;\n    onmspointerenter: ((this: Window, ev: Event) => any) | null;\n    onmspointerleave: ((this: Window, ev: Event) => any) | null;\n    onmspointermove: ((this: Window, ev: Event) => any) | null;\n    onmspointerout: ((this: Window, ev: Event) => any) | null;\n    onmspointerover: ((this: Window, ev: Event) => any) | null;\n    onmspointerup: ((this: Window, ev: Event) => any) | null;\n    /** @deprecated */\n    onorientationchange: ((this: Window, ev: Event) => any) | null;\n    onreadystatechange: ((this: Window, ev: ProgressEvent<Window>) => any) | null;\n    onvrdisplayactivate: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayblur: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayconnect: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaydeactivate: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaydisconnect: ((this: Window, ev: Event) => any) | null;\n    onvrdisplayfocus: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypointerrestricted: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypointerunrestricted: ((this: Window, ev: Event) => any) | null;\n    onvrdisplaypresentchange: ((this: Window, ev: Event) => any) | null;\n    opener: any;\n    /** @deprecated */\n    readonly orientation: string | number;\n    readonly outerHeight: number;\n    readonly outerWidth: number;\n    readonly pageXOffset: number;\n    readonly pageYOffset: number;\n    readonly parent: Window;\n    readonly personalbar: BarProp;\n    readonly screen: Screen;\n    readonly screenLeft: number;\n    readonly screenTop: number;\n    readonly screenX: number;\n    readonly screenY: number;\n    readonly scrollX: number;\n    readonly scrollY: number;\n    readonly scrollbars: BarProp;\n    readonly self: Window & typeof globalThis;\n    readonly speechSynthesis: SpeechSynthesis;\n    status: string;\n    readonly statusbar: BarProp;\n    readonly styleMedia: StyleMedia;\n    readonly toolbar: BarProp;\n    readonly top: Window;\n    readonly visualViewport: VisualViewport;\n    readonly window: Window & typeof globalThis;\n    alert(message?: any): void;\n    blur(): void;\n    /** @deprecated */\n    captureEvents(): void;\n    close(): void;\n    confirm(message?: string): boolean;\n    departFocus(navigationReason: NavigationReason, origin: FocusNavigationOrigin): void;\n    focus(): void;\n    getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\n    getMatchedCSSRules(elt: Element, pseudoElt?: string | null): CSSRuleList;\n    getSelection(): Selection | null;\n    matchMedia(query: string): MediaQueryList;\n    moveBy(x: number, y: number): void;\n    moveTo(x: number, y: number): void;\n    msWriteProfilerMark(profilerMarkName: string): void;\n    open(url?: string, target?: string, features?: string, replace?: boolean): Window | null;\n    postMessage(message: any, targetOrigin: string, transfer?: Transferable[]): void;\n    print(): void;\n    prompt(message?: string, _default?: string): string | null;\n    /** @deprecated */\n    releaseEvents(): void;\n    resizeBy(x: number, y: number): void;\n    resizeTo(width: number, height: number): void;\n    scroll(options?: ScrollToOptions): void;\n    scroll(x: number, y: number): void;\n    scrollBy(options?: ScrollToOptions): void;\n    scrollBy(x: number, y: number): void;\n    scrollTo(options?: ScrollToOptions): void;\n    scrollTo(x: number, y: number): void;\n    stop(): void;\n    webkitCancelAnimationFrame(handle: number): void;\n    webkitConvertPointFromNodeToPage(node: Node, pt: WebKitPoint): WebKitPoint;\n    webkitConvertPointFromPageToNode(node: Node, pt: WebKitPoint): WebKitPoint;\n    webkitRequestAnimationFrame(callback: FrameRequestCallback): number;\n    addEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n    [index: number]: Window;\n}\n\ndeclare var Window: {\n    prototype: Window;\n    new(): Window;\n};\n\ninterface WindowEventHandlersEventMap {\n    "afterprint": Event;\n    "beforeprint": Event;\n    "beforeunload": BeforeUnloadEvent;\n    "hashchange": HashChangeEvent;\n    "languagechange": Event;\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n    "offline": Event;\n    "online": Event;\n    "pagehide": PageTransitionEvent;\n    "pageshow": PageTransitionEvent;\n    "popstate": PopStateEvent;\n    "rejectionhandled": PromiseRejectionEvent;\n    "storage": StorageEvent;\n    "unhandledrejection": PromiseRejectionEvent;\n    "unload": Event;\n}\n\ninterface WindowEventHandlers {\n    onafterprint: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onbeforeprint: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onbeforeunload: ((this: WindowEventHandlers, ev: BeforeUnloadEvent) => any) | null;\n    onhashchange: ((this: WindowEventHandlers, ev: HashChangeEvent) => any) | null;\n    onlanguagechange: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onmessage: ((this: WindowEventHandlers, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: WindowEventHandlers, ev: MessageEvent) => any) | null;\n    onoffline: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    ononline: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    onpagehide: ((this: WindowEventHandlers, ev: PageTransitionEvent) => any) | null;\n    onpageshow: ((this: WindowEventHandlers, ev: PageTransitionEvent) => any) | null;\n    onpopstate: ((this: WindowEventHandlers, ev: PopStateEvent) => any) | null;\n    onrejectionhandled: ((this: WindowEventHandlers, ev: PromiseRejectionEvent) => any) | null;\n    onstorage: ((this: WindowEventHandlers, ev: StorageEvent) => any) | null;\n    onunhandledrejection: ((this: WindowEventHandlers, ev: PromiseRejectionEvent) => any) | null;\n    onunload: ((this: WindowEventHandlers, ev: Event) => any) | null;\n    addEventListener<K extends keyof WindowEventHandlersEventMap>(type: K, listener: (this: WindowEventHandlers, ev: WindowEventHandlersEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WindowEventHandlersEventMap>(type: K, listener: (this: WindowEventHandlers, ev: WindowEventHandlersEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface WindowLocalStorage {\n    readonly localStorage: Storage;\n}\n\ninterface WindowOrWorkerGlobalScope {\n    readonly caches: CacheStorage;\n    readonly crypto: Crypto;\n    readonly indexedDB: IDBFactory;\n    readonly isSecureContext: boolean;\n    readonly origin: string;\n    readonly performance: Performance;\n    atob(data: string): string;\n    btoa(data: string): string;\n    clearInterval(handle?: number): void;\n    clearTimeout(handle?: number): void;\n    createImageBitmap(image: ImageBitmapSource, options?: ImageBitmapOptions): Promise<ImageBitmap>;\n    createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number, options?: ImageBitmapOptions): Promise<ImageBitmap>;\n    fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\n    queueMicrotask(callback: VoidFunction): void;\n    setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n    setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n}\n\ninterface WindowSessionStorage {\n    readonly sessionStorage: Storage;\n}\n\ninterface WorkerEventMap extends AbstractWorkerEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** This Web Workers API interface represents a background task that can be easily created and can send messages back to its creator. Creating a worker is as simple as calling the Worker() constructor and specifying a script to be run in the worker thread. */\ninterface Worker extends EventTarget, AbstractWorker {\n    onmessage: ((this: Worker, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: Worker, ev: MessageEvent) => any) | null;\n    /**\n     * Clones message and transmits it to worker\'s global environment. transfer can be passed as a list of objects that are to be transferred rather than cloned.\n     */\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    /**\n     * Aborts worker\'s associated global environment.\n     */\n    terminate(): void;\n    addEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Worker: {\n    prototype: Worker;\n    new(stringUrl: string | URL, options?: WorkerOptions): Worker;\n};\n\ninterface Worklet {\n    addModule(moduleURL: string, options?: WorkletOptions): Promise<void>;\n}\n\ndeclare var Worklet: {\n    prototype: Worklet;\n    new(): Worklet;\n};\n\n/** This Streams API interface provides a standard abstraction for writing streaming data to a destination, known as a sink. This object comes with built-in backpressure and queuing. */\ninterface WritableStream<W = any> {\n    readonly locked: boolean;\n    abort(reason?: any): Promise<void>;\n    getWriter(): WritableStreamDefaultWriter<W>;\n}\n\ndeclare var WritableStream: {\n    prototype: WritableStream;\n    new<W = any>(underlyingSink?: UnderlyingSink<W>, strategy?: QueuingStrategy<W>): WritableStream<W>;\n};\n\n/** This Streams API interface represents a controller allowing control of a WritableStream\'s state. When constructing a WritableStream, the underlying sink is given a corresponding WritableStreamDefaultController instance to manipulate. */\ninterface WritableStreamDefaultController {\n    error(error?: any): void;\n}\n\ndeclare var WritableStreamDefaultController: {\n    prototype: WritableStreamDefaultController;\n    new(): WritableStreamDefaultController;\n};\n\n/** This Streams API interface is the object returned by WritableStream.getWriter() and once created locks the < writer to the WritableStream ensuring that no other streams can write to the underlying sink. */\ninterface WritableStreamDefaultWriter<W = any> {\n    readonly closed: Promise<void>;\n    readonly desiredSize: number | null;\n    readonly ready: Promise<void>;\n    abort(reason?: any): Promise<void>;\n    close(): Promise<void>;\n    releaseLock(): void;\n    write(chunk: W): Promise<void>;\n}\n\ndeclare var WritableStreamDefaultWriter: {\n    prototype: WritableStreamDefaultWriter;\n    new(): WritableStreamDefaultWriter;\n};\n\n/** An XML document. It inherits from the generic Document and does not add any specific methods or properties to it: nevertheless, several algorithms behave differently with the two types of documents. */\ninterface XMLDocument extends Document {\n    addEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: XMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DocumentEventMap>(type: K, listener: (this: XMLDocument, ev: DocumentEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLDocument: {\n    prototype: XMLDocument;\n    new(): XMLDocument;\n};\n\ninterface XMLHttpRequestEventMap extends XMLHttpRequestEventTargetEventMap {\n    "readystatechange": Event;\n}\n\n/** Use XMLHttpRequest (XHR) objects to interact with servers. You can retrieve data from a URL without having to do a full page refresh. This enables a Web page to update just part of a page without disrupting what the user is doing. */\ninterface XMLHttpRequest extends XMLHttpRequestEventTarget {\n    onreadystatechange: ((this: XMLHttpRequest, ev: Event) => any) | null;\n    /**\n     * Returns client\'s state.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the response\'s body.\n     */\n    readonly response: any;\n    /**\n     * Returns the text response.\n     * \n     * Throws an "InvalidStateError" DOMException if responseType is not the empty string or "text".\n     */\n    readonly responseText: string;\n    /**\n     * Returns the response type.\n     * \n     * Can be set to change the response type. Values are: the empty string (default), "arraybuffer", "blob", "document", "json", and "text".\n     * \n     * When set: setting to "document" is ignored if current global object is not a Window object.\n     * \n     * When set: throws an "InvalidStateError" DOMException if state is loading or done.\n     * \n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    responseType: XMLHttpRequestResponseType;\n    readonly responseURL: string;\n    /**\n     * Returns the document response.\n     * \n     * Throws an "InvalidStateError" DOMException if responseType is not the empty string or "document".\n     */\n    readonly responseXML: Document | null;\n    readonly status: number;\n    readonly statusText: string;\n    /**\n     * Can be set to a time in milliseconds. When set to a non-zero value will cause fetching to terminate after the given time has passed. When the time has passed, the request has not yet completed, and the synchronous flag is unset, a timeout event will then be dispatched, or a "TimeoutError" DOMException will be thrown otherwise (for the send() method).\n     * \n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    timeout: number;\n    /**\n     * Returns the associated XMLHttpRequestUpload object. It can be used to gather transmission information when data is transferred to a server.\n     */\n    readonly upload: XMLHttpRequestUpload;\n    /**\n     * True when credentials are to be included in a cross-origin request. False when they are to be excluded in a cross-origin request and when cookies are to be ignored in its response. Initially false.\n     * \n     * When set: throws an "InvalidStateError" DOMException if state is not unsent or opened, or if the send() flag is set.\n     */\n    withCredentials: boolean;\n    /**\n     * Cancels any network activity.\n     */\n    abort(): void;\n    getAllResponseHeaders(): string;\n    getResponseHeader(name: string): string | null;\n    /**\n     * Sets the request method, request URL, and synchronous flag.\n     * \n     * Throws a "SyntaxError" DOMException if either method is not a valid HTTP method or url cannot be parsed.\n     * \n     * Throws a "SecurityError" DOMException if method is a case-insensitive match for `CONNECT`, `TRACE`, or `TRACK`.\n     * \n     * Throws an "InvalidAccessError" DOMException if async is false, current global object is a Window object, and the timeout attribute is not zero or the responseType attribute is not the empty string.\n     */\n    open(method: string, url: string): void;\n    open(method: string, url: string, async: boolean, username?: string | null, password?: string | null): void;\n    /**\n     * Acts as if the `Content-Type` header value for response is mime. (It does not actually change the header though.)\n     * \n     * Throws an "InvalidStateError" DOMException if state is loading or done.\n     */\n    overrideMimeType(mime: string): void;\n    /**\n     * Initiates the request. The body argument provides the request body, if any, and is ignored if the request method is GET or HEAD.\n     * \n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     */\n    send(body?: Document | BodyInit | null): void;\n    /**\n     * Combines a header in author request headers.\n     * \n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     * \n     * Throws a "SyntaxError" DOMException if name is not a header name or if value is not a header value.\n     */\n    setRequestHeader(name: string, value: string): void;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n    addEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequest: {\n    prototype: XMLHttpRequest;\n    new(): XMLHttpRequest;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n};\n\ninterface XMLHttpRequestEventTargetEventMap {\n    "abort": ProgressEvent<XMLHttpRequestEventTarget>;\n    "error": ProgressEvent<XMLHttpRequestEventTarget>;\n    "load": ProgressEvent<XMLHttpRequestEventTarget>;\n    "loadend": ProgressEvent<XMLHttpRequestEventTarget>;\n    "loadstart": ProgressEvent<XMLHttpRequestEventTarget>;\n    "progress": ProgressEvent<XMLHttpRequestEventTarget>;\n    "timeout": ProgressEvent<XMLHttpRequestEventTarget>;\n}\n\ninterface XMLHttpRequestEventTarget extends EventTarget {\n    onabort: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onerror: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onload: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadend: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadstart: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onprogress: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    ontimeout: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestEventTarget: {\n    prototype: XMLHttpRequestEventTarget;\n    new(): XMLHttpRequestEventTarget;\n};\n\ninterface XMLHttpRequestUpload extends XMLHttpRequestEventTarget {\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestUpload: {\n    prototype: XMLHttpRequestUpload;\n    new(): XMLHttpRequestUpload;\n};\n\n/** Provides the serializeToString() method to construct an XML string representing a DOM tree. */\ninterface XMLSerializer {\n    serializeToString(root: Node): string;\n}\n\ndeclare var XMLSerializer: {\n    prototype: XMLSerializer;\n    new(): XMLSerializer;\n};\n\n/** The XPathEvaluator interface allows to compile and evaluate XPath expressions. */\ninterface XPathEvaluator extends XPathEvaluatorBase {\n}\n\ndeclare var XPathEvaluator: {\n    prototype: XPathEvaluator;\n    new(): XPathEvaluator;\n};\n\ninterface XPathEvaluatorBase {\n    createExpression(expression: string, resolver?: XPathNSResolver | null): XPathExpression;\n    createNSResolver(nodeResolver: Node): XPathNSResolver;\n    evaluate(expression: string, contextNode: Node, resolver?: XPathNSResolver | null, type?: number, result?: XPathResult | null): XPathResult;\n}\n\n/** This interface is a compiled XPath expression that can be evaluated on a document or specific node to return information its DOM tree. */\ninterface XPathExpression {\n    evaluate(contextNode: Node, type?: number, result?: XPathResult | null): XPathResult;\n}\n\ndeclare var XPathExpression: {\n    prototype: XPathExpression;\n    new(): XPathExpression;\n};\n\n/** The results generated by evaluating an XPath expression within the context of a given node. */\ninterface XPathResult {\n    readonly booleanValue: boolean;\n    readonly invalidIteratorState: boolean;\n    readonly numberValue: number;\n    readonly resultType: number;\n    readonly singleNodeValue: Node | null;\n    readonly snapshotLength: number;\n    readonly stringValue: string;\n    iterateNext(): Node | null;\n    snapshotItem(index: number): Node | null;\n    readonly ANY_TYPE: number;\n    readonly ANY_UNORDERED_NODE_TYPE: number;\n    readonly BOOLEAN_TYPE: number;\n    readonly FIRST_ORDERED_NODE_TYPE: number;\n    readonly NUMBER_TYPE: number;\n    readonly ORDERED_NODE_ITERATOR_TYPE: number;\n    readonly ORDERED_NODE_SNAPSHOT_TYPE: number;\n    readonly STRING_TYPE: number;\n    readonly UNORDERED_NODE_ITERATOR_TYPE: number;\n    readonly UNORDERED_NODE_SNAPSHOT_TYPE: number;\n}\n\ndeclare var XPathResult: {\n    prototype: XPathResult;\n    new(): XPathResult;\n    readonly ANY_TYPE: number;\n    readonly ANY_UNORDERED_NODE_TYPE: number;\n    readonly BOOLEAN_TYPE: number;\n    readonly FIRST_ORDERED_NODE_TYPE: number;\n    readonly NUMBER_TYPE: number;\n    readonly ORDERED_NODE_ITERATOR_TYPE: number;\n    readonly ORDERED_NODE_SNAPSHOT_TYPE: number;\n    readonly STRING_TYPE: number;\n    readonly UNORDERED_NODE_ITERATOR_TYPE: number;\n    readonly UNORDERED_NODE_SNAPSHOT_TYPE: number;\n};\n\n/** An XSLTProcessor applies an XSLT stylesheet transformation to an XML document to produce a new XML document as output. It has methods to load the XSLT stylesheet, to manipulate <xsl:param> parameter values, and to apply the transformation to documents. */\ninterface XSLTProcessor {\n    clearParameters(): void;\n    getParameter(namespaceURI: string, localName: string): any;\n    importStylesheet(style: Node): void;\n    removeParameter(namespaceURI: string, localName: string): void;\n    reset(): void;\n    setParameter(namespaceURI: string, localName: string, value: any): void;\n    transformToDocument(source: Node): Document;\n    transformToFragment(source: Node, document: Document): DocumentFragment;\n}\n\ndeclare var XSLTProcessor: {\n    prototype: XSLTProcessor;\n    new(): XSLTProcessor;\n};\n\ninterface webkitRTCPeerConnection extends RTCPeerConnection {\n    addEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: webkitRTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof RTCPeerConnectionEventMap>(type: K, listener: (this: webkitRTCPeerConnection, ev: RTCPeerConnectionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var webkitRTCPeerConnection: {\n    prototype: webkitRTCPeerConnection;\n    new(configuration: RTCConfiguration): webkitRTCPeerConnection;\n};\n\ndeclare type EventListenerOrEventListenerObject = EventListener | EventListenerObject;\n\ninterface Console {\n    memory: any;\n    assert(condition?: boolean, ...data: any[]): void;\n    clear(): void;\n    count(label?: string): void;\n    countReset(label?: string): void;\n    debug(...data: any[]): void;\n    dir(item?: any, options?: any): void;\n    dirxml(...data: any[]): void;\n    error(...data: any[]): void;\n    exception(message?: string, ...optionalParams: any[]): void;\n    group(...data: any[]): void;\n    groupCollapsed(...data: any[]): void;\n    groupEnd(): void;\n    info(...data: any[]): void;\n    log(...data: any[]): void;\n    table(tabularData?: any, properties?: string[]): void;\n    time(label?: string): void;\n    timeEnd(label?: string): void;\n    timeLog(label?: string, ...data: any[]): void;\n    timeStamp(label?: string): void;\n    trace(...data: any[]): void;\n    warn(...data: any[]): void;\n}\n\ndeclare var console: Console;\n\n/** Holds useful CSS-related methods. No object with this interface are implemented: it contains only static methods and therefore is a utilitarian interface. */\ndeclare namespace CSS {\n    function escape(ident: string): string;\n    function supports(property: string, value: string): boolean;\n    function supports(conditionText: string): boolean;\n}\n\ndeclare namespace WebAssembly {\n    interface CompileError {\n    }\n    \n    var CompileError: {\n        prototype: CompileError;\n        new(): CompileError;\n    };\n    \n    interface Global {\n        value: any;\n        valueOf(): any;\n    }\n    \n    var Global: {\n        prototype: Global;\n        new(descriptor: GlobalDescriptor, v?: any): Global;\n    };\n    \n    interface Instance {\n        readonly exports: Exports;\n    }\n    \n    var Instance: {\n        prototype: Instance;\n        new(module: Module, importObject?: Imports): Instance;\n    };\n    \n    interface LinkError {\n    }\n    \n    var LinkError: {\n        prototype: LinkError;\n        new(): LinkError;\n    };\n    \n    interface Memory {\n        readonly buffer: ArrayBuffer;\n        grow(delta: number): number;\n    }\n    \n    var Memory: {\n        prototype: Memory;\n        new(descriptor: MemoryDescriptor): Memory;\n    };\n    \n    interface Module {\n    }\n    \n    var Module: {\n        prototype: Module;\n        new(bytes: BufferSource): Module;\n        customSections(moduleObject: Module, sectionName: string): ArrayBuffer[];\n        exports(moduleObject: Module): ModuleExportDescriptor[];\n        imports(moduleObject: Module): ModuleImportDescriptor[];\n    };\n    \n    interface RuntimeError {\n    }\n    \n    var RuntimeError: {\n        prototype: RuntimeError;\n        new(): RuntimeError;\n    };\n    \n    interface Table {\n        readonly length: number;\n        get(index: number): Function | null;\n        grow(delta: number): number;\n        set(index: number, value: Function | null): void;\n    }\n    \n    var Table: {\n        prototype: Table;\n        new(descriptor: TableDescriptor): Table;\n    };\n    \n    interface GlobalDescriptor {\n        mutable?: boolean;\n        value: ValueType;\n    }\n    \n    interface MemoryDescriptor {\n        initial: number;\n        maximum?: number;\n    }\n    \n    interface ModuleExportDescriptor {\n        kind: ImportExportKind;\n        name: string;\n    }\n    \n    interface ModuleImportDescriptor {\n        kind: ImportExportKind;\n        module: string;\n        name: string;\n    }\n    \n    interface TableDescriptor {\n        element: TableKind;\n        initial: number;\n        maximum?: number;\n    }\n    \n    interface WebAssemblyInstantiatedSource {\n        instance: Instance;\n        module: Module;\n    }\n    \n    type ImportExportKind = "function" | "global" | "memory" | "table";\n    type TableKind = "anyfunc";\n    type ValueType = "f32" | "f64" | "i32" | "i64";\n    type ExportValue = Function | Global | Memory | Table;\n    type Exports = Record<string, ExportValue>;\n    type ImportValue = ExportValue | number;\n    type ModuleImports = Record<string, ImportValue>;\n    type Imports = Record<string, ModuleImports>;\n    function compile(bytes: BufferSource): Promise<Module>;\n    function compileStreaming(source: Response | Promise<Response>): Promise<Module>;\n    function instantiate(bytes: BufferSource, importObject?: Imports): Promise<WebAssemblyInstantiatedSource>;\n    function instantiate(moduleObject: Module, importObject?: Imports): Promise<Instance>;\n    function instantiateStreaming(response: Response | PromiseLike<Response>, importObject?: Imports): Promise<WebAssemblyInstantiatedSource>;\n    function validate(bytes: BufferSource): boolean;\n}\n\ninterface BlobCallback {\n    (blob: Blob | null): void;\n}\n\ninterface CustomElementConstructor {\n    new (...params: any[]): HTMLElement;\n}\n\ninterface DecodeErrorCallback {\n    (error: DOMException): void;\n}\n\ninterface DecodeSuccessCallback {\n    (decodedData: AudioBuffer): void;\n}\n\ninterface EventHandlerNonNull {\n    (event: Event): any;\n}\n\ninterface ForEachCallback {\n    (keyId: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, status: MediaKeyStatus): void;\n}\n\ninterface FrameRequestCallback {\n    (time: number): void;\n}\n\ninterface FunctionStringCallback {\n    (data: string): void;\n}\n\ninterface IntersectionObserverCallback {\n    (entries: IntersectionObserverEntry[], observer: IntersectionObserver): void;\n}\n\ninterface MSLaunchUriCallback {\n    (): void;\n}\n\ninterface MutationCallback {\n    (mutations: MutationRecord[], observer: MutationObserver): void;\n}\n\ninterface NavigatorUserMediaErrorCallback {\n    (error: MediaStreamError): void;\n}\n\ninterface NavigatorUserMediaSuccessCallback {\n    (stream: MediaStream): void;\n}\n\ninterface NotificationPermissionCallback {\n    (permission: NotificationPermission): void;\n}\n\ninterface OnBeforeUnloadEventHandlerNonNull {\n    (event: Event): string | null;\n}\n\ninterface OnErrorEventHandlerNonNull {\n    (event: Event | string, source?: string, lineno?: number, colno?: number, error?: Error): any;\n}\n\ninterface PerformanceObserverCallback {\n    (entries: PerformanceObserverEntryList, observer: PerformanceObserver): void;\n}\n\ninterface PositionCallback {\n    (position: GeolocationPosition): void;\n}\n\ninterface PositionErrorCallback {\n    (positionError: GeolocationPositionError): void;\n}\n\ninterface QueuingStrategySizeCallback<T = any> {\n    (chunk: T): number;\n}\n\ninterface RTCPeerConnectionErrorCallback {\n    (error: DOMException): void;\n}\n\ninterface RTCSessionDescriptionCallback {\n    (description: RTCSessionDescriptionInit): void;\n}\n\ninterface RTCStatsCallback {\n    (report: RTCStatsReport): void;\n}\n\ninterface ReadableByteStreamControllerCallback {\n    (controller: ReadableByteStreamController): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamDefaultControllerCallback<R> {\n    (controller: ReadableStreamDefaultController<R>): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerCallback<O> {\n    (controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerTransformCallback<I, O> {\n    (chunk: I, controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface VoidFunction {\n    (): void;\n}\n\ninterface WritableStreamDefaultControllerCloseCallback {\n    (): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerStartCallback {\n    (controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerWriteCallback<W> {\n    (chunk: W, controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\ninterface HTMLElementTagNameMap {\n    "a": HTMLAnchorElement;\n    "abbr": HTMLElement;\n    "address": HTMLElement;\n    "applet": HTMLAppletElement;\n    "area": HTMLAreaElement;\n    "article": HTMLElement;\n    "aside": HTMLElement;\n    "audio": HTMLAudioElement;\n    "b": HTMLElement;\n    "base": HTMLBaseElement;\n    "basefont": HTMLBaseFontElement;\n    "bdi": HTMLElement;\n    "bdo": HTMLElement;\n    "blockquote": HTMLQuoteElement;\n    "body": HTMLBodyElement;\n    "br": HTMLBRElement;\n    "button": HTMLButtonElement;\n    "canvas": HTMLCanvasElement;\n    "caption": HTMLTableCaptionElement;\n    "cite": HTMLElement;\n    "code": HTMLElement;\n    "col": HTMLTableColElement;\n    "colgroup": HTMLTableColElement;\n    "data": HTMLDataElement;\n    "datalist": HTMLDataListElement;\n    "dd": HTMLElement;\n    "del": HTMLModElement;\n    "details": HTMLDetailsElement;\n    "dfn": HTMLElement;\n    "dialog": HTMLDialogElement;\n    "dir": HTMLDirectoryElement;\n    "div": HTMLDivElement;\n    "dl": HTMLDListElement;\n    "dt": HTMLElement;\n    "em": HTMLElement;\n    "embed": HTMLEmbedElement;\n    "fieldset": HTMLFieldSetElement;\n    "figcaption": HTMLElement;\n    "figure": HTMLElement;\n    "font": HTMLFontElement;\n    "footer": HTMLElement;\n    "form": HTMLFormElement;\n    "frame": HTMLFrameElement;\n    "frameset": HTMLFrameSetElement;\n    "h1": HTMLHeadingElement;\n    "h2": HTMLHeadingElement;\n    "h3": HTMLHeadingElement;\n    "h4": HTMLHeadingElement;\n    "h5": HTMLHeadingElement;\n    "h6": HTMLHeadingElement;\n    "head": HTMLHeadElement;\n    "header": HTMLElement;\n    "hgroup": HTMLElement;\n    "hr": HTMLHRElement;\n    "html": HTMLHtmlElement;\n    "i": HTMLElement;\n    "iframe": HTMLIFrameElement;\n    "img": HTMLImageElement;\n    "input": HTMLInputElement;\n    "ins": HTMLModElement;\n    "kbd": HTMLElement;\n    "label": HTMLLabelElement;\n    "legend": HTMLLegendElement;\n    "li": HTMLLIElement;\n    "link": HTMLLinkElement;\n    "main": HTMLElement;\n    "map": HTMLMapElement;\n    "mark": HTMLElement;\n    "marquee": HTMLMarqueeElement;\n    "menu": HTMLMenuElement;\n    "meta": HTMLMetaElement;\n    "meter": HTMLMeterElement;\n    "nav": HTMLElement;\n    "noscript": HTMLElement;\n    "object": HTMLObjectElement;\n    "ol": HTMLOListElement;\n    "optgroup": HTMLOptGroupElement;\n    "option": HTMLOptionElement;\n    "output": HTMLOutputElement;\n    "p": HTMLParagraphElement;\n    "param": HTMLParamElement;\n    "picture": HTMLPictureElement;\n    "pre": HTMLPreElement;\n    "progress": HTMLProgressElement;\n    "q": HTMLQuoteElement;\n    "rp": HTMLElement;\n    "rt": HTMLElement;\n    "ruby": HTMLElement;\n    "s": HTMLElement;\n    "samp": HTMLElement;\n    "script": HTMLScriptElement;\n    "section": HTMLElement;\n    "select": HTMLSelectElement;\n    "slot": HTMLSlotElement;\n    "small": HTMLElement;\n    "source": HTMLSourceElement;\n    "span": HTMLSpanElement;\n    "strong": HTMLElement;\n    "style": HTMLStyleElement;\n    "sub": HTMLElement;\n    "summary": HTMLElement;\n    "sup": HTMLElement;\n    "table": HTMLTableElement;\n    "tbody": HTMLTableSectionElement;\n    "td": HTMLTableDataCellElement;\n    "template": HTMLTemplateElement;\n    "textarea": HTMLTextAreaElement;\n    "tfoot": HTMLTableSectionElement;\n    "th": HTMLTableHeaderCellElement;\n    "thead": HTMLTableSectionElement;\n    "time": HTMLTimeElement;\n    "title": HTMLTitleElement;\n    "tr": HTMLTableRowElement;\n    "track": HTMLTrackElement;\n    "u": HTMLElement;\n    "ul": HTMLUListElement;\n    "var": HTMLElement;\n    "video": HTMLVideoElement;\n    "wbr": HTMLElement;\n}\n\ninterface HTMLElementDeprecatedTagNameMap {\n    "listing": HTMLPreElement;\n    "xmp": HTMLPreElement;\n}\n\ninterface SVGElementTagNameMap {\n    "a": SVGAElement;\n    "circle": SVGCircleElement;\n    "clipPath": SVGClipPathElement;\n    "defs": SVGDefsElement;\n    "desc": SVGDescElement;\n    "ellipse": SVGEllipseElement;\n    "feBlend": SVGFEBlendElement;\n    "feColorMatrix": SVGFEColorMatrixElement;\n    "feComponentTransfer": SVGFEComponentTransferElement;\n    "feComposite": SVGFECompositeElement;\n    "feConvolveMatrix": SVGFEConvolveMatrixElement;\n    "feDiffuseLighting": SVGFEDiffuseLightingElement;\n    "feDisplacementMap": SVGFEDisplacementMapElement;\n    "feDistantLight": SVGFEDistantLightElement;\n    "feFlood": SVGFEFloodElement;\n    "feFuncA": SVGFEFuncAElement;\n    "feFuncB": SVGFEFuncBElement;\n    "feFuncG": SVGFEFuncGElement;\n    "feFuncR": SVGFEFuncRElement;\n    "feGaussianBlur": SVGFEGaussianBlurElement;\n    "feImage": SVGFEImageElement;\n    "feMerge": SVGFEMergeElement;\n    "feMergeNode": SVGFEMergeNodeElement;\n    "feMorphology": SVGFEMorphologyElement;\n    "feOffset": SVGFEOffsetElement;\n    "fePointLight": SVGFEPointLightElement;\n    "feSpecularLighting": SVGFESpecularLightingElement;\n    "feSpotLight": SVGFESpotLightElement;\n    "feTile": SVGFETileElement;\n    "feTurbulence": SVGFETurbulenceElement;\n    "filter": SVGFilterElement;\n    "foreignObject": SVGForeignObjectElement;\n    "g": SVGGElement;\n    "image": SVGImageElement;\n    "line": SVGLineElement;\n    "linearGradient": SVGLinearGradientElement;\n    "marker": SVGMarkerElement;\n    "mask": SVGMaskElement;\n    "metadata": SVGMetadataElement;\n    "path": SVGPathElement;\n    "pattern": SVGPatternElement;\n    "polygon": SVGPolygonElement;\n    "polyline": SVGPolylineElement;\n    "radialGradient": SVGRadialGradientElement;\n    "rect": SVGRectElement;\n    "script": SVGScriptElement;\n    "stop": SVGStopElement;\n    "style": SVGStyleElement;\n    "svg": SVGSVGElement;\n    "switch": SVGSwitchElement;\n    "symbol": SVGSymbolElement;\n    "text": SVGTextElement;\n    "textPath": SVGTextPathElement;\n    "title": SVGTitleElement;\n    "tspan": SVGTSpanElement;\n    "use": SVGUseElement;\n    "view": SVGViewElement;\n}\n\n/** @deprecated Directly use HTMLElementTagNameMap or SVGElementTagNameMap as appropriate, instead. */\ntype ElementTagNameMap = HTMLElementTagNameMap & Pick<SVGElementTagNameMap, Exclude<keyof SVGElementTagNameMap, keyof HTMLElementTagNameMap>>;\n\ndeclare var Audio: {\n    new(src?: string): HTMLAudioElement;\n};\ndeclare var Image: {\n    new(width?: number, height?: number): HTMLImageElement;\n};\ndeclare var Option: {\n    new(text?: string, value?: string, defaultSelected?: boolean, selected?: boolean): HTMLOptionElement;\n};\ndeclare var applicationCache: ApplicationCache;\ndeclare var clientInformation: Navigator;\ndeclare var closed: boolean;\ndeclare var customElements: CustomElementRegistry;\ndeclare var defaultStatus: string;\ndeclare var devicePixelRatio: number;\ndeclare var doNotTrack: string;\ndeclare var document: Document;\n/** @deprecated */\ndeclare var event: Event | undefined;\n/** @deprecated */\ndeclare var external: External;\ndeclare var frameElement: Element;\ndeclare var frames: Window;\ndeclare var history: History;\ndeclare var innerHeight: number;\ndeclare var innerWidth: number;\ndeclare var length: number;\ndeclare var location: Location;\ndeclare var locationbar: BarProp;\ndeclare var menubar: BarProp;\ndeclare var msContentScript: ExtensionScriptApis;\n/** @deprecated */\ndeclare const name: void;\ndeclare var navigator: Navigator;\ndeclare var offscreenBuffering: string | boolean;\ndeclare var oncompassneedscalibration: ((this: Window, ev: Event) => any) | null;\ndeclare var ondevicelight: ((this: Window, ev: DeviceLightEvent) => any) | null;\ndeclare var ondevicemotion: ((this: Window, ev: DeviceMotionEvent) => any) | null;\ndeclare var ondeviceorientation: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\ndeclare var ondeviceorientationabsolute: ((this: Window, ev: DeviceOrientationEvent) => any) | null;\ndeclare var onmousewheel: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturechange: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturedoubletap: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgestureend: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturehold: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturestart: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsgesturetap: ((this: Window, ev: Event) => any) | null;\ndeclare var onmsinertiastart: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointercancel: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerdown: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerenter: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerleave: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointermove: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerout: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerover: ((this: Window, ev: Event) => any) | null;\ndeclare var onmspointerup: ((this: Window, ev: Event) => any) | null;\n/** @deprecated */\ndeclare var onorientationchange: ((this: Window, ev: Event) => any) | null;\ndeclare var onreadystatechange: ((this: Window, ev: ProgressEvent<Window>) => any) | null;\ndeclare var onvrdisplayactivate: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayblur: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayconnect: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaydeactivate: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaydisconnect: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplayfocus: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypointerrestricted: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypointerunrestricted: ((this: Window, ev: Event) => any) | null;\ndeclare var onvrdisplaypresentchange: ((this: Window, ev: Event) => any) | null;\ndeclare var opener: any;\n/** @deprecated */\ndeclare var orientation: string | number;\ndeclare var outerHeight: number;\ndeclare var outerWidth: number;\ndeclare var pageXOffset: number;\ndeclare var pageYOffset: number;\ndeclare var parent: Window;\ndeclare var personalbar: BarProp;\ndeclare var screen: Screen;\ndeclare var screenLeft: number;\ndeclare var screenTop: number;\ndeclare var screenX: number;\ndeclare var screenY: number;\ndeclare var scrollX: number;\ndeclare var scrollY: number;\ndeclare var scrollbars: BarProp;\ndeclare var self: Window & typeof globalThis;\ndeclare var speechSynthesis: SpeechSynthesis;\ndeclare var status: string;\ndeclare var statusbar: BarProp;\ndeclare var styleMedia: StyleMedia;\ndeclare var toolbar: BarProp;\ndeclare var top: Window;\ndeclare var visualViewport: VisualViewport;\ndeclare var window: Window & typeof globalThis;\ndeclare function alert(message?: any): void;\ndeclare function blur(): void;\n/** @deprecated */\ndeclare function captureEvents(): void;\ndeclare function close(): void;\ndeclare function confirm(message?: string): boolean;\ndeclare function departFocus(navigationReason: NavigationReason, origin: FocusNavigationOrigin): void;\ndeclare function focus(): void;\ndeclare function getComputedStyle(elt: Element, pseudoElt?: string | null): CSSStyleDeclaration;\ndeclare function getMatchedCSSRules(elt: Element, pseudoElt?: string | null): CSSRuleList;\ndeclare function getSelection(): Selection | null;\ndeclare function matchMedia(query: string): MediaQueryList;\ndeclare function moveBy(x: number, y: number): void;\ndeclare function moveTo(x: number, y: number): void;\ndeclare function msWriteProfilerMark(profilerMarkName: string): void;\ndeclare function open(url?: string, target?: string, features?: string, replace?: boolean): Window | null;\ndeclare function postMessage(message: any, targetOrigin: string, transfer?: Transferable[]): void;\ndeclare function print(): void;\ndeclare function prompt(message?: string, _default?: string): string | null;\n/** @deprecated */\ndeclare function releaseEvents(): void;\ndeclare function resizeBy(x: number, y: number): void;\ndeclare function resizeTo(width: number, height: number): void;\ndeclare function scroll(options?: ScrollToOptions): void;\ndeclare function scroll(x: number, y: number): void;\ndeclare function scrollBy(options?: ScrollToOptions): void;\ndeclare function scrollBy(x: number, y: number): void;\ndeclare function scrollTo(options?: ScrollToOptions): void;\ndeclare function scrollTo(x: number, y: number): void;\ndeclare function stop(): void;\ndeclare function webkitCancelAnimationFrame(handle: number): void;\ndeclare function webkitConvertPointFromNodeToPage(node: Node, pt: WebKitPoint): WebKitPoint;\ndeclare function webkitConvertPointFromPageToNode(node: Node, pt: WebKitPoint): WebKitPoint;\ndeclare function webkitRequestAnimationFrame(callback: FrameRequestCallback): number;\ndeclare function toString(): string;\n/**\n * Dispatches a synthetic event event to target and returns true if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n */\ndeclare function dispatchEvent(event: Event): boolean;\ndeclare var sessionStorage: Storage;\ndeclare var localStorage: Storage;\n/**\n * Fires when the user aborts the download.\n * @param ev The event.\n */\ndeclare var onabort: ((this: Window, ev: UIEvent) => any) | null;\ndeclare var onanimationcancel: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationend: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationiteration: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onanimationstart: ((this: Window, ev: AnimationEvent) => any) | null;\ndeclare var onauxclick: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the object loses the input focus.\n * @param ev The focus event.\n */\ndeclare var onblur: ((this: Window, ev: FocusEvent) => any) | null;\ndeclare var oncancel: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when playback is possible, but would require further buffering.\n * @param ev The event.\n */\ndeclare var oncanplay: ((this: Window, ev: Event) => any) | null;\ndeclare var oncanplaythrough: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the contents of the object or selection have changed.\n * @param ev The event.\n */\ndeclare var onchange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user clicks the left mouse button on the object\n * @param ev The mouse event.\n */\ndeclare var onclick: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onclose: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user clicks the right mouse button in the client area, opening the context menu.\n * @param ev The mouse event.\n */\ndeclare var oncontextmenu: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var oncuechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user double-clicks the object.\n * @param ev The mouse event.\n */\ndeclare var ondblclick: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires on the source object continuously during a drag operation.\n * @param ev The event.\n */\ndeclare var ondrag: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the source object when the user releases the mouse at the close of a drag operation.\n * @param ev The event.\n */\ndeclare var ondragend: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the target element when the user drags the object to a valid drop target.\n * @param ev The drag event.\n */\ndeclare var ondragenter: ((this: Window, ev: DragEvent) => any) | null;\ndeclare var ondragexit: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires on the target object when the user moves the mouse out of a valid drop target during a drag operation.\n * @param ev The drag event.\n */\ndeclare var ondragleave: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the target element continuously while the user drags the object over a valid drop target.\n * @param ev The event.\n */\ndeclare var ondragover: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Fires on the source object when the user starts to drag a text selection or selected object.\n * @param ev The event.\n */\ndeclare var ondragstart: ((this: Window, ev: DragEvent) => any) | null;\ndeclare var ondrop: ((this: Window, ev: DragEvent) => any) | null;\n/**\n * Occurs when the duration attribute is updated.\n * @param ev The event.\n */\ndeclare var ondurationchange: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the media element is reset to its initial state.\n * @param ev The event.\n */\ndeclare var onemptied: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the end of playback is reached.\n * @param ev The event\n */\ndeclare var onended: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when an error occurs during object loading.\n * @param ev The event.\n */\ndeclare var onerror: OnErrorEventHandler;\n/**\n * Fires when the object receives focus.\n * @param ev The event.\n */\ndeclare var onfocus: ((this: Window, ev: FocusEvent) => any) | null;\ndeclare var ongotpointercapture: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var oninput: ((this: Window, ev: Event) => any) | null;\ndeclare var oninvalid: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user presses a key.\n * @param ev The keyboard event\n */\ndeclare var onkeydown: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires when the user presses an alphanumeric key.\n * @param ev The event.\n */\ndeclare var onkeypress: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires when the user releases a key.\n * @param ev The keyboard event\n */\ndeclare var onkeyup: ((this: Window, ev: KeyboardEvent) => any) | null;\n/**\n * Fires immediately after the browser loads the object.\n * @param ev The event.\n */\ndeclare var onload: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when media data is loaded at the current playback position.\n * @param ev The event.\n */\ndeclare var onloadeddata: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the duration and dimensions of the media have been determined.\n * @param ev The event.\n */\ndeclare var onloadedmetadata: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when Internet Explorer begins looking for media data.\n * @param ev The event.\n */\ndeclare var onloadstart: ((this: Window, ev: Event) => any) | null;\ndeclare var onlostpointercapture: ((this: Window, ev: PointerEvent) => any) | null;\n/**\n * Fires when the user clicks the object with either mouse button.\n * @param ev The mouse event.\n */\ndeclare var onmousedown: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onmouseenter: ((this: Window, ev: MouseEvent) => any) | null;\ndeclare var onmouseleave: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse over the object.\n * @param ev The mouse event.\n */\ndeclare var onmousemove: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse pointer outside the boundaries of the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseout: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user moves the mouse pointer into the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseover: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Fires when the user releases a mouse button while the mouse is over the object.\n * @param ev The mouse event.\n */\ndeclare var onmouseup: ((this: Window, ev: MouseEvent) => any) | null;\n/**\n * Occurs when playback is paused.\n * @param ev The event.\n */\ndeclare var onpause: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the play method is requested.\n * @param ev The event.\n */\ndeclare var onplay: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the audio or video has started playing.\n * @param ev The event.\n */\ndeclare var onplaying: ((this: Window, ev: Event) => any) | null;\ndeclare var onpointercancel: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerdown: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerenter: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerleave: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointermove: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerout: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerover: ((this: Window, ev: PointerEvent) => any) | null;\ndeclare var onpointerup: ((this: Window, ev: PointerEvent) => any) | null;\n/**\n * Occurs to indicate progress while downloading media data.\n * @param ev The event.\n */\ndeclare var onprogress: ((this: Window, ev: ProgressEvent) => any) | null;\n/**\n * Occurs when the playback rate is increased or decreased.\n * @param ev The event.\n */\ndeclare var onratechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the user resets a form.\n * @param ev The event.\n */\ndeclare var onreset: ((this: Window, ev: Event) => any) | null;\ndeclare var onresize: ((this: Window, ev: UIEvent) => any) | null;\n/**\n * Fires when the user repositions the scroll box in the scroll bar on the object.\n * @param ev The event.\n */\ndeclare var onscroll: ((this: Window, ev: Event) => any) | null;\ndeclare var onsecuritypolicyviolation: ((this: Window, ev: SecurityPolicyViolationEvent) => any) | null;\n/**\n * Occurs when the seek operation ends.\n * @param ev The event.\n */\ndeclare var onseeked: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the current playback position is moved.\n * @param ev The event.\n */\ndeclare var onseeking: ((this: Window, ev: Event) => any) | null;\n/**\n * Fires when the current selection changes.\n * @param ev The event.\n */\ndeclare var onselect: ((this: Window, ev: Event) => any) | null;\ndeclare var onselectionchange: ((this: Window, ev: Event) => any) | null;\ndeclare var onselectstart: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when the download has stopped.\n * @param ev The event.\n */\ndeclare var onstalled: ((this: Window, ev: Event) => any) | null;\ndeclare var onsubmit: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs if the load operation has been intentionally halted.\n * @param ev The event.\n */\ndeclare var onsuspend: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs to indicate the current playback position.\n * @param ev The event.\n */\ndeclare var ontimeupdate: ((this: Window, ev: Event) => any) | null;\ndeclare var ontoggle: ((this: Window, ev: Event) => any) | null;\ndeclare var ontouchcancel: ((this: Window, ev: TouchEvent) => any) | null | undefined;\ndeclare var ontouchend: ((this: Window, ev: TouchEvent) => any) | null | undefined;\ndeclare var ontouchmove: ((this: Window, ev: TouchEvent) => any) | null | undefined;\ndeclare var ontouchstart: ((this: Window, ev: TouchEvent) => any) | null | undefined;\ndeclare var ontransitioncancel: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionend: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionrun: ((this: Window, ev: TransitionEvent) => any) | null;\ndeclare var ontransitionstart: ((this: Window, ev: TransitionEvent) => any) | null;\n/**\n * Occurs when the volume is changed, or playback is muted or unmuted.\n * @param ev The event.\n */\ndeclare var onvolumechange: ((this: Window, ev: Event) => any) | null;\n/**\n * Occurs when playback stops because the next frame of a video resource is not available.\n * @param ev The event.\n */\ndeclare var onwaiting: ((this: Window, ev: Event) => any) | null;\ndeclare var onwheel: ((this: Window, ev: WheelEvent) => any) | null;\ndeclare function cancelAnimationFrame(handle: number): void;\ndeclare function requestAnimationFrame(callback: FrameRequestCallback): number;\ndeclare var caches: CacheStorage;\ndeclare var crypto: Crypto;\ndeclare var indexedDB: IDBFactory;\ndeclare var isSecureContext: boolean;\ndeclare var origin: string;\ndeclare var performance: Performance;\ndeclare function atob(data: string): string;\ndeclare function btoa(data: string): string;\ndeclare function clearInterval(handle?: number): void;\ndeclare function clearTimeout(handle?: number): void;\ndeclare function createImageBitmap(image: ImageBitmapSource, options?: ImageBitmapOptions): Promise<ImageBitmap>;\ndeclare function createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number, options?: ImageBitmapOptions): Promise<ImageBitmap>;\ndeclare function fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\ndeclare function queueMicrotask(callback: VoidFunction): void;\ndeclare function setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare function setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare var onafterprint: ((this: Window, ev: Event) => any) | null;\ndeclare var onbeforeprint: ((this: Window, ev: Event) => any) | null;\ndeclare var onbeforeunload: ((this: Window, ev: BeforeUnloadEvent) => any) | null;\ndeclare var onhashchange: ((this: Window, ev: HashChangeEvent) => any) | null;\ndeclare var onlanguagechange: ((this: Window, ev: Event) => any) | null;\ndeclare var onmessage: ((this: Window, ev: MessageEvent) => any) | null;\ndeclare var onmessageerror: ((this: Window, ev: MessageEvent) => any) | null;\ndeclare var onoffline: ((this: Window, ev: Event) => any) | null;\ndeclare var ononline: ((this: Window, ev: Event) => any) | null;\ndeclare var onpagehide: ((this: Window, ev: PageTransitionEvent) => any) | null;\ndeclare var onpageshow: ((this: Window, ev: PageTransitionEvent) => any) | null;\ndeclare var onpopstate: ((this: Window, ev: PopStateEvent) => any) | null;\ndeclare var onrejectionhandled: ((this: Window, ev: PromiseRejectionEvent) => any) | null;\ndeclare var onstorage: ((this: Window, ev: StorageEvent) => any) | null;\ndeclare var onunhandledrejection: ((this: Window, ev: PromiseRejectionEvent) => any) | null;\ndeclare var onunload: ((this: Window, ev: Event) => any) | null;\ndeclare function addEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\ndeclare function addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\ndeclare function removeEventListener<K extends keyof WindowEventMap>(type: K, listener: (this: Window, ev: WindowEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\ndeclare function removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\ntype HeadersInit = Headers | string[][] | Record<string, string>;\ntype BodyInit = Blob | BufferSource | FormData | URLSearchParams | ReadableStream<Uint8Array> | string;\ntype RequestInfo = Request | string;\ntype BlobPart = BufferSource | Blob | string;\ntype DOMHighResTimeStamp = number;\ntype RenderingContext = CanvasRenderingContext2D | ImageBitmapRenderingContext | WebGLRenderingContext | WebGL2RenderingContext;\ntype HTMLOrSVGImageElement = HTMLImageElement | SVGImageElement;\ntype CanvasImageSource = HTMLOrSVGImageElement | HTMLVideoElement | HTMLCanvasElement | ImageBitmap | OffscreenCanvas;\ntype OffscreenRenderingContext = OffscreenCanvasRenderingContext2D | ImageBitmapRenderingContext | WebGLRenderingContext | WebGL2RenderingContext;\ntype MessageEventSource = WindowProxy | MessagePort | ServiceWorker;\ntype HTMLOrSVGScriptElement = HTMLScriptElement | SVGScriptElement;\ntype ImageBitmapSource = CanvasImageSource | Blob | ImageData;\ntype MediaProvider = MediaStream | MediaSource | Blob;\ntype OnErrorEventHandler = OnErrorEventHandlerNonNull | null;\ntype OnBeforeUnloadEventHandler = OnBeforeUnloadEventHandlerNonNull | null;\ntype TimerHandler = string | Function;\ntype ConstrainULong = number | ConstrainULongRange;\ntype ConstrainDouble = number | ConstrainDoubleRange;\ntype ConstrainBoolean = boolean | ConstrainBooleanParameters;\ntype ConstrainDOMString = string | string[] | ConstrainDOMStringParameters;\ntype PerformanceEntryList = PerformanceEntry[];\ntype ReadableStreamReadResult<T> = ReadableStreamReadValueResult<T> | ReadableStreamReadDoneResult<T>;\ntype VibratePattern = number | number[];\ntype COSEAlgorithmIdentifier = number;\ntype UvmEntry = number[];\ntype UvmEntries = UvmEntry[];\ntype AlgorithmIdentifier = string | Algorithm;\ntype HashAlgorithmIdentifier = AlgorithmIdentifier;\ntype BigInteger = Uint8Array;\ntype NamedCurve = string;\ntype GLenum = number;\ntype GLboolean = boolean;\ntype GLbitfield = number;\ntype GLint = number;\ntype GLsizei = number;\ntype GLintptr = number;\ntype GLsizeiptr = number;\ntype GLuint = number;\ntype GLfloat = number;\ntype GLclampf = number;\ntype TexImageSource = ImageBitmap | ImageData | HTMLImageElement | HTMLCanvasElement | HTMLVideoElement | OffscreenCanvas;\ntype Float32List = Float32Array | GLfloat[];\ntype Int32List = Int32Array | GLint[];\ntype GLint64 = number;\ntype GLuint64 = number;\ntype Uint32List = Uint32Array | GLuint[];\ntype BufferSource = ArrayBufferView | ArrayBuffer;\ntype DOMTimeStamp = number;\ntype LineAndPositionSetting = number | AutoKeyword;\ntype FormDataEntryValue = File | string;\ntype InsertPosition = "beforebegin" | "afterbegin" | "beforeend" | "afterend";\ntype IDBValidKey = number | string | Date | BufferSource | IDBArrayKey;\ntype MutationRecordType = "attributes" | "characterData" | "childList";\ntype IDBKeyPath = string;\ntype Transferable = ArrayBuffer | MessagePort | ImageBitmap | OffscreenCanvas;\ntype RTCIceGatherCandidate = RTCIceCandidateDictionary | RTCIceCandidateComplete;\ntype RTCTransport = RTCDtlsTransport | RTCSrtpSdesTransport;\n/** @deprecated */\ntype MouseWheelEvent = WheelEvent;\ntype WindowProxy = Window;\ntype AlignSetting = "center" | "end" | "left" | "right" | "start";\ntype AnimationPlayState = "finished" | "idle" | "paused" | "running";\ntype AppendMode = "segments" | "sequence";\ntype AttestationConveyancePreference = "direct" | "enterprise" | "indirect" | "none";\ntype AudioContextLatencyCategory = "balanced" | "interactive" | "playback";\ntype AudioContextState = "closed" | "running" | "suspended";\ntype AuthenticatorAttachment = "cross-platform" | "platform";\ntype AuthenticatorTransport = "ble" | "internal" | "nfc" | "usb";\ntype AutoKeyword = "auto";\ntype AutomationRate = "a-rate" | "k-rate";\ntype BinaryType = "arraybuffer" | "blob";\ntype BiquadFilterType = "allpass" | "bandpass" | "highpass" | "highshelf" | "lowpass" | "lowshelf" | "notch" | "peaking";\ntype CanPlayTypeResult = "" | "maybe" | "probably";\ntype CanvasDirection = "inherit" | "ltr" | "rtl";\ntype CanvasFillRule = "evenodd" | "nonzero";\ntype CanvasLineCap = "butt" | "round" | "square";\ntype CanvasLineJoin = "bevel" | "miter" | "round";\ntype CanvasTextAlign = "center" | "end" | "left" | "right" | "start";\ntype CanvasTextBaseline = "alphabetic" | "bottom" | "hanging" | "ideographic" | "middle" | "top";\ntype ChannelCountMode = "clamped-max" | "explicit" | "max";\ntype ChannelInterpretation = "discrete" | "speakers";\ntype ClientTypes = "all" | "sharedworker" | "window" | "worker";\ntype ColorSpaceConversion = "default" | "none";\ntype CompositeOperation = "accumulate" | "add" | "replace";\ntype CompositeOperationOrAuto = "accumulate" | "add" | "auto" | "replace";\ntype CredentialMediationRequirement = "optional" | "required" | "silent";\ntype DOMParserSupportedType = "application/xhtml+xml" | "application/xml" | "image/svg+xml" | "text/html" | "text/xml";\ntype DirectionSetting = "" | "lr" | "rl";\ntype DisplayCaptureSurfaceType = "application" | "browser" | "monitor" | "window";\ntype DistanceModelType = "exponential" | "inverse" | "linear";\ntype DocumentReadyState = "complete" | "interactive" | "loading";\ntype EndOfStreamError = "decode" | "network";\ntype EndingType = "native" | "transparent";\ntype FillMode = "auto" | "backwards" | "both" | "forwards" | "none";\ntype FullscreenNavigationUI = "auto" | "hide" | "show";\ntype GamepadHand = "" | "left" | "right";\ntype GamepadHapticActuatorType = "vibration";\ntype GamepadMappingType = "" | "standard";\ntype IDBCursorDirection = "next" | "nextunique" | "prev" | "prevunique";\ntype IDBRequestReadyState = "done" | "pending";\ntype IDBTransactionMode = "readonly" | "readwrite" | "versionchange";\ntype ImageOrientation = "flipY" | "none";\ntype ImageSmoothingQuality = "high" | "low" | "medium";\ntype IterationCompositeOperation = "accumulate" | "replace";\ntype KeyFormat = "jwk" | "pkcs8" | "raw" | "spki";\ntype KeyType = "private" | "public" | "secret";\ntype KeyUsage = "decrypt" | "deriveBits" | "deriveKey" | "encrypt" | "sign" | "unwrapKey" | "verify" | "wrapKey";\ntype LineAlignSetting = "center" | "end" | "start";\ntype ListeningState = "active" | "disambiguation" | "inactive";\ntype MSCredentialType = "FIDO_2_0";\ntype MSTransportType = "BT" | "Embedded" | "NFC" | "USB";\ntype MSWebViewPermissionState = "allow" | "defer" | "deny" | "unknown";\ntype MSWebViewPermissionType = "geolocation" | "media" | "pointerlock" | "unlimitedIndexedDBQuota" | "webnotifications";\ntype MediaDeviceKind = "audioinput" | "audiooutput" | "videoinput";\ntype MediaKeyMessageType = "individualization-request" | "license-release" | "license-renewal" | "license-request";\ntype MediaKeySessionType = "persistent-license" | "temporary";\ntype MediaKeyStatus = "expired" | "internal-error" | "output-downscaled" | "output-restricted" | "released" | "status-pending" | "usable";\ntype MediaKeysRequirement = "not-allowed" | "optional" | "required";\ntype MediaStreamTrackState = "ended" | "live";\ntype NavigationReason = "down" | "left" | "right" | "up";\ntype NavigationType = "back_forward" | "navigate" | "prerender" | "reload";\ntype NotificationDirection = "auto" | "ltr" | "rtl";\ntype NotificationPermission = "default" | "denied" | "granted";\ntype OffscreenRenderingContextId = "2d" | "bitmaprenderer" | "webgl" | "webgl2";\ntype OrientationLockType = "any" | "landscape" | "landscape-primary" | "landscape-secondary" | "natural" | "portrait" | "portrait-primary" | "portrait-secondary";\ntype OrientationType = "landscape-primary" | "landscape-secondary" | "portrait-primary" | "portrait-secondary";\ntype OscillatorType = "custom" | "sawtooth" | "sine" | "square" | "triangle";\ntype OverSampleType = "2x" | "4x" | "none";\ntype PanningModelType = "HRTF" | "equalpower";\ntype PaymentComplete = "fail" | "success" | "unknown";\ntype PaymentShippingType = "delivery" | "pickup" | "shipping";\ntype PermissionName = "accelerometer" | "ambient-light-sensor" | "background-sync" | "bluetooth" | "camera" | "clipboard" | "device-info" | "geolocation" | "gyroscope" | "magnetometer" | "microphone" | "midi" | "notifications" | "persistent-storage" | "push" | "speaker";\ntype PermissionState = "denied" | "granted" | "prompt";\ntype PlaybackDirection = "alternate" | "alternate-reverse" | "normal" | "reverse";\ntype PositionAlignSetting = "auto" | "center" | "line-left" | "line-right";\ntype PremultiplyAlpha = "default" | "none" | "premultiply";\ntype PublicKeyCredentialType = "public-key";\ntype PushEncryptionKeyName = "auth" | "p256dh";\ntype PushPermissionState = "denied" | "granted" | "prompt";\ntype RTCBundlePolicy = "balanced" | "max-bundle" | "max-compat";\ntype RTCDataChannelState = "closed" | "closing" | "connecting" | "open";\ntype RTCDegradationPreference = "balanced" | "maintain-framerate" | "maintain-resolution";\ntype RTCDtlsRole = "auto" | "client" | "server";\ntype RTCDtlsTransportState = "closed" | "connected" | "connecting" | "failed" | "new";\ntype RTCDtxStatus = "disabled" | "enabled";\ntype RTCErrorDetailType = "data-channel-failure" | "dtls-failure" | "fingerprint-failure" | "hardware-encoder-error" | "hardware-encoder-not-available" | "idp-bad-script-failure" | "idp-execution-failure" | "idp-load-failure" | "idp-need-login" | "idp-timeout" | "idp-tls-failure" | "idp-token-expired" | "idp-token-invalid" | "sctp-failure" | "sdp-syntax-error";\ntype RTCIceCandidateType = "host" | "prflx" | "relay" | "srflx";\ntype RTCIceComponent = "rtcp" | "rtp";\ntype RTCIceConnectionState = "checking" | "closed" | "completed" | "connected" | "disconnected" | "failed" | "new";\ntype RTCIceCredentialType = "oauth" | "password";\ntype RTCIceGatherPolicy = "all" | "nohost" | "relay";\ntype RTCIceGathererState = "complete" | "gathering" | "new";\ntype RTCIceGatheringState = "complete" | "gathering" | "new";\ntype RTCIceProtocol = "tcp" | "udp";\ntype RTCIceRole = "controlled" | "controlling" | "unknown";\ntype RTCIceTcpCandidateType = "active" | "passive" | "so";\ntype RTCIceTransportPolicy = "all" | "relay";\ntype RTCIceTransportState = "checking" | "closed" | "completed" | "connected" | "disconnected" | "failed" | "new";\ntype RTCPeerConnectionState = "closed" | "connected" | "connecting" | "disconnected" | "failed" | "new";\ntype RTCPriorityType = "high" | "low" | "medium" | "very-low";\ntype RTCRtcpMuxPolicy = "negotiate" | "require";\ntype RTCRtpTransceiverDirection = "inactive" | "recvonly" | "sendonly" | "sendrecv" | "stopped";\ntype RTCSctpTransportState = "closed" | "connected" | "connecting";\ntype RTCSdpType = "answer" | "offer" | "pranswer" | "rollback";\ntype RTCSignalingState = "closed" | "have-local-offer" | "have-local-pranswer" | "have-remote-offer" | "have-remote-pranswer" | "stable";\ntype RTCStatsIceCandidatePairState = "cancelled" | "failed" | "frozen" | "inprogress" | "succeeded" | "waiting";\ntype RTCStatsIceCandidateType = "host" | "peerreflexive" | "relayed" | "serverreflexive";\ntype RTCStatsType = "candidatepair" | "datachannel" | "inboundrtp" | "localcandidate" | "outboundrtp" | "remotecandidate" | "session" | "track" | "transport";\ntype ReadyState = "closed" | "ended" | "open";\ntype ReferrerPolicy = "" | "no-referrer" | "no-referrer-when-downgrade" | "origin" | "origin-when-cross-origin" | "same-origin" | "strict-origin" | "strict-origin-when-cross-origin" | "unsafe-url";\ntype RequestCache = "default" | "force-cache" | "no-cache" | "no-store" | "only-if-cached" | "reload";\ntype RequestCredentials = "include" | "omit" | "same-origin";\ntype RequestDestination = "" | "audio" | "audioworklet" | "document" | "embed" | "font" | "image" | "manifest" | "object" | "paintworklet" | "report" | "script" | "sharedworker" | "style" | "track" | "video" | "worker" | "xslt";\ntype RequestMode = "cors" | "navigate" | "no-cors" | "same-origin";\ntype RequestRedirect = "error" | "follow" | "manual";\ntype ResidentKeyRequirement = "discouraged" | "preferred" | "required";\ntype ResizeQuality = "high" | "low" | "medium" | "pixelated";\ntype ResponseType = "basic" | "cors" | "default" | "error" | "opaque" | "opaqueredirect";\ntype ScopedCredentialType = "ScopedCred";\ntype ScrollBehavior = "auto" | "smooth";\ntype ScrollLogicalPosition = "center" | "end" | "nearest" | "start";\ntype ScrollRestoration = "auto" | "manual";\ntype ScrollSetting = "" | "up";\ntype SelectionMode = "end" | "preserve" | "select" | "start";\ntype ServiceWorkerState = "activated" | "activating" | "installed" | "installing" | "parsed" | "redundant";\ntype ServiceWorkerUpdateViaCache = "all" | "imports" | "none";\ntype ShadowRootMode = "closed" | "open";\ntype SpeechSynthesisErrorCode = "audio-busy" | "audio-hardware" | "canceled" | "interrupted" | "invalid-argument" | "language-unavailable" | "network" | "not-allowed" | "synthesis-failed" | "synthesis-unavailable" | "text-too-long" | "voice-unavailable";\ntype TextTrackKind = "captions" | "chapters" | "descriptions" | "metadata" | "subtitles";\ntype TextTrackMode = "disabled" | "hidden" | "showing";\ntype TouchType = "direct" | "stylus";\ntype Transport = "ble" | "nfc" | "usb";\ntype UserVerificationRequirement = "discouraged" | "preferred" | "required";\ntype VRDisplayEventReason = "mounted" | "navigation" | "requested" | "unmounted";\ntype VideoFacingModeEnum = "environment" | "left" | "right" | "user";\ntype VisibilityState = "hidden" | "visible";\ntype WebGLPowerPreference = "default" | "high-performance" | "low-power";\ntype WorkerType = "classic" | "module";\ntype XMLHttpRequestResponseType = "" | "arraybuffer" | "blob" | "document" | "json" | "text";\n',n.libFileMap["lib.dom.iterable.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// DOM Iterable APIs\n/////////////////////////////\n\ninterface AudioParam {\n    setValueCurveAtTime(values: Iterable<number>, startTime: number, duration: number): AudioParam;\n}\n\ninterface AudioParamMap extends ReadonlyMap<string, AudioParam> {\n}\n\ninterface BaseAudioContext {\n    createIIRFilter(feedforward: Iterable<number>, feedback: Iterable<number>): IIRFilterNode;\n    createPeriodicWave(real: Iterable<number>, imag: Iterable<number>, constraints?: PeriodicWaveConstraints): PeriodicWave;\n}\n\ninterface CSSRuleList {\n    [Symbol.iterator](): IterableIterator<CSSRule>;\n}\n\ninterface CSSStyleDeclaration {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface Cache {\n    addAll(requests: Iterable<RequestInfo>): Promise<void>;\n}\n\ninterface CanvasPathDrawingStyles {\n    setLineDash(segments: Iterable<number>): void;\n}\n\ninterface ClientRectList {\n    [Symbol.iterator](): IterableIterator<ClientRect>;\n}\n\ninterface DOMRectList {\n    [Symbol.iterator](): IterableIterator<DOMRect>;\n}\n\ninterface DOMStringList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface DOMTokenList {\n    [Symbol.iterator](): IterableIterator<string>;\n    entries(): IterableIterator<[number, string]>;\n    keys(): IterableIterator<number>;\n    values(): IterableIterator<string>;\n}\n\ninterface DataTransferItemList {\n    [Symbol.iterator](): IterableIterator<DataTransferItem>;\n}\n\ninterface FileList {\n    [Symbol.iterator](): IterableIterator<File>;\n}\n\ninterface FormData {\n    [Symbol.iterator](): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns a list of keys in the list.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the list.\n     */\n    values(): IterableIterator<FormDataEntryValue>;\n}\n\ninterface HTMLAllCollection {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLCollectionBase {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLCollectionOf<T extends Element> {\n    [Symbol.iterator](): IterableIterator<T>;\n}\n\ninterface HTMLFormElement {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface HTMLSelectElement {\n    [Symbol.iterator](): IterableIterator<Element>;\n}\n\ninterface Headers {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all key/value pairs contained in this object.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all keys of the key/value pairs contained in this object.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns an iterator allowing to go through all values of the key/value pairs contained in this object.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface IDBDatabase {\n    /**\n     * Returns a new transaction with the given mode ("readonly" or "readwrite") and scope which can be a single object store name or an array of names.\n     */\n    transaction(storeNames: string | Iterable<string>, mode?: IDBTransactionMode): IDBTransaction;\n}\n\ninterface IDBObjectStore {\n    /**\n     * Creates a new index in store with the given name, keyPath and options and returns a new IDBIndex. If the keyPath and options define constraints that cannot be satisfied with the data already in store the upgrade transaction will abort with a "ConstraintError" DOMException.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createIndex(name: string, keyPath: string | Iterable<string>, options?: IDBIndexParameters): IDBIndex;\n}\n\ninterface MediaKeyStatusMap {\n    [Symbol.iterator](): IterableIterator<[BufferSource, MediaKeyStatus]>;\n    entries(): IterableIterator<[BufferSource, MediaKeyStatus]>;\n    keys(): IterableIterator<BufferSource>;\n    values(): IterableIterator<MediaKeyStatus>;\n}\n\ninterface MediaList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface MimeTypeArray {\n    [Symbol.iterator](): IterableIterator<MimeType>;\n}\n\ninterface NamedNodeMap {\n    [Symbol.iterator](): IterableIterator<Attr>;\n}\n\ninterface Navigator {\n    requestMediaKeySystemAccess(keySystem: string, supportedConfigurations: Iterable<MediaKeySystemConfiguration>): Promise<MediaKeySystemAccess>;\n}\n\ninterface NodeList {\n    [Symbol.iterator](): IterableIterator<Node>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[number, Node]>;\n    /**\n     * Returns an list of keys in the list.\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the list.\n     */\n    values(): IterableIterator<Node>;\n}\n\ninterface NodeListOf<TNode extends Node> {\n    [Symbol.iterator](): IterableIterator<TNode>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[number, TNode]>;\n    /**\n     * Returns an list of keys in the list.\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the list.\n     */\n    values(): IterableIterator<TNode>;\n}\n\ninterface Plugin {\n    [Symbol.iterator](): IterableIterator<MimeType>;\n}\n\ninterface PluginArray {\n    [Symbol.iterator](): IterableIterator<Plugin>;\n}\n\ninterface RTCRtpTransceiver {\n    setCodecPreferences(codecs: Iterable<RTCRtpCodecCapability>): void;\n}\n\ninterface RTCStatsReport extends ReadonlyMap<string, any> {\n}\n\ninterface SVGLengthList {\n    [Symbol.iterator](): IterableIterator<SVGLength>;\n}\n\ninterface SVGNumberList {\n    [Symbol.iterator](): IterableIterator<SVGNumber>;\n}\n\ninterface SVGPointList {\n    [Symbol.iterator](): IterableIterator<DOMPoint>;\n}\n\ninterface SVGStringList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface SourceBufferList {\n    [Symbol.iterator](): IterableIterator<SourceBuffer>;\n}\n\ninterface SpeechGrammarList {\n    [Symbol.iterator](): IterableIterator<SpeechGrammar>;\n}\n\ninterface SpeechRecognitionResult {\n    [Symbol.iterator](): IterableIterator<SpeechRecognitionAlternative>;\n}\n\ninterface SpeechRecognitionResultList {\n    [Symbol.iterator](): IterableIterator<SpeechRecognitionResult>;\n}\n\ninterface StyleSheetList {\n    [Symbol.iterator](): IterableIterator<CSSStyleSheet>;\n}\n\ninterface TextTrackCueList {\n    [Symbol.iterator](): IterableIterator<TextTrackCue>;\n}\n\ninterface TextTrackList {\n    [Symbol.iterator](): IterableIterator<TextTrack>;\n}\n\ninterface TouchList {\n    [Symbol.iterator](): IterableIterator<Touch>;\n}\n\ninterface URLSearchParams {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the search params.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns a list of keys in the search params.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the search params.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface VRDisplay {\n    requestPresent(layers: Iterable<VRLayer>): Promise<void>;\n}\n\ninterface WEBGL_draw_buffers {\n    drawBuffersWEBGL(buffers: Iterable<GLenum>): void;\n}\n\ninterface WebAuthentication {\n    makeCredential(accountInformation: Account, cryptoParameters: Iterable<ScopedCredentialParameters>, attestationChallenge: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer | null, options?: ScopedCredentialOptions): Promise<ScopedCredentialInfo>;\n}\n\ninterface WebGL2RenderingContextBase {\n    clearBufferfv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLfloat>, srcOffset?: GLuint): void;\n    clearBufferiv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLint>, srcOffset?: GLuint): void;\n    clearBufferuiv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLuint>, srcOffset?: GLuint): void;\n    drawBuffers(buffers: Iterable<GLenum>): void;\n    getActiveUniforms(program: WebGLProgram, uniformIndices: Iterable<GLuint>, pname: GLenum): any;\n    getUniformIndices(program: WebGLProgram, uniformNames: Iterable<string>): Iterable<GLuint> | null;\n    invalidateFramebuffer(target: GLenum, attachments: Iterable<GLenum>): void;\n    invalidateSubFramebuffer(target: GLenum, attachments: Iterable<GLenum>, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    transformFeedbackVaryings(program: WebGLProgram, varyings: Iterable<string>, bufferMode: GLenum): void;\n    uniform1uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    vertexAttribI4iv(index: GLuint, values: Iterable<GLint>): void;\n    vertexAttribI4uiv(index: GLuint, values: Iterable<GLuint>): void;\n}\n\ninterface WebGL2RenderingContextOverloads {\n    uniform1fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform1iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n}\n\ninterface WebGLRenderingContextBase {\n    vertexAttrib1fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib2fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib3fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib4fv(index: GLuint, values: Iterable<GLfloat>): void;\n}\n\ninterface WebGLRenderingContextOverloads {\n    uniform1fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform1iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform2fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform2iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform3fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform3iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform4fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform4iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n}\n',n.libFileMap["lib.es2015.collection.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Map<K, V> {\n    clear(): void;\n    delete(key: K): boolean;\n    forEach(callbackfn: (value: V, key: K, map: Map<K, V>) => void, thisArg?: any): void;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    set(key: K, value: V): this;\n    readonly size: number;\n}\n\ninterface MapConstructor {\n    new(): Map<any, any>;\n    new<K, V>(entries?: readonly (readonly [K, V])[] | null): Map<K, V>;\n    readonly prototype: Map<any, any>;\n}\ndeclare var Map: MapConstructor;\n\ninterface ReadonlyMap<K, V> {\n    forEach(callbackfn: (value: V, key: K, map: ReadonlyMap<K, V>) => void, thisArg?: any): void;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    readonly size: number;\n}\n\ninterface WeakMap<K extends object, V> {\n    delete(key: K): boolean;\n    get(key: K): V | undefined;\n    has(key: K): boolean;\n    set(key: K, value: V): this;\n}\n\ninterface WeakMapConstructor {\n    new <K extends object = object, V = any>(entries?: readonly [K, V][] | null): WeakMap<K, V>;\n    readonly prototype: WeakMap<object, any>;\n}\ndeclare var WeakMap: WeakMapConstructor;\n\ninterface Set<T> {\n    add(value: T): this;\n    clear(): void;\n    delete(value: T): boolean;\n    forEach(callbackfn: (value: T, value2: T, set: Set<T>) => void, thisArg?: any): void;\n    has(value: T): boolean;\n    readonly size: number;\n}\n\ninterface SetConstructor {\n    new <T = any>(values?: readonly T[] | null): Set<T>;\n    readonly prototype: Set<any>;\n}\ndeclare var Set: SetConstructor;\n\ninterface ReadonlySet<T> {\n    forEach(callbackfn: (value: T, value2: T, set: ReadonlySet<T>) => void, thisArg?: any): void;\n    has(value: T): boolean;\n    readonly size: number;\n}\n\ninterface WeakSet<T extends object> {\n    add(value: T): this;\n    delete(value: T): boolean;\n    has(value: T): boolean;\n}\n\ninterface WeakSetConstructor {\n    new <T extends object = object>(values?: readonly T[] | null): WeakSet<T>;\n    readonly prototype: WeakSet<object>;\n}\ndeclare var WeakSet: WeakSetConstructor;\n',n.libFileMap["lib.es2015.core.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Array<T> {\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find<S extends T>(predicate: (this: void, value: T, index: number, obj: T[]) => value is S, thisArg?: any): S | undefined;\n    find(predicate: (value: T, index: number, obj: T[]) => unknown, thisArg?: any): T | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: T, index: number, obj: T[]) => unknown, thisArg?: any): number;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: T, start?: number, end?: number): this;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n}\n\ninterface ArrayConstructor {\n    /**\n     * Creates an array from an array-like object.\n     * @param arrayLike An array-like object to convert to an array.\n     */\n    from<T>(arrayLike: ArrayLike<T>): T[];\n\n    /**\n     * Creates an array from an iterable object.\n     * @param arrayLike An array-like object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of \'this\' used to invoke the mapfn.\n     */\n    from<T, U>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => U, thisArg?: any): U[];\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of<T>(...items: T[]): T[];\n}\n\ninterface DateConstructor {\n    new (value: number | string | Date): Date;\n}\n\ninterface Function {\n    /**\n     * Returns the name of the function. Function names are read-only and can not be changed.\n     */\n    readonly name: string;\n}\n\ninterface Math {\n    /**\n     * Returns the number of leading zero bits in the 32-bit binary representation of a number.\n     * @param x A numeric expression.\n     */\n    clz32(x: number): number;\n\n    /**\n     * Returns the result of 32-bit multiplication of two numbers.\n     * @param x First number\n     * @param y Second number\n     */\n    imul(x: number, y: number): number;\n\n    /**\n     * Returns the sign of the x, indicating whether x is positive, negative or zero.\n     * @param x The numeric expression to test\n     */\n    sign(x: number): number;\n\n    /**\n     * Returns the base 10 logarithm of a number.\n     * @param x A numeric expression.\n     */\n    log10(x: number): number;\n\n    /**\n     * Returns the base 2 logarithm of a number.\n     * @param x A numeric expression.\n     */\n    log2(x: number): number;\n\n    /**\n     * Returns the natural logarithm of 1 + x.\n     * @param x A numeric expression.\n     */\n    log1p(x: number): number;\n\n    /**\n     * Returns the result of (e^x - 1), which is an implementation-dependent approximation to\n     * subtracting 1 from the exponential function of x (e raised to the power of x, where e\n     * is the base of the natural logarithms).\n     * @param x A numeric expression.\n     */\n    expm1(x: number): number;\n\n    /**\n     * Returns the hyperbolic cosine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    cosh(x: number): number;\n\n    /**\n     * Returns the hyperbolic sine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    sinh(x: number): number;\n\n    /**\n     * Returns the hyperbolic tangent of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    tanh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic cosine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    acosh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic sine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    asinh(x: number): number;\n\n    /**\n     * Returns the inverse hyperbolic tangent of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    atanh(x: number): number;\n\n    /**\n     * Returns the square root of the sum of squares of its arguments.\n     * @param values Values to compute the square root for.\n     *     If no arguments are passed, the result is +0.\n     *     If there is only one argument, the result is the absolute value.\n     *     If any argument is +Infinity or -Infinity, the result is +Infinity.\n     *     If any argument is NaN, the result is NaN.\n     *     If all arguments are either +0 or −0, the result is +0.\n     */\n    hypot(...values: number[]): number;\n\n    /**\n     * Returns the integral part of the a numeric expression, x, removing any fractional digits.\n     * If x is already an integer, the result is x.\n     * @param x A numeric expression.\n     */\n    trunc(x: number): number;\n\n    /**\n     * Returns the nearest single precision float representation of a number.\n     * @param x A numeric expression.\n     */\n    fround(x: number): number;\n\n    /**\n     * Returns an implementation-dependent approximation to the cube root of number.\n     * @param x A numeric expression.\n     */\n    cbrt(x: number): number;\n}\n\ninterface NumberConstructor {\n    /**\n     * The value of Number.EPSILON is the difference between 1 and the smallest value greater than 1\n     * that is representable as a Number value, which is approximately:\n     * 2.2204460492503130808472633361816 x 10‍−‍16.\n     */\n    readonly EPSILON: number;\n\n    /**\n     * Returns true if passed value is finite.\n     * Unlike the global isFinite, Number.isFinite doesn\'t forcibly convert the parameter to a\n     * number. Only finite values of the type number, result in true.\n     * @param number A numeric value.\n     */\n    isFinite(number: unknown): boolean;\n\n    /**\n     * Returns true if the value passed is an integer, false otherwise.\n     * @param number A numeric value.\n     */\n    isInteger(number: unknown): boolean;\n\n    /**\n     * Returns a Boolean value that indicates whether a value is the reserved value NaN (not a\n     * number). Unlike the global isNaN(), Number.isNaN() doesn\'t forcefully convert the parameter\n     * to a number. Only values of the type number, that are also NaN, result in true.\n     * @param number A numeric value.\n     */\n    isNaN(number: unknown): boolean;\n\n    /**\n     * Returns true if the value passed is a safe integer.\n     * @param number A numeric value.\n     */\n    isSafeInteger(number: unknown): boolean;\n\n    /**\n     * The value of the largest integer n such that n and n + 1 are both exactly representable as\n     * a Number value.\n     * The value of Number.MAX_SAFE_INTEGER is 9007199254740991 2^53 − 1.\n     */\n    readonly MAX_SAFE_INTEGER: number;\n\n    /**\n     * The value of the smallest integer n such that n and n − 1 are both exactly representable as\n     * a Number value.\n     * The value of Number.MIN_SAFE_INTEGER is −9007199254740991 (−(2^53 − 1)).\n     */\n    readonly MIN_SAFE_INTEGER: number;\n\n    /**\n     * Converts a string to a floating-point number.\n     * @param string A string that contains a floating-point number.\n     */\n    parseFloat(string: string): number;\n\n    /**\n     * Converts A string to an integer.\n     * @param s A string to convert into a number.\n     * @param radix A value between 2 and 36 that specifies the base of the number in numString.\n     * If this argument is not supplied, strings with a prefix of \'0x\' are considered hexadecimal.\n     * All other strings are considered decimal.\n     */\n    parseInt(string: string, radix?: number): number;\n}\n\ninterface ObjectConstructor {\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source The source object from which to copy properties.\n     */\n    assign<T, U>(target: T, source: U): T & U;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source1 The first source object from which to copy properties.\n     * @param source2 The second source object from which to copy properties.\n     */\n    assign<T, U, V>(target: T, source1: U, source2: V): T & U & V;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param source1 The first source object from which to copy properties.\n     * @param source2 The second source object from which to copy properties.\n     * @param source3 The third source object from which to copy properties.\n     */\n    assign<T, U, V, W>(target: T, source1: U, source2: V, source3: W): T & U & V & W;\n\n    /**\n     * Copy the values of all of the enumerable own properties from one or more source objects to a\n     * target object. Returns the target object.\n     * @param target The target object to copy to.\n     * @param sources One or more source objects from which to copy properties\n     */\n    assign(target: object, ...sources: any[]): any;\n\n    /**\n     * Returns an array of all symbol properties found directly on object o.\n     * @param o Object to retrieve the symbols from.\n     */\n    getOwnPropertySymbols(o: any): symbol[];\n\n    /**\n     * Returns the names of the enumerable string properties and methods of an object.\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    keys(o: {}): string[];\n\n    /**\n     * Returns true if the values are the same value, false otherwise.\n     * @param value1 The first value.\n     * @param value2 The second value.\n     */\n    is(value1: any, value2: any): boolean;\n\n    /**\n     * Sets the prototype of a specified object o to object proto or null. Returns the object o.\n     * @param o The object to change its prototype.\n     * @param proto The value of the new prototype or null.\n     */\n    setPrototypeOf(o: any, proto: object | null): any;\n}\n\ninterface ReadonlyArray<T> {\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find<S extends T>(predicate: (this: void, value: T, index: number, obj: readonly T[]) => value is S, thisArg?: any): S | undefined;\n    find(predicate: (value: T, index: number, obj: readonly T[]) => unknown, thisArg?: any): T | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: T, index: number, obj: readonly T[]) => unknown, thisArg?: any): number;\n}\n\ninterface RegExp {\n    /**\n     * Returns a string indicating the flags of the regular expression in question. This field is read-only.\n     * The characters in this string are sequenced and concatenated in the following order:\n     *\n     *    - "g" for global\n     *    - "i" for ignoreCase\n     *    - "m" for multiline\n     *    - "u" for unicode\n     *    - "y" for sticky\n     *\n     * If no flags are set, the value is the empty string.\n     */\n    readonly flags: string;\n\n    /**\n     * Returns a Boolean value indicating the state of the sticky flag (y) used with a regular\n     * expression. Default is false. Read-only.\n     */\n    readonly sticky: boolean;\n\n    /**\n     * Returns a Boolean value indicating the state of the Unicode flag (u) used with a regular\n     * expression. Default is false. Read-only.\n     */\n    readonly unicode: boolean;\n}\n\ninterface RegExpConstructor {\n    new (pattern: RegExp | string, flags?: string): RegExp;\n    (pattern: RegExp | string, flags?: string): RegExp;\n}\n\ninterface String {\n    /**\n     * Returns a nonnegative integer Number less than 1114112 (0x110000) that is the code point\n     * value of the UTF-16 encoded code point starting at the string element at position pos in\n     * the String resulting from converting this object to a String.\n     * If there is no element at that position, the result is undefined.\n     * If a valid UTF-16 surrogate pair does not begin at pos, the result is the code unit at pos.\n     */\n    codePointAt(pos: number): number | undefined;\n\n    /**\n     * Returns true if searchString appears as a substring of the result of converting this\n     * object to a String, at one or more positions that are\n     * greater than or equal to position; otherwise, returns false.\n     * @param searchString search string\n     * @param position If position is undefined, 0 is assumed, so as to search all of the String.\n     */\n    includes(searchString: string, position?: number): boolean;\n\n    /**\n     * Returns true if the sequence of elements of searchString converted to a String is the\n     * same as the corresponding elements of this object (converted to a String) starting at\n     * endPosition – length(this). Otherwise returns false.\n     */\n    endsWith(searchString: string, endPosition?: number): boolean;\n\n    /**\n     * Returns the String value result of normalizing the string into the normalization form\n     * named by form as specified in Unicode Standard Annex #15, Unicode Normalization Forms.\n     * @param form Applicable values: "NFC", "NFD", "NFKC", or "NFKD", If not specified default\n     * is "NFC"\n     */\n    normalize(form: "NFC" | "NFD" | "NFKC" | "NFKD"): string;\n\n    /**\n     * Returns the String value result of normalizing the string into the normalization form\n     * named by form as specified in Unicode Standard Annex #15, Unicode Normalization Forms.\n     * @param form Applicable values: "NFC", "NFD", "NFKC", or "NFKD", If not specified default\n     * is "NFC"\n     */\n    normalize(form?: string): string;\n\n    /**\n     * Returns a String value that is made from count copies appended together. If count is 0,\n     * the empty string is returned.\n     * @param count number of copies to append\n     */\n    repeat(count: number): string;\n\n    /**\n     * Returns true if the sequence of elements of searchString converted to a String is the\n     * same as the corresponding elements of this object (converted to a String) starting at\n     * position. Otherwise returns false.\n     */\n    startsWith(searchString: string, position?: number): boolean;\n\n    /**\n     * Returns an `<a>` HTML anchor element and sets the name attribute to the text value\n     * @param name\n     */\n    anchor(name: string): string;\n\n    /** Returns a `<big>` HTML element */\n    big(): string;\n\n    /** Returns a `<blink>` HTML element */\n    blink(): string;\n\n    /** Returns a `<b>` HTML element */\n    bold(): string;\n\n    /** Returns a `<tt>` HTML element */\n    fixed(): string;\n\n    /** Returns a `<font>` HTML element and sets the color attribute value */\n    fontcolor(color: string): string;\n\n    /** Returns a `<font>` HTML element and sets the size attribute value */\n    fontsize(size: number): string;\n\n    /** Returns a `<font>` HTML element and sets the size attribute value */\n    fontsize(size: string): string;\n\n    /** Returns an `<i>` HTML element */\n    italics(): string;\n\n    /** Returns an `<a>` HTML element and sets the href attribute value */\n    link(url: string): string;\n\n    /** Returns a `<small>` HTML element */\n    small(): string;\n\n    /** Returns a `<strike>` HTML element */\n    strike(): string;\n\n    /** Returns a `<sub>` HTML element */\n    sub(): string;\n\n    /** Returns a `<sup>` HTML element */\n    sup(): string;\n}\n\ninterface StringConstructor {\n    /**\n     * Return the String value whose elements are, in order, the elements in the List elements.\n     * If length is 0, the empty string is returned.\n     */\n    fromCodePoint(...codePoints: number[]): string;\n\n    /**\n     * String.raw is intended for use as a tag function of a Tagged Template String. When called\n     * as such the first argument will be a well formed template call site object and the rest\n     * parameter will contain the substitution values.\n     * @param template A well-formed template string call site representation.\n     * @param substitutions A set of substitution values.\n     */\n    raw(template: TemplateStringsArray, ...substitutions: any[]): string;\n}\n',n.libFileMap["lib.es2015.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es5" />\n/// <reference lib="es2015.core" />\n/// <reference lib="es2015.collection" />\n/// <reference lib="es2015.iterable" />\n/// <reference lib="es2015.generator" />\n/// <reference lib="es2015.promise" />\n/// <reference lib="es2015.proxy" />\n/// <reference lib="es2015.reflect" />\n/// <reference lib="es2015.symbol" />\n/// <reference lib="es2015.symbol.wellknown" />\n',n.libFileMap["lib.es2015.generator.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.iterable" />\n\ninterface Generator<T = unknown, TReturn = any, TNext = unknown> extends Iterator<T, TReturn, TNext> {\n    // NOTE: \'next\' is defined using a tuple to ensure we report the correct assignability errors in all places.\n    next(...args: [] | [TNext]): IteratorResult<T, TReturn>;\n    return(value: TReturn): IteratorResult<T, TReturn>;\n    throw(e: any): IteratorResult<T, TReturn>;\n    [Symbol.iterator](): Generator<T, TReturn, TNext>;\n}\n\ninterface GeneratorFunction {\n    /**\n     * Creates a new Generator object.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: any[]): Generator;\n    /**\n     * Creates a new Generator object.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: any[]): Generator;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Generator;\n}\n\ninterface GeneratorFunctionConstructor {\n    /**\n     * Creates a new Generator function.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: string[]): GeneratorFunction;\n    /**\n     * Creates a new Generator function.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: string[]): GeneratorFunction;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: GeneratorFunction;\n}\n',n.libFileMap["lib.es2015.iterable.d.ts"]="/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the \"License\"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib=\"true\"/>\n\n\n/// <reference lib=\"es2015.symbol\" />\n\ninterface SymbolConstructor {\n    /**\n     * A method that returns the default iterator for an object. Called by the semantics of the\n     * for-of statement.\n     */\n    readonly iterator: symbol;\n}\n\ninterface IteratorYieldResult<TYield> {\n    done?: false;\n    value: TYield;\n}\n\ninterface IteratorReturnResult<TReturn> {\n    done: true;\n    value: TReturn;\n}\n\ntype IteratorResult<T, TReturn = any> = IteratorYieldResult<T> | IteratorReturnResult<TReturn>;\n\ninterface Iterator<T, TReturn = any, TNext = undefined> {\n    // NOTE: 'next' is defined using a tuple to ensure we report the correct assignability errors in all places.\n    next(...args: [] | [TNext]): IteratorResult<T, TReturn>;\n    return?(value?: TReturn): IteratorResult<T, TReturn>;\n    throw?(e?: any): IteratorResult<T, TReturn>;\n}\n\ninterface Iterable<T> {\n    [Symbol.iterator](): Iterator<T>;\n}\n\ninterface IterableIterator<T> extends Iterator<T> {\n    [Symbol.iterator](): IterableIterator<T>;\n}\n\ninterface Array<T> {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, T]>;\n\n    /**\n     * Returns an iterable of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an iterable of values in the array\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface ArrayConstructor {\n    /**\n     * Creates an array from an iterable object.\n     * @param iterable An iterable object to convert to an array.\n     */\n    from<T>(iterable: Iterable<T> | ArrayLike<T>): T[];\n\n    /**\n     * Creates an array from an iterable object.\n     * @param iterable An iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T, U>(iterable: Iterable<T> | ArrayLike<T>, mapfn: (v: T, k: number) => U, thisArg?: any): U[];\n}\n\ninterface ReadonlyArray<T> {\n    /** Iterator of values in the array. */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, T]>;\n\n    /**\n     * Returns an iterable of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an iterable of values in the array\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface IArguments {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<any>;\n}\n\ninterface Map<K, V> {\n    /** Returns an iterable of entries in the map. */\n    [Symbol.iterator](): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the map.\n     */\n    entries(): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of keys in the map\n     */\n    keys(): IterableIterator<K>;\n\n    /**\n     * Returns an iterable of values in the map\n     */\n    values(): IterableIterator<V>;\n}\n\ninterface ReadonlyMap<K, V> {\n    /** Returns an iterable of entries in the map. */\n    [Symbol.iterator](): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of key, value pairs for every entry in the map.\n     */\n    entries(): IterableIterator<[K, V]>;\n\n    /**\n     * Returns an iterable of keys in the map\n     */\n    keys(): IterableIterator<K>;\n\n    /**\n     * Returns an iterable of values in the map\n     */\n    values(): IterableIterator<V>;\n}\n\ninterface MapConstructor {\n    new <K, V>(iterable: Iterable<readonly [K, V]>): Map<K, V>;\n}\n\ninterface WeakMap<K extends object, V> { }\n\ninterface WeakMapConstructor {\n    new <K extends object, V>(iterable: Iterable<[K, V]>): WeakMap<K, V>;\n}\n\ninterface Set<T> {\n    /** Iterates over values in the set. */\n    [Symbol.iterator](): IterableIterator<T>;\n    /**\n     * Returns an iterable of [v,v] pairs for every value `v` in the set.\n     */\n    entries(): IterableIterator<[T, T]>;\n    /**\n     * Despite its name, returns an iterable of the values in the set.\n     */\n    keys(): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of values in the set.\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface ReadonlySet<T> {\n    /** Iterates over values in the set. */\n    [Symbol.iterator](): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of [v,v] pairs for every value `v` in the set.\n     */\n    entries(): IterableIterator<[T, T]>;\n\n    /**\n     * Despite its name, returns an iterable of the values in the set.\n     */\n    keys(): IterableIterator<T>;\n\n    /**\n     * Returns an iterable of values in the set.\n     */\n    values(): IterableIterator<T>;\n}\n\ninterface SetConstructor {\n    new <T>(iterable?: Iterable<T> | null): Set<T>;\n}\n\ninterface WeakSet<T extends object> { }\n\ninterface WeakSetConstructor {\n    new <T extends object = object>(iterable: Iterable<T>): WeakSet<T>;\n}\n\ninterface Promise<T> { }\n\ninterface PromiseConstructor {\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An iterable of Promises.\n     * @returns A new Promise.\n     */\n    all<T>(values: Iterable<T | PromiseLike<T>>): Promise<T[]>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An iterable of Promises.\n     * @returns A new Promise.\n     */\n    race<T>(values: Iterable<T>): Promise<T extends PromiseLike<infer U> ? U : T>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An iterable of Promises.\n     * @returns A new Promise.\n     */\n    race<T>(values: Iterable<T | PromiseLike<T>>): Promise<T>;\n}\n\ninterface String {\n    /** Iterator */\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface Int8Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int8ArrayConstructor {\n    new (elements: Iterable<number>): Int8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int8Array;\n}\n\ninterface Uint8Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint8ArrayConstructor {\n    new (elements: Iterable<number>): Uint8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint8Array;\n}\n\ninterface Uint8ClampedArray {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint8ClampedArrayConstructor {\n    new (elements: Iterable<number>): Uint8ClampedArray;\n\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint8ClampedArray;\n}\n\ninterface Int16Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int16ArrayConstructor {\n    new (elements: Iterable<number>): Int16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int16Array;\n}\n\ninterface Uint16Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint16ArrayConstructor {\n    new (elements: Iterable<number>): Uint16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint16Array;\n}\n\ninterface Int32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Int32ArrayConstructor {\n    new (elements: Iterable<number>): Int32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Int32Array;\n}\n\ninterface Uint32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Uint32ArrayConstructor {\n    new (elements: Iterable<number>): Uint32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Uint32Array;\n}\n\ninterface Float32Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Float32ArrayConstructor {\n    new (elements: Iterable<number>): Float32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Float32Array;\n}\n\ninterface Float64Array {\n    [Symbol.iterator](): IterableIterator<number>;\n    /**\n     * Returns an array of key, value pairs for every entry in the array\n     */\n    entries(): IterableIterator<[number, number]>;\n    /**\n     * Returns an list of keys in the array\n     */\n    keys(): IterableIterator<number>;\n    /**\n     * Returns an list of values in the array\n     */\n    values(): IterableIterator<number>;\n}\n\ninterface Float64ArrayConstructor {\n    new (elements: Iterable<number>): Float64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from(arrayLike: Iterable<number>, mapfn?: (v: number, k: number) => number, thisArg?: any): Float64Array;\n}\n",n.libFileMap["lib.es2015.promise.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface PromiseConstructor {\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Promise<any>;\n\n    /**\n     * Creates a new Promise.\n     * @param executor A callback used to initialize the promise. This callback is passed two arguments:\n     * a resolve callback used to resolve the promise with a value or the result of another promise,\n     * and a reject callback used to reject the promise with a provided reason or error.\n     */\n    new <T>(executor: (resolve: (value: T | PromiseLike<T>) => void, reject: (reason?: any) => void) => void): Promise<T>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8, T9, T10>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>, T10 | PromiseLike<T10>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8, T9>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>, T9 | PromiseLike<T9>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8, T9]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7, T8>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>, T8 | PromiseLike<T8>]): Promise<[T1, T2, T3, T4, T5, T6, T7, T8]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6, T7>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>, T7 | PromiseLike<T7>]): Promise<[T1, T2, T3, T4, T5, T6, T7]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5, T6>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>, T6 | PromiseLike<T6>]): Promise<[T1, T2, T3, T4, T5, T6]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4, T5>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>, T5 | PromiseLike<T5>]): Promise<[T1, T2, T3, T4, T5]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3, T4>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>, T4 | PromiseLike<T4>]): Promise<[T1, T2, T3, T4]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2, T3>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>, T3 | PromiseLike<T3>]): Promise<[T1, T2, T3]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T1, T2>(values: readonly [T1 | PromiseLike<T1>, T2 | PromiseLike<T2>]): Promise<[T1, T2]>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all of the provided Promises\n     * resolve, or rejected when any Promise is rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    all<T>(values: readonly (T | PromiseLike<T>)[]): Promise<T[]>;\n\n    // see: lib.es2015.iterable.d.ts\n    // all<T>(values: Iterable<T | PromiseLike<T>>): Promise<T[]>;\n\n    /**\n     * Creates a Promise that is resolved or rejected when any of the provided Promises are resolved\n     * or rejected.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    race<T>(values: readonly T[]): Promise<T extends PromiseLike<infer U> ? U : T>;\n\n    // see: lib.es2015.iterable.d.ts\n    // race<T>(values: Iterable<T>): Promise<T extends PromiseLike<infer U> ? U : T>;\n\n    /**\n     * Creates a new rejected promise for the provided reason.\n     * @param reason The reason the promise was rejected.\n     * @returns A new rejected Promise.\n     */\n    reject<T = never>(reason?: any): Promise<T>;\n\n    /**\n     * Creates a new resolved promise.\n     * @returns A resolved promise.\n     */\n    resolve(): Promise<void>;\n\n    /**\n     * Creates a new resolved promise for the provided value.\n     * @param value A promise.\n     * @returns A promise whose internal state matches the provided promise.\n     */\n    resolve<T>(value: T | PromiseLike<T>): Promise<T>;\n}\n\ndeclare var Promise: PromiseConstructor;\n',n.libFileMap["lib.es2015.proxy.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface ProxyHandler<T extends object> {\n    getPrototypeOf? (target: T): object | null;\n    setPrototypeOf? (target: T, v: any): boolean;\n    isExtensible? (target: T): boolean;\n    preventExtensions? (target: T): boolean;\n    getOwnPropertyDescriptor? (target: T, p: PropertyKey): PropertyDescriptor | undefined;\n    has? (target: T, p: PropertyKey): boolean;\n    get? (target: T, p: PropertyKey, receiver: any): any;\n    set? (target: T, p: PropertyKey, value: any, receiver: any): boolean;\n    deleteProperty? (target: T, p: PropertyKey): boolean;\n    defineProperty? (target: T, p: PropertyKey, attributes: PropertyDescriptor): boolean;\n    ownKeys? (target: T): PropertyKey[];\n    apply? (target: T, thisArg: any, argArray?: any): any;\n    construct? (target: T, argArray: any, newTarget?: any): object;\n}\n\ninterface ProxyConstructor {\n    revocable<T extends object>(target: T, handler: ProxyHandler<T>): { proxy: T; revoke: () => void; };\n    new <T extends object>(target: T, handler: ProxyHandler<T>): T;\n}\ndeclare var Proxy: ProxyConstructor;\n',n.libFileMap["lib.es2015.reflect.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Reflect {\n    function apply(target: Function, thisArgument: any, argumentsList: ArrayLike<any>): any;\n    function construct(target: Function, argumentsList: ArrayLike<any>, newTarget?: any): any;\n    function defineProperty(target: object, propertyKey: PropertyKey, attributes: PropertyDescriptor): boolean;\n    function deleteProperty(target: object, propertyKey: PropertyKey): boolean;\n    function get(target: object, propertyKey: PropertyKey, receiver?: any): any;\n    function getOwnPropertyDescriptor(target: object, propertyKey: PropertyKey): PropertyDescriptor | undefined;\n    function getPrototypeOf(target: object): object;\n    function has(target: object, propertyKey: PropertyKey): boolean;\n    function isExtensible(target: object): boolean;\n    function ownKeys(target: object): PropertyKey[];\n    function preventExtensions(target: object): boolean;\n    function set(target: object, propertyKey: PropertyKey, value: any, receiver?: any): boolean;\n    function setPrototypeOf(target: object, proto: any): boolean;\n}\n',n.libFileMap["lib.es2015.symbol.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface SymbolConstructor {\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: Symbol;\n\n    /**\n     * Returns a new unique Symbol value.\n     * @param  description Description of the new Symbol object.\n     */\n    (description?: string | number): symbol;\n\n    /**\n     * Returns a Symbol object from the global symbol registry matching the given key if found.\n     * Otherwise, returns a new symbol with this key.\n     * @param key key to search for.\n     */\n    for(key: string): symbol;\n\n    /**\n     * Returns a key from the global symbol registry matching the given Symbol if found.\n     * Otherwise, returns a undefined.\n     * @param sym Symbol to find the key for.\n     */\n    keyFor(sym: symbol): string | undefined;\n}\n\ndeclare var Symbol: SymbolConstructor;',n.libFileMap["lib.es2015.symbol.wellknown.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.symbol" />\n\ninterface SymbolConstructor {\n    /**\n     * A method that determines if a constructor object recognizes an object as one of the\n     * constructor’s instances. Called by the semantics of the instanceof operator.\n     */\n    readonly hasInstance: symbol;\n\n    /**\n     * A Boolean value that if true indicates that an object should flatten to its array elements\n     * by Array.prototype.concat.\n     */\n    readonly isConcatSpreadable: symbol;\n\n    /**\n     * A regular expression method that matches the regular expression against a string. Called\n     * by the String.prototype.match method.\n     */\n    readonly match: symbol;\n\n    /**\n     * A regular expression method that replaces matched substrings of a string. Called by the\n     * String.prototype.replace method.\n     */\n    readonly replace: symbol;\n\n    /**\n     * A regular expression method that returns the index within a string that matches the\n     * regular expression. Called by the String.prototype.search method.\n     */\n    readonly search: symbol;\n\n    /**\n     * A function valued property that is the constructor function that is used to create\n     * derived objects.\n     */\n    readonly species: symbol;\n\n    /**\n     * A regular expression method that splits a string at the indices that match the regular\n     * expression. Called by the String.prototype.split method.\n     */\n    readonly split: symbol;\n\n    /**\n     * A method that converts an object to a corresponding primitive value.\n     * Called by the ToPrimitive abstract operation.\n     */\n    readonly toPrimitive: symbol;\n\n    /**\n     * A String value that is used in the creation of the default string description of an object.\n     * Called by the built-in method Object.prototype.toString.\n     */\n    readonly toStringTag: symbol;\n\n    /**\n     * An Object whose own property names are property names that are excluded from the \'with\'\n     * environment bindings of the associated objects.\n     */\n    readonly unscopables: symbol;\n}\n\ninterface Symbol {\n    /**\n     * Converts a Symbol object to a symbol.\n     */\n    [Symbol.toPrimitive](hint: string): symbol;\n\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Array<T> {\n    /**\n     * Returns an object whose properties have the value \'true\'\n     * when they will be absent when used in a \'with\' statement.\n     */\n    [Symbol.unscopables](): {\n        copyWithin: boolean;\n        entries: boolean;\n        fill: boolean;\n        find: boolean;\n        findIndex: boolean;\n        keys: boolean;\n        values: boolean;\n    };\n}\n\ninterface Date {\n    /**\n     * Converts a Date object to a string.\n     */\n    [Symbol.toPrimitive](hint: "default"): string;\n    /**\n     * Converts a Date object to a string.\n     */\n    [Symbol.toPrimitive](hint: "string"): string;\n    /**\n     * Converts a Date object to a number.\n     */\n    [Symbol.toPrimitive](hint: "number"): number;\n    /**\n     * Converts a Date object to a string or number.\n     *\n     * @param hint The strings "number", "string", or "default" to specify what primitive to return.\n     *\n     * @throws {TypeError} If \'hint\' was given something other than "number", "string", or "default".\n     * @returns A number if \'hint\' was "number", a string if \'hint\' was "string" or "default".\n     */\n    [Symbol.toPrimitive](hint: string): string | number;\n}\n\ninterface Map<K, V> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface WeakMap<K extends object, V> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Set<T> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface WeakSet<T extends object> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface JSON {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Function {\n    /**\n     * Determines whether the given value inherits from this function if this function was used\n     * as a constructor function.\n     *\n     * A constructor function can control which objects are recognized as its instances by\n     * \'instanceof\' by overriding this method.\n     */\n    [Symbol.hasInstance](value: any): boolean;\n}\n\ninterface GeneratorFunction {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Math {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Promise<T> {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface PromiseConstructor {\n    readonly [Symbol.species]: PromiseConstructor;\n}\n\ninterface RegExp {\n    /**\n     * Matches a string with this regular expression, and returns an array containing the results of\n     * that search.\n     * @param string A string to search within.\n     */\n    [Symbol.match](string: string): RegExpMatchArray | null;\n\n    /**\n     * Replaces text in a string, using this regular expression.\n     * @param string A String object or string literal whose contents matching against\n     *               this regular expression will be replaced\n     * @param replaceValue A String object or string literal containing the text to replace for every\n     *                     successful match of this regular expression.\n     */\n    [Symbol.replace](string: string, replaceValue: string): string;\n\n    /**\n     * Replaces text in a string, using this regular expression.\n     * @param string A String object or string literal whose contents matching against\n     *               this regular expression will be replaced\n     * @param replacer A function that returns the replacement text.\n     */\n    [Symbol.replace](string: string, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n     * Finds the position beginning first substring match in a regular expression search\n     * using this regular expression.\n     *\n     * @param string The string to search within.\n     */\n    [Symbol.search](string: string): number;\n\n    /**\n     * Returns an array of substrings that were delimited by strings in the original input that\n     * match against this regular expression.\n     *\n     * If the regular expression contains capturing parentheses, then each time this\n     * regular expression matches, the results (including any undefined results) of the\n     * capturing parentheses are spliced.\n     *\n     * @param string string value to split\n     * @param limit if not undefined, the output array is truncated so that it contains no more\n     * than \'limit\' elements.\n     */\n    [Symbol.split](string: string, limit?: number): string[];\n}\n\ninterface RegExpConstructor {\n    readonly [Symbol.species]: RegExpConstructor;\n}\n\ninterface String {\n    /**\n     * Matches a string or an object that supports being matched against, and returns an array\n     * containing the results of that search, or null if no matches are found.\n     * @param matcher An object that supports being matched against.\n     */\n    match(matcher: { [Symbol.match](string: string): RegExpMatchArray | null; }): RegExpMatchArray | null;\n\n    /**\n     * Replaces text in a string, using an object that supports replacement within a string.\n     * @param searchValue A object can search for and replace matches within a string.\n     * @param replaceValue A string containing the text to replace for every successful match of searchValue in this string.\n     */\n    replace(searchValue: { [Symbol.replace](string: string, replaceValue: string): string; }, replaceValue: string): string;\n\n    /**\n     * Replaces text in a string, using an object that supports replacement within a string.\n     * @param searchValue A object can search for and replace matches within a string.\n     * @param replacer A function that returns the replacement text.\n     */\n    replace(searchValue: { [Symbol.replace](string: string, replacer: (substring: string, ...args: any[]) => string): string; }, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n     * Finds the first substring match in a regular expression search.\n     * @param searcher An object which supports searching within a string.\n     */\n    search(searcher: { [Symbol.search](string: string): number; }): number;\n\n    /**\n     * Split a string into substrings using the specified separator and return them as an array.\n     * @param splitter An object that can split a string.\n     * @param limit A value used to limit the number of elements returned in the array.\n     */\n    split(splitter: { [Symbol.split](string: string, limit?: number): string[]; }, limit?: number): string[];\n}\n\ninterface ArrayBuffer {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface DataView {\n    readonly [Symbol.toStringTag]: string;\n}\n\ninterface Int8Array {\n    readonly [Symbol.toStringTag]: "Int8Array";\n}\n\ninterface Uint8Array {\n    readonly [Symbol.toStringTag]: "Uint8Array";\n}\n\ninterface Uint8ClampedArray {\n    readonly [Symbol.toStringTag]: "Uint8ClampedArray";\n}\n\ninterface Int16Array {\n    readonly [Symbol.toStringTag]: "Int16Array";\n}\n\ninterface Uint16Array {\n    readonly [Symbol.toStringTag]: "Uint16Array";\n}\n\ninterface Int32Array {\n    readonly [Symbol.toStringTag]: "Int32Array";\n}\n\ninterface Uint32Array {\n    readonly [Symbol.toStringTag]: "Uint32Array";\n}\n\ninterface Float32Array {\n    readonly [Symbol.toStringTag]: "Float32Array";\n}\n\ninterface Float64Array {\n    readonly [Symbol.toStringTag]: "Float64Array";\n}\n\ninterface ArrayConstructor {\n    readonly [Symbol.species]: ArrayConstructor;\n}\ninterface MapConstructor {\n    readonly [Symbol.species]: MapConstructor;\n}\ninterface SetConstructor {\n    readonly [Symbol.species]: SetConstructor;\n}\ninterface ArrayBufferConstructor {\n    readonly [Symbol.species]: ArrayBufferConstructor;\n}\n',n.libFileMap["lib.es2016.array.include.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Array<T> {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: T, fromIndex?: number): boolean;\n}\n\ninterface ReadonlyArray<T> {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: T, fromIndex?: number): boolean;\n}\n\ninterface Int8Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Uint8Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Uint8ClampedArray {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Int16Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Uint16Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Int32Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Uint32Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Float32Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}\n\ninterface Float64Array {\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: number, fromIndex?: number): boolean;\n}',n.libFileMap["lib.es2016.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015" />\n/// <reference lib="es2016.array.include" />',n.libFileMap["lib.es2016.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2016" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />',n.libFileMap["lib.es2017.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2016" />\n/// <reference lib="es2017.object" />\n/// <reference lib="es2017.sharedmemory" />\n/// <reference lib="es2017.string" />\n/// <reference lib="es2017.intl" />\n/// <reference lib="es2017.typedarrays" />\n',n.libFileMap["lib.es2017.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2017" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />',n.libFileMap["lib.es2017.intl.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Intl {\n    type DateTimeFormatPartTypes = "day" | "dayPeriod" | "era" | "hour" | "literal" | "minute" | "month" | "second" | "timeZoneName" | "weekday" | "year";\n\n    interface DateTimeFormatPart {\n        type: DateTimeFormatPartTypes;\n        value: string;\n    }\n\n    interface DateTimeFormat {\n        formatToParts(date?: Date | number): DateTimeFormatPart[];\n    }\n}\n',n.libFileMap["lib.es2017.object.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface ObjectConstructor {\n    /**\n     * Returns an array of values of the enumerable properties of an object\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    values<T>(o: { [s: string]: T } | ArrayLike<T>): T[];\n\n    /**\n     * Returns an array of values of the enumerable properties of an object\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    values(o: {}): any[];\n\n    /**\n     * Returns an array of key/values of the enumerable properties of an object\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    entries<T>(o: { [s: string]: T } | ArrayLike<T>): [string, T][];\n\n    /**\n     * Returns an array of key/values of the enumerable properties of an object\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    entries(o: {}): [string, any][];\n\n    /**\n     * Returns an object containing all own property descriptors of an object\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    getOwnPropertyDescriptors<T>(o: T): {[P in keyof T]: TypedPropertyDescriptor<T[P]>} & { [x: string]: PropertyDescriptor };\n}\n',n.libFileMap["lib.es2017.sharedmemory.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.symbol" />\n/// <reference lib="es2015.symbol.wellknown" />\n\ninterface SharedArrayBuffer {\n    /**\n     * Read-only. The length of the ArrayBuffer (in bytes).\n     */\n    readonly byteLength: number;\n\n    /**\n     * Returns a section of an SharedArrayBuffer.\n     */\n    slice(begin: number, end?: number): SharedArrayBuffer;\n    readonly [Symbol.species]: SharedArrayBuffer;\n    readonly [Symbol.toStringTag]: "SharedArrayBuffer";\n}\n\ninterface SharedArrayBufferConstructor {\n    readonly prototype: SharedArrayBuffer;\n    new (byteLength: number): SharedArrayBuffer;\n}\ndeclare var SharedArrayBuffer: SharedArrayBufferConstructor;\n\ninterface ArrayBufferTypes {\n    SharedArrayBuffer: SharedArrayBuffer;\n}\n\ninterface Atomics {\n    /**\n     * Adds a value to the value at the given position in the array, returning the original value.\n     * Until this atomic operation completes, any other read or write operation against the array\n     * will block.\n     */\n    add(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * Stores the bitwise AND of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or\n     * write operation against the array will block.\n     */\n    and(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * Replaces the value at the given position in the array if the original value equals the given\n     * expected value, returning the original value. Until this atomic operation completes, any\n     * other read or write operation against the array will block.\n     */\n    compareExchange(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, expectedValue: number, replacementValue: number): number;\n\n    /**\n     * Replaces the value at the given position in the array, returning the original value. Until\n     * this atomic operation completes, any other read or write operation against the array will\n     * block.\n     */\n    exchange(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * Returns a value indicating whether high-performance algorithms can use atomic operations\n     * (`true`) or must use locks (`false`) for the given number of bytes-per-element of a typed\n     * array.\n     */\n    isLockFree(size: number): boolean;\n\n    /**\n     * Returns the value at the given position in the array. Until this atomic operation completes,\n     * any other read or write operation against the array will block.\n     */\n    load(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number): number;\n\n    /**\n     * Stores the bitwise OR of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or write\n     * operation against the array will block.\n     */\n    or(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * Stores a value at the given position in the array, returning the new value. Until this\n     * atomic operation completes, any other read or write operation against the array will block.\n     */\n    store(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * Subtracts a value from the value at the given position in the array, returning the original\n     * value. Until this atomic operation completes, any other read or write operation against the\n     * array will block.\n     */\n    sub(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    /**\n     * If the value at the given position in the array is equal to the provided value, the current\n     * agent is put to sleep causing execution to suspend until the timeout expires (returning\n     * `"timed-out"`) or until the agent is awoken (returning `"ok"`); otherwise, returns\n     * `"not-equal"`.\n     */\n    wait(typedArray: Int32Array, index: number, value: number, timeout?: number): "ok" | "not-equal" | "timed-out";\n\n    /**\n     * Wakes up sleeping agents that are waiting on the given index of the array, returning the\n     * number of agents that were awoken.\n     * @param typedArray A shared Int32Array.\n     * @param index The position in the typedArray to wake up on.\n     * @param count The number of sleeping agents to notify. Defaults to +Infinity.\n     */\n    notify(typedArray: Int32Array, index: number, count?: number): number;\n\n    /**\n     * Stores the bitwise XOR of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or write\n     * operation against the array will block.\n     */\n    xor(typedArray: Int8Array | Uint8Array | Int16Array | Uint16Array | Int32Array | Uint32Array, index: number, value: number): number;\n\n    readonly [Symbol.toStringTag]: "Atomics";\n}\n\ndeclare var Atomics: Atomics;\n',n.libFileMap["lib.es2017.string.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface String {\n    /**\n     * Pads the current string with a given string (possibly repeated) so that the resulting string reaches a given length.\n     * The padding is applied from the start (left) of the current string.\n     *\n     * @param maxLength The length of the resulting string once the current string has been padded.\n     *        If this parameter is smaller than the current string\'s length, the current string will be returned as it is.\n     *\n     * @param fillString The string to pad the current string with.\n     *        If this string is too long, it will be truncated and the left-most part will be applied.\n     *        The default value for this parameter is " " (U+0020).\n     */\n    padStart(maxLength: number, fillString?: string): string;\n\n    /**\n     * Pads the current string with a given string (possibly repeated) so that the resulting string reaches a given length.\n     * The padding is applied from the end (right) of the current string.\n     *\n     * @param maxLength The length of the resulting string once the current string has been padded.\n     *        If this parameter is smaller than the current string\'s length, the current string will be returned as it is.\n     *\n     * @param fillString The string to pad the current string with.\n     *        If this string is too long, it will be truncated and the left-most part will be applied.\n     *        The default value for this parameter is " " (U+0020).\n     */\n    padEnd(maxLength: number, fillString?: string): string;\n}\n',n.libFileMap["lib.es2017.typedarrays.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Int8ArrayConstructor {\n    new (): Int8Array;\n}\n\ninterface Uint8ArrayConstructor {\n    new (): Uint8Array;\n}\n\ninterface Uint8ClampedArrayConstructor {\n    new (): Uint8ClampedArray;\n}\n\ninterface Int16ArrayConstructor {\n    new (): Int16Array;\n}\n\ninterface Uint16ArrayConstructor {\n    new (): Uint16Array;\n}\n\ninterface Int32ArrayConstructor {\n    new (): Int32Array;\n}\n\ninterface Uint32ArrayConstructor {\n    new (): Uint32Array;\n}\n\ninterface Float32ArrayConstructor {\n    new (): Float32Array;\n}\n\ninterface Float64ArrayConstructor {\n    new (): Float64Array;\n}\n',n.libFileMap["lib.es2018.asyncgenerator.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2018.asynciterable" />\n\ninterface AsyncGenerator<T = unknown, TReturn = any, TNext = unknown> extends AsyncIterator<T, TReturn, TNext> {\n    // NOTE: \'next\' is defined using a tuple to ensure we report the correct assignability errors in all places.\n    next(...args: [] | [TNext]): Promise<IteratorResult<T, TReturn>>;\n    return(value: TReturn | PromiseLike<TReturn>): Promise<IteratorResult<T, TReturn>>;\n    throw(e: any): Promise<IteratorResult<T, TReturn>>;\n    [Symbol.asyncIterator](): AsyncGenerator<T, TReturn, TNext>;\n}\n\ninterface AsyncGeneratorFunction {\n    /**\n     * Creates a new AsyncGenerator object.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: any[]): AsyncGenerator;\n    /**\n     * Creates a new AsyncGenerator object.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: any[]): AsyncGenerator;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: AsyncGenerator;\n}\n\ninterface AsyncGeneratorFunctionConstructor {\n    /**\n     * Creates a new AsyncGenerator function.\n     * @param args A list of arguments the function accepts.\n     */\n    new (...args: string[]): AsyncGeneratorFunction;\n    /**\n     * Creates a new AsyncGenerator function.\n     * @param args A list of arguments the function accepts.\n     */\n    (...args: string[]): AsyncGeneratorFunction;\n    /**\n     * The length of the arguments.\n     */\n    readonly length: number;\n    /**\n     * Returns the name of the function.\n     */\n    readonly name: string;\n    /**\n     * A reference to the prototype.\n     */\n    readonly prototype: AsyncGeneratorFunction;\n}\n',n.libFileMap["lib.es2018.asynciterable.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.symbol" />\n/// <reference lib="es2015.iterable" />\n\ninterface SymbolConstructor {\n    /**\n     * A method that returns the default async iterator for an object. Called by the semantics of\n     * the for-await-of statement.\n     */\n    readonly asyncIterator: symbol;\n}\n\ninterface AsyncIterator<T, TReturn = any, TNext = undefined> {\n    // NOTE: \'next\' is defined using a tuple to ensure we report the correct assignability errors in all places.\n    next(...args: [] | [TNext]): Promise<IteratorResult<T, TReturn>>;\n    return?(value?: TReturn | PromiseLike<TReturn>): Promise<IteratorResult<T, TReturn>>;\n    throw?(e?: any): Promise<IteratorResult<T, TReturn>>;\n}\n\ninterface AsyncIterable<T> {\n    [Symbol.asyncIterator](): AsyncIterator<T>;\n}\n\ninterface AsyncIterableIterator<T> extends AsyncIterator<T> {\n    [Symbol.asyncIterator](): AsyncIterableIterator<T>;\n}',n.libFileMap["lib.es2018.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2017" />\n/// <reference lib="es2018.asynciterable" />\n/// <reference lib="es2018.asyncgenerator" />\n/// <reference lib="es2018.promise" />\n/// <reference lib="es2018.regexp" />\n/// <reference lib="es2018.intl" />\n',n.libFileMap["lib.es2018.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2018" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />',n.libFileMap["lib.es2018.intl.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Intl {\n\n    // http://cldr.unicode.org/index/cldr-spec/plural-rules#TOC-Determining-Plural-Categories\n    type LDMLPluralRule = "zero" | "one" | "two" | "few" | "many" | "other";\n    type PluralRuleType = "cardinal" | "ordinal";\n\n    interface PluralRulesOptions {\n        localeMatcher?: "lookup" | "best fit";\n        type?: PluralRuleType;\n        minimumIntegerDigits?: number;\n        minimumFractionDigits?: number;\n        maximumFractionDigits?: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n    }\n\n    interface ResolvedPluralRulesOptions {\n        locale: string;\n        pluralCategories: LDMLPluralRule[];\n        type: PluralRuleType;\n        minimumIntegerDigits: number;\n        minimumFractionDigits: number;\n        maximumFractionDigits: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n    }\n\n    interface PluralRules {\n        resolvedOptions(): ResolvedPluralRulesOptions;\n        select(n: number): LDMLPluralRule;\n    }\n\n    const PluralRules: {\n        new (locales?: string | string[], options?: PluralRulesOptions): PluralRules;\n        (locales?: string | string[], options?: PluralRulesOptions): PluralRules;\n        supportedLocalesOf(\n            locales: string | string[],\n            options?: PluralRulesOptions,\n        ): string[];\n    };\n}\n',n.libFileMap["lib.es2018.promise.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/**\n * Represents the completion of an asynchronous operation\n */\ninterface Promise<T> {\n    /**\n     * Attaches a callback that is invoked when the Promise is settled (fulfilled or rejected). The\n     * resolved value cannot be modified from the callback.\n     * @param onfinally The callback to execute when the Promise is settled (fulfilled or rejected).\n     * @returns A Promise for the completion of the callback.\n     */\n    finally(onfinally?: (() => void) | undefined | null): Promise<T>\n}\n',n.libFileMap["lib.es2018.regexp.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface RegExpMatchArray {\n    groups?: {\n        [key: string]: string\n    }\n}\n\ninterface RegExpExecArray {\n    groups?: {\n        [key: string]: string\n    }\n}\n\ninterface RegExp {\n    /**\n     * Returns a Boolean value indicating the state of the dotAll flag (s) used with a regular expression.\n     * Default is false. Read-only.\n     */\n    readonly dotAll: boolean;\n}',n.libFileMap["lib.es2019.array.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ntype FlatArray<Arr, Depth extends number> = {\n    "done": Arr,\n    "recur": Arr extends ReadonlyArray<infer InnerArr>\n        ? FlatArray<InnerArr, [-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20][Depth]>\n        : Arr\n}[Depth extends -1 ? "done" : "recur"];\n\ninterface ReadonlyArray<T> {\n\n    /**\n     * Calls a defined callback function on each element of an array. Then, flattens the result into\n     * a new array.\n     * This is identical to a map followed by flat with depth 1.\n     *\n     * @param callback A function that accepts up to three arguments. The flatMap method calls the\n     * callback function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callback function. If\n     * thisArg is omitted, undefined is used as the this value.\n     */\n    flatMap<U, This = undefined> (\n        callback: (this: This, value: T, index: number, array: T[]) => U | ReadonlyArray<U>,\n        thisArg?: This\n    ): U[]\n\n\n    /**\n     * Returns a new array with all sub-array elements concatenated into it recursively up to the\n     * specified depth.\n     *\n     * @param depth The maximum recursion depth\n     */\n    flat<A, D extends number = 1>(\n        this: A,\n        depth?: D\n    ): FlatArray<A, D>[]\n  }\n\ninterface Array<T> {\n\n    /**\n     * Calls a defined callback function on each element of an array. Then, flattens the result into\n     * a new array.\n     * This is identical to a map followed by flat with depth 1.\n     *\n     * @param callback A function that accepts up to three arguments. The flatMap method calls the\n     * callback function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callback function. If\n     * thisArg is omitted, undefined is used as the this value.\n     */\n    flatMap<U, This = undefined> (\n        callback: (this: This, value: T, index: number, array: T[]) => U | ReadonlyArray<U>,\n        thisArg?: This\n    ): U[]\n\n    /**\n     * Returns a new array with all sub-array elements concatenated into it recursively up to the\n     * specified depth.\n     *\n     * @param depth The maximum recursion depth\n     */\n    flat<A, D extends number = 1>(\n        this: A,\n        depth?: D\n    ): FlatArray<A, D>[]\n}\n',n.libFileMap["lib.es2019.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2018" />\n/// <reference lib="es2019.array" />\n/// <reference lib="es2019.object" />\n/// <reference lib="es2019.string" />\n/// <reference lib="es2019.symbol" />\n',n.libFileMap["lib.es2019.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2019" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />\n',n.libFileMap["lib.es2019.object.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.iterable" />\n\ninterface ObjectConstructor {\n    /**\n     * Returns an object created by key-value entries for properties and methods\n     * @param entries An iterable object that contains key-value entries for properties and methods.\n     */\n    fromEntries<T = any>(entries: Iterable<readonly [PropertyKey, T]>): { [k: string]: T };\n\n    /**\n     * Returns an object created by key-value entries for properties and methods\n     * @param entries An iterable object that contains key-value entries for properties and methods.\n     */\n    fromEntries(entries: Iterable<readonly any[]>): any;\n}\n',n.libFileMap["lib.es2019.string.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface String {\n    /** Removes the trailing white space and line terminator characters from a string. */\n    trimEnd(): string;\n\n    /** Removes the leading white space and line terminator characters from a string. */\n    trimStart(): string;\n\n    /** Removes the leading white space and line terminator characters from a string. */\n    trimLeft(): string;\n\n    /** Removes the trailing white space and line terminator characters from a string. */\n    trimRight(): string;\n}\n',n.libFileMap["lib.es2019.symbol.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Symbol {\n    /**\n     * Expose the [[Description]] internal slot of a symbol directly.\n     */\n    readonly description: string | undefined;\n}\n',n.libFileMap["lib.es2020.bigint.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface BigIntToLocaleStringOptions {\n    /**\n     * The locale matching algorithm to use.The default is "best fit". For information about this option, see the {@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_negotiation Intl page}.\n     */\n    localeMatcher?: string;\n    /**\n     * The formatting style to use , the default is "decimal".\n     */\n    style?: string;\n\n    numberingSystem?: string;\n    /**\n     * The unit to use in unit formatting, Possible values are core unit identifiers, defined in UTS #35, Part 2, Section 6. A subset of units from the full list was selected for use in ECMAScript. Pairs of simple units can be concatenated with "-per-" to make a compound unit. There is no default value; if the style is "unit", the unit property must be provided.\n     */\n    unit?: string;\n\n    /**\n     * The unit formatting style to use in unit formatting, the defaults is "short".\n     */\n    unitDisplay?: string;\n\n    /**\n     * The currency to use in currency formatting. Possible values are the ISO 4217 currency codes, such as "USD" for the US dollar, "EUR" for the euro, or "CNY" for the Chinese RMB — see the Current currency & funds code list. There is no default value; if the style is "currency", the currency property must be provided. It is only used when [[Style]] has the value "currency".\n     */\n    currency?: string;\n\n    /**\n     * How to display the currency in currency formatting. It is only used when [[Style]] has the value "currency". The default is "symbol".\n     *\n     * "symbol" to use a localized currency symbol such as €,\n     *\n     * "code" to use the ISO currency code,\n     *\n     * "name" to use a localized currency name such as "dollar"\n     */\n    currencyDisplay?: string;\n\n    /**\n     * Whether to use grouping separators, such as thousands separators or thousand/lakh/crore separators. The default is true.\n     */\n    useGrouping?: boolean;\n\n    /**\n     * The minimum number of integer digits to use. Possible values are from 1 to 21; the default is 1.\n     */\n    minimumIntegerDigits?: 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 | 21;\n\n    /**\n     * The minimum number of fraction digits to use. Possible values are from 0 to 20; the default for plain number and percent formatting is 0; the default for currency formatting is the number of minor unit digits provided by the {@link http://www.currency-iso.org/en/home/tables/table-a1.html ISO 4217 currency codes list} (2 if the list doesn\'t provide that information).\n     */\n    minimumFractionDigits?: 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20;\n\n    /**\n     * The maximum number of fraction digits to use. Possible values are from 0 to 20; the default for plain number formatting is the larger of minimumFractionDigits and 3; the default for currency formatting is the larger of minimumFractionDigits and the number of minor unit digits provided by the {@link http://www.currency-iso.org/en/home/tables/table-a1.html ISO 4217 currency codes list} (2 if the list doesn\'t provide that information); the default for percent formatting is the larger of minimumFractionDigits and 0.\n     */\n    maximumFractionDigits?: 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20;\n\n    /**\n     * The minimum number of significant digits to use. Possible values are from 1 to 21; the default is 1.\n     */\n    minimumSignificantDigits?: 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 | 21;\n\n    /**\n     * The maximum number of significant digits to use. Possible values are from 1 to 21; the default is 21.\n     */\n    maximumSignificantDigits?: 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 | 21;\n\n    /**\n     * The formatting that should be displayed for the number, the defaults is "standard"\n     *\n     *     "standard" plain number formatting\n     *\n     *     "scientific" return the order-of-magnitude for formatted number.\n     *\n     *     "engineering" return the exponent of ten when divisible by three\n     *\n     *     "compact" string representing exponent, defaults is using the "short" form\n     */\n    notation?: string;\n\n    /**\n     * used only when notation is "compact"\n     */\n    compactDisplay?: string;\n}\n\ninterface BigInt {\n    /**\n     * Returns a string representation of an object.\n     * @param radix Specifies a radix for converting numeric values to strings.\n     */\n    toString(radix?: number): string;\n\n    /** Returns a string representation appropriate to the host environment\'s current locale. */\n    toLocaleString(locales?: string, options?: BigIntToLocaleStringOptions): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): bigint;\n\n    readonly [Symbol.toStringTag]: "BigInt";\n}\n\ninterface BigIntConstructor {\n    (value?: any): bigint;\n    readonly prototype: BigInt;\n\n    /**\n     * Interprets the low bits of a BigInt as a 2\'s-complement signed integer.\n     * All higher bits are discarded.\n     * @param bits The number of low bits to use\n     * @param int The BigInt whose bits to extract\n     */\n    asIntN(bits: number, int: bigint): bigint;\n    /**\n     * Interprets the low bits of a BigInt as an unsigned integer.\n     * All higher bits are discarded.\n     * @param bits The number of low bits to use\n     * @param int The BigInt whose bits to extract\n     */\n    asUintN(bits: number, int: bigint): bigint;\n}\n\ndeclare var BigInt: BigIntConstructor;\n\n/**\n * A typed array of 64-bit signed integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated, an exception is raised.\n */\ninterface BigInt64Array {\n    /** The size in bytes of each element in the array. */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /** The ArrayBuffer instance referenced by the array. */\n    readonly buffer: ArrayBufferLike;\n\n    /** The length in bytes of the array. */\n    readonly byteLength: number;\n\n    /** The offset in bytes of the array. */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /** Yields index, value pairs for every entry in the array. */\n    entries(): IterableIterator<[number, bigint]>;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns false,\n     * or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: bigint, index: number, array: BigInt64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: bigint, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: bigint, index: number, array: BigInt64Array) => any, thisArg?: any): BigInt64Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: bigint, index: number, array: BigInt64Array) => boolean, thisArg?: any): bigint | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: bigint, index: number, array: BigInt64Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: bigint, index: number, array: BigInt64Array) => void, thisArg?: any): void;\n\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: bigint, fromIndex?: number): boolean;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    indexOf(searchElement: bigint, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /** Yields each index in the array. */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: bigint, fromIndex?: number): number;\n\n    /** The length of the array. */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: bigint, index: number, array: BigInt64Array) => bigint, thisArg?: any): BigInt64Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: bigint, currentValue: bigint, currentIndex: number, array: BigInt64Array) => bigint): bigint;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: bigint, currentIndex: number, array: BigInt64Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: bigint, currentValue: bigint, currentIndex: number, array: BigInt64Array) => bigint): bigint;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: bigint, currentIndex: number, array: BigInt64Array) => U, initialValue: U): U;\n\n    /** Reverses the elements in the array. */\n    reverse(): this;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<bigint>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array.\n     */\n    slice(start?: number, end?: number): BigInt64Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls the\n     * predicate function for each element in the array until the predicate returns true, or until\n     * the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: bigint, index: number, array: BigInt64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n     * Sorts the array.\n     * @param compareFn The function used to determine the order of the elements. If omitted, the elements are sorted in ascending order.\n     */\n    sort(compareFn?: (a: bigint, b: bigint) => number | bigint): this;\n\n    /**\n     * Gets a new BigInt64Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): BigInt64Array;\n\n    /** Converts the array to a string by using the current locale. */\n    toLocaleString(): string;\n\n    /** Returns a string representation of the array. */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): BigInt64Array;\n\n    /** Yields each value in the array. */\n    values(): IterableIterator<bigint>;\n\n    [Symbol.iterator](): IterableIterator<bigint>;\n\n    readonly [Symbol.toStringTag]: "BigInt64Array";\n\n    [index: number]: bigint;\n}\n\ninterface BigInt64ArrayConstructor {\n    readonly prototype: BigInt64Array;\n    new(length?: number): BigInt64Array;\n    new(array: Iterable<bigint>): BigInt64Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): BigInt64Array;\n\n    /** The size in bytes of each element in the array. */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: bigint[]): BigInt64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of \'this\' used to invoke the mapfn.\n     */\n    from(arrayLike: ArrayLike<bigint>): BigInt64Array;\n    from<U>(arrayLike: ArrayLike<U>, mapfn: (v: U, k: number) => bigint, thisArg?: any): BigInt64Array;\n}\n\ndeclare var BigInt64Array: BigInt64ArrayConstructor;\n\n/**\n * A typed array of 64-bit unsigned integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated, an exception is raised.\n */\ninterface BigUint64Array {\n    /** The size in bytes of each element in the array. */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /** The ArrayBuffer instance referenced by the array. */\n    readonly buffer: ArrayBufferLike;\n\n    /** The length in bytes of the array. */\n    readonly byteLength: number;\n\n    /** The offset in bytes of the array. */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /** Yields index, value pairs for every entry in the array. */\n    entries(): IterableIterator<[number, bigint]>;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns false,\n     * or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: bigint, index: number, array: BigUint64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: bigint, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: bigint, index: number, array: BigUint64Array) => any, thisArg?: any): BigUint64Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: bigint, index: number, array: BigUint64Array) => boolean, thisArg?: any): bigint | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: bigint, index: number, array: BigUint64Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: bigint, index: number, array: BigUint64Array) => void, thisArg?: any): void;\n\n    /**\n     * Determines whether an array includes a certain element, returning true or false as appropriate.\n     * @param searchElement The element to search for.\n     * @param fromIndex The position in this array at which to begin searching for searchElement.\n     */\n    includes(searchElement: bigint, fromIndex?: number): boolean;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    indexOf(searchElement: bigint, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /** Yields each index in the array. */\n    keys(): IterableIterator<number>;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: bigint, fromIndex?: number): number;\n\n    /** The length of the array. */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: bigint, index: number, array: BigUint64Array) => bigint, thisArg?: any): BigUint64Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: bigint, currentValue: bigint, currentIndex: number, array: BigUint64Array) => bigint): bigint;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: bigint, currentIndex: number, array: BigUint64Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: bigint, currentValue: bigint, currentIndex: number, array: BigUint64Array) => bigint): bigint;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: bigint, currentIndex: number, array: BigUint64Array) => U, initialValue: U): U;\n\n    /** Reverses the elements in the array. */\n    reverse(): this;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<bigint>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array.\n     */\n    slice(start?: number, end?: number): BigUint64Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls the\n     * predicate function for each element in the array until the predicate returns true, or until\n     * the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: bigint, index: number, array: BigUint64Array) => boolean, thisArg?: any): boolean;\n\n    /**\n     * Sorts the array.\n     * @param compareFn The function used to determine the order of the elements. If omitted, the elements are sorted in ascending order.\n     */\n    sort(compareFn?: (a: bigint, b: bigint) => number | bigint): this;\n\n    /**\n     * Gets a new BigUint64Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): BigUint64Array;\n\n    /** Converts the array to a string by using the current locale. */\n    toLocaleString(): string;\n\n    /** Returns a string representation of the array. */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): BigUint64Array;\n\n    /** Yields each value in the array. */\n    values(): IterableIterator<bigint>;\n\n    [Symbol.iterator](): IterableIterator<bigint>;\n\n    readonly [Symbol.toStringTag]: "BigUint64Array";\n\n    [index: number]: bigint;\n}\n\ninterface BigUint64ArrayConstructor {\n    readonly prototype: BigUint64Array;\n    new(length?: number): BigUint64Array;\n    new(array: Iterable<bigint>): BigUint64Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): BigUint64Array;\n\n    /** The size in bytes of each element in the array. */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: bigint[]): BigUint64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of \'this\' used to invoke the mapfn.\n     */\n    from(arrayLike: ArrayLike<bigint>): BigUint64Array;\n    from<U>(arrayLike: ArrayLike<U>, mapfn: (v: U, k: number) => bigint, thisArg?: any): BigUint64Array;\n}\n\ndeclare var BigUint64Array: BigUint64ArrayConstructor;\n\ninterface DataView {\n    /**\n     * Gets the BigInt64 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getBigInt64(byteOffset: number, littleEndian?: boolean): bigint;\n\n    /**\n     * Gets the BigUint64 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getBigUint64(byteOffset: number, littleEndian?: boolean): bigint;\n\n    /**\n     * Stores a BigInt64 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setBigInt64(byteOffset: number, value: bigint, littleEndian?: boolean): void;\n\n    /**\n     * Stores a BigUint64 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setBigUint64(byteOffset: number, value: bigint, littleEndian?: boolean): void;\n}\n\ndeclare namespace Intl{\n    interface NumberFormat {\n        format(value: number | bigint): string;\n        resolvedOptions(): ResolvedNumberFormatOptions;\n    }\n}\n',n.libFileMap["lib.es2020.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2019" />\n/// <reference lib="es2020.bigint" />\n/// <reference lib="es2020.promise" />\n/// <reference lib="es2020.sharedmemory" />\n/// <reference lib="es2020.string" />\n/// <reference lib="es2020.symbol.wellknown" />\n/// <reference lib="es2020.intl" />\n',n.libFileMap["lib.es2020.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2020" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />\n',n.libFileMap["lib.es2020.intl.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Intl {\n\n    /**\n     * [BCP 47 language tag](http://tools.ietf.org/html/rfc5646) definition.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl#locales_argument).\n     *\n     * [Wikipedia](https://en.wikipedia.org/wiki/IETF_language_tag).\n     */\n    type BCP47LanguageTag = string;\n\n    /**\n     * Unit to use in the relative time internationalized message.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/format#Parameters).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-singularrelativetimeunit).\n     */\n    type RelativeTimeFormatUnit =\n        | "year" | "years"\n        | "quarter" | "quarters"\n        | "month" | "months"\n        | "week" | "weeks"\n        | "day" | "days"\n        | "hour" | "hours"\n        | "minute" | "minutes"\n        | "second" | "seconds"\n        ;\n\n    /**\n     * The locale matching algorithm to use.\n     *\n     * [MDN](https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_negotiation).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-InitializeRelativeTimeFormat).\n     */\n    type RelativeTimeFormatLocaleMatcher = "lookup" | "best fit";\n\n    /**\n     * The format of output message.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat#Parameters).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-InitializeRelativeTimeFormat).\n     */\n    type RelativeTimeFormatNumeric = "always" | "auto";\n\n    /**\n     * The length of the internationalized message.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat#Parameters).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-InitializeRelativeTimeFormat).\n     */\n    type RelativeTimeFormatStyle = "long" | "short" | "narrow";\n\n    /**\n     * An object with some or all of properties of `options` parameter\n     * of `Intl.RelativeTimeFormat` constructor.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat#Parameters).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-InitializeRelativeTimeFormat).\n     */\n    interface RelativeTimeFormatOptions {\n        localeMatcher?: RelativeTimeFormatLocaleMatcher;\n        numeric?: RelativeTimeFormatNumeric;\n        style?: RelativeTimeFormatStyle;\n    }\n\n    /**\n     * An object with properties reflecting the locale\n     * and formatting options computed during initialization\n     * of the `Intel.RelativeTimeFormat` object\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/resolvedOptions#Description).\n     *\n     * [Specification](https://tc39.es/ecma402/#table-relativetimeformat-resolvedoptions-properties)\n     */\n    interface ResolvedRelativeTimeFormatOptions {\n        locale: BCP47LanguageTag;\n        style: RelativeTimeFormatStyle;\n        numeric: RelativeTimeFormatNumeric;\n        numberingSystem: string;\n    }\n\n    /**\n     * An object representing the relative time format in parts\n     * that can be used for custom locale-aware formatting.\n     *\n     * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/formatToParts#Using_formatToParts).\n     *\n     * [Specification](https://tc39.es/ecma402/#sec-FormatRelativeTimeToParts).\n     */\n    interface RelativeTimeFormatPart {\n        type: string;\n        value: string;\n        unit?: RelativeTimeFormatUnit;\n    }\n\n    interface RelativeTimeFormat {\n        /**\n         * Formats a value and a unit according to the locale\n         * and formatting options of the given\n         * [`Intl.RelativeTimeFormat`](https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/RelativeTimeFormat)\n         * object.\n         *\n         * While this method automatically provides the correct plural forms,\n         * the grammatical form is otherwise as neutral as possible.\n         * It is the caller\'s responsibility to handle cut-off logic\n         * such as deciding between displaying "in 7 days" or "in 1 week".\n         * This API does not support relative dates involving compound units.\n         * e.g "in 5 days and 4 hours".\n         *\n         * @param value -  Numeric value to use in the internationalized relative time message\n         *\n         * @param unit - [Unit](https://tc39.es/ecma402/#sec-singularrelativetimeunit)\n         *  to use in the relative time internationalized message.\n         *  Possible values are: `"year"`, `"quarter"`, `"month"`, `"week"`,\n         *  `"day"`, `"hour"`, `"minute"`, `"second"`.\n         *  Plural forms are also permitted.\n         *\n         * @throws `RangeError` if `unit` was given something other than `unit` possible values\n         *\n         * @returns Internationalized relative time message as string\n         *\n         * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/format).\n         *\n         * [Specification](https://tc39.es/ecma402/#sec-Intl.RelativeTimeFormat.prototype.format).\n         */\n        format(\n            value: number,\n            unit: RelativeTimeFormatUnit,\n        ): string;\n\n        /**\n         *  A version of the format method which it returns an array of objects\n         *  which represent "parts" of the object,\n         *  separating the formatted number into its constituent parts\n         *  and separating it from other surrounding text.\n         *  These objects have two properties:\n         * `type` a NumberFormat formatToParts type, and `value`,\n         *  which is the String which is the component of the output.\n         *  If a "part" came from NumberFormat,\n         *  it will have a unit property which indicates the `unit` being formatted;\n         *  literals which are part of the larger frame will not have this property.\n         *\n         *  @param value - Numeric value to use in the internationalized relative time message\n         *\n         *  @param unit - [Unit](https://tc39.es/ecma402/#sec-singularrelativetimeunit)\n         *   to use in the relative time internationalized message.\n         *   Possible values are: `"year"`, `"quarter"`, `"month"`, `"week"`,\n         *   `"day"`, `"hour"`, `"minute"`, `"second"`.\n         *   Plural forms are also permitted.\n         *\n         *  @throws `RangeError` if `unit` was given something other than `unit` possible values\n         *\n         *  @returns Array of [FormatRelativeTimeToParts](https://tc39.es/ecma402/#sec-FormatRelativeTimeToParts)\n         *\n         *  [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/formatToParts).\n         *\n         *  [Specification](https://tc39.es/ecma402/#sec-Intl.RelativeTimeFormat.prototype.formatToParts).\n         */\n        formatToParts(\n            value: number,\n            unit: RelativeTimeFormatUnit,\n        ): RelativeTimeFormatPart[];\n\n        /**\n         * Provides access to the locale and options computed during initialization of this `Intl.RelativeTimeFormat` object.\n         *\n         * @returns A new object with properties reflecting the locale\n         *  and formatting options computed during initialization\n         *  of the `Intel.RelativeTimeFormat` object.\n         *\n         * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/resolvedOptions).\n         *\n         * [Specification](https://tc39.es/ecma402/#sec-intl.relativetimeformat.prototype.resolvedoptions)\n         */\n        resolvedOptions(): ResolvedRelativeTimeFormatOptions;\n    }\n\n    /**\n     * The [`Intl.RelativeTimeFormat`](https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/RelativeTimeFormat)\n     * object is a constructor for objects that enable language-sensitive relative time formatting.\n     *\n     * Part of [Intl object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl)\n     * namespace and the [ECMAScript Internationalization API](https://www.ecma-international.org/publications/standards/Ecma-402.htm).\n     *\n     * [Compatibility](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat#Browser_compatibility).\n     *\n     * [Polyfills](https://github.com/tc39/proposal-intl-relative-time#polyfills).\n     */\n    const RelativeTimeFormat: {\n        /**\n         * Constructor creates [Intl.RelativeTimeFormat](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RelativeTimeFormat)\n         * objects\n         *\n         * @param locales - A string with a [BCP 47 language tag](http://tools.ietf.org/html/rfc5646), or an array of such strings.\n         *  For the general form and interpretation of the locales argument,\n         *  see the [`Intl` page](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_identification_and_negotiation).\n         *\n         * @param options - An [object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat#Parameters)\n         *  with some or all of options of the formatting.\n         *  An object with some or all of the following properties:\n         *  - `localeMatcher` - The locale matching algorithm to use.\n         *    Possible values are `"lookup"` and `"best fit"`; the default is `"best fit"`.\n         *    For information about this option, see [Intl page](https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_negotiation).\n         *  - `numeric` - The format of output message.\n         *    Possible values are: `"always"` (default, e.g., `1 day ago`) or `"auto"` (e.g., `yesterday`).\n         *    The `"auto"` value allows to not always have to use numeric values in the output.\n         *  - `style` - The length of the internationalized message. Possible values are:\n         *    `"long"` (default, e.g., in 1 month),\n         *    `"short"` (e.g., in 1 mo.)\n         *    or `"narrow"` (e.g., in 1 mo.). The narrow style could be similar to the short style for some locales.\n         *\n         * @returns [Intl.RelativeTimeFormat](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RelativeTimeFormat) object.\n         *\n         * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat).\n         *\n         * [Specification](https://tc39.es/ecma402/#sec-intl-relativetimeformat-constructor).\n         */\n        new(\n            locales?: BCP47LanguageTag | BCP47LanguageTag[],\n            options?: RelativeTimeFormatOptions,\n        ): RelativeTimeFormat;\n\n        /**\n         * Returns an array containing those of the provided locales\n         * that are supported in date and time formatting\n         * without having to fall back to the runtime\'s default locale.\n         *\n         * @param locales - A string with a [BCP 47 language tag](http://tools.ietf.org/html/rfc5646), or an array of such strings.\n         *  For the general form and interpretation of the locales argument,\n         *  see the [`Intl` page](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_identification_and_negotiation).\n         *\n         * @param options - An [object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/RelativeTimeFormat#Parameters)\n         *  with some or all of options of the formatting.\n         *  An object with some or all of the following properties:\n         *  - `localeMatcher` - The locale matching algorithm to use.\n         *    Possible values are `"lookup"` and `"best fit"`; the default is `"best fit"`.\n         *    For information about this option, see [Intl page](https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Intl#Locale_negotiation).\n         *  - `numeric` - The format of output message.\n         *    Possible values are: `"always"` (default, e.g., `1 day ago`) or `"auto"` (e.g., `yesterday`).\n         *    The `"auto"` value allows to not always have to use numeric values in the output.\n         *  - `style` - The length of the internationalized message. Possible values are:\n         *    `"long"` (default, e.g., in 1 month),\n         *    `"short"` (e.g., in 1 mo.)\n         *    or `"narrow"` (e.g., in 1 mo.). The narrow style could be similar to the short style for some locales.\n         *\n         * @returns An array containing those of the provided locales\n         *  that are supported in date and time formatting\n         *  without having to fall back to the runtime\'s default locale.\n         *\n         * [MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/RelativeTimeFormat/supportedLocalesOf).\n         *\n         * [Specification](https://tc39.es/ecma402/#sec-Intl.RelativeTimeFormat.supportedLocalesOf).\n         */\n        supportedLocalesOf(\n            locales: BCP47LanguageTag | BCP47LanguageTag[],\n            options?: RelativeTimeFormatOptions,\n        ): BCP47LanguageTag[];\n    };\n\n    interface NumberFormatOptions {\n        compactDisplay?: string;\n        notation?: string;\n        signDisplay?: string;\n        unit?: string;\n        unitDisplay?: string;\n    }\n\n    interface ResolvedNumberFormatOptions {\n        compactDisplay?: string;\n        notation?: string;\n        signDisplay?: string;\n        unit?: string;\n        unitDisplay?: string;\n    }\n}\n',n.libFileMap["lib.es2020.promise.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface PromiseFulfilledResult<T> {\n    status: "fulfilled";\n    value: T;\n}\n\ninterface PromiseRejectedResult {\n    status: "rejected";\n    reason: any;\n}\n\ntype PromiseSettledResult<T> = PromiseFulfilledResult<T> | PromiseRejectedResult;\n\ninterface PromiseConstructor {\n    /**\n     * Creates a Promise that is resolved with an array of results when all\n     * of the provided Promises resolve or reject.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    allSettled<T extends readonly unknown[] | readonly [unknown]>(values: T):\n        Promise<{ -readonly [P in keyof T]: PromiseSettledResult<T[P] extends PromiseLike<infer U> ? U : T[P]> }>;\n\n    /**\n     * Creates a Promise that is resolved with an array of results when all\n     * of the provided Promises resolve or reject.\n     * @param values An array of Promises.\n     * @returns A new Promise.\n     */\n    allSettled<T>(values: Iterable<T>): Promise<PromiseSettledResult<T extends PromiseLike<infer U> ? U : T>[]>;\n}\n',n.libFileMap["lib.es2020.sharedmemory.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface Atomics {\n    /**\n     * Adds a value to the value at the given position in the array, returning the original value.\n     * Until this atomic operation completes, any other read or write operation against the array\n     * will block.\n     */\n    add(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * Stores the bitwise AND of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or\n     * write operation against the array will block.\n     */\n    and(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * Replaces the value at the given position in the array if the original value equals the given\n     * expected value, returning the original value. Until this atomic operation completes, any\n     * other read or write operation against the array will block.\n     */\n    compareExchange(typedArray: BigInt64Array | BigUint64Array, index: number, expectedValue: bigint, replacementValue: bigint): bigint;\n\n    /**\n     * Replaces the value at the given position in the array, returning the original value. Until\n     * this atomic operation completes, any other read or write operation against the array will\n     * block.\n     */\n    exchange(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * Returns the value at the given position in the array. Until this atomic operation completes,\n     * any other read or write operation against the array will block.\n     */\n    load(typedArray: BigInt64Array | BigUint64Array, index: number): bigint;\n\n    /**\n     * Stores the bitwise OR of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or write\n     * operation against the array will block.\n     */\n    or(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * Stores a value at the given position in the array, returning the new value. Until this\n     * atomic operation completes, any other read or write operation against the array will block.\n     */\n    store(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * Subtracts a value from the value at the given position in the array, returning the original\n     * value. Until this atomic operation completes, any other read or write operation against the\n     * array will block.\n     */\n    sub(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n\n    /**\n     * If the value at the given position in the array is equal to the provided value, the current\n     * agent is put to sleep causing execution to suspend until the timeout expires (returning\n     * `"timed-out"`) or until the agent is awoken (returning `"ok"`); otherwise, returns\n     * `"not-equal"`.\n     */\n    wait(typedArray: BigInt64Array, index: number, value: bigint, timeout?: number): "ok" | "not-equal" | "timed-out";\n\n    /**\n     * Wakes up sleeping agents that are waiting on the given index of the array, returning the\n     * number of agents that were awoken.\n     * @param typedArray A shared BigInt64Array.\n     * @param index The position in the typedArray to wake up on.\n     * @param count The number of sleeping agents to notify. Defaults to +Infinity.\n     */\n    notify(typedArray: BigInt64Array, index: number, count?: number): number;\n\n    /**\n     * Stores the bitwise XOR of a value with the value at the given position in the array,\n     * returning the original value. Until this atomic operation completes, any other read or write\n     * operation against the array will block.\n     */\n    xor(typedArray: BigInt64Array | BigUint64Array, index: number, value: bigint): bigint;\n}\n',n.libFileMap["lib.es2020.string.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.iterable" />\n\ninterface String {\n    /**\n     * Matches a string with a regular expression, and returns an iterable of matches\n     * containing the results of that search.\n     * @param regexp A variable name or string literal containing the regular expression pattern and flags.\n     */\n    matchAll(regexp: RegExp): IterableIterator<RegExpMatchArray>;\n}\n',n.libFileMap["lib.es2020.symbol.wellknown.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015.iterable" />\n/// <reference lib="es2015.symbol" />\n\ninterface SymbolConstructor {\n    /**\n     * A regular expression method that matches the regular expression against a string. Called\n     * by the String.prototype.matchAll method.\n     */\n    readonly matchAll: symbol;\n}\n\ninterface RegExp {\n    /**\n     * Matches a string with this regular expression, and returns an iterable of matches\n     * containing the results of that search.\n     * @param string A string to search within.\n     */\n    [Symbol.matchAll](str: string): IterableIterator<RegExpMatchArray>;\n}\n',n.libFileMap["lib.es5.d.ts"]="/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the \"License\"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib=\"true\"/>\n\n\n/////////////////////////////\n/// ECMAScript APIs\n/////////////////////////////\n\ndeclare var NaN: number;\ndeclare var Infinity: number;\n\n/**\n * Evaluates JavaScript code and executes it.\n * @param x A String value that contains valid JavaScript code.\n */\ndeclare function eval(x: string): any;\n\n/**\n * Converts a string to an integer.\n * @param s A string to convert into a number.\n * @param radix A value between 2 and 36 that specifies the base of the number in numString.\n * If this argument is not supplied, strings with a prefix of '0x' are considered hexadecimal.\n * All other strings are considered decimal.\n */\ndeclare function parseInt(s: string, radix?: number): number;\n\n/**\n * Converts a string to a floating-point number.\n * @param string A string that contains a floating-point number.\n */\ndeclare function parseFloat(string: string): number;\n\n/**\n * Returns a Boolean value that indicates whether a value is the reserved value NaN (not a number).\n * @param number A numeric value.\n */\ndeclare function isNaN(number: number): boolean;\n\n/**\n * Determines whether a supplied number is finite.\n * @param number Any numeric value.\n */\ndeclare function isFinite(number: number): boolean;\n\n/**\n * Gets the unencoded version of an encoded Uniform Resource Identifier (URI).\n * @param encodedURI A value representing an encoded URI.\n */\ndeclare function decodeURI(encodedURI: string): string;\n\n/**\n * Gets the unencoded version of an encoded component of a Uniform Resource Identifier (URI).\n * @param encodedURIComponent A value representing an encoded URI component.\n */\ndeclare function decodeURIComponent(encodedURIComponent: string): string;\n\n/**\n * Encodes a text string as a valid Uniform Resource Identifier (URI)\n * @param uri A value representing an encoded URI.\n */\ndeclare function encodeURI(uri: string): string;\n\n/**\n * Encodes a text string as a valid component of a Uniform Resource Identifier (URI).\n * @param uriComponent A value representing an encoded URI component.\n */\ndeclare function encodeURIComponent(uriComponent: string | number | boolean): string;\n\n/**\n * Computes a new string in which certain characters have been replaced by a hexadecimal escape sequence.\n * @param string A string value\n */\ndeclare function escape(string: string): string;\n\n/**\n * Computes a new string in which hexadecimal escape sequences are replaced with the character that it represents.\n * @param string A string value\n */\ndeclare function unescape(string: string): string;\n\ninterface Symbol {\n    /** Returns a string representation of an object. */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): symbol;\n}\n\ndeclare type PropertyKey = string | number | symbol;\n\ninterface PropertyDescriptor {\n    configurable?: boolean;\n    enumerable?: boolean;\n    value?: any;\n    writable?: boolean;\n    get?(): any;\n    set?(v: any): void;\n}\n\ninterface PropertyDescriptorMap {\n    [s: string]: PropertyDescriptor;\n}\n\ninterface Object {\n    /** The initial value of Object.prototype.constructor is the standard built-in Object constructor. */\n    constructor: Function;\n\n    /** Returns a string representation of an object. */\n    toString(): string;\n\n    /** Returns a date converted to a string using the current locale. */\n    toLocaleString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Object;\n\n    /**\n     * Determines whether an object has a property with the specified name.\n     * @param v A property name.\n     */\n    hasOwnProperty(v: PropertyKey): boolean;\n\n    /**\n     * Determines whether an object exists in another object's prototype chain.\n     * @param v Another object whose prototype chain is to be checked.\n     */\n    isPrototypeOf(v: Object): boolean;\n\n    /**\n     * Determines whether a specified property is enumerable.\n     * @param v A property name.\n     */\n    propertyIsEnumerable(v: PropertyKey): boolean;\n}\n\ninterface ObjectConstructor {\n    new(value?: any): Object;\n    (): any;\n    (value: any): any;\n\n    /** A reference to the prototype for a class of objects. */\n    readonly prototype: Object;\n\n    /**\n     * Returns the prototype of an object.\n     * @param o The object that references the prototype.\n     */\n    getPrototypeOf(o: any): any;\n\n    /**\n     * Gets the own property descriptor of the specified object.\n     * An own property descriptor is one that is defined directly on the object and is not inherited from the object's prototype.\n     * @param o Object that contains the property.\n     * @param p Name of the property.\n     */\n    getOwnPropertyDescriptor(o: any, p: PropertyKey): PropertyDescriptor | undefined;\n\n    /**\n     * Returns the names of the own properties of an object. The own properties of an object are those that are defined directly\n     * on that object, and are not inherited from the object's prototype. The properties of an object include both fields (objects) and functions.\n     * @param o Object that contains the own properties.\n     */\n    getOwnPropertyNames(o: any): string[];\n\n    /**\n     * Creates an object that has the specified prototype or that has null prototype.\n     * @param o Object to use as a prototype. May be null.\n     */\n    create(o: object | null): any;\n\n    /**\n     * Creates an object that has the specified prototype, and that optionally contains specified properties.\n     * @param o Object to use as a prototype. May be null\n     * @param properties JavaScript object that contains one or more property descriptors.\n     */\n    create(o: object | null, properties: PropertyDescriptorMap & ThisType<any>): any;\n\n    /**\n     * Adds a property to an object, or modifies attributes of an existing property.\n     * @param o Object on which to add or modify the property. This can be a native JavaScript object (that is, a user-defined object or a built in object) or a DOM object.\n     * @param p The property name.\n     * @param attributes Descriptor for the property. It can be for a data property or an accessor property.\n     */\n    defineProperty(o: any, p: PropertyKey, attributes: PropertyDescriptor & ThisType<any>): any;\n\n    /**\n     * Adds one or more properties to an object, and/or modifies attributes of existing properties.\n     * @param o Object on which to add or modify the properties. This can be a native JavaScript object or a DOM object.\n     * @param properties JavaScript object that contains one or more descriptor objects. Each descriptor object describes a data property or an accessor property.\n     */\n    defineProperties(o: any, properties: PropertyDescriptorMap & ThisType<any>): any;\n\n    /**\n     * Prevents the modification of attributes of existing properties, and prevents the addition of new properties.\n     * @param o Object on which to lock the attributes.\n     */\n    seal<T>(o: T): T;\n\n    /**\n     * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n     * @param o Object on which to lock the attributes.\n     */\n    freeze<T>(a: T[]): readonly T[];\n\n    /**\n     * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n     * @param o Object on which to lock the attributes.\n     */\n    freeze<T extends Function>(f: T): T;\n\n    /**\n     * Prevents the modification of existing property attributes and values, and prevents the addition of new properties.\n     * @param o Object on which to lock the attributes.\n     */\n    freeze<T>(o: T): Readonly<T>;\n\n    /**\n     * Prevents the addition of new properties to an object.\n     * @param o Object to make non-extensible.\n     */\n    preventExtensions<T>(o: T): T;\n\n    /**\n     * Returns true if existing property attributes cannot be modified in an object and new properties cannot be added to the object.\n     * @param o Object to test.\n     */\n    isSealed(o: any): boolean;\n\n    /**\n     * Returns true if existing property attributes and values cannot be modified in an object, and new properties cannot be added to the object.\n     * @param o Object to test.\n     */\n    isFrozen(o: any): boolean;\n\n    /**\n     * Returns a value that indicates whether new properties can be added to an object.\n     * @param o Object to test.\n     */\n    isExtensible(o: any): boolean;\n\n    /**\n     * Returns the names of the enumerable string properties and methods of an object.\n     * @param o Object that contains the properties and methods. This can be an object that you created or an existing Document Object Model (DOM) object.\n     */\n    keys(o: object): string[];\n}\n\n/**\n * Provides functionality common to all JavaScript objects.\n */\ndeclare var Object: ObjectConstructor;\n\n/**\n * Creates a new function.\n */\ninterface Function {\n    /**\n     * Calls the function, substituting the specified object for the this value of the function, and the specified array for the arguments of the function.\n     * @param thisArg The object to be used as the this object.\n     * @param argArray A set of arguments to be passed to the function.\n     */\n    apply(this: Function, thisArg: any, argArray?: any): any;\n\n    /**\n     * Calls a method of an object, substituting another object for the current object.\n     * @param thisArg The object to be used as the current object.\n     * @param argArray A list of arguments to be passed to the method.\n     */\n    call(this: Function, thisArg: any, ...argArray: any[]): any;\n\n    /**\n     * For a given function, creates a bound function that has the same body as the original function.\n     * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n     * @param thisArg An object to which the this keyword can refer inside the new function.\n     * @param argArray A list of arguments to be passed to the new function.\n     */\n    bind(this: Function, thisArg: any, ...argArray: any[]): any;\n\n    /** Returns a string representation of a function. */\n    toString(): string;\n\n    prototype: any;\n    readonly length: number;\n\n    // Non-standard extensions\n    arguments: any;\n    caller: Function;\n}\n\ninterface FunctionConstructor {\n    /**\n     * Creates a new function.\n     * @param args A list of arguments the function accepts.\n     */\n    new(...args: string[]): Function;\n    (...args: string[]): Function;\n    readonly prototype: Function;\n}\n\ndeclare var Function: FunctionConstructor;\n\n/**\n * Extracts the type of the 'this' parameter of a function type, or 'unknown' if the function type has no 'this' parameter.\n */\ntype ThisParameterType<T> = T extends (this: infer U, ...args: any[]) => any ? U : unknown;\n\n/**\n * Removes the 'this' parameter from a function type.\n */\ntype OmitThisParameter<T> = unknown extends ThisParameterType<T> ? T : T extends (...args: infer A) => infer R ? (...args: A) => R : T;\n\ninterface CallableFunction extends Function {\n    /**\n     * Calls the function with the specified object as the this value and the elements of specified array as the arguments.\n     * @param thisArg The object to be used as the this object.\n     * @param args An array of argument values to be passed to the function.\n     */\n    apply<T, R>(this: (this: T) => R, thisArg: T): R;\n    apply<T, A extends any[], R>(this: (this: T, ...args: A) => R, thisArg: T, args: A): R;\n\n    /**\n     * Calls the function with the specified object as the this value and the specified rest arguments as the arguments.\n     * @param thisArg The object to be used as the this object.\n     * @param args Argument values to be passed to the function.\n     */\n    call<T, A extends any[], R>(this: (this: T, ...args: A) => R, thisArg: T, ...args: A): R;\n\n    /**\n     * For a given function, creates a bound function that has the same body as the original function.\n     * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n     * @param thisArg The object to be used as the this object.\n     * @param args Arguments to bind to the parameters of the function.\n     */\n    bind<T>(this: T, thisArg: ThisParameterType<T>): OmitThisParameter<T>;\n    bind<T, A0, A extends any[], R>(this: (this: T, arg0: A0, ...args: A) => R, thisArg: T, arg0: A0): (...args: A) => R;\n    bind<T, A0, A1, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1): (...args: A) => R;\n    bind<T, A0, A1, A2, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, arg2: A2, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1, arg2: A2): (...args: A) => R;\n    bind<T, A0, A1, A2, A3, A extends any[], R>(this: (this: T, arg0: A0, arg1: A1, arg2: A2, arg3: A3, ...args: A) => R, thisArg: T, arg0: A0, arg1: A1, arg2: A2, arg3: A3): (...args: A) => R;\n    bind<T, AX, R>(this: (this: T, ...args: AX[]) => R, thisArg: T, ...args: AX[]): (...args: AX[]) => R;\n}\n\ninterface NewableFunction extends Function {\n    /**\n     * Calls the function with the specified object as the this value and the elements of specified array as the arguments.\n     * @param thisArg The object to be used as the this object.\n     * @param args An array of argument values to be passed to the function.\n     */\n    apply<T>(this: new () => T, thisArg: T): void;\n    apply<T, A extends any[]>(this: new (...args: A) => T, thisArg: T, args: A): void;\n\n    /**\n     * Calls the function with the specified object as the this value and the specified rest arguments as the arguments.\n     * @param thisArg The object to be used as the this object.\n     * @param args Argument values to be passed to the function.\n     */\n    call<T, A extends any[]>(this: new (...args: A) => T, thisArg: T, ...args: A): void;\n\n    /**\n     * For a given function, creates a bound function that has the same body as the original function.\n     * The this object of the bound function is associated with the specified object, and has the specified initial parameters.\n     * @param thisArg The object to be used as the this object.\n     * @param args Arguments to bind to the parameters of the function.\n     */\n    bind<T>(this: T, thisArg: any): T;\n    bind<A0, A extends any[], R>(this: new (arg0: A0, ...args: A) => R, thisArg: any, arg0: A0): new (...args: A) => R;\n    bind<A0, A1, A extends any[], R>(this: new (arg0: A0, arg1: A1, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1): new (...args: A) => R;\n    bind<A0, A1, A2, A extends any[], R>(this: new (arg0: A0, arg1: A1, arg2: A2, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1, arg2: A2): new (...args: A) => R;\n    bind<A0, A1, A2, A3, A extends any[], R>(this: new (arg0: A0, arg1: A1, arg2: A2, arg3: A3, ...args: A) => R, thisArg: any, arg0: A0, arg1: A1, arg2: A2, arg3: A3): new (...args: A) => R;\n    bind<AX, R>(this: new (...args: AX[]) => R, thisArg: any, ...args: AX[]): new (...args: AX[]) => R;\n}\n\ninterface IArguments {\n    [index: number]: any;\n    length: number;\n    callee: Function;\n}\n\ninterface String {\n    /** Returns a string representation of a string. */\n    toString(): string;\n\n    /**\n     * Returns the character at the specified index.\n     * @param pos The zero-based index of the desired character.\n     */\n    charAt(pos: number): string;\n\n    /**\n     * Returns the Unicode value of the character at the specified location.\n     * @param index The zero-based index of the desired character. If there is no character at the specified index, NaN is returned.\n     */\n    charCodeAt(index: number): number;\n\n    /**\n     * Returns a string that contains the concatenation of two or more strings.\n     * @param strings The strings to append to the end of the string.\n     */\n    concat(...strings: string[]): string;\n\n    /**\n     * Returns the position of the first occurrence of a substring.\n     * @param searchString The substring to search for in the string\n     * @param position The index at which to begin searching the String object. If omitted, search starts at the beginning of the string.\n     */\n    indexOf(searchString: string, position?: number): number;\n\n    /**\n     * Returns the last occurrence of a substring in the string.\n     * @param searchString The substring to search for.\n     * @param position The index at which to begin searching. If omitted, the search begins at the end of the string.\n     */\n    lastIndexOf(searchString: string, position?: number): number;\n\n    /**\n     * Determines whether two strings are equivalent in the current locale.\n     * @param that String to compare to target string\n     */\n    localeCompare(that: string): number;\n\n    /**\n     * Matches a string with a regular expression, and returns an array containing the results of that search.\n     * @param regexp A variable name or string literal containing the regular expression pattern and flags.\n     */\n    match(regexp: string | RegExp): RegExpMatchArray | null;\n\n    /**\n     * Replaces text in a string, using a regular expression or search string.\n     * @param searchValue A string to search for.\n     * @param replaceValue A string containing the text to replace for every successful match of searchValue in this string.\n     */\n    replace(searchValue: string | RegExp, replaceValue: string): string;\n\n    /**\n     * Replaces text in a string, using a regular expression or search string.\n     * @param searchValue A string to search for.\n     * @param replacer A function that returns the replacement text.\n     */\n    replace(searchValue: string | RegExp, replacer: (substring: string, ...args: any[]) => string): string;\n\n    /**\n     * Finds the first substring match in a regular expression search.\n     * @param regexp The regular expression pattern and applicable flags.\n     */\n    search(regexp: string | RegExp): number;\n\n    /**\n     * Returns a section of a string.\n     * @param start The index to the beginning of the specified portion of stringObj.\n     * @param end The index to the end of the specified portion of stringObj. The substring includes the characters up to, but not including, the character indicated by end.\n     * If this value is not specified, the substring continues to the end of stringObj.\n     */\n    slice(start?: number, end?: number): string;\n\n    /**\n     * Split a string into substrings using the specified separator and return them as an array.\n     * @param separator A string that identifies character or characters to use in separating the string. If omitted, a single-element array containing the entire string is returned.\n     * @param limit A value used to limit the number of elements returned in the array.\n     */\n    split(separator: string | RegExp, limit?: number): string[];\n\n    /**\n     * Returns the substring at the specified location within a String object.\n     * @param start The zero-based index number indicating the beginning of the substring.\n     * @param end Zero-based index number indicating the end of the substring. The substring includes the characters up to, but not including, the character indicated by end.\n     * If end is omitted, the characters from start through the end of the original string are returned.\n     */\n    substring(start: number, end?: number): string;\n\n    /** Converts all the alphabetic characters in a string to lowercase. */\n    toLowerCase(): string;\n\n    /** Converts all alphabetic characters to lowercase, taking into account the host environment's current locale. */\n    toLocaleLowerCase(locales?: string | string[]): string;\n\n    /** Converts all the alphabetic characters in a string to uppercase. */\n    toUpperCase(): string;\n\n    /** Returns a string where all alphabetic characters have been converted to uppercase, taking into account the host environment's current locale. */\n    toLocaleUpperCase(locales?: string | string[]): string;\n\n    /** Removes the leading and trailing white space and line terminator characters from a string. */\n    trim(): string;\n\n    /** Returns the length of a String object. */\n    readonly length: number;\n\n    // IE extensions\n    /**\n     * Gets a substring beginning at the specified location and having the specified length.\n     * @param from The starting position of the desired substring. The index of the first character in the string is zero.\n     * @param length The number of characters to include in the returned substring.\n     */\n    substr(from: number, length?: number): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): string;\n\n    readonly [index: number]: string;\n}\n\ninterface StringConstructor {\n    new(value?: any): String;\n    (value?: any): string;\n    readonly prototype: String;\n    fromCharCode(...codes: number[]): string;\n}\n\n/**\n * Allows manipulation and formatting of text strings and determination and location of substrings within strings.\n */\ndeclare var String: StringConstructor;\n\ninterface Boolean {\n    /** Returns the primitive value of the specified object. */\n    valueOf(): boolean;\n}\n\ninterface BooleanConstructor {\n    new(value?: any): Boolean;\n    <T>(value?: T): boolean;\n    readonly prototype: Boolean;\n}\n\ndeclare var Boolean: BooleanConstructor;\n\ninterface Number {\n    /**\n     * Returns a string representation of an object.\n     * @param radix Specifies a radix for converting numeric values to strings. This value is only used for numbers.\n     */\n    toString(radix?: number): string;\n\n    /**\n     * Returns a string representing a number in fixed-point notation.\n     * @param fractionDigits Number of digits after the decimal point. Must be in the range 0 - 20, inclusive.\n     */\n    toFixed(fractionDigits?: number): string;\n\n    /**\n     * Returns a string containing a number represented in exponential notation.\n     * @param fractionDigits Number of digits after the decimal point. Must be in the range 0 - 20, inclusive.\n     */\n    toExponential(fractionDigits?: number): string;\n\n    /**\n     * Returns a string containing a number represented either in exponential or fixed-point notation with a specified number of digits.\n     * @param precision Number of significant digits. Must be in the range 1 - 21, inclusive.\n     */\n    toPrecision(precision?: number): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): number;\n}\n\ninterface NumberConstructor {\n    new(value?: any): Number;\n    (value?: any): number;\n    readonly prototype: Number;\n\n    /** The largest number that can be represented in JavaScript. Equal to approximately 1.79E+308. */\n    readonly MAX_VALUE: number;\n\n    /** The closest number to zero that can be represented in JavaScript. Equal to approximately 5.00E-324. */\n    readonly MIN_VALUE: number;\n\n    /**\n     * A value that is not a number.\n     * In equality comparisons, NaN does not equal any value, including itself. To test whether a value is equivalent to NaN, use the isNaN function.\n     */\n    readonly NaN: number;\n\n    /**\n     * A value that is less than the largest negative number that can be represented in JavaScript.\n     * JavaScript displays NEGATIVE_INFINITY values as -infinity.\n     */\n    readonly NEGATIVE_INFINITY: number;\n\n    /**\n     * A value greater than the largest number that can be represented in JavaScript.\n     * JavaScript displays POSITIVE_INFINITY values as infinity.\n     */\n    readonly POSITIVE_INFINITY: number;\n}\n\n/** An object that represents a number of any kind. All JavaScript numbers are 64-bit floating-point numbers. */\ndeclare var Number: NumberConstructor;\n\ninterface TemplateStringsArray extends ReadonlyArray<string> {\n    readonly raw: readonly string[];\n}\n\n/**\n * The type of `import.meta`.\n *\n * If you need to declare that a given property exists on `import.meta`,\n * this type may be augmented via interface merging.\n */\ninterface ImportMeta {\n}\n\ninterface Math {\n    /** The mathematical constant e. This is Euler's number, the base of natural logarithms. */\n    readonly E: number;\n    /** The natural logarithm of 10. */\n    readonly LN10: number;\n    /** The natural logarithm of 2. */\n    readonly LN2: number;\n    /** The base-2 logarithm of e. */\n    readonly LOG2E: number;\n    /** The base-10 logarithm of e. */\n    readonly LOG10E: number;\n    /** Pi. This is the ratio of the circumference of a circle to its diameter. */\n    readonly PI: number;\n    /** The square root of 0.5, or, equivalently, one divided by the square root of 2. */\n    readonly SQRT1_2: number;\n    /** The square root of 2. */\n    readonly SQRT2: number;\n    /**\n     * Returns the absolute value of a number (the value without regard to whether it is positive or negative).\n     * For example, the absolute value of -5 is the same as the absolute value of 5.\n     * @param x A numeric expression for which the absolute value is needed.\n     */\n    abs(x: number): number;\n    /**\n     * Returns the arc cosine (or inverse cosine) of a number.\n     * @param x A numeric expression.\n     */\n    acos(x: number): number;\n    /**\n     * Returns the arcsine of a number.\n     * @param x A numeric expression.\n     */\n    asin(x: number): number;\n    /**\n     * Returns the arctangent of a number.\n     * @param x A numeric expression for which the arctangent is needed.\n     */\n    atan(x: number): number;\n    /**\n     * Returns the angle (in radians) from the X axis to a point.\n     * @param y A numeric expression representing the cartesian y-coordinate.\n     * @param x A numeric expression representing the cartesian x-coordinate.\n     */\n    atan2(y: number, x: number): number;\n    /**\n     * Returns the smallest integer greater than or equal to its numeric argument.\n     * @param x A numeric expression.\n     */\n    ceil(x: number): number;\n    /**\n     * Returns the cosine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    cos(x: number): number;\n    /**\n     * Returns e (the base of natural logarithms) raised to a power.\n     * @param x A numeric expression representing the power of e.\n     */\n    exp(x: number): number;\n    /**\n     * Returns the greatest integer less than or equal to its numeric argument.\n     * @param x A numeric expression.\n     */\n    floor(x: number): number;\n    /**\n     * Returns the natural logarithm (base e) of a number.\n     * @param x A numeric expression.\n     */\n    log(x: number): number;\n    /**\n     * Returns the larger of a set of supplied numeric expressions.\n     * @param values Numeric expressions to be evaluated.\n     */\n    max(...values: number[]): number;\n    /**\n     * Returns the smaller of a set of supplied numeric expressions.\n     * @param values Numeric expressions to be evaluated.\n     */\n    min(...values: number[]): number;\n    /**\n     * Returns the value of a base expression taken to a specified power.\n     * @param x The base value of the expression.\n     * @param y The exponent value of the expression.\n     */\n    pow(x: number, y: number): number;\n    /** Returns a pseudorandom number between 0 and 1. */\n    random(): number;\n    /**\n     * Returns a supplied numeric expression rounded to the nearest integer.\n     * @param x The value to be rounded to the nearest integer.\n     */\n    round(x: number): number;\n    /**\n     * Returns the sine of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    sin(x: number): number;\n    /**\n     * Returns the square root of a number.\n     * @param x A numeric expression.\n     */\n    sqrt(x: number): number;\n    /**\n     * Returns the tangent of a number.\n     * @param x A numeric expression that contains an angle measured in radians.\n     */\n    tan(x: number): number;\n}\n/** An intrinsic object that provides basic mathematics functionality and constants. */\ndeclare var Math: Math;\n\n/** Enables basic storage and retrieval of dates and times. */\ninterface Date {\n    /** Returns a string representation of a date. The format of the string depends on the locale. */\n    toString(): string;\n    /** Returns a date as a string value. */\n    toDateString(): string;\n    /** Returns a time as a string value. */\n    toTimeString(): string;\n    /** Returns a value as a string value appropriate to the host environment's current locale. */\n    toLocaleString(): string;\n    /** Returns a date as a string value appropriate to the host environment's current locale. */\n    toLocaleDateString(): string;\n    /** Returns a time as a string value appropriate to the host environment's current locale. */\n    toLocaleTimeString(): string;\n    /** Returns the stored time value in milliseconds since midnight, January 1, 1970 UTC. */\n    valueOf(): number;\n    /** Gets the time value in milliseconds. */\n    getTime(): number;\n    /** Gets the year, using local time. */\n    getFullYear(): number;\n    /** Gets the year using Universal Coordinated Time (UTC). */\n    getUTCFullYear(): number;\n    /** Gets the month, using local time. */\n    getMonth(): number;\n    /** Gets the month of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMonth(): number;\n    /** Gets the day-of-the-month, using local time. */\n    getDate(): number;\n    /** Gets the day-of-the-month, using Universal Coordinated Time (UTC). */\n    getUTCDate(): number;\n    /** Gets the day of the week, using local time. */\n    getDay(): number;\n    /** Gets the day of the week using Universal Coordinated Time (UTC). */\n    getUTCDay(): number;\n    /** Gets the hours in a date, using local time. */\n    getHours(): number;\n    /** Gets the hours value in a Date object using Universal Coordinated Time (UTC). */\n    getUTCHours(): number;\n    /** Gets the minutes of a Date object, using local time. */\n    getMinutes(): number;\n    /** Gets the minutes of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMinutes(): number;\n    /** Gets the seconds of a Date object, using local time. */\n    getSeconds(): number;\n    /** Gets the seconds of a Date object using Universal Coordinated Time (UTC). */\n    getUTCSeconds(): number;\n    /** Gets the milliseconds of a Date, using local time. */\n    getMilliseconds(): number;\n    /** Gets the milliseconds of a Date object using Universal Coordinated Time (UTC). */\n    getUTCMilliseconds(): number;\n    /** Gets the difference in minutes between the time on the local computer and Universal Coordinated Time (UTC). */\n    getTimezoneOffset(): number;\n    /**\n     * Sets the date and time value in the Date object.\n     * @param time A numeric value representing the number of elapsed milliseconds since midnight, January 1, 1970 GMT.\n     */\n    setTime(time: number): number;\n    /**\n     * Sets the milliseconds value in the Date object using local time.\n     * @param ms A numeric value equal to the millisecond value.\n     */\n    setMilliseconds(ms: number): number;\n    /**\n     * Sets the milliseconds value in the Date object using Universal Coordinated Time (UTC).\n     * @param ms A numeric value equal to the millisecond value.\n     */\n    setUTCMilliseconds(ms: number): number;\n\n    /**\n     * Sets the seconds value in the Date object using local time.\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setSeconds(sec: number, ms?: number): number;\n    /**\n     * Sets the seconds value in the Date object using Universal Coordinated Time (UTC).\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setUTCSeconds(sec: number, ms?: number): number;\n    /**\n     * Sets the minutes value in the Date object using local time.\n     * @param min A numeric value equal to the minutes value.\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setMinutes(min: number, sec?: number, ms?: number): number;\n    /**\n     * Sets the minutes value in the Date object using Universal Coordinated Time (UTC).\n     * @param min A numeric value equal to the minutes value.\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setUTCMinutes(min: number, sec?: number, ms?: number): number;\n    /**\n     * Sets the hour value in the Date object using local time.\n     * @param hours A numeric value equal to the hours value.\n     * @param min A numeric value equal to the minutes value.\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setHours(hours: number, min?: number, sec?: number, ms?: number): number;\n    /**\n     * Sets the hours value in the Date object using Universal Coordinated Time (UTC).\n     * @param hours A numeric value equal to the hours value.\n     * @param min A numeric value equal to the minutes value.\n     * @param sec A numeric value equal to the seconds value.\n     * @param ms A numeric value equal to the milliseconds value.\n     */\n    setUTCHours(hours: number, min?: number, sec?: number, ms?: number): number;\n    /**\n     * Sets the numeric day-of-the-month value of the Date object using local time.\n     * @param date A numeric value equal to the day of the month.\n     */\n    setDate(date: number): number;\n    /**\n     * Sets the numeric day of the month in the Date object using Universal Coordinated Time (UTC).\n     * @param date A numeric value equal to the day of the month.\n     */\n    setUTCDate(date: number): number;\n    /**\n     * Sets the month value in the Date object using local time.\n     * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively.\n     * @param date A numeric value representing the day of the month. If this value is not supplied, the value from a call to the getDate method is used.\n     */\n    setMonth(month: number, date?: number): number;\n    /**\n     * Sets the month value in the Date object using Universal Coordinated Time (UTC).\n     * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively.\n     * @param date A numeric value representing the day of the month. If it is not supplied, the value from a call to the getUTCDate method is used.\n     */\n    setUTCMonth(month: number, date?: number): number;\n    /**\n     * Sets the year of the Date object using local time.\n     * @param year A numeric value for the year.\n     * @param month A zero-based numeric value for the month (0 for January, 11 for December). Must be specified if numDate is specified.\n     * @param date A numeric value equal for the day of the month.\n     */\n    setFullYear(year: number, month?: number, date?: number): number;\n    /**\n     * Sets the year value in the Date object using Universal Coordinated Time (UTC).\n     * @param year A numeric value equal to the year.\n     * @param month A numeric value equal to the month. The value for January is 0, and other month values follow consecutively. Must be supplied if numDate is supplied.\n     * @param date A numeric value equal to the day of the month.\n     */\n    setUTCFullYear(year: number, month?: number, date?: number): number;\n    /** Returns a date converted to a string using Universal Coordinated Time (UTC). */\n    toUTCString(): string;\n    /** Returns a date as a string value in ISO format. */\n    toISOString(): string;\n    /** Used by the JSON.stringify method to enable the transformation of an object's data for JavaScript Object Notation (JSON) serialization. */\n    toJSON(key?: any): string;\n}\n\ninterface DateConstructor {\n    new(): Date;\n    new(value: number | string): Date;\n    new(year: number, month: number, date?: number, hours?: number, minutes?: number, seconds?: number, ms?: number): Date;\n    (): string;\n    readonly prototype: Date;\n    /**\n     * Parses a string containing a date, and returns the number of milliseconds between that date and midnight, January 1, 1970.\n     * @param s A date string\n     */\n    parse(s: string): number;\n    /**\n     * Returns the number of milliseconds between midnight, January 1, 1970 Universal Coordinated Time (UTC) (or GMT) and the specified date.\n     * @param year The full year designation is required for cross-century date accuracy. If year is between 0 and 99 is used, then year is assumed to be 1900 + year.\n     * @param month The month as a number between 0 and 11 (January to December).\n     * @param date The date as a number between 1 and 31.\n     * @param hours Must be supplied if minutes is supplied. A number from 0 to 23 (midnight to 11pm) that specifies the hour.\n     * @param minutes Must be supplied if seconds is supplied. A number from 0 to 59 that specifies the minutes.\n     * @param seconds Must be supplied if milliseconds is supplied. A number from 0 to 59 that specifies the seconds.\n     * @param ms A number from 0 to 999 that specifies the milliseconds.\n     */\n    UTC(year: number, month: number, date?: number, hours?: number, minutes?: number, seconds?: number, ms?: number): number;\n    now(): number;\n}\n\ndeclare var Date: DateConstructor;\n\ninterface RegExpMatchArray extends Array<string> {\n    index?: number;\n    input?: string;\n}\n\ninterface RegExpExecArray extends Array<string> {\n    index: number;\n    input: string;\n}\n\ninterface RegExp {\n    /**\n     * Executes a search on a string using a regular expression pattern, and returns an array containing the results of that search.\n     * @param string The String object or string literal on which to perform the search.\n     */\n    exec(string: string): RegExpExecArray | null;\n\n    /**\n     * Returns a Boolean value that indicates whether or not a pattern exists in a searched string.\n     * @param string String on which to perform the search.\n     */\n    test(string: string): boolean;\n\n    /** Returns a copy of the text of the regular expression pattern. Read-only. The regExp argument is a Regular expression object. It can be a variable name or a literal. */\n    readonly source: string;\n\n    /** Returns a Boolean value indicating the state of the global flag (g) used with a regular expression. Default is false. Read-only. */\n    readonly global: boolean;\n\n    /** Returns a Boolean value indicating the state of the ignoreCase flag (i) used with a regular expression. Default is false. Read-only. */\n    readonly ignoreCase: boolean;\n\n    /** Returns a Boolean value indicating the state of the multiline flag (m) used with a regular expression. Default is false. Read-only. */\n    readonly multiline: boolean;\n\n    lastIndex: number;\n\n    // Non-standard extensions\n    compile(): this;\n}\n\ninterface RegExpConstructor {\n    new(pattern: RegExp | string): RegExp;\n    new(pattern: string, flags?: string): RegExp;\n    (pattern: RegExp | string): RegExp;\n    (pattern: string, flags?: string): RegExp;\n    readonly prototype: RegExp;\n\n    // Non-standard extensions\n    $1: string;\n    $2: string;\n    $3: string;\n    $4: string;\n    $5: string;\n    $6: string;\n    $7: string;\n    $8: string;\n    $9: string;\n    lastMatch: string;\n}\n\ndeclare var RegExp: RegExpConstructor;\n\ninterface Error {\n    name: string;\n    message: string;\n    stack?: string;\n}\n\ninterface ErrorConstructor {\n    new(message?: string): Error;\n    (message?: string): Error;\n    readonly prototype: Error;\n}\n\ndeclare var Error: ErrorConstructor;\n\ninterface EvalError extends Error {\n}\n\ninterface EvalErrorConstructor extends ErrorConstructor {\n    new(message?: string): EvalError;\n    (message?: string): EvalError;\n    readonly prototype: EvalError;\n}\n\ndeclare var EvalError: EvalErrorConstructor;\n\ninterface RangeError extends Error {\n}\n\ninterface RangeErrorConstructor extends ErrorConstructor {\n    new(message?: string): RangeError;\n    (message?: string): RangeError;\n    readonly prototype: RangeError;\n}\n\ndeclare var RangeError: RangeErrorConstructor;\n\ninterface ReferenceError extends Error {\n}\n\ninterface ReferenceErrorConstructor extends ErrorConstructor {\n    new(message?: string): ReferenceError;\n    (message?: string): ReferenceError;\n    readonly prototype: ReferenceError;\n}\n\ndeclare var ReferenceError: ReferenceErrorConstructor;\n\ninterface SyntaxError extends Error {\n}\n\ninterface SyntaxErrorConstructor extends ErrorConstructor {\n    new(message?: string): SyntaxError;\n    (message?: string): SyntaxError;\n    readonly prototype: SyntaxError;\n}\n\ndeclare var SyntaxError: SyntaxErrorConstructor;\n\ninterface TypeError extends Error {\n}\n\ninterface TypeErrorConstructor extends ErrorConstructor {\n    new(message?: string): TypeError;\n    (message?: string): TypeError;\n    readonly prototype: TypeError;\n}\n\ndeclare var TypeError: TypeErrorConstructor;\n\ninterface URIError extends Error {\n}\n\ninterface URIErrorConstructor extends ErrorConstructor {\n    new(message?: string): URIError;\n    (message?: string): URIError;\n    readonly prototype: URIError;\n}\n\ndeclare var URIError: URIErrorConstructor;\n\ninterface JSON {\n    /**\n     * Converts a JavaScript Object Notation (JSON) string into an object.\n     * @param text A valid JSON string.\n     * @param reviver A function that transforms the results. This function is called for each member of the object.\n     * If a member contains nested objects, the nested objects are transformed before the parent object is.\n     */\n    parse(text: string, reviver?: (this: any, key: string, value: any) => any): any;\n    /**\n     * Converts a JavaScript value to a JavaScript Object Notation (JSON) string.\n     * @param value A JavaScript value, usually an object or array, to be converted.\n     * @param replacer A function that transforms the results.\n     * @param space Adds indentation, white space, and line break characters to the return-value JSON text to make it easier to read.\n     */\n    stringify(value: any, replacer?: (this: any, key: string, value: any) => any, space?: string | number): string;\n    /**\n     * Converts a JavaScript value to a JavaScript Object Notation (JSON) string.\n     * @param value A JavaScript value, usually an object or array, to be converted.\n     * @param replacer An array of strings and numbers that acts as an approved list for selecting the object properties that will be stringified.\n     * @param space Adds indentation, white space, and line break characters to the return-value JSON text to make it easier to read.\n     */\n    stringify(value: any, replacer?: (number | string)[] | null, space?: string | number): string;\n}\n\n/**\n * An intrinsic object that provides functions to convert JavaScript values to and from the JavaScript Object Notation (JSON) format.\n */\ndeclare var JSON: JSON;\n\n\n/////////////////////////////\n/// ECMAScript Array API (specially handled by compiler)\n/////////////////////////////\n\ninterface ReadonlyArray<T> {\n    /**\n     * Gets the length of the array. This is a number one higher than the highest element defined in an array.\n     */\n    readonly length: number;\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n    /**\n     * Returns a string representation of an array. The elements are converted to string using their toLocalString methods.\n     */\n    toLocaleString(): string;\n    /**\n     * Combines two or more arrays.\n     * @param items Additional items to add to the end of array1.\n     */\n    concat(...items: ConcatArray<T>[]): T[];\n    /**\n     * Combines two or more arrays.\n     * @param items Additional items to add to the end of array1.\n     */\n    concat(...items: (T | ConcatArray<T>)[]): T[];\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): T[];\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at index 0.\n     */\n    indexOf(searchElement: T, fromIndex?: number): number;\n    /**\n     * Returns the index of the last occurrence of a specified value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at the last index in the array.\n     */\n    lastIndexOf(searchElement: T, fromIndex?: number): number;\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every<S extends T>(predicate: (value: T, index: number, array: readonly T[]) => value is S, thisArg?: any): this is readonly S[];\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: T, index: number, array: readonly T[]) => unknown, thisArg?: any): boolean;\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: T, index: number, array: readonly T[]) => unknown, thisArg?: any): boolean;\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: T, index: number, array: readonly T[]) => void, thisArg?: any): void;\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    map<U>(callbackfn: (value: T, index: number, array: readonly T[]) => U, thisArg?: any): U[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter<S extends T>(predicate: (value: T, index: number, array: readonly T[]) => value is S, thisArg?: any): S[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: T, index: number, array: readonly T[]) => unknown, thisArg?: any): T[];\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: readonly T[]) => T): T;\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: readonly T[]) => T, initialValue: T): T;\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: readonly T[]) => U, initialValue: U): U;\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: readonly T[]) => T): T;\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: readonly T[]) => T, initialValue: T): T;\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: readonly T[]) => U, initialValue: U): U;\n\n    readonly [n: number]: T;\n}\n\ninterface ConcatArray<T> {\n    readonly length: number;\n    readonly [n: number]: T;\n    join(separator?: string): string;\n    slice(start?: number, end?: number): T[];\n}\n\ninterface Array<T> {\n    /**\n     * Gets or sets the length of the array. This is a number one higher than the highest element defined in an array.\n     */\n    length: number;\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n    /**\n     * Returns a string representation of an array. The elements are converted to string using their toLocalString methods.\n     */\n    toLocaleString(): string;\n    /**\n     * Removes the last element from an array and returns it.\n     */\n    pop(): T | undefined;\n    /**\n     * Appends new elements to an array, and returns the new length of the array.\n     * @param items New elements of the Array.\n     */\n    push(...items: T[]): number;\n    /**\n     * Combines two or more arrays.\n     * @param items Additional items to add to the end of array1.\n     */\n    concat(...items: ConcatArray<T>[]): T[];\n    /**\n     * Combines two or more arrays.\n     * @param items Additional items to add to the end of array1.\n     */\n    concat(...items: (T | ConcatArray<T>)[]): T[];\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): T[];\n    /**\n     * Removes the first element from an array and returns it.\n     */\n    shift(): T | undefined;\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): T[];\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: T, b: T) => number): this;\n    /**\n     * Removes elements from an array and, if necessary, inserts new elements in their place, returning the deleted elements.\n     * @param start The zero-based location in the array from which to start removing elements.\n     * @param deleteCount The number of elements to remove.\n     */\n    splice(start: number, deleteCount?: number): T[];\n    /**\n     * Removes elements from an array and, if necessary, inserts new elements in their place, returning the deleted elements.\n     * @param start The zero-based location in the array from which to start removing elements.\n     * @param deleteCount The number of elements to remove.\n     * @param items Elements to insert into the array in place of the deleted elements.\n     */\n    splice(start: number, deleteCount: number, ...items: T[]): T[];\n    /**\n     * Inserts new elements at the start of an array.\n     * @param items  Elements to insert at the start of the Array.\n     */\n    unshift(...items: T[]): number;\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at index 0.\n     */\n    indexOf(searchElement: T, fromIndex?: number): number;\n    /**\n     * Returns the index of the last occurrence of a specified value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the search starts at the last index in the array.\n     */\n    lastIndexOf(searchElement: T, fromIndex?: number): number;\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every<S extends T>(predicate: (value: T, index: number, array: T[]) => value is S, thisArg?: any): this is S[];\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: T, index: number, array: T[]) => unknown, thisArg?: any): boolean;\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: T, index: number, array: T[]) => unknown, thisArg?: any): boolean;\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: T, index: number, array: T[]) => void, thisArg?: any): void;\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function. If thisArg is omitted, undefined is used as the this value.\n     */\n    map<U>(callbackfn: (value: T, index: number, array: T[]) => U, thisArg?: any): U[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter<S extends T>(predicate: (value: T, index: number, array: T[]) => value is S, thisArg?: any): S[];\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function. If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: T, index: number, array: T[]) => unknown, thisArg?: any): T[];\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T): T;\n    reduce(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T, initialValue: T): T;\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: T[]) => U, initialValue: U): U;\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T): T;\n    reduceRight(callbackfn: (previousValue: T, currentValue: T, currentIndex: number, array: T[]) => T, initialValue: T): T;\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order. The return value of the callback function is the accumulated result, and is provided as an argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start the accumulation. The first call to the callbackfn function provides this value as an argument instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: T, currentIndex: number, array: T[]) => U, initialValue: U): U;\n\n    [n: number]: T;\n}\n\ninterface ArrayConstructor {\n    new(arrayLength?: number): any[];\n    new <T>(arrayLength: number): T[];\n    new <T>(...items: T[]): T[];\n    (arrayLength?: number): any[];\n    <T>(arrayLength: number): T[];\n    <T>(...items: T[]): T[];\n    isArray(arg: any): arg is any[];\n    readonly prototype: any[];\n}\n\ndeclare var Array: ArrayConstructor;\n\ninterface TypedPropertyDescriptor<T> {\n    enumerable?: boolean;\n    configurable?: boolean;\n    writable?: boolean;\n    value?: T;\n    get?: () => T;\n    set?: (value: T) => void;\n}\n\ndeclare type ClassDecorator = <TFunction extends Function>(target: TFunction) => TFunction | void;\ndeclare type PropertyDecorator = (target: Object, propertyKey: string | symbol) => void;\ndeclare type MethodDecorator = <T>(target: Object, propertyKey: string | symbol, descriptor: TypedPropertyDescriptor<T>) => TypedPropertyDescriptor<T> | void;\ndeclare type ParameterDecorator = (target: Object, propertyKey: string | symbol, parameterIndex: number) => void;\n\ndeclare type PromiseConstructorLike = new <T>(executor: (resolve: (value: T | PromiseLike<T>) => void, reject: (reason?: any) => void) => void) => PromiseLike<T>;\n\ninterface PromiseLike<T> {\n    /**\n     * Attaches callbacks for the resolution and/or rejection of the Promise.\n     * @param onfulfilled The callback to execute when the Promise is resolved.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of which ever callback is executed.\n     */\n    then<TResult1 = T, TResult2 = never>(onfulfilled?: ((value: T) => TResult1 | PromiseLike<TResult1>) | undefined | null, onrejected?: ((reason: any) => TResult2 | PromiseLike<TResult2>) | undefined | null): PromiseLike<TResult1 | TResult2>;\n}\n\n/**\n * Represents the completion of an asynchronous operation\n */\ninterface Promise<T> {\n    /**\n     * Attaches callbacks for the resolution and/or rejection of the Promise.\n     * @param onfulfilled The callback to execute when the Promise is resolved.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of which ever callback is executed.\n     */\n    then<TResult1 = T, TResult2 = never>(onfulfilled?: ((value: T) => TResult1 | PromiseLike<TResult1>) | undefined | null, onrejected?: ((reason: any) => TResult2 | PromiseLike<TResult2>) | undefined | null): Promise<TResult1 | TResult2>;\n\n    /**\n     * Attaches a callback for only the rejection of the Promise.\n     * @param onrejected The callback to execute when the Promise is rejected.\n     * @returns A Promise for the completion of the callback.\n     */\n    catch<TResult = never>(onrejected?: ((reason: any) => TResult | PromiseLike<TResult>) | undefined | null): Promise<T | TResult>;\n}\n\ninterface ArrayLike<T> {\n    readonly length: number;\n    readonly [n: number]: T;\n}\n\n/**\n * Make all properties in T optional\n */\ntype Partial<T> = {\n    [P in keyof T]?: T[P];\n};\n\n/**\n * Make all properties in T required\n */\ntype Required<T> = {\n    [P in keyof T]-?: T[P];\n};\n\n/**\n * Make all properties in T readonly\n */\ntype Readonly<T> = {\n    readonly [P in keyof T]: T[P];\n};\n\n/**\n * From T, pick a set of properties whose keys are in the union K\n */\ntype Pick<T, K extends keyof T> = {\n    [P in K]: T[P];\n};\n\n/**\n * Construct a type with a set of properties K of type T\n */\ntype Record<K extends keyof any, T> = {\n    [P in K]: T;\n};\n\n/**\n * Exclude from T those types that are assignable to U\n */\ntype Exclude<T, U> = T extends U ? never : T;\n\n/**\n * Extract from T those types that are assignable to U\n */\ntype Extract<T, U> = T extends U ? T : never;\n\n/**\n * Construct a type with the properties of T except for those in type K.\n */\ntype Omit<T, K extends keyof any> = Pick<T, Exclude<keyof T, K>>;\n\n/**\n * Exclude null and undefined from T\n */\ntype NonNullable<T> = T extends null | undefined ? never : T;\n\n/**\n * Obtain the parameters of a function type in a tuple\n */\ntype Parameters<T extends (...args: any) => any> = T extends (...args: infer P) => any ? P : never;\n\n/**\n * Obtain the parameters of a constructor function type in a tuple\n */\ntype ConstructorParameters<T extends new (...args: any) => any> = T extends new (...args: infer P) => any ? P : never;\n\n/**\n * Obtain the return type of a function type\n */\ntype ReturnType<T extends (...args: any) => any> = T extends (...args: any) => infer R ? R : any;\n\n/**\n * Obtain the return type of a constructor function type\n */\ntype InstanceType<T extends new (...args: any) => any> = T extends new (...args: any) => infer R ? R : any;\n\n/**\n * Convert string literal type to uppercase\n */\ntype Uppercase<S extends string> = intrinsic;\n\n/**\n * Convert string literal type to lowercase\n */\ntype Lowercase<S extends string> = intrinsic;\n\n/**\n * Convert first character of string literal type to uppercase\n */\ntype Capitalize<S extends string> = intrinsic;\n\n/**\n * Convert first character of string literal type to lowercase\n */\ntype Uncapitalize<S extends string> = intrinsic;\n\n/**\n * Marker for contextual 'this' type\n */\ninterface ThisType<T> { }\n\n/**\n * Represents a raw buffer of binary data, which is used to store data for the\n * different typed arrays. ArrayBuffers cannot be read from or written to directly,\n * but can be passed to a typed array or DataView Object to interpret the raw\n * buffer as needed.\n */\ninterface ArrayBuffer {\n    /**\n     * Read-only. The length of the ArrayBuffer (in bytes).\n     */\n    readonly byteLength: number;\n\n    /**\n     * Returns a section of an ArrayBuffer.\n     */\n    slice(begin: number, end?: number): ArrayBuffer;\n}\n\n/**\n * Allowed ArrayBuffer types for the buffer of an ArrayBufferView and related Typed Arrays.\n */\ninterface ArrayBufferTypes {\n    ArrayBuffer: ArrayBuffer;\n}\ntype ArrayBufferLike = ArrayBufferTypes[keyof ArrayBufferTypes];\n\ninterface ArrayBufferConstructor {\n    readonly prototype: ArrayBuffer;\n    new(byteLength: number): ArrayBuffer;\n    isView(arg: any): arg is ArrayBufferView;\n}\ndeclare var ArrayBuffer: ArrayBufferConstructor;\n\ninterface ArrayBufferView {\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    byteOffset: number;\n}\n\ninterface DataView {\n    readonly buffer: ArrayBuffer;\n    readonly byteLength: number;\n    readonly byteOffset: number;\n    /**\n     * Gets the Float32 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getFloat32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n     * Gets the Float64 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getFloat64(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n     * Gets the Int8 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getInt8(byteOffset: number): number;\n\n    /**\n     * Gets the Int16 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getInt16(byteOffset: number, littleEndian?: boolean): number;\n    /**\n     * Gets the Int32 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getInt32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n     * Gets the Uint8 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getUint8(byteOffset: number): number;\n\n    /**\n     * Gets the Uint16 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getUint16(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n     * Gets the Uint32 value at the specified byte offset from the start of the view. There is\n     * no alignment constraint; multi-byte values may be fetched from any offset.\n     * @param byteOffset The place in the buffer at which the value should be retrieved.\n     */\n    getUint32(byteOffset: number, littleEndian?: boolean): number;\n\n    /**\n     * Stores an Float32 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setFloat32(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n     * Stores an Float64 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setFloat64(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n     * Stores an Int8 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     */\n    setInt8(byteOffset: number, value: number): void;\n\n    /**\n     * Stores an Int16 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setInt16(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n     * Stores an Int32 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setInt32(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n     * Stores an Uint8 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     */\n    setUint8(byteOffset: number, value: number): void;\n\n    /**\n     * Stores an Uint16 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setUint16(byteOffset: number, value: number, littleEndian?: boolean): void;\n\n    /**\n     * Stores an Uint32 value at the specified byte offset from the start of the view.\n     * @param byteOffset The place in the buffer at which the value should be set.\n     * @param value The value to set.\n     * @param littleEndian If false or undefined, a big-endian value should be written,\n     * otherwise a little-endian value should be written.\n     */\n    setUint32(byteOffset: number, value: number, littleEndian?: boolean): void;\n}\n\ninterface DataViewConstructor {\n    readonly prototype: DataView;\n    new(buffer: ArrayBufferLike, byteOffset?: number, byteLength?: number): DataView;\n}\ndeclare var DataView: DataViewConstructor;\n\n/**\n * A typed array of 8-bit integer values. The contents are initialized to 0. If the requested\n * number of bytes could not be allocated an exception is raised.\n */\ninterface Int8Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Int8Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Int8Array) => any, thisArg?: any): Int8Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Int8Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Int8Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Int8Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Int8Array) => number, thisArg?: any): Int8Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int8Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int8Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int8Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Int8Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Int8Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Int8Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Int8Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Int8Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Int8Array;\n\n    [index: number]: number;\n}\ninterface Int8ArrayConstructor {\n    readonly prototype: Int8Array;\n    new(length: number): Int8Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Int8Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Int8Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Int8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Int8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int8Array;\n\n\n}\ndeclare var Int8Array: Int8ArrayConstructor;\n\n/**\n * A typed array of 8-bit unsigned integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated an exception is raised.\n */\ninterface Uint8Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Uint8Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Uint8Array) => any, thisArg?: any): Uint8Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Uint8Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Uint8Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Uint8Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Uint8Array) => number, thisArg?: any): Uint8Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Uint8Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Uint8Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Uint8Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Uint8Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Uint8Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Uint8Array;\n\n    [index: number]: number;\n}\n\ninterface Uint8ArrayConstructor {\n    readonly prototype: Uint8Array;\n    new(length: number): Uint8Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Uint8Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Uint8Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Uint8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Uint8Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint8Array;\n\n}\ndeclare var Uint8Array: Uint8ArrayConstructor;\n\n/**\n * A typed array of 8-bit unsigned integer (clamped) values. The contents are initialized to 0.\n * If the requested number of bytes could not be allocated an exception is raised.\n */\ninterface Uint8ClampedArray {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Uint8ClampedArray) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Uint8ClampedArray) => any, thisArg?: any): Uint8ClampedArray;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Uint8ClampedArray) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Uint8ClampedArray) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Uint8ClampedArray) => number, thisArg?: any): Uint8ClampedArray;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint8ClampedArray) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Uint8ClampedArray;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Uint8ClampedArray;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Uint8ClampedArray) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Uint8ClampedArray view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Uint8ClampedArray;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Uint8ClampedArray;\n\n    [index: number]: number;\n}\n\ninterface Uint8ClampedArrayConstructor {\n    readonly prototype: Uint8ClampedArray;\n    new(length: number): Uint8ClampedArray;\n    new(array: ArrayLike<number> | ArrayBufferLike): Uint8ClampedArray;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Uint8ClampedArray;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Uint8ClampedArray;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Uint8ClampedArray;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint8ClampedArray;\n}\ndeclare var Uint8ClampedArray: Uint8ClampedArrayConstructor;\n\n/**\n * A typed array of 16-bit signed integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated an exception is raised.\n */\ninterface Int16Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Int16Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Int16Array) => any, thisArg?: any): Int16Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Int16Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Int16Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Int16Array) => void, thisArg?: any): void;\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Int16Array) => number, thisArg?: any): Int16Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int16Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int16Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int16Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Int16Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Int16Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Int16Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Int16Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Int16Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Int16Array;\n\n    [index: number]: number;\n}\n\ninterface Int16ArrayConstructor {\n    readonly prototype: Int16Array;\n    new(length: number): Int16Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Int16Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Int16Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Int16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Int16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int16Array;\n\n\n}\ndeclare var Int16Array: Int16ArrayConstructor;\n\n/**\n * A typed array of 16-bit unsigned integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated an exception is raised.\n */\ninterface Uint16Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Uint16Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Uint16Array) => any, thisArg?: any): Uint16Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Uint16Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Uint16Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Uint16Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Uint16Array) => number, thisArg?: any): Uint16Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint16Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint16Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint16Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Uint16Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Uint16Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Uint16Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Uint16Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Uint16Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Uint16Array;\n\n    [index: number]: number;\n}\n\ninterface Uint16ArrayConstructor {\n    readonly prototype: Uint16Array;\n    new(length: number): Uint16Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Uint16Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Uint16Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Uint16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Uint16Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint16Array;\n\n\n}\ndeclare var Uint16Array: Uint16ArrayConstructor;\n/**\n * A typed array of 32-bit signed integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated an exception is raised.\n */\ninterface Int32Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Int32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Int32Array) => any, thisArg?: any): Int32Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Int32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Int32Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Int32Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Int32Array) => number, thisArg?: any): Int32Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int32Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Int32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Int32Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Int32Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Int32Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Int32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Int32Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Int32Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Int32Array;\n\n    [index: number]: number;\n}\n\ninterface Int32ArrayConstructor {\n    readonly prototype: Int32Array;\n    new(length: number): Int32Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Int32Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Int32Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Int32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Int32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Int32Array;\n\n}\ndeclare var Int32Array: Int32ArrayConstructor;\n\n/**\n * A typed array of 32-bit unsigned integer values. The contents are initialized to 0. If the\n * requested number of bytes could not be allocated an exception is raised.\n */\ninterface Uint32Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Uint32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Uint32Array) => any, thisArg?: any): Uint32Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Uint32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Uint32Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Uint32Array) => void, thisArg?: any): void;\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Uint32Array) => number, thisArg?: any): Uint32Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint32Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Uint32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Uint32Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Uint32Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Uint32Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Uint32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Uint32Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Uint32Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Uint32Array;\n\n    [index: number]: number;\n}\n\ninterface Uint32ArrayConstructor {\n    readonly prototype: Uint32Array;\n    new(length: number): Uint32Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Uint32Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Uint32Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Uint32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Uint32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Uint32Array;\n\n}\ndeclare var Uint32Array: Uint32ArrayConstructor;\n\n/**\n * A typed array of 32-bit float values. The contents are initialized to 0. If the requested number\n * of bytes could not be allocated an exception is raised.\n */\ninterface Float32Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Float32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Float32Array) => any, thisArg?: any): Float32Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Float32Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Float32Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Float32Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Float32Array) => number, thisArg?: any): Float32Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float32Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float32Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float32Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Float32Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Float32Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Float32Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * Gets a new Float32Array view of the ArrayBuffer store for this array, referencing the elements\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Float32Array;\n\n    /**\n     * Converts a number to a string by using the current locale.\n     */\n    toLocaleString(): string;\n\n    /**\n     * Returns a string representation of an array.\n     */\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Float32Array;\n\n    [index: number]: number;\n}\n\ninterface Float32ArrayConstructor {\n    readonly prototype: Float32Array;\n    new(length: number): Float32Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Float32Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Float32Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Float32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Float32Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Float32Array;\n\n\n}\ndeclare var Float32Array: Float32ArrayConstructor;\n\n/**\n * A typed array of 64-bit float values. The contents are initialized to 0. If the requested\n * number of bytes could not be allocated an exception is raised.\n */\ninterface Float64Array {\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * The ArrayBuffer instance referenced by the array.\n     */\n    readonly buffer: ArrayBufferLike;\n\n    /**\n     * The length in bytes of the array.\n     */\n    readonly byteLength: number;\n\n    /**\n     * The offset in bytes of the array.\n     */\n    readonly byteOffset: number;\n\n    /**\n     * Returns the this object after copying a section of the array identified by start and end\n     * to the same array starting at position target\n     * @param target If target is negative, it is treated as length+target where length is the\n     * length of the array.\n     * @param start If start is negative, it is treated as length+start. If end is negative, it\n     * is treated as length+end.\n     * @param end If not specified, length of the this object is used as its default value.\n     */\n    copyWithin(target: number, start: number, end?: number): this;\n\n    /**\n     * Determines whether all the members of an array satisfy the specified test.\n     * @param predicate A function that accepts up to three arguments. The every method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value false, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    every(predicate: (value: number, index: number, array: Float64Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Returns the this object after filling the section identified by start and end with value\n     * @param value value to fill array section with\n     * @param start index to start filling the array at. If start is negative, it is treated as\n     * length+start where length is the length of the array.\n     * @param end index to stop filling the array at. If end is negative, it is treated as\n     * length+end.\n     */\n    fill(value: number, start?: number, end?: number): this;\n\n    /**\n     * Returns the elements of an array that meet the condition specified in a callback function.\n     * @param predicate A function that accepts up to three arguments. The filter method calls\n     * the predicate function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    filter(predicate: (value: number, index: number, array: Float64Array) => any, thisArg?: any): Float64Array;\n\n    /**\n     * Returns the value of the first element in the array where predicate is true, and undefined\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found, find\n     * immediately returns that element value. Otherwise, find returns undefined.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    find(predicate: (value: number, index: number, obj: Float64Array) => boolean, thisArg?: any): number | undefined;\n\n    /**\n     * Returns the index of the first element in the array where predicate is true, and -1\n     * otherwise.\n     * @param predicate find calls predicate once for each element of the array, in ascending\n     * order, until it finds one where predicate returns true. If such an element is found,\n     * findIndex immediately returns that element index. Otherwise, findIndex returns -1.\n     * @param thisArg If provided, it will be used as the this value for each invocation of\n     * predicate. If it is not provided, undefined is used instead.\n     */\n    findIndex(predicate: (value: number, index: number, obj: Float64Array) => boolean, thisArg?: any): number;\n\n    /**\n     * Performs the specified action for each element in an array.\n     * @param callbackfn  A function that accepts up to three arguments. forEach calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg  An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    forEach(callbackfn: (value: number, index: number, array: Float64Array) => void, thisArg?: any): void;\n\n    /**\n     * Returns the index of the first occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     *  search starts at index 0.\n     */\n    indexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * Adds all the elements of an array separated by the specified separator string.\n     * @param separator A string used to separate one element of an array from the next in the\n     * resulting String. If omitted, the array elements are separated with a comma.\n     */\n    join(separator?: string): string;\n\n    /**\n     * Returns the index of the last occurrence of a value in an array.\n     * @param searchElement The value to locate in the array.\n     * @param fromIndex The array index at which to begin the search. If fromIndex is omitted, the\n     * search starts at index 0.\n     */\n    lastIndexOf(searchElement: number, fromIndex?: number): number;\n\n    /**\n     * The length of the array.\n     */\n    readonly length: number;\n\n    /**\n     * Calls a defined callback function on each element of an array, and returns an array that\n     * contains the results.\n     * @param callbackfn A function that accepts up to three arguments. The map method calls the\n     * callbackfn function one time for each element in the array.\n     * @param thisArg An object to which the this keyword can refer in the callbackfn function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    map(callbackfn: (value: number, index: number, array: Float64Array) => number, thisArg?: any): Float64Array;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number): number;\n    reduce(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array. The return value of\n     * the callback function is the accumulated result, and is provided as an argument in the next\n     * call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduce method calls the\n     * callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduce<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float64Array) => U, initialValue: U): U;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an\n     * argument instead of an array value.\n     */\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number): number;\n    reduceRight(callbackfn: (previousValue: number, currentValue: number, currentIndex: number, array: Float64Array) => number, initialValue: number): number;\n\n    /**\n     * Calls the specified callback function for all the elements in an array, in descending order.\n     * The return value of the callback function is the accumulated result, and is provided as an\n     * argument in the next call to the callback function.\n     * @param callbackfn A function that accepts up to four arguments. The reduceRight method calls\n     * the callbackfn function one time for each element in the array.\n     * @param initialValue If initialValue is specified, it is used as the initial value to start\n     * the accumulation. The first call to the callbackfn function provides this value as an argument\n     * instead of an array value.\n     */\n    reduceRight<U>(callbackfn: (previousValue: U, currentValue: number, currentIndex: number, array: Float64Array) => U, initialValue: U): U;\n\n    /**\n     * Reverses the elements in an Array.\n     */\n    reverse(): Float64Array;\n\n    /**\n     * Sets a value or an array of values.\n     * @param array A typed or untyped array of values to set.\n     * @param offset The index in the current array at which the values are to be written.\n     */\n    set(array: ArrayLike<number>, offset?: number): void;\n\n    /**\n     * Returns a section of an array.\n     * @param start The beginning of the specified portion of the array.\n     * @param end The end of the specified portion of the array. This is exclusive of the element at the index 'end'.\n     */\n    slice(start?: number, end?: number): Float64Array;\n\n    /**\n     * Determines whether the specified callback function returns true for any element of an array.\n     * @param predicate A function that accepts up to three arguments. The some method calls\n     * the predicate function for each element in the array until the predicate returns a value\n     * which is coercible to the Boolean value true, or until the end of the array.\n     * @param thisArg An object to which the this keyword can refer in the predicate function.\n     * If thisArg is omitted, undefined is used as the this value.\n     */\n    some(predicate: (value: number, index: number, array: Float64Array) => unknown, thisArg?: any): boolean;\n\n    /**\n     * Sorts an array.\n     * @param compareFn Function used to determine the order of the elements. It is expected to return\n     * a negative value if first argument is less than second argument, zero if they're equal and a positive\n     * value otherwise. If omitted, the elements are sorted in ascending, ASCII character order.\n     * ```ts\n     * [11,2,22,1].sort((a, b) => a - b)\n     * ```\n     */\n    sort(compareFn?: (a: number, b: number) => number): this;\n\n    /**\n     * at begin, inclusive, up to end, exclusive.\n     * @param begin The index of the beginning of the array.\n     * @param end The index of the end of the array.\n     */\n    subarray(begin?: number, end?: number): Float64Array;\n\n    toString(): string;\n\n    /** Returns the primitive value of the specified object. */\n    valueOf(): Float64Array;\n\n    [index: number]: number;\n}\n\ninterface Float64ArrayConstructor {\n    readonly prototype: Float64Array;\n    new(length: number): Float64Array;\n    new(array: ArrayLike<number> | ArrayBufferLike): Float64Array;\n    new(buffer: ArrayBufferLike, byteOffset?: number, length?: number): Float64Array;\n\n    /**\n     * The size in bytes of each element in the array.\n     */\n    readonly BYTES_PER_ELEMENT: number;\n\n    /**\n     * Returns a new array from a set of elements.\n     * @param items A set of elements to include in the new array object.\n     */\n    of(...items: number[]): Float64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     */\n    from(arrayLike: ArrayLike<number>): Float64Array;\n\n    /**\n     * Creates an array from an array-like or iterable object.\n     * @param arrayLike An array-like or iterable object to convert to an array.\n     * @param mapfn A mapping function to call on every element of the array.\n     * @param thisArg Value of 'this' used to invoke the mapfn.\n     */\n    from<T>(arrayLike: ArrayLike<T>, mapfn: (v: T, k: number) => number, thisArg?: any): Float64Array;\n\n}\ndeclare var Float64Array: Float64ArrayConstructor;\n\n/////////////////////////////\n/// ECMAScript Internationalization API\n/////////////////////////////\n\ndeclare namespace Intl {\n    interface CollatorOptions {\n        usage?: string;\n        localeMatcher?: string;\n        numeric?: boolean;\n        caseFirst?: string;\n        sensitivity?: string;\n        ignorePunctuation?: boolean;\n    }\n\n    interface ResolvedCollatorOptions {\n        locale: string;\n        usage: string;\n        sensitivity: string;\n        ignorePunctuation: boolean;\n        collation: string;\n        caseFirst: string;\n        numeric: boolean;\n    }\n\n    interface Collator {\n        compare(x: string, y: string): number;\n        resolvedOptions(): ResolvedCollatorOptions;\n    }\n    var Collator: {\n        new(locales?: string | string[], options?: CollatorOptions): Collator;\n        (locales?: string | string[], options?: CollatorOptions): Collator;\n        supportedLocalesOf(locales: string | string[], options?: CollatorOptions): string[];\n    };\n\n    interface NumberFormatOptions {\n        localeMatcher?: string;\n        style?: string;\n        currency?: string;\n        currencyDisplay?: string;\n        currencySign?: string;\n        useGrouping?: boolean;\n        minimumIntegerDigits?: number;\n        minimumFractionDigits?: number;\n        maximumFractionDigits?: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n    }\n\n    interface ResolvedNumberFormatOptions {\n        locale: string;\n        numberingSystem: string;\n        style: string;\n        currency?: string;\n        currencyDisplay?: string;\n        minimumIntegerDigits: number;\n        minimumFractionDigits: number;\n        maximumFractionDigits: number;\n        minimumSignificantDigits?: number;\n        maximumSignificantDigits?: number;\n        useGrouping: boolean;\n    }\n\n    interface NumberFormat {\n        format(value: number): string;\n        resolvedOptions(): ResolvedNumberFormatOptions;\n    }\n    var NumberFormat: {\n        new(locales?: string | string[], options?: NumberFormatOptions): NumberFormat;\n        (locales?: string | string[], options?: NumberFormatOptions): NumberFormat;\n        supportedLocalesOf(locales: string | string[], options?: NumberFormatOptions): string[];\n    };\n\n    interface DateTimeFormatOptions {\n        localeMatcher?: string;\n        weekday?: string;\n        era?: string;\n        year?: string;\n        month?: string;\n        day?: string;\n        hour?: string;\n        minute?: string;\n        second?: string;\n        timeZoneName?: string;\n        formatMatcher?: string;\n        hour12?: boolean;\n        timeZone?: string;\n    }\n\n    interface ResolvedDateTimeFormatOptions {\n        locale: string;\n        calendar: string;\n        numberingSystem: string;\n        timeZone: string;\n        hour12?: boolean;\n        weekday?: string;\n        era?: string;\n        year?: string;\n        month?: string;\n        day?: string;\n        hour?: string;\n        minute?: string;\n        second?: string;\n        timeZoneName?: string;\n    }\n\n    interface DateTimeFormat {\n        format(date?: Date | number): string;\n        resolvedOptions(): ResolvedDateTimeFormatOptions;\n    }\n    var DateTimeFormat: {\n        new(locales?: string | string[], options?: DateTimeFormatOptions): DateTimeFormat;\n        (locales?: string | string[], options?: DateTimeFormatOptions): DateTimeFormat;\n        supportedLocalesOf(locales: string | string[], options?: DateTimeFormatOptions): string[];\n    };\n}\n\ninterface String {\n    /**\n     * Determines whether two strings are equivalent in the current or specified locale.\n     * @param that String to compare to target string\n     * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used. This parameter must conform to BCP 47 standards; see the Intl.Collator object for details.\n     * @param options An object that contains one or more properties that specify comparison options. see the Intl.Collator object for details.\n     */\n    localeCompare(that: string, locales?: string | string[], options?: Intl.CollatorOptions): number;\n}\n\ninterface Number {\n    /**\n     * Converts a number to a string by using the current or specified locale.\n     * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n     * @param options An object that contains one or more properties that specify comparison options.\n     */\n    toLocaleString(locales?: string | string[], options?: Intl.NumberFormatOptions): string;\n}\n\ninterface Date {\n    /**\n     * Converts a date and time to a string by using the current or specified locale.\n     * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n     * @param options An object that contains one or more properties that specify comparison options.\n     */\n    toLocaleString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n    /**\n     * Converts a date to a string by using the current or specified locale.\n     * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n     * @param options An object that contains one or more properties that specify comparison options.\n     */\n    toLocaleDateString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n\n    /**\n     * Converts a time to a string by using the current or specified locale.\n     * @param locales A locale string or array of locale strings that contain one or more language or locale tags. If you include more than one locale string, list them in descending order of priority so that the first entry is the preferred locale. If you omit this parameter, the default locale of the JavaScript runtime is used.\n     * @param options An object that contains one or more properties that specify comparison options.\n     */\n    toLocaleTimeString(locales?: string | string[], options?: Intl.DateTimeFormatOptions): string;\n}\n",n.libFileMap["lib.es6.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2015" />\n/// <reference lib="dom" />\n/// <reference lib="dom.iterable" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n',n.libFileMap["lib.esnext.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="es2020" />\n/// <reference lib="esnext.intl" />\n/// <reference lib="esnext.string" />\n/// <reference lib="esnext.promise" />\n/// <reference lib="esnext.weakref" />\n',n.libFileMap["lib.esnext.full.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/// <reference lib="esnext" />\n/// <reference lib="dom" />\n/// <reference lib="webworker.importscripts" />\n/// <reference lib="scripthost" />\n/// <reference lib="dom.iterable" />',n.libFileMap["lib.esnext.intl.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ndeclare namespace Intl {\n    type NumberFormatPartTypes = "currency" | "decimal" | "fraction" | "group" | "infinity" | "integer" | "literal" | "minusSign" | "nan" | "plusSign" | "percentSign";\n\n    interface NumberFormatPart {\n        type: NumberFormatPartTypes;\n        value: string;\n    }\n\n    interface NumberFormat {\n        formatToParts(number?: number): NumberFormatPart[];\n    }\n}\n',n.libFileMap["lib.esnext.promise.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface AggregateError extends Error {\n    errors: any[]\n}\n\ninterface AggregateErrorConstructor {\n    new(errors: Iterable<any>, message?: string): AggregateError;\n    (errors: Iterable<any>, message?: string): AggregateError;\n    readonly prototype: AggregateError;\n}\n\ndeclare var AggregateError: AggregateErrorConstructor;\n\n/**\n * Represents the completion of an asynchronous operation\n */\ninterface PromiseConstructor {\n    /**\n     * The any function returns a promise that is fulfilled by the first given promise to be fulfilled, or rejected with an AggregateError containing an array of rejection reasons if all of the given promises are rejected. It resolves all elements of the passed iterable to promises as it runs this algorithm.\n     * @param values An array or iterable of Promises.\n     * @returns A new Promise.\n     */\n    any<T>(values: (T | PromiseLike<T>)[] | Iterable<T | PromiseLike<T>>): Promise<T>\n}\n',n.libFileMap["lib.esnext.string.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface String {\n    /**\n     * Replace all instances of a substring in a string, using a regular expression or search string.\n     * @param searchValue A string to search for.\n     * @param replaceValue A string containing the text to replace for every successful match of searchValue in this string.\n     */\n    replaceAll(searchValue: string | RegExp, replaceValue: string): string;\n\n    /**\n     * Replace all instances of a substring in a string, using a regular expression or search string.\n     * @param searchValue A string to search for.\n     * @param replacer A function that returns the replacement text.\n     */\n    replaceAll(searchValue: string | RegExp, replacer: (substring: string, ...args: any[]) => string): string;\n}\n',n.libFileMap["lib.esnext.weakref.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\ninterface WeakRef<T extends object> {\n    readonly [Symbol.toStringTag]: "WeakRef";\n\n    /**\n     * Returns the WeakRef instance\'s target object, or undefined if the target object has been\n     * reclaimed.\n     */\n    deref(): T | undefined;\n}\n\ninterface WeakRefConstructor {\n    readonly prototype: WeakRef<any>;\n\n    /**\n     * Creates a WeakRef instance for the given target object.\n     * @param target The target object for the WeakRef instance.\n     */\n    new<T extends object>(target?: T): WeakRef<T>;\n}\n\ndeclare var WeakRef: WeakRefConstructor;\n\ninterface FinalizationRegistry {\n    readonly [Symbol.toStringTag]: "FinalizationRegistry";\n\n    /**\n     * Registers an object with the registry.\n     * @param target The target object to register.\n     * @param heldValue The value to pass to the finalizer for this object. This cannot be the\n     * target object.\n     * @param unregisterToken The token to pass to the unregister method to unregister the target\n     * object. If provided (and not undefined), this must be an object. If not provided, the target\n     * cannot be unregistered.\n     */\n    register(target: object, heldValue: any, unregisterToken?: object): void;\n\n    /**\n     * Unregisters an object from the registry.\n     * @param unregisterToken The token that was used as the unregisterToken argument when calling\n     * register to register the target object.\n     */\n    unregister(unregisterToken: object): void;\n}\n\ninterface FinalizationRegistryConstructor {\n    readonly prototype: FinalizationRegistry;\n\n    /**\n     * Creates a finalization registry with an associated cleanup callback\n     * @param cleanupCallback The callback to call after an object in the registry has been reclaimed.\n     */\n    new(cleanupCallback: (heldValue: any) => void): FinalizationRegistry;\n}\n\ndeclare var FinalizationRegistry: FinalizationRegistryConstructor;\n',n.libFileMap["lib.scripthost.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n\n\n/////////////////////////////\n/// Windows Script Host APIS\n/////////////////////////////\n\n\ninterface ActiveXObject {\n    new (s: string): any;\n}\ndeclare var ActiveXObject: ActiveXObject;\n\ninterface ITextWriter {\n    Write(s: string): void;\n    WriteLine(s: string): void;\n    Close(): void;\n}\n\ninterface TextStreamBase {\n    /**\n     * The column number of the current character position in an input stream.\n     */\n    Column: number;\n\n    /**\n     * The current line number in an input stream.\n     */\n    Line: number;\n\n    /**\n     * Closes a text stream.\n     * It is not necessary to close standard streams; they close automatically when the process ends. If\n     * you close a standard stream, be aware that any other pointers to that standard stream become invalid.\n     */\n    Close(): void;\n}\n\ninterface TextStreamWriter extends TextStreamBase {\n    /**\n     * Sends a string to an output stream.\n     */\n    Write(s: string): void;\n\n    /**\n     * Sends a specified number of blank lines (newline characters) to an output stream.\n     */\n    WriteBlankLines(intLines: number): void;\n\n    /**\n     * Sends a string followed by a newline character to an output stream.\n     */\n    WriteLine(s: string): void;\n}\n\ninterface TextStreamReader extends TextStreamBase {\n    /**\n     * Returns a specified number of characters from an input stream, starting at the current pointer position.\n     * Does not return until the ENTER key is pressed.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    Read(characters: number): string;\n\n    /**\n     * Returns all characters from an input stream.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    ReadAll(): string;\n\n    /**\n     * Returns an entire line from an input stream.\n     * Although this method extracts the newline character, it does not add it to the returned string.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     */\n    ReadLine(): string;\n\n    /**\n     * Skips a specified number of characters when reading from an input text stream.\n     * Can only be used on a stream in reading mode; causes an error in writing or appending mode.\n     * @param characters Positive number of characters to skip forward. (Backward skipping is not supported.)\n     */\n    Skip(characters: number): void;\n\n    /**\n     * Skips the next line when reading from an input text stream.\n     * Can only be used on a stream in reading mode, not writing or appending mode.\n     */\n    SkipLine(): void;\n\n    /**\n     * Indicates whether the stream pointer position is at the end of a line.\n     */\n    AtEndOfLine: boolean;\n\n    /**\n     * Indicates whether the stream pointer position is at the end of a stream.\n     */\n    AtEndOfStream: boolean;\n}\n\ndeclare var WScript: {\n    /**\n     * Outputs text to either a message box (under WScript.exe) or the command console window followed by\n     * a newline (under CScript.exe).\n     */\n    Echo(s: any): void;\n\n    /**\n     * Exposes the write-only error output stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdErr: TextStreamWriter;\n\n    /**\n     * Exposes the write-only output stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdOut: TextStreamWriter;\n    Arguments: { length: number; Item(n: number): string; };\n\n    /**\n     *  The full path of the currently running script.\n     */\n    ScriptFullName: string;\n\n    /**\n     * Forces the script to stop immediately, with an optional exit code.\n     */\n    Quit(exitCode?: number): number;\n\n    /**\n     * The Windows Script Host build version number.\n     */\n    BuildVersion: number;\n\n    /**\n     * Fully qualified path of the host executable.\n     */\n    FullName: string;\n\n    /**\n     * Gets/sets the script mode - interactive(true) or batch(false).\n     */\n    Interactive: boolean;\n\n    /**\n     * The name of the host executable (WScript.exe or CScript.exe).\n     */\n    Name: string;\n\n    /**\n     * Path of the directory containing the host executable.\n     */\n    Path: string;\n\n    /**\n     * The filename of the currently running script.\n     */\n    ScriptName: string;\n\n    /**\n     * Exposes the read-only input stream for the current script.\n     * Can be accessed only while using CScript.exe.\n     */\n    StdIn: TextStreamReader;\n\n    /**\n     * Windows Script Host version\n     */\n    Version: string;\n\n    /**\n     * Connects a COM object\'s event sources to functions named with a given prefix, in the form prefix_event.\n     */\n    ConnectObject(objEventSource: any, strPrefix: string): void;\n\n    /**\n     * Creates a COM object.\n     * @param strProgiID\n     * @param strPrefix Function names in the form prefix_event will be bound to this object\'s COM events.\n     */\n    CreateObject(strProgID: string, strPrefix?: string): any;\n\n    /**\n     * Disconnects a COM object from its event sources.\n     */\n    DisconnectObject(obj: any): void;\n\n    /**\n     * Retrieves an existing object with the specified ProgID from memory, or creates a new one from a file.\n     * @param strPathname Fully qualified path to the file containing the object persisted to disk.\n     *                       For objects in memory, pass a zero-length string.\n     * @param strProgID\n     * @param strPrefix Function names in the form prefix_event will be bound to this object\'s COM events.\n     */\n    GetObject(strPathname: string, strProgID?: string, strPrefix?: string): any;\n\n    /**\n     * Suspends script execution for a specified length of time, then continues execution.\n     * @param intTime Interval (in milliseconds) to suspend script execution.\n     */\n    Sleep(intTime: number): void;\n};\n\n/**\n * WSH is an alias for WScript under Windows Script Host\n */\ndeclare var WSH: typeof WScript;\n\n/**\n * Represents an Automation SAFEARRAY\n */\ndeclare class SafeArray<T = any> {\n    private constructor();\n    private SafeArray_typekey: SafeArray<T>;\n}\n\n/**\n * Allows enumerating over a COM collection, which may not have indexed item access.\n */\ninterface Enumerator<T = any> {\n    /**\n     * Returns true if the current item is the last one in the collection, or the collection is empty,\n     * or the current item is undefined.\n     */\n    atEnd(): boolean;\n\n    /**\n     * Returns the current item in the collection\n     */\n    item(): T;\n\n    /**\n     * Resets the current item in the collection to the first item. If there are no items in the collection,\n     * the current item is set to undefined.\n     */\n    moveFirst(): void;\n\n    /**\n     * Moves the current item to the next item in the collection. If the enumerator is at the end of\n     * the collection or the collection is empty, the current item is set to undefined.\n     */\n    moveNext(): void;\n}\n\ninterface EnumeratorConstructor {\n    new <T = any>(safearray: SafeArray<T>): Enumerator<T>;\n    new <T = any>(collection: { Item(index: any): T }): Enumerator<T>;\n    new <T = any>(collection: any): Enumerator<T>;\n}\n\ndeclare var Enumerator: EnumeratorConstructor;\n\n/**\n * Enables reading from a COM safe array, which might have an alternate lower bound, or multiple dimensions.\n */\ninterface VBArray<T = any> {\n    /**\n     * Returns the number of dimensions (1-based).\n     */\n    dimensions(): number;\n\n    /**\n     * Takes an index for each dimension in the array, and returns the item at the corresponding location.\n     */\n    getItem(dimension1Index: number, ...dimensionNIndexes: number[]): T;\n\n    /**\n     * Returns the smallest available index for a given dimension.\n     * @param dimension 1-based dimension (defaults to 1)\n     */\n    lbound(dimension?: number): number;\n\n    /**\n     * Returns the largest available index for a given dimension.\n     * @param dimension 1-based dimension (defaults to 1)\n     */\n    ubound(dimension?: number): number;\n\n    /**\n     * Returns a Javascript array with all the elements in the VBArray. If there are multiple dimensions,\n     * each successive dimension is appended to the end of the array.\n     * Example: [[1,2,3],[4,5,6]] becomes [1,2,3,4,5,6]\n     */\n    toArray(): T[];\n}\n\ninterface VBArrayConstructor {\n    new <T = any>(safeArray: SafeArray<T>): VBArray<T>;\n}\n\ndeclare var VBArray: VBArrayConstructor;\n\n/**\n * Automation date (VT_DATE)\n */\ndeclare class VarDate {\n    private constructor();\n    private VarDate_typekey: VarDate;\n}\n\ninterface DateConstructor {\n    new (vd: VarDate): Date;\n}\n\ninterface Date {\n    getVarDate: () => VarDate;\n}\n',n.libFileMap["lib.webworker.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// Worker APIs\n/////////////////////////////\n\ninterface AddEventListenerOptions extends EventListenerOptions {\n    once?: boolean;\n    passive?: boolean;\n}\n\ninterface AesCbcParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCtrParams extends Algorithm {\n    counter: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    length: number;\n}\n\ninterface AesDerivedKeyParams extends Algorithm {\n    length: number;\n}\n\ninterface AesGcmParams extends Algorithm {\n    additionalData?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    tagLength?: number;\n}\n\ninterface AesKeyAlgorithm extends KeyAlgorithm {\n    length: number;\n}\n\ninterface AesKeyGenParams extends Algorithm {\n    length: number;\n}\n\ninterface Algorithm {\n    name: string;\n}\n\ninterface BlobPropertyBag {\n    endings?: EndingType;\n    type?: string;\n}\n\ninterface CacheQueryOptions {\n    ignoreMethod?: boolean;\n    ignoreSearch?: boolean;\n    ignoreVary?: boolean;\n}\n\ninterface CanvasRenderingContext2DSettings {\n    alpha?: boolean;\n    desynchronized?: boolean;\n}\n\ninterface ClientQueryOptions {\n    includeUncontrolled?: boolean;\n    type?: ClientTypes;\n}\n\ninterface CloseEventInit extends EventInit {\n    code?: number;\n    reason?: string;\n    wasClean?: boolean;\n}\n\ninterface CryptoKeyPair {\n    privateKey?: CryptoKey;\n    publicKey?: CryptoKey;\n}\n\ninterface CustomEventInit<T = any> extends EventInit {\n    detail?: T;\n}\n\ninterface DOMMatrix2DInit {\n    a?: number;\n    b?: number;\n    c?: number;\n    d?: number;\n    e?: number;\n    f?: number;\n    m11?: number;\n    m12?: number;\n    m21?: number;\n    m22?: number;\n    m41?: number;\n    m42?: number;\n}\n\ninterface DOMMatrixInit extends DOMMatrix2DInit {\n    is2D?: boolean;\n    m13?: number;\n    m14?: number;\n    m23?: number;\n    m24?: number;\n    m31?: number;\n    m32?: number;\n    m33?: number;\n    m34?: number;\n    m43?: number;\n    m44?: number;\n}\n\ninterface DOMPointInit {\n    w?: number;\n    x?: number;\n    y?: number;\n    z?: number;\n}\n\ninterface DOMQuadInit {\n    p1?: DOMPointInit;\n    p2?: DOMPointInit;\n    p3?: DOMPointInit;\n    p4?: DOMPointInit;\n}\n\ninterface DOMRectInit {\n    height?: number;\n    width?: number;\n    x?: number;\n    y?: number;\n}\n\ninterface DevicePermissionDescriptor extends PermissionDescriptor {\n    deviceId?: string;\n    name: "camera" | "microphone" | "speaker";\n}\n\ninterface EcKeyGenParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcKeyImportParams extends Algorithm {\n    namedCurve: NamedCurve;\n}\n\ninterface EcdhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface EcdsaParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface ErrorEventInit extends EventInit {\n    colno?: number;\n    error?: any;\n    filename?: string;\n    lineno?: number;\n    message?: string;\n}\n\ninterface EventInit {\n    bubbles?: boolean;\n    cancelable?: boolean;\n    composed?: boolean;\n}\n\ninterface EventListenerOptions {\n    capture?: boolean;\n}\n\ninterface EventSourceInit {\n    withCredentials?: boolean;\n}\n\ninterface ExtendableEventInit extends EventInit {\n}\n\ninterface ExtendableMessageEventInit extends ExtendableEventInit {\n    data?: any;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[];\n    source?: Client | ServiceWorker | MessagePort | null;\n}\n\ninterface FetchEventInit extends ExtendableEventInit {\n    clientId?: string;\n    preloadResponse?: Promise<any>;\n    replacesClientId?: string;\n    request: Request;\n    resultingClientId?: string;\n}\n\ninterface FilePropertyBag extends BlobPropertyBag {\n    lastModified?: number;\n}\n\ninterface GetNotificationOptions {\n    tag?: string;\n}\n\ninterface HkdfParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    info: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface HmacImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface HmacKeyGenParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    length?: number;\n}\n\ninterface IDBIndexParameters {\n    multiEntry?: boolean;\n    unique?: boolean;\n}\n\ninterface IDBObjectStoreParameters {\n    autoIncrement?: boolean;\n    keyPath?: string | string[] | null;\n}\n\ninterface IDBVersionChangeEventInit extends EventInit {\n    newVersion?: number | null;\n    oldVersion?: number;\n}\n\ninterface ImageBitmapOptions {\n    colorSpaceConversion?: ColorSpaceConversion;\n    imageOrientation?: ImageOrientation;\n    premultiplyAlpha?: PremultiplyAlpha;\n    resizeHeight?: number;\n    resizeQuality?: ResizeQuality;\n    resizeWidth?: number;\n}\n\ninterface ImageBitmapRenderingContextSettings {\n    alpha?: boolean;\n}\n\ninterface ImageEncodeOptions {\n    quality?: number;\n    type?: string;\n}\n\ninterface JsonWebKey {\n    alg?: string;\n    crv?: string;\n    d?: string;\n    dp?: string;\n    dq?: string;\n    e?: string;\n    ext?: boolean;\n    k?: string;\n    key_ops?: string[];\n    kty?: string;\n    n?: string;\n    oth?: RsaOtherPrimesInfo[];\n    p?: string;\n    q?: string;\n    qi?: string;\n    use?: string;\n    x?: string;\n    y?: string;\n}\n\ninterface KeyAlgorithm {\n    name: string;\n}\n\ninterface MessageEventInit<T = any> extends EventInit {\n    data?: T;\n    lastEventId?: string;\n    origin?: string;\n    ports?: MessagePort[];\n    source?: MessageEventSource | null;\n}\n\ninterface MidiPermissionDescriptor extends PermissionDescriptor {\n    name: "midi";\n    sysex?: boolean;\n}\n\ninterface MultiCacheQueryOptions extends CacheQueryOptions {\n    cacheName?: string;\n}\n\ninterface NavigationPreloadState {\n    enabled?: boolean;\n    headerValue?: string;\n}\n\ninterface NotificationAction {\n    action: string;\n    icon?: string;\n    title: string;\n}\n\ninterface NotificationEventInit extends ExtendableEventInit {\n    action?: string;\n    notification: Notification;\n}\n\ninterface NotificationOptions {\n    actions?: NotificationAction[];\n    badge?: string;\n    body?: string;\n    data?: any;\n    dir?: NotificationDirection;\n    icon?: string;\n    image?: string;\n    lang?: string;\n    renotify?: boolean;\n    requireInteraction?: boolean;\n    silent?: boolean;\n    tag?: string;\n    timestamp?: number;\n    vibrate?: VibratePattern;\n}\n\ninterface Pbkdf2Params extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n    iterations: number;\n    salt: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface PerformanceObserverInit {\n    buffered?: boolean;\n    entryTypes?: string[];\n    type?: string;\n}\n\ninterface PermissionDescriptor {\n    name: PermissionName;\n}\n\ninterface PipeOptions {\n    preventAbort?: boolean;\n    preventCancel?: boolean;\n    preventClose?: boolean;\n    signal?: AbortSignal;\n}\n\ninterface PostMessageOptions {\n    transfer?: any[];\n}\n\ninterface ProgressEventInit extends EventInit {\n    lengthComputable?: boolean;\n    loaded?: number;\n    total?: number;\n}\n\ninterface PromiseRejectionEventInit extends EventInit {\n    promise: Promise<any>;\n    reason?: any;\n}\n\ninterface PushEventInit extends ExtendableEventInit {\n    data?: PushMessageDataInit;\n}\n\ninterface PushPermissionDescriptor extends PermissionDescriptor {\n    name: "push";\n    userVisibleOnly?: boolean;\n}\n\ninterface PushSubscriptionChangeEventInit extends ExtendableEventInit {\n    newSubscription?: PushSubscription;\n    oldSubscription?: PushSubscription;\n}\n\ninterface PushSubscriptionJSON {\n    endpoint?: string;\n    expirationTime?: number | null;\n    keys?: Record<string, string>;\n}\n\ninterface PushSubscriptionOptionsInit {\n    applicationServerKey?: BufferSource | string | null;\n    userVisibleOnly?: boolean;\n}\n\ninterface QueuingStrategy<T = any> {\n    highWaterMark?: number;\n    size?: QueuingStrategySizeCallback<T>;\n}\n\ninterface ReadableStreamReadDoneResult<T> {\n    done: true;\n    value?: T;\n}\n\ninterface ReadableStreamReadValueResult<T> {\n    done: false;\n    value: T;\n}\n\ninterface RegistrationOptions {\n    scope?: string;\n    type?: WorkerType;\n    updateViaCache?: ServiceWorkerUpdateViaCache;\n}\n\ninterface RequestInit {\n    /**\n     * A BodyInit object or null to set request\'s body.\n     */\n    body?: BodyInit | null;\n    /**\n     * A string indicating how the request will interact with the browser\'s cache to set request\'s cache.\n     */\n    cache?: RequestCache;\n    /**\n     * A string indicating whether credentials will be sent with the request always, never, or only when sent to a same-origin URL. Sets request\'s credentials.\n     */\n    credentials?: RequestCredentials;\n    /**\n     * A Headers object, an object literal, or an array of two-item arrays to set request\'s headers.\n     */\n    headers?: HeadersInit;\n    /**\n     * A cryptographic hash of the resource to be fetched by request. Sets request\'s integrity.\n     */\n    integrity?: string;\n    /**\n     * A boolean to set request\'s keepalive.\n     */\n    keepalive?: boolean;\n    /**\n     * A string to set request\'s method.\n     */\n    method?: string;\n    /**\n     * A string to indicate whether the request will use CORS, or will be restricted to same-origin URLs. Sets request\'s mode.\n     */\n    mode?: RequestMode;\n    /**\n     * A string indicating whether request follows redirects, results in an error upon encountering a redirect, or returns the redirect (in an opaque fashion). Sets request\'s redirect.\n     */\n    redirect?: RequestRedirect;\n    /**\n     * A string whose value is a same-origin URL, "about:client", or the empty string, to set request\'s referrer.\n     */\n    referrer?: string;\n    /**\n     * A referrer policy to set request\'s referrerPolicy.\n     */\n    referrerPolicy?: ReferrerPolicy;\n    /**\n     * An AbortSignal to set request\'s signal.\n     */\n    signal?: AbortSignal | null;\n    /**\n     * Can only be null. Used to disassociate request from any Window.\n     */\n    window?: any;\n}\n\ninterface ResponseInit {\n    headers?: HeadersInit;\n    status?: number;\n    statusText?: string;\n}\n\ninterface RsaHashedImportParams extends Algorithm {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaHashedKeyGenParams extends RsaKeyGenParams {\n    hash: HashAlgorithmIdentifier;\n}\n\ninterface RsaKeyGenParams extends Algorithm {\n    modulusLength: number;\n    publicExponent: BigInteger;\n}\n\ninterface RsaOaepParams extends Algorithm {\n    label?: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface RsaOtherPrimesInfo {\n    d?: string;\n    r?: string;\n    t?: string;\n}\n\ninterface RsaPssParams extends Algorithm {\n    saltLength: number;\n}\n\ninterface StorageEstimate {\n    quota?: number;\n    usage?: number;\n}\n\ninterface SyncEventInit extends ExtendableEventInit {\n    lastChance?: boolean;\n    tag: string;\n}\n\ninterface TextDecodeOptions {\n    stream?: boolean;\n}\n\ninterface TextDecoderOptions {\n    fatal?: boolean;\n    ignoreBOM?: boolean;\n}\n\ninterface TextEncoderEncodeIntoResult {\n    read?: number;\n    written?: number;\n}\n\ninterface Transformer<I = any, O = any> {\n    flush?: TransformStreamDefaultControllerCallback<O>;\n    readableType?: undefined;\n    start?: TransformStreamDefaultControllerCallback<O>;\n    transform?: TransformStreamDefaultControllerTransformCallback<I, O>;\n    writableType?: undefined;\n}\n\ninterface UnderlyingByteSource {\n    autoAllocateChunkSize?: number;\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableByteStreamControllerCallback;\n    start?: ReadableByteStreamControllerCallback;\n    type: "bytes";\n}\n\ninterface UnderlyingSink<W = any> {\n    abort?: WritableStreamErrorCallback;\n    close?: WritableStreamDefaultControllerCloseCallback;\n    start?: WritableStreamDefaultControllerStartCallback;\n    type?: undefined;\n    write?: WritableStreamDefaultControllerWriteCallback<W>;\n}\n\ninterface UnderlyingSource<R = any> {\n    cancel?: ReadableStreamErrorCallback;\n    pull?: ReadableStreamDefaultControllerCallback<R>;\n    start?: ReadableStreamDefaultControllerCallback<R>;\n    type?: undefined;\n}\n\ninterface WebGLContextAttributes {\n    alpha?: boolean;\n    antialias?: boolean;\n    depth?: boolean;\n    desynchronized?: boolean;\n    failIfMajorPerformanceCaveat?: boolean;\n    powerPreference?: WebGLPowerPreference;\n    premultipliedAlpha?: boolean;\n    preserveDrawingBuffer?: boolean;\n    stencil?: boolean;\n}\n\ninterface WebGLContextEventInit extends EventInit {\n    statusMessage?: string;\n}\n\ninterface WorkerOptions {\n    credentials?: RequestCredentials;\n    name?: string;\n    type?: WorkerType;\n}\n\ninterface EventListener {\n    (evt: Event): void;\n}\n\n/** The ANGLE_instanced_arrays extension is part of the WebGL API and allows to draw the same object, or groups of similar objects multiple times, if they share the same vertex data, primitive count and type. */\ninterface ANGLE_instanced_arrays {\n    drawArraysInstancedANGLE(mode: GLenum, first: GLint, count: GLsizei, primcount: GLsizei): void;\n    drawElementsInstancedANGLE(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr, primcount: GLsizei): void;\n    vertexAttribDivisorANGLE(index: GLuint, divisor: GLuint): void;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR_ANGLE: GLenum;\n}\n\n/** A controller object that allows you to abort one or more DOM requests as and when desired. */\ninterface AbortController {\n    /**\n     * Returns the AbortSignal object associated with this object.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Invoking this method will set this object\'s AbortSignal\'s aborted flag and signal to any observers that the associated activity is to be aborted.\n     */\n    abort(): void;\n}\n\ndeclare var AbortController: {\n    prototype: AbortController;\n    new(): AbortController;\n};\n\ninterface AbortSignalEventMap {\n    "abort": Event;\n}\n\n/** A signal object that allows you to communicate with a DOM request (such as a Fetch) and abort it if required via an AbortController object. */\ninterface AbortSignal extends EventTarget {\n    /**\n     * Returns true if this AbortSignal\'s AbortController has signaled to abort, and false otherwise.\n     */\n    readonly aborted: boolean;\n    onabort: ((this: AbortSignal, ev: Event) => any) | null;\n    addEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbortSignalEventMap>(type: K, listener: (this: AbortSignal, ev: AbortSignalEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var AbortSignal: {\n    prototype: AbortSignal;\n    new(): AbortSignal;\n};\n\ninterface AbstractWorkerEventMap {\n    "error": ErrorEvent;\n}\n\ninterface AbstractWorker {\n    onerror: ((this: AbstractWorker, ev: ErrorEvent) => any) | null;\n    addEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: AbstractWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ninterface AesCfbParams extends Algorithm {\n    iv: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer;\n}\n\ninterface AesCmacParams extends Algorithm {\n    length: number;\n}\n\ninterface AnimationFrameProvider {\n    cancelAnimationFrame(handle: number): void;\n    requestAnimationFrame(callback: FrameRequestCallback): number;\n}\n\n/** A file-like object of immutable, raw data. Blobs represent data that isn\'t necessarily in a JavaScript-native format. The File interface is based on Blob, inheriting blob functionality and expanding it to support files on the user\'s system. */\ninterface Blob {\n    readonly size: number;\n    readonly type: string;\n    arrayBuffer(): Promise<ArrayBuffer>;\n    slice(start?: number, end?: number, contentType?: string): Blob;\n    stream(): ReadableStream;\n    text(): Promise<string>;\n}\n\ndeclare var Blob: {\n    prototype: Blob;\n    new(blobParts?: BlobPart[], options?: BlobPropertyBag): Blob;\n};\n\ninterface Body {\n    readonly body: ReadableStream<Uint8Array> | null;\n    readonly bodyUsed: boolean;\n    arrayBuffer(): Promise<ArrayBuffer>;\n    blob(): Promise<Blob>;\n    formData(): Promise<FormData>;\n    json(): Promise<any>;\n    text(): Promise<string>;\n}\n\ninterface BroadcastChannelEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\ninterface BroadcastChannel extends EventTarget {\n    /**\n     * Returns the channel name (as passed to the constructor).\n     */\n    readonly name: string;\n    onmessage: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: BroadcastChannel, ev: MessageEvent) => any) | null;\n    /**\n     * Closes the BroadcastChannel object, opening it up to garbage collection.\n     */\n    close(): void;\n    /**\n     * Sends the given message to other BroadcastChannel objects set up for this channel. Messages can be structured objects, e.g. nested objects and arrays.\n     */\n    postMessage(message: any): void;\n    addEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof BroadcastChannelEventMap>(type: K, listener: (this: BroadcastChannel, ev: BroadcastChannelEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var BroadcastChannel: {\n    prototype: BroadcastChannel;\n    new(name: string): BroadcastChannel;\n};\n\n/** This Streams API interface provides a built-in byte length queuing strategy that can be used when constructing streams. */\ninterface ByteLengthQueuingStrategy extends QueuingStrategy<ArrayBufferView> {\n    highWaterMark: number;\n    size(chunk: ArrayBufferView): number;\n}\n\ndeclare var ByteLengthQueuingStrategy: {\n    prototype: ByteLengthQueuingStrategy;\n    new(options: { highWaterMark: number }): ByteLengthQueuingStrategy;\n};\n\n/** Provides a storage mechanism for Request / Response object pairs that are cached, for example as part of the ServiceWorker life cycle. Note that the Cache interface is exposed to windowed scopes as well as workers. You don\'t have to use it in conjunction with service workers, even though it is defined in the service worker spec. */\ninterface Cache {\n    add(request: RequestInfo): Promise<void>;\n    addAll(requests: RequestInfo[]): Promise<void>;\n    delete(request: RequestInfo, options?: CacheQueryOptions): Promise<boolean>;\n    keys(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Request>>;\n    match(request: RequestInfo, options?: CacheQueryOptions): Promise<Response | undefined>;\n    matchAll(request?: RequestInfo, options?: CacheQueryOptions): Promise<ReadonlyArray<Response>>;\n    put(request: RequestInfo, response: Response): Promise<void>;\n}\n\ndeclare var Cache: {\n    prototype: Cache;\n    new(): Cache;\n};\n\n/** The storage for Cache objects. */\ninterface CacheStorage {\n    delete(cacheName: string): Promise<boolean>;\n    has(cacheName: string): Promise<boolean>;\n    keys(): Promise<string[]>;\n    match(request: RequestInfo, options?: MultiCacheQueryOptions): Promise<Response | undefined>;\n    open(cacheName: string): Promise<Cache>;\n}\n\ndeclare var CacheStorage: {\n    prototype: CacheStorage;\n    new(): CacheStorage;\n};\n\ninterface CanvasCompositing {\n    globalAlpha: number;\n    globalCompositeOperation: string;\n}\n\ninterface CanvasDrawImage {\n    drawImage(image: CanvasImageSource, dx: number, dy: number): void;\n    drawImage(image: CanvasImageSource, dx: number, dy: number, dw: number, dh: number): void;\n    drawImage(image: CanvasImageSource, sx: number, sy: number, sw: number, sh: number, dx: number, dy: number, dw: number, dh: number): void;\n}\n\ninterface CanvasDrawPath {\n    beginPath(): void;\n    clip(fillRule?: CanvasFillRule): void;\n    clip(path: Path2D, fillRule?: CanvasFillRule): void;\n    fill(fillRule?: CanvasFillRule): void;\n    fill(path: Path2D, fillRule?: CanvasFillRule): void;\n    isPointInPath(x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInPath(path: Path2D, x: number, y: number, fillRule?: CanvasFillRule): boolean;\n    isPointInStroke(x: number, y: number): boolean;\n    isPointInStroke(path: Path2D, x: number, y: number): boolean;\n    stroke(): void;\n    stroke(path: Path2D): void;\n}\n\ninterface CanvasFillStrokeStyles {\n    fillStyle: string | CanvasGradient | CanvasPattern;\n    strokeStyle: string | CanvasGradient | CanvasPattern;\n    createLinearGradient(x0: number, y0: number, x1: number, y1: number): CanvasGradient;\n    createPattern(image: CanvasImageSource, repetition: string | null): CanvasPattern | null;\n    createRadialGradient(x0: number, y0: number, r0: number, x1: number, y1: number, r1: number): CanvasGradient;\n}\n\ninterface CanvasFilters {\n    filter: string;\n}\n\n/** An opaque object describing a gradient. It is returned by the methods CanvasRenderingContext2D.createLinearGradient() or CanvasRenderingContext2D.createRadialGradient(). */\ninterface CanvasGradient {\n    /**\n     * Adds a color stop with the given color to the gradient at the given offset. 0.0 is the offset at one end of the gradient, 1.0 is the offset at the other end.\n     * \n     * Throws an "IndexSizeError" DOMException if the offset is out of range. Throws a "SyntaxError" DOMException if the color cannot be parsed.\n     */\n    addColorStop(offset: number, color: string): void;\n}\n\ndeclare var CanvasGradient: {\n    prototype: CanvasGradient;\n    new(): CanvasGradient;\n};\n\ninterface CanvasImageData {\n    createImageData(sw: number, sh: number): ImageData;\n    createImageData(imagedata: ImageData): ImageData;\n    getImageData(sx: number, sy: number, sw: number, sh: number): ImageData;\n    putImageData(imagedata: ImageData, dx: number, dy: number): void;\n    putImageData(imagedata: ImageData, dx: number, dy: number, dirtyX: number, dirtyY: number, dirtyWidth: number, dirtyHeight: number): void;\n}\n\ninterface CanvasImageSmoothing {\n    imageSmoothingEnabled: boolean;\n    imageSmoothingQuality: ImageSmoothingQuality;\n}\n\ninterface CanvasPath {\n    arc(x: number, y: number, radius: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    arcTo(x1: number, y1: number, x2: number, y2: number, radius: number): void;\n    bezierCurveTo(cp1x: number, cp1y: number, cp2x: number, cp2y: number, x: number, y: number): void;\n    closePath(): void;\n    ellipse(x: number, y: number, radiusX: number, radiusY: number, rotation: number, startAngle: number, endAngle: number, anticlockwise?: boolean): void;\n    lineTo(x: number, y: number): void;\n    moveTo(x: number, y: number): void;\n    quadraticCurveTo(cpx: number, cpy: number, x: number, y: number): void;\n    rect(x: number, y: number, w: number, h: number): void;\n}\n\ninterface CanvasPathDrawingStyles {\n    lineCap: CanvasLineCap;\n    lineDashOffset: number;\n    lineJoin: CanvasLineJoin;\n    lineWidth: number;\n    miterLimit: number;\n    getLineDash(): number[];\n    setLineDash(segments: number[]): void;\n}\n\n/** An opaque object describing a pattern, based on an image, a canvas, or a video, created by the CanvasRenderingContext2D.createPattern() method. */\ninterface CanvasPattern {\n    /**\n     * Sets the transformation matrix that will be used when rendering the pattern during a fill or stroke painting operation.\n     */\n    setTransform(transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var CanvasPattern: {\n    prototype: CanvasPattern;\n    new(): CanvasPattern;\n};\n\ninterface CanvasRect {\n    clearRect(x: number, y: number, w: number, h: number): void;\n    fillRect(x: number, y: number, w: number, h: number): void;\n    strokeRect(x: number, y: number, w: number, h: number): void;\n}\n\ninterface CanvasShadowStyles {\n    shadowBlur: number;\n    shadowColor: string;\n    shadowOffsetX: number;\n    shadowOffsetY: number;\n}\n\ninterface CanvasState {\n    restore(): void;\n    save(): void;\n}\n\ninterface CanvasText {\n    fillText(text: string, x: number, y: number, maxWidth?: number): void;\n    measureText(text: string): TextMetrics;\n    strokeText(text: string, x: number, y: number, maxWidth?: number): void;\n}\n\ninterface CanvasTextDrawingStyles {\n    direction: CanvasDirection;\n    font: string;\n    textAlign: CanvasTextAlign;\n    textBaseline: CanvasTextBaseline;\n}\n\ninterface CanvasTransform {\n    getTransform(): DOMMatrix;\n    resetTransform(): void;\n    rotate(angle: number): void;\n    scale(x: number, y: number): void;\n    setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    setTransform(transform?: DOMMatrix2DInit): void;\n    transform(a: number, b: number, c: number, d: number, e: number, f: number): void;\n    translate(x: number, y: number): void;\n}\n\n/** The Client interface represents an executable context such as a Worker, or a SharedWorker. Window clients are represented by the more-specific WindowClient. You can get Client/WindowClient objects from methods such as Clients.matchAll() and Clients.get(). */\ninterface Client {\n    readonly frameType: FrameType;\n    readonly id: string;\n    readonly type: ClientTypes;\n    readonly url: string;\n    postMessage(message: any, transfer?: Transferable[]): void;\n}\n\ndeclare var Client: {\n    prototype: Client;\n    new(): Client;\n};\n\n/** Provides access to Client objects. Access it via self.clients within a service worker. */\ninterface Clients {\n    claim(): Promise<void>;\n    get(id: string): Promise<Client | undefined>;\n    matchAll(options?: ClientQueryOptions): Promise<ReadonlyArray<Client>>;\n    openWindow(url: string): Promise<WindowClient | null>;\n}\n\ndeclare var Clients: {\n    prototype: Clients;\n    new(): Clients;\n};\n\n/** A CloseEvent is sent to clients using WebSockets when the connection is closed. This is delivered to the listener indicated by the WebSocket object\'s onclose attribute. */\ninterface CloseEvent extends Event {\n    /**\n     * Returns the WebSocket connection close code provided by the server.\n     */\n    readonly code: number;\n    /**\n     * Returns the WebSocket connection close reason provided by the server.\n     */\n    readonly reason: string;\n    /**\n     * Returns true if the connection closed cleanly; false otherwise.\n     */\n    readonly wasClean: boolean;\n}\n\ndeclare var CloseEvent: {\n    prototype: CloseEvent;\n    new(type: string, eventInitDict?: CloseEventInit): CloseEvent;\n};\n\ninterface ConcatParams extends Algorithm {\n    algorithmId: Uint8Array;\n    hash?: string | Algorithm;\n    partyUInfo: Uint8Array;\n    partyVInfo: Uint8Array;\n    privateInfo?: Uint8Array;\n    publicInfo?: Uint8Array;\n}\n\n/** This Streams API interface provides a built-in byte length queuing strategy that can be used when constructing streams. */\ninterface CountQueuingStrategy extends QueuingStrategy {\n    highWaterMark: number;\n    size(chunk: any): 1;\n}\n\ndeclare var CountQueuingStrategy: {\n    prototype: CountQueuingStrategy;\n    new(options: { highWaterMark: number }): CountQueuingStrategy;\n};\n\n/** Basic cryptography features available in the current context. It allows access to a cryptographically strong random number generator and to cryptographic primitives. */\ninterface Crypto {\n    readonly subtle: SubtleCrypto;\n    getRandomValues<T extends Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | null>(array: T): T;\n}\n\ndeclare var Crypto: {\n    prototype: Crypto;\n    new(): Crypto;\n};\n\n/** The CryptoKey dictionary of the Web Crypto API represents a cryptographic key. */\ninterface CryptoKey {\n    readonly algorithm: KeyAlgorithm;\n    readonly extractable: boolean;\n    readonly type: KeyType;\n    readonly usages: KeyUsage[];\n}\n\ndeclare var CryptoKey: {\n    prototype: CryptoKey;\n    new(): CryptoKey;\n};\n\ninterface CustomEvent<T = any> extends Event {\n    /**\n     * Returns any custom data event was created with. Typically used for synthetic events.\n     */\n    readonly detail: T;\n    initCustomEvent(typeArg: string, canBubbleArg: boolean, cancelableArg: boolean, detailArg: T): void;\n}\n\ndeclare var CustomEvent: {\n    prototype: CustomEvent;\n    new<T>(typeArg: string, eventInitDict?: CustomEventInit<T>): CustomEvent<T>;\n};\n\n/** An abnormal event (called an exception) which occurs as a result of calling a method or accessing a property of a web API. */\ninterface DOMException {\n    readonly code: number;\n    readonly message: string;\n    readonly name: string;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n}\n\ndeclare var DOMException: {\n    prototype: DOMException;\n    new(message?: string, name?: string): DOMException;\n    readonly ABORT_ERR: number;\n    readonly DATA_CLONE_ERR: number;\n    readonly DOMSTRING_SIZE_ERR: number;\n    readonly HIERARCHY_REQUEST_ERR: number;\n    readonly INDEX_SIZE_ERR: number;\n    readonly INUSE_ATTRIBUTE_ERR: number;\n    readonly INVALID_ACCESS_ERR: number;\n    readonly INVALID_CHARACTER_ERR: number;\n    readonly INVALID_MODIFICATION_ERR: number;\n    readonly INVALID_NODE_TYPE_ERR: number;\n    readonly INVALID_STATE_ERR: number;\n    readonly NAMESPACE_ERR: number;\n    readonly NETWORK_ERR: number;\n    readonly NOT_FOUND_ERR: number;\n    readonly NOT_SUPPORTED_ERR: number;\n    readonly NO_DATA_ALLOWED_ERR: number;\n    readonly NO_MODIFICATION_ALLOWED_ERR: number;\n    readonly QUOTA_EXCEEDED_ERR: number;\n    readonly SECURITY_ERR: number;\n    readonly SYNTAX_ERR: number;\n    readonly TIMEOUT_ERR: number;\n    readonly TYPE_MISMATCH_ERR: number;\n    readonly URL_MISMATCH_ERR: number;\n    readonly VALIDATION_ERR: number;\n    readonly WRONG_DOCUMENT_ERR: number;\n};\n\ninterface DOMMatrix extends DOMMatrixReadOnly {\n    a: number;\n    b: number;\n    c: number;\n    d: number;\n    e: number;\n    f: number;\n    m11: number;\n    m12: number;\n    m13: number;\n    m14: number;\n    m21: number;\n    m22: number;\n    m23: number;\n    m24: number;\n    m31: number;\n    m32: number;\n    m33: number;\n    m34: number;\n    m41: number;\n    m42: number;\n    m43: number;\n    m44: number;\n    invertSelf(): DOMMatrix;\n    multiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    preMultiplySelf(other?: DOMMatrixInit): DOMMatrix;\n    rotateAxisAngleSelf(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVectorSelf(x?: number, y?: number): DOMMatrix;\n    rotateSelf(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    scale3dSelf(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scaleSelf(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    skewXSelf(sx?: number): DOMMatrix;\n    skewYSelf(sy?: number): DOMMatrix;\n    translateSelf(tx?: number, ty?: number, tz?: number): DOMMatrix;\n}\n\ndeclare var DOMMatrix: {\n    prototype: DOMMatrix;\n    new(init?: string | number[]): DOMMatrix;\n    fromFloat32Array(array32: Float32Array): DOMMatrix;\n    fromFloat64Array(array64: Float64Array): DOMMatrix;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrix;\n};\n\ninterface DOMMatrixReadOnly {\n    readonly a: number;\n    readonly b: number;\n    readonly c: number;\n    readonly d: number;\n    readonly e: number;\n    readonly f: number;\n    readonly is2D: boolean;\n    readonly isIdentity: boolean;\n    readonly m11: number;\n    readonly m12: number;\n    readonly m13: number;\n    readonly m14: number;\n    readonly m21: number;\n    readonly m22: number;\n    readonly m23: number;\n    readonly m24: number;\n    readonly m31: number;\n    readonly m32: number;\n    readonly m33: number;\n    readonly m34: number;\n    readonly m41: number;\n    readonly m42: number;\n    readonly m43: number;\n    readonly m44: number;\n    flipX(): DOMMatrix;\n    flipY(): DOMMatrix;\n    inverse(): DOMMatrix;\n    multiply(other?: DOMMatrixInit): DOMMatrix;\n    rotate(rotX?: number, rotY?: number, rotZ?: number): DOMMatrix;\n    rotateAxisAngle(x?: number, y?: number, z?: number, angle?: number): DOMMatrix;\n    rotateFromVector(x?: number, y?: number): DOMMatrix;\n    scale(scaleX?: number, scaleY?: number, scaleZ?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    scale3d(scale?: number, originX?: number, originY?: number, originZ?: number): DOMMatrix;\n    /** @deprecated */\n    scaleNonUniform(scaleX?: number, scaleY?: number): DOMMatrix;\n    skewX(sx?: number): DOMMatrix;\n    skewY(sy?: number): DOMMatrix;\n    toFloat32Array(): Float32Array;\n    toFloat64Array(): Float64Array;\n    toJSON(): any;\n    transformPoint(point?: DOMPointInit): DOMPoint;\n    translate(tx?: number, ty?: number, tz?: number): DOMMatrix;\n}\n\ndeclare var DOMMatrixReadOnly: {\n    prototype: DOMMatrixReadOnly;\n    new(init?: string | number[]): DOMMatrixReadOnly;\n    fromFloat32Array(array32: Float32Array): DOMMatrixReadOnly;\n    fromFloat64Array(array64: Float64Array): DOMMatrixReadOnly;\n    fromMatrix(other?: DOMMatrixInit): DOMMatrixReadOnly;\n};\n\ninterface DOMPoint extends DOMPointReadOnly {\n    w: number;\n    x: number;\n    y: number;\n    z: number;\n}\n\ndeclare var DOMPoint: {\n    prototype: DOMPoint;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPoint;\n    fromPoint(other?: DOMPointInit): DOMPoint;\n};\n\ninterface DOMPointReadOnly {\n    readonly w: number;\n    readonly x: number;\n    readonly y: number;\n    readonly z: number;\n    matrixTransform(matrix?: DOMMatrixInit): DOMPoint;\n    toJSON(): any;\n}\n\ndeclare var DOMPointReadOnly: {\n    prototype: DOMPointReadOnly;\n    new(x?: number, y?: number, z?: number, w?: number): DOMPointReadOnly;\n    fromPoint(other?: DOMPointInit): DOMPointReadOnly;\n};\n\ninterface DOMQuad {\n    readonly p1: DOMPoint;\n    readonly p2: DOMPoint;\n    readonly p3: DOMPoint;\n    readonly p4: DOMPoint;\n    getBounds(): DOMRect;\n    toJSON(): any;\n}\n\ndeclare var DOMQuad: {\n    prototype: DOMQuad;\n    new(p1?: DOMPointInit, p2?: DOMPointInit, p3?: DOMPointInit, p4?: DOMPointInit): DOMQuad;\n    fromQuad(other?: DOMQuadInit): DOMQuad;\n    fromRect(other?: DOMRectInit): DOMQuad;\n};\n\ninterface DOMRect extends DOMRectReadOnly {\n    height: number;\n    width: number;\n    x: number;\n    y: number;\n}\n\ndeclare var DOMRect: {\n    prototype: DOMRect;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRect;\n    fromRect(other?: DOMRectInit): DOMRect;\n};\n\ninterface DOMRectReadOnly {\n    readonly bottom: number;\n    readonly height: number;\n    readonly left: number;\n    readonly right: number;\n    readonly top: number;\n    readonly width: number;\n    readonly x: number;\n    readonly y: number;\n    toJSON(): any;\n}\n\ndeclare var DOMRectReadOnly: {\n    prototype: DOMRectReadOnly;\n    new(x?: number, y?: number, width?: number, height?: number): DOMRectReadOnly;\n    fromRect(other?: DOMRectInit): DOMRectReadOnly;\n};\n\n/** A type returned by some APIs which contains a list of DOMString (strings). */\ninterface DOMStringList {\n    /**\n     * Returns the number of strings in strings.\n     */\n    readonly length: number;\n    /**\n     * Returns true if strings contains string, and false otherwise.\n     */\n    contains(string: string): boolean;\n    /**\n     * Returns the string with index index from strings.\n     */\n    item(index: number): string | null;\n    [index: number]: string;\n}\n\ndeclare var DOMStringList: {\n    prototype: DOMStringList;\n    new(): DOMStringList;\n};\n\ninterface DedicatedWorkerGlobalScopeEventMap extends WorkerGlobalScopeEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** (the Worker global scope) is accessible through the self keyword. Some additional global functions, namespaces objects, and constructors, not typically associated with the worker global scope, but available on it, are listed in the JavaScript Reference. See also: Functions available to workers. */\ninterface DedicatedWorkerGlobalScope extends WorkerGlobalScope, AnimationFrameProvider {\n    /**\n     * Returns dedicatedWorkerGlobal\'s name, i.e. the value given to the Worker constructor. Primarily useful for debugging.\n     */\n    readonly name: string;\n    onmessage: ((this: DedicatedWorkerGlobalScope, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: DedicatedWorkerGlobalScope, ev: MessageEvent) => any) | null;\n    /**\n     * Aborts dedicatedWorkerGlobal.\n     */\n    close(): void;\n    /**\n     * Clones message and transmits it to the Worker object associated with dedicatedWorkerGlobal. transfer can be passed as a list of objects that are to be transferred rather than cloned.\n     */\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    addEventListener<K extends keyof DedicatedWorkerGlobalScopeEventMap>(type: K, listener: (this: DedicatedWorkerGlobalScope, ev: DedicatedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof DedicatedWorkerGlobalScopeEventMap>(type: K, listener: (this: DedicatedWorkerGlobalScope, ev: DedicatedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var DedicatedWorkerGlobalScope: {\n    prototype: DedicatedWorkerGlobalScope;\n    new(): DedicatedWorkerGlobalScope;\n};\n\ninterface DhImportKeyParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyAlgorithm extends KeyAlgorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface DhKeyDeriveParams extends Algorithm {\n    public: CryptoKey;\n}\n\ninterface DhKeyGenParams extends Algorithm {\n    generator: Uint8Array;\n    prime: Uint8Array;\n}\n\ninterface EXT_blend_minmax {\n    readonly MAX_EXT: GLenum;\n    readonly MIN_EXT: GLenum;\n}\n\n/** The EXT_frag_depth extension is part of the WebGL API and enables to set a depth value of a fragment from within the fragment shader. */\ninterface EXT_frag_depth {\n}\n\ninterface EXT_sRGB {\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING_EXT: GLenum;\n    readonly SRGB8_ALPHA8_EXT: GLenum;\n    readonly SRGB_ALPHA_EXT: GLenum;\n    readonly SRGB_EXT: GLenum;\n}\n\ninterface EXT_shader_texture_lod {\n}\n\n/** The EXT_texture_filter_anisotropic extension is part of the WebGL API and exposes two constants for anisotropic filtering (AF). */\ninterface EXT_texture_filter_anisotropic {\n    readonly MAX_TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n    readonly TEXTURE_MAX_ANISOTROPY_EXT: GLenum;\n}\n\n/** Events providing information related to errors in scripts or in files. */\ninterface ErrorEvent extends Event {\n    readonly colno: number;\n    readonly error: any;\n    readonly filename: string;\n    readonly lineno: number;\n    readonly message: string;\n}\n\ndeclare var ErrorEvent: {\n    prototype: ErrorEvent;\n    new(type: string, eventInitDict?: ErrorEventInit): ErrorEvent;\n};\n\n/** An event which takes place in the DOM. */\ninterface Event {\n    /**\n     * Returns true or false depending on how event was initialized. True if event goes through its target\'s ancestors in reverse tree order, and false otherwise.\n     */\n    readonly bubbles: boolean;\n    cancelBubble: boolean;\n    /**\n     * Returns true or false depending on how event was initialized. Its return value does not always carry meaning, but true can indicate that part of the operation during which event was dispatched, can be canceled by invoking the preventDefault() method.\n     */\n    readonly cancelable: boolean;\n    /**\n     * Returns true or false depending on how event was initialized. True if event invokes listeners past a ShadowRoot node that is the root of its target, and false otherwise.\n     */\n    readonly composed: boolean;\n    /**\n     * Returns the object whose event listener\'s callback is currently being invoked.\n     */\n    readonly currentTarget: EventTarget | null;\n    /**\n     * Returns true if preventDefault() was invoked successfully to indicate cancelation, and false otherwise.\n     */\n    readonly defaultPrevented: boolean;\n    /**\n     * Returns the event\'s phase, which is one of NONE, CAPTURING_PHASE, AT_TARGET, and BUBBLING_PHASE.\n     */\n    readonly eventPhase: number;\n    /**\n     * Returns true if event was dispatched by the user agent, and false otherwise.\n     */\n    readonly isTrusted: boolean;\n    returnValue: boolean;\n    /** @deprecated */\n    readonly srcElement: EventTarget | null;\n    /**\n     * Returns the object to which event is dispatched (its target).\n     */\n    readonly target: EventTarget | null;\n    /**\n     * Returns the event\'s timestamp as the number of milliseconds measured relative to the time origin.\n     */\n    readonly timeStamp: number;\n    /**\n     * Returns the type of event, e.g. "click", "hashchange", or "submit".\n     */\n    readonly type: string;\n    /**\n     * Returns the invocation target objects of event\'s path (objects on which listeners will be invoked), except for any nodes in shadow trees of which the shadow root\'s mode is "closed" that are not reachable from event\'s currentTarget.\n     */\n    composedPath(): EventTarget[];\n    initEvent(type: string, bubbles?: boolean, cancelable?: boolean): void;\n    /**\n     * If invoked when the cancelable attribute value is true, and while executing a listener for the event with passive set to false, signals to the operation that caused event to be dispatched that it needs to be canceled.\n     */\n    preventDefault(): void;\n    /**\n     * Invoking this method prevents event from reaching any registered event listeners after the current one finishes running and, when dispatched in a tree, also prevents event from reaching any other objects.\n     */\n    stopImmediatePropagation(): void;\n    /**\n     * When dispatched in a tree, invoking this method prevents event from reaching any objects other than the current object.\n     */\n    stopPropagation(): void;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n}\n\ndeclare var Event: {\n    prototype: Event;\n    new(type: string, eventInitDict?: EventInit): Event;\n    readonly AT_TARGET: number;\n    readonly BUBBLING_PHASE: number;\n    readonly CAPTURING_PHASE: number;\n    readonly NONE: number;\n};\n\ninterface EventListenerObject {\n    handleEvent(evt: Event): void;\n}\n\ninterface EventSourceEventMap {\n    "error": Event;\n    "message": MessageEvent;\n    "open": Event;\n}\n\ninterface EventSource extends EventTarget {\n    onerror: ((this: EventSource, ev: Event) => any) | null;\n    onmessage: ((this: EventSource, ev: MessageEvent) => any) | null;\n    onopen: ((this: EventSource, ev: Event) => any) | null;\n    /**\n     * Returns the state of this EventSource object\'s connection. It can have the values described below.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the URL providing the event stream.\n     */\n    readonly url: string;\n    /**\n     * Returns true if the credentials mode for connection requests to the URL providing the event stream is set to "include", and false otherwise.\n     */\n    readonly withCredentials: boolean;\n    /**\n     * Aborts any instances of the fetch algorithm started for this EventSource object, and sets the readyState attribute to CLOSED.\n     */\n    close(): void;\n    readonly CLOSED: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    addEventListener<K extends keyof EventSourceEventMap>(type: K, listener: (this: EventSource, ev: EventSourceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof EventSourceEventMap>(type: K, listener: (this: EventSource, ev: EventSourceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var EventSource: {\n    prototype: EventSource;\n    new(url: string, eventSourceInitDict?: EventSourceInit): EventSource;\n    readonly CLOSED: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n};\n\n/** EventTarget is a DOM interface implemented by objects that can receive events and may have listeners for them. */\ninterface EventTarget {\n    /**\n     * Appends an event listener for events whose type attribute value is type. The callback argument sets the callback that will be invoked when the event is dispatched.\n     * \n     * The options argument sets listener-specific options. For compatibility this can be a boolean, in which case the method behaves exactly as if the value was specified as options\'s capture.\n     * \n     * When set to true, options\'s capture prevents callback from being invoked when the event\'s eventPhase attribute value is BUBBLING_PHASE. When false (or not present), callback will not be invoked when event\'s eventPhase attribute value is CAPTURING_PHASE. Either way, callback will be invoked if event\'s eventPhase attribute value is AT_TARGET.\n     * \n     * When set to true, options\'s passive indicates that the callback will not cancel the event by invoking preventDefault(). This is used to enable performance optimizations described in § 2.8 Observing event listeners.\n     * \n     * When set to true, options\'s once indicates that the callback will only be invoked once after which the event listener will be removed.\n     * \n     * The event listener is appended to target\'s event listener list and is not appended if it has the same type, callback, and capture.\n     */\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject | null, options?: boolean | AddEventListenerOptions): void;\n    /**\n     * Dispatches a synthetic event event to target and returns true if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n     */\n    dispatchEvent(event: Event): boolean;\n    /**\n     * Removes the event listener in target\'s event listener list with the same type, callback, and options.\n     */\n    removeEventListener(type: string, callback: EventListenerOrEventListenerObject | null, options?: EventListenerOptions | boolean): void;\n}\n\ndeclare var EventTarget: {\n    prototype: EventTarget;\n    new(): EventTarget;\n};\n\n/** Extends the lifetime of the install and activate events dispatched on the global scope as part of the service worker lifecycle. This ensures that any functional events (like FetchEvent) are not dispatched until it upgrades database schemas and deletes the outdated cache entries. */\ninterface ExtendableEvent extends Event {\n    waitUntil(f: any): void;\n}\n\ndeclare var ExtendableEvent: {\n    prototype: ExtendableEvent;\n    new(type: string, eventInitDict?: ExtendableEventInit): ExtendableEvent;\n};\n\n/** This ServiceWorker API interface represents the event object of a message event fired on a service worker (when a channel message is received on the ServiceWorkerGlobalScope from another context) — extends the lifetime of such events. */\ninterface ExtendableMessageEvent extends ExtendableEvent {\n    readonly data: any;\n    readonly lastEventId: string;\n    readonly origin: string;\n    readonly ports: ReadonlyArray<MessagePort>;\n    readonly source: Client | ServiceWorker | MessagePort | null;\n}\n\ndeclare var ExtendableMessageEvent: {\n    prototype: ExtendableMessageEvent;\n    new(type: string, eventInitDict?: ExtendableMessageEventInit): ExtendableMessageEvent;\n};\n\n/** This is the event type for fetch events dispatched on the service worker global scope. It contains information about the fetch, including the request and how the receiver will treat the response. It provides the event.respondWith() method, which allows us to provide a response to this fetch. */\ninterface FetchEvent extends ExtendableEvent {\n    readonly clientId: string;\n    readonly preloadResponse: Promise<any>;\n    readonly replacesClientId: string;\n    readonly request: Request;\n    readonly resultingClientId: string;\n    respondWith(r: Response | Promise<Response>): void;\n}\n\ndeclare var FetchEvent: {\n    prototype: FetchEvent;\n    new(type: string, eventInitDict: FetchEventInit): FetchEvent;\n};\n\n/** Provides information about files and allows JavaScript in a web page to access their content. */\ninterface File extends Blob {\n    readonly lastModified: number;\n    readonly name: string;\n}\n\ndeclare var File: {\n    prototype: File;\n    new(fileBits: BlobPart[], fileName: string, options?: FilePropertyBag): File;\n};\n\n/** An object of this type is returned by the files property of the HTML <input> element; this lets you access the list of files selected with the <input type="file"> element. It\'s also used for a list of files dropped into web content when using the drag and drop API; see the DataTransfer object for details on this usage. */\ninterface FileList {\n    readonly length: number;\n    item(index: number): File | null;\n    [index: number]: File;\n}\n\ndeclare var FileList: {\n    prototype: FileList;\n    new(): FileList;\n};\n\ninterface FileReaderEventMap {\n    "abort": ProgressEvent<FileReader>;\n    "error": ProgressEvent<FileReader>;\n    "load": ProgressEvent<FileReader>;\n    "loadend": ProgressEvent<FileReader>;\n    "loadstart": ProgressEvent<FileReader>;\n    "progress": ProgressEvent<FileReader>;\n}\n\n/** Lets web applications asynchronously read the contents of files (or raw data buffers) stored on the user\'s computer, using File or Blob objects to specify the file or data to read. */\ninterface FileReader extends EventTarget {\n    readonly error: DOMException | null;\n    onabort: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onerror: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onload: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onloadend: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onloadstart: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    onprogress: ((this: FileReader, ev: ProgressEvent<FileReader>) => any) | null;\n    readonly readyState: number;\n    readonly result: string | ArrayBuffer | null;\n    abort(): void;\n    readAsArrayBuffer(blob: Blob): void;\n    readAsBinaryString(blob: Blob): void;\n    readAsDataURL(blob: Blob): void;\n    readAsText(blob: Blob, encoding?: string): void;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n    addEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof FileReaderEventMap>(type: K, listener: (this: FileReader, ev: FileReaderEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var FileReader: {\n    prototype: FileReader;\n    new(): FileReader;\n    readonly DONE: number;\n    readonly EMPTY: number;\n    readonly LOADING: number;\n};\n\n/** Allows to read File or Blob objects in a synchronous way. */\ninterface FileReaderSync {\n    readAsArrayBuffer(blob: Blob): ArrayBuffer;\n    readAsBinaryString(blob: Blob): string;\n    readAsDataURL(blob: Blob): string;\n    readAsText(blob: Blob, encoding?: string): string;\n}\n\ndeclare var FileReaderSync: {\n    prototype: FileReaderSync;\n    new(): FileReaderSync;\n};\n\n/** Provides a way to easily construct a set of key/value pairs representing form fields and their values, which can then be easily sent using the XMLHttpRequest.send() method. It uses the same format a form would use if the encoding type were set to "multipart/form-data". */\ninterface FormData {\n    append(name: string, value: string | Blob, fileName?: string): void;\n    delete(name: string): void;\n    get(name: string): FormDataEntryValue | null;\n    getAll(name: string): FormDataEntryValue[];\n    has(name: string): boolean;\n    set(name: string, value: string | Blob, fileName?: string): void;\n    forEach(callbackfn: (value: FormDataEntryValue, key: string, parent: FormData) => void, thisArg?: any): void;\n}\n\ndeclare var FormData: {\n    prototype: FormData;\n    new(): FormData;\n};\n\ninterface GenericTransformStream {\n    /**\n     * Returns a readable stream whose chunks are strings resulting from running encoding\'s decoder on the chunks written to writable.\n     */\n    readonly readable: ReadableStream;\n    /**\n     * Returns a writable stream which accepts [AllowShared] BufferSource chunks and runs them through encoding\'s decoder before making them available to readable.\n     * \n     * Typically this will be used via the pipeThrough() method on a ReadableStream source.\n     * \n     * ```\n     * var decoder = new TextDecoderStream(encoding);\n     * byteReadable\n     *   .pipeThrough(decoder)\n     *   .pipeTo(textWritable);\n     * ```\n     * \n     * If the error mode is "fatal" and encoding\'s decoder returns error, both readable and writable will be errored with a TypeError.\n     */\n    readonly writable: WritableStream;\n}\n\n/** This Fetch API interface allows you to perform various actions on HTTP request and response headers. These actions include retrieving, setting, adding to, and removing. A Headers object has an associated header list, which is initially empty and consists of zero or more name and value pairs.  You can add to this using methods like append() (see Examples.) In all methods of this interface, header names are matched by case-insensitive byte sequence. */\ninterface Headers {\n    append(name: string, value: string): void;\n    delete(name: string): void;\n    get(name: string): string | null;\n    has(name: string): boolean;\n    set(name: string, value: string): void;\n    forEach(callbackfn: (value: string, key: string, parent: Headers) => void, thisArg?: any): void;\n}\n\ndeclare var Headers: {\n    prototype: Headers;\n    new(init?: HeadersInit): Headers;\n};\n\ninterface IDBArrayKey extends Array<IDBValidKey> {\n}\n\n/** This IndexedDB API interface represents a cursor for traversing or iterating over multiple records in a database. */\ninterface IDBCursor {\n    /**\n     * Returns the direction ("next", "nextunique", "prev" or "prevunique") of the cursor.\n     */\n    readonly direction: IDBCursorDirection;\n    /**\n     * Returns the key of the cursor. Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly key: IDBValidKey;\n    /**\n     * Returns the effective key of the cursor. Throws a "InvalidStateError" DOMException if the cursor is advancing or is finished.\n     */\n    readonly primaryKey: IDBValidKey;\n    /**\n     * Returns the IDBObjectStore or IDBIndex the cursor was opened from.\n     */\n    readonly source: IDBObjectStore | IDBIndex;\n    /**\n     * Advances the cursor through the next count records in range.\n     */\n    advance(count: number): void;\n    /**\n     * Advances the cursor to the next record in range.\n     */\n    continue(key?: IDBValidKey): void;\n    /**\n     * Advances the cursor to the next record in range matching or after key and primaryKey. Throws an "InvalidAccessError" DOMException if the source is not an index.\n     */\n    continuePrimaryKey(key: IDBValidKey, primaryKey: IDBValidKey): void;\n    /**\n     * Delete the record pointed at by the cursor with a new value.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    delete(): IDBRequest<undefined>;\n    /**\n     * Updated the record pointed at by the cursor with a new value.\n     * \n     * Throws a "DataError" DOMException if the effective object store uses in-line keys and the key would have changed.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    update(value: any): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBCursor: {\n    prototype: IDBCursor;\n    new(): IDBCursor;\n};\n\n/** This IndexedDB API interface represents a cursor for traversing or iterating over multiple records in a database. It is the same as the IDBCursor, except that it includes the value property. */\ninterface IDBCursorWithValue extends IDBCursor {\n    /**\n     * Returns the cursor\'s current value.\n     */\n    readonly value: any;\n}\n\ndeclare var IDBCursorWithValue: {\n    prototype: IDBCursorWithValue;\n    new(): IDBCursorWithValue;\n};\n\ninterface IDBDatabaseEventMap {\n    "abort": Event;\n    "close": Event;\n    "error": Event;\n    "versionchange": IDBVersionChangeEvent;\n}\n\n/** This IndexedDB API interface provides a connection to a database; you can use an IDBDatabase object to open a transaction on your database then create, manipulate, and delete objects (data) in that database. The interface provides the only way to get and manage versions of the database. */\ninterface IDBDatabase extends EventTarget {\n    /**\n     * Returns the name of the database.\n     */\n    readonly name: string;\n    /**\n     * Returns a list of the names of object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBDatabase, ev: Event) => any) | null;\n    onclose: ((this: IDBDatabase, ev: Event) => any) | null;\n    onerror: ((this: IDBDatabase, ev: Event) => any) | null;\n    onversionchange: ((this: IDBDatabase, ev: IDBVersionChangeEvent) => any) | null;\n    /**\n     * Returns the version of the database.\n     */\n    readonly version: number;\n    /**\n     * Closes the connection once all running transactions have finished.\n     */\n    close(): void;\n    /**\n     * Creates a new object store with the given name and options and returns a new IDBObjectStore.\n     * \n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createObjectStore(name: string, options?: IDBObjectStoreParameters): IDBObjectStore;\n    /**\n     * Deletes the object store with the given name.\n     * \n     * Throws a "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    deleteObjectStore(name: string): void;\n    /**\n     * Returns a new transaction with the given mode ("readonly" or "readwrite") and scope which can be a single object store name or an array of names.\n     */\n    transaction(storeNames: string | string[], mode?: IDBTransactionMode): IDBTransaction;\n    addEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBDatabaseEventMap>(type: K, listener: (this: IDBDatabase, ev: IDBDatabaseEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBDatabase: {\n    prototype: IDBDatabase;\n    new(): IDBDatabase;\n};\n\n/** In the following code snippet, we make a request to open a database, and include handlers for the success and error cases. For a full working example, see our To-do Notifications app (view example live.) */\ninterface IDBFactory {\n    /**\n     * Compares two values as keys. Returns -1 if key1 precedes key2, 1 if key2 precedes key1, and 0 if the keys are equal.\n     * \n     * Throws a "DataError" DOMException if either input is not a valid key.\n     */\n    cmp(first: any, second: any): number;\n    /**\n     * Attempts to delete the named database. If the database already exists and there are open connections that don\'t close in response to a versionchange event, the request will be blocked until all they close. If the request is successful request\'s result will be null.\n     */\n    deleteDatabase(name: string): IDBOpenDBRequest;\n    /**\n     * Attempts to open a connection to the named database with the current version, or 1 if it does not already exist. If the request is successful request\'s result will be the connection.\n     */\n    open(name: string, version?: number): IDBOpenDBRequest;\n}\n\ndeclare var IDBFactory: {\n    prototype: IDBFactory;\n    new(): IDBFactory;\n};\n\n/** IDBIndex interface of the IndexedDB API provides asynchronous access to an index in a database. An index is a kind of object store for looking up records in another object store, called the referenced object store. You use this interface to retrieve data. */\ninterface IDBIndex {\n    readonly keyPath: string | string[];\n    readonly multiEntry: boolean;\n    /**\n     * Returns the name of the index.\n     */\n    name: string;\n    /**\n     * Returns the IDBObjectStore the index belongs to.\n     */\n    readonly objectStore: IDBObjectStore;\n    readonly unique: boolean;\n    /**\n     * Retrieves the number of records matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Retrieves the value of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(key: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(key: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    /**\n     * Opens a cursor over the records matching query, ordered by direction. If query is null, all records in index are matched.\n     * \n     * If successful, request\'s result will be an IDBCursorWithValue, or null if there were no matching records.\n     */\n    openCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in index are matched.\n     * \n     * If successful, request\'s result will be an IDBCursor, or null if there were no matching records.\n     */\n    openKeyCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n}\n\ndeclare var IDBIndex: {\n    prototype: IDBIndex;\n    new(): IDBIndex;\n};\n\n/** A key range can be a single value or a range with upper and lower bounds or endpoints. If the key range has both upper and lower bounds, then it is bounded; if it has no bounds, it is unbounded. A bounded key range can either be open (the endpoints are excluded) or closed (the endpoints are included). To retrieve all keys within a certain range, you can use the following code constructs: */\ninterface IDBKeyRange {\n    /**\n     * Returns lower bound, or undefined if none.\n     */\n    readonly lower: any;\n    /**\n     * Returns true if the lower open flag is set, and false otherwise.\n     */\n    readonly lowerOpen: boolean;\n    /**\n     * Returns upper bound, or undefined if none.\n     */\n    readonly upper: any;\n    /**\n     * Returns true if the upper open flag is set, and false otherwise.\n     */\n    readonly upperOpen: boolean;\n    /**\n     * Returns true if key is included in the range, and false otherwise.\n     */\n    includes(key: any): boolean;\n}\n\ndeclare var IDBKeyRange: {\n    prototype: IDBKeyRange;\n    new(): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning from lower to upper. If lowerOpen is true, lower is not included in the range. If upperOpen is true, upper is not included in the range.\n     */\n    bound(lower: any, upper: any, lowerOpen?: boolean, upperOpen?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange starting at key with no upper bound. If open is true, key is not included in the range.\n     */\n    lowerBound(lower: any, open?: boolean): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange spanning only key.\n     */\n    only(value: any): IDBKeyRange;\n    /**\n     * Returns a new IDBKeyRange with no lower bound and ending at key. If open is true, key is not included in the range.\n     */\n    upperBound(upper: any, open?: boolean): IDBKeyRange;\n};\n\n/** This example shows a variety of different uses of object stores, from updating the data structure with IDBObjectStore.createIndex inside an onupgradeneeded function, to adding a new item to our object store with IDBObjectStore.add. For a full working example, see our To-do Notifications app (view example live.) */\ninterface IDBObjectStore {\n    /**\n     * Returns true if the store has a key generator, and false otherwise.\n     */\n    readonly autoIncrement: boolean;\n    /**\n     * Returns a list of the names of indexes in the store.\n     */\n    readonly indexNames: DOMStringList;\n    /**\n     * Returns the key path of the store, or null if none.\n     */\n    readonly keyPath: string | string[];\n    /**\n     * Returns the name of the store.\n     */\n    name: string;\n    /**\n     * Returns the associated transaction.\n     */\n    readonly transaction: IDBTransaction;\n    /**\n     * Adds or updates a record in store with the given value and key.\n     * \n     * If the store uses in-line keys and key is specified a "DataError" DOMException will be thrown.\n     * \n     * If put() is used, any existing record with the key will be replaced. If add() is used, and if a record with the key already exists the request will fail, with request\'s error set to a "ConstraintError" DOMException.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    add(value: any, key?: IDBValidKey): IDBRequest<IDBValidKey>;\n    /**\n     * Deletes all records in store.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    clear(): IDBRequest<undefined>;\n    /**\n     * Retrieves the number of records matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the count.\n     */\n    count(key?: IDBValidKey | IDBKeyRange): IDBRequest<number>;\n    /**\n     * Creates a new index in store with the given name, keyPath and options and returns a new IDBIndex. If the keyPath and options define constraints that cannot be satisfied with the data already in store the upgrade transaction will abort with a "ConstraintError" DOMException.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createIndex(name: string, keyPath: string | string[], options?: IDBIndexParameters): IDBIndex;\n    /**\n     * Deletes records in store with the given key or in the given key range in query.\n     * \n     * If successful, request\'s result will be undefined.\n     */\n    delete(key: IDBValidKey | IDBKeyRange): IDBRequest<undefined>;\n    /**\n     * Deletes the index in store with the given name.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    deleteIndex(name: string): void;\n    /**\n     * Retrieves the value of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the value, or undefined if there was no matching record.\n     */\n    get(query: IDBValidKey | IDBKeyRange): IDBRequest<any | undefined>;\n    /**\n     * Retrieves the values of the records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the values.\n     */\n    getAll(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<any[]>;\n    /**\n     * Retrieves the keys of records matching the given key or key range in query (up to count if given).\n     * \n     * If successful, request\'s result will be an Array of the keys.\n     */\n    getAllKeys(query?: IDBValidKey | IDBKeyRange | null, count?: number): IDBRequest<IDBValidKey[]>;\n    /**\n     * Retrieves the key of the first record matching the given key or key range in query.\n     * \n     * If successful, request\'s result will be the key, or undefined if there was no matching record.\n     */\n    getKey(query: IDBValidKey | IDBKeyRange): IDBRequest<IDBValidKey | undefined>;\n    index(name: string): IDBIndex;\n    /**\n     * Opens a cursor over the records matching query, ordered by direction. If query is null, all records in store are matched.\n     * \n     * If successful, request\'s result will be an IDBCursorWithValue pointing at the first matching record, or null if there were no matching records.\n     */\n    openCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursorWithValue | null>;\n    /**\n     * Opens a cursor with key only flag set over the records matching query, ordered by direction. If query is null, all records in store are matched.\n     * \n     * If successful, request\'s result will be an IDBCursor pointing at the first matching record, or null if there were no matching records.\n     */\n    openKeyCursor(query?: IDBValidKey | IDBKeyRange | null, direction?: IDBCursorDirection): IDBRequest<IDBCursor | null>;\n    /**\n     * Adds or updates a record in store with the given value and key.\n     * \n     * If the store uses in-line keys and key is specified a "DataError" DOMException will be thrown.\n     * \n     * If put() is used, any existing record with the key will be replaced. If add() is used, and if a record with the key already exists the request will fail, with request\'s error set to a "ConstraintError" DOMException.\n     * \n     * If successful, request\'s result will be the record\'s key.\n     */\n    put(value: any, key?: IDBValidKey): IDBRequest<IDBValidKey>;\n}\n\ndeclare var IDBObjectStore: {\n    prototype: IDBObjectStore;\n    new(): IDBObjectStore;\n};\n\ninterface IDBOpenDBRequestEventMap extends IDBRequestEventMap {\n    "blocked": Event;\n    "upgradeneeded": IDBVersionChangeEvent;\n}\n\n/** Also inherits methods from its parents IDBRequest and EventTarget. */\ninterface IDBOpenDBRequest extends IDBRequest<IDBDatabase> {\n    onblocked: ((this: IDBOpenDBRequest, ev: Event) => any) | null;\n    onupgradeneeded: ((this: IDBOpenDBRequest, ev: IDBVersionChangeEvent) => any) | null;\n    addEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBOpenDBRequestEventMap>(type: K, listener: (this: IDBOpenDBRequest, ev: IDBOpenDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBOpenDBRequest: {\n    prototype: IDBOpenDBRequest;\n    new(): IDBOpenDBRequest;\n};\n\ninterface IDBRequestEventMap {\n    "error": Event;\n    "success": Event;\n}\n\n/** The request object does not initially contain any information about the result of the operation, but once information becomes available, an event is fired on the request, and the information becomes available through the properties of the IDBRequest instance. */\ninterface IDBRequest<T = any> extends EventTarget {\n    /**\n     * When a request is completed, returns the error (a DOMException), or null if the request succeeded. Throws a "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly error: DOMException | null;\n    onerror: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    onsuccess: ((this: IDBRequest<T>, ev: Event) => any) | null;\n    /**\n     * Returns "pending" until a request is complete, then returns "done".\n     */\n    readonly readyState: IDBRequestReadyState;\n    /**\n     * When a request is completed, returns the result, or undefined if the request failed. Throws a "InvalidStateError" DOMException if the request is still pending.\n     */\n    readonly result: T;\n    /**\n     * Returns the IDBObjectStore, IDBIndex, or IDBCursor the request was made against, or null if is was an open request.\n     */\n    readonly source: IDBObjectStore | IDBIndex | IDBCursor;\n    /**\n     * Returns the IDBTransaction the request was made within. If this as an open request, then it returns an upgrade transaction while it is running, or null otherwise.\n     */\n    readonly transaction: IDBTransaction | null;\n    addEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBRequestEventMap>(type: K, listener: (this: IDBRequest<T>, ev: IDBRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBRequest: {\n    prototype: IDBRequest;\n    new(): IDBRequest;\n};\n\ninterface IDBTransactionEventMap {\n    "abort": Event;\n    "complete": Event;\n    "error": Event;\n}\n\ninterface IDBTransaction extends EventTarget {\n    /**\n     * Returns the transaction\'s connection.\n     */\n    readonly db: IDBDatabase;\n    /**\n     * If the transaction was aborted, returns the error (a DOMException) providing the reason.\n     */\n    readonly error: DOMException;\n    /**\n     * Returns the mode the transaction was created with ("readonly" or "readwrite"), or "versionchange" for an upgrade transaction.\n     */\n    readonly mode: IDBTransactionMode;\n    /**\n     * Returns a list of the names of object stores in the transaction\'s scope. For an upgrade transaction this is all object stores in the database.\n     */\n    readonly objectStoreNames: DOMStringList;\n    onabort: ((this: IDBTransaction, ev: Event) => any) | null;\n    oncomplete: ((this: IDBTransaction, ev: Event) => any) | null;\n    onerror: ((this: IDBTransaction, ev: Event) => any) | null;\n    /**\n     * Aborts the transaction. All pending requests will fail with a "AbortError" DOMException and all changes made to the database will be reverted.\n     */\n    abort(): void;\n    /**\n     * Returns an IDBObjectStore in the transaction\'s scope.\n     */\n    objectStore(name: string): IDBObjectStore;\n    addEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof IDBTransactionEventMap>(type: K, listener: (this: IDBTransaction, ev: IDBTransactionEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var IDBTransaction: {\n    prototype: IDBTransaction;\n    new(): IDBTransaction;\n};\n\n/** This IndexedDB API interface indicates that the version of the database has changed, as the result of an IDBOpenDBRequest.onupgradeneeded event handler function. */\ninterface IDBVersionChangeEvent extends Event {\n    readonly newVersion: number | null;\n    readonly oldVersion: number;\n}\n\ndeclare var IDBVersionChangeEvent: {\n    prototype: IDBVersionChangeEvent;\n    new(type: string, eventInitDict?: IDBVersionChangeEventInit): IDBVersionChangeEvent;\n};\n\ninterface ImageBitmap {\n    /**\n     * Returns the intrinsic height of the image, in CSS pixels.\n     */\n    readonly height: number;\n    /**\n     * Returns the intrinsic width of the image, in CSS pixels.\n     */\n    readonly width: number;\n    /**\n     * Releases imageBitmap\'s underlying bitmap data.\n     */\n    close(): void;\n}\n\ndeclare var ImageBitmap: {\n    prototype: ImageBitmap;\n    new(): ImageBitmap;\n};\n\ninterface ImageBitmapRenderingContext {\n    /**\n     * Returns the canvas element that the context is bound to.\n     */\n    readonly canvas: OffscreenCanvas;\n    /**\n     * Transfers the underlying bitmap data from imageBitmap to context, and the bitmap becomes the contents of the canvas element to which context is bound.\n     */\n    transferFromImageBitmap(bitmap: ImageBitmap | null): void;\n}\n\ndeclare var ImageBitmapRenderingContext: {\n    prototype: ImageBitmapRenderingContext;\n    new(): ImageBitmapRenderingContext;\n};\n\n/** The underlying pixel data of an area of a <canvas> element. It is created using the ImageData() constructor or creator methods on the CanvasRenderingContext2D object associated with a canvas: createImageData() and getImageData(). It can also be used to set a part of the canvas by using putImageData(). */\ninterface ImageData {\n    /**\n     * Returns the one-dimensional array containing the data in RGBA order, as integers in the range 0 to 255.\n     */\n    readonly data: Uint8ClampedArray;\n    /**\n     * Returns the actual dimensions of the data in the ImageData object, in pixels.\n     */\n    readonly height: number;\n    /**\n     * Returns the actual dimensions of the data in the ImageData object, in pixels.\n     */\n    readonly width: number;\n}\n\ndeclare var ImageData: {\n    prototype: ImageData;\n    new(sw: number, sh: number): ImageData;\n    new(data: Uint8ClampedArray, sw: number, sh?: number): ImageData;\n};\n\n/** This Channel Messaging API interface allows us to create a new message channel and send data through it via its two MessagePort properties. */\ninterface MessageChannel {\n    /**\n     * Returns the first MessagePort object.\n     */\n    readonly port1: MessagePort;\n    /**\n     * Returns the second MessagePort object.\n     */\n    readonly port2: MessagePort;\n}\n\ndeclare var MessageChannel: {\n    prototype: MessageChannel;\n    new(): MessageChannel;\n};\n\n/** A message received by a target object. */\ninterface MessageEvent<T = any> extends Event {\n    /**\n     * Returns the data of the message.\n     */\n    readonly data: T;\n    /**\n     * Returns the last event ID string, for server-sent events.\n     */\n    readonly lastEventId: string;\n    /**\n     * Returns the origin of the message, for server-sent events and cross-document messaging.\n     */\n    readonly origin: string;\n    /**\n     * Returns the MessagePort array sent with the message, for cross-document messaging and channel messaging.\n     */\n    readonly ports: ReadonlyArray<MessagePort>;\n    /**\n     * Returns the WindowProxy of the source window, for cross-document messaging, and the MessagePort being attached, in the connect event fired at SharedWorkerGlobalScope objects.\n     */\n    readonly source: MessageEventSource | null;\n}\n\ndeclare var MessageEvent: {\n    prototype: MessageEvent;\n    new<T>(type: string, eventInitDict?: MessageEventInit<T>): MessageEvent<T>;\n};\n\ninterface MessagePortEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** This Channel Messaging API interface represents one of the two ports of a MessageChannel, allowing messages to be sent from one port and listening out for them arriving at the other. */\ninterface MessagePort extends EventTarget {\n    onmessage: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: MessagePort, ev: MessageEvent) => any) | null;\n    /**\n     * Disconnects the port, so that it is no longer active.\n     */\n    close(): void;\n    /**\n     * Posts a message through the channel. Objects listed in transfer are transferred, not just cloned, meaning that they are no longer usable on the sending side.\n     * \n     * Throws a "DataCloneError" DOMException if transfer contains duplicate objects or port, or if message could not be cloned.\n     */\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    /**\n     * Begins dispatching messages received on the port.\n     */\n    start(): void;\n    addEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof MessagePortEventMap>(type: K, listener: (this: MessagePort, ev: MessagePortEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var MessagePort: {\n    prototype: MessagePort;\n    new(): MessagePort;\n};\n\ninterface NavigationPreloadManager {\n    disable(): Promise<void>;\n    enable(): Promise<void>;\n    getState(): Promise<NavigationPreloadState>;\n    setHeaderValue(value: string): Promise<void>;\n}\n\ndeclare var NavigationPreloadManager: {\n    prototype: NavigationPreloadManager;\n    new(): NavigationPreloadManager;\n};\n\ninterface NavigatorConcurrentHardware {\n    readonly hardwareConcurrency: number;\n}\n\ninterface NavigatorID {\n    readonly appCodeName: string;\n    readonly appName: string;\n    readonly appVersion: string;\n    readonly platform: string;\n    readonly product: string;\n    readonly userAgent: string;\n}\n\ninterface NavigatorLanguage {\n    readonly language: string;\n    readonly languages: ReadonlyArray<string>;\n}\n\ninterface NavigatorOnLine {\n    readonly onLine: boolean;\n}\n\ninterface NavigatorStorage {\n    readonly storage: StorageManager;\n}\n\ninterface NotificationEventMap {\n    "click": Event;\n    "close": Event;\n    "error": Event;\n    "show": Event;\n}\n\n/** This Notifications API interface is used to configure and display desktop notifications to the user. */\ninterface Notification extends EventTarget {\n    readonly actions: ReadonlyArray<NotificationAction>;\n    readonly badge: string;\n    readonly body: string;\n    readonly data: any;\n    readonly dir: NotificationDirection;\n    readonly icon: string;\n    readonly image: string;\n    readonly lang: string;\n    onclick: ((this: Notification, ev: Event) => any) | null;\n    onclose: ((this: Notification, ev: Event) => any) | null;\n    onerror: ((this: Notification, ev: Event) => any) | null;\n    onshow: ((this: Notification, ev: Event) => any) | null;\n    readonly renotify: boolean;\n    readonly requireInteraction: boolean;\n    readonly silent: boolean;\n    readonly tag: string;\n    readonly timestamp: number;\n    readonly title: string;\n    readonly vibrate: ReadonlyArray<number>;\n    close(): void;\n    addEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof NotificationEventMap>(type: K, listener: (this: Notification, ev: NotificationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Notification: {\n    prototype: Notification;\n    new(title: string, options?: NotificationOptions): Notification;\n    readonly maxActions: number;\n    readonly permission: NotificationPermission;\n};\n\n/** The parameter passed into the onnotificationclick handler, the NotificationEvent interface represents a notification click event that is dispatched on the ServiceWorkerGlobalScope of a ServiceWorker. */\ninterface NotificationEvent extends ExtendableEvent {\n    readonly action: string;\n    readonly notification: Notification;\n}\n\ndeclare var NotificationEvent: {\n    prototype: NotificationEvent;\n    new(type: string, eventInitDict: NotificationEventInit): NotificationEvent;\n};\n\n/** The OES_element_index_uint extension is part of the WebGL API and adds support for gl.UNSIGNED_INT types to WebGLRenderingContext.drawElements(). */\ninterface OES_element_index_uint {\n}\n\n/** The OES_standard_derivatives extension is part of the WebGL API and adds the GLSL derivative functions dFdx, dFdy, and fwidth. */\ninterface OES_standard_derivatives {\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT_OES: GLenum;\n}\n\n/** The OES_texture_float extension is part of the WebGL API and exposes floating-point pixel types for textures. */\ninterface OES_texture_float {\n}\n\n/** The OES_texture_float_linear extension is part of the WebGL API and allows linear filtering with floating-point pixel types for textures. */\ninterface OES_texture_float_linear {\n}\n\n/** The OES_texture_half_float extension is part of the WebGL API and adds texture formats with 16- (aka half float) and 32-bit floating-point components. */\ninterface OES_texture_half_float {\n    readonly HALF_FLOAT_OES: GLenum;\n}\n\n/** The OES_texture_half_float_linear extension is part of the WebGL API and allows linear filtering with half floating-point pixel types for textures. */\ninterface OES_texture_half_float_linear {\n}\n\ninterface OES_vertex_array_object {\n    bindVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    createVertexArrayOES(): WebGLVertexArrayObjectOES | null;\n    deleteVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): void;\n    isVertexArrayOES(arrayObject: WebGLVertexArrayObjectOES | null): GLboolean;\n    readonly VERTEX_ARRAY_BINDING_OES: GLenum;\n}\n\ninterface OffscreenCanvas extends EventTarget {\n    /**\n     * These attributes return the dimensions of the OffscreenCanvas object\'s bitmap.\n     * \n     * They can be set, to replace the bitmap with a new, transparent black bitmap of the specified dimensions (effectively resizing it).\n     */\n    height: number;\n    /**\n     * These attributes return the dimensions of the OffscreenCanvas object\'s bitmap.\n     * \n     * They can be set, to replace the bitmap with a new, transparent black bitmap of the specified dimensions (effectively resizing it).\n     */\n    width: number;\n    /**\n     * Returns a promise that will fulfill with a new Blob object representing a file containing the image in the OffscreenCanvas object.\n     * \n     * The argument, if provided, is a dictionary that controls the encoding options of the image file to be created. The type field specifies the file format and has a default value of "image/png"; that type is also used if the requested type isn\'t supported. If the image format supports variable quality (such as "image/jpeg"), then the quality field is a number in the range 0.0 to 1.0 inclusive indicating the desired quality level for the resulting image.\n     */\n    convertToBlob(options?: ImageEncodeOptions): Promise<Blob>;\n    /**\n     * Returns an object that exposes an API for drawing on the OffscreenCanvas object. contextId specifies the desired API: "2d", "bitmaprenderer", "webgl", or "webgl2". options is handled by that API.\n     * \n     * This specification defines the "2d" context below, which is similar but distinct from the "2d" context that is created from a canvas element. The WebGL specifications define the "webgl" and "webgl2" contexts. [WEBGL]\n     * \n     * Returns null if the canvas has already been initialized with another context type (e.g., trying to get a "2d" context after getting a "webgl" context).\n     */\n    getContext(contextId: "2d", options?: CanvasRenderingContext2DSettings): OffscreenCanvasRenderingContext2D | null;\n    getContext(contextId: "bitmaprenderer", options?: ImageBitmapRenderingContextSettings): ImageBitmapRenderingContext | null;\n    getContext(contextId: "webgl", options?: WebGLContextAttributes): WebGLRenderingContext | null;\n    getContext(contextId: "webgl2", options?: WebGLContextAttributes): WebGL2RenderingContext | null;\n    getContext(contextId: OffscreenRenderingContextId, options?: any): OffscreenRenderingContext | null;\n    /**\n     * Returns a newly created ImageBitmap object with the image in the OffscreenCanvas object. The image in the OffscreenCanvas object is replaced with a new blank image.\n     */\n    transferToImageBitmap(): ImageBitmap;\n}\n\ndeclare var OffscreenCanvas: {\n    prototype: OffscreenCanvas;\n    new(width: number, height: number): OffscreenCanvas;\n};\n\ninterface OffscreenCanvasRenderingContext2D extends CanvasCompositing, CanvasDrawImage, CanvasDrawPath, CanvasFillStrokeStyles, CanvasFilters, CanvasImageData, CanvasImageSmoothing, CanvasPath, CanvasPathDrawingStyles, CanvasRect, CanvasShadowStyles, CanvasState, CanvasText, CanvasTextDrawingStyles, CanvasTransform {\n    readonly canvas: OffscreenCanvas;\n    commit(): void;\n}\n\ndeclare var OffscreenCanvasRenderingContext2D: {\n    prototype: OffscreenCanvasRenderingContext2D;\n    new(): OffscreenCanvasRenderingContext2D;\n};\n\n/** This Canvas 2D API interface is used to declare a path that can then be used on a CanvasRenderingContext2D object. The path methods of the CanvasRenderingContext2D interface are also present on this interface, which gives you the convenience of being able to retain and replay your path whenever desired. */\ninterface Path2D extends CanvasPath {\n    /**\n     * Adds to the path the path given by the argument.\n     */\n    addPath(path: Path2D, transform?: DOMMatrix2DInit): void;\n}\n\ndeclare var Path2D: {\n    prototype: Path2D;\n    new(path?: Path2D | string): Path2D;\n};\n\ninterface PerformanceEventMap {\n    "resourcetimingbufferfull": Event;\n}\n\n/** Provides access to performance-related information for the current page. It\'s part of the High Resolution Time API, but is enhanced by the Performance Timeline API, the Navigation Timing API, the User Timing API, and the Resource Timing API. */\ninterface Performance extends EventTarget {\n    onresourcetimingbufferfull: ((this: Performance, ev: Event) => any) | null;\n    readonly timeOrigin: number;\n    clearMarks(markName?: string): void;\n    clearMeasures(measureName?: string): void;\n    clearResourceTimings(): void;\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n    mark(markName: string): void;\n    measure(measureName: string, startMark?: string, endMark?: string): void;\n    now(): number;\n    setResourceTimingBufferSize(maxSize: number): void;\n    toJSON(): any;\n    addEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PerformanceEventMap>(type: K, listener: (this: Performance, ev: PerformanceEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Performance: {\n    prototype: Performance;\n    new(): Performance;\n};\n\n/** Encapsulates a single performance metric that is part of the performance timeline. A performance entry can be directly created by making a performance mark or measure (for example by calling the mark() method) at an explicit point in an application. Performance entries are also created in indirect ways such as loading a resource (such as an image). */\ninterface PerformanceEntry {\n    readonly duration: number;\n    readonly entryType: string;\n    readonly name: string;\n    readonly startTime: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceEntry: {\n    prototype: PerformanceEntry;\n    new(): PerformanceEntry;\n};\n\n/** PerformanceMark is an abstract interface for PerformanceEntry objects with an entryType of "mark". Entries of this type are created by calling performance.mark() to add a named DOMHighResTimeStamp (the mark) to the browser\'s performance timeline. */\ninterface PerformanceMark extends PerformanceEntry {\n}\n\ndeclare var PerformanceMark: {\n    prototype: PerformanceMark;\n    new(): PerformanceMark;\n};\n\n/** PerformanceMeasure is an abstract interface for PerformanceEntry objects with an entryType of "measure". Entries of this type are created by calling performance.measure() to add a named DOMHighResTimeStamp (the measure) between two marks to the browser\'s performance timeline. */\ninterface PerformanceMeasure extends PerformanceEntry {\n}\n\ndeclare var PerformanceMeasure: {\n    prototype: PerformanceMeasure;\n    new(): PerformanceMeasure;\n};\n\ninterface PerformanceObserver {\n    disconnect(): void;\n    observe(options?: PerformanceObserverInit): void;\n    takeRecords(): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserver: {\n    prototype: PerformanceObserver;\n    new(callback: PerformanceObserverCallback): PerformanceObserver;\n    readonly supportedEntryTypes: ReadonlyArray<string>;\n};\n\ninterface PerformanceObserverEntryList {\n    getEntries(): PerformanceEntryList;\n    getEntriesByName(name: string, type?: string): PerformanceEntryList;\n    getEntriesByType(type: string): PerformanceEntryList;\n}\n\ndeclare var PerformanceObserverEntryList: {\n    prototype: PerformanceObserverEntryList;\n    new(): PerformanceObserverEntryList;\n};\n\n/** Enables retrieval and analysis of detailed network timing data regarding the loading of an application\'s resources. An application can use the timing metrics to determine, for example, the length of time it takes to fetch a specific resource, such as an XMLHttpRequest, <SVG>, image, or script. */\ninterface PerformanceResourceTiming extends PerformanceEntry {\n    readonly connectEnd: number;\n    readonly connectStart: number;\n    readonly decodedBodySize: number;\n    readonly domainLookupEnd: number;\n    readonly domainLookupStart: number;\n    readonly encodedBodySize: number;\n    readonly fetchStart: number;\n    readonly initiatorType: string;\n    readonly nextHopProtocol: string;\n    readonly redirectEnd: number;\n    readonly redirectStart: number;\n    readonly requestStart: number;\n    readonly responseEnd: number;\n    readonly responseStart: number;\n    readonly secureConnectionStart: number;\n    readonly transferSize: number;\n    readonly workerStart: number;\n    toJSON(): any;\n}\n\ndeclare var PerformanceResourceTiming: {\n    prototype: PerformanceResourceTiming;\n    new(): PerformanceResourceTiming;\n};\n\ninterface PermissionStatusEventMap {\n    "change": Event;\n}\n\ninterface PermissionStatus extends EventTarget {\n    onchange: ((this: PermissionStatus, ev: Event) => any) | null;\n    readonly state: PermissionState;\n    addEventListener<K extends keyof PermissionStatusEventMap>(type: K, listener: (this: PermissionStatus, ev: PermissionStatusEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof PermissionStatusEventMap>(type: K, listener: (this: PermissionStatus, ev: PermissionStatusEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var PermissionStatus: {\n    prototype: PermissionStatus;\n    new(): PermissionStatus;\n};\n\ninterface Permissions {\n    query(permissionDesc: PermissionDescriptor | DevicePermissionDescriptor | MidiPermissionDescriptor | PushPermissionDescriptor): Promise<PermissionStatus>;\n}\n\ndeclare var Permissions: {\n    prototype: Permissions;\n    new(): Permissions;\n};\n\n/** Events measuring progress of an underlying process, like an HTTP request (for an XMLHttpRequest, or the loading of the underlying resource of an <img>, <audio>, <video>, <style> or <link>). */\ninterface ProgressEvent<T extends EventTarget = EventTarget> extends Event {\n    readonly lengthComputable: boolean;\n    readonly loaded: number;\n    readonly target: T | null;\n    readonly total: number;\n}\n\ndeclare var ProgressEvent: {\n    prototype: ProgressEvent;\n    new(type: string, eventInitDict?: ProgressEventInit): ProgressEvent;\n};\n\ninterface PromiseRejectionEvent extends Event {\n    readonly promise: Promise<any>;\n    readonly reason: any;\n}\n\ndeclare var PromiseRejectionEvent: {\n    prototype: PromiseRejectionEvent;\n    new(type: string, eventInitDict: PromiseRejectionEventInit): PromiseRejectionEvent;\n};\n\n/** This Push API interface represents a push message that has been received. This event is sent to the global scope of a ServiceWorker. It contains the information sent from an application server to a PushSubscription. */\ninterface PushEvent extends ExtendableEvent {\n    readonly data: PushMessageData | null;\n}\n\ndeclare var PushEvent: {\n    prototype: PushEvent;\n    new(type: string, eventInitDict?: PushEventInit): PushEvent;\n};\n\n/** This Push API interface provides a way to receive notifications from third-party servers as well as request URLs for push notifications. */\ninterface PushManager {\n    getSubscription(): Promise<PushSubscription | null>;\n    permissionState(options?: PushSubscriptionOptionsInit): Promise<PushPermissionState>;\n    subscribe(options?: PushSubscriptionOptionsInit): Promise<PushSubscription>;\n}\n\ndeclare var PushManager: {\n    prototype: PushManager;\n    new(): PushManager;\n    readonly supportedContentEncodings: ReadonlyArray<string>;\n};\n\n/** This Push API interface provides methods which let you retrieve the push data sent by a server in various formats. */\ninterface PushMessageData {\n    arrayBuffer(): ArrayBuffer;\n    blob(): Blob;\n    json(): any;\n    text(): string;\n}\n\ndeclare var PushMessageData: {\n    prototype: PushMessageData;\n    new(): PushMessageData;\n};\n\n/** This Push API interface provides a subcription\'s URL endpoint and allows unsubscription from a push service. */\ninterface PushSubscription {\n    readonly endpoint: string;\n    readonly expirationTime: number | null;\n    readonly options: PushSubscriptionOptions;\n    getKey(name: PushEncryptionKeyName): ArrayBuffer | null;\n    toJSON(): PushSubscriptionJSON;\n    unsubscribe(): Promise<boolean>;\n}\n\ndeclare var PushSubscription: {\n    prototype: PushSubscription;\n    new(): PushSubscription;\n};\n\ninterface PushSubscriptionChangeEvent extends ExtendableEvent {\n    readonly newSubscription: PushSubscription | null;\n    readonly oldSubscription: PushSubscription | null;\n}\n\ndeclare var PushSubscriptionChangeEvent: {\n    prototype: PushSubscriptionChangeEvent;\n    new(type: string, eventInitDict?: PushSubscriptionChangeEventInit): PushSubscriptionChangeEvent;\n};\n\ninterface PushSubscriptionOptions {\n    readonly applicationServerKey: ArrayBuffer | null;\n    readonly userVisibleOnly: boolean;\n}\n\ndeclare var PushSubscriptionOptions: {\n    prototype: PushSubscriptionOptions;\n    new(): PushSubscriptionOptions;\n};\n\ninterface ReadableByteStreamController {\n    readonly byobRequest: ReadableStreamBYOBRequest | undefined;\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: ArrayBufferView): void;\n    error(error?: any): void;\n}\n\ndeclare var ReadableByteStreamController: {\n    prototype: ReadableByteStreamController;\n    new(): ReadableByteStreamController;\n};\n\n/** This Streams API interface represents a readable stream of byte data. The Fetch API offers a concrete instance of a ReadableStream through the body property of a Response object. */\ninterface ReadableStream<R = any> {\n    readonly locked: boolean;\n    cancel(reason?: any): Promise<void>;\n    getReader(options: { mode: "byob" }): ReadableStreamBYOBReader;\n    getReader(): ReadableStreamDefaultReader<R>;\n    pipeThrough<T>({ writable, readable }: { writable: WritableStream<R>, readable: ReadableStream<T> }, options?: PipeOptions): ReadableStream<T>;\n    pipeTo(dest: WritableStream<R>, options?: PipeOptions): Promise<void>;\n    tee(): [ReadableStream<R>, ReadableStream<R>];\n}\n\ndeclare var ReadableStream: {\n    prototype: ReadableStream;\n    new(underlyingSource: UnderlyingByteSource, strategy?: { highWaterMark?: number, size?: undefined }): ReadableStream<Uint8Array>;\n    new<R = any>(underlyingSource?: UnderlyingSource<R>, strategy?: QueuingStrategy<R>): ReadableStream<R>;\n};\n\ninterface ReadableStreamBYOBReader {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read<T extends ArrayBufferView>(view: T): Promise<ReadableStreamReadResult<T>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamBYOBReader: {\n    prototype: ReadableStreamBYOBReader;\n    new(): ReadableStreamBYOBReader;\n};\n\ninterface ReadableStreamBYOBRequest {\n    readonly view: ArrayBufferView;\n    respond(bytesWritten: number): void;\n    respondWithNewView(view: ArrayBufferView): void;\n}\n\ndeclare var ReadableStreamBYOBRequest: {\n    prototype: ReadableStreamBYOBRequest;\n    new(): ReadableStreamBYOBRequest;\n};\n\ninterface ReadableStreamDefaultController<R = any> {\n    readonly desiredSize: number | null;\n    close(): void;\n    enqueue(chunk: R): void;\n    error(error?: any): void;\n}\n\ndeclare var ReadableStreamDefaultController: {\n    prototype: ReadableStreamDefaultController;\n    new(): ReadableStreamDefaultController;\n};\n\ninterface ReadableStreamDefaultReader<R = any> {\n    readonly closed: Promise<void>;\n    cancel(reason?: any): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamDefaultReader: {\n    prototype: ReadableStreamDefaultReader;\n    new(): ReadableStreamDefaultReader;\n};\n\ninterface ReadableStreamReader<R = any> {\n    cancel(): Promise<void>;\n    read(): Promise<ReadableStreamReadResult<R>>;\n    releaseLock(): void;\n}\n\ndeclare var ReadableStreamReader: {\n    prototype: ReadableStreamReader;\n    new(): ReadableStreamReader;\n};\n\n/** This Fetch API interface represents a resource request. */\ninterface Request extends Body {\n    /**\n     * Returns the cache mode associated with request, which is a string indicating how the request will interact with the browser\'s cache when fetching.\n     */\n    readonly cache: RequestCache;\n    /**\n     * Returns the credentials mode associated with request, which is a string indicating whether credentials will be sent with the request always, never, or only when sent to a same-origin URL.\n     */\n    readonly credentials: RequestCredentials;\n    /**\n     * Returns the kind of resource requested by request, e.g., "document" or "script".\n     */\n    readonly destination: RequestDestination;\n    /**\n     * Returns a Headers object consisting of the headers associated with request. Note that headers added in the network layer by the user agent will not be accounted for in this object, e.g., the "Host" header.\n     */\n    readonly headers: Headers;\n    /**\n     * Returns request\'s subresource integrity metadata, which is a cryptographic hash of the resource being fetched. Its value consists of multiple hashes separated by whitespace. [SRI]\n     */\n    readonly integrity: string;\n    /**\n     * Returns a boolean indicating whether or not request is for a history navigation (a.k.a. back-foward navigation).\n     */\n    readonly isHistoryNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request is for a reload navigation.\n     */\n    readonly isReloadNavigation: boolean;\n    /**\n     * Returns a boolean indicating whether or not request can outlive the global in which it was created.\n     */\n    readonly keepalive: boolean;\n    /**\n     * Returns request\'s HTTP method, which is "GET" by default.\n     */\n    readonly method: string;\n    /**\n     * Returns the mode associated with request, which is a string indicating whether the request will use CORS, or will be restricted to same-origin URLs.\n     */\n    readonly mode: RequestMode;\n    /**\n     * Returns the redirect mode associated with request, which is a string indicating how redirects for the request will be handled during fetching. A request will follow redirects by default.\n     */\n    readonly redirect: RequestRedirect;\n    /**\n     * Returns the referrer of request. Its value can be a same-origin URL if explicitly set in init, the empty string to indicate no referrer, and "about:client" when defaulting to the global\'s default. This is used during fetching to determine the value of the `Referer` header of the request being made.\n     */\n    readonly referrer: string;\n    /**\n     * Returns the referrer policy associated with request. This is used during fetching to compute the value of the request\'s referrer.\n     */\n    readonly referrerPolicy: ReferrerPolicy;\n    /**\n     * Returns the signal associated with request, which is an AbortSignal object indicating whether or not request has been aborted, and its abort event handler.\n     */\n    readonly signal: AbortSignal;\n    /**\n     * Returns the URL of request as a string.\n     */\n    readonly url: string;\n    clone(): Request;\n}\n\ndeclare var Request: {\n    prototype: Request;\n    new(input: RequestInfo, init?: RequestInit): Request;\n};\n\n/** This Fetch API interface represents the response to a request. */\ninterface Response extends Body {\n    readonly headers: Headers;\n    readonly ok: boolean;\n    readonly redirected: boolean;\n    readonly status: number;\n    readonly statusText: string;\n    readonly trailer: Promise<Headers>;\n    readonly type: ResponseType;\n    readonly url: string;\n    clone(): Response;\n}\n\ndeclare var Response: {\n    prototype: Response;\n    new(body?: BodyInit | null, init?: ResponseInit): Response;\n    error(): Response;\n    redirect(url: string, status?: number): Response;\n};\n\ninterface ServiceWorkerEventMap extends AbstractWorkerEventMap {\n    "statechange": Event;\n}\n\n/** This ServiceWorker API interface provides a reference to a service worker. Multiple browsing contexts (e.g. pages, workers, etc.) can be associated with the same service worker, each through a unique ServiceWorker object. */\ninterface ServiceWorker extends EventTarget, AbstractWorker {\n    onstatechange: ((this: ServiceWorker, ev: Event) => any) | null;\n    readonly scriptURL: string;\n    readonly state: ServiceWorkerState;\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    addEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerEventMap>(type: K, listener: (this: ServiceWorker, ev: ServiceWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorker: {\n    prototype: ServiceWorker;\n    new(): ServiceWorker;\n};\n\ninterface ServiceWorkerContainerEventMap {\n    "controllerchange": Event;\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** The ServiceWorkerContainer interface of the ServiceWorker API provides an object representing the service worker as an overall unit in the network ecosystem, including facilities to register, unregister and update service workers, and access the state of service workers and their registrations. */\ninterface ServiceWorkerContainer extends EventTarget {\n    readonly controller: ServiceWorker | null;\n    oncontrollerchange: ((this: ServiceWorkerContainer, ev: Event) => any) | null;\n    onmessage: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: ServiceWorkerContainer, ev: MessageEvent) => any) | null;\n    readonly ready: Promise<ServiceWorkerRegistration>;\n    getRegistration(clientURL?: string): Promise<ServiceWorkerRegistration | undefined>;\n    getRegistrations(): Promise<ReadonlyArray<ServiceWorkerRegistration>>;\n    register(scriptURL: string, options?: RegistrationOptions): Promise<ServiceWorkerRegistration>;\n    startMessages(): void;\n    addEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerContainerEventMap>(type: K, listener: (this: ServiceWorkerContainer, ev: ServiceWorkerContainerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerContainer: {\n    prototype: ServiceWorkerContainer;\n    new(): ServiceWorkerContainer;\n};\n\ninterface ServiceWorkerGlobalScopeEventMap extends WorkerGlobalScopeEventMap {\n    "activate": ExtendableEvent;\n    "fetch": FetchEvent;\n    "install": ExtendableEvent;\n    "message": ExtendableMessageEvent;\n    "messageerror": MessageEvent;\n    "notificationclick": NotificationEvent;\n    "notificationclose": NotificationEvent;\n    "push": PushEvent;\n    "pushsubscriptionchange": PushSubscriptionChangeEvent;\n    "sync": SyncEvent;\n}\n\n/** This ServiceWorker API interface represents the global execution context of a service worker. */\ninterface ServiceWorkerGlobalScope extends WorkerGlobalScope {\n    readonly clients: Clients;\n    onactivate: ((this: ServiceWorkerGlobalScope, ev: ExtendableEvent) => any) | null;\n    onfetch: ((this: ServiceWorkerGlobalScope, ev: FetchEvent) => any) | null;\n    oninstall: ((this: ServiceWorkerGlobalScope, ev: ExtendableEvent) => any) | null;\n    onmessage: ((this: ServiceWorkerGlobalScope, ev: ExtendableMessageEvent) => any) | null;\n    onmessageerror: ((this: ServiceWorkerGlobalScope, ev: MessageEvent) => any) | null;\n    onnotificationclick: ((this: ServiceWorkerGlobalScope, ev: NotificationEvent) => any) | null;\n    onnotificationclose: ((this: ServiceWorkerGlobalScope, ev: NotificationEvent) => any) | null;\n    onpush: ((this: ServiceWorkerGlobalScope, ev: PushEvent) => any) | null;\n    onpushsubscriptionchange: ((this: ServiceWorkerGlobalScope, ev: PushSubscriptionChangeEvent) => any) | null;\n    onsync: ((this: ServiceWorkerGlobalScope, ev: SyncEvent) => any) | null;\n    readonly registration: ServiceWorkerRegistration;\n    readonly serviceWorker: ServiceWorker;\n    skipWaiting(): Promise<void>;\n    addEventListener<K extends keyof ServiceWorkerGlobalScopeEventMap>(type: K, listener: (this: ServiceWorkerGlobalScope, ev: ServiceWorkerGlobalScopeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerGlobalScopeEventMap>(type: K, listener: (this: ServiceWorkerGlobalScope, ev: ServiceWorkerGlobalScopeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerGlobalScope: {\n    prototype: ServiceWorkerGlobalScope;\n    new(): ServiceWorkerGlobalScope;\n};\n\ninterface ServiceWorkerRegistrationEventMap {\n    "updatefound": Event;\n}\n\n/** This ServiceWorker API interface represents the service worker registration. You register a service worker to control one or more pages that share the same origin. */\ninterface ServiceWorkerRegistration extends EventTarget {\n    readonly active: ServiceWorker | null;\n    readonly installing: ServiceWorker | null;\n    readonly navigationPreload: NavigationPreloadManager;\n    onupdatefound: ((this: ServiceWorkerRegistration, ev: Event) => any) | null;\n    readonly pushManager: PushManager;\n    readonly scope: string;\n    readonly sync: SyncManager;\n    readonly updateViaCache: ServiceWorkerUpdateViaCache;\n    readonly waiting: ServiceWorker | null;\n    getNotifications(filter?: GetNotificationOptions): Promise<Notification[]>;\n    showNotification(title: string, options?: NotificationOptions): Promise<void>;\n    unregister(): Promise<boolean>;\n    update(): Promise<void>;\n    addEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof ServiceWorkerRegistrationEventMap>(type: K, listener: (this: ServiceWorkerRegistration, ev: ServiceWorkerRegistrationEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var ServiceWorkerRegistration: {\n    prototype: ServiceWorkerRegistration;\n    new(): ServiceWorkerRegistration;\n};\n\ninterface SharedWorker extends EventTarget, AbstractWorker {\n    /**\n     * Returns sharedWorker\'s MessagePort object which can be used to communicate with the global environment.\n     */\n    readonly port: MessagePort;\n    addEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: SharedWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof AbstractWorkerEventMap>(type: K, listener: (this: SharedWorker, ev: AbstractWorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SharedWorker: {\n    prototype: SharedWorker;\n    new(scriptURL: string, options?: string | WorkerOptions): SharedWorker;\n};\n\ninterface SharedWorkerGlobalScopeEventMap extends WorkerGlobalScopeEventMap {\n    "connect": MessageEvent;\n}\n\ninterface SharedWorkerGlobalScope extends WorkerGlobalScope {\n    /**\n     * Returns sharedWorkerGlobal\'s name, i.e. the value given to the SharedWorker constructor. Multiple SharedWorker objects can correspond to the same shared worker (and SharedWorkerGlobalScope), by reusing the same name.\n     */\n    readonly name: string;\n    onconnect: ((this: SharedWorkerGlobalScope, ev: MessageEvent) => any) | null;\n    /**\n     * Aborts sharedWorkerGlobal.\n     */\n    close(): void;\n    addEventListener<K extends keyof SharedWorkerGlobalScopeEventMap>(type: K, listener: (this: SharedWorkerGlobalScope, ev: SharedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof SharedWorkerGlobalScopeEventMap>(type: K, listener: (this: SharedWorkerGlobalScope, ev: SharedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var SharedWorkerGlobalScope: {\n    prototype: SharedWorkerGlobalScope;\n    new(): SharedWorkerGlobalScope;\n};\n\ninterface StorageManager {\n    estimate(): Promise<StorageEstimate>;\n    persisted(): Promise<boolean>;\n}\n\ndeclare var StorageManager: {\n    prototype: StorageManager;\n    new(): StorageManager;\n};\n\n/** This Web Crypto API interface provides a number of low-level cryptographic functions. It is accessed via the Crypto.subtle properties available in a window context (via Window.crypto). */\ninterface SubtleCrypto {\n    decrypt(algorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    deriveBits(algorithm: AlgorithmIdentifier | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfParams | Pbkdf2Params, baseKey: CryptoKey, length: number): Promise<ArrayBuffer>;\n    deriveKey(algorithm: AlgorithmIdentifier | EcdhKeyDeriveParams | DhKeyDeriveParams | ConcatParams | HkdfParams | Pbkdf2Params, baseKey: CryptoKey, derivedKeyType: string | AesDerivedKeyParams | HmacImportParams | ConcatParams | HkdfParams | Pbkdf2Params, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    digest(algorithm: AlgorithmIdentifier, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    encrypt(algorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    exportKey(format: "jwk", key: CryptoKey): Promise<JsonWebKey>;\n    exportKey(format: "raw" | "pkcs8" | "spki", key: CryptoKey): Promise<ArrayBuffer>;\n    exportKey(format: string, key: CryptoKey): Promise<JsonWebKey | ArrayBuffer>;\n    generateKey(algorithm: RsaHashedKeyGenParams | EcKeyGenParams | DhKeyGenParams, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKeyPair>;\n    generateKey(algorithm: AesKeyGenParams | HmacKeyGenParams | Pbkdf2Params, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    generateKey(algorithm: AlgorithmIdentifier, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKeyPair | CryptoKey>;\n    importKey(format: "jwk", keyData: JsonWebKey, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    importKey(format: "raw" | "pkcs8" | "spki", keyData: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    importKey(format: string, keyData: JsonWebKey | Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, algorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    sign(algorithm: AlgorithmIdentifier | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<ArrayBuffer>;\n    unwrapKey(format: "raw" | "pkcs8" | "spki" | "jwk" | string, wrappedKey: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, unwrappingKey: CryptoKey, unwrapAlgorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams, unwrappedKeyAlgorithm: AlgorithmIdentifier | RsaHashedImportParams | EcKeyImportParams | HmacImportParams | DhImportKeyParams | AesKeyAlgorithm, extractable: boolean, keyUsages: KeyUsage[]): Promise<CryptoKey>;\n    verify(algorithm: AlgorithmIdentifier | RsaPssParams | EcdsaParams | AesCmacParams, key: CryptoKey, signature: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer, data: Int8Array | Int16Array | Int32Array | Uint8Array | Uint16Array | Uint32Array | Uint8ClampedArray | Float32Array | Float64Array | DataView | ArrayBuffer): Promise<boolean>;\n    wrapKey(format: "raw" | "pkcs8" | "spki" | "jwk" | string, key: CryptoKey, wrappingKey: CryptoKey, wrapAlgorithm: AlgorithmIdentifier | RsaOaepParams | AesCtrParams | AesCbcParams | AesCmacParams | AesGcmParams | AesCfbParams): Promise<ArrayBuffer>;\n}\n\ndeclare var SubtleCrypto: {\n    prototype: SubtleCrypto;\n    new(): SubtleCrypto;\n};\n\n/** A sync action that is dispatched on the ServiceWorkerGlobalScope of a ServiceWorker.  */\ninterface SyncEvent extends ExtendableEvent {\n    readonly lastChance: boolean;\n    readonly tag: string;\n}\n\ndeclare var SyncEvent: {\n    prototype: SyncEvent;\n    new(type: string, init: SyncEventInit): SyncEvent;\n};\n\n/** This ServiceWorker API interface provides an interface for registering and listing sync registrations. */\ninterface SyncManager {\n    getTags(): Promise<string[]>;\n    register(tag: string): Promise<void>;\n}\n\ndeclare var SyncManager: {\n    prototype: SyncManager;\n    new(): SyncManager;\n};\n\n/** A decoder for a specific method, that is a specific character encoding, like utf-8, iso-8859-2, koi8, cp1261, gbk, etc. A decoder takes a stream of bytes as input and emits a stream of code points. For a more scalable, non-native library, see StringView – a C-like representation of strings based on typed arrays. */\ninterface TextDecoder extends TextDecoderCommon {\n    /**\n     * Returns the result of running encoding\'s decoder. The method can be invoked zero or more times with options\'s stream set to true, and then once without options\'s stream (or set to false), to process a fragmented stream. If the invocation without options\'s stream (or set to false) has no input, it\'s clearest to omit both arguments.\n     * \n     * ```\n     * var string = "", decoder = new TextDecoder(encoding), buffer;\n     * while(buffer = next_chunk()) {\n     *   string += decoder.decode(buffer, {stream:true});\n     * }\n     * string += decoder.decode(); // end-of-stream\n     * ```\n     * \n     * If the error mode is "fatal" and encoding\'s decoder returns error, throws a TypeError.\n     */\n    decode(input?: BufferSource, options?: TextDecodeOptions): string;\n}\n\ndeclare var TextDecoder: {\n    prototype: TextDecoder;\n    new(label?: string, options?: TextDecoderOptions): TextDecoder;\n};\n\ninterface TextDecoderCommon {\n    /**\n     * Returns encoding\'s name, lowercased.\n     */\n    readonly encoding: string;\n    /**\n     * Returns true if error mode is "fatal", and false otherwise.\n     */\n    readonly fatal: boolean;\n    /**\n     * Returns true if ignore BOM flag is set, and false otherwise.\n     */\n    readonly ignoreBOM: boolean;\n}\n\ninterface TextDecoderStream extends GenericTransformStream, TextDecoderCommon {\n    readonly readable: ReadableStream<string>;\n    readonly writable: WritableStream<BufferSource>;\n}\n\ndeclare var TextDecoderStream: {\n    prototype: TextDecoderStream;\n    new(label?: string, options?: TextDecoderOptions): TextDecoderStream;\n};\n\n/** TextEncoder takes a stream of code points as input and emits a stream of bytes. For a more scalable, non-native library, see StringView – a C-like representation of strings based on typed arrays. */\ninterface TextEncoder extends TextEncoderCommon {\n    /**\n     * Returns the result of running UTF-8\'s encoder.\n     */\n    encode(input?: string): Uint8Array;\n    /**\n     * Runs the UTF-8 encoder on source, stores the result of that operation into destination, and returns the progress made as a dictionary whereby read is the number of converted code units of source and written is the number of bytes modified in destination.\n     */\n    encodeInto(source: string, destination: Uint8Array): TextEncoderEncodeIntoResult;\n}\n\ndeclare var TextEncoder: {\n    prototype: TextEncoder;\n    new(): TextEncoder;\n};\n\ninterface TextEncoderCommon {\n    /**\n     * Returns "utf-8".\n     */\n    readonly encoding: string;\n}\n\ninterface TextEncoderStream extends GenericTransformStream, TextEncoderCommon {\n    readonly readable: ReadableStream<Uint8Array>;\n    readonly writable: WritableStream<string>;\n}\n\ndeclare var TextEncoderStream: {\n    prototype: TextEncoderStream;\n    new(): TextEncoderStream;\n};\n\n/** The dimensions of a piece of text in the canvas, as created by the CanvasRenderingContext2D.measureText() method. */\ninterface TextMetrics {\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxLeft: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly actualBoundingBoxRight: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly alphabeticBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly emHeightAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly emHeightDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly fontBoundingBoxAscent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly fontBoundingBoxDescent: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly hangingBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly ideographicBaseline: number;\n    /**\n     * Returns the measurement described below.\n     */\n    readonly width: number;\n}\n\ndeclare var TextMetrics: {\n    prototype: TextMetrics;\n    new(): TextMetrics;\n};\n\ninterface TransformStream<I = any, O = any> {\n    readonly readable: ReadableStream<O>;\n    readonly writable: WritableStream<I>;\n}\n\ndeclare var TransformStream: {\n    prototype: TransformStream;\n    new<I = any, O = any>(transformer?: Transformer<I, O>, writableStrategy?: QueuingStrategy<I>, readableStrategy?: QueuingStrategy<O>): TransformStream<I, O>;\n};\n\ninterface TransformStreamDefaultController<O = any> {\n    readonly desiredSize: number | null;\n    enqueue(chunk: O): void;\n    error(reason?: any): void;\n    terminate(): void;\n}\n\ndeclare var TransformStreamDefaultController: {\n    prototype: TransformStreamDefaultController;\n    new(): TransformStreamDefaultController;\n};\n\n/** The URL interface represents an object providing static methods used for creating object URLs. */\ninterface URL {\n    hash: string;\n    host: string;\n    hostname: string;\n    href: string;\n    toString(): string;\n    readonly origin: string;\n    password: string;\n    pathname: string;\n    port: string;\n    protocol: string;\n    search: string;\n    readonly searchParams: URLSearchParams;\n    username: string;\n    toJSON(): string;\n}\n\ndeclare var URL: {\n    prototype: URL;\n    new(url: string, base?: string | URL): URL;\n    createObjectURL(object: any): string;\n    revokeObjectURL(url: string): void;\n};\n\ninterface URLSearchParams {\n    /**\n     * Appends a specified key/value pair as a new search parameter.\n     */\n    append(name: string, value: string): void;\n    /**\n     * Deletes the given search parameter, and its associated value, from the list of all search parameters.\n     */\n    delete(name: string): void;\n    /**\n     * Returns the first value associated to the given search parameter.\n     */\n    get(name: string): string | null;\n    /**\n     * Returns all the values association with a given search parameter.\n     */\n    getAll(name: string): string[];\n    /**\n     * Returns a Boolean indicating if such a search parameter exists.\n     */\n    has(name: string): boolean;\n    /**\n     * Sets the value associated to a given search parameter to the given value. If there were several values, delete the others.\n     */\n    set(name: string, value: string): void;\n    sort(): void;\n    /**\n     * Returns a string containing a query string suitable for use in a URL. Does not include the question mark.\n     */\n    toString(): string;\n    forEach(callbackfn: (value: string, key: string, parent: URLSearchParams) => void, thisArg?: any): void;\n}\n\ndeclare var URLSearchParams: {\n    prototype: URLSearchParams;\n    new(init?: string[][] | Record<string, string> | string | URLSearchParams): URLSearchParams;\n    toString(): string;\n};\n\ninterface WEBGL_color_buffer_float {\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE_EXT: GLenum;\n    readonly RGBA32F_EXT: GLenum;\n    readonly UNSIGNED_NORMALIZED_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_astc {\n    getSupportedProfiles(): string[];\n    readonly COMPRESSED_RGBA_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_RGBA_ASTC_8x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_10x8_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x10_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_4x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x4_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_5x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x5_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x6_KHR: GLenum;\n    readonly COMPRESSED_SRGB8_ALPHA8_ASTC_8x8_KHR: GLenum;\n}\n\n/** The WEBGL_compressed_texture_s3tc extension is part of the WebGL API and exposes four S3TC compressed texture formats. */\ninterface WEBGL_compressed_texture_s3tc {\n    readonly COMPRESSED_RGBA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_RGBA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_RGB_S3TC_DXT1_EXT: GLenum;\n}\n\ninterface WEBGL_compressed_texture_s3tc_srgb {\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT1_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT3_EXT: GLenum;\n    readonly COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT: GLenum;\n    readonly COMPRESSED_SRGB_S3TC_DXT1_EXT: GLenum;\n}\n\n/** The WEBGL_debug_renderer_info extension is part of the WebGL API and exposes two constants with information about the graphics driver for debugging purposes. */\ninterface WEBGL_debug_renderer_info {\n    readonly UNMASKED_RENDERER_WEBGL: GLenum;\n    readonly UNMASKED_VENDOR_WEBGL: GLenum;\n}\n\ninterface WEBGL_debug_shaders {\n    getTranslatedShaderSource(shader: WebGLShader): string;\n}\n\n/** The WEBGL_depth_texture extension is part of the WebGL API and defines 2D depth and depth-stencil textures. */\ninterface WEBGL_depth_texture {\n    readonly UNSIGNED_INT_24_8_WEBGL: GLenum;\n}\n\ninterface WEBGL_draw_buffers {\n    drawBuffersWEBGL(buffers: GLenum[]): void;\n    readonly COLOR_ATTACHMENT0_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT10_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT11_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT12_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT13_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT14_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT15_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT1_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT2_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT3_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT4_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT5_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT6_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT7_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT8_WEBGL: GLenum;\n    readonly COLOR_ATTACHMENT9_WEBGL: GLenum;\n    readonly DRAW_BUFFER0_WEBGL: GLenum;\n    readonly DRAW_BUFFER10_WEBGL: GLenum;\n    readonly DRAW_BUFFER11_WEBGL: GLenum;\n    readonly DRAW_BUFFER12_WEBGL: GLenum;\n    readonly DRAW_BUFFER13_WEBGL: GLenum;\n    readonly DRAW_BUFFER14_WEBGL: GLenum;\n    readonly DRAW_BUFFER15_WEBGL: GLenum;\n    readonly DRAW_BUFFER1_WEBGL: GLenum;\n    readonly DRAW_BUFFER2_WEBGL: GLenum;\n    readonly DRAW_BUFFER3_WEBGL: GLenum;\n    readonly DRAW_BUFFER4_WEBGL: GLenum;\n    readonly DRAW_BUFFER5_WEBGL: GLenum;\n    readonly DRAW_BUFFER6_WEBGL: GLenum;\n    readonly DRAW_BUFFER7_WEBGL: GLenum;\n    readonly DRAW_BUFFER8_WEBGL: GLenum;\n    readonly DRAW_BUFFER9_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS_WEBGL: GLenum;\n    readonly MAX_DRAW_BUFFERS_WEBGL: GLenum;\n}\n\ninterface WEBGL_lose_context {\n    loseContext(): void;\n    restoreContext(): void;\n}\n\ninterface WebGL2RenderingContext extends WebGL2RenderingContextBase, WebGL2RenderingContextOverloads, WebGLRenderingContextBase {\n}\n\ndeclare var WebGL2RenderingContext: {\n    prototype: WebGL2RenderingContext;\n    new(): WebGL2RenderingContext;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n    readonly ACTIVE_UNIFORM_BLOCKS: GLenum;\n    readonly ALREADY_SIGNALED: GLenum;\n    readonly ANY_SAMPLES_PASSED: GLenum;\n    readonly ANY_SAMPLES_PASSED_CONSERVATIVE: GLenum;\n    readonly COLOR: GLenum;\n    readonly COLOR_ATTACHMENT1: GLenum;\n    readonly COLOR_ATTACHMENT10: GLenum;\n    readonly COLOR_ATTACHMENT11: GLenum;\n    readonly COLOR_ATTACHMENT12: GLenum;\n    readonly COLOR_ATTACHMENT13: GLenum;\n    readonly COLOR_ATTACHMENT14: GLenum;\n    readonly COLOR_ATTACHMENT15: GLenum;\n    readonly COLOR_ATTACHMENT2: GLenum;\n    readonly COLOR_ATTACHMENT3: GLenum;\n    readonly COLOR_ATTACHMENT4: GLenum;\n    readonly COLOR_ATTACHMENT5: GLenum;\n    readonly COLOR_ATTACHMENT6: GLenum;\n    readonly COLOR_ATTACHMENT7: GLenum;\n    readonly COLOR_ATTACHMENT8: GLenum;\n    readonly COLOR_ATTACHMENT9: GLenum;\n    readonly COMPARE_REF_TO_TEXTURE: GLenum;\n    readonly CONDITION_SATISFIED: GLenum;\n    readonly COPY_READ_BUFFER: GLenum;\n    readonly COPY_READ_BUFFER_BINDING: GLenum;\n    readonly COPY_WRITE_BUFFER: GLenum;\n    readonly COPY_WRITE_BUFFER_BINDING: GLenum;\n    readonly CURRENT_QUERY: GLenum;\n    readonly DEPTH: GLenum;\n    readonly DEPTH24_STENCIL8: GLenum;\n    readonly DEPTH32F_STENCIL8: GLenum;\n    readonly DEPTH_COMPONENT24: GLenum;\n    readonly DEPTH_COMPONENT32F: GLenum;\n    readonly DRAW_BUFFER0: GLenum;\n    readonly DRAW_BUFFER1: GLenum;\n    readonly DRAW_BUFFER10: GLenum;\n    readonly DRAW_BUFFER11: GLenum;\n    readonly DRAW_BUFFER12: GLenum;\n    readonly DRAW_BUFFER13: GLenum;\n    readonly DRAW_BUFFER14: GLenum;\n    readonly DRAW_BUFFER15: GLenum;\n    readonly DRAW_BUFFER2: GLenum;\n    readonly DRAW_BUFFER3: GLenum;\n    readonly DRAW_BUFFER4: GLenum;\n    readonly DRAW_BUFFER5: GLenum;\n    readonly DRAW_BUFFER6: GLenum;\n    readonly DRAW_BUFFER7: GLenum;\n    readonly DRAW_BUFFER8: GLenum;\n    readonly DRAW_BUFFER9: GLenum;\n    readonly DRAW_FRAMEBUFFER: GLenum;\n    readonly DRAW_FRAMEBUFFER_BINDING: GLenum;\n    readonly DYNAMIC_COPY: GLenum;\n    readonly DYNAMIC_READ: GLenum;\n    readonly FLOAT_32_UNSIGNED_INT_24_8_REV: GLenum;\n    readonly FLOAT_MAT2x3: GLenum;\n    readonly FLOAT_MAT2x4: GLenum;\n    readonly FLOAT_MAT3x2: GLenum;\n    readonly FLOAT_MAT3x4: GLenum;\n    readonly FLOAT_MAT4x2: GLenum;\n    readonly FLOAT_MAT4x3: GLenum;\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_BLUE_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_GREEN_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_RED_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER: GLenum;\n    readonly FRAMEBUFFER_DEFAULT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: GLenum;\n    readonly HALF_FLOAT: GLenum;\n    readonly INTERLEAVED_ATTRIBS: GLenum;\n    readonly INT_2_10_10_10_REV: GLenum;\n    readonly INT_SAMPLER_2D: GLenum;\n    readonly INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly INT_SAMPLER_3D: GLenum;\n    readonly INT_SAMPLER_CUBE: GLenum;\n    readonly INVALID_INDEX: GLenum;\n    readonly MAX: GLenum;\n    readonly MAX_3D_TEXTURE_SIZE: GLenum;\n    readonly MAX_ARRAY_TEXTURE_LAYERS: GLenum;\n    readonly MAX_CLIENT_WAIT_TIMEOUT_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS: GLenum;\n    readonly MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_COMBINED_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_DRAW_BUFFERS: GLenum;\n    readonly MAX_ELEMENTS_INDICES: GLenum;\n    readonly MAX_ELEMENTS_VERTICES: GLenum;\n    readonly MAX_ELEMENT_INDEX: GLenum;\n    readonly MAX_FRAGMENT_INPUT_COMPONENTS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly MAX_SAMPLES: GLenum;\n    readonly MAX_SERVER_WAIT_TIMEOUT: GLenum;\n    readonly MAX_TEXTURE_LOD_BIAS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS: GLenum;\n    readonly MAX_UNIFORM_BLOCK_SIZE: GLenum;\n    readonly MAX_UNIFORM_BUFFER_BINDINGS: GLenum;\n    readonly MAX_VARYING_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_OUTPUT_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MIN: GLenum;\n    readonly MIN_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly OBJECT_TYPE: GLenum;\n    readonly PACK_ROW_LENGTH: GLenum;\n    readonly PACK_SKIP_PIXELS: GLenum;\n    readonly PACK_SKIP_ROWS: GLenum;\n    readonly PIXEL_PACK_BUFFER: GLenum;\n    readonly PIXEL_PACK_BUFFER_BINDING: GLenum;\n    readonly PIXEL_UNPACK_BUFFER: GLenum;\n    readonly PIXEL_UNPACK_BUFFER_BINDING: GLenum;\n    readonly QUERY_RESULT: GLenum;\n    readonly QUERY_RESULT_AVAILABLE: GLenum;\n    readonly R11F_G11F_B10F: GLenum;\n    readonly R16F: GLenum;\n    readonly R16I: GLenum;\n    readonly R16UI: GLenum;\n    readonly R32F: GLenum;\n    readonly R32I: GLenum;\n    readonly R32UI: GLenum;\n    readonly R8: GLenum;\n    readonly R8I: GLenum;\n    readonly R8UI: GLenum;\n    readonly R8_SNORM: GLenum;\n    readonly RASTERIZER_DISCARD: GLenum;\n    readonly READ_BUFFER: GLenum;\n    readonly READ_FRAMEBUFFER: GLenum;\n    readonly READ_FRAMEBUFFER_BINDING: GLenum;\n    readonly RED: GLenum;\n    readonly RED_INTEGER: GLenum;\n    readonly RENDERBUFFER_SAMPLES: GLenum;\n    readonly RG: GLenum;\n    readonly RG16F: GLenum;\n    readonly RG16I: GLenum;\n    readonly RG16UI: GLenum;\n    readonly RG32F: GLenum;\n    readonly RG32I: GLenum;\n    readonly RG32UI: GLenum;\n    readonly RG8: GLenum;\n    readonly RG8I: GLenum;\n    readonly RG8UI: GLenum;\n    readonly RG8_SNORM: GLenum;\n    readonly RGB10_A2: GLenum;\n    readonly RGB10_A2UI: GLenum;\n    readonly RGB16F: GLenum;\n    readonly RGB16I: GLenum;\n    readonly RGB16UI: GLenum;\n    readonly RGB32F: GLenum;\n    readonly RGB32I: GLenum;\n    readonly RGB32UI: GLenum;\n    readonly RGB8: GLenum;\n    readonly RGB8I: GLenum;\n    readonly RGB8UI: GLenum;\n    readonly RGB8_SNORM: GLenum;\n    readonly RGB9_E5: GLenum;\n    readonly RGBA16F: GLenum;\n    readonly RGBA16I: GLenum;\n    readonly RGBA16UI: GLenum;\n    readonly RGBA32F: GLenum;\n    readonly RGBA32I: GLenum;\n    readonly RGBA32UI: GLenum;\n    readonly RGBA8: GLenum;\n    readonly RGBA8I: GLenum;\n    readonly RGBA8UI: GLenum;\n    readonly RGBA8_SNORM: GLenum;\n    readonly RGBA_INTEGER: GLenum;\n    readonly RGB_INTEGER: GLenum;\n    readonly RG_INTEGER: GLenum;\n    readonly SAMPLER_2D_ARRAY: GLenum;\n    readonly SAMPLER_2D_ARRAY_SHADOW: GLenum;\n    readonly SAMPLER_2D_SHADOW: GLenum;\n    readonly SAMPLER_3D: GLenum;\n    readonly SAMPLER_BINDING: GLenum;\n    readonly SAMPLER_CUBE_SHADOW: GLenum;\n    readonly SEPARATE_ATTRIBS: GLenum;\n    readonly SIGNALED: GLenum;\n    readonly SIGNED_NORMALIZED: GLenum;\n    readonly SRGB: GLenum;\n    readonly SRGB8: GLenum;\n    readonly SRGB8_ALPHA8: GLenum;\n    readonly STATIC_COPY: GLenum;\n    readonly STATIC_READ: GLenum;\n    readonly STENCIL: GLenum;\n    readonly STREAM_COPY: GLenum;\n    readonly STREAM_READ: GLenum;\n    readonly SYNC_CONDITION: GLenum;\n    readonly SYNC_FENCE: GLenum;\n    readonly SYNC_FLAGS: GLenum;\n    readonly SYNC_FLUSH_COMMANDS_BIT: GLenum;\n    readonly SYNC_GPU_COMMANDS_COMPLETE: GLenum;\n    readonly SYNC_STATUS: GLenum;\n    readonly TEXTURE_2D_ARRAY: GLenum;\n    readonly TEXTURE_3D: GLenum;\n    readonly TEXTURE_BASE_LEVEL: GLenum;\n    readonly TEXTURE_BINDING_2D_ARRAY: GLenum;\n    readonly TEXTURE_BINDING_3D: GLenum;\n    readonly TEXTURE_COMPARE_FUNC: GLenum;\n    readonly TEXTURE_COMPARE_MODE: GLenum;\n    readonly TEXTURE_IMMUTABLE_FORMAT: GLenum;\n    readonly TEXTURE_IMMUTABLE_LEVELS: GLenum;\n    readonly TEXTURE_MAX_LEVEL: GLenum;\n    readonly TEXTURE_MAX_LOD: GLenum;\n    readonly TEXTURE_MIN_LOD: GLenum;\n    readonly TEXTURE_WRAP_R: GLenum;\n    readonly TIMEOUT_EXPIRED: GLenum;\n    readonly TIMEOUT_IGNORED: GLint64;\n    readonly TRANSFORM_FEEDBACK: GLenum;\n    readonly TRANSFORM_FEEDBACK_ACTIVE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_MODE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_SIZE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_START: GLenum;\n    readonly TRANSFORM_FEEDBACK_PAUSED: GLenum;\n    readonly TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN: GLenum;\n    readonly TRANSFORM_FEEDBACK_VARYINGS: GLenum;\n    readonly UNIFORM_ARRAY_STRIDE: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORMS: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES: GLenum;\n    readonly UNIFORM_BLOCK_BINDING: GLenum;\n    readonly UNIFORM_BLOCK_DATA_SIZE: GLenum;\n    readonly UNIFORM_BLOCK_INDEX: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER: GLenum;\n    readonly UNIFORM_BUFFER: GLenum;\n    readonly UNIFORM_BUFFER_BINDING: GLenum;\n    readonly UNIFORM_BUFFER_OFFSET_ALIGNMENT: GLenum;\n    readonly UNIFORM_BUFFER_SIZE: GLenum;\n    readonly UNIFORM_BUFFER_START: GLenum;\n    readonly UNIFORM_IS_ROW_MAJOR: GLenum;\n    readonly UNIFORM_MATRIX_STRIDE: GLenum;\n    readonly UNIFORM_OFFSET: GLenum;\n    readonly UNIFORM_SIZE: GLenum;\n    readonly UNIFORM_TYPE: GLenum;\n    readonly UNPACK_IMAGE_HEIGHT: GLenum;\n    readonly UNPACK_ROW_LENGTH: GLenum;\n    readonly UNPACK_SKIP_IMAGES: GLenum;\n    readonly UNPACK_SKIP_PIXELS: GLenum;\n    readonly UNPACK_SKIP_ROWS: GLenum;\n    readonly UNSIGNALED: GLenum;\n    readonly UNSIGNED_INT_10F_11F_11F_REV: GLenum;\n    readonly UNSIGNED_INT_24_8: GLenum;\n    readonly UNSIGNED_INT_2_10_10_10_REV: GLenum;\n    readonly UNSIGNED_INT_5_9_9_9_REV: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_3D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_CUBE: GLenum;\n    readonly UNSIGNED_INT_VEC2: GLenum;\n    readonly UNSIGNED_INT_VEC3: GLenum;\n    readonly UNSIGNED_INT_VEC4: GLenum;\n    readonly UNSIGNED_NORMALIZED: GLenum;\n    readonly VERTEX_ARRAY_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_INTEGER: GLenum;\n    readonly WAIT_FAILED: GLenum;\n};\n\ninterface WebGL2RenderingContextBase {\n    beginQuery(target: GLenum, query: WebGLQuery): void;\n    beginTransformFeedback(primitiveMode: GLenum): void;\n    bindBufferBase(target: GLenum, index: GLuint, buffer: WebGLBuffer | null): void;\n    bindBufferRange(target: GLenum, index: GLuint, buffer: WebGLBuffer | null, offset: GLintptr, size: GLsizeiptr): void;\n    bindSampler(unit: GLuint, sampler: WebGLSampler | null): void;\n    bindTransformFeedback(target: GLenum, tf: WebGLTransformFeedback | null): void;\n    bindVertexArray(array: WebGLVertexArrayObject | null): void;\n    blitFramebuffer(srcX0: GLint, srcY0: GLint, srcX1: GLint, srcY1: GLint, dstX0: GLint, dstY0: GLint, dstX1: GLint, dstY1: GLint, mask: GLbitfield, filter: GLenum): void;\n    clearBufferfi(buffer: GLenum, drawbuffer: GLint, depth: GLfloat, stencil: GLint): void;\n    clearBufferfv(buffer: GLenum, drawbuffer: GLint, values: Float32List, srcOffset?: GLuint): void;\n    clearBufferiv(buffer: GLenum, drawbuffer: GLint, values: Int32List, srcOffset?: GLuint): void;\n    clearBufferuiv(buffer: GLenum, drawbuffer: GLint, values: Uint32List, srcOffset?: GLuint): void;\n    clientWaitSync(sync: WebGLSync, flags: GLbitfield, timeout: GLuint64): GLenum;\n    compressedTexImage3D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexImage3D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    compressedTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    copyBufferSubData(readTarget: GLenum, writeTarget: GLenum, readOffset: GLintptr, writeOffset: GLintptr, size: GLsizeiptr): void;\n    copyTexSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    createQuery(): WebGLQuery | null;\n    createSampler(): WebGLSampler | null;\n    createTransformFeedback(): WebGLTransformFeedback | null;\n    createVertexArray(): WebGLVertexArrayObject | null;\n    deleteQuery(query: WebGLQuery | null): void;\n    deleteSampler(sampler: WebGLSampler | null): void;\n    deleteSync(sync: WebGLSync | null): void;\n    deleteTransformFeedback(tf: WebGLTransformFeedback | null): void;\n    deleteVertexArray(vertexArray: WebGLVertexArrayObject | null): void;\n    drawArraysInstanced(mode: GLenum, first: GLint, count: GLsizei, instanceCount: GLsizei): void;\n    drawBuffers(buffers: GLenum[]): void;\n    drawElementsInstanced(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr, instanceCount: GLsizei): void;\n    drawRangeElements(mode: GLenum, start: GLuint, end: GLuint, count: GLsizei, type: GLenum, offset: GLintptr): void;\n    endQuery(target: GLenum): void;\n    endTransformFeedback(): void;\n    fenceSync(condition: GLenum, flags: GLbitfield): WebGLSync | null;\n    framebufferTextureLayer(target: GLenum, attachment: GLenum, texture: WebGLTexture | null, level: GLint, layer: GLint): void;\n    getActiveUniformBlockName(program: WebGLProgram, uniformBlockIndex: GLuint): string | null;\n    getActiveUniformBlockParameter(program: WebGLProgram, uniformBlockIndex: GLuint, pname: GLenum): any;\n    getActiveUniforms(program: WebGLProgram, uniformIndices: GLuint[], pname: GLenum): any;\n    getBufferSubData(target: GLenum, srcByteOffset: GLintptr, dstBuffer: ArrayBufferView, dstOffset?: GLuint, length?: GLuint): void;\n    getFragDataLocation(program: WebGLProgram, name: string): GLint;\n    getIndexedParameter(target: GLenum, index: GLuint): any;\n    getInternalformatParameter(target: GLenum, internalformat: GLenum, pname: GLenum): any;\n    getQuery(target: GLenum, pname: GLenum): WebGLQuery | null;\n    getQueryParameter(query: WebGLQuery, pname: GLenum): any;\n    getSamplerParameter(sampler: WebGLSampler, pname: GLenum): any;\n    getSyncParameter(sync: WebGLSync, pname: GLenum): any;\n    getTransformFeedbackVarying(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getUniformBlockIndex(program: WebGLProgram, uniformBlockName: string): GLuint;\n    getUniformIndices(program: WebGLProgram, uniformNames: string[]): GLuint[] | null;\n    invalidateFramebuffer(target: GLenum, attachments: GLenum[]): void;\n    invalidateSubFramebuffer(target: GLenum, attachments: GLenum[], x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    isQuery(query: WebGLQuery | null): GLboolean;\n    isSampler(sampler: WebGLSampler | null): GLboolean;\n    isSync(sync: WebGLSync | null): GLboolean;\n    isTransformFeedback(tf: WebGLTransformFeedback | null): GLboolean;\n    isVertexArray(vertexArray: WebGLVertexArrayObject | null): GLboolean;\n    pauseTransformFeedback(): void;\n    readBuffer(src: GLenum): void;\n    renderbufferStorageMultisample(target: GLenum, samples: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    resumeTransformFeedback(): void;\n    samplerParameterf(sampler: WebGLSampler, pname: GLenum, param: GLfloat): void;\n    samplerParameteri(sampler: WebGLSampler, pname: GLenum, param: GLint): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView | null): void;\n    texImage3D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    texStorage2D(target: GLenum, levels: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    texStorage3D(target: GLenum, levels: GLsizei, internalformat: GLenum, width: GLsizei, height: GLsizei, depth: GLsizei): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage3D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, zoffset: GLint, width: GLsizei, height: GLsizei, depth: GLsizei, format: GLenum, type: GLenum, srcData: ArrayBufferView | null, srcOffset?: GLuint): void;\n    transformFeedbackVaryings(program: WebGLProgram, varyings: string[], bufferMode: GLenum): void;\n    uniform1ui(location: WebGLUniformLocation | null, v0: GLuint): void;\n    uniform1uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint): void;\n    uniform2uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint, v2: GLuint): void;\n    uniform3uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4ui(location: WebGLUniformLocation | null, v0: GLuint, v1: GLuint, v2: GLuint, v3: GLuint): void;\n    uniform4uiv(location: WebGLUniformLocation | null, data: Uint32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformBlockBinding(program: WebGLProgram, uniformBlockIndex: GLuint, uniformBlockBinding: GLuint): void;\n    uniformMatrix2x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    vertexAttribDivisor(index: GLuint, divisor: GLuint): void;\n    vertexAttribI4i(index: GLuint, x: GLint, y: GLint, z: GLint, w: GLint): void;\n    vertexAttribI4iv(index: GLuint, values: Int32List): void;\n    vertexAttribI4ui(index: GLuint, x: GLuint, y: GLuint, z: GLuint, w: GLuint): void;\n    vertexAttribI4uiv(index: GLuint, values: Uint32List): void;\n    vertexAttribIPointer(index: GLuint, size: GLint, type: GLenum, stride: GLsizei, offset: GLintptr): void;\n    waitSync(sync: WebGLSync, flags: GLbitfield, timeout: GLint64): void;\n    readonly ACTIVE_UNIFORM_BLOCKS: GLenum;\n    readonly ALREADY_SIGNALED: GLenum;\n    readonly ANY_SAMPLES_PASSED: GLenum;\n    readonly ANY_SAMPLES_PASSED_CONSERVATIVE: GLenum;\n    readonly COLOR: GLenum;\n    readonly COLOR_ATTACHMENT1: GLenum;\n    readonly COLOR_ATTACHMENT10: GLenum;\n    readonly COLOR_ATTACHMENT11: GLenum;\n    readonly COLOR_ATTACHMENT12: GLenum;\n    readonly COLOR_ATTACHMENT13: GLenum;\n    readonly COLOR_ATTACHMENT14: GLenum;\n    readonly COLOR_ATTACHMENT15: GLenum;\n    readonly COLOR_ATTACHMENT2: GLenum;\n    readonly COLOR_ATTACHMENT3: GLenum;\n    readonly COLOR_ATTACHMENT4: GLenum;\n    readonly COLOR_ATTACHMENT5: GLenum;\n    readonly COLOR_ATTACHMENT6: GLenum;\n    readonly COLOR_ATTACHMENT7: GLenum;\n    readonly COLOR_ATTACHMENT8: GLenum;\n    readonly COLOR_ATTACHMENT9: GLenum;\n    readonly COMPARE_REF_TO_TEXTURE: GLenum;\n    readonly CONDITION_SATISFIED: GLenum;\n    readonly COPY_READ_BUFFER: GLenum;\n    readonly COPY_READ_BUFFER_BINDING: GLenum;\n    readonly COPY_WRITE_BUFFER: GLenum;\n    readonly COPY_WRITE_BUFFER_BINDING: GLenum;\n    readonly CURRENT_QUERY: GLenum;\n    readonly DEPTH: GLenum;\n    readonly DEPTH24_STENCIL8: GLenum;\n    readonly DEPTH32F_STENCIL8: GLenum;\n    readonly DEPTH_COMPONENT24: GLenum;\n    readonly DEPTH_COMPONENT32F: GLenum;\n    readonly DRAW_BUFFER0: GLenum;\n    readonly DRAW_BUFFER1: GLenum;\n    readonly DRAW_BUFFER10: GLenum;\n    readonly DRAW_BUFFER11: GLenum;\n    readonly DRAW_BUFFER12: GLenum;\n    readonly DRAW_BUFFER13: GLenum;\n    readonly DRAW_BUFFER14: GLenum;\n    readonly DRAW_BUFFER15: GLenum;\n    readonly DRAW_BUFFER2: GLenum;\n    readonly DRAW_BUFFER3: GLenum;\n    readonly DRAW_BUFFER4: GLenum;\n    readonly DRAW_BUFFER5: GLenum;\n    readonly DRAW_BUFFER6: GLenum;\n    readonly DRAW_BUFFER7: GLenum;\n    readonly DRAW_BUFFER8: GLenum;\n    readonly DRAW_BUFFER9: GLenum;\n    readonly DRAW_FRAMEBUFFER: GLenum;\n    readonly DRAW_FRAMEBUFFER_BINDING: GLenum;\n    readonly DYNAMIC_COPY: GLenum;\n    readonly DYNAMIC_READ: GLenum;\n    readonly FLOAT_32_UNSIGNED_INT_24_8_REV: GLenum;\n    readonly FLOAT_MAT2x3: GLenum;\n    readonly FLOAT_MAT2x4: GLenum;\n    readonly FLOAT_MAT3x2: GLenum;\n    readonly FLOAT_MAT3x4: GLenum;\n    readonly FLOAT_MAT4x2: GLenum;\n    readonly FLOAT_MAT4x3: GLenum;\n    readonly FRAGMENT_SHADER_DERIVATIVE_HINT: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_BLUE_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_GREEN_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_RED_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER: GLenum;\n    readonly FRAMEBUFFER_DEFAULT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: GLenum;\n    readonly HALF_FLOAT: GLenum;\n    readonly INTERLEAVED_ATTRIBS: GLenum;\n    readonly INT_2_10_10_10_REV: GLenum;\n    readonly INT_SAMPLER_2D: GLenum;\n    readonly INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly INT_SAMPLER_3D: GLenum;\n    readonly INT_SAMPLER_CUBE: GLenum;\n    readonly INVALID_INDEX: GLenum;\n    readonly MAX: GLenum;\n    readonly MAX_3D_TEXTURE_SIZE: GLenum;\n    readonly MAX_ARRAY_TEXTURE_LAYERS: GLenum;\n    readonly MAX_CLIENT_WAIT_TIMEOUT_WEBGL: GLenum;\n    readonly MAX_COLOR_ATTACHMENTS: GLenum;\n    readonly MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_COMBINED_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_DRAW_BUFFERS: GLenum;\n    readonly MAX_ELEMENTS_INDICES: GLenum;\n    readonly MAX_ELEMENTS_VERTICES: GLenum;\n    readonly MAX_ELEMENT_INDEX: GLenum;\n    readonly MAX_FRAGMENT_INPUT_COMPONENTS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_COMPONENTS: GLenum;\n    readonly MAX_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly MAX_SAMPLES: GLenum;\n    readonly MAX_SERVER_WAIT_TIMEOUT: GLenum;\n    readonly MAX_TEXTURE_LOD_BIAS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS: GLenum;\n    readonly MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS: GLenum;\n    readonly MAX_UNIFORM_BLOCK_SIZE: GLenum;\n    readonly MAX_UNIFORM_BUFFER_BINDINGS: GLenum;\n    readonly MAX_VARYING_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_OUTPUT_COMPONENTS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_BLOCKS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_COMPONENTS: GLenum;\n    readonly MIN: GLenum;\n    readonly MIN_PROGRAM_TEXEL_OFFSET: GLenum;\n    readonly OBJECT_TYPE: GLenum;\n    readonly PACK_ROW_LENGTH: GLenum;\n    readonly PACK_SKIP_PIXELS: GLenum;\n    readonly PACK_SKIP_ROWS: GLenum;\n    readonly PIXEL_PACK_BUFFER: GLenum;\n    readonly PIXEL_PACK_BUFFER_BINDING: GLenum;\n    readonly PIXEL_UNPACK_BUFFER: GLenum;\n    readonly PIXEL_UNPACK_BUFFER_BINDING: GLenum;\n    readonly QUERY_RESULT: GLenum;\n    readonly QUERY_RESULT_AVAILABLE: GLenum;\n    readonly R11F_G11F_B10F: GLenum;\n    readonly R16F: GLenum;\n    readonly R16I: GLenum;\n    readonly R16UI: GLenum;\n    readonly R32F: GLenum;\n    readonly R32I: GLenum;\n    readonly R32UI: GLenum;\n    readonly R8: GLenum;\n    readonly R8I: GLenum;\n    readonly R8UI: GLenum;\n    readonly R8_SNORM: GLenum;\n    readonly RASTERIZER_DISCARD: GLenum;\n    readonly READ_BUFFER: GLenum;\n    readonly READ_FRAMEBUFFER: GLenum;\n    readonly READ_FRAMEBUFFER_BINDING: GLenum;\n    readonly RED: GLenum;\n    readonly RED_INTEGER: GLenum;\n    readonly RENDERBUFFER_SAMPLES: GLenum;\n    readonly RG: GLenum;\n    readonly RG16F: GLenum;\n    readonly RG16I: GLenum;\n    readonly RG16UI: GLenum;\n    readonly RG32F: GLenum;\n    readonly RG32I: GLenum;\n    readonly RG32UI: GLenum;\n    readonly RG8: GLenum;\n    readonly RG8I: GLenum;\n    readonly RG8UI: GLenum;\n    readonly RG8_SNORM: GLenum;\n    readonly RGB10_A2: GLenum;\n    readonly RGB10_A2UI: GLenum;\n    readonly RGB16F: GLenum;\n    readonly RGB16I: GLenum;\n    readonly RGB16UI: GLenum;\n    readonly RGB32F: GLenum;\n    readonly RGB32I: GLenum;\n    readonly RGB32UI: GLenum;\n    readonly RGB8: GLenum;\n    readonly RGB8I: GLenum;\n    readonly RGB8UI: GLenum;\n    readonly RGB8_SNORM: GLenum;\n    readonly RGB9_E5: GLenum;\n    readonly RGBA16F: GLenum;\n    readonly RGBA16I: GLenum;\n    readonly RGBA16UI: GLenum;\n    readonly RGBA32F: GLenum;\n    readonly RGBA32I: GLenum;\n    readonly RGBA32UI: GLenum;\n    readonly RGBA8: GLenum;\n    readonly RGBA8I: GLenum;\n    readonly RGBA8UI: GLenum;\n    readonly RGBA8_SNORM: GLenum;\n    readonly RGBA_INTEGER: GLenum;\n    readonly RGB_INTEGER: GLenum;\n    readonly RG_INTEGER: GLenum;\n    readonly SAMPLER_2D_ARRAY: GLenum;\n    readonly SAMPLER_2D_ARRAY_SHADOW: GLenum;\n    readonly SAMPLER_2D_SHADOW: GLenum;\n    readonly SAMPLER_3D: GLenum;\n    readonly SAMPLER_BINDING: GLenum;\n    readonly SAMPLER_CUBE_SHADOW: GLenum;\n    readonly SEPARATE_ATTRIBS: GLenum;\n    readonly SIGNALED: GLenum;\n    readonly SIGNED_NORMALIZED: GLenum;\n    readonly SRGB: GLenum;\n    readonly SRGB8: GLenum;\n    readonly SRGB8_ALPHA8: GLenum;\n    readonly STATIC_COPY: GLenum;\n    readonly STATIC_READ: GLenum;\n    readonly STENCIL: GLenum;\n    readonly STREAM_COPY: GLenum;\n    readonly STREAM_READ: GLenum;\n    readonly SYNC_CONDITION: GLenum;\n    readonly SYNC_FENCE: GLenum;\n    readonly SYNC_FLAGS: GLenum;\n    readonly SYNC_FLUSH_COMMANDS_BIT: GLenum;\n    readonly SYNC_GPU_COMMANDS_COMPLETE: GLenum;\n    readonly SYNC_STATUS: GLenum;\n    readonly TEXTURE_2D_ARRAY: GLenum;\n    readonly TEXTURE_3D: GLenum;\n    readonly TEXTURE_BASE_LEVEL: GLenum;\n    readonly TEXTURE_BINDING_2D_ARRAY: GLenum;\n    readonly TEXTURE_BINDING_3D: GLenum;\n    readonly TEXTURE_COMPARE_FUNC: GLenum;\n    readonly TEXTURE_COMPARE_MODE: GLenum;\n    readonly TEXTURE_IMMUTABLE_FORMAT: GLenum;\n    readonly TEXTURE_IMMUTABLE_LEVELS: GLenum;\n    readonly TEXTURE_MAX_LEVEL: GLenum;\n    readonly TEXTURE_MAX_LOD: GLenum;\n    readonly TEXTURE_MIN_LOD: GLenum;\n    readonly TEXTURE_WRAP_R: GLenum;\n    readonly TIMEOUT_EXPIRED: GLenum;\n    readonly TIMEOUT_IGNORED: GLint64;\n    readonly TRANSFORM_FEEDBACK: GLenum;\n    readonly TRANSFORM_FEEDBACK_ACTIVE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_BINDING: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_MODE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_SIZE: GLenum;\n    readonly TRANSFORM_FEEDBACK_BUFFER_START: GLenum;\n    readonly TRANSFORM_FEEDBACK_PAUSED: GLenum;\n    readonly TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN: GLenum;\n    readonly TRANSFORM_FEEDBACK_VARYINGS: GLenum;\n    readonly UNIFORM_ARRAY_STRIDE: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORMS: GLenum;\n    readonly UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES: GLenum;\n    readonly UNIFORM_BLOCK_BINDING: GLenum;\n    readonly UNIFORM_BLOCK_DATA_SIZE: GLenum;\n    readonly UNIFORM_BLOCK_INDEX: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER: GLenum;\n    readonly UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER: GLenum;\n    readonly UNIFORM_BUFFER: GLenum;\n    readonly UNIFORM_BUFFER_BINDING: GLenum;\n    readonly UNIFORM_BUFFER_OFFSET_ALIGNMENT: GLenum;\n    readonly UNIFORM_BUFFER_SIZE: GLenum;\n    readonly UNIFORM_BUFFER_START: GLenum;\n    readonly UNIFORM_IS_ROW_MAJOR: GLenum;\n    readonly UNIFORM_MATRIX_STRIDE: GLenum;\n    readonly UNIFORM_OFFSET: GLenum;\n    readonly UNIFORM_SIZE: GLenum;\n    readonly UNIFORM_TYPE: GLenum;\n    readonly UNPACK_IMAGE_HEIGHT: GLenum;\n    readonly UNPACK_ROW_LENGTH: GLenum;\n    readonly UNPACK_SKIP_IMAGES: GLenum;\n    readonly UNPACK_SKIP_PIXELS: GLenum;\n    readonly UNPACK_SKIP_ROWS: GLenum;\n    readonly UNSIGNALED: GLenum;\n    readonly UNSIGNED_INT_10F_11F_11F_REV: GLenum;\n    readonly UNSIGNED_INT_24_8: GLenum;\n    readonly UNSIGNED_INT_2_10_10_10_REV: GLenum;\n    readonly UNSIGNED_INT_5_9_9_9_REV: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_2D_ARRAY: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_3D: GLenum;\n    readonly UNSIGNED_INT_SAMPLER_CUBE: GLenum;\n    readonly UNSIGNED_INT_VEC2: GLenum;\n    readonly UNSIGNED_INT_VEC3: GLenum;\n    readonly UNSIGNED_INT_VEC4: GLenum;\n    readonly UNSIGNED_NORMALIZED: GLenum;\n    readonly VERTEX_ARRAY_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_DIVISOR: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_INTEGER: GLenum;\n    readonly WAIT_FAILED: GLenum;\n}\n\ninterface WebGL2RenderingContextOverloads {\n    bufferData(target: GLenum, size: GLsizeiptr, usage: GLenum): void;\n    bufferData(target: GLenum, srcData: BufferSource | null, usage: GLenum): void;\n    bufferData(target: GLenum, srcData: ArrayBufferView, usage: GLenum, srcOffset: GLuint, length?: GLuint): void;\n    bufferSubData(target: GLenum, dstByteOffset: GLintptr, srcData: BufferSource): void;\n    bufferSubData(target: GLenum, dstByteOffset: GLintptr, srcData: ArrayBufferView, srcOffset: GLuint, length?: GLuint): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, imageSize: GLsizei, offset: GLintptr): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, srcData: ArrayBufferView, srcOffset?: GLuint, srcLengthOverride?: GLuint): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, dstData: ArrayBufferView | null): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, offset: GLintptr): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, dstData: ArrayBufferView, dstOffset: GLuint): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pboOffset: GLintptr): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, srcData: ArrayBufferView, srcOffset: GLuint): void;\n    uniform1fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform1iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4fv(location: WebGLUniformLocation | null, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4iv(location: WebGLUniformLocation | null, data: Int32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Float32List, srcOffset?: GLuint, srcLength?: GLuint): void;\n}\n\n/** Part of the WebGL API and represents the information returned by calling the WebGLRenderingContext.getActiveAttrib() and WebGLRenderingContext.getActiveUniform() methods. */\ninterface WebGLActiveInfo {\n    readonly name: string;\n    readonly size: GLint;\n    readonly type: GLenum;\n}\n\ndeclare var WebGLActiveInfo: {\n    prototype: WebGLActiveInfo;\n    new(): WebGLActiveInfo;\n};\n\n/** Part of the WebGL API and represents an opaque buffer object storing data such as vertices or colors. */\ninterface WebGLBuffer extends WebGLObject {\n}\n\ndeclare var WebGLBuffer: {\n    prototype: WebGLBuffer;\n    new(): WebGLBuffer;\n};\n\n/** The WebContextEvent interface is part of the WebGL API and is an interface for an event that is generated in response to a status change to the WebGL rendering context. */\ninterface WebGLContextEvent extends Event {\n    readonly statusMessage: string;\n}\n\ndeclare var WebGLContextEvent: {\n    prototype: WebGLContextEvent;\n    new(type: string, eventInit?: WebGLContextEventInit): WebGLContextEvent;\n};\n\n/** Part of the WebGL API and represents a collection of buffers that serve as a rendering destination. */\ninterface WebGLFramebuffer extends WebGLObject {\n}\n\ndeclare var WebGLFramebuffer: {\n    prototype: WebGLFramebuffer;\n    new(): WebGLFramebuffer;\n};\n\ninterface WebGLObject {\n}\n\ndeclare var WebGLObject: {\n    prototype: WebGLObject;\n    new(): WebGLObject;\n};\n\n/** The WebGLProgram is part of the WebGL API and is a combination of two compiled WebGLShaders consisting of a vertex shader and a fragment shader (both written in GLSL). */\ninterface WebGLProgram extends WebGLObject {\n}\n\ndeclare var WebGLProgram: {\n    prototype: WebGLProgram;\n    new(): WebGLProgram;\n};\n\ninterface WebGLQuery extends WebGLObject {\n}\n\ndeclare var WebGLQuery: {\n    prototype: WebGLQuery;\n    new(): WebGLQuery;\n};\n\n/** Part of the WebGL API and represents a buffer that can contain an image, or can be source or target of an rendering operation. */\ninterface WebGLRenderbuffer extends WebGLObject {\n}\n\ndeclare var WebGLRenderbuffer: {\n    prototype: WebGLRenderbuffer;\n    new(): WebGLRenderbuffer;\n};\n\n/** Provides an interface to the OpenGL ES 2.0 graphics rendering context for the drawing surface of an HTML <canvas> element. */\ninterface WebGLRenderingContext extends WebGLRenderingContextBase, WebGLRenderingContextOverloads {\n}\n\ndeclare var WebGLRenderingContext: {\n    prototype: WebGLRenderingContext;\n    new(): WebGLRenderingContext;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n};\n\ninterface WebGLRenderingContextBase {\n    readonly canvas: OffscreenCanvas;\n    readonly drawingBufferHeight: GLsizei;\n    readonly drawingBufferWidth: GLsizei;\n    activeTexture(texture: GLenum): void;\n    attachShader(program: WebGLProgram, shader: WebGLShader): void;\n    bindAttribLocation(program: WebGLProgram, index: GLuint, name: string): void;\n    bindBuffer(target: GLenum, buffer: WebGLBuffer | null): void;\n    bindFramebuffer(target: GLenum, framebuffer: WebGLFramebuffer | null): void;\n    bindRenderbuffer(target: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    bindTexture(target: GLenum, texture: WebGLTexture | null): void;\n    blendColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    blendEquation(mode: GLenum): void;\n    blendEquationSeparate(modeRGB: GLenum, modeAlpha: GLenum): void;\n    blendFunc(sfactor: GLenum, dfactor: GLenum): void;\n    blendFuncSeparate(srcRGB: GLenum, dstRGB: GLenum, srcAlpha: GLenum, dstAlpha: GLenum): void;\n    checkFramebufferStatus(target: GLenum): GLenum;\n    clear(mask: GLbitfield): void;\n    clearColor(red: GLclampf, green: GLclampf, blue: GLclampf, alpha: GLclampf): void;\n    clearDepth(depth: GLclampf): void;\n    clearStencil(s: GLint): void;\n    colorMask(red: GLboolean, green: GLboolean, blue: GLboolean, alpha: GLboolean): void;\n    compileShader(shader: WebGLShader): void;\n    copyTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, x: GLint, y: GLint, width: GLsizei, height: GLsizei, border: GLint): void;\n    copyTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    createBuffer(): WebGLBuffer | null;\n    createFramebuffer(): WebGLFramebuffer | null;\n    createProgram(): WebGLProgram | null;\n    createRenderbuffer(): WebGLRenderbuffer | null;\n    createShader(type: GLenum): WebGLShader | null;\n    createTexture(): WebGLTexture | null;\n    cullFace(mode: GLenum): void;\n    deleteBuffer(buffer: WebGLBuffer | null): void;\n    deleteFramebuffer(framebuffer: WebGLFramebuffer | null): void;\n    deleteProgram(program: WebGLProgram | null): void;\n    deleteRenderbuffer(renderbuffer: WebGLRenderbuffer | null): void;\n    deleteShader(shader: WebGLShader | null): void;\n    deleteTexture(texture: WebGLTexture | null): void;\n    depthFunc(func: GLenum): void;\n    depthMask(flag: GLboolean): void;\n    depthRange(zNear: GLclampf, zFar: GLclampf): void;\n    detachShader(program: WebGLProgram, shader: WebGLShader): void;\n    disable(cap: GLenum): void;\n    disableVertexAttribArray(index: GLuint): void;\n    drawArrays(mode: GLenum, first: GLint, count: GLsizei): void;\n    drawElements(mode: GLenum, count: GLsizei, type: GLenum, offset: GLintptr): void;\n    enable(cap: GLenum): void;\n    enableVertexAttribArray(index: GLuint): void;\n    finish(): void;\n    flush(): void;\n    framebufferRenderbuffer(target: GLenum, attachment: GLenum, renderbuffertarget: GLenum, renderbuffer: WebGLRenderbuffer | null): void;\n    framebufferTexture2D(target: GLenum, attachment: GLenum, textarget: GLenum, texture: WebGLTexture | null, level: GLint): void;\n    frontFace(mode: GLenum): void;\n    generateMipmap(target: GLenum): void;\n    getActiveAttrib(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getActiveUniform(program: WebGLProgram, index: GLuint): WebGLActiveInfo | null;\n    getAttachedShaders(program: WebGLProgram): WebGLShader[] | null;\n    getAttribLocation(program: WebGLProgram, name: string): GLint;\n    getBufferParameter(target: GLenum, pname: GLenum): any;\n    getContextAttributes(): WebGLContextAttributes | null;\n    getError(): GLenum;\n    getExtension(extensionName: "EXT_blend_minmax"): EXT_blend_minmax | null;\n    getExtension(extensionName: "EXT_texture_filter_anisotropic"): EXT_texture_filter_anisotropic | null;\n    getExtension(extensionName: "EXT_frag_depth"): EXT_frag_depth | null;\n    getExtension(extensionName: "EXT_shader_texture_lod"): EXT_shader_texture_lod | null;\n    getExtension(extensionName: "EXT_sRGB"): EXT_sRGB | null;\n    getExtension(extensionName: "OES_vertex_array_object"): OES_vertex_array_object | null;\n    getExtension(extensionName: "WEBGL_color_buffer_float"): WEBGL_color_buffer_float | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_astc"): WEBGL_compressed_texture_astc | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc_srgb"): WEBGL_compressed_texture_s3tc_srgb | null;\n    getExtension(extensionName: "WEBGL_debug_shaders"): WEBGL_debug_shaders | null;\n    getExtension(extensionName: "WEBGL_draw_buffers"): WEBGL_draw_buffers | null;\n    getExtension(extensionName: "WEBGL_lose_context"): WEBGL_lose_context | null;\n    getExtension(extensionName: "WEBGL_depth_texture"): WEBGL_depth_texture | null;\n    getExtension(extensionName: "WEBGL_debug_renderer_info"): WEBGL_debug_renderer_info | null;\n    getExtension(extensionName: "WEBGL_compressed_texture_s3tc"): WEBGL_compressed_texture_s3tc | null;\n    getExtension(extensionName: "OES_texture_half_float_linear"): OES_texture_half_float_linear | null;\n    getExtension(extensionName: "OES_texture_half_float"): OES_texture_half_float | null;\n    getExtension(extensionName: "OES_texture_float_linear"): OES_texture_float_linear | null;\n    getExtension(extensionName: "OES_texture_float"): OES_texture_float | null;\n    getExtension(extensionName: "OES_standard_derivatives"): OES_standard_derivatives | null;\n    getExtension(extensionName: "OES_element_index_uint"): OES_element_index_uint | null;\n    getExtension(extensionName: "ANGLE_instanced_arrays"): ANGLE_instanced_arrays | null;\n    getExtension(extensionName: string): any;\n    getFramebufferAttachmentParameter(target: GLenum, attachment: GLenum, pname: GLenum): any;\n    getParameter(pname: GLenum): any;\n    getProgramInfoLog(program: WebGLProgram): string | null;\n    getProgramParameter(program: WebGLProgram, pname: GLenum): any;\n    getRenderbufferParameter(target: GLenum, pname: GLenum): any;\n    getShaderInfoLog(shader: WebGLShader): string | null;\n    getShaderParameter(shader: WebGLShader, pname: GLenum): any;\n    getShaderPrecisionFormat(shadertype: GLenum, precisiontype: GLenum): WebGLShaderPrecisionFormat | null;\n    getShaderSource(shader: WebGLShader): string | null;\n    getSupportedExtensions(): string[] | null;\n    getTexParameter(target: GLenum, pname: GLenum): any;\n    getUniform(program: WebGLProgram, location: WebGLUniformLocation): any;\n    getUniformLocation(program: WebGLProgram, name: string): WebGLUniformLocation | null;\n    getVertexAttrib(index: GLuint, pname: GLenum): any;\n    getVertexAttribOffset(index: GLuint, pname: GLenum): GLintptr;\n    hint(target: GLenum, mode: GLenum): void;\n    isBuffer(buffer: WebGLBuffer | null): GLboolean;\n    isContextLost(): boolean;\n    isEnabled(cap: GLenum): GLboolean;\n    isFramebuffer(framebuffer: WebGLFramebuffer | null): GLboolean;\n    isProgram(program: WebGLProgram | null): GLboolean;\n    isRenderbuffer(renderbuffer: WebGLRenderbuffer | null): GLboolean;\n    isShader(shader: WebGLShader | null): GLboolean;\n    isTexture(texture: WebGLTexture | null): GLboolean;\n    lineWidth(width: GLfloat): void;\n    linkProgram(program: WebGLProgram): void;\n    pixelStorei(pname: GLenum, param: GLint | GLboolean): void;\n    polygonOffset(factor: GLfloat, units: GLfloat): void;\n    renderbufferStorage(target: GLenum, internalformat: GLenum, width: GLsizei, height: GLsizei): void;\n    sampleCoverage(value: GLclampf, invert: GLboolean): void;\n    scissor(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    shaderSource(shader: WebGLShader, source: string): void;\n    stencilFunc(func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilFuncSeparate(face: GLenum, func: GLenum, ref: GLint, mask: GLuint): void;\n    stencilMask(mask: GLuint): void;\n    stencilMaskSeparate(face: GLenum, mask: GLuint): void;\n    stencilOp(fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    stencilOpSeparate(face: GLenum, fail: GLenum, zfail: GLenum, zpass: GLenum): void;\n    texParameterf(target: GLenum, pname: GLenum, param: GLfloat): void;\n    texParameteri(target: GLenum, pname: GLenum, param: GLint): void;\n    uniform1f(location: WebGLUniformLocation | null, x: GLfloat): void;\n    uniform1i(location: WebGLUniformLocation | null, x: GLint): void;\n    uniform2f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat): void;\n    uniform2i(location: WebGLUniformLocation | null, x: GLint, y: GLint): void;\n    uniform3f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    uniform3i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint): void;\n    uniform4f(location: WebGLUniformLocation | null, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    uniform4i(location: WebGLUniformLocation | null, x: GLint, y: GLint, z: GLint, w: GLint): void;\n    useProgram(program: WebGLProgram | null): void;\n    validateProgram(program: WebGLProgram): void;\n    vertexAttrib1f(index: GLuint, x: GLfloat): void;\n    vertexAttrib1fv(index: GLuint, values: Float32List): void;\n    vertexAttrib2f(index: GLuint, x: GLfloat, y: GLfloat): void;\n    vertexAttrib2fv(index: GLuint, values: Float32List): void;\n    vertexAttrib3f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat): void;\n    vertexAttrib3fv(index: GLuint, values: Float32List): void;\n    vertexAttrib4f(index: GLuint, x: GLfloat, y: GLfloat, z: GLfloat, w: GLfloat): void;\n    vertexAttrib4fv(index: GLuint, values: Float32List): void;\n    vertexAttribPointer(index: GLuint, size: GLint, type: GLenum, normalized: GLboolean, stride: GLsizei, offset: GLintptr): void;\n    viewport(x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    readonly ACTIVE_ATTRIBUTES: GLenum;\n    readonly ACTIVE_TEXTURE: GLenum;\n    readonly ACTIVE_UNIFORMS: GLenum;\n    readonly ALIASED_LINE_WIDTH_RANGE: GLenum;\n    readonly ALIASED_POINT_SIZE_RANGE: GLenum;\n    readonly ALPHA: GLenum;\n    readonly ALPHA_BITS: GLenum;\n    readonly ALWAYS: GLenum;\n    readonly ARRAY_BUFFER: GLenum;\n    readonly ARRAY_BUFFER_BINDING: GLenum;\n    readonly ATTACHED_SHADERS: GLenum;\n    readonly BACK: GLenum;\n    readonly BLEND: GLenum;\n    readonly BLEND_COLOR: GLenum;\n    readonly BLEND_DST_ALPHA: GLenum;\n    readonly BLEND_DST_RGB: GLenum;\n    readonly BLEND_EQUATION: GLenum;\n    readonly BLEND_EQUATION_ALPHA: GLenum;\n    readonly BLEND_EQUATION_RGB: GLenum;\n    readonly BLEND_SRC_ALPHA: GLenum;\n    readonly BLEND_SRC_RGB: GLenum;\n    readonly BLUE_BITS: GLenum;\n    readonly BOOL: GLenum;\n    readonly BOOL_VEC2: GLenum;\n    readonly BOOL_VEC3: GLenum;\n    readonly BOOL_VEC4: GLenum;\n    readonly BROWSER_DEFAULT_WEBGL: GLenum;\n    readonly BUFFER_SIZE: GLenum;\n    readonly BUFFER_USAGE: GLenum;\n    readonly BYTE: GLenum;\n    readonly CCW: GLenum;\n    readonly CLAMP_TO_EDGE: GLenum;\n    readonly COLOR_ATTACHMENT0: GLenum;\n    readonly COLOR_BUFFER_BIT: GLenum;\n    readonly COLOR_CLEAR_VALUE: GLenum;\n    readonly COLOR_WRITEMASK: GLenum;\n    readonly COMPILE_STATUS: GLenum;\n    readonly COMPRESSED_TEXTURE_FORMATS: GLenum;\n    readonly CONSTANT_ALPHA: GLenum;\n    readonly CONSTANT_COLOR: GLenum;\n    readonly CONTEXT_LOST_WEBGL: GLenum;\n    readonly CULL_FACE: GLenum;\n    readonly CULL_FACE_MODE: GLenum;\n    readonly CURRENT_PROGRAM: GLenum;\n    readonly CURRENT_VERTEX_ATTRIB: GLenum;\n    readonly CW: GLenum;\n    readonly DECR: GLenum;\n    readonly DECR_WRAP: GLenum;\n    readonly DELETE_STATUS: GLenum;\n    readonly DEPTH_ATTACHMENT: GLenum;\n    readonly DEPTH_BITS: GLenum;\n    readonly DEPTH_BUFFER_BIT: GLenum;\n    readonly DEPTH_CLEAR_VALUE: GLenum;\n    readonly DEPTH_COMPONENT: GLenum;\n    readonly DEPTH_COMPONENT16: GLenum;\n    readonly DEPTH_FUNC: GLenum;\n    readonly DEPTH_RANGE: GLenum;\n    readonly DEPTH_STENCIL: GLenum;\n    readonly DEPTH_STENCIL_ATTACHMENT: GLenum;\n    readonly DEPTH_TEST: GLenum;\n    readonly DEPTH_WRITEMASK: GLenum;\n    readonly DITHER: GLenum;\n    readonly DONT_CARE: GLenum;\n    readonly DST_ALPHA: GLenum;\n    readonly DST_COLOR: GLenum;\n    readonly DYNAMIC_DRAW: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER: GLenum;\n    readonly ELEMENT_ARRAY_BUFFER_BINDING: GLenum;\n    readonly EQUAL: GLenum;\n    readonly FASTEST: GLenum;\n    readonly FLOAT: GLenum;\n    readonly FLOAT_MAT2: GLenum;\n    readonly FLOAT_MAT3: GLenum;\n    readonly FLOAT_MAT4: GLenum;\n    readonly FLOAT_VEC2: GLenum;\n    readonly FLOAT_VEC3: GLenum;\n    readonly FLOAT_VEC4: GLenum;\n    readonly FRAGMENT_SHADER: GLenum;\n    readonly FRAMEBUFFER: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_NAME: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE: GLenum;\n    readonly FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL: GLenum;\n    readonly FRAMEBUFFER_BINDING: GLenum;\n    readonly FRAMEBUFFER_COMPLETE: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_DIMENSIONS: GLenum;\n    readonly FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: GLenum;\n    readonly FRAMEBUFFER_UNSUPPORTED: GLenum;\n    readonly FRONT: GLenum;\n    readonly FRONT_AND_BACK: GLenum;\n    readonly FRONT_FACE: GLenum;\n    readonly FUNC_ADD: GLenum;\n    readonly FUNC_REVERSE_SUBTRACT: GLenum;\n    readonly FUNC_SUBTRACT: GLenum;\n    readonly GENERATE_MIPMAP_HINT: GLenum;\n    readonly GEQUAL: GLenum;\n    readonly GREATER: GLenum;\n    readonly GREEN_BITS: GLenum;\n    readonly HIGH_FLOAT: GLenum;\n    readonly HIGH_INT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_FORMAT: GLenum;\n    readonly IMPLEMENTATION_COLOR_READ_TYPE: GLenum;\n    readonly INCR: GLenum;\n    readonly INCR_WRAP: GLenum;\n    readonly INT: GLenum;\n    readonly INT_VEC2: GLenum;\n    readonly INT_VEC3: GLenum;\n    readonly INT_VEC4: GLenum;\n    readonly INVALID_ENUM: GLenum;\n    readonly INVALID_FRAMEBUFFER_OPERATION: GLenum;\n    readonly INVALID_OPERATION: GLenum;\n    readonly INVALID_VALUE: GLenum;\n    readonly INVERT: GLenum;\n    readonly KEEP: GLenum;\n    readonly LEQUAL: GLenum;\n    readonly LESS: GLenum;\n    readonly LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_LINEAR: GLenum;\n    readonly LINEAR_MIPMAP_NEAREST: GLenum;\n    readonly LINES: GLenum;\n    readonly LINE_LOOP: GLenum;\n    readonly LINE_STRIP: GLenum;\n    readonly LINE_WIDTH: GLenum;\n    readonly LINK_STATUS: GLenum;\n    readonly LOW_FLOAT: GLenum;\n    readonly LOW_INT: GLenum;\n    readonly LUMINANCE: GLenum;\n    readonly LUMINANCE_ALPHA: GLenum;\n    readonly MAX_COMBINED_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_CUBE_MAP_TEXTURE_SIZE: GLenum;\n    readonly MAX_FRAGMENT_UNIFORM_VECTORS: GLenum;\n    readonly MAX_RENDERBUFFER_SIZE: GLenum;\n    readonly MAX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_TEXTURE_SIZE: GLenum;\n    readonly MAX_VARYING_VECTORS: GLenum;\n    readonly MAX_VERTEX_ATTRIBS: GLenum;\n    readonly MAX_VERTEX_TEXTURE_IMAGE_UNITS: GLenum;\n    readonly MAX_VERTEX_UNIFORM_VECTORS: GLenum;\n    readonly MAX_VIEWPORT_DIMS: GLenum;\n    readonly MEDIUM_FLOAT: GLenum;\n    readonly MEDIUM_INT: GLenum;\n    readonly MIRRORED_REPEAT: GLenum;\n    readonly NEAREST: GLenum;\n    readonly NEAREST_MIPMAP_LINEAR: GLenum;\n    readonly NEAREST_MIPMAP_NEAREST: GLenum;\n    readonly NEVER: GLenum;\n    readonly NICEST: GLenum;\n    readonly NONE: GLenum;\n    readonly NOTEQUAL: GLenum;\n    readonly NO_ERROR: GLenum;\n    readonly ONE: GLenum;\n    readonly ONE_MINUS_CONSTANT_ALPHA: GLenum;\n    readonly ONE_MINUS_CONSTANT_COLOR: GLenum;\n    readonly ONE_MINUS_DST_ALPHA: GLenum;\n    readonly ONE_MINUS_DST_COLOR: GLenum;\n    readonly ONE_MINUS_SRC_ALPHA: GLenum;\n    readonly ONE_MINUS_SRC_COLOR: GLenum;\n    readonly OUT_OF_MEMORY: GLenum;\n    readonly PACK_ALIGNMENT: GLenum;\n    readonly POINTS: GLenum;\n    readonly POLYGON_OFFSET_FACTOR: GLenum;\n    readonly POLYGON_OFFSET_FILL: GLenum;\n    readonly POLYGON_OFFSET_UNITS: GLenum;\n    readonly RED_BITS: GLenum;\n    readonly RENDERBUFFER: GLenum;\n    readonly RENDERBUFFER_ALPHA_SIZE: GLenum;\n    readonly RENDERBUFFER_BINDING: GLenum;\n    readonly RENDERBUFFER_BLUE_SIZE: GLenum;\n    readonly RENDERBUFFER_DEPTH_SIZE: GLenum;\n    readonly RENDERBUFFER_GREEN_SIZE: GLenum;\n    readonly RENDERBUFFER_HEIGHT: GLenum;\n    readonly RENDERBUFFER_INTERNAL_FORMAT: GLenum;\n    readonly RENDERBUFFER_RED_SIZE: GLenum;\n    readonly RENDERBUFFER_STENCIL_SIZE: GLenum;\n    readonly RENDERBUFFER_WIDTH: GLenum;\n    readonly RENDERER: GLenum;\n    readonly REPEAT: GLenum;\n    readonly REPLACE: GLenum;\n    readonly RGB: GLenum;\n    readonly RGB565: GLenum;\n    readonly RGB5_A1: GLenum;\n    readonly RGBA: GLenum;\n    readonly RGBA4: GLenum;\n    readonly SAMPLER_2D: GLenum;\n    readonly SAMPLER_CUBE: GLenum;\n    readonly SAMPLES: GLenum;\n    readonly SAMPLE_ALPHA_TO_COVERAGE: GLenum;\n    readonly SAMPLE_BUFFERS: GLenum;\n    readonly SAMPLE_COVERAGE: GLenum;\n    readonly SAMPLE_COVERAGE_INVERT: GLenum;\n    readonly SAMPLE_COVERAGE_VALUE: GLenum;\n    readonly SCISSOR_BOX: GLenum;\n    readonly SCISSOR_TEST: GLenum;\n    readonly SHADER_TYPE: GLenum;\n    readonly SHADING_LANGUAGE_VERSION: GLenum;\n    readonly SHORT: GLenum;\n    readonly SRC_ALPHA: GLenum;\n    readonly SRC_ALPHA_SATURATE: GLenum;\n    readonly SRC_COLOR: GLenum;\n    readonly STATIC_DRAW: GLenum;\n    readonly STENCIL_ATTACHMENT: GLenum;\n    readonly STENCIL_BACK_FAIL: GLenum;\n    readonly STENCIL_BACK_FUNC: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_BACK_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_BACK_REF: GLenum;\n    readonly STENCIL_BACK_VALUE_MASK: GLenum;\n    readonly STENCIL_BACK_WRITEMASK: GLenum;\n    readonly STENCIL_BITS: GLenum;\n    readonly STENCIL_BUFFER_BIT: GLenum;\n    readonly STENCIL_CLEAR_VALUE: GLenum;\n    readonly STENCIL_FAIL: GLenum;\n    readonly STENCIL_FUNC: GLenum;\n    readonly STENCIL_INDEX8: GLenum;\n    readonly STENCIL_PASS_DEPTH_FAIL: GLenum;\n    readonly STENCIL_PASS_DEPTH_PASS: GLenum;\n    readonly STENCIL_REF: GLenum;\n    readonly STENCIL_TEST: GLenum;\n    readonly STENCIL_VALUE_MASK: GLenum;\n    readonly STENCIL_WRITEMASK: GLenum;\n    readonly STREAM_DRAW: GLenum;\n    readonly SUBPIXEL_BITS: GLenum;\n    readonly TEXTURE: GLenum;\n    readonly TEXTURE0: GLenum;\n    readonly TEXTURE1: GLenum;\n    readonly TEXTURE10: GLenum;\n    readonly TEXTURE11: GLenum;\n    readonly TEXTURE12: GLenum;\n    readonly TEXTURE13: GLenum;\n    readonly TEXTURE14: GLenum;\n    readonly TEXTURE15: GLenum;\n    readonly TEXTURE16: GLenum;\n    readonly TEXTURE17: GLenum;\n    readonly TEXTURE18: GLenum;\n    readonly TEXTURE19: GLenum;\n    readonly TEXTURE2: GLenum;\n    readonly TEXTURE20: GLenum;\n    readonly TEXTURE21: GLenum;\n    readonly TEXTURE22: GLenum;\n    readonly TEXTURE23: GLenum;\n    readonly TEXTURE24: GLenum;\n    readonly TEXTURE25: GLenum;\n    readonly TEXTURE26: GLenum;\n    readonly TEXTURE27: GLenum;\n    readonly TEXTURE28: GLenum;\n    readonly TEXTURE29: GLenum;\n    readonly TEXTURE3: GLenum;\n    readonly TEXTURE30: GLenum;\n    readonly TEXTURE31: GLenum;\n    readonly TEXTURE4: GLenum;\n    readonly TEXTURE5: GLenum;\n    readonly TEXTURE6: GLenum;\n    readonly TEXTURE7: GLenum;\n    readonly TEXTURE8: GLenum;\n    readonly TEXTURE9: GLenum;\n    readonly TEXTURE_2D: GLenum;\n    readonly TEXTURE_BINDING_2D: GLenum;\n    readonly TEXTURE_BINDING_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_NEGATIVE_Z: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_X: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Y: GLenum;\n    readonly TEXTURE_CUBE_MAP_POSITIVE_Z: GLenum;\n    readonly TEXTURE_MAG_FILTER: GLenum;\n    readonly TEXTURE_MIN_FILTER: GLenum;\n    readonly TEXTURE_WRAP_S: GLenum;\n    readonly TEXTURE_WRAP_T: GLenum;\n    readonly TRIANGLES: GLenum;\n    readonly TRIANGLE_FAN: GLenum;\n    readonly TRIANGLE_STRIP: GLenum;\n    readonly UNPACK_ALIGNMENT: GLenum;\n    readonly UNPACK_COLORSPACE_CONVERSION_WEBGL: GLenum;\n    readonly UNPACK_FLIP_Y_WEBGL: GLenum;\n    readonly UNPACK_PREMULTIPLY_ALPHA_WEBGL: GLenum;\n    readonly UNSIGNED_BYTE: GLenum;\n    readonly UNSIGNED_INT: GLenum;\n    readonly UNSIGNED_SHORT: GLenum;\n    readonly UNSIGNED_SHORT_4_4_4_4: GLenum;\n    readonly UNSIGNED_SHORT_5_5_5_1: GLenum;\n    readonly UNSIGNED_SHORT_5_6_5: GLenum;\n    readonly VALIDATE_STATUS: GLenum;\n    readonly VENDOR: GLenum;\n    readonly VERSION: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_BUFFER_BINDING: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_ENABLED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_NORMALIZED: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_POINTER: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_SIZE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_STRIDE: GLenum;\n    readonly VERTEX_ATTRIB_ARRAY_TYPE: GLenum;\n    readonly VERTEX_SHADER: GLenum;\n    readonly VIEWPORT: GLenum;\n    readonly ZERO: GLenum;\n}\n\ninterface WebGLRenderingContextOverloads {\n    bufferData(target: GLenum, size: GLsizeiptr, usage: GLenum): void;\n    bufferData(target: GLenum, data: BufferSource | null, usage: GLenum): void;\n    bufferSubData(target: GLenum, offset: GLintptr, data: BufferSource): void;\n    compressedTexImage2D(target: GLenum, level: GLint, internalformat: GLenum, width: GLsizei, height: GLsizei, border: GLint, data: ArrayBufferView): void;\n    compressedTexSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, data: ArrayBufferView): void;\n    readPixels(x: GLint, y: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, width: GLsizei, height: GLsizei, border: GLint, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texImage2D(target: GLenum, level: GLint, internalformat: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, width: GLsizei, height: GLsizei, format: GLenum, type: GLenum, pixels: ArrayBufferView | null): void;\n    texSubImage2D(target: GLenum, level: GLint, xoffset: GLint, yoffset: GLint, format: GLenum, type: GLenum, source: TexImageSource): void;\n    uniform1fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform1iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform2fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform2iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform3fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform3iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniform4fv(location: WebGLUniformLocation | null, v: Float32List): void;\n    uniform4iv(location: WebGLUniformLocation | null, v: Int32List): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Float32List): void;\n}\n\ninterface WebGLSampler extends WebGLObject {\n}\n\ndeclare var WebGLSampler: {\n    prototype: WebGLSampler;\n    new(): WebGLSampler;\n};\n\n/** The WebGLShader is part of the WebGL API and can either be a vertex or a fragment shader. A WebGLProgram requires both types of shaders. */\ninterface WebGLShader extends WebGLObject {\n}\n\ndeclare var WebGLShader: {\n    prototype: WebGLShader;\n    new(): WebGLShader;\n};\n\n/** Part of the WebGL API and represents the information returned by calling the WebGLRenderingContext.getShaderPrecisionFormat() method. */\ninterface WebGLShaderPrecisionFormat {\n    readonly precision: GLint;\n    readonly rangeMax: GLint;\n    readonly rangeMin: GLint;\n}\n\ndeclare var WebGLShaderPrecisionFormat: {\n    prototype: WebGLShaderPrecisionFormat;\n    new(): WebGLShaderPrecisionFormat;\n};\n\ninterface WebGLSync extends WebGLObject {\n}\n\ndeclare var WebGLSync: {\n    prototype: WebGLSync;\n    new(): WebGLSync;\n};\n\n/** Part of the WebGL API and represents an opaque texture object providing storage and state for texturing operations. */\ninterface WebGLTexture extends WebGLObject {\n}\n\ndeclare var WebGLTexture: {\n    prototype: WebGLTexture;\n    new(): WebGLTexture;\n};\n\ninterface WebGLTransformFeedback extends WebGLObject {\n}\n\ndeclare var WebGLTransformFeedback: {\n    prototype: WebGLTransformFeedback;\n    new(): WebGLTransformFeedback;\n};\n\n/** Part of the WebGL API and represents the location of a uniform variable in a shader program. */\ninterface WebGLUniformLocation {\n}\n\ndeclare var WebGLUniformLocation: {\n    prototype: WebGLUniformLocation;\n    new(): WebGLUniformLocation;\n};\n\ninterface WebGLVertexArrayObject extends WebGLObject {\n}\n\ndeclare var WebGLVertexArrayObject: {\n    prototype: WebGLVertexArrayObject;\n    new(): WebGLVertexArrayObject;\n};\n\ninterface WebGLVertexArrayObjectOES extends WebGLObject {\n}\n\ninterface WebSocketEventMap {\n    "close": CloseEvent;\n    "error": Event;\n    "message": MessageEvent;\n    "open": Event;\n}\n\n/** Provides the API for creating and managing a WebSocket connection to a server, as well as for sending and receiving data on the connection. */\ninterface WebSocket extends EventTarget {\n    /**\n     * Returns a string that indicates how binary data from the WebSocket object is exposed to scripts:\n     * \n     * Can be set, to change how binary data is returned. The default is "blob".\n     */\n    binaryType: BinaryType;\n    /**\n     * Returns the number of bytes of application data (UTF-8 text and binary data) that have been queued using send() but not yet been transmitted to the network.\n     * \n     * If the WebSocket connection is closed, this attribute\'s value will only increase with each call to the send() method. (The number does not reset to zero once the connection closes.)\n     */\n    readonly bufferedAmount: number;\n    /**\n     * Returns the extensions selected by the server, if any.\n     */\n    readonly extensions: string;\n    onclose: ((this: WebSocket, ev: CloseEvent) => any) | null;\n    onerror: ((this: WebSocket, ev: Event) => any) | null;\n    onmessage: ((this: WebSocket, ev: MessageEvent) => any) | null;\n    onopen: ((this: WebSocket, ev: Event) => any) | null;\n    /**\n     * Returns the subprotocol selected by the server, if any. It can be used in conjunction with the array form of the constructor\'s second argument to perform subprotocol negotiation.\n     */\n    readonly protocol: string;\n    /**\n     * Returns the state of the WebSocket object\'s connection. It can have the values described below.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the URL that was used to establish the WebSocket connection.\n     */\n    readonly url: string;\n    /**\n     * Closes the WebSocket connection, optionally using code as the the WebSocket connection close code and reason as the the WebSocket connection close reason.\n     */\n    close(code?: number, reason?: string): void;\n    /**\n     * Transmits data using the WebSocket connection. data can be a string, a Blob, an ArrayBuffer, or an ArrayBufferView.\n     */\n    send(data: string | ArrayBufferLike | Blob | ArrayBufferView): void;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n    addEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WebSocketEventMap>(type: K, listener: (this: WebSocket, ev: WebSocketEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var WebSocket: {\n    prototype: WebSocket;\n    new(url: string, protocols?: string | string[]): WebSocket;\n    readonly CLOSED: number;\n    readonly CLOSING: number;\n    readonly CONNECTING: number;\n    readonly OPEN: number;\n};\n\n/** This ServiceWorker API interface represents the scope of a service worker client that is a document in a browser context, controlled by an active worker. The service worker client independently selects and uses a service worker for its own loading and sub-resources. */\ninterface WindowClient extends Client {\n    readonly ancestorOrigins: ReadonlyArray<string>;\n    readonly focused: boolean;\n    readonly visibilityState: VisibilityState;\n    focus(): Promise<WindowClient>;\n    navigate(url: string): Promise<WindowClient | null>;\n}\n\ndeclare var WindowClient: {\n    prototype: WindowClient;\n    new(): WindowClient;\n};\n\ninterface WindowOrWorkerGlobalScope {\n    readonly caches: CacheStorage;\n    readonly crypto: Crypto;\n    readonly indexedDB: IDBFactory;\n    readonly isSecureContext: boolean;\n    readonly origin: string;\n    readonly performance: Performance;\n    atob(data: string): string;\n    btoa(data: string): string;\n    clearInterval(handle?: number): void;\n    clearTimeout(handle?: number): void;\n    createImageBitmap(image: ImageBitmapSource, options?: ImageBitmapOptions): Promise<ImageBitmap>;\n    createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number, options?: ImageBitmapOptions): Promise<ImageBitmap>;\n    fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\n    queueMicrotask(callback: VoidFunction): void;\n    setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n    setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\n}\n\ninterface WorkerEventMap extends AbstractWorkerEventMap {\n    "message": MessageEvent;\n    "messageerror": MessageEvent;\n}\n\n/** This Web Workers API interface represents a background task that can be easily created and can send messages back to its creator. Creating a worker is as simple as calling the Worker() constructor and specifying a script to be run in the worker thread. */\ninterface Worker extends EventTarget, AbstractWorker {\n    onmessage: ((this: Worker, ev: MessageEvent) => any) | null;\n    onmessageerror: ((this: Worker, ev: MessageEvent) => any) | null;\n    /**\n     * Clones message and transmits it to worker\'s global environment. transfer can be passed as a list of objects that are to be transferred rather than cloned.\n     */\n    postMessage(message: any, transfer: Transferable[]): void;\n    postMessage(message: any, options?: PostMessageOptions): void;\n    /**\n     * Aborts worker\'s associated global environment.\n     */\n    terminate(): void;\n    addEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WorkerEventMap>(type: K, listener: (this: Worker, ev: WorkerEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var Worker: {\n    prototype: Worker;\n    new(stringUrl: string | URL, options?: WorkerOptions): Worker;\n};\n\ninterface WorkerGlobalScopeEventMap {\n    "error": ErrorEvent;\n    "languagechange": Event;\n    "offline": Event;\n    "online": Event;\n    "rejectionhandled": PromiseRejectionEvent;\n    "unhandledrejection": PromiseRejectionEvent;\n}\n\n/** This Web Workers API interface is an interface representing the scope of any worker. Workers have no browsing context; this scope contains the information usually conveyed by Window objects — in this case event handlers, the console or the associated WorkerNavigator object. Each WorkerGlobalScope has its own event loop. */\ninterface WorkerGlobalScope extends EventTarget, WindowOrWorkerGlobalScope {\n    /**\n     * Returns workerGlobal\'s WorkerLocation object.\n     */\n    readonly location: WorkerLocation;\n    readonly navigator: WorkerNavigator;\n    onerror: ((this: WorkerGlobalScope, ev: ErrorEvent) => any) | null;\n    onlanguagechange: ((this: WorkerGlobalScope, ev: Event) => any) | null;\n    onoffline: ((this: WorkerGlobalScope, ev: Event) => any) | null;\n    ononline: ((this: WorkerGlobalScope, ev: Event) => any) | null;\n    onrejectionhandled: ((this: WorkerGlobalScope, ev: PromiseRejectionEvent) => any) | null;\n    onunhandledrejection: ((this: WorkerGlobalScope, ev: PromiseRejectionEvent) => any) | null;\n    /**\n     * Returns workerGlobal.\n     */\n    readonly self: WorkerGlobalScope & typeof globalThis;\n    /**\n     * Fetches each URL in urls, executes them one-by-one in the order they are passed, and then returns (or throws if something went amiss).\n     */\n    importScripts(...urls: string[]): void;\n    addEventListener<K extends keyof WorkerGlobalScopeEventMap>(type: K, listener: (this: WorkerGlobalScope, ev: WorkerGlobalScopeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof WorkerGlobalScopeEventMap>(type: K, listener: (this: WorkerGlobalScope, ev: WorkerGlobalScopeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var WorkerGlobalScope: {\n    prototype: WorkerGlobalScope;\n    new(): WorkerGlobalScope;\n};\n\n/** The absolute location of the script executed by the Worker. Such an object is initialized for each worker and is available via the WorkerGlobalScope.location property obtained by calling self.location. */\ninterface WorkerLocation {\n    readonly hash: string;\n    readonly host: string;\n    readonly hostname: string;\n    readonly href: string;\n    toString(): string;\n    readonly origin: string;\n    readonly pathname: string;\n    readonly port: string;\n    readonly protocol: string;\n    readonly search: string;\n}\n\ndeclare var WorkerLocation: {\n    prototype: WorkerLocation;\n    new(): WorkerLocation;\n};\n\n/** A subset of the Navigator interface allowed to be accessed from a Worker. Such an object is initialized for each worker and is available via the WorkerGlobalScope.navigator property obtained by calling window.self.navigator. */\ninterface WorkerNavigator extends NavigatorConcurrentHardware, NavigatorID, NavigatorLanguage, NavigatorOnLine, NavigatorStorage {\n    readonly permissions: Permissions;\n    readonly serviceWorker: ServiceWorkerContainer;\n}\n\ndeclare var WorkerNavigator: {\n    prototype: WorkerNavigator;\n    new(): WorkerNavigator;\n};\n\n/** This Streams API interface provides a standard abstraction for writing streaming data to a destination, known as a sink. This object comes with built-in backpressure and queuing. */\ninterface WritableStream<W = any> {\n    readonly locked: boolean;\n    abort(reason?: any): Promise<void>;\n    getWriter(): WritableStreamDefaultWriter<W>;\n}\n\ndeclare var WritableStream: {\n    prototype: WritableStream;\n    new<W = any>(underlyingSink?: UnderlyingSink<W>, strategy?: QueuingStrategy<W>): WritableStream<W>;\n};\n\n/** This Streams API interface represents a controller allowing control of a WritableStream\'s state. When constructing a WritableStream, the underlying sink is given a corresponding WritableStreamDefaultController instance to manipulate. */\ninterface WritableStreamDefaultController {\n    error(error?: any): void;\n}\n\ndeclare var WritableStreamDefaultController: {\n    prototype: WritableStreamDefaultController;\n    new(): WritableStreamDefaultController;\n};\n\n/** This Streams API interface is the object returned by WritableStream.getWriter() and once created locks the < writer to the WritableStream ensuring that no other streams can write to the underlying sink. */\ninterface WritableStreamDefaultWriter<W = any> {\n    readonly closed: Promise<void>;\n    readonly desiredSize: number | null;\n    readonly ready: Promise<void>;\n    abort(reason?: any): Promise<void>;\n    close(): Promise<void>;\n    releaseLock(): void;\n    write(chunk: W): Promise<void>;\n}\n\ndeclare var WritableStreamDefaultWriter: {\n    prototype: WritableStreamDefaultWriter;\n    new(): WritableStreamDefaultWriter;\n};\n\ninterface XMLHttpRequestEventMap extends XMLHttpRequestEventTargetEventMap {\n    "readystatechange": Event;\n}\n\n/** Use XMLHttpRequest (XHR) objects to interact with servers. You can retrieve data from a URL without having to do a full page refresh. This enables a Web page to update just part of a page without disrupting what the user is doing. */\ninterface XMLHttpRequest extends XMLHttpRequestEventTarget {\n    onreadystatechange: ((this: XMLHttpRequest, ev: Event) => any) | null;\n    /**\n     * Returns client\'s state.\n     */\n    readonly readyState: number;\n    /**\n     * Returns the response\'s body.\n     */\n    readonly response: any;\n    /**\n     * Returns the text response.\n     * \n     * Throws an "InvalidStateError" DOMException if responseType is not the empty string or "text".\n     */\n    readonly responseText: string;\n    /**\n     * Returns the response type.\n     * \n     * Can be set to change the response type. Values are: the empty string (default), "arraybuffer", "blob", "document", "json", and "text".\n     * \n     * When set: setting to "document" is ignored if current global object is not a Window object.\n     * \n     * When set: throws an "InvalidStateError" DOMException if state is loading or done.\n     * \n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    responseType: XMLHttpRequestResponseType;\n    readonly responseURL: string;\n    readonly status: number;\n    readonly statusText: string;\n    /**\n     * Can be set to a time in milliseconds. When set to a non-zero value will cause fetching to terminate after the given time has passed. When the time has passed, the request has not yet completed, and the synchronous flag is unset, a timeout event will then be dispatched, or a "TimeoutError" DOMException will be thrown otherwise (for the send() method).\n     * \n     * When set: throws an "InvalidAccessError" DOMException if the synchronous flag is set and current global object is a Window object.\n     */\n    timeout: number;\n    /**\n     * Returns the associated XMLHttpRequestUpload object. It can be used to gather transmission information when data is transferred to a server.\n     */\n    readonly upload: XMLHttpRequestUpload;\n    /**\n     * True when credentials are to be included in a cross-origin request. False when they are to be excluded in a cross-origin request and when cookies are to be ignored in its response. Initially false.\n     * \n     * When set: throws an "InvalidStateError" DOMException if state is not unsent or opened, or if the send() flag is set.\n     */\n    withCredentials: boolean;\n    /**\n     * Cancels any network activity.\n     */\n    abort(): void;\n    getAllResponseHeaders(): string;\n    getResponseHeader(name: string): string | null;\n    /**\n     * Sets the request method, request URL, and synchronous flag.\n     * \n     * Throws a "SyntaxError" DOMException if either method is not a valid HTTP method or url cannot be parsed.\n     * \n     * Throws a "SecurityError" DOMException if method is a case-insensitive match for `CONNECT`, `TRACE`, or `TRACK`.\n     * \n     * Throws an "InvalidAccessError" DOMException if async is false, current global object is a Window object, and the timeout attribute is not zero or the responseType attribute is not the empty string.\n     */\n    open(method: string, url: string): void;\n    open(method: string, url: string, async: boolean, username?: string | null, password?: string | null): void;\n    /**\n     * Acts as if the `Content-Type` header value for response is mime. (It does not actually change the header though.)\n     * \n     * Throws an "InvalidStateError" DOMException if state is loading or done.\n     */\n    overrideMimeType(mime: string): void;\n    /**\n     * Initiates the request. The body argument provides the request body, if any, and is ignored if the request method is GET or HEAD.\n     * \n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     */\n    send(body?: BodyInit | null): void;\n    /**\n     * Combines a header in author request headers.\n     * \n     * Throws an "InvalidStateError" DOMException if either state is not opened or the send() flag is set.\n     * \n     * Throws a "SyntaxError" DOMException if name is not a header name or if value is not a header value.\n     */\n    setRequestHeader(name: string, value: string): void;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n    addEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequest: {\n    prototype: XMLHttpRequest;\n    new(): XMLHttpRequest;\n    readonly DONE: number;\n    readonly HEADERS_RECEIVED: number;\n    readonly LOADING: number;\n    readonly OPENED: number;\n    readonly UNSENT: number;\n};\n\ninterface XMLHttpRequestEventTargetEventMap {\n    "abort": ProgressEvent<XMLHttpRequestEventTarget>;\n    "error": ProgressEvent<XMLHttpRequestEventTarget>;\n    "load": ProgressEvent<XMLHttpRequestEventTarget>;\n    "loadend": ProgressEvent<XMLHttpRequestEventTarget>;\n    "loadstart": ProgressEvent<XMLHttpRequestEventTarget>;\n    "progress": ProgressEvent<XMLHttpRequestEventTarget>;\n    "timeout": ProgressEvent<XMLHttpRequestEventTarget>;\n}\n\ninterface XMLHttpRequestEventTarget extends EventTarget {\n    onabort: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onerror: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onload: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadend: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onloadstart: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    onprogress: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    ontimeout: ((this: XMLHttpRequest, ev: ProgressEvent) => any) | null;\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestEventTarget, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestEventTarget: {\n    prototype: XMLHttpRequestEventTarget;\n    new(): XMLHttpRequestEventTarget;\n};\n\ninterface XMLHttpRequestUpload extends XMLHttpRequestEventTarget {\n    addEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\n    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\n    removeEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\n    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\n}\n\ndeclare var XMLHttpRequestUpload: {\n    prototype: XMLHttpRequestUpload;\n    new(): XMLHttpRequestUpload;\n};\n\ndeclare type EventListenerOrEventListenerObject = EventListener | EventListenerObject;\n\ninterface Console {\n    memory: any;\n    assert(condition?: boolean, ...data: any[]): void;\n    clear(): void;\n    count(label?: string): void;\n    countReset(label?: string): void;\n    debug(...data: any[]): void;\n    dir(item?: any, options?: any): void;\n    dirxml(...data: any[]): void;\n    error(...data: any[]): void;\n    exception(message?: string, ...optionalParams: any[]): void;\n    group(...data: any[]): void;\n    groupCollapsed(...data: any[]): void;\n    groupEnd(): void;\n    info(...data: any[]): void;\n    log(...data: any[]): void;\n    table(tabularData?: any, properties?: string[]): void;\n    time(label?: string): void;\n    timeEnd(label?: string): void;\n    timeLog(label?: string, ...data: any[]): void;\n    timeStamp(label?: string): void;\n    trace(...data: any[]): void;\n    warn(...data: any[]): void;\n}\n\ndeclare var console: Console;\n\ndeclare namespace WebAssembly {\n    interface Global {\n        value: any;\n        valueOf(): any;\n    }\n    \n    var Global: {\n        prototype: Global;\n        new(descriptor: GlobalDescriptor, v?: any): Global;\n    };\n    \n    interface Instance {\n        readonly exports: Exports;\n    }\n    \n    var Instance: {\n        prototype: Instance;\n        new(module: Module, importObject?: Imports): Instance;\n    };\n    \n    interface Memory {\n        readonly buffer: ArrayBuffer;\n        grow(delta: number): number;\n    }\n    \n    var Memory: {\n        prototype: Memory;\n        new(descriptor: MemoryDescriptor): Memory;\n    };\n    \n    interface Module {\n    }\n    \n    var Module: {\n        prototype: Module;\n        new(bytes: BufferSource): Module;\n        customSections(moduleObject: Module, sectionName: string): ArrayBuffer[];\n        exports(moduleObject: Module): ModuleExportDescriptor[];\n        imports(moduleObject: Module): ModuleImportDescriptor[];\n    };\n    \n    interface Table {\n        readonly length: number;\n        get(index: number): Function | null;\n        grow(delta: number): number;\n        set(index: number, value: Function | null): void;\n    }\n    \n    var Table: {\n        prototype: Table;\n        new(descriptor: TableDescriptor): Table;\n    };\n    \n    interface GlobalDescriptor {\n        mutable?: boolean;\n        value: ValueType;\n    }\n    \n    interface MemoryDescriptor {\n        initial: number;\n        maximum?: number;\n    }\n    \n    interface ModuleExportDescriptor {\n        kind: ImportExportKind;\n        name: string;\n    }\n    \n    interface ModuleImportDescriptor {\n        kind: ImportExportKind;\n        module: string;\n        name: string;\n    }\n    \n    interface TableDescriptor {\n        element: TableKind;\n        initial: number;\n        maximum?: number;\n    }\n    \n    interface WebAssemblyInstantiatedSource {\n        instance: Instance;\n        module: Module;\n    }\n    \n    type ImportExportKind = "function" | "global" | "memory" | "table";\n    type TableKind = "anyfunc";\n    type ValueType = "f32" | "f64" | "i32" | "i64";\n    type ExportValue = Function | Global | Memory | Table;\n    type Exports = Record<string, ExportValue>;\n    type ImportValue = ExportValue | number;\n    type ModuleImports = Record<string, ImportValue>;\n    type Imports = Record<string, ModuleImports>;\n    function compile(bytes: BufferSource): Promise<Module>;\n    function compileStreaming(source: Response | Promise<Response>): Promise<Module>;\n    function instantiate(bytes: BufferSource, importObject?: Imports): Promise<WebAssemblyInstantiatedSource>;\n    function instantiate(moduleObject: Module, importObject?: Imports): Promise<Instance>;\n    function instantiateStreaming(response: Response | PromiseLike<Response>, importObject?: Imports): Promise<WebAssemblyInstantiatedSource>;\n    function validate(bytes: BufferSource): boolean;\n}\n\ninterface FrameRequestCallback {\n    (time: number): void;\n}\n\ninterface OnErrorEventHandlerNonNull {\n    (event: Event | string, source?: string, lineno?: number, colno?: number, error?: Error): any;\n}\n\ninterface PerformanceObserverCallback {\n    (entries: PerformanceObserverEntryList, observer: PerformanceObserver): void;\n}\n\ninterface QueuingStrategySizeCallback<T = any> {\n    (chunk: T): number;\n}\n\ninterface ReadableByteStreamControllerCallback {\n    (controller: ReadableByteStreamController): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamDefaultControllerCallback<R> {\n    (controller: ReadableStreamDefaultController<R>): void | PromiseLike<void>;\n}\n\ninterface ReadableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerCallback<O> {\n    (controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface TransformStreamDefaultControllerTransformCallback<I, O> {\n    (chunk: I, controller: TransformStreamDefaultController<O>): void | PromiseLike<void>;\n}\n\ninterface VoidFunction {\n    (): void;\n}\n\ninterface WritableStreamDefaultControllerCloseCallback {\n    (): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerStartCallback {\n    (controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamDefaultControllerWriteCallback<W> {\n    (chunk: W, controller: WritableStreamDefaultController): void | PromiseLike<void>;\n}\n\ninterface WritableStreamErrorCallback {\n    (reason: any): void | PromiseLike<void>;\n}\n\n/**\n * Returns dedicatedWorkerGlobal\'s name, i.e. the value given to the Worker constructor. Primarily useful for debugging.\n */\ndeclare var name: string;\ndeclare var onmessage: ((this: DedicatedWorkerGlobalScope, ev: MessageEvent) => any) | null;\ndeclare var onmessageerror: ((this: DedicatedWorkerGlobalScope, ev: MessageEvent) => any) | null;\n/**\n * Aborts dedicatedWorkerGlobal.\n */\ndeclare function close(): void;\n/**\n * Clones message and transmits it to the Worker object associated with dedicatedWorkerGlobal. transfer can be passed as a list of objects that are to be transferred rather than cloned.\n */\ndeclare function postMessage(message: any, transfer: Transferable[]): void;\ndeclare function postMessage(message: any, options?: PostMessageOptions): void;\n/**\n * Dispatches a synthetic event event to target and returns true if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n */\ndeclare function dispatchEvent(event: Event): boolean;\n/**\n * Returns workerGlobal\'s WorkerLocation object.\n */\ndeclare var location: WorkerLocation;\ndeclare var navigator: WorkerNavigator;\ndeclare var onerror: ((this: DedicatedWorkerGlobalScope, ev: ErrorEvent) => any) | null;\ndeclare var onlanguagechange: ((this: DedicatedWorkerGlobalScope, ev: Event) => any) | null;\ndeclare var onoffline: ((this: DedicatedWorkerGlobalScope, ev: Event) => any) | null;\ndeclare var ononline: ((this: DedicatedWorkerGlobalScope, ev: Event) => any) | null;\ndeclare var onrejectionhandled: ((this: DedicatedWorkerGlobalScope, ev: PromiseRejectionEvent) => any) | null;\ndeclare var onunhandledrejection: ((this: DedicatedWorkerGlobalScope, ev: PromiseRejectionEvent) => any) | null;\n/**\n * Returns workerGlobal.\n */\ndeclare var self: WorkerGlobalScope & typeof globalThis;\n/**\n * Fetches each URL in urls, executes them one-by-one in the order they are passed, and then returns (or throws if something went amiss).\n */\ndeclare function importScripts(...urls: string[]): void;\n/**\n * Dispatches a synthetic event event to target and returns true if either event\'s cancelable attribute value is false or its preventDefault() method was not invoked, and false otherwise.\n */\ndeclare function dispatchEvent(event: Event): boolean;\ndeclare var caches: CacheStorage;\ndeclare var crypto: Crypto;\ndeclare var indexedDB: IDBFactory;\ndeclare var isSecureContext: boolean;\ndeclare var origin: string;\ndeclare var performance: Performance;\ndeclare function atob(data: string): string;\ndeclare function btoa(data: string): string;\ndeclare function clearInterval(handle?: number): void;\ndeclare function clearTimeout(handle?: number): void;\ndeclare function createImageBitmap(image: ImageBitmapSource, options?: ImageBitmapOptions): Promise<ImageBitmap>;\ndeclare function createImageBitmap(image: ImageBitmapSource, sx: number, sy: number, sw: number, sh: number, options?: ImageBitmapOptions): Promise<ImageBitmap>;\ndeclare function fetch(input: RequestInfo, init?: RequestInit): Promise<Response>;\ndeclare function queueMicrotask(callback: VoidFunction): void;\ndeclare function setInterval(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare function setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): number;\ndeclare function cancelAnimationFrame(handle: number): void;\ndeclare function requestAnimationFrame(callback: FrameRequestCallback): number;\ndeclare function addEventListener<K extends keyof DedicatedWorkerGlobalScopeEventMap>(type: K, listener: (this: DedicatedWorkerGlobalScope, ev: DedicatedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;\ndeclare function addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;\ndeclare function removeEventListener<K extends keyof DedicatedWorkerGlobalScopeEventMap>(type: K, listener: (this: DedicatedWorkerGlobalScope, ev: DedicatedWorkerGlobalScopeEventMap[K]) => any, options?: boolean | EventListenerOptions): void;\ndeclare function removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;\ntype HeadersInit = Headers | string[][] | Record<string, string>;\ntype BodyInit = Blob | BufferSource | FormData | URLSearchParams | ReadableStream<Uint8Array> | string;\ntype RequestInfo = Request | string;\ntype BlobPart = BufferSource | Blob | string;\ntype DOMHighResTimeStamp = number;\ntype CanvasImageSource = ImageBitmap | OffscreenCanvas;\ntype OffscreenRenderingContext = OffscreenCanvasRenderingContext2D | ImageBitmapRenderingContext | WebGLRenderingContext | WebGL2RenderingContext;\ntype MessageEventSource = MessagePort | ServiceWorker;\ntype ImageBitmapSource = CanvasImageSource | Blob | ImageData;\ntype OnErrorEventHandler = OnErrorEventHandlerNonNull | null;\ntype TimerHandler = string | Function;\ntype PerformanceEntryList = PerformanceEntry[];\ntype PushMessageDataInit = BufferSource | string;\ntype ReadableStreamReadResult<T> = ReadableStreamReadValueResult<T> | ReadableStreamReadDoneResult<T>;\ntype VibratePattern = number | number[];\ntype AlgorithmIdentifier = string | Algorithm;\ntype HashAlgorithmIdentifier = AlgorithmIdentifier;\ntype BigInteger = Uint8Array;\ntype NamedCurve = string;\ntype GLenum = number;\ntype GLboolean = boolean;\ntype GLbitfield = number;\ntype GLint = number;\ntype GLsizei = number;\ntype GLintptr = number;\ntype GLsizeiptr = number;\ntype GLuint = number;\ntype GLfloat = number;\ntype GLclampf = number;\ntype TexImageSource = ImageBitmap | ImageData | OffscreenCanvas;\ntype Float32List = Float32Array | GLfloat[];\ntype Int32List = Int32Array | GLint[];\ntype GLint64 = number;\ntype GLuint64 = number;\ntype Uint32List = Uint32Array | GLuint[];\ntype BufferSource = ArrayBufferView | ArrayBuffer;\ntype DOMTimeStamp = number;\ntype FormDataEntryValue = File | string;\ntype IDBValidKey = number | string | Date | BufferSource | IDBArrayKey;\ntype Transferable = ArrayBuffer | MessagePort | ImageBitmap | OffscreenCanvas;\ntype BinaryType = "arraybuffer" | "blob";\ntype CanvasDirection = "inherit" | "ltr" | "rtl";\ntype CanvasFillRule = "evenodd" | "nonzero";\ntype CanvasLineCap = "butt" | "round" | "square";\ntype CanvasLineJoin = "bevel" | "miter" | "round";\ntype CanvasTextAlign = "center" | "end" | "left" | "right" | "start";\ntype CanvasTextBaseline = "alphabetic" | "bottom" | "hanging" | "ideographic" | "middle" | "top";\ntype ClientTypes = "all" | "sharedworker" | "window" | "worker";\ntype ColorSpaceConversion = "default" | "none";\ntype EndingType = "native" | "transparent";\ntype FrameType = "auxiliary" | "nested" | "none" | "top-level";\ntype IDBCursorDirection = "next" | "nextunique" | "prev" | "prevunique";\ntype IDBRequestReadyState = "done" | "pending";\ntype IDBTransactionMode = "readonly" | "readwrite" | "versionchange";\ntype ImageOrientation = "flipY" | "none";\ntype ImageSmoothingQuality = "high" | "low" | "medium";\ntype KeyFormat = "jwk" | "pkcs8" | "raw" | "spki";\ntype KeyType = "private" | "public" | "secret";\ntype KeyUsage = "decrypt" | "deriveBits" | "deriveKey" | "encrypt" | "sign" | "unwrapKey" | "verify" | "wrapKey";\ntype NotificationDirection = "auto" | "ltr" | "rtl";\ntype NotificationPermission = "default" | "denied" | "granted";\ntype OffscreenRenderingContextId = "2d" | "bitmaprenderer" | "webgl" | "webgl2";\ntype PermissionName = "accelerometer" | "ambient-light-sensor" | "background-sync" | "bluetooth" | "camera" | "clipboard" | "device-info" | "geolocation" | "gyroscope" | "magnetometer" | "microphone" | "midi" | "notifications" | "persistent-storage" | "push" | "speaker";\ntype PermissionState = "denied" | "granted" | "prompt";\ntype PremultiplyAlpha = "default" | "none" | "premultiply";\ntype PushEncryptionKeyName = "auth" | "p256dh";\ntype PushPermissionState = "denied" | "granted" | "prompt";\ntype ReferrerPolicy = "" | "no-referrer" | "no-referrer-when-downgrade" | "origin" | "origin-when-cross-origin" | "same-origin" | "strict-origin" | "strict-origin-when-cross-origin" | "unsafe-url";\ntype RequestCache = "default" | "force-cache" | "no-cache" | "no-store" | "only-if-cached" | "reload";\ntype RequestCredentials = "include" | "omit" | "same-origin";\ntype RequestDestination = "" | "audio" | "audioworklet" | "document" | "embed" | "font" | "image" | "manifest" | "object" | "paintworklet" | "report" | "script" | "sharedworker" | "style" | "track" | "video" | "worker" | "xslt";\ntype RequestMode = "cors" | "navigate" | "no-cors" | "same-origin";\ntype RequestRedirect = "error" | "follow" | "manual";\ntype ResizeQuality = "high" | "low" | "medium" | "pixelated";\ntype ResponseType = "basic" | "cors" | "default" | "error" | "opaque" | "opaqueredirect";\ntype ServiceWorkerState = "activated" | "activating" | "installed" | "installing" | "parsed" | "redundant";\ntype ServiceWorkerUpdateViaCache = "all" | "imports" | "none";\ntype VisibilityState = "hidden" | "visible";\ntype WebGLPowerPreference = "default" | "high-performance" | "low-power";\ntype WorkerType = "classic" | "module";\ntype XMLHttpRequestResponseType = "" | "arraybuffer" | "blob" | "document" | "json" | "text";\n',n.libFileMap["lib.webworker.importscripts.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n\n/////////////////////////////\n/// WorkerGlobalScope APIs\n/////////////////////////////\n// These are only available in a Web Worker\ndeclare function importScripts(...urls: string[]): void;\n',n.libFileMap["lib.webworker.iterable.d.ts"]='/*! *****************************************************************************\nCopyright (c) Microsoft Corporation. All rights reserved.\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at http://www.apache.org/licenses/LICENSE-2.0\n\nTHIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY\nKIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED\nWARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,\nMERCHANTABLITY OR NON-INFRINGEMENT.\n\nSee the Apache Version 2.0 License for specific language governing permissions\nand limitations under the License.\n***************************************************************************** */\n\n\n\n/// <reference no-default-lib="true"/>\n\n\n/////////////////////////////\n/// Worker Iterable APIs\n/////////////////////////////\n\ninterface Cache {\n    addAll(requests: Iterable<RequestInfo>): Promise<void>;\n}\n\ninterface CanvasPathDrawingStyles {\n    setLineDash(segments: Iterable<number>): void;\n}\n\ninterface DOMStringList {\n    [Symbol.iterator](): IterableIterator<string>;\n}\n\ninterface FileList {\n    [Symbol.iterator](): IterableIterator<File>;\n}\n\ninterface FormData {\n    [Symbol.iterator](): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the list.\n     */\n    entries(): IterableIterator<[string, FormDataEntryValue]>;\n    /**\n     * Returns a list of keys in the list.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the list.\n     */\n    values(): IterableIterator<FormDataEntryValue>;\n}\n\ninterface Headers {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all key/value pairs contained in this object.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns an iterator allowing to go through all keys of the key/value pairs contained in this object.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns an iterator allowing to go through all values of the key/value pairs contained in this object.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface IDBDatabase {\n    /**\n     * Returns a new transaction with the given mode ("readonly" or "readwrite") and scope which can be a single object store name or an array of names.\n     */\n    transaction(storeNames: string | Iterable<string>, mode?: IDBTransactionMode): IDBTransaction;\n}\n\ninterface IDBObjectStore {\n    /**\n     * Creates a new index in store with the given name, keyPath and options and returns a new IDBIndex. If the keyPath and options define constraints that cannot be satisfied with the data already in store the upgrade transaction will abort with a "ConstraintError" DOMException.\n     * \n     * Throws an "InvalidStateError" DOMException if not called within an upgrade transaction.\n     */\n    createIndex(name: string, keyPath: string | Iterable<string>, options?: IDBIndexParameters): IDBIndex;\n}\n\ninterface URLSearchParams {\n    [Symbol.iterator](): IterableIterator<[string, string]>;\n    /**\n     * Returns an array of key, value pairs for every entry in the search params.\n     */\n    entries(): IterableIterator<[string, string]>;\n    /**\n     * Returns a list of keys in the search params.\n     */\n    keys(): IterableIterator<string>;\n    /**\n     * Returns a list of values in the search params.\n     */\n    values(): IterableIterator<string>;\n}\n\ninterface WEBGL_draw_buffers {\n    drawBuffersWEBGL(buffers: Iterable<GLenum>): void;\n}\n\ninterface WebGL2RenderingContextBase {\n    clearBufferfv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLfloat>, srcOffset?: GLuint): void;\n    clearBufferiv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLint>, srcOffset?: GLuint): void;\n    clearBufferuiv(buffer: GLenum, drawbuffer: GLint, values: Iterable<GLuint>, srcOffset?: GLuint): void;\n    drawBuffers(buffers: Iterable<GLenum>): void;\n    getActiveUniforms(program: WebGLProgram, uniformIndices: Iterable<GLuint>, pname: GLenum): any;\n    getUniformIndices(program: WebGLProgram, uniformNames: Iterable<string>): Iterable<GLuint> | null;\n    invalidateFramebuffer(target: GLenum, attachments: Iterable<GLenum>): void;\n    invalidateSubFramebuffer(target: GLenum, attachments: Iterable<GLenum>, x: GLint, y: GLint, width: GLsizei, height: GLsizei): void;\n    transformFeedbackVaryings(program: WebGLProgram, varyings: Iterable<string>, bufferMode: GLenum): void;\n    uniform1uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4uiv(location: WebGLUniformLocation | null, data: Iterable<GLuint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3x4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4x3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    vertexAttribI4iv(index: GLuint, values: Iterable<GLint>): void;\n    vertexAttribI4uiv(index: GLuint, values: Iterable<GLuint>): void;\n}\n\ninterface WebGL2RenderingContextOverloads {\n    uniform1fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform1iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform2iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform3iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4fv(location: WebGLUniformLocation | null, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniform4iv(location: WebGLUniformLocation | null, data: Iterable<GLint>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, data: Iterable<GLfloat>, srcOffset?: GLuint, srcLength?: GLuint): void;\n}\n\ninterface WebGLRenderingContextBase {\n    vertexAttrib1fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib2fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib3fv(index: GLuint, values: Iterable<GLfloat>): void;\n    vertexAttrib4fv(index: GLuint, values: Iterable<GLfloat>): void;\n}\n\ninterface WebGLRenderingContextOverloads {\n    uniform1fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform1iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform2fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform2iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform3fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform3iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniform4fv(location: WebGLUniformLocation | null, v: Iterable<GLfloat>): void;\n    uniform4iv(location: WebGLUniformLocation | null, v: Iterable<GLint>): void;\n    uniformMatrix2fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n    uniformMatrix3fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n    uniformMatrix4fv(location: WebGLUniformLocation | null, transpose: GLboolean, value: Iterable<GLfloat>): void;\n}\n'}));var __awaiter=this&&this.__awaiter||function(e,n,t,r){return new(t||(t=Promise))((function(a,i){function o(e){try{c(r.next(e))}catch(e){i(e)}}function s(e){try{c(r.throw(e))}catch(e){i(e)}}function c(e){var n;e.done?a(e.value):(n=e.value,n instanceof t?n:new t((function(e){e(n)}))).then(o,s)}c((r=r.apply(e,n||[])).next())}))};__generator=this&&this.__generator||function(e,n){var t,r,a,i,o={label:0,sent:function(){if(1&a[0])throw a[1];return a[1]},trys:[],ops:[]};return i={next:s(0),throw:s(1),return:s(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function s(i){return function(s){return function(i){if(t)throw new TypeError("Generator is already executing.");for(;o;)try{if(t=1,r&&(a=2&i[0]?r.return:i[0]?r.throw||((a=r.return)&&a.call(r),0):r.next)&&!(a=a.call(r,i[1])).done)return a;switch(r=0,a&&(i=[2&i[0],a.value]),i[0]){case 0:case 1:a=i;break;case 4:return o.label++,{value:i[1],done:!1};case 5:o.label++,r=i[1],i=[0];continue;case 7:i=o.ops.pop(),o.trys.pop();continue;default:if(!(a=o.trys,(a=a.length>0&&a[a.length-1])||6!==i[0]&&2!==i[0])){o=0;continue}if(3===i[0]&&(!a||i[1]>a[0]&&i[1]<a[3])){o.label=i[1];break}if(6===i[0]&&o.label<a[1]){o.label=a[1],a=i;break}if(a&&o.label<a[2]){o.label=a[2],o.ops.push(i);break}a[2]&&o.ops.pop(),o.trys.pop();continue}i=n.call(e,o)}catch(e){i=[6,e],r=0}finally{t=a=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,s])}}};define("vs/language/typescript/tsWorker",["require","exports","./lib/typescriptServices","./lib/lib"],(function(e,n,t,r){"use strict";function a(e){return"string"==typeof e?!!/^file:\/\/\//.test(e)&&!!r.libFileMap[e.substr(8)]:0===e.path.indexOf("/lib.")&&!!r.libFileMap[e.path.slice(1)]}Object.defineProperty(n,"__esModule",{value:!0}),n.create=n.TypeScriptWorker=void 0;var i=function(){function e(e,n){this._extraLibs=Object.create(null),this._languageService=t.createLanguageService(this),this._ctx=e,this._compilerOptions=n.compilerOptions,this._extraLibs=n.extraLibs}return e.prototype.getCompilationSettings=function(){return this._compilerOptions},e.prototype.getScriptFileNames=function(){return this._ctx.getMirrorModels().map((function(e){return e.uri})).filter((function(e){return!a(e)})).map((function(e){return e.toString()})).concat(Object.keys(this._extraLibs))},e.prototype._getModel=function(e){for(var n=this._ctx.getMirrorModels(),t=0;t<n.length;t++)if(n[t].uri.toString()===e)return n[t];return null},e.prototype.getScriptVersion=function(e){var n=this._getModel(e);return n?n.version.toString():this.isDefaultLibFileName(e)?"1":e in this._extraLibs?String(this._extraLibs[e].version):""},e.prototype.getScriptText=function(e){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(n){return[2,this._getScriptText(e)]}))}))},e.prototype._getScriptText=function(e){var n,t=this._getModel(e),a="lib."+e+".d.ts";if(t)n=t.getValue();else if(e in r.libFileMap)n=r.libFileMap[e];else if(a in r.libFileMap)n=r.libFileMap[a];else{if(!(e in this._extraLibs))return;n=this._extraLibs[e].content}return n},e.prototype.getScriptSnapshot=function(e){var n=this._getScriptText(e);if(void 0!==n)return{getText:function(e,t){return n.substring(e,t)},getLength:function(){return n.length},getChangeRange:function(){}}},e.prototype.getScriptKind=function(e){switch(e.substr(e.lastIndexOf(".")+1)){case"ts":return t.ScriptKind.TS;case"tsx":return t.ScriptKind.TSX;case"js":return t.ScriptKind.JS;case"jsx":return t.ScriptKind.JSX;default:return this.getCompilationSettings().allowJs?t.ScriptKind.JS:t.ScriptKind.TS}},e.prototype.getCurrentDirectory=function(){return""},e.prototype.getDefaultLibFileName=function(e){switch(e.target){case 99:var n="lib.esnext.full.d.ts";if(n in r.libFileMap||n in this._extraLibs)return n;case 7:case 6:case 5:case 4:case 3:case 2:default:var t="lib.es"+(2013+(e.target||99))+".full.d.ts";return t in r.libFileMap||t in this._extraLibs?t:"lib.es6.d.ts";case 1:case 0:return"lib.d.ts"}},e.prototype.isDefaultLibFileName=function(e){return e===this.getDefaultLibFileName(this._compilerOptions)},e.prototype.getLibFiles=function(){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(e){return[2,r.libFileMap]}))}))},e.clearFiles=function(e){return e.forEach((function(e){e.file=void 0;var n=e.relatedInformation;n&&n.forEach((function(e){return e.file=void 0}))})),e},e.prototype.getSyntacticDiagnostics=function(n){return __awaiter(this,void 0,void 0,(function(){var t;return __generator(this,(function(r){return a(n)?[2,[]]:(t=this._languageService.getSyntacticDiagnostics(n),[2,e.clearFiles(t)])}))}))},e.prototype.getSemanticDiagnostics=function(n){return __awaiter(this,void 0,void 0,(function(){var t;return __generator(this,(function(r){return a(n)?[2,[]]:(t=this._languageService.getSemanticDiagnostics(n),[2,e.clearFiles(t)])}))}))},e.prototype.getSuggestionDiagnostics=function(n){return __awaiter(this,void 0,void 0,(function(){var t;return __generator(this,(function(r){return a(n)?[2,[]]:(t=this._languageService.getSuggestionDiagnostics(n),[2,e.clearFiles(t)])}))}))},e.prototype.getCompilerOptionsDiagnostics=function(n){return __awaiter(this,void 0,void 0,(function(){var t;return __generator(this,(function(r){return a(n)?[2,[]]:(t=this._languageService.getCompilerOptionsDiagnostics(),[2,e.clearFiles(t)])}))}))},e.prototype.getCompletionsAtPosition=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,void 0]:[2,this._languageService.getCompletionsAtPosition(e,n,void 0)]}))}))},e.prototype.getCompletionEntryDetails=function(e,n,t){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(r){return[2,this._languageService.getCompletionEntryDetails(e,n,t,void 0,void 0,void 0)]}))}))},e.prototype.getSignatureHelpItems=function(e,n,t){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(r){return a(e)?[2,void 0]:[2,this._languageService.getSignatureHelpItems(e,n,t)]}))}))},e.prototype.getQuickInfoAtPosition=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,void 0]:[2,this._languageService.getQuickInfoAtPosition(e,n)]}))}))},e.prototype.getOccurrencesAtPosition=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,void 0]:[2,this._languageService.getOccurrencesAtPosition(e,n)]}))}))},e.prototype.getDefinitionAtPosition=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,void 0]:[2,this._languageService.getDefinitionAtPosition(e,n)]}))}))},e.prototype.getReferencesAtPosition=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,void 0]:[2,this._languageService.getReferencesAtPosition(e,n)]}))}))},e.prototype.getNavigationBarItems=function(e){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(n){return a(e)?[2,[]]:[2,this._languageService.getNavigationBarItems(e)]}))}))},e.prototype.getFormattingEditsForDocument=function(e,n){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(t){return a(e)?[2,[]]:[2,this._languageService.getFormattingEditsForDocument(e,n)]}))}))},e.prototype.getFormattingEditsForRange=function(e,n,t,r){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(i){return a(e)?[2,[]]:[2,this._languageService.getFormattingEditsForRange(e,n,t,r)]}))}))},e.prototype.getFormattingEditsAfterKeystroke=function(e,n,t,r){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(i){return a(e)?[2,[]]:[2,this._languageService.getFormattingEditsAfterKeystroke(e,n,t,r)]}))}))},e.prototype.findRenameLocations=function(e,n,t,r,i){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(o){return a(e)?[2,void 0]:[2,this._languageService.findRenameLocations(e,n,t,r,i)]}))}))},e.prototype.getRenameInfo=function(e,n,t){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(r){return a(e)?[2,{canRename:!1,localizedErrorMessage:"Cannot rename in lib file"}]:[2,this._languageService.getRenameInfo(e,n,t)]}))}))},e.prototype.getEmitOutput=function(e){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(n){return a(e)?[2,{outputFiles:[],emitSkipped:!0}]:[2,this._languageService.getEmitOutput(e)]}))}))},e.prototype.getCodeFixesAtPosition=function(e,n,t,r,i){return __awaiter(this,void 0,void 0,(function(){var o;return __generator(this,(function(s){if(a(e))return[2,[]];o={};try{return[2,this._languageService.getCodeFixesAtPosition(e,n,t,r,i,o)]}catch(e){return[2,[]]}return[2]}))}))},e.prototype.updateExtraLibs=function(e){return __awaiter(this,void 0,void 0,(function(){return __generator(this,(function(n){return this._extraLibs=e,[2]}))}))},e}();n.TypeScriptWorker=i,n.create=function(e,n){var a=i;if(n.customWorkerPath)if("undefined"==typeof importScripts)console.warn("Monaco is not using webworkers for background tasks, and that is needed to support the customWorkerPath flag");else{importScripts(n.customWorkerPath);var o=self.customTSWorkerFactory;if(!o)throw new Error("The script at "+n.customWorkerPath+" does not add customTSWorkerFactory to self");a=o(i,t,r.libFileMap)}return new a(e,n)}}));